(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.ModelXModel = {}));
})(this, (function (exports) { 'use strict';

    function _mergeNamespaces(n, m) {
        m.forEach(function (e) {
            e && typeof e !== 'string' && !Array.isArray(e) && Object.keys(e).forEach(function (k) {
                if (k !== 'default' && !(k in n)) {
                    var d = Object.getOwnPropertyDescriptor(e, k);
                    Object.defineProperty(n, k, d.get ? d : {
                        enumerable: true,
                        get: function () { return e[k]; }
                    });
                }
            });
        });
        return Object.freeze(n);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const EPSILON_FLOAT32$1 = 1e-7;
    const EPSILON_FLOAT16$1 = 1e-4;
    /** Convenient class for storing tensor-related data. */
    class DataStorage {
        constructor(backend, dataMover) {
            this.backend = backend;
            this.dataMover = dataMover;
            this.data = new WeakMap();
            this.dataIdsCount = 0;
        }
        get(dataId) {
            if (!this.data.has(dataId)) {
                this.dataMover.moveData(this.backend, dataId);
            }
            return this.data.get(dataId);
        }
        set(dataId, value) {
            this.dataIdsCount++;
            this.data.set(dataId, value);
        }
        has(dataId) {
            return this.data.has(dataId);
        }
        delete(dataId) {
            this.dataIdsCount--;
            return this.data.delete(dataId);
        }
        numDataIds() {
            return this.dataIdsCount;
        }
    }
    /**
     * The interface that defines the kernels that should be implemented when
     * adding a new backend. New backends don't need to implement every one of the
     * methods, this can be done gradually (throw an error for unimplemented
     * methods).
     */
    class KernelBackend {
        refCount(dataId) {
            return notYetImplemented('refCount');
        }
        incRef(dataId) {
            return notYetImplemented('incRef');
        }
        timerAvailable() {
            return true;
        }
        time(f) {
            return notYetImplemented('time');
        }
        read(dataId) {
            return notYetImplemented('read');
        }
        readSync(dataId) {
            return notYetImplemented('readSync');
        }
        readToGPU(dataId, options) {
            return notYetImplemented('readToGPU');
        }
        numDataIds() {
            return notYetImplemented('numDataIds');
        }
        disposeData(dataId, force) {
            return notYetImplemented('disposeData');
        }
        write(values, shape, dtype) {
            return notYetImplemented('write');
        }
        move(dataId, values, shape, dtype, refCount) {
            return notYetImplemented('move');
        }
        memory() {
            return notYetImplemented('memory');
        }
        /** Returns the highest precision for floats in bits (e.g. 16 or 32) */
        floatPrecision() {
            return notYetImplemented('floatPrecision');
        }
        /** Returns the smallest representable number.  */
        epsilon() {
            return this.floatPrecision() === 32 ? EPSILON_FLOAT32$1 : EPSILON_FLOAT16$1;
        }
        dispose() {
            return notYetImplemented('dispose');
        }
    }
    function notYetImplemented(kernelName) {
        throw new Error(`'${kernelName}' not yet implemented or not found in the registry. ` +
            `This kernel may not be supported by the tfjs backend you have chosen`);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Shuffles the array in-place using Fisher-Yates algorithm.
     *
     * ```js
     * const a = [1, 2, 3, 4, 5];
     * tf.util.shuffle(a);
     * console.log(a);
     * ```
     *
     * @param array The array to shuffle in-place.
     *
     * @doc {heading: 'Util', namespace: 'util'}
     */
    // tslint:disable-next-line:no-any
    function shuffle(array) {
        let counter = array.length;
        let index = 0;
        // While there are elements in the array
        while (counter > 0) {
            // Pick a random index
            index = (Math.random() * counter) | 0;
            // Decrease counter by 1
            counter--;
            // And swap the last element with it
            swap$1(array, counter, index);
        }
    }
    /**
     * Shuffles two arrays in-place the same way using Fisher-Yates algorithm.
     *
     * ```js
     * const a = [1,2,3,4,5];
     * const b = [11,22,33,44,55];
     * tf.util.shuffleCombo(a, b);
     * console.log(a, b);
     * ```
     *
     * @param array The first array to shuffle in-place.
     * @param array2 The second array to shuffle in-place with the same permutation
     *     as the first array.
     *
     * @doc {heading: 'Util', namespace: 'util'}
     */
    function shuffleCombo(
    // tslint:disable-next-line:no-any
    array, 
    // tslint:disable-next-line:no-any
    array2) {
        if (array.length !== array2.length) {
            throw new Error(`Array sizes must match to be shuffled together ` +
                `First array length was ${array.length}` +
                `Second array length was ${array2.length}`);
        }
        let counter = array.length;
        let index = 0;
        // While there are elements in the array
        while (counter > 0) {
            // Pick a random index
            index = (Math.random() * counter) | 0;
            // Decrease counter by 1
            counter--;
            // And swap the last element of each array with it
            swap$1(array, counter, index);
            swap$1(array2, counter, index);
        }
    }
    /** Clamps a value to a specified range. */
    function clamp(min, x, max) {
        return Math.max(min, Math.min(x, max));
    }
    function nearestLargerEven(val) {
        return val % 2 === 0 ? val : val + 1;
    }
    function swap$1(object, left, right) {
        const temp = object[left];
        object[left] = object[right];
        object[right] = temp;
    }
    function sum$3(arr) {
        let sum = 0;
        for (let i = 0; i < arr.length; i++) {
            sum += arr[i];
        }
        return sum;
    }
    /**
     * Returns a sample from a uniform [a, b) distribution.
     *
     * @param a The minimum support (inclusive).
     * @param b The maximum support (exclusive).
     * @return A pseudorandom number on the half-open interval [a,b).
     */
    function randUniform(a, b) {
        const r = Math.random();
        return (b * r) + (1 - r) * a;
    }
    /** Returns the squared Euclidean distance between two vectors. */
    function distSquared(a, b) {
        let result = 0;
        for (let i = 0; i < a.length; i++) {
            const diff = Number(a[i]) - Number(b[i]);
            result += diff * diff;
        }
        return result;
    }
    /**
     * Asserts that the expression is true. Otherwise throws an error with the
     * provided message.
     *
     * ```js
     * const x = 2;
     * tf.util.assert(x === 2, 'x is not 2');
     * ```
     *
     * @param expr The expression to assert (as a boolean).
     * @param msg A function that returns the message to report when throwing an
     *     error. We use a function for performance reasons.
     *
     * @doc {heading: 'Util', namespace: 'util'}
     */
    function assert$4(expr, msg) {
        if (!expr) {
            throw new Error(typeof msg === 'string' ? msg : msg());
        }
    }
    function assertShapesMatch(shapeA, shapeB, errorMessagePrefix = '') {
        assert$4(arraysEqual(shapeA, shapeB), () => errorMessagePrefix + ` Shapes ${shapeA} and ${shapeB} must match`);
    }
    function assertNonNull(a) {
        assert$4(a != null, () => `The input to the tensor constructor must be a non-null value.`);
    }
    // NOTE: We explicitly type out what T extends instead of any so that
    // util.flatten on a nested array of number doesn't try to infer T as a
    // number[][], causing us to explicitly type util.flatten<number>().
    /**
     *  Flattens an arbitrarily nested array.
     *
     * ```js
     * const a = [[1, 2], [3, 4], [5, [6, [7]]]];
     * const flat = tf.util.flatten(a);
     * console.log(flat);
     * ```
     *
     *  @param arr The nested array to flatten.
     *  @param result The destination array which holds the elements.
     *  @param skipTypedArray If true, avoids flattening the typed arrays. Defaults
     *      to false.
     *
     * @doc {heading: 'Util', namespace: 'util'}
     */
    function flatten$3(arr, result = [], skipTypedArray = false) {
        if (result == null) {
            result = [];
        }
        if (Array.isArray(arr) || isTypedArray$1(arr) && !skipTypedArray) {
            for (let i = 0; i < arr.length; ++i) {
                flatten$3(arr[i], result, skipTypedArray);
            }
        }
        else {
            result.push(arr);
        }
        return result;
    }
    /**
     * Returns the size (number of elements) of the tensor given its shape.
     *
     * ```js
     * const shape = [3, 4, 2];
     * const size = tf.util.sizeFromShape(shape);
     * console.log(size);
     * ```
     *
     * @doc {heading: 'Util', namespace: 'util'}
     */
    function sizeFromShape(shape) {
        if (shape.length === 0) {
            // Scalar.
            return 1;
        }
        let size = shape[0];
        for (let i = 1; i < shape.length; i++) {
            size *= shape[i];
        }
        return size;
    }
    function isScalarShape(shape) {
        return shape.length === 0;
    }
    function arraysEqual(n1, n2) {
        if (n1 === n2) {
            return true;
        }
        if (n1 == null || n2 == null) {
            return false;
        }
        if (n1.length !== n2.length) {
            return false;
        }
        for (let i = 0; i < n1.length; i++) {
            if (n1[i] !== n2[i]) {
                return false;
            }
        }
        return true;
    }
    function isInt(a) {
        return a % 1 === 0;
    }
    function tanh$3(x) {
        // tslint:disable-next-line:no-any
        if (Math.tanh != null) {
            // tslint:disable-next-line:no-any
            return Math.tanh(x);
        }
        if (x === Infinity) {
            return 1;
        }
        else if (x === -Infinity) {
            return -1;
        }
        else {
            const e2x = Math.exp(2 * x);
            return (e2x - 1) / (e2x + 1);
        }
    }
    function sizeToSquarishShape(size) {
        const width = Math.ceil(Math.sqrt(size));
        return [width, Math.ceil(size / width)];
    }
    /**
     * Creates a new array with randomized indicies to a given quantity.
     *
     * ```js
     * const randomTen = tf.util.createShuffledIndices(10);
     * console.log(randomTen);
     * ```
     *
     * @param number Quantity of how many shuffled indicies to create.
     *
     * @doc {heading: 'Util', namespace: 'util'}
     */
    function createShuffledIndices(n) {
        const shuffledIndices = new Uint32Array(n);
        for (let i = 0; i < n; ++i) {
            shuffledIndices[i] = i;
        }
        shuffle(shuffledIndices);
        return shuffledIndices;
    }
    function rightPad(a, size) {
        if (size <= a.length) {
            return a;
        }
        return a + ' '.repeat(size - a.length);
    }
    function repeatedTry(checkFn, delayFn = (counter) => 0, maxCounter) {
        return new Promise((resolve, reject) => {
            let tryCount = 0;
            const tryFn = () => {
                if (checkFn()) {
                    resolve();
                    return;
                }
                tryCount++;
                const nextBackoff = delayFn(tryCount);
                if (maxCounter != null && tryCount >= maxCounter) {
                    reject();
                    return;
                }
                setTimeout(tryFn, nextBackoff);
            };
            tryFn();
        });
    }
    /**
     * Given the full size of the array and a shape that may contain -1 as the
     * implicit dimension, returns the inferred shape where -1 is replaced.
     * E.g. For shape=[2, -1, 3] and size=24, it will return [2, 4, 3].
     *
     * @param shape The shape, which may contain -1 in some dimension.
     * @param size The full size (number of elements) of the array.
     * @return The inferred shape where -1 is replaced with the inferred size.
     */
    function inferFromImplicitShape(shape, size) {
        let shapeProd = 1;
        let implicitIdx = -1;
        for (let i = 0; i < shape.length; ++i) {
            if (shape[i] >= 0) {
                shapeProd *= shape[i];
            }
            else if (shape[i] === -1) {
                if (implicitIdx !== -1) {
                    throw Error(`Shapes can only have 1 implicit size. ` +
                        `Found -1 at dim ${implicitIdx} and dim ${i}`);
                }
                implicitIdx = i;
            }
            else if (shape[i] < 0) {
                throw Error(`Shapes can not be < 0. Found ${shape[i]} at dim ${i}`);
            }
        }
        if (implicitIdx === -1) {
            if (size > 0 && size !== shapeProd) {
                throw Error(`Size(${size}) must match the product of shape ${shape}`);
            }
            return shape;
        }
        if (shapeProd === 0) {
            throw Error(`Cannot infer the missing size in [${shape}] when ` +
                `there are 0 elements`);
        }
        if (size % shapeProd !== 0) {
            throw Error(`The implicit shape can't be a fractional number. ` +
                `Got ${size} / ${shapeProd}`);
        }
        const newShape = shape.slice();
        newShape[implicitIdx] = size / shapeProd;
        return newShape;
    }
    function parseAxisParam(axis, shape) {
        const rank = shape.length;
        // Normalize input
        axis = axis == null ? shape.map((s, i) => i) : [].concat(axis);
        // Check for valid range
        assert$4(axis.every(ax => ax >= -rank && ax < rank), () => `All values in axis param must be in range [-${rank}, ${rank}) but ` +
            `got axis ${axis}`);
        // Check for only integers
        assert$4(axis.every(ax => isInt(ax)), () => `All values in axis param must be integers but ` +
            `got axis ${axis}`);
        // Handle negative axis.
        return axis.map(a => a < 0 ? rank + a : a);
    }
    /** Reduces the shape by removing all dimensions of shape 1. */
    function squeezeShape(shape, axis) {
        const newShape = [];
        const keptDims = [];
        const isEmptyArray = axis != null && Array.isArray(axis) && axis.length === 0;
        const axes = (axis == null || isEmptyArray) ?
            null :
            parseAxisParam(axis, shape).sort();
        let j = 0;
        for (let i = 0; i < shape.length; ++i) {
            if (axes != null) {
                if (axes[j] === i && shape[i] !== 1) {
                    throw new Error(`Can't squeeze axis ${i} since its dim '${shape[i]}' is not 1`);
                }
                if ((axes[j] == null || axes[j] > i) && shape[i] === 1) {
                    newShape.push(shape[i]);
                    keptDims.push(i);
                }
                if (axes[j] <= i) {
                    j++;
                }
            }
            if (shape[i] !== 1) {
                newShape.push(shape[i]);
                keptDims.push(i);
            }
        }
        return { newShape, keptDims };
    }
    function getTypedArrayFromDType(dtype, size) {
        let values = null;
        if (dtype == null || dtype === 'float32') {
            values = new Float32Array(size);
        }
        else if (dtype === 'int32') {
            values = new Int32Array(size);
        }
        else if (dtype === 'bool') {
            values = new Uint8Array(size);
        }
        else {
            throw new Error(`Unknown data type ${dtype}`);
        }
        return values;
    }
    function getArrayFromDType(dtype, size) {
        let values = null;
        if (dtype == null || dtype === 'float32') {
            values = new Float32Array(size);
        }
        else if (dtype === 'int32') {
            values = new Int32Array(size);
        }
        else if (dtype === 'bool') {
            values = new Uint8Array(size);
        }
        else if (dtype === 'string') {
            values = new Array(size);
        }
        else {
            throw new Error(`Unknown data type ${dtype}`);
        }
        return values;
    }
    function checkConversionForErrors(vals, dtype) {
        for (let i = 0; i < vals.length; i++) {
            const num = vals[i];
            if (isNaN(num) || !isFinite(num)) {
                throw Error(`A tensor of type ${dtype} being uploaded contains ${num}.`);
            }
        }
    }
    /** Returns true if the dtype is valid. */
    function isValidDtype(dtype) {
        return dtype === 'bool' || dtype === 'complex64' || dtype === 'float32' ||
            dtype === 'int32' || dtype === 'string';
    }
    /**
     * Returns true if the new type can't encode the old type without loss of
     * precision.
     */
    function hasEncodingLoss(oldType, newType) {
        if (newType === 'complex64') {
            return false;
        }
        if (newType === 'float32' && oldType !== 'complex64') {
            return false;
        }
        if (newType === 'int32' && oldType !== 'float32' && oldType !== 'complex64') {
            return false;
        }
        if (newType === 'bool' && oldType === 'bool') {
            return false;
        }
        return true;
    }
    function isTypedArray$1(a) {
        return a instanceof Float32Array || a instanceof Int32Array ||
            a instanceof Uint8Array || a instanceof Uint8ClampedArray;
    }
    function bytesPerElement(dtype) {
        if (dtype === 'float32' || dtype === 'int32') {
            return 4;
        }
        else if (dtype === 'complex64') {
            return 8;
        }
        else if (dtype === 'bool') {
            return 1;
        }
        else {
            throw new Error(`Unknown dtype ${dtype}`);
        }
    }
    /**
     * Returns the approximate number of bytes allocated in the string array - 2
     * bytes per character. Computing the exact bytes for a native string in JS is
     * not possible since it depends on the encoding of the html page that serves
     * the website.
     */
    function bytesFromStringArray(arr) {
        if (arr == null) {
            return 0;
        }
        let bytes = 0;
        arr.forEach(x => bytes += x.length);
        return bytes;
    }
    /** Returns true if the value is a string. */
    function isString$2(value) {
        return typeof value === 'string' || value instanceof String;
    }
    function isBoolean$1(value) {
        return typeof value === 'boolean';
    }
    function isNumber$3(value) {
        return typeof value === 'number';
    }
    function inferDtype(values) {
        if (Array.isArray(values)) {
            return inferDtype(values[0]);
        }
        if (values instanceof Float32Array) {
            return 'float32';
        }
        else if (values instanceof Int32Array
            || values instanceof Uint8Array
            || values instanceof Uint8ClampedArray) {
            return 'int32';
        }
        else if (isNumber$3(values)) {
            return 'float32';
        }
        else if (isString$2(values)) {
            return 'string';
        }
        else if (isBoolean$1(values)) {
            return 'bool';
        }
        return 'float32';
    }
    function isFunction$5(f) {
        return !!(f && f.constructor && f.call && f.apply);
    }
    function nearestDivisor(size, start) {
        for (let i = start; i < size; ++i) {
            if (size % i === 0) {
                return i;
            }
        }
        return size;
    }
    function computeStrides(shape) {
        const rank = shape.length;
        if (rank < 2) {
            return [];
        }
        // Last dimension has implicit stride of 1, thus having D-1 (instead of D)
        // strides.
        const strides = new Array(rank - 1);
        strides[rank - 2] = shape[rank - 1];
        for (let i = rank - 3; i >= 0; --i) {
            strides[i] = strides[i + 1] * shape[i + 1];
        }
        return strides;
    }
    function createNestedArray(offset, shape, a, isComplex = false) {
        const ret = new Array();
        if (shape.length === 1) {
            const d = shape[0] * (isComplex ? 2 : 1);
            for (let i = 0; i < d; i++) {
                ret[i] = a[offset + i];
            }
        }
        else {
            const d = shape[0];
            const rest = shape.slice(1);
            const len = rest.reduce((acc, c) => acc * c) * (isComplex ? 2 : 1);
            for (let i = 0; i < d; i++) {
                ret[i] = createNestedArray(offset + i * len, rest, a, isComplex);
            }
        }
        return ret;
    }
    // Provide a nested array of TypedArray in given shape.
    function toNestedArray(shape, a, isComplex = false) {
        if (shape.length === 0) {
            // Scalar type should return a single number.
            return a[0];
        }
        const size = shape.reduce((acc, c) => acc * c) * (isComplex ? 2 : 1);
        if (size === 0) {
            // A tensor with shape zero should be turned into empty list.
            return [];
        }
        if (size !== a.length) {
            throw new Error(`[${shape}] does not match the input size ${a.length}${isComplex ? ' for a complex tensor' : ''}.`);
        }
        return createNestedArray(0, shape, a, isComplex);
    }
    function makeOnesTypedArray(size, dtype) {
        const array = makeZerosTypedArray(size, dtype);
        for (let i = 0; i < array.length; i++) {
            array[i] = 1;
        }
        return array;
    }
    function makeZerosTypedArray(size, dtype) {
        if (dtype == null || dtype === 'float32' || dtype === 'complex64') {
            return new Float32Array(size);
        }
        else if (dtype === 'int32') {
            return new Int32Array(size);
        }
        else if (dtype === 'bool') {
            return new Uint8Array(size);
        }
        else {
            throw new Error(`Unknown data type ${dtype}`);
        }
    }
    /**
     * Make nested `TypedArray` filled with zeros.
     * @param shape The shape information for the nested array.
     * @param dtype dtype of the array element.
     */
    function makeZerosNestedTypedArray(shape, dtype) {
        const size = shape.reduce((prev, curr) => prev * curr, 1);
        if (dtype == null || dtype === 'float32') {
            return toNestedArray(shape, new Float32Array(size));
        }
        else if (dtype === 'int32') {
            return toNestedArray(shape, new Int32Array(size));
        }
        else if (dtype === 'bool') {
            return toNestedArray(shape, new Uint8Array(size));
        }
        else {
            throw new Error(`Unknown data type ${dtype}`);
        }
    }
    function assertNonNegativeIntegerDimensions(shape) {
        shape.forEach(dimSize => {
            assert$4(Number.isInteger(dimSize) && dimSize >= 0, () => `Tensor must have a shape comprised of positive integers but got ` +
                `shape [${shape}].`);
        });
    }
    /**
     * Computes flat index for a given location (multidimentionsal index) in a
     * Tensor/multidimensional array.
     *
     * @param locs Location in the tensor.
     * @param rank Rank of the tensor.
     * @param strides Tensor strides.
     */
    function locToIndex(locs, rank, strides) {
        if (rank === 0) {
            return 0;
        }
        else if (rank === 1) {
            return locs[0];
        }
        let index = locs[locs.length - 1];
        for (let i = 0; i < locs.length - 1; ++i) {
            index += strides[i] * locs[i];
        }
        return index;
    }
    /**
     * Computes the location (multidimensional index) in a tensor/multidimentional
     * array for a given flat index.
     *
     * @param index Index in flat array.
     * @param rank Rank of tensor.
     * @param strides Strides of tensor.
     */
    function indexToLoc(index, rank, strides) {
        if (rank === 0) {
            return [];
        }
        else if (rank === 1) {
            return [index];
        }
        const locs = new Array(rank);
        for (let i = 0; i < locs.length - 1; ++i) {
            locs[i] = Math.floor(index / strides[i]);
            index -= locs[i] * strides[i];
        }
        locs[locs.length - 1] = index;
        return locs;
    }
    /**
     * This method asserts whether an object is a Promise instance.
     * @param object
     */
    // tslint:disable-next-line: no-any
    function isPromise(object) {
        //  We chose to not use 'obj instanceOf Promise' for two reasons:
        //  1. It only reliably works for es6 Promise, not other Promise
        //  implementations.
        //  2. It doesn't work with framework that uses zone.js. zone.js monkey patch
        //  the async calls, so it is possible the obj (patched) is comparing to a
        //  pre-patched Promise.
        return object && object.then && typeof object.then === 'function';
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Expects flags from URL in the format ?tfjsflags=FLAG1:1,FLAG2:true.
    const TENSORFLOWJS_FLAGS_PREFIX = 'tfjsflags';
    /**
     * The environment contains evaluated flags as well as the registered platform.
     * This is always used as a global singleton and can be retrieved with
     * `tf.env()`.
     *
     * @doc {heading: 'Environment'}
     */
    class Environment {
        // tslint:disable-next-line: no-any
        constructor(global) {
            this.global = global;
            this.flags = {};
            this.flagRegistry = {};
            this.urlFlags = {};
            // Jasmine spies on this in 'environment_test.ts'
            this.getQueryParams = getQueryParams;
            this.populateURLFlags();
        }
        setPlatform(platformName, platform) {
            if (this.platform != null) {
                if (!(env$1().getBool('IS_TEST') || env$1().getBool('PROD'))) {
                    console.warn(`Platform ${this.platformName} has already been set. ` +
                        `Overwriting the platform with ${platformName}.`);
                }
            }
            this.platformName = platformName;
            this.platform = platform;
        }
        registerFlag(flagName, evaluationFn, setHook) {
            this.flagRegistry[flagName] = { evaluationFn, setHook };
            // Override the flag value from the URL. This has to happen here because
            // the environment is initialized before flags get registered.
            if (this.urlFlags[flagName] != null) {
                const flagValue = this.urlFlags[flagName];
                if (!(env$1().getBool('IS_TEST') || env$1().getBool('PROD'))) {
                    console.warn(`Setting feature override from URL ${flagName}: ${flagValue}.`);
                }
                this.set(flagName, flagValue);
            }
        }
        async getAsync(flagName) {
            if (flagName in this.flags) {
                return this.flags[flagName];
            }
            this.flags[flagName] = await this.evaluateFlag(flagName);
            return this.flags[flagName];
        }
        get(flagName) {
            if (flagName in this.flags) {
                return this.flags[flagName];
            }
            const flagValue = this.evaluateFlag(flagName);
            if (isPromise(flagValue)) {
                throw new Error(`Flag ${flagName} cannot be synchronously evaluated. ` +
                    `Please use getAsync() instead.`);
            }
            this.flags[flagName] = flagValue;
            return this.flags[flagName];
        }
        getNumber(flagName) {
            return this.get(flagName);
        }
        getBool(flagName) {
            return this.get(flagName);
        }
        getFlags() {
            return this.flags;
        }
        // For backwards compatibility.
        get features() {
            return this.flags;
        }
        set(flagName, value) {
            if (this.flagRegistry[flagName] == null) {
                throw new Error(`Cannot set flag ${flagName} as it has not been registered.`);
            }
            this.flags[flagName] = value;
            if (this.flagRegistry[flagName].setHook != null) {
                this.flagRegistry[flagName].setHook(value);
            }
        }
        evaluateFlag(flagName) {
            if (this.flagRegistry[flagName] == null) {
                throw new Error(`Cannot evaluate flag '${flagName}': no evaluation function found.`);
            }
            return this.flagRegistry[flagName].evaluationFn();
        }
        setFlags(flags) {
            this.flags = Object.assign({}, flags);
        }
        reset() {
            this.flags = {};
            this.urlFlags = {};
            this.populateURLFlags();
        }
        populateURLFlags() {
            if (typeof this.global === 'undefined' ||
                typeof this.global.location === 'undefined' ||
                typeof this.global.location.search === 'undefined') {
                return;
            }
            const urlParams = this.getQueryParams(this.global.location.search);
            if (TENSORFLOWJS_FLAGS_PREFIX in urlParams) {
                const keyValues = urlParams[TENSORFLOWJS_FLAGS_PREFIX].split(',');
                keyValues.forEach(keyValue => {
                    const [key, value] = keyValue.split(':');
                    this.urlFlags[key] = parseValue(key, value);
                });
            }
        }
    }
    function getQueryParams(queryString) {
        const params = {};
        queryString.replace(/[?&]([^=?&]+)(?:=([^&]*))?/g, (s, ...t) => {
            decodeParam(params, t[0], t[1]);
            return t.join('=');
        });
        return params;
    }
    function decodeParam(params, name, value) {
        params[decodeURIComponent(name)] = decodeURIComponent(value || '');
    }
    function parseValue(flagName, value) {
        value = value.toLowerCase();
        if (value === 'true' || value === 'false') {
            return value === 'true';
        }
        else if (`${+value}` === value) {
            return +value;
        }
        throw new Error(`Could not parse value flag value ${value} for flag ${flagName}.`);
    }
    /**
     * Returns the current environment (a global singleton).
     *
     * The environment object contains the evaluated feature values as well as the
     * active platform.
     *
     * @doc {heading: 'Environment'}
     */
    function env$1() {
        return ENV$4;
    }
    let ENV$4 = null;
    function setEnvironmentGlobal(environment) {
        ENV$4 = environment;
    }

    var global$1 = (typeof global !== "undefined" ? global :
                typeof self !== "undefined" ? self :
                typeof window !== "undefined" ? window : {});

    // shim for using process in browser
    // based off https://github.com/defunctzombie/node-process/blob/master/browser.js

    function defaultSetTimout() {
        throw new Error('setTimeout has not been defined');
    }
    function defaultClearTimeout () {
        throw new Error('clearTimeout has not been defined');
    }
    var cachedSetTimeout = defaultSetTimout;
    var cachedClearTimeout = defaultClearTimeout;
    if (typeof global$1.setTimeout === 'function') {
        cachedSetTimeout = setTimeout;
    }
    if (typeof global$1.clearTimeout === 'function') {
        cachedClearTimeout = clearTimeout;
    }

    function runTimeout(fun) {
        if (cachedSetTimeout === setTimeout) {
            //normal enviroments in sane situations
            return setTimeout(fun, 0);
        }
        // if setTimeout wasn't available but was latter defined
        if ((cachedSetTimeout === defaultSetTimout || !cachedSetTimeout) && setTimeout) {
            cachedSetTimeout = setTimeout;
            return setTimeout(fun, 0);
        }
        try {
            // when when somebody has screwed with setTimeout but no I.E. maddness
            return cachedSetTimeout(fun, 0);
        } catch(e){
            try {
                // When we are in I.E. but the script has been evaled so I.E. doesn't trust the global object when called normally
                return cachedSetTimeout.call(null, fun, 0);
            } catch(e){
                // same as above but when it's a version of I.E. that must have the global object for 'this', hopfully our context correct otherwise it will throw a global error
                return cachedSetTimeout.call(this, fun, 0);
            }
        }


    }
    function runClearTimeout(marker) {
        if (cachedClearTimeout === clearTimeout) {
            //normal enviroments in sane situations
            return clearTimeout(marker);
        }
        // if clearTimeout wasn't available but was latter defined
        if ((cachedClearTimeout === defaultClearTimeout || !cachedClearTimeout) && clearTimeout) {
            cachedClearTimeout = clearTimeout;
            return clearTimeout(marker);
        }
        try {
            // when when somebody has screwed with setTimeout but no I.E. maddness
            return cachedClearTimeout(marker);
        } catch (e){
            try {
                // When we are in I.E. but the script has been evaled so I.E. doesn't  trust the global object when called normally
                return cachedClearTimeout.call(null, marker);
            } catch (e){
                // same as above but when it's a version of I.E. that must have the global object for 'this', hopfully our context correct otherwise it will throw a global error.
                // Some versions of I.E. have different rules for clearTimeout vs setTimeout
                return cachedClearTimeout.call(this, marker);
            }
        }



    }
    var queue = [];
    var draining = false;
    var currentQueue;
    var queueIndex = -1;

    function cleanUpNextTick() {
        if (!draining || !currentQueue) {
            return;
        }
        draining = false;
        if (currentQueue.length) {
            queue = currentQueue.concat(queue);
        } else {
            queueIndex = -1;
        }
        if (queue.length) {
            drainQueue();
        }
    }

    function drainQueue() {
        if (draining) {
            return;
        }
        var timeout = runTimeout(cleanUpNextTick);
        draining = true;

        var len = queue.length;
        while(len) {
            currentQueue = queue;
            queue = [];
            while (++queueIndex < len) {
                if (currentQueue) {
                    currentQueue[queueIndex].run();
                }
            }
            queueIndex = -1;
            len = queue.length;
        }
        currentQueue = null;
        draining = false;
        runClearTimeout(timeout);
    }
    function nextTick(fun) {
        var args = new Array(arguments.length - 1);
        if (arguments.length > 1) {
            for (var i = 1; i < arguments.length; i++) {
                args[i - 1] = arguments[i];
            }
        }
        queue.push(new Item(fun, args));
        if (queue.length === 1 && !draining) {
            runTimeout(drainQueue);
        }
    }
    // v8 likes predictible objects
    function Item(fun, array) {
        this.fun = fun;
        this.array = array;
    }
    Item.prototype.run = function () {
        this.fun.apply(null, this.array);
    };
    var title = 'browser';
    var platform$1 = 'browser';
    var browser$3 = true;
    var env = {};
    var argv = [];
    var version$8 = ''; // empty string to avoid regexp issues
    var versions = {};
    var release$1 = {};
    var config = {};

    function noop() {}

    var on = noop;
    var addListener = noop;
    var once = noop;
    var off = noop;
    var removeListener = noop;
    var removeAllListeners = noop;
    var emit = noop;

    function binding$1(name) {
        throw new Error('process.binding is not supported');
    }

    function cwd () { return '/' }
    function chdir (dir) {
        throw new Error('process.chdir is not supported');
    }function umask() { return 0; }

    // from https://github.com/kumavis/browser-process-hrtime/blob/master/index.js
    var performance$1 = global$1.performance || {};
    var performanceNow =
      performance$1.now        ||
      performance$1.mozNow     ||
      performance$1.msNow      ||
      performance$1.oNow       ||
      performance$1.webkitNow  ||
      function(){ return (new Date()).getTime() };

    // generate timestamp or delta
    // see http://nodejs.org/api/process.html#process_process_hrtime
    function hrtime(previousTimestamp){
      var clocktime = performanceNow.call(performance$1)*1e-3;
      var seconds = Math.floor(clocktime);
      var nanoseconds = Math.floor((clocktime%1)*1e9);
      if (previousTimestamp) {
        seconds = seconds - previousTimestamp[0];
        nanoseconds = nanoseconds - previousTimestamp[1];
        if (nanoseconds<0) {
          seconds--;
          nanoseconds += 1e9;
        }
      }
      return [seconds,nanoseconds]
    }

    var startTime = new Date();
    function uptime$1() {
      var currentTime = new Date();
      var dif = currentTime - startTime;
      return dif / 1000;
    }

    var process = {
      nextTick: nextTick,
      title: title,
      browser: browser$3,
      env: env,
      argv: argv,
      version: version$8,
      versions: versions,
      on: on,
      addListener: addListener,
      once: once,
      off: off,
      removeListener: removeListener,
      removeAllListeners: removeAllListeners,
      emit: emit,
      binding: binding$1,
      cwd: cwd,
      chdir: chdir,
      umask: umask,
      hrtime: hrtime,
      platform: platform$1,
      release: release$1,
      config: config,
      uptime: uptime$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Note that the identifier globalNameSpace is scoped to this module, but will
    // always resolve to the same global object regardless of how the module is
    // resolved.
    // tslint:disable-next-line:no-any
    let globalNameSpace;
    // tslint:disable-next-line:no-any
    function getGlobalNamespace() {
        if (globalNameSpace == null) {
            // tslint:disable-next-line:no-any
            let ns;
            if (typeof (window) !== 'undefined') {
                ns = window;
            }
            else if (typeof (global$1) !== 'undefined') {
                ns = global$1;
            }
            else if (typeof (process) !== 'undefined') {
                ns = process;
            }
            else if (typeof (self) !== 'undefined') {
                ns = self;
            }
            else {
                throw new Error('Could not find a global object');
            }
            globalNameSpace = ns;
        }
        return globalNameSpace;
    }
    // tslint:disable-next-line:no-any
    function getGlobalMap() {
        const ns = getGlobalNamespace();
        if (ns._tfGlobals == null) {
            ns._tfGlobals = new Map();
        }
        return ns._tfGlobals;
    }
    /**
     * Returns a globally accessible 'singleton' object.
     *
     * @param key the name of the object
     * @param init a function to initialize to initialize this object
     *             the first time it is fetched.
     */
    function getGlobal(key, init) {
        const globalMap = getGlobalMap();
        if (globalMap.has(key)) {
            return globalMap.get(key);
        }
        else {
            const singleton = init();
            globalMap.set(key, singleton);
            return globalMap.get(key);
        }
    }

    const Abs = 'Abs';
    const Acos = 'Acos';
    const Acosh = 'Acosh';
    const Add$1 = 'Add';
    const AddN = 'AddN';
    const All = 'All';
    const Any = 'Any';
    const ArgMax = 'ArgMax';
    const ArgMin = 'ArgMin';
    const Asin = 'Asin';
    const Asinh = 'Asinh';
    const Atan = 'Atan';
    const Atanh = 'Atanh';
    const Atan2 = 'Atan2';
    const AvgPool = 'AvgPool';
    const AvgPoolGrad = 'AvgPoolGrad';
    const AvgPool3D = 'AvgPool3D';
    const AvgPool3DGrad = 'AvgPool3DGrad';
    const BatchMatMul = 'BatchMatMul';
    const BatchToSpaceND = 'BatchToSpaceND';
    const Bincount = 'Bincount';
    const BroadcastTo = 'BroadcastTo';
    const BroadcastArgs = 'BroadcastArgs';
    const Cast = 'Cast';
    const Ceil = 'Ceil';
    const ClipByValue = 'ClipByValue';
    const Complex = 'Complex';
    const ComplexAbs = 'ComplexAbs';
    const Concat = 'Concat';
    const Conv2D$1 = 'Conv2D';
    const Conv2DBackpropFilter = 'Conv2DBackpropFilter';
    const Conv2DBackpropInput = 'Conv2DBackpropInput';
    const Conv3D$1 = 'Conv3D';
    const Conv3DBackpropFilterV2 = 'Conv3DBackpropFilterV2';
    const Conv3DBackpropInputV2 = 'Conv3DBackpropInputV2';
    const Cos = 'Cos';
    const Cosh = 'Cosh';
    const Cumprod = 'Cumprod';
    const Cumsum = 'Cumsum';
    const CropAndResize = 'CropAndResize';
    const DenseBincount = 'DenseBincount';
    const DepthToSpace = 'DepthToSpace';
    const DepthwiseConv2dNative = 'DepthwiseConv2dNative';
    const DepthwiseConv2dNativeBackpropFilter = 'DepthwiseConv2dNativeBackpropFilter';
    const DepthwiseConv2dNativeBackpropInput = 'DepthwiseConv2dNativeBackpropInput';
    const Diag = 'Diag';
    const Dilation2D = 'Dilation2D';
    const Dilation2DBackpropInput = 'Dilation2DBackpropInput';
    const Dilation2DBackpropFilter = 'Dilation2DBackpropFilter';
    const RealDiv = 'RealDiv';
    const Einsum = 'Einsum';
    const Elu$1 = 'Elu';
    const EluGrad = 'EluGrad';
    const Erf = 'Erf';
    const Equal = 'Equal';
    const Exp = 'Exp';
    const ExpandDims = 'ExpandDims';
    const Expm1 = 'Expm1';
    const FFT = 'FFT';
    const Fill = 'Fill';
    const FlipLeftRight = 'FlipLeftRight';
    const Floor = 'Floor';
    const FloorDiv = 'FloorDiv';
    const FusedBatchNorm = 'FusedBatchNorm';
    const GatherV2 = 'GatherV2';
    const GatherNd = 'GatherNd';
    const Greater = 'Greater';
    const GreaterEqual = 'GreaterEqual';
    const Identity$1 = 'Identity';
    const IFFT = 'IFFT';
    const Imag = 'Imag';
    const IsFinite = 'IsFinite';
    const IsInf = 'IsInf';
    const IsNan = 'IsNan';
    const LeakyRelu = 'LeakyRelu';
    const Less = 'Less';
    const LessEqual = 'LessEqual';
    const LinSpace = 'LinSpace';
    const Log = 'Log';
    const Log1p = 'Log1p';
    const LogicalAnd = 'LogicalAnd';
    const LogicalNot = 'LogicalNot';
    const LogicalOr = 'LogicalOr';
    const LogSoftmax$1 = 'LogSoftmax';
    const LowerBound = 'LowerBound';
    const LRN = 'LRN';
    const LRNGrad = 'LRNGrad';
    const Max = 'Max';
    const Maximum$1 = 'Maximum';
    const MaxPool = 'MaxPool';
    const MaxPoolGrad = 'MaxPoolGrad';
    const MaxPool3D = 'MaxPool3D';
    const MaxPool3DGrad = 'MaxPool3DGrad';
    const MaxPoolWithArgmax = 'MaxPoolWithArgmax';
    const Mean = 'Mean';
    const Min = 'Min';
    const Minimum$1 = 'Minimum';
    const MirrorPad = 'MirrorPad';
    const Mod = 'Mod';
    const Multinomial = 'Multinomial';
    const Multiply$1 = 'Multiply';
    const Neg = 'Neg';
    const NotEqual = 'NotEqual';
    const NonMaxSuppressionV3 = 'NonMaxSuppressionV3';
    const NonMaxSuppressionV4 = 'NonMaxSuppressionV4';
    const NonMaxSuppressionV5 = 'NonMaxSuppressionV5';
    const OnesLike = 'OnesLike';
    const OneHot = 'OneHot';
    const Pack = 'Pack';
    const PadV2 = 'PadV2';
    const Pool = 'Pool';
    const Pow = 'Pow';
    const Prelu = 'Prelu';
    const Prod = 'Prod';
    const Range = 'Range';
    const Real = 'Real';
    const Reciprocal = 'Reciprocal';
    const Relu$1 = 'Relu';
    const Reshape$1 = 'Reshape';
    const ResizeNearestNeighbor = 'ResizeNearestNeighbor';
    const ResizeNearestNeighborGrad = 'ResizeNearestNeighborGrad';
    const ResizeBilinear = 'ResizeBilinear';
    const ResizeBilinearGrad = 'ResizeBilinearGrad';
    const Relu6$1 = 'Relu6';
    const Reverse = 'Reverse';
    const Round = 'Round';
    const Rsqrt = 'Rsqrt';
    const ScatterNd = 'ScatterNd';
    const SearchSorted = 'SearchSorted';
    const Select = 'Select';
    const Selu$1 = 'Selu';
    const Slice = 'Slice';
    const Sin = 'Sin';
    const Sinh = 'Sinh';
    const Sign = 'Sign';
    const Sigmoid$1 = 'Sigmoid';
    const Softplus$1 = 'Softplus';
    const Sqrt = 'Sqrt';
    const Sum = 'Sum';
    const SpaceToBatchND = 'SpaceToBatchND';
    const SplitV = 'SplitV';
    const Softmax$2 = 'Softmax';
    const SparseFillEmptyRows = 'SparseFillEmptyRows';
    const SparseReshape = 'SparseReshape';
    const SparseSegmentMean = 'SparseSegmentMean';
    const SparseSegmentSum = 'SparseSegmentSum';
    const SparseToDense = 'SparseToDense';
    const SquaredDifference = 'SquaredDifference';
    const Square = 'Square';
    const StridedSlice = 'StridedSlice';
    const StringNGrams = 'StringNGrams';
    const StringSplit = 'StringSplit';
    const StringToHashBucketFast = 'StringToHashBucketFast';
    const Sub = 'Sub';
    const Tan = 'Tan';
    const Tanh$1 = 'Tanh';
    const Tile = 'Tile';
    const TopK = 'TopK';
    const Transform$1 = 'Transform';
    const Transpose = 'Transpose';
    const Unique = 'Unique';
    const Unpack = 'Unpack';
    const UnsortedSegmentSum = 'UnsortedSegmentSum';
    const UpperBound = 'UpperBound';
    const ZerosLike = 'ZerosLike';
    /**
     * TensorFlow.js-only kernels
     */
    const Step = 'Step';
    const FromPixels = 'FromPixels';
    const RotateWithOffset = 'RotateWithOffset';
    const _FusedMatMul = '_FusedMatMul';
    const FusedConv2D = 'FusedConv2D';
    const FusedDepthwiseConv2D = 'FusedDepthwiseConv2D';

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function warn(...msg) {
        if (!(env$1().getBool('IS_TEST') || env$1().getBool('PROD'))) {
            console.warn(...msg);
        }
    }
    function log$4(...msg) {
        if (!(env$1().getBool('IS_TEST') || env$1().getBool('PROD'))) {
            console.log(...msg);
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const kernelRegistry = getGlobal('kernelRegistry', () => new Map());
    const gradRegistry = getGlobal('gradRegistry', () => new Map());
    /**
     * Returns the kernel function (code) associated with the provided names.
     *
     * @param kernelName The official name of the kernel.
     * @param backendName The official name of the backend.
     */
    function getKernel(kernelName, backendName) {
        const key = makeKey(kernelName, backendName);
        return kernelRegistry.get(key);
    }
    /**
     * Returns the registered gradient info associated with the provided kernel.
     * @param kernelName The official TF kernel name.
     */
    function getGradient(kernelName) {
        return gradRegistry.get(kernelName);
    }
    function getKernelsForBackend(backendName) {
        const it = kernelRegistry.entries();
        const result = [];
        while (true) {
            const { done, value } = it.next();
            if (done) {
                break;
            }
            const [key, config] = value;
            const [backend,] = key.split('_');
            if (backend === backendName) {
                result.push(config);
            }
        }
        return result;
    }
    /**
     * Registers the function (forward pass) for the kernel in a global registry.
     *
     * @param config A config object with the following properties:
     * - `kernelName` The official name of the kernel.
     * - `backendName` The official name of the backend.
     * - `kernelFunc` The function to run during the forward pass of the kernel.
     * - `setupFunc` Optional. Gets called once, after the backend initializes.
     * - `disposeFunc` Optional. Gets called once, right before the backend is
     * disposed.
     */
    function registerKernel(config) {
        const { kernelName, backendName } = config;
        const key = makeKey(kernelName, backendName);
        if (kernelRegistry.has(key)) {
            warn(`The kernel '${kernelName}' for backend ` +
                `'${backendName}' is already registered`);
        }
        kernelRegistry.set(key, config);
    }
    /**
     * Registers a gradient function for a given kernel in the global registry,
     * to be used during the back-propagation of that kernel.
     *
     * @param config An object with the following properties:
     * - `kernelName` The name of the kernel that the gradient function is for.
     * - `gradFunc` The function to run during back-propagation.
     */
    function registerGradient(config) {
        const { kernelName } = config;
        if (gradRegistry.has(kernelName)) {
            // TODO (yassogba) after 3.0 assess whether we need to keep this gated
            // to debug mode.
            if (env$1().getBool('DEBUG')) {
                warn(`Overriding the gradient for '${kernelName}'`);
            }
        }
        gradRegistry.set(kernelName, config);
    }
    /**
     * Removes the kernel function from the registry.
     *
     * @param kernelName The official name of the kernel.
     * @param backendName The official name of the backend.
     *
     */
    function unregisterKernel(kernelName, backendName) {
        const key = makeKey(kernelName, backendName);
        if (!kernelRegistry.has(key)) {
            throw new Error(`The kernel '${kernelName}' for backend ` +
                `'${backendName}' is not registered`);
        }
        kernelRegistry.delete(key);
    }
    /** Removes the registered gradient from the global registry. */
    function unregisterGradient(kernelName) {
        if (!gradRegistry.has(kernelName)) {
            throw new Error(`The gradient '${kernelName}' for backend is not registered`);
        }
        gradRegistry.delete(kernelName);
    }
    /**
     * Finds kernels that have already been registered to a backend and re-registers
     * them for a new backend. Useful for registering custom backends.
     * @param registeredBackendName Already registered backend.
     * @param newBackendName New backend.
     */
    function copyRegisteredKernels(registeredBackendName, newBackendName) {
        const kernels = getKernelsForBackend(registeredBackendName);
        kernels.forEach(kernelConfig => {
            const newKernelConfig = Object.assign({}, kernelConfig, { backendName: newBackendName });
            registerKernel(newKernelConfig);
        });
    }
    function makeKey(kernelName, backendName) {
        return `${backendName}_${kernelName}`;
    }

    var commonjsGlobal = typeof globalThis !== 'undefined' ? globalThis : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

    function getDefaultExportFromCjs (x) {
    	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
    }

    function getAugmentedNamespace(n) {
      var f = n.default;
    	if (typeof f == "function") {
    		var a = function () {
    			return f.apply(this, arguments);
    		};
    		a.prototype = f.prototype;
      } else a = {};
      Object.defineProperty(a, '__esModule', {value: true});
    	Object.keys(n).forEach(function (k) {
    		var d = Object.getOwnPropertyDescriptor(n, k);
    		Object.defineProperty(a, k, d.get ? d : {
    			enumerable: true,
    			get: function () {
    				return n[k];
    			}
    		});
    	});
    	return a;
    }

    var long = Long$1;

    /**
     * wasm optimizations, to do native i64 multiplication and divide
     */
    var wasm = null;

    try {
      wasm = new WebAssembly.Instance(new WebAssembly.Module(new Uint8Array([
        0, 97, 115, 109, 1, 0, 0, 0, 1, 13, 2, 96, 0, 1, 127, 96, 4, 127, 127, 127, 127, 1, 127, 3, 7, 6, 0, 1, 1, 1, 1, 1, 6, 6, 1, 127, 1, 65, 0, 11, 7, 50, 6, 3, 109, 117, 108, 0, 1, 5, 100, 105, 118, 95, 115, 0, 2, 5, 100, 105, 118, 95, 117, 0, 3, 5, 114, 101, 109, 95, 115, 0, 4, 5, 114, 101, 109, 95, 117, 0, 5, 8, 103, 101, 116, 95, 104, 105, 103, 104, 0, 0, 10, 191, 1, 6, 4, 0, 35, 0, 11, 36, 1, 1, 126, 32, 0, 173, 32, 1, 173, 66, 32, 134, 132, 32, 2, 173, 32, 3, 173, 66, 32, 134, 132, 126, 34, 4, 66, 32, 135, 167, 36, 0, 32, 4, 167, 11, 36, 1, 1, 126, 32, 0, 173, 32, 1, 173, 66, 32, 134, 132, 32, 2, 173, 32, 3, 173, 66, 32, 134, 132, 127, 34, 4, 66, 32, 135, 167, 36, 0, 32, 4, 167, 11, 36, 1, 1, 126, 32, 0, 173, 32, 1, 173, 66, 32, 134, 132, 32, 2, 173, 32, 3, 173, 66, 32, 134, 132, 128, 34, 4, 66, 32, 135, 167, 36, 0, 32, 4, 167, 11, 36, 1, 1, 126, 32, 0, 173, 32, 1, 173, 66, 32, 134, 132, 32, 2, 173, 32, 3, 173, 66, 32, 134, 132, 129, 34, 4, 66, 32, 135, 167, 36, 0, 32, 4, 167, 11, 36, 1, 1, 126, 32, 0, 173, 32, 1, 173, 66, 32, 134, 132, 32, 2, 173, 32, 3, 173, 66, 32, 134, 132, 130, 34, 4, 66, 32, 135, 167, 36, 0, 32, 4, 167, 11
      ])), {}).exports;
    } catch (e) {
      // no wasm support :(
    }

    /**
     * Constructs a 64 bit two's-complement integer, given its low and high 32 bit values as *signed* integers.
     *  See the from* functions below for more convenient ways of constructing Longs.
     * @exports Long
     * @class A Long class for representing a 64 bit two's-complement integer value.
     * @param {number} low The low (signed) 32 bits of the long
     * @param {number} high The high (signed) 32 bits of the long
     * @param {boolean=} unsigned Whether unsigned or not, defaults to signed
     * @constructor
     */
    function Long$1(low, high, unsigned) {

        /**
         * The low 32 bits as a signed value.
         * @type {number}
         */
        this.low = low | 0;

        /**
         * The high 32 bits as a signed value.
         * @type {number}
         */
        this.high = high | 0;

        /**
         * Whether unsigned or not.
         * @type {boolean}
         */
        this.unsigned = !!unsigned;
    }

    // The internal representation of a long is the two given signed, 32-bit values.
    // We use 32-bit pieces because these are the size of integers on which
    // Javascript performs bit-operations.  For operations like addition and
    // multiplication, we split each number into 16 bit pieces, which can easily be
    // multiplied within Javascript's floating-point representation without overflow
    // or change in sign.
    //
    // In the algorithms below, we frequently reduce the negative case to the
    // positive case by negating the input(s) and then post-processing the result.
    // Note that we must ALWAYS check specially whether those values are MIN_VALUE
    // (-2^63) because -MIN_VALUE == MIN_VALUE (since 2^63 cannot be represented as
    // a positive number, it overflows back into a negative).  Not handling this
    // case would often result in infinite recursion.
    //
    // Common constant values ZERO, ONE, NEG_ONE, etc. are defined below the from*
    // methods on which they depend.

    /**
     * An indicator used to reliably determine if an object is a Long or not.
     * @type {boolean}
     * @const
     * @private
     */
    Long$1.prototype.__isLong__;

    Object.defineProperty(Long$1.prototype, "__isLong__", { value: true });

    /**
     * @function
     * @param {*} obj Object
     * @returns {boolean}
     * @inner
     */
    function isLong(obj) {
        return (obj && obj["__isLong__"]) === true;
    }

    /**
     * Tests if the specified object is a Long.
     * @function
     * @param {*} obj Object
     * @returns {boolean}
     */
    Long$1.isLong = isLong;

    /**
     * A cache of the Long representations of small integer values.
     * @type {!Object}
     * @inner
     */
    var INT_CACHE = {};

    /**
     * A cache of the Long representations of small unsigned integer values.
     * @type {!Object}
     * @inner
     */
    var UINT_CACHE = {};

    /**
     * @param {number} value
     * @param {boolean=} unsigned
     * @returns {!Long}
     * @inner
     */
    function fromInt(value, unsigned) {
        var obj, cachedObj, cache;
        if (unsigned) {
            value >>>= 0;
            if (cache = (0 <= value && value < 256)) {
                cachedObj = UINT_CACHE[value];
                if (cachedObj)
                    return cachedObj;
            }
            obj = fromBits(value, (value | 0) < 0 ? -1 : 0, true);
            if (cache)
                UINT_CACHE[value] = obj;
            return obj;
        } else {
            value |= 0;
            if (cache = (-128 <= value && value < 128)) {
                cachedObj = INT_CACHE[value];
                if (cachedObj)
                    return cachedObj;
            }
            obj = fromBits(value, value < 0 ? -1 : 0, false);
            if (cache)
                INT_CACHE[value] = obj;
            return obj;
        }
    }

    /**
     * Returns a Long representing the given 32 bit integer value.
     * @function
     * @param {number} value The 32 bit integer in question
     * @param {boolean=} unsigned Whether unsigned or not, defaults to signed
     * @returns {!Long} The corresponding Long value
     */
    Long$1.fromInt = fromInt;

    /**
     * @param {number} value
     * @param {boolean=} unsigned
     * @returns {!Long}
     * @inner
     */
    function fromNumber(value, unsigned) {
        if (isNaN(value))
            return unsigned ? UZERO : ZERO;
        if (unsigned) {
            if (value < 0)
                return UZERO;
            if (value >= TWO_PWR_64_DBL)
                return MAX_UNSIGNED_VALUE;
        } else {
            if (value <= -TWO_PWR_63_DBL)
                return MIN_VALUE;
            if (value + 1 >= TWO_PWR_63_DBL)
                return MAX_VALUE;
        }
        if (value < 0)
            return fromNumber(-value, unsigned).neg();
        return fromBits((value % TWO_PWR_32_DBL) | 0, (value / TWO_PWR_32_DBL) | 0, unsigned);
    }

    /**
     * Returns a Long representing the given value, provided that it is a finite number. Otherwise, zero is returned.
     * @function
     * @param {number} value The number in question
     * @param {boolean=} unsigned Whether unsigned or not, defaults to signed
     * @returns {!Long} The corresponding Long value
     */
    Long$1.fromNumber = fromNumber;

    /**
     * @param {number} lowBits
     * @param {number} highBits
     * @param {boolean=} unsigned
     * @returns {!Long}
     * @inner
     */
    function fromBits(lowBits, highBits, unsigned) {
        return new Long$1(lowBits, highBits, unsigned);
    }

    /**
     * Returns a Long representing the 64 bit integer that comes by concatenating the given low and high bits. Each is
     *  assumed to use 32 bits.
     * @function
     * @param {number} lowBits The low 32 bits
     * @param {number} highBits The high 32 bits
     * @param {boolean=} unsigned Whether unsigned or not, defaults to signed
     * @returns {!Long} The corresponding Long value
     */
    Long$1.fromBits = fromBits;

    /**
     * @function
     * @param {number} base
     * @param {number} exponent
     * @returns {number}
     * @inner
     */
    var pow_dbl = Math.pow; // Used 4 times (4*8 to 15+4)

    /**
     * @param {string} str
     * @param {(boolean|number)=} unsigned
     * @param {number=} radix
     * @returns {!Long}
     * @inner
     */
    function fromString$1(str, unsigned, radix) {
        if (str.length === 0)
            throw Error('empty string');
        if (str === "NaN" || str === "Infinity" || str === "+Infinity" || str === "-Infinity")
            return ZERO;
        if (typeof unsigned === 'number') {
            // For goog.math.long compatibility
            radix = unsigned,
            unsigned = false;
        } else {
            unsigned = !! unsigned;
        }
        radix = radix || 10;
        if (radix < 2 || 36 < radix)
            throw RangeError('radix');

        var p;
        if ((p = str.indexOf('-')) > 0)
            throw Error('interior hyphen');
        else if (p === 0) {
            return fromString$1(str.substring(1), unsigned, radix).neg();
        }

        // Do several (8) digits each time through the loop, so as to
        // minimize the calls to the very expensive emulated div.
        var radixToPower = fromNumber(pow_dbl(radix, 8));

        var result = ZERO;
        for (var i = 0; i < str.length; i += 8) {
            var size = Math.min(8, str.length - i),
                value = parseInt(str.substring(i, i + size), radix);
            if (size < 8) {
                var power = fromNumber(pow_dbl(radix, size));
                result = result.mul(power).add(fromNumber(value));
            } else {
                result = result.mul(radixToPower);
                result = result.add(fromNumber(value));
            }
        }
        result.unsigned = unsigned;
        return result;
    }

    /**
     * Returns a Long representation of the given string, written using the specified radix.
     * @function
     * @param {string} str The textual representation of the Long
     * @param {(boolean|number)=} unsigned Whether unsigned or not, defaults to signed
     * @param {number=} radix The radix in which the text is written (2-36), defaults to 10
     * @returns {!Long} The corresponding Long value
     */
    Long$1.fromString = fromString$1;

    /**
     * @function
     * @param {!Long|number|string|!{low: number, high: number, unsigned: boolean}} val
     * @param {boolean=} unsigned
     * @returns {!Long}
     * @inner
     */
    function fromValue(val, unsigned) {
        if (typeof val === 'number')
            return fromNumber(val, unsigned);
        if (typeof val === 'string')
            return fromString$1(val, unsigned);
        // Throws for non-objects, converts non-instanceof Long:
        return fromBits(val.low, val.high, typeof unsigned === 'boolean' ? unsigned : val.unsigned);
    }

    /**
     * Converts the specified value to a Long using the appropriate from* function for its type.
     * @function
     * @param {!Long|number|string|!{low: number, high: number, unsigned: boolean}} val Value
     * @param {boolean=} unsigned Whether unsigned or not, defaults to signed
     * @returns {!Long}
     */
    Long$1.fromValue = fromValue;

    // NOTE: the compiler should inline these constant values below and then remove these variables, so there should be
    // no runtime penalty for these.

    /**
     * @type {number}
     * @const
     * @inner
     */
    var TWO_PWR_16_DBL = 1 << 16;

    /**
     * @type {number}
     * @const
     * @inner
     */
    var TWO_PWR_24_DBL = 1 << 24;

    /**
     * @type {number}
     * @const
     * @inner
     */
    var TWO_PWR_32_DBL = TWO_PWR_16_DBL * TWO_PWR_16_DBL;

    /**
     * @type {number}
     * @const
     * @inner
     */
    var TWO_PWR_64_DBL = TWO_PWR_32_DBL * TWO_PWR_32_DBL;

    /**
     * @type {number}
     * @const
     * @inner
     */
    var TWO_PWR_63_DBL = TWO_PWR_64_DBL / 2;

    /**
     * @type {!Long}
     * @const
     * @inner
     */
    var TWO_PWR_24 = fromInt(TWO_PWR_24_DBL);

    /**
     * @type {!Long}
     * @inner
     */
    var ZERO = fromInt(0);

    /**
     * Signed zero.
     * @type {!Long}
     */
    Long$1.ZERO = ZERO;

    /**
     * @type {!Long}
     * @inner
     */
    var UZERO = fromInt(0, true);

    /**
     * Unsigned zero.
     * @type {!Long}
     */
    Long$1.UZERO = UZERO;

    /**
     * @type {!Long}
     * @inner
     */
    var ONE = fromInt(1);

    /**
     * Signed one.
     * @type {!Long}
     */
    Long$1.ONE = ONE;

    /**
     * @type {!Long}
     * @inner
     */
    var UONE = fromInt(1, true);

    /**
     * Unsigned one.
     * @type {!Long}
     */
    Long$1.UONE = UONE;

    /**
     * @type {!Long}
     * @inner
     */
    var NEG_ONE = fromInt(-1);

    /**
     * Signed negative one.
     * @type {!Long}
     */
    Long$1.NEG_ONE = NEG_ONE;

    /**
     * @type {!Long}
     * @inner
     */
    var MAX_VALUE = fromBits(0xFFFFFFFF|0, 0x7FFFFFFF|0, false);

    /**
     * Maximum signed value.
     * @type {!Long}
     */
    Long$1.MAX_VALUE = MAX_VALUE;

    /**
     * @type {!Long}
     * @inner
     */
    var MAX_UNSIGNED_VALUE = fromBits(0xFFFFFFFF|0, 0xFFFFFFFF|0, true);

    /**
     * Maximum unsigned value.
     * @type {!Long}
     */
    Long$1.MAX_UNSIGNED_VALUE = MAX_UNSIGNED_VALUE;

    /**
     * @type {!Long}
     * @inner
     */
    var MIN_VALUE = fromBits(0, 0x80000000|0, false);

    /**
     * Minimum signed value.
     * @type {!Long}
     */
    Long$1.MIN_VALUE = MIN_VALUE;

    /**
     * @alias Long.prototype
     * @inner
     */
    var LongPrototype = Long$1.prototype;

    /**
     * Converts the Long to a 32 bit integer, assuming it is a 32 bit integer.
     * @returns {number}
     */
    LongPrototype.toInt = function toInt() {
        return this.unsigned ? this.low >>> 0 : this.low;
    };

    /**
     * Converts the Long to a the nearest floating-point representation of this value (double, 53 bit mantissa).
     * @returns {number}
     */
    LongPrototype.toNumber = function toNumber() {
        if (this.unsigned)
            return ((this.high >>> 0) * TWO_PWR_32_DBL) + (this.low >>> 0);
        return this.high * TWO_PWR_32_DBL + (this.low >>> 0);
    };

    /**
     * Converts the Long to a string written in the specified radix.
     * @param {number=} radix Radix (2-36), defaults to 10
     * @returns {string}
     * @override
     * @throws {RangeError} If `radix` is out of range
     */
    LongPrototype.toString = function toString(radix) {
        radix = radix || 10;
        if (radix < 2 || 36 < radix)
            throw RangeError('radix');
        if (this.isZero())
            return '0';
        if (this.isNegative()) { // Unsigned Longs are never negative
            if (this.eq(MIN_VALUE)) {
                // We need to change the Long value before it can be negated, so we remove
                // the bottom-most digit in this base and then recurse to do the rest.
                var radixLong = fromNumber(radix),
                    div = this.div(radixLong),
                    rem1 = div.mul(radixLong).sub(this);
                return div.toString(radix) + rem1.toInt().toString(radix);
            } else
                return '-' + this.neg().toString(radix);
        }

        // Do several (6) digits each time through the loop, so as to
        // minimize the calls to the very expensive emulated div.
        var radixToPower = fromNumber(pow_dbl(radix, 6), this.unsigned),
            rem = this;
        var result = '';
        while (true) {
            var remDiv = rem.div(radixToPower),
                intval = rem.sub(remDiv.mul(radixToPower)).toInt() >>> 0,
                digits = intval.toString(radix);
            rem = remDiv;
            if (rem.isZero())
                return digits + result;
            else {
                while (digits.length < 6)
                    digits = '0' + digits;
                result = '' + digits + result;
            }
        }
    };

    /**
     * Gets the high 32 bits as a signed integer.
     * @returns {number} Signed high bits
     */
    LongPrototype.getHighBits = function getHighBits() {
        return this.high;
    };

    /**
     * Gets the high 32 bits as an unsigned integer.
     * @returns {number} Unsigned high bits
     */
    LongPrototype.getHighBitsUnsigned = function getHighBitsUnsigned() {
        return this.high >>> 0;
    };

    /**
     * Gets the low 32 bits as a signed integer.
     * @returns {number} Signed low bits
     */
    LongPrototype.getLowBits = function getLowBits() {
        return this.low;
    };

    /**
     * Gets the low 32 bits as an unsigned integer.
     * @returns {number} Unsigned low bits
     */
    LongPrototype.getLowBitsUnsigned = function getLowBitsUnsigned() {
        return this.low >>> 0;
    };

    /**
     * Gets the number of bits needed to represent the absolute value of this Long.
     * @returns {number}
     */
    LongPrototype.getNumBitsAbs = function getNumBitsAbs() {
        if (this.isNegative()) // Unsigned Longs are never negative
            return this.eq(MIN_VALUE) ? 64 : this.neg().getNumBitsAbs();
        var val = this.high != 0 ? this.high : this.low;
        for (var bit = 31; bit > 0; bit--)
            if ((val & (1 << bit)) != 0)
                break;
        return this.high != 0 ? bit + 33 : bit + 1;
    };

    /**
     * Tests if this Long's value equals zero.
     * @returns {boolean}
     */
    LongPrototype.isZero = function isZero() {
        return this.high === 0 && this.low === 0;
    };

    /**
     * Tests if this Long's value equals zero. This is an alias of {@link Long#isZero}.
     * @returns {boolean}
     */
    LongPrototype.eqz = LongPrototype.isZero;

    /**
     * Tests if this Long's value is negative.
     * @returns {boolean}
     */
    LongPrototype.isNegative = function isNegative() {
        return !this.unsigned && this.high < 0;
    };

    /**
     * Tests if this Long's value is positive.
     * @returns {boolean}
     */
    LongPrototype.isPositive = function isPositive() {
        return this.unsigned || this.high >= 0;
    };

    /**
     * Tests if this Long's value is odd.
     * @returns {boolean}
     */
    LongPrototype.isOdd = function isOdd() {
        return (this.low & 1) === 1;
    };

    /**
     * Tests if this Long's value is even.
     * @returns {boolean}
     */
    LongPrototype.isEven = function isEven() {
        return (this.low & 1) === 0;
    };

    /**
     * Tests if this Long's value equals the specified's.
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.equals = function equals(other) {
        if (!isLong(other))
            other = fromValue(other);
        if (this.unsigned !== other.unsigned && (this.high >>> 31) === 1 && (other.high >>> 31) === 1)
            return false;
        return this.high === other.high && this.low === other.low;
    };

    /**
     * Tests if this Long's value equals the specified's. This is an alias of {@link Long#equals}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.eq = LongPrototype.equals;

    /**
     * Tests if this Long's value differs from the specified's.
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.notEquals = function notEquals(other) {
        return !this.eq(/* validates */ other);
    };

    /**
     * Tests if this Long's value differs from the specified's. This is an alias of {@link Long#notEquals}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.neq = LongPrototype.notEquals;

    /**
     * Tests if this Long's value differs from the specified's. This is an alias of {@link Long#notEquals}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.ne = LongPrototype.notEquals;

    /**
     * Tests if this Long's value is less than the specified's.
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.lessThan = function lessThan(other) {
        return this.comp(/* validates */ other) < 0;
    };

    /**
     * Tests if this Long's value is less than the specified's. This is an alias of {@link Long#lessThan}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.lt = LongPrototype.lessThan;

    /**
     * Tests if this Long's value is less than or equal the specified's.
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.lessThanOrEqual = function lessThanOrEqual(other) {
        return this.comp(/* validates */ other) <= 0;
    };

    /**
     * Tests if this Long's value is less than or equal the specified's. This is an alias of {@link Long#lessThanOrEqual}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.lte = LongPrototype.lessThanOrEqual;

    /**
     * Tests if this Long's value is less than or equal the specified's. This is an alias of {@link Long#lessThanOrEqual}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.le = LongPrototype.lessThanOrEqual;

    /**
     * Tests if this Long's value is greater than the specified's.
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.greaterThan = function greaterThan(other) {
        return this.comp(/* validates */ other) > 0;
    };

    /**
     * Tests if this Long's value is greater than the specified's. This is an alias of {@link Long#greaterThan}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.gt = LongPrototype.greaterThan;

    /**
     * Tests if this Long's value is greater than or equal the specified's.
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.greaterThanOrEqual = function greaterThanOrEqual(other) {
        return this.comp(/* validates */ other) >= 0;
    };

    /**
     * Tests if this Long's value is greater than or equal the specified's. This is an alias of {@link Long#greaterThanOrEqual}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.gte = LongPrototype.greaterThanOrEqual;

    /**
     * Tests if this Long's value is greater than or equal the specified's. This is an alias of {@link Long#greaterThanOrEqual}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {boolean}
     */
    LongPrototype.ge = LongPrototype.greaterThanOrEqual;

    /**
     * Compares this Long's value with the specified's.
     * @param {!Long|number|string} other Other value
     * @returns {number} 0 if they are the same, 1 if the this is greater and -1
     *  if the given one is greater
     */
    LongPrototype.compare = function compare(other) {
        if (!isLong(other))
            other = fromValue(other);
        if (this.eq(other))
            return 0;
        var thisNeg = this.isNegative(),
            otherNeg = other.isNegative();
        if (thisNeg && !otherNeg)
            return -1;
        if (!thisNeg && otherNeg)
            return 1;
        // At this point the sign bits are the same
        if (!this.unsigned)
            return this.sub(other).isNegative() ? -1 : 1;
        // Both are positive if at least one is unsigned
        return (other.high >>> 0) > (this.high >>> 0) || (other.high === this.high && (other.low >>> 0) > (this.low >>> 0)) ? -1 : 1;
    };

    /**
     * Compares this Long's value with the specified's. This is an alias of {@link Long#compare}.
     * @function
     * @param {!Long|number|string} other Other value
     * @returns {number} 0 if they are the same, 1 if the this is greater and -1
     *  if the given one is greater
     */
    LongPrototype.comp = LongPrototype.compare;

    /**
     * Negates this Long's value.
     * @returns {!Long} Negated Long
     */
    LongPrototype.negate = function negate() {
        if (!this.unsigned && this.eq(MIN_VALUE))
            return MIN_VALUE;
        return this.not().add(ONE);
    };

    /**
     * Negates this Long's value. This is an alias of {@link Long#negate}.
     * @function
     * @returns {!Long} Negated Long
     */
    LongPrototype.neg = LongPrototype.negate;

    /**
     * Returns the sum of this and the specified Long.
     * @param {!Long|number|string} addend Addend
     * @returns {!Long} Sum
     */
    LongPrototype.add = function add(addend) {
        if (!isLong(addend))
            addend = fromValue(addend);

        // Divide each number into 4 chunks of 16 bits, and then sum the chunks.

        var a48 = this.high >>> 16;
        var a32 = this.high & 0xFFFF;
        var a16 = this.low >>> 16;
        var a00 = this.low & 0xFFFF;

        var b48 = addend.high >>> 16;
        var b32 = addend.high & 0xFFFF;
        var b16 = addend.low >>> 16;
        var b00 = addend.low & 0xFFFF;

        var c48 = 0, c32 = 0, c16 = 0, c00 = 0;
        c00 += a00 + b00;
        c16 += c00 >>> 16;
        c00 &= 0xFFFF;
        c16 += a16 + b16;
        c32 += c16 >>> 16;
        c16 &= 0xFFFF;
        c32 += a32 + b32;
        c48 += c32 >>> 16;
        c32 &= 0xFFFF;
        c48 += a48 + b48;
        c48 &= 0xFFFF;
        return fromBits((c16 << 16) | c00, (c48 << 16) | c32, this.unsigned);
    };

    /**
     * Returns the difference of this and the specified Long.
     * @param {!Long|number|string} subtrahend Subtrahend
     * @returns {!Long} Difference
     */
    LongPrototype.subtract = function subtract(subtrahend) {
        if (!isLong(subtrahend))
            subtrahend = fromValue(subtrahend);
        return this.add(subtrahend.neg());
    };

    /**
     * Returns the difference of this and the specified Long. This is an alias of {@link Long#subtract}.
     * @function
     * @param {!Long|number|string} subtrahend Subtrahend
     * @returns {!Long} Difference
     */
    LongPrototype.sub = LongPrototype.subtract;

    /**
     * Returns the product of this and the specified Long.
     * @param {!Long|number|string} multiplier Multiplier
     * @returns {!Long} Product
     */
    LongPrototype.multiply = function multiply(multiplier) {
        if (this.isZero())
            return ZERO;
        if (!isLong(multiplier))
            multiplier = fromValue(multiplier);

        // use wasm support if present
        if (wasm) {
            var low = wasm.mul(this.low,
                               this.high,
                               multiplier.low,
                               multiplier.high);
            return fromBits(low, wasm.get_high(), this.unsigned);
        }

        if (multiplier.isZero())
            return ZERO;
        if (this.eq(MIN_VALUE))
            return multiplier.isOdd() ? MIN_VALUE : ZERO;
        if (multiplier.eq(MIN_VALUE))
            return this.isOdd() ? MIN_VALUE : ZERO;

        if (this.isNegative()) {
            if (multiplier.isNegative())
                return this.neg().mul(multiplier.neg());
            else
                return this.neg().mul(multiplier).neg();
        } else if (multiplier.isNegative())
            return this.mul(multiplier.neg()).neg();

        // If both longs are small, use float multiplication
        if (this.lt(TWO_PWR_24) && multiplier.lt(TWO_PWR_24))
            return fromNumber(this.toNumber() * multiplier.toNumber(), this.unsigned);

        // Divide each long into 4 chunks of 16 bits, and then add up 4x4 products.
        // We can skip products that would overflow.

        var a48 = this.high >>> 16;
        var a32 = this.high & 0xFFFF;
        var a16 = this.low >>> 16;
        var a00 = this.low & 0xFFFF;

        var b48 = multiplier.high >>> 16;
        var b32 = multiplier.high & 0xFFFF;
        var b16 = multiplier.low >>> 16;
        var b00 = multiplier.low & 0xFFFF;

        var c48 = 0, c32 = 0, c16 = 0, c00 = 0;
        c00 += a00 * b00;
        c16 += c00 >>> 16;
        c00 &= 0xFFFF;
        c16 += a16 * b00;
        c32 += c16 >>> 16;
        c16 &= 0xFFFF;
        c16 += a00 * b16;
        c32 += c16 >>> 16;
        c16 &= 0xFFFF;
        c32 += a32 * b00;
        c48 += c32 >>> 16;
        c32 &= 0xFFFF;
        c32 += a16 * b16;
        c48 += c32 >>> 16;
        c32 &= 0xFFFF;
        c32 += a00 * b32;
        c48 += c32 >>> 16;
        c32 &= 0xFFFF;
        c48 += a48 * b00 + a32 * b16 + a16 * b32 + a00 * b48;
        c48 &= 0xFFFF;
        return fromBits((c16 << 16) | c00, (c48 << 16) | c32, this.unsigned);
    };

    /**
     * Returns the product of this and the specified Long. This is an alias of {@link Long#multiply}.
     * @function
     * @param {!Long|number|string} multiplier Multiplier
     * @returns {!Long} Product
     */
    LongPrototype.mul = LongPrototype.multiply;

    /**
     * Returns this Long divided by the specified. The result is signed if this Long is signed or
     *  unsigned if this Long is unsigned.
     * @param {!Long|number|string} divisor Divisor
     * @returns {!Long} Quotient
     */
    LongPrototype.divide = function divide(divisor) {
        if (!isLong(divisor))
            divisor = fromValue(divisor);
        if (divisor.isZero())
            throw Error('division by zero');

        // use wasm support if present
        if (wasm) {
            // guard against signed division overflow: the largest
            // negative number / -1 would be 1 larger than the largest
            // positive number, due to two's complement.
            if (!this.unsigned &&
                this.high === -0x80000000 &&
                divisor.low === -1 && divisor.high === -1) {
                // be consistent with non-wasm code path
                return this;
            }
            var low = (this.unsigned ? wasm.div_u : wasm.div_s)(
                this.low,
                this.high,
                divisor.low,
                divisor.high
            );
            return fromBits(low, wasm.get_high(), this.unsigned);
        }

        if (this.isZero())
            return this.unsigned ? UZERO : ZERO;
        var approx, rem, res;
        if (!this.unsigned) {
            // This section is only relevant for signed longs and is derived from the
            // closure library as a whole.
            if (this.eq(MIN_VALUE)) {
                if (divisor.eq(ONE) || divisor.eq(NEG_ONE))
                    return MIN_VALUE;  // recall that -MIN_VALUE == MIN_VALUE
                else if (divisor.eq(MIN_VALUE))
                    return ONE;
                else {
                    // At this point, we have |other| >= 2, so |this/other| < |MIN_VALUE|.
                    var halfThis = this.shr(1);
                    approx = halfThis.div(divisor).shl(1);
                    if (approx.eq(ZERO)) {
                        return divisor.isNegative() ? ONE : NEG_ONE;
                    } else {
                        rem = this.sub(divisor.mul(approx));
                        res = approx.add(rem.div(divisor));
                        return res;
                    }
                }
            } else if (divisor.eq(MIN_VALUE))
                return this.unsigned ? UZERO : ZERO;
            if (this.isNegative()) {
                if (divisor.isNegative())
                    return this.neg().div(divisor.neg());
                return this.neg().div(divisor).neg();
            } else if (divisor.isNegative())
                return this.div(divisor.neg()).neg();
            res = ZERO;
        } else {
            // The algorithm below has not been made for unsigned longs. It's therefore
            // required to take special care of the MSB prior to running it.
            if (!divisor.unsigned)
                divisor = divisor.toUnsigned();
            if (divisor.gt(this))
                return UZERO;
            if (divisor.gt(this.shru(1))) // 15 >>> 1 = 7 ; with divisor = 8 ; true
                return UONE;
            res = UZERO;
        }

        // Repeat the following until the remainder is less than other:  find a
        // floating-point that approximates remainder / other *from below*, add this
        // into the result, and subtract it from the remainder.  It is critical that
        // the approximate value is less than or equal to the real value so that the
        // remainder never becomes negative.
        rem = this;
        while (rem.gte(divisor)) {
            // Approximate the result of division. This may be a little greater or
            // smaller than the actual value.
            approx = Math.max(1, Math.floor(rem.toNumber() / divisor.toNumber()));

            // We will tweak the approximate result by changing it in the 48-th digit or
            // the smallest non-fractional digit, whichever is larger.
            var log2 = Math.ceil(Math.log(approx) / Math.LN2),
                delta = (log2 <= 48) ? 1 : pow_dbl(2, log2 - 48),

            // Decrease the approximation until it is smaller than the remainder.  Note
            // that if it is too large, the product overflows and is negative.
                approxRes = fromNumber(approx),
                approxRem = approxRes.mul(divisor);
            while (approxRem.isNegative() || approxRem.gt(rem)) {
                approx -= delta;
                approxRes = fromNumber(approx, this.unsigned);
                approxRem = approxRes.mul(divisor);
            }

            // We know the answer can't be zero... and actually, zero would cause
            // infinite recursion since we would make no progress.
            if (approxRes.isZero())
                approxRes = ONE;

            res = res.add(approxRes);
            rem = rem.sub(approxRem);
        }
        return res;
    };

    /**
     * Returns this Long divided by the specified. This is an alias of {@link Long#divide}.
     * @function
     * @param {!Long|number|string} divisor Divisor
     * @returns {!Long} Quotient
     */
    LongPrototype.div = LongPrototype.divide;

    /**
     * Returns this Long modulo the specified.
     * @param {!Long|number|string} divisor Divisor
     * @returns {!Long} Remainder
     */
    LongPrototype.modulo = function modulo(divisor) {
        if (!isLong(divisor))
            divisor = fromValue(divisor);

        // use wasm support if present
        if (wasm) {
            var low = (this.unsigned ? wasm.rem_u : wasm.rem_s)(
                this.low,
                this.high,
                divisor.low,
                divisor.high
            );
            return fromBits(low, wasm.get_high(), this.unsigned);
        }

        return this.sub(this.div(divisor).mul(divisor));
    };

    /**
     * Returns this Long modulo the specified. This is an alias of {@link Long#modulo}.
     * @function
     * @param {!Long|number|string} divisor Divisor
     * @returns {!Long} Remainder
     */
    LongPrototype.mod = LongPrototype.modulo;

    /**
     * Returns this Long modulo the specified. This is an alias of {@link Long#modulo}.
     * @function
     * @param {!Long|number|string} divisor Divisor
     * @returns {!Long} Remainder
     */
    LongPrototype.rem = LongPrototype.modulo;

    /**
     * Returns the bitwise NOT of this Long.
     * @returns {!Long}
     */
    LongPrototype.not = function not() {
        return fromBits(~this.low, ~this.high, this.unsigned);
    };

    /**
     * Returns the bitwise AND of this Long and the specified.
     * @param {!Long|number|string} other Other Long
     * @returns {!Long}
     */
    LongPrototype.and = function and(other) {
        if (!isLong(other))
            other = fromValue(other);
        return fromBits(this.low & other.low, this.high & other.high, this.unsigned);
    };

    /**
     * Returns the bitwise OR of this Long and the specified.
     * @param {!Long|number|string} other Other Long
     * @returns {!Long}
     */
    LongPrototype.or = function or(other) {
        if (!isLong(other))
            other = fromValue(other);
        return fromBits(this.low | other.low, this.high | other.high, this.unsigned);
    };

    /**
     * Returns the bitwise XOR of this Long and the given one.
     * @param {!Long|number|string} other Other Long
     * @returns {!Long}
     */
    LongPrototype.xor = function xor(other) {
        if (!isLong(other))
            other = fromValue(other);
        return fromBits(this.low ^ other.low, this.high ^ other.high, this.unsigned);
    };

    /**
     * Returns this Long with bits shifted to the left by the given amount.
     * @param {number|!Long} numBits Number of bits
     * @returns {!Long} Shifted Long
     */
    LongPrototype.shiftLeft = function shiftLeft(numBits) {
        if (isLong(numBits))
            numBits = numBits.toInt();
        if ((numBits &= 63) === 0)
            return this;
        else if (numBits < 32)
            return fromBits(this.low << numBits, (this.high << numBits) | (this.low >>> (32 - numBits)), this.unsigned);
        else
            return fromBits(0, this.low << (numBits - 32), this.unsigned);
    };

    /**
     * Returns this Long with bits shifted to the left by the given amount. This is an alias of {@link Long#shiftLeft}.
     * @function
     * @param {number|!Long} numBits Number of bits
     * @returns {!Long} Shifted Long
     */
    LongPrototype.shl = LongPrototype.shiftLeft;

    /**
     * Returns this Long with bits arithmetically shifted to the right by the given amount.
     * @param {number|!Long} numBits Number of bits
     * @returns {!Long} Shifted Long
     */
    LongPrototype.shiftRight = function shiftRight(numBits) {
        if (isLong(numBits))
            numBits = numBits.toInt();
        if ((numBits &= 63) === 0)
            return this;
        else if (numBits < 32)
            return fromBits((this.low >>> numBits) | (this.high << (32 - numBits)), this.high >> numBits, this.unsigned);
        else
            return fromBits(this.high >> (numBits - 32), this.high >= 0 ? 0 : -1, this.unsigned);
    };

    /**
     * Returns this Long with bits arithmetically shifted to the right by the given amount. This is an alias of {@link Long#shiftRight}.
     * @function
     * @param {number|!Long} numBits Number of bits
     * @returns {!Long} Shifted Long
     */
    LongPrototype.shr = LongPrototype.shiftRight;

    /**
     * Returns this Long with bits logically shifted to the right by the given amount.
     * @param {number|!Long} numBits Number of bits
     * @returns {!Long} Shifted Long
     */
    LongPrototype.shiftRightUnsigned = function shiftRightUnsigned(numBits) {
        if (isLong(numBits))
            numBits = numBits.toInt();
        numBits &= 63;
        if (numBits === 0)
            return this;
        else {
            var high = this.high;
            if (numBits < 32) {
                var low = this.low;
                return fromBits((low >>> numBits) | (high << (32 - numBits)), high >>> numBits, this.unsigned);
            } else if (numBits === 32)
                return fromBits(high, 0, this.unsigned);
            else
                return fromBits(high >>> (numBits - 32), 0, this.unsigned);
        }
    };

    /**
     * Returns this Long with bits logically shifted to the right by the given amount. This is an alias of {@link Long#shiftRightUnsigned}.
     * @function
     * @param {number|!Long} numBits Number of bits
     * @returns {!Long} Shifted Long
     */
    LongPrototype.shru = LongPrototype.shiftRightUnsigned;

    /**
     * Returns this Long with bits logically shifted to the right by the given amount. This is an alias of {@link Long#shiftRightUnsigned}.
     * @function
     * @param {number|!Long} numBits Number of bits
     * @returns {!Long} Shifted Long
     */
    LongPrototype.shr_u = LongPrototype.shiftRightUnsigned;

    /**
     * Converts this Long to signed.
     * @returns {!Long} Signed long
     */
    LongPrototype.toSigned = function toSigned() {
        if (!this.unsigned)
            return this;
        return fromBits(this.low, this.high, false);
    };

    /**
     * Converts this Long to unsigned.
     * @returns {!Long} Unsigned long
     */
    LongPrototype.toUnsigned = function toUnsigned() {
        if (this.unsigned)
            return this;
        return fromBits(this.low, this.high, true);
    };

    /**
     * Converts this Long to its byte representation.
     * @param {boolean=} le Whether little or big endian, defaults to big endian
     * @returns {!Array.<number>} Byte representation
     */
    LongPrototype.toBytes = function toBytes(le) {
        return le ? this.toBytesLE() : this.toBytesBE();
    };

    /**
     * Converts this Long to its little endian byte representation.
     * @returns {!Array.<number>} Little endian byte representation
     */
    LongPrototype.toBytesLE = function toBytesLE() {
        var hi = this.high,
            lo = this.low;
        return [
            lo        & 0xff,
            lo >>>  8 & 0xff,
            lo >>> 16 & 0xff,
            lo >>> 24       ,
            hi        & 0xff,
            hi >>>  8 & 0xff,
            hi >>> 16 & 0xff,
            hi >>> 24
        ];
    };

    /**
     * Converts this Long to its big endian byte representation.
     * @returns {!Array.<number>} Big endian byte representation
     */
    LongPrototype.toBytesBE = function toBytesBE() {
        var hi = this.high,
            lo = this.low;
        return [
            hi >>> 24       ,
            hi >>> 16 & 0xff,
            hi >>>  8 & 0xff,
            hi        & 0xff,
            lo >>> 24       ,
            lo >>> 16 & 0xff,
            lo >>>  8 & 0xff,
            lo        & 0xff
        ];
    };

    /**
     * Creates a Long from its byte representation.
     * @param {!Array.<number>} bytes Byte representation
     * @param {boolean=} unsigned Whether unsigned or not, defaults to signed
     * @param {boolean=} le Whether little or big endian, defaults to big endian
     * @returns {Long} The corresponding Long value
     */
    Long$1.fromBytes = function fromBytes(bytes, unsigned, le) {
        return le ? Long$1.fromBytesLE(bytes, unsigned) : Long$1.fromBytesBE(bytes, unsigned);
    };

    /**
     * Creates a Long from its little endian byte representation.
     * @param {!Array.<number>} bytes Little endian byte representation
     * @param {boolean=} unsigned Whether unsigned or not, defaults to signed
     * @returns {Long} The corresponding Long value
     */
    Long$1.fromBytesLE = function fromBytesLE(bytes, unsigned) {
        return new Long$1(
            bytes[0]       |
            bytes[1] <<  8 |
            bytes[2] << 16 |
            bytes[3] << 24,
            bytes[4]       |
            bytes[5] <<  8 |
            bytes[6] << 16 |
            bytes[7] << 24,
            unsigned
        );
    };

    /**
     * Creates a Long from its big endian byte representation.
     * @param {!Array.<number>} bytes Big endian byte representation
     * @param {boolean=} unsigned Whether unsigned or not, defaults to signed
     * @returns {Long} The corresponding Long value
     */
    Long$1.fromBytesBE = function fromBytesBE(bytes, unsigned) {
        return new Long$1(
            bytes[4] << 24 |
            bytes[5] << 16 |
            bytes[6] <<  8 |
            bytes[7],
            bytes[0] << 24 |
            bytes[1] << 16 |
            bytes[2] <<  8 |
            bytes[3],
            unsigned
        );
    };

    var LongExports = /*#__PURE__*/_mergeNamespaces({
        __proto__: null,
        'default': long
    }, [long]);

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // tslint:disable-next-line
    const Long = 
    // tslint:disable-next-line
    long || LongExports;
    function hexToLong(hex) {
        return Long.fromString(hex, true, 16);
    }
    // Some primes between 2^63 and 2^64 for various uses.
    // Hex 0xc3a5c85c97cb3127
    const k0 = hexToLong('c3a5c85c97cb3127');
    // Hex 0xb492b66fbe98f273
    const k1 = hexToLong('b492b66fbe98f273');
    // Hex 0x9ae16a3b2f90404f
    const k2 = hexToLong('9ae16a3b2f90404f');
    function shiftMix(val) {
        return val.xor(val.shru(47));
    }
    function fetch$2(s, offset, numBytes) {
        const bytes = s.slice(offset, offset + numBytes);
        return Long.fromBytes(Array.from(bytes), true, true);
    }
    function fetch64(s, offset) {
        return fetch$2(s, offset, 8);
    }
    function fetch32(s, offset) {
        return fetch$2(s, offset, 4);
    }
    function rotate64(val, shift) {
        // Avoid shifting by 64: doing so yields an undefined result.
        return shift === 0 ? val : val.shru(shift).or(val.shl(64 - shift));
    }
    function hashLen16(u, v, mul = hexToLong('9ddfea08eb382d69')) {
        // Murmur-inspired hashing.
        let a = u.xor(v).mul(mul);
        a = a.xor(a.shru(47));
        let b = v.xor(a).mul(mul);
        b = b.xor(b.shru(47));
        b = b.mul(mul);
        return b;
    }
    // Return a 16-byte hash for 48 bytes.  Quick and dirty.
    // Callers do best to use "random-looking" values for a and b.
    function weakHashLen32WithSeeds(w, x, y, z, a, b) {
        a = a.add(w);
        b = rotate64(b.add(a).add(z), 21);
        const c = a;
        a = a.add(x);
        a = a.add(y);
        b = b.add(rotate64(a, 44));
        return [a.add(z), b.add(c)];
    }
    function weakHashLen32WithSeedsStr(s, offset, a, b) {
        return weakHashLen32WithSeeds(fetch64(s, offset), fetch64(s, offset + 8), fetch64(s, offset + 16), fetch64(s, offset + 24), a, b);
    }
    function hashLen0to16(s, len = s.length) {
        if (len >= 8) {
            const mul = k2.add(len * 2);
            const a = fetch64(s, 0).add(k2);
            const b = fetch64(s, len - 8);
            const c = rotate64(b, 37).mul(mul).add(a);
            const d = rotate64(a, 25).add(b).mul(mul);
            return hashLen16(c, d, mul);
        }
        if (len >= 4) {
            const mul = k2.add(len * 2);
            const a = fetch32(s, 0);
            return hashLen16(a.shl(3).add(len), fetch32(s, len - 4), mul);
        }
        if (len > 0) {
            const a = s[0];
            const b = s[len >> 1];
            const c = s[len - 1];
            const y = a + (b << 8);
            const z = len + (c << 2);
            return shiftMix(k2.mul(y).xor(k0.mul(z))).mul(k2);
        }
        return k2;
    }
    function hashLen17to32(s, len = s.length) {
        const mul = k2.add(len * 2);
        const a = fetch64(s, 0).mul(k1);
        const b = fetch64(s, 8);
        const c = fetch64(s, len - 8).mul(mul);
        const d = fetch64(s, len - 16).mul(k2);
        return hashLen16(rotate64(a.add(b), 43).add(rotate64(c, 30)).add(d), a.add(rotate64(b.add(k2), 18)).add(c), mul);
    }
    function hashLen33to64(s, len = s.length) {
        const mul = k2.add(len * 2);
        const a = fetch64(s, 0).mul(k2);
        const b = fetch64(s, 8);
        const c = fetch64(s, len - 8).mul(mul);
        const d = fetch64(s, len - 16).mul(k2);
        const y = rotate64(a.add(b), 43).add(rotate64(c, 30)).add(d);
        const z = hashLen16(y, a.add(rotate64(b.add(k2), 18)).add(c), mul);
        const e = fetch64(s, 16).mul(mul);
        const f = fetch64(s, 24);
        const g = y.add(fetch64(s, len - 32)).mul(mul);
        const h = z.add(fetch64(s, len - 24)).mul(mul);
        return hashLen16(rotate64(e.add(f), 43).add(rotate64(g, 30)).add(h), e.add(rotate64(f.add(a), 18)).add(g), mul);
    }
    function fingerPrint64(s, len = s.length) {
        const seed = Long.fromNumber(81, true);
        if (len <= 32) {
            if (len <= 16) {
                return hashLen0to16(s, len);
            }
            else {
                return hashLen17to32(s, len);
            }
        }
        else if (len <= 64) {
            return hashLen33to64(s, len);
        }
        // For strings over 64 bytes we loop.  Internal state consists of
        // 56 bytes: v, w, x, y, and z.
        let x = seed;
        let y = seed.mul(k1).add(113);
        let z = shiftMix(y.mul(k2).add(113)).mul(k2);
        let v = [Long.UZERO, Long.UZERO];
        let w = [Long.UZERO, Long.UZERO];
        x = x.mul(k2).add(fetch64(s, 0));
        let offset = 0;
        // Set end so that after the loop we have 1 to 64 bytes left to process.
        const end = ((len - 1) >> 6) * 64;
        const last64 = end + ((len - 1) & 63) - 63;
        do {
            x = rotate64(x.add(y).add(v[0]).add(fetch64(s, offset + 8)), 37).mul(k1);
            y = rotate64(y.add(v[1]).add(fetch64(s, offset + 48)), 42).mul(k1);
            x = x.xor(w[1]);
            y = y.add(v[0]).add(fetch64(s, offset + 40));
            z = rotate64(z.add(w[0]), 33).mul(k1);
            v = weakHashLen32WithSeedsStr(s, offset, v[1].mul(k1), x.add(w[0]));
            w = weakHashLen32WithSeedsStr(s, offset + 32, z.add(w[1]), y.add(fetch64(s, offset + 16)));
            [z, x] = [x, z];
            offset += 64;
        } while (offset !== end);
        const mul = k1.add(z.and(0xff).shl(1));
        // Point to the last 64 bytes of input.
        offset = last64;
        w[0] = w[0].add((len - 1) & 63);
        v[0] = v[0].add(w[0]);
        w[0] = w[0].add(v[0]);
        x = rotate64(x.add(y).add(v[0]).add(fetch64(s, offset + 8)), 37).mul(mul);
        y = rotate64(y.add(v[1]).add(fetch64(s, offset + 48)), 42).mul(mul);
        x = x.xor(w[1].mul(9));
        y = y.add(v[0].mul(9).add(fetch64(s, offset + 40)));
        z = rotate64(z.add(w[0]), 33).mul(mul);
        v = weakHashLen32WithSeedsStr(s, offset, v[1].mul(mul), x.add(w[0]));
        w = weakHashLen32WithSeedsStr(s, offset + 32, z.add(w[1]), y.add(fetch64(s, offset + 16)));
        [z, x] = [x, z];
        return hashLen16(hashLen16(v[0], w[0], mul).add(shiftMix(y).mul(k0)).add(z), hashLen16(v[1], w[1], mul).add(x), mul);
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Create typed array for scalar value. Used for storing in `DataStorage`.
     */
    function createScalarValue(value, dtype) {
        if (dtype === 'string') {
            return encodeString(value);
        }
        return toTypedArray([value], dtype);
    }
    function noConversionNeeded(a, dtype) {
        return (a instanceof Float32Array && dtype === 'float32') ||
            (a instanceof Int32Array && dtype === 'int32') ||
            (a instanceof Uint8Array && dtype === 'bool');
    }
    function toTypedArray(a, dtype) {
        if (dtype === 'string') {
            throw new Error('Cannot convert a string[] to a TypedArray');
        }
        if (Array.isArray(a)) {
            a = flatten$3(a);
        }
        if (env$1().getBool('DEBUG')) {
            checkConversionForErrors(a, dtype);
        }
        if (noConversionNeeded(a, dtype)) {
            return a;
        }
        if (dtype == null || dtype === 'float32' || dtype === 'complex64') {
            return new Float32Array(a);
        }
        else if (dtype === 'int32') {
            return new Int32Array(a);
        }
        else if (dtype === 'bool') {
            const bool = new Uint8Array(a.length);
            for (let i = 0; i < bool.length; ++i) {
                if (Math.round(a[i]) !== 0) {
                    bool[i] = 1;
                }
            }
            return bool;
        }
        else {
            throw new Error(`Unknown data type ${dtype}`);
        }
    }
    /**
     * Returns the current high-resolution time in milliseconds relative to an
     * arbitrary time in the past. It works across different platforms (node.js,
     * browsers).
     *
     * ```js
     * console.log(tf.util.now());
     * ```
     *
     * @doc {heading: 'Util', namespace: 'util'}
     */
    function now() {
        return env$1().platform.now();
    }
    /**
     * Returns a platform-specific implementation of
     * [`fetch`](https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API).
     *
     * If `fetch` is defined on the global object (`window`, `process`, etc.),
     * `tf.util.fetch` returns that function.
     *
     * If not, `tf.util.fetch` returns a platform-specific solution.
     *
     * ```js
     * const resource = await tf.util.fetch('https://unpkg.com/@tensorflow/tfjs');
     * // handle response
     * ```
     *
     * @doc {heading: 'Util'}
     */
    function fetch$1(path, requestInits) {
        return env$1().platform.fetch(path, requestInits);
    }
    /**
     * Encodes the provided string into bytes using the provided encoding scheme.
     *
     * @param s The string to encode.
     * @param encoding The encoding scheme. Defaults to utf-8.
     *
     * @doc {heading: 'Util'}
     */
    function encodeString(s, encoding = 'utf-8') {
        encoding = encoding || 'utf-8';
        return env$1().platform.encode(s, encoding);
    }
    /**
     * Decodes the provided bytes into a string using the provided encoding scheme.
     * @param bytes The bytes to decode.
     *
     * @param encoding The encoding scheme. Defaults to utf-8.
     *
     * @doc {heading: 'Util'}
     */
    function decodeString(bytes, encoding = 'utf-8') {
        encoding = encoding || 'utf-8';
        return env$1().platform.decode(bytes, encoding);
    }

    var util$2 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        createScalarValue: createScalarValue,
        toTypedArray: toTypedArray,
        now: now,
        fetch: fetch$1,
        encodeString: encodeString,
        decodeString: decodeString,
        shuffle: shuffle,
        shuffleCombo: shuffleCombo,
        clamp: clamp,
        nearestLargerEven: nearestLargerEven,
        swap: swap$1,
        sum: sum$3,
        randUniform: randUniform,
        distSquared: distSquared,
        assert: assert$4,
        assertShapesMatch: assertShapesMatch,
        assertNonNull: assertNonNull,
        flatten: flatten$3,
        sizeFromShape: sizeFromShape,
        isScalarShape: isScalarShape,
        arraysEqual: arraysEqual,
        isInt: isInt,
        tanh: tanh$3,
        sizeToSquarishShape: sizeToSquarishShape,
        createShuffledIndices: createShuffledIndices,
        rightPad: rightPad,
        repeatedTry: repeatedTry,
        inferFromImplicitShape: inferFromImplicitShape,
        parseAxisParam: parseAxisParam,
        squeezeShape: squeezeShape,
        getTypedArrayFromDType: getTypedArrayFromDType,
        getArrayFromDType: getArrayFromDType,
        checkConversionForErrors: checkConversionForErrors,
        isValidDtype: isValidDtype,
        hasEncodingLoss: hasEncodingLoss,
        isTypedArray: isTypedArray$1,
        bytesPerElement: bytesPerElement,
        bytesFromStringArray: bytesFromStringArray,
        isString: isString$2,
        isBoolean: isBoolean$1,
        isNumber: isNumber$3,
        inferDtype: inferDtype,
        isFunction: isFunction$5,
        nearestDivisor: nearestDivisor,
        computeStrides: computeStrides,
        toNestedArray: toNestedArray,
        makeOnesTypedArray: makeOnesTypedArray,
        makeZerosTypedArray: makeZerosTypedArray,
        makeZerosNestedTypedArray: makeZerosNestedTypedArray,
        assertNonNegativeIntegerDimensions: assertNonNegativeIntegerDimensions,
        locToIndex: locToIndex,
        indexToLoc: indexToLoc,
        isPromise: isPromise,
        hexToLong: hexToLong,
        fingerPrint64: fingerPrint64
    });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class Profiler {
        constructor(backendTimer, logger) {
            this.backendTimer = backendTimer;
            this.logger = logger;
            if (logger == null) {
                this.logger = new Logger();
            }
        }
        profileKernel(kernelName, inputs, f) {
            let outputs;
            const holdResultWrapperFn = () => {
                outputs = f();
            };
            let timer;
            const start = now();
            if (this.backendTimer.timerAvailable()) {
                timer = this.backendTimer.time(holdResultWrapperFn);
            }
            else {
                holdResultWrapperFn();
                for (const output of outputs) {
                    output.dataSync();
                }
                timer = Promise.resolve({ kernelMs: now() - start });
            }
            if (env$1().getBool('CHECK_COMPUTATION_FOR_ERRORS')) {
                for (let i = 0; i < outputs.length; i++) {
                    const output = outputs[i];
                    // Dangling promise here because we don't want to propagate up
                    // asynchronicity.
                    output.data().then(tensorVals => {
                        checkComputationForErrors(tensorVals, output.dtype, kernelName);
                    });
                }
            }
            const kernelProfile = {
                kernelName,
                outputs,
                inputs,
                timeMs: timer.then(timing => timing.kernelMs),
                extraInfo: timer.then(timing => timing.getExtraProfileInfo != null ?
                    timing.getExtraProfileInfo() :
                    '')
            };
            return kernelProfile;
        }
        logKernelProfile(kernelProfile) {
            const { kernelName, outputs, timeMs, inputs, extraInfo } = kernelProfile;
            outputs.forEach(result => {
                Promise.all([result.data(), timeMs, extraInfo]).then(valueContainer => {
                    this.logger.logKernelProfile(kernelName, result, valueContainer[0], valueContainer[1], inputs, valueContainer[2]);
                });
            });
        }
    }
    function checkComputationForErrors(vals, dtype, kernelName) {
        if (dtype !== 'float32') {
            // Only floating point computations will generate NaN values
            return false;
        }
        for (let i = 0; i < vals.length; i++) {
            const num = vals[i];
            if (isNaN(num) || !isFinite(num)) {
                // Throwing custom exception so behavior is testable.
                console.warn(`Found ${num} in the result of '${kernelName}'`);
                return true;
            }
        }
        return false;
    }
    class Logger {
        logKernelProfile(name, result, vals, timeMs, inputs, extraInfo) {
            const time = typeof timeMs === 'number' ? rightPad(`${timeMs}ms`, 9) :
                timeMs['error'];
            const paddedName = rightPad(name, 25);
            const rank = result.rank;
            const size = result.size;
            const shape = rightPad(result.shape.toString(), 14);
            let inputShapesDescription = '';
            for (const name in inputs) {
                const input = inputs[name];
                if (input != null) {
                    // The input might be a non-tensor (e.g HTMLImageElement), in which case
                    // we claim the output shape as input shape.
                    const inputShape = input.shape || result.shape;
                    const inputRank = inputShape.length;
                    inputShapesDescription +=
                        `${name}: ${inputRank}D ${inputRank > 0 ? inputShape : ''} `;
                }
            }
            console.log(`%c${paddedName}\t%c${time}\t%c${rank}D ${shape}\t%c${size}\t%c${inputShapesDescription}\t%c${extraInfo}`, 'font-weight:bold', 'color:red', 'color:blue', 'color: orange', 'color: green', 'color: steelblue');
        }
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes a list of TapeNodes that connect x to y, filtering everything else
     * out and preserving the order of the original tape elements.
     *
     * @param tape The tape elements to filter.
     * @param xs The input Tensors.
     * @param y The output Tensor.
     */
    function getFilteredNodesXToY(tape, xs, y) {
        // Forward pass to compute all the nodes and Tensors that are transitively a
        // function of x.
        const tensorsFromX = {};
        const nodesFromX = {};
        for (let i = 0; i < xs.length; i++) {
            tensorsFromX[xs[i].id] = true;
        }
        for (let i = 0; i < tape.length; i++) {
            const node = tape[i];
            const nodeInputs = node.inputs;
            for (const inputName in nodeInputs) {
                const input = nodeInputs[inputName];
                let anyInputFromX = false;
                for (let j = 0; j < xs.length; j++) {
                    if (tensorsFromX[input.id]) {
                        node.outputs.forEach(output => tensorsFromX[output.id] = true);
                        anyInputFromX = true;
                        nodesFromX[node.id] = true;
                        break;
                    }
                }
                if (anyInputFromX) {
                    break;
                }
            }
        }
        // Backward pass to find all of the nodes and Tensors that lead to y.
        const tensorsLeadToY = {};
        tensorsLeadToY[y.id] = true;
        const nodesToY = {};
        for (let i = tape.length - 1; i >= 0; i--) {
            const node = tape[i];
            const nodeInputs = node.inputs;
            // If any of the outputs lead to y, mark all of the inputs as leading to y.
            for (let j = 0; j < node.outputs.length; j++) {
                if (tensorsLeadToY[node.outputs[j].id]) {
                    for (const inputName in nodeInputs) {
                        tensorsLeadToY[nodeInputs[inputName].id] = true;
                        nodesToY[node.id] = true;
                    }
                    break;
                }
            }
        }
        // Return the paths that come from x and lead to y.
        const filteredTape = [];
        for (let i = 0; i < tape.length; i++) {
            const node = tape[i];
            if (nodesFromX[node.id] && nodesToY[node.id]) {
                // Prune the inputs from the node that aren't a function of x.
                const prunedInputs = {};
                for (const inputName in node.inputs) {
                    const nodeInput = node.inputs[inputName];
                    if (tensorsFromX[nodeInput.id]) {
                        prunedInputs[inputName] = nodeInput;
                    }
                }
                // Copy the node and overwrite inputsAndArgs to the pruned version.
                const prunedNode = Object.assign({}, node);
                prunedNode.inputs = prunedInputs;
                prunedNode.outputs = node.outputs;
                filteredTape.push(prunedNode);
            }
        }
        return filteredTape;
    }
    /**
     * Backpropagate gradients through the filtered TapeNodes.
     *
     * @param tensorAccumulatedGradientMap A map of Tensor to its gradient. This map
     * is mutated by this method.
     * @param filteredTape The filtered TapeNodes to backprop through.
     */
    function backpropagateGradients(tensorAccumulatedGradientMap, filteredTape, tidy, add) {
        // Walk the tape backward and keep a map of Tensor to its gradient.
        for (let i = filteredTape.length - 1; i >= 0; i--) {
            const node = filteredTape[i];
            const dys = [];
            node.outputs.forEach(o => {
                const gradTensor = tensorAccumulatedGradientMap[o.id];
                if (gradTensor != null) {
                    dys.push(gradTensor);
                }
                else {
                    // This particular output is not in the back-propagation subgraph, so it
                    // does not affect the final output, thus we put null for its dy.
                    dys.push(null);
                }
            });
            if (node.gradient == null) {
                throw new Error(`Cannot compute gradient: gradient function not found ` +
                    `for ${node.kernelName}.`);
            }
            // Backprop dy through this node and accumulate gradients over the inputs.
            const inputGradients = node.gradient(dys);
            for (const inputName in node.inputs) {
                if (!(inputName in inputGradients)) {
                    throw new Error(`Cannot backprop through input ${inputName}. ` +
                        `Available gradients found: ${Object.keys(inputGradients)}.`);
                }
                // Call the gradient function.
                const dx = tidy(() => inputGradients[inputName]());
                if (dx.dtype !== 'float32') {
                    throw new Error(`Error in gradient for op ${node.kernelName}. The gradient of input ` +
                        `${inputName} must have 'float32' dtype, but has '${dx.dtype}'`);
                }
                const x = node.inputs[inputName];
                if (!arraysEqual(dx.shape, x.shape)) {
                    throw new Error(`Error in gradient for op ${node.kernelName}. The gradient of input ` +
                        `'${inputName}' has shape '${dx.shape}', which does not match ` +
                        `the shape of the input '${x.shape}'`);
                }
                if (tensorAccumulatedGradientMap[x.id] == null) {
                    tensorAccumulatedGradientMap[x.id] = dx;
                }
                else {
                    const curGradient = tensorAccumulatedGradientMap[x.id];
                    tensorAccumulatedGradientMap[x.id] = add(curGradient, dx);
                    curGradient.dispose();
                }
            }
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Maximum number of values before we decide to show ellipsis.
    const FORMAT_LIMIT_NUM_VALS = 20;
    // Number of first and last values to show when displaying a, b,...,y, z.
    const FORMAT_NUM_FIRST_LAST_VALS = 3;
    // Number of significant digits to show.
    const FORMAT_NUM_SIG_DIGITS = 7;
    function tensorToString(vals, shape, dtype, verbose) {
        const strides = computeStrides(shape);
        const padPerCol = computeMaxSizePerColumn(vals, shape, dtype, strides);
        const rank = shape.length;
        const valsLines = subTensorToString(vals, shape, dtype, strides, padPerCol);
        const lines = ['Tensor'];
        if (verbose) {
            lines.push(`  dtype: ${dtype}`);
            lines.push(`  rank: ${rank}`);
            lines.push(`  shape: [${shape}]`);
            lines.push(`  values:`);
        }
        lines.push(valsLines.map(l => '    ' + l).join('\n'));
        return lines.join('\n');
    }
    function computeMaxSizePerColumn(vals, shape, dtype, strides) {
        const n = sizeFromShape(shape);
        const numCols = strides[strides.length - 1];
        const padPerCol = new Array(numCols).fill(0);
        const rank = shape.length;
        const valuesOrTuples = dtype === 'complex64' ? createComplexTuples(vals) : vals;
        if (rank > 1) {
            for (let row = 0; row < n / numCols; row++) {
                const offset = row * numCols;
                for (let j = 0; j < numCols; j++) {
                    padPerCol[j] = Math.max(padPerCol[j], valToString(valuesOrTuples[offset + j], 0, dtype).length);
                }
            }
        }
        return padPerCol;
    }
    function valToString(val, pad, dtype) {
        let valStr;
        if (Array.isArray(val)) {
            valStr = `${parseFloat(val[0].toFixed(FORMAT_NUM_SIG_DIGITS))} + ` +
                `${parseFloat(val[1].toFixed(FORMAT_NUM_SIG_DIGITS))}j`;
        }
        else if (isString$2(val)) {
            valStr = `'${val}'`;
        }
        else if (dtype === 'bool') {
            valStr = boolNumToString(val);
        }
        else {
            valStr = parseFloat(val.toFixed(FORMAT_NUM_SIG_DIGITS)).toString();
        }
        return rightPad(valStr, pad);
    }
    function boolNumToString(v) {
        return v === 0 ? 'false' : 'true';
    }
    function subTensorToString(vals, shape, dtype, strides, padPerCol, isLast = true) {
        const storagePerElement = dtype === 'complex64' ? 2 : 1;
        const size = shape[0];
        const rank = shape.length;
        if (rank === 0) {
            if (dtype === 'complex64') {
                const complexTuple = createComplexTuples(vals);
                return [valToString(complexTuple[0], 0, dtype)];
            }
            if (dtype === 'bool') {
                return [boolNumToString(vals[0])];
            }
            return [vals[0].toString()];
        }
        if (rank === 1) {
            if (size > FORMAT_LIMIT_NUM_VALS) {
                const firstValsSize = FORMAT_NUM_FIRST_LAST_VALS * storagePerElement;
                let firstVals = Array.from(vals.slice(0, firstValsSize));
                let lastVals = Array.from(vals.slice((size - FORMAT_NUM_FIRST_LAST_VALS) * storagePerElement, size * storagePerElement));
                if (dtype === 'complex64') {
                    firstVals = createComplexTuples(firstVals);
                    lastVals = createComplexTuples(lastVals);
                }
                return [
                    '[' +
                        firstVals.map((x, i) => valToString(x, padPerCol[i], dtype))
                            .join(', ') +
                        ', ..., ' +
                        lastVals
                            .map((x, i) => valToString(x, padPerCol[size - FORMAT_NUM_FIRST_LAST_VALS + i], dtype))
                            .join(', ') +
                        ']'
                ];
            }
            const displayVals = dtype === 'complex64' ? createComplexTuples(vals) :
                Array.from(vals);
            return [
                '[' +
                    displayVals.map((x, i) => valToString(x, padPerCol[i], dtype))
                        .join(', ') +
                    ']'
            ];
        }
        // The array is rank 2 or more.
        const subshape = shape.slice(1);
        const substrides = strides.slice(1);
        const stride = strides[0] * storagePerElement;
        const lines = [];
        if (size > FORMAT_LIMIT_NUM_VALS) {
            for (let i = 0; i < FORMAT_NUM_FIRST_LAST_VALS; i++) {
                const start = i * stride;
                const end = start + stride;
                lines.push(...subTensorToString(vals.slice(start, end), subshape, dtype, substrides, padPerCol, false /* isLast */));
            }
            lines.push('...');
            for (let i = size - FORMAT_NUM_FIRST_LAST_VALS; i < size; i++) {
                const start = i * stride;
                const end = start + stride;
                lines.push(...subTensorToString(vals.slice(start, end), subshape, dtype, substrides, padPerCol, i === size - 1 /* isLast */));
            }
        }
        else {
            for (let i = 0; i < size; i++) {
                const start = i * stride;
                const end = start + stride;
                lines.push(...subTensorToString(vals.slice(start, end), subshape, dtype, substrides, padPerCol, i === size - 1 /* isLast */));
            }
        }
        const sep = rank === 2 ? ',' : '';
        lines[0] = '[' + lines[0] + sep;
        for (let i = 1; i < lines.length - 1; i++) {
            lines[i] = ' ' + lines[i] + sep;
        }
        let newLineSep = ',\n';
        for (let i = 2; i < rank; i++) {
            newLineSep += '\n';
        }
        lines[lines.length - 1] =
            ' ' + lines[lines.length - 1] + ']' + (isLast ? '' : newLineSep);
        return lines;
    }
    function createComplexTuples(vals) {
        const complexTuples = [];
        for (let i = 0; i < vals.length; i += 2) {
            complexTuples.push([vals[i], vals[i + 1]]);
        }
        return complexTuples;
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * A mutable object, similar to `tf.Tensor`, that allows users to set values
     * at locations before converting to an immutable `tf.Tensor`.
     *
     * See `tf.buffer` for creating a tensor buffer.
     *
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    class TensorBuffer {
        constructor(shape, dtype, values) {
            this.dtype = dtype;
            this.shape = shape.slice();
            this.size = sizeFromShape(shape);
            if (values != null) {
                const n = values.length;
                assert$4(n === this.size, () => `Length of values '${n}' does not match the size ` +
                    `inferred by the shape '${this.size}'.`);
            }
            if (dtype === 'complex64') {
                throw new Error(`complex64 dtype TensorBuffers are not supported. Please create ` +
                    `a TensorBuffer for the real and imaginary parts separately and ` +
                    `call tf.complex(real, imag).`);
            }
            this.values = values || getArrayFromDType(dtype, this.size);
            this.strides = computeStrides(shape);
        }
        /**
         * Sets a value in the buffer at a given location.
         *
         * @param value The value to set.
         * @param locs  The location indices.
         *
         * @doc {heading: 'Tensors', subheading: 'Creation'}
         */
        set(value, ...locs) {
            if (locs.length === 0) {
                locs = [0];
            }
            assert$4(locs.length === this.rank, () => `The number of provided coordinates (${locs.length}) must ` +
                `match the rank (${this.rank})`);
            const index = this.locToIndex(locs);
            this.values[index] = value;
        }
        /**
         * Returns the value in the buffer at the provided location.
         *
         * @param locs The location indices.
         *
         * @doc {heading: 'Tensors', subheading: 'Creation'}
         */
        get(...locs) {
            if (locs.length === 0) {
                locs = [0];
            }
            let i = 0;
            for (const loc of locs) {
                if (loc < 0 || loc >= this.shape[i]) {
                    const msg = `Requested out of range element at ${locs}. ` +
                        `  Buffer shape=${this.shape}`;
                    throw new Error(msg);
                }
                i++;
            }
            let index = locs[locs.length - 1];
            for (let i = 0; i < locs.length - 1; ++i) {
                index += this.strides[i] * locs[i];
            }
            return this.values[index];
        }
        locToIndex(locs) {
            if (this.rank === 0) {
                return 0;
            }
            else if (this.rank === 1) {
                return locs[0];
            }
            let index = locs[locs.length - 1];
            for (let i = 0; i < locs.length - 1; ++i) {
                index += this.strides[i] * locs[i];
            }
            return index;
        }
        indexToLoc(index) {
            if (this.rank === 0) {
                return [];
            }
            else if (this.rank === 1) {
                return [index];
            }
            const locs = new Array(this.shape.length);
            for (let i = 0; i < locs.length - 1; ++i) {
                locs[i] = Math.floor(index / this.strides[i]);
                index -= locs[i] * this.strides[i];
            }
            locs[locs.length - 1] = index;
            return locs;
        }
        get rank() {
            return this.shape.length;
        }
        /**
         * Creates an immutable `tf.Tensor` object from the buffer.
         *
         * @doc {heading: 'Tensors', subheading: 'Creation'}
         */
        toTensor() {
            return trackerFn().makeTensor(this.values, this.shape, this.dtype);
        }
    }
    // For tracking tensor creation and disposal.
    let trackerFn = null;
    // Used by chaining methods to call into ops.
    let opHandler$1 = null;
    /**
     * An external consumer can register itself as the tensor tracker. This way
     * the Tensor class can notify the tracker for every tensor created and
     * disposed.
     */
    function setTensorTracker(fn) {
        trackerFn = fn;
    }
    /**
     * An external consumer can register itself as the op handler. This way the
     * Tensor class can have chaining methods that call into ops via the op
     * handler.
     */
    function setOpHandler(handler) {
        opHandler$1 = handler;
    }
    /**
     * A `tf.Tensor` object represents an immutable, multidimensional array of
     * numbers that has a shape and a data type.
     *
     * For performance reasons, functions that create tensors do not necessarily
     * perform a copy of the data passed to them (e.g. if the data is passed as a
     * `Float32Array`), and changes to the data will change the tensor. This is not
     * a feature and is not supported. To avoid this behavior, use the tensor before
     * changing the input data or create a copy with `copy = tf.add(yourTensor, 0)`.
     *
     * See `tf.tensor` for details on how to create a `tf.Tensor`.
     *
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    class Tensor {
        constructor(shape, dtype, dataId, id) {
            /** Whether this tensor has been globally kept. */
            this.kept = false;
            this.isDisposedInternal = false;
            this.shape = shape.slice();
            this.dtype = dtype || 'float32';
            this.size = sizeFromShape(shape);
            this.strides = computeStrides(shape);
            this.dataId = dataId;
            this.id = id;
            this.rankType = (this.rank < 5 ? this.rank.toString() : 'higher');
        }
        get rank() {
            return this.shape.length;
        }
        /**
         * Returns a promise of `tf.TensorBuffer` that holds the underlying data.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        async buffer() {
            const vals = await this.data();
            return opHandler$1.buffer(this.shape, this.dtype, vals);
        }
        /**
         * Returns a `tf.TensorBuffer` that holds the underlying data.
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        bufferSync() {
            return opHandler$1.buffer(this.shape, this.dtype, this.dataSync());
        }
        /**
         * Returns the tensor data as a nested array. The transfer of data is done
         * asynchronously.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        async array() {
            const vals = await this.data();
            return toNestedArray(this.shape, vals, this.dtype === 'complex64');
        }
        /**
         * Returns the tensor data as a nested array. The transfer of data is done
         * synchronously.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        arraySync() {
            return toNestedArray(this.shape, this.dataSync(), this.dtype === 'complex64');
        }
        /**
         * Asynchronously downloads the values from the `tf.Tensor`. Returns a
         * promise of `TypedArray` that resolves when the computation has finished.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        async data() {
            this.throwIfDisposed();
            const data = trackerFn().read(this.dataId);
            if (this.dtype === 'string') {
                const bytes = await data;
                try {
                    return bytes.map(b => decodeString(b));
                }
                catch (_a) {
                    throw new Error('Failed to decode the string bytes into utf-8. ' +
                        'To get the original bytes, call tensor.bytes().');
                }
            }
            return data;
        }
        /**
         * Copy the tensor's data to a new GPU resource. Comparing to the `dataSync()`
         * and `data()`, this method prevents data from being downloaded to CPU.
         *
         * For WebGL backend, the data will be stored on a densely packed texture.
         * This means that the texture will use the RGBA channels to store value.
         *
         * @param options:
         *     For WebGL,
         *         - customTexShape: Optional. If set, will use the user defined
         *     texture shape to create the texture.
         *
         * @returns For WebGL backend, a GPUData contains the new texture and
         *     its information.
         *     {
         *        tensorRef: The tensor that is associated with this texture,
         *        texture: WebGLTexture,
         *        texShape: [number, number] // [height, width]
         *     }
         *     Remember to dispose the GPUData after it is used by
         *     `res.tensorRef.dispose()`.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        dataToGPU(options) {
            this.throwIfDisposed();
            return trackerFn().readToGPU(this.dataId, options);
        }
        /**
         * Synchronously downloads the values from the `tf.Tensor`. This blocks the
         * UI thread until the values are ready, which can cause performance issues.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        dataSync() {
            this.throwIfDisposed();
            const data = trackerFn().readSync(this.dataId);
            if (this.dtype === 'string') {
                try {
                    return data.map(b => decodeString(b));
                }
                catch (_a) {
                    throw new Error('Failed to decode the string bytes into utf-8. ' +
                        'To get the original bytes, call tensor.bytes().');
                }
            }
            return data;
        }
        /** Returns the underlying bytes of the tensor's data. */
        async bytes() {
            this.throwIfDisposed();
            const data = await trackerFn().read(this.dataId);
            if (this.dtype === 'string') {
                return data;
            }
            else {
                return new Uint8Array(data.buffer);
            }
        }
        /**
         * Disposes `tf.Tensor` from memory.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        dispose() {
            if (this.isDisposed) {
                return;
            }
            trackerFn().disposeTensor(this);
            this.isDisposedInternal = true;
        }
        get isDisposed() {
            return this.isDisposedInternal;
        }
        throwIfDisposed() {
            if (this.isDisposed) {
                throw new Error(`Tensor is disposed.`);
            }
        }
        /**
         * Prints the `tf.Tensor`. See `tf.print` for details.
         *
         * @param verbose Whether to print verbose information about the tensor,
         *    including dtype and size.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        print(verbose = false) {
            return opHandler$1.print(this, verbose);
        }
        /**
         * Returns a copy of the tensor. See `tf.clone` for details.
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        clone() {
            this.throwIfDisposed();
            return opHandler$1.clone(this);
        }
        /**
         * Returns a human-readable description of the tensor. Useful for logging.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        toString(verbose = false) {
            const vals = this.dataSync();
            return tensorToString(vals, this.shape, this.dtype, verbose);
        }
        cast(dtype) {
            this.throwIfDisposed();
            return opHandler$1.cast(this, dtype);
        }
        variable(trainable = true, name, dtype) {
            this.throwIfDisposed();
            return trackerFn().makeVariable(this, trainable, name, dtype);
        }
    }
    Object.defineProperty(Tensor, Symbol.hasInstance, {
        value: (instance) => {
            // Implementation note: we should use properties of the object that will be
            // defined before the constructor body has finished executing (methods).
            // This is because when this code is transpiled by babel, babel will call
            // classCallCheck before the constructor body is run.
            // See https://github.com/tensorflow/tfjs/issues/3384 for backstory.
            return !!instance && instance.data != null && instance.dataSync != null &&
                instance.throwIfDisposed != null;
        }
    });
    function getGlobalTensorClass() {
        // Use getGlobal so that we can augment the Tensor class across package
        // boundaries becase the node resolution alg may result in different modules
        // being returned for this file depending on the path they are loaded from.
        return getGlobal('Tensor', () => {
            return Tensor;
        });
    }
    // Global side effect. Cache global reference to Tensor class
    getGlobalTensorClass();
    /**
     * A mutable `tf.Tensor`, useful for persisting state, e.g. for training.
     *
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    class Variable extends Tensor {
        constructor(initialValue, trainable, name, tensorId) {
            super(initialValue.shape, initialValue.dtype, initialValue.dataId, tensorId);
            this.trainable = trainable;
            this.name = name;
        }
        /**
         * Assign a new `tf.Tensor` to this variable. The new `tf.Tensor` must have
         * the same shape and dtype as the old `tf.Tensor`.
         *
         * @param newValue New tensor to be assigned to this variable.
         *
         * @doc {heading: 'Tensors', subheading: 'Classes'}
         */
        assign(newValue) {
            if (newValue.dtype !== this.dtype) {
                throw new Error(`dtype of the new value (${newValue.dtype}) and ` +
                    `previous value (${this.dtype}) must match`);
            }
            if (!arraysEqual(newValue.shape, this.shape)) {
                throw new Error(`shape of the new value (${newValue.shape}) and ` +
                    `previous value (${this.shape}) must match`);
            }
            trackerFn().disposeTensor(this);
            this.dataId = newValue.dataId;
            trackerFn().incRef(this, null /* backend */);
        }
        dispose() {
            trackerFn().disposeVariable(this);
            this.isDisposedInternal = true;
        }
    }
    Object.defineProperty(Variable, Symbol.hasInstance, {
        value: (instance) => {
            return instance instanceof Tensor && instance.assign != null &&
                instance.assign instanceof Function;
        }
    });

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    var Rank;
    (function (Rank) {
        Rank["R0"] = "R0";
        Rank["R1"] = "R1";
        Rank["R2"] = "R2";
        Rank["R3"] = "R3";
        Rank["R4"] = "R4";
        Rank["R5"] = "R5";
        Rank["R6"] = "R6";
    })(Rank || (Rank = {}));
    // Looks for upcasting types. Used, for example, in operations with mixed dtype
    // inputs.
    var UpcastInt32AndMap;
    (function (UpcastInt32AndMap) {
        UpcastInt32AndMap["float32"] = "float32";
        UpcastInt32AndMap["int32"] = "int32";
        UpcastInt32AndMap["bool"] = "int32";
        UpcastInt32AndMap["complex64"] = "complex64";
    })(UpcastInt32AndMap || (UpcastInt32AndMap = {}));
    var UpcastBoolAndMap;
    (function (UpcastBoolAndMap) {
        UpcastBoolAndMap["float32"] = "float32";
        UpcastBoolAndMap["int32"] = "int32";
        UpcastBoolAndMap["bool"] = "bool";
        UpcastBoolAndMap["complex64"] = "complex64";
    })(UpcastBoolAndMap || (UpcastBoolAndMap = {}));
    var UpcastFloat32AndMap;
    (function (UpcastFloat32AndMap) {
        UpcastFloat32AndMap["float32"] = "float32";
        UpcastFloat32AndMap["int32"] = "float32";
        UpcastFloat32AndMap["bool"] = "float32";
        UpcastFloat32AndMap["complex64"] = "complex64";
    })(UpcastFloat32AndMap || (UpcastFloat32AndMap = {}));
    var UpcastComplex64AndMap;
    (function (UpcastComplex64AndMap) {
        UpcastComplex64AndMap["float32"] = "complex64";
        UpcastComplex64AndMap["int32"] = "complex64";
        UpcastComplex64AndMap["bool"] = "complex64";
        UpcastComplex64AndMap["complex64"] = "complex64";
    })(UpcastComplex64AndMap || (UpcastComplex64AndMap = {}));
    const upcastTypeMap = {
        'float32': UpcastFloat32AndMap,
        'int32': UpcastInt32AndMap,
        'bool': UpcastBoolAndMap,
        'complex64': UpcastComplex64AndMap
    };
    function upcastType(typeA, typeB) {
        if (typeA === 'string' || typeB === 'string') {
            if (typeA === 'string' && typeB === 'string') {
                return 'string';
            }
            throw new Error(`Can not upcast ${typeA} with ${typeB}`);
        }
        return upcastTypeMap[typeA][typeB];
    }
    /** Returns the output type after summation. */
    function sumOutType(type) {
        return upcastType(type, 'int32');
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function makeTypesMatch(a, b) {
        if (a.dtype === b.dtype) {
            return [a, b];
        }
        const dtype = upcastType(a.dtype, b.dtype);
        return [a.cast(dtype), b.cast(dtype)];
    }
    function assertTypesMatch(a, b) {
        assert$4(a.dtype === b.dtype, () => `The dtypes of the first(${a.dtype}) and` +
            ` second(${b.dtype}) input must match`);
    }
    function isTensorInList(tensor, tensorList) {
        return tensorList.some(x => x.id === tensor.id);
    }
    /**
     * Extracts any `Tensor`s found within the provided object.
     *
     * @param container an object that may be a `Tensor` or may directly contain
     *   `Tensor`s, such as a `Tensor[]` or `{key: Tensor, ...}`. In general it
     *   is safe to pass any object here, except that `Promise`s are not
     *   supported.
     * @returns An array of `Tensors` found within the passed object. If the
     *   argument is simply a `Tensor', a list containing that `Tensor` is
     *   returned. If the object is not a `Tensor` or does not
     *   contain `Tensors`, an empty list is returned.
     */
    function getTensorsInContainer(result) {
        const list = [];
        const seen = new Set();
        walkTensorContainer(result, list, seen);
        return list;
    }
    function walkTensorContainer(container, list, seen) {
        if (container == null) {
            return;
        }
        if (container instanceof Tensor) {
            list.push(container);
            return;
        }
        if (!isIterable$1(container)) {
            return;
        }
        // Iteration over keys works also for arrays.
        const iterable = container;
        for (const k in iterable) {
            const val = iterable[k];
            if (!seen.has(val)) {
                seen.add(val);
                walkTensorContainer(val, list, seen);
            }
        }
    }
    // tslint:disable-next-line:no-any
    function isIterable$1(obj) {
        return Array.isArray(obj) || typeof obj === 'object';
    }

    var tensor_util = /*#__PURE__*/Object.freeze({
        __proto__: null,
        makeTypesMatch: makeTypesMatch,
        assertTypesMatch: assertTypesMatch,
        isTensorInList: isTensorInList,
        getTensorsInContainer: getTensorsInContainer
    });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function isRegisteredKernelInvocation(kernelInvocation) {
        return kernelInvocation.kernelName != null;
    }
    class EngineState {
        constructor() {
            // Public since optimizers will use it.
            this.registeredVariables = {};
            this.nextTapeNodeId = 0;
            this.numBytes = 0;
            this.numTensors = 0;
            this.numStringTensors = 0;
            this.numDataBuffers = 0;
            // Number of nested tf.grad() statements when computing higher-order
            // gradients. E.g. `1` for first-order gradients and `2` for second-order
            // gradients. Used to track if the tape should be removed after a backprop.
            this.gradientDepth = 0;
            // Number of nested kernel calls. When kernel depth is greater than 1, we turn
            // off the tape.
            this.kernelDepth = 0;
            this.scopeStack = [];
            /**
             * Keeps track of the number of data moves during a kernel execution. We
             * maintain a stack since kernels can call other kernels, recursively.
             */
            this.numDataMovesStack = [];
            this.nextScopeId = 0;
            this.tensorInfo = new WeakMap();
            this.profiling = false;
            this.activeProfile = {
                newBytes: 0,
                newTensors: 0,
                peakBytes: 0,
                kernels: [],
                result: null,
                get kernelNames() {
                    return Array.from(new Set(this.kernels.map(k => k.name)));
                }
            };
        }
        dispose() {
            for (const variableName in this.registeredVariables) {
                this.registeredVariables[variableName].dispose();
            }
        }
    }
    class Engine {
        constructor(ENV) {
            this.ENV = ENV;
            this.registry = {};
            this.registryFactory = {};
            this.pendingBackendInitId = 0;
            this.state = new EngineState();
        }
        async ready() {
            if (this.pendingBackendInit != null) {
                return this.pendingBackendInit.then(() => { });
            }
            if (this.backendInstance != null) {
                return;
            }
            const sortedBackends = this.getSortedBackends();
            for (let i = 0; i < sortedBackends.length; i++) {
                const backendName = sortedBackends[i];
                const success = await this.initializeBackend(backendName).success;
                if (success) {
                    await this.setBackend(backendName);
                    return;
                }
            }
            throw new Error(`Could not initialize any backends, all backend initializations ` +
                `failed.`);
        }
        get backend() {
            if (this.pendingBackendInit != null) {
                throw new Error(`Backend '${this.backendName}' has not yet been initialized. Make ` +
                    `sure to await tf.ready() or await tf.setBackend() before calling ` +
                    `other methods`);
            }
            if (this.backendInstance == null) {
                const { name, asyncInit } = this.initializeBackendsAndReturnBest();
                if (asyncInit) {
                    throw new Error(`The highest priority backend '${name}' has not yet been ` +
                        `initialized. Make sure to await tf.ready() or ` +
                        `await tf.setBackend() before calling other methods`);
                }
                this.setBackend(name);
            }
            return this.backendInstance;
        }
        backendNames() {
            return Object.keys(this.registryFactory);
        }
        findBackend(backendName) {
            if (!(backendName in this.registry)) {
                // If the backend hasn't been initialized but we have a registry entry for
                // it, initialize it and return it.
                if (backendName in this.registryFactory) {
                    const { asyncInit } = this.initializeBackend(backendName);
                    if (asyncInit) {
                        // Backend is not ready yet.
                        return null;
                    }
                }
                else {
                    return null;
                }
            }
            return this.registry[backendName];
        }
        findBackendFactory(backendName) {
            if (!(backendName in this.registryFactory)) {
                return null;
            }
            return this.registryFactory[backendName].factory;
        }
        registerBackend(backendName, factory, priority = 1) {
            if (backendName in this.registryFactory) {
                warn(`${backendName} backend was already registered. ` +
                    `Reusing existing backend factory.`);
                return false;
            }
            this.registryFactory[backendName] = { factory, priority };
            return true;
        }
        async setBackend(backendName) {
            if (this.registryFactory[backendName] == null) {
                throw new Error(`Backend name '${backendName}' not found in registry`);
            }
            this.backendName = backendName;
            if (this.registry[backendName] == null) {
                this.backendInstance = null;
                const { success, asyncInit } = this.initializeBackend(backendName);
                const result = asyncInit ? await success : success;
                if (!result) {
                    return false;
                }
            }
            this.backendInstance = this.registry[backendName];
            this.setupRegisteredKernels();
            // Reset the profiler.
            this.profiler = new Profiler(this.backendInstance);
            return true;
        }
        setupRegisteredKernels() {
            const kernels = getKernelsForBackend(this.backendName);
            kernels.forEach(kernel => {
                if (kernel.setupFunc != null) {
                    kernel.setupFunc(this.backendInstance);
                }
            });
        }
        disposeRegisteredKernels(backendName) {
            const kernels = getKernelsForBackend(backendName);
            kernels.forEach(kernel => {
                if (kernel.disposeFunc != null) {
                    kernel.disposeFunc(this.registry[backendName]);
                }
            });
        }
        /**
         * Initializes a backend by looking up the backend name in the factory
         * registry and calling the factory method. Returns a boolean representing
         * whether the initialization of the backend suceeded. Throws an error if
         * there is no backend in the factory registry.
         */
        initializeBackend(backendName) {
            const registryFactoryEntry = this.registryFactory[backendName];
            if (registryFactoryEntry == null) {
                throw new Error(`Cannot initialize backend ${backendName}, no registration found.`);
            }
            try {
                const backend = registryFactoryEntry.factory();
                /* Test if the factory returns a promise.
                Done in a more liberal way than
                previous 'Promise.resolve(backend)===backend'
                as we needed to account for custom Promise
                implementations (e.g. Angular) */
                if (backend && !(backend instanceof KernelBackend) &&
                    typeof backend.then === 'function') {
                    const promiseId = ++this.pendingBackendInitId;
                    const success = backend
                        .then(backendInstance => {
                        // Outdated promise. Another backend was set in the meantime.
                        if (promiseId < this.pendingBackendInitId) {
                            return false;
                        }
                        this.registry[backendName] = backendInstance;
                        this.pendingBackendInit = null;
                        return true;
                    })
                        .catch(err => {
                        // Outdated promise. Another backend was set in the meantime.
                        if (promiseId < this.pendingBackendInitId) {
                            return false;
                        }
                        this.pendingBackendInit = null;
                        warn(`Initialization of backend ${backendName} failed`);
                        warn(err.stack || err.message);
                        return false;
                    });
                    this.pendingBackendInit = success;
                    return { success, asyncInit: true };
                }
                else {
                    this.registry[backendName] = backend;
                    return { success: true, asyncInit: false };
                }
            }
            catch (err) {
                warn(`Initialization of backend ${backendName} failed`);
                warn(err.stack || err.message);
                return { success: false, asyncInit: false };
            }
        }
        removeBackend(backendName) {
            if (!(backendName in this.registryFactory)) {
                throw new Error(`${backendName} backend not found in registry`);
            }
            if (this.backendName === backendName && this.pendingBackendInit != null) {
                // There is a pending promise of the backend we want to remove. Make it
                // obsolete.
                this.pendingBackendInitId++;
            }
            if (backendName in this.registry) {
                this.disposeRegisteredKernels(backendName);
                this.registry[backendName].dispose();
                delete this.registry[backendName];
            }
            delete this.registryFactory[backendName];
            // Unset the backend if it is active.
            if (this.backendName === backendName) {
                this.pendingBackendInit = null;
                this.backendName = null;
                this.backendInstance = null;
            }
        }
        getSortedBackends() {
            if (Object.keys(this.registryFactory).length === 0) {
                throw new Error('No backend found in registry.');
            }
            return Object.keys(this.registryFactory).sort((a, b) => {
                // Highest priority comes first.
                return this.registryFactory[b].priority -
                    this.registryFactory[a].priority;
            });
        }
        initializeBackendsAndReturnBest() {
            const sortedBackends = this.getSortedBackends();
            for (let i = 0; i < sortedBackends.length; i++) {
                const backendName = sortedBackends[i];
                const { success, asyncInit } = this.initializeBackend(backendName);
                if (asyncInit || success) {
                    return { name: backendName, asyncInit };
                }
            }
            throw new Error(`Could not initialize any backends, all backend initializations ` +
                `failed.`);
        }
        moveData(backend, dataId) {
            const info = this.state.tensorInfo.get(dataId);
            const srcBackend = info.backend;
            const values = this.readSync(dataId);
            const refCount = srcBackend.refCount(dataId);
            // Delete the tensor from the old backend and move it to the new
            // backend.
            srcBackend.disposeData(dataId, true);
            info.backend = backend;
            backend.move(dataId, values, info.shape, info.dtype, refCount);
            if (this.shouldCheckForMemLeaks()) {
                // Track the number of moves during a kernel execution to correctly
                // detect memory leaks.
                this.state.numDataMovesStack[this.state.numDataMovesStack.length - 1]++;
            }
        }
        tidy(nameOrFn, fn) {
            let name = null;
            if (fn == null) {
                // Called with only 1 argument.
                if (typeof nameOrFn !== 'function') {
                    throw new Error('Please provide a function to tidy()');
                }
                fn = nameOrFn;
            }
            else {
                // Called with 2 arguments.
                if (typeof nameOrFn !== 'string' && !(nameOrFn instanceof String)) {
                    throw new Error('When calling with two arguments, the first argument ' +
                        'to tidy() must be a string');
                }
                if (typeof fn !== 'function') {
                    throw new Error('When calling with two arguments, the 2nd argument ' +
                        'to tidy() must be a function');
                }
                name = nameOrFn;
                // TODO(nsthorat,smilkov): Do operation logging and performance
                // profiling.
            }
            let result;
            return this.scopedRun(() => this.startScope(name), () => this.endScope(result), () => {
                result = fn();
                if (result instanceof Promise) {
                    console.error('Cannot return a Promise inside of tidy.');
                }
                return result;
            });
        }
        scopedRun(start, end, f) {
            start();
            try {
                const res = f();
                end();
                return res;
            }
            catch (ex) {
                end();
                throw ex;
            }
        }
        nextTensorId() {
            return Engine.nextTensorId++;
        }
        nextVariableId() {
            return Engine.nextVariableId++;
        }
        /**
         * This method is called instead of the public-facing tensor.clone() when
         * saving a tensor for backwards pass. It makes sure to add the clone
         * operation to the tape regardless of being called inside a kernel
         * execution.
         */
        clone(x) {
            const y = ENGINE.runKernel(Identity$1, { x });
            const inputs = { x };
            const grad = (dy) => ({
                x: () => {
                    const dtype = 'float32';
                    const gradInputs = { x: dy };
                    const attrs = { dtype };
                    return ENGINE.runKernel(Cast, gradInputs, 
                    // tslint:disable-next-line: no-unnecessary-type-assertion
                    attrs);
                }
            });
            const saved = [];
            this.addTapeNode(this.state.activeScope.name, inputs, [y], grad, saved, {});
            return y;
        }
        /**
         * Execute a kernel with the given name and return the output tensor.
         *
         * @param kernelName The name of the kernel to execute.
         * @param inputs A map of input names to tensors.
         * @param attrs A map of attribute names to their values. An attribute is a
         *     primitive (non-tensor) input to the kernel.
         * @param inputsToSave A list of tensors, inputs to save for the backprop
         *     computation.
         * @param outputsToSave A list of booleans, specifying which output to save
         *     for the backprop computation. These are booleans since the output
         * tensors are not visible to the user.
         */
        runKernel(kernelName, inputs, attrs) {
            if (this.backendName == null) {
                // backend has not been initialized yet (backend initialization is lazy
                // can be deferred until an op/ kernel is run).
                // The below getter has side effects that will try to initialize the
                // backend and set properties like this.backendName
                // tslint:disable-next-line: no-unused-expression
                this.backend;
            }
            const hasKernel = getKernel(kernelName, this.backendName) != null;
            if (!hasKernel) {
                throw new Error(`Kernel '${kernelName}' not registered for backend '${this.backendName}'`);
            }
            return this.runKernelFunc({ kernelName, inputs, attrs });
        }
        shouldCheckForMemLeaks() {
            return this.ENV.getBool('IS_TEST');
        }
        checkKernelForMemLeak(kernelName, numDataIdsBefore, outInfos) {
            const numDataIdsAfter = this.backend.numDataIds();
            // Count the number of data ids associated with the result of the kernel.
            let numOutputDataIds = 0;
            outInfos.forEach(info => {
                // Complex numbers allocate 3 data ids, one for 'real', one for
                // 'imaginary', and one for the container that holds the former two.
                numOutputDataIds += (info.dtype === 'complex64' ? 3 : 1);
            });
            // Account for the number of moves during kernel execution. A "data move"
            // can happen in the middle of a kernel execution, placing a new (key,value)
            // pair in the data storage. Since data moves have net zero effect (we
            // always remove the data from the old backend), we have to cancel them out
            // when detecting memory leaks.
            const numMoves = this.state.numDataMovesStack[this.state.numDataMovesStack.length - 1];
            const dataIdsLeaked = numDataIdsAfter - numDataIdsBefore - numOutputDataIds - numMoves;
            if (dataIdsLeaked > 0) {
                throw new Error(`Backend '${this.backendName}' has an internal memory leak ` +
                    `(${dataIdsLeaked} data ids) after running '${kernelName}'`);
            }
        }
        /**
         * Internal helper method to execute a kernel Func
         *
         * Use `runKernel` to execute kernels from outside of engine.
         */
        runKernelFunc(kernelParams) {
            let outputs;
            let saved = [];
            const isTapeOn = this.isTapeOn();
            const startingBytecount = this.state.numBytes;
            const startingNumTensors = this.state.numTensors;
            if (this.shouldCheckForMemLeaks()) {
                this.state.numDataMovesStack.push(0);
            }
            let kernelFunc;
            if (this.backendName == null) {
                // backend has not been initialized yet (backend initialization is lazy
                // can be deferred until an op/ kernel is run).
                // The below getter has side effects that will try to initialize the
                // backend and set properties like this.backendName
                // tslint:disable-next-line: no-unused-expression
                this.backend;
            }
            let out;
            const kernelOrScopeName = isRegisteredKernelInvocation(kernelParams) ?
                kernelParams.kernelName :
                this.state.activeScope != null ? this.state.activeScope.name : '';
            // Create the kernelFunc from either a registered kernel OR passed in
            // forward/backward functions (used by custom grad). In this context a
            // kernelFunc wraps a kernel implementation with some bookkeeping.
            if (isRegisteredKernelInvocation(kernelParams)) {
                const { kernelName, inputs, attrs } = kernelParams;
                if (this.backendName == null) {
                    // backend has not been initialized yet (backend initialization is lazy
                    // can be deferred until an op/ kernel is run).
                    // The below getter has side effects that will try to initialize the
                    // backend and set properties like this.backendName
                    // tslint:disable-next-line: no-unused-expression
                    this.backend;
                }
                const kernel = getKernel(kernelName, this.backendName);
                assert$4(kernel != null, () => `Cannot find registered kernel '${kernelName}' for backend '${this.backendName}'`);
                kernelFunc = () => {
                    const numDataIdsBefore = this.backend.numDataIds();
                    out = kernel.kernelFunc({ inputs, attrs, backend: this.backend });
                    const outInfos = Array.isArray(out) ? out : [out];
                    if (this.shouldCheckForMemLeaks()) {
                        this.checkKernelForMemLeak(kernelName, numDataIdsBefore, outInfos);
                    }
                    const outTensors = outInfos.map((outInfo) => {
                        // todo (yassogba) remove this option (Tensor) when node backend
                        // methods have been modularized and they all return tensorInfo.
                        // TensorInfos do not have a rank attribute.
                        if (outInfo.rank != null) {
                            return outInfo;
                        }
                        return this.makeTensorFromTensorInfo(outInfo);
                    });
                    // Save any required inputs and outputs.
                    // Do not save unless we are recording to the tape. Otherwise it would
                    // cause a mem leak since there would be no backprop for these tensors
                    // (which would otherwise dispose them).
                    if (isTapeOn) {
                        const tensorsToSave = this.getTensorsForGradient(kernelName, inputs, outTensors);
                        saved = this.saveTensorsForBackwardMode(tensorsToSave);
                    }
                    return outTensors;
                };
            }
            else {
                const { forwardFunc } = kernelParams;
                // Running a customGrad op.
                const saveFunc = (tensors) => {
                    // Do not save unless we are recording to the tape. Otherwise it would
                    // cause a mem leak since we would never run backprop, which disposes
                    // the kept tensors.
                    if (!isTapeOn) {
                        return;
                    }
                    saved = tensors.map(tensor => this.keep(this.clone(tensor)));
                };
                kernelFunc = () => {
                    const numDataIdsBefore = this.backend.numDataIds();
                    out = this.tidy(() => forwardFunc(this.backend, saveFunc));
                    const outs = (Array.isArray(out) ? out : [out]);
                    if (this.shouldCheckForMemLeaks()) {
                        // Scope name is used to print a more helpful error message if needed.
                        this.checkKernelForMemLeak(kernelOrScopeName, numDataIdsBefore, outs);
                    }
                    return outs;
                };
            }
            //
            // Run the kernelFunc. Optionally profiling it.
            //
            const { inputs, attrs } = kernelParams;
            const backwardsFunc = isRegisteredKernelInvocation(kernelParams) ?
                null :
                kernelParams.backwardsFunc;
            let kernelProfile;
            this.scopedRun(
            // Stop recording to a tape when running a kernel.
            () => this.state.kernelDepth++, () => this.state.kernelDepth--, () => {
                if (!this.ENV.getBool('DEBUG') && !this.state.profiling) {
                    outputs = kernelFunc();
                }
                else {
                    kernelProfile = this.profiler.profileKernel(kernelOrScopeName, inputs, () => kernelFunc());
                    if (this.ENV.getBool('DEBUG')) {
                        this.profiler.logKernelProfile(kernelProfile);
                    }
                    outputs = kernelProfile.outputs;
                }
            });
            if (isTapeOn) {
                this.addTapeNode(kernelOrScopeName, inputs, outputs, backwardsFunc, saved, attrs);
            }
            if (this.state.profiling) {
                this.state.activeProfile.kernels.push({
                    name: kernelOrScopeName,
                    bytesAdded: this.state.numBytes - startingBytecount,
                    totalBytesSnapshot: this.state.numBytes,
                    tensorsAdded: this.state.numTensors - startingNumTensors,
                    totalTensorsSnapshot: this.state.numTensors,
                    inputShapes: Object.keys(inputs).map(key => inputs[key] != null ? inputs[key].shape : null),
                    outputShapes: outputs.map(item => item.shape),
                    kernelTimeMs: kernelProfile.timeMs,
                    extraInfo: kernelProfile.extraInfo
                });
            }
            return (Array.isArray(out) ? outputs : outputs[0]);
        }
        /**
         * Saves tensors used in forward mode for use in backward mode.
         *
         * @param tensors the list of tensors to save.
         */
        saveTensorsForBackwardMode(tensors) {
            const saved = tensors.map(tensor => this.keep(this.clone(tensor)));
            return saved;
        }
        /**
         * Returns a list of tensors to save for a given gradient calculation.
         *
         * @param kernelName name of kernel to look up gradient for.
         * @param inputs a map of input tensors.
         * @param outputs an array of output tensors from forward mode of kernel.
         */
        getTensorsForGradient(kernelName, inputs, outputs) {
            const gradConfig = getGradient(kernelName);
            if (gradConfig != null) {
                const inputsToSave = gradConfig.inputsToSave || [];
                const outputsToSave = gradConfig.outputsToSave || [];
                // If saveAllInputs is true, all inputs will be saved. Otherwise, inputs
                // specified in inputsToSave will be saved.
                let inputTensorsToSave;
                if (gradConfig.saveAllInputs) {
                    assert$4(Array.isArray(inputs), () => 'saveAllInputs is true, expected inputs to be an array.');
                    inputTensorsToSave = Object.keys(inputs).map((key) => inputs[key]);
                }
                else {
                    inputTensorsToSave = inputsToSave.map((inputName) => inputs[inputName]);
                }
                const outputTensorsToSave = outputs.filter((_, i) => outputsToSave[i]);
                return inputTensorsToSave.concat(outputTensorsToSave);
            }
            // We return an empty list rather than throw an error because the kernel we
            // are looking up may not actually be relevant to backproping through the
            // overall function
            //
            // See 'does not error if irrelevant (pruned) ops are missing grads' test
            // in gradients_test.ts for an example.
            return [];
        }
        /**
         * Internal method used by public APIs for tensor creation. Makes a new
         * tensor with the provided shape, dtype and values. It always
         * creates a new data id and writes the values to the underlying backend.
         */
        makeTensor(values, shape, dtype, backend) {
            if (values == null) {
                throw new Error('Values passed to engine.makeTensor() are null');
            }
            dtype = dtype || 'float32';
            backend = backend || this.backend;
            let backendVals = values;
            if (dtype === 'string' && isString$2(values[0])) {
                backendVals = values.map(d => encodeString(d));
            }
            const dataId = backend.write(backendVals, shape, dtype);
            const t = new Tensor(shape, dtype, dataId, this.nextTensorId());
            this.trackTensor(t, backend);
            // Count bytes for string tensors.
            if (dtype === 'string') {
                const info = this.state.tensorInfo.get(dataId);
                const newBytes = bytesFromStringArray(backendVals);
                this.state.numBytes += newBytes - info.bytes;
                info.bytes = newBytes;
            }
            return t;
        }
        /**
         * Internal method used by backends. Makes a new tensor that is a wrapper
         * around an existing data id in TensorInfo. It doesn't create a new data id,
         * only increments the ref count used in memory tracking.
         */
        makeTensorFromTensorInfo(tensorInfo, backend) {
            const { dataId, shape, dtype } = tensorInfo;
            const t = new Tensor(shape, dtype, dataId, this.nextTensorId());
            this.trackTensor(t, backend);
            return t;
        }
        makeVariable(initialValue, trainable = true, name, dtype) {
            name = name || this.nextVariableId().toString();
            if (dtype != null && dtype !== initialValue.dtype) {
                initialValue = initialValue.cast(dtype);
            }
            const v = new Variable(initialValue, trainable, name, this.nextTensorId());
            if (this.state.registeredVariables[v.name] != null) {
                throw new Error(`Variable with name ${v.name} was already registered`);
            }
            this.state.registeredVariables[v.name] = v;
            this.incRef(v, this.backend);
            return v;
        }
        trackTensor(a, backend) {
            this.state.numTensors++;
            if (a.dtype === 'string') {
                this.state.numStringTensors++;
            }
            // Bytes for complex numbers are counted by their components. Bytes for
            // string tensors are counted when writing values.
            let bytes = 0;
            if (a.dtype !== 'complex64' && a.dtype !== 'string') {
                bytes = a.size * bytesPerElement(a.dtype);
            }
            this.state.numBytes += bytes;
            if (!this.state.tensorInfo.has(a.dataId)) {
                this.state.numDataBuffers++;
                this.state.tensorInfo.set(a.dataId, {
                    backend: backend || this.backend,
                    dtype: a.dtype,
                    shape: a.shape,
                    bytes
                });
            }
            if (!(a instanceof Variable)) {
                this.track(a);
            }
        }
        // Track the tensor by dataId and increase the refCount for the dataId in the
        // backend.
        // TODO(pyu10055): This is currently used by makeVariable method, to increase
        // refCount on the backend for the dataId. It can potentially be replaced with
        // Identity op indead of calling backend directly.
        incRef(a, backend) {
            this.trackTensor(a, backend);
            this.backend.incRef(a.dataId);
        }
        removeDataId(dataId, backend) {
            if (this.state.tensorInfo.has(dataId) &&
                this.state.tensorInfo.get(dataId).backend === backend) {
                this.state.tensorInfo.delete(dataId);
                this.state.numDataBuffers--;
            }
        }
        disposeTensor(a) {
            if (!this.state.tensorInfo.has(a.dataId)) {
                return;
            }
            const info = this.state.tensorInfo.get(a.dataId);
            this.state.numTensors--;
            if (a.dtype === 'string') {
                this.state.numStringTensors--;
                this.state.numBytes -= info.bytes;
            }
            // Don't count bytes for complex numbers as they are counted by their
            // components.
            if (a.dtype !== 'complex64' && a.dtype !== 'string') {
                const bytes = a.size * bytesPerElement(a.dtype);
                this.state.numBytes -= bytes;
            }
            // Remove the reference to dataId if backend dispose the data successfully
            if (info.backend.disposeData(a.dataId)) {
                this.removeDataId(a.dataId, info.backend);
            }
            // TODO(nsthorat): Construct an error and save the stack trace for
            // debugging when in debug mode. Creating a stack trace is too expensive
            // to do unconditionally.
        }
        disposeVariables() {
            for (const varName in this.state.registeredVariables) {
                const v = this.state.registeredVariables[varName];
                this.disposeVariable(v);
            }
        }
        disposeVariable(v) {
            this.disposeTensor(v);
            if (this.state.registeredVariables[v.name] != null) {
                delete this.state.registeredVariables[v.name];
            }
        }
        memory() {
            const info = this.backend.memory();
            info.numTensors = this.state.numTensors;
            info.numDataBuffers = this.state.numDataBuffers;
            info.numBytes = this.state.numBytes;
            if (this.state.numStringTensors > 0) {
                info.unreliable = true;
                if (info.reasons == null) {
                    info.reasons = [];
                }
                info.reasons.push('Memory usage by string tensors is approximate ' +
                    '(2 bytes per character)');
            }
            return info;
        }
        async profile(query) {
            this.state.profiling = true;
            const startBytes = this.state.numBytes;
            const startNumTensors = this.state.numTensors;
            this.state.activeProfile.kernels = [];
            this.state.activeProfile.result = await query();
            this.state.profiling = false;
            this.state.activeProfile.peakBytes = Math.max(...this.state.activeProfile.kernels.map(d => d.totalBytesSnapshot));
            this.state.activeProfile.newBytes = this.state.numBytes - startBytes;
            this.state.activeProfile.newTensors =
                this.state.numTensors - startNumTensors;
            for (const kernel of this.state.activeProfile.kernels) {
                kernel.kernelTimeMs = await kernel.kernelTimeMs;
                kernel.extraInfo = await kernel.extraInfo;
            }
            return this.state.activeProfile;
        }
        isTapeOn() {
            return this.state.gradientDepth > 0 && this.state.kernelDepth === 0;
        }
        addTapeNode(kernelName, inputs, outputs, gradientsFunc, saved, attrs) {
            const tapeNode = { id: this.state.nextTapeNodeId++, kernelName, inputs, outputs, saved };
            const gradConfig = getGradient(kernelName);
            if (gradConfig != null) {
                gradientsFunc = gradConfig.gradFunc;
            }
            if (gradientsFunc != null) {
                tapeNode.gradient = (dys) => {
                    // TODO(smilkov): To optimize back-prop, pass dys that are not used in
                    // the backprop graph to the user as null instead of zeros
                    dys = dys.map((dy, i) => {
                        if (dy == null) {
                            const output = outputs[i];
                            const vals = makeZerosTypedArray(output.size, output.dtype);
                            return this.makeTensor(vals, output.shape, output.dtype);
                        }
                        return dy;
                    });
                    // Grad functions of ops with single outputs expect a dy, while ops
                    // with multiple outputs expect dys (array of dy).
                    return gradientsFunc(dys.length > 1 ? dys : dys[0], saved, attrs);
                };
            }
            this.state.activeTape.push(tapeNode);
        }
        keep(result) {
            result.kept = true;
            return result;
        }
        startTape() {
            if (this.state.gradientDepth === 0) {
                this.state.activeTape = [];
            }
            this.state.gradientDepth++;
        }
        endTape() {
            this.state.gradientDepth--;
        }
        /**
         * Start a scope. Use this with endScope() to achieve the same functionality
         * as scope() without the need for a function closure.
         */
        startScope(name) {
            const scopeInfo = {
                track: [],
                name: 'unnamed scope',
                id: this.state.nextScopeId++
            };
            if (name) {
                scopeInfo.name = name;
            }
            this.state.scopeStack.push(scopeInfo);
            this.state.activeScope = scopeInfo;
        }
        /**
         * End a scope. Use this with startScope() to achieve the same functionality
         * as scope() without the need for a function closure.
         */
        endScope(result) {
            const tensorsToTrackInParent = getTensorsInContainer(result);
            const tensorsToTrackInParentSet = new Set(tensorsToTrackInParent.map(t => t.id));
            // Dispose the arrays tracked in this scope.
            for (let i = 0; i < this.state.activeScope.track.length; i++) {
                const tensor = this.state.activeScope.track[i];
                if (!tensor.kept && !tensorsToTrackInParentSet.has(tensor.id)) {
                    tensor.dispose();
                }
            }
            const oldScope = this.state.scopeStack.pop();
            this.state.activeScope = this.state.scopeStack.length === 0 ?
                null :
                this.state.scopeStack[this.state.scopeStack.length - 1];
            // Track the current result in the parent scope.
            tensorsToTrackInParent.forEach(tensor => {
                // Only track the tensor if was allocated in the inner scope and is not
                // globally kept.
                if (!tensor.kept && tensor.scopeId === oldScope.id) {
                    this.track(tensor);
                }
            });
        }
        /**
         * Returns gradients of `f` with respect to each of the `xs`. The gradients
         * returned are of the same length as `xs`, but some might be null if `f`
         * was not a function of that `x`. It also takes optional dy to multiply the
         * gradient, which defaults to `1`.
         */
        gradients(f, xs, dy, allowNoGradients = false) {
            assert$4(xs.length > 0, () => 'gradients() received an empty list of xs.');
            if (dy != null && dy.dtype !== 'float32') {
                throw new Error(`dy must have 'float32' dtype, but has '${dy.dtype}'`);
            }
            const y = this.scopedRun(() => this.startTape(), () => this.endTape(), () => this.tidy('forward', f));
            assert$4(y instanceof Tensor, () => 'The result y returned by f() must be a tensor.');
            // Filter out the nodes that don't connect x => y.
            const filteredTape = getFilteredNodesXToY(this.state.activeTape, xs, y);
            if (!allowNoGradients && filteredTape.length === 0 && xs.length > 0) {
                throw new Error('Cannot compute gradient of y=f(x) with respect to x. Make sure ' +
                    'that the f you passed encloses all operations that lead from x ' +
                    'to y.');
            }
            return this.tidy('backward', () => {
                const accumulatedGradientMap = {};
                accumulatedGradientMap[y.id] = (dy == null) ? ones$2(y.shape) : dy;
                // Backprop gradients through the filtered nodes.
                backpropagateGradients(accumulatedGradientMap, filteredTape, 
                // Pass the tidy function to avoid circular dep with `tape.ts`.
                f => this.tidy(f), 
                // Pass an add function to avoide a circular dep with `tape.ts`.
                add$3);
                const grads = xs.map(x => accumulatedGradientMap[x.id]);
                if (this.state.gradientDepth === 0) {
                    // This means that we are not computing higher-order gradients
                    // and can clean up the tape.
                    this.state.activeTape.forEach(node => {
                        for (const tensor of node.saved) {
                            tensor.dispose();
                        }
                    });
                    this.state.activeTape = null;
                }
                return { value: y, grads };
            });
        }
        customGrad(f) {
            assert$4(isFunction$5(f), () => 'The f passed in customGrad(f) must be a function.');
            return (...inputs) => {
                assert$4(inputs.every(t => t instanceof Tensor), () => 'The args passed in customGrad(f)(x1, x2,...) must all be ' +
                    'tensors');
                let res;
                const inputMap = {};
                inputs.forEach((input, i) => {
                    inputMap[i] = input;
                });
                const forwardFunc = (_, save) => {
                    res = f(...[...inputs, save]);
                    assert$4(res.value instanceof Tensor, () => 'The function f passed in customGrad(f) must return an ' +
                        'object where `obj.value` is a tensor');
                    assert$4(isFunction$5(res.gradFunc), () => 'The function f passed in customGrad(f) must return an ' +
                        'object where `obj.gradFunc` is a function.');
                    return res.value;
                };
                const backwardsFunc = (dy, saved) => {
                    const gradRes = res.gradFunc(dy, saved);
                    const grads = Array.isArray(gradRes) ? gradRes : [gradRes];
                    assert$4(grads.length === inputs.length, () => 'The function f passed in customGrad(f) must return an ' +
                        'object where `obj.gradFunc` is a function that returns ' +
                        'the same number of tensors as inputs passed to f(...).');
                    assert$4(grads.every(t => t instanceof Tensor), () => 'The function f passed in customGrad(f) must return an ' +
                        'object where `obj.gradFunc` is a function that returns ' +
                        'a list of only tensors.');
                    const gradMap = {};
                    grads.forEach((grad, i) => {
                        gradMap[i] = () => grad;
                    });
                    return gradMap;
                };
                return this.runKernelFunc({
                    forwardFunc,
                    backwardsFunc,
                    inputs: inputMap,
                });
            };
        }
        readSync(dataId) {
            // Route the read to the correct backend.
            const info = this.state.tensorInfo.get(dataId);
            return info.backend.readSync(dataId);
        }
        read(dataId) {
            // Route the read to the correct backend.
            const info = this.state.tensorInfo.get(dataId);
            return info.backend.read(dataId);
        }
        readToGPU(dataId, options) {
            // Route the read to the correct backend.
            const info = this.state.tensorInfo.get(dataId);
            return info.backend.readToGPU(dataId, options);
        }
        async time(query) {
            const start = now();
            const timingInfo = await this.backend.time(query);
            timingInfo.wallMs = now() - start;
            return timingInfo;
        }
        /**
         * Tracks a Tensor in the current scope to be automatically cleaned up
         * when the current scope ends, and returns the value.
         *
         * @param result The Tensor to track in the current scope.
         */
        track(result) {
            if (this.state.activeScope != null) {
                result.scopeId = this.state.activeScope.id;
                this.state.activeScope.track.push(result);
            }
            return result;
        }
        get registeredVariables() {
            return this.state.registeredVariables;
        }
        /**
         * Resets the engine state. Removes all backends but does not remove
         * registered backend factories.
         */
        reset() {
            // Make any pending promise obsolete.
            this.pendingBackendInitId++;
            this.state.dispose();
            this.ENV.reset();
            this.state = new EngineState();
            for (const backendName in this.registry) {
                this.disposeRegisteredKernels(backendName);
                this.registry[backendName].dispose();
                delete this.registry[backendName];
            }
            this.backendName = null;
            this.backendInstance = null;
            this.pendingBackendInit = null;
        }
    }
    Engine.nextTensorId = 0;
    Engine.nextVariableId = 0;
    function ones$2(shape) {
        const values = makeOnesTypedArray(sizeFromShape(shape), 'float32');
        return ENGINE.makeTensor(values, shape, 'float32');
    }
    function getOrMakeEngine() {
        const ns = getGlobalNamespace();
        if (ns._tfengine == null) {
            const environment = new Environment(ns);
            ns._tfengine = new Engine(environment);
        }
        setEnvironmentGlobal(ns._tfengine.ENV);
        // Tell the current tensor interface that the global engine is responsible
        // for tracking.
        setTensorTracker(() => ns._tfengine);
        return ns._tfengine;
    }
    const ENGINE = getOrMakeEngine();
    /**
     * A implementation of the add op for use within engine and tape.
     *
     * This allows us to avoid a circular dependency between add.ts and engine.
     * It is exported to be available in tape tests.
     */
    function add$3(a, b) {
        // We duplicate Add here to avoid a circular dependency with add.ts.
        const inputs = { a, b };
        return ENGINE.runKernel(Add$1, inputs);
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // tslint:disable-next-line:no-any
    function _isNavigatorDefined() {
        return typeof navigator !== 'undefined' && navigator != null;
    }
    let isMobileMockValue;
    function mockIsMobile(value) {
        isMobileMockValue = value;
    }
    function isMobile(nav) {
        if (isMobileMockValue !== undefined) {
            return isMobileMockValue;
        }
        if (nav || _isNavigatorDefined()) {
            if (!nav) {
                nav = navigator;
            }
            if (nav.product === 'ReactNative') {
                return true;
            }
            const a = nav.userAgent || nav.vendor ||
                // tslint:disable-next-line:no-any
                (typeof window !== 'undefined' ? window.opera : '');
            // Use `navigator.userAgentData.mobile` as fallback.
            if (!a) {
                // tslint:disable-next-line:no-any
                const navAny = nav;
                return navAny.userAgentData && navAny.userAgentData.mobile;
            }
            // tslint:disable-next-line:max-line-length
            return /(android|bb\d+|meego).+mobile|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows ce|xda|xiino/i
                .test(a) ||
                // tslint:disable-next-line:max-line-length
                /1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i
                    .test(a.substr(0, 4));
        }
        return false;
    }
    function isBrowser() {
        return (typeof window !== 'undefined' && window.document != null) ||
            //@ts-ignore
            (typeof WorkerGlobalScope !== 'undefined');
    }

    var device_util = /*#__PURE__*/Object.freeze({
        __proto__: null,
        mockIsMobile: mockIsMobile,
        isMobile: isMobile,
        isBrowser: isBrowser
    });

    const ENV$3 = env$1();
    /**
     * This file contains environment-related flag registrations.
     */
    /** Whether to enable debug mode. */
    ENV$3.registerFlag('DEBUG', () => false, debugValue => {
        if (debugValue) {
            console.warn('Debugging mode is ON. The output of every math call will ' +
                'be downloaded to CPU and checked for NaNs. ' +
                'This significantly impacts performance.');
        }
    });
    /** Whether we are in a browser (as versus, say, node.js) environment. */
    ENV$3.registerFlag('IS_BROWSER', () => isBrowser());
    /** Whether we are in a browser (as versus, say, node.js) environment. */
    ENV$3.registerFlag('IS_NODE', () => (typeof process !== 'undefined') &&
        (typeof process.versions !== 'undefined') &&
        (typeof process.versions.node !== 'undefined'));
    /** Whether this browser is Chrome. */
    ENV$3.registerFlag('IS_CHROME', () => typeof navigator !== 'undefined' && navigator != null &&
        navigator.userAgent != null && /Chrome/.test(navigator.userAgent) &&
        /Google Inc/.test(navigator.vendor));
    /**
     * True when the environment is "production" where we disable safety checks
     * to gain performance.
     */
    ENV$3.registerFlag('PROD', () => false);
    /**
     * Whether to do sanity checks when inferring a shape from user-provided
     * values, used when creating a new tensor.
     */
    ENV$3.registerFlag('TENSORLIKE_CHECK_SHAPE_CONSISTENCY', () => ENV$3.getBool('DEBUG'));
    /** Whether deprecation warnings are enabled. */
    ENV$3.registerFlag('DEPRECATION_WARNINGS_ENABLED', () => true);
    /** True if running unit tests. */
    ENV$3.registerFlag('IS_TEST', () => false);
    /** Whether to check computation result for errors. */
    ENV$3.registerFlag('CHECK_COMPUTATION_FOR_ERRORS', () => true);
    /** Whether the backend needs to wrap input to imageBitmap. */
    ENV$3.registerFlag('WRAP_TO_IMAGEBITMAP', () => false);
    /** Experimental flag, whether enter compile only phase. */
    ENV$3.registerFlag('ENGINE_COMPILE_ONLY', () => false);

    var lookup = [];
    var revLookup = [];
    var Arr = typeof Uint8Array !== 'undefined' ? Uint8Array : Array;
    var inited = false;
    function init () {
      inited = true;
      var code = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
      for (var i = 0, len = code.length; i < len; ++i) {
        lookup[i] = code[i];
        revLookup[code.charCodeAt(i)] = i;
      }

      revLookup['-'.charCodeAt(0)] = 62;
      revLookup['_'.charCodeAt(0)] = 63;
    }

    function toByteArray (b64) {
      if (!inited) {
        init();
      }
      var i, j, l, tmp, placeHolders, arr;
      var len = b64.length;

      if (len % 4 > 0) {
        throw new Error('Invalid string. Length must be a multiple of 4')
      }

      // the number of equal signs (place holders)
      // if there are two placeholders, than the two characters before it
      // represent one byte
      // if there is only one, then the three characters before it represent 2 bytes
      // this is just a cheap hack to not do indexOf twice
      placeHolders = b64[len - 2] === '=' ? 2 : b64[len - 1] === '=' ? 1 : 0;

      // base64 is 4/3 + up to two characters of the original data
      arr = new Arr(len * 3 / 4 - placeHolders);

      // if there are placeholders, only get up to the last complete 4 chars
      l = placeHolders > 0 ? len - 4 : len;

      var L = 0;

      for (i = 0, j = 0; i < l; i += 4, j += 3) {
        tmp = (revLookup[b64.charCodeAt(i)] << 18) | (revLookup[b64.charCodeAt(i + 1)] << 12) | (revLookup[b64.charCodeAt(i + 2)] << 6) | revLookup[b64.charCodeAt(i + 3)];
        arr[L++] = (tmp >> 16) & 0xFF;
        arr[L++] = (tmp >> 8) & 0xFF;
        arr[L++] = tmp & 0xFF;
      }

      if (placeHolders === 2) {
        tmp = (revLookup[b64.charCodeAt(i)] << 2) | (revLookup[b64.charCodeAt(i + 1)] >> 4);
        arr[L++] = tmp & 0xFF;
      } else if (placeHolders === 1) {
        tmp = (revLookup[b64.charCodeAt(i)] << 10) | (revLookup[b64.charCodeAt(i + 1)] << 4) | (revLookup[b64.charCodeAt(i + 2)] >> 2);
        arr[L++] = (tmp >> 8) & 0xFF;
        arr[L++] = tmp & 0xFF;
      }

      return arr
    }

    function tripletToBase64 (num) {
      return lookup[num >> 18 & 0x3F] + lookup[num >> 12 & 0x3F] + lookup[num >> 6 & 0x3F] + lookup[num & 0x3F]
    }

    function encodeChunk (uint8, start, end) {
      var tmp;
      var output = [];
      for (var i = start; i < end; i += 3) {
        tmp = (uint8[i] << 16) + (uint8[i + 1] << 8) + (uint8[i + 2]);
        output.push(tripletToBase64(tmp));
      }
      return output.join('')
    }

    function fromByteArray (uint8) {
      if (!inited) {
        init();
      }
      var tmp;
      var len = uint8.length;
      var extraBytes = len % 3; // if we have 1 byte left, pad 2 bytes
      var output = '';
      var parts = [];
      var maxChunkLength = 16383; // must be multiple of 3

      // go through the array every three bytes, we'll deal with trailing stuff later
      for (var i = 0, len2 = len - extraBytes; i < len2; i += maxChunkLength) {
        parts.push(encodeChunk(uint8, i, (i + maxChunkLength) > len2 ? len2 : (i + maxChunkLength)));
      }

      // pad the end with zeros, but make sure to not forget the extra bytes
      if (extraBytes === 1) {
        tmp = uint8[len - 1];
        output += lookup[tmp >> 2];
        output += lookup[(tmp << 4) & 0x3F];
        output += '==';
      } else if (extraBytes === 2) {
        tmp = (uint8[len - 2] << 8) + (uint8[len - 1]);
        output += lookup[tmp >> 10];
        output += lookup[(tmp >> 4) & 0x3F];
        output += lookup[(tmp << 2) & 0x3F];
        output += '=';
      }

      parts.push(output);

      return parts.join('')
    }

    function read (buffer, offset, isLE, mLen, nBytes) {
      var e, m;
      var eLen = nBytes * 8 - mLen - 1;
      var eMax = (1 << eLen) - 1;
      var eBias = eMax >> 1;
      var nBits = -7;
      var i = isLE ? (nBytes - 1) : 0;
      var d = isLE ? -1 : 1;
      var s = buffer[offset + i];

      i += d;

      e = s & ((1 << (-nBits)) - 1);
      s >>= (-nBits);
      nBits += eLen;
      for (; nBits > 0; e = e * 256 + buffer[offset + i], i += d, nBits -= 8) {}

      m = e & ((1 << (-nBits)) - 1);
      e >>= (-nBits);
      nBits += mLen;
      for (; nBits > 0; m = m * 256 + buffer[offset + i], i += d, nBits -= 8) {}

      if (e === 0) {
        e = 1 - eBias;
      } else if (e === eMax) {
        return m ? NaN : ((s ? -1 : 1) * Infinity)
      } else {
        m = m + Math.pow(2, mLen);
        e = e - eBias;
      }
      return (s ? -1 : 1) * m * Math.pow(2, e - mLen)
    }

    function write (buffer, value, offset, isLE, mLen, nBytes) {
      var e, m, c;
      var eLen = nBytes * 8 - mLen - 1;
      var eMax = (1 << eLen) - 1;
      var eBias = eMax >> 1;
      var rt = (mLen === 23 ? Math.pow(2, -24) - Math.pow(2, -77) : 0);
      var i = isLE ? 0 : (nBytes - 1);
      var d = isLE ? 1 : -1;
      var s = value < 0 || (value === 0 && 1 / value < 0) ? 1 : 0;

      value = Math.abs(value);

      if (isNaN(value) || value === Infinity) {
        m = isNaN(value) ? 1 : 0;
        e = eMax;
      } else {
        e = Math.floor(Math.log(value) / Math.LN2);
        if (value * (c = Math.pow(2, -e)) < 1) {
          e--;
          c *= 2;
        }
        if (e + eBias >= 1) {
          value += rt / c;
        } else {
          value += rt * Math.pow(2, 1 - eBias);
        }
        if (value * c >= 2) {
          e++;
          c /= 2;
        }

        if (e + eBias >= eMax) {
          m = 0;
          e = eMax;
        } else if (e + eBias >= 1) {
          m = (value * c - 1) * Math.pow(2, mLen);
          e = e + eBias;
        } else {
          m = value * Math.pow(2, eBias - 1) * Math.pow(2, mLen);
          e = 0;
        }
      }

      for (; mLen >= 8; buffer[offset + i] = m & 0xff, i += d, m /= 256, mLen -= 8) {}

      e = (e << mLen) | m;
      eLen += mLen;
      for (; eLen > 0; buffer[offset + i] = e & 0xff, i += d, e /= 256, eLen -= 8) {}

      buffer[offset + i - d] |= s * 128;
    }

    var toString$1 = {}.toString;

    var isArray$3 = Array.isArray || function (arr) {
      return toString$1.call(arr) == '[object Array]';
    };

    var INSPECT_MAX_BYTES = 50;

    /**
     * If `Buffer.TYPED_ARRAY_SUPPORT`:
     *   === true    Use Uint8Array implementation (fastest)
     *   === false   Use Object implementation (most compatible, even IE6)
     *
     * Browsers that support typed arrays are IE 10+, Firefox 4+, Chrome 7+, Safari 5.1+,
     * Opera 11.6+, iOS 4.2+.
     *
     * Due to various browser bugs, sometimes the Object implementation will be used even
     * when the browser supports typed arrays.
     *
     * Note:
     *
     *   - Firefox 4-29 lacks support for adding new properties to `Uint8Array` instances,
     *     See: https://bugzilla.mozilla.org/show_bug.cgi?id=695438.
     *
     *   - Chrome 9-10 is missing the `TypedArray.prototype.subarray` function.
     *
     *   - IE10 has a broken `TypedArray.prototype.subarray` function which returns arrays of
     *     incorrect length in some situations.

     * We detect these buggy browsers and set `Buffer.TYPED_ARRAY_SUPPORT` to `false` so they
     * get the Object implementation, which is slower but behaves correctly.
     */
    Buffer$1.TYPED_ARRAY_SUPPORT = global$1.TYPED_ARRAY_SUPPORT !== undefined
      ? global$1.TYPED_ARRAY_SUPPORT
      : true;

    /*
     * Export kMaxLength after typed array support is determined.
     */
    var _kMaxLength = kMaxLength();

    function kMaxLength () {
      return Buffer$1.TYPED_ARRAY_SUPPORT
        ? 0x7fffffff
        : 0x3fffffff
    }

    function createBuffer (that, length) {
      if (kMaxLength() < length) {
        throw new RangeError('Invalid typed array length')
      }
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        // Return an augmented `Uint8Array` instance, for best performance
        that = new Uint8Array(length);
        that.__proto__ = Buffer$1.prototype;
      } else {
        // Fallback: Return an object instance of the Buffer class
        if (that === null) {
          that = new Buffer$1(length);
        }
        that.length = length;
      }

      return that
    }

    /**
     * The Buffer constructor returns instances of `Uint8Array` that have their
     * prototype changed to `Buffer.prototype`. Furthermore, `Buffer` is a subclass of
     * `Uint8Array`, so the returned instances will have all the node `Buffer` methods
     * and the `Uint8Array` methods. Square bracket notation works as expected -- it
     * returns a single octet.
     *
     * The `Uint8Array` prototype remains unmodified.
     */

    function Buffer$1 (arg, encodingOrOffset, length) {
      if (!Buffer$1.TYPED_ARRAY_SUPPORT && !(this instanceof Buffer$1)) {
        return new Buffer$1(arg, encodingOrOffset, length)
      }

      // Common case.
      if (typeof arg === 'number') {
        if (typeof encodingOrOffset === 'string') {
          throw new Error(
            'If encoding is specified then the first argument must be a string'
          )
        }
        return allocUnsafe(this, arg)
      }
      return from(this, arg, encodingOrOffset, length)
    }

    Buffer$1.poolSize = 8192; // not used by this implementation

    // TODO: Legacy, not needed anymore. Remove in next major version.
    Buffer$1._augment = function (arr) {
      arr.__proto__ = Buffer$1.prototype;
      return arr
    };

    function from (that, value, encodingOrOffset, length) {
      if (typeof value === 'number') {
        throw new TypeError('"value" argument must not be a number')
      }

      if (typeof ArrayBuffer !== 'undefined' && value instanceof ArrayBuffer) {
        return fromArrayBuffer(that, value, encodingOrOffset, length)
      }

      if (typeof value === 'string') {
        return fromString(that, value, encodingOrOffset)
      }

      return fromObject(that, value)
    }

    /**
     * Functionally equivalent to Buffer(arg, encoding) but throws a TypeError
     * if value is a number.
     * Buffer.from(str[, encoding])
     * Buffer.from(array)
     * Buffer.from(buffer)
     * Buffer.from(arrayBuffer[, byteOffset[, length]])
     **/
    Buffer$1.from = function (value, encodingOrOffset, length) {
      return from(null, value, encodingOrOffset, length)
    };

    if (Buffer$1.TYPED_ARRAY_SUPPORT) {
      Buffer$1.prototype.__proto__ = Uint8Array.prototype;
      Buffer$1.__proto__ = Uint8Array;
      if (typeof Symbol !== 'undefined' && Symbol.species &&
          Buffer$1[Symbol.species] === Buffer$1) ;
    }

    function assertSize (size) {
      if (typeof size !== 'number') {
        throw new TypeError('"size" argument must be a number')
      } else if (size < 0) {
        throw new RangeError('"size" argument must not be negative')
      }
    }

    function alloc (that, size, fill, encoding) {
      assertSize(size);
      if (size <= 0) {
        return createBuffer(that, size)
      }
      if (fill !== undefined) {
        // Only pay attention to encoding if it's a string. This
        // prevents accidentally sending in a number that would
        // be interpretted as a start offset.
        return typeof encoding === 'string'
          ? createBuffer(that, size).fill(fill, encoding)
          : createBuffer(that, size).fill(fill)
      }
      return createBuffer(that, size)
    }

    /**
     * Creates a new filled Buffer instance.
     * alloc(size[, fill[, encoding]])
     **/
    Buffer$1.alloc = function (size, fill, encoding) {
      return alloc(null, size, fill, encoding)
    };

    function allocUnsafe (that, size) {
      assertSize(size);
      that = createBuffer(that, size < 0 ? 0 : checked(size) | 0);
      if (!Buffer$1.TYPED_ARRAY_SUPPORT) {
        for (var i = 0; i < size; ++i) {
          that[i] = 0;
        }
      }
      return that
    }

    /**
     * Equivalent to Buffer(num), by default creates a non-zero-filled Buffer instance.
     * */
    Buffer$1.allocUnsafe = function (size) {
      return allocUnsafe(null, size)
    };
    /**
     * Equivalent to SlowBuffer(num), by default creates a non-zero-filled Buffer instance.
     */
    Buffer$1.allocUnsafeSlow = function (size) {
      return allocUnsafe(null, size)
    };

    function fromString (that, string, encoding) {
      if (typeof encoding !== 'string' || encoding === '') {
        encoding = 'utf8';
      }

      if (!Buffer$1.isEncoding(encoding)) {
        throw new TypeError('"encoding" must be a valid string encoding')
      }

      var length = byteLength(string, encoding) | 0;
      that = createBuffer(that, length);

      var actual = that.write(string, encoding);

      if (actual !== length) {
        // Writing a hex string, for example, that contains invalid characters will
        // cause everything after the first invalid character to be ignored. (e.g.
        // 'abxxcd' will be treated as 'ab')
        that = that.slice(0, actual);
      }

      return that
    }

    function fromArrayLike (that, array) {
      var length = array.length < 0 ? 0 : checked(array.length) | 0;
      that = createBuffer(that, length);
      for (var i = 0; i < length; i += 1) {
        that[i] = array[i] & 255;
      }
      return that
    }

    function fromArrayBuffer (that, array, byteOffset, length) {
      array.byteLength; // this throws if `array` is not a valid ArrayBuffer

      if (byteOffset < 0 || array.byteLength < byteOffset) {
        throw new RangeError('\'offset\' is out of bounds')
      }

      if (array.byteLength < byteOffset + (length || 0)) {
        throw new RangeError('\'length\' is out of bounds')
      }

      if (byteOffset === undefined && length === undefined) {
        array = new Uint8Array(array);
      } else if (length === undefined) {
        array = new Uint8Array(array, byteOffset);
      } else {
        array = new Uint8Array(array, byteOffset, length);
      }

      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        // Return an augmented `Uint8Array` instance, for best performance
        that = array;
        that.__proto__ = Buffer$1.prototype;
      } else {
        // Fallback: Return an object instance of the Buffer class
        that = fromArrayLike(that, array);
      }
      return that
    }

    function fromObject (that, obj) {
      if (internalIsBuffer(obj)) {
        var len = checked(obj.length) | 0;
        that = createBuffer(that, len);

        if (that.length === 0) {
          return that
        }

        obj.copy(that, 0, 0, len);
        return that
      }

      if (obj) {
        if ((typeof ArrayBuffer !== 'undefined' &&
            obj.buffer instanceof ArrayBuffer) || 'length' in obj) {
          if (typeof obj.length !== 'number' || isnan(obj.length)) {
            return createBuffer(that, 0)
          }
          return fromArrayLike(that, obj)
        }

        if (obj.type === 'Buffer' && isArray$3(obj.data)) {
          return fromArrayLike(that, obj.data)
        }
      }

      throw new TypeError('First argument must be a string, Buffer, ArrayBuffer, Array, or array-like object.')
    }

    function checked (length) {
      // Note: cannot use `length < kMaxLength()` here because that fails when
      // length is NaN (which is otherwise coerced to zero.)
      if (length >= kMaxLength()) {
        throw new RangeError('Attempt to allocate Buffer larger than maximum ' +
                             'size: 0x' + kMaxLength().toString(16) + ' bytes')
      }
      return length | 0
    }

    function SlowBuffer (length) {
      if (+length != length) { // eslint-disable-line eqeqeq
        length = 0;
      }
      return Buffer$1.alloc(+length)
    }
    Buffer$1.isBuffer = isBuffer$2;
    function internalIsBuffer (b) {
      return !!(b != null && b._isBuffer)
    }

    Buffer$1.compare = function compare (a, b) {
      if (!internalIsBuffer(a) || !internalIsBuffer(b)) {
        throw new TypeError('Arguments must be Buffers')
      }

      if (a === b) return 0

      var x = a.length;
      var y = b.length;

      for (var i = 0, len = Math.min(x, y); i < len; ++i) {
        if (a[i] !== b[i]) {
          x = a[i];
          y = b[i];
          break
        }
      }

      if (x < y) return -1
      if (y < x) return 1
      return 0
    };

    Buffer$1.isEncoding = function isEncoding (encoding) {
      switch (String(encoding).toLowerCase()) {
        case 'hex':
        case 'utf8':
        case 'utf-8':
        case 'ascii':
        case 'latin1':
        case 'binary':
        case 'base64':
        case 'ucs2':
        case 'ucs-2':
        case 'utf16le':
        case 'utf-16le':
          return true
        default:
          return false
      }
    };

    Buffer$1.concat = function concat (list, length) {
      if (!isArray$3(list)) {
        throw new TypeError('"list" argument must be an Array of Buffers')
      }

      if (list.length === 0) {
        return Buffer$1.alloc(0)
      }

      var i;
      if (length === undefined) {
        length = 0;
        for (i = 0; i < list.length; ++i) {
          length += list[i].length;
        }
      }

      var buffer = Buffer$1.allocUnsafe(length);
      var pos = 0;
      for (i = 0; i < list.length; ++i) {
        var buf = list[i];
        if (!internalIsBuffer(buf)) {
          throw new TypeError('"list" argument must be an Array of Buffers')
        }
        buf.copy(buffer, pos);
        pos += buf.length;
      }
      return buffer
    };

    function byteLength (string, encoding) {
      if (internalIsBuffer(string)) {
        return string.length
      }
      if (typeof ArrayBuffer !== 'undefined' && typeof ArrayBuffer.isView === 'function' &&
          (ArrayBuffer.isView(string) || string instanceof ArrayBuffer)) {
        return string.byteLength
      }
      if (typeof string !== 'string') {
        string = '' + string;
      }

      var len = string.length;
      if (len === 0) return 0

      // Use a for loop to avoid recursion
      var loweredCase = false;
      for (;;) {
        switch (encoding) {
          case 'ascii':
          case 'latin1':
          case 'binary':
            return len
          case 'utf8':
          case 'utf-8':
          case undefined:
            return utf8ToBytes(string).length
          case 'ucs2':
          case 'ucs-2':
          case 'utf16le':
          case 'utf-16le':
            return len * 2
          case 'hex':
            return len >>> 1
          case 'base64':
            return base64ToBytes(string).length
          default:
            if (loweredCase) return utf8ToBytes(string).length // assume utf8
            encoding = ('' + encoding).toLowerCase();
            loweredCase = true;
        }
      }
    }
    Buffer$1.byteLength = byteLength;

    function slowToString (encoding, start, end) {
      var loweredCase = false;

      // No need to verify that "this.length <= MAX_UINT32" since it's a read-only
      // property of a typed array.

      // This behaves neither like String nor Uint8Array in that we set start/end
      // to their upper/lower bounds if the value passed is out of range.
      // undefined is handled specially as per ECMA-262 6th Edition,
      // Section 13.3.3.7 Runtime Semantics: KeyedBindingInitialization.
      if (start === undefined || start < 0) {
        start = 0;
      }
      // Return early if start > this.length. Done here to prevent potential uint32
      // coercion fail below.
      if (start > this.length) {
        return ''
      }

      if (end === undefined || end > this.length) {
        end = this.length;
      }

      if (end <= 0) {
        return ''
      }

      // Force coersion to uint32. This will also coerce falsey/NaN values to 0.
      end >>>= 0;
      start >>>= 0;

      if (end <= start) {
        return ''
      }

      if (!encoding) encoding = 'utf8';

      while (true) {
        switch (encoding) {
          case 'hex':
            return hexSlice(this, start, end)

          case 'utf8':
          case 'utf-8':
            return utf8Slice(this, start, end)

          case 'ascii':
            return asciiSlice(this, start, end)

          case 'latin1':
          case 'binary':
            return latin1Slice(this, start, end)

          case 'base64':
            return base64Slice(this, start, end)

          case 'ucs2':
          case 'ucs-2':
          case 'utf16le':
          case 'utf-16le':
            return utf16leSlice(this, start, end)

          default:
            if (loweredCase) throw new TypeError('Unknown encoding: ' + encoding)
            encoding = (encoding + '').toLowerCase();
            loweredCase = true;
        }
      }
    }

    // The property is used by `Buffer.isBuffer` and `is-buffer` (in Safari 5-7) to detect
    // Buffer instances.
    Buffer$1.prototype._isBuffer = true;

    function swap (b, n, m) {
      var i = b[n];
      b[n] = b[m];
      b[m] = i;
    }

    Buffer$1.prototype.swap16 = function swap16 () {
      var len = this.length;
      if (len % 2 !== 0) {
        throw new RangeError('Buffer size must be a multiple of 16-bits')
      }
      for (var i = 0; i < len; i += 2) {
        swap(this, i, i + 1);
      }
      return this
    };

    Buffer$1.prototype.swap32 = function swap32 () {
      var len = this.length;
      if (len % 4 !== 0) {
        throw new RangeError('Buffer size must be a multiple of 32-bits')
      }
      for (var i = 0; i < len; i += 4) {
        swap(this, i, i + 3);
        swap(this, i + 1, i + 2);
      }
      return this
    };

    Buffer$1.prototype.swap64 = function swap64 () {
      var len = this.length;
      if (len % 8 !== 0) {
        throw new RangeError('Buffer size must be a multiple of 64-bits')
      }
      for (var i = 0; i < len; i += 8) {
        swap(this, i, i + 7);
        swap(this, i + 1, i + 6);
        swap(this, i + 2, i + 5);
        swap(this, i + 3, i + 4);
      }
      return this
    };

    Buffer$1.prototype.toString = function toString () {
      var length = this.length | 0;
      if (length === 0) return ''
      if (arguments.length === 0) return utf8Slice(this, 0, length)
      return slowToString.apply(this, arguments)
    };

    Buffer$1.prototype.equals = function equals (b) {
      if (!internalIsBuffer(b)) throw new TypeError('Argument must be a Buffer')
      if (this === b) return true
      return Buffer$1.compare(this, b) === 0
    };

    Buffer$1.prototype.inspect = function inspect () {
      var str = '';
      var max = INSPECT_MAX_BYTES;
      if (this.length > 0) {
        str = this.toString('hex', 0, max).match(/.{2}/g).join(' ');
        if (this.length > max) str += ' ... ';
      }
      return '<Buffer ' + str + '>'
    };

    Buffer$1.prototype.compare = function compare (target, start, end, thisStart, thisEnd) {
      if (!internalIsBuffer(target)) {
        throw new TypeError('Argument must be a Buffer')
      }

      if (start === undefined) {
        start = 0;
      }
      if (end === undefined) {
        end = target ? target.length : 0;
      }
      if (thisStart === undefined) {
        thisStart = 0;
      }
      if (thisEnd === undefined) {
        thisEnd = this.length;
      }

      if (start < 0 || end > target.length || thisStart < 0 || thisEnd > this.length) {
        throw new RangeError('out of range index')
      }

      if (thisStart >= thisEnd && start >= end) {
        return 0
      }
      if (thisStart >= thisEnd) {
        return -1
      }
      if (start >= end) {
        return 1
      }

      start >>>= 0;
      end >>>= 0;
      thisStart >>>= 0;
      thisEnd >>>= 0;

      if (this === target) return 0

      var x = thisEnd - thisStart;
      var y = end - start;
      var len = Math.min(x, y);

      var thisCopy = this.slice(thisStart, thisEnd);
      var targetCopy = target.slice(start, end);

      for (var i = 0; i < len; ++i) {
        if (thisCopy[i] !== targetCopy[i]) {
          x = thisCopy[i];
          y = targetCopy[i];
          break
        }
      }

      if (x < y) return -1
      if (y < x) return 1
      return 0
    };

    // Finds either the first index of `val` in `buffer` at offset >= `byteOffset`,
    // OR the last index of `val` in `buffer` at offset <= `byteOffset`.
    //
    // Arguments:
    // - buffer - a Buffer to search
    // - val - a string, Buffer, or number
    // - byteOffset - an index into `buffer`; will be clamped to an int32
    // - encoding - an optional encoding, relevant is val is a string
    // - dir - true for indexOf, false for lastIndexOf
    function bidirectionalIndexOf (buffer, val, byteOffset, encoding, dir) {
      // Empty buffer means no match
      if (buffer.length === 0) return -1

      // Normalize byteOffset
      if (typeof byteOffset === 'string') {
        encoding = byteOffset;
        byteOffset = 0;
      } else if (byteOffset > 0x7fffffff) {
        byteOffset = 0x7fffffff;
      } else if (byteOffset < -0x80000000) {
        byteOffset = -0x80000000;
      }
      byteOffset = +byteOffset;  // Coerce to Number.
      if (isNaN(byteOffset)) {
        // byteOffset: it it's undefined, null, NaN, "foo", etc, search whole buffer
        byteOffset = dir ? 0 : (buffer.length - 1);
      }

      // Normalize byteOffset: negative offsets start from the end of the buffer
      if (byteOffset < 0) byteOffset = buffer.length + byteOffset;
      if (byteOffset >= buffer.length) {
        if (dir) return -1
        else byteOffset = buffer.length - 1;
      } else if (byteOffset < 0) {
        if (dir) byteOffset = 0;
        else return -1
      }

      // Normalize val
      if (typeof val === 'string') {
        val = Buffer$1.from(val, encoding);
      }

      // Finally, search either indexOf (if dir is true) or lastIndexOf
      if (internalIsBuffer(val)) {
        // Special case: looking for empty string/buffer always fails
        if (val.length === 0) {
          return -1
        }
        return arrayIndexOf(buffer, val, byteOffset, encoding, dir)
      } else if (typeof val === 'number') {
        val = val & 0xFF; // Search for a byte value [0-255]
        if (Buffer$1.TYPED_ARRAY_SUPPORT &&
            typeof Uint8Array.prototype.indexOf === 'function') {
          if (dir) {
            return Uint8Array.prototype.indexOf.call(buffer, val, byteOffset)
          } else {
            return Uint8Array.prototype.lastIndexOf.call(buffer, val, byteOffset)
          }
        }
        return arrayIndexOf(buffer, [ val ], byteOffset, encoding, dir)
      }

      throw new TypeError('val must be string, number or Buffer')
    }

    function arrayIndexOf (arr, val, byteOffset, encoding, dir) {
      var indexSize = 1;
      var arrLength = arr.length;
      var valLength = val.length;

      if (encoding !== undefined) {
        encoding = String(encoding).toLowerCase();
        if (encoding === 'ucs2' || encoding === 'ucs-2' ||
            encoding === 'utf16le' || encoding === 'utf-16le') {
          if (arr.length < 2 || val.length < 2) {
            return -1
          }
          indexSize = 2;
          arrLength /= 2;
          valLength /= 2;
          byteOffset /= 2;
        }
      }

      function read (buf, i) {
        if (indexSize === 1) {
          return buf[i]
        } else {
          return buf.readUInt16BE(i * indexSize)
        }
      }

      var i;
      if (dir) {
        var foundIndex = -1;
        for (i = byteOffset; i < arrLength; i++) {
          if (read(arr, i) === read(val, foundIndex === -1 ? 0 : i - foundIndex)) {
            if (foundIndex === -1) foundIndex = i;
            if (i - foundIndex + 1 === valLength) return foundIndex * indexSize
          } else {
            if (foundIndex !== -1) i -= i - foundIndex;
            foundIndex = -1;
          }
        }
      } else {
        if (byteOffset + valLength > arrLength) byteOffset = arrLength - valLength;
        for (i = byteOffset; i >= 0; i--) {
          var found = true;
          for (var j = 0; j < valLength; j++) {
            if (read(arr, i + j) !== read(val, j)) {
              found = false;
              break
            }
          }
          if (found) return i
        }
      }

      return -1
    }

    Buffer$1.prototype.includes = function includes (val, byteOffset, encoding) {
      return this.indexOf(val, byteOffset, encoding) !== -1
    };

    Buffer$1.prototype.indexOf = function indexOf (val, byteOffset, encoding) {
      return bidirectionalIndexOf(this, val, byteOffset, encoding, true)
    };

    Buffer$1.prototype.lastIndexOf = function lastIndexOf (val, byteOffset, encoding) {
      return bidirectionalIndexOf(this, val, byteOffset, encoding, false)
    };

    function hexWrite (buf, string, offset, length) {
      offset = Number(offset) || 0;
      var remaining = buf.length - offset;
      if (!length) {
        length = remaining;
      } else {
        length = Number(length);
        if (length > remaining) {
          length = remaining;
        }
      }

      // must be an even number of digits
      var strLen = string.length;
      if (strLen % 2 !== 0) throw new TypeError('Invalid hex string')

      if (length > strLen / 2) {
        length = strLen / 2;
      }
      for (var i = 0; i < length; ++i) {
        var parsed = parseInt(string.substr(i * 2, 2), 16);
        if (isNaN(parsed)) return i
        buf[offset + i] = parsed;
      }
      return i
    }

    function utf8Write (buf, string, offset, length) {
      return blitBuffer(utf8ToBytes(string, buf.length - offset), buf, offset, length)
    }

    function asciiWrite (buf, string, offset, length) {
      return blitBuffer(asciiToBytes(string), buf, offset, length)
    }

    function latin1Write (buf, string, offset, length) {
      return asciiWrite(buf, string, offset, length)
    }

    function base64Write (buf, string, offset, length) {
      return blitBuffer(base64ToBytes(string), buf, offset, length)
    }

    function ucs2Write (buf, string, offset, length) {
      return blitBuffer(utf16leToBytes(string, buf.length - offset), buf, offset, length)
    }

    Buffer$1.prototype.write = function write (string, offset, length, encoding) {
      // Buffer#write(string)
      if (offset === undefined) {
        encoding = 'utf8';
        length = this.length;
        offset = 0;
      // Buffer#write(string, encoding)
      } else if (length === undefined && typeof offset === 'string') {
        encoding = offset;
        length = this.length;
        offset = 0;
      // Buffer#write(string, offset[, length][, encoding])
      } else if (isFinite(offset)) {
        offset = offset | 0;
        if (isFinite(length)) {
          length = length | 0;
          if (encoding === undefined) encoding = 'utf8';
        } else {
          encoding = length;
          length = undefined;
        }
      // legacy write(string, encoding, offset, length) - remove in v0.13
      } else {
        throw new Error(
          'Buffer.write(string, encoding, offset[, length]) is no longer supported'
        )
      }

      var remaining = this.length - offset;
      if (length === undefined || length > remaining) length = remaining;

      if ((string.length > 0 && (length < 0 || offset < 0)) || offset > this.length) {
        throw new RangeError('Attempt to write outside buffer bounds')
      }

      if (!encoding) encoding = 'utf8';

      var loweredCase = false;
      for (;;) {
        switch (encoding) {
          case 'hex':
            return hexWrite(this, string, offset, length)

          case 'utf8':
          case 'utf-8':
            return utf8Write(this, string, offset, length)

          case 'ascii':
            return asciiWrite(this, string, offset, length)

          case 'latin1':
          case 'binary':
            return latin1Write(this, string, offset, length)

          case 'base64':
            // Warning: maxLength not taken into account in base64Write
            return base64Write(this, string, offset, length)

          case 'ucs2':
          case 'ucs-2':
          case 'utf16le':
          case 'utf-16le':
            return ucs2Write(this, string, offset, length)

          default:
            if (loweredCase) throw new TypeError('Unknown encoding: ' + encoding)
            encoding = ('' + encoding).toLowerCase();
            loweredCase = true;
        }
      }
    };

    Buffer$1.prototype.toJSON = function toJSON () {
      return {
        type: 'Buffer',
        data: Array.prototype.slice.call(this._arr || this, 0)
      }
    };

    function base64Slice (buf, start, end) {
      if (start === 0 && end === buf.length) {
        return fromByteArray(buf)
      } else {
        return fromByteArray(buf.slice(start, end))
      }
    }

    function utf8Slice (buf, start, end) {
      end = Math.min(buf.length, end);
      var res = [];

      var i = start;
      while (i < end) {
        var firstByte = buf[i];
        var codePoint = null;
        var bytesPerSequence = (firstByte > 0xEF) ? 4
          : (firstByte > 0xDF) ? 3
          : (firstByte > 0xBF) ? 2
          : 1;

        if (i + bytesPerSequence <= end) {
          var secondByte, thirdByte, fourthByte, tempCodePoint;

          switch (bytesPerSequence) {
            case 1:
              if (firstByte < 0x80) {
                codePoint = firstByte;
              }
              break
            case 2:
              secondByte = buf[i + 1];
              if ((secondByte & 0xC0) === 0x80) {
                tempCodePoint = (firstByte & 0x1F) << 0x6 | (secondByte & 0x3F);
                if (tempCodePoint > 0x7F) {
                  codePoint = tempCodePoint;
                }
              }
              break
            case 3:
              secondByte = buf[i + 1];
              thirdByte = buf[i + 2];
              if ((secondByte & 0xC0) === 0x80 && (thirdByte & 0xC0) === 0x80) {
                tempCodePoint = (firstByte & 0xF) << 0xC | (secondByte & 0x3F) << 0x6 | (thirdByte & 0x3F);
                if (tempCodePoint > 0x7FF && (tempCodePoint < 0xD800 || tempCodePoint > 0xDFFF)) {
                  codePoint = tempCodePoint;
                }
              }
              break
            case 4:
              secondByte = buf[i + 1];
              thirdByte = buf[i + 2];
              fourthByte = buf[i + 3];
              if ((secondByte & 0xC0) === 0x80 && (thirdByte & 0xC0) === 0x80 && (fourthByte & 0xC0) === 0x80) {
                tempCodePoint = (firstByte & 0xF) << 0x12 | (secondByte & 0x3F) << 0xC | (thirdByte & 0x3F) << 0x6 | (fourthByte & 0x3F);
                if (tempCodePoint > 0xFFFF && tempCodePoint < 0x110000) {
                  codePoint = tempCodePoint;
                }
              }
          }
        }

        if (codePoint === null) {
          // we did not generate a valid codePoint so insert a
          // replacement char (U+FFFD) and advance only 1 byte
          codePoint = 0xFFFD;
          bytesPerSequence = 1;
        } else if (codePoint > 0xFFFF) {
          // encode to utf16 (surrogate pair dance)
          codePoint -= 0x10000;
          res.push(codePoint >>> 10 & 0x3FF | 0xD800);
          codePoint = 0xDC00 | codePoint & 0x3FF;
        }

        res.push(codePoint);
        i += bytesPerSequence;
      }

      return decodeCodePointsArray(res)
    }

    // Based on http://stackoverflow.com/a/22747272/680742, the browser with
    // the lowest limit is Chrome, with 0x10000 args.
    // We go 1 magnitude less, for safety
    var MAX_ARGUMENTS_LENGTH = 0x1000;

    function decodeCodePointsArray (codePoints) {
      var len = codePoints.length;
      if (len <= MAX_ARGUMENTS_LENGTH) {
        return String.fromCharCode.apply(String, codePoints) // avoid extra slice()
      }

      // Decode in chunks to avoid "call stack size exceeded".
      var res = '';
      var i = 0;
      while (i < len) {
        res += String.fromCharCode.apply(
          String,
          codePoints.slice(i, i += MAX_ARGUMENTS_LENGTH)
        );
      }
      return res
    }

    function asciiSlice (buf, start, end) {
      var ret = '';
      end = Math.min(buf.length, end);

      for (var i = start; i < end; ++i) {
        ret += String.fromCharCode(buf[i] & 0x7F);
      }
      return ret
    }

    function latin1Slice (buf, start, end) {
      var ret = '';
      end = Math.min(buf.length, end);

      for (var i = start; i < end; ++i) {
        ret += String.fromCharCode(buf[i]);
      }
      return ret
    }

    function hexSlice (buf, start, end) {
      var len = buf.length;

      if (!start || start < 0) start = 0;
      if (!end || end < 0 || end > len) end = len;

      var out = '';
      for (var i = start; i < end; ++i) {
        out += toHex(buf[i]);
      }
      return out
    }

    function utf16leSlice (buf, start, end) {
      var bytes = buf.slice(start, end);
      var res = '';
      for (var i = 0; i < bytes.length; i += 2) {
        res += String.fromCharCode(bytes[i] + bytes[i + 1] * 256);
      }
      return res
    }

    Buffer$1.prototype.slice = function slice (start, end) {
      var len = this.length;
      start = ~~start;
      end = end === undefined ? len : ~~end;

      if (start < 0) {
        start += len;
        if (start < 0) start = 0;
      } else if (start > len) {
        start = len;
      }

      if (end < 0) {
        end += len;
        if (end < 0) end = 0;
      } else if (end > len) {
        end = len;
      }

      if (end < start) end = start;

      var newBuf;
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        newBuf = this.subarray(start, end);
        newBuf.__proto__ = Buffer$1.prototype;
      } else {
        var sliceLen = end - start;
        newBuf = new Buffer$1(sliceLen, undefined);
        for (var i = 0; i < sliceLen; ++i) {
          newBuf[i] = this[i + start];
        }
      }

      return newBuf
    };

    /*
     * Need to make sure that buffer isn't trying to write out of bounds.
     */
    function checkOffset (offset, ext, length) {
      if ((offset % 1) !== 0 || offset < 0) throw new RangeError('offset is not uint')
      if (offset + ext > length) throw new RangeError('Trying to access beyond buffer length')
    }

    Buffer$1.prototype.readUIntLE = function readUIntLE (offset, byteLength, noAssert) {
      offset = offset | 0;
      byteLength = byteLength | 0;
      if (!noAssert) checkOffset(offset, byteLength, this.length);

      var val = this[offset];
      var mul = 1;
      var i = 0;
      while (++i < byteLength && (mul *= 0x100)) {
        val += this[offset + i] * mul;
      }

      return val
    };

    Buffer$1.prototype.readUIntBE = function readUIntBE (offset, byteLength, noAssert) {
      offset = offset | 0;
      byteLength = byteLength | 0;
      if (!noAssert) {
        checkOffset(offset, byteLength, this.length);
      }

      var val = this[offset + --byteLength];
      var mul = 1;
      while (byteLength > 0 && (mul *= 0x100)) {
        val += this[offset + --byteLength] * mul;
      }

      return val
    };

    Buffer$1.prototype.readUInt8 = function readUInt8 (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 1, this.length);
      return this[offset]
    };

    Buffer$1.prototype.readUInt16LE = function readUInt16LE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 2, this.length);
      return this[offset] | (this[offset + 1] << 8)
    };

    Buffer$1.prototype.readUInt16BE = function readUInt16BE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 2, this.length);
      return (this[offset] << 8) | this[offset + 1]
    };

    Buffer$1.prototype.readUInt32LE = function readUInt32LE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 4, this.length);

      return ((this[offset]) |
          (this[offset + 1] << 8) |
          (this[offset + 2] << 16)) +
          (this[offset + 3] * 0x1000000)
    };

    Buffer$1.prototype.readUInt32BE = function readUInt32BE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 4, this.length);

      return (this[offset] * 0x1000000) +
        ((this[offset + 1] << 16) |
        (this[offset + 2] << 8) |
        this[offset + 3])
    };

    Buffer$1.prototype.readIntLE = function readIntLE (offset, byteLength, noAssert) {
      offset = offset | 0;
      byteLength = byteLength | 0;
      if (!noAssert) checkOffset(offset, byteLength, this.length);

      var val = this[offset];
      var mul = 1;
      var i = 0;
      while (++i < byteLength && (mul *= 0x100)) {
        val += this[offset + i] * mul;
      }
      mul *= 0x80;

      if (val >= mul) val -= Math.pow(2, 8 * byteLength);

      return val
    };

    Buffer$1.prototype.readIntBE = function readIntBE (offset, byteLength, noAssert) {
      offset = offset | 0;
      byteLength = byteLength | 0;
      if (!noAssert) checkOffset(offset, byteLength, this.length);

      var i = byteLength;
      var mul = 1;
      var val = this[offset + --i];
      while (i > 0 && (mul *= 0x100)) {
        val += this[offset + --i] * mul;
      }
      mul *= 0x80;

      if (val >= mul) val -= Math.pow(2, 8 * byteLength);

      return val
    };

    Buffer$1.prototype.readInt8 = function readInt8 (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 1, this.length);
      if (!(this[offset] & 0x80)) return (this[offset])
      return ((0xff - this[offset] + 1) * -1)
    };

    Buffer$1.prototype.readInt16LE = function readInt16LE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 2, this.length);
      var val = this[offset] | (this[offset + 1] << 8);
      return (val & 0x8000) ? val | 0xFFFF0000 : val
    };

    Buffer$1.prototype.readInt16BE = function readInt16BE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 2, this.length);
      var val = this[offset + 1] | (this[offset] << 8);
      return (val & 0x8000) ? val | 0xFFFF0000 : val
    };

    Buffer$1.prototype.readInt32LE = function readInt32LE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 4, this.length);

      return (this[offset]) |
        (this[offset + 1] << 8) |
        (this[offset + 2] << 16) |
        (this[offset + 3] << 24)
    };

    Buffer$1.prototype.readInt32BE = function readInt32BE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 4, this.length);

      return (this[offset] << 24) |
        (this[offset + 1] << 16) |
        (this[offset + 2] << 8) |
        (this[offset + 3])
    };

    Buffer$1.prototype.readFloatLE = function readFloatLE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 4, this.length);
      return read(this, offset, true, 23, 4)
    };

    Buffer$1.prototype.readFloatBE = function readFloatBE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 4, this.length);
      return read(this, offset, false, 23, 4)
    };

    Buffer$1.prototype.readDoubleLE = function readDoubleLE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 8, this.length);
      return read(this, offset, true, 52, 8)
    };

    Buffer$1.prototype.readDoubleBE = function readDoubleBE (offset, noAssert) {
      if (!noAssert) checkOffset(offset, 8, this.length);
      return read(this, offset, false, 52, 8)
    };

    function checkInt (buf, value, offset, ext, max, min) {
      if (!internalIsBuffer(buf)) throw new TypeError('"buffer" argument must be a Buffer instance')
      if (value > max || value < min) throw new RangeError('"value" argument is out of bounds')
      if (offset + ext > buf.length) throw new RangeError('Index out of range')
    }

    Buffer$1.prototype.writeUIntLE = function writeUIntLE (value, offset, byteLength, noAssert) {
      value = +value;
      offset = offset | 0;
      byteLength = byteLength | 0;
      if (!noAssert) {
        var maxBytes = Math.pow(2, 8 * byteLength) - 1;
        checkInt(this, value, offset, byteLength, maxBytes, 0);
      }

      var mul = 1;
      var i = 0;
      this[offset] = value & 0xFF;
      while (++i < byteLength && (mul *= 0x100)) {
        this[offset + i] = (value / mul) & 0xFF;
      }

      return offset + byteLength
    };

    Buffer$1.prototype.writeUIntBE = function writeUIntBE (value, offset, byteLength, noAssert) {
      value = +value;
      offset = offset | 0;
      byteLength = byteLength | 0;
      if (!noAssert) {
        var maxBytes = Math.pow(2, 8 * byteLength) - 1;
        checkInt(this, value, offset, byteLength, maxBytes, 0);
      }

      var i = byteLength - 1;
      var mul = 1;
      this[offset + i] = value & 0xFF;
      while (--i >= 0 && (mul *= 0x100)) {
        this[offset + i] = (value / mul) & 0xFF;
      }

      return offset + byteLength
    };

    Buffer$1.prototype.writeUInt8 = function writeUInt8 (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 1, 0xff, 0);
      if (!Buffer$1.TYPED_ARRAY_SUPPORT) value = Math.floor(value);
      this[offset] = (value & 0xff);
      return offset + 1
    };

    function objectWriteUInt16 (buf, value, offset, littleEndian) {
      if (value < 0) value = 0xffff + value + 1;
      for (var i = 0, j = Math.min(buf.length - offset, 2); i < j; ++i) {
        buf[offset + i] = (value & (0xff << (8 * (littleEndian ? i : 1 - i)))) >>>
          (littleEndian ? i : 1 - i) * 8;
      }
    }

    Buffer$1.prototype.writeUInt16LE = function writeUInt16LE (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 2, 0xffff, 0);
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        this[offset] = (value & 0xff);
        this[offset + 1] = (value >>> 8);
      } else {
        objectWriteUInt16(this, value, offset, true);
      }
      return offset + 2
    };

    Buffer$1.prototype.writeUInt16BE = function writeUInt16BE (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 2, 0xffff, 0);
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        this[offset] = (value >>> 8);
        this[offset + 1] = (value & 0xff);
      } else {
        objectWriteUInt16(this, value, offset, false);
      }
      return offset + 2
    };

    function objectWriteUInt32 (buf, value, offset, littleEndian) {
      if (value < 0) value = 0xffffffff + value + 1;
      for (var i = 0, j = Math.min(buf.length - offset, 4); i < j; ++i) {
        buf[offset + i] = (value >>> (littleEndian ? i : 3 - i) * 8) & 0xff;
      }
    }

    Buffer$1.prototype.writeUInt32LE = function writeUInt32LE (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 4, 0xffffffff, 0);
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        this[offset + 3] = (value >>> 24);
        this[offset + 2] = (value >>> 16);
        this[offset + 1] = (value >>> 8);
        this[offset] = (value & 0xff);
      } else {
        objectWriteUInt32(this, value, offset, true);
      }
      return offset + 4
    };

    Buffer$1.prototype.writeUInt32BE = function writeUInt32BE (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 4, 0xffffffff, 0);
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        this[offset] = (value >>> 24);
        this[offset + 1] = (value >>> 16);
        this[offset + 2] = (value >>> 8);
        this[offset + 3] = (value & 0xff);
      } else {
        objectWriteUInt32(this, value, offset, false);
      }
      return offset + 4
    };

    Buffer$1.prototype.writeIntLE = function writeIntLE (value, offset, byteLength, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) {
        var limit = Math.pow(2, 8 * byteLength - 1);

        checkInt(this, value, offset, byteLength, limit - 1, -limit);
      }

      var i = 0;
      var mul = 1;
      var sub = 0;
      this[offset] = value & 0xFF;
      while (++i < byteLength && (mul *= 0x100)) {
        if (value < 0 && sub === 0 && this[offset + i - 1] !== 0) {
          sub = 1;
        }
        this[offset + i] = ((value / mul) >> 0) - sub & 0xFF;
      }

      return offset + byteLength
    };

    Buffer$1.prototype.writeIntBE = function writeIntBE (value, offset, byteLength, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) {
        var limit = Math.pow(2, 8 * byteLength - 1);

        checkInt(this, value, offset, byteLength, limit - 1, -limit);
      }

      var i = byteLength - 1;
      var mul = 1;
      var sub = 0;
      this[offset + i] = value & 0xFF;
      while (--i >= 0 && (mul *= 0x100)) {
        if (value < 0 && sub === 0 && this[offset + i + 1] !== 0) {
          sub = 1;
        }
        this[offset + i] = ((value / mul) >> 0) - sub & 0xFF;
      }

      return offset + byteLength
    };

    Buffer$1.prototype.writeInt8 = function writeInt8 (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 1, 0x7f, -0x80);
      if (!Buffer$1.TYPED_ARRAY_SUPPORT) value = Math.floor(value);
      if (value < 0) value = 0xff + value + 1;
      this[offset] = (value & 0xff);
      return offset + 1
    };

    Buffer$1.prototype.writeInt16LE = function writeInt16LE (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 2, 0x7fff, -0x8000);
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        this[offset] = (value & 0xff);
        this[offset + 1] = (value >>> 8);
      } else {
        objectWriteUInt16(this, value, offset, true);
      }
      return offset + 2
    };

    Buffer$1.prototype.writeInt16BE = function writeInt16BE (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 2, 0x7fff, -0x8000);
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        this[offset] = (value >>> 8);
        this[offset + 1] = (value & 0xff);
      } else {
        objectWriteUInt16(this, value, offset, false);
      }
      return offset + 2
    };

    Buffer$1.prototype.writeInt32LE = function writeInt32LE (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 4, 0x7fffffff, -0x80000000);
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        this[offset] = (value & 0xff);
        this[offset + 1] = (value >>> 8);
        this[offset + 2] = (value >>> 16);
        this[offset + 3] = (value >>> 24);
      } else {
        objectWriteUInt32(this, value, offset, true);
      }
      return offset + 4
    };

    Buffer$1.prototype.writeInt32BE = function writeInt32BE (value, offset, noAssert) {
      value = +value;
      offset = offset | 0;
      if (!noAssert) checkInt(this, value, offset, 4, 0x7fffffff, -0x80000000);
      if (value < 0) value = 0xffffffff + value + 1;
      if (Buffer$1.TYPED_ARRAY_SUPPORT) {
        this[offset] = (value >>> 24);
        this[offset + 1] = (value >>> 16);
        this[offset + 2] = (value >>> 8);
        this[offset + 3] = (value & 0xff);
      } else {
        objectWriteUInt32(this, value, offset, false);
      }
      return offset + 4
    };

    function checkIEEE754 (buf, value, offset, ext, max, min) {
      if (offset + ext > buf.length) throw new RangeError('Index out of range')
      if (offset < 0) throw new RangeError('Index out of range')
    }

    function writeFloat (buf, value, offset, littleEndian, noAssert) {
      if (!noAssert) {
        checkIEEE754(buf, value, offset, 4);
      }
      write(buf, value, offset, littleEndian, 23, 4);
      return offset + 4
    }

    Buffer$1.prototype.writeFloatLE = function writeFloatLE (value, offset, noAssert) {
      return writeFloat(this, value, offset, true, noAssert)
    };

    Buffer$1.prototype.writeFloatBE = function writeFloatBE (value, offset, noAssert) {
      return writeFloat(this, value, offset, false, noAssert)
    };

    function writeDouble (buf, value, offset, littleEndian, noAssert) {
      if (!noAssert) {
        checkIEEE754(buf, value, offset, 8);
      }
      write(buf, value, offset, littleEndian, 52, 8);
      return offset + 8
    }

    Buffer$1.prototype.writeDoubleLE = function writeDoubleLE (value, offset, noAssert) {
      return writeDouble(this, value, offset, true, noAssert)
    };

    Buffer$1.prototype.writeDoubleBE = function writeDoubleBE (value, offset, noAssert) {
      return writeDouble(this, value, offset, false, noAssert)
    };

    // copy(targetBuffer, targetStart=0, sourceStart=0, sourceEnd=buffer.length)
    Buffer$1.prototype.copy = function copy (target, targetStart, start, end) {
      if (!start) start = 0;
      if (!end && end !== 0) end = this.length;
      if (targetStart >= target.length) targetStart = target.length;
      if (!targetStart) targetStart = 0;
      if (end > 0 && end < start) end = start;

      // Copy 0 bytes; we're done
      if (end === start) return 0
      if (target.length === 0 || this.length === 0) return 0

      // Fatal error conditions
      if (targetStart < 0) {
        throw new RangeError('targetStart out of bounds')
      }
      if (start < 0 || start >= this.length) throw new RangeError('sourceStart out of bounds')
      if (end < 0) throw new RangeError('sourceEnd out of bounds')

      // Are we oob?
      if (end > this.length) end = this.length;
      if (target.length - targetStart < end - start) {
        end = target.length - targetStart + start;
      }

      var len = end - start;
      var i;

      if (this === target && start < targetStart && targetStart < end) {
        // descending copy from end
        for (i = len - 1; i >= 0; --i) {
          target[i + targetStart] = this[i + start];
        }
      } else if (len < 1000 || !Buffer$1.TYPED_ARRAY_SUPPORT) {
        // ascending copy from start
        for (i = 0; i < len; ++i) {
          target[i + targetStart] = this[i + start];
        }
      } else {
        Uint8Array.prototype.set.call(
          target,
          this.subarray(start, start + len),
          targetStart
        );
      }

      return len
    };

    // Usage:
    //    buffer.fill(number[, offset[, end]])
    //    buffer.fill(buffer[, offset[, end]])
    //    buffer.fill(string[, offset[, end]][, encoding])
    Buffer$1.prototype.fill = function fill (val, start, end, encoding) {
      // Handle string cases:
      if (typeof val === 'string') {
        if (typeof start === 'string') {
          encoding = start;
          start = 0;
          end = this.length;
        } else if (typeof end === 'string') {
          encoding = end;
          end = this.length;
        }
        if (val.length === 1) {
          var code = val.charCodeAt(0);
          if (code < 256) {
            val = code;
          }
        }
        if (encoding !== undefined && typeof encoding !== 'string') {
          throw new TypeError('encoding must be a string')
        }
        if (typeof encoding === 'string' && !Buffer$1.isEncoding(encoding)) {
          throw new TypeError('Unknown encoding: ' + encoding)
        }
      } else if (typeof val === 'number') {
        val = val & 255;
      }

      // Invalid ranges are not set to a default, so can range check early.
      if (start < 0 || this.length < start || this.length < end) {
        throw new RangeError('Out of range index')
      }

      if (end <= start) {
        return this
      }

      start = start >>> 0;
      end = end === undefined ? this.length : end >>> 0;

      if (!val) val = 0;

      var i;
      if (typeof val === 'number') {
        for (i = start; i < end; ++i) {
          this[i] = val;
        }
      } else {
        var bytes = internalIsBuffer(val)
          ? val
          : utf8ToBytes(new Buffer$1(val, encoding).toString());
        var len = bytes.length;
        for (i = 0; i < end - start; ++i) {
          this[i + start] = bytes[i % len];
        }
      }

      return this
    };

    // HELPER FUNCTIONS
    // ================

    var INVALID_BASE64_RE = /[^+\/0-9A-Za-z-_]/g;

    function base64clean (str) {
      // Node strips out invalid characters like \n and \t from the string, base64-js does not
      str = stringtrim(str).replace(INVALID_BASE64_RE, '');
      // Node converts strings with length < 2 to ''
      if (str.length < 2) return ''
      // Node allows for non-padded base64 strings (missing trailing ===), base64-js does not
      while (str.length % 4 !== 0) {
        str = str + '=';
      }
      return str
    }

    function stringtrim (str) {
      if (str.trim) return str.trim()
      return str.replace(/^\s+|\s+$/g, '')
    }

    function toHex (n) {
      if (n < 16) return '0' + n.toString(16)
      return n.toString(16)
    }

    function utf8ToBytes (string, units) {
      units = units || Infinity;
      var codePoint;
      var length = string.length;
      var leadSurrogate = null;
      var bytes = [];

      for (var i = 0; i < length; ++i) {
        codePoint = string.charCodeAt(i);

        // is surrogate component
        if (codePoint > 0xD7FF && codePoint < 0xE000) {
          // last char was a lead
          if (!leadSurrogate) {
            // no lead yet
            if (codePoint > 0xDBFF) {
              // unexpected trail
              if ((units -= 3) > -1) bytes.push(0xEF, 0xBF, 0xBD);
              continue
            } else if (i + 1 === length) {
              // unpaired lead
              if ((units -= 3) > -1) bytes.push(0xEF, 0xBF, 0xBD);
              continue
            }

            // valid lead
            leadSurrogate = codePoint;

            continue
          }

          // 2 leads in a row
          if (codePoint < 0xDC00) {
            if ((units -= 3) > -1) bytes.push(0xEF, 0xBF, 0xBD);
            leadSurrogate = codePoint;
            continue
          }

          // valid surrogate pair
          codePoint = (leadSurrogate - 0xD800 << 10 | codePoint - 0xDC00) + 0x10000;
        } else if (leadSurrogate) {
          // valid bmp char, but last char was a lead
          if ((units -= 3) > -1) bytes.push(0xEF, 0xBF, 0xBD);
        }

        leadSurrogate = null;

        // encode utf8
        if (codePoint < 0x80) {
          if ((units -= 1) < 0) break
          bytes.push(codePoint);
        } else if (codePoint < 0x800) {
          if ((units -= 2) < 0) break
          bytes.push(
            codePoint >> 0x6 | 0xC0,
            codePoint & 0x3F | 0x80
          );
        } else if (codePoint < 0x10000) {
          if ((units -= 3) < 0) break
          bytes.push(
            codePoint >> 0xC | 0xE0,
            codePoint >> 0x6 & 0x3F | 0x80,
            codePoint & 0x3F | 0x80
          );
        } else if (codePoint < 0x110000) {
          if ((units -= 4) < 0) break
          bytes.push(
            codePoint >> 0x12 | 0xF0,
            codePoint >> 0xC & 0x3F | 0x80,
            codePoint >> 0x6 & 0x3F | 0x80,
            codePoint & 0x3F | 0x80
          );
        } else {
          throw new Error('Invalid code point')
        }
      }

      return bytes
    }

    function asciiToBytes (str) {
      var byteArray = [];
      for (var i = 0; i < str.length; ++i) {
        // Node's code seems to be doing this and not & 0x7F..
        byteArray.push(str.charCodeAt(i) & 0xFF);
      }
      return byteArray
    }

    function utf16leToBytes (str, units) {
      var c, hi, lo;
      var byteArray = [];
      for (var i = 0; i < str.length; ++i) {
        if ((units -= 2) < 0) break

        c = str.charCodeAt(i);
        hi = c >> 8;
        lo = c % 256;
        byteArray.push(lo);
        byteArray.push(hi);
      }

      return byteArray
    }


    function base64ToBytes (str) {
      return toByteArray(base64clean(str))
    }

    function blitBuffer (src, dst, offset, length) {
      for (var i = 0; i < length; ++i) {
        if ((i + offset >= dst.length) || (i >= src.length)) break
        dst[i + offset] = src[i];
      }
      return i
    }

    function isnan (val) {
      return val !== val // eslint-disable-line no-self-compare
    }


    // the following is from is-buffer, also by Feross Aboukhadijeh and with same lisence
    // The _isBuffer check is for Safari 5-7 support, because it's missing
    // Object.prototype.constructor. Remove this eventually
    function isBuffer$2(obj) {
      return obj != null && (!!obj._isBuffer || isFastBuffer(obj) || isSlowBuffer(obj))
    }

    function isFastBuffer (obj) {
      return !!obj.constructor && typeof obj.constructor.isBuffer === 'function' && obj.constructor.isBuffer(obj)
    }

    // For Node v0.10 support. Remove this eventually.
    function isSlowBuffer (obj) {
      return typeof obj.readFloatLE === 'function' && typeof obj.slice === 'function' && isFastBuffer(obj.slice(0, 0))
    }

    var bufferEs6 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        INSPECT_MAX_BYTES: INSPECT_MAX_BYTES,
        kMaxLength: _kMaxLength,
        Buffer: Buffer$1,
        SlowBuffer: SlowBuffer,
        isBuffer: isBuffer$2
    });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function inferShape(val, dtype) {
        let firstElem = val;
        if (isTypedArray$1(val)) {
            return dtype === 'string' ? [] : [val.length];
        }
        if (!Array.isArray(val)) {
            return []; // Scalar.
        }
        const shape = [];
        while (Array.isArray(firstElem) ||
            isTypedArray$1(firstElem) && dtype !== 'string') {
            shape.push(firstElem.length);
            firstElem = firstElem[0];
        }
        if (Array.isArray(val) &&
            env$1().getBool('TENSORLIKE_CHECK_SHAPE_CONSISTENCY')) {
            deepAssertShapeConsistency(val, shape, []);
        }
        return shape;
    }
    function deepAssertShapeConsistency(val, shape, indices) {
        indices = indices || [];
        if (!(Array.isArray(val)) && !isTypedArray$1(val)) {
            assert$4(shape.length === 0, () => `Element arr[${indices.join('][')}] is a primitive, ` +
                `but should be an array/TypedArray of ${shape[0]} elements`);
            return;
        }
        assert$4(shape.length > 0, () => `Element arr[${indices.join('][')}] should be a primitive, ` +
            `but is an array of ${val.length} elements`);
        assert$4(val.length === shape[0], () => `Element arr[${indices.join('][')}] should have ${shape[0]} ` +
            `elements, but has ${val.length} elements`);
        const subShape = shape.slice(1);
        for (let i = 0; i < val.length; ++i) {
            deepAssertShapeConsistency(val[i], subShape, indices.concat(i));
        }
    }
    function assertDtype(expectedDtype, actualDType, argName, functionName) {
        if (expectedDtype === 'string_or_numeric') {
            return;
        }
        if (expectedDtype == null) {
            throw new Error(`Expected dtype cannot be null.`);
        }
        if (expectedDtype !== 'numeric' && expectedDtype !== actualDType ||
            expectedDtype === 'numeric' && actualDType === 'string') {
            throw new Error(`Argument '${argName}' passed to '${functionName}' must ` +
                `be ${expectedDtype} tensor, but got ${actualDType} tensor`);
        }
    }
    function convertToTensor(x, argName, functionName, parseAsDtype = 'numeric') {
        if (x instanceof Tensor) {
            assertDtype(parseAsDtype, x.dtype, argName, functionName);
            return x;
        }
        let inferredDtype = inferDtype(x);
        // If the user expects a bool/int/float, use that info to update the
        // inferredDtype when it is not a string.
        if (inferredDtype !== 'string' &&
            ['bool', 'int32', 'float32'].indexOf(parseAsDtype) >= 0) {
            inferredDtype = parseAsDtype;
        }
        assertDtype(parseAsDtype, inferredDtype, argName, functionName);
        if ((x == null) ||
            (!isTypedArray$1(x) && !Array.isArray(x) && typeof x !== 'number' &&
                typeof x !== 'boolean' && typeof x !== 'string')) {
            const type = x == null ? 'null' : x.constructor.name;
            throw new Error(`Argument '${argName}' passed to '${functionName}' must be a ` +
                `Tensor or TensorLike, but got '${type}'`);
        }
        const inferredShape = inferShape(x, inferredDtype);
        if (!isTypedArray$1(x) && !Array.isArray(x)) {
            x = [x];
        }
        const skipTypedArray = true;
        const values = inferredDtype !== 'string' ?
            toTypedArray(x, inferredDtype) :
            flatten$3(x, [], skipTypedArray);
        return ENGINE.makeTensor(values, inferredShape, inferredDtype);
    }
    function convertToTensorArray(arg, argName, functionName, parseAsDtype = 'numeric') {
        if (!Array.isArray(arg)) {
            throw new Error(`Argument ${argName} passed to ${functionName} must be a ` +
                '`Tensor[]` or `TensorLike[]`');
        }
        const tensors = arg;
        return tensors.map((t, i) => convertToTensor(t, `${argName}[${i}]`, functionName, parseAsDtype));
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const OP_SCOPE_SUFFIX = '__op';
    /**
     * Used for wrapping functions that perform math operations on
     * Tensors. The function will be wrapped in a named scope that cleans all
     * memory usage after the function is done.
     */
    function op(f) {
        const keys = Object.keys(f);
        if (keys.length !== 1) {
            throw new Error(`Please provide an object with a single key ` +
                `(operation name) mapping to a function. Got an object with ` +
                `${keys.length} keys.`);
        }
        let opName = keys[0];
        const fn = f[opName];
        // Strip the underscore from the end of the function name.
        if (opName.endsWith('_')) {
            opName = opName.substring(0, opName.length - 1);
        }
        // add an __op suffix to distinguish ops from kernels in tf.profile
        opName = opName + OP_SCOPE_SUFFIX;
        // tslint:disable-next-line:no-any
        const f2 = (...args) => {
            ENGINE.startScope(opName);
            try {
                const result = fn(...args);
                if (isPromise(result)) {
                    console.error('Cannot return a Promise inside of tidy.');
                }
                ENGINE.endScope(result);
                return result;
            }
            catch (ex) {
                ENGINE.endScope(null);
                throw ex;
            }
        };
        Object.defineProperty(f2, 'name', { value: opName, configurable: true });
        // tslint:disable-next-line:no-any
        return f2;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts two real numbers to a complex number.
     *
     * Given a tensor `real` representing the real part of a complex number, and a
     * tensor `imag` representing the imaginary part of a complex number, this
     * operation returns complex numbers elementwise of the form [r0, i0, r1, i1],
     * where r represents the real part and i represents the imag part.
     *
     * The input tensors real and imag must have the same shape.
     *
     * ```js
     * const real = tf.tensor1d([2.25, 3.25]);
     * const imag = tf.tensor1d([4.75, 5.75]);
     * const complex = tf.complex(real, imag);
     *
     * complex.print();
     * ```
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function complex_(real, imag) {
        const $real = convertToTensor(real, 'real', 'complex');
        const $imag = convertToTensor(imag, 'imag', 'complex');
        assertShapesMatch($real.shape, $imag.shape, `real and imag shapes, ${$real.shape} and ${$imag.shape}, ` +
            `must match in call to tf.complex().`);
        const inputs = { real: $real, imag: $imag };
        return ENGINE.runKernel(Complex, inputs);
    }
    const complex$2 = op({ complex_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /** This is shared code across all tensor creation methods. */
    function makeTensor(values, shape, inferredShape, dtype) {
        if (dtype == null) {
            dtype = inferDtype(values);
        }
        if (dtype === 'complex64') {
            throw new Error(`Cannot construct a complex64 tensor directly. ` +
                `Please use tf.complex(real, imag).`);
        }
        if (!isTypedArray$1(values) && !Array.isArray(values) &&
            typeof values !== 'number' && typeof values !== 'boolean' &&
            typeof values !== 'string') {
            throw new Error('values passed to tensor(values) must be a number/boolean/string or ' +
                'an array of numbers/booleans/strings, or a TypedArray');
        }
        if (shape != null) {
            assertNonNegativeIntegerDimensions(shape);
            const providedSize = sizeFromShape(shape);
            const inferredSize = sizeFromShape(inferredShape);
            assert$4(providedSize === inferredSize, () => `Based on the provided shape, [${shape}], the tensor should have ` +
                `${providedSize} values but has ${inferredSize}`);
            for (let i = 0; i < inferredShape.length; ++i) {
                const inferred = inferredShape[i];
                const flatDimsDontMatch = i === inferredShape.length - 1 ?
                    inferred !== sizeFromShape(shape.slice(i)) :
                    true;
                assert$4(inferredShape[i] === shape[i] || !flatDimsDontMatch, () => `Error creating a new Tensor. Inferred shape ` +
                    `(${inferredShape}) does not match the provided ` +
                    `shape (${shape}). `);
            }
        }
        if (!isTypedArray$1(values) && !Array.isArray(values)) {
            values = [values];
        }
        shape = shape || inferredShape;
        values = dtype !== 'string' ?
            toTypedArray(values, dtype) :
            flatten$3(values, [], true);
        return ENGINE.makeTensor(values, shape, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with the provided values, shape and dtype.
     *
     * ```js
     * // Pass an array of values to create a vector.
     * tf.tensor([1, 2, 3, 4]).print();
     * ```
     *
     * ```js
     * // Pass a nested array of values to make a matrix or a higher
     * // dimensional tensor.
     * tf.tensor([[1, 2], [3, 4]]).print();
     * ```
     *
     * ```js
     * // Pass a flat array and specify a shape yourself.
     * tf.tensor([1, 2, 3, 4], [2, 2]).print();
     * ```
     *
     * @param values The values of the tensor. Can be nested array of numbers,
     *     or a flat array, or a `TypedArray`. If the values are strings,
     *     they will be encoded as utf-8 and kept as `Uint8Array[]`.
     * @param shape The shape of the tensor. Optional. If not provided,
     *   it is inferred from `values`.
     * @param dtype The data type.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function tensor(values, shape, dtype) {
        const inferredShape = inferShape(values, dtype);
        return makeTensor(values, shape, inferredShape, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /* Type definitions for exporting and importing of models. */
    /**
     * A map from Tensor dtype to number of bytes per element of the Tensor.
     */
    const DTYPE_VALUE_SIZE_MAP = {
        'float32': 4,
        'float16': 2,
        'int32': 4,
        'uint16': 2,
        'uint8': 1,
        'bool': 1,
        'complex64': 8
    };

    /** Number of bytes reserved for the length of the string. (32bit integer). */
    const NUM_BYTES_STRING_LENGTH = 4;
    /**
     * Encode a map from names to weight values as an ArrayBuffer, along with an
     * `Array` of `WeightsManifestEntry` as specification of the encoded weights.
     *
     * This function does not perform sharding.
     *
     * This function is the reverse of `decodeWeights`.
     *
     * @param tensors A map ("dict") from names to tensors.
     * @param group Group to which the weights belong (optional).
     * @returns A `Promise` of
     *   - A flat `ArrayBuffer` with all the binary values of the `Tensor`s
     *     concatenated.
     *   - An `Array` of `WeightManifestEntry`s, carrying information including
     *     tensor names, `dtype`s and shapes.
     * @throws Error: on unsupported tensor `dtype`.
     */
    async function encodeWeights(tensors, group) {
        // TODO(adarob, cais): Support quantization.
        const specs = [];
        const dataPromises = [];
        const names = Array.isArray(tensors) ?
            tensors.map(tensor => tensor.name) :
            Object.keys(tensors);
        for (let i = 0; i < names.length; ++i) {
            const name = names[i];
            const t = Array.isArray(tensors) ? tensors[i].tensor : tensors[name];
            if (t.dtype !== 'float32' && t.dtype !== 'int32' && t.dtype !== 'bool' &&
                t.dtype !== 'string' && t.dtype !== 'complex64') {
                throw new Error(`Unsupported dtype in weight '${name}': ${t.dtype}`);
            }
            const spec = { name, shape: t.shape, dtype: t.dtype };
            if (t.dtype === 'string') {
                const utf8bytes = new Promise(async (resolve) => {
                    const vals = await t.bytes();
                    const totalNumBytes = vals.reduce((p, c) => p + c.length, 0) +
                        NUM_BYTES_STRING_LENGTH * vals.length;
                    const bytes = new Uint8Array(totalNumBytes);
                    let offset = 0;
                    for (let i = 0; i < vals.length; i++) {
                        const val = vals[i];
                        const bytesOfLength = new Uint8Array(new Uint32Array([val.length]).buffer);
                        bytes.set(bytesOfLength, offset);
                        offset += NUM_BYTES_STRING_LENGTH;
                        bytes.set(val, offset);
                        offset += val.length;
                    }
                    resolve(bytes);
                });
                dataPromises.push(utf8bytes);
            }
            else {
                dataPromises.push(t.data());
            }
            if (group != null) {
                spec.group = group;
            }
            specs.push(spec);
        }
        const tensorValues = await Promise.all(dataPromises);
        return { data: concatenateTypedArrays(tensorValues), specs };
    }
    /**
     * Decode flat ArrayBuffer as weights.
     *
     * This function does not handle sharding.
     *
     * This function is the reverse of `encodeWeights`.
     *
     * @param buffer A flat ArrayBuffer carrying the binary values of the tensors
     *   concatenated in the order specified in `specs`.
     * @param specs Specifications of the names, dtypes and shapes of the tensors
     *   whose value are encoded by `buffer`.
     * @return A map from tensor name to tensor value, with the names corresponding
     *   to names in `specs`.
     * @throws Error, if any of the tensors has unsupported dtype.
     */
    function decodeWeights(buffer, specs) {
        // TODO(adarob, cais): Support quantization.
        const out = {};
        let float16Decode;
        let offset = 0;
        for (const spec of specs) {
            const name = spec.name;
            const dtype = spec.dtype;
            const shape = spec.shape;
            const size = sizeFromShape(shape);
            let values;
            if ('quantization' in spec) {
                const quantization = spec.quantization;
                if (quantization.dtype === 'uint8' || quantization.dtype === 'uint16') {
                    if (!('min' in quantization && 'scale' in quantization)) {
                        throw new Error(`Weight ${spec.name} with quantization ${quantization.dtype} ` +
                            `doesn't have corresponding metadata min and scale.`);
                    }
                }
                else if (quantization.dtype === 'float16') {
                    if (dtype !== 'float32') {
                        throw new Error(`Weight ${spec.name} is quantized with ${quantization.dtype} ` +
                            `which only supports weights of type float32 not ${dtype}.`);
                    }
                }
                else {
                    throw new Error(`Weight ${spec.name} has unknown ` +
                        `quantization dtype ${quantization.dtype}. ` +
                        `Supported quantization dtypes are: ` +
                        `'uint8', 'uint16', and 'float16'.`);
                }
                const quantizationSizeFactor = DTYPE_VALUE_SIZE_MAP[quantization.dtype];
                const byteBuffer = buffer.slice(offset, offset + size * quantizationSizeFactor);
                const quantizedArray = (quantization.dtype === 'uint8') ?
                    new Uint8Array(byteBuffer) :
                    new Uint16Array(byteBuffer);
                if (dtype === 'float32') {
                    if (quantization.dtype === 'uint8' || quantization.dtype === 'uint16') {
                        values = new Float32Array(quantizedArray.length);
                        for (let i = 0; i < quantizedArray.length; i++) {
                            const v = quantizedArray[i];
                            values[i] = v * quantization.scale + quantization.min;
                        }
                    }
                    else if (quantization.dtype === 'float16') {
                        if (float16Decode === undefined) {
                            float16Decode = getFloat16Decoder();
                        }
                        values = float16Decode(quantizedArray);
                    }
                    else {
                        throw new Error(`Unsupported quantization type ${quantization.dtype} ` +
                            `for weight type float32.`);
                    }
                }
                else if (dtype === 'int32') {
                    if (quantization.dtype !== 'uint8' && quantization.dtype !== 'uint16') {
                        throw new Error(`Unsupported quantization type ${quantization.dtype} ` +
                            `for weight type int32.`);
                    }
                    values = new Int32Array(quantizedArray.length);
                    for (let i = 0; i < quantizedArray.length; i++) {
                        const v = quantizedArray[i];
                        values[i] = Math.round(v * quantization.scale + quantization.min);
                    }
                }
                else {
                    throw new Error(`Unsupported dtype in weight '${name}': ${dtype}`);
                }
                offset += size * quantizationSizeFactor;
            }
            else if (dtype === 'string') {
                const size = sizeFromShape(spec.shape);
                values = [];
                for (let i = 0; i < size; i++) {
                    const byteLength = new Uint32Array(buffer.slice(offset, offset + NUM_BYTES_STRING_LENGTH))[0];
                    offset += NUM_BYTES_STRING_LENGTH;
                    const bytes = new Uint8Array(buffer.slice(offset, offset + byteLength));
                    values.push(bytes);
                    offset += byteLength;
                }
            }
            else {
                const dtypeFactor = DTYPE_VALUE_SIZE_MAP[dtype];
                const byteBuffer = buffer.slice(offset, offset + size * dtypeFactor);
                if (dtype === 'float32') {
                    values = new Float32Array(byteBuffer);
                }
                else if (dtype === 'int32') {
                    values = new Int32Array(byteBuffer);
                }
                else if (dtype === 'bool') {
                    values = new Uint8Array(byteBuffer);
                }
                else if (dtype === 'complex64') {
                    values = new Float32Array(byteBuffer);
                    const real = new Float32Array(values.length / 2);
                    const image = new Float32Array(values.length / 2);
                    for (let i = 0; i < real.length; i++) {
                        real[i] = values[i * 2];
                        image[i] = values[i * 2 + 1];
                    }
                    const realTensor = tensor(real, shape, 'float32');
                    const imageTensor = tensor(image, shape, 'float32');
                    out[name] = complex$2(realTensor, imageTensor);
                    realTensor.dispose();
                    imageTensor.dispose();
                }
                else {
                    throw new Error(`Unsupported dtype in weight '${name}': ${dtype}`);
                }
                offset += size * dtypeFactor;
            }
            if (dtype !== 'complex64') {
                out[name] = tensor(values, shape, dtype);
            }
        }
        return out;
    }
    /**
     * Concatenate TypedArrays into an ArrayBuffer.
     */
    function concatenateTypedArrays(xs) {
        // TODO(adarob, cais): Support quantization.
        if (xs === null) {
            throw new Error(`Invalid input value: ${JSON.stringify(xs)}`);
        }
        let totalByteLength = 0;
        // `normalizedXs` is here for this reason: a `TypedArray`'s `buffer'
        // can have a different byte length from that of the `TypedArray` itself,
        // for example, when the `TypedArray` is created from an offset in an
        // `ArrayBuffer`. `normliazedXs` holds `TypedArray`s whose `buffer`s match
        // the `TypedArray` in byte length. If an element of `xs` does not show
        // this property, a new `TypedArray` that satisfy this property will be
        // constructed and pushed into `normalizedXs`.
        const normalizedXs = [];
        xs.forEach((x) => {
            totalByteLength += x.byteLength;
            // tslint:disable:no-any
            normalizedXs.push(x.byteLength === x.buffer.byteLength ? x :
                new x.constructor(x));
            if (!(x instanceof Float32Array || x instanceof Int32Array ||
                x instanceof Uint8Array)) {
                throw new Error(`Unsupported TypedArray subtype: ${x.constructor.name}`);
            }
            // tslint:enable:no-any
        });
        const y = new Uint8Array(totalByteLength);
        let offset = 0;
        normalizedXs.forEach((x) => {
            y.set(new Uint8Array(x.buffer), offset);
            offset += x.byteLength;
        });
        return y.buffer;
    }
    // Use Buffer on Node.js instead of Blob/atob/btoa
    const useNodeBuffer = typeof Buffer$1 !== 'undefined' &&
        (typeof Blob === 'undefined' || typeof atob === 'undefined' ||
            typeof btoa === 'undefined');
    /**
     * Calculate the byte length of a JavaScript string.
     *
     * Note that a JavaScript string can contain wide characters, therefore the
     * length of the string is not necessarily equal to the byte length.
     *
     * @param str Input string.
     * @returns Byte length.
     */
    function stringByteLength(str) {
        if (useNodeBuffer) {
            return Buffer$1.byteLength(str);
        }
        return new Blob([str]).size;
    }
    /**
     * Encode an ArrayBuffer as a base64 encoded string.
     *
     * @param buffer `ArrayBuffer` to be converted.
     * @returns A string that base64-encodes `buffer`.
     */
    function arrayBufferToBase64String(buffer) {
        if (useNodeBuffer) {
            return Buffer$1.from(buffer).toString('base64');
        }
        const buf = new Uint8Array(buffer);
        let s = '';
        for (let i = 0, l = buf.length; i < l; i++) {
            s += String.fromCharCode(buf[i]);
        }
        return btoa(s);
    }
    /**
     * Decode a base64 string as an ArrayBuffer.
     *
     * @param str Base64 string.
     * @returns Decoded `ArrayBuffer`.
     */
    function base64StringToArrayBuffer(str) {
        if (useNodeBuffer) {
            const buf = Buffer$1.from(str, 'base64');
            return buf.buffer.slice(buf.byteOffset, buf.byteOffset + buf.byteLength);
        }
        const s = atob(str);
        const buffer = new Uint8Array(s.length);
        for (let i = 0; i < s.length; ++i) {
            buffer.set([s.charCodeAt(i)], i);
        }
        return buffer.buffer;
    }
    /**
     * Concatenate a number of ArrayBuffers into one.
     *
     * @param buffers A number of array buffers to concatenate.
     * @returns Result of concatenating `buffers` in order.
     */
    function concatenateArrayBuffers(buffers) {
        if (buffers.length === 1) {
            return buffers[0];
        }
        let totalByteLength = 0;
        buffers.forEach((buffer) => {
            totalByteLength += buffer.byteLength;
        });
        const temp = new Uint8Array(totalByteLength);
        let offset = 0;
        buffers.forEach((buffer) => {
            temp.set(new Uint8Array(buffer), offset);
            offset += buffer.byteLength;
        });
        return temp.buffer;
    }
    /**
     * Get the basename of a path.
     *
     * Behaves in a way analogous to Linux's basename command.
     *
     * @param path
     */
    function basename$1(path) {
        const SEPARATOR = '/';
        path = path.trim();
        while (path.endsWith(SEPARATOR)) {
            path = path.slice(0, path.length - 1);
        }
        const items = path.split(SEPARATOR);
        return items[items.length - 1];
    }
    /**
     * Create `ModelJSON` from `ModelArtifacts`.
     *
     * @param artifacts Model artifacts, describing the model and its weights.
     * @param manifest Weight manifest, describing where the weights of the
     *     `ModelArtifacts` are stored, and some metadata about them.
     * @returns Object representing the `model.json` file describing the model
     *     artifacts and weights
     */
    function getModelJSONForModelArtifacts(artifacts, manifest) {
        const result = {
            modelTopology: artifacts.modelTopology,
            format: artifacts.format,
            generatedBy: artifacts.generatedBy,
            convertedBy: artifacts.convertedBy,
            weightsManifest: manifest
        };
        if (artifacts.signature != null) {
            result.signature = artifacts.signature;
        }
        if (artifacts.userDefinedMetadata != null) {
            result.userDefinedMetadata = artifacts.userDefinedMetadata;
        }
        if (artifacts.modelInitializer != null) {
            result.modelInitializer = artifacts.modelInitializer;
        }
        if (artifacts.trainingConfig != null) {
            result.trainingConfig = artifacts.trainingConfig;
        }
        return result;
    }
    /**
     * Create `ModelArtifacts` from a JSON file.
     *
     * @param modelJSON Object containing the parsed JSON of `model.json`
     * @param loadWeights Function that takes the JSON file's weights manifest,
     *     reads weights from the listed path(s), and returns a Promise of the
     *     weight manifest entries along with the weights data.
     * @returns A Promise of the `ModelArtifacts`, as described by the JSON file.
     */
    async function getModelArtifactsForJSON(modelJSON, loadWeights) {
        const modelArtifacts = {
            modelTopology: modelJSON.modelTopology,
            format: modelJSON.format,
            generatedBy: modelJSON.generatedBy,
            convertedBy: modelJSON.convertedBy
        };
        if (modelJSON.trainingConfig != null) {
            modelArtifacts.trainingConfig = modelJSON.trainingConfig;
        }
        if (modelJSON.weightsManifest != null) {
            const [weightSpecs, weightData] = await loadWeights(modelJSON.weightsManifest);
            modelArtifacts.weightSpecs = weightSpecs;
            modelArtifacts.weightData = weightData;
        }
        if (modelJSON.signature != null) {
            modelArtifacts.signature = modelJSON.signature;
        }
        if (modelJSON.userDefinedMetadata != null) {
            modelArtifacts.userDefinedMetadata = modelJSON.userDefinedMetadata;
        }
        if (modelJSON.modelInitializer != null) {
            modelArtifacts.modelInitializer = modelJSON.modelInitializer;
        }
        return modelArtifacts;
    }
    /**
     * Populate ModelArtifactsInfo fields for a model with JSON topology.
     * @param modelArtifacts
     * @returns A ModelArtifactsInfo object.
     */
    function getModelArtifactsInfoForJSON(modelArtifacts) {
        if (modelArtifacts.modelTopology instanceof ArrayBuffer) {
            throw new Error('Expected JSON model topology, received ArrayBuffer.');
        }
        return {
            dateSaved: new Date(),
            modelTopologyType: 'JSON',
            modelTopologyBytes: modelArtifacts.modelTopology == null ?
                0 :
                stringByteLength(JSON.stringify(modelArtifacts.modelTopology)),
            weightSpecsBytes: modelArtifacts.weightSpecs == null ?
                0 :
                stringByteLength(JSON.stringify(modelArtifacts.weightSpecs)),
            weightDataBytes: modelArtifacts.weightData == null ?
                0 :
                modelArtifacts.weightData.byteLength,
        };
    }
    /**
     * Computes mantisa table for casting Float16 to Float32
     * See http://www.fox-toolkit.org/ftp/fasthalffloatconversion.pdf
     *
     * @returns Uint32Array, 2048 mantissa lookup values.
     */
    function computeFloat16MantisaTable() {
        const convertMantissa = (i) => {
            let m = i << 13;
            let e = 0;
            while ((m & 0x00800000) === 0) {
                e -= 0x00800000;
                m <<= 1;
            }
            m &= ~0x00800000;
            e += 0x38800000;
            return m | e;
        };
        const mantisaTable = new Uint32Array(2048);
        mantisaTable[0] = 0;
        for (let i = 1; i < 1024; i++) {
            mantisaTable[i] = convertMantissa(i);
        }
        for (let i = 1024; i < 2048; i++) {
            mantisaTable[i] = 0x38000000 + ((i - 1024) << 13);
        }
        return mantisaTable;
    }
    /**
     * Computes exponent table for casting Float16 to Float32
     * See http://www.fox-toolkit.org/ftp/fasthalffloatconversion.pdf
     *
     * @returns Uint32Array, 64 exponent lookup values.
     */
    function computeFloat16ExponentTable() {
        const exponentTable = new Uint32Array(64);
        exponentTable[0] = 0;
        exponentTable[31] = 0x47800000;
        exponentTable[32] = 0x80000000;
        exponentTable[63] = 0xc7800000;
        for (let i = 1; i < 31; i++) {
            exponentTable[i] = i << 23;
        }
        for (let i = 33; i < 63; i++) {
            exponentTable[i] = 0x80000000 + ((i - 32) << 23);
        }
        return exponentTable;
    }
    /**
     * Computes offset table for casting Float16 to Float32
     * See http://www.fox-toolkit.org/ftp/fasthalffloatconversion.pdf
     *
     * @returns Uint32Array, 6d offset values.
     */
    function computeFloat16OffsetTable() {
        const offsetTable = new Uint32Array(64);
        for (let i = 0; i < 64; i++) {
            offsetTable[i] = 1024;
        }
        offsetTable[0] = offsetTable[32] = 0;
        return offsetTable;
    }
    /**
     * Retrieve a Float16 decoder which will decode a ByteArray of Float16 values
     * to a Float32Array.
     *
     * @returns Function (buffer: Uint16Array) => Float32Array which decodes
     *          the Uint16Array of Float16 bytes to a Float32Array.
     */
    function getFloat16Decoder() {
        // Algorithm is based off of
        // http://www.fox-toolkit.org/ftp/fasthalffloatconversion.pdf
        // Cache lookup tables
        const mantisaTable = computeFloat16MantisaTable();
        const exponentTable = computeFloat16ExponentTable();
        const offsetTable = computeFloat16OffsetTable();
        return (quantizedArray) => {
            const buffer = new ArrayBuffer(4 * quantizedArray.length);
            const bufferUint32View = new Uint32Array(buffer);
            for (let index = 0; index < quantizedArray.length; index++) {
                const float16Bits = quantizedArray[index];
                const float32Bits = mantisaTable[offsetTable[float16Bits >> 10] + (float16Bits & 0x3ff)] +
                    exponentTable[float16Bits >> 10];
                bufferUint32View[index] = float32Bits;
            }
            return new Float32Array(buffer);
        };
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class IORouterRegistry {
        constructor() {
            this.saveRouters = [];
            this.loadRouters = [];
        }
        static getInstance() {
            if (IORouterRegistry.instance == null) {
                IORouterRegistry.instance = new IORouterRegistry();
            }
            return IORouterRegistry.instance;
        }
        /**
         * Register a save-handler router.
         *
         * @param saveRouter A function that maps a URL-like string onto an instance
         * of `IOHandler` with the `save` method defined or `null`.
         */
        static registerSaveRouter(saveRouter) {
            IORouterRegistry.getInstance().saveRouters.push(saveRouter);
        }
        /**
         * Register a load-handler router.
         *
         * @param loadRouter A function that maps a URL-like string onto an instance
         * of `IOHandler` with the `load` method defined or `null`.
         */
        static registerLoadRouter(loadRouter) {
            IORouterRegistry.getInstance().loadRouters.push(loadRouter);
        }
        /**
         * Look up IOHandler for saving, given a URL-like string.
         *
         * @param url
         * @returns If only one match is found, an instance of IOHandler with the
         * `save` method defined. If no match is found, `null`.
         * @throws Error, if more than one match is found.
         */
        static getSaveHandlers(url) {
            return IORouterRegistry.getHandlers(url, 'save');
        }
        /**
         * Look up IOHandler for loading, given a URL-like string.
         *
         * @param url
         * @param loadOptions Optional, custom load options.
         * @returns All valid handlers for `url`, given the currently registered
         *   handler routers.
         */
        static getLoadHandlers(url, loadOptions) {
            return IORouterRegistry.getHandlers(url, 'load', loadOptions);
        }
        static getHandlers(url, handlerType, loadOptions) {
            const validHandlers = [];
            const routers = handlerType === 'load' ?
                IORouterRegistry.getInstance().loadRouters :
                IORouterRegistry.getInstance().saveRouters;
            routers.forEach(router => {
                const handler = router(url, loadOptions);
                if (handler !== null) {
                    validHandlers.push(handler);
                }
            });
            return validHandlers;
        }
    }
    const registerSaveRouter = (loudRouter) => IORouterRegistry.registerSaveRouter(loudRouter);
    const registerLoadRouter = (loudRouter) => IORouterRegistry.registerLoadRouter(loudRouter);
    const getSaveHandlers = (url) => IORouterRegistry.getSaveHandlers(url);
    const getLoadHandlers = (url, loadOptions) => IORouterRegistry.getLoadHandlers(url, loadOptions);

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const DATABASE_NAME = 'tensorflowjs';
    const DATABASE_VERSION = 1;
    // Model data and ModelArtifactsInfo (metadata) are stored in two separate
    // stores for efficient access of the list of stored models and their metadata.
    // 1. The object store for model data: topology, weights and weight manifests.
    const MODEL_STORE_NAME = 'models_store';
    // 2. The object store for ModelArtifactsInfo, including meta-information such
    //    as the type of topology (JSON vs binary), byte size of the topology, byte
    //    size of the weights, etc.
    const INFO_STORE_NAME = 'model_info_store';
    function getIndexedDBFactory() {
        if (!env$1().getBool('IS_BROWSER')) {
            // TODO(cais): Add more info about what IOHandler subtypes are available.
            //   Maybe point to a doc page on the web and/or automatically determine
            //   the available IOHandlers and print them in the error message.
            throw new Error('Failed to obtain IndexedDB factory because the current environment' +
                'is not a web browser.');
        }
        // tslint:disable-next-line:no-any
        const theWindow = typeof window === 'undefined' ? self : window;
        const factory = theWindow.indexedDB || theWindow.mozIndexedDB ||
            theWindow.webkitIndexedDB || theWindow.msIndexedDB ||
            theWindow.shimIndexedDB;
        if (factory == null) {
            throw new Error('The current browser does not appear to support IndexedDB.');
        }
        return factory;
    }
    function setUpDatabase(openRequest) {
        const db = openRequest.result;
        db.createObjectStore(MODEL_STORE_NAME, { keyPath: 'modelPath' });
        db.createObjectStore(INFO_STORE_NAME, { keyPath: 'modelPath' });
    }
    /**
     * IOHandler subclass: Browser IndexedDB.
     *
     * See the doc string of `browserIndexedDB` for more details.
     */
    class BrowserIndexedDB {
        constructor(modelPath) {
            this.indexedDB = getIndexedDBFactory();
            if (modelPath == null || !modelPath) {
                throw new Error('For IndexedDB, modelPath must not be null, undefined or empty.');
            }
            this.modelPath = modelPath;
        }
        async save(modelArtifacts) {
            // TODO(cais): Support saving GraphDef models.
            if (modelArtifacts.modelTopology instanceof ArrayBuffer) {
                throw new Error('BrowserLocalStorage.save() does not support saving model topology ' +
                    'in binary formats yet.');
            }
            return this.databaseAction(this.modelPath, modelArtifacts);
        }
        async load() {
            return this.databaseAction(this.modelPath);
        }
        /**
         * Perform database action to put model artifacts into or read model artifacts
         * from IndexedDB object store.
         *
         * Whether the action is put or get depends on whether `modelArtifacts` is
         * specified. If it is specified, the action will be put; otherwise the action
         * will be get.
         *
         * @param modelPath A unique string path for the model.
         * @param modelArtifacts If specified, it will be the model artifacts to be
         *   stored in IndexedDB.
         * @returns A `Promise` of `SaveResult`, if the action is put, or a `Promise`
         *   of `ModelArtifacts`, if the action is get.
         */
        databaseAction(modelPath, modelArtifacts) {
            return new Promise((resolve, reject) => {
                const openRequest = this.indexedDB.open(DATABASE_NAME, DATABASE_VERSION);
                openRequest.onupgradeneeded = () => setUpDatabase(openRequest);
                openRequest.onsuccess = () => {
                    const db = openRequest.result;
                    if (modelArtifacts == null) {
                        // Read model out from object store.
                        const modelTx = db.transaction(MODEL_STORE_NAME, 'readonly');
                        const modelStore = modelTx.objectStore(MODEL_STORE_NAME);
                        const getRequest = modelStore.get(this.modelPath);
                        getRequest.onsuccess = () => {
                            if (getRequest.result == null) {
                                db.close();
                                return reject(new Error(`Cannot find model with path '${this.modelPath}' ` +
                                    `in IndexedDB.`));
                            }
                            else {
                                resolve(getRequest.result.modelArtifacts);
                            }
                        };
                        getRequest.onerror = error => {
                            db.close();
                            return reject(getRequest.error);
                        };
                        modelTx.oncomplete = () => db.close();
                    }
                    else {
                        // Put model into object store.
                        const modelArtifactsInfo = getModelArtifactsInfoForJSON(modelArtifacts);
                        // First, put ModelArtifactsInfo into info store.
                        const infoTx = db.transaction(INFO_STORE_NAME, 'readwrite');
                        let infoStore = infoTx.objectStore(INFO_STORE_NAME);
                        const putInfoRequest = infoStore.put({ modelPath: this.modelPath, modelArtifactsInfo });
                        let modelTx;
                        putInfoRequest.onsuccess = () => {
                            // Second, put model data into model store.
                            modelTx = db.transaction(MODEL_STORE_NAME, 'readwrite');
                            const modelStore = modelTx.objectStore(MODEL_STORE_NAME);
                            const putModelRequest = modelStore.put({
                                modelPath: this.modelPath,
                                modelArtifacts,
                                modelArtifactsInfo
                            });
                            putModelRequest.onsuccess = () => resolve({ modelArtifactsInfo });
                            putModelRequest.onerror = error => {
                                // If the put-model request fails, roll back the info entry as
                                // well.
                                infoStore = infoTx.objectStore(INFO_STORE_NAME);
                                const deleteInfoRequest = infoStore.delete(this.modelPath);
                                deleteInfoRequest.onsuccess = () => {
                                    db.close();
                                    return reject(putModelRequest.error);
                                };
                                deleteInfoRequest.onerror = error => {
                                    db.close();
                                    return reject(putModelRequest.error);
                                };
                            };
                        };
                        putInfoRequest.onerror = error => {
                            db.close();
                            return reject(putInfoRequest.error);
                        };
                        infoTx.oncomplete = () => {
                            if (modelTx == null) {
                                db.close();
                            }
                            else {
                                modelTx.oncomplete = () => db.close();
                            }
                        };
                    }
                };
                openRequest.onerror = error => reject(openRequest.error);
            });
        }
    }
    BrowserIndexedDB.URL_SCHEME = 'indexeddb://';
    const indexedDBRouter = (url) => {
        if (!env$1().getBool('IS_BROWSER')) {
            return null;
        }
        else {
            if (!Array.isArray(url) && url.startsWith(BrowserIndexedDB.URL_SCHEME)) {
                return browserIndexedDB(url.slice(BrowserIndexedDB.URL_SCHEME.length));
            }
            else {
                return null;
            }
        }
    };
    IORouterRegistry.registerSaveRouter(indexedDBRouter);
    IORouterRegistry.registerLoadRouter(indexedDBRouter);
    /**
     * Creates a browser IndexedDB IOHandler for saving and loading models.
     *
     * ```js
     * const model = tf.sequential();
     * model.add(
     *     tf.layers.dense({units: 1, inputShape: [100], activation: 'sigmoid'}));
     *
     * const saveResult = await model.save('indexeddb://MyModel'));
     * console.log(saveResult);
     * ```
     *
     * @param modelPath A unique identifier for the model to be saved. Must be a
     *   non-empty string.
     * @returns An instance of `BrowserIndexedDB` (sublcass of `IOHandler`),
     *   which can be used with, e.g., `tf.Model.save`.
     */
    function browserIndexedDB(modelPath) {
        return new BrowserIndexedDB(modelPath);
    }
    function maybeStripScheme$1(key) {
        return key.startsWith(BrowserIndexedDB.URL_SCHEME) ?
            key.slice(BrowserIndexedDB.URL_SCHEME.length) :
            key;
    }
    class BrowserIndexedDBManager {
        constructor() {
            this.indexedDB = getIndexedDBFactory();
        }
        async listModels() {
            return new Promise((resolve, reject) => {
                const openRequest = this.indexedDB.open(DATABASE_NAME, DATABASE_VERSION);
                openRequest.onupgradeneeded = () => setUpDatabase(openRequest);
                openRequest.onsuccess = () => {
                    const db = openRequest.result;
                    const tx = db.transaction(INFO_STORE_NAME, 'readonly');
                    const store = tx.objectStore(INFO_STORE_NAME);
                    // tslint:disable:max-line-length
                    // Need to cast `store` as `any` here because TypeScript's DOM
                    // library does not have the `getAll()` method even though the
                    // method is supported in the latest version of most mainstream
                    // browsers:
                    // https://developer.mozilla.org/en-US/docs/Web/API/IDBObjectStore/getAll
                    // tslint:enable:max-line-length
                    // tslint:disable-next-line:no-any
                    const getAllInfoRequest = store.getAll();
                    getAllInfoRequest.onsuccess = () => {
                        const out = {};
                        for (const item of getAllInfoRequest.result) {
                            out[item.modelPath] = item.modelArtifactsInfo;
                        }
                        resolve(out);
                    };
                    getAllInfoRequest.onerror = error => {
                        db.close();
                        return reject(getAllInfoRequest.error);
                    };
                    tx.oncomplete = () => db.close();
                };
                openRequest.onerror = error => reject(openRequest.error);
            });
        }
        async removeModel(path) {
            path = maybeStripScheme$1(path);
            return new Promise((resolve, reject) => {
                const openRequest = this.indexedDB.open(DATABASE_NAME, DATABASE_VERSION);
                openRequest.onupgradeneeded = () => setUpDatabase(openRequest);
                openRequest.onsuccess = () => {
                    const db = openRequest.result;
                    const infoTx = db.transaction(INFO_STORE_NAME, 'readwrite');
                    const infoStore = infoTx.objectStore(INFO_STORE_NAME);
                    const getInfoRequest = infoStore.get(path);
                    let modelTx;
                    getInfoRequest.onsuccess = () => {
                        if (getInfoRequest.result == null) {
                            db.close();
                            return reject(new Error(`Cannot find model with path '${path}' ` +
                                `in IndexedDB.`));
                        }
                        else {
                            // First, delete the entry in the info store.
                            const deleteInfoRequest = infoStore.delete(path);
                            const deleteModelData = () => {
                                // Second, delete the entry in the model store.
                                modelTx = db.transaction(MODEL_STORE_NAME, 'readwrite');
                                const modelStore = modelTx.objectStore(MODEL_STORE_NAME);
                                const deleteModelRequest = modelStore.delete(path);
                                deleteModelRequest.onsuccess = () => resolve(getInfoRequest.result.modelArtifactsInfo);
                                deleteModelRequest.onerror = error => reject(getInfoRequest.error);
                            };
                            // Proceed with deleting model data regardless of whether deletion
                            // of info data succeeds or not.
                            deleteInfoRequest.onsuccess = deleteModelData;
                            deleteInfoRequest.onerror = error => {
                                deleteModelData();
                                db.close();
                                return reject(getInfoRequest.error);
                            };
                        }
                    };
                    getInfoRequest.onerror = error => {
                        db.close();
                        return reject(getInfoRequest.error);
                    };
                    infoTx.oncomplete = () => {
                        if (modelTx == null) {
                            db.close();
                        }
                        else {
                            modelTx.oncomplete = () => db.close();
                        }
                    };
                };
                openRequest.onerror = error => reject(openRequest.error);
            });
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const PATH_SEPARATOR = '/';
    const PATH_PREFIX = 'tensorflowjs_models';
    const INFO_SUFFIX = 'info';
    const MODEL_TOPOLOGY_SUFFIX = 'model_topology';
    const WEIGHT_SPECS_SUFFIX = 'weight_specs';
    const WEIGHT_DATA_SUFFIX = 'weight_data';
    const MODEL_METADATA_SUFFIX = 'model_metadata';
    function getModelKeys(path) {
        return {
            info: [PATH_PREFIX, path, INFO_SUFFIX].join(PATH_SEPARATOR),
            topology: [PATH_PREFIX, path, MODEL_TOPOLOGY_SUFFIX].join(PATH_SEPARATOR),
            weightSpecs: [PATH_PREFIX, path, WEIGHT_SPECS_SUFFIX].join(PATH_SEPARATOR),
            weightData: [PATH_PREFIX, path, WEIGHT_DATA_SUFFIX].join(PATH_SEPARATOR),
            modelMetadata: [PATH_PREFIX, path, MODEL_METADATA_SUFFIX].join(PATH_SEPARATOR)
        };
    }
    function removeItems(keys) {
        for (const key of Object.values(keys)) {
            window.localStorage.removeItem(key);
        }
    }
    /**
     * Get model path from a local-storage key.
     *
     * E.g., 'tensorflowjs_models/my/model/1/info' --> 'my/model/1'
     *
     * @param key
     */
    function getModelPathFromKey(key) {
        const items = key.split(PATH_SEPARATOR);
        if (items.length < 3) {
            throw new Error(`Invalid key format: ${key}`);
        }
        return items.slice(1, items.length - 1).join(PATH_SEPARATOR);
    }
    function maybeStripScheme(key) {
        return key.startsWith(BrowserLocalStorage.URL_SCHEME) ?
            key.slice(BrowserLocalStorage.URL_SCHEME.length) :
            key;
    }
    /**
     * IOHandler subclass: Browser Local Storage.
     *
     * See the doc string to `browserLocalStorage` for more details.
     */
    class BrowserLocalStorage {
        constructor(modelPath) {
            if (!env$1().getBool('IS_BROWSER') || typeof window === 'undefined' ||
                typeof window.localStorage === 'undefined') {
                // TODO(cais): Add more info about what IOHandler subtypes are
                // available.
                //   Maybe point to a doc page on the web and/or automatically determine
                //   the available IOHandlers and print them in the error message.
                throw new Error('The current environment does not support local storage.');
            }
            this.LS = window.localStorage;
            if (modelPath == null || !modelPath) {
                throw new Error('For local storage, modelPath must not be null, undefined or empty.');
            }
            this.modelPath = modelPath;
            this.keys = getModelKeys(this.modelPath);
        }
        /**
         * Save model artifacts to browser local storage.
         *
         * See the documentation to `browserLocalStorage` for details on the saved
         * artifacts.
         *
         * @param modelArtifacts The model artifacts to be stored.
         * @returns An instance of SaveResult.
         */
        async save(modelArtifacts) {
            if (modelArtifacts.modelTopology instanceof ArrayBuffer) {
                throw new Error('BrowserLocalStorage.save() does not support saving model topology ' +
                    'in binary formats yet.');
            }
            else {
                const topology = JSON.stringify(modelArtifacts.modelTopology);
                const weightSpecs = JSON.stringify(modelArtifacts.weightSpecs);
                const modelArtifactsInfo = getModelArtifactsInfoForJSON(modelArtifacts);
                try {
                    this.LS.setItem(this.keys.info, JSON.stringify(modelArtifactsInfo));
                    this.LS.setItem(this.keys.topology, topology);
                    this.LS.setItem(this.keys.weightSpecs, weightSpecs);
                    this.LS.setItem(this.keys.weightData, arrayBufferToBase64String(modelArtifacts.weightData));
                    // Note that JSON.stringify doesn't write out keys that have undefined
                    // values, so for some keys, we set undefined instead of a null-ish
                    // value.
                    const metadata = {
                        format: modelArtifacts.format,
                        generatedBy: modelArtifacts.generatedBy,
                        convertedBy: modelArtifacts.convertedBy,
                        signature: modelArtifacts.signature != null ?
                            modelArtifacts.signature :
                            undefined,
                        userDefinedMetadata: modelArtifacts.userDefinedMetadata != null ?
                            modelArtifacts.userDefinedMetadata :
                            undefined,
                        modelInitializer: modelArtifacts.modelInitializer != null ?
                            modelArtifacts.modelInitializer :
                            undefined,
                        trainingConfig: modelArtifacts.trainingConfig != null ?
                            modelArtifacts.trainingConfig :
                            undefined
                    };
                    this.LS.setItem(this.keys.modelMetadata, JSON.stringify(metadata));
                    return { modelArtifactsInfo };
                }
                catch (err) {
                    // If saving failed, clean up all items saved so far.
                    removeItems(this.keys);
                    throw new Error(`Failed to save model '${this.modelPath}' to local storage: ` +
                        `size quota being exceeded is a possible cause of this failure: ` +
                        `modelTopologyBytes=${modelArtifactsInfo.modelTopologyBytes}, ` +
                        `weightSpecsBytes=${modelArtifactsInfo.weightSpecsBytes}, ` +
                        `weightDataBytes=${modelArtifactsInfo.weightDataBytes}.`);
                }
            }
        }
        /**
         * Load a model from local storage.
         *
         * See the documentation to `browserLocalStorage` for details on the saved
         * artifacts.
         *
         * @returns The loaded model (if loading succeeds).
         */
        async load() {
            const info = JSON.parse(this.LS.getItem(this.keys.info));
            if (info == null) {
                throw new Error(`In local storage, there is no model with name '${this.modelPath}'`);
            }
            if (info.modelTopologyType !== 'JSON') {
                throw new Error('BrowserLocalStorage does not support loading non-JSON model ' +
                    'topology yet.');
            }
            const out = {};
            // Load topology.
            const topology = JSON.parse(this.LS.getItem(this.keys.topology));
            if (topology == null) {
                throw new Error(`In local storage, the topology of model '${this.modelPath}' ` +
                    `is missing.`);
            }
            out.modelTopology = topology;
            // Load weight specs.
            const weightSpecs = JSON.parse(this.LS.getItem(this.keys.weightSpecs));
            if (weightSpecs == null) {
                throw new Error(`In local storage, the weight specs of model '${this.modelPath}' ` +
                    `are missing.`);
            }
            out.weightSpecs = weightSpecs;
            // Load meta-data fields.
            const metadataString = this.LS.getItem(this.keys.modelMetadata);
            if (metadataString != null) {
                const metadata = JSON.parse(metadataString);
                out.format = metadata.format;
                out.generatedBy = metadata.generatedBy;
                out.convertedBy = metadata.convertedBy;
                if (metadata.signature != null) {
                    out.signature = metadata.signature;
                }
                if (metadata.userDefinedMetadata != null) {
                    out.userDefinedMetadata = metadata.userDefinedMetadata;
                }
                if (metadata.modelInitializer != null) {
                    out.modelInitializer = metadata.modelInitializer;
                }
                if (metadata.trainingConfig != null) {
                    out.trainingConfig = metadata.trainingConfig;
                }
            }
            // Load weight data.
            const weightDataBase64 = this.LS.getItem(this.keys.weightData);
            if (weightDataBase64 == null) {
                throw new Error(`In local storage, the binary weight values of model ` +
                    `'${this.modelPath}' are missing.`);
            }
            out.weightData = base64StringToArrayBuffer(weightDataBase64);
            return out;
        }
    }
    BrowserLocalStorage.URL_SCHEME = 'localstorage://';
    const localStorageRouter = (url) => {
        if (!env$1().getBool('IS_BROWSER')) {
            return null;
        }
        else {
            if (!Array.isArray(url) && url.startsWith(BrowserLocalStorage.URL_SCHEME)) {
                return browserLocalStorage(url.slice(BrowserLocalStorage.URL_SCHEME.length));
            }
            else {
                return null;
            }
        }
    };
    IORouterRegistry.registerSaveRouter(localStorageRouter);
    IORouterRegistry.registerLoadRouter(localStorageRouter);
    /**
     * Factory function for local storage IOHandler.
     *
     * This `IOHandler` supports both `save` and `load`.
     *
     * For each model's saved artifacts, four items are saved to local storage.
     *   - `${PATH_SEPARATOR}/${modelPath}/info`: Contains meta-info about the
     *     model, such as date saved, type of the topology, size in bytes, etc.
     *   - `${PATH_SEPARATOR}/${modelPath}/topology`: Model topology. For Keras-
     *     style models, this is a stringized JSON.
     *   - `${PATH_SEPARATOR}/${modelPath}/weight_specs`: Weight specs of the
     *     model, can be used to decode the saved binary weight values (see
     *     item below).
     *   - `${PATH_SEPARATOR}/${modelPath}/weight_data`: Concatenated binary
     *     weight values, stored as a base64-encoded string.
     *
     * Saving may throw an `Error` if the total size of the artifacts exceed the
     * browser-specific quota.
     *
     * @param modelPath A unique identifier for the model to be saved. Must be a
     *   non-empty string.
     * @returns An instance of `IOHandler`, which can be used with, e.g.,
     *   `tf.Model.save`.
     */
    function browserLocalStorage(modelPath) {
        return new BrowserLocalStorage(modelPath);
    }
    class BrowserLocalStorageManager {
        constructor() {
            assert$4(env$1().getBool('IS_BROWSER'), () => 'Current environment is not a web browser');
            assert$4(typeof window === 'undefined' ||
                typeof window.localStorage !== 'undefined', () => 'Current browser does not appear to support localStorage');
            this.LS = window.localStorage;
        }
        async listModels() {
            const out = {};
            const prefix = PATH_PREFIX + PATH_SEPARATOR;
            const suffix = PATH_SEPARATOR + INFO_SUFFIX;
            for (let i = 0; i < this.LS.length; ++i) {
                const key = this.LS.key(i);
                if (key.startsWith(prefix) && key.endsWith(suffix)) {
                    const modelPath = getModelPathFromKey(key);
                    out[modelPath] = JSON.parse(this.LS.getItem(key));
                }
            }
            return out;
        }
        async removeModel(path) {
            path = maybeStripScheme(path);
            const keys = getModelKeys(path);
            if (this.LS.getItem(keys.info) == null) {
                throw new Error(`Cannot find model at path '${path}'`);
            }
            const info = JSON.parse(this.LS.getItem(keys.info));
            removeItems(keys);
            return info;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const URL_SCHEME_SUFFIX = '://';
    class ModelStoreManagerRegistry {
        constructor() {
            this.managers = {};
        }
        static getInstance() {
            if (ModelStoreManagerRegistry.instance == null) {
                ModelStoreManagerRegistry.instance = new ModelStoreManagerRegistry();
            }
            return ModelStoreManagerRegistry.instance;
        }
        /**
         * Register a save-handler router.
         *
         * @param saveRouter A function that maps a URL-like string onto an instance
         * of `IOHandler` with the `save` method defined or `null`.
         */
        static registerManager(scheme, manager) {
            assert$4(scheme != null, () => 'scheme must not be undefined or null.');
            if (scheme.endsWith(URL_SCHEME_SUFFIX)) {
                scheme = scheme.slice(0, scheme.indexOf(URL_SCHEME_SUFFIX));
            }
            assert$4(scheme.length > 0, () => 'scheme must not be an empty string.');
            const registry = ModelStoreManagerRegistry.getInstance();
            assert$4(registry.managers[scheme] == null, () => `A model store manager is already registered for scheme '${scheme}'.`);
            registry.managers[scheme] = manager;
        }
        static getManager(scheme) {
            const manager = this.getInstance().managers[scheme];
            if (manager == null) {
                throw new Error(`Cannot find model manager for scheme '${scheme}'`);
            }
            return manager;
        }
        static getSchemes() {
            return Object.keys(this.getInstance().managers);
        }
    }
    /**
     * Helper method for parsing a URL string into a scheme and a path.
     *
     * @param url E.g., 'localstorage://my-model'
     * @returns A dictionary with two fields: scheme and path.
     *   Scheme: e.g., 'localstorage' in the example above.
     *   Path: e.g., 'my-model' in the example above.
     */
    function parseURL(url) {
        if (url.indexOf(URL_SCHEME_SUFFIX) === -1) {
            throw new Error(`The url string provided does not contain a scheme. ` +
                `Supported schemes are: ` +
                `${ModelStoreManagerRegistry.getSchemes().join(',')}`);
        }
        return {
            scheme: url.split(URL_SCHEME_SUFFIX)[0],
            path: url.split(URL_SCHEME_SUFFIX)[1],
        };
    }
    async function cloneModelInternal(sourceURL, destURL, deleteSource = false) {
        assert$4(sourceURL !== destURL, () => `Old path and new path are the same: '${sourceURL}'`);
        const loadHandlers = IORouterRegistry.getLoadHandlers(sourceURL);
        assert$4(loadHandlers.length > 0, () => `Copying failed because no load handler is found for source URL ${sourceURL}.`);
        assert$4(loadHandlers.length < 2, () => `Copying failed because more than one (${loadHandlers.length}) ` +
            `load handlers for source URL ${sourceURL}.`);
        const loadHandler = loadHandlers[0];
        const saveHandlers = IORouterRegistry.getSaveHandlers(destURL);
        assert$4(saveHandlers.length > 0, () => `Copying failed because no save handler is found for destination ` +
            `URL ${destURL}.`);
        assert$4(saveHandlers.length < 2, () => `Copying failed because more than one (${loadHandlers.length}) ` +
            `save handlers for destination URL ${destURL}.`);
        const saveHandler = saveHandlers[0];
        const sourceScheme = parseURL(sourceURL).scheme;
        const sourcePath = parseURL(sourceURL).path;
        const sameMedium = sourceScheme === parseURL(sourceURL).scheme;
        const modelArtifacts = await loadHandler.load();
        // If moving within the same storage medium, remove the old model as soon as
        // the loading is done. Without doing this, it is possible that the combined
        // size of the two models will cause the cloning to fail.
        if (deleteSource && sameMedium) {
            await ModelStoreManagerRegistry.getManager(sourceScheme)
                .removeModel(sourcePath);
        }
        const saveResult = await saveHandler.save(modelArtifacts);
        // If moving between mediums, the deletion is done after the save succeeds.
        // This guards against the case in which saving to the destination medium
        // fails.
        if (deleteSource && !sameMedium) {
            await ModelStoreManagerRegistry.getManager(sourceScheme)
                .removeModel(sourcePath);
        }
        return saveResult.modelArtifactsInfo;
    }
    /**
     * List all models stored in registered storage mediums.
     *
     * For a web browser environment, the registered mediums are Local Storage and
     * IndexedDB.
     *
     * ```js
     * // First create and save a model.
     * const model = tf.sequential();
     * model.add(tf.layers.dense(
     *     {units: 1, inputShape: [10], activation: 'sigmoid'}));
     * await model.save('localstorage://demo/management/model1');
     *
     * // Then list existing models.
     * console.log(JSON.stringify(await tf.io.listModels()));
     *
     * // Delete the model.
     * await tf.io.removeModel('localstorage://demo/management/model1');
     *
     * // List models again.
     * console.log(JSON.stringify(await tf.io.listModels()));
     * ```
     *
     * @returns A `Promise` of a dictionary mapping URLs of existing models to
     * their model artifacts info. URLs include medium-specific schemes, e.g.,
     *   'indexeddb://my/model/1'. Model artifacts info include type of the
     * model's topology, byte sizes of the topology, weights, etc.
     *
     * @doc {
     *   heading: 'Models',
     *   subheading: 'Management',
     *   namespace: 'io',
     *   ignoreCI: true
     * }
     */
    async function listModels() {
        const schemes = ModelStoreManagerRegistry.getSchemes();
        const out = {};
        for (const scheme of schemes) {
            const schemeOut = await ModelStoreManagerRegistry.getManager(scheme).listModels();
            for (const path in schemeOut) {
                const url = scheme + URL_SCHEME_SUFFIX + path;
                out[url] = schemeOut[path];
            }
        }
        return out;
    }
    /**
     * Remove a model specified by URL from a reigstered storage medium.
     *
     * ```js
     * // First create and save a model.
     * const model = tf.sequential();
     * model.add(tf.layers.dense(
     *     {units: 1, inputShape: [10], activation: 'sigmoid'}));
     * await model.save('localstorage://demo/management/model1');
     *
     * // Then list existing models.
     * console.log(JSON.stringify(await tf.io.listModels()));
     *
     * // Delete the model.
     * await tf.io.removeModel('localstorage://demo/management/model1');
     *
     * // List models again.
     * console.log(JSON.stringify(await tf.io.listModels()));
     * ```
     *
     * @param url A URL to a stored model, with a scheme prefix, e.g.,
     *   'localstorage://my-model-1', 'indexeddb://my/model/2'.
     * @returns ModelArtifactsInfo of the deleted model (if and only if deletion
     *   is successful).
     * @throws Error if deletion fails, e.g., if no model exists at `path`.
     *
     * @doc {
     *   heading: 'Models',
     *   subheading: 'Management',
     *   namespace: 'io',
     *   ignoreCI: true
     * }
     */
    async function removeModel(url) {
        const schemeAndPath = parseURL(url);
        const manager = ModelStoreManagerRegistry.getManager(schemeAndPath.scheme);
        return manager.removeModel(schemeAndPath.path);
    }
    /**
     * Copy a model from one URL to another.
     *
     * This function supports:
     *
     * 1. Copying within a storage medium, e.g.,
     *    `tf.io.copyModel('localstorage://model-1', 'localstorage://model-2')`
     * 2. Copying between two storage mediums, e.g.,
     *    `tf.io.copyModel('localstorage://model-1', 'indexeddb://model-1')`
     *
     * ```js
     * // First create and save a model.
     * const model = tf.sequential();
     * model.add(tf.layers.dense(
     *     {units: 1, inputShape: [10], activation: 'sigmoid'}));
     * await model.save('localstorage://demo/management/model1');
     *
     * // Then list existing models.
     * console.log(JSON.stringify(await tf.io.listModels()));
     *
     * // Copy the model, from Local Storage to IndexedDB.
     * await tf.io.copyModel(
     *     'localstorage://demo/management/model1',
     *     'indexeddb://demo/management/model1');
     *
     * // List models again.
     * console.log(JSON.stringify(await tf.io.listModels()));
     *
     * // Remove both models.
     * await tf.io.removeModel('localstorage://demo/management/model1');
     * await tf.io.removeModel('indexeddb://demo/management/model1');
     * ```
     *
     * @param sourceURL Source URL of copying.
     * @param destURL Destination URL of copying.
     * @returns ModelArtifactsInfo of the copied model (if and only if copying
     *   is successful).
     * @throws Error if copying fails, e.g., if no model exists at `sourceURL`, or
     *   if `oldPath` and `newPath` are identical.
     *
     * @doc {
     *   heading: 'Models',
     *   subheading: 'Management',
     *   namespace: 'io',
     *   ignoreCI: true
     * }
     */
    async function copyModel(sourceURL, destURL) {
        const deleteSource = false;
        return cloneModelInternal(sourceURL, destURL, deleteSource);
    }
    /**
     * Move a model from one URL to another.
     *
     * This function supports:
     *
     * 1. Moving within a storage medium, e.g.,
     *    `tf.io.moveModel('localstorage://model-1', 'localstorage://model-2')`
     * 2. Moving between two storage mediums, e.g.,
     *    `tf.io.moveModel('localstorage://model-1', 'indexeddb://model-1')`
     *
     * ```js
     * // First create and save a model.
     * const model = tf.sequential();
     * model.add(tf.layers.dense(
     *     {units: 1, inputShape: [10], activation: 'sigmoid'}));
     * await model.save('localstorage://demo/management/model1');
     *
     * // Then list existing models.
     * console.log(JSON.stringify(await tf.io.listModels()));
     *
     * // Move the model, from Local Storage to IndexedDB.
     * await tf.io.moveModel(
     *     'localstorage://demo/management/model1',
     *     'indexeddb://demo/management/model1');
     *
     * // List models again.
     * console.log(JSON.stringify(await tf.io.listModels()));
     *
     * // Remove the moved model.
     * await tf.io.removeModel('indexeddb://demo/management/model1');
     * ```
     *
     * @param sourceURL Source URL of moving.
     * @param destURL Destination URL of moving.
     * @returns ModelArtifactsInfo of the copied model (if and only if copying
     *   is successful).
     * @throws Error if moving fails, e.g., if no model exists at `sourceURL`, or
     *   if `oldPath` and `newPath` are identical.
     *
     * @doc {
     *   heading: 'Models',
     *   subheading: 'Management',
     *   namespace: 'io',
     *   ignoreCI: true
     * }
     */
    async function moveModel(sourceURL, destURL) {
        const deleteSource = true;
        return cloneModelInternal(sourceURL, destURL, deleteSource);
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class PlatformBrowser {
        fetch(path, init) {
            return fetch(path, init);
        }
        now() {
            return performance.now();
        }
        encode(text, encoding) {
            if (encoding !== 'utf-8' && encoding !== 'utf8') {
                throw new Error(`Browser's encoder only supports utf-8, but got ${encoding}`);
            }
            if (this.textEncoder == null) {
                this.textEncoder = new TextEncoder();
            }
            return this.textEncoder.encode(text);
        }
        decode(bytes, encoding) {
            return new TextDecoder(encoding).decode(bytes);
        }
    }
    if (env$1().get('IS_BROWSER')) {
        env$1().setPlatform('browser', new PlatformBrowser());
        // Register LocalStorage IOHandler
        try {
            ModelStoreManagerRegistry.registerManager(BrowserLocalStorage.URL_SCHEME, new BrowserLocalStorageManager());
        }
        catch (err) {
        }
        // Register IndexedDB IOHandler
        try {
            ModelStoreManagerRegistry.registerManager(BrowserIndexedDB.URL_SCHEME, new BrowserIndexedDBManager());
        }
        catch (err) {
        }
    }

    // We are wrapping this within an object so it can be stubbed by Jasmine.
    const getNodeFetch = {
        // tslint:disable-next-line:no-require-imports
        importFetch: () => require('node-fetch')
    };
    let systemFetch;
    class PlatformNode {
        constructor() {
            // tslint:disable-next-line:no-require-imports
            this.util = require('util');
            // According to the spec, the built-in encoder can do only UTF-8 encoding.
            // https://developer.mozilla.org/en-US/docs/Web/API/TextEncoder/TextEncoder
            this.textEncoder = new this.util.TextEncoder();
        }
        fetch(path, requestInits) {
            if (env$1().global.fetch != null) {
                return env$1().global.fetch(path, requestInits);
            }
            if (systemFetch == null) {
                systemFetch = getNodeFetch.importFetch();
            }
            return systemFetch(path, requestInits);
        }
        now() {
            const time = process.hrtime();
            return time[0] * 1000 + time[1] / 1000000;
        }
        encode(text, encoding) {
            if (encoding !== 'utf-8' && encoding !== 'utf8') {
                throw new Error(`Node built-in encoder only supports utf-8, but got ${encoding}`);
            }
            return this.textEncoder.encode(text);
        }
        decode(bytes, encoding) {
            if (bytes.length === 0) {
                return '';
            }
            return new this.util.TextDecoder(encoding).decode(bytes);
        }
    }
    if (env$1().get('IS_NODE') && !env$1().get('IS_BROWSER')) {
        env$1().setPlatform('node', new PlatformNode());
    }

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates an empty `tf.TensorBuffer` with the specified `shape` and `dtype`.
     *
     * The values are stored in CPU as `TypedArray`. Fill the buffer using
     * `buffer.set()`, or by modifying directly `buffer.values`.
     *
     * When done, call `buffer.toTensor()` to get an immutable `tf.Tensor` with
     * those values.
     *
     * ```js
     * // Create a buffer and set values at particular indices.
     * const buffer = tf.buffer([2, 2]);
     * buffer.set(3, 0, 0);
     * buffer.set(5, 1, 0);
     *
     * // Convert the buffer back to a tensor.
     * buffer.toTensor().print();
     * ```
     *
     * @param shape An array of integers defining the output tensor shape.
     * @param dtype The dtype of the buffer. Defaults to 'float32'.
     * @param values The values of the buffer as `TypedArray`. Defaults to
     * zeros.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function buffer(shape, dtype = 'float32', values) {
        dtype = dtype || 'float32';
        assertNonNegativeIntegerDimensions(shape);
        return new TensorBuffer(shape, dtype, values);
    }

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Casts a `tf.Tensor` to a new dtype.
     *
     * ```js
     * const x = tf.tensor1d([1.5, 2.5, 3]);
     * tf.cast(x, 'int32').print();
     * ```
     * @param x The input tensor to be casted.
     * @param dtype The dtype to cast the input tensor to.
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function cast_(x, dtype) {
        const $x = convertToTensor(x, 'x', 'cast');
        // Sanity checks.
        if (!isValidDtype(dtype)) {
            throw new Error(`Failed to cast to unknown dtype ${dtype}`);
        }
        if (dtype === 'string' && $x.dtype !== 'string' ||
            dtype !== 'string' && $x.dtype === 'string') {
            throw new Error('Only strings can be casted to strings');
        }
        const inputs = { x: $x };
        const attrs = { dtype };
        return ENGINE.runKernel(Cast, inputs, attrs);
    }
    const cast$3 = op({ cast_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a new tensor with the same values and shape as the specified
     * tensor.
     *
     * ```js
     * const x = tf.tensor([1, 2]);
     *
     * x.clone().print();
     * ```
     *
     * @param x The tensor to clone.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function clone_(x) {
        const $x = convertToTensor(x, 'x', 'clone', 'string_or_numeric');
        const inputs = { x: $x };
        // Note this op is called tf.identity in python. Hence the kernel name used
        // here.
        return ENGINE.runKernel(Identity$1, inputs);
    }
    const clone = op({ clone_ });

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Prints information about the `tf.Tensor` including its data.
     *
     * ```js
     * const verbose = true;
     * tf.tensor2d([1, 2, 3, 4], [2, 2]).print(verbose);
     * ```
     * @param x The tensor to be printed.
     * @param verbose Whether to print verbose information about the ` Tensor`,
     * including dtype and size.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function print(x, verbose = false) {
        console.log(x.toString(verbose));
    }

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getOrMakeEngine();
    const opHandler = {
        buffer,
        cast: cast$3,
        clone,
        print
    };
    setOpHandler(opHandler);

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const DEFAULT_FILE_NAME_PREFIX = 'model';
    const DEFAULT_JSON_EXTENSION_NAME = '.json';
    const DEFAULT_WEIGHT_DATA_EXTENSION_NAME = '.weights.bin';
    function defer(f) {
        return new Promise(resolve => setTimeout(resolve)).then(f);
    }
    class BrowserDownloads {
        constructor(fileNamePrefix) {
            if (!env$1().getBool('IS_BROWSER')) {
                // TODO(cais): Provide info on what IOHandlers are available under the
                //   current environment.
                throw new Error('browserDownloads() cannot proceed because the current environment ' +
                    'is not a browser.');
            }
            if (fileNamePrefix.startsWith(BrowserDownloads.URL_SCHEME)) {
                fileNamePrefix = fileNamePrefix.slice(BrowserDownloads.URL_SCHEME.length);
            }
            if (fileNamePrefix == null || fileNamePrefix.length === 0) {
                fileNamePrefix = DEFAULT_FILE_NAME_PREFIX;
            }
            this.modelJsonFileName = fileNamePrefix + DEFAULT_JSON_EXTENSION_NAME;
            this.weightDataFileName =
                fileNamePrefix + DEFAULT_WEIGHT_DATA_EXTENSION_NAME;
        }
        async save(modelArtifacts) {
            if (typeof (document) === 'undefined') {
                throw new Error('Browser downloads are not supported in ' +
                    'this environment since `document` is not present');
            }
            const weightsURL = window.URL.createObjectURL(new Blob([modelArtifacts.weightData], { type: 'application/octet-stream' }));
            if (modelArtifacts.modelTopology instanceof ArrayBuffer) {
                throw new Error('BrowserDownloads.save() does not support saving model topology ' +
                    'in binary formats yet.');
            }
            else {
                const weightsManifest = [{
                        paths: ['./' + this.weightDataFileName],
                        weights: modelArtifacts.weightSpecs
                    }];
                const modelJSON = getModelJSONForModelArtifacts(modelArtifacts, weightsManifest);
                const modelJsonURL = window.URL.createObjectURL(new Blob([JSON.stringify(modelJSON)], { type: 'application/json' }));
                // If anchor elements are not provided, create them without attaching them
                // to parents, so that the downloaded file names can be controlled.
                const jsonAnchor = this.modelJsonAnchor == null ?
                    document.createElement('a') :
                    this.modelJsonAnchor;
                jsonAnchor.download = this.modelJsonFileName;
                jsonAnchor.href = modelJsonURL;
                // Trigger downloads by evoking a click event on the download anchors.
                // When multiple downloads are started synchronously, Firefox will only
                // save the last one.
                await defer(() => jsonAnchor.dispatchEvent(new MouseEvent('click')));
                if (modelArtifacts.weightData != null) {
                    const weightDataAnchor = this.weightDataAnchor == null ?
                        document.createElement('a') :
                        this.weightDataAnchor;
                    weightDataAnchor.download = this.weightDataFileName;
                    weightDataAnchor.href = weightsURL;
                    await defer(() => weightDataAnchor.dispatchEvent(new MouseEvent('click')));
                }
                return { modelArtifactsInfo: getModelArtifactsInfoForJSON(modelArtifacts) };
            }
        }
    }
    BrowserDownloads.URL_SCHEME = 'downloads://';
    class BrowserFiles {
        constructor(files) {
            if (files == null || files.length < 1) {
                throw new Error(`When calling browserFiles, at least 1 file is required, ` +
                    `but received ${files}`);
            }
            this.jsonFile = files[0];
            this.weightsFiles = files.slice(1);
        }
        async load() {
            return new Promise((resolve, reject) => {
                const jsonReader = new FileReader();
                jsonReader.onload = (event) => {
                    // tslint:disable-next-line:no-any
                    const modelJSON = JSON.parse(event.target.result);
                    const modelTopology = modelJSON.modelTopology;
                    if (modelTopology == null) {
                        reject(new Error(`modelTopology field is missing from file ${this.jsonFile.name}`));
                        return;
                    }
                    const weightsManifest = modelJSON.weightsManifest;
                    if (weightsManifest == null) {
                        reject(new Error(`weightManifest field is missing from file ${this.jsonFile.name}`));
                        return;
                    }
                    if (this.weightsFiles.length === 0) {
                        resolve({ modelTopology });
                        return;
                    }
                    const modelArtifactsPromise = getModelArtifactsForJSON(modelJSON, (weightsManifest) => this.loadWeights(weightsManifest));
                    resolve(modelArtifactsPromise);
                };
                jsonReader.onerror = error => reject(`Failed to read model topology and weights manifest JSON ` +
                    `from file '${this.jsonFile.name}'. BrowserFiles supports loading ` +
                    `Keras-style tf.Model artifacts only.`);
                jsonReader.readAsText(this.jsonFile);
            });
        }
        loadWeights(weightsManifest) {
            const weightSpecs = [];
            const paths = [];
            for (const entry of weightsManifest) {
                weightSpecs.push(...entry.weights);
                paths.push(...entry.paths);
            }
            const pathToFile = this.checkManifestAndWeightFiles(weightsManifest);
            const promises = paths.map(path => this.loadWeightsFile(path, pathToFile[path]));
            return Promise.all(promises).then(buffers => [weightSpecs, concatenateArrayBuffers(buffers)]);
        }
        loadWeightsFile(path, file) {
            return new Promise((resolve, reject) => {
                const weightFileReader = new FileReader();
                weightFileReader.onload = (event) => {
                    // tslint:disable-next-line:no-any
                    const weightData = event.target.result;
                    resolve(weightData);
                };
                weightFileReader.onerror = error => reject(`Failed to weights data from file of path '${path}'.`);
                weightFileReader.readAsArrayBuffer(file);
            });
        }
        /**
         * Check the compatibility between weights manifest and weight files.
         */
        checkManifestAndWeightFiles(manifest) {
            const basenames = [];
            const fileNames = this.weightsFiles.map(file => basename$1(file.name));
            const pathToFile = {};
            for (const group of manifest) {
                group.paths.forEach(path => {
                    const pathBasename = basename$1(path);
                    if (basenames.indexOf(pathBasename) !== -1) {
                        throw new Error(`Duplicate file basename found in weights manifest: ` +
                            `'${pathBasename}'`);
                    }
                    basenames.push(pathBasename);
                    if (fileNames.indexOf(pathBasename) === -1) {
                        throw new Error(`Weight file with basename '${pathBasename}' is not provided.`);
                    }
                    else {
                        pathToFile[path] = this.weightsFiles[fileNames.indexOf(pathBasename)];
                    }
                });
            }
            if (basenames.length !== this.weightsFiles.length) {
                throw new Error(`Mismatch in the number of files in weights manifest ` +
                    `(${basenames.length}) and the number of weight files provided ` +
                    `(${this.weightsFiles.length}).`);
            }
            return pathToFile;
        }
    }
    const browserDownloadsRouter = (url) => {
        if (!env$1().getBool('IS_BROWSER')) {
            return null;
        }
        else {
            if (!Array.isArray(url) && url.startsWith(BrowserDownloads.URL_SCHEME)) {
                return browserDownloads(url.slice(BrowserDownloads.URL_SCHEME.length));
            }
            else {
                return null;
            }
        }
    };
    IORouterRegistry.registerSaveRouter(browserDownloadsRouter);
    /**
     * Creates an IOHandler that triggers file downloads from the browser.
     *
     * The returned `IOHandler` instance can be used as model exporting methods such
     * as `tf.Model.save` and supports only saving.
     *
     * ```js
     * const model = tf.sequential();
     * model.add(tf.layers.dense(
     *     {units: 1, inputShape: [10], activation: 'sigmoid'}));
     * const saveResult = await model.save('downloads://mymodel');
     * // This will trigger downloading of two files:
     * //   'mymodel.json' and 'mymodel.weights.bin'.
     * console.log(saveResult);
     * ```
     *
     * @param fileNamePrefix Prefix name of the files to be downloaded. For use with
     *   `tf.Model`, `fileNamePrefix` should follow either of the following two
     *   formats:
     *   1. `null` or `undefined`, in which case the default file
     *      names will be used:
     *      - 'model.json' for the JSON file containing the model topology and
     *        weights manifest.
     *      - 'model.weights.bin' for the binary file containing the binary weight
     *        values.
     *   2. A single string or an Array of a single string, as the file name prefix.
     *      For example, if `'foo'` is provided, the downloaded JSON
     *      file and binary weights file will be named 'foo.json' and
     *      'foo.weights.bin', respectively.
     * @param config Additional configuration for triggering downloads.
     * @returns An instance of `BrowserDownloads` `IOHandler`.
     *
     * @doc {
     *   heading: 'Models',
     *   subheading: 'Loading',
     *   namespace: 'io',
     *   ignoreCI: true
     * }
     */
    function browserDownloads(fileNamePrefix = 'model') {
        return new BrowserDownloads(fileNamePrefix);
    }
    /**
     * Creates an IOHandler that loads model artifacts from user-selected files.
     *
     * This method can be used for loading from files such as user-selected files
     * in the browser.
     * When used in conjunction with `tf.loadLayersModel`, an instance of
     * `tf.LayersModel` (Keras-style) can be constructed from the loaded artifacts.
     *
     * ```js
     * // Note: This code snippet won't run properly without the actual file input
     * //   elements in the HTML DOM.
     *
     * // Suppose there are two HTML file input (`<input type="file" ...>`)
     * // elements.
     * const uploadJSONInput = document.getElementById('upload-json');
     * const uploadWeightsInput = document.getElementById('upload-weights');
     * const model = await tf.loadLayersModel(tf.io.browserFiles(
     *     [uploadJSONInput.files[0], uploadWeightsInput.files[0]]));
     * ```
     *
     * @param files `File`s to load from. Currently, this function supports only
     *   loading from files that contain Keras-style models (i.e., `tf.Model`s), for
     *   which an `Array` of `File`s is expected (in that order):
     *   - A JSON file containing the model topology and weight manifest.
     *   - Optionally, One or more binary files containing the binary weights.
     *     These files must have names that match the paths in the `weightsManifest`
     *     contained by the aforementioned JSON file, or errors will be thrown
     *     during loading. These weights files have the same format as the ones
     *     generated by `tensorflowjs_converter` that comes with the `tensorflowjs`
     *     Python PIP package. If no weights files are provided, only the model
     *     topology will be loaded from the JSON file above.
     * @returns An instance of `Files` `IOHandler`.
     *
     * @doc {
     *   heading: 'Models',
     *   subheading: 'Loading',
     *   namespace: 'io',
     *   ignoreCI: true
     * }
     */
    function browserFiles(files) {
        return new BrowserFiles(files);
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Monitor Promise.all progress, fire onProgress callback function.
     *
     * @param promises Promise list going to be monitored
     * @param onProgress Callback function. Fired when a promise resolved.
     * @param startFraction Optional fraction start. Default to 0.
     * @param endFraction Optional fraction end. Default to 1.
     */
    function monitorPromisesProgress(promises, onProgress, startFraction, endFraction) {
        checkPromises(promises);
        startFraction = startFraction == null ? 0 : startFraction;
        endFraction = endFraction == null ? 1 : endFraction;
        checkFraction(startFraction, endFraction);
        let resolvedPromise = 0;
        const registerMonitor = (promise) => {
            promise.then(value => {
                const fraction = startFraction +
                    ++resolvedPromise / promises.length * (endFraction - startFraction);
                // pass fraction as parameter to callback function.
                onProgress(fraction);
                return value;
            });
            return promise;
        };
        function checkPromises(promises) {
            assert$4(promises != null && Array.isArray(promises) && promises.length > 0, () => 'promises must be a none empty array');
        }
        function checkFraction(startFraction, endFraction) {
            assert$4(startFraction >= 0 && startFraction <= 1, () => `Progress fraction must be in range [0, 1], but ` +
                `got startFraction ${startFraction}`);
            assert$4(endFraction >= 0 && endFraction <= 1, () => `Progress fraction must be in range [0, 1], but ` +
                `got endFraction ${endFraction}`);
            assert$4(endFraction >= startFraction, () => `startFraction must be no more than endFraction, but ` +
                `got startFraction ${startFraction} and endFraction ` +
                `${endFraction}`);
        }
        return Promise.all(promises.map(registerMonitor));
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Reads binary weights data from a number of URLs.
     *
     * @param fetchURLs URLs to send the HTTP requests at, using `fetch` calls.
     * @param requestOptions RequestInit (options) for the HTTP requests.
     * @param fetchFunc Optional overriding value for the `window.fetch` function.
     * @param onProgress Optional, progress callback function, fired periodically
     *   before the load is completed.
     * @returns A `Promise` of an Array of `ArrayBuffer`. The Array has the same
     *   length as `fetchURLs`.
     */
    async function loadWeightsAsArrayBuffer(fetchURLs, loadOptions) {
        if (loadOptions == null) {
            loadOptions = {};
        }
        const fetchFunc = loadOptions.fetchFunc == null ? env$1().platform.fetch :
            loadOptions.fetchFunc;
        // Create the requests for all of the weights in parallel.
        const requests = fetchURLs.map(fetchURL => fetchFunc(fetchURL, loadOptions.requestInit, { isBinary: true }));
        const fetchStartFraction = 0;
        const fetchEndFraction = 0.5;
        const responses = loadOptions.onProgress == null ?
            await Promise.all(requests) :
            await monitorPromisesProgress(requests, loadOptions.onProgress, fetchStartFraction, fetchEndFraction);
        const bufferPromises = responses.map(response => response.arrayBuffer());
        const bufferStartFraction = 0.5;
        const bufferEndFraction = 1;
        const buffers = loadOptions.onProgress == null ?
            await Promise.all(bufferPromises) :
            await monitorPromisesProgress(bufferPromises, loadOptions.onProgress, bufferStartFraction, bufferEndFraction);
        return buffers;
    }
    /**
     * Reads a weights manifest JSON configuration, fetches the weights and
     * returns them as `Tensor`s.
     *
     * @param manifest The weights manifest JSON.
     * @param filePathPrefix The path prefix for filenames given in the manifest.
     *     Defaults to the empty string.
     * @param weightNames The names of the weights to be fetched.
     */
    async function loadWeights(manifest, filePathPrefix = '', weightNames, requestInit) {
        // TODO(nsthorat): Groups are currently fetched atomically. If you need a
        // single weight from a group, the whole group will be fetched. At a future
        // date, we should support fetching only the individual shards within a
        // group that are needed to reconstruct the requested weight.
        // TODO(cais): Use `decodeWeights` for implementation.
        const fetchWeights = (fetchUrls) => loadWeightsAsArrayBuffer(fetchUrls, { requestInit });
        const loadWeights = weightsLoaderFactory(fetchWeights);
        return loadWeights(manifest, filePathPrefix, weightNames);
    }
    /**
     * Creates a function, which reads a weights manifest JSON configuration,
     * fetches the weight files using the specified function and returns them as
     * `Tensor`s.
     *
     * ```js
     * // example for creating a nodejs weight loader, which reads the weight files
     * // from disk using fs.readFileSync
     *
     * import * as fs from 'fs'
     *
     * const fetchWeightsFromDisk = (filePaths: string[]) =>
     *   filePaths.map(filePath => fs.readFileSync(filePath).buffer)
     *
     * const loadWeights = tf.io.weightsLoaderFactory(fetchWeightsFromDisk)
     *
     * const manifest = JSON.parse(
     *   fs.readFileSync('./my_model-weights_manifest').toString()
     * )
     * const weightMap = await loadWeights(manifest, './')
     * ```
     * @param fetchWeightsFunction The function used for fetching the weight files.
     * @returns Weight loading function.
     */
    function weightsLoaderFactory(fetchWeightsFunction) {
        return async (manifest, filePathPrefix = '', weightNames) => {
            // Collect all the groups, weights, and their relative offsets to be
            // fetched.
            const groupIndicesToFetchMap = manifest.map(() => false);
            const groupWeightsToFetch = {};
            const weightsFound = weightNames != null ? weightNames.map(() => false) : [];
            const allManifestWeightNames = [];
            manifest.forEach((manifestGroupConfig, groupIndex) => {
                let groupOffset = 0;
                manifestGroupConfig.weights.forEach(weightsEntry => {
                    const rawDtype = ('quantization' in weightsEntry) ?
                        weightsEntry.quantization.dtype :
                        weightsEntry.dtype;
                    const weightsBytes = DTYPE_VALUE_SIZE_MAP[rawDtype] *
                        sizeFromShape(weightsEntry.shape);
                    const enqueueWeightsForFetchingFn = () => {
                        groupIndicesToFetchMap[groupIndex] = true;
                        if (groupWeightsToFetch[groupIndex] == null) {
                            groupWeightsToFetch[groupIndex] = [];
                        }
                        groupWeightsToFetch[groupIndex].push({
                            manifestEntry: weightsEntry,
                            groupOffset,
                            sizeBytes: weightsBytes
                        });
                    };
                    if (weightNames != null) {
                        weightNames.forEach((weightName, weightIndex) => {
                            if (weightName === weightsEntry.name) {
                                enqueueWeightsForFetchingFn();
                                weightsFound[weightIndex] = true;
                            }
                        });
                    }
                    else {
                        enqueueWeightsForFetchingFn();
                    }
                    allManifestWeightNames.push(weightsEntry.name);
                    groupOffset += weightsBytes;
                });
            });
            if (!weightsFound.every(found => found)) {
                const weightsNotFound = weightNames.filter((_, i) => !weightsFound[i]);
                throw new Error(`Could not find weights in manifest with names: ` +
                    `${weightsNotFound.join(', ')}. \n` +
                    `Manifest JSON has weights with names: ` +
                    `${allManifestWeightNames.join(', ')}.`);
            }
            // Convert the one-hot boolean groupId => shouldFetch map to a list of group
            // IDs.
            const groupIndicesToFetch = groupIndicesToFetchMap.reduce((accumulator, shouldFetch, i) => {
                if (shouldFetch) {
                    accumulator.push(i);
                }
                return accumulator;
            }, []);
            const fetchUrls = [];
            groupIndicesToFetch.forEach(i => {
                manifest[i].paths.forEach(filepath => {
                    const fetchUrl = filePathPrefix +
                        (!filePathPrefix.endsWith('/') ? '/' : '') + filepath;
                    fetchUrls.push(fetchUrl);
                });
            });
            const buffers = await fetchWeightsFunction(fetchUrls);
            const weightsTensorMap = {};
            let bufferIndexOffset = 0;
            groupIndicesToFetch.forEach(i => {
                const numBuffers = manifest[i].paths.length;
                let groupBytes = 0;
                for (let i = 0; i < numBuffers; i++) {
                    groupBytes += buffers[bufferIndexOffset + i].byteLength;
                }
                // Create a buffer for the whole group.
                const groupBuffer = new ArrayBuffer(groupBytes);
                const groupByteBuffer = new Uint8Array(groupBuffer);
                let groupBufferOffset = 0;
                for (let i = 0; i < numBuffers; i++) {
                    const buffer = new Uint8Array(buffers[bufferIndexOffset + i]);
                    groupByteBuffer.set(buffer, groupBufferOffset);
                    groupBufferOffset += buffer.byteLength;
                }
                const weightsEntries = groupWeightsToFetch[i];
                weightsEntries.forEach(weightsEntry => {
                    const byteBuffer = groupBuffer.slice(weightsEntry.groupOffset, weightsEntry.groupOffset + weightsEntry.sizeBytes);
                    const nameToTensorMap = decodeWeights(byteBuffer, [weightsEntry.manifestEntry]);
                    for (const name in nameToTensorMap) {
                        weightsTensorMap[name] = nameToTensorMap[name];
                    }
                });
                bufferIndexOffset += numBuffers;
            });
            return weightsTensorMap;
        };
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const OCTET_STREAM_MIME_TYPE = 'application/octet-stream';
    const JSON_TYPE = 'application/json';
    class HTTPRequest {
        constructor(path, loadOptions) {
            this.DEFAULT_METHOD = 'POST';
            if (loadOptions == null) {
                loadOptions = {};
            }
            this.weightPathPrefix = loadOptions.weightPathPrefix;
            this.onProgress = loadOptions.onProgress;
            this.weightUrlConverter = loadOptions.weightUrlConverter;
            if (loadOptions.fetchFunc != null) {
                assert$4(typeof loadOptions.fetchFunc === 'function', () => 'Must pass a function that matches the signature of ' +
                    '`fetch` (see ' +
                    'https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API)');
                this.fetch = loadOptions.fetchFunc;
            }
            else {
                this.fetch = env$1().platform.fetch;
            }
            assert$4(path != null && path.length > 0, () => 'URL path for http must not be null, undefined or ' +
                'empty.');
            if (Array.isArray(path)) {
                assert$4(path.length === 2, () => 'URL paths for http must have a length of 2, ' +
                    `(actual length is ${path.length}).`);
            }
            this.path = path;
            if (loadOptions.requestInit != null &&
                loadOptions.requestInit.body != null) {
                throw new Error('requestInit is expected to have no pre-existing body, but has one.');
            }
            this.requestInit = loadOptions.requestInit || {};
        }
        async save(modelArtifacts) {
            if (modelArtifacts.modelTopology instanceof ArrayBuffer) {
                throw new Error('BrowserHTTPRequest.save() does not support saving model topology ' +
                    'in binary formats yet.');
            }
            const init = Object.assign({ method: this.DEFAULT_METHOD }, this.requestInit);
            init.body = new FormData();
            const weightsManifest = [{
                    paths: ['./model.weights.bin'],
                    weights: modelArtifacts.weightSpecs,
                }];
            const modelTopologyAndWeightManifest = getModelJSONForModelArtifacts(modelArtifacts, weightsManifest);
            init.body.append('model.json', new Blob([JSON.stringify(modelTopologyAndWeightManifest)], { type: JSON_TYPE }), 'model.json');
            if (modelArtifacts.weightData != null) {
                init.body.append('model.weights.bin', new Blob([modelArtifacts.weightData], { type: OCTET_STREAM_MIME_TYPE }), 'model.weights.bin');
            }
            const response = await this.fetch(this.path, init);
            if (response.ok) {
                return {
                    modelArtifactsInfo: getModelArtifactsInfoForJSON(modelArtifacts),
                    responses: [response],
                };
            }
            else {
                throw new Error(`BrowserHTTPRequest.save() failed due to HTTP response status ` +
                    `${response.status}.`);
            }
        }
        /**
         * Load model artifacts via HTTP request(s).
         *
         * See the documentation to `tf.io.http` for details on the saved
         * artifacts.
         *
         * @returns The loaded model artifacts (if loading succeeds).
         */
        async load() {
            const modelConfigRequest = await this.fetch(this.path, this.requestInit);
            if (!modelConfigRequest.ok) {
                throw new Error(`Request to ${this.path} failed with status code ` +
                    `${modelConfigRequest.status}. Please verify this URL points to ` +
                    `the model JSON of the model to load.`);
            }
            let modelJSON;
            try {
                modelJSON = await modelConfigRequest.json();
            }
            catch (e) {
                let message = `Failed to parse model JSON of response from ${this.path}.`;
                // TODO(nsthorat): Remove this after some time when we're comfortable that
                // .pb files are mostly gone.
                if (this.path.endsWith('.pb')) {
                    message += ' Your path contains a .pb file extension. ' +
                        'Support for .pb models have been removed in TensorFlow.js 1.0 ' +
                        'in favor of .json models. You can re-convert your Python ' +
                        'TensorFlow model using the TensorFlow.js 1.0 conversion scripts ' +
                        'or you can convert your.pb models with the \'pb2json\'' +
                        'NPM script in the tensorflow/tfjs-converter repository.';
                }
                else {
                    message += ' Please make sure the server is serving valid ' +
                        'JSON for this request.';
                }
                throw new Error(message);
            }
            // We do not allow both modelTopology and weightsManifest to be missing.
            const modelTopology = modelJSON.modelTopology;
            const weightsManifest = modelJSON.weightsManifest;
            if (modelTopology == null && weightsManifest == null) {
                throw new Error(`The JSON from HTTP path ${this.path} contains neither model ` +
                    `topology or manifest for weights.`);
            }
            return getModelArtifactsForJSON(modelJSON, (weightsManifest) => this.loadWeights(weightsManifest));
        }
        async loadWeights(weightsManifest) {
            const weightPath = Array.isArray(this.path) ? this.path[1] : this.path;
            const [prefix, suffix] = parseUrl(weightPath);
            const pathPrefix = this.weightPathPrefix || prefix;
            const weightSpecs = [];
            for (const entry of weightsManifest) {
                weightSpecs.push(...entry.weights);
            }
            const fetchURLs = [];
            const urlPromises = [];
            for (const weightsGroup of weightsManifest) {
                for (const path of weightsGroup.paths) {
                    if (this.weightUrlConverter != null) {
                        urlPromises.push(this.weightUrlConverter(path));
                    }
                    else {
                        fetchURLs.push(pathPrefix + path + suffix);
                    }
                }
            }
            if (this.weightUrlConverter) {
                fetchURLs.push(...await Promise.all(urlPromises));
            }
            const buffers = await loadWeightsAsArrayBuffer(fetchURLs, {
                requestInit: this.requestInit,
                fetchFunc: this.fetch,
                onProgress: this.onProgress
            });
            return [weightSpecs, concatenateArrayBuffers(buffers)];
        }
    }
    HTTPRequest.URL_SCHEME_REGEX = /^https?:\/\//;
    /**
     * Extract the prefix and suffix of the url, where the prefix is the path before
     * the last file, and suffix is the search params after the last file.
     * ```
     * const url = 'http://tfhub.dev/model/1/tensorflowjs_model.pb?tfjs-format=file'
     * [prefix, suffix] = parseUrl(url)
     * // prefix = 'http://tfhub.dev/model/1/'
     * // suffix = '?tfjs-format=file'
     * ```
     * @param url the model url to be parsed.
     */
    function parseUrl(url) {
        const lastSlash = url.lastIndexOf('/');
        const lastSearchParam = url.lastIndexOf('?');
        const prefix = url.substring(0, lastSlash);
        const suffix = lastSearchParam > lastSlash ? url.substring(lastSearchParam) : '';
        return [prefix + '/', suffix];
    }
    function isHTTPScheme(url) {
        return url.match(HTTPRequest.URL_SCHEME_REGEX) != null;
    }
    const httpRouter = (url, loadOptions) => {
        if (typeof fetch === 'undefined' &&
            (loadOptions == null || loadOptions.fetchFunc == null)) {
            // `http` uses `fetch` or `node-fetch`, if one wants to use it in
            // an environment that is not the browser or node they have to setup a
            // global fetch polyfill.
            return null;
        }
        else {
            let isHTTP = true;
            if (Array.isArray(url)) {
                isHTTP = url.every(urlItem => isHTTPScheme(urlItem));
            }
            else {
                isHTTP = isHTTPScheme(url);
            }
            if (isHTTP) {
                return http$2(url, loadOptions);
            }
        }
        return null;
    };
    IORouterRegistry.registerSaveRouter(httpRouter);
    IORouterRegistry.registerLoadRouter(httpRouter);
    /**
     * Creates an IOHandler subtype that sends model artifacts to HTTP server.
     *
     * An HTTP request of the `multipart/form-data` mime type will be sent to the
     * `path` URL. The form data includes artifacts that represent the topology
     * and/or weights of the model. In the case of Keras-style `tf.Model`, two
     * blobs (files) exist in form-data:
     *   - A JSON file consisting of `modelTopology` and `weightsManifest`.
     *   - A binary weights file consisting of the concatenated weight values.
     * These files are in the same format as the one generated by
     * [tfjs_converter](https://js.tensorflow.org/tutorials/import-keras.html).
     *
     * The following code snippet exemplifies the client-side code that uses this
     * function:
     *
     * ```js
     * const model = tf.sequential();
     * model.add(
     *     tf.layers.dense({units: 1, inputShape: [100], activation: 'sigmoid'}));
     *
     * const saveResult = await model.save(tf.io.http(
     *     'http://model-server:5000/upload', {requestInit: {method: 'PUT'}}));
     * console.log(saveResult);
     * ```
     *
     * If the default `POST` method is to be used, without any custom parameters
     * such as headers, you can simply pass an HTTP or HTTPS URL to `model.save`:
     *
     * ```js
     * const saveResult = await model.save('http://model-server:5000/upload');
     * ```
     *
     * The following GitHub Gist
     * https://gist.github.com/dsmilkov/1b6046fd6132d7408d5257b0976f7864
     * implements a server based on [flask](https://github.com/pallets/flask) that
     * can receive the request. Upon receiving the model artifacts via the requst,
     * this particular server reconsistutes instances of [Keras
     * Models](https://keras.io/models/model/) in memory.
     *
     *
     * @param path A URL path to the model.
     *   Can be an absolute HTTP path (e.g.,
     *   'http://localhost:8000/model-upload)') or a relative path (e.g.,
     *   './model-upload').
     * @param requestInit Request configurations to be used when sending
     *    HTTP request to server using `fetch`. It can contain fields such as
     *    `method`, `credentials`, `headers`, `mode`, etc. See
     *    https://developer.mozilla.org/en-US/docs/Web/API/Request/Request
     *    for more information. `requestInit` must not have a body, because the
     * body will be set by TensorFlow.js. File blobs representing the model
     * topology (filename: 'model.json') and the weights of the model (filename:
     * 'model.weights.bin') will be appended to the body. If `requestInit` has a
     * `body`, an Error will be thrown.
     * @param loadOptions Optional configuration for the loading. It includes the
     *   following fields:
     *   - weightPathPrefix Optional, this specifies the path prefix for weight
     *     files, by default this is calculated from the path param.
     *   - fetchFunc Optional, custom `fetch` function. E.g., in Node.js,
     *     the `fetch` from node-fetch can be used here.
     *   - onProgress Optional, progress callback function, fired periodically
     *     before the load is completed.
     * @returns An instance of `IOHandler`.
     *
     * @doc {
     *   heading: 'Models',
     *   subheading: 'Loading',
     *   namespace: 'io',
     *   ignoreCI: true
     * }
     */
    function http$2(path, loadOptions) {
        return new HTTPRequest(path, loadOptions);
    }
    /**
     * Deprecated. Use `tf.io.http`.
     * @param path
     * @param loadOptions
     */
    function browserHTTPRequest(path, loadOptions) {
        return http$2(path, loadOptions);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class PassthroughLoader {
        constructor(modelArtifacts) {
            this.modelArtifacts = modelArtifacts;
        }
        async load() {
            return this.modelArtifacts;
        }
    }
    class PassthroughSaver {
        constructor(saveHandler) {
            this.saveHandler = saveHandler;
        }
        async save(modelArtifacts) {
            return this.saveHandler(modelArtifacts);
        }
    }
    /**
     * Creates an IOHandler that loads model artifacts from memory.
     *
     * When used in conjunction with `tf.loadLayersModel`, an instance of
     * `tf.LayersModel` (Keras-style) can be constructed from the loaded artifacts.
     *
     * ```js
     * const model = await tf.loadLayersModel(tf.io.fromMemory(
     *     modelTopology, weightSpecs, weightData));
     * ```
     *
     * @param modelArtifacts a object containing model topology (i.e., parsed from
     *   the JSON format).
     * @param weightSpecs An array of `WeightsManifestEntry` objects describing the
     *   names, shapes, types, and quantization of the weight data.
     * @param weightData A single `ArrayBuffer` containing the weight data,
     *   concatenated in the order described by the weightSpecs.
     * @param trainingConfig Model training configuration. Optional.
     *
     * @returns A passthrough `IOHandler` that simply loads the provided data.
     */
    function fromMemory(modelArtifacts, weightSpecs, weightData, trainingConfig) {
        if (arguments.length === 1) {
            const isModelArtifacts = modelArtifacts.modelTopology != null ||
                modelArtifacts.weightSpecs != null;
            if (isModelArtifacts) {
                return new PassthroughLoader(modelArtifacts);
            }
            else {
                // Legacy support: with only modelTopology.
                // TODO(cais): Remove this deprecated API.
                console.warn('Please call tf.io.fromMemory() with only one argument. ' +
                    'The argument should be of type ModelArtifacts. ' +
                    'The multi-argument signature of tf.io.fromMemory() has been ' +
                    'deprecated and will be removed in a future release.');
                return new PassthroughLoader({ modelTopology: modelArtifacts });
            }
        }
        else {
            // Legacy support.
            // TODO(cais): Remove this deprecated API.
            console.warn('Please call tf.io.fromMemory() with only one argument. ' +
                'The argument should be of type ModelArtifacts. ' +
                'The multi-argument signature of tf.io.fromMemory() has been ' +
                'deprecated and will be removed in a future release.');
            return new PassthroughLoader({
                modelTopology: modelArtifacts,
                weightSpecs,
                weightData,
                trainingConfig
            });
        }
    }
    /**
     * Creates an IOHandler that passes saved model artifacts to a callback.
     *
     * ```js
     * function handleSave(artifacts) {
     *   // ... do something with the artifacts ...
     *   return {modelArtifactsInfo: {...}, ...};
     * }
     *
     * const saveResult = model.save(tf.io.withSaveHandler(handleSave));
     * ```
     *
     * @param saveHandler A function that accepts a `ModelArtifacts` and returns a
     *     `SaveResult`.
     */
    function withSaveHandler(saveHandler) {
        return new PassthroughSaver(saveHandler);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */

    var io = /*#__PURE__*/Object.freeze({
        __proto__: null,
        browserFiles: browserFiles,
        browserHTTPRequest: browserHTTPRequest,
        concatenateArrayBuffers: concatenateArrayBuffers,
        decodeWeights: decodeWeights,
        encodeWeights: encodeWeights,
        fromMemory: fromMemory,
        getLoadHandlers: getLoadHandlers,
        getModelArtifactsForJSON: getModelArtifactsForJSON,
        getModelArtifactsInfoForJSON: getModelArtifactsInfoForJSON,
        getSaveHandlers: getSaveHandlers,
        http: http$2,
        isHTTPScheme: isHTTPScheme,
        loadWeights: loadWeights,
        registerLoadRouter: registerLoadRouter,
        registerSaveRouter: registerSaveRouter,
        weightsLoaderFactory: weightsLoaderFactory,
        withSaveHandler: withSaveHandler,
        copyModel: copyModel,
        listModels: listModels,
        moveModel: moveModel,
        removeModel: removeModel
    });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the dot product of two matrices, A * B. These must be matrices.
     *
     * ```js
     * const a = tf.tensor2d([1, 2], [1, 2]);
     * const b = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * a.matMul(b).print();  // or tf.matMul(a, b)
     * ```
     * @param a First matrix in dot product operation.
     * @param b Second matrix in dot product operation.
     * @param transposeA If true, `a` is transposed before multiplication.
     * @param transposeB If true, `b` is transposed before multiplication.
     *
     * @doc {heading: 'Operations', subheading: 'Matrices'}
     */
    function matMul_(a, b, transposeA = false, transposeB = false) {
        let $a = convertToTensor(a, 'a', 'matMul');
        let $b = convertToTensor(b, 'b', 'matMul');
        [$a, $b] = makeTypesMatch($a, $b);
        const inputs = { a: $a, b: $b };
        const attrs = { transposeA, transposeB };
        return ENGINE.runKernel(BatchMatMul, inputs, attrs);
    }
    const matMul$1 = op({ matMul_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a one-hot `tf.Tensor`. The locations represented by `indices` take
     * value `onValue` (defaults to 1), while all other locations take value
     * `offValue` (defaults to 0). If `indices` is rank `R`, the output has rank
     * `R+1` with the last axis of size `depth`.
     * `indices` used to encode prediction class must start from 0. For example,
     *  if you have 3 classes of data, class 1 should be encoded as 0, class 2
     *  should be 1, and class 3 should be 2.
     *
     * ```js
     * tf.oneHot(tf.tensor1d([0, 1], 'int32'), 3).print();
     * ```
     *
     * @param indices `tf.Tensor` of indices with dtype `int32`. Indices must
     * start from 0.
     * @param depth The depth of the one hot dimension.
     * @param onValue A number used to fill in the output when the index matches
     * the location.
     * @param offValue A number used to fill in the output when the index does
     *     not match the location.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function oneHot_(indices, depth, onValue = 1, offValue = 0) {
        if (depth < 2) {
            throw new Error(`Error in oneHot: depth must be >=2, but it is ${depth}`);
        }
        const $indices = convertToTensor(indices, 'indices', 'oneHot', 'int32');
        const inputs = { indices: $indices };
        const attrs = { depth, onValue, offValue };
        return ENGINE.runKernel(OneHot, inputs, attrs);
    }
    const oneHot$2 = op({ oneHot_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Transposes the `tf.Tensor`. Permutes the dimensions according to `perm`.
     *
     * The returned `tf.Tensor`'s dimension `i` will correspond to the input
     * dimension `perm[i]`. If `perm` is not given, it is set to `[n-1...0]`,
     * where `n` is the rank of the input `tf.Tensor`. Hence by default, this
     * operation performs a regular matrix transpose on 2-D input `tf.Tensor`s.
     *
     * ```js
     * const a = tf.tensor2d([1, 2, 3, 4, 5, 6], [2, 3]);
     *
     * a.transpose().print();  // or tf.transpose(a)
     * ```
     *
     * @param x The tensor to transpose.
     * @param perm The permutation of the dimensions of a.
     *
     * @doc {heading: 'Operations', subheading: 'Matrices'}
     */
    function transpose_(x, perm) {
        const $x = convertToTensor(x, 'x', 'transpose');
        if (perm == null) {
            perm = $x.shape.map((s, i) => i).reverse();
        }
        assert$4($x.rank === perm.length, () => `Error in transpose: rank of input ${$x.rank} ` +
            `must match length of perm ${perm}.`);
        perm.forEach(axis => {
            assert$4(axis >= 0 && axis < $x.rank, () => `All entries in 'perm' must be between 0 and ${$x.rank - 1}` +
                ` but got ${perm}`);
        });
        if ($x.rank <= 1) {
            return $x.clone();
        }
        const inputs = { x: $x };
        const attrs = { perm };
        return ENGINE.runKernel(Transpose, inputs, attrs);
    }
    const transpose$2 = op({ transpose_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the confusion matrix from true labels and predicted labels.
     *
     * ```js
     * const labels = tf.tensor1d([0, 1, 2, 1, 0], 'int32');
     * const predictions = tf.tensor1d([0, 2, 2, 1, 0], 'int32');
     * const numClasses = 3;
     * const out = tf.math.confusionMatrix(labels, predictions, numClasses);
     * out.print();
     * // Expected output matrix:
     * // [[2, 0, 0],
     * //  [0, 1, 1],
     * //  [0, 0, 1]]
     * ```
     *
     * @param labels The target labels, assumed to be 0-based integers
     *   for the classes. The shape is `[numExamples]`, where
     *   `numExamples` is the number of examples included.
     * @param predictions The predicted classes, assumed to be
     *   0-based integers for the classes. Must have the same shape as `labels`.
     * @param numClasses Number of all classes, as an integer.
     *   Its value must be larger than the largest element in `labels` and
     *   `predictions`.
     * @returns The confusion matrix as a int32-type 2D tensor. The value at
     *   row `r` and column `c` is the number of times examples of actual class
     *   `r` were predicted as class `c`.
     *
     * @doc {heading: 'Operations', subheading: 'Evaluation'}
     */
    function confusionMatrix_(labels, predictions, numClasses) {
        const $labels = convertToTensor(labels, 'labels', 'confusionMatrix');
        const $predictions = convertToTensor(predictions, 'predictions', 'confusionMatrix');
        assert$4(numClasses == null || numClasses > 0 && Number.isInteger(numClasses), () => `If provided, numClasses must be a positive integer, ` +
            `but got ${numClasses}`);
        assert$4($labels.rank === 1, () => `Expected the rank of labels to be 1, but got ${$labels.rank}`);
        assert$4($predictions.rank === 1, () => `Expected the rank of predictions to be 1, ` +
            `but got ${$predictions.rank}`);
        assert$4($labels.shape[0] === $predictions.shape[0], () => `Mismatch in the number of examples: ` +
            `${$labels.shape[0]} vs. ${$predictions.shape[0]}. ` +
            `Labels and predictions should have the same number of elements.`);
        assert$4(numClasses > 0 && Number.isInteger(numClasses), () => `numClasses is required to be a positive integer, but got ` +
            `${numClasses}`);
        // TODO(cais): In the future, if oneHot supports tensors inputs for
        //   `numClasses`, `confusionMatrix` can make `numClasses` optional.
        const oneHotLabels = oneHot$2(cast$3($labels, 'int32'), numClasses);
        const oneHotPredictions = oneHot$2(cast$3($predictions, 'int32'), numClasses);
        const oneHotLabelsT = transpose$2(oneHotLabels);
        const product = matMul$1(oneHotLabelsT, oneHotPredictions);
        return cast$3(product, 'int32');
    }
    const confusionMatrix = op({ confusionMatrix_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */

    var math = /*#__PURE__*/Object.freeze({
        __proto__: null,
        confusionMatrix: confusionMatrix
    });

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the dimensions in the input shape that are broadcasted to
     * produce the provided output shape.
     *
     * The returned dimensions are 0-indexed and sorted. An example:
     * inShape = [4, 1, 3]
     * outShape = [5, 4, 3, 3]
     * result = [1]. Dimension 1 (2nd dimension of input) gets broadcasted 1 => 3.
     */
    function getBroadcastDims$1(inShape, outShape) {
        const inRank = inShape.length;
        const dims = [];
        for (let i = 0; i < inRank; i++) {
            const dim = inRank - 1 - i;
            const a = inShape[dim] || 1;
            const b = outShape[outShape.length - 1 - i] || 1;
            if (b > 1 && a === 1) {
                dims.unshift(dim);
            }
        }
        return dims;
    }
    /**
     * Returns the axes in the output space that should be reduced to produce
     * the input space.
     */
    function getReductionAxes(inShape, outShape) {
        const result = [];
        for (let i = 0; i < outShape.length; i++) {
            const inDim = inShape[inShape.length - i - 1];
            const outAxis = outShape.length - i - 1;
            const outDim = outShape[outAxis];
            if (inDim == null || (inDim === 1 && outDim > 1)) {
                result.unshift(outAxis);
            }
        }
        return result;
    }
    function assertAndGetBroadcastShape(shapeA, shapeB) {
        const result = [];
        const l = Math.max(shapeA.length, shapeB.length);
        for (let i = 0; i < l; i++) {
            let a = shapeA[shapeA.length - i - 1];
            if (a == null) {
                a = 1;
            }
            let b = shapeB[shapeB.length - i - 1];
            if (b == null) {
                b = 1;
            }
            if (a === 1) {
                result.unshift(b);
            }
            else if (b === 1) {
                result.unshift(a);
            }
            else if (a !== b) {
                const errMsg = `Operands could not be broadcast together with shapes ` +
                    `${shapeA} and ${shapeB}.`;
                throw Error(errMsg);
            }
            else {
                result.unshift(a);
            }
        }
        return result;
    }

    var broadcast_util = /*#__PURE__*/Object.freeze({
        __proto__: null,
        getBroadcastDims: getBroadcastDims$1,
        getReductionAxes: getReductionAxes,
        assertAndGetBroadcastShape: assertAndGetBroadcastShape
    });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates rank-3 `tf.Tensor` with the provided values, shape and dtype.
     *
     * The same functionality can be achieved with `tf.tensor`, but in general
     * we recommend using `tf.tensor3d` as it makes the code more readable.
     *
     *  ```js
     * // Pass a nested array.
     * tf.tensor3d([[[1], [2]], [[3], [4]]]).print();
     * ```
     * ```js
     * // Pass a flat array and specify a shape.
     * tf.tensor3d([1, 2, 3, 4], [2, 2, 1]).print();
     * ```
     *
     * @param values The values of the tensor. Can be nested array of numbers,
     *     or a flat array, or a `TypedArray`.
     * @param shape The shape of the tensor. If not provided,  it is inferred from
     *     `values`.
     * @param dtype The data type.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function tensor3d(values, shape, dtype) {
        assertNonNull(values);
        if (shape != null && shape.length !== 3) {
            throw new Error('tensor3d() requires shape to have three numbers');
        }
        const inferredShape = inferShape(values, dtype);
        if (inferredShape.length !== 3 && inferredShape.length !== 1) {
            throw new Error('tensor3d() requires values to be number[][][] or flat/TypedArray');
        }
        if (inferredShape.length === 1 && shape == null) {
            throw new Error('tensor3d() requires shape to be provided when `values` ' +
                'are a flat array');
        }
        return makeTensor(values, shape, inferredShape, dtype);
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    let fromPixels2DContext$1;
    /**
     * Creates a `tf.Tensor` from an image.
     *
     * ```js
     * const image = new ImageData(1, 1);
     * image.data[0] = 100;
     * image.data[1] = 150;
     * image.data[2] = 200;
     * image.data[3] = 255;
     *
     * tf.browser.fromPixels(image).print();
     * ```
     *
     * @param pixels The input image to construct the tensor from. The
     * supported image types are all 4-channel. You can also pass in an image
     * object with following attributes:
     * `{data: Uint8Array; width: number; height: number}`
     * @param numChannels The number of channels of the output tensor. A
     * numChannels value less than 4 allows you to ignore channels. Defaults to
     * 3 (ignores alpha channel of input image).
     *
     * @returns A Tensor3D with the shape `[height, width, numChannels]`.
     *
     * Note: fromPixels can be lossy in some cases, same image may result in
     * slightly different tensor values, if rendered by different rendering
     * engines. This means that results from different browsers, or even same
     * browser with CPU and GPU rendering engines can be different. See discussion
     * in details:
     * https://github.com/tensorflow/tfjs/issues/5482
     *
     * @doc {heading: 'Browser', namespace: 'browser', ignoreCI: true}
     */
    function fromPixels_(pixels, numChannels = 3) {
        // Sanity checks.
        if (numChannels > 4) {
            throw new Error('Cannot construct Tensor with more than 4 channels from pixels.');
        }
        if (pixels == null) {
            throw new Error('pixels passed to tf.browser.fromPixels() can not be null');
        }
        let isPixelData = false;
        let isImageData = false;
        let isVideo = false;
        let isImage = false;
        let isCanvasLike = false;
        let isImageBitmap = false;
        if (pixels.data instanceof Uint8Array) {
            isPixelData = true;
        }
        else if (typeof (ImageData) !== 'undefined' && pixels instanceof ImageData) {
            isImageData = true;
        }
        else if (typeof (HTMLVideoElement) !== 'undefined' &&
            pixels instanceof HTMLVideoElement) {
            isVideo = true;
        }
        else if (typeof (HTMLImageElement) !== 'undefined' &&
            pixels instanceof HTMLImageElement) {
            isImage = true;
            // tslint:disable-next-line: no-any
        }
        else if (pixels.getContext != null) {
            isCanvasLike = true;
        }
        else if (typeof (ImageBitmap) !== 'undefined' && pixels instanceof ImageBitmap) {
            isImageBitmap = true;
        }
        else {
            throw new Error('pixels passed to tf.browser.fromPixels() must be either an ' +
                `HTMLVideoElement, HTMLImageElement, HTMLCanvasElement, ImageData ` +
                `in browser, or OffscreenCanvas, ImageData in webworker` +
                ` or {data: Uint32Array, width: number, height: number}, ` +
                `but was ${pixels.constructor.name}`);
        }
        if (isVideo) {
            const HAVE_CURRENT_DATA_READY_STATE = 2;
            if (isVideo &&
                pixels.readyState <
                    HAVE_CURRENT_DATA_READY_STATE) {
                throw new Error('The video element has not loaded data yet. Please wait for ' +
                    '`loadeddata` event on the <video> element.');
            }
        }
        // If the current backend has 'FromPixels' registered, it has a more
        // efficient way of handling pixel uploads, so we call that.
        const kernel = getKernel(FromPixels, ENGINE.backendName);
        if (kernel != null) {
            const inputs = { pixels };
            const attrs = { numChannels };
            return ENGINE.runKernel(FromPixels, inputs, attrs);
        }
        const [width, height] = isVideo ?
            [
                pixels.videoWidth,
                pixels.videoHeight
            ] :
            [pixels.width, pixels.height];
        let vals;
        if (isCanvasLike) {
            vals =
                // tslint:disable-next-line:no-any
                pixels.getContext('2d').getImageData(0, 0, width, height).data;
        }
        else if (isImageData || isPixelData) {
            vals = pixels.data;
        }
        else if (isImage || isVideo || isImageBitmap) {
            if (fromPixels2DContext$1 == null) {
                if (typeof document === 'undefined') {
                    if (typeof OffscreenCanvas !== 'undefined' &&
                        typeof OffscreenCanvasRenderingContext2D !== 'undefined') {
                        // @ts-ignore
                        fromPixels2DContext$1 = new OffscreenCanvas(1, 1).getContext('2d');
                    }
                    else {
                        throw new Error('Cannot parse input in current context. ' +
                            'Reason: OffscreenCanvas Context2D rendering is not supported.');
                    }
                }
                else {
                    fromPixels2DContext$1 = document.createElement('canvas').getContext('2d');
                }
            }
            fromPixels2DContext$1.canvas.width = width;
            fromPixels2DContext$1.canvas.height = height;
            fromPixels2DContext$1.drawImage(pixels, 0, 0, width, height);
            vals = fromPixels2DContext$1.getImageData(0, 0, width, height).data;
        }
        let values;
        if (numChannels === 4) {
            values = new Int32Array(vals);
        }
        else {
            const numPixels = width * height;
            values = new Int32Array(numPixels * numChannels);
            for (let i = 0; i < numPixels; i++) {
                for (let channel = 0; channel < numChannels; ++channel) {
                    values[i * numChannels + channel] = vals[i * 4 + channel];
                }
            }
        }
        const outShape = [height, width, numChannels];
        return tensor3d(values, outShape, 'int32');
    }
    // Helper functions for |fromPixelsAsync| to check whether the input can
    // be wrapped into imageBitmap.
    function isPixelData(pixels) {
        return (pixels != null) && (pixels.data instanceof Uint8Array);
    }
    function isImageBitmapFullySupported() {
        return typeof window !== 'undefined' &&
            typeof (ImageBitmap) !== 'undefined' &&
            window.hasOwnProperty('createImageBitmap');
    }
    function isNonEmptyPixels(pixels) {
        return pixels != null && pixels.width !== 0 && pixels.height !== 0;
    }
    function canWrapPixelsToImageBitmap(pixels) {
        return isImageBitmapFullySupported() && !(pixels instanceof ImageBitmap) &&
            isNonEmptyPixels(pixels) && !isPixelData(pixels);
    }
    /**
     * Creates a `tf.Tensor` from an image in async way.
     *
     * ```js
     * const image = new ImageData(1, 1);
     * image.data[0] = 100;
     * image.data[1] = 150;
     * image.data[2] = 200;
     * image.data[3] = 255;
     *
     * (await tf.browser.fromPixelsAsync(image)).print();
     * ```
     * This API is the async version of fromPixels. The API will first
     * check |WRAP_TO_IMAGEBITMAP| flag, and try to wrap the input to
     * imageBitmap if the flag is set to true.
     *
     * @param pixels The input image to construct the tensor from. The
     * supported image types are all 4-channel. You can also pass in an image
     * object with following attributes:
     * `{data: Uint8Array; width: number; height: number}`
     * @param numChannels The number of channels of the output tensor. A
     * numChannels value less than 4 allows you to ignore channels. Defaults to
     * 3 (ignores alpha channel of input image).
     *
     * @doc {heading: 'Browser', namespace: 'browser', ignoreCI: true}
     */
    async function fromPixelsAsync(pixels, numChannels = 3) {
        let inputs = null;
        // Check whether the backend needs to wrap |pixels| to imageBitmap and
        // whether |pixels| can be wrapped to imageBitmap.
        if (env$1().getBool('WRAP_TO_IMAGEBITMAP') &&
            canWrapPixelsToImageBitmap(pixels)) {
            // Force the imageBitmap creation to not do any premultiply alpha
            // ops.
            let imageBitmap;
            try {
                // wrap in try-catch block, because createImageBitmap may not work
                // properly in some browsers, e.g.
                // https://bugzilla.mozilla.org/show_bug.cgi?id=1335594
                // tslint:disable-next-line: no-any
                imageBitmap = await createImageBitmap(pixels, { premultiplyAlpha: 'none' });
            }
            catch (e) {
                imageBitmap = null;
            }
            // createImageBitmap will clip the source size.
            // In some cases, the input will have larger size than its content.
            // E.g. new Image(10, 10) but with 1 x 1 content. Using
            // createImageBitmap will clip the size from 10 x 10 to 1 x 1, which
            // is not correct. We should avoid wrapping such resouce to
            // imageBitmap.
            if (imageBitmap != null && imageBitmap.width === pixels.width &&
                imageBitmap.height === pixels.height) {
                inputs = imageBitmap;
            }
            else {
                inputs = pixels;
            }
        }
        else {
            inputs = pixels;
        }
        return fromPixels_(inputs, numChannels);
    }
    /**
     * Draws a `tf.Tensor` of pixel values to a byte array or optionally a
     * canvas.
     *
     * When the dtype of the input is 'float32', we assume values in the range
     * [0-1]. Otherwise, when input is 'int32', we assume values in the range
     * [0-255].
     *
     * Returns a promise that resolves when the canvas has been drawn to.
     *
     * @param img A rank-2 tensor with shape `[height, width]`, or a rank-3 tensor
     * of shape `[height, width, numChannels]`. If rank-2, draws grayscale. If
     * rank-3, must have depth of 1, 3 or 4. When depth of 1, draws
     * grayscale. When depth of 3, we draw with the first three components of
     * the depth dimension corresponding to r, g, b and alpha = 1. When depth of
     * 4, all four components of the depth dimension correspond to r, g, b, a.
     * @param canvas The canvas to draw to.
     *
     * @doc {heading: 'Browser', namespace: 'browser'}
     */
    async function toPixels(img, canvas) {
        let $img = convertToTensor(img, 'img', 'toPixels');
        if (!(img instanceof Tensor)) {
            // Assume int32 if user passed a native array.
            const originalImgTensor = $img;
            $img = cast$3(originalImgTensor, 'int32');
            originalImgTensor.dispose();
        }
        if ($img.rank !== 2 && $img.rank !== 3) {
            throw new Error(`toPixels only supports rank 2 or 3 tensors, got rank ${$img.rank}.`);
        }
        const [height, width] = $img.shape.slice(0, 2);
        const depth = $img.rank === 2 ? 1 : $img.shape[2];
        if (depth > 4 || depth === 2) {
            throw new Error(`toPixels only supports depth of size ` +
                `1, 3 or 4 but got ${depth}`);
        }
        if ($img.dtype !== 'float32' && $img.dtype !== 'int32') {
            throw new Error(`Unsupported type for toPixels: ${$img.dtype}.` +
                ` Please use float32 or int32 tensors.`);
        }
        const data = await $img.data();
        const multiplier = $img.dtype === 'float32' ? 255 : 1;
        const bytes = new Uint8ClampedArray(width * height * 4);
        for (let i = 0; i < height * width; ++i) {
            const rgba = [0, 0, 0, 255];
            for (let d = 0; d < depth; d++) {
                const value = data[i * depth + d];
                if ($img.dtype === 'float32') {
                    if (value < 0 || value > 1) {
                        throw new Error(`Tensor values for a float32 Tensor must be in the ` +
                            `range [0 - 1] but encountered ${value}.`);
                    }
                }
                else if ($img.dtype === 'int32') {
                    if (value < 0 || value > 255) {
                        throw new Error(`Tensor values for a int32 Tensor must be in the ` +
                            `range [0 - 255] but encountered ${value}.`);
                    }
                }
                if (depth === 1) {
                    rgba[0] = value * multiplier;
                    rgba[1] = value * multiplier;
                    rgba[2] = value * multiplier;
                }
                else {
                    rgba[d] = value * multiplier;
                }
            }
            const j = i * 4;
            bytes[j + 0] = Math.round(rgba[0]);
            bytes[j + 1] = Math.round(rgba[1]);
            bytes[j + 2] = Math.round(rgba[2]);
            bytes[j + 3] = Math.round(rgba[3]);
        }
        if (canvas != null) {
            canvas.width = width;
            canvas.height = height;
            const ctx = canvas.getContext('2d');
            const imageData = new ImageData(bytes, width, height);
            ctx.putImageData(imageData, 0, 0);
        }
        if ($img !== img) {
            $img.dispose();
        }
        return bytes;
    }
    const fromPixels$1 = op({ fromPixels_ });

    var browser$2 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        fromPixelsAsync: fromPixelsAsync,
        toPixels: toPixels,
        fromPixels: fromPixels$1
    });

    /**
     * Validate gather nd inputs.
     *
     * @param tensor The tensor contains the source values.
     * @param indices The tensor contains the indices to slice the source.
     *
     * @returns [resultShape, numUpdates, sliceSize, strides]
     */
    function prepareAndValidate(tensor, indices) {
        const tensorRank = tensor.shape.length;
        const indicesRank = indices.shape.length;
        if (tensorRank < 1) {
            throw new Error('tf.gatherND() expects the input to be rank 1 or higher,' +
                ` but the rank was ${tensorRank}.`);
        }
        if (indicesRank < 1) {
            throw new Error('tf.gatherND() expects the indices to be rank 1 or higher,' +
                ` but the rank was ${indicesRank}.`);
        }
        if (indices.dtype !== 'int32') {
            throw new Error('tf.gatherND() expects the indices to be int32 type,' +
                ` but the dtype was ${indices.dtype}.`);
        }
        if (indices.shape[indicesRank - 1] > tensorRank) {
            throw new Error('index innermost dimension length must be <= tensor rank; saw: ' +
                `${indices.shape[indicesRank - 1]} vs. ${tensorRank}`);
        }
        if (sizeFromShape(tensor.shape) === 0) {
            throw new Error('Requested more than 0 entries, but input is empty.' +
                ` Input shape: ${tensor.shape}.`);
        }
        const indicesShape = indices.shape;
        const sliceRank = indicesShape[indicesShape.length - 1];
        // The result shape is
        //   indices.shape[:-1] + params.shape[indices.shape[-1]:]
        let nResult = 1;
        for (let i = 0; i < indicesShape.length - 1; ++i) {
            nResult *= indicesShape[i];
        }
        const inputShape = tensor.shape;
        const resultShape = indicesShape.slice();
        resultShape.pop();
        let sliceSize = 1;
        for (let i = sliceRank; i < tensorRank; ++i) {
            sliceSize *= inputShape[i];
            resultShape.push(inputShape[i]);
        }
        const strides = [...computeStrides(tensor.shape).map(stride => stride / sliceSize),
            1].slice(0, sliceRank);
        return [resultShape, nResult, sliceSize, strides];
    }

    var gather_nd_util = /*#__PURE__*/Object.freeze({
        __proto__: null,
        prepareAndValidate: prepareAndValidate
    });

    /**
     * Check whether updates.shape = indices.shape[:batchDim] +
     * shape[sliceDim:]
     *
     * @param x The input tensor.
     */
    function validateUpdateShape(shape, indices, updates) {
        const sliceDim = (indices.rank > 1) ? indices.shape[indices.rank - 1] : 1;
        const batchDim = (indices.rank > 1) ? indices.rank - 1 : 1;
        const shapeError = 'Must have updates.shape = indices.shape[:batchDim] + ' +
            `shape[sliceDim:], got updates.shape: ${updates.shape}` +
            `, indices.shape: ${indices.shape}, shape: ${shape}` +
            `, sliceDim: ${sliceDim}, and batchDim: ${batchDim}.`;
        if (updates.rank < batchDim) {
            throw new Error(shapeError + ` update.rank < ${batchDim}. `);
        }
        if (shape.length < sliceDim + (updates.rank - batchDim)) {
            throw new Error(shapeError +
                ` Output shape length < ${sliceDim + (updates.rank - batchDim)}`);
        }
        if (updates.rank !== batchDim + shape.length - sliceDim) {
            throw new Error(shapeError + ` update.rank != ${batchDim + shape.length - sliceDim}`);
        }
        for (let d = 0; d < batchDim; ++d) {
            if (updates.shape[d] !== indices.shape[d]) {
                throw new Error(shapeError +
                    ` updates.shape[${d}] (${updates.shape[d]}) != indices.shape[${d}] (${indices.shape[d]}).`);
            }
        }
        for (let d = 0; d < updates.rank - batchDim; ++d) {
            if (updates.shape[d + batchDim] !== shape[d + sliceDim]) {
                throw new Error(shapeError +
                    ` updates.shape[${d + batchDim}] (${updates.shape[d + batchDim]}) != shape[${d + batchDim}] (${shape[d + batchDim]})`);
            }
        }
    }
    /**
     * Validate scatter nd inputs.
     *
     * @param update The tensor contains the update values.
     * @param indices The tensor contains the indices for the update values.
     * @param shape The shape of the output tensor.
     */
    function validateInput$1(updates, indices, shape) {
        if (indices.rank < 1) {
            throw new Error('tf.scatterND() expects the indices to be rank 1 or higher,' +
                ` but the rank was ${indices.rank}.`);
        }
        if (updates.rank < 1) {
            throw new Error('tf.scatterND() expects the updates to be rank 1 or higher,' +
                ` but the rank was ${updates.rank}.`);
        }
        if (indices.dtype !== 'int32') {
            throw new Error(`The dtype of 'indices' should be int32, but got dtype: ${indices.dtype}`);
        }
        if (shape.length < 1) {
            throw new Error(`Output rank must be greater or equal to 1, but got shape: ${shape}`);
        }
        if (shape.length === 0) {
            if (indices.size === 0) {
                throw new Error(`Indices specified for empty output. indices shape: ${indices.shape}`);
            }
            if (updates.size === 0) {
                throw new Error(`Updates specified for empty output. updates shape: ${updates.shape}`);
            }
        }
        validateUpdateShape(shape, indices, updates);
    }
    /**
     * Calculate the shape information for the output.
     *
     * @param update The tensor contains the update values.
     * @param indices The tensor contains the indices for the update values.
     * @param shape The shape of the output tensor.
     *
     * @returns ScatterShapeInfo
     */
    function calculateShapes(updates, indices, shape) {
        // Calculate the number of dimensions in indices
        const indicesRank = indices.shape.length;
        const sliceRank = (indicesRank > 1) ? indices.shape[indicesRank - 1] : 1;
        // Calculate the number of elements that make up each slice of our updated
        // tensor. This allows us to work with flattened tensors and copy over whole
        // slices at a time.
        const totalNd = shape.length;
        let sliceSize = 1;
        for (let i = sliceRank; i < totalNd; ++i) {
            sliceSize *= shape[i];
        }
        const safeSliceDim = (sliceRank < 1) ? 1 : sliceRank;
        const numUpdates = sizeFromShape(indices.shape) / safeSliceDim;
        const strides = [...computeStrides(shape.slice(0, sliceRank)), 1];
        const outputSize = sizeFromShape(shape);
        return { sliceRank, numUpdates, sliceSize, strides, outputSize };
    }

    var scatter_nd_util = /*#__PURE__*/Object.freeze({
        __proto__: null,
        validateUpdateShape: validateUpdateShape,
        validateInput: validateInput$1,
        calculateShapes: calculateShapes
    });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const NEW_AXIS = -2;
    const SHRINK_AXIS = -1;
    function assertParamsValid(input, begin, size) {
        const inputRank = input.shape.length;
        assert$4(inputRank === begin.length, () => `Error in slice${inputRank}D: Length of begin ${begin} must ` +
            `match the rank of the array (${inputRank}).`);
        assert$4(inputRank === size.length, () => `Error in slice${inputRank}D: Length of size ${size} must ` +
            `match the rank of the array (${inputRank}).`);
        for (let i = 0; i < inputRank; ++i) {
            assert$4(begin[i] + size[i] <= input.shape[i], () => `Error in slice${inputRank}D: begin[${i}] + size[${i}] ` +
                `(${begin[i] + size[i]}) would overflow input.shape[${i}] (${input.shape[i]})`);
        }
    }
    /** Converts a binary mask to an array of axes. Used in stridedSlice(). */
    function maskToAxes(mask) {
        const axes = [];
        let axis = 0;
        while (mask > 0) {
            if (mask & 1) {
                axes.push(axis);
            }
            mask /= 2;
            axis++;
        }
        return axes;
    }
    /** Computes the output shape given the strided slice params. */
    function computeOutShape$2(begin, end, strides) {
        const size = [];
        for (let axis = 0; axis < begin.length; axis++) {
            size[axis] = Math.ceil((end[axis] - begin[axis]) / strides[axis]);
        }
        return size;
    }
    // Creates full selection at the elided dimensions. If the dimension matches
    // the ellipsis mask, override the current stride value. Otherwise, insert.
    function stridesWithElidedDims(strides, ellipsisInsertionIndex, numElidedAxes, inputShape) {
        const newStrides = [...strides];
        for (let i = newStrides.length; i < inputShape.length; i++) {
            newStrides.push(1);
        }
        for (let i = 0; i < numElidedAxes; i++) {
            if (i === 0) {
                newStrides[ellipsisInsertionIndex] = 1;
            }
            else {
                newStrides.splice(ellipsisInsertionIndex, 0 /* num elements to delete */, 1 /* element to add */);
                newStrides.pop();
            }
        }
        return newStrides;
    }
    function unnormalizeAxis(ellipsisInsertionIndex, numElidedAxes, normalizedAxis) {
        if (normalizedAxis <= ellipsisInsertionIndex) {
            return normalizedAxis;
        }
        return normalizedAxis - (numElidedAxes - 1);
    }
    function getElidedAxes(numElidedAxes, ellipsisInsertionIndex) {
        const elidedAxes = [];
        for (let i = 0; i < numElidedAxes; i++) {
            elidedAxes.push(ellipsisInsertionIndex + i);
        }
        return elidedAxes;
    }
    // Normalize the start, end and strides.
    function getNormalizedAxes(inputShape, ellipsisAxes, numInterpolatedAxes, begin, end, strides, beginMask, endMask, ellipsisMask) {
        const inputRank = inputShape.length;
        let normalizedBegin = new Array(inputRank), normalizedEnd = new Array(inputRank), normalizedStrides = new Array(inputRank);
        if (ellipsisAxes.length && numInterpolatedAxes > 0) {
            const fullIndex = ellipsisAxes[0];
            // The ellipsis applies to the masked index as well as any dimensions
            // that are interpolated.
            const numElidedAxes = numInterpolatedAxes + 1;
            normalizedBegin = startIndicesWithElidedDims(beginMask, fullIndex, numElidedAxes, begin, inputShape);
            normalizedEnd = stopIndicesWithElidedDims(endMask, fullIndex, numElidedAxes, end, inputShape);
            normalizedStrides =
                stridesWithElidedDims(strides, fullIndex, numElidedAxes, inputShape);
        }
        else {
            for (let axis = 0; axis < inputRank; axis++) {
                normalizedBegin[axis] = startForAxis(beginMask, begin, strides, inputShape, axis, ellipsisMask);
                normalizedEnd[axis] =
                    stopForAxis(endMask, end, strides, inputShape, axis, ellipsisMask);
                normalizedStrides[axis] = stridesForAxis(strides, axis, ellipsisMask);
            }
        }
        return {
            begin: normalizedBegin,
            end: normalizedEnd,
            strides: normalizedStrides
        };
    }
    // Creates full selection at the elided dimensions. If the dimension matches
    // the ellipsis mask, override the current start value. Otherwise, insert.
    function startIndicesWithElidedDims(beginMask, ellipsisInsertionIndex, numElidedAxes, originalBegin, inputShape) {
        const newIndices = [...inputShape];
        const elidedAxes = getElidedAxes(numElidedAxes, ellipsisInsertionIndex);
        for (let axis = 0; axis < newIndices.length; axis++) {
            if (elidedAxes.indexOf(axis) > -1) {
                newIndices[axis] = 0;
            }
            else {
                const originalAxis = unnormalizeAxis(ellipsisInsertionIndex, numElidedAxes, axis);
                let originalValue = originalBegin[originalAxis];
                if (beginMask & 1 << originalAxis) {
                    originalValue = 0;
                }
                newIndices[axis] = originalValue;
            }
        }
        return newIndices;
    }
    // Creates full selection at the elided dimensions. If the dimension matches
    // the ellipsis mask, override the current stop value. Otherwise, insert.
    function stopIndicesWithElidedDims(endMask, ellipsisInsertionIndex, numElidedAxes, originalEnd, inputShape) {
        const newIndices = [...inputShape];
        const elidedAxes = getElidedAxes(numElidedAxes, ellipsisInsertionIndex);
        for (let axis = 0; axis < newIndices.length; axis++) {
            if (elidedAxes.indexOf(axis) > -1) {
                newIndices[axis] = Number.MAX_SAFE_INTEGER;
            }
            else {
                const originalAxis = unnormalizeAxis(ellipsisInsertionIndex, numElidedAxes, axis);
                let originalValue = originalEnd[originalAxis];
                if (endMask & 1 << originalAxis) {
                    originalValue = Number.MAX_SAFE_INTEGER;
                }
                newIndices[axis] = originalValue;
            }
        }
        for (let i = 0; i < newIndices.length; i++) {
            // Handle negative indices
            const axisSize = inputShape[i];
            if (newIndices[i] < 0) {
                newIndices[i] += axisSize;
            }
            newIndices[i] = clamp(0, newIndices[i], inputShape[i]);
        }
        return newIndices;
    }
    function stridesForAxis(strides, axis, ellipsisMask) {
        let stride = strides[axis];
        if (ellipsisMask & (1 << axis) || stride == null) {
            stride = 1;
        }
        return stride;
    }
    function startForAxis(beginMask, startIndices, strides, inputShape, axis, ellipsisMask) {
        // Begin with the specified index
        let start = startIndices[axis];
        const stride = strides[axis] || 1;
        // Check the axis bit from right of masked axes, or the begin index is not set
        // for the axis.
        if (beginMask & 1 << axis || ellipsisMask & 1 << axis || start == null) {
            if (stride > 0) {
                // Forward iteration - use the first element. These values will get
                // clamped below (Note: We could have set them to 0 and axis_size-1, but
                // use lowest() and max() to maintain symmetry with StopForAxis())
                start = Number.MIN_SAFE_INTEGER;
            }
            else {
                // Backward iteration - use the last element.
                start = Number.MAX_SAFE_INTEGER;
            }
        }
        // Handle negative indices
        const axisSize = inputShape[axis];
        if (start < 0) {
            start += axisSize;
        }
        // Clamping
        start = clamp(0, start, axisSize - 1);
        return start;
    }
    function stopForAxis(endMask, stopIndices, strides, inputShape, axis, ellipsisMask) {
        // Begin with the specified index
        let stop = stopIndices[axis];
        const stride = strides[axis] || 1;
        // Check the axis bit from right of masked axes, or if the stop index is not
        // set for this axis.
        if (endMask & (1 << axis) || ellipsisMask & (1 << axis) || stop == null) {
            if (stride > 0) {
                // Forward iteration - use the last element. These values will get
                // clamped below
                stop = Number.MAX_SAFE_INTEGER;
            }
            else {
                // Backward iteration - use the first element.
                stop = Number.MIN_SAFE_INTEGER;
            }
        }
        // Handle negative indices
        const axisSize = inputShape[axis];
        if (stop < 0) {
            stop += axisSize;
        }
        // Clamping
        // Because the end index points one past the last element, we need slightly
        // different clamping ranges depending on the direction.
        if (stride > 0) {
            // Forward iteration
            stop = clamp(0, stop, axisSize);
        }
        else {
            // Backward iteration
            stop = clamp(-1, stop, axisSize - 1);
        }
        return stop;
    }
    /**
     * Returns true if the slice occupies a continous set of elements in the
     * 'flat' space.
     */
    function isSliceContinous(shape, begin, size) {
        // Index of the first axis that has size > 1.
        let firstNonOneAxis = size.length;
        for (let i = 0; i < size.length; i++) {
            if (size[i] > 1) {
                firstNonOneAxis = i;
                break;
            }
        }
        for (let i = firstNonOneAxis + 1; i < size.length; i++) {
            if (begin[i] > 0 || size[i] !== shape[i]) {
                return false;
            }
        }
        return true;
    }
    function computeFlatOffset(begin, strides) {
        let flatOffset = begin.length > 0 ? begin[begin.length - 1] : 1;
        for (let i = 0; i < begin.length - 1; i++) {
            flatOffset += begin[i] * strides[i];
        }
        return flatOffset;
    }
    function parseSliceParams(x, begin, size) {
        // The following logic allows for more ergonomic calls.
        let begin_;
        const xRank = x.shape.length;
        if (typeof begin === 'number') {
            begin_ = [begin, ...new Array(xRank - 1).fill(0)];
        }
        else if (begin.length < xRank) {
            begin_ = begin.concat(new Array(xRank - begin.length).fill(0));
        }
        else {
            begin_ = begin.slice();
        }
        begin_.forEach(d => {
            assert$4(d !== -1, () => 'slice() does not support negative begin indexing.');
        });
        let size_;
        if (size == null) {
            size_ = new Array(xRank).fill(-1);
        }
        else if (typeof size === 'number') {
            size_ = [size, ...new Array(xRank - 1).fill(-1)];
        }
        else if (size.length < xRank) {
            size_ = size.concat(new Array(xRank - size.length).fill(-1));
        }
        else {
            size_ = size;
        }
        size_ = size_.map((d, i) => {
            if (d >= 0) {
                return d;
            }
            else {
                assert$4(d === -1, () => `Negative size values should be exactly -1 but got ` +
                    `${d} for the slice() size at index ${i}.`);
                return x.shape[i] - begin_[i];
            }
        });
        return [begin_, size_];
    }
    // Convert the slicing specification from a sparse representation to a dense
    // representation. This means that all ellipses and newaxis are expanded out.
    function sliceInfo(xShape, begin, end, strides, beginMask, endMask, ellipsisMask, newAxisMask, shrinkAxisMask) {
        let stridesNonNull;
        if (strides == null) {
            stridesNonNull = new Array(begin.length);
            stridesNonNull.fill(1);
        }
        else {
            stridesNonNull = strides;
        }
        // Only one non-zero bit is allowed in ellipsisMask, which means ellipsisMask
        // is a power of 2. Use bit compares to ensure ellipsisMask is 0 or a power
        // of 2. When i is a power of 2, i & (i - 1) is always 0.
        // Also ref:
        // https://stackoverflow.com/questions/600293/how-to-check-if-a-number-is-a-power-of-2
        if (ellipsisMask != null && (ellipsisMask & (ellipsisMask - 1)) !== 0) {
            throw new Error('Multiple ellipses in slice is not allowed.');
        }
        // Step 1: Account for ellipsis and new axis.
        // Check for ellipsis and count how many non-newaxis there are after.
        let ellipsisSeen = false;
        const sparseSpec = {
            dims: stridesNonNull.length,
            numAddAxisAfterEllipsis: 0,
            begin: begin.slice(),
            end: end.slice(),
            strides: stridesNonNull.slice(),
            beginMask,
            endMask,
            ellipsisMask,
            newAxisMask,
            shrinkAxisMask
        };
        for (let i = 0; i < sparseSpec.dims; i++) {
            if (ellipsisSeen && ((1 << i) & newAxisMask) !== 0) {
                sparseSpec.numAddAxisAfterEllipsis++;
            }
            if ((1 << i) & ellipsisMask) {
                ellipsisSeen = true;
            }
        }
        // If no ellipsis insert one at the end.
        if (!ellipsisSeen) {
            sparseSpec.ellipsisMask |= (1 << sparseSpec.dims);
            sparseSpec.dims++; // this effects loop iteration below
        }
        // Step 2: Make a sparse spec into a full index spec.
        //
        // The sparse spec deos not correspond to the number of dimensions.
        // Make a dense spec that cooresponds to the number of dimensions.
        //
        // For example suppose foo[...,3:] on foo.shape = [2, 2, 3] then we need to
        // produce the missing beginMask for the first two dimensions i.e. from
        // beginMaskSpec = 0, endMaskSpec = 2, we achieve beginMask = 6 (110),
        // endMask = 7 (111).
        const denseSpec = {
            dims: xShape.length,
            beginMask: 0,
            endMask: 0,
            beginValid: false,
            endValid: false
        };
        buildDenseSpec(sparseSpec, denseSpec);
        // Step 3: Make implicit ranges (non-zero beginMasks and endMasks) explicit
        // and bounds check.
        let isIdentity = true;
        let sliceDim0 = true;
        let isSimpleSlice = true;
        const processingShape = [];
        const finalShape = [];
        for (let i = 0; i < xShape.length; ++i) {
            if (denseSpec.strides[i] === 0) {
                throw Error(`strides[${i}] must be non-zero`);
            }
            const shrinkI = !!(denseSpec.shrinkAxisMask & (1 << i));
            const dimI = xShape[i];
            if (dimI === -1) {
                processingShape.push(shrinkI ? 1 : -1);
                continue;
            }
            const masks = [denseSpec.beginMask & (1 << i), denseSpec.endMask & (1 << i)];
            const validRange = [
                denseSpec.strides[i] > 0 ? 0 : -1,
                denseSpec.strides[i] > 0 ? dimI : dimI - 1
            ];
            if (shrinkI && denseSpec.strides[i] <= 0) {
                throw Error('only stride 1 allowed on non-range indexing.');
            }
            isSimpleSlice = isSimpleSlice && (denseSpec.strides[i] === 1);
            const beginAndEndMasked = !!((denseSpec.beginMask & (1 << i)) && (denseSpec.endMask & (1 << i)));
            if (denseSpec.beginValid && denseSpec.endValid) {
                if (shrinkI) {
                    // If we are shrinking, the end index is now possibly incorrect. In
                    // particular foo[-1] produces sparseBegin = -1, sparseEnd = 0.
                    // and canonical puts these to n-1 and 0, which implies a degenerate
                    // interval. Fortunately, it is now safe to re-create end as begin + 1.
                    const xFwd = denseSpec.begin[i] < 0 ? dimI + denseSpec.begin[i] :
                        denseSpec.begin[i];
                    denseSpec.begin[i] = xFwd;
                    denseSpec.end[i] = denseSpec.begin[i] + 1;
                    if (xFwd < 0 || xFwd >= dimI) {
                        throw Error(`slice index ${denseSpec.begin[i]} of dimension ${i} out of bounds.`);
                    }
                }
                else {
                    denseSpec.begin[i] = canonical(denseSpec.begin[i], 0, denseSpec.strides[i], dimI, masks, validRange);
                    denseSpec.end[i] = canonical(denseSpec.end[i], 1, denseSpec.strides[i], dimI, masks, validRange);
                }
                // Update optimization values
                const takeAllInDimension = denseSpec.strides[i] === 1 &&
                    denseSpec.begin[i] === 0 && denseSpec.end[i] === dimI;
                isIdentity = isIdentity && takeAllInDimension;
                sliceDim0 = sliceDim0 &&
                    ((i === 0 && denseSpec.strides[i] === 1) || takeAllInDimension);
            }
            else {
                isIdentity =
                    isIdentity && ((denseSpec.strides[i] === 1) && beginAndEndMasked);
                sliceDim0 = sliceDim0 &&
                    ((i === 0 && denseSpec.strides[i] === 1) || beginAndEndMasked);
            }
            // Compute the processing shape (the intermediate Eigen will produce)
            let intervalLength;
            let knownInterval = false;
            if (denseSpec.beginValid && denseSpec.endValid) {
                intervalLength = denseSpec.end[i] - denseSpec.begin[i];
                knownInterval = true;
            }
            else if (shrinkI) {
                // The dimension is still known as 1 for the processingShape, but will be
                // discarded for the final shape.
                intervalLength = 1;
                knownInterval = true;
            }
            else if (beginAndEndMasked) {
                // Even if we don't have values for begin or end, we do know that this
                // dimension covers the whole interval. If we have shape information for
                // this dimension, that tells us the interval length.
                if (dimI >= 0) {
                    if (denseSpec.strides[i] < 0) {
                        intervalLength = -dimI;
                    }
                    else {
                        intervalLength = dimI;
                    }
                    knownInterval = true;
                }
            }
            if (knownInterval) {
                let sizeI;
                // Hold zero if the interval is degenerate, otherwise account for
                // remainder
                if (intervalLength === 0 ||
                    ((intervalLength < 0) !== (denseSpec.strides[i] < 0))) {
                    sizeI = 0;
                }
                else {
                    sizeI = Math.trunc(intervalLength / denseSpec.strides[i]) +
                        (intervalLength % denseSpec.strides[i] !== 0 ? 1 : 0);
                }
                processingShape.push(sizeI);
            }
            else {
                processingShape.push(-1);
            }
        }
        // Step 4: Compute the final shape
        //
        // newAxis will increase dimension by 1 (with a one-size dimension)
        // slices like foo[3, ...] will reduce dimension by 1.
        // This cannot be done earlier, because it depends on Step 3.
        for (let denseDim = 0; denseDim < denseSpec.finalShapeGatherIndices.length; ++denseDim) {
            const gatherIndex = denseSpec.finalShapeGatherIndices[denseDim];
            if (gatherIndex >= 0) {
                finalShape.push(processingShape[gatherIndex]);
            }
            else if (gatherIndex === NEW_AXIS) {
                finalShape.push(1);
            }
        }
        const finalShapeSparse = finalShape.filter((dim, i) => denseSpec.finalShapeGatherIndices[i] !== NEW_AXIS);
        return {
            finalShapeSparse,
            finalShape,
            isIdentity,
            sliceDim0,
            isSimpleSlice,
            begin: denseSpec.begin,
            end: denseSpec.end,
            strides: denseSpec.strides
        };
    }
    function buildDenseSpec(sparse, dense) {
        dense.beginMask = 0;
        dense.endMask = 0;
        dense.shrinkAxisMask = 0;
        let fullIndex = 0;
        dense.beginValid = sparse.begin != null;
        dense.endValid = sparse.end != null;
        dense.begin = new Array(dense.dims);
        dense.end = new Array(dense.dims);
        dense.strides = new Array(dense.dims);
        dense.finalShapeGatherIndices = [];
        dense.finalShapeGatherIndicesSparse = [];
        dense.inputShapeGatherIndicesSparse = new Array(dense.dims);
        for (let i = 0; i < sparse.dims; i++) {
            if ((1 << i) & sparse.ellipsisMask) {
                // Only the bit that has ellipsis will fall in this condition.
                // Expand the ellipsis into the appropriate indices
                // Note: this only works because we guaranteed one ellipsis.
                const nextIndex = Math.min(dense.dims - (sparse.dims - i) + 1 + sparse.numAddAxisAfterEllipsis, dense.dims);
                for (; fullIndex < nextIndex; fullIndex++) {
                    // newAxis aren't real axis so you have to skip.
                    dense.begin[fullIndex] = 0;
                    dense.end[fullIndex] = 0;
                    dense.strides[fullIndex] = 1;
                    dense.beginMask |= (1 << fullIndex);
                    dense.endMask |= (1 << fullIndex);
                    dense.finalShapeGatherIndices.push(fullIndex);
                    dense.finalShapeGatherIndicesSparse.push(-1);
                    dense.inputShapeGatherIndicesSparse[fullIndex] = i;
                }
            }
            else if ((1 << i) & sparse.newAxisMask) {
                // Only the bit that has newAxis will fall in this condition.
                dense.finalShapeGatherIndices.push(NEW_AXIS);
                dense.finalShapeGatherIndicesSparse.push(-1);
            }
            else {
                if (fullIndex === dense.begin.length) {
                    throw Error(`Index out of range using input dim ${fullIndex}; input ` +
                        `has only ${dense.dims} dims, ${dense.begin.length}.`);
                }
                // Gather slicing spec into appropriate index.
                if (sparse.begin != null) {
                    dense.begin[fullIndex] = sparse.begin[i];
                }
                if (sparse.end != null) {
                    dense.end[fullIndex] = sparse.end[i];
                }
                dense.strides[fullIndex] = sparse.strides[i];
                if (sparse.beginMask & (1 << i)) {
                    dense.beginMask |= (1 << fullIndex);
                }
                if (sparse.endMask & (1 << i)) {
                    dense.endMask |= (1 << fullIndex);
                }
                // If shrink, record where to get the dimensionality from (i.e. newAxis)
                // creates a fake 1 size dimension. Also remember shrink axis (now in
                // dense form) so we can ignore dense.end below.
                if (sparse.shrinkAxisMask & (1 << i)) {
                    dense.finalShapeGatherIndices.push(SHRINK_AXIS);
                    dense.finalShapeGatherIndicesSparse.push(-1);
                    dense.shrinkAxisMask |= (1 << fullIndex);
                }
                else {
                    dense.finalShapeGatherIndices.push(fullIndex);
                    // Remember that where in the sparse shape the dense dim comes from.
                    dense.finalShapeGatherIndicesSparse.push(i);
                }
                dense.inputShapeGatherIndicesSparse[fullIndex] = i;
                fullIndex++;
            }
        }
    }
    function canonical(x, c, strideI, dimI, masks, validRange) {
        if (masks[c]) {
            return strideI > 0 ? validRange[c] : validRange[(c + 1) & 1];
        }
        else {
            const xFwd = x < 0 ? dimI + x : x; // make negative indices positive
            return xFwd < validRange[0] ? validRange[0] :
                xFwd > validRange[1] ? validRange[1] : xFwd;
        }
    }

    var slice_util = /*#__PURE__*/Object.freeze({
        __proto__: null,
        assertParamsValid: assertParamsValid,
        maskToAxes: maskToAxes,
        computeOutShape: computeOutShape$2,
        stridesWithElidedDims: stridesWithElidedDims,
        getNormalizedAxes: getNormalizedAxes,
        startIndicesWithElidedDims: startIndicesWithElidedDims,
        stopIndicesWithElidedDims: stopIndicesWithElidedDims,
        stridesForAxis: stridesForAxis,
        startForAxis: startForAxis,
        stopForAxis: stopForAxis,
        isSliceContinous: isSliceContinous,
        computeFlatOffset: computeFlatOffset,
        parseSliceParams: parseSliceParams,
        sliceInfo: sliceInfo
    });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Serializable defines the serialization contract.
     *
     * TFJS requires serializable classes to return their className when asked
     * to avoid issues with minification.
     */
    class Serializable {
        /**
         * Return the class name for this class to use in serialization contexts.
         *
         * Generally speaking this will be the same thing that constructor.name
         * would have returned.  However, the class name needs to be robust
         * against minification for serialization/deserialization to work properly.
         *
         * There's also places such as initializers.VarianceScaling, where
         * implementation details between different languages led to different
         * class hierarchies and a non-leaf node is used for serialization purposes.
         */
        getClassName() {
            return this.constructor
                .className;
        }
        /**
         * Creates an instance of T from a ConfigDict.
         *
         * This works for most descendants of serializable.  A few need to
         * provide special handling.
         * @param cls A Constructor for the class to instantiate.
         * @param config The Configuration for the object.
         */
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config);
        }
    }
    /**
     * Maps string keys to class constructors.
     *
     * Used during (de)serialization from the cross-language JSON format, which
     * requires the class name in the serialization format matches the class
     * names as used in Python, should it exist.
     */
    class SerializationMap {
        constructor() {
            this.classNameMap = {};
        }
        /**
         * Returns the singleton instance of the map.
         */
        static getMap() {
            if (SerializationMap.instance == null) {
                SerializationMap.instance = new SerializationMap();
            }
            return SerializationMap.instance;
        }
        /**
         * Registers the class as serializable.
         */
        static register(cls) {
            SerializationMap.getMap().classNameMap[cls.className] =
                [cls, cls.fromConfig];
        }
    }
    /**
     * Register a class with the serialization map of TensorFlow.js.
     *
     * This is often used for registering custom Layers, so they can be
     * serialized and deserialized.
     *
     * Example:
     *
     * ```js
     * class MyCustomLayer extends tf.layers.Layer {
     *   static className = 'MyCustomLayer';
     *
     *   constructor(config) {
     *     super(config);
     *   }
     * }
     * tf.serialization.registerClass(MyCustomLayer);
     * ```
     *
     * @param cls The class to be registered. It must have a public static member
     *   called `className` defined and the value must be a non-empty string.
     *
     * @doc {heading: 'Models', subheading: 'Serialization', ignoreCI: true}
     */
    function registerClass(cls) {
        assert$4(cls.className != null, () => `Class being registered does not have the static className ` +
            `property defined.`);
        assert$4(typeof cls.className === 'string', () => `className is required to be a string, but got type ` +
            typeof cls.className);
        assert$4(cls.className.length > 0, () => `Class being registered has an empty-string as its className, ` +
            `which is disallowed.`);
        SerializationMap.register(cls);
    }

    var serialization = /*#__PURE__*/Object.freeze({
        __proto__: null,
        Serializable: Serializable,
        SerializationMap: SerializationMap,
        registerClass: registerClass
    });

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const TEST_EPSILON_FLOAT32 = 1e-3;
    const TEST_EPSILON_FLOAT16 = 1e-1;
    function expectArraysClose(actual, expected, epsilon) {
        if (epsilon == null) {
            epsilon = testEpsilon();
        }
        return expectArraysPredicate(actual, expected, (a, b) => areClose(a, b, epsilon));
    }
    function testEpsilon() {
        return ENGINE.backend.floatPrecision() === 32 ? TEST_EPSILON_FLOAT32 :
            TEST_EPSILON_FLOAT16;
    }
    function expectArraysPredicate(actual, expected, predicate) {
        let checkClassType = true;
        if (isTypedArray$1(actual) || isTypedArray$1(expected)) {
            checkClassType = false;
        }
        if (isTypedArray$1(actual) && isTypedArray$1(expected)) {
            checkClassType = true;
        }
        if (checkClassType) {
            const aType = actual.constructor.name;
            const bType = expected.constructor.name;
            if (aType !== bType) {
                throw new Error(`Arrays are of different type. Actual: ${aType}. ` +
                    `Expected: ${bType}`);
            }
        }
        if (Array.isArray(actual) && Array.isArray(expected)) {
            const actualShape = inferShape(actual);
            const expectedShape = inferShape(expected);
            if (!arraysEqual(actualShape, expectedShape)) {
                throw new Error(`Arrays have different shapes. ` +
                    `Actual: [${actualShape}]. Expected: [${expectedShape}]`);
            }
        }
        const actualFlat = isTypedArray$1(actual) ? actual : flatten$3(actual);
        const expectedFlat = isTypedArray$1(expected) ?
            expected :
            flatten$3(expected);
        if (actualFlat.length !== expectedFlat.length) {
            throw new Error(`Arrays have different lengths actual: ${actualFlat.length} vs ` +
                `expected: ${expectedFlat.length}.\n` +
                `Actual:   ${actualFlat}.\n` +
                `Expected: ${expectedFlat}.`);
        }
        for (let i = 0; i < expectedFlat.length; ++i) {
            const a = actualFlat[i];
            const e = expectedFlat[i];
            if (!predicate(a, e)) {
                throw new Error(`Arrays differ: actual[${i}] = ${a}, expected[${i}] = ${e}.\n` +
                    `Actual:   ${actualFlat}.\n` +
                    `Expected: ${expectedFlat}.`);
            }
        }
    }
    function expectPromiseToFail(fn, done) {
        fn().then(() => done.fail(), () => done());
    }
    function expectArraysEqual(actual, expected) {
        const exp = typeof expected === 'string' || typeof expected === 'number' ||
            typeof expected === 'boolean' ?
            [expected] :
            expected;
        if (isString$2(actual) || isString$2(actual[0]) ||
            isString$2(expected) || isString$2(expected[0])) {
            // tslint:disable-next-line: triple-equals
            return expectArraysPredicate(actual, exp, (a, b) => a == b);
        }
        return expectArraysPredicate(actual, expected, (a, b) => areClose(a, b, 0));
    }
    function expectNumbersClose(a, e, epsilon) {
        if (epsilon == null) {
            epsilon = testEpsilon();
        }
        if (!areClose(a, e, epsilon)) {
            throw new Error(`Numbers differ: actual === ${a}, expected === ${e}`);
        }
    }
    function areClose(a, e, epsilon) {
        if (!isFinite(a) && !isFinite(e)) {
            return true;
        }
        if (isNaN(a) || isNaN(e) || Math.abs(a - e) > epsilon) {
            return false;
        }
        return true;
    }
    function expectValuesInRange(actual, low, high) {
        for (let i = 0; i < actual.length; i++) {
            if (actual[i] < low || actual[i] > high) {
                throw new Error(`Value out of range:${actual[i]} low: ${low}, high: ${high}`);
            }
        }
    }
    function expectArrayBuffersEqual(actual, expected) {
        // Safari does not like comparing ArrayBuffers directly. Wrapping in
        // a Float32Array solves this issue.
        const actualArray = new Float32Array(actual);
        const expectedArray = new Float32Array(expected);
        if (actualArray.length !== expectedArray.length) {
            throw new Error('Expected ArrayBuffer to be of length ' +
                `${expectedArray.length}, but it was ${actualArray.length}`);
        }
        for (let i = 0; i < expectedArray.length; i++) {
            if (actualArray[i] !== expectedArray[i]) {
                throw new Error(`Expected ArrayBuffer value at ${i} to be ` +
                    `${expectedArray[i]} but got ${actualArray[i]} instead`);
            }
        }
    }
    /** Encodes strings into utf-8 bytes. */
    function encodeStrings(a) {
        for (let i = 0; i < a.length; i++) {
            const val = a[i];
            if (Array.isArray(val)) {
                encodeStrings(val);
            }
            else {
                a[i] = encodeString(val);
            }
        }
        return a;
    }

    var test_util = /*#__PURE__*/Object.freeze({
        __proto__: null,
        TEST_EPSILON_FLOAT16: TEST_EPSILON_FLOAT16,
        expectArraysClose: expectArraysClose,
        testEpsilon: testEpsilon,
        expectPromiseToFail: expectPromiseToFail,
        expectArraysEqual: expectArraysEqual,
        expectNumbersClose: expectNumbersClose,
        expectValuesInRange: expectValuesInRange,
        expectArrayBuffersEqual: expectArrayBuffersEqual,
        encodeStrings: encodeStrings
    });

    /** @license See the LICENSE file. */
    // This code is auto-generated, do not modify this file!
    const version$7 = '3.17.0';

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Enables production mode which disables correctness checks in favor of
     * performance.
     *
     * @doc {heading: 'Environment'}
     */
    function enableProdMode() {
        env$1().set('PROD', true);
    }
    /**
     * Enables debug mode which will log information about all executed kernels:
     * the elapsed time of the kernel execution, as well as the rank, shape, and
     * size of the output tensor.
     *
     * Debug mode will significantly slow down your application as it will
     * download the result of every operation to the CPU. This should not be used in
     * production. Debug mode does not affect the timing information of the kernel
     * execution as we do not measure download time in the kernel execution time.
     *
     * See also: `tf.profile`, `tf.memory`.
     *
     * @doc {heading: 'Environment'}
     */
    function enableDebugMode() {
        env$1().set('DEBUG', true);
    }
    /** Globally disables deprecation warnings */
    function disableDeprecationWarnings() {
        env$1().set('DEPRECATION_WARNINGS_ENABLED', false);
        console.warn(`TensorFlow.js deprecation warnings have been disabled.`);
    }
    /** Warn users about deprecated functionality. */
    function deprecationWarn(msg) {
        if (env$1().getBool('DEPRECATION_WARNINGS_ENABLED')) {
            console.warn(msg + ' You can disable deprecation warnings with ' +
                'tf.disableDeprecationWarnings().');
        }
    }
    /**
     * Dispose all variables kept in backend engine.
     *
     * @doc {heading: 'Environment'}
     */
    function disposeVariables() {
        ENGINE.disposeVariables();
    }
    /**
     * It returns the global engine that keeps track of all tensors and backends.
     *
     * @doc {heading: 'Environment'}
     */
    function engine() {
        return ENGINE;
    }
    /**
     * Returns memory info at the current time in the program. The result is an
     * object with the following properties:
     *
     * - `numBytes`: Number of bytes allocated (undisposed) at this time.
     * - `numTensors`: Number of unique tensors allocated.
     * - `numDataBuffers`: Number of unique data buffers allocated
     *   (undisposed) at this time, which is ≤ the number of tensors
     *   (e.g. `a.reshape(newShape)` makes a new Tensor that shares the same
     *   data buffer with `a`).
     * - `unreliable`: True if the memory usage is unreliable. See `reasons` when
     *    `unreliable` is true.
     * - `reasons`: `string[]`, reasons why the memory is unreliable, present if
     *    `unreliable` is true.
     *
     * WebGL Properties:
     * - `numBytesInGPU`: Number of bytes allocated (undisposed) in the GPU only at
     *     this time.
     *
     * @doc {heading: 'Performance', subheading: 'Memory'}
     */
    function memory() {
        return ENGINE.memory();
    }
    /**
     * Executes the provided function `f()` and returns a promise that resolves
     * with information about the function's memory use:
     * - `newBytes`: the number of new bytes allocated
     * - `newTensors`: the number of new tensors created
     * - `peakBytes`: the peak number of bytes allocated
     * - `kernels`: an array of objects for each kernel involved that reports
     * their input and output shapes, number of bytes used, and number of new
     * tensors created.
     * - `kernelNames`: an array of unique strings with just the names of the
     * kernels in the `kernels` array.
     *
     * ```js
     * const profile = await tf.profile(() => {
     *   const x = tf.tensor1d([1, 2, 3]);
     *   let x2 = x.square();
     *   x2.dispose();
     *   x2 = x.square();
     *   x2.dispose();
     *   return x;
     * });
     *
     * console.log(`newBytes: ${profile.newBytes}`);
     * console.log(`newTensors: ${profile.newTensors}`);
     * console.log(`byte usage over all kernels: ${profile.kernels.map(k =>
     * k.totalBytesSnapshot)}`);
     * ```
     *
     *
     * @doc {heading: 'Performance', subheading: 'Profile'}
     */
    function profile(f) {
        return ENGINE.profile(f);
    }
    /**
     * Executes the provided function `fn` and after it is executed, cleans up all
     * intermediate tensors allocated by `fn` except those returned by `fn`.
     * `fn` must not return a Promise (async functions not allowed). The returned
     * result can be a complex object.
     *
     * Using this method helps avoid memory leaks. In general, wrap calls to
     * operations in `tf.tidy` for automatic memory cleanup.
     *
     * NOTE: Variables do *not* get cleaned up when inside a tidy(). If you want to
     * dispose variables, please use `tf.disposeVariables` or call dispose()
     * directly on variables.
     *
     * ```js
     * // y = 2 ^ 2 + 1
     * const y = tf.tidy(() => {
     *   // a, b, and one will be cleaned up when the tidy ends.
     *   const one = tf.scalar(1);
     *   const a = tf.scalar(2);
     *   const b = a.square();
     *
     *   console.log('numTensors (in tidy): ' + tf.memory().numTensors);
     *
     *   // The value returned inside the tidy function will return
     *   // through the tidy, in this case to the variable y.
     *   return b.add(one);
     * });
     *
     * console.log('numTensors (outside tidy): ' + tf.memory().numTensors);
     * y.print();
     * ```
     *
     * @param nameOrFn The name of the closure, or the function to execute.
     *     If a name is provided, the 2nd argument should be the function.
     *     If debug mode is on, the timing and the memory usage of the function
     *     will be tracked and displayed on the console using the provided name.
     * @param fn The function to execute.
     *
     * @doc {heading: 'Performance', subheading: 'Memory'}
     */
    function tidy(nameOrFn, fn) {
        return ENGINE.tidy(nameOrFn, fn);
    }
    /**
     * Disposes any `tf.Tensor`s found within the provided object.
     *
     * @param container an object that may be a `tf.Tensor` or may directly
     *     contain `tf.Tensor`s, such as a `Tensor[]` or `{key: Tensor, ...}`. If
     *     the object is not a `tf.Tensor` or does not contain `Tensors`, nothing
     *     happens. In general it is safe to pass any object here, except that
     *     `Promise`s are not supported.
     *
     * @doc {heading: 'Performance', subheading: 'Memory'}
     */
    function dispose(container) {
        const tensors = getTensorsInContainer(container);
        tensors.forEach(tensor => tensor.dispose());
    }
    /**
     * Keeps a `tf.Tensor` generated inside a `tf.tidy` from being disposed
     * automatically.
     *
     * ```js
     * let b;
     * const y = tf.tidy(() => {
     *   const one = tf.scalar(1);
     *   const a = tf.scalar(2);
     *
     *   // b will not be cleaned up by the tidy. a and one will be cleaned up
     *   // when the tidy ends.
     *   b = tf.keep(a.square());
     *
     *   console.log('numTensors (in tidy): ' + tf.memory().numTensors);
     *
     *   // The value returned inside the tidy function will return
     *   // through the tidy, in this case to the variable y.
     *   return b.add(one);
     * });
     *
     * console.log('numTensors (outside tidy): ' + tf.memory().numTensors);
     * console.log('y:');
     * y.print();
     * console.log('b:');
     * b.print();
     * ```
     *
     * @param result The tensor to keep from being disposed.
     *
     * @doc {heading: 'Performance', subheading: 'Memory'}
     */
    function keep(result) {
        return ENGINE.keep(result);
    }
    /**
     * Executes `f()` and returns a promise that resolves with timing
     * information.
     *
     * The result is an object with the following properties:
     *
     * - `wallMs`: Wall execution time.
     * - `kernelMs`: Kernel execution time, ignoring data transfer. If using the
     * WebGL backend and the query timer extension is not available, this will
     * return an error object.
     * - On `WebGL` The following additional properties exist:
     *   - `uploadWaitMs`: CPU blocking time on texture uploads.
     *   - `downloadWaitMs`: CPU blocking time on texture downloads (readPixels).
     *
     * ```js
     * const x = tf.randomNormal([20, 20]);
     * const time = await tf.time(() => x.matMul(x));
     *
     * console.log(`kernelMs: ${time.kernelMs}, wallTimeMs: ${time.wallMs}`);
     * ```
     *
     * @param f The function to execute and time.
     *
     * @doc {heading: 'Performance', subheading: 'Timing'}
     */
    function time(f) {
        return ENGINE.time(f);
    }
    /**
     * Sets the backend (cpu, webgl, wasm, etc) responsible for creating tensors and
     * executing operations on those tensors. Returns a promise that resolves
     * to a boolean if the backend initialization was successful.
     *
     * Note this disposes the current backend, if any, as well as any tensors
     * associated with it. A new backend is initialized, even if it is of the
     * same type as the previous one.
     *
     * @param backendName The name of the backend. Currently supports
     *     `'webgl'|'cpu'` in the browser, `'tensorflow'` under node.js
     *     (requires tfjs-node), and `'wasm'` (requires tfjs-backend-wasm).
     *
     * @doc {heading: 'Backends'}
     */
    function setBackend(backendName) {
        return ENGINE.setBackend(backendName);
    }
    /**
     * Returns a promise that resolves when the currently selected backend (or the
     * highest priority one) has initialized. Await this promise when you are using
     * a backend that has async initialization.
     *
     * @doc {heading: 'Backends'}
     */
    function ready() {
        return ENGINE.ready();
    }
    /**
     * Returns the current backend name (cpu, webgl, etc). The backend is
     * responsible for creating tensors and executing operations on those tensors.
     *
     * @doc {heading: 'Backends'}
     */
    function getBackend() {
        return ENGINE.backendName;
    }
    /**
     * Removes a backend and the registered factory.
     *
     * @doc {heading: 'Backends'}
     */
    function removeBackend(name) {
        ENGINE.removeBackend(name);
    }
    /**
     * Finds the backend registered under the provided name. Returns null if the
     * name is not in the registry, or the registration hasn't finished yet.
     */
    function findBackend(name) {
        return ENGINE.findBackend(name);
    }
    /**
     * Finds the backend factory registered under the provided name. Returns a
     * function that produces a new backend when called. Returns null if the name
     * is not in the registry.
     */
    function findBackendFactory(name) {
        return ENGINE.findBackendFactory(name);
    }
    /**
     * Registers a global backend. The registration should happen when importing
     * a module file (e.g. when importing `backend_webgl.ts`), and is used for
     * modular builds (e.g. custom tfjs bundle with only webgl support).
     *
     * @param factory The backend factory function. When called, it should
     * return a backend instance, or a promise of an instance.
     * @param priority The priority of the backend (higher = more important).
     *     In case multiple backends are registered, the priority is used to find
     *     the best backend. Defaults to 1.
     * @return False if there is already a registered backend under this name, true
     *     if not.
     *
     * @doc {heading: 'Backends'}
     */
    function registerBackend(name, factory, priority = 1) {
        return ENGINE.registerBackend(name, factory, priority);
    }
    /**
     * Gets the current backend. If no backends have been initialized, this will
     * attempt to initialize the best backend. Will throw an error if the highest
     * priority backend has async initialization, in which case, you should call
     * 'await tf.ready()' before running other code.
     *
     * @doc {heading: 'Backends'}
     */
    function backend() {
        return ENGINE.backend;
    }
    /**
     * Sets the global platform.
     *
     * @param platformName The name of this platform.
     * @param platform A platform implementation.
     */
    function setPlatform(platformName, platform) {
        env$1().setPlatform(platformName, platform);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Adds two `tf.Tensor`s element-wise, A + B. Supports broadcasting.
     *
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3, 4]);
     * const b = tf.tensor1d([10, 20, 30, 40]);
     *
     * a.add(b).print();  // or tf.add(a, b)
     * ```
     *
     * ```js
     * // Broadcast add a with b.
     * const a = tf.scalar(5);
     * const b = tf.tensor1d([10, 20, 30, 40]);
     *
     * a.add(b).print();  // or tf.add(a, b)
     * ```
     * @param a The first `tf.Tensor` to add.
     * @param b The second `tf.Tensor` to add. Must have the same type as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function add_(a, b) {
        let $a = convertToTensor(a, 'a', 'add');
        let $b = convertToTensor(b, 'b', 'add');
        [$a, $b] = makeTypesMatch($a, $b);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Add$1, inputs);
    }
    const add$2 = op({ add_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Divides two `tf.Tensor`s element-wise, A / B. Supports broadcasting.
     * The result is rounded with floor function.
     *
     *
     * ```js
     * const a = tf.tensor1d([1, 4, 9, 16]);
     * const b = tf.tensor1d([1, 2, 3, 4]);
     *
     * a.floorDiv(b).print();  // or tf.div(a, b)
     * ```
     *
     * ```js
     * // Broadcast div a with b.
     * const a = tf.tensor1d([2, 4, 6, 8]);
     * const b = tf.scalar(2);
     *
     * a.floorDiv(b).print();  // or tf.floorDiv(a, b)
     * ```
     *
     * @param a The first tensor as the numerator.
     * @param b The second tensor as the denominator. Must have the same dtype as
     * `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function floorDiv_(a, b) {
        let $a = convertToTensor(a, 'a', 'floorDiv');
        let $b = convertToTensor(b, 'b', 'floorDiv');
        [$a, $b] = makeTypesMatch($a, $b);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(FloorDiv, inputs);
    }
    const floorDiv$2 = op({ floorDiv_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Divides two `tf.Tensor`s element-wise, A / B. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([1, 4, 9, 16]);
     * const b = tf.tensor1d([1, 2, 3, 4]);
     *
     * a.div(b).print();  // or tf.div(a, b)
     * ```
     *
     * ```js
     * // Broadcast div a with b.
     * const a = tf.tensor1d([2, 4, 6, 8]);
     * const b = tf.scalar(2);
     *
     * a.div(b).print();  // or tf.div(a, b)
     * ```
     *
     * @param a The first tensor as the numerator.
     * @param b The second tensor as the denominator. Must have the same dtype as
     * `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function div_(a, b) {
        let $a = convertToTensor(a, 'a', 'div');
        let $b = convertToTensor(b, 'b', 'div');
        [$a, $b] = makeTypesMatch($a, $b);
        if ($a.dtype === 'int32' && $b.dtype === 'int32') {
            return floorDiv$2($a, $b);
        }
        const inputs = { a: $a, b: $b };
        const attrs = {};
        // tslint:disable-next-line: no-unnecessary-type-assertion
        return ENGINE.runKernel(RealDiv, inputs, attrs);
    }
    const div$1 = op({ div_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Multiplies two `tf.Tensor`s element-wise, A * B. Supports broadcasting.
     *
     * We also expose `tf.mulStrict` which has the same signature as this op and
     * asserts that `a` and `b` are the same shape (does not broadcast).
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3, 4]);
     * const b = tf.tensor1d([2, 3, 4, 5]);
     *
     * a.mul(b).print();  // or tf.mul(a, b)
     * ```
     *
     * ```js
     * // Broadcast mul a with b.
     * const a = tf.tensor1d([1, 2, 3, 4]);
     * const b = tf.scalar(5);
     *
     * a.mul(b).print();  // or tf.mul(a, b)
     * ```
     * @param a The first tensor to multiply.
     * @param b The second tensor to multiply. Must have the same dtype as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function mul_(a, b) {
        let $a = convertToTensor(a, 'a', 'mul');
        let $b = convertToTensor(b, 'b', 'mul');
        [$a, $b] = makeTypesMatch($a, $b);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Multiply$1, inputs);
    }
    const mul = op({ mul_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes absolute value element-wise: `abs(x)`
     *
     * ```js
     * const x = tf.tensor1d([-1, 2, -3, 4]);
     *
     * x.abs().print();  // or tf.abs(x)
     * ```
     * @param x The input `tf.Tensor`.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function abs_(x) {
        const $x = convertToTensor(x, 'x', 'abs');
        if ($x.dtype === 'complex64') {
            const inputs = { x: $x };
            return ENGINE.runKernel(ComplexAbs, inputs);
        }
        else {
            const inputs = { x: $x };
            return ENGINE.runKernel(Abs, inputs);
        }
    }
    const abs$2 = op({ abs_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes acos of the input `tf.Tensor` element-wise: `acos(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, 1, -1, .7]);
     *
     * x.acos().print();  // or tf.acos(x)
     * ```
     * @param x The input tensor.
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function acos_(x) {
        const $x = convertToTensor(x, 'x', 'acos');
        const inputs = { x: $x };
        return ENGINE.runKernel(Acos, inputs);
    }
    const acos$2 = op({ acos_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the inverse hyperbolic cos of the input `tf.Tensor` element-wise:
     * `acosh(x)`
     *
     * ```js
     * const x = tf.tensor1d([10, 1, 3, 5.7]);
     *
     * x.acosh().print();  // or tf.acosh(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function acosh_(x) {
        const $x = convertToTensor(x, 'x', 'acosh');
        const inputs = { x: $x };
        return ENGINE.runKernel(Acosh, inputs);
    }
    const acosh$2 = op({ acosh_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Adds a list of `tf.Tensor`s element-wise, each with the same shape and dtype.
     *
     * ```js
     * const a = tf.tensor1d([1, 2]);
     * const b = tf.tensor1d([3, 4]);
     * const c = tf.tensor1d([5, 6]);
     *
     * tf.addN([a, b, c]).print();
     * ```
     * @param tensors A list of tensors with the same shape and dtype.
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function addN_(tensors) {
        assert$4(Array.isArray(tensors), () => 'The argument passed to tf.addN() must be a list of tensors');
        assert$4(tensors.length >= 1, () => `Must pass at least one tensor to tf.addN(), but got ` +
            `${tensors.length}`);
        const $tensors = tensors.map((t, i) => convertToTensor(t, `tensors${i}`, 'addN'));
        const firstTensor = $tensors[0];
        $tensors.forEach(t => {
            if (t.dtype !== firstTensor.dtype) {
                throw new Error('All tensors passed to tf.addN() must have the same dtype');
            }
        });
        $tensors.forEach(t => {
            if (!arraysEqual(t.shape, firstTensor.shape)) {
                throw new Error('All tensors passed to tf.addN() must have the same shape');
            }
        });
        const inputs = $tensors;
        return ENGINE.runKernel(AddN, inputs);
    }
    const addN$2 = op({ addN_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the logical and of elements across dimensions of a `tf.Tensor`.
     *
     * Reduces the input along the dimensions given in `axes`. Unless `keepDims`
     * is true, the rank of the `tf.Tensor` is reduced by 1 for each entry in
     * `axes`. If `keepDims` is true, the reduced dimensions are retained with
     * length 1. If `axes` has no entries, all dimensions are reduced, and an
     * `tf.Tensor` with a single element is returned.
     *
     * ```js
     * const x = tf.tensor1d([1, 1, 1], 'bool');
     *
     * x.all().print();  // or tf.all(x)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 1, 0, 0], [2, 2], 'bool');
     *
     * const axis = 1;
     * x.all(axis).print();  // or tf.all(x, axis)
     * ```
     *
     * @param x The input tensor. Must be of dtype bool.
     * @param axis The dimension(s) to reduce. By default it reduces
     *     all dimensions.
     * @param keepDims If true, retains reduced dimensions with size 1.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function all_(x, axis = null, keepDims = false) {
        const $x = convertToTensor(x, 'x', 'all', 'bool');
        const inputs = { x: $x };
        const attrs = { axis, keepDims };
        return ENGINE.runKernel(All, inputs, attrs);
    }
    const all$2 = op({ all_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the logical or of elements across dimensions of a `tf.Tensor`.
     *
     * Reduces the input along the dimensions given in `axes`. Unless `keepDims`
     * is true, the rank of the `tf.Tensor` is reduced by 1 for each entry in
     * `axes`. If `keepDims` is true, the reduced dimensions are retained with
     * length 1. If `axes` has no entries, all dimensions are reduced, and an
     * `tf.Tensor` with a single element is returned.
     *
     * ```js
     * const x = tf.tensor1d([1, 1, 1], 'bool');
     *
     * x.any().print();  // or tf.any(x)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 1, 0, 0], [2, 2], 'bool');
     *
     * const axis = 1;
     * x.any(axis).print();  // or tf.any(x, axis)
     * ```
     *
     * @param x The input tensor. Must be of dtype bool.
     * @param axis The dimension(s) to reduce. By default it reduces
     *     all dimensions.
     * @param keepDims If true, retains reduced dimensions with size 1.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function any_(x, axis = null, keepDims = false) {
        const $x = convertToTensor(x, 'x', 'any', 'bool');
        const inputs = { x: $x };
        const attrs = { axis, keepDims };
        return ENGINE.runKernel(Any, inputs, attrs);
    }
    // tslint:disable-next-line:variable-name
    const any$2 = op({ any_ });

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the indices of the maximum values along an `axis`.
     *
     * The result has the same shape as `input` with the dimension along `axis`
     * removed.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3]);
     *
     * x.argMax().print();  // or tf.argMax(x)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 4, 3], [2, 2]);
     *
     * const axis = 1;
     * x.argMax(axis).print();  // or tf.argMax(x, axis)
     * ```
     *
     * @param x The input tensor.
     * @param axis The dimension to reduce. Defaults to 0 (outer-most dimension).
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function argMax_(x, axis = 0) {
        const $x = convertToTensor(x, 'x', 'argMax');
        const inputs = { x: $x };
        const attrs = { axis };
        return ENGINE.runKernel(ArgMax, inputs, attrs);
    }
    const argMax$2 = op({ argMax_ });

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the indices of the minimum values along an `axis`.
     *
     * The result has the same shape as `input` with the dimension along `axis`
     * removed.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3]);
     *
     * x.argMin().print();  // or tf.argMin(x)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 4, 3], [2, 2]);
     *
     * const axis = 1;
     * x.argMin(axis).print();  // or tf.argMin(x, axis)
     * ```
     *
     * @param x The input tensor.
     * @param axis The dimension to reduce. Defaults to 0 (outer-most dimension).
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function argMin_(x, axis = 0) {
        const $x = convertToTensor(x, 'x', 'argMin');
        const inputs = { x: $x };
        const attrs = { axis };
        return ENGINE.runKernel(ArgMin, inputs, attrs);
    }
    const argMin$2 = op({ argMin_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes asin of the input `tf.Tensor` element-wise: `asin(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, 1, -1, .7]);
     *
     * x.asin().print();  // or tf.asin(x)
     * ```
     * @param x The input tensor.
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function asin_(x) {
        const $x = convertToTensor(x, 'x', 'asin');
        const inputs = { x: $x };
        return ENGINE.runKernel(Asin, inputs);
    }
    const asin$2 = op({ asin_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes inverse hyperbolic sin of the input `tf.Tensor` element-wise:
     * `asinh(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, 1, -1, .7]);
     *
     * x.asinh().print();  // or tf.asinh(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function asinh_(x) {
        const $x = convertToTensor(x, 'x', 'asinh');
        const inputs = { x: $x };
        return ENGINE.runKernel(Asinh, inputs);
    }
    const asinh$2 = op({ asinh_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes atan of the input `tf.Tensor` element-wise: `atan(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, 1, -1, .7]);
     *
     * x.atan().print();  // or tf.atan(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function atan_(x) {
        const $x = convertToTensor(x, 'x', 'atan');
        const inputs = { x: $x };
        return ENGINE.runKernel(Atan, inputs);
    }
    const atan$2 = op({ atan_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes arctangent of `tf.Tensor`s a / b element-wise: `atan2(a, b)`.
     * Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([1.0, 1.0, -1.0, .7]);
     * const b = tf.tensor1d([2.0, 13.0, 3.5, .21]);
     *
     * tf.atan2(a, b).print()
     * ```
     *
     * @param a The first tensor.
     * @param b The second tensor. Must have the same dtype as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function atan2_(a, b) {
        let $a = convertToTensor(a, 'a', 'atan2');
        let $b = convertToTensor(b, 'b', 'atan2');
        [$a, $b] = makeTypesMatch($a, $b);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Atan2, inputs);
    }
    const atan2$2 = op({ atan2_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes inverse hyperbolic tan of the input `tf.Tensor` element-wise:
     * `atanh(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, .1, -.1, .7]);
     *
     * x.atanh().print();  // or tf.atanh(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function atanh_(x) {
        const $x = convertToTensor(x, 'x', 'atanh');
        const inputs = { x: $x };
        return ENGINE.runKernel(Atanh, inputs);
    }
    const atanh$2 = op({ atanh_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     *
     * @param inputShape Input tensor shape is of the following dimensions:
     *     `[batch, height, width, inChannels]`.
     * @param filterShape The filter shape is of the following dimensions:
     *     `[filterHeight, filterWidth, depth]`.
     * @param strides The strides of the sliding window for each dimension of the
     *     input tensor: `[strideHeight, strideWidth]`.
     *     If `strides` is a single number,
     *     then `strideHeight == strideWidth`.
     * @param pad The type of padding algorithm.
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1*1x1.
     *    - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dataFormat The data format of the input and output data.
     *     Defaults to 'NHWC'.
     * @param dilations The dilation rates: `[dilationHeight, dilationWidth]`.
     *     Defaults to `[1, 1]`. If `dilations` is a single number, then
     *     `dilationHeight == dilationWidth`.
     */
    function computeDilation2DInfo(inputShape, filterShape, strides, pad, dataFormat = 'NHWC', dilations) {
        // `computerConv2DInfo` require filterShape to be in the dimension of:
        // `[filterHeight, filterWidth, depth, outDepth]`, dilation2d doesn't have
        // outDepth, it should have the same depth as the input.
        // Input shape: [batch, height, width, inChannels]
        const inputChannels = inputShape[3];
        const $filterShape = [...filterShape, inputChannels];
        const $dataFormat = convertConv2DDataFormat(dataFormat);
        return computeConv2DInfo(inputShape, $filterShape, strides, dilations, pad, null /* roundingMode */, null /* depthWise */, $dataFormat);
    }
    function computePool2DInfo(inShape, filterSize, strides, dilations, pad, roundingMode, dataFormat = 'channelsLast') {
        const [filterHeight, filterWidth] = parseTupleParam(filterSize);
        let filterShape;
        if (dataFormat === 'channelsLast') {
            filterShape = [filterHeight, filterWidth, inShape[3], inShape[3]];
        }
        else if (dataFormat === 'channelsFirst') {
            filterShape = [filterHeight, filterWidth, inShape[1], inShape[1]];
        }
        else {
            throw new Error(`Unknown dataFormat ${dataFormat}`);
        }
        return computeConv2DInfo(inShape, filterShape, strides, dilations, pad, roundingMode, false, dataFormat);
    }
    /**
     * Computes the information for a forward pass of a pooling3D operation.
     */
    function computePool3DInfo(inShape, filterSize, strides, dilations, pad, roundingMode, dataFormat = 'NDHWC') {
        const [filterDepth, filterHeight, filterWidth] = parse3TupleParam(filterSize);
        let filterShape;
        let $dataFormat;
        if (dataFormat === 'NDHWC') {
            $dataFormat = 'channelsLast';
            filterShape =
                [filterDepth, filterHeight, filterWidth, inShape[4], inShape[4]];
        }
        else if (dataFormat === 'NCDHW') {
            $dataFormat = 'channelsFirst';
            filterShape =
                [filterDepth, filterHeight, filterWidth, inShape[1], inShape[1]];
        }
        else {
            throw new Error(`Unknown dataFormat ${dataFormat}`);
        }
        return computeConv3DInfo(inShape, filterShape, strides, dilations, pad, false, $dataFormat, roundingMode);
    }
    /**
     * Computes the information for a forward pass of a convolution/pooling
     * operation.
     */
    function computeConv2DInfo(inShape, filterShape, strides, dilations, pad, roundingMode, depthwise = false, dataFormat = 'channelsLast') {
        let [batchSize, inHeight, inWidth, inChannels] = [-1, -1, -1, -1];
        if (dataFormat === 'channelsLast') {
            [batchSize, inHeight, inWidth, inChannels] = inShape;
        }
        else if (dataFormat === 'channelsFirst') {
            [batchSize, inChannels, inHeight, inWidth] = inShape;
        }
        else {
            throw new Error(`Unknown dataFormat ${dataFormat}`);
        }
        const [filterHeight, filterWidth, , filterChannels] = filterShape;
        const [strideHeight, strideWidth] = parseTupleParam(strides);
        const [dilationHeight, dilationWidth] = parseTupleParam(dilations);
        const effectiveFilterHeight = getEffectiveFilterSize(filterHeight, dilationHeight);
        const effectiveFilterWidth = getEffectiveFilterSize(filterWidth, dilationWidth);
        const { padInfo, outHeight, outWidth } = getPadAndOutInfo(pad, inHeight, inWidth, strideHeight, strideWidth, effectiveFilterHeight, effectiveFilterWidth, roundingMode, dataFormat);
        const outChannels = depthwise ? filterChannels * inChannels : filterChannels;
        let outShape;
        if (dataFormat === 'channelsFirst') {
            outShape = [batchSize, outChannels, outHeight, outWidth];
        }
        else if (dataFormat === 'channelsLast') {
            outShape = [batchSize, outHeight, outWidth, outChannels];
        }
        return {
            batchSize,
            dataFormat,
            inHeight,
            inWidth,
            inChannels,
            outHeight,
            outWidth,
            outChannels,
            padInfo,
            strideHeight,
            strideWidth,
            filterHeight,
            filterWidth,
            effectiveFilterHeight,
            effectiveFilterWidth,
            dilationHeight,
            dilationWidth,
            inShape,
            outShape,
            filterShape
        };
    }
    /**
     * Computes the information for a forward pass of a 3D convolution/pooling
     * operation.
     */
    function computeConv3DInfo(inShape, filterShape, strides, dilations, pad, depthwise = false, dataFormat = 'channelsLast', roundingMode) {
        let [batchSize, inDepth, inHeight, inWidth, inChannels] = [-1, -1, -1, -1, -1];
        if (dataFormat === 'channelsLast') {
            [batchSize, inDepth, inHeight, inWidth, inChannels] = inShape;
        }
        else if (dataFormat === 'channelsFirst') {
            [batchSize, inChannels, inDepth, inHeight, inWidth] = inShape;
        }
        else {
            throw new Error(`Unknown dataFormat ${dataFormat}`);
        }
        const [filterDepth, filterHeight, filterWidth, , filterChannels] = filterShape;
        const [strideDepth, strideHeight, strideWidth] = parse3TupleParam(strides);
        const [dilationDepth, dilationHeight, dilationWidth] = parse3TupleParam(dilations);
        const effectiveFilterDepth = getEffectiveFilterSize(filterDepth, dilationDepth);
        const effectiveFilterHeight = getEffectiveFilterSize(filterHeight, dilationHeight);
        const effectiveFilterWidth = getEffectiveFilterSize(filterWidth, dilationWidth);
        const { padInfo, outDepth, outHeight, outWidth } = get3DPadAndOutInfo(pad, inDepth, inHeight, inWidth, strideDepth, strideHeight, strideWidth, effectiveFilterDepth, effectiveFilterHeight, effectiveFilterWidth, roundingMode);
        const outChannels = depthwise ? filterChannels * inChannels : filterChannels;
        let outShape;
        if (dataFormat === 'channelsFirst') {
            outShape = [batchSize, outChannels, outDepth, outHeight, outWidth];
        }
        else if (dataFormat === 'channelsLast') {
            outShape = [batchSize, outDepth, outHeight, outWidth, outChannels];
        }
        return {
            batchSize,
            dataFormat,
            inDepth,
            inHeight,
            inWidth,
            inChannels,
            outDepth,
            outHeight,
            outWidth,
            outChannels,
            padInfo,
            strideDepth,
            strideHeight,
            strideWidth,
            filterDepth,
            filterHeight,
            filterWidth,
            effectiveFilterDepth,
            effectiveFilterHeight,
            effectiveFilterWidth,
            dilationDepth,
            dilationHeight,
            dilationWidth,
            inShape,
            outShape,
            filterShape
        };
    }
    function computeOutputShape2D(inShape, fieldSize, stride, zeroPad, roundingMode) {
        if (zeroPad == null) {
            zeroPad = computeDefaultPad(inShape, fieldSize, stride);
        }
        const inputRows = inShape[0];
        const inputCols = inShape[1];
        const outputRows = round$3((inputRows - fieldSize + 2 * zeroPad) / stride + 1, roundingMode);
        const outputCols = round$3((inputCols - fieldSize + 2 * zeroPad) / stride + 1, roundingMode);
        return [outputRows, outputCols];
    }
    function computeOutputShape4D(inShape, fieldSize, outChannels, stride, zeroPad, roundingMode) {
        if (zeroPad == null) {
            zeroPad = computeDefaultPad(inShape, fieldSize, stride);
        }
        const inputDepth = inShape[0];
        const inputRows = inShape[1];
        const inputCols = inShape[2];
        const outputDepths = round$3((inputDepth - fieldSize + 2 * zeroPad) / stride + 1, roundingMode);
        const outputRows = round$3((inputRows - fieldSize + 2 * zeroPad) / stride + 1, roundingMode);
        const outputCols = round$3((inputCols - fieldSize + 2 * zeroPad) / stride + 1, roundingMode);
        return [outputDepths, outputRows, outputCols, outChannels];
    }
    function computeDefaultPad(inputShape, fieldSize, stride, dilation = 1) {
        const effectiveFieldSize = getEffectiveFilterSize(fieldSize, dilation);
        return Math.floor((inputShape[0] * (stride - 1) - stride + effectiveFieldSize) / 2);
    }
    function parseTupleParam(param) {
        if (typeof param === 'number') {
            return [param, param, param];
        }
        if (param.length === 2) {
            return [param[0], param[1], 1];
        }
        return param;
    }
    function parse3TupleParam(param) {
        return typeof param === 'number' ? [param, param, param] : param;
    }
    /* See https://www.tensorflow.org/api_docs/python/tf/nn/atrous_conv2d
     * Atrous convolution is equivalent to standard convolution with upsampled
     * filters with effective_filter_height =
     * filter_height + (filter_height - 1) * (dilation - 1)
     * and effective_filter_width =
     * filter_width + (filter_width - 1) * (dilation - 1),
     * produced by inserting dilation - 1 zeros along consecutive elements across
     * the filters' spatial dimensions.
     * When there is a dilation, this converts a filter dimension to the
     * effective filter dimension, so it can be used in a standard convolution.
     */
    function getEffectiveFilterSize(filterSize, dilation) {
        if (dilation <= 1) {
            return filterSize;
        }
        return filterSize + (filterSize - 1) * (dilation - 1);
    }
    function getPadAndOutInfo(pad, inHeight, inWidth, strideHeight, strideWidth, filterHeight, filterWidth, roundingMode, dataFormat) {
        let padInfo;
        let outHeight;
        let outWidth;
        if (typeof pad === 'number') {
            const padType = (pad === 0) ? 'VALID' : 'NUMBER';
            padInfo = { top: pad, bottom: pad, left: pad, right: pad, type: padType };
            const outShape = computeOutputShape2D([inHeight, inWidth], filterHeight, strideHeight, pad, roundingMode);
            outHeight = outShape[0];
            outWidth = outShape[1];
        }
        else if (pad === 'same') {
            outHeight = Math.ceil(inHeight / strideHeight);
            outWidth = Math.ceil(inWidth / strideWidth);
            const padAlongHeight = Math.max(0, (outHeight - 1) * strideHeight + filterHeight - inHeight);
            const padAlongWidth = Math.max(0, (outWidth - 1) * strideWidth + filterWidth - inWidth);
            const top = Math.floor(padAlongHeight / 2);
            const bottom = padAlongHeight - top;
            const left = Math.floor(padAlongWidth / 2);
            const right = padAlongWidth - left;
            padInfo = { top, bottom, left, right, type: 'SAME' };
        }
        else if (pad === 'valid') {
            padInfo = { top: 0, bottom: 0, left: 0, right: 0, type: 'VALID' };
            outHeight = Math.ceil((inHeight - filterHeight + 1) / strideHeight);
            outWidth = Math.ceil((inWidth - filterWidth + 1) / strideWidth);
        }
        else if (typeof pad === 'object') {
            const top = dataFormat === 'channelsLast' ? pad[1][0] : pad[2][0];
            const bottom = dataFormat === 'channelsLast' ? pad[1][1] : pad[2][1];
            const left = dataFormat === 'channelsLast' ? pad[2][0] : pad[3][0];
            const right = dataFormat === 'channelsLast' ? pad[2][1] : pad[3][1];
            const padType = (top === 0 && bottom === 0 && left === 0 && right === 0) ?
                'VALID' :
                'EXPLICIT';
            padInfo = { top, bottom, left, right, type: padType };
            outHeight = round$3((inHeight - filterHeight + top + bottom) / strideHeight + 1, roundingMode);
            outWidth = round$3((inWidth - filterWidth + left + right) / strideWidth + 1, roundingMode);
        }
        else {
            throw Error(`Unknown padding parameter: ${pad}`);
        }
        return { padInfo, outHeight, outWidth };
    }
    function get3DPadAndOutInfo(pad, inDepth, inHeight, inWidth, strideDepth, strideHeight, strideWidth, filterDepth, filterHeight, filterWidth, roundingMode) {
        let padInfo;
        let outDepth;
        let outHeight;
        let outWidth;
        if (typeof pad === 'number') {
            const padType = (pad === 0) ? 'VALID' : 'NUMBER';
            padInfo = {
                top: pad,
                bottom: pad,
                left: pad,
                right: pad,
                front: pad,
                back: pad,
                type: padType
            };
            const outShape = computeOutputShape4D([inDepth, inHeight, inWidth, 1], filterDepth, 1, strideDepth, pad, roundingMode);
            outDepth = outShape[0];
            outHeight = outShape[1];
            outWidth = outShape[2];
        }
        else if (pad === 'same') {
            outDepth = Math.ceil(inDepth / strideDepth);
            outHeight = Math.ceil(inHeight / strideHeight);
            outWidth = Math.ceil(inWidth / strideWidth);
            const padAlongDepth = (outDepth - 1) * strideDepth + filterDepth - inDepth;
            const padAlongHeight = (outHeight - 1) * strideHeight + filterHeight - inHeight;
            const padAlongWidth = (outWidth - 1) * strideWidth + filterWidth - inWidth;
            const front = Math.floor(padAlongDepth / 2);
            const back = padAlongDepth - front;
            const top = Math.floor(padAlongHeight / 2);
            const bottom = padAlongHeight - top;
            const left = Math.floor(padAlongWidth / 2);
            const right = padAlongWidth - left;
            padInfo = { top, bottom, left, right, front, back, type: 'SAME' };
        }
        else if (pad === 'valid') {
            padInfo = {
                top: 0,
                bottom: 0,
                left: 0,
                right: 0,
                front: 0,
                back: 0,
                type: 'VALID'
            };
            outDepth = Math.ceil((inDepth - filterDepth + 1) / strideDepth);
            outHeight = Math.ceil((inHeight - filterHeight + 1) / strideHeight);
            outWidth = Math.ceil((inWidth - filterWidth + 1) / strideWidth);
        }
        else {
            throw Error(`Unknown padding parameter: ${pad}`);
        }
        return { padInfo, outDepth, outHeight, outWidth };
    }
    /**
     * Rounds a value depending on the rounding mode
     * @param value
     * @param roundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     */
    function round$3(value, roundingMode) {
        if (!roundingMode) {
            return Math.trunc(value);
        }
        switch (roundingMode) {
            case 'round':
                // used for Caffe Conv
                return Math.round(value);
            case 'ceil':
                // used for Caffe Pool
                return Math.ceil(value);
            case 'floor':
                return Math.floor(value);
            default:
                throw new Error(`Unknown roundingMode ${roundingMode}`);
        }
    }
    function tupleValuesAreOne(param) {
        const [dimA, dimB, dimC] = parseTupleParam(param);
        return dimA === 1 && dimB === 1 && dimC === 1;
    }
    function eitherStridesOrDilationsAreOne(strides, dilations) {
        return tupleValuesAreOne(strides) || tupleValuesAreOne(dilations);
    }
    /**
     * Convert Conv2D dataFormat from 'NHWC'|'NCHW' to
     *    'channelsLast'|'channelsFirst'
     * @param dataFormat in 'NHWC'|'NCHW' mode
     * @return dataFormat in 'channelsLast'|'channelsFirst' mode
     * @throws unknown dataFormat
     */
    function convertConv2DDataFormat(dataFormat) {
        if (dataFormat === 'NHWC') {
            return 'channelsLast';
        }
        else if (dataFormat === 'NCHW') {
            return 'channelsFirst';
        }
        else {
            throw new Error(`Unknown dataFormat ${dataFormat}`);
        }
    }
    /**
     * Check validity of pad when using dimRoundingMode.
     * @param opDesc A string of op description
     * @param pad The type of padding algorithm.
     *   - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *   - `valid` output will be smaller than input if filter is larger
     *       than 1x1.
     *   - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     * @throws unknown padding parameter
     */
    function checkPadOnDimRoundingMode(opDesc, pad, dimRoundingMode) {
        if (dimRoundingMode != null) {
            if (typeof pad === 'string') {
                throw Error(`Error in ${opDesc}: pad must be an integer when using ` +
                    `dimRoundingMode ${dimRoundingMode} but got pad ${pad}.`);
            }
            else if (typeof pad === 'number') {
                assert$4(isInt(pad), () => `Error in ${opDesc}: pad must be an integer when using ` +
                    `dimRoundingMode ${dimRoundingMode} but got pad ${pad}.`);
            }
            else if (typeof pad === 'object') {
                pad.forEach(p => {
                    p.forEach(v => {
                        assert$4(isInt(v), () => `Error in ${opDesc}: pad must be an integer when using ` +
                            `dimRoundingMode ${dimRoundingMode} but got pad ${v}.`);
                    });
                });
            }
            else {
                throw Error(`Error in ${opDesc}: Unknown padding parameter: ${pad}`);
            }
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Reshapes a `tf.Tensor` to a given shape.
     *
     * Given an input tensor, returns a new tensor with the same values as the
     * input tensor with shape `shape`.
     *
     * If one component of shape is the special value -1, the size of that
     * dimension is computed so that the total size remains constant. In
     * particular, a shape of [-1] flattens into 1-D. At most one component of
     * shape can be -1.
     *
     * If shape is 1-D or higher, then the operation returns a tensor with shape
     * shape filled with the values of tensor. In this case, the number of
     * elements implied by shape must be the same as the number of elements in
     * tensor.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     * x.reshape([2, 2]).print();
     * ```
     *
     * @param x The input tensor to be reshaped.
     * @param shape An array of integers defining the output tensor shape.
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function reshape_(x, shape) {
        const $x = convertToTensor(x, 'x', 'reshape', 'string_or_numeric');
        const inputs = { x: $x };
        const attrs = { shape };
        return ENGINE.runKernel(Reshape$1, inputs, attrs);
    }
    const reshape$3 = op({ reshape_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the 2D average pooling of an image.
     *
     * @param x The input tensor, of rank 4 or rank 3 of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is assumed.
     * @param filterSize The filter size: `[filterHeight, filterWidth]`. If
     *     `filterSize` is a single number, then `filterHeight == filterWidth`.
     * @param strides The strides of the pooling: `[strideHeight, strideWidth]`. If
     *     `strides` is a single number, then `strideHeight == strideWidth`.
     * @param pad The type of padding algorithm:
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *    - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *         https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     */
    function avgPool_(x, filterSize, strides, pad, dimRoundingMode) {
        const $x = convertToTensor(x, 'x', 'avgPool', 'float32');
        const dilations = 1;
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in avgPool: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            reshapedTo4D = true;
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
        }
        assert$4(x4D.rank === 4, () => `Error in avgPool: x must be rank 4 but got rank ${x4D.rank}.`);
        checkPadOnDimRoundingMode('avgPool', pad, dimRoundingMode);
        const inputs = { x: x4D };
        const attrs = { filterSize, strides, pad, dimRoundingMode };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        let res = ENGINE.runKernel(AvgPool, inputs, attrs);
        res = cast$3(res, $x.dtype);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const avgPool$2 = op({ avgPool_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the 3D average pooling.
     *
     * ```js
     * const x = tf.tensor5d([1, 2, 3, 4, 5, 6, 7, 8], [1, 2, 2, 2, 1]);
     * const result = tf.avgPool3d(x, 2, 1, 'valid');
     * result.print();
     * ```
     *
     * @param x The input tensor, of rank 5 or rank 4 of shape
     *     `[batch, depth, height, width, inChannels]`.
     * @param filterSize The filter size:
     *     `[filterDepth, filterHeight, filterWidth]`.
     *     If `filterSize` is a single number,
     *     then `filterDepth == filterHeight == filterWidth`.
     * @param strides The strides of the pooling:
     *     `[strideDepth, strideHeight, strideWidth]`.
     *     If `strides` is a single number,
     *     then `strideDepth == strideHeight == strideWidth`.
     * @param pad The type of padding algorithm.
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1*1x1.
     *    - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     * @param dataFormat An optional string from: "NDHWC", "NCDHW". Defaults to
     *     "NDHWC". Specify the data format of the input and output data. With the
     *     default format "NDHWC", the data is stored in the order of: [batch,
     *     depth, height, width, channels]. Only "NDHWC" is currently supported.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function avgPool3d_(x, filterSize, strides, pad, dimRoundingMode, dataFormat = 'NDHWC') {
        const $x = convertToTensor(x, 'x', 'avgPool3d', 'float32');
        let x5D = $x;
        let reshapedTo5D = false;
        if ($x.rank === 4) {
            reshapedTo5D = true;
            x5D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2], $x.shape[3]]);
        }
        assert$4(x5D.rank === 5, () => `Error in avgPool3d: x must be rank 5 but got rank ${x5D.rank}.`);
        assert$4(dataFormat === 'NDHWC', () => `Error in avgPool3d: Only NDHWC is currently supported, ` +
            `but got dataFormat of ${dataFormat}`);
        checkPadOnDimRoundingMode('avgPool3d', pad, dimRoundingMode);
        const inputs = { x: x5D };
        const attrs = { filterSize, strides, pad, dimRoundingMode, dataFormat };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        let res = ENGINE.runKernel(AvgPool3D, inputs, attrs);
        res = cast$3(res, x5D.dtype);
        if (reshapedTo5D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3], res.shape[4]]);
        }
        return res;
    }
    const avgPool3d$1 = op({ avgPool3d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Concatenates a list of `tf.Tensor`s along a given axis.
     *
     * The tensors ranks and types must match, and their sizes must match in all
     * dimensions except `axis`.
     *
     * Also available are stricter rank-specific methods that assert that
     * `tensors` are of the given rank:
     *   - `tf.concat1d`
     *   - `tf.concat2d`
     *   - `tf.concat3d`
     *   - `tf.concat4d`
     *
     * Except `tf.concat1d` (which does not have axis param), all methods have
     * same signature as this method.
     *
     * ```js
     * const a = tf.tensor1d([1, 2]);
     * const b = tf.tensor1d([3, 4]);
     * a.concat(b).print();  // or a.concat(b)
     * ```
     *
     * ```js
     * const a = tf.tensor1d([1, 2]);
     * const b = tf.tensor1d([3, 4]);
     * const c = tf.tensor1d([5, 6]);
     * tf.concat([a, b, c]).print();
     * ```
     *
     * ```js
     * const a = tf.tensor2d([[1, 2], [10, 20]]);
     * const b = tf.tensor2d([[3, 4], [30, 40]]);
     * const axis = 1;
     * tf.concat([a, b], axis).print();
     * ```
     * @param tensors A list of tensors to concatenate.
     * @param axis The axis to concate along. Defaults to 0 (the first dim).
     *
     * @doc {heading: 'Tensors', subheading: 'Slicing and Joining'}
     */
    function concat_(tensors, axis = 0) {
        assert$4(tensors.length >= 1, () => 'Pass at least one tensor to concat');
        const $tensors = convertToTensorArray(tensors, 'tensors', 'concat', 'string_or_numeric');
        if ($tensors[0].dtype === 'complex64') {
            $tensors.forEach(tensor => {
                if (tensor.dtype !== 'complex64') {
                    throw new Error(`Cannot concatenate complex64 tensors with a tensor
          with dtype ${tensor.dtype}. `);
                }
            });
        }
        if ($tensors.length === 1) {
            return clone($tensors[0]);
        }
        const inputs = $tensors;
        const attr = { axis };
        return ENGINE.runKernel(Concat, inputs, attr);
    }
    const concat$2 = op({ concat_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes sigmoid element-wise, `1 / (1 + exp(-x))`
     *
     * ```js
     * const x = tf.tensor1d([0, -1, 2, -3]);
     *
     * x.sigmoid().print();  // or tf.sigmoid(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function sigmoid_(x) {
        const $x = convertToTensor(x, 'x', 'sigmoid', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Sigmoid$1, inputs);
    }
    const sigmoid$2 = op({ sigmoid_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Extracts a slice from a `tf.Tensor` starting at coordinates `begin`
     * and is of size `size`.
     *
     * Also available are stricter rank-specific methods with the same signature
     * as this method that assert that `x` is of the given rank:
     *   - `tf.slice1d`
     *   - `tf.slice2d`
     *   - `tf.slice3d`
     *   - `tf.slice4d`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     *
     * x.slice([1], [2]).print();
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * x.slice([1, 0], [1, 2]).print();
     * ```
     * @param x The input `tf.Tensor` to slice from.
     * @param begin The coordinates to start the slice from. The length can be
     *     less than the rank of x - the rest of the axes will have implicit 0 as
     *     start. Can also be a single number, in which case it specifies the
     *     first axis.
     * @param size The size of the slice. The length can be less than the rank of
     *     x - the rest of the axes will have implicit -1. A value of -1 requests
     *     the rest of the dimensions in the axis. Can also be a single number,
     *     in which case it specifies the size of the first axis.
     *
     * @doc {heading: 'Tensors', subheading: 'Slicing and Joining'}
     */
    function slice_(x, begin, size) {
        const $x = convertToTensor(x, 'x', 'slice', 'string_or_numeric');
        if ($x.rank === 0) {
            throw new Error('Slicing scalar is not possible');
        }
        const inputs = { x: $x };
        const attrs = { begin, size };
        return ENGINE.runKernel(Slice, inputs, attrs);
    }
    const slice$2 = op({ slice_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes hyperbolic tangent of the input `tf.Tensor` element-wise: `tanh(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, 1, -1, 70]);
     *
     * x.tanh().print();  // or tf.tanh(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function tanh_(x) {
        const $x = convertToTensor(x, 'x', 'tanh', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Tanh$1, inputs);
    }
    const tanh$2 = op({ tanh_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the next state and output of a BasicLSTMCell.
     *
     * Returns `[newC, newH]`.
     *
     * Derived from tf.contrib.rnn.BasicLSTMCell.
     *
     * @param forgetBias Forget bias for the cell.
     * @param lstmKernel The weights for the cell.
     * @param lstmBias The bias for the cell.
     * @param data The input to the cell.
     * @param c Previous cell state.
     * @param h Previous cell output.
     *
     * @doc {heading: 'Operations', subheading: 'RNN'}
     */
    function basicLSTMCell_(forgetBias, lstmKernel, lstmBias, data, c, h) {
        const $forgetBias = convertToTensor(forgetBias, 'forgetBias', 'basicLSTMCell');
        const $lstmKernel = convertToTensor(lstmKernel, 'lstmKernel', 'basicLSTMCell');
        const $lstmBias = convertToTensor(lstmBias, 'lstmBias', 'basicLSTMCell');
        const $data = convertToTensor(data, 'data', 'basicLSTMCell');
        const $c = convertToTensor(c, 'c', 'basicLSTMCell');
        const $h = convertToTensor(h, 'h', 'basicLSTMCell');
        const combined = concat$2([$data, $h], 1);
        const weighted = matMul$1(combined, $lstmKernel);
        const res = add$2(weighted, $lstmBias);
        // i = input_gate, j = new_input, f = forget_gate, o = output_gate
        const batchSize = res.shape[0];
        const sliceCols = res.shape[1] / 4;
        const sliceSize = [batchSize, sliceCols];
        const i = slice$2(res, [0, 0], sliceSize);
        const j = slice$2(res, [0, sliceCols], sliceSize);
        const f = slice$2(res, [0, sliceCols * 2], sliceSize);
        const o = slice$2(res, [0, sliceCols * 3], sliceSize);
        const newC = add$2(mul(sigmoid$2(i), tanh$2(j)), mul($c, sigmoid$2(add$2($forgetBias, f))));
        const newH = mul(tanh$2(newC), sigmoid$2(o));
        return [newC, newH];
    }
    const basicLSTMCell = op({ basicLSTMCell_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * This operation reshapes the "batch" dimension 0 into `M + 1` dimensions of
     * shape `blockShape + [batch]`, interleaves these blocks back into the grid
     * defined by the spatial dimensions `[1, ..., M]`, to obtain a result with
     * the same rank as the input. The spatial dimensions of this intermediate
     * result are then optionally cropped according to `crops` to produce the
     * output. This is the reverse of `tf.spaceToBatchND`. See below for a precise
     * description.
     *
     * ```js
     * const x = tf.tensor4d([1, 2, 3, 4], [4, 1, 1, 1]);
     * const blockShape = [2, 2];
     * const crops = [[0, 0], [0, 0]];
     *
     * x.batchToSpaceND(blockShape, crops).print();
     * ```
     *
     * @param x A `tf.Tensor`. N-D with `x.shape` = `[batch] + spatialShape +
     * remainingShape`, where spatialShape has `M` dimensions.
     * @param blockShape A 1-D array. Must have shape `[M]`, all values must
     * be >= 1.
     * @param crops A 2-D array.  Must have shape `[M, 2]`, all values must be >= 0.
     * `crops[i] = [cropStart, cropEnd]` specifies the amount to crop from input
     * dimension `i + 1`, which corresponds to spatial dimension `i`. It is required
     * that `cropStart[i] + cropEnd[i] <= blockShape[i] * inputShape[i + 1]`
     *
     * This operation is equivalent to the following steps:
     *
     * 1. Reshape `x` to `reshaped` of shape: `[blockShape[0], ...,
     * blockShape[M-1], batch / prod(blockShape), x.shape[1], ...,
     * x.shape[N-1]]`
     *
     * 2. Permute dimensions of `reshaped`to produce `permuted` of shape `[batch /
     * prod(blockShape),x.shape[1], blockShape[0], ..., x.shape[M],
     * blockShape[M-1],x.shape[M+1], ..., x.shape[N-1]]`
     *
     * 3. Reshape `permuted` to produce `reshapedPermuted` of shape `[batch /
     * prod(blockShape),x.shape[1] * blockShape[0], ..., x.shape[M] *
     * blockShape[M-1],x.shape[M+1], ..., x.shape[N-1]]`
     *
     * 4. Crop the start and end of dimensions `[1, ..., M]` of `reshapedPermuted`
     * according to `crops` to produce the output of shape: `[batch /
     * prod(blockShape),x.shape[1] * blockShape[0] - crops[0,0] - crops[0,1],
     * ..., x.shape[M] * blockShape[M-1] - crops[M-1,0] -
     * crops[M-1,1],x.shape[M+1], ..., x.shape[N-1]]`
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function batchToSpaceND_(x, blockShape, crops) {
        const $x = convertToTensor(x, 'x', 'batchToSpaceND');
        const prod = blockShape.reduce((a, b) => a * b);
        assert$4($x.rank >= 1 + blockShape.length, () => `input rank is ${$x.rank} but should be > than blockShape.length ${blockShape.length}`);
        assert$4(crops.length === blockShape.length, () => `crops.length is ${crops.length} but should be equal to blockShape.length  ${blockShape.length}`);
        assert$4($x.shape[0] % prod === 0, () => `input tensor batch is ${$x.shape[0]} but is not divisible by the product of ` +
            `the elements of blockShape ${blockShape.join(' * ')} === ${prod}`);
        const inputs = { x: $x };
        const attrs = { blockShape, crops };
        return ENGINE.runKernel(BatchToSpaceND, inputs, attrs);
    }
    const batchToSpaceND$2 = op({ batchToSpaceND_ });

    function xAs4D(x) {
        let x4D;
        if (x.rank === 0 || x.rank === 1) {
            x4D = reshape$3(x, [1, 1, 1, x.size]);
        }
        else if (x.rank === 2) {
            x4D = reshape$3(x, [1, 1, x.shape[0], x.shape[1]]);
        }
        else if (x.rank === 3) {
            x4D = reshape$3(x, [1, x.shape[0], x.shape[1], x.shape[2]]);
        }
        else {
            x4D = x;
        }
        return x4D;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Batch normalization.
     *
     * As described in
     * [http://arxiv.org/abs/1502.03167](http://arxiv.org/abs/1502.03167).
     *
     * Mean, variance, scale, and offset can be of two shapes:
     *   - The same shape as the input.
     *   - In the common case, the depth dimension is the last dimension of x, so
     *     the values would be an `tf.Tensor1D` of shape [depth].
     *
     * Also available are stricter rank-specific methods with the same signature
     * as this method that assert that parameters passed are of given rank
     *   - `tf.batchNorm2d`
     *   - `tf.batchNorm3d`
     *   - `tf.batchNorm4d`
     *
     * @param x The input Tensor.
     * @param mean A mean Tensor.
     * @param variance A variance Tensor.
     * @param offset An offset Tensor.
     * @param scale A scale Tensor.
     * @param varianceEpsilon A small float number to avoid dividing by 0.
     *
     * @doc {heading: 'Operations', subheading: 'Normalization'}
     */
    function batchNorm_(x, mean, variance, offset, scale, varianceEpsilon) {
        if (varianceEpsilon == null) {
            varianceEpsilon = 0.001;
        }
        const $x = convertToTensor(x, 'x', 'batchNorm');
        const $mean = convertToTensor(mean, 'mean', 'batchNorm');
        const $variance = convertToTensor(variance, 'variance', 'batchNorm');
        let $scale;
        if (scale != null) {
            $scale = convertToTensor(scale, 'scale', 'batchNorm');
        }
        let $offset;
        if (offset != null) {
            $offset = convertToTensor(offset, 'offset', 'batchNorm');
        }
        assert$4($mean.rank === $variance.rank, () => 'Batch normalization gradient requires mean and variance to have ' +
            'equal ranks.');
        assert$4($offset == null || $mean.rank === $offset.rank, () => 'Batch normalization gradient requires mean and offset to have ' +
            'equal ranks.');
        assert$4($scale == null || $mean.rank === $scale.rank, () => 'Batch normalization gradient requires mean and scale to have ' +
            'equal ranks.');
        const x4D = xAs4D($x);
        const inputs = {
            x: x4D,
            scale: $scale,
            offset: $offset,
            mean: $mean,
            variance: $variance
        };
        const attrs = { varianceEpsilon };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(FusedBatchNorm, inputs, attrs);
        return reshape$3(res, $x.shape);
    }
    const batchNorm$2 = op({ batchNorm_ });

    /**
     * Batch normalization, strictly for 2D. For the more relaxed version, see
     * `tf.batchNorm`.
     *
     * @param x The input Tensor.
     * @param mean A mean Tensor.
     * @param variance A variance Tensor.
     * @param offset An offset Tensor.
     * @param scale A scale Tensor.
     * @param varianceEpsilon A small float number to avoid dividing by 0.
     */
    function batchNorm2d_(x, mean, variance, offset, scale, varianceEpsilon) {
        const $x = convertToTensor(x, 'x', 'batchNorm');
        const $mean = convertToTensor(mean, 'mean', 'batchNorm');
        const $variance = convertToTensor(variance, 'variance', 'batchNorm');
        let $scale;
        if (scale != null) {
            $scale = convertToTensor(scale, 'scale', 'batchNorm');
        }
        let $offset;
        if (offset != null) {
            $offset = convertToTensor(offset, 'offset', 'batchNorm');
        }
        assert$4($x.rank === 2, () => `Error in batchNorm2D: x must be rank 2 but got rank ` +
            `${$x.rank}.`);
        assert$4($mean.rank === 2 || $mean.rank === 1, () => `Error in batchNorm2D: mean must be rank 2 or rank 1 but ` +
            `got rank ${$mean.rank}.`);
        assert$4($variance.rank === 2 || $variance.rank === 1, () => `Error in batchNorm2D: variance must be rank 2 or rank 1 ` +
            `but got rank ${$variance.rank}.`);
        if ($scale != null) {
            assert$4($scale.rank === 2 || $scale.rank === 1, () => `Error in batchNorm2D: scale must be rank 2 or rank 1 ` +
                `but got rank ${$scale.rank}.`);
        }
        if ($offset != null) {
            assert$4($offset.rank === 2 || $offset.rank === 1, () => `Error in batchNorm2D: offset must be rank 2 or rank 1 ` +
                `but got rank ${$offset.rank}.`);
        }
        return batchNorm$2($x, $mean, $variance, $offset, $scale, varianceEpsilon);
    }
    const batchNorm2d = op({ batchNorm2d_ });

    /**
     * Batch normalization, strictly for 3D. For the more relaxed version, see
     * `tf.batchNorm`.
     *
     * @param x The input Tensor.
     * @param mean A mean Tensor.
     * @param variance A variance Tensor.
     * @param offset An offset Tensor.
     * @param scale A scale Tensor.
     * @param varianceEpsilon A small float number to avoid dividing by 0.
     */
    function batchNorm3d_(x, mean, variance, offset, scale, varianceEpsilon) {
        const $x = convertToTensor(x, 'x', 'batchNorm');
        const $mean = convertToTensor(mean, 'mean', 'batchNorm');
        const $variance = convertToTensor(variance, 'variance', 'batchNorm');
        let $scale;
        if (scale != null) {
            $scale = convertToTensor(scale, 'scale', 'batchNorm');
        }
        let $offset;
        if (offset != null) {
            $offset = convertToTensor(offset, 'offset', 'batchNorm');
        }
        assert$4($x.rank === 3, () => `Error in batchNorm3D: x must be rank 3 but got rank ` +
            `${$x.rank}.`);
        assert$4($mean.rank === 3 || $mean.rank === 1, () => `Error in batchNorm3D: mean must be rank 3 or rank 1 but ` +
            `got rank ${$mean.rank}.`);
        assert$4($variance.rank === 3 || $variance.rank === 1, () => `Error in batchNorm3D: variance must be rank 3 or rank 1 ` +
            `but got rank ${$variance.rank}.`);
        if ($scale != null) {
            assert$4($scale.rank === 3 || $scale.rank === 1, () => `Error in batchNorm3D: scale must be rank 3 or rank 1 ` +
                `but got rank ${$scale.rank}.`);
        }
        if ($offset != null) {
            assert$4($offset.rank === 3 || $offset.rank === 1, () => `Error in batchNorm3D: offset must be rank 3 or rank 1 ` +
                `but got rank ${$offset.rank}.`);
        }
        return batchNorm$2($x, $mean, $variance, $offset, $scale, varianceEpsilon);
    }
    const batchNorm3d = op({ batchNorm3d_ });

    /**
     * Batch normalization, strictly for 4D. For the more relaxed version, see
     * `tf.batchNorm`.
     *
     * @param x The input Tensor.
     * @param mean A mean Tensor.
     * @param variance A variance Tensor.
     * @param offset An offset Tensor.
     * @param scale A scale Tensor.
     * @param varianceEpsilon A small float number to avoid dividing by 0.
     */
    function batchNorm4d_(x, mean, variance, offset, scale, varianceEpsilon) {
        const $x = convertToTensor(x, 'x', 'batchNorm');
        const $mean = convertToTensor(mean, 'mean', 'batchNorm');
        const $variance = convertToTensor(variance, 'variance', 'batchNorm');
        let $scale;
        if (scale != null) {
            $scale = convertToTensor(scale, 'scale', 'batchNorm');
        }
        let $offset;
        if (offset != null) {
            $offset = convertToTensor(offset, 'offset', 'batchNorm');
        }
        assert$4($x.rank === 4, () => `Error in batchNorm4D: x must be rank 4 but got rank ` +
            `${$x.rank}.`);
        assert$4($mean.rank === 4 || $mean.rank === 1, () => `Error in batchNorm4D: mean must be rank 4 or rank 1 but ` +
            `got rank ${$mean.rank}.`);
        assert$4($variance.rank === 4 || $variance.rank === 1, () => `Error in batchNorm4D: variance must be rank 4 or rank 1 ` +
            `but got rank ${$variance.rank}.`);
        if ($scale != null) {
            assert$4($scale.rank === 4 || $scale.rank === 1, () => `Error in batchNorm4D: scale must be rank 4 or rank 1 ` +
                `but got rank ${$scale.rank}.`);
        }
        if ($offset != null) {
            assert$4($offset.rank === 4 || $offset.rank === 1, () => `Error in batchNorm4D: offset must be rank 4 or rank 1 ` +
                `but got rank ${$offset.rank}.`);
        }
        return batchNorm$2($x, $mean, $variance, $offset, $scale, varianceEpsilon);
    }
    const batchNorm4d = op({ batchNorm4d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Outputs a vector with length `size` and the same dtype as `weights`.
     *
     * If `weights` are empty, then index `i` stores the number of times the value
     * `i` is counted in `x`. If `weights` are non-empty, then index `i` stores the
     * sum of the value in `weights` at each index where the corresponding value in
     * `x` is `i`.
     *
     * Values in `x` outside of the range [0, size) are ignored.
     *
     * @param x The input int tensor, rank 1.
     * @param weights The weights tensor, must have the same shape as x, or a
     *     length-0 Tensor, in which case it acts as all weights equal to 1.
     * @param size Non-negative integer.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function bincount_(x, weights, size) {
        const $x = convertToTensor(x, 'x', 'bincount');
        const $weights = convertToTensor(weights, 'weights', 'bincount');
        assert$4($x.dtype === 'int32', () => `Error in bincount: input ` +
            `dtype must be int32, but got ${$x.dtype}`);
        assert$4(size >= 0, () => `size must be non-negative, but got ${size}.`);
        assert$4($weights.size === $x.size || $weights.size === 0, () => `Error in bincount: weights must have the same size as input or` +
            `0-length, but got input shape: ${$x.shape}, weights shape: ` +
            `${$weights.shape}.`);
        const inputs = { x: $x, weights: $weights };
        const attrs = { size };
        return ENGINE.runKernel(Bincount, inputs, attrs);
    }
    const bincount$2 = op({ bincount_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Return the shape of s0 op s1 with broadcast.
     *
     * compute r0, the broadcasted shape as a tensor.
     * s0, s1 and r0 are all integer vectors.
     *
     * This function returns the shape of the result of an operation between
     * two tensors of size s0 and s1 performed with broadcast.
     *
     * @param s0 A tensor representing a shape
     * @param s1 A tensor representing a shape
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function broadcastArgs_(s0, s1) {
        const shape1Input = convertToTensor(s0, 's0', 'broadcastArgs', 'int32');
        const shape2Input = convertToTensor(s1, 's1', 'broadcastArgs', 'int32');
        if (shape1Input.rank !== 1) {
            throw new Error('broadcastArgs(): first input must be a vector (rank=1). ' +
                `Has rank ${shape1Input.rank}`);
        }
        if (shape2Input.rank !== 1) {
            throw new Error('broadcastArgs(): second input must be a vector (rank=1). ' +
                `Has rank ${shape2Input.rank}`);
        }
        const inputs = { s0: shape1Input, s1: shape2Input };
        return ENGINE.runKernel(BroadcastArgs, inputs);
    }
    const broadcastArgs$2 = op({ broadcastArgs_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Broadcast an array to a compatible shape NumPy-style.
     *
     * The tensor's shape is compared to the broadcast shape from end to beginning.
     * Ones are prepended to the tensor's shape until is has the same length as
     * the broadcast shape. If input.shape[i]==shape[i], the (i+1)-th axis is
     * already broadcast-compatible. If input.shape[i]==1 and shape[i]==N, then
     * the input tensor is tiled N times along that axis (using tf.tile).
     *
     * @param input The tensor that is to be broadcasted.
     * @param shape The input is to be broadcast to this shape.
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function broadcastTo_(x, shape) {
        let input = convertToTensor(x, 'broadcastTo', 'x');
        const xShape = input.shape;
        if (shape.some(d => !(d > 0) || d % 1 !== 0)) {
            throw new Error(`broadcastTo(): Invalid broadcast shape [${shape}].`);
        }
        if (shape.length < input.rank) {
            throw new Error(`broadcastTo(): shape.length=${shape.length} < input.rank=${input.rank}.`);
        }
        if (shape.length > input.rank) {
            const newShape = input.shape.slice();
            while (newShape.length < shape.length) {
                newShape.unshift(1);
            }
            input = reshape$3(input, newShape);
        }
        const inputShape = input.shape;
        const reps = Array.from(shape);
        for (let i = shape.length - 1; i >= 0; i--) {
            if (inputShape[i] === shape[i]) {
                reps[i] = 1;
            }
            else if (input.shape[i] !== 1) {
                throw new Error(`broadcastTo(): [${xShape}] cannot be broadcast to [${shape}].`);
            }
        }
        const axes = reps.map((n, i) => n > 1 ? i : -1).filter(i => i >= 0);
        if (axes.length === 0) {
            return clone(input);
        }
        // TODO call broadcastTo kernel directly once backends implement broadcstTo
        const inputs = { x: input };
        const attrs = { reps };
        return ENGINE.runKernel(Tile, inputs, attrs);
    }
    const broadcastTo = op({ broadcastTo_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes ceiling of input `tf.Tensor` element-wise: `ceil(x)`
     *
     * ```js
     * const x = tf.tensor1d([.6, 1.1, -3.3]);
     *
     * x.ceil().print();  // or tf.ceil(x)
     * ```
     * @param x The input Tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function ceil_(x) {
        const $x = convertToTensor(x, 'x', 'ceil', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Ceil, inputs);
    }
    const ceil$2 = op({ ceil_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Clips values element-wise. `max(min(x, clipValueMax), clipValueMin)`
     *
     * ```js
     * const x = tf.tensor1d([-1, 2, -3, 4]);
     *
     * x.clipByValue(-2, 3).print();  // or tf.clipByValue(x, -2, 3)
     * ```
     * @param x The input tensor.
     * @param clipValueMin Lower-bound of range to be clipped to.
     * @param clipValueMax Upper-bound of range to be clipped to.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function clipByValue_(x, clipValueMin, clipValueMax) {
        const $x = convertToTensor(x, 'x', 'clipByValue');
        assert$4((clipValueMin <= clipValueMax), () => `Error in clip: min (${clipValueMin}) must be ` +
            `less than or equal to max (${clipValueMax}).`);
        const inputs = { x: $x };
        const attrs = { clipValueMin, clipValueMax };
        return ENGINE.runKernel(ClipByValue, inputs, attrs);
    }
    const clipByValue$2 = op({ clipByValue_ });

    /**
     * Concatenates a list of`tf.Tensor1D`s along an axis. See `concat` for details.
     *
     * For example, if:
     * A: shape(3) = |r1, g1, b1|
     * B: shape(2) = |r2, g2|
     * C = tf.concat1d([A, B]) == |r1, g1, b1, r2, g2|
     *
     * @param tensors A list of`tf.Tensor`s to concatenate.
     * @return The concatenated array.
     */
    function concat1d_(tensors) {
        return concat$2(tensors, 0 /* axis */);
    }
    const concat1d = op({ concat1d_ });

    /**
     * Concatenates a list of`tf.Tensor2D`s along an axis. See `concat` for details.
     *
     * For example, if:
     * A: shape(2, 3) = | r1, g1, b1 |
     *                  | r2, g2, b2 |
     *
     * B: shape(2, 3) = | r3, g3, b3 |
     *                  | r4, g4, b4 |
     *
     * C = tf.concat2d([A, B], axis)
     *
     * if axis = 0:
     * C: shape(4, 3) = | r1, g1, b1 |
     *                  | r2, g2, b2 |
     *                  | r3, g3, b3 |
     *                  | r4, g4, b4 |
     *
     * if axis = 1:
     * C = shape(2, 6) = | r1, g1, b1, r3, g3, b3 |
     *                   | r2, g2, b2, r4, g4, b4 |
     *
     *
     * @param tensors A list of `tf.Tensor`s to concatenate.
     * @param axis The axis to concatenate along.
     * @return The concatenated array.
     */
    function concat2d_(tensors, axis) {
        return concat$2(tensors, axis);
    }
    const concat2d = op({ concat2d_ });

    /**
     * Concatenates a list of `tf.Tensor3D`s along an axis.
     * See `concat` for details.
     *
     * For example, if:
     * A: shape(2, 1, 3) = | r1, g1, b1 |
     *                     | r2, g2, b2 |
     *
     * B: shape(2, 1, 3) = | r3, g3, b3 |
     *                     | r4, g4, b4 |
     *
     * C = tf.concat3d([A, B], axis)
     *
     * if axis = 0:
     * C: shape(4, 1, 3) = | r1, g1, b1 |
     *                     | r2, g2, b2 |
     *                     | r3, g3, b3 |
     *                     | r4, g4, b4 |
     *
     * if axis = 1:
     * C: shape(2, 2, 3) = | r1, g1, b1, r3, g3, b3 |
     *                     | r2, g2, b2, r4, g4, b4 |
     *
     * if axis = 2:
     * C = shape(2, 1, 6) = | r1, g1, b1, r3, g3, b3 |
     *                      | r2, g2, b2, r4, g4, b4 |
     *
     * @param tensors A list of`tf.Tensor`s to concatenate.
     * @param axis The axis to concate along.
     * @return The concatenated array.
     */
    function concat3d_(tensors, axis) {
        return concat$2(tensors, axis);
    }
    const concat3d = op({ concat3d_ });

    /**
     * Concatenates a list of `tf.Tensor4D`s along an axis.
     * See `concat` for details.
     *
     * @param tensors A list of `tf.Tensor`s to concatenate.
     * @param axis The axis to concate along.
     * @return The concatenated array.
     */
    function concat4d_(tensors, axis) {
        return concat$2(tensors, axis);
    }
    const concat4d = op({ concat4d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes a 2D convolution over the input x.
     *
     * @param x The input tensor, of rank 4 or rank 3, of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is
     * assumed.
     * @param filter The filter, rank 4, of shape
     *     `[filterHeight, filterWidth, inDepth, outDepth]`.
     * @param strides The strides of the convolution: `[strideHeight,
     * strideWidth]`.
     * @param pad The type of padding algorithm.
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *   - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dataFormat: An optional string from: "NHWC", "NCHW". Defaults to
     *     "NHWC". Specify the data format of the input and output data. With the
     *     default format "NHWC", the data is stored in the order of: [batch,
     *     height, width, channels].
     * @param dilations The dilation rates: `[dilationHeight, dilationWidth]`
     *     in which we sample input values across the height and width dimensions
     *     in atrous convolution. Defaults to `[1, 1]`. If `dilations` is a single
     *     number, then `dilationHeight == dilationWidth`. If it is greater than
     *     1, then all values of `strides` must be 1.
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function conv2d_(x, filter, strides, pad, dataFormat = 'NHWC', dilations = [1, 1], dimRoundingMode) {
        const $x = convertToTensor(x, 'x', 'conv2d', 'float32');
        const $filter = convertToTensor(filter, 'filter', 'conv2d', 'float32');
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            reshapedTo4D = true;
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
        }
        assert$4(x4D.rank === 4, () => `Error in conv2d: input must be rank 4, but got rank ${x4D.rank}.`);
        assert$4($filter.rank === 4, () => `Error in conv2d: filter must be rank 4, but got rank ` +
            `${$filter.rank}.`);
        checkPadOnDimRoundingMode('conv2d', pad, dimRoundingMode);
        const inDepth = dataFormat === 'NHWC' ? x4D.shape[3] : x4D.shape[1];
        assert$4(inDepth === $filter.shape[2], () => `Error in conv2d: depth of input (${inDepth}) must match ` +
            `input depth for filter ${$filter.shape[2]}.`);
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in conv2D: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        const inputs = { x: x4D, filter: $filter };
        const attrs = { strides, pad, dataFormat, dilations, dimRoundingMode };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(Conv2D$1, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const conv2d$3 = op({ conv2d_ });

    /**
     * Computes a 1D convolution over the input x.
     *
     * @param x The input tensor, of rank 3 or rank 2, of shape
     *     `[batch, width, inChannels]`. If rank 2, batch of 1 is assumed.
     * @param filter The filter, rank 3, of shape
     *     `[filterWidth, inDepth, outDepth]`.
     * @param stride The number of entries by which the filter is moved right at
     *     each step.
     * @param pad The type of padding algorithm.
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *   - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dataFormat An optional string from "NWC", "NCW". Defaults to "NWC",
     *     the data is stored in the order of [batch, in_width, in_channels]. Only
     *     "NWC" is currently supported.
     * @param dilation The dilation rate in which we sample input values in
     *     atrous convolution. Defaults to `1`. If it is greater than 1, then
     *     stride must be `1`.
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function conv1d_(x, filter, stride, pad, dataFormat = 'NWC', dilation = 1, dimRoundingMode) {
        const $x = convertToTensor(x, 'x', 'conv1d');
        const $filter = convertToTensor(filter, 'filter', 'conv1d');
        let x3D = $x;
        let reshapedTo3D = false;
        if ($x.rank === 2) {
            reshapedTo3D = true;
            x3D = reshape$3($x, [1, $x.shape[0], $x.shape[1]]);
        }
        assert$4(x3D.rank === 3, () => `Error in conv1d: input must be rank 3, but got rank ${x3D.rank}.`);
        assert$4($filter.rank === 3, () => `Error in conv1d: filter must be rank 3, but got rank ` +
            `${$filter.rank}.`);
        checkPadOnDimRoundingMode('conv1d', pad, dimRoundingMode);
        assert$4(x3D.shape[2] === $filter.shape[1], () => `Error in conv1d: depth of input (${x3D.shape[2]}) must match ` +
            `input depth for filter ${$filter.shape[1]}.`);
        assert$4(eitherStridesOrDilationsAreOne(stride, dilation), () => 'Error in conv1D: Either stride or dilation must be 1. ' +
            `Got stride ${stride} and dilation '${dilation}'`);
        assert$4(dataFormat === 'NWC', () => `Error in conv1d: got dataFormat of ${dataFormat} but only NWC is currently supported.`);
        const filter4D = reshape$3($filter, [1, $filter.shape[0], $filter.shape[1], $filter.shape[2]]);
        const input4D = reshape$3(x3D, [x3D.shape[0], 1, x3D.shape[1], x3D.shape[2]]);
        const strides = [1, stride];
        const dilations = [1, dilation];
        const conv2dDataFormat = 'NHWC';
        const res = conv2d$3(input4D, filter4D, strides, pad, conv2dDataFormat, dilations, dimRoundingMode);
        if (reshapedTo3D) {
            return reshape$3(res, [res.shape[2], res.shape[3]]);
        }
        return reshape$3(res, [res.shape[0], res.shape[2], res.shape[3]]);
    }
    const conv1d$1 = op({ conv1d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the derivative of the input of a 2D convolution.
     *
     * @param xShape The shape of the input: [batch, height, width, inDepth].
     * If length of 3, batch of 1 is assumed.
     * @param dy The derivative of the output, of rank 4 or rank 3 of shape
     *   `[batch, outHeight, outWidth, outDepth]`. If rank 3, batch of 1 is
     * assumed.
     * @param filter The filter, rank 4, of shape
     *     `[filterHeight, filterWidth, inDepth, outDepth]`.
     * @param strides The strides of the convolution: `[strideHeight,
     * strideWidth]`.
     * @param pad The type of padding algorithm used:
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     * @param dataFormat: An optional string from: "NHWC", "NCHW". Defaults to
     *     "NHWC". Specify the data format of the input and output data. With the
     *     default format "NHWC", the data is stored in the order of: [batch,
     *     height, width, channels].
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     */
    function conv2DBackpropInput_(xShape, dy, filter, strides, pad, dataFormat = 'NHWC', dimRoundingMode) {
        assert$4(xShape.length === dy.rank, () => `Length of inShape ` +
            `(${xShape.length}) and rank of dy (${dy.rank}) must match`);
        let xShape4D = xShape;
        let dy4D = dy;
        let reshapedTo4D = false;
        if (dy.rank === 3) {
            reshapedTo4D = true;
            dy4D = reshape$3(dy, [1, dy.shape[0], dy.shape[1], dy.shape[2]]);
            xShape4D = [1, xShape[0], xShape[1], xShape[2]];
        }
        assert$4(xShape4D.length === 4, () => `Error in conv2dDerInput: inShape must be length 4, but got length ` +
            `${xShape4D.length}.`);
        assert$4(dy4D.rank === 4, () => `Error in conv2dDerInput: dy must be rank 4, but got ` +
            `rank ${dy4D.rank}`);
        assert$4(filter.rank === 4, () => `Error in conv2dDerInput: filter must be rank 4, but got ` +
            `rank ${filter.rank}`);
        const inDepth = dataFormat === 'NHWC' ? xShape4D[3] : xShape4D[1];
        const outDepth = dataFormat === 'NHWC' ? dy4D.shape[3] : dy4D.shape[1];
        assert$4(inDepth === filter.shape[2], () => `Error in conv2dDerInput: depth of input (${inDepth}) must ` +
            `match input depth for filter ${filter.shape[2]}.`);
        assert$4(outDepth === filter.shape[3], () => `Error in conv2dDerInput: depth of output (${outDepth}) must ` +
            `match output depth for filter ${filter.shape[3]}.`);
        checkPadOnDimRoundingMode('conv2dDerInput', pad, dimRoundingMode);
        const inputs = { dy: dy4D, filter };
        const attrs = { strides, pad, dataFormat, dimRoundingMode, inputShape: xShape4D };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(Conv2DBackpropInput, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const conv2DBackpropInput$2 = op({ conv2DBackpropInput_ });

    /**
     * Computes the transposed 2D convolution of an image, also known as a
     * deconvolution.
     *
     * @param x The input image, of rank 4 or rank 3, of shape
     *   `[batch, height, width, inDepth]`. If rank 3, batch of 1 is assumed.
     * @param filter The filter, rank 4, of shape
     *     `[filterHeight, filterWidth, outDepth, inDepth]`.
     *     `inDepth` must match `inDepth` in `x`.
     * @param outputShape Output shape, of rank 4 or rank 3:
     *     `[batch, height, width, outDepth]`. If rank 3, batch of 1 is assumed.
     * @param strides The strides of the original convolution:
     *     `[strideHeight, strideWidth]`.
     * @param pad  The type of padding algorithm used in the non-transpose version
     *    of the op.
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function conv2dTranspose_(x, filter, outputShape, strides, pad, dimRoundingMode) {
        const $x = convertToTensor(x, 'x', 'conv2dTranspose');
        const $filter = convertToTensor(filter, 'filter', 'conv2dTranspose');
        return conv2DBackpropInput$2(outputShape, $x, $filter, strides, pad, 'NHWC', dimRoundingMode);
    }
    const conv2dTranspose$1 = op({ conv2dTranspose_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes a 3D convolution over the input x.
     *
     * @param x The input tensor, of rank 5 or rank 4, of shape
     *     `[batch, depth, height, width, channels]`. If rank 4,
     * batch of 1 is assumed.
     * @param filter The filter, rank 5, of shape
     *     `[filterDepth, filterHeight, filterWidth, inChannels, outChannels]`.
     *      inChannels must match between input and filter.
     * @param strides The strides of the convolution: `[strideDepth, strideHeight,
     * strideWidth]`.
     * @param pad The type of padding algorithm.
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *   - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dataFormat: An optional string from: "NDHWC", "NCDHW". Defaults to
     *     "NDHWC". Specify the data format of the input and output data. With the
     *     default format "NDHWC", the data is stored in the order of: [batch,
     *     depth, height, width, channels]. Only "NDHWC" is currently supported.
     * @param dilations The dilation rates: `[dilationDepth, dilationHeight,
     *     dilationWidth]` in which we sample input values across the height
     *     and width dimensions in atrous convolution. Defaults to `[1, 1, 1]`.
     *     If `dilations` is a single number, then
     *     `dilationDepth == dilationHeight == dilationWidth`. If it is greater
     *     than 1, then all values of `strides` must be 1.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function conv3d_(x, filter, strides, pad, dataFormat = 'NDHWC', dilations = [1, 1, 1]) {
        const $x = convertToTensor(x, 'x', 'conv3d');
        const $filter = convertToTensor(filter, 'filter', 'conv3d');
        let x5D = $x;
        let reshapedTo5D = false;
        if ($x.rank === 4) {
            reshapedTo5D = true;
            x5D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2], $x.shape[3]]);
        }
        assert$4(x5D.rank === 5, () => `Error in conv3d: input must be rank 5, but got rank ${x5D.rank}.`);
        assert$4($filter.rank === 5, () => `Error in conv3d: filter must be rank 5, but got rank ` +
            `${$filter.rank}.`);
        assert$4(x5D.shape[4] === $filter.shape[3], () => `Error in conv3d: depth of input (${x5D.shape[4]}) must match ` +
            `input depth for filter ${$filter.shape[3]}.`);
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in conv3D: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        assert$4(dataFormat === 'NDHWC', () => `Error in conv3d: got dataFormat of ${dataFormat} but only NDHWC is currently supported.`);
        const inputs = { x: x5D, filter: $filter };
        const attrs = { strides, pad, dataFormat, dilations };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(Conv3D$1, inputs, attrs);
        if (reshapedTo5D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3], res.shape[4]]);
        }
        return res;
    }
    const conv3d$1 = op({ conv3d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the derivative of the input of a 3D convolution.
     *
     * @param xShape The shape of the input: [batch, depth, height, width,
     * in_channels]. If length of 4, batch of 1 is assumed.
     * @param dy The derivative of the output, of rank 5 or rank 4 of shape
     *   `[batch, outDepth, outHeight, outWidth, in_channels]`.
     * If rank 4, batch of 1 is assumed.
     * @param filter The filter, rank 5, of shape
     *     `[filterDepth, filterHeight, filterWidth, inDepth, outDepth]`.
     * @param strides The strides of the convolution: `[strideDepth, strideHeight,
     * strideWidth]`.
     * @param pad The type of padding algorithm used:
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     */
    function conv3DBackpropInput_(xShape, dy, filter, strides, pad) {
        assert$4(xShape.length === dy.rank, () => `Length of inShape ` +
            `(${xShape.length}) and rank of dy (${dy.rank}) must match`);
        let xShape5D = xShape;
        let dy5D = dy;
        let reshapedTo5D = false;
        if (dy.rank === 4) {
            reshapedTo5D = true;
            dy5D = reshape$3(dy, [1, dy.shape[0], dy.shape[1], dy.shape[2], dy.shape[3]]);
            xShape5D = [1, xShape[0], xShape[1], xShape[2], xShape[3]];
        }
        const inDepth = xShape5D[4];
        const outDepth = dy5D.shape[4];
        assert$4(xShape5D.length === 5, () => `Error in conv3dDerInput: inShape must be length 5, but got length ` +
            `${xShape5D.length}.`);
        assert$4(dy5D.rank === 5, () => `Error in conv3dDerInput: dy must be rank 5, but got ` +
            `rank ${dy5D.rank}`);
        assert$4(filter.rank === 5, () => `Error in conv3dDerInput: filter must be rank 5, but got ` +
            `rank ${filter.rank}`);
        assert$4(inDepth === filter.shape[3], () => `Error in conv3dDerInput: depth of input (${inDepth}) must ` +
            `match input depth for filter ${filter.shape[3]}.`);
        assert$4(outDepth === filter.shape[4], () => `Error in conv3dDerInput: depth of output (${outDepth}) must ` +
            `match output depth for filter ${filter.shape[4]}.`);
        const inputs = { dy: dy5D, filter };
        const attrs = { pad, strides, inputShape: xShape5D };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(Conv3DBackpropInputV2, inputs, attrs);
        if (reshapedTo5D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3], res.shape[4]]);
        }
        return res;
    }
    const conv3DBackpropInput$1 = op({ conv3DBackpropInput_ });

    /**
     * Computes the transposed 3D convolution of a volume, also known as a
     * deconvolution.
     *
     * @param x The input image, of rank 5 or rank 4, of shape
     *   `[batch, depth, height, width, inDepth]`. If rank 4, batch of 1 is assumed.
     * @param filter The filter, rank 4, of shape
     *     `[depth, filterHeight, filterWidth, outDepth, inDepth]`.
     *     `inDepth` must match `inDepth` in `x`.
     * @param outputShape Output shape, of rank 5 or rank 4:
     *     `[batch, depth, height, width, outDepth]`. If rank 3, batch of 1 is
     *    assumed.
     * @param strides The strides of the original convolution:
     *     `[strideDepth, strideHeight, strideWidth]`.
     * @param pad  The type of padding algorithm used in the non-transpose version
     *    of the op.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function conv3dTranspose_(x, filter, outputShape, strides, pad) {
        const $x = convertToTensor(x, 'x', 'conv3dTranspose');
        const $filter = convertToTensor(filter, 'filter', 'conv3dTranspose');
        return conv3DBackpropInput$1(outputShape, $x, $filter, strides, pad);
    }
    const conv3dTranspose$1 = op({ conv3dTranspose_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes cos of the input `tf.Tensor` element-wise: `cos(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, Math.PI / 2, Math.PI * 3 / 4]);
     *
     * x.cos().print();  // or tf.cos(x)
     * ```
     * @param x The input tensor. Must be float32 type.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function cos_(x) {
        const $x = convertToTensor(x, 'x', 'cos', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Cos, inputs);
    }
    const cos$2 = op({ cos_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes hyperbolic cos of the input `tf.Tensor` element-wise: `cosh(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, 1, -1, .7]);
     *
     * x.cosh().print();  // or tf.cosh(x)
     * ```
     * @param x The input tensor. Must be float32 type.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function cosh_(x) {
        const $x = convertToTensor(x, 'x', 'cosh', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Cosh, inputs);
    }
    const cosh$2 = op({ cosh_ });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the 'License');
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an 'AS IS' BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the cumulative product of a `tf.Tensor` along `axis`.
     *
     * ```js
     * const x = tf.tensor([1, 2, 3, 4]);
     * x.cumprod().print();
     * ```
     * ```js
     * const x = tf.tensor([[1, 2], [3, 4]]);
     * x.cumprod().print();
     * ```
     *
     * @param x The input tensor to cumulatively multiply.
     * @param axis The axis along which to multiply. Optional. Defaults to 0.
     * @param exclusive Whether to perform exclusive cumulative product. Optional.
     *     Defaults to false. If set to true then the product of each tensor entry
     *     does not include its own value, but only the values previous to it
     *     along the specified axis.
     * @param reverse Whether to multiply in the opposite direction. Optional.
     *     Defaults to false.
     *
     * @doc {heading: 'Operations', subheading: 'Scan'}
     */
    function cumprod_(x, axis = 0, exclusive = false, reverse = false) {
        const $x = convertToTensor(x, 'x', 'cumprod');
        const inputs = { x: $x };
        const attrs = { axis, exclusive, reverse };
        return ENGINE.runKernel(Cumprod, inputs, attrs);
    }
    const cumprod$2 = op({ cumprod_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the cumulative sum of a `tf.Tensor` along `axis`.
     *
     * ```js
     * const x = tf.tensor([1, 2, 3, 4]);
     * x.cumsum().print();
     * ```
     * ```js
     * const x = tf.tensor([[1, 2], [3, 4]]);
     * x.cumsum().print();
     * ```
     *
     * @param x The input tensor to be summed.
     * @param axis The axis along which to sum. Optional. Defaults to 0.
     * @param exclusive Whether to perform exclusive cumulative sum. Optional.
     *     Defaults to false. If set to true then the sum of each tensor entry
     *     does not include its own value, but only the values previous to it
     *     along the specified axis.
     * @param reverse Whether to sum in the opposite direction. Optional.
     *     Defaults to false.
     *
     * @doc {heading: 'Operations', subheading: 'Scan'}
     */
    function cumsum_(x, axis = 0, exclusive = false, reverse = false) {
        const $x = convertToTensor(x, 'x', 'cumsum');
        const inputs = { x: $x };
        const attrs = { axis, exclusive, reverse };
        return ENGINE.runKernel(Cumsum, inputs, attrs);
    }
    const cumsum$2 = op({ cumsum_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Outputs a vector with length `size` and the same dtype as `weights`.
     *
     * If `weights` are empty, then index `i` stores the number of times the value
     * `i` is counted in `x`. If `weights` are non-empty, then index `i` stores the
     * sum of the value in `weights` at each index where the corresponding value in
     * `x` is `i`.
     *
     * Values in `x` outside of the range [0, size) are ignored.
     *
     * @param x The input int tensor, rank 1 or rank 2.
     * @param weights The weights tensor, must have the same shape as x, or a
     *     length-0 Tensor, in which case it acts as all weights equal to 1.
     * @param size Non-negative integer.
     * @param binaryOutput Optional. Whether the kernel should count the appearance
     *     or number of occurrences. Defaults to False.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function denseBincount_(x, weights, size, binaryOutput = false) {
        const $x = convertToTensor(x, 'x', 'denseBincount');
        const $weights = convertToTensor(weights, 'weights', 'denseBincount');
        assert$4($x.dtype === 'int32', () => `Error in denseBincount: input ` +
            `dtype must be int32, but got ${$x.dtype}`);
        assert$4($x.rank <= 2, () => `Error in denseBincount: input must be at most rank 2, but got ` +
            `rank ${$x.rank}.`);
        assert$4(size >= 0, () => `size must be non-negative, but got ${size}.`);
        assert$4($weights.size === $x.size || $weights.size === 0, () => `Error in denseBincount: weights must have the same shape as x or ` +
            `0-length, but got x shape: ${$x.shape}, weights shape: ` +
            `${$weights.shape}.`);
        const inputs = { x: $x, weights: $weights };
        const attrs = { size, binaryOutput };
        return ENGINE.runKernel(DenseBincount, inputs, attrs);
    }
    const denseBincount$2 = op({ denseBincount_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Rearranges data from depth into blocks of spatial data. More specifically,
     * this op outputs a copy of the input tensor where values from the `depth`
     * dimension are moved in spatial blocks to the `height` and `width` dimensions.
     * The attr `blockSize` indicates the input block size and how the data is
     * moved.
     *
     *  - Chunks of data of size `blockSize * blockSize` from depth are rearranged
     * into non-overlapping blocks of size `blockSize x blockSize`
     *
     *  - The width the output tensor is `inputWidth * blockSize`, whereas the
     * height is `inputHeight * blockSize`
     *
     *  - The Y, X coordinates within each block of the output image are determined
     * by the high order component of the input channel index
     *
     *  - The depth of the input tensor must be divisible by `blockSize *
     * blockSize`
     *
     * The `dataFormat` attr specifies the layout of the input and output tensors
     * with the following options: "NHWC": [ `batch, height, width, channels` ]
     * "NCHW": [ `batch, channels, height, width` ]
     *
     * ```js
     * const x = tf.tensor4d([1, 2, 3, 4], [1, 1, 1, 4]);
     * const blockSize = 2;
     * const dataFormat = "NHWC";
     *
     * tf.depthToSpace(x, blockSize, dataFormat).print();
     * ```
     *
     * @param x The input tensor of rank 4
     * @param blockSIze  An `int` that is `>= 2`. The size of the spatial block
     * @param dataFormat An optional string from: "NHWC", "NCHW". Defaults to "NHWC"
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function depthToSpace_(x, blockSize, dataFormat = 'NHWC') {
        const $x = convertToTensor(x, 'x', 'depthToSpace', 'float32');
        const inputHeight = (dataFormat === 'NHWC') ? $x.shape[1] : $x.shape[2];
        const inputWidth = (dataFormat === 'NHWC') ? $x.shape[2] : $x.shape[3];
        const inputDepth = (dataFormat === 'NHWC') ? $x.shape[3] : $x.shape[1];
        assert$4(blockSize > 1, () => `blockSize should be > 1 for depthToSpace, but was: ${blockSize}`);
        assert$4(inputHeight * blockSize >= 0, () => `Negative dimension size caused by overflow when multiplying
    ${inputHeight} and ${blockSize}  for depthToSpace with input shape
    ${$x.shape}`);
        assert$4(inputWidth * blockSize >= 0, () => `Negative dimension size caused by overflow when multiplying
    ${inputWidth} and ${blockSize} for depthToSpace with input shape
        ${$x.shape}`);
        assert$4((inputDepth % (blockSize * blockSize) === 0), () => `Dimension size must be evenly divisible by ${blockSize * blockSize} but is ${inputDepth} for depthToSpace with input shape ${$x.shape}`);
        const inputs = { x: $x };
        const attrs = { blockSize, dataFormat };
        return ENGINE.runKernel(DepthToSpace, inputs, attrs);
    }
    const depthToSpace$2 = op({ depthToSpace_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Depthwise 2D convolution.
     *
     * Given a 4D `input` array and a `filter` array of shape
     * `[filterHeight, filterWidth, inChannels, channelMultiplier]` containing
     * `inChannels` convolutional filters of depth 1, this op applies a
     * different filter to each input channel (expanding from 1 channel to
     * `channelMultiplier` channels for each), then concatenates the results
     * together. The output has `inChannels * channelMultiplier` channels.
     *
     * See
     * [https://www.tensorflow.org/api_docs/python/tf/nn/depthwise_conv2d](
     *     https://www.tensorflow.org/api_docs/python/tf/nn/depthwise_conv2d)
     * for more details.
     *
     * @param x The input tensor, of rank 4 or rank 3, of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is
     * assumed.
     * @param filter The filter tensor, rank 4, of shape
     *     `[filterHeight, filterWidth, inChannels, channelMultiplier]`.
     * @param strides The strides of the convolution: `[strideHeight,
     * strideWidth]`. If strides is a single number, then `strideHeight ==
     * strideWidth`.
     * @param pad The type of padding algorithm.
     *   - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *   - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *   - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dilations The dilation rates: `[dilationHeight, dilationWidth]`
     *     in which we sample input values across the height and width dimensions
     *     in atrous convolution. Defaults to `[1, 1]`. If `rate` is a single
     *     number, then `dilationHeight == dilationWidth`. If it is greater than
     *     1, then all values of `strides` must be 1.
     * @param dataFormat: An optional string from: "NHWC", "NCHW". Defaults to
     *     "NHWC". Specify the data format of the input and output data. With the
     *     default format "NHWC", the data is stored in the order of: [batch,
     *     height, width, channels]. Only "NHWC" is currently supported.
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function depthwiseConv2d_(x, filter, strides, pad, dataFormat = 'NHWC', dilations = [1, 1], dimRoundingMode) {
        const $x = convertToTensor(x, 'x', 'depthwiseConv2d', 'float32');
        const $filter = convertToTensor(filter, 'filter', 'depthwiseConv2d', 'float32');
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            reshapedTo4D = true;
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
        }
        assert$4(x4D.rank === 4, () => `Error in depthwiseConv2d: input must be rank 4, but got ` +
            `rank ${x4D.rank}.`);
        assert$4($filter.rank === 4, () => `Error in depthwiseConv2d: filter must be rank 4, but got rank ` +
            `${$filter.rank}.`);
        assert$4(x4D.shape[3] === $filter.shape[2], () => `Error in depthwiseConv2d: number of input channels ` +
            `(${x4D.shape[3]}) must match the inChannels dimension in ` +
            `filter ${$filter.shape[2]}.`);
        checkPadOnDimRoundingMode('depthwiseConv2d', pad, dimRoundingMode);
        const inputs = { x: x4D, filter: $filter };
        const attrs = { strides, pad, dataFormat, dilations, dimRoundingMode };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(DepthwiseConv2dNative, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const depthwiseConv2d$3 = op({ depthwiseConv2d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns a diagonal tensor with a given diagonal values.
     *
     * Given a diagonal, this operation returns a tensor with the diagonal and
     * everything else padded with zeros.
     *
     * Assume the input has dimensions `[D1,..., Dk]`, then the output is a tensor
     * of rank 2k with dimensions `[D1,..., Dk, D1,..., Dk]`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     *
     * tf.diag(x).print()
     * ```
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4, 5, 6, 6, 8], [4, 2])
     *
     * tf.diag(x).print()
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function diag_(x) {
        const $x = convertToTensor(x, 'x', 'diag');
        const inputs = { x: $x };
        return ENGINE.runKernel(Diag, inputs);
    }
    const diag$2 = op({ diag_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the grayscale dilation over the input `x`.
     *
     * @param x The input tensor, rank 3 or rank 4 of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is assumed.
     * @param filter The filter tensor, rank 3, of shape
     *     `[filterHeight, filterWidth, depth]`.
     * @param strides The strides of the sliding window for each dimension of the
     *     input tensor: `[strideHeight, strideWidth]`.
     *     If `strides` is a single number,
     *     then `strideHeight == strideWidth`.
     * @param pad The type of padding algorithm.
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1*1x1.
     *    - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dataFormat Specify the data format of the input and output data.
     *      Defaults to 'NHWC'. Only 'NHWC' is currently supported. With the
     *      default format "NHWC", the data is stored in the order of: [batch,
     *      height, width, channels].
     * @param dilations The dilation rates: `[dilationHeight, dilationWidth]`
     *     in which we sample input values across the height and width dimensions
     *     for atrous morphological dilation. Defaults to `[1, 1]`. If `dilations`
     *     is a single number, then `dilationHeight == dilationWidth`. If it is
     *     greater than 1, then all values of `strides` must be 1.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function dilation2d_(x, filter, strides, pad, dilations = [1, 1], dataFormat = 'NHWC') {
        const $x = convertToTensor(x, 'x', 'dilation2d');
        const $filter = convertToTensor(filter, 'filter', 'dilation2d');
        assert$4($x.rank === 3 || $x.rank === 4, () => `Error in dilation2d: input must be rank 3 or 4, but got rank ` +
            `${$x.rank}.`);
        assert$4($filter.rank === 3, () => `Error in dilation2d: filter must be rank 3, but got rank ` +
            `${$filter.rank}.`);
        assert$4(dataFormat === 'NHWC', () => `Error in dilation2d: Only NHWC is currently supported, ` +
            `but got dataFormat of ${dataFormat}`);
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
            reshapedTo4D = true;
        }
        const inputs = { x: x4D, filter: $filter };
        const attrs = { strides, pad, dilations };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(Dilation2D, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const dilation2d = op({ dilation2d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of (a == b) element-wise. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3]);
     * const b = tf.tensor1d([2, 2, 2]);
     *
     * a.equal(b).print();
     * ```
     *
     * @param a The first input tensor.
     * @param b The second input tensor. Must have the same dtype as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function equal_(a, b) {
        let $a = convertToTensor(a, 'a', 'equal', 'string_or_numeric');
        let $b = convertToTensor(b, 'b', 'equal', 'string_or_numeric');
        [$a, $b] = makeTypesMatch($a, $b);
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Equal, inputs);
    }
    const equal$3 = op({ equal_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the elements, either `a` or `b` depending on the `condition`.
     *
     * If the condition is true, select from `a`, otherwise select from `b`.
     *
     * ```js
     * const cond = tf.tensor1d([false, false, true], 'bool');
     * const a = tf.tensor1d([1 , 2, 3]);
     * const b = tf.tensor1d([-1, -2, -3]);
     *
     * a.where(cond, b).print();
     * ```
     *
     * @param condition The input condition. Must be of dtype bool.
     * @param a If `condition` is rank 1, `a` may have a higher rank but
     *     its first dimension must match the size of `condition`.
     * @param b A tensor with the same dtype as `a` and with shape that is
     *     compatible with `a`.
     * @return A tensor with same dtype as `a` and `b`, and shape that is
     *     broadcastable from `a` and `b`.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function where_(condition, a, b) {
        const $a = convertToTensor(a, 'a', 'where');
        const $b = convertToTensor(b, 'b', 'where');
        const $condition = convertToTensor(condition, 'condition', 'where', 'bool');
        // TODO: move this logic to forward function when the broadcastTo op is
        // implemented in WASM.
        // Find the broadcastable shape for $condition, $a, and $b.
        const broadcastShape = assertAndGetBroadcastShape(assertAndGetBroadcastShape($condition.shape, $a.shape), $b.shape);
        const $broadcastedCondition = broadcastTo($condition, broadcastShape);
        const $broadcastedA = broadcastTo($a, broadcastShape);
        const $broadcastedB = broadcastTo($b, broadcastShape);
        const inputs = {
            condition: $broadcastedCondition,
            t: $broadcastedA,
            e: $broadcastedB
        };
        return ENGINE.runKernel(Select, inputs);
    }
    const where = op({ where_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with all elements set to 0 with the same shape as the
     * given tensor.
     *
     * ```js
     * const x = tf.tensor([1, 2]);
     * tf.zerosLike(x).print();
     * ```
     *
     * @param x The tensor of required shape.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function zerosLike_(x) {
        const $x = convertToTensor(x, 'x', 'zerosLike');
        const inputs = { x: $x };
        return ENGINE.runKernel(ZerosLike, inputs);
    }
    const zerosLike$2 = op({ zerosLike_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Divides two `tf.Tensor`s element-wise, A / B. Supports broadcasting. Return 0
     * if denominator is 0.
     *
     *
     * ```js
     * const a = tf.tensor1d([1, 4, 9, 16]);
     * const b = tf.tensor1d([1, 2, 3, 4]);
     * const c = tf.tensor1d([0, 0, 0, 0]);
     *
     * a.divNoNan(b).print();  // or tf.divNoNan(a, b)
     * a.divNoNan(c).print();  // or tf.divNoNan(a, c)
     * ```
     *
     * ```js
     * // Broadcast div a with b.
     * const a = tf.tensor1d([2, 4, 6, 8]);
     * const b = tf.scalar(2);
     * const c = tf.scalar(0);
     *
     * a.divNoNan(b).print();  // or tf.divNoNan(a, b)
     * a.divNoNan(c).print();  // or tf.divNoNan(a, c)
     * ```
     *
     * @param a The first tensor as the numerator.
     * @param b The second tensor as the denominator. Must have the same dtype as
     * `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function divNoNan_(a, b) {
        // TODO: Make this into its own kernel.
        let $a = convertToTensor(a, 'a', 'div');
        let $b = convertToTensor(b, 'b', 'div');
        [$a, $b] = makeTypesMatch($a, $b);
        const divResult = div$1($a, $b);
        const zeros = zerosLike$2(divResult);
        const bEqualsZero = equal$3($b, zeros);
        return where(bEqualsZero, zeros, divResult);
    }
    const divNoNan = op({ divNoNan_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the dot product of two matrices and/or vectors, `t1` and `t2`.
     *
     * ```js
     * const a = tf.tensor1d([1, 2]);
     * const b = tf.tensor2d([[1, 2], [3, 4]]);
     * const c = tf.tensor2d([[1, 2, 3], [4, 5, 6]]);
     *
     * a.dot(b).print();  // or tf.dot(a, b)
     * b.dot(a).print();
     * b.dot(c).print();
     * ```
     * @param t1 The first tensor in the dot operation.
     * @param t2 The second tensor in the dot operation.
     *
     * @doc {heading: 'Operations', subheading: 'Matrices'}
     */
    function dot_(t1, t2) {
        const $t1 = convertToTensor(t1, 't1', 'dot');
        const $t2 = convertToTensor(t2, 't2', 'dot');
        assert$4(($t1.rank === 1 || $t1.rank === 2) && ($t2.rank === 1 || $t2.rank === 2), () => `Error in dot: inputs must all be rank 1 or 2, but got ranks ` +
            `${$t1.rank} and ${$t2.rank}.`);
        const t1Inner = ($t1.rank === 1 ? $t1.size : $t1.shape[1]);
        const t2Inner = ($t2.rank === 1 ? $t2.size : $t2.shape[0]);
        assert$4(t1Inner === t2Inner, () => `Error in dot: inner dimensions of inputs must match, but got ` +
            `${t1Inner} and ${t2Inner}.`);
        if ($t1.rank === 1 && $t2.rank === 1) {
            const t12D = reshape$3($t1, [1, -1]);
            const t22D = reshape$3($t2, [-1, 1]);
            const t1t2 = matMul$1(t12D, t22D);
            return reshape$3(t1t2, []);
        }
        else if ($t1.rank === 1 && $t2.rank === 2) {
            const t12D = reshape$3($t1, [1, -1]);
            const t22D = reshape$3($t2, [$t2.shape[0], $t2.shape[1]]);
            const t1t2 = matMul$1(t12D, t22D);
            return reshape$3(t1t2, [t1t2.size]);
        }
        else if ($t1.rank === 2 && $t2.rank === 1) {
            const t22D = reshape$3($t2, [-1, 1]);
            const t1t2 = matMul$1($t1, t22D);
            return reshape$3(t1t2, [t1t2.size]);
        }
        else {
            const t22D = reshape$3($t2, [$t2.shape[0], $t2.shape[1]]);
            const t1t2 = matMul$1($t1, t22D);
            return t1t2;
        }
    }
    const dot$2 = op({ dot_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Tensor contraction over specified indices and outer product.
     *
     * `einsum` allows defining Tensors by defining their element-wise computation.
     * This computation is based on
     * [Einstein summation](https://en.wikipedia.org/wiki/Einstein_notation).
     *
     * Some special cases include:
     *
     * Matrix multiplication:
     * ```js
     * const x = tf.tensor2d([[1, 2, 3], [4, 5, 6]]);
     * const y = tf.tensor2d([[0, 1], [2, 3], [4, 5]]);
     * x.print();
     * y.print();
     * tf.einsum('ij,jk->ik', x, y).print();
     * ```
     *
     * Dot product:
     * ```js
     * const x = tf.tensor1d([1, 2, 3]);
     * const y = tf.tensor1d([0, 1, 2]);
     * x.print();
     * y.print();
     * tf.einsum('i,i->', x, y).print();
     * ```
     *
     * Batch dot product:
     * ```js
     * const x = tf.tensor2d([[1, 2, 3], [4, 5, 6]]);
     * const y = tf.tensor2d([[0, 1, 2], [3, 4, 5]]);
     * x.print();
     * y.print();
     * tf.einsum('bi,bi->b', x, y).print();
     * ```
     *
     * Outer prouduct:
     * ```js
     * const x = tf.tensor1d([1, 3, 5]);
     * const y = tf.tensor1d([2, 4, 6]);
     * x.print();
     * y.print();
     * tf.einsum('i,j->ij', x, y).print();
     * ```
     *
     * Matrix transpose:
     * ```js
     * const x = tf.tensor2d([[1, 2], [3, 4]]);
     * x.print();
     * tf.einsum('ij->ji', x).print();
     * ```
     *
     * Batch matrix transpose:
     * ```js
     * const x = tf.tensor3d([[[1, 2], [3, 4]], [[-1, -2], [-3, -4]]]);
     * x.print();
     * tf.einsum('bij->bji', x).print();
     * ```
     *
     * Limitations:
     *
     * This implementation of einsum has the following limitations:
     *
     * - Does not support >2 input tensors.
     * - Does not support duplicate axes for any given input tensor. E.g., equation
     *   'ii->' is not suppoted.
     * - The `...` notation is not supported.
     *
     * @param equation a string describing the contraction, in the same format as
     * [numpy.einsum](https://numpy.org/doc/stable/reference/generated/numpy.einsum.html).
     * @param tensors the input(s) to contract (each one a Tensor), whose shapes
     *     should be consistent with equation.
     * @returns The output tensor.
     *
     * @doc {heading: 'Tensors', subheading: 'Matrices'}
     */
    function einsum_(equation, ...tensors) {
        const $tensors = tensors.map((t, i) => convertToTensor(t, `tensors${i}`, 'einsum'));
        const attrs = { equation };
        return ENGINE.runKernel(Einsum, $tensors, attrs);
    }
    const einsum$2 = op({ einsum_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes exponential linear element-wise: `x > 0 ? x : (e ^ x) - 1`.
     *
     * ```js
     * const x = tf.tensor1d([-1, 1, -3, 2]);
     *
     * x.elu().print();  // or tf.elu(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function elu_(x) {
        const $x = convertToTensor(x, 'x', 'elu', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Elu$1, inputs);
    }
    const elu$4 = op({ elu_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes gause error function of the input `tf.Tensor` element-wise:
     * `erf(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, .1, -.1, .7]);
     *
     * x.erf().print(); // or tf.erf(x);
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function erf_(x) {
        let $x = convertToTensor(x, 'x', 'erf');
        assert$4($x.dtype === 'int32' || $x.dtype === 'float32', () => 'Input dtype must be `int32` or `float32`.');
        if ($x.dtype === 'int32') {
            $x = cast$3($x, 'float32');
        }
        const inputs = { x: $x };
        return ENGINE.runKernel(Erf, inputs);
    }
    const erf$2 = op({ erf_ });

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns true if the axis specifies the inner most dimensions of the
     * array.
     */
    function axesAreInnerMostDims(axes, rank) {
        for (let i = 0; i < axes.length; ++i) {
            if (axes[axes.length - i - 1] !== rank - 1 - i) {
                return false;
            }
        }
        return true;
    }
    function combineLocations(outputLoc, reduceLoc, axes) {
        const rank = outputLoc.length + reduceLoc.length;
        const loc = [];
        let outIdx = 0;
        let reduceIdx = 0;
        for (let dim = 0; dim < rank; dim++) {
            if (axes.indexOf(dim) === -1) {
                loc.push(outputLoc[outIdx++]);
            }
            else {
                loc.push(reduceLoc[reduceIdx++]);
            }
        }
        return loc;
    }
    function computeOutAndReduceShapes(aShape, axes) {
        const outShape = [];
        const rank = aShape.length;
        for (let dim = 0; dim < rank; dim++) {
            if (axes.indexOf(dim) === -1) {
                outShape.push(aShape[dim]);
            }
        }
        const reduceShape = axes.map(dim => aShape[dim]);
        return [outShape, reduceShape];
    }
    function expandShapeToKeepDim(shape, axes) {
        const reduceSubShape = axes.map(x => 1);
        return combineLocations(shape, reduceSubShape, axes);
    }
    function assertAxesAreInnerMostDims(msg, axes, rank) {
        assert$4(axesAreInnerMostDims(axes, rank), () => `${msg} supports only inner-most axes for now. ` +
            `Got axes ${axes} and rank-${rank} input.`);
    }
    /**
     * Returns the axes permutation to be used with `tf.transpose`, if such
     * permutation is necessary. Otherwise it returns null. This method is used by
     * operations that operate only on inner-most axes.
     */
    function getAxesPermutation(axes, rank) {
        if (axesAreInnerMostDims(axes, rank)) {
            return null;
        }
        const result = [];
        for (let i = 0; i < rank; ++i) {
            if (axes.indexOf(i) === -1) {
                result.push(i);
            }
        }
        axes.forEach(axis => result.push(axis));
        return result;
    }
    /** Returns the axes permutation that undoes the original permutation. */
    function getUndoAxesPermutation(axes) {
        return axes.map((axis, i) => [i, axis])
            .sort((a, b) => a[1] - b[1])
            .map(x => x[0]);
    }
    function getInnerMostAxes(numAxes, rank) {
        const res = [];
        for (let i = rank - numAxes; i < rank; ++i) {
            res.push(i);
        }
        return res;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the maximum of elements across dimensions of a `tf.Tensor`.
     *
     * Reduces the input along the dimensions given in `axes`. Unless `keepDims`
     * is true, the rank of the `tf.Tensor` is reduced by 1 for each entry in
     * `axes`. If `keepDims` is true, the reduced dimensions are retained with
     * length 1. If `axes` has no entries, all dimensions are reduced, and an
     * `tf.Tensor` with a single element is returned.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3]);
     *
     * x.max().print();  // or tf.max(x)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * const axis = 1;
     * x.max(axis).print();  // or tf.max(x, axis)
     * ```
     *
     * @param x The input tensor.
     * @param axis The dimension(s) to reduce. By default it reduces
     *     all dimensions.
     * @param keepDims If true, retains reduced dimensions with size 1.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function max_(x, axis = null, keepDims = false) {
        const $x = convertToTensor(x, 'x', 'max');
        const inputs = { x: $x };
        const attrs = { reductionIndices: axis, keepDims };
        return ENGINE.runKernel(Max, inputs, attrs);
    }
    const max$3 = op({ max_ });

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the minimum value from the input.
     *
     * Reduces the input along the dimensions given in `axes`. Unless `keepDims`
     * is true, the rank of the array is reduced by 1 for each entry in `axes`.
     * If `keepDims` is true, the reduced dimensions are retained with length 1.
     * If `axes` has no entries, all dimensions are reduced, and an array with a
     * single element is returned.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3]);
     *
     * x.min().print();  // or tf.min(x)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * const axis = 1;
     * x.min(axis).print();  // or tf.min(x, axis)
     * ```
     *
     * @param x The input Tensor.
     * @param axis The dimension(s) to reduce. By default it reduces
     *     all dimensions.
     * @param keepDims If true, retains reduced dimensions with size 1.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function min_(x, axis = null, keepDims = false) {
        const $x = convertToTensor(x, 'x', 'min');
        const inputs = { x: $x };
        const attrs = { axis, keepDims };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        return ENGINE.runKernel(Min, inputs, attrs);
    }
    const min$3 = op({ min_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the power of one `tf.Tensor` to another. Supports broadcasting.
     *
     * Given a `tf.Tensor` x and a `tf.Tensor` y, this operation computes x^y for
     * corresponding elements in x and y. The result's dtype will be the upcasted
     * type of the `base` and `exp` dtypes.
     *
     * ```js
     * const a = tf.tensor([[2, 3], [4, 5]])
     * const b = tf.tensor([[1, 2], [3, 0]]).toInt();
     *
     * a.pow(b).print();  // or tf.pow(a, b)
     * ```
     *
     * ```js
     * const a = tf.tensor([[1, 2], [3, 4]])
     * const b = tf.tensor(2).toInt();
     *
     * a.pow(b).print();  // or tf.pow(a, b)
     * ```
     * We also expose `powStrict` which has the same signature as this op and
     * asserts that `base` and `exp` are the same shape (does not broadcast).
     *
     * @param base The base `tf.Tensor` to pow element-wise.
     * @param exp The exponent `tf.Tensor` to pow element-wise.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function pow_(base, exp) {
        let $base = convertToTensor(base, 'base', 'pow');
        let $exp = convertToTensor(exp, 'exp', 'pow');
        [$base, $exp] = makeTypesMatch($base, $exp);
        const inputs = { a: $base, b: $exp };
        return ENGINE.runKernel(Pow, inputs);
    }
    const pow$2 = op({ pow_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates rank-0 `tf.Tensor` (scalar) with the provided value and dtype.
     *
     * The same functionality can be achieved with `tf.tensor`, but in general
     * we recommend using `tf.scalar` as it makes the code more readable.
     *
     * ```js
     * tf.scalar(3.14).print();
     * ```
     *
     * @param value The value of the scalar.
     * @param dtype The data type.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function scalar(value, dtype) {
        if (((isTypedArray$1(value) && dtype !== 'string') || Array.isArray(value)) &&
            dtype !== 'complex64') {
            throw new Error('Error creating a new Scalar: value must be a primitive ' +
                '(number|boolean|string)');
        }
        if (dtype === 'string' && isTypedArray$1(value) &&
            !(value instanceof Uint8Array)) {
            throw new Error('When making a scalar from encoded string, ' +
                'the value must be `Uint8Array`.');
        }
        const shape = [];
        const inferredShape = [];
        return makeTensor(value, shape, inferredShape, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes square root of the input `tf.Tensor` element-wise: `y = sqrt(x)`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 4, -1]);
     *
     * x.sqrt().print();  // or tf.sqrt(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function sqrt_(x) {
        const $x = convertToTensor(x, 'x', 'sqrt', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Sqrt, inputs);
    }
    const sqrt$2 = op({ sqrt_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes square of `x` element-wise: `x ^ 2`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, Math.sqrt(2), -1]);
     *
     * x.square().print();  // or tf.square(x)
     * ```
     * @param x The input Tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function square_(x) {
        const $x = convertToTensor(x, 'x', 'square');
        const attrs = {};
        return ENGINE.runKernel('Square', { x: $x }, attrs);
    }
    const square$2 = op({ square_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the sum of elements across dimensions of a `tf.Tensor`.
     *
     * Reduces the input along the dimensions given in `axes`. Unless `keepDims`
     * is true, the rank of the `tf.Tensor` is reduced by 1 for each entry in
     * `axes`. If `keepDims` is true, the reduced dimensions are retained with
     * length 1. If axes has no entries, all dimensions are reduced, and a
     * `tf.Tensor` with a single element is returned.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3]);
     *
     * x.sum().print();  // or tf.sum(x)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * const axis = 1;
     * x.sum(axis).print();  // or tf.sum(x, axis)
     * ```
     *
     * @param x The input tensor to compute the sum over. If the dtype is `bool`
     *   it will be converted to `int32` and the output dtype will be `int32`.
     * @param axis The dimension(s) to reduce. By default it reduces
     *     all dimensions.
     * @param keepDims If true, retains reduced dimensions with size 1.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function sum_(x, axis = null, keepDims = false) {
        let $x = convertToTensor(x, 'x', 'sum');
        if ($x.dtype === 'bool') {
            $x = cast$3($x, 'int32');
        }
        const inputs = { x: $x };
        const attrs = { axis, keepDims };
        return ENGINE.runKernel(Sum, inputs, attrs);
    }
    const sum$2 = op({ sum_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the norm of scalar, vectors, and matrices.
     * This function can compute several different vector norms (the 1-norm, the
     * Euclidean or 2-norm, the inf-norm, and in general the p-norm for p > 0)
     * and matrix norms (Frobenius, 1-norm, and inf-norm).
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     *
     * x.norm().print();  // or tf.norm(x)
     * ```
     *
     * @param x The input array.
     * @param ord Optional. Order of the norm. Supported norm types are
     * following:
     *
     *  | ord        | norm for matrices         | norm for vectors
     *  |------------|---------------------------|---------------------
     *  |'euclidean' |Frobenius norm             |2-norm
     *  |'fro'       |Frobenius norm	           |
     *  |Infinity    |max(sum(abs(x), axis=1))   |max(abs(x))
     *  |-Infinity   |min(sum(abs(x), axis=1))   |min(abs(x))
     *  |1           |max(sum(abs(x), axis=0))   |sum(abs(x))
     *  |2           |                           |sum(abs(x)^2)^1/2*
     *
     * @param axis Optional. If axis is null (the default), the input is
     * considered a vector and a single vector norm is computed over the entire
     * set of values in the Tensor, i.e. norm(x, ord) is equivalent
     * to norm(x.reshape([-1]), ord). If axis is a integer, the input
     * is considered a batch of vectors, and axis determines the axis in x
     * over which to compute vector norms. If axis is a 2-tuple of integer it is
     * considered a batch of matrices and axis determines the axes in NDArray
     * over which to compute a matrix norm.
     * @param keepDims Optional. If true, the norm have the same dimensionality
     * as the input.
     *
     * @doc {heading: 'Operations', subheading: 'Matrices'}
     */
    function norm_(x, ord = 'euclidean', axis = null, keepDims = false) {
        x = convertToTensor(x, 'x', 'norm');
        const norm = normImpl(x, ord, axis);
        let keepDimsShape = norm.shape;
        if (keepDims) {
            const axes = parseAxisParam(axis, x.shape);
            keepDimsShape = expandShapeToKeepDim(norm.shape, axes);
        }
        return reshape$3(norm, keepDimsShape);
    }
    function normImpl(x, p, axis = null) {
        if (x.rank === 0) {
            return abs$2(x);
        }
        // consider vector when no axis is specified
        if (x.rank !== 1 && axis === null) {
            return normImpl(reshape$3(x, [-1]), p, axis);
        }
        // vector
        if (x.rank === 1 || typeof axis === 'number' ||
            Array.isArray(axis) && axis.length === 1) {
            if (p === 1) {
                return sum$2(abs$2(x), axis);
            }
            if (p === Infinity) {
                return max$3(abs$2(x), axis);
            }
            if (p === -Infinity) {
                return min$3(abs$2(x), axis);
            }
            if (p === 'euclidean' || p === 2) {
                // norm(x, 2) = sum(abs(xi) ^ 2) ^ 1/2
                return sqrt$2(sum$2(pow$2(abs$2(x), scalar(2, 'int32')), axis));
            }
            throw new Error(`Error in norm: invalid ord value: ${p}`);
        }
        // matrix (assumption axis[0] < axis[1])
        if (Array.isArray(axis) && axis.length === 2) {
            if (p === 1) {
                return max$3(sum$2(abs$2(x), axis[0]), axis[1] - 1);
            }
            if (p === Infinity) {
                return max$3(sum$2(abs$2(x), axis[1]), axis[0]);
            }
            if (p === -Infinity) {
                return min$3(sum$2(abs$2(x), axis[1]), axis[0]);
            }
            if (p === 'fro' || p === 'euclidean') {
                // norm(x) = sqrt(sum(pow(x, 2)))
                return sqrt$2(sum$2(square$2(x), axis));
            }
            throw new Error(`Error in norm: invalid ord value: ${p}`);
        }
        throw new Error(`Error in norm: invalid axis: ${axis}`);
    }
    const norm = op({ norm_ });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the euclidean norm of scalar, vectors, and matrices.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     *
     * x.euclideanNorm().print();  // or tf.euclideanNorm(x)
     * ```
     *
     * @param x The input array.
     * @param axis Optional. If axis is null (the default), the input is
     * considered a vector and a single vector norm is computed over the entire
     * set of values in the Tensor, i.e. euclideanNorm(x) is equivalent
     * to euclideanNorm(x.reshape([-1])). If axis is a integer, the input
     * is considered a batch of vectors, and axis determines the axis in x
     * over which to compute vector norms. If axis is a 2-tuple of integer it is
     * considered a batch of matrices and axis determines the axes in NDArray
     * over which to compute a matrix norm.
     * @param keepDims Optional. If true, the norm have the same dimensionality
     * as the input.
     *
     * @doc {heading: 'Operations', subheading: 'Matrices'}
     */
    function euclideanNorm_(x, axis = null, keepDims = false) {
        return norm(x, 'euclidean', axis, keepDims);
    }
    const euclideanNorm = op({ euclideanNorm_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes exponential of the input `tf.Tensor` element-wise. `e ^ x`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, -3]);
     *
     * x.exp().print();  // or tf.exp(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function exp_(x) {
        const $x = convertToTensor(x, 'x', 'exp');
        const inputs = { x: $x };
        return ENGINE.runKernel(Exp, inputs);
    }
    const exp$2 = op({ exp_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns a `tf.Tensor` that has expanded rank, by inserting a dimension
     * into the tensor's shape.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     * const axis = 1;
     * x.expandDims(axis).print();
     * ```
     *
     * @param x The input tensor whose dimensions to be expanded.
     * @param axis The dimension index at which to insert shape of `1`. Defaults
     *     to 0 (the first dimension).
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function expandDims_(x, axis = 0) {
        const $x = convertToTensor(x, 'x', 'expandDims', 'string_or_numeric');
        assert$4(axis <= $x.rank, () => 'Axis must be <= rank of the tensor');
        const inputs = { input: $x };
        const attrs = { dim: axis };
        return ENGINE.runKernel(ExpandDims, inputs, attrs);
    }
    const expandDims$3 = op({ expandDims_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes exponential of the input `tf.Tensor` minus one element-wise.
     * `e ^ x - 1`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, -3]);
     *
     * x.expm1().print();  // or tf.expm1(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function expm1_(x) {
        const $x = convertToTensor(x, 'x', 'expm1');
        const inputs = { x: $x };
        return ENGINE.runKernel(Expm1, inputs);
    }
    const expm1$2 = op({ expm1_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Construct a tensor by repeating it the number of times given by reps.
     *
     * This operation creates a new tensor by replicating `input` `reps`
     * times. The output tensor's i'th dimension has `input.shape[i] *
     * reps[i]` elements, and the values of `input` are replicated
     * `reps[i]` times along the i'th dimension. For example, tiling
     * `[a, b, c, d]` by `[2]` produces `[a, b, c, d, a, b, c, d]`.
     *
     * ```js
     * const a = tf.tensor1d([1, 2]);
     *
     * a.tile([2]).print();    // or a.tile([2])
     * ```
     *
     * ```js
     * const a = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * a.tile([1, 2]).print();  // or a.tile([1, 2])
     * ```
     * @param x The tensor to tile.
     * @param reps Determines the number of replications per dimension.
     *
     * @doc {heading: 'Tensors', subheading: 'Slicing and Joining'}
     */
    function tile_(x, reps) {
        const $x = convertToTensor(x, 'x', 'tile', 'string_or_numeric');
        assert$4($x.rank === reps.length, () => `Error in transpose: rank of input ${$x.rank} ` +
            `must match length of reps ${reps}.`);
        const inputs = { x: $x };
        const attrs = { reps };
        return ENGINE.runKernel(Tile, inputs, attrs);
    }
    const tile$3 = op({ tile_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Create an identity matrix.
     *
     * @param numRows Number of rows.
     * @param numColumns Number of columns. Defaults to `numRows`.
     * @param batchShape If provided, will add the batch shape to the beginning
     *   of the shape of the returned `tf.Tensor` by repeating the identity
     *   matrix.
     * @param dtype Data type.
     * @returns Identity matrix of the specified size and data type, possibly
     *   with batch repetition if `batchShape` is specified.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function eye_(numRows, numColumns, batchShape, dtype = 'float32') {
        if (numColumns == null) {
            numColumns = numRows;
        }
        const buff = buffer([numRows, numColumns], dtype);
        const n = numRows <= numColumns ? numRows : numColumns;
        for (let i = 0; i < n; ++i) {
            buff.set(1, i, i);
        }
        const out = reshape$3(buff.toTensor(), [numRows, numColumns]);
        if (batchShape == null) {
            return out;
        }
        else {
            if (batchShape.length === 1) {
                return tile$3(expandDims$3(out, 0), [batchShape[0], 1, 1]);
            }
            else if (batchShape.length === 2) {
                // tslint:disable-next-line:no-unnecessary-type-assertion
                return tile$3(expandDims$3(expandDims$3(out, 0), 0), [batchShape[0], batchShape[1], 1, 1]);
            }
            else if (batchShape.length === 3) {
                // tslint:disable-next-line:no-unnecessary-type-assertion
                return tile$3(expandDims$3(expandDims$3(expandDims$3(out, 0), 0), 0), [
                    batchShape[0], batchShape[1], batchShape[2], 1, 1
                ]);
            }
            else {
                throw new Error(`eye() currently supports only 1D and 2D ` +
                    // tslint:disable-next-line:no-any
                    `batchShapes, but received ${batchShape.length}D.`);
            }
        }
    }
    const eye = op({ eye_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` filled with a scalar value.
     *
     * ```js
     * tf.fill([2, 2], 4).print();
     * ```
     *
     * @param shape An array of integers defining the output tensor shape.
     * @param value The scalar value to fill the tensor with.
     * @param dtype The type of an element in the resulting tensor. Defaults to
     * 'float'.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function fill$2(shape, value, dtype) {
        const attrs = { shape, value, dtype };
        return ENGINE.runKernel(Fill, {}, attrs);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes floor of input `tf.Tensor` element-wise: `floor(x)`.
     *
     * ```js
     * const x = tf.tensor1d([.6, 1.1, -3.3]);
     *
     * x.floor().print();  // or tf.floor(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function floor_(x) {
        const $x = convertToTensor(x, 'x', 'floor', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Floor, inputs);
    }
    const floor$3 = op({ floor_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Gather slices from tensor `x`'s axis `axis` according to `indices`.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     * const indices = tf.tensor1d([1, 3, 3], 'int32');
     *
     * x.gather(indices).print();
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     * const indices = tf.tensor1d([1, 1, 0], 'int32');
     *
     * x.gather(indices).print();
     * ```
     * @param x The input tensor whose slices to be gathered.
     * @param indices The indices of the values to extract.
     * @param axis The axis over which to select values. Defaults to 0.
     * @param batchDims Optional. The number of batch dimensions. It must be less
     *     than or equal to rank(indices). Defaults to 0.
     *     The output tensor will have shape of
     *     `x.shape[:axis] + indices.shape[batchDims:] + x.shape[axis + 1:]`
     *
     * @doc {heading: 'Tensors', subheading: 'Slicing and Joining'}
     */
    function gather_(x, indices, axis = 0, batchDims = 0) {
        const $x = convertToTensor(x, 'x', 'gather');
        const $indices = convertToTensor(indices, 'indices', 'gather', 'int32');
        const inputs = { x: $x, indices: $indices };
        const attrs = { axis, batchDims };
        return ENGINE.runKernel(GatherV2, inputs, attrs);
    }
    const gather$1 = op({ gather_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of (a > b) element-wise. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3]);
     * const b = tf.tensor1d([2, 2, 2]);
     *
     * a.greater(b).print();
     * ```
     *
     * @param a The first input tensor.
     * @param b The second input tensor. Must have the same dtype as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function greater_(a, b) {
        let $a = convertToTensor(a, 'a', 'greater', 'string_or_numeric');
        let $b = convertToTensor(b, 'b', 'greater', 'string_or_numeric');
        [$a, $b] = makeTypesMatch($a, $b);
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Greater, inputs);
    }
    const greater$3 = op({ greater_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of (a >= b) element-wise. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3]);
     * const b = tf.tensor1d([2, 2, 2]);
     *
     * a.greaterEqual(b).print();
     * ```
     *
     * @param a The first input tensor.
     * @param b The second input tensor. Must have the same dtype as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function greaterEqual_(a, b) {
        let $a = convertToTensor(a, 'a', 'greaterEqual', 'string_or_numeric');
        let $b = convertToTensor(b, 'b', 'greaterEqual', 'string_or_numeric');
        [$a, $b] = makeTypesMatch($a, $b);
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(GreaterEqual, inputs);
    }
    const greaterEqual$2 = op({ greaterEqual_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the imaginary part of a complex (or real) tensor.
     *
     * Given a tensor input, this operation returns a tensor of type float that is
     * the imaginary part of each element in input considered as a complex number.
     * If input is real, a tensor of all zeros is returned.
     *
     * ```js
     * const x = tf.complex([-2.25, 3.25], [4.75, 5.75]);
     * tf.imag(x).print();
     * ```
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function imag_(input) {
        const $input = convertToTensor(input, 'input', 'imag');
        const inputs = { input: $input };
        return ENGINE.runKernel(Imag, inputs);
    }
    const imag$2 = op({ imag_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns which elements of x are finite.
     *
     * ```js
     * const x = tf.tensor1d([NaN, Infinity, -Infinity, 0, 1]);
     *
     * x.isFinite().print();  // or tf.isNaN(x)
     * ```
     * @param x The input Tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function isFinite_(x) {
        const $x = convertToTensor(x, 'x', 'isFinite');
        const inputs = { x: $x };
        return ENGINE.runKernel(IsFinite, inputs);
    }
    const isFinite$3 = op({ isFinite_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns which elements of x are Infinity or -Infinity.
     *
     * ```js
     * const x = tf.tensor1d([NaN, Infinity, -Infinity, 0, 1]);
     *
     * x.isInf().print();  // or tf.isNaN(x)
     * ```
     * @param x The input Tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function isInf_(x) {
        const $x = convertToTensor(x, 'x', 'isInf');
        const inputs = { x: $x };
        return ENGINE.runKernel(IsInf, inputs);
    }
    const isInf$2 = op({ isInf_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * RReturns which elements of x are NaN.
     *
     * ```js
     * const x = tf.tensor1d([NaN, Infinity, -Infinity, 0, 1]);
     *
     * x.isNaN().print();  // or tf.isNaN(x)
     * ```
     * @param x The input Tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function isNaN_(x) {
        const $x = convertToTensor(x, 'x', 'isNaN');
        const inputs = { x: $x };
        return ENGINE.runKernel(IsNan, inputs);
    }
    const isNaN$3 = op({ isNaN_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes leaky rectified linear element-wise.
     *
     * See
     * [http://web.stanford.edu/~awni/papers/relu_hybrid_icml2013_final.pdf](
     *     http://web.stanford.edu/~awni/papers/relu_hybrid_icml2013_final.pdf)
     *
     * ```js
     * const x = tf.tensor1d([-1, 2, -3, 4]);
     *
     * x.leakyRelu(0.1).print();  // or tf.leakyRelu(x, 0.1)
     * ```
     * @param x The input tensor.
     * @param alpha The scaling factor for negative values, defaults to 0.2.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function leakyRelu_(x, alpha = 0.2) {
        const $x = convertToTensor(x, 'x', 'leakyRelu');
        const inputs = { x: $x };
        const attrs = { alpha };
        return ENGINE.runKernel(LeakyRelu, inputs, attrs);
    }
    const leakyRelu$2 = op({ leakyRelu_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of (a < b) element-wise. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3]);
     * const b = tf.tensor1d([2, 2, 2]);
     *
     * a.less(b).print();
     * ```
     * @param a The first input tensor.
     * @param b The second input tensor. Must have the same dtype as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function less_(a, b) {
        let $a = convertToTensor(a, 'a', 'less', 'string_or_numeric');
        let $b = convertToTensor(b, 'b', 'less', 'string_or_numeric');
        [$a, $b] = makeTypesMatch($a, $b);
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Less, inputs);
    }
    const less$3 = op({ less_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of (a <= b) element-wise. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3]);
     * const b = tf.tensor1d([2, 2, 2]);
     *
     * a.lessEqual(b).print();
     * ```
     *
     * @param a The first input tensor.
     * @param b The second input tensor. Must have the same dtype as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function lessEqual_(a, b) {
        let $a = convertToTensor(a, 'a', 'lessEqual', 'string_or_numeric');
        let $b = convertToTensor(b, 'b', 'lessEqual', 'string_or_numeric');
        [$a, $b] = makeTypesMatch($a, $b);
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(LessEqual, inputs);
    }
    const lessEqual$2 = op({ lessEqual_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Return an evenly spaced sequence of numbers over the given interval.
     *
     * ```js
     * tf.linspace(0, 9, 10).print();
     * ```
     * @param start The start value of the sequence.
     * @param stop The end value of the sequence.
     * @param num The number of values to generate.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function linspace(start, stop, num) {
        if (num <= 0) {
            throw new Error('The number of values should be positive.');
        }
        const attrs = { start, stop, num };
        return ENGINE.runKernel(LinSpace, {}, attrs);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Normalizes the activation of a local neighborhood across or within
     * channels.
     *
     * @param x The input tensor. The 4-D input tensor is treated as a 3-D array
     *     of 1D vectors (along the last dimension), and each vector is
     *     normalized independently.
     * @param depthRadius The number of adjacent channels in the 1D normalization
     *     window.
     * @param bias A constant bias term for the basis.
     * @param alpha A scale factor, usually positive.
     * @param beta An exponent.
     *
     * @doc {heading: 'Operations', subheading: 'Normalization'}
     */
    function localResponseNormalization_(x, depthRadius = 5, bias = 1, alpha = 1, beta = 0.5) {
        const $x = convertToTensor(x, 'x', 'localResponseNormalization');
        assert$4($x.rank === 4 || $x.rank === 3, () => `Error in localResponseNormalization: x must be rank 3 or 4 but got
               rank ${$x.rank}.`);
        assert$4(isInt(depthRadius), () => `Error in localResponseNormalization: depthRadius must be an ` +
            `integer but got depthRadius ${depthRadius}.`);
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            reshapedTo4D = true;
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
        }
        const inputs = { x: x4D };
        const attrs = { depthRadius, bias, alpha, beta };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(LRN, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        else {
            return res;
        }
    }
    const localResponseNormalization = op({ localResponseNormalization_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes natural logarithm of the input `tf.Tensor` element-wise: `ln(x)`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, Math.E]);
     *
     * x.log().print();  // or tf.log(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function log_(x) {
        const $x = convertToTensor(x, 'x', 'log', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Log, inputs);
    }
    const log$3 = op({ log_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes natural logarithm of the input `tf.Tensor` plus one
     * element-wise: `ln(1 + x)`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, Math.E - 1]);
     *
     * x.log1p().print();  // or tf.log1p(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function log1p_(x) {
        const $x = convertToTensor(x, 'x', 'log1p');
        const inputs = { x: $x };
        return ENGINE.runKernel(Log1p, inputs);
    }
    const log1p$2 = op({ log1p_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Provided `f(x)`, returns another function `g(x, dy?)`, which gives the
     * gradient of `f(x)` with respect to `x`.
     *
     * If `dy` is provided, the gradient of `f(x).mul(dy).sum()` with respect to
     * `x` is computed instead. `f(x)` must take a single tensor `x` and return a
     * single tensor `y`. If `f()` takes multiple inputs, use `tf.grads` instead.
     *
     * ```js
     * // f(x) = x ^ 2
     * const f = x => x.square();
     * // f'(x) = 2x
     * const g = tf.grad(f);
     *
     * const x = tf.tensor1d([2, 3]);
     * g(x).print();
     * ```
     *
     * ```js
     * // f(x) = x ^ 3
     * const f = x => x.pow(tf.scalar(3, 'int32'));
     * // f'(x) = 3x ^ 2
     * const g = tf.grad(f);
     * // f''(x) = 6x
     * const gg = tf.grad(g);
     *
     * const x = tf.tensor1d([2, 3]);
     * gg(x).print();
     * ```
     *
     * @param f The function f(x), to compute gradient for.
     *
     * @doc {heading: 'Training', subheading: 'Gradients'}
     */
    function grad(f) {
        assert$4(isFunction$5(f), () => 'The f passed in grad(f) must be a function');
        return (x, dy) => {
            // x can be of any dtype, thus null as the last argument.
            const $x = convertToTensor(x, 'x', 'tf.grad', 'string_or_numeric');
            const $dy = (dy != null) ? convertToTensor(dy, 'dy', 'tf.grad') : null;
            return ENGINE.tidy(() => {
                const { value, grads } = ENGINE.gradients(() => f($x), [$x], $dy);
                if ($dy != null) {
                    assertShapesMatch(value.shape, $dy.shape, 'The shape of dy passed in grad(f)(x, dy) must match the shape ' +
                        'returned by f(x)');
                }
                checkGrads(grads);
                return grads[0];
            });
        };
    }
    /**
     * Provided `f(x1, x2,...)`, returns another function `g([x1, x2,...], dy?)`,
     * which gives an array of gradients of `f()` with respect to each input
     * [`x1`,`x2`,...].
     *
     * If `dy` is passed when calling `g()`, the gradient of
     * `f(x1,...).mul(dy).sum()` with respect to each input is computed instead.
     * The provided `f` must take one or more tensors and return a single tensor
     * `y`. If `f()` takes a single input, we recommend using `tf.grad` instead.
     *
     * ```js
     * // f(a, b) = a * b
     * const f = (a, b) => a.mul(b);
     * // df / da = b, df / db = a
     * const g = tf.grads(f);
     *
     * const a = tf.tensor1d([2, 3]);
     * const b = tf.tensor1d([-2, -3]);
     * const [da, db] = g([a, b]);
     * console.log('da');
     * da.print();
     * console.log('db');
     * db.print();
     * ```
     *
     * @param f The function `f(x1, x2,...)` to compute gradients for.
     *
     * @doc {heading: 'Training', subheading: 'Gradients'}
     */
    function grads(f) {
        assert$4(isFunction$5(f), () => 'The f passed in grads(f) must be a function');
        return (args, dy) => {
            assert$4(Array.isArray(args), () => 'The args passed in grads(f)(args) must be an array ' +
                'of `Tensor`s or `TensorLike`s');
            // args can be of any dtype, thus null as the last argument.
            const $args = convertToTensorArray(args, 'args', 'tf.grads', 'string_or_numeric');
            const $dy = (dy != null) ? convertToTensor(dy, 'dy', 'tf.grads') : null;
            return ENGINE.tidy(() => {
                const { value, grads } = ENGINE.gradients(() => f(...$args), $args, $dy);
                if ($dy != null) {
                    assertShapesMatch(value.shape, $dy.shape, 'The shape of dy passed in grads(f)([x1,...], dy) must ' +
                        'match the shape returned by f([x1,...])');
                }
                checkGrads(grads);
                return grads;
            });
        };
    }
    /**
     * Like `tf.grad`, but also returns the value of `f()`. Useful when `f()`
     * returns a metric you want to show.
     *
     * The result is a rich object with the following properties:
     * - grad: The gradient of `f(x)` w.r.t `x` (result of `tf.grad`).
     * - value: The value returned by `f(x)`.
     *
     * ```js
     * // f(x) = x ^ 2
     * const f = x => x.square();
     * // f'(x) = 2x
     * const g = tf.valueAndGrad(f);
     *
     * const x = tf.tensor1d([2, 3]);
     * const {value, grad} = g(x);
     *
     * console.log('value');
     * value.print();
     * console.log('grad');
     * grad.print();
     * ```
     *
     * @doc {heading: 'Training', subheading: 'Gradients'}
     */
    function valueAndGrad(f) {
        assert$4(isFunction$5(f), () => 'The f passed in valueAndGrad(f) must be a function');
        return (x, dy) => {
            assert$4(x instanceof Tensor, () => 'The x passed in valueAndGrad(f)(x) must be a tensor');
            assert$4(dy == null || dy instanceof Tensor, () => 'The dy passed in valueAndGrad(f)(x, dy) must be a tensor');
            const { grads, value } = ENGINE.gradients(() => f(x), [x], dy);
            checkGrads(grads);
            return { grad: grads[0], value };
        };
    }
    /**
     * Like `tf.grads`, but returns also the value of `f()`. Useful when `f()`
     * returns a metric you want to show.
     *
     * The result is a rich object with the following properties:
     * - grads: The gradients of `f()` w.r.t each input (result of `tf.grads`).
     * - value: The value returned by `f(x)`.
     *
     * ```js
     * // f(a, b) = a * b
     * const f = (a, b) => a.mul(b);
     * // df/da = b, df/db = a
     * const g = tf.valueAndGrads(f);
     *
     * const a = tf.tensor1d([2, 3]);
     * const b = tf.tensor1d([-2, -3]);
     * const {value, grads} = g([a, b]);
     *
     * const [da, db] = grads;
     *
     * console.log('value');
     * value.print();
     *
     * console.log('da');
     * da.print();
     * console.log('db');
     * db.print();
     * ```
     *
     * @doc {heading: 'Training', subheading: 'Gradients'}
     */
    function valueAndGrads(f) {
        assert$4(isFunction$5(f), () => 'The f passed in valueAndGrads(f) must be a function');
        return (args, dy) => {
            assert$4(Array.isArray(args) && args.every(arg => arg instanceof Tensor), () => 'The args passed in valueAndGrads(f)(args) must be array of ' +
                'tensors');
            assert$4(dy == null || dy instanceof Tensor, () => 'The dy passed in valueAndGrads(f)(args, dy) must be a tensor');
            const res = ENGINE.gradients(() => f(...args), args, dy);
            if (dy != null) {
                assertShapesMatch(res.value.shape, dy.shape, 'The shape of dy passed in valueAndGrads(f)([x1,...], dy) must ' +
                    'match the shape returned by f([x1,...])');
            }
            checkGrads(res.grads);
            return res;
        };
    }
    /**
     * Computes and returns the gradient of f(x) with respect to the list of
     * trainable variables provided by `varList`. If no list is provided, it
     * defaults to all trainable variables.
     *
     * ```js
     * const a = tf.variable(tf.tensor1d([3, 4]));
     * const b = tf.variable(tf.tensor1d([5, 6]));
     * const x = tf.tensor1d([1, 2]);
     *
     * // f(a, b) = a * x ^ 2 + b * x
     * const f = () => a.mul(x.square()).add(b.mul(x)).sum();
     * // df/da = x ^ 2, df/db = x
     * const {value, grads} = tf.variableGrads(f);
     *
     * Object.keys(grads).forEach(varName => grads[varName].print());
     * ```
     *
     * @param f The function to execute. f() should return a scalar.
     * @param varList The list of variables to compute the gradients with respect
     *     to. Defaults to all trainable variables.
     * @returns An object with the following keys and values:
     *   - `value`: The value of the function `f`.
     *   - `grads`: A map from the names of the variables to the gradients.
     *     If the `varList` argument is provided explicitly and contains a subset of
     *     non-trainable variables, this map in the return value will contain keys
     *     that map the names of the non-trainable variables to `null`.
     *
     * @doc {heading: 'Training', subheading: 'Gradients'}
     */
    function variableGrads(f, varList) {
        assert$4(isFunction$5(f), () => 'The f passed in variableGrads(f) must be a function');
        assert$4(varList == null ||
            Array.isArray(varList) && varList.every(v => v instanceof Variable), () => 'The varList passed in variableGrads(f, varList) must be an array ' +
            'of variables');
        const specifiedVarList = varList != null;
        if (!specifiedVarList) {
            // Get all of the trainable variables.
            varList = [];
            for (const varName in ENGINE.registeredVariables) {
                varList.push(ENGINE.registeredVariables[varName]);
            }
        }
        const specifiedNonTrainable = specifiedVarList ? varList.filter(variable => !variable.trainable) : null;
        // Prune non-trainable variables.
        const originalVarCount = varList.length;
        varList = varList.filter(variable => variable.trainable);
        assert$4(varList.length > 0, () => `variableGrads() expects at least one of the input variables to ` +
            `be trainable, but none of the ${originalVarCount} variables is ` +
            `trainable.`);
        const allowNoGradients = true;
        const { value, grads } = ENGINE.gradients(f, varList, null, allowNoGradients);
        assert$4(grads.some(g => g != null), () => 'Cannot find a connection between any variable and the result of ' +
            'the loss function y=f(x). Please make sure the operations that ' +
            'use variables are inside the function f passed to minimize().');
        assert$4(value.rank === 0, () => `The f passed in variableGrads(f) must return a scalar, but it ` +
            `returned a rank-${value.rank} tensor`);
        const namedGrads = {};
        varList.forEach((v, i) => {
            if (grads[i] != null) {
                namedGrads[v.name] = grads[i];
            }
        });
        if (specifiedNonTrainable != null) {
            // If varList is explicitly provided and contains non-trainable values,
            // add them to the returned gradients with `null` values.
            specifiedNonTrainable.forEach(v => namedGrads[v.name] = null);
        }
        return { value, grads: namedGrads };
    }
    /**
     * Overrides the gradient computation of a function `f`.
     *
     * Takes a function
     * `f(...inputs, save) => {value: Tensor, gradFunc: (dy, saved) => Tensor[]}`
     * and returns another function `g(...inputs)` which takes the same inputs as
     * `f`. When called, `g` returns `f().value`. In backward mode, custom gradients
     * with respect to each input of `f` are computed using `f().gradFunc`.
     *
     * The `save` function passsed to `f` should be used for saving tensors needed
     * in the gradient. And the `saved` passed to the `gradFunc` is a
     * `NamedTensorMap`, which contains those saved tensor.
     *
     * ```js
     * const customOp = tf.customGrad((x, save) => {
     *   // Save x to make sure it's available later for the gradient.
     *   save([x]);
     *   // Override gradient of our custom x ^ 2 op to be dy * abs(x);
     *   return {
     *     value: x.square(),
     *     // Note `saved.x` which points to the `x` we saved earlier.
     *     gradFunc: (dy, saved) => [dy.mul(saved[0].abs())]
     *   };
     * });
     *
     * const x = tf.tensor1d([-1, -2, 3]);
     * const dx = tf.grad(x => customOp(x));
     *
     * console.log(`f(x):`);
     * customOp(x).print();
     * console.log(`f'(x):`);
     * dx(x).print();
     * ```
     *
     * @param f The function to evaluate in forward mode, which should return
     *     `{value: Tensor, gradFunc: (dy, saved) => Tensor[]}`, where `gradFunc`
     *     returns the custom gradients of `f` with respect to its inputs.
     *
     * @doc {heading: 'Training', subheading: 'Gradients'}
     */
    function customGrad(f) {
        return ENGINE.customGrad(f);
    }
    function checkGrads(grads) {
        const numNullGradients = grads.filter(g => g == null).length;
        if (numNullGradients > 0) {
            throw new Error(`Cannot compute gradient of y=f(x) with respect to x. Make sure that
    the f you passed encloses all operations that lead from x to y.`);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes `-1 * x` element-wise.
     *
     * ```js
     * const x = tf.tensor2d([1, 2, -2, 0], [2, 2]);
     *
     * x.neg().print();  // or tf.neg(x)
     * ```
     *
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function neg_(x) {
        const $x = convertToTensor(x, 'x', 'neg');
        const inputs = { x: $x };
        return ENGINE.runKernel(Neg, inputs);
    }
    const neg$2 = op({ neg_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes softplus of the input `tf.Tensor` element-wise: `log(exp(x) + 1)`
     *
     * ```js
     * const x = tf.tensor1d([0, 1, -1, .7]);
     *
     * x.softplus().print();  // or tf.softplus(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function softplus_(x) {
        const $x = convertToTensor(x, 'x', 'softplus');
        const inputs = { x: $x };
        return ENGINE.runKernel(Softplus$1, inputs);
    }
    const softplus$2 = op({ softplus_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes log sigmoid of the input `tf.Tensor` element-wise:
     * `logSigmoid(x)`. For numerical stability, we use `-tf.softplus(-x)`.
     *
     * ```js
     * const x = tf.tensor1d([0, 1, -1, .7]);
     *
     * x.logSigmoid().print();  // or tf.logSigmoid(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function logSigmoid_(x) {
        const $x = convertToTensor(x, 'x', 'logSigmoid');
        // Use a custom gradient to maintain previous implementation.
        // There is no LogSigmoid kernel in TF so we can't use engine.runKernel
        // directly
        const customOp = customGrad((x) => {
            // TODO(yassogba) we can remove the chained softplus call here only
            // after backends have modualrized softplus at which point we can call
            // engine runKernel(..., Sotfplus, ...) directly.
            const value = neg$2(softplus$2(neg$2(x)));
            const gradFunc = (dy) => {
                const derX = mul(dy, sigmoid$2(neg$2(x)));
                return derX;
            };
            return { value, gradFunc };
        });
        return customOp($x);
    }
    const logSigmoid = op({ logSigmoid_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Subtracts two `tf.Tensor`s element-wise, A - B. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([10, 20, 30, 40]);
     * const b = tf.tensor1d([1, 2, 3, 4]);
     *
     * a.sub(b).print();  // or tf.sub(a, b)
     * ```
     *
     * ```js
     * // Broadcast subtract a with b.
     * const a = tf.tensor1d([10, 20, 30, 40]);
     * const b = tf.scalar(5);
     *
     * a.sub(b).print();  // or tf.sub(a, b)
     * ```
     * @param a The first `tf.Tensor` to subtract from.
     * @param b The second `tf.Tensor` to be subtracted. Must have the same dtype as
     * `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function sub_(a, b) {
        let $a = convertToTensor(a, 'a', 'sub');
        let $b = convertToTensor(b, 'b', 'sub');
        [$a, $b] = makeTypesMatch($a, $b);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Sub, inputs);
    }
    const sub$2 = op({ sub_ });

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the log softmax.
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3]);
     *
     * a.logSoftmax().print();  // or tf.logSoftmax(a)
     * ```
     *
     * ```js
     * const a = tf.tensor2d([2, 4, 6, 1, 2, 3], [2, 3]);
     *
     * a.logSoftmax().print();  // or tf.logSoftmax(a)
     * ```
     *
     * @param logits The logits array.
     * @param axis The dimension softmax would be performed on. Defaults to `-1`
     *     which indicates the last dimension.
     *
     * @doc {heading: 'Operations', subheading: 'Normalization'}
     */
    function logSoftmax_(logits, axis = -1) {
        const $logits = convertToTensor(logits, 'logits', 'logSoftmax');
        if (axis === -1) {
            axis = $logits.rank - 1;
        }
        if (axis !== $logits.rank - 1) {
            throw Error('Log Softmax along a non-last dimension is not yet supported. ' +
                `Logits was rank ${$logits.rank} and axis was ${axis}`);
        }
        // const forward: ForwardFunc<Tensor> = (backend, save) => {
        //   const keepDims = true;
        //   const xMax = max(logits, axis, true);
        //   const shifted = sub(logits, xMax);
        //   const value =
        //       sub(cast(shifted, 'float32'), log(sum(exp(shifted), axis,
        //       keepDims)));
        //   save([value]);
        //   return value;
        // };
        // Use a custom gradient for numerical stability.
        const customOp = customGrad((logits, save) => {
            const keepDims = true;
            const xMax = max$3(logits, axis, true);
            const shifted = sub$2(logits, xMax);
            const value = sub$2(cast$3(shifted, 'float32'), log$3(sum$2(exp$2(shifted), axis, keepDims)));
            save([value]);
            const gradFunc = (dy, saved) => {
                const [value] = saved;
                const keepDims = true;
                const softmax = exp$2(value);
                return sub$2(dy, mul(sum$2(dy, axis, keepDims), softmax));
            };
            return { value, gradFunc };
        });
        return customOp($logits);
        // TODO Use Engine.runKernel when CPU/WebGL/WASM backends implement this.
        // const inputs: LogSoftmaxInputs = {logits: $logits};
        // const attrs: LogSoftmaxAttrs = {axis};
        // return ENGINE.runKernel(
        //            LogSoftmax, inputs as {} as NamedTensorMap,
        //            attrs as {} as NamedAttrMap);
    }
    const logSoftmax = op({ logSoftmax_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the log(sum(exp(elements across the reduction dimensions)).
     *
     * Reduces the input along the dimensions given in `axis`. Unless `keepDims`
     * is true, the rank of the array is reduced by 1 for each entry in `axis`.
     * If `keepDims` is true, the reduced dimensions are retained with length 1.
     * If `axis` has no entries, all dimensions are reduced, and an array with a
     * single element is returned.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3]);
     *
     * x.logSumExp().print();  // or tf.logSumExp(x)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * const axis = 1;
     * x.logSumExp(axis).print();  // or tf.logSumExp(a, axis)
     * ```
     * @param x The input tensor.
     * @param axis The dimension(s) to reduce. If null (the default),
     *     reduces all dimensions.
     * @param keepDims If true, retains reduced dimensions with length
     *     of 1. Defaults to false.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function logSumExp_(x, axis = null, keepDims = false) {
        const $x = convertToTensor(x, 'x', 'logSumExp');
        const axes = parseAxisParam(axis, $x.shape);
        const xMax = max$3($x, axes, true /* keepDims */);
        const a = sub$2($x, xMax);
        const b = exp$2(a);
        const c = sum$2(b, axes);
        const d = log$3(c);
        const res = add$2(reshape$3(xMax, d.shape), d);
        if (keepDims) {
            const newShape = expandShapeToKeepDim(res.shape, axes);
            return reshape$3(res, newShape);
        }
        return res;
    }
    const logSumExp = op({ logSumExp_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of `a AND b` element-wise. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([false, false, true, true], 'bool');
     * const b = tf.tensor1d([false, true, false, true], 'bool');
     *
     * a.logicalAnd(b).print();
     * ```
     *
     * @param a The first input tensor. Must be of dtype bool.
     * @param b The second input tensor. Must be of dtype bool.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function logicalAnd_(a, b) {
        const $a = convertToTensor(a, 'a', 'logicalAnd', 'bool');
        const $b = convertToTensor(b, 'b', 'logicalAnd', 'bool');
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(LogicalAnd, inputs);
    }
    const logicalAnd$2 = op({ logicalAnd_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of `NOT x` element-wise.
     *
     * ```js
     * const a = tf.tensor1d([false, true], 'bool');
     *
     * a.logicalNot().print();
     * ```
     *
     * @param x The input tensor. Must be of dtype 'bool'.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function logicalNot_(x) {
        const $x = convertToTensor(x, 'x', 'logicalNot', 'bool');
        const inputs = { x: $x };
        return ENGINE.runKernel(LogicalNot, inputs);
    }
    const logicalNot$2 = op({ logicalNot_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of `a OR b` element-wise. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([false, false, true, true], 'bool');
     * const b = tf.tensor1d([false, true, false, true], 'bool');
     *
     * a.logicalOr(b).print();
     * ```
     * @param a The first input tensor. Must be of dtype bool.
     * @param b The second input tensor. Must be of dtype bool.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function logicalOr_(a, b) {
        const $a = convertToTensor(a, 'a', 'logicalOr', 'bool');
        const $b = convertToTensor(b, 'b', 'logicalOr', 'bool');
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(LogicalOr, inputs);
    }
    const logicalOr$2 = op({ logicalOr_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of `a XOR b` element-wise. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([false, false, true, true], 'bool');
     * const b = tf.tensor1d([false, true, false, true], 'bool');
     *
     * a.logicalXor(b).print();
     * ```
     *
     * @param a The first input tensor. Must be of dtype bool.
     * @param b The second input tensor. Must be of dtype bool.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function logicalXor_(a, b) {
        const $a = convertToTensor(a, 'a', 'logicalXor', 'bool');
        const $b = convertToTensor(b, 'b', 'logicalXor', 'bool');
        assertAndGetBroadcastShape($a.shape, $b.shape);
        // x ^ y = (x | y) & ~(x & y)
        return logicalAnd$2(logicalOr$2(a, b), logicalNot$2(logicalAnd$2(a, b)));
    }
    const logicalXor = op({ logicalXor_ });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const INT32_MAX = 2147483648;
    /**
     * Searches for where a value would go in a sorted sequence.
     *
     * This is not a method for checking containment (like javascript in).
     *
     * The typical use case for this operation is "binning", "bucketing", or
     * "discretizing". The values are assigned to bucket-indices based on the edges
     * listed in 'sortedSequence'. This operation returns the bucket-index for each
     * value.
     *
     * The side argument controls which index is returned if a value lands exactly
     * on an edge.
     *
     * The axis is not settable for this operation. It always operates on the
     * innermost dimension (axis=-1). The operation will accept any number of outer
     * dimensions.
     *
     * Note: This operation assumes that 'sortedSequence' is sorted along the
     * innermost axis, maybe using 'sort(..., axis=-1)'. If the sequence is not
     * sorted no error is raised and the content of the returned tensor is not well
     * defined.
     *
     * ```js
     * const edges = tf.tensor1d([-1, 3.3, 9.1, 10.0]);
     * let values = tf.tensor1d([0.0, 4.1, 12.0]);
     * const result1 = tf.searchSorted(edges, values, 'left');
     * result1.print(); // [1, 2, 4]
     *
     * const seq = tf.tensor1d([0, 3, 9, 10, 10]);
     * values = tf.tensor1d([0, 4, 10]);
     * const result2 = tf.searchSorted(seq, values, 'left');
     * result2.print(); // [0, 2, 3]
     * const result3 = tf.searchSorted(seq, values, 'right');
     * result3.print(); // [1, 2, 5]
     *
     * const sortedSequence = tf.tensor2d([[0., 3., 8., 9., 10.],
     *                                     [1., 2., 3., 4., 5.]]);
     * values = tf.tensor2d([[9.8, 2.1, 4.3],
     *                       [0.1, 6.6, 4.5, ]]);
     * const result4 = tf.searchSorted(sortedSequence, values, 'left');
     * result4.print(); // [[4, 1, 2], [0, 5, 4]]
     * ```
     * @param sortedSequence: N-D. Sorted sequence.
     * @param values: N-D. Search values.
     * @param side: 'left'|'right'. Defaults to 'left'. 'left' corresponds to lower
     *     bound and 'right' to upper bound.
     * @return An N-D int32 tensor the size of values containing the result of
     *     applying either lower bound or upper bound (depending on side) to each
     *     value. The result is not a global index to the entire Tensor, but the
     *     index in the last dimension.
     * @doc {heading: 'Operations', subheading: 'Evaluation'}
     */
    function searchSorted_(sortedSequence, values, side = 'left') {
        const $sortedSequence = convertToTensor(sortedSequence, 'sortedSequence', 'searchSorted');
        const $values = convertToTensor(values, 'values', 'searchSorted');
        const sequenceSize = $sortedSequence.shape[$sortedSequence.shape.length - 1];
        const valuesSize = $values.shape[$values.shape.length - 1];
        const $sortedSequence2D = reshape$3($sortedSequence, [-1, sequenceSize]);
        const $values2D = reshape$3($values, [-1, valuesSize]);
        if ($sortedSequence2D.rank < 2) {
            throw new Error(`Sorted input argument must be at least 2-dimensional`);
        }
        if ($sortedSequence2D.shape[0] !== $values2D.shape[0]) {
            throw new Error(`Leading dimension of 'sortedSequence' and 'values' must match.`);
        }
        if (sizeFromShape($values2D.shape) >= INT32_MAX) {
            throw new Error(`values tensor size must less than ${INT32_MAX}`);
        }
        if ($sortedSequence2D.shape[1] >= INT32_MAX) {
            throw new Error(`trailing dim_size must less than ${INT32_MAX} for int32 output type, was ${$sortedSequence2D.shape[1]}`);
        }
        const inputs = {
            sortedSequence: $sortedSequence2D,
            values: $values2D,
        };
        const attrs = { side };
        return ENGINE.runKernel(SearchSorted, inputs, attrs);
    }
    const searchSorted$2 = op({ searchSorted_ });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Searches for where a value would go in a sorted sequence.
     *
     * This is not a method for checking containment (like javascript in).
     *
     * The typical use case for this operation is "binning", "bucketing", or
     * "discretizing". The values are assigned to bucket-indices based on the edges
     * listed in 'sortedSequence'. This operation returns the bucket-index for each
     * value.
     *
     * The index returned corresponds to the first edge greater than or equal to the
     * value.
     *
     * The axis is not settable for this operation. It always operates on the
     * innermost dimension (axis=-1). The operation will accept any number of outer
     * dimensions.
     *
     * Note: This operation assumes that 'lowerBound' is sorted along the
     * innermost axis, maybe using 'sort(..., axis=-1)'. If the sequence is not
     * sorted no error is raised and the content of the returned tensor is not well
     * defined.
     *
     * ```js
     * const edges = tf.tensor1d([-1, 3.3, 9.1, 10.0]);
     * let values = tf.tensor1d([0.0, 4.1, 12.0]);
     * const result1 = tf.lowerBound(edges, values);
     * result1.print(); // [1, 2, 4]
     *
     * const seq = tf.tensor1d([0, 3, 9, 10, 10]);
     * values = tf.tensor1d([0, 4, 10]);
     * const result2 = tf.lowerBound(seq, values);
     * result2.print(); // [0, 2, 3]
     *
     * const sortedSequence = tf.tensor2d([[0., 3., 8., 9., 10.],
     *                                     [1., 2., 3., 4., 5.]]);
     * values = tf.tensor2d([[9.8, 2.1, 4.3],
     *                       [0.1, 6.6, 4.5, ]]);
     * const result3 = tf.lowerBound(sortedSequence, values);
     * result3.print(); // [[4, 1, 2], [0, 5, 4]]
     * ```
     * @param sortedSequence: N-D. Sorted sequence.
     * @param values: N-D. Search values.
     * @return An N-D int32 tensor the size of values containing the result of
     *     applying lower bound to each value. The result is not a global index to
     *     the entire Tensor, but the index in the last dimension.
     * @doc {heading: 'Operations', subheading: 'Evaluation'}
     */
    function lowerBound$1(sortedSequence, values) {
        return searchSorted$2(sortedSequence, values, 'left');
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the 2D max pooling of an image.
     *
     * @param x The input tensor, of rank 4 or rank 3 of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is assumed.
     * @param filterSize The filter size: `[filterHeight, filterWidth]`. If
     *     `filterSize` is a single number, then `filterHeight == filterWidth`.
     * @param strides The strides of the pooling: `[strideHeight, strideWidth]`. If
     *     `strides` is a single number, then `strideHeight == strideWidth`.
     * @param dilations The dilation rates: `[dilationHeight, dilationWidth]`
     *     in which we sample input values across the height and width dimensions
     *     in dilated pooling. Defaults to `[1, 1]`. If `dilations` is a single
     *     number, then `dilationHeight == dilationWidth`. If it is greater than
     *     1, then all values of `strides` must be 1.
     * @param pad The type of padding algorithm.
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *    - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     */
    function maxPool_(x, filterSize, strides, pad, dimRoundingMode) {
        const $x = convertToTensor(x, 'x', 'maxPool');
        const dilations = 1;
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            reshapedTo4D = true;
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
        }
        assert$4(x4D.rank === 4, () => `Error in maxPool: input must be rank 4 but got rank ${x4D.rank}.`);
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in maxPool: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        checkPadOnDimRoundingMode('maxPool', pad, dimRoundingMode);
        const inputs = { x: x4D };
        const attrs = { filterSize, strides, pad, dimRoundingMode };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(MaxPool, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const maxPool$2 = op({ maxPool_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the 3D max pooling.
     *
     * ```js
     * const x = tf.tensor5d([1, 2, 3, 4, 5, 6, 7, 8], [1, 2, 2, 2, 1]);
     * const result = tf.maxPool3d(x, 2, 1, 'valid');
     * result.print();
     * ```
     *
     * @param x The input tensor, of rank 5 or rank 4 of shape
     *     `[batch, depth, height, width, inChannels]`.
     * @param filterSize The filter size:
     *     `[filterDepth, filterHeight, filterWidth]`.
     *     If `filterSize` is a single number,
     *     then `filterDepth == filterHeight == filterWidth`.
     * @param strides The strides of the pooling:
     *     `[strideDepth, strideHeight, strideWidth]`.
     *     If `strides` is a single number,
     *     then `strideDepth == strideHeight == strideWidth`.
     * @param pad The type of padding algorithm.
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1*1x1.
     *    - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     * @param dataFormat An optional string from: "NDHWC", "NCDHW". Defaults to
     *     "NDHWC". Specify the data format of the input and output data. With the
     *     default format "NDHWC", the data is stored in the order of: [batch,
     *     depth, height, width, channels]. Only "NDHWC" is currently supported.
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function maxPool3d_(x, filterSize = [1, 1, 1], strides, pad, dimRoundingMode, dataFormat = 'NDHWC') {
        const $x = convertToTensor(x, 'x', 'maxPool3d');
        let x5D = $x;
        let reshapedTo5D = false;
        if ($x.rank === 4) {
            reshapedTo5D = true;
            x5D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2], $x.shape[3]]);
        }
        assert$4(x5D.rank === 5, () => `Error in maxPool3d: x must be rank 5 but got rank ${x5D.rank}.`);
        assert$4(dataFormat === 'NDHWC', () => `Error in maxPool3d: Only NDHWC is currently supported, ` +
            `but got dataFormat of ${dataFormat}`);
        checkPadOnDimRoundingMode('maxPool3d', pad, dimRoundingMode);
        const inputs = { x: x5D };
        const attrs = { filterSize, strides, pad, dimRoundingMode, dataFormat };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(MaxPool3D, inputs, attrs);
        if (reshapedTo5D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3], res.shape[4]]);
        }
        return res;
    }
    const maxPool3d$1 = op({ maxPool3d_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the 2D max pooling of an image with Argmax index.
     * The indices in argmax are flattened, so that a maximum value at position `[b,
     * y, x, c]` becomes flattened index: `(y * width + x) * channels + c` if
     * include_batch_in_index is False; `((b * height + y) * width + x) * channels
     * +c` if include_batch_in_index is True.
     *
     * The indices returned are always in `[0, height) x [0, width)` before
     * flattening.
     *
     * @param x The input tensor, of rank 4 or rank 3 of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is assumed.
     * @param filterSize The filter size: `[filterHeight, filterWidth]`. If
     *     `filterSize` is a single number, then `filterHeight == filterWidth`.
     * @param strides The strides of the pooling: `[strideHeight, strideWidth]`. If
     *     `strides` is a single number, then `strideHeight == strideWidth`.
     * @param dataFormat An optional string from: "NDHWC", "NCDHW". Defaults to
     *     "NDHWC". Specify the data format of the input and output data. With the
     *     default format "NDHWC", the data is stored in the order of: [batch,
     *     depth, height, width, channels]. Only "NDHWC" is currently supported.
     * @param pad The type of padding algorithm.
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *    - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param includeBatchIndex Defaults to False. Whether to include batch
     *    dimension in flattened index of argmax.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function maxPoolWithArgmax_(x, filterSize, strides, pad, includeBatchInIndex = false) {
        const $x = convertToTensor(x, 'x', 'maxPoolWithArgmax');
        const inputs = { x: $x };
        const attrs = { filterSize, strides, pad, includeBatchInIndex };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const result = ENGINE.runKernel(MaxPoolWithArgmax, inputs, attrs);
        return { result: result[0], indexes: result[1] };
    }
    const maxPoolWithArgmax = op({ maxPoolWithArgmax_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the max of a and b (`a > b ? a : b`) element-wise.
     * Supports broadcasting.
     *
     * We also expose `tf.maximumStrict` which has the same signature as this op and
     * asserts that `a` and `b` are the same shape (does not broadcast).
     *
     * ```js
     * const a = tf.tensor1d([1, 4, 3, 16]);
     * const b = tf.tensor1d([1, 2, 9, 4]);
     *
     * a.maximum(b).print();  // or tf.maximum(a, b)
     * ```
     *
     * ```js
     * // Broadcast maximum a with b.
     * const a = tf.tensor1d([2, 4, 6, 8]);
     * const b = tf.scalar(5);
     *
     * a.maximum(b).print();  // or tf.maximum(a, b)
     * ```
     *
     * @param a The first tensor.
     * @param b The second tensor. Must have the same type as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function maximum_(a, b) {
        let $a = convertToTensor(a, 'a', 'maximum');
        let $b = convertToTensor(b, 'b', 'maximum');
        [$a, $b] = makeTypesMatch($a, $b);
        if ($a.dtype === 'bool') {
            $a = cast$3($a, 'int32');
            $b = cast$3($b, 'int32');
        }
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Maximum$1, inputs);
    }
    const maximum$3 = op({ maximum_ });

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the mean of elements across dimensions of a `tf.Tensor`.
     *
     * Reduces `x` along the dimensions given in `axis`. Unless `keepDims` is
     * true, the rank of the `tf.Tensor` is reduced by 1 for each entry in `axis`.
     * If `keepDims` is true, the reduced dimensions are retained with length 1.
     * If `axis` has no entries, all dimensions are reduced, and a `tf.Tensor` with
     * a single element is returned.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3]);
     *
     * x.mean().print();  // or tf.mean(a)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * const axis = 1;
     * x.mean(axis).print();  // or tf.mean(x, axis)
     * ```
     *
     * @param x The input tensor.
     * @param axis The dimension(s) to reduce. By default it reduces
     *     all dimensions.
     * @param keepDims If true, retains reduced dimensions with size 1.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function mean_(x, axis = null, keepDims = false) {
        const $x = convertToTensor(x, 'x', 'mean');
        const inputs = { x: $x };
        const attrs = { axis, keepDims };
        return ENGINE.runKernel(Mean, inputs, attrs);
    }
    const mean$1 = op({ mean_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with all elements set to 0.
     *
     * ```js
     * tf.zeros([2, 2]).print();
     * ```
     *
     * @param shape An array of integers defining the output tensor shape.
     * @param dtype The type of an element in the resulting tensor. Can
     *     be 'float32', 'int32' or 'bool'. Defaults to 'float'.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function zeros$2(shape, dtype = 'float32') {
        if (dtype === 'complex64') {
            const real = zeros$2(shape, 'float32');
            const imag = zeros$2(shape, 'float32');
            return complex$2(real, imag);
        }
        const values = makeZerosTypedArray(sizeFromShape(shape), dtype);
        return ENGINE.makeTensor(values, shape, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with all elements set to 1.
     *
     * ```js
     * tf.ones([2, 2]).print();
     * ```
     *
     * @param shape An array of integers defining the output tensor shape.
     * @param dtype The type of an element in the resulting tensor. Defaults to
     *     'float'.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function ones$1(shape, dtype = 'float32') {
        if (dtype === 'complex64') {
            const real = ones$1(shape, 'float32');
            const imag = zeros$2(shape, 'float32');
            return complex$2(real, imag);
        }
        const values = makeOnesTypedArray(sizeFromShape(shape), dtype);
        return ENGINE.makeTensor(values, shape, dtype);
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Broadcasts parameters for evaluation on an N-D grid.
     *
     * Given N one-dimensional coordinate arrays `*args`, returns a list `outputs`
     * of N-D coordinate arrays for evaluating expressions on an N-D grid.
     *
     * Notes:
     * `meshgrid` supports cartesian ('xy') and matrix ('ij') indexing conventions.
     * When the `indexing` argument is set to 'xy' (the default), the broadcasting
     * instructions for the first two dimensions are swapped.
     * Examples:
     * Calling `const [X, Y] = meshgrid(x, y)` with the tensors
     *
     * ```javascript
     * const x = [1, 2, 3];
     * const y = [4, 5, 6];
     * const [X, Y] = tf.meshgrid(x, y);
     * // X = [[1, 2, 3],
     * //      [1, 2, 3],
     * //      [1, 2, 3]]
     * // Y = [[4, 4, 4],
     * //      [5, 5, 5],
     * //      [6, 6, 6]]
     * ```
     *
     * @param x Tensor with rank geq 1.
     * @param y Tensor with rank geq 1.
     * @param indexing
     *
     * @doc {heading: 'Operations', subheading: 'Slicing and Joining'}
     */
    function meshgrid(x, y, { indexing = 'xy' } = {}) {
        if (indexing !== 'xy' && indexing !== 'ij') {
            throw new TypeError(`${indexing} is not a valid third argument to meshgrid`);
        }
        if (x === undefined) {
            return [];
        }
        let $x = convertToTensor(x, 'x', 'meshgrid', x instanceof Tensor ? x.dtype : 'float32');
        if (y === undefined) {
            return [$x];
        }
        let $y = convertToTensor(y, 'y', 'meshgrid', y instanceof Tensor ? y.dtype : 'float32');
        const w = sizeFromShape($x.shape);
        const h = sizeFromShape($y.shape);
        if (indexing === 'xy') {
            $x = reshape$3($x, [1, -1]);
            $y = reshape$3($y, [-1, 1]);
            return [
                matMul$1(ones$1([h, 1], $x.dtype), $x),
                matMul$1($y, ones$1([1, w], $y.dtype)),
            ];
        }
        $x = reshape$3($x, [-1, 1]);
        $y = reshape$3($y, [1, -1]);
        return [
            matMul$1($x, ones$1([1, h], $x.dtype)),
            matMul$1(ones$1([w, 1], $y.dtype), $y),
        ];
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the min of a and b (`a < b ? a : b`) element-wise.
     * Supports broadcasting.
     *
     * We also expose `minimumStrict` which has the same signature as this op and
     * asserts that `a` and `b` are the same shape (does not broadcast).
     *
     * ```js
     * const a = tf.tensor1d([1, 4, 3, 16]);
     * const b = tf.tensor1d([1, 2, 9, 4]);
     *
     * a.minimum(b).print();  // or tf.minimum(a, b)
     * ```
     *
     * ```js
     * // Broadcast minimum a with b.
     * const a = tf.tensor1d([2, 4, 6, 8]);
     * const b = tf.scalar(5);
     *
     * a.minimum(b).print();  // or tf.minimum(a, b)
     * ```
     *
     * @param a The first tensor.
     * @param b The second tensor. Must have the same type as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function minimum_(a, b) {
        let $a = convertToTensor(a, 'a', 'minimum');
        let $b = convertToTensor(b, 'b', 'minimum');
        [$a, $b] = makeTypesMatch($a, $b);
        if ($a.dtype === 'bool') {
            $a = cast$3($a, 'int32');
            $b = cast$3($b, 'int32');
        }
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Minimum$1, inputs);
    }
    const minimum$3 = op({ minimum_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Pads a `tf.Tensor` using mirror padding.
     *
     * This operation implements the `REFLECT` and `SYMMETRIC` modes of pad.
     *
     * ```js
     * const x = tf.range(0, 9).reshape([1, 1, 3, 3]);
     * x.mirrorPad([[0, 0], [0, 0], [2, 2], [2, 2]], 'reflect').print();
     * ```
     * @param x The tensor to pad.
     * @param paddings An array of length `R` (the rank of the tensor), where
     * each element is a length-2 tuple of ints `[padBefore, padAfter]`,
     * specifying how much to pad along each dimension of the tensor.
     * In "reflect" mode, the padded regions do not include the borders,
     * while in "symmetric" mode the padded regions do include the borders.
     * For example, if the input is `[1, 2, 3]` and paddings is `[0, 2]`,
     * then the output is `[1, 2, 3, 2, 1]` in "reflect" mode, and
     * `[1, 2, 3, 3, 2]` in "symmetric" mode.
     * If `mode` is "reflect" then both `paddings[D, 0]` and `paddings[D, 1]`
     * must be no greater than `x.shape[D] - 1`. If mode is "symmetric"
     * then both `paddings[D, 0]` and `paddings[D, 1]` must be no greater than
     * `x.shape[D]`
     * @param mode String to specify padding mode. Can be `'reflect' | 'symmetric'`
     */
    /** @doc {heading: 'Tensors', subheading: 'Transformations'} */
    function mirrorPad_(x, paddings, mode) {
        assert$4(mode === 'reflect' || mode === 'symmetric', () => `Invalid mode. Mode must be either reflect or symmetric. ` +
            `Got ${mode}.`);
        const $x = convertToTensor(x, 'x', 'mirrorPad');
        if ($x.rank === 0) {
            throw new Error('mirrorPad(scalar) is not defined. ' +
                'Pass non-scalar to mirrorPad');
        }
        assert$4(paddings.length === $x.rank, () => `Padding doesn't match input. Must be ${$x.rank}. ` +
            `Got ${paddings.length}.`);
        const shapeOffset = mode === 'reflect' ? 1 : 0;
        for (let i = 0; i < $x.rank; i++) {
            assert$4(paddings[i].length === 2, () => `Invalid number of paddings. Must be length of 2 each.`);
            assert$4(paddings[i][0] >= 0 && paddings[i][0] <= $x.shape[i] - shapeOffset &&
                paddings[i][1] >= 0 && paddings[i][1] <= $x.shape[i] - shapeOffset, () => `Padding in dimension ${i} cannot be greater than or equal ` +
                `to ${$x.shape[i] - shapeOffset} or less than 0 for input of ` +
                `shape ${$x.shape}`);
        }
        const attrs = { paddings, mode };
        const inputs = { x: $x };
        return ENGINE.runKernel(MirrorPad, inputs, attrs);
    }
    const mirrorPad$1 = op({ mirrorPad_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the mod of a and b element-wise.
     * `floor(x / y) * y + mod(x, y) = x`
     * Supports broadcasting.
     *
     * We also expose `tf.modStrict` which has the same signature as this op and
     * asserts that `a` and `b` are the same shape (does not broadcast).
     *
     * ```js
     * const a = tf.tensor1d([1, 4, 3, 16]);
     * const b = tf.tensor1d([1, 2, 9, 4]);
     *
     * a.mod(b).print();  // or tf.mod(a, b)
     * ```
     *
     * ```js
     * // Broadcast a mod b.
     * const a = tf.tensor1d([2, 4, 6, 8]);
     * const b = tf.scalar(5);
     *
     * a.mod(b).print();  // or tf.mod(a, b)
     * ```
     *
     * @param a The first tensor.
     * @param b The second tensor. Must have the same type as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function mod_(a, b) {
        let $a = convertToTensor(a, 'a', 'mod');
        let $b = convertToTensor(b, 'b', 'mod');
        [$a, $b] = makeTypesMatch($a, $b);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(Mod, inputs);
    }
    const mod$2 = op({ mod_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Calculates the mean and variance of `x`. The mean and variance are
     * calculated by aggregating the contents of `x` across `axes`. If `x` is
     * 1-D and `axes = [0]` this is just the mean and variance of a vector.
     *
     * @param x The input tensor.
     * @param axis The dimension(s) along with to compute mean and
     *     variance. By default it reduces all dimensions.
     * @param keepDims If true, the moments have the same dimensionality as the
     *     input.
     * @return An object with two keys: `mean` and `variance`.
     *
     * @doc {heading: 'Operations', subheading: 'Normalization'}
     */
    function moments_(x, axis = null, keepDims = false) {
        x = convertToTensor(x, 'x', 'moments');
        const axes = parseAxisParam(axis, x.shape);
        const xMean = mean$1(x, axes, keepDims);
        let keepDimsShape = xMean.shape;
        if (!keepDims) {
            keepDimsShape = expandShapeToKeepDim(xMean.shape, axes);
        }
        const devSquared = square$2(sub$2(cast$3(x, 'float32'), reshape$3(xMean, keepDimsShape)));
        const variance = mean$1(devSquared, axes, keepDims);
        return { mean: xMean, variance };
    }
    const moments = op({ moments_ });

    /**
     * Computes the next states and outputs of a stack of LSTMCells.
     *
     * Each cell output is used as input to the next cell.
     *
     * Returns `[cellState, cellOutput]`.
     *
     * Derived from tf.contrib.rn.MultiRNNCell.
     *
     * @param lstmCells Array of LSTMCell functions.
     * @param data The input to the cell.
     * @param c Array of previous cell states.
     * @param h Array of previous cell outputs.
     *
     * @doc {heading: 'Operations', subheading: 'RNN'}
     */
    function multiRNNCell_(lstmCells, data, c, h) {
        const $data = convertToTensor(data, 'data', 'multiRNNCell');
        const $c = convertToTensorArray(c, 'c', 'multiRNNCell');
        const $h = convertToTensorArray(h, 'h', 'multiRNNCell');
        let input = $data;
        const newStates = [];
        for (let i = 0; i < lstmCells.length; i++) {
            const output = lstmCells[i](input, $c[i], $h[i]);
            newStates.push(output[0]);
            newStates.push(output[1]);
            input = output[1];
        }
        const newC = [];
        const newH = [];
        for (let i = 0; i < newStates.length; i += 2) {
            newC.push(newStates[i]);
            newH.push(newStates[i + 1]);
        }
        return [newC, newH];
    }
    const multiRNNCell = op({ multiRNNCell_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with values drawn from a multinomial distribution.
     *
     * ```js
     * const probs = tf.tensor([.75, .25]);
     * tf.multinomial(probs, 3).print();
     * ```
     *
     * @param logits 1D array with unnormalized log-probabilities, or
     *     2D array of shape `[batchSize, numOutcomes]`. See the `normalized`
     *     parameter.
     * @param numSamples Number of samples to draw for each row slice.
     * @param seed The seed number.
     * @param normalized Whether the provided `logits` are normalized true
     *     probabilities (sum to 1). Defaults to false.
     * @return 1D array of shape `[numSamples]`, or 2D array of shape
     *     `[batchSize, numSamples]`, depending on the rank of the input.
     *
     * @doc {heading: 'Tensors', subheading: 'Random'}
     */
    function multinomial_(logits, numSamples, seed, normalized = false) {
        const $logits = convertToTensor(logits, 'logits', 'multinomial');
        const numOutcomes = $logits.size;
        const origRank = $logits.rank;
        if (numOutcomes < 2) {
            throw new Error(`Error in multinomial: you need at least 2 outcomes, but got ` +
                `${numOutcomes}.`);
        }
        if (origRank > 2) {
            throw new Error(`Rank of probabilities must be 1 or 2, but is ${origRank}`);
        }
        // TODO(lina128): Investigate correct seed behavior. The code seems not allow
        // setting see to 0.
        seed = seed || Math.random();
        // The kernel only accepts (and returns) rank 2 tensors.
        const logits2D = origRank === 1 ? reshape$3($logits, [1, -1]) : $logits;
        const inputs = { logits: logits2D };
        const attrs = { numSamples, seed, normalized };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(Multinomial, inputs, attrs);
        // tslint:disable-next-line:no-unnecessary-type-assertion
        return origRank === 1 ? reshape$3(res, [res.size]) : res;
    }
    const multinomial$2 = op({ multinomial_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the truth value of (a != b) element-wise. Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3]);
     * const b = tf.tensor1d([0, 2, 3]);
     *
     * a.notEqual(b).print();
     * ```
     * @param a The first input tensor.
     * @param b The second input tensor. Must have the same dtype as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    function notEqual_(a, b) {
        let $a = convertToTensor(a, 'a', 'notEqual', 'string_or_numeric');
        let $b = convertToTensor(b, 'b', 'notEqual', 'string_or_numeric');
        [$a, $b] = makeTypesMatch($a, $b);
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        return ENGINE.runKernel(NotEqual, inputs);
    }
    const notEqual$3 = op({ notEqual_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with all elements set to 1 with the same shape as the
     * given tensor.
     *
     * ```js
     * const x = tf.tensor([1, 2]);
     * tf.onesLike(x).print();
     * ```
     * @param x A tensor.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function onesLike_(x) {
        const $x = convertToTensor(x, 'x', 'onesLike');
        const inputs = { x: $x };
        return ENGINE.runKernel(OnesLike, inputs);
    }
    const onesLike$2 = op({ onesLike_ });

    /**
     * Computes the outer product of two vectors, `v1` and `v2`.
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3]);
     * const b = tf.tensor1d([3, 4, 5]);
     *
     * tf.outerProduct(a, b).print();
     * ```
     * @param v1 The first vector in the outer product operation.
     * @param v2 The second vector in the outer product operation.
     *
     * @doc {heading: 'Operations', subheading: 'Matrices'}
     */
    function outerProduct_(v1, v2) {
        const $v1 = convertToTensor(v1, 'v1', 'outerProduct');
        const $v2 = convertToTensor(v2, 'v2', 'outerProduct');
        assert$4($v1.rank === 1 && $v2.rank === 1, () => `Error in outerProduct: inputs must be rank 1, but got ranks ` +
            `${$v1.rank} and ${$v2.rank}.`);
        const v12D = reshape$3($v1, [-1, 1]);
        const v22D = reshape$3($v2, [1, -1]);
        return matMul$1(v12D, v22D);
    }
    const outerProduct = op({ outerProduct_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Pads a `tf.Tensor` with a given value and paddings.
     *
     * This operation implements `CONSTANT` mode. For `REFLECT` and `SYMMETRIC`,
     * refer to `tf.mirrorPad`
     *
     * Also available are stricter rank-specific methods with the same signature
     * as this method that assert that `paddings` is of given length.
     *   - `tf.pad1d`
     *   - `tf.pad2d`
     *   - `tf.pad3d`
     *   - `tf.pad4d`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     * x.pad([[1, 2]]).print();
     * ```
     * @param x The tensor to pad.
     * @param paddings An array of length `R` (the rank of the tensor), where
     * each element is a length-2 tuple of ints `[padBefore, padAfter]`,
     * specifying how much to pad along each dimension of the tensor.
     * @param constantValue The pad value to use. Defaults to 0.
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function pad_(x, paddings, constantValue = 0) {
        const $x = convertToTensor(x, 'x', 'pad');
        if ($x.rank === 0) {
            throw new Error('pad(scalar) is not defined. Pass non-scalar to pad');
        }
        const attrs = { paddings, constantValue };
        const inputs = { x: $x };
        return ENGINE.runKernel(PadV2, inputs, attrs);
    }
    const pad$1 = op({ pad_ });

    /**
     * Pads a `tf.Tensor1D` with a given value and paddings. See `pad` for details.
     */
    function pad1d_(x, paddings, constantValue = 0) {
        assert$4(paddings.length === 2, () => 'Invalid number of paddings. Must be length of 2.');
        return pad$1(x, [paddings], constantValue);
    }
    const pad1d = op({ pad1d_ });

    /**
     * Pads a `tf.Tensor2D` with a given value and paddings. See `pad` for details.
     */
    function pad2d_(x, paddings, constantValue = 0) {
        assert$4(paddings.length === 2 && paddings[0].length === 2 &&
            paddings[1].length === 2, () => 'Invalid number of paddings. Must be length of 2 each.');
        return pad$1(x, paddings, constantValue);
    }
    const pad2d = op({ pad2d_ });

    /**
     * Pads a `tf.Tensor3D` with a given value and paddings. See `pad` for details.
     */
    function pad3d_(x, paddings, constantValue = 0) {
        assert$4(paddings.length === 3 && paddings[0].length === 2 &&
            paddings[1].length === 2 && paddings[2].length === 2, () => 'Invalid number of paddings. Must be length of 2 each.');
        return pad$1(x, paddings, constantValue);
    }
    const pad3d = op({ pad3d_ });

    /**
     * Pads a `tf.Tensor4D` with a given value and paddings. See `pad` for details.
     */
    function pad4d_(x, paddings, constantValue = 0) {
        assert$4(paddings.length === 4 && paddings[0].length === 2 &&
            paddings[1].length === 2 && paddings[2].length === 2 &&
            paddings[3].length === 2, () => 'Invalid number of paddings. Must be length of 2 each.');
        return pad$1(x, paddings, constantValue);
    }
    const pad4d = op({ pad4d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * This operation divides "spatial" dimensions `[1, ..., M]` of the input into
     * a grid of blocks of shape `blockShape`, and interleaves these blocks with
     * the "batch" dimension (0) such that in the output, the spatial
     * dimensions `[1, ..., M]` correspond to the position within the grid,
     * and the batch dimension combines both the position within a spatial block
     * and the original batch position. Prior to division into blocks,
     * the spatial dimensions of the input are optionally zero padded
     * according to `paddings`. See below for a precise description.
     *
     * ```js
     * const x = tf.tensor4d([1, 2, 3, 4], [1, 2, 2, 1]);
     * const blockShape = [2, 2];
     * const paddings = [[0, 0], [0, 0]];
     *
     * x.spaceToBatchND(blockShape, paddings).print();
     * ```
     *
     * @param x A `tf.Tensor`. N-D with `x.shape` = `[batch] + spatialShape +
     * remainingShape`, where spatialShape has `M` dimensions.
     * @param blockShape A 1-D array. Must have shape `[M]`, all values must
     * be >= 1.
     * @param paddings A 2-D array. Must have shape `[M, 2]`, all values must be >=
     *     0. `paddings[i] = [padStart, padEnd]` specifies the amount to zero-pad
     * from input dimension `i + 1`, which corresponds to spatial dimension `i`. It
     * is required that
     * `(inputShape[i + 1] + padStart + padEnd) % blockShape[i] === 0`
     *
     * This operation is equivalent to the following steps:
     *
     * 1. Zero-pad the start and end of dimensions `[1, ..., M]` of the input
     * according to `paddings` to produce `padded` of shape paddedShape.
     *
     * 2. Reshape `padded` to `reshapedPadded` of shape:
     * `[batch] + [paddedShape[1] / blockShape[0], blockShape[0], ...,
     * paddedShape[M] / blockShape[M-1], blockShape[M-1]] + remainingShape`
     *
     * 3. Permute dimensions of `reshapedPadded` to produce `permutedReshapedPadded`
     * of shape: `blockShape + [batch] + [paddedShape[1] / blockShape[0], ...,
     * paddedShape[M] / blockShape[M-1]] + remainingShape`
     *
     * 4. Reshape `permutedReshapedPadded` to flatten `blockShape` into the
     * batch dimension, producing an output tensor of shape:
     * `[batch * prod(blockShape)] + [paddedShape[1] / blockShape[0], ...,
     * paddedShape[M] / blockShape[M-1]] + remainingShape`
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function spaceToBatchND_(x, blockShape, paddings) {
        const $x = convertToTensor(x, 'x', 'spaceToBatchND');
        assert$4($x.rank >= 1 + blockShape.length, () => `input rank ${$x.rank} should be > than [blockShape] ${blockShape.length}`);
        assert$4(paddings.length === blockShape.length, () => `paddings.shape[0] ${paddings.length} must be equal to [blockShape] ${blockShape.length}`);
        assert$4($x.shape.reduce((a, b, i) => {
            if (i > 0 && i <= blockShape.length) {
                return a &&
                    ((b + paddings[i - 1][0] + paddings[i - 1][1]) %
                        blockShape[i - 1] ===
                        0);
            }
            return a;
        }, true), () => `input spatial dimensions ${$x.shape.slice(1)} with paddings ${paddings.toString()} must be divisible by blockShapes ${blockShape.toString()}`);
        const inputs = { x: $x };
        const attrs = { blockShape, paddings };
        return ENGINE.runKernel(SpaceToBatchND, inputs, attrs);
    }
    const spaceToBatchND$2 = op({ spaceToBatchND_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Performs an N-D pooling operation
     *
     * @param input The input tensor, of rank 4 or rank 3 of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is assumed.
     * @param windowShape The filter size: `[filterHeight, filterWidth]`. If
     *     `filterSize` is a single number, then `filterHeight == filterWidth`.
     * @param poolingType The type of pooling, either 'max' or 'avg'.
     * @param pad The type of padding algorithm:
     *    - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *    - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *    - For more info, see this guide:
     *     [https://www.tensorflow.org/api_guides/python/nn#Convolution](
     *         https://www.tensorflow.org/api_guides/python/nn#Convolution)
     * @param dilations The dilation rates: `[dilationHeight, dilationWidth]`
     *     in which we sample input values across the height and width dimensions
     *     in dilated pooling. Defaults to `[1, 1]`. If `dilationRate` is a single
     *     number, then `dilationHeight == dilationWidth`. If it is greater than
     *     1, then all values of `strides` must be 1.
     * @param strides The strides of the pooling: `[strideHeight, strideWidth]`. If
     *     `strides` is a single number, then `strideHeight == strideWidth`.
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function pool_(input, windowShape, poolingType, pad, dilations, strides, dimRoundingMode) {
        if (dilations == null) {
            dilations = [1, 1];
        }
        if (strides == null) {
            strides = 1;
        }
        if (pad === 0) {
            pad = 'valid';
        }
        const $x = convertToTensor(input, 'x', 'maxPool');
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            reshapedTo4D = true;
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
        }
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in pool: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        const convInfo = computePool2DInfo(x4D.shape, windowShape, strides, dilations, pad);
        const dilation = [convInfo.dilationHeight, convInfo.dilationWidth];
        // The following implementation does batchToSpace(pool(spaceToBatch(x)))
        // whenever dilation > 1 since the TF kernels do not support dilation > 1.
        // tslint:disable-next-line:max-line-length
        // https://github.com/tensorflow/tensorflow/blob/50f6bb67dc98c9b74630b6047aae7a4f8a40fd02/tensorflow/python/ops/nn_ops.py#L1037
        let basePadding;
        if (pad === 'same') {
            basePadding = withSpaceToBatchBasePaddings([convInfo.filterHeight, convInfo.filterWidth], dilation);
        }
        else {
            basePadding = [[0, 0], [0, 0]];
        }
        const isDilationOne = dilation[0] === 1 && dilation[1] === 1;
        const [adjustedPadding, adjustedCrops] = requiredSpaceToBatchPaddings([convInfo.inHeight, convInfo.inWidth], dilation, basePadding);
        const convertedPad = isDilationOne ? pad : 'valid';
        const convertedX = isDilationOne ? x4D : spaceToBatchND$2(x4D, dilation, adjustedPadding);
        const forwardOp = poolingType === 'avg' ?
            () => avgPool$2(convertedX, windowShape, strides, convertedPad, dimRoundingMode) :
            () => maxPool$2(convertedX, windowShape, strides, convertedPad, dimRoundingMode);
        const y = forwardOp();
        const res = isDilationOne ? y : batchToSpaceND$2(y, dilation, adjustedCrops);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    // Helper function to compute crops and paddings for pool with dilation > 1.
    // tslint:disable-next-line:max-line-length
    // https://github.com/tensorflow/tensorflow/blob/50f6bb67dc98c9b74630b6047aae7a4f8a40fd02/tensorflow/python/ops/array_ops.py#L2184
    function requiredSpaceToBatchPaddings(inputShape, blockShape, basePadding) {
        const padStart = basePadding.map(b => b[0]);
        const origPadEnd = basePadding.map(b => b[1]);
        const fullInputShape = inputShape.concat(padStart, origPadEnd);
        const padEndExtra = blockShape.map((b, i) => (b - fullInputShape[i] % b) % b);
        const padEnd = origPadEnd.map((s, i) => s + padEndExtra[i]);
        const paddings = blockShape.map((_, i) => [padStart[i], padEnd[i]]);
        const crops = blockShape.map((_, i) => [0, padEndExtra[i]]);
        return [paddings, crops];
    }
    // Helper function to compute base paddings for pool with dilation > 1.
    // tslint:disable-next-line:max-line-length
    // https://github.com/tensorflow/tensorflow/blob/50f6bb67dc98c9b74630b6047aae7a4f8a40fd02/tensorflow/python/ops/nn_ops.py#L524
    function withSpaceToBatchBasePaddings(filterShape, dilation) {
        // Spatial dimensions of the filters and the upsampled filters in which we
        // introduce (rate - 1) zeros between consecutive filter values.
        const dilatedFilterShape = filterShape.map((s, i) => {
            return s + (s - 1) * (dilation[i] - 1);
        });
        const padExtraShape = dilatedFilterShape.map(s => s - 1);
        // When padding is odd, we pad more at end, following the same
        // convention as conv2d.
        const padExtraStart = padExtraShape.map(s => Math.floor(s / 2));
        const padExtraEnd = padExtraShape.map((s, i) => s - padExtraStart[i]);
        return padExtraShape.map((_, i) => {
            return [padExtraStart[i], padExtraEnd[i]];
        });
    }
    const pool$1 = op({ pool_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes leaky rectified linear element-wise with parametric alphas.
     *
     * `x < 0 ? alpha * x : f(x) = x`
     *
     * ```js
     * const x = tf.tensor1d([-1, 2, -3, 4]);
     * const alpha = tf.scalar(0.1);
     *
     * x.prelu(alpha).print();  // or tf.prelu(x, alpha)
     * ```
     * @param x The input tensor.
     * @param alpha Scaling factor for negative values.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function prelu_(x, alpha) {
        const $x = convertToTensor(x, 'x', 'prelu');
        const $alpha = convertToTensor(alpha, 'alpha', 'prelu');
        const inputs = { x: $x, alpha: $alpha };
        return ENGINE.runKernel(Prelu, inputs);
    }
    const prelu$3 = op({ prelu_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the product of elements across dimensions of a `tf.Tensor`.
     *
     * Reduces the input along the dimensions given in `axes`. Unless `keepDims`
     * is true, the rank of the `tf.Tensor` is reduced by 1 for each entry in
     * `axes`. If `keepDims` is true, the reduced dimensions are retained with
     * length 1. If `axes` has no entries, all dimensions are reduced, and a
     * `tf.Tensor` with a single element is returned.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3]);
     *
     * x.prod().print();  // or tf.prod(x)
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * const axis = 1;
     * x.prod(axis).print();  // or tf.prod(x, axis)
     * ```
     *
     * @param x The input tensor to compute the product over. If the dtype is `bool`
     *   it will be converted to `int32` and the output dtype will be `int32`.
     * @param axis The dimension(s) to reduce. By default it reduces
     *     all dimensions.
     * @param keepDims If true, retains reduced dimensions with size 1.
     *
     * @doc {heading: 'Operations', subheading: 'Reduction'}
     */
    function prod_(x, axis = null, keepDims = false) {
        let $x = convertToTensor(x, 'x', 'prod');
        if ($x.dtype === 'bool') {
            // bool is not an allowed type for the underlying kernel.
            $x = cast$3($x, 'int32');
        }
        const inputs = { x: $x };
        const attrs = { axis, keepDims };
        return ENGINE.runKernel(Prod, inputs, attrs);
    }
    const prod$2 = op({ prod_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with values sampled from a random number generator
     * function defined by the user.
     *
     * @param shape An array of integers defining the output tensor shape.
     * @param randFunction A random number generator function which is called
     * for each element in the output tensor.
     * @param dtype The data type of the output tensor. Defaults to 'float32'.
     *
     * @doc {heading: 'Tensors', subheading: 'Random'}
     */
    function rand_(shape, randFunction, dtype) {
        const size = sizeFromShape(shape);
        let values = null;
        if (dtype == null || dtype === 'float32') {
            values = new Float32Array(size);
        }
        else if (dtype === 'int32') {
            values = new Int32Array(size);
        }
        else if (dtype === 'bool') {
            values = new Uint8Array(size);
        }
        else {
            throw new Error(`Unknown data type ${dtype}`);
        }
        for (let i = 0; i < size; i++) {
            values[i] = randFunction();
        }
        return ENGINE.makeTensor(values, shape, dtype);
    }
    const rand = op({ rand_ });

    var alea$1 = {exports: {}};

    (function (module) {
    	// A port of an algorithm by Johannes Baagøe <baagoe@baagoe.com>, 2010
    	// http://baagoe.com/en/RandomMusings/javascript/
    	// https://github.com/nquinlan/better-random-numbers-for-javascript-mirror
    	// Original work is under MIT license -

    	// Copyright (C) 2010 by Johannes Baagøe <baagoe@baagoe.org>
    	//
    	// Permission is hereby granted, free of charge, to any person obtaining a copy
    	// of this software and associated documentation files (the "Software"), to deal
    	// in the Software without restriction, including without limitation the rights
    	// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    	// copies of the Software, and to permit persons to whom the Software is
    	// furnished to do so, subject to the following conditions:
    	// 
    	// The above copyright notice and this permission notice shall be included in
    	// all copies or substantial portions of the Software.
    	// 
    	// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    	// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    	// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    	// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    	// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    	// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    	// THE SOFTWARE.



    	(function(global, module, define) {

    	function Alea(seed) {
    	  var me = this, mash = Mash();

    	  me.next = function() {
    	    var t = 2091639 * me.s0 + me.c * 2.3283064365386963e-10; // 2^-32
    	    me.s0 = me.s1;
    	    me.s1 = me.s2;
    	    return me.s2 = t - (me.c = t | 0);
    	  };

    	  // Apply the seeding algorithm from Baagoe.
    	  me.c = 1;
    	  me.s0 = mash(' ');
    	  me.s1 = mash(' ');
    	  me.s2 = mash(' ');
    	  me.s0 -= mash(seed);
    	  if (me.s0 < 0) { me.s0 += 1; }
    	  me.s1 -= mash(seed);
    	  if (me.s1 < 0) { me.s1 += 1; }
    	  me.s2 -= mash(seed);
    	  if (me.s2 < 0) { me.s2 += 1; }
    	  mash = null;
    	}

    	function copy(f, t) {
    	  t.c = f.c;
    	  t.s0 = f.s0;
    	  t.s1 = f.s1;
    	  t.s2 = f.s2;
    	  return t;
    	}

    	function impl(seed, opts) {
    	  var xg = new Alea(seed),
    	      state = opts && opts.state,
    	      prng = xg.next;
    	  prng.int32 = function() { return (xg.next() * 0x100000000) | 0; };
    	  prng.double = function() {
    	    return prng() + (prng() * 0x200000 | 0) * 1.1102230246251565e-16; // 2^-53
    	  };
    	  prng.quick = prng;
    	  if (state) {
    	    if (typeof(state) == 'object') copy(state, xg);
    	    prng.state = function() { return copy(xg, {}); };
    	  }
    	  return prng;
    	}

    	function Mash() {
    	  var n = 0xefc8249d;

    	  var mash = function(data) {
    	    data = data.toString();
    	    for (var i = 0; i < data.length; i++) {
    	      n += data.charCodeAt(i);
    	      var h = 0.02519603282416938 * n;
    	      n = h >>> 0;
    	      h -= n;
    	      h *= n;
    	      n = h >>> 0;
    	      h -= n;
    	      n += h * 0x100000000; // 2^32
    	    }
    	    return (n >>> 0) * 2.3283064365386963e-10; // 2^-32
    	  };

    	  return mash;
    	}


    	if (module && module.exports) {
    	  module.exports = impl;
    	} else if (define && define.amd) {
    	  define(function() { return impl; });
    	} else {
    	  this.alea = impl;
    	}

    	})(
    	  commonjsGlobal,
    	  module,    // present in node.js
    	  (typeof undefined) == 'function'    // present with an AMD loader
    	);
    } (alea$1));

    var xor128$1 = {exports: {}};

    (function (module) {
    	// A Javascript implementaion of the "xor128" prng algorithm by
    	// George Marsaglia.  See http://www.jstatsoft.org/v08/i14/paper

    	(function(global, module, define) {

    	function XorGen(seed) {
    	  var me = this, strseed = '';

    	  me.x = 0;
    	  me.y = 0;
    	  me.z = 0;
    	  me.w = 0;

    	  // Set up generator function.
    	  me.next = function() {
    	    var t = me.x ^ (me.x << 11);
    	    me.x = me.y;
    	    me.y = me.z;
    	    me.z = me.w;
    	    return me.w ^= (me.w >>> 19) ^ t ^ (t >>> 8);
    	  };

    	  if (seed === (seed | 0)) {
    	    // Integer seed.
    	    me.x = seed;
    	  } else {
    	    // String seed.
    	    strseed += seed;
    	  }

    	  // Mix in string seed, then discard an initial batch of 64 values.
    	  for (var k = 0; k < strseed.length + 64; k++) {
    	    me.x ^= strseed.charCodeAt(k) | 0;
    	    me.next();
    	  }
    	}

    	function copy(f, t) {
    	  t.x = f.x;
    	  t.y = f.y;
    	  t.z = f.z;
    	  t.w = f.w;
    	  return t;
    	}

    	function impl(seed, opts) {
    	  var xg = new XorGen(seed),
    	      state = opts && opts.state,
    	      prng = function() { return (xg.next() >>> 0) / 0x100000000; };
    	  prng.double = function() {
    	    do {
    	      var top = xg.next() >>> 11,
    	          bot = (xg.next() >>> 0) / 0x100000000,
    	          result = (top + bot) / (1 << 21);
    	    } while (result === 0);
    	    return result;
    	  };
    	  prng.int32 = xg.next;
    	  prng.quick = prng;
    	  if (state) {
    	    if (typeof(state) == 'object') copy(state, xg);
    	    prng.state = function() { return copy(xg, {}); };
    	  }
    	  return prng;
    	}

    	if (module && module.exports) {
    	  module.exports = impl;
    	} else if (define && define.amd) {
    	  define(function() { return impl; });
    	} else {
    	  this.xor128 = impl;
    	}

    	})(
    	  commonjsGlobal,
    	  module,    // present in node.js
    	  (typeof undefined) == 'function'    // present with an AMD loader
    	);
    } (xor128$1));

    var xorwow$1 = {exports: {}};

    (function (module) {
    	// A Javascript implementaion of the "xorwow" prng algorithm by
    	// George Marsaglia.  See http://www.jstatsoft.org/v08/i14/paper

    	(function(global, module, define) {

    	function XorGen(seed) {
    	  var me = this, strseed = '';

    	  // Set up generator function.
    	  me.next = function() {
    	    var t = (me.x ^ (me.x >>> 2));
    	    me.x = me.y; me.y = me.z; me.z = me.w; me.w = me.v;
    	    return (me.d = (me.d + 362437 | 0)) +
    	       (me.v = (me.v ^ (me.v << 4)) ^ (t ^ (t << 1))) | 0;
    	  };

    	  me.x = 0;
    	  me.y = 0;
    	  me.z = 0;
    	  me.w = 0;
    	  me.v = 0;

    	  if (seed === (seed | 0)) {
    	    // Integer seed.
    	    me.x = seed;
    	  } else {
    	    // String seed.
    	    strseed += seed;
    	  }

    	  // Mix in string seed, then discard an initial batch of 64 values.
    	  for (var k = 0; k < strseed.length + 64; k++) {
    	    me.x ^= strseed.charCodeAt(k) | 0;
    	    if (k == strseed.length) {
    	      me.d = me.x << 10 ^ me.x >>> 4;
    	    }
    	    me.next();
    	  }
    	}

    	function copy(f, t) {
    	  t.x = f.x;
    	  t.y = f.y;
    	  t.z = f.z;
    	  t.w = f.w;
    	  t.v = f.v;
    	  t.d = f.d;
    	  return t;
    	}

    	function impl(seed, opts) {
    	  var xg = new XorGen(seed),
    	      state = opts && opts.state,
    	      prng = function() { return (xg.next() >>> 0) / 0x100000000; };
    	  prng.double = function() {
    	    do {
    	      var top = xg.next() >>> 11,
    	          bot = (xg.next() >>> 0) / 0x100000000,
    	          result = (top + bot) / (1 << 21);
    	    } while (result === 0);
    	    return result;
    	  };
    	  prng.int32 = xg.next;
    	  prng.quick = prng;
    	  if (state) {
    	    if (typeof(state) == 'object') copy(state, xg);
    	    prng.state = function() { return copy(xg, {}); };
    	  }
    	  return prng;
    	}

    	if (module && module.exports) {
    	  module.exports = impl;
    	} else if (define && define.amd) {
    	  define(function() { return impl; });
    	} else {
    	  this.xorwow = impl;
    	}

    	})(
    	  commonjsGlobal,
    	  module,    // present in node.js
    	  (typeof undefined) == 'function'    // present with an AMD loader
    	);
    } (xorwow$1));

    var xorshift7$1 = {exports: {}};

    (function (module) {
    	// A Javascript implementaion of the "xorshift7" algorithm by
    	// François Panneton and Pierre L'ecuyer:
    	// "On the Xorgshift Random Number Generators"
    	// http://saluc.engr.uconn.edu/refs/crypto/rng/panneton05onthexorshift.pdf

    	(function(global, module, define) {

    	function XorGen(seed) {
    	  var me = this;

    	  // Set up generator function.
    	  me.next = function() {
    	    // Update xor generator.
    	    var X = me.x, i = me.i, t, v;
    	    t = X[i]; t ^= (t >>> 7); v = t ^ (t << 24);
    	    t = X[(i + 1) & 7]; v ^= t ^ (t >>> 10);
    	    t = X[(i + 3) & 7]; v ^= t ^ (t >>> 3);
    	    t = X[(i + 4) & 7]; v ^= t ^ (t << 7);
    	    t = X[(i + 7) & 7]; t = t ^ (t << 13); v ^= t ^ (t << 9);
    	    X[i] = v;
    	    me.i = (i + 1) & 7;
    	    return v;
    	  };

    	  function init(me, seed) {
    	    var j, X = [];

    	    if (seed === (seed | 0)) {
    	      // Seed state array using a 32-bit integer.
    	      X[0] = seed;
    	    } else {
    	      // Seed state using a string.
    	      seed = '' + seed;
    	      for (j = 0; j < seed.length; ++j) {
    	        X[j & 7] = (X[j & 7] << 15) ^
    	            (seed.charCodeAt(j) + X[(j + 1) & 7] << 13);
    	      }
    	    }
    	    // Enforce an array length of 8, not all zeroes.
    	    while (X.length < 8) X.push(0);
    	    for (j = 0; j < 8 && X[j] === 0; ++j);
    	    if (j == 8) X[7] = -1;

    	    me.x = X;
    	    me.i = 0;

    	    // Discard an initial 256 values.
    	    for (j = 256; j > 0; --j) {
    	      me.next();
    	    }
    	  }

    	  init(me, seed);
    	}

    	function copy(f, t) {
    	  t.x = f.x.slice();
    	  t.i = f.i;
    	  return t;
    	}

    	function impl(seed, opts) {
    	  if (seed == null) seed = +(new Date);
    	  var xg = new XorGen(seed),
    	      state = opts && opts.state,
    	      prng = function() { return (xg.next() >>> 0) / 0x100000000; };
    	  prng.double = function() {
    	    do {
    	      var top = xg.next() >>> 11,
    	          bot = (xg.next() >>> 0) / 0x100000000,
    	          result = (top + bot) / (1 << 21);
    	    } while (result === 0);
    	    return result;
    	  };
    	  prng.int32 = xg.next;
    	  prng.quick = prng;
    	  if (state) {
    	    if (state.x) copy(state, xg);
    	    prng.state = function() { return copy(xg, {}); };
    	  }
    	  return prng;
    	}

    	if (module && module.exports) {
    	  module.exports = impl;
    	} else if (define && define.amd) {
    	  define(function() { return impl; });
    	} else {
    	  this.xorshift7 = impl;
    	}

    	})(
    	  commonjsGlobal,
    	  module,    // present in node.js
    	  (typeof undefined) == 'function'    // present with an AMD loader
    	);
    } (xorshift7$1));

    var xor4096$1 = {exports: {}};

    (function (module) {
    	// A Javascript implementaion of Richard Brent's Xorgens xor4096 algorithm.
    	//
    	// This fast non-cryptographic random number generator is designed for
    	// use in Monte-Carlo algorithms. It combines a long-period xorshift
    	// generator with a Weyl generator, and it passes all common batteries
    	// of stasticial tests for randomness while consuming only a few nanoseconds
    	// for each prng generated.  For background on the generator, see Brent's
    	// paper: "Some long-period random number generators using shifts and xors."
    	// http://arxiv.org/pdf/1004.3115v1.pdf
    	//
    	// Usage:
    	//
    	// var xor4096 = require('xor4096');
    	// random = xor4096(1);                        // Seed with int32 or string.
    	// assert.equal(random(), 0.1520436450538547); // (0, 1) range, 53 bits.
    	// assert.equal(random.int32(), 1806534897);   // signed int32, 32 bits.
    	//
    	// For nonzero numeric keys, this impelementation provides a sequence
    	// identical to that by Brent's xorgens 3 implementaion in C.  This
    	// implementation also provides for initalizing the generator with
    	// string seeds, or for saving and restoring the state of the generator.
    	//
    	// On Chrome, this prng benchmarks about 2.1 times slower than
    	// Javascript's built-in Math.random().

    	(function(global, module, define) {

    	function XorGen(seed) {
    	  var me = this;

    	  // Set up generator function.
    	  me.next = function() {
    	    var w = me.w,
    	        X = me.X, i = me.i, t, v;
    	    // Update Weyl generator.
    	    me.w = w = (w + 0x61c88647) | 0;
    	    // Update xor generator.
    	    v = X[(i + 34) & 127];
    	    t = X[i = ((i + 1) & 127)];
    	    v ^= v << 13;
    	    t ^= t << 17;
    	    v ^= v >>> 15;
    	    t ^= t >>> 12;
    	    // Update Xor generator array state.
    	    v = X[i] = v ^ t;
    	    me.i = i;
    	    // Result is the combination.
    	    return (v + (w ^ (w >>> 16))) | 0;
    	  };

    	  function init(me, seed) {
    	    var t, v, i, j, w, X = [], limit = 128;
    	    if (seed === (seed | 0)) {
    	      // Numeric seeds initialize v, which is used to generates X.
    	      v = seed;
    	      seed = null;
    	    } else {
    	      // String seeds are mixed into v and X one character at a time.
    	      seed = seed + '\0';
    	      v = 0;
    	      limit = Math.max(limit, seed.length);
    	    }
    	    // Initialize circular array and weyl value.
    	    for (i = 0, j = -32; j < limit; ++j) {
    	      // Put the unicode characters into the array, and shuffle them.
    	      if (seed) v ^= seed.charCodeAt((j + 32) % seed.length);
    	      // After 32 shuffles, take v as the starting w value.
    	      if (j === 0) w = v;
    	      v ^= v << 10;
    	      v ^= v >>> 15;
    	      v ^= v << 4;
    	      v ^= v >>> 13;
    	      if (j >= 0) {
    	        w = (w + 0x61c88647) | 0;     // Weyl.
    	        t = (X[j & 127] ^= (v + w));  // Combine xor and weyl to init array.
    	        i = (0 == t) ? i + 1 : 0;     // Count zeroes.
    	      }
    	    }
    	    // We have detected all zeroes; make the key nonzero.
    	    if (i >= 128) {
    	      X[(seed && seed.length || 0) & 127] = -1;
    	    }
    	    // Run the generator 512 times to further mix the state before using it.
    	    // Factoring this as a function slows the main generator, so it is just
    	    // unrolled here.  The weyl generator is not advanced while warming up.
    	    i = 127;
    	    for (j = 4 * 128; j > 0; --j) {
    	      v = X[(i + 34) & 127];
    	      t = X[i = ((i + 1) & 127)];
    	      v ^= v << 13;
    	      t ^= t << 17;
    	      v ^= v >>> 15;
    	      t ^= t >>> 12;
    	      X[i] = v ^ t;
    	    }
    	    // Storing state as object members is faster than using closure variables.
    	    me.w = w;
    	    me.X = X;
    	    me.i = i;
    	  }

    	  init(me, seed);
    	}

    	function copy(f, t) {
    	  t.i = f.i;
    	  t.w = f.w;
    	  t.X = f.X.slice();
    	  return t;
    	}
    	function impl(seed, opts) {
    	  if (seed == null) seed = +(new Date);
    	  var xg = new XorGen(seed),
    	      state = opts && opts.state,
    	      prng = function() { return (xg.next() >>> 0) / 0x100000000; };
    	  prng.double = function() {
    	    do {
    	      var top = xg.next() >>> 11,
    	          bot = (xg.next() >>> 0) / 0x100000000,
    	          result = (top + bot) / (1 << 21);
    	    } while (result === 0);
    	    return result;
    	  };
    	  prng.int32 = xg.next;
    	  prng.quick = prng;
    	  if (state) {
    	    if (state.X) copy(state, xg);
    	    prng.state = function() { return copy(xg, {}); };
    	  }
    	  return prng;
    	}

    	if (module && module.exports) {
    	  module.exports = impl;
    	} else if (define && define.amd) {
    	  define(function() { return impl; });
    	} else {
    	  this.xor4096 = impl;
    	}

    	})(
    	  commonjsGlobal,                                     // window object or global
    	  module,    // present in node.js
    	  (typeof undefined) == 'function'    // present with an AMD loader
    	);
    } (xor4096$1));

    var tychei$1 = {exports: {}};

    (function (module) {
    	// A Javascript implementaion of the "Tyche-i" prng algorithm by
    	// Samuel Neves and Filipe Araujo.
    	// See https://eden.dei.uc.pt/~sneves/pubs/2011-snfa2.pdf

    	(function(global, module, define) {

    	function XorGen(seed) {
    	  var me = this, strseed = '';

    	  // Set up generator function.
    	  me.next = function() {
    	    var b = me.b, c = me.c, d = me.d, a = me.a;
    	    b = (b << 25) ^ (b >>> 7) ^ c;
    	    c = (c - d) | 0;
    	    d = (d << 24) ^ (d >>> 8) ^ a;
    	    a = (a - b) | 0;
    	    me.b = b = (b << 20) ^ (b >>> 12) ^ c;
    	    me.c = c = (c - d) | 0;
    	    me.d = (d << 16) ^ (c >>> 16) ^ a;
    	    return me.a = (a - b) | 0;
    	  };

    	  /* The following is non-inverted tyche, which has better internal
    	   * bit diffusion, but which is about 25% slower than tyche-i in JS.
    	  me.next = function() {
    	    var a = me.a, b = me.b, c = me.c, d = me.d;
    	    a = (me.a + me.b | 0) >>> 0;
    	    d = me.d ^ a; d = d << 16 ^ d >>> 16;
    	    c = me.c + d | 0;
    	    b = me.b ^ c; b = b << 12 ^ d >>> 20;
    	    me.a = a = a + b | 0;
    	    d = d ^ a; me.d = d = d << 8 ^ d >>> 24;
    	    me.c = c = c + d | 0;
    	    b = b ^ c;
    	    return me.b = (b << 7 ^ b >>> 25);
    	  }
    	  */

    	  me.a = 0;
    	  me.b = 0;
    	  me.c = 2654435769 | 0;
    	  me.d = 1367130551;

    	  if (seed === Math.floor(seed)) {
    	    // Integer seed.
    	    me.a = (seed / 0x100000000) | 0;
    	    me.b = seed | 0;
    	  } else {
    	    // String seed.
    	    strseed += seed;
    	  }

    	  // Mix in string seed, then discard an initial batch of 64 values.
    	  for (var k = 0; k < strseed.length + 20; k++) {
    	    me.b ^= strseed.charCodeAt(k) | 0;
    	    me.next();
    	  }
    	}

    	function copy(f, t) {
    	  t.a = f.a;
    	  t.b = f.b;
    	  t.c = f.c;
    	  t.d = f.d;
    	  return t;
    	}
    	function impl(seed, opts) {
    	  var xg = new XorGen(seed),
    	      state = opts && opts.state,
    	      prng = function() { return (xg.next() >>> 0) / 0x100000000; };
    	  prng.double = function() {
    	    do {
    	      var top = xg.next() >>> 11,
    	          bot = (xg.next() >>> 0) / 0x100000000,
    	          result = (top + bot) / (1 << 21);
    	    } while (result === 0);
    	    return result;
    	  };
    	  prng.int32 = xg.next;
    	  prng.quick = prng;
    	  if (state) {
    	    if (typeof(state) == 'object') copy(state, xg);
    	    prng.state = function() { return copy(xg, {}); };
    	  }
    	  return prng;
    	}

    	if (module && module.exports) {
    	  module.exports = impl;
    	} else if (define && define.amd) {
    	  define(function() { return impl; });
    	} else {
    	  this.tychei = impl;
    	}

    	})(
    	  commonjsGlobal,
    	  module,    // present in node.js
    	  (typeof undefined) == 'function'    // present with an AMD loader
    	);
    } (tychei$1));

    var seedrandom$1 = {exports: {}};

    var empty = {};

    var empty$1 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        'default': empty
    });

    var require$$6 = /*@__PURE__*/getAugmentedNamespace(empty$1);

    /*
    Copyright 2014 David Bau.

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    */

    (function (module) {
    	(function (pool, math) {
    	//
    	// The following constants are related to IEEE 754 limits.
    	//
    	var global = this,
    	    width = 256,        // each RC4 output is 0 <= x < 256
    	    chunks = 6,         // at least six RC4 outputs for each double
    	    digits = 52,        // there are 52 significant digits in a double
    	    rngname = 'random', // rngname: name for Math.random and Math.seedrandom
    	    startdenom = math.pow(width, chunks),
    	    significance = math.pow(2, digits),
    	    overflow = significance * 2,
    	    mask = width - 1,
    	    nodecrypto;         // node.js crypto module, initialized at the bottom.

    	//
    	// seedrandom()
    	// This is the seedrandom function described above.
    	//
    	function seedrandom(seed, options, callback) {
    	  var key = [];
    	  options = (options == true) ? { entropy: true } : (options || {});

    	  // Flatten the seed string or build one from local entropy if needed.
    	  var shortseed = mixkey(flatten(
    	    options.entropy ? [seed, tostring(pool)] :
    	    (seed == null) ? autoseed() : seed, 3), key);

    	  // Use the seed to initialize an ARC4 generator.
    	  var arc4 = new ARC4(key);

    	  // This function returns a random double in [0, 1) that contains
    	  // randomness in every bit of the mantissa of the IEEE 754 value.
    	  var prng = function() {
    	    var n = arc4.g(chunks),             // Start with a numerator n < 2 ^ 48
    	        d = startdenom,                 //   and denominator d = 2 ^ 48.
    	        x = 0;                          //   and no 'extra last byte'.
    	    while (n < significance) {          // Fill up all significant digits by
    	      n = (n + x) * width;              //   shifting numerator and
    	      d *= width;                       //   denominator and generating a
    	      x = arc4.g(1);                    //   new least-significant-byte.
    	    }
    	    while (n >= overflow) {             // To avoid rounding up, before adding
    	      n /= 2;                           //   last byte, shift everything
    	      d /= 2;                           //   right using integer math until
    	      x >>>= 1;                         //   we have exactly the desired bits.
    	    }
    	    return (n + x) / d;                 // Form the number within [0, 1).
    	  };

    	  prng.int32 = function() { return arc4.g(4) | 0; };
    	  prng.quick = function() { return arc4.g(4) / 0x100000000; };
    	  prng.double = prng;

    	  // Mix the randomness into accumulated entropy.
    	  mixkey(tostring(arc4.S), pool);

    	  // Calling convention: what to return as a function of prng, seed, is_math.
    	  return (options.pass || callback ||
    	      function(prng, seed, is_math_call, state) {
    	        if (state) {
    	          // Load the arc4 state from the given state if it has an S array.
    	          if (state.S) { copy(state, arc4); }
    	          // Only provide the .state method if requested via options.state.
    	          prng.state = function() { return copy(arc4, {}); };
    	        }

    	        // If called as a method of Math (Math.seedrandom()), mutate
    	        // Math.random because that is how seedrandom.js has worked since v1.0.
    	        if (is_math_call) { math[rngname] = prng; return seed; }

    	        // Otherwise, it is a newer calling convention, so return the
    	        // prng directly.
    	        else return prng;
    	      })(
    	  prng,
    	  shortseed,
    	  'global' in options ? options.global : (this == math),
    	  options.state);
    	}
    	math['seed' + rngname] = seedrandom;

    	//
    	// ARC4
    	//
    	// An ARC4 implementation.  The constructor takes a key in the form of
    	// an array of at most (width) integers that should be 0 <= x < (width).
    	//
    	// The g(count) method returns a pseudorandom integer that concatenates
    	// the next (count) outputs from ARC4.  Its return value is a number x
    	// that is in the range 0 <= x < (width ^ count).
    	//
    	function ARC4(key) {
    	  var t, keylen = key.length,
    	      me = this, i = 0, j = me.i = me.j = 0, s = me.S = [];

    	  // The empty key [] is treated as [0].
    	  if (!keylen) { key = [keylen++]; }

    	  // Set up S using the standard key scheduling algorithm.
    	  while (i < width) {
    	    s[i] = i++;
    	  }
    	  for (i = 0; i < width; i++) {
    	    s[i] = s[j = mask & (j + key[i % keylen] + (t = s[i]))];
    	    s[j] = t;
    	  }

    	  // The "g" method returns the next (count) outputs as one number.
    	  (me.g = function(count) {
    	    // Using instance members instead of closure state nearly doubles speed.
    	    var t, r = 0,
    	        i = me.i, j = me.j, s = me.S;
    	    while (count--) {
    	      t = s[i = mask & (i + 1)];
    	      r = r * width + s[mask & ((s[i] = s[j = mask & (j + t)]) + (s[j] = t))];
    	    }
    	    me.i = i; me.j = j;
    	    return r;
    	    // For robust unpredictability, the function call below automatically
    	    // discards an initial batch of values.  This is called RC4-drop[256].
    	    // See http://google.com/search?q=rsa+fluhrer+response&btnI
    	  })(width);
    	}

    	//
    	// copy()
    	// Copies internal state of ARC4 to or from a plain object.
    	//
    	function copy(f, t) {
    	  t.i = f.i;
    	  t.j = f.j;
    	  t.S = f.S.slice();
    	  return t;
    	}
    	//
    	// flatten()
    	// Converts an object tree to nested arrays of strings.
    	//
    	function flatten(obj, depth) {
    	  var result = [], typ = (typeof obj), prop;
    	  if (depth && typ == 'object') {
    	    for (prop in obj) {
    	      try { result.push(flatten(obj[prop], depth - 1)); } catch (e) {}
    	    }
    	  }
    	  return (result.length ? result : typ == 'string' ? obj : obj + '\0');
    	}

    	//
    	// mixkey()
    	// Mixes a string seed into a key that is an array of integers, and
    	// returns a shortened string seed that is equivalent to the result key.
    	//
    	function mixkey(seed, key) {
    	  var stringseed = seed + '', smear, j = 0;
    	  while (j < stringseed.length) {
    	    key[mask & j] =
    	      mask & ((smear ^= key[mask & j] * 19) + stringseed.charCodeAt(j++));
    	  }
    	  return tostring(key);
    	}

    	//
    	// autoseed()
    	// Returns an object for autoseeding, using window.crypto and Node crypto
    	// module if available.
    	//
    	function autoseed() {
    	  try {
    	    var out;
    	    if (nodecrypto && (out = nodecrypto.randomBytes)) {
    	      // The use of 'out' to remember randomBytes makes tight minified code.
    	      out = out(width);
    	    } else {
    	      out = new Uint8Array(width);
    	      (global.crypto || global.msCrypto).getRandomValues(out);
    	    }
    	    return tostring(out);
    	  } catch (e) {
    	    var browser = global.navigator,
    	        plugins = browser && browser.plugins;
    	    return [+new Date, global, plugins, global.screen, tostring(pool)];
    	  }
    	}

    	//
    	// tostring()
    	// Converts an array of charcodes to a string
    	//
    	function tostring(a) {
    	  return String.fromCharCode.apply(0, a);
    	}

    	//
    	// When seedrandom.js is loaded, we immediately mix a few bits
    	// from the built-in RNG into the entropy pool.  Because we do
    	// not want to interfere with deterministic PRNG state later,
    	// seedrandom will not call math.random on its own again after
    	// initialization.
    	//
    	mixkey(math.random(), pool);

    	//
    	// Nodejs and AMD support: export the implementation as a module using
    	// either convention.
    	//
    	if (module.exports) {
    	  module.exports = seedrandom;
    	  // When in node.js, try using crypto package for autoseeding.
    	  try {
    	    nodecrypto = require$$6;
    	  } catch (ex) {}
    	}

    	// End anonymous scope, and pass initial values.
    	})(
    	  [],     // pool: entropy pool starts empty
    	  Math    // math: package containing random, pow, and seedrandom
    	);
    } (seedrandom$1));

    // A library of seedable RNGs implemented in Javascript.
    //
    // Usage:
    //
    // var seedrandom = require('seedrandom');
    // var random = seedrandom(1); // or any seed.
    // var x = random();       // 0 <= x < 1.  Every bit is random.
    // var x = random.quick(); // 0 <= x < 1.  32 bits of randomness.

    // alea, a 53-bit multiply-with-carry generator by Johannes Baagøe.
    // Period: ~2^116
    // Reported to pass all BigCrush tests.
    var alea = alea$1.exports;

    // xor128, a pure xor-shift generator by George Marsaglia.
    // Period: 2^128-1.
    // Reported to fail: MatrixRank and LinearComp.
    var xor128 = xor128$1.exports;

    // xorwow, George Marsaglia's 160-bit xor-shift combined plus weyl.
    // Period: 2^192-2^32
    // Reported to fail: CollisionOver, SimpPoker, and LinearComp.
    var xorwow = xorwow$1.exports;

    // xorshift7, by François Panneton and Pierre L'ecuyer, takes
    // a different approach: it adds robustness by allowing more shifts
    // than Marsaglia's original three.  It is a 7-shift generator
    // with 256 bits, that passes BigCrush with no systmatic failures.
    // Period 2^256-1.
    // No systematic BigCrush failures reported.
    var xorshift7 = xorshift7$1.exports;

    // xor4096, by Richard Brent, is a 4096-bit xor-shift with a
    // very long period that also adds a Weyl generator. It also passes
    // BigCrush with no systematic failures.  Its long period may
    // be useful if you have many generators and need to avoid
    // collisions.
    // Period: 2^4128-2^32.
    // No systematic BigCrush failures reported.
    var xor4096 = xor4096$1.exports;

    // Tyche-i, by Samuel Neves and Filipe Araujo, is a bit-shifting random
    // number generator derived from ChaCha, a modern stream cipher.
    // https://eden.dei.uc.pt/~sneves/pubs/2011-snfa2.pdf
    // Period: ~2^127
    // No systematic BigCrush failures reported.
    var tychei = tychei$1.exports;

    // The original ARC4-based prng included in this library.
    // Period: ~2^1600
    var sr = seedrandom$1.exports;

    sr.alea = alea;
    sr.xor128 = xor128;
    sr.xorwow = xorwow;
    sr.xorshift7 = xorshift7;
    sr.xor4096 = xor4096;
    sr.tychei = tychei;

    var seedrandom = sr;

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // https://en.wikipedia.org/wiki/Marsaglia_polar_method
    class MPRandGauss {
        constructor(mean, stdDeviation, dtype, truncated, seed) {
            this.mean = mean;
            this.stdDev = stdDeviation;
            this.dtype = dtype;
            this.nextVal = NaN;
            this.truncated = truncated;
            if (this.truncated) {
                this.upper = this.mean + this.stdDev * 2;
                this.lower = this.mean - this.stdDev * 2;
            }
            const seedValue = seed ? seed : Math.random();
            this.random = seedrandom.alea(seedValue.toString());
        }
        /** Returns next sample from a Gaussian distribution. */
        nextValue() {
            if (!isNaN(this.nextVal)) {
                const value = this.nextVal;
                this.nextVal = NaN;
                return value;
            }
            let resultX, resultY;
            let isValid = false;
            while (!isValid) {
                let v1, v2, s;
                do {
                    v1 = 2 * this.random() - 1;
                    v2 = 2 * this.random() - 1;
                    s = v1 * v1 + v2 * v2;
                } while (s >= 1 || s === 0);
                const mul = Math.sqrt(-2.0 * Math.log(s) / s);
                resultX = this.mean + this.stdDev * v1 * mul;
                resultY = this.mean + this.stdDev * v2 * mul;
                if (!this.truncated || this.isValidTruncated(resultX)) {
                    isValid = true;
                }
            }
            if (!this.truncated || this.isValidTruncated(resultY)) {
                this.nextVal = this.convertValue(resultY);
            }
            return this.convertValue(resultX);
        }
        /** Handles proper rounding for non-floating-point numbers. */
        convertValue(value) {
            if (this.dtype == null || this.dtype === 'float32') {
                return value;
            }
            return Math.round(value);
        }
        /** Returns true if less than 2-standard-deviations from the mean. */
        isValidTruncated(value) {
            return value <= this.upper && value >= this.lower;
        }
    }
    // Marsaglia, George, and Wai Wan Tsang. 2000. "A Simple Method for Generating
    // Gamma Variables."
    class RandGamma {
        constructor(alpha, beta, dtype, seed) {
            this.alpha = alpha;
            this.beta = 1 / beta; // convert rate to scale parameter
            this.dtype = dtype;
            const seedValue = seed ? seed : Math.random();
            this.randu = seedrandom.alea(seedValue.toString());
            this.randn = new MPRandGauss(0, 1, dtype, false, this.randu());
            if (alpha < 1) {
                this.d = alpha + (2 / 3);
            }
            else {
                this.d = alpha - (1 / 3);
            }
            this.c = 1 / Math.sqrt(9 * this.d);
        }
        /** Returns next sample from a gamma distribution. */
        nextValue() {
            let x2, v0, v1, x, u, v;
            while (true) {
                do {
                    x = this.randn.nextValue();
                    v = 1 + (this.c * x);
                } while (v <= 0);
                v *= v * v;
                x2 = x * x;
                v0 = 1 - (0.331 * x2 * x2);
                v1 = (0.5 * x2) + (this.d * (1 - v + Math.log(v)));
                u = this.randu();
                if (u < v0 || Math.log(u) < v1) {
                    break;
                }
            }
            v = (1 / this.beta) * this.d * v;
            if (this.alpha < 1) {
                v *= Math.pow(this.randu(), 1 / this.alpha);
            }
            return this.convertValue(v);
        }
        /** Handles proper rounding for non-floating-point numbers. */
        convertValue(value) {
            if (this.dtype === 'float32') {
                return value;
            }
            return Math.round(value);
        }
    }
    class UniformRandom {
        constructor(min = 0, max = 1, dtype, seed) {
            /** Handles proper rounding for non floating point numbers. */
            this.canReturnFloat = () => (this.dtype == null || this.dtype === 'float32');
            this.min = min;
            this.range = max - min;
            this.dtype = dtype;
            if (seed == null) {
                seed = Math.random();
            }
            if (typeof seed === 'number') {
                seed = seed.toString();
            }
            if (!this.canReturnFloat() && this.range <= 1) {
                throw new Error(`The difference between ${min} - ${max} <= 1 and dtype is not float`);
            }
            this.random = seedrandom.alea(seed);
        }
        convertValue(value) {
            if (this.canReturnFloat()) {
                return value;
            }
            return Math.round(value);
        }
        nextValue() {
            return this.convertValue(this.min + this.range * this.random());
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with values sampled from a gamma distribution.
     *
     * ```js
     * tf.randomGamma([2, 2], 1).print();
     * ```
     *
     * @param shape An array of integers defining the output tensor shape.
     * @param alpha The shape parameter of the gamma distribution.
     * @param beta The inverse scale parameter of the gamma distribution. Defaults
     *     to 1.
     * @param dtype The data type of the output. Defaults to float32.
     * @param seed The seed for the random number generator.
     *
     * @doc {heading: 'Tensors', subheading: 'Random'}
     */
    function randomGamma_(shape, alpha, beta = 1, dtype = 'float32', seed) {
        if (beta == null) {
            beta = 1;
        }
        if (dtype == null) {
            dtype = 'float32';
        }
        if (dtype !== 'float32' && dtype !== 'int32') {
            throw new Error(`Unsupported data type ${dtype}`);
        }
        const rgamma = new RandGamma(alpha, beta, dtype, seed);
        const res = buffer(shape, dtype);
        for (let i = 0; i < res.values.length; i++) {
            res.values[i] = rgamma.nextValue();
        }
        return res.toTensor();
    }
    const randomGamma = op({ randomGamma_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with values sampled from a normal distribution.
     *
     * ```js
     * tf.randomNormal([2, 2]).print();
     * ```
     *
     * @param shape An array of integers defining the output tensor shape.
     * @param mean The mean of the normal distribution.
     * @param stdDev The standard deviation of the normal distribution.
     * @param dtype The data type of the output.
     * @param seed The seed for the random number generator.
     *
     * @doc {heading: 'Tensors', subheading: 'Random'}
     */
    function randomNormal_(shape, mean = 0, stdDev = 1, dtype, seed) {
        if (dtype != null && dtype === 'bool') {
            throw new Error(`Unsupported data type ${dtype}`);
        }
        const randGauss = new MPRandGauss(mean, stdDev, dtype, false /* truncated */, seed);
        const res = buffer(shape, dtype);
        for (let i = 0; i < res.values.length; i++) {
            res.values[i] = randGauss.nextValue();
        }
        return res.toTensor();
    }
    const randomNormal$2 = op({ randomNormal_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with values sampled from a uniform distribution.
     *
     * The generated values follow a uniform distribution in the range [minval,
     * maxval). The lower bound minval is included in the range, while the upper
     * bound maxval is excluded.
     *
     * ```js
     * tf.randomUniform([2, 2]).print();
     * ```
     *
     * @param shape An array of integers defining the output tensor shape.
     * @param minval The lower bound on the range of random values to generate.
     *   Defaults to 0.
     * @param maxval The upper bound on the range of random values to generate.
     *   Defaults to 1.
     * @param dtype The data type of the output tensor. Defaults to 'float32'.
     *
     * @doc {heading: 'Tensors', subheading: 'Random'}
     */
    function randomUniform_(shape, minval = 0, maxval = 1, dtype = 'float32', seed) {
        const res = buffer(shape, dtype);
        const random = new UniformRandom(minval, maxval, null, seed);
        for (let i = 0; i < res.values.length; i++) {
            res.values[i] = random.nextValue();
        }
        return res.toTensor();
    }
    const randomUniform$1 = op({ randomUniform_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a new `tf.Tensor1D` filled with the numbers in the range provided.
     *
     * The tensor is a is half-open interval meaning it includes start, but
     * excludes stop. Decrementing ranges and negative step values are also
     * supported.sv
     *
     *
     * ```js
     * tf.range(0, 9, 2).print();
     * ```
     *
     * @param start An integer start value
     * @param stop An integer stop value
     * @param step An integer increment (will default to 1 or -1)
     * @param dtype The data type of the output tensor. Defaults to 'float32'.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function range$4(start, stop, step = 1, dtype = 'float32') {
        if (step === 0) {
            throw new Error('Cannot have a step of zero');
        }
        const attrs = { start, stop, step, dtype };
        return ENGINE.runKernel(Range, {} /* inputs */, attrs);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the real part of a complex (or real) tensor.
     *
     * Given a tensor input, this operation returns a tensor of type float that is
     * the real part of each element in input considered as a complex number.
     *
     * If the input is real, it simply makes a clone.
     *
     * ```js
     * const x = tf.complex([-2.25, 3.25], [4.75, 5.75]);
     * tf.real(x).print();
     * ```
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function real_(input) {
        const $input = convertToTensor(input, 'input', 'real');
        const inputs = { input: $input };
        return ENGINE.runKernel(Real, inputs);
    }
    const real$2 = op({ real_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes reciprocal of x element-wise: `1 / x`
     *
     * ```js
     * const x = tf.tensor1d([0, 1, 2]);
     *
     * x.reciprocal().print();  // or tf.reciprocal(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function reciprocal_(x) {
        const $x = convertToTensor(x, 'x', 'reciprocal');
        const inputs = { x: $x };
        return ENGINE.runKernel(Reciprocal, inputs);
    }
    const reciprocal$2 = op({ reciprocal_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes rectified linear element-wise: `max(x, 0)`.
     *
     * ```js
     * const x = tf.tensor1d([-1, 2, -3, 4]);
     *
     * x.relu().print();  // or tf.relu(x)
     * ```
     * @param x The input tensor. If the dtype is `bool`, the output dtype will be
     *     `int32'.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function relu_(x) {
        const $x = convertToTensor(x, 'x', 'relu');
        const inputs = { x: $x };
        return ENGINE.runKernel(Relu$1, inputs);
    }
    const relu$2 = op({ relu_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes rectified linear 6 element-wise: `min(max(x, 0), 6)`.
     *
     * ```js
     * const x = tf.tensor1d([-1, 2, -3, 8]);
     *
     * x.relu6().print();  // or tf.relu6(x)
     * ```
     * @param x The input tensor. If the dtype is `bool`, the output dtype will be
     *     `int32'.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function relu6_(x) {
        const $x = convertToTensor(x, 'x', 'relu6');
        const inputs = { x: $x };
        return ENGINE.runKernel(Relu6$1, inputs);
    }
    const relu6$2 = op({ relu6_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Reverses a `tf.Tensor` along a specified axis.
     *
     * Also available are stricter rank-specific methods that assert that `x` is
     * of the given rank:
     *   - `tf.reverse1d`
     *   - `tf.reverse2d`
     *   - `tf.reverse3d`
     *   - `tf.reverse4d`
     *
     * Except `tf.reverse1d` (which does not have axis param), all methods have
     * same signature as this method.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     *
     * x.reverse().print();
     * ```
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * const axis = 1;
     * x.reverse(axis).print();
     * ```
     * @param x The input tensor to be reversed.
     * @param axis The set of dimensions to reverse. Must be in the
     *     range [-rank(x), rank(x)). Defaults to all axes.
     *
     * @doc {heading: 'Tensors', subheading: 'Slicing and Joining'}
     */
    function reverse_(x, axis) {
        const $x = convertToTensor(x, 'x', 'reverse');
        const inputs = { x: $x };
        const attrs = { dims: axis };
        return ENGINE.runKernel(Reverse, inputs, attrs);
    }
    const reverse$2 = op({ reverse_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Reverses a `tf.Tensor1D`.
     *
     * @param x The input tensor.
     */
    function reverse1d_(x) {
        const $x = convertToTensor(x, 'x', 'reverse');
        assert$4($x.rank === 1, () => `Error in reverse1D: x must be rank 1 but got rank ${$x.rank}.`);
        return reverse$2($x, 0);
    }
    const reverse1d = op({ reverse1d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Reverses a `tf.Tensor2D` along a specified axis.
     *
     * @param x The input tensor.
     * @param axis The set of dimensions to reverse. Must be in the
     *     range [-rank(x), rank(x)). Defaults to all axes.
     */
    function reverse2d_(x, axis) {
        const $x = convertToTensor(x, 'x', 'reverse');
        assert$4($x.rank === 2, () => `Error in reverse2D: x must be rank 2 but got rank ${$x.rank}.`);
        return reverse$2($x, axis);
    }
    const reverse2d = op({ reverse2d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Reverses a `tf.Tensor3D` along a specified axis.
     *
     * @param x The input tensor.
     * @param axis The set of dimensions to reverse. Must be in the
     *     range [-rank(x), rank(x)). Defaults to all axes.
     */
    function reverse3d_(x, axis) {
        const $x = convertToTensor(x, 'x', 'reverse');
        assert$4($x.rank === 3, () => `Error in reverse3D: x must be rank 3 but got rank ${$x.rank}.`);
        return reverse$2($x, axis);
    }
    const reverse3d = op({ reverse3d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Reverses a `tf.Tensor4D` along a specified axis.
     *
     * @param x The input tensor.
     * @param axis The set of dimensions to reverse. Must be in the
     *     range [-rank(x), rank(x)). Defaults to all axes.
     */
    function reverse4d_(x, axis) {
        const $x = convertToTensor(x, 'x', 'reverse');
        assert$4($x.rank === 4, () => `Error in reverse4D: x must be rank 4 but got rank ${$x.rank}.`);
        return reverse$2($x, axis);
    }
    const reverse4d = op({ reverse4d_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes round of input `tf.Tensor` element-wise: `round(x)`.
     * It implements banker's rounding.
     *
     * ```js
     * const x = tf.tensor1d([.6, 1.1, -3.3]);
     *
     * x.round().print();  // or tf.round(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function round_(x) {
        const $x = convertToTensor(x, 'x', 'round');
        const inputs = { x: $x };
        return ENGINE.runKernel(Round, inputs);
    }
    const round$2 = op({ round_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes reciprocal of square root of the input `tf.Tensor` element-wise:
     * `y = 1 / sqrt(x)`
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 4, -1]);
     *
     * x.rsqrt().print();  // or tf.rsqrt(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function rsqrt_(x) {
        const $x = convertToTensor(x, 'x', 'rsqrt', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Rsqrt, inputs);
    }
    const rsqrt$2 = op({ rsqrt_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes scaled exponential linear element-wise.
     *
     * `x < 0 ? scale * alpha * (exp(x) - 1) : x`
     *
     * ```js
     * const x = tf.tensor1d([-1, 2, -3, 4]);
     *
     * x.selu().print();  // or tf.selu(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function selu_(x) {
        const $x = convertToTensor(x, 'x', 'selu');
        const inputs = { x: $x };
        return ENGINE.runKernel(Selu$1, inputs);
    }
    const selu$2 = op({ selu_ });

    /**
     * 2-D convolution with separable filters.
     *
     * Performs a depthwise convolution that acts separately on channels followed
     * by a pointwise convolution that mixes channels. Note that this is
     * separability between dimensions [1, 2] and 3, not spatial separability
     * between dimensions 1 and 2.
     *
     * See
     * [https://www.tensorflow.org/api_docs/python/tf/nn/separable_conv2d](
     *     https://www.tensorflow.org/api_docs/python/tf/nn/separable_conv2d)
     * for more details.
     *
     * @param x The input tensor, of rank 4 or rank 3, of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is
     * assumed.
     * @param depthwiseFilter The depthwise filter tensor, rank 4, of shape
     *     `[filterHeight, filterWidth, inChannels, channelMultiplier]`. This is
     *     the filter used in the first step.
     * @param pointwiseFilter The pointwise filter tensor, rank 4, of shape
     *     `[1, 1, inChannels * channelMultiplier, outChannels]`. This is
     *     the filter used in the second step.
     * @param strides The strides of the convolution: `[strideHeight,
     * strideWidth]`. If strides is a single number, then `strideHeight ==
     * strideWidth`.
     * @param pad The type of padding algorithm.
     *   - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *   - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *   - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dilations The dilation rates: `[dilationHeight, dilationWidth]`
     *     in which we sample input values across the height and width dimensions
     *     in atrous convolution. Defaults to `[1, 1]`. If `rate` is a single
     *     number, then `dilationHeight == dilationWidth`. If it is greater than
     *     1, then all values of `strides` must be 1.
     * @param dataFormat: An optional string from: "NHWC", "NCHW". Defaults to
     *     "NHWC". Specify the data format of the input and output data. With the
     *     default format "NHWC", the data is stored in the order of: [batch,
     *     height, width, channels]. Only "NHWC" is currently supported.
     *
     * @doc {heading: 'Operations', subheading: 'Convolution'}
     */
    function separableConv2d_(x, depthwiseFilter, pointwiseFilter, strides, pad, dilation = [1, 1], dataFormat = 'NHWC') {
        const $x = convertToTensor(x, 'x', 'separableConv2d');
        const $depthwiseFilter = convertToTensor(depthwiseFilter, 'depthwiseFilter', 'separableConv2d');
        const $pointwiseFilter = convertToTensor(pointwiseFilter, 'pointwiseFilter', 'separableConv2d');
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            reshapedTo4D = true;
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
        }
        if (dataFormat === 'NCHW') {
            throw new Error('separableConv2d currently does not support dataFormat NCHW; only ' +
                'NHWC is supported');
        }
        assert$4(x4D.rank === 4, () => `Error in separableConv2d: input must be rank 4, but got ` +
            `rank ${x4D.rank}.`);
        assert$4($depthwiseFilter.rank === 4, () => `Error in separableConv2d: depthwise filter must be rank 4, but ` +
            `got rank ${$depthwiseFilter.rank}.`);
        assert$4($pointwiseFilter.rank === 4, () => `Error in separableConv2d: pointwise filter must be rank 4, but ` +
            `got rank ${$depthwiseFilter.rank}.`);
        assert$4($pointwiseFilter.shape[0] === 1, () => `Error in separableConv2d: the first dimension of pointwise filter ` +
            ` must be 1, but got ${$pointwiseFilter.shape[0]}.`);
        assert$4($pointwiseFilter.shape[1] === 1, () => `Error in separableConv2d: the second dimension of pointwise ` +
            `filter must be 1, but got ${$pointwiseFilter.shape[1]}.`);
        const inChannels = $depthwiseFilter.shape[2];
        const channelMultiplier = $depthwiseFilter.shape[3];
        assert$4($pointwiseFilter.shape[2] === inChannels * channelMultiplier, () => `Error in separableConv2d: the third dimension of pointwise filter ` +
            `must be ${inChannels * channelMultiplier}, ` +
            `but got ${$pointwiseFilter.shape[2]}.`);
        const depthwise = depthwiseConv2d$3(x4D, $depthwiseFilter, strides, pad, dataFormat, dilation);
        const pointwiseStride = 1;
        const res = conv2d$3(depthwise, $pointwiseFilter, pointwiseStride, 'valid', dataFormat);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const separableConv2d$1 = op({ separableConv2d_ });

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the difference between two lists of numbers.
     *
     * Given a Tensor `x` and a Tensor `y`, this operation returns a Tensor `out`
     * that represents all values that are in `x` but not in `y`. The returned
     * Tensor `out` is sorted in the same order that the numbers appear in `x`
     * (duplicates are preserved). This operation also returns a Tensor indices that
     * represents the position of each out element in `x`. In other words:
     *
     * `out[i] = x[idx[i]] for i in [0, 1, ..., out.length - 1]`
     *
     * ```js
     * const x = [1, 2, 3, 4, 5, 6];
     * const y = [1, 3, 5];
     *
     * const [out, indices] = await tf.setdiff1dAsync(x, y);
     * out.print(); // [2, 4, 6]
     * indices.print(); // [1, 3, 5]
     * ```
     *
     * @param x 1-D Tensor. Values to keep.
     * @param y 1-D Tensor. Must have the same type as x. Values to exclude in the
     *     output.
     * @returns Promise of Tensor tuple [out, indices].
     *  out: Tensor with the same type as x.
     *  indices: A Tensor of type int32.
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    async function setdiff1dAsync_(x, y) {
        const $x = convertToTensor(x, 'x', 'setdiff1d');
        const $y = convertToTensor(y, 'y', 'setdiff1d');
        assert$4($x.dtype === $y.dtype, () => `x and y should have the same dtype, but got x (${$x.dtype}) and y (${$y.dtype}).`);
        assert$4($x.rank === 1, () => `x should be 1D tensor, but got x (${$x.shape}).`);
        assert$4($y.rank === 1, () => `y should be 1D tensor, but got y (${$y.shape}).`);
        const xVals = await $x.data();
        const yVals = await $y.data();
        const ySet = new Set(yVals);
        let outputSize = 0;
        for (let i = 0; i < xVals.length; i++) {
            if (!ySet.has(xVals[i])) {
                outputSize++;
            }
        }
        const buffer = new TensorBuffer([outputSize], $x.dtype);
        const indices = new TensorBuffer([outputSize], 'int32');
        for (let i = 0, p = 0; i < xVals.length; i++) {
            if (!ySet.has(xVals[i])) {
                buffer.values[p] = xVals[i];
                indices.values[p] = i;
                p++;
            }
        }
        return [buffer.toTensor(), indices.toTensor()];
    }
    const setdiff1dAsync = setdiff1dAsync_;

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns an element-wise indication of the sign of a number.
     *
     * ```js
     * const x = tf.tensor1d([.6, 1.1, -3.3, NaN, 0]);
     *
     * x.sign().print();  // or tf.sign(x)
     * ```
     * @param x The input Tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function sign_(x) {
        const $x = convertToTensor(x, 'x', 'sign');
        const inputs = { x: $x };
        return ENGINE.runKernel(Sign, inputs);
    }
    const sign$2 = op({ sign_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes sin of the input Tensor element-wise: `sin(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, Math.PI / 2, Math.PI * 3 / 4]);
     *
     * x.sin().print();  // or tf.sin(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function sin_(x) {
        const $x = convertToTensor(x, 'x', 'sin', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Sin, inputs);
    }
    const sin$2 = op({ sin_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes hyperbolic sin of the input `tf.Tensor` element-wise: `sinh(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, 1, -1, .7]);
     *
     * x.sinh().print();  // or tf.sinh(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function sinh_(x) {
        const $x = convertToTensor(x, 'x', 'sinh');
        const inputs = { x: $x };
        return ENGINE.runKernel(Sinh, inputs);
    }
    const sinh$2 = op({ sinh_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Extracts a 1D slice from 1D array starting at coordinates `begin` and is
     * of length `size`. See `slice` for details.
     */
    function slice1d_(x, begin, size) {
        const $x = convertToTensor(x, 'x', 'slice1d');
        assert$4($x.rank === 1, () => `slice1d expects a rank-1 tensor, but got a rank-${$x.rank} tensor`);
        return slice$2($x, [begin], [size]);
    }
    const slice1d = op({ slice1d_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Extracts a 2D slice from a 2D array starting at coordinates `begin` and
     * is of size `size`. See `slice` for details.
     */
    function slice2d_(x, begin, size) {
        const $x = convertToTensor(x, 'x', 'slice2d');
        assert$4($x.rank === 2, () => `slice2d expects a rank-2 tensor, but got a rank-${$x.rank} tensor`);
        return slice$2($x, begin, size);
    }
    const slice2d = op({ slice2d_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Extracts a 3D slice from a 3D array starting at coordinates `begin` and
     * is of size `size`. See `slice` for details.
     */
    function slice3d_(x, begin, size) {
        const $x = convertToTensor(x, 'x', 'slice3d');
        assert$4($x.rank === 3, () => `slice3d expects a rank-3 tensor, but got a rank-${$x.rank} tensor`);
        return slice$2($x, begin, size);
    }
    const slice3d = op({ slice3d_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Extracts a 4D slice from a 4D array starting at coordinates `begin` and
     * is of size `size`. See `slice` for details.
     */
    function slice4d_(x, begin, size) {
        const $x = convertToTensor(x, 'x', 'slice4d');
        assert$4($x.rank === 4, () => `slice4d expects a rank-4 tensor, but got a rank-${$x.rank} tensor`);
        return slice$2($x, begin, size);
    }
    const slice4d = op({ slice4d_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the softmax normalized vector given the logits.
     *
     * ```js
     * const a = tf.tensor1d([1, 2, 3]);
     *
     * a.softmax().print();  // or tf.softmax(a)
     * ```
     *
     * ```js
     * const a = tf.tensor2d([2, 4, 6, 1, 2, 3], [2, 3]);
     *
     * a.softmax().print();  // or tf.softmax(a)
     * ```
     *
     * @param logits The logits array.
     * @param dim The dimension softmax would be performed on. Defaults to `-1`
     *     which indicates the last dimension.
     *
     * @doc {heading: 'Operations', subheading: 'Normalization'}
     */
    function softmax_(logits, dim = -1) {
        const $logits = convertToTensor(logits, 'logits', 'softmax', 'float32');
        if (dim === -1) {
            dim = $logits.rank - 1;
        }
        if (dim !== $logits.rank - 1) {
            throw Error('Softmax along a non-last dimension is not yet supported. ' +
                `Logits was rank ${$logits.rank} and dim was ${dim}`);
        }
        const inputs = { logits: $logits };
        const attrs = { dim };
        return ENGINE.runKernel(Softmax$2, inputs, attrs);
    }
    const softmax$3 = op({ softmax_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Fast Fourier transform.
     *
     * Computes the 1-dimensional discrete Fourier transform over the inner-most
     * dimension of input.
     *
     * ```js
     * const real = tf.tensor1d([1, 2, 3]);
     * const imag = tf.tensor1d([1, 2, 3]);
     * const x = tf.complex(real, imag);
     *
     * x.fft().print();  // tf.spectral.fft(x).print();
     * ```
     * @param input The complex input to compute an fft over.
     *
     * @doc {heading: 'Operations', subheading: 'Spectral', namespace: 'spectral'}
     */
    function fft_(input) {
        assert$4(input.dtype === 'complex64', () => `The dtype for tf.spectral.fft() must be complex64 ` +
            `but got ${input.dtype}.`);
        const inputs = { input };
        return ENGINE.runKernel(FFT, inputs);
    }
    const fft$2 = op({ fft_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Inverse fast Fourier transform.
     *
     * Computes the inverse 1-dimensional discrete Fourier transform over the
     * inner-most dimension of input.
     *
     * ```js
     * const real = tf.tensor1d([1, 2, 3]);
     * const imag = tf.tensor1d([1, 2, 3]);
     * const x = tf.complex(real, imag);
     *
     * x.ifft().print();  // tf.spectral.ifft(x).print();
     * ```
     * @param input The complex input to compute an ifft over.
     *
     * @doc {heading: 'Operations', subheading: 'Spectral', namespace: 'spectral'}
     */
    function ifft_(input) {
        assert$4(input.dtype === 'complex64', () => `The dtype for tf.spectral.ifft() must be complex64 ` +
            `but got ${input.dtype}.`);
        const inputs = { input };
        return ENGINE.runKernel(IFFT, inputs);
    }
    const ifft$2 = op({ ifft_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Inversed real value input fast Fourier transform.
     *
     * Computes the 1-dimensional inversed discrete Fourier transform over the
     * inner-most dimension of the real input.
     *
     * ```js
     * const real = tf.tensor1d([1, 2, 3]);
     * const imag = tf.tensor1d([0, 0, 0]);
     * const x = tf.complex(real, imag);
     *
     * x.irfft().print();
     * ```
     * @param input The real value input to compute an irfft over.
     *
     * @doc {heading: 'Operations', subheading: 'Spectral', namespace: 'spectral'}
     */
    function irfft_(input) {
        const innerDimensionSize = input.shape[input.shape.length - 1];
        const batch = input.size / innerDimensionSize;
        let ret;
        if (innerDimensionSize <= 2) {
            const complexInput = reshape$3(input, [batch, innerDimensionSize]);
            ret = ifft$2(complexInput);
        }
        else {
            // The length of unique components of the DFT of a real-valued signal
            // is 2 * (input_len - 1)
            const outputShape = [batch, 2 * (innerDimensionSize - 1)];
            const realInput = reshape$3(real$2(input), [batch, innerDimensionSize]);
            const imagInput = reshape$3(imag$2(input), [batch, innerDimensionSize]);
            const realConjugate = reverse$2(slice$2(realInput, [0, 1], [batch, innerDimensionSize - 2]), 1);
            const imagConjugate = mul(reverse$2(slice$2(imagInput, [0, 1], [batch, innerDimensionSize - 2]), 1), scalar(-1));
            const r = concat$2([realInput, realConjugate], 1);
            const i = concat$2([imagInput, imagConjugate], 1);
            const complexInput = reshape$3(complex$2(r, i), [outputShape[0], outputShape[1]]);
            ret = ifft$2(complexInput);
        }
        ret = real$2(ret);
        // reshape the result if the input is 3D tensor.
        if (input.rank === 3 && input.shape[0] !== 0) {
            const temp = ret;
            const batch = input.shape[0];
            ret = reshape$3(ret, [batch, ret.shape[0] / batch, ret.shape[1]]);
            temp.dispose();
        }
        return ret;
    }
    const irfft = op({ irfft_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Splits a `tf.Tensor` into sub tensors.
     *
     * If `numOrSizeSplits` is a number, splits `x` along dimension `axis`
     * into `numOrSizeSplits` smaller tensors.
     * Requires that `numOrSizeSplits` evenly divides `x.shape[axis]`.
     *
     * If `numOrSizeSplits` is a number array, splits `x` into
     * `numOrSizeSplits.length` pieces. The shape of the `i`-th piece has the
     * same size as `x` except along dimension `axis` where the size is
     * `numOrSizeSplits[i]`.
     *
     * ```js
     * const x = tf.tensor2d([1, 2, 3, 4, 5, 6, 7, 8], [2, 4]);
     * const [a, b] = tf.split(x, 2, 1);
     * a.print();
     * b.print();
     *
     * const [c, d, e] = tf.split(x, [1, 2, 1], 1);
     * c.print();
     * d.print();
     * e.print();
     * ```
     *
     * @param x The input tensor to split.
     * @param numOrSizeSplits Either an integer indicating the number of
     * splits along the axis or an array of integers containing the sizes of
     * each output tensor along the axis. If a number then it must evenly divide
     * `x.shape[axis]`; otherwise the sum of sizes must match `x.shape[axis]`.
     * Can contain one -1 indicating that dimension is to be inferred.
     * @param axis The dimension along which to split. Defaults to 0 (the first
     * dim).
     *
     * @doc {heading: 'Tensors', subheading: 'Slicing and Joining'}
     */
    function split_(x, numOrSizeSplits, axis = 0) {
        const $x = convertToTensor(x, 'x', 'split');
        const inputs = { x: $x };
        const attr = { numOrSizeSplits, axis };
        return ENGINE.runKernel(SplitV, inputs, attr);
    }
    const split$2 = op({ split_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Real value input fast Fourier transform.
     *
     * Computes the 1-dimensional discrete Fourier transform over the
     * inner-most dimension of the real input.
     *
     * ```js
     * const real = tf.tensor1d([1, 2, 3]);
     *
     * real.rfft().print();
     * ```
     * @param input The real value input to compute an rfft over.
     *
     * @doc {heading: 'Operations', subheading: 'Spectral', namespace: 'spectral'}
     */
    function rfft_(input, fftLength) {
        assert$4(input.dtype === 'float32', () => `The dtype for rfft() must be real value but got ${input.dtype}`);
        let innerDimensionSize = input.shape[input.shape.length - 1];
        const batch = input.size / innerDimensionSize;
        let adjustedInput;
        if (fftLength != null && fftLength < innerDimensionSize) {
            // Need to crop
            const begin = input.shape.map(v => 0);
            const size = input.shape.map(v => v);
            size[input.shape.length - 1] = fftLength;
            adjustedInput = slice$2(input, begin, size);
            innerDimensionSize = fftLength;
        }
        else if (fftLength != null && fftLength > innerDimensionSize) {
            // Need to pad with zeros
            const zerosShape = input.shape.map(v => v);
            zerosShape[input.shape.length - 1] = fftLength - innerDimensionSize;
            adjustedInput = concat$2([input, zeros$2(zerosShape)], input.shape.length - 1);
            innerDimensionSize = fftLength;
        }
        else {
            adjustedInput = input;
        }
        // Complement the input with zero imaginary numbers.
        const zerosInput = zerosLike$2(adjustedInput);
        const complexInput = reshape$3(complex$2(adjustedInput, zerosInput), [batch, innerDimensionSize]);
        const ret = fft$2(complexInput);
        // Exclude complex conjugations. These conjugations are put symmetrically.
        const half = Math.floor(innerDimensionSize / 2) + 1;
        const realValues = real$2(ret);
        const imagValues = imag$2(ret);
        const realComplexConjugate = split$2(realValues, [half, innerDimensionSize - half], realValues.shape.length - 1);
        const imagComplexConjugate = split$2(imagValues, [half, innerDimensionSize - half], imagValues.shape.length - 1);
        const outputShape = adjustedInput.shape.slice();
        outputShape[adjustedInput.shape.length - 1] = half;
        return reshape$3(complex$2(realComplexConjugate[0], imagComplexConjugate[0]), outputShape);
    }
    const rfft = op({ rfft_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns (a - b) * (a - b) element-wise.
     * Supports broadcasting.
     *
     * ```js
     * const a = tf.tensor1d([1, 4, 3, 16]);
     * const b = tf.tensor1d([1, 2, 9, 4]);
     *
     * a.squaredDifference(b).print();  // or tf.squaredDifference(a, b)
     * ```
     *
     * ```js
     * // Broadcast squared difference  a with b.
     * const a = tf.tensor1d([2, 4, 6, 8]);
     * const b = tf.scalar(5);
     *
     * a.squaredDifference(b).print();  // or tf.squaredDifference(a, b)
     * ```
     *
     * @param a The first tensor.
     * @param b The second tensor. Must have the same type as `a`.
     *
     * @doc {heading: 'Operations', subheading: 'Arithmetic'}
     */
    function squaredDifference_(a, b) {
        let $a = convertToTensor(a, 'a', 'squaredDifference');
        let $b = convertToTensor(b, 'b', 'squaredDifference');
        [$a, $b] = makeTypesMatch($a, $b);
        assertAndGetBroadcastShape($a.shape, $b.shape);
        const inputs = { a: $a, b: $b };
        const attrs = {};
        return ENGINE.runKernel(SquaredDifference, inputs, attrs);
    }
    const squaredDifference$2 = op({ squaredDifference_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Removes dimensions of size 1 from the shape of a `tf.Tensor`.
     *
     * ```js
     * const x = tf.tensor([1, 2, 3, 4], [1, 1, 4]);
     * x.squeeze().print();
     * ```
     *
     * @param x The input tensor to be squeezed.
     * @param axis An optional list of numbers. If specified, only
     *     squeezes the dimensions listed. The dimension index starts at 0. It
     * is an error to squeeze a dimension that is not 1.
     *
     * @doc {heading: 'Tensors', subheading: 'Transformations'}
     */
    function squeeze_(x, axis) {
        const $x = convertToTensor(x, 'x', 'squeeze');
        return reshape$3($x, squeezeShape($x.shape, axis).newShape);
    }
    const squeeze = op({ squeeze_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Stacks a list of rank-`R` `tf.Tensor`s into one rank-`(R+1)` `tf.Tensor`.
     *
     * ```js
     * const a = tf.tensor1d([1, 2]);
     * const b = tf.tensor1d([3, 4]);
     * const c = tf.tensor1d([5, 6]);
     * tf.stack([a, b, c]).print();
     * ```
     *
     * @param tensors A list of tensor objects with the same shape and dtype.
     * @param axis The axis to stack along. Defaults to 0 (the first dim).
     *
     * @doc {heading: 'Tensors', subheading: 'Slicing and Joining'}
     */
    function stack_(tensors, axis = 0) {
        const $tensors = convertToTensorArray(tensors, 'tensors', 'stack', 'string_or_numeric');
        assert$4($tensors.length >= 1, () => 'Pass at least one tensor to tf.stack');
        if ($tensors.length > 0) {
            assert$4(axis <= $tensors[0].rank, () => 'Axis must be <= rank of the tensor');
        }
        const inputs = $tensors;
        const attrs = { axis };
        return ENGINE.runKernel(Pack, inputs, attrs);
    }
    const stack = op({ stack_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes step of the input `tf.Tensor` element-wise: `x > 0 ? 1 : alpha * x`
     *
     * ```js
     * const x = tf.tensor1d([0, 2, -1, -3]);
     *
     * x.step(.5).print();  // or tf.step(x, .5)
     * ```
     * @param x The input tensor.
     * @param alpha The gradient when input is negative.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function step_(x, alpha = 0.0) {
        const $x = convertToTensor(x, 'x', 'step');
        const inputs = { x: $x };
        const attrs = { alpha };
        return ENGINE.runKernel(Step, inputs, attrs);
    }
    const step$2 = op({ step_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Extracts a strided slice of a tensor.
     *
     * Roughly speaking, this op extracts a slice of size (end-begin)/stride from
     * the given input tensor (x). Starting at the location specified by begin the
     * slice continues by adding stride to the index until all dimensions are not
     * less than end. Note that a stride can be negative, which causes a reverse
     * slice.
     *
     * ```js
     * const t = tf.tensor3d([1, 1, 1 ,2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6],
     *    [3, 2, 3]);
     * t.stridedSlice([1, 0, 0], [2, 1, 3], [1, 1, 1]).print()  // [[[3, 3, 3]]]
     * t.stridedSlice([1, 0, 0], [2, 2, 3], [1, 1, 1]).print()  // [[[3, 3, 3],
     *                                                     // [4, 4, 4]]]
     * t.stridedSlice([1, -1, 0], [2, -3, 3], [1, -1, 1]).print() // [[[4, 4, 4],
     *                                                     // [3, 3, 3]]]
     * ```
     *
     * @param x The tensor to stride slice.
     * @param begin The coordinates to start the slice from.
     * @param end: The coordinates to end the slice at.
     * @param strides: The size of the slice.
     * @param beginMask: If the ith bit of beginMask is set, begin[i] is ignored
     *      and the fullest possible range in that dimension is used instead.
     * @param endMask: If the ith bit of endMask is set, end[i] is ignored
     *      and the fullest possible range in that dimension is used instead.
     * @param shrinkAxisMask: a bitmask where bit i implies that
     * the ith specification should shrink the dimensionality. begin and end must
     * imply a slice of size 1 in the dimension.
     *
     * @doc {heading: 'Operations', subheading: 'Slicing and Joining'}
     */
    function stridedSlice_(x, begin, end, strides, beginMask = 0, endMask = 0, ellipsisMask = 0, newAxisMask = 0, shrinkAxisMask = 0) {
        const $x = convertToTensor(x, 'x', 'stridedSlice', 'string_or_numeric');
        const inputs = { x: $x };
        const attrs = {
            begin,
            end,
            strides,
            beginMask,
            endMask,
            ellipsisMask,
            newAxisMask,
            shrinkAxisMask
        };
        return ENGINE.runKernel(StridedSlice, inputs, attrs);
    }
    const stridedSlice$2 = op({ stridedSlice_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes tan of the input `tf.Tensor` element-wise, `tan(x)`
     *
     * ```js
     * const x = tf.tensor1d([0, Math.PI / 2, Math.PI * 3 / 4]);
     *
     * x.tan().print();  // or tf.tan(x)
     * ```
     * @param x The input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Basic math'}
     */
    function tan_(x) {
        const $x = convertToTensor(x, 'x', 'tan', 'float32');
        const inputs = { x: $x };
        return ENGINE.runKernel(Tan, inputs);
    }
    const tan$2 = op({ tan_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates rank-1 `tf.Tensor` with the provided values, shape and dtype.
     *
     * The same functionality can be achieved with `tf.tensor`, but in general
     * we recommend using `tf.tensor1d` as it makes the code more readable.
     *
     * ```js
     * tf.tensor1d([1, 2, 3]).print();
     * ```
     *
     * @param values The values of the tensor. Can be array of numbers,
     *     or a `TypedArray`.
     * @param dtype The data type.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function tensor1d(values, dtype) {
        assertNonNull(values);
        const inferredShape = inferShape(values, dtype);
        if (inferredShape.length !== 1) {
            throw new Error('tensor1d() requires values to be a flat/TypedArray');
        }
        const shape = null;
        return makeTensor(values, shape, inferredShape, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates rank-2 `tf.Tensor` with the provided values, shape and dtype.
     *
     * The same functionality can be achieved with `tf.tensor`, but in general
     * we recommend using `tf.tensor2d` as it makes the code more readable.
     *
     *  ```js
     * // Pass a nested array.
     * tf.tensor2d([[1, 2], [3, 4]]).print();
     * ```
     * ```js
     * // Pass a flat array and specify a shape.
     * tf.tensor2d([1, 2, 3, 4], [2, 2]).print();
     * ```
     *
     * @param values The values of the tensor. Can be nested array of numbers,
     *     or a flat array, or a `TypedArray`.
     * @param shape The shape of the tensor. If not provided, it is inferred from
     *     `values`.
     * @param dtype The data type.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function tensor2d(values, shape, dtype) {
        assertNonNull(values);
        if (shape != null && shape.length !== 2) {
            throw new Error('tensor2d() requires shape to have two numbers');
        }
        const inferredShape = inferShape(values, dtype);
        if (inferredShape.length !== 2 && inferredShape.length !== 1) {
            throw new Error('tensor2d() requires values to be number[][] or flat/TypedArray');
        }
        if (inferredShape.length === 1 && shape == null) {
            throw new Error('tensor2d() requires shape to be provided when `values` ' +
                'are a flat/TypedArray');
        }
        return makeTensor(values, shape, inferredShape, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates rank-4 `tf.Tensor` with the provided values, shape and dtype.
     *
     * The same functionality can be achieved with `tf.tensor`, but in general
     * we recommend using `tf.tensor4d` as it makes the code more readable.
     *
     *  ```js
     * // Pass a nested array.
     * tf.tensor4d([[[[1], [2]], [[3], [4]]]]).print();
     * ```
     * ```js
     * // Pass a flat array and specify a shape.
     * tf.tensor4d([1, 2, 3, 4], [1, 2, 2, 1]).print();
     * ```
     *
     * @param values The values of the tensor. Can be nested array of numbers,
     *     or a flat array, or a `TypedArray`.
     * @param shape The shape of the tensor. Optional. If not provided,
     *   it is inferred from `values`.
     * @param dtype The data type.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function tensor4d(values, shape, dtype) {
        assertNonNull(values);
        if (shape != null && shape.length !== 4) {
            throw new Error('tensor4d() requires shape to have four numbers');
        }
        const inferredShape = inferShape(values, dtype);
        if (inferredShape.length !== 4 && inferredShape.length !== 1) {
            throw new Error('tensor4d() requires values to be number[][][][] or flat/TypedArray');
        }
        if (inferredShape.length === 1 && shape == null) {
            throw new Error('tensor4d() requires shape to be provided when `values` ' +
                'are a flat array');
        }
        return makeTensor(values, shape, inferredShape, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates rank-5 `tf.Tensor` with the provided values, shape and dtype.
     *
     * The same functionality can be achieved with `tf.tensor`, but in general
     * we recommend using `tf.tensor5d` as it makes the code more readable.
     *
     *  ```js
     * // Pass a nested array.
     * tf.tensor5d([[[[[1],[2]],[[3],[4]]],[[[5],[6]],[[7],[8]]]]]).print();
     * ```
     * ```js
     * // Pass a flat array and specify a shape.
     * tf.tensor5d([1, 2, 3, 4, 5, 6, 7, 8], [1, 2, 2, 2, 1]).print();
     * ```
     *
     * @param values The values of the tensor. Can be nested array of numbers,
     *     or a flat array, or a `TypedArray`.
     * @param shape The shape of the tensor. Optional. If not provided,
     *   it is inferred from `values`.
     * @param dtype The data type.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function tensor5d(values, shape, dtype) {
        assertNonNull(values);
        if (shape != null && shape.length !== 5) {
            throw new Error('tensor5d() requires shape to have five numbers');
        }
        const inferredShape = inferShape(values, dtype);
        if (inferredShape.length !== 5 && inferredShape.length !== 1) {
            throw new Error('tensor5d() requires values to be ' +
                'number[][][][][] or flat/TypedArray');
        }
        if (inferredShape.length === 1 && shape == null) {
            throw new Error('tensor5d() requires shape to be provided when `values` ' +
                'are a flat array');
        }
        return makeTensor(values, shape, inferredShape, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates rank-6 `tf.Tensor` with the provided values, shape and dtype.
     *
     * The same functionality can be achieved with `tf.tensor`, but in general
     * we recommend using `tf.tensor6d` as it makes the code more readable.
     *
     *  ```js
     * // Pass a nested array.
     * tf.tensor6d([[[[[[1],[2]],[[3],[4]]],[[[5],[6]],[[7],[8]]]]]]).print();
     * ```
     * ```js
     * // Pass a flat array and specify a shape.
     * tf.tensor6d([1, 2, 3, 4, 5, 6, 7, 8], [1, 1, 2, 2, 2, 1]).print();
     * ```
     *
     * @param values The values of the tensor. Can be nested array of numbers,
     *     or a flat array, or a `TypedArray`.
     * @param shape The shape of the tensor. Optional. If not provided,
     *   it is inferred from `values`.
     * @param dtype The data type.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function tensor6d(values, shape, dtype) {
        assertNonNull(values);
        if (shape != null && shape.length !== 6) {
            throw new Error('tensor6d() requires shape to have six numbers');
        }
        const inferredShape = inferShape(values, dtype);
        if (inferredShape.length !== 6 && inferredShape.length !== 1) {
            throw new Error('tensor6d() requires values to be number[][][][][][] or ' +
                'flat/TypedArray');
        }
        if (inferredShape.length === 1 && shape == null) {
            throw new Error('tensor6d() requires shape to be provided when `values` ' +
                'are a flat array');
        }
        shape = shape ||
            inferredShape;
        return makeTensor(values, shape, inferredShape, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Finds the values and indices of the `k` largest entries along the last
     * dimension.
     *
     * If the input is a vector (rank=1), finds the k largest entries in the vector
     * and outputs their values and indices as vectors. Thus values[j] is the j-th
     * largest entry in input, and its index is indices[j].
     * For higher rank inputs, computes the top k entries along the last dimension.
     *
     * If two elements are equal, the lower-index element appears first.
     *
     * ```js
     * const a = tf.tensor2d([[1, 5], [4, 3]]);
     * const {values, indices} = tf.topk(a);
     * values.print();
     * indices.print();
     * ```
     * @param x 1-D or higher `tf.Tensor` with last dimension being at least `k`.
     * @param k Number of top elements to look for along the last dimension.
     * @param sorted If true, the resulting `k` elements will be sorted by the
     *     values in descending order.
     *
     * @doc {heading: 'Operations', subheading: 'Evaluation'}
     */
    function topk_(x, k = 1, sorted = true) {
        const $x = convertToTensor(x, 'x', 'topk');
        if ($x.rank === 0) {
            throw new Error('topk() expects the input to be of rank 1 or higher');
        }
        const lastDim = $x.shape[$x.shape.length - 1];
        if (k < 0) {
            throw new Error(`'k' passed to topk() must be >= 0 but got ${k}`);
        }
        if (k > lastDim) {
            throw new Error(`'k' passed to topk() must be <= the last dimension (${lastDim}) ` +
                `but got ${k}`);
        }
        const inputs = { x: $x };
        const attrs = { k, sorted };
        const [values, indices] = ENGINE.runKernel(TopK, inputs, attrs);
        return { values, indices };
    }
    const topk = op({ topk_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a `tf.Tensor` with values sampled from a truncated normal
     * distribution.
     *
     * ```js
     * tf.truncatedNormal([2, 2]).print();
     * ```
     *
     * The generated values follow a normal distribution with specified mean and
     * standard deviation, except that values whose magnitude is more than 2
     * standard deviations from the mean are dropped and re-picked.
     *
     * @param shape An array of integers defining the output tensor shape.
     * @param mean The mean of the normal distribution.
     * @param stdDev The standard deviation of the normal distribution.
     * @param dtype The data type of the output tensor.
     * @param seed The seed for the random number generator.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function truncatedNormal_(shape, mean = 0, stdDev = 1, dtype, seed) {
        if (dtype != null && dtype === 'bool') {
            throw new Error(`Unsupported data type $ { dtype }`);
        }
        const randGauss = new MPRandGauss(mean, stdDev, dtype, true /* truncated */, seed);
        const res = buffer(shape, dtype);
        for (let i = 0; i < res.values.length; i++) {
            res.values[i] = randGauss.nextValue();
        }
        return res.toTensor();
    }
    const truncatedNormal$1 = op({ truncatedNormal_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Finds unique elements along an axis of a tensor.
     *
     * It returns a tensor `values` containing all of the unique elements along the
     * `axis` of the given tensor `x` in the same order that they occur along the
     * `axis` in `x`; `x` does not need to be sorted. It also returns a tensor
     * `indices` the same size as the number of the elements in `x` along the `axis`
     * dimension. It contains the index in the unique output `values`.
     *
     * ```js
     * // A 1-D tensor
     * const a = tf.tensor1d([1, 1, 2, 4, 4, 4, 7, 8, 8]);
     * const {values, indices} = tf.unique(a);
     * values.print();   // [1, 2, 4, 7, 8,]
     * indices.print();  // [0, 0, 1, 2, 2, 2, 3, 4, 4]
     * ```
     *
     * ```js
     * // A 2-D tensor with axis=0
     * //
     * // 'a' is: [[1, 0, 0],
     * //          [1, 0, 0],
     * //          [2, 0, 0]]
     * const a = tf.tensor2d([[1, 0, 0], [1, 0, 0], [2, 0, 0]]);
     * const {values, indices} = tf.unique(a, 0)
     * values.print();   // [[1, 0, 0],
     *                   //  [2, 0, 0]]
     * indices.print();  // [0, 0, 1]
     * ```
     *
     * ```js
     * // A 2-D tensor with axis=1
     * //
     * // 'a' is: [[1, 0, 0],
     * //          [1, 0, 0],
     * //          [2, 0, 0]]
     * const a = tf.tensor2d([[1, 0, 0], [1, 0, 0], [2, 0, 0]]);
     * const {values, indices} = tf.unique(a, 1)
     * values.print();   // [[1, 0],
     *                   //  [1, 0],
     *                   //  [2, 0]]
     * indices.print();  // [0, 1, 1]
     * ```
     * @param x A tensor (int32, string, bool).
     * @param axis The axis of the tensor to find the unique elements.
     * @returns [uniqueElements, indices] (see above for details)
     *
     * @doc {heading: 'Operations', subheading: 'Evaluation'}
     */
    function unique_(x, axis = 0) {
        const $x = convertToTensor(x, 'x', 'unique', 'string_or_numeric');
        assert$4($x.rank > 0, () => 'The input tensor must be at least 1D');
        const inputs = { x: $x };
        const attrs = { axis };
        const [values, indices] = ENGINE.runKernel(Unique, inputs, attrs);
        return { values, indices };
    }
    const unique$3 = op({ unique_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the sum along segments of a `tf.Tensor`.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 3, 4]);
     * const segmentIds = tf.tensor1d([1, 2, 0, 1], 'int32');
     * const numSegments = 3;
     *
     * x.unsortedSegmentSum(segmentIds, numSegments).print()
     * //or tf.unsortedSegmentSum(x, segmentIds, numSegments)
     * ```
     * @param x The `tf.Tensor` that will be summed along its segments.
     * @param segmentIds A `tf.Tensor1D` whose rank is equal to the rank of `x`'s
     * dimension along the `axis`.  Maps each element of `x` to a segment.
     * @param numSegments The number of distinct `segmentIds`.
     *
     * @doc {heading: 'Operations', subheading: 'Segment'}
     */
    function unsortedSegmentSum_(x, segmentIds, numSegments) {
        const $x = convertToTensor(x, 'x', 'unsortedSegmentSum');
        const $segmentIds = convertToTensor(segmentIds, 'segmentIds', 'unsortedSegmentSum', 'int32');
        assert$4(isInt(numSegments), () => 'numSegments must be of dtype int');
        const inputs = { x: $x, segmentIds: $segmentIds };
        const attrs = { numSegments };
        return ENGINE.runKernel(UnsortedSegmentSum, inputs, attrs);
    }
    const unsortedSegmentSum$2 = op({ unsortedSegmentSum_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Unstacks a `tf.Tensor` of rank-`R` into a list of rank-`(R-1)` `tf.Tensor`s.
     *
     * ```js
     * const a = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     *
     * tf.unstack(a).forEach(tensor => tensor.print());
     * ```
     *
     * @param x A tensor object.
     * @param axis The axis to unstack along. Defaults to 0 (the first dim).
     *
     * @doc {heading: 'Tensors', subheading: 'Slicing and Joining'}
     */
    function unstack_(x, axis = 0) {
        const $x = convertToTensor(x, 'x', 'unstack', 'string_or_numeric');
        assert$4(axis >= -$x.shape.length && axis < $x.shape.length, () => `Axis = ${axis} is not in [-${$x.shape.length}, ${$x.shape.length})`);
        const inputs = { value: $x };
        const attrs = { axis };
        return ENGINE.runKernel(Unpack, inputs, attrs);
    }
    const unstack = op({ unstack_ });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Searches for where a value would go in a sorted sequence.
     *
     * This is not a method for checking containment (like javascript in).
     *
     * The typical use case for this operation is "binning", "bucketing", or
     * "discretizing". The values are assigned to bucket-indices based on the edges
     * listed in 'sortedSequence'. This operation returns the bucket-index for each
     * value.
     *
     * The index returned corresponds to the first edge greater than the value.
     *
     * The axis is not settable for this operation. It always operates on the
     * innermost dimension (axis=-1). The operation will accept any number of outer
     * dimensions.
     *
     * Note: This operation assumes that 'upperBound' is sorted along the
     * innermost axis, maybe using 'sort(..., axis=-1)'. If the sequence is not
     * sorted no error is raised and the content of the returned tensor is not well
     * defined.
     *
     * ```js
     * const seq = tf.tensor1d([0, 3, 9, 10, 10]);
     * const values = tf.tensor1d([0, 4, 10]);
     * const result = tf.upperBound(seq, values);
     * result.print(); // [1, 2, 5]
     * ```
     * @param sortedSequence: N-D. Sorted sequence.
     * @param values: N-D. Search values.
     * @return An N-D int32 tensor the size of values containing the result of
     *     applying upper bound to each value. The result is not a global index to
     *     the entire Tensor, but the index in the last dimension.
     * @doc {heading: 'Operations', subheading: 'Evaluation'}
     */
    function upperBound$1(sortedSequence, values) {
        return searchSorted$2(sortedSequence, values, 'right');
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a new variable with the provided initial value.
     * ```js
     * const x = tf.variable(tf.tensor([1, 2, 3]));
     * x.assign(tf.tensor([4, 5, 6]));
     *
     * x.print();
     * ```
     *
     * @param initialValue Initial value for the tensor.
     * @param trainable If true, optimizers are allowed to update it.
     * @param name Name of the variable. Defaults to a unique id.
     * @param dtype If set, initialValue will be converted to the given type.
     *
     * @doc {heading: 'Tensors', subheading: 'Creation'}
     */
    function variable(initialValue, trainable = true, name, dtype) {
        return ENGINE.makeVariable(initialValue, trainable, name, dtype);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function whereImpl$2(condShape, condVals) {
        const indices = [];
        for (let i = 0; i < condVals.length; i++) {
            if (condVals[i]) {
                indices.push(i);
            }
        }
        const inBuffer = buffer(condShape, 'int32');
        const out = buffer([indices.length, condShape.length], 'int32');
        for (let i = 0; i < indices.length; i++) {
            const loc = inBuffer.indexToLoc(indices[i]);
            const offset = i * condShape.length;
            out.values.set(loc, offset);
        }
        return out.toTensor();
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns the coordinates of true elements of condition.
     *
     * The coordinates are returned in a 2-D tensor where the first dimension (rows)
     * represents the number of true elements, and the second dimension (columns)
     * represents the coordinates of the true elements. Keep in mind, the shape of
     * the output tensor can vary depending on how many true values there are in
     * input. Indices are output in row-major order. The resulting tensor has the
     * shape `[numTrueElems, condition.rank]`.
     *
     * This is analogous to calling the python `tf.where(cond)` without an x or y.
     *
     * ```js
     * const cond = tf.tensor1d([false, false, true], 'bool');
     * const result = await tf.whereAsync(cond);
     * result.print();
     * ```
     *
     * @doc {heading: 'Operations', subheading: 'Logical'}
     */
    async function whereAsync_(condition) {
        const $condition = convertToTensor(condition, 'condition', 'whereAsync', 'bool');
        const vals = await $condition.data();
        const res = whereImpl$2($condition.shape, vals);
        if (condition !== $condition) {
            $condition.dispose();
        }
        return res;
    }
    const whereAsync = whereAsync_;

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Apply boolean mask to tensor.
     *
     * ```js
     * const tensor = tf.tensor2d([1, 2, 3, 4, 5, 6], [3, 2]);
     * const mask = tf.tensor1d([1, 0, 1], 'bool');
     * const result = await tf.booleanMaskAsync(tensor, mask);
     * result.print();
     * ```
     *
     * @param tensor N-D tensor.
     * @param mask K-D boolean tensor, K <= N and K must be known statically.
     * @param axis A 0-D int Tensor representing the axis in tensor to mask from.
     *     By default, axis is 0 which will mask from the first dimension.
     *     Otherwise K + axis <= N.
     *
     * @doc {heading: 'Tensors', subheading: 'Slicing and Joining'}
     */
    async function booleanMaskAsync_(tensor, mask, axis) {
        const $tensor = convertToTensor(tensor, 'tensor', 'boolMask');
        const $mask = convertToTensor(mask, 'mask', 'boolMask', 'bool');
        const axisFrom = axis == null ? 0 : axis;
        const maskDim = $mask.rank;
        const tensorShape = $tensor.shape;
        assert$4(maskDim > 0, () => 'mask cannot be scalar');
        assertShapesMatch(tensorShape.slice(axisFrom, axisFrom + maskDim), $mask.shape, `mask's shape must match the first K dimensions of tensor's shape,`);
        let leadingSize = 1;
        for (let i = axisFrom; i < axisFrom + maskDim; i++) {
            leadingSize *= tensorShape[i];
        }
        const targetTensorShape = tensorShape.slice(0, axisFrom)
            .concat([leadingSize], tensorShape.slice(axisFrom + maskDim));
        const reshapedTensor = reshape$3($tensor, targetTensorShape);
        const reshapedMask = reshape$3($mask, [-1]);
        const positivePositions = await whereAsync(reshapedMask);
        const indices = squeeze(positivePositions, [1]);
        const res = gather$1(reshapedTensor, indices, axisFrom);
        // Ensure no memory leak.
        if (tensor !== $tensor) {
            $tensor.dispose();
        }
        if (mask !== $mask) {
            $mask.dispose();
        }
        indices.dispose();
        reshapedTensor.dispose();
        reshapedMask.dispose();
        positivePositions.dispose();
        return res;
    }
    const booleanMaskAsync = booleanMaskAsync_;

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Compute the moving average of a variable.
     *
     * Without zeroDebias, the moving average operation is defined by:
     *   `v += delta`
     * where
     *   `delta = (1 - decay) * (x - v)`
     *
     * With zeroDebias (default), the `delta` term is scaled to debias the
     * effect of the (assumed) zero-initialization of `v`.
     *   `delta /= (1 - decay ^ step)`
     *
     * For more details on the zero-debiasing algorithm, see:
     *   https://arxiv.org/abs/1412.6980
     *
     * Note that this function is completely stateless and does not keep track of
     * step count. The step count needs to be maintained by the caller and passed
     * in as `step`.
     *
     * @param v The current moving average value.
     * @param x New input value, must have the same shape and dtype as `v`.
     * @param decay The decay factor. Typical values are 0.95 and 0.99.
     * @param step Step count.
     * @param zeroDebias: Whether zeroDebias is to be performed (default: `true`).
     * @returns The new moving average value.
     *
     * @doc {heading: 'Operations', subheading: 'Moving Average'}
     */
    function movingAverage_(v, x, decay, step, zeroDebias = true) {
        const $v = convertToTensor(v, 'v', 'movingAverage');
        const $x = convertToTensor(x, 'x', 'movingAverage');
        const $decay = convertToTensor(decay, 'decay', 'movingAverage');
        assertTypesMatch($v, $x);
        assert$4(arraysEqual($v.shape, $x.shape), () => 'Shape mismatch in v and x');
        const one = scalar(1);
        const oneMinusDecay = sub$2(one, $decay);
        let update = mul(sub$2($x, $v), oneMinusDecay);
        if (zeroDebias) {
            assert$4(step != null, () => 'When using zeroDebias: true, step is required.');
            const $step = convertToTensor(step, 'step', 'movingAverage');
            update = div$1(update, sub$2(one, pow$2($decay, $step)));
        }
        return add$2($v, update);
    }
    const movingAverage = op({ movingAverage_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates a new tensor by applying sparse updates to individual
     * values or slices within a zero tensor of the given shape tensor according to
     * indices. This operator is the inverse of the `tf.gatherND` operator which
     * extracts values or slices from a given tensor.
     *
     * ```js
     * const indices = tf.tensor2d([4, 3, 1, 7], [4, 1], 'int32');
     * const updates = tf.tensor1d([9, 10, 11, 12]);
     * const shape = [8];
     * tf.scatterND(indices, updates, shape).print() //[0, 11, 0, 10, 9, 0, 0, 12]
     * ```
     *
     * @param indices The tensor contains the indices into the output tensor.
     * @param updates The tensor contains the value for the indices.
     * @param shape: The shape of the output tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Slicing and Joining'}
     */
    function scatterND_(indices, updates, shape) {
        const $indices = convertToTensor(indices, 'indices', 'scatterND', 'int32');
        const $updates = convertToTensor(updates, 'updates', 'scatterND');
        validateInput$1($updates, $indices, shape);
        const inputs = { indices: $indices, updates: $updates };
        const attrs = { shape };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        return ENGINE.runKernel(ScatterNd, inputs, attrs);
    }
    const scatterND = op({ scatterND_ });

    /**
     * Validate sparseToDense inputs.
     *
     * @param sparseIndices A 0-D, 1-D, or 2-D Tensor of type int32.
     * sparseIndices[i] contains the complete index where sparseValues[i] will be
     * placed.
     * @param sparseValues A 0-D or 1-D Tensor. Values
     * corresponding to each row of sparseIndices, or a scalar value to be used for
     * all sparse indices.
     * @param outputShape number[]. Shape of the dense output tensor.
     * @param validateIndices boolean. indice validation is not supported, error
     * will be thrown if it is set.
     */
    function validateInput(sparseIndices, sparseValues, outputShape, defaultValues) {
        if (sparseIndices.dtype !== 'int32') {
            throw new Error('tf.sparseToDense() expects the indices to be int32 type,' +
                ` but the dtype was ${sparseIndices.dtype}.`);
        }
        if (sparseIndices.rank > 2) {
            throw new Error('sparseIndices should be a scalar, vector, or matrix,' +
                ` but got shape ${sparseIndices.shape}.`);
        }
        const numElems = sparseIndices.rank > 0 ? sparseIndices.shape[0] : 1;
        const numDims = sparseIndices.rank > 1 ? sparseIndices.shape[1] : 1;
        if (outputShape.length !== numDims) {
            throw new Error('outputShape has incorrect number of elements:,' +
                ` ${outputShape.length}, should be: ${numDims}.`);
        }
        const numValues = sparseValues.size;
        if (!(sparseValues.rank === 0 ||
            sparseValues.rank === 1 && numValues === numElems)) {
            throw new Error('sparseValues has incorrect shape ' +
                `${sparseValues.shape}, should be [] or [${numElems}]`);
        }
        if (sparseValues.dtype !== defaultValues.dtype) {
            throw new Error('sparseValues.dtype must match defaultValues.dtype');
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts a sparse representation into a dense tensor.
     *
     * Builds an array dense with shape outputShape such that:
     *
     * // If sparseIndices is scalar
     * dense[i] = (i == sparseIndices ? sparseValues : defaultValue)
     *
     * // If sparseIndices is a vector, then for each i
     * dense[sparseIndices[i]] = sparseValues[i]
     *
     * // If sparseIndices is an n by d matrix, then for each i in [0, n)
     * dense[sparseIndices[i][0], ..., sparseIndices[i][d-1]] = sparseValues[i]
     * All other values in dense are set to defaultValue. If sparseValues is a
     * scalar, all sparse indices are set to this single value.
     *
     * If indices are repeated the final value is summed over all values for those
     * indices.
     *
     * ```js
     * const indices = tf.tensor1d([4, 5, 6, 1, 2, 3], 'int32');
     * const values = tf.tensor1d([10, 11, 12, 13, 14, 15], 'float32');
     * const shape = [8];
     * tf.sparseToDense(indices, values, shape).print();
     * ```
     *
     * @param sparseIndices A 0-D, 1-D, or 2-D Tensor of type int32.
     * sparseIndices[i] contains the complete index where sparseValues[i] will be
     * placed.
     * @param sparseValues A 0-D or 1-D Tensor. Values
     * corresponding to each row of sparseIndices, or a scalar value to be used for
     * all sparse indices.
     * @param outputShape Shape of the dense output tensor. the type is inferred.
     * @param defaultValue Scalar. Value to set for indices not specified in
     * sparseIndices. Defaults to zero.
     *
     * @doc {heading: 'Operations', subheading: 'Normalization'}
     */
    function sparseToDense_(sparseIndices, sparseValues, outputShape, defaultValue = 0) {
        const $sparseIndices = convertToTensor(sparseIndices, 'sparseIndices', 'sparseToDense', 'int32');
        const $sparseValues = convertToTensor(sparseValues, 'sparseValues', 'sparseToDense', 'string_or_numeric');
        const $defaultValue = convertToTensor(defaultValue, 'defaultValue', 'sparseToDense', $sparseValues.dtype);
        validateInput($sparseIndices, $sparseValues, outputShape, $defaultValue);
        const inputs = {
            sparseIndices: $sparseIndices,
            sparseValues: $sparseValues,
            defaultValue: $defaultValue
        };
        const attrs = { outputShape };
        return ENGINE.runKernel(SparseToDense, inputs, attrs);
    }
    const sparseToDense$2 = op({ sparseToDense_ });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Gather slices from input tensor into a Tensor with shape specified by
     * `indices`.
     *
     * `indices` is an K-dimensional integer tensor, best thought of as a
     * (K-1)-dimensional tensor of indices into input, where each element defines a
     * slice of input:
     * output[\\(i_0, ..., i_{K-2}\\)] = input[indices[\\(i_0, ..., i_{K-2}\\)]]
     *
     * Whereas in `tf.gather`, `indices` defines slices into the first dimension of
     * input, in `tf.gatherND`, `indices` defines slices into the first N dimensions
     * of input, where N = indices.shape[-1].
     *
     * The last dimension of indices can be at most the rank of input:
     * indices.shape[-1] <= input.rank
     *
     * The last dimension of `indices` corresponds to elements
     * (if indices.shape[-1] == input.rank) or slices
     * (if indices.shape[-1] < input.rank) along dimension indices.shape[-1] of
     * input.
     * The output tensor has shape
     * indices.shape[:-1] + input.shape[indices.shape[-1]:]
     *
     * Note that on CPU, if an out of bound index is found, an error is returned. On
     * GPU, if an out of bound index is found, a 0 is stored in the corresponding
     * output value.
     *
     * ```js
     * const indices = tf.tensor2d([0, 1, 1, 0], [2,2], 'int32');
     * const input = tf.tensor2d([9, 10, 11, 12], [2, 2]);
     * tf.gatherND(input, indices).print() // [10, 11]
     * ```
     *
     * @param x The tensor from which to gather values.
     * @param indices Index tensor, must be of type int32.
     *
     * @doc {heading: 'Operations', subheading: 'Slicing and Joining'}
     */
    function gatherND_(x, indices) {
        const $indices = convertToTensor(indices, 'indices', 'gatherND', 'int32');
        const $x = convertToTensor(x, 'x', 'gatherND', 'string_or_numeric');
        const inputs = { params: $x, indices: $indices };
        return ENGINE.runKernel(GatherNd, inputs);
    }
    const gatherND = op({ gatherND_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Normalize noise shape based on provided tensor and noise shape.
     *
     * @param x Tensor.
     * @param noiseShape The shape for the randomly generated keep/drop flags, as
     *   an array of numbers. Optional.
     * @returns Normalized noise shape.
     */
    function getNoiseShape(x, noiseShape) {
        if (noiseShape == null) {
            return x.shape.slice();
        }
        if (arraysEqual(x.shape, noiseShape)) {
            return noiseShape;
        }
        if (x.shape.length === noiseShape.length) {
            const newDimension = [];
            for (let i = 0; i < x.shape.length; i++) {
                if (noiseShape[i] == null && x.shape[i] != null) {
                    newDimension.push(x.shape[i]);
                }
                else {
                    newDimension.push(noiseShape[i]);
                }
            }
            return newDimension;
        }
        return noiseShape;
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes dropout.
     *
     * ```js
     * const x = tf.tensor1d([1, 2, 2, 1]);
     * const rate = 0.75;
     * const output = tf.dropout(x, rate);
     * output.print();
     * ```
     *
     * @param x A floating point Tensor or TensorLike.
     * @param rate A float in the range [0, 1). The probability that each element
     *   of x is discarded.
     * @param noiseShape An array of numbers of type int32, representing the
     * shape for randomly generated keep/drop flags. If the noiseShape has null
     * value, it will be automatically replaced with the x's relative dimension
     * size. Optional.
     * @param seed Used to create random seeds. Optional.
     * @returns A Tensor of the same shape of x.
     *
     * @doc {heading: 'Operations', subheading: 'Dropout'}
     */
    function dropout_(x, rate, noiseShape, seed) {
        const $x = convertToTensor(x, 'x', 'dropout');
        assert$4($x.dtype === 'float32', () => `x has to be a floating point tensor since it's going to be ` +
            `scaled, but got a ${$x.dtype} tensor instead.`);
        assert$4(rate >= 0 && rate < 1, () => `rate must be a float in the range [0, 1), but got ${rate}.`);
        if (rate === 0) {
            return x instanceof Tensor ? $x.clone() : $x;
        }
        const $noiseShape = getNoiseShape($x, noiseShape);
        const keepProb = 1 - rate;
        const multiplier = div$1(floor$3(add$2(randomUniform$1($noiseShape, 0, 1, 'float32', seed), keepProb)), keepProb);
        return mul($x, multiplier);
    }
    const dropout$2 = op({ dropout_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function enclosingPowerOfTwo(value) {
        // Return 2**N for integer N such that 2**N >= value.
        return Math.floor(Math.pow(2, Math.ceil(Math.log(value) / Math.log(2.0))));
    }
    function cosineWindow(windowLength, a, b) {
        const even = 1 - windowLength % 2;
        const newValues = new Float32Array(windowLength);
        for (let i = 0; i < windowLength; ++i) {
            const cosArg = (2.0 * Math.PI * i) / (windowLength + even - 1);
            newValues[i] = a - b * Math.cos(cosArg);
        }
        return tensor1d(newValues, 'float32');
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Returns whether the targets are in the top K predictions.
     *
     * ```js
     * const predictions = tf.tensor2d([[20, 10, 40, 30], [30, 50, -20, 10]]);
     * const targets = tf.tensor1d([2, 0]);
     * const precision = await tf.inTopKAsync(predictions, targets);
     * precision.print();
     * ```
     * @param predictions 2-D or higher `tf.Tensor` with last dimension being
     *     at least `k`.
     * @param targets 1-D or higher `tf.Tensor`.
     * @param k Optional Number of top elements to look at for computing precision,
     *     default to 1.
     *
     * @doc {heading: 'Operations', subheading: 'Evaluation'}
     */
    async function inTopKAsync_(predictions, targets, k = 1) {
        const $predictions = convertToTensor(predictions, 'predictions', 'inTopK');
        const $targets = convertToTensor(targets, 'targets', 'inTopK');
        assert$4($predictions.rank > 1, () => 'inTopK() expects the predictions to be of rank 2 or higher, ' +
            `but got ${$predictions.rank}`);
        assert$4($predictions.rank - 1 === $targets.rank, () => `predictions rank should be 1 larger than ` +
            `targets rank, but got predictions rank ` +
            `${$predictions.rank} and targets rank ${$targets.rank}`);
        assertShapesMatch($predictions.shape.slice(0, $predictions.shape.length - 1), $targets.shape, `predictions's shape should be align with the targets' shape, ` +
            'except the last dimension.');
        const lastDim = $predictions.shape[$predictions.shape.length - 1];
        assert$4(k > 0 && k <= lastDim, () => `'k' passed to inTopK() must be > 0 && <= the predictions last ` +
            `dimension (${lastDim}), but got ${k}`);
        const predictionsVals = await $predictions.data();
        const targetsVals = await $targets.data();
        // Reshape predictionsVals into a 2d tensor [batch, lastDim]
        // and look up topK along lastDim.
        const [batch, size] = [predictionsVals.length / lastDim, lastDim];
        const precision = getTypedArrayFromDType('bool', batch);
        for (let b = 0; b < batch; b++) {
            const offset = b * size;
            const vals = predictionsVals.subarray(offset, offset + size);
            const valAndInd = [];
            for (let i = 0; i < vals.length; i++) {
                valAndInd.push({ value: vals[i], index: i });
            }
            valAndInd.sort((a, b) => b.value - a.value);
            precision[b] = 0;
            for (let i = 0; i < k; i++) {
                if (valAndInd[i].index === targetsVals[b]) {
                    precision[b] = 1;
                    break;
                }
            }
        }
        if (predictions !== $predictions) {
            $predictions.dispose();
        }
        if (targets !== $targets) {
            $targets.dispose();
        }
        // Output precision has the same shape as targets.
        return tensor(precision, $targets.shape, 'bool');
    }
    const inTopKAsync = inTopKAsync_;

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the derivative of the filter of a 2D convolution.
     *
     * @param x The input tensor, of rank 4 or rank 3 of shape
     *     [batch, height, width, inChannels]. If rank 3, batch of 1 is assumed.
     * @param dy The dy image, of rank 4 or rank 3, of shape
     *     [batch, height, width, outDepth]. If rank 3, batch of 1 is assumed.
     * @param filterShape The shape of the filter, length 4,
     *     [filterHeight, filterWidth, inDepth, outDepth].
     * @param strides The strides of the convolution: [strideHeight,
     * strideWidth].
     * @param pad A string from: 'same', 'valid'. The type of padding algorithm
     *     used in the forward prop of the op.
     * @param dataFormat: An optional string from: "NHWC", "NCHW". Defaults to
     *     "NHWC". Specify the data format of the input and output data. With the
     *     default format "NHWC", the data is stored in the order of: [batch,
     *     height, width, channels].
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     */
    function conv2DBackpropFilter_(x, dy, filterShape, strides, pad, dataFormat = 'NHWC', dimRoundingMode) {
        let x4D = x;
        if (x.rank === 3) {
            x4D = reshape$3(x, [1, x.shape[0], x.shape[1], x.shape[2]]);
        }
        let dy4D = dy;
        if (dy4D.rank === 3) {
            dy4D = reshape$3(dy, [1, dy.shape[0], dy.shape[1], dy.shape[2]]);
        }
        assert$4(x4D.rank === 4, () => `Error in conv2dDerFilter: input must be rank 4, but got shape ` +
            `${x4D.shape}.`);
        assert$4(dy4D.rank === 4, () => `Error in conv2dDerFilter: dy must be rank 4, but got shape ` +
            `${dy4D.shape}.`);
        assert$4(filterShape.length === 4, () => `Error in conv2dDerFilter: filterShape must be length 4, but got ` +
            `${filterShape}.`);
        const inDepth = dataFormat === 'NHWC' ? x4D.shape[3] : x4D.shape[1];
        const outDepth = dataFormat === 'NHWC' ? dy4D.shape[3] : dy4D.shape[1];
        assert$4(inDepth === filterShape[2], () => `Error in conv2dDerFilter: depth of input ${inDepth}) must ` +
            `match input depth in filter (${filterShape[2]}.`);
        assert$4(outDepth === filterShape[3], () => `Error in conv2dDerFilter: depth of dy (${outDepth}) must ` +
            `match output depth for filter (${filterShape[3]}).`);
        checkPadOnDimRoundingMode('conv2dDerFilter', pad, dimRoundingMode);
        const inputs = { x: x4D, dy: dy4D };
        const attrs = { strides, pad, dataFormat, dimRoundingMode, filterShape };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        return ENGINE.runKernel(Conv2DBackpropFilter, inputs, attrs);
    }
    const conv2DBackpropFilter$2 = op({ conv2DBackpropFilter_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Returns gradient for fused activation.
    function getFusedDyActivation(dy, y, activation) {
        if (activation == null || activation === 'linear') {
            return dy;
        }
        if (activation === 'relu') {
            return mul(dy, step$2(y));
        }
        throw new Error(`Cannot compute gradient for fused activation ${activation}.`);
    }
    // Returns gradient for fused bias.
    function getFusedBiasGradient(bias, dyActivation) {
        let res = dyActivation;
        const reduceAxes = getReductionAxes(bias.shape, dyActivation.shape);
        if (reduceAxes.length > 0) {
            res = sum$2(res, reduceAxes);
        }
        return reshape$3(res, bias.shape);
    }
    function applyActivation$1(x, activation, preluActivationWeights, leakyreluAlpha) {
        if (activation === 'linear') {
            return x;
        }
        else if (activation === 'relu') {
            return relu$2(x);
        }
        else if (activation === 'elu') {
            return elu$4(x);
        }
        else if (activation === 'relu6') {
            return relu6$2(x);
        }
        else if (activation === 'prelu') {
            return prelu$3(x, preluActivationWeights);
        }
        else if (activation === 'leakyrelu') {
            return leakyRelu$2(x, leakyreluAlpha);
        }
        else if (activation === 'sigmoid') {
            return sigmoid$2(x);
        }
        throw new Error(`Unknown fused activation ${activation}.`);
    }
    // Whether we should call fused ops.
    const shouldFuse = (gradientDepth, activation) => {
        const gradientMode = gradientDepth > 0;
        return !gradientMode || activation === 'linear';
    };

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes a 2D convolution over the input x, optionally fused with adding a
     * bias and applying an activation.
     *
     * ```js
     * const inputDepth = 2;
     * const inShape = [2, 2, 2, inputDepth];
     * const outputDepth = 2;
     * const fSize = 1;
     * const pad = 0;
     * const strides = 1;
     *
     * const x = tf.tensor4d( [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
     * 16], inShape);
     * const w = tf.tensor4d([-1, 1, -2, 0.5], [fSize, fSize, inputDepth,
     * outputDepth]);
     *
     * tf.fused.conv2d({ x, filter: w, strides, pad, dataFormat: 'NHWC',
     * dilations: [1, 1], bias: tf.scalar(5), activation: 'relu' }).print();
     * ```
     *
     * @param obj An object with the following properties:
     * @param x The input tensor, of rank 4 or rank 3, of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is
     * assumed.
     * @param filter The filter, rank 4, of shape
     *     `[filterHeight, filterWidth, inDepth, outDepth]`.
     * @param strides The strides of the convolution: `[strideHeight,
     * strideWidth]`.
     * @param pad The type of padding algorithm.
     *   - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *   - `valid` output will be smaller than input if filter is larger
     *       than 1x1.
     *   - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dataFormat An optional string from: "NHWC", "NCHW". Defaults to
     *     "NHWC". Specify the data format of the input and output data. With the
     *     default format "NHWC", the data is stored in the order of: [batch,
     *     height, width, channels]. Only "NHWC" is currently supported.
     * @param dilations The dilation rates: `[dilationHeight, dilationWidth]`
     *     in which we sample input values across the height and width dimensions
     *     in atrous convolution. Defaults to `[1, 1]`. If `dilations` is a single
     *     number, then `dilationHeight == dilationWidth`. If it is greater than
     *     1, then all values of `strides` must be 1.
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     * @param bias Tensor to be added to the result.
     * @param activation Name of activation kernel (defaults to `linear`) to be
     *     applied
     *      after biasAdd.
     * @param preluActivationWeights Tensor of prelu weights to be applied as part
     *     of a `prelu` activation, typically the same shape as `x`.
     * @param leakyreluAlpha Optional. Alpha to be applied as part of a `leakyrelu`
     *     activation.
     */
    function fusedConv2d_({ x, filter, strides, pad, dataFormat = 'NHWC', dilations = [1, 1], dimRoundingMode, bias, activation = 'linear', preluActivationWeights, leakyreluAlpha }) {
        activation = activation || 'linear';
        if (shouldFuse(ENGINE.state.gradientDepth, activation) === false) {
            let result = conv2d$3(x, filter, strides, pad, dataFormat, dilations, dimRoundingMode);
            if (bias != null) {
                result = add$2(result, bias);
            }
            return applyActivation$1(result, activation, preluActivationWeights, leakyreluAlpha);
        }
        const $x = convertToTensor(x, 'x', 'conv2d', 'float32');
        const $filter = convertToTensor(filter, 'filter', 'conv2d', 'float32');
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            reshapedTo4D = true;
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
        }
        assert$4(x4D.rank === 4, () => `Error in fused conv2d: input must be rank 4, but got rank ` +
            `${x4D.rank}.`);
        assert$4($filter.rank === 4, () => `Error in fused conv2d: filter must be rank 4, but got rank ` +
            `${$filter.rank}.`);
        checkPadOnDimRoundingMode('fused conv2d', pad, dimRoundingMode);
        assert$4(x4D.shape[3] === $filter.shape[2], () => `Error in conv2d: depth of input (${x4D.shape[3]}) must match ` +
            `input depth for filter ${$filter.shape[2]}.`);
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in conv2D: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        assert$4(dataFormat === 'NHWC', () => `Error in conv2d: got dataFormat of ${dataFormat} but only NHWC is currently supported.`);
        const convInfo = computeConv2DInfo(x4D.shape, $filter.shape, strides, dilations, pad, dimRoundingMode);
        let $bias;
        if (bias != null) {
            $bias = convertToTensor(bias, 'bias', 'fused conv2d');
            [$bias] = makeTypesMatch($bias, $x);
            // According to TensorFlow, the bias is supposed be a 1-D tensor or a
            // scalar.
            if (dataFormat === 'NHWC') {
                assertAndGetBroadcastShape(convInfo.outShape, $bias.shape);
            }
            else {
                assert$4($bias.shape.length <= 1, () => `Error in fused conv2d: only supports scalar or 1-D Tensor ` +
                    `bias for NCHW format but got the bias of ` +
                    `rank-${$bias.shape.length}.`);
                assert$4($bias.shape.length === 0 || $bias.shape[0] === convInfo.outChannels ||
                    $bias.shape[0] === 1, () => `Error in fused conv2d: bias shape (${$bias.shape}) is not ` +
                    `compatible with the number of output channels ` +
                    `(${convInfo.outChannels})`);
            }
        }
        let $preluActivationWeights;
        if (preluActivationWeights != null) {
            $preluActivationWeights = convertToTensor(preluActivationWeights, 'prelu weights', 'fused conv2d');
        }
        const grad = (dy, saved) => {
            const [$filter, x4D, y, $bias] = saved;
            const dyActivation = getFusedDyActivation(dy, y, activation);
            assert$4(tupleValuesAreOne(dilations), () => 'Error in gradient of fused conv2D: ' +
                `dilation rates greater than 1 ` +
                `are not yet supported in gradients. Got dilations '${dilations}'`);
            const xDer = conv2DBackpropInput$2(x4D.shape, dyActivation, $filter, strides, pad);
            const filterDer = conv2DBackpropFilter$2(x4D, dyActivation, $filter.shape, strides, pad);
            const der = [xDer, filterDer];
            if ($bias != null) {
                const biasDer = getFusedBiasGradient($bias, dyActivation);
                der.push(biasDer);
            }
            return der;
        };
        const inputs = {
            x: x4D,
            filter: $filter,
            bias: $bias,
            preluActivationWeights: $preluActivationWeights
        };
        const attrs = {
            strides,
            pad,
            dataFormat,
            dilations,
            dimRoundingMode,
            activation,
            leakyreluAlpha
        };
        // Depending on the the params passed in we will have different number of
        // inputs and thus a a different number of elements in the gradient.
        if (bias == null) {
            const customOp = customGrad((x4D, filter, save) => {
                let res = 
                // tslint:disable-next-line: no-unnecessary-type-assertion
                ENGINE.runKernel(FusedConv2D, inputs, attrs);
                save([filter, x4D, res]);
                if (reshapedTo4D) {
                    // tslint:disable-next-line: no-unnecessary-type-assertion
                    res = reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
                }
                return { value: res, gradFunc: grad };
            });
            return customOp(x4D, $filter);
        }
        else {
            const customOpWithBias = customGrad((x4D, filter, bias, save) => {
                let res = ENGINE.runKernel(FusedConv2D, inputs, attrs);
                save([filter, x4D, res, bias]);
                if (reshapedTo4D) {
                    // tslint:disable-next-line: no-unnecessary-type-assertion
                    res = reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
                }
                return { value: res, gradFunc: grad };
            });
            return customOpWithBias(x4D, $filter, $bias);
        }
    }
    const conv2d$2 = op({ fusedConv2d_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthwiseConv2dNativeBackpropFilter_(x, dy, filterShape, strides, pad, dilations = [1, 1], dimRoundingMode) {
        let x4D = x;
        if (x.rank === 3) {
            x4D = reshape$3(x, [1, x.shape[0], x.shape[1], x.shape[2]]);
        }
        let dy4D = dy;
        if (dy4D.rank === 3) {
            dy4D = reshape$3(dy, [1, dy.shape[0], dy.shape[1], dy.shape[2]]);
        }
        const inputs = { x: x4D, dy: dy4D };
        const attrs = { strides, pad, dimRoundingMode, dilations, filterShape };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        return ENGINE.runKernel(DepthwiseConv2dNativeBackpropFilter, inputs, attrs);
    }
    const depthwiseConv2dNativeBackpropFilter$2 = op({ depthwiseConv2dNativeBackpropFilter_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthwiseConv2dNativeBackpropInput_(xShape, dy, filter, strides, pad, dilations = [1, 1], dimRoundingMode) {
        let dy4D = dy;
        let reshapedTo4D = false;
        if (dy.rank === 3) {
            reshapedTo4D = true;
            dy4D = reshape$3(dy, [1, dy.shape[0], dy.shape[1], dy.shape[2]]);
        }
        const inputs = { dy: dy4D, filter };
        const attrs = { strides, pad, dimRoundingMode, dilations, inputShape: xShape };
        const res = 
        // tslint:disable-next-line: no-unnecessary-type-assertion
        ENGINE.runKernel(DepthwiseConv2dNativeBackpropInput, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const depthwiseConv2dNativeBackpropInput$2 = op({ depthwiseConv2dNativeBackpropInput_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes depthwise 2D convolution, optionally fused with adding a
     * bias and applying an activation.
     *
     * Given a 4D `input` array and a `filter` array of shape
     * `[filterHeight, filterWidth, inChannels, channelMultiplier]` containing
     * `inChannels` convolutional filters of depth 1, this op applies a
     * different filter to each input channel (expanding from 1 channel to
     * `channelMultiplier` channels for each), then concatenates the results
     * together. The output has `inChannels * channelMultiplier` channels.
     *
     * See
     * [https://www.tensorflow.org/api_docs/python/tf/nn/depthwise_conv2d](
     *     https://www.tensorflow.org/api_docs/python/tf/nn/depthwise_conv2d)
     * for more details.
     *
     * @param obj An object with the following properties:
     * @param x The input tensor, of rank 4 or rank 3, of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is
     * assumed.
     * @param filter The filter tensor, rank 4, of shape
     *     `[filterHeight, filterWidth, inChannels, channelMultiplier]`.
     * @param strides The strides of the convolution: `[strideHeight,
     * strideWidth]`. If strides is a single number, then `strideHeight ==
     * strideWidth`.
     * @param pad The type of padding algorithm.
     *   - `same` and stride 1: output will be of same size as input,
     *       regardless of filter size.
     *   - `valid`: output will be smaller than input if filter is larger
     *       than 1x1.
     *   - For more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dilations The dilation rates: `[dilationHeight, dilationWidth]`
     *     in which we sample input values across the height and width dimensions
     *     in atrous convolution. Defaults to `[1, 1]`. If `rate` is a single
     *     number, then `dilationHeight == dilationWidth`. If it is greater than
     *     1, then all values of `strides` must be 1.
     * @param dataFormat: An optional string from: "NHWC", "NCHW". Defaults to
     *     "NHWC". Specify the data format of the input and output data. With the
     *     default format "NHWC", the data is stored in the order of: [batch,
     *     height, width, channels]. Only "NHWC" is currently supported.
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     * @param bias Tensor to be added to the result.
     * @param activation Name of activation kernel (defaults to `linear`).
     * @param preluActivationWeights Tensor of prelu weights to be applied as part
     *     of a `prelu` activation, typically the same shape as `x`.
     * @param leakyreluAlpha Optional. Alpha to be applied as part of a `leakyrelu`
     *     activation.
     */
    function fusedDepthwiseConv2d_({ x, filter, strides, pad, dataFormat = 'NHWC', dilations = [1, 1], dimRoundingMode, bias, activation = 'linear', preluActivationWeights, leakyreluAlpha }) {
        if (shouldFuse(ENGINE.state.gradientDepth, activation) === false) {
            let result = depthwiseConv2d$3(x, filter, strides, pad, dataFormat, dilations, dimRoundingMode);
            if (bias != null) {
                result = add$2(result, bias);
            }
            return applyActivation$1(result, activation, preluActivationWeights, leakyreluAlpha);
        }
        const $x = convertToTensor(x, 'x', 'depthwiseConv2d', 'float32');
        const $filter = convertToTensor(filter, 'filter', 'depthwiseConv2d', 'float32');
        let x4D = $x;
        let reshapedTo4D = false;
        if ($x.rank === 3) {
            reshapedTo4D = true;
            x4D = reshape$3($x, [1, $x.shape[0], $x.shape[1], $x.shape[2]]);
        }
        assert$4(x4D.rank === 4, () => `Error in fused depthwiseConv2d: input must be rank 4, but got ` +
            `rank ${x4D.rank}.`);
        assert$4($filter.rank === 4, () => `Error in fused depthwiseConv2d: filter must be rank 4, ` +
            `but got rank ${$filter.rank}.`);
        assert$4(x4D.shape[3] === $filter.shape[2], () => `Error in fused depthwiseConv2d: number of input channels ` +
            `(${x4D.shape[3]}) must match the inChannels dimension in ` +
            `filter ${$filter.shape[2]}.`);
        if (dilations == null) {
            dilations = [1, 1];
        }
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in fused depthwiseConv2d: Either strides or dilations must ' +
            `be 1. Got strides ${strides} and dilations '${dilations}'`);
        checkPadOnDimRoundingMode('fused depthwiseConv2d', pad, dimRoundingMode);
        const convInfo = computeConv2DInfo(x4D.shape, $filter.shape, strides, dilations, pad, dimRoundingMode, true /* depthwise */);
        let $bias;
        if (bias != null) {
            $bias = convertToTensor(bias, 'bias', 'fused conv2d');
            [$bias] = makeTypesMatch($bias, $x);
            assertAndGetBroadcastShape(convInfo.outShape, $bias.shape);
        }
        let $preluActivationWeights;
        if (preluActivationWeights != null) {
            $preluActivationWeights = convertToTensor(preluActivationWeights, 'prelu weights', 'fused depthwiseConv2d');
        }
        const grad = (dy, saved) => {
            assert$4(tupleValuesAreOne(dilations), () => 'Error in gradient of fused depthwiseConv2d: dilation rates ' +
                `greater than 1 are not yet supported. Got dilations ` +
                `'${dilations}'`);
            const [$filter, x4D, y, bias] = saved;
            const dyActivation = getFusedDyActivation(dy, y, activation);
            const xDer = depthwiseConv2dNativeBackpropInput$2(x4D.shape, dyActivation, $filter, strides, pad, dilations, dimRoundingMode);
            const filterDer = depthwiseConv2dNativeBackpropFilter$2(x4D, dyActivation, $filter.shape, strides, pad, dilations, dimRoundingMode);
            if (bias != null) {
                const biasDer = getFusedBiasGradient($bias, dyActivation);
                return [xDer, filterDer, biasDer];
            }
            return [xDer, filterDer];
        };
        const inputs = {
            x: x4D,
            filter: $filter,
            bias: $bias,
            preluActivationWeights: $preluActivationWeights
        };
        const attrs = {
            strides,
            pad,
            dataFormat,
            dilations,
            dimRoundingMode,
            activation,
            leakyreluAlpha
        };
        // Depending on the the params passed in we will have different number of
        // inputs and thus a a different number of elements in the gradient.
        if (bias == null) {
            const customOp = customGrad((x4D, filter, save) => {
                // tslint:disable-next-line: no-unnecessary-type-assertion
                let res = ENGINE.runKernel(FusedDepthwiseConv2D, inputs, attrs);
                save([filter, x4D, res]);
                if (reshapedTo4D) {
                    // tslint:disable-next-line: no-unnecessary-type-assertion
                    res = reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
                }
                return { value: res, gradFunc: grad };
            });
            return customOp(x4D, $filter);
        }
        else {
            const customOpWithBias = customGrad((x4D, filter, bias, save) => {
                // tslint:disable-next-line: no-unnecessary-type-assertion
                let res = ENGINE.runKernel(FusedDepthwiseConv2D, inputs, attrs);
                save([filter, x4D, res, bias]);
                if (reshapedTo4D) {
                    // tslint:disable-next-line: no-unnecessary-type-assertion
                    res = reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
                }
                return { value: res, gradFunc: grad };
            });
            return customOpWithBias(x4D, $filter, $bias);
        }
    }
    const depthwiseConv2d$2 = op({ fusedDepthwiseConv2d_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the dot product of two matrices with optional activation and bias.
     *
     * ```js
     * const a = tf.tensor2d([-1, -2], [1, 2]);
     * const b = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     * const bias = tf.tensor2d([1, 2], [1, 2]);
     *
     * tf.fused.matMul({a, b, bias, activation: 'relu'}).print();
     * ```
     *
     * @param obj An object with the following properties:
     * - `a` First matrix in dot product operation.
     * - `b` Second matrix in dot product operation.
     * - `transposeA` If true, `a` is transposed before multiplication.
     * - `transposeB` If true, `b` is transposed before multiplication.
     * - `bias` Matrix to be added to the result.
     * - `activation` Name of activation kernel (defaults to `linear`).
     * - `preluActivationWeights` Tensor of prelu weights.
     * - `leakyreluAlpha` Alpha of leakyrelu.
     */
    function fusedMatMul_({ a, b, transposeA = false, transposeB = false, bias, activation = 'linear', preluActivationWeights, leakyreluAlpha, }) {
        if (shouldFuse(ENGINE.state.gradientDepth, activation) === false) {
            let result = matMul$1(a, b, transposeA, transposeB);
            if (bias != null) {
                result = add$2(result, bias);
            }
            return applyActivation$1(result, activation, preluActivationWeights, leakyreluAlpha);
        }
        let $a = convertToTensor(a, 'a', 'fused matMul');
        let $b = convertToTensor(b, 'b', 'fused matMul');
        [$a, $b] = makeTypesMatch($a, $b);
        const innerShapeA = transposeA ? $a.shape[$a.rank - 2] : $a.shape[$a.rank - 1];
        const innerShapeB = transposeB ? $b.shape[$b.rank - 1] : $b.shape[$b.rank - 2];
        const outerShapeA = transposeA ? $a.shape[$a.rank - 1] : $a.shape[$a.rank - 2];
        const outerShapeB = transposeB ? $b.shape[$b.rank - 2] : $b.shape[$b.rank - 1];
        const outerDimsA = $a.shape.slice(0, -2);
        const outerDimsB = $b.shape.slice(0, -2);
        const batchDimA = sizeFromShape(outerDimsA);
        const batchDimB = sizeFromShape(outerDimsB);
        assert$4(innerShapeA === innerShapeB, () => `Error in fused matMul: inner shapes (${innerShapeA}) and (` +
            `${innerShapeB}) of Tensors with shapes ${$a.shape} and ` +
            `${$b.shape} and transposeA=${transposeA}` +
            ` and transposeB=${transposeB} must match.`);
        const outShapeOuterDims = assertAndGetBroadcastShape($a.shape.slice(0, -2), $b.shape.slice(0, -2));
        const outShape = outShapeOuterDims.concat([outerShapeA, outerShapeB]);
        const a3D = transposeA ?
            reshape$3($a, [batchDimA, innerShapeA, outerShapeA]) :
            reshape$3($a, [batchDimA, outerShapeA, innerShapeA]);
        const b3D = transposeB ?
            reshape$3($b, [batchDimB, outerShapeB, innerShapeB]) :
            reshape$3($b, [batchDimB, innerShapeB, outerShapeB]);
        let $bias;
        if (bias != null) {
            $bias = convertToTensor(bias, 'bias', 'fused matMul');
            [$bias] = makeTypesMatch($bias, $a);
            assertAndGetBroadcastShape(outShape, $bias.shape);
        }
        let $preluActivationWeights;
        if (preluActivationWeights != null) {
            $preluActivationWeights = convertToTensor(preluActivationWeights, 'prelu weights', 'fused matMul');
        }
        const grad = (dy, saved) => {
            const [a3D, b3D, y, $bias] = saved;
            // we reshape dy because the result of the forward is not
            // necessarily going to be a 3d tensor due to a reshape done at the end of
            // the customOp.
            const dyActivation = getFusedDyActivation(reshape$3(dy, y.shape), y, activation);
            let aDer;
            let bDer;
            if (!transposeA && !transposeB) {
                aDer = matMul$1(dyActivation, b3D, false, true);
                bDer = matMul$1(a3D, dyActivation, true, false);
            }
            else if (!transposeA && transposeB) {
                aDer = matMul$1(dyActivation, b3D, false, false);
                bDer = matMul$1(dyActivation, a3D, true, false);
            }
            else if (transposeA && !transposeB) {
                aDer = matMul$1(b3D, dyActivation, false, true);
                bDer = matMul$1(a3D, dyActivation, false, false);
            }
            else {
                aDer = matMul$1(b3D, dyActivation, true, true);
                bDer = matMul$1(dyActivation, a3D, true, true);
            }
            if (bias != null) {
                const biasDer = getFusedBiasGradient($bias, dyActivation);
                return [aDer, bDer, biasDer];
            }
            else {
                return [aDer, bDer];
            }
        };
        const inputs = {
            a: a3D,
            b: b3D,
            bias: $bias,
            preluActivationWeights: $preluActivationWeights
        };
        const attrs = { transposeA, transposeB, activation, leakyreluAlpha };
        // Depending on the the params passed in we will have different number of
        // inputs and thus a a different number of elements in the gradient.
        if (bias == null) {
            const customOp = customGrad((a3D, b3D, save) => {
                const res = 
                // tslint:disable-next-line: no-unnecessary-type-assertion
                ENGINE.runKernel(_FusedMatMul, inputs, attrs);
                save([a3D, b3D, res]);
                return { value: reshape$3(res, outShape), gradFunc: grad };
            });
            return customOp(a3D, b3D);
        }
        else {
            const customOpWithBias = customGrad((a3D, b3D, $bias, save) => {
                const res = 
                // tslint:disable-next-line: no-unnecessary-type-assertion
                ENGINE.runKernel(_FusedMatMul, inputs, attrs);
                save([a3D, b3D, res, $bias]);
                return { value: reshape$3(res, outShape), gradFunc: grad };
            });
            return customOpWithBias(a3D, b3D, $bias);
        }
    }
    const matMul = op({ fusedMatMul_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */

    var fused_ops = /*#__PURE__*/Object.freeze({
        __proto__: null,
        conv2d: conv2d$2,
        depthwiseConv2d: depthwiseConv2d$2,
        matMul: matMul
    });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Generate a hamming window.
     *
     * See: https://en.wikipedia.org/wiki/Window_function#Hann_and_Hamming_windows
     *
     * ```js
     * tf.signal.hammingWindow(10).print();
     * ```
     * @param The length of window
     *
     * @doc {heading: 'Operations', subheading: 'Signal', namespace: 'signal'}
     */
    function hammingWindow_(windowLength) {
        return cosineWindow(windowLength, 0.54, 0.46);
    }
    const hammingWindow = op({ hammingWindow_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Generate a Hann window.
     *
     * See: https://en.wikipedia.org/wiki/Window_function#Hann_and_Hamming_windows
     *
     * ```js
     * tf.signal.hannWindow(10).print();
     * ```
     * @param The length of window
     *
     * @doc {heading: 'Operations', subheading: 'Signal', namespace: 'signal'}
     */
    function hannWindow_(windowLength) {
        return cosineWindow(windowLength, 0.5, 0.5);
    }
    const hannWindow = op({ hannWindow_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Expands input into frames of frameLength.
     * Slides a window size with frameStep.
     *
     * ```js
     * tf.signal.frame([1, 2, 3], 2, 1).print();
     * ```
     * @param signal The input tensor to be expanded
     * @param frameLength Length of each frame
     * @param frameStep The frame hop size in samples.
     * @param padEnd Whether to pad the end of signal with padValue.
     * @param padValue An number to use where the input signal does
     *     not exist when padEnd is True.
     *
     * @doc {heading: 'Operations', subheading: 'Signal', namespace: 'signal'}
     */
    function frame_(signal, frameLength, frameStep, padEnd = false, padValue = 0) {
        let start = 0;
        const output = [];
        while (start + frameLength <= signal.size) {
            output.push(slice$2(signal, start, frameLength));
            start += frameStep;
        }
        if (padEnd) {
            while (start < signal.size) {
                const padLen = (start + frameLength) - signal.size;
                const pad = concat$2([
                    slice$2(signal, start, frameLength - padLen), fill$2([padLen], padValue)
                ]);
                output.push(pad);
                start += frameStep;
            }
        }
        if (output.length === 0) {
            return tensor2d([], [0, frameLength]);
        }
        return reshape$3(concat$2(output), [output.length, frameLength]);
    }
    const frame = op({ frame_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the Short-time Fourier Transform of signals
     * See: https://en.wikipedia.org/wiki/Short-time_Fourier_transform
     *
     * ```js
     * const input = tf.tensor1d([1, 1, 1, 1, 1])
     * tf.signal.stft(input, 3, 1).print();
     * ```
     * @param signal 1-dimensional real value tensor.
     * @param frameLength The window length of samples.
     * @param frameStep The number of samples to step.
     * @param fftLength The size of the FFT to apply.
     * @param windowFn A callable that takes a window length and returns 1-d tensor.
     *
     * @doc {heading: 'Operations', subheading: 'Signal', namespace: 'signal'}
     */
    function stft_(signal, frameLength, frameStep, fftLength, windowFn = hannWindow) {
        if (fftLength == null) {
            fftLength = enclosingPowerOfTwo(frameLength);
        }
        const framedSignal = frame(signal, frameLength, frameStep);
        const windowedSignal = mul(framedSignal, windowFn(frameLength));
        return rfft(windowedSignal, fftLength);
    }
    const stft = op({ stft_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Extracts crops from the input image tensor and resizes them using bilinear
     * sampling or nearest neighbor sampling (possibly with aspect ratio change)
     * to a common output size specified by cropSize.
     *
     * @param image 4d tensor of shape `[batch,imageHeight,imageWidth, depth]`,
     *     where imageHeight and imageWidth must be positive, specifying the
     *     batch of images from which to take crops
     * @param boxes 2d float32 tensor of shape `[numBoxes, 4]`. Each entry is
     *     `[y1, x1, y2, x2]`, where `(y1, x1)` and `(y2, x2)` are the normalized
     *     coordinates of the box in the boxInd[i]'th image in the batch
     * @param boxInd 1d int32 tensor of shape `[numBoxes]` with values in range
     *     `[0, batch)` that specifies the image that the `i`-th box refers to.
     * @param cropSize 1d int32 tensor of 2 elements `[cropHeigh, cropWidth]`
     *     specifying the size to which all crops are resized to.
     * @param method Optional string from `'bilinear' | 'nearest'`,
     *     defaults to bilinear, which specifies the sampling method for resizing
     * @param extrapolationValue A threshold for deciding when to remove boxes based
     *     on score. Defaults to 0.
     * @return A 4D tensor of the shape `[numBoxes,cropHeight,cropWidth,depth]`
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    function cropAndResize_(image, boxes, boxInd, cropSize, method = 'bilinear', extrapolationValue = 0) {
        const $image = convertToTensor(image, 'image', 'cropAndResize');
        const $boxes = convertToTensor(boxes, 'boxes', 'cropAndResize', 'float32');
        const $boxInd = convertToTensor(boxInd, 'boxInd', 'cropAndResize', 'int32');
        const numBoxes = $boxes.shape[0];
        assert$4($image.rank === 4, () => 'Error in cropAndResize: image must be rank 4,' +
            `but got rank ${$image.rank}.`);
        assert$4($boxes.rank === 2 && $boxes.shape[1] === 4, () => `Error in cropAndResize: boxes must be have size [${numBoxes},4] ` +
            `but had shape ${$boxes.shape}.`);
        assert$4($boxInd.rank === 1 && $boxInd.shape[0] === numBoxes, () => `Error in cropAndResize: boxInd must be have size [${numBoxes}] ` +
            `but had shape ${$boxes.shape}.`);
        assert$4(cropSize.length === 2, () => `Error in cropAndResize: cropSize must be of length 2, but got ` +
            `length ${cropSize.length}.`);
        assert$4(cropSize[0] >= 1 && cropSize[1] >= 1, () => `cropSize must be atleast [1,1], but was ${cropSize}`);
        assert$4(method === 'bilinear' || method === 'nearest', () => `method must be bilinear or nearest, but was ${method}`);
        const inputs = { image: $image, boxes: $boxes, boxInd: $boxInd };
        const attrs = { method, extrapolationValue, cropSize };
        const res = ENGINE.runKernel(CropAndResize, inputs, attrs);
        return res;
    }
    const cropAndResize$2 = op({ cropAndResize_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Flips the image left to right. Currently available in the CPU, WebGL, and
     * WASM backends.
     *
     * @param image 4d tensor of shape `[batch, imageHeight, imageWidth, depth]`.
     */
    /** @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'} */
    function flipLeftRight_(image) {
        const $image = convertToTensor(image, 'image', 'flipLeftRight', 'float32');
        assert$4($image.rank === 4, () => 'Error in flipLeftRight: image must be rank 4,' +
            `but got rank ${$image.rank}.`);
        const inputs = { image: $image };
        const res = ENGINE.runKernel(FlipLeftRight, inputs, {});
        return res;
    }
    const flipLeftRight = op({ flipLeftRight_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts images from grayscale to RGB format.
     *
     * @param image A grayscale tensor to convert. The `image`'s last dimension must
     *     be size 1 with at least a two-dimensional shape.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    function grayscaleToRGB_(image) {
        const $image = convertToTensor(image, 'image', 'grayscaleToRGB');
        const lastDimsIdx = $image.rank - 1;
        const lastDims = $image.shape[lastDimsIdx];
        assert$4($image.rank >= 2, () => 'Error in grayscaleToRGB: images must be at least rank 2, ' +
            `but got rank ${$image.rank}.`);
        assert$4(lastDims === 1, () => 'Error in grayscaleToRGB: last dimension of a grayscale image ' +
            `should be size 1, but got size ${lastDims}.`);
        const reps = new Array($image.rank);
        reps.fill(1, 0, lastDimsIdx);
        reps[lastDimsIdx] = 3;
        return tile$3($image, reps);
    }
    const grayscaleToRGB = op({ grayscaleToRGB_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Rotates the input image tensor counter-clockwise with an optional offset
     * center of rotation. Currently available in the CPU, WebGL, and WASM backends.
     *
     * @param image 4d tensor of shape `[batch, imageHeight, imageWidth, depth]`.
     * @param radians The amount of rotation.
     * @param fillValue The value to fill in the empty space leftover
     *     after rotation. Can be either a single grayscale value (0-255), or an
     *     array of three numbers `[red, green, blue]` specifying the red, green,
     *     and blue channels. Defaults to `0` (black).
     * @param center The center of rotation. Can be either a single value (0-1), or
     *     an array of two numbers `[centerX, centerY]`. Defaults to `0.5` (rotates
     *     the image around its center).
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    function rotateWithOffset_(image, radians, fillValue = 0, center = 0.5) {
        const $image = convertToTensor(image, 'image', 'rotateWithOffset', 'float32');
        assert$4($image.rank === 4, () => 'Error in rotateWithOffset: image must be rank 4,' +
            `but got rank ${$image.rank}.`);
        const inputs = { image: $image };
        const attrs = { radians, fillValue, center };
        const res = ENGINE.runKernel(RotateWithOffset, inputs, attrs);
        return res;
    }
    const rotateWithOffset = op({ rotateWithOffset_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function nonMaxSuppSanityCheck(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma) {
        if (iouThreshold == null) {
            iouThreshold = 0.5;
        }
        if (scoreThreshold == null) {
            scoreThreshold = Number.NEGATIVE_INFINITY;
        }
        if (softNmsSigma == null) {
            softNmsSigma = 0.0;
        }
        const numBoxes = boxes.shape[0];
        maxOutputSize = Math.min(maxOutputSize, numBoxes);
        assert$4(0 <= iouThreshold && iouThreshold <= 1, () => `iouThreshold must be in [0, 1], but was '${iouThreshold}'`);
        assert$4(boxes.rank === 2, () => `boxes must be a 2D tensor, but was of rank '${boxes.rank}'`);
        assert$4(boxes.shape[1] === 4, () => `boxes must have 4 columns, but 2nd dimension was ${boxes.shape[1]}`);
        assert$4(scores.rank === 1, () => 'scores must be a 1D tensor');
        assert$4(scores.shape[0] === numBoxes, () => `scores has incompatible shape with boxes. Expected ${numBoxes}, ` +
            `but was ${scores.shape[0]}`);
        assert$4(0 <= softNmsSigma && softNmsSigma <= 1, () => `softNmsSigma must be in [0, 1], but was '${softNmsSigma}'`);
        return { maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Performs non maximum suppression of bounding boxes based on
     * iou (intersection over union).
     *
     * @param boxes a 2d tensor of shape `[numBoxes, 4]`. Each entry is
     *     `[y1, x1, y2, x2]`, where `(y1, x1)` and `(y2, x2)` are the corners of
     *     the bounding box.
     * @param scores a 1d tensor providing the box scores of shape `[numBoxes]`.
     * @param maxOutputSize The maximum number of boxes to be selected.
     * @param iouThreshold A float representing the threshold for deciding whether
     *     boxes overlap too much with respect to IOU. Must be between [0, 1].
     *     Defaults to 0.5 (50% box overlap).
     * @param scoreThreshold A threshold for deciding when to remove boxes based
     *     on score. Defaults to -inf, which means any score is accepted.
     * @return A 1D tensor with the selected box indices.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    function nonMaxSuppression_(boxes, scores, maxOutputSize, iouThreshold = 0.5, scoreThreshold = Number.NEGATIVE_INFINITY) {
        const $boxes = convertToTensor(boxes, 'boxes', 'nonMaxSuppression', 'float32');
        const $scores = convertToTensor(scores, 'scores', 'nonMaxSuppression', 'float32');
        const inputs = nonMaxSuppSanityCheck($boxes, $scores, maxOutputSize, iouThreshold, scoreThreshold);
        maxOutputSize = inputs.maxOutputSize;
        iouThreshold = inputs.iouThreshold;
        scoreThreshold = inputs.scoreThreshold;
        const attrs = { maxOutputSize, iouThreshold, scoreThreshold };
        return ENGINE.runKernel(NonMaxSuppressionV3, { boxes: $boxes, scores: $scores }, attrs);
    }
    const nonMaxSuppression = op({ nonMaxSuppression_ });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Inserts a value into a sorted array. This method allows duplicate, meaning it
     * allows inserting duplicate value, in which case, the element will be inserted
     * at the lowest index of the value.
     * @param arr The array to modify.
     * @param element The element to insert.
     * @param comparator Optional. If no comparator is specified, elements are
     * compared using array_util.defaultComparator, which is suitable for Strings
     * and Numbers in ascending arrays. If the array contains multiple instances of
     * the target value, the left-most instance will be returned. To provide a
     * comparator, it should take 2 arguments to compare and return a negative,
     * zero, or a positive number.
     */
    function binaryInsert(arr, element, comparator) {
        const index = binarySearch(arr, element, comparator);
        const insertionPoint = index < 0 ? -(index + 1) : index;
        arr.splice(insertionPoint, 0, element);
    }
    /**
     * Searches the array for the target using binary search, returns the index
     * of the found element, or position to insert if element not found. If no
     * comparator is specified, elements are compared using array_
     * util.defaultComparator, which is suitable for Strings and Numbers in
     * ascending arrays. If the array contains multiple instances of the target
     * value, the left-most instance will be returned.
     * @param arr The array to be searched in.
     * @param target The target to be searched for.
     * @param comparator Should take 2 arguments to compare and return a negative,
     *    zero, or a positive number.
     * @return Lowest index of the target value if found, otherwise the insertion
     *    point where the target should be inserted, in the form of
     *    (-insertionPoint - 1).
     */
    function binarySearch(arr, target, comparator) {
        return binarySearch_(arr, target, comparator || defaultComparator);
    }
    /**
     * Compares its two arguments for order.
     * @param a The first element to be compared.
     * @param b The second element to be compared.
     * @return A negative number, zero, or a positive number as the first
     *     argument is less than, equal to, or greater than the second.
     */
    function defaultComparator(a, b) {
        return a > b ? 1 : a < b ? -1 : 0;
    }
    function binarySearch_(arr, target, comparator) {
        let left = 0;
        let right = arr.length;
        let middle = 0;
        let found = false;
        while (left < right) {
            middle = left + ((right - left) >>> 1);
            const compareResult = comparator(target, arr[middle]);
            if (compareResult > 0) {
                left = middle + 1;
            }
            else {
                right = middle;
                // If compareResult is 0, the value is found. We record it is found,
                // and then keep looking because there may be duplicate.
                found = !compareResult;
            }
        }
        return found ? left : -left - 1;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function nonMaxSuppressionV3Impl$2(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold) {
        return nonMaxSuppressionImpl_(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, 0 /* softNmsSigma */);
    }
    function nonMaxSuppressionV4Impl$2(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, padToMaxOutputSize) {
        return nonMaxSuppressionImpl_(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, 0 /* softNmsSigma */, false /* returnScoresTensor */, padToMaxOutputSize /* padToMaxOutputSize */, true
        /* returnValidOutputs */ );
    }
    function nonMaxSuppressionV5Impl$2(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma) {
        return nonMaxSuppressionImpl_(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma, true /* returnScoresTensor */);
    }
    function nonMaxSuppressionImpl_(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma, returnScoresTensor = false, padToMaxOutputSize = false, returnValidOutputs = false) {
        // The list is sorted in ascending order, so that we can always pop the
        // candidate with the largest score in O(1) time.
        const candidates = [];
        for (let i = 0; i < scores.length; i++) {
            if (scores[i] > scoreThreshold) {
                candidates.push({ score: scores[i], boxIndex: i, suppressBeginIndex: 0 });
            }
        }
        candidates.sort(ascendingComparator);
        // If softNmsSigma is 0, the outcome of this algorithm is exactly same as
        // before.
        const scale = softNmsSigma > 0 ? (-0.5 / softNmsSigma) : 0.0;
        const selectedIndices = [];
        const selectedScores = [];
        while (selectedIndices.length < maxOutputSize && candidates.length > 0) {
            const candidate = candidates.pop();
            const { score: originalScore, boxIndex, suppressBeginIndex } = candidate;
            if (originalScore < scoreThreshold) {
                break;
            }
            // Overlapping boxes are likely to have similar scores, therefore we
            // iterate through the previously selected boxes backwards in order to
            // see if candidate's score should be suppressed. We use
            // suppressBeginIndex to track and ensure a candidate can be suppressed
            // by a selected box no more than once. Also, if the overlap exceeds
            // iouThreshold, we simply ignore the candidate.
            let ignoreCandidate = false;
            for (let j = selectedIndices.length - 1; j >= suppressBeginIndex; --j) {
                const iou = intersectionOverUnion(boxes, boxIndex, selectedIndices[j]);
                if (iou >= iouThreshold) {
                    ignoreCandidate = true;
                    break;
                }
                candidate.score =
                    candidate.score * suppressWeight(iouThreshold, scale, iou);
                if (candidate.score <= scoreThreshold) {
                    break;
                }
            }
            // At this point, if `candidate.score` has not dropped below
            // `scoreThreshold`, then we know that we went through all of the
            // previous selections and can safely update `suppressBeginIndex` to the
            // end of the selected array. Then we can re-insert the candidate with
            // the updated score and suppressBeginIndex back in the candidate list.
            // If on the other hand, `candidate.score` has dropped below the score
            // threshold, we will not add it back to the candidates list.
            candidate.suppressBeginIndex = selectedIndices.length;
            if (!ignoreCandidate) {
                // Candidate has passed all the tests, and is not suppressed, so
                // select the candidate.
                if (candidate.score === originalScore) {
                    selectedIndices.push(boxIndex);
                    selectedScores.push(candidate.score);
                }
                else if (candidate.score > scoreThreshold) {
                    // Candidate's score is suppressed but is still high enough to be
                    // considered, so add back to the candidates list.
                    binaryInsert(candidates, candidate, ascendingComparator);
                }
            }
        }
        // NonMaxSuppressionV4 feature: padding output to maxOutputSize.
        const validOutputs = selectedIndices.length;
        const elemsToPad = maxOutputSize - validOutputs;
        if (padToMaxOutputSize && elemsToPad > 0) {
            selectedIndices.push(...new Array(elemsToPad).fill(0));
            selectedScores.push(...new Array(elemsToPad).fill(0.0));
        }
        const result = { selectedIndices };
        if (returnScoresTensor) {
            result['selectedScores'] = selectedScores;
        }
        if (returnValidOutputs) {
            result['validOutputs'] = validOutputs;
        }
        return result;
    }
    function intersectionOverUnion(boxes, i, j) {
        const iCoord = boxes.subarray(i * 4, i * 4 + 4);
        const jCoord = boxes.subarray(j * 4, j * 4 + 4);
        const yminI = Math.min(iCoord[0], iCoord[2]);
        const xminI = Math.min(iCoord[1], iCoord[3]);
        const ymaxI = Math.max(iCoord[0], iCoord[2]);
        const xmaxI = Math.max(iCoord[1], iCoord[3]);
        const yminJ = Math.min(jCoord[0], jCoord[2]);
        const xminJ = Math.min(jCoord[1], jCoord[3]);
        const ymaxJ = Math.max(jCoord[0], jCoord[2]);
        const xmaxJ = Math.max(jCoord[1], jCoord[3]);
        const areaI = (ymaxI - yminI) * (xmaxI - xminI);
        const areaJ = (ymaxJ - yminJ) * (xmaxJ - xminJ);
        if (areaI <= 0 || areaJ <= 0) {
            return 0.0;
        }
        const intersectionYmin = Math.max(yminI, yminJ);
        const intersectionXmin = Math.max(xminI, xminJ);
        const intersectionYmax = Math.min(ymaxI, ymaxJ);
        const intersectionXmax = Math.min(xmaxI, xmaxJ);
        const intersectionArea = Math.max(intersectionYmax - intersectionYmin, 0.0) *
            Math.max(intersectionXmax - intersectionXmin, 0.0);
        return intersectionArea / (areaI + areaJ - intersectionArea);
    }
    // A Gaussian penalty function, this method always returns values in [0, 1].
    // The weight is a function of similarity, the more overlap two boxes are, the
    // smaller the weight is, meaning highly overlapping boxe will be significantly
    // penalized. On the other hand, a non-overlapping box will not be penalized.
    function suppressWeight(iouThreshold, scale, iou) {
        const weight = Math.exp(scale * iou * iou);
        return iou <= iouThreshold ? weight : 0.0;
    }
    function ascendingComparator(c1, c2) {
        // For objects with same scores, we make the object with the larger index go
        // first. In an array that pops from the end, this means that the object with
        // the smaller index will be popped first. This ensures the same output as
        // the TensorFlow python version.
        return (c1.score - c2.score) ||
            ((c1.score === c2.score) && (c2.boxIndex - c1.boxIndex));
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Performs non maximum suppression of bounding boxes based on
     * iou (intersection over union).
     *
     * This is the async version of `nonMaxSuppression`
     *
     * @param boxes a 2d tensor of shape `[numBoxes, 4]`. Each entry is
     *     `[y1, x1, y2, x2]`, where `(y1, x1)` and `(y2, x2)` are the corners of
     *     the bounding box.
     * @param scores a 1d tensor providing the box scores of shape `[numBoxes]`.
     * @param maxOutputSize The maximum number of boxes to be selected.
     * @param iouThreshold A float representing the threshold for deciding whether
     *     boxes overlap too much with respect to IOU. Must be between [0, 1].
     *     Defaults to 0.5 (50% box overlap).
     * @param scoreThreshold A threshold for deciding when to remove boxes based
     *     on score. Defaults to -inf, which means any score is accepted.
     * @return A 1D tensor with the selected box indices.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    async function nonMaxSuppressionAsync_(boxes, scores, maxOutputSize, iouThreshold = 0.5, scoreThreshold = Number.NEGATIVE_INFINITY) {
        const $boxes = convertToTensor(boxes, 'boxes', 'nonMaxSuppressionAsync');
        const $scores = convertToTensor(scores, 'scores', 'nonMaxSuppressionAsync');
        const inputs = nonMaxSuppSanityCheck($boxes, $scores, maxOutputSize, iouThreshold, scoreThreshold);
        maxOutputSize = inputs.maxOutputSize;
        iouThreshold = inputs.iouThreshold;
        scoreThreshold = inputs.scoreThreshold;
        const boxesAndScores = await Promise.all([$boxes.data(), $scores.data()]);
        const boxesVals = boxesAndScores[0];
        const scoresVals = boxesAndScores[1];
        // We call a cpu based impl directly with the typedarray data  here rather
        // than a kernel because all kernels are synchronous (and thus cannot await
        // .data()).
        const { selectedIndices } = nonMaxSuppressionV3Impl$2(boxesVals, scoresVals, maxOutputSize, iouThreshold, scoreThreshold);
        if ($boxes !== boxes) {
            $boxes.dispose();
        }
        if ($scores !== scores) {
            $scores.dispose();
        }
        return tensor1d(selectedIndices, 'int32');
    }
    const nonMaxSuppressionAsync = nonMaxSuppressionAsync_;

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Performs non maximum suppression of bounding boxes based on
     * iou (intersection over union).
     *
     * This op also supports a Soft-NMS mode (c.f.
     * Bodla et al, https://arxiv.org/abs/1704.04503) where boxes reduce the score
     * of other overlapping boxes, therefore favoring different regions of the image
     * with high scores. To enable this Soft-NMS mode, set the `softNmsSigma`
     * parameter to be larger than 0.
     *
     * @param boxes a 2d tensor of shape `[numBoxes, 4]`. Each entry is
     *     `[y1, x1, y2, x2]`, where `(y1, x1)` and `(y2, x2)` are the corners of
     *     the bounding box.
     * @param scores a 1d tensor providing the box scores of shape `[numBoxes]`.
     * @param maxOutputSize The maximum number of boxes to be selected.
     * @param iouThreshold A float representing the threshold for deciding whether
     *     boxes overlap too much with respect to IOU. Must be between [0, 1].
     *     Defaults to 0.5 (50% box overlap).
     * @param scoreThreshold A threshold for deciding when to remove boxes based
     *     on score. Defaults to -inf, which means any score is accepted.
     * @param softNmsSigma A float representing the sigma parameter for Soft NMS.
     *     When sigma is 0, it falls back to nonMaxSuppression.
     * @return A map with the following properties:
     *     - selectedIndices: A 1D tensor with the selected box indices.
     *     - selectedScores: A 1D tensor with the corresponding scores for each
     *       selected box.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    function nonMaxSuppressionWithScore_(boxes, scores, maxOutputSize, iouThreshold = 0.5, scoreThreshold = Number.NEGATIVE_INFINITY, softNmsSigma = 0.0) {
        const $boxes = convertToTensor(boxes, 'boxes', 'nonMaxSuppression');
        const $scores = convertToTensor(scores, 'scores', 'nonMaxSuppression');
        const params = nonMaxSuppSanityCheck($boxes, $scores, maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma);
        maxOutputSize = params.maxOutputSize;
        iouThreshold = params.iouThreshold;
        scoreThreshold = params.scoreThreshold;
        softNmsSigma = params.softNmsSigma;
        const inputs = { boxes: $boxes, scores: $scores };
        const attrs = { maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const result = ENGINE.runKernel(NonMaxSuppressionV5, inputs, attrs);
        return { selectedIndices: result[0], selectedScores: result[1] };
    }
    const nonMaxSuppressionWithScore = op({ nonMaxSuppressionWithScore_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Asynchronously performs non maximum suppression of bounding boxes based on
     * iou (intersection over union).
     *
     * This op also supports a Soft-NMS mode (c.f.
     * Bodla et al, https://arxiv.org/abs/1704.04503) where boxes reduce the score
     * of other overlapping boxes, therefore favoring different regions of the image
     * with high scores. To enable this Soft-NMS mode, set the `softNmsSigma`
     * parameter to be larger than 0.
     *
     * @param boxes a 2d tensor of shape `[numBoxes, 4]`. Each entry is
     *     `[y1, x1, y2, x2]`, where `(y1, x1)` and `(y2, x2)` are the corners of
     *     the bounding box.
     * @param scores a 1d tensor providing the box scores of shape `[numBoxes]`.
     * @param maxOutputSize The maximum number of boxes to be selected.
     * @param iouThreshold A float representing the threshold for deciding whether
     *     boxes overlap too much with respect to IOU. Must be between [0, 1].
     *     Defaults to 0.5 (50% box overlap).
     * @param scoreThreshold A threshold for deciding when to remove boxes based
     *     on score. Defaults to -inf, which means any score is accepted.
     * @param softNmsSigma A float representing the sigma parameter for Soft NMS.
     *     When sigma is 0, it falls back to nonMaxSuppression.
     * @return A map with the following properties:
     *     - selectedIndices: A 1D tensor with the selected box indices.
     *     - selectedScores: A 1D tensor with the corresponding scores for each
     *       selected box.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    async function nonMaxSuppressionWithScoreAsync_(boxes, scores, maxOutputSize, iouThreshold = 0.5, scoreThreshold = Number.NEGATIVE_INFINITY, softNmsSigma = 0.0) {
        const $boxes = convertToTensor(boxes, 'boxes', 'nonMaxSuppressionAsync');
        const $scores = convertToTensor(scores, 'scores', 'nonMaxSuppressionAsync');
        const params = nonMaxSuppSanityCheck($boxes, $scores, maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma);
        maxOutputSize = params.maxOutputSize;
        iouThreshold = params.iouThreshold;
        scoreThreshold = params.scoreThreshold;
        softNmsSigma = params.softNmsSigma;
        const boxesAndScores = await Promise.all([$boxes.data(), $scores.data()]);
        const boxesVals = boxesAndScores[0];
        const scoresVals = boxesAndScores[1];
        // We call a cpu based impl directly with the typedarray data  here rather
        // than a kernel because all kernels are synchronous (and thus cannot await
        // .data()).
        const { selectedIndices, selectedScores } = nonMaxSuppressionV5Impl$2(boxesVals, scoresVals, maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma);
        if ($boxes !== boxes) {
            $boxes.dispose();
        }
        if ($scores !== scores) {
            $scores.dispose();
        }
        return {
            selectedIndices: tensor1d(selectedIndices, 'int32'),
            selectedScores: tensor1d(selectedScores)
        };
    }
    const nonMaxSuppressionWithScoreAsync = nonMaxSuppressionWithScoreAsync_;

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Asynchronously performs non maximum suppression of bounding boxes based on
     * iou (intersection over union), with an option to pad results.
     *
     * @param boxes a 2d tensor of shape `[numBoxes, 4]`. Each entry is
     *     `[y1, x1, y2, x2]`, where `(y1, x1)` and `(y2, x2)` are the corners of
     *     the bounding box.
     * @param scores a 1d tensor providing the box scores of shape `[numBoxes]`.
     * @param maxOutputSize The maximum number of boxes to be selected.
     * @param iouThreshold A float representing the threshold for deciding whether
     *     boxes overlap too much with respect to IOU. Must be between [0, 1].
     *     Defaults to 0.5 (50% box overlap).
     * @param scoreThreshold A threshold for deciding when to remove boxes based
     *     on score. Defaults to -inf, which means any score is accepted.
     * @param padToMaxOutputSize Defalts to false. If true, size of output
     *     `selectedIndices` is padded to maxOutputSize.
     * @return A map with the following properties:
     *     - selectedIndices: A 1D tensor with the selected box indices.
     *     - validOutputs: A scalar denoting how many elements in `selectedIndices`
     *       are valid. Valid elements occur first, then padding.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    function nonMaxSuppressionPadded_(boxes, scores, maxOutputSize, iouThreshold = 0.5, scoreThreshold = Number.NEGATIVE_INFINITY, padToMaxOutputSize = false) {
        const $boxes = convertToTensor(boxes, 'boxes', 'nonMaxSuppression');
        const $scores = convertToTensor(scores, 'scores', 'nonMaxSuppression');
        const params = nonMaxSuppSanityCheck($boxes, $scores, maxOutputSize, iouThreshold, scoreThreshold, null /* softNmsSigma */);
        const $maxOutputSize = params.maxOutputSize;
        const $iouThreshold = params.iouThreshold;
        const $scoreThreshold = params.scoreThreshold;
        const inputs = { boxes: $boxes, scores: $scores };
        const attrs = {
            maxOutputSize: $maxOutputSize,
            iouThreshold: $iouThreshold,
            scoreThreshold: $scoreThreshold,
            padToMaxOutputSize
        };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const result = ENGINE.runKernel(NonMaxSuppressionV4, inputs, attrs);
        return { selectedIndices: result[0], validOutputs: result[1] };
    }
    const nonMaxSuppressionPadded = op({ nonMaxSuppressionPadded_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Asynchronously performs non maximum suppression of bounding boxes based on
     * iou (intersection over union), with an option to pad results.
     *
     * @param boxes a 2d tensor of shape `[numBoxes, 4]`. Each entry is
     *     `[y1, x1, y2, x2]`, where `(y1, x1)` and `(y2, x2)` are the corners of
     *     the bounding box.
     * @param scores a 1d tensor providing the box scores of shape `[numBoxes]`.
     * @param maxOutputSize The maximum number of boxes to be selected.
     * @param iouThreshold A float representing the threshold for deciding whether
     *     boxes overlap too much with respect to IOU. Must be between [0, 1].
     *     Defaults to 0.5 (50% box overlap).
     * @param scoreThreshold A threshold for deciding when to remove boxes based
     *     on score. Defaults to -inf, which means any score is accepted.
     * @param padToMaxOutputSize Defalts to false. If true, size of output
     *     `selectedIndices` is padded to maxOutputSize.
     * @return A map with the following properties:
     *     - selectedIndices: A 1D tensor with the selected box indices.
     *     - validOutputs: A scalar denoting how many elements in `selectedIndices`
     *       are valid. Valid elements occur first, then padding.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    async function nonMaxSuppressionPaddedAsync_(boxes, scores, maxOutputSize, iouThreshold = 0.5, scoreThreshold = Number.NEGATIVE_INFINITY, padToMaxOutputSize = false) {
        const $boxes = convertToTensor(boxes, 'boxes', 'nonMaxSuppressionAsync');
        const $scores = convertToTensor(scores, 'scores', 'nonMaxSuppressionAsync');
        const params = nonMaxSuppSanityCheck($boxes, $scores, maxOutputSize, iouThreshold, scoreThreshold, null /* softNmsSigma */);
        const $maxOutputSize = params.maxOutputSize;
        const $iouThreshold = params.iouThreshold;
        const $scoreThreshold = params.scoreThreshold;
        const [boxesVals, scoresVals] = await Promise.all([$boxes.data(), $scores.data()]);
        // We call a cpu based impl directly with the typedarray data here rather
        // than a kernel because all kernels are synchronous (and thus cannot await
        // .data()).
        const { selectedIndices, validOutputs } = nonMaxSuppressionV4Impl$2(boxesVals, scoresVals, $maxOutputSize, $iouThreshold, $scoreThreshold, padToMaxOutputSize);
        if ($boxes !== boxes) {
            $boxes.dispose();
        }
        if ($scores !== scores) {
            $scores.dispose();
        }
        return {
            selectedIndices: tensor1d(selectedIndices, 'int32'),
            validOutputs: scalar(validOutputs, 'int32')
        };
    }
    const nonMaxSuppressionPaddedAsync = nonMaxSuppressionPaddedAsync_;

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Bilinear resize a single 3D image or a batch of 3D images to a new shape.
     *
     * @param images The images, of rank 4 or rank 3, of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is assumed.
     * @param size The new shape `[newHeight, newWidth]` to resize the
     *     images to. Each channel is resized individually.
     * @param alignCorners Defaults to `false`. If true, rescale
     *     input by `(new_height - 1) / (height - 1)`, which exactly aligns the 4
     *     corners of images and resized images. If false, rescale by
     *     `new_height / height`. Treat similarly the width dimension.
     * @param halfPixelCenters Defaults to `false`. Whether to assume pixel centers
     *     are at 0.5, which would make the floating point coordinates of the top
     *     left pixel 0.5, 0.5.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    function resizeBilinear_(images, size, alignCorners = false, halfPixelCenters = false) {
        const $images = convertToTensor(images, 'images', 'resizeBilinear');
        assert$4($images.rank === 3 || $images.rank === 4, () => `Error in resizeBilinear: x must be rank 3 or 4, but got ` +
            `rank ${$images.rank}.`);
        assert$4(size.length === 2, () => `Error in resizeBilinear: new shape must 2D, but got shape ` +
            `${size}.`);
        assert$4(halfPixelCenters === false || alignCorners === false, () => `Error in resizeBilinear: If halfPixelCenters is true, ` +
            `alignCorners must be false.`);
        let batchImages = $images;
        let reshapedTo4D = false;
        if ($images.rank === 3) {
            reshapedTo4D = true;
            batchImages = reshape$3($images, [1, $images.shape[0], $images.shape[1], $images.shape[2]]);
        }
        const inputs = { images: batchImages };
        const attrs = { alignCorners, halfPixelCenters, size };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(ResizeBilinear, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const resizeBilinear$2 = op({ resizeBilinear_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * NearestNeighbor resize a batch of 3D images to a new shape.
     *
     * @param images The images, of rank 4 or rank 3, of shape
     *     `[batch, height, width, inChannels]`. If rank 3, batch of 1 is assumed.
     * @param size The new shape `[newHeight, newWidth]` to resize the
     *     images to. Each channel is resized individually.
     * @param alignCorners Defaults to False. If true, rescale
     *     input by `(new_height - 1) / (height - 1)`, which exactly aligns the 4
     *     corners of images and resized images. If false, rescale by
     *     `new_height / height`. Treat similarly the width dimension.
     * @param halfPixelCenters Defaults to `false`. Whether to assumes pixels are of
     *      half the actual dimensions, and yields more accurate resizes. This flag
     *      would also make the floating point coordinates of the top left pixel
     *      0.5, 0.5.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    function resizeNearestNeighbor_(images, size, alignCorners = false, halfPixelCenters = false) {
        const $images = convertToTensor(images, 'images', 'resizeNearestNeighbor');
        assert$4($images.rank === 3 || $images.rank === 4, () => `Error in resizeNearestNeighbor: x must be rank 3 or 4, but got ` +
            `rank ${$images.rank}.`);
        assert$4(size.length === 2, () => `Error in resizeNearestNeighbor: new shape must 2D, but got shape ` +
            `${size}.`);
        assert$4($images.dtype === 'float32' || $images.dtype === 'int32', () => '`images` must have `int32` or `float32` as dtype');
        assert$4(halfPixelCenters === false || alignCorners === false, () => `Error in resizeNearestNeighbor: If halfPixelCenters is true, ` +
            `alignCorners must be false.`);
        let batchImages = $images;
        let reshapedTo4D = false;
        if ($images.rank === 3) {
            reshapedTo4D = true;
            batchImages = reshape$3($images, [1, $images.shape[0], $images.shape[1], $images.shape[2]]);
        }
        const inputs = { images: batchImages };
        const attrs = { alignCorners, halfPixelCenters, size };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(ResizeNearestNeighbor, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const resizeNearestNeighbor$2 = op({ resizeNearestNeighbor_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * https://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Performs image binarization with corresponding threshold
     * (depends on the method)value, which creates a binary image from a grayscale.
     * @param image 3d tensor of shape [imageHeight,imageWidth, depth],
     * where imageHeight and imageWidth must be positive.The image color
     * range should be [0, 255].
     * @param method Optional string from `'binary' | 'otsu'`
     * which specifies the method for thresholding. Defaults to 'binary'.
     * @param inverted Optional boolean whichspecifies
     * if colours should be inverted. Defaults to false.
     * @param threshValue Optional number which defines threshold value from 0 to 1.
     * Defaults to 0.5.
     * @return A 3d tensor of shape [imageHeight,imageWidth, depth], which
     * contains binarized image.
     */
    function threshold_(image, method = 'binary', inverted = false, threshValue = 0.5) {
        const $image = convertToTensor(image, 'image', 'threshold');
        /* 0.2989, 0.5870, 0.1140 are represent luma coefficients in CCIR601.
        Reference for converting between RGB and grayscale: https://en.wikipedia.org/wiki/Luma_%28video%29  */
        const RED_INTENCITY_COEF = 0.2989;
        const GREEN_INTENCITY_COEF = 0.5870;
        const BLUE_INTENCITY_COEF = 0.1140;
        const totalPixelsInImage = $image.shape[0] * $image.shape[1];
        let $threshold = mul(tensor1d([threshValue]), 255);
        let r, g, b, grayscale;
        assert$4($image.rank === 3, () => 'Error in threshold: image must be rank 3,' +
            `but got rank ${$image.rank}.`);
        assert$4($image.shape[2] === 3 || $image.shape[2] === 1, () => 'Error in threshold: ' +
            'image color channel must be equal to 3 or 1' +
            `but got ${$image.shape[2]}.`);
        assert$4($image.dtype === 'int32' || $image.dtype === 'float32', () => 'Error in dtype: image dtype must be int32 or float32,' +
            `but got dtype ${$image.dtype}.`);
        assert$4(method === 'otsu' || method === 'binary', () => `Method must be binary or otsu, but was ${method}`);
        if ($image.shape[2] === 3) {
            [r, g, b] = split$2($image, [1, 1, 1], -1);
            const $r = mul(r, RED_INTENCITY_COEF);
            const $g = mul(g, GREEN_INTENCITY_COEF);
            const $b = mul(b, BLUE_INTENCITY_COEF);
            grayscale = add$2(add$2($r, $g), $b);
        }
        else {
            grayscale = image;
        }
        if (method === 'otsu') {
            const $histogram = bincount$2(cast$3(round$2(grayscale), 'int32'), tensor([]), 256);
            $threshold = otsu($histogram, totalPixelsInImage);
        }
        const invCondition = inverted ?
            lessEqual$2(grayscale, $threshold) : greater$3(grayscale, $threshold);
        const result = cast$3(mul(invCondition, 255), 'int32');
        return result;
    }
    function otsu(histogram, total) {
        let bestThresh = tensor1d([-1]);
        let bestInBetVar = tensor1d([0]);
        let cInBetVar = tensor1d([0]);
        let classFirst, classSecond, meanFirst, meanSec, weightForeground, weightBack;
        for (let index = 0; index < histogram.size - 1; index++) {
            classFirst = slice$2(histogram, 0, index + 1);
            classSecond = slice$2(histogram, index + 1);
            weightForeground = div$1(sum$2(classFirst), total);
            weightBack = div$1(sum$2(classSecond), total);
            const meanFirstDivA = sum$2(mul(classFirst, range$4(0, classFirst.size)));
            meanFirst = div$1(meanFirstDivA, sum$2(classFirst));
            const meanSecFill = fill$2(classSecond.shape, classFirst.size);
            const meanSecAdd = add$2(range$4(0, classSecond.size), meanSecFill);
            const meanSecMul = mul(classSecond, (meanSecAdd));
            meanSec = div$1(sum$2(meanSecMul), sum$2(classSecond));
            const cInBetVarSubA = sub$2(meanFirst, meanSec);
            const cInBetVarSubB = sub$2(meanFirst, meanSec);
            const cInBetVarMul = mul(weightForeground, weightBack);
            cInBetVar = mul(mul(cInBetVarMul, cInBetVarSubA), cInBetVarSubB);
            const condition = greater$3(cInBetVar, bestInBetVar);
            bestInBetVar = where(condition, cInBetVar, bestInBetVar);
            bestThresh = where(condition, tensor1d([index]), bestThresh);
        }
        return bestThresh;
    }
    const threshold$1 = op({ threshold_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Applies the given transform(s) to the image(s).
     *
     * @param image 4d tensor of shape `[batch, imageHeight, imageWidth, depth]`.
     * @param transforms Projective transform matrix/matrices. A tensor1d of length
     *     8 or tensor of size N x 8. If one row of transforms is [a0, a1, a2, b0
     *     b1, b2, c0, c1], then it maps the output point (x, y) to a transformed
     *     input point (x', y') = ((a0 x + a1 y + a2) / k, (b0 x + b1 y + b2) / k),
     *     where k = c0 x + c1 y + 1. The transforms are inverted compared to the
     *     transform mapping input points to output points.
     * @param interpolation Interpolation mode.
     *     Supported values: 'nearest', 'bilinear'. Default to 'nearest'.
     * @param fillMode Points outside the boundaries of the input are filled
     *     according to the given mode, one of 'constant', 'reflect', 'wrap',
     *     'nearest'. Default to 'constant'.
     *     'reflect': (d c b a | a b c d | d c b a ) The input is extended by
     *     reflecting about the edge of the last pixel.
     *     'constant': (k k k k | a b c d | k k k k) The input is extended by
     *     filling all values beyond the edge with the same constant value k.
     *     'wrap': (a b c d | a b c d | a b c d) The input is extended by
     *     wrapping around to the opposite edge.
     *     'nearest': (a a a a | a b c d | d d d d) The input is extended by
     *     the nearest pixel.
     * @param fillValue A float represents the value to be filled outside the
     *     boundaries when fillMode is 'constant'.
     * @param Output dimension after the transform, [height, width]. If undefined,
     *     output is the same size as input image.
     *
     * @doc {heading: 'Operations', subheading: 'Images', namespace: 'image'}
     */
    function transform_(image, transforms, interpolation = 'nearest', fillMode = 'constant', fillValue = 0, outputShape) {
        const $image = convertToTensor(image, 'image', 'transform', 'float32');
        const $transforms = convertToTensor(transforms, 'transforms', 'transform', 'float32');
        assert$4($image.rank === 4, () => 'Error in transform: image must be rank 4,' +
            `but got rank ${$image.rank}.`);
        assert$4($transforms.rank === 2 &&
            ($transforms.shape[0] === $image.shape[0] ||
                $transforms.shape[0] === 1) &&
            $transforms.shape[1] === 8, () => `Error in transform: Input transform should be batch x 8 or 1 x 8`);
        assert$4(outputShape == null || outputShape.length === 2, () => 'Error in transform: outputShape must be [height, width] or null, ' +
            `but got ${outputShape}.`);
        const inputs = { image: $image, transforms: $transforms };
        const attrs = { interpolation, fillMode, fillValue, outputShape };
        return ENGINE.runKernel(Transform$1, inputs, attrs);
    }
    const transform$2 = op({ transform_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Copy a tensor setting everything outside a central band in each innermost
     * matrix to zero.
     *
     * The band part is computed as follows: Assume input has `k` dimensions
     * `[I, J, K, ..., M, N]`, then the output is a tensor with the same shape where
     * `band[i, j, k, ..., m, n] = in_band(m, n) * input[i, j, k, ..., m, n]`.
     * The indicator function
     * `in_band(m, n) = (num_lower < 0 || (m-n) <= num_lower))`
     * `&& (num_upper < 0 || (n-m) <= num_upper)`
     *
     * ```js
     * const x = tf.tensor2d([[ 0,  1,  2, 3],
     *                        [-1,  0,  1, 2],
     *                        [-2, -1,  0, 1],
     *                        [-3, -2, -1, 0]]);
     * let y = tf.linalg.bandPart(x, 1, -1);
     * y.print(); // [[ 0,  1,  2, 3],
     *            //  [-1,  0,  1, 2],
     *            //  [ 0, -1,  0, 1],
     *            //  [ 0, 0 , -1, 0]]
     * let z = tf.linalg.bandPart(x, 2, 1);
     * z.print(); // [[ 0,  1,  0, 0],
     *            //  [-1,  0,  1, 0],
     *            //  [-2, -1,  0, 1],
     *            //  [ 0, -2, -1, 0]]
     * ```
     *
     * @param x Rank `k` tensor
     * @param numLower Number of subdiagonals to keep.
     *   If negative, keep entire lower triangle.
     * @param numUpper Number of subdiagonals to keep.
     *   If negative, keep entire upper triangle.
     * @returns Rank `k` tensor of the same shape as input.
     *   The extracted banded tensor.
     *
     * @doc {heading:'Operations', subheading:'Linear Algebra', namespace:'linalg'}
     */
    function bandPart_(a, numLower, numUpper) {
        assert$4(numLower % 1 === 0, () => `bandPart(): numLower must be an integer, got ${numLower}.`);
        assert$4(numUpper % 1 === 0, () => `bandPart(): numUpper must be an integer, got ${numUpper}.`);
        const $a = convertToTensor(a, 'a', 'bandPart');
        assert$4($a.rank >= 2, () => `bandPart(): Rank must be at least 2, got ${$a.rank}.`);
        const shape = $a.shape;
        const [M, N] = $a.shape.slice(-2);
        if (!(numLower <= M)) {
            throw new Error(`bandPart(): numLower (${numLower})` +
                ` must not be greater than the number of rows (${M}).`);
        }
        if (!(numUpper <= N)) {
            throw new Error(`bandPart(): numUpper (${numUpper})` +
                ` must not be greater than the number of columns (${N}).`);
        }
        if (numLower < 0) {
            numLower = M;
        }
        if (numUpper < 0) {
            numUpper = N;
        }
        const i = reshape$3(range$4(0, M, 1, 'int32'), [-1, 1]);
        const j = range$4(0, N, 1, 'int32');
        const ij = sub$2(i, j);
        const inBand = logicalAnd$2(lessEqual$2(ij, scalar(+numLower, 'int32')), greaterEqual$2(ij, scalar(-numUpper, 'int32')));
        const zero = zeros$2([M, N], $a.dtype);
        return reshape$3(stack(unstack(reshape$3($a, [-1, M, N]))
            .map(mat => where(inBand, mat, zero))), shape);
    }
    const bandPart = op({ bandPart_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Gram-Schmidt orthogonalization.
     *
     * ```js
     * const x = tf.tensor2d([[1, 2], [3, 4]]);
     * let y = tf.linalg.gramSchmidt(x);
     * y.print();
     * console.log('Othogonalized:');
     * y.dot(y.transpose()).print();  // should be nearly the identity matrix.
     * console.log('First row direction maintained:');
     * const data = await y.array();
     * console.log(data[0][1] / data[0][0]);  // should be nearly 2.
     * ```
     *
     * @param xs The vectors to be orthogonalized, in one of the two following
     *   formats:
     *   - An Array of `tf.Tensor1D`.
     *   - A `tf.Tensor2D`, i.e., a matrix, in which case the vectors are the rows
     *     of `xs`.
     *   In each case, all the vectors must have the same length and the length
     *   must be greater than or equal to the number of vectors.
     * @returns The orthogonalized and normalized vectors or matrix.
     *   Orthogonalization means that the vectors or the rows of the matrix
     *   are orthogonal (zero inner products). Normalization means that each
     *   vector or each row of the matrix has an L2 norm that equals `1`.
     *
     * @doc {heading:'Operations', subheading:'Linear Algebra', namespace:'linalg'}
     */
    function gramSchmidt_(xs) {
        let inputIsTensor2D;
        if (Array.isArray(xs)) {
            inputIsTensor2D = false;
            assert$4(xs != null && xs.length > 0, () => 'Gram-Schmidt process: input must not be null, undefined, or ' +
                'empty');
            const dim = xs[0].shape[0];
            for (let i = 1; i < xs.length; ++i) {
                assert$4(xs[i].shape[0] === dim, () => 'Gram-Schmidt: Non-unique lengths found in the input vectors: ' +
                    `(${xs[i].shape[0]} vs. ${dim})`);
            }
        }
        else {
            inputIsTensor2D = true;
            xs = split$2(xs, xs.shape[0], 0).map(x => squeeze(x, [0]));
        }
        assert$4(xs.length <= xs[0].shape[0], () => `Gram-Schmidt: Number of vectors (${xs.length}) exceeds ` +
            `number of dimensions (${xs[0].shape[0]}).`);
        const ys = [];
        const xs1d = xs;
        for (let i = 0; i < xs.length; ++i) {
            ys.push(ENGINE.tidy(() => {
                let x = xs1d[i];
                if (i > 0) {
                    for (let j = 0; j < i; ++j) {
                        const proj = mul(sum$2(mul(ys[j], x)), ys[j]);
                        x = sub$2(x, proj);
                    }
                }
                return div$1(x, norm(x, 'euclidean'));
            }));
        }
        if (inputIsTensor2D) {
            return stack(ys, 0);
        }
        else {
            return ys;
        }
    }
    const gramSchmidt = op({ gramSchmidt_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Compute QR decomposition of m-by-n matrix using Householder transformation.
     *
     * Implementation based on
     *   [http://www.cs.cornell.edu/~bindel/class/cs6210-f09/lec18.pdf]
     * (http://www.cs.cornell.edu/~bindel/class/cs6210-f09/lec18.pdf)
     *
     * ```js
     * const a = tf.tensor2d([[1, 2], [3, 4]]);
     * let [q, r] = tf.linalg.qr(a);
     * console.log('Q');
     * q.print();
     * console.log('R');
     * r.print();
     * console.log('Orthogonalized');
     * q.dot(q.transpose()).print()  // should be nearly the identity matrix.
     * console.log('Reconstructed');
     * q.dot(r).print(); // should be nearly [[1, 2], [3, 4]];
     * ```
     *
     * @param x The `tf.Tensor` to be QR-decomposed. Must have rank >= 2. Suppose
     *   it has the shape `[..., M, N]`.
     * @param fullMatrices An optional boolean parameter. Defaults to `false`.
     *   If `true`, compute full-sized `Q`. If `false` (the default),
     *   compute only the leading N columns of `Q` and `R`.
     * @returns An `Array` of two `tf.Tensor`s: `[Q, R]`. `Q` is a unitary matrix,
     *   i.e., its columns all have unit norm and are mutually orthogonal.
     *   If `M >= N`,
     *     If `fullMatrices` is `false` (default),
     *       - `Q` has a shape of `[..., M, N]`,
     *       - `R` has a shape of `[..., N, N]`.
     *     If `fullMatrices` is `true` (default),
     *       - `Q` has a shape of `[..., M, M]`,
     *       - `R` has a shape of `[..., M, N]`.
     *   If `M < N`,
     *     - `Q` has a shape of `[..., M, M]`,
     *     - `R` has a shape of `[..., M, N]`.
     * @throws If the rank of `x` is less than 2.
     *
     * @doc {heading:'Operations',
     *       subheading:'Linear Algebra',
     *       namespace:'linalg'}
     */
    function qr_(x, fullMatrices = false) {
        assert$4(x.rank >= 2, () => `qr() requires input tensor to have a rank >= 2, but got rank ${x.rank}`);
        if (x.rank === 2) {
            return qr2d(x, fullMatrices);
        }
        else {
            // Rank > 2.
            // TODO(cais): Below we split the input into individual 2D tensors,
            //   perform QR decomposition on them and then stack the results back
            //   together. We should explore whether this can be parallelized.
            const outerDimsProd = x.shape.slice(0, x.shape.length - 2)
                .reduce((value, prev) => value * prev);
            const x2ds = unstack(reshape$3(x, [
                outerDimsProd, x.shape[x.shape.length - 2],
                x.shape[x.shape.length - 1]
            ]), 0);
            const q2ds = [];
            const r2ds = [];
            x2ds.forEach(x2d => {
                const [q2d, r2d] = qr2d(x2d, fullMatrices);
                q2ds.push(q2d);
                r2ds.push(r2d);
            });
            const q = reshape$3(stack(q2ds, 0), x.shape);
            const r = reshape$3(stack(r2ds, 0), x.shape);
            return [q, r];
        }
    }
    function qr2d(x, fullMatrices = false) {
        return ENGINE.tidy(() => {
            assert$4(x.shape.length === 2, () => `qr2d() requires a 2D Tensor, but got a ${x.shape.length}D Tensor.`);
            const m = x.shape[0];
            const n = x.shape[1];
            let q = eye(m); // Orthogonal transform so far.
            let r = clone(x); // Transformed matrix so far.
            const one2D = tensor2d([[1]], [1, 1]);
            let w = clone(one2D);
            const iters = m >= n ? n : m;
            for (let j = 0; j < iters; ++j) {
                // This tidy within the for-loop ensures we clean up temporary
                // tensors as soon as they are no longer needed.
                const rTemp = r;
                const wTemp = w;
                const qTemp = q;
                [w, r, q] = ENGINE.tidy(() => {
                    // Find H = I - tau * w * w', to put zeros below R(j, j).
                    const rjEnd1 = slice$2(r, [j, j], [m - j, 1]);
                    const normX = norm(rjEnd1);
                    const rjj = slice$2(r, [j, j], [1, 1]);
                    // The sign() function returns 0 on 0, which causes division by zero.
                    const s = where(greater$3(rjj, 0), tensor2d([[-1]]), tensor2d([[1]]));
                    const u1 = sub$2(rjj, mul(s, normX));
                    const wPre = div$1(rjEnd1, u1);
                    if (wPre.shape[0] === 1) {
                        w = clone(one2D);
                    }
                    else {
                        w = concat$2([
                            one2D,
                            slice$2(wPre, [1, 0], [wPre.shape[0] - 1, wPre.shape[1]])
                        ], 0);
                    }
                    const tau = neg$2(div$1(matMul$1(s, u1), normX));
                    // -- R := HR, Q := QH.
                    const rjEndAll = slice$2(r, [j, 0], [m - j, n]);
                    const tauTimesW = mul(tau, w);
                    const wT = transpose$2(w);
                    if (j === 0) {
                        r = sub$2(rjEndAll, matMul$1(tauTimesW, matMul$1(wT, rjEndAll)));
                    }
                    else {
                        const rTimesTau = sub$2(rjEndAll, matMul$1(tauTimesW, matMul$1(wT, rjEndAll)));
                        r = concat$2([slice$2(r, [0, 0], [j, n]), rTimesTau], 0);
                    }
                    const tawTimesWT = transpose$2(tauTimesW);
                    const qAllJEnd = slice$2(q, [0, j], [m, q.shape[1] - j]);
                    if (j === 0) {
                        q = sub$2(qAllJEnd, matMul$1(matMul$1(qAllJEnd, w), tawTimesWT));
                    }
                    else {
                        const qTimesTau = sub$2(qAllJEnd, matMul$1(matMul$1(qAllJEnd, w), tawTimesWT));
                        q = concat$2([slice$2(q, [0, 0], [m, j]), qTimesTau], 1);
                    }
                    return [w, r, q];
                });
                dispose([rTemp, wTemp, qTemp]);
            }
            if (!fullMatrices && m > n) {
                q = slice$2(q, [0, 0], [m, n]);
                r = slice$2(r, [0, 0], [n, n]);
            }
            return [q, r];
        });
    }
    const qr = op({ qr_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    var Reduction;
    (function (Reduction) {
        Reduction[Reduction["NONE"] = 0] = "NONE";
        Reduction[Reduction["MEAN"] = 1] = "MEAN";
        Reduction[Reduction["SUM"] = 2] = "SUM";
        Reduction[Reduction["SUM_BY_NONZERO_WEIGHTS"] = 3] = "SUM_BY_NONZERO_WEIGHTS";
    })(Reduction || (Reduction = {}));

    /**
     * Computes the weighted loss between two tensors.
     *
     * @param losses Tensor of shape `[batch_size, d1, ... dN]`.
     * @param weights Tensor whose rank is either 0, or the same rank as
     *    `losses`, and must be broadcastable to `losses` (i.e., all
     *    dimensions must be either `1`, or the same as the corresponding
     *    `losses` dimension).
     *
     * @doc {heading: 'Training', subheading: 'Losses', namespace: 'losses'}
     */
    function computeWeightedLoss_(losses, weights, reduction = Reduction.SUM_BY_NONZERO_WEIGHTS) {
        const $losses = convertToTensor(losses, 'losses', 'computeWeightedLoss');
        let $weights = null;
        if (weights != null) {
            $weights = convertToTensor(weights, 'weights', 'computeWeightedLoss');
        }
        const weightedLoss = ($weights == null) ? $losses : mul($losses, $weights);
        if (reduction === Reduction.NONE) {
            return weightedLoss;
        }
        if (reduction === Reduction.SUM) {
            return sum$2(weightedLoss);
        }
        if (reduction === Reduction.MEAN) {
            if ($weights == null) {
                return mean$1(weightedLoss);
            }
            else {
                const broadcastFactor = $losses.size / $weights.size;
                const result = div$1(sum$2(weightedLoss), sum$2($weights));
                return broadcastFactor > 1 ? div$1(result, scalar(broadcastFactor)) :
                    result;
            }
        }
        if (reduction === Reduction.SUM_BY_NONZERO_WEIGHTS) {
            if ($weights == null) {
                return div$1(sum$2(weightedLoss), scalar($losses.size));
            }
            else {
                const broadcastedWeights = mul($weights, ones$1($losses.shape));
                const numNonZeros = cast$3(sum$2(notEqual$3(broadcastedWeights, scalar(0))), 'float32');
                return div$1(sum$2(weightedLoss), numNonZeros);
            }
        }
        throw Error(`Unknown reduction: ${reduction}`);
    }
    const computeWeightedLoss$1 = op({ computeWeightedLoss_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the absolute difference loss between two tensors.
     *
     * @param labels The ground truth output tensor, same dimensions as
     *    'predictions'.
     * @param predictions The predicted outputs.
     * @param weights Tensor whose rank is either 0, or the same rank as
     *    `labels`, and must be broadcastable to `labels` (i.e., all dimensions
     *    must be either `1`, or the same as the corresponding `losses`
     *    dimension).
     * @param reduction Type of reduction to apply to loss. Should be of type
     *    `Reduction`
     *
     * @doc {heading: 'Training', subheading: 'Losses', namespace: 'losses'}
     */
    function absoluteDifference_(labels, predictions, weights, reduction = Reduction.SUM_BY_NONZERO_WEIGHTS) {
        const $labels = convertToTensor(labels, 'labels', 'absoluteDifference');
        const $predictions = convertToTensor(predictions, 'predictions', 'absoluteDifference');
        let $weights = null;
        if (weights != null) {
            $weights = convertToTensor(weights, 'weights', 'absoluteDifference');
        }
        assertShapesMatch($labels.shape, $predictions.shape, 'Error in absoluteDifference: ');
        const losses = abs$2(sub$2($labels, $predictions));
        return computeWeightedLoss$1(losses, $weights, reduction);
    }
    const absoluteDifference = op({ absoluteDifference_ });

    /**
     * Computes the cosine distance loss between two tensors.
     *
     * @param labels The ground truth output tensor, same dimensions as
     *    'predictions'.
     * @param predictions The predicted outputs.
     * @param axis The dimension along which the cosine distance is computed.
     * @param weights Tensor whose rank is either 0, or the same rank as
     *    `labels`, and must be broadcastable to `labels` (i.e., all dimensions
     *    must be either `1`, or the same as the corresponding `losses`
     *    dimension).
     * @param reduction Type of reduction to apply to loss. Should be of type
     *    `Reduction`
     *
     * @doc {heading: 'Training', subheading: 'Losses', namespace: 'losses'}
     */
    function cosineDistance_(labels, predictions, axis, weights, reduction = Reduction.SUM_BY_NONZERO_WEIGHTS) {
        const $labels = convertToTensor(labels, 'labels', 'cosineDistance');
        const $predictions = convertToTensor(predictions, 'predictions', 'cosineDistance');
        let $weights = null;
        if (weights != null) {
            $weights = convertToTensor(weights, 'weights', 'cosineDistance');
        }
        assertShapesMatch($labels.shape, $predictions.shape, 'Error in cosineDistance: ');
        const one = scalar(1);
        const losses = sub$2(one, sum$2(mul($labels, $predictions), axis, true));
        return computeWeightedLoss$1(losses, $weights, reduction);
    }
    const cosineDistance = op({ cosineDistance_ });

    /**
     * Computes the Hinge loss between two tensors.
     *
     * @param labels The ground truth output tensor, same dimensions as
     *    'predictions'.
     * @param predictions The predicted outputs.
     * @param weights Tensor whose rank is either 0, or the same rank as
     *    `labels`, and must be broadcastable to `labels` (i.e., all dimensions
     *    must be either `1`, or the same as the corresponding `losses`
     *    dimension).
     * @param reduction Type of reduction to apply to loss. Should be of type
     *    `Reduction`
     *
     * @doc {heading: 'Training', subheading: 'Losses', namespace: 'losses'}
     */
    function hingeLoss_(labels, predictions, weights, reduction = Reduction.SUM_BY_NONZERO_WEIGHTS) {
        let $labels = convertToTensor(labels, 'labels', 'hingeLoss');
        const $predictions = convertToTensor(predictions, 'predictions', 'hingeLoss');
        let $weights = null;
        if (weights != null) {
            $weights = convertToTensor(weights, 'weights', 'hingeLoss');
        }
        assertShapesMatch($labels.shape, $predictions.shape, 'Error in hingeLoss: ');
        const one = scalar(1);
        // Convert binary labels to (-1, 1)
        $labels = sub$2(mul(scalar(2), $labels), one);
        const losses = relu$2(sub$2(one, mul($labels, $predictions)));
        return computeWeightedLoss$1(losses, $weights, reduction);
    }
    const hingeLoss = op({ hingeLoss_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the huber loss between two tensors.
     *
     * @param labels The ground truth output tensor, same dimensions as
     *    'predictions'.
     * @param predictions The predicted outputs.
     * @param weights Tensor whose rank is either 0, or the same rank as
     *    `labels`, and must be broadcastable to `labels` (i.e., all dimensions
     *    must be either `1`, or the same as the corresponding `losses`
     *    dimension).
     * @param delta Point where huber loss changes from quadratic to linear.
     * @param reduction Type of reduction to apply to loss. Should be of type
     *    `Reduction`.
     *
     * @doc {heading: 'Training', subheading: 'Losses', namespace: 'losses'}
     */
    function huberLoss_(labels, predictions, weights, delta = 1.0, reduction = Reduction.SUM_BY_NONZERO_WEIGHTS) {
        const $labels = convertToTensor(labels, 'labels', 'huberLoss');
        const $predictions = convertToTensor(predictions, 'predictions', 'huberLoss');
        let $weights = null;
        if (weights != null) {
            $weights = convertToTensor(weights, 'weights', 'huberLoss');
        }
        assertShapesMatch($labels.shape, $predictions.shape, 'Error in huberLoss: ');
        const deltaScalar = scalar(delta);
        const error = abs$2(sub$2($predictions, $labels));
        const quadratic = minimum$3(error, deltaScalar);
        const linear = sub$2(error, quadratic);
        const losses = add$2(mul(scalar(0.5), square$2(quadratic)), mul(deltaScalar, linear));
        return computeWeightedLoss$1(losses, $weights, reduction);
    }
    const huberLoss = op({ huberLoss_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the log loss between two tensors.
     *
     * @param labels The ground truth output tensor, same dimensions as
     *    'predictions'.
     * @param predictions The predicted outputs.
     * @param weights Tensor whose rank is either 0, or the same rank as
     *    `labels`, and must be broadcastable to `labels` (i.e., all dimensions
     *    must be either `1`, or the same as the corresponding `losses`
     *    dimension).
     * @param epsilon A small increment to avoid taking log of zero
     * @param reduction Type of reduction to apply to loss. Should be of type
     *    `Reduction`
     *
     * @doc {heading: 'Training', subheading: 'Losses', namespace: 'losses'}
     */
    function logLoss_(labels, predictions, weights, epsilon = 1e-7, reduction = Reduction.SUM_BY_NONZERO_WEIGHTS) {
        const $labels = convertToTensor(labels, 'labels', 'logLoss');
        const $predictions = convertToTensor(predictions, 'predictions', 'logLoss');
        let $weights = null;
        if (weights != null) {
            $weights = convertToTensor(weights, 'weights', 'logLoss');
        }
        assertShapesMatch($labels.shape, $predictions.shape, 'Error in logLoss: ');
        const one = scalar(1);
        const epsilonScalar = scalar(epsilon);
        const l1 = neg$2(mul($labels, log$3(add$2($predictions, epsilonScalar))));
        const l2 = mul(sub$2(one, $labels), log$3(add$2(sub$2(one, $predictions), epsilonScalar)));
        const losses = sub$2(l1, l2);
        return computeWeightedLoss$1(losses, $weights, reduction);
    }
    const logLoss = op({ logLoss_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the mean squared error between two tensors.
     *
     * @param labels The ground truth output tensor, same dimensions as
     *    'predictions'.
     * @param predictions The predicted outputs.
     * @param weights Tensor whose rank is either 0, or the same rank as
     *    `labels`, and must be broadcastable to `labels` (i.e., all dimensions
     *    must be either `1`, or the same as the corresponding `losses`
     *    dimension).
     * @param reduction Type of reduction to apply to loss. Should be of type
     *    `Reduction`
     *
     * @doc {heading: 'Training', subheading: 'Losses', namespace: 'losses'}
     */
    function meanSquaredError_(labels, predictions, weights, reduction = Reduction.SUM_BY_NONZERO_WEIGHTS) {
        const $labels = convertToTensor(labels, 'labels', 'meanSquaredError');
        const $predictions = convertToTensor(predictions, 'predictions', 'meanSquaredError');
        let $weights = null;
        if (weights != null) {
            $weights = convertToTensor(weights, 'weights', 'meanSquaredError');
        }
        assertShapesMatch($labels.shape, $predictions.shape, 'Error in meanSquaredError: ');
        const losses = squaredDifference$2($labels, $predictions);
        return computeWeightedLoss$1(losses, $weights, reduction);
    }
    const meanSquaredError$2 = op({ meanSquaredError_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sigmoidCrossEntropyWithLogits_(labels, logits) {
        const $labels = convertToTensor(labels, 'labels', 'sigmoidCrossEntropyWithLogits');
        const $logits = convertToTensor(logits, 'logits', 'sigmoidCrossEntropyWithLogits');
        assertShapesMatch($labels.shape, $logits.shape, 'Error in sigmoidCrossEntropyWithLogits: ');
        /**
         * Implementation Details:
         *
         * For brevity, let `x = logits`, `z = labels`.  The logistic loss is
         *     z * -log(sigmoid(x)) + (1 - z) * -log(1 - sigmoid(x))
         *   = z * -log(1 / (1 + exp(-x))) + (1 - z) * -log(exp(-x) / (1 + exp(-x)))
         *   = z * log(1 + exp(-x)) + (1 - z) * (-log(exp(-x)) + log(1 + exp(-x)))
         *   = z * log(1 + exp(-x)) + (1 - z) * (x + log(1 + exp(-x))
         *   = (1 - z) * x + log(1 + exp(-x))
         *   = x - x * z + log(1 + exp(-x))
         *
         *   For x < 0, to avoid overflow in exp(-x), we reformulate the above
         *     x - x * z + log(1 + exp(-x))
         *   = log(exp(x)) - x * z + log(1 + exp(-x))
         *   = - x * z + log(1 + exp(x))
         *
         * Hence, to ensure stability and avoid overflow, the implementation uses
         * this equivalent formulation:
         *     max(x, 0) - x * z + log(1 + exp(-abs(x)))
         */
        const maxOutput = relu$2($logits);
        const outputXTarget = mul($logits, $labels);
        const sigmoidOutput = log1p$2(exp$2(neg$2(abs$2($logits))));
        return add$2(sub$2(maxOutput, outputXTarget), sigmoidOutput);
    }
    /**
     * Computes the sigmoid cross entropy loss between two tensors.
     *
     * If labelSmoothing is nonzero, smooth the labels towards 1/2:
     *
     *   newMulticlassLabels = multiclassLabels * (1 - labelSmoothing)
     *                         + 0.5 * labelSmoothing
     *
     * @param multiClassLabels The ground truth output tensor of shape
     * [batch_size, num_classes], same dimensions as 'predictions'.
     * @param logits The predicted outputs.
     * @param weights Tensor whose rank is either 0, or the same rank as
     *    `labels`, and must be broadcastable to `labels` (i.e., all dimensions
     *    must be either `1`, or the same as the corresponding `losses`
     *    dimension).
     * @param labelSmoothing If greater than 0, then smooth the labels.
     * @param reduction Type of reduction to apply to loss. Should be of type
     *    `Reduction`
     *
     * @doc { heading: 'Training', subheading: 'Losses', namespace: 'losses' }
     */
    function sigmoidCrossEntropy_(multiClassLabels, logits, weights, labelSmoothing = 0, reduction = Reduction.SUM_BY_NONZERO_WEIGHTS) {
        let $multiClassLabels = convertToTensor(multiClassLabels, 'multiClassLabels', 'sigmoidCrossEntropy');
        const $logits = convertToTensor(logits, 'logits', 'sigmoidCrossEntropy');
        let $weights = null;
        if (weights != null) {
            $weights = convertToTensor(weights, 'weights', 'sigmoidCrossEntropy');
        }
        assertShapesMatch($multiClassLabels.shape, $logits.shape, 'Error in sigmoidCrossEntropy: ');
        if (labelSmoothing > 0) {
            const labelSmoothingScalar = scalar(labelSmoothing);
            const one = scalar(1);
            const half = scalar(0.5);
            $multiClassLabels =
                add$2(mul($multiClassLabels, sub$2(one, labelSmoothingScalar)), mul(half, labelSmoothingScalar));
        }
        const losses = sigmoidCrossEntropyWithLogits_($multiClassLabels, $logits);
        return computeWeightedLoss$1(losses, $weights, reduction);
    }
    const sigmoidCrossEntropy = op({ sigmoidCrossEntropy_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes softmax cross entropy between logits and labels.
     *
     * Measures the probability error in discrete classification tasks in which
     * the classes are mutually exclusive (each entry is in exactly one class).
     * For example, each CIFAR-10 image is labeled with one and only one label: an
     * image can be a dog or a truck, but not both.
     *
     * `NOTE`: While the classes are mutually exclusive, their probabilities need
     * not be. All that is required is that each row of labels is a valid
     * probability distribution. If they are not, the computation of the gradient
     * will be incorrect.
     *
     * `WARNING`: This op expects unscaled logits, since it performs a softmax on
     * logits internally for efficiency. Do not call this op with the output of
     * softmax, as it will produce incorrect results.
     *
     * logits and labels must have the same shape, e.g. [batch_size, num_classes]
     * and the same dtype.
     * @param labels The labels array.
     * @param logits The logits array.
     * @param dim The dimension softmax would be performed on. Defaults to `-1`
     *     which indicates the last dimension.
     */
    function softmaxCrossEntropyWithLogits_(labels, logits, dim = -1) {
        if (dim === -1) {
            dim = logits.rank - 1;
        }
        if (dim !== logits.rank - 1) {
            throw Error(`Softmax cross entropy along a non-last dimension is not yet ` +
                `supported. Labels / logits was rank ${logits.rank} ` +
                `and dim was ${dim}`);
        }
        // Use a custom gradient for numerical stability.
        const customOp = customGrad((labels, logits, save) => {
            // Reference:
            //   1. http://cs231n.github.io/linear-classify/#softmax
            //   2. https://blog.feedly.com/tricks-of-the-trade-logsumexp/
            const keepDims = true;
            const lse = logSumExp(logits, [dim], keepDims);
            const logResult = sub$2(cast$3(logits, 'float32'), lse);
            save([labels, logResult]);
            const costVector = neg$2(mul(logResult, labels));
            const value = sum$2(costVector, [dim]);
            const gradFunc = (dy, saved) => {
                const [labels, logResult] = saved;
                const dyShape = expandShapeToKeepDim(dy.shape, [dim]);
                return [
                    mul(reshape$3(dy, dyShape), sub$2(cast$3(labels, 'float32'), exp$2(logResult))),
                    mul(reshape$3(dy, dyShape), sub$2(exp$2(logResult), cast$3(labels, 'float32'))),
                ];
            };
            return { value, gradFunc };
        });
        return customOp(labels, logits);
    }
    /**
     * Computes the softmax cross entropy loss between two tensors.
     *
     * If labelSmoothing is nonzero, smooth the labels towards 1/2:
     *
     *   newOnehotLabels = onehotLabels * (1 - labelSmoothing)
     *                         + labelSmoothing / numClasses
     *
     * @param onehotLabels One hot encoded labels
     *    [batch_size, num_classes], same dimensions as 'predictions'.
     * @param logits The predicted outputs.
     * @param weights Tensor whose rank is either 0, or 1, and must be
     *    broadcastable to `loss`  of shape [batch_size]
     * @param labelSmoothing If greater than 0, then smooth the labels.
     * @param reduction Type of reduction to apply to loss. Should be of type
     *    `Reduction`
     *
     * @doc { heading: 'Training', subheading: 'Losses', namespace: 'losses' }
     */
    function softmaxCrossEntropy_(onehotLabels, logits, weights, labelSmoothing = 0, reduction = Reduction.SUM_BY_NONZERO_WEIGHTS) {
        let $onehotLabels = convertToTensor(onehotLabels, 'onehotLabels', 'softmaxCrossEntropy');
        const $logits = convertToTensor(logits, 'logits', 'softmaxCrossEntropy');
        let $weights = null;
        if (weights != null) {
            $weights = convertToTensor(weights, 'weights', 'softmaxCrossEntropy');
        }
        assertShapesMatch($onehotLabels.shape, $logits.shape, 'Error in softmaxCrossEntropy: ');
        if (labelSmoothing > 0) {
            const labelSmoothingScalar = scalar(labelSmoothing);
            const one = scalar(1);
            const numClasses = scalar($onehotLabels.shape[1]);
            $onehotLabels =
                add$2(mul($onehotLabels, sub$2(one, labelSmoothingScalar)), div$1(labelSmoothingScalar, numClasses));
        }
        const losses = softmaxCrossEntropyWithLogits_($onehotLabels, $logits);
        return computeWeightedLoss$1(losses, $weights, reduction);
    }
    const softmaxCrossEntropy = op({ softmaxCrossEntropy_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * The input SparseTensor is represented via the map of inputs {`indices`,
     * `values`, `denseShape`}. The output SparseTensor has the same `denseShape`
     * but with indices `outputIndices` and values `outputValues`. This op inserts a
     * single entry for every row that doesn't have any values. The index is created
     * as `[row, 0, ..., 0]` and the inserted value is `defaultValue`.
     *
     * For example, suppose `spInput` has shape [5, 6] and non-empty values:
     * [0, 1]: a
     * [0, 3]: b
     * [2, 0]: c
     * [3, 1]: d
     *
     * Rows 1 and 4 are empty, so the output will be of shape [5, 6] with values:
     * [0, 1]: a
     * [0, 3]: b
     * [1, 0]: `defaultValue`
     * [2, 0]: c
     * [3, 1]: d
     * [4, 0]: `defaultValue`
     *
     * The output SparseTensor will be in row-major order and will have the same
     * shape as the input.
     *
     * This op also returns an indicator vector shaped [dense_shape[0]] such that
     * emptyRowIndicator[i] = True iff row i was an empty row.
     *
     * And a reverse index map vector shaped [indices.shape[0]] that is used during
     * backpropagation, reverseIndexMap[i] = outi s.t. indices[i, j] ==
     * outputIndices[outi, j] for all j
     *
     * ```js
     * const result = tf.sparse.sparseFillEmptyRows(
     *   [[0, 0], [1, 0], [1, 3], [1, 4], [3, 2], [3, 3]],
     *   [0, 10, 13, 14, 32, 33], [5, 6], -1);
     * console.log(result);
     * result['outputIndices'].print(); // [[0, 0], [1, 0], [1, 3], [1, 4],
     *                                  //  [2, 0], [3, 2], [3, 3], [4, 0]]
     * result['outputValues'].print(); // [0, 10, 13, 14,-1, 32, 33, -1]
     * result['emptyRowIndicator'].print(); // [false, false, true, false, true]
     * result['reverseIndexMap'].print(); // [0, 1, 2, 3, 5, 6]
     * ```
     * @param indices: 2-D. the indices of the sparse tensor.
     * @param values: 1-D. the values of the sparse tensor.
     * @param denseShape: 1-D. the shape of the sparse tensor.
     * @param defaultValue: 0-D. default value to insert into location [row, 0, ...,
     *     0] for rows missing from the input sparse tensor.
     * @return A map with the following properties:
     *     - outputIndices
     *     - outputValues: 1-D. the values of the filled sparse tensor.
     *     - emptyRowIndicator: 1-D. whether the dense row was missing in the input
     * sparse tensor.
     *     - reverseIndexMap: 1-D. a map from the input indices to the output
     * indices.
     * @doc {heading: 'Operations', subheading: 'Sparse'}
     */
    function sparseFillEmptyRows_(indices, values, denseShape, defaultValue) {
        const $indices = convertToTensor(indices, 'indices', 'sparseFillEmptyRows', 'int32');
        const $values = convertToTensor(values, 'values', 'sparseFillEmptyRows');
        const $denseShape = convertToTensor(denseShape, 'denseShape', 'sparseFillEmptyRows', 'int32');
        const $defaultValue = convertToTensor(defaultValue, 'defaultValue', 'sparseFillEmptyRows', $values.dtype);
        if ($indices.rank !== 2) {
            throw new Error(`Indices should be Tensor2D but received shape
        ${$indices.shape}`);
        }
        if ($values.rank !== 1) {
            throw new Error(`Values should be Tensor1D but received shape ${$values.shape}`);
        }
        if ($denseShape.rank !== 1) {
            throw new Error(`Dense shape should be Tensor1D but received shape ${$denseShape.shape}`);
        }
        if ($defaultValue.rank !== 0) {
            throw new Error(`Default value should be a scalar but received shape ${$defaultValue.shape}`);
        }
        const inputs = {
            indices: $indices,
            values: $values,
            denseShape: $denseShape,
            defaultValue: $defaultValue
        };
        const result = ENGINE.runKernel(SparseFillEmptyRows, inputs);
        return {
            outputIndices: result[0],
            outputValues: result[1],
            emptyRowIndicator: result[2],
            reverseIndexMap: result[3]
        };
    }
    const sparseFillEmptyRows$2 = op({ sparseFillEmptyRows_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * This operation has the same semantics as reshape on the represented dense
     * tensor. The `inputIndices` are recomputed based on the requested `newShape`.
     * If one component of `newShape` is the special value -1, the size of that
     * dimension is computed so that the total dense size remains constant. At most
     * one component of `newShape` can be -1. The number of dense elements implied
     * by `newShape` must be the same as the number of dense elements originally
     * implied by `inputShape`. Reshaping does not affect the order of values in the
     * SparseTensor. If the input tensor has rank R_in and N non-empty values, and
     * `newShape` has length R_out, then `inputIndices` has shape [N, R_in],
     * `inputShape` has length R_in, `outputIndices` has shape [N, R_out], and
     * `outputShape` has length R_out.
     *
     * ```js
     * const result = tf.sparse.sparseReshape(
     *   [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0], [1, 2, 3]],
     *   [2, 3, 6], [9, -1]);
     * console.log(result);
     * result['outputIndices'].print(); //[[0, 0], [0, 1], [1, 2], [4, 2], [8, 1]]
     * result['outputShape'].print(); // [9, 4]
     * ```
     * @param inputIndices: 2-D. N x R_in matrix with the indices of non-empty
     * values in a SparseTensor.
     * @param inputShape: 1-D. R_in Tensor1D with the input SparseTensor's dense
     * shape.
     * @param newShape: 1-D. R_out Tensor1D with the requested new dense shape.
     * @return A map with the following properties:
     *     - outputIndices: 2-D. N x R_out matrix with the updated indices of
     *       non-empty values in the output SparseTensor.
     *     - outputShape: 1-D. R_out vector with the full dense shape of the output
     *       SparseTensor. This is the same as newShape but with any -1 dimensions
     *        filled in.
     * @doc {heading: 'Operations', subheading: 'Sparse'}
     */
    function sparseReshape_(inputIndices, inputShape, newShape) {
        const $inputIndices = convertToTensor(inputIndices, 'inputIndices', 'sparseReshape', 'int32');
        const $inputShape = convertToTensor(inputShape, 'inputShape', 'sparseReshape', 'int32');
        const $newShape = convertToTensor(newShape, 'newShape', 'sparseReshape', 'int32');
        if ($inputIndices.rank !== 2) {
            throw new Error(`Input indices should be Tensor2D but received shape
        ${$inputIndices.shape}`);
        }
        if ($inputShape.rank !== 1) {
            throw new Error(`Input shape should be Tensor1D but received shape ${$inputShape.shape}`);
        }
        if ($newShape.rank !== 1) {
            throw new Error(`New shape should be Tensor1D but received shape ${$newShape.shape}`);
        }
        const inputs = {
            inputIndices: $inputIndices,
            inputShape: $inputShape,
            newShape: $newShape
        };
        const result = ENGINE.runKernel(SparseReshape, inputs);
        return { outputIndices: result[0], outputShape: result[1] };
    }
    const sparseReshape$2 = op({ sparseReshape_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the mean along sparse segments of a tensor.
     *
     * ```js
     * const c = tf.tensor2d([[1,2,3,4], [-1,-2,-3,-4], [6,7,8,9]]);
     * // Select two rows, one segment.
     * const result1 = tf.sparse.sparseSegmentMean(c,
     *                                           tf.tensor1d([0, 1], 'int32'),
     *                                           tf.tensor1d([0, 0], 'int32'));
     * result1.print(); // [[0, 0, 0, 0]]
     *
     * // Select two rows, two segments.
     * const result2 = tf.sparse.sparseSegmentMean(c,
     *                                             tf.tensor1d([0, 1], 'int32'),
     *                                             tf.tensor1d([0, 1], 'int32'));
     * result2.print(); // [[1, 2, 3, 4], [-1, -2, -3, -4]]
     *
     * // Select all rows, two segments.
     * const result3 = tf.sparse.sparseSegmentMean(c,
     *                                             tf.tensor1d([0, 1, 2], 'int32'),
     *                                             tf.tensor1d([0, 1, 1], 'int32'));
     * result3.print(); // [[1.0, 2.0, 3.0, 4.0], [2.5, 2.5, 2.5, 2.5]]
     * ```
     * @param data: A Tensor of at least one dimension with data that will be
     *     assembled in the output.
     * @param indices: A 1-D Tensor with indices into data. Has same rank as
     *     segmentIds.
     * @param segmentIds: A 1-D Tensor with indices into the output Tensor. Values
     *     should be sorted and can be repeated.
     * @return Has same shape as data, except for dimension 0 which has equal to
     *         the number of segments.
     *
     * @doc {heading: 'Operations', subheading: 'Sparse'}
     */
    function sparseSegmentMean_(data, indices, segmentIds) {
        const $data = convertToTensor(data, 'data', 'sparseSegmentMean');
        const $indices = convertToTensor(indices, 'indices', 'sparseSegmentMean', 'int32');
        const $segmentIds = convertToTensor(segmentIds, 'segmentIds', 'sparseSegmentMean', 'int32');
        if ($data.rank < 1) {
            throw new Error(`Data should be at least 1 dimensional but received scalar`);
        }
        if ($indices.rank !== 1) {
            throw new Error(`Indices should be Tensor1D but received shape
          ${$indices.shape}`);
        }
        if ($segmentIds.rank !== 1) {
            throw new Error(`Segment ids should be Tensor1D but received shape
          ${$segmentIds.shape}`);
        }
        const inputs = {
            data: $data,
            indices: $indices,
            segmentIds: $segmentIds
        };
        return ENGINE.runKernel(SparseSegmentMean, inputs);
    }
    const sparseSegmentMean$2 = op({ sparseSegmentMean_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the sum along sparse segments of a tensor.
     *
     * ```js
     * const c = tf.tensor2d([[1,2,3,4], [-1,-2,-3,-4], [5,6,7,8]]);
     * // Select two rows, one segment.
     * const result1 = tf.sparse.sparseSegmentSum(c,
     *                                           tf.tensor1d([0, 1], 'int32'),
     *                                           tf.tensor1d([0, 0], 'int32'));
     * result1.print(); // [[0, 0, 0, 0]]
     *
     * // Select two rows, two segment.
     * const result2 = tf.sparse.sparseSegmentSum(c,
     *                                           tf.tensor1d([0, 1], 'int32'),
     *                                           tf.tensor1d([0, 1], 'int32'));
     * result2.print(); // [[1, 2, 3, 4], [-1, -2, -3, -4]]
     *
     * // Select all rows, two segments.
     * const result3 = tf.sparse.sparseSegmentSum(c,
     *                                           tf.tensor1d([0, 1, 2], 'int32'),
     *                                           tf.tensor1d([0, 0, 1], 'int32'));
     * result3.print(); // [[0, 0, 0, 0], [5, 6, 7, 8]]
     * ```
     * @param data: A Tensor of at least one dimension with data that will be
     *     assembled in the output.
     * @param indices: A 1-D Tensor with indices into data. Has same rank as
     *     segmentIds.
     * @param segmentIds: A 1-D Tensor with indices into the output Tensor. Values
     *     should be sorted and can be repeated.
     * @return Has same shape as data, except for dimension 0 which has equal to
     *         the number of segments.
     *
     * @doc {heading: 'Operations', subheading: 'Sparse'}
     */
    function sparseSegmentSum_(data, indices, segmentIds) {
        const $data = convertToTensor(data, 'data', 'sparseSegmentSum');
        const $indices = convertToTensor(indices, 'indices', 'sparseSegmentSum', 'int32');
        const $segmentIds = convertToTensor(segmentIds, 'segmentIds', 'sparseSegmentSum', 'int32');
        if ($data.rank < 1) {
            throw new Error(`Data should be at least 1 dimensional but received scalar`);
        }
        if ($indices.rank !== 1) {
            throw new Error(`Indices should be Tensor1D but received shape
         ${$indices.shape}`);
        }
        if ($segmentIds.rank !== 1) {
            throw new Error(`Segment ids should be Tensor1D but received shape
         ${$segmentIds.shape}`);
        }
        const inputs = {
            data: $data,
            indices: $indices,
            segmentIds: $segmentIds
        };
        return ENGINE.runKernel(SparseSegmentSum, inputs);
    }
    const sparseSegmentSum$2 = op({ sparseSegmentSum_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Creates ngrams from ragged string data.
     *
     * This op accepts a ragged tensor with 1 ragged dimension containing only
     * strings and outputs a ragged tensor with 1 ragged dimension containing ngrams
     * of that string, joined along the innermost axis.
     *
     * ```js
     * const result = tf.string.stringNGrams(
     *   ['a', 'b', 'c', 'd'], tf.tensor1d([0, 2, 4], 'int32'),
     *   '|', [1, 2], 'LP', 'RP', -1, false);
     * result['nGrams'].print(); // ['a', 'b', 'LP|a', 'a|b', 'b|RP',
     *                           //  'c', 'd', 'LP|c', 'c|d', 'd|RP']
     * result['nGramsSplits'].print(); // [0, 5, 10]
     * ```
     * @param data: The values tensor of the ragged string tensor to make ngrams out
     *     of. Must be a 1D string tensor.
     * @param dataSplits: The splits tensor of the ragged string tensor to make
     *     ngrams out of.
     * @param separator: The string to append between elements of the token. Use ""
     *     for no separator.
     * @param nGramWidths: The sizes of the ngrams to create.
     * @param leftPad: The string to use to pad the left side of the ngram sequence.
     *     Only used if pad_width !== 0.
     * @param rightPad: The string to use to pad the right side of the ngram
     *     sequence. Only used if pad_width !== 0.
     * @param padWidth: The number of padding elements to add to each side of each
     *     sequence. Note that padding will never be greater than `nGramWidths`-1
     *     regardless of this value. If `padWidth`=-1 , then add max(`nGramWidths)-1
     *     elements.
     * @param preserveShortSequences: If true, then ensure that at least one ngram
     *     is generated for each input sequence. In particular, if an input sequence
     *     is shorter than min(ngramWidth) + 2*padWidth, then generate a single
     *     ngram containing the entire sequence. If false, then no ngrams are
     *     generated for these short input sequences.
     * @return A map with the following properties:
     *     - nGrams: The values tensor of the output ngrams ragged tensor.
     *     - nGramsSplits: The splits tensor of the output ngrams ragged tensor.
     *
     * @doc {heading: 'Operations', subheading: 'String'}
     */
    function stringNGrams_(data, dataSplits, separator, nGramWidths, leftPad, rightPad, padWidth, preserveShortSequences) {
        const $data = convertToTensor(data, 'data', 'stringNGrams', 'string');
        if ($data.dtype !== 'string') {
            throw new Error('Data must be of datatype string');
        }
        if ($data.shape.length !== 1) {
            throw new Error(`Data must be a vector, saw: ${$data.shape}`);
        }
        const $dataSplits = convertToTensor(dataSplits, 'dataSplits', 'stringNGrams');
        if ($dataSplits.dtype !== 'int32') {
            throw new Error('Data splits must be of datatype int32');
        }
        const attrs = {
            separator,
            nGramWidths,
            leftPad,
            rightPad,
            padWidth,
            preserveShortSequences
        };
        const inputs = { data: $data, dataSplits: $dataSplits };
        const result = ENGINE.runKernel(StringNGrams, inputs, attrs);
        return { nGrams: result[0], nGramsSplits: result[1] };
    }
    const stringNGrams$2 = op({ stringNGrams_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Split elements of `input` based on `delimiter` into a SparseTensor .
     *
     * Let N be the size of source (typically N will be the batch size). Split each
     * element of `input` based on `delimiter` and return a SparseTensor containing
     * the splitted tokens. Empty tokens are ignored if `skipEmpty` is set to True.
     *
     * `delimiter` can be empty, or a string of split characters. If `delimiter` is
     * an empty string, each element of `input` is split into individual
     * character strings. Otherwise every character of `delimiter` is a potential
     * split point.
     *
     * ```js
     * const result = tf.string.stringSplit(['hello world',  'a b c'], ' ');
     * result['indices'].print(); // [[0, 0], [0, 1], [1, 0], [1, 1], [1, 2]]
     * result['values'].print(); // ['hello', 'world', 'a', 'b', 'c']
     * result['shape'].print(); // [2, 3]
     * ```
     * @param input: 1-D. Strings to split.
     * @param delimiter: 0-D. Delimiter characters, or empty string.
     * @param skipEmpty: Optional. If true, skip the empty strings from the result.
     *     Defaults to true.
     * @return A map with the following properties:
     *     - indices: A dense matrix of int32 representing the indices of the sparse
     *       tensor.
     *     - values: A vector of strings corresponding to the splited values.
     *     - shape: a length-2 vector of int32 representing the shape of the sparse
     * tensor, where the first value is N and the second value is the maximum number
     * of tokens in a single input entry.
     *
     * @doc {heading: 'Operations', subheading: 'String'}
     */
    function stringSplit_(input, delimiter, skipEmpty = true) {
        const $input = convertToTensor(input, 'input', 'stringSplit', 'string');
        const $delimiter = convertToTensor(delimiter, 'delimiter', 'stringSplit', 'string');
        if ($input.rank !== 1) {
            throw new Error(`Input should be Tensor1D but received shape ${$input.shape}`);
        }
        if ($delimiter.rank !== 0) {
            throw new Error(`Delimiter should be a scalar but received shape ${$delimiter.shape}`);
        }
        const attrs = { skipEmpty };
        const inputs = { input: $input, delimiter: $delimiter };
        const result = ENGINE.runKernel(StringSplit, inputs, attrs);
        return { indices: result[0], values: result[1], shape: result[2] };
    }
    const stringSplit$2 = op({ stringSplit_ });

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts each string in the input Tensor to its hash mod by a number of
     * buckets.
     *
     * The hash function is deterministic on the content of the string within the
     * process and will never change. However, it is not suitable for cryptography.
     * This function may be used when CPU time is scarce and inputs are trusted or
     * unimportant. There is a risk of adversaries constructing inputs that all hash
     * to the same bucket.
     *
     * ```js
     * const result = tf.string.stringToHashBucketFast(
     *   ['Hello', 'TensorFlow', '2.x'], 3);
     * result.print(); // [0, 2, 2]
     * ```
     * @param input: The strings to assign a hash bucket.
     * @param numBuckets: The number of buckets.
     * @return A Tensor of the same shape as the input tensor.
     *
     * @doc {heading: 'Operations', subheading: 'String'}
     */
    function stringToHashBucketFast_(input, numBuckets) {
        const $input = convertToTensor(input, 'input', 'stringToHashBucketFast', 'string');
        const attrs = { numBuckets };
        if (numBuckets <= 0) {
            throw new Error(`Number of buckets must be at least 1`);
        }
        const inputs = { input: $input };
        return ENGINE.runKernel(StringToHashBucketFast, inputs, attrs);
    }
    const stringToHashBucketFast$2 = op({ stringToHashBucketFast_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const spectral$1 = {
        fft: fft$2,
        ifft: ifft$2,
        rfft,
        irfft
    };
    const signal = {
        hammingWindow,
        hannWindow,
        frame,
        stft,
    };
    const image$1 = {
        flipLeftRight,
        grayscaleToRGB,
        resizeNearestNeighbor: resizeNearestNeighbor$2,
        resizeBilinear: resizeBilinear$2,
        rotateWithOffset,
        cropAndResize: cropAndResize$2,
        nonMaxSuppression,
        nonMaxSuppressionAsync,
        nonMaxSuppressionWithScore,
        nonMaxSuppressionWithScoreAsync,
        nonMaxSuppressionPadded,
        nonMaxSuppressionPaddedAsync,
        threshold: threshold$1,
        transform: transform$2
    };
    const linalg = {
        bandPart,
        gramSchmidt,
        qr
    };
    const losses = {
        absoluteDifference,
        computeWeightedLoss: computeWeightedLoss$1,
        cosineDistance,
        hingeLoss,
        huberLoss,
        logLoss,
        meanSquaredError: meanSquaredError$2,
        sigmoidCrossEntropy,
        softmaxCrossEntropy
    };
    const sparse$1 = {
        sparseFillEmptyRows: sparseFillEmptyRows$2,
        sparseReshape: sparseReshape$2,
        sparseSegmentMean: sparseSegmentMean$2,
        sparseSegmentSum: sparseSegmentSum$2
    };
    // tslint:disable-next-line:variable-name
    const string$1 = {
        stringNGrams: stringNGrams$2,
        stringSplit: stringSplit$2,
        stringToHashBucketFast: stringToHashBucketFast$2
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /** @doc {heading: 'Training', subheading: 'Classes', namespace: 'train'} */
    class Optimizer extends Serializable {
        /**
         * Executes `f()` and minimizes the scalar output of `f()` by computing
         * gradients of y with respect to the list of trainable variables provided by
         * `varList`. If no list is provided, it defaults to all trainable variables.
         *
         * @param f The function to execute and whose output to minimize.
         * @param returnCost Whether to return the scalar cost value produced by
         * executing `f()`.
         * @param varList An optional list of variables to update. If specified, only
         * the trainable variables in varList will be updated by minimize. Defaults to
         * all trainable variables.
         *
         * @doc {heading: 'Training', subheading: 'Optimizers'}
         */
        minimize(f, returnCost = false, varList) {
            const { value, grads } = this.computeGradients(f, varList);
            if (varList != null) {
                const gradArray = varList.map(v => ({ name: v.name, tensor: grads[v.name] }));
                this.applyGradients(gradArray);
            }
            else {
                this.applyGradients(grads);
            }
            // Dispose gradients.
            dispose(grads);
            if (returnCost) {
                return value;
            }
            else {
                value.dispose();
                return null;
            }
        }
        /**
         * The number of iterations that this optimizer instance has been invoked for.
         */
        get iterations() {
            if (this.iterations_ == null) {
                this.iterations_ = 0;
            }
            return this.iterations_;
        }
        incrementIterations() {
            this.iterations_ = this.iterations + 1;
        }
        /**
         * Executes f() and computes the gradient of the scalar output of f() with
         * respect to the list of trainable variables provided by `varList`. If no
         * list is provided, it defaults to all trainable variables.
         *
         * @param f The function to execute and whose output to use for computing
         * gradients with respect to variables.
         * @param varList An optional list of variables to compute gradients with
         * respect to. If specified, only the trainable variables in varList will have
         * gradients computed with respect to. Defaults to all trainable variables.
         *
         * @doc {heading: 'Training', subheading: 'Optimizers'}
         */
        computeGradients(f, varList) {
            return variableGrads(f, varList);
        }
        /**
         * Dispose the variables (if any) owned by this optimizer instance.
         */
        dispose() {
            if (this.iterations_ != null) {
                dispose(this.iterations_);
            }
        }
        async saveIterations() {
            if (this.iterations_ == null) {
                this.iterations_ = 0;
            }
            return {
                name: 'iter',
                // TODO(cais): Use 'int64' type when available.
                tensor: scalar(this.iterations_, 'int32')
            };
        }
        async getWeights() {
            throw new Error('getWeights() is not implemented for this optimizer yet.');
        }
        async setWeights(weightValues) {
            throw new Error(`setWeights() is not implemented for this optimizer class ` +
                `${this.getClassName()}`);
        }
        /**
         * Extract the first element of the weight values and set it
         * as the iterations counter variable of this instance of optimizer.
         *
         * @param weightValues
         * @returns Weight values with the first element consumed and excluded.
         */
        async extractIterations(weightValues) {
            this.iterations_ = (await weightValues[0].tensor.data())[0];
            return weightValues.slice(1);
        }
    }
    Object.defineProperty(Optimizer, Symbol.hasInstance, {
        value: (instance) => {
            return instance.minimize != null && instance.computeGradients != null &&
                instance.applyGradients != null;
        }
    });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /** @doclink Optimizer */
    class AdadeltaOptimizer extends Optimizer {
        constructor(learningRate, rho, epsilon = null) {
            super();
            this.learningRate = learningRate;
            this.rho = rho;
            this.epsilon = epsilon;
            this.accumulatedGrads = [];
            this.accumulatedUpdates = [];
            if (epsilon == null) {
                this.epsilon = ENGINE.backend.epsilon();
            }
        }
        applyGradients(variableGradients) {
            const variableNames = Array.isArray(variableGradients) ?
                variableGradients.map(item => item.name) :
                Object.keys(variableGradients);
            variableNames.forEach((name, i) => {
                const value = ENGINE.registeredVariables[name];
                const trainable = false;
                if (this.accumulatedGrads[i] == null) {
                    this.accumulatedGrads[i] = {
                        originalName: `${name}/accum_grad`,
                        variable: tidy(() => zerosLike$2(value).variable(trainable))
                    };
                }
                if (this.accumulatedUpdates[i] == null) {
                    this.accumulatedUpdates[i] = {
                        originalName: `${name}/accum_var`,
                        variable: tidy(() => zerosLike$2(value).variable(trainable))
                    };
                }
                const gradient = Array.isArray(variableGradients) ?
                    variableGradients[i].tensor :
                    variableGradients[name];
                if (gradient == null) {
                    return;
                }
                const accumulatedGrad = this.accumulatedGrads[i].variable;
                const accumulatedUpdate = this.accumulatedUpdates[i].variable;
                tidy(() => {
                    const newAccumulatedGrad = add$2(mul(accumulatedGrad, this.rho), mul(square$2(gradient), 1 - this.rho));
                    const updates = mul(div$1(sqrt$2(add$2(accumulatedUpdate, this.epsilon)), sqrt$2(add$2(accumulatedGrad, this.epsilon))), gradient);
                    const newAccumulatedUpdate = add$2(mul(accumulatedUpdate, this.rho), mul(square$2(updates), 1 - this.rho));
                    accumulatedGrad.assign(newAccumulatedGrad);
                    accumulatedUpdate.assign(newAccumulatedUpdate);
                    const newValue = add$2(mul(updates, -this.learningRate), value);
                    value.assign(newValue);
                });
            });
            this.incrementIterations();
        }
        dispose() {
            if (this.accumulatedUpdates != null) {
                dispose(this.accumulatedGrads.map(v => v.variable));
                dispose(this.accumulatedUpdates.map(v => v.variable));
            }
        }
        async getWeights() {
            // Order matters for Python compatibility.
            const variables = [...this.accumulatedGrads, ...this.accumulatedUpdates];
            return [await this.saveIterations()].concat(variables.map(v => ({ name: v.originalName, tensor: v.variable })));
        }
        async setWeights(weightValues) {
            weightValues = await this.extractIterations(weightValues);
            const variableCount = weightValues.length / 2;
            const trainable = false;
            this.accumulatedGrads =
                weightValues.slice(0, variableCount).map(v => ({
                    originalName: v.name,
                    variable: v.tensor.variable(trainable)
                }));
            this.accumulatedUpdates =
                weightValues.slice(variableCount, variableCount * 2)
                    .map(v => ({
                    originalName: v.name,
                    variable: v.tensor.variable(trainable)
                }));
        }
        getConfig() {
            return {
                'learningRate': this.learningRate,
                'rho': this.rho,
                'epsilon': this.epsilon
            };
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config['learningRate'], config['rho'], config['epsilon']);
        }
    }
    /** @nocollapse */
    AdadeltaOptimizer.className = 'Adadelta'; // Name matters for Python compatibility.
    registerClass(AdadeltaOptimizer);

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /** @doclink Optimizer */
    class AdagradOptimizer extends Optimizer {
        constructor(learningRate, initialAccumulatorValue = 0.1) {
            super();
            this.learningRate = learningRate;
            this.initialAccumulatorValue = initialAccumulatorValue;
            this.accumulatedGrads = [];
        }
        applyGradients(variableGradients) {
            const variableNames = Array.isArray(variableGradients) ?
                variableGradients.map(item => item.name) :
                Object.keys(variableGradients);
            variableNames.forEach((name, i) => {
                const value = ENGINE.registeredVariables[name];
                if (this.accumulatedGrads[i] == null) {
                    const trainable = false;
                    this.accumulatedGrads[i] = {
                        originalName: `${name}/accumulator`,
                        variable: tidy(() => fill$2(value.shape, this.initialAccumulatorValue)
                            .variable(trainable))
                    };
                }
                const gradient = Array.isArray(variableGradients) ?
                    variableGradients[i].tensor :
                    variableGradients[name];
                if (gradient == null) {
                    return;
                }
                const accumulatedGrad = this.accumulatedGrads[i].variable;
                tidy(() => {
                    const newAccumulatedGrad = add$2(accumulatedGrad, square$2(gradient));
                    accumulatedGrad.assign(newAccumulatedGrad);
                    const newValue = add$2(mul(div$1(gradient, sqrt$2(add$2(newAccumulatedGrad, ENGINE.backend.epsilon()))), -this.learningRate), value);
                    value.assign(newValue);
                });
            });
            this.incrementIterations();
        }
        dispose() {
            if (this.accumulatedGrads != null) {
                dispose(this.accumulatedGrads.map(v => v.variable));
            }
        }
        async getWeights() {
            // Order matters for Python compatibility.
            return [await this.saveIterations()].concat(this.accumulatedGrads.map(v => ({ name: v.originalName, tensor: v.variable })));
        }
        async setWeights(weightValues) {
            weightValues = await this.extractIterations(weightValues);
            const trainable = false;
            this.accumulatedGrads = weightValues.map(v => ({ originalName: v.name, variable: v.tensor.variable(trainable) }));
        }
        getConfig() {
            return {
                'learningRate': this.learningRate,
                'initialAccumulatorValue': this.initialAccumulatorValue,
            };
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config['learningRate'], config['initialAccumulatorValue']);
        }
    }
    /** @nocollapse */
    AdagradOptimizer.className = 'Adagrad'; // Note: Name matters for Python compatibility.
    registerClass(AdagradOptimizer);

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class AdamOptimizer extends Optimizer {
        constructor(learningRate, beta1, beta2, epsilon = null) {
            super();
            this.learningRate = learningRate;
            this.beta1 = beta1;
            this.beta2 = beta2;
            this.epsilon = epsilon;
            this.accumulatedFirstMoment = [];
            this.accumulatedSecondMoment = [];
            tidy(() => {
                // accB* will be updated by batch.
                this.accBeta1 = scalar(beta1).variable();
                this.accBeta2 = scalar(beta2).variable();
            });
            if (epsilon == null) {
                this.epsilon = ENGINE.backend.epsilon();
            }
        }
        applyGradients(variableGradients) {
            const varNames = Array.isArray(variableGradients) ?
                variableGradients.map(v => v.name) :
                Object.keys(variableGradients);
            tidy(() => {
                const oneMinusAccBeta1 = sub$2(1, this.accBeta1);
                const oneMinusAccBeta2 = sub$2(1, this.accBeta2);
                varNames.forEach((name, i) => {
                    const value = ENGINE.registeredVariables[name];
                    const trainable = false;
                    if (this.accumulatedFirstMoment[i] == null) {
                        this.accumulatedFirstMoment[i] = {
                            originalName: `${name}/m`,
                            variable: tidy(() => zerosLike$2(value).variable(trainable))
                        };
                    }
                    if (this.accumulatedSecondMoment[i] == null) {
                        this.accumulatedSecondMoment[i] = {
                            originalName: `${name}/v`,
                            variable: tidy(() => zerosLike$2(value).variable(trainable))
                        };
                    }
                    const gradient = Array.isArray(variableGradients) ?
                        variableGradients[i].tensor :
                        variableGradients[name];
                    if (gradient == null) {
                        return;
                    }
                    const firstMoment = this.accumulatedFirstMoment[i].variable;
                    const secondMoment = this.accumulatedSecondMoment[i].variable;
                    const newFirstMoment = add$2(mul(firstMoment, this.beta1), mul(gradient, 1 - this.beta1));
                    const newSecondMoment = add$2(mul(secondMoment, this.beta2), mul(square$2(gradient), 1 - this.beta2));
                    const biasCorrectedFirstMoment = div$1(newFirstMoment, oneMinusAccBeta1);
                    const biasCorrectedSecondMoment = div$1(newSecondMoment, oneMinusAccBeta2);
                    firstMoment.assign(newFirstMoment);
                    secondMoment.assign(newSecondMoment);
                    const newValue = add$2(mul(div$1(biasCorrectedFirstMoment, add$2(sqrt$2(biasCorrectedSecondMoment), this.epsilon)), -this.learningRate), value);
                    value.assign(newValue);
                });
                this.accBeta1.assign(mul(this.accBeta1, this.beta1));
                this.accBeta2.assign(mul(this.accBeta2, this.beta2));
            });
            this.incrementIterations();
        }
        dispose() {
            this.accBeta1.dispose();
            this.accBeta2.dispose();
            if (this.accumulatedFirstMoment != null) {
                dispose(this.accumulatedFirstMoment.map(v => v.variable));
            }
            if (this.accumulatedSecondMoment != null) {
                dispose(this.accumulatedSecondMoment.map(v => v.variable));
            }
        }
        async getWeights() {
            // Order matters for Python compatibility.
            const variables = [...this.accumulatedFirstMoment, ...this.accumulatedSecondMoment];
            return [await this.saveIterations()].concat(variables.map(v => ({ name: v.originalName, tensor: v.variable })));
        }
        async setWeights(weightValues) {
            weightValues = await this.extractIterations(weightValues);
            tidy(() => {
                this.accBeta1.assign(pow$2(this.beta1, this.iterations_ + 1));
                this.accBeta2.assign(pow$2(this.beta2, this.iterations_ + 1));
            });
            const variableCount = weightValues.length / 2;
            const trainable = false;
            this.accumulatedFirstMoment =
                weightValues.slice(0, variableCount).map(v => ({
                    originalName: v.name,
                    variable: v.tensor.variable(trainable)
                }));
            this.accumulatedSecondMoment =
                weightValues.slice(variableCount, variableCount * 2)
                    .map(v => ({
                    originalName: v.name,
                    variable: v.tensor.variable(trainable)
                }));
        }
        getConfig() {
            return {
                'learningRate': this.learningRate,
                'beta1': this.beta1,
                'beta2': this.beta2,
                'epsilon': this.epsilon,
            };
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config['learningRate'], config['beta1'], config['beta2'], config['epsilon']);
        }
    }
    /** @nocollapse */
    AdamOptimizer.className = 'Adam'; // Note: Name matters for Python compatibility.
    registerClass(AdamOptimizer);

    /**
    * @license
    * Copyright 2018 Google LLC. All Rights Reserved.
    * Licensed under the Apache License, Version 2.0 (the "License");
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    * http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
    * =============================================================================
    */
    class AdamaxOptimizer extends Optimizer {
        constructor(learningRate, beta1, beta2, epsilon = null, decay = 0.0) {
            super();
            this.learningRate = learningRate;
            this.beta1 = beta1;
            this.beta2 = beta2;
            this.epsilon = epsilon;
            this.decay = decay;
            this.accumulatedFirstMoment = [];
            this.accumulatedWeightedInfNorm = [];
            tidy(() => {
                this.iteration = scalar(0).variable();
                this.accBeta1 = scalar(beta1).variable();
            });
            if (epsilon == null) {
                this.epsilon = ENGINE.backend.epsilon();
            }
        }
        applyGradients(variableGradients) {
            const variableNames = Array.isArray(variableGradients) ?
                variableGradients.map(item => item.name) :
                Object.keys(variableGradients);
            tidy(() => {
                const oneMinusAccBeta1 = sub$2(1, this.accBeta1);
                const lr = div$1(-this.learningRate, add$2(mul(this.iteration, this.decay), 1));
                variableNames.forEach((name, i) => {
                    const value = ENGINE.registeredVariables[name];
                    const trainable = false;
                    if (this.accumulatedFirstMoment[i] == null) {
                        this.accumulatedFirstMoment[i] = {
                            originalName: `${name}/m`,
                            variable: zerosLike$2(value).variable(trainable)
                        };
                    }
                    if (this.accumulatedWeightedInfNorm[i] == null) {
                        this.accumulatedWeightedInfNorm[i] = {
                            originalName: `${name}/v`,
                            variable: zerosLike$2(value).variable(trainable)
                        };
                    }
                    const gradient = Array.isArray(variableGradients) ?
                        variableGradients[i].tensor :
                        variableGradients[name];
                    if (gradient == null) {
                        return;
                    }
                    const firstMoment = this.accumulatedFirstMoment[i].variable;
                    const weightedInfNorm = this.accumulatedWeightedInfNorm[i].variable;
                    const newFirstMoment = add$2(mul(firstMoment, this.beta1), mul(gradient, 1 - this.beta1));
                    const ut0 = mul(weightedInfNorm, this.beta2);
                    const ut1 = abs$2(gradient);
                    const newWeightedInfNorm = maximum$3(ut0, ut1);
                    firstMoment.assign(newFirstMoment);
                    weightedInfNorm.assign(newWeightedInfNorm);
                    const newValue = add$2(mul(div$1(lr, oneMinusAccBeta1), div$1(newFirstMoment, add$2(newWeightedInfNorm, this.epsilon))), value);
                    value.assign(newValue);
                });
                this.iteration.assign(add$2(this.iteration, 1));
                this.accBeta1.assign(mul(this.accBeta1, this.beta1));
            });
            this.incrementIterations();
        }
        dispose() {
            this.accBeta1.dispose();
            this.iteration.dispose();
            if (this.accumulatedFirstMoment != null) {
                dispose(this.accumulatedFirstMoment.map(v => v.variable));
            }
            if (this.accumulatedWeightedInfNorm != null) {
                dispose(this.accumulatedWeightedInfNorm.map(v => v.variable));
            }
        }
        async getWeights() {
            throw new Error('getWeights() is not implemented for Adamax yet.');
        }
        async setWeights(weightValues) {
            throw new Error('setWeights() is not implemented for Adamax yet.');
        }
        getConfig() {
            return {
                'learningRate': this.learningRate,
                'beta1': this.beta1,
                'beta2': this.beta2,
                'epsilon': this.epsilon,
                'decay': this.decay
            };
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config['learningRate'], config['beta1'], config['beta2'], config['epsilon'], config['decay']);
        }
    }
    /** @nocollapse */
    AdamaxOptimizer.className = 'Adamax'; // Note: Name matters for Python compatbility.
    registerClass(AdamaxOptimizer);

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /** @doclink Optimizer */
    class SGDOptimizer extends Optimizer {
        constructor(learningRate) {
            super();
            this.learningRate = learningRate;
            this.setLearningRate(learningRate);
        }
        applyGradients(variableGradients) {
            const varNames = Array.isArray(variableGradients) ?
                variableGradients.map(v => v.name) :
                Object.keys(variableGradients);
            varNames.forEach((name, i) => {
                const gradient = Array.isArray(variableGradients) ?
                    variableGradients[i].tensor :
                    variableGradients[name];
                if (gradient == null) {
                    return;
                }
                const value = ENGINE.registeredVariables[name];
                tidy(() => {
                    const newValue = add$2(mul(this.c, gradient), value);
                    value.assign(newValue);
                });
            });
            this.incrementIterations();
        }
        /**
         * Sets the learning rate of the optimizer.
         */
        setLearningRate(learningRate) {
            this.learningRate = learningRate;
            if (this.c != null) {
                this.c.dispose();
            }
            this.c = keep(scalar(-learningRate));
        }
        dispose() {
            this.c.dispose();
        }
        async getWeights() {
            return [await this.saveIterations()];
        }
        async setWeights(weightValues) {
            weightValues = await this.extractIterations(weightValues);
            if (weightValues.length !== 0) {
                throw new Error('SGD optimizer does not have settable weights.');
            }
        }
        getConfig() {
            return { 'learningRate': this.learningRate };
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config['learningRate']);
        }
    }
    /** @nocollapse */
    SGDOptimizer.className = 'SGD'; // Note: Name matters for Python compatibility.
    registerClass(SGDOptimizer);

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /** @doclink Optimizer */
    class MomentumOptimizer extends SGDOptimizer {
        constructor(learningRate, momentum, useNesterov = false) {
            super(learningRate);
            this.learningRate = learningRate;
            this.momentum = momentum;
            this.useNesterov = useNesterov;
            this.accumulations = [];
            this.m = scalar(this.momentum);
        }
        applyGradients(variableGradients) {
            const variableNames = Array.isArray(variableGradients) ?
                variableGradients.map(item => item.name) :
                Object.keys(variableGradients);
            variableNames.forEach((name, i) => {
                const value = ENGINE.registeredVariables[name];
                if (this.accumulations[i] == null) {
                    const trainable = false;
                    this.accumulations[i] = {
                        originalName: `${name}/momentum`,
                        variable: tidy(() => zerosLike$2(value).variable(trainable))
                    };
                }
                const accumulation = this.accumulations[i].variable;
                const gradient = Array.isArray(variableGradients) ?
                    variableGradients[i].tensor :
                    variableGradients[name];
                if (gradient == null) {
                    return;
                }
                tidy(() => {
                    let newValue;
                    const newAccumulation = add$2(mul(this.m, accumulation), gradient);
                    if (this.useNesterov) {
                        newValue = add$2(mul(this.c, add$2(gradient, mul(newAccumulation, this.m))), value);
                    }
                    else {
                        newValue = add$2(mul(this.c, newAccumulation), value);
                    }
                    accumulation.assign(newAccumulation);
                    value.assign(newValue);
                });
            });
            this.incrementIterations();
        }
        dispose() {
            this.m.dispose();
            if (this.accumulations != null) {
                dispose(this.accumulations.map(v => v.variable));
            }
        }
        /**
         * Sets the momentum of the optimizer.
         *
         * @param momentum
         */
        setMomentum(momentum) {
            this.momentum = momentum;
        }
        async getWeights() {
            // Order matters for Python compatibility.
            return [await this.saveIterations()].concat(this.accumulations.map(v => ({ name: v.originalName, tensor: v.variable })));
        }
        async setWeights(weightValues) {
            weightValues = await this.extractIterations(weightValues);
            const trainable = false;
            this.accumulations = weightValues.map(v => ({ originalName: v.name, variable: v.tensor.variable(trainable) }));
        }
        getConfig() {
            return {
                'learningRate': this.learningRate,
                'momentum': this.momentum,
                'useNesterov': this.useNesterov
            };
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config['learningRate'], config['momentum'], config['useNesterov']);
        }
    }
    /** @nocollapse */
    MomentumOptimizer.className = 'Momentum'; // Name matters for Python compatibility.
    registerClass(MomentumOptimizer);

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /** @doclink Optimizer */
    class RMSPropOptimizer extends Optimizer {
        constructor(learningRate, decay = 0.9, momentum = 0.0, epsilon = null, centered = false) {
            super();
            this.learningRate = learningRate;
            this.decay = decay;
            this.momentum = momentum;
            this.epsilon = epsilon;
            this.accumulatedMeanSquares = [];
            this.accumulatedMoments = [];
            this.accumulatedMeanGrads = [];
            this.centered = centered;
            if (epsilon == null) {
                this.epsilon = ENGINE.backend.epsilon();
            }
            if (learningRate == null) {
                throw new Error(`learningRate for RMSPropOptimizer must be defined.`);
            }
        }
        applyGradients(variableGradients) {
            const variableNames = Array.isArray(variableGradients) ?
                variableGradients.map(item => item.name) :
                Object.keys(variableGradients);
            variableNames.forEach((name, i) => {
                const value = ENGINE.registeredVariables[name];
                const trainable = false;
                if (this.accumulatedMeanSquares[i] == null) {
                    this.accumulatedMeanSquares[i] = {
                        originalName: `${name}/rms`,
                        variable: tidy(() => zerosLike$2(value).variable(trainable))
                    };
                }
                if (this.accumulatedMoments[i] == null) {
                    this.accumulatedMoments[i] = {
                        originalName: `${name}/momentum`,
                        variable: tidy(() => zerosLike$2(value).variable(trainable))
                    };
                }
                if (this.accumulatedMeanGrads[i] == null && this.centered) {
                    this.accumulatedMeanGrads[i] = {
                        originalName: `${name}/mg`,
                        variable: tidy(() => zerosLike$2(value).variable(trainable))
                    };
                }
                const gradient = Array.isArray(variableGradients) ?
                    variableGradients[i].tensor :
                    variableGradients[name];
                if (gradient == null) {
                    return;
                }
                const accumulatedMeanSquare = this.accumulatedMeanSquares[i].variable;
                const accumulatedMoments = this.accumulatedMoments[i].variable;
                tidy(() => {
                    const newAccumulatedMeanSquare = add$2(mul(accumulatedMeanSquare, this.decay), mul(square$2(gradient), 1 - this.decay));
                    if (this.centered) {
                        const accumulatedMeanGrad = this.accumulatedMeanGrads[i].variable;
                        // Centered gradient
                        const newAccumulatedMeanGrad = add$2(mul(accumulatedMeanGrad, this.decay), mul(gradient, 1 - this.decay));
                        const gradContribution = div$1(mul(gradient, this.learningRate), sqrt$2(sub$2(newAccumulatedMeanSquare, add$2(square$2(newAccumulatedMeanGrad), this.epsilon))));
                        const newAccumulatedMoments = add$2(mul(accumulatedMoments, this.momentum), gradContribution);
                        accumulatedMeanSquare.assign(newAccumulatedMeanSquare);
                        accumulatedMeanGrad.assign(newAccumulatedMeanGrad);
                        accumulatedMoments.assign(newAccumulatedMoments);
                        const newValue = sub$2(value, newAccumulatedMoments);
                        value.assign(newValue);
                    }
                    else {
                        // Plain gradient
                        const newAccumulatedMeanSquare = add$2(mul(accumulatedMeanSquare, this.decay), mul(square$2(gradient), 1 - this.decay));
                        const newAccumulatedMoments = add$2(mul(accumulatedMoments, this.momentum), div$1(mul(gradient, this.learningRate), sqrt$2(add$2(newAccumulatedMeanSquare, this.epsilon))));
                        accumulatedMeanSquare.assign(newAccumulatedMeanSquare);
                        accumulatedMoments.assign(newAccumulatedMoments);
                        const newValue = sub$2(value, newAccumulatedMoments);
                        value.assign(newValue);
                    }
                });
            });
            this.incrementIterations();
        }
        dispose() {
            if (this.accumulatedMeanSquares != null) {
                dispose(this.accumulatedMeanSquares.map(v => v.variable));
            }
            if (this.accumulatedMeanGrads != null && this.centered) {
                dispose(this.accumulatedMeanGrads.map(v => v.variable));
            }
            if (this.accumulatedMoments != null) {
                dispose(this.accumulatedMoments.map(v => v.variable));
            }
        }
        async getWeights() {
            // Order matters for Python compatibility.
            const variables = [...this.accumulatedMeanSquares, ...this.accumulatedMoments];
            if (this.centered) {
                variables.push(...this.accumulatedMeanGrads);
            }
            return [await this.saveIterations()].concat(variables.map(v => ({ name: v.originalName, tensor: v.variable })));
        }
        async setWeights(weightValues) {
            weightValues = await this.extractIterations(weightValues);
            const variableCount = this.centered ? weightValues.length / 3 : weightValues.length / 2;
            const trainable = false;
            this.accumulatedMeanSquares =
                weightValues.slice(0, variableCount).map(v => ({
                    originalName: v.name,
                    variable: v.tensor.variable(trainable)
                }));
            this.accumulatedMoments =
                weightValues.slice(variableCount, variableCount * 2)
                    .map(v => ({
                    originalName: v.name,
                    variable: v.tensor.variable(trainable)
                }));
            if (this.centered) {
                this.accumulatedMeanGrads =
                    weightValues.slice(variableCount * 2, variableCount * 3)
                        .map(v => ({
                        originalName: v.name,
                        variable: v.tensor.variable(trainable)
                    }));
            }
        }
        getConfig() {
            return {
                'learningRate': this.learningRate,
                'decay': this.decay,
                'momentum': this.momentum,
                'epsilon': this.epsilon,
                'centered': this.centered
            };
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config['learningRate'], config['decay'], config['momentum'], config['epsilon'], config['centered']);
        }
    }
    /** @nocollapse */
    RMSPropOptimizer.className = 'RMSProp'; // Note: Name matters for Python compatibility.
    registerClass(RMSPropOptimizer);

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class OptimizerConstructors {
        /**
         * Constructs a `tf.SGDOptimizer` that uses stochastic gradient descent.
         *
         * ```js
         * // Fit a quadratic function by learning the coefficients a, b, c.
         * const xs = tf.tensor1d([0, 1, 2, 3]);
         * const ys = tf.tensor1d([1.1, 5.9, 16.8, 33.9]);
         *
         * const a = tf.scalar(Math.random()).variable();
         * const b = tf.scalar(Math.random()).variable();
         * const c = tf.scalar(Math.random()).variable();
         *
         * // y = a * x^2 + b * x + c.
         * const f = x => a.mul(x.square()).add(b.mul(x)).add(c);
         * const loss = (pred, label) => pred.sub(label).square().mean();
         *
         * const learningRate = 0.01;
         * const optimizer = tf.train.sgd(learningRate);
         *
         * // Train the model.
         * for (let i = 0; i < 10; i++) {
         *   optimizer.minimize(() => loss(f(xs), ys));
         * }
         *
         * // Make predictions.
         * console.log(
         *     `a: ${a.dataSync()}, b: ${b.dataSync()}, c: ${c.dataSync()}`);
         * const preds = f(xs).dataSync();
         * preds.forEach((pred, i) => {
         *   console.log(`x: ${i}, pred: ${pred}`);
         * });
         * ```
         *
         * @param learningRate The learning rate to use for the SGD algorithm.
         *
         * @doc {heading: 'Training', subheading: 'Optimizers', namespace: 'train'}
         */
        static sgd(learningRate) {
            return new SGDOptimizer(learningRate);
        }
        /**
         * Constructs a `tf.MomentumOptimizer` that uses momentum gradient
         * descent.
         *
         * See
         * [http://proceedings.mlr.press/v28/sutskever13.pdf](
         * http://proceedings.mlr.press/v28/sutskever13.pdf)
         *
         * @param learningRate The learning rate to use for the Momentum gradient
         * descent algorithm.
         * @param momentum The momentum to use for the momentum gradient descent
         * algorithm.
         *
         * @doc {heading: 'Training', subheading: 'Optimizers', namespace: 'train'}
         */
        static momentum(learningRate, momentum, useNesterov = false) {
            return new MomentumOptimizer(learningRate, momentum, useNesterov);
        }
        /**
         * Constructs a `tf.RMSPropOptimizer` that uses RMSProp gradient
         * descent. This implementation uses plain momentum and is not centered
         * version of RMSProp.
         *
         * See
         * [http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf](
         * http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf)
         *
         * @param learningRate The learning rate to use for the RMSProp gradient
         * descent algorithm.
         * @param decay The discounting factor for the history/coming gradient.
         * @param momentum The momentum to use for the RMSProp gradient descent
         * algorithm.
         * @param epsilon Small value to avoid zero denominator.
         * @param centered If true, gradients are normalized by the estimated
         * variance of the gradient.
         *
         * @doc {heading: 'Training', subheading: 'Optimizers', namespace: 'train'}
         */
        static rmsprop(learningRate, decay = .9, momentum = 0.0, epsilon = null, centered = false) {
            return new RMSPropOptimizer(learningRate, decay, momentum, epsilon, centered);
        }
        /**
         * Constructs a `tf.AdamOptimizer` that uses the Adam algorithm.
         * See [https://arxiv.org/abs/1412.6980](https://arxiv.org/abs/1412.6980)
         *
         * @param learningRate The learning rate to use for the Adam gradient
         * descent algorithm.
         * @param beta1 The exponential decay rate for the 1st moment estimates.
         * @param beta2 The exponential decay rate for the 2nd moment estimates.
         * @param epsilon A small constant for numerical stability.
         *
         * @doc {heading: 'Training', subheading: 'Optimizers', namespace: 'train'}
         */
        static adam(learningRate = 0.001, beta1 = 0.9, beta2 = 0.999, epsilon = null) {
            return new AdamOptimizer(learningRate, beta1, beta2, epsilon);
        }
        /**
         * Constructs a `tf.AdadeltaOptimizer` that uses the Adadelta algorithm.
         * See [https://arxiv.org/abs/1212.5701](https://arxiv.org/abs/1212.5701)
         *
         * @param learningRate The learning rate to use for the Adadelta gradient
         * descent algorithm.
         * @param rho The learning rate decay over each update.
         * @param epsilon A constant epsilon used to better condition the grad
         * update.
         *
         * @doc {heading: 'Training', subheading: 'Optimizers', namespace: 'train'}
         */
        static adadelta(learningRate = .001, rho = .95, epsilon = null) {
            return new AdadeltaOptimizer(learningRate, rho, epsilon);
        }
        /**
         * Constructs a `tf.AdamaxOptimizer` that uses the Adamax algorithm.
         * See [https://arxiv.org/abs/1412.6980](https://arxiv.org/abs/1412.6980)
         *
         * @param learningRate The learning rate to use for the Adamax gradient
         * descent algorithm.
         * @param beta1 The exponential decay rate for the 1st moment estimates.
         * @param beta2 The exponential decay rate for the 2nd moment estimates.
         * @param epsilon A small constant for numerical stability.
         * @param decay The learning rate decay over each update.
         *
         * @doc {heading: 'Training', subheading: 'Optimizers', namespace: 'train'}
         */
        static adamax(learningRate = 0.002, beta1 = 0.9, beta2 = 0.999, epsilon = null, decay = 0.0) {
            return new AdamaxOptimizer(learningRate, beta1, beta2, epsilon, decay);
        }
        /**
         * Constructs a `tf.AdagradOptimizer` that uses the Adagrad algorithm.
         * See
         * [http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf](
         * http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf)
         * or
         * [http://ruder.io/optimizing-gradient-descent/index.html#adagrad](
         * http://ruder.io/optimizing-gradient-descent/index.html#adagrad)
         *
         * @param learningRate The learning rate to use for the Adagrad gradient
         * descent algorithm.
         * @param initialAccumulatorValue Starting value for the accumulators, must be
         * positive.
         *
         * @doc {heading: 'Training', subheading: 'Optimizers', namespace: 'train'}
         */
        static adagrad(learningRate, initialAccumulatorValue = 0.1) {
            return new AdagradOptimizer(learningRate, initialAccumulatorValue);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const train = {
        sgd: OptimizerConstructors.sgd,
        momentum: OptimizerConstructors.momentum,
        adadelta: OptimizerConstructors.adadelta,
        adagrad: OptimizerConstructors.adagrad,
        rmsprop: OptimizerConstructors.rmsprop,
        adamax: OptimizerConstructors.adamax,
        adam: OptimizerConstructors.adam
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const delayCallback = (() => {
        if (typeof requestAnimationFrame !== 'undefined') {
            return requestAnimationFrame;
        }
        else if (typeof setImmediate !== 'undefined') {
            return setImmediate;
        }
        return (f) => f(); // no delays
    })();
    /**
     * Returns a promise that resolve when a requestAnimationFrame has completed.
     *
     * On Node.js this uses setImmediate instead of requestAnimationFrame.
     *
     * This is simply a sugar method so that users can do the following:
     * `await tf.nextFrame();`
     *
     * @doc {heading: 'Performance', subheading: 'Timing'}
     */
    function nextFrame() {
        return new Promise(resolve => delayCallback(() => resolve()));
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function assertParamsConsistent(shapes, axis) {
        const rank = shapes[0].length;
        shapes.forEach((shape, i) => {
            assert$4(shape.length === rank, () => `Error in concat${rank}D: rank of tensors[${i}] must be the same ` +
                `as the rank of the rest (${rank})`);
        });
        assert$4(axis >= 0 && axis < rank, () => `Error in concat${rank}D: axis must be between 0 and ${rank - 1}.`);
        const firstShape = shapes[0];
        shapes.forEach((shape, i) => {
            for (let r = 0; r < rank; r++) {
                assert$4((r === axis) || (shape[r] === firstShape[r]), () => `Error in concat${rank}D: Shape of tensors[${i}] (${shape}) ` +
                    `does not match the shape of the rest (${firstShape}) ` +
                    `along the non-concatenated axis ${i}.`);
            }
        });
    }
    function computeOutShape$1(shapes, axis) {
        const outputShape = shapes[0].slice();
        for (let i = 1; i < shapes.length; i++) {
            outputShape[axis] += shapes[i][axis];
        }
        return outputShape;
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const PARALLELIZE_THRESHOLD = 30;
    function computeOptimalWindowSize(inSize) {
        if (inSize <= PARALLELIZE_THRESHOLD) {
            return inSize;
        }
        return nearestDivisor(inSize, Math.floor(Math.sqrt(inSize)));
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Returns the image center in pixels.
    function getImageCenter(center, imageHeight, imageWidth) {
        const centerX = imageWidth * (typeof center === 'number' ? center : center[0]);
        const centerY = imageHeight * (typeof center === 'number' ? center : center[1]);
        return [centerX, centerY];
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Gets the new shape of the input Tensor after it's been reshaped
     * to:
     * [blockShape[0], ..., blockShape[M-1], batch / prod(blockShape),
     * inputShape[1], ..., inputShape[N-1]]
     *
     * See step 1: https://www.tensorflow.org/api_docs/python/tf/batch_to_space_nd
     */
    function getReshaped(inputShape, blockShape, prod, batchToSpace = true) {
        let reshaped = [];
        if (batchToSpace) {
            reshaped = reshaped.concat(blockShape.slice(0));
            reshaped.push(inputShape[0] / prod);
            reshaped = reshaped.concat(inputShape.slice(1));
        }
        else {
            reshaped = reshaped.concat(inputShape[0]);
            const spatialLength = blockShape.length;
            for (let i = 0; i < spatialLength; ++i) {
                reshaped =
                    reshaped.concat([inputShape[i + 1] / blockShape[i], blockShape[i]]);
            }
            reshaped = reshaped.concat(inputShape.slice(spatialLength + 1));
        }
        return reshaped;
    }
    /**
     * Gets the permutation that will transpose the dimensions of the
     * reshaped tensor to shape:
     *
     * [batch / prod(block_shape),inputShape[1], blockShape[0], ...,
     * inputShape[M], blockShape[M-1],inputShape[M+1], ..., inputShape[N-1]]
     *
     * see step 2: https://www.tensorflow.org/api_docs/python/tf/batch_to_space_nd
     */
    function getPermuted(reshapedRank, blockShapeRank, batchToSpace = true) {
        const permuted = [];
        if (batchToSpace) {
            permuted.push(blockShapeRank);
            for (let i = blockShapeRank + 1; i < reshapedRank; ++i) {
                if (i <= 2 * blockShapeRank) {
                    permuted.push(i);
                    permuted.push(i - (blockShapeRank + 1));
                }
                else {
                    permuted.push(i);
                }
            }
        }
        else {
            const permutedBeforeBatch = [];
            const permutedAfterBatch = [];
            for (let i = 1; i < reshapedRank; ++i) {
                if (i >= blockShapeRank * 2 + 1 || i % 2 === 1) {
                    permutedAfterBatch.push(i);
                }
                else {
                    permutedBeforeBatch.push(i);
                }
            }
            permuted.push(...permutedBeforeBatch);
            permuted.push(0);
            permuted.push(...permutedAfterBatch);
        }
        return permuted;
    }
    /**
     * Gets the shape of the reshaped and permuted input Tensor before any cropping
     * is applied.  The new shape will be:
     *
     * [batch / prod(blockShape),inputShape[1] * blockShape[0], ...,
     * inputShape[M] * blockShape[M-1],inputShape[M+1], ..., inputShape[N-1]]
     *
     * See step 3: https://www.tensorflow.org/api_docs/python/tf/batch_to_space_nd
     */
    function getReshapedPermuted(inputShape, blockShape, prod, batchToSpace = true) {
        const reshapedPermuted = [];
        if (batchToSpace) {
            reshapedPermuted.push(inputShape[0] / prod);
        }
        else {
            reshapedPermuted.push(inputShape[0] * prod);
        }
        for (let i = 1; i < inputShape.length; ++i) {
            if (i <= blockShape.length) {
                if (batchToSpace) {
                    reshapedPermuted.push(blockShape[i - 1] * inputShape[i]);
                }
                else {
                    reshapedPermuted.push(inputShape[i] / blockShape[i - 1]);
                }
            }
            else {
                reshapedPermuted.push(inputShape[i]);
            }
        }
        return reshapedPermuted;
    }
    /**
     * Converts the crops argument into the beginning coordinates of a slice
     * operation.
     */
    function getSliceBeginCoords(crops, blockShape) {
        const sliceBeginCoords = [0];
        for (let i = 0; i < blockShape; ++i) {
            sliceBeginCoords.push(crops[i][0]);
        }
        return sliceBeginCoords;
    }
    /**
     * Converts the crops argument into the size of a slice operation.  When
     * combined with getSliceBeginCoords this function allows the reshaped and
     * permuted Tensor to be cropped to its final output shape of:
     *
     * inputShape[1] * blockShape[0] - crops[0,0] - crops[0,1], ...,
     * inputShape[M] * blockShape[M-1] -crops[M-1,0] -
     * crops[M-1,1],inputShape[M+1], ..., inputShape[N-1]]
     *
     * See step 4: https://www.tensorflow.org/api_docs/python/tf/batch_to_space_nd
     */
    function getSliceSize(uncroppedShape, crops, blockShape) {
        const sliceSize = uncroppedShape.slice(0, 1);
        for (let i = 0; i < blockShape; ++i) {
            sliceSize.push(uncroppedShape[i + 1] - crops[i][0] - crops[i][1]);
        }
        return sliceSize;
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SELU_SCALEALPHA = 1.7580993408473768599402175208123;
    const SELU_SCALE = 1.0507009873554804934193349852946;

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ERF_P = 0.3275911;
    const ERF_A1 = 0.254829592;
    const ERF_A2 = -0.284496736;
    const ERF_A3 = 1.421413741;
    const ERF_A4 = -1.453152027;
    const ERF_A5 = 1.061405429;

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Merges real and imaginary Float32Arrays into a single complex Float32Array.
     *
     * The memory layout is interleaved as follows:
     * real: [r0, r1, r2]
     * imag: [i0, i1, i2]
     * complex: [r0, i0, r1, i1, r2, i2]
     *
     * This is the inverse of splitRealAndImagArrays.
     *
     * @param real The real values of the complex tensor values.
     * @param imag The imag values of the complex tensor values.
     * @returns A complex tensor as a Float32Array with merged values.
     */
    function mergeRealAndImagArrays(real, imag) {
        if (real.length !== imag.length) {
            throw new Error(`Cannot merge real and imag arrays of different lengths. real:` +
                `${real.length}, imag: ${imag.length}.`);
        }
        const result = new Float32Array(real.length * 2);
        for (let i = 0; i < result.length; i += 2) {
            result[i] = real[i / 2];
            result[i + 1] = imag[i / 2];
        }
        return result;
    }
    /**
     * Splits a complex Float32Array into real and imag parts.
     *
     * The memory layout is interleaved as follows:
     * complex: [r0, i0, r1, i1, r2, i2]
     * real: [r0, r1, r2]
     * imag: [i0, i1, i2]
     *
     * This is the inverse of mergeRealAndImagArrays.
     *
     * @param complex The complex tensor values.
     * @returns An object with real and imag Float32Array components of the complex
     *     tensor.
     */
    function splitRealAndImagArrays(complex) {
        const real = new Float32Array(complex.length / 2);
        const imag = new Float32Array(complex.length / 2);
        for (let i = 0; i < complex.length; i += 2) {
            real[i / 2] = complex[i];
            imag[i / 2] = complex[i + 1];
        }
        return { real, imag };
    }
    /**
     * Extracts even indexed complex values in the given array.
     * @param complex The complex tensor values
     */
    function complexWithEvenIndex(complex) {
        const len = Math.ceil(complex.length / 4);
        const real = new Float32Array(len);
        const imag = new Float32Array(len);
        for (let i = 0; i < complex.length; i += 4) {
            real[Math.floor(i / 4)] = complex[i];
            imag[Math.floor(i / 4)] = complex[i + 1];
        }
        return { real, imag };
    }
    /**
     * Extracts odd indexed comple values in the given array.
     * @param complex The complex tensor values
     */
    function complexWithOddIndex(complex) {
        const len = Math.floor(complex.length / 4);
        const real = new Float32Array(len);
        const imag = new Float32Array(len);
        for (let i = 2; i < complex.length; i += 4) {
            real[Math.floor(i / 4)] = complex[i];
            imag[Math.floor(i / 4)] = complex[i + 1];
        }
        return { real, imag };
    }
    /**
     * Get the map representing a complex value in the given array.
     * @param complex The complex tensor values.
     * @param index An index of the target complex value.
     */
    function getComplexWithIndex(complex, index) {
        const real = complex[index * 2];
        const imag = complex[index * 2 + 1];
        return { real, imag };
    }
    /**
     * Insert a given complex value into the TypedArray.
     * @param data The array in which the complex value is inserted.
     * @param c The complex value to be inserted.
     * @param index An index of the target complex value.
     */
    function assignToTypedArray(data, real, imag, index) {
        data[index * 2] = real;
        data[index * 2 + 1] = imag;
    }
    /**
     * Make the list of exponent terms used by FFT.
     */
    function exponents(n, inverse) {
        const real = new Float32Array(n / 2);
        const imag = new Float32Array(n / 2);
        for (let i = 0; i < Math.ceil(n / 2); i++) {
            const x = (inverse ? 2 : -2) * Math.PI * (i / n);
            real[i] = Math.cos(x);
            imag[i] = Math.sin(x);
        }
        return { real, imag };
    }
    /**
     * Make the exponent term used by FFT.
     */
    function exponent(k, n, inverse) {
        const x = (inverse ? 2 : -2) * Math.PI * (k / n);
        const real = Math.cos(x);
        const imag = Math.sin(x);
        return { real, imag };
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ARROW = '->';
    const ARROW_REGEX = /->/g;
    const COMMA = ',';
    const ELLIPSIS = '...';
    /**
     * Parse an equation for einsum.
     *
     * @param equation The einsum equation (e.g., "ij,jk->ik").
     * @param numTensors Number of tensors provided along with `equation`. Used to
     *   check matching number of input tensors.
     * @returns An object consisting of the following fields:
     *   - allDims: all dimension names as strings.
     *   - summedDims: a list of all dimensions being summed over, as indices to
     *     the elements of `allDims`.
     *   - idDims: indices of the dimensions in each input tensor, as indices to
     *     the elements of `allDims.
     */
    function decodeEinsumEquation(equation, numTensors) {
        equation = equation.replace(/\s/g, ''); // Remove witespace in equation.
        const numArrows = (equation.length - equation.replace(ARROW_REGEX, '').length) /
            ARROW.length;
        if (numArrows < 1) {
            throw new Error('Equations without an arrow are not supported.');
        }
        else if (numArrows > 1) {
            throw new Error(`Equation must contain exactly one arrow ("${ARROW}").`);
        }
        const [inputString, outputString] = equation.split(ARROW);
        assert$4(inputString.indexOf(ELLIPSIS) === -1, () => `The ellipsis notation ("${ELLIPSIS}") is not supported yet.`);
        const inputTerms = inputString.split(COMMA);
        const numInputs = inputTerms.length;
        if (numTensors !== numInputs) {
            throw new Error(`Expected ${numInputs} input tensors, received ${numTensors}`);
        }
        if (numInputs > 2) {
            throw new Error('Support for more than 2 input tensors is not implemented yet.');
        }
        const allDims = [];
        for (let i = 0; i < outputString.length; ++i) {
            const dimName = outputString[i];
            if (!inputTerms.some(inputTerm => inputTerm.indexOf(dimName) !== -1)) {
                throw new Error(`Output subscripts contain the label ${dimName} ` +
                    `not present in the input subscripts.`);
            }
            if (allDims.indexOf(dimName) === -1) {
                allDims.push(dimName);
            }
        }
        for (let i = 0; i < inputString.length; ++i) {
            const dimName = inputString[i];
            if (allDims.indexOf(dimName) === -1 && dimName !== COMMA) {
                allDims.push(dimName);
            }
        }
        const idDims = new Array(inputTerms.length);
        for (let i = 0; i < numInputs; ++i) {
            if (new Set(inputTerms[i].split('')).size !== inputTerms[i].length) {
                throw new Error(`Found duplicate axes in input component ${inputTerms[i]}. ` +
                    `Support for duplicate axes in input is not implemented yet.`);
            }
            idDims[i] = [];
            for (let j = 0; j < inputTerms[i].length; ++j) {
                idDims[i].push(allDims.indexOf(inputTerms[i][j]));
            }
        }
        const numDims = allDims.length; // Number of unique dimensions.
        const numOutDims = outputString.length; // Number of output dimensions.
        const summedDims = []; // Dimensions being summed over.
        for (let i = numOutDims; i < numDims; ++i) {
            summedDims.push(i);
        }
        return { allDims, summedDims, idDims };
    }
    /**
     * Get the permutation for a given input tensor.
     *
     * @param nDims Total number of dimension of all tensors involved in the einsum
     *   operation.
     * @param idDims Dimension indices involve in the tensor in question.
     * @returns An object consisting of the following fields:
     *   - permutationIndices: Indices to permute the axes of the tensor with.
     *   - expandDims: Indices to the dimension that need to be expanded from the
     *     tensor after permutation.
     */
    function getEinsumPermutation(nDims, idDims) {
        let permutationIndices = new Array(nDims);
        permutationIndices.fill(-1);
        for (let i = 0; i < idDims.length; ++i) {
            permutationIndices[idDims[i]] = i;
        }
        const expandDims = [];
        for (let i = 0; i < nDims; ++i) {
            if (permutationIndices[i] === -1) {
                expandDims.push(i);
            }
        }
        permutationIndices = permutationIndices.filter(d => d !== -1);
        return { permutationIndices, expandDims };
    }
    /**
     * Checks that the dimension sizes from different input tensors match the
     * equation.
     */
    function checkEinsumDimSizes(nDims, idDims, tensors) {
        const dimSizes = new Array(nDims);
        for (let i = 0; i < tensors.length; ++i) {
            const shape = tensors[i].shape;
            for (let j = 0; j < idDims[i].length; ++j) {
                if (dimSizes[idDims[i][j]] === undefined) {
                    dimSizes[idDims[i][j]] = shape[j];
                }
                else {
                    assert$4(dimSizes[idDims[i][j]] === shape[j], () => `Expected dimension ${dimSizes[idDims[i][j]]} at axis ${j} ` +
                        `of input shaped ${JSON.stringify(shape)}, ` +
                        `but got dimension ${shape[j]}`);
                }
            }
        }
    }
    /**
     * Gets path of computation for einsum.
     *
     * @param summedDims indices to the dimensions being summed over.
     * @param idDims A look up table for the dimensions present in each input
     *     tensor. Each consituent array contains indices for the dimensions in the
     *     corresponding input tensor.
     *
     * @return A map with two fields:
     *   - path: The path of computation, with each element indicating the dimension
     *     being summed over after the element-wise multiplication in that step.
     *   - steps: With the same length as `path`. Each element contains the indices
     *     to the input tensors being used for element-wise multiplication in the
     *     corresponding step.
     */
    function getEinsumComputePath(summedDims, idDims) {
        const path = summedDims;
        const steps = [];
        let nSteps = 0;
        if (summedDims.length === 0) {
            // Einsum that involes no summing: e.g., transpose and outer product.
            path.push(-1);
        }
        nSteps = summedDims.length + 1;
        for (let i = 0; i < nSteps; ++i) {
            steps.push([]);
        }
        const computedTermIndices = [];
        for (let i = 0; i < path.length; ++i) {
            const summedDim = path[i];
            const termIndices = findTermsWithDim(idDims, summedDim);
            for (const termIndex of termIndices) {
                if (computedTermIndices.indexOf(termIndex) === -1) {
                    steps[i].push(termIndex);
                    computedTermIndices.push(termIndex);
                }
            }
        }
        return { path, steps };
    }
    /** Determines if an axes permutation is the identity permutation. */
    function isIdentityPermutation(perm) {
        return perm.every((dim, index) => dim === index);
    }
    function findTermsWithDim(idDims, dim) {
        const termIndices = [];
        for (let i = 0; i < idDims.length; ++i) {
            if (idDims[i].length === 0 || idDims[i].indexOf(dim) !== -1 || dim === -1) {
                termIndices.push(i);
            }
        }
        return termIndices;
    }

    /**
     * Prepare the split size array. When the input is a number, the axis is evenly
     * divided among the split size. When the input contains the negative value, the
     * rest of the axis is allocated toward that.
     */
    function prepareSplitSize(x, numOrSizeSplits, axis = 0) {
        let splitSizes = [];
        if (typeof (numOrSizeSplits) === 'number') {
            assert$4(x.shape[axis] % numOrSizeSplits === 0, () => 'Number of splits must evenly divide the axis.');
            splitSizes =
                new Array(numOrSizeSplits).fill(x.shape[axis] / numOrSizeSplits);
        }
        else {
            const numOfNegs = numOrSizeSplits.reduce((count, value) => {
                if (value === -1) {
                    count += 1;
                }
                return count;
            }, 0);
            assert$4(numOfNegs <= 1, () => 'There should be only one negative value in split array.');
            const negIndex = numOrSizeSplits.indexOf(-1);
            // Allow the number of split array to be -1, which indicates the rest
            // of dimension is allocated to that split.
            if (negIndex !== -1) {
                const total = numOrSizeSplits.reduce((a, b) => b > 0 ? a + b : a);
                numOrSizeSplits[negIndex] = x.shape[axis] - total;
            }
            assert$4(x.shape[axis] === numOrSizeSplits.reduce((a, b) => a + b), () => 'The sum of sizes must match the size of the axis dimension.');
            splitSizes = numOrSizeSplits;
        }
        return splitSizes;
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Generates sparse fill empty rows indices, dense shape mismatch error message.
     *
     * @param indicesLength The first dimension of indices.
     */
    function getSparseFillEmptyRowsIndicesDenseShapeMismatch(indicesLength) {
        return `Received SparseTensor with denseShape[0] = 0 but
  indices.shape[0] = ${indicesLength}`;
    }
    /**
     * Generates sparse fill empty rows negative index error message.
     *
     * @param index The index with a negative value.
     * @param value The negative value.
     */
    function getSparseFillEmptyRowsNegativeIndexErrorMessage(index, value) {
        return `indices(${index}, 0) is invalid: ${value} < 0`;
    }
    /**
     * Generates sparse fill empty rows out of range index error message.
     *
     * @param index The index with an out of range value.
     * @param value The out of range value.
     * @param limit The upper limit for indices.
     */
    function getSparseFillEmptyRowsOutOfRangeIndexErrorMessage(index, value, limit) {
        return `indices(${index}, 0) is invalid: ${value} >= ${limit}`;
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Generates sparse reshape multiple negative 1 output dimension error message.
     *
     * @param dim1 The first dimension with a negative 1 value.
     * @param dim2 The second dimension with a negative 1 value.
     */
    function getSparseReshapeMultipleNegativeOneOutputDimErrorMessage(dim1, dim2) {
        return `only one output dimension may be -1, not both ${dim1} and ${dim2}`;
    }
    /**
     * Generates sparse reshape negative output dimension error message.
     *
     * @param dim The dimension with a negative value.
     * @param value The negative value.
     */
    function getSparseReshapeNegativeOutputDimErrorMessage(dim, value) {
        return `size ${dim} must be non-negative, not ${value}`;
    }
    /**
     * Generates sparse reshape empty tensor zero output dimension error message.
     *
     */
    function getSparseReshapeEmptyTensorZeroOutputDimErrorMessage() {
        return 'reshape cannot infer the missing input size for an empty tensor ' +
            'unless all specified input sizes are non-zero';
    }
    /**
     * Generates sparse reshape input output multiple mismatch error message.
     *
     * @param inputShape the input shape.
     * @param outputShape the requested output shape.
     */
    function getSparseReshapeInputOutputMultipleErrorMessage(inputShape, outputShape) {
        const inputSize = sizeFromShape(inputShape);
        const outputSize = sizeFromShape(outputShape);
        return `Input to reshape is a SparseTensor with ${inputSize}
  dense values, but the requested shape requires a multiple of ${outputSize}. inputShape=${inputShape} outputShape= ${outputShape}`;
    }
    /**
     * Generates sparse reshape input output inequality error message.
     *
     * @param inputShape the input shape.
     * @param outputShape the requested output shape.
     */
    function getSparseReshapeInputOutputMismatchErrorMessage(inputShape, outputShape) {
        const inputSize = sizeFromShape(inputShape);
        const outputSize = sizeFromShape(outputShape);
        return `Input to reshape is a tensor with ${inputSize} dense values, but the requested shape has ${outputSize}. inputShape=${inputShape} outputShape=${outputShape}`;
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Generates sparse segment reduction negative segment ids error message.
     *
     */
    function getSparseSegmentReductionNegativeSegmentIdsErrorMessage() {
        return `segment ids must be >= 0`;
    }
    /**
     * Generates sparse segment reduction non increasing segment ids error message.
     *
     */
    function getSparseSegmentReductionNonIncreasingSegmentIdsErrorMessage() {
        return `segment ids are not increasing`;
    }
    /**
     * Generates sparse segment reduction segment id out of range error message.
     *
     * @param segmentId The segment id index that is out of range.
     * @param outputRows Upper bound of valid segment id values.
     */
    function getSparseSegmentReductionSegmentIdOutOfRangeErrorMessage(segmentId, outputRows) {
        return `Segment id ${segmentId} out of range [0, ${outputRows}), possibly because segmentIds input is not sorted.`;
    }
    /**
     * Generates sparse segment reduction input indice out of range error message.
     *
     * @param index The index that holds the out of range value.
     * @param indexValue The value that is out of range.
     * @param inputRows Upper bound of valid index values.
     */
    function getSparseSegmentReductionIndicesOutOfRangeErrorMessage(index, indexValue, inputRows) {
        return `Bad: indices[${index}] == ${indexValue} out of range [0, ${inputRows})`;
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function segOpComputeOptimalWindowSize(inSize, numSegments) {
        let done = false;
        let res;
        if (inSize <= PARALLELIZE_THRESHOLD) {
            res = inSize;
            done = true;
        }
        else {
            res = nearestDivisor(inSize, Math.floor(Math.sqrt(inSize)));
        }
        while (!done) {
            if (res > numSegments || res === inSize) {
                done = true;
            }
            else {
                res = nearestDivisor(inSize, res + 1);
            }
        }
        return res;
    }
    function computeOutShape(aShape, axis, numSegments) {
        const outShape = [];
        const rank = aShape.length;
        for (let dim = 0; dim < rank; dim++) {
            if (dim !== axis) {
                outShape.push(aShape[dim]);
            }
            else {
                outShape.push(numSegments);
            }
        }
        return outShape;
    }
    function collectGatherOpShapeInfo(x, indices, axis, batchDims) {
        const indicesRank = indices.shape.length;
        const xRank = x.shape.length;
        if (batchDims !== 0) {
            if (batchDims < -indicesRank || batchDims > indicesRank) {
                throw new Error(`Expect batchDims in the range of [-${indicesRank}, ${indicesRank}], but got ${batchDims}`);
            }
        }
        if (batchDims < 0) {
            batchDims += indicesRank;
        }
        if (batchDims > xRank) {
            throw new Error(`batchDims (${batchDims}) must be less than rank(x) (
    ${xRank}).`);
        }
        if (axis < batchDims) {
            throw new Error(`batchDims (${batchDims}) must be less than or equal to axis (${axis}).`);
        }
        for (let i = 0; i < batchDims; ++i) {
            if (x.shape[i] !== indices.shape[i]) {
                throw new Error(`x.shape[${i}]: ${x.shape[i]} should be equal to indices.shape[${i}]: ${indices.shape[i]}.`);
            }
        }
        const dimSize = x.shape[axis];
        const outputShape = [];
        let batchSize = 1;
        let outerSize = 1;
        let sliceSize = 1;
        for (let i = 0; i < batchDims; ++i) {
            outputShape.push(x.shape[i]);
            batchSize *= x.shape[i];
        }
        for (let i = batchDims; i < axis; i++) {
            outputShape.push(x.shape[i]);
            outerSize *= x.shape[i];
        }
        for (let i = batchDims; i < indicesRank; i++) {
            outputShape.push(indices.shape[i]);
        }
        for (let i = axis + 1; i < xRank; i++) {
            outputShape.push(x.shape[i]);
            sliceSize *= x.shape[i];
        }
        return { batchSize, sliceSize, outerSize, dimSize, outputShape };
    }

    var segment_util = /*#__PURE__*/Object.freeze({
        __proto__: null,
        segOpComputeOptimalWindowSize: segOpComputeOptimalWindowSize,
        computeOutShape: computeOutShape,
        collectGatherOpShapeInfo: collectGatherOpShapeInfo
    });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fromUint8ToStringArray(vals) {
        try {
            // Decode the bytes into string.
            return vals.map(val => decodeString(val));
        }
        catch (err) {
            throw new Error(`Failed to decode encoded string bytes into utf-8, error: ${err}`);
        }
    }
    function fromStringArrayToUint8(strings) {
        return strings.map(s => encodeString(s));
    }

    var backend_util = /*#__PURE__*/Object.freeze({
        __proto__: null,
        slice_util: slice_util,
        segment_util: segment_util,
        fromUint8ToStringArray: fromUint8ToStringArray,
        fromStringArrayToUint8: fromStringArrayToUint8,
        upcastType: upcastType,
        axesAreInnerMostDims: axesAreInnerMostDims,
        combineLocations: combineLocations,
        computeOutAndReduceShapes: computeOutAndReduceShapes,
        expandShapeToKeepDim: expandShapeToKeepDim,
        assertAxesAreInnerMostDims: assertAxesAreInnerMostDims,
        getAxesPermutation: getAxesPermutation,
        getUndoAxesPermutation: getUndoAxesPermutation,
        getInnerMostAxes: getInnerMostAxes,
        getBroadcastDims: getBroadcastDims$1,
        getReductionAxes: getReductionAxes,
        assertAndGetBroadcastShape: assertAndGetBroadcastShape,
        assertParamsConsistent: assertParamsConsistent,
        computeOutShape: computeOutShape$1,
        computeDilation2DInfo: computeDilation2DInfo,
        computePool2DInfo: computePool2DInfo,
        computePool3DInfo: computePool3DInfo,
        computeConv2DInfo: computeConv2DInfo,
        computeConv3DInfo: computeConv3DInfo,
        computeDefaultPad: computeDefaultPad,
        tupleValuesAreOne: tupleValuesAreOne,
        eitherStridesOrDilationsAreOne: eitherStridesOrDilationsAreOne,
        convertConv2DDataFormat: convertConv2DDataFormat,
        checkPadOnDimRoundingMode: checkPadOnDimRoundingMode,
        getFusedDyActivation: getFusedDyActivation,
        getFusedBiasGradient: getFusedBiasGradient,
        applyActivation: applyActivation$1,
        shouldFuse: shouldFuse,
        PARALLELIZE_THRESHOLD: PARALLELIZE_THRESHOLD,
        computeOptimalWindowSize: computeOptimalWindowSize,
        getImageCenter: getImageCenter,
        getReshaped: getReshaped,
        getPermuted: getPermuted,
        getReshapedPermuted: getReshapedPermuted,
        getSliceBeginCoords: getSliceBeginCoords,
        getSliceSize: getSliceSize,
        prepareAndValidate: prepareAndValidate,
        validateUpdateShape: validateUpdateShape,
        validateInput: validateInput$1,
        calculateShapes: calculateShapes,
        SELU_SCALEALPHA: SELU_SCALEALPHA,
        SELU_SCALE: SELU_SCALE,
        ERF_P: ERF_P,
        ERF_A1: ERF_A1,
        ERF_A2: ERF_A2,
        ERF_A3: ERF_A3,
        ERF_A4: ERF_A4,
        ERF_A5: ERF_A5,
        warn: warn,
        log: log$4,
        mergeRealAndImagArrays: mergeRealAndImagArrays,
        splitRealAndImagArrays: splitRealAndImagArrays,
        complexWithEvenIndex: complexWithEvenIndex,
        complexWithOddIndex: complexWithOddIndex,
        getComplexWithIndex: getComplexWithIndex,
        assignToTypedArray: assignToTypedArray,
        exponents: exponents,
        exponent: exponent,
        decodeEinsumEquation: decodeEinsumEquation,
        getEinsumPermutation: getEinsumPermutation,
        checkEinsumDimSizes: checkEinsumDimSizes,
        getEinsumComputePath: getEinsumComputePath,
        isIdentityPermutation: isIdentityPermutation,
        prepareSplitSize: prepareSplitSize,
        getSparseFillEmptyRowsIndicesDenseShapeMismatch: getSparseFillEmptyRowsIndicesDenseShapeMismatch,
        getSparseFillEmptyRowsNegativeIndexErrorMessage: getSparseFillEmptyRowsNegativeIndexErrorMessage,
        getSparseFillEmptyRowsOutOfRangeIndexErrorMessage: getSparseFillEmptyRowsOutOfRangeIndexErrorMessage,
        getSparseReshapeMultipleNegativeOneOutputDimErrorMessage: getSparseReshapeMultipleNegativeOneOutputDimErrorMessage,
        getSparseReshapeNegativeOutputDimErrorMessage: getSparseReshapeNegativeOutputDimErrorMessage,
        getSparseReshapeEmptyTensorZeroOutputDimErrorMessage: getSparseReshapeEmptyTensorZeroOutputDimErrorMessage,
        getSparseReshapeInputOutputMultipleErrorMessage: getSparseReshapeInputOutputMultipleErrorMessage,
        getSparseReshapeInputOutputMismatchErrorMessage: getSparseReshapeInputOutputMismatchErrorMessage,
        getSparseSegmentReductionNegativeSegmentIdsErrorMessage: getSparseSegmentReductionNegativeSegmentIdsErrorMessage,
        getSparseSegmentReductionNonIncreasingSegmentIdsErrorMessage: getSparseSegmentReductionNonIncreasingSegmentIdsErrorMessage,
        getSparseSegmentReductionSegmentIdOutOfRangeErrorMessage: getSparseSegmentReductionSegmentIdOutOfRangeErrorMessage,
        getSparseSegmentReductionIndicesOutOfRangeErrorMessage: getSparseSegmentReductionIndicesOutOfRangeErrorMessage
    });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */

    var kernel_impls = /*#__PURE__*/Object.freeze({
        __proto__: null,
        nonMaxSuppressionV3Impl: nonMaxSuppressionV3Impl$2,
        nonMaxSuppressionV4Impl: nonMaxSuppressionV4Impl$2,
        nonMaxSuppressionV5Impl: nonMaxSuppressionV5Impl$2,
        whereImpl: whereImpl$2
    });

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */

    var dist$1 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        AdadeltaOptimizer: AdadeltaOptimizer,
        AdagradOptimizer: AdagradOptimizer,
        AdamOptimizer: AdamOptimizer,
        AdamaxOptimizer: AdamaxOptimizer,
        MomentumOptimizer: MomentumOptimizer,
        Optimizer: Optimizer,
        OptimizerConstructors: OptimizerConstructors,
        RMSPropOptimizer: RMSPropOptimizer,
        SGDOptimizer: SGDOptimizer,
        Tensor: Tensor,
        TensorBuffer: TensorBuffer,
        Variable: Variable,
        get Rank () { return Rank; },
        sumOutType: sumOutType,
        upcastType: upcastType,
        get Reduction () { return Reduction; },
        customGrad: customGrad,
        grad: grad,
        grads: grads,
        valueAndGrad: valueAndGrad,
        valueAndGrads: valueAndGrads,
        variableGrads: variableGrads,
        Environment: Environment,
        env: env$1,
        get ENV () { return ENV$4; },
        nextFrame: nextFrame,
        KernelBackend: KernelBackend,
        DataStorage: DataStorage,
        abs: abs$2,
        acos: acos$2,
        acosh: acosh$2,
        add: add$2,
        addN: addN$2,
        all: all$2,
        any: any$2,
        argMax: argMax$2,
        argMin: argMin$2,
        asin: asin$2,
        asinh: asinh$2,
        atan: atan$2,
        atan2: atan2$2,
        atanh: atanh$2,
        avgPool: avgPool$2,
        avgPool3d: avgPool3d$1,
        basicLSTMCell: basicLSTMCell,
        batchToSpaceND: batchToSpaceND$2,
        batchNorm: batchNorm$2,
        batchNorm2d: batchNorm2d,
        batchNorm3d: batchNorm3d,
        batchNorm4d: batchNorm4d,
        bincount: bincount$2,
        broadcastArgs: broadcastArgs$2,
        broadcastTo: broadcastTo,
        buffer: buffer,
        cast: cast$3,
        ceil: ceil$2,
        clipByValue: clipByValue$2,
        clone: clone,
        complex: complex$2,
        concat: concat$2,
        concat1d: concat1d,
        concat2d: concat2d,
        concat3d: concat3d,
        concat4d: concat4d,
        conv1d: conv1d$1,
        conv2d: conv2d$3,
        conv2dTranspose: conv2dTranspose$1,
        conv3d: conv3d$1,
        conv3dTranspose: conv3dTranspose$1,
        cos: cos$2,
        cosh: cosh$2,
        cumprod: cumprod$2,
        cumsum: cumsum$2,
        denseBincount: denseBincount$2,
        depthToSpace: depthToSpace$2,
        depthwiseConv2d: depthwiseConv2d$3,
        diag: diag$2,
        dilation2d: dilation2d,
        div: div$1,
        divNoNan: divNoNan,
        dot: dot$2,
        einsum: einsum$2,
        elu: elu$4,
        equal: equal$3,
        erf: erf$2,
        euclideanNorm: euclideanNorm,
        exp: exp$2,
        expandDims: expandDims$3,
        expm1: expm1$2,
        eye: eye,
        fill: fill$2,
        floor: floor$3,
        floorDiv: floorDiv$2,
        gather: gather$1,
        greater: greater$3,
        greaterEqual: greaterEqual$2,
        imag: imag$2,
        isFinite: isFinite$3,
        isInf: isInf$2,
        isNaN: isNaN$3,
        leakyRelu: leakyRelu$2,
        less: less$3,
        lessEqual: lessEqual$2,
        linspace: linspace,
        localResponseNormalization: localResponseNormalization,
        log: log$3,
        log1p: log1p$2,
        logSigmoid: logSigmoid,
        logSoftmax: logSoftmax,
        logSumExp: logSumExp,
        logicalAnd: logicalAnd$2,
        logicalNot: logicalNot$2,
        logicalOr: logicalOr$2,
        logicalXor: logicalXor,
        lowerBound: lowerBound$1,
        matMul: matMul$1,
        max: max$3,
        maxPool: maxPool$2,
        maxPool3d: maxPool3d$1,
        maxPoolWithArgmax: maxPoolWithArgmax,
        maximum: maximum$3,
        mean: mean$1,
        meshgrid: meshgrid,
        min: min$3,
        minimum: minimum$3,
        mirrorPad: mirrorPad$1,
        mod: mod$2,
        moments: moments,
        mul: mul,
        multiRNNCell: multiRNNCell,
        multinomial: multinomial$2,
        neg: neg$2,
        notEqual: notEqual$3,
        oneHot: oneHot$2,
        ones: ones$1,
        onesLike: onesLike$2,
        outerProduct: outerProduct,
        pad: pad$1,
        pad1d: pad1d,
        pad2d: pad2d,
        pad3d: pad3d,
        pad4d: pad4d,
        pool: pool$1,
        pow: pow$2,
        prelu: prelu$3,
        print: print,
        prod: prod$2,
        rand: rand,
        randomGamma: randomGamma,
        randomNormal: randomNormal$2,
        randomUniform: randomUniform$1,
        range: range$4,
        real: real$2,
        reciprocal: reciprocal$2,
        relu: relu$2,
        relu6: relu6$2,
        reshape: reshape$3,
        reverse: reverse$2,
        reverse1d: reverse1d,
        reverse2d: reverse2d,
        reverse3d: reverse3d,
        reverse4d: reverse4d,
        round: round$2,
        rsqrt: rsqrt$2,
        scalar: scalar,
        selu: selu$2,
        separableConv2d: separableConv2d$1,
        setdiff1dAsync: setdiff1dAsync,
        sigmoid: sigmoid$2,
        sign: sign$2,
        sin: sin$2,
        sinh: sinh$2,
        slice: slice$2,
        slice1d: slice1d,
        slice2d: slice2d,
        slice3d: slice3d,
        slice4d: slice4d,
        softmax: softmax$3,
        softplus: softplus$2,
        spaceToBatchND: spaceToBatchND$2,
        fft: fft$2,
        ifft: ifft$2,
        irfft: irfft,
        rfft: rfft,
        split: split$2,
        sqrt: sqrt$2,
        square: square$2,
        squaredDifference: squaredDifference$2,
        squeeze: squeeze,
        stack: stack,
        step: step$2,
        stridedSlice: stridedSlice$2,
        sub: sub$2,
        sum: sum$2,
        tan: tan$2,
        tanh: tanh$2,
        tensor: tensor,
        tensor1d: tensor1d,
        tensor2d: tensor2d,
        tensor3d: tensor3d,
        tensor4d: tensor4d,
        tensor5d: tensor5d,
        tensor6d: tensor6d,
        tile: tile$3,
        topk: topk,
        truncatedNormal: truncatedNormal$1,
        unique: unique$3,
        unsortedSegmentSum: unsortedSegmentSum$2,
        unstack: unstack,
        upperBound: upperBound$1,
        variable: variable,
        where: where,
        whereAsync: whereAsync,
        zeros: zeros$2,
        zerosLike: zerosLike$2,
        op: op,
        OP_SCOPE_SUFFIX: OP_SCOPE_SUFFIX,
        booleanMaskAsync: booleanMaskAsync,
        transpose: transpose$2,
        norm: norm,
        movingAverage: movingAverage,
        scatterND: scatterND,
        searchSorted: searchSorted$2,
        sparseToDense: sparseToDense$2,
        gatherND: gatherND,
        dropout: dropout$2,
        enclosingPowerOfTwo: enclosingPowerOfTwo,
        cosineWindow: cosineWindow,
        inTopKAsync: inTopKAsync,
        image: image$1,
        linalg: linalg,
        losses: losses,
        spectral: spectral$1,
        fused: fused_ops,
        signal: signal,
        sparse: sparse$1,
        string: string$1,
        train: train,
        enableProdMode: enableProdMode,
        enableDebugMode: enableDebugMode,
        disableDeprecationWarnings: disableDeprecationWarnings,
        deprecationWarn: deprecationWarn,
        disposeVariables: disposeVariables,
        engine: engine,
        memory: memory,
        profile: profile,
        tidy: tidy,
        dispose: dispose,
        keep: keep,
        time: time,
        setBackend: setBackend,
        ready: ready,
        getBackend: getBackend,
        removeBackend: removeBackend,
        findBackend: findBackend,
        findBackendFactory: findBackendFactory,
        registerBackend: registerBackend,
        backend: backend,
        setPlatform: setPlatform,
        getKernel: getKernel,
        getGradient: getGradient,
        getKernelsForBackend: getKernelsForBackend,
        registerKernel: registerKernel,
        registerGradient: registerGradient,
        unregisterKernel: unregisterKernel,
        unregisterGradient: unregisterGradient,
        copyRegisteredKernels: copyRegisteredKernels,
        Abs: Abs,
        Acos: Acos,
        Acosh: Acosh,
        Add: Add$1,
        AddN: AddN,
        All: All,
        Any: Any,
        ArgMax: ArgMax,
        ArgMin: ArgMin,
        Asin: Asin,
        Asinh: Asinh,
        Atan: Atan,
        Atanh: Atanh,
        Atan2: Atan2,
        AvgPool: AvgPool,
        AvgPoolGrad: AvgPoolGrad,
        AvgPool3D: AvgPool3D,
        AvgPool3DGrad: AvgPool3DGrad,
        BatchMatMul: BatchMatMul,
        BatchToSpaceND: BatchToSpaceND,
        Bincount: Bincount,
        BroadcastTo: BroadcastTo,
        BroadcastArgs: BroadcastArgs,
        Cast: Cast,
        Ceil: Ceil,
        ClipByValue: ClipByValue,
        Complex: Complex,
        ComplexAbs: ComplexAbs,
        Concat: Concat,
        Conv2D: Conv2D$1,
        Conv2DBackpropFilter: Conv2DBackpropFilter,
        Conv2DBackpropInput: Conv2DBackpropInput,
        Conv3D: Conv3D$1,
        Conv3DBackpropFilterV2: Conv3DBackpropFilterV2,
        Conv3DBackpropInputV2: Conv3DBackpropInputV2,
        Cos: Cos,
        Cosh: Cosh,
        Cumprod: Cumprod,
        Cumsum: Cumsum,
        CropAndResize: CropAndResize,
        DenseBincount: DenseBincount,
        DepthToSpace: DepthToSpace,
        DepthwiseConv2dNative: DepthwiseConv2dNative,
        DepthwiseConv2dNativeBackpropFilter: DepthwiseConv2dNativeBackpropFilter,
        DepthwiseConv2dNativeBackpropInput: DepthwiseConv2dNativeBackpropInput,
        Diag: Diag,
        Dilation2D: Dilation2D,
        Dilation2DBackpropInput: Dilation2DBackpropInput,
        Dilation2DBackpropFilter: Dilation2DBackpropFilter,
        RealDiv: RealDiv,
        Einsum: Einsum,
        Elu: Elu$1,
        EluGrad: EluGrad,
        Erf: Erf,
        Equal: Equal,
        Exp: Exp,
        ExpandDims: ExpandDims,
        Expm1: Expm1,
        FFT: FFT,
        Fill: Fill,
        FlipLeftRight: FlipLeftRight,
        Floor: Floor,
        FloorDiv: FloorDiv,
        FusedBatchNorm: FusedBatchNorm,
        GatherV2: GatherV2,
        GatherNd: GatherNd,
        Greater: Greater,
        GreaterEqual: GreaterEqual,
        Identity: Identity$1,
        IFFT: IFFT,
        Imag: Imag,
        IsFinite: IsFinite,
        IsInf: IsInf,
        IsNan: IsNan,
        LeakyRelu: LeakyRelu,
        Less: Less,
        LessEqual: LessEqual,
        LinSpace: LinSpace,
        Log: Log,
        Log1p: Log1p,
        LogicalAnd: LogicalAnd,
        LogicalNot: LogicalNot,
        LogicalOr: LogicalOr,
        LogSoftmax: LogSoftmax$1,
        LowerBound: LowerBound,
        LRN: LRN,
        LRNGrad: LRNGrad,
        Max: Max,
        Maximum: Maximum$1,
        MaxPool: MaxPool,
        MaxPoolGrad: MaxPoolGrad,
        MaxPool3D: MaxPool3D,
        MaxPool3DGrad: MaxPool3DGrad,
        MaxPoolWithArgmax: MaxPoolWithArgmax,
        Mean: Mean,
        Min: Min,
        Minimum: Minimum$1,
        MirrorPad: MirrorPad,
        Mod: Mod,
        Multinomial: Multinomial,
        Multiply: Multiply$1,
        Neg: Neg,
        NotEqual: NotEqual,
        NonMaxSuppressionV3: NonMaxSuppressionV3,
        NonMaxSuppressionV4: NonMaxSuppressionV4,
        NonMaxSuppressionV5: NonMaxSuppressionV5,
        OnesLike: OnesLike,
        OneHot: OneHot,
        Pack: Pack,
        PadV2: PadV2,
        Pool: Pool,
        Pow: Pow,
        Prelu: Prelu,
        Prod: Prod,
        Range: Range,
        Real: Real,
        Reciprocal: Reciprocal,
        Relu: Relu$1,
        Reshape: Reshape$1,
        ResizeNearestNeighbor: ResizeNearestNeighbor,
        ResizeNearestNeighborGrad: ResizeNearestNeighborGrad,
        ResizeBilinear: ResizeBilinear,
        ResizeBilinearGrad: ResizeBilinearGrad,
        Relu6: Relu6$1,
        Reverse: Reverse,
        Round: Round,
        Rsqrt: Rsqrt,
        ScatterNd: ScatterNd,
        SearchSorted: SearchSorted,
        Select: Select,
        Selu: Selu$1,
        Slice: Slice,
        Sin: Sin,
        Sinh: Sinh,
        Sign: Sign,
        Sigmoid: Sigmoid$1,
        Softplus: Softplus$1,
        Sqrt: Sqrt,
        Sum: Sum,
        SpaceToBatchND: SpaceToBatchND,
        SplitV: SplitV,
        Softmax: Softmax$2,
        SparseFillEmptyRows: SparseFillEmptyRows,
        SparseReshape: SparseReshape,
        SparseSegmentMean: SparseSegmentMean,
        SparseSegmentSum: SparseSegmentSum,
        SparseToDense: SparseToDense,
        SquaredDifference: SquaredDifference,
        Square: Square,
        StridedSlice: StridedSlice,
        StringNGrams: StringNGrams,
        StringSplit: StringSplit,
        StringToHashBucketFast: StringToHashBucketFast,
        Sub: Sub,
        Tan: Tan,
        Tanh: Tanh$1,
        Tile: Tile,
        TopK: TopK,
        Transform: Transform$1,
        Transpose: Transpose,
        Unique: Unique,
        Unpack: Unpack,
        UnsortedSegmentSum: UnsortedSegmentSum,
        UpperBound: UpperBound,
        ZerosLike: ZerosLike,
        Step: Step,
        FromPixels: FromPixels,
        RotateWithOffset: RotateWithOffset,
        _FusedMatMul: _FusedMatMul,
        FusedConv2D: FusedConv2D,
        FusedDepthwiseConv2D: FusedDepthwiseConv2D,
        version_core: version$7,
        browser: browser$2,
        io: io,
        math: math,
        serialization: serialization,
        test_util: test_util,
        util: util$2,
        backend_util: backend_util,
        broadcast_util: broadcast_util,
        tensor_util: tensor_util,
        slice_util: slice_util,
        gather_util: gather_nd_util,
        scatter_util: scatter_nd_util,
        device_util: device_util,
        kernel_impls: kernel_impls
    });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const absGradConfig = {
        kernelName: Abs,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => mul(dy, step$2(cast$3(x, 'float32'), -1)) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const acosGradConfig = {
        kernelName: Acos,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return {
                x: () => {
                    const a = square$2(cast$3(x, 'float32'));
                    const b = sqrt$2(sub$2(scalar(1), a));
                    return neg$2(div$1(dy, b));
                }
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const acoshGradConfig = {
        kernelName: Acosh,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return {
                x: () => {
                    const a = sqrt$2(sub$2(square$2(cast$3(x, 'float32')), 1));
                    return div$1(dy, a);
                }
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const addGradConfig = {
        kernelName: Add$1,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const outShape = assertAndGetBroadcastShape(a.shape, b.shape);
            const derA = () => {
                let res = dy;
                const reduceAxes = getReductionAxes(a.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = sum$2(res, reduceAxes);
                }
                return reshape$3(res, a.shape);
            };
            const derB = () => {
                let res = dy;
                const reduceAxes = getReductionAxes(b.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = sum$2(res, reduceAxes);
                }
                return reshape$3(res, b.shape);
            };
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const addNGradConfig = {
        kernelName: AddN,
        saveAllInputs: true,
        gradFunc: (dy, saved) => {
            const ders = {};
            saved.forEach((_, i) => {
                ders[i] = () => dy.clone();
            });
            return ders;
        }
    };

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const argMaxGradConfig = {
        kernelName: ArgMax,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => zerosLike$2(x) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const argMinGradConfig = {
        kernelName: ArgMin,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => zerosLike$2(x) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const asinGradConfig = {
        kernelName: Asin,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => div$1(dy, sqrt$2(sub$2(scalar(1), square$2(cast$3(x, 'float32'))))) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const asinhGradConfig = {
        kernelName: Asinh,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return {
                x: () => {
                    const a = sqrt$2(add$2(scalar(1), square$2(cast$3(x, 'float32'))));
                    return div$1(dy, a);
                }
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const atan2GradConfig = {
        kernelName: Atan2,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const outShape = assertAndGetBroadcastShape(a.shape, b.shape);
            const derA = () => {
                const d = add$2(square$2(a), square$2(b));
                let res = mul(dy, div$1(b, d));
                const reduceAxes = getReductionAxes(a.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = sum$2(res, reduceAxes);
                }
                return reshape$3(res, a.shape);
            };
            const derB = () => {
                const d = add$2(square$2(a), square$2(b));
                let res = neg$2(mul(dy, div$1(a, d)));
                const reduceAxes = getReductionAxes(b.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = sum$2(res, reduceAxes);
                }
                return reshape$3(res, b.shape);
            };
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const atanGradConfig = {
        kernelName: Atan,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => div$1(dy, add$2(square$2(cast$3(x, 'float32')), 1)) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const atanhGradConfig = {
        kernelName: Atanh,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => div$1(dy, sub$2(scalar(1), square$2(cast$3(x, 'float32')))) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the backprop of a 3d avg pool.
     *
     * @param dy The dy error, of rank 5 of shape
     *     [batchSize, depth, height, width, channels].
     * assumed.
     * @param input The original input image, of rank 5 or rank4 of shape
     *     [batchSize, depth, height, width, channels].
     * @param filterSize The filter size:
     *     `[filterDepth, filterHeight, filterWidth]`.
     *     `filterSize` is a single number,
     *     then `filterDepth == filterHeight == filterWidth`.
     * @param strides The strides of the pooling:
     *     `[strideDepth, strideHeight, strideWidth]`. If
     *     `strides` is a single number, then `strideHeight == strideWidth`.
     * @param pad A string from: 'same', 'valid'. The type of padding algorithm
     *     used in the forward prop of the op.
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     */
    function avgPool3dGrad_(dy, input, filterSize, strides, pad, dimRoundingMode) {
        const $dy = convertToTensor(dy, 'dy', 'avgPool3dGrad');
        const $input = convertToTensor(input, 'input', 'avgPool3dGrad');
        let dy5D = $dy;
        let input5D = $input;
        let reshapedTo5D = false;
        if ($input.rank === 4) {
            reshapedTo5D = true;
            dy5D = reshape$3($dy, [1, $dy.shape[0], $dy.shape[1], $dy.shape[2], $dy.shape[3]]);
            input5D = reshape$3($input, [
                1, $input.shape[0], $input.shape[1], $input.shape[2], $input.shape[3]
            ]);
        }
        assert$4(dy5D.rank === 5, () => `Error in avgPool3dGrad: dy must be rank 5 but got rank ` +
            `${dy5D.rank}.`);
        assert$4(input5D.rank === 5, () => `Error in avgPool3dGrad: input must be rank 5 but got rank ` +
            `${input5D.rank}.`);
        checkPadOnDimRoundingMode('avgPool3dGrad', pad, dimRoundingMode);
        const inputs = { dy: dy5D, input: input5D };
        const attrs = { filterSize, strides, pad, dimRoundingMode };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(AvgPool3DGrad, inputs, attrs);
        if (reshapedTo5D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3], res.shape[4]]);
        }
        return res;
    }
    const avgPool3dGrad = op({ avgPool3dGrad_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const avgPool3DGradConfig$2 = {
        kernelName: AvgPool3D,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const { filterSize, strides, pad, dimRoundingMode } = attrs;
            return {
                x: () => avgPool3dGrad(dy, x, filterSize, strides, pad, dimRoundingMode)
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the backprop of an 2D avg pool.
     *
     * @param dy The dy error, of rank 4 or rank 3 of shape
     *     [batchSize, height, width, channels]. If rank 3, batch of 1 is
     * assumed.
     * @param input The input image, of rank 4 or rank 3 of shape
     *     [batchSize, height, width, channels]. If rank 3, batch of 1 is
     * assumed.
     * @param filterSize The filter size: `[filterHeight, filterWidth]`. If
     *     `filterSize` is a single number, then `filterHeight == filterWidth`.
     * @param strides The strides of the pooling: `[strideHeight, strideWidth]`. If
     *     `strides` is a single number, then `strideHeight == strideWidth`.
     * @param pad The type of padding algorithm used in the forward prop of the op.
     *     'same', 'valid', for more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *         https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     */
    function avgPoolGrad_(dy, input, filterSize, strides, pad) {
        const $dy = convertToTensor(dy, 'dy', 'avgPoolGrad');
        const $input = convertToTensor(input, 'input', 'avgPoolGrad');
        assert$4($input.rank === $dy.rank, () => `Rank of input (${$input.rank}) does not match rank of dy (${$dy.rank})`);
        let input4D = $input;
        let dy4D = $dy;
        let reshapedTo4D = false;
        if ($input.rank === 3) {
            reshapedTo4D = true;
            input4D =
                reshape$3($input, [1, $input.shape[0], $input.shape[1], $input.shape[2]]);
            dy4D = reshape$3($dy, [1, $dy.shape[0], $dy.shape[1], $dy.shape[2]]);
        }
        assert$4(dy4D.rank === 4, () => `Error in avgPoolGrad: dy must be rank 4 but got rank ` +
            `${dy4D.rank}.`);
        assert$4(input4D.rank === 4, () => `Error in avgPoolGrad: input must be rank 4 but got rank ` +
            `${input4D.rank}.`);
        const inputs = { dy: dy4D, input: input4D };
        const attrs = { filterSize, strides, pad };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(AvgPoolGrad, inputs, attrs);
        if (reshapedTo4D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3]]);
        }
        return res;
    }
    const avgPoolGrad$2 = op({ avgPoolGrad_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const avgPoolGradConfig$2 = {
        kernelName: AvgPool,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const { filterSize, strides, pad } = attrs;
            return { x: () => avgPoolGrad$2(dy, x, filterSize, strides, pad) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const batchMatMulGradConfig = {
        kernelName: BatchMatMul,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved, attrs) => {
            const [a, b] = saved;
            const { transposeA, transposeB } = attrs;
            if (!transposeA && !transposeB) {
                return {
                    a: () => matMul$1(dy, b, false, true),
                    b: () => matMul$1(a, dy, true, false)
                };
            }
            else if (!transposeA && transposeB) {
                return {
                    a: () => matMul$1(dy, b, false, false),
                    b: () => matMul$1(dy, a, true, false)
                };
            }
            else if (transposeA && !transposeB) {
                return {
                    a: () => matMul$1(b, dy, false, true),
                    b: () => matMul$1(a, dy, false, false)
                };
            }
            else {
                return {
                    a: () => matMul$1(b, dy, true, true),
                    b: () => matMul$1(dy, a, true, true)
                };
            }
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const batchToSpaceNDGradConfig = {
        kernelName: BatchToSpaceND,
        gradFunc: (dy, saved, attrs) => {
            const { blockShape, crops } = attrs;
            return { x: () => spaceToBatchND$2(dy, blockShape, crops) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const broadcastToGradConfig = {
        kernelName: BroadcastTo,
        gradFunc: (dy, saved, attrs) => {
            const broadCastToAttrs = attrs;
            const inputShape = broadCastToAttrs.inputShape;
            const outputShape = broadCastToAttrs.shape;
            const reps = Array.from(outputShape);
            for (let i = inputShape.length - 1; i >= 0; i--) {
                if (inputShape[i] === outputShape[i]) {
                    reps[i] = 1;
                }
                else if (inputShape[i] !== 1) {
                    throw new Error(`broadcastTo(): [${inputShape}] cannot be broadcast to [${outputShape}].`);
                }
            }
            const axes = [];
            for (let i = 0; i < reps.length; i++) {
                if (reps[i] > 1) {
                    axes.push(i);
                }
            }
            return { x: () => sum$2(dy, axes, true /* keepDims */) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const castGradConfig = {
        kernelName: Cast,
        gradFunc: (dy) => {
            return { x: () => dy.clone() };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ceilGradConfig = {
        kernelName: Ceil,
        gradFunc: (dy) => {
            // TODO(manrajgrover): Return null for gradients when backprop supports it.
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const clipByValueGradConfig = {
        kernelName: ClipByValue,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const { clipValueMin, clipValueMax } = attrs;
            return {
                x: () => where(logicalAnd$2(greaterEqual$2(x, clipValueMin), lessEqual$2(x, clipValueMax)), dy, zerosLike$2(dy)),
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const complexAbsGradConfig = {
        kernelName: ComplexAbs,
        inputsToSave: ['x'],
        gradFunc: absGradConfig.gradFunc,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const concatGradConfig = {
        kernelName: Concat,
        saveAllInputs: true,
        gradFunc: (dy, saved, attrs) => {
            const shapes = saved.map(t => t.shape);
            const { axis } = attrs;
            const $axis = parseAxisParam(axis, saved[0].shape)[0];
            const sizeSplits = shapes.map(s => s[$axis]);
            const derTensors = split$2(dy, sizeSplits, $axis);
            return derTensors.map(t => () => t);
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const conv2DGradConfig = {
        kernelName: Conv2D$1,
        inputsToSave: ['x', 'filter'],
        gradFunc: (dy, saved, attrs) => {
            const [x4D, $filter] = saved;
            const { dilations, strides, pad, dataFormat } = attrs;
            assert$4(tupleValuesAreOne(dilations), () => 'Error in gradient of conv2D: dilation rates greater than 1 ' +
                `are not yet supported in gradients. Got dilations '${dilations}'`);
            return {
                x: () => conv2DBackpropInput$2(x4D.shape, dy, $filter, strides, pad, dataFormat),
                filter: () => conv2DBackpropFilter$2(x4D, dy, $filter.shape, strides, pad, dataFormat)
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const conv2DBackpropInputGradConfig = {
        kernelName: Conv2DBackpropInput,
        inputsToSave: ['dy', 'filter'],
        gradFunc: (ddx, saved, attrs) => {
            const [dy, filter] = saved;
            const { strides, pad, dataFormat, dimRoundingMode } = attrs;
            return {
                dy: () => conv2d$3(ddx, filter, strides, pad, dataFormat, 1 /* dilations */, dimRoundingMode),
                filter: () => conv2DBackpropFilter$2(ddx, dy, filter.shape, strides, pad, dataFormat, dimRoundingMode)
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the derivative of the filter of a 3D convolution.
     *
     * @param x The input tensor, of rank 5 or rank 4 of shape
     *     [batch, depth, height, width, inChannels]. If rank 4, batch of 1 is
     *     assumed.
     * @param dy The dy image, of rank 5 or rank 4, of shape
     *     [batch, depth, height, width, outDepth]. If rank 4, batch of 1 is
     *     assumed.
     * @param filterShape The shape of the filter, length 5,
     *     [filterDepth, filterHeight, filterWidth, inDepth, outDepth].
     * @param strides The strides of the convolution: [strideDepth, strideHeight,
     * strideWidth].
     * @param pad A string from: 'same', 'valid'. The type of padding algorithm
     *     used in the forward prop of the op.
     */
    function conv3DBackpropFilter_(x, dy, filterShape, strides, pad) {
        let x5D = x;
        if (x.rank === 4) {
            x5D = reshape$3(x, [1, x.shape[0], x.shape[1], x.shape[2], x.shape[3]]);
        }
        let dy5D = dy;
        if (dy5D.rank === 4) {
            dy5D = reshape$3(dy, [1, dy.shape[0], dy.shape[1], dy.shape[2], dy.shape[3]]);
        }
        assert$4(x5D.rank === 5, () => `Error in conv3dDerFilter: input must be rank 5, but got shape ` +
            `${x5D.shape}.`);
        assert$4(dy5D.rank === 5, () => `Error in conv3dDerFilter: dy must be rank 5, but got shape ` +
            `${dy5D.shape}.`);
        assert$4(filterShape.length === 5, () => `Error in conv3dDerFilter: filterShape must be length 5, but got ` +
            `${filterShape}.`);
        assert$4(x5D.shape[4] === filterShape[3], () => `Error in conv3dDerFilter: depth of input ${x5D.shape[4]}) must ` +
            `match input depth in filter (${filterShape[3]}.`);
        assert$4(dy5D.shape[4] === filterShape[4], () => `Error in conv3dDerFilter: depth of dy (${dy5D.shape[4]}) must ` +
            `match output depth for filter (${filterShape[4]}).`);
        const inputs = { x: x5D, dy: dy5D };
        const attrs = { strides, pad, filterShape };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        return ENGINE.runKernel(Conv3DBackpropFilterV2, inputs, attrs);
    }
    const conv3DBackpropFilter = op({ conv3DBackpropFilter_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const conv3DGradConfig = {
        kernelName: Conv3D$1,
        inputsToSave: ['x', 'filter'],
        gradFunc: (dy, saved, attrs) => {
            const { dilations, strides, pad } = attrs;
            assert$4(tupleValuesAreOne(dilations), () => 'Error in gradient of conv3D: dilation rates greater than 1 are ' +
                `not yet supported in gradients. Got dilations '${dilations}'`);
            const [x5D, $filter] = saved;
            return {
                x: () => conv3DBackpropInput$1(x5D.shape, dy, $filter, strides, pad),
                filter: () => conv3DBackpropFilter(x5D, dy, $filter.shape, strides, pad)
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const cosGradConfig = {
        kernelName: Cos,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => mul(neg$2(sin$2(cast$3(x, 'float32'))), dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const coshGradConfig = {
        kernelName: Cosh,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => mul(sinh$2(cast$3(x, 'float32')), dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const cumsumGradConfig = {
        kernelName: Cumsum,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const { axis, exclusive, reverse } = attrs;
            return {
                x: () => {
                    const permutation = getAxesPermutation([axis], x.rank);
                    let out = cumsum$2(dy, axis, exclusive, !reverse);
                    if (permutation != null) {
                        out = transpose$2(out, permutation);
                    }
                    return out;
                }
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const depthwiseConv2dNativeGradConfig = {
        kernelName: DepthwiseConv2dNative,
        inputsToSave: ['x', 'filter'],
        gradFunc: (dy, saved, attrs) => {
            const { dilations, strides, pad, dimRoundingMode } = attrs;
            const $dilations = dilations == null ? [1, 1] : dilations;
            assert$4(tupleValuesAreOne($dilations), () => 'Error in gradient of depthwiseConv2dNative: dilation rates ' +
                `greater than 1 are not yet supported. Got dilations ` +
                `'${$dilations}'`);
            const [x, filter] = saved;
            assert$4(x.rank === 4, () => `Error in gradient of depthwiseConv2dNative: input must be ` +
                `rank 4, but got rank ${x.rank}.`);
            assert$4(filter.rank === 4, () => `Error in gradient of depthwiseConv2dNative: filter must be ` +
                `rank 4, but got rank ${filter.rank}.`);
            assert$4(x.shape[3] === filter.shape[2], () => `Error in gradient of depthwiseConv2d: number of input ` +
                `channels (${x.shape[3]}) must match the inChannels dimension ` +
                `in filter ${filter.shape[2]}.`);
            assert$4(eitherStridesOrDilationsAreOne(strides, $dilations), () => 'Error in gradient of depthwiseConv2d: Either strides or ' +
                `dilations must be  1. Got strides ${strides} and dilations ` +
                `'${$dilations}'.`);
            checkPadOnDimRoundingMode('depthwiseConv2d', pad, dimRoundingMode);
            return {
                x: () => depthwiseConv2dNativeBackpropInput$2(x.shape, dy, filter, strides, pad, $dilations, dimRoundingMode),
                filter: () => depthwiseConv2dNativeBackpropFilter$2(x, dy, filter.shape, strides, pad, $dilations, dimRoundingMode),
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const dilation2dGradConfig = {
        kernelName: Dilation2D,
        inputsToSave: ['x', 'filter'],
        gradFunc: (dy, saved, attrs) => {
            const [x, filter] = saved;
            const inputInputs = { x, filter, dy };
            const filterInputs = { x, filter, dy };
            return {
                x: () => ENGINE.runKernel(Dilation2DBackpropInput, inputInputs, attrs),
                filter: () => ENGINE.runKernel(Dilation2DBackpropFilter, filterInputs, attrs)
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const eluGradConfig$2 = {
        kernelName: Elu$1,
        outputsToSave: [true],
        gradFunc: (dy, saved) => {
            const [y] = saved;
            const inputs = { dy, y };
            return { x: () => ENGINE.runKernel(EluGrad, inputs) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const erfGradConfig = {
        kernelName: Erf,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            const a = mul(exp$2(neg$2(square$2(x))), 2 / Math.sqrt(Math.PI));
            return { x: () => mul(dy, a) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const expGradConfig = {
        kernelName: Exp,
        outputsToSave: [true],
        gradFunc: (dy, saved) => {
            const [y] = saved;
            return { x: () => mul(dy, y) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const expandDimsGradConfig = {
        kernelName: ExpandDims,
        inputsToSave: ['input'],
        gradFunc: (dy, saved) => {
            const [input] = saved;
            return { input: () => reshape$3(dy, input.shape) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const expm1GradConfig = {
        kernelName: Expm1,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => mul(dy, exp$2(x)) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const floorGradConfig = {
        kernelName: Floor,
        gradFunc: (dy) => {
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const floorDivGradConfig = {
        kernelName: FloorDiv,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const outShape = assertAndGetBroadcastShape(a.shape, b.shape);
            const derA = () => {
                const res = div$1(dy, cast$3(b, 'float32'));
                const reduceAxes = getReductionAxes(a.shape, outShape);
                if (reduceAxes.length > 0) {
                    return reshape$3(sum$2(res, reduceAxes), a.shape);
                }
                return res;
            };
            const derB = () => {
                let res = mul(dy, cast$3(a, 'float32'));
                const reduceAxes = getReductionAxes(b.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = reshape$3(sum$2(res, reduceAxes), b.shape);
                }
                const tmp = square$2(b);
                return neg$2(div$1(res, cast$3(tmp, 'float32')));
            };
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const fusedBatchNormGradConfig = {
        kernelName: FusedBatchNorm,
        inputsToSave: ['x', 'mean', 'variance', 'scale'],
        gradFunc: (dy, saved, attrs) => {
            const { varianceEpsilon } = attrs;
            const [x, mean, variance, scale] = saved;
            const scaleValue = scale == null ? scalar(1) : scale;
            const reductionAxes = getReductionAxes(mean.shape, x.shape);
            const tileShape = [];
            if (mean.rank === 1) {
                for (let i = 0; i < x.shape.length - 1; ++i) {
                    tileShape.push(x.shape[i]);
                }
                tileShape.push(1);
            }
            const xMinusMean = sub$2(x, mean);
            const dyTimesScaleValue = mul(dy, scaleValue);
            const oneOverSqrtVariance = rsqrt$2(add$2(variance, scalar(varianceEpsilon)));
            const minusHalfRCube = mul(mul(mul(oneOverSqrtVariance, oneOverSqrtVariance), oneOverSqrtVariance), scalar(-0.5));
            const derX = () => {
                if (mean.rank === 1) {
                    return reshape$3(mul(mul(dy, tile$3(reshape$3(oneOverSqrtVariance, [1, 1, 1, mean.shape[0]]), tileShape)), scaleValue), x.shape);
                }
                else {
                    return reshape$3(mul(mul(dy, oneOverSqrtVariance), scaleValue), x.shape);
                }
            };
            const derMean = () => {
                let meanDer = mul(mul(oneOverSqrtVariance, scalar(-1)), dyTimesScaleValue);
                if (mean.rank === 1) {
                    meanDer = sum$2(meanDer, reductionAxes);
                }
                return reshape$3(meanDer, mean.shape);
            };
            const derVariance = () => {
                let varianceDer = mul(mul(minusHalfRCube, xMinusMean), dyTimesScaleValue);
                if (mean.rank === 1) {
                    varianceDer = sum$2(varianceDer, reductionAxes);
                }
                return reshape$3(varianceDer, mean.shape);
            };
            const derScale = () => {
                const xMinusMean2TimesRsqrt = mul(xMinusMean, oneOverSqrtVariance);
                let scaleDer = mul(dy, xMinusMean2TimesRsqrt);
                if (mean.rank === 1) {
                    scaleDer = sum$2(scaleDer, reductionAxes);
                }
                return reshape$3(scaleDer, mean.shape);
            };
            const derOffset = () => {
                let offsetDer = dy;
                if (mean.rank === 1) {
                    offsetDer = sum$2(offsetDer, reductionAxes);
                }
                return reshape$3(offsetDer, mean.shape);
            };
            return {
                x: derX,
                mean: derMean,
                variance: derVariance,
                scale: derScale,
                offset: derOffset
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const gatherGradConfig = {
        kernelName: GatherV2,
        inputsToSave: ['x', 'indices'],
        gradFunc: (dy, saved, attrs) => {
            const [x, indices] = saved;
            const { axis } = attrs;
            const parsedAxis = parseAxisParam(axis, x.shape)[0];
            const derX = () => {
                const paramsShape = x.shape;
                const indicesSize = indices.size;
                const outerShape = paramsShape.slice(0, parsedAxis);
                const outerDims = outerShape.length;
                const innerShape = paramsShape.slice(axis, paramsShape.length).slice(1);
                const innerDims = innerShape.length;
                const outerAxesIndices = arrayRange(0, outerDims);
                const innerAxesIndices = arrayRange(outerDims + 1, outerDims + 1 + innerDims);
                const valuesShape = arrayConcat([outerShape, [indicesSize], innerShape]);
                const values = reshape$3(dy, valuesShape);
                const reshapedIndices = reshape$3(indices, [indicesSize]);
                const transposeDims = arrayConcat([[outerDims], outerAxesIndices, innerAxesIndices]);
                const valuesTranspose = transpose$2(values, transposeDims);
                let paramsGrad = unsortedSegmentSum$2(valuesTranspose, reshapedIndices, x.shape[parsedAxis]);
                const invertTransposeDims = getUndoAxesPermutation(transposeDims);
                paramsGrad = transpose$2(paramsGrad, invertTransposeDims);
                return paramsGrad;
            };
            return { x: derX, indices: () => indices };
        }
    };
    function arrayRange(start, stop) {
        const result = [];
        for (let i = start; i < stop; ++i) {
            result.push(i);
        }
        return result;
    }
    function arrayConcat(arrays) {
        const result = [];
        for (let i = 0; i < arrays.length; ++i) {
            for (let j = 0; j < arrays[i].length; ++j) {
                result.push(arrays[i][j]);
            }
        }
        return result;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const greaterEqualGradConfig = {
        kernelName: GreaterEqual,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            return { a: () => zerosLike$2(a), b: () => zerosLike$2(b) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const identityGradConfig = {
        kernelName: Identity$1,
        gradFunc: (dy) => {
            return { x: () => cast$3(dy, 'float32') };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const isFiniteGradConfig = {
        kernelName: IsFinite,
        gradFunc: (dy) => {
            // TODO(nsthorat): Let gradients be null for cases where we want to stop
            // backpropgation.
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const isInfGradConfig = {
        kernelName: IsInf,
        gradFunc: (dy) => {
            // TODO(nsthorat): Let gradients be null for cases where we want to stop
            // backpropgation.
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const isNanGradConfig = {
        kernelName: IsNan,
        gradFunc: (dy) => {
            // TODO(nsthorat): Let gradients be null for cases where we want to stop
            // backpropgation.
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const leakyReluGradConfig = {
        kernelName: LeakyRelu,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const { alpha } = attrs;
            const mask = greater$3(x, 0);
            // Returns `gradients * (features > 0) + alpha * gradients * (features <=
            // 0)`.
            return { x: () => where(mask, dy, mul(dy, alpha)) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const log1pGradConfig = {
        kernelName: Log1p,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => div$1(dy, add$2(x, 1)) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const logGradConfig = {
        kernelName: Log,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => div$1(dy, cast$3(x, 'float32')) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const logSoftmaxGradConfig = {
        kernelName: LogSoftmax$1,
        inputsToSave: [],
        outputsToSave: [true],
        gradFunc: (dy, saved, attrs) => {
            const [value] = saved;
            const { axis } = attrs;
            return {
                logits: () => {
                    const keepDims = true;
                    const softmax = exp$2(value);
                    return sub$2(dy, mul(sum$2(dy, axis, keepDims), softmax));
                }
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function localResponseNormalizationBackprop_(x, y, dy, depthRadius = 5, bias = 1, alpha = 1, beta = 0.5) {
        const inputs = { x, y, dy };
        const attrs = { depthRadius, bias, alpha, beta };
        return ENGINE.runKernel(LRNGrad, inputs, attrs);
    }
    const localResponseNormalizationBackprop = op({ localResponseNormalizationBackprop_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const lrnGradConfig = {
        kernelName: LRN,
        inputsToSave: ['x'],
        outputsToSave: [true],
        gradFunc: (dy, saved, attrs) => {
            const [x, y] = saved;
            const { depthRadius, bias, alpha, beta } = attrs;
            return {
                x: () => localResponseNormalizationBackprop(x, y, dy, depthRadius, bias, alpha, beta)
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Gradient helper function for the min and max operations.
     */
    function gradForMinAndMax(dy, y, xOrig, origAxes) {
        if (y.rank < xOrig.rank) {
            y = reshape$3(y, expandShapeToKeepDim(y.shape, origAxes));
        }
        if (dy.rank < xOrig.rank) {
            dy = reshape$3(dy, expandShapeToKeepDim(dy.shape, origAxes));
        }
        return {
            x: () => {
                const dx = mul(dy, cast$3(equal$3(xOrig, y), dy.dtype));
                return dx;
            }
        };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const maxGradConfig = {
        kernelName: Max,
        inputsToSave: ['x'],
        outputsToSave: [true],
        gradFunc: (dy, saved, attrs) => {
            const maxAttrs = attrs;
            const { reductionIndices } = maxAttrs;
            const x = saved[0];
            const y = saved[1];
            const origAxes = parseAxisParam(reductionIndices, x.shape);
            const maxGrad = gradForMinAndMax(dy, y, x, origAxes);
            return {
                x: () => {
                    return maxGrad['x']();
                }
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const maximumGradConfig = {
        kernelName: Maximum$1,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const derA = () => mul(dy, cast$3(greaterEqual$2(a, b), 'float32'));
            const derB = () => mul(dy, cast$3(less$3(a, b), 'float32'));
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the backprop of a 3d max pool.
     *
     * @param dy The dy error, of rank 5 of shape
     *     [batchSize, depth, height, width, channels].
     * assumed.
     * @param input The original input image, of rank 5 or rank 4 of shape
     *     [batchSize, depth, height, width, channels].
     * @param output The original output image, of rank 5 of shape
     *     [batchSize, outDepth, outHeight, outWidth, channels].
     * @param filterSize The filter size:
     *     `[filterDepth, filterHeight, filterWidth]`.
     *     `filterSize` is a single number,
     *     then `filterDepth == filterHeight == filterWidth`.
     * @param strides The strides of the pooling:
     *     `[strideDepth, strideHeight, strideWidth]`. If
     *     `strides` is a single number, then `strideHeight == strideWidth`.
     * @param pad A string from: 'same', 'valid'. The type of padding algorithm
     *     used in the forward prop of the op.
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     */
    function maxPool3dGrad_(dy, input, output, filterSize, strides, pad, dimRoundingMode) {
        const $dy = convertToTensor(dy, 'dy', 'maxPool3dGrad');
        const $input = convertToTensor(input, 'input', 'maxPool3dGrad');
        const $output = convertToTensor(output, 'output', 'maxPool3dGrad');
        let dy5D = $dy;
        let input5D = $input;
        let output5D = $output;
        let reshapedTo5D = false;
        if ($input.rank === 4) {
            reshapedTo5D = true;
            dy5D = reshape$3($dy, [1, $dy.shape[0], $dy.shape[1], $dy.shape[2], $dy.shape[3]]);
            input5D = reshape$3($input, [
                1, $input.shape[0], $input.shape[1], $input.shape[2], $input.shape[3]
            ]);
            output5D = reshape$3($output, [
                1, $output.shape[0], $output.shape[1], $output.shape[2], $output.shape[3]
            ]);
        }
        assert$4(dy5D.rank === 5, () => `Error in maxPool3dGrad: dy must be rank 5 but got rank ` +
            `${dy5D.rank}.`);
        assert$4(input5D.rank === 5, () => `Error in maxPool3dGrad: input must be rank 5 but got rank ` +
            `${input5D.rank}.`);
        assert$4(output5D.rank === 5, () => `Error in maxPool3dGrad: output must be rank 5 but got rank ` +
            `${output5D.rank}.`);
        checkPadOnDimRoundingMode('maxPool3dGrad', pad, dimRoundingMode);
        const inputs = { dy: dy5D, input: input5D, output: output5D };
        const attrs = { filterSize, strides, pad, dimRoundingMode };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        const res = ENGINE.runKernel(MaxPool3DGrad, inputs, attrs);
        if (reshapedTo5D) {
            return reshape$3(res, [res.shape[1], res.shape[2], res.shape[3], res.shape[4]]);
        }
        return res;
    }
    const maxPool3dGrad = op({ maxPool3dGrad_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const maxPool3DGradConfig$2 = {
        kernelName: MaxPool3D,
        inputsToSave: ['x'],
        outputsToSave: [true],
        gradFunc: (dy, saved, attrs) => {
            const [x, y] = saved;
            const { filterSize, strides, pad, dimRoundingMode } = attrs;
            return {
                x: () => maxPool3dGrad(dy, x, y, filterSize, strides, pad, dimRoundingMode)
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Computes the backprop of a 2D max pool.
     *
     * @param dy The dy error, of rank 4 or rank 3 of shape
     *     [batchSize, height, width, channels]. If rank 3, batch of 1 is
     * assumed.
     * @param input The original input image, of rank 4, of shape
     *     [batchSize, height, width, channels].
     * @param output The original output image, of rank 4, of shape
     *     [batchSize, outHeight, outWidth, channels].
     * @param filterSize The filter size: `[filterHeight, filterWidth]`. If
     *     `filterSize` is a single number, then `filterHeight == filterWidth`.
     * @param strides The strides of the pooling: `[strideHeight, strideWidth]`. If
     *     `strides` is a single number, then `strideHeight == strideWidth`.
     * @param pad The type of padding algorithm used in the forward prop of the op.
     *     'same', 'valid', for more info, see this guide:
     *     [https://www.tensorflow.org/api_docs/python/tf/nn/convolution](
     *          https://www.tensorflow.org/api_docs/python/tf/nn/convolution)
     * @param dimRoundingMode A string from: 'ceil', 'round', 'floor'. If none is
     *     provided, it will default to truncate.
     */
    function maxPoolGrad_(dy, input, output, filterSize, strides, pad, dimRoundingMode) {
        const $dy = convertToTensor(dy, 'dy', 'maxPoolGrad');
        const $input = convertToTensor(input, 'input', 'maxPoolGrad');
        const $output = convertToTensor(output, 'output', 'maxPoolGrad');
        assert$4($input.rank === $dy.rank, () => `Rank of input (${$input.rank}) does not match rank of dy ` +
            `(${$dy.rank})`);
        assert$4($dy.rank === 4, () => `Error in maxPoolGrad: dy must be rank 4 but got rank ` +
            `${$dy.rank}.`);
        assert$4($input.rank === 4, () => `Error in maxPoolGrad: input must be rank 4 but got rank ` +
            `${$input.rank}.`);
        checkPadOnDimRoundingMode('maxPoolGrad', pad, dimRoundingMode);
        const inputs = { dy: $dy, input: $input, output: $output };
        const attrs = { filterSize, strides, pad, dimRoundingMode };
        // tslint:disable-next-line: no-unnecessary-type-assertion
        return ENGINE.runKernel(MaxPoolGrad, inputs, attrs);
    }
    const maxPoolGrad$2 = op({ maxPoolGrad_ });

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const maxPoolGradConfig$2 = {
        kernelName: MaxPool,
        inputsToSave: ['x'],
        outputsToSave: [true],
        gradFunc: (dy, saved, attrs) => {
            const [x, y] = saved;
            const { filterSize, strides, pad } = attrs;
            return {
                x: () => maxPoolGrad$2(dy, x, y, filterSize, strides, pad)
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const meanGradConfig = {
        kernelName: Mean,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const { axis } = attrs;
            const axes = parseAxisParam(axis, x.shape);
            const shapes = computeOutAndReduceShapes(x.shape, axes);
            const reduceShape = shapes[1];
            const reduceSize = sizeFromShape(reduceShape);
            const derX = () => {
                const expandedDyShape = x.shape.slice();
                axes.forEach(axis => {
                    expandedDyShape[axis] = 1;
                });
                const expandedDy = reshape$3(dy, expandedDyShape);
                const res = div$1(mul(expandedDy, ones$1(x.shape, 'float32')), reduceSize);
                return res;
            };
            return { x: derX };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const minGradConfig = {
        kernelName: Min,
        inputsToSave: ['x'],
        outputsToSave: [true],
        gradFunc: (dy, saved, attrs) => {
            const minAttrs = attrs;
            const { axis } = minAttrs;
            const [x, y] = saved;
            const origAxes = parseAxisParam(axis, x.shape);
            const minGrad = gradForMinAndMax(dy, y, x, origAxes);
            return {
                x: () => {
                    return minGrad['x']();
                }
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const minimumGradConfig = {
        kernelName: Minimum$1,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const derA = () => mul(dy, cast$3(lessEqual$2(a, b), 'float32'));
            const derB = () => mul(dy, cast$3(greater$3(a, b), 'float32'));
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const mirrorPadGradConfig = {
        kernelName: MirrorPad,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            // Pad introduces values around the original tensor, so the gradient
            // slices the original shape out of the gradient.
            const x = saved[0];
            const { paddings } = attrs;
            const begin = paddings.map(p => p[0]);
            return { x: () => slice$2(dy, begin, x.shape) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const modGradConfig = {
        kernelName: Mod,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const outShape = assertAndGetBroadcastShape(a.shape, b.shape);
            const derA = () => {
                const reduceAxes = getReductionAxes(a.shape, outShape);
                if (reduceAxes.length > 0) {
                    return reshape$3(sum$2(dy, reduceAxes), a.shape);
                }
                return dy;
            };
            const derB = () => {
                const res = mul(dy, neg$2(floor$3(div$1(a, b))));
                const reduceAxes = getReductionAxes(b.shape, outShape);
                if (reduceAxes.length > 0) {
                    return reshape$3(sum$2(res, reduceAxes), b.shape);
                }
                return res;
            };
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const multiplyGradConfig = {
        kernelName: Multiply$1,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const outShape = assertAndGetBroadcastShape(a.shape, b.shape);
            const derA = () => {
                const res = mul(dy, cast$3(b, 'float32'));
                const reduceAxes = getReductionAxes(a.shape, outShape);
                if (reduceAxes.length > 0) {
                    return reshape$3(sum$2(res, reduceAxes), a.shape);
                }
                return res;
            };
            const derB = () => {
                const res = mul(dy, cast$3(a, 'float32'));
                const reduceAxes = getReductionAxes(b.shape, outShape);
                if (reduceAxes.length > 0) {
                    return reshape$3(sum$2(res, reduceAxes), b.shape);
                }
                return res;
            };
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const negGradConfig = {
        kernelName: Neg,
        gradFunc: (dy) => {
            return { x: () => neg$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const oneHotGradConfig = {
        kernelName: OneHot,
        inputsToSave: ['indices'],
        gradFunc: (dy, saved) => {
            const indices = saved[0];
            return { indices: () => zeros$2(indices.shape, 'float32') };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const onesLikeGradConfig = {
        kernelName: OnesLike,
        gradFunc: (dy) => {
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const packGradConfig = {
        kernelName: Pack,
        saveAllInputs: true,
        gradFunc: (dy, saved, attrs) => {
            const { axis } = attrs;
            const derTensors = unstack(dy, axis);
            return derTensors.map(t => () => t);
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const padV2GradConfig = {
        kernelName: PadV2,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            // Pad introduces values around the original tensor, so the gradient
            // slices the original shape out of the gradient.
            const x = saved[0];
            const { paddings } = attrs;
            const begin = paddings.map(p => p[0]);
            return { x: () => slice$2(dy, begin, x.shape) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const powGradConfig = {
        kernelName: Pow,
        inputsToSave: ['a', 'b'],
        outputsToSave: [true],
        gradFunc: (dy, saved) => {
            const [a, b, y] = saved;
            const base = a;
            const exp = b;
            const outShape = assertAndGetBroadcastShape(base.shape, exp.shape);
            const derBase = () => {
                const expFloat = cast$3(exp, 'float32');
                let res = mul(dy, mul(expFloat, pow$2(base, sub$2(expFloat, scalar(1)))));
                const reduceAxes = getReductionAxes(base.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = sum$2(res, reduceAxes);
                }
                return reshape$3(res, base.shape);
            };
            const derExp = () => {
                const condition = greater$3(base, 0);
                const logBase = where(condition, log$3(base), zerosLike$2(base));
                let res = mul(dy, mul(y, logBase));
                const reduceAxes = getReductionAxes(exp.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = sum$2(res, reduceAxes);
                }
                return reshape$3(res, exp.shape);
            };
            return { a: derBase, b: derExp };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const preluGradConfig = {
        kernelName: Prelu,
        inputsToSave: ['x', 'alpha'],
        gradFunc: (dy, saved) => {
            const [x, alpha] = saved;
            const mask = greater$3(x, 0);
            return {
                x: () => where(mask, dy, mul(dy, alpha)),
                alpha: () => {
                    let res = where(mask, zerosLike$2(dy), mul(dy, x));
                    const reduceAxes = getReductionAxes(alpha.shape, dy.shape);
                    if (reduceAxes.length > 0) {
                        res = sum$2(res, reduceAxes);
                    }
                    return reshape$3(res, alpha.shape);
                }
            };
        }
    };

    /**
     * @license
     * Copyright 2022 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Gradient for product operation on a single axis.
    function prodGradFn_(x, dy, axis) {
        // The gradient tensor (dy) has a set of axes removed, so we create re-shaped
        // versions (of size 1) for the removed axis; this supports broadcasting over
        // those dimensions.
        const expandedYShape = x.shape.slice();
        expandedYShape[axis] = 1;
        // The actual gradient computation.
        const expandedDy = reshape$3(dy, expandedYShape);
        const xCumProd = cumprod$2(x, axis, true, false);
        const xCumRevProd = cumprod$2(x, axis, true, true);
        const dx = mul(xCumProd, xCumRevProd);
        return mul(expandedDy, dx);
    }
    // Support gradients when the product is done on many axes at once.
    // This done py pushing all the axes on which the product is applied into a
    // single axis.
    function prodsGradFn_(x, dy, axis) {
        // Move all axes for doing prod over to the end of the tensor.
        const xRank = x.shape.length;
        const finalProdAxis = xRank - axis.length;
        const xPermutation = getAxesPermutation(axis, xRank);
        let permutedX = x;
        if (xPermutation != null) {
            permutedX = transpose$2(x, xPermutation);
        }
        // Reshape all the prod dimensions into a single one, and do compute prod
        // gradients on that.
        const newShape = permutedX.shape.slice();
        const removedShape = newShape.splice(xRank - axis.length, axis.length);
        const endPartShape = removedShape.reduce((p, c) => p * c, 1);
        newShape.push(endPartShape);
        const reshapedPermutedX = permutedX.reshape(newShape);
        let prodGrad = prodGradFn_(reshapedPermutedX, dy, finalProdAxis);
        // Undo the re-shaping now we have the dx vector, and permute back to
        // original axes order.
        prodGrad = prodGrad.reshape(permutedX.shape);
        if (xPermutation != null) {
            const undoPermutation = getUndoAxesPermutation(xPermutation);
            prodGrad = transpose$2(prodGrad, undoPermutation);
        }
        return prodGrad;
    }
    // Running example:
    // [
    //   [
    //     [3.0, 4.0],
    //     [5.0, 6.0],
    //     [7.0, 8.0]
    //   ],
    //   [
    //     [3.0, 5.0],
    //     [0.0, 6.0],
    //     [5.0, 6.0]
    //   ]
    // ]
    //
    const prodGradConfig = {
        kernelName: Prod,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const { axis } = attrs;
            let axisArr = [];
            if (axis === undefined || axis === null) {
                axisArr = x.shape.map((_, i) => i);
            }
            else if (typeof axis === 'number') {
                axisArr = [axis];
            }
            else {
                axisArr = axis;
            }
            return { x: () => prodsGradFn_(x, dy, axisArr) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const divGradConfig = {
        kernelName: RealDiv,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const outShape = assertAndGetBroadcastShape(a.shape, b.shape);
            const derA = () => {
                const res = div$1(dy, cast$3(b, 'float32'));
                const reduceAxes = getReductionAxes(a.shape, outShape);
                if (reduceAxes.length > 0) {
                    return reshape$3(sum$2(res, reduceAxes), a.shape);
                }
                return res;
            };
            const derB = () => {
                let res = mul(dy, cast$3(a, 'float32'));
                const reduceAxes = getReductionAxes(b.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = reshape$3(sum$2(res, reduceAxes), b.shape);
                }
                const tmp = square$2(b);
                return neg$2(div$1(res, cast$3(tmp, 'float32')));
            };
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const reciprocalGradConfig = {
        kernelName: Reciprocal,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => div$1(dy, neg$2(square$2(x))) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const relu6GradConfig = {
        kernelName: Relu6$1,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            const mask = mul(lessEqual$2(x, 6), step$2(x));
            return { x: () => mul(dy, cast$3(mask, 'float32')) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const reluGradConfig = {
        kernelName: Relu$1,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => mul(dy, cast$3(step$2(x), 'float32')) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const reshapeGradConfig = {
        kernelName: Reshape$1,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => reshape$3(dy, x.shape) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const resizeBilinearGradConfig$2 = {
        kernelName: ResizeBilinear,
        inputsToSave: ['images'],
        gradFunc: (dy, saved, attrs) => {
            const [images] = saved;
            const inputs = { dy, images };
            const imagesDer = () => 
            // tslint:disable-next-line: no-unnecessary-type-assertion
            ENGINE.runKernel(ResizeBilinearGrad, inputs, attrs);
            return { images: imagesDer };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const resizeNearestNeighborGradConfig$2 = {
        kernelName: ResizeNearestNeighbor,
        inputsToSave: ['images'],
        gradFunc: (dy, saved, attrs) => {
            const [images] = saved;
            const inputs = { dy, images };
            const imagesDer = () => 
            // tslint:disable-next-line: no-unnecessary-type-assertion
            ENGINE.runKernel(ResizeNearestNeighborGrad, inputs, attrs);
            return { images: imagesDer };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const reverseGradConfig = {
        kernelName: Reverse,
        gradFunc: (dy, saved, attrs) => {
            const { dims } = attrs;
            const axes = parseAxisParam(dims, dy.shape);
            return { x: () => reverse$2(dy, axes) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const roundGradConfig = {
        kernelName: Round,
        gradFunc: (dy) => {
            // TODO(nsthorat): Let gradients be null for cases where we want to stop
            // backpropgation.
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const rsqrtGradConfig = {
        kernelName: Rsqrt,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => neg$2(div$1(dy, mul(pow$2(x, 1.5), 2))) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const selectGradConfig = {
        kernelName: Select,
        inputsToSave: ['condition'],
        gradFunc: (dy, saved) => {
            const [condition] = saved;
            return {
                // TODO(julianoks): Return null for condition gradient
                // when backprop supports it.
                condition: () => cast$3(zerosLike$2(condition), 'float32'),
                t: () => mul(dy, cast$3(condition, dy.dtype)),
                e: () => mul(dy, cast$3(logicalNot$2(condition), dy.dtype))
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const seluGradConfig = {
        kernelName: Selu$1,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return {
                x: () => {
                    const mask = greater$3(x, scalar(0));
                    const scaleAlpha = scalar(SELU_SCALEALPHA);
                    const scale = scalar(SELU_SCALE);
                    const greaterThanZeroDer = mul(dy, scale);
                    const lessEqualZeroDer = mul(mul(dy, scaleAlpha), exp$2(cast$3(x, 'float32')));
                    return where(mask, greaterThanZeroDer, lessEqualZeroDer);
                }
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sigmoidGradConfig = {
        kernelName: Sigmoid$1,
        outputsToSave: [true],
        gradFunc: (dy, saved) => {
            const [y] = saved;
            return { x: () => mul(dy, mul(y, sub$2(scalar(1), y))) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const signGradConfig = {
        kernelName: Sign,
        gradFunc: (dy) => {
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sinGradConfig = {
        kernelName: Sin,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => mul(cos$2(cast$3(x, 'float32')), dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sinhGradConfig = {
        kernelName: Sinh,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => mul(cosh$2(cast$3(x, 'float32')), dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sliceGradConfig = {
        kernelName: Slice,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const { begin, size } = attrs;
            const inputShape = x.shape;
            const [begin_, size_] = parseSliceParams(x, begin, size);
            // Create an Nx2 padding where the first column represents how many
            // zeros are prepended (at start) for each dimension, and the second
            // column indicates how many zeros are appended (at end).
            // The number of zeros to append is the shape of the input
            // elementwise-subtracted by both the begin vector and sizes vector.
            const paddings = [];
            for (let i = 0; i < dy.rank; i++) {
                paddings.push([begin_[i], inputShape[i] - begin_[i] - size_[i]]);
            }
            return { x: () => pad$1(dy, paddings) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const softmaxGradConfig = {
        kernelName: Softmax$2,
        outputsToSave: [true],
        gradFunc: (dy, saved, attrs) => {
            const [y] = saved;
            const { dim } = attrs;
            const keepDims = true;
            const dyTimesY = mul(dy, y);
            return {
                logits: () => sub$2(dyTimesY, mul(sum$2(dyTimesY, [dim], keepDims), y))
            };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const softplusGradConfig = {
        kernelName: Softplus$1,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => mul(dy, sigmoid$2(x)) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const spaceToBatchNDGradConfig = {
        kernelName: SpaceToBatchND,
        gradFunc: (dy, saved, attrs) => {
            const { blockShape, paddings } = attrs;
            return { x: () => batchToSpaceND$2(dy, blockShape, paddings) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const splitVGradConfig = {
        kernelName: SplitV,
        gradFunc: (dy, saved, attrs) => {
            const { axis } = attrs;
            return { x: () => concat$2(dy, axis) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sqrtGradConfig = {
        kernelName: Sqrt,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => div$1(dy, mul(sqrt$2(cast$3(x, 'float32')), 2)) };
        }
    };

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const squareGradConfig = {
        kernelName: Square,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => mul(dy, mul(cast$3(x, 'float32'), 2)) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const squaredDifferenceGradConfig = {
        kernelName: SquaredDifference,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const two = scalar(2);
            const derA = () => mul(dy, mul(two, sub$2(a, b)));
            const derB = () => mul(dy, mul(two, sub$2(b, a)));
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const stepGradConfig = {
        kernelName: Step,
        gradFunc: (dy) => {
            // TODO(manrajgrover): Return null for gradients when backprop supports
            // it.
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const subGradConfig = {
        kernelName: Sub,
        inputsToSave: ['a', 'b'],
        gradFunc: (dy, saved) => {
            const [a, b] = saved;
            const outShape = assertAndGetBroadcastShape(a.shape, b.shape);
            const derA = () => {
                let res = dy;
                const reduceAxes = getReductionAxes(a.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = sum$2(res, reduceAxes);
                }
                return reshape$3(res, a.shape);
            };
            const derB = () => {
                let res = dy;
                const reduceAxes = getReductionAxes(b.shape, outShape);
                if (reduceAxes.length > 0) {
                    res = sum$2(res, reduceAxes);
                }
                return reshape$3(neg$2(res), b.shape);
            };
            return { a: derA, b: derB };
        }
    };

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sumGradConfig = {
        kernelName: Sum,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const expandedDyShape = x.shape.slice();
            const { axis } = attrs;
            const axes = parseAxisParam(axis, x.shape);
            axes.forEach(axis => {
                expandedDyShape[axis] = 1;
            });
            const expandedDy = reshape$3(dy, expandedDyShape);
            const derX = mul(expandedDy, ones$1(x.shape, 'float32'));
            return { x: () => derX };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const tanGradConfig = {
        kernelName: Tan,
        inputsToSave: ['x'],
        gradFunc: (dy, saved) => {
            const [x] = saved;
            return { x: () => div$1(dy, square$2(cos$2(x))) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const tanhGradConfig = {
        kernelName: Tanh$1,
        outputsToSave: [true],
        gradFunc: (dy, saved) => {
            const [y] = saved;
            return { x: () => mul(sub$2(scalar(1), square$2(y)), dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const tileGradConfig = {
        kernelName: Tile,
        inputsToSave: ['x'],
        gradFunc: (dy, saved, attrs) => {
            const [x] = saved;
            const { reps } = attrs;
            const derX = () => {
                let xGrad = zerosLike$2(x);
                // TODO(cais): Maybe reduce memory footprint by avoiding repeated
                // slicing.
                if (x.rank === 1) {
                    for (let i = 0; i < reps[0]; ++i) {
                        xGrad = add$2(xGrad, slice$2(dy, [i * x.shape[0]], [x.shape[0]]));
                    }
                }
                else if (x.rank === 2) {
                    for (let i = 0; i < reps[0]; ++i) {
                        for (let j = 0; j < reps[1]; ++j) {
                            xGrad = add$2(xGrad, slice$2(dy, [i * x.shape[0], j * x.shape[1]], [
                                x.shape[0], x.shape[1]
                            ]));
                        }
                    }
                }
                else if (x.rank === 3) {
                    for (let i = 0; i < reps[0]; ++i) {
                        for (let j = 0; j < reps[1]; ++j) {
                            for (let k = 0; k < reps[2]; ++k) {
                                xGrad =
                                    add$2(xGrad, slice$2(dy, [i * x.shape[0], j * x.shape[1], k * x.shape[2]], [x.shape[0], x.shape[1], x.shape[2]]));
                            }
                        }
                    }
                }
                else if (x.rank === 4) {
                    for (let i = 0; i < reps[0]; ++i) {
                        for (let j = 0; j < reps[1]; ++j) {
                            for (let k = 0; k < reps[2]; ++k) {
                                for (let l = 0; l < reps[3]; ++l) {
                                    xGrad =
                                        add$2(xGrad, slice$2(dy, [
                                            i * x.shape[0], j * x.shape[1], k * x.shape[2],
                                            l * x.shape[3]
                                        ], [x.shape[0], x.shape[1], x.shape[2], x.shape[3]]));
                                }
                            }
                        }
                    }
                }
                else {
                    throw new Error(`Gradient for tile operation is not implemented for rank-` +
                        `${x.rank} tensors yet.`);
                }
                return xGrad;
            };
            return { x: derX };
        },
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const transposeGradConfig = {
        kernelName: Transpose,
        gradFunc: (dy, saved, attrs) => {
            const transposeAttrs = attrs;
            const { perm } = transposeAttrs;
            const undoPerm = getUndoAxesPermutation(perm);
            return { x: () => transpose$2(dy, undoPerm) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const unpackGradConfig = {
        kernelName: Unpack,
        gradFunc: (dy, saved, attrs) => {
            const unpackAttrs = attrs;
            const { axis } = unpackAttrs;
            return { value: () => stack(dy, axis) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const unsortedSegmentSumGradConfig = {
        kernelName: UnsortedSegmentSum,
        inputsToSave: ['segmentIds'],
        gradFunc: (dy, saved) => {
            const [segmentIds] = saved;
            const derX = () => {
                return gatherDropNegatives(dy, segmentIds);
            };
            return { x: derX };
        }
    };
    function gatherDropNegatives(x, indices) {
        // Helper function for unsorted segment ops. Gathers params for
        // positive segment ids and gathers 0 for inputs with negative segment id.
        // Mirrors _GatherDropNegatives from tensorflow/python/ops/math_grad.py
        const zeroClippedIndices = maximum$3(indices, zerosLike$2(indices));
        const gathered = gather$1(x, zeroClippedIndices);
        let isPositive = greaterEqual$2(indices, scalar(0, 'int32'));
        const numIters = gathered.rank - isPositive.rank;
        for (let i = 0; i < numIters; ++i) {
            isPositive = expandDims$3(isPositive, i + 1);
        }
        isPositive = logicalAnd$2(isPositive, ones$1(gathered.shape, 'bool'));
        const zeroSlice = zerosLike$2(gathered);
        return where(isPositive, gathered, zeroSlice);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const zerosLikeGradConfig = {
        kernelName: ZerosLike,
        gradFunc: (dy) => {
            return { x: () => zerosLike$2(dy) };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Export all kernel configs here so that the package can auto register them
    const gradConfigs = [
        absGradConfig,
        acosGradConfig,
        acoshGradConfig,
        addGradConfig,
        addNGradConfig,
        argMaxGradConfig,
        argMinGradConfig,
        asinGradConfig,
        asinhGradConfig,
        atan2GradConfig,
        atanGradConfig,
        atanhGradConfig,
        avgPool3DGradConfig$2,
        avgPoolGradConfig$2,
        batchMatMulGradConfig,
        batchToSpaceNDGradConfig,
        broadcastToGradConfig,
        castGradConfig,
        ceilGradConfig,
        clipByValueGradConfig,
        complexAbsGradConfig,
        concatGradConfig,
        conv2DBackpropInputGradConfig,
        conv2DGradConfig,
        conv3DGradConfig,
        cosGradConfig,
        coshGradConfig,
        cumsumGradConfig,
        depthwiseConv2dNativeGradConfig,
        dilation2dGradConfig,
        divGradConfig,
        eluGradConfig$2,
        erfGradConfig,
        expGradConfig,
        expandDimsGradConfig,
        expm1GradConfig,
        floorDivGradConfig,
        floorGradConfig,
        fusedBatchNormGradConfig,
        gatherGradConfig,
        greaterEqualGradConfig,
        identityGradConfig,
        isFiniteGradConfig,
        isInfGradConfig,
        isNanGradConfig,
        leakyReluGradConfig,
        log1pGradConfig,
        logGradConfig,
        logSoftmaxGradConfig,
        lrnGradConfig,
        maxGradConfig,
        maxGradConfig,
        maximumGradConfig,
        maxPool3DGradConfig$2,
        maxPoolGradConfig$2,
        meanGradConfig,
        minGradConfig,
        minimumGradConfig,
        mirrorPadGradConfig,
        modGradConfig,
        multiplyGradConfig,
        negGradConfig,
        oneHotGradConfig,
        onesLikeGradConfig,
        packGradConfig,
        padV2GradConfig,
        padV2GradConfig,
        powGradConfig,
        preluGradConfig,
        prodGradConfig,
        reciprocalGradConfig,
        relu6GradConfig,
        reluGradConfig,
        reshapeGradConfig,
        resizeBilinearGradConfig$2,
        resizeNearestNeighborGradConfig$2,
        reverseGradConfig,
        roundGradConfig,
        rsqrtGradConfig,
        selectGradConfig,
        seluGradConfig,
        sigmoidGradConfig,
        signGradConfig,
        sinGradConfig,
        sinhGradConfig,
        sliceGradConfig,
        softmaxGradConfig,
        softplusGradConfig,
        spaceToBatchNDGradConfig,
        spaceToBatchNDGradConfig,
        splitVGradConfig,
        splitVGradConfig,
        sqrtGradConfig,
        squaredDifferenceGradConfig,
        squareGradConfig,
        stepGradConfig,
        subGradConfig,
        sumGradConfig,
        tanGradConfig,
        tanhGradConfig,
        tileGradConfig,
        transposeGradConfig,
        unpackGradConfig,
        unsortedSegmentSumGradConfig,
        zerosLikeGradConfig
    ];
    for (const gradientConfig of gradConfigs) {
        registerGradient(gradientConfig);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.abs = function () {
        this.throwIfDisposed();
        return abs$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.acos = function () {
        this.throwIfDisposed();
        return acos$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.acosh = function () {
        this.throwIfDisposed();
        return acosh$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.add = function (b) {
        this.throwIfDisposed();
        return add$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.all = function (axis, keepDims) {
        this.throwIfDisposed();
        return all$2(this, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.any = function (axis, keepDims) {
        this.throwIfDisposed();
        return any$2(this, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.argMax = function (axis) {
        this.throwIfDisposed();
        return argMax$2(this, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.argMin = function (axis) {
        this.throwIfDisposed();
        return argMin$2(this, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts a size-1 `tf.Tensor` to a `tf.Scalar`.
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.asScalar = function () {
        this.throwIfDisposed();
        assert$4(this.size === 1, () => 'The array must have only 1 element.');
        return reshape$3(this, []);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Casts a `tf.Tensor` to a specified dtype.
     *
     * @param dtype Data-type to cast the tensor to.
     *
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.asType = function (dtype) {
        this.throwIfDisposed();
        return cast$3(this, dtype);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts a `tf.Tensor` to a `tf.Tensor1D`.
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.as1D = function () {
        this.throwIfDisposed();
        return reshape$3(this, [this.size]);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts a `tf.Tensor` to a `tf.Tensor2D`.
     *
     * @param rows Number of rows in `tf.Tensor2D`.
     * @param columns Number of columns in `tf.Tensor2D`.
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.as2D = function (rows, columns) {
        this.throwIfDisposed();
        return reshape$3(this, [rows, columns]);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts a `tf.Tensor` to a `tf.Tensor3D`.
     *
     * @param rows Number of rows in `tf.Tensor3D`.
     * @param columns Number of columns in `tf.Tensor3D`.
     * @param depth Depth of `tf.Tensor3D`.
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.as3D = function (rows, columns, depth) {
        this.throwIfDisposed();
        return reshape$3(this, [rows, columns, depth]);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts a `tf.Tensor` to a `tf.Tensor4D`.
     *
     * @param rows Number of rows in `tf.Tensor4D`.
     * @param columns Number of columns in `tf.Tensor4D`.
     * @param depth Depth of `tf.Tensor4D`.
     * @param depth2 4th dimension of `tf.Tensor4D`.
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.as4D = function (rows, columns, depth, depth2) {
        this.throwIfDisposed();
        return reshape$3(this, [rows, columns, depth, depth2]);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Converts a `tf.Tensor` to a `tf.Tensor5D`.
     *
     * @param rows Number of rows in `tf.Tensor5D`.
     * @param columns Number of columns in `tf.Tensor5D`.
     * @param depth Depth of `tf.Tensor5D`.
     * @param depth2 4th dimension of `tf.Tensor5D`.
     * @param depth3 5th dimension of 'tf.Tensor5D'
     *
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.as5D = function (rows, columns, depth, depth2, depth3) {
        this.throwIfDisposed();
        return reshape$3(this, [rows, columns, depth, depth2, depth3]);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.asin = function () {
        this.throwIfDisposed();
        return asin$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.asinh = function () {
        this.throwIfDisposed();
        return asinh$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.atan = function () {
        this.throwIfDisposed();
        return atan$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.atan2 = function (b) {
        this.throwIfDisposed();
        return atan2$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.atanh = function () {
        this.throwIfDisposed();
        return atanh$2(this);
    };

    getGlobalTensorClass().prototype.avgPool =
        function (filterSize, strides, pad, dimRoundingMode) {
            this.throwIfDisposed();
            return avgPool$2(this, filterSize, strides, pad, dimRoundingMode);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.batchToSpaceND = function (blockShape, crops) {
        this.throwIfDisposed();
        return batchToSpaceND$2(this, blockShape, crops);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.batchNorm = function (mean, variance, offset, scale, varianceEpsilon) {
        this.throwIfDisposed();
        return batchNorm$2(this, mean, variance, offset, scale, varianceEpsilon);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.broadcastTo = function (shape) {
        this.throwIfDisposed();
        return broadcastTo(this, shape);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.cast = function (dtype) {
        this.throwIfDisposed();
        return cast$3(this, dtype);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.ceil = function () {
        this.throwIfDisposed();
        return ceil$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.clipByValue = function (min, max) {
        this.throwIfDisposed();
        return clipByValue$2(this, min, max);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.concat = function (x, axis) {
        this.throwIfDisposed();
        if (x instanceof Tensor) {
            x = [x];
        }
        return concat$2([this, ...x], axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.conv1d = function (filter, stride, pad, dataFormat, dilation, dimRoundingMode) {
        this.throwIfDisposed();
        return conv1d$1(this, filter, stride, pad, dataFormat, dilation, dimRoundingMode);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.conv2dTranspose =
        function (filter, outputShape, strides, pad, dimRoundingMode) {
            this.throwIfDisposed();
            return conv2dTranspose$1(this, filter, outputShape, strides, pad, dimRoundingMode);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.conv2d = function (filter, strides, pad, dataFormat, dilations, dimRoundingMode) {
        this.throwIfDisposed();
        return conv2d$3(this, filter, strides, pad, dataFormat, dilations, dimRoundingMode);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.cos = function () {
        this.throwIfDisposed();
        return cos$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.cosh = function () {
        this.throwIfDisposed();
        return cosh$2(this);
    };

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the 'License');
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an 'AS IS' BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.cumprod = function (axis, exclusive, reverse) {
        this.throwIfDisposed();
        return cumprod$2(this, axis, exclusive, reverse);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.cumsum = function (axis, exclusive, reverse) {
        this.throwIfDisposed();
        return cumsum$2(this, axis, exclusive, reverse);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.depthToSpace = function (blockSize, dataFormat) {
        this.throwIfDisposed();
        return depthToSpace$2(this, blockSize, dataFormat);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.depthwiseConv2d =
        function (filter, strides, pad, dataFormat, dilations, dimRoundingMode) {
            this.throwIfDisposed();
            return depthwiseConv2d$3(this, filter, strides, pad, dataFormat, dilations, dimRoundingMode);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.dilation2d =
        function (filter, strides, pad, dilations, dataFormat) {
            this.throwIfDisposed();
            return dilation2d(this, filter, strides, pad, dilations, dataFormat);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.divNoNan = function (b) {
        this.throwIfDisposed();
        return divNoNan(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.div = function (b) {
        this.throwIfDisposed();
        return div$1(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.dot = function (b) {
        this.throwIfDisposed();
        return dot$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.elu = function () {
        this.throwIfDisposed();
        return elu$4(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.equal = function (b) {
        this.throwIfDisposed();
        return equal$3(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.erf = function () {
        this.throwIfDisposed();
        return erf$2(this);
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.euclideanNorm = function (axis, keepDims) {
        this.throwIfDisposed();
        return euclideanNorm(this, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.exp = function () {
        this.throwIfDisposed();
        return exp$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.expandDims = function (axis) {
        this.throwIfDisposed();
        return expandDims$3(this, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.expm1 = function () {
        this.throwIfDisposed();
        return expm1$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.fft = function () {
        this.throwIfDisposed();
        return fft$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Flatten a Tensor to a 1D array.
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.flatten = function () {
        this.throwIfDisposed();
        return reshape$3(this, [this.size]);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.floor = function () {
        this.throwIfDisposed();
        return floor$3(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.floorDiv = function (b) {
        this.throwIfDisposed();
        return floorDiv$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.gather = function (indices, axis) {
        this.throwIfDisposed();
        return gather$1(this, indices, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.greaterEqual = function (b) {
        this.throwIfDisposed();
        return greaterEqual$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.greater = function (b) {
        this.throwIfDisposed();
        return greater$3(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.ifft = function () {
        this.throwIfDisposed();
        return ifft$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.irfft = function () {
        this.throwIfDisposed();
        return irfft(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.isFinite = function () {
        this.throwIfDisposed();
        return isFinite$3(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.isInf = function () {
        this.throwIfDisposed();
        return isInf$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.isNaN = function () {
        this.throwIfDisposed();
        return isNaN$3(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.leakyRelu = function (alpha) {
        this.throwIfDisposed();
        return leakyRelu$2(this, alpha);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.lessEqual = function (b) {
        this.throwIfDisposed();
        return lessEqual$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.less = function (b) {
        this.throwIfDisposed();
        return less$3(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.localResponseNormalization =
        function (depthRadius, bias, alpha, beta) {
            this.throwIfDisposed();
            return localResponseNormalization(this, depthRadius, bias, alpha, beta);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.logSigmoid = function () {
        this.throwIfDisposed();
        return logSigmoid(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.logSoftmax = function (axis) {
        this.throwIfDisposed();
        return logSoftmax(this, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.logSumExp = function (axis, keepDims) {
        this.throwIfDisposed();
        return logSumExp(this, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.log = function () {
        this.throwIfDisposed();
        return log$3(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.log1p = function () {
        this.throwIfDisposed();
        return log1p$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.logicalAnd = function (b) {
        this.throwIfDisposed();
        return logicalAnd$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.logicalNot = function () {
        this.throwIfDisposed();
        return logicalNot$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.logicalOr = function (b) {
        this.throwIfDisposed();
        return logicalOr$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.logicalXor = function (b) {
        this.throwIfDisposed();
        return logicalXor(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.matMul = function (b, transposeA, transposeB) {
        this.throwIfDisposed();
        return matMul$1(this, b, transposeA, transposeB);
    };

    getGlobalTensorClass().prototype.maxPool =
        function (filterSize, strides, pad, dimRoundingMode) {
            this.throwIfDisposed();
            return maxPool$2(this, filterSize, strides, pad, dimRoundingMode);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.max = function (axis, keepDims) {
        this.throwIfDisposed();
        return max$3(this, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.maximum = function (b) {
        this.throwIfDisposed();
        return maximum$3(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.mean = function (axis, keepDims) {
        this.throwIfDisposed();
        return mean$1(this, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.min = function (axis, keepDims) {
        this.throwIfDisposed();
        return min$3(this, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.minimum = function (b) {
        this.throwIfDisposed();
        return minimum$3(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.mirrorPad = function (paddings, mode) {
        this.throwIfDisposed();
        return mirrorPad$1(this, paddings, mode);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.mod = function (b) {
        this.throwIfDisposed();
        return mod$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.mul = function (b) {
        this.throwIfDisposed();
        return mul(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.neg = function () {
        this.throwIfDisposed();
        return neg$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.norm = function (ord, axis, keepDims) {
        this.throwIfDisposed();
        return norm(this, ord, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.notEqual = function (b) {
        this.throwIfDisposed();
        return notEqual$3(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.oneHot = function (depth, onValue = 1, offValue = 0) {
        this.throwIfDisposed();
        return oneHot$2(this, depth, onValue, offValue);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.onesLike = function () {
        this.throwIfDisposed();
        return onesLike$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.pad = function (paddings, constantValue) {
        this.throwIfDisposed();
        return pad$1(this, paddings, constantValue);
    };

    getGlobalTensorClass().prototype.pool = function (windowShape, poolingType, padding, dilationRate, strides, dimRoundingMode) {
        this.throwIfDisposed();
        return pool$1(this, windowShape, poolingType, padding, dilationRate, strides, dimRoundingMode);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.pow = function (exp) {
        this.throwIfDisposed();
        return pow$2(this, exp);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.prelu = function (alpha) {
        this.throwIfDisposed();
        return prelu$3(this, alpha);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.prod = function (axis, keepDims) {
        this.throwIfDisposed();
        return prod$2(this, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.reciprocal = function () {
        this.throwIfDisposed();
        return reciprocal$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.relu = function () {
        this.throwIfDisposed();
        return relu$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.relu6 = function () {
        this.throwIfDisposed();
        return relu6$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Reshapes the tensor into the shape of the provided tensor.
     *
     * @param x The tensor of required shape.
     *
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.reshapeAs = function (x) {
        this.throwIfDisposed();
        return reshape$3(this, x.shape);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.reshape = function (shape) {
        this.throwIfDisposed();
        return reshape$3(this, shape);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.resizeBilinear =
        function (newShape2D, alignCorners, halfPixelCenters) {
            this.throwIfDisposed();
            return resizeBilinear$2(this, newShape2D, alignCorners, halfPixelCenters);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.resizeNearestNeighbor =
        function (newShape2D, alignCorners, halfFloatCenters) {
            this.throwIfDisposed();
            return resizeNearestNeighbor$2(this, newShape2D, alignCorners, halfFloatCenters);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.reverse = function (axis) {
        this.throwIfDisposed();
        return reverse$2(this, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.rfft = function () {
        this.throwIfDisposed();
        return rfft(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.round = function () {
        this.throwIfDisposed();
        return round$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.rsqrt = function () {
        this.throwIfDisposed();
        return rsqrt$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.selu = function () {
        this.throwIfDisposed();
        return selu$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.separableConv2d =
        function (depthwiseFilter, pointwiseFilter, strides, pad, dilation, dataFormat) {
            this.throwIfDisposed();
            return separableConv2d$1(this, depthwiseFilter, pointwiseFilter, strides, pad, dilation, dataFormat);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.sigmoid = function () {
        this.throwIfDisposed();
        return sigmoid$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.sign = function () {
        this.throwIfDisposed();
        return sign$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.sin = function () {
        this.throwIfDisposed();
        return sin$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.sinh = function () {
        this.throwIfDisposed();
        return sinh$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.slice = function (begin, size) {
        this.throwIfDisposed();
        return slice$2(this, begin, size);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.softmax = function (dim) {
        this.throwIfDisposed();
        return softmax$3(this, dim);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.softplus = function () {
        this.throwIfDisposed();
        return softplus$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.spaceToBatchND = function (blockShape, paddings) {
        this.throwIfDisposed();
        return spaceToBatchND$2(this, blockShape, paddings);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.split = function (numOrSizeSplits, axis) {
        this.throwIfDisposed();
        return split$2(this, numOrSizeSplits, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.sqrt = function () {
        this.throwIfDisposed();
        return sqrt$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.square = function () {
        this.throwIfDisposed();
        return square$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.squaredDifference = function (b) {
        this.throwIfDisposed();
        return squaredDifference$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.squeeze = function (axis) {
        this.throwIfDisposed();
        return squeeze(this, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.stack = function (x, axis) {
        this.throwIfDisposed();
        const tensorsToBeStacked = x instanceof Tensor ? [this, x] : [this, ...x];
        return stack(tensorsToBeStacked, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.step = function (alpha) {
        this.throwIfDisposed();
        return step$2(this, alpha);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.stridedSlice = function (begin, end, strides, beginMask, endMask, ellipsisMask, newAxisMask, shrinkAxisMask) {
        this.throwIfDisposed();
        return stridedSlice$2(this, begin, end, strides, beginMask, endMask, ellipsisMask, newAxisMask, shrinkAxisMask);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.sub = function (b) {
        this.throwIfDisposed();
        return sub$2(this, b);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.sum = function (axis, keepDims) {
        this.throwIfDisposed();
        return sum$2(this, axis, keepDims);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.tan = function () {
        this.throwIfDisposed();
        return tan$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.tanh = function () {
        this.throwIfDisposed();
        return tanh$2(this);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.tile = function (reps) {
        this.throwIfDisposed();
        return tile$3(this, reps);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Casts the array to type `bool`
     *
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.toBool = function () {
        this.throwIfDisposed();
        return cast$3(this, 'bool');
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Casts the array to type `float32`
     *
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.toFloat = function () {
        this.throwIfDisposed();
        return cast$3(this, 'float32');
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Casts the array to type `int32`
     *
     * @doc {heading: 'Tensors', subheading: 'Classes'}
     */
    getGlobalTensorClass().prototype.toInt = function () {
        this.throwIfDisposed();
        return cast$3(this, 'int32');
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.topk = function (k, sorted) {
        this.throwIfDisposed();
        return topk(this, k, sorted);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.transpose = function (perm) {
        this.throwIfDisposed();
        return transpose$2(this, perm);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.unique = function (axis) {
        this.throwIfDisposed();
        return unique$3(this, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.unsortedSegmentSum =
        function (segmentIds, numSegments) {
            this.throwIfDisposed();
            return unsortedSegmentSum$2(this, segmentIds, numSegments);
        };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.unstack = function (axis) {
        this.throwIfDisposed();
        return unstack(this, axis);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.where = function (condition, x) {
        this.throwIfDisposed();
        return where(condition, this, x);
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    getGlobalTensorClass().prototype.zerosLike = function () {
        this.throwIfDisposed();
        return zerosLike$2(this);
    };

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Explicit error types.
     *
     * See the following link for more information about why the code includes
     * calls to setPrototypeOf:
     *
     * https://github.com/Microsoft/TypeScript-wiki/blob/master/Breaking-Changes.md#extending-built-ins-like-error-array-and-map-may-no-longer-work
     */
    // tslint:enable
    /**
     * Equivalent of Python's AttributeError.
     */
    class AttributeError extends Error {
        constructor(message) {
            super(message);
            // Set the prototype explicitly.
            Object.setPrototypeOf(this, AttributeError.prototype);
        }
    }
    /**
     * Equivalent of Python's RuntimeError.
     */
    class RuntimeError extends Error {
        constructor(message) {
            super(message);
            // Set the prototype explicitly.
            Object.setPrototypeOf(this, RuntimeError.prototype);
        }
    }
    /**
     * Equivalent of Python's ValueError.
     */
    class ValueError extends Error {
        constructor(message) {
            super(message);
            // Set the prototype explicitly.
            Object.setPrototypeOf(this, ValueError.prototype);
        }
    }
    /**
     * Equivalent of Python's NotImplementedError.
     */
    class NotImplementedError extends Error {
        constructor(message) {
            super(message);
            // Set the prototype explicitly.
            Object.setPrototypeOf(this, NotImplementedError.prototype);
        }
    }
    /**
     * Equivalent of Python's AssertionError.
     */
    class AssertionError$1 extends Error {
        constructor(message) {
            super(message);
            // Set the prototype explicitly.
            Object.setPrototypeOf(this, AssertionError$1.prototype);
        }
    }

    /**
     * @license
     * Copyright 2022 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * LruCache: A mapping from the String to T. If the number of the entries is
     * exceeding the `maxEntries`, the LruCache will delete the least recently
     * used entry.
     */
    class LruCache {
        constructor(maxEntries) {
            this.maxEntries = maxEntries || 100;
            this.cache = new Map();
        }
        /**
         * Get the entry for the key and mark it as used recently.
         */
        get(key) {
            let entry;
            if (this.cache.has(key)) {
                entry = this.cache.get(key);
                this.cache.delete(key);
                this.cache.set(key, entry);
            }
            return entry;
        }
        /**
         * Put the entry into the cache. If the key already existed, mark the key as
         * used recently.
         */
        put(key, value) {
            if (this.cache.has(key)) {
                this.cache.delete(key);
            }
            else if (this.cache.size >= this.maxEntries) {
                const keyToDelete = this.cache.keys().next().value;
                this.cache.delete(keyToDelete);
            }
            this.cache.set(key, value);
        }
        /**
         * Get the MaxEntries of the cache.
         */
        getMaxEntries() {
            return this.maxEntries;
        }
        /**
         * Set the MaxEntries of the cache. If the maxEntries is decreased, reduce
         * entries in the cache.
         */
        setMaxEntries(maxEntries) {
            if (maxEntries < 0) {
                throw new Error(`The maxEntries of LRU caches must be at least 0, but got ${maxEntries}.`);
            }
            if (this.maxEntries > maxEntries) {
                for (let i = 0; i < this.maxEntries - maxEntries; i++) {
                    const keyToDelete = this.cache.keys().next().value;
                    this.cache.delete(keyToDelete);
                }
            }
            this.maxEntries = maxEntries;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    // tslint:enable
    /**
     * If `value` is an Array, equivalent to Python's `value * numValues`.
     * If `value` is not an Array, equivalent to Python's `[value] * numValues`
     */
    // tslint:disable-next-line:no-any
    function pyListRepeat(value, numValues) {
        if (Array.isArray(value)) {
            // tslint:disable-next-line:no-any
            let newArray = [];
            for (let i = 0; i < numValues; i++) {
                newArray = newArray.concat(value);
            }
            return newArray;
        }
        else {
            const newArray = new Array(numValues);
            newArray.fill(value);
            return newArray;
        }
    }
    function assert$3(val, message) {
        if (!val) {
            throw new AssertionError$1(message);
        }
    }
    /**
     * Count the number of elements of the `array` that are equal to `reference`.
     */
    function count(array, refernce) {
        let counter = 0;
        for (const item of array) {
            if (item === refernce) {
                counter++;
            }
        }
        return counter;
    }
    /**
     * If an array is of length 1, just return the first element. Otherwise, return
     * the full array.
     * @param tensors
     */
    function singletonOrArray(xs) {
        if (xs.length === 1) {
            return xs[0];
        }
        return xs;
    }
    /**
     * Normalizes a list/tensor into a list.
     *
     * If a tensor is passed, we return
     * a list of size 1 containing the tensor.
     *
     * @param x target object to be normalized.
     */
    // tslint:disable-next-line:no-any
    function toList(x) {
        if (Array.isArray(x)) {
            return x;
        }
        return [x];
    }
    /**
     * Converts string to snake-case.
     * @param name
     */
    function toSnakeCase(name) {
        const intermediate = name.replace(/(.)([A-Z][a-z0-9]+)/g, '$1_$2');
        const insecure = intermediate.replace(/([a-z])([A-Z])/g, '$1_$2').toLowerCase();
        /*
         If the class is private the name starts with "_" which is not secure
         for creating scopes. We prefix the name with "private" in this case.
         */
        if (insecure[0] !== '_') {
            return insecure;
        }
        return 'private' + insecure;
    }
    function toCamelCase(identifier) {
        // quick return for empty string or single character strings
        if (identifier.length <= 1) {
            return identifier;
        }
        // Check for the underscore indicating snake_case
        if (identifier.indexOf('_') === -1) {
            return identifier;
        }
        return identifier.replace(/[_]+(\w|$)/g, (m, p1) => p1.toUpperCase());
    }
    // tslint:disable-next-line:no-any
    let _GLOBAL_CUSTOM_OBJECTS = {};
    function serializeKerasObject(instance) {
        if (instance === null || instance === undefined) {
            return null;
        }
        const dict = {};
        dict['className'] = instance.getClassName();
        dict['config'] = instance.getConfig();
        return dict;
    }
    /**
     * Replace ndarray-style scalar objects in serialization objects with numbers.
     *
     * Background: In some versions of tf.keras, certain scalar values in the HDF5
     * model save file can be serialized as: `{'type': 'ndarray', 'value': num}`,
     * where in `num` is a plain number. This method converts such serialization
     * to a `number`.
     *
     * @param config The keras-format serialization object to be processed
     *   (in place).
     */
    function convertNDArrayScalarsInConfig(config) {
        if (config == null || typeof config !== 'object') {
            return;
        }
        else if (Array.isArray(config)) {
            config.forEach(configItem => convertNDArrayScalarsInConfig(configItem));
        }
        else {
            const fields = Object.keys(config);
            for (const field of fields) {
                const value = config[field];
                if (value != null && typeof value === 'object') {
                    if (!Array.isArray(value) && value['type'] === 'ndarray' &&
                        typeof value['value'] === 'number') {
                        config[field] = value['value'];
                    }
                    else {
                        convertNDArrayScalarsInConfig(value);
                    }
                }
            }
        }
    }
    /**
     * Deserialize a saved Keras Object
     * @param identifier either a string ID or a saved Keras dictionary
     * @param moduleObjects a list of Python class names to object constructors
     * @param customObjects a list of Python class names to object constructors
     * @param printableModuleName debug text for the object being reconstituted
     * @param fastWeightInit Optional flag to use fast weight initialization
     *   during deserialization. This is applicable to cases in which
     *   the initialization will be immediately overwritten by loaded weight
     *   values. Default: `false`.
     * @returns a TensorFlow.js Layers object
     */
    // tslint:disable:no-any
    function deserializeKerasObject(identifier, moduleObjects = {}, customObjects = {}, printableModuleName = 'object', fastWeightInit = false) {
        // tslint:enable
        if (typeof identifier === 'string') {
            const functionName = identifier;
            let fn;
            if (functionName in customObjects) {
                fn = customObjects[functionName];
            }
            else if (functionName in _GLOBAL_CUSTOM_OBJECTS) {
                fn = _GLOBAL_CUSTOM_OBJECTS[functionName];
            }
            else {
                fn = moduleObjects[functionName];
                if (fn == null) {
                    throw new ValueError(`Unknown ${printableModuleName}: ${identifier}. ` +
                        `This may be due to one of the following reasons:\n` +
                        `1. The ${printableModuleName} is defined in Python, in which ` +
                        `case it needs to be ported to TensorFlow.js or your JavaScript ` +
                        `code.\n` +
                        `2. The custom ${printableModuleName} is defined in JavaScript, ` +
                        `but is not registered properly with ` +
                        `tf.serialization.registerClass().`);
                    // TODO(cais): Add link to tutorial page on custom layers.
                }
            }
            return fn;
        }
        else {
            // In this case we are dealing with a Keras config dictionary.
            const config = identifier;
            if (config['className'] == null || config['config'] == null) {
                throw new ValueError(`${printableModuleName}: Improper config format: ` +
                    `${JSON.stringify(config)}.\n` +
                    `'className' and 'config' must set.`);
            }
            const className = config['className'];
            let cls, fromConfig;
            if (className in customObjects) {
                [cls, fromConfig] = customObjects[className];
            }
            else if (className in _GLOBAL_CUSTOM_OBJECTS) {
                [cls, fromConfig] = _GLOBAL_CUSTOM_OBJECTS['className'];
            }
            else if (className in moduleObjects) {
                [cls, fromConfig] = moduleObjects[className];
            }
            if (cls == null) {
                throw new ValueError(`Unknown ${printableModuleName}: ${className}. ` +
                    `This may be due to one of the following reasons:\n` +
                    `1. The ${printableModuleName} is defined in Python, in which ` +
                    `case it needs to be ported to TensorFlow.js or your JavaScript ` +
                    `code.\n` +
                    `2. The custom ${printableModuleName} is defined in JavaScript, ` +
                    `but is not registered properly with ` +
                    `tf.serialization.registerClass().`);
                // TODO(cais): Add link to tutorial page on custom layers.
            }
            if (fromConfig != null) {
                // Porting notes: Instead of checking to see whether fromConfig accepts
                // customObjects, we create a customObjects dictionary and tack it on to
                // config['config'] as config['config'].customObjects. Objects can use it,
                // if they want.
                // tslint:disable-next-line:no-any
                const customObjectsCombined = {};
                for (const key of Object.keys(_GLOBAL_CUSTOM_OBJECTS)) {
                    customObjectsCombined[key] = _GLOBAL_CUSTOM_OBJECTS[key];
                }
                for (const key of Object.keys(customObjects)) {
                    customObjectsCombined[key] = customObjects[key];
                }
                // Add the customObjects to config
                const nestedConfig = config['config'];
                nestedConfig['customObjects'] = customObjectsCombined;
                const backupCustomObjects = Object.assign({}, _GLOBAL_CUSTOM_OBJECTS);
                for (const key of Object.keys(customObjects)) {
                    _GLOBAL_CUSTOM_OBJECTS[key] = customObjects[key];
                }
                convertNDArrayScalarsInConfig(config['config']);
                const returnObj = fromConfig(cls, config['config'], customObjects, fastWeightInit);
                _GLOBAL_CUSTOM_OBJECTS = Object.assign({}, backupCustomObjects);
                return returnObj;
            }
            else {
                // Then `cls` may be a function returning a class.
                // In this case by convention `config` holds
                // the kwargs of the function.
                const backupCustomObjects = Object.assign({}, _GLOBAL_CUSTOM_OBJECTS);
                for (const key of Object.keys(customObjects)) {
                    _GLOBAL_CUSTOM_OBJECTS[key] = customObjects[key];
                }
                // In python this is **config['config'], for tfjs-layers we require
                // classes that use this fall-through construction method to take
                // a config interface that mimics the expansion of named parameters.
                const returnObj = new cls(config['config']);
                _GLOBAL_CUSTOM_OBJECTS = Object.assign({}, backupCustomObjects);
                return returnObj;
            }
        }
    }
    /**
     * Compares two numbers for sorting.
     * @param a
     * @param b
     */
    function numberCompare(a, b) {
        return (a < b) ? -1 : ((a > b) ? 1 : 0);
    }
    /**
     * Comparison of two numbers for reverse sorting.
     * @param a
     * @param b
     */
    function reverseNumberCompare(a, b) {
        return -1 * numberCompare(a, b);
    }
    /**
     * Get the unique elements of an array.
     * @param xs Array.
     * @returns An Array consisting of the unique elements in `xs`.
     */
    function unique$2(xs) {
        if (xs == null) {
            return xs;
        }
        const out = [];
        // TODO(cais): Maybe improve performance by sorting.
        for (const x of xs) {
            if (out.indexOf(x) === -1) {
                out.push(x);
            }
        }
        return out;
    }
    /**
     * Determine if an Object is empty (i.e., does not have own properties).
     * @param obj Object
     * @returns Whether the Object is empty.
     * @throws ValueError: If object is `null` or `undefined`.
     */
    function isObjectEmpty(obj) {
        if (obj == null) {
            throw new ValueError(`Invalid value in obj: ${JSON.stringify(obj)}`);
        }
        for (const key in obj) {
            if (obj.hasOwnProperty(key)) {
                return false;
            }
        }
        return true;
    }
    /**
     * Helper function used to build type union/enum run-time checkers.
     * @param values The list of allowed values.
     * @param label A string name for the type
     * @param value The value to test.
     * @throws ValueError: If the value is not in values nor `undefined`/`null`.
     */
    function checkStringTypeUnionValue(values, label, value) {
        if (value == null) {
            return;
        }
        if (values.indexOf(value) < 0) {
            throw new ValueError(`${value} is not a valid ${label}.  Valid values are ${values} or null/undefined.`);
        }
    }
    /**
     * Helper function for verifying the types of inputs.
     *
     * Ensures that the elements of `x` are all of type `expectedType`.
     * Also verifies that the length of `x` is within bounds.
     *
     * @param x Object to test.
     * @param expectedType The string expected type of all of the elements in the
     * Array.
     * @param minLength Return false if x.length is less than this.
     * @param maxLength Return false if x.length is greater than this.
     * @returns true if and only if `x` is an `Array<expectedType>` with
     * length >= `minLength` and <= `maxLength`.
     */
    // tslint:disable:no-any
    function checkArrayTypeAndLength(x, expectedType, minLength = 0, maxLength = Infinity) {
        assert$3(minLength >= 0);
        assert$3(maxLength >= minLength);
        return (Array.isArray(x) && x.length >= minLength && x.length <= maxLength &&
            x.every(e => typeof e === expectedType));
    }
    // tslint:enable:no-any
    /**
     * Assert that a value or an array of value are positive integer.
     *
     * @param value The value being asserted on. May be a single number or an array
     *   of numbers.
     * @param name Name of the value, used to make the error message.
     */
    function assertPositiveInteger(value, name) {
        if (Array.isArray(value)) {
            assert$4(value.length > 0, () => `${name} is unexpectedly an empty array.`);
            value.forEach((v, i) => assertPositiveInteger(v, `element ${i + 1} of ${name}`));
        }
        else {
            assert$4(Number.isInteger(value) && value > 0, () => `Expected ${name} to be a positive integer, but got ` +
                `${formatAsFriendlyString(value)}.`);
        }
    }
    /**
     * Format a value into a display-friendly, human-readable fashion.
     *
     * - `null` is formatted as `'null'`
     * - Strings are formated with flanking pair of quotes.
     * - Arrays are formatted with flanking pair of square brackets.
     *
     * @param value The value to display.
     * @return Formatted string.
     */
    // tslint:disable-next-line:no-any
    function formatAsFriendlyString(value) {
        if (value === null) {
            return 'null';
        }
        else if (Array.isArray(value)) {
            return '[' + value.map(v => formatAsFriendlyString(v)).join(',') + ']';
        }
        else if (typeof value === 'string') {
            return `"${value}"`;
        }
        else {
            return `${value}`;
        }
    }
    /**
     * Returns a function `f2` (decorator) which wraps the original function
     * `f`. `f2` guarantees that `f` can be called at most once
     * every `waitMs` ms. If `f2` is called more often, it will return
     * the last returned result of `f`.
     *
     * @param f The original function `f` to wrap.
     * @param waitMs The time between two consecutive calls to `f` in ms.
     */
    function debounce(f, waitMs, nowFunc) {
        let lastTime = nowFunc != null ? nowFunc() : now();
        let lastResult;
        const f2 = (...args) => {
            const now$1 = nowFunc != null ? nowFunc() : now();
            if (now$1 - lastTime < waitMs) {
                return lastResult;
            }
            lastTime = now$1;
            lastResult = f(...args);
            return lastResult;
        };
        return f2;
    }
    /**
     * Returns the fusable activation given a layers identifier.
     *
     * @param activationName The layers identifier string.
     * @return The name of the fusable activation.
     */
    function mapActivationToFusedKernel(activationName) {
        if (activationName === 'relu') {
            return 'relu';
        }
        if (activationName === 'linear') {
            return 'linear';
        }
        if (activationName === 'elu') {
            return 'elu';
        }
        return null;
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Utilities related to persistent state in the backend.
     */
    /**
     * An ID to track `tf.SymbolicTensor`s and derived classes.
     * Required in different places in engine/topology.ts to identify unique
     * tensors.
     */
    let _nextUniqueTensorId = 0;
    function getNextUniqueTensorId() {
        return _nextUniqueTensorId++;
    }
    const _uidPrefixes = {};
    /**
     * Provides a unique UID given a string prefix.
     *
     * @param prefix
     */
    function getUid(prefix = '') {
        if (!(prefix in _uidPrefixes)) {
            _uidPrefixes[prefix] = 0;
        }
        _uidPrefixes[prefix] += 1;
        return prefix + _uidPrefixes[prefix].toString();
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    const VALID_DATA_FORMAT_VALUES = ['channelsFirst', 'channelsLast'];
    const VALID_INTERPOLATION_FORMAT_VALUES = ['nearest', 'bilinear'];
    const VALID_PADDING_MODE_VALUES = ['valid', 'same', 'causal'];
    const VALID_POOL_MODE_VALUES = ['max', 'avg'];
    const VALID_BIDIRECTIONAL_MERGE_MODES = ['sum', 'mul', 'concat', 'ave'];

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    // A map from the requested scoped name of a Tensor to the number of Tensors
    // wanting that name so far.  This allows enforcing name uniqueness by appending
    // an incrementing index, e.g. scope/name, scope/name_1, scope/name_2, etc.
    const nameMap = new Map();
    function checkDataFormat(value) {
        checkStringTypeUnionValue(VALID_DATA_FORMAT_VALUES, 'DataFormat', value);
    }
    function checkInterpolationFormat(value) {
        checkStringTypeUnionValue(VALID_INTERPOLATION_FORMAT_VALUES, 'InterpolationFormat', value);
    }
    function checkPaddingMode(value) {
        checkStringTypeUnionValue(VALID_PADDING_MODE_VALUES, 'PaddingMode', value);
    }
    function checkPoolMode(value) {
        checkStringTypeUnionValue(VALID_POOL_MODE_VALUES, 'PoolMode', value);
    }
    const _nameScopeStack = [];
    const _nameScopeDivider = '/';
    /**
     * Enter namescope, which can be nested.
     */
    function nameScope(name, fn) {
        _nameScopeStack.push(name);
        try {
            const val = fn();
            _nameScopeStack.pop();
            return val;
        }
        catch (e) {
            _nameScopeStack.pop();
            throw e;
        }
    }
    /**
     * Get the current namescope as a flat, concatenated string.
     */
    function currentNameScopePrefix() {
        if (_nameScopeStack.length === 0) {
            return '';
        }
        else {
            return _nameScopeStack.join(_nameScopeDivider) + _nameScopeDivider;
        }
    }
    /**
     * Get the name a Tensor (or Variable) would have if not uniqueified.
     * @param tensorName
     * @return Scoped name string.
     */
    function getScopedTensorName(tensorName) {
        if (!isValidTensorName(tensorName)) {
            throw new Error('Not a valid tensor name: \'' + tensorName + '\'');
        }
        return currentNameScopePrefix() + tensorName;
    }
    /**
     * Get unique names for Tensors and Variables.
     * @param scopedName The fully-qualified name of the Tensor, i.e. as produced by
     *  `getScopedTensorName()`.
     * @return A unique version of the given fully scoped name.
     *   If this is the first time that the scoped name is seen in this session,
     *   then the given `scopedName` is returned unaltered.  If the same name is
     *   seen again (producing a collision), an incrementing suffix is added to the
     *   end of the name, so it takes the form 'scope/name_1', 'scope/name_2', etc.
     */
    function getUniqueTensorName(scopedName) {
        if (!isValidTensorName(scopedName)) {
            throw new Error('Not a valid tensor name: \'' + scopedName + '\'');
        }
        if (!nameMap.has(scopedName)) {
            nameMap.set(scopedName, 0);
        }
        const index = nameMap.get(scopedName);
        nameMap.set(scopedName, nameMap.get(scopedName) + 1);
        if (index > 0) {
            const result = `${scopedName}_${index}`;
            // Mark the composed name as used in case someone wants
            // to call getUniqueTensorName("name_1").
            nameMap.set(result, 1);
            return result;
        }
        else {
            return scopedName;
        }
    }
    const tensorNameRegex = new RegExp(/^[A-Za-z0-9][-A-Za-z0-9\._\/]*$/);
    /**
     * Determine whether a string is a valid tensor name.
     * @param name
     * @returns A Boolean indicating whether `name` is a valid tensor name.
     */
    function isValidTensorName(name) {
        return !!name.match(tensorNameRegex);
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Determine if a number is an integer.
     */
    function isInteger(x) {
        return x === parseInt(x.toString(), 10);
    }
    /**
     * Calculate the product of an array of numbers.
     * @param array The array to calculate the product over.
     * @param begin Beginning index, inclusive.
     * @param end Ending index, exclusive.
     * @return The product.
     */
    function arrayProd(array, begin, end) {
        if (begin == null) {
            begin = 0;
        }
        if (end == null) {
            end = array.length;
        }
        let prod = 1;
        for (let i = begin; i < end; ++i) {
            prod *= array[i];
        }
        return prod;
    }
    /**
     * Compute minimum value.
     * @param array
     * @return minimum value.
     */
    function min$2(array) {
        // same behavior as tf.min()
        if (array.length === 0) {
            return Number.NaN;
        }
        let min = Number.POSITIVE_INFINITY;
        for (let i = 0; i < array.length; i++) {
            const value = array[i];
            if (value < min) {
                min = value;
            }
        }
        return min;
    }
    /**
     * Compute maximum value.
     * @param array
     * @return maximum value
     */
    function max$2(array) {
        // same behavior as tf.max()
        if (array.length === 0) {
            return Number.NaN;
        }
        let max = Number.NEGATIVE_INFINITY;
        for (let i = 0; i < array.length; i++) {
            const value = array[i];
            if (value > max) {
                max = value;
            }
        }
        return max;
    }
    /**
     * Generate an array of integers in [begin, end).
     * @param begin Beginning integer, inclusive.
     * @param end Ending integer, exclusive.
     * @returns Range array.
     * @throws ValueError, iff `end` < `begin`.
     */
    function range$3(begin, end) {
        if (end < begin) {
            throw new ValueError(`end (${end}) < begin (${begin}) is forbidden.`);
        }
        const out = [];
        for (let i = begin; i < end; ++i) {
            out.push(i);
        }
        return out;
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    let _epsilon;
    /**
     * Returns the value of the fuzz factor used in numeric expressions.
     */
    function epsilon$1() {
        if (_epsilon == null) {
            _epsilon = backend().epsilon();
        }
        return _epsilon;
    }
    /**
     * Returns the default image data format convention.
     */
    function imageDataFormat() {
        return 'channelsLast';
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Casts a tensor to a different dtype and returns it.
     * @param x Input tensor.
     * @param dtype String: 'float32'|'int32'|'bool'.
     * @returns Tensor of the specified `dtype`.
     */
    function cast$2(x, dtype) {
        return cast$3(x, dtype);
    }
    /**
     * Adds a 1-sized dimension at index "axis".
     * @param x Input tensor.
     * @param axis Position where to add the new axis.
     * @returns Result of the dimension expansion.
     */
    function expandDims$2(x, axis = -1) {
        const outShape = x.shape.slice();
        if (axis < 0) {
            axis = outShape.length + axis + 1;
        }
        outShape.splice(axis, 0, 1);
        return reshape$3(x, outShape);
    }
    /**
     * Repeats a 2D tensor.
     *
     * If `x` has shape `[samples, dim]` and `n` is 2, for example, the output
     * will have shape `[samples, 2, dim]`.
     *
     * @param x Input tensor.
     * @param n Integer, number of times to repeat.
     * @returns The result of the repeat operation.
     * @throws ValueError: If input tensor is not 2D.
     */
    function repeat(x, n) {
        return tidy(() => {
            if (x.shape.length !== 2) {
                throw new ValueError(`repeat() expects a rank-2 tensor, but received a ` +
                    `rank-${x.shape.length} tensor.`);
            }
            const y = expandDims$2(x, 1);
            return tile$2(y, [1, n, 1]);
        });
    }
    /**
     * Flatten a Tensor into 1D.
     * @param x Input tensor.
     * @return The result of the flattening `x`.
     */
    function flatten$2(x) {
        const newShape = [arrayProd(x.shape)];
        return reshape$3(x, newShape);
    }
    /**
     * Turn a nD tensor into a 2D tensor with same 0th dimension.
     * In other words, it flattens each data samples of a batch.
     *
     * @param x The tensor to flatten. The rank of this tensor is required to be 2
     *   or higher.
     * @return The result of the flattening.
     */
    function batchFlatten(x) {
        if (x.rank <= 1) {
            throw new ValueError(`batchFlatten requires a minimum rank of 2. Got rank: ${x.rank}.`);
        }
        const newShape = [x.shape[0], arrayProd(x.shape, 1)];
        return reshape$3(x, newShape);
    }
    /**
     * Do slicing along the first axis.
     * @param array input `tf.Tensor`.
     * @param start starting index, inclusive.
     * @param size size of the slice along the first axis.
     * @returns result of the slicing.
     * @throws ValueError: If `array` is of an unsupported subtype of `tf.Tensor`.
     */
    function sliceAlongFirstAxis(array, start, size) {
        return tidy(() => {
            switch (array.rank) {
                case 1:
                    return slice1d(array, start, size);
                case 2:
                    return slice2d(array, [start, 0], [size, array.shape[1]]);
                case 3:
                    return slice3d(array, [start, 0, 0], [size, array.shape[1], array.shape[2]]);
                case 4:
                    return slice4d(array, [start, 0, 0, 0], [size, array.shape[1], array.shape[2], array.shape[3]]);
                case 5:
                    return slice$2(array, [start, 0, 0, 0, 0], [
                        size, array.shape[1], array.shape[2], array.shape[3], array.shape[4]
                    ]);
                case 6:
                    return slice$2(array, [start, 0, 0, 0, 0, 0], [
                        size, array.shape[1], array.shape[2], array.shape[3], array.shape[4],
                        array.shape[5]
                    ]);
                default:
                    throw new ValueError(`sliceAlongFirstAxis() received an unsupported tensor rank: ` +
                        `${array.rank}`);
            }
        });
    }
    /**
     * Do slicing along the last axis.
     * @param array input `tf.Tensor`.
     * @param start starting index, inclusive.
     * @param size size of the slice along the last axis.
     * @returns result of the slicing.
     * @throws ValueError: If `array` is of an unsupported subtype of `tf.Tensor`.
     */
    function sliceAlongLastAxis(array, start, size) {
        return tidy(() => {
            switch (array.rank) {
                case 1:
                    return slice1d(array, start, size);
                case 2:
                    return slice2d(array, [0, start], [array.shape[0], size]);
                case 3:
                    return slice3d(array, [0, 0, start], [array.shape[0], array.shape[1], size]);
                case 4:
                    return slice4d(array, [0, 0, 0, start], [array.shape[0], array.shape[1], array.shape[2], size]);
                default:
                    throw new ValueError(`sliceAlongLastAxis() received an unsupported tensor rank: ` +
                        `${array.rank}`);
            }
        });
    }
    /**
     * Do slicing along the sepcified axis.
     * @param array input `tf.Tensor`.
     * @param start starting index, inclusive.
     * @param size of the slice along the chosen axis.
     * @param choose an axis.
     * @returns result of the slicing.
     * @throws ValueError: If `array` is of an unsupported subtype of `tf.Tensor`.
     */
    function sliceAlongAxis(array, start, size, axis) {
        return tidy(() => {
            switch (array.rank) {
                case 1:
                    return slice1d(array, start, size);
                case 2:
                    switch (axis) {
                        case 1:
                            return sliceAlongFirstAxis(array, start, size);
                        case 2:
                            return sliceAlongLastAxis(array, start, size);
                        default:
                            throw new ValueError(`The axis is not within the rank of the tensor ` +
                                `${axis}`);
                    }
                case 3:
                    switch (axis) {
                        case 1:
                            return sliceAlongFirstAxis(array, start, size);
                        case 2:
                            return slice3d(array, [0, start, 0], [array.shape[0], size, array.shape[2]]);
                        case 3:
                            return sliceAlongLastAxis(array, start, size);
                        default:
                            throw new ValueError(`The axis is not within the rank of the tensor ` +
                                `${axis}`);
                    }
                case 4:
                    switch (axis) {
                        case 1:
                            return sliceAlongFirstAxis(array, start, size);
                        case 2:
                            return slice4d(array, [0, start, 0, 0], [array.shape[0], size, array.shape[2], array.shape[3]]);
                        case 3:
                            return slice4d(array, [0, 0, start, 0], [array.shape[0], array.shape[1], size, array.shape[3]]);
                        case 4:
                            return sliceAlongLastAxis(array, start, size);
                        default:
                            throw new ValueError(`The axis is not within the rank of the tensor ` +
                                `${axis}`);
                    }
                default:
                    throw new ValueError(`sliceAlongLastAxis() received an unsupported tensor rank: ` +
                        `${array.rank}`);
            }
        });
    }
    /**
     * Concatenates a list of tensors alongside the specified axis.
     * @param tensors `Array` of tensors to concatenate.
     * @param axis Concatenation axis.
     * @returns The result of the concatenation.
     */
    function concatenate$1(tensors, axis = -1) {
        let rank;
        if (axis < 0) {
            rank = tensors[0].rank;
            if (rank !== 0) {
                axis = rank;
            }
            else {
                axis = 0;
            }
        }
        if (axis === tensors[0].rank) {
            // Porting Note: This is necessary because tfc.concat() requires axis to be
            //   in the interval [-rank, rank).
            axis = -1;
        }
        // Porting Note: Sparse concat is not supported yet.
        return concat$2(tensors, axis);
    }
    /**
     * Concatenate two arrays along the first dimension.
     * @param a The 1st `tf.Tensor` to concatenate.
     * @param b The 2nd `tf.Tensor` to concatenate.
     * @returns Result of the concatenation.
     * @throws ValueError: If `a` is of an unsupported subtype of `tf.Tensor`.
     */
    function concatAlongFirstAxis(a, b) {
        switch (a.rank) {
            case 1:
                return concat1d([a, b]);
            case 2:
                return concat2d([a, b], 0);
            case 3:
                return concat3d([a, b], 0);
            case 4:
                return concat4d([a, b], 0);
            default:
                throw new ValueError(`concatAlongFirstAxis() received an unsupported ` +
                    `tensor rank: ${a.rank}`);
        }
    }
    /**
     * Creates a tensor by tiling `x` by `n`.
     * @param x A tensor.
     * @param n An Array of integers or a single integer. If an Array, the length
     *   must be the same as the number of dimensions in `x`. If a single integer,
     *   it will be treated as an Array of length 1.
     */
    function tile$2(x, n) {
        if (!Array.isArray(n)) {
            n = [n];
        }
        if (x.rank !== n.length) {
            throw new ValueError(`The length of input n (${n.length}) does not match ` +
                `the number of dimensions in input x (${x.rank})`);
        }
        return tile$3(x, n);
    }
    /* Creation of random tensors. */
    /**
     * Get a tensor with normal distribution of values.
     *
     * @param shape Shape of the tensor.
     * @param mean mean value of the normal distribution.
     * @param stddev standard deviation of the normal distribution.
     * @param dtype
     * @param seed
     * @return The normal tensor.
     */
    function randomNormal$1(shape, mean = 0.0, stddev = 1.0, dtype, seed) {
        return randomNormal$2(shape, mean, stddev, dtype, seed);
    }
    /* Linear Algebra */
    /**
     * Multiply two tensors and returns the result as a tensor.
     *
     * For 2D tensors, this is equivalent to matrix multiplication (matMul).
     * For tensors of higher ranks, it follows the Theano behavior,
     * (e.g. `(2, 3) * (4, 3, 5) -> (2, 4, 5)`).  From the Theano documentation:
     *
     * For N dimensions it is a sum product over the last axis of x and the
     * second-to-last of y:
     *
     * @param a A tensor of at least rank 2.
     * @param b A tensor of at least rank 2.
     * @param activation (optional) A string identifying the activation
     *   function.
     * @return Result of the dot operation.
     */
    function dot$1(a, b, activation, bias) {
        if ((a.rank < 2) || (b.rank < 2)) {
            throw new NotImplementedError(`dot requires both inputs to be rank >= 2` +
                ` but got x shape = ${a.shape} and y shape = ${b.shape}`);
        }
        if (b.rank >= 3) {
            const xLastDim = a.shape.slice(-1)[0];
            const ySecondLastDim = b.shape.slice(-2)[0];
            if (xLastDim !== ySecondLastDim) {
                throw new NotImplementedError(`If rank y >= 3, then the second last dim` +
                    ` of y must equal the last dim of x but got x shape = ${a.shape} and ` +
                    ` y shape = ${b.shape}`);
            }
        }
        // Handle basic 2D x 2D case.
        if ((a.rank === 2) && (b.rank === 2)) {
            const transposeA = false;
            const transposeB = false;
            // tfc.fused.matMul only fuses certain activation functions. Unsupported
            // activation functions are treated as 'linear' activations, which is
            // equivalent to a no-op.
            return matMul({
                a,
                b: b,
                transposeA,
                transposeB,
                bias: bias ? reshapeBias(a.rank, bias, imageDataFormat()) : null,
                activation
            });
        }
        else {
            // Reshape x into the analogous 2D Tensor.
            const aFirstDims = a.shape.slice(); // Holds all but the last dim of x.
            const aLastDim = aFirstDims.pop();
            a = reshape$3(a, [-1, aLastDim]);
            // Reshape y into the analogous 2D Tensor, and keep track of the
            // required dimensions to reproduce the output shape.
            const bShape = b.shape.slice();
            const bLastDim = bShape.pop();
            const ySecondLastDim = bShape.pop();
            const yOtherDims = [...bShape, bLastDim];
            // permutation should be like [r-2, 0, 1, 2, ... r-4, r-3, r-1]
            // where r is the rank of y.
            const perm = Array.from({ length: b.rank }, (_, i) => {
                if (i === 0) {
                    return b.rank - 2;
                }
                else if (i <= b.rank - 2) {
                    return i - 1;
                }
                return i;
            });
            b = reshape$3(transpose$2(b, perm), [ySecondLastDim, -1]);
            // Multiply x and y as 2D Tensors, and then reshape back to original.
            const outputShape = [...aFirstDims, ...yOtherDims];
            const transposeA = false;
            const transposeB = false;
            return reshape$3(matMul({
                a,
                b,
                transposeA,
                transposeB,
                bias: bias ? reshapeBias(a.rank, bias, imageDataFormat()) : null,
                activation
            }), outputShape);
        }
    }
    /* Elementary math functions. */
    /**
     * Retrieves the elements of indices `indices` in the tensor `reference`.
     * @param reference A tensor.
     * @param indices An integer tensor of indices or an `Array` of integers.
     * @param axis Axis along which to perform the gather operation.
     * @returns The result of the gathering as a tensor.
     */
    function gather(reference, indices, axis) {
        return tidy(() => {
            if (Array.isArray(indices)) {
                indices = tensor1d(indices, 'int32');
            }
            else {
                indices = cast$3(indices, 'int32');
            }
            return gather$1(reference, indices, axis);
        });
    }
    /**
     * Element-wise square.
     * @param x Input tensor.
     * @return element-wise x^2
     */
    function square$1(x) {
        return mul(x, x);
    }
    /**
     * Reshapes bias tensor according to rank of x.
     */
    function reshapeBias(xRank, bias, dataFormat) {
        const biasShape = bias.shape;
        if (bias.rank !== 1 && bias.rank !== xRank) {
            throw new ValueError(`Unexpected bias dimensions: ${bias.rank}` +
                `; expected it to be 1 or ${xRank}`);
        }
        if (xRank === 5) {
            if (dataFormat === 'channelsFirst') {
                if (biasShape.length === 1) {
                    return reshape$3(bias, [1, biasShape[0], 1, 1, 1]);
                }
                else {
                    return reshape$3(bias, [1, biasShape[3], biasShape[0], biasShape[1], biasShape[2]]);
                }
            }
            else if (dataFormat === 'channelsLast') {
                if (biasShape.length === 1) {
                    return reshape$3(bias, [1, 1, 1, 1, biasShape[0]]);
                }
                else {
                    return reshape$3(bias, [1].concat(biasShape));
                }
            }
        }
        else if (xRank === 4) {
            if (dataFormat === 'channelsFirst') {
                if (biasShape.length === 1) {
                    return reshape$3(bias, [1, biasShape[0], 1, 1]);
                }
                else {
                    return reshape$3(bias, [1, biasShape[2], biasShape[0], biasShape[1]]);
                }
            }
            else if (dataFormat === 'channelsLast') {
                if (biasShape.length === 1) {
                    return reshape$3(bias, [1, 1, 1, biasShape[0]]);
                }
                else {
                    return reshape$3(bias, [1].concat(biasShape));
                }
            }
        }
        else if (xRank === 3) {
            if (dataFormat === 'channelsFirst') {
                if (biasShape.length === 1) {
                    return reshape$3(bias, [1, biasShape[0], 1]);
                }
                else {
                    return reshape$3(bias, [1, biasShape[1], biasShape[0]]);
                }
            }
            else if (dataFormat === 'channelsLast') {
                if (biasShape.length === 1) {
                    return reshape$3(bias, [1, 1, biasShape[0]]);
                }
                else {
                    return reshape$3(bias, [1].concat(biasShape));
                }
            }
        }
        else if (xRank < 3) {
            return bias;
        }
        throw new ValueError(`Unsupported input rank by biasAdd: ${bias.rank}`);
    }
    /* Neural-network operations. */
    /**
     * Add a bias to a tensor.
     *
     * @param x The tensor to add the bias to.
     * @param bias The bias to add to `x`. Must be 1D or the same rank as `x`.
     * @return Result of the bias adding.
     * @throws ValueError: If the rank of `bias` is incorrect.
     */
    function biasAdd(x, bias, dataFormat) {
        return tidy(() => {
            if (dataFormat == null) {
                dataFormat = imageDataFormat();
            }
            checkDataFormat(dataFormat);
            return add$2(x, reshapeBias(x.rank, bias, dataFormat));
        });
    }
    /**
     * Exponential linear unit (ELU).
     * @param x A tensor or variable to compute the activation function for.
     * @param alpha: A scalar, a scaling factor for the negative section.
     * @return Output of the ELU operation.
     */
    function elu$3(x, alpha = 1) {
        // TODO(cais): Add support for alpha values other than 1.
        if (alpha !== 1) {
            throw new NotImplementedError(`Support for alpha values other than 1 (${alpha}) is not implemented ` +
                `yet.`);
        }
        return elu$4(x);
    }
    /**
     * Softsign of a tensor.
     *
     * Defined as x / (abs(x) + 1), element-wise.
     *
     * @param x: Input.
     * @returns Output.
     */
    function softsign(x) {
        return tidy(() => div$1(x, add$2(abs$2(x), 1)));
    }
    /**
     * Sets entries in `x` to zero at random, while scaling the entire tensor.
     *
     * @param x input tensor.
     * @param level fraction of the entries in the tensor that will be set to 0.
     * @param noiseShape shape of randomly generated keep/drop flags, must be
     *   broadcastable to the shape of `x`. Optional.
     * @param seed random seed to ensure determinism. Optional.
     * @returns Result of the dropout operation.
     */
    function dropout$1(x, level, noiseShape, seed) {
        return tidy(() => dropout$2(x, level, noiseShape, seed));
    }
    /**
     * Element-wise, segment-wise linear approximation of sigmoid.
     *
     * Returns `0.` if `x < -2.5`, `1.` if `x > 2.5`.
     * In `-2.5 <= x <= 2.5`, returns `0.2 * x + 0.5`.
     *
     * @param x Input tensor.
     * @returns Output tensor.
     */
    function hardSigmoid(x) {
        return tidy(() => {
            const y = add$2(.5, mul(.2, x));
            return clipByValue$2(y, 0, 1);
        });
    }
    /**
     * Invoke `x` in the training phase, and `alt` otherwise.
     *
     * Porting Note: We do not create placeholder tensors for the `training`
     * boolean flag here, because there is no such thing in the TF.js imperative
     * backend.
     *
     * @param x The function to invoke iff `training` is `true`.
     * @param alt The function to invoke iff `training` is `false`.
     * @param training Boolean flag for whether training phase is active.
     * @returns The return value of `x()` if `training` is `true`, or the return
     *   value of `alt()` if `training` is `false`.
     */
    function inTrainPhase(x, alt, training = false) {
        return training ? x() : alt();
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    const VALID_FAN_MODE_VALUES = ['fanIn', 'fanOut', 'fanAvg'];
    const VALID_DISTRIBUTION_VALUES = ['normal', 'uniform', 'truncatedNormal'];

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    function checkFanMode(value) {
        checkStringTypeUnionValue(VALID_FAN_MODE_VALUES, 'FanMode', value);
    }
    function checkDistribution(value) {
        checkStringTypeUnionValue(VALID_DISTRIBUTION_VALUES, 'Distribution', value);
    }
    /**
     * Initializer base class.
     *
     * @doc {
     *   heading: 'Initializers', subheading: 'Classes', namespace: 'initializers'}
     */
    class Initializer extends Serializable {
        fromConfigUsesCustomObjects() {
            return false;
        }
        getConfig() {
            return {};
        }
    }
    class Zeros extends Initializer {
        apply(shape, dtype) {
            return zeros$2(shape, dtype);
        }
    }
    /** @nocollapse */
    Zeros.className = 'Zeros';
    registerClass(Zeros);
    class Ones extends Initializer {
        apply(shape, dtype) {
            return ones$1(shape, dtype);
        }
    }
    /** @nocollapse */
    Ones.className = 'Ones';
    registerClass(Ones);
    class Constant extends Initializer {
        constructor(args) {
            super();
            if (typeof args !== 'object') {
                throw new ValueError(`Expected argument of type ConstantConfig but got ${args}`);
            }
            if (args.value === undefined) {
                throw new ValueError(`config must have value set but got ${args}`);
            }
            this.value = args.value;
        }
        apply(shape, dtype) {
            return tidy(() => mul(scalar(this.value), ones$1(shape, dtype)));
        }
        getConfig() {
            return {
                value: this.value,
            };
        }
    }
    /** @nocollapse */
    Constant.className = 'Constant';
    registerClass(Constant);
    class RandomUniform extends Initializer {
        constructor(args) {
            super();
            this.DEFAULT_MINVAL = -0.05;
            this.DEFAULT_MAXVAL = 0.05;
            this.minval = args.minval || this.DEFAULT_MINVAL;
            this.maxval = args.maxval || this.DEFAULT_MAXVAL;
            this.seed = args.seed;
        }
        apply(shape, dtype) {
            return randomUniform$1(shape, this.minval, this.maxval, dtype);
        }
        getConfig() {
            return { minval: this.minval, maxval: this.maxval, seed: this.seed };
        }
    }
    /** @nocollapse */
    RandomUniform.className = 'RandomUniform';
    registerClass(RandomUniform);
    class RandomNormal extends Initializer {
        constructor(args) {
            super();
            this.DEFAULT_MEAN = 0.;
            this.DEFAULT_STDDEV = 0.05;
            this.mean = args.mean || this.DEFAULT_MEAN;
            this.stddev = args.stddev || this.DEFAULT_STDDEV;
            this.seed = args.seed;
        }
        apply(shape, dtype) {
            dtype = dtype || 'float32';
            if (dtype !== 'float32' && dtype !== 'int32') {
                throw new NotImplementedError(`randomNormal does not support dType ${dtype}.`);
            }
            return randomNormal$1(shape, this.mean, this.stddev, dtype, this.seed);
        }
        getConfig() {
            return { mean: this.mean, stddev: this.stddev, seed: this.seed };
        }
    }
    /** @nocollapse */
    RandomNormal.className = 'RandomNormal';
    registerClass(RandomNormal);
    class TruncatedNormal extends Initializer {
        constructor(args) {
            super();
            this.DEFAULT_MEAN = 0.;
            this.DEFAULT_STDDEV = 0.05;
            this.mean = args.mean || this.DEFAULT_MEAN;
            this.stddev = args.stddev || this.DEFAULT_STDDEV;
            this.seed = args.seed;
        }
        apply(shape, dtype) {
            dtype = dtype || 'float32';
            if (dtype !== 'float32' && dtype !== 'int32') {
                throw new NotImplementedError(`truncatedNormal does not support dType ${dtype}.`);
            }
            return truncatedNormal$1(shape, this.mean, this.stddev, dtype, this.seed);
        }
        getConfig() {
            return { mean: this.mean, stddev: this.stddev, seed: this.seed };
        }
    }
    /** @nocollapse */
    TruncatedNormal.className = 'TruncatedNormal';
    registerClass(TruncatedNormal);
    class Identity extends Initializer {
        constructor(args) {
            super();
            this.gain = args.gain != null ? args.gain : 1.0;
        }
        apply(shape, dtype) {
            return tidy(() => {
                if (shape.length !== 2 || shape[0] !== shape[1]) {
                    throw new ValueError('Identity matrix initializer can only be used for' +
                        ' 2D square matrices.');
                }
                else {
                    return mul(this.gain, eye(shape[0]));
                }
            });
        }
        getConfig() {
            return { gain: this.gain };
        }
    }
    /** @nocollapse */
    Identity.className = 'Identity';
    registerClass(Identity);
    /**
     * Computes the number of input and output units for a weight shape.
     * @param shape Shape of weight.
     * @param dataFormat data format to use for convolution kernels.
     *   Note that all kernels in Keras are standardized on the
     *   CHANNEL_LAST ordering (even when inputs are set to CHANNEL_FIRST).
     * @return An length-2 array: fanIn, fanOut.
     */
    function computeFans(shape, dataFormat = 'channelsLast') {
        let fanIn;
        let fanOut;
        checkDataFormat(dataFormat);
        if (shape.length === 2) {
            fanIn = shape[0];
            fanOut = shape[1];
        }
        else if ([3, 4, 5].indexOf(shape.length) !== -1) {
            if (dataFormat === 'channelsFirst') {
                const receptiveFieldSize = arrayProd(shape, 2);
                fanIn = shape[1] * receptiveFieldSize;
                fanOut = shape[0] * receptiveFieldSize;
            }
            else if (dataFormat === 'channelsLast') {
                const receptiveFieldSize = arrayProd(shape, 0, shape.length - 2);
                fanIn = shape[shape.length - 2] * receptiveFieldSize;
                fanOut = shape[shape.length - 1] * receptiveFieldSize;
            }
        }
        else {
            const shapeProd = arrayProd(shape);
            fanIn = Math.sqrt(shapeProd);
            fanOut = Math.sqrt(shapeProd);
        }
        return [fanIn, fanOut];
    }
    class VarianceScaling extends Initializer {
        /**
         * Constructor of VarianceScaling.
         * @throws ValueError for invalid value in scale.
         */
        constructor(args) {
            super();
            if (args.scale < 0.0) {
                throw new ValueError(`scale must be a positive float. Got: ${args.scale}`);
            }
            this.scale = args.scale == null ? 1.0 : args.scale;
            this.mode = args.mode == null ? 'fanIn' : args.mode;
            checkFanMode(this.mode);
            this.distribution =
                args.distribution == null ? 'normal' : args.distribution;
            checkDistribution(this.distribution);
            this.seed = args.seed;
        }
        apply(shape, dtype) {
            const fans = computeFans(shape);
            const fanIn = fans[0];
            const fanOut = fans[1];
            let scale = this.scale;
            if (this.mode === 'fanIn') {
                scale /= Math.max(1, fanIn);
            }
            else if (this.mode === 'fanOut') {
                scale /= Math.max(1, fanOut);
            }
            else {
                scale /= Math.max(1, (fanIn + fanOut) / 2);
            }
            if (this.distribution === 'normal') {
                const stddev = Math.sqrt(scale);
                dtype = dtype || 'float32';
                if (dtype !== 'float32' && dtype !== 'int32') {
                    throw new NotImplementedError(`${this.getClassName()} does not support dType ${dtype}.`);
                }
                return truncatedNormal$1(shape, 0, stddev, dtype, this.seed);
            }
            else {
                const limit = Math.sqrt(3 * scale);
                return randomUniform$1(shape, -limit, limit, dtype);
            }
        }
        getConfig() {
            return {
                scale: this.scale,
                mode: this.mode,
                distribution: this.distribution,
                seed: this.seed
            };
        }
    }
    /** @nocollapse */
    VarianceScaling.className = 'VarianceScaling';
    registerClass(VarianceScaling);
    class GlorotUniform extends VarianceScaling {
        /**
         * Constructor of GlorotUniform
         * @param scale
         * @param mode
         * @param distribution
         * @param seed
         */
        constructor(args) {
            super({
                scale: 1.0,
                mode: 'fanAvg',
                distribution: 'uniform',
                seed: args == null ? null : args.seed
            });
        }
        getClassName() {
            // In Python Keras, GlorotUniform is not a class, but a helper method
            // that creates a VarianceScaling object. Use 'VarianceScaling' as
            // class name to be compatible with that.
            return VarianceScaling.className;
        }
    }
    /** @nocollapse */
    GlorotUniform.className = 'GlorotUniform';
    registerClass(GlorotUniform);
    class GlorotNormal extends VarianceScaling {
        /**
         * Constructor of GlorotNormal.
         * @param scale
         * @param mode
         * @param distribution
         * @param seed
         */
        constructor(args) {
            super({
                scale: 1.0,
                mode: 'fanAvg',
                distribution: 'normal',
                seed: args == null ? null : args.seed
            });
        }
        getClassName() {
            // In Python Keras, GlorotNormal is not a class, but a helper method
            // that creates a VarianceScaling object. Use 'VarianceScaling' as
            // class name to be compatible with that.
            return VarianceScaling.className;
        }
    }
    /** @nocollapse */
    GlorotNormal.className = 'GlorotNormal';
    registerClass(GlorotNormal);
    class HeNormal extends VarianceScaling {
        constructor(args) {
            super({
                scale: 2.0,
                mode: 'fanIn',
                distribution: 'normal',
                seed: args == null ? null : args.seed
            });
        }
        getClassName() {
            // In Python Keras, HeNormal is not a class, but a helper method
            // that creates a VarianceScaling object. Use 'VarianceScaling' as
            // class name to be compatible with that.
            return VarianceScaling.className;
        }
    }
    /** @nocollapse */
    HeNormal.className = 'HeNormal';
    registerClass(HeNormal);
    class HeUniform extends VarianceScaling {
        constructor(args) {
            super({
                scale: 2.0,
                mode: 'fanIn',
                distribution: 'uniform',
                seed: args == null ? null : args.seed
            });
        }
        getClassName() {
            // In Python Keras, HeUniform is not a class, but a helper method
            // that creates a VarianceScaling object. Use 'VarianceScaling' as
            // class name to be compatible with that.
            return VarianceScaling.className;
        }
    }
    /** @nocollapse */
    HeUniform.className = 'HeUniform';
    registerClass(HeUniform);
    class LeCunNormal extends VarianceScaling {
        constructor(args) {
            super({
                scale: 1.0,
                mode: 'fanIn',
                distribution: 'normal',
                seed: args == null ? null : args.seed
            });
        }
        getClassName() {
            // In Python Keras, LeCunNormal is not a class, but a helper method
            // that creates a VarianceScaling object. Use 'VarianceScaling' as
            // class name to be compatible with that.
            return VarianceScaling.className;
        }
    }
    /** @nocollapse */
    LeCunNormal.className = 'LeCunNormal';
    registerClass(LeCunNormal);
    class LeCunUniform extends VarianceScaling {
        constructor(args) {
            super({
                scale: 1.0,
                mode: 'fanIn',
                distribution: 'uniform',
                seed: args == null ? null : args.seed
            });
        }
        getClassName() {
            // In Python Keras, LeCunUniform is not a class, but a helper method
            // that creates a VarianceScaling object. Use 'VarianceScaling' as
            // class name to be compatible with that.
            return VarianceScaling.className;
        }
    }
    /** @nocollapse */
    LeCunUniform.className = 'LeCunNormal';
    registerClass(LeCunUniform);
    class Orthogonal extends Initializer {
        constructor(args) {
            super();
            this.DEFAULT_GAIN = 1;
            this.gain = args.gain == null ? this.DEFAULT_GAIN : args.gain;
            this.seed = args.seed;
            if (this.seed != null) {
                throw new NotImplementedError('Random seed is not implemented for Orthogonal Initializer yet.');
            }
        }
        apply(shape, dtype) {
            return tidy(() => {
                if (shape.length < 2) {
                    throw new NotImplementedError('Shape must be at least 2D.');
                }
                if (shape[0] * shape[1] > 2000) {
                    console.warn(`Orthogonal initializer is being called on a matrix with more ` +
                        `than 2000 (${shape[0] * shape[1]}) elements: ` +
                        `Slowness may result.`);
                }
                // TODO(cais): Add seed support.
                const normalizedShape = shape[0] > shape[1] ? [shape[1], shape[0]] : shape;
                const a = randomNormal$1(normalizedShape, 0, 1, 'float32');
                let q = linalg.gramSchmidt(a);
                if (shape[0] > shape[1]) {
                    q = transpose$2(q);
                }
                return mul(this.gain, q);
            });
        }
        getConfig() {
            return {
                gain: this.gain,
                seed: this.seed,
            };
        }
    }
    /** @nocollapse */
    Orthogonal.className = 'Orthogonal';
    registerClass(Orthogonal);
    // Maps the JavaScript-like identifier keys to the corresponding registry
    // symbols.
    const INITIALIZER_IDENTIFIER_REGISTRY_SYMBOL_MAP = {
        'constant': 'Constant',
        'glorotNormal': 'GlorotNormal',
        'glorotUniform': 'GlorotUniform',
        'heNormal': 'HeNormal',
        'heUniform': 'HeUniform',
        'identity': 'Identity',
        'leCunNormal': 'LeCunNormal',
        'leCunUniform': 'LeCunUniform',
        'ones': 'Ones',
        'orthogonal': 'Orthogonal',
        'randomNormal': 'RandomNormal',
        'randomUniform': 'RandomUniform',
        'truncatedNormal': 'TruncatedNormal',
        'varianceScaling': 'VarianceScaling',
        'zeros': 'Zeros'
    };
    function deserializeInitializer(config, customObjects = {}) {
        return deserializeKerasObject(config, SerializationMap.getMap().classNameMap, customObjects, 'initializer');
    }
    function serializeInitializer(initializer) {
        return serializeKerasObject(initializer);
    }
    function getInitializer(identifier) {
        if (typeof identifier === 'string') {
            const className = identifier in INITIALIZER_IDENTIFIER_REGISTRY_SYMBOL_MAP ?
                INITIALIZER_IDENTIFIER_REGISTRY_SYMBOL_MAP[identifier] :
                identifier;
            /* We have four 'helper' classes for common initializers that
            all get serialized as 'VarianceScaling' and shouldn't go through
            the deserializeInitializer pathway. */
            if (className === 'GlorotNormal') {
                return new GlorotNormal();
            }
            else if (className === 'GlorotUniform') {
                return new GlorotUniform();
            }
            else if (className === 'HeNormal') {
                return new HeNormal();
            }
            else if (className === 'HeUniform') {
                return new HeUniform();
            }
            else if (className === 'LeCunNormal') {
                return new LeCunNormal();
            }
            else if (className === 'LeCunUniform') {
                return new LeCunUniform();
            }
            else {
                const config = {};
                config['className'] = className;
                config['config'] = {};
                return deserializeInitializer(config);
            }
        }
        else if (identifier instanceof Initializer) {
            return identifier;
        }
        else {
            return deserializeInitializer(identifier);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    // tslint:enable
    /**
     * Determine whether the input is an Array of Shapes.
     */
    function isArrayOfShapes(x) {
        return Array.isArray(x) && Array.isArray(x[0]);
    }
    /**
     * Special case of normalizing shapes to lists.
     *
     * @param x A shape or list of shapes to normalize into a list of Shapes.
     * @return A list of Shapes.
     */
    function normalizeShapeList(x) {
        if (x.length === 0) {
            return [];
        }
        if (!Array.isArray(x[0])) {
            return [x];
        }
        return x;
    }
    /**
     * Helper function to obtain exactly one Tensor.
     * @param xs: A single `tf.Tensor` or an `Array` of `tf.Tensor`s.
     * @return A single `tf.Tensor`. If `xs` is an `Array`, return the first one.
     * @throws ValueError: If `xs` is an `Array` and its length is not 1.
     */
    function getExactlyOneTensor(xs) {
        let x;
        if (Array.isArray(xs)) {
            if (xs.length !== 1) {
                throw new ValueError(`Expected Tensor length to be 1; got ${xs.length}`);
            }
            x = xs[0];
        }
        else {
            x = xs;
        }
        return x;
    }
    /**
     * Helper function to obtain exactly on instance of Shape.
     *
     * @param shapes Input single `Shape` or Array of `Shape`s.
     * @returns If input is a single `Shape`, return it unchanged. If the input is
     *   an `Array` containing exactly one instance of `Shape`, return the instance.
     *   Otherwise, throw a `ValueError`.
     * @throws ValueError: If input is an `Array` of `Shape`s, and its length is not
     *   1.
     */
    function getExactlyOneShape(shapes) {
        if (Array.isArray(shapes) && Array.isArray(shapes[0])) {
            if (shapes.length === 1) {
                shapes = shapes;
                return shapes[0];
            }
            else {
                throw new ValueError(`Expected exactly 1 Shape; got ${shapes.length}`);
            }
        }
        else {
            return shapes;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Count the elements in an Array of LayerVariables.
     *
     * @param weights: The LayerVariables of which the constituent numbers are to
     *   be counted.
     * @returns A count of the elements in all the LayerVariables
     */
    function countParamsInWeights(weights) {
        let count = 0;
        for (const weight of weights) {
            if (weight.shape.length === 0) {
                count += 1;
            }
            else {
                count += weight.shape.reduce((a, b) => a * b);
            }
        }
        return count;
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    const DEFAULT_VARIABLE_NAME_PREFIX = 'Variable';
    /**
     * A `tf.layers.LayerVariable` is similar to a `tf.Tensor` in that it has a
     * dtype and shape, but its value is mutable.  The value is itself represented
     * as a`tf.Tensor`, and can be read with the `read()` method and updated with
     * the `write()` method.
     */
    class LayerVariable {
        /**
         * Construct Variable from a `tf.Tensor`.
         *
         * If not explicitly named, the Variable will be given a name with the
         * prefix 'Variable'. Variable names are unique. In the case of name
         * collision, suffixies '_<num>' will be added to the name.
         *
         * @param val Initial value of the Variable.
         * @param name Name of the variable. If `null` or `undefined` is provided, it
         *   will default a name with the prefix 'Variable'.
         * @param constraint Optional, projection function to be applied to the
         * variable after optimize updates
         * @throws ValueError if `name` is `null` or `undefined`.
         */
        constructor(val, dtype = 'float32', name = DEFAULT_VARIABLE_NAME_PREFIX, trainable = true, constraint = null) {
            this.dtype = dtype == null ? 'float32' : dtype;
            this.shape = val.shape;
            this.id = getNextUniqueTensorId();
            name = name == null ? DEFAULT_VARIABLE_NAME_PREFIX : name;
            this.originalName = getScopedTensorName(name);
            this.name = getUniqueTensorName(this.originalName);
            this.trainable_ = trainable;
            this.constraint = constraint;
            this.val = variable(val, this.trainable_, this.name, this.dtype);
        }
        /**
         * Get a snapshot of the Variable's value.
         *
         * The returned value is a snapshot of the Variable's value at the time of
         * the invocation. Future mutations in the value of the tensor will only
         * be reflected by future calls to this method.
         */
        read() {
            this.assertNotDisposed();
            return this.val;
        }
        /**
         * Update the value of the Variable.
         *
         * @param newVal: The new value to update to. Must be consistent with the
         *   dtype and shape of the Variable.
         * @return This Variable.
         */
        write(newVal) {
            // TODO(cais): Once  TF.js Core supports Tensor.dtype, check dtype match.
            this.assertNotDisposed();
            checkShapesMatch(this.val, newVal);
            // Skip updating if this is the exact same tensor.
            if (this.val.id !== newVal.id) {
                this.val.assign(newVal);
                if (this.constraint != null) {
                    this.val.assign(this.constraint.apply(this.val));
                }
            }
            return this;
        }
        /**
         * Dispose this LayersVariable instance from memory.
         */
        dispose() {
            this.assertNotDisposed();
            this.val.dispose();
        }
        assertNotDisposed() {
            if (this.val.isDisposed) {
                throw new Error(`LayersVariable ${this.name} is already disposed.`);
            }
        }
        get trainable() {
            return this.trainable_;
        }
        set trainable(trainable) {
            this.trainable_ = trainable;
            this.val.trainable = trainable;
        }
    }
    function checkShapesMatch(x, y) {
        if (x.shape.toString() !== y.shape.toString()) {
            throw new Error('Shape mismatch: ' + JSON.stringify(x.shape) + ' vs. ' +
                JSON.stringify(y.shape));
        }
    }
    /**
     * Get the values of an array of Variables.
     *
     * @param tensors An `Array` of `Variable`s to get the values of.
     * @return The values of the inputs, as an `Array` of`tf.Tensor`s.
     */
    function batchGetValue(xs) {
        return xs.map(x => x.read());
    }
    /**
     * Update the value of multiple Variables at once.
     *
     * @param variablesAndValues An `Array`, each element is of type
     *   [Variable, Tensor]. The first item is the
     *   `Variable` of which the value is to be updated. The second item
     *   carries the new value.
     */
    function batchSetValue(variablesAndValues) {
        variablesAndValues.forEach(variableAndValue => {
            const variable = variableAndValue[0];
            variable.write(variableAndValue[1]);
        });
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Specifies the ndim, dtype and shape of every input to a layer.
     *
     * Every layer should expose (if appropriate) an `inputSpec` attribute:
     * a list of instances of InputSpec (one per input tensor).
     *
     * A null entry in a shape is compatible with any dimension,
     * a null shape is compatible with any shape.
     */
    class InputSpec {
        constructor(args) {
            this.dtype = args.dtype;
            this.shape = args.shape;
            /*
              TODO(michaelterry): Could throw error if ndim and shape are both defined
                (then backport).
            */
            if (args.shape != null) {
                this.ndim = args.shape.length;
            }
            else {
                this.ndim = args.ndim;
            }
            this.maxNDim = args.maxNDim;
            this.minNDim = args.minNDim;
            this.axes = args.axes || {};
        }
    }
    /**
     * `tf.SymbolicTensor` is a placeholder for a Tensor without any concrete value.
     *
     * They are most often encountered when building a graph of `Layer`s for a
     * a `tf.LayersModel` and the input data's shape, but not values are known.
     *
     * @doc {heading: 'Models', 'subheading': 'Classes'}
     */
    class SymbolicTensor {
        /**
         *
         * @param dtype
         * @param shape
         * @param sourceLayer The Layer that produced this symbolic tensor.
         * @param inputs The inputs passed to sourceLayer's __call__() method.
         * @param nodeIndex
         * @param tensorIndex
         * @param callArgs The keyword arguments passed to the __call__() method.
         * @param name
         * @param outputTensorIndex The index of this tensor in the list of outputs
         *   returned by apply().
         */
        constructor(dtype, shape, sourceLayer, inputs, callArgs, name, outputTensorIndex) {
            this.dtype = dtype;
            this.shape = shape;
            this.sourceLayer = sourceLayer;
            this.inputs = inputs;
            this.callArgs = callArgs;
            this.outputTensorIndex = outputTensorIndex;
            this.id = getNextUniqueTensorId();
            if (name != null) {
                this.originalName = getScopedTensorName(name);
                this.name = getUniqueTensorName(this.originalName);
            }
            this.rank = shape.length;
        }
    }
    let _nextNodeID = 0;
    /**
     * A `Node` describes the connectivity between two layers.
     *
     * Each time a layer is connected to some new input,
     * a node is added to `layer.inboundNodes`.
     *
     * Each time the output of a layer is used by another layer,
     * a node is added to `layer.outboundNodes`.
     *
     * `nodeIndices` and `tensorIndices` are basically fine-grained coordinates
     * describing the origin of the `inputTensors`, verifying the following:
     *
     * `inputTensors[i] ==
     * inboundLayers[i].inboundNodes[nodeIndices[i]].outputTensors[
     *   tensorIndices[i]]`
     *
     * A node from layer A to layer B is added to:
     *     A.outboundNodes
     *     B.inboundNodes
     */
    class Node {
        constructor(args, 
        // TODO(michaelterry): Define actual type for this.
        callArgs) {
            this.callArgs = callArgs;
            this.id = _nextNodeID++;
            /*
              Layer instance (NOT a list).
              this is the layer that takes a list of input tensors
              and turns them into a list of output tensors.
              the current node will be added to
              the inboundNodes of outboundLayer.
            */
            this.outboundLayer = args.outboundLayer;
            /*
                The following 3 properties describe where
                the input tensors come from: which layers,
                and for each layer, which node and which
                tensor output of each node.
            */
            // List of layer instances.
            this.inboundLayers = args.inboundLayers;
            // List of integers, 1:1 mapping with inboundLayers.
            this.nodeIndices = args.nodeIndices;
            // List of integers, 1:1 mapping with inboundLayers.
            this.tensorIndices = args.tensorIndices;
            /*
                Following 2 properties:
                tensor inputs and outputs of outboundLayer.
            */
            // List of tensors. 1:1 mapping with inboundLayers.
            this.inputTensors = args.inputTensors;
            // List of tensors, created by outboundLayer.call().
            this.outputTensors = args.outputTensors;
            /*
                Following 2 properties: input and output masks.
                List of tensors, 1:1 mapping with inputTensor.
            */
            this.inputMasks = args.inputMasks;
            // List of tensors, created by outboundLayer.computeMask().
            this.outputMasks = args.outputMasks;
            // Following 2 properties: input and output shapes.
            // List of shape tuples, shapes of inputTensors.
            this.inputShapes = args.inputShapes;
            // List of shape tuples, shapes of outputTensors.
            this.outputShapes = args.outputShapes;
            // Add nodes to all layers involved.
            for (const layer of args.inboundLayers) {
                if (layer != null) {
                    layer.outboundNodes.push(this);
                }
            }
            args.outboundLayer.inboundNodes.push(this);
        }
        getConfig() {
            const inboundNames = [];
            for (const layer of this.inboundLayers) {
                if (layer != null) {
                    inboundNames.push(layer.name);
                }
                else {
                    inboundNames.push(null);
                }
            }
            return {
                outboundLayer: this.outboundLayer ? this.outboundLayer.name : null,
                inboundLayers: inboundNames,
                nodeIndices: this.nodeIndices,
                tensorIndices: this.tensorIndices
            };
        }
    }
    let _nextLayerID = 0;
    /**
     * A layer is a grouping of operations and weights that can be composed to
     * create a `tf.LayersModel`.
     *
     * Layers are constructed by using the functions under the
     * [tf.layers](#Layers-Basic) namespace.
     *
     * @doc {heading: 'Layers', subheading: 'Classes', namespace: 'layers'}
     */
    class Layer extends Serializable {
        constructor(args = {}) {
            super();
            this._callHook = null;
            this._addedWeightNames = [];
            // Porting Notes: PyKeras does not have this property in this base Layer
            //   class. Instead lets Layer subclass set it dynamically and checks the
            //   value with `hasattr`. In tfjs-layers, we let this be a member of this
            //   base class.
            this._stateful = false;
            this.id = _nextLayerID++;
            this.activityRegularizer = null;
            this.inputSpec = null;
            this.supportsMasking = false;
            // These properties will be set upon call of this.build()
            this._trainableWeights = [];
            this._nonTrainableWeights = [];
            this._losses = [];
            this._updates = [];
            this._built = false;
            /*
              These lists will be filled via successive calls
              to this.addInboundNode().
             */
            this.inboundNodes = [];
            this.outboundNodes = [];
            let name = args.name;
            if (!name) {
                const prefix = this.getClassName();
                name = toSnakeCase(prefix) + '_' + getUid(prefix);
            }
            this.name = name;
            this.trainable_ = args.trainable == null ? true : args.trainable;
            if (args.inputShape != null || args.batchInputShape != null) {
                /*
                  In this case we will later create an input layer
                  to insert before the current layer
                 */
                let batchInputShape;
                if (args.batchInputShape != null) {
                    batchInputShape = args.batchInputShape;
                }
                else if (args.inputShape != null) {
                    let batchSize = null;
                    if (args.batchSize != null) {
                        batchSize = args.batchSize;
                    }
                    batchInputShape = [batchSize].concat(args.inputShape);
                }
                this.batchInputShape = batchInputShape;
                // Set dtype.
                let dtype = args.dtype;
                if (dtype == null) {
                    dtype = args.inputDType;
                }
                if (dtype == null) {
                    dtype = 'float32';
                }
                this.dtype = dtype;
            }
            if (args.weights != null) {
                this.initialWeights = args.weights;
            }
            else {
                this.initialWeights = null;
            }
            // The value of `_refCount` is initialized to null. When the layer is used
            // in a symbolic way for the first time, it will be set to 1.
            this._refCount = null;
            this.fastWeightInitDuringBuild = false;
        }
        /**
         * Converts a layer and its index to a unique (immutable type) name.
         * This function is used internally with `this.containerNodes`.
         * @param layer The layer.
         * @param nodeIndex The layer's position (e.g. via enumerate) in a list of
         *   nodes.
         *
         * @returns The unique name.
         */
        static nodeKey(layer, nodeIndex) {
            return layer.name + '_ib-' + nodeIndex.toString();
        }
        /**
         * Returns this.inboundNode at index nodeIndex.
         *
         * Porting note: This is a replacement for _get_node_attribute_at_index()
         * @param nodeIndex
         * @param attrName The name of the attribute related to request for this node.
         */
        getNodeAtIndex(nodeIndex, attrName) {
            if (this.inboundNodes.length === 0) {
                throw new RuntimeError('The layer has never been called ' +
                    `and thus has no defined ${attrName}.`);
            }
            if (this.inboundNodes.length <= nodeIndex) {
                throw new ValueError(`Asked to get ${attrName} at node ${nodeIndex}, ` +
                    `but the layer has only ${this.inboundNodes.length} inbound nodes.`);
            }
            return this.inboundNodes[nodeIndex];
        }
        /**
         * Retrieves the input tensor(s) of a layer at a given node.
         *
         * @param nodeIndex Integer, index of the node from which to retrieve the
         *   attribute. E.g. `nodeIndex=0` will correspond to the first time the layer
         *   was called.
         *
         * @return A tensor (or list of tensors if the layer has multiple inputs).
         */
        getInputAt(nodeIndex) {
            return singletonOrArray(this.getNodeAtIndex(nodeIndex, 'input').inputTensors);
        }
        /**
         * Retrieves the output tensor(s) of a layer at a given node.
         *
         * @param nodeIndex Integer, index of the node from which to retrieve the
         *   attribute. E.g. `nodeIndex=0` will correspond to the first time the layer
         *   was called.
         *
         * @return A tensor (or list of tensors if the layer has multiple outputs).
         */
        getOutputAt(nodeIndex) {
            return singletonOrArray(this.getNodeAtIndex(nodeIndex, 'output').outputTensors);
        }
        // Properties
        /**
         * Retrieves the input tensor(s) of a layer.
         *
         * Only applicable if the layer has exactly one inbound node,
         * i.e. if it is connected to one incoming layer.
         *
         * @return Input tensor or list of input tensors.
         *
         * @exception AttributeError if the layer is connected to more than one
         *   incoming layers.
         */
        get input() {
            if (this.inboundNodes.length > 1) {
                throw new AttributeError(`Layer ${this.name}` +
                    ' has multiple inbound nodes, ' +
                    'hence the notion of "layer input" ' +
                    'is ill-defined. ' +
                    'Use `getInputAt(nodeIndex)` instead.');
            }
            else if (this.inboundNodes.length === 0) {
                throw new AttributeError(`Layer ${this.name}` +
                    ' is not connected, no input to return.');
            }
            return singletonOrArray(this.getNodeAtIndex(0, 'input').inputTensors);
        }
        /**
         * Retrieves the output tensor(s) of a layer.
         *
         * Only applicable if the layer has exactly one inbound node,
         * i.e. if it is connected to one incoming layer.
         *
         * @return Output tensor or list of output tensors.
         *
         * @exception AttributeError if the layer is connected to more than one
         *   incoming layers.
         */
        get output() {
            if (this.inboundNodes.length === 0) {
                throw new AttributeError(`Layer ${this.name}` +
                    ' has no inbound nodes.');
            }
            if (this.inboundNodes.length > 1) {
                throw new AttributeError(`Layer ${this.name}` +
                    ' has multiple inbound nodes, ' +
                    'hence the notion of "layer output" ' +
                    'is ill-defined. ' +
                    'Use `getOutputAt(nodeIndex)` instead.');
            }
            return singletonOrArray(this.getNodeAtIndex(0, 'output').outputTensors);
        }
        get losses() {
            return this._losses;
        }
        /**
         * Retrieves the Layer's current loss values.
         *
         * Used for regularizers during training.
         */
        calculateLosses() {
            // Porting Node: This is an augmentation to Layer.loss in PyKeras.
            //   In PyKeras, Layer.loss returns symbolic tensors. Here a concrete
            //   Tensor (specifically Scalar) values are returned. This is due to the
            //   imperative backend.
            return this.losses.map(lossFn => lossFn());
        }
        get updates() {
            return this._updates;
        }
        get built() {
            return this._built;
        }
        set built(built) {
            this._built = built;
        }
        get trainable() {
            return this.trainable_;
        }
        set trainable(trainable) {
            this._trainableWeights.forEach(w => w.trainable = trainable);
            this.trainable_ = trainable;
        }
        get trainableWeights() {
            if (this.trainable_) {
                return this._trainableWeights.filter(w => w.trainable);
            }
            else {
                return [];
            }
        }
        set trainableWeights(weights) {
            this._trainableWeights = weights;
        }
        get nonTrainableWeights() {
            if (this.trainable) {
                return this._trainableWeights.filter(w => !w.trainable)
                    .concat(this._nonTrainableWeights);
            }
            else {
                return this._trainableWeights.concat(this._nonTrainableWeights);
            }
        }
        set nonTrainableWeights(weights) {
            this._nonTrainableWeights = weights;
        }
        /**
         * The concatenation of the lists trainableWeights and nonTrainableWeights
         * (in this order).
         */
        get weights() {
            return this.trainableWeights.concat(this.nonTrainableWeights);
        }
        get stateful() {
            return this._stateful;
        }
        /**
         * Reset the states of the layer.
         *
         * This method of the base Layer class is essentially a no-op.
         * Subclasses that are stateful (e.g., stateful RNNs) should override this
         * method.
         */
        resetStates() {
            if (!this.stateful) {
                throw new Error('Cannot call the resetStates() method of a non-stateful Layer ' +
                    'object.');
            }
        }
        /**
         * Checks compatibility between the layer and provided inputs.
         *
         * This checks that the tensor(s) `input`
         * verify the input assumptions of the layer
         * (if any). If not, exceptions are raised.
         *
         * @param inputs Input tensor or list of input tensors.
         *
         * @exception ValueError in case of mismatch between
         *   the provided inputs and the expectations of the layer.
         */
        assertInputCompatibility(inputs) {
            inputs = toList(inputs);
            if (this.inputSpec == null || this.inputSpec.length === 0) {
                return;
            }
            const inputSpec = toList(this.inputSpec);
            if (inputs.length !== inputSpec.length) {
                throw new ValueError(`Layer ${this.name} expects ${inputSpec.length} inputs, ` +
                    `but it received ${inputs.length} input tensors. ` +
                    `Input received: ${inputs}`);
            }
            for (let inputIndex = 0; inputIndex < inputs.length; inputIndex++) {
                const x = inputs[inputIndex];
                const spec = inputSpec[inputIndex];
                if (spec == null) {
                    continue;
                }
                // Check ndim.
                const ndim = x.rank;
                if (spec.ndim != null) {
                    if (ndim !== spec.ndim) {
                        throw new ValueError(`Input ${inputIndex} is incompatible with layer ${this.name}: ` +
                            `expected ndim=${spec.ndim}, found ndim=${ndim}`);
                    }
                }
                if (spec.maxNDim != null) {
                    if (ndim > spec.maxNDim) {
                        throw new ValueError(`Input ${inputIndex} is incompatible with layer ${this.name}` +
                            `: expected max_ndim=${spec.maxNDim}, found ndim=${ndim}`);
                    }
                }
                if (spec.minNDim != null) {
                    if (ndim < spec.minNDim) {
                        throw new ValueError(`Input ${inputIndex} is incompatible with layer ${this.name}` +
                            `: expected min_ndim=${spec.minNDim}, found ndim=${ndim}.`);
                    }
                }
                // Check dtype.
                if (spec.dtype != null) {
                    if (x.dtype !== spec.dtype) {
                        throw new ValueError(`Input ${inputIndex} is incompatible with layer ${this.name} ` +
                            `: expected dtype=${spec.dtype}, found dtype=${x.dtype}.`);
                    }
                }
                // Check specific shape axes.
                if (spec.axes) {
                    const xShape = x.shape;
                    for (const key in spec.axes) {
                        const axis = Number(key);
                        const value = spec.axes[key];
                        // Perform Python-style slicing in case axis < 0;
                        // TODO(cais): Use https://github.com/alvivi/typescript-underscore to
                        // ensure type safety through Underscore calls.
                        const xShapeAtAxis = axis >= 0 ? xShape[axis] : xShape[xShape.length + axis];
                        if (value != null && [value, null].indexOf(xShapeAtAxis) === -1) {
                            throw new ValueError(`Input ${inputIndex} is incompatible with layer ` +
                                `${this.name}: expected axis ${axis} of input shape to ` +
                                `have value ${value} but got shape ${xShape}.`);
                        }
                    }
                }
                // Check shape.
                if (spec.shape != null) {
                    for (let i = 0; i < spec.shape.length; ++i) {
                        const specDim = spec.shape[i];
                        const dim = x.shape[i];
                        if (specDim != null && dim != null) {
                            if (specDim !== dim) {
                                throw new ValueError(`Input ${inputIndex} is incompatible with layer ` +
                                    `${this.name}: expected shape=${spec.shape}, ` +
                                    `found shape=${x.shape}.`);
                            }
                        }
                    }
                }
            }
        }
        /**
         * This is where the layer's logic lives.
         *
         * @param inputs Input tensor, or list/tuple of input tensors.
         * @param kwargs Additional keyword arguments.
         *
         * @return A tensor or list/tuple of tensors.
         */
        call(inputs, kwargs) {
            return inputs;
        }
        invokeCallHook(inputs, kwargs) {
            if (this._callHook != null) {
                this._callHook(inputs, kwargs);
            }
        }
        /**
         * Set call hook.
         * This is currently used for testing only.
         * @param callHook
         */
        setCallHook(callHook) {
            this._callHook = callHook;
        }
        /**
         * Clear call hook.
         * This is currently used for testing only.
         */
        clearCallHook() {
            this._callHook = null;
        }
        /**
         * Builds or executes a `Layer's logic.
         *
         * When called with `tf.Tensor`(s), execute the `Layer`s computation and
         * return Tensor(s). For example:
         *
         * ```js
         * const denseLayer = tf.layers.dense({
         *   units: 1,
         *   kernelInitializer: 'zeros',
         *   useBias: false
         * });
         *
         * // Invoke the layer's apply() method with a `tf.Tensor` (with concrete
         * // numeric values).
         * const input = tf.ones([2, 2]);
         * const output = denseLayer.apply(input);
         *
         * // The output's value is expected to be [[0], [0]], due to the fact that
         * // the dense layer has a kernel initialized to all-zeros and does not have
         * // a bias.
         * output.print();
         * ```
         *
         * When called with `tf.SymbolicTensor`(s), this will prepare the layer for
         * future execution.  This entails internal book-keeping on shapes of
         * expected Tensors, wiring layers together, and initializing weights.
         *
         * Calling `apply` with `tf.SymbolicTensor`s are typically used during the
         * building of non-`tf.Sequential` models. For example:
         *
         * ```js
         * const flattenLayer = tf.layers.flatten();
         * const denseLayer = tf.layers.dense({units: 1});
         *
         * // Use tf.layers.input() to obtain a SymbolicTensor as input to apply().
         * const input = tf.input({shape: [2, 2]});
         * const output1 = flattenLayer.apply(input);
         *
         * // output1.shape is [null, 4]. The first dimension is the undetermined
         * // batch size. The second dimension comes from flattening the [2, 2]
         * // shape.
         * console.log(JSON.stringify(output1.shape));
         *
         * // The output SymbolicTensor of the flatten layer can be used to call
         * // the apply() of the dense layer:
         * const output2 = denseLayer.apply(output1);
         *
         * // output2.shape is [null, 1]. The first dimension is the undetermined
         * // batch size. The second dimension matches the number of units of the
         * // dense layer.
         * console.log(JSON.stringify(output2.shape));
         *
         * // The input and output and be used to construct a model that consists
         * // of the flatten and dense layers.
         * const model = tf.model({inputs: input, outputs: output2});
         * ```
         *
         * @param inputs a `tf.Tensor` or `tf.SymbolicTensor` or an Array of them.
         * @param kwargs Additional keyword arguments to be passed to `call()`.
         *
         * @return Output of the layer's `call` method.
         *
         * @exception ValueError error in case the layer is missing shape information
         *   for its `build` call.
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        // Porting Note: This is a replacement for __call__() in Python.
        apply(inputs, kwargs) {
            kwargs = kwargs || {};
            this.assertNotDisposed();
            // Ensure inputs are all the same type.
            const inputsList = toList(inputs);
            let allAreSymbolic = true;
            for (const input of inputsList) {
                if (!(input instanceof SymbolicTensor)) {
                    allAreSymbolic = false;
                    break;
                }
            }
            let noneAreSymbolic = true;
            for (const input of inputsList) {
                if (input instanceof SymbolicTensor) {
                    noneAreSymbolic = false;
                    break;
                }
            }
            if (allAreSymbolic === noneAreSymbolic) {
                throw new ValueError('Arguments to apply() must be all ' +
                    'SymbolicTensors or all Tensors');
            }
            // TODO(michaelterry): nameScope() may not be necessary.
            return nameScope(this.name, () => {
                // Handle laying building (weight creating, input spec locking).
                if (!this.built) {
                    /*
                      Throw exceptions in case the input is not compatible
                      with the inputSpec specified in the layer constructor.
                     */
                    this.assertInputCompatibility(inputs);
                    // Collect input shapes to build layer.
                    const inputShapes = [];
                    for (const xElem of toList(inputs)) {
                        inputShapes.push(xElem.shape);
                    }
                    this.build(singletonOrArray(inputShapes));
                    this.built = true;
                    // Load weights that were specified at layer instantiation.
                    if (this.initialWeights) {
                        this.setWeights(this.initialWeights);
                    }
                    if (this._refCount === null && noneAreSymbolic) {
                        // The first use of this layer is a non-symbolic call, set ref count
                        // to 1 so the Layer can be properly disposed if its dispose() method
                        // is called.
                        this._refCount = 1;
                    }
                }
                /*
                  Throw exceptions in case the input is not compatible
                  with the inputSpec set at build time.
                */
                this.assertInputCompatibility(inputs);
                // Handle mask propagation.
                // TODO(michaelterry): Mask propagation not currently implemented.
                // Actually call the layer, collecting output(s), mask(s), and shape(s).
                if (noneAreSymbolic) {
                    let output = this.call(inputs, kwargs);
                    // TODO(michaelterry): Compute the outputMask
                    // If the layer returns tensors from its inputs, unmodified,
                    // we copy them to avoid loss of tensor metadata.
                    const outputList = toList(output);
                    const outputListCopy = [];
                    // TODO(michaelterry): This copying may not be necessary given our eager
                    // backend.
                    for (let x of outputList) {
                        if (inputsList.indexOf(x) !== -1) {
                            x = x.clone();
                        }
                        outputListCopy.push(x);
                    }
                    output = singletonOrArray(outputListCopy);
                    if (this.activityRegularizer != null) {
                        throw new NotImplementedError('Layer invocation in the presence of activity ' +
                            'regularizer(s) is not supported yet.');
                    }
                    // TODO(michaelterry): Call addInboundNode()?
                    return output;
                }
                else {
                    const inputShape = collectInputShape(inputs);
                    const outputShape = this.computeOutputShape(inputShape);
                    let output;
                    const outputDType = guessOutputDType(inputs);
                    this.warnOnIncompatibleInputShape(Array.isArray(inputs) ? inputShape[0] :
                        inputShape);
                    if (outputShape != null && outputShape.length > 0 &&
                        Array.isArray(outputShape[0])) {
                        // We have multiple output shapes. Create multiple output tensors.
                        output = outputShape
                            .map((shape, index) => new SymbolicTensor(outputDType, shape, this, toList(inputs), kwargs, this.name, index));
                    }
                    else {
                        output = new SymbolicTensor(outputDType, outputShape, this, toList(inputs), kwargs, this.name);
                    }
                    /*
                      Add an inbound node to the layer, so that it keeps track
                      of the call and of all new variables created during the call.
                      This also updates the layer history of the output tensor(s).
                      If the input tensor(s) had no previous history,
                      this does nothing.
                    */
                    this.addInboundNode(inputs, output, null, null, inputShape, outputShape, kwargs);
                    this._refCount++;
                    if (this.activityRegularizer != null) {
                        throw new NotImplementedError('Layer invocation in the presence of activity ' +
                            'regularizer(s) is not supported yet.');
                    }
                    return output;
                }
            });
        }
        /**
         * Check compatibility between input shape and this layer's batchInputShape.
         *
         * Print warning if any incompatibility is found.
         *
         * @param inputShape Input shape to be checked.
         */
        warnOnIncompatibleInputShape(inputShape) {
            if (this.batchInputShape == null) {
                return;
            }
            else if (inputShape.length !== this.batchInputShape.length) {
                console.warn(`The rank of the input tensor provided (shape: ` +
                    `${JSON.stringify(inputShape)}) does not match that of the ` +
                    `batchInputShape (${JSON.stringify(this.batchInputShape)}) ` +
                    `of the layer ${this.name}`);
            }
            else {
                let dimMismatch = false;
                this.batchInputShape.forEach((dimension, i) => {
                    if (dimension != null && inputShape[i] != null &&
                        inputShape[i] !== dimension) {
                        dimMismatch = true;
                    }
                });
                if (dimMismatch) {
                    console.warn(`The shape of the input tensor ` +
                        `(${JSON.stringify(inputShape)}) does not ` +
                        `match the expectation of layer ${this.name}: ` +
                        `${JSON.stringify(this.batchInputShape)}`);
                }
            }
        }
        /**
         * Retrieves the output shape(s) of a layer.
         *
         * Only applicable if the layer has only one inbound node, or if all inbound
         * nodes have the same output shape.
         *
         * @returns Output shape or shapes.
         * @throws AttributeError: if the layer is connected to more than one incoming
         *   nodes.
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        get outputShape() {
            if (this.inboundNodes == null || this.inboundNodes.length === 0) {
                throw new AttributeError(`The layer ${this.name} has never been called and thus has no ` +
                    `defined output shape.`);
            }
            const allOutputShapes = [];
            for (const node of this.inboundNodes) {
                const shapeString = JSON.stringify(node.outputShapes);
                if (allOutputShapes.indexOf(shapeString) === -1) {
                    allOutputShapes.push(shapeString);
                }
            }
            if (allOutputShapes.length === 1) {
                const outputShapes = this.inboundNodes[0].outputShapes;
                if (Array.isArray(outputShapes) && Array.isArray(outputShapes[0]) &&
                    outputShapes.length === 1) {
                    return outputShapes[0];
                }
                else {
                    return outputShapes;
                }
            }
            else {
                throw new AttributeError(`The layer ${this.name} has multiple inbound nodes with different ` +
                    `output shapes. Hence the notion of "output shape" is ill-defined ` +
                    `for the layer.`);
                // TODO(cais): Implement getOutputShapeAt().
            }
        }
        /**
         * Counts the total number of numbers (e.g., float32, int32) in the
         * weights.
         *
         * @returns An integer count.
         * @throws RuntimeError: If the layer is not built yet (in which case its
         *   weights are not defined yet.)
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        countParams() {
            if (!this.built) {
                throw new RuntimeError(`You tried to call countParams() on ${this.name}, ` +
                    `but the layer is not built yet. Build it first by calling ` +
                    `build(batchInputShape).`);
            }
            return countParamsInWeights(this.weights);
        }
        /**
         * Creates the layer weights.
         *
         * Must be implemented on all layers that have weights.
         *
         * Called when apply() is called to construct the weights.
         *
         * @param inputShape A `Shape` or array of `Shape` (unused).
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        build(inputShape) {
            this.built = true;
        }
        /**
         * Returns the current values of the weights of the layer.
         *
         * @param trainableOnly Whether to get the values of only trainable weights.
         * @returns Weight values as an `Array` of `tf.Tensor`s.
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        getWeights(trainableOnly = false) {
            return batchGetValue(trainableOnly ? this.trainableWeights : this.weights);
        }
        /**
         * Sets the weights of the layer, from Tensors.
         *
         * @param weights a list of Tensors. The number of arrays and their shape
         *   must match number of the dimensions of the weights of the layer (i.e.
         *   it should match the output of `getWeights`).
         *
         * @exception ValueError If the provided weights list does not match the
         *   layer's specifications.
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        setWeights(weights) {
            tidy(() => {
                const params = this.weights;
                if (params.length !== weights.length) {
                    // TODO(cais): Restore the following and use `providedWeights`, instead
                    // of `weights` in the error message, once the deeplearn.js bug is
                    // fixed: https://github.com/PAIR-code/deeplearnjs/issues/498 const
                    // providedWeights = JSON.stringify(weights).slice(0, 50);
                    throw new ValueError(`You called setWeights(weights) on layer "${this.name}" ` +
                        `with a weight list of length ${weights.length}, ` +
                        `but the layer was expecting ${params.length} weights. ` +
                        `Provided weights: ${weights}...`);
                }
                if (params.length === 0) {
                    return;
                }
                const weightValueTuples = [];
                const paramValues = batchGetValue(params);
                for (let i = 0; i < paramValues.length; ++i) {
                    const pv = paramValues[i];
                    const p = params[i];
                    const w = weights[i];
                    if (!arraysEqual(pv.shape, w.shape)) {
                        throw new ValueError(`Layer weight shape ${pv.shape} ` +
                            `not compatible with provided weight shape ${w.shape}`);
                    }
                    weightValueTuples.push([p, w]);
                }
                batchSetValue(weightValueTuples);
            });
        }
        /**
         * Adds a weight variable to the layer.
         *
         * @param name Name of the new weight variable.
         * @param shape The shape of the weight.
         * @param dtype The dtype of the weight.
         * @param initializer An initializer instance.
         * @param regularizer A regularizer instance.
         * @param trainable Whether the weight should be trained via backprop or not
         *   (assuming that the layer itself is also trainable).
         * @param constraint An optional trainable.
         * @return The created weight variable.
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        addWeight(name, shape, dtype, initializer, regularizer, trainable, constraint, getInitializerFunc) {
            // Reject duplicate weight names.
            if (this._addedWeightNames.indexOf(name) !== -1) {
                throw new ValueError(`Duplicate weight name ${name} for layer ${this.name}`);
            }
            this._addedWeightNames.push(name);
            if (dtype == null) {
                dtype = 'float32';
            }
            if (this.fastWeightInitDuringBuild) {
                initializer = getInitializerFunc != null ? getInitializerFunc() :
                    getInitializer('zeros');
            }
            const initValue = initializer.apply(shape, dtype);
            const weight = new LayerVariable(initValue, dtype, name, trainable, constraint);
            initValue.dispose();
            // Request backend not to dispose the weights of the model on scope() exit.
            if (regularizer != null) {
                this.addLoss(() => regularizer.apply(weight.read()));
            }
            if (trainable == null) {
                trainable = true;
            }
            if (trainable) {
                this._trainableWeights.push(weight);
            }
            else {
                this._nonTrainableWeights.push(weight);
            }
            return weight;
        }
        /**
         * Set the fast-weight-initialization flag.
         *
         * In cases where the initialized weight values will be immediately
         * overwritten by loaded weight values during model loading, setting
         * the flag to `true` saves unnecessary calls to potentially expensive
         * initializers and speeds up the loading process.
         *
         * @param value Target value of the flag.
         */
        setFastWeightInitDuringBuild(value) {
            this.fastWeightInitDuringBuild = value;
        }
        /**
         * Add losses to the layer.
         *
         * The loss may potentionally be conditional on some inputs tensors,
         * for instance activity losses are conditional on the layer's inputs.
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        addLoss(losses) {
            if (losses == null || Array.isArray(losses) && losses.length === 0) {
                return;
            }
            // Update this.losses
            losses = toList(losses);
            if (this._losses !== undefined && this._losses !== null) {
                this.losses.push(...losses);
            }
        }
        /**
         * Computes the output shape of the layer.
         *
         * Assumes that the layer will be built to match that input shape provided.
         *
         * @param inputShape A shape (tuple of integers) or a list of shape tuples
         *   (one per output tensor of the layer). Shape tuples can include null for
         *   free dimensions, instead of an integer.
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        computeOutputShape(inputShape) {
            return inputShape;
        }
        /**
         * Computes an output mask tensor.
         *
         * @param inputs Tensor or list of tensors.
         * @param mask Tensor or list of tensors.
         *
         * @return null or a tensor (or list of tensors, one per output tensor of the
         * layer).
         */
        computeMask(inputs, mask) {
            if (!this.supportsMasking) {
                if (mask != null) {
                    if (Array.isArray(mask)) {
                        mask.forEach(maskElement => {
                            if (maskElement != null) {
                                throw new TypeError(`Layer ${this.name} does not support masking, ` +
                                    'but was passed an inputMask.');
                            }
                        });
                    }
                    else {
                        throw new TypeError(`Layer ${this.name} does not support masking, ` +
                            'but was passed an inputMask.');
                    }
                }
                // masking not explicitly supported: return null as mask
                return null;
            }
            // if masking is explictly supported, by default
            // carry over the input mask
            return mask;
        }
        /**
         * Internal method to create an inbound node for the layer.
         *
         * @param inputTensors List of input tensors.
         * @param outputTensors List of output tensors.
         * @param inputMasks List of input masks (a mask can be a tensor, or null).
         * @param outputMasks List of output masks (a mask can be a tensor, or null).
         * @param inputShapes List of input shape tuples.
         * @param outputShapes List of output shape tuples.
         * @param kwargs Dictionary of keyword arguments that were passed to the
         *   `call` method of the layer at the call that created the node.
         */
        addInboundNode(inputTensors, outputTensors, inputMasks, outputMasks, inputShapes, outputShapes, kwargs = null) {
            const inputTensorList = toList(inputTensors);
            outputTensors = toList(outputTensors);
            inputMasks = toList(inputMasks);
            outputMasks = toList(outputMasks);
            inputShapes = normalizeShapeList(inputShapes);
            outputShapes = normalizeShapeList(outputShapes);
            // Collect input tensor(s) coordinates.
            const inboundLayers = [];
            const nodeIndices = [];
            const tensorIndices = [];
            for (const x of inputTensorList) {
                /*
                 * TODO(michaelterry): Keras adds this value to tensors; it's not
                 * clear whether we'll use this or not.
                 */
                inboundLayers.push(x.sourceLayer);
                nodeIndices.push(x.nodeIndex);
                tensorIndices.push(x.tensorIndex);
            }
            // Create node, add it to inbound nodes.
            // (This call has side effects.)
            // tslint:disable-next-line:no-unused-expression
            new Node({
                outboundLayer: this,
                inboundLayers,
                nodeIndices,
                tensorIndices,
                inputTensors: inputTensorList,
                outputTensors,
                inputMasks,
                outputMasks,
                inputShapes,
                outputShapes
            }, kwargs);
            // Update tensor history
            for (let i = 0; i < outputTensors.length; i++) {
                // TODO(michaelterry: _uses_learning_phase not tracked.
                outputTensors[i].sourceLayer = this;
                outputTensors[i].nodeIndex = this.inboundNodes.length - 1;
                outputTensors[i].tensorIndex = i;
            }
        }
        /**
         * Returns the config of the layer.
         *
         * A layer config is a TS dictionary (serializable)
         * containing the configuration of a layer.
         * The same layer can be reinstantiated later
         * (without its trained weights) from this configuration.
         *
         * The config of a layer does not include connectivity
         * information, nor the layer class name.  These are handled
         * by 'Container' (one layer of abstraction above).
         *
         * Porting Note: The TS dictionary follows TS naming standrds for
         * keys, and uses tfjs-layers type-safe Enums.  Serialization methods
         * should use a helper function to convert to the pythonic storage
         * standard. (see serialization_utils.convertTsToPythonic)
         *
         * @returns TS dictionary of configuration.
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        getConfig() {
            const config = { name: this.name, trainable: this.trainable };
            if (this.batchInputShape != null) {
                config['batchInputShape'] = this.batchInputShape;
            }
            if (this.dtype != null) {
                config['dtype'] = this.dtype;
            }
            return config;
        }
        /**
         * Dispose the weight variables that this Layer instance holds.
         *
         * @returns {number} Number of disposed variables.
         */
        disposeWeights() {
            this.weights.forEach(weight => weight.dispose());
            return this.weights.length;
        }
        assertNotDisposed() {
            if (this._refCount === 0) {
                throw new Error(`Layer '${this.name}' is already disposed.`);
            }
        }
        /**
         * Attempt to dispose layer's weights.
         *
         * This method decrease the reference count of the Layer object by 1.
         *
         * A Layer is reference-counted. Its reference count is incremented by 1
         * the first item its `apply()` method is called and when it becomes a part
         * of a new `Node` (through calling the `apply()`) method on a
         * `tf.SymbolicTensor`).
         *
         * If the reference count of a Layer becomes 0, all the weights will be
         * disposed and the underlying memory (e.g., the textures allocated in WebGL)
         * will be freed.
         *
         * Note: If the reference count is greater than 0 after the decrement, the
         * weights of the Layer will *not* be disposed.
         *
         * After a Layer is disposed, it cannot be used in calls such as `apply()`,
         * `getWeights()` or `setWeights()` anymore.
         *
         * @returns A DisposeResult Object with the following fields:
         *   - refCountAfterDispose: The reference count of the Container after this
         *     `dispose()` call.
         *   - numDisposedVariables: Number of `tf.Variable`s (i.e., weights) disposed
         *     during this `dispose()` call.
         * @throws {Error} If the layer is not built yet, or if the layer has already
         *   been disposed.
         *
         * @doc {heading: 'Models', 'subheading': 'Classes'}
         */
        dispose() {
            if (!this.built) {
                throw new Error(`Cannot dispose Layer ${this.name} because it has not been ` +
                    `built yet.`);
            }
            if (this._refCount === null) {
                throw new Error(`Cannot dispose Layer ${this.name} because it has not been used ` +
                    `yet.`);
            }
            this.assertNotDisposed();
            let numDisposedVariables = 0;
            if (--this._refCount === 0) {
                numDisposedVariables = this.disposeWeights();
            }
            return { refCountAfterDispose: this._refCount, numDisposedVariables };
        }
    }
    /**
     * Collects the input shape(s) of a list of `tf.Tensor`s or
     * `tf.SymbolicTensor`s.
     *
     * TODO(michaelterry): Update PyKeras docs (backport).
     *
     * @param inputTensors List of input tensors (or single input tensor).
     *
     * @return List of shape tuples (or single tuple), one tuple per input.
     */
    function collectInputShape(inputTensors) {
        inputTensors =
            toList(inputTensors);
        const shapes = [];
        for (const x of inputTensors) {
            shapes.push(x.shape);
        }
        return singletonOrArray(shapes);
    }
    /**
     * Guesses output dtype based on inputs.
     *
     * At present, just returns 'float32' for any input.
     *
     * @param inputTensors List of input tensors (or single input tensor).
     *
     * @return The guessed DType. At present, always returns 'float32'.
     */
    function guessOutputDType(inputTensors) {
        return 'float32';
    }
    /**
     * Returns the list of input tensors necessary to compute `tensor`.
     *
     * Output will always be a list of tensors (potentially with 1 element).
     *
     * @param tensor The tensor to start from.
     * @param layer Origin layer of the tensor.
     * @param nodeIndex Origin node index of the tensor.
     *
     * @return Array of input tensors.
     */
    function getSourceInputs(tensor, layer, nodeIndex) {
        if (layer == null || (nodeIndex != null && nodeIndex > 0)) {
            layer = tensor.sourceLayer;
            nodeIndex = tensor.nodeIndex;
        }
        if (layer.inboundNodes.length === 0) {
            return [tensor];
        }
        else {
            const node = layer.inboundNodes[nodeIndex];
            if (node.inboundLayers.length === 0) {
                return node.inputTensors;
            }
            else {
                const sourceTensors = [];
                for (let i = 0; i < node.inboundLayers.length; i++) {
                    const x = node.inputTensors[i];
                    const layer = node.inboundLayers[i];
                    const nodeIndex = node.nodeIndices[i];
                    const previousSources = getSourceInputs(x, layer, nodeIndex);
                    // Avoid input redundancy.
                    for (const x of previousSources) {
                        if (sourceTensors.indexOf(x) === -1) {
                            sourceTensors.push(x);
                        }
                    }
                }
                return sourceTensors;
            }
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    class InputLayer extends Layer {
        constructor(args) {
            super({
                dtype: args.dtype,
                name: args.name != null ? args.name : getUid('input').toString()
            });
            // Normalize config.batchSize and config.sparse
            if (args.batchSize == null) {
                args.batchSize = null;
            }
            if (args.sparse == null) {
                args.sparse = false;
            }
            this.trainable = false;
            this.built = true;
            this.sparse = args.sparse;
            if (args.inputShape != null && args.batchInputShape != null) {
                throw new ValueError('Only provide the inputShape OR ' +
                    'batchInputShape argument to inputLayer, not both at the same time.');
            }
            let batchInputShape = args.batchInputShape;
            if (batchInputShape == null) {
                if (args.inputShape == null) {
                    throw new ValueError('An InputLayer should be passed either a ' +
                        '`batchInputShape` or an `inputShape`.');
                }
                else {
                    batchInputShape = [args.batchSize].concat(args.inputShape);
                }
            }
            else {
                // TODO(michaelterry): Backport to PyKeras
                if (args.batchSize != null) {
                    throw new ValueError('Cannot specify batchSize if batchInputShape is ' +
                        'specified when creating an InputLayer.');
                }
            }
            const dtype = args.dtype || 'float32';
            this.batchInputShape = batchInputShape;
            this.dtype = dtype;
            // TODO(michaelterry): Backport this to PyKeras?
            this.inputSpec = [{ shape: batchInputShape }];
            const inputTensor = new SymbolicTensor(this.dtype, this.batchInputShape, this, [], {}, this.name);
            inputTensor.nodeIndex = 0;
            inputTensor.tensorIndex = 0;
            // Create an input node to add to this.outboundNode.
            // (This call has side effects.)
            // tslint:disable-next-line:no-unused-expression
            new Node({
                outboundLayer: this,
                inboundLayers: [],
                nodeIndices: [],
                tensorIndices: [],
                inputTensors: [inputTensor],
                outputTensors: [inputTensor],
                inputMasks: [null],
                outputMasks: [null],
                inputShapes: [batchInputShape],
                outputShapes: [batchInputShape]
            });
        }
        apply(inputs, kwargs) {
            throw new ValueError('Cannot pass any input to an ' +
                `InputLayer's apply() method. InputLayer name: ${this.name}`);
        }
        dispose() {
            // dispose() for InputLayer is overridden as no-op.
            return { refCountAfterDispose: this._refCount, numDisposedVariables: 0 };
        }
        getConfig() {
            return {
                batchInputShape: this.batchInputShape,
                dtype: this.dtype,
                sparse: this.sparse,
                name: this.name
            };
        }
    }
    /** @nocollapse */
    InputLayer.className = 'InputLayer';
    registerClass(InputLayer);
    function Input(config) {
        if (config.batchShape == null && config.shape == null) {
            throw new Error('Please provide to Input either a `shape`' +
                ' or a `batchShape` argument. Note that ' +
                '`shape` does not include the batch ' +
                'dimension.');
        }
        if (config.batchShape != null && config.shape != null) {
            // TODO(michaelterry): Backport to PyKeras.
            throw new ValueError('Please provide either a `shape` or `batchShape` ' +
                'argument to Input, but not both.');
        }
        let batchShape = config.batchShape;
        if (config.shape != null && batchShape == null) {
            batchShape = [null].concat(config.shape);
        }
        let dtype = config.dtype;
        if (dtype == null) {
            dtype = 'float32';
        }
        const inputLayer = new InputLayer({
            batchInputShape: batchShape,
            name: config.name,
            dtype,
            sparse: config.sparse
        });
        const outputs = inputLayer.inboundNodes[0].outputTensors;
        return outputs[0];
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Helper function to check the dtype and shape compatibility of a feed value.
     */
    function assertFeedCompatibility(key, val) {
        // Check dtype compatibility.
        if (key.dtype == null || key.dtype === val.dtype) {
            //  a.  If types match, return val tensor as is.
            return val;
        }
        try {
            //  b. Attempt to convert to expected type.
            return cast$3(val, key.dtype);
        }
        catch (err) {
            //  c. If conversion fails, return helpful error.
            throw new ValueError(`The dtype of the feed (${val.dtype}) can not be cast to the dtype ` +
                `of the key '${key.name}' (${key.dtype}).`);
        }
    }
    /**
     * FeedDict: A mapping from unique SymbolicTensors to feed values for them.
     * A feed value is a concrete value represented as an `Tensor`.
     */
    class FeedDict {
        /**
         * Constructor, optionally does copy-construction.
         * @param feeds An Array of `Feed`s, or another `FeedDict`, in which case
         *   copy-construction will be performed.
         */
        constructor(feeds) {
            this.id2Value = {};
            this.id2Mask = {};
            this.name2Id = {};
            if (feeds instanceof FeedDict) {
                for (const id in feeds.id2Value) {
                    this.id2Value[id] = feeds.id2Value[id];
                    if (id in feeds.id2Mask) {
                        this.id2Mask[id] = feeds.id2Mask[id];
                    }
                }
            }
            else {
                if (feeds == null) {
                    return;
                }
                for (const feed of feeds) {
                    this.add(feed.key, feed.value);
                }
            }
        }
        /**
         * Add a key-value pair to the FeedDict.
         *
         * @param key The key of the feed.
         * @param value The value of the tensor feed.
         * @param mask The value of the mask feed (optional).
         * @returns This `FeedDict`.
         * @throws ValueError: If the key `SymbolicTensor` already exists in the
         *   `FeedDict`.
         */
        add(key, value, mask) {
            if (this.id2Value[key.id] == null) {
                this.id2Value[key.id] = assertFeedCompatibility(key, value);
                this.name2Id[key.name] = key.id;
                if (mask != null) {
                    this.id2Mask[key.id] = mask;
                }
            }
            else {
                throw new ValueError(`Duplicate key: name=${key.name}, id=${key.id}`);
            }
            return this;
        }
        /**
         * Add a Feed to the FeedDict.
         * @param feed The new `Feed` to add.
         * @returns This `FeedDict`.
         */
        addFeed(feed) {
            this.add(feed.key, feed.value);
        }
        /**
         * Probe whether a key already exists in the FeedDict.
         * @param key
         */
        hasKey(key) {
            return this.id2Value[key.id] != null;
        }
        /**
         * Get all the SymbolicTensor available in this FeedDict.
         */
        names() {
            return Object.keys(this.name2Id);
        }
        /**
         * Get the feed value for given key.
         * @param key The SymbolicTensor, or its name (as a string), of which the
         *     value is sought.
         * @returns If `key` exists, the corresponding feed value.
         * @throws ValueError: If `key` does not exist in this `FeedDict`.
         */
        getValue(key) {
            if (key instanceof SymbolicTensor) {
                if (this.id2Value[key.id] == null) {
                    throw new ValueError(`Nonexistent key: ${key.name}`);
                }
                else {
                    return this.id2Value[key.id];
                }
            }
            else {
                const id = this.name2Id[key];
                if (id == null) {
                    throw new ValueError(`Feed dict has no SymbolicTensor name: ${key}`);
                }
                return this.id2Value[id];
            }
        }
        /**
         * Get the feed mask for given key.
         * @param key The SymbolicTensor, or its name (as a string), of which the
         *     value is sought.
         * @returns If `key` exists, the corresponding feed mask.
         * @throws ValueError: If `key` does not exist in this `FeedDict`.
         */
        getMask(key) {
            if (key instanceof SymbolicTensor) {
                if (this.id2Value[key.id] == null) {
                    throw new ValueError(`Nonexistent key: ${key.name}`);
                }
                else {
                    return this.id2Mask[key.id];
                }
            }
            else {
                const id = this.name2Id[key];
                if (id == null) {
                    throw new ValueError(`Feed dict has no SymbolicTensor name: ${key}`);
                }
                return this.id2Mask[id];
            }
        }
        /** Dispose all mask Tensors held by this object. */
        disposeMasks() {
            if (this.id2Mask != null) {
                dispose(this.id2Mask);
            }
        }
    }
    // Cache for topologically sorted SymbolicTensors for given execution
    // targets (i.e., fetches).
    const cachedSorted = new LruCache();
    // Cache for recipient count maps for given execution targets (i.e., fetches).
    const cachedRecipientCounts = new LruCache();
    function updateCacheMaxEntries(maxEntries) {
        if (cachedSorted != null) {
            cachedSorted.setMaxEntries(maxEntries);
        }
        if (cachedRecipientCounts != null) {
            cachedRecipientCounts.setMaxEntries(maxEntries);
        }
    }
    /**
     * Execute a SymbolicTensor by using concrete feed values.
     *
     * A `SymbolicTensor` object is a node in a computation graph of TF.js
     * Layers. The object is backed by a source layer and input
     * `SymbolicTensor`s to the source layer. This method evaluates
     * the `call()` method of the source layer, using concrete values of the
     * inputs obtained from either
     * * `feedDict`, if the input key exists in `feedDict`, or else,
     * * a recursive call to `execute()` itself.
     *
     * @param x: The `SymbolicTensor` to execute.
     * @param feedDict: The feed values, as base condition of the recursion.
     *   execution.
     * @param kwargs: Optional keyword arguments.
     * @param probe: A probe object (of interface `ExecutionProbe`) used for
     *   testing memory footprint of `execute` calls.
     * @returns Result of the execution.
     * @throws ValueError: If any `SymbolicTensor`s from `InputLayer`s
     *   encountered during the execution lacks a feed value in `feedDict`.
     */
    function execute(fetches, feedDict, kwargs, probe) {
        const training = kwargs == null ? false : kwargs['training'];
        const arrayFetches = Array.isArray(fetches);
        const fetchArray = arrayFetches ? fetches : [fetches];
        const outputNames = fetchArray.map(t => t.name);
        const finalOutputs = [];
        const feedNames = feedDict.names();
        for (const outputName of outputNames) {
            if (feedNames.indexOf(outputName) !== -1) {
                finalOutputs.push(feedDict.getValue(outputName));
            }
            else {
                finalOutputs.push(null);
            }
        }
        if (probe != null) {
            // For optional probing of memory footprint during execution.
            probe.maxNumTensors = -Infinity;
            probe.minNumTensors = Infinity;
        }
        // Check cache.
        const fetchAndFeedKey = outputNames.join(',') + '|' + feedDict.names().sort().join(',');
        let sorted = cachedSorted.get(fetchAndFeedKey);
        let recipientCounts;
        if (sorted == null) {
            // Cache doesn't contain the desired combination of fetches. Compute
            // topological sort for the combination for the first time.
            const out = getTopologicalSortAndRecipientCounts(fetchArray, feedDict);
            sorted = out.sorted;
            recipientCounts = out.recipientCounts;
            // Store results in cache for future use.
            cachedSorted.put(fetchAndFeedKey, sorted);
            cachedRecipientCounts.put(fetchAndFeedKey, recipientCounts);
        }
        recipientCounts = {};
        if (!training) {
            Object.assign(recipientCounts, cachedRecipientCounts.get(fetchAndFeedKey));
        }
        const internalFeedDict = new FeedDict(feedDict);
        // Start iterative execution on the topologically-sorted SymbolicTensors.
        for (let i = 0; i < sorted.length; ++i) {
            if (probe != null) {
                // For optional probing of memory usage during execution.
                const numTensors = memory().numTensors;
                if (numTensors > probe.maxNumTensors) {
                    probe.maxNumTensors = numTensors;
                }
                if (numTensors < probe.minNumTensors) {
                    probe.minNumTensors = numTensors;
                }
            }
            const symbolic = sorted[i];
            const srcLayer = symbolic.sourceLayer;
            if (srcLayer instanceof InputLayer) {
                continue;
            }
            const inputValues = [];
            const inputMasks = [];
            const tensorsToDispose = [];
            let maskExists = false;
            for (const input of symbolic.inputs) {
                const value = internalFeedDict.getValue(input);
                const mask = internalFeedDict.getMask(input);
                inputValues.push(value);
                inputMasks.push(mask);
                if (mask != null) {
                    maskExists = true;
                }
                if (!training) {
                    recipientCounts[input.name]--;
                    if (recipientCounts[input.name] === 0 && !feedDict.hasKey(input) &&
                        outputNames.indexOf(input.name) === -1 && !value.isDisposed &&
                        input.sourceLayer.stateful !== true) {
                        tensorsToDispose.push(value);
                    }
                }
            }
            if (maskExists) {
                kwargs = kwargs || {};
                kwargs['mask'] = inputMasks[0];
            }
            const outputTensors = toList(srcLayer.apply(inputValues, kwargs));
            let outputMask = null;
            if (srcLayer.supportsMasking) {
                outputMask = srcLayer.computeMask(inputValues, inputMasks);
            }
            const layerOutputs = getNodeOutputs(symbolic);
            const outputSymbolicTensors = Array.isArray(layerOutputs) ? layerOutputs : [layerOutputs];
            for (let i = 0; i < outputSymbolicTensors.length; ++i) {
                if (!internalFeedDict.hasKey(outputSymbolicTensors[i])) {
                    internalFeedDict.add(outputSymbolicTensors[i], outputTensors[i], Array.isArray(outputMask) ? outputMask[0] : outputMask);
                }
                const index = outputNames.indexOf(outputSymbolicTensors[i].name);
                if (index !== -1) {
                    finalOutputs[index] = outputTensors[i];
                }
            }
            if (!training) {
                // Clean up Tensors that are no longer needed.
                dispose(tensorsToDispose);
            }
        }
        // NOTE(cais): Unlike intermediate tensors, we don't discard mask
        // tensors as we go, because these tensors are sometimes passed over a
        // series of mutliple layers, i.e., not obeying the immediate input
        // relations in the graph. If this becomes a memory-usage concern,
        // we can improve this in the future.
        internalFeedDict.disposeMasks();
        return arrayFetches ? finalOutputs : finalOutputs[0];
    }
    /**
     * Sort the `SymbolicTensor`s topologically, for an array of fetches.
     *
     * This function calls getTopologicalSortAndRecipientCountsForOneFetch and
     * merges their results.
     *
     * @param fetch The array of fetches requested. Must be a non-empty array.
     * @param feedDict The dictionary of fed values.
     * @returns sorted: Topologically-sorted array of SymbolicTensors.
     *   recipientCounts: Recipient counts for all SymbolicTensors in `sorted`.
     */
    function getTopologicalSortAndRecipientCounts(fetches, feedDict) {
        assert$4(fetches != null && fetches.length > 0, () => `Expected at least one fetch, got none`);
        let finalSorted = [];
        let finalRecipientMap = {};
        if (fetches.length === 1) {
            // Special-casing 1 fetch for efficiency.
            const out = getTopologicalSortAndRecipientCountsForOneFetch(fetches[0], feedDict);
            finalSorted = out.sorted;
            finalRecipientMap = out.recipientMap;
        }
        else {
            const visited = new Set();
            for (const fetch of fetches) {
                const { sorted, recipientMap } = getTopologicalSortAndRecipientCountsForOneFetch(fetch, feedDict);
                // Merge sorted SymbolicTensor Arrays.
                for (const symbolicTensor of sorted) {
                    if (!visited.has(symbolicTensor.name)) {
                        finalSorted.push(symbolicTensor);
                        visited.add(symbolicTensor.name);
                    }
                }
                // Merge recipient maps.
                for (const name in recipientMap) {
                    if (finalRecipientMap[name] == null) {
                        finalRecipientMap[name] = new Set();
                    }
                    recipientMap[name].forEach(recipient => finalRecipientMap[name].add(recipient));
                }
            }
        }
        return {
            sorted: finalSorted,
            recipientCounts: recipientMap2Counts(finalRecipientMap)
        };
    }
    function recipientMap2Counts(recipientMap) {
        const recipientCounts = {};
        for (const name in recipientMap) {
            recipientCounts[name] = recipientMap[name].size;
        }
        return recipientCounts;
    }
    /**
     * Sort the `SymbolicTensor`s topologically, for a single fetch.
     *
     * This helper function processes the upstream SymbolicTensors of a single
     * fetch.
     *
     * @param fetch The single fetch requested.
     * @param feedDict The dictionary of fed values.
     * @returns sorted: Topologically-sorted array of SymbolicTensors.
     *   recipientMap: Recipient names for all SymbolicTensors in `sorted`.
     */
    function getTopologicalSortAndRecipientCountsForOneFetch(fetch, feedDict) {
        const visited = new Set();
        const sorted = [];
        const recipientMap = {};
        // Put keys of the feedDict into visited first, so they don't have to be
        // walked. This is needed in case where there are feeds for intermediate
        // SymbolicTensors of the graph.
        for (const key of feedDict.names()) {
            visited.add(key);
        }
        const stack = [];
        const marks = [];
        // Initial population of stack and marks.
        stack.push(fetch);
        while (stack.length > 0) {
            const top = stack[stack.length - 1];
            if (visited.has(top.name)) {
                stack.pop();
                continue;
            }
            const topIsMarked = marks[marks.length - 1] === stack.length - 1;
            if (top.inputs.length === 0 || topIsMarked) {
                // Input SymbolicTensor or all children have been visited.
                stack.pop();
                sorted.push(top);
                visited.add(top.name);
                if (topIsMarked) {
                    marks.pop();
                }
            }
            else {
                // A non-input SymbolicTensor whose upstream SymbolicTensors haven't
                // been visited yet. Push them onto the stack.
                marks.push(stack.length - 1);
                for (const input of top.inputs) {
                    // Increment the recipient count. Note that this needs to happen
                    // regardless of whether the SymbolicTensor has been visited before.
                    if (recipientMap[input.name] == null) {
                        recipientMap[input.name] = new Set();
                    }
                    recipientMap[input.name].add(top.name);
                    if (visited.has(input.name)) {
                        continue; // Avoid repeated visits to the same SymbolicTensor.
                    }
                    stack.push(input);
                }
            }
        }
        return { sorted, recipientMap };
    }
    /**
     * Get the symbolic output tensors of the node to which a given fetch belongs.
     * @param fetch The fetched symbolic tensor.
     * @returns The Array of symbolic tensors output by the node to which `fetch`
     *   belongs.
     */
    function getNodeOutputs(fetch) {
        let layerOutputs;
        if (fetch.sourceLayer.inboundNodes.length === 1) {
            layerOutputs = fetch.sourceLayer.output;
        }
        else {
            let nodeIndex = null;
            for (let i = 0; i < fetch.sourceLayer.inboundNodes.length; ++i) {
                for (const outputTensor of fetch.sourceLayer.inboundNodes[i]
                    .outputTensors) {
                    if (outputTensor.id === fetch.id) {
                        nodeIndex = i;
                        break;
                    }
                }
            }
            layerOutputs = fetch.sourceLayer.getOutputAt(nodeIndex);
        }
        return layerOutputs;
    }

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ENV$2 = env$1();
    /** The max number of entries for the caches of layers' topological sort. */
    ENV$2.registerFlag('TOPOLOGICAL_SORT_CACHE_MAX_ENTRIES', () => 100, updateCacheMaxEntries);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Helper function used by many of the Constraints to find the L2Norms.
     */
    function calcL2Norms(w, axis) {
        return tidy(() => sqrt$2(sum$2(mul(w, w), axis, true)));
    }
    /**
     * Base class for functions that impose constraints on weight values
     *
     * @doc {
     *   heading: 'Constraints',
     *   subheading: 'Classes',
     *   namespace: 'constraints'
     * }
     */
    class Constraint extends Serializable {
        getConfig() {
            return {};
        }
    }
    class MaxNorm extends Constraint {
        constructor(args) {
            super();
            this.defaultMaxValue = 2;
            this.defaultAxis = 0;
            this.maxValue =
                args.maxValue != null ? args.maxValue : this.defaultMaxValue;
            this.axis = args.axis != null ? args.axis : this.defaultAxis;
        }
        apply(w) {
            return tidy(() => {
                const norms = calcL2Norms(w, this.axis);
                const desired = clipByValue$2(norms, 0, this.maxValue);
                return mul(w, div$1(desired, add$2(epsilon$1(), norms)));
            });
        }
        getConfig() {
            return { maxValue: this.maxValue, axis: this.axis };
        }
    }
    /** @nocollapse */
    MaxNorm.className = 'MaxNorm';
    registerClass(MaxNorm);
    class UnitNorm extends Constraint {
        constructor(args) {
            super();
            this.defaultAxis = 0;
            this.axis = args.axis != null ? args.axis : this.defaultAxis;
        }
        apply(w) {
            return tidy(() => div$1(w, add$2(epsilon$1(), calcL2Norms(w, this.axis))));
        }
        getConfig() {
            return { axis: this.axis };
        }
    }
    /** @nocollapse */
    UnitNorm.className = 'UnitNorm';
    registerClass(UnitNorm);
    class NonNeg extends Constraint {
        apply(w) {
            return relu$2(w);
        }
    }
    /** @nocollapse */
    NonNeg.className = 'NonNeg';
    registerClass(NonNeg);
    class MinMaxNorm extends Constraint {
        constructor(args) {
            super();
            this.defaultMinValue = 0.0;
            this.defaultMaxValue = 1.0;
            this.defaultRate = 1.0;
            this.defaultAxis = 0;
            this.minValue =
                args.minValue != null ? args.minValue : this.defaultMinValue;
            this.maxValue =
                args.maxValue != null ? args.maxValue : this.defaultMaxValue;
            this.rate = args.rate != null ? args.rate : this.defaultRate;
            this.axis = args.axis != null ? args.axis : this.defaultAxis;
        }
        apply(w) {
            return tidy(() => {
                const norms = calcL2Norms(w, this.axis);
                const desired = add$2(mul(this.rate, clipByValue$2(norms, this.minValue, this.maxValue)), mul(1.0 - this.rate, norms));
                return mul(w, div$1(desired, add$2(epsilon$1(), norms)));
            });
        }
        getConfig() {
            return {
                minValue: this.minValue,
                maxValue: this.maxValue,
                rate: this.rate,
                axis: this.axis
            };
        }
    }
    /** @nocollapse */
    MinMaxNorm.className = 'MinMaxNorm';
    registerClass(MinMaxNorm);
    // Maps the JavaScript-like identifier keys to the corresponding registry
    // symbols.
    const CONSTRAINT_IDENTIFIER_REGISTRY_SYMBOL_MAP = {
        'maxNorm': 'MaxNorm',
        'minMaxNorm': 'MinMaxNorm',
        'nonNeg': 'NonNeg',
        'unitNorm': 'UnitNorm'
    };
    function serializeConstraint(constraint) {
        return serializeKerasObject(constraint);
    }
    function deserializeConstraint(config, customObjects = {}) {
        return deserializeKerasObject(config, SerializationMap.getMap().classNameMap, customObjects, 'constraint');
    }
    function getConstraint(identifier) {
        if (identifier == null) {
            return null;
        }
        if (typeof identifier === 'string') {
            const className = identifier in CONSTRAINT_IDENTIFIER_REGISTRY_SYMBOL_MAP ?
                CONSTRAINT_IDENTIFIER_REGISTRY_SYMBOL_MAP[identifier] :
                identifier;
            const config = { className, config: {} };
            return deserializeConstraint(config);
        }
        else if (identifier instanceof Constraint) {
            return identifier;
        }
        else {
            return deserializeConstraint(identifier);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * MaxNorm weight constraint.
     *
     * Constrains the weights incident to each hidden unit
     * to have a norm less than or equal to a desired value.
     *
     * References
     *       - [Dropout: A Simple Way to Prevent Neural Networks from Overfitting
     * Srivastava, Hinton, et al.
     * 2014](http://www.cs.toronto.edu/~rsalakhu/papers/srivastava14a.pdf)
     *
     * @doc {heading: 'Constraints',namespace: 'constraints'}
     */
    function maxNorm(args) {
        return new MaxNorm(args);
    }
    /**
     * Constrains the weights incident to each hidden unit to have unit norm.
     *
     * @doc {heading: 'Constraints', namespace: 'constraints'}
     */
    function unitNorm(args) {
        return new UnitNorm(args);
    }
    /**
     * Constains the weight to be non-negative.
     *
     * @doc {heading: 'Constraints', namespace: 'constraints'}
     */
    function nonNeg() {
        return new NonNeg();
    }
    /** @doc {heading: 'Constraints', namespace: 'constraints'} */
    function minMaxNorm(config) {
        return new MinMaxNorm(config);
    }

    var exports_constraints = /*#__PURE__*/Object.freeze({
        __proto__: null,
        maxNorm: maxNorm,
        unitNorm: unitNorm,
        nonNeg: nonNeg,
        minMaxNorm: minMaxNorm
    });

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Initializer that generates tensors initialized to 0.
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function zeros$1() {
        return new Zeros();
    }
    /**
     * Initializer that generates tensors initialized to 1.
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function ones() {
        return new Ones();
    }
    /**
     * Initializer that generates values initialized to some constant.
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function constant(args) {
        return new Constant(args);
    }
    /**
     * Initializer that generates random values initialized to a uniform
     * distribution.
     *
     * Values will be distributed uniformly between the configured minval and
     * maxval.
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function randomUniform(args) {
        return new RandomUniform(args);
    }
    /**
     * Initializer that generates random values initialized to a normal
     * distribution.
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function randomNormal(args) {
        return new RandomNormal(args);
    }
    /**
     * Initializer that generates random values initialized to a truncated normal.
     * distribution.
     *
     * These values are similar to values from a `RandomNormal` except that values
     * more than two standard deviations from the mean are discarded and re-drawn.
     * This is the recommended initializer for neural network weights and filters.
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function truncatedNormal(args) {
        return new TruncatedNormal(args);
    }
    /**
     * Initializer that generates the identity matrix.
     * Only use for square 2D matrices.
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function identity$2(args) {
        return new Identity(args);
    }
    /**
     * Initializer capable of adapting its scale to the shape of weights.
     * With distribution=NORMAL, samples are drawn from a truncated normal
     * distribution centered on zero, with `stddev = sqrt(scale / n)` where n is:
     *   - number of input units in the weight tensor, if mode = FAN_IN.
     *   - number of output units, if mode = FAN_OUT.
     *   - average of the numbers of input and output units, if mode = FAN_AVG.
     * With distribution=UNIFORM,
     * samples are drawn from a uniform distribution
     * within [-limit, limit], with `limit = sqrt(3 * scale / n)`.
     *
     * @doc {heading: 'Initializers',namespace: 'initializers'}
     */
    function varianceScaling(config) {
        return new VarianceScaling(config);
    }
    /**
     * Glorot uniform initializer, also called Xavier uniform initializer.
     * It draws samples from a uniform distribution within [-limit, limit]
     * where `limit` is `sqrt(6 / (fan_in + fan_out))`
     * where `fan_in` is the number of input units in the weight tensor
     * and `fan_out` is the number of output units in the weight tensor
     *
     * Reference:
     *   Glorot & Bengio, AISTATS 2010
     *       http://jmlr.org/proceedings/papers/v9/glorot10a/glorot10a.pdf.
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function glorotUniform(args) {
        return new GlorotUniform(args);
    }
    /**
     * Glorot normal initializer, also called Xavier normal initializer.
     * It draws samples from a truncated normal distribution centered on 0
     * with `stddev = sqrt(2 / (fan_in + fan_out))`
     * where `fan_in` is the number of input units in the weight tensor
     * and `fan_out` is the number of output units in the weight tensor.
     *
     * Reference:
     *   Glorot & Bengio, AISTATS 2010
     *       http://jmlr.org/proceedings/papers/v9/glorot10a/glorot10a.pdf
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function glorotNormal(args) {
        return new GlorotNormal(args);
    }
    /**
     * He normal initializer.
     *
     * It draws samples from a truncated normal distribution centered on 0
     * with `stddev = sqrt(2 / fanIn)`
     * where `fanIn` is the number of input units in the weight tensor.
     *
     * Reference:
     *     He et al., http://arxiv.org/abs/1502.01852
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function heNormal(args) {
        return new HeNormal(args);
    }
    /**
     * He uniform initializer.
     *
     * It draws samples from a uniform distribution within [-limit, limit]
     * where `limit` is `sqrt(6 / fan_in)`
     * where `fanIn` is the number of input units in the weight tensor.
     *
     * Reference:
     *     He et al., http://arxiv.org/abs/1502.01852
     *
     * @doc {heading: 'Initializers',namespace: 'initializers'}
     */
    function heUniform(args) {
        return new HeUniform(args);
    }
    /**
     * LeCun normal initializer.
     *
     * It draws samples from a truncated normal distribution centered on 0
     * with `stddev = sqrt(1 / fanIn)`
     * where `fanIn` is the number of input units in the weight tensor.
     *
     * References:
     *   [Self-Normalizing Neural Networks](https://arxiv.org/abs/1706.02515)
     *   [Efficient Backprop](http://yann.lecun.com/exdb/publis/pdf/lecun-98b.pdf)
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function leCunNormal(args) {
        return new LeCunNormal(args);
    }
    /**
     * LeCun uniform initializer.
     *
     * It draws samples from a uniform distribution in the interval
     * `[-limit, limit]` with `limit = sqrt(3 / fanIn)`,
     * where `fanIn` is the number of input units in the weight tensor.
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function leCunUniform(args) {
        return new LeCunUniform(args);
    }
    /**
     * Initializer that generates a random orthogonal matrix.
     *
     * Reference:
     * [Saxe et al., http://arxiv.org/abs/1312.6120](http://arxiv.org/abs/1312.6120)
     *
     * @doc {heading: 'Initializers', namespace: 'initializers'}
     */
    function orthogonal(args) {
        return new Orthogonal(args);
    }

    var exports_initializers = /*#__PURE__*/Object.freeze({
        __proto__: null,
        zeros: zeros$1,
        ones: ones,
        constant: constant,
        randomUniform: randomUniform,
        randomNormal: randomNormal,
        truncatedNormal: truncatedNormal,
        identity: identity$2,
        varianceScaling: varianceScaling,
        glorotUniform: glorotUniform,
        glorotNormal: glorotNormal,
        heNormal: heNormal,
        heUniform: heUniform,
        leCunNormal: leCunNormal,
        leCunUniform: leCunUniform,
        orthogonal: orthogonal
    });

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Turn any Scalar values in a Logs object into actual number values.
     *
     * @param logs The `Logs` object to be resolved in place.
     */
    async function resolveScalarsInLogs(logs) {
        if (logs == null) {
            return;
        }
        const promises = [];
        const keys = [];
        const scalarsToDispose = [];
        for (const key in logs) {
            const value = logs[key];
            if (typeof value !== 'number') {
                const valueScalar = value;
                promises.push(valueScalar.data());
                keys.push(key);
                scalarsToDispose.push(valueScalar);
            }
        }
        if (promises.length > 0) {
            const values = await Promise.all(promises);
            for (let i = 0; i < values.length; ++i) {
                logs[keys[i]] = values[i][0];
            }
            // Dispose the original scalar tensors.
            dispose(scalarsToDispose);
        }
    }
    /**
     * Dispose all Tensors in an UnresolvedLogs object.
     *
     * @param logs An `UnresolvedLogs` object potentially containing `tf.Tensor`s in
     *   places where the values can be `tf.Tensor` or `number`.
     */
    function disposeTensorsInLogs(logs) {
        if (logs == null) {
            return;
        }
        for (const key in logs) {
            const value = logs[key];
            if (typeof value !== 'number') {
                value.dispose();
            }
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /** Verbosity logging level when fitting a model. */
    var ModelLoggingVerbosity;
    (function (ModelLoggingVerbosity) {
        ModelLoggingVerbosity[ModelLoggingVerbosity["SILENT"] = 0] = "SILENT";
        ModelLoggingVerbosity[ModelLoggingVerbosity["VERBOSE"] = 1] = "VERBOSE";
    })(ModelLoggingVerbosity || (ModelLoggingVerbosity = {}));
    /** How often to yield to the main thread when training (in ms). */
    const DEFAULT_YIELD_EVERY_MS = 125;
    /**
     * Abstract base class used to build new callbacks.
     *
     * The `logs` dictionary that callback methods take as argument will contain
     * keys for quantities relevant to the current batch or epoch.
     *
     * Currently, the `.fit()` method of the `Sequential` model class
     * will include the following quantities in the `logs` that
     * it passes to its callbacks:
     *
     * onEpochEnd: Logs include `acc` and `loss`, and optionally include `valLoss`
     *   (if validation is enabled in `fit`), and `valAcc` (if validation and
     *   accuracy monitoring are enabled).
     * onBatchBegin: Logs include `size`, the number of samples in the current
     *   batch.
     * onBatchEnd: Logs include `loss`, and optionally `acc` (if accuracy monitoring
     *   is enabled).
     */
    class BaseCallback {
        constructor() {
            // TODO(michaelterry): This type is a best guess.
            this.validationData = null;
        }
        setParams(params) {
            this.params = params;
        }
        async onEpochBegin(epoch, logs) { }
        async onEpochEnd(epoch, logs) { }
        async onBatchBegin(batch, logs) { }
        async onBatchEnd(batch, logs) { }
        async onTrainBegin(logs) { }
        async onTrainEnd(logs) { }
        // LayersModel needs to call Callback.setModel(), but cannot actually depend
        // on Callback because that creates a cyclic dependency.  Providing this no-op
        // method on BaseCallback breaks the cycle: this way LayersModel can depend on
        // BaseCallback but not on Callback.  The argument is typed as `Container`
        // (the superclass of LayersModel) to avoid recapitulating the cycle. Callback
        // overrides this method and enforces that the argument is really a
        // LayersModel.
        setModel(model) {
            // Do nothing. Use Callback instead of BaseCallback to track the model.
        }
    }
    /**
     * Container abstracting a list of callbacks.
     */
    class CallbackList {
        // TODO(cais): When the need arises, uncomment the following lines and
        // implement the queue for time values.
        // private deltaTBatch: number;
        // private deltaTsBatchBegin: Array<number>;
        // private deltaTsBatchEnd: Array<number>;
        /**
         * Constructor of CallbackList.
         * @param callbacks Array of `Callback` instances.
         * @param queueLength Queue length for keeping running statistics over
         *   callback execution time.
         */
        constructor(callbacks, queueLength = 10) {
            // TODO(cais): Make use of queueLength when implementing the queue for time
            // values.
            if (callbacks == null) {
                callbacks = [];
            }
            this.callbacks = callbacks;
            this.queueLength = queueLength;
        }
        append(callback) {
            this.callbacks.push(callback);
        }
        setParams(params) {
            for (const callback of this.callbacks) {
                callback.setParams(params);
            }
        }
        setModel(model) {
            for (const callback of this.callbacks) {
                callback.setModel(model);
            }
        }
        /**
         * Called at the start of an epoch.
         * @param epoch Index of epoch.
         * @param logs Dictionary of logs.
         */
        async onEpochBegin(epoch, logs) {
            if (logs == null) {
                logs = {};
            }
            for (const callback of this.callbacks) {
                await callback.onEpochBegin(epoch, logs);
            }
        }
        /**
         * Called at the end of an epoch.
         * @param epoch Index of epoch.
         * @param logs Dictionary of logs.
         */
        async onEpochEnd(epoch, logs) {
            if (logs == null) {
                logs = {};
            }
            for (const callback of this.callbacks) {
                await callback.onEpochEnd(epoch, logs);
            }
        }
        /**
         * Called  right before processing a batch.
         * @param batch Index of batch within the current epoch.
         * @param logs Dictionary of logs.
         */
        async onBatchBegin(batch, logs) {
            if (logs == null) {
                logs = {};
            }
            for (const callback of this.callbacks) {
                await callback.onBatchBegin(batch, logs);
            }
        }
        /**
         * Called at the end of a batch.
         * @param batch Index of batch within the current epoch.
         * @param logs Dictionary of logs.
         */
        async onBatchEnd(batch, logs) {
            if (logs == null) {
                logs = {};
            }
            for (const callback of this.callbacks) {
                await callback.onBatchEnd(batch, logs);
            }
        }
        /**
         * Called at the beginning of training.
         * @param logs Dictionary of logs.
         */
        async onTrainBegin(logs) {
            if (logs == null) {
                logs = {};
            }
            for (const callback of this.callbacks) {
                await callback.onTrainBegin(logs);
            }
        }
        /**
         * Called at the end of training.
         * @param logs Dictionary of logs.
         */
        async onTrainEnd(logs) {
            if (logs == null) {
                logs = {};
            }
            for (const callback of this.callbacks) {
                await callback.onTrainEnd(logs);
            }
        }
    }
    /**
     * Callback that accumulates epoch averages of metrics.
     *
     * This callback is automatically applied to every LayersModel.
     */
    class BaseLogger extends BaseCallback {
        constructor() {
            super();
        }
        async onEpochBegin(epoch) {
            this.seen = 0;
            this.totals = {};
        }
        async onBatchEnd(batch, logs) {
            if (logs == null) {
                logs = {};
            }
            const batchSize = logs['size'] == null ? 0 : logs['size'];
            this.seen += batchSize;
            for (const key in logs) {
                const value = logs[key];
                if (typeof value === 'number') {
                    if (!this.totals.hasOwnProperty(key)) {
                        this.totals[key] = 0;
                    }
                    this.totals[key] = this.totals[key] + value * batchSize;
                }
                else {
                    let oldTotalsToDispose;
                    if (key in this.totals) {
                        oldTotalsToDispose = this.totals[key];
                    }
                    else {
                        this.totals[key] = 0;
                    }
                    const total = tidy(() => add$2((this.totals[key]), mul(value, batchSize)));
                    this.totals[key] = total;
                    if (oldTotalsToDispose != null) {
                        oldTotalsToDispose.dispose();
                    }
                }
            }
        }
        async onEpochEnd(epoch, logs) {
            if (logs != null) {
                for (const key of this.params['metrics']) {
                    if (this.totals[key] == null) {
                        continue;
                    }
                    if (typeof this.totals[key] === 'number') {
                        logs[key] = this.totals[key] / this.seen;
                    }
                    else {
                        tidy(() => {
                            const log = mul(div$1(1, this.seen), this.totals[key]);
                            logs[key] = log;
                            this.totals[key].dispose();
                            keep(logs[key]);
                        });
                    }
                }
            }
        }
    }
    /**
     * Callback that records events into a `History` object. This callback is
     * automatically applied to every TF.js Layers model. The `History` object
     * gets returned by the `fit` method of models.
     */
    class History extends BaseCallback {
        async onTrainBegin(logs) {
            this.epoch = [];
            this.history = {};
        }
        async onEpochEnd(epoch, logs) {
            if (logs == null) {
                logs = {};
            }
            this.epoch.push(epoch);
            for (const key in logs) {
                if (this.history[key] == null) {
                    this.history[key] = [];
                }
                this.history[key].push(logs[key]);
            }
        }
        /**
         * Await the values of all losses and metrics.
         */
        async syncData() {
            const promises = [];
            const keys = [];
            const indices = [];
            for (const key in this.history) {
                const valueArray = this.history[key];
                for (let i = 0; i < valueArray.length; ++i) {
                    if (typeof valueArray[i] !== 'number') {
                        const valueScalar = valueArray[i];
                        promises.push(valueScalar.data());
                        keys.push(key);
                        indices.push(i);
                    }
                }
            }
            const values = await Promise.all(promises);
            for (let n = 0; n < values.length; ++n) {
                const tensorToDispose = this.history[keys[n]][indices[n]];
                tensorToDispose.dispose();
                this.history[keys[n]][indices[n]] = values[n][0];
            }
        }
    }
    /**
     * Custom callback for training.
     */
    class CustomCallback extends BaseCallback {
        constructor(args, yieldEvery) {
            super();
            this.currentEpoch = 0;
            this.nowFunc = args.nowFunc;
            this.nextFrameFunc = args.nextFrameFunc || nextFrame;
            this.yieldEvery = yieldEvery || 'auto';
            if (this.yieldEvery === 'auto') {
                this.yieldEvery = DEFAULT_YIELD_EVERY_MS;
            }
            if (this.yieldEvery === 'never' && args.onYield != null) {
                throw new Error('yieldEvery is `never` but you provided an `onYield` callback. ' +
                    'Either change `yieldEvery` or remove the callback');
            }
            if (isNumber$3(this.yieldEvery)) {
                // Decorate `maybeWait` so it will be called at most once every
                // `yieldEvery` ms.
                this.maybeWait = debounce(this.maybeWait.bind(this), this.yieldEvery, this.nowFunc);
            }
            this.trainBegin = args.onTrainBegin;
            this.trainEnd = args.onTrainEnd;
            this.epochBegin = args.onEpochBegin;
            this.epochEnd = args.onEpochEnd;
            this.batchBegin = args.onBatchBegin;
            this.batchEnd = args.onBatchEnd;
            this.yield = args.onYield;
        }
        async maybeWait(epoch, batch, logs) {
            const ps = [];
            if (this.yield != null) {
                await resolveScalarsInLogs(logs);
                ps.push(this.yield(epoch, batch, logs));
            }
            ps.push(this.nextFrameFunc());
            await Promise.all(ps);
        }
        async onEpochBegin(epoch, logs) {
            this.currentEpoch = epoch;
            if (this.epochBegin != null) {
                await resolveScalarsInLogs(logs);
                await this.epochBegin(epoch, logs);
            }
        }
        async onEpochEnd(epoch, logs) {
            const ps = [];
            if (this.epochEnd != null) {
                await resolveScalarsInLogs(logs);
                ps.push(this.epochEnd(epoch, logs));
            }
            if (this.yieldEvery === 'epoch') {
                ps.push(this.nextFrameFunc());
            }
            await Promise.all(ps);
        }
        async onBatchBegin(batch, logs) {
            if (this.batchBegin != null) {
                await resolveScalarsInLogs(logs);
                await this.batchBegin(batch, logs);
            }
        }
        async onBatchEnd(batch, logs) {
            const ps = [];
            if (this.batchEnd != null) {
                await resolveScalarsInLogs(logs);
                ps.push(this.batchEnd(batch, logs));
            }
            if (this.yieldEvery === 'batch') {
                ps.push(this.nextFrameFunc());
            }
            else if (isNumber$3(this.yieldEvery)) {
                ps.push(this.maybeWait(this.currentEpoch, batch, logs));
            }
            await Promise.all(ps);
        }
        async onTrainBegin(logs) {
            if (this.trainBegin != null) {
                await resolveScalarsInLogs(logs);
                await this.trainBegin(logs);
            }
        }
        async onTrainEnd(logs) {
            if (this.trainEnd != null) {
                await resolveScalarsInLogs(logs);
                await this.trainEnd(logs);
            }
        }
    }
    /**
     * Standardize callbacks or configurations of them to an Array of callbacks.
     */
    function standardizeCallbacks(callbacks, yieldEvery) {
        if (callbacks == null) {
            callbacks = {};
        }
        if (callbacks instanceof BaseCallback) {
            return [callbacks];
        }
        if (Array.isArray(callbacks) && callbacks[0] instanceof BaseCallback) {
            return callbacks;
        }
        // Convert custom callback configs to custom callback objects.
        const callbackConfigs = toList(callbacks);
        return callbackConfigs.map(callbackConfig => new CustomCallback(callbackConfig, yieldEvery));
    }
    /**
     * A global registry for callback constructors to be used during
     * LayersModel.fit().
     */
    class CallbackConstructorRegistry {
        /**
         * Blocks public access to constructor.
         */
        constructor() { }
        /**
         * Register a tf.LayersModel.fit() callback constructor.
         *
         * The registered callback constructor will be used to instantiate
         * callbacks for every tf.LayersModel.fit() call afterwards.
         *
         * @param verbosityLevel Level of verbosity at which the `callbackConstructor`
         *   is to be reigstered.
         * @param callbackConstructor A no-arg constructor for `tf.Callback`.
         * @throws Error, if the same callbackConstructor has been registered before,
         *   either at the same or a different `verbosityLevel`.
         */
        static registerCallbackConstructor(verbosityLevel, callbackConstructor) {
            assert$4(verbosityLevel >= 0 && Number.isInteger(verbosityLevel), () => `Verbosity level is expected to be an integer >= 0, ` +
                `but got ${verbosityLevel}`);
            CallbackConstructorRegistry.checkForDuplicate(callbackConstructor);
            if (CallbackConstructorRegistry.constructors[verbosityLevel] == null) {
                CallbackConstructorRegistry.constructors[verbosityLevel] = [];
            }
            CallbackConstructorRegistry.constructors[verbosityLevel].push(callbackConstructor);
        }
        static checkForDuplicate(callbackConstructor) {
            for (const levelName in CallbackConstructorRegistry.constructors) {
                const constructors = CallbackConstructorRegistry.constructors[+levelName];
                constructors.forEach(ctor => {
                    if (ctor === callbackConstructor) {
                        throw new ValueError('Duplicate callback constructor.');
                    }
                });
            }
        }
        /**
         * Clear all registered callback constructors.
         */
        static clear() {
            CallbackConstructorRegistry.constructors = {};
        }
        /**
         * Create callbacks using the registered callback constructors.
         *
         * Given `verbosityLevel`, all constructors registered at that level or above
         * will be called and the instantiated callbacks will be used.
         *
         * @param verbosityLevel: Level of verbosity.
         */
        static createCallbacks(verbosityLevel) {
            const constructors = [];
            for (const levelName in CallbackConstructorRegistry.constructors) {
                const level = +levelName;
                if (verbosityLevel >= level) {
                    constructors.push(...CallbackConstructorRegistry.constructors[level]);
                }
            }
            return constructors.map(ctor => new ctor());
        }
    }
    CallbackConstructorRegistry.constructors = {};
    function configureCallbacks(callbacks, verbose, epochs, initialEpoch, numTrainSamples, stepsPerEpoch, batchSize, doValidation, callbackMetrics) {
        const history = new History();
        const actualCallbacks = [
            new BaseLogger(), ...CallbackConstructorRegistry.createCallbacks(verbose)
        ];
        if (callbacks != null) {
            actualCallbacks.push(...callbacks);
        }
        actualCallbacks.push(history);
        const callbackList = new CallbackList(actualCallbacks);
        // TODO(cais): Figure out when this LayersModel instance can have a
        // dynamically
        //   set property called 'callback_model' as in PyKeras.
        callbackList.setParams({
            epochs,
            initialEpoch,
            samples: numTrainSamples,
            steps: stepsPerEpoch,
            batchSize,
            verbose,
            doValidation,
            metrics: callbackMetrics,
        });
        return { callbackList, history };
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Instantiate a layer from a config dictionary.
     * @param config dict of the form {class_name: str, config: dict}
     * @param customObjects dict mapping class names (or function names)
     *   of custom (non-Keras) objects to class/functions
     * @param fastWeightInit Optional flag to use fast weight initialization
     *   during deserialization. This is applicable to cases in which
     *   the initialization will be immediately overwritten by loaded weight
     *   values. Default: `false`.
     * @returns Layer instance (may be LayersModel, Sequential, Layer...)
     */
    function deserialize(config, customObjects = {}, fastWeightInit = false) {
        return deserializeKerasObject(config, SerializationMap.getMap().classNameMap, customObjects, 'layer', fastWeightInit);
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Normalizes a tensor wrt the L2 norm alongside the specified axis.
     * @param x
     * @param axis Axis along which to perform normalization.
     */
    function l2Normalize(x, axis) {
        return tidy(() => {
            if (x.dtype !== 'float32') {
                x = cast$3(x, 'float32');
            }
            const squareSum = sum$2(square$1(x), axis, true);
            const epsilonTensor = fill$2(squareSum.shape, epsilon$1());
            const norm = sqrt$2(maximum$3(squareSum, epsilonTensor));
            return div$1(x, norm);
        });
    }
    function meanSquaredError$1(yTrue, yPred) {
        return tidy(() => mean$1(square$1(sub$2(yPred, yTrue)), -1));
    }
    function meanAbsoluteError$1(yTrue, yPred) {
        return tidy(() => mean$1(abs$2(sub$2(yPred, yTrue)), -1));
    }
    function meanAbsolutePercentageError$1(yTrue, yPred) {
        return tidy(() => {
            const diff = sub$2(yTrue, yPred);
            const clippedTrue = clipByValue$2(abs$2(yTrue), epsilon$1(), Number.MAX_VALUE);
            const absResult = abs$2(div$1(diff, clippedTrue));
            return mul(100, mean$1(absResult, -1));
        });
    }
    function meanSquaredLogarithmicError(yTrue, yPred) {
        return tidy(() => {
            const clippedPred = clipByValue$2(yPred, epsilon$1(), Number.MAX_VALUE);
            const firstLog = log$3(add$2(1, clippedPred));
            const clippedTrue = clipByValue$2(yTrue, epsilon$1(), Number.MAX_VALUE);
            const secondLog = log$3(add$2(1, clippedTrue));
            return mean$1(square$1(sub$2(firstLog, secondLog)), -1);
        });
    }
    function squaredHinge(yTrue, yPred) {
        return tidy(() => {
            const maxResult = maximum$3(0, sub$2(1, mul(yTrue, yPred)));
            return mean$1(square$1(maxResult), -1);
        });
    }
    function hinge(yTrue, yPred) {
        return tidy(() => {
            const maxResult = maximum$3(0, sub$2(1, mul(yTrue, yPred)));
            return mean$1(maxResult, -1);
        });
    }
    function categoricalHinge(yTrue, yPred) {
        return tidy(() => {
            const pos = sum$2(mul(yTrue, yPred), -1);
            const neg = max$3(mul(sub$2(1, yTrue), yPred), -1);
            return maximum$3(0, add$2(1, sub$2(neg, pos)));
        });
    }
    /**
     * Logarithm of the hyperbolic cosine of the prediction error.
     *
     * `log(cosh(x))` is approximately equal to `(x ** 2) / 2` for small `x` and
     * to `abs(x) - log(2)` for large `x`. This means that 'logcosh' works mostly
     * like the mean squared error, but will not be so strongly affected by the
     * occasional wildly incorrect prediction.
     */
    function logcosh(yTrue, yPred) {
        return tidy(() => {
            const log2 = Math.log(2);
            const predictionDiff = sub$2(yPred, yTrue);
            const logcoshResult = sub$2(add$2(predictionDiff, softplus$2(mul(-2, predictionDiff))), log2);
            return mean$1(logcoshResult, -1);
        });
    }
    function categoricalCrossentropy$2(target, output, fromLogits = false) {
        return tidy(() => {
            if (fromLogits) {
                output = softmax$3(output);
            }
            else {
                // scale preds so that the class probabilities of each sample sum to 1.
                const outputSum = sum$2(output, output.shape.length - 1, true);
                output = div$1(output, outputSum);
            }
            output = clipByValue$2(output, epsilon$1(), 1 - epsilon$1());
            return neg$2(sum$2(mul(cast$3(target, 'float32'), log$3(output)), output.shape.length - 1));
        });
    }
    /**
     * Categorical crossentropy with integer targets.
     *
     * @param target An integer tensor.
     * @param output A tensor resulting from a softmax (unless `fromLogits` is
     *  `true`, in which case `output` is expected to be the logits).
     * @param fromLogits Boolean, whether `output` is the result of a softmax, or is
     *   a tensor of logits.
     */
    function sparseCategoricalCrossentropy$1(target, output, fromLogits = false) {
        return tidy(() => {
            const flatTarget = cast$3(floor$3(flatten$2(target)), 'int32');
            output = clipByValue$2(output, epsilon$1(), 1 - epsilon$1());
            const outputShape = output.shape;
            const oneHotTarget = reshape$3(oneHot$2(flatTarget, outputShape[outputShape.length - 1]), outputShape);
            return categoricalCrossentropy$2(oneHotTarget, output, fromLogits);
        });
    }
    /**
     * From TensorFlow's implementation in nn_impl.py:
     *
     * For brevity, let `x = logits`, `z = labels`.  The logistic loss is
     *      z * -log(sigmoid(x)) + (1 - z) * -log(1 - sigmoid(x))
     *    = z * -log(1 / (1 + exp(-x))) + (1 - z) * -log(exp(-x) / (1 + exp(-x)))
     *    = z * log(1 + exp(-x)) + (1 - z) * (-log(exp(-x)) + log(1 + exp(-x)))
     *    = z * log(1 + exp(-x)) + (1 - z) * (x + log(1 + exp(-x))
     *    = (1 - z) * x + log(1 + exp(-x))
     *    = x - x * z + log(1 + exp(-x))
     * For x < 0, to avoid overflow in exp(-x), we reformulate the above
     *      x - x * z + log(1 + exp(-x))
     *    = log(exp(x)) - x * z + log(1 + exp(-x))
     *    = - x * z + log(1 + exp(x))
     * Hence, to ensure stability and avoid overflow, the implementation uses this
     * equivalent formulation
     *    max(x, 0) - x * z + log(1 + exp(-abs(x)))
     *
     * @param labels The labels.
     * @param logits The logits.
     */
    function sigmoidCrossEntropyWithLogits(labels, logits) {
        if (!arraysEqual(labels.shape, logits.shape)) {
            throw new ValueError(`logits and labels must have the same shape, but got shapes ` +
                `${JSON.stringify(labels.shape)} and ${JSON.stringify(logits.shape)}`);
        }
        return tidy(() => {
            // The logistic loss formula from above is
            //   x - x * z + log(1 + exp(-x))
            // For x < 0, a more numerically stable formula is
            //   -x * z + log(1 + exp(x))
            // Note that these two expressions can be combined into the following:
            //   max(x, 0) - x * z + log(1 + exp(-abs(x)))
            const reluLogits = relu$2(logits);
            const negAbsLogits = neg$2(abs$2(logits));
            return add$2(sub$2(reluLogits, mul(logits, labels)), log1p$2(exp$2(negAbsLogits)));
        });
    }
    function binaryCrossentropy$2(yTrue, yPred) {
        return tidy(() => {
            let y;
            y = clipByValue$2(yPred, epsilon$1(), 1 - epsilon$1());
            y = log$3(div$1(y, sub$2(1, y)));
            return mean$1(sigmoidCrossEntropyWithLogits(yTrue, y), -1);
        });
    }
    function kullbackLeiblerDivergence(yTrue, yPred) {
        return tidy(() => {
            const clippedTrue = clipByValue$2(yTrue, epsilon$1(), 1);
            const clippedPred = clipByValue$2(yPred, epsilon$1(), 1);
            return sum$2(mul(yTrue, log$3(div$1(clippedTrue, clippedPred))), -1);
        });
    }
    function poisson(yTrue, yPred) {
        return tidy(() => {
            const logPred = log$3(add$2(epsilon$1(), yPred));
            return mean$1(sub$2(yPred, mul(yTrue, logPred)), -1);
        });
    }
    function cosineProximity$1(yTrue, yPred) {
        return tidy(() => {
            const trueNormalized = l2Normalize(yTrue, -1);
            const predNormalized = l2Normalize(yPred, -1);
            const trueXPred = mul(trueNormalized, predNormalized);
            return neg$2(sum$2(trueXPred, -1));
        });
    }
    // TODO(michaelterry): Add deserialize() function.
    const lossesMap = {
        meanSquaredError: meanSquaredError$1,
        meanAbsoluteError: meanAbsoluteError$1,
        meanAbsolutePercentageError: meanAbsolutePercentageError$1,
        meanSquaredLogarithmicError,
        squaredHinge,
        hinge,
        categoricalHinge,
        logcosh,
        categoricalCrossentropy: categoricalCrossentropy$2,
        sparseCategoricalCrossentropy: sparseCategoricalCrossentropy$1,
        binaryCrossentropy: binaryCrossentropy$2,
        kullbackLeiblerDivergence,
        poisson,
        cosineProximity: cosineProximity$1
    };
    // Porting note: This diverges from the PyKeras implementation and may need to
    // change based on (de)serialization requirements.
    function get$2(identifierOrFn) {
        if (typeof identifierOrFn === 'string') {
            if (identifierOrFn in lossesMap) {
                return lossesMap[identifierOrFn];
            }
            let errMsg = `Unknown loss ${identifierOrFn}`;
            if (identifierOrFn.toLowerCase().includes('softmaxcrossentropy')) {
                errMsg = `Unknown loss ${identifierOrFn}. ` +
                    'Use "categoricalCrossentropy" as the string name for ' +
                    'tf.losses.softmaxCrossEntropy';
            }
            throw new ValueError(errMsg);
        }
        else {
            return identifierOrFn;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    function binaryAccuracy$1(yTrue, yPred) {
        return tidy(() => {
            const threshold = mul(.5, onesLike$2(yPred));
            const yPredThresholded = cast$2(greater$3(yPred, threshold), yTrue.dtype);
            return mean$1(equal$3(yTrue, yPredThresholded), -1);
        });
    }
    function categoricalAccuracy$1(yTrue, yPred) {
        return tidy(() => cast$2(equal$3(argMax$2(yTrue, -1), argMax$2(yPred, -1)), 'float32'));
    }
    function truePositives(yTrue, yPred) {
        return tidy(() => {
            return cast$3(sum$2(logicalAnd$2(equal$3(yTrue, 1), equal$3(yPred, 1))), 'float32');
        });
    }
    function falseNegatives(yTrue, yPred) {
        return tidy(() => {
            return cast$3(sum$2(logicalAnd$2(equal$3(yTrue, 1), equal$3(yPred, 0))), 'float32');
        });
    }
    function falsePositives(yTrue, yPred) {
        return tidy(() => {
            return cast$3(sum$2(logicalAnd$2(equal$3(yTrue, 0), equal$3(yPred, 1))), 'float32');
        });
    }
    function precision$1(yTrue, yPred) {
        return tidy(() => {
            const tp = truePositives(yTrue, yPred);
            const fp = falsePositives(yTrue, yPred);
            const denominator = add$2(tp, fp);
            return cast$3(where(greater$3(denominator, 0), div$1(tp, denominator), 0), 'float32');
        });
    }
    function recall$1(yTrue, yPred) {
        return tidy(() => {
            const tp = truePositives(yTrue, yPred);
            const fn = falseNegatives(yTrue, yPred);
            const denominator = add$2(tp, fn);
            return cast$3(where(greater$3(denominator, 0), div$1(tp, denominator), 0), 'float32');
        });
    }
    function binaryCrossentropy$1(yTrue, yPred) {
        return binaryCrossentropy$2(yTrue, yPred);
    }
    function sparseCategoricalAccuracy$1(yTrue, yPred) {
        if (yTrue.rank === yPred.rank) {
            yTrue = squeeze(yTrue, [yTrue.rank - 1]);
        }
        yPred = argMax$2(yPred, -1);
        if (yPred.dtype !== yTrue.dtype) {
            yPred = cast$3(yPred, yTrue.dtype);
        }
        return cast$3(equal$3(yTrue, yPred), 'float32');
    }
    // Aliases.
    const mse$1 = meanSquaredError$1;
    const MSE$1 = meanSquaredError$1;
    const mae = meanAbsoluteError$1;
    const MAE = meanAbsoluteError$1;
    const mape$1 = meanAbsolutePercentageError$1;
    const MAPE$1 = meanAbsolutePercentageError$1;
    const categoricalCrossentropy$1 = categoricalCrossentropy$2;
    const cosine = cosineProximity$1;
    const sparseCategoricalCrossentropy = sparseCategoricalCrossentropy$1;
    // TODO(cais, nielsene): Add serialize().
    const metricsMap = {
        binaryAccuracy: binaryAccuracy$1,
        categoricalAccuracy: categoricalAccuracy$1,
        precision: precision$1,
        categoricalCrossentropy: categoricalCrossentropy$1,
        sparseCategoricalCrossentropy,
        mse: mse$1,
        MSE: MSE$1,
        mae,
        MAE,
        mape: mape$1,
        MAPE: MAPE$1,
        cosine
    };
    function get$1(identifier) {
        if (typeof identifier === 'string' && identifier in metricsMap) {
            return metricsMap[identifier];
        }
        else if (typeof identifier !== 'string' && identifier != null) {
            return identifier;
        }
        else {
            throw new ValueError(`Unknown metric ${identifier}`);
        }
    }
    /**
     * Get the shortcut function name.
     *
     * If the fn name is a string,
     *   directly return the string name.
     * If the function is included in metricsMap or lossesMap,
     *   return key of the map.
     *   - If the function relative to multiple keys,
     *     return the first found key as the function name.
     *   - If the function exists in both lossesMap and metricsMap,
     *     search lossesMap first.
     * If the function is not included in metricsMap or lossesMap,
     *   return the function name.
     *
     * @param fn loss function, metric function, or short cut name.
     * @returns Loss or Metric name in string.
     */
    function getLossOrMetricName(fn) {
        assert$3(fn !== null, `Unknown LossOrMetricFn ${fn}`);
        if (typeof fn === 'string') {
            return fn;
        }
        else {
            let fnName;
            for (const key of Object.keys(lossesMap)) {
                if (lossesMap[key] === fn) {
                    fnName = key;
                    break;
                }
            }
            if (fnName !== undefined) {
                return fnName;
            }
            for (const key of Object.keys(metricsMap)) {
                if (metricsMap[key] === fn) {
                    fnName = key;
                    break;
                }
            }
            if (fnName !== undefined) {
                return fnName;
            }
            return fn.name;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    // Add (de)serialize()
    // Porting note: This diverges from the PyKeras implementation and may need to
    // change based on (de)serialization requirements.
    function getOptimizer(identifier) {
        const optimizerMap = {
            'Adagrad': () => train.adagrad(0.01),
            'Adadelta': () => train.adadelta(1, 0.95, epsilon$1()),
            'Adam': () => train.adam(0.001, 0.9, 0.999, epsilon$1()),
            'Adamax': () => train.adamax(0.002, 0.9, 0.999, epsilon$1(), 0),
            'RMSProp': () => train.rmsprop(0.001, 0.9, 0, epsilon$1()),
            'SGD': () => train.sgd(0.01)
        };
        optimizerMap['adagrad'] = optimizerMap['Adagrad'];
        optimizerMap['adadelta'] = optimizerMap['Adadelta'];
        optimizerMap['adam'] = optimizerMap['Adam'];
        optimizerMap['adamax'] = optimizerMap['Adamax'];
        optimizerMap['rmsprop'] = optimizerMap['RMSProp'];
        optimizerMap['sgd'] = optimizerMap['SGD'];
        if (identifier in optimizerMap) {
            return optimizerMap[identifier]();
        }
        throw new ValueError(`Unknown Optimizer ${identifier}`);
    }

    /**
     * @license
     * Copyright 2019 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /** Utility functions related to user-defined metadata. */
    // Maximum recommended serialized size for user-defined metadata.
    // Beyond this limit, a warning message will be printed during model loading and
    // saving.
    const MAX_USER_DEFINED_METADATA_SERIALIZED_LENGTH = 1 * 1024 * 1024;
    /**
     * Check validity of user-defined metadata.
     *
     * @param userDefinedMetadata
     * @param modelName Name of the model that the user-defined metadata belongs to.
     *   Used during construction of error messages.
     * @param checkSize Whether to check the size of the metadata is under
     *   recommended limit. Default: `false`. If `true`, will try stringify the
     *   JSON object and print a console warning if the serialzied size is above the
     *   limit.
     * @throws Error if `userDefinedMetadata` is not a plain JSON object.
     */
    function checkUserDefinedMetadata(userDefinedMetadata, modelName, checkSize = false) {
        if (userDefinedMetadata == null ||
            typeof userDefinedMetadata !== 'object' ||
            Object.getPrototypeOf(userDefinedMetadata) !== Object.prototype ||
            !plainObjectCheck(userDefinedMetadata)) {
            throw new Error('User-defined metadata is expected to be a JSON object, but is not.');
        }
        if (checkSize) {
            const out = JSON.stringify(userDefinedMetadata);
            if (out.length > MAX_USER_DEFINED_METADATA_SERIALIZED_LENGTH) {
                console.warn(`User-defined metadata of model "${modelName}" is too large in ` +
                    `size (length=${out.length} when serialized). It is not ` +
                    `recommended to store such large objects in user-defined metadata. ` +
                    `Please make sure its serialized length is <= ` +
                    `${MAX_USER_DEFINED_METADATA_SERIALIZED_LENGTH}.`);
            }
        }
    }
    /**
     * Check if an input is plain JSON object or any valid subfield of it.
     *
     * @param x The input to be checked.
     * @param assertObject Whether to assert `x` is a JSON object, i.e., reject
     *   cases of arrays and primitives.
     * @return Returns `true` if and only if `x` is a plain JSON object,
     *   a JSON-valid primitive including string, number, boolean and null,
     *   or an array of the said types.
     */
    // tslint:disable-next-line:no-any
    function plainObjectCheck(x) {
        if (x === null) {
            // Note: typeof `null` is 'object', and `null` is valid in JSON.
            return true;
        }
        else if (typeof x === 'object') {
            if (Object.getPrototypeOf(x) === Object.prototype) {
                // `x` is a JavaScript object and its prototype is Object.
                const keys = Object.keys(x);
                for (const key of keys) {
                    if (typeof key !== 'string') {
                        // JSON keys must be strings.
                        return false;
                    }
                    if (!plainObjectCheck(x[key])) { // Recursive call.
                        return false;
                    }
                }
                return true;
            }
            else {
                // `x` is a JavaScript object but its prototype is not Object.
                if (Array.isArray(x)) {
                    // `x` is a JavaScript array.
                    for (const item of x) {
                        if (!plainObjectCheck(item)) { // Recursive call.
                            return false;
                        }
                    }
                    return true;
                }
                else {
                    // `x` is a JavaScript object and its prototype is not Object,
                    // and it's not an Array. I.e., it's a complex object such as
                    // `Error` and `Date`.
                    return false;
                }
            }
        }
        else {
            // `x` is not a JavaScript object or `null`.
            const xType = typeof x;
            return xType === 'string' || xType === 'number' || xType === 'boolean';
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Print the summary of a LayersModel object.
     *
     * @param model tf.LayersModel instance.
     * @param lineLength Total length of printed lines. Set this to adapt to the
     *   display to different terminal or console sizes.
     * @param positions Relative or absolute positions of log elements in each
     *   line. Each number corresponds to right-most (i.e., ending) position of a
     *   column.
     *   If not provided, defaults to `[0.45, 0.85, 1]` for sequential-like
     *   models and `[0.33, 0.55, 0.67, 1]` for non-sequential like models.
     * @param printFn Print function to use.
     *   It will be called on each line of the summary. You can provide a custom
     *   function in order to capture the string summary. Defaults to `console.log`.
     */
    function printSummary(model, lineLength, positions, 
    // tslint:disable-next-line:no-any
    printFn = console.log) {
        const sequentialLike = isModelSequentialLike(model);
        // Header names for different log elements.
        const toDisplay = ['Layer (type)', 'Input Shape', 'Output shape', 'Param #'];
        if (sequentialLike) {
            lineLength = lineLength || 90;
            positions = positions || [0.32, 0.61, 0.89, 1];
        }
        else {
            lineLength = lineLength || 115;
            positions = positions || [0.24, 0.48, 0.70, 0.80, 1];
            // Header names for different log elements.
        }
        if (positions[positions.length - 1] <= 1) {
            // `positions` is relative. Convert it to absolute positioning.
            positions = positions.map(p => Math.floor(lineLength * p));
        }
        let relevantNodes;
        if (!sequentialLike) {
            toDisplay.push('Receives inputs');
            relevantNodes = [];
            for (const depth in model.nodesByDepth) {
                relevantNodes.push(...model.nodesByDepth[depth]);
            }
        }
        printFn('_'.repeat(lineLength));
        printRow(toDisplay, positions, printFn);
        printFn('='.repeat(lineLength));
        const layers = model.layers;
        for (let i = 0; i < layers.length; ++i) {
            if (sequentialLike) {
                printLayerSummary(layers[i], positions, printFn);
            }
            else {
                printLayerSummaryWithConnections(layers[i], positions, relevantNodes, printFn);
            }
            printFn((i === layers.length - 1 ? '=' : '_').repeat(lineLength));
        }
        // tslint:disable-next-line:no-any
        model.checkTrainableWeightsConsistency();
        const trainableCount = countTrainableParams(model);
        const nonTrainableCount = countParamsInWeights(model.nonTrainableWeights);
        printFn(`Total params: ${trainableCount + nonTrainableCount}`);
        printFn(`Trainable params: ${trainableCount}`);
        printFn(`Non-trainable params: ${nonTrainableCount}`);
        printFn('_'.repeat(lineLength));
    }
    function countTrainableParams(model) {
        let trainableCount;
        // tslint:disable:no-any
        if (model.collectedTrainableWeights != null) {
            trainableCount =
                countParamsInWeights(model.collectedTrainableWeights);
        }
        else {
            trainableCount = countParamsInWeights(model.trainableWeights);
        }
        // tslint:enable:no-any
        return trainableCount;
    }
    function isModelSequentialLike(model) {
        let sequentialLike = true;
        const nodesByDepth = [];
        const nodes = [];
        for (const depth in model.nodesByDepth) {
            nodesByDepth.push(model.nodesByDepth[depth]);
        }
        for (const depthNodes of nodesByDepth) {
            if (depthNodes.length > 1 ||
                depthNodes.length === 1 && depthNodes[0].inboundLayers.length > 1) {
                sequentialLike = false;
                break;
            }
            nodes.push(...depthNodes);
        }
        if (sequentialLike) {
            // Search for shared layers.
            for (const layer of model.layers) {
                let flag = false;
                for (const node of layer.inboundNodes) {
                    if (nodes.indexOf(node) !== -1) {
                        if (flag) {
                            sequentialLike = false;
                            break;
                        }
                        else {
                            flag = true;
                        }
                    }
                }
                if (!sequentialLike) {
                    break;
                }
            }
        }
        return sequentialLike;
    }
    function printRow(fields, positions, 
    // tslint:disable-next-line:no-any
    printFn = console.log) {
        let line = '';
        for (let i = 0; i < fields.length; ++i) {
            if (i > 0) {
                line = line.slice(0, line.length - 1) + ' ';
            }
            line += fields[i];
            line = line.slice(0, positions[i]);
            line += ' '.repeat(positions[i] - line.length);
        }
        printFn(line);
    }
    /**
     * Prints a summary for a single Layer, without connectivity information.
     *
     * @param layer: Layer instance to print.
     */
    function printLayerSummary(layer, positions, 
    // tslint:disable-next-line:no-any
    printFn) {
        let outputShape;
        let inputShape;
        try {
            inputShape = (layer.inboundNodes.map(x => JSON.stringify(x.inputShapes))).join(',');
        }
        catch (err) {
            inputShape = 'multiple';
        }
        try {
            outputShape = JSON.stringify(layer.outputShape);
        }
        catch (err) {
            outputShape = 'multiple';
        }
        const name = layer.name;
        const className = layer.getClassName();
        const fields = [`${name} (${className})`, inputShape,
            outputShape, layer.countParams().toString()];
        printRow(fields, positions, printFn);
    }
    /**
     * Prints a summary for a single Layer, with connectivity information.
     */
    function printLayerSummaryWithConnections(layer, positions, relevantNodes, 
    // tslint:disable-next-line:no-any
    printFn) {
        let outputShape;
        let inputShape;
        try {
            inputShape = (layer.inboundNodes.map(x => JSON.stringify(x.inputShapes))).join(',');
        }
        catch (err) {
            inputShape = 'multiple';
        }
        try {
            outputShape = JSON.stringify(layer.outputShape);
        }
        catch (err) {
            outputShape = 'multiple';
        }
        const connections = [];
        for (const node of layer.inboundNodes) {
            if (relevantNodes != null && relevantNodes.length > 0 &&
                relevantNodes.indexOf(node) === -1) {
                continue;
            }
            for (let i = 0; i < node.inboundLayers.length; ++i) {
                const inboundLayer = node.inboundLayers[i].name;
                const inboundLayerIndex = node.nodeIndices[i];
                const inboundTensorIndex = node.tensorIndices[i];
                connections.push(`${inboundLayer}[${inboundLayerIndex}][${inboundTensorIndex}]`);
            }
        }
        const name = layer.name;
        const className = layer.getClassName();
        const firstConnection = connections.length === 0 ? '' : connections[0];
        const fields = [
            `${name} (${className})`, inputShape,
            outputShape, layer.countParams().toString(),
            firstConnection
        ];
        printRow(fields, positions, printFn);
        for (let i = 1; i < connections.length; ++i) {
            printRow(['', '', '', '', connections[i]], positions, printFn);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    // tslint:enable
    /**
     * Test whether a value in an array is the name of a LayersModel or Layer.
     * @param key The key name that the value is found under. Note that the key
     *   may not be at the level immediately above the value, if the value is in a
     *   nested array.
     * @param index Index of the value in the Array that it is found in.
     * @param value The value object.
     * @returns A boolean indicating whether value is a name.
     */
    function isArrayItemInputOrOutputName(key, index, value) {
        return (key === 'inboundNodes' || key === 'outputLayers' ||
            key === 'inputLayers') &&
            index === 0 && typeof value === 'string';
    }
    /**
     * Convert a Pythonic config object to TypeScript config object.
     * @param pythonicConfig The config object to convert.
     * @param key Optional key name of the object being converted.
     * @returns Result of the conversion.
     */
    function convertPythonicToTs(pythonicConfig, key) {
        if (pythonicConfig === null) {
            return null;
        }
        else if (typeof pythonicConfig === 'string') {
            return toCamelCase(pythonicConfig);
        }
        else if ((typeof pythonicConfig === 'number') ||
            (typeof pythonicConfig === 'boolean')) {
            return pythonicConfig;
        }
        else if (pythonicConfig instanceof Array) {
            const tsArray = [];
            const arrayLength = pythonicConfig.length;
            for (let i = 0; i < arrayLength; ++i) {
                const item = pythonicConfig[i];
                if (isArrayItemInputOrOutputName(key, i, item)) {
                    tsArray.push(item);
                }
                else {
                    tsArray.push(convertPythonicToTs(item, key));
                }
            }
            return tsArray;
        }
        else {
            const tsDict = {};
            for (const pythonicKey of Object.keys(pythonicConfig)) {
                const pythonicValue = pythonicConfig[pythonicKey];
                if (pythonicKey === 'name' && typeof pythonicValue === 'string') {
                    // Special case the 'name' key with a string value. Name values, such as
                    // the names of LayersModel and Layer instances, should not undergo the
                    // camel-case conversion.
                    tsDict[pythonicKey] = pythonicValue;
                }
                else {
                    const tsKey = toCamelCase(pythonicKey);
                    tsDict[tsKey] = convertPythonicToTs(pythonicValue, tsKey);
                }
            }
            return tsDict;
        }
    }
    /**
     * Convert a TypeScript config object to Python config object.
     * @param tsConfig The config object to convert.
     * @param key Optional key name of the object being converted.
     * @returns Result of the conversion.
     */
    function convertTsToPythonic(tsConfig, key) {
        if (tsConfig === null || tsConfig === undefined) {
            return null;
        }
        else if (typeof tsConfig === 'string') {
            return toSnakeCase(tsConfig);
        }
        else if ((typeof tsConfig === 'number') || (typeof tsConfig === 'boolean')) {
            return tsConfig;
        }
        else if (tsConfig instanceof Array) {
            const pyArray = [];
            const arrayLength = tsConfig.length;
            for (let i = 0; i < arrayLength; ++i) {
                const item = tsConfig[i];
                if (isArrayItemInputOrOutputName(key, i, item)) {
                    pyArray.push(item);
                }
                else {
                    pyArray.push(convertTsToPythonic(item, key));
                }
            }
            return pyArray;
        }
        else {
            const pyDict = {};
            for (const tsKey of Object.keys(tsConfig)) {
                const tsValue = tsConfig[tsKey];
                const pyKey = toSnakeCase(tsKey);
                if ((tsKey === 'name' || tsKey === 'className') &&
                    typeof tsValue === 'string') {
                    // Special case the 'name' key with a string value. Name values, such as
                    // the names of LayersModel and Layer instances, should not undergo the
                    // snake-case conversion.
                    pyDict[pyKey] = tsValue;
                }
                else {
                    pyDict[pyKey] = convertTsToPythonic(tsValue, tsKey);
                }
            }
            return pyDict;
        }
    }

    /** @license See the LICENSE file. */
    // This code is auto-generated, do not modify this file!
    const version$6 = '3.17.0';

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * A Container is a directed acyclic graph of layers.
     *
     * It is the topological form of a "model". A LayersModel
     * is simply a Container with added training routines.
     *
     */
    class Container extends Layer {
        constructor(args) {
            // No args passed to super's constructor.
            super({});
            this.containerNodes = new Set();
            this.name = args.name;
            if (this.name == null) {
                const prefix = this.getClassName().toLowerCase();
                this.name = getUid(prefix);
            }
            this.supportsMasking = false;
            this.trainable_ = true;
            // TODO(michaelterry): Initialize perInputLosses/Updates here.
            // Container-specific properties.
            if (Array.isArray(args.inputs)) {
                this.inputs = args.inputs.slice();
            }
            else {
                this.inputs = [args.inputs];
            }
            if (Array.isArray(args.outputs)) {
                this.outputs = args.outputs.slice();
            }
            else {
                this.outputs = [args.outputs];
            }
            // Check for redundancy in inputs.
            if (unique$2(this.inputs).length !== this.inputs.length) {
                throw new ValueError('The list of inputs passed to the model is ' +
                    'redundant. All inputs should only appear once. Found: ' +
                    `${this.inputs.map(x => x.name)}`);
            }
            // Check for redundancy in outputs.
            if (unique$2(this.outputs).length !== this.outputs.length) {
                console.warn('The list of outputs passed to the model is redundant. ' +
                    'All outputs should only appear once. Found: ' +
                    `${this.outputs.map(x => x.name)}`);
            }
            /*
              List of initial layers (1 to 1 mapping with this.inputs, hence the same
              layer might appear twice)
            */
            this.inputLayers = [];
            this.inputLayersNodeIndices = [];
            this.inputLayersTensorIndices = [];
            /*
              List of layers (1 to 1 mapping with this.outputs, hence the same layer
              might appear twice)
            */
            this.outputLayers = [];
            this.outputLayersNodeIndices = [];
            this.outputLayersTensorIndices = [];
            /*
              All layers in order of horizontal graph traversal. Entries are unique.
              Includes input and output layers.
            */
            this.layers = [];
            /*
              References to container layers that were constructed internally. We need
              these to properly dispose of tensors from nested containers.
            */
            this.internalContainerRefs = [];
            // TODO(michaelterry): Determine if caching still needed with eager
            // backend.
            /*
              This is for performance optimization when calling the Container on new
              inputs. Every time the Container is called on a set on input tensors,
              we compute the output tensors, output masks and output shapes in one pass,
              then cache them here. When one of these outputs is queried later,
              we retrieve it from there instead of recomputing it.
            */
            // this.outputTensorCache = {};
            // this.outputShapeCache = {};
            // Build this.outputLayers:
            for (const x of this.outputs) {
                const layer = x.sourceLayer;
                const nodeIndex = x.nodeIndex;
                const tensorIndex = x.tensorIndex;
                this.outputLayers.push(layer);
                this.outputLayersNodeIndices.push(nodeIndex);
                this.outputLayersTensorIndices.push(tensorIndex);
            }
            // TODO(michaelterry): Add output mask cache code.
            // Build this.inputLayers:
            for (const x of this.inputs) {
                const layer = x.sourceLayer;
                const nodeIndex = x.nodeIndex;
                const tensorIndex = x.tensorIndex;
                /*
                  It's supposed to be an input layer, so only one node
                  and one tensor output.
                */
                assert$3(nodeIndex === 0, 'input layer has >1 nodes');
                assert$3(tensorIndex === 0, 'input layer has >1 tensors');
                this.inputLayers.push(layer);
                this.inputLayersNodeIndices.push(nodeIndex);
                this.inputLayersTensorIndices.push(tensorIndex);
            }
            // Build this.inputNames and this.outputNames.
            this.inputNames = [];
            this.outputNames = [];
            this.feedInputShapes = [];
            this.feedInputNames = [];
            this.feedOutputNames = [];
            for (let i = 0; i < this.inputLayers.length; i++) {
                const layer = this.inputLayers[i];
                // Check that layer is an InputLayer.
                if (!(layer instanceof InputLayer)) {
                    throw new TypeError('Input layers to a LayersModel must be InputLayer objects. ' +
                        `Received inputs: ${args.inputs}. ` +
                        `Input ${i} (0-based) originates ` +
                        `from layer type ${layer.getClassName()}.`);
                }
                this.inputNames.push(layer.name);
                this.feedInputShapes.push(layer.batchInputShape);
                this.feedInputNames.push(layer.name);
            }
            for (const layer of this.outputLayers) {
                this.outputNames.push(layer.name);
            }
            this.internalInputShapes = this.inputs.map(x => x.shape);
            this.internalOutputShapes = this.outputs.map(x => x.shape);
            /*
              Container_nodes: set of nodes included in the graph (not all nodes
              included in the layers are relevant to the current graph).
            */
            // ids of all nodes relevant to the Container:
            const nodesDepths = {};
            // To recover nodes from their ID.
            const nodeIDToNode = {};
            const layersDepths = {};
            // To layers from their ID.
            const layerIDToLayer = {};
            const layerIndices = {};
            const nodesInDecreasingDepth = [];
            /**
             * Builds a map of the graph of layers.
             *
             * This recursively updates the map `layerIndices`,
             * the list `nodesInDecreasingDepth` and the set `containerNodes`.
             *
             * @param tensor Some tensor in a graph.
             * @param finishedNodes Set of nodes whose subgraphs have been traversed
             *         completely. Useful to prevent duplicated work.
             * @param nodesInProgress Set of nodes that are currently active on the
             *         recursion stack. Useful to detect cycles.
             * @param layer Layer from which `tensor` comes from. If not provided,
             *   will be obtained from tensor.sourceLayer.
             * @param nodeIndex Node index from which `tensor` comes from.
             * @param tensorIndex TensorIndex from which `tensor` comes from.
             *
             * @exception RuntimeError if a cycle is detected.
             */
            const buildMapOfGraph = (tensor, finishedNodes, nodesInProgress, layer, nodeIndex, tensorIndex) => {
                if (layer == null || nodeIndex == null || tensorIndex == null) {
                    layer = tensor.sourceLayer;
                    nodeIndex = tensor.nodeIndex;
                    tensorIndex = tensor.tensorIndex;
                }
                const node = layer.inboundNodes[nodeIndex];
                // Prevent cycles.
                if (nodesInProgress.indexOf(node) !== -1) {
                    throw new RuntimeError(`The tensor ${tensor.name} at layer "${layer.name}" ` +
                        'is part of a cycle.');
                }
                // Don't repeat work for shared subgraphs
                if (finishedNodes.indexOf(node) !== -1) {
                    return;
                }
                // Update containerNodes.
                this.containerNodes.add(Container.nodeKey(layer, nodeIndex));
                // Store the traversal order for layer sorting.
                if (!(layer.id in layerIndices)) {
                    layerIndices[layer.id] = Object.keys(layerIndices).length;
                }
                if (nodesInProgress.indexOf(node) === -1) {
                    nodesInProgress.push(node);
                }
                // Propagate to all previous tensors connected to this node.
                const numInboundLayers = node.inboundLayers.length;
                for (let i = 0; i < numInboundLayers; i++) {
                    const x = node.inputTensors[i];
                    const layer = node.inboundLayers[i];
                    const nodeIndex = node.nodeIndices[i];
                    const tensorIndex = node.tensorIndices[i];
                    buildMapOfGraph(x, finishedNodes, nodesInProgress, layer, nodeIndex, tensorIndex);
                }
                finishedNodes.push(node);
                while (nodesInProgress.indexOf(node) >= 0) {
                    nodesInProgress.splice(nodesInProgress.indexOf(node), 1);
                }
                nodesInDecreasingDepth.push(node);
            };
            const finishedNodes = [];
            const nodesInProgress = [];
            for (const x of this.outputs) {
                buildMapOfGraph(x, finishedNodes, nodesInProgress);
            }
            const reversedNodesInDecreasingDepth = nodesInDecreasingDepth.slice().reverse();
            for (const node of reversedNodesInDecreasingDepth) {
                nodeIDToNode[node.id] = node;
                // If the depth is not set, the node has no outbound nodes (depth 0).
                if (!(node.id in nodesDepths)) {
                    nodesDepths[node.id] = 0;
                }
                let depth = nodesDepths[node.id];
                // Update the depth of the corresponding layer
                const previousDepth = (layersDepths[node.outboundLayer.id] == null ?
                    0 :
                    layersDepths[node.outboundLayer.id]);
                /*
                  If we've seen this layer before at a higher depth, we should use that
                  depth instead of the node depth.  This is necessary for shared layers
                  that have inputs at different depth levels in the graph.
                */
                depth = Math.max(depth, previousDepth);
                layersDepths[node.outboundLayer.id] = depth;
                layerIDToLayer[node.outboundLayer.id] = node.outboundLayer;
                nodesDepths[node.id] = depth;
                // Update the depth of inbound nodes.
                for (let i = 0; i < node.inboundLayers.length; i++) {
                    const inboundLayer = node.inboundLayers[i];
                    const nodeIndex = node.nodeIndices[i];
                    const inboundNode = inboundLayer.inboundNodes[nodeIndex];
                    const previousDepth = (nodesDepths[inboundNode.id] == null ? 0 :
                        nodesDepths[inboundNode.id]);
                    nodesDepths[inboundNode.id] = Math.max(depth + 1, previousDepth);
                    nodeIDToNode[inboundNode.id] = inboundNode;
                }
            }
            // Build a dict {depth: list of nodes with this depth}
            const nodesByDepth = {};
            for (const nodeID in nodesDepths) {
                const depth = nodesDepths[nodeID];
                if (!(depth in nodesByDepth)) {
                    nodesByDepth[depth] = [];
                }
                nodesByDepth[depth].push(nodeIDToNode[nodeID]);
            }
            // Build a dict {depth: list of layers with this depth}
            const layersByDepth = {};
            for (const layerID in layersDepths) {
                const depth = layersDepths[layerID];
                if (!(depth in layersByDepth)) {
                    layersByDepth[depth] = [];
                }
                layersByDepth[depth].push(layerIDToLayer[layerID]);
            }
            // Get sorted list of layer depths.
            let depthKeys = Object.keys(layersByDepth)
                .map(x => parseInt(x, 10))
                .sort(reverseNumberCompare);
            // Set this.layers and this.layersByDepth.
            this.layers = [];
            for (const depth of depthKeys) {
                const layersForDepth = layersByDepth[depth];
                // Container.layers needs to have a deterministic order:
                // here we order them by traversal order.
                layersForDepth.sort((a, b) => {
                    const aIndex = layerIndices[a.id];
                    const bIndex = layerIndices[b.id];
                    if (aIndex < bIndex) {
                        return -1;
                    }
                    if (aIndex > bIndex) {
                        return 1;
                    }
                    return 0;
                });
                for (const layer of layersForDepth) {
                    if (layer instanceof Container) {
                        this.internalContainerRefs.push(layer);
                    }
                    this.layers.push(layer);
                }
            }
            this.layersByDepth = layersByDepth;
            // Get sorted list of node depths;
            depthKeys = Object.keys(nodesByDepth)
                .map(x => parseInt(x, 10))
                .sort(reverseNumberCompare);
            // Check that all tensors required are computable.
            // computable_tensors: all tensors in the graph
            // that can be computed from the inputs provided.
            const computableTensors = this.inputs.slice();
            // To provide a better error msg.
            const layersWithCompleteInput = [];
            for (const depth of depthKeys) {
                for (const node of nodesByDepth[depth]) {
                    const layer = node.outboundLayer;
                    if (layer != null) {
                        for (const x of node.inputTensors) {
                            if (computableTensors.indexOf(x) === -1) {
                                throw new RuntimeError(`Graph disconnected: cannot obtain value for tensor ${x}` +
                                    ` at layer "${layer.name}". ` +
                                    'The following previous layers were accessed without ' +
                                    `issue: ${layersWithCompleteInput}`);
                            }
                        }
                        for (const x of node.outputTensors) {
                            computableTensors.push(x);
                        }
                        layersWithCompleteInput.push(layer.name);
                    }
                }
            }
            // Set this.containerNodes and this.nodesByDepth.
            this.nodesByDepth = nodesByDepth;
            // Ensure name unicity, which will be crucial for serialization
            // (since serialized nodes refer to layers by their name).
            const allNames = this.layers.map(x => x.name);
            for (const name of allNames) {
                const numOccurrences = allNames.filter(x => x === name).length;
                if (numOccurrences !== 1) {
                    throw new RuntimeError(`The name "${name}" is used ${numOccurrences} times ` +
                        'in the model. All layer names should be unique. Layer names: ' +
                        JSON.stringify(allNames));
                }
            }
            // Layer parameters.
            // The new container starts with a single inbound node
            // for its inputs, and no outbound nodes.
            // Will be appended to by future calls to apply().
            this.outboundNodes = [];
            // Will be appended to below, and by future calls to apply().
            this.inboundNodes = [];
            // Create the node linking internal inputs to internal outputs.
            // (This call has side effects.)
            // tslint:disable-next-line:no-unused-expression
            new Node({
                outboundLayer: this,
                inboundLayers: [],
                nodeIndices: [],
                tensorIndices: [],
                inputTensors: this.inputs,
                outputTensors: this.outputs,
                inputMasks: this.inputs.map(x => null),
                outputMasks: this.outputs.map(x => null),
                inputShapes: this.inputs.map(x => x.shape),
                outputShapes: this.outputs.map(x => x.shape)
            });
            this.built = true;
            this._refCount = 1; // The ref count of a container always start at 1.
        }
        assertNotDisposed() {
            if (this._refCount === 0) {
                throw new Error(`Container '${this.name}' is already disposed.`);
            }
        }
        /**
         * Attempt to dispose a LayersModel's weights.
         *
         * This method decrease the reference count of the LayersModel object by 1.
         *
         * A LayersModel is reference-counted. Its reference count is incremented by 1
         * when it is first constructed and when it is used as a Layer of another
         * LayersModel.
         *
         * If the reference count of a LayersModel becomes 0, the `dispose` method of
         * all its constituent `Layer`s will be called.
         *
         * Note: If the reference count is greater than 0 after the decrement, the
         * `dispose` method of its constituent `Layer`s will *not* be called.
         *
         * After a LayersModel is disposed, it cannot be used in calls such as
         * 'predict`, `evaluate` or `fit` anymore.
         *
         * @returns A DisposeResult Object with the following fields:
         *   - refCountAfterDispose: The reference count of the LayersModel after this
         *     `dispose()` call.
         *   - numDisposedVariables: Number of `tf.Variable`s (i.e., weights) disposed
         *     during this `dispose()` call.
         * @throws {Error} If the layer is not built yet, or if the LayersModel has
         *   already been disposed.
         */
        dispose() {
            this.assertNotDisposed();
            const result = { refCountAfterDispose: null, numDisposedVariables: 0 };
            if (--this._refCount === 0) {
                for (const layer of this.layers) {
                    result.numDisposedVariables += layer.dispose().numDisposedVariables;
                }
                // Call dispose on each internally created container layer again to ensure
                // their refCounts hit zero and their tensors are subsequently deleted.
                for (const container of this.internalContainerRefs) {
                    result.numDisposedVariables += container.dispose().numDisposedVariables;
                }
            }
            result.refCountAfterDispose = this._refCount;
            return result;
        }
        get trainable() {
            return this.trainable_;
        }
        set trainable(trainable) {
            this.layers.forEach(layer => {
                // tslint:disable-next-line:no-any
                layer._trainableWeights
                    .forEach(w => w.trainable = trainable);
            });
            this.trainable_ = trainable;
        }
        get trainableWeights() {
            // Porting Note: This check below is to prevent errors where the
            //   _trainableWeights inherited from the parent class (Layer) gets
            //   inadvertently used.
            if (this._trainableWeights.length > 0) {
                throw new ValueError('Container instance unexpectedly contains _trainableWeights.' +
                    'The trainable weights of a Container are a union of the ' +
                    'trainable weights of its consituent Layers. Its own ' +
                    '_trainableWeights must remain an empty Array.');
            }
            if (!this.trainable) {
                return [];
            }
            let weights = [];
            for (const layer of this.layers) {
                weights = weights.concat(layer.trainableWeights);
            }
            return weights;
        }
        get nonTrainableWeights() {
            const weights = [];
            for (const layer of this.layers) {
                weights.push(...layer.nonTrainableWeights);
            }
            if (!this.trainable) {
                const trainableWeights = [];
                for (const layer of this.layers) {
                    trainableWeights.push(...layer.trainableWeights);
                }
                return trainableWeights.concat(weights);
            }
            return weights;
        }
        get weights() {
            return this.trainableWeights.concat(this.nonTrainableWeights);
        }
        /**
         * Loads all layer weights from a JSON object.
         *
         * Porting Note: HDF5 weight files cannot be directly loaded in JavaScript /
         *   TypeScript. The utility script at `scripts/pykeras.py` offers means
         *   to convert them into JSON strings compatible with this method.
         * Porting Note: TensorFlow.js Layers supports only loading by name currently.
         *
         * @param weights A JSON mapping weight names to weight values as nested
         *   arrays of numbers, or a `NamedTensorMap`, i.e., a JSON mapping weight
         *   names to `tf.Tensor` objects.
         * @param strict Require that the provided weights exactly match those
         *   required by the container.  Default: `true`.  Passing `false` means that
         *   extra weights and missing weights will be silently ignored.
         */
        loadWeights(weights, strict = true) {
            const nameToWeight = {};
            let totalWeightsCount = 0;
            for (const layer of this.layers) {
                for (const weight of layer.weights) {
                    if (nameToWeight[weight.originalName] != null) {
                        throw new ValueError(`Duplicate weight name: ${weight.originalName}`);
                    }
                    nameToWeight[weight.originalName] = weight;
                    totalWeightsCount++;
                }
            }
            const weightValueTuples = [];
            for (const name in weights) {
                // TF 2.2.0 added cell name to the weight name in the format of
                // layer_name/cell_name/weight_name, we need to remove
                // the inner cell name.
                let validatedName = name;
                if (nameToWeight[name] == null) {
                    const tokens = name.split('/');
                    const shortenNameArray = tokens.slice(0, -2).concat([tokens[tokens.length - 1]]);
                    validatedName = shortenNameArray.join('/');
                }
                if (nameToWeight[validatedName] != null) {
                    weightValueTuples.push([nameToWeight[validatedName], weights[name]]);
                }
                else if (strict) {
                    throw new ValueError(`Provided weight data has no target variable: ${name}`);
                }
                delete nameToWeight[validatedName];
            }
            if (strict) {
                // Check that all weights are set.
                const unsetNames = [];
                for (const name in nameToWeight) {
                    unsetNames.push(name);
                }
                if (unsetNames.length > 0) {
                    throw new ValueError(`${unsetNames.length} of ${totalWeightsCount} weights are not set: ` +
                        `${unsetNames}`);
                }
            }
            batchSetValue(weightValueTuples);
        }
        /**
         * Util shared between different serialization methods.
         * @returns LayersModel config with Keras version information added.
         */
        updatedConfig() {
            const theConfig = this.getConfig();
            const modelConfig = {};
            modelConfig['className'] = this.getClassName();
            modelConfig['config'] = theConfig;
            modelConfig['kerasVersion'] = `tfjs-layers ${version$6}`;
            // TODO(nielsene): Replace something like K.backend() once
            // possible.
            modelConfig['backend'] = 'TensorFlow.js';
            return modelConfig;
        }
        /**
         * Returns a JSON string containing the network configuration.
         *
         * To load a network from a JSON save file, use
         * models.modelFromJSON(jsonString);
         * @param extraJsonArgs Unused in tfjs-layers, maintained for PyKeras
         * @param returnString Whether the return value should be stringified
         *    (default: `true`).
         * @returns a JSON string if `returnString` (default), or a JSON object if
         *   `!returnString`.
         */
        // tslint:disable-next-line:no-any
        toJSON(unused, returnString = true) {
            const modelConfig = convertTsToPythonic(this.updatedConfig());
            return returnString ? JSON.stringify(modelConfig) : modelConfig;
        }
        /**
         * Call the model on new inputs.
         *
         * In this case `call` just reapplies all ops in the graph to the new inputs
         * (e.g. build a new computational graph from the provided inputs).
         *
         * @param inputs A tensor or list of tensors.
         * @param mask A mask or list of masks. A mask can be either a tensor or null
         *   (no mask).
         *
         * @return A tensor if there is a single output, or a list of tensors if there
         *   are more than one outputs.
         */
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = toList(inputs);
                const feedDict = new FeedDict();
                for (let i = 0; i < this.inputs.length; ++i) {
                    feedDict.add(this.inputs[i], inputs[i]);
                }
                return execute(this.outputs, feedDict, kwargs);
            });
        }
        /**
         * Computes an output mask tensor.
         *
         * @param inputs Tensor or list of tensors.
         * @param mask Tensor or list of tensors.
         *
         * @return null or a tensor (or list of tensors, one per output tensor of the
         * layer).
         */
        computeMask(inputs, mask) {
            return tidy(() => {
                inputs = toList(inputs);
                let masks;
                if (mask == null) {
                    masks = pyListRepeat(null, inputs.length);
                }
                else {
                    masks = toList(mask);
                }
                // TODO(michaelterry): Add support for mask caching.
                return this.runInternalGraph(inputs, masks)[1];
            });
        }
        /**
         * Computes the output shape of the layer.
         *
         * Assumes that the layer will be built to match that input shape provided.
         *
         * @param inputShape A shape (tuple of integers) or a list of shape tuples
         *   (one per output tensor of the layer). Shape tuples can include null for
         *   free dimensions, instead of an integer.
         */
        computeOutputShape(inputShape) {
            const inputShapes = normalizeShapeList(inputShape);
            if (inputShapes.length !== this.inputLayers.length) {
                throw new ValueError(`Invalid inputShape argument ${inputShape}: ` +
                    `model has ${this.inputLayers.length} tensor inputs.`);
            }
            // TODO(michaelterry): Add caching
            const layersToOutputShapes = {};
            for (let i = 0; i < inputShapes.length; i++) {
                const layer = this.inputLayers[i];
                const inputShape = inputShapes[i];
                // It's an input layer: computeOutputShape is identity,
                // and there is only one node and one tensor output.
                const shapeKey = layer.name + '_0_0';
                layersToOutputShapes[shapeKey] = inputShape;
            }
            const depthKeys = Object.keys(this.nodesByDepth)
                .map(x => parseInt(x, 10))
                .sort(reverseNumberCompare);
            // Iterate over nodes, by depth level.
            if (depthKeys.length > 1) {
                for (const depth of depthKeys) {
                    const nodes = this.nodesByDepth[depth];
                    for (const node of nodes) {
                        // This is always a single layer, never a list.
                        const layer = node.outboundLayer;
                        if (this.inputLayers.map(x => x.id).indexOf(layer.id) !== -1) {
                            // We've already covered the input layers a few lines above.
                            continue;
                        }
                        // Potentially redundant list, same size of node.inputTensors.
                        const inputShapes = [];
                        for (let j = 0; j < node.inboundLayers.length; j++) {
                            const inboundLayer = node.inboundLayers[j];
                            const nodeIndex = node.nodeIndices[j];
                            const tensorIndex = node.tensorIndices[j];
                            const shapeKey = `${inboundLayer.name}_${nodeIndex}_${tensorIndex}`;
                            const inputShape = layersToOutputShapes[shapeKey];
                            inputShapes.push(inputShape);
                        }
                        const outputShape = layer.computeOutputShape(singletonOrArray(inputShapes));
                        const outputShapes = normalizeShapeList(outputShape);
                        const nodeIndex = layer.inboundNodes.indexOf(node);
                        for (let j = 0; j < outputShapes.length; j++) {
                            const shapeKey = `${layer.name}_${nodeIndex}_${j}`;
                            layersToOutputShapes[shapeKey] = outputShapes[j];
                        }
                    }
                }
            }
            // Read final output shapes from layersToOutputShapes.
            const outputShapes = [];
            const outputShapeKeys = [];
            for (let i = 0; i < this.outputLayers.length; i++) {
                const layer = this.outputLayers[i];
                const nodeIndex = this.outputLayersNodeIndices[i];
                const tensorIndex = this.outputLayersTensorIndices[i];
                const shapeKey = `${layer.name}_${nodeIndex}_${tensorIndex}`;
                outputShapeKeys.push(shapeKey);
            }
            for (let i = 0; i < outputShapeKeys.length; i++) {
                const key = outputShapeKeys[i];
                assert$3(key in layersToOutputShapes);
                outputShapes.push(layersToOutputShapes[key]);
            }
            // TODO(michaelterry): Update cache
            return singletonOrArray(outputShapes);
        }
        /**
         * Computes output tensors for new inputs.
         *
         * Note:
         *   - Expects `inputs` to be a list (potentially with 1 element).
         *
         * @param inputs List of tensors
         * @param masks List of masks (tensors or null).
         * @return Three lists: outputTensors, outputMasks, outputShapes
         */
        runInternalGraph(inputs, masks) {
            if (masks == null) {
                masks = pyListRepeat(null, inputs.length);
            }
            // Dictionary mapping reference tensors to tuples
            // (computed tensor, compute mask)
            // we assume a 1:1 mapping from tensor to mask
            // TODO: raise exception when a `.computeMask()` call
            // does not return a list the same size as `call`
            const tensorMap = {};
            for (let i = 0; i < this.inputs.length; ++i) {
                const x = this.inputs[i];
                const y = inputs[i];
                const mask = masks[i];
                tensorMap[x.id] = [y, mask];
            }
            const depthKeys = Object.keys(this.nodesByDepth)
                .map(x => parseInt(x, 10))
                .sort(reverseNumberCompare);
            for (const depth of depthKeys) {
                const nodes = this.nodesByDepth[depth];
                for (const node of nodes) {
                    // This is always a single layer, never a list.
                    const layer = node.outboundLayer;
                    const referenceInputTensors = node.inputTensors;
                    const referenceOutputTensors = node.outputTensors;
                    // If all previous input tensors are available in tensorMap,
                    // then call node.inboundLayer on them.
                    // List of tuples [input, mask]:
                    const computedData = new Array();
                    for (const x of referenceInputTensors) {
                        if (x.id in tensorMap) {
                            computedData.push(tensorMap[x.id]);
                        }
                    }
                    if (computedData.length === referenceInputTensors.length) {
                        // TODO(michaelterry): Add K.name_scope here, if we need it.
                        let kwargs = {};
                        let computedTensors;
                        let computedMasks;
                        let outputTensors;
                        let outputMasks;
                        // call layer
                        if (node.callArgs != null) {
                            kwargs = node.callArgs;
                        }
                        if (computedData.length === 1) {
                            const [computedTensor, computedMask] = computedData[0];
                            if (kwargs['mask'] == null) {
                                kwargs['mask'] = computedMask;
                            }
                            outputTensors =
                                toList(layer.call(computedTensor, kwargs));
                            outputMasks = toList(layer.computeMask(computedTensor, computedMask));
                            computedTensors = [computedTensor];
                            computedMasks = [computedMask];
                        }
                        else {
                            computedTensors = computedData.map(x => x[0]);
                            computedMasks = computedData.map(x => x[1]);
                            if (kwargs['mask'] == null) {
                                kwargs['mask'] = computedMasks;
                            }
                            outputTensors =
                                toList(layer.call(computedTensors, kwargs));
                            outputMasks = toList(layer.computeMask(computedTensors, computedMasks));
                        }
                        if (layer.activityRegularizer) {
                            throw new NotImplementedError('LayersModel invocation with concrete Tensor value(s) in the ' +
                                'presence of activity regularizer(s) is not supported yet.');
                        }
                        // TODO(michaelterry): Add model updates and losses
                        // Update tensor map.
                        for (let i = 0; i < referenceOutputTensors.length; ++i) {
                            const x = referenceOutputTensors[i];
                            const y = outputTensors[i];
                            const mask = outputMasks[i];
                            tensorMap[x.id] = [y, mask];
                        }
                    }
                }
            }
            const outputTensors = [];
            const outputMasks = [];
            const outputShapes = [];
            for (const x of this.outputs) {
                assert$3(x.id in tensorMap, `Could not compute output ${x.name} : ${x.id}`);
                const [tensor, mask] = tensorMap[x.id];
                outputShapes.push(tensor.shape);
                outputTensors.push(tensor);
                outputMasks.push(mask);
            }
            // TODO(michaelterry): Add support for caches.
            return [outputTensors, outputMasks, outputShapes];
        }
        /**
         * Builds a map of internal node keys to node ordering.
         * Used in serializaion a node orderings may change as unused nodes are
         * dropped. Porting Note:  This helper method was pulled out of getConfig to
         * improve readability.
         * @param layers An array of Layers in the model.
         * @returns Map of Node Keys to index order within the layer.
         */
        buildNodeConversionMap(layers) {
            const nodeConversionMap = {};
            let keptNodes;
            for (const layer of this.layers) {
                keptNodes = layer instanceof Container ? 1 : 0;
                for (let originalNodeIndex = 0; originalNodeIndex < layer.inboundNodes.length; originalNodeIndex++) {
                    const nodeKey = Container.nodeKey(layer, originalNodeIndex);
                    if (this.containerNodes.has(nodeKey)) {
                        // i.e. we mark it to be saved
                        nodeConversionMap[nodeKey] = keptNodes;
                        keptNodes += 1;
                    }
                }
            }
            return nodeConversionMap;
        }
        /**
         * Retrieves a layer based on either its name (unique) or index.
         *
         * Indices are based on order of horizontal graph traversal (bottom-up).
         *
         * If both `name` and `index` are specified, `index` takes precedence.
         *
         * @param name Name of layer.
         * @param index Index of layer.
         * @returns A Layer instance.
         * @throws ValueError: In case of invalid layer name or index.
         *
         * @doc {
         *    heading: 'Layers',
         *    subheading: 'Classes',
         *    namespace: 'layers',
         *    subclasses: ['LayersModel']
         * }
         */
        getLayer(name, index) {
            if (index != null) {
                if (this.layers.length <= index) {
                    throw new ValueError(`Was asked to retrieve layer at index ${index}, but model only ` +
                        `has ${this.layers.length} layer(s).`);
                }
                else {
                    return this.layers[index];
                }
            }
            else {
                if (name == null) {
                    throw new ValueError('Provide either a layer name or layer index');
                }
            }
            for (const layer of this.layers) {
                if (layer.name === name) {
                    return layer;
                }
            }
            throw new ValueError(`No such layer: ${name}`);
        }
        /**
         * Retrieves the Container's current loss values.
         *
         * Used for regularizers during training.
         */
        calculateLosses() {
            // Porting Node: This is an augmentation to Container.loss in PyKeras.
            //   In PyKeras, Container.loss returns symbolic tensors. Here a concrete
            //   Tensor (specifically Scalar) values are returned. This is due to the
            //   imperative backend.
            return tidy(() => {
                const losses = [];
                for (const layer of this.layers) {
                    for (let nodeIndex = 0; nodeIndex < layer.inboundNodes.length; ++nodeIndex) {
                        const nodeKey = Container.nodeKey(layer, nodeIndex);
                        if (this.containerNodes.has(nodeKey)) {
                            losses.push(...layer.calculateLosses());
                        }
                    }
                }
                // TODO(cais): Add any unconditional model-level losses?
                return losses;
            });
        }
        getConfig() {
            const config = { name: this.name };
            // Build a map from layer unique name (self._node_key)
            // to the index of the nodes that are saved in the config.
            // Only nodes in container_nodes are saved.
            const nodeConversionMap = this.buildNodeConversionMap(this.layers);
            // Serialize and save the layers in layerConfigs
            const layerConfigs = [];
            for (const layer of this.layers) {
                const layerClassName = layer.getClassName();
                const layerConfig = layer.getConfig();
                const filteredInboundNodes = [];
                for (let originalNodeIndex = 0; originalNodeIndex < layer.inboundNodes.length; originalNodeIndex++) {
                    const node = layer.inboundNodes[originalNodeIndex];
                    const nodeKey = Container.nodeKey(layer, originalNodeIndex);
                    let kwargs = {};
                    if (this.containerNodes.has(nodeKey)) {
                        // The node is relevant to the model:
                        // add to filteredInboundNodes.
                        if (node.callArgs) {
                            try {
                                JSON.stringify(node.callArgs);
                                kwargs = node.callArgs;
                            }
                            catch (err) {
                                console.warn(`Layer ${layer.name} was passed ` +
                                    `non-serializable keyword arguments: ` +
                                    `${node.callArgs}. They will not be included ` +
                                    `in the serialized model (and thus will be ` +
                                    `missing at deserialization time).`);
                                kwargs = {};
                            }
                        }
                        if (node.inboundLayers.length > 0) {
                            const nodeData = [];
                            for (let i = 0; i < node.inboundLayers.length; i++) {
                                const inboundLayer = node.inboundLayers[i];
                                const nodeIndex = node.nodeIndices[i];
                                const tensorIndex = node.tensorIndices[i];
                                const nodeKey = Container.nodeKey(inboundLayer, nodeIndex);
                                let newNodeIndex = nodeConversionMap[nodeKey];
                                if (newNodeIndex == null) {
                                    newNodeIndex = 0;
                                }
                                nodeData.push([inboundLayer.name, newNodeIndex, tensorIndex, kwargs]);
                            }
                            filteredInboundNodes.push(nodeData);
                        }
                    }
                }
                const dict = {};
                dict['name'] = layer.name;
                dict['className'] = layerClassName;
                dict['config'] = layerConfig;
                dict['inboundNodes'] = filteredInboundNodes;
                layerConfigs.push(dict);
            }
            config['layers'] = layerConfigs;
            // Gather info about inputs and outputs
            const modelInputs = [];
            for (let i = 0; i < this.inputLayers.length; i++) {
                const layer = this.inputLayers[i];
                const nodeIndex = this.inputLayersNodeIndices[i];
                const nodeKey = Container.nodeKey(layer, nodeIndex);
                if (!this.containerNodes.has(nodeKey)) {
                    continue;
                }
                let newNodeIndex = nodeConversionMap[nodeKey];
                if (newNodeIndex === null || newNodeIndex === undefined) {
                    newNodeIndex = 0;
                }
                const tensorIndex = this.inputLayersTensorIndices[i];
                modelInputs.push([layer.name, newNodeIndex, tensorIndex]);
            }
            config['inputLayers'] = modelInputs;
            const modelOutputs = [];
            for (let i = 0; i < this.outputLayers.length; i++) {
                const layer = this.outputLayers[i];
                const nodeIndex = this.outputLayersNodeIndices[i];
                const nodeKey = Container.nodeKey(layer, nodeIndex);
                if (!this.containerNodes.has(nodeKey)) {
                    continue;
                }
                let newNodeIndex = nodeConversionMap[nodeKey];
                if (newNodeIndex === null || newNodeIndex === undefined) {
                    newNodeIndex = 0;
                }
                const tensorIndex = this.outputLayersTensorIndices[i];
                modelOutputs.push([layer.name, newNodeIndex, tensorIndex]);
            }
            config['outputLayers'] = modelOutputs;
            return config;
        }
        /**
         * Instantiates a LayersModel from its config (output of `get_config()`).
         * @param cls the class to create
         * @param config LayersModel config dictionary.
         * @param customObjects An optional dictionary of custom objects.
         * @param fastWeightInit Optional flag to use fast weight initialization
         *   during deserialization. This is applicable to cases in which
         *   the initialization will be immediately overwritten by loaded weight
         *   values. Default: `false`.
         * @returns A LayersModel instance.
         * @throws ValueError: In case of improperly formatted config dict.
         */
        /** @nocollapse */
        static fromConfig(cls, config, customObjects = {}, fastWeightInit = false) {
            // Layer instances created during
            // the graph reconstruction process
            const createdLayers = {};
            // Dictionary mapping layer instances to
            // node data that specifies a layer call.
            // It acts as a queue that maintains any unprocessed
            // layer call until it becomes possible to process it
            // (i.e. until the input tensors to the call all exist).
            const unprocessedNodes = {};
            function addUnprocessedNode(layer, nodeData) {
                if (!(layer.name in unprocessedNodes)) {
                    unprocessedNodes[layer.name] = [nodeData];
                }
                else {
                    unprocessedNodes[layer.name].push(nodeData);
                }
            }
            function processNode(layer, nodeData) {
                const inputTensors = [];
                let kwargs;
                for (const inputData of nodeData) {
                    const inboundLayerName = inputData[0];
                    const inboundNodeIndex = inputData[1];
                    const inboundTensorIndex = inputData[2];
                    kwargs = inputData[3] == null ?
                        {} :
                        inputData[3];
                    if (!(inboundLayerName in createdLayers)) {
                        addUnprocessedNode(layer, nodeData);
                        return;
                    }
                    const inboundLayer = createdLayers[inboundLayerName];
                    if (inboundLayer.inboundNodes.length <= inboundNodeIndex) {
                        addUnprocessedNode(layer, nodeData);
                        return;
                    }
                    const inboundNode = inboundLayer.inboundNodes[inboundNodeIndex];
                    inputTensors.push(inboundNode.outputTensors[inboundTensorIndex]);
                }
                // Call layer on its inputs, thus creating the node
                // and building the layer if needed.
                // Note: This has Eager vs Graph Implications.
                if (inputTensors.length > 0) {
                    layer.apply(singletonOrArray(inputTensors), kwargs); // was ** kwargs
                }
            }
            /**
             * Deserialize a layer, then call it on appropriate inputs.
             * @param layerData: layer config dict.
             * @throws ValueError: In case of improperly formatted `layer_data`
             * dict.
             */
            function processLayer(layerData) {
                const layerName = layerData['name'];
                // Instantiate layer.
                const layer = deserialize(layerData, config['customObjects'] != null ?
                    config['customObjects'] :
                    {});
                layer.setFastWeightInitDuringBuild(fastWeightInit);
                createdLayers[layerName] = layer;
                // Gather layer inputs.
                const inboundNodesData = layerData['inboundNodes'];
                inboundNodesData.forEach(nodeData => {
                    if (!(nodeData instanceof Array)) {
                        throw new ValueError(`Corrupted configuration, expected array for nodeData: ${nodeData}`);
                    }
                    // We don't process nodes (i.e. make layer calls)
                    // on the fly because the inbound node may not yet exist,
                    // in case of layer shared at different topological depths
                    // (e.g.a model such as A(B(A(B(x)))))
                    addUnprocessedNode(layer, nodeData);
                });
            }
            // First, we create all layers and enqueue nodes to be processed.
            const name = config['name'];
            const layersFromConfig = config['layers'];
            for (const layerData of layersFromConfig) {
                processLayer(layerData);
            }
            // Then we process nodes in order of layer depth.
            // Nodes that cannot yet be processed(if the inbound node
            // does not yet exist) are re - enqueued, and the process
            // is repeated until all nodes are processed.
            while (!isObjectEmpty(unprocessedNodes)) {
                for (const layerData of layersFromConfig) {
                    const layer = createdLayers[layerData['name']];
                    if (layer.name in unprocessedNodes) {
                        const currentUnprocessedNodesForLayer = unprocessedNodes[layer.name];
                        delete unprocessedNodes[layer.name];
                        for (const nodeData of currentUnprocessedNodesForLayer) {
                            processNode(layer, nodeData);
                        }
                    }
                }
            }
            const inputTensors = [];
            const outputTensors = [];
            const inputLayersFromConfig = config['inputLayers'];
            for (const layerData of inputLayersFromConfig) {
                const layerName = layerData[0];
                const nodeIndex = layerData[1];
                const tensorIndex = layerData[2];
                assert$3(layerName in createdLayers);
                const layer = createdLayers[layerName];
                const layerOutputTensors = layer.inboundNodes[nodeIndex].outputTensors;
                inputTensors.push(layerOutputTensors[tensorIndex]);
            }
            const outputLayersFromConfig = config['outputLayers'];
            for (const layerData of outputLayersFromConfig) {
                const layerName = layerData[0];
                const nodeIndex = layerData[1];
                const tensorIndex = layerData[2];
                assert$3(layerName in createdLayers);
                const layer = createdLayers[layerName];
                const layerOutputTensors = layer.inboundNodes[nodeIndex].outputTensors;
                outputTensors.push(layerOutputTensors[tensorIndex]);
            }
            return new cls({ inputs: inputTensors, outputs: outputTensors, name });
        }
        /**
         * Determine whether the container is stateful.
         *
         * Porting Note: this is the equivalent of the stateful @property of
         *   the Container class in PyKeras.
         */
        get stateful() {
            // Porting Note: This check is to prevent inadvertent setting of the
            //   _stateful property of the Container instance.
            if (this._stateful) {
                throw new ValueError('Container instance unexpectedly has _stateful = true. The ' +
                    'statefulness of a Container is determined by the Layers it ' +
                    'contains. Its _stateful property must remain the default false.');
            }
            for (const layer of this.layers) {
                if (layer.stateful) {
                    return true;
                }
            }
            return false;
        }
        /**
         * Reset the state of all stateful constituent layers (if any).
         *
         * Examples of stateful layers include RNN layers whose `stateful` property
         * is set as `true`.
         */
        resetStates() {
            tidy(() => {
                this.layers.forEach(layer => {
                    // tslint:disable:no-any
                    if (layer.stateful) {
                        layer.resetStates();
                    }
                    // tslint:enable:no-any
                });
            });
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    function standardizeSampleOrClassWeights(xWeight, outputNames, weightType) {
        const numOutputs = outputNames.length;
        if (xWeight == null || (Array.isArray(xWeight) && xWeight.length === 0)) {
            return outputNames.map(name => null);
        }
        if (numOutputs === 1) {
            if (Array.isArray(xWeight) && xWeight.length === 1) {
                return xWeight;
            }
            else if (typeof xWeight === 'object' && outputNames[0] in xWeight) {
                return [xWeight[outputNames[0]]];
            }
            else {
                return [xWeight];
            }
        }
        if (Array.isArray(xWeight)) {
            if (xWeight.length !== numOutputs) {
                throw new Error(`Provided ${weightType} is an array of ${xWeight.length} ` +
                    `element(s), but the model has ${numOutputs} outputs. ` +
                    `Make sure a set of weights is provided for each model output.`);
            }
            return xWeight;
        }
        else if (typeof xWeight === 'object' && Object.keys(xWeight).length > 0 &&
            typeof xWeight[Object.keys(xWeight)[0]] ===
                'object') {
            const output = [];
            outputNames.forEach(outputName => {
                if (outputName in xWeight) {
                    output.push(xWeight[outputName]);
                }
                else {
                    output.push(null);
                }
            });
            return output;
        }
        else {
            throw new Error(`The model has multiple (${numOutputs}) outputs, ` +
                `so ${weightType} must be either an array with ` +
                `${numOutputs} elements or an object with ${outputNames} keys. ` +
                `Provided ${weightType} not understood: ${JSON.stringify(xWeight)}`);
        }
    }
    /**
     * Standardize class weighting objects.
     *
     * This function takes a single class-weighting object, an array of them,
     * or a map from output name to class-weighting object. It compares it to the
     * output name(s) of the model, base on which it outputs an array of
     * class-weighting objects of which the length matches the number of outputs.
     *
     * @param classWeight Input class-weighting object(s).
     * @param outputNames All output name(s) of the model.
     * @return An array of class-weighting objects. The length of the array matches
     *   the model's number of outputs.
     */
    function standardizeClassWeights(classWeight, outputNames) {
        return standardizeSampleOrClassWeights(classWeight, outputNames, 'classWeight');
    }
    /**
     * Standardize by-sample and/or by-class weights for training.
     *
     * Note that this function operates on one model output at a time. For a model
     * with multiple outputs, you must call this function multiple times.
     *
     * @param y The target tensor that the by-sample and/or by-class weight is for.
     *     The values of y are assumed to encode the classes, either directly
     *     as an integer index, or as one-hot encoding.
     * @param sampleWeight By-sample weights.
     * @param classWeight By-class weights: an object mapping class indices
     *     (integers) to a weight (float) to apply to the model's loss for the
     *     samples from this class during training. This can be useful to tell the
     *     model to "pay more attention" to samples from an under-represented class.
     * @param sampleWeightMode The mode for the sample weights.
     * @return A Promise of weight tensor, of which the size of the first dimension
     *     matches that of `y`.
     */
    async function standardizeWeights(y, sampleWeight, classWeight, sampleWeightMode) {
        if (sampleWeight != null || sampleWeightMode != null) {
            // TODO(cais): Once 'temporal' mode is implemented, document it in the doc
            // string.
            throw new Error('Support sampleWeight is not implemented yet');
        }
        if (classWeight != null) {
            // Apply class weights per sample.
            const yClasses = tidy(() => {
                if (y.shape.length === 1) {
                    // Assume class indices.
                    return clone(y);
                }
                else if (y.shape.length === 2) {
                    if (y.shape[1] > 1) {
                        // Assume one-hot encoding of classes.
                        const axis = 1;
                        return argMax$2(y, axis);
                    }
                    else if (y.shape[1] === 1) {
                        // Class index.
                        return reshape$3(y, [y.shape[0]]);
                    }
                    else {
                        throw new Error(`Encountered unexpected last-dimension size (${y.shape[1]}) ` +
                            `during handling of class weights. The size is expected to be ` +
                            `>= 1.`);
                    }
                }
                else {
                    throw new Error(`Unexpected rank of target (y) tensor (${y.rank}) during ` +
                        `handling of class weights. The rank is expected to be 1 or 2.`);
                }
            });
            const yClassIndices = Array.from(await yClasses.data());
            dispose(yClasses);
            const classSampleWeight = [];
            yClassIndices.forEach(classIndex => {
                if (classWeight[classIndex] == null) {
                    throw new Error(`classWeight must contain all classes in the training data. ` +
                        `The class ${classIndex} exists in the data but not in ` +
                        `classWeight`);
                }
                else {
                    classSampleWeight.push(classWeight[classIndex]);
                }
            });
            return tensor1d(classSampleWeight, 'float32');
        }
        else {
            return null;
        }
    }
    /**
     * Apply per-sample weights on the loss values from a number of samples.
     *
     * @param losses Loss tensor of shape `[batchSize]`.
     * @param sampleWeights Per-sample weight tensor of shape `[batchSize]`.
     * @returns Tensor of the same shape as`losses`.
     */
    function computeWeightedLoss(losses, sampleWeights) {
        return mul(losses, sampleWeights);
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    // Default batch size used during tensor-based validation.
    const DEFAULT_VALIDATION_BATCH_SIZE = 32;
    /**
     * Standardize the output of a dataset iterator for use by
     * LayersModel.fitDataset().
     *
     * @param model: A `tf.LayersModel` object.
     * @param iteratorOut The output of a dataset iterator. It is required to be
     *   an object of the form `{xs: TensorOrArrayOrMap, ys:
     * TensorOrArrayOrMap}`, where `TensorOrArrayOrMap` is a single `tf.Tensor`,
     * a `tf.Tensor[]`, or a flat map from string names to `tf.Tensor`s.
     * @returns A flat array of `tf.Tensor` objects: the input `tf.Tensor`s
     *   followed by the target `tf.Tensor`s.  When `tf.Tensor`s are provided
     *   as a map, the order in the resulting array is taken from the `inputNames`
     *   and `outputNames` of the model.
     */
    function standardizeDataIteratorOutput(
    // Type `model` as `any` here to avoid circular dependency w/
    // training.ts.
    // tslint:disable-next-line:no-any
    model, iteratorOut) {
        let xs;
        let ys;
        const iteratorOutObj = iteratorOut;
        xs = iteratorOutObj['xs'];
        ys = iteratorOutObj['ys'];
        assert$4(xs != null && ys != null, () => 'A Dataset iterator for fitDataset() is expected to generate ' +
            'objects of the form `{xs: xVal, ys: yVal}`, where the two ' +
            'values may be `tf.Tensor`, an array of Tensors, or a map of ' +
            'string to Tensor.  The provided Dataset instead generates ' +
            `${iteratorOut}`);
        const flattenedXs = flattenTensorOrArrayOrMap('input', model.inputNames, xs);
        const flattenedYs = flattenTensorOrArrayOrMap('output', model.outputNames, ys);
        const batchSize = flattenedXs[0].shape[0];
        assert$4(flattenedXs.length === model.inputs.length, () => `LayersModel has ${model.inputs.length} inputs, but the dataset ` +
            `provides ${flattenedXs.length} inputs.  (Expected input keys: ` +
            `${JSON.stringify(model.inputNames)})`);
        assert$4(flattenedYs.length === model.outputs.length, () => `LayersModel has ${model.outputs.length} outputs, but the dataset ` +
            `provides ${flattenedYs.length} outputs.  (Expected output keys: ` +
            `${JSON.stringify(model.outputNames)})`);
        for (let xIndex = 0; xIndex < flattenedXs.length; xIndex++) {
            assert$4(flattenedXs[xIndex].shape[0] === batchSize, () => `Batch size mismatch: input ` +
                `${model.inputNames[xIndex]} has ${flattenedXs[xIndex].shape[0]}; ` +
                `expected  ${batchSize} based on input ${model.inputNames[0]}.`);
        }
        for (let yIndex = 0; yIndex < flattenedYs.length; yIndex++) {
            assert$4(flattenedYs[yIndex].shape[0] === batchSize, () => `Batch size mismatch: output ` +
                `${model.outputNames[yIndex]} has ${flattenedYs[yIndex].shape[0]}; ` +
                `expected  ${batchSize} based on input ${model.inputNames[0]}.`);
        }
        return { xs: flattenedXs, ys: flattenedYs };
    }
    function flattenTensorOrArrayOrMap(inputOrOutput, names, values) {
        if (values instanceof Tensor) {
            return [values];
        }
        else if (Array.isArray(values)) {
            assert$4(values.length === names.length, () => `Received an array of ${values.length} Tensors, but expected ${names.length} to match the ${inputOrOutput} keys ${names}.`);
            return values;
        }
        else {
            const result = [];
            // Check that all the required keys are available.
            for (const name of names) {
                if (values[name] == null) {
                    throw new ValueError(`The feature data generated by the dataset lacks the required ` +
                        `${inputOrOutput} key '${name}'.`);
                }
                result.push(values[name]);
            }
            return result;
        }
    }
    function standardizeTensorValidationData(data) {
        if (data.length === 3) {
            throw new NotImplementedError('Validation with sample weights is not implemented yet.');
        }
        return { xs: data[0], ys: data[1] };
    }
    async function fitDataset(
    // Type `model` as `any` here to avoid circular dependency w/
    // training.ts.
    // tslint:disable-next-line:no-any
    model, dataset, args) {
        const hasBatchesPerEpoch = args.batchesPerEpoch != null;
        assert$4(model.optimizer != null, () => 'You must compile a model before training/testing. Use ' +
            'LayersModel.compile(modelCompileConfig).');
        assert$4(args != null, () => `For fitDataset(), the 2nd argument (config) is required, ` +
            `but it is not provided in this call.`);
        assert$4(args.epochs != null && args.epochs > 0 && Number.isInteger(args.epochs), () => `For fitDataset(), config.epochs is expected to be a positive ` +
            `integer, but got ${args.epochs}`);
        assert$4(!hasBatchesPerEpoch ||
            (args.batchesPerEpoch > 0 && Number.isInteger(args.batchesPerEpoch)), () => `For fitDataset(), config.batchesPerEpoch is expected to be a ` +
            `positive integer if specified, but got ${args.batchesPerEpoch}`);
        assert$4(
        // tslint:disable-next-line:no-any
        args['validationSplit'] == null, () => '`validationSplit` is not supported by `fitDataset()`. ' +
            'Use validationData instead.');
        if (model.isTraining) {
            throw new Error('Cannot start training because another fit() call is ongoing.');
        }
        model.isTraining = true;
        try {
            const doValidation = args.validationData != null;
            let valXs;
            let valYs;
            if (doValidation) {
                if (isDatasetObject(args.validationData)) {
                    assert$4(args.validationBatches == null ||
                        (args.validationBatches > 0 &&
                            Number.isInteger(args.validationBatches)), () => `For fitDataset() with dataset-based validation, ` +
                        `config.validationBatches is expected not to be provided, ` +
                        `or to be a positive integer, ` +
                        `but got ${args.validationBatches}`);
                }
                else {
                    const validationData = standardizeTensorValidationData(args.validationData);
                    valXs = validationData.xs;
                    valYs = validationData.ys;
                }
            }
            const trainFunction = model.makeTrainFunction();
            const outLabels = model.getDedupedMetricsNames();
            let callbackMetrics;
            if (doValidation) {
                callbackMetrics =
                    outLabels.slice().concat(outLabels.map(n => 'val_' + n));
            }
            else {
                callbackMetrics = outLabels.slice();
            }
            const callbacks = standardizeCallbacks(args.callbacks, args.yieldEvery);
            const verbose = args.verbose == null ? 1 : args.verbose;
            const { callbackList, history } = configureCallbacks(callbacks, verbose, args.epochs, null, null, getStepsPerEpoch(dataset, args), null, // Batch size determined by the dataset itself.
            doValidation, callbackMetrics);
            callbackList.setModel(model);
            model.history = history;
            await callbackList.onTrainBegin();
            model.stopTraining_ = false;
            let epoch = args.initialEpoch == null ? 0 : args.initialEpoch;
            let dataIterator = await dataset.iterator();
            while (epoch < args.epochs) {
                const epochLogs = {};
                await callbackList.onEpochBegin(epoch);
                let stepsDone = 0;
                let batchIndex = 0;
                if (!hasBatchesPerEpoch) {
                    dataIterator = await dataset.iterator();
                }
                while (hasBatchesPerEpoch ? stepsDone < args.batchesPerEpoch : true) {
                    const iteratorOut = await dataIterator.next();
                    // If `batchesPerEpoch` is specified, the dataset should not be
                    // exhausted until all epoches are done.
                    if (hasBatchesPerEpoch && iteratorOut.done) {
                        console.warn('You provided `batchesPerEpoch` as ' +
                            `${args.batchesPerEpoch}, ` +
                            'but your dataset iterator ran out of data after ' +
                            `${stepsDone} batches; ` +
                            'interrupting training. Make sure that your ' +
                            'dataset can generate at least `batchesPerEpoch * epochs` ' +
                            'batches (in this case, ' +
                            `${args.batchesPerEpoch * args.epochs} batches). ` +
                            'You may need to use the repeat() function when building ' +
                            'your dataset.');
                        break;
                    }
                    if (iteratorOut.value != null) {
                        const { xs, ys } = standardizeDataIteratorOutput(model, iteratorOut.value);
                        const batchLogs = {};
                        batchLogs['batch'] = batchIndex;
                        batchLogs['size'] = xs[0].shape[0];
                        await callbackList.onBatchBegin(batchIndex, batchLogs);
                        const sampleWeights = [];
                        if (args.classWeight != null) {
                            const standardClassWeights = standardizeClassWeights(args.classWeight, model.outputNames);
                            for (let i = 0; i < standardClassWeights.length; ++i) {
                                sampleWeights.push(await standardizeWeights(ys[i], null, standardClassWeights[i]));
                            }
                        }
                        // Train on batch.
                        const ins = xs.concat(ys).concat(sampleWeights);
                        const outs = trainFunction(ins);
                        dispose(ins);
                        for (let i = 0; i < outLabels.length; ++i) {
                            const label = outLabels[i];
                            const out = outs[i];
                            batchLogs[label] = out;
                            keep(out);
                        }
                        await callbackList.onBatchEnd(batchIndex, batchLogs);
                        disposeTensorsInLogs(batchLogs);
                        batchIndex++;
                        stepsDone++;
                    }
                    if (hasBatchesPerEpoch ? stepsDone >= args.batchesPerEpoch :
                        iteratorOut.done) {
                        // Epoch finished. Perform validation.
                        if (doValidation) {
                            let valOuts;
                            if (isDatasetObject(args.validationData)) {
                                valOuts = toList(await model.evaluateDataset(args.validationData, { batches: args.validationBatches }));
                            }
                            else {
                                valOuts = toList(model.evaluate(valXs, valYs, {
                                    batchSize: args.validationBatchSize == null ?
                                        DEFAULT_VALIDATION_BATCH_SIZE :
                                        args.validationBatchSize,
                                    verbose: 0
                                }));
                            }
                            for (let i = 0; i < model.metricsNames.length; ++i) {
                                epochLogs[`val_${model.metricsNames[i]}`] = valOuts[i];
                            }
                        }
                        // Call `break` to exit one epoch lopp after validation is done. If
                        // config.batchesPerEpoch is specified, an epoch while loop will
                        // stop when `stepsDone >= config.batchesPerEpoch`. When
                        // config.batchesPerEpoch is not provided, the following `break` is
                        // required to exit the while lopp after dataset is exhausted.
                        break;
                    }
                    if (model.stopTraining_) {
                        break;
                    }
                }
                await callbackList.onEpochEnd(epoch, epochLogs);
                epoch++;
                if (model.stopTraining_) {
                    break;
                }
            }
            await callbackList.onTrainEnd();
            await model.history.syncData();
            return model.history;
        }
        finally {
            model.isTraining = false;
        }
    }
    /** Helper function that determines number of steps (batches) per epoch. */
    function getStepsPerEpoch(dataset, args) {
        // Attempt to determine # of batches in an epoch.
        let stepsPerEpoch = null;
        if (args.batchesPerEpoch != null) {
            stepsPerEpoch = args.batchesPerEpoch;
        }
        else if (Number.isFinite(dataset.size)) {
            stepsPerEpoch = dataset.size;
        }
        return stepsPerEpoch;
    }
    // Check if provided object is a Dataset object by checking its .iterator
    // element.
    function isDatasetObject(dataset) {
        return (typeof dataset.iterator === 'function');
    }
    // Check if provided object is a LazyIterator object by checking it's .next
    // element.
    function isLazyIteratorObject(iterator) {
        return (typeof iterator.next === 'function');
    }
    async function evaluateDataset(
    // Type `model` as `any` here to avoid circular dependency w/
    // training.ts.
    // tslint:disable-next-line:no-any
    model, dataset, args) {
        args = args || {};
        const hasBatches = args.batches != null;
        const f = model.testFunction;
        let outs = [];
        if (args.verbose > 0) {
            throw new NotImplementedError('Verbose mode is not implemented yet.');
        }
        assert$4(!hasBatches || (args.batches > 0 && Number.isInteger(args.batches)), () => 'Test loop expects `batches` to be a positive integer, but ' +
            `received ${JSON.stringify(args.batches)}`);
        const dataIterator = isLazyIteratorObject(dataset) ?
            dataset :
            await dataset.iterator();
        // Keeps track of number of examples used in this evaluation.
        let numExamples = 0;
        let batch = 0;
        while (hasBatches ? batch < args.batches : true) {
            const iteratorOut = await dataIterator.next();
            outs = tidy(() => {
                if (iteratorOut.value) {
                    // TODO(cais): Once real dataset is available, use
                    //   `map(x => standardizeDataIteratorOutput(model, x).map(f)`.
                    const { xs, ys } = standardizeDataIteratorOutput(model, iteratorOut.value);
                    const xsAndYs = xs.concat(ys);
                    const batchOuts = tidy(() => f(xsAndYs));
                    dispose(xsAndYs);
                    if (batch === 0) {
                        for (let i = 0; i < batchOuts.length; ++i) {
                            outs.push(scalar(0));
                        }
                    }
                    const batchSize = xsAndYs[0].shape[0];
                    for (let i = 0; i < batchOuts.length; ++i) {
                        const batchOut = batchOuts[i];
                        const oldScalar = outs[i];
                        outs[i] =
                            tidy(() => add$2(outs[i], mul(batchSize, batchOut)));
                        if (batch > 0) {
                            dispose(oldScalar);
                        }
                    }
                    dispose(batchOuts);
                    numExamples += batchSize;
                    ++batch;
                }
                return outs;
            });
            if (iteratorOut.done) {
                if (hasBatches) {
                    console.warn('Your dataset iterator ran out of data during evaluateDataset(). ' +
                        'Interrupting evalution. Make sure that your ' +
                        'dataset can generate at least `batches` ' +
                        `batches (in this case, ${args.batches} batches). ` +
                        'You may need to use the repeat() function when building ' +
                        'your dataset.');
                }
                break;
            }
        }
        for (let i = 0; i < outs.length; ++i) {
            const oldScalar = outs[i];
            outs[i] = div$1(outs[i], numExamples);
            dispose(oldScalar);
        }
        return singletonOrArray(outs);
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    function checkBatchSize(batchSize) {
        assert$4(batchSize > 0 && Number.isInteger(batchSize), () => `batchSize is required to be a positive integer, but got ${batchSize}`);
    }
    /**
     * Slice a Tensor or an Array of Tensors, by start and stop indices.
     *
     * Porting Note: The `_slice_arrays` function in PyKeras is covered by this
     *   function and `sliceArraysByIndices()` together.
     *
     * @param arrays: the input.
     * @param start: the starting index (inclusive).
     * @param stop: the stopping index (exclusive).
     * @returns The result of the slicing. If `arrays` is an `Array` of
     *   `tf.Tensor`s, the slicing will be applied to all elements of the `Array`
     *   in the same way.
     */
    function sliceArrays(arrays, start, stop) {
        if (arrays == null) {
            return [null];
        }
        else if (Array.isArray(arrays)) {
            return arrays.map(array => sliceAlongFirstAxis(array, start, stop - start));
        }
        else { // Tensor.
            return sliceAlongFirstAxis(arrays, start, stop - start);
        }
    }
    /**
     * Slice a Tensor or an Array of Tensors, by random-order indices.
     *
     * Porting Note: The `_slice_arrays` function in PyKeras is covered by this
     *   function and `sliceArrays()` together.
     *
     * @param arrays The input `tf.Tensor` or `Array` of `tf.Tensor`s to slice.
     *   If an `Array` of `tf.Tensor`s, all `tf.Tensor`s will be sliced in the
     *   same fashion.
     * @param indices The indices to use for slicing along the first (batch)
     *   dimension.
     * @returns Result(s) of the slicing.
     */
    function sliceArraysByIndices(arrays, indices) {
        return tidy(() => {
            if (arrays == null) {
                return null;
            }
            else if (Array.isArray(arrays)) {
                return arrays.map(array => sliceArraysByIndices(array, indices));
            }
            else {
                // TODO(cais): indices should be a pre-constructed Tensor1D to avoid
                //   tensor1d() calls.
                return gather(arrays, indices.dtype === 'int32' ? indices : cast$3(indices, 'int32'));
            }
        });
    }
    /**
     * Returns a list of batch indices (tuples of indices).
     * @param size: Integer, total size of the data to slice into batches.
     * @param batchSize: Integer, batch size.
     * @returns An Array of [batchStart, batchEnd] tuples. batchStart is
     *   inclusive; batchEnd is exclusive. I.e., each batch consists of indices x
     *   that satisfy batchStart <= x < batchEnd.
     */
    function makeBatches(size, batchSize) {
        const output = [];
        let batchStart = 0;
        let batchEnd = null;
        while (batchStart < size) {
            batchEnd = batchStart + batchSize;
            if (batchEnd >= size) {
                batchEnd = size;
            }
            output.push([batchStart, batchEnd]);
            batchStart = batchEnd;
        }
        return output;
    }
    /**
     * Abstract fit function for `f(ins)`.
     * @param f A Function returning a list of tensors. For training, this
     *   function is expected to perform the updates to the variables.
     * @param ins List of tensors to be fed to `f`.
     * @param outLabels List of strings, display names of the outputs of `f`.
     * @param batchSize Integer batch size or `== null` if unknown. Default : 32.
     * @param epochs Number of times to iterate over the data. Default : 1.
     * @param verbose Verbosity mode: 0, 1, or 2. Default: 1.
     * @param callbacks List of callbacks to be called during training.
     * @param valF Function to call for validation.
     * @param valIns List of tensors to be fed to `valF`.
     * @param shuffle Whether to shuffle the data at the beginning of every
     * epoch. Default : true.
     * @param callbackMetrics List of strings, the display names of the metrics
     *   passed to the callbacks. They should be the concatenation of the
     *   display names of the outputs of `f` and the list of display names
     *   of the outputs of `valF`.
     * @param initialEpoch Epoch at which to start training (useful for
     *   resuming a previous training run). Default : 0.
     * @param stepsPerEpoch Total number of steps (batches on samples) before
     *   declaring one epoch finished and starting the next epoch. Ignored with
     *   the default value of `undefined` or `null`.
     * @param validationSteps Number of steps to run validation for (only if
     *   doing validation from data tensors). Not applicable for tfjs-layers.
     * @returns A `History` object.
     */
    async function fitLoop(
    // Type `model` as `any` here to avoid circular dependency w/ training.ts.
    // tslint:disable-next-line:no-any
    model, f, ins, outLabels, batchSize, epochs, verbose, callbacks, valF, valIns, shuffle$1, callbackMetrics, initialEpoch, stepsPerEpoch, validationSteps) {
        if (batchSize == null) {
            batchSize = 32;
        }
        if (epochs == null) {
            epochs = 1;
        }
        if (shuffle$1 == null) {
            shuffle$1 = true;
        }
        if (initialEpoch == null) {
            initialEpoch = 0;
        }
        // TODO(cais): Change const to let below when implementing validation.
        let doValidation = false;
        if (valF != null && valIns != null) {
            doValidation = true;
            // TODO(cais): verbose message.
        }
        if (validationSteps != null) {
            doValidation = true;
            if (stepsPerEpoch == null) {
                throw new ValueError('Can only use `validationSteps` when doing step-wise training, ' +
                    'i.e., `stepsPerEpoch` must be set.');
            }
        }
        const numTrainSamples = model.checkNumSamples(ins, batchSize, stepsPerEpoch, 'steps_per_epoch');
        let indexArray;
        if (numTrainSamples != null) {
            indexArray = range$3(0, numTrainSamples);
        }
        if (verbose == null) {
            verbose = 1;
        }
        const { callbackList, history } = configureCallbacks(callbacks, verbose, epochs, initialEpoch, numTrainSamples, stepsPerEpoch, batchSize, doValidation, callbackMetrics);
        callbackList.setModel(model);
        model.history = history;
        await callbackList.onTrainBegin();
        model.stopTraining_ = false;
        // TODO(cais): Take care of callbacks.validation_data as in PyKeras.
        // TODO(cais): Pre-convert feeds for performance as in PyKeras.
        for (let epoch = initialEpoch; epoch < epochs; ++epoch) {
            await callbackList.onEpochBegin(epoch);
            const epochLogs = {};
            if (stepsPerEpoch != null) {
                throw new NotImplementedError('stepsPerEpoch mode is not implemented yet.');
            }
            else {
                if (shuffle$1 === 'batch') {
                    throw new NotImplementedError('batch shuffling is not implemneted yet');
                }
                else if (shuffle$1) {
                    shuffle(indexArray);
                }
                // Convert the potentially shuffled indices to Tensor1D, to avoid the
                // cost of repeated creation of Array1Ds later on.
                const epochIndexArray1D = tensor1d(indexArray);
                const batches = makeBatches(numTrainSamples, batchSize);
                for (let batchIndex = 0; batchIndex < batches.length; ++batchIndex) {
                    const batchLogs = {};
                    await callbackList.onBatchBegin(batchIndex, batchLogs);
                    tidy(() => {
                        const batchStart = batches[batchIndex][0];
                        const batchEnd = batches[batchIndex][1];
                        const batchIds = sliceAlongFirstAxis(epochIndexArray1D, batchStart, batchEnd - batchStart);
                        batchLogs['batch'] = batchIndex;
                        batchLogs['size'] = batchEnd - batchStart;
                        // TODO(cais): In ins, train flag can be a number, instead of an
                        //   Tensor? Do we need to handle this in tfjs-layers?
                        const insBatch = sliceArraysByIndices(ins, batchIds);
                        const outs = f(insBatch);
                        for (let i = 0; i < outLabels.length; ++i) {
                            const label = outLabels[i];
                            const out = outs[i];
                            batchLogs[label] = out;
                            keep(out);
                            // TODO(cais): Use scope() to avoid ownership.
                        }
                        if (batchIndex === batches.length - 1) { // Last batch.
                            if (doValidation) {
                                const valOuts = model.testLoop(valF, valIns, batchSize);
                                // Porting Notes: In tfjs-layers, valOuts is always an Array.
                                for (let i = 0; i < outLabels.length; ++i) {
                                    const label = outLabels[i];
                                    const out = valOuts[i];
                                    keep(out);
                                    // TODO(cais): Use scope() to avoid ownership.
                                    epochLogs['val_' + label] = out;
                                }
                            }
                        }
                    });
                    await callbackList.onBatchEnd(batchIndex, batchLogs);
                    disposeTensorsInLogs(batchLogs);
                    if (model.stopTraining_) {
                        break;
                    }
                    // TODO(cais): return outs as list of Tensor.
                }
                epochIndexArray1D.dispose();
            }
            // TODO(cais): Run validation at the end of the epoch.
            await callbackList.onEpochEnd(epoch, epochLogs);
            if (model.stopTraining_) {
                break;
            }
        }
        await callbackList.onTrainEnd();
        await model.history.syncData();
        return model.history;
    }
    async function fitTensors(
    // Type `model` as `any` here to avoid circular dependency w/ training.ts.
    // tslint:disable-next-line:no-any
    model, x, y, args = {}) {
        if (model.isTraining) {
            throw new Error('Cannot start training because another fit() call is ongoing.');
        }
        model.isTraining = true;
        let inputs;
        let targets;
        let originalInputs;
        let originalTargets;
        let inputValX;
        let inputValY;
        let valX;
        let valY;
        let sampleWeights;
        try {
            const batchSize = args.batchSize == null ? 32 : args.batchSize;
            checkBatchSize(batchSize);
            // Validate user data.
            // TODO(cais): Support sampleWeight.
            const checkBatchAxis = false;
            const standardizedOuts = await model.standardizeUserData(x, y, args.sampleWeight, args.classWeight, checkBatchAxis, batchSize);
            inputs = standardizedOuts[0];
            targets = standardizedOuts[1];
            sampleWeights = standardizedOuts[2];
            // Prepare validation data.
            let doValidation = false;
            let valIns;
            if (args.validationData != null && args.validationData.length > 0) {
                doValidation = true;
                if (args.validationData.length === 2) {
                    // config.validationData consists of valX and valY.
                    inputValX = args.validationData[0];
                    inputValY = args.validationData[1];
                }
                else if (args.validationData.length === 3) {
                    throw new NotImplementedError('validationData including sample weights is not supported yet.');
                }
                else {
                    throw new ValueError(`When passing validation data, it must contain 2 (valX, valY) ` +
                        `or 3 (valX, valY, valSampleWeight) items; ` +
                        `${args.validationData} is invalid.`);
                }
                const checkBatchAxis = true;
                const valStandardized = await model.standardizeUserData(inputValX, inputValY, null, /** Unused sample weights. */ null, /** Unused class weights. */ checkBatchAxis, batchSize);
                valX = valStandardized[0];
                valY = valStandardized[1];
                valIns = valX.concat(valY);
                // TODO(cais): Add useLearningPhase data properly.
            }
            else if (args.validationSplit != null && args.validationSplit > 0 &&
                args.validationSplit < 1) {
                doValidation = true;
                // Porting Note: In tfjs-layers, inputs[0] is always a Tensor.
                const splitAt = Math.floor(inputs[0].shape[0] * (1 - args.validationSplit));
                const originalBatchSize = inputs[0].shape[0];
                valX = sliceArrays(inputs, splitAt, originalBatchSize);
                originalInputs = inputs;
                inputs = sliceArrays(inputs, 0, splitAt);
                valY = sliceArrays(targets, splitAt, originalBatchSize);
                originalTargets = targets;
                targets = sliceArrays(targets, 0, splitAt);
                // TODO(cais): Once sampleWeights becomes available, slice it to get
                //   valSampleWeights.
                valIns = valX.concat(valY);
                // TODO(cais): Add useLearningPhase data properly.
            }
            else if (args.validationSteps != null) {
                doValidation = true;
                // TODO(cais): Add useLearningPhase.
            }
            const ins = inputs.concat(targets).concat(sampleWeights);
            model.checkTrainableWeightsConsistency();
            // TODO(cais): Handle use_learning_phase and learning_phase?
            // Porting Note: Here we see a key deviation of tfjs-layers from
            // Keras.
            //  Due to the imperative nature of tfjs-layers' backend (tfjs-core),
            //  we do not construct symbolic computation graphs to embody the
            //  training process. Instead, we define a function that performs the
            //  training action. In PyKeras, the data (inputs and targets) are fed
            //  through graph placeholders. In tfjs-layers, the data are fed as
            //  function arguments. Since the function are defined below in the
            //  scope, we don't have equivalents of PyKeras's
            //  `_make_train_funciton`.
            const trainFunction = model.makeTrainFunction();
            const outLabels = model.getDedupedMetricsNames();
            let valFunction;
            let callbackMetrics;
            if (doValidation) {
                model.makeTestFunction();
                valFunction = model.testFunction;
                callbackMetrics =
                    outLabels.slice().concat(outLabels.map(n => 'val_' + n));
            }
            else {
                valFunction = null;
                valIns = [];
                callbackMetrics = outLabels.slice();
            }
            const callbacks = standardizeCallbacks(args.callbacks, args.yieldEvery);
            const out = await fitLoop(model, trainFunction, ins, outLabels, batchSize, args.epochs, args.verbose, callbacks, valFunction, valIns, args.shuffle, callbackMetrics, args.initialEpoch, null, null);
            return out;
        }
        finally {
            model.isTraining = false;
            // Memory clean up.
            disposeNewTensors(inputs, x);
            disposeNewTensors(targets, y);
            disposeNewTensors(originalInputs, x);
            disposeNewTensors(originalTargets, y);
            disposeNewTensors(valX, inputValX);
            disposeNewTensors(valY, inputValY);
            if (sampleWeights != null) {
                dispose(sampleWeights);
            }
        }
        // TODO(cais): Add value to outLabels.
    }
    /**
     * Ensure tensors all have a rank of at least 2.
     *
     * If a tensor has a rank of 1, it is dimension-expanded to rank 2.
     * If any tensor has a rank of 0 (i.e., is a scalar), an error will be thrown.
     */
    function ensureTensorsRank2OrHigher(tensors) {
        const outs = [];
        if (tensors instanceof Tensor) {
            tensors = [tensors];
        }
        // Make Tensors at least 2D.
        for (let i = 0; i < tensors.length; ++i) {
            const tensor = tensors[i];
            if (tensor.rank === 1) {
                outs.push(expandDims$2(tensor, 1));
            }
            else if (tensor.rank === 0) {
                throw new Error('Expected tensor to be at least 1D, but received a 0D tensor ' +
                    '(scalar).');
            }
            else {
                outs.push(tensor);
            }
        }
        return outs;
    }
    /**
     * Compare a set of tensors with a reference (old) set, discard the ones
     * in the new set that are not present in the reference set.
     *
     * This method is used for memory clenaup during calls such as
     * LayersModel.fit().
     *
     * @param tensors New set which may contain Tensors not present in
     *   `refTensors`.
     * @param refTensors Reference Tensor set.
     */
    // TODO(cais, kangyizhang): Deduplicate with tfjs-data.
    function disposeNewTensors(tensors, refTensors) {
        if (tensors == null) {
            return;
        }
        const oldTensorIds = [];
        if (refTensors instanceof Tensor) {
            oldTensorIds.push(refTensors.id);
        }
        else if (Array.isArray(refTensors)) {
            refTensors.forEach(t => oldTensorIds.push(t.id));
        }
        else if (refTensors != null) {
            // `oldTensors` is a map from string name to Tensor.
            for (const name in refTensors) {
                const oldTensor = refTensors[name];
                oldTensorIds.push(oldTensor.id);
            }
        }
        const tensorsToDispose = [];
        if (tensors instanceof Tensor) {
            if (oldTensorIds.indexOf(tensors.id) === -1) {
                tensorsToDispose.push(tensors);
            }
        }
        else if (Array.isArray(tensors)) {
            tensors.forEach(t => {
                if (oldTensorIds.indexOf(t.id) === -1) {
                    tensorsToDispose.push(t);
                }
            });
        }
        else if (tensors != null) {
            // `oldTensors` is a map from string name to Tensor.
            for (const name in tensors) {
                const tensor = tensors[name];
                if (oldTensorIds.indexOf(tensor.id) === -1) {
                    tensorsToDispose.push(tensor);
                }
            }
        }
        tensorsToDispose.forEach(t => {
            if (!t.isDisposed) {
                t.dispose();
            }
        });
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Helper function for polymorphic input data: 1. singleton Tensor.
     */
    function isDataTensor(x) {
        return x instanceof Tensor;
    }
    /**
     * Helper function for polymorphic input data: 2. Array of Tensor.
     */
    function isDataArray(x) {
        return Array.isArray(x);
    }
    /**
     * Helper function for polymorphic input data: 3. "dict" of Tensor.
     */
    function isDataDict(x) {
        return !isDataTensor(x) && !isDataArray(x);
    }
    /**
     * Normalizes inputs and targets provided by users.
     * @param data User-provided input data (polymorphic).
     * @param names An Array of expected Tensor names.
     * @param shapes Optional Array of expected Tensor shapes.
     * @param checkBatchAxis Whether to check that the batch axis of the arrays
     *   match  the expected value found in `shapes`.
     * @param exceptionPrefix String prefix used for exception formatting.
     * @returns List of standardized input Tensors (one Tensor per model input).
     * @throws ValueError: in case of improperly formatted user data.
     */
    function standardizeInputData(data, names, shapes, checkBatchAxis = true, exceptionPrefix = '') {
        if (names == null || names.length === 0) {
            // Check for the case where the model expected no data, but some data got
            // sent.
            if (data != null) {
                let gotUnexpectedData = false;
                if (isDataArray(data) && data.length > 0) {
                    gotUnexpectedData = true;
                }
                else if (isDataDict(data)) {
                    for (const key in data) {
                        if (data.hasOwnProperty(key)) {
                            gotUnexpectedData = true;
                            break;
                        }
                    }
                }
                else {
                    // `data` is a singleton Tensor in this case.
                    gotUnexpectedData = true;
                }
                if (gotUnexpectedData) {
                    throw new ValueError(`Error when checking model ${exceptionPrefix} expected no data, ` +
                        `but got ${data}`);
                }
            }
            return [];
        }
        if (data == null) {
            return names.map(name => null);
        }
        let arrays;
        if (isDataDict(data)) {
            data = data;
            arrays = [];
            for (const name of names) {
                if (data[name] == null) {
                    throw new ValueError(`No data provided for "${name}". Need data for each key in: ` +
                        `${names}`);
                }
                arrays.push(data[name]);
            }
        }
        else if (isDataArray(data)) {
            data = data;
            if (data.length !== names.length) {
                throw new ValueError(`Error when checking model ${exceptionPrefix}: the Array of ` +
                    `Tensors that you are passing to your model is not the size the ` +
                    `model expected. Expected to see ${names.length} Tensor(s), but ` +
                    `instead got the following list of Tensor(s): ${data}`);
            }
            arrays = data;
        }
        else {
            data = data;
            if (names.length > 1) {
                throw new ValueError(`The model ${exceptionPrefix} expects ${names.length} Tensor(s), ` +
                    `but only received one Tensor. Found: Tensor with shape ${data.shape}`);
            }
            arrays = [data];
        }
        arrays = ensureTensorsRank2OrHigher(arrays);
        // Check shape compatibility.
        if (shapes != null) {
            for (let i = 0; i < names.length; ++i) {
                if (shapes[i] == null) {
                    continue;
                }
                const array = arrays[i];
                if (array.shape.length !== shapes[i].length) {
                    throw new ValueError(`Error when checking ${exceptionPrefix}: expected ${names[i]} ` +
                        `to have ${shapes[i].length} dimension(s). but got array with ` +
                        `shape ${array.shape}`);
                }
                for (let j = 0; j < shapes[i].length; ++j) {
                    if (j === 0 && !checkBatchAxis) {
                        // Skip the first (batch) axis.
                        continue;
                    }
                    const dim = array.shape[j];
                    const refDim = shapes[i][j];
                    if (refDim != null && refDim >= 0 && dim !== refDim) {
                        throw new ValueError(`${exceptionPrefix} expected a batch of elements where each ` +
                            `example has shape [${shapes[i].slice(1, shapes[i].length)}] ` +
                            `(i.e.,tensor shape [*,${shapes[i].slice(1, shapes[i].length)}])` +
                            ` but the ${exceptionPrefix} received an input with ${array.shape[0]}` +
                            ` examples, each with shape [${array.shape.slice(1, array.shape.length)}]` +
                            ` (tensor shape [${array.shape}])`);
                    }
                }
            }
        }
        return arrays;
    }
    /**
     * User input validation for Tensors.
     * @param inputs `Array` of `tf.Tensor`s for inputs.
     * @param targets `Array` of `tf.Tensor`s for targets.
     * @param weights Optional `Array` of `tf.Tensor`s for sample weights.
     * @throws ValueError: in case of incorrectly formatted data.
     */
    function checkArrayLengths(inputs, targets, weights) {
        const setX = unique$2(inputs.map(input => input.shape[0]));
        setX.sort();
        const setY = unique$2(targets.map(target => target.shape[0]));
        setY.sort();
        // TODO(cais): Check `weights` as well.
        if (setX.length > 1) {
            throw new ValueError(`All input Tensors (x) should have the same number of samples. ` +
                `Got array shapes: ` +
                `${JSON.stringify(inputs.map(input => input.shape))}`);
        }
        if (setY.length > 1) {
            throw new ValueError(`All target Tensors (y) should have the same number of samples. ` +
                `Got array shapes: ` +
                `${JSON.stringify(targets.map(target => target.shape))}`);
        }
        if (setX.length > 0 && setY.length > 0 && !arraysEqual(setX, setY)) {
            throw new ValueError(`Input Tensors should have the same number of samples as target ` +
                `Tensors. Found ${setX[0]} input sample(s) and ${setY[0]} target ` +
                `sample(s).`);
        }
    }
    /**
     * Validation on the compatibility of targes and loss functions.
     *
     * This helps prevent users from using loss functions incorrectly.
     *
     * @param targets `Array` of `tf.Tensor`s of targets.
     * @param lossFns `Array` of loss functions.
     * @param outputShapes `Array` of shapes of model outputs.
     */
    function checkLossAndTargetCompatibility(targets, lossFns, outputShapes) {
        // TODO(cais): Dedicated test coverage?
        const keyLosses = [
            meanSquaredError$1, binaryCrossentropy$2,
            categoricalCrossentropy$2
        ];
        for (let i = 0; i < targets.length; ++i) {
            const y = targets[i];
            const loss = lossFns[i];
            const shape = outputShapes[i];
            if (loss == null) {
                continue;
            }
            if (loss === categoricalCrossentropy$2) {
                if (y.shape[y.shape.length - 1] === 1) {
                    throw new ValueError(`You are passing a target array of shape ${y.shape} while using ` +
                        `a loss 'categorical_crossentropy'. 'categorical_crossentropy'` +
                        `expects targets to be binary matrices (1s and 0s) of shape ` +
                        `[samples, classes].`);
                    // TODO(cais): Example code in error message.
                }
            }
            if (keyLosses.indexOf(loss) !== -1) {
                const slicedYShape = y.shape.slice(1);
                const slicedShape = shape.slice(1);
                for (let j = 0; j < slicedYShape.length; ++j) {
                    const targetDim = slicedYShape[j];
                    const outDim = slicedShape[j];
                    if (outDim != null && targetDim !== outDim) {
                        throw new ValueError(`A target Tensor with shape ${y.shape} was passed for an ` +
                            `output of shape ${shape}, while using a loss function that ` +
                            `expects targets to have the same shape as the output.`);
                    }
                }
            }
        }
    }
    /**
     * Check inputs provided by the user.
     *
     * Porting Note: This corresponds to _standardize_input_data() in Python
     *   Keras. Because of the strong typing in TF.js, we do not need to convert
     *   the data. Specifically:
     *   1) in PyKeras, `data` can be `DataFrame` instances from pandas, for
     *      example. We don't need to worry about that here because there is no
     *      widely popular javascript/typesdcript equivalent of pandas (so far).
     *      If one becomes available in the future, we can add support.
     *   2) in PyKeras, inputs can be Python dict. But here we are stipulating
     * that the data is either a single `tf.Tensor` or an Array of `tf.Tensor`s. We
     * may add support for `Object` data inputs in the future when the need
     * arises.
     *
     * Instead, we perform basic checks for number of parameters and shapes.
     *
     * @param data: The input data.
     * @param names: Name for the inputs, from the model.
     * @param shapes: Expected shapes for the input data, from the model.
     * @param checkBatchAxis: Whether the size along the batch axis (i.e., the
     *   first dimension) will be checked for matching.
     * @param exceptionPrefix: Execption prefix message, used in generating error
     *   messages.
     * @throws ValueError: on incorrect number of inputs or mismatches in shapes.
     */
    function checkInputData(data, names, shapes, checkBatchAxis = true, exceptionPrefix = '') {
        let arrays;
        if (Array.isArray(data)) {
            if (data.length !== names.length) {
                throw new ValueError(`Error when checking model ${exceptionPrefix}: the Array of ` +
                    `Tensors that you are passing to your model is not the size the ` +
                    `the model expected. Expected to see ${names.length} Tensor(s),` +
                    ` but instead got ${data.length} Tensors(s).`);
            }
            arrays = data;
        }
        else {
            if (names.length > 1) {
                throw new ValueError(`The model expects ${names.length} ${exceptionPrefix} Tensors, ` +
                    `but only received one Tensor. Found: array with shape ` +
                    `${JSON.stringify(data.shape)}.`);
            }
            arrays = [data];
        }
        if (shapes != null) {
            for (let i = 0; i < names.length; ++i) {
                if (shapes[i] == null) {
                    continue;
                }
                const array = arrays[i];
                if (array.shape.length !== shapes[i].length) {
                    throw new ValueError(`Error when checking ${exceptionPrefix}: expected ${names[i]} ` +
                        `to have ${shapes[i].length} dimension(s), but got array with ` +
                        `shape ${JSON.stringify(array.shape)}`);
                }
                for (let j = 0; j < shapes[i].length; ++j) {
                    if (j === 0 && !checkBatchAxis) {
                        continue;
                    }
                    const dim = array.shape[j];
                    const refDim = shapes[i][j];
                    if (refDim != null) {
                        if (refDim !== dim) {
                            throw new ValueError(`Error when checking ${exceptionPrefix}: expected ` +
                                `${names[i]} to have shape ${JSON.stringify(shapes[i])} but ` +
                                `got array with shape ${JSON.stringify(array.shape)}.`);
                        }
                    }
                }
            }
        }
    }
    /**
     * Maps metric functions to model outputs.
     * @param metrics An shortcut strings name, metric function, `Array` or dict
     *   (`Object`) of metric functions.
     * @param outputNames An `Array` of the names of model outputs.
     * @returns An `Array` (one entry per model output) of `Array` of metric
     *   functions. For instance, if the model has 2 outputs, and for the first
     *   output we want to compute `binaryAccuracy` and `binaryCrossentropy`,
     *   and just `binaryAccuracy` for the second output, the `Array` would look
     *   like:
     *     `[[binaryAccuracy, binaryCrossentropy],  [binaryAccuracy]]`
     * @throws TypeError: incompatible metrics format.
     */
    function collectMetrics(metrics, outputNames) {
        if (metrics == null || Array.isArray(metrics) && metrics.length === 0) {
            return outputNames.map(name => []);
        }
        let wrappedMetrics;
        if (typeof metrics === 'string' || typeof metrics === 'function') {
            wrappedMetrics = [metrics];
        }
        else if (Array.isArray(metrics) || typeof metrics === 'object') {
            wrappedMetrics = metrics;
        }
        else {
            throw new TypeError('Type of metrics argument not understood. Expected an string,' +
                `function, Array, or Object, found: ${metrics}`);
        }
        if (Array.isArray(wrappedMetrics)) {
            // We then apply all metrics to all outputs.
            return outputNames.map(name => wrappedMetrics);
        }
        else {
            // In this case, metrics is a dict.
            const nestedMetrics = [];
            for (const name of outputNames) {
                let outputMetrics = wrappedMetrics.hasOwnProperty(name) ? wrappedMetrics[name] : [];
                if (!Array.isArray(outputMetrics)) {
                    outputMetrics = [outputMetrics];
                }
                nestedMetrics.push(outputMetrics);
            }
            return nestedMetrics;
        }
    }
    const LAYERS_MODEL_FORMAT_NAME = 'layers-model';
    /**
     * A `tf.LayersModel` is a directed, acyclic graph of `tf.Layer`s plus methods
     * for training, evaluation, prediction and saving.
     *
     * `tf.LayersModel` is the basic unit of training, inference and evaluation in
     * TensorFlow.js. To create a `tf.LayersModel`, use `tf.LayersModel`.
     *
     * See also:
     *   `tf.Sequential`, `tf.loadLayersModel`.
     *
     * @doc {heading: 'Models', subheading: 'Classes'}
     */
    class LayersModel extends Container {
        constructor(args) {
            super(args);
            this.isTraining = false;
        }
        /**
         * Print a text summary of the model's layers.
         *
         * The summary includes
         * - Name and type of all layers that comprise the model.
         * - Output shape(s) of the layers
         * - Number of weight parameters of each layer
         * - If the model has non-sequential-like topology, the inputs each layer
         *   receives
         * - The total number of trainable and non-trainable parameters of the model.
         *
         * ```js
         * const input1 = tf.input({shape: [10]});
         * const input2 = tf.input({shape: [20]});
         * const dense1 = tf.layers.dense({units: 4}).apply(input1);
         * const dense2 = tf.layers.dense({units: 8}).apply(input2);
         * const concat = tf.layers.concatenate().apply([dense1, dense2]);
         * const output =
         *     tf.layers.dense({units: 3, activation: 'softmax'}).apply(concat);
         *
         * const model = tf.model({inputs: [input1, input2], outputs: output});
         * model.summary();
         * ```
         *
         * @param lineLength Custom line length, in number of characters.
         * @param positions Custom widths of each of the columns, as either
         *   fractions of `lineLength` (e.g., `[0.5, 0.75, 1]`) or absolute number
         *   of characters (e.g., `[30, 50, 65]`). Each number corresponds to
         *   right-most (i.e., ending) position of a column.
         * @param printFn Custom print function. Can be used to replace the default
         *   `console.log`. For example, you can use `x => {}` to mute the printed
         *   messages in the console.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        summary(lineLength, positions, printFn = console.log) {
            if (!this.built) {
                throw new ValueError(`This model has never been called, thus its weights have not been ` +
                    `created yet. So no summary can be displayed. Build the model ` +
                    `first (e.g., by calling it on some test data).`);
            }
            printSummary(this, lineLength, positions, printFn);
        }
        /**
         * Configures and prepares the model for training and evaluation.  Compiling
         * outfits the model with an optimizer, loss, and/or metrics.  Calling `fit`
         * or `evaluate` on an un-compiled model will throw an error.
         *
         * @param args a `ModelCompileArgs` specifying the loss, optimizer, and
         * metrics to be used for fitting and evaluating this model.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        compile(args) {
            if (args.loss == null) {
                args.loss = [];
            }
            this.loss = args.loss;
            if (typeof args.optimizer === 'string') {
                this.optimizer_ = getOptimizer(args.optimizer);
                this.isOptimizerOwned = true;
            }
            else {
                if (!(args.optimizer instanceof Optimizer)) {
                    throw new ValueError(`User-defined optimizer must be an instance of tf.Optimizer.`);
                }
                this.optimizer_ = args.optimizer;
                this.isOptimizerOwned = false;
            }
            // TODO(cais): Add lossWeights.
            // TODO(cais): Add sampleWeightMode.
            // Prepare loss functions.
            let lossFunctions = [];
            if (!Array.isArray(args.loss) && typeof args.loss !== 'string' &&
                typeof args.loss !== 'function') {
                args.loss = args.loss;
                for (const name in args.loss) {
                    if (this.outputNames.indexOf(name) === -1) {
                        throw new ValueError(`Unknown entry in loss dictionary: "${name}". ` +
                            `Only expected the following keys: ${this.outputNames}`);
                    }
                }
                for (const name of this.outputNames) {
                    if (args.loss[name] == null) {
                        console.warn(`Output "${name}" is missing from loss dictionary. We assume ` +
                            `this was done on purpose, and we will not be expecting data ` +
                            `to be passed to ${name} during training`);
                    }
                    lossFunctions.push(get$2(args.loss[name]));
                }
            }
            else if (Array.isArray(args.loss)) {
                if (args.loss.length !== this.outputs.length) {
                    throw new ValueError(`When passing an Array as loss, it should have one entry per ` +
                        `model output. The model has ${this.outputs.length} output(s), ` +
                        `but you passed loss=${args.loss}.`);
                }
                const theLosses = args.loss;
                lossFunctions = theLosses.map(l => get$2(l));
            }
            else {
                const lossFunction = get$2(args.loss);
                this.outputs.forEach(_ => {
                    lossFunctions.push(lossFunction);
                });
            }
            this.lossFunctions = lossFunctions;
            this.feedOutputNames = [];
            this.feedOutputShapes = [];
            this.feedLossFns = [];
            for (let i = 0; i < this.outputs.length; ++i) {
                // TODO(cais): Logic for skipping target(s).
                const shape = this.internalOutputShapes[i];
                const name = this.outputNames[i];
                this.feedOutputNames.push(name);
                this.feedOutputShapes.push(shape);
                this.feedLossFns.push(this.lossFunctions[i]);
            }
            // TODO(cais): Add logic for output masks.
            // TODO(cais): Add logic for sample weights.
            const skipTargetIndices = [];
            // Prepare metrics.
            this.metrics = args.metrics;
            // TODO(cais): Add weightedMetrics.
            this.metricsNames = ['loss'];
            this.metricsTensors = [];
            // Compute total loss.
            // Porting Note: In PyKeras, metrics_tensors are symbolic tensor objects.
            //   Here, metricsTensors are TypeScript functions. This difference is due
            //   to the difference in symbolic/imperative property of the backends.
            nameScope('loss', () => {
                for (let i = 0; i < this.outputs.length; ++i) {
                    if (skipTargetIndices.indexOf(i) !== -1) {
                        continue;
                    }
                    // TODO(cais): Add weightedLoss, sampleWeight and mask.
                    //   The following line should be weightedLoss
                    const weightedLoss = this.lossFunctions[i];
                    if (this.outputs.length > 1) {
                        this.metricsTensors.push([weightedLoss, i]);
                        this.metricsNames.push(this.outputNames[i] + '_loss');
                    }
                }
                // Porting Note: Due to the imperative nature of the backend, we calculate
                //   the regularizer penalties in the totalLossFunction, instead of here.
            });
            const nestedMetrics = collectMetrics(args.metrics, this.outputNames);
            // TODO(cais): Add nestedWeightedMetrics.
            /**
             * Helper function used in loop below.
             */
            const appendMetric = (outputIndex, metricName, metricTensor) => {
                if (this.outputNames.length > 1) {
                    metricName = this.outputNames[outputIndex] + '_' + metricName;
                }
                this.metricsNames.push(metricName);
                this.metricsTensors.push([metricTensor, outputIndex]);
            };
            nameScope('metric', () => {
                for (let i = 0; i < this.outputs.length; ++i) {
                    if (skipTargetIndices.indexOf(i) !== -1) {
                        continue;
                    }
                    const outputMetrics = nestedMetrics[i];
                    // TODO(cais): Add weights and outputWeightedMetrics.
                    // TODO(cais): Add optional arg `weights` to the following function.
                    const handleMetrics = (metrics) => {
                        const metricNamePrefix = '';
                        let metricName;
                        let accFn;
                        let weightedMetricFn;
                        //  TODO(cais): Use 'weights_' for weighted metrics.
                        for (const metric of metrics) {
                            if (typeof metric === 'string' &&
                                ['accuracy', 'acc', 'crossentropy', 'ce'].indexOf(metric) !==
                                    -1) {
                                const outputShape = this.internalOutputShapes[i];
                                if (outputShape[outputShape.length - 1] === 1 ||
                                    this.lossFunctions[i] === binaryCrossentropy$2) {
                                    // case: binary accuracy/crossentropy.
                                    if (['accuracy', 'acc'].indexOf(metric) !== -1) {
                                        accFn = binaryAccuracy$1;
                                    }
                                    else if (['crossentropy', 'ce'].indexOf(metric) !== -1) {
                                        accFn = binaryCrossentropy$1;
                                    }
                                }
                                else if (this.lossFunctions[i] ===
                                    sparseCategoricalCrossentropy$1) {
                                    // case: categorical accuracy / crossentropy with sparse
                                    // targets.
                                    if (['accuracy', 'acc'].indexOf(metric) !== -1) {
                                        accFn = sparseCategoricalAccuracy$1;
                                    }
                                    else if (['crossentropy', 'ce'].indexOf(metric) !== -1) {
                                        accFn = sparseCategoricalCrossentropy;
                                    }
                                }
                                else {
                                    // case: categorical accuracy / crossentropy.
                                    if (['accuracy', 'acc'].indexOf(metric) !== -1) {
                                        accFn = categoricalAccuracy$1;
                                    }
                                    else if (['crossentropy', 'ce'].indexOf(metric) !== -1) {
                                        accFn = categoricalCrossentropy$1;
                                    }
                                }
                                let suffix;
                                if (['accuracy', 'acc'].indexOf(metric) !== -1) {
                                    suffix = 'acc';
                                }
                                else if (['crossentropy', 'ce'].indexOf(metric) !== -1) {
                                    suffix = 'ce';
                                }
                                // TODO(cais): Add weighting actually.
                                weightedMetricFn = accFn;
                                metricName = metricNamePrefix + suffix;
                            }
                            else {
                                const metricFn = get$1(metric);
                                // TODO(cais): Add weighting actually.
                                weightedMetricFn = metricFn;
                                metricName =
                                    metricNamePrefix + getLossOrMetricName(metric);
                            }
                            // TODO(cais): Add weighting and masking to metricResult.
                            let metricResult;
                            nameScope(metricName, () => {
                                metricResult = weightedMetricFn;
                            });
                            appendMetric(i, metricName, metricResult);
                        }
                    };
                    handleMetrics(outputMetrics);
                    // TODO(cais): Call handleMetrics with weights.
                }
            });
            // Porting Notes: Given the imperative backend of tfjs-core,
            //   there is no need for constructing the symbolic graph and placeholders.
            this.collectedTrainableWeights = this.trainableWeights;
        }
        /**
         * Check trainable weights count consistency.
         *
         * This will raise a warning if `this.trainableWeights` and
         * `this.collectedTrainableWeights` are inconsistent (i.e., have different
         * numbers of parameters).
         * Inconsistency will typically arise when one modifies `model.trainable`
         * without calling `model.compile()` again.
         */
        checkTrainableWeightsConsistency() {
            if (this.collectedTrainableWeights == null) {
                return;
            }
            if (this.trainableWeights.length !==
                this.collectedTrainableWeights.length) {
                console.warn('Discrepancy between trainableweights and collected trainable ' +
                    'weights. Did you set `model.trainable` without calling ' +
                    '`model.compile()` afterwards?');
            }
        }
        /**
         * Returns the loss value & metrics values for the model in test mode.
         *
         * Loss and metrics are specified during `compile()`, which needs to happen
         * before calls to `evaluate()`.
         *
         * Computation is done in batches.
         *
         * ```js
         * const model = tf.sequential({
         *   layers: [tf.layers.dense({units: 1, inputShape: [10]})]
         * });
         * model.compile({optimizer: 'sgd', loss: 'meanSquaredError'});
         * const result = model.evaluate(
         *     tf.ones([8, 10]), tf.ones([8, 1]), {batchSize: 4});
         * result.print();
         * ```
         *
         * @param x `tf.Tensor` of test data, or an `Array` of `tf.Tensor`s if the
         * model has multiple inputs.
         * @param y `tf.Tensor` of target data, or an `Array` of `tf.Tensor`s if the
         * model has multiple outputs.
         * @param args A `ModelEvaluateArgs`, containing optional fields.
         *
         * @return `Scalar` test loss (if the model has a single output and no
         *   metrics) or `Array` of `Scalar`s (if the model has multiple outputs
         *   and/or metrics). The attribute `model.metricsNames`
         *   will give you the display labels for the scalar outputs.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        evaluate(x, y, args = {}) {
            const batchSize = args.batchSize == null ? 32 : args.batchSize;
            checkBatchSize(batchSize);
            // TODO(cais): Standardize `config.sampleWeights` as well.
            // Validate user data.
            const checkBatchAxis = true;
            const standardizedOuts = this.standardizeUserDataXY(x, y, checkBatchAxis, batchSize);
            try {
                // TODO(cais): If uses `useLearningPhase`, set the corresponding element
                // of the input to 0.
                const ins = standardizedOuts[0].concat(standardizedOuts[1]);
                this.makeTestFunction();
                const f = this.testFunction;
                const testOuts = this.testLoop(f, ins, batchSize, args.verbose, args.steps);
                return singletonOrArray(testOuts);
            }
            finally {
                disposeNewTensors(standardizedOuts[0], x);
                disposeNewTensors(standardizedOuts[1], y);
            }
        }
        // TODO(cais): Add code snippet below once real dataset objects are
        //   available.
        /**
         * Evaluate model using a dataset object.
         *
         * Note: Unlike `evaluate()`, this method is asynchronous (`async`);
         *
         * @param dataset A dataset object. Its `iterator()` method is expected
         *   to generate a dataset iterator object, the `next()` method of which
         *   is expected to produce data batches for evaluation. The return value
         *   of the `next()` call ought to contain a boolean `done` field and a
         *   `value` field. The `value` field is expected to be an array of two
         *   `tf.Tensor`s or an array of two nested `tf.Tensor` structures. The former
         *   case is for models with exactly one input and one output (e.g..
         *   a sequential model). The latter case is for models with multiple
         *   inputs and/or multiple outputs. Of the two items in the array, the
         *   first is the input feature(s) and the second is the output target(s).
         * @param args A configuration object for the dataset-based evaluation.
         * @returns Loss and metric values as an Array of `Scalar` objects.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        async evaluateDataset(dataset, args) {
            this.makeTestFunction();
            return evaluateDataset(this, dataset, args);
        }
        /**
         * Get number of samples provided for training, evaluation or prediction.
         *
         * @param ins Input `tf.Tensor`.
         * @param batchSize Integer batch size, optional.
         * @param steps Total number of steps (batches of samples) before
         * declaring loop finished. Optional.
         * @param stepsName The public API's parameter name for `steps`.
         * @returns Number of samples provided.
         */
        checkNumSamples(ins, batchSize, steps, stepsName = 'steps') {
            let numSamples;
            if (steps != null) {
                numSamples = null;
                if (batchSize != null) {
                    throw new ValueError(`If ${stepsName} is set, batchSize must be null or undefined.` +
                        `Got batchSize = ${batchSize}`);
                }
            }
            else if (ins != null) {
                if (Array.isArray(ins)) {
                    numSamples = ins[0].shape[0];
                }
                else {
                    numSamples = ins.shape[0];
                }
            }
            else {
                throw new ValueError(`Either the input data should have a defined shape, or ` +
                    `${stepsName} shoud be specified.`);
            }
            return numSamples;
        }
        /**
         * Execute internal tensors of the model with input data feed.
         * @param inputs Input data feed. Must match the inputs of the model.
         * @param outputs Names of the output tensors to be fetched. Must match
         *   names of the SymbolicTensors that belong to the graph.
         * @returns Fetched values for `outputs`.
         */
        execute(inputs, outputs) {
            if (Array.isArray(outputs) && outputs.length === 0) {
                throw new ValueError('`outputs` is an empty Array, which is not allowed.');
            }
            const outputsIsArray = Array.isArray(outputs);
            const outputNames = (outputsIsArray ? outputs : [outputs]);
            const outputSymbolicTensors = this.retrieveSymbolicTensors(outputNames);
            // Format the input into a FeedDict.
            const feedDict = new FeedDict();
            if (inputs instanceof Tensor) {
                inputs = [inputs];
            }
            if (Array.isArray(inputs)) {
                if (inputs.length !== this.inputs.length) {
                    throw new ValueError(`The number of inputs provided (${inputs.length}) ` +
                        `does not match the number of inputs of this model ` +
                        `(${this.inputs.length}).`);
                }
                for (let i = 0; i < this.inputs.length; ++i) {
                    feedDict.add(this.inputs[i], inputs[i]);
                }
            }
            else {
                for (const input of this.inputs) {
                    const tensorValue = inputs[input.name];
                    if (tensorValue == null) {
                        throw new ValueError(`No value is provided for the model's input ${input.name}`);
                    }
                    feedDict.add(input, tensorValue);
                }
            }
            // Run execution.
            const executeOutputs = execute(outputSymbolicTensors, feedDict);
            return outputsIsArray ? executeOutputs : executeOutputs[0];
        }
        /**
         * Retrieve the model's internal symbolic tensors from symbolic-tensor names.
         */
        retrieveSymbolicTensors(symbolicTensorNames) {
            const outputSymbolicTensors = pyListRepeat(null, symbolicTensorNames.length);
            let outputsRemaining = symbolicTensorNames.length;
            for (const layer of this.layers) {
                const layerOutputs = Array.isArray(layer.output) ? layer.output : [layer.output];
                const layerOutputNames = layerOutputs.map(output => output.name);
                for (let i = 0; i < symbolicTensorNames.length; ++i) {
                    const index = layerOutputNames.indexOf(symbolicTensorNames[i]);
                    if (index !== -1) {
                        outputSymbolicTensors[i] = layerOutputs[index];
                        outputsRemaining--;
                    }
                    if (outputsRemaining === 0) {
                        break;
                    }
                }
                if (outputsRemaining === 0) {
                    break;
                }
            }
            if (outputsRemaining > 0) {
                const remainingNames = [];
                outputSymbolicTensors.forEach((tensor, i) => {
                    if (tensor == null) {
                        remainingNames.push(symbolicTensorNames[i]);
                    }
                });
                throw new ValueError(`Cannot find SymbolicTensors for output name(s): ` +
                    `${JSON.stringify(remainingNames)}`);
            }
            return outputSymbolicTensors;
        }
        /**
         * Helper method to loop over some data in batches.
         *
         * Porting Note: Not using the functional approach in the Python equivalent
         *   due to the imperative backend.
         * Porting Note: Does not support step mode currently.
         *
         * @param ins: input data
         * @param batchSize: integer batch size.
         * @param verbose: verbosity model
         * @returns: Predictions as `tf.Tensor` (if a single output) or an `Array` of
         *   `tf.Tensor` (if multipe outputs).
         */
        predictLoop(ins, batchSize = 32, verbose = false) {
            return tidy(() => {
                const numSamples = this.checkNumSamples(ins);
                if (verbose) {
                    throw new NotImplementedError('Verbose predictLoop() is not implemented yet.');
                }
                // Sample-based predictions.
                // Porting Note: Tensor currently does not support sliced assignments as
                //   in numpy, e.g., x[1:3] = y. Therefore we use concatenation while
                //   iterating over the batches.
                const batches = makeBatches(numSamples, batchSize);
                const outsBatches = this.outputs.map(output => []);
                // TODO(cais): Can the scope() be pushed down inside the for loop?
                for (let batchIndex = 0; batchIndex < batches.length; ++batchIndex) {
                    const batchOuts = tidy(() => {
                        const batchStart = batches[batchIndex][0];
                        const batchEnd = batches[batchIndex][1];
                        // TODO(cais): Take care of the case of the last element is a flag for
                        //   training/test.
                        const insBatch = sliceArrays(ins, batchStart, batchEnd);
                        // Construct the feeds for execute();
                        const feeds = [];
                        if (Array.isArray(insBatch)) {
                            for (let i = 0; i < insBatch.length; ++i) {
                                feeds.push({ key: this.inputs[i], value: insBatch[i] });
                            }
                        }
                        else {
                            feeds.push({ key: this.inputs[0], value: insBatch });
                        }
                        const feedDict = new FeedDict(feeds);
                        return execute(this.outputs, feedDict);
                    });
                    batchOuts.forEach((batchOut, i) => outsBatches[i].push(batchOut));
                }
                return singletonOrArray(outsBatches.map(batches => concat$2(batches, 0)));
            });
        }
        /**
         * Generates output predictions for the input samples.
         *
         * Computation is done in batches.
         *
         * Note: the "step" mode of predict() is currently not supported.
         *   This is because the TensorFlow.js core backend is imperative only.
         *
         * ```js
         * const model = tf.sequential({
         *   layers: [tf.layers.dense({units: 1, inputShape: [10]})]
         * });
         * model.predict(tf.ones([8, 10]), {batchSize: 4}).print();
         * ```
         *
         * @param x The input data, as a Tensor, or an `Array` of `tf.Tensor`s if
         *   the model has multiple inputs.
         * @param args A `ModelPredictArgs` object containing optional fields.
         *
         * @return Prediction results as a `tf.Tensor`(s).
         *
         * @exception ValueError In case of mismatch between the provided input data
         *   and the model's expectations, or in case a stateful model receives a
         *   number of samples that is not a multiple of the batch size.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        predict(x, args = {}) {
            const xsRank2OrHigher = ensureTensorsRank2OrHigher(x);
            checkInputData(xsRank2OrHigher, this.inputNames, this.feedInputShapes, false);
            try {
                // TODO(cais): Take care of stateful models.
                //   if (this.stateful) ...
                // TODO(cais): Take care of the learning_phase boolean flag.
                //   if (this.useLearningPhase) ...
                const batchSize = args.batchSize == null ? 32 : args.batchSize;
                checkBatchSize(batchSize);
                return this.predictLoop(xsRank2OrHigher, batchSize);
            }
            finally {
                disposeNewTensors(xsRank2OrHigher, x);
            }
        }
        /**
         * Returns predictions for a single batch of samples.
         *
         * ```js
         * const model = tf.sequential({
         *   layers: [tf.layers.dense({units: 1, inputShape: [10]})]
         * });
         * model.predictOnBatch(tf.ones([8, 10])).print();
         * ```
         * @param x: Input samples, as a Tensor (for models with exactly one
         *   input) or an array of Tensors (for models with more than one input).
         * @return Tensor(s) of predictions
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        predictOnBatch(x) {
            checkInputData(x, this.inputNames, this.feedInputShapes, true);
            // TODO(cais): Take care of the learning_phase boolean flag.
            //   if (this.useLearningPhase) ...
            const batchSize = (Array.isArray(x) ? x[0] : x).shape[0];
            return this.predictLoop(x, batchSize);
        }
        standardizeUserDataXY(x, y, checkBatchAxis = true, batchSize) {
            // TODO(cais): Add sampleWeight, classWeight
            if (this.optimizer_ == null) {
                throw new RuntimeError('You must compile a model before training/testing. Use ' +
                    'LayersModel.compile(modelCompileArgs).');
            }
            const outputShapes = [];
            for (let i = 0; i < this.feedOutputShapes.length; ++i) {
                const outputShape = this.feedOutputShapes[i];
                const lossFn = this.feedLossFns[i];
                if (lossFn === sparseCategoricalCrossentropy$1) {
                    outputShapes.push(outputShape.slice(0, outputShape.length - 1).concat([1]));
                }
                else {
                    // Porting Note: Because of strong typing `lossFn` must be a function.
                    outputShapes.push(outputShape);
                }
            }
            x = standardizeInputData(x, this.feedInputNames, this.feedInputShapes, false, 'input');
            y = standardizeInputData(y, this.feedOutputNames, outputShapes, false, 'target');
            // TODO(cais): Standardize sampleWeights & classWeights.
            checkArrayLengths(x, y);
            // TODO(cais): Check sampleWeights as well.
            checkLossAndTargetCompatibility(y, this.feedLossFns, this.feedOutputShapes);
            if (this.stateful && batchSize != null && batchSize > 0) {
                if (x[0].shape[0] % batchSize !== 0) {
                    throw new ValueError(`In a stateful network, you should only pass inputs with a ` +
                        `number of samples that is divisible by the batch size ` +
                        `${batchSize}. Found: ${x[0].shape[0]} sample(s).`);
                }
            }
            return [x, y];
        }
        async standardizeUserData(x, y, sampleWeight, classWeight, checkBatchAxis = true, batchSize) {
            const [standardXs, standardYs] = this.standardizeUserDataXY(x, y, checkBatchAxis, batchSize);
            // TODO(cais): Handle sampleWeights.
            if (sampleWeight != null) {
                throw new Error('sample weight is not supported yet.');
            }
            let standardSampleWeights = null;
            if (classWeight != null) {
                const classWeights = standardizeClassWeights(classWeight, this.outputNames);
                standardSampleWeights = [];
                for (let i = 0; i < classWeights.length; ++i) {
                    standardSampleWeights.push(await standardizeWeights(standardYs[i], null, classWeights[i]));
                }
            }
            // TODO(cais): Deal with the case of model.stateful == true.
            return [standardXs, standardYs, standardSampleWeights];
        }
        /**
         * Loop over some test data in batches.
         * @param f A Function returning a list of tensors.
         * @param ins Array of tensors to be fed to `f`.
         * @param batchSize Integer batch size or `null` / `undefined`.
         * @param verbose verbosity mode.
         * @param steps Total number of steps (batches of samples) before
         * declaring test finished. Ignored with the default value of `null` /
         * `undefined`.
         * @returns Array of Scalars.
         */
        testLoop(f, ins, batchSize, verbose = 0, steps) {
            return tidy(() => {
                const numSamples = this.checkNumSamples(ins, batchSize, steps, 'steps');
                const outs = [];
                if (verbose > 0) {
                    throw new NotImplementedError('Verbose mode is not implemented yet.');
                }
                // TODO(cais): Use `indicesForConversionToDense' to prevent slow down.
                if (steps != null) {
                    throw new NotImplementedError('steps mode in testLoop() is not implemented yet');
                }
                else {
                    const batches = makeBatches(numSamples, batchSize);
                    const indexArray = tensor1d(range$3(0, numSamples));
                    for (let batchIndex = 0; batchIndex < batches.length; ++batchIndex) {
                        const batchStart = batches[batchIndex][0];
                        const batchEnd = batches[batchIndex][1];
                        const batchIds = sliceAlongFirstAxis(indexArray, batchStart, batchEnd - batchStart);
                        // TODO(cais): In ins, train flag can be a number, instead of an
                        //   Tensor? Do we need to handle this in tfjs-layers?
                        const insBatch = sliceArraysByIndices(ins, batchIds);
                        const batchOuts = f(insBatch);
                        if (batchIndex === 0) {
                            for (let i = 0; i < batchOuts.length; ++i) {
                                outs.push(scalar(0));
                            }
                        }
                        for (let i = 0; i < batchOuts.length; ++i) {
                            const batchOut = batchOuts[i];
                            outs[i] =
                                add$2(outs[i], mul(batchEnd - batchStart, batchOut));
                        }
                    }
                    for (let i = 0; i < outs.length; ++i) {
                        outs[i] = div$1(outs[i], numSamples);
                    }
                }
                return outs;
            });
        }
        getDedupedMetricsNames() {
            const outLabels = this.metricsNames;
            // Rename duplicated metrics names (can happen with an output layer
            // shared among multiple dataflows).
            const dedupedOutLabels = [];
            for (let i = 0; i < outLabels.length; ++i) {
                const label = outLabels[i];
                let newLabel = label;
                if (count(outLabels, label) > 1) {
                    const dupIndex = count(outLabels.slice(0, i), label);
                    newLabel += `_${dupIndex}`;
                }
                dedupedOutLabels.push(newLabel);
            }
            return dedupedOutLabels;
        }
        /**
         * Creates a function that performs the following actions:
         *
         * 1. computes the losses
         * 2. sums them to get the total loss
         * 3. call the optimizer computes the gradients of the LayersModel's
         *    trainable weights w.r.t. the total loss and update the variables
         * 4. calculates the metrics
         * 5. returns the values of the losses and metrics.
         */
        makeTrainFunction() {
            return (data) => {
                const lossValues = [];
                const inputs = data.slice(0, this.inputs.length);
                const targets = data.slice(this.inputs.length, this.inputs.length + this.outputs.length);
                const sampleWeights = data.slice(this.inputs.length + this.outputs.length, this.inputs.length + this.outputs.length * 2);
                const metricsValues = [];
                // Create a function that computes the total loss based on the
                // inputs. This function is used for obtaining gradients through
                // backprop.
                const totalLossFunction = () => {
                    const feeds = [];
                    for (let i = 0; i < this.inputs.length; ++i) {
                        feeds.push({ key: this.inputs[i], value: inputs[i] });
                    }
                    const feedDict = new FeedDict(feeds);
                    const outputs = execute(this.outputs, feedDict, { 'training': true });
                    // TODO(cais): Take care of the case of multiple outputs from a
                    //   single layer?
                    let totalLoss;
                    for (let i = 0; i < this.lossFunctions.length; ++i) {
                        const lossFunction = this.lossFunctions[i];
                        let loss = lossFunction(targets[i], outputs[i]);
                        if (sampleWeights[i] != null) {
                            loss = computeWeightedLoss(loss, sampleWeights[i]);
                        }
                        // TODO(cais): push Scalar instead.
                        const meanLoss = mean$1(loss);
                        // TODO(cais): Use a scope() instead, to avoid ownership.
                        lossValues.push(meanLoss);
                        if (i === 0) {
                            totalLoss = loss;
                        }
                        else {
                            totalLoss = add$2(totalLoss, loss);
                        }
                    }
                    // Compute the metrics.
                    // TODO(cais): These should probably be calculated outside
                    //   totalLossFunction to benefit speed?
                    for (let i = 0; i < this.metricsTensors.length; ++i) {
                        let weightedMetric;
                        if (this.outputs.length > 1 && i < this.outputs.length) {
                            weightedMetric = lossValues[i];
                        }
                        else {
                            const metric = this.metricsTensors[i][0];
                            const outputIndex = this.metricsTensors[i][1];
                            weightedMetric =
                                mean$1(metric(targets[outputIndex], outputs[outputIndex]));
                        }
                        keep(weightedMetric);
                        // TODO(cais): Use a scope() instead, to avoid ownership.
                        metricsValues.push(weightedMetric);
                    }
                    totalLoss = mean$1(totalLoss);
                    // Add regularizer penalties.
                    this.calculateLosses().forEach(regularizerLoss => {
                        totalLoss = add$2(totalLoss, regularizerLoss);
                    });
                    return totalLoss;
                };
                const variables = this.collectedTrainableWeights.map(param => param.read());
                const returnCost = true;
                const totalLossValue = this.optimizer_.minimize(totalLossFunction, returnCost, variables);
                return [totalLossValue].concat(metricsValues);
            };
        }
        /**
         * Create a function which, when invoked with an array of `tf.Tensor`s as a
         * batch of inputs, returns the prespecified loss and metrics of the model
         * under the batch of input data.
         */
        makeTestFunction() {
            this.testFunction = (data) => {
                return tidy(() => {
                    const valOutputs = [];
                    let totalLoss;
                    const inputs = data.slice(0, this.inputs.length);
                    const targets = data.slice(this.inputs.length, this.inputs.length + this.outputs.length);
                    const feeds = [];
                    for (let i = 0; i < this.inputs.length; ++i) {
                        feeds.push({ key: this.inputs[i], value: inputs[i] });
                    }
                    const feedDict = new FeedDict(feeds);
                    const outputs = execute(this.outputs, feedDict);
                    // Compute total loss.
                    for (let i = 0; i < this.lossFunctions.length; ++i) {
                        const lossFunction = this.lossFunctions[i];
                        // TODO(cais): Add sample weighting and replace the simple
                        // averaging.
                        const loss = mean$1(lossFunction(targets[i], outputs[i]));
                        if (i === 0) {
                            totalLoss = loss;
                        }
                        else {
                            totalLoss = add$2(totalLoss, loss);
                        }
                        valOutputs.push(totalLoss);
                    }
                    // Compute the metrics.
                    for (let i = 0; i < this.metricsTensors.length; ++i) {
                        const metric = this.metricsTensors[i][0];
                        const outputIndex = this.metricsTensors[i][1];
                        // TODO(cais): Replace K.mean() with a proper weighting function.
                        const meanMetric = mean$1(metric(targets[outputIndex], outputs[outputIndex]));
                        valOutputs.push(meanMetric);
                    }
                    return valOutputs;
                });
            };
        }
        /**
         * Trains the model for a fixed number of epochs (iterations on a
         * dataset).
         *
         * ```js
         * const model = tf.sequential({
         *     layers: [tf.layers.dense({units: 1, inputShape: [10]})]
         * });
         * model.compile({optimizer: 'sgd', loss: 'meanSquaredError'});
         * for (let i = 1; i < 5 ; ++i) {
         *   const h = await model.fit(tf.ones([8, 10]), tf.ones([8, 1]), {
         *       batchSize: 4,
         *       epochs: 3
         *   });
         *   console.log("Loss after Epoch " + i + " : " + h.history.loss[0]);
         * }
         * ```
         *
         * @param x `tf.Tensor` of training data, or an array of `tf.Tensor`s if the
         * model has multiple inputs. If all inputs in the model are named, you
         * can also pass a dictionary mapping input names to `tf.Tensor`s.
         * @param y `tf.Tensor` of target (label) data, or an array of `tf.Tensor`s if
         * the model has multiple outputs. If all outputs in the model are named,
         * you can also pass a dictionary mapping output names to `tf.Tensor`s.
         * @param args A `ModelFitArgs`, containing optional fields.
         *
         * @return A `History` instance. Its `history` attribute contains all
         *   information collected during training.
         *
         * @exception ValueError In case of mismatch between the provided input
         * data and what the model expects.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        async fit(x, y, args = {}) {
            return fitTensors(this, x, y, args);
        }
        // TODO(cais): Add code snippet below when it's possible to instantiate
        //   actual dataset objects.
        /**
         * Trains the model using a dataset object.
         *
         * @param dataset A dataset object. Its `iterator()` method is expected
         *   to generate a dataset iterator object, the `next()` method of which
         *   is expected to produce data batches for training. The return value
         *   of the `next()` call ought to contain a boolean `done` field and a
         *   `value` field. The `value` field is expected to be an array of two
         *   `tf.Tensor`s or an array of two nested `tf.Tensor` structures. The former
         *   case is for models with exactly one input and one output (e.g..
         *   a sequential model). The latter case is for models with multiple
         *   inputs and/or multiple outputs.
         *   Of the two items in the array, the first is the input feature(s) and
         *   the second is the output target(s).
         * @param args A `ModelFitDatasetArgs`, containing optional fields.
         *
         * @return A `History` instance. Its `history` attribute contains all
         *   information collected during training.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        async fitDataset(dataset, args) {
            return fitDataset(this, dataset, args);
        }
        /**
         * Runs a single gradient update on a single batch of data.
         *
         * This method differs from `fit()` and `fitDataset()` in the following
         * regards:
         *   - It operates on exactly one batch of data.
         *   - It returns only the loss and matric values, instead of
         *     returning the batch-by-batch loss and metric values.
         *   - It doesn't support fine-grained options such as verbosity and
         *     callbacks.
         *
         * @param x Input data. It could be one of the following:
         *   - A `tf.Tensor`, or an Array of `tf.Tensor`s (in case the model has
         *     multiple inputs).
         *   - An Object mapping input names to corresponding `tf.Tensor` (if the
         *     model has named inputs).
         * @param y Target darta. It could be either a `tf.Tensor` a multiple
         *   `tf.Tensor`s. It should be consistent with `x`.
         * @returns Training loss or losses (in case the model has
         *   multiple outputs), along with metrics (if any), as numbers.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        async trainOnBatch(x, y) {
            // TODO(cais): Support sampleWeight and classWeight.
            // TODO(cais): Support Dataset objects.
            const standardizeOut = await this.standardizeUserData(x, y);
            const inputs = standardizeOut[0];
            const targets = standardizeOut[1];
            const trainFunction = this.makeTrainFunction();
            const losses = trainFunction(inputs.concat(targets));
            const lossValues = [];
            for (const loss of losses) {
                const v = await loss.data();
                lossValues.push(v[0]);
            }
            dispose(losses);
            disposeNewTensors(standardizeOut[0], x);
            disposeNewTensors(standardizeOut[1], y);
            return singletonOrArray(lossValues);
        }
        /**
         * Extract weight values of the model.
         *
         * @param config: An instance of `io.SaveConfig`, which specifies
         * model-saving options such as whether only trainable weights are to be
         * saved.
         * @returns A `NamedTensorMap` mapping original weight names (i.e.,
         *   non-uniqueified weight names) to their values.
         */
        getNamedWeights(config) {
            const namedWeights = [];
            const trainableOnly = config != null && config.trainableOnly;
            const weights = trainableOnly ? this.trainableWeights : this.weights;
            const weightValues = this.getWeights(trainableOnly);
            for (let i = 0; i < weights.length; ++i) {
                if (trainableOnly && !weights[i].trainable) {
                    // Optionally skip non-trainable weights.
                    continue;
                }
                namedWeights.push({ name: weights[i].originalName, tensor: weightValues[i] });
            }
            return namedWeights;
        }
        /**
         * Setter used for force stopping of LayersModel.fit() (i.e., training).
         *
         * Example:
         *
         * ```js
         * const input = tf.input({shape: [10]});
         * const output = tf.layers.dense({units: 1}).apply(input);
         * const model = tf.model({inputs: [input], outputs: [output]});
         * model.compile({loss: 'meanSquaredError', optimizer: 'sgd'});
         * const xs = tf.ones([8, 10]);
         * const ys = tf.zeros([8, 1]);
         *
         * const history = await model.fit(xs, ys, {
         *   epochs: 10,
         *   callbacks: {
         *     onEpochEnd: async (epoch, logs) => {
         *       if (epoch === 2) {
         *         model.stopTraining = true;
         *       }
         *     }
         *   }
         * });
         *
         * // There should be only 3 values in the loss array, instead of 10
         * values,
         * // due to the stopping after 3 epochs.
         * console.log(history.history.loss);
         * ```
         */
        set stopTraining(stop) {
            this.stopTraining_ = stop;
        }
        get stopTraining() {
            return this.stopTraining_;
        }
        get optimizer() {
            return this.optimizer_;
        }
        set optimizer(optimizer) {
            if (this.optimizer_ !== optimizer) {
                this.optimizer_ = optimizer;
                this.isOptimizerOwned = false;
            }
        }
        dispose() {
            const result = super.dispose();
            if (result.refCountAfterDispose === 0 && this.optimizer != null &&
                this.isOptimizerOwned) {
                const numTensorsBeforeOptmizerDisposal = memory().numTensors;
                this.optimizer_.dispose();
                result.numDisposedVariables +=
                    numTensorsBeforeOptmizerDisposal - memory().numTensors;
            }
            return result;
        }
        getLossIdentifiers() {
            let lossNames;
            if (typeof this.loss === 'string') {
                lossNames = toSnakeCase(this.loss);
            }
            else if (Array.isArray(this.loss)) {
                for (const loss of this.loss) {
                    if (typeof loss !== 'string') {
                        throw new Error('Serialization of non-string loss is not supported.');
                    }
                }
                lossNames = this.loss.map(name => toSnakeCase(name));
            }
            else {
                const outputNames = Object.keys(this.loss);
                lossNames = {};
                const losses = this.loss;
                for (const outputName of outputNames) {
                    if (typeof losses[outputName] === 'string') {
                        lossNames[outputName] =
                            toSnakeCase(losses[outputName]);
                    }
                    else {
                        throw new Error('Serialization of non-string loss is not supported.');
                    }
                }
            }
            return lossNames;
        }
        getMetricIdentifiers() {
            if (typeof this.metrics === 'string' ||
                typeof this.metrics === 'function') {
                return [toSnakeCase(getLossOrMetricName(this.metrics))];
            }
            else if (Array.isArray(this.metrics)) {
                return this.metrics.map(metric => toSnakeCase(getLossOrMetricName(metric)));
            }
            else {
                const metricsIdentifiers = {};
                for (const key in this.metrics) {
                    metricsIdentifiers[key] =
                        toSnakeCase(getLossOrMetricName(this.metrics[key]));
                }
                return metricsIdentifiers;
            }
        }
        getTrainingConfig() {
            return {
                loss: this.getLossIdentifiers(),
                metrics: this.getMetricIdentifiers(),
                optimizer_config: {
                    class_name: this.optimizer.getClassName(),
                    config: this.optimizer.getConfig()
                }
            };
            // TODO(cais): Add weight_metrics when they are supported.
            // TODO(cais): Add sample_weight_mode when it's supported.
            // TODO(cais): Add loss_weights when it's supported.
        }
        loadTrainingConfig(trainingConfig) {
            if (trainingConfig.weighted_metrics != null) {
                throw new Error('Loading weight_metrics is not supported yet.');
            }
            if (trainingConfig.loss_weights != null) {
                throw new Error('Loading loss_weights is not supported yet.');
            }
            if (trainingConfig.sample_weight_mode != null) {
                throw new Error('Loading sample_weight_mode is not supported yet.');
            }
            const tsConfig = convertPythonicToTs(trainingConfig.optimizer_config);
            const optimizer = deserialize(tsConfig);
            let loss;
            if (typeof trainingConfig.loss === 'string') {
                loss = toCamelCase(trainingConfig.loss);
            }
            else if (Array.isArray(trainingConfig.loss)) {
                loss = trainingConfig.loss.map(lossEntry => toCamelCase(lossEntry));
            }
            else if (trainingConfig.loss != null) {
                loss = {};
                for (const key in trainingConfig.loss) {
                    loss[key] = toCamelCase(trainingConfig.loss[key]);
                }
            }
            let metrics;
            if (Array.isArray(trainingConfig.metrics)) {
                metrics = trainingConfig.metrics.map(metric => toCamelCase(metric));
            }
            else if (trainingConfig.metrics != null) {
                metrics = {};
                for (const key in trainingConfig.metrics) {
                    metrics[key] = toCamelCase(trainingConfig.metrics[key]);
                }
            }
            this.compile({ loss, metrics, optimizer });
        }
        /**
         * Save the configuration and/or weights of the LayersModel.
         *
         * An `IOHandler` is an object that has a `save` method of the proper
         * signature defined. The `save` method manages the storing or
         * transmission of serialized data ("artifacts") that represent the
         * model's topology and weights onto or via a specific medium, such as
         * file downloads, local storage, IndexedDB in the web browser and HTTP
         * requests to a server. TensorFlow.js provides `IOHandler`
         * implementations for a number of frequently used saving mediums, such as
         * `tf.io.browserDownloads` and `tf.io.browserLocalStorage`. See `tf.io`
         * for more details.
         *
         * This method also allows you to refer to certain types of `IOHandler`s
         * as URL-like string shortcuts, such as 'localstorage://' and
         * 'indexeddb://'.
         *
         * Example 1: Save `model`'s topology and weights to browser [local
         * storage](https://developer.mozilla.org/en-US/docs/Web/API/Window/localStorage);
         * then load it back.
         *
         * ```js
         * const model = tf.sequential(
         *     {layers: [tf.layers.dense({units: 1, inputShape: [3]})]});
         * console.log('Prediction from original model:');
         * model.predict(tf.ones([1, 3])).print();
         *
         * const saveResults = await model.save('localstorage://my-model-1');
         *
         * const loadedModel = await tf.loadLayersModel('localstorage://my-model-1');
         * console.log('Prediction from loaded model:');
         * loadedModel.predict(tf.ones([1, 3])).print();
         * ```
         *
         * Example 2. Saving `model`'s topology and weights to browser
         * [IndexedDB](https://developer.mozilla.org/en-US/docs/Web/API/IndexedDB_API);
         * then load it back.
         *
         * ```js
         * const model = tf.sequential(
         *     {layers: [tf.layers.dense({units: 1, inputShape: [3]})]});
         * console.log('Prediction from original model:');
         * model.predict(tf.ones([1, 3])).print();
         *
         * const saveResults = await model.save('indexeddb://my-model-1');
         *
         * const loadedModel = await tf.loadLayersModel('indexeddb://my-model-1');
         * console.log('Prediction from loaded model:');
         * loadedModel.predict(tf.ones([1, 3])).print();
         * ```
         *
         * Example 3. Saving `model`'s topology and weights as two files
         * (`my-model-1.json` and `my-model-1.weights.bin`) downloaded from
         * browser.
         *
         * ```js
         * const model = tf.sequential(
         *     {layers: [tf.layers.dense({units: 1, inputShape: [3]})]});
         * const saveResults = await model.save('downloads://my-model-1');
         * ```
         *
         * Example 4. Send  `model`'s topology and weights to an HTTP server.
         * See the documentation of `tf.io.http` for more details
         * including specifying request parameters and implementation of the
         * server.
         *
         * ```js
         * const model = tf.sequential(
         *     {layers: [tf.layers.dense({units: 1, inputShape: [3]})]});
         * const saveResults = await model.save('http://my-server/model/upload');
         * ```
         *
         * @param handlerOrURL An instance of `IOHandler` or a URL-like,
         * scheme-based string shortcut for `IOHandler`.
         * @param config Options for saving the model.
         * @returns A `Promise` of `SaveResult`, which summarizes the result of
         * the saving, such as byte sizes of the saved artifacts for the model's
         *   topology and weight values.
         *
         * @doc {heading: 'Models', subheading: 'Classes', ignoreCI: true}
         */
        async save(handlerOrURL, config) {
            if (typeof handlerOrURL === 'string') {
                const handlers = getSaveHandlers(handlerOrURL);
                if (handlers.length === 0) {
                    throw new ValueError(`Cannot find any save handlers for URL '${handlerOrURL}'`);
                }
                else if (handlers.length > 1) {
                    throw new ValueError(`Found more than one (${handlers.length}) save handlers for ` +
                        `URL '${handlerOrURL}'`);
                }
                handlerOrURL = handlers[0];
            }
            if (handlerOrURL.save == null) {
                throw new ValueError('LayersModel.save() cannot proceed because the IOHandler ' +
                    'provided does not have the `save` attribute defined.');
            }
            const weightDataAndSpecs = await encodeWeights(this.getNamedWeights(config));
            const returnString = false;
            const unusedArg = null;
            const modelConfig = this.toJSON(unusedArg, returnString);
            const modelArtifacts = {
                modelTopology: modelConfig,
                format: LAYERS_MODEL_FORMAT_NAME,
                generatedBy: `TensorFlow.js tfjs-layers v${version$6}`,
                convertedBy: null,
            };
            const includeOptimizer = config == null ? false : config.includeOptimizer;
            if (includeOptimizer && this.optimizer != null) {
                modelArtifacts.trainingConfig = this.getTrainingConfig();
                const weightType = 'optimizer';
                const { data: optimizerWeightData, specs: optimizerWeightSpecs } = await encodeWeights(await this.optimizer.getWeights(), weightType);
                weightDataAndSpecs.specs.push(...optimizerWeightSpecs);
                weightDataAndSpecs.data = concatenateArrayBuffers([weightDataAndSpecs.data, optimizerWeightData]);
            }
            if (this.userDefinedMetadata != null) {
                // Check serialized size of user-defined metadata.
                const checkSize = true;
                checkUserDefinedMetadata(this.userDefinedMetadata, this.name, checkSize);
                modelArtifacts.userDefinedMetadata = this.userDefinedMetadata;
            }
            modelArtifacts.weightData = weightDataAndSpecs.data;
            modelArtifacts.weightSpecs = weightDataAndSpecs.specs;
            return handlerOrURL.save(modelArtifacts);
        }
        /**
         * Set user-defined metadata.
         *
         * The set metadata will be serialized together with the topology
         * and weights of the model during `save()` calls.
         *
         * @param setUserDefinedMetadata
         */
        setUserDefinedMetadata(userDefinedMetadata) {
            checkUserDefinedMetadata(userDefinedMetadata, this.name);
            this.userDefinedMetadata = userDefinedMetadata;
        }
        /**
         * Get user-defined metadata.
         *
         * The metadata is supplied via one of the two routes:
         *   1. By calling `setUserDefinedMetadata()`.
         *   2. Loaded during model loading (if the model is constructed
         *      via `tf.loadLayersModel()`.)
         *
         * If no user-defined metadata is available from either of the
         * two routes, this function will return `undefined`.
         */
        getUserDefinedMetadata() {
            return this.userDefinedMetadata;
        }
    }
    // The class name is 'Model' rather than 'LayersModel' for backwards
    // compatibility since this class name shows up in the serialization format.
    /** @nocollapse */
    LayersModel.className = 'Model';
    registerClass(LayersModel);
    /**
     * A `tf.Functional` is an alias to `tf.LayersModel`.
     *
     * See also:
     *   `tf.LayersModel`, `tf.Sequential`, `tf.loadLayersModel`.
     */
    /** @doc {heading: 'Models', subheading: 'Classes'} */
    class Functional extends LayersModel {
    }
    Functional.className = 'Functional';
    registerClass(Functional);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Parses a JSON model configuration file and returns a model instance.
     *
     * ```js
     * // This example shows how to serialize a model using `toJSON()` and
     * // deserialize it as another model using `tf.models.modelFromJSON()`.
     * // Note: this example serializes and deserializes only the topology
     * // of the model; the weights of the loaded model will be different
     * // from those of the the original model, due to random weight
     * // initialization.
     * // To load the topology and weights of a model, use `tf.loadLayersModel()`.
     * const model1 = tf.sequential();
     * model1.add(tf.layers.repeatVector({inputShape: [2], n: 4}));
     * // Serialize `model1` as a JSON object.
     * const model1JSON = model1.toJSON(null, false);
     * model1.summary();
     *
     * const model2 = await tf.models.modelFromJSON(model1JSON);
     * model2.summary();
     * ```
     *
     *  @param modelAndWeightsConfig JSON object or string encoding a model and
     *       weights configuration. It can also be only the topology JSON of the
     *       model, in which case the weights will not be loaded.
     *  @param custom_objects Optional dictionary mapping names
     *       (strings) to custom classes or functions to be
     *       considered during deserialization.
     * @returns A TensorFlow.js Layers `tf.LayersModel` instance (uncompiled).
     */
    async function modelFromJSON(modelAndWeightsConfig, customObjects) {
        if (!('modelTopology' in modelAndWeightsConfig)) {
            modelAndWeightsConfig = { modelTopology: modelAndWeightsConfig };
        }
        modelAndWeightsConfig = modelAndWeightsConfig;
        let modelTopology = modelAndWeightsConfig.modelTopology;
        if (modelTopology['model_config'] != null) {
            // If the model-topology JSON contains a 'model_config' field, then it is
            // a full model JSON (e.g., from `keras.Model.save()`), which contains
            // not only the model's architecture in its 'model_config' field, but
            // additional information such as the model's optimizer. We use only the
            // 'model_config' field currently.
            modelTopology = modelTopology['model_config'];
        }
        const tsConfig = convertPythonicToTs(modelTopology);
        const model = deserialize(tsConfig, customObjects);
        if (modelAndWeightsConfig.weightsManifest != null) {
            // Load the weight values keyed by the original tensor names in the model
            // file that was loaded.  These should match the keys of the weight
            // manifest.
            const weightValues = await loadWeights(modelAndWeightsConfig.weightsManifest, modelAndWeightsConfig.pathPrefix, model.weights.map(weight => weight.originalName));
            // Map the weights to the unique tensor names generated during model loading
            const uniqueWeightValues = {};
            for (const weight of model.weights) {
                uniqueWeightValues[weight.originalName] =
                    weightValues[weight.originalName];
            }
            model.loadWeights(uniqueWeightValues);
            // Dispose temporary weight values.
            dispose(weightValues);
        }
        return model;
    }
    /**
     * Load a model, including its topology and optionally weights.  See the
     * Tutorial named "How to import a Keras Model" for usage examples.
     *
     * Example 1: Save `model`'s topology and weights to browser [local
     * storage](https://developer.mozilla.org/en-US/docs/Web/API/Window/localStorage);
     * then load it back.
     *
     * ```js
     * const model = tf.sequential(
     *     {layers: [tf.layers.dense({units: 1, inputShape: [3]})]});
     * console.log('Prediction from original model:');
     * model.predict(tf.ones([1, 3])).print();
     *
     * const saveResults = await model.save('localstorage://my-model-1');
     *
     * const loadedModel = await tf.loadLayersModel('localstorage://my-model-1');
     * console.log('Prediction from loaded model:');
     * loadedModel.predict(tf.ones([1, 3])).print();
     * ```
     *
     * Example 2. Saving `model`'s topology and weights to browser
     * [IndexedDB](https://developer.mozilla.org/en-US/docs/Web/API/IndexedDB_API);
     * then load it back.
     *
     * ```js
     * const model = tf.sequential(
     *     {layers: [tf.layers.dense({units: 1, inputShape: [3]})]});
     * console.log('Prediction from original model:');
     * model.predict(tf.ones([1, 3])).print();
     *
     * const saveResults = await model.save('indexeddb://my-model-1');
     *
     * const loadedModel = await tf.loadLayersModel('indexeddb://my-model-1');
     * console.log('Prediction from loaded model:');
     * loadedModel.predict(tf.ones([1, 3])).print();
     * ```
     *
     * Example 3. Load a model from user-selected files from HTML
     * [file input
     * elements](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/file).
     *
     * ```js
     * // Note: this code snippet will not work without the HTML elements in the
     * //   page
     * const jsonUpload = document.getElementById('json-upload');
     * const weightsUpload = document.getElementById('weights-upload');
     *
     * const model = await tf.loadLayersModel(
     *     tf.io.browserFiles([jsonUpload.files[0], weightsUpload.files[0]]));
     * ```
     *
     * Example 4. Load a model from an HTTP server.
     *
     * ```js
     * const model = await
     *     tf.loadLayersModel('https://storage.googleapis.com/tfjs-models/tfjs/iris_v1/model.json');
     * model.summary();
     * ```
     *
     * @param pathOrIOHandler Can be either of the two formats
     *   1. A string path to the `ModelAndWeightsConfig` JSON describing
     *      the model in the canonical TensorFlow.js format. This path will be
     *      interpreted as a relative HTTP path, to which `fetch` will be used to
     *      request the model topology and weight manifest JSON.
     *      The content of the JSON file is assumed to be a JSON object with the
     *      following fields and values:
     *      - 'modelTopology': A JSON object that can be either of:
     *        1. a model architecture JSON consistent with the format of the return
     *            value of `keras.Model.to_json()`
     *        2. a full model JSON in the format of `keras.models.save_model()`.
     *      - 'weightsManifest': A TensorFlow.js weights manifest.
     *      See the Python converter function `save_model()` for more details.
     *      It is also assumed that model weights can be accessed from relative
     *      paths described by the `paths` fields in weights manifest.
     *   2. An `tf.io.IOHandler` object that loads model artifacts with its `load`
     *      method.
     * @param options Optional configuration arguments for the model loading,
     *   including:
     *   - `strict`: Require that the provided weights exactly match those required
     *     by the layers.  Default true.  Passing false means that both extra
     *     weights and missing weights will be silently ignored.
     *   - `onProgress`: A progress callback of the form:
     *     `(fraction: number) => void`. This callback can be used to monitor the
     *     model-loading process.
     * @returns A `Promise` of `tf.LayersModel`, with the topology and weights
     *     loaded.
     */
    async function loadLayersModelInternal(pathOrIOHandler, options) {
        if (options == null) {
            options = {};
        }
        if (typeof pathOrIOHandler === 'string') {
            const handlers = getLoadHandlers(pathOrIOHandler, options);
            if (handlers.length === 0) {
                // For backward compatibility: if no load handler can be found,
                // assume it is a relative http path.
                // TODO(cais): Reformat the args into a single `LoadOptions` once the core
                // is refactored.
                handlers.push(browserHTTPRequest(pathOrIOHandler, options));
            }
            else if (handlers.length > 1) {
                throw new ValueError(`Found more than one (${handlers.length}) load handlers for ` +
                    `URL '${pathOrIOHandler}'`);
            }
            pathOrIOHandler = handlers[0];
        }
        return loadLayersModelFromIOHandler(pathOrIOHandler, undefined, options);
    }
    /**
     * Load a model and optionally its weights, using an IOHandler object.
     *
     * @param handler The instance of `IOHandler` to be used during the model
     *   loading.
     * @param customObjects Any optional custom objects to be used during model
     *   loading.
     * @param strict Whether the weight loading will be done in strict mode.
     *   Default: `true`.
     */
    async function loadLayersModelFromIOHandler(handler, customObjects, options) {
        if (options == null) {
            options = {};
        }
        if (handler.load == null) {
            throw new ValueError('Cannot proceed with model loading because the IOHandler provided ' +
                'does not have the `load` method implemented.');
        }
        const artifacts = await handler.load();
        let modelTopology = artifacts.modelTopology;
        if (modelTopology['model_config'] != null) {
            modelTopology = modelTopology['model_config'];
        }
        const strict = options.strict == null ? true : options.strict;
        // If weights are provided and the weight-loading mode is strict, use
        // fast weight initialization. This skips costly initializers such as
        // 'orthogonal' and saves unnecessary computation in cases where
        // the initialized weight values will immediately be overwritten by
        // loaded weight values.
        const fastWeightInit = artifacts.weightData != null && artifacts.weightSpecs != null && strict;
        const model = deserialize(convertPythonicToTs(modelTopology), customObjects, fastWeightInit);
        const trainingConfig = artifacts.trainingConfig;
        if (trainingConfig != null) {
            model.loadTrainingConfig(trainingConfig);
        }
        if (artifacts.userDefinedMetadata != null) {
            model.setUserDefinedMetadata(artifacts.userDefinedMetadata);
        }
        // If weightData is present, load the weights into the model.
        if (artifacts.weightData != null) {
            // Loading weights requires weightSpecs.
            if (artifacts.weightSpecs == null) {
                throw new ValueError('LayersModel artifacts contains weight data, but not weight specs. ' +
                    'Therefore loading of weights cannot proceed.');
            }
            const { modelWeights, optimizerWeights } = decodeModelAndOptimizerWeights(artifacts.weightData, artifacts.weightSpecs);
            model.loadWeights(modelWeights, strict);
            if (model.optimizer != null && optimizerWeights.length > 0) {
                await model.optimizer.setWeights(optimizerWeights);
            }
            // Dispose temporary weight values.
            dispose(modelWeights);
            dispose(optimizerWeights.map(w => w.tensor));
        }
        return model;
    }
    function decodeModelAndOptimizerWeights(buffer, specs) {
        const name2Tensor = decodeWeights(buffer, specs);
        const modelWeights = {};
        const optimizerWeights = [];
        specs.forEach(spec => {
            if (spec.group === 'optimizer') {
                optimizerWeights.push({ name: spec.name, tensor: name2Tensor[spec.name] });
            }
            else {
                modelWeights[spec.name] = name2Tensor[spec.name];
            }
        });
        return { modelWeights, optimizerWeights };
    }
    /**
     * A model with a stack of layers, feeding linearly from one to the next.
     *
     * `tf.sequential` is a factory function that creates an instance of
     * `tf.Sequential`.
     *
     * ```js
     *  // Define a model for linear regression.
     *  const model = tf.sequential();
     *  model.add(tf.layers.dense({units: 1, inputShape: [1]}));
     *
     *  // Prepare the model for training: Specify the loss and the optimizer.
     *  model.compile({loss: 'meanSquaredError', optimizer: 'sgd'});
     *
     *  // Generate some synthetic data for training.
     *  const xs = tf.tensor2d([1, 2, 3, 4], [4, 1]);
     *  const ys = tf.tensor2d([1, 3, 5, 7], [4, 1]);
     *
     *  // Train the model using the data then do inference on a data point the
     *  // model hasn't seen:
     *  await model.fit(xs, ys);
     *  model.predict(tf.tensor2d([5], [1, 1])).print();
     * ```
     *
     * @doc {heading: 'Models', subheading: 'Classes'}
     */
    class Sequential extends LayersModel {
        constructor(args) {
            super({ inputs: [], outputs: [] });
            args = args || {};
            this.trainable = true;
            this.built = false;
            // Set model name.
            this.name = (args.name != null) ? args.name : getUid('sequential_');
            // Add to the model any layers passed to the constructor.
            if (args.layers != null) {
                for (const layer of args.layers) {
                    this.add(layer);
                }
            }
        }
        // Helper function to Sequential.add  Throws if the new output shape will be
        // invalid.
        checkShape(layer) {
            const shape = layer.inboundNodes[0].outputTensors[0].shape;
            if (shape.some(x => x < 0)) {
                throw new ValueError('Negative dimension size caused by adding layer ' +
                    `${layer.name} with input shape [` +
                    `${layer.inboundNodes[0].inputTensors[0].shape}]`);
            }
        }
        /**
         * Adds a layer instance on top of the layer stack.
         *
         * ```js
         *  const model = tf.sequential();
         *  model.add(tf.layers.dense({units: 8, inputShape: [1]}));
         *  model.add(tf.layers.dense({units: 4, activation: 'relu6'}));
         *  model.add(tf.layers.dense({units: 1, activation: 'relu6'}));
         *  // Note that the untrained model is random at this point.
         *  model.predict(tf.randomNormal([10, 1])).print();
         * ```
         * @param layer Layer instance.
         *
         * @exception ValueError In case the `layer` argument does not know its
         * input shape.
         * @exception ValueError In case the `layer` argument has multiple output
         *   tensors, or is already connected somewhere else (forbidden in
         *   `Sequential` models).
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        add(layer) {
            const isLayerModelInstance = layer instanceof Sequential || layer instanceof LayersModel;
            let modelLayer;
            if (isLayerModelInstance) {
                modelLayer = layer;
                if (modelLayer.outputs.length !== 1) {
                    throw new ValueError('All layers in a Sequential model ' +
                        'should have a single output tensor. ' +
                        'For multi-output layers, ' +
                        'use the functional API.');
                }
                if (modelLayer.inputs.length !== 1) {
                    throw new ValueError('All layers in a Sequential model ' +
                        'should have a single input tensor. ' +
                        'For multi-input layers, ' +
                        'use the functional API.');
                }
            }
            if (this.outputs.length === 0) {
                // first layer in model: check that it is an input layer
                if (layer.inboundNodes.length === 0) {
                    // create an input layer
                    if (layer.batchInputShape == null) {
                        throw new ValueError('The first layer in a Sequential model must ' +
                            'get an `inputShape` or `batchInputShape` argument.');
                    }
                    // Instantiate the input layer.
                    const x = Input({
                        batchShape: layer.batchInputShape,
                        dtype: layer.dtype,
                        name: layer.name + '_input'
                    });
                    // This will build the current layer and create the node connecting
                    // the current layer to the input layer we just created.
                    layer.apply(x);
                }
                if (isLayerModelInstance) {
                    this.outputs = modelLayer.outputs;
                    this.inputs = modelLayer.inputs;
                }
                else {
                    if (layer.inboundNodes.length !== 1) {
                        throw new ValueError('A layer added to a Sequential model must not already be ' +
                            `connected somewhere else. LayersModel received layer ${layer.name} ` +
                            `which has ${layer.inboundNodes.length} pre-existing inbound ` +
                            'connections.');
                    }
                    if (layer.inboundNodes[0].outputTensors.length !== 1) {
                        throw new ValueError('All layers in a Sequential model ' +
                            'should have a single output tensor. ' +
                            'For multi-output layers, ' +
                            'use the functional API.');
                    }
                    this.checkShape(layer);
                    this.outputs = [layer.inboundNodes[0].outputTensors[0]];
                    this.inputs = getSourceInputs(this.outputs[0]);
                }
                this.inboundNodes = [];
                // We create an input node, which we will keep updated
                // as we add more layers.
                // (This call has side effects.)
                // tslint:disable-next-line:no-unused-expression
                new Node({
                    outboundLayer: this,
                    inboundLayers: [],
                    nodeIndices: [],
                    tensorIndices: [],
                    inputTensors: this.inputs,
                    outputTensors: this.outputs,
                    // no model-level masking for now
                    inputMasks: pyListRepeat(null, this.inputs.length),
                    outputMasks: [null],
                    inputShapes: this.inputs.map(x => x.shape),
                    outputShapes: this.outputs[0].shape
                });
            }
            else {
                const outputTensor = layer.apply(this.outputs[0]);
                if (Array.isArray(outputTensor)) {
                    throw new TypeError('All layers in a Sequential model ' +
                        'should have a single output tensor. ' +
                        'For multi-output layers, ' +
                        'use the functional API.');
                }
                this.checkShape(layer);
                this.outputs = [outputTensor];
                // update self.inbound_nodes
                this.inboundNodes[0].outputTensors = this.outputs;
                this.inboundNodes[0].outputShapes = [this.outputs[0].shape];
            }
            this.layers.push(layer);
            this.built = false;
        }
        /**
         * Removes the last layer in the model.
         *
         * @exception TypeError if there are no layers in the model.
         */
        pop() {
            if (this.layers.length === 0) {
                throw new TypeError('There are no layers in the model.');
            }
            this.layers.pop();
            if (this.layers.length === 0) {
                this.outputs = [];
                this.inboundNodes = [];
                this.outboundNodes = [];
            }
            else {
                const lastLayerIndex = this.layers.length - 1;
                this.layers[lastLayerIndex].outboundNodes = [];
                this.outputs = [this.layers[lastLayerIndex].output];
                // update self.inbound_nodes
                this.inboundNodes[0].outputTensors = this.outputs;
                this.inboundNodes[0].outputShapes = [this.outputs[0].shape];
            }
        }
        call(inputs, kwargs) {
            if (this.model == null) {
                this.build();
            }
            return this.model.call(inputs, kwargs);
        }
        build(inputShape) {
            // Call `getExactlyOneShape` without using its return value,
            // to verify that exactly one input shape is provided.
            getExactlyOneShape(inputShape);
            if (this.inputs.length === 0 || this.outputs.length === 0) {
                throw new TypeError('Sequential model cannot be built: model is empty.' +
                    ' Add some layers first.');
            }
            // actually create the model
            this.model = new LayersModel({
                inputs: this.inputs,
                outputs: this.outputs[0],
                name: this.name + '_model'
            });
            this.model.trainable = this.trainable;
            // mirror model attributes
            this.supportsMasking = this.model.supportsMasking;
            // TODO(michaelterry): Add caches
            this.inputLayers = this.model.inputLayers;
            this.inputLayersNodeIndices = this.model.inputLayersNodeIndices;
            this.inputLayersTensorIndices = this.model.inputLayersTensorIndices;
            this.outputLayers = this.model.outputLayers;
            this.outputLayersNodeIndices = this.model.outputLayersNodeIndices;
            this.outputLayersTensorIndices = this.model.outputLayersTensorIndices;
            this.nodesByDepth = this.model.nodesByDepth;
            this.containerNodes = this.model.containerNodes;
            this.outputNames = this.model.outputNames;
            this.inputNames = this.model.inputNames;
            // TODO(michaelterry): Add feedInputNames, feedInputs, if needed.
            // TODO(michaelterry): Add callbackModel if needed.
            this.built = true;
        }
        countParams() {
            if (!this.built) {
                this.build();
            }
            return super.countParams();
        }
        /**
         * Print a text summary of the Sequential model's layers.
         *
         * The summary includes
         * - Name and type of all layers that comprise the model.
         * - Output shape(s) of the layers
         * - Number of weight parameters of each layer
         * - The total number of trainable and non-trainable parameters of the
         * model.
         *
         * ```js
         * const model = tf.sequential();
         * model.add(
         *     tf.layers.dense({units: 100, inputShape: [10], activation: 'relu'}));
         * model.add(tf.layers.dense({units: 1, activation: 'sigmoid'}));
         *
         * model.summary();
         * ```
         *
         * @param lineLength Custom line length, in number of characters.
         * @param positions Custom widths of each of the columns, as either
         *   fractions of `lineLength` (e.g., `[0.5, 0.75, 1]`) or absolute number
         *   of characters (e.g., `[30, 50, 65]`). Each number corresponds to
         *   right-most (i.e., ending) position of a column.
         * @param printFn Custom print function. Can be used to replace the default
         *   `console.log`. For example, you can use `x => {}` to mute the printed
         *   messages in the console.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        summary(lineLength, positions, printFn = console.log) {
            if (!this.built) {
                this.build();
            }
            super.summary(lineLength, positions, printFn);
        }
        /**
         * Sets the weights of the model.
         *
         * @param weights Should be a list of Tensors with shapes and types matching
         *   the output of `model.getWeights()`.
         */
        setWeights(weights) {
            if (this.model == null) {
                this.build();
            }
            this.model.setWeights(weights);
        }
        /**
         * Returns the loss value & metrics values for the model in test mode.
         *
         * Loss and metrics are specified during `compile()`, which needs to happen
         * before calls to `evaluate()`.
         *
         * Computation is done in batches.
         *
         * ```js
         * const model = tf.sequential({
         *   layers: [tf.layers.dense({units: 1, inputShape: [10]})]
         * });
         * model.compile({optimizer: 'sgd', loss: 'meanSquaredError'});
         * const result = model.evaluate(tf.ones([8, 10]), tf.ones([8, 1]), {
         *   batchSize: 4,
         * });
         * result.print();
         * ```
         *
         * @param x `tf.Tensor` of test data, or an `Array` of `tf.Tensor`s if the
         * model has multiple inputs.
         * @param y `tf.Tensor` of target data, or an `Array` of `tf.Tensor`s if the
         * model has multiple outputs.
         * @param args A `ModelEvaluateConfig`, containing optional fields.
         *
         * @return `Scalar` test loss (if the model has a single output and no
         *   metrics) or `Array` of `Scalar`s (if the model has multiple outputs
         *   and/or metrics). The attribute `model.metricsNames`
         *   will give you the display labels for the scalar outputs.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        evaluate(x, y, args = {}) {
            if (!this.built) {
                throw new RuntimeError('The model needs to be compiled before being used.');
            }
            return this.model.evaluate(x, y, args);
        }
        // TODO(cais): Add code snippet below once real dataset objects are
        //   available.
        /**
         * Evaluate model using a dataset object.
         *
         * Note: Unlike `evaluate()`, this method is asynchronous (`async`);
         *
         * @param dataset A dataset object. Its `iterator()` method is expected
         *   to generate a dataset iterator object, the `next()` method of which
         *   is expected to produce data batches for evaluation. The return value
         *   of the `next()` call ought to contain a boolean `done` field and a
         *   `value` field. The `value` field is expected to be an array of two
         *   `tf.Tensor`s or an array of two nested `tf.Tensor` structures. The former
         *   case is for models with exactly one input and one output (e.g..
         *   a sequential model). The latter case is for models with multiple
         *   inputs and/or multiple outputs. Of the two items in the array, the
         *   first is the input feature(s) and the second is the output target(s).
         * @param args A configuration object for the dataset-based evaluation.
         * @returns Loss and metric values as an Array of `Scalar` objects.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        async evaluateDataset(dataset, args) {
            if (!this.built) {
                throw new RuntimeError('The model needs to be compiled before being used.');
            }
            return this.model.evaluateDataset(dataset, args);
        }
        /**
         * Generates output predictions for the input samples.
         *
         * Computation is done in batches.
         *
         * Note: the "step" mode of predict() is currently not supported.
         *   This is because the TensorFlow.js core backend is imperative only.
         *
         * ```js
         * const model = tf.sequential({
         *   layers: [tf.layers.dense({units: 1, inputShape: [10]})]
         * });
         * model.predict(tf.ones([2, 10])).print();
         * ```
         *
         * @param x The input data, as a Tensor, or an `Array` of `tf.Tensor`s if
         *   the model has multiple inputs.
         * @param conifg A `ModelPredictConfig` object containing optional fields.
         *
         * @return `tf.Tensor`(s) of predictions.
         *
         * @exception ValueError In case of mismatch between the provided input data
         *   and the model's expectations, or in case a stateful model receives a
         *   number of samples that is not a multiple of the batch size.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        predict(x, args = {}) {
            if (this.model == null) {
                this.build();
            }
            return this.model.predict(x, args);
        }
        /**
         * Returns predictions for a single batch of samples.
         *
         * @param x: Input samples, as a Tensor, or list of Tensors (if the model
         *   has multiple inputs).
         * @return Tensor(s) of predictions
         */
        predictOnBatch(x) {
            if (this.model == null) {
                this.build();
            }
            return this.model.predictOnBatch(x);
        }
        /**
         * See `LayersModel.compile`.
         *
         * @param args
         */
        compile(args) {
            this.build();
            this.model.compile(args);
            this.optimizer_ = this.model.optimizer;
            // tslint:disable-next-line:no-any
            this.isOptimizerOwned = this.model.isOptimizerOwned;
            this.loss = this.model.loss;
            this.metrics = this.model.metrics;
            // TODO(cais): Add this.lossWeights, this.sampleWeightMode,
            //   this.weightedMetrics, this.targets.
            this.metricsTensors = this.model.metricsTensors;
            this.metricsNames = this.model.metricsNames;
            // TODO(cais): Add sampleWeights.
        }
        get optimizer() {
            return this.model == null ? undefined : this.model.optimizer;
        }
        set optimizer(optimizer) {
            this.model.optimizer = optimizer;
        }
        /**
         * Trains the model for a fixed number of epochs (iterations on a dataset).
         *
         * ```js
         * const model = tf.sequential({
         *   layers: [tf.layers.dense({units: 1, inputShape: [10]})]
         * });
         * model.compile({optimizer: 'sgd', loss: 'meanSquaredError'});
         * const history = await model.fit(tf.ones([8, 10]), tf.ones([8, 1]), {
         *   batchSize: 4,
         *   epochs: 3
         * });
         * console.log(history.history.loss[0]);
         * ```
         *
         * @param x `tf.Tensor` of training data, or an array of `tf.Tensor`s if the
         * model has multiple inputs. If all inputs in the model are named, you can
         * also pass a dictionary mapping input names to `tf.Tensor`s.
         * @param y `tf.Tensor` of target (label) data, or an array of `tf.Tensor`s if
         * the model has multiple outputs. If all outputs in the model are named, you
         *  can also pass a dictionary mapping output names to `tf.Tensor`s.
         * @param args  A `ModelFitConfig`, containing optional fields.
         *
         * @return A `History` instance. Its `history` attribute contains all
         *   information collected during training.
         *
         * @exception ValueError In case of mismatch between the provided input data
         *   and what the model expects.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        async fit(x, y, args = {}) {
            if (!this.built) {
                throw new RuntimeError('The model needs to be compiled before ' +
                    'being used.');
            }
            return this.model.fit(x, y, args);
        }
        /**
         * Trains the model using a dataset object.
         *
         * ```js
         * const xArray = [
         *   [1, 1, 1, 1, 1, 1, 1, 1, 1],
         *   [1, 1, 1, 1, 1, 1, 1, 1, 1],
         *   [1, 1, 1, 1, 1, 1, 1, 1, 1],
         *   [1, 1, 1, 1, 1, 1, 1, 1, 1],
         * ];
         * const yArray = [1, 1, 1, 1];
         * // Create a dataset from the JavaScript array.
         * const xDataset = tf.data.array(xArray);
         * const yDataset = tf.data.array(yArray);
         * // Zip combines the `x` and `y` Datasets into a single Dataset, the
         * // iterator of which will return an object containing of two tensors,
         * // corresponding to `x` and `y`.  The call to `batch(4)` will bundle
         * // four such samples into a single object, with the same keys now pointing
         * // to tensors that hold 4 examples, organized along the batch dimension.
         * // The call to `shuffle(4)` causes each iteration through the dataset to
         * // happen in a different order.  The size of the shuffle window is 4.
         * const xyDataset = tf.data.zip({xs: xDataset, ys: yDataset})
         *     .batch(4)
         *     .shuffle(4);
         * const model = tf.sequential({
         *   layers: [tf.layers.dense({units: 1, inputShape: [9]})]
         * });
         * model.compile({optimizer: 'sgd', loss: 'meanSquaredError'});
         * const history = await model.fitDataset(xyDataset, {
         *   epochs: 4,
         *   callbacks: {onEpochEnd: (epoch, logs) => console.log(logs.loss)}
         * });
         * ```
         *
         * @param dataset A dataset object. Its `iterator()` method is expected to
         *   generate a dataset iterator object, the `next()` method of which is
         *   expected to produce data batches for evaluation. The return value of the
         *   `next()` call ought to contain a boolean `done` field and a `value`
         *   field.
         *
         *   The `value` field is expected to be an object of with fields
         *   `xs` and `ys`, which point to the feature tensor and the target tensor,
         *   respectively. This case is for models with exactly one input and one
         *   output (e.g.. a sequential model). For example:
         *   ```js
         *   {value: {xs: xsTensor, ys: ysTensor}, done: false}
         *   ```
         *
         *   If the model has multiple inputs, the `xs` field of `value` should
         *   be an object mapping input names to their respective feature tensors.
         *   For example:
         *   ```js
         *   {
         *     value: {
         *       xs: {
         *         input_1: xsTensor1,
         *         input_2: xsTensor2
         *       },
         *       ys: ysTensor
         *     },
         *     done: false
         *   }
         *   ```
         *   If the model has multiple outputs, the `ys` field of `value` should
         *   be an object mapping output names to their respective target tensors.
         *   For example:
         *   ```js
         *   {
         *     value: {
         *       xs: xsTensor,
         *       ys: {
         *         output_1: ysTensor1,
         *         output_2: ysTensor2
         *       },
         *     },
         *     done: false
         *   }
         *   ```
         * @param args A `ModelFitDatasetArgs`, containing optional fields.
         *
         * @return A `History` instance. Its `history` attribute contains all
         *   information collected during training.
         *
         * @doc {heading: 'Models', subheading: 'Classes', ignoreCI: true}
         */
        async fitDataset(dataset, args) {
            if (!this.built) {
                throw new RuntimeError('The model needs to be compiled before ' +
                    'being used.');
            }
            return this.model.fitDataset(dataset, args);
        }
        /**
         * Runs a single gradient update on a single batch of data.
         *
         * This method differs from `fit()` and `fitDataset()` in the following
         * regards:
         *   - It operates on exactly one batch of data.
         *   - It returns only the loss and matric values, instead of
         *     returning the batch-by-batch loss and metric values.
         *   - It doesn't support fine-grained options such as verbosity and
         *     callbacks.
         *
         * @param x Input data. It could be one of the following:
         *   - A `tf.Tensor`, or an Array of `tf.Tensor`s (in case the model has
         *     multiple inputs).
         *   - An Object mapping input names to corresponding `tf.Tensor` (if the
         *     model has named inputs).
         * @param y Target darta. It could be either a `tf.Tensor` a multiple
         *   `tf.Tensor`s. It should be consistent with `x`.
         * @returns Training loss or losses (in case the model has
         *   multiple outputs), along with metrics (if any), as numbers.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        async trainOnBatch(x, y) {
            return this.model.trainOnBatch(x, y);
        }
        /* See parent class for JsDoc */
        /** @nocollapse */
        static fromConfig(cls, config, customObjects = {}, fastWeightInit = false) {
            let configArray;
            let extraModelConfig = {};
            if (config instanceof Array) {
                if (!(config[0].className != null) ||
                    config[0]['className'] === 'Merge') {
                    throw new ValueError('Legacy serialization format not supported yet.');
                }
                configArray = config;
            }
            else {
                assert$4(config['layers'] != null, () => `When the config data for a Sequential model is not an Array, ` +
                    `it must be an Object that contains the 'layers' field.`);
                configArray = config['layers'];
                delete config['layers'];
                extraModelConfig = config;
            }
            const model = new cls(extraModelConfig);
            if (!(model instanceof Sequential)) {
                throw new NotImplementedError(`Sequential.fromConfig called on non-Sequential input: ${model}`);
            }
            for (const conf of configArray) {
                const customObjects = undefined;
                const layer = deserialize(conf, customObjects, fastWeightInit);
                if (fastWeightInit) {
                    layer.setFastWeightInitDuringBuild(true);
                }
                model.add(layer);
            }
            return model;
        }
        /**
         * Setter used for force stopping of LayersModel.fit() (i.e., training).
         *
         * Example:
         *
         * ```js
         * const model = tf.sequential();
         * model.add(tf.layers.dense({units: 1, inputShape: [10]}));
         * model.compile({loss: 'meanSquaredError', optimizer: 'sgd'});
         * const xs = tf.ones([8, 10]);
         * const ys = tf.zeros([8, 1]);
         *
         * const history = await model.fit(xs, ys, {
         *   epochs: 10,
         *   callbacks: {
         *     onEpochEnd: async (epoch, logs) => {
         *       if (epoch === 2) {
         *         model.stopTraining = true;
         *       }
         *     }
         *   }
         * });
         *
         * // There should be only 3 values in the loss array, instead of 10 values,
         * // due to the stopping after 3 epochs.
         * console.log(history.history.loss);
         * ```
         */
        set stopTraining(stop) {
            // TODO(cais): When refactoring to remove the composition pattern happens,
            // remove this method overriding.
            if (this.model == null) {
                throw new ValueError('Cannot set the stopTraining property of a sequential model before ' +
                    'it is compiled.');
            }
            this.model.stopTraining = stop;
        }
        get stopTraining() {
            if (this.model == null) {
                throw new ValueError('Cannot get the stopTraining property of a sequential model before ' +
                    'it is compiled.');
            }
            return this.model.stopTraining;
        }
        // TODO(cais): Override get trainableWeights() here
        // tslint:disable-next-line:no-any
        getConfig() {
            // NOTE(cais): We override the return type of getConfig() to `any` here,
            //   because the `Sequential` class is a special case among `Container`
            //   subtypes in that its getConfig() method returns an Array (not a
            //   dict).
            const layers = [];
            for (const layer of this.layers) {
                const dict = {};
                dict['className'] = layer.getClassName();
                dict['config'] = layer.getConfig();
                layers.push(dict);
            }
            return { name: this.name, layers };
        }
    }
    /** @nocollapse */
    Sequential.className = 'Sequential';
    registerClass(Sequential);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    // TODO(cais): Add doc string to all the public static functions in this
    //   class; include exectuable JavaScript code snippets where applicable
    //   (b/74074458).
    // LayersModel and related factory methods.
    /**
     * A model is a data structure that consists of `Layers` and defines inputs
     * and outputs.
     *
     * The key difference between `tf.model` and `tf.sequential` is that
     * `tf.model` is more generic, supporting an arbitrary graph (without
     * cycles) of layers. `tf.sequential` is less generic and supports only a linear
     * stack of layers.
     *
     * When creating a `tf.LayersModel`, specify its input(s) and output(s). Layers
     * are used to wire input(s) to output(s).
     *
     * For example, the following code snippet defines a model consisting of
     * two `dense` layers, with 10 and 4 units, respectively.
     *
     * ```js
     * // Define input, which has a size of 5 (not including batch dimension).
     * const input = tf.input({shape: [5]});
     *
     * // First dense layer uses relu activation.
     * const denseLayer1 = tf.layers.dense({units: 10, activation: 'relu'});
     * // Second dense layer uses softmax activation.
     * const denseLayer2 = tf.layers.dense({units: 4, activation: 'softmax'});
     *
     * // Obtain the output symbolic tensor by applying the layers on the input.
     * const output = denseLayer2.apply(denseLayer1.apply(input));
     *
     * // Create the model based on the inputs.
     * const model = tf.model({inputs: input, outputs: output});
     *
     * // The model can be used for training, evaluation and prediction.
     * // For example, the following line runs prediction with the model on
     * // some fake data.
     * model.predict(tf.ones([2, 5])).print();
     * ```
     * See also:
     *   `tf.sequential`, `tf.loadLayersModel`.
     *
     * @doc {heading: 'Models', subheading: 'Creation'}
     */
    function model$1(args) {
        return new LayersModel(args);
    }
    /**
     * Creates a `tf.Sequential` model.  A sequential model is any model where the
     * outputs of one layer are the inputs to the next layer, i.e. the model
     * topology is a simple 'stack' of layers, with no branching or skipping.
     *
     * This means that the first layer passed to a `tf.Sequential` model should have
     * a defined input shape. What that means is that it should have received an
     * `inputShape` or `batchInputShape` argument, or for some type of layers
     * (recurrent, Dense...) an `inputDim` argument.
     *
     * The key difference between `tf.model` and `tf.sequential` is that
     * `tf.sequential` is less generic, supporting only a linear stack of layers.
     * `tf.model` is more generic and supports an arbitrary graph (without
     * cycles) of layers.
     *
     * Examples:
     *
     * ```js
     * const model = tf.sequential();
     *
     * // First layer must have an input shape defined.
     * model.add(tf.layers.dense({units: 32, inputShape: [50]}));
     * // Afterwards, TF.js does automatic shape inference.
     * model.add(tf.layers.dense({units: 4}));
     *
     * // Inspect the inferred shape of the model's output, which equals
     * // `[null, 4]`. The 1st dimension is the undetermined batch dimension; the
     * // 2nd is the output size of the model's last layer.
     * console.log(JSON.stringify(model.outputs[0].shape));
     * ```
     *
     * It is also possible to specify a batch size (with potentially undetermined
     * batch dimension, denoted by "null") for the first layer using the
     * `batchInputShape` key. The following example is equivalent to the above:
     *
     * ```js
     * const model = tf.sequential();
     *
     * // First layer must have a defined input shape
     * model.add(tf.layers.dense({units: 32, batchInputShape: [null, 50]}));
     * // Afterwards, TF.js does automatic shape inference.
     * model.add(tf.layers.dense({units: 4}));
     *
     * // Inspect the inferred shape of the model's output.
     * console.log(JSON.stringify(model.outputs[0].shape));
     * ```
     *
     * You can also use an `Array` of already-constructed `Layer`s to create
     * a `tf.Sequential` model:
     *
     * ```js
     * const model = tf.sequential({
     *   layers: [tf.layers.dense({units: 32, inputShape: [50]}),
     *            tf.layers.dense({units: 4})]
     * });
     * console.log(JSON.stringify(model.outputs[0].shape));
     * ```
     *
     * @doc {heading: 'Models', subheading: 'Creation'}
     */
    function sequential(config) {
        return new Sequential(config);
    }
    /**
     * Load a model composed of Layer objects, including its topology and optionally
     * weights. See the Tutorial named "How to import a Keras Model" for usage
     * examples.
     *
     * This method is applicable to:
     *
     * 1. Models created with the `tf.layers.*`, `tf.sequential`, and
     * `tf.model` APIs of TensorFlow.js and later saved with the
     * `tf.LayersModel.save` method.
     * 2. Models converted from Keras or TensorFlow tf.keras using the
     * [tensorflowjs_converter](https://github.com/tensorflow/tfjs/tree/master/tfjs-converter).
     *
     * This mode is *not* applicable to TensorFlow `SavedModel`s or their converted
     * forms. For those models, use `tf.loadGraphModel`.
     *
     * Example 1. Load a model from an HTTP server.
     *
     * ```js
     * const model = await tf.loadLayersModel(
     *     'https://storage.googleapis.com/tfjs-models/tfjs/iris_v1/model.json');
     * model.summary();
     * ```
     *
     * Example 2: Save `model`'s topology and weights to browser [local
     * storage](https://developer.mozilla.org/en-US/docs/Web/API/Window/localStorage);
     * then load it back.
     *
     * ```js
     * const model = tf.sequential(
     *     {layers: [tf.layers.dense({units: 1, inputShape: [3]})]});
     * console.log('Prediction from original model:');
     * model.predict(tf.ones([1, 3])).print();
     *
     * const saveResults = await model.save('localstorage://my-model-1');
     *
     * const loadedModel = await tf.loadLayersModel('localstorage://my-model-1');
     * console.log('Prediction from loaded model:');
     * loadedModel.predict(tf.ones([1, 3])).print();
     * ```
     *
     * Example 3. Saving `model`'s topology and weights to browser
     * [IndexedDB](https://developer.mozilla.org/en-US/docs/Web/API/IndexedDB_API);
     * then load it back.
     *
     * ```js
     * const model = tf.sequential(
     *     {layers: [tf.layers.dense({units: 1, inputShape: [3]})]});
     * console.log('Prediction from original model:');
     * model.predict(tf.ones([1, 3])).print();
     *
     * const saveResults = await model.save('indexeddb://my-model-1');
     *
     * const loadedModel = await tf.loadLayersModel('indexeddb://my-model-1');
     * console.log('Prediction from loaded model:');
     * loadedModel.predict(tf.ones([1, 3])).print();
     * ```
     *
     * Example 4. Load a model from user-selected files from HTML
     * [file input
     * elements](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/file).
     *
     * ```js
     * // Note: this code snippet will not work without the HTML elements in the
     * //   page
     * const jsonUpload = document.getElementById('json-upload');
     * const weightsUpload = document.getElementById('weights-upload');
     *
     * const model = await tf.loadLayersModel(
     *     tf.io.browserFiles([jsonUpload.files[0], weightsUpload.files[0]]));
     * ```
     *
     * @param pathOrIOHandler Can be either of the two formats
     *   1. A string path to the `ModelAndWeightsConfig` JSON describing
     *      the model in the canonical TensorFlow.js format. For file://
     *      (tfjs-node-only), http:// and https:// schemas, the path can be
     *      either absolute or relative.
     *   2. An `tf.io.IOHandler` object that loads model artifacts with its `load`
     *      method.
     * @param options Optional configuration arguments for the model loading,
     *   including:
     *   - `strict`: Require that the provided weights exactly match those required
     *     by the layers.  Default true.  Passing false means that both extra
     *     weights and missing weights will be silently ignored.
     *   - `onProgress`: A function of the signature `(fraction: number) => void',
     *     that can be used as the progress callback for the model loading.
     * @returns A `Promise` of `tf.LayersModel`, with the topology and weights
     *     loaded.
     *
     * @doc {heading: 'Models', subheading: 'Loading'}
     */
    function loadLayersModel(pathOrIOHandler, options) {
        if (options == null) {
            options = {};
        }
        return loadLayersModelInternal(pathOrIOHandler, options);
    }
    /**
     * Used to instantiate an input to a model as a `tf.SymbolicTensor`.
     *
     * Users should call the `input` factory function for
     * consistency with other generator functions.
     *
     * Example:
     *
     * ```js
     * // Defines a simple logistic regression model with 32 dimensional input
     * // and 3 dimensional output.
     * const x = tf.input({shape: [32]});
     * const y = tf.layers.dense({units: 3, activation: 'softmax'}).apply(x);
     * const model = tf.model({inputs: x, outputs: y});
     * model.predict(tf.ones([2, 32])).print();
     * ```
     *
     * Note: `input` is only necessary when using `model`. When using
     * `sequential`, specify `inputShape` for the first layer or use `inputLayer`
     * as the first layer.
     *
     * @doc {heading: 'Models', subheading: 'Inputs'}
     */
    function input(config) {
        return Input(config);
    }
    function registerCallbackConstructor(verbosityLevel, callbackConstructor) {
        CallbackConstructorRegistry.registerCallbackConstructor(verbosityLevel, callbackConstructor);
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Base class for Activations.
     *
     * Special note: due to cross-language compatibility reasons, the
     * static readonly className field in this family of classes must be set to
     * the initialLowerCamelCase name of the activation.
     */
    class Activation$1 extends Serializable {
        getConfig() {
            return {};
        }
    }
    /**
     * Exponential linear unit (ELU).
     * Reference: https://arxiv.org/abs/1511.07289
     */
    class Elu extends Activation$1 {
        /**
         * Calculate the activation function.
         *
         * @param x: Input.
         * @param alpha: Scaling factor the negative section.
         * @return Output of the ELU activation.
         */
        apply(x, alpha = 1) {
            return elu$3(x, alpha);
        }
    }
    /** @nocollapse */
    Elu.className = 'elu';
    registerClass(Elu);
    /**
     * Scaled Exponential Linear Unit. (Klambauer et al., 2017).
     * Reference: Self-Normalizing Neural Networks, https://arxiv.org/abs/1706.02515
     * Notes:
     *   - To be used together with the initialization "lecunNormal".
     *   - To be used together with the dropout variant "AlphaDropout".
     */
    class Selu extends Activation$1 {
        apply(x) {
            return selu$2(x);
        }
    }
    /** @nocollapse */
    Selu.className = 'selu';
    registerClass(Selu);
    /**
     *  Rectified linear unit
     */
    class Relu extends Activation$1 {
        apply(x) {
            return relu$2(x);
        }
    }
    /** @nocollapse */
    Relu.className = 'relu';
    registerClass(Relu);
    /**
     * Rectified linear unit activation maxing out at 6.0.
     */
    class Relu6 extends Activation$1 {
        apply(x) {
            return tidy(() => minimum$3(6.0, relu$2(x)));
        }
    }
    /** @nocollapse */
    Relu6.className = 'relu6';
    registerClass(Relu6);
    //* Linear activation (no-op) */
    class Linear extends Activation$1 {
        apply(x) {
            return x;
        }
    }
    /** @nocollapse */
    Linear.className = 'linear';
    registerClass(Linear);
    /**
     * Sigmoid activation function.
     */
    class Sigmoid extends Activation$1 {
        apply(x) {
            return sigmoid$2(x);
        }
    }
    /** @nocollapse */
    Sigmoid.className = 'sigmoid';
    registerClass(Sigmoid);
    /**
     * Segment-wise linear approximation of sigmoid.
     */
    class HardSigmoid extends Activation$1 {
        apply(x) {
            return hardSigmoid(x);
        }
    }
    /** @nocollapse */
    HardSigmoid.className = 'hardSigmoid';
    registerClass(HardSigmoid);
    /**
     * Softplus activation function.
     */
    class Softplus extends Activation$1 {
        apply(x) {
            return softplus$2(x);
        }
    }
    /** @nocollapse */
    Softplus.className = 'softplus';
    registerClass(Softplus);
    /**
     * Softsign activation function.
     */
    class Softsign extends Activation$1 {
        apply(x) {
            return softsign(x);
        }
    }
    /** @nocollapse */
    Softsign.className = 'softsign';
    registerClass(Softsign);
    /**
     * Hyperbolic tangent function.
     */
    class Tanh extends Activation$1 {
        apply(x) {
            return tanh$2(x);
        }
    }
    /** @nocollapse */
    Tanh.className = 'tanh';
    registerClass(Tanh);
    /**
     * Softmax activation function
     */
    class Softmax$1 extends Activation$1 {
        /**
         * Calculate the activation function.
         *
         * @param x Tensor.
         * @param axis Integer, axis along which the softmax normalization is applied.
         * Invalid if < 2, as softmax across 1 (the batch dimension) is assumed to be
         * an error.
         *
         * @returns a Tensor of the same shape as x
         *
         * @throws ValueError: In case `dim(x) < 2`.
         */
        apply(x, axis = (-1)) {
            return softmax$3(x, axis);
        }
    }
    /** @nocollapse */
    Softmax$1.className = 'softmax';
    registerClass(Softmax$1);
    /**
     * Log softmax activation function
     */
    class LogSoftmax extends Activation$1 {
        /**
         * Calculate the activation function of log softmax:
         * log( exp(x_i) / sum(exp(x)) )
         *
         * @param x Tensor.
         * @param axis Integer, axis along which the softmax normalization is applied.
         * Invalid if < 2, as softmax across 1 (the batch dimension) is assumed to be
         * an error.
         *
         * @returns a Tensor of the same shape as x
         *
         * @throws ValueError: In case `dim(x) < 2`.
         */
        apply(x, axis = (-1)) {
            return logSoftmax(x, axis);
        }
    }
    /** @nocollapse */
    LogSoftmax.className = 'logSoftmax';
    registerClass(LogSoftmax);
    /**
     * Swish activation function
     */
    class Swish extends Activation$1 {
        /**
         * Calculate the activation function.
         *
         * @param x Tensor.
         * @param alpha Scaling factor for the sigmoid function.
         * @returns a Tensor of the same shape as x
         */
        apply(x, alpha = 1) {
            return tidy(() => mul(sigmoid$2(mul(x, alpha)), x));
        }
    }
    /** @nocollapse */
    Swish.className = 'swish';
    registerClass(Swish);
    /**
     * Mish activation function
     */
    class Mish extends Activation$1 {
        /**
         * Calculate the activation function.
         *
         * @param x Tensor.
         * @returns a Tensor of the same shape as x
         */
        apply(x) {
            return tidy(() => mul(x, tanh$2(softplus$2(x))));
        }
    }
    /** @nocollapse */
    Mish.className = 'mish';
    registerClass(Mish);
    function serializeActivation(activation) {
        return activation.getClassName();
    }
    function deserializeActivation(config, customObjects = {}) {
        return deserializeKerasObject(config, SerializationMap.getMap().classNameMap, customObjects, 'activation');
    }
    function getActivation(identifier) {
        if (identifier == null) {
            const config = {};
            config['className'] = 'linear';
            config['config'] = {};
            return deserializeActivation(config);
        }
        if (typeof identifier === 'string') {
            const config = {};
            config['className'] = identifier;
            config['config'] = {};
            return deserializeActivation(config);
        }
        else if (identifier instanceof Activation$1) {
            return identifier;
        }
        else {
            return deserializeActivation(identifier);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    function assertObjectArgs(args) {
        if (args != null && typeof args !== 'object') {
            throw new Error(`Argument to L1L2 regularizer's constructor is expected to be an ` +
                `object, but received: ${args}`);
        }
    }
    /**
     * Regularizer base class.
     */
    class Regularizer extends Serializable {
    }
    class L1L2 extends Regularizer {
        constructor(args) {
            super();
            assertObjectArgs(args);
            this.l1 = args == null || args.l1 == null ? 0.01 : args.l1;
            this.l2 = args == null || args.l2 == null ? 0.01 : args.l2;
            this.hasL1 = this.l1 !== 0;
            this.hasL2 = this.l2 !== 0;
        }
        /**
         * Porting note: Renamed from __call__.
         * @param x Variable of which to calculate the regularization score.
         */
        apply(x) {
            return tidy(() => {
                let regularization = zeros$2([1]);
                if (this.hasL1) {
                    regularization = add$2(regularization, sum$2(mul(this.l1, abs$2(x))));
                }
                if (this.hasL2) {
                    regularization =
                        add$2(regularization, sum$2(mul(this.l2, square$1(x))));
                }
                return reshape$3(regularization, []);
            });
        }
        getConfig() {
            return { 'l1': this.l1, 'l2': this.l2 };
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls({ l1: config['l1'], l2: config['l2'] });
        }
    }
    /** @nocollapse */
    L1L2.className = 'L1L2';
    registerClass(L1L2);
    function l1$1(args) {
        assertObjectArgs(args);
        return new L1L2({ l1: args != null ? args.l1 : null, l2: 0 });
    }
    function l2$1(args) {
        assertObjectArgs(args);
        return new L1L2({ l2: args != null ? args.l2 : null, l1: 0 });
    }
    // Maps the JavaScript-like identifier keys to the corresponding keras symbols.
    const REGULARIZER_IDENTIFIER_REGISTRY_SYMBOL_MAP = {
        'l1l2': 'L1L2'
    };
    function serializeRegularizer(constraint) {
        return serializeKerasObject(constraint);
    }
    function deserializeRegularizer(config, customObjects = {}) {
        return deserializeKerasObject(config, SerializationMap.getMap().classNameMap, customObjects, 'regularizer');
    }
    function getRegularizer(identifier) {
        if (identifier == null) {
            return null;
        }
        if (typeof identifier === 'string') {
            const className = identifier in REGULARIZER_IDENTIFIER_REGISTRY_SYMBOL_MAP ?
                REGULARIZER_IDENTIFIER_REGISTRY_SYMBOL_MAP[identifier] :
                identifier;
            const config = { className, config: {} };
            return deserializeRegularizer(config);
        }
        else if (identifier instanceof Regularizer) {
            return identifier;
        }
        else {
            return deserializeRegularizer(identifier);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    class ReLU extends Layer {
        constructor(args) {
            super(args == null ? {} : args);
            this.supportsMasking = true;
            if (args != null) {
                this.maxValue = args.maxValue;
            }
        }
        call(inputs, kwargs) {
            inputs = getExactlyOneTensor(inputs);
            let output = relu$2(inputs);
            if (this.maxValue != null) {
                output = clipByValue$2(output, 0, this.maxValue);
            }
            return output;
        }
        computeOutputShape(inputShape) {
            return inputShape;
        }
        getConfig() {
            const config = { maxValue: this.maxValue };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    ReLU.className = 'ReLU';
    registerClass(ReLU);
    class LeakyReLU extends Layer {
        constructor(args) {
            super(args == null ? {} : args);
            this.DEFAULT_ALPHA = 0.3;
            if (args == null) {
                args = {};
            }
            this.alpha = args.alpha == null ? this.DEFAULT_ALPHA : args.alpha;
        }
        call(inputs, kwargs) {
            const x = getExactlyOneTensor(inputs);
            return leakyRelu$2(x, this.alpha);
        }
        computeOutputShape(inputShape) {
            return inputShape;
        }
        getConfig() {
            const config = { alpha: this.alpha };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    LeakyReLU.className = 'LeakyReLU';
    registerClass(LeakyReLU);
    class PReLU extends Layer {
        constructor(args) {
            super(args == null ? {} : args);
            this.DEFAULT_ALPHA_INITIALIZER = 'zeros';
            if (args == null) {
                args = {};
            }
            this.supportsMasking = true;
            this.alphaInitializer =
                getInitializer(args.alphaInitializer || this.DEFAULT_ALPHA_INITIALIZER);
            this.alphaRegularizer = getRegularizer(args.alphaRegularizer);
            this.alphaConstraint = getConstraint(args.alphaConstraint);
            if (args.sharedAxes == null) {
                this.sharedAxes = null;
            }
            else if (Array.isArray(args.sharedAxes)) {
                this.sharedAxes = args.sharedAxes;
            }
            else if (typeof args.sharedAxes === 'number') {
                this.sharedAxes = [args.sharedAxes];
            }
            else {
                throw new ValueError(`Expected sharedAxes to be a number or an array of numbers, ` +
                    `but got ${args.sharedAxes}`);
            }
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const paramShape = inputShape.slice(1);
            if (this.sharedAxes != null) {
                for (const i of this.sharedAxes) {
                    paramShape[i - 1] = 1;
                }
            }
            this.alpha = this.addWeight('alpha', paramShape, 'float32', this.alphaInitializer, this.alphaRegularizer, true, this.alphaConstraint);
            // Set input spec.
            const axes = {};
            if (this.sharedAxes != null) {
                for (let i = 1; i < inputShape.length; ++i) {
                    axes[i] = inputShape[i];
                }
            }
            this.inputSpec = [new InputSpec({
                    ndim: inputShape.length,
                    axes,
                })];
            this.built = true;
        }
        call(inputs, kwargs) {
            inputs = getExactlyOneTensor(inputs);
            return prelu$3(inputs, this.alpha.read());
        }
        getConfig() {
            const config = {
                alphaInitializer: serializeInitializer(this.alphaInitializer),
                alphaRegularizer: serializeRegularizer(this.alphaRegularizer),
                alphaConstraint: serializeConstraint(this.alphaConstraint),
                sharedAxes: this.sharedAxes
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    PReLU.className = 'PReLU';
    registerClass(PReLU);
    class ELU$3 extends Layer {
        constructor(args) {
            super(args == null ? {} : args);
            this.DEFAULT_ALPHA = 1.0;
            if (args == null) {
                args = {};
            }
            if (args.alpha != null && args.alpha !== this.DEFAULT_ALPHA) {
                throw new NotImplementedError(`Non-default alpha value (${args.alpha}) is not supported by the ` +
                    `ELU layer yet.`);
            }
            this.alpha = args.alpha == null ? this.DEFAULT_ALPHA : args.alpha;
        }
        call(inputs, kwargs) {
            const x = getExactlyOneTensor(inputs);
            return elu$4(x);
        }
        computeOutputShape(inputShape) {
            return inputShape;
        }
        getConfig() {
            const config = { alpha: this.alpha };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    ELU$3.className = 'ELU';
    registerClass(ELU$3);
    class ThresholdedReLU extends Layer {
        constructor(args) {
            super(args == null ? {} : args);
            this.DEFAULT_THETA = 1.0;
            if (args == null) {
                args = {};
            }
            this.theta = args.theta == null ? this.DEFAULT_THETA : args.theta;
        }
        call(inputs, kwargs) {
            const x = getExactlyOneTensor(inputs);
            return mul(x, cast$3(greater$3(x, this.theta), 'float32'));
        }
        computeOutputShape(inputShape) {
            return inputShape;
        }
        getConfig() {
            const config = { theta: this.theta };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    ThresholdedReLU.className = 'ThresholdedReLU';
    registerClass(ThresholdedReLU);
    class Softmax extends Layer {
        constructor(args) {
            super(args == null ? {} : args);
            this.DEFAULT_AXIS = 1.0;
            if (args == null) {
                args = {};
            }
            this.softmax = new Softmax$1().apply;
            this.axis = args.axis == null ? this.DEFAULT_AXIS : args.axis;
        }
        call(inputs, kwargs) {
            const x = getExactlyOneTensor(inputs);
            return this.softmax(x, this.axis);
        }
        computeOutputShape(inputShape) {
            return inputShape;
        }
        getConfig() {
            const config = { axis: this.axis };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Softmax.className = 'Softmax';
    registerClass(Softmax);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Transforms a single number of array of numbers into an array of numbers.
     * @param value
     * @param n: The size of the tuple to be returned.
     * @param name: Name of the parameter, used for generating error messages.
     * @returns An array of numbers.
     */
    function normalizeArray$1(value, n, name) {
        if (typeof value === 'number') {
            return pyListRepeat(value, n);
        }
        else {
            if (value.length !== n) {
                throw new ValueError(`The ${name} argument must be an integer or tuple of ${n} integers.` +
                    ` Received: ${value.length} elements.`);
            }
            for (let i = 0; i < n; ++i) {
                const singleValue = value[i];
                if (!isInteger(singleValue)) {
                    throw new ValueError(`The ${name} argument must be an integer or tuple of ${n}` +
                        ` integers. Received: ${JSON.stringify(value)} including a` +
                        ` non-integer number ${singleValue}`);
                }
            }
            return value;
        }
    }
    /**
     * Determines output length of a convolution given input length.
     * @param inputLength
     * @param filterSize
     * @param padding
     * @param stride
     * @param dilation: dilation rate.
     */
    function convOutputLength(inputLength, filterSize, padding, stride, dilation = 1) {
        if (inputLength == null) {
            return inputLength;
        }
        const dilatedFilterSize = filterSize + (filterSize - 1) * (dilation - 1);
        let outputLength;
        if (padding === 'same') {
            outputLength = inputLength;
        }
        else { // VALID
            outputLength = inputLength - dilatedFilterSize + 1;
        }
        return Math.floor((outputLength + stride - 1) / stride);
    }
    function deconvLength(dimSize, strideSize, kernelSize, padding) {
        if (dimSize == null) {
            return null;
        }
        if (padding === 'valid') {
            dimSize = dimSize * strideSize + max$2([kernelSize - strideSize, 0]);
        }
        else if (padding === 'same') {
            dimSize = dimSize * strideSize;
        }
        else {
            throw new ValueError(`Unsupport padding mode: ${padding}.`);
        }
        return dimSize;
    }

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Transpose and cast the input before the conv2d.
     * @param x Input image tensor.
     * @param dataFormat
     */
    function preprocessConv2DInput(x, dataFormat) {
        // TODO(cais): Cast type to float32 if not.
        return tidy(() => {
            checkDataFormat(dataFormat);
            if (dataFormat === 'channelsFirst') {
                return transpose$2(x, [0, 2, 3, 1]); // NCHW -> NHWC.
            }
            else {
                return x;
            }
        });
    }
    /**
     * Transpose and cast the input before the conv3d.
     * @param x Input image tensor.
     * @param dataFormat
     */
    function preprocessConv3DInput(x, dataFormat) {
        return tidy(() => {
            checkDataFormat(dataFormat);
            if (dataFormat === 'channelsFirst') {
                return transpose$2(x, [0, 2, 3, 4, 1]); // NCDHW -> NDHWC.
            }
            else {
                return x;
            }
        });
    }
    /**
     * 1D-convolution with bias added.
     *
     * Porting Note: This function does not exist in the Python Keras backend.
     *   It is exactly the same as `conv2d`, except the added `bias`.
     *
     * @param x Input tensor, rank-3, of shape `[batchSize, width, inChannels]`.
     * @param kernel Kernel, rank-3, of shape `[filterWidth, inDepth, outDepth]`.
     * @param bias Bias, rank-3, of shape `[outDepth]`.
     * @param strides
     * @param padding Padding mode.
     * @param dataFormat Data format.
     * @param dilationRate
     * @returns The result of the 1D convolution.
     * @throws ValueError, if `x`, `kernel` or `bias` is not of the correct rank.
     */
    function conv1dWithBias(x, kernel, bias, strides = 1, padding = 'valid', dataFormat, dilationRate = 1) {
        return tidy(() => {
            if (dataFormat == null) {
                dataFormat = imageDataFormat();
            }
            checkDataFormat(dataFormat);
            // Check the ranks of x, kernel and bias.
            if (x.shape.length !== 3) {
                throw new ValueError(`The input of a conv1dWithBias operation should be 3, but is ` +
                    `${x.shape.length} instead.`);
            }
            if (kernel.shape.length !== 3) {
                throw new ValueError(`The kernel for a conv1dWithBias operation should be 3, but is ` +
                    `${kernel.shape.length} instead`);
            }
            if (bias != null && bias.shape.length !== 1) {
                throw new ValueError(`The bias for a conv1dWithBias operation should be 1, but is ` +
                    `${kernel.shape.length} instead`);
            }
            // TODO(cais): Support CAUSAL padding mode.
            if (dataFormat === 'channelsFirst') {
                x = transpose$2(x, [0, 2, 1]); // NCW -> NWC.
            }
            if (padding === 'causal') {
                throw new NotImplementedError('The support for CAUSAL padding mode in conv1dWithBias is not ' +
                    'implemented yet.');
            }
            let y = conv1d$1(x, kernel, strides, padding === 'same' ? 'same' : 'valid', 'NWC', dilationRate);
            if (bias != null) {
                y = biasAdd(y, bias);
            }
            return y;
        });
    }
    /**
     * 2D Convolution with an added bias and optional activation.
     * Note: This function does not exist in the Python Keras Backend. This function
     * is exactly the same as `conv2d`, except the added `bias`.
     */
    function conv2dWithBiasActivation(x, kernel, bias, strides = [1, 1], padding = 'valid', dataFormat, dilationRate, activation = null) {
        return tidy(() => {
            if (dataFormat == null) {
                dataFormat = imageDataFormat();
            }
            checkDataFormat(dataFormat);
            if (x.rank !== 3 && x.rank !== 4) {
                throw new ValueError(`conv2dWithBiasActivation expects input to be of rank 3 or 4, ` +
                    `but received ${x.rank}.`);
            }
            if (kernel.rank !== 3 && kernel.rank !== 4) {
                throw new ValueError(`conv2dWithBiasActivation expects kernel to be of rank 3 or 4, ` +
                    `but received ${x.rank}.`);
            }
            let y = preprocessConv2DInput(x, dataFormat);
            if (padding === 'causal') {
                throw new NotImplementedError('The support for CAUSAL padding mode in conv1dWithBias is not ' +
                    'implemented yet.');
            }
            y = conv2d$2({
                x: y,
                filter: kernel,
                strides: strides,
                pad: padding === 'same' ? 'same' : 'valid',
                dilations: dilationRate,
                dataFormat: 'NHWC',
                bias,
                activation
            });
            if (dataFormat === 'channelsFirst') {
                y = transpose$2(y, [0, 3, 1, 2]);
            }
            return y;
        });
    }
    /**
     * 3D Convolution with an added bias.
     * Note: This function does not exist in the Python Keras Backend. This function
     * is exactly the same as `conv3d`, except the added `bias`.
     */
    function conv3dWithBias(x, kernel, bias, strides = [1, 1, 1], padding = 'valid', dataFormat, dilationRate) {
        return tidy(() => {
            if (dataFormat == null) {
                dataFormat = imageDataFormat();
            }
            checkDataFormat(dataFormat);
            if (x.rank !== 4 && x.rank !== 5) {
                throw new ValueError(`conv3dWithBias expects input to be of rank 4 or 5, but received ` +
                    `${x.rank}.`);
            }
            if (kernel.rank !== 4 && kernel.rank !== 5) {
                throw new ValueError(`conv3dWithBias expects kernel to be of rank 4 or 5, but received ` +
                    `${x.rank}.`);
            }
            let y = preprocessConv3DInput(x, dataFormat);
            if (padding === 'causal') {
                throw new NotImplementedError('The support for CAUSAL padding mode in conv3dWithBias is not ' +
                    'implemented yet.');
            }
            y = conv3d$1(y, kernel, strides, padding === 'same' ? 'same' : 'valid', 'NDHWC', dilationRate);
            if (bias != null) {
                y = biasAdd(y, bias);
            }
            if (dataFormat === 'channelsFirst') {
                y = transpose$2(y, [0, 4, 1, 2, 3]);
            }
            return y;
        });
    }
    /**
     * Abstract convolution layer.
     */
    class BaseConv extends Layer {
        constructor(rank, args) {
            super(args);
            this.bias = null;
            this.DEFAULT_KERNEL_INITIALIZER = 'glorotNormal';
            this.DEFAULT_BIAS_INITIALIZER = 'zeros';
            BaseConv.verifyArgs(args);
            this.rank = rank;
            assertPositiveInteger(this.rank, 'rank');
            if (this.rank !== 1 && this.rank !== 2 && this.rank !== 3) {
                throw new NotImplementedError(`Convolution layer for rank other than 1, 2, or 3 (${this.rank}) is ` +
                    `not implemented yet.`);
            }
            this.kernelSize = normalizeArray$1(args.kernelSize, rank, 'kernelSize');
            this.strides = normalizeArray$1(args.strides == null ? 1 : args.strides, rank, 'strides');
            this.padding = args.padding == null ? 'valid' : args.padding;
            checkPaddingMode(this.padding);
            this.dataFormat =
                args.dataFormat == null ? 'channelsLast' : args.dataFormat;
            checkDataFormat(this.dataFormat);
            this.activation = getActivation(args.activation);
            this.useBias = args.useBias == null ? true : args.useBias;
            this.biasInitializer =
                getInitializer(args.biasInitializer || this.DEFAULT_BIAS_INITIALIZER);
            this.biasConstraint = getConstraint(args.biasConstraint);
            this.biasRegularizer = getRegularizer(args.biasRegularizer);
            this.activityRegularizer = getRegularizer(args.activityRegularizer);
            this.dilationRate = normalizeArray$1(args.dilationRate == null ? 1 : args.dilationRate, rank, 'dilationRate');
            if (this.rank === 1 &&
                (Array.isArray(this.dilationRate) && this.dilationRate.length !== 1)) {
                throw new ValueError(`dilationRate must be a number or an array of a single number ` +
                    `for 1D convolution, but received ` +
                    `${JSON.stringify(this.dilationRate)}`);
            }
            else if (this.rank === 2) {
                if (typeof this.dilationRate === 'number') {
                    this.dilationRate = [this.dilationRate, this.dilationRate];
                }
                else if (this.dilationRate.length !== 2) {
                    throw new ValueError(`dilationRate must be a number or array of two numbers for 2D ` +
                        `convolution, but received ${JSON.stringify(this.dilationRate)}`);
                }
            }
            else if (this.rank === 3) {
                if (typeof this.dilationRate === 'number') {
                    this.dilationRate =
                        [this.dilationRate, this.dilationRate, this.dilationRate];
                }
                else if (this.dilationRate.length !== 3) {
                    throw new ValueError(`dilationRate must be a number or array of three numbers for 3D ` +
                        `convolution, but received ${JSON.stringify(this.dilationRate)}`);
                }
            }
        }
        static verifyArgs(args) {
            // Check config.kernelSize type and shape.
            assert$3('kernelSize' in args, `required key 'kernelSize' not in config`);
            if (typeof args.kernelSize !== 'number' &&
                !checkArrayTypeAndLength(args.kernelSize, 'number', 1, 3)) {
                throw new ValueError(`BaseConv expects config.kernelSize to be number or number[] with ` +
                    `length 1, 2, or 3, but received ${JSON.stringify(args.kernelSize)}.`);
            }
        }
        getConfig() {
            const config = {
                kernelSize: this.kernelSize,
                strides: this.strides,
                padding: this.padding,
                dataFormat: this.dataFormat,
                dilationRate: this.dilationRate,
                activation: serializeActivation(this.activation),
                useBias: this.useBias,
                biasInitializer: serializeInitializer(this.biasInitializer),
                biasRegularizer: serializeRegularizer(this.biasRegularizer),
                activityRegularizer: serializeRegularizer(this.activityRegularizer),
                biasConstraint: serializeConstraint(this.biasConstraint)
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /**
     * Abstract nD convolution layer.  Ancestor of convolution layers which reduce
     * across channels, i.e., Conv1D and Conv2D, but not DepthwiseConv2D.
     */
    class Conv extends BaseConv {
        constructor(rank, args) {
            super(rank, args);
            this.kernel = null;
            Conv.verifyArgs(args);
            this.filters = args.filters;
            assertPositiveInteger(this.filters, 'filters');
            this.kernelInitializer = getInitializer(args.kernelInitializer || this.DEFAULT_KERNEL_INITIALIZER);
            this.kernelConstraint = getConstraint(args.kernelConstraint);
            this.kernelRegularizer = getRegularizer(args.kernelRegularizer);
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const channelAxis = this.dataFormat === 'channelsFirst' ? 1 : inputShape.length - 1;
            if (inputShape[channelAxis] == null) {
                throw new ValueError(`The channel dimension of the input should be defined. ` +
                    `Found ${inputShape[channelAxis]}`);
            }
            const inputDim = inputShape[channelAxis];
            const kernelShape = this.kernelSize.concat([inputDim, this.filters]);
            this.kernel = this.addWeight('kernel', kernelShape, null, this.kernelInitializer, this.kernelRegularizer, true, this.kernelConstraint);
            if (this.useBias) {
                this.bias = this.addWeight('bias', [this.filters], null, this.biasInitializer, this.biasRegularizer, true, this.biasConstraint);
            }
            this.inputSpec = [{ ndim: this.rank + 2, axes: { [channelAxis]: inputDim } }];
            this.built = true;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = getExactlyOneTensor(inputs);
                let outputs;
                const biasValue = this.bias == null ? null : this.bias.read();
                const fusedActivationName = mapActivationToFusedKernel(this.activation.getClassName());
                if (fusedActivationName != null && this.rank === 2) {
                    outputs = conv2dWithBiasActivation(inputs, this.kernel.read(), biasValue, this.strides, this.padding, this.dataFormat, this.dilationRate, fusedActivationName);
                }
                else {
                    if (this.rank === 1) {
                        outputs = conv1dWithBias(inputs, this.kernel.read(), biasValue, this.strides[0], this.padding, this.dataFormat, this.dilationRate[0]);
                    }
                    else if (this.rank === 2) {
                        // TODO(cais): Move up to constructor.
                        outputs = conv2dWithBiasActivation(inputs, this.kernel.read(), biasValue, this.strides, this.padding, this.dataFormat, this.dilationRate);
                    }
                    else if (this.rank === 3) {
                        outputs = conv3dWithBias(inputs, this.kernel.read(), biasValue, this.strides, this.padding, this.dataFormat, this.dilationRate);
                    }
                    else {
                        throw new NotImplementedError('convolutions greater than 3D are not implemented yet.');
                    }
                    if (this.activation != null) {
                        outputs = this.activation.apply(outputs);
                    }
                }
                return outputs;
            });
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const newSpace = [];
            const space = (this.dataFormat === 'channelsLast') ?
                inputShape.slice(1, inputShape.length - 1) :
                inputShape.slice(2);
            for (let i = 0; i < space.length; ++i) {
                const newDim = convOutputLength(space[i], this.kernelSize[i], this.padding, this.strides[i], typeof this.dilationRate === 'number' ? this.dilationRate :
                    this.dilationRate[i]);
                newSpace.push(newDim);
            }
            let outputShape = [inputShape[0]];
            if (this.dataFormat === 'channelsLast') {
                outputShape = outputShape.concat(newSpace);
                outputShape.push(this.filters);
            }
            else {
                outputShape.push(this.filters);
                outputShape = outputShape.concat(newSpace);
            }
            return outputShape;
        }
        getConfig() {
            const config = {
                filters: this.filters,
                kernelInitializer: serializeInitializer(this.kernelInitializer),
                kernelRegularizer: serializeRegularizer(this.kernelRegularizer),
                kernelConstraint: serializeConstraint(this.kernelConstraint)
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
        static verifyArgs(args) {
            // Check config.filters type, shape, and value.
            if (!('filters' in args) || typeof args.filters !== 'number' ||
                args.filters < 1) {
                throw new ValueError(`Convolution layer expected config.filters to be a 'number' > 0 ` +
                    `but got ${JSON.stringify(args.filters)}`);
            }
        }
    }
    class Conv2D extends Conv {
        constructor(args) {
            super(2, args);
            Conv2D.verifyArgs(args);
        }
        getConfig() {
            const config = super.getConfig();
            delete config['rank'];
            return config;
        }
        static verifyArgs(args) {
            // config.kernelSize must be a number or array of numbers.
            if ((typeof args.kernelSize !== 'number') &&
                !checkArrayTypeAndLength(args.kernelSize, 'number', 1, 2)) {
                throw new ValueError(`Conv2D expects config.kernelSize to be number or number[] with ` +
                    `length 1 or 2, but received ${JSON.stringify(args.kernelSize)}.`);
            }
        }
    }
    /** @nocollapse */
    Conv2D.className = 'Conv2D';
    registerClass(Conv2D);
    class Conv3D extends Conv {
        constructor(args) {
            super(3, args);
            Conv3D.verifyArgs(args);
        }
        getConfig() {
            const config = super.getConfig();
            delete config['rank'];
            return config;
        }
        static verifyArgs(args) {
            // config.kernelSize must be a number or array of numbers.
            if (typeof args.kernelSize !== 'number') {
                if (!(Array.isArray(args.kernelSize) &&
                    (args.kernelSize.length === 1 || args.kernelSize.length === 3))) {
                    throw new ValueError(`Conv3D expects config.kernelSize to be number or` +
                        ` [number, number, number], but received ${JSON.stringify(args.kernelSize)}.`);
                }
            }
        }
    }
    /** @nocollapse */
    Conv3D.className = 'Conv3D';
    registerClass(Conv3D);
    class Conv2DTranspose extends Conv2D {
        constructor(args) {
            super(args);
            this.inputSpec = [new InputSpec({ ndim: 4 })];
            if (this.padding !== 'same' && this.padding !== 'valid') {
                throw new ValueError(`Conv2DTranspose currently supports only padding modes 'same' ` +
                    `and 'valid', but received padding mode ${this.padding}`);
            }
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            if (inputShape.length !== 4) {
                throw new ValueError('Input should have rank 4; Received input shape: ' +
                    JSON.stringify(inputShape));
            }
            const channelAxis = this.dataFormat === 'channelsFirst' ? 1 : inputShape.length - 1;
            if (inputShape[channelAxis] == null) {
                throw new ValueError('The channel dimension of the inputs should be defined. ' +
                    'Found `None`.');
            }
            const inputDim = inputShape[channelAxis];
            const kernelShape = this.kernelSize.concat([this.filters, inputDim]);
            this.kernel = this.addWeight('kernel', kernelShape, 'float32', this.kernelInitializer, this.kernelRegularizer, true, this.kernelConstraint);
            if (this.useBias) {
                this.bias = this.addWeight('bias', [this.filters], 'float32', this.biasInitializer, this.biasRegularizer, true, this.biasConstraint);
            }
            // Set input spec.
            this.inputSpec =
                [new InputSpec({ ndim: 4, axes: { [channelAxis]: inputDim } })];
            this.built = true;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                let input = getExactlyOneTensor(inputs);
                if (input.shape.length !== 4) {
                    throw new ValueError(`Conv2DTranspose.call() expects input tensor to be rank-4, but ` +
                        `received a tensor of rank-${input.shape.length}`);
                }
                const inputShape = input.shape;
                const batchSize = inputShape[0];
                let hAxis;
                let wAxis;
                if (this.dataFormat === 'channelsFirst') {
                    hAxis = 2;
                    wAxis = 3;
                }
                else {
                    hAxis = 1;
                    wAxis = 2;
                }
                const height = inputShape[hAxis];
                const width = inputShape[wAxis];
                const kernelH = this.kernelSize[0];
                const kernelW = this.kernelSize[1];
                const strideH = this.strides[0];
                const strideW = this.strides[1];
                // Infer the dynamic output shape.
                const outHeight = deconvLength(height, strideH, kernelH, this.padding);
                const outWidth = deconvLength(width, strideW, kernelW, this.padding);
                // Porting Note: We don't branch based on `this.dataFormat` here,
                // because
                //   the tjfs-core function `conv2dTranspose` called below always
                //   assumes channelsLast.
                const outputShape = [batchSize, outHeight, outWidth, this.filters];
                if (this.dataFormat !== 'channelsLast') {
                    input = transpose$2(input, [0, 2, 3, 1]);
                }
                let outputs = conv2dTranspose$1(input, this.kernel.read(), outputShape, this.strides, this.padding);
                if (this.dataFormat !== 'channelsLast') {
                    outputs = transpose$2(outputs, [0, 3, 1, 2]);
                }
                if (this.bias != null) {
                    outputs =
                        biasAdd(outputs, this.bias.read(), this.dataFormat);
                }
                if (this.activation != null) {
                    outputs = this.activation.apply(outputs);
                }
                return outputs;
            });
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const outputShape = inputShape.slice();
            let channelAxis;
            let heightAxis;
            let widthAxis;
            if (this.dataFormat === 'channelsFirst') {
                channelAxis = 1;
                heightAxis = 2;
                widthAxis = 3;
            }
            else {
                channelAxis = 3;
                heightAxis = 1;
                widthAxis = 2;
            }
            const kernelH = this.kernelSize[0];
            const kernelW = this.kernelSize[1];
            const strideH = this.strides[0];
            const strideW = this.strides[1];
            outputShape[channelAxis] = this.filters;
            outputShape[heightAxis] =
                deconvLength(outputShape[heightAxis], strideH, kernelH, this.padding);
            outputShape[widthAxis] =
                deconvLength(outputShape[widthAxis], strideW, kernelW, this.padding);
            return outputShape;
        }
        getConfig() {
            const config = super.getConfig();
            delete config['dilationRate'];
            return config;
        }
    }
    /** @nocollapse */
    Conv2DTranspose.className = 'Conv2DTranspose';
    registerClass(Conv2DTranspose);
    class Conv3DTranspose extends Conv3D {
        constructor(args) {
            super(args);
            this.inputSpec = [new InputSpec({ ndim: 5 })];
            if (this.padding !== 'same' && this.padding !== 'valid') {
                throw new ValueError(`Conv3DTranspose currently supports only padding modes 'same' ` +
                    `and 'valid', but received padding mode ${this.padding}`);
            }
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            if (inputShape.length !== 5) {
                throw new ValueError('Input should have rank 5; Received input shape: ' +
                    JSON.stringify(inputShape));
            }
            const channelAxis = this.dataFormat === 'channelsFirst' ? 1 : inputShape.length - 1;
            if (inputShape[channelAxis] == null) {
                throw new ValueError('The channel dimension of the inputs should be defined. ' +
                    'Found `None`.');
            }
            const inputDim = inputShape[channelAxis];
            const kernelShape = this.kernelSize.concat([this.filters, inputDim]);
            this.kernel = this.addWeight('kernel', kernelShape, 'float32', this.kernelInitializer, this.kernelRegularizer, true, this.kernelConstraint);
            if (this.useBias) {
                this.bias = this.addWeight('bias', [this.filters], 'float32', this.biasInitializer, this.biasRegularizer, true, this.biasConstraint);
            }
            // Set input spec.
            this.inputSpec =
                [new InputSpec({ ndim: 5, axes: { [channelAxis]: inputDim } })];
            this.built = true;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                let input = getExactlyOneTensor(inputs);
                if (input.shape.length !== 5) {
                    throw new ValueError(`Conv3DTranspose.call() expects input tensor to be rank-4, but ` +
                        `received a tensor of rank-${input.shape.length}`);
                }
                const inputShape = input.shape;
                const batchSize = inputShape[0];
                let hAxis;
                let wAxis;
                let dAxis;
                if (this.dataFormat === 'channelsFirst') {
                    dAxis = 2;
                    hAxis = 3;
                    wAxis = 4;
                }
                else {
                    dAxis = 1;
                    hAxis = 2;
                    wAxis = 3;
                }
                const depth = inputShape[dAxis];
                const height = inputShape[hAxis];
                const width = inputShape[wAxis];
                const kernelD = this.kernelSize[0];
                const kernelH = this.kernelSize[1];
                const kernelW = this.kernelSize[2];
                const strideD = this.strides[0];
                const strideH = this.strides[1];
                const strideW = this.strides[2];
                // Infer the dynamic output shape.
                const outDepth = deconvLength(depth, strideD, kernelD, this.padding);
                const outHeight = deconvLength(height, strideH, kernelH, this.padding);
                const outWidth = deconvLength(width, strideW, kernelW, this.padding);
                // Same as `conv2dTranspose`. We always assumes channelsLast.
                const outputShape = [batchSize, outDepth, outHeight, outWidth, this.filters];
                if (this.dataFormat !== 'channelsLast') {
                    input = transpose$2(input, [0, 2, 3, 4, 1]);
                }
                let outputs = conv3dTranspose$1(input, this.kernel.read(), outputShape, this.strides, this.padding);
                if (this.dataFormat !== 'channelsLast') {
                    outputs = transpose$2(outputs, [0, 4, 1, 2, 3]);
                }
                if (this.bias !== null) {
                    outputs =
                        biasAdd(outputs, this.bias.read(), this.dataFormat);
                }
                if (this.activation !== null) {
                    outputs = this.activation.apply(outputs);
                }
                return outputs;
            });
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const outputShape = inputShape.slice();
            let channelAxis;
            let depthAxis;
            let heightAxis;
            let widthAxis;
            if (this.dataFormat === 'channelsFirst') {
                channelAxis = 1;
                depthAxis = 2;
                heightAxis = 3;
                widthAxis = 4;
            }
            else {
                channelAxis = 4;
                depthAxis = 1;
                heightAxis = 2;
                widthAxis = 3;
            }
            const kernelD = this.kernelSize[0];
            const kernelH = this.kernelSize[1];
            const kernelW = this.kernelSize[2];
            const strideD = this.strides[0];
            const strideH = this.strides[1];
            const strideW = this.strides[2];
            outputShape[channelAxis] = this.filters;
            outputShape[depthAxis] =
                deconvLength(outputShape[depthAxis], strideD, kernelD, this.padding);
            outputShape[heightAxis] =
                deconvLength(outputShape[heightAxis], strideH, kernelH, this.padding);
            outputShape[widthAxis] =
                deconvLength(outputShape[widthAxis], strideW, kernelW, this.padding);
            return outputShape;
        }
        getConfig() {
            const config = super.getConfig();
            delete config['dilationRate'];
            return config;
        }
    }
    /** @nocollapse */
    Conv3DTranspose.className = 'Conv3DTranspose';
    registerClass(Conv3DTranspose);
    class SeparableConv extends Conv {
        constructor(rank, config) {
            super(rank, config);
            this.DEFAULT_DEPTHWISE_INITIALIZER = 'glorotUniform';
            this.DEFAULT_POINTWISE_INITIALIZER = 'glorotUniform';
            this.depthwiseKernel = null;
            this.pointwiseKernel = null;
            if (config.filters == null) {
                throw new ValueError('The `filters` configuration field is required by SeparableConv, ' +
                    'but is unspecified.');
            }
            if (config.kernelInitializer != null || config.kernelRegularizer != null ||
                config.kernelConstraint != null) {
                throw new ValueError('Fields kernelInitializer, kernelRegularizer and kernelConstraint ' +
                    'are invalid for SeparableConv2D. Use depthwiseInitializer, ' +
                    'depthwiseRegularizer, depthwiseConstraint, pointwiseInitializer, ' +
                    'pointwiseRegularizer and pointwiseConstraint instead.');
            }
            if (config.padding != null && config.padding !== 'same' &&
                config.padding !== 'valid') {
                throw new ValueError(`SeparableConv${this.rank}D supports only padding modes: ` +
                    `'same' and 'valid', but received ${JSON.stringify(config.padding)}`);
            }
            this.depthMultiplier =
                config.depthMultiplier == null ? 1 : config.depthMultiplier;
            this.depthwiseInitializer = getInitializer(config.depthwiseInitializer || this.DEFAULT_DEPTHWISE_INITIALIZER);
            this.depthwiseRegularizer = getRegularizer(config.depthwiseRegularizer);
            this.depthwiseConstraint = getConstraint(config.depthwiseConstraint);
            this.pointwiseInitializer = getInitializer(config.depthwiseInitializer || this.DEFAULT_POINTWISE_INITIALIZER);
            this.pointwiseRegularizer = getRegularizer(config.pointwiseRegularizer);
            this.pointwiseConstraint = getConstraint(config.pointwiseConstraint);
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            if (inputShape.length < this.rank + 2) {
                throw new ValueError(`Inputs to SeparableConv${this.rank}D should have rank ` +
                    `${this.rank + 2}, but received input shape: ` +
                    `${JSON.stringify(inputShape)}`);
            }
            const channelAxis = this.dataFormat === 'channelsFirst' ? 1 : inputShape.length - 1;
            if (inputShape[channelAxis] == null || inputShape[channelAxis] < 0) {
                throw new ValueError(`The channel dimension of the inputs should be defined, ` +
                    `but found ${JSON.stringify(inputShape[channelAxis])}`);
            }
            const inputDim = inputShape[channelAxis];
            const depthwiseKernelShape = this.kernelSize.concat([inputDim, this.depthMultiplier]);
            const pointwiseKernelShape = [];
            for (let i = 0; i < this.rank; ++i) {
                pointwiseKernelShape.push(1);
            }
            pointwiseKernelShape.push(inputDim * this.depthMultiplier, this.filters);
            const trainable = true;
            this.depthwiseKernel = this.addWeight('depthwise_kernel', depthwiseKernelShape, 'float32', this.depthwiseInitializer, this.depthwiseRegularizer, trainable, this.depthwiseConstraint);
            this.pointwiseKernel = this.addWeight('pointwise_kernel', pointwiseKernelShape, 'float32', this.pointwiseInitializer, this.pointwiseRegularizer, trainable, this.pointwiseConstraint);
            if (this.useBias) {
                this.bias = this.addWeight('bias', [this.filters], 'float32', this.biasInitializer, this.biasRegularizer, trainable, this.biasConstraint);
            }
            else {
                this.bias = null;
            }
            this.inputSpec =
                [new InputSpec({ ndim: this.rank + 2, axes: { [channelAxis]: inputDim } })];
            this.built = true;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = getExactlyOneTensor(inputs);
                let output;
                if (this.rank === 1) {
                    throw new NotImplementedError('1D separable convolution is not implemented yet.');
                }
                else if (this.rank === 2) {
                    if (this.dataFormat === 'channelsFirst') {
                        inputs = transpose$2(inputs, [0, 2, 3, 1]); // NCHW -> NHWC.
                    }
                    output = separableConv2d$1(inputs, this.depthwiseKernel.read(), this.pointwiseKernel.read(), this.strides, this.padding, this.dilationRate, 'NHWC');
                }
                if (this.useBias) {
                    output = biasAdd(output, this.bias.read(), this.dataFormat);
                }
                if (this.activation != null) {
                    output = this.activation.apply(output);
                }
                if (this.dataFormat === 'channelsFirst') {
                    output = transpose$2(output, [0, 3, 1, 2]); // NHWC -> NCHW.
                }
                return output;
            });
        }
        getConfig() {
            const config = super.getConfig();
            delete config['rank'];
            delete config['kernelInitializer'];
            delete config['kernelRegularizer'];
            delete config['kernelConstraint'];
            config['depthwiseInitializer'] =
                serializeInitializer(this.depthwiseInitializer);
            config['pointwiseInitializer'] =
                serializeInitializer(this.pointwiseInitializer);
            config['depthwiseRegularizer'] =
                serializeRegularizer(this.depthwiseRegularizer);
            config['pointwiseRegularizer'] =
                serializeRegularizer(this.pointwiseRegularizer);
            config['depthwiseConstraint'] =
                serializeConstraint(this.depthwiseConstraint);
            config['pointwiseConstraint'] =
                serializeConstraint(this.pointwiseConstraint);
            return config;
        }
    }
    /** @nocollapse */
    SeparableConv.className = 'SeparableConv';
    class SeparableConv2D extends SeparableConv {
        constructor(args) {
            super(2, args);
        }
    }
    /** @nocollapse */
    SeparableConv2D.className = 'SeparableConv2D';
    registerClass(SeparableConv2D);
    class Conv1D extends Conv {
        constructor(args) {
            super(1, args);
            Conv1D.verifyArgs(args);
            this.inputSpec = [{ ndim: 3 }];
        }
        getConfig() {
            const config = super.getConfig();
            delete config['rank'];
            delete config['dataFormat'];
            return config;
        }
        static verifyArgs(args) {
            // config.kernelSize must be a number or array of numbers.
            if (typeof args.kernelSize !== 'number' &&
                !checkArrayTypeAndLength(args.kernelSize, 'number', 1, 1)) {
                throw new ValueError(`Conv1D expects config.kernelSize to be number or number[] with ` +
                    `length 1, but received ${JSON.stringify(args.kernelSize)}.`);
            }
        }
    }
    /** @nocollapse */
    Conv1D.className = 'Conv1D';
    registerClass(Conv1D);
    class Cropping2D extends Layer {
        constructor(args) {
            super(args);
            if (typeof args.cropping === 'number') {
                this.cropping =
                    [[args.cropping, args.cropping], [args.cropping, args.cropping]];
            }
            else if (typeof args.cropping[0] === 'number') {
                this.cropping = [
                    [args.cropping[0], args.cropping[0]],
                    [args.cropping[1], args.cropping[1]]
                ];
            }
            else {
                this.cropping = args.cropping;
            }
            this.dataFormat =
                args.dataFormat === undefined ? 'channelsLast' : args.dataFormat;
            this.inputSpec = [{ ndim: 4 }];
        }
        computeOutputShape(inputShape) {
            if (this.dataFormat === 'channelsFirst') {
                return [
                    inputShape[0], inputShape[1],
                    inputShape[2] - this.cropping[0][0] - this.cropping[0][1],
                    inputShape[3] - this.cropping[1][0] - this.cropping[1][1]
                ];
            }
            else {
                return [
                    inputShape[0],
                    inputShape[1] - this.cropping[0][0] - this.cropping[0][1],
                    inputShape[2] - this.cropping[1][0] - this.cropping[1][1], inputShape[3]
                ];
            }
        }
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = getExactlyOneTensor(inputs);
                if (this.dataFormat === 'channelsLast') {
                    const hSliced = sliceAlongAxis(inputs, this.cropping[0][0], inputs.shape[1] - this.cropping[0][0] - this.cropping[0][1], 2);
                    return sliceAlongAxis(hSliced, this.cropping[1][0], inputs.shape[2] - this.cropping[1][1] - this.cropping[1][0], 3);
                }
                else {
                    const hSliced = sliceAlongAxis(inputs, this.cropping[0][0], inputs.shape[2] - this.cropping[0][0] - this.cropping[0][1], 3);
                    return sliceAlongAxis(hSliced, this.cropping[1][0], inputs.shape[3] - this.cropping[1][1] - this.cropping[1][0], 4);
                }
            });
        }
        getConfig() {
            const config = { cropping: this.cropping, dataFormat: this.dataFormat };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Cropping2D.className = 'Cropping2D';
    registerClass(Cropping2D);
    class UpSampling2D extends Layer {
        constructor(args) {
            super(args);
            this.DEFAULT_SIZE = [2, 2];
            this.inputSpec = [{ ndim: 4 }];
            this.size = args.size == null ? this.DEFAULT_SIZE : args.size;
            this.dataFormat =
                args.dataFormat == null ? 'channelsLast' : args.dataFormat;
            checkDataFormat(this.dataFormat);
            this.interpolation =
                args.interpolation == null ? 'nearest' : args.interpolation;
            checkInterpolationFormat(this.interpolation);
        }
        computeOutputShape(inputShape) {
            if (this.dataFormat === 'channelsFirst') {
                const height = inputShape[2] == null ? null : this.size[0] * inputShape[2];
                const width = inputShape[3] == null ? null : this.size[1] * inputShape[3];
                return [inputShape[0], inputShape[1], height, width];
            }
            else {
                const height = inputShape[1] == null ? null : this.size[0] * inputShape[1];
                const width = inputShape[2] == null ? null : this.size[1] * inputShape[2];
                return [inputShape[0], height, width, inputShape[3]];
            }
        }
        call(inputs, kwargs) {
            return tidy(() => {
                let input = getExactlyOneTensor(inputs);
                const inputShape = input.shape;
                if (this.dataFormat === 'channelsFirst') {
                    input = transpose$2(input, [0, 2, 3, 1]);
                    const height = this.size[0] * inputShape[2];
                    const width = this.size[1] * inputShape[3];
                    const resized = this.interpolation === 'nearest' ?
                        image$1.resizeNearestNeighbor(input, [height, width]) :
                        image$1.resizeBilinear(input, [height, width]);
                    return transpose$2(resized, [0, 3, 1, 2]);
                }
                else {
                    const height = this.size[0] * inputShape[1];
                    const width = this.size[1] * inputShape[2];
                    return this.interpolation === 'nearest' ?
                        image$1.resizeNearestNeighbor(input, [height, width]) :
                        image$1.resizeBilinear(input, [height, width]);
                }
            });
        }
        getConfig() {
            const config = {
                size: this.size,
                dataFormat: this.dataFormat,
                interpolation: this.interpolation
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    UpSampling2D.className = 'UpSampling2D';
    registerClass(UpSampling2D);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * 2D convolution with separable filters.
     * @param x Input tensor.
     * @param depthwiseKernel Convolution kernel for depthwise convolution.
     * @param strides Strides (Array of two integers).
     * @param padding Padding model.
     * @param dataFormat Data format.
     * @param dilationRate Array of two integers, dilation rates for the separable
     *   convolution.
     * @returns Output tensor.
     * @throws ValueError If depthwiseKernel is not a 4D array.
     */
    function depthwiseConv2d$1(x, depthwiseKernel, strides = [1, 1], padding = 'valid', dataFormat, dilationRate) {
        return tidy(() => {
            if (dataFormat == null) {
                dataFormat = imageDataFormat();
            }
            checkDataFormat(dataFormat);
            let y = preprocessConv2DInput(x, dataFormat);
            if (x.rank !== 4) {
                throw new ValueError(`Input for depthwiseConv2d is required to be 4-D, but is instead ` +
                    `${x.rank}-D`);
            }
            if (depthwiseKernel.rank !== 4) {
                throw new ValueError(`depthwiseKernel is required to be 4-D, but is instead ` +
                    `${depthwiseKernel.rank}-D`);
            }
            y = depthwiseConv2d$3(y, depthwiseKernel, strides, padding === 'same' ? 'same' : 'valid', 'NHWC', dilationRate);
            if (dataFormat === 'channelsFirst') {
                y = transpose$2(y, [0, 3, 1, 2]);
            }
            return y;
        });
    }
    class DepthwiseConv2D extends BaseConv {
        constructor(args) {
            super(2, args);
            this.depthwiseKernel = null;
            this.depthMultiplier =
                args.depthMultiplier == null ? 1 : args.depthMultiplier;
            this.depthwiseInitializer = getInitializer(args.depthwiseInitializer || this.DEFAULT_KERNEL_INITIALIZER);
            this.depthwiseConstraint = getConstraint(args.depthwiseConstraint);
            this.depthwiseRegularizer = getRegularizer(args.depthwiseRegularizer);
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            if (inputShape.length < 4) {
                throw new ValueError(`Inputs to DepthwiseConv2D should have rank 4. ` +
                    `Received input shape: ${JSON.stringify(inputShape)}.`);
            }
            const channelAxis = this.dataFormat === 'channelsFirst' ? 1 : 3;
            if (inputShape[channelAxis] == null || inputShape[channelAxis] < 0) {
                throw new ValueError('The channel dimension of the inputs to DepthwiseConv2D should ' +
                    `be defined, but is not (${inputShape[channelAxis]}).`);
            }
            const inputDim = inputShape[channelAxis];
            const depthwiseKernelShape = [
                this.kernelSize[0], this.kernelSize[1], inputDim, this.depthMultiplier
            ];
            this.depthwiseKernel = this.addWeight('depthwise_kernel', depthwiseKernelShape, null, this.depthwiseInitializer, this.depthwiseRegularizer, true, this.depthwiseConstraint);
            if (this.useBias) {
                this.bias = this.addWeight('bias', [inputDim * this.depthMultiplier], null, this.biasInitializer, this.biasRegularizer, true, this.biasConstraint);
            }
            else {
                this.bias = null;
            }
            this.built = true;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = getExactlyOneTensor(inputs);
                let outputs = depthwiseConv2d$1(inputs, this.depthwiseKernel.read(), this.strides, this.padding, this.dataFormat, null);
                // TODO(cais): Add support for dilation.
                if (this.useBias) {
                    outputs = biasAdd(outputs, this.bias.read(), this.dataFormat);
                }
                if (this.activation != null) {
                    outputs = this.activation.apply(outputs);
                }
                return outputs;
            });
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const rows = this.dataFormat === 'channelsFirst' ? inputShape[2] : inputShape[1];
            const cols = this.dataFormat === 'channelsFirst' ? inputShape[3] : inputShape[2];
            const outFilters = this.dataFormat === 'channelsFirst' ?
                inputShape[1] * this.depthMultiplier :
                inputShape[3] * this.depthMultiplier;
            const outRows = convOutputLength(rows, this.kernelSize[0], this.padding, this.strides[0]);
            const outCols = convOutputLength(cols, this.kernelSize[1], this.padding, this.strides[1]);
            if (this.dataFormat === 'channelsFirst') {
                return [inputShape[0], outFilters, outRows, outCols];
            }
            else {
                // In this case, assume 'channelsLast'.
                return [inputShape[0], outRows, outCols, outFilters];
            }
        }
        getConfig() {
            const config = super.getConfig();
            config['depthMultiplier'] = this.depthMultiplier;
            config['depthwiseInitializer'] =
                serializeInitializer(this.depthwiseInitializer);
            config['depthwiseRegularizer'] =
                serializeRegularizer(this.depthwiseRegularizer);
            config['depthwiseConstraint'] =
                serializeConstraint(this.depthwiseRegularizer);
            return config;
        }
    }
    /** @nocollapse */
    DepthwiseConv2D.className = 'DepthwiseConv2D';
    registerClass(DepthwiseConv2D);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Standardize `apply()` args to a single list of tensor inputs.
     *
     * When running a model loaded from file, the input tensors `initialState` and
     * `constants` are passed to `RNN.apply()` as part of `inputs` instead of the
     * dedicated kwargs fields. `inputs` consists of
     * `[inputs, initialState0, initialState1, ..., constant0, constant1]` in this
     * case.
     * This method makes sure that arguments are
     * separated and that `initialState` and `constants` are `Array`s of tensors
     * (or None).
     *
     * @param inputs Tensor or `Array` of  tensors.
     * @param initialState Tensor or `Array` of tensors or `null`/`undefined`.
     * @param constants Tensor or `Array` of tensors or `null`/`undefined`.
     * @returns An object consisting of
     *   inputs: A tensor.
     *   initialState: `Array` of tensors or `null`.
     *   constants: `Array` of tensors or `null`.
     * @throws ValueError, if `inputs` is an `Array` but either `initialState` or
     *   `constants` is provided.
     */
    function standardizeArgs(inputs, initialState, constants, numConstants) {
        if (Array.isArray(inputs)) {
            if (initialState != null || constants != null) {
                throw new ValueError('When inputs is an array, neither initialState or constants ' +
                    'should be provided');
            }
            if (numConstants != null) {
                constants = inputs.slice(inputs.length - numConstants, inputs.length);
                inputs = inputs.slice(0, inputs.length - numConstants);
            }
            if (inputs.length > 1) {
                initialState = inputs.slice(1, inputs.length);
            }
            inputs = inputs[0];
        }
        function toListOrNull(x) {
            if (x == null || Array.isArray(x)) {
                return x;
            }
            else {
                return [x];
            }
        }
        initialState = toListOrNull(initialState);
        constants = toListOrNull(constants);
        return { inputs, initialState, constants };
    }
    /**
     * Iterates over the time dimension of a tensor.
     *
     * @param stepFunction RNN step function.
     *   Parameters:
     *     inputs: tensor with shape `[samples, ...]` (no time dimension),
     *       representing input for the batch of samples at a certain time step.
     *     states: an Array of tensors.
     *   Returns:
     *     outputs: tensor with shape `[samples, outputDim]` (no time dimension).
     *     newStates: list of tensors, same length and shapes as `states`. The first
     *       state in the list must be the output tensor at the previous timestep.
     * @param inputs Tensor of temporal data of shape `[samples, time, ...]` (at
     *   least 3D).
     * @param initialStates Tensor with shape `[samples, outputDim]` (no time
     *   dimension), containing the initial values of the states used in the step
     *   function.
     * @param goBackwards If `true`, do the iteration over the time dimension in
     *   reverse order and return the reversed sequence.
     * @param mask Binary tensor with shape `[sample, time, 1]`, with a zero for
     *   every element that is masked.
     * @param constants An Array of constant values passed at each step.
     * @param unroll Whether to unroll the RNN or to use a symbolic loop. *Not*
     *   applicable to this imperative deeplearn.js backend. Its value is ignored.
     * @param needPerStepOutputs Whether the per-step outputs are to be
     *   concatenated into a single tensor and returned (as the second return
     *   value). Default: `false`. This arg is included so that the relatively
     *   expensive concatenation of the stepwise outputs can be omitted unless
     *   the stepwise outputs need to be kept (e.g., for an LSTM layer of which
     *   `returnSequence` is `true`.)
     * @returns An Array: `[lastOutput, outputs, newStates]`.
     *   lastOutput: the lastest output of the RNN, of shape `[samples, ...]`.
     *   outputs: tensor with shape `[samples, time, ...]` where each entry
     *     `output[s, t]` is the output of the step function at time `t` for sample
     *     `s`. This return value is provided if and only if the
     *     `needPerStepOutputs` is set as `true`. If it is set as `false`, this
     *     return value will be `undefined`.
     *   newStates: Array of tensors, latest states returned by the step function,
     *      of shape `(samples, ...)`.
     * @throws ValueError If input dimension is less than 3.
     *
     * TODO(nielsene): This needs to be tidy-ed.
     */
    function rnn$1(stepFunction, inputs, initialStates, goBackwards = false, mask, constants, unroll = false, needPerStepOutputs = false) {
        return tidy(() => {
            const ndim = inputs.shape.length;
            if (ndim < 3) {
                throw new ValueError(`Input should be at least 3D, but is ${ndim}D.`);
            }
            // Transpose to time-major, i.e., from [batch, time, ...] to [time, batch,
            // ...].
            const axes = [1, 0].concat(range$3(2, ndim));
            inputs = transpose$2(inputs, axes);
            if (constants != null) {
                throw new NotImplementedError('The rnn() functoin of the deeplearn.js backend does not support ' +
                    'constants yet.');
            }
            // Porting Note: the unroll option is ignored by the imperative backend.
            if (unroll) {
                console.warn('Backend rnn(): the unroll = true option is not applicable to the ' +
                    'imperative deeplearn.js backend.');
            }
            if (mask != null) {
                mask = cast$3(cast$3(mask, 'bool'), 'float32');
                if (mask.rank === ndim - 1) {
                    mask = expandDims$3(mask, -1);
                }
                mask = transpose$2(mask, axes);
            }
            if (goBackwards) {
                inputs = reverse$2(inputs, 0);
                if (mask != null) {
                    mask = reverse$2(mask, 0);
                }
            }
            // Porting Note: PyKeras with TensorFlow backend uses a symbolic loop
            //   (tf.while_loop). But for the imperative deeplearn.js backend, we just
            //   use the usual TypeScript control flow to iterate over the time steps in
            //   the inputs.
            // Porting Note: PyKeras patches a "_use_learning_phase" attribute to
            // outputs.
            //   This is not idiomatic in TypeScript. The info regarding whether we are
            //   in a learning (i.e., training) phase for RNN is passed in a different
            //   way.
            const perStepOutputs = [];
            let lastOutput;
            let states = initialStates;
            const timeSteps = inputs.shape[0];
            const perStepInputs = unstack(inputs);
            let perStepMasks;
            if (mask != null) {
                perStepMasks = unstack(mask);
            }
            for (let t = 0; t < timeSteps; ++t) {
                const currentInput = perStepInputs[t];
                const stepOutputs = tidy(() => stepFunction(currentInput, states));
                if (mask == null) {
                    lastOutput = stepOutputs[0];
                    states = stepOutputs[1];
                }
                else {
                    const maskedOutputs = tidy(() => {
                        const stepMask = perStepMasks[t];
                        const negStepMask = sub$2(onesLike$2(stepMask), stepMask);
                        // TODO(cais): Would tfc.where() be better for performance?
                        const output = add$2(mul(stepOutputs[0], stepMask), mul(states[0], negStepMask));
                        const newStates = states.map((state, i) => {
                            return add$2(mul(stepOutputs[1][i], stepMask), mul(state, negStepMask));
                        });
                        return { output, newStates };
                    });
                    lastOutput = maskedOutputs.output;
                    states = maskedOutputs.newStates;
                }
                if (needPerStepOutputs) {
                    perStepOutputs.push(lastOutput);
                }
            }
            let outputs;
            if (needPerStepOutputs) {
                const axis = 1;
                outputs = stack(perStepOutputs, axis);
            }
            return [lastOutput, outputs, states];
        });
    }
    class RNN extends Layer {
        constructor(args) {
            super(args);
            let cell;
            if (args.cell == null) {
                throw new ValueError('cell property is missing for the constructor of RNN.');
            }
            else if (Array.isArray(args.cell)) {
                cell = new StackedRNNCells({ cells: args.cell });
            }
            else {
                cell = args.cell;
            }
            if (cell.stateSize == null) {
                throw new ValueError('The RNN cell should have an attribute `stateSize` (tuple of ' +
                    'integers, one integer per RNN state).');
            }
            this.cell = cell;
            this.returnSequences =
                args.returnSequences == null ? false : args.returnSequences;
            this.returnState = args.returnState == null ? false : args.returnState;
            this.goBackwards = args.goBackwards == null ? false : args.goBackwards;
            this._stateful = args.stateful == null ? false : args.stateful;
            this.unroll = args.unroll == null ? false : args.unroll;
            this.supportsMasking = true;
            this.inputSpec = [new InputSpec({ ndim: 3 })];
            this.stateSpec = null;
            this.states_ = null;
            // TODO(cais): Add constantsSpec and numConstants.
            this.numConstants = null;
            // TODO(cais): Look into the use of initial_state in the kwargs of the
            //   constructor.
            this.keptStates = [];
        }
        // Porting Note: This is the equivalent of `RNN.states` property getter in
        //   PyKeras.
        getStates() {
            if (this.states_ == null) {
                const numStates = Array.isArray(this.cell.stateSize) ? this.cell.stateSize.length : 1;
                return range$3(0, numStates).map(x => null);
            }
            else {
                return this.states_;
            }
        }
        // Porting Note: This is the equivalent of the `RNN.states` property setter in
        //   PyKeras.
        setStates(states) {
            this.states_ = states;
        }
        computeOutputShape(inputShape) {
            if (isArrayOfShapes(inputShape)) {
                inputShape = inputShape[0];
            }
            inputShape = inputShape;
            // TODO(cais): Remove the casting once stacked RNN cells become supported.
            let stateSize = this.cell.stateSize;
            if (!Array.isArray(stateSize)) {
                stateSize = [stateSize];
            }
            const outputDim = stateSize[0];
            let outputShape;
            if (this.returnSequences) {
                outputShape = [inputShape[0], inputShape[1], outputDim];
            }
            else {
                outputShape = [inputShape[0], outputDim];
            }
            if (this.returnState) {
                const stateShape = [];
                for (const dim of stateSize) {
                    stateShape.push([inputShape[0], dim]);
                }
                return [outputShape].concat(stateShape);
            }
            else {
                return outputShape;
            }
        }
        computeMask(inputs, mask) {
            return tidy(() => {
                if (Array.isArray(mask)) {
                    mask = mask[0];
                }
                const outputMask = this.returnSequences ? mask : null;
                if (this.returnState) {
                    const stateMask = this.states.map(s => null);
                    return [outputMask].concat(stateMask);
                }
                else {
                    return outputMask;
                }
            });
        }
        /**
         * Get the current state tensors of the RNN.
         *
         * If the state hasn't been set, return an array of `null`s of the correct
         * length.
         */
        get states() {
            if (this.states_ == null) {
                const numStates = Array.isArray(this.cell.stateSize) ? this.cell.stateSize.length : 1;
                const output = [];
                for (let i = 0; i < numStates; ++i) {
                    output.push(null);
                }
                return output;
            }
            else {
                return this.states_;
            }
        }
        set states(s) {
            this.states_ = s;
        }
        build(inputShape) {
            if (this.numConstants != null) {
                throw new NotImplementedError('Constants support is not implemented in RNN yet.');
            }
            if (isArrayOfShapes(inputShape)) {
                inputShape = inputShape[0];
            }
            inputShape = inputShape;
            const batchSize = this.stateful ? inputShape[0] : null;
            const inputDim = inputShape.slice(2);
            this.inputSpec[0] = new InputSpec({ shape: [batchSize, null, ...inputDim] });
            // Allow cell (if RNNCell Layer) to build before we set or validate
            // stateSpec.
            const stepInputShape = [inputShape[0]].concat(inputShape.slice(2));
            {
                this.cell.build(stepInputShape);
            }
            // Set or validate stateSpec.
            let stateSize;
            if (Array.isArray(this.cell.stateSize)) {
                stateSize = this.cell.stateSize;
            }
            else {
                stateSize = [this.cell.stateSize];
            }
            if (this.stateSpec != null) {
                if (!arraysEqual(this.stateSpec.map(spec => spec.shape[spec.shape.length - 1]), stateSize)) {
                    throw new ValueError(`An initialState was passed that is not compatible with ` +
                        `cell.stateSize. Received stateSpec=${this.stateSpec}; ` +
                        `However cell.stateSize is ${this.cell.stateSize}`);
                }
            }
            else {
                this.stateSpec =
                    stateSize.map(dim => new InputSpec({ shape: [null, dim] }));
            }
            if (this.stateful) {
                this.resetStates();
            }
        }
        /**
         * Reset the state tensors of the RNN.
         *
         * If the `states` argument is `undefined` or `null`, will set the
         * state tensor(s) of the RNN to all-zero tensors of the appropriate
         * shape(s).
         *
         * If `states` is provided, will set the state tensors of the RNN to its
         * value.
         *
         * @param states Optional externally-provided initial states.
         * @param training Whether this call is done during training. For stateful
         *   RNNs, this affects whether the old states are kept or discarded. In
         *   particular, if `training` is `true`, the old states will be kept so
         *   that subsequent backpropgataion through time (BPTT) may work properly.
         *   Else, the old states will be discarded.
         */
        resetStates(states, training = false) {
            tidy(() => {
                if (!this.stateful) {
                    throw new AttributeError('Cannot call resetStates() on an RNN Layer that is not stateful.');
                }
                const batchSize = this.inputSpec[0].shape[0];
                if (batchSize == null) {
                    throw new ValueError('If an RNN is stateful, it needs to know its batch size. Specify ' +
                        'the batch size of your input tensors: \n' +
                        '- If using a Sequential model, specify the batch size by ' +
                        'passing a `batchInputShape` option to your first layer.\n' +
                        '- If using the functional API, specify the batch size by ' +
                        'passing a `batchShape` option to your Input layer.');
                }
                // Initialize state if null.
                if (this.states_ == null) {
                    if (Array.isArray(this.cell.stateSize)) {
                        this.states_ =
                            this.cell.stateSize.map(dim => zeros$2([batchSize, dim]));
                    }
                    else {
                        this.states_ = [zeros$2([batchSize, this.cell.stateSize])];
                    }
                }
                else if (states == null) {
                    // Dispose old state tensors.
                    dispose(this.states_);
                    // For stateful RNNs, fully dispose kept old states.
                    if (this.keptStates != null) {
                        dispose(this.keptStates);
                        this.keptStates = [];
                    }
                    if (Array.isArray(this.cell.stateSize)) {
                        this.states_ =
                            this.cell.stateSize.map(dim => zeros$2([batchSize, dim]));
                    }
                    else {
                        this.states_[0] = zeros$2([batchSize, this.cell.stateSize]);
                    }
                }
                else {
                    if (!Array.isArray(states)) {
                        states = [states];
                    }
                    if (states.length !== this.states_.length) {
                        throw new ValueError(`Layer ${this.name} expects ${this.states_.length} state(s), ` +
                            `but it received ${states.length} state value(s). Input ` +
                            `received: ${states}`);
                    }
                    if (training === true) {
                        // Store old state tensors for complete disposal later, i.e., during
                        // the next no-arg call to this method. We do not dispose the old
                        // states immediately because that BPTT (among other things) require
                        // them.
                        this.keptStates.push(this.states_.slice());
                    }
                    else {
                        dispose(this.states_);
                    }
                    for (let index = 0; index < this.states_.length; ++index) {
                        const value = states[index];
                        const dim = Array.isArray(this.cell.stateSize) ?
                            this.cell.stateSize[index] :
                            this.cell.stateSize;
                        const expectedShape = [batchSize, dim];
                        if (!arraysEqual(value.shape, expectedShape)) {
                            throw new ValueError(`State ${index} is incompatible with layer ${this.name}: ` +
                                `expected shape=${expectedShape}, received shape=${value.shape}`);
                        }
                        this.states_[index] = value;
                    }
                }
                this.states_ = this.states_.map(state => keep(state.clone()));
            });
        }
        apply(inputs, kwargs) {
            // TODO(cais): Figure out whether initialState is in kwargs or inputs.
            let initialState = kwargs == null ? null : kwargs['initialState'];
            let constants = kwargs == null ? null : kwargs['constants'];
            if (kwargs == null) {
                kwargs = {};
            }
            const standardized = standardizeArgs(inputs, initialState, constants, this.numConstants);
            inputs = standardized.inputs;
            initialState = standardized.initialState;
            constants = standardized.constants;
            // If any of `initial_state` or `constants` are specified and are
            // `tf.SymbolicTensor`s, then add them to the inputs and temporarily modify
            // the input_spec to include them.
            let additionalInputs = [];
            let additionalSpecs = [];
            if (initialState != null) {
                kwargs['initialState'] = initialState;
                additionalInputs = additionalInputs.concat(initialState);
                this.stateSpec = [];
                for (const state of initialState) {
                    this.stateSpec.push(new InputSpec({ shape: state.shape }));
                }
                // TODO(cais): Use the following instead.
                // this.stateSpec = initialState.map(state => new InputSpec({shape:
                // state.shape}));
                additionalSpecs = additionalSpecs.concat(this.stateSpec);
            }
            if (constants != null) {
                kwargs['constants'] = constants;
                additionalInputs = additionalInputs.concat(constants);
                // TODO(cais): Add this.constantsSpec.
                this.numConstants = constants.length;
            }
            const isTensor = additionalInputs[0] instanceof SymbolicTensor;
            if (isTensor) {
                // Compute full input spec, including state and constants.
                const fullInput = [inputs].concat(additionalInputs);
                const fullInputSpec = this.inputSpec.concat(additionalSpecs);
                // Perform the call with temporarily replaced inputSpec.
                const originalInputSpec = this.inputSpec;
                this.inputSpec = fullInputSpec;
                const output = super.apply(fullInput, kwargs);
                this.inputSpec = originalInputSpec;
                return output;
            }
            else {
                return super.apply(inputs, kwargs);
            }
        }
        // tslint:disable-next-line:no-any
        call(inputs, kwargs) {
            // Input shape: `[samples, time (padded with zeros), input_dim]`.
            // Note that the .build() method of subclasses **must** define
            // this.inputSpec and this.stateSpec owith complete input shapes.
            return tidy(() => {
                const mask = kwargs == null ? null : kwargs['mask'];
                const training = kwargs == null ? null : kwargs['training'];
                let initialState = kwargs == null ? null : kwargs['initialState'];
                inputs = getExactlyOneTensor(inputs);
                if (initialState == null) {
                    if (this.stateful) {
                        initialState = this.states_;
                    }
                    else {
                        initialState = this.getInitialState(inputs);
                    }
                }
                const numStates = Array.isArray(this.cell.stateSize) ? this.cell.stateSize.length : 1;
                if (initialState.length !== numStates) {
                    throw new ValueError(`RNN Layer has ${numStates} state(s) but was passed ` +
                        `${initialState.length} initial state(s).`);
                }
                if (this.unroll) {
                    console.warn('Ignoring unroll = true for RNN layer, due to imperative backend.');
                }
                const cellCallKwargs = { training };
                // TODO(cais): Add support for constants.
                const step = (inputs, states) => {
                    // `inputs` and `states` are concatenated to form a single `Array` of
                    // `tf.Tensor`s as the input to `cell.call()`.
                    const outputs = this.cell.call([inputs].concat(states), cellCallKwargs);
                    // Marshall the return value into output and new states.
                    return [outputs[0], outputs.slice(1)];
                };
                // TODO(cais): Add support for constants.
                const rnnOutputs = rnn$1(step, inputs, initialState, this.goBackwards, mask, null, this.unroll, this.returnSequences);
                const lastOutput = rnnOutputs[0];
                const outputs = rnnOutputs[1];
                const states = rnnOutputs[2];
                if (this.stateful) {
                    this.resetStates(states, training);
                }
                const output = this.returnSequences ? outputs : lastOutput;
                // TODO(cais): Porperty set learning phase flag.
                if (this.returnState) {
                    return [output].concat(states);
                }
                else {
                    return output;
                }
            });
        }
        getInitialState(inputs) {
            return tidy(() => {
                // Build an all-zero tensor of shape [samples, outputDim].
                // [Samples, timeSteps, inputDim].
                let initialState = zeros$2(inputs.shape);
                // [Samples].
                initialState = sum$2(initialState, [1, 2]);
                initialState = expandDims$2(initialState); // [Samples, 1].
                if (Array.isArray(this.cell.stateSize)) {
                    return this.cell.stateSize.map(dim => dim > 1 ? tile$2(initialState, [1, dim]) : initialState);
                }
                else {
                    return this.cell.stateSize > 1 ?
                        [tile$2(initialState, [1, this.cell.stateSize])] :
                        [initialState];
                }
            });
        }
        get trainableWeights() {
            if (!this.trainable) {
                return [];
            }
            // Porting Note: In TypeScript, `this` is always an instance of `Layer`.
            return this.cell.trainableWeights;
        }
        get nonTrainableWeights() {
            // Porting Note: In TypeScript, `this` is always an instance of `Layer`.
            if (!this.trainable) {
                return this.cell.weights;
            }
            return this.cell.nonTrainableWeights;
        }
        setFastWeightInitDuringBuild(value) {
            super.setFastWeightInitDuringBuild(value);
            if (this.cell != null) {
                this.cell.setFastWeightInitDuringBuild(value);
            }
        }
        getConfig() {
            const baseConfig = super.getConfig();
            const config = {
                returnSequences: this.returnSequences,
                returnState: this.returnState,
                goBackwards: this.goBackwards,
                stateful: this.stateful,
                unroll: this.unroll,
            };
            if (this.numConstants != null) {
                config['numConstants'] = this.numConstants;
            }
            const cellConfig = this.cell.getConfig();
            if (this.getClassName() === RNN.className) {
                config['cell'] = {
                    'className': this.cell.getClassName(),
                    'config': cellConfig,
                };
            }
            // this order is necessary, to prevent cell name from replacing layer name
            return Object.assign({}, cellConfig, baseConfig, config);
        }
        /** @nocollapse */
        static fromConfig(cls, config, customObjects = {}) {
            const cellConfig = config['cell'];
            const cell = deserialize(cellConfig, customObjects);
            return new cls(Object.assign(config, { cell }));
        }
    }
    /** @nocollapse */
    RNN.className = 'RNN';
    registerClass(RNN);
    // Porting Note: This is a common parent class for RNN cells. There is no
    // equivalent of this in PyKeras. Having a common parent class forgoes the
    //  need for `has_attr(cell, ...)` checks or its TypeScript equivalent.
    /**
     * An RNNCell layer.
     *
     * @doc {heading: 'Layers', subheading: 'Classes'}
     */
    class RNNCell extends Layer {
    }
    class SimpleRNNCell extends RNNCell {
        constructor(args) {
            super(args);
            this.DEFAULT_ACTIVATION = 'tanh';
            this.DEFAULT_KERNEL_INITIALIZER = 'glorotNormal';
            this.DEFAULT_RECURRENT_INITIALIZER = 'orthogonal';
            this.DEFAULT_BIAS_INITIALIZER = 'zeros';
            this.units = args.units;
            assertPositiveInteger(this.units, `units`);
            this.activation = getActivation(args.activation == null ? this.DEFAULT_ACTIVATION : args.activation);
            this.useBias = args.useBias == null ? true : args.useBias;
            this.kernelInitializer = getInitializer(args.kernelInitializer || this.DEFAULT_KERNEL_INITIALIZER);
            this.recurrentInitializer = getInitializer(args.recurrentInitializer || this.DEFAULT_RECURRENT_INITIALIZER);
            this.biasInitializer =
                getInitializer(args.biasInitializer || this.DEFAULT_BIAS_INITIALIZER);
            this.kernelRegularizer = getRegularizer(args.kernelRegularizer);
            this.recurrentRegularizer = getRegularizer(args.recurrentRegularizer);
            this.biasRegularizer = getRegularizer(args.biasRegularizer);
            this.kernelConstraint = getConstraint(args.kernelConstraint);
            this.recurrentConstraint = getConstraint(args.recurrentConstraint);
            this.biasConstraint = getConstraint(args.biasConstraint);
            this.dropout = min$2([1, max$2([0, args.dropout == null ? 0 : args.dropout])]);
            this.recurrentDropout = min$2([
                1,
                max$2([0, args.recurrentDropout == null ? 0 : args.recurrentDropout])
            ]);
            this.dropoutFunc = args.dropoutFunc;
            this.stateSize = this.units;
            this.dropoutMask = null;
            this.recurrentDropoutMask = null;
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            // TODO(cais): Use regularizer.
            this.kernel = this.addWeight('kernel', [inputShape[inputShape.length - 1], this.units], null, this.kernelInitializer, this.kernelRegularizer, true, this.kernelConstraint);
            this.recurrentKernel = this.addWeight('recurrent_kernel', [this.units, this.units], null, this.recurrentInitializer, this.recurrentRegularizer, true, this.recurrentConstraint);
            if (this.useBias) {
                this.bias = this.addWeight('bias', [this.units], null, this.biasInitializer, this.biasRegularizer, true, this.biasConstraint);
            }
            else {
                this.bias = null;
            }
            this.built = true;
        }
        // Porting Note: PyKeras' equivalent of this method takes two tensor inputs:
        //   `inputs` and `states`. Here, the two tensors are combined into an
        //   `Tensor[]` Array as the first input argument.
        //   Similarly, PyKeras' equivalent of this method returns two values:
        //    `output` and `[output]`. Here the two are combined into one length-2
        //    `Tensor[]`, consisting of `output` repeated.
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = inputs;
                if (inputs.length !== 2) {
                    throw new ValueError(`SimpleRNNCell expects 2 input Tensors, got ${inputs.length}.`);
                }
                let prevOutput = inputs[1];
                inputs = inputs[0];
                const training = kwargs['training'] == null ? false : kwargs['training'];
                if (0 < this.dropout && this.dropout < 1 && this.dropoutMask == null) {
                    this.dropoutMask = generateDropoutMask({
                        ones: () => onesLike$2(inputs),
                        rate: this.dropout,
                        training,
                        dropoutFunc: this.dropoutFunc,
                    });
                }
                if (0 < this.recurrentDropout && this.recurrentDropout < 1 &&
                    this.recurrentDropoutMask == null) {
                    this.recurrentDropoutMask = generateDropoutMask({
                        ones: () => onesLike$2(prevOutput),
                        rate: this.recurrentDropout,
                        training,
                        dropoutFunc: this.dropoutFunc,
                    });
                }
                let h;
                const dpMask = this.dropoutMask;
                const recDpMask = this.recurrentDropoutMask;
                if (dpMask != null) {
                    h = dot$1(mul(inputs, dpMask), this.kernel.read());
                }
                else {
                    h = dot$1(inputs, this.kernel.read());
                }
                if (this.bias != null) {
                    h = biasAdd(h, this.bias.read());
                }
                if (recDpMask != null) {
                    prevOutput = mul(prevOutput, recDpMask);
                }
                let output = add$2(h, dot$1(prevOutput, this.recurrentKernel.read()));
                if (this.activation != null) {
                    output = this.activation.apply(output);
                }
                // TODO(cais): Properly set learning phase on output tensor?
                return [output, output];
            });
        }
        getConfig() {
            const baseConfig = super.getConfig();
            const config = {
                units: this.units,
                activation: serializeActivation(this.activation),
                useBias: this.useBias,
                kernelInitializer: serializeInitializer(this.kernelInitializer),
                recurrentInitializer: serializeInitializer(this.recurrentInitializer),
                biasInitializer: serializeInitializer(this.biasInitializer),
                kernelRegularizer: serializeRegularizer(this.kernelRegularizer),
                recurrentRegularizer: serializeRegularizer(this.recurrentRegularizer),
                biasRegularizer: serializeRegularizer(this.biasRegularizer),
                activityRegularizer: serializeRegularizer(this.activityRegularizer),
                kernelConstraint: serializeConstraint(this.kernelConstraint),
                recurrentConstraint: serializeConstraint(this.recurrentConstraint),
                biasConstraint: serializeConstraint(this.biasConstraint),
                dropout: this.dropout,
                recurrentDropout: this.recurrentDropout,
            };
            return Object.assign({}, baseConfig, config);
        }
    }
    /** @nocollapse */
    SimpleRNNCell.className = 'SimpleRNNCell';
    registerClass(SimpleRNNCell);
    class SimpleRNN extends RNN {
        constructor(args) {
            args.cell = new SimpleRNNCell(args);
            super(args);
            // TODO(cais): Add activityRegularizer.
        }
        call(inputs, kwargs) {
            return tidy(() => {
                if (this.cell.dropoutMask != null) {
                    dispose(this.cell.dropoutMask);
                    this.cell.dropoutMask = null;
                }
                if (this.cell.recurrentDropoutMask != null) {
                    dispose(this.cell.recurrentDropoutMask);
                    this.cell.recurrentDropoutMask = null;
                }
                const mask = kwargs == null ? null : kwargs['mask'];
                const training = kwargs == null ? null : kwargs['training'];
                const initialState = kwargs == null ? null : kwargs['initialState'];
                return super.call(inputs, { mask, training, initialState });
            });
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config);
        }
    }
    /** @nocollapse */
    SimpleRNN.className = 'SimpleRNN';
    registerClass(SimpleRNN);
    class GRUCell extends RNNCell {
        constructor(args) {
            super(args);
            this.DEFAULT_ACTIVATION = 'tanh';
            this.DEFAULT_RECURRENT_ACTIVATION = 'hardSigmoid';
            this.DEFAULT_KERNEL_INITIALIZER = 'glorotNormal';
            this.DEFAULT_RECURRENT_INITIALIZER = 'orthogonal';
            this.DEFAULT_BIAS_INITIALIZER = 'zeros';
            if (args.resetAfter) {
                throw new ValueError(`GRUCell does not support reset_after parameter set to true.`);
            }
            this.units = args.units;
            assertPositiveInteger(this.units, 'units');
            this.activation = getActivation(args.activation === undefined ? this.DEFAULT_ACTIVATION :
                args.activation);
            this.recurrentActivation = getActivation(args.recurrentActivation === undefined ?
                this.DEFAULT_RECURRENT_ACTIVATION :
                args.recurrentActivation);
            this.useBias = args.useBias == null ? true : args.useBias;
            this.kernelInitializer = getInitializer(args.kernelInitializer || this.DEFAULT_KERNEL_INITIALIZER);
            this.recurrentInitializer = getInitializer(args.recurrentInitializer || this.DEFAULT_RECURRENT_INITIALIZER);
            this.biasInitializer =
                getInitializer(args.biasInitializer || this.DEFAULT_BIAS_INITIALIZER);
            this.kernelRegularizer = getRegularizer(args.kernelRegularizer);
            this.recurrentRegularizer = getRegularizer(args.recurrentRegularizer);
            this.biasRegularizer = getRegularizer(args.biasRegularizer);
            this.kernelConstraint = getConstraint(args.kernelConstraint);
            this.recurrentConstraint = getConstraint(args.recurrentConstraint);
            this.biasConstraint = getConstraint(args.biasConstraint);
            this.dropout = min$2([1, max$2([0, args.dropout == null ? 0 : args.dropout])]);
            this.recurrentDropout = min$2([
                1,
                max$2([0, args.recurrentDropout == null ? 0 : args.recurrentDropout])
            ]);
            this.dropoutFunc = args.dropoutFunc;
            this.implementation = args.implementation;
            this.stateSize = this.units;
            this.dropoutMask = null;
            this.recurrentDropoutMask = null;
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const inputDim = inputShape[inputShape.length - 1];
            this.kernel = this.addWeight('kernel', [inputDim, this.units * 3], null, this.kernelInitializer, this.kernelRegularizer, true, this.kernelConstraint);
            this.recurrentKernel = this.addWeight('recurrent_kernel', [this.units, this.units * 3], null, this.recurrentInitializer, this.recurrentRegularizer, true, this.recurrentConstraint);
            if (this.useBias) {
                this.bias = this.addWeight('bias', [this.units * 3], null, this.biasInitializer, this.biasRegularizer, true, this.biasConstraint);
            }
            else {
                this.bias = null;
            }
            // Porting Notes: Unlike the PyKeras implementation, we perform slicing
            //   of the weights and bias in the call() method, at execution time.
            this.built = true;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = inputs;
                if (inputs.length !== 2) {
                    throw new ValueError(`GRUCell expects 2 input Tensors (inputs, h, c), got ` +
                        `${inputs.length}.`);
                }
                const training = kwargs['training'] == null ? false : kwargs['training'];
                let hTMinus1 = inputs[1]; // Previous memory state.
                inputs = inputs[0];
                // Note: For superior performance, TensorFlow.js always uses
                // implementation 2, regardless of the actual value of
                // config.implementation.
                if (0 < this.dropout && this.dropout < 1 && this.dropoutMask == null) {
                    this.dropoutMask = generateDropoutMask({
                        ones: () => onesLike$2(inputs),
                        rate: this.dropout,
                        training,
                        count: 3,
                        dropoutFunc: this.dropoutFunc,
                    });
                }
                if (0 < this.recurrentDropout && this.recurrentDropout < 1 &&
                    this.recurrentDropoutMask == null) {
                    this.recurrentDropoutMask = generateDropoutMask({
                        ones: () => onesLike$2(hTMinus1),
                        rate: this.recurrentDropout,
                        training,
                        count: 3,
                        dropoutFunc: this.dropoutFunc,
                    });
                }
                const dpMask = this.dropoutMask;
                const recDpMask = this.recurrentDropoutMask;
                let z;
                let r;
                let hh;
                if (0 < this.dropout && this.dropout < 1) {
                    inputs = mul(inputs, dpMask[0]);
                }
                let matrixX = dot$1(inputs, this.kernel.read());
                if (this.useBias) {
                    matrixX = biasAdd(matrixX, this.bias.read());
                }
                if (0 < this.recurrentDropout && this.recurrentDropout < 1) {
                    hTMinus1 = mul(hTMinus1, recDpMask[0]);
                }
                const recurrentKernelValue = this.recurrentKernel.read();
                const [rk1, rk2] = split$2(recurrentKernelValue, [2 * this.units, this.units], recurrentKernelValue.rank - 1);
                const matrixInner = dot$1(hTMinus1, rk1);
                const [xZ, xR, xH] = split$2(matrixX, 3, matrixX.rank - 1);
                const [recurrentZ, recurrentR] = split$2(matrixInner, 2, matrixInner.rank - 1);
                z = this.recurrentActivation.apply(add$2(xZ, recurrentZ));
                r = this.recurrentActivation.apply(add$2(xR, recurrentR));
                const recurrentH = dot$1(mul(r, hTMinus1), rk2);
                hh = this.activation.apply(add$2(xH, recurrentH));
                const h = add$2(mul(z, hTMinus1), mul(add$2(1, neg$2(z)), hh));
                // TODO(cais): Add use_learning_phase flag properly.
                return [h, h];
            });
        }
        getConfig() {
            const baseConfig = super.getConfig();
            const config = {
                units: this.units,
                activation: serializeActivation(this.activation),
                recurrentActivation: serializeActivation(this.recurrentActivation),
                useBias: this.useBias,
                kernelInitializer: serializeInitializer(this.kernelInitializer),
                recurrentInitializer: serializeInitializer(this.recurrentInitializer),
                biasInitializer: serializeInitializer(this.biasInitializer),
                kernelRegularizer: serializeRegularizer(this.kernelRegularizer),
                recurrentRegularizer: serializeRegularizer(this.recurrentRegularizer),
                biasRegularizer: serializeRegularizer(this.biasRegularizer),
                activityRegularizer: serializeRegularizer(this.activityRegularizer),
                kernelConstraint: serializeConstraint(this.kernelConstraint),
                recurrentConstraint: serializeConstraint(this.recurrentConstraint),
                biasConstraint: serializeConstraint(this.biasConstraint),
                dropout: this.dropout,
                recurrentDropout: this.recurrentDropout,
                implementation: this.implementation,
                resetAfter: false
            };
            return Object.assign({}, baseConfig, config);
        }
    }
    /** @nocollapse */
    GRUCell.className = 'GRUCell';
    registerClass(GRUCell);
    class GRU extends RNN {
        constructor(args) {
            if (args.implementation === 0) {
                console.warn('`implementation=0` has been deprecated, and now defaults to ' +
                    '`implementation=1`. Please update your layer call.');
            }
            args.cell = new GRUCell(args);
            super(args);
            // TODO(cais): Add activityRegularizer.
        }
        call(inputs, kwargs) {
            return tidy(() => {
                if (this.cell.dropoutMask != null) {
                    dispose(this.cell.dropoutMask);
                    this.cell.dropoutMask = null;
                }
                if (this.cell.recurrentDropoutMask != null) {
                    dispose(this.cell.recurrentDropoutMask);
                    this.cell.recurrentDropoutMask = null;
                }
                const mask = kwargs == null ? null : kwargs['mask'];
                const training = kwargs == null ? null : kwargs['training'];
                const initialState = kwargs == null ? null : kwargs['initialState'];
                return super.call(inputs, { mask, training, initialState });
            });
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            if (config['implmentation'] === 0) {
                config['implementation'] = 1;
            }
            return new cls(config);
        }
    }
    /** @nocollapse */
    GRU.className = 'GRU';
    registerClass(GRU);
    class LSTMCell extends RNNCell {
        constructor(args) {
            super(args);
            this.DEFAULT_ACTIVATION = 'tanh';
            this.DEFAULT_RECURRENT_ACTIVATION = 'hardSigmoid';
            this.DEFAULT_KERNEL_INITIALIZER = 'glorotNormal';
            this.DEFAULT_RECURRENT_INITIALIZER = 'orthogonal';
            this.DEFAULT_BIAS_INITIALIZER = 'zeros';
            this.units = args.units;
            assertPositiveInteger(this.units, 'units');
            this.activation = getActivation(args.activation === undefined ? this.DEFAULT_ACTIVATION :
                args.activation);
            this.recurrentActivation = getActivation(args.recurrentActivation === undefined ?
                this.DEFAULT_RECURRENT_ACTIVATION :
                args.recurrentActivation);
            this.useBias = args.useBias == null ? true : args.useBias;
            this.kernelInitializer = getInitializer(args.kernelInitializer || this.DEFAULT_KERNEL_INITIALIZER);
            this.recurrentInitializer = getInitializer(args.recurrentInitializer || this.DEFAULT_RECURRENT_INITIALIZER);
            this.biasInitializer =
                getInitializer(args.biasInitializer || this.DEFAULT_BIAS_INITIALIZER);
            this.unitForgetBias = args.unitForgetBias;
            this.kernelRegularizer = getRegularizer(args.kernelRegularizer);
            this.recurrentRegularizer = getRegularizer(args.recurrentRegularizer);
            this.biasRegularizer = getRegularizer(args.biasRegularizer);
            this.kernelConstraint = getConstraint(args.kernelConstraint);
            this.recurrentConstraint = getConstraint(args.recurrentConstraint);
            this.biasConstraint = getConstraint(args.biasConstraint);
            this.dropout = min$2([1, max$2([0, args.dropout == null ? 0 : args.dropout])]);
            this.recurrentDropout = min$2([
                1,
                max$2([0, args.recurrentDropout == null ? 0 : args.recurrentDropout])
            ]);
            this.dropoutFunc = args.dropoutFunc;
            this.implementation = args.implementation;
            this.stateSize = [this.units, this.units];
            this.dropoutMask = null;
            this.recurrentDropoutMask = null;
        }
        build(inputShape) {
            var _a;
            inputShape = getExactlyOneShape(inputShape);
            const inputDim = inputShape[inputShape.length - 1];
            this.kernel = this.addWeight('kernel', [inputDim, this.units * 4], null, this.kernelInitializer, this.kernelRegularizer, true, this.kernelConstraint);
            this.recurrentKernel = this.addWeight('recurrent_kernel', [this.units, this.units * 4], null, this.recurrentInitializer, this.recurrentRegularizer, true, this.recurrentConstraint);
            let biasInitializer;
            if (this.useBias) {
                if (this.unitForgetBias) {
                    const capturedBiasInit = this.biasInitializer;
                    const capturedUnits = this.units;
                    biasInitializer = new (_a = class CustomInit extends Initializer {
                            apply(shape, dtype) {
                                // TODO(cais): More informative variable names?
                                const bI = capturedBiasInit.apply([capturedUnits]);
                                const bF = (new Ones()).apply([capturedUnits]);
                                const bCAndH = capturedBiasInit.apply([capturedUnits * 2]);
                                return concatAlongFirstAxis(concatAlongFirstAxis(bI, bF), bCAndH);
                            }
                        },
                        /** @nocollapse */
                        _a.className = 'CustomInit',
                        _a)();
                }
                else {
                    biasInitializer = this.biasInitializer;
                }
                this.bias = this.addWeight('bias', [this.units * 4], null, biasInitializer, this.biasRegularizer, true, this.biasConstraint);
            }
            else {
                this.bias = null;
            }
            // Porting Notes: Unlike the PyKeras implementation, we perform slicing
            //   of the weights and bias in the call() method, at execution time.
            this.built = true;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                const training = kwargs['training'] == null ? false : kwargs['training'];
                inputs = inputs;
                if (inputs.length !== 3) {
                    throw new ValueError(`LSTMCell expects 3 input Tensors (inputs, h, c), got ` +
                        `${inputs.length}.`);
                }
                let hTMinus1 = inputs[1]; // Previous memory state.
                const cTMinus1 = inputs[2]; // Previous carry state.
                inputs = inputs[0];
                if (0 < this.dropout && this.dropout < 1 && this.dropoutMask == null) {
                    this.dropoutMask = generateDropoutMask({
                        ones: () => onesLike$2(inputs),
                        rate: this.dropout,
                        training,
                        count: 4,
                        dropoutFunc: this.dropoutFunc
                    });
                }
                if (0 < this.recurrentDropout && this.recurrentDropout < 1 &&
                    this.recurrentDropoutMask == null) {
                    this.recurrentDropoutMask = generateDropoutMask({
                        ones: () => onesLike$2(hTMinus1),
                        rate: this.recurrentDropout,
                        training,
                        count: 4,
                        dropoutFunc: this.dropoutFunc
                    });
                }
                const dpMask = this.dropoutMask;
                const recDpMask = this.recurrentDropoutMask;
                // Note: For superior performance, TensorFlow.js always uses
                // implementation 2 regardless of the actual value of
                // config.implementation.
                let i;
                let f;
                let c;
                let o;
                if (0 < this.dropout && this.dropout < 1) {
                    inputs = mul(inputs, dpMask[0]);
                }
                let z = dot$1(inputs, this.kernel.read());
                if (0 < this.recurrentDropout && this.recurrentDropout < 1) {
                    hTMinus1 = mul(hTMinus1, recDpMask[0]);
                }
                z = add$2(z, dot$1(hTMinus1, this.recurrentKernel.read()));
                if (this.useBias) {
                    z = biasAdd(z, this.bias.read());
                }
                const [z0, z1, z2, z3] = split$2(z, 4, z.rank - 1);
                i = this.recurrentActivation.apply(z0);
                f = this.recurrentActivation.apply(z1);
                c = add$2(mul(f, cTMinus1), mul(i, this.activation.apply(z2)));
                o = this.recurrentActivation.apply(z3);
                const h = mul(o, this.activation.apply(c));
                // TODO(cais): Add use_learning_phase flag properly.
                return [h, h, c];
            });
        }
        getConfig() {
            const baseConfig = super.getConfig();
            const config = {
                units: this.units,
                activation: serializeActivation(this.activation),
                recurrentActivation: serializeActivation(this.recurrentActivation),
                useBias: this.useBias,
                kernelInitializer: serializeInitializer(this.kernelInitializer),
                recurrentInitializer: serializeInitializer(this.recurrentInitializer),
                biasInitializer: serializeInitializer(this.biasInitializer),
                unitForgetBias: this.unitForgetBias,
                kernelRegularizer: serializeRegularizer(this.kernelRegularizer),
                recurrentRegularizer: serializeRegularizer(this.recurrentRegularizer),
                biasRegularizer: serializeRegularizer(this.biasRegularizer),
                activityRegularizer: serializeRegularizer(this.activityRegularizer),
                kernelConstraint: serializeConstraint(this.kernelConstraint),
                recurrentConstraint: serializeConstraint(this.recurrentConstraint),
                biasConstraint: serializeConstraint(this.biasConstraint),
                dropout: this.dropout,
                recurrentDropout: this.recurrentDropout,
                implementation: this.implementation,
            };
            return Object.assign({}, baseConfig, config);
        }
    }
    /** @nocollapse */
    LSTMCell.className = 'LSTMCell';
    registerClass(LSTMCell);
    class LSTM extends RNN {
        constructor(args) {
            if (args.implementation === 0) {
                console.warn('`implementation=0` has been deprecated, and now defaults to ' +
                    '`implementation=1`. Please update your layer call.');
            }
            args.cell = new LSTMCell(args);
            super(args);
            // TODO(cais): Add activityRegularizer.
        }
        call(inputs, kwargs) {
            return tidy(() => {
                if (this.cell.dropoutMask != null) {
                    dispose(this.cell.dropoutMask);
                    this.cell.dropoutMask = null;
                }
                if (this.cell.recurrentDropoutMask != null) {
                    dispose(this.cell.recurrentDropoutMask);
                    this.cell.recurrentDropoutMask = null;
                }
                const mask = kwargs == null ? null : kwargs['mask'];
                const training = kwargs == null ? null : kwargs['training'];
                const initialState = kwargs == null ? null : kwargs['initialState'];
                return super.call(inputs, { mask, training, initialState });
            });
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            if (config['implmentation'] === 0) {
                config['implementation'] = 1;
            }
            return new cls(config);
        }
    }
    /** @nocollapse */
    LSTM.className = 'LSTM';
    registerClass(LSTM);
    class StackedRNNCells extends RNNCell {
        constructor(args) {
            super(args);
            this.cells = args.cells;
        }
        get stateSize() {
            // States are a flat list in reverse order of the cell stack.
            // This allows perserving the requirement `stack.statesize[0] ===
            // outputDim`. E.g., states of a 2-layer LSTM would be `[h2, c2, h1, c1]`,
            // assuming one LSTM has states `[h, c]`.
            const stateSize = [];
            for (const cell of this.cells.slice().reverse()) {
                if (Array.isArray(cell.stateSize)) {
                    stateSize.push(...cell.stateSize);
                }
                else {
                    stateSize.push(cell.stateSize);
                }
            }
            return stateSize;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = inputs;
                let states = inputs.slice(1);
                // Recover per-cell states.
                const nestedStates = [];
                for (const cell of this.cells.slice().reverse()) {
                    if (Array.isArray(cell.stateSize)) {
                        nestedStates.push(states.splice(0, cell.stateSize.length));
                    }
                    else {
                        nestedStates.push(states.splice(0, 1));
                    }
                }
                nestedStates.reverse();
                // Call the cells in order and store the returned states.
                const newNestedStates = [];
                let callInputs;
                for (let i = 0; i < this.cells.length; ++i) {
                    const cell = this.cells[i];
                    states = nestedStates[i];
                    // TODO(cais): Take care of constants.
                    if (i === 0) {
                        callInputs = [inputs[0]].concat(states);
                    }
                    else {
                        callInputs = [callInputs[0]].concat(states);
                    }
                    callInputs = cell.call(callInputs, kwargs);
                    newNestedStates.push(callInputs.slice(1));
                }
                // Format the new states as a flat list in reverse cell order.
                states = [];
                for (const cellStates of newNestedStates.slice().reverse()) {
                    states.push(...cellStates);
                }
                return [callInputs[0]].concat(states);
            });
        }
        build(inputShape) {
            if (isArrayOfShapes(inputShape)) {
                // TODO(cais): Take care of input constants.
                // const constantShape = inputShape.slice(1);
                inputShape = inputShape[0];
            }
            inputShape = inputShape;
            let outputDim;
            this.cells.forEach((cell, i) => {
                nameScope(`RNNCell_${i}`, () => {
                    // TODO(cais): Take care of input constants.
                    cell.build(inputShape);
                    if (Array.isArray(cell.stateSize)) {
                        outputDim = cell.stateSize[0];
                    }
                    else {
                        outputDim = cell.stateSize;
                    }
                    inputShape = [inputShape[0], outputDim];
                });
            });
            this.built = true;
        }
        getConfig() {
            const baseConfig = super.getConfig();
            const getCellConfig = (cell) => {
                return {
                    'className': cell.getClassName(),
                    'config': cell.getConfig(),
                };
            };
            const cellConfigs = this.cells.map(getCellConfig);
            const config = { 'cells': cellConfigs };
            return Object.assign({}, baseConfig, config);
        }
        /** @nocollapse */
        static fromConfig(cls, config, customObjects = {}) {
            const cells = [];
            for (const cellConfig of config['cells']) {
                cells.push(deserialize(cellConfig, customObjects));
            }
            return new cls({ cells });
        }
        get trainableWeights() {
            if (!this.trainable) {
                return [];
            }
            const weights = [];
            for (const cell of this.cells) {
                weights.push(...cell.trainableWeights);
            }
            return weights;
        }
        get nonTrainableWeights() {
            const weights = [];
            for (const cell of this.cells) {
                weights.push(...cell.nonTrainableWeights);
            }
            if (!this.trainable) {
                const trainableWeights = [];
                for (const cell of this.cells) {
                    trainableWeights.push(...cell.trainableWeights);
                }
                return trainableWeights.concat(weights);
            }
            return weights;
        }
        /**
         * Retrieve the weights of a the model.
         *
         * @returns A flat `Array` of `tf.Tensor`s.
         */
        getWeights() {
            const weights = [];
            for (const cell of this.cells) {
                weights.push(...cell.weights);
            }
            return batchGetValue(weights);
        }
        /**
         * Set the weights of the model.
         *
         * @param weights An `Array` of `tf.Tensor`s with shapes and types matching
         *     the output of `getWeights()`.
         */
        setWeights(weights) {
            const tuples = [];
            for (const cell of this.cells) {
                const numParams = cell.weights.length;
                const inputWeights = weights.splice(numParams);
                for (let i = 0; i < cell.weights.length; ++i) {
                    tuples.push([cell.weights[i], inputWeights[i]]);
                }
            }
            batchSetValue(tuples);
        }
    }
    /** @nocollapse */
    StackedRNNCells.className = 'StackedRNNCells';
    registerClass(StackedRNNCells);
    function generateDropoutMask(args) {
        const { ones, rate, training = false, count = 1, dropoutFunc } = args;
        const droppedInputs = () => dropoutFunc != null ? dropoutFunc(ones(), rate) : dropout$1(ones(), rate);
        const createMask = () => inTrainPhase(droppedInputs, ones, training);
        // just in case count is provided with null or undefined
        if (!count || count <= 1) {
            return keep(createMask().clone());
        }
        const masks = Array(count).fill(undefined).map(createMask);
        return masks.map(m => keep(m.clone()));
    }

    /**
     * @license
     * Copyright 2020 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    var __rest = (undefined && undefined.__rest) || function (s, e) {
        var t = {};
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
            t[p] = s[p];
        if (s != null && typeof Object.getOwnPropertySymbols === "function")
            for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
                if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                    t[p[i]] = s[p[i]];
            }
        return t;
    };
    /**
     * Base class for convolutional-recurrent layers.
     */
    class ConvRNN2D extends RNN {
        constructor(args) {
            if (args.unroll) {
                throw new NotImplementedError('Unrolling is not possible with convolutional RNNs.');
            }
            if (Array.isArray(args.cell)) {
                throw new NotImplementedError('It is not possible at the moment to stack convolutional cells.');
            }
            super(args);
            this.inputSpec = [new InputSpec({ ndim: 5 })];
        }
        call(inputs, kwargs) {
            return tidy(() => {
                if (this.cell.dropoutMask != null) {
                    dispose(this.cell.dropoutMask);
                    this.cell.dropoutMask = null;
                }
                if (this.cell.recurrentDropoutMask != null) {
                    dispose(this.cell.recurrentDropoutMask);
                    this.cell.recurrentDropoutMask = null;
                }
                if (kwargs && kwargs['constants']) {
                    throw new ValueError('ConvRNN2D cell does not support constants');
                }
                const mask = kwargs == null ? null : kwargs['mask'];
                const training = kwargs == null ? null : kwargs['training'];
                const initialState = kwargs == null ? null : kwargs['initialState'];
                return super.call(inputs, { mask, training, initialState });
            });
        }
        computeOutputShape(inputShape) {
            let outShape = this.computeSingleOutputShape(inputShape);
            if (!this.returnSequences) {
                outShape = [outShape[0], ...outShape.slice(2)];
            }
            if (this.returnState) {
                outShape =
                    [outShape, ...Array(2).fill([inputShape[0], ...outShape.slice(-3)])];
            }
            return outShape;
        }
        getInitialState(inputs) {
            return tidy(() => {
                const { stateSize } = this.cell;
                const inputShape = inputs.shape;
                const outputShape = this.computeSingleOutputShape(inputShape);
                const stateShape = [outputShape[0], ...outputShape.slice(2)];
                const initialState = zeros$2(stateShape);
                if (Array.isArray(stateSize)) {
                    return Array(stateSize.length).fill(initialState);
                }
                return [initialState];
            });
        }
        resetStates(states, training = false) {
            tidy(() => {
                if (!this.stateful) {
                    throw new AttributeError('Cannot call resetStates() on an RNN Layer that is not stateful.');
                }
                const inputShape = this.inputSpec[0].shape;
                const outputShape = this.computeSingleOutputShape(inputShape);
                const stateShape = [outputShape[0], ...outputShape.slice(2)];
                const batchSize = inputShape[0];
                if (batchSize == null) {
                    throw new ValueError('If an RNN is stateful, it needs to know its batch size. Specify ' +
                        'the batch size of your input tensors: \n' +
                        '- If using a Sequential model, specify the batch size by ' +
                        'passing a `batchInputShape` option to your first layer.\n' +
                        '- If using the functional API, specify the batch size by ' +
                        'passing a `batchShape` option to your Input layer.');
                }
                // Initialize state if null.
                if (this.getStates() == null) {
                    if (Array.isArray(this.cell.stateSize)) {
                        this.states_ = this.cell.stateSize.map(() => zeros$2(stateShape));
                    }
                    else {
                        this.states_ = [zeros$2(stateShape)];
                    }
                }
                else if (states == null) {
                    // Dispose old state tensors.
                    dispose(this.states_);
                    // For stateful RNNs, fully dispose kept old states.
                    if (this.keptStates != null) {
                        dispose(this.keptStates);
                        this.keptStates = [];
                    }
                    if (Array.isArray(this.cell.stateSize)) {
                        this.states_ = this.cell.stateSize.map(() => zeros$2(stateShape));
                    }
                    else {
                        this.states_[0] = zeros$2(stateShape);
                    }
                }
                else {
                    if (!Array.isArray(states)) {
                        states = [states];
                    }
                    if (states.length !== this.states_.length) {
                        throw new ValueError(`Layer ${this.name} expects ${this.states_.length} state(s), ` +
                            `but it received ${states.length} state value(s). Input ` +
                            `received: ${states}`);
                    }
                    if (training) {
                        // Store old state tensors for complete disposal later, i.e., during
                        // the next no-arg call to this method. We do not dispose the old
                        // states immediately because that BPTT (among other things) require
                        // them.
                        this.keptStates.push(this.states_.slice());
                    }
                    else {
                        dispose(this.states_);
                    }
                    for (let index = 0; index < this.states_.length; ++index) {
                        const value = states[index];
                        const expectedShape = stateShape;
                        if (!arraysEqual(value.shape, expectedShape)) {
                            throw new ValueError(`State ${index} is incompatible with layer ${this.name}: ` +
                                `expected shape=${expectedShape}, received shape=${value.shape}`);
                        }
                        this.states_[index] = value;
                    }
                }
                this.states_ = this.states_.map(state => keep(state.clone()));
            });
        }
        computeSingleOutputShape(inputShape) {
            const { dataFormat, filters, kernelSize, padding, strides, dilationRate } = this.cell;
            const isChannelsFirst = dataFormat === 'channelsFirst';
            const h = inputShape[isChannelsFirst ? 3 : 2];
            const w = inputShape[isChannelsFirst ? 4 : 3];
            const hOut = convOutputLength(h, kernelSize[0], padding, strides[0], dilationRate[0]);
            const wOut = convOutputLength(w, kernelSize[1], padding, strides[1], dilationRate[1]);
            const outShape = [
                ...inputShape.slice(0, 2),
                ...(isChannelsFirst ? [filters, hOut, wOut] : [hOut, wOut, filters])
            ];
            return outShape;
        }
    }
    /** @nocollapse */
    ConvRNN2D.className = 'ConvRNN2D';
    class ConvLSTM2DCell extends LSTMCell {
        constructor(args) {
            const { filters, kernelSize, strides, padding, dataFormat, dilationRate, } = args;
            super(Object.assign({}, args, { units: filters }));
            this.filters = filters;
            assertPositiveInteger(this.filters, 'filters');
            this.kernelSize = normalizeArray$1(kernelSize, 2, 'kernelSize');
            this.kernelSize.forEach(size => assertPositiveInteger(size, 'kernelSize'));
            this.strides = normalizeArray$1(strides || 1, 2, 'strides');
            this.strides.forEach(stride => assertPositiveInteger(stride, 'strides'));
            this.padding = padding || 'valid';
            checkPaddingMode(this.padding);
            this.dataFormat = dataFormat || 'channelsLast';
            checkDataFormat(this.dataFormat);
            this.dilationRate = normalizeArray$1(dilationRate || 1, 2, 'dilationRate');
            this.dilationRate.forEach(rate => assertPositiveInteger(rate, 'dilationRate'));
        }
        build(inputShape) {
            var _a;
            inputShape = getExactlyOneShape(inputShape);
            const channelAxis = this.dataFormat === 'channelsFirst' ? 1 : inputShape.length - 1;
            if (inputShape[channelAxis] == null) {
                throw new ValueError(`The channel dimension of the input should be defined. ` +
                    `Found ${inputShape[channelAxis]}`);
            }
            const inputDim = inputShape[channelAxis];
            const numOfKernels = 4;
            const kernelShape = this.kernelSize.concat([inputDim, this.filters * numOfKernels]);
            this.kernel = this.addWeight('kernel', kernelShape, null, this.kernelInitializer, this.kernelRegularizer, true, this.kernelConstraint);
            const recurrentKernelShape = this.kernelSize.concat([this.filters, this.filters * numOfKernels]);
            this.recurrentKernel = this.addWeight('recurrent_kernel', recurrentKernelShape, null, this.recurrentInitializer, this.recurrentRegularizer, true, this.recurrentConstraint);
            if (this.useBias) {
                let biasInitializer;
                if (this.unitForgetBias) {
                    const init = this.biasInitializer;
                    const filters = this.filters;
                    biasInitializer = new (_a = class CustomInit extends Initializer {
                            apply(shape, dtype) {
                                const biasI = init.apply([filters]);
                                const biasF = ones$1([filters]);
                                const biasCAndO = init.apply([filters * 2]);
                                return concatenate$1([biasI, biasF, biasCAndO]);
                            }
                        },
                        /** @nocollapse */
                        _a.className = 'CustomInit',
                        _a)();
                }
                else {
                    biasInitializer = this.biasInitializer;
                }
                this.bias = this.addWeight('bias', [this.filters * numOfKernels], null, biasInitializer, this.biasRegularizer, true, this.biasConstraint);
            }
            this.built = true;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                if (inputs.length !== 3) {
                    throw new ValueError(`ConvLSTM2DCell expects 3 input Tensors (inputs, h, c), got ` +
                        `${inputs.length}.`);
                }
                const training = kwargs['training'] || false;
                const x = inputs[0]; // Current input
                const hTMinus1 = inputs[1]; // Previous memory state.
                const cTMinus1 = inputs[2]; // Previous carry state.
                const numOfKernels = 4;
                if (0 < this.dropout && this.dropout < 1 && this.dropoutMask == null) {
                    this.dropoutMask = generateDropoutMask({
                        ones: () => onesLike$2(x),
                        rate: this.dropout,
                        training,
                        count: numOfKernels,
                        dropoutFunc: this.dropoutFunc
                    });
                }
                const dropoutMask = this.dropoutMask;
                const applyDropout = (x, mask, index) => {
                    if (!mask || !mask[index]) {
                        return x;
                    }
                    return mul(mask[index], x);
                };
                let xI = applyDropout(x, dropoutMask, 0);
                let xF = applyDropout(x, dropoutMask, 1);
                let xC = applyDropout(x, dropoutMask, 2);
                let xO = applyDropout(x, dropoutMask, 3);
                if (0 < this.recurrentDropout && this.recurrentDropout < 1 &&
                    this.recurrentDropoutMask == null) {
                    this.recurrentDropoutMask = generateDropoutMask({
                        ones: () => onesLike$2(hTMinus1),
                        rate: this.recurrentDropout,
                        training,
                        count: numOfKernels,
                        dropoutFunc: this.dropoutFunc
                    });
                }
                const recDropoutMask = this.recurrentDropoutMask;
                let hI = applyDropout(hTMinus1, recDropoutMask, 0);
                let hF = applyDropout(hTMinus1, recDropoutMask, 1);
                let hC = applyDropout(hTMinus1, recDropoutMask, 2);
                let hO = applyDropout(hTMinus1, recDropoutMask, 3);
                const kernelChannelAxis = 3;
                const [kernelI, kernelF, kernelC, kernelO] = split$2(this.kernel.read(), numOfKernels, kernelChannelAxis);
                const [biasI, biasF, biasC, biasO] = this.useBias ?
                    split$2(this.bias.read(), numOfKernels) :
                    [null, null, null, null];
                xI = this.inputConv(xI, kernelI, biasI, this.padding);
                xF = this.inputConv(xF, kernelF, biasF, this.padding);
                xC = this.inputConv(xC, kernelC, biasC, this.padding);
                xO = this.inputConv(xO, kernelO, biasO, this.padding);
                const [recKernelI, recKernelF, recKernelC, recKernelO] = split$2(this.recurrentKernel.read(), numOfKernels, kernelChannelAxis);
                hI = this.recurrentConv(hI, recKernelI);
                hF = this.recurrentConv(hF, recKernelF);
                hC = this.recurrentConv(hC, recKernelC);
                hO = this.recurrentConv(hO, recKernelO);
                const i = this.recurrentActivation.apply(add$2(xI, hI));
                const f = this.recurrentActivation.apply(add$2(xF, hF));
                const c = add$2(mul(f, cTMinus1), mul(i, this.activation.apply(add$2(xC, hC))));
                const h = mul(this.recurrentActivation.apply(add$2(xO, hO)), this.activation.apply(c));
                return [h, h, c];
            });
        }
        getConfig() {
            const _a = super.getConfig(), baseConfig = __rest(_a, ['units']);
            const config = {
                filters: this.filters,
                kernelSize: this.kernelSize,
                padding: this.padding,
                dataFormat: this.dataFormat,
                dilationRate: this.dilationRate,
                strides: this.strides,
            };
            return Object.assign({}, baseConfig, config);
        }
        inputConv(x, w, b, padding) {
            const out = conv2d$3(x, w, this.strides, (padding || 'valid'), this.dataFormat === 'channelsFirst' ? 'NCHW' : 'NHWC', this.dilationRate);
            if (b) {
                return biasAdd(out, b, this.dataFormat);
            }
            return out;
        }
        recurrentConv(x, w) {
            const strides = 1;
            return conv2d$3(x, w, strides, 'same', this.dataFormat === 'channelsFirst' ? 'NCHW' : 'NHWC');
        }
    }
    /** @nocollapse */
    ConvLSTM2DCell.className = 'ConvLSTM2DCell';
    registerClass(ConvLSTM2DCell);
    class ConvLSTM2D extends ConvRNN2D {
        constructor(args) {
            const cell = new ConvLSTM2DCell(args);
            super(Object.assign({}, args, { cell }));
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            return new cls(config);
        }
    }
    /** @nocollapse */
    ConvLSTM2D.className = 'ConvLSTM2D';
    registerClass(ConvLSTM2D);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    class Dropout extends Layer {
        constructor(args) {
            super(args);
            this.rate = Math.max(Math.min(args.rate, 1), 0);
            // So that the scalar doesn't get tidied up between executions.
            this.noiseShape = args.noiseShape;
            this.seed = args.seed;
            this.supportsMasking = true;
        }
        getNoiseShape(input) {
            if (this.noiseShape == null) {
                return this.noiseShape;
            }
            const inputShape = input.shape;
            const noiseShape = [];
            for (let i = 0; i < this.noiseShape.length; ++i) {
                noiseShape.push(this.noiseShape[i] == null ? inputShape[i] : this.noiseShape[i]);
            }
            return noiseShape;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                const input = getExactlyOneTensor(inputs);
                if (0 < this.rate && this.rate < 1) {
                    const training = kwargs['training'] == null ? false : kwargs['training'];
                    const noiseShape = this.getNoiseShape(input);
                    const output = inTrainPhase(() => dropout$1(input, this.rate, noiseShape, this.seed), () => input, training);
                    return output;
                }
                return inputs;
            });
        }
        getConfig() {
            const config = {
                rate: this.rate,
                noiseShape: this.noiseShape,
                seed: this.seed,
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
        dispose() {
            return super.dispose();
        }
    }
    /** @nocollapse */
    Dropout.className = 'Dropout';
    registerClass(Dropout);
    class SpatialDropout1D extends Dropout {
        constructor(args) {
            super(args);
            this.inputSpec = [{ ndim: 3 }];
        }
        getNoiseShape(input) {
            const inputShape = input.shape;
            return [inputShape[0], 1, inputShape[2]];
        }
    }
    /** @nocollapse */
    SpatialDropout1D.className = 'SpatialDropout1D';
    registerClass(SpatialDropout1D);
    class Dense extends Layer {
        constructor(args) {
            super(args);
            // Default activation: Linear (none).
            this.activation = null;
            this.useBias = true;
            this.kernel = null;
            this.bias = null;
            this.DEFAULT_KERNEL_INITIALIZER = 'glorotNormal';
            this.DEFAULT_BIAS_INITIALIZER = 'zeros';
            if (args.batchInputShape == null && args.inputShape == null &&
                args.inputDim != null) {
                // This logic is copied from Layer's constructor, since we can't
                // do exactly what the Python constructor does for Dense().
                let batchSize = null;
                if (args.batchSize != null) {
                    batchSize = args.batchSize;
                }
                this.batchInputShape = [batchSize, args.inputDim];
            }
            this.units = args.units;
            assertPositiveInteger(this.units, 'units');
            this.activation = getActivation(args.activation);
            if (args.useBias != null) {
                this.useBias = args.useBias;
            }
            this.kernelInitializer = getInitializer(args.kernelInitializer || this.DEFAULT_KERNEL_INITIALIZER);
            this.biasInitializer =
                getInitializer(args.biasInitializer || this.DEFAULT_BIAS_INITIALIZER);
            this.kernelConstraint = getConstraint(args.kernelConstraint);
            this.biasConstraint = getConstraint(args.biasConstraint);
            this.kernelRegularizer = getRegularizer(args.kernelRegularizer);
            this.biasRegularizer = getRegularizer(args.biasRegularizer);
            this.activityRegularizer = getRegularizer(args.activityRegularizer);
            this.supportsMasking = true;
            this.inputSpec = [{ minNDim: 2 }];
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const inputLastDim = inputShape[inputShape.length - 1];
            if (this.kernel == null) {
                this.kernel = this.addWeight('kernel', [inputLastDim, this.units], null, this.kernelInitializer, this.kernelRegularizer, true, this.kernelConstraint);
                if (this.useBias) {
                    this.bias = this.addWeight('bias', [this.units], null, this.biasInitializer, this.biasRegularizer, true, this.biasConstraint);
                }
            }
            this.inputSpec = [{ minNDim: 2, axes: { [-1]: inputLastDim } }];
            this.built = true;
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const outputShape = inputShape.slice();
            outputShape[outputShape.length - 1] = this.units;
            return outputShape;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                // Dense layer accepts only a single input.
                const input = getExactlyOneTensor(inputs);
                const fusedActivationName = mapActivationToFusedKernel(this.activation.getClassName());
                let output;
                if (fusedActivationName != null) {
                    output = dot$1(input, this.kernel.read(), fusedActivationName, this.bias ? this.bias.read() : null);
                }
                else {
                    output = dot$1(input, this.kernel.read());
                    if (this.bias != null) {
                        output = biasAdd(output, this.bias.read());
                    }
                    if (this.activation != null) {
                        output = this.activation.apply(output);
                    }
                }
                return output;
            });
        }
        getConfig() {
            const config = {
                units: this.units,
                activation: serializeActivation(this.activation),
                useBias: this.useBias,
                kernelInitializer: serializeInitializer(this.kernelInitializer),
                biasInitializer: serializeInitializer(this.biasInitializer),
                kernelRegularizer: serializeRegularizer(this.kernelRegularizer),
                biasRegularizer: serializeRegularizer(this.biasRegularizer),
                activityRegularizer: serializeRegularizer(this.activityRegularizer),
                kernelConstraint: serializeConstraint(this.kernelConstraint),
                biasConstraint: serializeConstraint(this.biasConstraint)
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Dense.className = 'Dense';
    registerClass(Dense);
    class Flatten extends Layer {
        constructor(args) {
            args = args || {};
            super(args);
            this.inputSpec = [{ minNDim: 3 }];
            this.dataFormat = args.dataFormat;
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            for (const dim of inputShape.slice(1)) {
                if (dim == null) {
                    throw new ValueError(`The shape of the input to "Flatten" is not fully defined ` +
                        `(got ${inputShape.slice(1)}). Make sure to pass a complete ` +
                        `"input_shape" or "batch_input_shape" argument to the first ` +
                        `layer in your model.`);
                }
            }
            return [inputShape[0], arrayProd(inputShape, 1)];
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                let input = getExactlyOneTensor(inputs);
                if (this.dataFormat === 'channelsFirst' && input.rank > 1) {
                    const permutation = [0];
                    for (let i = 2; i < input.rank; ++i) {
                        permutation.push(i);
                    }
                    permutation.push(1);
                    input = transpose$2(input, permutation);
                }
                return batchFlatten(input);
            });
        }
        getConfig() {
            const config = {};
            if (this.dataFormat != null) {
                config['dataFormat'] = this.dataFormat;
            }
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Flatten.className = 'Flatten';
    registerClass(Flatten);
    class Activation extends Layer {
        constructor(args) {
            super(args);
            this.supportsMasking = true;
            this.activation = getActivation(args.activation);
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                const input = getExactlyOneTensor(inputs);
                return this.activation.apply(input);
            });
        }
        getConfig() {
            const config = { activation: serializeActivation(this.activation) };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Activation.className = 'Activation';
    registerClass(Activation);
    class RepeatVector extends Layer {
        constructor(args) {
            super(args);
            this.n = args.n;
            this.inputSpec = [{ ndim: 2 }];
        }
        computeOutputShape(inputShape) {
            return [inputShape[0], this.n, inputShape[1]];
        }
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = getExactlyOneTensor(inputs);
                return repeat(inputs, this.n);
            });
        }
        getConfig() {
            const config = {
                n: this.n,
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    RepeatVector.className = 'RepeatVector';
    registerClass(RepeatVector);
    class Reshape extends Layer {
        constructor(args) {
            super(args);
            this.targetShape = args.targetShape;
            // Make sure that all unknown dimensions are represented as `null`.
            for (let i = 0; i < this.targetShape.length; ++i) {
                if (this.isUnknown(this.targetShape[i])) {
                    this.targetShape[i] = null;
                }
            }
        }
        isUnknown(dim) {
            return dim < 0 || dim == null;
        }
        /**
         * Finds and replaces a missing dimension in output shape.
         *
         * This is a near direct port of the internal Numpy function
         * `_fix_unknown_dimension` in `numpy/core/src/multiarray/shape.c`.
         *
         * @param inputShape: Original shape of array begin reshape.
         * @param outputShape: Target shape of the array, with at most a single
         * `null` or negative number, which indicates an underdetermined dimension
         * that should be derived from `inputShape` and the known dimensions of
         *   `outputShape`.
         * @returns: The output shape with `null` replaced with its computed value.
         * @throws: ValueError: If `inputShape` and `outputShape` do not match.
         */
        fixUnknownDimension(inputShape, outputShape) {
            const errorMsg = 'Total size of new array must be unchanged.';
            const finalShape = outputShape.slice();
            let known = 1;
            let unknown = null;
            for (let i = 0; i < finalShape.length; ++i) {
                const dim = finalShape[i];
                if (this.isUnknown(dim)) {
                    if (unknown === null) {
                        unknown = i;
                    }
                    else {
                        throw new ValueError('Can only specifiy one unknown dimension.');
                    }
                }
                else {
                    known *= dim;
                }
            }
            const originalSize = arrayProd(inputShape);
            if (unknown !== null) {
                if (known === 0 || originalSize % known !== 0) {
                    throw new ValueError(errorMsg);
                }
                finalShape[unknown] = originalSize / known;
            }
            else if (originalSize !== known) {
                throw new ValueError(errorMsg);
            }
            return finalShape;
        }
        computeOutputShape(inputShape) {
            let anyUnknownDims = false;
            for (let i = 0; i < inputShape.length; ++i) {
                if (this.isUnknown(inputShape[i])) {
                    anyUnknownDims = true;
                    break;
                }
            }
            if (anyUnknownDims) {
                return inputShape.slice(0, 1).concat(this.targetShape);
            }
            else {
                return inputShape.slice(0, 1).concat(this.fixUnknownDimension(inputShape.slice(1), this.targetShape));
            }
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                const input = getExactlyOneTensor(inputs);
                const inputShape = input.shape;
                const outputShape = inputShape.slice(0, 1).concat(this.fixUnknownDimension(inputShape.slice(1), this.targetShape));
                return reshape$3(input, outputShape);
            });
        }
        getConfig() {
            const config = {
                targetShape: this.targetShape,
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Reshape.className = 'Reshape';
    registerClass(Reshape);
    class Permute extends Layer {
        constructor(args) {
            super(args);
            if (args.dims == null) {
                throw new Error('Required configuration field `dims` is missing during Permute ' +
                    'constructor call.');
            }
            if (!Array.isArray(args.dims)) {
                throw new Error('Permute constructor requires `dims` to be an Array, but received ' +
                    `${args.dims} instead.`);
            }
            // Check the validity of the permutation indices.
            const expectedSortedIndices = range$3(1, args.dims.length + 1);
            if (!arraysEqual(args.dims.slice().sort(), expectedSortedIndices)) {
                throw new Error('Invalid permutation `dims`: ' + JSON.stringify(args.dims) +
                    ' `dims` must contain consecutive integers starting from 1.');
            }
            this.dims = args.dims;
            this.dimsIncludingBatch = [0].concat(this.dims);
            this.inputSpec = [new InputSpec({ ndim: this.dims.length + 1 })];
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const outputShape = inputShape.slice();
            this.dims.forEach((dim, i) => {
                outputShape[i + 1] = inputShape[dim];
            });
            return outputShape;
        }
        call(inputs, kwargs) {
            return transpose$2(getExactlyOneTensor(inputs), this.dimsIncludingBatch);
        }
        getConfig() {
            const config = {
                dims: this.dims,
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Permute.className = 'Permute';
    registerClass(Permute);
    class Masking extends Layer {
        constructor(args) {
            super(args == null ? {} : args);
            this.supportsMasking = true;
            if (args != null) {
                this.maskValue = args.maskValue == null ? 0 : args.maskValue;
            }
            else {
                this.maskValue = 0;
            }
        }
        computeOutputShape(inputShape) {
            return inputShape;
        }
        getConfig() {
            const baseConfig = super.getConfig();
            const config = { maskValue: this.maskValue };
            Object.assign(config, baseConfig);
            return config;
        }
        computeMask(inputs, mask) {
            const input = getExactlyOneTensor(inputs);
            const axis = -1;
            return any$2(notEqual$3(input, this.maskValue), axis);
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                const input = getExactlyOneTensor(inputs);
                const axis = -1;
                const keepDims = true;
                const booleanMask = any$2(notEqual$3(input, this.maskValue), axis, keepDims);
                const output = mul(input, cast$3(booleanMask, input.dtype));
                return output;
            });
        }
    }
    /** @nocollapse */
    Masking.className = 'Masking';
    registerClass(Masking);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    class Embedding extends Layer {
        constructor(args) {
            super(args);
            this.embeddings = null;
            this.DEFAULT_EMBEDDINGS_INITIALIZER = 'randomUniform';
            if (args.batchInputShape == null && args.inputShape == null) {
                // Porting Note: This logic is copied from Layer's constructor, since we
                // can't do exactly what the Python constructor does for Embedding().
                // Specifically, the super constructor can not be called after the
                // mutation of the `config` argument.
                let batchSize = null;
                if (args.batchSize != null) {
                    batchSize = args.batchSize;
                }
                if (args.inputLength == null) {
                    // Fix super-constructor to what it would have done if
                    // 'config.inputShape' were (None, )
                    this.batchInputShape = [batchSize, null];
                }
                else {
                    // Fix super-constructor to what it would have done if
                    // 'config.inputShape' were (config.inputLength, )
                    this.batchInputShape =
                        [batchSize].concat(toList(args.inputLength));
                }
            }
            this.inputDim = args.inputDim;
            assertPositiveInteger(this.inputDim, 'inputDim');
            this.outputDim = args.outputDim;
            assertPositiveInteger(this.outputDim, 'outputDim');
            this.embeddingsInitializer = getInitializer(args.embeddingsInitializer || this.DEFAULT_EMBEDDINGS_INITIALIZER);
            this.embeddingsRegularizer = getRegularizer(args.embeddingsRegularizer);
            this.activityRegularizer = getRegularizer(args.activityRegularizer);
            this.embeddingsConstraint = getConstraint(args.embeddingsConstraint);
            this.maskZero = args.maskZero;
            this.supportsMasking = args.maskZero;
            this.inputLength = args.inputLength;
        }
        build(inputShape) {
            this.embeddings = this.addWeight('embeddings', [this.inputDim, this.outputDim], this.dtype, this.embeddingsInitializer, this.embeddingsRegularizer, true, this.embeddingsConstraint);
            this.built = true;
        }
        // Override warnOnIncompatibleInputShape because an embedding layer allows
        // the input to have varying ranks.
        warnOnIncompatibleInputShape(inputShape) { }
        computeMask(inputs, mask) {
            return tidy(() => {
                if (!this.maskZero) {
                    return null;
                }
                else {
                    inputs = getExactlyOneTensor(inputs);
                    return notEqual$3(inputs, zerosLike$2(inputs));
                }
            });
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            if (this.inputLength == null) {
                return [...inputShape, this.outputDim];
            }
            // inputLength can be an array if input is 3D or higher.
            const inLens = toList(this.inputLength);
            if (inLens.length !== inputShape.length - 1) {
                throw new ValueError(`"inputLength" is ${this.inputLength}, but received ` +
                    `input shape has shape ${inputShape}`);
            }
            else {
                let i = 0;
                for (let k = 0; k < inLens.length; ++k) {
                    const s1 = inLens[k];
                    const s2 = inputShape[k + 1];
                    if ((s1 != null) && (s2 != null) && (s1 !== s2)) {
                        throw new ValueError(`"inputLength" is ${this.inputLength}, but received ` +
                            `input shape has shape ${inputShape}`);
                    }
                    else if (s1 == null) {
                        inLens[i] = s2;
                    }
                    i++;
                }
            }
            return [inputShape[0], ...inLens, this.outputDim];
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                // Embedding layer accepts only a single input.
                let input = getExactlyOneTensor(inputs);
                if (input.dtype !== 'int32') {
                    input = cast$2(input, 'int32');
                }
                const output = gather(this.embeddings.read(), reshape$3(input, [input.size]));
                return reshape$3(output, getExactlyOneShape(this.computeOutputShape(input.shape)));
            });
        }
        getConfig() {
            const config = {
                inputDim: this.inputDim,
                outputDim: this.outputDim,
                embeddingsInitializer: serializeInitializer(this.embeddingsInitializer),
                embeddingsRegularizer: serializeRegularizer(this.embeddingsRegularizer),
                activityRegularizer: serializeRegularizer(this.activityRegularizer),
                embeddingsConstraint: serializeConstraint(this.embeddingsConstraint),
                maskZero: this.maskZero,
                inputLength: this.inputLength
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Embedding.className = 'Embedding';
    registerClass(Embedding);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Generic Merge layer for element-wise merge functions.
     *
     * Used to implement `Sum`, `Average`, `Concatenate`, etc.
     */
    class Merge extends Layer {
        constructor(args) {
            super(args || {});
            this.supportsMasking = true;
        }
        /**
         * Logic for merging multiple tensors, to be overridden by subclasses.
         * @param inputs
         */
        mergeFunction(inputs) {
            throw new NotImplementedError();
        }
        /**
         * Computes the shape of the result of an elementwise operation.
         *
         * @param shape1: Shape of the first tensor.
         * @param shape2: Shape of the second tensor.
         * @returns Expected output shape when an elementwise operation is carried
         *   out on 2 tensors with shapes `shape1` and `shape2`.
         * @throws ValueError: If `shape1` and `shape2` are not compatible for
         *   element-wise operations.
         */
        computeElementwiseOpOutputShape(shape1, shape2) {
            if (shape1 == null || shape2 == null) {
                return null;
            }
            else if (shape1.length < shape2.length) {
                return this.computeElementwiseOpOutputShape(shape2, shape1);
            }
            else if (shape2.length === 0) {
                return shape1;
            }
            const outputShape = shape1.slice(0, shape1.length - shape2.length);
            for (let k = 0; k < shape2.length; ++k) {
                const i = shape1[shape1.length - shape2.length + k];
                const j = shape2[k];
                if (i == null || j == null || i < 0 || j < 0) {
                    outputShape.push(null);
                }
                else if (i === 1) {
                    outputShape.push(j);
                }
                else if (j === 1) {
                    outputShape.push(i);
                }
                else {
                    if (i !== j) {
                        throw new ValueError('Operands could not be broadcast together with shapes ' +
                            JSON.stringify(shape1) + ' ' + JSON.stringify(shape2));
                    }
                    outputShape.push(i);
                }
            }
            return outputShape;
        }
        build(inputShape) {
            // Used purely for shape validation.
            if (Array.isArray(inputShape) && !Array.isArray(inputShape[0])) {
                // Make sure that inputShape is an Array of shape.
                inputShape = [getExactlyOneShape(inputShape)];
            }
            inputShape = inputShape;
            if (inputShape.length < 2) {
                throw new ValueError('A merge layer should be called on an Array of at least 2 inputs.' +
                    ` Got ${inputShape.length} input(s).`);
            }
            // Make sure that there is at most one unique batch size among the input
            // shapes.
            let batchSizes = [];
            for (const shape of inputShape) {
                if (shape != null && shape[0] !== null) {
                    batchSizes.push(shape[0]);
                }
            }
            batchSizes = unique$2(batchSizes);
            if (batchSizes.length > 1) {
                throw new ValueError(`Can not merge tensors with different batch sizes. ` +
                    `Got tensors with shapes: ${JSON.stringify(inputShape)}.`);
            }
            let outputShape = inputShape[0] == null ? null : inputShape[0].slice(1);
            for (let i = 1; i < inputShape.length; ++i) {
                const shape = inputShape[i] == null ? null : inputShape[i].slice(1);
                outputShape = this.computeElementwiseOpOutputShape(outputShape, shape);
            }
            // If the inputs have different ranks, we have to reshape them to make them
            // broadcastable.
            const allRanks = inputShape.map(shape => shape.length);
            if (inputShape.indexOf(null) === -1 &&
                unique$2(allRanks).length === 1) {
                this.reshapeRequired = false;
            }
            else {
                this.reshapeRequired = true;
            }
        }
        call(inputs, kwargs) {
            return tidy(() => {
                inputs = inputs;
                if (this.reshapeRequired) {
                    const reshapedInputs = [];
                    const inputDims = inputs.map(input => input.rank);
                    if (inputDims.indexOf(null) === -1) {
                        // If ranks of all inputs are available, we simply expand each of them
                        // at axis=1 until all of them have the same rank.
                        const maxNDim = max$2(inputDims);
                        for (let x of inputs) {
                            const xNDim = x.rank;
                            for (let k = 0; k < maxNDim - xNDim; ++k) {
                                x = expandDims$2(x, 1);
                            }
                            reshapedInputs.push(x);
                        }
                        return this.mergeFunction(reshapedInputs);
                    }
                    else {
                        // Transpose all inputs so that batch size is the last dimension.
                        // [batchSize, dim1, dim2, ...] -> [dim1, dim2, ..., batchSize]
                        let transposed = false;
                        for (const x of inputs) {
                            const xNDim = x.rank;
                            if (xNDim == null) {
                                const xShape = x.shape;
                                const batchSize = xShape[0];
                                const newShape = xShape.slice(1).concat([batchSize]);
                                let xTransposed = reshape$3(x, [batchSize].concat(arrayProd(xShape.slice(1))));
                                xTransposed = transpose$2(xTransposed, [1, 0]);
                                xTransposed = reshape$3(xTransposed, newShape);
                                reshapedInputs.push(xTransposed);
                                transposed = true;
                            }
                            else if (xNDim > 1) {
                                const dims = range$3(1, xNDim).concat([0]);
                                reshapedInputs.push(transpose$2(x, dims));
                                transposed = true;
                            }
                            else {
                                // We don't transpose inputs if they are 1D vectors or scalars.
                                reshapedInputs.push(x);
                            }
                        }
                        let y = this.mergeFunction(reshapedInputs);
                        const yNDim = y.rank;
                        if (transposed) {
                            // If inputs have been transposed, we have to transpose the output
                            // too.
                            if (yNDim == null) {
                                const yShape = y.shape;
                                const yNDim = yShape.length;
                                const batchSize = yShape[yNDim - 1];
                                const newShape = [batchSize].concat(yShape.slice(0, yShape.length - 1));
                                y = reshape$3(transpose$2(reshape$3(y, [-1, batchSize]), [1, 0]), newShape);
                            }
                            else if (yNDim > 1) {
                                const dims = [yNDim - 1].concat(range$3(0, yNDim - 1));
                                y = transpose$2(y, dims);
                            }
                        }
                        return y;
                    }
                }
                else {
                    return this.mergeFunction(inputs);
                }
            });
        }
        computeOutputShape(inputShape) {
            inputShape = inputShape;
            let outputShape;
            if (inputShape[0] == null) {
                outputShape = null;
            }
            else {
                outputShape = inputShape[0].slice(1);
            }
            for (let i = 1; i < inputShape.length; ++i) {
                const shape = inputShape[i] == null ? null : inputShape[i].slice(1);
                outputShape = this.computeElementwiseOpOutputShape(outputShape, shape);
            }
            let batchSizes = [];
            for (const shape of inputShape) {
                if (shape != null && shape[0] !== null) {
                    batchSizes.push(shape[0]);
                }
            }
            batchSizes = unique$2(batchSizes);
            if (batchSizes.length === 1) {
                outputShape = batchSizes.concat(outputShape);
            }
            else {
                outputShape = [null].concat(outputShape);
            }
            return outputShape;
        }
        computeMask(inputs, mask) {
            return tidy(() => {
                if (mask == null) {
                    return null;
                }
                if (!Array.isArray(mask)) {
                    throw new ValueError('`mask` should be an Array');
                }
                if (!Array.isArray(inputs)) {
                    throw new ValueError('`inputs` should be an Array');
                }
                if (mask.length !== inputs.length) {
                    throw new ValueError(`The Array 'inputs' and 'mask' are expected to have the same ` +
                        `length, but have different lengths ` +
                        `(${inputs.length} vs ${mask.length})`);
                }
                if (mask.every(m => m == null)) {
                    return null;
                }
                mask = mask.map(m => m == null ? m : expandDims$3(m, 0));
                let output = mask[0];
                for (let i = 1; i < mask.length - 1; ++i) {
                    output = logicalAnd$2(output, mask[i]);
                }
                return output;
            });
        }
    }
    class Add extends Merge {
        constructor(args) {
            super(args);
        }
        mergeFunction(inputs) {
            return tidy(() => {
                let output = inputs[0].clone();
                for (let i = 1; i < inputs.length; ++i) {
                    output = add$2(output, inputs[i]);
                }
                return output;
            });
        }
    }
    /** @nocollapse */
    Add.className = 'Add';
    registerClass(Add);
    class Multiply extends Merge {
        constructor(args) {
            super(args);
        }
        mergeFunction(inputs) {
            return tidy(() => {
                let output = inputs[0].clone();
                for (let i = 1; i < inputs.length; ++i) {
                    output = mul(output, inputs[i]);
                }
                return output;
            });
        }
    }
    /** @nocollapse */
    Multiply.className = 'Multiply';
    registerClass(Multiply);
    class Average extends Merge {
        constructor(args) {
            super(args);
        }
        mergeFunction(inputs) {
            return tidy(() => {
                let output = inputs[0].clone();
                for (let i = 1; i < inputs.length; ++i) {
                    output = add$2(output, inputs[i]);
                }
                return mul(1 / inputs.length, output);
            });
        }
    }
    /** @nocollapse */
    Average.className = 'Average';
    registerClass(Average);
    class Maximum extends Merge {
        constructor(args) {
            super(args);
        }
        mergeFunction(inputs) {
            return tidy(() => {
                let output = inputs[0];
                for (let i = 1; i < inputs.length; ++i) {
                    output = maximum$3(output, inputs[i]);
                }
                return output;
            });
        }
    }
    /** @nocollapse */
    Maximum.className = 'Maximum';
    registerClass(Maximum);
    class Minimum extends Merge {
        constructor(args) {
            super(args);
        }
        mergeFunction(inputs) {
            return tidy(() => {
                let output = inputs[0];
                for (let i = 1; i < inputs.length; ++i) {
                    output = minimum$3(output, inputs[i]);
                }
                return output;
            });
        }
    }
    /** @nocollapse */
    Minimum.className = 'Minimum';
    registerClass(Minimum);
    class Concatenate extends Merge {
        constructor(args) {
            super(args);
            this.DEFAULT_AXIS = -1;
            if (args == null) {
                args = {};
            }
            this.axis = args.axis == null ? this.DEFAULT_AXIS : args.axis;
            this.supportsMasking = true;
            this.reshapeRequired = false;
        }
        build(inputShape) {
            // Used purely for shape validation.]
            if (!(Array.isArray(inputShape) && Array.isArray(inputShape[0])) ||
                inputShape.length === 1) {
                throw new ValueError('A `Concatenate` layer should be called on a list of at least 2 ' +
                    'inputs');
            }
            inputShape = inputShape;
            let allNoneShape = true;
            for (const shape of inputShape) {
                if (shape != null) {
                    allNoneShape = false;
                    break;
                }
            }
            if (allNoneShape) {
                return;
            }
            const shapeSet = [];
            for (let i = 0; i < inputShape.length; ++i) {
                const shapeWithoutConcatAxis = inputShape[i].slice();
                shapeWithoutConcatAxis.splice(this.axis, 1);
                let exists = false;
                for (const shape of shapeSet) {
                    if (arraysEqual(shape, shapeWithoutConcatAxis)) {
                        exists = true;
                        break;
                    }
                }
                if (!exists) {
                    shapeSet.push(shapeWithoutConcatAxis);
                }
            }
            if (shapeSet.length > 1) {
                throw new ValueError('A `Concatenate` layer requires inputs with matching shapes ' +
                    'except for the concat axis. Got input shapes: ' +
                    JSON.stringify(inputShape));
            }
        }
        mergeFunction(inputs) {
            return tidy(() => {
                return concatenate$1(inputs, this.axis);
            });
        }
        computeOutputShape(inputShape) {
            if (!(Array.isArray(inputShape) && Array.isArray(inputShape[0]))) {
                throw new ValueError('A `Concatenate` layer should be called on a list of inputs.');
            }
            const inputShapes = inputShape;
            const outputShape = inputShapes[0].slice();
            const axis = this.axis < 0 ? outputShape.length + this.axis : this.axis;
            // Porting Note: the line above is because TypeScript doesn't support
            //   negative indices.
            for (const shape of inputShapes.slice(1)) {
                if (outputShape[axis] == null || shape[axis] == null) {
                    outputShape[axis] = null;
                    break;
                }
                outputShape[axis] += shape[axis];
            }
            return outputShape;
        }
        computeMask(inputs, mask) {
            if (mask == null) {
                return null;
            }
            if (!Array.isArray(mask)) {
                throw new ValueError('`mask` should be an array for Concatenate');
            }
            if (!Array.isArray(inputs)) {
                throw new ValueError('`inputs` should be an array for Concatenate');
            }
            if (mask.length !== inputs.length) {
                throw new ValueError(`Mismatch in the length of mask (${mask.length}) ` +
                    `and the legnth of inputs (${inputs.length})`);
            }
            return tidy(() => {
                let allNullMasks = true;
                mask.forEach(m => {
                    if (m != null) {
                        allNullMasks = false;
                        return;
                    }
                });
                if (allNullMasks) {
                    return null;
                }
                const outputMasks = [];
                for (let i = 0; i < inputs.length; ++i) {
                    if (mask[i] == null) {
                        // Input is unmasked. Append all 1's to masks.
                        outputMasks.push(cast$3(onesLike$2(inputs[i]), 'bool'));
                    }
                    else if (mask[i].rank < inputs[i].rank) {
                        // Mask is smaller than the input, expand it.
                        outputMasks.push(expandDims$3(mask[i], -1));
                    }
                    else {
                        outputMasks.push(mask[i]);
                    }
                }
                const concatenatedMasks = concat$2(outputMasks, this.axis);
                return all$2(concatenatedMasks, -1, false);
            });
        }
        getConfig() {
            const config = {
                'axis': this.axis,
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Concatenate.className = 'Concatenate';
    registerClass(Concatenate);
    /**
     * Interpretable potentially negative axis index.
     *
     * For example, given axis = -1, and dim = 3, this function will return 2.
     *
     * @param axis The axis index, may be a positive, zero or negative integer.
     * @param dim Total number of dimensions, a positive integer.
     * @returns A non-negative axis index equivalent to the input `axis`.
     */
    function interpretAxis(axis, dim) {
        while (axis < 0) {
            axis += dim;
        }
        return axis;
    }
    function batchDot(x, y, axes) {
        if (x.shape.length > 3 || y.shape.length > 3) {
            throw new NotImplementedError('batchDot is not implemented for tensors of 4D or higher rank yet');
        }
        assert$4(x.shape.length >= 2, () => `batchDot requires the rank of x to be >= 2, ` +
            `but got ${x.shape.length}`);
        assert$4(x.shape.length >= 2, () => `batchDot requires the rank of y to be >= 2, ` +
            `but got ${y.shape.length}`);
        if (typeof axes === 'number') {
            axes = [axes, axes];
        }
        if (x.dtype === 'complex64' || y.dtype === 'complex64') {
            throw new NotImplementedError('batchDot is not implemented for complex64-type Tensors yet.');
        }
        const xNDim = x.shape.length;
        const yNDim = y.shape.length;
        if (axes == null) {
            // Behave like batchMatmul by default.
            axes = [xNDim - 1, yNDim - 2];
        }
        const axesArray = axes;
        return tidy(() => {
            let diff;
            if (xNDim > yNDim) {
                diff = xNDim - yNDim;
                const diffShape = [];
                for (let i = 0; i < diff; ++i) {
                    diffShape.push(1);
                }
                y = reshape$3(y, y.shape.concat(diffShape));
            }
            else if (yNDim > xNDim) {
                diff = yNDim - xNDim;
                const diffShape = [];
                for (let i = 0; i < diff; ++i) {
                    diffShape.push(1);
                }
                x = reshape$3(x, x.shape.concat(diffShape));
            }
            else {
                diff = 0;
            }
            let out;
            if (x.shape.length === 2 && y.shape.length === 2) {
                if (axesArray[0] === axesArray[1]) {
                    out = sum$2(mul(x, y), axesArray[0]);
                }
                else {
                    out = sum$2(mul(transpose$2(x, [1, 0]), y), axesArray[1]);
                }
            }
            else {
                const adjX = axesArray[0] !== x.shape.length - 1;
                const adjY = axesArray[1] === y.shape.length - 1;
                out = matMul$1(x, y, adjX, adjY);
            }
            if (diff > 0) {
                let idx;
                if (xNDim > yNDim) {
                    idx = xNDim + yNDim - 3;
                }
                else {
                    idx = xNDim - 1;
                }
                const squeezeAxes = [];
                for (let i = idx; i < idx + diff; ++i) {
                    squeezeAxes.push(i);
                }
                out = squeeze(out, squeezeAxes);
            }
            if (out.shape.length === 1) {
                out = expandDims$3(out, 1);
            }
            return out;
        });
    }
    class Dot extends Merge {
        constructor(args) {
            super(args);
            this.axes = args.axes;
            this.normalize = args.normalize == null ? false : args.normalize;
            this.supportsMasking = true;
            this.reshapeRequired = false;
        }
        build(inputShape) {
            assert$4(Array.isArray(inputShape) && inputShape.length === 2 &&
                Array.isArray(inputShape[0]) && Array.isArray(inputShape[1]), () => 'A `Dot` layer should be called on a list of exactly 2 inputs.');
            const shape1 = inputShape[0];
            const shape2 = inputShape[1];
            if (shape1.length > 3 || shape2.length > 3) {
                throw new NotImplementedError('Dot layer does not support tensors of 4D or higher rank yet.');
            }
            const axes = this.interpretAxes(shape1, shape2);
            if (shape1[axes[0]] !== shape2[axes[1]]) {
                throw new ValueError(`Dimension incompatibility: ` +
                    `${shape1[axes[0]]} !== ${shape2[axes[1]]}`);
            }
        }
        mergeFunction(inputs) {
            if (inputs.length !== 2) {
                throw new ValueError('A `Dot` layer must be called on exactly 2 inputs, ' +
                    `but received ${inputs.length} input(s).`);
            }
            let x1 = inputs[0];
            let x2 = inputs[1];
            let axes;
            if (!Array.isArray(this.axes)) {
                axes = [
                    interpretAxis(this.axes, x1.shape.length),
                    interpretAxis(this.axes, x2.shape.length)
                ];
            }
            else {
                axes = this.axes.map((axis, i) => interpretAxis(axis, inputs[i].shape.length));
            }
            if (this.normalize) {
                x1 = l2Normalize(x1, axes[0]);
                x2 = l2Normalize(x2, axes[1]);
            }
            return batchDot(x1, x2, axes);
        }
        interpretAxes(shape1, shape2) {
            let axes;
            if (!Array.isArray(this.axes)) {
                // `this.axes` is a single integer.
                axes = [
                    interpretAxis(this.axes, shape1.length),
                    interpretAxis(this.axes, shape2.length)
                ];
            }
            else {
                // `this.axes` is an Array of integers.
                axes = this.axes;
            }
            return axes;
        }
        computeOutputShape(inputShape) {
            assert$4(Array.isArray(inputShape) && inputShape.length === 2 &&
                Array.isArray(inputShape[0]) && Array.isArray(inputShape[1]), () => 'A `Dot` layer should be called on a list of exactly 2 inputs.');
            const shape1 = inputShape[0].slice();
            const shape2 = inputShape[1].slice();
            if (shape1.length > 3 || shape2.length > 3) {
                throw new NotImplementedError('Dot layer does not support tensors of 4D or higher rank yet.');
            }
            const axes = this.interpretAxes(shape1, shape2);
            shape1.splice(axes[0], 1);
            shape2.splice(axes[1], 1);
            shape2.splice(0, 1);
            const outputShape = shape1.concat(shape2);
            if (outputShape.length === 1) {
                outputShape.push(1);
            }
            return outputShape;
        }
        computeMask(inputs, mask) {
            return null;
        }
        getConfig() {
            const config = {
                'axes': this.axes,
                'normalize': this.normalize
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    Dot.className = 'Dot';
    registerClass(Dot);
    // TODO(cais): Add functional interfaces for the merge layers.

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    class GaussianNoise extends Layer {
        constructor(args) {
            super(args);
            this.supportsMasking = true;
            this.stddev = args.stddev;
        }
        computeOutputShape(inputShape) {
            return inputShape;
        }
        getConfig() {
            const baseConfig = super.getConfig();
            const config = { stddev: this.stddev };
            Object.assign(config, baseConfig);
            return config;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                const input = getExactlyOneTensor(inputs);
                const noised = () => add$2(randomNormal$1(input.shape, 0, this.stddev), input);
                const output = inTrainPhase(noised, () => input, kwargs['training'] || false);
                return output;
            });
        }
    }
    /** @nocollapse */
    GaussianNoise.className = 'GaussianNoise';
    registerClass(GaussianNoise);
    class GaussianDropout extends Layer {
        constructor(args) {
            super(args);
            this.supportsMasking = true;
            this.rate = args.rate;
        }
        computeOutputShape(inputShape) {
            return inputShape;
        }
        getConfig() {
            const baseConfig = super.getConfig();
            const config = { rate: this.rate };
            Object.assign(config, baseConfig);
            return config;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                const input = getExactlyOneTensor(inputs);
                if (this.rate > 0 && this.rate < 1) {
                    const noised = () => {
                        const stddev = Math.sqrt(this.rate / (1 - this.rate));
                        return mul(input, randomNormal$1(input.shape, 1, stddev));
                    };
                    return inTrainPhase(noised, () => input, kwargs['training'] || false);
                }
                return input;
            });
        }
    }
    /** @nocollapse */
    GaussianDropout.className = 'GaussianDropout';
    registerClass(GaussianDropout);
    /**
     * Applies Alpha Dropout to the input.
     *
     * As it is a regularization layer, it is only active at training time.
     *
     * Alpha Dropout is a `Dropout` that keeps mean and variance of inputs
     * to their original values, in order to ensure the self-normalizing property
     * even after this dropout.
     * Alpha Dropout fits well to Scaled Exponential Linear Units
     * by randomly setting activations to the negative saturation value.
     *
     * Arguments:
     *   - `rate`: float, drop probability (as with `Dropout`).
     *     The multiplicative noise will have
     *     standard deviation `sqrt(rate / (1 - rate))`.
     *   - `noise_shape`: A 1-D `Tensor` of type `int32`, representing the
     *     shape for randomly generated keep/drop flags.
     *
     * Input shape:
     *   Arbitrary. Use the keyword argument `inputShape`
     *   (tuple of integers, does not include the samples axis)
     *   when using this layer as the first layer in a model.
     *
     * Output shape:
     *   Same shape as input.
     *
     * References:
     *   - [Self-Normalizing Neural Networks](https://arxiv.org/abs/1706.02515)
     */
    class AlphaDropout extends Layer {
        constructor(args) {
            super(args);
            this.supportsMasking = true;
            this.rate = args.rate;
            this.noiseShape = args.noiseShape;
        }
        _getNoiseShape(inputs) {
            return this.noiseShape || getExactlyOneTensor(inputs).shape;
        }
        computeOutputShape(inputShape) {
            return inputShape;
        }
        getConfig() {
            const baseConfig = super.getConfig();
            const config = { rate: this.rate };
            Object.assign(config, baseConfig);
            return config;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                if (this.rate < 1 && this.rate > 0) {
                    const noiseShape = this._getNoiseShape(inputs);
                    const droppedInputs = () => {
                        const input = getExactlyOneTensor(inputs);
                        const alpha = 1.6732632423543772848170429916717;
                        const scale = 1.0507009873554804934193349852946;
                        const alphaP = -alpha * scale;
                        let keptIdx = greaterEqual$2(randomUniform$1(noiseShape), this.rate);
                        keptIdx = cast$2(keptIdx, 'float32'); // get default dtype.
                        // Get affine transformation params.
                        const a = ((1 - this.rate) * (1 + this.rate * alphaP ** 2)) ** -0.5;
                        const b = -a * alphaP * this.rate;
                        // Apply mask.
                        const x = add$2(mul(input, keptIdx), mul(add$2(keptIdx, -1), alphaP));
                        return add$2(mul(x, a), b);
                    };
                    return inTrainPhase(droppedInputs, () => getExactlyOneTensor(inputs), kwargs['training'] || false);
                }
                return inputs;
            });
        }
    }
    /** @nocollapse */
    AlphaDropout.className = 'AlphaDropout';
    registerClass(AlphaDropout);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Applies batch normalization on x given mean, var, beta and gamma.
     *
     * I.e. returns:
     *   `output = (x - mean) / (sqrt(var) + epsilon) * gamma + beta`
     *
     * @param x Input tensor.
     * @param mean Mean of batch.
     * @param variance Variance of batch.
     * @param beta Tensor with which to center the input.
     * @param gamma Tensor by which to scale the input.
     * @param epsilon Fuzz factor.
     * @returns The result of the batch normalization.
     */
    function batchNormalization$1(x, mean, variance, beta, gamma, epsilon = 1e-3) {
        let out;
        if (x.rank === 2) {
            out = batchNorm2d(x, mean, variance, beta, gamma, epsilon);
        }
        else if (x.rank === 3) {
            // TODO(cais): Check rank; give proper error message.
            out = batchNorm3d(x, mean, variance, beta, gamma, epsilon);
        }
        else if (x.rank === 4) {
            out = batchNorm4d(x, mean, variance, beta, gamma, epsilon);
        }
        else {
            throw new NotImplementedError(`batchNormalization is not implemented for array of rank ${x.rank} ` +
                `yet`);
        }
        return out;
    }
    /**
     * Non-broadcasting batch normalization for use in training (not inference).
     *
     * The input is normalized to zero mean and unit variance along the
     * `reductionAxes`, followed by scaling with `gamma` and shifted by `beta`.
     * The result of that is returned as the first element
     * of the returned `Array`. The other two elements are the mean and variance,
     * respectively.
     *
     * @param x Input tensor to be normalized.
     * @param gamma Tensor by which to scale the input.
     * @param beta Tensor by which to center the input.
     * @param reductionAxes Axes over which to normalize.
     * @param epsilon Fuzz factor.
     * @returns An `Array` of three `Tensors`:
     *   [normalized tensor, mean of input, variance of input].
     */
    function regularNormalizeBatchInTraining(x, gamma, beta, reductionAxes, epsilon = 1e-3) {
        return tidy(() => {
            const meanAndVariance = moments(x, reductionAxes);
            const mean = meanAndVariance.mean;
            const variance = meanAndVariance.variance;
            const normed = batchNormalization$1(x, mean, variance, beta, gamma, epsilon);
            return [normed, mean, variance];
        });
    }
    /**
     * Broadcasting batch normalization for use in training (not inference).
     *
     * The input is normalized to zero mean and unit variance along the
     * `reductionAxes`, followed by scaling with `gamma` and shifted by `beta`.
     * The result of that is returned as the first element
     * of the returned `Array`. The other two elements are the mean and variance,
     * respectively.
     *
     * @param x Input tensor to be normalized.
     * @param gamma Tensor by which to scale the input.
     * @param beta Tensor by which to center the input.
     * @param reductionAxes Axes over which to normalize.
     * @param epsilon Fuzz factor.
     * @returns An `Array` of three `Tensors`:
     *   [normalized tensor, mean of input, variance of input].
     */
    function broadcastNormalizeBatchInTraining(x, gamma, beta, reductionAxes, epsilon = 1e-3) {
        return tidy(() => {
            const meanAndVariance = moments(x, reductionAxes);
            const mean = meanAndVariance.mean;
            const variance = meanAndVariance.variance;
            const targetShape = [];
            for (const axis of range$3(0, x.rank)) {
                if (reductionAxes.indexOf(axis) !== -1) {
                    targetShape.push(1);
                }
                else {
                    targetShape.push(x.shape[axis]);
                }
            }
            const broadcastMean = reshape$3(mean, targetShape);
            const broadcastVariance = reshape$3(variance, targetShape);
            const broadcastGamma = gamma == null ? null : reshape$3(gamma, targetShape);
            const broadcastBeta = beta == null ? null : reshape$3(beta, targetShape);
            const normed = batchNormalization$1(x, broadcastMean, broadcastVariance, broadcastBeta, broadcastGamma, epsilon);
            return [normed, mean, variance];
        });
    }
    /**
     * Batch normalization for use in training (not inference).
     *
     * @param x Input tensor to be normalized.
     * @param gamma Tensor by which to scale the input.
     * @param beta Tensor by which to center the input.
     * @param reductionAxes Axes over which to normalize.
     * @param epsilon Fuzz factor.
     * @returns An `Array` of three `Tensors`:
     *   [normalized tensor, mean of input, variance of input].
     */
    function normalizeBatchInTraining(x, gamma, beta, reductionAxes, epsilon = 1e-3) {
        if (arraysEqual(reductionAxes.slice().sort(), range$3(0, x.rank - 1))) {
            return regularNormalizeBatchInTraining(x, gamma, beta, reductionAxes, epsilon);
        }
        else {
            return broadcastNormalizeBatchInTraining(x, gamma, beta, reductionAxes, epsilon);
        }
    }
    class BatchNormalization extends Layer {
        constructor(args) {
            if (args == null) {
                args = {};
            }
            super(args);
            this.supportsMasking = true;
            this.axis = args.axis == null ? -1 : args.axis;
            this.momentum = args.momentum == null ? 0.99 : args.momentum;
            this.epsilon = args.epsilon == null ? 1e-3 : args.epsilon;
            this.center = args.center == null ? true : args.center;
            this.scale = args.scale == null ? true : args.scale;
            this.betaInitializer = getInitializer(args.betaInitializer || 'zeros');
            this.gammaInitializer = getInitializer(args.gammaInitializer || 'ones');
            this.movingMeanInitializer =
                getInitializer(args.movingMeanInitializer || 'zeros');
            this.movingVarianceInitializer =
                getInitializer(args.movingVarianceInitializer || 'ones');
            this.betaConstraint = getConstraint(args.betaConstraint);
            this.gammaConstraint = getConstraint(args.gammaConstraint);
            this.betaRegularizer = getRegularizer(args.betaRegularizer);
            this.gammaRegularizer = getRegularizer(args.gammaRegularizer);
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const axis = this.axis >= 0 ? this.axis : (this.axis + inputShape.length);
            const dim = inputShape[axis];
            if (dim == null) {
                throw new ValueError(`Axis ${axis} of input tensor should have a defined dimension but ` +
                    `the layer received an input with shape ` +
                    `${JSON.stringify(inputShape)}.`);
            }
            this.inputSpec =
                [new InputSpec({ ndim: inputShape.length, axes: { [axis]: dim } })];
            const shape = [dim];
            if (this.scale) {
                this.gamma = this.addWeight('gamma', shape, null, this.gammaInitializer, this.gammaRegularizer, true, this.gammaConstraint);
            }
            if (this.center) {
                this.beta = this.addWeight('beta', shape, null, this.betaInitializer, this.betaRegularizer, true, this.betaConstraint);
            }
            this.movingMean = this.addWeight('moving_mean', shape, null, this.movingMeanInitializer, null, false);
            this.movingVariance = this.addWeight('moving_variance', shape, null, this.movingVarianceInitializer, null, false);
            this.built = true;
        }
        call(inputs, kwargs) {
            return tidy(() => {
                const training = kwargs['training'] == null ? false : kwargs['training'];
                const input = getExactlyOneTensor(inputs);
                const inputShape = input.shape;
                const ndim = inputShape.length;
                const reductionAxes = range$3(0, ndim);
                const axis = this.axis >= 0 ? this.axis : (this.axis + ndim);
                reductionAxes.splice(axis, 1);
                const broadcastShape = pyListRepeat(1, ndim);
                broadcastShape[axis] = inputShape[axis];
                const sortedReductionAxes = reductionAxes.slice();
                sortedReductionAxes.sort();
                const needsBroadcasting = !arraysEqual(sortedReductionAxes, range$3(0, ndim).slice(0, ndim - 1));
                const normalizeInference = () => {
                    if (needsBroadcasting) {
                        const broadcastMovingMean = reshape$3(this.movingMean.read(), broadcastShape);
                        const broadcastMovingVariance = reshape$3(this.movingVariance.read(), broadcastShape);
                        const broadcastBeta = this.center ? reshape$3(this.beta.read(), broadcastShape) : null;
                        const broadcastGamma = this.scale ? reshape$3(this.gamma.read(), broadcastShape) : null;
                        return batchNormalization$1(input, broadcastMovingMean, broadcastMovingVariance, broadcastBeta, broadcastGamma, this.epsilon);
                    }
                    else {
                        return batchNormalization$1(input, this.movingMean.read(), this.movingVariance.read(), this.beta == null ? null : this.beta.read(), this.gamma == null ? null : this.gamma.read(), this.epsilon);
                    }
                };
                if (!training) {
                    return normalizeInference();
                }
                const [normedTraining, mean, variance] = normalizeBatchInTraining(input, this.gamma.read(), this.beta.read(), reductionAxes, this.epsilon);
                const doMovingAverage = (variable, value, momentum) => {
                    tidy(() => {
                        const decay = 1 - momentum;
                        const origValue = variable.read();
                        const updateDelta = mul(sub$2(origValue, value), decay);
                        variable.write(sub$2(origValue, updateDelta));
                    });
                };
                // Perform updates to moving mean and moving variance for training.
                // Porting Note: In PyKeras, these updates to `movingMean` and
                //   `movingAverage` are done as a deferred Graph, added to the `Layer`'s
                //   `update`s using the `add_update()` method. Here we do it imperatively
                //   and encapsulate the updates in a function that is invoked
                //   immediately.
                const updateMovingMeanAndVariance = () => {
                    doMovingAverage(this.movingMean, mean, this.momentum);
                    doMovingAverage(this.movingVariance, variance, this.momentum);
                };
                updateMovingMeanAndVariance();
                return normedTraining;
            });
        }
        getConfig() {
            const config = {
                axis: this.axis,
                momentum: this.momentum,
                epsilon: this.epsilon,
                center: this.center,
                scale: this.scale,
                betaInitializer: serializeInitializer(this.betaInitializer),
                gammaInitializer: serializeInitializer(this.gammaInitializer),
                movingMeanInitializer: serializeInitializer(this.movingMeanInitializer),
                movingVarianceInitializer: serializeInitializer(this.movingVarianceInitializer),
                betaRegularizer: serializeRegularizer(this.betaRegularizer),
                gammaRegularizer: serializeRegularizer(this.gammaRegularizer),
                betaConstraint: serializeConstraint(this.betaConstraint),
                gammaConstraint: serializeConstraint(this.gammaConstraint)
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    BatchNormalization.className = 'BatchNormalization';
    registerClass(BatchNormalization);
    class LayerNormalization extends Layer {
        constructor(args) {
            if (args == null) {
                args = {};
            }
            super(args);
            this.axis = args.axis == null ? -1 : args.axis;
            if (typeof this.axis === 'number') {
                if (!Number.isInteger(this.axis)) {
                    throw new Error(`Expected axis to be an integer, but received ${this.axis}`);
                }
            }
            else if (Array.isArray(this.axis)) {
                for (const axis of this.axis) {
                    if (!Number.isInteger(axis)) {
                        throw new Error(`Expected axis to be an array of integers, ` +
                            `but received ${JSON.stringify(this.axis)}`);
                    }
                }
            }
            else {
                throw new Error(`Expected axis to be an integer or an array of integers, ` +
                    `but received ${JSON.stringify(this.axis)}`);
            }
            this.epsilon = args.epsilon == null ? 1e-3 : args.epsilon;
            this.center = args.center == null ? true : args.center;
            this.scale = args.scale == null ? true : args.scale;
            this.betaInitializer = getInitializer(args.betaInitializer || 'zeros');
            this.gammaInitializer = getInitializer(args.gammaInitializer || 'ones');
            this.betaRegularizer = getRegularizer(args.betaRegularizer);
            this.gammaRegularizer = getRegularizer(args.gammaRegularizer);
            this.supportsMasking = true;
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const nDims = inputShape.length;
            // Convert axis to array and resolve negatives.
            if (typeof this.axis === 'number') {
                this.axis = [this.axis];
            }
            for (let i = 0; i < this.axis.length; ++i) {
                if (this.axis[i] < 0) {
                    this.axis[i] += nDims;
                }
            }
            // Further validate axes.
            for (const axis of this.axis) {
                if (axis < 0 || axis >= nDims) {
                    throw new Error(`Invalid axis: ${axis}`);
                }
            }
            if (this.axis.length !== unique$2(this.axis).length) {
                throw new Error(`Found duplicate axes in: ${this.axis}`);
            }
            const paramShape = this.axis.map(axis => inputShape[axis]);
            const trainable = true;
            if (this.scale) {
                this.gamma = this.addWeight('gamma', paramShape, 'float32', this.gammaInitializer, this.gammaRegularizer, trainable);
            }
            else {
                this.gamma = null;
            }
            if (this.center) {
                this.beta = this.addWeight('beta', paramShape, 'float32', this.betaInitializer, this.betaRegularizer, trainable);
            }
            else {
                this.beta = null;
            }
            this.built = true;
        }
        call(inputs, kwargs) {
            const input = getExactlyOneTensor(inputs);
            const inputShape = input.shape;
            const nDims = inputShape.length;
            return tidy(() => {
                const keepDims = true;
                let { mean, variance } = moments(input, this.axis, keepDims);
                const broadcastShape = pyListRepeat(1, nDims);
                for (const dim of this.axis) {
                    broadcastShape[dim] = inputShape[dim];
                }
                const broadcast = (v) => {
                    if (v != null && v.shape.length !== nDims) {
                        return reshape$3(v, broadcastShape);
                    }
                    else {
                        return v;
                    }
                };
                let scale = this.scale ? broadcast(this.gamma.read()) : null;
                let offset = this.center ? broadcast(this.beta.read()) : null;
                // TODO(https://github.com/tensorflow/tfjs/issues/2120): The tiling below
                // is a workaround for the limitation of core's batchNormalization?d don't
                // support broadcasting in their gradients. In addition, the tiling is
                // necessary to ensure correctness on the browser CPU backend regardless
                // of forward or backward computation. Remove this workaround once the
                // limitation is addressed. See .
                const momentsTiling = [];
                const scaleOffsetTiling = [];
                for (let i = 0; i < nDims; ++i) {
                    if (this.axis.indexOf(i) !== -1) {
                        momentsTiling.push(inputShape[i]);
                        scaleOffsetTiling.push(1);
                    }
                    else {
                        momentsTiling.push(1);
                        scaleOffsetTiling.push(inputShape[i]);
                    }
                }
                mean = tile$3(mean, momentsTiling);
                variance = tile$3(variance, momentsTiling);
                if (scale != null) {
                    scale = tile$3(scale, scaleOffsetTiling);
                }
                if (offset != null) {
                    offset = tile$3(offset, scaleOffsetTiling);
                }
                return batchNormalization$1(input, mean, variance, offset, scale, this.epsilon);
            });
        }
        getConfig() {
            const config = {
                axis: this.axis,
                epsilon: this.epsilon,
                center: this.center,
                scale: this.scale,
                betaInitializer: serializeInitializer(this.betaInitializer),
                gammaInitializer: serializeInitializer(this.gammaInitializer),
                betaRegularizer: serializeRegularizer(this.betaRegularizer),
                gammaRegularizer: serializeRegularizer(this.gammaRegularizer)
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    LayerNormalization.className = 'LayerNormalization';
    registerClass(LayerNormalization);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Pads the 2nd and 3rd dimensions of a 4D tensor.
     *
     * @param x Input `tf.Tensor` to be padded.
     * @param padding `Array` of two `Array`s, each of which is an `Array` of two
     *   integers. The amount of padding at the beginning and end of the 2nd and 3rd
     *   dimensions, respectively.
     * @param dataFormat 'channelsLast' (default) or 'channelsFirst'.
     * @return Padded 4D `tf.Tensor`.
     */
    function spatial2dPadding(x, padding, dataFormat) {
        return tidy(() => {
            if (x.rank !== 4) {
                throw new ValueError(`temporalPadding expects input tensor to be 4-D, but received a ` +
                    `${x.rank}-D tensor.`);
            }
            if (padding == null) {
                padding = [[1, 1], [1, 1]];
            }
            if (padding.length !== 2 || padding[0].length !== 2 ||
                padding[1].length !== 2) {
                throw new ValueError('spatial2dPadding expects `padding` to be an Array of two Arrays, ' +
                    'each of which is an Array of two integers.');
            }
            if (dataFormat == null) {
                dataFormat = imageDataFormat();
            }
            if (dataFormat !== 'channelsLast' && dataFormat !== 'channelsFirst') {
                throw new ValueError(`Unknown data format: ${dataFormat}. ` +
                    `Supported data formats are 'channelsLast' and 'channelsFirst.`);
            }
            let pattern;
            if (dataFormat === 'channelsFirst') {
                pattern = [[0, 0], [0, 0], padding[0], padding[1]];
            }
            else {
                pattern = [[0, 0], padding[0], padding[1], [0, 0]];
            }
            return pad$1(x, pattern);
        });
    }
    class ZeroPadding2D extends Layer {
        constructor(args) {
            if (args == null) {
                args = {};
            }
            super(args);
            this.dataFormat =
                args.dataFormat == null ? imageDataFormat() : args.dataFormat;
            // TODO(cais): Maybe refactor the following logic surrounding `padding`
            //   into a helper method.
            if (args.padding == null) {
                this.padding = [[1, 1], [1, 1]];
            }
            else if (typeof args.padding === 'number') {
                this.padding =
                    [[args.padding, args.padding], [args.padding, args.padding]];
            }
            else {
                args.padding = args.padding;
                if (args.padding.length !== 2) {
                    throw new ValueError(`ZeroPadding2D expects padding to be a length-2 array, but ` +
                        `received a length-${args.padding.length} array.`);
                }
                let heightPadding;
                let widthPadding;
                if (typeof args.padding[0] === 'number') {
                    heightPadding = [args.padding[0], args.padding[0]];
                    widthPadding = [args.padding[1], args.padding[1]];
                }
                else {
                    args.padding = args.padding;
                    if (args.padding[0].length !== 2) {
                        throw new ValueError(`ZeroPadding2D expects height padding to be a length-2 array, ` +
                            `but received a length-${args.padding[0].length} array.`);
                    }
                    heightPadding = args.padding[0];
                    if (args.padding[1].length !== 2) {
                        throw new ValueError(`ZeroPadding2D expects width padding to be a length-2 array, ` +
                            `but received a length-${args.padding[1].length} array.`);
                    }
                    widthPadding = args.padding[1];
                }
                this.padding = [heightPadding, widthPadding];
            }
            this.inputSpec = [new InputSpec({ ndim: 4 })];
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            let rows;
            let cols;
            if (this.dataFormat === 'channelsFirst') {
                if (inputShape[2] != null && inputShape[2] >= 0) {
                    rows = inputShape[2] + this.padding[0][0] + this.padding[0][1];
                }
                else {
                    rows = null;
                }
                if (inputShape[3] != null && inputShape[3] >= 0) {
                    cols = inputShape[3] + this.padding[1][0] + this.padding[1][1];
                }
                else {
                    cols = null;
                }
                return [inputShape[0], inputShape[1], rows, cols];
            }
            else {
                if (inputShape[1] != null && inputShape[1] >= 0) {
                    rows = inputShape[1] + this.padding[0][0] + this.padding[0][1];
                }
                else {
                    rows = null;
                }
                if (inputShape[2] != null && inputShape[2] >= 0) {
                    cols = inputShape[2] + this.padding[1][0] + this.padding[1][1];
                }
                else {
                    cols = null;
                }
                return [inputShape[0], rows, cols, inputShape[3]];
            }
        }
        call(inputs, kwargs) {
            return tidy(() => spatial2dPadding(getExactlyOneTensor(inputs), this.padding, this.dataFormat));
        }
        getConfig() {
            const config = {
                padding: this.padding,
                dataFormat: this.dataFormat,
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    /** @nocollapse */
    ZeroPadding2D.className = 'ZeroPadding2D';
    registerClass(ZeroPadding2D);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * 2D pooling.
     * @param x
     * @param poolSize
     * @param stridesdes strides. Defaults to [1, 1].
     * @param padding padding. Defaults to 'valid'.
     * @param dataFormat data format. Defaults to 'channelsLast'.
     * @param poolMode Mode of pooling. Defaults to 'max'.
     * @returns Result of the 2D pooling.
     */
    function pool2d(x, poolSize, strides, padding, dataFormat, poolMode) {
        return tidy(() => {
            checkDataFormat(dataFormat);
            checkPoolMode(poolMode);
            checkPaddingMode(padding);
            if (strides == null) {
                strides = [1, 1];
            }
            if (padding == null) {
                padding = 'valid';
            }
            if (dataFormat == null) {
                dataFormat = imageDataFormat();
            }
            if (poolMode == null) {
                poolMode = 'max';
            }
            // TODO(cais): Remove the preprocessing step once deeplearn.js supports
            // dataFormat as an input argument.
            x = preprocessConv2DInput(x, dataFormat); // x is NHWC after preprocessing.
            let y;
            const paddingString = (padding === 'same') ? 'same' : 'valid';
            if (poolMode === 'max') {
                // TODO(cais): Rank check?
                y = maxPool$2(x, poolSize, strides, paddingString);
            }
            else { // 'avg'
                // TODO(cais): Check the dtype and rank of x and give clear error message
                //   if those are incorrect.
                y = avgPool$2(
                // TODO(cais): Rank check?
                x, poolSize, strides, paddingString);
            }
            if (dataFormat === 'channelsFirst') {
                y = transpose$2(y, [0, 3, 1, 2]); // NHWC -> NCHW.
            }
            return y;
        });
    }
    /**
     * 3D pooling.
     * @param x
     * @param poolSize. Default to [1, 1, 1].
     * @param strides strides. Defaults to [1, 1, 1].
     * @param padding padding. Defaults to 'valid'.
     * @param dataFormat data format. Defaults to 'channelsLast'.
     * @param poolMode Mode of pooling. Defaults to 'max'.
     * @returns Result of the 3D pooling.
     */
    function pool3d$1(x, poolSize, strides, padding, dataFormat, poolMode) {
        return tidy(() => {
            checkDataFormat(dataFormat);
            checkPoolMode(poolMode);
            checkPaddingMode(padding);
            if (strides == null) {
                strides = [1, 1, 1];
            }
            if (padding == null) {
                padding = 'valid';
            }
            if (dataFormat == null) {
                dataFormat = imageDataFormat();
            }
            if (poolMode == null) {
                poolMode = 'max';
            }
            // x is NDHWC after preprocessing.
            x = preprocessConv3DInput(x, dataFormat);
            let y;
            const paddingString = (padding === 'same') ? 'same' : 'valid';
            if (poolMode === 'max') {
                y = maxPool3d$1(x, poolSize, strides, paddingString);
            }
            else { // 'avg'
                y = avgPool3d$1(x, poolSize, strides, paddingString);
            }
            if (dataFormat === 'channelsFirst') {
                y = transpose$2(y, [0, 4, 1, 2, 3]); // NDHWC -> NCDHW.
            }
            return y;
        });
    }
    /**
     * Abstract class for different pooling 1D layers.
     */
    class Pooling1D extends Layer {
        /**
         *
         * @param args Parameters for the Pooling layer.
         *
         * config.poolSize defaults to 2.
         */
        constructor(args) {
            if (args.poolSize == null) {
                args.poolSize = 2;
            }
            super(args);
            if (typeof args.poolSize === 'number') {
                this.poolSize = [args.poolSize];
            }
            else if (Array.isArray(args.poolSize) &&
                args.poolSize.length === 1 &&
                typeof args.poolSize[0] === 'number') {
                this.poolSize = args.poolSize;
            }
            else {
                throw new ValueError(`poolSize for 1D convolutional layer must be a number or an ` +
                    `Array of a single number, but received ` +
                    `${JSON.stringify(args.poolSize)}`);
            }
            assertPositiveInteger(this.poolSize, 'poolSize');
            if (args.strides == null) {
                this.strides = this.poolSize;
            }
            else {
                if (typeof args.strides === 'number') {
                    this.strides = [args.strides];
                }
                else if (Array.isArray(args.strides) &&
                    args.strides.length === 1 &&
                    typeof args.strides[0] === 'number') {
                    this.strides = args.strides;
                }
                else {
                    throw new ValueError(`strides for 1D convolutional layer must be a number or an ` +
                        `Array of a single number, but received ` +
                        `${JSON.stringify(args.strides)}`);
                }
            }
            assertPositiveInteger(this.strides, 'strides');
            this.padding = args.padding == null ? 'valid' : args.padding;
            checkPaddingMode(this.padding);
            this.inputSpec = [new InputSpec({ ndim: 3 })];
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const length = convOutputLength(inputShape[1], this.poolSize[0], this.padding, this.strides[0]);
            return [inputShape[0], length, inputShape[2]];
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                // Add dummy last dimension.
                inputs = expandDims$2(getExactlyOneTensor(inputs), 2);
                const output = this.poolingFunction(getExactlyOneTensor(inputs), [this.poolSize[0], 1], [this.strides[0], 1], this.padding, 'channelsLast');
                // Remove dummy last dimension.
                return squeeze(output, [2]);
            });
        }
        getConfig() {
            const config = {
                poolSize: this.poolSize,
                padding: this.padding,
                strides: this.strides,
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    class MaxPooling1D extends Pooling1D {
        constructor(args) {
            super(args);
        }
        poolingFunction(inputs, poolSize, strides, padding, dataFormat) {
            checkDataFormat(dataFormat);
            checkPaddingMode(padding);
            return pool2d(inputs, poolSize, strides, padding, dataFormat, 'max');
        }
    }
    /** @nocollapse */
    MaxPooling1D.className = 'MaxPooling1D';
    registerClass(MaxPooling1D);
    class AveragePooling1D extends Pooling1D {
        constructor(args) {
            super(args);
        }
        poolingFunction(inputs, poolSize, strides, padding, dataFormat) {
            checkDataFormat(dataFormat);
            checkPaddingMode(padding);
            return pool2d(inputs, poolSize, strides, padding, dataFormat, 'avg');
        }
    }
    /** @nocollapse */
    AveragePooling1D.className = 'AveragePooling1D';
    registerClass(AveragePooling1D);
    /**
     * Abstract class for different pooling 2D layers.
     */
    class Pooling2D extends Layer {
        constructor(args) {
            if (args.poolSize == null) {
                args.poolSize = [2, 2];
            }
            super(args);
            this.poolSize = Array.isArray(args.poolSize) ?
                args.poolSize :
                [args.poolSize, args.poolSize];
            if (args.strides == null) {
                this.strides = this.poolSize;
            }
            else if (Array.isArray(args.strides)) {
                if (args.strides.length !== 2) {
                    throw new ValueError(`If the strides property of a 2D pooling layer is an Array, ` +
                        `it is expected to have a length of 2, but received length ` +
                        `${args.strides.length}.`);
                }
                this.strides = args.strides;
            }
            else {
                // `config.strides` is a number.
                this.strides = [args.strides, args.strides];
            }
            assertPositiveInteger(this.poolSize, 'poolSize');
            assertPositiveInteger(this.strides, 'strides');
            this.padding = args.padding == null ? 'valid' : args.padding;
            this.dataFormat =
                args.dataFormat == null ? 'channelsLast' : args.dataFormat;
            checkDataFormat(this.dataFormat);
            checkPaddingMode(this.padding);
            this.inputSpec = [new InputSpec({ ndim: 4 })];
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            let rows = this.dataFormat === 'channelsFirst' ? inputShape[2] : inputShape[1];
            let cols = this.dataFormat === 'channelsFirst' ? inputShape[3] : inputShape[2];
            rows =
                convOutputLength(rows, this.poolSize[0], this.padding, this.strides[0]);
            cols =
                convOutputLength(cols, this.poolSize[1], this.padding, this.strides[1]);
            if (this.dataFormat === 'channelsFirst') {
                return [inputShape[0], inputShape[1], rows, cols];
            }
            else {
                return [inputShape[0], rows, cols, inputShape[3]];
            }
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                return this.poolingFunction(getExactlyOneTensor(inputs), this.poolSize, this.strides, this.padding, this.dataFormat);
            });
        }
        getConfig() {
            const config = {
                poolSize: this.poolSize,
                padding: this.padding,
                strides: this.strides,
                dataFormat: this.dataFormat
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    class MaxPooling2D extends Pooling2D {
        constructor(args) {
            super(args);
        }
        poolingFunction(inputs, poolSize, strides, padding, dataFormat) {
            checkDataFormat(dataFormat);
            checkPaddingMode(padding);
            return pool2d(inputs, poolSize, strides, padding, dataFormat, 'max');
        }
    }
    /** @nocollapse */
    MaxPooling2D.className = 'MaxPooling2D';
    registerClass(MaxPooling2D);
    class AveragePooling2D extends Pooling2D {
        constructor(args) {
            super(args);
        }
        poolingFunction(inputs, poolSize, strides, padding, dataFormat) {
            checkDataFormat(dataFormat);
            checkPaddingMode(padding);
            return pool2d(inputs, poolSize, strides, padding, dataFormat, 'avg');
        }
    }
    /** @nocollapse */
    AveragePooling2D.className = 'AveragePooling2D';
    registerClass(AveragePooling2D);
    /**
     * Abstract class for different pooling 3D layers.
     */
    class Pooling3D extends Layer {
        constructor(args) {
            if (args.poolSize == null) {
                args.poolSize = [2, 2, 2];
            }
            super(args);
            this.poolSize = Array.isArray(args.poolSize) ?
                args.poolSize :
                [args.poolSize, args.poolSize, args.poolSize];
            if (args.strides == null) {
                this.strides = this.poolSize;
            }
            else if (Array.isArray(args.strides)) {
                if (args.strides.length !== 3) {
                    throw new ValueError(`If the strides property of a 3D pooling layer is an Array, ` +
                        `it is expected to have a length of 3, but received length ` +
                        `${args.strides.length}.`);
                }
                this.strides = args.strides;
            }
            else {
                // `config.strides` is a number.
                this.strides = [args.strides, args.strides, args.strides];
            }
            assertPositiveInteger(this.poolSize, 'poolSize');
            assertPositiveInteger(this.strides, 'strides');
            this.padding = args.padding == null ? 'valid' : args.padding;
            this.dataFormat =
                args.dataFormat == null ? 'channelsLast' : args.dataFormat;
            checkDataFormat(this.dataFormat);
            checkPaddingMode(this.padding);
            this.inputSpec = [new InputSpec({ ndim: 5 })];
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            let depths = this.dataFormat === 'channelsFirst' ? inputShape[2] : inputShape[1];
            let rows = this.dataFormat === 'channelsFirst' ? inputShape[3] : inputShape[2];
            let cols = this.dataFormat === 'channelsFirst' ? inputShape[4] : inputShape[3];
            depths = convOutputLength(depths, this.poolSize[0], this.padding, this.strides[0]);
            rows =
                convOutputLength(rows, this.poolSize[1], this.padding, this.strides[1]);
            cols =
                convOutputLength(cols, this.poolSize[2], this.padding, this.strides[2]);
            if (this.dataFormat === 'channelsFirst') {
                return [inputShape[0], inputShape[1], depths, rows, cols];
            }
            else {
                return [inputShape[0], depths, rows, cols, inputShape[4]];
            }
        }
        call(inputs, kwargs) {
            return tidy(() => {
                this.invokeCallHook(inputs, kwargs);
                return this.poolingFunction(getExactlyOneTensor(inputs), this.poolSize, this.strides, this.padding, this.dataFormat);
            });
        }
        getConfig() {
            const config = {
                poolSize: this.poolSize,
                padding: this.padding,
                strides: this.strides,
                dataFormat: this.dataFormat
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    class MaxPooling3D extends Pooling3D {
        constructor(args) {
            super(args);
        }
        poolingFunction(inputs, poolSize, strides, padding, dataFormat) {
            checkDataFormat(dataFormat);
            checkPaddingMode(padding);
            return pool3d$1(inputs, poolSize, strides, padding, dataFormat, 'max');
        }
    }
    /** @nocollapse */
    MaxPooling3D.className = 'MaxPooling3D';
    registerClass(MaxPooling3D);
    class AveragePooling3D extends Pooling3D {
        constructor(args) {
            super(args);
        }
        poolingFunction(inputs, poolSize, strides, padding, dataFormat) {
            checkDataFormat(dataFormat);
            checkPaddingMode(padding);
            return pool3d$1(inputs, poolSize, strides, padding, dataFormat, 'avg');
        }
    }
    /** @nocollapse */
    AveragePooling3D.className = 'AveragePooling3D';
    registerClass(AveragePooling3D);
    /**
     * Abstract class for different global pooling 1D layers.
     */
    class GlobalPooling1D extends Layer {
        constructor(args) {
            super(args);
            this.inputSpec = [new InputSpec({ ndim: 3 })];
        }
        computeOutputShape(inputShape) {
            return [inputShape[0], inputShape[2]];
        }
        call(inputs, kwargs) {
            throw new NotImplementedError();
        }
    }
    class GlobalAveragePooling1D extends GlobalPooling1D {
        constructor(args) {
            super(args || {});
        }
        call(inputs, kwargs) {
            return tidy(() => {
                const input = getExactlyOneTensor(inputs);
                return mean$1(input, 1);
            });
        }
    }
    /** @nocollapse */
    GlobalAveragePooling1D.className = 'GlobalAveragePooling1D';
    registerClass(GlobalAveragePooling1D);
    class GlobalMaxPooling1D extends GlobalPooling1D {
        constructor(args) {
            super(args || {});
        }
        call(inputs, kwargs) {
            return tidy(() => {
                const input = getExactlyOneTensor(inputs);
                return max$3(input, 1);
            });
        }
    }
    /** @nocollapse */
    GlobalMaxPooling1D.className = 'GlobalMaxPooling1D';
    registerClass(GlobalMaxPooling1D);
    /**
     * Abstract class for different global pooling 2D layers.
     */
    class GlobalPooling2D extends Layer {
        constructor(args) {
            super(args);
            this.dataFormat =
                args.dataFormat == null ? 'channelsLast' : args.dataFormat;
            checkDataFormat(this.dataFormat);
            this.inputSpec = [new InputSpec({ ndim: 4 })];
        }
        computeOutputShape(inputShape) {
            inputShape = inputShape;
            if (this.dataFormat === 'channelsLast') {
                return [inputShape[0], inputShape[3]];
            }
            else {
                return [inputShape[0], inputShape[1]];
            }
        }
        call(inputs, kwargs) {
            throw new NotImplementedError();
        }
        getConfig() {
            const config = { dataFormat: this.dataFormat };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
    }
    class GlobalAveragePooling2D extends GlobalPooling2D {
        call(inputs, kwargs) {
            return tidy(() => {
                const input = getExactlyOneTensor(inputs);
                if (this.dataFormat === 'channelsLast') {
                    return mean$1(input, [1, 2]);
                }
                else {
                    return mean$1(input, [2, 3]);
                }
            });
        }
    }
    /** @nocollapse */
    GlobalAveragePooling2D.className = 'GlobalAveragePooling2D';
    registerClass(GlobalAveragePooling2D);
    class GlobalMaxPooling2D extends GlobalPooling2D {
        call(inputs, kwargs) {
            return tidy(() => {
                const input = getExactlyOneTensor(inputs);
                if (this.dataFormat === 'channelsLast') {
                    return max$3(input, [1, 2]);
                }
                else {
                    return max$3(input, [2, 3]);
                }
            });
        }
    }
    /** @nocollapse */
    GlobalMaxPooling2D.className = 'GlobalMaxPooling2D';
    registerClass(GlobalMaxPooling2D);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Abstract wrapper base class.
     *
     * Wrappers take another layer and augment it in various ways.
     * Do not use this class as a layer, it is only an abstract base class.
     * Two usable wrappers are the `TimeDistributed` and `Bidirectional` wrappers.
     */
    class Wrapper extends Layer {
        constructor(args) {
            // Porting Note: In PyKeras, `self.layer` is set prior to the calling
            //   `super()`. But we can't do that here due to TypeScript's restriction.
            //   See: https://github.com/Microsoft/TypeScript/issues/8277
            //   As a result, we have to add checks in `get trainable()` and
            //   `set trainable()` below in order to prevent using `this.layer` when
            //   its value is `undefined`. The super constructor does use the getter
            //   and the setter of `this.layer`.
            super(args);
            this.layer = args.layer;
        }
        build(inputShape) {
            this.built = true;
        }
        // TODO(cais): Implement activityRegularizer getter.
        get trainable() {
            // Porting Note: the check of `this.layer` here is necessary due to the
            //   way the `constructor` of this class is written (see Porting Note
            //   above).
            if (this.layer != null) {
                return this.layer.trainable;
            }
            else {
                return false;
            }
        }
        set trainable(value) {
            // Porting Note: the check of `this.layer` here is necessary due to the
            //   way the `constructor` of this class is written (see Porting Note
            //   above).
            if (this.layer != null) {
                this.layer.trainable = value;
            }
        }
        get trainableWeights() {
            return this.layer.trainableWeights;
        }
        // TODO(cais): Implement setter for trainableWeights.
        get nonTrainableWeights() {
            return this.layer.nonTrainableWeights;
        }
        // TODO(cais): Implement setter for nonTrainableWeights.
        get updates() {
            // tslint:disable-next-line:no-any
            return this.layer._updates;
        }
        // TODO(cais): Implement getUpdatesFor().
        get losses() {
            return this.layer.losses;
        }
        // TODO(cais): Implement getLossesFor().
        getWeights() {
            return this.layer.getWeights();
        }
        setWeights(weights) {
            this.layer.setWeights(weights);
        }
        getConfig() {
            const config = {
                'layer': {
                    'className': this.layer.getClassName(),
                    'config': this.layer.getConfig(),
                }
            };
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
        setFastWeightInitDuringBuild(value) {
            super.setFastWeightInitDuringBuild(value);
            if (this.layer != null) {
                this.layer.setFastWeightInitDuringBuild(value);
            }
        }
        /** @nocollapse */
        static fromConfig(cls, config, customObjects = {}) {
            const layerConfig = config['layer'];
            const layer = deserialize(layerConfig, customObjects);
            delete config['layer'];
            const newConfig = { layer };
            Object.assign(newConfig, config);
            return new cls(newConfig);
        }
    }
    class TimeDistributed extends Wrapper {
        constructor(args) {
            super(args);
            this.supportsMasking = true;
        }
        build(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            if (inputShape.length < 3) {
                throw new ValueError(`TimeDistributed layer expects an input shape >= 3D, but received ` +
                    `input shape ${JSON.stringify(inputShape)}`);
            }
            this.inputSpec = [{ shape: inputShape }];
            const childInputShape = [inputShape[0]].concat(inputShape.slice(2));
            if (!this.layer.built) {
                this.layer.build(childInputShape);
                this.layer.built = true;
            }
            super.build(inputShape);
        }
        computeOutputShape(inputShape) {
            inputShape = getExactlyOneShape(inputShape);
            const childInputShape = [inputShape[0]].concat(inputShape.slice(2));
            const childOutputShape = this.layer.computeOutputShape(childInputShape);
            const timesteps = inputShape[1];
            return [childOutputShape[0], timesteps].concat(childOutputShape.slice(1));
        }
        call(inputs, kwargs) {
            return tidy(() => {
                // TODO(cais): Add 'training' and 'useLearningPhase' to kwargs.
                inputs = getExactlyOneTensor(inputs);
                // Porting Note: In tfjs-layers, `inputs` are always concrete tensor
                // values. Hence the inputs can't have an undetermined first (batch)
                // dimension, which is why we always use the K.rnn approach here.
                const step = (inputs, states) => {
                    // TODO(cais): Add useLearningPhase.
                    // NOTE(cais): `layer.call` may return a length-1 array of Tensor in
                    //   some cases (e.g., `layer` is a `Sequential` instance), which is
                    //   why `getExactlyOneTensor` is used below.
                    const output = getExactlyOneTensor(this.layer.call(inputs, kwargs));
                    return [output, []];
                };
                const rnnOutputs = rnn$1(step, inputs, [], false /* goBackwards */, null /* mask */, null /* constants */, false /* unroll */, true /* needPerStepOutputs */);
                const y = rnnOutputs[1];
                // TODO(cais): Add activity regularization.
                // TODO(cais): Add useLearningPhase.
                return y;
            });
        }
    }
    /** @nocollapse */
    TimeDistributed.className = 'TimeDistributed';
    registerClass(TimeDistributed);
    function checkBidirectionalMergeMode(value) {
        checkStringTypeUnionValue(VALID_BIDIRECTIONAL_MERGE_MODES, 'BidirectionalMergeMode', value);
    }
    const DEFAULT_BIDIRECTIONAL_MERGE_MODE = 'concat';
    class Bidirectional extends Wrapper {
        constructor(args) {
            super(args);
            // Note: When creating `this.forwardLayer`, the original Layer object
            //   (`config.layer`) ought to be cloned. This is why we call
            //   `getConfig()` followed by `deserialize()`. Without this cloning,
            //   the layer names saved during serialization will incorrectly contain
            //   the 'forward_' prefix. In Python Keras, this is done using
            //   `copy.copy` (shallow copy), which does not have a simple equivalent
            //   in JavaScript. JavaScript's `Object.assign()` does not copy
            //   methods.
            const layerConfig = args.layer.getConfig();
            const forwDict = {};
            forwDict['className'] = args.layer.getClassName();
            forwDict['config'] = layerConfig;
            this.forwardLayer = deserialize(forwDict);
            layerConfig['goBackwards'] =
                layerConfig['goBackwards'] === true ? false : true;
            const backDict = {};
            backDict['className'] = args.layer.getClassName();
            backDict['config'] = layerConfig;
            this.backwardLayer = deserialize(backDict);
            this.forwardLayer.name = 'forward_' + this.forwardLayer.name;
            this.backwardLayer.name = 'backward_' + this.backwardLayer.name;
            this.mergeMode = args.mergeMode === undefined ?
                DEFAULT_BIDIRECTIONAL_MERGE_MODE :
                args.mergeMode;
            checkBidirectionalMergeMode(this.mergeMode);
            if (args.weights) {
                throw new NotImplementedError('weights support is not implemented for Bidirectional layer yet.');
            }
            this._stateful = args.layer.stateful;
            this.returnSequences = args.layer.returnSequences;
            this.returnState = args.layer.returnState;
            this.supportsMasking = true;
            this._trainable = true;
            this.inputSpec = args.layer.inputSpec;
            this.numConstants = null;
        }
        get trainable() {
            return this._trainable;
        }
        set trainable(value) {
            // Porting Note: the check of `this.layer` here is necessary due to the
            //   way the `constructor` of this class is written (see Porting Note
            //   above).
            this._trainable = value;
            if (this.forwardLayer != null) {
                this.forwardLayer.trainable = value;
            }
            if (this.backwardLayer != null) {
                this.backwardLayer.trainable = value;
            }
        }
        getWeights() {
            return this.forwardLayer.getWeights().concat(this.backwardLayer.getWeights());
        }
        setWeights(weights) {
            const numWeights = weights.length;
            const numeightsOver2 = Math.floor(numWeights / 2);
            this.forwardLayer.setWeights(weights.slice(0, numeightsOver2));
            this.backwardLayer.setWeights(weights.slice(numeightsOver2));
        }
        computeOutputShape(inputShape) {
            let layerShapes = this.forwardLayer.computeOutputShape(inputShape);
            if (!(Array.isArray(layerShapes) && Array.isArray(layerShapes[0]))) {
                layerShapes = [layerShapes];
            }
            layerShapes = layerShapes;
            let outputShape;
            let outputShapes;
            let stateShape;
            if (this.returnState) {
                stateShape = layerShapes.slice(1);
                outputShape = layerShapes[0];
            }
            else {
                outputShape = layerShapes[0];
            }
            outputShape = outputShape;
            if (this.mergeMode === 'concat') {
                outputShape[outputShape.length - 1] *= 2;
                outputShapes = [outputShape];
            }
            else if (this.mergeMode == null) {
                outputShapes = [outputShape, outputShape.slice()];
            }
            else {
                outputShapes = [outputShape];
            }
            if (this.returnState) {
                if (this.mergeMode == null) {
                    return outputShapes.concat(stateShape).concat(stateShape.slice());
                }
                return [outputShape].concat(stateShape).concat(stateShape.slice());
            }
            return singletonOrArray(outputShapes);
        }
        apply(inputs, kwargs) {
            let initialState = kwargs == null ? null : kwargs['initialState'];
            let constants = kwargs == null ? null : kwargs['constants'];
            if (kwargs == null) {
                kwargs = {};
            }
            const standardized = standardizeArgs(inputs, initialState, constants, this.numConstants);
            inputs = standardized.inputs;
            initialState = standardized.initialState;
            constants = standardized.constants;
            if (Array.isArray(inputs)) {
                initialState = inputs.slice(1);
                inputs = inputs[0];
            }
            if ((initialState == null || initialState.length === 0) &&
                constants == null) {
                return super.apply(inputs, kwargs);
            }
            const additionalInputs = [];
            const additionalSpecs = [];
            if (initialState != null) {
                const numStates = initialState.length;
                if (numStates % 2 > 0) {
                    throw new ValueError('When passing `initialState` to a Bidrectional RNN, ' +
                        'the state should be an Array containing the states of ' +
                        'the underlying RNNs.');
                }
                kwargs['initialState'] = initialState;
                additionalInputs.push(...initialState);
                const stateSpecs = initialState
                    .map(state => new InputSpec({ shape: state.shape }));
                this.forwardLayer.stateSpec = stateSpecs.slice(0, numStates / 2);
                this.backwardLayer.stateSpec = stateSpecs.slice(numStates / 2);
                additionalSpecs.push(...stateSpecs);
            }
            if (constants != null) {
                throw new NotImplementedError('Support for constants in Bidirectional layers is not ' +
                    'implemented yet.');
            }
            const isSymbolicTensor = additionalInputs[0] instanceof SymbolicTensor;
            for (const tensor of additionalInputs) {
                if (tensor instanceof SymbolicTensor !== isSymbolicTensor) {
                    throw new ValueError('The initial state of a Bidirectional layer cannot be ' +
                        'specified as a mix of symbolic and non-symbolic tensors');
                }
            }
            if (isSymbolicTensor) {
                // Compute the full input and specs, including the states.
                const fullInput = [inputs].concat(additionalInputs);
                const fullInputSpec = this.inputSpec.concat(additionalSpecs);
                // Perform the call temporarily and replace inputSpec.
                // Note: with initial states symbolic calls and non-symbolic calls to
                // this method differ in how the initial states are passed. For
                // symbolic calls, the initial states are passed in the first arg, as
                // an Array of SymbolicTensors; for non-symbolic calls, they are
                // passed in the second arg as a part of the kwargs. Hence the need to
                // temporarily modify inputSpec here.
                // TODO(cais): Make refactoring so that this hacky code below is no
                // longer needed.
                const originalInputSpec = this.inputSpec;
                this.inputSpec = fullInputSpec;
                const output = super.apply(fullInput, kwargs);
                this.inputSpec = originalInputSpec;
                return output;
            }
            else {
                return super.apply(inputs, kwargs);
            }
        }
        call(inputs, kwargs) {
            return tidy(() => {
                const initialState = kwargs['initialState'];
                let y;
                let yRev;
                if (initialState == null) {
                    y = this.forwardLayer.call(inputs, kwargs);
                    yRev = this.backwardLayer.call(inputs, kwargs);
                }
                else {
                    const forwardState = initialState.slice(0, initialState.length / 2);
                    const backwardState = initialState.slice(initialState.length / 2);
                    y = this.forwardLayer.call(inputs, Object.assign(kwargs, { initialState: forwardState }));
                    yRev = this.backwardLayer.call(inputs, Object.assign(kwargs, { initialState: backwardState }));
                }
                let states;
                if (this.returnState) {
                    if (Array.isArray(y)) {
                        states = y.slice(1).concat(yRev.slice(1));
                    }
                    y = y[0];
                    yRev = yRev[0];
                }
                if (this.returnSequences) {
                    yRev = reverse$2(yRev, 1);
                }
                let output;
                if (this.mergeMode === 'concat') {
                    output = concatenate$1([y, yRev]);
                }
                else if (this.mergeMode === 'sum') {
                    output = add$2(y, yRev);
                }
                else if (this.mergeMode === 'ave') {
                    output = mul(.5, add$2(y, yRev));
                }
                else if (this.mergeMode === 'mul') {
                    output = mul(y, yRev);
                }
                else if (this.mergeMode == null) {
                    output = [y, yRev];
                }
                // TODO(cais): Properly set learning phase.
                if (this.returnState) {
                    if (this.mergeMode == null) {
                        return output.concat(states);
                    }
                    return [output].concat(states);
                }
                return output;
            });
        }
        resetStates(states) {
            this.forwardLayer.resetStates();
            this.backwardLayer.resetStates();
        }
        build(inputShape) {
            nameScope(this.forwardLayer.name, () => {
                this.forwardLayer.build(inputShape);
            });
            nameScope(this.backwardLayer.name, () => {
                this.backwardLayer.build(inputShape);
            });
            this.built = true;
        }
        computeMask(inputs, mask) {
            if (Array.isArray(mask)) {
                mask = mask[0];
            }
            let outputMask;
            if (this.returnSequences) {
                if (this.mergeMode == null) {
                    outputMask = [mask, mask];
                }
                else {
                    outputMask = mask;
                }
            }
            else {
                if (this.mergeMode == null) {
                    outputMask = [null, null];
                }
                else {
                    outputMask = null;
                }
            }
            if (this.returnState) {
                const states = this.forwardLayer.states;
                const stateMask = states.map(state => null);
                if (Array.isArray(outputMask)) {
                    return outputMask.concat(stateMask).concat(stateMask);
                }
                else {
                    return [outputMask].concat(stateMask).concat(stateMask);
                }
            }
            else {
                return outputMask;
            }
        }
        get trainableWeights() {
            return this.forwardLayer.trainableWeights.concat(this.backwardLayer.trainableWeights);
        }
        get nonTrainableWeights() {
            return this.forwardLayer.nonTrainableWeights.concat(this.backwardLayer.nonTrainableWeights);
        }
        // TODO(cais): Implement constraints().
        setFastWeightInitDuringBuild(value) {
            super.setFastWeightInitDuringBuild(value);
            if (this.forwardLayer != null) {
                this.forwardLayer.setFastWeightInitDuringBuild(value);
            }
            if (this.backwardLayer != null) {
                this.backwardLayer.setFastWeightInitDuringBuild(value);
            }
        }
        getConfig() {
            const config = {
                'mergeMode': this.mergeMode,
            };
            // TODO(cais): Add logic for `numConstants` once the property is added.
            const baseConfig = super.getConfig();
            Object.assign(config, baseConfig);
            return config;
        }
        /** @nocollapse */
        static fromConfig(cls, config) {
            const rnnLayer = deserialize(config['layer']);
            delete config['layer'];
            // TODO(cais): Add logic for `numConstants` once the property is added.
            if (config['numConstants'] != null) {
                throw new NotImplementedError(`Deserialization of a Bidirectional layer with numConstants ` +
                    `present is not supported yet.`);
            }
            // tslint:disable-next-line:no-any
            const newConfig = config;
            newConfig['layer'] = rnnLayer;
            return new cls(newConfig);
        }
    }
    /** @nocollapse */
    Bidirectional.className = 'Bidirectional';
    registerClass(Bidirectional);

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    // TODO(cais): Add doc string to all the public static functions in this
    //   class; include exectuable JavaScript code snippets where applicable
    //   (b/74074458).
    // Input Layer.
    /**
     * An input layer is an entry point into a `tf.LayersModel`.
     *
     * `InputLayer` is generated automatically for `tf.Sequential`` models by
     * specifying the `inputshape` or `batchInputShape` for the first layer.  It
     * should not be specified explicitly. However, it can be useful sometimes,
     * e.g., when constructing a sequential model from a subset of another
     * sequential model's layers. Like the code snippet below shows.
     *
     * ```js
     * // Define a model which simply adds two inputs.
     * const model1 = tf.sequential();
     * model1.add(tf.layers.dense({inputShape: [4], units: 3, activation: 'relu'}));
     * model1.add(tf.layers.dense({units: 1, activation: 'sigmoid'}));
     * model1.summary();
     * model1.predict(tf.zeros([1, 4])).print();
     *
     * // Construct another model, reusing the second layer of `model1` while
     * // not using the first layer of `model1`. Note that you cannot add the second
     * // layer of `model` directly as the first layer of the new sequential model,
     * // because doing so will lead to an error related to the fact that the layer
     * // is not an input layer. Instead, you need to create an `inputLayer` and add
     * // it to the new sequential model before adding the reused layer.
     * const model2 = tf.sequential();
     * // Use an inputShape that matches the input shape of `model1`'s second
     * // layer.
     * model2.add(tf.layers.inputLayer({inputShape: [3]}));
     * model2.add(model1.layers[1]);
     * model2.summary();
     * model2.predict(tf.zeros([1, 3])).print();
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Inputs', namespace: 'layers'}
     */
    function inputLayer(args) {
        return new InputLayer(args);
    }
    // Advanced Activation Layers.
    /**
     * Exponetial Linear Unit (ELU).
     *
     * It follows:
     * `f(x) =  alpha * (exp(x) - 1.) for x < 0`,
     * `f(x) = x for x >= 0`.
     *
     * Input shape:
     *   Arbitrary. Use the configuration `inputShape` when using this layer as the
     *   first layer in a model.
     *
     * Output shape:
     *   Same shape as the input.
     *
     * References:
     *   - [Fast and Accurate Deep Network Learning by Exponential Linear Units
     * (ELUs)](https://arxiv.org/abs/1511.07289v1)
     *
     * @doc {
     *   heading: 'Layers',
     *   subheading: 'Advanced Activation',
     *   namespace: 'layers'
     * }
     */
    function elu$2(args) {
        return new ELU$3(args);
    }
    /**
     * Rectified Linear Unit activation function.
     *
     * Input shape:
     *   Arbitrary. Use the config field `inputShape` (Array of integers, does
     *   not include the sample axis) when using this layer as the first layer
     *   in a model.
     *
     * Output shape:
     *   Same shape as the input.
     *
     * @doc {
     *   heading: 'Layers',
     *   subheading: 'Advanced Activation',
     *   namespace: 'layers'
     * }
     */
    function reLU(args) {
        return new ReLU(args);
    }
    /**
     * Leaky version of a rectified linear unit.
     *
     * It allows a small gradient when the unit is not active:
     * `f(x) = alpha * x for x < 0.`
     * `f(x) = x for x >= 0.`
     *
     * Input shape:
     *   Arbitrary. Use the configuration `inputShape` when using this layer as the
     *   first layer in a model.
     *
     * Output shape:
     *   Same shape as the input.
     *
     * @doc {
     *   heading: 'Layers',
     *   subheading: 'Advanced Activation',
     *   namespace: 'layers'
     * }
     */
    function leakyReLU(args) {
        return new LeakyReLU(args);
    }
    /**
     * Parameterized version of a leaky rectified linear unit.
     *
     * It follows
     * `f(x) = alpha * x for x < 0.`
     * `f(x) = x for x >= 0.`
     * wherein `alpha` is a trainable weight.
     *
     * Input shape:
     *   Arbitrary. Use the configuration `inputShape` when using this layer as the
     *   first layer in a model.
     *
     * Output shape:
     *   Same shape as the input.
     *
     * @doc {
     *   heading: 'Layers',
     *   subheading: 'Advanced Activation',
     *   namespace: 'layers'
     * }
     */
    function prelu$2(args) {
        return new PReLU(args);
    }
    /**
     * Softmax activation layer.
     *
     * Input shape:
     *   Arbitrary. Use the configuration `inputShape` when using this layer as the
     *   first layer in a model.
     *
     * Output shape:
     *   Same shape as the input.
     *
     * @doc {
     *   heading: 'Layers',
     *   subheading: 'Advanced Activation',
     *   namespace: 'layers'
     * }
     */
    function softmax$2(args) {
        return new Softmax(args);
    }
    /**
     * Thresholded Rectified Linear Unit.
     *
     * It follows:
     * `f(x) = x for x > theta`,
     * `f(x) = 0 otherwise`.
     *
     * Input shape:
     *   Arbitrary. Use the configuration `inputShape` when using this layer as the
     *   first layer in a model.
     *
     * Output shape:
     *   Same shape as the input.
     *
     * References:
     *   - [Zero-Bias Autoencoders and the Benefits of Co-Adapting
     * Features](http://arxiv.org/abs/1402.3337)
     *
     * @doc {
     *   heading: 'Layers',
     *   subheading: 'Advanced Activation',
     *   namespace: 'layers'
     * }
     */
    function thresholdedReLU(args) {
        return new ThresholdedReLU(args);
    }
    // Convolutional Layers.
    /**
     * 1D convolution layer (e.g., temporal convolution).
     *
     * This layer creates a convolution kernel that is convolved
     * with the layer input over a single spatial (or temporal) dimension
     * to produce a tensor of outputs.
     *
     * If `use_bias` is True, a bias vector is created and added to the outputs.
     *
     * If `activation` is not `null`, it is applied to the outputs as well.
     *
     * When using this layer as the first layer in a model, provide an
     * `inputShape` argument `Array` or `null`.
     *
     * For example, `inputShape` would be:
     * - `[10, 128]` for sequences of 10 vectors of 128-dimensional vectors
     * - `[null, 128]` for variable-length sequences of 128-dimensional vectors.
     *
     * @doc {heading: 'Layers', subheading: 'Convolutional',  namespace: 'layers'}
     */
    function conv1d(args) {
        return new Conv1D(args);
    }
    /**
     * 2D convolution layer (e.g. spatial convolution over images).
     *
     * This layer creates a convolution kernel that is convolved
     * with the layer input to produce a tensor of outputs.
     *
     * If `useBias` is True, a bias vector is created and added to the outputs.
     *
     * If `activation` is not `null`, it is applied to the outputs as well.
     *
     * When using this layer as the first layer in a model,
     * provide the keyword argument `inputShape`
     * (Array of integers, does not include the sample axis),
     * e.g. `inputShape=[128, 128, 3]` for 128x128 RGB pictures
     * in `dataFormat='channelsLast'`.
     *
     * @doc {heading: 'Layers', subheading: 'Convolutional', namespace: 'layers'}
     */
    function conv2d$1(args) {
        return new Conv2D(args);
    }
    /**
     * Transposed convolutional layer (sometimes called Deconvolution).
     *
     * The need for transposed convolutions generally arises
     * from the desire to use a transformation going in the opposite direction of
     * a normal convolution, i.e., from something that has the shape of the output
     * of some convolution to something that has the shape of its input while
     * maintaining a connectivity pattern that is compatible with said
     * convolution.
     *
     * When using this layer as the first layer in a model, provide the
     * configuration `inputShape` (`Array` of integers, does not include the
     * sample axis), e.g., `inputShape: [128, 128, 3]` for 128x128 RGB pictures in
     * `dataFormat: 'channelsLast'`.
     *
     * Input shape:
     *   4D tensor with shape:
     *   `[batch, channels, rows, cols]` if `dataFormat` is `'channelsFirst'`.
     *   or 4D tensor with shape
     *   `[batch, rows, cols, channels]` if `dataFormat` is `'channelsLast`.
     *
     * Output shape:
     *   4D tensor with shape:
     *   `[batch, filters, newRows, newCols]` if `dataFormat` is
     * `'channelsFirst'`. or 4D tensor with shape:
     *   `[batch, newRows, newCols, filters]` if `dataFormat` is `'channelsLast'`.
     *
     * References:
     *   - [A guide to convolution arithmetic for deep
     * learning](https://arxiv.org/abs/1603.07285v1)
     *   - [Deconvolutional
     * Networks](http://www.matthewzeiler.com/pubs/cvpr2010/cvpr2010.pdf)
     *
     * @doc {heading: 'Layers', subheading: 'Convolutional', namespace: 'layers'}
     */
    function conv2dTranspose(args) {
        return new Conv2DTranspose(args);
    }
    /**
     * 3D convolution layer (e.g. spatial convolution over volumes).
     *
     * This layer creates a convolution kernel that is convolved
     * with the layer input to produce a tensor of outputs.
     *
     * If `useBias` is True, a bias vector is created and added to the outputs.
     *
     * If `activation` is not `null`, it is applied to the outputs as well.
     *
     * When using this layer as the first layer in a model,
     * provide the keyword argument `inputShape`
     * (Array of integers, does not include the sample axis),
     * e.g. `inputShape=[128, 128, 128, 1]` for 128x128x128 grayscale volumes
     * in `dataFormat='channelsLast'`.
     *
     * @doc {heading: 'Layers', subheading: 'Convolutional', namespace: 'layers'}
     */
    function conv3d(args) {
        return new Conv3D(args);
    }
    function conv3dTranspose(args) {
        return new Conv3DTranspose(args);
    }
    /**
     * Depthwise separable 2D convolution.
     *
     * Separable convolution consists of first performing
     * a depthwise spatial convolution
     * (which acts on each input channel separately)
     * followed by a pointwise convolution which mixes together the resulting
     * output channels. The `depthMultiplier` argument controls how many
     * output channels are generated per input channel in the depthwise step.
     *
     * Intuitively, separable convolutions can be understood as
     * a way to factorize a convolution kernel into two smaller kernels,
     * or as an extreme version of an Inception block.
     *
     * Input shape:
     *   4D tensor with shape:
     *     `[batch, channels, rows, cols]` if data_format='channelsFirst'
     *   or 4D tensor with shape:
     *     `[batch, rows, cols, channels]` if data_format='channelsLast'.
     *
     * Output shape:
     *   4D tensor with shape:
     *     `[batch, filters, newRows, newCols]` if data_format='channelsFirst'
     *   or 4D tensor with shape:
     *     `[batch, newRows, newCols, filters]` if data_format='channelsLast'.
     *     `rows` and `cols` values might have changed due to padding.
     *
     * @doc {heading: 'Layers', subheading: 'Convolutional', namespace: 'layers'}
     */
    function separableConv2d(args) {
        return new SeparableConv2D(args);
    }
    /**
     * Cropping layer for 2D input (e.g., image).
     *
     * This layer can crop an input
     * at the top, bottom, left and right side of an image tensor.
     *
     * Input shape:
     *   4D tensor with shape:
     *   - If `dataFormat` is `"channelsLast"`:
     *     `[batch, rows, cols, channels]`
     *   - If `data_format` is `"channels_first"`:
     *     `[batch, channels, rows, cols]`.
     *
     * Output shape:
     *   4D with shape:
     *   - If `dataFormat` is `"channelsLast"`:
     *     `[batch, croppedRows, croppedCols, channels]`
     *    - If `dataFormat` is `"channelsFirst"`:
     *     `[batch, channels, croppedRows, croppedCols]`.
     *
     * Examples
     * ```js
     *
     * const model = tf.sequential();
     * model.add(tf.layers.cropping2D({cropping:[[2, 2], [2, 2]],
     *                                inputShape: [128, 128, 3]}));
     * //now output shape is [batch, 124, 124, 3]
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Convolutional', namespace: 'layers'}
     */
    function cropping2D(args) {
        return new Cropping2D(args);
    }
    /**
     * Upsampling layer for 2D inputs.
     *
     * Repeats the rows and columns of the data
     * by size[0] and size[1] respectively.
     *
     *
     * Input shape:
     *    4D tensor with shape:
     *     - If `dataFormat` is `"channelsLast"`:
     *         `[batch, rows, cols, channels]`
     *     - If `dataFormat` is `"channelsFirst"`:
     *        `[batch, channels, rows, cols]`
     *
     * Output shape:
     *     4D tensor with shape:
     *     - If `dataFormat` is `"channelsLast"`:
     *        `[batch, upsampledRows, upsampledCols, channels]`
     *     - If `dataFormat` is `"channelsFirst"`:
     *         `[batch, channels, upsampledRows, upsampledCols]`
     *
     *
     * @doc {heading: 'Layers', subheading: 'Convolutional', namespace: 'layers'}
     */
    function upSampling2d(args) {
        return new UpSampling2D(args);
    }
    // Convolutional(depthwise) Layers.
    /**
     * Depthwise separable 2D convolution.
     *
     * Depthwise Separable convolutions consists in performing just the first step
     * in a depthwise spatial convolution (which acts on each input channel
     * separately). The `depthMultplier` argument controls how many output channels
     * are generated per input channel in the depthwise step.
     *
     * @doc {heading: 'Layers', subheading: 'Convolutional', namespace: 'layers'}
     */
    function depthwiseConv2d(args) {
        return new DepthwiseConv2D(args);
    }
    // Basic Layers.
    /**
     * Applies an activation function to an output.
     *
     * This layer applies element-wise activation function.  Other layers, notably
     * `dense` can also apply activation functions.  Use this isolated activation
     * function to extract the values before and after the
     * activation. For instance:
     *
     * ```js
     * const input = tf.input({shape: [5]});
     * const denseLayer = tf.layers.dense({units: 1});
     * const activationLayer = tf.layers.activation({activation: 'relu6'});
     *
     * // Obtain the output symbolic tensors by applying the layers in order.
     * const denseOutput = denseLayer.apply(input);
     * const activationOutput = activationLayer.apply(denseOutput);
     *
     * // Create the model based on the inputs.
     * const model = tf.model({
     *     inputs: input,
     *     outputs: [denseOutput, activationOutput]
     * });
     *
     * // Collect both outputs and print separately.
     * const [denseOut, activationOut] = model.predict(tf.randomNormal([6, 5]));
     * denseOut.print();
     * activationOut.print();
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Basic', namespace: 'layers'}
     */
    function activation(args) {
        return new Activation(args);
    }
    /**
     * Creates a dense (fully connected) layer.
     *
     * This layer implements the operation:
     *   `output = activation(dot(input, kernel) + bias)`
     *
     * `activation` is the element-wise activation function
     *   passed as the `activation` argument.
     *
     * `kernel` is a weights matrix created by the layer.
     *
     * `bias` is a bias vector created by the layer (only applicable if `useBias`
     * is `true`).
     *
     * **Input shape:**
     *
     *   nD `tf.Tensor` with shape: `(batchSize, ..., inputDim)`.
     *
     *   The most common situation would be
     *   a 2D input with shape `(batchSize, inputDim)`.
     *
     * **Output shape:**
     *
     *   nD tensor with shape: `(batchSize, ..., units)`.
     *
     *   For instance, for a 2D input with shape `(batchSize, inputDim)`,
     *   the output would have shape `(batchSize, units)`.
     *
     * Note: if the input to the layer has a rank greater than 2, then it is
     * flattened prior to the initial dot product with the kernel.
     *
     * @doc {heading: 'Layers', subheading: 'Basic', namespace: 'layers'}
     */
    function dense(args) {
        return new Dense(args);
    }
    /**
     * Applies
     * [dropout](http://www.cs.toronto.edu/~rsalakhu/papers/srivastava14a.pdf) to
     * the input.
     *
     * Dropout consists in randomly setting a fraction `rate` of input units to 0 at
     * each update during training time, which helps prevent overfitting.
     *
     * @doc {heading: 'Layers', subheading: 'Basic', namespace: 'layers'}
     */
    function dropout(args) {
        return new Dropout(args);
    }
    /**
     * Spatial 1D version of Dropout.
     *
     * This Layer type performs the same function as the Dropout layer, but it drops
     * entire 1D feature maps instead of individual elements. For example, if an
     * input example consists of 3 timesteps and the feature map for each timestep
     * has a size of 4, a `spatialDropout1d` layer may zero out the feature maps
     * of the 1st timesteps and 2nd timesteps completely while sparing all feature
     * elements of the 3rd timestep.
     *
     * If adjacent frames (timesteps) are strongly correlated (as is normally the
     * case in early convolution layers), regular dropout will not regularize the
     * activation and will otherwise just result in merely an effective learning
     * rate decrease. In this case, `spatialDropout1d` will help promote
     * independence among feature maps and should be used instead.
     *
     * **Arguments:**
     *   rate: A floating-point number >=0 and <=1. Fraction of the input elements
     *     to drop.
     *
     * **Input shape:**
     *   3D tensor with shape `(samples, timesteps, channels)`.
     *
     * **Output shape:**
     *   Same as the input shape.
     *
     * References:
     *   - [Efficient Object Localization Using Convolutional
     *      Networks](https://arxiv.org/abs/1411.4280)
     *
     * @doc {heading: 'Layers', subheading: 'Basic', namespace: 'layers'}
     */
    function spatialDropout1d(args) {
        return new SpatialDropout1D(args);
    }
    /**
     * Flattens the input. Does not affect the batch size.
     *
     * A `Flatten` layer flattens each batch in its inputs to 1D (making the output
     * 2D).
     *
     * For example:
     *
     * ```js
     * const input = tf.input({shape: [4, 3]});
     * const flattenLayer = tf.layers.flatten();
     * // Inspect the inferred output shape of the flatten layer, which
     * // equals `[null, 12]`. The 2nd dimension is 4 * 3, i.e., the result of the
     * // flattening. (The 1st dimension is the undermined batch size.)
     * console.log(JSON.stringify(flattenLayer.apply(input).shape));
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Basic', namespace: 'layers'}
     */
    function flatten$1(args) {
        return new Flatten(args);
    }
    /**
     * Repeats the input n times in a new dimension.
     *
     * ```js
     *  const model = tf.sequential();
     *  model.add(tf.layers.repeatVector({n: 4, inputShape: [2]}));
     *  const x = tf.tensor2d([[10, 20]]);
     *  // Use the model to do inference on a data point the model hasn't see
     *  model.predict(x).print();
     *  // output shape is now [batch, 2, 4]
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Basic', namespace: 'layers'}
     */
    function repeatVector(args) {
        return new RepeatVector(args);
    }
    /**
     * Reshapes an input to a certain shape.
     *
     * ```js
     * const input = tf.input({shape: [4, 3]});
     * const reshapeLayer = tf.layers.reshape({targetShape: [2, 6]});
     * // Inspect the inferred output shape of the Reshape layer, which
     * // equals `[null, 2, 6]`. (The 1st dimension is the undermined batch size.)
     * console.log(JSON.stringify(reshapeLayer.apply(input).shape));
     * ```
     *
     * Input shape:
     *   Arbitrary, although all dimensions in the input shape must be fixed.
     *   Use the configuration `inputShape` when using this layer as the
     *   first layer in a model.
     *
     *
     * Output shape:
     *   [batchSize, targetShape[0], targetShape[1], ...,
     *    targetShape[targetShape.length - 1]].
     *
     * @doc {heading: 'Layers', subheading: 'Basic', namespace: 'layers'}
     */
    function reshape$2(args) {
        return new Reshape(args);
    }
    /**
     * Permutes the dimensions of the input according to a given pattern.
     *
     * Useful for, e.g., connecting RNNs and convnets together.
     *
     * Example:
     *
     * ```js
     * const model = tf.sequential();
     * model.add(tf.layers.permute({
     *   dims: [2, 1],
     *   inputShape: [10, 64]
     * }));
     * console.log(model.outputShape);
     * // Now model's output shape is [null, 64, 10], where null is the
     * // unpermuted sample (batch) dimension.
     * ```
     *
     * Input shape:
     *   Arbitrary. Use the configuration field `inputShape` when using this
     *   layer as the first layer in a model.
     *
     * Output shape:
     *   Same rank as the input shape, but with the dimensions re-ordered (i.e.,
     *   permuted) according to the `dims` configuration of this layer.
     *
     * @doc {heading: 'Layers', subheading: 'Basic', namespace: 'layers'}
     */
    function permute(args) {
        return new Permute(args);
    }
    /**
     * Maps positive integers (indices) into dense vectors of fixed size.
     * eg. [[4], [20]] -> [[0.25, 0.1], [0.6, -0.2]]
     *
     * **Input shape:** 2D tensor with shape: `[batchSize, sequenceLength]`.
     *
     * **Output shape:** 3D tensor with shape: `[batchSize, sequenceLength,
     * outputDim]`.
     *
     * @doc {heading: 'Layers', subheading: 'Basic', namespace: 'layers'}
     */
    function embedding(args) {
        return new Embedding(args);
    }
    // Merge Layers.
    /**
     * Layer that performs element-wise addition on an `Array` of inputs.
     *
     * It takes as input a list of tensors, all of the same shape, and returns a
     * single tensor (also of the same shape). The inputs are specified as an
     * `Array` when the `apply` method of the `Add` layer instance is called. For
     * example:
     *
     * ```js
     * const input1 = tf.input({shape: [2, 2]});
     * const input2 = tf.input({shape: [2, 2]});
     * const addLayer = tf.layers.add();
     * const sum = addLayer.apply([input1, input2]);
     * console.log(JSON.stringify(sum.shape));
     * // You get [null, 2, 2], with the first dimension as the undetermined batch
     * // dimension.
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Merge', namespace: 'layers'}
     */
    function add$1(args) {
        return new Add(args);
    }
    /**
     * Layer that performs element-wise averaging on an `Array` of inputs.
     *
     * It takes as input a list of tensors, all of the same shape, and returns a
     * single tensor (also of the same shape). For example:
     *
     * ```js
     * const input1 = tf.input({shape: [2, 2]});
     * const input2 = tf.input({shape: [2, 2]});
     * const averageLayer = tf.layers.average();
     * const average = averageLayer.apply([input1, input2]);
     * console.log(JSON.stringify(average.shape));
     * // You get [null, 2, 2], with the first dimension as the undetermined batch
     * // dimension.
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Merge', namespace: 'layers'}
     */
    function average(args) {
        return new Average(args);
    }
    /**
     * Layer that concatenates an `Array` of inputs.
     *
     * It takes a list of tensors, all of the same shape except for the
     * concatenation axis, and returns a single tensor, the concatenation
     * of all inputs. For example:
     *
     * ```js
     * const input1 = tf.input({shape: [2, 2]});
     * const input2 = tf.input({shape: [2, 3]});
     * const concatLayer = tf.layers.concatenate();
     * const output = concatLayer.apply([input1, input2]);
     * console.log(JSON.stringify(output.shape));
     * // You get [null, 2, 5], with the first dimension as the undetermined batch
     * // dimension. The last dimension (5) is the result of concatenating the
     * // last dimensions of the inputs (2 and 3).
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Merge', namespace: 'layers'}
     */
    function concatenate(args) {
        return new Concatenate(args);
    }
    /**
     * Layer that computes the element-wise maximum an `Array` of inputs.
     *
     * It takes as input a list of tensors, all of the same shape and returns a
     * single tensor (also of the same shape). For example:
     *
     * ```js
     * const input1 = tf.input({shape: [2, 2]});
     * const input2 = tf.input({shape: [2, 2]});
     * const maxLayer = tf.layers.maximum();
     * const max = maxLayer.apply([input1, input2]);
     * console.log(JSON.stringify(max.shape));
     * // You get [null, 2, 2], with the first dimension as the undetermined batch
     * // dimension.
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Merge', namespace: 'layers'}
     */
    function maximum$2(args) {
        return new Maximum(args);
    }
    /**
     * Layer that computes the element-wise minimum of an `Array` of inputs.
     *
     * It takes as input a list of tensors, all of the same shape and returns a
     * single tensor (also of the same shape). For example:
     *
     * ```js
     * const input1 = tf.input({shape: [2, 2]});
     * const input2 = tf.input({shape: [2, 2]});
     * const minLayer = tf.layers.minimum();
     * const min = minLayer.apply([input1, input2]);
     * console.log(JSON.stringify(min.shape));
     * // You get [null, 2, 2], with the first dimension as the undetermined batch
     * // dimension.
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Merge', namespace: 'layers'}
     */
    function minimum$2(args) {
        return new Minimum(args);
    }
    /**
     * Layer that multiplies (element-wise) an `Array` of inputs.
     *
     * It takes as input an Array of tensors, all of the same
     * shape, and returns a single tensor (also of the same shape).
     * For example:
     *
     * ```js
     * const input1 = tf.input({shape: [2, 2]});
     * const input2 = tf.input({shape: [2, 2]});
     * const input3 = tf.input({shape: [2, 2]});
     * const multiplyLayer = tf.layers.multiply();
     * const product = multiplyLayer.apply([input1, input2, input3]);
     * console.log(product.shape);
     * // You get [null, 2, 2], with the first dimension as the undetermined batch
     * // dimension.
     *
     * @doc {heading: 'Layers', subheading: 'Merge', namespace: 'layers'}
     */
    function multiply$2(args) {
        return new Multiply(args);
    }
    /**
     * Layer that computes a dot product between samples in two tensors.
     *
     * E.g., if applied to a list of two tensors `a` and `b` both of shape
     * `[batchSize, n]`, the output will be a tensor of shape `[batchSize, 1]`,
     * where each entry at index `[i, 0]` will be the dot product between
     * `a[i, :]` and `b[i, :]`.
     *
     * Example:
     *
     * ```js
     * const dotLayer = tf.layers.dot({axes: -1});
     * const x1 = tf.tensor2d([[10, 20], [30, 40]]);
     * const x2 = tf.tensor2d([[-1, -2], [-3, -4]]);
     *
     * // Invoke the layer's apply() method in eager (imperative) mode.
     * const y = dotLayer.apply([x1, x2]);
     * y.print();
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Merge', namespace: 'layers'}
     */
    function dot(args) {
        return new Dot(args);
    }
    // Normalization Layers.
    /**
     * Batch normalization layer (Ioffe and Szegedy, 2014).
     *
     * Normalize the activations of the previous layer at each batch,
     * i.e. applies a transformation that maintains the mean activation
     * close to 0 and the activation standard deviation close to 1.
     *
     * Input shape:
     *   Arbitrary. Use the keyword argument `inputShape` (Array of integers, does
     *   not include the sample axis) when calling the constructor of this class,
     *   if this layer is used as a first layer in a model.
     *
     * Output shape:
     *   Same shape as input.
     *
     * References:
     *   - [Batch Normalization: Accelerating Deep Network Training by Reducing
     * Internal Covariate Shift](https://arxiv.org/abs/1502.03167)
     *
     * @doc {heading: 'Layers', subheading: 'Normalization', namespace: 'layers'}
     */
    function batchNormalization(args) {
        return new BatchNormalization(args);
    }
    /**
     * Layer-normalization layer (Ba et al., 2016).
     *
     * Normalizes the activations of the previous layer for each given example in a
     * batch independently, instead of across a batch like in `batchNormalization`.
     * In other words, this layer applies a transformation that maintanis the mean
     * activation within each example close to0 and activation variance close to 1.
     *
     * Input shape:
     *   Arbitrary. Use the argument `inputShape` when using this layer as the first
     *   layer in a model.
     *
     * Output shape:
     *   Same as input.
     *
     * References:
     *   - [Layer Normalization](https://arxiv.org/abs/1607.06450)
     *
     * @doc {heading: 'Layers', subheading: 'Normalization', namespace: 'layers'}
     */
    function layerNormalization(args) {
        return new LayerNormalization(args);
    }
    // Padding Layers.
    /**
     * Zero-padding layer for 2D input (e.g., image).
     *
     * This layer can add rows and columns of zeros
     * at the top, bottom, left and right side of an image tensor.
     *
     * Input shape:
     *   4D tensor with shape:
     *   - If `dataFormat` is `"channelsLast"`:
     *     `[batch, rows, cols, channels]`
     *   - If `data_format` is `"channels_first"`:
     *     `[batch, channels, rows, cols]`.
     *
     * Output shape:
     *   4D with shape:
     *   - If `dataFormat` is `"channelsLast"`:
     *     `[batch, paddedRows, paddedCols, channels]`
     *    - If `dataFormat` is `"channelsFirst"`:
     *     `[batch, channels, paddedRows, paddedCols]`.
     *
     * @doc {heading: 'Layers', subheading: 'Padding', namespace: 'layers'}
     */
    function zeroPadding2d(args) {
        return new ZeroPadding2D(args);
    }
    // Pooling Layers.
    /**
     * Average pooling operation for spatial data.
     *
     * Input shape: `[batchSize, inLength, channels]`
     *
     * Output shape: `[batchSize, pooledLength, channels]`
     *
     * `tf.avgPool1d` is an alias.
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function averagePooling1d(args) {
        return new AveragePooling1D(args);
    }
    function avgPool1d(args) {
        return averagePooling1d(args);
    }
    // For backwards compatibility.
    // See https://github.com/tensorflow/tfjs/issues/152
    function avgPooling1d(args) {
        return averagePooling1d(args);
    }
    /**
     * Average pooling operation for spatial data.
     *
     * Input shape:
     *  - If `dataFormat === CHANNEL_LAST`:
     *      4D tensor with shape:
     *      `[batchSize, rows, cols, channels]`
     *  - If `dataFormat === CHANNEL_FIRST`:
     *      4D tensor with shape:
     *      `[batchSize, channels, rows, cols]`
     *
     * Output shape
     *  - If `dataFormat === CHANNEL_LAST`:
     *      4D tensor with shape:
     *      `[batchSize, pooleRows, pooledCols, channels]`
     *  - If `dataFormat === CHANNEL_FIRST`:
     *      4D tensor with shape:
     *      `[batchSize, channels, pooleRows, pooledCols]`
     *
     * `tf.avgPool2d` is an alias.
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function averagePooling2d(args) {
        return new AveragePooling2D(args);
    }
    function avgPool2d(args) {
        return averagePooling2d(args);
    }
    // For backwards compatibility.
    // See https://github.com/tensorflow/tfjs/issues/152
    function avgPooling2d(args) {
        return averagePooling2d(args);
    }
    /**
     * Average pooling operation for 3D data.
     *
     * Input shape
     *   - If `dataFormat === channelsLast`:
     *       5D tensor with shape:
     *       `[batchSize, depths, rows, cols, channels]`
     *   - If `dataFormat === channelsFirst`:
     *      4D tensor with shape:
     *       `[batchSize, channels, depths, rows, cols]`
     *
     * Output shape
     *   - If `dataFormat=channelsLast`:
     *       5D tensor with shape:
     *       `[batchSize, pooledDepths, pooledRows, pooledCols, channels]`
     *   - If `dataFormat=channelsFirst`:
     *       5D tensor with shape:
     *       `[batchSize, channels, pooledDepths, pooledRows, pooledCols]`
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function averagePooling3d(args) {
        return new AveragePooling3D(args);
    }
    function avgPool3d(args) {
        return averagePooling3d(args);
    }
    // For backwards compatibility.
    // See https://github.com/tensorflow/tfjs/issues/152
    function avgPooling3d(args) {
        return averagePooling3d(args);
    }
    /**
     * Global average pooling operation for temporal data.
     *
     * Input Shape: 3D tensor with shape: `[batchSize, steps, features]`.
     *
     * Output Shape:2D tensor with shape: `[batchSize, features]`.
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function globalAveragePooling1d(args) {
        return new GlobalAveragePooling1D(args);
    }
    /**
     * Global average pooling operation for spatial data.
     *
     * Input shape:
     *   - If `dataFormat` is `CHANNEL_LAST`:
     *       4D tensor with shape: `[batchSize, rows, cols, channels]`.
     *   - If `dataFormat` is `CHANNEL_FIRST`:
     *       4D tensor with shape: `[batchSize, channels, rows, cols]`.
     *
     * Output shape:
     *   2D tensor with shape: `[batchSize, channels]`.
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function globalAveragePooling2d(args) {
        return new GlobalAveragePooling2D(args);
    }
    /**
     * Global max pooling operation for temporal data.
     *
     * Input Shape: 3D tensor with shape: `[batchSize, steps, features]`.
     *
     * Output Shape:2D tensor with shape: `[batchSize, features]`.
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function globalMaxPooling1d(args) {
        return new GlobalMaxPooling1D(args);
    }
    /**
     * Global max pooling operation for spatial data.
     *
     * Input shape:
     *   - If `dataFormat` is `CHANNEL_LAST`:
     *       4D tensor with shape: `[batchSize, rows, cols, channels]`.
     *   - If `dataFormat` is `CHANNEL_FIRST`:
     *       4D tensor with shape: `[batchSize, channels, rows, cols]`.
     *
     * Output shape:
     *   2D tensor with shape: `[batchSize, channels]`.
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function globalMaxPooling2d(args) {
        return new GlobalMaxPooling2D(args);
    }
    /**
     * Max pooling operation for temporal data.
     *
     * Input shape:  `[batchSize, inLength, channels]`
     *
     * Output shape: `[batchSize, pooledLength, channels]`
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function maxPooling1d(args) {
        return new MaxPooling1D(args);
    }
    /**
     * Max pooling operation for spatial data.
     *
     * Input shape
     *   - If `dataFormat === CHANNEL_LAST`:
     *       4D tensor with shape:
     *       `[batchSize, rows, cols, channels]`
     *   - If `dataFormat === CHANNEL_FIRST`:
     *      4D tensor with shape:
     *       `[batchSize, channels, rows, cols]`
     *
     * Output shape
     *   - If `dataFormat=CHANNEL_LAST`:
     *       4D tensor with shape:
     *       `[batchSize, pooleRows, pooledCols, channels]`
     *   - If `dataFormat=CHANNEL_FIRST`:
     *       4D tensor with shape:
     *       `[batchSize, channels, pooleRows, pooledCols]`
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function maxPooling2d(args) {
        return new MaxPooling2D(args);
    }
    /**
     * Max pooling operation for 3D data.
     *
     * Input shape
     *   - If `dataFormat === channelsLast`:
     *       5D tensor with shape:
     *       `[batchSize, depths, rows, cols, channels]`
     *   - If `dataFormat === channelsFirst`:
     *      5D tensor with shape:
     *       `[batchSize, channels, depths, rows, cols]`
     *
     * Output shape
     *   - If `dataFormat=channelsLast`:
     *       5D tensor with shape:
     *       `[batchSize, pooledDepths, pooledRows, pooledCols, channels]`
     *   - If `dataFormat=channelsFirst`:
     *       5D tensor with shape:
     *       `[batchSize, channels, pooledDepths, pooledRows, pooledCols]`
     *
     * @doc {heading: 'Layers', subheading: 'Pooling', namespace: 'layers'}
     */
    function maxPooling3d(args) {
        return new MaxPooling3D(args);
    }
    // Recurrent Layers.
    /**
     * Gated Recurrent Unit - Cho et al. 2014.
     *
     * This is an `RNN` layer consisting of one `GRUCell`. However, unlike
     * the underlying `GRUCell`, the `apply` method of `SimpleRNN` operates
     * on a sequence of inputs. The shape of the input (not including the first,
     * batch dimension) needs to be at least 2-D, with the first dimension being
     * time steps. For example:
     *
     * ```js
     * const rnn = tf.layers.gru({units: 8, returnSequences: true});
     *
     * // Create an input with 10 time steps.
     * const input = tf.input({shape: [10, 20]});
     * const output = rnn.apply(input);
     *
     * console.log(JSON.stringify(output.shape));
     * // [null, 10, 8]: 1st dimension is unknown batch size; 2nd dimension is the
     * // same as the sequence length of `input`, due to `returnSequences`: `true`;
     * // 3rd dimension is the `GRUCell`'s number of units.
     *
     * @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'}
     */
    function gru(args) {
        return new GRU(args);
    }
    /**
     * Cell class for `GRU`.
     *
     * `GRUCell` is distinct from the `RNN` subclass `GRU` in that its
     * `apply` method takes the input data of only a single time step and returns
     * the cell's output at the time step, while `GRU` takes the input data
     * over a number of time steps. For example:
     *
     * ```js
     * const cell = tf.layers.gruCell({units: 2});
     * const input = tf.input({shape: [10]});
     * const output = cell.apply(input);
     *
     * console.log(JSON.stringify(output.shape));
     * // [null, 10]: This is the cell's output at a single time step. The 1st
     * // dimension is the unknown batch size.
     * ```
     *
     * Instance(s) of `GRUCell` can be used to construct `RNN` layers. The
     * most typical use of this workflow is to combine a number of cells into a
     * stacked RNN cell (i.e., `StackedRNNCell` internally) and use it to create an
     * RNN. For example:
     *
     * ```js
     * const cells = [
     *   tf.layers.gruCell({units: 4}),
     *   tf.layers.gruCell({units: 8}),
     * ];
     * const rnn = tf.layers.rnn({cell: cells, returnSequences: true});
     *
     * // Create an input with 10 time steps and a length-20 vector at each step.
     * const input = tf.input({shape: [10, 20]});
     * const output = rnn.apply(input);
     *
     * console.log(JSON.stringify(output.shape));
     * // [null, 10, 8]: 1st dimension is unknown batch size; 2nd dimension is the
     * // same as the sequence length of `input`, due to `returnSequences`: `true`;
     * // 3rd dimension is the last `gruCell`'s number of units.
     * ```
     *
     * To create an `RNN` consisting of only *one* `GRUCell`, use the
     * `tf.layers.gru`.
     *
     * @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'}
     */
    function gruCell(args) {
        return new GRUCell(args);
    }
    /**
     * Long-Short Term Memory layer - Hochreiter 1997.
     *
     * This is an `RNN` layer consisting of one `LSTMCell`. However, unlike
     * the underlying `LSTMCell`, the `apply` method of `LSTM` operates
     * on a sequence of inputs. The shape of the input (not including the first,
     * batch dimension) needs to be at least 2-D, with the first dimension being
     * time steps. For example:
     *
     * ```js
     * const lstm = tf.layers.lstm({units: 8, returnSequences: true});
     *
     * // Create an input with 10 time steps.
     * const input = tf.input({shape: [10, 20]});
     * const output = lstm.apply(input);
     *
     * console.log(JSON.stringify(output.shape));
     * // [null, 10, 8]: 1st dimension is unknown batch size; 2nd dimension is the
     * // same as the sequence length of `input`, due to `returnSequences`: `true`;
     * // 3rd dimension is the `LSTMCell`'s number of units.
     *
     * @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'}
     */
    function lstm(args) {
        return new LSTM(args);
    }
    /**
     * Cell class for `LSTM`.
     *
     * `LSTMCell` is distinct from the `RNN` subclass `LSTM` in that its
     * `apply` method takes the input data of only a single time step and returns
     * the cell's output at the time step, while `LSTM` takes the input data
     * over a number of time steps. For example:
     *
     * ```js
     * const cell = tf.layers.lstmCell({units: 2});
     * const input = tf.input({shape: [10]});
     * const output = cell.apply(input);
     *
     * console.log(JSON.stringify(output.shape));
     * // [null, 10]: This is the cell's output at a single time step. The 1st
     * // dimension is the unknown batch size.
     * ```
     *
     * Instance(s) of `LSTMCell` can be used to construct `RNN` layers. The
     * most typical use of this workflow is to combine a number of cells into a
     * stacked RNN cell (i.e., `StackedRNNCell` internally) and use it to create an
     * RNN. For example:
     *
     * ```js
     * const cells = [
     *   tf.layers.lstmCell({units: 4}),
     *   tf.layers.lstmCell({units: 8}),
     * ];
     * const rnn = tf.layers.rnn({cell: cells, returnSequences: true});
     *
     * // Create an input with 10 time steps and a length-20 vector at each step.
     * const input = tf.input({shape: [10, 20]});
     * const output = rnn.apply(input);
     *
     * console.log(JSON.stringify(output.shape));
     * // [null, 10, 8]: 1st dimension is unknown batch size; 2nd dimension is the
     * // same as the sequence length of `input`, due to `returnSequences`: `true`;
     * // 3rd dimension is the last `lstmCell`'s number of units.
     * ```
     *
     * To create an `RNN` consisting of only *one* `LSTMCell`, use the
     * `tf.layers.lstm`.
     *
     * @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'}
     */
    function lstmCell(args) {
        return new LSTMCell(args);
    }
    /**
     * Fully-connected RNN where the output is to be fed back to input.
     *
     * This is an `RNN` layer consisting of one `SimpleRNNCell`. However, unlike
     * the underlying `SimpleRNNCell`, the `apply` method of `SimpleRNN` operates
     * on a sequence of inputs. The shape of the input (not including the first,
     * batch dimension) needs to be at least 2-D, with the first dimension being
     * time steps. For example:
     *
     * ```js
     * const rnn = tf.layers.simpleRNN({units: 8, returnSequences: true});
     *
     * // Create an input with 10 time steps.
     * const input = tf.input({shape: [10, 20]});
     * const output = rnn.apply(input);
     *
     * console.log(JSON.stringify(output.shape));
     * // [null, 10, 8]: 1st dimension is unknown batch size; 2nd dimension is the
     * // same as the sequence length of `input`, due to `returnSequences`: `true`;
     * // 3rd dimension is the `SimpleRNNCell`'s number of units.
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'}
     */
    function simpleRNN(args) {
        return new SimpleRNN(args);
    }
    /**
     * Cell class for `SimpleRNN`.
     *
     * `SimpleRNNCell` is distinct from the `RNN` subclass `SimpleRNN` in that its
     * `apply` method takes the input data of only a single time step and returns
     * the cell's output at the time step, while `SimpleRNN` takes the input data
     * over a number of time steps. For example:
     *
     * ```js
     * const cell = tf.layers.simpleRNNCell({units: 2});
     * const input = tf.input({shape: [10]});
     * const output = cell.apply(input);
     *
     * console.log(JSON.stringify(output.shape));
     * // [null, 10]: This is the cell's output at a single time step. The 1st
     * // dimension is the unknown batch size.
     * ```
     *
     * Instance(s) of `SimpleRNNCell` can be used to construct `RNN` layers. The
     * most typical use of this workflow is to combine a number of cells into a
     * stacked RNN cell (i.e., `StackedRNNCell` internally) and use it to create an
     * RNN. For example:
     *
     * ```js
     * const cells = [
     *   tf.layers.simpleRNNCell({units: 4}),
     *   tf.layers.simpleRNNCell({units: 8}),
     * ];
     * const rnn = tf.layers.rnn({cell: cells, returnSequences: true});
     *
     * // Create an input with 10 time steps and a length-20 vector at each step.
     * const input = tf.input({shape: [10, 20]});
     * const output = rnn.apply(input);
     *
     * console.log(JSON.stringify(output.shape));
     * // [null, 10, 8]: 1st dimension is unknown batch size; 2nd dimension is the
     * // same as the sequence length of `input`, due to `returnSequences`: `true`;
     * // 3rd dimension is the last `SimpleRNNCell`'s number of units.
     * ```
     *
     * To create an `RNN` consisting of only *one* `SimpleRNNCell`, use the
     * `tf.layers.simpleRNN`.
     *
     * @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'}
     */
    function simpleRNNCell(args) {
        return new SimpleRNNCell(args);
    }
    /**
     * Convolutional LSTM layer - Xingjian Shi 2015.
     *
     * This is an `ConvRNN2D` layer consisting of one `ConvLSTM2DCell`. However,
     * unlike the underlying `ConvLSTM2DCell`, the `apply` method of `ConvLSTM2D`
     * operates on a sequence of inputs. The shape of the input (not including the
     * first, batch dimension) needs to be 4-D, with the first dimension being time
     * steps. For example:
     *
     * ```js
     * const filters = 3;
     * const kernelSize = 3;
     *
     * const batchSize = 4;
     * const sequenceLength = 2;
     * const size = 5;
     * const channels = 3;
     *
     * const inputShape = [batchSize, sequenceLength, size, size, channels];
     * const input = tf.ones(inputShape);
     *
     * const layer = tf.layers.convLstm2d({filters, kernelSize});
     *
     * const output = layer.apply(input);
     * ```
     */
    /** @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'} */
    function convLstm2d(args) {
        return new ConvLSTM2D(args);
    }
    /**
     * Cell class for `ConvLSTM2D`.
     *
     * `ConvLSTM2DCell` is distinct from the `ConvRNN2D` subclass `ConvLSTM2D` in
     * that its `call` method takes the input data of only a single time step and
     * returns the cell's output at the time step, while `ConvLSTM2D` takes the
     * input data over a number of time steps. For example:
     *
     * ```js
     * const filters = 3;
     * const kernelSize = 3;
     *
     * const sequenceLength = 1;
     * const size = 5;
     * const channels = 3;
     *
     * const inputShape = [sequenceLength, size, size, channels];
     * const input = tf.ones(inputShape);
     *
     * const cell = tf.layers.convLstm2dCell({filters, kernelSize});
     *
     * cell.build(input.shape);
     *
     * const outputSize = size - kernelSize + 1;
     * const outShape = [sequenceLength, outputSize, outputSize, filters];
     *
     * const initialH = tf.zeros(outShape);
     * const initialC = tf.zeros(outShape);
     *
     * const [o, h, c] = cell.call([input, initialH, initialC], {});
     * ```
     */
    /** @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'} */
    function convLstm2dCell(args) {
        return new ConvLSTM2DCell(args);
    }
    /**
     * Base class for recurrent layers.
     *
     * Input shape:
     *   3D tensor with shape `[batchSize, timeSteps, inputDim]`.
     *
     * Output shape:
     *   - if `returnState`, an Array of tensors (i.e., `tf.Tensor`s). The first
     *     tensor is the output. The remaining tensors are the states at the
     *     last time step, each with shape `[batchSize, units]`.
     *   - if `returnSequences`, the output will have shape
     *     `[batchSize, timeSteps, units]`.
     *   - else, the output will have shape `[batchSize, units]`.
     *
     * Masking:
     *   This layer supports masking for input data with a variable number
     *   of timesteps. To introduce masks to your data,
     *   use an embedding layer with the `mask_zero` parameter
     *   set to `True`.
     *
     * Notes on using statefulness in RNNs:
     *   You can set RNN layers to be 'stateful', which means that the states
     *   computed for the samples in one batch will be reused as initial states
     *   for the samples in the next batch. This assumes a one-to-one mapping
     *   between samples in different successive batches.
     *
     *   To enable statefulness:
     *     - specify `stateful: true` in the layer constructor.
     *     - specify a fixed batch size for your model, by passing
     *       if sequential model:
     *         `batchInputShape=[...]` to the first layer in your model.
     *       else for functional model with 1 or more Input layers:
     *         `batchShape=[...]` to all the first layers in your model.
     *       This is the expected shape of your inputs *including the batch size*.
     *       It should be a tuple of integers, e.g. `(32, 10, 100)`.
     *     - specify `shuffle=False` when calling fit().
     *
     *   To reset the states of your model, call `.resetStates()` on either
     *   a specific layer, or on your entire model.
     *
     * Note on specifying the initial state of RNNs
     *   You can specify the initial state of RNN layers symbolically by
     *   calling them with the option `initialState`. The value of
     *   `initialState` should be a tensor or list of tensors representing
     *   the initial state of the RNN layer.
     *
     *   You can specify the initial state of RNN layers numerically by
     *   calling `resetStates` with the keyword argument `states`. The value of
     *   `states` should be a numpy array or list of numpy arrays representing
     *   the initial state of the RNN layer.
     *
     * Note on passing external constants to RNNs
     *   You can pass "external" constants to the cell using the `constants`
     *   keyword argument of `RNN.call` method. This requires that the `cell.call`
     *   method accepts the same keyword argument `constants`. Such constants
     *   can be used to conditon the cell transformation on additional static inputs
     *   (not changing over time), a.k.a an attention mechanism.
     *
     * @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'}
     */
    function rnn(args) {
        return new RNN(args);
    }
    /**
     * Wrapper allowing a stack of RNN cells to behave as a single cell.
     *
     * Used to implement efficient stacked RNNs.
     *
     * @doc {heading: 'Layers', subheading: 'Recurrent', namespace: 'layers'}
     */
    function stackedRNNCells(args) {
        return new StackedRNNCells(args);
    }
    // Wrapper Layers.
    /** @doc {heading: 'Layers', subheading: 'Wrapper', namespace: 'layers'} */
    function bidirectional(args) {
        return new Bidirectional(args);
    }
    /**
     * This wrapper applies a layer to every temporal slice of an input.
     *
     * The input should be at least 3D,  and the dimension of the index `1` will be
     * considered to be the temporal dimension.
     *
     * Consider a batch of 32 samples, where each sample is a sequence of 10 vectors
     * of 16 dimensions. The batch input shape of the layer is then `[32,  10,
     * 16]`, and the `inputShape`, not including the sample dimension, is
     * `[10, 16]`.
     *
     * You can then use `TimeDistributed` to apply a `Dense` layer to each of the 10
     * timesteps, independently:
     *
     * ```js
     * const model = tf.sequential();
     * model.add(tf.layers.timeDistributed({
     *   layer: tf.layers.dense({units: 8}),
     *   inputShape: [10, 16],
     * }));
     *
     * // Now model.outputShape = [null, 10, 8].
     * // The output will then have shape `[32, 10, 8]`.
     *
     * // In subsequent layers, there is no need for `inputShape`:
     * model.add(tf.layers.timeDistributed({layer: tf.layers.dense({units: 32})}));
     * console.log(JSON.stringify(model.outputs[0].shape));
     * // Now model.outputShape = [null, 10, 32].
     * ```
     *
     * The output will then have shape `[32, 10, 32]`.
     *
     * `TimeDistributed` can be used with arbitrary layers, not just `Dense`, for
     * instance a `Conv2D` layer.
     *
     * ```js
     * const model = tf.sequential();
     * model.add(tf.layers.timeDistributed({
     *   layer: tf.layers.conv2d({filters: 64, kernelSize: [3, 3]}),
     *   inputShape: [10, 299, 299, 3],
     * }));
     * console.log(JSON.stringify(model.outputs[0].shape));
     * ```
     *
     * @doc {heading: 'Layers', subheading: 'Wrapper', namespace: 'layers'}
     */
    function timeDistributed(args) {
        return new TimeDistributed(args);
    }
    // Aliases for pooling.
    const globalMaxPool1d = globalMaxPooling1d;
    const globalMaxPool2d = globalMaxPooling2d;
    const maxPool1d = maxPooling1d;
    const maxPool2d = maxPooling2d;
    /**
     * Apply additive zero-centered Gaussian noise.
     *
     * As it is a regularization layer, it is only active at training time.
     *
     * This is useful to mitigate overfitting
     * (you could see it as a form of random data augmentation).
     * Gaussian Noise (GS) is a natural choice as corruption process
     * for real valued inputs.
     *
     * # Arguments
     *     stddev: float, standard deviation of the noise distribution.
     *
     * # Input shape
     *         Arbitrary. Use the keyword argument `input_shape`
     *         (tuple of integers, does not include the samples axis)
     *         when using this layer as the first layer in a model.
     *
     * # Output shape
     *         Same shape as input.
     *
     * @doc {heading: 'Layers', subheading: 'Noise', namespace: 'layers'}
     */
    function gaussianNoise(args) {
        return new GaussianNoise(args);
    }
    /**
     * Apply multiplicative 1-centered Gaussian noise.
     *
     * As it is a regularization layer, it is only active at training time.
     *
     * Arguments:
     *   - `rate`: float, drop probability (as with `Dropout`).
     *     The multiplicative noise will have
     *     standard deviation `sqrt(rate / (1 - rate))`.
     *
     * Input shape:
     *   Arbitrary. Use the keyword argument `inputShape`
     *   (tuple of integers, does not include the samples axis)
     *   when using this layer as the first layer in a model.
     *
     * Output shape:
     *   Same shape as input.
     *
     * References:
     *   - [Dropout: A Simple Way to Prevent Neural Networks from Overfitting](
     *      http://www.cs.toronto.edu/~rsalakhu/papers/srivastava14a.pdf)
     *
     * @doc {heading: 'Layers', subheading: 'Noise', namespace: 'layers'}
     */
    function gaussianDropout(args) {
        return new GaussianDropout(args);
    }
    /**
     * Applies Alpha Dropout to the input.
     *
     * As it is a regularization layer, it is only active at training time.
     *
     * Alpha Dropout is a `Dropout` that keeps mean and variance of inputs
     * to their original values, in order to ensure the self-normalizing property
     * even after this dropout.
     * Alpha Dropout fits well to Scaled Exponential Linear Units
     * by randomly setting activations to the negative saturation value.
     *
     * Arguments:
     *   - `rate`: float, drop probability (as with `Dropout`).
     *     The multiplicative noise will have
     *     standard deviation `sqrt(rate / (1 - rate))`.
     *   - `noise_shape`: A 1-D `Tensor` of type `int32`, representing the
     *     shape for randomly generated keep/drop flags.
     *
     * Input shape:
     *   Arbitrary. Use the keyword argument `inputShape`
     *   (tuple of integers, does not include the samples axis)
     *   when using this layer as the first layer in a model.
     *
     * Output shape:
     *   Same shape as input.
     *
     * References:
     *   - [Self-Normalizing Neural Networks](https://arxiv.org/abs/1706.02515)
     *
     * @doc {heading: 'Layers', subheading: 'Noise', namespace: 'layers'}
     */
    function alphaDropout(args) {
        return new AlphaDropout(args);
    }
    /**
     * Masks a sequence by using a mask value to skip timesteps.
     *
     * If all features for a given sample timestep are equal to `mask_value`,
     * then the sample timestep will be masked (skipped) in all downstream layers
     * (as long as they support masking).
     *
     * If any downstream layer does not support masking yet receives such
     * an input mask, an exception will be raised.
     *
     * Arguments:
     *   - `maskValue`: Either None or mask value to skip.
     *
     * Input shape:
     *   Arbitrary. Use the keyword argument `inputShape`
     *   (tuple of integers, does not include the samples axis)
     *   when using this layer as the first layer in a model.
     *
     * Output shape:
     *   Same shape as input.
     *
     * @doc {heading: 'Layers', subheading: 'Mask', namespace: 'layers'}
     */
    function masking(args) {
        return new Masking(args);
    }

    var exports_layers = /*#__PURE__*/Object.freeze({
        __proto__: null,
        inputLayer: inputLayer,
        elu: elu$2,
        reLU: reLU,
        leakyReLU: leakyReLU,
        prelu: prelu$2,
        softmax: softmax$2,
        thresholdedReLU: thresholdedReLU,
        conv1d: conv1d,
        conv2d: conv2d$1,
        conv2dTranspose: conv2dTranspose,
        conv3d: conv3d,
        conv3dTranspose: conv3dTranspose,
        separableConv2d: separableConv2d,
        cropping2D: cropping2D,
        upSampling2d: upSampling2d,
        depthwiseConv2d: depthwiseConv2d,
        activation: activation,
        dense: dense,
        dropout: dropout,
        spatialDropout1d: spatialDropout1d,
        flatten: flatten$1,
        repeatVector: repeatVector,
        reshape: reshape$2,
        permute: permute,
        embedding: embedding,
        add: add$1,
        average: average,
        concatenate: concatenate,
        maximum: maximum$2,
        minimum: minimum$2,
        multiply: multiply$2,
        dot: dot,
        batchNormalization: batchNormalization,
        layerNormalization: layerNormalization,
        zeroPadding2d: zeroPadding2d,
        averagePooling1d: averagePooling1d,
        avgPool1d: avgPool1d,
        avgPooling1d: avgPooling1d,
        averagePooling2d: averagePooling2d,
        avgPool2d: avgPool2d,
        avgPooling2d: avgPooling2d,
        averagePooling3d: averagePooling3d,
        avgPool3d: avgPool3d,
        avgPooling3d: avgPooling3d,
        globalAveragePooling1d: globalAveragePooling1d,
        globalAveragePooling2d: globalAveragePooling2d,
        globalMaxPooling1d: globalMaxPooling1d,
        globalMaxPooling2d: globalMaxPooling2d,
        maxPooling1d: maxPooling1d,
        maxPooling2d: maxPooling2d,
        maxPooling3d: maxPooling3d,
        gru: gru,
        gruCell: gruCell,
        lstm: lstm,
        lstmCell: lstmCell,
        simpleRNN: simpleRNN,
        simpleRNNCell: simpleRNNCell,
        convLstm2d: convLstm2d,
        convLstm2dCell: convLstm2dCell,
        rnn: rnn,
        stackedRNNCells: stackedRNNCells,
        bidirectional: bidirectional,
        timeDistributed: timeDistributed,
        globalMaxPool1d: globalMaxPool1d,
        globalMaxPool2d: globalMaxPool2d,
        maxPool1d: maxPool1d,
        maxPool2d: maxPool2d,
        Layer: Layer,
        RNN: RNN,
        RNNCell: RNNCell,
        input: input,
        gaussianNoise: gaussianNoise,
        gaussianDropout: gaussianDropout,
        alphaDropout: alphaDropout,
        masking: masking
    });

    /**
     * Binary accuracy metric function.
     *
     * `yTrue` and `yPred` can have 0-1 values. Example:
     * ```js
     * const x = tf.tensor2d([[1, 1, 1, 1], [0, 0, 0, 0]], [2, 4]);
     * const y = tf.tensor2d([[1, 0, 1, 0], [0, 0, 0, 1]], [2, 4]);
     * const accuracy = tf.metrics.binaryAccuracy(x, y);
     * accuracy.print();
     * ```
     *
     * `yTrue` and `yPred` can also have floating-number values between 0 and 1, in
     * which case the values will be thresholded at 0.5 to yield 0-1 values (i.e.,
     * a value >= 0.5 and <= 1.0 is interpreted as 1.
     * )
     * Example:
     * ```js
     * const x = tf.tensor1d([1, 1, 1, 1, 0, 0, 0, 0]);
     * const y = tf.tensor1d([0.2, 0.4, 0.6, 0.8, 0.2, 0.3, 0.4, 0.7]);
     * const accuracy = tf.metrics.binaryAccuracy(x, y);
     * accuracy.print();
     * ```
     *
     * @param yTrue Binary Tensor of truth.
     * @param yPred Binary Tensor of prediction.
     * @return Accuracy Tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function binaryAccuracy(yTrue, yPred) {
        return binaryAccuracy$1(yTrue, yPred);
    }
    /**
     * Binary crossentropy metric function.
     *
     * Example:
     * ```js
     * const x = tf.tensor2d([[0], [1], [1], [1]]);
     * const y = tf.tensor2d([[0], [0], [0.5], [1]]);
     * const crossentropy = tf.metrics.binaryCrossentropy(x, y);
     * crossentropy.print();
     * ```
     *
     * @param yTrue Binary Tensor of truth.
     * @param yPred Binary Tensor of prediction, probabilities for the `1` case.
     * @return Accuracy Tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function binaryCrossentropy(yTrue, yPred) {
        return binaryCrossentropy$1(yTrue, yPred);
    }
    /**
     * Sparse categorical accuracy metric function.
     *
     * Example:
     * ```js
     *
     * const yTrue = tf.tensor1d([1, 1, 2, 2, 0]);
     * const yPred = tf.tensor2d(
     *      [[0, 1, 0], [1, 0, 0], [0, 0.4, 0.6], [0, 0.6, 0.4], [0.7, 0.3, 0]]);
     * const crossentropy = tf.metrics.sparseCategoricalAccuracy(yTrue, yPred);
     * crossentropy.print();
     * ```
     *
     * @param yTrue True labels: indices.
     * @param yPred Predicted probabilities or logits.
     * @returns Accuracy tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function sparseCategoricalAccuracy(yTrue, yPred) {
        return sparseCategoricalAccuracy$1(yTrue, yPred);
    }
    /**
     * Categorical accuracy metric function.
     *
     * Example:
     * ```js
     * const x = tf.tensor2d([[0, 0, 0, 1], [0, 0, 0, 1]]);
     * const y = tf.tensor2d([[0.1, 0.8, 0.05, 0.05], [0.1, 0.05, 0.05, 0.8]]);
     * const accuracy = tf.metrics.categoricalAccuracy(x, y);
     * accuracy.print();
     * ```
     *
     * @param yTrue Binary Tensor of truth: one-hot encoding of categories.
     * @param yPred Binary Tensor of prediction: probabilities or logits for the
     *   same categories as in `yTrue`.
     * @return Accuracy Tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function categoricalAccuracy(yTrue, yPred) {
        return categoricalAccuracy$1(yTrue, yPred);
    }
    /**
     * Categorical crossentropy between an output tensor and a target tensor.
     *
     * @param target A tensor of the same shape as `output`.
     * @param output A tensor resulting from a softmax (unless `fromLogits` is
     *  `true`, in which case `output` is expected to be the logits).
     * @param fromLogits Boolean, whether `output` is the result of a softmax, or is
     *   a tensor of logits.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function categoricalCrossentropy(yTrue, yPred) {
        return categoricalCrossentropy$1(yTrue, yPred);
    }
    /**
     * Computes the precision of the predictions with respect to the labels.
     *
     * Example:
     * ```js
     * const x = tf.tensor2d(
     *    [
     *      [0, 0, 0, 1],
     *      [0, 1, 0, 0],
     *      [0, 0, 0, 1],
     *      [1, 0, 0, 0],
     *      [0, 0, 1, 0]
     *    ]
     * );
     *
     * const y = tf.tensor2d(
     *    [
     *      [0, 0, 1, 0],
     *      [0, 1, 0, 0],
     *      [0, 0, 0, 1],
     *      [0, 1, 0, 0],
     *      [0, 1, 0, 0]
     *    ]
     * );
     *
     * const precision = tf.metrics.precision(x, y);
     * precision.print();
     * ```
     *
     * @param yTrue The ground truth values. Expected to be contain only 0-1 values.
     * @param yPred The predicted values. Expected to be contain only 0-1 values.
     * @return Precision Tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function precision(yTrue, yPred) {
        return precision$1(yTrue, yPred);
    }
    /**
     * Computes the recall of the predictions with respect to the labels.
     *
     * Example:
     * ```js
     * const x = tf.tensor2d(
     *    [
     *      [0, 0, 0, 1],
     *      [0, 1, 0, 0],
     *      [0, 0, 0, 1],
     *      [1, 0, 0, 0],
     *      [0, 0, 1, 0]
     *    ]
     * );
     *
     * const y = tf.tensor2d(
     *    [
     *      [0, 0, 1, 0],
     *      [0, 1, 0, 0],
     *      [0, 0, 0, 1],
     *      [0, 1, 0, 0],
     *      [0, 1, 0, 0]
     *    ]
     * );
     *
     * const recall = tf.metrics.recall(x, y);
     * recall.print();
     * ```
     *
     * @param yTrue The ground truth values. Expected to be contain only 0-1 values.
     * @param yPred The predicted values. Expected to be contain only 0-1 values.
     * @return Recall Tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function recall(yTrue, yPred) {
        return recall$1(yTrue, yPred);
    }
    /**
     * Loss or metric function: Cosine proximity.
     *
     * Mathematically, cosine proximity is defined as:
     *   `-sum(l2Normalize(yTrue) * l2Normalize(yPred))`,
     * wherein `l2Normalize()` normalizes the L2 norm of the input to 1 and `*`
     * represents element-wise multiplication.
     *
     * ```js
     * const yTrue = tf.tensor2d([[1, 0], [1, 0]]);
     * const yPred = tf.tensor2d([[1 / Math.sqrt(2), 1 / Math.sqrt(2)], [0, 1]]);
     * const proximity = tf.metrics.cosineProximity(yTrue, yPred);
     * proximity.print();
     * ```
     *
     * @param yTrue Truth Tensor.
     * @param yPred Prediction Tensor.
     * @return Cosine proximity Tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function cosineProximity(yTrue, yPred) {
        return cosineProximity$1(yTrue, yPred);
    }
    /**
     * Loss or metric function: Mean absolute error.
     *
     * Mathematically, mean absolute error is defined as:
     *   `mean(abs(yPred - yTrue))`,
     * wherein the `mean` is applied over feature dimensions.
     *
     * ```js
     * const yTrue = tf.tensor2d([[0, 1], [0, 0], [2, 3]]);
     * const yPred = tf.tensor2d([[0, 1], [0, 1], [-2, -3]]);
     * const mse = tf.metrics.meanAbsoluteError(yTrue, yPred);
     * mse.print();
     * ```
     *
     * @param yTrue Truth Tensor.
     * @param yPred Prediction Tensor.
     * @return Mean absolute error Tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function meanAbsoluteError(yTrue, yPred) {
        return meanAbsoluteError$1(yTrue, yPred);
    }
    /**
     * Loss or metric function: Mean absolute percentage error.
     *
     * ```js
     * const yTrue = tf.tensor2d([[0, 1], [10, 20]]);
     * const yPred = tf.tensor2d([[0, 1], [11, 24]]);
     * const mse = tf.metrics.meanAbsolutePercentageError(yTrue, yPred);
     * mse.print();
     * ```
     *
     * Aliases: `tf.metrics.MAPE`, `tf.metrics.mape`.
     *
     * @param yTrue Truth Tensor.
     * @param yPred Prediction Tensor.
     * @return Mean absolute percentage error Tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function meanAbsolutePercentageError(yTrue, yPred) {
        return meanAbsolutePercentageError$1(yTrue, yPred);
    }
    function MAPE(yTrue, yPred) {
        return meanAbsolutePercentageError$1(yTrue, yPred);
    }
    function mape(yTrue, yPred) {
        return meanAbsolutePercentageError$1(yTrue, yPred);
    }
    /**
     * Loss or metric function: Mean squared error.
     *
     * ```js
     * const yTrue = tf.tensor2d([[0, 1], [3, 4]]);
     * const yPred = tf.tensor2d([[0, 1], [-3, -4]]);
     * const mse = tf.metrics.meanSquaredError(yTrue, yPred);
     * mse.print();
     * ```
     *
     * Aliases: `tf.metrics.MSE`, `tf.metrics.mse`.
     *
     * @param yTrue Truth Tensor.
     * @param yPred Prediction Tensor.
     * @return Mean squared error Tensor.
     *
     * @doc {heading: 'Metrics', namespace: 'metrics'}
     */
    function meanSquaredError(yTrue, yPred) {
        return meanSquaredError$1(yTrue, yPred);
    }
    function MSE(yTrue, yPred) {
        return meanSquaredError$1(yTrue, yPred);
    }
    function mse(yTrue, yPred) {
        return meanSquaredError$1(yTrue, yPred);
    }

    var exports_metrics = /*#__PURE__*/Object.freeze({
        __proto__: null,
        binaryAccuracy: binaryAccuracy,
        binaryCrossentropy: binaryCrossentropy,
        sparseCategoricalAccuracy: sparseCategoricalAccuracy,
        categoricalAccuracy: categoricalAccuracy,
        categoricalCrossentropy: categoricalCrossentropy,
        precision: precision,
        recall: recall,
        cosineProximity: cosineProximity,
        meanAbsoluteError: meanAbsoluteError,
        meanAbsolutePercentageError: meanAbsolutePercentageError,
        MAPE: MAPE,
        mape: mape,
        meanSquaredError: meanSquaredError,
        MSE: MSE,
        mse: mse
    });

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */

    var exports_models = /*#__PURE__*/Object.freeze({
        __proto__: null,
        modelFromJSON: modelFromJSON
    });

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    /**
     * Regularizer for L1 and L2 regularization.
     *
     * Adds a term to the loss to penalize large weights:
     * loss += sum(l1 * abs(x)) + sum(l2 * x^2)
     *
     * @doc {heading: 'Regularizers', namespace: 'regularizers'}
     */
    function l1l2(config) {
        return new L1L2(config);
    }
    /**
     * Regularizer for L1 regularization.
     *
     * Adds a term to the loss to penalize large weights:
     * loss += sum(l1 * abs(x))
     * @param args l1 config.
     *
     * @doc {heading: 'Regularizers', namespace: 'regularizers'}
     */
    function l1(config) {
        return l1$1(config);
    }
    /**
     * Regularizer for L2 regularization.
     *
     * Adds a term to the loss to penalize large weights:
     * loss += sum(l2 * x^2)
     * @param args l2 config.
     *
     * @doc {heading: 'Regularizers', namespace: 'regularizers'}
     */
    function l2(config) {
        return l2$1(config);
    }

    var exports_regularizers = /*#__PURE__*/Object.freeze({
        __proto__: null,
        l1l2: l1l2,
        l1: l1,
        l2: l2
    });

    /**
     * @license
     * Copyright 2018 Google LLC
     *
     * Use of this source code is governed by an MIT-style
     * license that can be found in the LICENSE file or at
     * https://opensource.org/licenses/MIT.
     * =============================================================================
     */
    class Callback extends BaseCallback {
        constructor() {
            super(...arguments);
            /** Instance of `keras.models.Model`. Reference of the model being trained. */
            this.model = null;
        }
        setModel(model) {
            if (!(model instanceof LayersModel)) {
                throw new Error('model must be a LayersModel, not some other Container');
            }
            this.model = model;
        }
    }
    function less$2(currVal, prevVal) {
        return currVal < prevVal;
    }
    function greater$2(currVal, prevVal) {
        return currVal > prevVal;
    }
    /**
     * A Callback that stops training when a monitored quantity has stopped
     * improving.
     */
    class EarlyStopping extends Callback {
        constructor(args) {
            super();
            if (args == null) {
                args = {};
            }
            if (args.restoreBestWeights) {
                throw new NotImplementedError('restoreBestWeights = True is not implemented in EarlyStopping yet.');
            }
            this.monitor = args.monitor || 'val_loss';
            this.minDelta = Math.abs(args.minDelta || 0);
            this.patience = args.patience || 0;
            this.verbose = args.verbose || 0;
            this.mode = args.mode || 'auto';
            this.baseline = args.baseline;
            if (['auto', 'min', 'max'].indexOf(this.mode) === -1) {
                console.warn(`EarlyStopping mode '${this.mode}' is invalid. ` +
                    `Falling back to mode 'auto'.`);
                this.mode = 'auto';
            }
            if (this.mode === 'min') {
                this.monitorFunc = less$2;
            }
            else if (this.mode === 'max') {
                this.monitorFunc = greater$2;
            }
            else {
                // For mode === 'auto'.
                if (this.monitor.indexOf('acc') !== -1) {
                    this.monitorFunc = greater$2;
                }
                else {
                    this.monitorFunc = less$2;
                }
            }
            if (this.monitorFunc === less$2) {
                this.minDelta *= -1;
            }
        }
        async onTrainBegin(logs) {
            this.wait = 0;
            this.stoppedEpoch = 0;
            if (this.baseline != null) {
                this.best = this.baseline;
            }
            else {
                this.best = this.monitorFunc === less$2 ? Infinity : -Infinity;
            }
        }
        async onEpochEnd(epoch, logs) {
            await resolveScalarsInLogs(logs);
            const current = this.getMonitorValue(logs);
            if (current == null) {
                return;
            }
            if (this.monitorFunc(current - this.minDelta, this.best)) {
                this.best = current;
                this.wait = 0;
                // TODO(cais): Logic for restoreBestWeights.
            }
            else {
                this.wait++;
                if (this.wait >= this.patience) {
                    this.stoppedEpoch = epoch;
                    this.model.stopTraining = true;
                }
                // TODO(cais): Logic for restoreBestWeights.
            }
        }
        async onTrainEnd(logs) {
            if (this.stoppedEpoch > 0 && this.verbose) {
                console.log(`Epoch ${this.stoppedEpoch}: early stopping.`);
            }
        }
        getMonitorValue(logs) {
            if (logs == null) {
                logs = {};
            }
            const monitorValue = logs[this.monitor];
            if (monitorValue == null) {
                console.warn(`Metric for EarlyStopping ${this.monitor} is not available. ` +
                    `Available metrics are: ${Object.keys(logs)}`);
            }
            return monitorValue;
        }
    }
    /**
     * Factory function for a Callback that stops training when a monitored
     * quantity has stopped improving.
     *
     * Early stopping is a type of regularization, and protects model against
     * overfitting.
     *
     * The following example based on fake data illustrates how this callback
     * can be used during `tf.LayersModel.fit()`:
     *
     * ```js
     * const model = tf.sequential();
     * model.add(tf.layers.dense({
     *   units: 3,
     *   activation: 'softmax',
     *   kernelInitializer: 'ones',
     *   inputShape: [2]
     * }));
     * const xs = tf.tensor2d([1, 2, 3, 4], [2, 2]);
     * const ys = tf.tensor2d([[1, 0, 0], [0, 1, 0]], [2, 3]);
     * const xsVal = tf.tensor2d([4, 3, 2, 1], [2, 2]);
     * const ysVal = tf.tensor2d([[0, 0, 1], [0, 1, 0]], [2, 3]);
     * model.compile(
     *     {loss: 'categoricalCrossentropy', optimizer: 'sgd', metrics: ['acc']});
     *
     * // Without the EarlyStopping callback, the val_acc value would be:
     * //   0.5, 0.5, 0.5, 0.5, ...
     * // With val_acc being monitored, training should stop after the 2nd epoch.
     * const history = await model.fit(xs, ys, {
     *   epochs: 10,
     *   validationData: [xsVal, ysVal],
     *   callbacks: tf.callbacks.earlyStopping({monitor: 'val_acc'})
     * });
     *
     * // Expect to see a length-2 array.
     * console.log(history.history.val_acc);
     * ```
     *
     * @doc {
     *   heading: 'Callbacks',
     *   namespace: 'callbacks'
     * }
     */
    function earlyStopping(args) {
        return new EarlyStopping(args);
    }
    const callbacks = { earlyStopping };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ENV$1 = env$1();
    /** Whether to keep intermediate tensors. */
    ENV$1.registerFlag('KEEP_INTERMEDIATE_TENSORS', () => false, debugValue => {
        if (debugValue) {
            console.warn('Keep intermediate tensors is ON. This will print the values of all ' +
                'intermediate tensors during model inference. Not all models ' +
                'support this mode. For details, check e2e/benchmarks/ ' +
                'model_config.js. This significantly impacts performance.');
        }
    });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /** DataType enum. */
    var DataType;
    (function (DataType) {
        // Not a legal value for DataType.  Used to indicate a DataType field
        // has not been set.
        DataType[DataType["DT_INVALID"] = 0] = "DT_INVALID";
        // Data types that all computation devices are expected to be
        // capable to support.
        DataType[DataType["DT_FLOAT"] = 1] = "DT_FLOAT";
        DataType[DataType["DT_DOUBLE"] = 2] = "DT_DOUBLE";
        DataType[DataType["DT_INT32"] = 3] = "DT_INT32";
        DataType[DataType["DT_UINT8"] = 4] = "DT_UINT8";
        DataType[DataType["DT_INT16"] = 5] = "DT_INT16";
        DataType[DataType["DT_INT8"] = 6] = "DT_INT8";
        DataType[DataType["DT_STRING"] = 7] = "DT_STRING";
        DataType[DataType["DT_COMPLEX64"] = 8] = "DT_COMPLEX64";
        DataType[DataType["DT_INT64"] = 9] = "DT_INT64";
        DataType[DataType["DT_BOOL"] = 10] = "DT_BOOL";
        DataType[DataType["DT_QINT8"] = 11] = "DT_QINT8";
        DataType[DataType["DT_QUINT8"] = 12] = "DT_QUINT8";
        DataType[DataType["DT_QINT32"] = 13] = "DT_QINT32";
        DataType[DataType["DT_BFLOAT16"] = 14] = "DT_BFLOAT16";
        DataType[DataType["DT_QINT16"] = 15] = "DT_QINT16";
        DataType[DataType["DT_QUINT16"] = 16] = "DT_QUINT16";
        DataType[DataType["DT_UINT16"] = 17] = "DT_UINT16";
        DataType[DataType["DT_COMPLEX128"] = 18] = "DT_COMPLEX128";
        DataType[DataType["DT_HALF"] = 19] = "DT_HALF";
        DataType[DataType["DT_RESOURCE"] = 20] = "DT_RESOURCE";
        DataType[DataType["DT_VARIANT"] = 21] = "DT_VARIANT";
        DataType[DataType["DT_UINT32"] = 22] = "DT_UINT32";
        DataType[DataType["DT_UINT64"] = 23] = "DT_UINT64";
        // Do not use!  These are only for parameters.  Every enum above
        // should have a corresponding value below (verified by types_test).
        DataType[DataType["DT_FLOAT_REF"] = 101] = "DT_FLOAT_REF";
        DataType[DataType["DT_DOUBLE_REF"] = 102] = "DT_DOUBLE_REF";
        DataType[DataType["DT_INT32_REF"] = 103] = "DT_INT32_REF";
        DataType[DataType["DT_UINT8_REF"] = 104] = "DT_UINT8_REF";
        DataType[DataType["DT_INT16_REF"] = 105] = "DT_INT16_REF";
        DataType[DataType["DT_INT8_REF"] = 106] = "DT_INT8_REF";
        DataType[DataType["DT_STRING_REF"] = 107] = "DT_STRING_REF";
        DataType[DataType["DT_COMPLEX64_REF"] = 108] = "DT_COMPLEX64_REF";
        DataType[DataType["DT_INT64_REF"] = 109] = "DT_INT64_REF";
        DataType[DataType["DT_BOOL_REF"] = 110] = "DT_BOOL_REF";
        DataType[DataType["DT_QINT8_REF"] = 111] = "DT_QINT8_REF";
        DataType[DataType["DT_QUINT8_REF"] = 112] = "DT_QUINT8_REF";
        DataType[DataType["DT_QINT32_REF"] = 113] = "DT_QINT32_REF";
        DataType[DataType["DT_BFLOAT16_REF"] = 114] = "DT_BFLOAT16_REF";
        DataType[DataType["DT_QINT16_REF"] = 115] = "DT_QINT16_REF";
        DataType[DataType["DT_QUINT16_REF"] = 116] = "DT_QUINT16_REF";
        DataType[DataType["DT_UINT16_REF"] = 117] = "DT_UINT16_REF";
        DataType[DataType["DT_COMPLEX128_REF"] = 118] = "DT_COMPLEX128_REF";
        DataType[DataType["DT_HALF_REF"] = 119] = "DT_HALF_REF";
        DataType[DataType["DT_RESOURCE_REF"] = 120] = "DT_RESOURCE_REF";
        DataType[DataType["DT_VARIANT_REF"] = 121] = "DT_VARIANT_REF";
        DataType[DataType["DT_UINT32_REF"] = 122] = "DT_UINT32_REF";
        DataType[DataType["DT_UINT64_REF"] = 123] = "DT_UINT64_REF";
    })(DataType || (DataType = {}));
    var SaverDef;
    (function (SaverDef) {
        (function (CheckpointFormatVersion) {
            CheckpointFormatVersion[CheckpointFormatVersion["LEGACY"] = 0] = "LEGACY";
            CheckpointFormatVersion[CheckpointFormatVersion["V1"] = 1] = "V1";
            CheckpointFormatVersion[CheckpointFormatVersion["V2"] = 2] = "V2";
        })(SaverDef.CheckpointFormatVersion || (SaverDef.CheckpointFormatVersion = {}));
    })(SaverDef || (SaverDef = {}));

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const CUSTOM_OPS = {};
    /**
     * Register an Op for graph model executor. This allow you to register
     * TensorFlow custom op or override existing op.
     *
     * Here is an example of registering a new MatMul Op.
     * ```js
     * const customMatmul = (node) =>
     *    tf.matMul(
     *        node.inputs[0], node.inputs[1],
     *        node.attrs['transpose_a'], node.attrs['transpose_b']);
     *
     * tf.registerOp('MatMul', customMatmul);
     * ```
     * The inputs and attrs of the node object is based on the TensorFlow op
     * registry.
     *
     * @param name The Tensorflow Op name.
     * @param opFunc An op function which is called with the current graph node
     * during execution and needs to return a tensor or a list of tensors. The node
     * has the following attributes:
     *    - attr: A map from attribute name to its value
     *    - inputs: A list of input tensors
     *
     * @doc {heading: 'Models', subheading: 'Op Registry'}
     */
    function registerOp(name, opFunc) {
        const opMapper = {
            tfOpName: name,
            category: 'custom',
            inputs: [],
            attrs: [],
            customExecutor: opFunc
        };
        CUSTOM_OPS[name] = opMapper;
    }
    /**
     * Retrieve the OpMapper object for the registered op.
     *
     * @param name The Tensorflow Op name.
     *
     * @doc {heading: 'Models', subheading: 'Op Registry'}
     */
    function getRegisteredOp(name) {
        return CUSTOM_OPS[name];
    }
    /**
     * Deregister the Op for graph model executor.
     *
     * @param name The Tensorflow Op name.
     *
     * @doc {heading: 'Models', subheading: 'Op Registry'}
     */
    function deregisterOp(name) {
        delete CUSTOM_OPS[name];
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function getParamValue(paramName, node, tensorMap, context, resourceManager) {
        const inputParam = node.inputParams[paramName];
        if (inputParam && inputParam.inputIndexStart !== undefined) {
            const start = inputParam.inputIndexStart;
            const end = inputParam.inputIndexEnd === 0 ?
                undefined :
                (inputParam.inputIndexEnd === undefined ? start + 1 :
                    inputParam.inputIndexEnd);
            if (inputParam.type === 'tensor') {
                return getTensor(node.inputNames[inputParam.inputIndexStart], tensorMap, context, resourceManager);
            }
            if (inputParam.type === 'tensors') {
                const inputs = node.inputNames.slice(start, end);
                return inputs.map(name => getTensor(name, tensorMap, context, resourceManager));
            }
            const tensor = getTensor(node.inputNames.slice(start)[0], tensorMap, context, resourceManager);
            const data = tensor.dataSync();
            return inputParam.type === 'number' ?
                data[0] :
                toNestedArray(tensor.shape, data);
        }
        const attrParam = node.attrParams[paramName];
        return attrParam && attrParam.value;
    }
    /**
     * Retrieve the tensor from tensorsMap based on input name.
     * @param name Node input name
     * @param tensorsMap Tensors map keyed by the node
     * @param context contains tensors and information for running the current node.
     * @param resourceManager Optional. Contains global resources of the model.
     */
    function getTensor(name, tensorsMap, context, resourceManager) {
        const [nodeName, index] = parseNodeName(name);
        if (resourceManager != null) {
            const tensor = resourceManager.getHashTableHandleByName(nodeName);
            if (tensor != null) {
                return tensor;
            }
        }
        const contextId = context.currentContextIds.find(contextId => {
            return !!tensorsMap[getNodeNameWithContextId(nodeName, contextId)];
        });
        return contextId !== undefined ?
            tensorsMap[getNodeNameWithContextId(nodeName, contextId)][index] :
            undefined;
    }
    /**
     * Retrieve the tensors based on input name for current context.
     * @param name Node input name
     * @param tensorsMap Tensors map keyed by the node
     */
    function getTensorsForCurrentContenxt(name, tensorsMap, context) {
        return tensorsMap[getNodeNameWithContextId(name, context.currentContextId)];
    }
    /**
     * Returns the node name, outputName and index from the Node input name.
     * @param inputName The input name of the node, in format of
     * node_name:output_index, i.e. MatMul:0, if the output_index is not set, it is
     * default to 0.
     * If the input name contains output name i.e. StringSplit:indices:0, it will
     * return ['StringSplit', 0, 'indices'].
     */
    function getNodeNameAndIndex(inputName, context) {
        const [nodeName, index, outputName] = parseNodeName(inputName);
        return [
            getNodeNameWithContextId(nodeName, context && context.currentContextId),
            index, outputName
        ];
    }
    function getNodeNameWithContextId(name, contextId) {
        return !!contextId ? `${name}-${contextId}` : name;
    }
    function parseNodeName(name) {
        const parts = name.split(':');
        if (parts.length === 1) {
            return [name, 0, undefined];
        }
        const nodeName = parts[0];
        const outputName = parts.length === 3 ? parts[1] : undefined;
        const index = Number(parts[parts.length - 1]);
        return [nodeName, index, outputName];
    }
    function getPadding(node, tensorMap, context) {
        let pad = getParamValue('pad', node, tensorMap, context);
        if (pad === 'explicit') {
            // This is 1d array, we need to convert it to 2d array
            pad = getParamValue('explicitPaddings', node, tensorMap, context);
            const explicitPadding = [[0, 0], [0, 0], [0, 0], [0, 0]];
            for (let i = 0; i < 4; i++) {
                explicitPadding[i][0] = pad[i * 2];
                explicitPadding[i][1] = pad[i * 2 + 1];
            }
            return explicitPadding;
        }
        return pad;
    }
    /**
     *  Reuse the tensor if it is marked as keep, otherwise clone the tensor to
     *  avoid disposal. This is important for TensorArray and TensorList ops, since
     *  internally they use a tensor as the id for TensorArray and TensorList, and
     * to simplify lookup, they also use Tensor.id as the key to the internal map.
     * These id tensors have been marked as kept in the backend, we need avoid clone
     * them in order to create new Tensor.id.
     * @param tensor
     */
    function cloneTensor(tensor) {
        return tensor.kept ? tensor : clone(tensor);
    }

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$i = [
        {
            'tfOpName': 'Add',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'AddV2',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'AddN',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'end': 0,
                    'name': 'tensors',
                    'type': 'tensors'
                }
            ]
        },
        {
            'tfOpName': 'BiasAdd',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Sub',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'RealDiv',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Div',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'DivNoNan',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'FloorDiv',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Mul',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Maximum',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Minimum',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Pow',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'SquaredDifference',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Mod',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'FloorMod',
            'category': 'arithmetic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        }
    ];

    var arithmetic = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$i
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$h = [
        {
            'tfOpName': 'Abs',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Acos',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Asin',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Atan',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Atan2',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'y',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Ceil',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'ClipByValue',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'clipValueMin',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'clipValueMax',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Complex',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'real',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'imag',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'ComplexAbs',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Cos',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Cosh',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Elu',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Exp',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Floor',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Log',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Imag',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'Tout',
                    'name': 'outputType',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Neg',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Real',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'Tout',
                    'name': 'outputType',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Prelu',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'alpha',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Relu',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Relu6',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Selu',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Sigmoid',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Sin',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Sinh',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Sqrt',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Rsqrt',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Square',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Tan',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Tanh',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Sign',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Round',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Expm1',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Log1p',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Reciprocal',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Softplus',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Asinh',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Acosh',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Atanh',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Erf',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Prod',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axes',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'keep_dims',
                    'name': 'keepDims',
                    'type': 'bool',
                    'notSupported': true
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LeakyRelu',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'alpha',
                    'name': 'alpha',
                    'type': 'number',
                    'defaultValue': 0.2
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'IsNan',
            'category': 'basic_math',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        }
    ];

    var basicMath = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$h
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$g = [
        {
            'tfOpName': 'EmptyTensorList',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'elementShape',
                    'type': 'shape'
                },
                {
                    'start': 1,
                    'name': 'maxNumElements',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'LoopCond',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'pred',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'Switch',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'data',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'pred',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'Merge',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'end': 0,
                    'name': 'tensors',
                    'type': 'tensors'
                }
            ]
        },
        {
            'tfOpName': 'Enter',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensor',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'frame_name',
                    'name': 'frameName',
                    'type': 'string'
                },
                {
                    'tfName': 'is_constant',
                    'name': 'isConstant',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'Exit',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensor',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'NextIteration',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensor',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'TensorArrayV3',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'size',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'dtype',
                    'name': 'dtype',
                    'type': 'dtype'
                },
                {
                    'tfName': 'element_shape',
                    'name': 'elementShape',
                    'type': 'shape'
                },
                {
                    'tfName': 'dynamic_size',
                    'name': 'dynamicSize',
                    'type': 'bool'
                },
                {
                    'tfName': 'clear_after_read',
                    'name': 'clearAfterRead',
                    'type': 'bool'
                },
                {
                    'tfName': 'identical_element_shapes',
                    'name': 'identicalElementShapes',
                    'type': 'bool'
                },
                {
                    'tfName': 'tensor_array_name',
                    'name': 'name',
                    'type': 'string'
                }
            ]
        },
        {
            'tfOpName': 'TensorArrayWriteV3',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorArrayId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'index',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'tensor',
                    'type': 'tensor'
                },
                {
                    'start': 3,
                    'name': 'flowIn',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'TensorArrayReadV3',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorArrayId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'index',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'flowIn',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'dtype',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'TensorArrayGatherV3',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorArrayId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'flowIn',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'dtype',
                    'name': 'dtype',
                    'type': 'dtype'
                },
                {
                    'tfName': 'element_shape',
                    'name': 'elementShape',
                    'type': 'shape'
                }
            ]
        },
        {
            'tfOpName': 'TensorArrayScatterV3',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorArrayId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'tensor',
                    'type': 'tensor'
                },
                {
                    'start': 3,
                    'name': 'flowIn',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorArrayConcatV3',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorArrayId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'flowIn',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'dtype',
                    'name': 'dtype',
                    'type': 'dtype'
                },
                {
                    'tfName': 'element_shape_except0',
                    'name': 'elementShapeExcept0',
                    'type': 'shape',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'TensorArraySplitV3',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorArrayId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'tensor',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'lengths',
                    'type': 'number[]'
                },
                {
                    'start': 3,
                    'name': 'flowIn',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorArraySizeV3',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorArrayId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'flowIn',
                    'type': 'number'
                }
            ]
        },
        {
            'tfOpName': 'TensorArrayCloseV3',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorArrayId',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'StatelessIf',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'cond',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'end': 0,
                    'name': 'args',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'then_branch',
                    'name': 'thenBranch',
                    'type': 'func'
                },
                {
                    'tfName': 'else_branch',
                    'name': 'elseBranch',
                    'type': 'func'
                }
            ]
        },
        {
            'tfOpName': 'If',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'cond',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'end': 0,
                    'name': 'args',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'then_branch',
                    'name': 'thenBranch',
                    'type': 'func'
                },
                {
                    'tfName': 'else_branch',
                    'name': 'elseBranch',
                    'type': 'func'
                }
            ]
        },
        {
            'tfOpName': 'StatelessWhile',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'end': 0,
                    'name': 'args',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'cond',
                    'name': 'cond',
                    'type': 'func'
                },
                {
                    'tfName': 'body',
                    'name': 'body',
                    'type': 'func'
                }
            ]
        },
        {
            'tfOpName': 'While',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'end': 0,
                    'name': 'args',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'cond',
                    'name': 'cond',
                    'type': 'func'
                },
                {
                    'tfName': 'body',
                    'name': 'body',
                    'type': 'func'
                }
            ]
        },
        {
            'tfOpName': 'TensorListScatter',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensor',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'elementShape',
                    'type': 'shape'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListScatterV2',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensor',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'elementShape',
                    'type': 'shape'
                },
                {
                    'start': 3,
                    'name': 'numElements',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListGather',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'elementShape',
                    'type': 'shape'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListGetItem',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'index',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'elementShape',
                    'type': 'shape'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListSetItem',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'index',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'tensor',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListReserve',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'elementShape',
                    'type': 'shape'
                },
                {
                    'start': 1,
                    'name': 'numElements',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListFromTensor',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensor',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'elementShape',
                    'type': 'shape'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListStack',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'elementShape',
                    'type': 'shape'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                },
                {
                    'tfName': 'num_elements',
                    'name': 'numElements',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListSplit',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensor',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'elementShape',
                    'type': 'shape'
                },
                {
                    'start': 2,
                    'name': 'lengths',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListConcat',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_shape',
                    'name': 'elementShape',
                    'type': 'shape'
                },
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListConcatV2',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_shape',
                    'name': 'elementShape',
                    'type': 'shape'
                },
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListPopBack',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'elementShape',
                    'type': 'shape'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListPushBack',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'tensor',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'element_dtype',
                    'name': 'elementDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TensorListLength',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'TensorListResize',
            'category': 'control',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensorListId',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'size',
                    'type': 'number'
                }
            ]
        }
    ];

    var control = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$g
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$f = [
        {
            'tfOpName': 'AvgPool',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'notSupported': true
                },
                {
                    'tfName': 'ksize',
                    'name': 'kernelSize',
                    'type': 'number[]'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'MaxPool',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'notSupported': true
                },
                {
                    'tfName': 'ksize',
                    'name': 'kernelSize',
                    'type': 'number[]'
                },
                {
                    'tfName': 'explicit_paddings',
                    'name': 'explicitPaddings',
                    'type': 'number[]',
                    'defaultValue': [],
                    'notSupported': true
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'MaxPoolWithArgmax',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'ksize',
                    'name': 'kernelSize',
                    'type': 'number[]'
                },
                {
                    'tfName': 'include_batch_in_index',
                    'name': 'includeBatchInIndex',
                    'type': 'bool'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'AvgPool3D',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'notSupported': true
                },
                {
                    'tfName': 'ksize',
                    'name': 'kernelSize',
                    'type': 'number[]'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'MaxPool3D',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'notSupported': true
                },
                {
                    'tfName': 'ksize',
                    'name': 'kernelSize',
                    'type': 'number[]'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Conv1D',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'filter',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'stride',
                    'name': 'stride',
                    'type': 'number'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'defaultValue': 'NWC'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'dilation',
                    'name': 'dilation',
                    'type': 'number',
                    'defaultValue': 1
                }
            ]
        },
        {
            'tfOpName': 'Conv2D',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'filter',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'useCudnnOnGpu',
                    'name': 'useCudnnOnGpu',
                    'type': 'bool'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'defaultValue': 'NHWC'
                },
                {
                    'tfName': 'explicit_paddings',
                    'name': 'explicitPaddings',
                    'type': 'number[]',
                    'defaultValue': []
                },
                {
                    'tfName': 'dilations',
                    'name': 'dilations',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': '_FusedConv2D',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'filter',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'end': 0,
                    'name': 'args',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'num_args',
                    'name': 'numArgs',
                    'type': 'number'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'explicit_paddings',
                    'name': 'explicitPaddings',
                    'type': 'number[]',
                    'defaultValue': []
                },
                {
                    'tfName': 'use_cudnn_on_gpu',
                    'name': 'useCudnnOnGpu',
                    'type': 'bool',
                    'defaultValue': true
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'defaultValue': 'NHWC'
                },
                {
                    'tfName': 'dilations',
                    'name': 'dilations',
                    'type': 'number[]',
                    'defaultValue': [
                        1,
                        1,
                        1,
                        1
                    ]
                },
                {
                    'tfName': 'fused_ops',
                    'name': 'fusedOps',
                    'type': 'string[]',
                    'defaultValue': []
                },
                {
                    'tfName': 'epsilon',
                    'name': 'epsilon',
                    'type': 'number',
                    'defaultValue': 0.0001
                },
                {
                    'tfName': 'leakyrelu_alpha',
                    'name': 'leakyreluAlpha',
                    'type': 'number'
                }
            ]
        },
        {
            'tfOpName': 'Conv2DBackpropInput',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 2,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'filter',
                    'type': 'tensor'
                },
                {
                    'start': 0,
                    'name': 'outputShape',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'notSupported': true
                },
                {
                    'tfName': 'explicit_paddings',
                    'name': 'explicitPaddings',
                    'type': 'number[]',
                    'defaultValue': []
                },
                {
                    'tfName': 'dilations',
                    'name': 'dilations',
                    'type': 'number[]',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'DepthwiseConv2d',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'input',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'filter',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'defaultValue': 'NHWC'
                },
                {
                    'tfName': 'explicit_paddings',
                    'name': 'explicitPaddings',
                    'type': 'number[]',
                    'defaultValue': []
                },
                {
                    'tfName': 'dilations',
                    'name': 'dilations',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'DepthwiseConv2dNative',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'input',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'filter',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'defaultValue': 'NHWC'
                },
                {
                    'tfName': 'explicit_paddings',
                    'name': 'explicitPaddings',
                    'type': 'number[]',
                    'defaultValue': []
                },
                {
                    'tfName': 'dilations',
                    'name': 'dilations',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'FusedDepthwiseConv2dNative',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'filter',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'end': 0,
                    'name': 'args',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'num_args',
                    'name': 'numArgs',
                    'type': 'number'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'defaultValue': 'NHWC'
                },
                {
                    'tfName': 'dilations',
                    'name': 'dilations',
                    'type': 'number[]',
                    'defaultValue': [
                        1,
                        1,
                        1,
                        1
                    ]
                },
                {
                    'tfName': 'fused_ops',
                    'name': 'fusedOps',
                    'type': 'string[]',
                    'defaultValue': []
                },
                {
                    'tfName': 'explicit_paddings',
                    'name': 'explicitPaddings',
                    'type': 'number[]',
                    'defaultValue': []
                }
            ]
        },
        {
            'tfOpName': 'Conv3D',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'filter',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'defaultValue': 'NHWC'
                },
                {
                    'tfName': 'dilations',
                    'name': 'dilations',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'Dilation2D',
            'category': 'convolution',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'filter',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'strides',
                    'name': 'strides',
                    'type': 'number[]'
                },
                {
                    'tfName': 'rates',
                    'name': 'dilations',
                    'type': 'number[]'
                },
                {
                    'tfName': 'padding',
                    'name': 'pad',
                    'type': 'string'
                }
            ]
        }
    ];

    var convolution = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$f
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$e = [
        {
            'tfOpName': 'Fill',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'shape',
                    'type': 'number[]'
                },
                {
                    'start': 1,
                    'name': 'value',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'LinSpace',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'start',
                    'type': 'number'
                },
                {
                    'start': 1,
                    'name': 'stop',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'num',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'OneHot',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'indices',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'depth',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'onValue',
                    'type': 'number',
                    'defaultValue': 1
                },
                {
                    'start': 3,
                    'name': 'offValue',
                    'type': 'number',
                    'defaultValue': 0
                }
            ],
            'attrs': [
                {
                    'tfName': 'axis',
                    'name': 'axis',
                    'type': 'number',
                    'notSupported': true
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Ones',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'shape',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'OnesLike',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'dtype',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'RandomUniform',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'shape',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'minval',
                    'name': 'minval',
                    'type': 'number',
                    'defaultValue': 0
                },
                {
                    'tfName': 'maxval',
                    'name': 'maxval',
                    'type': 'number',
                    'defaultValue': 1
                },
                {
                    'tfName': 'dtype',
                    'name': 'dtype',
                    'type': 'dtype'
                },
                {
                    'tfName': 'seed',
                    'name': 'seed',
                    'type': 'number',
                    'defaultValue': 0
                },
                {
                    'tfName': 'seed2',
                    'name': 'seed2',
                    'type': 'number',
                    'defaultValue': 0,
                    'notSupported': true
                },
                {
                    'tfName': 'T',
                    'name': 'T',
                    'type': 'number',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Range',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'start',
                    'type': 'number'
                },
                {
                    'start': 1,
                    'name': 'stop',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'step',
                    'type': 'number',
                    'defaultValue': 0
                }
            ],
            'attrs': [
                {
                    'tfName': 'Tidx',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'TruncatedNormal',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'shape',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'means',
                    'name': 'mean',
                    'type': 'number',
                    'defaultValue': 0
                },
                {
                    'tfName': 'stddev',
                    'name': 'stdDev',
                    'type': 'number',
                    'defaultValue': 1
                },
                {
                    'tfName': 'seed',
                    'name': 'seed',
                    'type': 'number'
                },
                {
                    'tfName': 'seed2',
                    'name': 'seed2',
                    'type': 'number',
                    'defaultValue': 0,
                    'notSupported': true
                },
                {
                    'tfName': 'dtype',
                    'name': 'dtype',
                    'type': 'dtype'
                },
                {
                    'tfName': 'T',
                    'name': 'T',
                    'type': 'number',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Zeros',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'shape',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'ZerosLike',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'Multinomial',
            'category': 'creation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'logits',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'numSamples',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'seed',
                    'name': 'seed',
                    'type': 'number'
                },
                {
                    'tfName': 'seed2',
                    'name': 'seed2',
                    'type': 'number'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype'
                },
                {
                    'tfName': 'output_dtype',
                    'name': 'output_dtype',
                    'type': 'dtype'
                }
            ]
        }
    ];

    var creation = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$e
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$d = [
        {
            'tfOpName': 'NonMaxSuppressionV2',
            'category': 'dynamic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'boxes',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'scores',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'maxOutputSize',
                    'type': 'number'
                },
                {
                    'start': 3,
                    'name': 'iouThreshold',
                    'type': 'number'
                }
            ]
        },
        {
            'tfOpName': 'NonMaxSuppressionV3',
            'category': 'dynamic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'boxes',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'scores',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'maxOutputSize',
                    'type': 'number'
                },
                {
                    'start': 3,
                    'name': 'iouThreshold',
                    'type': 'number'
                },
                {
                    'start': 4,
                    'name': 'scoreThreshold',
                    'type': 'number'
                }
            ]
        },
        {
            'tfOpName': 'NonMaxSuppressionV4',
            'category': 'dynamic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'boxes',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'scores',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'maxOutputSize',
                    'type': 'number'
                },
                {
                    'start': 3,
                    'name': 'iouThreshold',
                    'type': 'number'
                },
                {
                    'start': 4,
                    'name': 'scoreThreshold',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'T_threshold',
                    'name': 'threshold',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'pad_to_max_output_size',
                    'name': 'padToMaxOutputSize',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'NonMaxSuppressionV5',
            'category': 'dynamic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'boxes',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'scores',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'maxOutputSize',
                    'type': 'number'
                },
                {
                    'start': 3,
                    'name': 'iouThreshold',
                    'type': 'number'
                },
                {
                    'start': 4,
                    'name': 'scoreThreshold',
                    'type': 'number'
                },
                {
                    'start': 5,
                    'name': 'softNmsSigma',
                    'type': 'number'
                }
            ]
        },
        {
            'tfOpName': 'Where',
            'category': 'dynamic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'condition',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'ListDiff',
            'category': 'dynamic',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'y',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        }
    ];

    var dynamic = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$d
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$c = [
        {
            'tfOpName': 'LowerBound',
            'category': 'evaluation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'sortedSequence',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'values',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'TopKV2',
            'category': 'evaluation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'k',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'sorted',
                    'name': 'sorted',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'UpperBound',
            'category': 'evaluation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'sortedSequence',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'values',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'Unique',
            'category': 'evaluation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'UniqueV2',
            'category': 'evaluation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number'
                }
            ]
        }
    ];

    var evaluation = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$c
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$b = [
        {
            'tfOpName': 'PlaceholderWithDefault',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'name': 'default',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'shape',
                    'name': 'shape',
                    'type': 'shape'
                },
                {
                    'tfName': 'dtype',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'Placeholder',
            'category': 'graph',
            'attrs': [
                {
                    'tfName': 'shape',
                    'name': 'shape',
                    'type': 'shape'
                },
                {
                    'tfName': 'dtype',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'Const',
            'category': 'graph'
        },
        {
            'tfOpName': 'Identity',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'IdentityN',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'end': 0,
                    'name': 'x',
                    'type': 'tensors'
                }
            ]
        },
        {
            'tfOpName': 'Snapshot',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'Rank',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'Size',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'Shape',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'ShapeN',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'end': 0,
                    'name': 'x',
                    'type': 'tensors'
                }
            ]
        },
        {
            'tfOpName': 'Print',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'data',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'message',
                    'name': 'message',
                    'type': 'string'
                },
                {
                    'tfName': 'first_n',
                    'name': 'firstN',
                    'type': 'number',
                    'notSupported': true
                },
                {
                    'tfName': 'summarize',
                    'name': 'summarize',
                    'type': 'number',
                    'defaultValue': 3
                }
            ]
        },
        {
            'tfOpName': 'NoOp',
            'category': 'graph',
            'inputs': []
        },
        {
            'tfOpName': 'StopGradient',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'FakeQuantWithMinMaxVars',
            'category': 'graph',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'min',
                    'name': 'min',
                    'type': 'number'
                },
                {
                    'tfName': 'max',
                    'name': 'max',
                    'type': 'number'
                }
            ]
        }
    ];

    var graph = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$b
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$a = [
        {
            'tfOpName': 'HashTable',
            'category': 'hash_table',
            'inputs': [],
            'attrs': [
                {
                    'tfName': 'shared_name',
                    'name': 'sharedName',
                    'type': 'string'
                },
                {
                    'tfName': 'use_node_name_sharing',
                    'name': 'useNodeNameSharing',
                    'type': 'bool'
                },
                {
                    'tfName': 'key_dtype',
                    'name': 'keyDType',
                    'type': 'dtype'
                },
                {
                    'tfName': 'value_dtype',
                    'name': 'valueDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'HashTableV2',
            'category': 'hash_table',
            'inputs': [],
            'attrs': [
                {
                    'tfName': 'shared_name',
                    'name': 'sharedName',
                    'type': 'string'
                },
                {
                    'tfName': 'use_node_name_sharing',
                    'name': 'useNodeNameSharing',
                    'type': 'bool'
                },
                {
                    'tfName': 'key_dtype',
                    'name': 'keyDType',
                    'type': 'dtype'
                },
                {
                    'tfName': 'value_dtype',
                    'name': 'valueDType',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'LookupTableImport',
            'category': 'hash_table',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tableHandle',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'keys',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'values',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'Tin',
                    'name': 'tIn',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'Tout',
                    'name': 'tOut',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LookupTableImportV2',
            'category': 'hash_table',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tableHandle',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'keys',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'values',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'Tin',
                    'name': 'tIn',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'Tout',
                    'name': 'tOut',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LookupTableFind',
            'category': 'hash_table',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tableHandle',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'keys',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'defaultValue',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'Tin',
                    'name': 'tIn',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'Tout',
                    'name': 'tOut',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LookupTableFindV2',
            'category': 'hash_table',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tableHandle',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'keys',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'defaultValue',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'Tin',
                    'name': 'tIn',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'Tout',
                    'name': 'tOut',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LookupTableSize',
            'category': 'hash_table',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tableHandle',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'LookupTableSizeV2',
            'category': 'hash_table',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tableHandle',
                    'type': 'tensor'
                }
            ]
        }
    ];

    var hashTable = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$a
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$9 = [
        {
            'tfOpName': 'ResizeBilinear',
            'category': 'image',
            'inputs': [
                {
                    'start': 0,
                    'name': 'images',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'size',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'align_corners',
                    'name': 'alignCorners',
                    'type': 'bool'
                },
                {
                    'tfName': 'half_pixel_centers',
                    'name': 'halfPixelCenters',
                    'type': 'bool'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'ResizeNearestNeighbor',
            'category': 'image',
            'inputs': [
                {
                    'start': 0,
                    'name': 'images',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'size',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'align_corners',
                    'name': 'alignCorners',
                    'type': 'bool'
                },
                {
                    'tfName': 'half_pixel_centers',
                    'name': 'halfPixelCenters',
                    'type': 'bool'
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'CropAndResize',
            'category': 'image',
            'inputs': [
                {
                    'start': 0,
                    'name': 'image',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'boxes',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'boxInd',
                    'type': 'tensor'
                },
                {
                    'start': 3,
                    'name': 'cropSize',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'method',
                    'name': 'method',
                    'type': 'string'
                },
                {
                    'tfName': 'extrapolation_value',
                    'name': 'extrapolationValue',
                    'type': 'number'
                }
            ]
        },
        {
            'tfOpName': 'ImageProjectiveTransformV3',
            'category': 'image',
            'inputs': [
                {
                    'start': 0,
                    'name': 'images',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'transforms',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'outputShape',
                    'type': 'number[]'
                },
                {
                    'start': 3,
                    'name': 'fillValue',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'interpolation',
                    'name': 'interpolation',
                    'type': 'string'
                },
                {
                    'tfName': 'fill_mode',
                    'name': 'fillMode',
                    'type': 'string'
                }
            ]
        }
    ];

    var image = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$9
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$8 = [
        {
            'tfOpName': 'Equal',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'NotEqual',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Greater',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'GreaterEqual',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Less',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LessEqual',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LogicalAnd',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LogicalNot',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LogicalOr',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Select',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'condition',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'SelectV2',
            'category': 'logical',
            'inputs': [
                {
                    'start': 0,
                    'name': 'condition',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        }
    ];

    var logical = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$8
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$7 = [
        {
            'tfOpName': '_FusedMatMul',
            'category': 'matrices',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'end': 0,
                    'name': 'args',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'num_args',
                    'name': 'numArgs',
                    'type': 'number'
                },
                {
                    'tfName': 'fused_ops',
                    'name': 'fusedOps',
                    'type': 'string[]',
                    'defaultValue': []
                },
                {
                    'tfName': 'epsilon',
                    'name': 'epsilon',
                    'type': 'number',
                    'defaultValue': 0.0001
                },
                {
                    'tfName': 'transpose_a',
                    'name': 'transposeA',
                    'type': 'bool',
                    'defaultValue': false
                },
                {
                    'tfName': 'transpose_b',
                    'name': 'transposeB',
                    'type': 'bool',
                    'defaultValue': false
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'MatMul',
            'category': 'matrices',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'transpose_a',
                    'name': 'transposeA',
                    'type': 'bool',
                    'defaultValue': false
                },
                {
                    'tfName': 'transpose_b',
                    'name': 'transposeB',
                    'type': 'bool',
                    'defaultValue': false
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'BatchMatMul',
            'category': 'matrices',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'adj_x',
                    'name': 'transposeA',
                    'type': 'bool',
                    'defaultValue': false
                },
                {
                    'tfName': 'adj_y',
                    'name': 'transposeB',
                    'type': 'bool',
                    'defaultValue': false
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'BatchMatMulV2',
            'category': 'matrices',
            'inputs': [
                {
                    'start': 0,
                    'name': 'a',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'b',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'adj_x',
                    'name': 'transposeA',
                    'type': 'bool',
                    'defaultValue': false
                },
                {
                    'tfName': 'adj_y',
                    'name': 'transposeB',
                    'type': 'bool',
                    'defaultValue': false
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Transpose',
            'category': 'matrices',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'perm',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Einsum',
            'category': 'matrices',
            'inputs': [
                {
                    'start': 0,
                    'end': 0,
                    'name': 'tensors',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'equation',
                    'name': 'equation',
                    'type': 'string'
                },
                {
                    'tfName': 'N',
                    'name': 'n',
                    'type': 'number',
                    'defaultValue': 2
                },
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        }
    ];

    var matrices = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$7
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$6 = [
        {
            'tfOpName': 'EuclideanNorm',
            'category': 'normalization',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'keep_dims',
                    'name': 'keepDims',
                    'type': 'bool',
                    'defaultValue': false
                }
            ]
        },
        {
            'tfOpName': 'FusedBatchNorm',
            'category': 'normalization',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'scale',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'offset',
                    'type': 'tensor'
                },
                {
                    'start': 3,
                    'name': 'mean',
                    'type': 'tensor'
                },
                {
                    'start': 4,
                    'name': 'variance',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'epsilon',
                    'name': 'epsilon',
                    'type': 'number',
                    'defaultValue': 0.001
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'FusedBatchNormV2',
            'category': 'normalization',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'scale',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'offset',
                    'type': 'tensor'
                },
                {
                    'start': 3,
                    'name': 'mean',
                    'type': 'tensor'
                },
                {
                    'start': 4,
                    'name': 'variance',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'epsilon',
                    'name': 'epsilon',
                    'type': 'number',
                    'defaultValue': 0.001
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'FusedBatchNormV3',
            'category': 'normalization',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'scale',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'offset',
                    'type': 'tensor'
                },
                {
                    'start': 3,
                    'name': 'mean',
                    'type': 'tensor'
                },
                {
                    'start': 4,
                    'name': 'variance',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'epsilon',
                    'name': 'epsilon',
                    'type': 'number',
                    'defaultValue': 0.001
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'LRN',
            'category': 'normalization',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'depth_radius',
                    'name': 'radius',
                    'type': 'number',
                    'defaultValue': 5
                },
                {
                    'tfName': 'bias',
                    'name': 'bias',
                    'type': 'number',
                    'defaultValue': 1
                },
                {
                    'tfName': 'alpha',
                    'name': 'alpha',
                    'type': 'number',
                    'defaultValue': 1
                },
                {
                    'tfName': 'beta',
                    'name': 'beta',
                    'type': 'number',
                    'defaultValue': 0.5
                }
            ]
        },
        {
            'tfOpName': 'Softmax',
            'category': 'normalization',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'LogSoftmax',
            'category': 'normalization',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'SparseToDense',
            'category': 'normalization',
            'inputs': [
                {
                    'start': 0,
                    'name': 'sparseIndices',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'outputShape',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'sparseValues',
                    'type': 'tensor'
                },
                {
                    'start': 3,
                    'name': 'defaultValue',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'validate_indices',
                    'name': 'validateIndices',
                    'type': 'bool',
                    'defaultValue': true,
                    'notSupported': true
                }
            ]
        }
    ];

    var normalization = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$6
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$5 = [
        {
            'tfOpName': 'Bincount',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'size',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'weights',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'DenseBincount',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'size',
                    'type': 'number'
                },
                {
                    'start': 2,
                    'name': 'weights',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'binary_output',
                    'name': 'binaryOutput',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'Max',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'keep_dims',
                    'name': 'keepDims',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'Mean',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'keep_dims',
                    'name': 'keepDims',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'Min',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'keep_dims',
                    'name': 'keepDims',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'Sum',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'keep_dims',
                    'name': 'keepDims',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'All',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'keep_dims',
                    'name': 'keepDims',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'Any',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'keep_dims',
                    'name': 'keepDims',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'ArgMax',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number'
                }
            ]
        },
        {
            'tfOpName': 'ArgMin',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number'
                }
            ]
        },
        {
            'tfOpName': 'Prod',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'keep_dims',
                    'name': 'keepDims',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'Cumprod',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'exclusive',
                    'name': 'exclusive',
                    'type': 'bool'
                },
                {
                    'tfName': 'reverse',
                    'name': 'reverse',
                    'type': 'bool'
                }
            ]
        },
        {
            'tfOpName': 'Cumsum',
            'category': 'reduction',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'exclusive',
                    'name': 'exclusive',
                    'type': 'bool'
                },
                {
                    'tfName': 'reverse',
                    'name': 'reverse',
                    'type': 'bool'
                }
            ]
        }
    ];

    var reduction = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$5
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$4 = [
        {
            'tfOpName': 'ConcatV2',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'end': -1,
                    'name': 'tensors',
                    'type': 'tensors'
                },
                {
                    'start': -1,
                    'name': 'axis',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'N',
                    'name': 'n',
                    'type': 'number',
                    'defaultValue': 2
                }
            ]
        },
        {
            'tfOpName': 'Concat',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 1,
                    'end': 0,
                    'name': 'tensors',
                    'type': 'tensors'
                },
                {
                    'start': 0,
                    'name': 'axis',
                    'type': 'number'
                }
            ],
            'attrs': [
                {
                    'tfName': 'N',
                    'name': 'n',
                    'type': 'number',
                    'defaultValue': 2
                }
            ]
        },
        {
            'tfOpName': 'GatherV2',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'axis',
                    'type': 'number',
                    'defaultValue': 0
                }
            ],
            'attrs': [
                {
                    'tfName': 'batch_dims',
                    'name': 'batchDims',
                    'type': 'number',
                    'defaultValue': 0
                }
            ]
        },
        {
            'tfOpName': 'Gather',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'validate_indices',
                    'name': 'validateIndices',
                    'type': 'bool',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Reverse',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'dims',
                    'type': 'bool[]'
                }
            ]
        },
        {
            'tfOpName': 'ReverseV2',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'Slice',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'begin',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'size',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'StridedSlice',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'begin',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'end',
                    'type': 'number[]'
                },
                {
                    'start': 3,
                    'name': 'strides',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'begin_mask',
                    'name': 'beginMask',
                    'type': 'number',
                    'defaultValue': 0
                },
                {
                    'tfName': 'end_mask',
                    'name': 'endMask',
                    'type': 'number',
                    'defaultValue': 0
                },
                {
                    'tfName': 'new_axis_mask',
                    'name': 'newAxisMask',
                    'type': 'number',
                    'defaultValue': 0
                },
                {
                    'tfName': 'ellipsis_mask',
                    'name': 'ellipsisMask',
                    'type': 'number',
                    'defaultValue': 0
                },
                {
                    'tfName': 'shrink_axis_mask',
                    'name': 'shrinkAxisMask',
                    'type': 'number',
                    'defaultValue': 0
                }
            ]
        },
        {
            'tfOpName': 'Pack',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'end': 0,
                    'name': 'tensors',
                    'type': 'tensors'
                }
            ],
            'attrs': [
                {
                    'tfName': 'axis',
                    'name': 'axis',
                    'type': 'number',
                    'defaultValue': 0
                }
            ]
        },
        {
            'tfOpName': 'Unpack',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'tensor',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'axis',
                    'name': 'axis',
                    'type': 'number',
                    'defaultValue': 0
                },
                {
                    'tfName': 'num',
                    'name': 'num',
                    'type': 'number',
                    'defaultValue': 0,
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'Tile',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'reps',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'Split',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'axis',
                    'type': 'number',
                    'defaultValue': 0
                },
                {
                    'start': 1,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'num_split',
                    'name': 'numOrSizeSplits',
                    'type': 'number',
                    'defaultValue': 1
                }
            ]
        },
        {
            'tfOpName': 'SplitV',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'numOrSizeSplits',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'axis',
                    'type': 'number',
                    'defaultValue': 0
                }
            ]
        },
        {
            'tfOpName': 'ScatterNd',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'indices',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'values',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'shape',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'GatherNd',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'SparseToDense',
            'category': 'slice_join',
            'inputs': [
                {
                    'start': 0,
                    'name': 'sparseIndices',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'outputShape',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'sparseValues',
                    'type': 'tensor'
                },
                {
                    'start': 3,
                    'name': 'defaultValue',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'validate_indices',
                    'name': 'validateIndices',
                    'type': 'bool',
                    'defaultValue': false,
                    'notSupported': true
                }
            ]
        }
    ];

    var sliceJoin = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$4
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$3 = [
        {
            'tfOpName': 'SparseFillEmptyRows',
            'category': 'sparse',
            'inputs': [
                {
                    'start': 0,
                    'name': 'indices',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'values',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'denseShape',
                    'type': 'tensor'
                },
                {
                    'start': 3,
                    'name': 'defaultValue',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'SparseReshape',
            'category': 'sparse',
            'inputs': [
                {
                    'start': 0,
                    'name': 'inputIndices',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'inputShape',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'newShape',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'T',
                    'name': 'dtype',
                    'type': 'dtype',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'SparseSegmentMean',
            'category': 'sparse',
            'inputs': [
                {
                    'start': 0,
                    'name': 'data',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'segmentIds',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'SparseSegmentSum',
            'category': 'sparse',
            'inputs': [
                {
                    'start': 0,
                    'name': 'data',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'indices',
                    'type': 'tensor'
                },
                {
                    'start': 2,
                    'name': 'segmentIds',
                    'type': 'tensor'
                }
            ]
        }
    ];

    var sparse = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$3
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$2 = [
        {
            'tfOpName': 'FFT',
            'category': 'spectral',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'IFFT',
            'category': 'spectral',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ]
        },
        {
            'tfOpName': 'RFFT',
            'category': 'spectral',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'fft_length',
                    'type': 'number',
                    'notSupported': true
                }
            ]
        },
        {
            'tfOpName': 'IRFFT',
            'category': 'spectral',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'fft_length',
                    'type': 'number',
                    'notSupported': true
                }
            ]
        }
    ];

    var spectral = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$2
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json$1 = [
        {
            'tfOpName': 'StringNGrams',
            'category': 'string',
            'inputs': [
                {
                    'start': 0,
                    'name': 'data',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'dataSplits',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'separator',
                    'name': 'separator',
                    'type': 'string'
                },
                {
                    'tfName': 'ngram_widths',
                    'name': 'nGramWidths',
                    'type': 'number[]'
                },
                {
                    'tfName': 'left_pad',
                    'name': 'leftPad',
                    'type': 'string'
                },
                {
                    'tfName': 'right_pad',
                    'name': 'rightPad',
                    'type': 'string'
                },
                {
                    'tfName': 'pad_width',
                    'name': 'padWidth',
                    'type': 'number'
                },
                {
                    'tfName': 'preserve_short_sequences',
                    'name': 'preserveShortSequences',
                    'type': 'bool'
                }
            ],
            'outputs': [
                'ngrams',
                'ngrams_splits'
            ]
        },
        {
            'tfOpName': 'StringSplit',
            'category': 'string',
            'inputs': [
                {
                    'start': 0,
                    'name': 'input',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'delimiter',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'skip_empty',
                    'name': 'skipEmpty',
                    'type': 'bool'
                }
            ],
            'outputs': [
                'indices',
                'values',
                'shape'
            ]
        },
        {
            'tfOpName': 'StringToHashBucketFast',
            'category': 'string',
            'inputs': [
                {
                    'start': 0,
                    'name': 'input',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'num_buckets',
                    'name': 'numBuckets',
                    'type': 'number'
                }
            ]
        }
    ];

    var string = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json$1
    });

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const json = [
        {
            'tfOpName': 'Cast',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'SrcT',
                    'name': 'sdtype',
                    'type': 'dtype',
                    'notSupported': true
                },
                {
                    'tfName': 'DstT',
                    'name': 'dtype',
                    'type': 'dtype'
                }
            ]
        },
        {
            'tfOpName': 'ExpandDims',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'axis',
                    'type': 'number'
                }
            ]
        },
        {
            'tfOpName': 'MirrorPad',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'padding',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'mode',
                    'name': 'mode',
                    'type': 'string'
                }
            ]
        },
        {
            'tfOpName': 'Pad',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'padding',
                    'type': 'number[]'
                }
            ],
            'attrs': [
                {
                    'tfName': 'constant_value',
                    'name': 'constantValue',
                    'type': 'number',
                    'defaultValue': 0
                }
            ]
        },
        {
            'tfOpName': 'PadV2',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'padding',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'constantValue',
                    'type': 'number',
                    'defaultValue': 0
                }
            ]
        },
        {
            'tfOpName': 'Reshape',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'shape',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'Squeeze',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'axis',
                    'tfDeprecatedName': 'squeeze_dims',
                    'name': 'axis',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'SpaceToBatchND',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'blockShape',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'paddings',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'BatchToSpaceND',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'blockShape',
                    'type': 'number[]'
                },
                {
                    'start': 2,
                    'name': 'crops',
                    'type': 'number[]'
                }
            ]
        },
        {
            'tfOpName': 'DepthToSpace',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                }
            ],
            'attrs': [
                {
                    'tfName': 'block_size',
                    'name': 'blockSize',
                    'type': 'number'
                },
                {
                    'tfName': 'data_format',
                    'name': 'dataFormat',
                    'type': 'string'
                }
            ]
        },
        {
            'tfOpName': 'BroadcastTo',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 'x',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 'shape',
                    'type': 'number[]'
                }
            ],
            'attrs': []
        },
        {
            'tfOpName': 'BroadcastArgs',
            'category': 'transformation',
            'inputs': [
                {
                    'start': 0,
                    'name': 's0',
                    'type': 'tensor'
                },
                {
                    'start': 1,
                    'name': 's1',
                    'type': 'tensor'
                }
            ],
            'attrs': []
        }
    ];

    var transformation = /*#__PURE__*/Object.freeze({
        __proto__: null,
        json: json
    });

    class OperationMapper {
        // Singleton instance for the mapper
        static get Instance() {
            return this._instance || (this._instance = new this());
        }
        // Loads the op mapping from the JSON file.
        constructor() {
            const ops = [
                arithmetic, basicMath, control, convolution, creation, dynamic,
                evaluation, graph, hashTable, image, logical, matrices, normalization,
                reduction, sliceJoin, sparse, spectral, string, transformation
            ];
            const mappersJson = [].concat(...ops.map(op => op.json));
            this.opMappers = mappersJson.reduce((map, mapper) => {
                map[mapper.tfOpName] = mapper;
                return map;
            }, {});
        }
        // Converts the model inference graph from Tensorflow GraphDef to local
        // representation for TensorFlow.js API
        transformGraph(graph, signature = {}) {
            const tfNodes = graph.node;
            const placeholders = [];
            const weights = [];
            const initNodes = [];
            const nodes = tfNodes.reduce((map, node) => {
                map[node.name] = this.mapNode(node);
                if (node.op.startsWith('Placeholder')) {
                    placeholders.push(map[node.name]);
                }
                else if (node.op === 'Const') {
                    weights.push(map[node.name]);
                }
                else if (node.input == null || node.input.length === 0) {
                    initNodes.push(map[node.name]);
                }
                return map;
            }, {});
            let inputs = [];
            const outputs = [];
            let inputNodeNameToKey = {};
            let outputNodeNameToKey = {};
            if (signature != null) {
                inputNodeNameToKey = this.mapSignatureEntries(signature.inputs);
                outputNodeNameToKey = this.mapSignatureEntries(signature.outputs);
            }
            const allNodes = Object.keys(nodes);
            allNodes.forEach(key => {
                const node = nodes[key];
                node.inputNames.forEach((name, index) => {
                    const [nodeName, , outputName] = getNodeNameAndIndex(name);
                    const inputNode = nodes[nodeName];
                    if (inputNode.outputs != null) {
                        const outputIndex = inputNode.outputs.indexOf(outputName);
                        if (outputIndex !== -1) {
                            const inputName = `${nodeName}:${outputIndex}`;
                            // update the input name to use the mapped output index directly.
                            node.inputNames[index] = inputName;
                        }
                    }
                    node.inputs.push(inputNode);
                    inputNode.children.push(node);
                });
            });
            // if signature has not outputs set, add any node that does not have
            // outputs.
            if (Object.keys(outputNodeNameToKey).length === 0) {
                allNodes.forEach(key => {
                    const node = nodes[key];
                    if (node.children.length === 0) {
                        outputs.push(node);
                    }
                });
            }
            else {
                Object.keys(outputNodeNameToKey).forEach(name => {
                    const [nodeName,] = getNodeNameAndIndex(name);
                    const node = nodes[nodeName];
                    if (node != null) {
                        node.signatureKey = outputNodeNameToKey[name];
                        outputs.push(node);
                    }
                });
            }
            if (Object.keys(inputNodeNameToKey).length > 0) {
                Object.keys(inputNodeNameToKey).forEach(name => {
                    const [nodeName,] = getNodeNameAndIndex(name);
                    const node = nodes[nodeName];
                    if (node) {
                        node.signatureKey = inputNodeNameToKey[name];
                        inputs.push(node);
                    }
                });
            }
            else {
                inputs = placeholders;
            }
            let functions = {};
            if (graph.library != null && graph.library.function != null) {
                functions = graph.library.function.reduce((functions, func) => {
                    functions[func.signature.name] = this.mapFunction(func);
                    return functions;
                }, {});
            }
            const result = { nodes, inputs, outputs, weights, placeholders, signature, functions };
            if (initNodes.length > 0) {
                result.initNodes = initNodes;
            }
            return result;
        }
        mapSignatureEntries(entries) {
            return Object.keys(entries || {})
                .reduce((prev, curr) => {
                prev[entries[curr].name] = curr;
                return prev;
            }, {});
        }
        mapNode(node) {
            // Unsupported ops will cause an error at run-time (not parse time), since
            // they may not be used by the actual execution subgraph.
            const mapper = getRegisteredOp(node.op) || this.opMappers[node.op] || {};
            if (node.attr == null) {
                node.attr = {};
            }
            const newNode = {
                name: node.name,
                op: node.op,
                category: mapper.category,
                inputNames: (node.input ||
                    []).map(input => input.startsWith('^') ? input.slice(1) : input),
                inputs: [],
                children: [],
                inputParams: {},
                attrParams: {},
                rawAttrs: node.attr,
                outputs: mapper.outputs
            };
            if (mapper.inputs != null) {
                newNode.inputParams =
                    mapper.inputs.reduce((map, param) => {
                        map[param.name] = {
                            type: param.type,
                            inputIndexStart: param.start,
                            inputIndexEnd: param.end
                        };
                        return map;
                    }, {});
            }
            if (mapper.attrs != null) {
                newNode.attrParams =
                    mapper.attrs.reduce((map, param) => {
                        const type = param.type;
                        let value = undefined;
                        switch (param.type) {
                            case 'string':
                                value = getStringParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getStringParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'string[]':
                                value = getStringArrayParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getStringArrayParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'number':
                                value = getNumberParam(node.attr, param.tfName, (param.defaultValue || 0));
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getNumberParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'number[]':
                                value = getNumericArrayParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getNumericArrayParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'bool':
                                value = getBoolParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getBoolParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'bool[]':
                                value = getBoolArrayParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getBoolArrayParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'shape':
                                value = getTensorShapeParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getTensorShapeParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'shape[]':
                                value = getTensorShapeArrayParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getTensorShapeArrayParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'dtype':
                                value = getDtypeParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getDtypeParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'dtype[]':
                                value = getDtypeArrayParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getDtypeArrayParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'func':
                                value = getFuncParam(node.attr, param.tfName, param.defaultValue);
                                if (value === undefined && !!param.tfDeprecatedName) {
                                    value = getFuncParam(node.attr, param.tfDeprecatedName, param.defaultValue);
                                }
                                break;
                            case 'tensor':
                            case 'tensors':
                                break;
                            default:
                                throw new Error(`Unsupported param type: ${param.type} for op: ${node.op}`);
                        }
                        map[param.name] = { value, type };
                        return map;
                    }, {});
            }
            return newNode;
        }
        // map the TFunctionDef to TFJS graph object
        mapFunction(functionDef) {
            const tfNodes = functionDef.nodeDef;
            const placeholders = [];
            const weights = [];
            let nodes = {};
            if (tfNodes != null) {
                nodes = tfNodes.reduce((map, node) => {
                    map[node.name] = this.mapNode(node);
                    if (node.op === 'Const') {
                        weights.push(map[node.name]);
                    }
                    return map;
                }, {});
            }
            const inputs = [];
            const outputs = [];
            functionDef.signature.inputArg.forEach(arg => {
                const [nodeName,] = getNodeNameAndIndex(arg.name);
                const node = {
                    name: nodeName,
                    op: 'Placeholder',
                    inputs: [],
                    inputNames: [],
                    category: 'graph',
                    inputParams: {},
                    attrParams: { dtype: { value: parseDtypeParam(arg.type), type: 'dtype' } },
                    children: []
                };
                node.signatureKey = arg.name;
                inputs.push(node);
                nodes[nodeName] = node;
            });
            const allNodes = Object.keys(nodes);
            allNodes.forEach(key => {
                const node = nodes[key];
                node.inputNames.forEach((name, index) => {
                    const [nodeName, , outputName] = getNodeNameAndIndex(name);
                    const inputNode = nodes[nodeName];
                    if (inputNode.outputs != null) {
                        const outputIndex = inputNode.outputs.indexOf(outputName);
                        if (outputIndex !== -1) {
                            const inputName = `${nodeName}:${outputIndex}`;
                            // update the input name to use the mapped output index directly.
                            node.inputNames[index] = inputName;
                        }
                    }
                    node.inputs.push(inputNode);
                    inputNode.children.push(node);
                });
            });
            const returnNodeMap = functionDef.ret;
            functionDef.signature.outputArg.forEach(output => {
                const [nodeName, index] = getNodeNameAndIndex(returnNodeMap[output.name]);
                const node = nodes[nodeName];
                if (node != null) {
                    node.defaultOutput = index;
                    outputs.push(node);
                }
            });
            const signature = this.mapArgsToSignature(functionDef);
            return { nodes, inputs, outputs, weights, placeholders, signature };
        }
        mapArgsToSignature(functionDef) {
            return {
                methodName: functionDef.signature.name,
                inputs: functionDef.signature.inputArg.reduce((map, arg) => {
                    map[arg.name] = this.mapArgToTensorInfo(arg);
                    return map;
                }, {}),
                outputs: functionDef.signature.outputArg.reduce((map, arg) => {
                    map[arg.name] = this.mapArgToTensorInfo(arg, functionDef.ret);
                    return map;
                }, {}),
            };
        }
        mapArgToTensorInfo(arg, nameMap) {
            let name = arg.name;
            if (nameMap != null) {
                name = nameMap[name];
            }
            return { name, dtype: arg.type };
        }
    }
    function decodeBase64(text) {
        const global = env$1().global;
        if (typeof global.atob !== 'undefined') {
            return global.atob(text);
        }
        else if (typeof Buffer$1 !== 'undefined') {
            return new Buffer$1(text, 'base64').toString();
        }
        else {
            throw new Error('Unable to decode base64 in this environment. ' +
                'Missing built-in atob() or Buffer()');
        }
    }
    function parseStringParam(s, keepCase) {
        const value = Array.isArray(s) ? String.fromCharCode.apply(null, s) : decodeBase64(s);
        return keepCase ? value : value.toLowerCase();
    }
    function getStringParam(attrs, name, def, keepCase = false) {
        const param = attrs[name];
        if (param != null) {
            return parseStringParam(param.s, keepCase);
        }
        return def;
    }
    function getBoolParam(attrs, name, def) {
        const param = attrs[name];
        return param ? param.b : def;
    }
    function getNumberParam(attrs, name, def) {
        const param = attrs[name] || {};
        const value = param['i'] != null ? param['i'] : (param['f'] != null ? param['f'] : def);
        return (typeof value === 'number') ? value : parseInt(value, 10);
    }
    function parseDtypeParam(value) {
        if (typeof (value) === 'string') {
            // tslint:disable-next-line:no-any
            value = DataType[value];
        }
        switch (value) {
            case DataType.DT_FLOAT:
            case DataType.DT_HALF:
                return 'float32';
            case DataType.DT_INT32:
            case DataType.DT_INT64:
            case DataType.DT_INT8:
            case DataType.DT_UINT8:
                return 'int32';
            case DataType.DT_BOOL:
                return 'bool';
            case DataType.DT_DOUBLE:
                return 'float32';
            case DataType.DT_STRING:
                return 'string';
            default:
                // Unknown dtype error will happen at runtime (instead of parse time),
                // since these nodes might not be used by the actual subgraph execution.
                return null;
        }
    }
    function getFuncParam(attrs, name, def) {
        const param = attrs[name];
        if (param && param.func) {
            return param.func.name;
        }
        return def;
    }
    function getDtypeParam(attrs, name, def) {
        const param = attrs[name];
        if (param && param.type) {
            return parseDtypeParam(param.type);
        }
        return def;
    }
    function getDtypeArrayParam(attrs, name, def) {
        const param = attrs[name];
        if (param && param.list && param.list.type) {
            return param.list.type.map(v => parseDtypeParam(v));
        }
        return def;
    }
    function parseTensorShapeParam(shape) {
        if (shape.unknownRank) {
            return undefined;
        }
        if (shape.dim != null) {
            return shape.dim.map(dim => (typeof dim.size === 'number') ? dim.size : parseInt(dim.size, 10));
        }
        return [];
    }
    function getTensorShapeParam(attrs, name, def) {
        const param = attrs[name];
        if (param && param.shape) {
            return parseTensorShapeParam(param.shape);
        }
        return def;
    }
    function getNumericArrayParam(attrs, name, def) {
        const param = attrs[name];
        if (param) {
            return ((param.list.f && param.list.f.length ? param.list.f :
                param.list.i) ||
                [])
                .map(v => (typeof v === 'number') ? v : parseInt(v, 10));
        }
        return def;
    }
    function getStringArrayParam(attrs, name, def, keepCase = false) {
        const param = attrs[name];
        if (param && param.list && param.list.s) {
            return param.list.s.map((v) => {
                return parseStringParam(v, keepCase);
            });
        }
        return def;
    }
    function getTensorShapeArrayParam(attrs, name, def) {
        const param = attrs[name];
        if (param && param.list && param.list.shape) {
            return param.list.shape.map((v) => {
                return parseTensorShapeParam(v);
            });
        }
        return def;
    }
    function getBoolArrayParam(attrs, name, def) {
        const param = attrs[name];
        if (param && param.list && param.list.b) {
            return param.list.b;
        }
        return def;
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Helper class for lookup inputs and params for nodes in the model graph.
     */
    class NodeValueImpl {
        constructor(node, tensorMap, context) {
            this.node = node;
            this.tensorMap = tensorMap;
            this.context = context;
            this.inputs = [];
            this.attrs = {};
            this.inputs = node.inputNames.map(name => this.getInput(name));
            if (node.rawAttrs != null) {
                this.attrs = Object.keys(node.rawAttrs)
                    .reduce((attrs, key) => {
                    attrs[key] = this.getAttr(key);
                    return attrs;
                }, {});
            }
        }
        /**
         * Return the value of the attribute or input param.
         * @param name String: name of attribute or input param.
         */
        getInput(name) {
            return getTensor(name, this.tensorMap, this.context);
        }
        /**
         * Return the value of the attribute or input param.
         * @param name String: name of attribute or input param.
         */
        getAttr(name, defaultValue) {
            const value = this.node.rawAttrs[name];
            if (value.tensor != null) {
                return getTensor(name, this.tensorMap, this.context);
            }
            if (value.i != null || value.f != null) {
                return getNumberParam(this.node.rawAttrs, name, defaultValue);
            }
            if (value.s != null) {
                return getStringParam(this.node.rawAttrs, name, defaultValue);
            }
            if (value.b != null) {
                return getBoolParam(this.node.rawAttrs, name, defaultValue);
            }
            if (value.shape != null) {
                return getTensorShapeParam(this.node.rawAttrs, name, defaultValue);
            }
            if (value.type != null) {
                return getDtypeParam(this.node.rawAttrs, name, defaultValue);
            }
            if (value.list != null) {
                if (value.list.i != null || value.list.f != null) {
                    return getNumericArrayParam(this.node.rawAttrs, name, defaultValue);
                }
                if (value.list.s != null) {
                    return getStringArrayParam(this.node.rawAttrs, name, defaultValue);
                }
                if (value.list.shape != null) {
                    return getTensorShapeArrayParam(this.node.rawAttrs, name, defaultValue);
                }
                if (value.list.b != null) {
                    return getBoolArrayParam(this.node.rawAttrs, name, defaultValue);
                }
                if (value.list.type != null) {
                    return getDtypeArrayParam(this.node.rawAttrs, name, defaultValue);
                }
            }
            return defaultValue;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$j = (node, tensorMap, context) => {
        switch (node.op) {
            case 'BiasAdd':
            case 'AddV2':
            case 'Add': {
                return [add$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'AddN': {
                return [addN$2(getParamValue('tensors', node, tensorMap, context))];
            }
            case 'FloorMod':
            case 'Mod':
                return [mod$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            case 'Mul':
                return [mul(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            case 'RealDiv':
            case 'Div': {
                return [div$1(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'DivNoNan': {
                return [divNoNan(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'FloorDiv': {
                return [floorDiv$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'Sub': {
                return [sub$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'Minimum': {
                return [minimum$3(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'Maximum': {
                return [maximum$3(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'Pow': {
                return [pow$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'SquaredDifference': {
                return [squaredDifference$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$i = (node, tensorMap, context) => {
        switch (node.op) {
            case 'Abs':
            case 'ComplexAbs':
                return [abs$2(getParamValue('x', node, tensorMap, context))];
            case 'Acos':
                return [acos$2(getParamValue('x', node, tensorMap, context))];
            case 'Acosh':
                return [acosh$2(getParamValue('x', node, tensorMap, context))];
            case 'Asin':
                return [asin$2(getParamValue('x', node, tensorMap, context))];
            case 'Asinh':
                return [asinh$2(getParamValue('x', node, tensorMap, context))];
            case 'Atan':
                return [atan$2(getParamValue('x', node, tensorMap, context))];
            case 'Atan2':
                return [atan2$2(getParamValue('x', node, tensorMap, context), getParamValue('y', node, tensorMap, context))];
            case 'Atanh':
                return [atanh$2(getParamValue('x', node, tensorMap, context))];
            case 'Ceil':
                return [ceil$2(getParamValue('x', node, tensorMap, context))];
            case 'Complex':
                return [complex$2(getParamValue('real', node, tensorMap, context), getParamValue('imag', node, tensorMap, context))];
            case 'Cos':
                return [cos$2(getParamValue('x', node, tensorMap, context))];
            case 'Cosh':
                return [cosh$2(getParamValue('x', node, tensorMap, context))];
            case 'Elu':
                return [elu$4(getParamValue('x', node, tensorMap, context))];
            case 'Erf':
                return [erf$2(getParamValue('x', node, tensorMap, context))];
            case 'Exp':
                return [exp$2(getParamValue('x', node, tensorMap, context))];
            case 'Expm1': {
                return [expm1$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Floor':
                return [floor$3(getParamValue('x', node, tensorMap, context))];
            case 'Log':
                return [log$3(getParamValue('x', node, tensorMap, context))];
            case 'Log1p': {
                return [log1p$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Imag':
                return [imag$2(getParamValue('x', node, tensorMap, context))];
            case 'Neg':
                return [neg$2(getParamValue('x', node, tensorMap, context))];
            case 'Reciprocal': {
                return [reciprocal$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Real':
                return [real$2(getParamValue('x', node, tensorMap, context))];
            case 'Relu':
                return [relu$2(getParamValue('x', node, tensorMap, context))];
            case 'Round': {
                return [round$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Selu':
                return [selu$2(getParamValue('x', node, tensorMap, context))];
            case 'Sigmoid':
                return [sigmoid$2(getParamValue('x', node, tensorMap, context))];
            case 'Sin':
                return [sin$2(getParamValue('x', node, tensorMap, context))];
            case 'Sign': {
                return [sign$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Sinh': {
                return [sinh$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Softplus': {
                return [softplus$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Sqrt': {
                return [sqrt$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Square': {
                return [square$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Tanh': {
                return [tanh$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'Tan':
                return [tan$2(getParamValue('x', node, tensorMap, context))];
            case 'ClipByValue':
                return [clipByValue$2(getParamValue('x', node, tensorMap, context), getParamValue('clipValueMin', node, tensorMap, context), getParamValue('clipValueMax', node, tensorMap, context))];
            case 'Relu6':
                return [relu6$2(getParamValue('x', node, tensorMap, context))];
            case 'Rsqrt':
                return [rsqrt$2(getTensor(node.inputNames[0], tensorMap, context))];
            case 'Prod':
                return [prod$2(getParamValue('x', node, tensorMap, context), getParamValue('axes', node, tensorMap, context))];
            case 'LeakyRelu':
                return [leakyRelu$2(getParamValue('x', node, tensorMap, context), getParamValue('alpha', node, tensorMap, context))];
            case 'Prelu':
                return [prelu$3(getParamValue('x', node, tensorMap, context), getParamValue('alpha', node, tensorMap, context))];
            case 'IsNan':
                return [isNaN$3(getTensor(node.inputNames[0], tensorMap, context))];
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Used by TensorList and TensorArray to verify if elementShape matches, support
     * negative value as the dim shape.
     * @param shapeA
     * @param shapeB
     * @param errorMessagePrefix
     */
    function assertShapesMatchAllowUndefinedSize(shapeA, shapeB, errorMessagePrefix = '') {
        // constant shape means unknown rank
        if (typeof shapeA === 'number' || typeof shapeB === 'number') {
            return;
        }
        assert$4(shapeA.length === shapeB.length, () => errorMessagePrefix + ` Shapes ${shapeA} and ${shapeB} must match`);
        for (let i = 0; i < shapeA.length; i++) {
            const dim0 = shapeA[i];
            const dim1 = shapeB[i];
            assert$4(dim0 < 0 || dim1 < 0 || dim0 === dim1, () => errorMessagePrefix + ` Shapes ${shapeA} and ${shapeB} must match`);
        }
    }
    function fullDefinedShape(elementShape) {
        if (typeof elementShape === 'number' || elementShape.some(dim => dim < 0)) {
            return false;
        }
        return true;
    }
    /**
     * Generate the output element shape from the list elementShape, list tensors
     * and input param.
     * @param listElementShape
     * @param tensors
     * @param elementShape
     */
    function inferElementShape(listElementShape, tensors, elementShape) {
        let partialShape = mergeElementShape(listElementShape, elementShape);
        const notfullDefinedShape = !fullDefinedShape(partialShape);
        if (notfullDefinedShape && tensors.length === 0) {
            throw new Error(`Tried to calculate elements of an empty list` +
                ` with non-fully-defined elementShape: ${partialShape}`);
        }
        if (notfullDefinedShape) {
            tensors.forEach(tensor => {
                partialShape = mergeElementShape(tensor.shape, partialShape);
            });
        }
        if (!fullDefinedShape(partialShape)) {
            throw new Error(`Non-fully-defined elementShape: ${partialShape}`);
        }
        return partialShape;
    }
    function mergeElementShape(elementShapeA, elementShapeB) {
        if (typeof elementShapeA === 'number') {
            return elementShapeB;
        }
        if (typeof elementShapeB === 'number') {
            return elementShapeA;
        }
        if (elementShapeA.length !== elementShapeB.length) {
            throw new Error(`Incompatible ranks during merge: ${elementShapeA} vs. ${elementShapeB}`);
        }
        const result = [];
        for (let i = 0; i < elementShapeA.length; ++i) {
            const dim0 = elementShapeA[i];
            const dim1 = elementShapeB[i];
            if (dim0 >= 0 && dim1 >= 0 && dim0 !== dim1) {
                throw new Error(`Incompatible shape during merge: ${elementShapeA} vs. ${elementShapeB}`);
            }
            result[i] = dim0 >= 0 ? dim0 : dim1;
        }
        return result;
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * The TensorArray object keeps an array of Tensors.  It
     * allows reading from the array and writing to the array.
     */
    class TensorArray {
        constructor(name, dtype, maxSize, elementShape, identicalElementShapes, dynamicSize, clearAfterRead) {
            this.name = name;
            this.dtype = dtype;
            this.maxSize = maxSize;
            this.elementShape = elementShape;
            this.identicalElementShapes = identicalElementShapes;
            this.dynamicSize = dynamicSize;
            this.clearAfterRead = clearAfterRead;
            this.tensors = [];
            this.closed_ = false;
            this.idTensor = scalar(0);
            keep(this.idTensor);
        }
        get id() {
            return this.idTensor.id;
        }
        get closed() {
            return this.closed_;
        }
        /**
         * Dispose the tensors and idTensor and mark the TensoryArray as closed.
         */
        clearAndClose(keepIds) {
            this.tensors.forEach(tensor => {
                if (keepIds == null || !keepIds.has(tensor.tensor.id)) {
                    tensor.tensor.dispose();
                }
            });
            this.tensors = [];
            this.closed_ = true;
            this.idTensor.dispose();
        }
        size() {
            return this.tensors.length;
        }
        /**
         * Read the value at location index in the TensorArray.
         * @param index Number the index to read from.
         */
        read(index) {
            if (this.closed_) {
                throw new Error(`TensorArray ${this.name} has already been closed.`);
            }
            if (index < 0 || index >= this.size()) {
                throw new Error(`Tried to read from index ${index}, but array size is: ${this.size()}`);
            }
            const tensorWithState = this.tensors[index];
            if (tensorWithState.cleared) {
                throw new Error(`TensorArray ${this.name}: Could not read index ${index} twice because it was cleared after a previous read ` +
                    `(perhaps try setting clear_after_read = false?).`);
            }
            if (this.clearAfterRead) {
                tensorWithState.cleared = true;
            }
            tensorWithState.read = true;
            return tensorWithState.tensor;
        }
        /**
         * Helper method to read multiple tensors from the specified indices.
         */
        readMany(indices) {
            return indices.map(index => this.read(index));
        }
        /**
         * Write value into the index of the TensorArray.
         * @param index number the index to write to.
         * @param tensor
         */
        write(index, tensor) {
            if (this.closed_) {
                throw new Error(`TensorArray ${this.name} has already been closed.`);
            }
            if (index < 0 || !this.dynamicSize && index >= this.maxSize) {
                throw new Error(`Tried to write to index ${index}, but array is not resizeable and size is: ${this.maxSize}`);
            }
            const t = this.tensors[index] || {};
            if (tensor.dtype !== this.dtype) {
                throw new Error(`TensorArray ${this.name}: Could not write to TensorArray index ${index},
          because the value dtype is ${tensor.dtype}, but TensorArray dtype is ${this.dtype}.`);
            }
            // Set the shape for the first time write to unknow shape tensor array
            if (this.size() === 0 &&
                (this.elementShape == null || this.elementShape.length === 0)) {
                this.elementShape = tensor.shape;
            }
            assertShapesMatchAllowUndefinedSize(this.elementShape, tensor.shape, `TensorArray ${this.name}: Could not write to TensorArray index ${index}.`);
            if (t.read) {
                throw new Error(`TensorArray ${this.name}: Could not write to TensorArray index ${index}, because it has already been read.`);
            }
            if (t.written) {
                throw new Error(`TensorArray ${this.name}: Could not write to TensorArray index ${index}, because it has already been written.`);
            }
            t.tensor = tensor;
            keep(tensor);
            t.written = true;
            this.tensors[index] = t;
        }
        /**
         * Helper method to write multiple tensors to the specified indices.
         */
        writeMany(indices, tensors) {
            if (indices.length !== tensors.length) {
                throw new Error(`TensorArray ${this.name}: could not write multiple tensors,` +
                    `because the index size: ${indices.length} is not the same as tensors size: ${tensors.length}.`);
            }
            indices.forEach((i, index) => this.write(i, tensors[index]));
        }
        /**
         * Return selected values in the TensorArray as a packed Tensor. All of
         * selected values must have been written and their shapes must all match.
         * @param [indices] number[] Optional. Taking values in [0, max_value). If the
         *    TensorArray is not dynamic, max_value=size(). If not specified returns
         *    all tensors in the original order.
         * @param [dtype]
         */
        gather(indices, dtype) {
            if (!!dtype && dtype !== this.dtype) {
                throw new Error(`TensorArray dtype is ${this.dtype} but gather requested dtype ${dtype}`);
            }
            if (!indices) {
                indices = [];
                for (let i = 0; i < this.size(); i++) {
                    indices.push(i);
                }
            }
            else {
                indices = indices.slice(0, this.size());
            }
            if (indices.length === 0) {
                return tensor([], [0].concat(this.elementShape));
            }
            // Read all the PersistentTensors into a vector to keep track of
            // their memory.
            const tensors = this.readMany(indices);
            assertShapesMatchAllowUndefinedSize(this.elementShape, tensors[0].shape, 'TensorArray shape mismatch: ');
            return stack(tensors, 0);
        }
        /**
         * Return the values in the TensorArray as a concatenated Tensor.
         */
        concat(dtype) {
            if (!!dtype && dtype !== this.dtype) {
                throw new Error(`TensorArray dtype is ${this.dtype} but concat requested dtype ${dtype}`);
            }
            if (this.size() === 0) {
                return tensor([], [0].concat(this.elementShape));
            }
            const indices = [];
            for (let i = 0; i < this.size(); i++) {
                indices.push(i);
            }
            // Collect all the tensors from the tensors array.
            const tensors = this.readMany(indices);
            assertShapesMatchAllowUndefinedSize(this.elementShape, tensors[0].shape, `TensorArray shape mismatch: tensor array shape (${this.elementShape}) vs first tensor shape (${tensors[0].shape})`);
            return concat$2(tensors, 0);
        }
        /**
         * Scatter the values of a Tensor in specific indices of a TensorArray.
         * @param indices nummber[] values in [0, max_value). If the
         *    TensorArray is not dynamic, max_value=size().
         * @param tensor Tensor input tensor.
         */
        scatter(indices, tensor) {
            if (tensor.dtype !== this.dtype) {
                throw new Error(`TensorArray dtype is ${this.dtype} but tensor has dtype ${tensor.dtype}`);
            }
            if (indices.length !== tensor.shape[0]) {
                throw new Error(`Expected len(indices) == tensor.shape[0], but saw: ${indices.length} vs. ${tensor.shape[0]}`);
            }
            const maxIndex = Math.max(...indices);
            if (!this.dynamicSize && maxIndex >= this.maxSize) {
                throw new Error(`Max index must be < array size (${maxIndex}  vs. ${this.maxSize})`);
            }
            this.writeMany(indices, unstack(tensor, 0));
        }
        /**
         * Split the values of a Tensor into the TensorArray.
         * @param length number[] with the lengths to use when splitting value along
         *    its first dimension.
         * @param tensor Tensor, the tensor to split.
         */
        split(length, tensor) {
            if (tensor.dtype !== this.dtype) {
                throw new Error(`TensorArray dtype is ${this.dtype} but tensor has dtype ${tensor.dtype}`);
            }
            let totalLength = 0;
            const cumulativeLengths = length.map(len => {
                totalLength += len;
                return totalLength;
            });
            if (totalLength !== tensor.shape[0]) {
                throw new Error(`Expected sum of lengths to be equal to
          tensor.shape[0], but sum of lengths is
        ${totalLength}, and tensor's shape is: ${tensor.shape}`);
            }
            if (!this.dynamicSize && length.length !== this.maxSize) {
                throw new Error(`TensorArray's size is not equal to the size of lengths (${this.maxSize} vs. ${length.length}), ` +
                    'and the TensorArray is not marked as dynamically resizeable');
            }
            const elementPerRow = totalLength === 0 ? 0 : tensor.size / totalLength;
            const tensors = [];
            tidy(() => {
                tensor = reshape$3(tensor, [1, totalLength, elementPerRow]);
                for (let i = 0; i < length.length; ++i) {
                    const previousLength = (i === 0) ? 0 : cumulativeLengths[i - 1];
                    const indices = [0, previousLength, 0];
                    const sizes = [1, length[i], elementPerRow];
                    tensors[i] = reshape$3(slice$2(tensor, indices, sizes), this.elementShape);
                }
                return tensors;
            });
            const indices = [];
            for (let i = 0; i < length.length; i++) {
                indices[i] = i;
            }
            this.writeMany(indices, tensors);
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * TensorList stores a container of `tf.Tensor` objects, which are accessible
     * via tensors field.
     *
     * In order to get a copy of the underlying list, use the copy method:
     * ```
     *    TensorList b = a.copy();
     *    b.tensors().pushBack(t);  // This does not modify a.tensors().
     * ```
     *
     * Note that this is not a deep copy: the memory locations of the underlying
     * tensors will still point to the same locations of the corresponding tensors
     * in the original.
     */
    class TensorList {
        /**
         *
         * @param tensors list of tensors
         * @param elementShape shape of each tensor, this can be a single number (any
         * shape is allowed) or partial shape (dim = -1).
         * @param elementDtype data type of each tensor
         * @param maxNumElements The maximum allowed size of `tensors`. Defaults to -1
         *   meaning that the size of `tensors` is unbounded.
         */
        constructor(tensors, elementShape, elementDtype, maxNumElements = -1) {
            this.tensors = tensors;
            this.elementShape = elementShape;
            this.elementDtype = elementDtype;
            if (tensors != null) {
                tensors.forEach(tensor => {
                    if (elementDtype !== tensor.dtype) {
                        throw new Error(`Invalid data types; op elements ${elementDtype}, but list elements ${tensor.dtype}`);
                    }
                    assertShapesMatchAllowUndefinedSize(elementShape, tensor.shape, 'TensorList shape mismatch: ');
                    keep(tensor);
                });
            }
            this.idTensor = scalar(0);
            this.maxNumElements = maxNumElements;
            keep(this.idTensor);
        }
        get id() {
            return this.idTensor.id;
        }
        /**
         * Get a new TensorList containing a copy of the underlying tensor container.
         */
        copy() {
            return new TensorList([...this.tensors], this.elementShape, this.elementDtype);
        }
        /**
         * Dispose the tensors and idTensor and clear the tensor list.
         */
        clearAndClose(keepIds) {
            this.tensors.forEach(tensor => {
                if (keepIds == null || !keepIds.has(tensor.id)) {
                    tensor.dispose();
                }
            });
            this.tensors.length = 0;
            this.idTensor.dispose();
        }
        /**
         * The size of the tensors in the tensor list.
         */
        size() {
            return this.tensors.length;
        }
        /**
         * Return a tensor that stacks a list of rank-R tf.Tensors into one rank-(R+1)
         * tf.Tensor.
         * @param elementShape shape of each tensor
         * @param elementDtype data type of each tensor
         * @param numElements the number of elements to stack
         */
        stack(elementShape, elementDtype, numElements = -1) {
            if (elementDtype !== this.elementDtype) {
                throw new Error(`Invalid data types; op elements ${elementDtype}, but list elements ${this.elementDtype}`);
            }
            if (numElements !== -1 && this.tensors.length !== numElements) {
                throw new Error(`Operation expected a list with ${numElements} elements but got a list with ${this.tensors.length} elements.`);
            }
            assertShapesMatchAllowUndefinedSize(elementShape, this.elementShape, 'TensorList shape mismatch: ');
            const outputElementShape = inferElementShape(this.elementShape, this.tensors, elementShape);
            return tidy(() => {
                const reshapedTensors = this.tensors.map(tensor => reshape$3(tensor, outputElementShape));
                return stack(reshapedTensors, 0);
            });
        }
        /**
         * Pop a tensor from the end of the list.
         * @param elementShape shape of the tensor
         * @param elementDtype data type of the tensor
         */
        popBack(elementShape, elementDtype) {
            if (elementDtype !== this.elementDtype) {
                throw new Error(`Invalid data types; op elements ${elementDtype}, but list elements ${this.elementDtype}`);
            }
            if (this.size() === 0) {
                throw new Error('Trying to pop from an empty list.');
            }
            const outputElementShape = inferElementShape(this.elementShape, this.tensors, elementShape);
            const tensor = this.tensors.pop();
            assertShapesMatchAllowUndefinedSize(tensor.shape, elementShape, 'TensorList shape mismatch: ');
            return reshape$3(tensor, outputElementShape);
        }
        /**
         * Push a tensor to the end of the list.
         * @param tensor Tensor to be pushed.
         */
        pushBack(tensor) {
            if (tensor.dtype !== this.elementDtype) {
                throw new Error(`Invalid data types; op elements ${tensor.dtype}, but list elements ${this.elementDtype}`);
            }
            assertShapesMatchAllowUndefinedSize(tensor.shape, this.elementShape, 'TensorList shape mismatch: ');
            if (this.maxNumElements === this.size()) {
                throw new Error(`Trying to push element into a full list.`);
            }
            keep(tensor);
            this.tensors.push(tensor);
        }
        /**
         * Update the size of the list.
         * @param size the new size of the list.
         */
        resize(size) {
            if (size < 0) {
                throw new Error(`TensorListResize expects size to be non-negative. Got: ${size}`);
            }
            if (this.maxNumElements !== -1 && size > this.maxNumElements) {
                throw new Error(`TensorListResize input size ${size} is greater maxNumElement ${this.maxNumElements}.`);
            }
            const destTensorList = new TensorList([], this.elementShape, this.elementDtype, this.maxNumElements);
            destTensorList.tensors.length = size;
            for (let i = 0; i < Math.min(this.tensors.length, size); ++i) {
                destTensorList.tensors[i] = this.tensors[i];
            }
            return destTensorList;
        }
        /**
         * Retrieve the element at the provided index
         * @param elementShape shape of the tensor
         * @param elementDtype dtype of the tensor
         * @param elementIndex index of the tensor
         */
        getItem(elementIndex, elementShape, elementDtype) {
            if (elementDtype !== this.elementDtype) {
                throw new Error(`Invalid data types; op elements ${elementDtype}, but list elements ${this.elementDtype}`);
            }
            if (elementIndex < 0 || elementIndex > this.tensors.length) {
                throw new Error(`Trying to access element ${elementIndex} in a list with ${this.tensors.length} elements.`);
            }
            if (this.tensors[elementIndex] == null) {
                throw new Error(`element at index ${elementIndex} is null.`);
            }
            assertShapesMatchAllowUndefinedSize(this.tensors[elementIndex].shape, elementShape, 'TensorList shape mismatch: ');
            const outputElementShape = inferElementShape(this.elementShape, this.tensors, elementShape);
            return reshape$3(this.tensors[elementIndex], outputElementShape);
        }
        /**
         * Set the tensor at the index
         * @param elementIndex index of the tensor
         * @param tensor the tensor to be inserted into the list
         */
        setItem(elementIndex, tensor) {
            if (tensor.dtype !== this.elementDtype) {
                throw new Error(`Invalid data types; op elements ${tensor.dtype}, but list elements ${this.elementDtype}`);
            }
            if (elementIndex < 0 ||
                this.maxNumElements !== -1 && elementIndex >= this.maxNumElements) {
                throw new Error(`Trying to set element ${elementIndex} in a list with max ${this.maxNumElements} elements.`);
            }
            assertShapesMatchAllowUndefinedSize(this.elementShape, tensor.shape, 'TensorList shape mismatch: ');
            keep(tensor);
            this.tensors[elementIndex] = tensor;
        }
        /**
         * Return selected values in the TensorList as a stacked Tensor. All of
         * selected values must have been written and their shapes must all match.
         * @param indices indices of tensors to gather
         * @param elementDtype output tensor dtype
         * @param elementShape output tensor element shape
         */
        gather(indices, elementDtype, elementShape) {
            if (elementDtype !== this.elementDtype) {
                throw new Error(`Invalid data types; op elements ${elementDtype}, but list elements ${this.elementDtype}`);
            }
            assertShapesMatchAllowUndefinedSize(this.elementShape, elementShape, 'TensorList shape mismatch: ');
            // When indices is greater than the size of the list, indices beyond the
            // size of the list are ignored.
            indices = indices.slice(0, this.size());
            const outputElementShape = inferElementShape(this.elementShape, this.tensors, elementShape);
            if (indices.length === 0) {
                return tensor([], [0].concat(outputElementShape));
            }
            return tidy(() => {
                const tensors = indices.map(i => reshape$3(this.tensors[i], outputElementShape));
                return stack(tensors, 0);
            });
        }
        /**
         * Return the values in the TensorList as a concatenated Tensor.
         * @param elementDtype output tensor dtype
         * @param elementShape output tensor element shape
         */
        concat(elementDtype, elementShape) {
            if (!!elementDtype && elementDtype !== this.elementDtype) {
                throw new Error(`TensorList dtype is ${this.elementDtype} but concat requested dtype ${elementDtype}`);
            }
            assertShapesMatchAllowUndefinedSize(this.elementShape, elementShape, 'TensorList shape mismatch: ');
            const outputElementShape = inferElementShape(this.elementShape, this.tensors, elementShape);
            if (this.size() === 0) {
                return tensor([], [0].concat(outputElementShape));
            }
            return tidy(() => {
                const tensors = this.tensors.map(t => reshape$3(t, outputElementShape));
                return concat$2(tensors, 0);
            });
        }
    }
    /**
     * Creates a TensorList which, when stacked, has the value of tensor.
     * @param tensor from tensor
     * @param elementShape output tensor element shape
     */
    function fromTensor(tensor, elementShape, elementDtype) {
        const dtype = tensor.dtype;
        if (tensor.shape.length < 1) {
            throw new Error(`Tensor must be at least a vector, but saw shape: ${tensor.shape}`);
        }
        if (tensor.dtype !== elementDtype) {
            throw new Error(`Invalid data types; op elements ${tensor.dtype}, but list elements ${elementDtype}`);
        }
        const tensorElementShape = tensor.shape.slice(1);
        assertShapesMatchAllowUndefinedSize(tensorElementShape, elementShape, 'TensorList shape mismatch: ');
        const tensorList = unstack(tensor);
        return new TensorList(tensorList, elementShape, dtype);
    }
    /**
     * Return a TensorList of the given size with empty elements.
     * @param elementShape the shape of the future elements of the list
     * @param elementDtype the desired type of elements in the list
     * @param numElements the number of elements to reserve
     */
    function reserve(elementShape, elementDtype, numElements) {
        return new TensorList([], elementShape, elementDtype, numElements);
    }
    /**
     * Put tensors at specific indices of a stacked tensor into a TensorList.
     * @param indices list of indices on how to scatter the tensor.
     * @param tensor input tensor.
     * @param elementShape the shape of the future elements of the list
     * @param numElements the number of elements to scatter
     */
    function scatter(tensor, indices, elementShape, numElements) {
        if (indices.length !== tensor.shape[0]) {
            throw new Error(`Expected len(indices) == tensor.shape[0], but saw: ${indices.length} vs. ${tensor.shape[0]}`);
        }
        const maxIndex = Math.max(...indices);
        if (numElements != null && numElements !== -1 && maxIndex >= numElements) {
            throw new Error(`Max index must be < array size (${maxIndex}  vs. ${numElements})`);
        }
        const list = new TensorList([], elementShape, tensor.dtype, numElements);
        const tensors = unstack(tensor, 0);
        indices.forEach((value, index) => {
            list.setItem(value, tensors[index]);
        });
        return list;
    }
    /**
     * Split the values of a Tensor into a TensorList.
     * @param length the lengths to use when splitting value along
     *    its first dimension.
     * @param tensor the tensor to split.
     * @param elementShape the shape of the future elements of the list
     */
    function split$1(tensor, length, elementShape) {
        let totalLength = 0;
        const cumulativeLengths = length.map(len => {
            totalLength += len;
            return totalLength;
        });
        if (totalLength !== tensor.shape[0]) {
            throw new Error(`Expected sum of lengths to be equal to
          tensor.shape[0], but sum of lengths is
        ${totalLength}, and tensor's shape is: ${tensor.shape}`);
        }
        const shapeWithoutFirstDim = tensor.shape.slice(1);
        const outputElementShape = mergeElementShape(shapeWithoutFirstDim, elementShape);
        const elementPerRow = totalLength === 0 ? 0 : tensor.size / totalLength;
        const tensors = tidy(() => {
            const tensors = [];
            tensor = reshape$3(tensor, [1, totalLength, elementPerRow]);
            for (let i = 0; i < length.length; ++i) {
                const previousLength = (i === 0) ? 0 : cumulativeLengths[i - 1];
                const indices = [0, previousLength, 0];
                const sizes = [1, length[i], elementPerRow];
                tensors[i] = reshape$3(slice$2(tensor, indices, sizes), outputElementShape);
            }
            tensor.dispose();
            return tensors;
        });
        const list = new TensorList([], elementShape, tensor.dtype, length.length);
        for (let i = 0; i < tensors.length; i++) {
            list.setItem(i, tensors[i]);
        }
        return list;
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$h = async (node, tensorMap, context) => {
        switch (node.op) {
            case 'If':
            case 'StatelessIf': {
                const thenFunc = getParamValue('thenBranch', node, tensorMap, context);
                const elseFunc = getParamValue('elseBranch', node, tensorMap, context);
                const cond = getParamValue('cond', node, tensorMap, context);
                const args = getParamValue('args', node, tensorMap, context);
                const condValue = await cond.data();
                if (condValue[0]) {
                    return context.functionMap[thenFunc].executeFunctionAsync(args, context.tensorArrayMap, context.tensorListMap);
                }
                else {
                    return context.functionMap[elseFunc].executeFunctionAsync(args, context.tensorArrayMap, context.tensorListMap);
                }
            }
            case 'While':
            case 'StatelessWhile': {
                const bodyFunc = getParamValue('body', node, tensorMap, context);
                const condFunc = getParamValue('cond', node, tensorMap, context);
                const args = getParamValue('args', node, tensorMap, context);
                // Calculate the condition of the loop
                const condResult = (await context.functionMap[condFunc].executeFunctionAsync(args, context.tensorArrayMap, context.tensorListMap));
                const argIds = args.map(tensor => tensor.id);
                let condValue = await condResult[0].data();
                // Dispose the intermediate tensors for condition function
                condResult.forEach(tensor => {
                    if (!tensor.kept && argIds.indexOf(tensor.id) === -1) {
                        tensor.dispose();
                    }
                });
                let result = args;
                while (condValue[0]) {
                    // Record the previous result for intermediate tensor tracking
                    const origResult = result;
                    // Execution the body of the loop
                    result = await context.functionMap[bodyFunc].executeFunctionAsync(result, context.tensorArrayMap, context.tensorListMap);
                    const resultIds = result.map(tensor => tensor.id);
                    // Dispose the intermediate tensor for body function that is not global
                    // kept, not input/output of the body function
                    origResult.forEach(tensor => {
                        if (!tensor.kept && argIds.indexOf(tensor.id) === -1 &&
                            resultIds.indexOf(tensor.id) === -1) {
                            tensor.dispose();
                        }
                    });
                    // Recalcuate the condition of the loop using the latest results.
                    const condResult = (await context.functionMap[condFunc].executeFunctionAsync(result, context.tensorArrayMap, context.tensorListMap));
                    condValue = await condResult[0].data();
                    // Dispose the intermediate tensors for condition function
                    condResult.forEach(tensor => {
                        if (!tensor.kept && argIds.indexOf(tensor.id) === -1 &&
                            resultIds.indexOf(tensor.id) === -1) {
                            tensor.dispose();
                        }
                    });
                }
                return result;
            }
            case 'LoopCond': {
                const pred = getParamValue('pred', node, tensorMap, context);
                return [cloneTensor(pred)];
            }
            case 'Switch': {
                const pred = getParamValue('pred', node, tensorMap, context);
                let data = getParamValue('data', node, tensorMap, context);
                if (!data.kept) {
                    data = cloneTensor(data);
                }
                // Outputs nodes :0 => false, :1 => true
                return (await pred.data())[0] ? [undefined, data] : [data, undefined];
            }
            case 'Merge': {
                const inputName = node.inputNames.find(name => getTensor(name, tensorMap, context) !== undefined);
                if (inputName) {
                    const data = getTensor(inputName, tensorMap, context);
                    return [cloneTensor(data)];
                }
                return undefined;
            }
            case 'Enter': {
                const frameId = getParamValue('frameName', node, tensorMap, context);
                const data = getParamValue('tensor', node, tensorMap, context);
                context.enterFrame(frameId);
                return [cloneTensor(data)];
            }
            case 'Exit': {
                const data = getParamValue('tensor', node, tensorMap, context);
                context.exitFrame();
                return [cloneTensor(data)];
            }
            case 'NextIteration': {
                const data = getParamValue('tensor', node, tensorMap, context);
                context.nextIteration();
                return [cloneTensor(data)];
            }
            case 'TensorArrayV3': {
                const size = getParamValue('size', node, tensorMap, context);
                const dtype = getParamValue('dtype', node, tensorMap, context);
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                const dynamicSize = getParamValue('dynamicSize', node, tensorMap, context);
                const clearAfterRead = getParamValue('clearAfterRead', node, tensorMap, context);
                const identicalElementShapes = getParamValue('identicalElementShapes', node, tensorMap, context);
                const name = getParamValue('name', node, tensorMap, context);
                const tensorArray = new TensorArray(name, dtype, size, elementShape, identicalElementShapes, dynamicSize, clearAfterRead);
                context.addTensorArray(tensorArray);
                return [tensorArray.idTensor, scalar(1.0)];
            }
            case 'TensorArrayWriteV3': {
                const id = getParamValue('tensorArrayId', node, tensorMap, context);
                const index = getParamValue('index', node, tensorMap, context);
                const writeTensor = getParamValue('tensor', node, tensorMap, context);
                const writeTensorArray = context.getTensorArray(id.id);
                writeTensorArray.write(index, writeTensor);
                return [writeTensorArray.idTensor];
            }
            case 'TensorArrayReadV3': {
                const readId = getParamValue('tensorArrayId', node, tensorMap, context);
                const readIndex = getParamValue('index', node, tensorMap, context);
                const readTensorArray = context.getTensorArray(readId.id);
                return [readTensorArray.read(readIndex)];
            }
            case 'TensorArrayGatherV3': {
                const gatherId = getParamValue('tensorArrayId', node, tensorMap, context);
                const gatherIndices = getParamValue('indices', node, tensorMap, context);
                const gatherDtype = getParamValue('dtype', node, tensorMap, context);
                const gatherTensorArray = context.getTensorArray(gatherId.id);
                return [gatherTensorArray.gather(gatherIndices, gatherDtype)];
            }
            case 'TensorArrayScatterV3': {
                const scatterId = getParamValue('tensorArrayId', node, tensorMap, context);
                const scatterIndices = getParamValue('indices', node, tensorMap, context);
                const scatterTensor = getParamValue('tensor', node, tensorMap, context);
                const scatterTensorArray = context.getTensorArray(scatterId.id);
                scatterTensorArray.scatter(scatterIndices, scatterTensor);
                return [scatterTensorArray.idTensor];
            }
            case 'TensorArrayConcatV3': {
                const concatId = getParamValue('tensorArrayId', node, tensorMap, context);
                const concatTensorArray = context.getTensorArray(concatId.id);
                const concatDtype = getParamValue('dtype', node, tensorMap, context);
                return [concatTensorArray.concat(concatDtype)];
            }
            case 'TensorArraySplitV3': {
                const splitId = getParamValue('tensorArrayId', node, tensorMap, context);
                const splitTensor = getParamValue('tensor', node, tensorMap, context);
                const lengths = getParamValue('lengths', node, tensorMap, context);
                const splitTensorArray = context.getTensorArray(splitId.id);
                splitTensorArray.split(lengths, splitTensor);
                return [splitTensorArray.idTensor];
            }
            case 'TensorArraySizeV3': {
                const sizeId = getParamValue('tensorArrayId', node, tensorMap, context);
                const sizeTensorArray = context.getTensorArray(sizeId.id);
                return [scalar(sizeTensorArray.size(), 'int32')];
            }
            case 'TensorArrayCloseV3': {
                const closeId = getParamValue('tensorArrayId', node, tensorMap, context);
                const closeTensorArray = context.getTensorArray(closeId.id);
                closeTensorArray.clearAndClose();
                return [closeTensorArray.idTensor];
            }
            case 'TensorListSetItem': {
                const idTensor = getParamValue('tensorListId', node, tensorMap, context);
                const index = getParamValue('index', node, tensorMap, context);
                const writeTensor = getParamValue('tensor', node, tensorMap, context);
                const tensorList = context.getTensorList(idTensor.id);
                tensorList.setItem(index, writeTensor);
                return [tensorList.idTensor];
            }
            case 'TensorListGetItem': {
                const idTensor = getParamValue('tensorListId', node, tensorMap, context);
                const readIndex = getParamValue('index', node, tensorMap, context);
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                const elementDType = getParamValue('elementDType', node, tensorMap, context);
                const tensorList = context.getTensorList(idTensor.id);
                return [tensorList.getItem(readIndex, elementShape, elementDType)];
            }
            case 'TensorListScatterV2':
            case 'TensorListScatter': {
                const scatterIndices = getParamValue('indices', node, tensorMap, context);
                const scatterTensor = getParamValue('tensor', node, tensorMap, context);
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                const numElements = getParamValue('numElements', node, tensorMap, context);
                const tensorList = scatter(scatterTensor, scatterIndices, elementShape, numElements);
                context.addTensorList(tensorList);
                return [tensorList.idTensor];
            }
            case 'TensorListReserve':
            case 'EmptyTensorList': {
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                const elementDtype = getParamValue('elementDType', node, tensorMap, context);
                let numElementsParam;
                if (node.op === 'TensorListReserve') {
                    numElementsParam = 'numElements';
                }
                else {
                    numElementsParam = 'maxNumElements';
                }
                const numElements = getParamValue(numElementsParam, node, tensorMap, context);
                const tensorList = reserve(elementShape, elementDtype, numElements);
                context.addTensorList(tensorList);
                return [tensorList.idTensor];
            }
            case 'TensorListGather': {
                const gatherId = getParamValue('tensorListId', node, tensorMap, context);
                const gatherIndices = getParamValue('indices', node, tensorMap, context);
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                const elementDtype = getParamValue('elementDType', node, tensorMap, context);
                const tensorList = context.getTensorList(gatherId.id);
                return [tensorList.gather(gatherIndices, elementDtype, elementShape)];
            }
            case 'TensorListStack': {
                const idTensor = getParamValue('tensorListId', node, tensorMap, context);
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                const elementDtype = getParamValue('elementDType', node, tensorMap, context);
                const numElements = getParamValue('numElements', node, tensorMap, context);
                const tensorList = context.getTensorList(idTensor.id);
                return [tensorList.stack(elementShape, elementDtype, numElements)];
            }
            case 'TensorListFromTensor': {
                const tensor = getParamValue('tensor', node, tensorMap, context);
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                const elementDtype = getParamValue('elementDType', node, tensorMap, context);
                const tensorList = fromTensor(tensor, elementShape, elementDtype);
                context.addTensorList(tensorList);
                return [tensorList.idTensor];
            }
            case 'TensorListConcat':
            case 'TensorListConcatV2': {
                const concatId = getParamValue('tensorListId', node, tensorMap, context);
                const tensorList = context.getTensorList(concatId.id);
                const concatDtype = getParamValue('dtype', node, tensorMap, context);
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                return [tensorList.concat(concatDtype, elementShape)];
            }
            case 'TensorListPushBack': {
                const idTensor = getParamValue('tensorListId', node, tensorMap, context);
                const writeTensor = getParamValue('tensor', node, tensorMap, context);
                const tensorList = context.getTensorList(idTensor.id);
                tensorList.pushBack(writeTensor);
                return [tensorList.idTensor];
            }
            case 'TensorListPopBack': {
                const idTensor = getParamValue('tensorListId', node, tensorMap, context);
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                const elementDType = getParamValue('elementDType', node, tensorMap, context);
                const tensorList = context.getTensorList(idTensor.id);
                return [tensorList.popBack(elementShape, elementDType)];
            }
            case 'TensorListSplit': {
                const splitTensor = getParamValue('tensor', node, tensorMap, context);
                const elementShape = getParamValue('elementShape', node, tensorMap, context);
                const lengths = getParamValue('lengths', node, tensorMap, context);
                const tensorList = split$1(splitTensor, lengths, elementShape);
                context.addTensorList(tensorList);
                return [tensorList.idTensor];
            }
            case 'TensorListLength': {
                const idTensor = getParamValue('tensorListId', node, tensorMap, context);
                const tensorList = context.getTensorList(idTensor.id);
                return [scalar(tensorList.size(), 'int32')];
            }
            case 'TensorListResize': {
                const idTensor = getParamValue('tensorListId', node, tensorMap, context);
                const size = getParamValue('size', node, tensorMap, context);
                const srcTensorList = context.getTensorList(idTensor.id);
                const destTensorList = srcTensorList.resize(size);
                context.addTensorList(destTensorList);
                return [destTensorList.idTensor];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fusedConvAndDepthWiseParams(node, tensorMap, context) {
        const [extraOp, activationFunc] = getParamValue('fusedOps', node, tensorMap, context);
        const isBiasAdd = extraOp === 'biasadd';
        const noBiasAdd = !isBiasAdd;
        const isPrelu = activationFunc === 'prelu';
        const isBatchNorm = extraOp === 'fusedbatchnorm';
        const numArgs = getParamValue('numArgs', node, tensorMap, context);
        if (isBiasAdd) {
            if (isPrelu && numArgs !== 2) {
                throw new Error('FusedConv2d and DepthwiseConv2d with BiasAdd and Prelu ' +
                    'must have two extra arguments: bias and alpha.');
            }
            if (!isPrelu && isBiasAdd && numArgs !== 1) {
                throw new Error('FusedConv2d and DepthwiseConv2d with BiasAdd must have ' +
                    'one extra argument: bias.');
            }
        }
        if (isBatchNorm) {
            throw new Error('FusedConv2d and DepthwiseConv2d with FusedBatchNorm is not supported');
        }
        const stride = getParamValue('strides', node, tensorMap, context);
        const pad = getPadding(node, tensorMap, context);
        const dataFormat = getParamValue('dataFormat', node, tensorMap, context)
            .toUpperCase();
        const dilations = getParamValue('dilations', node, tensorMap, context);
        let [biasArg, preluArg] = getParamValue('args', node, tensorMap, context);
        if (noBiasAdd) {
            preluArg = biasArg;
            biasArg = undefined;
        }
        const leakyreluAlpha = getParamValue('leakyreluAlpha', node, tensorMap, context);
        return {
            stride,
            pad,
            dataFormat,
            dilations,
            biasArg,
            preluArg,
            activationFunc,
            leakyreluAlpha
        };
    }
    const executeOp$g = (node, tensorMap, context) => {
        switch (node.op) {
            case 'Conv1D': {
                const stride = getParamValue('stride', node, tensorMap, context);
                const pad = getParamValue('pad', node, tensorMap, context);
                const dataFormat = getParamValue('dataFormat', node, tensorMap, context)
                    .toUpperCase();
                const dilation = getParamValue('dilation', node, tensorMap, context);
                return [conv1d$1(getParamValue('x', node, tensorMap, context), getParamValue('filter', node, tensorMap, context), stride, pad, dataFormat, dilation)];
            }
            case 'Conv2D': {
                const stride = getParamValue('strides', node, tensorMap, context);
                const pad = getPadding(node, tensorMap, context);
                const dataFormat = getParamValue('dataFormat', node, tensorMap, context)
                    .toUpperCase();
                const dilations = getParamValue('dilations', node, tensorMap, context);
                return [conv2d$3(getParamValue('x', node, tensorMap, context), getParamValue('filter', node, tensorMap, context), [stride[1], stride[2]], pad, dataFormat, [dilations[1], dilations[2]])];
            }
            case '_FusedConv2D': {
                const { stride, pad, dataFormat, dilations, biasArg, preluArg, activationFunc, leakyreluAlpha } = fusedConvAndDepthWiseParams(node, tensorMap, context);
                return [conv2d$2({
                        x: getParamValue('x', node, tensorMap, context),
                        filter: getParamValue('filter', node, tensorMap, context),
                        strides: [stride[1], stride[2]],
                        pad: pad,
                        dataFormat: dataFormat,
                        dilations: [dilations[1], dilations[2]],
                        bias: biasArg,
                        activation: activationFunc,
                        preluActivationWeights: preluArg,
                        leakyreluAlpha
                    })];
            }
            case 'FusedDepthwiseConv2dNative': {
                const { stride, pad, dataFormat, dilations, biasArg, preluArg, activationFunc, leakyreluAlpha, } = fusedConvAndDepthWiseParams(node, tensorMap, context);
                return [depthwiseConv2d$2({
                        x: getParamValue('x', node, tensorMap, context),
                        filter: getParamValue('filter', node, tensorMap, context),
                        strides: [stride[1], stride[2]],
                        pad: pad,
                        dataFormat: dataFormat,
                        dilations: [dilations[1], dilations[2]],
                        bias: biasArg,
                        activation: activationFunc,
                        preluActivationWeights: preluArg,
                        leakyreluAlpha
                    })];
            }
            case 'Conv2DBackpropInput':
            case 'Conv2dTranspose': {
                const shape = getParamValue('outputShape', node, tensorMap, context);
                const stride = getParamValue('strides', node, tensorMap, context);
                const pad = getPadding(node, tensorMap, context);
                return [conv2dTranspose$1(getParamValue('x', node, tensorMap, context), getParamValue('filter', node, tensorMap, context), shape, [stride[1], stride[2]], pad)];
            }
            case 'DepthwiseConv2dNative':
            case 'DepthwiseConv2d': {
                const stride = getParamValue('strides', node, tensorMap, context);
                const pad = getPadding(node, tensorMap, context);
                const dilations = getParamValue('dilations', node, tensorMap, context);
                const dataFormat = getParamValue('dataFormat', node, tensorMap, context)
                    .toUpperCase();
                return [depthwiseConv2d$3(getParamValue('input', node, tensorMap, context), getParamValue('filter', node, tensorMap, context), [stride[1], stride[2]], pad, dataFormat, [dilations[1], dilations[2]])];
            }
            case 'Conv3D': {
                const stride = getParamValue('strides', node, tensorMap, context);
                const pad = getParamValue('pad', node, tensorMap, context);
                const dataFormat = getParamValue('dataFormat', node, tensorMap, context)
                    .toUpperCase();
                const dilations = getParamValue('dilations', node, tensorMap, context);
                return [conv3d$1(getParamValue('x', node, tensorMap, context), getParamValue('filter', node, tensorMap, context), [stride[1], stride[2], stride[3]], pad, dataFormat, [dilations[1], dilations[2], dilations[3]])];
            }
            case 'AvgPool': {
                const stride = getParamValue('strides', node, tensorMap, context);
                const pad = getParamValue('pad', node, tensorMap, context);
                const kernelSize = getParamValue('kernelSize', node, tensorMap, context);
                return [avgPool$2(getParamValue('x', node, tensorMap, context), [kernelSize[1], kernelSize[2]], [stride[1], stride[2]], pad)];
            }
            case 'MaxPool': {
                const stride = getParamValue('strides', node, tensorMap, context);
                const pad = getParamValue('pad', node, tensorMap, context);
                const kernelSize = getParamValue('kernelSize', node, tensorMap, context);
                return [maxPool$2(getParamValue('x', node, tensorMap, context), [kernelSize[1], kernelSize[2]], [stride[1], stride[2]], pad)];
            }
            case 'MaxPoolWithArgmax': {
                const stride = getParamValue('strides', node, tensorMap, context);
                const pad = getParamValue('pad', node, tensorMap, context);
                const kernelSize = getParamValue('kernelSize', node, tensorMap, context);
                const includeBatchInIndex = getParamValue('includeBatchInIndex', node, tensorMap, context);
                const { result, indexes } = maxPoolWithArgmax(getParamValue('x', node, tensorMap, context), [kernelSize[1], kernelSize[2]], [stride[1], stride[2]], pad, includeBatchInIndex);
                return [result, indexes];
            }
            case 'AvgPool3D': {
                const stride = getParamValue('strides', node, tensorMap, context);
                const pad = getParamValue('pad', node, tensorMap, context);
                const kernelSize = getParamValue('kernelSize', node, tensorMap, context);
                return [avgPool3d$1(getParamValue('x', node, tensorMap, context), [kernelSize[1], kernelSize[2], kernelSize[3]], [stride[1], stride[2], stride[3]], pad)];
            }
            case 'MaxPool3D': {
                const stride = getParamValue('strides', node, tensorMap, context);
                const pad = getParamValue('pad', node, tensorMap, context);
                const kernelSize = getParamValue('kernelSize', node, tensorMap, context);
                return [maxPool3d$1(getParamValue('x', node, tensorMap, context), [kernelSize[1], kernelSize[2], kernelSize[3]], [stride[1], stride[2], stride[3]], pad)];
            }
            case 'Dilation2D': {
                const strides = getParamValue('strides', node, tensorMap, context);
                const pad = getParamValue('pad', node, tensorMap, context);
                const dilations = getParamValue('dilations', node, tensorMap, context);
                // strides: [1, stride_height, stride_width, 1].
                const strideHeight = strides[1];
                const strideWidth = strides[2];
                // dilations: [1, dilation_height, dilation_width, 1].
                const dilationHeight = dilations[1];
                const dilationWidth = dilations[2];
                return [dilation2d(getParamValue('x', node, tensorMap, context), getParamValue('filter', node, tensorMap, context), [strideHeight, strideWidth], pad, [dilationHeight, dilationWidth], 'NHWC' /* dataFormat */)];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$f = (node, tensorMap, context) => {
        switch (node.op) {
            case 'Fill': {
                const shape = getParamValue('shape', node, tensorMap, context);
                const dtype = getParamValue('dtype', node, tensorMap, context);
                const value = getParamValue('value', node, tensorMap, context);
                return [fill$2(shape, value, dtype)];
            }
            case 'LinSpace': {
                const start = getParamValue('start', node, tensorMap, context);
                const stop = getParamValue('stop', node, tensorMap, context);
                const num = getParamValue('num', node, tensorMap, context);
                return [linspace(start, stop, num)];
            }
            case 'Multinomial': {
                const logits = getParamValue('logits', node, tensorMap, context);
                const numSamples = getParamValue('numSamples', node, tensorMap, context);
                const seed = getParamValue('seed', node, tensorMap, context);
                return [multinomial$2(logits, numSamples, seed)];
            }
            case 'OneHot': {
                const indices = getParamValue('indices', node, tensorMap, context);
                const depth = getParamValue('depth', node, tensorMap, context);
                const onValue = getParamValue('onValue', node, tensorMap, context);
                const offValue = getParamValue('offValue', node, tensorMap, context);
                return [oneHot$2(indices, depth, onValue, offValue)];
            }
            case 'Ones': {
                return [ones$1(getParamValue('shape', node, tensorMap, context), getParamValue('dtype', node, tensorMap, context))];
            }
            case 'OnesLike': {
                return [onesLike$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'RandomUniform': {
                return [randomUniform$1(
                    // tslint:disable-next-line:no-any
                    getParamValue('shape', node, tensorMap, context), getParamValue('minval', node, tensorMap, context), getParamValue('maxval', node, tensorMap, context), getParamValue('dtype', node, tensorMap, context))];
            }
            case 'Range': {
                const start = getParamValue('start', node, tensorMap, context);
                const stop = getParamValue('stop', node, tensorMap, context);
                const step = getParamValue('step', node, tensorMap, context);
                return [range$4(start, stop, step, getParamValue('dtype', node, tensorMap, context))];
            }
            case 'TruncatedNormal': {
                const shape = getParamValue('shape', node, tensorMap, context);
                const mean = getParamValue('mean', node, tensorMap, context);
                const stdDev = getParamValue('stdDev', node, tensorMap, context);
                const seed = getParamValue('seed', node, tensorMap, context);
                return [truncatedNormal$1(shape, mean, stdDev, getParamValue('dtype', node, tensorMap, context), seed)];
            }
            case 'Zeros': {
                return [zeros$2(getParamValue('shape', node, tensorMap, context), getParamValue('dtype', node, tensorMap, context))];
            }
            case 'ZerosLike': {
                return [zerosLike$2(getParamValue('x', node, tensorMap, context))];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function nmsParams(node, tensorMap, context) {
        const boxes = getParamValue('boxes', node, tensorMap, context);
        const scores = getParamValue('scores', node, tensorMap, context);
        const maxOutputSize = getParamValue('maxOutputSize', node, tensorMap, context);
        const iouThreshold = getParamValue('iouThreshold', node, tensorMap, context);
        const scoreThreshold = getParamValue('scoreThreshold', node, tensorMap, context);
        const softNmsSigma = getParamValue('softNmsSigma', node, tensorMap, context);
        return {
            boxes,
            scores,
            maxOutputSize,
            iouThreshold,
            scoreThreshold,
            softNmsSigma
        };
    }
    const executeOp$e = async (node, tensorMap, context) => {
        switch (node.op) {
            case 'NonMaxSuppressionV5': {
                const { boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma } = nmsParams(node, tensorMap, context);
                const result = await image$1.nonMaxSuppressionWithScoreAsync(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma);
                return [result.selectedIndices, result.selectedScores];
            }
            case 'NonMaxSuppressionV4': {
                const { boxes, scores, maxOutputSize, iouThreshold, scoreThreshold } = nmsParams(node, tensorMap, context);
                const padToMaxOutputSize = getParamValue('padToMaxOutputSize', node, tensorMap, context);
                const result = await image$1.nonMaxSuppressionPaddedAsync(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold, padToMaxOutputSize);
                return [result.selectedIndices, result.validOutputs];
            }
            case 'NonMaxSuppressionV3':
            case 'NonMaxSuppressionV2': {
                const { boxes, scores, maxOutputSize, iouThreshold, scoreThreshold } = nmsParams(node, tensorMap, context);
                return [await image$1.nonMaxSuppressionAsync(boxes, scores, maxOutputSize, iouThreshold, scoreThreshold)];
            }
            case 'Where': {
                const condition = cast$3(getParamValue('condition', node, tensorMap, context), 'bool');
                const result = [await whereAsync(condition)];
                condition.dispose();
                return result;
            }
            case 'ListDiff': {
                return setdiff1dAsync(getParamValue('x', node, tensorMap, context), getParamValue('y', node, tensorMap, context));
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$d = (node, tensorMap, context) => {
        switch (node.op) {
            case 'LowerBound': {
                const sortedSequence = getParamValue('sortedSequence', node, tensorMap, context);
                const values = getParamValue('values', node, tensorMap, context);
                return [lowerBound$1(sortedSequence, values)];
            }
            case 'TopKV2': {
                const x = getParamValue('x', node, tensorMap, context);
                const k = getParamValue('k', node, tensorMap, context);
                const sorted = getParamValue('sorted', node, tensorMap, context);
                const result = topk(x, k, sorted);
                return [result.values, result.indices];
            }
            case 'UpperBound': {
                const sortedSequence = getParamValue('sortedSequence', node, tensorMap, context);
                const values = getParamValue('values', node, tensorMap, context);
                return [upperBound$1(sortedSequence, values)];
            }
            case 'Unique': {
                const x = getParamValue('x', node, tensorMap, context);
                const result = unique$3(x);
                return [result.values, result.indices];
            }
            case 'UniqueV2': {
                const x = getParamValue('x', node, tensorMap, context);
                const axis = getParamValue('axis', node, tensorMap, context);
                const result = unique$3(x, axis);
                return [result.values, result.indices];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$c = (node, tensorMap, context) => {
        switch (node.op) {
            case 'Const': {
                return tensorMap[node.name];
            }
            case 'PlaceholderWithDefault':
                const def = getParamValue('default', node, tensorMap, context);
                return [getTensor(node.name, tensorMap, context) || def];
            case 'Placeholder':
                return [getTensor(node.name, tensorMap, context)];
            case 'Identity':
            case 'StopGradient':
            case 'FakeQuantWithMinMaxVars': { // This op is currently ignored.
                const data = getParamValue('x', node, tensorMap, context);
                return [cloneTensor(data)];
            }
            case 'IdentityN':
                return getParamValue('x', node, tensorMap, context)
                    .map((t) => cloneTensor(t));
            case 'Snapshot':
                const snapshot = getParamValue('x', node, tensorMap, context);
                return [cloneTensor(snapshot)];
            case 'Shape':
                return [tensor1d(getParamValue('x', node, tensorMap, context).shape, 'int32')];
            case 'ShapeN':
                return getParamValue('x', node, tensorMap, context)
                    .map((t) => tensor1d(t.shape));
            case 'Size':
                return [scalar(getParamValue('x', node, tensorMap, context).size, 'int32')];
            case 'Rank':
                return [scalar(getParamValue('x', node, tensorMap, context).rank, 'int32')];
            case 'NoOp':
                return [scalar(1)];
            case 'Print':
                const input = getParamValue('x', node, tensorMap, context);
                const data = getParamValue('data', node, tensorMap, context);
                const message = getParamValue('message', node, tensorMap, context);
                const summarize = getParamValue('summarize', node, tensorMap, context);
                console.warn('The graph has a tf.print() operation,' +
                    'usually used for debugging, which slows down performance.');
                console.log(message);
                for (let i = 0; i < data.length; i++) {
                    console.log(Array.prototype.slice.call(data[i].dataSync())
                        .slice(0, summarize));
                }
                return [input];
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Hashtable contains a set of tensors, which can be accessed by key.
     */
    class HashTable {
        /**
         * Constructor of HashTable. Creates a hash table.
         *
         * @param keyDType `dtype` of the table keys.
         * @param valueDType `dtype` of the table values.
         */
        constructor(keyDType, valueDType) {
            this.keyDType = keyDType;
            this.valueDType = valueDType;
            this.handle = scalar(0);
            // tslint:disable-next-line: no-any
            this.tensorMap = new Map();
            keep(this.handle);
        }
        get id() {
            return this.handle.id;
        }
        /**
         * Dispose the tensors and handle and clear the hashtable.
         */
        clearAndClose() {
            this.tensorMap.forEach(value => value.dispose());
            this.tensorMap.clear();
            this.handle.dispose();
        }
        /**
         * The number of items in the hash table.
         */
        size() {
            return this.tensorMap.size;
        }
        /**
         * The number of items in the hash table as a rank-0 tensor.
         */
        tensorSize() {
            return scalar(this.size(), 'int32');
        }
        /**
         * Replaces the contents of the table with the specified keys and values.
         * @param keys Keys to store in the hashtable.
         * @param values Values to store in the hashtable.
         */
        async import(keys, values) {
            this.checkKeyAndValueTensor(keys, values);
            // We only store the primitive values of the keys, this allows lookup
            // to be O(1).
            const $keys = await keys.data();
            // Clear the hashTable before inserting new values.
            this.tensorMap.forEach(value => value.dispose());
            this.tensorMap.clear();
            return tidy(() => {
                const $values = unstack(values);
                const keysLength = $keys.length;
                const valuesLength = $values.length;
                assert$4(keysLength === valuesLength, () => `The number of elements doesn't match, keys has ` +
                    `${keysLength} elements, the values has ${valuesLength} ` +
                    `elements.`);
                for (let i = 0; i < keysLength; i++) {
                    const key = $keys[i];
                    const value = $values[i];
                    keep(value);
                    this.tensorMap.set(key, value);
                }
                return this.handle;
            });
        }
        /**
         * Looks up keys in a hash table, outputs the corresponding values.
         *
         * Performs batch lookups, for every element in the key tensor, `find`
         * stacks the corresponding value into the return tensor.
         *
         * If an element is not present in the table, the given `defaultValue` is
         * used.
         *
         * @param keys Keys to look up. Must have the same type as the keys of the
         *     table.
         * @param defaultValue The scalar `defaultValue` is the value output for keys
         *     not present in the table. It must also be of the same type as the
         *     table values.
         */
        async find(keys, defaultValue) {
            this.checkKeyAndValueTensor(keys, defaultValue);
            const $keys = await keys.data();
            return tidy(() => {
                const result = [];
                for (let i = 0; i < $keys.length; i++) {
                    const key = $keys[i];
                    const value = this.findWithDefault(key, defaultValue);
                    result.push(value);
                }
                return stack(result);
            });
        }
        // tslint:disable-next-line: no-any
        findWithDefault(key, defaultValue) {
            const result = this.tensorMap.get(key);
            return result != null ? result : defaultValue;
        }
        checkKeyAndValueTensor(key, value) {
            if (key.dtype !== this.keyDType) {
                throw new Error(`Expect key dtype ${this.keyDType}, but got ` +
                    `${key.dtype}`);
            }
            if (value.dtype !== this.valueDType) {
                throw new Error(`Expect value dtype ${this.valueDType}, but got ` +
                    `${value.dtype}`);
            }
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$b = async (node, tensorMap, context, resourceManager) => {
        switch (node.op) {
            case 'HashTable':
            case 'HashTableV2': {
                const keyDType = getParamValue('keyDType', node, tensorMap, context);
                const valueDType = getParamValue('valueDType', node, tensorMap, context);
                const hashTable = new HashTable(keyDType, valueDType);
                resourceManager.addHashTable(node.name, hashTable);
                return [hashTable.handle];
            }
            case 'LookupTableImport':
            case 'LookupTableImportV2': {
                const handle = getParamValue('tableHandle', node, tensorMap, context, resourceManager);
                const keys = getParamValue('keys', node, tensorMap, context);
                const values = getParamValue('values', node, tensorMap, context);
                const hashTable = resourceManager.getHashTableById(handle.id);
                return [await hashTable.import(keys, values)];
            }
            case 'LookupTableFind':
            case 'LookupTableFindV2': {
                const handle = getParamValue('tableHandle', node, tensorMap, context, resourceManager);
                const keys = getParamValue('keys', node, tensorMap, context);
                const defaultValue = getParamValue('defaultValue', node, tensorMap, context);
                const hashTable = resourceManager.getHashTableById(handle.id);
                return [await hashTable.find(keys, defaultValue)];
            }
            case 'LookupTableSize':
            case 'LookupTableSizeV2': {
                const handle = getParamValue('tableHandle', node, tensorMap, context, resourceManager);
                const hashTable = resourceManager.getHashTableById(handle.id);
                return [hashTable.tensorSize()];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$a = (node, tensorMap, context) => {
        switch (node.op) {
            case 'ResizeBilinear': {
                const images = getParamValue('images', node, tensorMap, context);
                const size = getParamValue('size', node, tensorMap, context);
                const alignCorners = getParamValue('alignCorners', node, tensorMap, context);
                const halfPixelCenters = getParamValue('halfPixelCenters', node, tensorMap, context);
                return [image$1.resizeBilinear(images, [size[0], size[1]], alignCorners, halfPixelCenters)];
            }
            case 'ResizeNearestNeighbor': {
                const images = getParamValue('images', node, tensorMap, context);
                const size = getParamValue('size', node, tensorMap, context);
                const alignCorners = getParamValue('alignCorners', node, tensorMap, context);
                const halfPixelCenters = getParamValue('halfPixelCenters', node, tensorMap, context);
                return [image$1.resizeNearestNeighbor(images, [size[0], size[1]], alignCorners, halfPixelCenters)];
            }
            case 'CropAndResize': {
                const image = getParamValue('image', node, tensorMap, context);
                const boxes = getParamValue('boxes', node, tensorMap, context);
                const boxInd = getParamValue('boxInd', node, tensorMap, context);
                const cropSize = getParamValue('cropSize', node, tensorMap, context);
                const method = getParamValue('method', node, tensorMap, context);
                const extrapolationValue = getParamValue('extrapolationValue', node, tensorMap, context);
                return [image$1.cropAndResize(image, boxes, boxInd, cropSize, method, extrapolationValue)];
            }
            case 'ImageProjectiveTransformV3': {
                const images = getParamValue('images', node, tensorMap, context);
                const transforms = getParamValue('transforms', node, tensorMap, context);
                const outputShape = getParamValue('outputShape', node, tensorMap, context);
                const fillValue = getParamValue('fillValue', node, tensorMap, context);
                const interpolation = getParamValue('interpolation', node, tensorMap, context);
                const fillMode = getParamValue('fillMode', node, tensorMap, context);
                return [image$1.transform(images, transforms, interpolation.toLowerCase(), fillMode.toLowerCase(), fillValue, outputShape)];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$9 = (node, tensorMap, context) => {
        switch (node.op) {
            case 'Equal': {
                return [equal$3(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'NotEqual': {
                return [notEqual$3(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'Greater': {
                return [greater$3(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'GreaterEqual': {
                return [greaterEqual$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'Less': {
                return [less$3(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'LessEqual': {
                return [lessEqual$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'LogicalAnd': {
                return [logicalAnd$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'LogicalNot': {
                return [logicalNot$2(getParamValue('a', node, tensorMap, context))];
            }
            case 'LogicalOr': {
                return [logicalOr$2(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            case 'Select':
            case 'SelectV2': {
                return [where(getParamValue('condition', node, tensorMap, context), getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context))];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$8 = (node, tensorMap, context) => {
        switch (node.op) {
            case 'BatchMatMul':
            case 'BatchMatMulV2':
            case 'MatMul':
                return [matMul$1(getParamValue('a', node, tensorMap, context), getParamValue('b', node, tensorMap, context), getParamValue('transposeA', node, tensorMap, context), getParamValue('transposeB', node, tensorMap, context))];
            case 'Einsum':
                return [einsum$2(getParamValue('equation', node, tensorMap, context), ...getParamValue('tensors', node, tensorMap, context))];
            case 'Transpose':
                return [transpose$2(getParamValue('x', node, tensorMap, context), getParamValue('perm', node, tensorMap, context))];
            case '_FusedMatMul':
                const [extraOp, activationFunc] = getParamValue('fusedOps', node, tensorMap, context);
                const isBiasAdd = extraOp === 'biasadd';
                const isPrelu = activationFunc === 'prelu';
                const numArgs = getParamValue('numArgs', node, tensorMap, context);
                const leakyreluAlpha = getParamValue('leakyreluAlpha', node, tensorMap, context);
                if (isBiasAdd) {
                    if (isPrelu && numArgs !== 2) {
                        throw new Error('Fused MatMul with BiasAdd and Prelu must have two ' +
                            'extra arguments: bias and alpha.');
                    }
                    if (!isPrelu && numArgs !== 1) {
                        throw new Error('Fused MatMul with BiasAdd must have one extra argument: bias.');
                    }
                }
                const [biasArg, preluArg] = getParamValue('args', node, tensorMap, context);
                return [matMul({
                        a: getParamValue('a', node, tensorMap, context),
                        b: getParamValue('b', node, tensorMap, context),
                        transposeA: getParamValue('transposeA', node, tensorMap, context),
                        transposeB: getParamValue('transposeB', node, tensorMap, context),
                        bias: biasArg,
                        activation: activationFunc,
                        preluActivationWeights: preluArg,
                        leakyreluAlpha
                    })];
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$7 = (node, tensorMap, context) => {
        switch (node.op) {
            case 'EuclideanNorm':
                return [euclideanNorm(getParamValue('x', node, tensorMap, context), getParamValue('axis', node, tensorMap, context), getParamValue('keepDims', node, tensorMap, context))];
            case 'FusedBatchNorm':
            case 'FusedBatchNormV2': {
                return [batchNorm$2(getParamValue('x', node, tensorMap, context), getParamValue('mean', node, tensorMap, context), getParamValue('variance', node, tensorMap, context), getParamValue('offset', node, tensorMap, context), getParamValue('scale', node, tensorMap, context), getParamValue('epsilon', node, tensorMap, context))];
            }
            case 'FusedBatchNormV3': {
                return [batchNorm$2(getParamValue('x', node, tensorMap, context), getParamValue('mean', node, tensorMap, context), getParamValue('variance', node, tensorMap, context), getParamValue('offset', node, tensorMap, context), getParamValue('scale', node, tensorMap, context), getParamValue('epsilon', node, tensorMap, context))];
            }
            case 'LRN': {
                return [localResponseNormalization(getParamValue('x', node, tensorMap, context), getParamValue('radius', node, tensorMap, context), getParamValue('bias', node, tensorMap, context), getParamValue('alpha', node, tensorMap, context), getParamValue('beta', node, tensorMap, context))];
            }
            case 'Softmax': {
                return [softmax$3(getParamValue('x', node, tensorMap, context))];
            }
            case 'LogSoftmax': {
                return [logSoftmax(getParamValue('x', node, tensorMap, context))];
            }
            case 'SparseToDense': {
                return [sparseToDense$2(getParamValue('sparseIndices', node, tensorMap, context), getParamValue('outputShape', node, tensorMap, context), getParamValue('sparseValues', node, tensorMap, context), getParamValue('defaultValue', node, tensorMap, context))];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$6 = (node, tensorMap, context) => {
        switch (node.op) {
            case 'Max': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const keepDims = getParamValue('keepDims', node, tensorMap, context);
                return [max$3(getParamValue('x', node, tensorMap, context), axis, keepDims)];
            }
            case 'Mean': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const keepDims = getParamValue('keepDims', node, tensorMap, context);
                return [mean$1(getParamValue('x', node, tensorMap, context), axis, keepDims)];
            }
            case 'Min': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const keepDims = getParamValue('keepDims', node, tensorMap, context);
                return [min$3(getParamValue('x', node, tensorMap, context), axis, keepDims)];
            }
            case 'Sum': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const keepDims = getParamValue('keepDims', node, tensorMap, context);
                return [sum$2(getParamValue('x', node, tensorMap, context), axis, keepDims)];
            }
            case 'All': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const keepDims = getParamValue('keepDims', node, tensorMap, context);
                return [all$2(getParamValue('x', node, tensorMap, context), axis, keepDims)];
            }
            case 'Any': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const keepDims = getParamValue('keepDims', node, tensorMap, context);
                return [any$2(getParamValue('x', node, tensorMap, context), axis, keepDims)];
            }
            case 'ArgMax': {
                const axis = getParamValue('axis', node, tensorMap, context);
                return [argMax$2(getParamValue('x', node, tensorMap, context), axis)];
            }
            case 'ArgMin': {
                const axis = getParamValue('axis', node, tensorMap, context);
                return [argMin$2(getParamValue('x', node, tensorMap, context), axis)];
            }
            case 'Prod': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const keepDims = getParamValue('keepDims', node, tensorMap, context);
                return [prod$2(getParamValue('x', node, tensorMap, context), axis, keepDims)];
            }
            case 'Cumprod': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const exclusive = getParamValue('exclusive', node, tensorMap, context);
                const reverse = getParamValue('reverse', node, tensorMap, context);
                return [cumprod$2(getParamValue('x', node, tensorMap, context), axis, exclusive, reverse)];
            }
            case 'Cumsum': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const exclusive = getParamValue('exclusive', node, tensorMap, context);
                const reverse = getParamValue('reverse', node, tensorMap, context);
                return [cumsum$2(getParamValue('x', node, tensorMap, context), axis, exclusive, reverse)];
            }
            case 'Bincount':
                const x = getParamValue('x', node, tensorMap, context);
                const weights = getParamValue('weights', node, tensorMap, context);
                const size = getParamValue('size', node, tensorMap, context);
                return [bincount$2(x, weights, size)];
            case 'DenseBincount': {
                const x = getParamValue('x', node, tensorMap, context);
                const weights = getParamValue('weights', node, tensorMap, context);
                const size = getParamValue('size', node, tensorMap, context);
                const binaryOutput = getParamValue('binaryOutput', node, tensorMap, context);
                return [denseBincount$2(x, weights, size, binaryOutput)];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$5 = (node, tensorMap, context) => {
        switch (node.op) {
            case 'ConcatV2':
            case 'Concat': {
                const n = getParamValue('n', node, tensorMap, context);
                const axis = getParamValue('axis', node, tensorMap, context);
                let inputs = getParamValue('tensors', node, tensorMap, context);
                inputs = inputs.slice(0, n);
                return [concat$2(inputs, axis)];
            }
            case 'Gather': {
                const input = getParamValue('x', node, tensorMap, context);
                const indices = getParamValue('indices', node, tensorMap, context);
                return [gather$1(input, cast$3(indices, 'int32'), 0)];
            }
            case 'GatherV2': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const batchDims = getParamValue('batchDims', node, tensorMap, context);
                const input = getParamValue('x', node, tensorMap, context);
                const indices = getParamValue('indices', node, tensorMap, context);
                return [gather$1(input, cast$3(indices, 'int32'), axis, batchDims)];
            }
            case 'Reverse': {
                const dims = getParamValue('dims', node, tensorMap, context);
                const axis = [];
                for (let i = 0; i < dims.length; i++) {
                    if (dims[i]) {
                        axis.push(i);
                    }
                }
                const input = getParamValue('x', node, tensorMap, context);
                return [reverse$2(input, axis)];
            }
            case 'ReverseV2': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const input = getParamValue('x', node, tensorMap, context);
                return [reverse$2(input, axis)];
            }
            case 'Slice': {
                // tslint:disable-next-line:no-any
                const begin = getParamValue('begin', node, tensorMap, context);
                // tslint:disable-next-line:no-any
                const size = getParamValue('size', node, tensorMap, context);
                return [slice$2(getParamValue('x', node, tensorMap, context), begin, size)];
            }
            case 'StridedSlice': {
                const begin = getParamValue('begin', node, tensorMap, context);
                const end = getParamValue('end', node, tensorMap, context);
                const strides = getParamValue('strides', node, tensorMap, context);
                const beginMask = getParamValue('beginMask', node, tensorMap, context);
                const endMask = getParamValue('endMask', node, tensorMap, context);
                const ellipsisMask = getParamValue('ellipsisMask', node, tensorMap, context);
                const newAxisMask = getParamValue('newAxisMask', node, tensorMap, context);
                const shrinkAxisMask = getParamValue('shrinkAxisMask', node, tensorMap, context);
                const tensor = getParamValue('x', node, tensorMap, context);
                return [stridedSlice$2(tensor, begin, end, strides, beginMask, endMask, ellipsisMask, newAxisMask, shrinkAxisMask)];
            }
            case 'Pack': {
                return tidy(() => {
                    const axis = getParamValue('axis', node, tensorMap, context);
                    const tensors = getParamValue('tensors', node, tensorMap, context);
                    // Reshape the tensors to the first tensor's shape if they don't
                    // match.
                    const shape = tensors[0].shape;
                    const squeezedShape = squeeze(tensors[0]).shape;
                    const mapped = tensors.map(tensor => {
                        const sameShape = arraysEqual(tensor.shape, shape);
                        if (!sameShape &&
                            !arraysEqual(squeeze(tensor).shape, squeezedShape)) {
                            throw new Error('the input tensors shape does not match');
                        }
                        return sameShape ? tensor : reshape$3(tensor, shape);
                    });
                    return [stack(mapped, axis)];
                });
            }
            case 'Unpack': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const tensor = getParamValue('tensor', node, tensorMap, context);
                return unstack(tensor, axis);
            }
            case 'Tile': {
                const reps = getParamValue('reps', node, tensorMap, context);
                return [tile$3(getParamValue('x', node, tensorMap, context), reps)];
            }
            case 'Split':
            case 'SplitV': {
                const axis = getParamValue('axis', node, tensorMap, context);
                const numOrSizeSplits = getParamValue('numOrSizeSplits', node, tensorMap, context);
                const tensor = getParamValue('x', node, tensorMap, context);
                return split$2(tensor, numOrSizeSplits, axis);
            }
            case 'ScatterNd': {
                const indices = getParamValue('indices', node, tensorMap, context);
                const values = getParamValue('values', node, tensorMap, context);
                const shape = getParamValue('shape', node, tensorMap, context);
                return [scatterND(indices, values, shape)];
            }
            case 'GatherNd': {
                const x = getParamValue('x', node, tensorMap, context);
                const indices = getParamValue('indices', node, tensorMap, context);
                return [gatherND(x, indices)];
            }
            case 'SparseToDense': {
                const indices = getParamValue('sparseIndices', node, tensorMap, context);
                const shape = getParamValue('outputShape', node, tensorMap, context);
                const sparseValues = getParamValue('sparseValues', node, tensorMap, context);
                const defaultValue = getParamValue('defaultValue', node, tensorMap, context);
                return [sparseToDense$2(indices, sparseValues, shape, sparseValues.dtype === defaultValue.dtype ?
                        defaultValue :
                        cast$3(defaultValue, sparseValues.dtype))];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$4 = (node, tensorMap, context) => {
        switch (node.op) {
            case 'SparseFillEmptyRows': {
                const { outputIndices, outputValues, emptyRowIndicator, reverseIndexMap } = sparse$1.sparseFillEmptyRows(getParamValue('indices', node, tensorMap, context), getParamValue('values', node, tensorMap, context), getParamValue('denseShape', node, tensorMap, context), getParamValue('defaultValue', node, tensorMap, context));
                return [
                    outputIndices, outputValues, emptyRowIndicator, reverseIndexMap
                ];
            }
            case 'SparseReshape': {
                const { outputIndices, outputShape } = sparse$1.sparseReshape(getParamValue('inputIndices', node, tensorMap, context), getParamValue('inputShape', node, tensorMap, context), getParamValue('newShape', node, tensorMap, context));
                return [outputIndices, outputShape];
            }
            case 'SparseSegmentMean': {
                const outputData = sparse$1.sparseSegmentMean(getParamValue('data', node, tensorMap, context), getParamValue('indices', node, tensorMap, context), getParamValue('segmentIds', node, tensorMap, context));
                return [outputData];
            }
            case 'SparseSegmentSum': {
                const outputData = sparse$1.sparseSegmentSum(getParamValue('data', node, tensorMap, context), getParamValue('indices', node, tensorMap, context), getParamValue('segmentIds', node, tensorMap, context));
                return [outputData];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$3 = (node, tensorMap, context) => {
        switch (node.op) {
            case 'FFT': {
                return [fft$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'IFFT': {
                return [ifft$2(getParamValue('x', node, tensorMap, context))];
            }
            case 'RFFT': {
                return [rfft(getParamValue('x', node, tensorMap, context))];
            }
            case 'IRFFT': {
                return [irfft(getParamValue('x', node, tensorMap, context))];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$2 = (node, tensorMap, context) => {
        switch (node.op) {
            case 'StringNGrams': {
                const { nGrams, nGramsSplits } = string$1.stringNGrams(getParamValue('data', node, tensorMap, context), getParamValue('dataSplits', node, tensorMap, context), getParamValue('separator', node, tensorMap, context), getParamValue('nGramWidths', node, tensorMap, context), getParamValue('leftPad', node, tensorMap, context), getParamValue('rightPad', node, tensorMap, context), getParamValue('padWidth', node, tensorMap, context), getParamValue('preserveShortSequences', node, tensorMap, context));
                return [nGrams, nGramsSplits];
            }
            case 'StringSplit': {
                const { indices, values, shape } = string$1.stringSplit(getParamValue('input', node, tensorMap, context), getParamValue('delimiter', node, tensorMap, context), getParamValue('skipEmpty', node, tensorMap, context));
                return [indices, values, shape];
            }
            case 'StringToHashBucketFast': {
                const output = string$1.stringToHashBucketFast(getParamValue('input', node, tensorMap, context), getParamValue('numBuckets', node, tensorMap, context));
                return [output];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const executeOp$1 = (node, tensorMap, context) => {
        switch (node.op) {
            case 'Cast': {
                return [cast$3(getParamValue('x', node, tensorMap, context), getParamValue('dtype', node, tensorMap, context))];
            }
            case 'ExpandDims': {
                const axis = getParamValue('axis', node, tensorMap, context);
                return [expandDims$3(getParamValue('x', node, tensorMap, context), axis)];
            }
            case 'Squeeze': {
                const axis = getParamValue('axis', node, tensorMap, context);
                return [squeeze(getParamValue('x', node, tensorMap, context), axis)];
            }
            case 'Reshape': {
                return [reshape$3(getParamValue('x', node, tensorMap, context), getParamValue('shape', node, tensorMap, context))];
            }
            case 'MirrorPad': {
                return [mirrorPad$1(getParamValue('x', node, tensorMap, context), getParamValue('padding', node, tensorMap, context), getParamValue('mode', node, tensorMap, context))];
            }
            case 'PadV2':
            case 'Pad': {
                return [pad$1(getParamValue('x', node, tensorMap, context), getParamValue('padding', node, tensorMap, context), getParamValue('constantValue', node, tensorMap, context))];
            }
            case 'SpaceToBatchND': {
                const blockShape = getParamValue('blockShape', node, tensorMap, context);
                const paddings = getParamValue('paddings', node, tensorMap, context);
                return [spaceToBatchND$2(getParamValue('x', node, tensorMap, context), blockShape, paddings)];
            }
            case 'BatchToSpaceND': {
                const blockShape = getParamValue('blockShape', node, tensorMap, context);
                const crops = getParamValue('crops', node, tensorMap, context);
                return [batchToSpaceND$2(getParamValue('x', node, tensorMap, context), blockShape, crops)];
            }
            case 'DepthToSpace': {
                const blockSize = getParamValue('blockSize', node, tensorMap, context);
                const dataFormat = getParamValue('dataFormat', node, tensorMap, context).toUpperCase();
                return [depthToSpace$2(getParamValue('x', node, tensorMap, context), blockSize, dataFormat)];
            }
            case 'BroadcastTo': {
                return [broadcastTo(getParamValue('x', node, tensorMap, context), getParamValue('shape', node, tensorMap, context))];
            }
            case 'BroadcastArgs': {
                return [broadcastArgs$2(getParamValue('s0', node, tensorMap, context), getParamValue('s1', node, tensorMap, context))];
            }
            default:
                throw TypeError(`Node type ${node.op} is not implemented`);
        }
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Executes the op defined by the node object.
     * @param node
     * @param tensorMap contains tensors for executed nodes and weights
     * @param context contains tensors and information for running the current node.
     * @param resourceManager Optional. Contains global resources of the model.
     */
    function executeOp(node, tensorMap, context, resourceManager) {
        const value = ((node, tensorMap, context) => {
            switch (node.category) {
                case 'arithmetic':
                    return tidy(() => executeOp$j(node, tensorMap, context));
                case 'basic_math':
                    return tidy(() => executeOp$i(node, tensorMap, context));
                case 'control':
                    return executeOp$h(node, tensorMap, context);
                case 'convolution':
                    return tidy(() => executeOp$g(node, tensorMap, context));
                case 'creation':
                    return tidy(() => executeOp$f(node, tensorMap, context));
                case 'dynamic':
                    return executeOp$e(node, tensorMap, context);
                case 'evaluation':
                    return tidy(() => executeOp$d(node, tensorMap, context));
                case 'image':
                    return tidy(() => executeOp$a(node, tensorMap, context));
                case 'graph':
                    return tidy(() => executeOp$c(node, tensorMap, context));
                case 'logical':
                    return tidy(() => executeOp$9(node, tensorMap, context));
                case 'matrices':
                    return tidy(() => executeOp$8(node, tensorMap, context));
                case 'normalization':
                    return tidy(() => executeOp$7(node, tensorMap, context));
                case 'reduction':
                    return tidy(() => executeOp$6(node, tensorMap, context));
                case 'slice_join':
                    return tidy(() => executeOp$5(node, tensorMap, context));
                case 'sparse':
                    return tidy(() => executeOp$4(node, tensorMap, context));
                case 'spectral':
                    return tidy(() => executeOp$3(node, tensorMap, context));
                case 'string':
                    return tidy(() => executeOp$2(node, tensorMap, context));
                case 'transformation':
                    return tidy(() => executeOp$1(node, tensorMap, context));
                case 'hash_table':
                    return executeOp$b(node, tensorMap, context, resourceManager);
                case 'custom':
                    const opMapper = getRegisteredOp(node.op);
                    if (opMapper && opMapper.customExecutor) {
                        return opMapper.customExecutor(new NodeValueImpl(node, tensorMap, context));
                    }
                    else {
                        throw TypeError(`Custom op ${node.op} is not registered.`);
                    }
                default:
                    throw TypeError(`Unknown op '${node.op}'. File an issue at ` +
                        `https://github.com/tensorflow/tfjs/issues so we can add it` +
                        `, or register a custom execution with tf.registerOp()`);
            }
        })(node, tensorMap, context);
        if (isPromise(value)) {
            return value.then((data) => [].concat(data));
        }
        return [].concat(value);
    }

    /**
     * ExecutionContext captures the runtime environment of the node. It keeps
     * track of the current frame and iteration for the control flow ops.
     *
     * For example, typical Dynamic RNN model may contain loops, for which
     * TensorFlow will generate graphs with Enter/Exit nodes to control the
     * current execution frame, and NextIteration Nodes for iteration id increment.
     * For model with branch logic, TensorFLow will generate Switch/Merge ops.
     */
    class ExecutionContext {
        constructor(weightMap = {}, tensorArrayMap = {}, tensorListMap = {}, functionMap = {}) {
            this.weightMap = weightMap;
            this.tensorArrayMap = tensorArrayMap;
            this.tensorListMap = tensorListMap;
            this.functionMap = functionMap;
            this.rootContext = { id: 0, frameName: '', iterationId: 0 };
            this.contexts = [this.rootContext];
            this.lastId = 0;
            this.generateCurrentContextIds();
        }
        newFrame(id, frameName) {
            return { id, frameName, iterationId: 0 };
        }
        /**
         * Set the current context
         * @param contexts: ExecutionContextInfo[] the current path of execution
         * frames
         */
        set currentContext(contexts) {
            if (this.contexts !== contexts) {
                this.contexts = contexts;
                this.generateCurrentContextIds();
            }
        }
        get currentContext() {
            return this.contexts;
        }
        /**
         * Returns the current context in string format.
         */
        get currentContextId() {
            return this._currentContextIds[0];
        }
        /**
         * Returns the current context and all parent contexts in string format.
         * This allow access to the nodes in the current and parent frames.
         */
        get currentContextIds() {
            return this._currentContextIds;
        }
        generateCurrentContextIds() {
            const names = [];
            for (let i = 0; i < this.contexts.length - 1; i++) {
                const contexts = this.contexts.slice(0, this.contexts.length - i);
                names.push(this.contextIdforContexts(contexts));
            }
            names.push('');
            this._currentContextIds = names;
        }
        contextIdforContexts(contexts) {
            return contexts ?
                contexts
                    .map(context => (context.id === 0 && context.iterationId === 0) ?
                    '' :
                    `${context.frameName}-${context.iterationId}`)
                    .join('/') :
                '';
        }
        /**
         * Enter a new frame, a new context is pushed on the current context list.
         * @param frameId new frame id
         */
        enterFrame(frameId) {
            if (this.contexts) {
                this.lastId++;
                this.contexts = this.contexts.slice();
                this.contexts.push(this.newFrame(this.lastId, frameId));
                this._currentContextIds.unshift(this.contextIdforContexts(this.contexts));
            }
        }
        /**
         * Exit the current frame, the last context is removed from the current
         * context list.
         */
        exitFrame() {
            if (this.contexts && this.contexts.length > 1) {
                this.contexts = this.contexts.slice();
                this.contexts.splice(-1);
                this.currentContextIds.shift();
            }
            else {
                throw new Error('Cannot exit frame, the context is empty');
            }
        }
        /**
         * Enter the next iteration of a loop, the iteration id of last context is
         * increased.
         */
        nextIteration() {
            if (this.contexts && this.contexts.length > 0) {
                this.contexts = this.contexts.slice();
                this.lastId++;
                const context = Object.assign({}, this.contexts[this.contexts.length - 1]);
                context.iterationId += 1;
                context.id = this.lastId;
                this.contexts.splice(-1, 1, context);
                this._currentContextIds.splice(0, 1, this.contextIdforContexts(this.contexts));
            }
            else {
                throw new Error('Cannot increase frame iteration, the context is empty');
            }
        }
        getWeight(name) {
            return this.weightMap[name];
        }
        addTensorArray(tensorArray) {
            this.tensorArrayMap[tensorArray.id] = tensorArray;
        }
        getTensorArray(id) {
            return this.tensorArrayMap[id];
        }
        addTensorList(tensorList) {
            this.tensorListMap[tensorList.id] = tensorList;
        }
        getTensorList(id) {
            return this.tensorListMap[id];
        }
        dispose(keepIds) {
            for (const key in this.tensorArrayMap) {
                this.tensorArrayMap[key].clearAndClose(keepIds);
            }
            for (const key in this.tensorListMap) {
                this.tensorListMap[key].clearAndClose(keepIds);
            }
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Given graph inputs and desired outputs, find the minimal set of nodes
     * to execute in order to compute the outputs. In addition return other useful
     * info such:
     * - Missing inputs needed to compute the output.
     * - Whether the subgraph contains dynamic ops (control flow, dynamic shape).
     * - Alternative inputs in order to avoid async (dynamic op) execution.
     */
    function getExecutionSubgraph(inputs, outputs, weightMap, initNodes) {
        const usedNodes = new Set();
        const missingInputs = [];
        let dynamicNode = null;
        let syncInputs = null;
        // Start with the outputs, going backwards and find all the nodes that are
        // needed to compute those outputs.
        const seen = new Set();
        const inputNodeNames = Object.keys(inputs).map(name => parseNodeName(name)[0]);
        let initNodeNames = [];
        if (initNodes != null) {
            initNodeNames = initNodes.map(node => parseNodeName(node.name)[0]);
        }
        const frontier = [...outputs];
        while (frontier.length > 0) {
            const node = frontier.pop();
            if (isControlFlow(node) || isDynamicShape(node) || isHashTable(node)) {
                if (dynamicNode == null) {
                    dynamicNode = node;
                    syncInputs = dynamicNode.children.map(child => child.name)
                        .filter(name => usedNodes.has(name));
                }
            }
            usedNodes.add(node.name);
            // Weights are dead end since we already have their values.
            if (weightMap[node.name] != null) {
                continue;
            }
            // This node is a dead end since it's one of the user-provided inputs.
            if (inputNodeNames.indexOf(node.name) !== -1) {
                continue;
            }
            // This node is a dead end since it doesn't have any inputs.
            if (initNodeNames.indexOf(node.name) !== -1) {
                continue;
            }
            if (node.inputs.length === 0) {
                missingInputs.push(node.name);
                continue;
            }
            node.inputs.forEach(input => {
                // Don't add to the frontier if it is already there.
                if (seen.has(input.name)) {
                    return;
                }
                seen.add(input.name);
                frontier.push(input);
            });
        }
        return { inputs, outputs, usedNodes, missingInputs, dynamicNode, syncInputs };
    }
    /**
     * Given the execution info, return a list of nodes in topological order that
     * need to be executed to compute the output.
     */
    function getNodesInTopologicalOrder(graph, weightMap, executionInfo) {
        const { usedNodes, inputs } = executionInfo;
        const frontier = [];
        const inputNodes = Object.keys(inputs)
            .map(name => parseNodeName(name)[0])
            .map(name => graph.nodes[name]);
        const initNodes = graph.initNodes;
        inputNodes.forEach(input => {
            if (usedNodes.has(input.name)) {
                frontier.push(input);
            }
        });
        graph.weights.forEach(weight => {
            if (usedNodes.has(weight.name)) {
                frontier.push(weight);
            }
        });
        if (initNodes != null) {
            initNodes.forEach(node => {
                if (usedNodes.has(node.name)) {
                    frontier.push(node);
                }
            });
        }
        const seen = new Set();
        const orderedNodes = [];
        while (frontier.length > 0) {
            const node = frontier.pop();
            seen.add(node.name);
            if (!weightMap[node.name]) {
                orderedNodes.push(node);
            }
            node.children.forEach(child => {
                if (!seen.has(child.name) && usedNodes.has(child.name) &&
                    child.inputs.every(input => seen.has(input.name))) {
                    frontier.push(child);
                }
            });
        }
        return orderedNodes;
    }
    const CONTROL_FLOW_OPS = [
        'Switch', 'Merge', 'Enter', 'Exit', 'NextIteration', 'StatelessIf',
        'StatelessWhile', 'if', 'While'
    ];
    const DYNAMIC_SHAPE_OPS = [
        'NonMaxSuppressionV2', 'NonMaxSuppressionV3', 'NonMaxSuppressionV5', 'Where'
    ];
    const HASH_TABLE_OPS = [
        'HashTable', 'HashTableV2', 'LookupTableImport', 'LookupTableImportV2',
        'LookupTableFind', 'LookupTableFindV2', 'LookupTableSize', 'LookupTableSizeV2'
    ];
    function isControlFlow(node) {
        return CONTROL_FLOW_OPS.indexOf(node.op) >= 0;
    }
    function isDynamicShape(node) {
        return DYNAMIC_SHAPE_OPS.indexOf(node.op) >= 0;
    }
    function isHashTable(node) {
        return HASH_TABLE_OPS.indexOf(node.op) >= 0;
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class GraphExecutor {
        /**
         *
         * @param graph Graph the model or function graph to be executed.
         * @param parent When building function exector you need to set the parent
         * executor. Since the weights and function executor maps are set at parant
         * level, that function executor can access the function maps and weight maps
         * through the parent.
         */
        constructor(graph, parent) {
            this.graph = graph;
            this.parent = parent;
            this.compiledMap = new Map();
            this._weightMap = {};
            this.SEPERATOR = ',';
            this._functions = {};
            this._functionExecutorMap = {};
            this.intermediateTensors = {};
            this.keepTensorForDebug = false;
            this._outputs = graph.outputs;
            this._inputs = graph.inputs;
            this._initNodes = graph.initNodes;
            this._signature = graph.signature;
            this._functions = graph.functions;
            // create sub-graph executors
            if (graph.functions != null) {
                Object.keys(graph.functions).forEach(name => {
                    this._functionExecutorMap[name] =
                        new GraphExecutor(graph.functions[name], this);
                });
            }
        }
        get weightIds() {
            return this.parent ? this.parent.weightIds : this._weightIds;
        }
        get functionExecutorMap() {
            return this.parent ? this.parent.functionExecutorMap :
                this._functionExecutorMap;
        }
        get weightMap() {
            return this.parent ? this.parent.weightMap : this._weightMap;
        }
        set weightMap(weightMap) {
            const weightIds = Object.keys(weightMap).map(key => weightMap[key].map(tensor => tensor.id));
            this._weightIds = [].concat(...weightIds);
            this._weightMap = weightMap;
        }
        /**
         * Set `ResourceManager` shared by executors of a model.
         * @param resourceManager: `ResourceManager` of the `GraphModel`.
         */
        set resourceManager(resourceManager) {
            this._resourceManager = resourceManager;
        }
        get inputs() {
            return this._inputs.map(node => {
                return {
                    name: node.name,
                    shape: node.attrParams['shape'] ?
                        node.attrParams['shape'].value :
                        undefined,
                    dtype: node.attrParams['dtype'] ?
                        node.attrParams['dtype'].value :
                        undefined
                };
            });
        }
        get outputs() {
            return this._outputs.map(node => {
                return {
                    name: node.name,
                    shape: node.attrParams['shape'] ?
                        node.attrParams['shape'].value :
                        undefined,
                    dtype: node.attrParams['dtype'] ?
                        node.attrParams['dtype'].value :
                        undefined
                };
            });
        }
        get inputNodes() {
            return this._inputs.map(node => node.signatureKey || node.name);
        }
        get outputNodes() {
            return this._outputs.map((node) => {
                const name = node.signatureKey || node.name;
                return node.defaultOutput ? (`${name}:${node.defaultOutput}`) : name;
            });
        }
        get functions() {
            return Object.keys(this._functions).reduce((map, key) => {
                map[key] = this._functions[key].signature;
                return map;
            }, {});
        }
        getCompilationKey(inputs, outputs) {
            const sortedInputs = inputs.map(node => node.name).sort();
            const sortedOutputs = outputs.map(node => node.name).sort();
            return sortedInputs.join(this.SEPERATOR) + '--' +
                sortedOutputs.join(this.SEPERATOR);
        }
        /**
         * Compiles the inference graph and returns the minimal set of nodes that are
         * required for execution, in the correct execution order.
         */
        compile(inputs, outputs) {
            const executionInfo = getExecutionSubgraph(inputs, outputs, this.weightMap, this._initNodes);
            const { missingInputs, dynamicNode, syncInputs } = executionInfo;
            if (dynamicNode != null) {
                throw new Error(`This execution contains the node '${dynamicNode.name}', which has ` +
                    `the dynamic op '${dynamicNode.op}'. Please use ` +
                    `model.executeAsync() instead. Alternatively, to avoid the ` +
                    `dynamic ops, specify the inputs [${syncInputs}]`);
            }
            if (missingInputs.length > 0) {
                const outNames = outputs.map(n => n.name);
                const inNames = Object.keys(inputs);
                throw new Error(`Cannot compute the outputs [${outNames}] from the provided inputs ` +
                    `[${inNames}]. Missing the following inputs: [${missingInputs}]`);
            }
            return getNodesInTopologicalOrder(this.graph, this.weightMap, executionInfo);
        }
        /**
         * Executes the inference for given input tensors.
         * @param inputs Tensor map for the model inputs, keyed by the input node
         * names.
         * @param outputs Optional. output node name from the Tensorflow model, if
         * no outputs are specified, the default outputs of the model would be used.
         * You can inspect intermediate nodes of the model by adding them to the
         * outputs array.
         */
        execute(inputs, outputs) {
            inputs = this.mapInputs(inputs);
            const names = Object.keys(inputs).sort();
            this.checkInputs(inputs);
            this.checkInputShapeAndType(inputs);
            outputs = this.mapOutputs(outputs);
            this.checkOutputs(outputs);
            const inputNodes = names.map(name => this.graph.nodes[parseNodeName(name)[0]]);
            const outputNodeNames = outputs.map(name => parseNodeName(name)[0]);
            let outputNodes = outputNodeNames.map(name => this.graph.nodes[name]);
            this.resetIntermediateTensors();
            // If no outputs are specified, then use the default outputs of the model.
            if (outputNodes.length === 0) {
                outputNodes = this._outputs;
            }
            const compilationKey = this.getCompilationKey(inputNodes, outputNodes);
            // Do nothing if the compiled graph cache contains the input.
            let orderedNodes = this.compiledMap.get(compilationKey);
            if (orderedNodes == null) {
                orderedNodes = this.compile(inputs, outputNodes);
                this.compiledMap.set(compilationKey, orderedNodes);
            }
            const tensorArrayMap = {};
            const tensorListMap = {};
            return tidy(() => {
                const context = new ExecutionContext(this.weightMap, tensorArrayMap, tensorListMap, this.functionExecutorMap);
                const tensorsMap = Object.assign({}, this.weightMap);
                Object.keys(inputs).forEach(name => {
                    const [nodeName, index] = parseNodeName(name);
                    const tensors = [];
                    tensors[index] = inputs[name];
                    tensorsMap[nodeName] = tensors;
                });
                const tensorsToKeep = this.getFrozenTensorIds(tensorsMap);
                const intermediateTensorConsumerCount = {};
                for (let i = 0; i < orderedNodes.length; i++) {
                    const node = orderedNodes[i];
                    if (!tensorsMap[node.name]) {
                        const tensors = executeOp(node, tensorsMap, context, this._resourceManager);
                        if (isPromise(tensors)) {
                            throw new Error(`The execution of the op '${node.op}' returned a promise. ` +
                                `Please use model.executeAsync() instead.`);
                        }
                        tensorsMap[node.name] = tensors;
                        this.checkTensorForDisposal(node.name, node, tensorsMap, context, tensorsToKeep, outputNodeNames, intermediateTensorConsumerCount);
                    }
                }
                // dispose the context for the root executor
                if (this.parent == null) {
                    context.dispose(tensorsToKeep);
                }
                return outputs.map(name => getTensor(name, tensorsMap, context));
            });
        }
        getFrozenTensorIds(tensorMap) {
            const ids = [].concat.apply([], Object.keys(tensorMap)
                .map(key => tensorMap[key])
                .map(tensors => tensors.map(tensor => tensor.id)));
            return new Set(ids);
        }
        checkTensorForDisposal(nodeName, node, tensorMap, context, tensorsToKeep, outputNames, intermediateTensorConsumerCount) {
            // Skip output nodes and any control flow nodes, since its dependency is
            // tricky to track correctly.
            if (node.category === 'control' || outputNames.indexOf(nodeName) !== -1) {
                return;
            }
            tensorMap[nodeName].forEach(tensor => {
                if (tensor != null) {
                    intermediateTensorConsumerCount[tensor.id] =
                        (intermediateTensorConsumerCount[tensor.id] || 0) +
                            node.children.length;
                }
            });
            node.inputs.forEach(input => {
                // Skip any control flow nodes, since its dependency is tricky to track
                // correctly.
                if (input.category !== 'control') {
                    const tensors = getTensorsForCurrentContenxt(input.name, tensorMap, context);
                    if (tensors != null) {
                        tensors.forEach(tensor => {
                            if (tensor && !tensor.kept && !tensorsToKeep.has(tensor.id)) {
                                const count = intermediateTensorConsumerCount[tensor.id];
                                if (count === 1) {
                                    if (!this.keepTensorForDebug) {
                                        tensor.dispose();
                                    }
                                    else {
                                        const [nodeName, index] = getNodeNameAndIndex(node.name, context);
                                        if (this.intermediateTensors[nodeName]) {
                                            this.intermediateTensors[nodeName][index] = tensor;
                                        }
                                        else {
                                            this.intermediateTensors[nodeName] = [];
                                            this.intermediateTensors[nodeName][index] = tensor;
                                        }
                                    }
                                    delete intermediateTensorConsumerCount[tensor.id];
                                }
                                else if (count != null) {
                                    // only intermediate nodes has count set, inputs and weights are
                                    // not.
                                    intermediateTensorConsumerCount[tensor.id]--;
                                }
                            }
                        });
                    }
                }
            });
        }
        /**
         * Executes the inference for given input tensors in Async fashion.
         * @param inputs Tensor map for the model inputs, keyed by the input node
         * names.
         * @param outputs output node name from the Tensorflow model, if no outputs
         * are specified, the default outputs of the model would be used. You can
         * inspect intermediate nodes of the model by adding them to the outputs
         * array.
         */
        async executeAsync(inputs, outputs) {
            return this._executeAsync(inputs, outputs);
        }
        disposeIntermediateTensors() {
            if (!this.intermediateTensors) {
                return;
            }
            Object.keys(this.intermediateTensors)
                .forEach(key => this.intermediateTensors[key].forEach(tensor => tensor.dispose()));
            this.disposeTensorsMap();
        }
        disposeTensorsMap() {
            if (!this.tensorsMap) {
                return;
            }
            Object.keys(this.tensorsMap).forEach(key => {
                const tensorArray = this.tensorsMap[key];
                tensorArray.forEach(tensor => {
                    if (tensor && !tensor.kept && !tensor.isDisposed &&
                        !this.keepIds.has(tensor.id)) {
                        tensor.dispose();
                    }
                });
            });
        }
        getIntermediateTensors() {
            return this.tensorsMap;
        }
        resetIntermediateTensors() {
            for (const key in this.intermediateTensors) {
                this.intermediateTensors[key].forEach(tensor => tensor.dispose());
                delete this.intermediateTensors[key];
            }
        }
        /**
         * Executes the inference for given input tensors in Async fashion.
         * @param inputs Tensor map for the model inputs, keyed by the input node
         * names.
         * @param outputs Optional. output node name from the Tensorflow model,
         * if no outputs are specified, the default outputs of the model would be
         * used. You can inspect intermediate nodes of the model by adding them to the
         * outputs array.
         * @param isFunctionExecution Optional. Flag for executing a function.
         * @param tensorArrayMap Optional, global TensorArray map by id. Used for
         * function execution.
         * @param tensorArrayMap Optinal global TensorList map by id. Used for
         * function execution.
         */
        async _executeAsync(inputs, outputs, isFunctionExecution = false, tensorArrayMap = {}, tensorListMap = {}) {
            if (!isFunctionExecution) {
                inputs = this.mapInputs(inputs);
                this.checkInputs(inputs);
                this.checkInputShapeAndType(inputs);
                outputs = this.mapOutputs(outputs);
                this.checkOutputs(outputs);
            }
            // For model debug.
            try {
                this.keepTensorForDebug = env$1().getBool('KEEP_INTERMEDIATE_TENSORS');
            }
            catch (e) {
                console.warn(e.message);
            }
            this.resetIntermediateTensors();
            const context = new ExecutionContext(this.weightMap, tensorArrayMap, tensorListMap, this.functionExecutorMap);
            // Graph with control flow op requires runtime evaluation of the execution
            // order, while without control flow the execution order is pre-determined
            // in the compile method.
            this.tensorsMap = await this.executeWithControlFlow(inputs, context, outputs, isFunctionExecution);
            const results = outputs.map(name => getTensor(name, this.tensorsMap, context));
            // dispose all the intermediate tensors
            const outputIds = results.map(t => t.id);
            const inputIds = Object.keys(inputs).map(name => inputs[name].id);
            this.keepIds =
                new Set([...outputIds, ...inputIds, ...this.weightIds]);
            if (!this.keepTensorForDebug) {
                this.disposeTensorsMap();
            }
            // dispose the context for the root executor
            if (this.parent == null) {
                context.dispose(this.keepIds);
            }
            return results;
        }
        async executeFunctionAsync(inputs, tensorArrayMap, tensorListMap) {
            const mappedInputs = inputs.reduce((map, tensor, index) => {
                map[this.inputs[index].name] = tensor;
                return map;
            }, {});
            return this._executeAsync(mappedInputs, this.outputNodes, true, tensorArrayMap, tensorListMap);
        }
        /**
         * When there are control flow nodes in the graph, the graph execution use
         * ExecutionContext to keep track of the frames and loop iterators.
         * @param inputs placeholder tensors for the graph.
         * @param context the execution context object for current execution.
         * @param outputNames Optional. output node name from the Tensorflow model,
         * if no outputs are specified, the default outputs of the model would be
         * used. You can inspect intermediate nodes of the model by adding them to the
         * outputs array.
         * @param isFunctionExecution Flag for executing a function.
         */
        async executeWithControlFlow(inputs, context, outputNames, isFunctionExecution) {
            const names = Object.keys(inputs);
            const inputNodes = names.map(name => this.graph.nodes[parseNodeName(name)[0]]);
            const outputNodeNames = outputNames.map(name => parseNodeName(name)[0]);
            let outputNodes = outputNodeNames.map(name => this.graph.nodes[name]);
            // If no outputs are specified, then use the default outputs of the model.
            if (outputNodes.length === 0) {
                outputNodes = this._outputs;
            }
            const { usedNodes, missingInputs, dynamicNode, syncInputs } = getExecutionSubgraph(inputs, outputNodes, this.weightMap, this._initNodes);
            // First nodes to execute include inputNodes, weights, and initNodes.
            const stack = [
                ...inputNodes, ...this.graph.weights, ...(this._initNodes || [])
            ].map(node => {
                return { node, contexts: context.currentContext };
            });
            const tensorsMap = Object.assign({}, this.weightMap);
            Object.keys(inputs).forEach(name => {
                const [nodeName, index] = parseNodeName(name);
                const tensors = [];
                tensors[index] = inputs[name];
                tensorsMap[nodeName] = tensors;
            });
            const intermediateTensorConsumerCount = {};
            const tensorsToKeep = this.getFrozenTensorIds(tensorsMap);
            const added = {};
            while (stack.length > 0) {
                const promises = this.processStack(inputNodes, stack, context, tensorsMap, added, tensorsToKeep, outputNodeNames, intermediateTensorConsumerCount, usedNodes);
                await Promise.all(promises);
            }
            if (dynamicNode == null && !isFunctionExecution) {
                console.warn(`This model execution did not contain any nodes with control flow ` +
                    `or dynamic output shapes. You can use model.execute() instead.`);
            }
            const missingOutputs = outputNodes
                .filter(node => !isControlFlow(node) &&
                !getTensor(node.name, tensorsMap, context))
                .map(node => node.name);
            if (missingOutputs.length > 0) {
                let alternativeMsg = '';
                if (dynamicNode != null) {
                    alternativeMsg =
                        `Alternatively, to avoid the dynamic ops, use model.execute() ` +
                            `and specify the inputs [${syncInputs}]`;
                }
                throw new Error(`Cannot compute the outputs [${missingOutputs}] from the provided ` +
                    `inputs [${names}]. Consider providing the following inputs: ` +
                    `[${missingInputs}]. ${alternativeMsg}`);
            }
            return tensorsMap;
        }
        processStack(inputNodes, stack, context, tensorMap, added, tensorsToKeep, outputNames, intermediateTensorConsumerCount, usedNodes) {
            const promises = [];
            while (stack.length > 0) {
                const item = stack.pop();
                context.currentContext = item.contexts;
                let nodeName = '';
                // The tensor of the Enter op with isConstant set should be set
                // in the parent scope, so it will be available as constant for the
                // whole loop.
                if (item.node.op === 'Enter' &&
                    getParamValue('isConstant', item.node, tensorMap, context)) {
                    [nodeName] = getNodeNameAndIndex(item.node.name, context);
                }
                // only process nodes that are not in the tensorMap yet, this include
                // inputNodes and internal initNodes.
                if (tensorMap[item.node.name] == null) {
                    const tensors = executeOp(item.node, tensorMap, context, this._resourceManager);
                    if (!nodeName) {
                        [nodeName] = getNodeNameAndIndex(item.node.name, context);
                    }
                    const currentContext = context.currentContext;
                    if (isPromise(tensors)) {
                        promises.push(tensors.then(t => {
                            tensorMap[nodeName] = t;
                            context.currentContext = currentContext;
                            this.checkTensorForDisposal(nodeName, item.node, tensorMap, context, tensorsToKeep, outputNames, intermediateTensorConsumerCount);
                            this.processChildNodes(item.node, stack, context, tensorMap, added, usedNodes);
                            return t;
                        }));
                    }
                    else {
                        tensorMap[nodeName] = tensors;
                        this.checkTensorForDisposal(nodeName, item.node, tensorMap, context, tensorsToKeep, outputNames, intermediateTensorConsumerCount);
                        this.processChildNodes(item.node, stack, context, tensorMap, added, usedNodes);
                    }
                }
                else {
                    this.processChildNodes(item.node, stack, context, tensorMap, added, usedNodes);
                }
            }
            return promises;
        }
        processChildNodes(node, stack, context, tensorMap, added, usedNodes) {
            node.children.forEach((childNode) => {
                const [nodeName,] = getNodeNameAndIndex(childNode.name, context);
                if (added[nodeName] || !usedNodes.has(childNode.name)) {
                    return;
                }
                // Merge op can be pushed if any of its inputs has value.
                if (childNode.op === 'Merge') {
                    if (childNode.inputNames.some(name => {
                        return !!getTensor(name, tensorMap, context);
                    })) {
                        added[nodeName] = true;
                        stack.push({ contexts: context.currentContext, node: childNode });
                    }
                }
                else // Otherwise all inputs must to have value.
                 if (childNode.inputNames.every(name => {
                    return !!getTensor(name, tensorMap, context);
                })) {
                    added[nodeName] = true;
                    stack.push({ contexts: context.currentContext, node: childNode });
                }
            });
        }
        /**
         * Releases the memory used by the weight tensors.
         */
        dispose() {
            Object.keys(this.weightMap)
                .forEach(key => this.weightMap[key].forEach(tensor => tensor.dispose()));
        }
        checkInputShapeAndType(inputs) {
            Object.keys(inputs).forEach(name => {
                const input = inputs[name];
                const [nodeName,] = parseNodeName(name);
                const node = this.graph.nodes[nodeName];
                if (node.attrParams['shape'] && node.attrParams['shape'].value) {
                    const shape = node.attrParams['shape'].value;
                    const match = shape.length === input.shape.length &&
                        input.shape.every((dim, index) => shape[index] === -1 || shape[index] === dim);
                    assert$4(match, () => `The shape of dict['${node.name}'] provided in ` +
                        `model.execute(dict) must be [${shape}], but was ` +
                        `[${input.shape}]`);
                }
                if (node.attrParams['dtype'] && node.attrParams['dtype'].value) {
                    assert$4(input.dtype === node.attrParams['dtype'].value, () => `The dtype of dict['${node.name}'] provided in ` +
                        `model.execute(dict) must be ` +
                        `${node.attrParams['dtype'].value}, but was ${input.dtype}`);
                }
            });
        }
        mapInputs(inputs) {
            const result = {};
            for (const inputName in inputs) {
                if (this._signature != null && this._signature.inputs != null &&
                    this._signature.inputs[inputName] != null) {
                    const tensor = this._signature.inputs[inputName];
                    result[tensor.name] = inputs[inputName];
                }
                else {
                    result[inputName] = inputs[inputName];
                }
            }
            return result;
        }
        checkInputs(inputs) {
            const notInGraph = Object.keys(inputs).filter(name => {
                const [nodeName] = parseNodeName(name);
                return this.graph.nodes[nodeName] == null;
            });
            if (notInGraph.length > 0) {
                throw new Error(`The dict provided in model.execute(dict) has ` +
                    `keys: [${notInGraph}] that are not part of graph`);
            }
        }
        mapOutputs(outputs) {
            return outputs.map(name => {
                if (this._signature != null && this._signature.outputs != null &&
                    this._signature.outputs[name] != null) {
                    const tensor = this._signature.outputs[name];
                    return tensor.name;
                }
                return name;
            }, {});
        }
        checkOutputs(outputs) {
            outputs.forEach(name => {
                const [normalizedName] = parseNodeName(name);
                if (!this.graph.nodes[normalizedName]) {
                    throw new Error(`The output '${name}' is not found in the graph`);
                }
            });
        }
    }

    /**
     * Contains global resources of a model.
     */
    class ResourceManager {
        constructor(hashTableNameToHandle = {}, hashTableMap = {}) {
            this.hashTableNameToHandle = hashTableNameToHandle;
            this.hashTableMap = hashTableMap;
        }
        /**
         * Register a `HashTable` in the resource manager.
         *
         * The `HashTable` can be retrieved by `resourceManager.getHashTableById`,
         * where id is the table handle tensor's id.
         *
         * @param name Op node name that creates the `HashTable`.
         * @param hashTable The `HashTable` to be added to resource manager.
         */
        addHashTable(name, hashTable) {
            this.hashTableNameToHandle[name] = hashTable.handle;
            this.hashTableMap[hashTable.id] = hashTable;
        }
        /**
         * Get the table handle by node name.
         * @param name Op node name that creates the `HashTable`. This name is also
         *     used in the inputs list of lookup and import `HashTable` ops.
         */
        getHashTableHandleByName(name) {
            return this.hashTableNameToHandle[name];
        }
        /**
         * Get the actual `HashTable` by its handle tensor's id.
         * @param id The id of the handle tensor.
         */
        getHashTableById(id) {
            return this.hashTableMap[id];
        }
        /**
         * Dispose `ResourceManager`, including its hashTables and tensors in them.
         */
        dispose() {
            for (const key in this.hashTableMap) {
                this.hashTableMap[key].clearAndClose();
                delete this.hashTableMap[key];
            }
            for (const name in this.hashTableNameToHandle) {
                this.hashTableNameToHandle[name].dispose();
                delete this.hashTableNameToHandle[name];
            }
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const TFHUB_SEARCH_PARAM = '?tfjs-format=file';
    const DEFAULT_MODEL_NAME = 'model.json';
    /**
     * A `tf.GraphModel` is a directed, acyclic graph built from a
     * SavedModel GraphDef and allows inference execution.
     *
     * A `tf.GraphModel` can only be created by loading from a model converted from
     * a [TensorFlow SavedModel](https://www.tensorflow.org/guide/saved_model) using
     * the command line converter tool and loaded via `tf.loadGraphModel`.
     *
     * @doc {heading: 'Models', subheading: 'Classes'}
     */
    class GraphModel {
        /**
         * @param modelUrl url for the model, or an `io.IOHandler`.
         * @param weightManifestUrl url for the weight file generated by
         * scripts/convert.py script.
         * @param requestOption options for Request, which allows to send credentials
         * and custom headers.
         * @param onProgress Optional, progress callback function, fired periodically
         * before the load is completed.
         */
        constructor(modelUrl, loadOptions = {}) {
            this.modelUrl = modelUrl;
            this.loadOptions = loadOptions;
            this.version = 'n/a';
            if (loadOptions == null) {
                this.loadOptions = {};
            }
            this.resourceManager = new ResourceManager();
        }
        // Returns the version information for the tensorflow model GraphDef.
        get modelVersion() {
            return this.version;
        }
        get inputNodes() {
            return this.executor.inputNodes;
        }
        get outputNodes() {
            return this.executor.outputNodes;
        }
        get inputs() {
            return this.executor.inputs;
        }
        get outputs() {
            return this.executor.outputs;
        }
        get weights() {
            return this.executor.weightMap;
        }
        get metadata() {
            return this.artifacts.userDefinedMetadata;
        }
        get modelSignature() {
            return this.signature;
        }
        findIOHandler() {
            const path = this.modelUrl;
            if (path.load != null) {
                // Path is an IO Handler.
                this.handler = path;
            }
            else if (this.loadOptions.requestInit != null) {
                this.handler = browserHTTPRequest(path, this.loadOptions);
            }
            else {
                const handlers = getLoadHandlers(path, this.loadOptions);
                if (handlers.length === 0) {
                    // For backward compatibility: if no load handler can be found,
                    // assume it is a relative http path.
                    handlers.push(browserHTTPRequest(path, this.loadOptions));
                }
                else if (handlers.length > 1) {
                    throw new Error(`Found more than one (${handlers.length}) load handlers for ` +
                        `URL '${[path]}'`);
                }
                this.handler = handlers[0];
            }
        }
        /**
         * Loads the model and weight files, construct the in memory weight map and
         * compile the inference graph.
         */
        async load() {
            this.findIOHandler();
            if (this.handler.load == null) {
                throw new Error('Cannot proceed with model loading because the IOHandler provided ' +
                    'does not have the `load` method implemented.');
            }
            const artifacts = await this.handler.load();
            return this.loadSync(artifacts);
        }
        /**
         * Synchronously construct the in memory weight map and
         * compile the inference graph. Also initialize hashtable if any.
         *
         * @doc {heading: 'Models', subheading: 'Classes', ignoreCI: true}
         */
        loadSync(artifacts) {
            this.artifacts = artifacts;
            const graph = this.artifacts.modelTopology;
            let signature;
            if (this.artifacts.userDefinedMetadata != null &&
                this.artifacts.userDefinedMetadata.signature != null) {
                signature = // tslint:disable-next-line:no-any
                    this.artifacts.userDefinedMetadata.signature;
            }
            else {
                signature = this.artifacts.signature;
            }
            this.signature = signature;
            this.version = `${graph.versions.producer}.${graph.versions.minConsumer}`;
            const weightMap = decodeWeights(this.artifacts.weightData, this.artifacts.weightSpecs);
            this.executor = new GraphExecutor(OperationMapper.Instance.transformGraph(graph, this.signature));
            this.executor.weightMap = this.convertTensorMapToTensorsMap(weightMap);
            // Attach a model-level resourceManager to each executor to share resources,
            // such as `HashTable`.
            this.executor.resourceManager = this.resourceManager;
            if (artifacts.modelInitializer != null &&
                artifacts.modelInitializer.node != null) {
                const initializer = OperationMapper.Instance.transformGraph(artifacts.modelInitializer);
                this.initializer = new GraphExecutor(initializer);
                this.initializer.weightMap = this.executor.weightMap;
                // Attach a model-level resourceManager to the initializer, the
                // hashTables created from when executing the initializer will be stored
                // in the resourceManager.
                this.initializer.resourceManager = this.resourceManager;
                this.initializer.executeAsync({}, []);
            }
            return true;
        }
        /**
         * Save the configuration and/or weights of the GraphModel.
         *
         * An `IOHandler` is an object that has a `save` method of the proper
         * signature defined. The `save` method manages the storing or
         * transmission of serialized data ("artifacts") that represent the
         * model's topology and weights onto or via a specific medium, such as
         * file downloads, local storage, IndexedDB in the web browser and HTTP
         * requests to a server. TensorFlow.js provides `IOHandler`
         * implementations for a number of frequently used saving mediums, such as
         * `tf.io.browserDownloads` and `tf.io.browserLocalStorage`. See `tf.io`
         * for more details.
         *
         * This method also allows you to refer to certain types of `IOHandler`s
         * as URL-like string shortcuts, such as 'localstorage://' and
         * 'indexeddb://'.
         *
         * Example 1: Save `model`'s topology and weights to browser [local
         * storage](https://developer.mozilla.org/en-US/docs/Web/API/Window/localStorage);
         * then load it back.
         *
         * ```js
         * const modelUrl =
         *    'https://storage.googleapis.com/tfjs-models/savedmodel/mobilenet_v2_1.0_224/model.json';
         * const model = await tf.loadGraphModel(modelUrl);
         * const zeros = tf.zeros([1, 224, 224, 3]);
         * model.predict(zeros).print();
         *
         * const saveResults = await model.save('localstorage://my-model-1');
         *
         * const loadedModel = await tf.loadGraphModel('localstorage://my-model-1');
         * console.log('Prediction from loaded model:');
         * model.predict(zeros).print();
         * ```
         *
         * @param handlerOrURL An instance of `IOHandler` or a URL-like,
         * scheme-based string shortcut for `IOHandler`.
         * @param config Options for saving the model.
         * @returns A `Promise` of `SaveResult`, which summarizes the result of
         * the saving, such as byte sizes of the saved artifacts for the model's
         *   topology and weight values.
         *
         * @doc {heading: 'Models', subheading: 'Classes', ignoreCI: true}
         */
        async save(handlerOrURL, config) {
            if (typeof handlerOrURL === 'string') {
                const handlers = getSaveHandlers(handlerOrURL);
                if (handlers.length === 0) {
                    throw new Error(`Cannot find any save handlers for URL '${handlerOrURL}'`);
                }
                else if (handlers.length > 1) {
                    throw new Error(`Found more than one (${handlers.length}) save handlers for ` +
                        `URL '${handlerOrURL}'`);
                }
                handlerOrURL = handlers[0];
            }
            if (handlerOrURL.save == null) {
                throw new Error('GraphModel.save() cannot proceed because the IOHandler ' +
                    'provided does not have the `save` attribute defined.');
            }
            return handlerOrURL.save(this.artifacts);
        }
        /**
         * Execute the inference for the input tensors.
         *
         * @param input The input tensors, when there is single input for the model,
         * inputs param should be a `tf.Tensor`. For models with mutliple inputs,
         * inputs params should be in either `tf.Tensor`[] if the input order is
         * fixed, or otherwise NamedTensorMap format.
         *
         * For model with multiple inputs, we recommend you use NamedTensorMap as the
         * input type, if you use `tf.Tensor`[], the order of the array needs to
         * follow the
         * order of inputNodes array. @see {@link GraphModel.inputNodes}
         *
         * You can also feed any intermediate nodes using the NamedTensorMap as the
         * input type. For example, given the graph
         *    InputNode => Intermediate => OutputNode,
         * you can execute the subgraph Intermediate => OutputNode by calling
         *    model.execute('IntermediateNode' : tf.tensor(...));
         *
         * This is useful for models that uses tf.dynamic_rnn, where the intermediate
         * state needs to be fed manually.
         *
         * For batch inference execution, the tensors for each input need to be
         * concatenated together. For example with mobilenet, the required input shape
         * is [1, 244, 244, 3], which represents the [batch, height, width, channel].
         * If we are provide a batched data of 100 images, the input tensor should be
         * in the shape of [100, 244, 244, 3].
         *
         * @param config Prediction configuration for specifying the batch size and
         * output node names. Currently the batch size option is ignored for graph
         * model.
         *
         * @returns Inference result tensors. The output would be single `tf.Tensor`
         * if model has single output node, otherwise Tensor[] or NamedTensorMap[]
         * will be returned for model with multiple outputs.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        predict(inputs, config) {
            return this.execute(inputs, this.outputNodes);
        }
        normalizeInputs(inputs) {
            if (!(inputs instanceof Tensor) && !Array.isArray(inputs)) {
                // The input is already a NamedTensorMap.
                return inputs;
            }
            inputs = Array.isArray(inputs) ? inputs : [inputs];
            if (inputs.length !== this.inputNodes.length) {
                throw new Error('Input tensor count mismatch,' +
                    `the graph model has ${this.inputNodes.length} placeholders, ` +
                    `while there are ${inputs.length} input tensors.`);
            }
            return this.inputNodes.reduce((map, inputName, i) => {
                map[inputName] = inputs[i];
                return map;
            }, {});
        }
        normalizeOutputs(outputs) {
            outputs = outputs || this.outputNodes;
            return !Array.isArray(outputs) ? [outputs] : outputs;
        }
        /**
         * Executes inference for the model for given input tensors.
         * @param inputs tensor, tensor array or tensor map of the inputs for the
         * model, keyed by the input node names.
         * @param outputs output node name from the Tensorflow model, if no
         * outputs are specified, the default outputs of the model would be used.
         * You can inspect intermediate nodes of the model by adding them to the
         * outputs array.
         *
         * @returns A single tensor if provided with a single output or no outputs
         * are provided and there is only one default output, otherwise return a
         * tensor array. The order of the tensor array is the same as the outputs
         * if provided, otherwise the order of outputNodes attribute of the model.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        execute(inputs, outputs) {
            inputs = this.normalizeInputs(inputs);
            outputs = this.normalizeOutputs(outputs);
            const result = this.executor.execute(inputs, outputs);
            return result.length > 1 ? result : result[0];
        }
        /**
         * Executes inference for the model for given input tensors in async
         * fashion, use this method when your model contains control flow ops.
         * @param inputs tensor, tensor array or tensor map of the inputs for the
         * model, keyed by the input node names.
         * @param outputs output node name from the Tensorflow model, if no outputs
         * are specified, the default outputs of the model would be used. You can
         * inspect intermediate nodes of the model by adding them to the outputs
         * array.
         *
         * @returns A Promise of single tensor if provided with a single output or
         * no outputs are provided and there is only one default output, otherwise
         * return a tensor map.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        async executeAsync(inputs, outputs) {
            inputs = this.normalizeInputs(inputs);
            outputs = this.normalizeOutputs(outputs);
            const result = await this.executor.executeAsync(inputs, outputs);
            return result.length > 1 ? result : result[0];
        }
        /**
         * Get intermediate tensors for model debugging mode (flag
         * KEEP_INTERMEDIATE_TENSORS is true).
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        getIntermediateTensors() {
            return this.executor.getIntermediateTensors();
        }
        /**
         * Dispose intermediate tensors for model debugging mode (flag
         * KEEP_INTERMEDIATE_TENSORS is true).
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        disposeIntermediateTensors() {
            this.executor.disposeIntermediateTensors();
        }
        convertTensorMapToTensorsMap(map) {
            return Object.keys(map).reduce((newMap, key) => {
                newMap[key] = [map[key]];
                return newMap;
            }, {});
        }
        /**
         * Releases the memory used by the weight tensors and resourceManager.
         *
         * @doc {heading: 'Models', subheading: 'Classes'}
         */
        dispose() {
            this.executor.dispose();
            if (this.initializer) {
                this.initializer.dispose();
            }
            this.resourceManager.dispose();
        }
    }
    /**
     * Load a graph model given a URL to the model definition.
     *
     * Example of loading MobileNetV2 from a URL and making a prediction with a
     * zeros input:
     *
     * ```js
     * const modelUrl =
     *    'https://storage.googleapis.com/tfjs-models/savedmodel/mobilenet_v2_1.0_224/model.json';
     * const model = await tf.loadGraphModel(modelUrl);
     * const zeros = tf.zeros([1, 224, 224, 3]);
     * model.predict(zeros).print();
     * ```
     *
     * Example of loading MobileNetV2 from a TF Hub URL and making a prediction with
     * a zeros input:
     *
     * ```js
     * const modelUrl =
     *    'https://tfhub.dev/google/imagenet/mobilenet_v2_140_224/classification/2';
     * const model = await tf.loadGraphModel(modelUrl, {fromTFHub: true});
     * const zeros = tf.zeros([1, 224, 224, 3]);
     * model.predict(zeros).print();
     * ```
     * @param modelUrl The url or an `io.IOHandler` that loads the model.
     * @param options Options for the HTTP request, which allows to send credentials
     *    and custom headers.
     *
     * @doc {heading: 'Models', subheading: 'Loading'}
     */
    async function loadGraphModel(modelUrl, options = {}) {
        if (modelUrl == null) {
            throw new Error('modelUrl in loadGraphModel() cannot be null. Please provide a url ' +
                'or an IOHandler that loads the model');
        }
        if (options == null) {
            options = {};
        }
        if (options.fromTFHub) {
            if (modelUrl.load == null) {
                if (!modelUrl.endsWith('/')) {
                    modelUrl = modelUrl + '/';
                }
                modelUrl = `${modelUrl}${DEFAULT_MODEL_NAME}${TFHUB_SEARCH_PARAM}`;
            }
        }
        const model = new GraphModel(modelUrl, options);
        await model.load();
        return model;
    }

    /** @license See the LICENSE file. */
    // This code is auto-generated, do not modify this file!
    const version$5 = '3.17.0';

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */

    var dist = /*#__PURE__*/Object.freeze({
        __proto__: null,
        GraphModel: GraphModel,
        loadGraphModel: loadGraphModel,
        deregisterOp: deregisterOp,
        registerOp: registerOp,
        version_converter: version$5
    });

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * Apply a mapping function to a nested structure in a recursive manner.
     *
     * The result of the mapping is an object with the same nested structure (i.e.,
     * of arrays and dicts) as the input, except that some subtrees are replaced,
     * according to the results of the mapping function.
     *
     * Mappings are memoized.  Thus, if the nested structure contains the same
     * object in multiple positions, the output will contain the same mapped object
     * in those positions.  Cycles are not supported, however.
     *
     * @param input: The object to which to apply the mapping function.
     * @param mapFn: A function that expects a single node of the object tree, and
     *   returns a `DeepMapResult`.  The `DeepMapResult` either provides a
     *   replacement value for that node (i.e., replacing the subtree), or indicates
     *   that the node should be processed recursively.
     */
    function deepMap(input, mapFn) {
        return deepMapInternal(input, mapFn);
    }
    /**
     * @param seen: A Map of known object mappings (i.e., memoized results of
     *   `mapFn()`)
     * @param containedIn: An set containing objects on the reference path currently
     *   being processed (used to detect cycles).
     */
    function deepMapInternal(input, mapFn, seen = new Map(), containedIn = new Set()) {
        if (input == null) {
            return null;
        }
        if (typeof Blob === 'function' && input instanceof Blob) {
            return input.slice();
        }
        if (containedIn.has(input)) {
            throw new Error('Circular references are not supported.');
        }
        if (seen.has(input)) {
            return seen.get(input);
        }
        const result = mapFn(input);
        if (result.recurse && result.value !== null) {
            throw new Error('A deep map function may not return both a value and recurse=true.');
        }
        if (!result.recurse) {
            seen.set(input, result.value);
            return result.value;
        }
        else if (isIterable(input)) {
            // tslint:disable-next-line:no-any
            const mappedIterable = Array.isArray(input) ? [] : {};
            containedIn.add(input);
            for (const k in input) {
                const child = input[k];
                const childResult = deepMapInternal(child, mapFn, seen, containedIn);
                mappedIterable[k] = childResult;
            }
            containedIn.delete(input);
            if (input.__proto__) {
                mappedIterable.__proto__ = input.__proto__;
            }
            return mappedIterable;
        }
        else {
            throw new Error(`Can't recurse into non-iterable type: ${input}`);
        }
    }
    // TODO(soergel, kangyizhang) Reconsider naming of deepZip() to avoid confusion
    // with zip()
    /**
     * Zip nested structures together in a recursive manner.
     *
     * This has the effect of transposing or pivoting data, e.g. converting it from
     * a row-major representation to a column-major representation.
     *
     * For example, `deepZip([{a: 1, b: 2}, {a: 3, b: 4}])` returns
     * `{a: [1, 3], b: [2, 4]}`.
     *
     * The inputs should all have the same nested structure (i.e., of arrays and
     * dicts).  The result is a single object with the same nested structure, where
     * the leaves are arrays collecting the values of the inputs at that location
     * (or, optionally, the result of a custom function applied to those arrays).
     *
     * @param inputs: An array of the objects to zip together.
     * @param zipFn: (optional) A function that expects an array of elements at a
     *   single node of the object tree, and returns a `DeepMapResult`.  The
     *   `DeepMapResult` either provides a result value for that node (i.e.,
     *   representing the subtree), or indicates that the node should be processed
     *   recursively.  The default zipFn recurses as far as possible and places
     *   arrays at the leaves.
     */
    function deepZip(inputs, zipFn = zipToList) {
        return deepZipInternal(inputs, zipFn);
    }
    /**
     * @param containedIn: An set containing objects on the reference path currently
     *   being processed (used to detect cycles).
     */
    function deepZipInternal(inputs, zipFn, containedIn = new Set()) {
        // The recursion follows the structure of input 0; it's assumed that all the
        // other inputs have the same structure.
        const input = inputs[0];
        if (containedIn.has(input)) {
            throw new Error('Circular references are not supported.');
        }
        const result = zipFn(inputs);
        if (result.recurse && result.value !== null) {
            throw new Error('A deep zip function may not return both a value and recurse=true.');
        }
        if (!result.recurse) {
            return result.value;
        }
        else if (isIterable(input)) {
            // tslint:disable-next-line:no-any
            const mappedIterable = Array.isArray(input) ? [] : {};
            containedIn.add(input);
            for (const k in input) {
                const children = inputs.map(x => x[k]);
                const childResult = deepZipInternal(children, zipFn, containedIn);
                mappedIterable[k] = childResult;
            }
            containedIn.delete(input);
            return mappedIterable;
        }
        else {
            throw new Error(`Can't recurse into non-iterable type: ${input}`);
        }
    }
    // tslint:disable-next-line:no-any
    function zipToList(x) {
        if (x === null) {
            return null;
        }
        // TODO(soergel): validate array type?
        if (isIterable(x[0])) {
            return { value: null, recurse: true };
        }
        else {
            return { value: x, recurse: false };
        }
    }
    /**
     * Apply an async mapping function to a nested structure in a recursive manner.
     *
     * This first creates a nested structure of Promises, and then awaits all of
     * those, resulting in a single Promise for a resolved nested structure.
     *
     * The result of the mapping is an object with the same nested structure (i.e.,
     * of arrays and dicts) as the input, except that some subtrees are replaced,
     * according to the results of the mapping function.
     *
     * Mappings are memoized.  Thus, if the nested structure contains the same
     * object in multiple positions, the output will contain the same mapped object
     * in those positions.  Cycles are not supported, however.
     *
     * @param input: The object to which to apply the mapping function.
     * @param mapFn: A function that expects a single node of the object tree, and
     *   returns a `DeepMapAsyncResult`.  The `DeepMapAsyncResult` either provides
     *   a `Promise` for a replacement value for that node (i.e., replacing the
     *   subtree), or indicates that the node should be processed recursively.  Note
     *   that the decision whether or not to recurse must be made immediately; only
     *   the mapped value may be promised.
     */
    async function deepMapAndAwaitAll(input, mapFn) {
        const seen = new Map();
        // First do a normal deepMap, collecting Promises in 'seen' as a side effect.
        deepMapInternal(input, mapFn, seen);
        // Replace the Promises in 'seen' in place.
        // Note TypeScript provides no async map iteration, and regular map iteration
        // is broken too, so sadly we have to do Array.from() to make it work.
        // (There's no advantage to Promise.all(), and that would be tricky anyway.)
        for (const key of Array.from(seen.keys())) {
            const value = seen.get(key);
            if (isPromise(value)) {
                const mappedValue = await value;
                seen.set(key, mappedValue);
            }
        }
        // Normal deepMap again, this time filling in the resolved values.
        // It's unfortunate that we have to do two passes.
        // TODO(soergel): test performance and think harder about a fast solution.
        const result = deepMapInternal(input, mapFn, seen);
        return result;
    }
    /**
     * Determine whether the argument is iterable.
     *
     * @returns true if the argument is an array or any non-Tensor object.
     */
    // tslint:disable-next-line:no-any
    function isIterable(obj) {
        let isTextDecoder = false;
        if (env$1().get('IS_BROWSER')) {
            isTextDecoder = obj instanceof TextDecoder;
        }
        else {
            // tslint:disable-next-line:no-require-imports
            const { StringDecoder } = require('string_decoder');
            isTextDecoder = obj instanceof StringDecoder;
        }
        return obj != null && (!ArrayBuffer.isView(obj)) &&
            (Array.isArray(obj) ||
                (typeof obj === 'object' && !(obj instanceof Tensor) &&
                    !(obj instanceof Promise) && !isTextDecoder));
    }
    /**
     * Determine whether the argument can be converted to Tensor.
     *
     * Tensors, primitives, arrays, and TypedArrays all qualify; anything else does
     * not.
     *
     * @returns true if the argument can be converted to Tensor.
     */
    // tslint:disable-next-line:no-any
    function canTensorify(obj) {
        return obj == null || isPrimitive$1(obj) || Array.isArray(obj) ||
            (typeof obj === 'object' && (obj instanceof Tensor)) ||
            isTypedArray$1(obj);
    }
    /**
     * Returns true if the given `value` is a primitive type. Otherwise returns
     * false. This is equivalant to node util.isPrimitive
     */
    function isPrimitive$1(value) {
        return (value === null ||
            (typeof value !== 'object' && typeof value !== 'function'));
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    function deepClone(container) {
        return deepMap(container, cloneIfTensor);
    }
    // tslint:disable-next-line: no-any
    function cloneIfTensor(item) {
        if (item instanceof Tensor) {
            return ({ value: item.clone(), recurse: false });
        }
        else if (isIterable(item)) {
            return { value: null, recurse: true };
        }
        else {
            return { value: item, recurse: false };
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * A ring buffer, providing O(1) FIFO, LIFO, and related operations.
     */
    class RingBuffer {
        /**
         * Constructs a `RingBuffer`.
         * @param capacity The number of items that the buffer can accomodate.
         */
        constructor(capacity) {
            this.capacity = capacity;
            // Note we store the indices in the range 0 <= index < 2*capacity.
            // This allows us to distinguish the full from the empty case.
            // See https://www.snellman.net/blog/archive/2016-12-13-ring-buffers/
            this.begin = 0; // inclusive
            this.end = 0; // exclusive
            if (capacity == null) {
                throw new RangeError('Can\'t create a ring buffer of unknown capacity.');
            }
            if (capacity < 1) {
                throw new RangeError('Can\'t create ring buffer of capacity < 1.');
            }
            this.data = new Array(capacity);
            this.doubledCapacity = 2 * capacity;
        }
        /**
         * Map any index into the range 0 <= index < 2*capacity.
         */
        wrap(index) {
            // don't trust % on negative numbers
            while (index < 0) {
                index += this.doubledCapacity;
            }
            return index % this.doubledCapacity;
        }
        get(index) {
            if (index < 0) {
                throw new RangeError('Can\'t get item at a negative index.');
            }
            return this.data[index % this.capacity];
        }
        set(index, value) {
            if (index < 0) {
                throw new RangeError('Can\'t set item at a negative index.');
            }
            this.data[index % this.capacity] = value;
        }
        /**
         * Returns the current number of items in the buffer.
         */
        length() {
            let length = this.end - this.begin;
            if (length < 0) {
                length = this.doubledCapacity + length;
            }
            return length;
        }
        /**
         * Reports whether the buffer is full.
         * @returns true if the number of items in the buffer equals its capacity, and
         *   false otherwise.
         */
        isFull() {
            return this.length() === this.capacity;
        }
        /**
         * Reports whether the buffer is empty.
         * @returns true if the number of items in the buffer equals zero, and
         *   false otherwise.
         */
        isEmpty() {
            return this.length() === 0;
        }
        /**
         * Adds an item to the end of the buffer.
         */
        push(value) {
            if (this.isFull()) {
                throw new RangeError('Ring buffer is full.');
            }
            this.set(this.end, value);
            this.end = this.wrap(this.end + 1);
        }
        /**
         * Adds many items to the end of the buffer, in order.
         */
        pushAll(values) {
            for (const value of values) {
                this.push(value);
            }
        }
        /**
         * Removes and returns the last item in the buffer.
         */
        pop() {
            if (this.isEmpty()) {
                throw new RangeError('Ring buffer is empty.');
            }
            this.end = this.wrap(this.end - 1);
            const result = this.get(this.end);
            this.set(this.end, undefined);
            return result;
        }
        /**
         * Adds an item to the beginning of the buffer.
         */
        unshift(value) {
            if (this.isFull()) {
                throw new RangeError('Ring buffer is full.');
            }
            this.begin = this.wrap(this.begin - 1);
            this.set(this.begin, value);
        }
        /**
         * Removes and returns the first item in the buffer.
         */
        shift() {
            if (this.isEmpty()) {
                throw new RangeError('Ring buffer is empty.');
            }
            const result = this.get(this.begin);
            this.set(this.begin, undefined);
            this.begin = this.wrap(this.begin + 1);
            return result;
        }
        /**
         * Removes and returns a specific item in the buffer, and moves the last item
         * to the vacated slot.  This is useful for implementing a shuffling stream.
         * Note that this operation necessarily scrambles the original order.
         *
         * @param relativeIndex: the index of the item to remove, relative to the
         *   first item in the buffer (e.g., hiding the ring nature of the underlying
         *   storage).
         */
        shuffleExcise(relativeIndex) {
            if (this.isEmpty()) {
                throw new RangeError('Ring buffer is empty.');
            }
            const index = this.wrap(this.begin + relativeIndex);
            const result = this.get(index);
            this.set(index, this.pop());
            return result;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    class GrowingRingBuffer extends RingBuffer {
        /**
         * Constructs a `GrowingRingBuffer`.
         */
        constructor() {
            super(GrowingRingBuffer.INITIAL_CAPACITY);
        }
        isFull() {
            return false;
        }
        push(value) {
            if (super.isFull()) {
                this.expand();
            }
            super.push(value);
        }
        unshift(value) {
            if (super.isFull()) {
                this.expand();
            }
            super.unshift(value);
        }
        /**
         * Doubles the capacity of the buffer.
         */
        expand() {
            const newCapacity = this.capacity * 2;
            const newData = new Array(newCapacity);
            const len = this.length();
            // Rotate the buffer to start at index 0 again, since we can't just
            // allocate more space at the end.
            for (let i = 0; i < len; i++) {
                newData[i] = this.get(this.wrap(this.begin + i));
            }
            this.data = newData;
            this.capacity = newCapacity;
            this.doubledCapacity = 2 * this.capacity;
            this.begin = 0;
            this.end = len;
        }
    }
    GrowingRingBuffer.INITIAL_CAPACITY = 32;

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    // Here we implement a simple asynchronous iterator.
    // This lets us avoid using either third-party stream libraries or
    // recent TypeScript language support requiring polyfills.
    /**
     * Create a `LazyIterator` from an array of items.
     */
    function iteratorFromItems(items) {
        return new ArrayIterator(items);
    }
    /**
     * Create a `LazyIterator` from a function.
     *
     * ```js
     * let i = -1;
     * const func = () =>
     *    ++i < 5 ? {value: i, done: false} : {value: null, done: true};
     * const iter = tf.data.iteratorFromFunction(func);
     * await iter.forEachAsync(e => console.log(e));
     * ```
     *
     * @param func A function that produces data on each call.
     */
    function iteratorFromFunction(func) {
        return new FunctionCallIterator(func);
    }
    /**
     * Create a `LazyIterator` by concatenating underlying streams, which are
     * themselves provided as a stream.
     *
     * This can also be thought of as a "stream flatten" operation.
     *
     * @param baseIterators A stream of streams to be concatenated.
     * @param baseErrorHandler An optional function that can intercept `Error`s
     *   raised during a `next()` call on the base stream.  This function can decide
     *   whether the error should be propagated, whether the error should be
     *   ignored, or whether the base stream should be terminated.
     */
    function iteratorFromConcatenated(baseIterators, baseErrorHandler) {
        return new ChainedIterator(baseIterators, baseErrorHandler);
    }
    /**
     * Create a `LazyIterator` by zipping together an array, dict, or nested
     * structure of `LazyIterator`s (and perhaps additional constants).
     *
     * The underlying streams must provide elements in a consistent order such
     * that they correspond.
     *
     * Typically, the underlying streams should have the same number of
     * elements. If they do not, the behavior is determined by the
     * `mismatchMode` argument.
     *
     * The nested structure of the `iterators` argument determines the
     * structure of elements in the resulting iterator.
     *
     * @param iterators: An array or object containing LazyIterators at the
     * leaves.
     * @param mismatchMode: Determines what to do when one underlying iterator
     * is exhausted before the others.  `ZipMismatchMode.FAIL` (the default)
     * causes an error to be thrown in this case.  `ZipMismatchMode.SHORTEST`
     * causes the zipped iterator to terminate with the furst underlying
     * streams, so elements remaining on the longer streams are ignored.
     * `ZipMismatchMode.LONGEST` causes the zipped stream to continue, filling
     * in nulls for the exhausted streams, until all streams are exhausted.
     */
    function iteratorFromZipped(iterators, mismatchMode = ZipMismatchMode.FAIL) {
        return new ZipIterator(iterators, mismatchMode);
    }
    /**
     * An asynchronous iterator, providing lazy access to a potentially
     * unbounded stream of elements.
     *
     * Iterator can be obtained from a dataset:
     * `const iter = await dataset.iterator();`
     */
    class LazyIterator {
        /**
         * Collect all remaining elements of a bounded stream into an array.
         * Obviously this will succeed only for small streams that fit in memory.
         * Useful for testing.
         *
         * @returns A Promise for an array of stream elements, which will resolve
         *   when the stream is exhausted.
         */
        async toArray() {
            const result = [];
            let x = await this.next();
            while (!x.done) {
                result.push(x.value);
                x = await this.next();
            }
            return result;
        }
        /**
         * Collect all elements of this dataset into an array with prefetching 100
         * elements. This is useful for testing, because the prefetch changes the
         * order in which the Promises are resolved along the processing pipeline.
         * This may help expose bugs where results are dependent on the order of
         * Promise resolution rather than on the logical order of the stream (i.e.,
         * due to hidden mutable state).
         *
         * @returns A Promise for an array of stream elements, which will resolve
         *   when the stream is exhausted.
         */
        async toArrayForTest() {
            const stream = this.prefetch(100);
            const result = [];
            let x = await stream.next();
            while (!x.done) {
                result.push(x.value);
                x = await stream.next();
            }
            return result;
        }
        /**
         * Draw items from the stream until it is exhausted.
         *
         * This can be useful when the stream has side effects but no output.  In
         * that case, calling this function guarantees that the stream will be
         * fully processed.
         */
        async resolveFully() {
            let x = await this.next();
            while (!x.done) {
                x = await this.next();
            }
        }
        /**
         * Draw items from the stream until it is exhausted, or a predicate fails.
         *
         * This can be useful when the stream has side effects but no output.  In
         * that case, calling this function guarantees that the stream will be
         * fully processed.
         */
        async resolveWhile(predicate) {
            let x = await this.next();
            let shouldContinue = predicate(x.value);
            while ((!x.done) && shouldContinue) {
                x = await this.next();
                shouldContinue = predicate(x.value);
            }
        }
        /**
         * Handles errors thrown on this stream using a provided handler function.
         *
         * @param handler A function that handles any `Error` thrown during a `next()`
         *   call and returns true if the stream should continue (dropping the failed
         *   call) or false if the stream should quietly terminate.  If the handler
         *   itself throws (or rethrows) an `Error`, that will be propagated.
         *
         * @returns A `LazyIterator` of elements passed through from upstream,
         *   possibly filtering or terminating on upstream `next()` calls that
         *   throw an `Error`.
         */
        handleErrors(handler) {
            return new ErrorHandlingLazyIterator(this, handler);
        }
        // TODO(soergel): Implement reduce() etc.
        /**
         * Filters this stream according to `predicate`.
         *
         * @param predicate A function mapping a stream element to a boolean or a
         * `Promise` for one.
         *
         * @returns A `LazyIterator` of elements for which the predicate was true.
         */
        filter(predicate) {
            return new FilterIterator(this, predicate);
        }
        /**
         * Maps this stream through a 1-to-1 transform.
         *
         * @param transform A function mapping a stream element to a transformed
         *   element.
         *
         * @returns A `LazyIterator` of transformed elements.
         */
        map(transform) {
            return new MapIterator(this, transform);
        }
        /**
         * Maps this stream through an async 1-to-1 transform.
         *
         * @param transform A function mapping a stream element to a `Promise` for a
         *   transformed stream element.
         *
         * @returns A `LazyIterator` of transformed elements.
         */
        mapAsync(transform) {
            return new AsyncMapIterator(this, transform);
        }
        /**
         * Maps this stream through a 1-to-1 transform, forcing serial execution.
         *
         * @param transform A function mapping a stream element to a transformed
         *   element.
         *
         * @returns A `LazyIterator` of transformed elements.
         */
        serialMapAsync(transform) {
            return new AsyncMapIterator(this, transform).serial();
        }
        /**
         * Maps this stream through a 1-to-many transform.
         *
         * @param transform A function mapping a stream element to an array of
         *   transformed elements.
         *
         * @returns A `DataStream` of transformed elements.
         */
        flatmap(transform) {
            return new FlatmapIterator(this, transform);
        }
        /**
         * Apply a function to every element of the stream.
         *
         * @param f A function to apply to each stream element.
         */
        async forEachAsync(f) {
            return this.map(f).resolveFully();
        }
        /**
         * Apply a function to every element of the stream, forcing serial execution.
         *
         * @param f A function to apply to each stream element.  Should return 'true'
         *   to indicate that the stream should continue, or 'false' to cause it to
         *   terminate.
         */
        async serialForEach(f) {
            return this.serialMapAsync(f).resolveWhile(x => (x === true));
        }
        /**
         * Groups elements into batches, represented as arrays of elements.
         *
         * We can think of the elements of this iterator as 'rows' (even if they are
         * nested structures).  By the same token, consecutive values for a given
         * key within the elements form a 'column'.  This matches the usual sense of
         * 'row' and 'column' when processing tabular data (e.g., parsing a CSV).
         *
         * Thus, "Row-major" means that the resulting batch is simply a collection of
         * rows: `[row1, row2, row3, ...]`.  This is contrast to the column-major
         * form, which is needed for vectorized computation.
         *
         * @param batchSize The number of elements desired per batch.
         * @param smallLastBatch Whether to emit the final batch when it has fewer
         *   than batchSize elements. Default true.
         * @returns A `LazyIterator` of batches of elements, represented as arrays
         *   of the original element type.
         */
        rowMajorBatch(batchSize, smallLastBatch = true) {
            return new RowMajorBatchIterator(this, batchSize, smallLastBatch);
        }
        /**
         * Groups elements into batches, represented in column-major form.
         *
         * We can think of the elements of this iterator as 'rows' (even if they are
         * nested structures).  By the same token, consecutive values for a given
         * key within the elements form a 'column'.  This matches the usual sense of
         * 'row' and 'column' when processing tabular data (e.g., parsing a CSV).
         *
         * Thus, "column-major" means that the resulting batch is a (potentially
         * nested) structure representing the columns.  Each column entry, then,
         * contains a collection of the values found in that column for a range of
         * input elements.  This representation allows for vectorized computation, in
         * contrast to the row-major form.
         *
         * The inputs should all have the same nested structure (i.e., of arrays and
         * dicts).  The result is a single object with the same nested structure,
         * where the leaves are arrays collecting the values of the inputs at that
         * location (or, optionally, the result of a custom function applied to those
         * arrays).
         *
         * @param batchSize The number of elements desired per batch.
         * @param smallLastBatch Whether to emit the final batch when it has fewer
         *   than batchSize elements. Default true.
         * @param zipFn: (optional) A function that expects an array of elements at a
         *   single node of the object tree, and returns a `DeepMapResult`.  The
         *   `DeepMapResult` either provides a result value for that node (i.e.,
         *   representing the subtree), or indicates that the node should be processed
         *   recursively.  The default zipFn recurses as far as possible and places
         *   arrays at the leaves.
         * @returns A `LazyIterator` of batches of elements, represented as an object
         *   with collections at the leaves.
         */
        columnMajorBatch(batchSize, smallLastBatch = true, 
        // tslint:disable-next-line:no-any
        zipFn = zipToList) {
            // First collect the desired number of input elements as a row-major batch.
            const rowBatches = this.rowMajorBatch(batchSize, smallLastBatch);
            // Now 'rotate' or 'pivot' the data, collecting all values from each column
            // in the batch (i.e., for each key within the elements) into an array.
            return rowBatches.map(x => deepZip(x, zipFn));
        }
        /**
         * Concatenate this `LazyIterator` with another.
         *
         * @param iterator A `LazyIterator` to be concatenated onto this one.
         * @param baseErrorHandler An optional function that can intercept `Error`s
         *   raised during a `next()` call on the base stream.  This function can
         *   decide whether the error should be propagated, whether the error should
         *   be ignored, or whether the base stream should be terminated.
         * @returns A `LazyIterator`.
         */
        concatenate(iterator, baseErrorHandler) {
            return new ChainedIterator(iteratorFromItems([this, iterator]), baseErrorHandler);
        }
        /**
         * Limits this stream to return at most `count` items.
         *
         * @param count The maximum number of items to provide from the stream. If
         * a negative or undefined value is given, the entire stream is returned
         *   unaltered.
         */
        take(count) {
            if (count < 0 || count == null) {
                return this;
            }
            return new TakeIterator(this, count);
        }
        /**
         * Skips the first `count` items in this stream.
         *
         * @param count The number of items to skip.  If a negative or undefined
         * value is given, the entire stream is returned unaltered.
         */
        skip(count) {
            if (count < 0 || count == null) {
                return this;
            }
            return new SkipIterator(this, count);
        }
        /**
         * Prefetch the first `bufferSize` items in this stream.
         *
         * Note this prefetches Promises, but makes no guarantees about when those
         * Promises resolve.
         *
         * @param bufferSize: An integer specifying the number of elements to be
         *   prefetched.
         */
        prefetch(bufferSize) {
            return new PrefetchIterator(this, bufferSize);
        }
        // TODO(soergel): deep sharded shuffle, where supported
        /**
         * Randomly shuffles the elements of this stream.
         *
         * @param bufferSize: An integer specifying the number of elements from
         * this stream from which the new stream will sample.
         * @param seed: (Optional.) An integer specifying the random seed that
         * will be used to create the distribution.
         */
        shuffle(windowSize, seed) {
            return new ShuffleIterator(this, windowSize, seed);
        }
        /**
         * Force an iterator to execute serially: each next() call will await the
         * prior one, so that they cannot execute concurrently.
         */
        serial() {
            return new SerialIterator(this);
        }
    }
    // ============================================================================
    // The following private classes serve to implement the chainable methods
    // on LazyIterator.  Unfortunately they can't be placed in separate files,
    // due to resulting trouble with circular imports.
    // ============================================================================
    // Iterators that just extend LazyIterator directly
    // ============================================================================
    class ArrayIterator extends LazyIterator {
        constructor(items) {
            super();
            this.items = items;
            this.trav = 0;
        }
        summary() {
            return `Array of ${this.items.length} items`;
        }
        async next() {
            if (this.trav >= this.items.length) {
                return { value: null, done: true };
            }
            const item = this.items[this.trav];
            this.trav++;
            return { value: deepClone(item), done: false };
        }
    }
    class FunctionCallIterator extends LazyIterator {
        constructor(nextFn) {
            super();
            this.nextFn = nextFn;
        }
        summary() {
            return `Function call`;
        }
        async next() {
            try {
                return this.nextFn();
            }
            catch (e) {
                // Modify the error message but leave the stack trace intact
                e.message =
                    `Error thrown while iterating through a dataset: ${e.message}`;
                throw e;
            }
        }
    }
    class SerialIterator extends LazyIterator {
        constructor(upstream) {
            super();
            this.upstream = upstream;
            this.lastRead = Promise.resolve({ value: null, done: false });
        }
        summary() {
            return `${this.upstream.summary()} -> Serial`;
        }
        async next() {
            // This sets this.lastRead to a new Promise right away, as opposed to
            // saying `await this.lastRead; this.lastRead = this.serialNext();` which
            // would not work because this.nextRead would be updated only after the
            // promise resolves.
            this.lastRead = this.lastRead.then(() => this.serialNext());
            return this.lastRead;
        }
        async serialNext() {
            return this.upstream.next();
        }
    }
    class SkipIterator extends LazyIterator {
        constructor(upstream, maxCount) {
            super();
            this.upstream = upstream;
            this.maxCount = maxCount;
            // Local state that should not be clobbered by out-of-order execution.
            this.count = 0;
            this.lastRead = Promise.resolve({ value: null, done: false });
        }
        summary() {
            return `${this.upstream.summary()} -> Skip`;
        }
        async next() {
            // This sets this.lastRead to a new Promise right away, as opposed to
            // saying `await this.lastRead; this.lastRead = this.serialNext();` which
            // would not work because this.nextRead would be updated only after the
            // promise resolves.
            this.lastRead = this.lastRead.then(() => this.serialNext());
            return this.lastRead;
        }
        async serialNext() {
            // TODO(soergel): consider tradeoffs of reading in parallel, eg.
            // collecting next() promises in an Array and then waiting for
            // Promise.all() of those. Benefit: pseudo-parallel execution.  Drawback:
            // maybe delayed GC.
            while (this.count++ < this.maxCount) {
                const skipped = await this.upstream.next();
                // short-circuit if upstream is already empty
                if (skipped.done) {
                    return skipped;
                }
                dispose(skipped.value);
            }
            return this.upstream.next();
        }
    }
    class TakeIterator extends LazyIterator {
        constructor(upstream, maxCount) {
            super();
            this.upstream = upstream;
            this.maxCount = maxCount;
            this.count = 0;
        }
        summary() {
            return `${this.upstream.summary()} -> Take`;
        }
        async next() {
            if (this.count++ >= this.maxCount) {
                return { value: null, done: true };
            }
            return this.upstream.next();
        }
    }
    // Note this batch just groups items into row-wise element arrays.
    // Rotating these to a column-wise representation happens only at the dataset
    // level.
    class RowMajorBatchIterator extends LazyIterator {
        constructor(upstream, batchSize, enableSmallLastBatch = true) {
            super();
            this.upstream = upstream;
            this.batchSize = batchSize;
            this.enableSmallLastBatch = enableSmallLastBatch;
            this.lastRead = Promise.resolve({ value: null, done: false });
        }
        summary() {
            return `${this.upstream.summary()} -> RowMajorBatch`;
        }
        async next() {
            // This sets this.lastRead to a new Promise right away, as opposed to
            // saying `await this.lastRead; this.lastRead = this.serialNext();` which
            // would not work because this.nextRead would be updated only after the
            // promise resolves.
            this.lastRead = this.lastRead.then(() => this.serialNext());
            return this.lastRead;
        }
        async serialNext() {
            const batch = [];
            while (batch.length < this.batchSize) {
                const item = await this.upstream.next();
                if (item.done) {
                    if (this.enableSmallLastBatch && batch.length > 0) {
                        return { value: batch, done: false };
                    }
                    return { value: null, done: true };
                }
                batch.push(item.value);
            }
            return { value: batch, done: false };
        }
    }
    class FilterIterator extends LazyIterator {
        constructor(upstream, predicate) {
            super();
            this.upstream = upstream;
            this.predicate = predicate;
            this.lastRead = Promise.resolve({ value: null, done: false });
        }
        summary() {
            return `${this.upstream.summary()} -> Filter`;
        }
        async next() {
            // This sets this.lastRead to a new Promise right away, as opposed to
            // saying `await this.lastRead; this.lastRead = this.serialNext();` which
            // would not work because this.nextRead would be updated only after the
            // promise resolves.
            this.lastRead = this.lastRead.then(() => this.serialNext());
            return this.lastRead;
        }
        async serialNext() {
            while (true) {
                const item = await this.upstream.next();
                if (item.done || this.predicate(item.value)) {
                    return item;
                }
                dispose(item.value);
            }
        }
    }
    class MapIterator extends LazyIterator {
        constructor(upstream, transform) {
            super();
            this.upstream = upstream;
            this.transform = transform;
        }
        summary() {
            return `${this.upstream.summary()} -> Map`;
        }
        async next() {
            const item = await this.upstream.next();
            if (item.done) {
                return { value: null, done: true };
            }
            const inputTensors = getTensorsInContainer(item.value);
            // Careful: the transform may mutate the item in place.
            // That's why we have to remember the input Tensors above, and then
            // below dispose only those that were not passed through to the output.
            // Note too that the transform function is responsible for tidying
            // any intermediate Tensors.  Here we are concerned only about the
            // inputs.
            const mapped = this.transform(item.value);
            const outputTensors = getTensorsInContainer(mapped);
            // TODO(soergel) faster intersection
            // TODO(soergel) move to tf.disposeExcept(in, out)?
            for (const t of inputTensors) {
                if (!isTensorInList(t, outputTensors)) {
                    t.dispose();
                }
            }
            return { value: mapped, done: false };
        }
    }
    class ErrorHandlingLazyIterator extends LazyIterator {
        constructor(upstream, handler) {
            super();
            this.upstream = upstream;
            this.handler = handler;
            this.count = 0;
            this.lastRead = Promise.resolve({ value: null, done: false });
        }
        summary() {
            return `${this.upstream.summary()} -> handleErrors`;
        }
        async next() {
            // This sets this.lastRead to a new Promise right away, as opposed to
            // saying `await this.lastRead; this.lastRead = this.serialNext();` which
            // would not work because this.nextRead would be updated only after the
            // promise resolves.
            this.lastRead = this.lastRead.then(() => this.serialNext());
            return this.lastRead;
        }
        async serialNext() {
            while (true) {
                try {
                    return await this.upstream.next();
                }
                catch (e) {
                    if (!this.handler(e)) {
                        return { value: null, done: true };
                    }
                    // If the handler returns true, loop and fetch the next upstream item.
                    // If the upstream iterator throws an endless stream of errors, and if
                    // the handler says to ignore them, then we loop forever here.  That is
                    // the correct behavior-- it's up to the handler to decide when to stop.
                }
            }
        }
    }
    class AsyncMapIterator extends LazyIterator {
        constructor(upstream, transform) {
            super();
            this.upstream = upstream;
            this.transform = transform;
        }
        summary() {
            return `${this.upstream.summary()} -> AsyncMap`;
        }
        async next() {
            const item = await this.upstream.next();
            if (item.done) {
                return { value: null, done: true };
            }
            const inputTensors = getTensorsInContainer(item.value);
            // Careful: the transform may mutate the item in place.
            // That's why we have to remember the input Tensors above, and then
            // below dispose only those that were not passed through to the output.
            // Note too that the transform function is responsible for tidying
            // any intermediate Tensors.  Here we are concerned only about the
            // inputs.
            const mapped = await this.transform(item.value);
            const outputTensors = getTensorsInContainer(mapped);
            // TODO(soergel) faster intersection
            // TODO(soergel) move to tf.disposeExcept(in, out)?
            for (const t of inputTensors) {
                if (!isTensorInList(t, outputTensors)) {
                    t.dispose();
                }
            }
            return { value: mapped, done: false };
        }
    }
    // Iterators that maintain a queue of pending items
    // ============================================================================
    /**
     * A base class for transforming streams that operate by maintaining an
     * output queue of elements that are ready to return via next().  This is
     * commonly required when the transformation is 1-to-many:  A call to next()
     * may trigger a call to the underlying stream, which will produce many
     * mapped elements of this stream-- of which we need to return only one, so
     * we have to queue the rest.
     */
    class OneToManyIterator extends LazyIterator {
        constructor() {
            super();
            this.outputQueue = new GrowingRingBuffer();
            this.lastRead = Promise.resolve({ value: null, done: false });
        }
        async next() {
            // This sets this.lastRead to a new Promise right away, as opposed to
            // saying `await this.lastRead; this.lastRead = this.serialNext();` which
            // would not work because this.nextRead would be updated only after the
            // promise resolves.
            this.lastRead = this.lastRead.then(() => this.serialNext());
            return this.lastRead;
        }
        async serialNext() {
            // Fetch so that the queue contains at least one item if possible.
            // If the upstream source is exhausted, AND there are no items left in
            // the output queue, then this stream is also exhausted.
            while (this.outputQueue.length() === 0) {
                // TODO(soergel): consider parallel reads.
                if (!await this.pump()) {
                    return { value: null, done: true };
                }
            }
            return { value: this.outputQueue.shift(), done: false };
        }
    }
    class FlatmapIterator extends OneToManyIterator {
        constructor(upstream, transform) {
            super();
            this.upstream = upstream;
            this.transform = transform;
        }
        summary() {
            return `${this.upstream.summary()} -> Flatmap`;
        }
        async pump() {
            const item = await this.upstream.next();
            if (item.done) {
                return false;
            }
            const inputTensors = getTensorsInContainer(item.value);
            // Careful: the transform may mutate the item in place.
            // that's why we have to remember the input Tensors above, and then
            // below dispose only those that were not passed through to the output.
            // Note too that the transform function is responsible for tidying any
            // intermediate Tensors.  Here we are concerned only about the inputs.
            const mappedArray = this.transform(item.value);
            const outputTensors = getTensorsInContainer(mappedArray);
            this.outputQueue.pushAll(mappedArray);
            // TODO(soergel) faster intersection, and deduplicate outputTensors
            // TODO(soergel) move to tf.disposeExcept(in, out)?
            for (const t of inputTensors) {
                if (!isTensorInList(t, outputTensors)) {
                    t.dispose();
                }
            }
            return true;
        }
    }
    /**
     * Provides a `LazyIterator` that concatenates a stream of underlying
     * streams.
     *
     * Doing this in a concurrency-safe way requires some trickery.  In
     * particular, we want this stream to return the elements from the
     * underlying streams in the correct order according to when next() was
     * called, even if the resulting Promises resolve in a different order.
     */
    class ChainedIterator extends LazyIterator {
        constructor(iterators, baseErrorHandler) {
            super();
            this.baseErrorHandler = baseErrorHandler;
            // Strict Promise execution order:
            // a next() call may not even begin until the previous one completes.
            this.lastRead = null;
            // Local state that should not be clobbered by out-of-order execution.
            this.iterator = null;
            this.moreIterators = iterators;
        }
        summary() {
            const upstreamSummaries = 'TODO: fill in upstream of chained summaries';
            return `${upstreamSummaries} -> Chained`;
        }
        async next() {
            this.lastRead = this.readFromChain(this.lastRead);
            return this.lastRead;
        }
        async readFromChain(lastRead) {
            // Must await on the previous read since the previous read may have advanced
            // the stream of streams, from which we need to read.
            // This is unfortunate since we can't parallelize reads. Which means
            // prefetching of chained streams is a no-op.
            // One solution is to prefetch immediately upstream of this.
            await lastRead;
            if (this.iterator == null) {
                const iteratorResult = await this.moreIterators.next();
                if (iteratorResult.done) {
                    // No more streams to stream from.
                    return { value: null, done: true };
                }
                this.iterator = iteratorResult.value;
                if (this.baseErrorHandler != null) {
                    this.iterator = this.iterator.handleErrors(this.baseErrorHandler);
                }
            }
            const itemResult = await this.iterator.next();
            if (itemResult.done) {
                this.iterator = null;
                return this.readFromChain(lastRead);
            }
            return itemResult;
        }
    }
    var ZipMismatchMode;
    (function (ZipMismatchMode) {
        ZipMismatchMode[ZipMismatchMode["FAIL"] = 0] = "FAIL";
        ZipMismatchMode[ZipMismatchMode["SHORTEST"] = 1] = "SHORTEST";
        ZipMismatchMode[ZipMismatchMode["LONGEST"] = 2] = "LONGEST"; // use nulls for exhausted streams; use up the longest stream.
    })(ZipMismatchMode || (ZipMismatchMode = {}));
    /**
     * Provides a `LazyIterator` that zips together an array, dict, or nested
     * structure of `LazyIterator`s (and perhaps additional constants).
     *
     * The underlying streams must provide elements in a consistent order such
     * that they correspond.
     *
     * Typically, the underlying streams should have the same number of
     * elements. If they do not, the behavior is determined by the
     * `mismatchMode` argument.
     *
     * The nested structure of the `iterators` argument determines the
     * structure of elements in the resulting iterator.
     *
     * Doing this in a concurrency-safe way requires some trickery.  In
     * particular, we want this stream to return the elements from the
     * underlying streams in the correct order according to when next() was
     * called, even if the resulting Promises resolve in a different order.
     *
     * @param iterators: An array or object containing LazyIterators at the
     * leaves.
     * @param mismatchMode: Determines what to do when one underlying iterator
     * is exhausted before the others.  `ZipMismatchMode.FAIL` (the default)
     * causes an error to be thrown in this case.  `ZipMismatchMode.SHORTEST`
     * causes the zipped iterator to terminate with the furst underlying
     * streams, so elements remaining on the longer streams are ignored.
     * `ZipMismatchMode.LONGEST` causes the zipped stream to continue, filling
     * in nulls for the exhausted streams, until all streams are exhausted.
     */
    class ZipIterator extends LazyIterator {
        constructor(iterators, mismatchMode = ZipMismatchMode.FAIL) {
            super();
            this.iterators = iterators;
            this.mismatchMode = mismatchMode;
            this.count = 0;
            this.currentPromise = null;
        }
        summary() {
            const upstreamSummaries = 'TODO: fill in upstream of zip summaries';
            return `{${upstreamSummaries}} -> Zip`;
        }
        async nextState(afterState) {
            // This chaining ensures that the underlying next() are not even called
            // before the previous ones have resolved.
            await afterState;
            // Collect underlying iterator "done" signals as a side effect in
            // getNext()
            let numIterators = 0;
            let iteratorsDone = 0;
            function getNext(container) {
                if (container instanceof LazyIterator) {
                    const result = container.next();
                    return {
                        value: result.then(x => {
                            numIterators++;
                            if (x.done) {
                                iteratorsDone++;
                            }
                            return x.value;
                        }),
                        recurse: false
                    };
                }
                else {
                    return { value: null, recurse: true };
                }
            }
            const mapped = await deepMapAndAwaitAll(this.iterators, getNext);
            if (numIterators === iteratorsDone) {
                // The streams have all ended.
                return { value: null, done: true };
            }
            if (iteratorsDone > 0) {
                switch (this.mismatchMode) {
                    case ZipMismatchMode.FAIL:
                        throw new Error('Zipped streams should have the same length. ' +
                            `Mismatched at element ${this.count}.`);
                    case ZipMismatchMode.SHORTEST:
                        return { value: null, done: true };
                    case ZipMismatchMode.LONGEST:
                    // Continue.  The exhausted streams already produced value: null.
                }
            }
            this.count++;
            return { value: mapped, done: false };
        }
        async next() {
            this.currentPromise = this.nextState(this.currentPromise);
            return this.currentPromise;
        }
    }
    // Iterators that maintain a ring buffer of pending promises
    // ============================================================================
    /**
     * A stream that prefetches a given number of items from an upstream source,
     * returning them in FIFO order.
     *
     * Note this prefetches Promises, but makes no guarantees about when those
     * Promises resolve.
     */
    class PrefetchIterator extends LazyIterator {
        constructor(upstream, bufferSize) {
            super();
            this.upstream = upstream;
            this.bufferSize = bufferSize;
            this.buffer = new RingBuffer(bufferSize);
        }
        summary() {
            return `${this.upstream.summary()} -> Prefetch`;
        }
        /**
         * Refill the prefetch buffer.  Returns only after the buffer is full, or
         * the upstream source is exhausted.
         */
        refill() {
            while (!this.buffer.isFull()) {
                const v = this.upstream.next();
                this.buffer.push(v);
            }
        }
        next() {
            this.refill();
            // This shift will never throw an error because the buffer is always
            // full after a refill. If the stream is exhausted, the buffer will be
            // full of Promises that will resolve to the end-of-stream signal.
            return this.buffer.shift();
        }
    }
    /**
     * A stream that performs a sliding-window random shuffle on an upstream
     * source. This is like a `PrefetchIterator` except that the items are
     * returned in randomized order.  Mixing naturally improves as the buffer
     * size increases.
     */
    class ShuffleIterator extends PrefetchIterator {
        constructor(upstream, windowSize, seed) {
            super(upstream, windowSize);
            this.upstream = upstream;
            this.windowSize = windowSize;
            // Local state that should not be clobbered by out-of-order execution.
            this.upstreamExhausted = false;
            this.random = seedrandom.alea(seed || now().toString());
            this.lastRead = Promise.resolve({ value: null, done: false });
        }
        async next() {
            // This sets this.lastRead to a new Promise right away, as opposed to
            // saying `await this.lastRead; this.lastRead = this.serialNext();` which
            // would not work because this.nextRead would be updated only after the
            // promise resolves.
            this.lastRead = this.lastRead.then(() => this.serialNext());
            return this.lastRead;
        }
        randomInt(max) {
            return Math.floor(this.random() * max);
        }
        chooseIndex() {
            return this.randomInt(this.buffer.length());
        }
        async serialNext() {
            // TODO(soergel): consider performance
            if (!this.upstreamExhausted) {
                this.refill();
            }
            while (!this.buffer.isEmpty()) {
                const chosenIndex = this.chooseIndex();
                const result = await this.buffer.shuffleExcise(chosenIndex);
                if (result.done) {
                    this.upstreamExhausted = true;
                }
                else {
                    this.refill();
                    return result;
                }
            }
            return { value: null, done: true };
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    // TODO(soergel): consider vectorized operations within the pipeline.
    /**
     * Represents a potentially large list of independent data elements (typically
     * 'samples' or 'examples').
     *
     * A 'data example' may be a primitive, an array, a map from string keys to
     * values, or any nested structure of these.
     *
     * A `Dataset` represents an ordered collection of elements, together with a
     * chain of transformations to be performed on those elements. Each
     * transformation is a method of `Dataset` that returns another `Dataset`, so
     * these may be chained, e.g.
     * `const processedDataset = rawDataset.filter(...).map(...).batch(...)`.
     *
     * Data loading and transformation is done in a lazy, streaming fashion.  The
     * dataset may be iterated over multiple times; each iteration starts the data
     * loading anew and recapitulates the transformations.
     *
     * A `Dataset` is typically processed as a stream of unbatched examples --i.e.,
     * its transformations are applied one example at a time. Batching produces a
     * new `Dataset` where each element is a batch. Batching should usually come
     * last in a pipeline, because data transformations are easier to express on a
     * per-example basis than on a per-batch basis.
     *
     * The following code examples are calling `await dataset.forEachAsync(...)` to
     * iterate once over the entire dataset in order to print out the data.
     *
     * @doc {heading: 'Data', subheading: 'Classes', namespace: 'data'}
     */
    class Dataset {
        constructor() {
            this.size = null;
        }
        // TODO(soergel): Make Datasets report whether repeated iterator() calls
        // produce the same result (e.g., reading from a file) or different results
        // (e.g., from the webcam).  Currently we don't make this distinction but it
        // could be important for the user to know.
        // abstract isDeterministic(): boolean;
        /**
         * Groups elements into batches.
         *
         * It is assumed that each of the incoming dataset elements has the same
         * structure-- i.e. the same set of keys at each location in an object
         * hierarchy.  For each key, the resulting `Dataset` provides a batched
         * element collecting all of the incoming values for that key.
         *
         *  * Incoming primitives are grouped into a 1-D Tensor.
         *  * Incoming Tensors are grouped into a new Tensor where the 0'th axis is
         *    the batch dimension.
         *  * Incoming arrays are converted to Tensor and then batched.
         *  * A nested array is interpreted as an n-D Tensor, so the batched result
         *    has n+1 dimensions.
         *  * An array that cannot be converted to Tensor produces an error.
         *
         * If an array should not be batched as a unit, it should first be converted
         * to an object with integer keys.
         *
         * Here are a few examples:
         *
         * Batch a dataset of numbers:
         * ```js
         * const a = tf.data.array([1, 2, 3, 4, 5, 6, 7, 8]).batch(4);
         * await a.forEachAsync(e => e.print());
         * ```
         *
         * Batch a dataset of arrays:
         * ```js
         * const b = tf.data.array([[1], [2], [3], [4], [5], [6], [7], [8]]).batch(4);
         * await b.forEachAsync(e => e.print());
         * ```
         *
         * Batch a dataset of objects:
         * ```js
         * const c = tf.data.array([{a: 1, b: 11}, {a: 2, b: 12}, {a: 3, b: 13},
         *   {a: 4, b: 14}, {a: 5, b: 15}, {a: 6, b: 16}, {a: 7, b: 17},
         *   {a: 8, b: 18}]).batch(4);
         * await c.forEachAsync(e => {
         *   console.log('{');
         *   for(var key in e) {
         *     console.log(key+':');
         *     e[key].print();
         *   }
         *   console.log('}');
         * })
         * ```
         *
         * @param batchSize The number of elements desired per batch.
         * @param smallLastBatch Whether to emit the final batch when it has fewer
         *   than batchSize elements. Default true.
         * @returns A `Dataset`, from which a stream of batches can be obtained.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        batch(batchSize, smallLastBatch = true) {
            const base = this;
            assert$4(batchSize > 0, () => `batchSize needs to be positive, but it is
      ${batchSize}`);
            let size;
            if (this.size === Infinity || this.size == null) {
                // If the size of this dataset is infinity or null, the new size keeps the
                // same.
                size = this.size;
            }
            else if (smallLastBatch) {
                // If the size of this dataset is known and include small last batch, the
                // new size is full batch count plus last batch.
                size = Math.ceil(this.size / batchSize);
            }
            else {
                // If the size of this dataset is known and not include small last batch,
                // the new size is full batch count.
                size = Math.floor(this.size / batchSize);
            }
            return datasetFromIteratorFn(async () => {
                return (await base.iterator())
                    .columnMajorBatch(batchSize, smallLastBatch, deepBatchConcat);
            }, size);
        }
        /**
         * Concatenates this `Dataset` with another.
         *
         * ```js
         * const a = tf.data.array([1, 2, 3]);
         * const b = tf.data.array([4, 5, 6]);
         * const c = a.concatenate(b);
         * await c.forEachAsync(e => console.log(e));
         * ```
         *
         * @param dataset A `Dataset` to be concatenated onto this one.
         * @returns A `Dataset`.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        concatenate(dataset) {
            const base = this;
            let size;
            if (this.size === Infinity || dataset.size === Infinity) {
                // If the size of any of these two dataset is infinity, new size is
                // infinity.
                size = Infinity;
            }
            else if (this.size != null && dataset.size != null) {
                // If the size of both datasets are known and not infinity, new size is
                // sum the size of these two datasets.
                size = this.size + dataset.size;
            }
            else {
                // If neither of these two datasets has infinite size and any of these two
                // datasets' size is null, the new size is null.
                size = null;
            }
            return datasetFromIteratorFn(async () => (await base.iterator()).concatenate(await dataset.iterator()), size);
        }
        /**
         * Filters this dataset according to `predicate`.
         *
         * ```js
         * const a = tf.data.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
         *   .filter(x => x%2 === 0);
         * await a.forEachAsync(e => console.log(e));
         * ```
         *
         * @param predicate A function mapping a dataset element to a boolean or a
         * `Promise` for one.
         *
         * @returns A `Dataset` of elements for which the predicate was true.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        filter(predicate) {
            const base = this;
            let size;
            if (this.size === Infinity) {
                // If the size of this dataset is infinity, new size is infinity
                size = Infinity;
            }
            else {
                // If this dataset has limited elements, new size is null because it might
                // exhausted randomly.
                size = null;
            }
            return datasetFromIteratorFn(async () => {
                return (await base.iterator()).filter(x => tidy(() => predicate(x)));
            }, size);
        }
        /**
         * Apply a function to every element of the dataset.
         *
         * After the function is applied to a dataset element, any Tensors contained
         * within that element are disposed.
         *
         * ```js
         * const a = tf.data.array([1, 2, 3]);
         * await a.forEachAsync(e => console.log(e));
         * ```
         *
         * @param f A function to apply to each dataset element.
         * @returns A `Promise` that resolves after all elements have been processed.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        async forEachAsync(f) {
            return (await this.iterator()).forEachAsync(f);
        }
        /**
         * Maps this dataset through a 1-to-1 transform.
         *
         * ```js
         * const a = tf.data.array([1, 2, 3]).map(x => x*x);
         * await a.forEachAsync(e => console.log(e));
         * ```
         *
         * @param transform A function mapping a dataset element to a transformed
         *   dataset element.
         *
         * @returns A `Dataset` of transformed elements.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        map(transform) {
            const base = this;
            return datasetFromIteratorFn(async () => {
                return (await base.iterator()).map(x => tidy(() => transform(x)));
            }, this.size);
        }
        /**
         * Maps this dataset through an async 1-to-1 transform.
         *
         * ```js
         * const a =
         *  tf.data.array([1, 2, 3]).mapAsync(x => new Promise(function(resolve){
         *    setTimeout(() => {
         *      resolve(x * x);
         *    }, Math.random()*1000 + 500);
         *  }));
         * console.log(await a.toArray());
         * ```
         *
         * @param transform A function mapping a dataset element to a `Promise` for a
         *   transformed dataset element.  This transform is responsible for disposing
         *   any intermediate `Tensor`s, i.e. by wrapping its computation in
         *   `tf.tidy()`; that cannot be automated here (as it is in the synchronous
         *   `map()` case).
         *
         * @returns A `Dataset` of transformed elements.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        mapAsync(transform) {
            const base = this;
            return datasetFromIteratorFn(async () => {
                return (await base.iterator()).mapAsync(transform);
            }, this.size);
        }
        /**
         *  Creates a `Dataset` that prefetches elements from this dataset.
         *
         * @param bufferSize: An integer specifying the number of elements to be
         *   prefetched.
         * @returns A `Dataset`.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        prefetch(bufferSize) {
            if (bufferSize == null) {
                throw new RangeError('`Dataset.prefetch()` requires bufferSize to be specified.');
            }
            const base = this;
            return datasetFromIteratorFn(async () => (await base.iterator()).prefetch(bufferSize), this.size);
        }
        /**
         * Repeats this dataset `count` times.
         *
         * NOTE: If this dataset is a function of global state (e.g. a random number
         * generator), then different repetitions may produce different elements.
         *
         * ```js
         * const a = tf.data.array([1, 2, 3]).repeat(3);
         * await a.forEachAsync(e => console.log(e));
         * ```
         *
         * @param count: (Optional) An integer, representing the number of times
         *   the dataset should be repeated. The default behavior (if `count` is
         *   `undefined` or negative) is for the dataset be repeated indefinitely.
         * @returns A `Dataset`.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        repeat(count) {
            const base = this;
            let size;
            if (this.size != null && count > 0) {
                // If this dataset has size and count is positive, new size is current
                // size multiply count. This also covers the case that current size is
                // infinity.
                size = this.size * count;
            }
            else if (count === 0) {
                // If count is 0, new size is 0.
                size = 0;
            }
            else if (this.size != null && (count === undefined || count < 0)) {
                // If this dataset has size and count is undefined or negative, the
                // dataset will be repeated indefinitely and new size is infinity.
                size = Infinity;
            }
            else {
                // If the size of this dataset is null, the new dataset's size is null.
                size = null;
            }
            return datasetFromIteratorFn(async () => {
                const iteratorIterator = iteratorFromFunction(async () => ({ value: await base.iterator(), done: false }));
                return iteratorFromConcatenated(iteratorIterator.take(count));
            }, size);
        }
        /**
         * Creates a `Dataset` that skips `count` initial elements from this dataset.
         *
         * ```js
         * const a = tf.data.array([1, 2, 3, 4, 5, 6]).skip(3);
         * await a.forEachAsync(e => console.log(e));
         * ```
         *
         * @param count: The number of elements of this dataset that should be skipped
         *   to form the new dataset.  If `count` is greater than the size of this
         *   dataset, the new dataset will contain no elements.  If `count`
         *   is `undefined` or negative, skips the entire dataset.
         *
         * @returns A `Dataset`.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        skip(count) {
            const base = this;
            let size;
            if (this.size != null && count >= 0 && this.size >= count) {
                // If the size of this dataset is greater than count, the new dataset's
                // size is current size minus skipped size.This also covers the case that
                // current size is infinity.
                size = this.size - count;
            }
            else if (this.size != null &&
                (this.size < count || count === undefined || count < 0)) {
                // If the size of this dataset is smaller than count, or count is
                // undefined or negative, skips the entire dataset and the new size is 0.
                size = 0;
            }
            else {
                // If the size of this dataset is null, the new dataset's size is null.
                size = null;
            }
            return datasetFromIteratorFn(async () => (await base.iterator()).skip(count), size);
        }
        /**
         * Pseudorandomly shuffles the elements of this dataset. This is done in a
         * streaming manner, by sampling from a given number of prefetched elements.
         *
         * ```js
         * const a = tf.data.array([1, 2, 3, 4, 5, 6]).shuffle(3);
         * await a.forEachAsync(e => console.log(e));
         * ```
         *
         * @param bufferSize: An integer specifying the number of elements from this
         *   dataset from which the new dataset will sample.
         * @param seed: (Optional) An integer specifying the random seed that will
         *   be used to create the distribution.
         * @param reshuffleEachIteration: (Optional) A boolean, which if true
         *   indicates that the dataset should be pseudorandomly reshuffled each time
         *   it is iterated over. If false, elements will be returned in the same
         *   shuffled order on each iteration. (Defaults to `true`.)
         * @returns A `Dataset`.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        shuffle(bufferSize, seed, reshuffleEachIteration = true) {
            if (bufferSize == null || bufferSize < 0) {
                if (this.size == null) {
                    throw new RangeError('`Dataset.shuffle()` requires bufferSize to be specified.');
                }
                else {
                    throw new RangeError('`Dataset.shuffle()` requires bufferSize to be specified.  ' +
                        'If your data fits in main memory (for regular JS objects), ' +
                        'and/or GPU memory (for `tf.Tensor`s), consider setting ' +
                        `bufferSize to the dataset size (${this.size} elements)`);
                }
            }
            const base = this;
            const random = seedrandom.alea(seed || now().toString());
            return datasetFromIteratorFn(async () => {
                let seed2 = random.int32();
                if (reshuffleEachIteration) {
                    seed2 += random.int32();
                }
                return (await base.iterator()).shuffle(bufferSize, seed2.toString());
            }, this.size);
        }
        /**
         * Creates a `Dataset` with at most `count` initial elements from this
         * dataset.
         *
         * ```js
         * const a = tf.data.array([1, 2, 3, 4, 5, 6]).take(3);
         * await a.forEachAsync(e => console.log(e));
         * ```
         *
         * @param count: The number of elements of this dataset that should be taken
         *   to form the new dataset.  If `count` is `undefined` or negative, or if
         *   `count` is greater than the size of this dataset, the new dataset will
         *   contain all elements of this dataset.
         * @returns A `Dataset`.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        take(count) {
            const base = this;
            let size;
            if (this.size != null && this.size > count) {
                // If the size of this dataset is greater than count, the new dataset's
                // size is count.
                size = count;
            }
            else if (this.size != null && this.size <= count) {
                // If the size of this dataset is equal or smaller than count, the new
                // dataset's size is the size of this dataset.
                size = this.size;
            }
            else {
                // If the size of this dataset is null, the new dataset's size is null.
                size = null;
            }
            return datasetFromIteratorFn(async () => (await base.iterator()).take(count), size);
        }
        /**
         * Collect all elements of this dataset into an array.
         *
         * Obviously this will succeed only for small datasets that fit in memory.
         * Useful for testing and generally should be avoided if possible.
         *
         * ```js
         * const a = tf.data.array([1, 2, 3, 4, 5, 6]);
         * console.log(await a.toArray());
         * ```
         *
         * @returns A Promise for an array of elements, which will resolve
         *   when a new stream has been obtained and fully consumed.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        async toArray() {
            if (this.size === Infinity) {
                throw new Error('Can not convert infinite data stream to array.');
            }
            return (await this.iterator()).toArray();
        }
        /**
         * Collect all elements of this dataset into an array with prefetching 100
         * elements. This is useful for testing, because the prefetch changes the
         * order in which the Promises are resolved along the processing pipeline.
         * This may help expose bugs where results are dependent on the order of
         * Promise resolution rather than on the logical order of the stream (i.e.,
         * due to hidden mutable state).
         *
         * @returns A Promise for an array of elements, which will resolve
         *   when a new stream has been obtained and fully consumed.
         */
        async toArrayForTest() {
            if (this.size === Infinity) {
                throw new Error('Can not convert infinite data stream to array.');
            }
            return (await this.iterator()).toArrayForTest();
        }
    }
    // TODO(soergel): deep sharded shuffle, where supported
    Dataset.MAX_BUFFER_SIZE = 10000;
    /**
     * Create a `Dataset` defined by a provided iterator() function.
     *
     * ```js
     * let i = -1;
     * const func = () =>
     *    ++i < 5 ? {value: i, done: false} : {value: null, done: true};
     * const iter = tf.data.iteratorFromFunction(func);
     * const ds = tf.data.datasetFromIteratorFn(iter);
     * await ds.forEachAsync(e => console.log(e));
     * ```
     */
    function datasetFromIteratorFn(iteratorFn, size = null) {
        return new class extends Dataset {
            constructor() {
                super(...arguments);
                this.size = size;
            }
            /*
             * Provide a new stream of elements.  Note this will also start new streams
             * from any underlying `Dataset`s.
             */
            async iterator() {
                return iteratorFn();
            }
        }();
    }
    /**
     * Create a `Dataset` from an array of elements.
     *
     * Create a Dataset from an array of objects:
     * ```js
     * const a = tf.data.array([{'item': 1}, {'item': 2}, {'item': 3}]);
     * await a.forEachAsync(e => console.log(e));
     * ```
     *
     * Create a Dataset from an array of numbers:
     * ```js
     * const a = tf.data.array([4, 5, 6]);
     * await a.forEachAsync(e => console.log(e));
     * ```
     * @param items An array of elements that will be parsed as items in a dataset.
     *
     * @doc {heading: 'Data', subheading: 'Creation', namespace: 'data'}
     */
    function array(items) {
        return datasetFromIteratorFn(async () => iteratorFromItems(items), items.length);
    }
    /**
     * Create a `Dataset` by zipping together an array, dict, or nested
     * structure of `Dataset`s (and perhaps additional constants).
     * The underlying datasets must provide elements in a consistent order such that
     * they correspond.
     *
     * The number of elements in the resulting dataset is the same as the size of
     * the smallest dataset in datasets.
     *
     * The nested structure of the `datasets` argument determines the
     * structure of elements in the resulting iterator.
     *
     * Note this means that, given an array of two datasets that produce dict
     * elements, the result is a dataset that produces elements that are arrays
     * of two dicts:
     *
     * Zip an array of datasets:
     * ```js
     * console.log('Zip two datasets of objects:');
     * const ds1 = tf.data.array([{a: 1}, {a: 2}, {a: 3}]);
     * const ds2 = tf.data.array([{b: 4}, {b: 5}, {b: 6}]);
     * const ds3 = tf.data.zip([ds1, ds2]);
     * await ds3.forEachAsync(e => console.log(JSON.stringify(e)));
     *
     * // If the goal is to merge the dicts in order to produce elements like
     * // {a: ..., b: ...}, this requires a second step such as:
     * console.log('Merge the objects:');
     * const ds4 = ds3.map(x => {return {a: x[0].a, b: x[1].b}});
     * await ds4.forEachAsync(e => console.log(e));
     * ```
     *
     * Zip a dict of datasets:
     * ```js
     * const a = tf.data.array([{a: 1}, {a: 2}, {a: 3}]);
     * const b = tf.data.array([{b: 4}, {b: 5}, {b: 6}]);
     * const c = tf.data.zip({c: a, d: b});
     * await c.forEachAsync(e => console.log(JSON.stringify(e)));
     * ```
     *
     * @doc {heading: 'Data', subheading: 'Operations', namespace: 'data'}
     */
    function zip(datasets) {
        // manually type-check the argument for JS users
        if (!isIterable(datasets)) {
            throw new Error('The argument to zip() must be an object or array.');
        }
        let size;
        if (Array.isArray(datasets)) {
            for (let i = 0; i < datasets.length; i++) {
                size = size == null ? datasets[i].size :
                    Math.min(size, datasets[i].size);
            }
        }
        else if (datasets instanceof Object) {
            for (const ds in datasets) {
                size = size == null ? datasets[ds].size :
                    Math.min(size, datasets[ds].size);
            }
        }
        return datasetFromIteratorFn(async () => {
            const streams = await deepMapAndAwaitAll(datasets, d => {
                if (d instanceof Dataset) {
                    return { value: d.iterator(), recurse: false };
                }
                else if (isIterable(d)) {
                    return { value: null, recurse: true };
                }
                else {
                    throw new Error('Leaves of the structure passed to zip() must be Datasets, ' +
                        'not primitives.');
                }
            });
            return iteratorFromZipped(streams, ZipMismatchMode.SHORTEST);
        }, size);
    }
    /**
     * A zip function for use with deepZip, passed via the columnMajorBatch call.
     *
     * Accepts an array of identically-structured nested elements and either batches
     * them (if they are primitives, numeric arrays, or Tensors) or requests
     * recursion (if not).
     */
    // tslint:disable-next-line:no-any
    function deepBatchConcat(rows) {
        if (rows === null) {
            return null;
        }
        // use the first item to decide whether to recurse or batch here.
        const exampleRow = rows[0];
        if (canTensorify(exampleRow)) {
            // rows is an array of primitives, Tensors, or arrays.  Batch them.
            const value = batchConcat(rows);
            return { value, recurse: false };
        }
        // the example row is an object, so recurse into it.
        return { value: null, recurse: true };
    }
    /**
     * Assembles a list of same-shaped numbers, number arrays, or Tensors
     * into a single new Tensor where axis 0 is the batch dimension.
     */
    function batchConcat(arrays) {
        if (arrays.length === 0) {
            // We can't return an empty Tensor because we don't know the element shape.
            throw new Error('Can\'t make a batch of zero elements.');
        }
        if (arrays[0] instanceof Tensor) {
            // Input is an array of Tensors
            return stack(arrays);
        }
        else {
            // Input is a possibly-nested array of numbers.
            return tensor(arrays);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * Represents a potentially large collection of text lines.
     *
     * The results are not batched.
     */
    class TextLineDataset extends Dataset {
        /**
         * Create a `TextLineDataset`.
         *
         * @param input A `DataSource` providing a chunked, UTF8-encoded byte stream.
         */
        constructor(input) {
            super();
            this.input = input;
        }
        async iterator() {
            const inputIterator = await this.input.iterator();
            const utf8Iterator = inputIterator.decodeUTF8();
            const lineIterator = utf8Iterator.split('\n').map(line => {
                // Windows/DOS format text file has extra line breaker at the end of line.
                if (line.endsWith('\r')) {
                    line = line.slice(0, -1);
                }
                return line;
            });
            return lineIterator;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    const CODE_QUOTE = '"';
    const STATE_OUT = Symbol('out');
    const STATE_FIELD = Symbol('field');
    const STATE_QUOTE = Symbol('quote');
    const STATE_QUOTE_AFTER_QUOTE = Symbol('quoteafterquote');
    const STATE_WITHIN_QUOTE_IN_QUOTE = Symbol('quoteinquote');
    /**
     * Represents a potentially large collection of delimited text records.
     *
     * The produced `TensorContainer`s each contain one key-value pair for
     * every column of the table.  When a field is empty in the incoming data, the
     * resulting value is `undefined`, or throw error if it is required.  Values
     * that can be parsed as numbers are emitted as type `number`, other values
     * are parsed as `string`.
     *
     * The results are not batched.
     *
     * @doc {heading: 'Data', subheading: 'Classes', namespace: 'data'}
     */
    class CSVDataset extends Dataset {
        /**
         * Create a `CSVDataset`.
         *
         * @param input A `DataSource` providing a chunked, UTF8-encoded byte stream.
         * @param csvConfig (Optional) A CSVConfig object that contains configurations
         *     of reading and decoding from CSV file(s).
         *
         *     hasHeader: (Optional) A boolean value that indicates whether the first
         *     row of provided CSV file is a header line with column names, and should
         *     not be included in the data. Defaults to `true`.
         *
         *     columnNames: (Optional) A list of strings that corresponds to
         *     the CSV column names, in order. If provided, it ignores the column
         *     names inferred from the header row. If not provided, infers the column
         *     names from the first row of the records. If hasHeader is false and
         *     columnNames is not provided, this method throws an error.
         *
         *     columnConfigs: (Optional) A dictionary whose key is column names, value
         *     is an object stating if this column is required, column's data type,
         *     default value, and if this column is label. If provided, keys must
         *     correspond to names provided in columnNames or inferred from the file
         *     header lines. If isLabel is true any column, returns an array of two
         *     items: the first item is a dict of features key/value pairs, the second
         *     item is a dict of labels key/value pairs. If no feature is marked as
         *     label, returns a dict of features only.
         *
         *     configuredColumnsOnly (Optional) If true, only columns provided in
         *     columnConfigs will be parsed and provided during iteration.
         *
         *     delimiter (Optional) The string used to parse each line of the input
         *     file. Defaults to `,`.
         */
        constructor(input, csvConfig) {
            super();
            this.input = input;
            this.hasHeader = true;
            this.fullColumnNames = null;
            this.columnNamesValidated = false;
            this.columnConfigs = null;
            this.configuredColumnsOnly = false;
            this.delimiter = ',';
            this.delimWhitespace = false;
            this.base = new TextLineDataset(input);
            if (!csvConfig) {
                csvConfig = {};
            }
            this.hasHeader = csvConfig.hasHeader === false ? false : true;
            this.fullColumnNames = csvConfig.columnNames;
            this.columnConfigs = csvConfig.columnConfigs;
            this.configuredColumnsOnly = csvConfig.configuredColumnsOnly;
            if (csvConfig.delimWhitespace) {
                assert$4(csvConfig.delimiter == null, () => 'Delimiter should not be provided when delimWhitespace is true.');
                this.delimWhitespace = true;
                this.delimiter = ' ';
            }
            else {
                this.delimiter = csvConfig.delimiter ? csvConfig.delimiter : ',';
            }
        }
        /**
         * Returns column names of the csv dataset. If `configuredColumnsOnly` is
         * true, return column names in `columnConfigs`. If `configuredColumnsOnly` is
         * false and `columnNames` is provided, `columnNames`. If
         * `configuredColumnsOnly` is false and `columnNames` is not provided, return
         * all column names parsed from the csv file. For example usage please go to
         * `tf.data.csv`.
         *
         * @doc {heading: 'Data', subheading: 'Classes'}
         */
        async columnNames() {
            if (!this.columnNamesValidated) {
                await this.setColumnNames();
            }
            return this.configuredColumnsOnly ? Object.keys(this.columnConfigs) :
                this.fullColumnNames;
        }
        /* 1) If `columnNames` is provided as string[], use this string[] as output
         * keys in corresponding order. The length must match the number of inferred
         * columns if `hasHeader` is true .
         * 2) If `columnNames` is not provided, parse header line as `columnNames` if
         * hasHeader is true. If `hasHeader` is false, throw an error.
         * 3) If `columnConfigs` is provided, all the keys in `columnConfigs` must
         * exist in parsed `columnNames`.
         */
        async setColumnNames() {
            const columnNamesFromFile = await this.maybeReadHeaderLine();
            if (!this.fullColumnNames && !columnNamesFromFile) {
                // Throw an error if columnNames is not provided and no header line.
                throw new Error('Column names must be provided if there is no header line.');
            }
            else if (this.fullColumnNames && columnNamesFromFile) {
                // Check provided columnNames match header line.
                assert$4(columnNamesFromFile.length === this.fullColumnNames.length, () => 'The length of provided columnNames (' +
                    this.fullColumnNames.length.toString() +
                    ') does not match the length of the header line read from ' +
                    'file (' + columnNamesFromFile.length.toString() + ').');
            }
            if (!this.fullColumnNames) {
                this.fullColumnNames = columnNamesFromFile;
            }
            // Check if there are duplicate column names.
            const counts = this.fullColumnNames.reduce((countAcc, name) => {
                countAcc[name] = (countAcc[name] + 1) || 1;
                return countAcc;
            }, {});
            const duplicateNames = Object.keys(counts).filter((name) => (counts[name] > 1));
            assert$4(duplicateNames.length === 0, () => 'Duplicate column names found: ' + duplicateNames.toString());
            // Check if keys in columnConfigs match columnNames.
            if (this.columnConfigs) {
                for (const key of Object.keys(this.columnConfigs)) {
                    const index = this.fullColumnNames.indexOf(key);
                    if (index === -1) {
                        throw new Error('The key "' + key +
                            '" provided in columnConfigs does not match any of the column ' +
                            'names (' + this.fullColumnNames.toString() + ').');
                    }
                }
            }
            this.columnNamesValidated = true;
        }
        async maybeReadHeaderLine() {
            if (this.hasHeader) {
                const iter = await this.base.iterator();
                const firstElement = await iter.next();
                if (firstElement.done) {
                    throw new Error('No data was found for CSV parsing.');
                }
                const firstLine = firstElement.value;
                const headers = this.parseRow(firstLine, false);
                return headers;
            }
            else {
                return null;
            }
        }
        async iterator() {
            if (!this.columnNamesValidated) {
                await this.setColumnNames();
            }
            let lines = await this.base.iterator();
            if (this.hasHeader) {
                // We previously read the first line to get the columnNames.
                // Now that we're providing data, skip it.
                lines = lines.skip(1);
            }
            return lines.map(x => this.makeDataElement(x));
        }
        makeDataElement(line) {
            const values = this.parseRow(line);
            const features = {};
            const labels = {};
            for (let i = 0; i < this.fullColumnNames.length; i++) {
                const key = this.fullColumnNames[i];
                const config = this.columnConfigs ? this.columnConfigs[key] : null;
                if (this.configuredColumnsOnly && !config) {
                    // This column is not selected.
                    continue;
                }
                else {
                    const value = values[i];
                    let parsedValue = null;
                    if (value === '') {
                        // If default value is provided, use it. If default value is not
                        // provided, set as undefined.
                        if (config && config.default !== undefined) {
                            parsedValue = config.default;
                        }
                        else if (config && (config.required || config.isLabel)) {
                            throw new Error(`Required column ${key} is empty in this line: ${line}`);
                        }
                        else {
                            parsedValue = undefined;
                        }
                    }
                    else {
                        // A value is present, so parse it based on type
                        const valueAsNum = Number(value);
                        if (isNaN(valueAsNum)) {
                            // The value is a string and this column is declared as boolean
                            // in config, parse it as boolean.
                            if (config && config.dtype === 'bool') {
                                parsedValue = this.getBoolean(value);
                            }
                            else {
                                // Set value as string
                                parsedValue = value;
                            }
                        }
                        else if (!config || !config.dtype) {
                            // If this value is a number and no type config is provided, return
                            // it as number.
                            parsedValue = valueAsNum;
                        }
                        else {
                            // If this value is a number and data type is provided, parse it
                            // according to provided data type.
                            switch (config.dtype) {
                                case 'float32':
                                    parsedValue = valueAsNum;
                                    break;
                                case 'int32':
                                    parsedValue = Math.floor(valueAsNum);
                                    break;
                                case 'bool':
                                    parsedValue = this.getBoolean(value);
                                    break;
                                default:
                                    parsedValue = valueAsNum;
                            }
                        }
                    }
                    // Check if this column is label.
                    (config && config.isLabel) ? labels[key] = parsedValue :
                        features[key] = parsedValue;
                }
            }
            // If label exists, return an object of features and labels as {xs:features,
            // ys:labels}, otherwise return features only.
            if (Object.keys(labels).length === 0) {
                return features;
            }
            else {
                return { xs: features, ys: labels };
            }
        }
        getBoolean(value) {
            if (value === '1' || value.toLowerCase() === 'true') {
                return 1;
            }
            else {
                return 0;
            }
        }
        // adapted from https://beta.observablehq.com/@mbostock/streaming-csv
        parseRow(line, validateElementCount = true) {
            const result = [];
            let readOffset = 0;
            const readLength = line.length;
            let currentState = STATE_OUT;
            // Goes through the line to parse quote.
            for (let i = 0; i < readLength; i++) {
                switch (currentState) {
                    // Before enter a new field
                    case STATE_OUT:
                        switch (line.charAt(i)) {
                            // Enter a quoted field
                            case CODE_QUOTE:
                                readOffset = i + 1;
                                currentState = STATE_QUOTE;
                                break;
                            // Read an empty field
                            case this.delimiter:
                                readOffset = i + 1;
                                // If delimiter is white space and configured to collapse
                                // multiple white spaces, ignore this white space.
                                if (this.delimiter === ' ' && this.delimWhitespace) {
                                    break;
                                }
                                result.push('');
                                currentState = STATE_OUT;
                                break;
                            // Enter an unquoted field
                            default:
                                currentState = STATE_FIELD;
                                readOffset = i;
                                break;
                        }
                        break;
                    // In an unquoted field
                    case STATE_FIELD:
                        switch (line.charAt(i)) {
                            // Exit an unquoted field, add it to result
                            case this.delimiter:
                                result.push(line.substring(readOffset, i));
                                currentState = STATE_OUT;
                                readOffset = i + 1;
                                break;
                        }
                        break;
                    // In a quoted field
                    case STATE_QUOTE:
                        switch (line.charAt(i)) {
                            // Read a quote after a quote
                            case CODE_QUOTE:
                                currentState = STATE_QUOTE_AFTER_QUOTE;
                                break;
                        }
                        break;
                    // This state means it's right after a second quote in a field
                    case STATE_QUOTE_AFTER_QUOTE:
                        switch (line.charAt(i)) {
                            // Finished a quoted field
                            case this.delimiter:
                                result.push(line.substring(readOffset, i - 1));
                                currentState = STATE_OUT;
                                readOffset = i + 1;
                                break;
                            // Finished a quoted part in a quoted field
                            case CODE_QUOTE:
                                currentState = STATE_QUOTE;
                                break;
                            // In a quoted part in a quoted field
                            default:
                                currentState = STATE_WITHIN_QUOTE_IN_QUOTE;
                                break;
                        }
                        break;
                    case STATE_WITHIN_QUOTE_IN_QUOTE:
                        switch (line.charAt(i)) {
                            // Exit a quoted part in a quoted field
                            case CODE_QUOTE:
                                currentState = STATE_QUOTE;
                                break;
                        }
                        break;
                }
            }
            // Adds last item based on if it is quoted.
            if (currentState === STATE_QUOTE_AFTER_QUOTE) {
                result.push(line.substring(readOffset, readLength - 1));
            }
            else {
                result.push(line.substring(readOffset));
            }
            // Check if each row has the same number of elements as column names.
            if (validateElementCount && result.length !== this.fullColumnNames.length) {
                throw new Error(`Invalid row in csv file. Should have ${this.fullColumnNames.length} elements in a row, but got ${result}`);
            }
            return result;
        }
    }
    // TODO(soergel): add more basic datasets for parity with tf.data
    // tf.data.FixedLengthRecordDataset()
    // tf.data.TFRecordDataset()

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * Provide a stream of tensors from microphone audio stream. The tensors are
     * representing audio data as frequency-domain spectrogram generated with
     * browser's native FFT. Tensors representing time-domain waveform is available
     * based on configuration. Only works in browser environment.
     */
    class MicrophoneIterator extends LazyIterator {
        constructor(microphoneConfig) {
            super();
            this.microphoneConfig = microphoneConfig;
            this.isClosed = false;
            this.fftSize = microphoneConfig.fftSize || 1024;
            const fftSizeLog2 = Math.log2(this.fftSize);
            if (this.fftSize < 0 || fftSizeLog2 < 4 || fftSizeLog2 > 14 ||
                !Number.isInteger(fftSizeLog2)) {
                throw new Error(`Invalid fftSize: it must be a power of 2 between ` +
                    `2 to 4 and 2 to 14, but got ${this.fftSize}`);
            }
            this.numFrames = microphoneConfig.numFramesPerSpectrogram || 43;
            this.sampleRateHz = microphoneConfig.sampleRateHz;
            this.columnTruncateLength =
                microphoneConfig.columnTruncateLength || this.fftSize;
            this.audioTrackConstraints = microphoneConfig.audioTrackConstraints;
            this.smoothingTimeConstant = microphoneConfig.smoothingTimeConstant || 0;
            this.includeSpectrogram =
                microphoneConfig.includeSpectrogram === false ? false : true;
            this.includeWaveform =
                microphoneConfig.includeWaveform === true ? true : false;
            if (!this.includeSpectrogram && !this.includeWaveform) {
                throw new Error('Both includeSpectrogram and includeWaveform are false. ' +
                    'At least one type of data should be returned.');
            }
        }
        summary() {
            return `microphone`;
        }
        // Construct a MicrophoneIterator and start the audio stream.
        static async create(microphoneConfig = {}) {
            if (!env$1().get('IS_BROWSER')) {
                throw new Error('microphone API is only supported in browser environment.');
            }
            const microphoneIterator = new MicrophoneIterator(microphoneConfig);
            // Call async function start() to initialize the audio stream.
            await microphoneIterator.start();
            return microphoneIterator;
        }
        // Start the audio stream and FFT.
        async start() {
            try {
                this.stream = await navigator.mediaDevices.getUserMedia({
                    audio: this.audioTrackConstraints == null ? true :
                        this.audioTrackConstraints,
                    video: false
                });
            }
            catch (e) {
                throw new Error(`Error thrown while initializing video stream: ${e.message}`);
            }
            if (!this.stream) {
                throw new Error('Could not obtain audio from microphone.');
            }
            const ctxConstructor = 
            // tslint:disable-next-line:no-any
            window.AudioContext || window.webkitAudioContext;
            this.audioContext = new ctxConstructor();
            if (!this.sampleRateHz) {
                // If sample rate is not provided, use the available sample rate on
                // device.
                this.sampleRateHz = this.audioContext.sampleRate;
            }
            else if (this.audioContext.sampleRate !== this.sampleRateHz) {
                throw new Error(`Mismatch in sampling rate: ` +
                    `Expected: ${this.sampleRateHz}; ` +
                    `Actual: ${this.audioContext.sampleRate}`);
            }
            const streamSource = this.audioContext.createMediaStreamSource(this.stream);
            this.analyser = this.audioContext.createAnalyser();
            this.analyser.fftSize = this.fftSize * 2;
            this.analyser.smoothingTimeConstant = this.smoothingTimeConstant;
            streamSource.connect(this.analyser);
            this.freqData = new Float32Array(this.fftSize);
            this.timeData = new Float32Array(this.fftSize);
            return;
        }
        async next() {
            if (this.isClosed) {
                return { value: null, done: true };
            }
            let spectrogramTensor;
            let waveformTensor;
            const audioDataQueue = await this.getAudioData();
            if (this.includeSpectrogram) {
                const freqData = this.flattenQueue(audioDataQueue.freqDataQueue);
                spectrogramTensor = this.getTensorFromAudioDataArray(freqData, [this.numFrames, this.columnTruncateLength, 1]);
            }
            if (this.includeWaveform) {
                const timeData = this.flattenQueue(audioDataQueue.timeDataQueue);
                waveformTensor = this.getTensorFromAudioDataArray(timeData, [this.numFrames * this.fftSize, 1]);
            }
            return {
                value: { 'spectrogram': spectrogramTensor, 'waveform': waveformTensor },
                done: false
            };
        }
        // Capture one result from the audio stream, and extract the value from
        // iterator.next() result.
        async capture() {
            return (await this.next()).value;
        }
        async getAudioData() {
            const freqDataQueue = [];
            const timeDataQueue = [];
            let currentFrames = 0;
            return new Promise(resolve => {
                const intervalID = setInterval(() => {
                    if (this.includeSpectrogram) {
                        this.analyser.getFloatFrequencyData(this.freqData);
                        // If the audio stream is initializing, return empty queue.
                        if (this.freqData[0] === -Infinity) {
                            resolve({ freqDataQueue, timeDataQueue });
                        }
                        freqDataQueue.push(this.freqData.slice(0, this.columnTruncateLength));
                    }
                    if (this.includeWaveform) {
                        this.analyser.getFloatTimeDomainData(this.timeData);
                        timeDataQueue.push(this.timeData.slice());
                    }
                    // Clean interval and return when all frames have been collected
                    if (++currentFrames === this.numFrames) {
                        clearInterval(intervalID);
                        resolve({ freqDataQueue, timeDataQueue });
                    }
                }, this.fftSize / this.sampleRateHz * 1e3);
            });
        }
        // Stop the audio stream and pause the iterator.
        stop() {
            if (!this.isClosed) {
                this.isClosed = true;
                this.analyser.disconnect();
                this.audioContext.close();
                if (this.stream != null && this.stream.getTracks().length > 0) {
                    this.stream.getTracks()[0].stop();
                }
            }
        }
        // Override toArray() function to prevent collecting.
        toArray() {
            throw new Error('Can not convert infinite audio stream to array.');
        }
        // Return audio sampling rate in Hz
        getSampleRate() {
            return this.sampleRateHz;
        }
        flattenQueue(queue) {
            const frameSize = queue[0].length;
            const freqData = new Float32Array(queue.length * frameSize);
            queue.forEach((data, i) => freqData.set(data, i * frameSize));
            return freqData;
        }
        getTensorFromAudioDataArray(freqData, shape) {
            const vals = new Float32Array(sizeFromShape(shape));
            // If the data is less than the output shape, the rest is padded with zeros.
            vals.set(freqData, vals.length - freqData.length);
            return tensor(vals, shape);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * Provide a stream of image tensors from webcam video stream. Only works in
     * browser environment.
     */
    class WebcamIterator extends LazyIterator {
        constructor(webcamVideoElement, webcamConfig) {
            super();
            this.webcamVideoElement = webcamVideoElement;
            this.webcamConfig = webcamConfig;
            this.isClosed = true;
            this.resize = false;
            if (this.needToResize()) {
                this.resize = true;
                this.cropSize =
                    [this.webcamConfig.resizeHeight, this.webcamConfig.resizeWidth];
                this.cropBoxInd = tensor1d([0], 'int32');
                if (this.webcamConfig.centerCrop) {
                    // Calculate the box based on resizing shape.
                    const widthCroppingRatio = this.webcamConfig.resizeWidth * 1.0 / this.webcamVideoElement.width;
                    const heightCroppingRatio = this.webcamConfig.resizeHeight * 1.0 /
                        this.webcamVideoElement.height;
                    const widthCropStart = (1 - widthCroppingRatio) / 2;
                    const heightCropStart = (1 - heightCroppingRatio) / 2;
                    const widthCropEnd = widthCropStart + widthCroppingRatio;
                    const heightCropEnd = heightCroppingRatio + heightCropStart;
                    this.cropBox = tensor2d([heightCropStart, widthCropStart, heightCropEnd, widthCropEnd], [1, 4]);
                }
                else {
                    this.cropBox = tensor2d([0, 0, 1, 1], [1, 4]);
                }
            }
        }
        summary() {
            return `webcam`;
        }
        // Construct a WebcamIterator and start it's video stream.
        static async create(webcamVideoElement, webcamConfig = {}) {
            if (!env$1().get('IS_BROWSER')) {
                throw new Error('tf.data.webcam is only supported in browser environment.');
            }
            if (!webcamVideoElement) {
                // If webcam video element is not provided, create a hidden video element
                // with provided width and height.
                webcamVideoElement = document.createElement('video');
                if (!webcamConfig.resizeWidth || !webcamConfig.resizeHeight) {
                    throw new Error('Please provide webcam video element, or resizeWidth and ' +
                        'resizeHeight to create a hidden video element.');
                }
                webcamVideoElement.width = webcamConfig.resizeWidth;
                webcamVideoElement.height = webcamConfig.resizeHeight;
            }
            const webcamIterator = new WebcamIterator(webcamVideoElement, webcamConfig);
            // Call async function to initialize the video stream.
            await webcamIterator.start();
            return webcamIterator;
        }
        // Async function to start video stream.
        async start() {
            if (this.webcamConfig.facingMode) {
                assert$4((this.webcamConfig.facingMode === 'user') ||
                    (this.webcamConfig.facingMode === 'environment'), () => `Invalid webcam facing mode: ${this.webcamConfig.facingMode}. ` +
                    `Please provide 'user' or 'environment'`);
            }
            try {
                this.stream = await navigator.mediaDevices.getUserMedia({
                    video: {
                        deviceId: this.webcamConfig.deviceId,
                        facingMode: this.webcamConfig.facingMode ?
                            this.webcamConfig.facingMode :
                            'user',
                        width: this.webcamVideoElement.width,
                        height: this.webcamVideoElement.height
                    }
                });
            }
            catch (e) {
                // Modify the error message but leave the stack trace intact
                e.message = `Error thrown while initializing video stream: ${e.message}`;
                throw e;
            }
            if (!this.stream) {
                throw new Error('Could not obtain video from webcam.');
            }
            // Older browsers may not have srcObject
            try {
                this.webcamVideoElement.srcObject = this.stream;
            }
            catch (error) {
                console.log(error);
                this.webcamVideoElement.src = window.URL.createObjectURL(this.stream);
            }
            // Start the webcam video stream
            this.webcamVideoElement.play();
            this.isClosed = false;
            return new Promise(resolve => {
                // Add event listener to make sure the webcam has been fully initialized.
                this.webcamVideoElement.onloadedmetadata = () => {
                    resolve();
                };
            });
        }
        async next() {
            if (this.isClosed) {
                return { value: null, done: true };
            }
            let img;
            try {
                img = fromPixels$1(this.webcamVideoElement);
            }
            catch (e) {
                throw new Error(`Error thrown converting video to pixels: ${JSON.stringify(e)}`);
            }
            if (this.resize) {
                try {
                    return { value: this.cropAndResizeFrame(img), done: false };
                }
                catch (e) {
                    throw new Error(`Error thrown cropping the video: ${e.message}`);
                }
                finally {
                    img.dispose();
                }
            }
            else {
                return { value: img, done: false };
            }
        }
        needToResize() {
            // If resizeWidth and resizeHeight are provided, and different from the
            // width and height of original HTMLVideoElement, then resizing and cropping
            // is required.
            if (this.webcamConfig.resizeWidth && this.webcamConfig.resizeHeight &&
                (this.webcamVideoElement.width !== this.webcamConfig.resizeWidth ||
                    this.webcamVideoElement.height !== this.webcamConfig.resizeHeight)) {
                return true;
            }
            return false;
        }
        // Cropping and resizing each frame based on config
        cropAndResizeFrame(img) {
            return tidy(() => {
                const expandedImage = expandDims$3(cast$3(img, 'float32'), (0));
                let resizedImage;
                resizedImage = image$1.cropAndResize(expandedImage, this.cropBox, this.cropBoxInd, this.cropSize, 'bilinear');
                // Extract image from batch cropping.
                const shape = resizedImage.shape;
                return reshape$3(resizedImage, shape.slice(1));
            });
        }
        // Capture one frame from the video stream, and extract the value from
        // iterator.next() result.
        async capture() {
            return (await this.next()).value;
        }
        // Stop the video stream and pause webcam iterator.
        stop() {
            const tracks = this.stream.getTracks();
            tracks.forEach(track => track.stop());
            try {
                this.webcamVideoElement.srcObject = null;
            }
            catch (error) {
                console.log(error);
                this.webcamVideoElement.src = null;
            }
            this.isClosed = true;
        }
        // Override toArray() function to prevent collecting.
        toArray() {
            throw new Error('Can not convert infinite video stream to array.');
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * Represents a data source readable as a stream of binary data chunks.
     *
     * Because `Dataset`s can be read repeatedly (via `Dataset.iterator()`), this
     * provides a means to repeatedly create streams from the underlying data
     * sources.
     */
    class DataSource {
    }
    // TODO(soergel): consider convenience factory functions here
    // in combination with chainable source->dataset above, e.g.:
    // tf.data.url(...).asCsvDataset().shuffle().batch()

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    class StringIterator extends LazyIterator {
        /**
         * Splits a string stream on a given separator.
         *
         * It is assumed that the incoming chunk boundaries have no semantic meaning,
         * so conceptually the incoming stream is treated simply as the concatenation
         * of its elements.
         *
         * The outgoing stream provides chunks corresponding to the results of the
         * standard string split() operation (even if such a chunk spanned incoming
         * chunks).  The separators are not included.
         *
         * A typical usage is to split a text file (represented as a stream with
         * arbitrary chunk boundaries) into lines.
         *
         * @param upstream A readable stream of strings that can be treated as
         *   concatenated.
         * @param separator A character to split on.
         */
        split(separator) {
            return new SplitIterator(this, separator);
        }
    }
    // ============================================================================
    // The following private classes serve to implement the chainable methods
    // on StringIterator.  Unfortunately they can't be placed in separate files, due
    // to resulting trouble with circular imports.
    // ============================================================================
    // We wanted multiple inheritance, e.g.
    //   class SplitIterator extends QueueIterator<string>, StringIterator
    // but the TypeScript mixin approach is a bit hacky, so we take this adapter
    // approach instead.
    class SplitIterator extends StringIterator {
        constructor(upstream, separator) {
            super();
            this.upstream = upstream;
            this.impl = new SplitIteratorImpl(upstream, separator);
        }
        summary() {
            return this.impl.summary();
        }
        async next() {
            return this.impl.next();
        }
    }
    class SplitIteratorImpl extends OneToManyIterator {
        constructor(upstream, separator) {
            super();
            this.upstream = upstream;
            this.separator = separator;
            // A partial string at the end of an upstream chunk
            this.carryover = '';
        }
        summary() {
            return `${this.upstream.summary()} -> Split('${this.separator}')`;
        }
        async pump() {
            const chunkResult = await this.upstream.next();
            if (chunkResult.done) {
                if (this.carryover === '') {
                    return false;
                }
                // Pretend that the pump succeeded in order to emit the small last batch.
                // The next pump() call will actually fail.
                this.outputQueue.push(this.carryover);
                this.carryover = '';
                return true;
            }
            const lines = chunkResult.value.split(this.separator);
            // Note the behavior: " ab ".split(' ') === ['', 'ab', '']
            // Thus the carryover may be '' if the separator falls on a chunk
            // boundary; this produces the correct result.
            lines[0] = this.carryover + lines[0];
            for (const line of lines.slice(0, -1)) {
                this.outputQueue.push(line);
            }
            this.carryover = lines[lines.length - 1];
            return true;
        }
    }

    class ByteChunkIterator extends LazyIterator {
        /**
         * Decode a stream of UTF8-encoded byte arrays to a stream of strings.
         *
         * The byte arrays producetd from the ByteChunkIterator on which this is
         * called will be interpreted as concatenated.  No assumptions are made about
         * the boundaries of the incoming chunks, so a multi-byte UTF8 encoding of a
         * character may span the boundary between chunks.  This naturally happens,
         * for instance, when reading fixed-size byte arrays from a file.
         */
        decodeUTF8() {
            return new Utf8Iterator(this);
        }
    }
    // ============================================================================
    // The following private classes serve to implement the chainable methods
    // on ByteChunkIterator.  Unfortunately they can't be placed in separate files,
    // due to resulting trouble with circular imports.
    // ============================================================================
    // We wanted multiple inheritance, e.g.
    //   class Utf8Iterator extends QueueIterator<string>, StringIterator
    // but the TypeScript mixin approach is a bit hacky, so we take this adapter
    // approach instead.
    class Utf8Iterator extends StringIterator {
        constructor(upstream) {
            super();
            this.upstream = upstream;
            this.impl = new Utf8IteratorImpl(upstream);
        }
        summary() {
            return this.impl.summary();
        }
        async next() {
            return this.impl.next();
        }
    }
    /**
     * Decode a stream of UTF8-encoded byte arrays to a stream of strings.
     *
     * This is tricky because the incoming byte array boundaries may disrupt a
     * multi-byte UTF8 character. Thus any incomplete character data at the end of
     * a chunk must be carried over and prepended to the next chunk before
     * decoding. Luckily with native decoder, TextDecoder in browser and
     * string_decoder in node, byte array boundaries are handled automatically.
     *
     * In the context of an input pipeline for machine learning, UTF8 decoding is
     * needed to parse text files containing training examples or prediction
     * requests (e.g., formatted as CSV or JSON). We cannot use the built-in
     * decoding provided by FileReader.readAsText() because here we are in a
     * streaming context, which FileReader does not support.
     *
     * @param upstream A `LazyIterator` of `Uint8Arrays` containing UTF8-encoded
     *   text, which should be interpreted as concatenated.  No assumptions are
     *   made about the boundaries of the incoming chunks, so a multi-byte UTF8
     *   encoding of a character may span the boundary between chunks.  This
     *   naturally happens, for instance, when reading fixed-size byte arrays from a
     *   file.
     */
    class Utf8IteratorImpl extends OneToManyIterator {
        constructor(upstream) {
            super();
            this.upstream = upstream;
            if (env$1().get('IS_BROWSER')) {
                this.decoder = new TextDecoder('utf-8');
            }
            else {
                // tslint:disable-next-line:no-require-imports
                const { StringDecoder } = require('string_decoder');
                this.decoder = new StringDecoder('utf8');
            }
        }
        summary() {
            return `${this.upstream.summary()} -> Utf8`;
        }
        async pump() {
            const chunkResult = await this.upstream.next();
            let chunk;
            if (chunkResult.done) {
                return false;
            }
            else {
                chunk = chunkResult.value;
            }
            let text;
            if (env$1().get('IS_BROWSER')) {
                text = this.decoder.decode(chunk, { stream: true });
            }
            else {
                text = this.decoder.write(Buffer$1.from(chunk.buffer));
            }
            this.outputQueue.push(text);
            return true;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * Provide a stream of chunks from a File, Blob, or Uint8Array.
     * @param file The source File, Blob or Uint8Array.
     * @param options Optional settings controlling file reading.
     * @returns a lazy Iterator of Uint8Arrays containing sequential chunks of the
     *   input File, Blob or Uint8Array.
     */
    class FileChunkIterator extends ByteChunkIterator {
        constructor(file, options = {}) {
            super();
            this.file = file;
            this.options = options;
            assert$4((file instanceof Uint8Array) ||
                (env$1().get('IS_BROWSER') ?
                    (file instanceof File || file instanceof Blob) :
                    false), () => 'FileChunkIterator only supports File, Blob and Uint8Array ' +
                'right now.');
            this.offset = options.offset || 0;
            // default 1MB chunk has tolerable perf on large files
            this.chunkSize = options.chunkSize || 1024 * 1024;
        }
        summary() {
            return `FileChunks ${this.file}`;
        }
        async next() {
            if (this.offset >= ((this.file instanceof Uint8Array) ?
                this.file.byteLength :
                this.file.size)) {
                return { value: null, done: true };
            }
            const chunk = new Promise((resolve, reject) => {
                const end = this.offset + this.chunkSize;
                if (this.file instanceof Uint8Array) {
                    // Note if end > this.uint8Array.byteLength, we just get a small last
                    // chunk.
                    resolve(new Uint8Array(this.file.slice(this.offset, end)));
                }
                else {
                    // This branch assumes that this.file type is File or Blob, which
                    // means it is in the browser environment.
                    // TODO(soergel): is this a performance issue?
                    const fileReader = new FileReader();
                    fileReader.onload = (event) => {
                        let data = fileReader.result;
                        // Not sure we can trust the return type of
                        // FileReader.readAsArrayBuffer See e.g.
                        // https://github.com/node-file-api/FileReader/issues/2
                        if (data instanceof ArrayBuffer) {
                            data = new Uint8Array(data);
                        }
                        if (!(data instanceof Uint8Array)) {
                            return reject(new TypeError('FileReader returned unknown type.'));
                        }
                        resolve(data);
                    };
                    fileReader.onabort = (event) => {
                        return reject(new Error('Aborted'));
                    };
                    fileReader.onerror = (event) => {
                        return reject(new Error(event.type));
                    };
                    // TODO(soergel): better handle onabort, onerror
                    // Note if end > this.file.size, we just get a small last chunk.
                    const slice = this.file.slice(this.offset, end);
                    // We can't use readAsText here (even if we know the file is text)
                    // because the slice boundary may fall within a multi-byte character.
                    fileReader.readAsArrayBuffer(slice);
                }
                this.offset = end;
            });
            return { value: (await chunk), done: false };
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * Provide a stream of chunks from a URL.
     *
     * Note this class first downloads the entire file into memory before providing
     * the first element from the stream.  This is because the Fetch API does not
     * yet reliably provide a reader stream for the response body.
     */
    async function urlChunkIterator(url, options = {}, fetchFunc) {
        let urlString;
        let requestInit;
        if ((typeof url) === 'string') {
            urlString = url;
        }
        else {
            urlString = url.url;
            requestInit = getRequestInitFromRequest(url);
        }
        const response = await (fetchFunc || fetch$1)(urlString, requestInit);
        if (response.ok) {
            const uint8Array = new Uint8Array(await response.arrayBuffer());
            return new FileChunkIterator(uint8Array, options);
        }
        else {
            throw new Error(response.statusText);
        }
    }
    // Generate RequestInit from Request to match tf.util.fetch signature.
    const getRequestInitFromRequest = (request) => {
        const init = {
            method: request.method,
            headers: request.headers,
            body: request.body,
            mode: request.mode,
            credentials: request.credentials,
            cache: request.cache,
            redirect: request.redirect,
            referrer: request.referrer,
            integrity: request.integrity,
        };
        return init;
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    // Skip tslint any type check cause this method is aiming to check type of
    // input.
    // tslint:disable-next-line:no-any
    function isLocalPath(source) {
        return (typeof source === 'string') && source.slice(0, 7) === 'file://';
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * Represents a file, blob, or Uint8Array readable as a stream of binary data
     * chunks.
     */
    class FileDataSource extends DataSource {
        /**
         * Create a `FileDataSource`.
         *
         * @param input Local file path, or `File`/`Blob`/`Uint8Array` object to
         *     read. Local file only works in node environment.
         * @param options Options passed to the underlying `FileChunkIterator`s,
         *   such as {chunksize: 1024}.
         */
        constructor(input, options = {}) {
            super();
            this.input = input;
            this.options = options;
        }
        async iterator() {
            if (isLocalPath(this.input) && env$1().get('IS_NODE')) {
                // tslint:disable-next-line:no-require-imports
                const fs = require('fs');
                this.input = fs.readFileSync(this.input.slice(7));
            }
            // TODO(kangyizhang): Add LocalFileChunkIterator to split local streaming
            // with file in browser.
            return new FileChunkIterator(this.input, this.options);
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /*
     * Represents a URL readable as a stream of binary data chunks.
     */
    class URLDataSource extends DataSource {
        /**
         * Create a `URLDataSource`.
         *
         * @param url A source URL string, or a `Request` object.
         * @param options Options passed to the underlying `FileChunkIterator`s,
         *   such as {chunksize: 1024}.
         */
        constructor(url, fileOptions = {}) {
            super();
            this.url = url;
            this.fileOptions = fileOptions;
        }
        // TODO(soergel): provide appropriate caching options.  Currently this
        // will download the URL anew for each call to iterator().  Since we have
        // to treat the downloaded file as a blob/buffer anyway, we may as well retain
        // it-- but that raises GC issues.  Also we may want a persistent disk cache.
        async iterator() {
            if (isLocalPath(this.url)) {
                return (new FileDataSource(this.url, this.fileOptions))
                    .iterator();
            }
            else {
                return urlChunkIterator(this.url, this.fileOptions);
            }
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     *
     * =============================================================================
     */
    /**
     * Create a `CSVDataset` by reading and decoding CSV file(s) from provided URL
     * or local path if it's in Node environment.
     *
     * Note: If isLabel in columnConfigs is `true` for at least one column, the
     * element in returned `CSVDataset` will be an object of
     * `{xs:features, ys:labels}`: xs is a dict of features key/value pairs, ys
     * is a dict of labels key/value pairs. If no column is marked as label,
     * returns a dict of features only.
     *
     * ```js
     * const csvUrl =
     * 'https://storage.googleapis.com/tfjs-examples/multivariate-linear-regression/data/boston-housing-train.csv';
     *
     * async function run() {
     *   // We want to predict the column "medv", which represents a median value of
     *   // a home (in $1000s), so we mark it as a label.
     *   const csvDataset = tf.data.csv(
     *     csvUrl, {
     *       columnConfigs: {
     *         medv: {
     *           isLabel: true
     *         }
     *       }
     *     });
     *
     *   // Number of features is the number of column names minus one for the label
     *   // column.
     *   const numOfFeatures = (await csvDataset.columnNames()).length - 1;
     *
     *   // Prepare the Dataset for training.
     *   const flattenedDataset =
     *     csvDataset
     *     .map(({xs, ys}) =>
     *       {
     *         // Convert xs(features) and ys(labels) from object form (keyed by
     *         // column name) to array form.
     *         return {xs:Object.values(xs), ys:Object.values(ys)};
     *       })
     *     .batch(10);
     *
     *   // Define the model.
     *   const model = tf.sequential();
     *   model.add(tf.layers.dense({
     *     inputShape: [numOfFeatures],
     *     units: 1
     *   }));
     *   model.compile({
     *     optimizer: tf.train.sgd(0.000001),
     *     loss: 'meanSquaredError'
     *   });
     *
     *   // Fit the model using the prepared Dataset
     *   return model.fitDataset(flattenedDataset, {
     *     epochs: 10,
     *     callbacks: {
     *       onEpochEnd: async (epoch, logs) => {
     *         console.log(epoch + ':' + logs.loss);
     *       }
     *     }
     *   });
     * }
     *
     * await run();
     * ```
     *
     * @param source URL or local path to get CSV file. If it's a local path, it
     * must have prefix `file://` and it only works in node environment.
     * @param csvConfig (Optional) A CSVConfig object that contains configurations
     *     of reading and decoding from CSV file(s).
     *
     * @doc {
     *   heading: 'Data',
     *   subheading: 'Creation',
     *   namespace: 'data',
     *   configParamIndices: [1]
     *  }
     */
    function csv(source, csvConfig = {}) {
        return new CSVDataset(new URLDataSource(source), csvConfig);
    }
    /**
     * Create a `Dataset` that produces each element by calling a provided function.
     *
     * Note that repeated iterations over this `Dataset` may produce different
     * results, because the function will be called anew for each element of each
     * iteration.
     *
     * Also, beware that the sequence of calls to this function may be out of order
     * in time with respect to the logical order of the Dataset. This is due to the
     * asynchronous lazy nature of stream processing, and depends on downstream
     * transformations (e.g. .shuffle()). If the provided function is pure, this is
     * no problem, but if it is a closure over a mutable state (e.g., a traversal
     * pointer), then the order of the produced elements may be scrambled.
     *
     * ```js
     * let i = -1;
     * const func = () =>
     *    ++i < 5 ? {value: i, done: false} : {value: null, done: true};
     * const ds = tf.data.func(func);
     * await ds.forEachAsync(e => console.log(e));
     * ```
     *
     * @param f A function that produces one data element on each call.
     */
    function func(f) {
        const iter = iteratorFromFunction(f);
        return datasetFromIteratorFn(async () => iter);
    }
    /**
     * Create a `Dataset` that produces each element from provided JavaScript
     * generator, which is a function*
     * (https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Iterators_and_Generators#Generator_functions),
     * or a function that returns an
     * iterator
     * (https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Iterators_and_Generators#Generator_functions).
     *
     * The returned iterator should have `.next()` function that returns element in
     * format of `{value: TensorContainer, done:boolean}`.
     *
     * Example of creating a dataset from an iterator factory:
     * ```js
     * function makeIterator() {
     *   const numElements = 10;
     *   let index = 0;
     *
     *   const iterator = {
     *     next: () => {
     *       let result;
     *       if (index < numElements) {
     *         result = {value: index, done: false};
     *         index++;
     *         return result;
     *       }
     *       return {value: index, done: true};
     *     }
     *   };
     *   return iterator;
     * }
     * const ds = tf.data.generator(makeIterator);
     * await ds.forEachAsync(e => console.log(e));
     * ```
     *
     * Example of creating a dataset from a generator:
     * ```js
     * function* dataGenerator() {
     *   const numElements = 10;
     *   let index = 0;
     *   while (index < numElements) {
     *     const x = index;
     *     index++;
     *     yield x;
     *   }
     * }
     *
     * const ds = tf.data.generator(dataGenerator);
     * await ds.forEachAsync(e => console.log(e));
     * ```
     *
     * @param generator A Javascript generator function that returns a JavaScript
     *     iterator.
     *
     * @doc {
     *   heading: 'Data',
     *   subheading: 'Creation',
     *   namespace: 'data',
     *   configParamIndices: [1]
     *  }
     */
    function generator(generator) {
        return datasetFromIteratorFn(async () => {
            const gen = await generator();
            return iteratorFromFunction(() => gen.next());
        });
    }
    /**
     * Create an iterator that generate `Tensor`s from webcam video stream. This API
     * only works in Browser environment when the device has webcam.
     *
     * Note: this code snippet only works when the device has a webcam. It will
     * request permission to open the webcam when running.
     * ```js
     * const videoElement = document.createElement('video');
     * videoElement.width = 100;
     * videoElement.height = 100;
     * const cam = await tf.data.webcam(videoElement);
     * const img = await cam.capture();
     * img.print();
     * cam.stop();
     * ```
     *
     * @param webcamVideoElement A `HTMLVideoElement` used to play video from
     *     webcam. If this element is not provided, a hidden `HTMLVideoElement` will
     *     be created. In that case, `resizeWidth` and `resizeHeight` must be
     *     provided to set the generated tensor shape.
     * @param webcamConfig A `WebcamConfig` object that contains configurations of
     *     reading and manipulating data from webcam video stream.
     *
     * @doc {
     *   heading: 'Data',
     *   subheading: 'Creation',
     *   namespace: 'data',
     *   ignoreCI: true
     *  }
     */
    async function webcam(webcamVideoElement, webcamConfig) {
        return WebcamIterator.create(webcamVideoElement, webcamConfig);
    }
    /**
     * Create an iterator that generate frequency-domain spectrogram `Tensor`s from
     * microphone audio stream with browser's native FFT. This API only works in
     * browser environment when the device has microphone.
     *
     * Note: this code snippet only works when the device has a microphone. It will
     * request permission to open the microphone when running.
     * ```js
     * const mic = await tf.data.microphone({
     *   fftSize: 1024,
     *   columnTruncateLength: 232,
     *   numFramesPerSpectrogram: 43,
     *   sampleRateHz:44100,
     *   includeSpectrogram: true,
     *   includeWaveform: true
     * });
     * const audioData = await mic.capture();
     * const spectrogramTensor = audioData.spectrogram;
     * spectrogramTensor.print();
     * const waveformTensor = audioData.waveform;
     * waveformTensor.print();
     * mic.stop();
     * ```
     *
     * @param microphoneConfig A `MicrophoneConfig` object that contains
     *     configurations of reading audio data from microphone.
     *
     * @doc {
     *   heading: 'Data',
     *   subheading: 'Creation',
     *   namespace: 'data',
     *   ignoreCI: true
     *  }
     */
    async function microphone(microphoneConfig) {
        return MicrophoneIterator.create(microphoneConfig);
    }

    /** @license See the LICENSE file. */
    // This code is auto-generated, do not modify this file!
    const version$4 = '3.17.0';

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */

    var index = /*#__PURE__*/Object.freeze({
        __proto__: null,
        array: array,
        Dataset: Dataset,
        zip: zip,
        CSVDataset: CSVDataset,
        TextLineDataset: TextLineDataset,
        csv: csv,
        func: func,
        generator: generator,
        microphone: microphone,
        webcam: webcam,
        FileDataSource: FileDataSource,
        URLDataSource: URLDataSource,
        version_data: version$4
    });

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function assertNotComplex$1(tensor, opName) {
        if (!Array.isArray(tensor)) {
            tensor = [tensor];
        }
        tensor.forEach(t => {
            if (t != null) {
                assert$4(t.dtype !== 'complex64', () => `${opName} does not support complex64 tensors in the CPU backend.`);
            }
        });
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const whereImpl$1 = whereImpl$2;
    class MathBackendCPU extends KernelBackend {
        constructor() {
            super();
            this.blockSize = 48;
            this.firstUse = true;
            this.data = new DataStorage(this, engine());
        }
        nextDataId() {
            return MathBackendCPU.nextDataId++;
        }
        write(values, shape, dtype) {
            if (this.firstUse) {
                this.firstUse = false;
                if (env$1().get('IS_NODE')) {
                    warn('\n============================\n' +
                        'Hi, looks like you are running TensorFlow.js in ' +
                        'Node.js. To speed things up dramatically, install our node ' +
                        'backend, visit https://github.com/tensorflow/tfjs-node for more details. ' +
                        '\n============================');
                }
            }
            const dataId = { id: this.nextDataId() };
            this.data.set(dataId, { values, dtype, refCount: 1 });
            return dataId;
        }
        /**
         * Create a data bucket in cpu backend.
         * @param shape Shape of the `TensorInfo`.
         * @param dtype DType of the `TensorInfo`.
         * @param values The value of the `TensorInfo` stored as a flattened array.
         */
        makeTensorInfo(shape, dtype, values) {
            let outId;
            if (dtype === 'string' && values != null && values.length > 0 &&
                isString$2(values[0])) {
                const encodedValues = values.map(d => encodeString(d));
                outId = this.write(encodedValues, shape, dtype);
            }
            else {
                outId = this.write(values, shape, dtype);
            }
            return { dataId: outId, shape, dtype };
        }
        /** Return refCount of a `TensorData`. */
        refCount(dataId) {
            if (this.data.has(dataId)) {
                const tensorData = this.data.get(dataId);
                return tensorData.refCount;
            }
            return 0;
        }
        /** Increase refCount of a `TensorData`. */
        incRef(dataId) {
            const tensorData = this.data.get(dataId);
            tensorData.refCount++;
        }
        /** Decrease refCount of a `TensorData`. */
        decRef(dataId) {
            if (this.data.has(dataId)) {
                const tensorData = this.data.get(dataId);
                tensorData.refCount--;
            }
        }
        move(dataId, values, shape, dtype, refCount) {
            this.data.set(dataId, { values, dtype, refCount });
        }
        numDataIds() {
            return this.data.numDataIds();
        }
        async read(dataId) {
            return this.readSync(dataId);
        }
        readSync(dataId) {
            const { dtype, complexTensorInfos } = this.data.get(dataId);
            if (dtype === 'complex64') {
                const realValues = this.readSync(complexTensorInfos.real.dataId);
                const imagValues = this.readSync(complexTensorInfos.imag.dataId);
                return mergeRealAndImagArrays(realValues, imagValues);
            }
            return this.data.get(dataId).values;
        }
        bufferSync(t) {
            const data = this.readSync(t.dataId);
            if (t.dtype === 'string') {
                try {
                    // Decode the bytes into string.
                    const strings = data.map(d => decodeString(d));
                    return buffer(t.shape, t.dtype, strings);
                }
                catch (_a) {
                    throw new Error('Failed to decode encoded string bytes into utf-8');
                }
            }
            return buffer(t.shape, t.dtype, data);
        }
        makeOutput(values, shape, dtype) {
            return engine().makeTensorFromTensorInfo(this.makeTensorInfo(shape, dtype, values), this);
        }
        /**
         * Dispose the memory if the dataId has 0 refCount. Return true if the memory
         * is released or memory is not managed in this backend, false if memory is
         * not cleared.
         * @param dataId
         * @oaram force Optional, remove the data regardless of refCount
         */
        disposeData(dataId, force = false) {
            if (this.data.has(dataId)) {
                this.data.get(dataId).refCount--;
                if (!force && this.data.get(dataId).refCount > 0) {
                    return false;
                }
                const { complexTensorInfos } = this.data.get(dataId);
                if (complexTensorInfos != null) {
                    this.disposeData(complexTensorInfos.real.dataId, true);
                    this.disposeData(complexTensorInfos.imag.dataId, true);
                }
                this.data.delete(dataId);
            }
            return true;
        }
        disposeIntermediateTensorInfo(tensorInfo) {
            this.disposeData(tensorInfo.dataId);
        }
        async time(f) {
            const start = now();
            f();
            const kernelMs = now() - start;
            return { kernelMs };
        }
        memory() {
            return {
                // Unreliable due to automatic gc. The numbers above are cumulative.
                unreliable: true,
                reasons: ['The reported memory is an upper bound. Due to automatic garbage ' +
                        'collection, the true allocated memory may be less.']
            };
        }
        where(condition) {
            assertNotComplex$1([condition], 'where');
            const condVals = this.readSync(condition.dataId);
            return whereImpl$1(condition.shape, condVals);
        }
        dispose() { }
        floatPrecision() {
            return 32;
        }
        /** Returns the smallest representable number.  */
        epsilon() {
            return super.epsilon();
        }
    }
    MathBackendCPU.nextDataId = 0;

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function simpleAbsImpl(vals) {
        const resultValues = new Float32Array(vals.length);
        for (let i = 0; i < vals.length; ++i) {
            resultValues[i] = Math.abs(vals[i]);
        }
        return resultValues;
    }
    const abs$1 = (args) => {
        const { x } = args.inputs;
        const cpuBackend = args.backend;
        assertNotComplex$1(x, 'abs');
        let resultValues = new Float32Array(sizeFromShape(x.shape));
        const values = cpuBackend.data.get(x.dataId).values;
        resultValues = simpleAbsImpl(values);
        return cpuBackend.makeOutput(resultValues, x.shape, x.dtype);
    };
    const absConfig$1 = {
        kernelName: Abs,
        backendName: 'cpu',
        kernelFunc: abs$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Template that creates implementation for binary ops. Supports broadcast.
     */
    function createSimpleBinaryKernelImpl(op) {
        return (aShape, bShape, aVals, bVals, dtype) => {
            const newShape = assertAndGetBroadcastShape(aShape, bShape);
            const resultRank = newShape.length;
            const resultStrides = computeStrides(newShape);
            const resultSize = sizeFromShape(newShape);
            const result = getTypedArrayFromDType(dtype, resultSize);
            const aRank = aShape.length;
            const bRank = bShape.length;
            const aStrides = computeStrides(aShape);
            const bStrides = computeStrides(bShape);
            const aBroadcastDims = getBroadcastDims$1(aShape, newShape);
            const bBroadcastDims = getBroadcastDims$1(bShape, newShape);
            if (aBroadcastDims.length + bBroadcastDims.length === 0) {
                for (let i = 0; i < result.length; ++i) {
                    result[i] = op(aVals[i % aVals.length], bVals[i % bVals.length]);
                }
            }
            else {
                for (let i = 0; i < result.length; ++i) {
                    const loc = indexToLoc(i, resultRank, resultStrides);
                    const aLoc = loc.slice(-aRank);
                    aBroadcastDims.forEach(d => aLoc[d] = 0);
                    const aIndex = locToIndex(aLoc, aRank, aStrides);
                    const bLoc = loc.slice(-bRank);
                    bBroadcastDims.forEach(d => bLoc[d] = 0);
                    const bIndex = locToIndex(bLoc, bRank, bStrides);
                    result[i] = op(aVals[aIndex], bVals[bIndex]);
                }
            }
            return [result, newShape];
        };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function complex$1(args) {
        const { inputs, backend } = args;
        const { real, imag } = inputs;
        const realVals = backend.data.get(real.dataId).values;
        const imagVals = backend.data.get(imag.dataId).values;
        const complexInfo = backend.makeTensorInfo(real.shape, 'complex64');
        const complex = backend.data.get(complexInfo.dataId);
        // The complex tensor owns the underlying real and imag tensorInfos, only the
        // complex tensor tracks refCount, when complexData is disposed the
        // underlying tensorData will be disposed.
        complex.complexTensorInfos = {
            real: backend.makeTensorInfo(real.shape, 'float32', realVals),
            imag: backend.makeTensorInfo(imag.shape, 'float32', imagVals)
        };
        return complexInfo;
    }
    const complexConfig$1 = {
        kernelName: Complex,
        backendName: 'cpu',
        kernelFunc: complex$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Generates a tensorInfo with all zeros value.
     * @param backend cpu backend.
     * @param shape Shape for the zeros tensor.
     * @param dtype Optional. If set, the result has this dtype.
     */
    function zeros(backend, shape, dtype = 'float32') {
        if (dtype === 'complex64') {
            const real = zeros(backend, shape, 'float32');
            const imag = zeros(backend, shape, 'float32');
            return complex$1({ inputs: { real, imag }, backend });
        }
        const values = makeZerosTypedArray(sizeFromShape(shape), dtype);
        return backend.makeTensorInfo(shape, dtype, values);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function identity$1(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        backend.incRef(x.dataId);
        return { dataId: x.dataId, shape: x.shape, dtype: x.dtype };
    }
    const identityConfig$1 = {
        kernelName: Identity$1,
        backendName: 'cpu',
        kernelFunc: identity$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function real$1(args) {
        const { inputs, backend } = args;
        const { input } = inputs;
        const real = backend.data.get(input.dataId).complexTensorInfos.real;
        const realVal = backend.data.get(real.dataId).values;
        // When complex tensor is disposed, its underlying parts will be disposed too.
        // Make new tensor out of the real value of the complex. This makes sure the
        // value is still accessible even if complex tensor is disposed.
        return backend.makeTensorInfo(real.shape, real.dtype, realVal);
    }
    const realConfig$1 = {
        kernelName: Real,
        backendName: 'cpu',
        kernelFunc: real$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function cast$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { dtype } = attrs;
        // Casting to complex64.
        if (dtype === 'complex64') {
            if (x.dtype === 'complex64') {
                return identity$1({ inputs: { x }, backend });
            }
            const zerosTensorInfo = zeros(backend, x.shape, x.dtype);
            const floatX = cast$1({ inputs: { x }, backend, attrs: { dtype: 'float32' } });
            const result = complex$1({ inputs: { real: floatX, imag: zerosTensorInfo }, backend });
            backend.disposeIntermediateTensorInfo(zerosTensorInfo);
            backend.disposeIntermediateTensorInfo(floatX);
            return result;
        }
        // Casting from complex64
        if (x.dtype === 'complex64') {
            const realPart = real$1({ inputs: { input: x }, backend });
            const result = cast$1({ inputs: { x: realPart }, backend, attrs: { dtype } });
            backend.disposeIntermediateTensorInfo(realPart);
            return result;
        }
        if (!hasEncodingLoss(x.dtype, dtype)) {
            // We don't change the underlying data, since we cast to higher
            // precision.
            const result = identity$1({ inputs: { x }, backend });
            return { dataId: result.dataId, shape: result.shape, dtype };
        }
        if (dtype === 'int32') {
            const values = backend.data.get(x.dataId).values;
            const resultValues = Int32Array.from(values);
            return backend.makeTensorInfo(x.shape, 'int32', resultValues);
        }
        if (dtype === 'bool') {
            // This is essentially the result of notEqual(x, 0). We avoid using
            // kernel notEqual to avoid circular dependency, i.e. binary_utils ->
            // cast -> notEqual -> binary_utils.
            const xVals = backend.data.get(x.dataId).values;
            const zero = toTypedArray([0], x.dtype);
            const [resultData, resultShape] = createSimpleBinaryKernelImpl((a, b) => (a !== b) ? 1 : 0)(x.shape, [], xVals, zero, 'bool');
            return backend.makeTensorInfo(resultShape, 'bool', resultData);
        }
        throw new Error(`Error in Cast: failed to cast ${x.dtype} to ${dtype}`);
    }
    const castConfig$1 = {
        kernelName: Cast,
        backendName: 'cpu',
        kernelFunc: cast$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Template that creates a `KernelFunc` for binary ops.
     * @param name Kernel name.
     * @param binaryKernelImpl A `SimpleBinaryKernelImpl` for the kernel.
     * @param binaryKernelComplexImpl Optional. If exists, represents a
     *     `ComplexBinaryKernelImpl` for the kernel, will be used when input dtype
     *     is `complex64`.
     * @param dtype Optional. If set, the result has this dtype. Otherwise, the
     *     result has the same dtype as the first input. This is mainly used in
     *     comparison kernels, such as Equal, Less, Greater, etc.
     */
    function binaryKernelFunc$1(name, simpleImpl, complexImpl, dtype) {
        if (complexImpl == null) {
            return ({ inputs, backend }) => {
                const { a, b } = inputs;
                const cpuBackend = backend;
                assertNotComplex$1([a, b], name);
                const aVals = cpuBackend.data.get(a.dataId).values;
                const bVals = cpuBackend.data.get(b.dataId).values;
                const decodedAVals = a.dtype === 'string' ?
                    // tslint:disable-next-line: no-any
                    fromUint8ToStringArray(aVals) :
                    aVals;
                const decodedBVals = a.dtype === 'string' ?
                    // tslint:disable-next-line: no-any
                    fromUint8ToStringArray(bVals) :
                    bVals;
                const $dtype = dtype || a.dtype;
                const [resultData, resultShape] = simpleImpl(a.shape, b.shape, decodedAVals, decodedBVals, $dtype);
                return cpuBackend.makeTensorInfo(resultShape, $dtype, resultData);
            };
        }
        return ({ inputs, backend }) => {
            const { a, b } = inputs;
            const cpuBackend = backend;
            if (a.dtype === 'complex64' || b.dtype === 'complex64') {
                const $aComplex = cast$1({ inputs: { x: a }, backend: cpuBackend, attrs: { dtype: 'complex64' } });
                const $aComplexVals = cpuBackend.data.get($aComplex.dataId);
                const aReal = $aComplexVals.complexTensorInfos.real;
                const aImag = $aComplexVals.complexTensorInfos.imag;
                const aRealVals = cpuBackend.data.get(aReal.dataId).values;
                const aImagVals = cpuBackend.data.get(aImag.dataId).values;
                const $bComplex = cast$1({ inputs: { x: b }, backend: cpuBackend, attrs: { dtype: 'complex64' } });
                const $bComplexVals = cpuBackend.data.get($bComplex.dataId);
                const bReal = $bComplexVals.complexTensorInfos.real;
                const bImag = $bComplexVals.complexTensorInfos.imag;
                const bRealVals = cpuBackend.data.get(bReal.dataId).values;
                const bImagVals = cpuBackend.data.get(bImag.dataId).values;
                const [resultRealData, resultImagData, resultShape] = complexImpl(a.shape, b.shape, aRealVals, aImagVals, bRealVals, bImagVals);
                const resultReal = cpuBackend.makeTensorInfo(resultShape, 'float32', resultRealData);
                const resultImag = cpuBackend.makeTensorInfo(resultShape, 'float32', resultImagData);
                const result = complex$1({ inputs: { real: resultReal, imag: resultImag }, backend: cpuBackend });
                cpuBackend.disposeIntermediateTensorInfo($aComplex);
                cpuBackend.disposeIntermediateTensorInfo($bComplex);
                cpuBackend.disposeIntermediateTensorInfo(resultReal);
                cpuBackend.disposeIntermediateTensorInfo(resultImag);
                return result;
            }
            else {
                const aVals = cpuBackend.data.get(a.dataId).values;
                const bVals = cpuBackend.data.get(b.dataId).values;
                const $dtype = dtype || a.dtype;
                const [resultData, resultShape] = simpleImpl(a.shape, b.shape, aVals, bVals, $dtype);
                return cpuBackend.makeTensorInfo(resultShape, $dtype, resultData);
            }
        };
    }
    /**
     * Template that creates the complex type implementation for binary ops.
     * Supports broadcast.
     */
    function createComplexBinaryKernelImpl(op) {
        return (aShape, bShape, aRealVals, aImagVals, bRealVals, bImagVals) => {
            const resultShape = assertAndGetBroadcastShape(aShape, bShape);
            const resultSize = sizeFromShape(resultShape);
            const resultRank = resultShape.length;
            const resultStrides = computeStrides(resultShape);
            const resultRealVals = getTypedArrayFromDType('float32', resultSize);
            const resultImagVals = getTypedArrayFromDType('float32', resultSize);
            const aBroadcastDims = getBroadcastDims$1(aShape, resultShape);
            const bBroadcastDims = getBroadcastDims$1(bShape, resultShape);
            const aVals = mergeRealAndImagArrays(aRealVals, aImagVals);
            const bVals = mergeRealAndImagArrays(bRealVals, bImagVals);
            const aRank = aShape.length;
            const aStrides = computeStrides(aShape);
            const bRank = bShape.length;
            const bStrides = computeStrides(bShape);
            if (aBroadcastDims.length + bBroadcastDims.length === 0) {
                for (let i = 0; i < resultRealVals.length; i++) {
                    const aIdx = i % aVals.length;
                    const bIdx = i % bVals.length;
                    const result = op(aVals[aIdx * 2], aVals[aIdx * 2 + 1], bVals[bIdx * 2], bVals[bIdx * 2 + 1]);
                    resultRealVals[i] = result.real;
                    resultImagVals[i] = result.imag;
                }
            }
            else {
                for (let i = 0; i < resultRealVals.length; i++) {
                    const loc = indexToLoc(i, resultRank, resultStrides);
                    const aLoc = loc.slice(-aRank);
                    aBroadcastDims.forEach(d => aLoc[d] = 0);
                    const aIndex = locToIndex(aLoc, aRank, aStrides);
                    const bLoc = loc.slice(-bRank);
                    bBroadcastDims.forEach(d => bLoc[d] = 0);
                    const bIndex = locToIndex(bLoc, bRank, bStrides);
                    const opResult = op(aVals[aIndex * 2], aVals[aIndex * 2 + 1], bVals[bIndex * 2], bVals[bIndex * 2 + 1]);
                    resultRealVals[i] = opResult.real;
                    resultImagVals[i] = opResult.imag;
                }
            }
            return [resultRealVals, resultImagVals, resultShape];
        };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const addImpl = createSimpleBinaryKernelImpl(((a, b) => a + b));
    const addComplexImpl = createComplexBinaryKernelImpl(((aReal, aImag, bReal, bImag) => {
        return { real: aReal + bReal, imag: aImag + bImag };
    }));
    const add = binaryKernelFunc$1(Add$1, addImpl, addComplexImpl);
    const addConfig$1 = {
        kernelName: Add$1,
        backendName: 'cpu',
        kernelFunc: add
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function bincountImpl(xVals, weightsVals, weightsDtype, weightsShape, size) {
        const weightsSize = sizeFromShape(weightsShape);
        const outVals = makeZerosTypedArray(size, weightsDtype);
        for (let i = 0; i < xVals.length; i++) {
            const value = xVals[i];
            if (value < 0) {
                throw new Error('Input x must be non-negative!');
            }
            if (value >= size) {
                continue;
            }
            if (weightsSize > 0) {
                outVals[value] += weightsVals[i];
            }
            else {
                outVals[value] += 1;
            }
        }
        return outVals;
    }
    function bincountReduceImpl(xBuf, weightsBuf, size, binaryOutput = false) {
        const numRows = xBuf.shape[0];
        const numCols = xBuf.shape[1];
        const outBuf = buffer([numRows, size], weightsBuf.dtype);
        for (let i = 0; i < numRows; i++) {
            for (let j = 0; j < numCols; j++) {
                const value = xBuf.get(i, j);
                if (value < 0) {
                    throw new Error('Input x must be non-negative!');
                }
                if (value >= size) {
                    continue;
                }
                if (binaryOutput) {
                    outBuf.set(1, i, value);
                }
                else {
                    if (weightsBuf.size > 0) {
                        outBuf.set(outBuf.get(i, value) + weightsBuf.get(i, j), i, value);
                    }
                    else {
                        outBuf.set(outBuf.get(i, value) + 1, i, value);
                    }
                }
            }
        }
        return outBuf;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Template that creates implementation for unary op.
     */
    function createSimpleUnaryImpl(op) {
        return (values, dtype, attrs) => {
            const newValues = getTypedArrayFromDType(dtype, values.length);
            for (let i = 0; i < values.length; ++i) {
                newValues[i] = op(values[i], attrs);
            }
            return newValues;
        };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Template that creates a `KernelFunc` for unary ops.
     * @param name Kernel name.
     * @param op A `SimpleUnaryOperation` for the kernel.
     * @param dtype Optional. If set, the result has this dtype. Otherwise, the
     *     result has the same dtype as the input. This is mainly used in certain
     *     kernels that return bool type, such as isFinite, isInf, etc.
     */
    function unaryKernelFunc$1(name, op, dtype) {
        return ({ inputs, attrs, backend }) => {
            const { x } = inputs;
            assertNotComplex$1(x, name);
            if (x.dtype === 'string' || dtype === 'string') {
                throw new Error('unaryKernelFunc does not support string input/output');
            }
            const cpuBackend = backend;
            const values = cpuBackend.data.get(x.dataId).values;
            const xSize = sizeFromShape(x.shape);
            const $dtype = dtype || x.dtype;
            const newValues = getArrayFromDType($dtype, xSize);
            for (let i = 0; i < xSize; ++i) {
                newValues[i] = op(values[i], attrs);
            }
            return cpuBackend.makeTensorInfo(x.shape, $dtype, newValues);
        };
    }
    /**
     * Template that creates a `KernelFunc` for unary ops from the given
     * `SimpleUnaryImpl`..
     * @param name Kernel name.
     * @param unaryImpl A `SimpleUnaryImpl` that implements the op.
     * @param dtype Optional. If set, the result has this dtype. Otherwise, the
     *     result has the same dtype as the input. This is mainly used in certain
     *     kernels that return bool type, such as isFinite, isInf, etc.
     */
    function unaryKernelFuncFromImpl(name, unaryImpl, dtype) {
        return ({ inputs, attrs, backend }) => {
            const { x } = inputs;
            assertNotComplex$1(x, name);
            if (x.dtype === 'string' || dtype === 'string') {
                throw new Error('unaryKernelFunc does not support string input/output');
            }
            const cpuBackend = backend;
            const values = cpuBackend.data.get(x.dataId).values;
            const $dtype = dtype || x.dtype;
            const newValues = unaryImpl(values, $dtype, attrs);
            return cpuBackend.makeTensorInfo(x.shape, $dtype, newValues);
        };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ceilImpl = createSimpleUnaryImpl((xi) => Math.ceil(xi));
    const ceil$1 = unaryKernelFuncFromImpl(Ceil, ceilImpl);
    const ceilConfig$1 = {
        kernelName: Ceil,
        backendName: 'cpu',
        kernelFunc: ceil$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function concatImpl$1(inputs, outShape, dtype, simplyConcat) {
        const outVals = getArrayFromDType(dtype, sizeFromShape(outShape));
        if (simplyConcat && dtype !== 'string') {
            // Use built-in TypedArray.set() method for speed.
            let offset = 0;
            inputs.forEach(input => {
                const size = sizeFromShape(input.shape);
                outVals.set(input.vals, offset);
                offset += size;
            });
        }
        else {
            let colOffset = 0;
            inputs.forEach(input => {
                const decodedData = dtype === 'string' ?
                    fromUint8ToStringArray(input.vals) :
                    input.vals;
                let tIdx = 0;
                for (let row = 0; row < input.shape[0]; ++row) {
                    const resIdx = row * outShape[1] + colOffset;
                    for (let col = 0; col < input.shape[1]; ++col) {
                        outVals[resIdx + col] = decodedData[tIdx++];
                    }
                }
                colOffset += input.shape[1];
            });
        }
        return outVals;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const equalImpl = createSimpleBinaryKernelImpl((a, b) => (a === b) ? 1 : 0);
    const equal$2 = binaryKernelFunc$1(Equal, equalImpl, null /* complexImpl */, 'bool');
    const equalConfig$1 = {
        kernelName: Equal,
        backendName: 'cpu',
        kernelFunc: equal$2
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const expImpl = createSimpleUnaryImpl((xi) => Math.exp(xi));
    const exp$1 = unaryKernelFuncFromImpl(Exp, expImpl, 'float32');
    const expConfig$1 = {
        kernelName: Exp,
        backendName: 'cpu',
        kernelFunc: exp$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const expm1Impl = createSimpleUnaryImpl((xi) => Math.expm1(xi));
    const expm1$1 = unaryKernelFuncFromImpl(Expm1, expm1Impl);
    const expm1Config$1 = {
        kernelName: Expm1,
        backendName: 'cpu',
        kernelFunc: expm1$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const floorImpl = createSimpleUnaryImpl((xi) => Math.floor(xi));
    const floor$2 = unaryKernelFuncFromImpl(Floor, floorImpl);
    const floorConfig$1 = {
        kernelName: Floor,
        backendName: 'cpu',
        kernelFunc: floor$2,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function gatherNdImpl(indicesData, paramsBuf, dtype, numSlices, sliceRank, sliceSize, strides, paramsShape, paramsSize) {
        const outBuf = buffer([numSlices, sliceSize], dtype);
        for (let i = 0; i < numSlices; i++) {
            const index = [];
            let flattenIndex = 0;
            for (let j = 0; j < sliceRank; j++) {
                const dim = indicesData[i * sliceRank + j];
                flattenIndex += dim * strides[j];
                index.push(dim);
            }
            if (flattenIndex < 0 || flattenIndex >= paramsSize / sliceSize) {
                throw new Error(`Invalid indices: ${index} does not index into ${paramsShape}`);
            }
            for (let k = 0; k < sliceSize; k++) {
                outBuf.values[i * sliceSize + k] =
                    paramsBuf.get(...paramsBuf.indexToLoc(flattenIndex * sliceSize + k));
            }
        }
        return outBuf;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function gatherV2Impl(xBuf, indicesBuf, flattenOutputShape) {
        const outBuf = buffer(flattenOutputShape, xBuf.dtype);
        for (let i = 0; i < outBuf.size; ++i) {
            const newLoc = outBuf.indexToLoc(i);
            const originalLoc = newLoc.slice();
            const batchIdx = originalLoc[0];
            const indicesIdx = originalLoc[2];
            const indicesIndex = indicesBuf.locToIndex([batchIdx, indicesIdx]);
            originalLoc[2] = indicesBuf.values[indicesIndex];
            const originalIndex = xBuf.locToIndex(originalLoc);
            if (0 <= originalIndex && originalIndex < xBuf.values.length) {
                outBuf.values[i] = xBuf.values[originalIndex];
            } // Else, index is out of bounds, so leave the default zero val in outBuf.
        }
        return outBuf;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const greaterImpl = createSimpleBinaryKernelImpl((a, b) => (a > b) ? 1 : 0);
    const greater$1 = binaryKernelFunc$1(Greater, greaterImpl, null /* complexImpl */, 'bool');
    const greaterConfig$1 = {
        kernelName: Greater,
        backendName: 'cpu',
        kernelFunc: greater$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const greaterEqualImpl = createSimpleBinaryKernelImpl((a, b) => (a >= b) ? 1 : 0);
    const greaterEqual$1 = binaryKernelFunc$1(GreaterEqual, greaterEqualImpl, null /* complexImpl */, 'bool');
    const greaterEqualConfig$1 = {
        kernelName: GreaterEqual,
        backendName: 'cpu',
        kernelFunc: greaterEqual$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const lessImpl = createSimpleBinaryKernelImpl((a, b) => (a < b) ? 1 : 0);
    const less$1 = binaryKernelFunc$1(Less, lessImpl, null /* complexImpl */, 'bool');
    const lessConfig$1 = {
        kernelName: Less,
        backendName: 'cpu',
        kernelFunc: less$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const lessEqualImpl = createSimpleBinaryKernelImpl((a, b) => (a <= b) ? 1 : 0);
    const lessEqual$1 = binaryKernelFunc$1(LessEqual, lessEqualImpl, null /* complexImpl */, 'bool');
    const lessEqualConfig$1 = {
        kernelName: LessEqual,
        backendName: 'cpu',
        kernelFunc: lessEqual$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function linSpaceImpl(start, stop, num) {
        const step = (stop - start) / (num - 1);
        const values = makeZerosTypedArray(num, 'float32');
        values[0] = start;
        for (let i = 1; i < values.length; i++) {
            values[i] = values[i - 1] + step;
        }
        return values;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const logImpl = createSimpleUnaryImpl((xi) => Math.log(xi));
    const log$2 = unaryKernelFuncFromImpl(Log, logImpl);
    const logConfig$1 = {
        kernelName: Log,
        backendName: 'cpu',
        kernelFunc: log$2,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxImpl$1(aVals, reduceSize, outShape, dtype) {
        const vals = getTypedArrayFromDType(dtype, sizeFromShape(outShape));
        for (let i = 0; i < vals.length; ++i) {
            const offset = i * reduceSize;
            let max = aVals[offset];
            for (let j = 0; j < reduceSize; ++j) {
                const value = aVals[offset + j];
                if (Number.isNaN(value) ||
                    value > max) { // comparison with NaN always return false
                    max = value;
                }
            }
            vals[i] = max;
        }
        return vals;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const maximumImpl = createSimpleBinaryKernelImpl(((aValue, bValue) => Math.max(aValue, bValue)));
    const maximum$1 = binaryKernelFunc$1(Maximum$1, maximumImpl);
    const maximumConfig$1 = {
        kernelName: Maximum$1,
        backendName: 'cpu',
        kernelFunc: maximum$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const minimumImpl = createSimpleBinaryKernelImpl(((aValue, bValue) => Math.min(aValue, bValue)));
    const minimum$1 = binaryKernelFunc$1(Minimum$1, minimumImpl);
    const minimumConfig$1 = {
        kernelName: Minimum$1,
        backendName: 'cpu',
        kernelFunc: minimum$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const multiplyImpl = createSimpleBinaryKernelImpl(((aValue, bValue) => aValue * bValue));
    const multiplyComplexImpl = createComplexBinaryKernelImpl(((aReal, aImag, bReal, bImag) => {
        return {
            real: aReal * bReal - aImag * bImag,
            imag: aReal * bImag + aImag * bReal
        };
    }));
    const multiply$1 = binaryKernelFunc$1(Multiply$1, multiplyImpl, multiplyComplexImpl);
    const multiplyConfig$1 = {
        kernelName: Multiply$1,
        backendName: 'cpu',
        kernelFunc: multiply$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function negImpl(xVals, xShape, xDtype) {
        const minusOne = createScalarValue(-1, xDtype);
        return multiplyImpl([], xShape, minusOne, xVals, xDtype);
    }
    function neg$1(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        assertNotComplex$1(x, 'neg');
        const xVals = backend.data.get(x.dataId).values;
        const [res, newShape] = negImpl(xVals, x.shape, x.dtype);
        return backend.makeTensorInfo(newShape, x.dtype, res);
    }
    const negConfig$1 = {
        kernelName: Neg,
        backendName: 'cpu',
        kernelFunc: neg$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const notEqualImpl = createSimpleBinaryKernelImpl(((a, b) => (a !== b) ? 1 : 0));
    const notEqual$2 = binaryKernelFunc$1(NotEqual, notEqualImpl, null /* complexOp */, 'bool');
    const notEqualConfig$1 = {
        kernelName: NotEqual,
        backendName: 'cpu',
        kernelFunc: notEqual$2
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function transposeImpl$1(xVals, xShape, dtype, perm, newShape) {
        const xRank = xShape.length;
        const xSize = sizeFromShape(xShape);
        const xStrides = computeStrides(xShape);
        const newStrides = computeStrides(newShape);
        const result = getTypedArrayFromDType(dtype, sizeFromShape(newShape));
        for (let i = 0; i < xSize; ++i) {
            const loc = indexToLoc(i, xRank, xStrides);
            // Permute location.
            const newLoc = new Array(loc.length);
            for (let i = 0; i < newLoc.length; i++) {
                newLoc[i] = loc[perm[i]];
            }
            const newIndex = locToIndex(newLoc, xRank, newStrides);
            result[newIndex] = xVals[i];
        }
        return result;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function transpose$1(args) {
        const { inputs, attrs, backend } = args;
        const { x } = inputs;
        const { perm } = attrs;
        assertNotComplex$1(x, 'transpose');
        const xRank = x.shape.length;
        const newShape = new Array(xRank);
        for (let i = 0; i < newShape.length; i++) {
            newShape[i] = x.shape[perm[i]];
        }
        const values = backend.data.get(x.dataId).values;
        const result = transposeImpl$1(values, x.shape, x.dtype, perm, newShape);
        const dataId = backend.write(result, newShape, x.dtype);
        return { dataId, shape: newShape, dtype: x.dtype };
    }
    const transposeConfig$1 = {
        kernelName: Transpose,
        backendName: 'cpu',
        kernelFunc: transpose$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function prodImpl(xShape, xDtype, xVals, reductionAxes) {
        const [outShape, reduceShape] = computeOutAndReduceShapes(xShape, reductionAxes);
        const outDtype = upcastType(xDtype, 'int32');
        const outVals = makeZerosTypedArray(sizeFromShape(outShape), outDtype);
        const reduceSize = sizeFromShape(reduceShape);
        for (let i = 0; i < outVals.length; ++i) {
            const offset = i * reduceSize;
            let prod = 1;
            for (let j = 0; j < reduceSize; ++j) {
                prod *= xVals[offset + j];
            }
            outVals[i] = prod;
        }
        return { outVals, outShape, outDtype };
    }
    function prod$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        assertNotComplex$1(x, 'prod');
        const xRank = x.shape.length;
        const axes = parseAxisParam(axis, x.shape);
        const permutation = getAxesPermutation(axes, xRank);
        let reductionAxes = axes;
        let permutedX = x;
        const intermediateTensorInfos = [];
        if (permutation != null) {
            permutedX = transpose$1({ inputs: { x }, backend, attrs: { perm: permutation } });
            intermediateTensorInfos.push(permutedX);
            reductionAxes = getInnerMostAxes(reductionAxes.length, xRank);
        }
        const xVals = backend.data.get(permutedX.dataId).values;
        const { outVals, outShape, outDtype } = prodImpl(permutedX.shape, permutedX.dtype, xVals, reductionAxes);
        let resultShape = outShape;
        if (keepDims) {
            resultShape = expandShapeToKeepDim(outShape, axes);
        }
        intermediateTensorInfos.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return backend.makeTensorInfo(resultShape, outDtype, outVals);
    }
    const prodConfig$1 = {
        kernelName: Prod,
        backendName: 'cpu',
        kernelFunc: prod$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function rangeImpl(start, stop, step, dtype) {
        const sameStartStop = start === stop;
        const increasingRangeNegativeStep = start < stop && step < 0;
        const decreasingRangePositiveStep = stop < start && step > 1;
        if (sameStartStop || increasingRangeNegativeStep ||
            decreasingRangePositiveStep) {
            return makeZerosTypedArray(0, dtype);
        }
        const numElements = Math.abs(Math.ceil((stop - start) / step));
        const values = makeZerosTypedArray(numElements, dtype);
        if (stop < start && step === 1) {
            // Auto adjust the step's sign if it hasn't been set
            // (or was set to 1)
            step = -1;
        }
        values[0] = start;
        for (let i = 1; i < values.length; i++) {
            values[i] = values[i - 1] + step;
        }
        return values;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const rsqrtImpl = createSimpleUnaryImpl((xi) => 1 / Math.sqrt(xi));
    const rsqrt$1 = unaryKernelFuncFromImpl(Rsqrt, rsqrtImpl);
    const rsqrtConfig$1 = {
        kernelName: Rsqrt,
        backendName: 'cpu',
        kernelFunc: rsqrt$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function scatterImpl(indices, updates, shape, outputSize, sliceSize, numUpdates, sliceRank, strides, defaultValue, sumDupeIndices) {
        const flattenShape = [outputSize / sliceSize, sliceSize];
        const indicesData = indices.values;
        const updatesData = updates.values;
        if (outputSize === 0) {
            return buffer(shape, updates.dtype);
        }
        const outBuf = buffer(flattenShape, updates.dtype);
        if (typeof defaultValue === 'string') {
            outBuf.values.fill(defaultValue);
        }
        else if (typeof defaultValue === 'number') {
            outBuf.values.fill(defaultValue);
        }
        else if (typeof defaultValue === 'boolean') {
            outBuf.values.fill(+defaultValue);
        }
        for (let i = 0; i < numUpdates; i++) {
            const index = [];
            let flattenIndex = 0;
            for (let j = 0; j < sliceRank; j++) {
                const dim = indicesData[i * sliceRank + j];
                index.push(dim);
                flattenIndex += dim * strides[j];
            }
            if (flattenIndex < 0 || flattenIndex >= outputSize / sliceSize) {
                throw new Error(`Invalid indices: ${index} does not index into ${shape}`);
            }
            for (let k = 0; k < sliceSize; k++) {
                if (sumDupeIndices) {
                    outBuf.values[flattenIndex * sliceSize + k] +=
                        updatesData[i * sliceSize + k];
                }
                else {
                    outBuf.values[flattenIndex * sliceSize + k] = updates.rank === 0 ?
                        updatesData[0] :
                        updatesData[i * sliceSize + k];
                }
            }
        }
        return outBuf;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sigmoidImpl = createSimpleUnaryImpl((xi) => 1 / (1 + Math.exp(-xi)));
    const sigmoid$1 = unaryKernelFunc$1(Sigmoid$1, (xi) => 1 / (1 + Math.exp(-xi)));
    const sigmoidConfig$1 = {
        kernelName: Sigmoid$1,
        backendName: 'cpu',
        kernelFunc: sigmoid$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sliceImpl(vals, begin, size, shape, dtype) {
        const isContinous = isSliceContinous(shape, begin, size);
        const length = sizeFromShape(size);
        const xStrides = computeStrides(shape);
        if (isContinous) {
            const flatOffset = computeFlatOffset(begin, xStrides);
            if (dtype === 'string') {
                return vals.slice(flatOffset, flatOffset + length);
            }
            return vals.subarray(flatOffset, flatOffset + length);
        }
        const decodedData = dtype === 'string' ?
            fromUint8ToStringArray(vals) :
            vals;
        const inBuf = buffer(shape, dtype, decodedData);
        const outBuf = buffer(size, dtype);
        for (let i = 0; i < outBuf.size; ++i) {
            const outLoc = outBuf.indexToLoc(i);
            const inLoc = outLoc.map((idx, j) => idx + begin[j]);
            outBuf.set(inBuf.get(...inLoc), ...outLoc);
        }
        if (dtype === 'string') {
            return fromStringArrayToUint8(outBuf.values);
        }
        return outBuf.values;
    }
    function slice$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { begin, size } = attrs;
        assertNotComplex$1(x, 'slice');
        const [$begin, $size] = parseSliceParams(x, begin, size);
        assertParamsValid(x, $begin, $size);
        const vals = backend.data.get(x.dataId).values;
        const outVals = sliceImpl(vals, $begin, $size, x.shape, x.dtype);
        return backend.makeTensorInfo($size, x.dtype, outVals);
    }
    const sliceConfig$1 = {
        kernelName: Slice,
        backendName: 'cpu',
        kernelFunc: slice$1
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseFillEmptyRowsImpl(indices, indicesShape, indicesDType, values, valuesDType, denseShape, defaultValue) {
        const indicesCount = indicesShape[0];
        const denseRows = denseShape[0];
        const emptyRowIndicator = new Array(denseRows);
        const reverseIndexMap = new Array(indicesCount);
        const rank = indicesShape[1];
        if (denseRows === 0) {
            if (indicesCount !== 0) {
                throw new Error(getSparseFillEmptyRowsIndicesDenseShapeMismatch(indicesCount));
            }
            const outputIndices = getArrayFromDType(indicesDType, 0);
            const outputValues = getArrayFromDType(valuesDType, 0);
            return [
                outputIndices, [0, rank], outputValues, emptyRowIndicator, reverseIndexMap
            ];
        }
        let rowsAreOrdered = true;
        let lastIndicesRow = 0;
        const csrOffset = new Array(denseRows).fill(0);
        for (let i = 0; i < indicesCount; ++i) {
            // indices is a 2d tensor with shape of [N, rank]
            const row = indices[i * rank];
            if (row < 0) {
                throw new Error(getSparseFillEmptyRowsNegativeIndexErrorMessage(i, row));
            }
            if (row >= denseRows) {
                throw new Error(getSparseFillEmptyRowsOutOfRangeIndexErrorMessage(i, row, denseRows));
            }
            ++csrOffset[row];
            rowsAreOrdered = rowsAreOrdered && (row >= lastIndicesRow);
            lastIndicesRow = row;
        }
        let allRowsFull = true;
        for (let row = 0; row < denseRows; ++row) {
            // csrOffset here describes the number of elements in this dense row
            const rowEmpty = (csrOffset[row] === 0);
            emptyRowIndicator[row] = rowEmpty;
            allRowsFull = allRowsFull && !rowEmpty;
            // In filled version, each row has at least one element.
            csrOffset[row] = Math.max(csrOffset[row], 1);
            // Update csrOffset to represent the number of elements up to and
            // including denseRows + 1:
            //  csrOffset[0] == #{elements of row 0}
            //  csrOffset[1] == #{elements of row 1} + #{elements of row 0}
            //  ..
            //  csrOffset[i] == starting index for elements in row i + 1.
            if (row > 0) {
                csrOffset[row] += csrOffset[row - 1];
            }
        }
        if (allRowsFull && rowsAreOrdered) {
            const outputIndices = indices;
            const outputValues = values;
            for (let i = 0; i < indicesCount; ++i) {
                reverseIndexMap[i] = i;
            }
            return [
                outputIndices, [indicesCount, rank], outputValues, emptyRowIndicator,
                reverseIndexMap
            ];
        }
        else {
            const fullIndicesCount = csrOffset[denseRows - 1];
            const outputIndices = getArrayFromDType(indicesDType, fullIndicesCount * rank);
            const outputValues = getArrayFromDType(valuesDType, fullIndicesCount);
            const filledCount = new Array(denseRows).fill(0);
            // Fill in values for rows that are not missing
            for (let i = 0; i < indicesCount; ++i) {
                // indices is a 2d tensor with shape of [N, rank]
                const row = indices[i * rank];
                const offset = filledCount[row];
                const outputI = ((row === 0) ? 0 : csrOffset[row - 1]) + offset;
                filledCount[row]++; // Increment the filled count for this row.
                for (let j = 0; j < rank; ++j) {
                    // indices and outputIndices are 2d tensors with shape of [N, rank]
                    outputIndices[outputI * rank + j] = indices[i * rank + j];
                }
                outputValues[outputI] = values[i];
                // We'll need this reverse index map to backprop correctly.
                reverseIndexMap[i] = outputI;
            }
            // Fill in values for rows that are missing
            for (let row = 0; row < denseRows; ++row) {
                const rowCount = filledCount[row];
                if (rowCount === 0) { // We haven't filled this row
                    const startingIndex = (row === 0) ? 0 : csrOffset[row - 1];
                    // Remaining index values were set to zero already.
                    // Just need to set the row index in the right location.
                    // outputIndices is a 2d tensor with shape of [N, rank]
                    outputIndices[startingIndex * rank + 0] = row;
                    for (let col = 1; col < rank; ++col) {
                        outputIndices[startingIndex * rank + col] = 0;
                    }
                    outputValues[startingIndex] = defaultValue;
                }
            }
            return [
                outputIndices, [fullIndicesCount, rank], outputValues, emptyRowIndicator,
                reverseIndexMap
            ];
        }
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseReshapeImpl(inputIndices, inputIndicesShape, inputDType, inputShape, targetShape) {
        const denseSize = sizeFromShape(inputShape);
        const nnz = inputIndicesShape[0];
        const outputRank = targetShape.length;
        // Compute the output shape. Determine product of specified dimensions, and
        // find the index of the unspecified one.
        const outputShape = [];
        let product = 1;
        let unknownIndex = -1;
        for (let d = 0; d < outputRank; ++d) {
            const size = targetShape[d];
            if (size === -1) {
                if (unknownIndex !== -1) {
                    throw new Error(getSparseReshapeMultipleNegativeOneOutputDimErrorMessage(unknownIndex, d));
                }
                unknownIndex = d;
                outputShape.push(1);
            }
            else {
                if (size < 0) {
                    throw new Error(getSparseReshapeNegativeOutputDimErrorMessage(d, size));
                }
                product *= size;
                outputShape.push(size);
            }
        }
        if (unknownIndex !== -1) {
            if (product <= 0) {
                throw new Error(getSparseReshapeEmptyTensorZeroOutputDimErrorMessage());
            }
            const missing = Math.trunc(denseSize / product);
            if (product * missing !== denseSize) {
                throw new Error(getSparseReshapeInputOutputMultipleErrorMessage(inputShape, outputShape));
            }
            outputShape[unknownIndex] = missing;
        }
        const outputSize = sizeFromShape(outputShape);
        if (outputSize !== denseSize) {
            throw new Error(getSparseReshapeInputOutputMismatchErrorMessage(inputShape, outputShape));
        }
        const inputRank = inputShape.length;
        const inputStrides = [];
        if (inputRank > 0) {
            inputStrides[inputRank - 1] = 1;
            for (let d = inputRank - 2; d >= 0; --d) {
                inputStrides[d] = inputStrides[d + 1] * inputShape[d + 1];
            }
        }
        const outputStrides = [];
        if (outputRank > 0) {
            outputStrides[outputRank - 1] = 1;
            for (let d = outputRank - 2; d >= 0; --d) {
                outputStrides[d] = outputStrides[d + 1] * outputShape[d + 1];
            }
        }
        const newIndices = getArrayFromDType(inputDType, nnz * outputRank);
        for (let i = 0; i < nnz; ++i) {
            let id = 0;
            for (let j = 0; j < inputRank; ++j) {
                // inputIndices is a 2d tensor with shape of [nnz, inputRank]
                id += inputIndices[i * inputRank + j] * inputStrides[j];
            }
            for (let j = 0; j < outputRank; ++j) {
                // newIndices is a 2d tensor with shape of [nnz, outputRank]
                newIndices[i * outputRank + j] = Math.trunc(id / outputStrides[j]);
                id %= outputStrides[j];
            }
        }
        return [newIndices, [nnz, outputRank], outputShape];
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseSegmentReductionImpl(input, inputShape, inputDType, indices, segmentIds, isMean = false, defaultValue = 0) {
        const numIndices = indices.length;
        // Flatten the array to two dimensions
        const inputFlat = [inputShape[0], input.length / inputShape[0]];
        const numCol = inputFlat[1];
        // Note that the current implementation assumes that segmentIds values are
        // sorted.
        const lastSegmentIdPlusOne = numIndices > 0 ? segmentIds[numIndices - 1] + 1 : 0;
        const outputRows = lastSegmentIdPlusOne;
        if (outputRows < 0) {
            throw new Error(getSparseSegmentReductionNegativeSegmentIdsErrorMessage());
        }
        const outputShape = inputShape.slice();
        outputShape[0] = outputRows;
        const outputLength = outputShape.reduce((product, value) => product * value, 1);
        // Output array is initialized with the value 0 by default.
        const output = getArrayFromDType(inputDType, outputLength);
        // Note that we do not initialize the output buffer with a default value, so
        // we need to explicitly set missing indices to the default value.
        if (numIndices === 0) {
            if (outputRows > 0) {
                output.fill(defaultValue);
            }
            return [output, outputShape];
        }
        if (outputRows <= 0) {
            throw new Error(getSparseSegmentReductionNegativeSegmentIdsErrorMessage());
        }
        let start = 0, end = 1;
        // Index from which the output is not initialized.
        let uninitializedIndex = 0;
        let outIndex = segmentIds[start];
        while (true) {
            // We initialize nextIndex to 0 to avoid may be uninitialized warning
            let nextIndex = 0;
            if (end < numIndices) {
                nextIndex = segmentIds[end];
                if (outIndex === nextIndex) {
                    ++end;
                    continue;
                }
                // We have a new segment here.  Verify that the segment ids are growing.
                if (outIndex >= nextIndex) {
                    throw new Error(getSparseSegmentReductionNonIncreasingSegmentIdsErrorMessage());
                }
            }
            if (outIndex < 0 || outIndex >= outputRows) {
                throw new Error(getSparseSegmentReductionSegmentIdOutOfRangeErrorMessage(outIndex, outputRows));
            }
            // If there is a gap between two indices, we need to set that gap to the
            // default value.
            if (outIndex > uninitializedIndex) {
                output.fill(defaultValue, uninitializedIndex * numCol, outIndex * numCol);
            }
            for (let i = start; i < end; ++i) {
                const index = indices[i];
                if (index < 0 || index >= inputFlat[0]) {
                    throw new Error(getSparseSegmentReductionIndicesOutOfRangeErrorMessage(i, indices[i], inputFlat[0]));
                }
                for (let j = 0; j < numCol; j++) {
                    output[outIndex * numCol + j] += input[index * numCol + j];
                }
            }
            if (isMean) {
                for (let j = 0; j < numCol; j++) {
                    output[outIndex * numCol + j] /= end - start;
                }
            }
            start = end;
            ++end;
            uninitializedIndex = outIndex + 1;
            outIndex = nextIndex;
            if (end > numIndices) {
                break;
            }
        }
        // Fill the gap at the end with the default value.
        if (uninitializedIndex < outputRows) {
            output.fill(defaultValue, uninitializedIndex * numCol, outputRows * numCol);
        }
        return [output, outputShape];
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sqrtImpl = createSimpleUnaryImpl((xi) => Math.sqrt(xi));
    const sqrt$1 = unaryKernelFunc$1(Sqrt, (xi) => Math.sqrt(xi));
    const sqrtConfig$1 = {
        kernelName: Sqrt,
        backendName: 'cpu',
        kernelFunc: sqrt$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const squaredDifferenceImpl = createSimpleBinaryKernelImpl(((a, b) => {
        const diff = a - b;
        return diff * diff;
    }));
    const squaredDifference$1 = binaryKernelFunc$1(SquaredDifference, squaredDifferenceImpl);
    const squaredDifferenceConfig$1 = {
        kernelName: SquaredDifference,
        backendName: 'cpu',
        kernelFunc: squaredDifference$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stridedSliceImpl(outShape, xBuf, strides, begin) {
        const outBuf = buffer(outShape, xBuf.dtype);
        for (let i = 0; i < outBuf.size; i++) {
            const loc = outBuf.indexToLoc(i);
            const newLoc = new Array(loc.length);
            for (let j = 0; j < newLoc.length; j++) {
                newLoc[j] = loc[j] * strides[j] + begin[j];
            }
            outBuf.set(xBuf.get(...newLoc), ...loc);
        }
        return outBuf;
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * The StringNGramsOp class creates ngrams from ragged string data.
     * The constructor contains all attributes related to the operation such as
     * padding widths and strings, and the compute function can be used to
     * compute the ngrams for different ragged tensor inputs.
     */
    class StringNGramsOp {
        constructor(separator, nGramWidths, leftPad, rightPad, padWidth, preserveShortSequences) {
            this.separator = encodeString(separator);
            this.nGramWidths = nGramWidths;
            this.leftPad = encodeString(leftPad);
            this.rightPad = encodeString(rightPad);
            this.padWidth = padWidth;
            this.preserveShort = preserveShortSequences;
        }
        getPadWidth(nGramWidth) {
            // Ngrams can be padded with either a fixed pad width or a dynamic pad
            // width depending on the 'padWidth' arg, but in no case should the padding
            // ever be wider than 'nGramWidth' - 1.
            return Math.min(this.padWidth < 0 ? nGramWidth - 1 : this.padWidth, nGramWidth - 1);
        }
        getNumNGrams(length, nGramWidth) {
            const padWidth = this.getPadWidth(nGramWidth);
            return Math.max(0, ((length + 2 * padWidth) - nGramWidth) + 1);
        }
        createNGrams(data, splitIndex, output, outputStartIndex, numNGrams, nGramWidth) {
            for (let nGramIndex = 0; nGramIndex < numNGrams; ++nGramIndex) {
                const padWidth = this.getPadWidth(nGramWidth);
                const leftPadding = Math.max(0, padWidth - nGramIndex);
                const rightPadding = Math.max(0, padWidth - (numNGrams - (nGramIndex + 1)));
                const numTokens = nGramWidth - (leftPadding + rightPadding);
                const dataStartIndex = splitIndex + (leftPadding > 0 ? 0 : nGramIndex - padWidth);
                // Calculate the total expected size of the nGram so we can reserve the
                // correct amount of space in the string.
                let nGramSize = 0;
                // Size of the left padding.
                nGramSize += leftPadding * this.leftPad.length;
                // Size of the tokens.
                for (let n = 0; n < numTokens; ++n) {
                    nGramSize += data[dataStartIndex + n].length;
                }
                // Size of the right padding.
                nGramSize += rightPadding * this.rightPad.length;
                // Size of the separators.
                const numSeparators = leftPadding + rightPadding + numTokens - 1;
                nGramSize += numSeparators * this.separator.length;
                // Build the nGram.
                output[outputStartIndex + nGramIndex] = new Uint8Array(nGramSize);
                const nGram = output[outputStartIndex + nGramIndex];
                let nextNGramIndex = 0;
                const appendToNGram = (str) => str.forEach((value) => nGram[nextNGramIndex++] = value);
                for (let n = 0; n < leftPadding; ++n) {
                    appendToNGram(this.leftPad);
                    appendToNGram(this.separator);
                }
                // Only output first numTokens - 1 pairs of data and separator
                for (let n = 0; n < numTokens - 1; ++n) {
                    appendToNGram(data[dataStartIndex + n]);
                    appendToNGram(this.separator);
                }
                // Handle case when there are no tokens or no right padding as these
                // can result in consecutive separators.
                if (numTokens > 0) {
                    // If we have tokens, then output last and then pair each separator
                    // with the right padding that follows, to ensure nGram ends either with
                    // the token or with the right pad.
                    appendToNGram(data[dataStartIndex + numTokens - 1]);
                    for (let n = 0; n < rightPadding; ++n) {
                        appendToNGram(this.separator);
                        appendToNGram(this.rightPad);
                    }
                }
                else {
                    // If we don't have tokens, then the last item inserted into the nGram
                    // has been the separator from the left padding loop above. Hence,
                    // output right pad and separator and make sure to finish with a
                    // padding, not a separator.
                    for (let n = 0; n < rightPadding - 1; ++n) {
                        appendToNGram(this.rightPad);
                        appendToNGram(this.separator);
                    }
                    appendToNGram(this.rightPad);
                }
            }
        }
        // Data and splits together form the definition of the ragged tensor,
        // where data is 1 dimensional and contains the values of the tensor
        // and splits denotes the indices at which each row starts.
        compute(data, splits) {
            // Validate that the splits are valid indices into data, only if there are
            // splits specified.
            const inputDataSize = data.length;
            const splitsSize = splits.length;
            if (splitsSize > 0) {
                let prevSplit = splits[0];
                if (prevSplit !== 0) {
                    throw new Error(`First split value must be 0, got ${prevSplit}`);
                }
                for (let i = 1; i < splitsSize; ++i) {
                    let validSplits = splits[i] >= prevSplit;
                    validSplits = validSplits && (splits[i] <= inputDataSize);
                    if (!validSplits) {
                        throw new Error(`Invalid split value ${splits[i]}, must be in [${prevSplit}, ${inputDataSize}]`);
                    }
                    prevSplit = splits[i];
                }
                if (prevSplit !== inputDataSize) {
                    throw new Error(`Last split value must be data size. Expected ${inputDataSize}, got ${prevSplit}`);
                }
            }
            const numBatchItems = splitsSize - 1;
            const nGramsSplits = getArrayFromDType('int32', splitsSize);
            // If there is no data or size, return an empty ragged tensor.
            if (inputDataSize === 0 || splitsSize === 0) {
                const empty = new Array(inputDataSize);
                for (let i = 0; i <= numBatchItems; ++i) {
                    nGramsSplits[i] = 0;
                }
                return [empty, nGramsSplits];
            }
            nGramsSplits[0] = 0;
            for (let i = 1; i <= numBatchItems; ++i) {
                const length = splits[i] - splits[i - 1];
                let numNGrams = 0;
                this.nGramWidths.forEach((nGramWidth) => {
                    numNGrams += this.getNumNGrams(length, nGramWidth);
                });
                if (this.preserveShort && length > 0 && numNGrams === 0) {
                    numNGrams = 1;
                }
                nGramsSplits[i] = nGramsSplits[i - 1] + numNGrams;
            }
            const nGrams = new Array(nGramsSplits[numBatchItems]);
            for (let i = 0; i < numBatchItems; ++i) {
                const splitIndex = splits[i];
                let outputStartIdx = nGramsSplits[i];
                this.nGramWidths.forEach((nGramWidth) => {
                    const length = splits[i + 1] - splits[i];
                    const numNGrams = this.getNumNGrams(length, nGramWidth);
                    this.createNGrams(data, splitIndex, nGrams, outputStartIdx, numNGrams, nGramWidth);
                    outputStartIdx += numNGrams;
                });
                // If we're preserving short sequences, check to see if no sequence was
                // generated by comparing the current output start idx to the original
                // one (nGramSplitsdata). If no ngrams were generated, then they will
                // be equal (since we increment outputStartIdx by numNGrams every
                // time we create a set of ngrams.)
                if (this.preserveShort && outputStartIdx === nGramsSplits[i]) {
                    const dataLength = splits[i + 1] - splits[i];
                    // One legitimate reason to not have any ngrams when this.preserveShort
                    // is true is if the sequence itself is empty. In that case, move on.
                    if (dataLength === 0) {
                        continue;
                    }
                    // We don't have to worry about dynamic padding sizes here: if padding
                    // was dynamic, every sequence would have had sufficient padding to
                    // generate at least one nGram.
                    const nGramWidth = dataLength + 2 * this.padWidth;
                    const numNGrams = 1;
                    this.createNGrams(data, splitIndex, nGrams, outputStartIdx, numNGrams, nGramWidth);
                }
            }
            return [nGrams, nGramsSplits];
        }
    }
    function stringNGramsImpl(data, dataSplits, separator, nGramWidths, leftPad, rightPad, padWidth, preserveShortSequences) {
        return new StringNGramsOp(separator, nGramWidths, leftPad, rightPad, padWidth, preserveShortSequences)
            .compute(data, dataSplits);
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function split(str, delimiters, skipEmpty, result) {
        if (!str.length) {
            return;
        }
        // When the delimiter is empty, the input is split into individual characters.
        if (delimiters.length === 0) {
            for (let i = 0; i < str.length; ++i) {
                result.push(str.subarray(i, i + 1));
            }
            return;
        }
        // When there is one delimiter, the input is split only at that delimiter.
        if (delimiters.length === 1) {
            const delimiter = delimiters[0];
            let f = str.indexOf(delimiter);
            while (f !== -1) {
                const token = str.subarray(0, f);
                if (!skipEmpty || token.length !== 0) {
                    result.push(token);
                }
                str = str.subarray(f + 1);
                f = str.indexOf(delimiter);
            }
            if (!skipEmpty || str.length !== 0) {
                result.push(str);
            }
            return;
        }
        // When there are multiple delimiters, the input is split at every instance
        // one of the delimiters appears.
        let tokenStart = 0;
        for (let i = 0; i < str.length + 1; i++) {
            if ((i === str.length) || (delimiters.indexOf(str[i]) !== -1)) {
                const token = str.subarray(tokenStart, i);
                if (!skipEmpty || token.length !== 0) {
                    result.push(token);
                }
                tokenStart = i + 1;
            }
        }
    }
    function stringSplitImpl(input, delimiter, skipEmpty) {
        const batchSize = input.length;
        // Empty delimiter means split the input character by character.
        const tokens = [];
        let outputSize = 0;
        let maxNumEntries = 0;
        const numIndices = new Array(batchSize);
        for (let i = 0; i < batchSize; ++i) {
            const prevTokensLength = tokens.length;
            split(input[i], delimiter, skipEmpty, tokens);
            const nEntries = tokens.length - prevTokensLength;
            numIndices[i] = nEntries;
            outputSize += nEntries;
            maxNumEntries = Math.max(maxNumEntries, nEntries);
        }
        const indices = getArrayFromDType('int32', outputSize * 2);
        const values = new Array(outputSize);
        const shape = [batchSize, maxNumEntries];
        let c = 0;
        for (let i = 0; i < batchSize; ++i) {
            for (let j = 0; j < numIndices[i]; ++j) {
                // indices is a 2d tensor with shape of [outputSize, 2]
                indices[c * 2] = i;
                indices[c * 2 + 1] = j;
                values[c] = tokens[c];
                ++c;
            }
        }
        return [indices, values, shape];
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stringToHashBucketFastImpl(input, numBuckets) {
        const output = getArrayFromDType('int32', input.length);
        for (let i = 0; i < input.length; ++i) {
            output[i] =
                fingerPrint64(input[i]).modulo(numBuckets).getLowBitsUnsigned();
        }
        return output;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const subImpl = createSimpleBinaryKernelImpl(((aValue, bValue) => aValue - bValue));
    const subComplexImpl = createComplexBinaryKernelImpl(((aReal, aImag, bReal, bImag) => {
        return { real: aReal - bReal, imag: aImag - bImag };
    }));
    const sub$1 = binaryKernelFunc$1(Sub, subImpl, subComplexImpl);
    const subConfig$1 = {
        kernelName: Sub,
        backendName: 'cpu',
        kernelFunc: sub$1
    };

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * An implementation of the tile kernel shared between webgl and cpu for string
     * tensors only.
     */
    function tileImpl(xBuf, reps) {
        const newShape = new Array(xBuf.rank);
        for (let i = 0; i < newShape.length; i++) {
            newShape[i] = xBuf.shape[i] * reps[i];
        }
        const result = buffer(newShape, xBuf.dtype);
        for (let i = 0; i < result.values.length; ++i) {
            const newLoc = result.indexToLoc(i);
            const originalLoc = new Array(xBuf.rank);
            for (let j = 0; j < originalLoc.length; j++) {
                originalLoc[j] = newLoc[j] % xBuf.shape[j];
            }
            const originalIndex = xBuf.locToIndex(originalLoc);
            result.values[i] = xBuf.values[originalIndex];
        }
        return result;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const comparePair = (a, b) => {
        const valueDiff = b.value - a.value;
        return valueDiff === 0 ? a.index - b.index : valueDiff;
    };
    /**
     * Partitions array where all elements smaller than the (k+1) smallest element
     * are found to the left of it, and all larger to the right of it.
     * Based on the Floyd-Rivest Algorithm, ref:
     * https://en.wikipedia.org/wiki/Floyd%E2%80%93Rivest_algorithm
     * @param array: Array to partition
     * @param left: Left index for the interval
     * @param right: Right index for the interval
     * @param k: Desired index value, where array[k] is the (k+1)th smallest element
     *           when left = 0
     */
    function select$2(array, k, left = 0, right = array.length - 1) {
        while (right > left) {
            // Use select recursively to sample a smaller set of size s
            // the arbitrary constants 600 and 0.5 are used in the original
            // version to minimize execution time.
            if (right - left > 600) {
                const n = right - left + 1;
                const i = k - left + 1;
                const z = Math.log(n);
                const s = 0.5 * Math.exp(2 * z / 3);
                const sd = 0.5 * Math.sqrt(z * s * (n - s) / n) * Math.sign(i - n / 2);
                const newLeft = Math.max(left, Math.floor(k - i * s / n + sd));
                const newRight = Math.min(right, Math.floor(k + (n - i) * s / n + sd));
                select$2(array, k, newLeft, newRight);
            }
            // partition the elements between left and right around t
            const t = array[k];
            let i = left;
            let j = right;
            swap$1(array, left, k);
            if (comparePair(array[right], t) > 0) {
                swap$1(array, left, right);
            }
            while (i < j) {
                swap$1(array, i, j);
                i++;
                j--;
                while (comparePair(array[i], t) < 0) {
                    i = i + 1;
                }
                while (comparePair(array[j], t) > 0) {
                    j = j - 1;
                }
            }
            if (comparePair(array[left], t) === 0) {
                swap$1(array, left, j);
            }
            else {
                j = j + 1;
                swap$1(array, j, right);
            }
            // Adjust left and right towards the boundaries of the subset
            // containing the (k - left + 1)th smallest element.
            if (j <= k) {
                left = j + 1;
            }
            if (k <= j) {
                right = j - 1;
            }
        }
    }
    function topKImpl(x, xShape, xDtype, k, sorted) {
        // Reshape into a 2d tensor [batch, lastDim] and compute topk along lastDim.
        const lastDim = xShape[xShape.length - 1];
        const [batch, size] = [x.length / lastDim, lastDim];
        const allTopKVals = getTypedArrayFromDType(xDtype, batch * k);
        const allTopKIndices = getTypedArrayFromDType('int32', batch * k);
        for (let b = 0; b < batch; b++) {
            const offset = b * size;
            const vals = x.subarray(offset, offset + size);
            let valAndInd = new Array(vals.length);
            vals.forEach((value, index) => valAndInd[index] = { value, index });
            if (k < valAndInd.length) {
                select$2(valAndInd, k);
                valAndInd = valAndInd.slice(0, k);
            }
            if (sorted) {
                valAndInd.sort(comparePair);
            }
            const outOffset = b * k;
            const topKVals = allTopKVals.subarray(outOffset, outOffset + k);
            const topKIndices = allTopKIndices.subarray(outOffset, outOffset + k);
            for (let i = 0; i < k; i++) {
                topKVals[i] = valAndInd[i].value;
                topKIndices[i] = valAndInd[i].index;
            }
        }
        // Reshape back to the original input shape, except that the last
        // dimension is k.
        const outputShape = xShape.slice();
        outputShape[outputShape.length - 1] = k;
        return [
            buffer(outputShape, xDtype, allTopKVals),
            buffer(outputShape, 'int32', allTopKIndices)
        ];
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function uniqueImpl(values, axis, shape, dtype) {
        // Normalize and validate axis.
        const $axis = parseAxisParam(axis, shape)[0];
        // Calculate the new shape that is suitable for extracting data along the
        // given axis.
        //
        // The rank is 3.
        // The size of the 1st dimension is the size of all the axes < the given axis.
        // The size of the 2nd dimension is the same as the size of the given axis.
        // The size of the 3rd dimension is the size of all the axes > the given axis.
        //
        // For example, for a 4D tensor with shape=[2, 3, 5, 4] and axis=2, the
        // newShape would be: [2*3, 5, 4].
        //
        // Note that this is not the final output shape. This will be the shape for an
        // intermediate TensorBuffer (see inputBuffer below) to allow us to extract
        // values along the given axis. To demonstrate how it works, consider the
        // following example:
        //
        // Input: a 3D tensor, with shape [1, 2, 3]
        // [
        //   [
        //      [1,2,3],
        //      [4,5,6]
        //   ]
        // ]
        // Axis: 2 (the last axis).
        // Along axis 2, we expect to extract 3 tensors: [1,4], [2,5], [3,6].
        //
        // For this example, newShape would be: [2, 3, 1], where 2 is calculated from
        // 1*2. The re-shaped data would look like:
        //
        // [
        //   [
        //     [1], [2], [3]
        //   ],
        //   [
        //     [4], [5], [6]
        //   ]
        // ]
        //
        // Then, we can construct a 3-level nested loop by the following dimension
        // order to extract the values along the axis (dimension1):
        // i: dimension1       // 0,1,2 (newShape[1])
        //   m: dimension0     // 0,1   (newShape[0])
        //     n: dimension2   // 0     (newShape[2])
        //
        //                       m, i, n
        //                      ---------
        // Iteration 0: data at [0, 0, 0] => "1"
        // Iteration 1: data at [1, 0, 0] => "4"
        // We got [1,4].
        // Iteration 2: data at [0, 1, 0] => "2"
        // Iteration 3: data at [1, 1, 0] => "5"
        // We got [2,5].
        // Iteration 4: data at [0, 2, 0] => "3"
        // Iteration 5: data at [1, 2, 0] => "6"
        // We got [3,6].
        const newShape = [1, shape[0], 1];
        for (let i = 0; i < $axis; i++) {
            newShape[0] *= shape[i];
        }
        newShape[1] = shape[$axis];
        for (let i = $axis + 1; i < shape.length; i++) {
            newShape[2] *= shape[i];
        }
        // A map from unique elements (their string representations) to their values
        // in "indices" (below).
        const uniqueElements = {};
        // The indices of each unique element in the original tensor along the given
        // axis. It is 1D and has the same size as the given axis.
        const indices = new Int32Array(shape[$axis]);
        // Create a buffer so we can easily extract value at a given location.
        const inputBuffer = new TensorBuffer(newShape, dtype, values);
        // The indices along the given axis that have unique elements. This is a
        // de-duped version of "indices" above.
        const uniqueIndices = [];
        const is1DTensor = newShape[0] === 1 && newShape[2] === 1;
        for (let i = 0; i < shape[$axis]; i++) {
            // Extract values along the axis.
            let element;
            if (is1DTensor) {
                // Fast path for 1D tensor input.
                element = values[i].toString();
            }
            else {
                const axisValues = [];
                for (let m = 0; m < newShape[0]; m++) {
                    for (let n = 0; n < newShape[2]; n++) {
                        axisValues.push(inputBuffer.get(m, i, n));
                    }
                }
                element = axisValues.join(',');
            }
            // Dedup and update various indices.
            if (uniqueElements[element] !== undefined) {
                indices[i] = uniqueElements[element];
            }
            else {
                const uniqueIndex = Object.keys(uniqueElements).length;
                uniqueElements[element] = uniqueIndex;
                indices[i] = uniqueIndex;
                uniqueIndices.push(i);
            }
        }
        // Now we know where each of the unique elements are located along the axis
        // (uniqueIndices). Extract them from input buffer and store them in the
        // output buffer.
        const outputTmpShape = newShape.slice();
        outputTmpShape[1] = Object.keys(uniqueElements).length;
        const outputBuffer = new TensorBuffer(outputTmpShape, dtype);
        uniqueIndices.forEach((uniqueElementIndex, i) => {
            for (let m = 0; m < newShape[0]; m++) {
                for (let n = 0; n < newShape[2]; n++) {
                    outputBuffer.set(inputBuffer.get(m, uniqueElementIndex, n), m, i, n);
                }
            }
        });
        // The output shape can be calculated from the input shape with the size of
        // the given axis replaced by the number of unique elements along that axis.
        const outputShape = shape.slice();
        outputShape[$axis] = outputTmpShape[1];
        return {
            outputValues: outputBuffer.values,
            outputShape,
            indices,
        };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */

    var shared = /*#__PURE__*/Object.freeze({
        __proto__: null,
        simpleAbsImpl: simpleAbsImpl,
        addImpl: addImpl,
        bincountImpl: bincountImpl,
        bincountReduceImpl: bincountReduceImpl,
        ceilImpl: ceilImpl,
        concatImpl: concatImpl$1,
        equalImpl: equalImpl,
        expImpl: expImpl,
        expm1Impl: expm1Impl,
        floorImpl: floorImpl,
        gatherNdImpl: gatherNdImpl,
        gatherV2Impl: gatherV2Impl,
        greaterImpl: greaterImpl,
        greaterEqualImpl: greaterEqualImpl,
        lessImpl: lessImpl,
        lessEqualImpl: lessEqualImpl,
        linSpaceImpl: linSpaceImpl,
        logImpl: logImpl,
        maxImpl: maxImpl$1,
        maximumImpl: maximumImpl,
        minimumImpl: minimumImpl,
        multiplyImpl: multiplyImpl,
        negImpl: negImpl,
        notEqualImpl: notEqualImpl,
        prodImpl: prodImpl,
        rangeImpl: rangeImpl,
        rsqrtImpl: rsqrtImpl,
        scatterImpl: scatterImpl,
        sigmoidImpl: sigmoidImpl,
        sliceImpl: sliceImpl,
        sparseFillEmptyRowsImpl: sparseFillEmptyRowsImpl,
        sparseReshapeImpl: sparseReshapeImpl,
        sparseSegmentReductionImpl: sparseSegmentReductionImpl,
        sqrtImpl: sqrtImpl,
        squaredDifferenceImpl: squaredDifferenceImpl,
        stridedSliceImpl: stridedSliceImpl,
        stringNGramsImpl: stringNGramsImpl,
        stringSplitImpl: stringSplitImpl,
        stringToHashBucketFastImpl: stringToHashBucketFastImpl,
        subImpl: subImpl,
        tileImpl: tileImpl,
        topKImpl: topKImpl,
        transposeImpl: transposeImpl$1,
        uniqueImpl: uniqueImpl
    });

    /** @license See the LICENSE file. */
    // This code is auto-generated, do not modify this file!
    const version$3 = '3.17.0';

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Side effects for default initialization of MathBackendCPU
    registerBackend('cpu', () => new MathBackendCPU(), 1 /* priority */);

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const elu$1 = unaryKernelFunc$1(Elu$1, (xi) => xi >= 0 ? xi : (Math.exp(xi) - 1));
    const eluConfig$1 = {
        kernelName: Elu$1,
        backendName: 'cpu',
        kernelFunc: elu$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function leakyRelu$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { alpha } = attrs;
        assertNotComplex$1([x], 'leakyRelu');
        const xSize = sizeFromShape(x.shape);
        const xVals = backend.data.get(x.dataId).values;
        const outVals = getTypedArrayFromDType('float32', xSize);
        for (let i = 0; i < xVals.length; i++) {
            outVals[i] = xVals[i] < 0 ? alpha * xVals[i] : xVals[i];
        }
        return backend.makeTensorInfo(x.shape, 'float32', outVals);
    }
    const leakyReluConfig$1 = {
        kernelName: LeakyRelu,
        backendName: 'cpu',
        kernelFunc: leakyRelu$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const preluImpl = createSimpleBinaryKernelImpl((xValue, aValue) => xValue < 0 ? aValue * xValue : xValue);
    function prelu$1(args) {
        const { inputs, backend } = args;
        const { x, alpha } = inputs;
        assertNotComplex$1([x, alpha], 'prelu');
        const aVals = backend.data.get(x.dataId).values;
        const bVals = backend.data.get(alpha.dataId).values;
        const [resultData, resultShape] = preluImpl(x.shape, alpha.shape, aVals, bVals, 'float32');
        return backend.makeTensorInfo(resultShape, 'float32', resultData);
    }
    const preluConfig$1 = {
        kernelName: Prelu,
        backendName: 'cpu',
        kernelFunc: prelu$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const relu$1 = unaryKernelFunc$1(Relu$1, (xi) => Math.max(0, xi));
    const reluConfig$1 = {
        kernelName: Relu$1,
        backendName: 'cpu',
        kernelFunc: relu$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const relu6$1 = unaryKernelFunc$1(Relu6$1, (xi) => Math.min(Math.max(0, xi), 6));
    const relu6Config$1 = {
        kernelName: Relu6$1,
        backendName: 'cpu',
        kernelFunc: relu6$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function applyActivation(backend, x, activation, preluActivationWeights, leakyreluAlpha) {
        if (activation === 'linear') {
            return identity$1({ inputs: { x }, backend });
        }
        else if (activation === 'relu') {
            return relu$1({ inputs: { x }, backend });
        }
        else if (activation === 'elu') {
            return elu$1({ inputs: { x }, backend });
        }
        else if (activation === 'relu6') {
            return relu6$1({ inputs: { x }, backend });
        }
        else if (activation === 'prelu') {
            return prelu$1({ inputs: { x, alpha: preluActivationWeights }, backend });
        }
        else if (activation === 'leakyrelu') {
            return leakyRelu$1({ inputs: { x }, backend, attrs: { alpha: leakyreluAlpha } });
        }
        else if (activation === 'sigmoid') {
            return sigmoid$1({ inputs: { x }, backend });
        }
        throw new Error(`Activation ${activation} has not been implemented for the CPU backend.`);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function reshape$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { shape } = attrs;
        const xSize = sizeFromShape(x.shape);
        const $shape = inferFromImplicitShape(shape, xSize);
        const $xSize = sizeFromShape($shape);
        assert$4(xSize === $xSize, () => `The new shape (${$shape}) has ${$xSize} elements and the old ` +
            `shape (${x.shape}) has ${xSize} elements. The new shape and old ` +
            `shape must have the same number of elements.`);
        backend.incRef(x.dataId);
        const xData = backend.data.get(x.dataId);
        if (xData.complexTensorInfos != null) {
            const real = xData.complexTensorInfos.real;
            const imag = xData.complexTensorInfos.imag;
            real.shape = $shape;
            imag.shape = $shape;
        }
        return { dataId: x.dataId, shape: $shape, dtype: x.dtype };
    }
    const reshapeConfig$1 = {
        kernelName: Reshape$1,
        backendName: 'cpu',
        kernelFunc: reshape$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function batchMatMul$1(args) {
        const { inputs, backend, attrs } = args;
        const { a, b } = inputs;
        const { transposeA, transposeB } = attrs;
        assertNotComplex$1([a, b], 'matMul');
        const aRank = a.shape.length;
        const bRank = b.shape.length;
        const innerShapeA = transposeA ? a.shape[aRank - 2] : a.shape[aRank - 1];
        const innerShapeB = transposeB ? b.shape[bRank - 1] : b.shape[bRank - 2];
        const outerShapeA = transposeA ? a.shape[aRank - 1] : a.shape[aRank - 2];
        const outerShapeB = transposeB ? b.shape[bRank - 2] : b.shape[bRank - 1];
        const outerDimsA = a.shape.slice(0, -2);
        const outerDimsB = b.shape.slice(0, -2);
        const batchDimA = sizeFromShape(outerDimsA);
        const batchDimB = sizeFromShape(outerDimsB);
        const outShapeOuterDims = assertAndGetBroadcastShape(a.shape.slice(0, -2), b.shape.slice(0, -2));
        const outShape = outShapeOuterDims.concat([outerShapeA, outerShapeB]);
        assert$4(innerShapeA === innerShapeB, () => `Error in matMul: inner shapes (${innerShapeA}) and (` +
            `${innerShapeB}) of Tensors with shapes ${a.shape} and ` +
            `${b.shape} and transposeA=${transposeA}` +
            ` and transposeB=${transposeB} must match.`);
        const a3dShape = transposeA ? [batchDimA, innerShapeA, outerShapeA] :
            [batchDimA, outerShapeA, innerShapeA];
        const b3dShape = transposeB ? [batchDimB, outerShapeB, innerShapeB] :
            [batchDimB, innerShapeB, outerShapeB];
        // The rest of the implementation is designed to operate on rank-3 tensors
        const a3d = reshape$1({ inputs: { x: a }, backend, attrs: { shape: a3dShape } });
        const b3d = reshape$1({ inputs: { x: b }, backend, attrs: { shape: b3dShape } });
        const sharedDim = transposeA ? a3d.shape[1] : a3d.shape[2];
        const leftDim = transposeA ? a3d.shape[2] : a3d.shape[1];
        const rightDim = transposeB ? b3d.shape[1] : b3d.shape[2];
        const batchDim = Math.max(batchDimA, batchDimB);
        const a3dValues = backend.data.get(a3d.dataId).values;
        const b3dValues = backend.data.get(b3d.dataId).values;
        const a3dStrides = computeStrides(a3d.shape);
        const b3dStrides = computeStrides(b3d.shape);
        const [aBatch, aOuterStep, aInnerStep] = transposeA ?
            [a3dStrides[0], 1, a3dStrides[1]] :
            [a3dStrides[0], a3dStrides[1], 1];
        const [bInnerStep, bOuterStep, bBatch] = transposeB ?
            [1, b3dStrides[1], b3dStrides[0]] :
            [b3dStrides[1], 1, b3dStrides[0]];
        const size = leftDim * rightDim;
        const result = buffer([batchDim, leftDim, rightDim], a3d.dtype);
        const resVals = result.values;
        const blockSize = backend.blockSize;
        for (let bi = 0; bi < batchDim; bi++) {
            for (let i0 = 0; i0 < leftDim; i0 += blockSize) {
                for (let j0 = 0; j0 < rightDim; j0 += blockSize) {
                    for (let k0 = 0; k0 < sharedDim; k0 += blockSize) {
                        // for when blockSize doesn't evenly divide the input
                        const iBlock = Math.min(i0 + blockSize, leftDim);
                        const jBlock = Math.min(j0 + blockSize, rightDim);
                        const kBlock = Math.min(k0 + blockSize, sharedDim);
                        for (let i = i0; i < iBlock; i++) {
                            for (let j = j0; j < jBlock; j++) {
                                let sum = 0.0;
                                for (let k = k0; k < kBlock; k++) {
                                    const batchOffsetA = Math.min(bi, batchDimA - 1) * aBatch;
                                    const batchOffsetB = Math.min(bi, batchDimB - 1) * bBatch;
                                    const aVal = a3dValues[batchOffsetA + i * aOuterStep + k * aInnerStep];
                                    const bVal = b3dValues[k * bInnerStep + j * bOuterStep + batchOffsetB];
                                    sum += aVal * bVal;
                                }
                                resVals[bi * size + (i * rightDim + j)] += sum;
                            }
                        }
                    }
                }
            }
        }
        backend.disposeIntermediateTensorInfo(a3d);
        backend.disposeIntermediateTensorInfo(b3d);
        // set correct shape on output.
        return backend.makeTensorInfo(outShape, result.dtype, result.values);
    }
    const batchMatMulConfig$1 = {
        kernelName: BatchMatMul,
        backendName: 'cpu',
        kernelFunc: batchMatMul$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function _fusedMatMul$1(args) {
        const { inputs, backend, attrs } = args;
        const { a, b, bias, preluActivationWeights } = inputs;
        const { transposeA, transposeB, activation, leakyreluAlpha } = attrs;
        let current;
        let addRes;
        let activationRes;
        const intermediates = [];
        const matMulRes = batchMatMul$1({ inputs: { a, b }, attrs: { transposeA, transposeB }, backend });
        current = matMulRes;
        if (bias) {
            addRes = add({ inputs: { a: current, b: bias }, backend });
            intermediates.push(current);
            current = addRes;
        }
        if (activation) {
            activationRes = applyActivation(backend, current, activation, preluActivationWeights, leakyreluAlpha);
            intermediates.push(current);
            current = activationRes;
        }
        for (const i of intermediates) {
            backend.disposeIntermediateTensorInfo(i);
        }
        return current;
    }
    const _fusedMatMulConfig$1 = {
        kernelName: _FusedMatMul,
        backendName: 'cpu',
        kernelFunc: _fusedMatMul$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const acos$1 = unaryKernelFunc$1(Acos, (xi) => Math.acos(xi));
    const acosConfig$1 = {
        kernelName: Acos,
        backendName: 'cpu',
        kernelFunc: acos$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const acosh$1 = unaryKernelFunc$1(Acosh, (xi) => Math.acosh(xi));
    const acoshConfig$1 = {
        kernelName: Acosh,
        backendName: 'cpu',
        kernelFunc: acosh$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function addN$1(args) {
        const { inputs, backend } = args;
        const tensors = inputs;
        assertNotComplex$1(inputs, 'addN');
        const vals = tensors.map(t => backend.data.get(t.dataId).values);
        const outBuf = buffer(tensors[0].shape, tensors[0].dtype);
        const outVals = outBuf.values;
        for (let i = 0; i < tensors.length; i++) {
            const currVals = vals[i];
            for (let j = 0; j < outVals.length; j++) {
                outVals[j] += currVals[j];
            }
        }
        return backend.makeTensorInfo(outBuf.shape, outBuf.dtype, outBuf.values);
    }
    const addNConfig$1 = {
        kernelName: AddN,
        backendName: 'cpu',
        kernelFunc: addN$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function all$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        assertNotComplex$1(x, 'all');
        const origAxes = parseAxisParam(axis, x.shape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, x.shape.length);
        let $x = x;
        if (permutedAxes != null) {
            $x = transpose$1({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            axes = getInnerMostAxes(axes.length, x.shape.length);
        }
        assertAxesAreInnerMostDims('all', axes, $x.shape.length);
        const [outShape, reduceShape] = computeOutAndReduceShapes($x.shape, axes);
        const reduceSize = sizeFromShape(reduceShape);
        const vals = makeZerosTypedArray(sizeFromShape(outShape), $x.dtype);
        const aVals = backend.data.get($x.dataId).values;
        for (let i = 0; i < vals.length; ++i) {
            const offset = i * reduceSize;
            let all = aVals[offset];
            for (let j = 0; j < reduceSize; ++j) {
                const value = aVals[offset + j];
                all = all && value;
            }
            vals[i] = all;
        }
        if (permutedAxes != null) {
            backend.disposeIntermediateTensorInfo($x);
        }
        const result = backend.makeTensorInfo(outShape, $x.dtype, vals);
        if (keepDims) {
            const expandedShape = expandShapeToKeepDim(outShape, origAxes);
            const reshapedResult = reshape$1({ inputs: { x: result }, backend, attrs: { shape: expandedShape } });
            backend.disposeIntermediateTensorInfo(result);
            return reshapedResult;
        }
        return result;
    }
    const allConfig$1 = {
        kernelName: All,
        backendName: 'cpu',
        kernelFunc: all$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function any$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        assertNotComplex$1(x, 'any');
        const origAxes = parseAxisParam(axis, x.shape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, x.shape.length);
        let $x = x;
        if (permutedAxes != null) {
            $x = transpose$1({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            axes = getInnerMostAxes(axes.length, x.shape.length);
        }
        assertAxesAreInnerMostDims('any', axes, $x.shape.length);
        const [outShape, reduceShape] = computeOutAndReduceShapes($x.shape, axes);
        const reduceSize = sizeFromShape(reduceShape);
        const vals = makeZerosTypedArray(sizeFromShape(outShape), $x.dtype);
        const aVals = backend.data.get($x.dataId).values;
        for (let i = 0; i < vals.length; ++i) {
            const offset = i * reduceSize;
            let anyVal = aVals[offset];
            for (let j = 0; j < reduceSize; ++j) {
                const value = aVals[offset + j];
                anyVal = anyVal || value;
            }
            vals[i] = anyVal;
        }
        if (permutedAxes != null) {
            backend.disposeIntermediateTensorInfo($x);
        }
        const result = backend.makeTensorInfo(outShape, $x.dtype, vals);
        if (keepDims) {
            const expandedShape = expandShapeToKeepDim(outShape, origAxes);
            const reshapedResult = reshape$1({ inputs: { x: result }, backend, attrs: { shape: expandedShape } });
            backend.disposeIntermediateTensorInfo(result);
            return reshapedResult;
        }
        return result;
    }
    const anyConfig$1 = {
        kernelName: Any,
        backendName: 'cpu',
        kernelFunc: any$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function argMax$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis } = attrs;
        assertNotComplex$1(x, 'argMax');
        let axes = parseAxisParam(axis, x.shape);
        const permutedAxes = getAxesPermutation(axes, x.shape.length);
        let $x = x;
        const intermediateTensorInfos = [];
        if (permutedAxes != null) {
            $x = transpose$1({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            intermediateTensorInfos.push($x);
            axes = getInnerMostAxes(axes.length, $x.shape.length);
        }
        axes = [axes[0]];
        assertAxesAreInnerMostDims('argMax', axes, $x.shape.length);
        const [outShape, reduceShape] = computeOutAndReduceShapes($x.shape, axes);
        const outSize = sizeFromShape(outShape);
        const vals = makeZerosTypedArray(outSize, 'int32');
        const reduceSize = sizeFromShape(reduceShape);
        const aVals = backend.data.get($x.dataId).values;
        for (let i = 0; i < vals.length; ++i) {
            const offset = i * reduceSize;
            let max = aVals[offset];
            let maxIndex = 0;
            for (let j = 0; j < reduceSize; ++j) {
                const value = aVals[offset + j];
                if (value > max) {
                    max = value;
                    maxIndex = j;
                }
            }
            vals[i] = maxIndex;
        }
        intermediateTensorInfos.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return backend.makeTensorInfo(outShape, 'int32', vals);
    }
    const argMaxConfig$1 = {
        kernelName: ArgMax,
        backendName: 'cpu',
        kernelFunc: argMax$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function argMin$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis } = attrs;
        assertNotComplex$1(x, 'argMin');
        let axes = parseAxisParam(axis, x.shape);
        const permutedAxes = getAxesPermutation(axes, x.shape.length);
        let $x = x;
        const intermediateTensorInfos = [];
        if (permutedAxes != null) {
            $x = transpose$1({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            intermediateTensorInfos.push($x);
            axes = getInnerMostAxes(axes.length, $x.shape.length);
        }
        axes = [axes[0]];
        assertAxesAreInnerMostDims('argMin', axes, $x.shape.length);
        const [outShape, reduceShape] = computeOutAndReduceShapes($x.shape, axes);
        const outSize = sizeFromShape(outShape);
        const vals = makeZerosTypedArray(outSize, 'int32');
        const reduceSize = sizeFromShape(reduceShape);
        const aVals = backend.data.get($x.dataId).values;
        for (let i = 0; i < vals.length; ++i) {
            const offset = i * reduceSize;
            let min = aVals[offset];
            let minIndex = 0;
            for (let j = 0; j < reduceSize; ++j) {
                const value = aVals[offset + j];
                if (value < min) {
                    min = value;
                    minIndex = j;
                }
            }
            vals[i] = minIndex;
        }
        intermediateTensorInfos.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return backend.makeTensorInfo(outShape, 'int32', vals);
    }
    const argMinConfig$1 = {
        kernelName: ArgMin,
        backendName: 'cpu',
        kernelFunc: argMin$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const asin$1 = unaryKernelFunc$1(Asin, (xi) => Math.asin(xi));
    const asinConfig$1 = {
        kernelName: Asin,
        backendName: 'cpu',
        kernelFunc: asin$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const asinh$1 = unaryKernelFunc$1(Asinh, (xi) => Math.asinh(xi));
    const asinhConfig$1 = {
        kernelName: Asinh,
        backendName: 'cpu',
        kernelFunc: asinh$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const atan$1 = unaryKernelFunc$1(Atan, (xi) => Math.atan(xi));
    const atanConfig$1 = {
        kernelName: Atan,
        backendName: 'cpu',
        kernelFunc: atan$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const atan2Impl = createSimpleBinaryKernelImpl((aValue, bValue) => Math.atan2(aValue, bValue));
    const atan2$1 = binaryKernelFunc$1(Atan2, atan2Impl);
    const atan2Config$1 = {
        kernelName: Atan2,
        backendName: 'cpu',
        kernelFunc: atan2$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const atanh$1 = unaryKernelFunc$1(Atanh, (xi) => Math.atanh(xi));
    const atanhConfig$1 = {
        kernelName: Atanh,
        backendName: 'cpu',
        kernelFunc: atanh$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function pool(xValues, xShape, dtype, strides, convInfo, poolType) {
        const strideHeight = convInfo.strideHeight;
        const strideWidth = convInfo.strideWidth;
        const dilationHeight = convInfo.dilationHeight;
        const dilationWidth = convInfo.dilationWidth;
        const effectiveFilterHeight = convInfo.effectiveFilterHeight;
        const effectiveFilterWidth = convInfo.effectiveFilterWidth;
        const padTop = convInfo.padInfo.top;
        const padLeft = convInfo.padInfo.left;
        const initialValue = (poolType === 'max' ? Number.NEGATIVE_INFINITY :
            Number.POSITIVE_INFINITY);
        const output = buffer(convInfo.outShape, dtype);
        const outputVals = output.values;
        const outputBatchStrides = convInfo.outShape[1] * convInfo.outShape[2] * convInfo.outShape[3];
        const outputRowStrides = convInfo.outShape[2] * convInfo.outShape[3];
        const outputColStrides = convInfo.outShape[3];
        for (let b = 0; b < convInfo.batchSize; ++b) {
            const outputBatchOffset = b * outputBatchStrides;
            const inputBatchOffset = b * strides[0];
            for (let d = 0; d < convInfo.inChannels; ++d) {
                for (let yR = 0; yR < convInfo.outHeight; ++yR) {
                    const xRCorner = yR * strideHeight - padTop;
                    const xRMin = Math.max(0, xRCorner);
                    const xRMax = Math.min(convInfo.inHeight, effectiveFilterHeight + xRCorner);
                    const outputRowOffset = outputBatchOffset + yR * outputRowStrides;
                    for (let yC = 0; yC < convInfo.outWidth; ++yC) {
                        const xCCorner = yC * strideWidth - padLeft;
                        const xCMin = Math.max(0, xCCorner);
                        const xCMax = Math.min(convInfo.inWidth, effectiveFilterWidth + xCCorner);
                        let minMaxValue = initialValue;
                        let avgValue = 0;
                        let count = 0;
                        for (let xR = xRMin; xR < xRMax; xR += dilationHeight) {
                            const xROffset = inputBatchOffset + xR * strides[1];
                            for (let xC = xCMin; xC < xCMax; xC += dilationWidth) {
                                const xCOffset = xROffset + xC * strides[2];
                                const pixel = xValues[xCOffset + d];
                                if ((poolType === 'max' && pixel > minMaxValue)) {
                                    minMaxValue = pixel;
                                }
                                else if (poolType === 'avg') {
                                    avgValue += pixel;
                                    count++;
                                }
                            }
                            if (isNaN(minMaxValue)) {
                                break;
                            }
                        }
                        const outputOffset = outputRowOffset + yC * outputColStrides + d;
                        outputVals[outputOffset] =
                            poolType === 'avg' ? avgValue / count : minMaxValue;
                    }
                }
            }
        }
        return output;
    }
    function maxPoolPositions(xValues, xShape, dtype, convInfo, flattenPositions = false, includeBatchInIndex = false) {
        const maxPositions = buffer(convInfo.outShape, 'int32');
        const strideHeight = convInfo.strideHeight;
        const strideWidth = convInfo.strideWidth;
        const dilationHeight = convInfo.dilationHeight;
        const dilationWidth = convInfo.dilationWidth;
        const effectiveFilterHeight = convInfo.effectiveFilterHeight;
        const effectiveFilterWidth = convInfo.effectiveFilterWidth;
        const padTop = convInfo.padInfo.top;
        const padLeft = convInfo.padInfo.left;
        const xBuf = buffer(xShape, dtype, xValues);
        for (let b = 0; b < convInfo.batchSize; ++b) {
            for (let d = 0; d < convInfo.inChannels; ++d) {
                for (let yR = 0; yR < convInfo.outHeight; ++yR) {
                    const xRCorner = yR * strideHeight - padTop;
                    let xRMin = xRCorner;
                    while (xRMin < 0) {
                        xRMin += dilationHeight;
                    }
                    // const xRMin = Math.max(0, xRCorner);
                    const xRMax = Math.min(convInfo.inHeight, effectiveFilterHeight + xRCorner);
                    for (let yC = 0; yC < convInfo.outWidth; ++yC) {
                        const xCCorner = yC * strideWidth - padLeft;
                        let xCMin = xCCorner;
                        while (xCMin < 0) {
                            xCMin += dilationWidth;
                        }
                        const xCMax = Math.min(convInfo.inWidth, effectiveFilterWidth + xCCorner);
                        let maxValue = Number.NEGATIVE_INFINITY;
                        let maxPosition = -1;
                        for (let xR = xRMin; xR < xRMax; xR += dilationHeight) {
                            const wR = xR - xRCorner;
                            for (let xC = xCMin; xC < xCMax; xC += dilationWidth) {
                                const wC = xC - xCCorner;
                                const pixel = xBuf.get(b, xR, xC, d);
                                if (pixel > maxValue) {
                                    maxValue = pixel;
                                    if (flattenPositions) {
                                        maxPosition = includeBatchInIndex ?
                                            ((b * convInfo.inHeight + xR) * convInfo.inWidth + xC) *
                                                convInfo.inChannels +
                                                d :
                                            (xR * convInfo.inWidth + xC) * convInfo.inChannels + d;
                                    }
                                    else {
                                        maxPosition = wR * effectiveFilterWidth + wC;
                                    }
                                }
                            }
                        }
                        maxPositions.set(maxPosition, b, yR, yC, d);
                    }
                }
            }
        }
        return maxPositions;
    }
    function pool3d(xValues, xShape, dtype, strides, convInfo, poolType) {
        const strideDepth = convInfo.strideDepth;
        const strideHeight = convInfo.strideHeight;
        const strideWidth = convInfo.strideWidth;
        const dilationDepth = convInfo.dilationDepth;
        const dilationHeight = convInfo.dilationHeight;
        const dilationWidth = convInfo.dilationWidth;
        const effectiveFilterDepth = convInfo.effectiveFilterDepth;
        const effectiveFilterHeight = convInfo.effectiveFilterHeight;
        const effectiveFilterWidth = convInfo.effectiveFilterWidth;
        const padFront = convInfo.padInfo.front;
        const padTop = convInfo.padInfo.top;
        const padLeft = convInfo.padInfo.left;
        const initialValue = (poolType === 'max' ? Number.NEGATIVE_INFINITY :
            Number.POSITIVE_INFINITY);
        const output = buffer(convInfo.outShape, dtype);
        const outputVals = output.values;
        const outputBatchStrides = convInfo.outShape[1] * convInfo.outShape[2] *
            convInfo.outShape[3] * convInfo.outShape[4];
        const outputDepthStrides = convInfo.outShape[2] * convInfo.outShape[3] * convInfo.outShape[4];
        const outputRowStrides = convInfo.outShape[3] * convInfo.outShape[4];
        const outputColStrides = convInfo.outShape[4];
        for (let batch = 0; batch < convInfo.batchSize; ++batch) {
            const outputBatchOffset = batch * outputBatchStrides;
            const inputBatchOffset = batch * strides[0];
            for (let channel = 0; channel < convInfo.inChannels; ++channel) {
                for (let yDepth = 0; yDepth < convInfo.outDepth; ++yDepth) {
                    const xDepthCorner = yDepth * strideDepth - padFront;
                    let xDepthMin = xDepthCorner;
                    while (xDepthMin < 0) {
                        xDepthMin += dilationDepth;
                    }
                    const xDepthMax = Math.min(convInfo.inDepth, effectiveFilterDepth + xDepthCorner);
                    const outputDepthOffset = outputBatchOffset + yDepth * outputDepthStrides;
                    for (let yRow = 0; yRow < convInfo.outHeight; ++yRow) {
                        const xRowCorner = yRow * strideHeight - padTop;
                        let xRowMin = xRowCorner;
                        while (xRowMin < 0) {
                            xRowMin += dilationHeight;
                        }
                        const xRowMax = Math.min(convInfo.inHeight, effectiveFilterHeight + xRowCorner);
                        const outputRowOffset = outputDepthOffset + yRow * outputRowStrides;
                        for (let yCol = 0; yCol < convInfo.outWidth; ++yCol) {
                            const xColCorner = yCol * strideWidth - padLeft;
                            let xColMin = xColCorner;
                            while (xColMin < 0) {
                                xColMin += dilationWidth;
                            }
                            const xColMax = Math.min(convInfo.inWidth, effectiveFilterWidth + xColCorner);
                            // Shader code begins
                            const outputColOffset = outputRowOffset + yCol * outputColStrides;
                            let minMaxValue = initialValue;
                            let avgValue = 0;
                            let count = 0;
                            for (let xDepth = xDepthMin; xDepth < xDepthMax; xDepth += dilationDepth) {
                                const xDepthOffset = inputBatchOffset + xDepth * strides[1];
                                for (let xRow = xRowMin; xRow < xRowMax; xRow += dilationHeight) {
                                    const xRowOffset = xDepthOffset + xRow * strides[2];
                                    for (let xCol = xColMin; xCol < xColMax; xCol += dilationWidth) {
                                        const xColOffset = xRowOffset + xCol * strides[3];
                                        const pixel = xValues[xColOffset + channel];
                                        if ((poolType === 'max' && pixel > minMaxValue)) {
                                            minMaxValue = pixel;
                                        }
                                        else if (poolType === 'avg') {
                                            avgValue += pixel;
                                            count++;
                                        }
                                        if (isNaN(minMaxValue)) {
                                            break;
                                        }
                                    }
                                    if (isNaN(minMaxValue)) {
                                        break;
                                    }
                                }
                                if (isNaN(minMaxValue)) {
                                    break;
                                }
                            }
                            const outputOffset = outputColOffset + channel;
                            outputVals[outputOffset] =
                                poolType === 'avg' ? avgValue / count : minMaxValue;
                        }
                    }
                }
            }
        }
        return output;
    }
    function maxPool3dPositions(xBuf, convInfo) {
        const maxPositions = buffer(convInfo.outShape, 'int32');
        const strideDepth = convInfo.strideDepth;
        const strideHeight = convInfo.strideHeight;
        const strideWidth = convInfo.strideWidth;
        const dilationDepth = convInfo.dilationDepth;
        const dilationHeight = convInfo.dilationHeight;
        const dilationWidth = convInfo.dilationWidth;
        const effectiveFilterDepth = convInfo.effectiveFilterDepth;
        const effectiveFilterHeight = convInfo.effectiveFilterHeight;
        const effectiveFilterWidth = convInfo.effectiveFilterWidth;
        const padFront = convInfo.padInfo.front;
        const padTop = convInfo.padInfo.top;
        const padLeft = convInfo.padInfo.left;
        for (let batch = 0; batch < convInfo.batchSize; ++batch) {
            for (let channel = 0; channel < convInfo.inChannels; ++channel) {
                for (let yDepth = 0; yDepth < convInfo.outDepth; ++yDepth) {
                    const xDepthCorner = yDepth * strideDepth - padFront;
                    let xDepthMin = xDepthCorner;
                    while (xDepthMin < 0) {
                        xDepthMin += dilationDepth;
                    }
                    const xDepthMax = Math.min(convInfo.inDepth, effectiveFilterDepth + xDepthCorner);
                    for (let yRow = 0; yRow < convInfo.outHeight; ++yRow) {
                        const xRowCorner = yRow * strideHeight - padTop;
                        let xRowMin = xRowCorner;
                        while (xRowMin < 0) {
                            xRowMin += dilationHeight;
                        }
                        const xRowMax = Math.min(convInfo.inHeight, effectiveFilterHeight + xRowCorner);
                        for (let yCol = 0; yCol < convInfo.outWidth; ++yCol) {
                            const xColCorner = yCol * strideWidth - padLeft;
                            let xColMin = xColCorner;
                            while (xColMin < 0) {
                                xColMin += dilationWidth;
                            }
                            const xColMax = Math.min(convInfo.inWidth, effectiveFilterWidth + xColCorner);
                            // Shader code begins
                            let maxValue = Number.NEGATIVE_INFINITY;
                            let maxPosition = -1;
                            for (let xDepth = xDepthMin; xDepth < xDepthMax; xDepth += dilationDepth) {
                                const wDepth = xDepth - xDepthCorner;
                                for (let xRow = xRowMin; xRow < xRowMax; xRow += dilationHeight) {
                                    const wRow = xRow - xRowCorner;
                                    for (let xCol = xColMin; xCol < xColMax; xCol += dilationWidth) {
                                        const wCol = xCol - xColCorner;
                                        const pixel = xBuf.get(batch, xDepth, xRow, xCol, channel);
                                        if (pixel >= maxValue) {
                                            maxValue = pixel;
                                            maxPosition =
                                                wDepth * effectiveFilterHeight * effectiveFilterWidth +
                                                    wRow * effectiveFilterHeight + wCol;
                                        }
                                    }
                                }
                            }
                            maxPositions.set(maxPosition, batch, yDepth, yRow, yCol, channel);
                        }
                    }
                }
            }
        }
        return maxPositions;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function avgPool$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        assertNotComplex$1(x, 'avgPool');
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        const dilations = 1;
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in avgPool: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        const convInfo = computePool2DInfo(x.shape, filterSize, strides, dilations, pad, dimRoundingMode);
        let res;
        if (convInfo.filterWidth === 1 && convInfo.filterHeight === 1 &&
            arraysEqual(convInfo.inShape, convInfo.outShape)) {
            res = identity$1({ inputs: { x }, backend });
        }
        else {
            const xValues = backend.data.get(x.dataId).values;
            const strides = computeStrides(x.shape);
            const buffer = pool(xValues, x.shape, x.dtype, strides, convInfo, 'avg');
            res = backend.makeTensorInfo(convInfo.outShape, x.dtype, buffer.values);
        }
        return res;
    }
    const avgPoolConfig$1 = {
        kernelName: AvgPool,
        backendName: 'cpu',
        kernelFunc: avgPool$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function avgPool3D$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { filterSize, strides, pad, dimRoundingMode, dataFormat } = attrs;
        assertNotComplex$1(x, 'avgPool3d');
        const convInfo = computePool3DInfo(x.shape, filterSize, strides, 1 /* dilations */, pad, dimRoundingMode, dataFormat);
        const xValues = backend.data.get(x.dataId).values;
        const outBuf = pool3d(xValues, x.shape, x.dtype, computeStrides(x.shape), convInfo, 'avg');
        return backend.makeTensorInfo(outBuf.shape, 'float32', outBuf.values);
    }
    const avgPool3DConfig$1 = {
        kernelName: AvgPool3D,
        backendName: 'cpu',
        kernelFunc: avgPool3D$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function avgPool3DGrad$1(args) {
        const { inputs, backend, attrs } = args;
        const { dy, input } = inputs;
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        assertNotComplex$1([dy, input], 'avgPool3DGrad');
        const convInfo = computePool3DInfo(input.shape, filterSize, strides, 1 /* dilations */, pad, dimRoundingMode);
        const strideDepth = convInfo.strideDepth;
        const strideHeight = convInfo.strideHeight;
        const strideWidth = convInfo.strideWidth;
        const filterDepth = convInfo.filterDepth;
        const filterHeight = convInfo.filterHeight;
        const filterWidth = convInfo.filterWidth;
        const dilationDepth = convInfo.dilationDepth;
        const dilationHeight = convInfo.dilationHeight;
        const dilationWidth = convInfo.dilationWidth;
        const effectiveFilterDepth = convInfo.effectiveFilterDepth;
        const effectiveFilterHeight = convInfo.effectiveFilterHeight;
        const effectiveFilterWidth = convInfo.effectiveFilterWidth;
        const padFront = effectiveFilterDepth - 1 - convInfo.padInfo.front;
        const padLeft = effectiveFilterWidth - 1 - convInfo.padInfo.left;
        const padTop = effectiveFilterHeight - 1 - convInfo.padInfo.top;
        const dx = buffer(input.shape, 'float32');
        const avgMultiplier = 1 / (filterDepth * filterHeight * filterWidth);
        const dyBuf = backend.bufferSync(dy);
        for (let batch = 0; batch < convInfo.batchSize; ++batch) {
            for (let channel = 0; channel < convInfo.inChannels; ++channel) {
                for (let dxDepth = 0; dxDepth < convInfo.inDepth; ++dxDepth) {
                    for (let dxRow = 0; dxRow < convInfo.inHeight; ++dxRow) {
                        for (let dxCol = 0; dxCol < convInfo.inWidth; ++dxCol) {
                            // Shader code begins.
                            const dyDepthCorner = dxDepth - padFront;
                            const dyRowCorner = dxRow - padTop;
                            const dyColCorner = dxCol - padLeft;
                            let dotProd = 0;
                            for (let wDepth = 0; wDepth < effectiveFilterDepth; wDepth += dilationDepth) {
                                const dyDepth = (dyDepthCorner + wDepth) / strideDepth;
                                if (dyDepth < 0 || dyDepth >= convInfo.outDepth ||
                                    Math.floor(dyDepth) !== dyDepth) {
                                    continue;
                                }
                                for (let wRow = 0; wRow < effectiveFilterHeight; wRow += dilationHeight) {
                                    const dyRow = (dyRowCorner + wRow) / strideHeight;
                                    if (dyRow < 0 || dyRow >= convInfo.outHeight ||
                                        Math.floor(dyRow) !== dyRow) {
                                        continue;
                                    }
                                    for (let wCol = 0; wCol < effectiveFilterWidth; wCol += dilationWidth) {
                                        const dyCol = (dyColCorner + wCol) / strideWidth;
                                        if (dyCol < 0 || dyCol >= convInfo.outWidth ||
                                            Math.floor(dyCol) !== dyCol) {
                                            continue;
                                        }
                                        const pixel = dyBuf.get(batch, dyDepth, dyRow, dyCol, channel);
                                        dotProd += pixel;
                                    }
                                }
                            }
                            dx.set(dotProd * avgMultiplier, batch, dxDepth, dxRow, dxCol, channel);
                        }
                    }
                }
            }
        }
        return backend.makeTensorInfo(dx.shape, dx.dtype, dx.values);
    }
    const avgPool3DGradConfig$1 = {
        kernelName: AvgPool3DGrad,
        backendName: 'cpu',
        kernelFunc: avgPool3DGrad$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function avgPoolGrad$1(args) {
        const { inputs, backend, attrs } = args;
        const { dy, input } = inputs;
        const x = input;
        assertNotComplex$1([dy, input], 'avgPoolGrad');
        const { filterSize, strides, pad } = attrs;
        const convInfo = computePool2DInfo(x.shape, filterSize, strides, 1 /* dilations */, pad);
        const strideHeight = convInfo.strideHeight;
        const strideWidth = convInfo.strideWidth;
        const filterHeight = convInfo.filterHeight;
        const filterWidth = convInfo.filterWidth;
        const dilationHeight = convInfo.dilationHeight;
        const dilationWidth = convInfo.dilationWidth;
        const effectiveFilterHeight = convInfo.effectiveFilterHeight;
        const effectiveFilterWidth = convInfo.effectiveFilterWidth;
        const padLeft = effectiveFilterWidth - 1 - convInfo.padInfo.left;
        const padTop = effectiveFilterHeight - 1 - convInfo.padInfo.top;
        const dx = buffer(x.shape, 'float32');
        const avgMultiplier = 1 / (filterHeight * filterWidth);
        const dyData = backend.data.get(dy.dataId).values;
        const dyBuf = buffer(dy.shape, 'float32', dyData);
        for (let b = 0; b < convInfo.batchSize; ++b) {
            for (let d = 0; d < convInfo.inChannels; ++d) {
                for (let dxR = 0; dxR < convInfo.inHeight; ++dxR) {
                    for (let dxC = 0; dxC < convInfo.inWidth; ++dxC) {
                        // Shader code begins.
                        const dyRCorner = dxR - padTop;
                        const dyCCorner = dxC - padLeft;
                        let dotProd = 0;
                        for (let wR = 0; wR < effectiveFilterHeight; wR += dilationHeight) {
                            const dyR = (dyRCorner + wR) / strideHeight;
                            if (dyR < 0 || dyR >= convInfo.outHeight ||
                                Math.floor(dyR) !== dyR) {
                                continue;
                            }
                            for (let wC = 0; wC < effectiveFilterWidth; wC += dilationWidth) {
                                const dyC = (dyCCorner + wC) / strideWidth;
                                if (dyC < 0 || dyC >= convInfo.outWidth ||
                                    Math.floor(dyC) !== dyC) {
                                    continue;
                                }
                                const pixel = dyBuf.get(b, dyR, dyC, d);
                                dotProd += pixel;
                            }
                        }
                        dx.set(dotProd * avgMultiplier, b, dxR, dxC, d);
                    }
                }
            }
        }
        return backend.makeTensorInfo(dx.shape, dx.dtype, dx.values);
    }
    const avgPoolGradConfig$1 = {
        kernelName: AvgPoolGrad,
        backendName: 'cpu',
        kernelFunc: avgPoolGrad$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function batchNorm$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, scale, offset, mean, variance } = inputs;
        assert$4(mean.shape.length === variance.shape.length, () => 'Batch normalization gradient requires mean and variance to have ' +
            'equal ranks.');
        assert$4(offset == null || mean.shape.length === offset.shape.length, () => 'Batch normalization gradient requires mean and offset to have ' +
            'equal ranks.');
        assert$4(scale == null || mean.shape.length === scale.shape.length, () => 'Batch normalization gradient requires mean and scale to have ' +
            'equal ranks.');
        assertNotComplex$1([x, mean, variance, scale, offset], 'batchNorm');
        let { varianceEpsilon } = attrs;
        if (varianceEpsilon == null) {
            varianceEpsilon = 0.001;
        }
        const xVals = backend.data.get(x.dataId).values;
        const mVals = backend.data.get(mean.dataId).values;
        const varVals = backend.data.get(variance.dataId).values;
        const sVals = scale ? backend.data.get(scale.dataId).values :
            new Float32Array([1]);
        const offVals = offset ?
            backend.data.get(offset.dataId).values :
            new Float32Array([0]);
        const outVals = new Float32Array(xVals.length);
        const offValsLength = offVals.length;
        const sValsLength = sVals.length;
        const varValsLength = varVals.length;
        const mValsLength = mVals.length;
        let offi = 0;
        let mi = 0;
        let si = 0;
        let vi = 0;
        for (let i = 0; i < xVals.length; ++i) {
            outVals[i] = offVals[offi++] +
                (xVals[i] - mVals[mi++]) * sVals[si++] /
                    Math.sqrt(varVals[vi++] + varianceEpsilon);
            if (offi >= offValsLength) {
                offi = 0;
            }
            if (mi >= mValsLength) {
                mi = 0;
            }
            if (si >= sValsLength) {
                si = 0;
            }
            if (vi >= varValsLength) {
                vi = 0;
            }
        }
        return backend.makeTensorInfo(x.shape, x.dtype, outVals);
    }
    const batchNormConfig$1 = {
        kernelName: FusedBatchNorm,
        backendName: 'cpu',
        kernelFunc: batchNorm$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function batchToSpaceND$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { blockShape, crops } = attrs;
        assertNotComplex$1([x], 'batchToSpaceND');
        const prod = blockShape.reduce((a, b) => a * b);
        const reshaped = getReshaped(x.shape, blockShape, prod);
        const permuted = getPermuted(reshaped.length, blockShape.length);
        const reshapedPermuted = getReshapedPermuted(x.shape, blockShape, prod);
        const sliceBeginCoords = getSliceBeginCoords(crops, blockShape.length);
        const sliceSize = getSliceSize(reshapedPermuted, crops, blockShape.length);
        const xReshaped = reshape$1({ inputs: { x }, backend, attrs: { shape: reshaped } });
        const xTransposed = transpose$1({ inputs: { x: xReshaped }, backend, attrs: { perm: permuted } });
        const xTransposedReshaped = reshape$1({ inputs: { x: xTransposed }, backend, attrs: { shape: reshapedPermuted } });
        const result = slice$1({
            inputs: { x: xTransposedReshaped },
            backend,
            attrs: { begin: sliceBeginCoords, size: sliceSize }
        });
        backend.disposeIntermediateTensorInfo(xReshaped);
        backend.disposeIntermediateTensorInfo(xTransposed);
        backend.disposeIntermediateTensorInfo(xTransposedReshaped);
        return result;
    }
    const batchToSpaceNDConfig$1 = {
        kernelName: BatchToSpaceND,
        backendName: 'cpu',
        kernelFunc: batchToSpaceND$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function bincount$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, weights } = inputs;
        const { size } = attrs;
        const xVals = backend.data.get(x.dataId).values;
        const weightsVals = backend.data.get(weights.dataId).values;
        const outVals = bincountImpl(xVals, weightsVals, weights.dtype, weights.shape, size);
        return backend.makeTensorInfo([size], weights.dtype, outVals);
    }
    const bincountConfig$1 = {
        kernelName: Bincount,
        backendName: 'cpu',
        kernelFunc: bincount$1
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function broadcastArgs$1(args) {
        const { inputs, backend } = args;
        const { s0, s1 } = inputs;
        const s0Vals = backend.data.get(s0.dataId).values;
        const s1Vals = backend.data.get(s1.dataId).values;
        const broadcastShape = assertAndGetBroadcastShape(Array.from(s0Vals), Array.from(s1Vals));
        return backend.makeTensorInfo([broadcastShape.length], 'int32', Int32Array.from(broadcastShape));
    }
    const broadcastArgsConfig$1 = {
        kernelName: BroadcastArgs,
        backendName: 'cpu',
        kernelFunc: broadcastArgs$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const clipByValue$1 = unaryKernelFunc$1(ClipByValue, (xi, attrs) => {
        const clipAttrs = attrs;
        if (xi > clipAttrs.clipValueMax) {
            return clipAttrs.clipValueMax;
        }
        return xi < clipAttrs.clipValueMin ? clipAttrs.clipValueMin : xi;
    });
    const clipByValueConfig$1 = {
        kernelName: ClipByValue,
        backendName: 'cpu',
        kernelFunc: clipByValue$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const complexAbs$1 = (args) => {
        const { x } = args.inputs;
        const cpuBackend = args.backend;
        const resultValues = new Float32Array(sizeFromShape(x.shape));
        const complexVals = cpuBackend.data.get(x.dataId);
        const real = complexVals.complexTensorInfos.real;
        const imag = complexVals.complexTensorInfos.imag;
        const realVals = cpuBackend.data.get(real.dataId).values;
        const imagVals = cpuBackend.data.get(imag.dataId).values;
        for (let i = 0; i < realVals.length; i++) {
            const real = realVals[i];
            const imag = imagVals[i];
            resultValues[i] = Math.hypot(real, imag);
        }
        return cpuBackend.makeOutput(resultValues, x.shape, 'float32');
    };
    const complexAbsConfig$1 = {
        kernelName: ComplexAbs,
        backendName: 'cpu',
        kernelFunc: complexAbs$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function imag$1(args) {
        const { inputs, backend } = args;
        const { input } = inputs;
        const imag = backend.data.get(input.dataId).complexTensorInfos.imag;
        const imagVal = backend.data.get(imag.dataId).values;
        // When complex tensor is disposed, its underlying parts will be disposed too.
        // Make new tensor out of the imag value of the complex. This makes sure the
        // value is still accessible even if complex tensor is disposed.
        return backend.makeTensorInfo(imag.shape, imag.dtype, imagVal);
    }
    const imagConfig$1 = {
        kernelName: Imag,
        backendName: 'cpu',
        kernelFunc: imag$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function concat$1(args) {
        const { inputs, backend, attrs } = args;
        const { axis } = attrs;
        const $axis = parseAxisParam(axis, inputs[0].shape)[0];
        let outShape = computeOutShape$1(inputs.map(t => t.shape), $axis);
        if (sizeFromShape(outShape) === 0) {
            return backend.makeTensorInfo(outShape, inputs[0].dtype, []);
        }
        // Keep only non-empty tensors (ignore tensors with 0 in their shape).
        const $inputs = inputs.filter(t => sizeFromShape(t.shape) > 0);
        if ($inputs.length === 1) {
            return identity$1({ inputs: { x: $inputs[0] }, backend });
        }
        const shapes = $inputs.map(t => t.shape);
        assertParamsConsistent(shapes, $axis);
        if ($inputs[0].dtype === 'complex64') {
            const reals = $inputs.map((t) => real$1({ inputs: { input: t }, backend }));
            const imags = $inputs.map((t) => imag$1({ inputs: { input: t }, backend }));
            const realConcated = concat$1({ inputs: reals, backend, attrs: { axis: $axis } });
            const imagConcated = concat$1({ inputs: imags, backend, attrs: { axis: $axis } });
            const result = complex$1({ inputs: { real: realConcated, imag: imagConcated }, backend });
            reals.forEach(r => backend.disposeIntermediateTensorInfo(r));
            imags.forEach(i => backend.disposeIntermediateTensorInfo(i));
            backend.disposeIntermediateTensorInfo(realConcated);
            backend.disposeIntermediateTensorInfo(imagConcated);
            return result;
        }
        // Any concat of n-dimensional tensors across any axis can be reduced to
        // a concatenation of two-dimensional tensors across the axis 1 by first
        // partitioning the axes of the original tensors into those less than the
        // axis to be concatenated and the rest. Then reshape the tensors
        // into a two-dimensional tensor by collapsing these two sets of axes and
        // concatenate the resulting matrices across the axis 1, finally reshaping
        // the result to have the proper shape.
        const inputs2D = $inputs.map(t => {
            const innerSize = sizeFromShape(t.shape.slice($axis));
            const shape = [-1, innerSize];
            return reshape$1({ inputs: { x: t }, backend, attrs: { shape } });
        });
        const inputsValShapes = inputs2D.map(t => {
            return { vals: backend.data.get(t.dataId).values, shape: t.shape };
        });
        // Concats 2d tensors along axis=1.
        outShape =
            computeOutShape$1(inputs2D.map(t => t.shape), 1 /* axis */);
        const simplyConcat = inputs2D[0].shape[0] === 1;
        const outVals = concatImpl$1(inputsValShapes, outShape, inputs[0].dtype, simplyConcat);
        const finalOutShape = computeOutShape$1($inputs.map(t => t.shape), $axis);
        const outInfo = backend.makeTensorInfo(finalOutShape, inputs[0].dtype, outVals);
        inputs2D.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return outInfo;
    }
    const concatConfig$1 = {
        kernelName: Concat,
        backendName: 'cpu',
        kernelFunc: concat$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv2D(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter } = inputs;
        const { strides, pad, dataFormat, dilations, dimRoundingMode } = attrs;
        assertNotComplex$1([x, filter], 'conv2d');
        const $dataFormat = convertConv2DDataFormat(dataFormat);
        const convInfo = computeConv2DInfo(x.shape, filter.shape, strides, dilations, pad, dimRoundingMode, false /* depthwise */, $dataFormat);
        const filterHeight = convInfo.filterHeight;
        const filterWidth = convInfo.filterWidth;
        const dilationHeight = convInfo.dilationHeight;
        const dilationWidth = convInfo.dilationWidth;
        const padLeft = convInfo.padInfo.left;
        const padTop = convInfo.padInfo.top;
        const isChannelsLast = convInfo.dataFormat === 'channelsLast';
        const y = new TensorBuffer(convInfo.outShape, x.dtype);
        const xStrides = computeStrides(x.shape);
        const filterStrides = computeStrides(filter.shape);
        const xBatchStride = xStrides[0];
        const xRowStride = isChannelsLast ? xStrides[1] : xStrides[2];
        const xColStride = isChannelsLast ? xStrides[2] : 1;
        const xChannelStride = isChannelsLast ? 1 : xStrides[1];
        const yBatchStride = y.strides[0];
        const yRowStride = isChannelsLast ? y.strides[1] : y.strides[2];
        const yColStride = isChannelsLast ? y.strides[2] : 1;
        const yChannelStride = isChannelsLast ? 1 : y.strides[1];
        const xVals = backend.data.get(x.dataId).values;
        const wVals = backend.data.get(filter.dataId).values;
        const yVals = y.values;
        for (let b = 0; b < convInfo.batchSize; ++b) {
            const xOffset1 = b * xBatchStride;
            const yOffset1 = b * yBatchStride;
            for (let yR = 0; yR < convInfo.outHeight; ++yR) {
                const yOffset2 = yOffset1 + yR * yRowStride;
                const xRCorner = yR * convInfo.strideHeight - padTop;
                for (let wR = 0; wR < filterHeight; ++wR) {
                    const xR = xRCorner + wR * dilationHeight;
                    if (xR < 0 || xR >= convInfo.inHeight) {
                        continue;
                    }
                    const wOffset1 = wR * filterStrides[0];
                    const xOffset2 = xOffset1 + xR * xRowStride;
                    for (let yC = 0; yC < convInfo.outWidth; ++yC) {
                        const yOffset3 = yOffset2 + yC * yColStride;
                        const xCCorner = yC * convInfo.strideWidth - padLeft;
                        for (let wC = 0; wC < filterWidth; ++wC) {
                            const xC = xCCorner + wC * dilationWidth;
                            if (xC < 0 || xC >= convInfo.inWidth) {
                                continue;
                            }
                            const wOffset2 = wOffset1 + wC * filterStrides[1];
                            const xOffset3 = xOffset2 + xC * xColStride;
                            let wOffset3 = wOffset2;
                            for (let d1 = 0; d1 < convInfo.inChannels; ++d1) {
                                const xVal = xVals[xOffset3 + d1 * xChannelStride];
                                for (let d2 = 0; d2 < convInfo.outChannels; ++d2) {
                                    yVals[yOffset3 + d2 * yChannelStride] +=
                                        xVal * wVals[wOffset3 + d2];
                                }
                                wOffset3 += convInfo.outChannels;
                            }
                        }
                    }
                }
            }
        }
        return backend.makeTensorInfo(y.shape, y.dtype, yVals);
    }
    const conv2DConfig$1 = {
        kernelName: Conv2D$1,
        backendName: 'cpu',
        kernelFunc: conv2D
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv2DBackpropFilter$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, dy } = inputs;
        const { strides, pad, dataFormat, dimRoundingMode, filterShape } = attrs;
        assertNotComplex$1([x, dy], 'conv2dBackpropFilter');
        const $dataFormat = convertConv2DDataFormat(dataFormat);
        const convInfo = computeConv2DInfo(x.shape, filterShape, strides, 1 /* dilations */, pad, dimRoundingMode, false /* depthwise */, $dataFormat);
        const { strideHeight, strideWidth, filterHeight, filterWidth } = convInfo;
        const isChannelsLast = convInfo.dataFormat === 'channelsLast';
        const dW = new TensorBuffer(convInfo.filterShape, 'float32');
        const leftPad = convInfo.padInfo.left;
        const topPad = convInfo.padInfo.top;
        const xVals = backend.data.get(x.dataId).values;
        const dyVals = backend.data.get(dy.dataId).values;
        const xBuf = new TensorBuffer(x.shape, x.dtype, xVals);
        const dyBuf = new TensorBuffer(dy.shape, dy.dtype, dyVals);
        for (let wR = 0; wR < filterHeight; ++wR) {
            const yRMin = Math.max(0, Math.ceil((topPad - wR) / strideHeight));
            const yRMax = Math.min(convInfo.outHeight, (convInfo.inHeight + topPad - wR) / strideHeight);
            for (let wC = 0; wC < filterWidth; ++wC) {
                const yCMin = Math.max(0, Math.ceil((leftPad - wC) / strideWidth));
                const yCMax = Math.min(convInfo.outWidth, (convInfo.inWidth + leftPad - wC) / strideWidth);
                for (let d1 = 0; d1 < convInfo.inChannels; ++d1) {
                    for (let d2 = 0; d2 < convInfo.outChannels; ++d2) {
                        let dotProd = 0;
                        for (let b = 0; b < convInfo.batchSize; ++b) {
                            for (let yR = yRMin; yR < yRMax; ++yR) {
                                const xR = wR + yR * strideHeight - topPad;
                                for (let yC = yCMin; yC < yCMax; ++yC) {
                                    const xC = wC + yC * strideWidth - leftPad;
                                    if (isChannelsLast) {
                                        dotProd += xBuf.get(b, xR, xC, d1) *
                                            dyBuf.get(b, yR, yC, d2);
                                    }
                                    else {
                                        dotProd += xBuf.get(b, d1, xR, xC) *
                                            dyBuf.get(b, d2, yR, yC);
                                    }
                                }
                            }
                        }
                        dW.set(dotProd, wR, wC, d1, d2);
                    }
                }
            }
        }
        return backend.makeTensorInfo(dW.shape, dW.dtype, dW.values);
    }
    const conv2DBackpropFilterConfig$1 = {
        kernelName: Conv2DBackpropFilter,
        backendName: 'cpu',
        kernelFunc: conv2DBackpropFilter$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv2DBackpropInput$1(args) {
        const { inputs, backend, attrs } = args;
        const { dy, filter } = inputs;
        const { inputShape, strides, pad, dataFormat, dimRoundingMode } = attrs;
        assertNotComplex$1([dy, filter], 'conv2dBackpropInput');
        const filterStrides = computeStrides(filter.shape);
        const dyStrides = computeStrides(dy.shape);
        let $dataFormat = convertConv2DDataFormat(dataFormat);
        const convInfo = computeConv2DInfo(inputShape, filter.shape, strides, 1 /* dilations */, pad, dimRoundingMode, false, $dataFormat);
        const dx = new TensorBuffer(convInfo.inShape, 'float32');
        const dxValues = dx.values;
        const dyValues = backend.data.get(dy.dataId).values;
        const fltValues = backend.data.get(filter.dataId).values;
        const [fltS0, fltS1, fltS2] = filterStrides;
        const { batchSize, filterHeight, filterWidth, inChannels, inHeight, inWidth, outChannels, outHeight, outWidth, strideHeight, strideWidth } = convInfo;
        $dataFormat = convInfo.dataFormat;
        const topPad = filterHeight - 1 - convInfo.padInfo.top;
        const leftPad = filterWidth - 1 - convInfo.padInfo.left;
        const isChannelsLast = $dataFormat === 'channelsLast';
        const xBatchStride = dx.strides[0];
        const xRowStride = isChannelsLast ? dx.strides[1] : dx.strides[2];
        const xColStride = isChannelsLast ? dx.strides[2] : 1;
        const xChannelStride = isChannelsLast ? 1 : dx.strides[1];
        const yBatchStride = dyStrides[0];
        const yRowStride = isChannelsLast ? dyStrides[1] : dyStrides[2];
        const yColStride = isChannelsLast ? dyStrides[2] : 1;
        const yChannelStride = isChannelsLast ? 1 : dyStrides[1];
        for (let b = 0; b < batchSize; ++b) {
            for (let d1 = 0; d1 < inChannels; ++d1) {
                for (let xR = 0; xR < inHeight; ++xR) {
                    const xRCorner = xR - topPad;
                    const xRMin = Math.max(0, Math.ceil(xRCorner / strideHeight));
                    const yRMax = Math.min(outHeight, (filterHeight + xRCorner) / strideHeight);
                    for (let xC = 0; xC < inWidth; ++xC) {
                        const xCCorner = xC - leftPad;
                        const xCMin = Math.max(0, Math.ceil(xCCorner / strideWidth));
                        const yCMax = Math.min(outWidth, (filterWidth + xCCorner) / strideWidth);
                        let dotProd = 0;
                        for (let yR = xRMin; yR < yRMax; ++yR) {
                            const wR = yR * strideHeight - xRCorner;
                            for (let yC = xCMin; yC < yCMax; ++yC) {
                                const wC = yC * strideWidth - xCCorner;
                                const dyOffset = yBatchStride * b + yRowStride * yR + yColStride * yC;
                                const fltOffset = fltS0 * (filterHeight - 1 - wR) +
                                    fltS1 * (filterWidth - 1 - wC) + fltS2 * d1;
                                for (let d2 = 0; d2 < outChannels; ++d2) {
                                    const pixel = dyValues[dyOffset + yChannelStride * d2];
                                    const weight = fltValues[fltOffset + d2];
                                    dotProd += pixel * weight;
                                }
                            }
                        }
                        const dxOffset = xBatchStride * b + xRowStride * xR +
                            xColStride * xC + xChannelStride * d1;
                        dxValues[dxOffset] = dotProd;
                    }
                }
            }
        }
        return backend.makeTensorInfo(dx.shape, dx.dtype, dx.values);
    }
    const conv2DBackpropInputConfig$1 = {
        kernelName: Conv2DBackpropInput,
        backendName: 'cpu',
        kernelFunc: conv2DBackpropInput$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv3D$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter } = inputs;
        const { strides, pad, dilations } = attrs;
        assertNotComplex$1([x, filter], 'conv3d');
        const convInfo = computeConv3DInfo(x.shape, filter.shape, strides, dilations, pad);
        const { filterDepth, filterHeight, filterWidth, dilationDepth, dilationHeight, dilationWidth, padInfo } = convInfo;
        const padFront = padInfo.front;
        const padLeft = padInfo.left;
        const padTop = padInfo.top;
        const y = new TensorBuffer(convInfo.outShape, x.dtype);
        const xVals = backend.data.get(x.dataId).values;
        const wVals = backend.data.get(filter.dataId).values;
        const yVals = y.values;
        const xStrides = computeStrides(x.shape);
        const filterStrides = computeStrides(filter.shape);
        for (let b = 0; b < convInfo.batchSize; ++b) {
            const xOffset1 = b * xStrides[0];
            const yOffset1 = b * y.strides[0];
            for (let yF = 0; yF < convInfo.outDepth; ++yF) {
                const yOffset2 = yOffset1 + yF * y.strides[1];
                const xFCorner = yF * convInfo.strideDepth - padFront;
                for (let wF = 0; wF < filterDepth; ++wF) {
                    const xF = xFCorner + wF * dilationDepth;
                    if (xF < 0 || xF >= convInfo.inDepth) {
                        continue;
                    }
                    const wOffset1 = wF * filterStrides[0];
                    const xOffset2 = xOffset1 + xF * xStrides[1];
                    for (let yR = 0; yR < convInfo.outHeight; ++yR) {
                        const yOffset3 = yOffset2 + yR * y.strides[2];
                        const xRCorner = yR * convInfo.strideHeight - padTop;
                        for (let wR = 0; wR < filterHeight; ++wR) {
                            const xR = xRCorner + wR * dilationHeight;
                            if (xR < 0 || xR >= convInfo.inHeight) {
                                continue;
                            }
                            const wOffset2 = wOffset1 + wR * filterStrides[1];
                            const xOffset3 = xOffset2 + xR * xStrides[2];
                            for (let yC = 0; yC < convInfo.outWidth; ++yC) {
                                const yOffset4 = yOffset3 + yC * convInfo.outChannels;
                                const xCCorner = yC * convInfo.strideWidth - padLeft;
                                for (let wC = 0; wC < filterWidth; ++wC) {
                                    const xC = xCCorner + wC * dilationWidth;
                                    if (xC < 0 || xC >= convInfo.inWidth) {
                                        continue;
                                    }
                                    const wOffset3 = wOffset2 + wC * filterStrides[2];
                                    const xOffset4 = xOffset3 + xC * convInfo.inChannels;
                                    let wOffset4 = wOffset3;
                                    for (let d1 = 0; d1 < convInfo.inChannels; ++d1) {
                                        const xVal = xVals[xOffset4 + d1];
                                        for (let d2 = 0; d2 < convInfo.outChannels; ++d2) {
                                            yVals[yOffset4 + d2] += xVal * wVals[wOffset4 + d2];
                                        }
                                        wOffset4 += convInfo.outChannels;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return backend.makeTensorInfo(y.shape, y.dtype, y.values);
    }
    const conv3DConfig$1 = {
        kernelName: Conv3D$1,
        backendName: 'cpu',
        kernelFunc: conv3D$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv3DBackpropFilterV2$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, dy } = inputs;
        const { strides, pad, filterShape } = attrs;
        assertNotComplex$1([x, dy], 'conv3dBackpropFilterV2');
        const xStrides = computeStrides(x.shape);
        const dyStrides = computeStrides(dy.shape);
        const convInfo = computeConv3DInfo(x.shape, filterShape, strides, 1 /* dilations */, pad);
        const strideDepth = convInfo.strideDepth;
        const strideHeight = convInfo.strideHeight;
        const strideWidth = convInfo.strideWidth;
        const filterDepth = convInfo.filterDepth;
        const filterHeight = convInfo.filterHeight;
        const filterWidth = convInfo.filterWidth;
        const dw = new TensorBuffer(convInfo.filterShape, 'float32');
        const dwValues = dw.values;
        const [dwS0, dwS1, dwS2, dwS3] = dw.strides;
        const dyValues = backend.data.get(dy.dataId).values;
        const [dyS0, dyS1, dyS2, dyS3] = dyStrides;
        const xValues = backend.data.get(x.dataId).values;
        const [xS0, xS1, xS2, xS3] = xStrides;
        const frontPad = convInfo.padInfo.front;
        const leftPad = convInfo.padInfo.left;
        const topPad = convInfo.padInfo.top;
        for (let wF = 0; wF < filterDepth; ++wF) {
            const yFMin = Math.max(0, Math.ceil((frontPad - wF) / strideDepth));
            const yFMax = Math.min(convInfo.outDepth, (convInfo.inDepth + frontPad - wF) / strideDepth);
            const wOffset1 = wF * dwS0;
            for (let wR = 0; wR < filterHeight; ++wR) {
                const yRMin = Math.max(0, Math.ceil((topPad - wR) / strideHeight));
                const yRMax = Math.min(convInfo.outHeight, (convInfo.inHeight + topPad - wR) / strideHeight);
                const wOffset2 = wR * dwS1 + wOffset1;
                for (let wC = 0; wC < filterWidth; ++wC) {
                    const yCMin = Math.max(0, Math.ceil((leftPad - wC) / strideWidth));
                    const yCMax = Math.min(convInfo.outWidth, (convInfo.inWidth + leftPad - wC) / strideWidth);
                    const wOffset3 = wC * dwS2 + wOffset2;
                    for (let d1 = 0; d1 < convInfo.inChannels; ++d1) {
                        const wOffset4 = d1 * dwS3 + wOffset3;
                        for (let d2 = 0; d2 < convInfo.outChannels; ++d2) {
                            let dotProd = 0;
                            for (let b = 0; b < convInfo.batchSize; ++b) {
                                const xOffset1 = b * xS0;
                                const yOffset1 = b * dyS0;
                                for (let yF = yFMin; yF < yFMax; ++yF) {
                                    const xF = wF + yF * strideDepth - frontPad;
                                    const xOffset2 = xF * xS1 + xOffset1;
                                    const yOffset2 = yF * dyS1 + yOffset1;
                                    for (let yR = yRMin; yR < yRMax; ++yR) {
                                        const xR = wR + yR * strideHeight - topPad;
                                        const xOffset3 = xR * xS2 + xOffset2;
                                        const yOffset3 = yR * dyS2 + yOffset2;
                                        for (let yC = yCMin; yC < yCMax; ++yC) {
                                            const xC = wC + yC * strideWidth - leftPad;
                                            const xOffset4 = xC * xS3 + xOffset3;
                                            const yOffset4 = yC * dyS3 + yOffset3;
                                            dotProd += xValues[xOffset4 + d1] * dyValues[yOffset4 + d2];
                                        }
                                    }
                                }
                            }
                            dwValues[wOffset4 + d2] = dotProd;
                        }
                    }
                }
            }
        }
        return backend.makeTensorInfo(dw.shape, dw.dtype, dw.values);
    }
    const conv3DBackpropFilterV2Config$1 = {
        kernelName: Conv3DBackpropFilterV2,
        backendName: 'cpu',
        kernelFunc: conv3DBackpropFilterV2$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv3DBackpropInputV2(args) {
        const { inputs, backend, attrs } = args;
        const { dy, filter } = inputs;
        const { pad, strides, inputShape } = attrs;
        assertNotComplex$1([dy], 'conv3dBackpropInputV2');
        const dyStrides = computeStrides(dy.shape);
        const filterStrides = computeStrides(filter.shape);
        const convInfo = computeConv3DInfo(inputShape, filter.shape, strides, 1 /* dilations */, pad);
        const dx = new TensorBuffer(convInfo.inShape, 'float32');
        const dxValues = dx.values;
        const [dxS0, dxS1, dxS2, dxS3] = dx.strides;
        const dyValues = backend.data.get(dy.dataId).values;
        const [dyS0, dyS1, dyS2, dyS3] = dyStrides;
        const fltValues = backend.data.get(filter.dataId).values;
        const [fltS0, fltS1, fltS2, fltS3] = filterStrides;
        const { batchSize, filterDepth, filterHeight, filterWidth, inChannels, inDepth, inHeight, inWidth, outChannels, outDepth, outHeight, outWidth, strideDepth, strideHeight, strideWidth } = convInfo;
        const frontPad = filterDepth - 1 - convInfo.padInfo.front;
        const topPad = filterHeight - 1 - convInfo.padInfo.top;
        const leftPad = filterWidth - 1 - convInfo.padInfo.left;
        for (let b = 0; b < batchSize; ++b) {
            for (let d1 = 0; d1 < inChannels; ++d1) {
                // Frames of depth
                for (let xF = 0; xF < inDepth; ++xF) {
                    const xFCorner = xF - frontPad;
                    const xFMin = Math.max(0, Math.ceil(xFCorner / strideDepth));
                    const yFMax = Math.min(outDepth, (filterDepth + xFCorner) / strideDepth);
                    // Rows as per standard 2d matrix notation
                    for (let xR = 0; xR < inHeight; ++xR) {
                        const xRCorner = xR - topPad;
                        const xRMin = Math.max(0, Math.ceil(xRCorner / strideHeight));
                        const yRMax = Math.min(outHeight, (filterHeight + xRCorner) / strideHeight);
                        // Columns as per standard 2d matrix notation
                        for (let xC = 0; xC < inWidth; ++xC) {
                            const xCCorner = xC - leftPad;
                            const xCMin = Math.max(0, Math.ceil(xCCorner / strideWidth));
                            const yCMax = Math.min(outWidth, (filterWidth + xCCorner) / strideWidth);
                            let dotProd = 0;
                            for (let yF = xFMin; yF < yFMax; ++yF) {
                                const wF = yF * strideDepth - xFCorner;
                                for (let yR = xRMin; yR < yRMax; ++yR) {
                                    const wR = yR * strideHeight - xRCorner;
                                    for (let yC = xCMin; yC < yCMax; ++yC) {
                                        const wC = yC * strideWidth - xCCorner;
                                        const dyOffset = dyS0 * b + dyS1 * yF + dyS2 * yR + dyS3 * yC;
                                        const fltOffset = fltS0 * (filterDepth - 1 - wF) +
                                            fltS1 * (filterHeight - 1 - wR) +
                                            fltS2 * (filterWidth - 1 - wC) + fltS3 * d1;
                                        for (let d2 = 0; d2 < outChannels; ++d2) {
                                            const pixel = dyValues[dyOffset + d2];
                                            const weight = fltValues[fltOffset + d2];
                                            dotProd += pixel * weight;
                                        }
                                    }
                                }
                            }
                            dxValues[dxS0 * b + dxS1 * xF + dxS2 * xR + dxS3 * xC + d1] =
                                dotProd;
                        }
                    }
                }
            }
        }
        return backend.makeTensorInfo(dx.shape, dx.dtype, dx.values);
    }
    const conv3DBackpropInputV2Config = {
        kernelName: Conv3DBackpropInputV2,
        backendName: 'cpu',
        kernelFunc: conv3DBackpropInputV2
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const cos$1 = unaryKernelFunc$1(Cos, (xi) => Math.cos(xi));
    const cosConfig$1 = {
        kernelName: Cos,
        backendName: 'cpu',
        kernelFunc: cos$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const cosh$1 = unaryKernelFunc$1(Cosh, (xi) => Math.cosh(xi));
    const coshConfig$1 = {
        kernelName: Cosh,
        backendName: 'cpu',
        kernelFunc: cosh$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function cropAndResize$1(args) {
        const { inputs, backend, attrs } = args;
        const { image, boxes, boxInd } = inputs;
        const { cropSize, method, extrapolationValue } = attrs;
        const [batch, imageHeight, imageWidth, numChannels] = image.shape;
        const numBoxes = boxes.shape[0];
        const [cropHeight, cropWidth] = cropSize;
        const output = buffer([numBoxes, cropHeight, cropWidth, numChannels], 'float32');
        const boxVals = backend.data.get(boxes.dataId).values;
        const boxIndVals = backend.data.get(boxInd.dataId).values;
        const imageVals = backend.data.get(image.dataId).values;
        const inStride = computeStrides(image.shape); // to calculate flat indexes into image
        const outStride = computeStrides(output.shape); // to calculate flat indexes into output
        // Reference implementation
        // tslint:disable-next-line:max-line-length
        // https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/kernels/crop_and_resize_op.cc
        for (let b = 0; b < numBoxes; b++) {
            const startInd = b * 4;
            const y1 = boxVals[startInd];
            const x1 = boxVals[startInd + 1];
            const y2 = boxVals[startInd + 2];
            const x2 = boxVals[startInd + 3];
            const bInd = boxIndVals[b];
            if (bInd >= batch) {
                continue;
            }
            const heightScale = (cropHeight > 1) ? (y2 - y1) * (imageHeight - 1) / (cropHeight - 1) : 0;
            const widthScale = (cropWidth > 1) ? (x2 - x1) * (imageWidth - 1) / (cropWidth - 1) : 0;
            for (let y = 0; y < cropHeight; y++) {
                const yInd = (cropHeight > 1) ?
                    y1 * (imageHeight - 1) + y * (heightScale) :
                    0.5 * (y1 + y2) * (imageHeight - 1);
                if (yInd < 0 || yInd > imageHeight - 1) {
                    for (let x = 0; x < cropWidth; x++) {
                        for (let c = 0; c < numChannels; c++) {
                            const ind = c + x * outStride[2] + y * outStride[1] + b * outStride[0];
                            output.values[ind] = extrapolationValue;
                        }
                    }
                    continue;
                }
                if (method === 'bilinear') {
                    const topInd = Math.floor(yInd);
                    const bottomInd = Math.ceil(yInd);
                    const yLerp = yInd - topInd;
                    for (let x = 0; x < cropWidth; x++) {
                        const xInd = (cropWidth > 1) ?
                            x1 * (imageWidth - 1) + x * widthScale :
                            0.5 * (x1 + x2) * (imageWidth - 1);
                        if (xInd < 0 || xInd > imageWidth - 1) {
                            for (let c = 0; c < numChannels; c++) {
                                const ind = c + x * outStride[2] + y * outStride[1] + b * outStride[0];
                                output.values[ind] = extrapolationValue;
                            }
                            continue;
                        }
                        const leftInd = Math.floor(xInd);
                        const rightInd = Math.ceil(xInd);
                        const xLerp = xInd - leftInd;
                        for (let c = 0; c < numChannels; c++) {
                            let ind = c + leftInd * inStride[2] + topInd * inStride[1] +
                                bInd * inStride[0];
                            const topLeft = imageVals[ind];
                            ind = c + rightInd * inStride[2] + topInd * inStride[1] +
                                bInd * inStride[0];
                            const topRight = imageVals[ind];
                            ind = c + leftInd * inStride[2] + bottomInd * inStride[1] +
                                bInd * inStride[0];
                            const bottomLeft = imageVals[ind];
                            ind = c + rightInd * inStride[2] + bottomInd * inStride[1] +
                                bInd * inStride[0];
                            const bottomRight = imageVals[ind];
                            const top = topLeft + (topRight - topLeft) * xLerp;
                            const bottom = bottomLeft + (bottomRight - bottomLeft) * xLerp;
                            ind = c + x * outStride[2] + y * outStride[1] + b * outStride[0];
                            output.values[ind] = top + ((bottom - top) * yLerp);
                        }
                    }
                }
                else { // method == "nearest"
                    for (let x = 0; x < cropWidth; ++x) {
                        const xInd = (cropWidth > 1) ?
                            x1 * (imageWidth - 1) + x * widthScale :
                            0.5 * (x1 + x2) * (imageWidth - 1);
                        if (xInd < 0 || xInd > imageWidth - 1) {
                            for (let c = 0; c < numChannels; c++) {
                                const ind = c + x * outStride[2] + y * outStride[1] + b * outStride[0];
                                output.values[ind] = extrapolationValue;
                            }
                            continue;
                        }
                        const closestX = Math.round(xInd);
                        const closestY = Math.round(yInd);
                        for (let c = 0; c < numChannels; c++) {
                            const inInd = c + closestX * inStride[2] + closestY * inStride[1] +
                                bInd * inStride[0];
                            const outInd = c + x * outStride[2] + y * outStride[1] + b * outStride[0];
                            output.values[outInd] = imageVals[inInd];
                        }
                    }
                }
            }
        }
        return backend.makeTensorInfo(output.shape, output.dtype, output.values);
    }
    const cropAndResizeConfig$1 = {
        kernelName: CropAndResize,
        backendName: 'cpu',
        kernelFunc: cropAndResize$1
    };

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function cumprod$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, exclusive, reverse } = attrs;
        assertNotComplex$1(x, 'cumprod');
        const permutation = getAxesPermutation([axis], x.shape.length);
        let $x = x;
        if (permutation != null) {
            $x = transpose$1({ inputs: { x }, backend, attrs: { perm: permutation } });
        }
        const permutedAxis = getInnerMostAxes(1, x.shape.length)[0];
        if (permutedAxis !== $x.shape.length - 1) {
            throw new Error(`backend.cumprod in CPU expects an inner-most ` +
                `axis=${$x.shape.length - 1} but got axis=${permutedAxis}`);
        }
        const resultDtype = upcastType($x.dtype, 'int32');
        const vals = makeOnesTypedArray(sizeFromShape($x.shape), resultDtype);
        const aVals = backend.data.get($x.dataId).values;
        const finalDim = $x.shape[$x.shape.length - 1];
        const indexAdjuster = reverse ?
            (i, j) => i + finalDim - j - 1 :
            (i, j) => i + j;
        for (let i = 0; i < aVals.length; i += finalDim) {
            for (let j = 0; j < finalDim; j++) {
                const idx = indexAdjuster(i, j);
                if (j === 0) {
                    vals[idx] = exclusive ? 1 : aVals[idx];
                }
                else {
                    const prevIdx = indexAdjuster(i, j - 1);
                    vals[idx] = exclusive ? aVals[prevIdx] * vals[prevIdx] :
                        aVals[idx] * vals[prevIdx];
                }
            }
        }
        const result = backend.makeTensorInfo($x.shape, resultDtype, vals);
        if (permutation != null) {
            const reversePermutation = getUndoAxesPermutation(permutation);
            const reverseTransposedResult = transpose$1({ inputs: { x: result }, backend, attrs: { perm: reversePermutation } });
            backend.disposeIntermediateTensorInfo(result);
            backend.disposeIntermediateTensorInfo($x);
            return reverseTransposedResult;
        }
        return result;
    }
    const cumprodConfig$1 = {
        kernelName: Cumprod,
        backendName: 'cpu',
        kernelFunc: cumprod$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function cumsum$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, exclusive, reverse } = attrs;
        assertNotComplex$1(x, 'cumsum');
        const permutation = getAxesPermutation([axis], x.shape.length);
        let $x = x;
        if (permutation != null) {
            $x = transpose$1({ inputs: { x }, backend, attrs: { perm: permutation } });
        }
        const permutedAxis = getInnerMostAxes(1, x.shape.length)[0];
        if (permutedAxis !== $x.shape.length - 1) {
            throw new Error(`backend.cumsum in CPU expects an inner-most ` +
                `axis=${$x.shape.length - 1} but got axis=${permutedAxis}`);
        }
        const resultDtype = upcastType($x.dtype, 'int32');
        const vals = makeZerosTypedArray(sizeFromShape($x.shape), resultDtype);
        const aVals = backend.data.get($x.dataId).values;
        const finalDim = $x.shape[$x.shape.length - 1];
        const indexAdjuster = reverse ?
            (i, j) => i + finalDim - j - 1 :
            (i, j) => i + j;
        for (let i = 0; i < aVals.length; i += finalDim) {
            for (let j = 0; j < finalDim; j++) {
                const idx = indexAdjuster(i, j);
                if (j === 0) {
                    vals[idx] = exclusive ? 0 : aVals[idx];
                }
                else {
                    const prevIdx = indexAdjuster(i, j - 1);
                    vals[idx] = exclusive ? aVals[prevIdx] + vals[prevIdx] :
                        aVals[idx] + vals[prevIdx];
                }
            }
        }
        const result = backend.makeTensorInfo($x.shape, resultDtype, vals);
        if (permutation != null) {
            const reversePermutation = getUndoAxesPermutation(permutation);
            const reverseTransposedResult = transpose$1({ inputs: { x: result }, backend, attrs: { perm: reversePermutation } });
            backend.disposeIntermediateTensorInfo(result);
            backend.disposeIntermediateTensorInfo($x);
            return reverseTransposedResult;
        }
        return result;
    }
    const cumsumConfig$1 = {
        kernelName: Cumsum,
        backendName: 'cpu',
        kernelFunc: cumsum$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function denseBincount$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, weights } = inputs;
        const { size, binaryOutput } = attrs;
        if (x.shape.length === 1) {
            const xVals = backend.data.get(x.dataId).values;
            const weightsVals = backend.data.get(weights.dataId).values;
            const outVals = bincountImpl(xVals, weightsVals, weights.dtype, weights.shape, size);
            return backend.makeTensorInfo([size], weights.dtype, outVals);
        }
        else if (x.shape.length === 2) {
            const xBuf = backend.bufferSync(x);
            const weightsBuf = backend.bufferSync(weights);
            const outBuf = bincountReduceImpl(xBuf, weightsBuf, size, binaryOutput);
            return backend.makeTensorInfo(outBuf.shape, weights.dtype, outBuf.values);
        }
        throw new Error(`Error in denseBincount: input must be at most rank 2, but got rank` +
            `${x.shape.length}.`);
    }
    const denseBincountConfig$1 = {
        kernelName: DenseBincount,
        backendName: 'cpu',
        kernelFunc: denseBincount$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthToSpace$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { blockSize, dataFormat } = attrs;
        assert$4(dataFormat === 'NHWC', () => `Only NHWC dataFormat supported on CPU for depthToSpace. Got ${dataFormat}`);
        const batchSize = x.shape[0];
        const inputHeight = x.shape[1];
        const inputWidth = x.shape[2];
        const inputDepth = x.shape[3];
        const outputHeight = inputHeight * blockSize;
        const outputWidth = inputWidth * blockSize;
        const outputDepth = inputDepth / (blockSize * blockSize);
        const xValues = backend.data.get(x.dataId).values;
        const result = new Float32Array(batchSize * outputHeight * outputWidth * outputDepth);
        let outputIdx = 0;
        for (let b = 0; b < batchSize; ++b) {
            for (let h = 0; h < outputHeight; ++h) {
                const inH = Math.floor(h / blockSize);
                const offsetH = (h % blockSize);
                for (let w = 0; w < outputWidth; ++w) {
                    const inW = Math.floor(w / blockSize);
                    const offsetW = (w % blockSize);
                    const offsetD = (offsetH * blockSize + offsetW) * outputDepth;
                    for (let d = 0; d < outputDepth; ++d) {
                        const inD = d + offsetD;
                        const inputIdx = inD + inputDepth * (inW + inputWidth * (inH + inputHeight * b));
                        result[outputIdx++] = xValues[inputIdx];
                    }
                }
            }
        }
        return backend.makeTensorInfo([batchSize, outputHeight, outputWidth, outputDepth], x.dtype, result);
    }
    const depthToSpaceConfig$1 = {
        kernelName: DepthToSpace,
        backendName: 'cpu',
        kernelFunc: depthToSpace$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthwiseConv2dNative$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter } = inputs;
        const { strides, pad, dilations, dimRoundingMode } = attrs;
        assertNotComplex$1([x, filter], 'depthwiseConv2DNative');
        const xStrides = computeStrides(x.shape);
        const filterStrides = computeStrides(filter.shape);
        let $dilations = dilations;
        if ($dilations == null) {
            $dilations = [1, 1];
        }
        assert$4(eitherStridesOrDilationsAreOne(strides, $dilations), () => 'Error in depthwiseConv2d: Either strides or dilations must be ' +
            `1. Got strides ${strides} and dilations '${$dilations}'`);
        const convInfo = computeConv2DInfo(x.shape, filter.shape, strides, $dilations, pad, dimRoundingMode, true /* depthwise */);
        const { filterHeight, filterWidth, dilationHeight, dilationWidth, padInfo } = convInfo;
        const padLeft = padInfo.left;
        const padTop = padInfo.top;
        const chMul = convInfo.outChannels / convInfo.inChannels;
        const y = new TensorBuffer(convInfo.outShape, x.dtype);
        const xVals = backend.data.get(x.dataId).values;
        const wVals = backend.data.get(filter.dataId).values;
        const yVals = y.values;
        for (let b = 0; b < convInfo.batchSize; ++b) {
            const xOffset1 = b * xStrides[0];
            const yOffset1 = b * y.strides[0];
            for (let yR = 0; yR < convInfo.outHeight; ++yR) {
                const yOffset2 = yOffset1 + yR * y.strides[1];
                const xRCorner = yR * convInfo.strideHeight - padTop;
                for (let wR = 0; wR < filterHeight; ++wR) {
                    const xR = xRCorner + wR * dilationHeight;
                    if (xR < 0 || xR >= convInfo.inHeight) {
                        continue;
                    }
                    const wOffset1 = wR * filterStrides[0];
                    const xOffset2 = xOffset1 + xR * xStrides[1];
                    for (let yC = 0; yC < convInfo.outWidth; ++yC) {
                        const yOffset3 = yOffset2 + yC * y.strides[2];
                        const xCCorner = yC * convInfo.strideWidth - padLeft;
                        for (let wC = 0; wC < filterWidth; ++wC) {
                            const xC = xCCorner + wC * dilationWidth;
                            if (xC < 0 || xC >= convInfo.inWidth) {
                                continue;
                            }
                            const wOffset2 = wOffset1 + wC * filterStrides[1];
                            const xOffset3 = xOffset2 + xC * convInfo.inChannels;
                            let yOffset4 = yOffset3;
                            let wOffset3 = wOffset2;
                            for (let d1 = 0; d1 < convInfo.inChannels; ++d1) {
                                const xVal = xVals[xOffset3 + d1];
                                for (let q = 0; q < chMul; ++q) {
                                    yVals[yOffset4 + q] += xVal * wVals[wOffset3 + q];
                                }
                                yOffset4 += chMul;
                                wOffset3 += chMul;
                            }
                        }
                    }
                }
            }
        }
        return backend.makeTensorInfo(y.shape, y.dtype, y.values);
    }
    const depthwiseConv2dNativeConfig$1 = {
        kernelName: DepthwiseConv2dNative,
        backendName: 'cpu',
        kernelFunc: depthwiseConv2dNative$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthwiseConv2dNativeBackpropFilter$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, dy } = inputs;
        const { strides, dilations, pad, dimRoundingMode, filterShape } = attrs;
        assertNotComplex$1([x, dy], 'depthwiseConv2dNativeBackpropFilter');
        const convInfo = computeConv2DInfo(x.shape, filterShape, strides, dilations, pad, dimRoundingMode, true /* depthwise */);
        const { strideHeight, strideWidth, filterHeight, filterWidth } = convInfo;
        const dW = new TensorBuffer(convInfo.filterShape, 'float32');
        const leftPad = convInfo.padInfo.left;
        const topPad = convInfo.padInfo.top;
        const chMul = convInfo.outChannels / convInfo.inChannels;
        const xVals = backend.data.get(x.dataId).values;
        const xBuf = new TensorBuffer(x.shape, x.dtype, xVals);
        const dyVals = backend.data.get(dy.dataId).values;
        const dyBuf = new TensorBuffer(dy.shape, dy.dtype, dyVals);
        for (let wR = 0; wR < filterHeight; ++wR) {
            const yRMin = Math.max(0, Math.ceil((topPad - wR) / strideHeight));
            const yRMax = Math.min(convInfo.outHeight, (convInfo.inHeight + topPad - wR) / strideHeight);
            for (let wC = 0; wC < filterWidth; ++wC) {
                const yCMin = Math.max(0, Math.ceil((leftPad - wC) / strideWidth));
                const yCMax = Math.min(convInfo.outWidth, (convInfo.inWidth + leftPad - wC) / strideWidth);
                for (let d2 = 0; d2 < convInfo.outChannels; ++d2) {
                    const d1 = Math.trunc(d2 / chMul);
                    const dm = d2 % chMul;
                    let dotProd = 0;
                    for (let b = 0; b < convInfo.batchSize; ++b) {
                        for (let yR = yRMin; yR < yRMax; ++yR) {
                            const xR = wR + yR * strideHeight - topPad;
                            for (let yC = yCMin; yC < yCMax; ++yC) {
                                const xC = wC + yC * strideWidth - leftPad;
                                dotProd += xBuf.get(b, xR, xC, d1) *
                                    dyBuf.get(b, yR, yC, d2);
                            }
                        }
                    }
                    dW.set(dotProd, wR, wC, d1, dm);
                }
            }
        }
        return backend.makeTensorInfo(dW.shape, dW.dtype, dW.values);
    }
    const depthwiseConv2dNativeBackpropFilterConfig$1 = {
        kernelName: DepthwiseConv2dNativeBackpropFilter,
        backendName: 'cpu',
        kernelFunc: depthwiseConv2dNativeBackpropFilter$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthwiseConv2dNativeBackpropInput$1(args) {
        const { inputs, backend, attrs } = args;
        const { dy, filter } = inputs;
        const { strides, dilations, pad, dimRoundingMode, inputShape } = attrs;
        assertNotComplex$1([dy, filter], 'depthwiseConv2DNativeBackpropInput');
        const dyStrides = computeStrides(dy.shape);
        const filterStrides = computeStrides(filter.shape);
        const convInfo = computeConv2DInfo(inputShape, filter.shape, strides, dilations, pad, dimRoundingMode, true /* depthwise */);
        const dx = new TensorBuffer(convInfo.inShape, 'float32');
        const dxValues = dx.values;
        const [dxS0, dxS1, dxS2] = dx.strides;
        const dyValues = backend.data.get(dy.dataId).values;
        const [dyS0, dyS1, dyS2] = dyStrides;
        const fltValues = backend.data.get(filter.dataId).values;
        const [fltS0, fltS1, fltS2] = filterStrides;
        const { batchSize, filterHeight, filterWidth, inChannels, inHeight, inWidth, outChannels, outHeight, outWidth, strideHeight, strideWidth } = convInfo;
        const topPad = filterHeight - 1 - convInfo.padInfo.top;
        const leftPad = filterWidth - 1 - convInfo.padInfo.left;
        const chMul = outChannels / inChannels;
        for (let b = 0; b < batchSize; ++b) {
            for (let d1 = 0; d1 < inChannels; ++d1) {
                for (let xR = 0; xR < inHeight; ++xR) {
                    const xRCorner = xR - topPad;
                    const xRMin = Math.max(0, Math.ceil(xRCorner / strideHeight));
                    const yRMax = Math.min(outHeight, (filterHeight + xRCorner) / strideHeight);
                    for (let xC = 0; xC < inWidth; ++xC) {
                        const xCCorner = xC - leftPad;
                        const xCMin = Math.max(0, Math.ceil(xCCorner / strideWidth));
                        const yCMax = Math.min(outWidth, (filterWidth + xCCorner) / strideWidth);
                        let dotProd = 0;
                        for (let yR = xRMin; yR < yRMax; ++yR) {
                            const wR = yR * strideHeight - xRCorner;
                            for (let yC = xCMin; yC < yCMax; ++yC) {
                                const wC = yC * strideWidth - xCCorner;
                                const dyOffset = dyS0 * b + dyS1 * yR + dyS2 * yC;
                                const fltOffset = fltS0 * (filterHeight - 1 - wR) +
                                    fltS1 * (filterWidth - 1 - wC) + fltS2 * d1;
                                for (let dm = 0; dm < chMul; ++dm) {
                                    const d2 = d1 * chMul + dm;
                                    const pixel = dyValues[dyOffset + d2];
                                    const weight = fltValues[fltOffset + dm];
                                    dotProd += pixel * weight;
                                }
                            }
                        }
                        dxValues[dxS0 * b + dxS1 * xR + dxS2 * xC + d1] = dotProd;
                    }
                }
            }
        }
        return backend.makeTensorInfo(dx.shape, dx.dtype, dx.values);
    }
    const depthwiseConv2dNativeBackpropInputConfig$1 = {
        kernelName: DepthwiseConv2dNativeBackpropInput,
        backendName: 'cpu',
        kernelFunc: depthwiseConv2dNativeBackpropInput$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function diag$1(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        const xSize = sizeFromShape(x.shape);
        const xVals = backend.data.get(x.dataId).values;
        const outBuf = buffer([xSize, xSize], x.dtype);
        const vals = outBuf.values;
        for (let i = 0; i < xVals.length; i++) {
            vals[i * xSize + i] = xVals[i];
        }
        const outShape = [...x.shape, ...x.shape];
        return backend.makeTensorInfo(outShape, outBuf.dtype, outBuf.values);
    }
    const diagConfig$1 = {
        kernelName: Diag,
        backendName: 'cpu',
        kernelFunc: diag$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const dilation2DConfig$1 = {
        kernelName: Dilation2D,
        backendName: 'cpu',
        kernelFunc: ({ inputs, backend, attrs }) => {
            const { x, filter } = inputs;
            const { strides, pad, dilations } = attrs;
            const cpuBackend = backend;
            const xVals = cpuBackend.data.get(x.dataId).values;
            const xRank = x.shape.length;
            const filterVals = cpuBackend.data.get(filter.dataId).values;
            const filterRank = filter.shape.length;
            const { batchSize, inHeight, inWidth, inChannels, outHeight, outWidth, padInfo, strideHeight, strideWidth, filterHeight, filterWidth, dilationHeight, dilationWidth, outShape } = computeDilation2DInfo(x.shape, filter.shape, strides, pad, 'NHWC' /* dataFormat */, dilations);
            const outSize = sizeFromShape(outShape);
            const outRank = outShape.length;
            const outputVals = getArrayFromDType(x.dtype, outSize);
            // Upsampling the input by fill in `dilation size - 1` values between each
            // input value.
            // This implementation follows the TF c++ implementation:
            // https://github.com/tensorflow/tensorflow/blob/d9a3a849edc198e90172bc58eb293de457f9d986/tensorflow/core/kernels/dilation_ops.cc
            for (let b = 0; b < batchSize; ++b) {
                for (let hOut = 0; hOut < outHeight; ++hOut) {
                    const hBeg = hOut * strideHeight - padInfo.top;
                    for (let wOut = 0; wOut < outWidth; ++wOut) {
                        const wBeg = wOut * strideWidth - padInfo.left;
                        for (let d = 0; d < inChannels; ++d) {
                            let curVal = Number.MIN_SAFE_INTEGER;
                            for (let h = 0; h < filterHeight; ++h) {
                                const hIn = hBeg + h * dilationHeight;
                                if (hIn >= 0 && hIn < inHeight) {
                                    for (let w = 0; w < filterWidth; ++w) {
                                        const wIn = wBeg + w * dilationWidth;
                                        if (wIn >= 0 && wIn < inWidth) {
                                            const xIndex = locToIndex([b, hIn, wIn, d], xRank, computeStrides(x.shape));
                                            const filterIndex = locToIndex([h, w, d], filterRank, computeStrides(filter.shape));
                                            const val = xVals[xIndex] + filterVals[filterIndex];
                                            if (val > curVal) {
                                                curVal = val;
                                            }
                                        }
                                    }
                                }
                            }
                            const outputIndex = locToIndex([b, hOut, wOut, d], outRank, computeStrides(outShape));
                            outputVals[outputIndex] = curVal;
                        }
                    }
                }
            }
            const dataId = cpuBackend.write(toTypedArray(outputVals, x.dtype), outShape, x.dtype);
            return { dataId, shape: outShape, dtype: x.dtype };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const dilation2DBackpropFilterConfig = {
        kernelName: Dilation2DBackpropFilter,
        backendName: 'cpu',
        kernelFunc: ({ inputs, backend, attrs }) => {
            const { x, filter, dy } = inputs;
            const { strides, pad, dilations } = attrs;
            const cpuBackend = backend;
            const $x = toNestedArray(x.shape, cpuBackend.data.get(x.dataId).values);
            const $filter = toNestedArray(filter.shape, cpuBackend.data.get(filter.dataId).values);
            const { batchSize, inHeight, inWidth, inChannels, outHeight, outWidth, padInfo, strideHeight, strideWidth, filterHeight, filterWidth, dilationHeight, dilationWidth, outShape } = computeDilation2DInfo(x.shape, filter.shape, strides, pad, 'NHWC' /* dataFormat */, dilations);
            assert$4(dy.rank === outShape.length, () => `Error in ${Dilation2DBackpropFilter}, dy ` +
                `must have the same rank as output ${outShape.length}, but got ` +
                `${dy.rank}`);
            const $dy = toNestedArray(outShape, cpuBackend.data.get(dy.dataId).values);
            // The computed filter gradients has the same dimensions as the filter:
            // [filterHeight, filterWidth, depth]
            const gradients = makeZerosNestedTypedArray(filter.shape, filter.dtype);
            // In the case of multiple argmax branches, we only back-propagate along the
            // last branch, i.e., the one with largest value of `h * filter_cols + w`,
            // similarly to the max-pooling backward routines.
            // This implementation follows the TF c++ implementation:
            // https://github.com/tensorflow/tensorflow/blob/d9a3a849edc198e90172bc58eb293de457f9d986/tensorflow/core/kernels/dilation_ops.cc
            for (let b = 0; b < batchSize; ++b) {
                for (let hOut = 0; hOut < outHeight; ++hOut) {
                    const hBeg = hOut * strideHeight - padInfo.top;
                    for (let wOut = 0; wOut < outWidth; ++wOut) {
                        const wBeg = wOut * strideWidth - padInfo.left;
                        for (let d = 0; d < inChannels; ++d) {
                            let curVal = Number.MIN_SAFE_INTEGER;
                            let hMax = 0;
                            let wMax = 0;
                            for (let h = 0; h < filterHeight; ++h) {
                                const hIn = hBeg + h * dilationHeight;
                                if (hIn >= 0 && hIn < inHeight) {
                                    for (let w = 0; w < filterWidth; ++w) {
                                        const wIn = wBeg + w * dilationWidth;
                                        if (wIn >= 0 && wIn < inWidth) {
                                            const val = $x[b][hIn][wIn][d] + $filter[h][w][d];
                                            if (val > curVal) {
                                                curVal = val;
                                                hMax = h;
                                                wMax = w;
                                            }
                                        }
                                    }
                                }
                            }
                            gradients[hMax][wMax][d] += $dy[b][hOut][wOut][d];
                        }
                    }
                }
            }
            const dataId = cpuBackend.write(toTypedArray(gradients, x.dtype), filter.shape, filter.dtype);
            return { dataId, shape: filter.shape, dtype: filter.dtype };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const dilation2DBackpropInputConfig = {
        kernelName: Dilation2DBackpropInput,
        backendName: 'cpu',
        kernelFunc: ({ inputs, backend, attrs }) => {
            const { x, filter, dy } = inputs;
            const { strides, pad, dilations } = attrs;
            const cpuBackend = backend;
            const $x = toNestedArray(x.shape, cpuBackend.data.get(x.dataId).values);
            const $filter = toNestedArray(filter.shape, cpuBackend.data.get(filter.dataId).values);
            const { batchSize, inHeight, inWidth, inChannels, outHeight, outWidth, padInfo, strideHeight, strideWidth, filterHeight, filterWidth, dilationHeight, dilationWidth, outShape } = computeDilation2DInfo(x.shape, filter.shape, strides, pad, 'NHWC' /* dataFormat */, dilations);
            assert$4(dy.rank === outShape.length, () => `Error in ${Dilation2DBackpropInput}, dy ` +
                `must have the same rank as output ${outShape.length}, but got ` +
                `${dy.rank}`);
            const $dy = toNestedArray(outShape, cpuBackend.data.get(dy.dataId).values);
            // The computed gradients has the same dimensions as the input:
            // [batch, inputHeight, inputCols, inChannel]
            const gradients = makeZerosNestedTypedArray(x.shape, x.dtype);
            // In the case of multiple argmax branches, we only back-propagate along the
            // last branch, i.e., the one with largest value of `h * filter_cols + w`,
            // similarly to the max-pooling backward routines.
            // This implementation follows the TF c++ implementation:
            // https://github.com/tensorflow/tensorflow/blob/d9a3a849edc198e90172bc58eb293de457f9d986/tensorflow/core/kernels/dilation_ops.cc
            for (let b = 0; b < batchSize; ++b) {
                for (let hOut = 0; hOut < outHeight; ++hOut) {
                    const hBeg = hOut * strideHeight - padInfo.top;
                    for (let wOut = 0; wOut < outWidth; ++wOut) {
                        const wBeg = wOut * strideWidth - padInfo.left;
                        for (let d = 0; d < inChannels; ++d) {
                            let curVal = Number.MIN_SAFE_INTEGER;
                            let hInMax = (hBeg < 0) ? 0 : hBeg;
                            let wInMax = (wBeg < 0) ? 0 : wBeg;
                            for (let h = 0; h < filterHeight; ++h) {
                                const hIn = hBeg + h * dilationHeight;
                                if (hIn >= 0 && hIn < inHeight) {
                                    for (let w = 0; w < filterWidth; ++w) {
                                        const wIn = wBeg + w * dilationWidth;
                                        if (wIn >= 0 && wIn < inWidth) {
                                            const val = $x[b][hIn][wIn][d] + $filter[h][w][d];
                                            if (val > curVal) {
                                                curVal = val;
                                                hInMax = hIn;
                                                wInMax = wIn;
                                            }
                                        }
                                    }
                                }
                            }
                            gradients[b][hInMax][wInMax][d] += $dy[b][hOut][wOut][d];
                        }
                    }
                }
            }
            const dataId = cpuBackend.write(toTypedArray(gradients, x.dtype), x.shape, x.dtype);
            return { dataId, shape: x.shape, dtype: x.dtype };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sum$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        assertNotComplex$1(x, 'sum');
        let $x;
        if (x.dtype === 'bool') {
            $x = cast$1({ inputs: { x }, backend, attrs: { dtype: 'int32' } });
        }
        else {
            $x = identity$1({ inputs: { x }, backend });
        }
        const xRank = $x.shape.length;
        const axes = parseAxisParam(axis, $x.shape);
        const permutation = getAxesPermutation(axes, xRank);
        let reductionAxes = axes;
        let permutedX = $x;
        if (permutation != null) {
            permutedX =
                transpose$1({ inputs: { x: $x }, backend, attrs: { perm: permutation } });
            reductionAxes = getInnerMostAxes(reductionAxes.length, xRank);
        }
        assertAxesAreInnerMostDims('sum', reductionAxes, permutedX.shape.length);
        const [outShape, reduceShape] = computeOutAndReduceShapes(permutedX.shape, reductionAxes);
        const resultDtype = upcastType(permutedX.dtype, 'int32');
        let result = zeros(backend, outShape, resultDtype);
        const reduceSize = sizeFromShape(reduceShape);
        const vals = backend.data.get(result.dataId).values;
        const aVals = backend.data.get(permutedX.dataId).values;
        for (let i = 0; i < vals.length; ++i) {
            const offset = i * reduceSize;
            let sum = 0;
            for (let j = 0; j < reduceSize; ++j) {
                sum += aVals[offset + j];
            }
            vals[i] = sum;
        }
        if (keepDims) {
            const newShape = expandShapeToKeepDim(result.shape, axes);
            const oldResult = result;
            result = reshape$1({ inputs: { x: result }, backend, attrs: { shape: newShape } });
            backend.disposeIntermediateTensorInfo(oldResult);
        }
        backend.disposeIntermediateTensorInfo($x);
        if (permutation != null) {
            backend.disposeIntermediateTensorInfo(permutedX);
        }
        return result;
    }
    const sumConfig$1 = {
        kernelName: Sum,
        backendName: 'cpu',
        kernelFunc: sum$1
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function einsum$1(args) {
        const { inputs, backend, attrs } = args;
        const { equation } = attrs;
        const tensors = inputs;
        const { allDims, summedDims, idDims } = decodeEinsumEquation(equation, tensors.length);
        checkEinsumDimSizes(allDims.length, idDims, tensors);
        const { path, steps } = getEinsumComputePath(summedDims, idDims);
        const nSteps = steps.length;
        let out = null;
        let numDimsRemaining = allDims.length;
        const tensorsToDispose = [];
        for (let i = 0; i < nSteps; ++i) {
            for (const idTerm of steps[i]) {
                const { permutationIndices: perm, expandDims: dimsToExpand } = getEinsumPermutation(numDimsRemaining, idDims[idTerm]);
                let x;
                if (isIdentityPermutation(perm)) {
                    x = tensors[idTerm];
                }
                else {
                    x = transpose$1({ inputs: { x: tensors[idTerm] }, backend, attrs: { perm } });
                    tensorsToDispose.push(x);
                }
                const targetShape = x.shape.slice();
                for (let k = 0; k < dimsToExpand.length; ++k) {
                    targetShape.splice(dimsToExpand[k], 0, 1);
                }
                if (!arraysEqual(x.shape, targetShape)) {
                    x = reshape$1({ inputs: { x }, backend, attrs: { shape: targetShape } });
                    tensorsToDispose.push(x);
                }
                if (out === null) {
                    out = x;
                }
                else {
                    // tslint:disable-next-line: no-unnecessary-type-assertion
                    out = multiply$1({ inputs: { a: x, b: out }, backend });
                    tensorsToDispose.push(out);
                }
            }
            if (i < nSteps - 1) {
                if (path[i] >= 0) {
                    out = sum$1({
                        inputs: { x: out },
                        backend,
                        attrs: {
                            axis: path[i] - (allDims.length - numDimsRemaining),
                            keepDims: false
                        }
                    });
                    tensorsToDispose.push(out);
                }
                numDimsRemaining--;
            }
        }
        // Clean up intermediate tensors.
        for (const tensorInfo of tensorsToDispose) {
            if (tensorInfo === out) {
                continue;
            }
            backend.disposeIntermediateTensorInfo(tensorInfo);
        }
        return out;
    }
    const einsumConfig$1 = {
        kernelName: Einsum,
        backendName: 'cpu',
        kernelFunc: einsum$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function eluGrad$1(args) {
        const { inputs, backend } = args;
        const { dy, y } = inputs;
        assertNotComplex$1([dy, y], 'eluGrad');
        const resultValues = new Float32Array(sizeFromShape(y.shape));
        const values = backend.data.get(y.dataId).values;
        const dyValues = backend.data.get(dy.dataId).values;
        for (let i = 0; i < values.length; ++i) {
            const v = values[i];
            if (v >= 1) {
                resultValues[i] = dyValues[i];
            }
            else {
                resultValues[i] = dyValues[i] * (v + 1);
            }
        }
        return backend.makeTensorInfo(y.shape, 'float32', resultValues);
    }
    const eluGradConfig$1 = {
        kernelName: EluGrad,
        backendName: 'cpu',
        kernelFunc: eluGrad$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const p = ERF_P;
    const a1 = ERF_A1;
    const a2 = ERF_A2;
    const a3 = ERF_A3;
    const a4 = ERF_A4;
    const a5 = ERF_A5;
    const erf$1 = unaryKernelFunc$1(Erf, (xi) => {
        const sign = Math.sign(xi);
        const v = Math.abs(xi);
        const t = 1.0 / (1.0 + p * v);
        return sign *
            (1.0 -
                (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t *
                    Math.exp(-v * v));
    });
    const erfConfig$1 = {
        kernelName: Erf,
        backendName: 'cpu',
        kernelFunc: erf$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function expandDims$1(args) {
        const { inputs, backend, attrs } = args;
        const { input } = inputs;
        const { dim } = attrs;
        const inputRank = input.shape.length;
        const newShape = input.shape.slice();
        let $dim = dim;
        if (dim < 0) {
            // Negative value is counted from the tail of rank.
            assert$4(-(inputRank + 1) <= dim, () => `Axis must be in the interval [${-(inputRank + 1)}, ${inputRank}]`);
            $dim = inputRank + dim + 1;
        }
        newShape.splice($dim, 0, 1);
        return reshape$1({ inputs: { x: input }, backend, attrs: { shape: newShape } });
    }
    const expandDimsConfig$1 = {
        kernelName: ExpandDims,
        backendName: 'cpu',
        kernelFunc: expandDims$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const realDivImpl = createSimpleBinaryKernelImpl((a, b) => a / b);
    const div = binaryKernelFunc$1(RealDiv, realDivImpl);
    const realDivConfig$1 = {
        kernelName: RealDiv,
        backendName: 'cpu',
        kernelFunc: div
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Calculate FFT of inner most elements of batch tensor.
     */
    function fftBatch(input, inverse, cpuBackend) {
        const inputShape = input.shape;
        const batch = inputShape[0];
        const innerDim = inputShape[1];
        const inputVals = cpuBackend.data.get(input.dataId);
        const real2D = inputVals.complexTensorInfos.real;
        const imag2D = inputVals.complexTensorInfos.imag;
        // Collects real and imaginary values separately.
        const resultShape = [batch, innerDim];
        const resultSize = sizeFromShape(resultShape);
        const resultReal = getTypedArrayFromDType('float32', resultSize);
        const resultImag = getTypedArrayFromDType('float32', resultSize);
        for (let b = 0; b < batch; b++) {
            // TODO: Support slice ops for complex type.
            const r = slice$1({
                inputs: { x: real2D },
                backend: cpuBackend,
                attrs: { begin: [b, 0], size: [1, innerDim] }
            });
            const i = slice$1({
                inputs: { x: imag2D },
                backend: cpuBackend,
                attrs: { begin: [b, 0], size: [1, innerDim] }
            });
            const input = complex$1({ inputs: { real: r, imag: i }, backend: cpuBackend });
            // Run FFT by batch element.
            const { real, imag } = fftImpl$1(input, inverse, cpuBackend);
            const res = mergeRealAndImagArrays(real, imag);
            for (let d = 0; d < innerDim; d++) {
                const c = getComplexWithIndex(res, d);
                resultReal[b * innerDim + d] = c.real;
                resultImag[b * innerDim + d] = c.imag;
            }
            cpuBackend.disposeIntermediateTensorInfo(r);
            cpuBackend.disposeIntermediateTensorInfo(i);
            cpuBackend.disposeIntermediateTensorInfo(input);
        }
        const $realInfo = cpuBackend.makeTensorInfo(resultShape, 'float32', resultReal);
        const $imagInfo = cpuBackend.makeTensorInfo(resultShape, 'float32', resultImag);
        const result = complex$1({ inputs: { real: $realInfo, imag: $imagInfo }, backend: cpuBackend });
        cpuBackend.disposeIntermediateTensorInfo($realInfo);
        cpuBackend.disposeIntermediateTensorInfo($imagInfo);
        return result;
    }
    function fftImpl$1(input, inverse, cpuBackend) {
        const inputSize = sizeFromShape(input.shape);
        const inputVals = cpuBackend.data.get(input.dataId);
        const realVals = cpuBackend.data.get(inputVals.complexTensorInfos.real.dataId).values;
        const imagVals = cpuBackend.data.get(inputVals.complexTensorInfos.imag.dataId).values;
        if (isExponentOf2(inputSize)) {
            const result = fftRadix2(realVals, imagVals, inputSize, inverse, cpuBackend);
            const resultShape = [input.shape[0], input.shape[1]];
            if (inverse) {
                const realInfo = cpuBackend.makeTensorInfo(resultShape, 'float32', result.real);
                const imagInfo = cpuBackend.makeTensorInfo(resultShape, 'float32', result.imag);
                const sizeInfo = cpuBackend.makeTensorInfo([], 'float32', createScalarValue(inputSize, 'float32'));
                const sizeInfoCopy = identity$1({ inputs: { x: sizeInfo }, backend: cpuBackend });
                const divRealInfo = realDivConfig$1.kernelFunc({ inputs: { a: realInfo, b: sizeInfo }, backend: cpuBackend });
                const divImagInfo = realDivConfig$1.kernelFunc({ inputs: { a: imagInfo, b: sizeInfoCopy }, backend: cpuBackend });
                const divRealVals = cpuBackend.data.get(divRealInfo.dataId).values;
                const divImagVals = cpuBackend.data.get(divImagInfo.dataId).values;
                cpuBackend.disposeIntermediateTensorInfo(realInfo);
                cpuBackend.disposeIntermediateTensorInfo(imagInfo);
                cpuBackend.disposeIntermediateTensorInfo(sizeInfo);
                cpuBackend.disposeIntermediateTensorInfo(sizeInfoCopy);
                cpuBackend.disposeIntermediateTensorInfo(divRealInfo);
                cpuBackend.disposeIntermediateTensorInfo(divImagInfo);
                return { real: divRealVals, imag: divImagVals };
            }
            return result;
        }
        else {
            const data = mergeRealAndImagArrays(realVals, imagVals);
            const rawOutput = fourierTransformByMatmul(data, inputSize, inverse);
            return splitRealAndImagArrays(rawOutput);
        }
    }
    function isExponentOf2(size) {
        return (size & size - 1) === 0;
    }
    // FFT using Cooley-Tukey algorithm on radix 2 dimensional input.
    function fftRadix2(realVals, imagVals, size, inverse, cpuBackend) {
        if (size === 1) {
            return { real: realVals, imag: imagVals };
        }
        const data = mergeRealAndImagArrays(realVals, imagVals);
        const half = size / 2;
        const evenComplex = complexWithEvenIndex(data);
        const evenRealVals = evenComplex.real;
        const evenImagVals = evenComplex.imag;
        const evenShape = [evenRealVals.length];
        const evenRealInfo = cpuBackend.makeTensorInfo(evenShape, 'float32', evenRealVals);
        const evenImagInfo = cpuBackend.makeTensorInfo(evenShape, 'float32', evenImagVals);
        const evenTensorInfo = complex$1({ inputs: { real: evenRealInfo, imag: evenImagInfo }, backend: cpuBackend });
        const oddComplex = complexWithOddIndex(data);
        const oddRealVals = oddComplex.real;
        const oddImagVals = oddComplex.imag;
        const oddShape = [oddRealVals.length];
        const oddRealInfo = cpuBackend.makeTensorInfo(oddShape, 'float32', oddRealVals);
        const oddImagInfo = cpuBackend.makeTensorInfo(oddShape, 'float32', oddImagVals);
        const oddTensorInfo = complex$1({ inputs: { real: oddRealInfo, imag: oddImagInfo }, backend: cpuBackend });
        // Recursive call for half part of original input.
        const $evenComplex = fftRadix2(evenRealVals, evenImagVals, half, inverse, cpuBackend);
        const $evenRealVals = $evenComplex.real;
        const $evenImagVals = $evenComplex.imag;
        const $evenShape = [$evenRealVals.length];
        const $evenRealInfo = cpuBackend.makeTensorInfo($evenShape, 'float32', $evenRealVals);
        const $evenImagInfo = cpuBackend.makeTensorInfo($evenShape, 'float32', $evenImagVals);
        const $evenTensorInfo = complex$1({
            inputs: { real: $evenRealInfo, imag: $evenImagInfo },
            backend: cpuBackend
        });
        const $oddComplex = fftRadix2(oddRealVals, oddImagVals, half, inverse, cpuBackend);
        const $oddRealVals = $oddComplex.real;
        const $oddImagVals = $oddComplex.imag;
        const $oddShape = [$oddRealVals.length];
        const $oddRealInfo = cpuBackend.makeTensorInfo($oddShape, 'float32', $oddRealVals);
        const $oddImagInfo = cpuBackend.makeTensorInfo($oddShape, 'float32', $oddImagVals);
        const $oddTensorInfo = complex$1({ inputs: { real: $oddRealInfo, imag: $oddImagInfo }, backend: cpuBackend });
        const e = exponents(size, inverse);
        const eShape = [e.real.length];
        const eRealInfo = cpuBackend.makeTensorInfo(eShape, 'float32', e.real);
        const eImagInfo = cpuBackend.makeTensorInfo(eShape, 'float32', e.imag);
        const complexInfo = complex$1({ inputs: { real: eRealInfo, imag: eImagInfo }, backend: cpuBackend });
        const exponentInfo = multiply$1({ inputs: { a: complexInfo, b: $oddTensorInfo }, backend: cpuBackend });
        const addPart = add({
            inputs: { a: $evenTensorInfo, b: exponentInfo },
            backend: cpuBackend
        });
        const subPart = sub$1({
            inputs: { a: $evenTensorInfo, b: exponentInfo },
            backend: cpuBackend
        });
        const addPartReal = real$1({ inputs: { input: addPart }, backend: cpuBackend });
        const subPartReal = real$1({ inputs: { input: subPart }, backend: cpuBackend });
        const addPartImag = imag$1({ inputs: { input: addPart }, backend: cpuBackend });
        const subPartImag = imag$1({ inputs: { input: subPart }, backend: cpuBackend });
        const $real = concat$1({
            inputs: [addPartReal, subPartReal],
            backend: cpuBackend,
            attrs: { axis: 0 }
        });
        const $imag = concat$1({
            inputs: [addPartImag, subPartImag],
            backend: cpuBackend,
            attrs: { axis: 0 }
        });
        const $realVals = cpuBackend.data.get($real.dataId).values;
        const $imagVals = cpuBackend.data.get($imag.dataId).values;
        cpuBackend.disposeIntermediateTensorInfo(evenRealInfo);
        cpuBackend.disposeIntermediateTensorInfo(evenImagInfo);
        cpuBackend.disposeIntermediateTensorInfo(evenTensorInfo);
        cpuBackend.disposeIntermediateTensorInfo(oddRealInfo);
        cpuBackend.disposeIntermediateTensorInfo(oddImagInfo);
        cpuBackend.disposeIntermediateTensorInfo(oddTensorInfo);
        cpuBackend.disposeIntermediateTensorInfo($evenRealInfo);
        cpuBackend.disposeIntermediateTensorInfo($evenImagInfo);
        cpuBackend.disposeIntermediateTensorInfo($evenTensorInfo);
        cpuBackend.disposeIntermediateTensorInfo($oddRealInfo);
        cpuBackend.disposeIntermediateTensorInfo($oddImagInfo);
        cpuBackend.disposeIntermediateTensorInfo($oddTensorInfo);
        cpuBackend.disposeIntermediateTensorInfo(eRealInfo);
        cpuBackend.disposeIntermediateTensorInfo(eImagInfo);
        cpuBackend.disposeIntermediateTensorInfo(complexInfo);
        cpuBackend.disposeIntermediateTensorInfo(exponentInfo);
        cpuBackend.disposeIntermediateTensorInfo(addPart);
        cpuBackend.disposeIntermediateTensorInfo(subPart);
        cpuBackend.disposeIntermediateTensorInfo(addPartReal);
        cpuBackend.disposeIntermediateTensorInfo(addPartImag);
        cpuBackend.disposeIntermediateTensorInfo(subPartReal);
        cpuBackend.disposeIntermediateTensorInfo(subPartImag);
        cpuBackend.disposeIntermediateTensorInfo($real);
        cpuBackend.disposeIntermediateTensorInfo($imag);
        return { real: $realVals, imag: $imagVals };
    }
    // Calculate fourier transform by multplying sinusoid matrix.
    function fourierTransformByMatmul(data, size, inverse) {
        const ret = new Float32Array(size * 2);
        // TODO: Use matmul instead once it supports complex64 type.
        for (let r = 0; r < size; r++) {
            let real = 0.0;
            let imag = 0.0;
            for (let c = 0; c < size; c++) {
                const e = exponent(r * c, size, inverse);
                const term = getComplexWithIndex(data, c);
                real += term.real * e.real - term.imag * e.imag;
                imag += term.real * e.imag + term.imag * e.real;
            }
            if (inverse) {
                real /= size;
                imag /= size;
            }
            assignToTypedArray(ret, real, imag, r);
        }
        return ret;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fft$1(args) {
        const { inputs, backend } = args;
        const { input } = inputs;
        const inputSize = sizeFromShape(input.shape);
        // Collapse all outer dimensions to a single batch dimension.
        const innerDimensionSize = input.shape[input.shape.length - 1];
        const batch = inputSize / innerDimensionSize;
        const input2D = reshape$1({
            inputs: { x: input },
            backend,
            attrs: { shape: [batch, innerDimensionSize] }
        });
        const result = fftBatch(input2D, false, backend);
        const resultReshaped = reshape$1({ inputs: { x: result }, backend, attrs: { shape: input.shape } });
        backend.disposeIntermediateTensorInfo(input2D);
        backend.disposeIntermediateTensorInfo(result);
        return resultReshaped;
    }
    const fftConfig$1 = {
        kernelName: FFT,
        backendName: 'cpu',
        kernelFunc: fft$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fill$1(args) {
        const { backend, attrs } = args;
        const { shape, value, dtype } = attrs;
        const $dtype = dtype || inferDtype(value);
        const values = getArrayFromDType($dtype, sizeFromShape(shape));
        fillValues(values, value, $dtype);
        return backend.makeTensorInfo(shape, $dtype, values);
    }
    const fillConfig$1 = {
        kernelName: Fill,
        backendName: 'cpu',
        kernelFunc: fill$1
    };
    function fillValues(values, value, dtype) {
        if (dtype === 'string') {
            values.fill(value);
        }
        else {
            values.fill(value);
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const flipLeftRightConfig$1 = {
        kernelName: FlipLeftRight,
        backendName: 'cpu',
        kernelFunc: ({ inputs, attrs, backend }) => {
            const { image } = inputs;
            const cpuBackend = backend;
            const output = getTypedArrayFromDType(image.dtype, sizeFromShape(image.shape));
            const [batch, imageHeight, imageWidth, numChannels] = image.shape;
            const imageVals = cpuBackend.data.get(image.dataId).values;
            for (let batchIdx = 0; batchIdx < batch; batchIdx++) {
                const batchOffset = batchIdx * imageWidth * imageHeight * numChannels;
                for (let row = 0; row < imageHeight; row++) {
                    const rowOffset = row * (imageWidth * numChannels);
                    for (let col = 0; col < imageWidth; col++) {
                        const colOffset = col * numChannels;
                        for (let channel = 0; channel < numChannels; channel++) {
                            const coordX = Math.round(imageWidth - col - 1);
                            const outIdx = batchOffset + rowOffset + colOffset + channel;
                            let outputValue = imageVals[outIdx];
                            // If the coordinate position falls within the image boundaries...
                            if (coordX >= 0 && coordX < imageWidth) {
                                // set the output to the image value at the coordinate position.
                                const rotatedColOffset = coordX * numChannels;
                                const imageIdx = batchOffset + rowOffset + rotatedColOffset + channel;
                                outputValue = imageVals[imageIdx];
                            }
                            output[outIdx] = outputValue;
                        }
                    }
                }
            }
            const dataId = cpuBackend.write(output, image.shape, image.dtype);
            return { dataId, shape: image.shape, dtype: image.dtype };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const floorDivImpl = createSimpleBinaryKernelImpl((a, b) => Math.floor(a / b));
    const floorDiv$1 = binaryKernelFunc$1(FloorDiv, floorDivImpl, null /* complexImpl */, 'int32');
    const floorDivConfig$1 = {
        kernelName: FloorDiv,
        backendName: 'cpu',
        kernelFunc: floorDiv$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fusedConv2D(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter, bias, preluActivationWeights } = inputs;
        const { strides, pad, dataFormat, dilations, dimRoundingMode, activation, leakyreluAlpha } = attrs;
        let result = conv2D({
            inputs: { x, filter },
            backend,
            attrs: { strides, pad, dataFormat, dilations, dimRoundingMode }
        });
        if (bias) {
            const resultOld = result;
            if (dataFormat === 'NCHW' && bias.shape.length === 1 &&
                bias.shape[0] !== 1) {
                // For NCHW format, if bias is a 1-D tensor, it is supposed to be aligned
                // to the channel of the conv2d's result; if bias is a scalar, the
                // bias_add is computed as if the bias was broadcasted to the shape of the
                // conv2d's result.
                const reshapedBias = reshape$1({ inputs: { x: bias }, backend, attrs: { shape: [bias.shape[0], 1, 1] } });
                result =
                    add({ inputs: { a: result, b: reshapedBias }, backend });
                backend.disposeIntermediateTensorInfo(reshapedBias);
            }
            else {
                result = add({ inputs: { a: result, b: bias }, backend });
            }
            backend.disposeIntermediateTensorInfo(resultOld);
        }
        if (activation) {
            const resultOld = result;
            result = applyActivation(backend, result, activation, preluActivationWeights, leakyreluAlpha);
            backend.disposeIntermediateTensorInfo(resultOld);
        }
        return result;
    }
    const fusedConv2DConfig$1 = {
        kernelName: FusedConv2D,
        backendName: 'cpu',
        kernelFunc: fusedConv2D
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fusedDepthwiseConv2D$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter, bias, preluActivationWeights } = inputs;
        const { strides, pad, dataFormat, dilations, dimRoundingMode, activation, leakyreluAlpha } = attrs;
        let result = depthwiseConv2dNative$1({
            inputs: { x, filter },
            backend,
            attrs: { strides, pad, dataFormat, dilations, dimRoundingMode }
        });
        if (bias) {
            const oldResult = result;
            result = add({ inputs: { a: result, b: bias }, backend });
            backend.disposeIntermediateTensorInfo(oldResult);
        }
        if (activation) {
            const oldResult = result;
            result = applyActivation(backend, result, activation, preluActivationWeights, leakyreluAlpha);
            backend.disposeIntermediateTensorInfo(oldResult);
        }
        return result;
    }
    const fusedDepthwiseConv2DConfig$1 = {
        kernelName: FusedDepthwiseConv2D,
        backendName: 'cpu',
        kernelFunc: fusedDepthwiseConv2D$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function gatherNd$1(args) {
        const { inputs, backend } = args;
        const { params, indices } = inputs;
        const paramsSize = sizeFromShape(params.shape);
        const indicesShape = indices.shape;
        const sliceRank = indicesShape[indicesShape.length - 1];
        const [resultShape, numSlices, sliceSize, strides] = prepareAndValidate(params, indices);
        if (numSlices === 0) {
            return backend.makeTensorInfo(resultShape, params.dtype, []);
        }
        const indicesData = backend.data.get(indices.dataId).values;
        const paramsBuf = backend.bufferSync(params);
        const outBuf = gatherNdImpl(indicesData, paramsBuf, params.dtype, numSlices, sliceRank, sliceSize, strides, params.shape, paramsSize);
        return backend.makeTensorInfo(resultShape, params.dtype, outBuf.values);
    }
    const gatherNdConfig$1 = {
        kernelName: GatherNd,
        backendName: 'cpu',
        kernelFunc: gatherNd$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function gatherV2$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, indices } = inputs;
        const { axis, batchDims } = attrs;
        assertNotComplex$1([x, indices], 'gatherV2');
        // Throw error when any index is out of bound.
        const parsedAxis = parseAxisParam(axis, x.shape)[0];
        const indicesVals = backend.data.get(indices.dataId).values;
        const axisDim = x.shape[parsedAxis];
        for (let i = 0; i < indicesVals.length; ++i) {
            const index = indicesVals[i];
            assert$4(index <= axisDim - 1 && index >= 0, () => `GatherV2: the index value ${index} is not in [0, ${axisDim - 1}]`);
        }
        let $batchDims = batchDims;
        if (batchDims == null) {
            $batchDims = 0;
        }
        const indicesSize = sizeFromShape(indices.shape);
        const shapeInfo = collectGatherOpShapeInfo(x, indices, parsedAxis, $batchDims);
        const flattenX = reshape$1({
            inputs: { x },
            backend,
            attrs: {
                shape: [
                    shapeInfo.batchSize, shapeInfo.outerSize, shapeInfo.dimSize,
                    shapeInfo.sliceSize
                ]
            }
        });
        const flattenIndex = reshape$1({
            inputs: { x: indices },
            backend,
            attrs: { shape: [shapeInfo.batchSize, indicesSize / shapeInfo.batchSize] }
        });
        const flattenOutputShape = [
            shapeInfo.batchSize, shapeInfo.outerSize, indicesSize / shapeInfo.batchSize,
            shapeInfo.sliceSize
        ];
        const indicesBuf = backend.bufferSync(flattenIndex);
        const xBuf = backend.bufferSync(flattenX);
        const outBuf = gatherV2Impl(xBuf, indicesBuf, flattenOutputShape);
        backend.disposeIntermediateTensorInfo(flattenX);
        backend.disposeIntermediateTensorInfo(flattenIndex);
        return backend.makeTensorInfo(shapeInfo.outputShape, outBuf.dtype, outBuf.values);
    }
    const gatherV2Config$1 = {
        kernelName: GatherV2,
        backendName: 'cpu',
        kernelFunc: gatherV2$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function ifft$1(args) {
        const { inputs, backend } = args;
        const { input } = inputs;
        const inputSize = sizeFromShape(input.shape);
        // Collapse all outer dimensions to a single batch dimension.
        const innerDimensionSize = input.shape[input.shape.length - 1];
        const batch = inputSize / innerDimensionSize;
        const input2D = reshape$1({
            inputs: { x: input },
            backend,
            attrs: { shape: [batch, innerDimensionSize] }
        });
        const result = fftBatch(input2D, true, backend);
        const resultReshaped = reshape$1({ inputs: { x: result }, backend, attrs: { shape: input.shape } });
        backend.disposeIntermediateTensorInfo(input2D);
        backend.disposeIntermediateTensorInfo(result);
        return resultReshaped;
    }
    const ifftConfig$1 = {
        kernelName: IFFT,
        backendName: 'cpu',
        kernelFunc: ifft$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const isFinite$2 = unaryKernelFunc$1(IsFinite, (xi) => Number.isFinite(xi) ? 1 : 0, 'bool');
    const isFiniteConfig$1 = {
        kernelName: IsFinite,
        backendName: 'cpu',
        kernelFunc: isFinite$2,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const isInf$1 = unaryKernelFunc$1(IsInf, (xi) => Math.abs(xi) === Infinity ? 1 : 0, 'bool');
    const isInfConfig$1 = {
        kernelName: IsInf,
        backendName: 'cpu',
        kernelFunc: isInf$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const isNaN$2 = unaryKernelFunc$1(IsNan, (xi) => Number.isNaN(xi) ? 1 : 0, 'bool');
    const isNaNConfig$1 = {
        kernelName: IsNan,
        backendName: 'cpu',
        kernelFunc: isNaN$2,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function linSpace$1(args) {
        const { backend, attrs } = args;
        const { start, stop, num } = attrs;
        const outVals = linSpaceImpl(start, stop, num);
        return backend.makeTensorInfo([outVals.length], 'float32', outVals);
    }
    const linSpaceConfig$1 = {
        kernelName: LinSpace,
        backendName: 'cpu',
        kernelFunc: linSpace$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const log1p$1 = unaryKernelFunc$1(Log1p, (xi) => Math.log1p(xi));
    const log1pConfig$1 = {
        kernelName: Log1p,
        backendName: 'cpu',
        kernelFunc: log1p$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const logicalAndImpl = createSimpleBinaryKernelImpl((a, b) => a && b);
    const logicalAnd$1 = binaryKernelFunc$1(LogicalAnd, logicalAndImpl, null /* complexImpl */, 'bool');
    const logicalAndConfig$1 = {
        kernelName: LogicalAnd,
        backendName: 'cpu',
        kernelFunc: logicalAnd$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const logicalNot$1 = unaryKernelFunc$1(LogicalNot, (xi) => xi ? 0 : 1, 'bool');
    const logicalNotConfig$1 = {
        kernelName: LogicalNot,
        backendName: 'cpu',
        kernelFunc: logicalNot$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const logicalOrImpl = createSimpleBinaryKernelImpl((a, b) => a || b);
    const logicalOr$1 = binaryKernelFunc$1(LogicalOr, logicalOrImpl, null /* complexImpl */, 'bool');
    const logicalOrConfig$1 = {
        kernelName: LogicalOr,
        backendName: 'cpu',
        kernelFunc: logicalOr$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function lRN(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { depthRadius, bias, alpha, beta } = attrs;
        assertNotComplex$1(x, 'LRN');
        const channels = x.shape[3];
        const maxD = channels - 1;
        const xValues = backend.data.get(x.dataId).values;
        const size = sizeFromShape(x.shape);
        const result = new Float32Array(size);
        function sumAcrossChannels(offset) {
            const currentChannel = offset % channels;
            let beginSumOffset = offset - currentChannel + Math.max(0, currentChannel - depthRadius);
            const endSumOffset = offset - currentChannel + Math.min(currentChannel + depthRadius, maxD);
            let sum = 0.0;
            for (; beginSumOffset <= endSumOffset; beginSumOffset++) {
                const z = xValues[beginSumOffset];
                sum += z * z;
            }
            return sum;
        }
        for (let offset = 0; offset < size; offset++) {
            const sum = sumAcrossChannels(offset);
            const val = xValues[offset] * Math.pow(bias + alpha * sum, -beta);
            result[offset] = val;
        }
        return backend.makeTensorInfo(x.shape, x.dtype, result);
    }
    // tslint:disable-next-line: variable-name
    const LRNConfig$1 = {
        kernelName: LRN,
        backendName: 'cpu',
        kernelFunc: lRN
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function lRNGrad(args) {
        const { inputs, backend, attrs } = args;
        const { x, y, dy } = inputs;
        const { depthRadius, bias, alpha, beta } = attrs;
        assertNotComplex$1(dy, 'LRNGrad');
        const dySize = sizeFromShape(dy.shape);
        const channels = dy.shape[3];
        const dyValues = backend.data.get(dy.dataId).values;
        const xValues = backend.data.get(x.dataId).values;
        const yValues = backend.data.get(y.dataId).values;
        const result = new Float32Array(dySize);
        const size = dySize;
        for (let offset = 0; offset < size; offset++) {
            const currentChannel = offset % channels;
            const depthBegin = (offset - currentChannel) + Math.max(0, currentChannel - depthRadius);
            const depthEnd = (offset - currentChannel) +
                Math.min(channels, currentChannel + depthRadius + 1);
            let norm = 0;
            for (let k = depthBegin; k < depthEnd; k++) {
                norm += Math.pow(xValues[k], 2);
            }
            norm = alpha * norm + bias;
            for (let k = depthBegin; k < depthEnd; k++) {
                let dyi = -2 * alpha * beta * xValues[k] * yValues[offset] / norm;
                if (offset === k) {
                    dyi += Math.pow(norm, -beta);
                }
                dyi *= dyValues[offset];
                result[k] += dyi;
            }
        }
        return backend.makeTensorInfo(dy.shape, x.dtype, result);
    }
    // tslint:disable-next-line: variable-name
    const LRNGradConfig$1 = {
        kernelName: LRNGrad,
        backendName: 'cpu',
        kernelFunc: lRNGrad
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function max$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { reductionIndices, keepDims } = attrs;
        const cpuBackend = backend;
        let xShape = x.shape;
        const xRank = xShape.length;
        const origAxes = parseAxisParam(reductionIndices, xShape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, xRank);
        let xVals = cpuBackend.data.get(x.dataId).values;
        if (permutedAxes != null) {
            const newShape = new Array(xRank);
            for (let i = 0; i < newShape.length; i++) {
                newShape[i] = xShape[permutedAxes[i]];
            }
            xVals = transposeImpl$1(xVals, xShape, x.dtype, permutedAxes, newShape);
            axes = getInnerMostAxes(axes.length, xRank);
            xShape = newShape;
        }
        assertNotComplex$1(x, 'max');
        assertAxesAreInnerMostDims('max', axes, xRank);
        const [maxOutShape, reduceShape] = computeOutAndReduceShapes(xShape, axes);
        const reduceSize = sizeFromShape(reduceShape);
        const result = maxImpl$1(xVals, reduceSize, maxOutShape, x.dtype);
        const dataId = cpuBackend.write(result, maxOutShape, x.dtype);
        let outShape = maxOutShape;
        if (keepDims) {
            // reshape
            const newShape = expandShapeToKeepDim(maxOutShape, origAxes);
            outShape = newShape;
        }
        return { dataId, shape: outShape, dtype: x.dtype };
    }
    const maxConfig$1 = {
        kernelName: Max,
        backendName: 'cpu',
        kernelFunc: max$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPool$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        assertNotComplex$1(x, 'maxPool');
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        const dilations = 1;
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in maxPool: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        const convInfo = computePool2DInfo(x.shape, filterSize, strides, dilations, pad, dimRoundingMode);
        let res;
        if (convInfo.filterWidth === 1 && convInfo.filterHeight === 1 &&
            arraysEqual(convInfo.inShape, convInfo.outShape)) {
            res = identity$1({ inputs: { x }, backend });
        }
        else {
            const xValues = backend.data.get(x.dataId).values;
            const strides = computeStrides(x.shape);
            const buffer = pool(xValues, x.shape, x.dtype, strides, convInfo, 'max');
            res = backend.makeTensorInfo(convInfo.outShape, x.dtype, buffer.values);
        }
        return res;
    }
    const maxPoolConfig$1 = {
        kernelName: MaxPool,
        backendName: 'cpu',
        kernelFunc: maxPool$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPool3D(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { filterSize, strides, pad, dimRoundingMode, dataFormat } = attrs;
        assertNotComplex$1(x, 'maxPool3d');
        const convInfo = computePool3DInfo(x.shape, filterSize, strides, 1 /* dilations */, pad, dimRoundingMode, dataFormat);
        const xValues = backend.data.get(x.dataId).values;
        const outBuf = pool3d(xValues, x.shape, x.dtype, computeStrides(x.shape), convInfo, 'max');
        return backend.makeTensorInfo(outBuf.shape, 'float32', outBuf.values);
    }
    const maxPool3DConfig$1 = {
        kernelName: MaxPool3D,
        backendName: 'cpu',
        kernelFunc: maxPool3D
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPool3DGrad$1(args) {
        const { inputs, backend, attrs } = args;
        const { dy, input } = inputs;
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        assertNotComplex$1([dy, input], 'maxPool3DGrad');
        const convInfo = computePool3DInfo(input.shape, filterSize, strides, 1 /* dilations */, pad, dimRoundingMode);
        const inputBuf = backend.bufferSync(input);
        const maxPosBuf = maxPool3dPositions(inputBuf, convInfo);
        const strideDepth = convInfo.strideDepth;
        const strideHeight = convInfo.strideHeight;
        const strideWidth = convInfo.strideWidth;
        const dilationDepth = convInfo.dilationDepth;
        const dilationHeight = convInfo.dilationHeight;
        const dilationWidth = convInfo.dilationWidth;
        const effectiveFilterDepth = convInfo.effectiveFilterDepth;
        const effectiveFilterHeight = convInfo.effectiveFilterHeight;
        const effectiveFilterWidth = convInfo.effectiveFilterWidth;
        const padFront = effectiveFilterDepth - 1 - convInfo.padInfo.front;
        const padLeft = effectiveFilterWidth - 1 - convInfo.padInfo.left;
        const padTop = effectiveFilterHeight - 1 - convInfo.padInfo.top;
        const dx = buffer(input.shape, 'float32');
        const dyBuf = backend.bufferSync(dy);
        for (let batch = 0; batch < convInfo.batchSize; ++batch) {
            for (let channel = 0; channel < convInfo.inChannels; ++channel) {
                for (let dxDepth = 0; dxDepth < convInfo.inDepth; ++dxDepth) {
                    for (let dxRow = 0; dxRow < convInfo.inHeight; ++dxRow) {
                        for (let dxCol = 0; dxCol < convInfo.inWidth; ++dxCol) {
                            // Shader code begins
                            const dyDepthCorner = dxDepth - padFront;
                            const dyRowCorner = dxRow - padTop;
                            const dyColCorner = dxCol - padLeft;
                            let dotProd = 0;
                            for (let wDepth = 0; wDepth < effectiveFilterDepth; wDepth += dilationDepth) {
                                const dyDepth = (dyDepthCorner + wDepth) / strideDepth;
                                if (dyDepth < 0 || dyDepth >= convInfo.outDepth ||
                                    Math.floor(dyDepth) !== dyDepth) {
                                    continue;
                                }
                                for (let wRow = 0; wRow < effectiveFilterHeight; wRow += dilationHeight) {
                                    const dyRow = (dyRowCorner + wRow) / strideHeight;
                                    if (dyRow < 0 || dyRow >= convInfo.outHeight ||
                                        Math.floor(dyRow) !== dyRow) {
                                        continue;
                                    }
                                    for (let wCol = 0; wCol < effectiveFilterWidth; wCol += dilationWidth) {
                                        const dyCol = (dyColCorner + wCol) / strideWidth;
                                        if (dyCol < 0 || dyCol >= convInfo.outWidth ||
                                            Math.floor(dyCol) !== dyCol) {
                                            continue;
                                        }
                                        const maxPos = effectiveFilterDepth * effectiveFilterHeight *
                                            effectiveFilterWidth -
                                            1 -
                                            maxPosBuf.get(batch, dyDepth, dyRow, dyCol, channel);
                                        const curPos = wDepth * effectiveFilterHeight * effectiveFilterWidth +
                                            wRow * effectiveFilterWidth + wCol;
                                        const mask = maxPos === curPos ? 1 : 0;
                                        if (mask === 0) {
                                            continue;
                                        }
                                        const pixel = dyBuf.get(batch, dyDepth, dyRow, dyCol, channel);
                                        dotProd += pixel * mask;
                                    }
                                }
                            }
                            dx.set(dotProd, batch, dxDepth, dxRow, dxCol, channel);
                        }
                    }
                }
            }
        }
        return backend.makeTensorInfo(dx.shape, dx.dtype, dx.values);
    }
    const maxPool3DGradConfig$1 = {
        kernelName: MaxPool3DGrad,
        backendName: 'cpu',
        kernelFunc: maxPool3DGrad$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPoolGrad$1(args) {
        const { inputs, backend, attrs } = args;
        const { dy, input, output } = inputs;
        const x = input;
        assertNotComplex$1([input, output], 'maxPoolGrad');
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        const convInfo = computePool2DInfo(x.shape, filterSize, strides, 1 /* dilations */, pad, dimRoundingMode);
        const xValues = backend.data.get(x.dataId).values;
        const maxPosBuf = buffer(convInfo.outShape, x.dtype, maxPoolPositions(xValues, x.shape, x.dtype, convInfo).values);
        const strideHeight = convInfo.strideHeight;
        const strideWidth = convInfo.strideWidth;
        const dilationHeight = convInfo.dilationHeight;
        const dilationWidth = convInfo.dilationWidth;
        const effectiveFilterHeight = convInfo.effectiveFilterHeight;
        const effectiveFilterWidth = convInfo.effectiveFilterWidth;
        const padLeft = effectiveFilterWidth - 1 - convInfo.padInfo.left;
        const padTop = effectiveFilterHeight - 1 - convInfo.padInfo.top;
        const dx = buffer(x.shape, 'float32');
        const dyData = backend.data.get(dy.dataId).values;
        const dyBuf = buffer(dy.shape, 'float32', dyData);
        for (let b = 0; b < convInfo.batchSize; ++b) {
            for (let d = 0; d < convInfo.inChannels; ++d) {
                for (let dxR = 0; dxR < convInfo.inHeight; ++dxR) {
                    for (let dxC = 0; dxC < convInfo.inWidth; ++dxC) {
                        // Shader code begins.
                        const dyRCorner = dxR - padTop;
                        const dyCCorner = dxC - padLeft;
                        let dotProd = 0;
                        for (let wR = 0; wR < effectiveFilterHeight; wR += dilationHeight) {
                            const dyR = (dyRCorner + wR) / strideHeight;
                            if (dyR < 0 || dyR >= convInfo.outHeight ||
                                Math.floor(dyR) !== dyR) {
                                continue;
                            }
                            for (let wC = 0; wC < effectiveFilterWidth; wC += dilationWidth) {
                                const dyC = (dyCCorner + wC) / strideWidth;
                                if (dyC < 0 || dyC >= convInfo.outWidth ||
                                    Math.floor(dyC) !== dyC) {
                                    continue;
                                }
                                const maxPos = effectiveFilterHeight * effectiveFilterWidth - 1 -
                                    maxPosBuf.get(b, dyR, dyC, d);
                                const curPos = wR * effectiveFilterWidth + wC;
                                const mask = maxPos === curPos ? 1 : 0;
                                if (mask === 0) {
                                    continue;
                                }
                                const pixel = dyBuf.get(b, dyR, dyC, d);
                                dotProd += pixel * mask;
                            }
                        }
                        dx.set(dotProd, b, dxR, dxC, d);
                    }
                }
            }
        }
        return backend.makeTensorInfo(dx.shape, dx.dtype, dx.values);
    }
    const maxPoolGradConfig$1 = {
        kernelName: MaxPoolGrad,
        backendName: 'cpu',
        kernelFunc: maxPoolGrad$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPoolWithArgmaxImpl$1(xValues, xShape, dtype, includeBatchInIndex, convInfo) {
        const strides = computeStrides(xShape);
        const maxPools = pool(xValues, xShape, dtype, strides, convInfo, 'max');
        const maxPositions = maxPoolPositions(xValues, xShape, dtype, convInfo, true, includeBatchInIndex);
        return [maxPools.values, maxPositions.values];
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const maxPoolWithArgmaxConfig$1 = {
        kernelName: MaxPoolWithArgmax,
        backendName: 'cpu',
        kernelFunc: ({ inputs, attrs, backend }) => {
            const { x } = inputs;
            const { filterSize, strides, pad, includeBatchInIndex } = attrs;
            const cpuBackend = backend;
            assertNotComplex$1(x, 'MaxPoolWithArgmax');
            const values = cpuBackend.data.get(x.dataId).values;
            const convInfo = computePool2DInfo(x.shape, filterSize, strides, [1, 1], pad);
            const [pooled, indexes] = maxPoolWithArgmaxImpl$1(values, x.shape, x.dtype, includeBatchInIndex, convInfo);
            const pooledDataId = cpuBackend.write(pooled, convInfo.outShape, x.dtype);
            const indexesDataId = cpuBackend.write(indexes, convInfo.outShape, x.dtype);
            return [
                { dataId: pooledDataId, shape: convInfo.outShape, dtype: x.dtype },
                { dataId: indexesDataId, shape: convInfo.outShape, dtype: 'int32' }
            ];
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function mean(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        const axes = parseAxisParam(axis, x.shape);
        const shapes = computeOutAndReduceShapes(x.shape, axes);
        const reduceShape = shapes[1];
        const reduceSize = sizeFromShape(reduceShape);
        const toDispose = [];
        const reduceSizeScalar = backend.makeTensorInfo([], 'float32', new Float32Array([reduceSize]));
        toDispose.push(reduceSizeScalar);
        const $x = cast$1({ inputs: { x }, backend, attrs: { dtype: 'float32' } });
        toDispose.push($x);
        const res = div({ inputs: { a: $x, b: reduceSizeScalar }, backend });
        toDispose.push(res);
        const result = sum$1({ inputs: { x: res }, backend, attrs: { axis, keepDims } });
        toDispose.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return result;
    }
    const meanConfig$1 = {
        kernelName: Mean,
        backendName: 'cpu',
        kernelFunc: mean
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function min$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        assertNotComplex$1(x, 'min');
        const origAxes = parseAxisParam(axis, x.shape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, x.shape.length);
        let $x = x;
        if (permutedAxes != null) {
            $x = transpose$1({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            axes = getInnerMostAxes(axes.length, x.shape.length);
        }
        assertAxesAreInnerMostDims('min', axes, $x.shape.length);
        const [outShape, reduceShape] = computeOutAndReduceShapes($x.shape, axes);
        const reduceSize = sizeFromShape(reduceShape);
        const vals = makeZerosTypedArray(sizeFromShape(outShape), $x.dtype);
        const aVals = backend.data.get($x.dataId).values;
        for (let i = 0; i < vals.length; ++i) {
            const offset = i * reduceSize;
            let min = aVals[offset];
            for (let j = 0; j < reduceSize; ++j) {
                const value = aVals[offset + j];
                if (Number.isNaN(value) ||
                    value < min) { // comparison with NaN always return false
                    min = value;
                }
            }
            vals[i] = min;
        }
        if (permutedAxes != null) {
            backend.disposeIntermediateTensorInfo($x);
        }
        const result = backend.makeTensorInfo(outShape, $x.dtype, vals);
        if (keepDims) {
            const expandedShape = expandShapeToKeepDim(outShape, origAxes);
            const reshapedResult = reshape$1({ inputs: { x: result }, backend, attrs: { shape: expandedShape } });
            backend.disposeIntermediateTensorInfo(result);
            return reshapedResult;
        }
        return result;
    }
    const minConfig$1 = {
        kernelName: Min,
        backendName: 'cpu',
        kernelFunc: min$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function mirrorPad(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { paddings, mode } = attrs;
        assertNotComplex$1(x, 'mirrorPad');
        const outShape = paddings.map((p, i) => p[0] /* beforePad */ + x.shape[i] + p[1] /* afterPad */);
        const start = paddings.map(p => p[0]);
        const end = paddings.map((p, i) => p[0] + x.shape[i]);
        const offset = mode === 'reflect' ? 0 : 1;
        const xVals = backend.data.get(x.dataId).values;
        const xRank = x.shape.length;
        const xStrides = computeStrides(x.shape);
        const resultSize = sizeFromShape(outShape);
        const resultRank = outShape.length;
        const resultStrides = computeStrides(outShape);
        const resVals = getTypedArrayFromDType(x.dtype, resultSize);
        for (let i = 0; i < resultSize; i++) {
            let coords = indexToLoc(i, resultRank, resultStrides);
            for (let i = 0; i < resultRank; i++) {
                if (coords[i] < start[i]) {
                    coords[i] = start[i] * 2 - coords[i] - offset;
                }
                else if (coords[i] >= end[i]) {
                    coords[i] = (end[i] - 1) * 2 - coords[i] + offset;
                }
            }
            coords = coords.map((c, i) => c - start[i]);
            const inIndex = locToIndex(coords, xRank, xStrides);
            resVals[i] = xVals[inIndex];
        }
        const outId = backend.write(resVals, outShape, x.dtype);
        return { dataId: outId, shape: outShape, dtype: x.dtype };
    }
    const mirrorPadConfig$1 = {
        kernelName: MirrorPad,
        backendName: 'cpu',
        kernelFunc: mirrorPad
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const modImpl = createSimpleBinaryKernelImpl(((aValue, bValue) => {
        const rem = aValue % bValue;
        if ((aValue < 0 && bValue < 0) || (aValue >= 0 && bValue >= 0)) {
            return rem;
        }
        else {
            return (rem + bValue) % bValue;
        }
    }));
    const mod$1 = binaryKernelFunc$1(Mod, modImpl);
    const modConfig$1 = {
        kernelName: Mod,
        backendName: 'cpu',
        kernelFunc: mod$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function softmax$1(args) {
        const { inputs, backend, attrs } = args;
        const { logits } = inputs;
        const { dim } = attrs;
        const logitsRank = logits.shape.length;
        let $dim = dim;
        if ($dim === -1) {
            $dim = logitsRank - 1;
        }
        if ($dim !== logitsRank - 1) {
            throw Error('Softmax along a non-last dimension is not yet supported. ' +
                `Logits was rank ${logitsRank} and dim was ${$dim}`);
        }
        const axes = parseAxisParam([$dim], logits.shape);
        const maxLogit = max$1({
            inputs: { x: logits },
            backend,
            attrs: { reductionIndices: axes, keepDims: false }
        });
        const expandedShape = expandShapeToKeepDim(maxLogit.shape, axes);
        const maxLogitReshaped = reshape$1({ inputs: { x: maxLogit }, backend, attrs: { shape: expandedShape } });
        const a = sub$1({ inputs: { a: logits, b: maxLogitReshaped }, backend });
        const b = exp$1({ inputs: { x: a }, backend });
        const sumExp = sum$1({ inputs: { x: b }, backend, attrs: { axis: axes, keepDims: false } });
        const sumReshaped = reshape$1({ inputs: { x: sumExp }, backend, attrs: { shape: expandedShape } });
        const result = div({ inputs: { a: b, b: sumReshaped }, backend });
        backend.disposeIntermediateTensorInfo(maxLogit);
        backend.disposeIntermediateTensorInfo(maxLogitReshaped);
        backend.disposeIntermediateTensorInfo(a);
        backend.disposeIntermediateTensorInfo(b);
        backend.disposeIntermediateTensorInfo(sumExp);
        backend.disposeIntermediateTensorInfo(sumReshaped);
        return result;
    }
    const softmaxConfig$1 = {
        kernelName: Softmax$2,
        backendName: 'cpu',
        kernelFunc: softmax$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function multinomial$1(args) {
        const { inputs, backend, attrs } = args;
        const { logits } = inputs;
        const { numSamples, seed, normalized } = attrs;
        assertNotComplex$1(logits, 'multinomial');
        const probabilities = normalized ?
            logits :
            softmax$1({ inputs: { logits }, backend, attrs: { dim: -1 } });
        const batchSize = probabilities.shape[0];
        const numEvents = probabilities.shape[1];
        const probVals = backend.data.get(probabilities.dataId).values;
        const resShape = [batchSize, numSamples];
        const resVals = makeZerosTypedArray(sizeFromShape(resShape), 'int32');
        for (let b = 0; b < batchSize; ++b) {
            const offset = b * numEvents;
            // The cdf won't include the last event. It will be implicit if no other
            // event happened.
            const cdf = new Float32Array(numEvents - 1);
            cdf[0] = probVals[offset];
            for (let event = 1; event < cdf.length; ++event) {
                cdf[event] = cdf[event - 1] + probVals[offset + event];
            }
            const random = seedrandom.alea(seed.toString());
            const outOffset = b * numSamples;
            for (let sampleId = 0; sampleId < numSamples; ++sampleId) {
                const r = random();
                // Assume last event happened by default.
                resVals[outOffset + sampleId] = cdf.length;
                for (let event = 0; event < cdf.length; event++) {
                    if (r < cdf[event]) {
                        resVals[outOffset + sampleId] = event;
                        break;
                    }
                }
            }
        }
        if (!normalized) {
            backend.disposeIntermediateTensorInfo(probabilities);
        }
        return backend.makeTensorInfo(resShape, 'int32', resVals);
    }
    const multinomialConfig$1 = {
        kernelName: Multinomial,
        backendName: 'cpu',
        kernelFunc: multinomial$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const nonMaxSuppressionV3Impl$1 = nonMaxSuppressionV3Impl$2;
    function nonMaxSuppressionV3$1(args) {
        const { inputs, backend, attrs } = args;
        const { boxes, scores } = inputs;
        const { maxOutputSize, iouThreshold, scoreThreshold } = attrs;
        assertNotComplex$1(boxes, 'NonMaxSuppression');
        const boxesVals = backend.data.get(boxes.dataId).values;
        const scoresVals = backend.data.get(scores.dataId).values;
        const { selectedIndices } = nonMaxSuppressionV3Impl$1(boxesVals, scoresVals, maxOutputSize, iouThreshold, scoreThreshold);
        return backend.makeTensorInfo([selectedIndices.length], 'int32', new Int32Array(selectedIndices));
    }
    const nonMaxSuppressionV3Config$1 = {
        kernelName: NonMaxSuppressionV3,
        backendName: 'cpu',
        kernelFunc: nonMaxSuppressionV3$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const nonMaxSuppressionV4Impl$1 = nonMaxSuppressionV4Impl$2;
    function nonMaxSuppressionV4$1(args) {
        const { inputs, backend, attrs } = args;
        const { boxes, scores } = inputs;
        const { maxOutputSize, iouThreshold, scoreThreshold, padToMaxOutputSize } = attrs;
        assertNotComplex$1(boxes, 'NonMaxSuppressionPadded');
        const boxesVals = backend.data.get(boxes.dataId).values;
        const scoresVals = backend.data.get(scores.dataId).values;
        const { selectedIndices, validOutputs } = nonMaxSuppressionV4Impl$1(boxesVals, scoresVals, maxOutputSize, iouThreshold, scoreThreshold, padToMaxOutputSize);
        return [
            backend.makeTensorInfo([selectedIndices.length], 'int32', new Int32Array(selectedIndices)),
            backend.makeTensorInfo([], 'int32', new Int32Array([validOutputs]))
        ];
    }
    const nonMaxSuppressionV4Config$1 = {
        kernelName: NonMaxSuppressionV4,
        backendName: 'cpu',
        kernelFunc: nonMaxSuppressionV4$1
    };

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const nonMaxSuppressionV5Impl$1 = nonMaxSuppressionV5Impl$2;
    function nonMaxSuppressionV5$1(args) {
        const { inputs, backend, attrs } = args;
        const { boxes, scores } = inputs;
        const { maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma } = attrs;
        assertNotComplex$1(boxes, 'NonMaxSuppressionWithScore');
        const boxesVals = backend.data.get(boxes.dataId).values;
        const scoresVals = backend.data.get(scores.dataId).values;
        const maxOutputSizeVal = maxOutputSize;
        const iouThresholdVal = iouThreshold;
        const scoreThresholdVal = scoreThreshold;
        const softNmsSigmaVal = softNmsSigma;
        const { selectedIndices, selectedScores } = nonMaxSuppressionV5Impl$1(boxesVals, scoresVals, maxOutputSizeVal, iouThresholdVal, scoreThresholdVal, softNmsSigmaVal);
        return [
            backend.makeTensorInfo([selectedIndices.length], 'int32', new Int32Array(selectedIndices)),
            backend.makeTensorInfo([selectedScores.length], 'float32', new Float32Array(selectedScores))
        ];
    }
    const nonMaxSuppressionV5Config$1 = {
        kernelName: NonMaxSuppressionV5,
        backendName: 'cpu',
        kernelFunc: nonMaxSuppressionV5$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function oneHot$1(args) {
        const { inputs, backend, attrs } = args;
        const { indices } = inputs;
        const { depth, onValue, offValue } = attrs;
        assertNotComplex$1(indices, 'oneHot');
        const indicesSize = sizeFromShape(indices.shape);
        const res = new Float32Array(indicesSize * depth);
        res.fill(offValue);
        const indicesVal = backend.data.get(indices.dataId).values;
        for (let event = 0; event < indicesSize; ++event) {
            if (indicesVal[event] >= 0 && indicesVal[event] < depth) {
                res[event * depth + indicesVal[event]] = onValue;
            }
        }
        return backend.makeTensorInfo([...indices.shape, depth], 'int32', res);
    }
    const oneHotConfig$1 = {
        kernelName: OneHot,
        backendName: 'cpu',
        kernelFunc: oneHot$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function zerosLike$1(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        if (x.dtype === 'string') {
            throw new Error('zerosLike is not supported for string tensors');
        }
        else if (x.dtype === 'complex64') {
            const realPart = real$1({ inputs: { input: x }, backend });
            const r = zerosLike$1({ inputs: { x: realPart }, backend });
            const imagPart = imag$1({ inputs: { input: x }, backend });
            const i = zerosLike$1({ inputs: { x: imagPart }, backend });
            const result = complex$1({ inputs: { real: r, imag: i }, backend });
            backend.disposeIntermediateTensorInfo(realPart);
            backend.disposeIntermediateTensorInfo(r);
            backend.disposeIntermediateTensorInfo(imagPart);
            backend.disposeIntermediateTensorInfo(i);
            return result;
        }
        else {
            return fill$1({ backend, attrs: { shape: x.shape, value: 0, dtype: x.dtype } });
        }
    }
    const zerosLikeConfig$1 = {
        kernelName: ZerosLike,
        backendName: 'cpu',
        kernelFunc: zerosLike$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function onesLike$1(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        if (x.dtype === 'string') {
            throw new Error('onesLike is not supported for string tensors');
        }
        else if (x.dtype === 'complex64') {
            const realPart = real$1({ inputs: { input: x }, backend });
            const r = onesLike$1({ inputs: { x: realPart }, backend });
            const imagPart = imag$1({ inputs: { input: x }, backend });
            const i = zerosLike$1({ inputs: { x: imagPart }, backend });
            const result = complex$1({ inputs: { real: r, imag: i }, backend });
            backend.disposeIntermediateTensorInfo(realPart);
            backend.disposeIntermediateTensorInfo(r);
            backend.disposeIntermediateTensorInfo(imagPart);
            backend.disposeIntermediateTensorInfo(i);
            return result;
        }
        else {
            return fill$1({ backend, attrs: { shape: x.shape, value: 1, dtype: x.dtype } });
        }
    }
    const onesLikeConfig$1 = {
        kernelName: OnesLike,
        backendName: 'cpu',
        kernelFunc: onesLike$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function pack$1(args) {
        const { inputs, backend, attrs } = args;
        const { axis } = attrs;
        if (inputs.length === 1) {
            return expandDims$1({ inputs: { input: inputs[0] }, backend, attrs: { dim: axis } });
        }
        const shape = inputs[0].shape;
        const dtype = inputs[0].dtype;
        inputs.forEach(t => {
            assertShapesMatch(shape, t.shape, 'All tensors passed to stack must have matching shapes');
            assert$4(dtype === t.dtype, () => 'All tensors passed to stack must have matching dtypes');
        });
        const intermediateTensorInfos = [];
        const expandedTensors = inputs.map(t => {
            const expandedT = expandDims$1({ inputs: { input: t }, backend, attrs: { dim: axis } });
            intermediateTensorInfos.push(expandedT);
            return expandedT;
        });
        const result = concat$1({ inputs: expandedTensors, backend, attrs: { axis } });
        intermediateTensorInfos.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return result;
    }
    const packConfig$1 = {
        kernelName: Pack,
        backendName: 'cpu',
        kernelFunc: pack$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function padV2$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { paddings, constantValue } = attrs;
        assertNotComplex$1(x, 'pad');
        const outShape = paddings.map((p, i) => p[0] /* beforePad */ + x.shape[i] + p[1] /* afterPad */);
        const start = paddings.map(p => p[0]);
        const xVals = backend.data.get(x.dataId).values;
        const xSize = sizeFromShape(x.shape);
        const xRank = x.shape.length;
        const xStrides = computeStrides(x.shape);
        const resultSize = sizeFromShape(outShape);
        const resultRank = outShape.length;
        const resultStrides = computeStrides(outShape);
        const resVals = getTypedArrayFromDType(x.dtype, resultSize);
        if (constantValue !== 0) {
            resVals.fill(constantValue);
        }
        for (let i = 0; i < xSize; i++) {
            const coords = indexToLoc(i, xRank, xStrides);
            const outCoords = coords.map((c, i) => c + start[i]);
            const outIndex = locToIndex(outCoords, resultRank, resultStrides);
            resVals[outIndex] = xVals[i];
        }
        const outId = backend.write(resVals, outShape, x.dtype);
        return { dataId: outId, shape: outShape, dtype: x.dtype };
    }
    const padV2Config$1 = {
        kernelName: PadV2,
        backendName: 'cpu',
        kernelFunc: padV2$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const powImpl = createSimpleBinaryKernelImpl((a, b) => Math.pow(a, b));
    const pow$1 = binaryKernelFunc$1(Pow, powImpl);
    const powConfig$1 = {
        kernelName: Pow,
        backendName: 'cpu',
        kernelFunc: pow$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function range$2(args) {
        const { backend, attrs } = args;
        const { start, stop, dtype, step } = attrs;
        const values = rangeImpl(start, stop, step, dtype);
        return backend.makeTensorInfo([values.length], dtype, values);
    }
    const rangeConfig$1 = {
        kernelName: Range,
        backendName: 'cpu',
        kernelFunc: range$2
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const reciprocal$1 = unaryKernelFunc$1(Reciprocal, (xi) => 1 / xi);
    const reciprocalConfig$1 = {
        kernelName: Reciprocal,
        backendName: 'cpu',
        kernelFunc: reciprocal$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function resizeBilinear$1(args) {
        const { inputs, backend, attrs } = args;
        const { images } = inputs;
        const { alignCorners, halfPixelCenters, size } = attrs;
        assertNotComplex$1(images, 'resizeBilinear');
        const imagesStrides = computeStrides(images.shape);
        const [newHeight, newWidth] = size;
        const [batch, oldHeight, oldWidth, numChannels] = images.shape;
        const xValues = backend.data.get(images.dataId).values;
        const result = new Float32Array(sizeFromShape([batch, newHeight, newWidth, numChannels]));
        const effectiveInputSize = [
            (alignCorners && newHeight > 1) ? oldHeight - 1 : oldHeight,
            (alignCorners && newWidth > 1) ? oldWidth - 1 : oldWidth
        ];
        const effectiveOutputSize = [
            (alignCorners && newHeight > 1) ? newHeight - 1 : newHeight,
            (alignCorners && newWidth > 1) ? newWidth - 1 : newWidth
        ];
        let outputIdx = 0;
        const effectiveRowSizeRatio = effectiveInputSize[0] / effectiveOutputSize[0];
        const effectiveColSizeRatio = effectiveInputSize[1] / effectiveOutputSize[1];
        for (let b = 0; b < batch; b++) {
            for (let r = 0; r < newHeight; r++) {
                let sourceFracRow;
                if (halfPixelCenters) {
                    sourceFracRow = effectiveRowSizeRatio * (r + 0.5) - 0.5;
                }
                else {
                    sourceFracRow = effectiveRowSizeRatio * r;
                }
                const sourceRowFloor = Math.max(0, Math.floor(sourceFracRow));
                const rowFrac = sourceFracRow - sourceRowFloor;
                const sourceRowCeil = Math.min(oldHeight - 1, Math.ceil(sourceFracRow));
                const topRowOffset = b * imagesStrides[0] + sourceRowFloor * imagesStrides[1];
                const botRowOffset = b * imagesStrides[0] + sourceRowCeil * imagesStrides[1];
                for (let c = 0; c < newWidth; c++) {
                    let sourceFracCol;
                    if (halfPixelCenters) {
                        sourceFracCol = effectiveColSizeRatio * (c + 0.5) - 0.5;
                    }
                    else {
                        sourceFracCol = effectiveColSizeRatio * c;
                    }
                    const sourceColFloor = Math.max(0, Math.floor(sourceFracCol));
                    const colFrac = sourceFracCol - sourceColFloor;
                    const sourceColCeil = Math.min(oldWidth - 1, Math.ceil(sourceFracCol));
                    const topLeftOffest = topRowOffset + sourceColFloor * imagesStrides[2];
                    const botLeftOffset = botRowOffset + sourceColFloor * imagesStrides[2];
                    const topRightOffset = topRowOffset + sourceColCeil * imagesStrides[2];
                    const botRightOffest = botRowOffset + sourceColCeil * imagesStrides[2];
                    for (let d = 0; d < numChannels; d++) {
                        // Begin shader.
                        // Compute the fractional index of the source.
                        const topLeft = xValues[topLeftOffest + d];
                        const bottomLeft = xValues[botLeftOffset + d];
                        const topRight = xValues[topRightOffset + d];
                        const bottomRight = xValues[botRightOffest + d];
                        const top = topLeft + (topRight - topLeft) * colFrac;
                        const bottom = bottomLeft + (bottomRight - bottomLeft) * colFrac;
                        const newValue = top + (bottom - top) * rowFrac;
                        result[outputIdx++] = newValue;
                    }
                }
            }
        }
        return backend.makeTensorInfo([batch, newHeight, newWidth, numChannels], 'float32', result);
    }
    const resizeBilinearConfig$1 = {
        kernelName: ResizeBilinear,
        backendName: 'cpu',
        kernelFunc: resizeBilinear$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function resizeBilinearGrad$1(args) {
        const { inputs, backend, attrs } = args;
        const { images, dy } = inputs;
        const { alignCorners } = attrs;
        assertNotComplex$1([dy, images], 'resizeBilinearGrad');
        const imagesStrides = computeStrides(images.shape);
        const [batch, xHeight, xWidth, depth] = images.shape;
        const [, yHeight, yWidth] = dy.shape;
        const output = new Float32Array(batch * xHeight * xWidth * depth);
        // In the backwards pass, we want to find the pixels that were generated
        // for each pixel in the input image the forward pass and add the
        // corresponding coefficient from dy to the gradient (with some
        // interpolation).
        const effectiveXSize = [
            (alignCorners && yHeight > 1) ? xHeight - 1 : xHeight,
            (alignCorners && yWidth > 1) ? xWidth - 1 : xWidth
        ];
        const effectiveYSize = [
            (alignCorners && yHeight > 1) ? yHeight - 1 : yHeight,
            (alignCorners && yWidth > 1) ? yWidth - 1 : yWidth
        ];
        const heightScale = effectiveXSize[0] / effectiveYSize[0];
        const widthScale = effectiveXSize[1] / effectiveYSize[1];
        // Reference implementation
        // tslint:disable-next-line:max-line-length
        // https://github.com/tensorflow/tensorflow/blob/3039375c86a5bbc9610c7725dcaa95d635f87ba2/tensorflow/core/kernels/resize_bilinear_op.cc#L275
        const dyValues = backend.data.get(dy.dataId).values;
        let offset = 0;
        for (let b = 0; b < batch; b++) {
            const bOffset = b * imagesStrides[0];
            for (let r = 0; r < yHeight; r++) {
                const dxR = r * heightScale;
                const topDxRIndex = Math.floor(dxR);
                const bottomDxRIndex = Math.min(Math.ceil(dxR), xHeight - 1);
                const topDxROffset = bOffset + topDxRIndex * imagesStrides[1];
                const bottomDxROffset = bOffset + bottomDxRIndex * imagesStrides[1];
                const dxRLerp = dxR - topDxRIndex;
                const inverseDxRLerp = 1.0 - dxRLerp;
                for (let c = 0; c < yWidth; c++) {
                    const dxC = c * widthScale;
                    const leftDxCIndex = Math.floor(dxC);
                    const rightDxCIndex = Math.min(Math.ceil(dxC), xWidth - 1);
                    const dxCLerp = dxC - leftDxCIndex;
                    const inverseDxCLerp = 1.0 - dxCLerp;
                    const topLeftRCOffset = topDxROffset + leftDxCIndex * imagesStrides[2];
                    const topRightRCOffset = topDxROffset + rightDxCIndex * imagesStrides[2];
                    const bottomLeftRCOffset = bottomDxROffset + leftDxCIndex * imagesStrides[2];
                    const bottomRightRCOffset = bottomDxROffset + rightDxCIndex * imagesStrides[2];
                    const inverseDxRLerpTimesInverseDxCLerp = inverseDxRLerp * inverseDxCLerp;
                    const inverseDxRLerpTimesDxCLerp = inverseDxRLerp * dxCLerp;
                    const dxRLerpTimesInverseDxCLerp = dxRLerp * inverseDxCLerp;
                    const dxRLerpTimesDxCLerp = dxRLerp * dxCLerp;
                    for (let d = 0; d < depth; d++) {
                        const dyVal = dyValues[offset++];
                        output[topLeftRCOffset + d] +=
                            dyVal * inverseDxRLerpTimesInverseDxCLerp;
                        output[topRightRCOffset + d] += dyVal * inverseDxRLerpTimesDxCLerp;
                        output[bottomLeftRCOffset + d] += dyVal * dxRLerpTimesInverseDxCLerp;
                        output[bottomRightRCOffset + d] += dyVal * dxRLerpTimesDxCLerp;
                    }
                }
            }
        }
        return backend.makeTensorInfo([batch, xWidth, xHeight, depth], 'float32', output);
    }
    const resizeBilinearGradConfig$1 = {
        kernelName: ResizeBilinearGrad,
        backendName: 'cpu',
        kernelFunc: resizeBilinearGrad$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function resizeNearestNeighbor$1(args) {
        const { inputs, backend, attrs } = args;
        const { images } = inputs;
        const { alignCorners, halfPixelCenters, size } = attrs;
        assertNotComplex$1(images, 'resizeNearestNeighbor');
        const imagesStrides = computeStrides(images.shape);
        const [newHeight, newWidth] = size;
        const [batch, oldHeight, oldWidth, numChannels] = images.shape;
        const xValues = backend.data.get(images.dataId).values;
        const output = new Float32Array(batch * newHeight * newWidth * numChannels);
        const effectiveInputSize = [
            (alignCorners && newHeight > 1) ? oldHeight - 1 : oldHeight,
            (alignCorners && newWidth > 1) ? oldWidth - 1 : oldWidth
        ];
        const effectiveOutputSize = [
            (alignCorners && newHeight > 1) ? newHeight - 1 : newHeight,
            (alignCorners && newWidth > 1) ? newWidth - 1 : newWidth
        ];
        const effectiveRowSizeRatio = effectiveInputSize[0] / effectiveOutputSize[0];
        const effectiveColSizeRatio = effectiveInputSize[1] / effectiveOutputSize[1];
        let outputOffset = 0;
        for (let b = 0; b < batch; b++) {
            const batchOffset = b * imagesStrides[0];
            for (let r = 0; r < newHeight; r++) {
                const sourceFracRow = halfPixelCenters ?
                    effectiveRowSizeRatio * (r + 0.5) :
                    effectiveRowSizeRatio * r;
                let sourceNearestRow = Math.min(oldHeight - 1, alignCorners ? Math.round(sourceFracRow) : Math.floor(sourceFracRow));
                if (halfPixelCenters) {
                    sourceNearestRow = Math.max(0, sourceNearestRow);
                }
                const rowOffset = batchOffset + sourceNearestRow * imagesStrides[1];
                for (let c = 0; c < newWidth; c++) {
                    const sourceFracCol = halfPixelCenters ?
                        effectiveColSizeRatio * (c + 0.5) :
                        effectiveColSizeRatio * c;
                    let sourceNearestCol = Math.min(oldWidth - 1, alignCorners ? Math.round(sourceFracCol) :
                        Math.floor(sourceFracCol));
                    if (halfPixelCenters) {
                        sourceNearestCol = Math.max(0, sourceNearestCol);
                    }
                    const colOffset = rowOffset + sourceNearestCol * imagesStrides[2];
                    for (let d = 0; d < numChannels; d++) {
                        // Begin shader.
                        // Compute the fractional index of the source.
                        const newVal = xValues[colOffset + d];
                        output[outputOffset++] = newVal;
                    }
                }
            }
        }
        return backend.makeTensorInfo([batch, newHeight, newWidth, numChannels], images.dtype, output);
    }
    const resizeNearestNeighborConfig$1 = {
        kernelName: ResizeNearestNeighbor,
        backendName: 'cpu',
        kernelFunc: resizeNearestNeighbor$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function resizeNearestNeighborGrad$1(args) {
        const { inputs, backend, attrs } = args;
        const { images, dy } = inputs;
        const { alignCorners } = attrs;
        assertNotComplex$1([dy, images], 'resizeNearestNeighborGrad');
        const imagesStrides = computeStrides(images.shape);
        const dyStrides = computeStrides(dy.shape);
        const [batch, xHeight, xWidth, depth] = images.shape;
        const [, yHeight, yWidth] = dy.shape;
        const output = new Float32Array(batch * xHeight * xWidth * depth);
        const dyValues = backend.data.get(dy.dataId).values;
        // In the backwards pass, we want to find the pixels that were generated
        // for each pixel in the input image the forward pass
        const effectiveXSize = [
            (alignCorners && yHeight > 1) ? xHeight - 1 : xHeight,
            (alignCorners && yWidth > 1) ? xWidth - 1 : xWidth
        ];
        const effectiveYSize = [
            (alignCorners && yHeight > 1) ? yHeight - 1 : yHeight,
            (alignCorners && yWidth > 1) ? yWidth - 1 : yWidth
        ];
        const heightScale = effectiveXSize[0] / effectiveYSize[0];
        const widthScale = effectiveXSize[1] / effectiveYSize[1];
        const invHeightScale = 1 / heightScale;
        const invWidthScale = 1 / widthScale;
        // This defines the size of the window of values around a particular
        // index in dy that we want to search for contributions to dx.
        const winHeight = (Math.ceil(invHeightScale) * 2) + 2;
        const winWidth = (Math.ceil(invWidthScale) * 2) + 2;
        // Loop over the output space.
        for (let b = 0; b < batch; b++) {
            const batchOffset = b * imagesStrides[0];
            for (let r = 0; r < xHeight; r++) {
                const rowOffset = batchOffset + r * imagesStrides[1];
                // Compute bounds for where in dy we will look
                const startRLerp = Math.floor(r * invHeightScale);
                const startDyR = Math.floor(startRLerp - (winHeight / 2));
                for (let c = 0; c < xWidth; c++) {
                    const colOffset = rowOffset + c * imagesStrides[2];
                    // Compute bounds for where in dy we will look
                    const startCLerp = Math.floor(c * invWidthScale);
                    const startDyC = Math.floor(startCLerp - (winWidth / 2));
                    for (let d = 0; d < depth; d++) {
                        let accum = 0;
                        // loop over dy
                        for (let dyRIndex = 0; dyRIndex < winHeight; dyRIndex++) {
                            const dyR = dyRIndex + startDyR;
                            // Guard against the window exceeding the bounds of dy
                            if (dyR < 0 || dyR >= yHeight) {
                                continue;
                            }
                            const dyROffset = batchOffset + dyR * dyStrides[1];
                            const sourceFracRow = dyR * heightScale;
                            const sourceNearestRow = Math.min(xHeight - 1, alignCorners ? Math.round(sourceFracRow) :
                                Math.floor(sourceFracRow));
                            if (r !== sourceNearestRow) {
                                continue;
                            }
                            for (let dyCIndex = 0; dyCIndex < winWidth; dyCIndex++) {
                                const dyC = dyCIndex + startDyC;
                                // Guard against the window exceeding the bounds of dy
                                if (dyC < 0 || dyC >= yWidth) {
                                    continue;
                                }
                                const dyCOffset = dyROffset + dyC * dyStrides[2];
                                const sourceFracCol = dyC * widthScale;
                                const sourceNearestCol = Math.min(xWidth - 1, alignCorners ? Math.round(sourceFracCol) :
                                    Math.floor(sourceFracCol));
                                if (c === sourceNearestCol) {
                                    accum += dyValues[dyCOffset + d];
                                }
                            }
                        }
                        output[colOffset + d] = accum;
                    }
                }
            }
        }
        return backend.makeTensorInfo(images.shape, images.dtype, output);
    }
    const resizeNearestNeighborGradConfig$1 = {
        kernelName: ResizeNearestNeighborGrad,
        backendName: 'cpu',
        kernelFunc: resizeNearestNeighborGrad$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function reverse$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { dims } = attrs;
        assertNotComplex$1(x, 'reverse');
        const xRank = x.shape.length;
        const $dims = parseAxisParam(dims, x.shape);
        if (xRank === 0) {
            return identity$1({ inputs: { x }, backend });
        }
        const outBuf = new TensorBuffer(x.shape, x.dtype);
        const xBuf = backend.bufferSync(x);
        for (let i = 0; i < outBuf.size; i++) {
            const outLoc = outBuf.indexToLoc(i);
            const inLoc = outLoc.slice();
            $dims.forEach(d => inLoc[d] = x.shape[d] - 1 - inLoc[d]);
            outBuf.set(xBuf.get(...inLoc), ...outLoc);
        }
        return backend.makeTensorInfo(outBuf.shape, outBuf.dtype, outBuf.values);
    }
    const reverseConfig$1 = {
        kernelName: Reverse,
        backendName: 'cpu',
        kernelFunc: reverse$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const rotateWithOffsetConfig$1 = {
        kernelName: RotateWithOffset,
        backendName: 'cpu',
        kernelFunc: ({ inputs, attrs, backend }) => {
            const { image } = inputs;
            const { radians, fillValue, center } = attrs;
            const cpuBackend = backend;
            const output = getTypedArrayFromDType(image.dtype, sizeFromShape(image.shape));
            const [batch, imageHeight, imageWidth, numChannels] = image.shape;
            const [centerX, centerY] = getImageCenter(center, imageHeight, imageWidth);
            const fullOpacityValue = 255;
            const sinFactor = Math.sin(radians);
            const cosFactor = Math.cos(radians);
            const imageVals = cpuBackend.data.get(image.dataId).values;
            for (let batchIdx = 0; batchIdx < batch; batchIdx++) {
                const batchOffset = batchIdx * imageWidth * imageHeight * numChannels;
                for (let row = 0; row < imageHeight; row++) {
                    const rowOffset = row * (imageWidth * numChannels);
                    for (let col = 0; col < imageWidth; col++) {
                        const colOffset = col * numChannels;
                        for (let channel = 0; channel < numChannels; channel++) {
                            const coords = [batch, row, col, channel];
                            const x = coords[2];
                            const y = coords[1];
                            // coordX/coordY are the result of rotating and translating x/y.
                            let coordX = (x - centerX) * cosFactor - (y - centerY) * sinFactor;
                            let coordY = (x - centerX) * sinFactor + (y - centerY) * cosFactor;
                            coordX = Math.round(coordX + centerX);
                            coordY = Math.round(coordY + centerY);
                            let outputValue = fillValue;
                            if (typeof fillValue !== 'number') {
                                if (channel === 3) {
                                    outputValue = fullOpacityValue;
                                }
                                else {
                                    outputValue = fillValue[channel];
                                }
                            }
                            // If the coordinate position falls within the image boundaries...
                            if (coordX >= 0 && coordX < imageWidth && coordY >= 0 &&
                                coordY < imageHeight) {
                                // set the output to the image value at the coordinate position.
                                const rotatedRowOffset = coordY * (imageWidth * numChannels);
                                const rotatedColOffset = coordX * numChannels;
                                const imageIdx = batchOffset + rotatedRowOffset + rotatedColOffset + channel;
                                outputValue = imageVals[imageIdx];
                            }
                            const outIdx = batchOffset + rowOffset + colOffset + channel;
                            output[outIdx] = outputValue;
                        }
                    }
                }
            }
            const dataId = cpuBackend.write(output, image.shape, image.dtype);
            return { dataId, shape: image.shape, dtype: image.dtype };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const round$1 = unaryKernelFunc$1(Round, (xi) => {
        // The algorithm is based on banker's rounding.
        const base = Math.floor(xi);
        if (xi - base < 0.5) {
            return Math.floor(xi);
        }
        else if (xi - base > 0.5) {
            return Math.ceil(xi);
        }
        else {
            if (base % 2.0 === 0.0) {
                return base;
            }
            else {
                return base + 1.0;
            }
        }
    });
    const roundConfig$1 = {
        kernelName: Round,
        backendName: 'cpu',
        kernelFunc: round$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function scatterNd$1(args) {
        const { inputs, backend, attrs } = args;
        const { indices, updates } = inputs;
        const { shape } = attrs;
        const { sliceRank, numUpdates, sliceSize, strides, outputSize } = calculateShapes(updates, indices, shape);
        const sumDupeIndices = true;
        const indicesBuf = backend.bufferSync(indices);
        const updatesBuf = backend.bufferSync(updates);
        const outBuf = scatterImpl(indicesBuf, updatesBuf, shape, outputSize, sliceSize, numUpdates, sliceRank, strides, 0 /* defaultValue */, sumDupeIndices);
        return backend.makeTensorInfo(shape, outBuf.dtype, outBuf.values);
    }
    const scatterNdConfig$1 = {
        kernelName: ScatterNd,
        backendName: 'cpu',
        kernelFunc: scatterNd$1
    };

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function lowerBound(array, value) {
        let left = 0;
        let right = array.length;
        let mid = 0;
        while (left < right) {
            mid = Math.floor((left + right) / 2);
            if (array[mid] < value) {
                left = mid + 1;
            }
            else {
                right = mid;
            }
        }
        return right;
    }
    function upperBound(array, value) {
        let left = 0;
        let right = array.length;
        let mid = 0;
        while (left < right) {
            mid = Math.floor((left + right) / 2);
            if (array[mid] <= value) {
                left = mid + 1;
            }
            else {
                right = mid;
            }
        }
        return right;
    }
    function searchSortedImpl(sortedInputs, values, batchSize, numInputs, numValues, side) {
        const output = getArrayFromDType('int32', batchSize * numValues);
        for (let b = 0; b < batchSize; ++b) {
            const sortedInputsSlice = sortedInputs.slice(b * numInputs, (b + 1) * numInputs);
            const outputOffset = b * numValues;
            for (let i = 0; i < numValues; ++i) {
                output[outputOffset + i] = side === 'left' ?
                    lowerBound(sortedInputsSlice, values[i + outputOffset]) :
                    upperBound(sortedInputsSlice, values[i + outputOffset]);
            }
        }
        return output;
    }

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function searchSorted$1(args) {
        const { inputs, backend, attrs } = args;
        const { sortedSequence, values } = inputs;
        const { side } = attrs;
        const $sortedSequence = backend.data.get(sortedSequence.dataId).values;
        const $values = backend.data.get(values.dataId).values;
        const output = searchSortedImpl($sortedSequence, $values, sortedSequence.shape[0], sortedSequence.shape[1], values.shape[1], side);
        return backend.makeTensorInfo(values.shape, 'int32', output);
    }
    const searchSortedConfig$1 = {
        kernelName: SearchSorted,
        backendName: 'cpu',
        kernelFunc: searchSorted$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function select$1(args) {
        const { inputs, backend } = args;
        const { condition, t, e } = inputs;
        assertNotComplex$1([condition, t, e], 'select');
        const conditionRank = condition.shape.length;
        const values = backend.data.get(condition.dataId).values;
        const tValues = backend.data.get(t.dataId).values;
        const eValues = backend.data.get(e.dataId).values;
        const resultDtype = upcastType(t.dtype, e.dtype);
        const newValues = makeZerosTypedArray(sizeFromShape(t.shape), resultDtype);
        let index = 0;
        const offset = conditionRank === 0 || conditionRank > 1 || t.shape.length === 1 ?
            1 :
            sizeFromShape(t.shape.slice(1));
        for (let i = 0; i < values.length; i++) {
            for (let j = 0; j < offset; j++) {
                if (values[i] === 1) {
                    newValues[index++] = tValues[i];
                }
                else {
                    newValues[index++] = eValues[i];
                }
            }
        }
        return backend.makeTensorInfo(t.shape, resultDtype, newValues);
    }
    const selectConfig$1 = {
        kernelName: Select,
        backendName: 'cpu',
        kernelFunc: select$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const scaleAlpha = SELU_SCALEALPHA;
    const scale = SELU_SCALE;
    const selu$1 = unaryKernelFunc$1(Selu$1, (xi) => {
        if (xi >= 0) {
            return scale * xi;
        }
        else {
            return scaleAlpha * (Math.exp(xi) - 1);
        }
    });
    const seluConfig$1 = {
        kernelName: Selu$1,
        backendName: 'cpu',
        kernelFunc: selu$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sign$1 = unaryKernelFunc$1(Sign, (xi) => {
        if (xi < 0) {
            return -1;
        }
        else if (xi > 0) {
            return 1;
        }
        else {
            return 0;
        }
    });
    const signConfig$1 = {
        kernelName: Sign,
        backendName: 'cpu',
        kernelFunc: sign$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sin$1 = unaryKernelFunc$1(Sin, (xi) => Math.sin(xi));
    const sinConfig$1 = {
        kernelName: Sin,
        backendName: 'cpu',
        kernelFunc: sin$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const sinh$1 = unaryKernelFunc$1(Sinh, (xi) => Math.sinh(xi));
    const sinhConfig$1 = {
        kernelName: Sinh,
        backendName: 'cpu',
        kernelFunc: sinh$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // mirrors the implementation of tf.nn.softplus: https://goo.gl/vkcvwX
    // epsilon is the difference between 1.0 and the next representable float.
    // For a single precision 32 bit float this should be 2^-23, see:
    // https://math.byu.edu/~schow/work/IEEEFloatingPoint.htm
    const epsilon = 1.1920928955078125e-7;
    const threshold = Math.log(epsilon) + 2.0;
    const softplus$1 = unaryKernelFunc$1(Softplus$1, (xi) => {
        // Value above which exp(x) may overflow, but softplus(x) == x
        // is within machine epsilon.
        const tooLarge = xi > -threshold;
        // Value below which exp(x) may underflow, but softplus(x) == exp(x)
        // is within machine epsilon.
        const tooSmall = xi < threshold;
        const expX = Math.exp(xi);
        let result;
        if (tooSmall) {
            result = expX;
        }
        else if (tooLarge) {
            result = xi;
        }
        else {
            result = Math.log(1.0 + expX);
        }
        return result;
    });
    const softplusConfig$1 = {
        kernelName: Softplus$1,
        backendName: 'cpu',
        kernelFunc: softplus$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function spaceToBatchND$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { blockShape, paddings } = attrs;
        assertNotComplex$1([x], 'spaceToBatchND');
        const prod = sizeFromShape(blockShape);
        const completePaddings = [[0, 0]];
        completePaddings.push(...paddings);
        for (let i = 1 + blockShape.length; i < x.shape.length; ++i) {
            completePaddings.push([0, 0]);
        }
        const paddedX = padV2Config$1.kernelFunc({
            inputs: { x },
            backend,
            attrs: { paddings: completePaddings, constantValue: 0 }
        });
        const reshapedPaddedShape = getReshaped(paddedX.shape, blockShape, prod, false);
        const permutedReshapedPaddedPermutation = getPermuted(reshapedPaddedShape.length, blockShape.length, false);
        const flattenShape = getReshapedPermuted(paddedX.shape, blockShape, prod, false);
        const reshapeInputs = { x: paddedX };
        const reshapeAttrs = { shape: reshapedPaddedShape };
        const paddedXReshaped = reshape$1({ inputs: reshapeInputs, backend, attrs: reshapeAttrs });
        const transposeInputs = { x: paddedXReshaped };
        const transposeAttrs = { perm: permutedReshapedPaddedPermutation };
        const paddedXT = transpose$1({ inputs: transposeInputs, backend, attrs: transposeAttrs });
        const resultReshapeInputs = { x: paddedXT };
        const resultReshapeAttrs = { shape: flattenShape };
        const result = reshape$1({ inputs: resultReshapeInputs, backend, attrs: resultReshapeAttrs });
        backend.disposeIntermediateTensorInfo(paddedX);
        backend.disposeIntermediateTensorInfo(paddedXReshaped);
        backend.disposeIntermediateTensorInfo(paddedXT);
        return result;
    }
    const spaceToBatchNDConfig$1 = {
        kernelName: SpaceToBatchND,
        backendName: 'cpu',
        kernelFunc: spaceToBatchND$1
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseFillEmptyRows$1(args) {
        const { inputs, backend } = args;
        const { indices, values, denseShape, defaultValue } = inputs;
        if (denseShape.shape.length !== 1) {
            throw new Error(`Dense shape must be a vector, saw:
        ${denseShape.shape}`);
        }
        if (indices.shape.length !== 2) {
            throw new Error(`Indices must be a matrix, saw:
        ${indices.shape}`);
        }
        if (values.shape.length !== 1) {
            throw new Error(`Values must be a vector, saw:
        ${values.shape}`);
        }
        if (defaultValue.shape.length !== 0) {
            throw new Error(`Default value must be a scalar, saw:
        ${defaultValue.shape}`);
        }
        const $indices = backend.data.get(indices.dataId).values;
        const $values = backend.data.get(values.dataId).values;
        const $denseShape = backend.data.get(denseShape.dataId).values;
        const $defaultValue = backend.data.get(defaultValue.dataId).values[0];
        const [outputIndices, outputIndicesShape, outputValues, emptyRowIndicator, reverseIndexMap] = sparseFillEmptyRowsImpl($indices, indices.shape, indices.dtype, $values, values.dtype, $denseShape, $defaultValue);
        return [
            backend.makeTensorInfo(outputIndicesShape, indices.dtype, outputIndices),
            backend.makeTensorInfo([outputIndicesShape[0]], values.dtype, outputValues),
            backend.makeTensorInfo([emptyRowIndicator.length], 'bool', new Uint8Array(emptyRowIndicator.map((value) => Number(value)))),
            backend.makeTensorInfo([reverseIndexMap.length], indices.dtype, new Int32Array(reverseIndexMap)),
        ];
    }
    const sparseFillEmptyRowsConfig$1 = {
        kernelName: SparseFillEmptyRows,
        backendName: 'cpu',
        kernelFunc: sparseFillEmptyRows$1,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseReshape$1(args) {
        const { inputs, backend } = args;
        const { inputIndices, inputShape, newShape } = inputs;
        if (inputIndices.shape.length !== 2) {
            throw new Error(`Input indices should be a matrix but received shape
        ${inputIndices.shape}`);
        }
        if (inputShape.shape.length !== 1) {
            throw new Error(`Input shape should be a vector but received shape
        ${inputShape.shape}`);
        }
        if (newShape.shape.length !== 1) {
            throw new Error(`Target shape should be a vector but received shape ${newShape.shape}`);
        }
        const $inputShape = Array.from(backend.data.get(inputShape.dataId).values);
        const $inputIndices = backend.data.get(inputIndices.dataId).values;
        const targetShape = Array.from(backend.data.get(newShape.dataId).values);
        const [newIndices, indicesShape, outputShape] = sparseReshapeImpl($inputIndices, inputIndices.shape, inputIndices.dtype, $inputShape, targetShape);
        return [
            backend.makeTensorInfo(indicesShape, inputIndices.dtype, newIndices),
            backend.makeTensorInfo([outputShape.length], newShape.dtype, new Int32Array(outputShape)),
        ];
    }
    const sparseReshapeConfig$1 = {
        kernelName: SparseReshape,
        backendName: 'cpu',
        kernelFunc: sparseReshape$1,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseSegmentMean$1(args) {
        const { inputs, backend } = args;
        const { data, indices, segmentIds } = inputs;
        if (data.shape.length < 1) {
            throw new Error(`Data should be at least 1 dimensional but received scalar`);
        }
        if (indices.shape.length !== 1) {
            throw new Error(`Indices should be a vector but received shape
          ${indices.shape}`);
        }
        if (segmentIds.shape.length !== 1) {
            throw new Error(`Segment ids should be a vector but received shape
          ${segmentIds.shape}`);
        }
        if (indices.shape[0] !== segmentIds.shape[0]) {
            throw new Error(`segmentIds and indices should have same size.`);
        }
        const $data = backend.data.get(data.dataId).values;
        const $indices = backend.data.get(indices.dataId).values;
        const $segmentIds = backend.data.get(segmentIds.dataId).values;
        const [outputData, outputDataShape] = sparseSegmentReductionImpl($data, data.shape, data.dtype, $indices, $segmentIds, true);
        return backend.makeTensorInfo(outputDataShape, data.dtype, outputData);
    }
    const sparseSegmentMeanConfig$1 = {
        kernelName: SparseSegmentMean,
        backendName: 'cpu',
        kernelFunc: sparseSegmentMean$1,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseSegmentSum$1(args) {
        const { inputs, backend } = args;
        const { data, indices, segmentIds } = inputs;
        if (data.shape.length < 1) {
            throw new Error(`Data should be at least 1 dimensional but received scalar`);
        }
        if (indices.shape.length !== 1) {
            throw new Error(`Indices should be a vector but received shape
         ${indices.shape}`);
        }
        if (segmentIds.shape.length !== 1) {
            throw new Error(`Segment ids should be a vector but received shape
         ${segmentIds.shape}`);
        }
        if (indices.shape[0] !== segmentIds.shape[0]) {
            throw new Error(`segmentIds and indices should have same size.`);
        }
        const $data = backend.data.get(data.dataId).values;
        const $indices = backend.data.get(indices.dataId).values;
        const $segmentIds = backend.data.get(segmentIds.dataId).values;
        const [outputData, outputDataShape] = sparseSegmentReductionImpl($data, data.shape, data.dtype, $indices, $segmentIds);
        return backend.makeTensorInfo(outputDataShape, data.dtype, outputData);
    }
    const sparseSegmentSumConfig$1 = {
        kernelName: SparseSegmentSum,
        backendName: 'cpu',
        kernelFunc: sparseSegmentSum$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseToDense$1(args) {
        const { inputs, backend, attrs } = args;
        const { sparseIndices, sparseValues, defaultValue } = inputs;
        const { outputShape } = attrs;
        const { sliceRank, numUpdates, sliceSize, strides, outputSize } = calculateShapes(sparseValues, sparseIndices, outputShape);
        const sumDupeIndices = false;
        const indicesBuf = backend.bufferSync(sparseIndices);
        let outBuf;
        switch (sparseValues.dtype) {
            case 'bool': {
                const updatesBuf = backend.bufferSync(sparseValues);
                const $defaultValue = Boolean(backend.data.get(defaultValue.dataId).values[0]);
                outBuf = scatterImpl(indicesBuf, updatesBuf, outputShape, outputSize, sliceSize, numUpdates, sliceRank, strides, $defaultValue, sumDupeIndices);
                break;
            }
            case 'float32': {
                const updatesBuf = backend.bufferSync(sparseValues);
                const $defaultValue = backend.data.get(defaultValue.dataId).values[0];
                outBuf = scatterImpl(indicesBuf, updatesBuf, outputShape, outputSize, sliceSize, numUpdates, sliceRank, strides, $defaultValue, sumDupeIndices);
                break;
            }
            case 'int32': {
                const updatesBuf = backend.bufferSync(sparseValues);
                const $defaultValue = backend.data.get(defaultValue.dataId).values[0];
                outBuf = scatterImpl(indicesBuf, updatesBuf, outputShape, outputSize, sliceSize, numUpdates, sliceRank, strides, $defaultValue, sumDupeIndices);
                break;
            }
            case 'string': {
                const updatesBuf = backend.bufferSync(sparseValues);
                const $defaultValue = decodeString(backend.data.get(defaultValue.dataId).values[0]);
                outBuf = scatterImpl(indicesBuf, updatesBuf, outputShape, outputSize, sliceSize, numUpdates, sliceRank, strides, $defaultValue, sumDupeIndices);
                break;
            }
            default:
                throw new Error(`Unsupported type ${sparseValues.dtype}`);
        }
        return backend.makeTensorInfo(outputShape, outBuf.dtype, outBuf.values);
    }
    const sparseToDenseConfig$1 = {
        kernelName: SparseToDense,
        backendName: 'cpu',
        kernelFunc: sparseToDense$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function splitV$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { numOrSizeSplits, axis } = attrs;
        const $axis = parseAxisParam(axis, x.shape)[0];
        const splitSizes = prepareSplitSize(x, numOrSizeSplits, $axis);
        const begin = new Array(x.shape.length).fill(0);
        const size = x.shape.slice();
        return splitSizes.map(s => {
            const sliceSize = [...size];
            sliceSize[$axis] = s;
            const sliceT = slice$1({ inputs: { x }, backend, attrs: { begin, size: sliceSize } });
            begin[$axis] += s;
            return sliceT;
        });
    }
    const splitVConfig$1 = {
        kernelName: SplitV,
        backendName: 'cpu',
        kernelFunc: splitV$1
    };

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const squareConfig$1 = {
        kernelName: Square,
        backendName: 'cpu',
        kernelFunc: ({ inputs, backend }) => {
            const { x } = inputs;
            const cpuBackend = backend;
            assertNotComplex$1(x, 'square');
            const values = cpuBackend.data.get(x.dataId).values;
            const newValues = new Float32Array(values.length);
            for (let i = 0; i < values.length; ++i) {
                const value = values[i];
                newValues[i] = value * value;
            }
            const dataId = cpuBackend.write(newValues, x.shape, x.dtype);
            return { dataId, shape: x.shape, dtype: x.dtype };
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const step$1 = unaryKernelFunc$1(Step, (xi, attrs) => {
        const stepAttrs = attrs;
        if (isNaN(xi)) {
            return NaN;
        }
        else {
            return xi > 0 ? 1 : stepAttrs.alpha;
        }
    });
    const stepConfig$1 = {
        kernelName: Step,
        backendName: 'cpu',
        kernelFunc: step$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stridedSlice$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { begin, end, strides, beginMask, endMask, ellipsisMask, newAxisMask, shrinkAxisMask } = attrs;
        assertNotComplex$1(x, 'stridedSlice');
        const { finalShapeSparse, finalShape, isIdentity, sliceDim0, isSimpleSlice, begin: $begin, end: $end, strides: $strides } = sliceInfo(x.shape, begin, end, strides, beginMask, endMask, ellipsisMask, newAxisMask, shrinkAxisMask);
        let result;
        // ref:
        // https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/kernels/strided_slice_op.cc
        if (isIdentity) {
            // Optimization #1, slice is a no-op plus reshape
            result = reshape$1({ inputs: { x }, backend, attrs: { shape: finalShape } });
        }
        else if (sliceDim0 || isSimpleSlice) {
            // Optimization #2, slice is memory contiguous (only occurs in dim 0)
            assert$4(x.shape.length >= 1, () => `Input must have rank at least 1, got: ${x.shape.length}`);
            const size = computeOutShape$2($begin, $end, $strides);
            // To tolerate begin[0] > end[0] (a 0-output slice), we min(begin, end).
            const sliced = slice$1({ inputs: { x }, backend, attrs: { begin: $begin, size } });
            result =
                reshape$1({ inputs: { x: sliced }, backend, attrs: { shape: finalShape } });
            backend.disposeIntermediateTensorInfo(sliced);
        }
        else {
            const xBuf = backend.bufferSync(x);
            const outBuf = stridedSliceImpl(finalShapeSparse, xBuf, $strides, $begin);
            result = backend.makeTensorInfo(finalShape, outBuf.dtype, outBuf.values);
        }
        return result;
    }
    const stridedSliceConfig$1 = {
        kernelName: StridedSlice,
        backendName: 'cpu',
        kernelFunc: stridedSlice$1
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stringNGrams$1(args) {
        const { inputs, backend, attrs } = args;
        const { separator, nGramWidths, leftPad, rightPad, padWidth, preserveShortSequences } = attrs;
        const { data, dataSplits } = inputs;
        const $data = backend.data.get(data.dataId).values;
        const $dataSplits = backend.data.get(dataSplits.dataId).values;
        const [nGrams, nGramsSplits] = stringNGramsImpl($data, $dataSplits, separator, nGramWidths, leftPad, rightPad, padWidth, preserveShortSequences);
        return [
            backend.makeTensorInfo([nGrams.length], 'string', nGrams),
            backend.makeTensorInfo(dataSplits.shape, 'int32', nGramsSplits),
        ];
    }
    const stringNGramsConfig$1 = {
        kernelName: StringNGrams,
        backendName: 'cpu',
        kernelFunc: stringNGrams$1,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stringSplit$1(args) {
        const { inputs, backend, attrs } = args;
        const { skipEmpty } = attrs;
        const { input, delimiter } = inputs;
        if (input.dtype !== 'string') {
            throw new Error('Input must be of datatype string');
        }
        if (input.shape.length !== 1) {
            throw new Error(`Input must be a vector, got shape: ${input.shape}`);
        }
        if (delimiter.shape.length !== 0) {
            throw new Error(`Delimiter must be a scalar, got shape: ${delimiter.shape}`);
        }
        const $input = backend.data.get(input.dataId).values;
        const $delimiter = backend.data.get(delimiter.dataId).values[0];
        const [indices, values, shape] = stringSplitImpl($input, $delimiter, skipEmpty);
        const outputSize = values.length;
        return [
            backend.makeTensorInfo([outputSize, 2], 'int32', indices),
            backend.makeTensorInfo([outputSize], 'string', values),
            backend.makeTensorInfo([2], 'int32', new Int32Array(shape))
        ];
    }
    const stringSplitConfig$1 = {
        kernelName: StringSplit,
        backendName: 'cpu',
        kernelFunc: stringSplit$1,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stringToHashBucketFast$1(args) {
        const { inputs, backend, attrs } = args;
        const { numBuckets } = attrs;
        const { input } = inputs;
        if (input.dtype !== 'string') {
            throw new Error('Input must be of datatype string');
        }
        if (numBuckets <= 0) {
            throw new Error(`Number of buckets must be at least 1`);
        }
        const $input = backend.data.get(input.dataId).values;
        const output = stringToHashBucketFastImpl($input, numBuckets);
        return backend.makeTensorInfo(input.shape, 'int32', output);
    }
    const stringToHashBucketFastConfig$1 = {
        kernelName: StringToHashBucketFast,
        backendName: 'cpu',
        kernelFunc: stringToHashBucketFast$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const tan$1 = unaryKernelFunc$1(Tan, (xi) => Math.tan(xi));
    const tanConfig$1 = {
        kernelName: Tan,
        backendName: 'cpu',
        kernelFunc: tan$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const tanh$1 = unaryKernelFunc$1(Tanh$1, (xi) => Math.tanh(xi));
    const tanhConfig$1 = {
        kernelName: Tanh$1,
        backendName: 'cpu',
        kernelFunc: tanh$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function tile$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { reps } = attrs;
        assertNotComplex$1(x, 'tile');
        const outBuf = tileImpl(backend.bufferSync(x), reps);
        return backend.makeTensorInfo(outBuf.shape, outBuf.dtype, outBuf.values);
    }
    const tileConfig$1 = {
        kernelName: Tile,
        backendName: 'cpu',
        kernelFunc: tile$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function topK$1(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { k, sorted } = attrs;
        assertNotComplex$1(x, 'topk');
        const xVals = backend.data.get(x.dataId).values;
        const [allTopKVals, allTopKIndices] = topKImpl(xVals, x.shape, x.dtype, k, sorted);
        return [
            backend.makeTensorInfo(allTopKVals.shape, allTopKVals.dtype, allTopKVals.values),
            backend.makeTensorInfo(allTopKIndices.shape, allTopKIndices.dtype, allTopKIndices.values)
        ];
    }
    const topKConfig$1 = {
        kernelName: TopK,
        backendName: 'cpu',
        kernelFunc: topK$1
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function transform$1(args) {
        const { inputs, attrs, backend } = args;
        const { image, transforms } = inputs;
        const { interpolation, fillMode, fillValue, outputShape } = attrs;
        const [batch, imageHeight, imageWidth, numChannels] = image.shape;
        const [outHeight, outWidth] = outputShape != null ? outputShape : [imageHeight, imageWidth];
        const outShape = [batch, outHeight, outWidth, numChannels];
        const strides = computeStrides(image.shape);
        const batchStride = strides[0];
        const rowStride = strides[1];
        const colStride = strides[2];
        const outVals = getTypedArrayFromDType(image.dtype, sizeFromShape(outShape));
        outVals.fill(fillValue);
        const imageVals = backend.data.get(image.dataId).values;
        const transformVals = backend.data.get(transforms.dataId).values;
        // Ref TF implementation:
        // https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/kernels/image/image_ops.h
        for (let b = 0; b < batch; ++b) {
            const transform = transforms.shape[0] === 1 ?
                transformVals :
                transformVals.subarray(b * 8, b * 8 + 8);
            for (let outY = 0; outY < outHeight; ++outY) {
                for (let outX = 0; outX < outWidth; ++outX) {
                    for (let channel = 0; channel < numChannels; ++channel) {
                        let val;
                        const projection = transform[6] * outX + transform[7] * outY + 1;
                        if (projection === 0) {
                            // Return the fill value for infinite coordinates,
                            // which are outside the input image
                            continue;
                        }
                        const inX = (transform[0] * outX + transform[1] * outY + transform[2]) /
                            projection;
                        const inY = (transform[3] * outX + transform[4] * outY + transform[5]) /
                            projection;
                        const x = mapCoord(inX, imageWidth, fillMode);
                        const y = mapCoord(inY, imageHeight, fillMode);
                        switch (interpolation) {
                            case 'nearest':
                                val = nearestInterpolation(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, b, y, x, channel, fillValue);
                                break;
                            case 'bilinear':
                                val = bilinearInterpolation(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, b, y, x, channel, fillValue);
                                break;
                            default:
                                throw new Error(`Error in Transform: Expect 'nearest' or ` +
                                    `'bilinear', but got ${interpolation}`);
                        }
                        const ind = b * batchStride + outY * rowStride + outX * colStride + channel;
                        outVals[ind] = val;
                    }
                }
            }
            return backend.makeTensorInfo(outShape, image.dtype, outVals);
        }
        const dataId = backend.write(outVals, outShape, image.dtype);
        return { dataId, shape: image.shape, dtype: image.dtype };
    }
    const transformConfig$1 = {
        kernelName: Transform$1,
        backendName: 'cpu',
        kernelFunc: transform$1
    };
    function mapCoord(outCoord, len, mode) {
        switch (mode) {
            case 'reflect':
                return mapCoordReflect(outCoord, len);
            case 'wrap':
                return mapCoordWrap(outCoord, len);
            case 'nearest':
                return mapCoordNearest(outCoord, len);
            case 'constant':
            default:
                return mapCoordConstant(outCoord);
        }
    }
    function mapCoordReflect(outCoord, len) {
        // Reflect [abcd] to [dcba|abcd|dcba].
        let inCoord = outCoord;
        if (inCoord < 0) {
            if (len <= 1) {
                inCoord = 0;
            }
            else {
                const sz2 = 2 * len;
                if (inCoord < sz2) {
                    inCoord = sz2 * Math.trunc(-inCoord / sz2) + inCoord;
                }
                inCoord = inCoord < -len ? inCoord + sz2 : -inCoord - 1;
            }
        }
        else if (inCoord > len - 1) {
            if (len <= 1) {
                inCoord = 0;
            }
            else {
                const sz2 = 2 * len;
                inCoord -= sz2 * Math.trunc(inCoord / sz2);
                if (inCoord >= len) {
                    inCoord = sz2 - inCoord - 1;
                }
            }
        }
        // clamp is necessary because when outCoord = 3.5 and len = 4,
        // inCoord = 3.5 and will be rounded to 4 in nearest interpolation.
        return clamp(0, inCoord, len - 1);
    }
    function mapCoordWrap(outCoord, len) {
        // Wrap [abcd] to [abcd|abcd|abcd].
        let inCoord = outCoord;
        if (inCoord < 0) {
            if (len <= 1) {
                inCoord = 0;
            }
            else {
                const sz = len - 1;
                inCoord += len * (Math.trunc(-inCoord / sz) + 1);
            }
        }
        else if (inCoord > len - 1) {
            if (len <= 1) {
                inCoord = 0;
            }
            else {
                const sz = len - 1;
                inCoord -= len * Math.trunc(inCoord / sz);
            }
        }
        // clamp is necessary because when outCoord = -0.5 and len = 4,
        // inCoord = 3.5 and will be rounded to 4 in nearest interpolation.
        return clamp(0, inCoord, len - 1);
    }
    function mapCoordConstant(outCoord, len) {
        return outCoord;
    }
    function mapCoordNearest(outCoord, len) {
        return clamp(0, outCoord, len - 1);
    }
    function readWithFillValue(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, batch, y, x, channel, fillValue) {
        const ind = batch * batchStride + y * rowStride + x * colStride + channel;
        if (0 <= y && y < imageHeight && 0 <= x && x < imageWidth) {
            return imageVals[ind];
        }
        else {
            return fillValue;
        }
    }
    function nearestInterpolation(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, batch, y, x, channel, fillValue) {
        const $y = Math.round(y);
        const $x = Math.round(x);
        return readWithFillValue(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, batch, $y, $x, channel, fillValue);
    }
    function bilinearInterpolation(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, batch, y, x, channel, fillValue) {
        const yFloor = Math.floor(y);
        const xFloor = Math.floor(x);
        const yCeil = yFloor + 1;
        const xCeil = xFloor + 1;
        // f(x, yFloor) = (xCeil - x) / (xCeil - xFloor) * f(xFloor, yFloor)
        //               + (x - xFloor) / (xCeil - xFloor) * f(xCeil, yFloor)
        const valueYFloor = (xCeil - x) *
            readWithFillValue(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, batch, yFloor, xFloor, channel, fillValue) +
            (x - xFloor) *
                readWithFillValue(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, batch, yFloor, xCeil, channel, fillValue);
        // f(x, yCeil) = (xCeil - x) / (xCeil - xFloor) * f(xFloor, yCeil)
        //             + (x - xFloor) / (xCeil - xFloor) * f(xCeil, yCeil)
        const valueYCeil = (xCeil - x) *
            readWithFillValue(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, batch, yCeil, xFloor, channel, fillValue) +
            (x - xFloor) *
                readWithFillValue(imageVals, imageHeight, imageWidth, batchStride, rowStride, colStride, batch, yCeil, xCeil, channel, fillValue);
        // f(x, y) = (yCeil - y) / (yCeil - yFloor) * f(x, yFloor)
        //         + (y - yFloor) / (yCeil - yFloor) * f(x, yCeil)
        return (yCeil - y) * valueYFloor + (y - yFloor) * valueYCeil;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function unique$1(args) {
        const { inputs, attrs, backend } = args;
        const { axis } = attrs;
        const { x } = inputs;
        assertNotComplex$1(x, 'unique');
        const values = backend.data.get(x.dataId).values;
        const { outputValues, outputShape, indices } = uniqueImpl(values, axis, x.shape, x.dtype);
        return [
            backend.makeTensorInfo(outputShape, x.dtype, outputValues),
            backend.makeTensorInfo([indices.length], 'int32', indices),
        ];
    }
    const uniqueConfig$1 = {
        kernelName: Unique,
        backendName: 'cpu',
        kernelFunc: unique$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function unpack$1(args) {
        const { inputs, backend, attrs } = args;
        const { value } = inputs;
        let { axis } = attrs;
        if (axis < 0) {
            axis += value.shape.length;
        }
        const valueRank = value.shape.length;
        const num = value.shape[axis];
        const outShape = new Array(valueRank - 1);
        let outIndex = 0;
        for (let i = 0; i < valueRank; i++) {
            if (i !== axis) {
                outShape[outIndex++] = value.shape[i];
            }
        }
        const begin = new Array(valueRank).fill(0);
        const size = value.shape.slice();
        size[axis] = 1;
        const res = new Array(num);
        for (let i = 0; i < res.length; i++) {
            begin[axis] = i;
            const tempRes = slice$1({ inputs: { x: value }, backend, attrs: { begin, size } });
            res[i] = reshape$1({ inputs: { x: tempRes }, backend, attrs: { shape: outShape } });
            backend.disposeIntermediateTensorInfo(tempRes);
        }
        return res;
    }
    const unpackConfig$1 = {
        kernelName: Unpack,
        backendName: 'cpu',
        kernelFunc: unpack$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function unsortedSegmentSum$1(args) {
        const { inputs, backend, attrs } = args;
        const { x, segmentIds } = inputs;
        const { numSegments } = attrs;
        assertNotComplex$1(x, 'unsortedSegmentSum');
        const xRank = x.shape.length;
        const segmentIdsRank = segmentIds.shape.length;
        const res = [];
        const intermediates = [];
        // Reshape the segment id's so that they can be broadcast with
        // x. The new shape should be [segmentIds.shape, 1, ..., 1]
        const numIters = xRank - segmentIdsRank;
        let $segmentIds = segmentIds;
        for (let i = 0; i < numIters; ++i) {
            const expanded = expandDims$1({ inputs: { input: $segmentIds }, backend, attrs: { dim: i + 1 } });
            $segmentIds = expanded;
            intermediates.push(expanded);
        }
        for (let i = 0; i < numSegments; ++i) {
            const scalarValue = createScalarValue(i, 'int32');
            const segmentId = backend.makeTensorInfo([], 'int32', scalarValue);
            const mask = equal$2({ inputs: { a: segmentId, b: $segmentIds }, backend });
            const maskCasted = cast$1({ inputs: { x: mask }, backend, attrs: { dtype: 'float32' } });
            const mul = multiply$1({ inputs: { a: maskCasted, b: x }, backend });
            const sumTensorInfo = sum$1({ inputs: { x: mul }, backend, attrs: { axis: 0, keepDims: false } });
            res.push(sumTensorInfo);
            intermediates.push(segmentId);
            intermediates.push(mask);
            intermediates.push(maskCasted);
            intermediates.push(mul);
            intermediates.push(sumTensorInfo);
        }
        const result = pack$1({ inputs: res, backend, attrs: { axis: 0 } });
        intermediates.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return result;
    }
    const unsortedSegmentSumConfig$1 = {
        kernelName: UnsortedSegmentSum,
        backendName: 'cpu',
        kernelFunc: unsortedSegmentSum$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // List all kernel configs here
    const kernelConfigs$1 = [
        _fusedMatMulConfig$1,
        absConfig$1,
        acosConfig$1,
        acoshConfig$1,
        addConfig$1,
        addNConfig$1,
        allConfig$1,
        anyConfig$1,
        argMaxConfig$1,
        argMinConfig$1,
        asinConfig$1,
        asinhConfig$1,
        atanConfig$1,
        atan2Config$1,
        atanhConfig$1,
        avgPoolConfig$1,
        avgPool3DConfig$1,
        avgPool3DGradConfig$1,
        avgPoolGradConfig$1,
        batchMatMulConfig$1,
        batchNormConfig$1,
        batchToSpaceNDConfig$1,
        bincountConfig$1,
        broadcastArgsConfig$1,
        castConfig$1,
        ceilConfig$1,
        clipByValueConfig$1,
        complexConfig$1,
        complexAbsConfig$1,
        concatConfig$1,
        conv2DConfig$1,
        conv2DBackpropFilterConfig$1,
        conv2DBackpropInputConfig$1,
        conv3DConfig$1,
        conv3DBackpropFilterV2Config$1,
        conv3DBackpropInputV2Config,
        cosConfig$1,
        coshConfig$1,
        cropAndResizeConfig$1,
        cumprodConfig$1,
        cumsumConfig$1,
        denseBincountConfig$1,
        depthToSpaceConfig$1,
        depthwiseConv2dNativeConfig$1,
        depthwiseConv2dNativeBackpropFilterConfig$1,
        depthwiseConv2dNativeBackpropInputConfig$1,
        diagConfig$1,
        dilation2DConfig$1,
        dilation2DBackpropFilterConfig,
        dilation2DBackpropInputConfig,
        einsumConfig$1,
        eluConfig$1,
        eluGradConfig$1,
        equalConfig$1,
        erfConfig$1,
        expConfig$1,
        expandDimsConfig$1,
        expm1Config$1,
        fftConfig$1,
        fillConfig$1,
        flipLeftRightConfig$1,
        floorConfig$1,
        floorDivConfig$1,
        fusedConv2DConfig$1,
        fusedDepthwiseConv2DConfig$1,
        gatherNdConfig$1,
        gatherV2Config$1,
        greaterConfig$1,
        greaterEqualConfig$1,
        identityConfig$1,
        ifftConfig$1,
        imagConfig$1,
        isFiniteConfig$1,
        isInfConfig$1,
        isNaNConfig$1,
        leakyReluConfig$1,
        lessConfig$1,
        lessEqualConfig$1,
        linSpaceConfig$1,
        logConfig$1,
        log1pConfig$1,
        logicalAndConfig$1,
        logicalNotConfig$1,
        logicalOrConfig$1,
        LRNConfig$1,
        LRNGradConfig$1,
        maxConfig$1,
        maximumConfig$1,
        maxPoolConfig$1,
        maxPool3DConfig$1,
        maxPool3DGradConfig$1,
        maxPoolGradConfig$1,
        maxPoolWithArgmaxConfig$1,
        meanConfig$1,
        minConfig$1,
        minimumConfig$1,
        mirrorPadConfig$1,
        modConfig$1,
        multinomialConfig$1,
        multiplyConfig$1,
        negConfig$1,
        nonMaxSuppressionV3Config$1,
        nonMaxSuppressionV4Config$1,
        nonMaxSuppressionV5Config$1,
        notEqualConfig$1,
        oneHotConfig$1,
        onesLikeConfig$1,
        packConfig$1,
        padV2Config$1,
        powConfig$1,
        preluConfig$1,
        prodConfig$1,
        rangeConfig$1,
        realConfig$1,
        realDivConfig$1,
        reciprocalConfig$1,
        reluConfig$1,
        relu6Config$1,
        reshapeConfig$1,
        resizeBilinearConfig$1,
        resizeBilinearGradConfig$1,
        resizeNearestNeighborConfig$1,
        resizeNearestNeighborGradConfig$1,
        reverseConfig$1,
        rotateWithOffsetConfig$1,
        roundConfig$1,
        rsqrtConfig$1,
        scatterNdConfig$1,
        searchSortedConfig$1,
        selectConfig$1,
        seluConfig$1,
        sigmoidConfig$1,
        signConfig$1,
        sinConfig$1,
        sinhConfig$1,
        sliceConfig$1,
        softmaxConfig$1,
        softplusConfig$1,
        spaceToBatchNDConfig$1,
        sparseFillEmptyRowsConfig$1,
        sparseReshapeConfig$1,
        sparseSegmentMeanConfig$1,
        sparseSegmentSumConfig$1,
        sparseToDenseConfig$1,
        splitVConfig$1,
        sqrtConfig$1,
        squareConfig$1,
        squaredDifferenceConfig$1,
        stepConfig$1,
        stridedSliceConfig$1,
        stringNGramsConfig$1,
        stringSplitConfig$1,
        stringToHashBucketFastConfig$1,
        subConfig$1,
        sumConfig$1,
        tanConfig$1,
        tanhConfig$1,
        tileConfig$1,
        topKConfig$1,
        transformConfig$1,
        transposeConfig$1,
        uniqueConfig$1,
        unpackConfig$1,
        unsortedSegmentSumConfig$1,
        zerosLikeConfig$1
    ];
    for (const kernelConfig of kernelConfigs$1) {
        registerKernel(kernelConfig);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const contexts = {};
    const WEBGL_ATTRIBUTES = {
        alpha: false,
        antialias: false,
        premultipliedAlpha: false,
        preserveDrawingBuffer: false,
        depth: false,
        stencil: false,
        failIfMajorPerformanceCaveat: true
    };
    function setWebGLContext(webGLVersion, gl) {
        contexts[webGLVersion] = gl;
    }
    function getWebGLContext(webGLVersion, customCanvas) {
        if (!(webGLVersion in contexts) || customCanvas != null) {
            const newCtx = getWebGLRenderingContext(webGLVersion, customCanvas);
            if (newCtx !== null) {
                contexts[webGLVersion] = newCtx;
            }
            else {
                console.log('Could not get context for WebGL version', webGLVersion);
                return null;
            }
        }
        const gl = contexts[webGLVersion];
        if (gl == null || gl.isContextLost()) {
            delete contexts[webGLVersion];
            return getWebGLContext(webGLVersion);
        }
        gl.disable(gl.DEPTH_TEST);
        gl.disable(gl.STENCIL_TEST);
        gl.disable(gl.BLEND);
        gl.disable(gl.DITHER);
        gl.disable(gl.POLYGON_OFFSET_FILL);
        gl.disable(gl.SAMPLE_COVERAGE);
        gl.enable(gl.SCISSOR_TEST);
        gl.enable(gl.CULL_FACE);
        gl.cullFace(gl.BACK);
        return contexts[webGLVersion];
    }
    function createCanvas(webGLVersion) {
        if (typeof OffscreenCanvas !== 'undefined' && webGLVersion === 2) {
            return new OffscreenCanvas(300, 150);
        }
        else if (typeof document !== 'undefined') {
            return document.createElement('canvas');
        }
        else {
            throw new Error('Cannot create a canvas in this context');
        }
    }
    function getWebGLRenderingContext(webGLVersion, customCanvas) {
        if (webGLVersion !== 1 && webGLVersion !== 2) {
            throw new Error('Cannot get WebGL rendering context, WebGL is disabled.');
        }
        const canvas = customCanvas == null ? createCanvas(webGLVersion) : customCanvas;
        canvas.addEventListener('webglcontextlost', (ev) => {
            ev.preventDefault();
            delete contexts[webGLVersion];
        }, false);
        if (webGLVersion === 1) {
            return (canvas.getContext('webgl', WEBGL_ATTRIBUTES) ||
                canvas.getContext('experimental-webgl', WEBGL_ATTRIBUTES));
        }
        return canvas.getContext('webgl2', WEBGL_ATTRIBUTES);
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    var PackingScheme;
    (function (PackingScheme) {
        /**
         * All values in a single texel are densely packed without any constraints.
         *
         * This is how the shader encodes a tensor with shape = [2, 3, 4]
         * (indices are [batch, row, col]).
         *
         * 000|001   010|011   020|021
         * -------   -------   -------
         * 002|003   012|013   022|023
         *
         * 100|101   110|111   120|121
         * -------   -------   -------
         * 102|103   112|113   122|123
         *
         */
        PackingScheme[PackingScheme["DENSE"] = 0] = "DENSE";
        /**
         * Single texels contain only values from the same batch, and from adjacent
         * rows and columns.
         *
         * This is how the shader encodes a tensor with shape = [2, 3, 5]
         * (indices are [batch, row, col]).
         *
         * 000|001   002|003   004|xxx   020|021   022|023   024|xxx
         * -------   -------   -------   -------   -------   -------
         * 010|011   012|013   014|xxx   xxx|xxx   xxx|xxx   xxx|xxx
         *
         * 100|101   102|103   104|xxx   120|121   122|123   124|xxx
         * -------   -------   -------   -------   -------   -------
         * 110|111   112|113   114|xxx   xxx|xxx   xxx|xxx   xxx|xxx
         *
         */
        PackingScheme[PackingScheme["SHARED_BATCH"] = 1] = "SHARED_BATCH";
    })(PackingScheme || (PackingScheme = {}));
    var TextureUsage;
    (function (TextureUsage) {
        TextureUsage[TextureUsage["RENDER"] = 0] = "RENDER";
        TextureUsage[TextureUsage["UPLOAD"] = 1] = "UPLOAD";
        TextureUsage[TextureUsage["PIXELS"] = 2] = "PIXELS";
        TextureUsage[TextureUsage["DOWNLOAD"] = 3] = "DOWNLOAD";
    })(TextureUsage || (TextureUsage = {}));
    var PhysicalTextureType;
    (function (PhysicalTextureType) {
        PhysicalTextureType[PhysicalTextureType["UNPACKED_FLOAT16"] = 0] = "UNPACKED_FLOAT16";
        PhysicalTextureType[PhysicalTextureType["UNPACKED_FLOAT32"] = 1] = "UNPACKED_FLOAT32";
        PhysicalTextureType[PhysicalTextureType["PACKED_4X1_UNSIGNED_BYTE"] = 2] = "PACKED_4X1_UNSIGNED_BYTE";
        PhysicalTextureType[PhysicalTextureType["PACKED_2X2_FLOAT32"] = 3] = "PACKED_2X2_FLOAT32";
        PhysicalTextureType[PhysicalTextureType["PACKED_2X2_FLOAT16"] = 4] = "PACKED_2X2_FLOAT16";
    })(PhysicalTextureType || (PhysicalTextureType = {}));
    function getUnpackedMatrixTextureShapeWidthHeight(rows, columns) {
        return [columns, rows];
    }
    function getUnpackedArraySizeFromMatrixSize(matrixSize, channelsPerTexture) {
        return matrixSize * channelsPerTexture;
    }
    /**
     * Get shape for densely packed RGBA texture.
     */
    function getDenseTexShape(shape) {
        const size = sizeFromShape(shape);
        const texelsNeeded = Math.ceil(size / 4);
        return sizeToSquarishShape(texelsNeeded);
    }
    function getPackedMatrixTextureShapeWidthHeight(rows, columns) {
        return [
            Math.max(1, Math.ceil(columns / 2)), Math.max(1, Math.ceil(rows / 2))
        ];
    }
    function getPackedRGBAArraySizeFromMatrixShape(rows, columns) {
        const [w, h] = getPackedMatrixTextureShapeWidthHeight(rows, columns);
        return w * h * 4;
    }
    function getTextureConfig(
    // tslint:disable-next-line:no-any
    gl, textureHalfFloatExtension) {
        // tslint:disable-next-line:no-any
        const glany = gl;
        let internalFormatFloat;
        let internalFormatHalfFloat;
        let internalFormatPackedHalfFloat;
        let internalFormatPackedFloat;
        let textureFormatFloat;
        let downloadTextureFormat;
        let downloadUnpackNumChannels;
        let defaultNumChannels;
        let textureTypeHalfFloat;
        let textureTypeFloat;
        if (env$1().getNumber('WEBGL_VERSION') === 2) {
            internalFormatFloat = glany.R32F;
            internalFormatHalfFloat = glany.R16F;
            internalFormatPackedHalfFloat = glany.RGBA16F;
            internalFormatPackedFloat = glany.RGBA32F;
            textureFormatFloat = glany.RED;
            downloadUnpackNumChannels = 4;
            defaultNumChannels = 1;
            textureTypeHalfFloat = glany.HALF_FLOAT;
            textureTypeFloat = glany.FLOAT;
            downloadTextureFormat = glany.RGBA8;
        }
        else {
            internalFormatFloat = gl.RGBA;
            internalFormatHalfFloat = gl.RGBA;
            internalFormatPackedHalfFloat = gl.RGBA;
            internalFormatPackedFloat = glany.RGBA;
            textureFormatFloat = gl.RGBA;
            downloadUnpackNumChannels = 4;
            defaultNumChannels = 4;
            textureTypeHalfFloat = textureHalfFloatExtension != null ?
                textureHalfFloatExtension.HALF_FLOAT_OES :
                null;
            textureTypeFloat = gl.FLOAT;
            downloadTextureFormat = gl.RGBA;
        }
        return {
            internalFormatFloat,
            internalFormatHalfFloat,
            internalFormatPackedHalfFloat,
            internalFormatPackedFloat,
            textureFormatFloat,
            downloadTextureFormat,
            downloadUnpackNumChannels,
            defaultNumChannels,
            textureTypeHalfFloat,
            textureTypeFloat
        };
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function callAndCheck(gl, func) {
        const returnValue = func();
        if (env$1().getBool('DEBUG')) {
            checkWebGLError(gl);
        }
        return returnValue;
    }
    function checkWebGLError(gl) {
        const error = gl.getError();
        if (error !== gl.NO_ERROR) {
            throw new Error('WebGL Error: ' + getWebGLErrorMessage(gl, error));
        }
    }
    // https://en.wikipedia.org/wiki/Half-precision_floating-point_format
    const MIN_FLOAT16 = 5.96e-8;
    const MAX_FLOAT16 = 65504;
    function canBeRepresented(num) {
        if (env$1().getBool('WEBGL_RENDER_FLOAT32_ENABLED') || num === 0 ||
            (MIN_FLOAT16 < Math.abs(num) && Math.abs(num) < MAX_FLOAT16)) {
            return true;
        }
        return false;
    }
    function getWebGLErrorMessage(gl, status) {
        switch (status) {
            case gl.NO_ERROR:
                return 'NO_ERROR';
            case gl.INVALID_ENUM:
                return 'INVALID_ENUM';
            case gl.INVALID_VALUE:
                return 'INVALID_VALUE';
            case gl.INVALID_OPERATION:
                return 'INVALID_OPERATION';
            case gl.INVALID_FRAMEBUFFER_OPERATION:
                return 'INVALID_FRAMEBUFFER_OPERATION';
            case gl.OUT_OF_MEMORY:
                return 'OUT_OF_MEMORY';
            case gl.CONTEXT_LOST_WEBGL:
                return 'CONTEXT_LOST_WEBGL';
            default:
                return `Unknown error code ${status}`;
        }
    }
    function getExtensionOrThrow(gl, extensionName) {
        return throwIfNull(gl, () => gl.getExtension(extensionName), 'Extension "' + extensionName + '" not supported on this browser.');
    }
    function createVertexShader$1(gl, vertexShaderSource) {
        const vertexShader = throwIfNull(gl, () => gl.createShader(gl.VERTEX_SHADER), 'Unable to create vertex WebGLShader.');
        callAndCheck(gl, () => gl.shaderSource(vertexShader, vertexShaderSource));
        callAndCheck(gl, () => gl.compileShader(vertexShader));
        if (gl.getShaderParameter(vertexShader, gl.COMPILE_STATUS) === false) {
            console.log(gl.getShaderInfoLog(vertexShader));
            throw new Error('Failed to compile vertex shader.');
        }
        return vertexShader;
    }
    function createFragmentShader(gl, fragmentShaderSource) {
        const fragmentShader = throwIfNull(gl, () => gl.createShader(gl.FRAGMENT_SHADER), 'Unable to create fragment WebGLShader.');
        callAndCheck(gl, () => gl.shaderSource(fragmentShader, fragmentShaderSource));
        callAndCheck(gl, () => gl.compileShader(fragmentShader));
        if (env$1().get('ENGINE_COMPILE_ONLY')) {
            return fragmentShader;
        }
        if (gl.getShaderParameter(fragmentShader, gl.COMPILE_STATUS) === false) {
            logShaderSourceAndInfoLog(fragmentShaderSource, gl.getShaderInfoLog(fragmentShader));
            throw new Error('Failed to compile fragment shader.');
        }
        return fragmentShader;
    }
    const lineNumberRegex = /ERROR: [0-9]+:([0-9]+):/g;
    function logShaderSourceAndInfoLog(shaderSource, shaderInfoLog) {
        const lineNumberRegexResult = lineNumberRegex.exec(shaderInfoLog);
        if (lineNumberRegexResult == null) {
            console.log(`Couldn't parse line number in error: ${shaderInfoLog}`);
            console.log(shaderSource);
            return;
        }
        const lineNumber = +lineNumberRegexResult[1];
        const shaderLines = shaderSource.split('\n');
        const pad = shaderLines.length.toString().length + 2;
        const linesWithLineNumbers = shaderLines.map((line, lineNumber) => rightPad((lineNumber + 1).toString(), pad) + line);
        let maxLineLength = 0;
        for (let i = 0; i < linesWithLineNumbers.length; i++) {
            maxLineLength = Math.max(linesWithLineNumbers[i].length, maxLineLength);
        }
        const beforeErrorLines = linesWithLineNumbers.slice(0, lineNumber - 1);
        const errorLine = linesWithLineNumbers.slice(lineNumber - 1, lineNumber);
        const afterErrorLines = linesWithLineNumbers.slice(lineNumber);
        console.log(beforeErrorLines.join('\n'));
        console.log(shaderInfoLog.split('\n')[0]);
        console.log(`%c ${rightPad(errorLine[0], maxLineLength)}`, 'border:1px solid red; background-color:#e3d2d2; color:#a61717');
        console.log(afterErrorLines.join('\n'));
    }
    function createProgram(gl) {
        return throwIfNull(gl, () => gl.createProgram(), 'Unable to create WebGLProgram.');
    }
    function linkProgram(gl, program) {
        callAndCheck(gl, () => gl.linkProgram(program));
        if (env$1().get('ENGINE_COMPILE_ONLY')) {
            return;
        }
        if (gl.getProgramParameter(program, gl.LINK_STATUS) === false) {
            console.log(gl.getProgramInfoLog(program));
            throw new Error('Failed to link vertex and fragment shaders.');
        }
    }
    function validateProgram(gl, program) {
        callAndCheck(gl, () => gl.validateProgram(program));
        if (gl.getProgramParameter(program, gl.VALIDATE_STATUS) === false) {
            console.log(gl.getProgramInfoLog(program));
            throw new Error('Shader program validation failed.');
        }
    }
    function createStaticVertexBuffer(gl, data) {
        const buffer = throwIfNull(gl, () => gl.createBuffer(), 'Unable to create WebGLBuffer');
        callAndCheck(gl, () => gl.bindBuffer(gl.ARRAY_BUFFER, buffer));
        callAndCheck(gl, () => gl.bufferData(gl.ARRAY_BUFFER, data, gl.STATIC_DRAW));
        return buffer;
    }
    function createStaticIndexBuffer(gl, data) {
        const buffer = throwIfNull(gl, () => gl.createBuffer(), 'Unable to create WebGLBuffer');
        callAndCheck(gl, () => gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, buffer));
        callAndCheck(gl, () => gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, data, gl.STATIC_DRAW));
        return buffer;
    }
    function createTexture(gl) {
        return throwIfNull(gl, () => gl.createTexture(), 'Unable to create WebGLTexture.');
    }
    function validateTextureSize(width, height) {
        const maxTextureSize = env$1().getNumber('WEBGL_MAX_TEXTURE_SIZE');
        if ((width <= 0) || (height <= 0)) {
            const requested = `[${width}x${height}]`;
            throw new Error('Requested texture size ' + requested + ' is invalid.');
        }
        if ((width > maxTextureSize) || (height > maxTextureSize)) {
            const requested = `[${width}x${height}]`;
            const max = `[${maxTextureSize}x${maxTextureSize}]`;
            throw new Error('Requested texture size ' + requested +
                ' greater than WebGL maximum on this browser / GPU ' + max + '.');
        }
    }
    function createFramebuffer(gl) {
        return throwIfNull(gl, () => gl.createFramebuffer(), 'Unable to create WebGLFramebuffer.');
    }
    function bindVertexBufferToProgramAttribute(gl, program, attribute, buffer, arrayEntriesPerItem, itemStrideInBytes, itemOffsetInBytes) {
        const loc = gl.getAttribLocation(program, attribute);
        if (loc === -1) {
            // The GPU compiler decided to strip out this attribute because it's unused,
            // thus no need to bind.
            return false;
        }
        callAndCheck(gl, () => gl.bindBuffer(gl.ARRAY_BUFFER, buffer));
        callAndCheck(gl, () => gl.vertexAttribPointer(loc, arrayEntriesPerItem, gl.FLOAT, false, itemStrideInBytes, itemOffsetInBytes));
        callAndCheck(gl, () => gl.enableVertexAttribArray(loc));
        return true;
    }
    function bindTextureUnit(gl, texture, textureUnit) {
        validateTextureUnit(gl, textureUnit);
        callAndCheck(gl, () => gl.activeTexture(gl.TEXTURE0 + textureUnit));
        callAndCheck(gl, () => gl.bindTexture(gl.TEXTURE_2D, texture));
    }
    function getProgramUniformLocationOrThrow(gl, program, uniformName) {
        return throwIfNull(gl, () => gl.getUniformLocation(program, uniformName), 'uniform "' + uniformName + '" not present in program.');
    }
    function getProgramUniformLocation(gl, program, uniformName) {
        return gl.getUniformLocation(program, uniformName);
    }
    function bindTextureToProgramUniformSampler(gl, texture, uniformSamplerLocation, textureUnit) {
        callAndCheck(gl, () => bindTextureUnit(gl, texture, textureUnit));
        callAndCheck(gl, () => gl.uniform1i(uniformSamplerLocation, textureUnit));
    }
    function bindColorTextureToFramebuffer(gl, texture, framebuffer) {
        callAndCheck(gl, () => gl.bindFramebuffer(gl.FRAMEBUFFER, framebuffer));
        callAndCheck(gl, () => gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, texture, 0));
    }
    function unbindColorTextureFromFramebuffer(gl, framebuffer) {
        callAndCheck(gl, () => gl.bindFramebuffer(gl.FRAMEBUFFER, framebuffer));
        callAndCheck(gl, () => gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, null, 0));
    }
    function validateFramebuffer(gl) {
        const status = gl.checkFramebufferStatus(gl.FRAMEBUFFER);
        if (status !== gl.FRAMEBUFFER_COMPLETE) {
            throw new Error('Error binding framebuffer: ' + getFramebufferErrorMessage(gl, status));
        }
    }
    function getFramebufferErrorMessage(gl, status) {
        switch (status) {
            case gl.FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
                return 'FRAMEBUFFER_INCOMPLETE_ATTACHMENT';
            case gl.FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
                return 'FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT';
            case gl.FRAMEBUFFER_INCOMPLETE_DIMENSIONS:
                return 'FRAMEBUFFER_INCOMPLETE_DIMENSIONS';
            case gl.FRAMEBUFFER_UNSUPPORTED:
                return 'FRAMEBUFFER_UNSUPPORTED';
            default:
                return `unknown error ${status}`;
        }
    }
    function throwIfNull(gl, returnTOrNull, failureMessage) {
        const tOrNull = callAndCheck(gl, () => returnTOrNull());
        if (tOrNull == null) {
            throw new Error(failureMessage);
        }
        return tOrNull;
    }
    function validateTextureUnit(gl, textureUnit) {
        const maxTextureUnit = gl.MAX_COMBINED_TEXTURE_IMAGE_UNITS - 1;
        const glTextureUnit = textureUnit + gl.TEXTURE0;
        if (glTextureUnit < gl.TEXTURE0 || glTextureUnit > maxTextureUnit) {
            const textureUnitRange = `[gl.TEXTURE0, gl.TEXTURE${maxTextureUnit}]`;
            throw new Error(`textureUnit must be in ${textureUnitRange}.`);
        }
    }
    function getBatchDim(shape, dimsToSkip = 2) {
        return sizeFromShape(shape.slice(0, shape.length - dimsToSkip));
    }
    function getRowsCols(shape) {
        if (shape.length === 0) {
            throw Error('Cannot get rows and columns of an empty shape array.');
        }
        return [
            shape.length > 1 ? shape[shape.length - 2] : 1, shape[shape.length - 1]
        ];
    }
    function getShapeAs3D(shape) {
        let shapeAs3D = [1, 1, 1];
        const isScalar = shape.length === 0 || (shape.length === 1 && shape[0] === 1);
        if (!isScalar) {
            shapeAs3D =
                [getBatchDim(shape), ...getRowsCols(shape)];
        }
        return shapeAs3D;
    }
    function getTextureShapeFromLogicalShape(logShape, isPacked = false) {
        let maxTexSize = env$1().getNumber('WEBGL_MAX_TEXTURE_SIZE');
        if (isPacked) {
            maxTexSize = maxTexSize * 2;
            // This logic ensures we accurately count the number of packed texels needed
            // to accommodate the tensor. We can only pack values in the same texel if
            // they are from adjacent pairs of rows/cols within the same batch. So if a
            // tensor has 3 rows, we pretend it has 4 rows in order to account for the
            // fact that the texels containing the third row are half empty.
            logShape = logShape.map((d, i) => i >= logShape.length - 2 ?
                nearestLargerEven(logShape[i]) :
                logShape[i]);
            // Packed texture height is at least 2 (the channel height of a single
            // texel).
            if (logShape.length === 1) {
                logShape = [2, logShape[0]];
            }
        }
        // If logical shape is 2, we don't squeeze, since we want to match physical.
        if (logShape.length !== 2) {
            const squeezeResult = squeezeShape(logShape);
            logShape = squeezeResult.newShape;
        }
        let size = sizeFromShape(logShape);
        if (logShape.length <= 1 && size <= maxTexSize) {
            return [1, size];
        }
        else if (logShape.length === 2 && logShape[0] <= maxTexSize &&
            logShape[1] <= maxTexSize) {
            return logShape;
        }
        else if (logShape.length === 3 && logShape[0] * logShape[1] <= maxTexSize &&
            logShape[2] <= maxTexSize) {
            return [logShape[0] * logShape[1], logShape[2]];
        }
        else if (logShape.length === 3 && logShape[0] <= maxTexSize &&
            logShape[1] * logShape[2] <= maxTexSize) {
            return [logShape[0], logShape[1] * logShape[2]];
        }
        else if (logShape.length === 4 &&
            logShape[0] * logShape[1] * logShape[2] <= maxTexSize &&
            logShape[3] <= maxTexSize) {
            return [logShape[0] * logShape[1] * logShape[2], logShape[3]];
        }
        else if (logShape.length === 4 && logShape[0] <= maxTexSize &&
            logShape[1] * logShape[2] * logShape[3] <= maxTexSize) {
            return [logShape[0], logShape[1] * logShape[2] * logShape[3]];
        }
        else {
            if (isPacked) {
                // For packed textures size equals the number of channels required to
                // accommodate the texture data. However in order to squarify such that
                // inner dimensions stay even, we rewrite size to equal the number of
                // texels. Then in the return statement we rehydrate the squarified
                // dimensions to channel units.
                const batchDim = getBatchDim(logShape);
                let rows = 2, cols = 2;
                if (logShape.length) {
                    [rows, cols] = getRowsCols(logShape);
                }
                size = batchDim * (rows / 2) * (cols / 2);
                return sizeToSquarishShape(size).map(d => d * 2);
            }
            return sizeToSquarishShape(size);
        }
    }
    function isEven(n) {
        return n % 2 === 0;
    }
    /**
     * This determines whether reshaping a packed texture requires rearranging
     * the data within the texture, assuming 2x2 packing.
     */
    function isReshapeFree(shape1, shape2) {
        shape1 = shape1.slice(-2);
        shape2 = shape2.slice(-2);
        if (arraysEqual(shape1, shape2)) {
            return true;
        }
        if (!shape1.length || !shape2.length) { // One of the shapes is a scalar.
            return true;
        }
        if (shape1[0] === 0 || shape1[1] === 0 || shape2[0] === 0 ||
            shape2[1] === 0) {
            return true;
        }
        if (shape1.length !== shape2.length) { // One of the shapes is a vector.
            const shape1Cols = shape1.slice(-1)[0];
            const shape2Cols = shape2.slice(-1)[0];
            if (shape1Cols === shape2Cols) {
                return true;
            }
            if (isEven(shape1Cols) && isEven(shape2Cols) &&
                (shape1[0] === 1 || shape2[0] === 1)) {
                return true;
            }
        }
        return shape1[1] === shape2[1] && isEven(shape1[0]) && isEven(shape2[0]);
    }
    // We cache webgl params because the environment gets reset between
    // unit tests and we don't want to constantly query the WebGLContext for
    // MAX_TEXTURE_SIZE.
    let MAX_TEXTURE_SIZE;
    let MAX_TEXTURES_IN_SHADER;
    function getWebGLMaxTextureSize(webGLVersion) {
        if (MAX_TEXTURE_SIZE == null) {
            const gl = getWebGLContext(webGLVersion);
            MAX_TEXTURE_SIZE = gl.getParameter(gl.MAX_TEXTURE_SIZE);
        }
        return MAX_TEXTURE_SIZE;
    }
    function getMaxTexturesInShader(webGLVersion) {
        if (MAX_TEXTURES_IN_SHADER == null) {
            const gl = getWebGLContext(webGLVersion);
            MAX_TEXTURES_IN_SHADER = gl.getParameter(gl.MAX_TEXTURE_IMAGE_UNITS);
        }
        // We cap at 16 to avoid spurious runtime "memory exhausted" error.
        return Math.min(16, MAX_TEXTURES_IN_SHADER);
    }
    function getWebGLDisjointQueryTimerVersion(webGLVersion) {
        if (webGLVersion === 0) {
            return 0;
        }
        let queryTimerVersion;
        const gl = getWebGLContext(webGLVersion);
        if (hasExtension(gl, 'EXT_disjoint_timer_query_webgl2') &&
            webGLVersion === 2) {
            queryTimerVersion = 2;
        }
        else if (hasExtension(gl, 'EXT_disjoint_timer_query')) {
            queryTimerVersion = 1;
        }
        else {
            queryTimerVersion = 0;
        }
        return queryTimerVersion;
    }
    function hasExtension(gl, extensionName) {
        const ext = gl.getExtension(extensionName);
        return ext != null;
    }
    function isWebGLVersionEnabled(webGLVersion) {
        try {
            const gl = getWebGLContext(webGLVersion);
            if (gl != null) {
                return true;
            }
        }
        catch (e) {
            console.log('Error when getting WebGL context: ', e);
            return false;
        }
        return false;
    }
    function isCapableOfRenderingToFloatTexture(webGLVersion) {
        if (webGLVersion === 0) {
            return false;
        }
        const gl = getWebGLContext(webGLVersion);
        if (webGLVersion === 1) {
            if (!hasExtension(gl, 'OES_texture_float')) {
                return false;
            }
        }
        else {
            if (!hasExtension(gl, 'EXT_color_buffer_float')) {
                return false;
            }
        }
        const isFrameBufferComplete = createFloatTextureAndBindToFramebuffer(gl);
        return isFrameBufferComplete;
    }
    /**
     * Check if we can download values from a float/half-float texture.
     *
     * Note that for performance reasons we use binding a texture to a framebuffer
     * as a proxy for ability to download float values later using readPixels. The
     * texture params of this texture will not match those in readPixels exactly
     * but if we are unable to bind some kind of float texture to the frameBuffer
     * then we definitely will not be able to read float values from it.
     */
    function isDownloadFloatTextureEnabled(webGLVersion) {
        if (webGLVersion === 0) {
            return false;
        }
        const gl = getWebGLContext(webGLVersion);
        if (webGLVersion === 1) {
            if (!hasExtension(gl, 'OES_texture_float')) {
                return false;
            }
            if (!hasExtension(gl, 'WEBGL_color_buffer_float')) {
                return false;
            }
        }
        else {
            if (hasExtension(gl, 'EXT_color_buffer_float')) {
                return createFloatTextureAndBindToFramebuffer(gl);
            }
            const COLOR_BUFFER_HALF_FLOAT = 'EXT_color_buffer_half_float';
            if (hasExtension(gl, COLOR_BUFFER_HALF_FLOAT)) {
                const textureHalfFloatExtension = gl.getExtension(COLOR_BUFFER_HALF_FLOAT);
                return createHalfFloatTextureAndBindToFramebuffer(gl, textureHalfFloatExtension);
            }
            return false;
        }
        const isFrameBufferComplete = createFloatTextureAndBindToFramebuffer(gl);
        return isFrameBufferComplete;
    }
    function createFloatTextureAndBindToFramebuffer(gl) {
        const texConfig = getTextureConfig(gl);
        const texture = gl.createTexture();
        gl.bindTexture(gl.TEXTURE_2D, texture);
        const width = 1;
        const height = 1;
        gl.texImage2D(gl.TEXTURE_2D, 0, texConfig.internalFormatFloat, width, height, 0, texConfig.textureFormatFloat, texConfig.textureTypeFloat, null);
        const frameBuffer = gl.createFramebuffer();
        gl.bindFramebuffer(gl.FRAMEBUFFER, frameBuffer);
        gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, texture, 0);
        const isFrameBufferComplete = gl.checkFramebufferStatus(gl.FRAMEBUFFER) === gl.FRAMEBUFFER_COMPLETE;
        gl.bindTexture(gl.TEXTURE_2D, null);
        gl.bindFramebuffer(gl.FRAMEBUFFER, null);
        gl.deleteTexture(texture);
        gl.deleteFramebuffer(frameBuffer);
        return isFrameBufferComplete;
    }
    function createHalfFloatTextureAndBindToFramebuffer(
    // tslint:disable-next-line:no-any
    gl, textureHalfFloatExtension) {
        const texConfig = getTextureConfig(gl, textureHalfFloatExtension);
        const texture = gl.createTexture();
        gl.bindTexture(gl.TEXTURE_2D, texture);
        const width = 1;
        const height = 1;
        gl.texImage2D(gl.TEXTURE_2D, 0, texConfig.internalFormatHalfFloat, width, height, 0, texConfig.textureFormatFloat, texConfig.textureTypeHalfFloat, null);
        const frameBuffer = gl.createFramebuffer();
        gl.bindFramebuffer(gl.FRAMEBUFFER, frameBuffer);
        gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, texture, 0);
        const isFrameBufferComplete = gl.checkFramebufferStatus(gl.FRAMEBUFFER) === gl.FRAMEBUFFER_COMPLETE;
        gl.bindTexture(gl.TEXTURE_2D, null);
        gl.bindFramebuffer(gl.FRAMEBUFFER, null);
        gl.deleteTexture(texture);
        gl.deleteFramebuffer(frameBuffer);
        return isFrameBufferComplete;
    }
    function isWebGLFenceEnabled(webGLVersion) {
        if (webGLVersion !== 2) {
            return false;
        }
        const gl = getWebGLContext(webGLVersion);
        // tslint:disable-next-line:no-any
        const isEnabled = gl.fenceSync != null;
        return isEnabled;
    }
    function assertNotComplex(tensor, opName) {
        if (!Array.isArray(tensor)) {
            tensor = [tensor];
        }
        tensor.forEach(t => {
            if (t != null) {
                assert$4(t.dtype !== 'complex64', () => `${opName} does not support complex64 tensors ` +
                    'in the WebGL backend.');
            }
        });
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ENV = env$1();
    /**
     * This file contains WebGL-specific flag registrations.
     */
    /**
     * True if WebGL is supported.
     */
    ENV.registerFlag('HAS_WEBGL', () => ENV.getNumber('WEBGL_VERSION') > 0);
    /** 0: No WebGL, 1: WebGL 1.0, 2: WebGL 2.0. */
    ENV.registerFlag('WEBGL_VERSION', () => {
        if (isWebGLVersionEnabled(2)) {
            return 2;
        }
        else if (isWebGLVersionEnabled(1)) {
            return 1;
        }
        return 0;
    });
    /** Whether to check for numerical representation problems. */
    ENV.registerFlag('WEBGL_CHECK_NUMERICAL_PROBLEMS', () => false);
    ENV.registerFlag('WEBGL_BUFFER_SUPPORTED', () => ENV.get('WEBGL_VERSION') === 2);
    /** Whether the WebGL backend will sometimes forward ops to the CPU. */
    ENV.registerFlag('WEBGL_CPU_FORWARD', () => true);
    /** Whether the WebGL backend will always use f16 textures for rendering. */
    ENV.registerFlag('WEBGL_FORCE_F16_TEXTURES', () => false);
    /** Whether to turn all packing related flags on. */
    ENV.registerFlag('WEBGL_PACK', () => ENV.getBool('HAS_WEBGL'));
    /** Whether we will pack the batchnormalization op. */
    ENV.registerFlag('WEBGL_PACK_NORMALIZATION', () => ENV.getBool('WEBGL_PACK'));
    /** Whether we will pack the clip op. */
    ENV.registerFlag('WEBGL_PACK_CLIP', () => ENV.getBool('WEBGL_PACK'));
    /** Whether we will pack the depthwise conv op. */
    ENV.registerFlag('WEBGL_PACK_DEPTHWISECONV', () => ENV.getBool('WEBGL_PACK'));
    /** Whether we will pack binary ops. */
    ENV.registerFlag('WEBGL_PACK_BINARY_OPERATIONS', () => ENV.getBool('WEBGL_PACK'));
    /** Whether we will pack unary ops. */
    ENV.registerFlag('WEBGL_PACK_UNARY_OPERATIONS', () => ENV.getBool('WEBGL_PACK'));
    /** Whether we will pack array ops. */
    ENV.registerFlag('WEBGL_PACK_ARRAY_OPERATIONS', () => ENV.getBool('WEBGL_PACK'));
    /** Whether we will pack image ops. */
    ENV.registerFlag('WEBGL_PACK_IMAGE_OPERATIONS', () => ENV.getBool('WEBGL_PACK'));
    /** Whether we will pack reduce ops. */
    ENV.registerFlag('WEBGL_PACK_REDUCE', () => ENV.getBool('WEBGL_PACK'));
    /** Whether packed WebGL kernels lazily unpack their outputs. */
    ENV.registerFlag('WEBGL_LAZILY_UNPACK', () => ENV.getBool('WEBGL_PACK'));
    /** Whether we will use the im2col algorithm to speed up convolutions. */
    ENV.registerFlag('WEBGL_CONV_IM2COL', () => ENV.getBool('WEBGL_PACK'));
    /** The maximum texture dimension. */
    ENV.registerFlag('WEBGL_MAX_TEXTURE_SIZE', () => getWebGLMaxTextureSize(ENV.getNumber('WEBGL_VERSION')));
    /** The maximum texture dimension. */
    ENV.registerFlag('WEBGL_MAX_TEXTURES_IN_SHADER', () => getMaxTexturesInShader(ENV.getNumber('WEBGL_VERSION')));
    /**
     * The disjoint_query_timer extension version.
     * 0: disabled, 1: EXT_disjoint_timer_query, 2:
     * EXT_disjoint_timer_query_webgl2.
     * In Firefox with WebGL 2.0,
     * EXT_disjoint_timer_query_webgl2 is not available, so we must use the
     * WebGL 1.0 extension.
     */
    ENV.registerFlag('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_VERSION', () => {
        const webGLVersion = ENV.getNumber('WEBGL_VERSION');
        if (webGLVersion === 0) {
            return 0;
        }
        return getWebGLDisjointQueryTimerVersion(webGLVersion);
    });
    /**
     * Whether the timer object from the disjoint_query_timer extension gives
     * timing information that is reliable.
     */
    ENV.registerFlag('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_RELIABLE', () => ENV.getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_VERSION') > 0 &&
        !isMobile());
    /**
     * Whether the device is physically capable of rendering to float32 textures.
     */
    ENV.registerFlag('WEBGL_RENDER_FLOAT32_CAPABLE', () => isCapableOfRenderingToFloatTexture(ENV.getNumber('WEBGL_VERSION')));
    /**
     * Whether rendering to float32 textures is enabled. If disabled, renders to
     * float16 textures.
     */
    ENV.registerFlag('WEBGL_RENDER_FLOAT32_ENABLED', () => {
        return ENV.getBool('WEBGL_FORCE_F16_TEXTURES') ?
            false :
            ENV.getBool('WEBGL_RENDER_FLOAT32_CAPABLE');
    });
    /**
     * Whether downloading float textures is enabled (16 or 32 bit). If disabled,
     * uses IEEE 754 encoding of the float32 values to 4 uint8 when downloading.
     */
    ENV.registerFlag('WEBGL_DOWNLOAD_FLOAT_ENABLED', () => isDownloadFloatTextureEnabled(ENV.getNumber('WEBGL_VERSION')));
    /** Whether the fence API is available. */
    ENV.registerFlag('WEBGL_FENCE_API_ENABLED', () => isWebGLFenceEnabled(ENV.getNumber('WEBGL_VERSION')));
    /**
     * Tensors with size <= than this will be uploaded as uniforms, not textures.
     */
    ENV.registerFlag('WEBGL_SIZE_UPLOAD_UNIFORM', () => {
        // Use uniform uploads only when 32bit floats are supported. In
        // 16bit
        // environments there are problems with comparing a 16bit texture value
        // with a 32bit uniform value.
        const useUniforms = ENV.getBool('WEBGL_RENDER_FLOAT32_ENABLED');
        return useUniforms ? 4 : 0;
    });
    /**
     * If the total number of bytes allocated on the GPU is greater than this
     * number, we will aggressively delete textures upon disposal with
     * gl.deleteMatrixTexture, rather than making them available for reuse.
     *
     * Default value -1 indicates that we will never aggressively delete textures.
     */
    ENV.registerFlag('WEBGL_DELETE_TEXTURE_THRESHOLD', () => {
        return -1;
    }, threshold => {
        if (threshold < 0 && threshold !== -1) {
            throw new Error(`WEBGL_DELETE_TEXTURE_THRESHOLD must be -1 (indicating never ` +
                `delete) or at least 0, but got ${threshold}.`);
        }
    });
    /**
     * Trigger a manual GL command flush if the threshold of time has passed since
     * previous Kernel execution. This can be useful for Andorid device where GL
     * command flush are delayed un til the end of javascript task. This value is
     * measured in millisecond. Typically you want to set this value to close to 1.
     *
     * Default value 1 for mobile chrome, and -1 for rest cases. -1 indicates that
     * we will not enforce manual flush and depend on system default flush schedule.
     */
    ENV.registerFlag('WEBGL_FLUSH_THRESHOLD', () => {
        return isMobile() ? 1 : -1;
    }, threshold => {
        if (threshold < 0 && threshold !== -1) {
            throw new Error(`WEBGL_FLUSH_THRESHOLD must be -1 (indicating never ` +
                `manual flush) or at least 0, but got ${threshold}.`);
        }
    });
    /**
     * Threshold for input tensor size that determines whether WebGL backend will
     * delegate computation to CPU.
     *
     * Default value is 128.
     */
    ENV.registerFlag('CPU_HANDOFF_SIZE_THRESHOLD', () => 128);
    /** Whether we will use shapes uniforms. */
    ENV.registerFlag('WEBGL_USE_SHAPES_UNIFORMS', () => false);
    /**
     * Threshold for last dimension of input tensor that determines whether
     * WebGL backend for the Top K op will delegate computation to CPU. If input
     * is smaller than threshold then CPU will be used
     *
     * Default value is 100000.
     */
    ENV.registerFlag('TOPK_LAST_DIM_CPU_HANDOFF_SIZE_THRESHOLD', () => 100000);
    /**
     * Threshold for K that determines whether
     * WebGL backend for the Top K op will delegate computation to CPU. If k
     * is larger than threshold then CPU will be used
     *
     * Default value is 128.
     */
    ENV.registerFlag('TOPK_K_CPU_HANDOFF_THRESHOLD', () => 128);

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function getGlslDifferences() {
        let version;
        let attribute;
        let varyingVs;
        let varyingFs;
        let texture2D;
        let output;
        let defineOutput;
        let defineSpecialNaN;
        let defineSpecialInf;
        let defineRound;
        if (env$1().getNumber('WEBGL_VERSION') === 2) {
            version = '#version 300 es';
            attribute = 'in';
            varyingVs = 'out';
            varyingFs = 'in';
            texture2D = 'texture';
            output = 'outputColor';
            defineOutput = 'out vec4 outputColor;';
            // Use custom isnan definition to work across differences between
            // implementations on various platforms. While this should happen in ANGLE
            // we still see differences between android and windows (on chrome) when
            // using isnan directly. Since WebGL2 supports uint type and
            // floatBitsToUinT built-in function, we could implment isnan following
            // IEEE 754 rules.
            // NaN defination in IEEE 754-1985 is :
            //   - sign = either 0 or 1.
            //   - biased exponent = all 1 bits.
            //   - fraction = anything except all 0 bits (since all 0 bits represents
            //   infinity).
            // https://en.wikipedia.org/wiki/IEEE_754-1985#Representation_of_non-numbers
            defineSpecialNaN = `
      bool isnan_custom(float val) {
        uint floatToUint = floatBitsToUint(val);
        return (floatToUint & 0x7fffffffu) > 0x7f800000u;
      }

      bvec4 isnan_custom(vec4 val) {
        return bvec4(isnan_custom(val.x),
          isnan_custom(val.y), isnan_custom(val.z), isnan_custom(val.w));
      }

      #define isnan(value) isnan_custom(value)
    `;
            // In webgl 2 we do not need to specify a custom isinf so there is no
            // need for a special INFINITY constant.
            defineSpecialInf = ``;
            defineRound = `
      #define round(value) newRound(value)
      int newRound(float value) {
        return int(floor(value + 0.5));
      }

      ivec4 newRound(vec4 value) {
        return ivec4(floor(value + vec4(0.5)));
      }
    `;
        }
        else {
            version = '';
            attribute = 'attribute';
            varyingVs = 'varying';
            varyingFs = 'varying';
            texture2D = 'texture2D';
            output = 'gl_FragColor';
            defineOutput = '';
            // WebGL1 has no built in isnan so we define one here.
            defineSpecialNaN = `
      #define isnan(value) isnan_custom(value)
      bool isnan_custom(float val) {
        return (val > 0. || val < 1. || val == 0.) ? false : true;
      }
      bvec4 isnan_custom(vec4 val) {
        return bvec4(isnan(val.x), isnan(val.y), isnan(val.z), isnan(val.w));
      }
    `;
            defineSpecialInf = `
      uniform float INFINITY;

      bool isinf(float val) {
        return abs(val) == INFINITY;
      }
      bvec4 isinf(vec4 val) {
        return equal(abs(val), vec4(INFINITY));
      }
    `;
            defineRound = `
      int round(float value) {
        return int(floor(value + 0.5));
      }

      ivec4 round(vec4 value) {
        return ivec4(floor(value + vec4(0.5)));
      }
    `;
        }
        return {
            version,
            attribute,
            varyingVs,
            varyingFs,
            texture2D,
            output,
            defineOutput,
            defineSpecialNaN,
            defineSpecialInf,
            defineRound
        };
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Produces GLSL code that derives logical coordinates from a flat
     * index. The code performs integer division with each stride and decrements
     * the index until the index equals the final dimension coordinate.
     */
    function getLogicalCoordinatesFromFlatIndex(coords, shape, index = 'index') {
        const strides = computeStrides(shape);
        return strides
            .map((stride, i) => {
            const line1 = `int ${coords[i]} = ${index} / ${stride}`;
            const line2 = i === strides.length - 1 ?
                `int ${coords[i + 1]} = ${index} - ${coords[i]} * ${stride}` :
                `index -= ${coords[i]} * ${stride}`;
            return `${line1}; ${line2};`;
        })
            .join('');
    }
    function getOutputLogicalCoordinatesFromFlatIndexByUniform(coords, shape, index = 'index') {
        const strides = computeStrides(shape);
        return strides
            .map((_, i) => {
            const line1 = `int ${coords[i]} = ${index} / outShapeStrides[${i}]`;
            const line2 = i === strides.length - 1 ?
                `int ${coords[i + 1]} = ${index} - ${coords[i]} * outShapeStrides[${i}]` :
                `index -= ${coords[i]} * outShapeStrides[${i}]`;
            return `${line1}; ${line2};`;
        })
            .join('');
    }
    // Produces GLSL code that computes strides.
    function symbolicallyComputeStrides(indicesArr, variableName) {
        const numCoords = indicesArr.length;
        const shape = indicesArr.map(d => `${variableName}[${d}]`);
        const strides = new Array(numCoords - 1);
        strides[numCoords - 2] = shape[numCoords - 1];
        for (let i = numCoords - 3; i >= 0; --i) {
            strides[i] = `(${strides[i + 1]} * ${shape[i + 1]})`;
        }
        return strides;
    }
    function getLogicalCoordinatesFromFlatIndexByUniform(coords, variableName, index = 'index') {
        const indicesArray = coords.map((_, i) => i);
        const strides = symbolicallyComputeStrides(indicesArray, variableName);
        return strides
            .map((_, i) => {
            const line1 = `int ${coords[i]} = ${index} / ${strides[i]}`;
            const line2 = i === strides.length - 1 ?
                `int ${coords[i + 1]} = ${index} - ${coords[i]} * ${strides[i]}` :
                `index -= ${coords[i]} * ${strides[i]}`;
            return `${line1}; ${line2};`;
        })
            .join('');
    }
    /**
     * Produces GLSL that computes the flat index from 3D coordinates.
     */
    function getFlatIndexFrom3D(shape) {
        const strides = computeStrides(shape).map(d => d.toString());
        return `
  int getFlatIndex(ivec3 coords) {
    return coords.x * ${strides[0]} + coords.y * ${strides[1]} + coords.z;
  }
`;
    }
    function getFlatIndexFrom3DOutput() {
        return `
  int getFlatIndex(ivec3 coords) {
    return coords.x * outShapeStrides[0] + coords.y * outShapeStrides[1] + coords.z;
  }
`;
    }
    const ENCODE_FLOAT_SNIPPET = `
  const float FLOAT_MAX = 1.70141184e38;
  const float FLOAT_MIN = 1.17549435e-38;

  lowp vec4 encode_float(highp float v) {
    if (isnan(v)) {
      return vec4(255, 255, 255, 255);
    }

    highp float av = abs(v);

    if(av < FLOAT_MIN) {
      return vec4(0.0, 0.0, 0.0, 0.0);
    } else if(v > FLOAT_MAX) {
      return vec4(0.0, 0.0, 128.0, 127.0) / 255.0;
    } else if(v < -FLOAT_MAX) {
      return vec4(0.0, 0.0,  128.0, 255.0) / 255.0;
    }

    highp vec4 c = vec4(0,0,0,0);

    highp float e = floor(log2(av));
    highp float m = exp2(fract(log2(av))) - 1.0;

    c[2] = floor(128.0 * m);
    m -= c[2] / 128.0;
    c[1] = floor(32768.0 * m);
    m -= c[1] / 32768.0;
    c[0] = floor(8388608.0 * m);

    highp float ebias = e + 127.0;
    c[3] = floor(ebias / 2.0);
    ebias -= c[3] * 2.0;
    c[2] += floor(ebias) * 128.0;

    c[3] += 128.0 * step(0.0, -v);

    return c / 255.0;
  }
`;

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const { getBroadcastDims } = backend_util;
    function makeShader(inputsInfo, outputShape, program) {
        const prefixSnippets = [];
        inputsInfo.forEach(x => {
            const size = sizeFromShape(x.shapeInfo.logicalShape);
            // Snippet when we decided to upload the values as uniform.
            if (x.shapeInfo.isUniform) {
                prefixSnippets.push(`uniform float ${x.name}${size > 1 ? `[${size}]` : ''};`);
            }
            else {
                prefixSnippets.push(`uniform sampler2D ${x.name};`);
                prefixSnippets.push(`uniform int offset${x.name};`);
            }
            if (program.enableShapeUniforms) {
                const { uniformShape } = getUniformInfoFromShape(program.packedInputs, x.shapeInfo.logicalShape, x.shapeInfo.texShape);
                switch (uniformShape.length) {
                    case 1:
                        prefixSnippets.push(`uniform int ${x.name}Shape;`);
                        break;
                    case 2:
                        prefixSnippets.push(`uniform ivec2 ${x.name}Shape;`);
                        break;
                    case 3:
                        prefixSnippets.push(`uniform ivec3 ${x.name}Shape;`);
                        break;
                    case 4:
                        prefixSnippets.push(`uniform ivec4 ${x.name}Shape;`);
                        break;
                }
                prefixSnippets.push(`uniform ivec2 ${x.name}TexShape;`);
            }
        });
        if (program.enableShapeUniforms) {
            switch (outputShape.logicalShape.length) {
                case 1:
                    prefixSnippets.push(`uniform int outShape;`);
                    break;
                case 2:
                    prefixSnippets.push(`uniform ivec2 outShape;`);
                    prefixSnippets.push(`uniform int outShapeStrides;`);
                    break;
                case 3:
                    prefixSnippets.push(`uniform ivec3 outShape;`);
                    prefixSnippets.push(`uniform ivec2 outShapeStrides;`);
                    break;
                case 4:
                    prefixSnippets.push(`uniform ivec4 outShape;`);
                    prefixSnippets.push(`uniform ivec3 outShapeStrides;`);
                    break;
            }
            prefixSnippets.push(`uniform ivec2 outTexShape;`);
        }
        if (program.customUniforms) {
            program.customUniforms.forEach((d) => {
                prefixSnippets.push(`uniform ${d.type} ${d.name}${d.arrayIndex ? `[${d.arrayIndex}]` : ''};`);
            });
        }
        const inputPrefixSnippet = prefixSnippets.join('\n');
        const inputSamplingSnippet = inputsInfo
            .map(x => getInputSamplingSnippet(x, outputShape, program.packedInputs, program.enableShapeUniforms))
            .join('\n');
        const outTexShape = outputShape.texShape;
        const glsl = getGlslDifferences();
        const floatTextureSampleSnippet = getFloatTextureSampleSnippet(glsl);
        let outputSamplingSnippet;
        let floatTextureSetOutputSnippet;
        let shaderPrefix = getShaderPrefix(glsl);
        if (outputShape.isPacked) {
            outputSamplingSnippet = getPackedOutputSamplingSnippet(outputShape.logicalShape, outTexShape, program.enableShapeUniforms);
            floatTextureSetOutputSnippet = getFloatTextureSetRGBASnippet(glsl);
        }
        else {
            outputSamplingSnippet = getOutputSamplingSnippet(outputShape.logicalShape, outTexShape, program.enableShapeUniforms);
            floatTextureSetOutputSnippet = getFloatTextureSetRSnippet(glsl);
        }
        if (program.packedInputs) {
            shaderPrefix += SHADER_PACKED_PREFIX;
        }
        const source = [
            shaderPrefix, floatTextureSampleSnippet, floatTextureSetOutputSnippet,
            inputPrefixSnippet, outputSamplingSnippet, inputSamplingSnippet,
            program.userCode
        ].join('\n');
        return source;
    }
    function getSamplerFromInInfo(inInfo, enableShapeUniforms = false) {
        const shape = inInfo.shapeInfo.logicalShape;
        switch (shape.length) {
            case 0:
                return getSamplerScalar(inInfo, enableShapeUniforms);
            case 1:
                return getSampler1D(inInfo, enableShapeUniforms);
            case 2:
                return getSampler2D(inInfo, enableShapeUniforms);
            case 3:
                return getSampler3D(inInfo, enableShapeUniforms);
            case 4:
                return getSampler4D(inInfo, enableShapeUniforms);
            case 5:
                return getSampler5D(inInfo);
            case 6:
                return getSampler6D(inInfo);
            default:
                throw new Error(`${shape.length}-D input sampling` +
                    ` is not yet supported`);
        }
    }
    function getPackedSamplerFromInInfo(inInfo, enableShapeUniforms) {
        const shape = inInfo.shapeInfo.logicalShape;
        switch (shape.length) {
            case 0:
                return getPackedSamplerScalar(inInfo);
            case 1:
                return getPackedSampler1D(inInfo, enableShapeUniforms);
            case 2:
                return getPackedSampler2D(inInfo, enableShapeUniforms);
            case 3:
                return getPackedSampler3D(inInfo, enableShapeUniforms);
            default:
                return getPackedSamplerND(inInfo, enableShapeUniforms);
        }
    }
    function getInputSamplingSnippet(inInfo, outShapeInfo, usesPackedTextures = false, enableShapeUniforms) {
        let res = '';
        if (usesPackedTextures) {
            res += getPackedSamplerFromInInfo(inInfo, enableShapeUniforms);
        }
        else {
            res += getSamplerFromInInfo(inInfo, enableShapeUniforms);
        }
        const inShape = inInfo.shapeInfo.logicalShape;
        const outShape = outShapeInfo.logicalShape;
        if (inShape.length <= outShape.length) {
            if (usesPackedTextures) {
                res += getPackedSamplerAtOutputCoords(inInfo, outShapeInfo);
            }
            else {
                res += getSamplerAtOutputCoords(inInfo, outShapeInfo);
            }
        }
        return res;
    }
    function getPackedOutputSamplingSnippet(outShape, outTexShape, enableShapeUniforms) {
        switch (outShape.length) {
            case 0:
                return getOutputScalarCoords();
            case 1:
                return getOutputPacked1DCoords(outShape, outTexShape, enableShapeUniforms);
            case 2:
                return getOutputPacked2DCoords(outShape, outTexShape, enableShapeUniforms);
            case 3:
                return getOutputPacked3DCoords(outShape, outTexShape, enableShapeUniforms);
            default:
                return getOutputPackedNDCoords(outShape, outTexShape, enableShapeUniforms);
        }
    }
    function getOutputSamplingSnippet(outShape, outTexShape, enableShapeUniforms) {
        switch (outShape.length) {
            case 0:
                return getOutputScalarCoords();
            case 1:
                return getOutput1DCoords(outShape, outTexShape, enableShapeUniforms);
            case 2:
                return getOutput2DCoords(outShape, outTexShape, enableShapeUniforms);
            case 3:
                return getOutput3DCoords(outShape, outTexShape, enableShapeUniforms);
            case 4:
                return getOutput4DCoords(outShape, outTexShape, enableShapeUniforms);
            case 5:
                return getOutput5DCoords(outShape, outTexShape);
            case 6:
                return getOutput6DCoords(outShape, outTexShape);
            default:
                throw new Error(`${outShape.length}-D output sampling is not yet supported`);
        }
    }
    function getFloatTextureSampleSnippet(glsl) {
        return `
    float sampleTexture(sampler2D textureSampler, vec2 uv) {
      return ${glsl.texture2D}(textureSampler, uv).r;
    }
  `;
    }
    function getFloatTextureSetRSnippet(glsl) {
        return `
    void setOutput(float val) {
      ${glsl.output} = vec4(val, 0, 0, 0);
    }
  `;
    }
    function getFloatTextureSetRGBASnippet(glsl) {
        return `
    void setOutput(vec4 val) {
      ${glsl.output} = val;
    }
  `;
    }
    function getShaderPrefix(glsl) {
        const SHADER_PREFIX = `${glsl.version}
    precision highp float;
    precision highp int;
    precision highp sampler2D;
    ${glsl.varyingFs} vec2 resultUV;
    ${glsl.defineOutput}
    const vec2 halfCR = vec2(0.5, 0.5);

    struct ivec5
    {
      int x;
      int y;
      int z;
      int w;
      int u;
    };

    struct ivec6
    {
      int x;
      int y;
      int z;
      int w;
      int u;
      int v;
    };

    uniform float NAN;
    ${glsl.defineSpecialNaN}
    ${glsl.defineSpecialInf}
    ${glsl.defineRound}

    int imod(int x, int y) {
      return x - y * (x / y);
    }

    int idiv(int a, int b, float sign) {
      int res = a / b;
      int mod = imod(a, b);
      if (sign < 0. && mod != 0) {
        res -= 1;
      }
      return res;
    }

    //Based on the work of Dave Hoskins
    //https://www.shadertoy.com/view/4djSRW
    #define HASHSCALE1 443.8975
    float random(float seed){
      vec2 p = resultUV * seed;
      vec3 p3  = fract(vec3(p.xyx) * HASHSCALE1);
      p3 += dot(p3, p3.yzx + 19.19);
      return fract((p3.x + p3.y) * p3.z);
    }

    ${SAMPLE_1D_SNIPPET}
    ${SAMPLE_2D_SNIPPET}
    ${SAMPLE_3D_SNIPPET}
  `;
        return SHADER_PREFIX;
    }
    const SAMPLE_1D_SNIPPET = `
vec2 uvFromFlat(int texNumR, int texNumC, int index) {
  int texR = index / texNumC;
  int texC = index - texR * texNumC;
  return (vec2(texC, texR) + halfCR) / vec2(texNumC, texNumR);
}
vec2 packedUVfrom1D(int texNumR, int texNumC, int index) {
  int texelIndex = index / 2;
  int texR = texelIndex / texNumC;
  int texC = texelIndex - texR * texNumC;
  return (vec2(texC, texR) + halfCR) / vec2(texNumC, texNumR);
}
`;
    const SAMPLE_2D_SNIPPET = `
vec2 packedUVfrom2D(int texelsInLogicalRow, int texNumR,
  int texNumC, int row, int col) {
  int texelIndex = (row / 2) * texelsInLogicalRow + (col / 2);
  int texR = texelIndex / texNumC;
  int texC = texelIndex - texR * texNumC;
  return (vec2(texC, texR) + halfCR) / vec2(texNumC, texNumR);
}
`;
    const SAMPLE_3D_SNIPPET = `
vec2 packedUVfrom3D(int texNumR, int texNumC,
    int texelsInBatch, int texelsInLogicalRow, int b,
    int row, int col) {
  int index = b * texelsInBatch + (row / 2) * texelsInLogicalRow + (col / 2);
  int texR = index / texNumC;
  int texC = index - texR * texNumC;
  return (vec2(texC, texR) + halfCR) / vec2(texNumC, texNumR);
}
`;
    const SHADER_PACKED_PREFIX = `
  float getChannel(vec4 frag, vec2 innerDims) {
    vec2 modCoord = mod(innerDims, 2.);
    return modCoord.x == 0. ?
      (modCoord.y == 0. ? frag.r : frag.g) :
      (modCoord.y == 0. ? frag.b : frag.a);
  }
  float getChannel(vec4 frag, int dim) {
    float modCoord = mod(float(dim), 2.);
    return modCoord == 0. ? frag.r : frag.g;
  }
`;
    function getOutputScalarCoords() {
        return `
    int getOutputCoords() {
      return 0;
    }
  `;
    }
    function getOutputPacked1DCoords(shape, texShape, enableShapeUniforms) {
        const packedTexShape = [Math.ceil(texShape[0] / 2), Math.ceil(texShape[1] / 2)];
        if (packedTexShape[0] === 1) {
            if (enableShapeUniforms) {
                return `
      int getOutputCoords() {
        return 2 * int(resultUV.x * ceil(float(outTexShape[1]) / 2.0));
      }
    `;
            }
            return `
      int getOutputCoords() {
        return 2 * int(resultUV.x * ${packedTexShape[1]}.0);
      }
    `;
        }
        if (packedTexShape[1] === 1) {
            if (enableShapeUniforms) {
                return `
      int getOutputCoords() {
        return 2 * int(resultUV.y * ceil(float(outTexShape[0]) / 2.0));
      }
    `;
            }
            return `
      int getOutputCoords() {
        return 2 * int(resultUV.y * ${packedTexShape[0]}.0);
      }
    `;
        }
        if (enableShapeUniforms) {
            return `
    int getOutputCoords() {
      ivec2 packedTexShape = ivec2(ceil(float(outTexShape[0]) / 2.0), ceil(float(outTexShape[1]) / 2.0));
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(packedTexShape[0], packedTexShape[1]));
      return 2 * (resTexRC.x * packedTexShape[1] + resTexRC.y);
    }
  `;
        }
        return `
    int getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(${packedTexShape[0]}, ${packedTexShape[1]}));
      return 2 * (resTexRC.x * ${packedTexShape[1]} + resTexRC.y);
    }
  `;
    }
    function getOutput1DCoords(shape, texShape, enableShapeUniforms) {
        if (texShape[0] === 1) {
            if (enableShapeUniforms) {
                return `
      int getOutputCoords() {
        return int(resultUV.x * float(outTexShape[1]));
      }
    `;
            }
            return `
      int getOutputCoords() {
        return int(resultUV.x * ${texShape[1]}.0);
      }
    `;
        }
        if (texShape[1] === 1) {
            if (enableShapeUniforms) {
                return `
      int getOutputCoords() {
        return int(resultUV.y * float(outTexShape[0]));
      }
    `;
            }
            return `
      int getOutputCoords() {
        return int(resultUV.y * ${texShape[0]}.0);
      }
    `;
        }
        if (enableShapeUniforms) {
            return `
    int getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(outTexShape[0], outTexShape[1]));
      return resTexRC.x * outTexShape[1] + resTexRC.y;
    }
  `;
        }
        return `
    int getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(${texShape[0]}, ${texShape[1]}));
      return resTexRC.x * ${texShape[1]} + resTexRC.y;
    }
  `;
    }
    function getOutputPacked3DCoords(shape, texShape, enableShapeUniforms) {
        if (enableShapeUniforms) {
            return `
    ivec3 getOutputCoords() {
      ivec2 packedTexShape = ivec2(ceil(float(outTexShape[0]) / 2.0), ceil(float(outTexShape[1]) / 2.0));
      int texelsInLogicalRow = int(ceil(float(outShape[2]) / 2.0));
      int texelsInBatch = texelsInLogicalRow * int(ceil(float(outShape[1]) / 2.0));
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(packedTexShape[0], packedTexShape[1]));
      int index = resTexRC.x * packedTexShape[1] + resTexRC.y;

      int b = index / texelsInBatch;
      index -= b * texelsInBatch;

      int r = 2 * (index / texelsInLogicalRow);
      int c = imod(index, texelsInLogicalRow) * 2;

      return ivec3(b, r, c);
    }
  `;
        }
        const packedTexShape = [Math.ceil(texShape[0] / 2), Math.ceil(texShape[1] / 2)];
        const texelsInLogicalRow = Math.ceil(shape[2] / 2);
        const texelsInBatch = texelsInLogicalRow * Math.ceil(shape[1] / 2);
        return `
    ivec3 getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(${packedTexShape[0]}, ${packedTexShape[1]}));
      int index = resTexRC.x * ${packedTexShape[1]} + resTexRC.y;

      int b = index / ${texelsInBatch};
      index -= b * ${texelsInBatch};

      int r = 2 * (index / ${texelsInLogicalRow});
      int c = imod(index, ${texelsInLogicalRow}) * 2;

      return ivec3(b, r, c);
    }
  `;
    }
    function getOutput3DCoords(shape, texShape, enableShapeUniforms) {
        if (enableShapeUniforms) {
            const coordsFromIndexSnippet = getOutputLogicalCoordinatesFromFlatIndexByUniform(['r', 'c', 'd'], shape);
            return `
  ivec3 getOutputCoords() {
    ivec2 resTexRC = ivec2(resultUV.yx *
                           vec2(outTexShape[0], outTexShape[1]));
    int index = resTexRC.x * outTexShape[1] + resTexRC.y;
    ${coordsFromIndexSnippet}
    return ivec3(r, c, d);
  }
`;
        }
        const coordsFromIndexSnippet = getLogicalCoordinatesFromFlatIndex(['r', 'c', 'd'], shape);
        return `
    ivec3 getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(${texShape[0]}, ${texShape[1]}));
      int index = resTexRC.x * ${texShape[1]} + resTexRC.y;
      ${coordsFromIndexSnippet}
      return ivec3(r, c, d);
    }
  `;
    }
    function getOutputPackedNDCoords(shape, texShape, enableShapeUniforms) {
        if (enableShapeUniforms) {
            // TODO: support 5d and 6d
            return `
    ivec4 getOutputCoords() {
      ivec2 packedTexShape = ivec2(ceil(float(outTexShape[0]) / 2.0), ceil(float(outTexShape[1]) / 2.0));
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(packedTexShape[0], packedTexShape[1]));
      int index = resTexRC.x * packedTexShape[1] + resTexRC.y;

      int texelsInLogicalRow = int(ceil(float(outShape[3]) / 2.0));
      int texelsInBatch = texelsInLogicalRow * int(ceil(float(outShape[2]) / 2.0));
      int texelsInBatchN = texelsInBatch * outShape[1];

      int b2 = index / texelsInBatchN;
      index -= b2 * texelsInBatchN;

      int b = index / texelsInBatch;
      index -= b * texelsInBatch;

      int r = 2 * (index / texelsInLogicalRow);
      int c = imod(index, texelsInLogicalRow) * 2;

      return ivec4(b2, b, r, c);
    }
  `;
        }
        const packedTexShape = [Math.ceil(texShape[0] / 2), Math.ceil(texShape[1] / 2)];
        const texelsInLogicalRow = Math.ceil(shape[shape.length - 1] / 2);
        const texelsInBatch = texelsInLogicalRow * Math.ceil(shape[shape.length - 2] / 2);
        let texelsInBatchN = texelsInBatch;
        let batches = ``;
        let coords = 'b, r, c';
        for (let b = 2; b < shape.length - 1; b++) {
            texelsInBatchN *= shape[shape.length - b - 1];
            batches = `
      int b${b} = index / ${texelsInBatchN};
      index -= b${b} * ${texelsInBatchN};
    ` + batches;
            coords = `b${b}, ` + coords;
        }
        return `
    ivec${shape.length} getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(${packedTexShape[0]}, ${packedTexShape[1]}));
      int index = resTexRC.x * ${packedTexShape[1]} + resTexRC.y;

      ${batches}

      int b = index / ${texelsInBatch};
      index -= b * ${texelsInBatch};

      int r = 2 * (index / ${texelsInLogicalRow});
      int c = imod(index, ${texelsInLogicalRow}) * 2;

      return ivec${shape.length}(${coords});
    }
  `;
    }
    function getOutput4DCoords(shape, texShape, enableShapeUniforms) {
        if (enableShapeUniforms) {
            const coordsFromIndexSnippet = getOutputLogicalCoordinatesFromFlatIndexByUniform(['r', 'c', 'd', 'd2'], shape);
            return `
    ivec4 getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
        vec2(outTexShape[0], outTexShape[1]));
      int index = resTexRC.x * outTexShape[1] + resTexRC.y;
      ${coordsFromIndexSnippet}
      return ivec4(r, c, d, d2);
    }
  `;
        }
        const coordsFromIndexSnippet = getLogicalCoordinatesFromFlatIndex(['r', 'c', 'd', 'd2'], shape);
        return `
    ivec4 getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
        vec2(${texShape[0]}, ${texShape[1]}));
      int index = resTexRC.x * ${texShape[1]} + resTexRC.y;
      ${coordsFromIndexSnippet}
      return ivec4(r, c, d, d2);
    }
  `;
    }
    function getOutput5DCoords(shape, texShape) {
        const coordsFromIndexSnippet = getLogicalCoordinatesFromFlatIndex(['r', 'c', 'd', 'd2', 'd3'], shape);
        return `
    ivec5 getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx * vec2(${texShape[0]},
                             ${texShape[1]}));

      int index = resTexRC.x * ${texShape[1]} + resTexRC.y;

      ${coordsFromIndexSnippet}

      ivec5 outShape = ivec5(r, c, d, d2, d3);
      return outShape;
    }
  `;
    }
    function getOutput6DCoords(shape, texShape) {
        const coordsFromIndexSnippet = getLogicalCoordinatesFromFlatIndex(['r', 'c', 'd', 'd2', 'd3', 'd4'], shape);
        return `
    ivec6 getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
        vec2(${texShape[0]}, ${texShape[1]}));
      int index = resTexRC.x * ${texShape[1]} + resTexRC.y;

      ${coordsFromIndexSnippet}

      ivec6 result = ivec6(r, c, d, d2, d3, d4);
      return result;
    }
  `;
    }
    function getOutputPacked2DCoords(shape, texShape, enableShapeUniforms) {
        const packedTexShape = [Math.ceil(texShape[0] / 2), Math.ceil(texShape[1] / 2)];
        if (arraysEqual(shape, texShape)) {
            if (enableShapeUniforms) {
                return `
      ivec2 getOutputCoords() {
        ivec2 packedTexShape = ivec2(ceil(float(outTexShape[0]) / 2.0), ceil(float(outTexShape[1]) / 2.0));
        return 2 * ivec2(resultUV.yx * vec2(packedTexShape[0], packedTexShape[1]));
      }
    `;
            }
            return `
      ivec2 getOutputCoords() {
        return 2 * ivec2(resultUV.yx * vec2(${packedTexShape[0]}, ${packedTexShape[1]}));
      }
    `;
        }
        // texels needed to accommodate a logical row
        const texelsInLogicalRow = Math.ceil(shape[1] / 2);
        /**
         * getOutputCoords
         *
         * resTexRC: The rows and columns of the texels. If you move over one
         * texel to the right in the packed texture, you are moving over one column
         * (not two).
         *
         * index: The texel index
         */
        if (enableShapeUniforms) {
            return `
    ivec2 getOutputCoords() {
      ivec2 packedTexShape = ivec2(ceil(float(outTexShape[0]) / 2.0), ceil(float(outTexShape[1]) / 2.0));
      int texelsInLogicalRow = int(ceil(float(outShape[1]) / 2.0));
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(packedTexShape[0], packedTexShape[1]));

      int index = resTexRC.x * packedTexShape[1] + resTexRC.y;
      int r = 2 * (index / texelsInLogicalRow);
      int c = imod(index, texelsInLogicalRow) * 2;

      return ivec2(r, c);
    }
  `;
        }
        return `
    ivec2 getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(${packedTexShape[0]}, ${packedTexShape[1]}));

      int index = resTexRC.x * ${packedTexShape[1]} + resTexRC.y;
      int r = 2 * (index / ${texelsInLogicalRow});
      int c = imod(index, ${texelsInLogicalRow}) * 2;

      return ivec2(r, c);
    }
  `;
    }
    function getOutput2DCoords(shape, texShape, enableShapeUniforms) {
        if (arraysEqual(shape, texShape)) {
            if (enableShapeUniforms) {
                return `
      ivec2 getOutputCoords() {
        return ivec2(resultUV.yx * vec2(outTexShape[0], outTexShape[1]));
      }
    `;
            }
            return `
      ivec2 getOutputCoords() {
        return ivec2(resultUV.yx * vec2(${texShape[0]}, ${texShape[1]}));
      }
    `;
        }
        if (shape[1] === 1) {
            if (enableShapeUniforms) {
                return `
      ivec2 getOutputCoords() {
        ivec2 resTexRC = ivec2(resultUV.yx *
                               vec2(outTexShape[0], outTexShape[1]));
        int index = resTexRC.x * outTexShape[1] + resTexRC.y;
        return ivec2(index, 0);
      }
    `;
            }
            return `
      ivec2 getOutputCoords() {
        ivec2 resTexRC = ivec2(resultUV.yx *
                               vec2(${texShape[0]}, ${texShape[1]}));
        int index = resTexRC.x * ${texShape[1]} + resTexRC.y;
        return ivec2(index, 0);
      }
    `;
        }
        if (shape[0] === 1) {
            if (enableShapeUniforms) {
                return `
      ivec2 getOutputCoords() {
        ivec2 resTexRC = ivec2(resultUV.yx *
                               vec2(outTexShape[0], outTexShape[1]));
        int index = resTexRC.x * outTexShape[1] + resTexRC.y;
        return ivec2(0, index);
      }
    `;
            }
            return `
      ivec2 getOutputCoords() {
        ivec2 resTexRC = ivec2(resultUV.yx *
                               vec2(${texShape[0]}, ${texShape[1]}));
        int index = resTexRC.x * ${texShape[1]} + resTexRC.y;
        return ivec2(0, index);
      }
    `;
        }
        if (enableShapeUniforms) {
            return `
    ivec2 getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(outTexShape[0], outTexShape[1]));
      int index = resTexRC.x * outTexShape[1] + resTexRC.y;
      int r = index / outShape[1];
      int c = index - r * outShape[1];
      return ivec2(r, c);
    }
  `;
        }
        return `
    ivec2 getOutputCoords() {
      ivec2 resTexRC = ivec2(resultUV.yx *
                             vec2(${texShape[0]}, ${texShape[1]}));
      int index = resTexRC.x * ${texShape[1]} + resTexRC.y;
      int r = index / ${shape[1]};
      int c = index - r * ${shape[1]};
      return ivec2(r, c);
    }
  `;
    }
    function getFlatOffsetUniformName(texName) {
        return `offset${texName}`;
    }
    function getPackedSamplerScalar(inputInfo) {
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const glsl = getGlslDifferences();
        return `
    vec4 ${funcName}() {
      return ${glsl.texture2D}(${texName}, halfCR);
    }
  `;
    }
    function getSamplerScalar(inputInfo, enableShapeUniforms) {
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        if (inputInfo.shapeInfo.isUniform) {
            return `float ${funcName}() {return ${texName};}`;
        }
        const [texNumR, texNumC] = inputInfo.shapeInfo.texShape;
        if (texNumR === 1 && texNumC === 1) {
            return `
      float ${funcName}() {
        return sampleTexture(${texName}, halfCR);
      }
    `;
        }
        const offset = getFlatOffsetUniformName(texName);
        if (enableShapeUniforms) {
            return `
    float ${funcName}() {
      vec2 uv = uvFromFlat(${texName}TexShape[0], ${texName}TexShape[1], ${offset});
      return sampleTexture(${texName}, uv);
    }
  `;
        }
        const [tNumR, tNumC] = inputInfo.shapeInfo.texShape;
        return `
    float ${funcName}() {
      vec2 uv = uvFromFlat(${tNumR}, ${tNumC}, ${offset});
      return sampleTexture(${texName}, uv);
    }
  `;
    }
    function getPackedSampler1D(inputInfo, enableShapeUniforms) {
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const texShape = inputInfo.shapeInfo.texShape;
        const glsl = getGlslDifferences();
        if (enableShapeUniforms) {
            return `
    vec4 ${funcName}(int index) {
      ivec2 packedTexShape = ivec2(ceil(float(${texName}TexShape[0]) / 2.0), ceil(float(${texName}TexShape[1]) / 2.0));
      vec2 uv = packedUVfrom1D(
        packedTexShape[0], packedTexShape[1], index);
      return ${glsl.texture2D}(${texName}, uv);
    }
  `;
        }
        const packedTexShape = [Math.ceil(texShape[0] / 2), Math.ceil(texShape[1] / 2)];
        return `
    vec4 ${funcName}(int index) {
      vec2 uv = packedUVfrom1D(
        ${packedTexShape[0]}, ${packedTexShape[1]}, index);
      return ${glsl.texture2D}(${texName}, uv);
    }
  `;
    }
    function getSampler1D(inputInfo, enableShapeUniforms) {
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        if (inputInfo.shapeInfo.isUniform) {
            // Uniform arrays will be less than 65505 (no risk of float16 overflow).
            return `
      float ${funcName}(int index) {
        ${getUniformSampler(inputInfo)}
      }
    `;
        }
        const texShape = inputInfo.shapeInfo.texShape;
        const tNumR = texShape[0];
        const tNumC = texShape[1];
        if (tNumC === 1 && tNumR === 1) {
            return `
      float ${funcName}(int index) {
        return sampleTexture(${texName}, halfCR);
      }
    `;
        }
        const offset = getFlatOffsetUniformName(texName);
        if (tNumC === 1) {
            if (enableShapeUniforms) {
                return `
      float ${funcName}(int index) {
        vec2 uv = vec2(0.5, (float(index + ${offset}) + 0.5) / float(${texName}TexShape[0]));
        return sampleTexture(${texName}, uv);
      }
    `;
            }
            return `
      float ${funcName}(int index) {
        vec2 uv = vec2(0.5, (float(index + ${offset}) + 0.5) / ${tNumR}.0);
        return sampleTexture(${texName}, uv);
      }
    `;
        }
        if (tNumR === 1) {
            if (enableShapeUniforms) {
                return `
      float ${funcName}(int index) {
        vec2 uv = vec2((float(index + ${offset}) + 0.5) / float(${texName}TexShape[1]), 0.5);
        return sampleTexture(${texName}, uv);
      }
    `;
            }
            return `
      float ${funcName}(int index) {
        vec2 uv = vec2((float(index + ${offset}) + 0.5) / ${tNumC}.0, 0.5);
        return sampleTexture(${texName}, uv);
      }
    `;
        }
        if (enableShapeUniforms) {
            return `
    float ${funcName}(int index) {
      vec2 uv = uvFromFlat(${texName}TexShape[0], ${texName}TexShape[1], index + ${offset});
      return sampleTexture(${texName}, uv);
    }
  `;
        }
        return `
    float ${funcName}(int index) {
      vec2 uv = uvFromFlat(${tNumR}, ${tNumC}, index + ${offset});
      return sampleTexture(${texName}, uv);
    }
  `;
    }
    function getPackedSampler2D(inputInfo, enableShapeUniforms) {
        const shape = inputInfo.shapeInfo.logicalShape;
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const texShape = inputInfo.shapeInfo.texShape;
        const texNumR = texShape[0];
        const texNumC = texShape[1];
        const glsl = getGlslDifferences();
        if (texShape != null && arraysEqual(shape, texShape)) {
            if (enableShapeUniforms) {
                return `
      vec4 ${funcName}(int row, int col) {
        vec2 uv = (vec2(col, row) + halfCR) / vec2(${texName}TexShape[1], ${texName}TexShape[0]);

        return ${glsl.texture2D}(${texName}, uv);
      }
    `;
            }
            return `
      vec4 ${funcName}(int row, int col) {
        vec2 uv = (vec2(col, row) + halfCR) / vec2(${texNumC}.0, ${texNumR}.0);

        return ${glsl.texture2D}(${texName}, uv);
      }
    `;
        }
        if (enableShapeUniforms) {
            return `
    vec4 ${funcName}(int row, int col) {
      ivec2 packedTexShape = ivec2(ceil(float(${texName}TexShape[0]) / 2.0), ceil(float(${texName}TexShape[1]) / 2.0));
      int valuesPerRow = int(ceil(float(${texName}Shape[1]) / 2.0));
      vec2 uv = packedUVfrom2D(valuesPerRow, packedTexShape[0], packedTexShape[1], row, col);
      return ${glsl.texture2D}(${texName}, uv);
    }
  `;
        }
        const packedTexShape = [Math.ceil(texShape[0] / 2), Math.ceil(texShape[1] / 2)];
        const valuesPerRow = Math.ceil(shape[1] / 2);
        return `
    vec4 ${funcName}(int row, int col) {
      vec2 uv = packedUVfrom2D(${valuesPerRow}, ${packedTexShape[0]}, ${packedTexShape[1]}, row, col);
      return ${glsl.texture2D}(${texName}, uv);
    }
  `;
    }
    function getSampler2D(inputInfo, enableShapeUniforms) {
        const shape = inputInfo.shapeInfo.logicalShape;
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const texShape = inputInfo.shapeInfo.texShape;
        if (texShape != null && arraysEqual(shape, texShape)) {
            if (enableShapeUniforms) {
                return `
      float ${funcName}(int row, int col) {
        vec2 uv = (vec2(col, row) + halfCR) / vec2(${texName}TexShape[1], ${texName}TexShape[0]);
        return sampleTexture(${texName}, uv);
      }
    `;
            }
            const texNumR = texShape[0];
            const texNumC = texShape[1];
            return `
    float ${funcName}(int row, int col) {
      vec2 uv = (vec2(col, row) + halfCR) / vec2(${texNumC}.0, ${texNumR}.0);
      return sampleTexture(${texName}, uv);
    }
  `;
        }
        const { newShape, keptDims } = squeezeShape(shape);
        const squeezedShape = newShape;
        if (squeezedShape.length < shape.length) {
            const newInputInfo = squeezeInputInfo(inputInfo, squeezedShape);
            const params = ['row', 'col'];
            return `
      ${getSamplerFromInInfo(newInputInfo, enableShapeUniforms)}
      float ${funcName}(int row, int col) {
        return ${funcName}(${getSqueezedParams(params, keptDims)});
      }
    `;
        }
        if (inputInfo.shapeInfo.isUniform) {
            // Uniform arrays will be less than 65505 (no risk of float16 overflow).
            return `
      float ${funcName}(int row, int col) {
        int index = round(dot(vec2(row, col), vec2(${shape[1]}, 1)));
        ${getUniformSampler(inputInfo)}
      }
    `;
        }
        const texNumR = texShape[0];
        const texNumC = texShape[1];
        const offset = getFlatOffsetUniformName(texName);
        if (texNumC === 1) {
            // index is used directly as physical (no risk of float16 overflow).
            if (enableShapeUniforms) {
                return `
      float ${funcName}(int row, int col) {
        float index = dot(vec3(row, col, ${offset}), vec3(${texName}Shape[1], 1, 1));
        vec2 uv = vec2(0.5, (index + 0.5) / float(${texName}TexShape[0]));
        return sampleTexture(${texName}, uv);
      }
    `;
            }
            return `
    float ${funcName}(int row, int col) {
      float index = dot(vec3(row, col, ${offset}), vec3(${shape[1]}, 1, 1));
      vec2 uv = vec2(0.5, (index + 0.5) / ${texNumR}.0);
      return sampleTexture(${texName}, uv);
    }
  `;
        }
        if (texNumR === 1) {
            // index is used directly as physical (no risk of float16 overflow).
            if (enableShapeUniforms) {
                return `
      float ${funcName}(int row, int col) {
        float index = dot(vec3(row, col, ${offset}), vec3(${texName}Shape[1], 1, 1));
        vec2 uv = vec2((index + 0.5) / float(${texName}TexShape[1]), 0.5);
        return sampleTexture(${texName}, uv);
      }
    `;
            }
            return `
    float ${funcName}(int row, int col) {
      float index = dot(vec3(row, col, ${offset}), vec3(${shape[1]}, 1, 1));
      vec2 uv = vec2((index + 0.5) / ${texNumC}.0, 0.5);
      return sampleTexture(${texName}, uv);
    }
  `;
        }
        if (enableShapeUniforms) {
            return `
      float ${funcName}(int row, int col) {
        // Explicitly use integer operations as dot() only works on floats.
        int index = row * ${texName}Shape[1] + col + ${offset};
        vec2 uv = uvFromFlat(${texName}TexShape[0], ${texName}TexShape[1], index);
        return sampleTexture(${texName}, uv);
      }
    `;
        }
        return `
  float ${funcName}(int row, int col) {
    // Explicitly use integer operations as dot() only works on floats.
    int index = row * ${shape[1]} + col + ${offset};
    vec2 uv = uvFromFlat(${texNumR}, ${texNumC}, index);
    return sampleTexture(${texName}, uv);
  }
`;
    }
    function getPackedSampler3D(inputInfo, enableShapeUniforms) {
        const shape = inputInfo.shapeInfo.logicalShape;
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const texShape = inputInfo.shapeInfo.texShape;
        const packedTexShape = [Math.ceil(texShape[0] / 2), Math.ceil(texShape[1] / 2)];
        if (shape[0] === 1) {
            const squeezedShape = shape.slice(1);
            const keptDims = [1, 2];
            const newInputInfo = squeezeInputInfo(inputInfo, squeezedShape);
            const params = ['b', 'row', 'col'];
            return `
        ${getPackedSamplerFromInInfo(newInputInfo, enableShapeUniforms)}
        vec4 ${funcName}(int b, int row, int col) {
          return ${funcName}(${getSqueezedParams(params, keptDims)});
        }
      `;
        }
        const glsl = getGlslDifferences();
        if (enableShapeUniforms) {
            return `
    vec4 ${funcName}(int b, int row, int col) {
      ivec2 packedTexShape = ivec2(ceil(float(${texName}TexShape[0]) / 2.0), ceil(float(${texName}TexShape[1]) / 2.0));
      int valuesPerRow = int(ceil(float(${texName}Shape[2]) / 2.0));
      int texelsInBatch = valuesPerRow * int(ceil(float(${texName}Shape[1]) / 2.0));
      vec2 uv = packedUVfrom3D(
        packedTexShape[0], packedTexShape[1], texelsInBatch, valuesPerRow, b, row, col);
      return ${glsl.texture2D}(${texName}, uv);
    }
  `;
        }
        const texNumR = packedTexShape[0];
        const texNumC = packedTexShape[1];
        const valuesPerRow = Math.ceil(shape[2] / 2);
        const texelsInBatch = valuesPerRow * Math.ceil(shape[1] / 2);
        return `
    vec4 ${funcName}(int b, int row, int col) {
      vec2 uv = packedUVfrom3D(
        ${texNumR}, ${texNumC}, ${texelsInBatch}, ${valuesPerRow}, b, row, col);
      return ${glsl.texture2D}(${texName}, uv);
    }
  `;
    }
    function getSampler3D(inputInfo, enableShapeUniforms) {
        const shape = inputInfo.shapeInfo.logicalShape;
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const stride0 = shape[1] * shape[2];
        const stride1 = shape[2];
        const { newShape, keptDims } = squeezeShape(shape);
        const squeezedShape = newShape;
        if (squeezedShape.length < shape.length) {
            const newInputInfo = squeezeInputInfo(inputInfo, squeezedShape);
            const params = ['row', 'col', 'depth'];
            return `
        ${getSamplerFromInInfo(newInputInfo, enableShapeUniforms)}
        float ${funcName}(int row, int col, int depth) {
          return ${funcName}(${getSqueezedParams(params, keptDims)});
        }
      `;
        }
        if (inputInfo.shapeInfo.isUniform) {
            // Uniform arrays will be less than 65505 (no risk of float16 overflow).
            return `
      float ${funcName}(int row, int col, int depth) {
        int index = round(dot(vec3(row, col, depth),
                          vec3(${stride0}, ${stride1}, 1)));
        ${getUniformSampler(inputInfo)}
      }
    `;
        }
        const texShape = inputInfo.shapeInfo.texShape;
        const texNumR = texShape[0];
        const texNumC = texShape[1];
        const flatOffset = inputInfo.shapeInfo.flatOffset;
        if (texNumC === stride0 && flatOffset == null) {
            // texC is used directly as physical (no risk of float16 overflow).
            if (enableShapeUniforms) {
                return `
      float ${funcName}(int row, int col, int depth) {
        int stride1 = ${texName}Shape[2];
        float texR = float(row);
        float texC = dot(vec2(col, depth), vec2(stride1, 1));
        vec2 uv = (vec2(texC, texR) + halfCR) /
                   vec2(${texName}TexShape[1], ${texName}TexShape[0]);
        return sampleTexture(${texName}, uv);
      }
    `;
            }
            return `
        float ${funcName}(int row, int col, int depth) {
          float texR = float(row);
          float texC = dot(vec2(col, depth), vec2(${stride1}, 1));
          vec2 uv = (vec2(texC, texR) + halfCR) /
                     vec2(${texNumC}.0, ${texNumR}.0);
          return sampleTexture(${texName}, uv);
        }
      `;
        }
        if (texNumC === stride1 && flatOffset == null) {
            // texR is used directly as physical (no risk of float16 overflow).
            if (enableShapeUniforms) {
                return `
      float ${funcName}(int row, int col, int depth) {
        float texR = dot(vec2(row, col), vec2(${texName}Shape[1], 1));
        float texC = float(depth);
        vec2 uv = (vec2(texC, texR) + halfCR) / vec2(${texName}TexShape[1], ${texName}TexShape[0]);
        return sampleTexture(${texName}, uv);
      }
    `;
            }
            return `
    float ${funcName}(int row, int col, int depth) {
      float texR = dot(vec2(row, col), vec2(${shape[1]}, 1));
      float texC = float(depth);
      vec2 uv = (vec2(texC, texR) + halfCR) / vec2(${texNumC}.0, ${texNumR}.0);
      return sampleTexture(${texName}, uv);
    }
  `;
        }
        const offset = getFlatOffsetUniformName(texName);
        if (enableShapeUniforms) {
            return `
    float ${funcName}(int row, int col, int depth) {
      // Explicitly use integer operations as dot() only works on floats.
      int stride0 = ${texName}Shape[1] * ${texName}Shape[2];
      int stride1 = ${texName}Shape[2];
      int index = row * ${stride0} + col * ${stride1} + depth + ${offset};
      vec2 uv = uvFromFlat(${texName}TexShape[0], ${texName}TexShape[1], index);
      return sampleTexture(${texName}, uv);
    }
    `;
        }
        return `
      float ${funcName}(int row, int col, int depth) {
        // Explicitly use integer operations as dot() only works on floats.
        int index = row * ${stride0} + col * ${stride1} + depth + ${offset};
        vec2 uv = uvFromFlat(${texNumR}, ${texNumC}, index);
        return sampleTexture(${texName}, uv);
      }
  `;
    }
    function getPackedSamplerND(inputInfo, enableShapeUniforms) {
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const glsl = getGlslDifferences();
        if (enableShapeUniforms) {
            // TODO: support 5d and 6d
            return `
    vec4 ${funcName}(int b2, int b, int row, int col) {
      int valuesPerRow = int(ceil(float(${texName}Shape[3]) / 2.0));
      int texelsInBatch = valuesPerRow * int(ceil(float(${texName}Shape[2]) / 2.0));
      int index = b * texelsInBatch + (row / 2) * valuesPerRow + (col / 2);
      texelsInBatch *= ${texName}Shape[1];
      index = b2 * texelsInBatch + index;
      ivec2 packedTexShape = ivec2(ceil(float(${texName}TexShape[0]) / 2.0), ceil(float(${texName}TexShape[1]) / 2.0));
      int texR = index / packedTexShape[1];
      int texC = index - texR * packedTexShape[1];
      vec2 uv = (vec2(texC, texR) + halfCR) / vec2(packedTexShape[1], packedTexShape[0]); return ${glsl.texture2D}(${texName}, uv);
    }
  `;
        }
        const shape = inputInfo.shapeInfo.logicalShape;
        const rank = shape.length;
        const texShape = inputInfo.shapeInfo.texShape;
        const packedTexShape = [Math.ceil(texShape[0] / 2), Math.ceil(texShape[1] / 2)];
        const texNumR = packedTexShape[0];
        const texNumC = packedTexShape[1];
        const valuesPerRow = Math.ceil(shape[rank - 1] / 2);
        let texelsInBatch = valuesPerRow * Math.ceil(shape[rank - 2] / 2);
        let params = `int b, int row, int col`;
        let index = `b * ${texelsInBatch} + (row / 2) * ${valuesPerRow} + (col / 2)`;
        for (let b = 2; b < rank - 1; b++) {
            params = `int b${b}, ` + params;
            texelsInBatch *= shape[rank - b - 1];
            index = `b${b} * ${texelsInBatch} + ` + index;
        }
        return `
    vec4 ${funcName}(${params}) {
      int index = ${index};
      int texR = index / ${texNumC};
      int texC = index - texR * ${texNumC};
      vec2 uv = (vec2(texC, texR) + halfCR) / vec2(${texNumC}, ${texNumR});
      return ${glsl.texture2D}(${texName}, uv);
    }
  `;
    }
    function getSampler4D(inputInfo, enableShapeUniforms) {
        const shape = inputInfo.shapeInfo.logicalShape;
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const stride2 = shape[3];
        const stride1 = shape[2] * stride2;
        const stride0 = shape[1] * stride1;
        const { newShape, keptDims } = squeezeShape(shape);
        if (newShape.length < shape.length) {
            const newInputInfo = squeezeInputInfo(inputInfo, newShape);
            const params = ['row', 'col', 'depth', 'depth2'];
            return `
      ${getSamplerFromInInfo(newInputInfo, enableShapeUniforms)}
      float ${funcName}(int row, int col, int depth, int depth2) {
        return ${funcName}(${getSqueezedParams(params, keptDims)});
      }
    `;
        }
        if (inputInfo.shapeInfo.isUniform) {
            // Uniform arrays will be less than 65505 (no risk of float16 overflow).
            return `
      float ${funcName}(int row, int col, int depth, int depth2) {
        int index = round(dot(vec4(row, col, depth, depth2),
                          vec4(${stride0}, ${stride1}, ${stride2}, 1)));
        ${getUniformSampler(inputInfo)}
      }
    `;
        }
        const flatOffset = inputInfo.shapeInfo.flatOffset;
        const texShape = inputInfo.shapeInfo.texShape;
        const texNumR = texShape[0];
        const texNumC = texShape[1];
        const stride2Str = `int stride2 = ${texName}Shape[3];`;
        const stride1Str = `int stride1 = ${texName}Shape[2] * stride2;`;
        const stride0Str = `int stride0 = ${texName}Shape[1] * stride1;`;
        if (texNumC === stride0 && flatOffset == null) {
            // texC is used directly as physical (no risk of float16 overflow).
            if (enableShapeUniforms) {
                return `
      float ${funcName}(int row, int col, int depth, int depth2) {
        ${stride2Str}
        ${stride1Str}
        float texR = float(row);
        float texC =
            dot(vec3(col, depth, depth2),
                vec3(stride1, stride2, 1));
        vec2 uv = (vec2(texC, texR) + halfCR) /
                   vec2(${texName}TexShape[1], ${texName}TexShape[0]);
        return sampleTexture(${texName}, uv);
      }
    `;
            }
            return `
      float ${funcName}(int row, int col, int depth, int depth2) {
        float texR = float(row);
        float texC =
            dot(vec3(col, depth, depth2),
                vec3(${stride1}, ${stride2}, 1));
        vec2 uv = (vec2(texC, texR) + halfCR) /
                   vec2(${texNumC}.0, ${texNumR}.0);
        return sampleTexture(${texName}, uv);
      }
    `;
        }
        if (texNumC === stride2 && flatOffset == null) {
            // texR is used directly as physical (no risk of float16 overflow).
            if (enableShapeUniforms) {
                return `
      float ${funcName}(int row, int col, int depth, int depth2) {
        float texR = dot(vec3(row, col, depth),
                         vec3(${texName}Shape[1] * ${texName}Shape[2], ${texName}Shape[2], 1));
        float texC = float(depth2);
        vec2 uv = (vec2(texC, texR) + halfCR) /
                  vec2(${texName}TexShape[1], ${texName}TexShape[0]);
        return sampleTexture(${texName}, uv);
      }
    `;
            }
            return `
      float ${funcName}(int row, int col, int depth, int depth2) {
        float texR = dot(vec3(row, col, depth),
                         vec3(${shape[1] * shape[2]}, ${shape[2]}, 1));
        float texC = float(depth2);
        vec2 uv = (vec2(texC, texR) + halfCR) /
                  vec2(${texNumC}.0, ${texNumR}.0);
        return sampleTexture(${texName}, uv);
      }
    `;
        }
        const offset = getFlatOffsetUniformName(texName);
        if (enableShapeUniforms) {
            return `
    float ${funcName}(int row, int col, int depth, int depth2) {
      // Explicitly use integer operations as dot() only works on floats.
      ${stride2Str}
      ${stride1Str}
      ${stride0Str}
      int index = row * stride0 + col * stride1 +
          depth * stride2 + depth2;
      vec2 uv = uvFromFlat(${texName}TexShape[0], ${texName}TexShape[1], index + ${offset});
      return sampleTexture(${texName}, uv);
    }
  `;
        }
        return `
    float ${funcName}(int row, int col, int depth, int depth2) {
      // Explicitly use integer operations as dot() only works on floats.
      int index = row * ${stride0} + col * ${stride1} +
          depth * ${stride2} + depth2;
      vec2 uv = uvFromFlat(${texNumR}, ${texNumC}, index + ${offset});
      return sampleTexture(${texName}, uv);
    }
  `;
    }
    function getSampler5D(inputInfo) {
        const shape = inputInfo.shapeInfo.logicalShape;
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const stride3 = shape[4];
        const stride2 = shape[3] * stride3;
        const stride1 = shape[2] * stride2;
        const stride0 = shape[1] * stride1;
        const { newShape, keptDims } = squeezeShape(shape);
        if (newShape.length < shape.length) {
            const newInputInfo = squeezeInputInfo(inputInfo, newShape);
            const params = ['row', 'col', 'depth', 'depth2', 'depth3'];
            return `
      ${getSamplerFromInInfo(newInputInfo)}
      float ${funcName}(int row, int col, int depth, int depth2, int depth3) {
        return ${funcName}(${getSqueezedParams(params, keptDims)});
      }
    `;
        }
        if (inputInfo.shapeInfo.isUniform) {
            // Uniform arrays will be less than 65505 (no risk of float16 overflow).
            return `
      float ${funcName}(int row, int col, int depth, int depth2, int depth3) {
        float index = dot(
          vec4(row, col, depth, depth2),
          vec4(${stride0}, ${stride1}, ${stride2}, ${stride3})) +
          depth3;
        ${getUniformSampler(inputInfo)}
      }
    `;
        }
        const flatOffset = inputInfo.shapeInfo.flatOffset;
        const texShape = inputInfo.shapeInfo.texShape;
        const texNumR = texShape[0];
        const texNumC = texShape[1];
        if (texNumC === stride0 && flatOffset == null) {
            // texC is used directly as physical (no risk of float16 overflow).
            return `
      float ${funcName}(int row, int col, int depth, int depth2, int depth3) {
        int texR = row;
        float texC = dot(vec4(col, depth, depth2, depth3),
                         vec4(${stride1}, ${stride2}, ${stride3}, 1));
        vec2 uv = (vec2(texC, texR) + halfCR) /
                   vec2(${texNumC}.0, ${texNumR}.0);
        return sampleTexture(${texName}, uv);
      }
    `;
        }
        if (texNumC === stride3 && flatOffset == null) {
            // texR is used directly as physical (no risk of float16 overflow).
            return `
      float ${funcName}(int row, int col, int depth, int depth2, int depth3) {
        float texR = dot(
          vec4(row, col, depth, depth2),
          vec4(${shape[1] * shape[2] * shape[3]},
               ${shape[2] * shape[3]}, ${shape[3]}, 1));
        int texC = depth3;
        vec2 uv = (vec2(texC, texR) + halfCR) /
                  vec2(${texNumC}.0, ${texNumR}.0);
        return sampleTexture(${texName}, uv);
      }
    `;
        }
        const offset = getFlatOffsetUniformName(texName);
        return `
    float ${funcName}(int row, int col, int depth, int depth2, int depth3) {
      // Explicitly use integer operations as dot() only works on floats.
      int index = row * ${stride0} + col * ${stride1} + depth * ${stride2} +
          depth2 * ${stride3} + depth3 + ${offset};
      vec2 uv = uvFromFlat(${texNumR}, ${texNumC}, index);
      return sampleTexture(${texName}, uv);
    }
  `;
    }
    function getSampler6D(inputInfo) {
        const shape = inputInfo.shapeInfo.logicalShape;
        const texName = inputInfo.name;
        const funcName = 'get' + texName.charAt(0).toUpperCase() + texName.slice(1);
        const { newShape, keptDims } = squeezeShape(shape);
        if (newShape.length < shape.length) {
            const newInputInfo = squeezeInputInfo(inputInfo, newShape);
            const params = ['row', 'col', 'depth', 'depth2', 'depth3', 'depth4'];
            return `
      ${getSamplerFromInInfo(newInputInfo)}
      float ${funcName}(int row, int col, int depth,
                    int depth2, int depth3, int depth4) {
        return ${funcName}(${getSqueezedParams(params, keptDims)});
      }
    `;
        }
        const stride4 = shape[5];
        const stride3 = shape[4] * stride4;
        const stride2 = shape[3] * stride3;
        const stride1 = shape[2] * stride2;
        const stride0 = shape[1] * stride1;
        if (inputInfo.shapeInfo.isUniform) {
            // Uniform arrays will be less than 65505 (no risk of float16 overflow).
            return `
      float ${funcName}(int row, int col, int depth,
                  int depth2, int depth3, int depth4) {
        int index = round(dot(
          vec4(row, col, depth, depth2),
          vec4(${stride0}, ${stride1}, ${stride2}, ${stride3})) +
          dot(
            vec2(depth3, depth4),
            vec2(${stride4}, 1)));
        ${getUniformSampler(inputInfo)}
      }
    `;
        }
        const flatOffset = inputInfo.shapeInfo.flatOffset;
        const texShape = inputInfo.shapeInfo.texShape;
        const texNumR = texShape[0];
        const texNumC = texShape[1];
        if (texNumC === stride0 && flatOffset == null) {
            // texC is used directly as physical (no risk of float16 overflow).
            return `
      float ${funcName}(int row, int col, int depth,
                    int depth2, int depth3, int depth4) {
        int texR = row;
        float texC = dot(vec4(col, depth, depth2, depth3),
          vec4(${stride1}, ${stride2}, ${stride3}, ${stride4})) +
               float(depth4);
        vec2 uv = (vec2(texC, texR) + halfCR) /
                   vec2(${texNumC}.0, ${texNumR}.0);
        return sampleTexture(${texName}, uv);
      }
    `;
        }
        if (texNumC === stride4 && flatOffset == null) {
            // texR is used directly as physical (no risk of float16 overflow).
            return `
      float ${funcName}(int row, int col, int depth,
                    int depth2, int depth3, int depth4) {
        float texR = dot(vec4(row, col, depth, depth2),
          vec4(${shape[1] * shape[2] * shape[3] * shape[4]},
               ${shape[2] * shape[3] * shape[4]},
               ${shape[3] * shape[4]},
               ${shape[4]})) + float(depth3);
        int texC = depth4;
        vec2 uv = (vec2(texC, texR) + halfCR) /
                  vec2(${texNumC}.0, ${texNumR}.0);
        return sampleTexture(${texName}, uv);
      }
    `;
        }
        const offset = getFlatOffsetUniformName(texName);
        return `
    float ${funcName}(int row, int col, int depth,
                  int depth2, int depth3, int depth4) {
      // Explicitly use integer operations as dot() only works on floats.
      int index = row * ${stride0} + col * ${stride1} + depth * ${stride2} +
          depth2 * ${stride3} + depth3 * ${stride4} + depth4 + ${offset};
      vec2 uv = uvFromFlat(${texNumR}, ${texNumC}, index);
      return sampleTexture(${texName}, uv);
    }
  `;
    }
    function getUniformSampler(inputInfo) {
        const texName = inputInfo.name;
        const inSize = sizeFromShape(inputInfo.shapeInfo.logicalShape);
        if (inSize < 2) {
            return `return ${texName};`;
        }
        return `
    for (int i = 0; i < ${inSize}; i++) {
      if (i == index) {
        return ${texName}[i];
      }
    }
  `;
    }
    function getPackedSamplerAtOutputCoords(inputInfo, outShapeInfo) {
        const texName = inputInfo.name;
        const texFuncSnippet = texName.charAt(0).toUpperCase() + texName.slice(1);
        const funcName = 'get' + texFuncSnippet + 'AtOutCoords';
        const inRank = inputInfo.shapeInfo.logicalShape.length;
        const outRank = outShapeInfo.logicalShape.length;
        const broadcastDims = getBroadcastDims(inputInfo.shapeInfo.logicalShape, outShapeInfo.logicalShape);
        const type = getCoordsDataType(outRank);
        const rankDiff = outRank - inRank;
        let coordsSnippet;
        const fields = ['x', 'y', 'z', 'w', 'u', 'v'];
        if (inRank === 0) {
            coordsSnippet = '';
        }
        else if (outRank < 2 && broadcastDims.length >= 1) {
            coordsSnippet = 'coords = 0;';
        }
        else {
            coordsSnippet =
                broadcastDims.map(d => `coords.${fields[d + rankDiff]} = 0;`)
                    .join('\n');
        }
        let unpackedCoordsSnippet = '';
        if (outRank < 2 && inRank > 0) {
            unpackedCoordsSnippet = 'coords';
        }
        else {
            unpackedCoordsSnippet = inputInfo.shapeInfo.logicalShape
                .map((s, i) => `coords.${fields[i + rankDiff]}`)
                .join(', ');
        }
        let output = `return outputValue;`;
        const inSize = sizeFromShape(inputInfo.shapeInfo.logicalShape);
        const isInputScalar = inSize === 1;
        const outSize = sizeFromShape(outShapeInfo.logicalShape);
        const isOutputScalar = outSize === 1;
        if (inRank === 1 && !isInputScalar && !isOutputScalar) {
            output = `
      return vec4(outputValue.xy, outputValue.xy);
    `;
        }
        else if (isInputScalar && !isOutputScalar) {
            if (outRank === 1) {
                output = `
        return vec4(outputValue.x, outputValue.x, 0., 0.);
      `;
            }
            else {
                output = `
        return vec4(outputValue.x);
      `;
            }
        }
        else if (broadcastDims.length) {
            const rows = inRank - 2;
            const cols = inRank - 1;
            if (broadcastDims.indexOf(rows) > -1 && broadcastDims.indexOf(cols) > -1) {
                output = `return vec4(outputValue.x);`;
            }
            else if (broadcastDims.indexOf(rows) > -1) {
                output = `return vec4(outputValue.x, outputValue.y, ` +
                    `outputValue.x, outputValue.y);`;
            }
            else if (broadcastDims.indexOf(cols) > -1) {
                output = `return vec4(outputValue.xx, outputValue.zz);`;
            }
        }
        return `
    vec4 ${funcName}() {
      ${type} coords = getOutputCoords();
      ${coordsSnippet}
      vec4 outputValue = get${texFuncSnippet}(${unpackedCoordsSnippet});
      ${output}
    }
  `;
    }
    function getSamplerAtOutputCoords(inputInfo, outShapeInfo) {
        const texName = inputInfo.name;
        const texFuncSnippet = texName.charAt(0).toUpperCase() + texName.slice(1);
        const funcName = 'get' + texFuncSnippet + 'AtOutCoords';
        const outTexShape = outShapeInfo.texShape;
        const inTexShape = inputInfo.shapeInfo.texShape;
        const inRank = inputInfo.shapeInfo.logicalShape.length;
        const outRank = outShapeInfo.logicalShape.length;
        if (!inputInfo.shapeInfo.isUniform && inRank === outRank &&
            inputInfo.shapeInfo.flatOffset == null &&
            arraysEqual(inTexShape, outTexShape)) {
            return `
      float ${funcName}() {
        return sampleTexture(${texName}, resultUV);
      }
    `;
        }
        const type = getCoordsDataType(outRank);
        const broadcastDims = getBroadcastDims(inputInfo.shapeInfo.logicalShape, outShapeInfo.logicalShape);
        const rankDiff = outRank - inRank;
        let coordsSnippet;
        const fields = ['x', 'y', 'z', 'w', 'u', 'v'];
        if (inRank === 0) {
            coordsSnippet = '';
        }
        else if (outRank < 2 && broadcastDims.length >= 1) {
            coordsSnippet = 'coords = 0;';
        }
        else {
            coordsSnippet =
                broadcastDims.map(d => `coords.${fields[d + rankDiff]} = 0;`)
                    .join('\n');
        }
        let unpackedCoordsSnippet = '';
        if (outRank < 2 && inRank > 0) {
            unpackedCoordsSnippet = 'coords';
        }
        else {
            unpackedCoordsSnippet = inputInfo.shapeInfo.logicalShape
                .map((s, i) => `coords.${fields[i + rankDiff]}`)
                .join(', ');
        }
        return `
    float ${funcName}() {
      ${type} coords = getOutputCoords();
      ${coordsSnippet}
      return get${texFuncSnippet}(${unpackedCoordsSnippet});
    }
  `;
    }
    function getCoordsDataType(rank) {
        if (rank <= 1) {
            return 'int';
        }
        else if (rank === 2) {
            return 'ivec2';
        }
        else if (rank === 3) {
            return 'ivec3';
        }
        else if (rank === 4) {
            return 'ivec4';
        }
        else if (rank === 5) {
            return 'ivec5';
        }
        else if (rank === 6) {
            return 'ivec6';
        }
        else {
            throw Error(`GPU for rank ${rank} is not yet supported`);
        }
    }
    function getUniformInfoFromShape(isPacked, shape, texShape) {
        const { newShape, keptDims } = squeezeShape(shape);
        const rank = shape.length;
        const useSqueezePackedShape = isPacked && rank === 3 && shape[0] === 1;
        const squeezeShape$1 = useSqueezePackedShape ? shape.slice(1) : newShape;
        const useSqueezeShape = (!isPacked && rank > 1 && !arraysEqual(shape, texShape) &&
            newShape.length < rank) ||
            useSqueezePackedShape;
        const uniformShape = useSqueezeShape ? squeezeShape$1 : shape;
        return { useSqueezeShape, uniformShape, keptDims };
    }
    /** Returns a new input info (a copy) that has a squeezed logical shape. */
    function squeezeInputInfo(inInfo, squeezedShape) {
        // Deep copy.
        const newInputInfo = JSON.parse(JSON.stringify(inInfo));
        newInputInfo.shapeInfo.logicalShape = squeezedShape;
        return newInputInfo;
    }
    function getSqueezedParams(params, keptDims) {
        return keptDims.map(d => params[d]).join(', ');
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function compileProgram(gpgpu, program, inputs, output) {
        const inputInfos = inputs.map((input, i) => {
            const shapeInfo = {
                logicalShape: input.shape,
                texShape: input.isUniform ? null : input.texData.texShape,
                isUniform: input.isUniform,
                isPacked: input.isUniform ? false : input.texData.isPacked,
                flatOffset: null
            };
            if (input.texData != null && input.texData.slice != null &&
                input.texData.slice.flatOffset > 0) {
                shapeInfo.flatOffset = input.texData.slice.flatOffset;
            }
            return { name: program.variableNames[i], shapeInfo };
        });
        const inShapeInfos = inputInfos.map(x => x.shapeInfo);
        const outShapeInfo = {
            logicalShape: output.shape,
            texShape: output.texData.texShape,
            isUniform: false,
            isPacked: output.texData.isPacked,
            flatOffset: null
        };
        const source = makeShader(inputInfos, outShapeInfo, program);
        const fragmentShader = createFragmentShader(gpgpu.gl, source);
        const webGLProgram = gpgpu.createProgram(fragmentShader);
        if (!env$1().get('ENGINE_COMPILE_ONLY')) {
            return Object.assign({ program,
                fragmentShader,
                source,
                webGLProgram,
                inShapeInfos,
                outShapeInfo }, getUniformLocations(gpgpu, program, webGLProgram));
        }
        else {
            return {
                program,
                fragmentShader,
                source,
                webGLProgram,
                inShapeInfos,
                outShapeInfo,
                uniformLocations: null,
                customUniformLocations: null,
                infLoc: null,
                nanLoc: null,
                inShapesLocations: null,
                inTexShapesLocations: null,
                outShapeLocation: null,
                outShapeStridesLocation: null,
                outTexShapeLocation: null
            };
        }
    }
    function getUniformLocations(gpgpu, program, webGLProgram) {
        const uniformLocations = {};
        const inShapesLocations = {};
        const inTexShapesLocations = {};
        const customUniformLocations = [];
        let outShapeLocation;
        let outTexShapeLocation;
        let outShapeStridesLocation;
        let infLoc = null;
        let nanLoc = null;
        // Add special uniforms (NAN, INFINITY)
        nanLoc = gpgpu.getUniformLocation(webGLProgram, 'NAN', false);
        if (env$1().getNumber('WEBGL_VERSION') === 1) {
            infLoc = gpgpu.getUniformLocation(webGLProgram, 'INFINITY', false);
        }
        // Add user-defined uniforms
        const shouldThrow = false;
        for (let i = 0; i < program.variableNames.length; i++) {
            const varName = program.variableNames[i];
            uniformLocations[varName] =
                gpgpu.getUniformLocation(webGLProgram, varName, shouldThrow);
            uniformLocations[`offset${varName}`] =
                gpgpu.getUniformLocation(webGLProgram, `offset${varName}`, shouldThrow);
            if (program.enableShapeUniforms) {
                inShapesLocations[`${varName}Shape`] = gpgpu.getUniformLocation(webGLProgram, `${varName}Shape`, shouldThrow);
                inTexShapesLocations[`${varName}TexShape`] = gpgpu.getUniformLocation(webGLProgram, `${varName}TexShape`, shouldThrow);
            }
        }
        if (program.enableShapeUniforms) {
            outShapeLocation =
                gpgpu.getUniformLocation(webGLProgram, 'outShape', shouldThrow);
            outShapeStridesLocation =
                gpgpu.getUniformLocation(webGLProgram, 'outShapeStrides', shouldThrow);
            outTexShapeLocation =
                gpgpu.getUniformLocation(webGLProgram, 'outTexShape', shouldThrow);
        }
        if (program.customUniforms) {
            program.customUniforms.forEach((d, i) => {
                customUniformLocations[i] =
                    gpgpu.getUniformLocation(webGLProgram, d.name, shouldThrow);
            });
        }
        return {
            uniformLocations,
            customUniformLocations,
            infLoc,
            nanLoc,
            inShapesLocations,
            inTexShapesLocations,
            outShapeLocation,
            outShapeStridesLocation,
            outTexShapeLocation
        };
    }
    function validateBinaryAndProgram(shapeInfos, inputs) {
        if (shapeInfos.length !== inputs.length) {
            throw Error(`Binary was compiled with ${shapeInfos.length} inputs, but ` +
                `was executed with ${inputs.length} inputs`);
        }
        shapeInfos.forEach((s, i) => {
            const shapeA = s.logicalShape;
            const input = inputs[i];
            const shapeB = input.shape;
            if (!arraysEqual(shapeA, shapeB)) {
                throw Error(`Binary was compiled with different shapes than ` +
                    `the current args. Shapes ${shapeA} and ${shapeB} must match`);
            }
            // The input is uploaded as uniform.
            if (s.isUniform && input.isUniform) {
                return;
            }
            const texShapeA = s.texShape;
            const texShapeB = input.isUniform ? null : input.texData.texShape;
            if (!arraysEqual(texShapeA, texShapeB)) {
                throw Error(`Binary was compiled with different texture shapes than the` +
                    ` current args. Shape ${texShapeA} and ${texShapeB} must match`);
            }
        });
    }
    function runProgram(gpgpu, binary, inputs, output, customUniformValues) {
        if (!binary.program.enableShapeUniforms) {
            validateBinaryAndProgram(binary.inShapeInfos, inputs);
            validateBinaryAndProgram([binary.outShapeInfo], [output]);
        }
        const outTex = output.texData.texture;
        const outTexShape = output.texData.texShape;
        if (output.texData.isPacked) {
            gpgpu.setOutputPackedMatrixTexture(outTex.texture, outTexShape[0], outTexShape[1]);
        }
        else {
            gpgpu.setOutputMatrixTexture(outTex.texture, outTexShape[0], outTexShape[1]);
        }
        gpgpu.setProgram(binary.webGLProgram);
        // Set special uniforms (NAN, INFINITY)
        if (env$1().getNumber('WEBGL_VERSION') === 1) {
            if (binary.infLoc !== null) {
                gpgpu.gl.uniform1f(binary.infLoc, Infinity);
            }
        }
        if (binary.nanLoc !== null) {
            gpgpu.gl.uniform1f(binary.nanLoc, NaN);
        }
        // Set user-defined inputs
        inputs.forEach((input, i) => {
            const varName = binary.program.variableNames[i];
            const varLoc = binary.uniformLocations[varName];
            const varOffsetLoc = binary.uniformLocations[`offset${varName}`];
            const varShapeLoc = binary.inShapesLocations[`${varName}Shape`];
            const varTexShapeLoc = binary.inTexShapesLocations[`${varName}TexShape`];
            if (varShapeLoc) {
                const { uniformShape } = getUniformInfoFromShape(binary.program.packedInputs, input.shape, input.texData.texShape);
                switch (uniformShape.length) {
                    case 1:
                        gpgpu.gl.uniform1iv(varShapeLoc, new Int32Array(uniformShape));
                        break;
                    case 2:
                        gpgpu.gl.uniform2iv(varShapeLoc, new Int32Array(uniformShape));
                        break;
                    case 3:
                        gpgpu.gl.uniform3iv(varShapeLoc, new Int32Array(uniformShape));
                        break;
                    case 4:
                        gpgpu.gl.uniform4iv(varShapeLoc, new Int32Array(uniformShape));
                        break;
                }
            }
            if (varTexShapeLoc) {
                gpgpu.gl.uniform2i(varTexShapeLoc, input.texData.texShape[0], input.texData.texShape[1]);
            }
            if (varLoc == null) {
                // The compiler inferred that this variable is not used in this shader.
                return;
            }
            if (input.isUniform) {
                // Upload the values of the tensor as uniform.
                if (sizeFromShape(input.shape) < 2) {
                    gpgpu.gl.uniform1f(varLoc, input.uniformValues[0]);
                }
                else {
                    let vals = input.uniformValues;
                    if (!(vals instanceof Float32Array)) {
                        vals = new Float32Array(vals);
                    }
                    gpgpu.gl.uniform1fv(varLoc, vals);
                }
                return;
            }
            // If the input was sliced, upload the flat offset index.
            if (input.texData.slice != null && varOffsetLoc != null) {
                gpgpu.gl.uniform1i(varOffsetLoc, input.texData.slice.flatOffset);
            }
            gpgpu.setInputMatrixTexture(input.texData.texture.texture, varLoc, i);
        });
        const outShapeLoc = binary.outShapeLocation;
        if (outShapeLoc) {
            switch (output.shape.length) {
                case 1:
                    gpgpu.gl.uniform1iv(outShapeLoc, new Int32Array(output.shape));
                    break;
                case 2:
                    gpgpu.gl.uniform2iv(outShapeLoc, new Int32Array(output.shape));
                    break;
                case 3:
                    gpgpu.gl.uniform3iv(outShapeLoc, new Int32Array(output.shape));
                    break;
                case 4:
                    gpgpu.gl.uniform4iv(outShapeLoc, new Int32Array(output.shape));
                    break;
            }
        }
        if (binary.outShapeStridesLocation) {
            const strides = computeStrides(output.shape);
            switch (output.shape.length) {
                case 2:
                    gpgpu.gl.uniform1iv(binary.outShapeStridesLocation, new Int32Array(strides));
                    break;
                case 3:
                    gpgpu.gl.uniform2iv(binary.outShapeStridesLocation, new Int32Array(strides));
                    break;
                case 4:
                    gpgpu.gl.uniform3iv(binary.outShapeStridesLocation, new Int32Array(strides));
                    break;
            }
        }
        if (binary.outTexShapeLocation) {
            gpgpu.gl.uniform2i(binary.outTexShapeLocation, output.texData.texShape[0], output.texData.texShape[1]);
        }
        if (binary.program.customUniforms && customUniformValues) {
            binary.program.customUniforms.forEach((d, i) => {
                const customLoc = binary.customUniformLocations[i];
                const customValue = customUniformValues[i];
                if (d.type === 'float') {
                    gpgpu.gl.uniform1fv(customLoc, customValue);
                }
                else if (d.type === 'vec2') {
                    gpgpu.gl.uniform2fv(customLoc, customValue);
                }
                else if (d.type === 'vec3') {
                    gpgpu.gl.uniform3fv(customLoc, customValue);
                }
                else if (d.type === 'vec4') {
                    gpgpu.gl.uniform4fv(customLoc, customValue);
                }
                else if (d.type === 'int') {
                    gpgpu.gl.uniform1iv(customLoc, customValue);
                }
                else if (d.type === 'ivec2') {
                    gpgpu.gl.uniform2iv(customLoc, customValue);
                }
                else if (d.type === 'ivec3') {
                    gpgpu.gl.uniform3iv(customLoc, customValue);
                }
                else if (d.type === 'ivec4') {
                    gpgpu.gl.uniform4iv(customLoc, customValue);
                }
                else {
                    throw Error(`uniform type ${d.type} is not supported yet.`);
                }
            });
        }
        gpgpu.executeProgram();
    }
    function makeShaderKey(program, inputs, output) {
        let keyInputs = '';
        inputs.concat(output).forEach(x => {
            const hasOffset = x.texData != null && x.texData.slice != null &&
                x.texData.slice.flatOffset > 0;
            // TODO: Remove the condition of !x.isUniform.
            if (program.enableShapeUniforms && !x.isUniform) {
                const xTexShape = x.texData.texShape;
                const { useSqueezeShape, uniformShape, keptDims } = getUniformInfoFromShape(program.packedInputs, x.shape, xTexShape);
                let rank1 = '', rank2 = '', rank34 = '';
                if (uniformShape.length === 1 && program.packedInputs) {
                    const packedTexShape = [Math.ceil(xTexShape[0] / 2), Math.ceil(xTexShape[1] / 2)];
                    rank1 = `${packedTexShape[0] > 1}_${packedTexShape[1] > 1}`;
                }
                else if (uniformShape.length === 2 && !program.packedInputs) {
                    rank2 = `${uniformShape[0] > 1}_${uniformShape[1] > 1}`;
                }
                else if (uniformShape.length > 2 && !program.packedInputs) {
                    const strides = computeStrides(uniformShape);
                    rank34 = `${strides[0] === xTexShape[1]}_${strides[strides.length - 1] === xTexShape[1]}`;
                }
                const xRank = x.shape.length;
                const isLogicalShapTexShapeEqual = uniformShape.length === 2 && arraysEqual(x.shape, xTexShape);
                const isScalar = sizeFromShape(x.shape) === 1;
                const broadcastDims = getBroadcastDims$1(x.shape, output.shape);
                const isInOutTexShapeEqual = !program.packedInputs &&
                    xRank === output.shape.length &&
                    arraysEqual(xTexShape, output.texData.texShape);
                const isTexShapeGreaterThanOne = program.packedInputs || uniformShape.length > 2 ?
                    '' :
                    `${xTexShape[0] > 1}_${xTexShape[1] > 1}`;
                // These key components are needed due to shader_compiler is embedding
                // them in the shader.
                // |xRank| is used to determine the coords length. See
                // get[Packed]SamplerAtOutputCoords.
                // |isInOutTexShapeEqual| is used to determine whether going to an
                // optimization path in getSamplerAtOutputCoords.
                // |useSqueezeShape| is extracted from squeezeInputInfo of
                // getSampler[2|3|4]D/getPackedSampler3D.
                // |isScalar| is extracted from isInputScalar/isOutputScalar in
                // getPackedSamplerAtOutputCoords.
                // |broadcastDims| is extracted from get[Packed]SamplerAtOutputCoords.
                // |isLogicalShapTexShapeEqual| is used in
                // getOutput[Packed]2DCoords/get[Packed]Sampler2D.
                // |rank1| is used in getOutputPacked1DCoords.
                // |rank2| is used in getOutput2DCoords.
                // |rank34| is used in getSampler3D/getSampler4D.
                // |isTexShapeGreaterThanOne| are used in
                // getSampler[Scalar|1D|2D]/getOutput1DCoords.
                keyInputs += `${xRank}_${isInOutTexShapeEqual}_${useSqueezeShape ? keptDims : ''}_${uniformShape.length}_${isScalar}_${broadcastDims}_${isLogicalShapTexShapeEqual}_${rank1}_${rank2}_${rank34}_${isTexShapeGreaterThanOne}_${hasOffset}`;
            }
            else {
                const texShape = x.isUniform ? 'uniform' : x.texData.texShape;
                keyInputs += `${x.shape}_${texShape}_${hasOffset}`;
            }
        });
        const keyUserCode = program.userCode;
        let key = program.constructor.name;
        // Fast string concat. See https://jsperf.com/string-concatenation/14.
        key += '_' + keyInputs + '_' + keyUserCode +
            `${env$1().getNumber('WEBGL_VERSION')}`;
        return key;
    }
    function useShapeUniforms(rank) {
        // TODO: Remove the limitaion of rank <= 4.
        return env$1().getBool('WEBGL_USE_SHAPES_UNIFORMS') && rank <= 4;
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class DecodeMatrixProgram {
        constructor(outputShape) {
            this.variableNames = ['A'];
            this.packedInputs = false;
            this.packedOutput = true;
            this.outPackingScheme = PackingScheme.DENSE;
            this.customUniforms = [{ name: 'texShape', type: 'ivec2' }];
            const glsl = getGlslDifferences();
            this.outputShape = outputShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            this.userCode = `
      ivec3 outCoordsFromFlatIndex(int index) {
        ${this.enableShapeUniforms ?
            getOutputLogicalCoordinatesFromFlatIndexByUniform(['r', 'c', 'd'], outputShape) :
            getLogicalCoordinatesFromFlatIndex(['r', 'c', 'd'], outputShape)}
        return ivec3(r, c, d);
      }

      void main() {
        ivec2 resTexRC = ivec2(resultUV.yx * vec2(texShape[0], texShape[1]));
        int index = 4 * (resTexRC.x * texShape[1] + resTexRC.y);

        vec4 result = vec4(0.);

        for (int i=0; i<4; i++) {
          int flatIndex = index + i;
          ivec3 rc = outCoordsFromFlatIndex(flatIndex);
          result[i] = getA(rc.x, rc.y, rc.z);
        }

        ${glsl.output} = result;
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class DecodeMatrixPackedProgram {
        constructor(outputShape) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.outPackingScheme = PackingScheme.DENSE;
            this.customUniforms = [{ name: 'texShape', type: 'ivec2' }];
            const glsl = getGlslDifferences();
            this.outputShape = outputShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            this.userCode = `
      ivec3 outCoordsFromFlatIndex(int index) {
        ${this.enableShapeUniforms ?
            getOutputLogicalCoordinatesFromFlatIndexByUniform(['r', 'c', 'd'], outputShape) :
            getLogicalCoordinatesFromFlatIndex(['r', 'c', 'd'], outputShape)}
        return ivec3(r, c, d);
      }

      void main() {
        ivec2 resTexRC = ivec2(resultUV.yx * vec2(texShape[0], texShape[1]));
        int index = 4 * (resTexRC.x * texShape[1] + resTexRC.y);

        vec4 result = vec4(0.);

        for (int i=0; i<4; i++) {
          int flatIndex = index + i;
          ivec3 rc = outCoordsFromFlatIndex(flatIndex);
          result[i] = getChannel(getA(rc.x, rc.y, rc.z), vec2(rc.y, rc.z));
        }

        ${glsl.output} = result;
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class EncodeFloatProgram {
        constructor(outputShape) {
            this.variableNames = ['A'];
            this.outTexUsage = TextureUsage.DOWNLOAD;
            const glsl = getGlslDifferences();
            this.outputShape = outputShape;
            this.userCode = `
      ${ENCODE_FLOAT_SNIPPET}

      void main() {
        float x = getAAtOutCoords();
        ${glsl.output} = encode_float(x);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class EncodeFloatPackedProgram {
        constructor(outputShape) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = false;
            this.outTexUsage = TextureUsage.DOWNLOAD;
            const glsl = getGlslDifferences();
            this.outputShape = outputShape;
            this.userCode = `
      ${ENCODE_FLOAT_SNIPPET}

      void main() {
        ivec3 coords = getOutputCoords();
        float x = getChannel(getAAtOutCoords(), vec2(coords.y, coords.z));
        ${glsl.output} = encode_float(x);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class EncodeMatrixProgram {
        constructor(outputShape, inputIsUnsignedByte = false) {
            this.variableNames = ['A'];
            this.customUniforms = [{ name: 'texShape', type: 'ivec2' }];
            const glsl = getGlslDifferences();
            this.outputShape = outputShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            let output = `result`;
            if (inputIsUnsignedByte) {
                output = `floor(result * 255. + 0.5)`;
            }
            this.userCode = `
      ${this.enableShapeUniforms ? getFlatIndexFrom3DOutput() :
            getFlatIndexFrom3D(outputShape)}

      void main() {
        ivec3 coords = getOutputCoords();

        int flatIndex = getFlatIndex(coords);
        int offset = imod(flatIndex, 4);

        flatIndex = idiv(flatIndex, 4, 1.);

        int r = flatIndex / texShape[1];
        int c = imod(flatIndex, texShape[1]);
        vec2 uv = (vec2(c, r) + halfCR) / vec2(texShape[1], texShape[0]);
        vec4 values = ${glsl.texture2D}(A, uv);

        float result;

        if(offset == 0) {
          result = values[0];
        } else if(offset == 1) {
          result = values[1];
        } else if(offset == 2) {
          result = values[2];
        } else {
          result = values[3];
        }

        ${glsl.output} = vec4(${output}, 0., 0., 0.);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /*
    This is how the shader encodes a tensor with shape = [2, 3, 5]
    (indices are [batch, row, col]).

    000|001   002|003   004|xxx   020|021   022|023   024|xxx
    -------   -------   -------   -------   -------   -------
    010|011   012|013   014|xxx   xxx|xxx   xxx|xxx   xxx|xxx

    100|101   102|103   104|xxx   120|121   122|123   124|xxx
    -------   -------   -------   -------   -------   -------
    110|111   112|113   114|xxx   xxx|xxx   xxx|xxx   xxx|xxx

    Single texels contain only values from the same batch, and from adjacent rows
    and columns.
     */
    class EncodeMatrixPackedProgram {
        constructor(outputShape, inputIsUnsignedByte = false) {
            this.variableNames = ['A'];
            this.packedInputs = false;
            this.packedOutput = true;
            this.customUniforms = [{ name: 'texShape', type: 'ivec2' }];
            const glsl = getGlslDifferences();
            this.outputShape = outputShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            let mainLoop = '';
            let output = 'result';
            if (inputIsUnsignedByte) {
                output = 'floor(result * 255. + 0.5)';
            }
            for (let row = 0; row <= 1; row++) {
                for (let col = 0; col <= 1; col++) {
                    const channel = row * 2 + col;
                    mainLoop += `
          localCoords = coords;
          if(localCoords[2] + ${col} < ${this.enableShapeUniforms ? 'outShape[2]' : `${outputShape[2]}`}) {
          localCoords[2] += ${col};
          if (localCoords[1] + ${row} < ${this.enableShapeUniforms ? 'outShape[1]' : `${outputShape[1]}`}) {
            localCoords[1] += ${row};

            flatIndex = getFlatIndex(localCoords);
            offset = imod(flatIndex, 4);

            flatIndex = idiv(flatIndex, 4, 1.);

            int r = flatIndex / texShape[1];
            int c = imod(flatIndex, texShape[1]);
            vec2 uv = (vec2(c, r) + halfCR) / vec2(texShape[1], texShape[0]);
            values = ${glsl.texture2D}(A, uv);

            if (offset == 0) {
              result[${channel}] = values[0];
            } else if (offset == 1) {
              result[${channel}] = values[1];
            } else if (offset == 2) {
              result[${channel}] = values[2];
            } else {
              result[${channel}] = values[3];
            }
          }
        }
        `;
                }
            }
            this.userCode = `
        ${this.enableShapeUniforms ? getFlatIndexFrom3DOutput() :
            getFlatIndexFrom3D(outputShape)}

        void main() {
          ivec3 coords = getOutputCoords();

          vec4 result = vec4(0.);
          int flatIndex, r, c, offset;
          ivec3 localCoords;
          vec2 uv;
          vec4 values;

          ${mainLoop}

          ${glsl.output} = ${output};
        }
    `;
        }
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function createVertexShader(gl) {
        const glsl = getGlslDifferences();
        const vertexShaderSource = `${glsl.version}
    precision highp float;
    ${glsl.attribute} vec3 clipSpacePos;
    ${glsl.attribute} vec2 uv;
    ${glsl.varyingVs} vec2 resultUV;

    void main() {
      gl_Position = vec4(clipSpacePos, 1);
      resultUV = uv;
    }`;
        return createVertexShader$1(gl, vertexShaderSource);
    }
    function createVertexBuffer(gl) {
        // [x y z u v] * [upper-left, lower-left, upper-right, lower-right]
        const vertexArray = new Float32Array([-1, 1, 0, 0, 1, -1, -1, 0, 0, 0, 1, 1, 0, 1, 1, 1, -1, 0, 1, 0]);
        return createStaticVertexBuffer(gl, vertexArray);
    }
    function createIndexBuffer(gl) {
        // OpenGL (and WebGL) have "CCW == front" winding
        const triangleVertexIndices = new Uint16Array([0, 1, 2, 2, 1, 3]);
        return createStaticIndexBuffer(gl, triangleVertexIndices);
    }
    function createAndConfigureTexture(gl, width, height, internalFormat, textureFormat, textureType) {
        validateTextureSize(width, height);
        const texture = createTexture(gl);
        const tex2d = gl.TEXTURE_2D;
        callAndCheck(gl, () => gl.bindTexture(tex2d, texture));
        callAndCheck(gl, () => gl.texParameteri(tex2d, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE));
        callAndCheck(gl, () => gl.texParameteri(tex2d, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE));
        callAndCheck(gl, () => gl.texParameteri(tex2d, gl.TEXTURE_MIN_FILTER, gl.NEAREST));
        callAndCheck(gl, () => gl.texParameteri(tex2d, gl.TEXTURE_MAG_FILTER, gl.NEAREST));
        if (env$1().getNumber('WEBGL_VERSION') === 1) {
            callAndCheck(gl, () => gl.texImage2D(tex2d, 0, internalFormat, width, height, 0, textureFormat, textureType, null));
        }
        else {
            callAndCheck(gl, () => gl
                .texStorage2D(tex2d, 1, internalFormat, width, height));
        }
        callAndCheck(gl, () => gl.bindTexture(gl.TEXTURE_2D, null));
        return { texture, texShape: [height, width] };
    }
    function getInternalFormatForFloat32MatrixTexture(textureConfig) {
        return textureConfig.internalFormatFloat;
    }
    function createFloat32MatrixTexture(gl, rows, columns, textureConfig) {
        const [width, height] = getUnpackedMatrixTextureShapeWidthHeight(rows, columns);
        return createAndConfigureTexture(gl, width, height, getInternalFormatForFloat32MatrixTexture(textureConfig), textureConfig.textureFormatFloat, gl.FLOAT);
    }
    function getInternalFormatForFloat16MatrixTexture(textureConfig) {
        return textureConfig.internalFormatHalfFloat;
    }
    function createFloat16MatrixTexture(gl, rows, columns, textureConfig) {
        const [width, height] = getUnpackedMatrixTextureShapeWidthHeight(rows, columns);
        return createAndConfigureTexture(gl, width, height, getInternalFormatForFloat16MatrixTexture(textureConfig), textureConfig.textureFormatFloat, textureConfig.textureTypeHalfFloat);
    }
    function getInternalFormatForUnsignedBytesMatrixTexture(textureConfig) {
        return textureConfig.downloadTextureFormat;
    }
    function createUnsignedBytesMatrixTexture(gl, rows, columns, textureConfig) {
        const [width, height] = getUnpackedMatrixTextureShapeWidthHeight(rows, columns);
        return createAndConfigureTexture(gl, width, height, getInternalFormatForUnsignedBytesMatrixTexture(textureConfig), gl.RGBA, gl.UNSIGNED_BYTE);
    }
    function getInternalFormatForPackedMatrixTexture(textureConfig) {
        return textureConfig.internalFormatPackedFloat;
    }
    function createPackedMatrixTexture(gl, rows, columns, textureConfig) {
        const [width, height] = getPackedMatrixTextureShapeWidthHeight(rows, columns);
        return createAndConfigureTexture(gl, width, height, getInternalFormatForPackedMatrixTexture(textureConfig), gl.RGBA, gl.FLOAT);
    }
    function getInternalFormatForFloat16PackedMatrixTexture(textureConfig) {
        return textureConfig.internalFormatPackedHalfFloat;
    }
    function createFloat16PackedMatrixTexture(gl, rows, columns, textureConfig) {
        const [width, height] = getPackedMatrixTextureShapeWidthHeight(rows, columns);
        return createAndConfigureTexture(gl, width, height, getInternalFormatForFloat16PackedMatrixTexture(textureConfig), gl.RGBA, textureConfig.textureTypeHalfFloat);
    }
    function bindVertexProgramAttributeStreams(gl, program, vertexBuffer) {
        const posOffset = 0; // x is the first buffer element
        const uvOffset = 3 * 4; // uv comes after [x y z]
        const stride = (3 * 4) + (2 * 4); // xyz + uv, each entry is 4-byte float.
        callAndCheck(gl, () => gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer));
        const success = bindVertexBufferToProgramAttribute(gl, program, 'clipSpacePos', vertexBuffer, 3, stride, posOffset);
        return success &&
            bindVertexBufferToProgramAttribute(gl, program, 'uv', vertexBuffer, 2, stride, uvOffset);
    }
    function uploadDenseMatrixToTexture(gl, texture, width, height, data, textureConfig) {
        callAndCheck(gl, () => gl.bindTexture(gl.TEXTURE_2D, texture));
        let dataForUpload, texelDataType, internalFormat;
        if (data instanceof Uint8Array) {
            dataForUpload = new Uint8Array(width * height * 4);
            texelDataType = gl.UNSIGNED_BYTE;
            internalFormat = gl.RGBA;
        }
        else {
            dataForUpload = new Float32Array(width * height * 4);
            texelDataType = gl.FLOAT;
            internalFormat = textureConfig.internalFormatPackedFloat;
        }
        dataForUpload.set(data);
        if (env$1().getNumber('WEBGL_VERSION') === 2) {
            callAndCheck(gl, () => gl.texSubImage2D(gl.TEXTURE_2D, 0, 0, 0, width, height, gl.RGBA, texelDataType, dataForUpload));
        }
        else {
            callAndCheck(gl, () => gl.texImage2D(gl.TEXTURE_2D, 0, internalFormat, width, height, 0, gl.RGBA, texelDataType, dataForUpload));
        }
        callAndCheck(gl, () => gl.bindTexture(gl.TEXTURE_2D, null));
    }
    function uploadPixelDataToTexture(gl, texture, pixels) {
        callAndCheck(gl, () => gl.bindTexture(gl.TEXTURE_2D, texture));
        if (pixels.data instanceof Uint8Array) {
            if (env$1().getNumber('WEBGL_VERSION') === 2) {
                callAndCheck(gl, () => gl.texSubImage2D(gl.TEXTURE_2D, 0, 0, 0, pixels.width, pixels.height, gl.RGBA, gl.UNSIGNED_BYTE, pixels.data));
            }
            else {
                callAndCheck(gl, () => gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, pixels.width, pixels.height, 0, gl.RGBA, gl.UNSIGNED_BYTE, pixels.data));
            }
        }
        else {
            if (env$1().getNumber('WEBGL_VERSION') === 2) {
                callAndCheck(gl, () => gl.texSubImage2D(gl.TEXTURE_2D, 0, 0, 0, gl.RGBA, gl.UNSIGNED_BYTE, pixels));
            }
            else {
                callAndCheck(gl, () => gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, pixels));
            }
        }
        callAndCheck(gl, () => gl.bindTexture(gl.TEXTURE_2D, null));
    }
    function createBufferFromOutputTexture(gl2, rows, columns, textureConfig) {
        // Create and bind the buffer.
        const buffer = gl2.createBuffer();
        callAndCheck(gl2, () => gl2.bindBuffer(gl2.PIXEL_PACK_BUFFER, buffer));
        // Initialize the buffer to the size of the texture in bytes.
        const bytesPerFloat = 4;
        const valuesPerTexel = 4;
        const bufferSizeBytes = bytesPerFloat * valuesPerTexel * rows * columns;
        callAndCheck(gl2, () => gl2.bufferData(gl2.PIXEL_PACK_BUFFER, bufferSizeBytes, gl2.STREAM_READ));
        // Enqueue a command on the GPU command queue to copy of texture into the
        // buffer.
        callAndCheck(gl2, () => gl2.readPixels(0, 0, columns, rows, gl2.RGBA, gl2.FLOAT, 0));
        callAndCheck(gl2, () => gl2.bindBuffer(gl2.PIXEL_PACK_BUFFER, null));
        return buffer;
    }
    function downloadFloat32MatrixFromBuffer(gl, buffer, size) {
        const gl2 = gl;
        const downloadTarget = new Float32Array(size);
        gl2.bindBuffer(gl2.PIXEL_PACK_BUFFER, buffer);
        gl2.getBufferSubData(gl2.PIXEL_PACK_BUFFER, 0, downloadTarget);
        gl2.bindBuffer(gl2.PIXEL_PACK_BUFFER, null);
        return downloadTarget;
    }
    function downloadByteEncodedFloatMatrixFromOutputTexture(gl, rows, columns, textureConfig) {
        const [w, h] = getUnpackedMatrixTextureShapeWidthHeight(rows, columns);
        const numChannels = 4;
        const downloadTarget = new Uint8Array(getUnpackedArraySizeFromMatrixSize(rows * columns, numChannels));
        callAndCheck(gl, () => gl.readPixels(0, 0, w, h, textureConfig.downloadTextureFormat, gl.UNSIGNED_BYTE, downloadTarget));
        // By wrapping the buffer in a Float32Array, we use native browser IEEE 754
        // decoding of the 4 bytes that back each 32 bit float.
        return new Float32Array(downloadTarget.buffer);
    }
    function downloadPackedMatrixFromBuffer(gl, buffer, batch, rows, cols, physicalRows, physicalCols, textureConfig) {
        const gl2 = gl;
        const downloadTarget = new Float32Array(getPackedRGBAArraySizeFromMatrixShape(physicalRows, physicalCols));
        gl2.bindBuffer(gl2.PIXEL_PACK_BUFFER, buffer);
        gl2.getBufferSubData(gl2.PIXEL_PACK_BUFFER, 0, downloadTarget);
        gl2.bindBuffer(gl2.PIXEL_PACK_BUFFER, null);
        return downloadTarget;
    }
    function downloadMatrixFromPackedOutputTexture(gl, physicalRows, physicalCols) {
        const packedRGBA = new Float32Array(physicalRows * physicalCols * 4);
        callAndCheck(gl, () => gl.readPixels(0, 0, physicalCols, physicalRows, gl.RGBA, gl.FLOAT, packedRGBA));
        return packedRGBA;
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class GPGPUContext {
        constructor(gl) {
            this.outputTexture = null;
            this.program = null;
            this.disposed = false;
            this.vertexAttrsAreBound = false;
            this.itemsToPoll = [];
            const glVersion = env$1().getNumber('WEBGL_VERSION');
            if (gl != null) {
                this.gl = gl;
                setWebGLContext(glVersion, gl);
            }
            else {
                this.gl = getWebGLContext(glVersion);
            }
            // WebGL 2.0 enables texture floats without an extension.
            let COLOR_BUFFER_FLOAT = 'WEBGL_color_buffer_float';
            const COLOR_BUFFER_HALF_FLOAT = 'EXT_color_buffer_half_float';
            this.parallelCompilationExtension =
                this.gl.getExtension('KHR_parallel_shader_compile');
            if (env$1().getNumber('WEBGL_VERSION') === 1) {
                const TEXTURE_FLOAT = 'OES_texture_float';
                const TEXTURE_HALF_FLOAT = 'OES_texture_half_float';
                this.textureFloatExtension =
                    getExtensionOrThrow(this.gl, TEXTURE_FLOAT);
                if (hasExtension(this.gl, TEXTURE_HALF_FLOAT)) {
                    this.textureHalfFloatExtension =
                        getExtensionOrThrow(this.gl, TEXTURE_HALF_FLOAT);
                }
                else if (env$1().get('WEBGL_FORCE_F16_TEXTURES')) {
                    throw new Error('GL context does not support half float textures, yet the ' +
                        'environment flag WEBGL_FORCE_F16_TEXTURES is set to true.');
                }
                this.colorBufferFloatExtension = this.gl.getExtension(COLOR_BUFFER_FLOAT);
                if (hasExtension(this.gl, COLOR_BUFFER_HALF_FLOAT)) {
                    this.colorBufferHalfFloatExtension =
                        getExtensionOrThrow(this.gl, COLOR_BUFFER_HALF_FLOAT);
                }
                else if (env$1().get('WEBGL_FORCE_F16_TEXTURES')) {
                    throw new Error('GL context does not support color renderable half floats, yet ' +
                        'the environment flag WEBGL_FORCE_F16_TEXTURES is set to true.');
                }
            }
            else {
                COLOR_BUFFER_FLOAT = 'EXT_color_buffer_float';
                if (hasExtension(this.gl, COLOR_BUFFER_FLOAT)) {
                    this.colorBufferFloatExtension =
                        this.gl.getExtension(COLOR_BUFFER_FLOAT);
                }
                else if (hasExtension(this.gl, COLOR_BUFFER_HALF_FLOAT)) {
                    this.colorBufferHalfFloatExtension =
                        this.gl.getExtension(COLOR_BUFFER_HALF_FLOAT);
                }
                else {
                    throw new Error('GL context does not support color renderable floats');
                }
            }
            this.vertexBuffer = createVertexBuffer(this.gl);
            this.indexBuffer = createIndexBuffer(this.gl);
            this.framebuffer = createFramebuffer(this.gl);
            this.textureConfig =
                getTextureConfig(this.gl, this.textureHalfFloatExtension);
        }
        get debug() {
            return env$1().getBool('DEBUG');
        }
        dispose() {
            if (this.disposed) {
                return;
            }
            if (this.program != null) {
                console.warn('Disposing a GPGPUContext that still has a bound WebGLProgram.' +
                    ' This is probably a resource leak, delete the program with ' +
                    'GPGPUContext.deleteProgram before disposing.');
            }
            if (this.outputTexture != null) {
                console.warn('Disposing a GPGPUContext that still has a bound output matrix ' +
                    'texture.  This is probably a resource leak, delete the output ' +
                    'matrix texture with GPGPUContext.deleteMatrixTexture before ' +
                    'disposing.');
            }
            const gl = this.gl;
            callAndCheck(gl, () => gl.finish());
            callAndCheck(gl, () => gl.bindFramebuffer(gl.FRAMEBUFFER, null));
            callAndCheck(gl, () => gl.deleteFramebuffer(this.framebuffer));
            callAndCheck(gl, () => gl.bindBuffer(gl.ARRAY_BUFFER, null));
            callAndCheck(gl, () => gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, null));
            callAndCheck(gl, () => gl.deleteBuffer(this.indexBuffer));
            this.disposed = true;
        }
        createFloat32MatrixTexture(rows, columns) {
            this.throwIfDisposed();
            return createFloat32MatrixTexture(this.gl, rows, columns, this.textureConfig);
        }
        createFloat16MatrixTexture(rows, columns) {
            this.throwIfDisposed();
            return createFloat16MatrixTexture(this.gl, rows, columns, this.textureConfig);
        }
        createUnsignedBytesMatrixTexture(rows, columns) {
            this.throwIfDisposed();
            return createUnsignedBytesMatrixTexture(this.gl, rows, columns, this.textureConfig);
        }
        uploadPixelDataToTexture(texture, pixels) {
            this.throwIfDisposed();
            uploadPixelDataToTexture(this.gl, texture, pixels);
        }
        uploadDenseMatrixToTexture(texture, width, height, data) {
            this.throwIfDisposed();
            uploadDenseMatrixToTexture(this.gl, texture, width, height, data, this.textureConfig);
        }
        createFloat16PackedMatrixTexture(rows, columns) {
            this.throwIfDisposed();
            return createFloat16PackedMatrixTexture(this.gl, rows, columns, this.textureConfig);
        }
        createPackedMatrixTexture(rows, columns) {
            this.throwIfDisposed();
            return createPackedMatrixTexture(this.gl, rows, columns, this.textureConfig);
        }
        deleteMatrixTexture(texture) {
            this.throwIfDisposed();
            if (this.outputTexture === texture) {
                unbindColorTextureFromFramebuffer(this.gl, this.framebuffer);
                this.outputTexture = null;
            }
            callAndCheck(this.gl, () => this.gl.deleteTexture(texture));
        }
        downloadByteEncodedFloatMatrixFromOutputTexture(texture, rows, columns) {
            return this.downloadMatrixDriver(texture, () => downloadByteEncodedFloatMatrixFromOutputTexture(this.gl, rows, columns, this.textureConfig));
        }
        downloadPackedMatrixFromBuffer(buffer, batch, rows, columns, physicalRows, physicalCols) {
            return downloadPackedMatrixFromBuffer(this.gl, buffer, batch, rows, columns, physicalRows, physicalCols, this.textureConfig);
        }
        downloadFloat32MatrixFromBuffer(buffer, size) {
            return downloadFloat32MatrixFromBuffer(this.gl, buffer, size);
        }
        createBufferFromTexture(texture, rows, columns) {
            this.bindTextureToFrameBuffer(texture);
            const result = createBufferFromOutputTexture(this.gl, rows, columns, this.textureConfig);
            this.unbindTextureToFrameBuffer();
            return result;
        }
        createAndWaitForFence() {
            const fenceContext = this.createFence(this.gl);
            return this.pollFence(fenceContext);
        }
        createFence(gl) {
            let query;
            let isFencePassed;
            if (env$1().getBool('WEBGL_FENCE_API_ENABLED')) {
                const gl2 = gl;
                const sync = gl2.fenceSync(gl2.SYNC_GPU_COMMANDS_COMPLETE, 0);
                gl.flush();
                isFencePassed = () => {
                    const status = gl2.clientWaitSync(sync, 0, 0);
                    return status === gl2.ALREADY_SIGNALED ||
                        status === gl2.CONDITION_SATISFIED;
                };
                query = sync;
            }
            else if (env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_VERSION') > 0) {
                query = this.beginQuery();
                this.endQuery();
                isFencePassed = () => this.isQueryAvailable(query, env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_VERSION'));
            }
            else {
                // If we have no way to fence, return true immediately. This will fire in
                // WebGL 1.0 when there is no disjoint query timer. In this case, because
                // the fence passes immediately, we'll immediately ask for a download of
                // the texture, which will cause the UI thread to hang.
                isFencePassed = () => true;
            }
            return { query, isFencePassed };
        }
        downloadMatrixFromPackedTexture(texture, physicalRows, physicalCols) {
            return this.downloadMatrixDriver(texture, () => downloadMatrixFromPackedOutputTexture(this.gl, physicalRows, physicalCols));
        }
        createProgram(fragmentShader) {
            this.throwIfDisposed();
            const gl = this.gl;
            if (this.vertexShader == null) {
                this.vertexShader = createVertexShader(gl);
            }
            const program = createProgram(gl);
            callAndCheck(gl, () => gl.attachShader(program, this.vertexShader));
            callAndCheck(gl, () => gl.attachShader(program, fragmentShader));
            linkProgram(gl, program);
            if (this.debug) {
                validateProgram(gl, program);
            }
            if (!this.vertexAttrsAreBound) {
                this.setProgram(program);
                this.vertexAttrsAreBound = bindVertexProgramAttributeStreams(gl, this.program, this.vertexBuffer);
            }
            return program;
        }
        deleteProgram(program) {
            this.throwIfDisposed();
            if (program === this.program) {
                this.program = null;
            }
            if (program != null) {
                callAndCheck(this.gl, () => this.gl.deleteProgram(program));
            }
        }
        setProgram(program) {
            this.throwIfDisposed();
            this.program = program;
            if ((this.program != null) && this.debug) {
                validateProgram(this.gl, this.program);
            }
            callAndCheck(this.gl, () => this.gl.useProgram(program));
        }
        getUniformLocation(program, uniformName, shouldThrow = true) {
            this.throwIfDisposed();
            if (shouldThrow) {
                return getProgramUniformLocationOrThrow(this.gl, program, uniformName);
            }
            else {
                return getProgramUniformLocation(this.gl, program, uniformName);
            }
        }
        getAttributeLocation(program, attribute) {
            this.throwIfDisposed();
            return callAndCheck(this.gl, () => this.gl.getAttribLocation(program, attribute));
        }
        getUniformLocationNoThrow(program, uniformName) {
            this.throwIfDisposed();
            return this.gl.getUniformLocation(program, uniformName);
        }
        setInputMatrixTexture(inputMatrixTexture, uniformLocation, textureUnit) {
            this.throwIfDisposed();
            this.throwIfNoProgram();
            bindTextureToProgramUniformSampler(this.gl, inputMatrixTexture, uniformLocation, textureUnit);
        }
        setOutputMatrixTexture(outputMatrixTexture, rows, columns) {
            this.setOutputMatrixTextureDriver(outputMatrixTexture, columns, rows);
        }
        setOutputPackedMatrixTexture(outputPackedMatrixTexture, rows, columns) {
            this.throwIfDisposed();
            const [width, height] = getPackedMatrixTextureShapeWidthHeight(rows, columns);
            this.setOutputMatrixTextureDriver(outputPackedMatrixTexture, width, height);
        }
        setOutputMatrixWriteRegion(startRow, numRows, startColumn, numColumns) {
            this.setOutputMatrixWriteRegionDriver(startColumn, startRow, numColumns, numRows);
        }
        setOutputPackedMatrixWriteRegion(startRow, numRows, startColumn, numColumns) {
            throw new Error('setOutputPackedMatrixWriteRegion not implemented.');
        }
        debugValidate() {
            if (this.program != null) {
                validateProgram(this.gl, this.program);
            }
            validateFramebuffer(this.gl);
        }
        executeProgram() {
            this.throwIfDisposed();
            this.throwIfNoProgram();
            const gl = this.gl;
            if (this.debug) {
                this.debugValidate();
            }
            callAndCheck(gl, () => gl.drawElements(gl.TRIANGLES, 6, gl.UNSIGNED_SHORT, 0));
        }
        blockUntilAllProgramsCompleted() {
            this.throwIfDisposed();
            callAndCheck(this.gl, () => this.gl.finish());
        }
        getQueryTimerExtension() {
            if (this.disjointQueryTimerExtension == null) {
                this.disjointQueryTimerExtension =
                    getExtensionOrThrow(this.gl, env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_VERSION') === 2 ?
                        'EXT_disjoint_timer_query_webgl2' :
                        'EXT_disjoint_timer_query');
            }
            return this.disjointQueryTimerExtension;
        }
        getQueryTimerExtensionWebGL2() {
            return this.getQueryTimerExtension();
        }
        getQueryTimerExtensionWebGL1() {
            return this.getQueryTimerExtension();
        }
        beginQuery() {
            if (env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_VERSION') === 2) {
                const gl2 = this.gl;
                const ext = this.getQueryTimerExtensionWebGL2();
                const query = gl2.createQuery();
                gl2.beginQuery(ext.TIME_ELAPSED_EXT, query);
                return query;
            }
            const ext = this.getQueryTimerExtensionWebGL1();
            const query = ext.createQueryEXT();
            ext.beginQueryEXT(ext.TIME_ELAPSED_EXT, query);
            return query;
        }
        endQuery() {
            if (env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_VERSION') === 2) {
                const gl2 = this.gl;
                const ext = this.getQueryTimerExtensionWebGL2();
                gl2.endQuery(ext.TIME_ELAPSED_EXT);
                return;
            }
            const ext = this.getQueryTimerExtensionWebGL1();
            ext.endQueryEXT(ext.TIME_ELAPSED_EXT);
        }
        async waitForQueryAndGetTime(query) {
            await repeatedTry(() => this.disposed || // while testing contexts are created / disposed
                // in rapid succession, so without this check we
                // may poll for the query timer indefinitely
                this.isQueryAvailable(query, env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_VERSION')));
            return this.getQueryTime(query, env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_VERSION'));
        }
        getQueryTime(query, queryTimerVersion) {
            if (queryTimerVersion === 0) {
                return null;
            }
            if (queryTimerVersion === 2) {
                const gl2 = this.gl;
                const timeElapsedNanos = gl2.getQueryParameter(query, gl2.QUERY_RESULT);
                // Return milliseconds.
                return timeElapsedNanos / 1000000;
            }
            else {
                const ext = this.getQueryTimerExtensionWebGL1();
                const timeElapsedNanos = ext.getQueryObjectEXT(query, ext.QUERY_RESULT_EXT);
                // Return milliseconds.
                return timeElapsedNanos / 1000000;
            }
        }
        isQueryAvailable(query, queryTimerVersion) {
            if (queryTimerVersion === 0) {
                return true;
            }
            if (queryTimerVersion === 2) {
                const gl2 = this.gl;
                const ext = this.getQueryTimerExtensionWebGL2();
                const available = gl2.getQueryParameter(query, gl2.QUERY_RESULT_AVAILABLE);
                if (this.disjoint == null) {
                    this.disjoint = this.gl.getParameter(ext.GPU_DISJOINT_EXT);
                }
                return available && !this.disjoint;
            }
            else {
                const ext = this.getQueryTimerExtensionWebGL1();
                const available = ext.getQueryObjectEXT(query, ext.QUERY_RESULT_AVAILABLE_EXT);
                if (this.disjoint == null) {
                    this.disjoint = this.gl.getParameter(ext.GPU_DISJOINT_EXT);
                }
                return available && !this.disjoint;
            }
        }
        pollFence(fenceContext) {
            return new Promise(resolve => {
                this.addItemToPoll(() => fenceContext.isFencePassed(), () => resolve());
            });
        }
        pollItems() {
            // Find the last query that has finished.
            const index = linearSearchLastTrue(this.itemsToPoll.map(x => x.isDoneFn));
            for (let i = 0; i <= index; ++i) {
                const { resolveFn } = this.itemsToPoll[i];
                resolveFn();
            }
            this.itemsToPoll = this.itemsToPoll.slice(index + 1);
        }
        addItemToPoll(isDoneFn, resolveFn) {
            this.itemsToPoll.push({ isDoneFn, resolveFn });
            if (this.itemsToPoll.length > 1) {
                // We already have a running loop that polls.
                return;
            }
            // Start a new loop that polls.
            repeatedTry(() => {
                this.pollItems();
                // End the loop if no more items to poll.
                return this.itemsToPoll.length === 0;
            });
        }
        bindTextureToFrameBuffer(texture) {
            this.throwIfDisposed();
            bindColorTextureToFramebuffer(this.gl, texture, this.framebuffer);
            if (this.debug) {
                validateFramebuffer(this.gl);
            }
        }
        unbindTextureToFrameBuffer() {
            if (this.outputTexture != null) {
                bindColorTextureToFramebuffer(this.gl, this.outputTexture, this.framebuffer);
                if (this.debug) {
                    validateFramebuffer(this.gl);
                }
            }
            else {
                unbindColorTextureFromFramebuffer(this.gl, this.framebuffer);
            }
        }
        downloadMatrixDriver(texture, downloadAndDecode) {
            this.bindTextureToFrameBuffer(texture);
            const result = downloadAndDecode();
            this.unbindTextureToFrameBuffer();
            return result;
        }
        setOutputMatrixTextureDriver(outputMatrixTextureMaybePacked, width, height) {
            this.throwIfDisposed();
            const gl = this.gl;
            bindColorTextureToFramebuffer(gl, outputMatrixTextureMaybePacked, this.framebuffer);
            if (this.debug) {
                validateFramebuffer(gl);
            }
            this.outputTexture = outputMatrixTextureMaybePacked;
            callAndCheck(gl, () => gl.viewport(0, 0, width, height));
            callAndCheck(gl, () => gl.scissor(0, 0, width, height));
        }
        setOutputMatrixWriteRegionDriver(x, y, width, height) {
            this.throwIfDisposed();
            callAndCheck(this.gl, () => this.gl.scissor(x, y, width, height));
        }
        throwIfDisposed() {
            if (this.disposed) {
                throw new Error('Attempted to use disposed GPGPUContext.');
            }
        }
        throwIfNoProgram() {
            if (this.program == null) {
                throw new Error('No GPU program is currently set.');
            }
        }
    }
    /**
     * Finds the index of the last true element using linear search.
     * Note: We can't do binary search because Chrome expects us to explicitly
     * test all fences before download:
     * https://github.com/tensorflow/tfjs/issues/1145
     */
    function linearSearchLastTrue(arr) {
        let i = 0;
        for (; i < arr.length; ++i) {
            const isDone = arr[i]();
            if (!isDone) {
                break;
            }
        }
        return i - 1;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const { addImpl: addImplCPU, bincountImpl: bincountImplCPU, bincountReduceImpl: bincountReduceImplCPU, ceilImpl: ceilImplCPU, concatImpl: concatImplCPU, equalImpl: equalImplCPU, expImpl: expImplCPU, expm1Impl: expm1ImplCPU, floorImpl: floorImplCPU, gatherNdImpl: gatherNdImplCPU, gatherV2Impl: gatherV2ImplCPU, greaterImpl: greaterImplCPU, greaterEqualImpl: greaterEqualImplCPU, lessImpl: lessImplCPU, lessEqualImpl: lessEqualImplCPU, linSpaceImpl: linSpaceImplCPU, logImpl: logImplCPU, maxImpl: maxImplCPU, maximumImpl: maximumImplCPU, minimumImpl: minimumImplCPU, multiplyImpl: multiplyImplCPU, negImpl: negImplCPU, notEqualImpl: notEqualImplCPU, prodImpl: prodImplCPU, rangeImpl: rangeImplCPU, rsqrtImpl: rsqrtImplCPU, scatterImpl: scatterImplCPU, sigmoidImpl: sigmoidImplCPU, simpleAbsImpl: simpleAbsImplCPU, sliceImpl: sliceImplCPU, sparseFillEmptyRowsImpl: sparseFillEmptyRowsImplCPU, sparseReshapeImpl: sparseReshapeImplCPU, sparseSegmentReductionImpl: sparseSegmentReductionImplCPU, sqrtImpl: sqrtImplCPU, stridedSliceImpl: stridedSliceImplCPU, stringNGramsImpl: stringNGramsImplCPU, stringSplitImpl: stringSplitImplCPU, stringToHashBucketFastImpl: stringToHashBucketFastImplCPU, subImpl: subImplCPU, tileImpl: tileImplCPU, topKImpl: topKImplCPU, transposeImpl: transposeImplCPU, uniqueImpl: uniqueImplCPU, } = shared;

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function getVecChannels(name, rank) {
        return ['x', 'y', 'z', 'w', 'u', 'v'].slice(0, rank).map(d => `${name}.${d}`);
    }
    function getChannels(name, rank) {
        if (rank === 1) {
            return [name];
        }
        return getVecChannels(name, rank);
    }
    function getSourceCoords$2(rank, dims) {
        if (rank === 1) {
            return 'rc';
        }
        let coords = '';
        for (let i = 0; i < rank; i++) {
            coords += dims[i];
            if (i < rank - 1) {
                coords += ',';
            }
        }
        return coords;
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class PackProgram {
        constructor(outputShape) {
            this.variableNames = ['A'];
            this.packedInputs = false;
            this.packedOutput = true;
            // Only input / output 3D tensors.
            this.outputShape = outputShape;
            this.rank = outputShape.length;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            if (this.rank === 0) {
                this.userCode = `
        void main() {
          setOutput(vec4(getA(), 0., 0., 0.));
        }
      `;
            }
            else {
                const channels = getChannels('rc', this.rank);
                const dtype = getCoordsDataType(this.rank);
                const outOfBoundsCondition = this.getOutOfBoundsCondition(channels);
                const setup = this.getSetup(channels);
                const output = this.getOutput(channels);
                this.userCode = `
        void main() {
          ${dtype} rc = getOutputCoords();

          if(${outOfBoundsCondition}) {
            setOutput(vec4(0));
          } else {
            ${setup}

            setOutput(vec4(${output}));
          }
        }
      `;
            }
        }
        getSourceCoordsArr(dims) {
            const coords = [];
            for (let row = 0; row <= 1; row++) {
                for (let col = 0; col <= 1; col++) {
                    let coord = `${row === 0 ? 'r' : 'rp1'}, ${col === 0 ? 'c' : 'cp1'}`;
                    for (let d = 2; d < this.rank; d++) {
                        coord = `${dims[dims.length - 1 - d]},` + coord;
                    }
                    coords.push(coord);
                }
            }
            return coords;
        }
        getOutOfBoundsCondition(dims) {
            if (this.rank === 1) {
                return `rc > ${this.enableShapeUniforms ? 'outShape' : this.outputShape[0]}`;
            }
            let cond = '';
            for (let i = this.rank - 2; i < this.rank; i++) {
                cond += `${dims[i]} >= ${this.enableShapeUniforms ? `outShape[${i}]` : this.outputShape[i]}`;
                if (i < this.rank - 1) {
                    cond += '||';
                }
            }
            return cond;
        }
        getSetup(dims) {
            if (this.rank === 1) {
                return '';
            }
            const innerDims = dims.slice(-2);
            const col = this.enableShapeUniforms ? `outShape[${this.rank} - 1]` :
                this.outputShape[this.rank - 1];
            const row = this.enableShapeUniforms ? `outShape[${this.rank} - 2]` :
                this.outputShape[this.rank - 2];
            return `
      int r = ${innerDims[0]};
      int c = ${innerDims[1]};
      int rp1 = r + 1;
      int cp1 = c + 1;

      bool cEdge = cp1 >= ${col};
      bool rEdge = rp1 >= ${row};
    `;
        }
        getOutput(dims) {
            const sourceCoords = this.getSourceCoordsArr(dims);
            if (this.rank === 1) {
                const outShape = this.enableShapeUniforms ? 'outShape' : this.outputShape[0];
                return `getA(rc), (rc + 1 >= ${outShape} ? 0. : getA(rc + 1)), 0, 0`;
            }
            return `getA(${sourceCoords[0]}),
            cEdge ? 0. : getA(${sourceCoords[1]}),
            rEdge ? 0. : getA(${sourceCoords[2]}),
            rEdge || cEdge ? 0. : getA(${sourceCoords[3]})`;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ReshapePackedProgram {
        constructor(outputShape, inputShape) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.customUniforms = [{ name: 'inputShape', type: 'ivec3' }];
            this.outputShape = outputShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            let mainLoop = ``;
            for (let i = 0; i < 4; i++) {
                let thisRC = `thisRC = rc;`;
                if (i % 2 === 1) {
                    thisRC += `thisRC.z += 1;`;
                }
                if (i > 1) {
                    thisRC += `thisRC.y += 1;`;
                }
                mainLoop += `
        ${thisRC}
        ${i > 0 ? `if(thisRC.y < rows && thisRC.z < cols){` : ''}
          int flatIndex = getFlatIndex(thisRC);

          ivec3 inputRC = inputCoordsFromReshapedOutCoords(flatIndex);
          vec2 inputRCInnerDims = vec2(float(inputRC.y),float(inputRC.z));

          result[${i}] =
            getChannel(getA(inputRC.x, inputRC.y, inputRC.z), inputRCInnerDims);
        ${i > 0 ? '}' : ''}
      `;
            }
            this.userCode = `
      ${getReshapedInputCoords(inputShape, this.enableShapeUniforms)}
      ${this.enableShapeUniforms ? getFlatIndexFrom3DOutput() :
            getFlatIndexFrom3D(outputShape)}

      void main() {
        ivec3 rc = getOutputCoords();

        vec4 result = vec4(0.);

        ivec3 thisRC;
        int rows = ${this.enableShapeUniforms ? 'outShape[1]' : outputShape[1]};
        int cols = ${this.enableShapeUniforms ? 'outShape[2]' : outputShape[2]};

        ${mainLoop}

        setOutput(result);
      }
    `;
        }
    }
    function getReshapedInputCoords(shape, enableShapeUniforms) {
        const coordsFromIndexSnippet = enableShapeUniforms ?
            getLogicalCoordinatesFromFlatIndexByUniform(['r', 'c', 'd'], 'inputShape') :
            getLogicalCoordinatesFromFlatIndex(['r', 'c', 'd'], shape);
        return `
    ivec3 inputCoordsFromReshapedOutCoords(int index) {
      ${coordsFromIndexSnippet}
      return ivec3(r, c, d);
    }
  `;
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class TextureManager {
        constructor(gpgpu) {
            this.gpgpu = gpgpu;
            this.numUsedTextures = 0;
            this.numFreeTextures = 0;
            this._numBytesAllocated = 0;
            this._numBytesFree = 0; // How many bytes that have been allocated
            // are available for reuse.
            this.freeTextures = {};
            this.logEnabled = false;
            this.usedTextures = {};
        }
        acquireTexture(shapeRC, usage, isPacked) {
            const physicalTexType = getPhysicalFromLogicalTextureType(usage, isPacked);
            const shapeKey = getKeyFromTextureShape(shapeRC, physicalTexType, isPacked);
            if (!(shapeKey in this.freeTextures)) {
                this.freeTextures[shapeKey] = [];
            }
            if (!(shapeKey in this.usedTextures)) {
                this.usedTextures[shapeKey] = [];
            }
            const texBytes = computeBytes(shapeRC, physicalTexType, this.gpgpu.gl, this.gpgpu.textureConfig, isPacked);
            if (this.freeTextures[shapeKey].length > 0) {
                this.numFreeTextures--;
                this.numUsedTextures++;
                this._numBytesFree -= texBytes;
                this.log();
                const newTexture = this.freeTextures[shapeKey].shift();
                this.usedTextures[shapeKey].push(newTexture);
                return newTexture;
            }
            let newTexture;
            if (physicalTexType === PhysicalTextureType.PACKED_2X2_FLOAT32) {
                newTexture = this.gpgpu.createPackedMatrixTexture(shapeRC[0], shapeRC[1]);
            }
            else if (physicalTexType === PhysicalTextureType.PACKED_2X2_FLOAT16) {
                newTexture =
                    this.gpgpu.createFloat16PackedMatrixTexture(shapeRC[0], shapeRC[1]);
            }
            else if (physicalTexType === PhysicalTextureType.UNPACKED_FLOAT32) {
                newTexture =
                    this.gpgpu.createFloat32MatrixTexture(shapeRC[0], shapeRC[1]);
            }
            else if (physicalTexType === PhysicalTextureType.UNPACKED_FLOAT16) {
                newTexture =
                    this.gpgpu.createFloat16MatrixTexture(shapeRC[0], shapeRC[1]);
            }
            else if (physicalTexType === PhysicalTextureType.PACKED_4X1_UNSIGNED_BYTE) {
                newTexture =
                    this.gpgpu.createUnsignedBytesMatrixTexture(shapeRC[0], shapeRC[1]);
            }
            this.usedTextures[shapeKey].push(newTexture);
            this.numUsedTextures++;
            this._numBytesAllocated += texBytes;
            this.log();
            return newTexture;
        }
        releaseTexture(texture, shape, logicalTexType, isPacked) {
            if (this.freeTextures == null) {
                // Already disposed.
                return;
            }
            const physicalTexType = getPhysicalFromLogicalTextureType(logicalTexType, isPacked);
            const shapeKey = getKeyFromTextureShape(shape, physicalTexType, isPacked);
            if (!(shapeKey in this.freeTextures)) {
                this.freeTextures[shapeKey] = [];
            }
            const texBytes = computeBytes(shape, physicalTexType, this.gpgpu.gl, this.gpgpu.textureConfig, isPacked);
            const deleteTexThreshold = env$1().get('WEBGL_DELETE_TEXTURE_THRESHOLD');
            if (deleteTexThreshold !== -1 &&
                this._numBytesAllocated > deleteTexThreshold) {
                this.gpgpu.deleteMatrixTexture(texture.texture);
                this._numBytesAllocated -= texBytes;
            }
            else {
                this.freeTextures[shapeKey].push(texture);
                this.numFreeTextures++;
                this._numBytesFree += texBytes;
            }
            this.numUsedTextures--;
            const texList = this.usedTextures[shapeKey];
            const texIndex = texList.indexOf(texture);
            if (texIndex < 0) {
                throw new Error('Cannot release a texture that was never provided by this ' +
                    'texture manager');
            }
            texList.splice(texIndex, 1);
            this.log();
        }
        log() {
            if (!this.logEnabled) {
                return;
            }
            const total = this.numFreeTextures + this.numUsedTextures;
            console.log('Free/Used', `${this.numFreeTextures} / ${this.numUsedTextures}`, `(${total})`);
            const freeRatio = this._numBytesFree / this._numBytesAllocated;
            console.log(`Bytes allocated: ${this._numBytesAllocated}`);
            console.log(`Bytes unused: ${this._numBytesFree} (${Math.round(100 * freeRatio)}%)`);
        }
        get numBytesAllocated() {
            return this._numBytesAllocated;
        }
        get numBytesFree() {
            return this._numBytesFree;
        }
        getNumUsedTextures() {
            return this.numUsedTextures;
        }
        getNumFreeTextures() {
            return this.numFreeTextures;
        }
        dispose() {
            if (this.freeTextures == null) {
                // Already disposed.
                return;
            }
            for (const texShape in this.freeTextures) {
                this.freeTextures[texShape].forEach(tex => {
                    this.gpgpu.deleteMatrixTexture(tex.texture);
                });
            }
            for (const texShape in this.usedTextures) {
                this.usedTextures[texShape].forEach(tex => {
                    this.gpgpu.deleteMatrixTexture(tex.texture);
                });
            }
            this.freeTextures = null;
            this.usedTextures = null;
            this.numUsedTextures = 0;
            this.numFreeTextures = 0;
            this._numBytesAllocated = 0;
            this._numBytesFree = 0;
        }
    }
    function numBytesForInternalFormat(gl, internalFormat) {
        // tslint:disable-next-line:no-any
        const glany = gl;
        if (internalFormat === glany.R32F) {
            return 4;
        }
        else if (internalFormat === glany.R16F) {
            return 2;
        }
        else if (internalFormat === glany.RGBA32F) {
            return 16;
        }
        else if (internalFormat === gl.RGBA) {
            return 16;
        }
        else if (internalFormat === glany.RGBA16F) {
            return 8;
        }
        else if (internalFormat === glany.RGBA8) {
            return 4;
        }
        throw new Error(`Unknown internal format ${internalFormat}`);
    }
    function computeBytes(shape, physicalTexType, gl, textureConfig, isPacked) {
        // It is not possible to infer packed status from the texture type because
        // depending on the textureConfig, different  texture types may resolve to the
        // same internal format (e.g. in WebGL1, the internal format for
        // UNPACKED_FLOAT16 textures is gl.RGBA). Therefore we pass in `isPacked`
        // explicitly.
        const internalFormat = internalFormatForPhysicalTexType(physicalTexType, textureConfig);
        let numElements;
        if (isPacked) {
            const [packedWidth, packedHeight] = getPackedMatrixTextureShapeWidthHeight(shape[0], shape[1]);
            numElements = packedWidth * packedHeight;
        }
        else {
            const [width, height] = getUnpackedMatrixTextureShapeWidthHeight(shape[0], shape[1]);
            numElements = width * height;
        }
        const bytesPerElement = numBytesForInternalFormat(gl, internalFormat);
        return numElements * bytesPerElement;
    }
    function internalFormatForPhysicalTexType(physicalTexType, textureConfig) {
        switch (physicalTexType) {
            case PhysicalTextureType.PACKED_2X2_FLOAT32:
                return getInternalFormatForPackedMatrixTexture(textureConfig);
            case PhysicalTextureType.PACKED_2X2_FLOAT16:
                return getInternalFormatForFloat16PackedMatrixTexture(textureConfig);
            case PhysicalTextureType.UNPACKED_FLOAT32:
                return getInternalFormatForFloat32MatrixTexture(textureConfig);
            case PhysicalTextureType.UNPACKED_FLOAT16:
                return getInternalFormatForFloat16MatrixTexture(textureConfig);
            case PhysicalTextureType.PACKED_4X1_UNSIGNED_BYTE:
                return getInternalFormatForUnsignedBytesMatrixTexture(textureConfig);
            default:
                throw new Error(`Unknown physical texture type ${physicalTexType}`);
        }
    }
    function getPhysicalTextureForRendering(isPacked) {
        if (env$1().getBool('WEBGL_RENDER_FLOAT32_ENABLED')) {
            if (isPacked) {
                return PhysicalTextureType.PACKED_2X2_FLOAT32;
            }
            return PhysicalTextureType.UNPACKED_FLOAT32;
        }
        if (isPacked) {
            return PhysicalTextureType.PACKED_2X2_FLOAT16;
        }
        return PhysicalTextureType.UNPACKED_FLOAT16;
    }
    function getPhysicalFromLogicalTextureType(logicalTexType, isPacked) {
        if (logicalTexType === TextureUsage.UPLOAD) {
            return PhysicalTextureType.PACKED_2X2_FLOAT32;
        }
        else if (logicalTexType === TextureUsage.RENDER || logicalTexType == null) {
            return getPhysicalTextureForRendering(isPacked);
        }
        else if (logicalTexType === TextureUsage.DOWNLOAD ||
            logicalTexType === TextureUsage.PIXELS) {
            return PhysicalTextureType.PACKED_4X1_UNSIGNED_BYTE;
        }
        throw new Error(`Unknown logical texture type ${logicalTexType}`);
    }
    function getKeyFromTextureShape(shapeRowsCol, physicalTexType, isPacked) {
        return `${shapeRowsCol[0]}_${shapeRowsCol[1]}_${physicalTexType}_${isPacked}`;
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class UnaryOpProgram {
        constructor(aShape, opSnippet) {
            this.variableNames = ['A'];
            this.outputShape = aShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            this.userCode = `
      float unaryOperation(float x) {
        ${opSnippet}
      }

      void main() {
        float x = getAAtOutCoords();
        float y = unaryOperation(x);

        setOutput(y);
      }
    `;
        }
    }
    const CHECK_NAN_SNIPPET$2 = `if (isnan(x)) return x;`;
    const LINEAR$1 = `return x;`;
    const ABS$1 = `return abs(x);`;
    const ELU$2 = `return (x >= 0.0) ? x : (exp(x) - 1.0);`;
    const RELU$2 = CHECK_NAN_SNIPPET$2 + `
  return (x < 0.0) ? 0.0 : x;
`;
    const RELU6$2 = CHECK_NAN_SNIPPET$2 + `
  return (x < 0.0) ? 0.0 : min(6.0, x);
`;
    const CLONE = 'return x;';
    const SIGMOID$2 = `return 1.0 / (1.0 + exp(-1.0 * x));`;

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const LINEAR = `return x;`;
    const ELU$1 = `
  vec4 result;

  result.r = (x.r >= 0.0) ? x.r : (exp(x.r) - 1.0);
  result.g = (x.g >= 0.0) ? x.g : (exp(x.g) - 1.0);
  result.b = (x.b >= 0.0) ? x.b : (exp(x.b) - 1.0);
  result.a = (x.a >= 0.0) ? x.a : (exp(x.a) - 1.0);

  return result;
`;
    const RELU$1 = `
  vec4 result = x * vec4(greaterThanEqual(x, vec4(0.0)));
  bvec4 isNaN = isnan(x);

  result.r = isNaN.r ? x.r : result.r;
  result.g = isNaN.g ? x.g : result.g;
  result.b = isNaN.b ? x.b : result.b;
  result.a = isNaN.a ? x.a : result.a;

  return result;
`;
    const RELU6$1 = `
  vec4 result = min(x, vec4(6.)) * vec4(greaterThanEqual(x, vec4(0.0)));
  bvec4 isNaN = isnan(x);

  result.r = isNaN.r ? x.r : result.r;
  result.g = isNaN.g ? x.g : result.g;
  result.b = isNaN.b ? x.b : result.b;
  result.a = isNaN.a ? x.a : result.a;

  return result;
`;
    const SIGMOID$1 = `return 1.0 / (1.0 + exp(-1.0 * x));`;
    class UnaryOpPackedProgram {
        constructor(aShape, opSnippet) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.outputShape = aShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            this.userCode = `
      vec4 unaryOperation(vec4 x) {
        ${opSnippet}
      }

      void main() {
        vec4 x = getAAtOutCoords();
        vec4 y = unaryOperation(x);

        setOutput(y);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class UnpackProgram {
        constructor(outputShape) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = false;
            this.outputShape = outputShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            const rank = outputShape.length;
            const channels = getChannels('rc', rank);
            const dtype = getCoordsDataType(rank);
            const sourceCoords = getSourceCoords$2(rank, channels);
            const innerDims = channels.slice(-2);
            const coords = rank <= 1 ? 'rc' : `vec2(${innerDims.join(',')})`;
            this.userCode = `
      void main() {
        ${dtype} rc = getOutputCoords();
        vec4 packedInput = getA(${sourceCoords});

        setOutput(getChannel(packedInput, ${coords}));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const whereImpl = whereImpl$2;
    const EPSILON_FLOAT32 = 1e-7;
    const EPSILON_FLOAT16 = 1e-4;
    const binaryCaches = {};
    function getBinaryCache(webGLVersion) {
        if (webGLVersion in binaryCaches) {
            return binaryCaches[webGLVersion];
        }
        binaryCaches[webGLVersion] = {};
        return binaryCaches[webGLVersion];
    }
    // Empirically determined constant used to determine size threshold for handing
    // off execution to the CPU.
    const CPU_HANDOFF_SIZE_THRESHOLD = env$1().getNumber('CPU_HANDOFF_SIZE_THRESHOLD');
    // Empirically determined constant used to decide the number of MB on GPU
    // before we warn about high memory use. The MB are this constant * screen area
    // * dpi / 1024 / 1024.
    const BEFORE_PAGING_CONSTANT = 600;
    function numMBBeforeWarning() {
        if (env$1().global.screen == null) {
            return 1024; // 1 GB.
        }
        return (env$1().global.screen.height * env$1().global.screen.width *
            window.devicePixelRatio) *
            BEFORE_PAGING_CONSTANT / 1024 / 1024;
    }
    class MathBackendWebGL extends KernelBackend {
        constructor(gpuResource) {
            super();
            // Maps data ids that have a pending read operation, to list of subscribers.
            this.pendingRead = new WeakMap();
            // List of data ids that are scheduled for disposal, but are waiting on a
            // pending read operation.
            this.pendingDisposal = new WeakSet();
            // Used to count the number of 'shallow' sliced tensors that point to the
            // same data id.
            this.dataRefCount = new WeakMap();
            this.numBytesInGPU = 0;
            // Accumulated time spent (including blocking) in uploading data to webgl.
            this.uploadWaitMs = 0;
            // Accumulated time spent (including blocking in downloading data from webgl.
            this.downloadWaitMs = 0;
            // record the last manual GL Flush time.
            this.lastGlFlushTime = 0;
            this.warnedAboutMemory = false;
            this.pendingDeletes = 0;
            this.disposed = false;
            if (!env$1().getBool('HAS_WEBGL')) {
                throw new Error('WebGL is not supported on this device');
            }
            let newGPGPU;
            if (gpuResource != null) {
                if (gpuResource instanceof GPGPUContext) {
                    newGPGPU = gpuResource;
                }
                else {
                    const gl = getWebGLContext(env$1().getNumber('WEBGL_VERSION'), gpuResource);
                    newGPGPU = new GPGPUContext(gl);
                }
                this.binaryCache = {};
                this.gpgpuCreatedLocally = false;
            }
            else {
                const gl = getWebGLContext(env$1().getNumber('WEBGL_VERSION'));
                newGPGPU = new GPGPUContext(gl);
                this.binaryCache = getBinaryCache(env$1().getNumber('WEBGL_VERSION'));
                this.gpgpuCreatedLocally = true;
            }
            this.gpgpu = newGPGPU;
            this.canvas = this.gpgpu.gl.canvas;
            this.textureManager = new TextureManager(this.gpgpu);
            this.numMBBeforeWarning = numMBBeforeWarning();
            this.texData = new DataStorage(this, engine());
        }
        nextDataId() {
            return MathBackendWebGL.nextDataId++;
        }
        numDataIds() {
            return this.texData.numDataIds() - this.pendingDeletes;
        }
        write(values, shape, dtype) {
            if (env$1().getBool('WEBGL_CHECK_NUMERICAL_PROBLEMS') ||
                env$1().getBool('DEBUG')) {
                this.checkNumericalProblems(values);
            }
            if (dtype === 'complex64' && values != null) {
                throw new Error(`Cannot write to a complex64 dtype. ` +
                    `Please use tf.complex(real, imag).`);
            }
            const dataId = { id: this.nextDataId() };
            this.texData.set(dataId, { shape, dtype, values, usage: TextureUsage.UPLOAD, refCount: 1 });
            return dataId;
        }
        /** Return refCount of a `TensorData`. */
        refCount(dataId) {
            if (this.texData.has(dataId)) {
                const tensorData = this.texData.get(dataId);
                return tensorData.refCount;
            }
            return 0;
        }
        /** Increase refCount of a `TextureData`. */
        incRef(dataId) {
            const texData = this.texData.get(dataId);
            texData.refCount++;
        }
        /** Decrease refCount of a `TextureData`. */
        decRef(dataId) {
            if (this.texData.has(dataId)) {
                const texData = this.texData.get(dataId);
                texData.refCount--;
            }
        }
        move(dataId, values, shape, dtype, refCount) {
            if (env$1().getBool('DEBUG')) {
                this.checkNumericalProblems(values);
            }
            if (dtype === 'complex64') {
                throw new Error(`Cannot write to a complex64 dtype. ` +
                    `Please use tf.complex(real, imag).`);
            }
            this.texData.set(dataId, { shape, dtype, values, usage: TextureUsage.UPLOAD, refCount });
        }
        disposeIntermediateTensorInfo(tensorInfo) {
            this.disposeData(tensorInfo.dataId);
        }
        readSync(dataId) {
            const texData = this.texData.get(dataId);
            const { values, dtype, complexTensorInfos, slice, shape, isPacked } = texData;
            // The presence of `slice` indicates this tensor is a shallow slice of a
            // different tensor, and is using that original tensor's texture. Run
            // `clone` in order to copy that texture and read from it.
            if (slice != null) {
                let program;
                if (isPacked) {
                    program = new UnaryOpPackedProgram(shape, CLONE);
                }
                else {
                    program = new UnaryOpProgram(shape, CLONE);
                }
                const res = this.runWebGLProgram(program, [{ dataId, shape, dtype }], dtype);
                const data = this.readSync(res.dataId);
                this.disposeIntermediateTensorInfo(res);
                return data;
            }
            if (values != null) {
                return this.convertAndCacheOnCPU(dataId);
            }
            if (dtype === 'string') {
                return values;
            }
            const shouldTimeProgram = this.activeTimers != null;
            let start;
            if (shouldTimeProgram) {
                start = now();
            }
            let result;
            if (dtype === 'complex64') {
                const realValues = this.readSync(complexTensorInfos.real.dataId);
                const imagValues = this.readSync(complexTensorInfos.imag.dataId);
                result = mergeRealAndImagArrays(realValues, imagValues);
            }
            else {
                result = this.getValuesFromTexture(dataId);
            }
            if (shouldTimeProgram) {
                this.downloadWaitMs += now() - start;
            }
            return this.convertAndCacheOnCPU(dataId, result);
        }
        async read(dataId) {
            if (this.pendingRead.has(dataId)) {
                const subscribers = this.pendingRead.get(dataId);
                return new Promise(resolve => subscribers.push(resolve));
            }
            const texData = this.texData.get(dataId);
            const { values, shape, slice, dtype, complexTensorInfos, isPacked } = texData;
            // The presence of `slice` indicates this tensor is a shallow slice of a
            // different tensor, and is using that original tensor's texture. Run
            // `clone` in order to copy that texture and read from it.
            if (slice != null) {
                let program;
                if (isPacked) {
                    program = new UnaryOpPackedProgram(shape, CLONE);
                }
                else {
                    program = new UnaryOpProgram(shape, CLONE);
                }
                const res = this.runWebGLProgram(program, [{ dataId, shape, dtype }], dtype);
                const data = this.read(res.dataId);
                this.disposeIntermediateTensorInfo(res);
                return data;
            }
            if (values != null) {
                return this.convertAndCacheOnCPU(dataId);
            }
            if (env$1().getBool('DEBUG')) {
                // getBool('WEBGL_DOWNLOAD_FLOAT_ENABLED') caused a blocking GPU call.
                // For performance reason, only check it for debugging. In production,
                // it doesn't handle this use case anyway, so behavior is not changed.
                if (!env$1().getBool('WEBGL_DOWNLOAD_FLOAT_ENABLED') &&
                    env$1().getNumber('WEBGL_VERSION') === 2) {
                    throw new Error(`tensor.data() with WEBGL_DOWNLOAD_FLOAT_ENABLED=false and ` +
                        `WEBGL_VERSION=2 not yet supported.`);
                }
            }
            let buffer = null;
            let tmpDownloadTarget;
            if (dtype !== 'complex64' && env$1().get('WEBGL_BUFFER_SUPPORTED')) {
                // Possibly copy the texture into a buffer before inserting a fence.
                tmpDownloadTarget = this.decode(dataId);
                const tmpData = this.texData.get(tmpDownloadTarget.dataId);
                buffer = this.gpgpu.createBufferFromTexture(tmpData.texture.texture, ...getDenseTexShape(shape));
            }
            this.pendingRead.set(dataId, []);
            if (dtype !== 'complex64') {
                // Create a fence and wait for it to resolve.
                await this.gpgpu.createAndWaitForFence();
            }
            // Download the values from the GPU.
            let vals;
            if (dtype === 'complex64') {
                const ps = await Promise.all([
                    this.read(complexTensorInfos.real.dataId),
                    this.read(complexTensorInfos.imag.dataId)
                ]);
                const realValues = ps[0];
                const imagValues = ps[1];
                vals = mergeRealAndImagArrays(realValues, imagValues);
            }
            else if (buffer == null) {
                vals = this.getValuesFromTexture(dataId);
            }
            else {
                const size = sizeFromShape(shape);
                vals = this.gpgpu.downloadFloat32MatrixFromBuffer(buffer, size);
            }
            if (tmpDownloadTarget != null) {
                this.disposeIntermediateTensorInfo(tmpDownloadTarget);
            }
            if (buffer != null) {
                const gl = this.gpgpu.gl;
                callAndCheck(gl, () => gl.deleteBuffer(buffer));
            }
            const dTypeVals = this.convertAndCacheOnCPU(dataId, vals);
            const subscribers = this.pendingRead.get(dataId);
            this.pendingRead.delete(dataId);
            // Notify all pending reads.
            subscribers.forEach(resolve => resolve(dTypeVals));
            if (this.pendingDisposal.has(dataId)) {
                this.pendingDisposal.delete(dataId);
                if (this.disposeData(dataId)) {
                    engine().removeDataId(dataId, this);
                }
                this.pendingDeletes--;
            }
            return dTypeVals;
        }
        /**
         * Read tensor to a new texture that is densely packed for ease of use.
         * @param dataId The source tensor.
         * @param options
         *     customTexShape: Optional. If set, will use the user defined texture
         *     shape to create the texture.
         */
        readToGPU(dataId, options = {}) {
            const texData = this.texData.get(dataId);
            const { values, shape, slice, dtype, isPacked, texture } = texData;
            if (dtype === 'complex64') {
                throw new Error('Does not support reading texture for complex64 dtype.');
            }
            // The presence of `slice` indicates this tensor is a shallow slice of a
            // different tensor, and is using that original tensor's texture. Run
            // `clone` in order to copy that texture and read from it.
            if (slice != null) {
                let program;
                if (isPacked) {
                    program = new UnaryOpPackedProgram(shape, CLONE);
                }
                else {
                    program = new UnaryOpProgram(shape, CLONE);
                }
                const res = this.runWebGLProgram(program, [{ dataId, shape, dtype }], dtype);
                const gpuResouorce = this.readToGPU(res, options);
                this.disposeIntermediateTensorInfo(res);
                return gpuResouorce;
            }
            if (texture == null) {
                if (values != null) {
                    throw new Error('Data is not on GPU but on CPU.');
                }
                else {
                    throw new Error('There is no data on GPU or CPU.');
                }
            }
            // Decode the texture so that it is stored densely (using four channels).
            const tmpTarget = this.decode(dataId, options.customTexShape);
            // Make engine track this tensor, so that we can dispose it later.
            const tensorRef = engine().makeTensorFromTensorInfo(tmpTarget);
            const tmpData = this.texData.get(tmpTarget.dataId);
            return Object.assign({ tensorRef }, tmpData.texture);
        }
        bufferSync(t) {
            const data = this.readSync(t.dataId);
            if (t.dtype === 'string') {
                try {
                    // Decode the bytes into string.
                    const strings = data.map(d => decodeString(d));
                    return buffer(t.shape, t.dtype, strings);
                }
                catch (_a) {
                    throw new Error('Failed to decode encoded string bytes into utf-8');
                }
            }
            return buffer(t.shape, t.dtype, data);
        }
        checkNumericalProblems(values) {
            if (values == null) {
                return;
            }
            for (let i = 0; i < values.length; i++) {
                const num = values[i];
                if (!canBeRepresented(num)) {
                    if (env$1().getBool('WEBGL_RENDER_FLOAT32_CAPABLE')) {
                        throw Error(`The value ${num} cannot be represented with your ` +
                            `current settings. Consider enabling float32 rendering: ` +
                            `'tf.env().set('WEBGL_RENDER_FLOAT32_ENABLED', true);'`);
                    }
                    throw Error(`The value ${num} cannot be represented on this device.`);
                }
            }
        }
        getValuesFromTexture(dataId) {
            const { shape, dtype, isPacked } = this.texData.get(dataId);
            const size = sizeFromShape(shape);
            if (env$1().getBool('WEBGL_DOWNLOAD_FLOAT_ENABLED')) {
                const tmpTarget = this.decode(dataId);
                const tmpData = this.texData.get(tmpTarget.dataId);
                const vals = this.gpgpu
                    .downloadMatrixFromPackedTexture(tmpData.texture.texture, ...getDenseTexShape(shape))
                    .subarray(0, size);
                this.disposeIntermediateTensorInfo(tmpTarget);
                return vals;
            }
            const shouldUsePackedProgram = env$1().getBool('WEBGL_PACK') && isPacked === true;
            const outputShape = shouldUsePackedProgram ? getShapeAs3D(shape) : shape;
            const program = shouldUsePackedProgram ?
                new EncodeFloatPackedProgram(outputShape) :
                new EncodeFloatProgram(outputShape);
            const output = this.runWebGLProgram(program, [{ shape: outputShape, dtype, dataId }], 'float32');
            const tmpData = this.texData.get(output.dataId);
            const vals = this.gpgpu
                .downloadByteEncodedFloatMatrixFromOutputTexture(tmpData.texture.texture, tmpData.texShape[0], tmpData.texShape[1])
                .subarray(0, size);
            this.disposeIntermediateTensorInfo(output);
            return vals;
        }
        timerAvailable() {
            return env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_RELIABLE') > 0;
        }
        time(f) {
            const oldActiveTimers = this.activeTimers;
            const newActiveTimers = [];
            let outerMostTime = false;
            if (this.programTimersStack == null) {
                this.programTimersStack = newActiveTimers;
                outerMostTime = true;
            }
            else {
                this.activeTimers.push(newActiveTimers);
            }
            this.activeTimers = newActiveTimers;
            f();
            // needing to split these up because util.flatten only accepts certain types
            const flattenedActiveTimerQueries = flatten$3(this.activeTimers.map((d) => d.query))
                .filter(d => d != null);
            const flattenedActiveTimerNames = flatten$3(this.activeTimers.map((d) => d.name))
                .filter(d => d != null);
            this.activeTimers = oldActiveTimers;
            if (outerMostTime) {
                this.programTimersStack = null;
            }
            const res = {
                uploadWaitMs: this.uploadWaitMs,
                downloadWaitMs: this.downloadWaitMs,
                kernelMs: null,
                wallMs: null // will be filled by the engine
            };
            return (async () => {
                if (env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_RELIABLE') >
                    0) {
                    const kernelMs = await Promise.all(flattenedActiveTimerQueries);
                    res['kernelMs'] = sum$3(kernelMs);
                    res['getExtraProfileInfo'] = () => kernelMs
                        .map((d, i) => ({ name: flattenedActiveTimerNames[i], ms: d }))
                        .map(d => `${d.name}: ${d.ms}`)
                        .join(', ');
                }
                else {
                    res['kernelMs'] = {
                        error: 'WebGL query timers are not supported in this environment.'
                    };
                }
                this.uploadWaitMs = 0;
                this.downloadWaitMs = 0;
                return res;
            })();
        }
        memory() {
            return {
                unreliable: false,
                numBytesInGPU: this.numBytesInGPU,
                numBytesInGPUAllocated: this.textureManager.numBytesAllocated,
                numBytesInGPUFree: this.textureManager.numBytesFree
            };
        }
        startTimer() {
            if (env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_RELIABLE') > 0) {
                return this.gpgpu.beginQuery();
            }
            return { startMs: now(), endMs: null };
        }
        endTimer(query) {
            if (env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_RELIABLE') > 0) {
                this.gpgpu.endQuery();
                return query;
            }
            query.endMs = now();
            return query;
        }
        async getQueryTime(query) {
            if (env$1().getNumber('WEBGL_DISJOINT_QUERY_TIMER_EXTENSION_RELIABLE') > 0) {
                return this.gpgpu.waitForQueryAndGetTime(query);
            }
            const timerQuery = query;
            return timerQuery.endMs - timerQuery.startMs;
        }
        /**
         * Decrease the RefCount on the dataId and dispose the memory if the dataId
         * has 0 refCount. If there are pending read on the data, the disposal would
         * added to the pending delete queue. Return true if the dataId is removed
         * from backend or the backend does not contain the dataId, false if the
         * dataId is not removed. Memory may or may not be released even when dataId
         * is removed, which also depends on dataRefCount, see `releaseGPU`.
         * @param dataId
         * @oaram force Optional, remove the data regardless of refCount
         */
        disposeData(dataId, force = false) {
            if (this.pendingDisposal.has(dataId)) {
                return false;
            }
            // No-op if already disposed.
            if (!this.texData.has(dataId)) {
                return true;
            }
            // if force flag is set, change refCount to 0, this would ensure disposal
            // when added to the pendingDisposal queue. Memory may or may not be
            // released, which also depends on dataRefCount, see `releaseGPU`.
            if (force) {
                this.texData.get(dataId).refCount = 0;
            }
            else {
                this.texData.get(dataId).refCount--;
            }
            if (!force && this.texData.get(dataId).refCount > 0) {
                return false;
            }
            if (this.pendingRead.has(dataId)) {
                this.pendingDisposal.add(dataId);
                this.pendingDeletes++;
                return false;
            }
            this.releaseGPUData(dataId);
            const { complexTensorInfos } = this.texData.get(dataId);
            if (complexTensorInfos != null) {
                this.disposeData(complexTensorInfos.real.dataId, force);
                this.disposeData(complexTensorInfos.imag.dataId, force);
            }
            this.texData.delete(dataId);
            return true;
        }
        releaseGPUData(dataId) {
            const { texture, dtype, texShape, usage, isPacked, slice } = this.texData.get(dataId);
            const key = slice && slice.origDataId || dataId;
            const refCount = this.dataRefCount.get(key);
            if (refCount > 1) {
                this.dataRefCount.set(key, refCount - 1);
            }
            else {
                this.dataRefCount.delete(key);
                if (texture != null) {
                    this.numBytesInGPU -= this.computeBytes(texShape, dtype);
                    this.textureManager.releaseTexture(texture, texShape, usage, isPacked);
                }
            }
            const texData = this.texData.get(dataId);
            texData.texture = null;
            texData.texShape = null;
            texData.isPacked = false;
            texData.slice = null;
        }
        getTexture(dataId) {
            this.uploadToGPU(dataId);
            return this.texData.get(dataId).texture.texture;
        }
        /**
         * Returns internal information for the specific data bucket. Used in unit
         * tests.
         */
        getDataInfo(dataId) {
            return this.texData.get(dataId);
        }
        /*
        Tests whether all the inputs to an op are small and on the CPU. This heuristic
        determines when it would be faster to execute a kernel on the CPU. WebGL
        kernels opt into running this check and forwarding when appropriate.
        TODO(https://github.com/tensorflow/tfjs/issues/872): Develop a more
        sustainable strategy for optimizing backend execution of ops.
         */
        shouldExecuteOnCPU(inputs, sizeThreshold = CPU_HANDOFF_SIZE_THRESHOLD) {
            return env$1().getBool('WEBGL_CPU_FORWARD') &&
                inputs.every(input => this.texData.get(input.dataId).texture == null &&
                    sizeFromShape(input.shape) < sizeThreshold);
        }
        getGPGPUContext() {
            return this.gpgpu;
        }
        where(condition) {
            warn('tf.where() in webgl locks the UI thread. ' +
                'Call tf.whereAsync() instead');
            const condVals = condition.dataSync();
            return whereImpl(condition.shape, condVals);
        }
        packedUnaryOp(x, op, dtype) {
            const program = new UnaryOpPackedProgram(x.shape, op);
            const outInfo = this.compileAndRun(program, [x], dtype);
            return engine().makeTensorFromTensorInfo(outInfo);
        }
        // TODO(msoulanille) remove this once the backend has been modularized
        // a copy is needed here to break a circular dependency.
        // Also remove the op from unary_op.
        abs(x) {
            // TODO: handle cases when x is complex.
            if (this.shouldExecuteOnCPU([x]) && x.dtype !== 'complex64') {
                const outValues = simpleAbsImplCPU(this.texData.get(x.dataId).values);
                return this.makeOutput(x.shape, x.dtype, outValues);
            }
            if (env$1().getBool('WEBGL_PACK_UNARY_OPERATIONS')) {
                return this.packedUnaryOp(x, ABS$1, x.dtype);
            }
            const program = new UnaryOpProgram(x.shape, ABS$1);
            const outInfo = this.compileAndRun(program, [x]);
            return engine().makeTensorFromTensorInfo(outInfo);
        }
        makeTensorInfo(shape, dtype, values) {
            let dataId;
            if (dtype === 'string' && values != null && values.length > 0 &&
                isString$2(values[0])) {
                const encodedValues = values.map(d => encodeString(d));
                dataId = this.write(encodedValues, shape, dtype);
            }
            else {
                dataId = this.write(values, shape, dtype);
            }
            this.texData.get(dataId).usage = null;
            return { dataId, shape, dtype };
        }
        makeOutput(shape, dtype, values) {
            return engine().makeTensorFromTensorInfo(this.makeTensorInfo(shape, dtype, values), this);
        }
        unpackTensor(input) {
            const program = new UnpackProgram(input.shape);
            return this.runWebGLProgram(program, [input], input.dtype);
        }
        packTensor(input) {
            const program = new PackProgram(input.shape);
            const preventEagerUnpackingOutput = true;
            return this.runWebGLProgram(program, [input], input.dtype, null /* customUniformValues */, preventEagerUnpackingOutput);
        }
        packedReshape(input, afterShape) {
            const input3DShape = [
                getBatchDim(input.shape),
                ...getRowsCols(input.shape)
            ];
            const input3D = {
                dtype: input.dtype,
                shape: input3DShape,
                dataId: input.dataId
            };
            const afterShapeAs3D = [
                getBatchDim(afterShape), ...getRowsCols(afterShape)
            ];
            const program = new ReshapePackedProgram(afterShapeAs3D, input3DShape);
            const preventEagerUnpackingOfOutput = true;
            const customValues = [input3DShape];
            const output = this.runWebGLProgram(program, [input3D], input.dtype, customValues, preventEagerUnpackingOfOutput);
            return { dataId: output.dataId, shape: afterShape, dtype: output.dtype };
        }
        decode(dataId, customTexShape) {
            const texData = this.texData.get(dataId);
            const { isPacked, shape, dtype } = texData;
            if (customTexShape != null) {
                const size = sizeFromShape(shape);
                const texSize = customTexShape[0] * customTexShape[1] * 4;
                assert$4(size <= texSize, () => 'customTexShape is too small. ' +
                    'Row * Column * 4 should be equal or larger than the ' +
                    'size of the tensor data.');
            }
            const shapeAs3D = getShapeAs3D(shape);
            let program;
            if (isPacked) {
                program = new DecodeMatrixPackedProgram(shapeAs3D);
            }
            else {
                program = new DecodeMatrixProgram(shapeAs3D);
            }
            const preventEagerUnpackingOfOutput = true;
            const customValues = [customTexShape != null ? customTexShape :
                    getDenseTexShape(shapeAs3D)];
            const out = this.runWebGLProgram(program, [{ shape: shapeAs3D, dtype, dataId }], dtype, customValues, preventEagerUnpackingOfOutput, customTexShape);
            return { dtype, shape, dataId: out.dataId };
        }
        runWebGLProgram(program, inputs, outputDtype, customUniformValues, preventEagerUnpackingOfOutput = false, customTexShape) {
            const output = this.makeTensorInfo(program.outputShape, outputDtype);
            const outData = this.texData.get(output.dataId);
            if (program.packedOutput) {
                outData.isPacked = true;
            }
            if (program.outPackingScheme === PackingScheme.DENSE) {
                const texelShape = customTexShape != null ?
                    customTexShape :
                    getDenseTexShape(program.outputShape);
                // For a densely packed output, we explicitly set texShape
                // so it doesn't get assigned later according to our typical packing
                // scheme wherein a single texel can only contain values from adjacent
                // rows/cols.
                outData.texShape = texelShape.map(d => d * 2);
            }
            if (program.outTexUsage != null) {
                outData.usage = program.outTexUsage;
            }
            if (sizeFromShape(output.shape) === 0) {
                // Short-circuit the computation since the result is empty (has 0 in its
                // shape).
                outData.values =
                    getTypedArrayFromDType(output.dtype, 0);
                return output;
            }
            const dataToDispose = [];
            const inputsData = inputs.map(input => {
                if (input.dtype === 'complex64') {
                    throw new Error(`GPGPUProgram does not support complex64 input. For complex64 ` +
                        `dtypes, please separate the program into real and imaginary ` +
                        `parts.`);
                }
                let texData = this.texData.get(input.dataId);
                if (texData.texture == null) {
                    if (!program.packedInputs &&
                        sizeFromShape(input.shape) <=
                            env$1().getNumber('WEBGL_SIZE_UPLOAD_UNIFORM')) {
                        // Upload small tensors that live on the CPU as uniforms, not as
                        // textures. Do this only when the environment supports 32bit floats
                        // due to problems when comparing 16bit floats with 32bit floats.
                        // TODO(https://github.com/tensorflow/tfjs/issues/821): Make it
                        // possible for packed shaders to sample from uniforms.
                        return {
                            shape: input.shape,
                            texData: null,
                            isUniform: true,
                            uniformValues: texData.values
                        };
                    }
                    // This ensures that if a packed program's inputs have not yet been
                    // uploaded to the GPU, they get uploaded as packed right off the bat.
                    if (program.packedInputs) {
                        texData.isPacked = true;
                        texData.shape = input.shape;
                    }
                }
                this.uploadToGPU(input.dataId);
                if (!!texData.isPacked !== !!program.packedInputs) {
                    input = texData.isPacked ? this.unpackTensor(input) :
                        this.packTensor(input);
                    dataToDispose.push(input);
                    texData = this.texData.get(input.dataId);
                }
                else if (texData.isPacked &&
                    !isReshapeFree(texData.shape, input.shape)) {
                    // This is a special case where a texture exists for a tensor
                    // but the shapes are incompatible (due to packing constraints) because
                    // the tensor did not have a chance to go through the packed reshape
                    // shader. This only happens when we reshape the *same* tensor to form
                    // *distinct* inputs to an op, e.g. dotting a vector with itself. This
                    // case will disappear once packed uploading is the default.
                    const savedInput = input;
                    const targetShape = input.shape;
                    input.shape = texData.shape;
                    input = this.packedReshape(input, targetShape);
                    dataToDispose.push(input);
                    texData = this.texData.get(input.dataId);
                    savedInput.shape = targetShape;
                }
                return { shape: input.shape, texData, isUniform: false };
            });
            this.uploadToGPU(output.dataId);
            const outputData = { shape: output.shape, texData: outData, isUniform: false };
            const key = makeShaderKey(program, inputsData, outputData);
            const binary = this.getAndSaveBinary(key, () => {
                return compileProgram(this.gpgpu, program, inputsData, outputData);
            });
            const shouldTimeProgram = this.activeTimers != null;
            let query;
            if (shouldTimeProgram) {
                query = this.startTimer();
            }
            if (!env$1().get('ENGINE_COMPILE_ONLY')) {
                runProgram(this.gpgpu, binary, inputsData, outputData, customUniformValues);
            }
            dataToDispose.forEach(info => this.disposeIntermediateTensorInfo(info));
            if (shouldTimeProgram) {
                query = this.endTimer(query);
                this.activeTimers.push({ name: program.constructor.name, query: this.getQueryTime(query) });
            }
            const glFlushThreshold = env$1().get('WEBGL_FLUSH_THRESHOLD');
            // Manually GL flush requested
            if (glFlushThreshold > 0) {
                const time = now();
                if ((time - this.lastGlFlushTime) > glFlushThreshold) {
                    this.gpgpu.gl.flush();
                    this.lastGlFlushTime = time;
                }
            }
            if (!env$1().getBool('WEBGL_LAZILY_UNPACK') && outData.isPacked &&
                preventEagerUnpackingOfOutput === false) {
                const unpacked = this.unpackTensor(output);
                this.disposeIntermediateTensorInfo(output);
                return unpacked;
            }
            return output;
        }
        compileAndRun(program, inputs, outputDtype, customUniformValues, preventEagerUnpackingOfOutput = false) {
            outputDtype = outputDtype || inputs[0].dtype;
            const outInfo = this.runWebGLProgram(program, inputs, outputDtype, customUniformValues, preventEagerUnpackingOfOutput);
            return outInfo;
        }
        getAndSaveBinary(key, getBinary) {
            if (!(key in this.binaryCache)) {
                this.binaryCache[key] = getBinary();
            }
            return this.binaryCache[key];
        }
        getTextureManager() {
            return this.textureManager;
        }
        dispose() {
            if (this.disposed) {
                return;
            }
            // Avoid disposing the compiled webgl programs during unit testing because
            // it slows down test execution.
            if (!env$1().getBool('IS_TEST')) {
                const allKeys = Object.keys(this.binaryCache);
                allKeys.forEach(key => {
                    this.gpgpu.deleteProgram(this.binaryCache[key].webGLProgram);
                    delete this.binaryCache[key];
                });
            }
            this.textureManager.dispose();
            if (this.canvas != null &&
                (typeof (HTMLCanvasElement) !== 'undefined' &&
                    this.canvas instanceof HTMLCanvasElement)) {
                this.canvas.remove();
            }
            else {
                this.canvas = null;
            }
            if (this.gpgpuCreatedLocally) {
                this.gpgpu.program = null;
                this.gpgpu.dispose();
            }
            this.disposed = true;
        }
        floatPrecision() {
            if (this.floatPrecisionValue == null) {
                this.floatPrecisionValue = tidy(() => {
                    if (!env$1().get('WEBGL_RENDER_FLOAT32_ENABLED')) {
                        // Momentarily switching DEBUG flag to false so we don't throw an
                        // error trying to upload a small value.
                        const debugFlag = env$1().getBool('DEBUG');
                        env$1().set('DEBUG', false);
                        const underflowCheckValue = this.abs(scalar(1e-8)).dataSync()[0];
                        env$1().set('DEBUG', debugFlag);
                        if (underflowCheckValue > 0) {
                            return 32;
                        }
                    }
                    return 16;
                });
            }
            return this.floatPrecisionValue;
        }
        /** Returns the smallest representable number.  */
        epsilon() {
            return this.floatPrecision() === 32 ? EPSILON_FLOAT32 : EPSILON_FLOAT16;
        }
        uploadToGPU(dataId) {
            const texData = this.texData.get(dataId);
            const { shape, dtype, values, texture, usage, isPacked } = texData;
            if (texture != null) {
                // Array is already on GPU. No-op.
                return;
            }
            const shouldTimeProgram = this.activeTimers != null;
            let start;
            if (shouldTimeProgram) {
                start = now();
            }
            let texShape = texData.texShape;
            if (texShape == null) {
                // This texShape may not be the final texture shape. For packed or dense
                // textures, the texShape will be changed when textures are created.
                texShape = getTextureShapeFromLogicalShape(shape, isPacked);
                texData.texShape = texShape;
            }
            if (values != null) {
                const shapeAs3D = getShapeAs3D(shape);
                let program;
                let width = texShape[1], height = texShape[0];
                const isByteArray = values instanceof Uint8Array || values instanceof Uint8ClampedArray;
                // texture for float array is PhysicalTextureType.PACKED_2X2_FLOAT32, we
                // need to make sure the upload uses the same packed size
                if (isPacked || !isByteArray) {
                    [width, height] = getPackedMatrixTextureShapeWidthHeight(texShape[0], texShape[1]);
                }
                if (isPacked) {
                    program = new EncodeMatrixPackedProgram(shapeAs3D, isByteArray);
                }
                else {
                    program = new EncodeMatrixProgram(shapeAs3D, isByteArray);
                }
                // TexShape for float array needs to be the original shape, which byte
                // array needs to be packed size. This allow the data upload shape to be
                // matched with texture creation logic.
                const tempDenseInputTexShape = isByteArray ? [height, width] : texShape;
                const tempDenseInputHandle = this.makeTensorInfo(tempDenseInputTexShape, dtype);
                const tempDenseInputTexData = this.texData.get(tempDenseInputHandle.dataId);
                if (isByteArray) {
                    tempDenseInputTexData.usage = TextureUsage.PIXELS;
                }
                else {
                    tempDenseInputTexData.usage = TextureUsage.UPLOAD;
                }
                tempDenseInputTexData.texShape = tempDenseInputTexShape;
                this.gpgpu.uploadDenseMatrixToTexture(this.getTexture(tempDenseInputHandle.dataId), width, height, values);
                const customValues = [[height, width]];
                // We want the output to remain packed regardless of the value of
                // WEBGL_PACK.
                const preventEagerUnpacking = true;
                const encodedOutputTarget = this.runWebGLProgram(program, [tempDenseInputHandle], dtype, customValues, preventEagerUnpacking);
                // Have the original texture assume the identity of the encoded output.
                const outputTexData = this.texData.get(encodedOutputTarget.dataId);
                texData.texShape = outputTexData.texShape;
                texData.isPacked = outputTexData.isPacked;
                texData.usage = outputTexData.usage;
                if (!env$1().get('ENGINE_COMPILE_ONLY')) {
                    texData.texture = outputTexData.texture;
                    // Once uploaded, don't store the values on cpu.
                    texData.values = null;
                    this.texData.delete(encodedOutputTarget.dataId);
                }
                else {
                    this.disposeData(encodedOutputTarget.dataId);
                }
                this.disposeIntermediateTensorInfo(tempDenseInputHandle);
                if (shouldTimeProgram) {
                    this.uploadWaitMs += now() - start;
                }
            }
            else {
                const newTexture = this.acquireTexture(texShape, usage, dtype, isPacked);
                texData.texture = newTexture;
            }
        }
        convertAndCacheOnCPU(dataId, float32Values) {
            const texData = this.texData.get(dataId);
            const { dtype } = texData;
            this.releaseGPUData(dataId);
            if (float32Values != null) {
                texData.values = float32ToTypedArray(float32Values, dtype);
            }
            return texData.values;
        }
        acquireTexture(texShape, texType, dtype, isPacked) {
            this.numBytesInGPU += this.computeBytes(texShape, dtype);
            if (!this.warnedAboutMemory &&
                this.numBytesInGPU > this.numMBBeforeWarning * 1024 * 1024) {
                const mb = (this.numBytesInGPU / 1024 / 1024).toFixed(2);
                this.warnedAboutMemory = true;
                console.warn(`High memory usage in GPU: ${mb} MB, ` +
                    `most likely due to a memory leak`);
            }
            return this.textureManager.acquireTexture(texShape, texType, isPacked);
        }
        computeBytes(shape, dtype) {
            return shape[0] * shape[1] * bytesPerElement(dtype);
        }
        checkCompileCompletion() {
            for (const [, binary] of Object.entries(this.binaryCache)) {
                this.checkCompletion_(binary);
            }
        }
        async checkCompileCompletionAsync() {
            const ps = [];
            if (this.gpgpu.parallelCompilationExtension) {
                for (const [, binary] of Object.entries(this.binaryCache)) {
                    ps.push(this.checkCompletionAsync_(binary));
                }
                return Promise.all(ps);
            }
            else {
                for (const [, binary] of Object.entries(this.binaryCache)) {
                    const p = new Promise((resolve) => {
                        try {
                            this.checkCompletion_(binary);
                            resolve(true);
                        }
                        catch (error) {
                            throw error;
                        }
                    });
                    ps.push(p);
                }
                return Promise.all(ps);
            }
        }
        async checkCompletionAsync_(binary) {
            if (this.gpgpu.gl.getProgramParameter(binary.webGLProgram, this.gpgpu.parallelCompilationExtension.COMPLETION_STATUS_KHR)) {
                return this.checkCompletion_(binary);
            }
            else {
                await nextFrame();
                return this.checkCompletionAsync_(binary);
            }
        }
        checkCompletion_(binary) {
            if (this.gpgpu.gl.getProgramParameter(binary.webGLProgram, this.gpgpu.gl.LINK_STATUS) === false) {
                console.log(this.gpgpu.gl.getProgramInfoLog(binary.webGLProgram));
                if (this.gpgpu.gl.getShaderParameter(binary.fragmentShader, this.gpgpu.gl.COMPILE_STATUS) === false) {
                    logShaderSourceAndInfoLog(binary.source, this.gpgpu.gl.getShaderInfoLog(binary.fragmentShader));
                    throw new Error('Failed to compile fragment shader.');
                }
                throw new Error('Failed to link vertex and fragment shaders.');
            }
            return true;
        }
        getUniformLocations() {
            for (const [, binary] of Object.entries(this.binaryCache)) {
                const { uniformLocations, customUniformLocations, infLoc, nanLoc, inShapesLocations, inTexShapesLocations, outShapeLocation, outShapeStridesLocation, outTexShapeLocation } = getUniformLocations(this.gpgpu, binary.program, binary.webGLProgram);
                binary.uniformLocations = uniformLocations;
                binary.customUniformLocations = customUniformLocations;
                binary.infLoc = infLoc;
                binary.nanLoc = nanLoc;
                binary.inShapesLocations = inShapesLocations;
                binary.inTexShapesLocations = inTexShapesLocations;
                binary.outShapeLocation = outShapeLocation;
                binary.outShapeStridesLocation = outShapeStridesLocation;
                binary.outTexShapeLocation = outTexShapeLocation;
            }
        }
    }
    MathBackendWebGL.nextDataId = 0;
    function float32ToTypedArray(a, dtype) {
        if (dtype === 'float32' || dtype === 'complex64') {
            return a;
        }
        else if (dtype === 'int32' || dtype === 'bool') {
            const result = (dtype === 'int32') ? new Int32Array(a.length) :
                new Uint8Array(a.length);
            for (let i = 0; i < result.length; ++i) {
                result[i] = Math.round(a[i]);
            }
            return result;
        }
        else {
            throw new Error(`Unknown dtype ${dtype}`);
        }
    }

    /** @license See the LICENSE file. */
    // This code is auto-generated, do not modify this file!
    const version$2 = '3.17.0';

    /**
     * @license
     * Copyright 2020 Google Inc. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    if (isBrowser()) {
        registerBackend('webgl', () => new MathBackendWebGL(), 2 /* priority */);
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const CHECK_NAN_SNIPPET$1 = `
  if (isnan(a)) return a;
  if (isnan(b)) return b;
`;
    class BinaryOpProgram {
        constructor(op, aShape, bShape) {
            this.variableNames = ['A', 'B'];
            this.outputShape = assertAndGetBroadcastShape(aShape, bShape);
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            this.userCode = `
      float binaryOperation(float a, float b) {
        ${op}
      }

      void main() {
        float a = getAAtOutCoords();
        float b = getBAtOutCoords();
        setOutput(binaryOperation(a, b));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const CHECK_NAN_SNIPPET = `
  result.r = isNaN.r > 0. ? NAN : result.r;
  result.g = isNaN.g > 0. ? NAN : result.g;
  result.b = isNaN.b > 0. ? NAN : result.b;
  result.a = isNaN.a > 0. ? NAN : result.a;
`;
    class BinaryOpPackedProgram {
        constructor(op, aShape, bShape, checkOutOfBounds = false) {
            this.variableNames = ['A', 'B'];
            this.supportsBroadcasting = true;
            this.packedInputs = true;
            this.packedOutput = true;
            this.outputShape = assertAndGetBroadcastShape(aShape, bShape);
            const rank = this.outputShape.length;
            this.enableShapeUniforms = useShapeUniforms(rank);
            let checkOutOfBoundsString = '';
            if (checkOutOfBounds) {
                if (rank === 0 || sizeFromShape(this.outputShape) === 1) {
                    checkOutOfBoundsString = `
          result.y = 0.;
          result.z = 0.;
          result.w = 0.;
        `;
                }
                else {
                    const dtype = getCoordsDataType(rank);
                    checkOutOfBoundsString = `
          ${dtype} coords = getOutputCoords();
        `;
                    if (rank === 1) {
                        if (this.enableShapeUniforms) {
                            checkOutOfBoundsString += `
            result.y = (coords + 1) >= outShape ? 0. : result.y;
            result.z = 0.;
            result.w = 0.;
          `;
                        }
                        else {
                            checkOutOfBoundsString += `
            result.y = (coords + 1) >= ${this.outputShape[0]} ? 0. : result.y;
            result.z = 0.;
            result.w = 0.;
          `;
                        }
                    }
                    else {
                        const channels = getChannels('coords', rank);
                        if (this.enableShapeUniforms) {
                            checkOutOfBoundsString += `
            bool nextRowOutOfBounds =
              (${channels[rank - 2]} + 1) >= outShape[${rank} - 2];
            bool nextColOutOfBounds =
              (${channels[rank - 1]} + 1) >= outShape[${rank} - 1];
            result.y = nextColOutOfBounds ? 0. : result.y;
            result.z = nextRowOutOfBounds ? 0. : result.z;
            result.w = nextColOutOfBounds || nextRowOutOfBounds ? 0. : result.w;
          `;
                        }
                        else {
                            checkOutOfBoundsString += `
            bool nextRowOutOfBounds =
              (${channels[rank - 2]} + 1) >= ${this.outputShape[rank - 2]};
            bool nextColOutOfBounds =
              (${channels[rank - 1]} + 1) >= ${this.outputShape[rank - 1]};
            result.y = nextColOutOfBounds ? 0. : result.y;
            result.z = nextRowOutOfBounds ? 0. : result.z;
            result.w = nextColOutOfBounds || nextRowOutOfBounds ? 0. : result.w;
          `;
                        }
                    }
                }
            }
            this.userCode = `
      vec4 binaryOperation(vec4 a, vec4 b) {
        ${op}
      }

      void main() {
        vec4 a = getAAtOutCoords();
        vec4 b = getBAtOutCoords();

        vec4 result = binaryOperation(a, b);
        ${checkOutOfBoundsString}

        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function identity(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        backend.incRef(x.dataId);
        return { dataId: x.dataId, shape: x.shape, dtype: x.dtype };
    }
    const identityConfig = {
        kernelName: Identity$1,
        backendName: 'webgl',
        kernelFunc: identity
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * In WebGL data is stored in GPU textures which can't be efficiently copied, so
     * complex tensors share data with their real and imaginary components. Complex
     * tensors' reference to the components is tracked by refCount on the individual
     * component. The refCounts are increased by the identity call.
     *
     * When a complex tensor is disposed, it will reduce the refCount on the
     * components by calling disposeData on each.
     */
    function complex(args) {
        const { inputs, backend } = args;
        const { real, imag } = inputs;
        const complexInfo = backend.makeTensorInfo(real.shape, 'complex64');
        const complex = backend.texData.get(complexInfo.dataId);
        const realTensorInfo = identity({ inputs: { x: real }, backend });
        const imagTensorInfo = identity({ inputs: { x: imag }, backend });
        complex.complexTensorInfos = { real: realTensorInfo, imag: imagTensorInfo };
        return complexInfo;
    }
    const complexConfig = {
        kernelName: Complex,
        backendName: 'webgl',
        kernelFunc: complex
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const LEAKYRELU = `return (a < 0.) ? b * a : a;`;
    const LEAKYRELU_PACKED = `
  vec4 aLessThanZero = vec4(lessThan(a, vec4(0.)));
  return (aLessThanZero * (b * a)) + ((vec4(1.0) - aLessThanZero) * a);
`;
    function leakyRelu(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { alpha } = attrs;
        const $alpha = backend.makeTensorInfo([], 'float32', createScalarValue(alpha, 'float32'));
        const program = env$1().getBool('WEBGL_PACK_BINARY_OPERATIONS') ?
            new BinaryOpPackedProgram(LEAKYRELU_PACKED, x.shape, $alpha.shape) :
            new BinaryOpProgram(LEAKYRELU, x.shape, $alpha.shape);
        const result = backend.runWebGLProgram(program, [x, $alpha], 'float32');
        backend.disposeIntermediateTensorInfo($alpha);
        return result;
    }
    const leakyReluConfig = {
        kernelName: LeakyRelu,
        backendName: 'webgl',
        kernelFunc: leakyRelu
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const PRELU = `return (a < 0.) ? b * a : a;`;
    const PRELU_PACKED = `
  vec4 aLessThanZero = vec4(lessThan(a, vec4(0.)));
  return (aLessThanZero * (b * a)) + ((vec4(1.0) - aLessThanZero) * a);
`;
    function prelu(args) {
        const { inputs, backend } = args;
        const { x, alpha } = inputs;
        const program = env$1().getBool('WEBGL_PACK_BINARY_OPERATIONS') ?
            new BinaryOpPackedProgram(PRELU_PACKED, x.shape, alpha.shape) :
            new BinaryOpProgram(PRELU, x.shape, alpha.shape);
        return backend.runWebGLProgram(program, [x, alpha], 'float32');
    }
    const preluConfig = {
        kernelName: Prelu,
        backendName: 'webgl',
        kernelFunc: prelu
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const CHECK_NAN_SNIPPET_UNARY = `if (isnan(x)) return x;`;
    const CHECK_NAN_SNIPPET_BINARY = `
  if (isnan(a)) return a;
  if (isnan(b)) return b;
`;
    const CHECK_NAN_SNIPPET_BINARY_PACKED = `
  result.r = isNaN.r > 0. ? NAN : result.r;
  result.g = isNaN.g > 0. ? NAN : result.g;
  result.b = isNaN.b > 0. ? NAN : result.b;
  result.a = isNaN.a > 0. ? NAN : result.a;
`;
    /**
     * Template that creates a `KernelFunc` for unary ops.
     * @param opSnippet Op snippet to create `UnaryOpProgram`.
     * @param packedOpSnippet Op snippet to create `UnaryOpPackedProgram`.
     * @param dtype Optional. If set, the result has this dtype. Otherwise, the
     *     result has the same dtype as the first input. This is mainly used in
     *     comparison kernels, such as Equal, Less, Greater, etc.
     */
    function unaryKernelFunc({ opSnippet, packedOpSnippet, cpuKernelImpl, dtype }) {
        return ({ inputs, backend }) => {
            const { x } = inputs;
            const webglBackend = backend;
            const $dtype = dtype || x.dtype;
            if (webglBackend.shouldExecuteOnCPU([x]) && cpuKernelImpl != null) {
                const xData = webglBackend.texData.get(x.dataId);
                const outValues = cpuKernelImpl(xData.values, $dtype);
                return webglBackend.makeTensorInfo(x.shape, $dtype, outValues);
            }
            const shouldUsePackedProgram = env$1().getBool('WEBGL_PACK_UNARY_OPERATIONS') && packedOpSnippet != null;
            let program;
            if (shouldUsePackedProgram) {
                program = new UnaryOpPackedProgram(x.shape, packedOpSnippet);
            }
            else {
                program = new UnaryOpProgram(x.shape, opSnippet);
            }
            return webglBackend.runWebGLProgram(program, [x], $dtype);
        };
    }
    /**
     * Template that creates a `KernelFunc` for binary ops.
     * @param opSnippet Op snippet to create `BinaryOpProgram`.
     * @param packedOpSnippet Op snippet to create `BinaryOpPackedProgram`.
     * @param checkOutOfBoundsForPackedProgram Whether to set checkOutOfBounds=true
     *     when creating BinaryOpPackedProgram.
     * @param dtype Optional. If set, the result has this dtype. Otherwise, the
     *     result has the same dtype as the first input. This is mainly used in
     *     comparison kernels, such as Equal, Less, Greater, etc.
     */
    function binaryKernelFunc({ opSnippet, packedOpSnippet, checkOutOfBounds = false, supportsComplex = false, cpuKernelImpl, dtype }) {
        return ({ inputs, backend }) => {
            const { a, b } = inputs;
            const webglBackend = backend;
            if (supportsComplex && a.dtype === 'complex64') {
                const aData = webglBackend.texData.get(a.dataId);
                const bData = webglBackend.texData.get(b.dataId);
                const [real, imag] = [
                    [aData.complexTensorInfos.real, bData.complexTensorInfos.real],
                    [aData.complexTensorInfos.imag, bData.complexTensorInfos.imag]
                ].map(complexParts => {
                    const [aPart, bPart] = complexParts;
                    const aHandle = {
                        dataId: aPart.dataId,
                        dtype: aPart.dtype,
                        shape: a.shape
                    };
                    const bHandle = {
                        dataId: bPart.dataId,
                        dtype: bPart.dtype,
                        shape: b.shape
                    };
                    const program = new BinaryOpProgram(opSnippet, a.shape, b.shape);
                    return webglBackend.runWebGLProgram(program, [aHandle, bHandle], upcastType(aPart.dtype, bPart.dtype));
                });
                const complexOutput = complex({ inputs: { real, imag }, backend: webglBackend });
                webglBackend.disposeIntermediateTensorInfo(real);
                webglBackend.disposeIntermediateTensorInfo(imag);
                // TODO(annxingyuan): Implement CPU forwarding for complex inputs.
                return complexOutput;
            }
            const $dtype = dtype || upcastType(a.dtype, b.dtype);
            if ((a.dtype === 'string' || b.dtype === 'string' ||
                webglBackend.shouldExecuteOnCPU([a, b])) &&
                cpuKernelImpl != null) {
                const aVals = webglBackend.texData.get(a.dataId).values;
                const bVals = webglBackend.texData.get(b.dataId).values;
                const decodedAVals = a.dtype === 'string' ?
                    // tslint:disable-next-line: no-any
                    fromUint8ToStringArray(aVals) :
                    aVals;
                const decodedBVals = a.dtype === 'string' ?
                    // tslint:disable-next-line: no-any
                    fromUint8ToStringArray(bVals) :
                    bVals;
                const [outValues, outShape] = cpuKernelImpl(a.shape, b.shape, decodedAVals, decodedBVals, $dtype);
                const out = webglBackend.makeTensorInfo(outShape, $dtype);
                const outData = webglBackend.texData.get(out.dataId);
                outData.values = outValues;
                return out;
            }
            const shouldUsePackedProgram = env$1().getBool('WEBGL_PACK_BINARY_OPERATIONS') &&
                packedOpSnippet != null;
            let program;
            if (shouldUsePackedProgram) {
                program = new BinaryOpPackedProgram(packedOpSnippet, a.shape, b.shape, checkOutOfBounds);
            }
            else {
                program = new BinaryOpProgram(opSnippet, a.shape, b.shape);
            }
            return webglBackend.runWebGLProgram(program, [a, b], $dtype);
        };
    }
    function mapActivationToShaderProgram(activation, packed = false) {
        if (activation === 'linear') {
            if (packed) {
                return LINEAR;
            }
            return LINEAR$1;
        }
        else if (activation === 'relu') {
            if (packed) {
                return RELU$1;
            }
            return RELU$2;
        }
        else if (activation === 'elu') {
            if (packed) {
                return ELU$1;
            }
            return ELU$2;
        }
        else if (activation === 'relu6') {
            if (packed) {
                return RELU6$1;
            }
            return RELU6$2;
        }
        else if (activation === 'prelu') {
            if (packed) {
                return PRELU_PACKED;
            }
            return PRELU;
        }
        else if (activation === 'leakyrelu') {
            if (packed) {
                return LEAKYRELU_PACKED;
            }
            return LEAKYRELU;
        }
        else if (activation === 'sigmoid') {
            if (packed) {
                return SIGMOID$1;
            }
            return SIGMOID$2;
        }
        throw new Error(`Activation ${activation} has not been implemented for the WebGL backend.`);
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class MatMulPackedProgram {
        constructor(aShape, bShape, outputShape, transposeA = false, transposeB = false, addBias = false, activation = null, hasPreluActivation = false, hasLeakyreluActivation = false) {
            this.variableNames = ['matrixA', 'matrixB'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.outputShape = outputShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            const sharedDim = transposeA ? aShape[1] : aShape[2];
            const sharedDimensionPacked = Math.ceil(sharedDim / 2);
            const aSample = transposeA ? 'i * 2, rc.y' : 'rc.y, i * 2';
            const bSample = transposeB ? 'rc.z, i * 2' : 'i * 2, rc.z';
            const aSwizzle = transposeA ? ['a.xxyy', 'a.zzww'] : ['a.xxzz', 'a.yyww'];
            const bSwizzle = transposeB ? ['b.xzxz', 'b.ywyw'] : ['b.xyxy', 'b.zwzw'];
            let activationSnippet = '', applyActivationSnippet = '';
            if (activation) {
                if (hasPreluActivation) {
                    activationSnippet = `vec4 activation(vec4 a) {
          vec4 b = getPreluActivationWeightsAtOutCoords();
          ${activation}
        }`;
                }
                else if (hasLeakyreluActivation) {
                    activationSnippet = `vec4 activation(vec4 a) {
          vec4 b = getLeakyreluAlphaAtOutCoords();
          ${activation}
        }`;
                }
                else {
                    activationSnippet = `vec4 activation(vec4 x) {
          ${activation}
        }`;
                }
                applyActivationSnippet = `result = activation(result);`;
            }
            const addBiasSnippet = addBias ? 'result += getBiasAtOutCoords();' : '';
            if (addBias) {
                this.variableNames.push('bias');
            }
            if (hasPreluActivation) {
                this.variableNames.push('preluActivationWeights');
            }
            if (hasLeakyreluActivation) {
                this.variableNames.push('leakyreluAlpha');
            }
            let batchASnippet = 'rc.x';
            let batchBSnippet = 'rc.x';
            if (aShape[0] < bShape[0]) {
                batchASnippet = `int(min(float(rc.x), ${aShape[0] - 1}.))`;
            }
            else if (bShape[0] < aShape[0]) {
                batchBSnippet = `int(min(float(rc.x), ${bShape[0] - 1}.))`;
            }
            this.userCode = `
      ${activationSnippet}
      // Don't use uniform for sharedDimensionPacked for performance.
      const float sharedDimension = ${sharedDimensionPacked}.0;

      vec4 dot2x2ARowBCol(ivec3 rc) {
        vec4 result = vec4(0);
        for (int i = 0; i < ${sharedDimensionPacked}; i++) {
          int batchA = ${batchASnippet};
          int batchB = ${batchBSnippet};
          vec4 a = getMatrixA(batchA, ${aSample});
          vec4 b = getMatrixB(batchB, ${bSample});

          // These swizzled products need to be separately added.
          // See: https://github.com/tensorflow/tfjs/issues/1735
          result += (${aSwizzle[0]} * ${bSwizzle[0]});
          result += (${aSwizzle[1]} * ${bSwizzle[1]});
        }
        return result;
      }

      void main() {
        ivec3 rc = getOutputCoords();
        vec4 result = dot2x2ARowBCol(rc);

        ${addBiasSnippet}

        ${applyActivationSnippet}

        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // (Ar + Ai)(Br + Bi) =
    // ArBr + ArBi + AiBr + AiBi = ArBr - AB + ArBi + AiBr
    // Yr = ArBr - AB
    // Yi = ArBi + AiBr
    const COMPLEX_MULTIPLY = {
        REAL: 'return areal * breal - aimag * bimag;',
        IMAG: 'return areal * bimag + aimag * breal;'
    };
    class BinaryOpComplexProgram {
        constructor(op, aShape, bShape) {
            this.variableNames = ['AReal', 'AImag', 'BReal', 'BImag'];
            this.outputShape = assertAndGetBroadcastShape(aShape, bShape);
            this.userCode = `
      float binaryOpComplex(
          float areal, float aimag, float breal, float bimag) {
        ${op}
      }

      void main() {
        float areal = getARealAtOutCoords();
        float aimag = getAImagAtOutCoords();
        float breal = getBRealAtOutCoords();
        float bimag = getBImagAtOutCoords();
        setOutput(binaryOpComplex(areal, aimag, breal, bimag));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const MUL = 'return a * b;';
    function multiply(args) {
        const { inputs, backend } = args;
        const { a, b } = inputs;
        const dtype = upcastType(a.dtype, b.dtype);
        if (a.dtype === 'complex64') {
            const aData = backend.texData.get(a.dataId);
            const bData = backend.texData.get(b.dataId);
            const realProgram = new BinaryOpComplexProgram(COMPLEX_MULTIPLY.REAL, a.shape, b.shape);
            const imagProgram = new BinaryOpComplexProgram(COMPLEX_MULTIPLY.IMAG, a.shape, b.shape);
            const inputs = [
                {
                    dataId: aData.complexTensorInfos.real.dataId,
                    dtype: aData.complexTensorInfos.real.dtype,
                    shape: a.shape
                },
                {
                    dataId: aData.complexTensorInfos.imag.dataId,
                    dtype: aData.complexTensorInfos.imag.dtype,
                    shape: a.shape
                },
                {
                    dataId: bData.complexTensorInfos.real.dataId,
                    dtype: bData.complexTensorInfos.real.dtype,
                    shape: b.shape
                },
                {
                    dataId: bData.complexTensorInfos.imag.dataId,
                    dtype: bData.complexTensorInfos.imag.dtype,
                    shape: b.shape
                }
            ];
            const realPart = backend.runWebGLProgram(realProgram, inputs, 'float32');
            const imagPart = backend.runWebGLProgram(imagProgram, inputs, 'float32');
            const complexOutput = complex({ inputs: { real: realPart, imag: imagPart }, backend });
            backend.disposeIntermediateTensorInfo(realPart);
            backend.disposeIntermediateTensorInfo(imagPart);
            // TODO(annxingyuan): CPU forwarding for complex inputs.
            return complexOutput;
        }
        if (backend.shouldExecuteOnCPU([a, b])) {
            const aData = backend.texData.get(a.dataId);
            const bData = backend.texData.get(b.dataId);
            const [outValues, outShape] = multiplyImplCPU(a.shape, b.shape, aData.values, bData.values, dtype);
            const out = backend.makeTensorInfo(outShape, dtype);
            const outData = backend.texData.get(out.dataId);
            outData.values = outValues;
            return out;
        }
        let program;
        if (env$1().getBool('WEBGL_PACK_BINARY_OPERATIONS')) {
            program = new BinaryOpPackedProgram(MUL, a.shape, b.shape);
        }
        else {
            program = new BinaryOpProgram(MUL, a.shape, b.shape);
        }
        return backend.runWebGLProgram(program, [a, b], dtype);
    }
    const multiplyConfig = {
        kernelName: Multiply$1,
        backendName: 'webgl',
        kernelFunc: multiply
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function packedReshape(input, afterShape, backend) {
        const input3DShape = [getBatchDim(input.shape),
            ...getRowsCols(input.shape)];
        const input3D = {
            dtype: input.dtype,
            shape: input3DShape,
            dataId: input.dataId
        };
        const afterShapeAs3D = [getBatchDim(afterShape),
            ...getRowsCols(afterShape)];
        const program = new ReshapePackedProgram(afterShapeAs3D, input3DShape);
        const preventEagerUnpackingOfOutput = true;
        const customValues = [input3DShape];
        const output = backend.runWebGLProgram(program, [input3D], input.dtype, customValues, preventEagerUnpackingOfOutput);
        return { dataId: output.dataId, shape: afterShape, dtype: output.dtype };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function reshape(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { shape } = attrs;
        const webglBackend = backend;
        const xSize = sizeFromShape(x.shape);
        const $shape = inferFromImplicitShape(shape, xSize);
        const $xSize = sizeFromShape($shape);
        assert$4(xSize === $xSize, () => `The new shape (${$shape}) has ${$xSize} elements and the old ` +
            `shape (${x.shape}) has ${xSize} elements. The new shape and old ` +
            `shape must have the same number of elements.`);
        const xTexData = webglBackend.texData.get(x.dataId);
        if (xTexData.isPacked && !isReshapeFree(x.shape, $shape) &&
            !(xTexData.texture !== null && isReshapeFree(xTexData.shape, $shape))) {
            return packedReshape(x, $shape, webglBackend);
        }
        webglBackend.incRef(x.dataId);
        return { dataId: x.dataId, shape: $shape, dtype: x.dtype };
    }
    const reshapeConfig = {
        kernelName: Reshape$1,
        backendName: 'webgl',
        kernelFunc: reshape
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class MeanProgram {
        constructor(reduceInfo, divisor) {
            this.variableNames = ['x'];
            const { windowSize, batchSize, inSize, outSize } = reduceInfo;
            this.outputShape = [batchSize, outSize];
            const windowSizeNearestVec4 = Math.floor(windowSize / 4) * 4;
            const windowSizeVec4Remainder = windowSize % 4;
            let updateSnippet = `sumValue += dot(values, ones);`;
            if (divisor != null) {
                const denominator = 1 / divisor;
                updateSnippet = `sumValue += dot(values * ${isInt(denominator) ? denominator.toPrecision(2) :
                denominator}, ones);`;
            }
            let checkOutOfBounds = '';
            if (inSize % windowSize > 0) {
                checkOutOfBounds = `
        if (inIdx < 0 || inIdx >= ${inSize}) {
          return 0.0;
        }
      `;
            }
            this.userCode = `
      const vec4 ones = vec4(1.0, 1.0, 1.0, 1.0);

      float getValue(int batch, int inIdx) {
        ${checkOutOfBounds}
        return getX(batch, inIdx);
      }

      void main() {
        ivec2 coords = getOutputCoords();
        int batch = coords[0];
        int outIdx = coords[1];
        int inOffset = outIdx * ${windowSize};

        float sumValue = 0.0;

        for (int i = 0; i < ${windowSizeNearestVec4}; i += 4) {
          int inIdx = inOffset + i;
          vec4 values = vec4(
            getValue(batch, inIdx),
            getValue(batch, inIdx + 1),
            getValue(batch, inIdx + 2),
            getValue(batch, inIdx + 3)
          );

          ${updateSnippet}
        }

        int inIdx = inOffset + ${windowSizeNearestVec4};
        if (${windowSizeVec4Remainder === 1}) {
          vec4 values = vec4(getValue(batch, inIdx), 0.0, 0.0, 0.0);

          ${updateSnippet}
        } else if (${windowSizeVec4Remainder === 2}) {
          vec4 values = vec4(
            getValue(batch, inIdx),
            getValue(batch, inIdx + 1), 0.0, 0.0);

          ${updateSnippet}
        } else if (${windowSizeVec4Remainder === 3}) {
          vec4 values = vec4(
            getValue(batch, inIdx),
            getValue(batch, inIdx + 1),
            getValue(batch, inIdx + 2), 0.0);

          ${updateSnippet}
        }
        setOutput(sumValue);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ReduceProgram {
        constructor(reduceInfo, reduceType) {
            this.variableNames = ['x'];
            const { windowSize, batchSize, inSize, outSize } = reduceInfo;
            this.outputShape = [batchSize, outSize];
            let initializationValue = '0.0';
            let compareOp = ``;
            if (reduceType === 'prod') {
                initializationValue = '1.0';
            }
            else if (reduceType === 'min') {
                // WebGL on Firefox Linux can't compile 1/0 so we do 1/eps.
                initializationValue = '1.0 / 1e-20';
                compareOp = `min`;
            }
            else if (reduceType === 'max') {
                // WebGL on Firefox Linux can't compile 1/0 so we do 1/eps.
                initializationValue = '-1.0 / 1e-20';
                compareOp = `max`;
            }
            let returnValue = `${reduceType}(${reduceType}(${reduceType}(` +
                'minMaxValue[0], minMaxValue[1]), minMaxValue[2]), minMaxValue[3])';
            if (reduceType === 'sum') {
                returnValue = `sumValue`;
            }
            else if (reduceType === 'prod') {
                returnValue = `prodValue`;
            }
            else if (reduceType === 'all') {
                returnValue = `allValue`;
            }
            else if (reduceType === 'any') {
                returnValue = `anyValue`;
            }
            const windowSizeNearestVec4 = Math.floor(windowSize / 4) * 4;
            const windowSizeVec4Remainder = windowSize % 4;
            let updateSnippet = `
      if (${reduceType === 'sum'}) {
        sumValue += dot(values, ones);
      } else if (${reduceType === 'prod'}) {
        vec2 tmp = vec2(values[0], values[1]) * vec2(values[2], values[3]);
        prodValue *= tmp[0] * tmp[1];
      } else {
        minMaxValue = ${compareOp}(values, minMaxValue);
        if (${reduceType === 'min'} || ${reduceType === 'max'}) {
          minMaxValue = ${compareOp}(values, minMaxValue);
          bvec4 isNaN = isnan(values);
          if (isNaN.r || isNaN.g || isNaN.b || isNaN.a) {
            minMaxValue = vec4(NAN);
          }
        }
      }
    `;
            let vecType = `vec4`;
            if (reduceType === 'all') {
                initializationValue = '1.0';
                updateSnippet = `
        bool reducedAllValue = all(values);
        float floatedReducedAllValue = float(reducedAllValue);
        allValue = float(allValue >= 1.0 && floatedReducedAllValue >= 1.0);
      `;
                vecType = `bvec4`;
            }
            else if (reduceType === 'any') {
                initializationValue = '0.0';
                updateSnippet = `
        bool reducedAnyValue = any(values);
        float floatedReducedAnyValue = float(reducedAnyValue);
        anyValue = float(anyValue >= 1.0 || floatedReducedAnyValue >= 1.0);
      `;
                vecType = `bvec4`;
            }
            let checkOutOfBounds = '';
            if (inSize % windowSize > 0) {
                checkOutOfBounds = `
        if (inIdx < 0 || inIdx >= ${inSize}) {
          return initializationValue;
        }
      `;
            }
            this.userCode = `
      const float initializationValue = ${initializationValue};
      const vec4 ones = vec4(1.0, 1.0, 1.0, 1.0);

      float getValue(int batch, int inIdx) {
        ${checkOutOfBounds}
        return getX(batch, inIdx);
      }

      void main() {
        ivec2 coords = getOutputCoords();
        int batch = coords[0];
        int outIdx = coords[1];
        int inOffset = outIdx * ${windowSize};

        vec4 minMaxValue = vec4(${initializationValue});
        float prodValue = 1.0;
        float sumValue = 0.0;
        float allValue = 1.0;
        float anyValue = 0.0;

        for (int i = 0; i < ${windowSizeNearestVec4}; i += 4) {
          int inIdx = inOffset + i;
          ${vecType} values = ${vecType}(
            getValue(batch, inIdx),
            getValue(batch, inIdx + 1),
            getValue(batch, inIdx + 2),
            getValue(batch, inIdx + 3)
          );

          ${updateSnippet}
        }

        int inIdx = inOffset + ${windowSizeNearestVec4};
        if (${windowSizeVec4Remainder === 1}) {
          ${vecType} values = ${vecType}(
            getValue(batch, inIdx),
            initializationValue,
            initializationValue,
            initializationValue
          );

          ${updateSnippet}
        } else if (${windowSizeVec4Remainder === 2}) {
          ${vecType} values = ${vecType}(
            getValue(batch, inIdx),
            getValue(batch, inIdx + 1),
            initializationValue,
            initializationValue
          );

          ${updateSnippet}
        } else if (${windowSizeVec4Remainder === 3}) {
          ${vecType} values = ${vecType}(
            getValue(batch, inIdx),
            getValue(batch, inIdx + 1),
            getValue(batch, inIdx + 2),
            initializationValue
          );

          ${updateSnippet}
        }
        setOutput(${returnValue});
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Returns an array of configuration objects that describe each stage of the
    // reduction.
    function getReductionStages(inShape) {
        const stages = [];
        while (stages.length === 0 || stages[stages.length - 1].outSize !== 1) {
            const outSize = stages.length ? stages[stages.length - 1].outSize : inShape[1];
            const windowSize = computeOptimalWindowSize(outSize);
            stages.push({
                inSize: outSize,
                windowSize,
                outSize: Math.ceil(outSize / windowSize)
            });
        }
        return stages;
    }
    function reduce(x, dtype, reductionType, backend) {
        const reductionStages = getReductionStages(x.shape);
        let result = x;
        for (let i = 0; i < reductionStages.length; i++) {
            const { inSize, windowSize, outSize } = reductionStages[i];
            let program;
            let previousResult;
            if (reductionType === 'mean') {
                program = i === 0 ?
                    new MeanProgram({ windowSize, inSize, batchSize: x.shape[0], outSize }, inSize) :
                    new MeanProgram({ windowSize, inSize, batchSize: x.shape[0], outSize });
            }
            else {
                program = new ReduceProgram({ windowSize, inSize, batchSize: x.shape[0], outSize }, reductionType);
            }
            previousResult = result;
            result = backend.runWebGLProgram(program, [result], dtype);
            if (previousResult.dataId !== x.dataId) {
                backend.disposeIntermediateTensorInfo(previousResult);
            }
        }
        return result;
    }

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class TransposeProgram {
        constructor(aShape, newDim) {
            this.variableNames = ['A'];
            const outputShape = new Array(aShape.length);
            for (let i = 0; i < outputShape.length; i++) {
                outputShape[i] = aShape[newDim[i]];
            }
            this.outputShape = outputShape;
            this.rank = outputShape.length;
            const dtype = getCoordsDataType(this.rank);
            const switched = getSwitchedCoords(newDim);
            this.userCode = `
    void main() {
      ${dtype} resRC = getOutputCoords();
      setOutput(getA(${switched}));
    }
    `;
        }
    }
    function getSwitchedCoords(newDim) {
        const rank = newDim.length;
        if (rank > 6) {
            throw Error(`Transpose for rank ${rank} is not yet supported`);
        }
        const originalOrder = ['resRC.x', 'resRC.y', 'resRC.z', 'resRC.w', 'resRC.u', 'resRC.v'];
        const switchedCoords = new Array(rank);
        for (let i = 0; i < newDim.length; i++) {
            switchedCoords[newDim[i]] = originalOrder[i];
        }
        return switchedCoords.join();
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class TransposePackedProgram {
        constructor(aShape, newDim) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = true;
            const outputShape = new Array(aShape.length);
            for (let i = 0; i < outputShape.length; i++) {
                outputShape[i] = aShape[newDim[i]];
            }
            this.outputShape = outputShape;
            this.rank = outputShape.length;
            if (this.rank > 6) {
                throw Error(`Packed transpose for rank ${this.rank} is not yet supported.`);
            }
            const dtype = getCoordsDataType(this.rank);
            const outputOrder = getVecChannels('rc', this.rank);
            const switchedOrder = new Array(this.rank);
            for (let i = 0; i < newDim.length; i++) {
                switchedOrder[newDim[i]] = outputOrder[i];
            }
            const innerDims = `vec2(${switchedOrder.slice(-2).join()})`;
            const nextColumn = `++${outputOrder[this.rank - 1]} < ${outputShape[this.rank - 1]}`;
            const getc = `getChannel(getA(${switchedOrder.join()}), ${innerDims})`;
            this.userCode = `
    void main() {
      ${dtype} rc = getOutputCoords();
      vec4 result = vec4(0.);
      result[0] = ${getc};
      if(${nextColumn}) {
        result[1] = ${getc};
      }
      --${outputOrder[this.rank - 1]};
      if(++${outputOrder[this.rank - 2]} < ${outputShape[this.rank - 2]}) {
        result[2] = ${getc};
        if(${nextColumn}) {
          result[3] = ${getc};
        }
      }
      setOutput(result);
    }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function transposeImpl(x, perm, backend) {
        const program = env$1().getBool('WEBGL_PACK_ARRAY_OPERATIONS') ?
            new TransposePackedProgram(x.shape, perm) :
            new TransposeProgram(x.shape, perm);
        return backend.runWebGLProgram(program, [x], x.dtype);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sumImpl(x, axis, keepDims, backend) {
        const reductionIndices = axis;
        const xRank = x.shape.length;
        const origAxes = parseAxisParam(reductionIndices, x.shape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, xRank);
        const sumInputIsTransposed = permutedAxes != null;
        let sumInput = x;
        if (sumInputIsTransposed) {
            sumInput = transposeImpl(x, permutedAxes, backend);
            axes = getInnerMostAxes(axes.length, xRank);
        }
        assertAxesAreInnerMostDims('sum', axes, xRank);
        const [sumOutShape, reduceShape] = computeOutAndReduceShapes(sumInput.shape, axes);
        let outShape = sumOutShape;
        if (keepDims) {
            // rather than reshape at the end, set the target shape here.
            outShape = expandShapeToKeepDim(sumOutShape, origAxes);
        }
        const inSize = sizeFromShape(reduceShape);
        const xSize = sizeFromShape(x.shape);
        const batchSize = xSize / inSize;
        const reshapedInput = reshape({ inputs: { x: sumInput }, attrs: { shape: [batchSize, inSize] }, backend });
        const outType = sumOutType(x.dtype);
        const reduced = reduce(reshapedInput, outType, 'sum', backend);
        const out = reshape({ inputs: { x: reduced }, attrs: { shape: outShape }, backend });
        backend.disposeIntermediateTensorInfo(reshapedInput);
        backend.disposeIntermediateTensorInfo(reduced);
        if (sumInputIsTransposed) {
            backend.disposeIntermediateTensorInfo(sumInput);
        }
        return out;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sum(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        return sumImpl(x, axis, keepDims, backend);
    }
    const sumConfig = {
        kernelName: Sum,
        backendName: 'webgl',
        kernelFunc: sum
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function transpose(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { perm } = attrs;
        const webglBackend = backend;
        const xRank = x.shape.length;
        const newShape = new Array(xRank);
        for (let i = 0; i < newShape.length; i++) {
            newShape[i] = x.shape[perm[i]];
        }
        let out;
        if (webglBackend.shouldExecuteOnCPU([x])) {
            const xTexData = webglBackend.texData.get(x.dataId);
            const values = xTexData.values;
            const outValues = transposeImplCPU(values, x.shape, x.dtype, perm, newShape);
            out = webglBackend.makeTensorInfo(newShape, x.dtype);
            const outData = webglBackend.texData.get(out.dataId);
            outData.values = outValues;
        }
        else {
            out = transposeImpl(x, perm, webglBackend);
        }
        return out;
    }
    const transposeConfig = {
        kernelName: Transpose,
        backendName: 'webgl',
        kernelFunc: transpose
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Empirically determined minimal shared dimension in matmul before we forward
    // to a.mul(b).sum() in order to take advantage of GPU parallelism. See
    // https://github.com/tensorflow/tfjs-core/pull/1379 for benchmarks.
    const MATMUL_SHARED_DIM_THRESHOLD = 1000;
    function batchMatMulImpl({ a, b, transposeA, transposeB, backend, bias = null, preluActivationWeights = null, leakyreluAlpha = 0, activation = null }) {
        const aRank = a.shape.length;
        const bRank = b.shape.length;
        const innerShapeA = transposeA ? a.shape[aRank - 2] : a.shape[aRank - 1];
        const innerShapeB = transposeB ? b.shape[bRank - 1] : b.shape[bRank - 2];
        const outerShapeA = transposeA ? a.shape[aRank - 1] : a.shape[aRank - 2];
        const outerShapeB = transposeB ? b.shape[bRank - 2] : b.shape[bRank - 1];
        const outerDimsA = a.shape.slice(0, -2);
        const outerDimsB = b.shape.slice(0, -2);
        const batchDimA = sizeFromShape(outerDimsA);
        const batchDimB = sizeFromShape(outerDimsB);
        const outShapeOuterDims = assertAndGetBroadcastShape(a.shape.slice(0, -2), b.shape.slice(0, -2));
        const outShape = outShapeOuterDims.concat([outerShapeA, outerShapeB]);
        assert$4(innerShapeA === innerShapeB, () => `Error in matMul: inner shapes (${innerShapeA}) and (` +
            `${innerShapeB}) of Tensors with shapes ${a.shape} and ` +
            `${b.shape} and transposeA=${transposeA}` +
            ` and transposeB=${transposeB} must match.`);
        const a3dShape = transposeA ?
            [batchDimA, innerShapeA, outerShapeA] :
            [batchDimA, outerShapeA, innerShapeA];
        const b3dShape = transposeB ?
            [batchDimB, outerShapeB, innerShapeB] :
            [batchDimB, innerShapeB, outerShapeB];
        // The rest of the implementation is designed to operate on rank-3 tensors
        const a3d = reshape({ inputs: { x: a }, backend, attrs: { shape: a3dShape } });
        const b3d = reshape({ inputs: { x: b }, backend, attrs: { shape: b3dShape } });
        const intermediates = [a3d, b3d];
        const batchDim = Math.max(batchDimA, batchDimB);
        const sharedDim = transposeA ? a3d.shape[1] : a3d.shape[2];
        const hasBias = bias != null;
        const hasPreluActivationWeights = preluActivationWeights != null;
        const hasLeakyreluAlpha = activation === 'leakyrelu';
        const fusedActivation = activation != null ?
            mapActivationToShaderProgram(activation, true) :
            null;
        const containsFusedOps = hasBias || hasPreluActivationWeights ||
            hasLeakyreluAlpha || fusedActivation != null;
        let out;
        // Since the matrices are vectors, it is faster to call mul().sum()
        // because sum() is O(sqrt(N)) due to divide-and-conquer.
        if ((outerShapeA === 1 || outerShapeB === 1) &&
            sharedDim > MATMUL_SHARED_DIM_THRESHOLD && containsFusedOps === false) {
            let aVec = a3d;
            let bVec = b3d;
            if (transposeA) {
                aVec = transpose({ inputs: { x: a3d }, backend, attrs: { perm: [0, 2, 1] } });
                intermediates.push(aVec);
            }
            if (transposeB) {
                bVec = transpose({ inputs: { x: b3d }, backend, attrs: { perm: [0, 2, 1] } });
                intermediates.push(bVec);
            }
            const shouldReshapeA = outerShapeB !== 1;
            const shouldReshapeB = outerShapeB === 1;
            let aVec3d = aVec;
            if (shouldReshapeA) {
                aVec3d = reshape({
                    inputs: { x: aVec },
                    backend,
                    attrs: { shape: [batchDim, sharedDim, 1] }
                });
                intermediates.push(aVec3d);
            }
            const axis = outerShapeB === 1 ? 2 : 1;
            let bVec3d = bVec;
            if (shouldReshapeB) {
                bVec3d = reshape({
                    inputs: { x: bVec },
                    backend,
                    attrs: { shape: [batchDim, 1, sharedDim] }
                });
                intermediates.push(bVec3d);
            }
            const product = multiply({ inputs: { a: aVec3d, b: bVec3d }, backend });
            out = sum({ inputs: { x: product }, backend, attrs: { axis, keepDims: true } });
            intermediates.push(product);
        }
        else {
            const dtype = upcastType(a.dtype, b.dtype);
            const program = new MatMulPackedProgram(a3dShape, b3dShape, [batchDim, outerShapeA, outerShapeB], transposeA, transposeB, hasBias, fusedActivation, hasPreluActivationWeights, hasLeakyreluAlpha);
            const inputs = [a3d, b3d];
            if (bias != null) {
                inputs.push(bias);
            }
            if (hasPreluActivationWeights) {
                inputs.push(preluActivationWeights);
            }
            if (hasLeakyreluAlpha) {
                const $leakyreluAlpha = backend.makeTensorInfo([], 'float32', createScalarValue(leakyreluAlpha, 'float32'));
                inputs.push($leakyreluAlpha);
                intermediates.push($leakyreluAlpha);
            }
            out = backend.runWebGLProgram(program, inputs, dtype);
        }
        const outReshaped = reshape({ inputs: { x: out }, backend, attrs: { shape: outShape } });
        intermediates.push(out);
        for (const i of intermediates) {
            backend.disposeIntermediateTensorInfo(i);
        }
        return outReshaped;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function _fusedMatMul(args) {
        const { inputs, backend, attrs } = args;
        const { a, b, bias, preluActivationWeights } = inputs;
        const { transposeA, transposeB, activation, leakyreluAlpha } = attrs;
        return batchMatMulImpl({
            a,
            b,
            transposeA,
            transposeB,
            backend,
            bias,
            preluActivationWeights,
            leakyreluAlpha,
            activation
        });
    }
    const _fusedMatMulConfig = {
        kernelName: _FusedMatMul,
        backendName: 'webgl',
        kernelFunc: _fusedMatMul,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ABS = `return abs(x);`;
    function abs(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        // TODO: handle cases when x is complex. Once the cpu implementation
        // can handle complex values, refactor to use unaryKernelFunc.
        if (backend.shouldExecuteOnCPU([x]) && x.dtype !== 'complex64') {
            const xData = backend.texData.get(x.dataId);
            const outValues = simpleAbsImplCPU(xData.values);
            return backend.makeTensorInfo(x.shape, x.dtype, outValues);
        }
        let program;
        if (env$1().getBool('WEBGL_PACK_UNARY_OPERATIONS')) {
            program = new UnaryOpPackedProgram(x.shape, ABS);
        }
        else {
            program = new UnaryOpProgram(x.shape, ABS);
        }
        return backend.runWebGLProgram(program, [x], x.dtype);
    }
    const absConfig = {
        kernelName: Abs,
        backendName: 'webgl',
        kernelFunc: abs
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ACOS = CHECK_NAN_SNIPPET$2 + `
  if (abs(x) > 1.) {
    return NAN;
  }
  return acos(x);
`;
    const acos = unaryKernelFunc({ opSnippet: ACOS });
    const acosConfig = {
        kernelName: Acos,
        backendName: 'webgl',
        kernelFunc: acos,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ACOSH = CHECK_NAN_SNIPPET$2 + `
  if (x < 1.0) return NAN;
return log(x + sqrt(x * x - 1.0));`;
    const acosh = unaryKernelFunc({ opSnippet: ACOSH });
    const acoshConfig = {
        kernelName: Acosh,
        backendName: 'webgl',
        kernelFunc: acosh,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ADD = 'return a + b;';
    const addKernelFunc = binaryKernelFunc({
        opSnippet: ADD,
        packedOpSnippet: ADD,
        supportsComplex: true,
        cpuKernelImpl: addImplCPU
    });
    const addConfig = {
        kernelName: Add$1,
        backendName: 'webgl',
        kernelFunc: addKernelFunc
    };

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class AddNProgram {
        constructor(outputShape, shapes) {
            this.outputShape = [];
            this.outputShape = outputShape;
            this.variableNames = shapes.map((_, i) => `T${i}`);
            const snippets = [];
            // Get target elements from every input tensor.
            this.variableNames.forEach(variable => {
                snippets.push(`float v${variable} = get${variable}AtOutCoords();`);
            });
            // Calculate the sum of all elements.
            const operation = this.variableNames
                .map(variable => {
                return `v${variable}`;
            })
                .join(' + ');
            this.userCode = `
      void main() {
        ${snippets.join('\n        ')}

        float result = ${operation};
        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class AddNPackedProgram {
        constructor(outputShape, shapes) {
            this.outputShape = [];
            this.packedInputs = true;
            this.packedOutput = true;
            this.outputShape = outputShape;
            this.variableNames = shapes.map((_, i) => `T${i}`);
            const snippets = [];
            // Get target elements from every input tensor.
            this.variableNames.forEach(variable => {
                snippets.push(`vec4 v${variable} = get${variable}AtOutCoords();`);
            });
            // Calculate the sum of all elements.
            const operation = this.variableNames
                .map(variable => {
                return `v${variable}`;
            })
                .join(' + ');
            this.userCode = `
      void main() {
        ${snippets.join('\n        ')}

        vec4 result = ${operation};
        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function addN(args) {
        const { inputs, backend } = args;
        const tensors = inputs;
        if (tensors.length === 1) {
            return identity({ inputs: { x: tensors[0] }, backend });
        }
        // Limit the number of uploaded textures for optimization.
        if (tensors.length > env$1().get('WEBGL_MAX_TEXTURES_IN_SHADER')) {
            const midIndex = Math.floor(tensors.length / 2);
            const leftSide = addN({ inputs: tensors.slice(0, midIndex), backend });
            const rightSide = addN({ inputs: tensors.slice(midIndex), backend });
            return addN({ inputs: [leftSide, rightSide], backend });
        }
        const dtype = tensors.map(t => t.dtype).reduce((d1, d2) => upcastType(d1, d2));
        const shapes = tensors.map(t => t.shape);
        // We can make sure shapes are identical in op level.
        const usePackedOp = env$1().getBool('WEBGL_PACK');
        const program = usePackedOp ?
            new AddNPackedProgram(tensors[0].shape, shapes) :
            new AddNProgram(tensors[0].shape, shapes);
        return backend.runWebGLProgram(program, tensors, dtype);
    }
    const addNConfig = {
        kernelName: AddN,
        backendName: 'webgl',
        kernelFunc: addN
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function all(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        const xRank = x.shape.length;
        const origAxes = parseAxisParam(axis, x.shape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, xRank);
        let permutedX = x;
        if (permutedAxes != null) {
            permutedX = transpose({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            axes = getInnerMostAxes(axes.length, xRank);
        }
        assertAxesAreInnerMostDims('all', axes, xRank);
        const [outShape, reduceShape] = computeOutAndReduceShapes(permutedX.shape, axes);
        const inSize = sizeFromShape(reduceShape);
        const a2D = reshape({ inputs: { x: permutedX }, backend, attrs: { shape: [-1, inSize] } });
        const reduced = reduce(a2D, a2D.dtype, 'all', backend);
        let res;
        if (keepDims) {
            const newShape = expandShapeToKeepDim(outShape, origAxes);
            res = reshape({ inputs: { x: reduced }, backend, attrs: { shape: newShape } });
        }
        else {
            res = reshape({ inputs: { x: reduced }, backend, attrs: { shape: outShape } });
        }
        backend.disposeIntermediateTensorInfo(a2D);
        backend.disposeIntermediateTensorInfo(reduced);
        if (permutedAxes != null) {
            backend.disposeIntermediateTensorInfo(permutedX);
        }
        return res;
    }
    const allConfig = {
        kernelName: All,
        backendName: 'webgl',
        kernelFunc: all
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function any(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        const xRank = x.shape.length;
        const origAxes = parseAxisParam(axis, x.shape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, xRank);
        let permutedX = x;
        if (permutedAxes != null) {
            permutedX = transpose({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            axes = getInnerMostAxes(axes.length, xRank);
        }
        assertAxesAreInnerMostDims('any', axes, xRank);
        const [outShape, reduceShape] = computeOutAndReduceShapes(permutedX.shape, axes);
        const inSize = sizeFromShape(reduceShape);
        const a2D = reshape({ inputs: { x: permutedX }, backend, attrs: { shape: [-1, inSize] } });
        const reduced = reduce(a2D, a2D.dtype, 'any', backend);
        let res;
        if (keepDims) {
            const newShape = expandShapeToKeepDim(outShape, origAxes);
            res = reshape({ inputs: { x: reduced }, backend, attrs: { shape: newShape } });
        }
        else {
            res = reshape({ inputs: { x: reduced }, backend, attrs: { shape: outShape } });
        }
        backend.disposeIntermediateTensorInfo(a2D);
        backend.disposeIntermediateTensorInfo(reduced);
        if (permutedAxes != null) {
            backend.disposeIntermediateTensorInfo(permutedX);
        }
        return res;
    }
    const anyConfig = {
        kernelName: Any,
        backendName: 'webgl',
        kernelFunc: any
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ArgMinMaxProgram {
        constructor(reduceInfo, op, firstPass) {
            this.variableNames = ['A'];
            const { windowSize, batchSize, outSize } = reduceInfo;
            if (!firstPass) {
                this.variableNames.push('bestIndicesA');
            }
            this.outputShape = [batchSize, outSize];
            const compOp = (op === 'max') ? '>' : '<';
            const indexSnippet = firstPass ?
                'inOffset + i;' :
                'round(getBestIndicesA(batch, inOffset + i));';
            this.userCode = `
      void main() {
        ivec2 coords = getOutputCoords();
        int batch = coords[0];
        int outIdx = coords[1];
        int inOffset = outIdx * ${windowSize};

        int bestIndex = inOffset;
        float bestValue = getA(batch, bestIndex);

        for (int i = 0; i < ${windowSize}; i++) {
          int inIdx = ${indexSnippet};
          float candidate = getA(batch, inIdx);
          if (candidate ${compOp} bestValue) {
            bestValue = candidate;
            bestIndex = inIdx;
          }
        }
        setOutput(float(bestIndex));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ArgMinMaxPackedProgram {
        constructor(shape, windowSize, op, firstPass) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = true;
            assert$4(shape.length > 2, () => `Packed arg${op.charAt(0).toUpperCase() +
            op.slice(1)} supports only inputs with rank above 2.`);
            const inSize = shape[shape.length - 1];
            const outSize = Math.ceil(inSize / windowSize);
            this.outputShape = shape.slice(0, -1);
            if (outSize > 1) {
                this.outputShape.push(outSize);
            }
            if (!firstPass) {
                this.variableNames.push('bestIndicesA');
            }
            const outShape = this.outputShape;
            const rank = outShape.length;
            const dtype = getCoordsDataType(rank);
            const coords = getChannels('coords', rank);
            let sourceLocSetup;
            let sourceRank;
            if (outSize === 1) {
                sourceRank = rank + 1;
                const sourceLocDType = getCoordsDataType(sourceRank);
                sourceLocSetup = `
        ${sourceLocDType} sourceLocR = ${sourceLocDType}(${coords.join()}, 0);
        ++${coords[rank - 1]};
        ${sourceLocDType} sourceLocG = ${sourceLocDType}(${coords.join()}, 0);
        ++${coords[rank - 2]};
        ${sourceLocDType} sourceLocA = ${sourceLocDType}(${coords.join()}, 0);
        --${coords[rank - 1]};
        ${sourceLocDType} sourceLocB = ${sourceLocDType}(${coords.join()}, 0);
        --${coords[rank - 2]};`;
            }
            else {
                sourceRank = rank;
                sourceLocSetup = `
        ${dtype} sourceLocR = coords;
        ++${coords[rank - 1]};
        ${dtype} sourceLocG = coords;
        ++${coords[rank - 2]};
        ${dtype} sourceLocA = coords;
        --${coords[rank - 1]};
        ${dtype} sourceLocB = coords;
        --${coords[rank - 2]};`;
            }
            const channels = ['x', 'y', 'z', 'w', 'u', 'v'].slice(0, sourceRank);
            const inChannel = '.' + channels[sourceRank - 1]; // e.g. ".b" for rank 3.
            const intChannels = channels.map(x => 'int ' + x);
            const srcRCoords = getChannels('sourceLocR', sourceRank - 1).concat('inIdx.r');
            const srcGCoords = getChannels('sourceLocG', sourceRank - 1).concat('inIdx.g');
            const srcBCoords = getChannels('sourceLocB', sourceRank - 1).concat('inIdx.b');
            const srcACoords = getChannels('sourceLocA', sourceRank - 1).concat('inIdx.a');
            const compOp = (op === 'max') ? 'greaterThan' : 'lessThan';
            const fetchCandidateIdx = firstPass ? '' : `
          inIdx = round(vec4(getBestIndicesAChannel(${srcRCoords.join()}),
                             getBestIndicesAChannel(${srcGCoords.join()}),
                             getBestIndicesAChannel(${srcBCoords.join()}),
                             getBestIndicesAChannel(${srcACoords.join()})));`;
            const fetchValue = `vec4(
            getAChannel(${srcRCoords.join()}),
            hasNextCol ? getAChannel(${srcGCoords.join()}) : 0.,
            hasNextRow ? getAChannel(${srcBCoords.join()}) : 0.,
            hasNextRow && hasNextCol ? getAChannel(${srcACoords.join()}) : 0.)`;
            const getBestIndicesAChannelSnippet = firstPass ? '' : `
      float getBestIndicesAChannel(${intChannels.join()}) {
        return getChannel(getBestIndicesA(${channels.join()}),
                                          vec2(${channels.slice(-2).join()}));
      }`;
            this.userCode = `
      float getAChannel(${intChannels.join()}) {
        return getChannel(getA(${channels.join()}),
                               vec2(${channels.slice(-2).join()}));
      }
      ${getBestIndicesAChannelSnippet}
      void main() {
        ${dtype} coords = getOutputCoords();
        bool hasNextCol = ${coords[rank - 1]} < ${outShape[rank - 1] - 1};
        bool hasNextRow = ${coords[rank - 2]} < ${outShape[rank - 2] - 1};
        ${sourceLocSetup}
        ivec4 srcIdx = ivec4(sourceLocR${inChannel}, sourceLocG${inChannel},
          sourceLocB${inChannel}, sourceLocA${inChannel}) * ${windowSize};
        ivec4 inIdx = srcIdx;
        vec4 bestIndex = vec4(inIdx);
        vec4 bestValue = ${fetchValue};

        for (int i = 0; i < ${windowSize}; i++) {
          inIdx = srcIdx;
          ${fetchCandidateIdx}
          vec4 candidate = ${fetchValue};
          bvec4 nan = isnan(candidate);
          bvec4 replace = bvec4(
            vec4(${compOp}(candidate, bestValue)) * (vec4(1.0) - vec4(nan)));

          bestValue = vec4(replace.x  ? candidate.x : bestValue.x,
                           replace.y  ? candidate.y : bestValue.y,
                           replace.z  ? candidate.z : bestValue.z,
                           replace.w  ? candidate.w : bestValue.w);
          bestIndex = mix(bestIndex, vec4(inIdx), vec4(replace));
          srcIdx++;
        }
        setOutput(bestIndex);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function argReduce(backend, x, reduceType, bestIndicesA = null) {
        let batchSize = x.shape[0];
        let inSize = x.shape[1];
        if (bestIndicesA != null) {
            batchSize = bestIndicesA.shape[0];
            inSize = bestIndicesA.shape[1];
        }
        const windowSize = computeOptimalWindowSize(inSize);
        const reduceInfo = { windowSize, inSize, batchSize, outSize: Math.ceil(inSize / windowSize) };
        const program = new ArgMinMaxProgram(reduceInfo, reduceType, bestIndicesA == null);
        const inputs = [x];
        if (bestIndicesA != null) {
            inputs.push(bestIndicesA);
        }
        const output = backend.runWebGLProgram(program, inputs, 'int32');
        // No need to run another GPGPU program.
        if (output.shape[1] === 1) {
            return output;
        }
        const result = argReduce(backend, x, reduceType, output);
        backend.disposeIntermediateTensorInfo(output);
        return result;
    }
    function argReducePacked(backend, x, reduceType, bestIndicesA = null) {
        const inShape = bestIndicesA != null ? bestIndicesA.shape : x.shape;
        const inSize = inShape[inShape.length - 1];
        const windowSize = computeOptimalWindowSize(inSize);
        const program = new ArgMinMaxPackedProgram(inShape, windowSize, reduceType, bestIndicesA == null);
        const inputs = bestIndicesA == null ? [x] : [x, bestIndicesA];
        const output = backend.runWebGLProgram(program, inputs, 'int32');
        if (output.shape.length === x.shape.length) {
            const result = argReducePacked(backend, x, reduceType, output);
            backend.disposeIntermediateTensorInfo(output);
            return result;
        }
        return output;
    }
    function argMinMaxReduce(backend, x, axis, reduceType) {
        const axes = [axis];
        assertAxesAreInnerMostDims('arg' + reduceType.charAt(0).toUpperCase() + reduceType.slice(1), axes, x.shape.length);
        if (!env$1().getBool('WEBGL_PACK_REDUCE') || x.shape.length <= 2) {
            const intermediateTensorInfos = [];
            // Eagerly unpack x input since it is passed in to all the shaders which
            // require unpacked inputs.
            const xtexData = backend.texData.get(x.dataId);
            const xIsPacked = xtexData !== null && xtexData.isPacked;
            let xUnPacked = x;
            if (xIsPacked) {
                xUnPacked = backend.unpackTensor(x);
                intermediateTensorInfos.push(xUnPacked);
            }
            const [outShape, reduceShape] = computeOutAndReduceShapes(xUnPacked.shape, axes);
            const inSize = sizeFromShape(reduceShape);
            const a2D = reshape({ inputs: { x: xUnPacked }, backend, attrs: { shape: [-1, inSize] } });
            intermediateTensorInfos.push(a2D);
            const reduced = argReduce(backend, a2D, reduceType);
            intermediateTensorInfos.push(reduced);
            const reshaped = reshape({ inputs: { x: reduced }, backend, attrs: { shape: outShape } });
            intermediateTensorInfos.forEach(t => backend.disposeIntermediateTensorInfo(t));
            return reshaped;
        }
        return argReducePacked(backend, x, reduceType);
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function argMax(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis } = attrs;
        let axes = parseAxisParam(axis, x.shape);
        const permutedAxes = getAxesPermutation(axes, x.shape.length);
        let $x = x;
        const intermediateTensorInfos = [];
        if (permutedAxes != null) {
            $x = transpose({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            intermediateTensorInfos.push($x);
            axes = getInnerMostAxes(axes.length, $x.shape.length);
        }
        assertAxesAreInnerMostDims('argMax', [axes[0]], $x.shape.length);
        const out = argMinMaxReduce(backend, $x, axes[0], 'max');
        intermediateTensorInfos.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return out;
    }
    const argMaxConfig = {
        kernelName: ArgMax,
        backendName: 'webgl',
        kernelFunc: argMax
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function argMin(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis } = attrs;
        let axes = parseAxisParam(axis, x.shape);
        const permutedAxes = getAxesPermutation(axes, x.shape.length);
        let $x = x;
        const intermediateTensorInfos = [];
        if (permutedAxes != null) {
            $x = transpose({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            intermediateTensorInfos.push($x);
            axes = getInnerMostAxes(axes.length, $x.shape.length);
        }
        assertAxesAreInnerMostDims('argMin', [axes[0]], $x.shape.length);
        const out = argMinMaxReduce(backend, $x, axes[0], 'min');
        intermediateTensorInfos.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return out;
    }
    const argMinConfig = {
        kernelName: ArgMin,
        backendName: 'webgl',
        kernelFunc: argMin
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ASIN = CHECK_NAN_SNIPPET$2 + `
  if (abs(x) > 1.) {
    return NAN;
  }
  return asin(x);
`;
    const asin = unaryKernelFunc({ opSnippet: ASIN });
    const asinConfig = {
        kernelName: Asin,
        backendName: 'webgl',
        kernelFunc: asin,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ASINH = CHECK_NAN_SNIPPET$2 + `return log(x + sqrt(x * x + 1.0));`;
    const asinh = unaryKernelFunc({ opSnippet: ASINH });
    const asinhConfig = {
        kernelName: Asinh,
        backendName: 'webgl',
        kernelFunc: asinh,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ATAN = CHECK_NAN_SNIPPET$2 + `
  return atan(x);
`;
    const atan = unaryKernelFunc({ opSnippet: ATAN });
    const atanConfig = {
        kernelName: Atan,
        backendName: 'webgl',
        kernelFunc: atan,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ATAN2 = CHECK_NAN_SNIPPET_BINARY + `
  return atan(a, b);
`;
    const ATAN2_PACKED = `
  vec4 result = atan(a, b);
  vec4 isNaN = min(vec4(isnan(a)) + vec4(isnan(b)), vec4(1.0));
  ` +
        CHECK_NAN_SNIPPET_BINARY_PACKED + `
  return result;
`;
    const atan2 = binaryKernelFunc({ opSnippet: ATAN2, packedOpSnippet: ATAN2_PACKED });
    const atan2Config = {
        kernelName: Atan2,
        backendName: 'webgl',
        kernelFunc: atan2,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ATANH = CHECK_NAN_SNIPPET$2 + `
  if ((x < -1.0) || (x > 1.0)) return NAN;
return (log(1.0 + x) - log(1.0 - x)) / 2.0;`;
    const atanh = unaryKernelFunc({ opSnippet: ATANH });
    const atanhConfig = {
        kernelName: Atanh,
        backendName: 'webgl',
        kernelFunc: atanh,
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class Pool2DProgram {
        constructor(convInfo, poolType, computePositions, flattenPositions = false, includeBatchInIndex = false) {
            this.variableNames = ['x'];
            if (poolType === 'avg' && computePositions) {
                throw new Error('Cannot compute positions for average pool.');
            }
            const filterWidth = convInfo.filterWidth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const dilationHeight = convInfo.dilationHeight;
            const dilationWidth = convInfo.dilationWidth;
            const effectiveFilterHeight = convInfo.effectiveFilterHeight;
            const effectiveFilterWidth = convInfo.effectiveFilterWidth;
            const padTop = convInfo.padInfo.top;
            const padLeft = convInfo.padInfo.left;
            this.outputShape = convInfo.outShape;
            const isAvgPool = poolType === 'avg';
            const batchFlattenPositionStr = `((batch  * ${convInfo.inHeight} + xR) * ${convInfo.inWidth} + xC) * ${convInfo.inChannels} + d`;
            const flattenPositionStr = `(xR * ${convInfo.inWidth} + xC) * ${convInfo.inChannels} + d`;
            let initializationValue = '0.0';
            if (!isAvgPool) {
                // WebGL on Firefox Linux can't compile 1/0 so we do 1/eps.
                initializationValue = '-1.0 / 1e-20';
            }
            if (computePositions) {
                const compareOp = '>=';
                this.userCode = `
        const ivec2 strides = ivec2(${strideHeight}, ${strideWidth});
        const ivec2 pads = ivec2(${padTop}, ${padLeft});

        void main() {
          ivec4 coords = getOutputCoords();
          int batch = coords[0];
          int d = coords[3];

          ivec2 xRCCorner = coords.yz * strides - pads;
          int xRCorner = xRCCorner.x;
          int xCCorner = xRCCorner.y;

          // max/min x(?, ?, d) to get y(yR, yC, d).
          // ? = to be determined
          float minMaxValue = 0.0;
          float minMaxValueFound = 0.0;
          int minMaxPosition = 0;
          float avgValue = 0.0;

          for (int wR = 0; wR < ${effectiveFilterHeight};
              wR += ${dilationHeight}) {
            int xR = xRCorner + wR;

            if (xR < 0 || xR >= ${convInfo.inHeight}) {
              continue;
            }

            for (int wC = 0; wC < ${effectiveFilterWidth};
                wC += ${dilationWidth}) {
              int xC = xCCorner + wC;

              if (xC < 0 || xC >= ${convInfo.inWidth}) {
                continue;
              }

              float value = getX(batch, xR, xC, d);

              // If a min / max value has already been found, use it. If not,
              // use the current value.
              float currMinMaxValue = mix(
                  value, minMaxValue, minMaxValueFound);
              if (value ${compareOp} currMinMaxValue) {
                minMaxValue = value;
                minMaxValueFound = 1.0;
                minMaxPosition = ${flattenPositions ? (includeBatchInIndex ? batchFlattenPositionStr :
                flattenPositionStr) :
                `wR * ${effectiveFilterWidth} + wC`};
              }
            }
          }
          setOutput(float(minMaxPosition));
        }
      `;
                return;
            }
            const compareOp = 'max';
            let returnValue = `${poolType}(${poolType}(${poolType}(` +
                'minMaxValue[0], minMaxValue[1]), minMaxValue[2]), minMaxValue[3])';
            if (poolType === 'avg') {
                returnValue = `avgValue / count`;
            }
            const filterWidthNearestVec4 = Math.floor(filterWidth / 4) * 4;
            const filterWidthVec4Remainder = filterWidth % 4;
            const updateSnippet = `
      if (${isAvgPool}) {
        avgValue += dot(values, ones);
      } else {
        minMaxValue = ${compareOp}(values, minMaxValue);
      }
    `;
            this.userCode = `
      const ivec2 strides = ivec2(${strideHeight}, ${strideWidth});
      const ivec2 pads = ivec2(${padTop}, ${padLeft});
      const float initializationValue = ${initializationValue};
      const vec4 ones = vec4(1.0, 1.0, 1.0, 1.0);

      float count = 0.0;

      float getValue(int batch, int xR, int xC, int d) {
        if (xC < 0 || xC >= ${convInfo.inWidth}) {
          return initializationValue;
        }
        count += 1.0;
        return getX(batch, xR, xC, d);
      }

      void main() {
        ivec4 coords = getOutputCoords();
        int batch = coords[0];
        int d = coords[3];

        ivec2 xRCCorner = coords.yz * strides - pads;
        int xRCorner = xRCCorner.x;
        int xCCorner = xRCCorner.y;

        // max/min x(?, ?, d) to get y(yR, yC, d).
        // ? = to be determined
        vec4 minMaxValue = vec4(${initializationValue});
        float avgValue = 0.0;
        count = 0.0;

        for (int wR = 0; wR < ${effectiveFilterHeight};
            wR += ${dilationHeight}) {
          int xR = xRCorner + wR;

          if (xR < 0 || xR >= ${convInfo.inHeight}) {
            continue;
          }

          for (int wC = 0; wC < ${filterWidthNearestVec4}; wC += 4) {
            int xC = xCCorner + wC * ${dilationWidth};

            vec4 values = vec4(
              getValue(batch, xR, xC, d),
              getValue(batch, xR, xC + ${dilationWidth}, d),
              getValue(batch, xR, xC + 2 * ${dilationWidth}, d),
              getValue(batch, xR, xC + 3 * ${dilationWidth}, d)
            );

            ${updateSnippet}
          }

          int xC = xCCorner + ${filterWidthNearestVec4};
          if (${filterWidthVec4Remainder === 1}) {
            vec4 values = vec4(
              getValue(batch, xR, xC, d),
              initializationValue,
              initializationValue,
              initializationValue
            );

            ${updateSnippet}
          } else if (${filterWidthVec4Remainder === 2}) {
            vec4 values = vec4(
              getValue(batch, xR, xC, d),
              getValue(batch, xR, xC + ${dilationWidth}, d),
              initializationValue,
              initializationValue
            );

            ${updateSnippet}
          } else if (${filterWidthVec4Remainder === 3}) {
            vec4 values = vec4(
              getValue(batch, xR, xC, d),
              getValue(batch, xR, xC + ${dilationWidth}, d),
              getValue(batch, xR, xC + 2 * ${dilationWidth}, d),
              initializationValue
            );

            ${updateSnippet}
          }
        }
        setOutput(${returnValue});
      }
    `;
        }
    }
    class Pool3DProgram {
        constructor(convInfo, poolType, computePositions, flattenPositions = false, includeBatchInIndex = false) {
            this.variableNames = ['x'];
            if (poolType === 'avg' && computePositions) {
                throw new Error('Cannot compute positions for average pool.');
            }
            const filterWidth = convInfo.filterWidth;
            const strideDepth = convInfo.strideDepth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const dilationDepth = convInfo.dilationDepth;
            const dilationHeight = convInfo.dilationHeight;
            const dilationWidth = convInfo.dilationWidth;
            const effectiveFilterDepth = convInfo.effectiveFilterDepth;
            const effectiveFilterHeight = convInfo.effectiveFilterHeight;
            const effectiveFilterWidth = convInfo.effectiveFilterWidth;
            const padFront = convInfo.padInfo.front;
            const padTop = convInfo.padInfo.top;
            const padLeft = convInfo.padInfo.left;
            this.outputShape = convInfo.outShape;
            const isAvgPool = poolType === 'avg';
            let initializationValue = '0.0';
            if (!isAvgPool) {
                // WebGL on Firefox Linux can't compile 1/0 so we do 1/eps.
                initializationValue = '-1.0 / 1e-20';
            }
            if (computePositions) {
                const compareOp = '>=';
                this.userCode = `
        const ivec3 strides =
            ivec3(${strideDepth}, ${strideHeight}, ${strideWidth});
        const ivec3 pads = ivec3(${padFront}, ${padTop}, ${padLeft});

        void main() {
          ivec5 coords = getOutputCoords();
          int batch = coords.x;
          int ch = coords.u;

          ivec3 xCorner = ivec3(coords.y, coords.z, coords.w) * strides - pads;
          int xDCorner = xCorner.x;
          int xRCorner = xCorner.y;
          int xCCorner = xCorner.z;

          // max/min x(?, ?, ?, ch) to get y(yD, yR, yC, ch).
          // ? = to be determined
          float minMaxValue = 0.0;
          float minMaxValueFound = 0.0;
          int minMaxPosition = 0;

          for (int wD = 0; wD < ${effectiveFilterDepth};
              wD += ${dilationDepth}) {
            int xD = xDCorner + wD;

            if (xD < 0 || xD >= ${convInfo.inDepth}) {
              continue;
            }

            for (int wR = 0; wR < ${effectiveFilterHeight};
                wR += ${dilationHeight}) {
              int xR = xRCorner + wR;

              if (xR < 0 || xR >= ${convInfo.inHeight}) {
                continue;
              }

              for (int wC = 0; wC < ${effectiveFilterWidth};
                  wC += ${dilationWidth}) {
                int xC = xCCorner + wC;

                if (xC < 0 || xC >= ${convInfo.inWidth}) {
                  continue;
                }

                float value = getX(batch, xD, xR, xC, ch);

                // If a min / max value has already been found, use it. If not,
                // use the current value.
                float currMinMaxValue = mix(
                    value, minMaxValue, minMaxValueFound);
                if (value ${compareOp} currMinMaxValue) {
                  minMaxValue = value;
                  minMaxValueFound = 1.0;
                  minMaxPosition = ${flattenPositions ?
                (includeBatchInIndex ?
                    `(((batch * ${convInfo.inDepth} + xD) * ${convInfo.inHeight} + xR) * ${convInfo.inWidth} + xC) * ${convInfo.inChannels} + ch` :
                    `((xD * ${convInfo.inHeight} + xR) * ${convInfo.inWidth} + xC) * ${convInfo.inChannels} + ch`) :
                `wD * ${effectiveFilterHeight} * ${effectiveFilterWidth} +
                      wR * ${effectiveFilterWidth} + wC`};
                }
              }
            }
          }
          setOutput(float(minMaxPosition));
        }
      `;
                return;
            }
            const compareOp = 'max';
            let returnValue = `${poolType}(${poolType}(${poolType}(` +
                'minMaxValue[0], minMaxValue[1]), minMaxValue[2]), minMaxValue[3])';
            if (poolType === 'avg') {
                returnValue = `avgValue / count`;
            }
            const filterWidthNearestVec4 = Math.floor(filterWidth / 4) * 4;
            const filterWidthVec4Remainder = filterWidth % 4;
            const updateSnippet = `
      if (${isAvgPool}) {
        avgValue += dot(values, ones);
      } else {
        minMaxValue = ${compareOp}(values, minMaxValue);
      }
    `;
            this.userCode = `
      const ivec3 strides =
        ivec3(${strideDepth}, ${strideHeight}, ${strideWidth});
      const ivec3 pads = ivec3(${padFront}, ${padTop}, ${padLeft});
      const float initializationValue = ${initializationValue};
      const vec4 ones = vec4(1.0, 1.0, 1.0, 1.0);

      float count = 0.0;

      float getValue(int batch, int xD, int xR, int xC, int ch) {
        if (xC < 0 || xC >= ${convInfo.inWidth}) {
          return initializationValue;
        }
        count += 1.0;
        return getX(batch, xD, xR, xC, ch);
      }

      void main() {
        ivec5 coords = getOutputCoords();
        int batch = coords.x;
        int ch = coords.u;

        ivec3 xCorner = ivec3(coords.y, coords.z, coords.w) * strides - pads;
        int xDCorner = xCorner.x;
        int xRCorner = xCorner.y;
        int xCCorner = xCorner.z;

        // max/min x(?, ?, ?, d) to get y(yD, yR, yC, ch).
        // ? = to be determined
        vec4 minMaxValue = vec4(${initializationValue});
        float avgValue = 0.0;
        count = 0.0;

        for (int wD = 0; wD < ${effectiveFilterDepth};
            wD += ${dilationDepth}) {
          int xD = xDCorner + wD;

          if (xD < 0 || xD >= ${convInfo.inDepth}) {
            continue;
          }

          for (int wR = 0; wR < ${effectiveFilterHeight};
            wR += ${dilationHeight}) {
            int xR = xRCorner + wR;

            if (xR < 0 || xR >= ${convInfo.inHeight}) {
              continue;
            }

            for (int wC = 0; wC < ${filterWidthNearestVec4}; wC += 4) {
              int xC = xCCorner + wC * ${dilationWidth};

              vec4 values = vec4(
                getValue(batch, xD, xR, xC, ch),
                getValue(batch, xD, xR, xC + ${dilationWidth}, ch),
                getValue(batch, xD, xR, xC + 2 * ${dilationWidth}, ch),
                getValue(batch, xD, xR, xC + 3 * ${dilationWidth}, ch)
              );

              ${updateSnippet}
            }

            int xC = xCCorner + ${filterWidthNearestVec4};
            if (${filterWidthVec4Remainder === 1}) {
              vec4 values = vec4(
                getValue(batch, xD, xR, xC, ch),
                initializationValue,
                initializationValue,
                initializationValue
              );

              ${updateSnippet}
            } else if (${filterWidthVec4Remainder === 2}) {
              vec4 values = vec4(
                getValue(batch, xD, xR, xC, ch),
                getValue(batch, xD, xR, xC + ${dilationWidth}, ch),
                initializationValue,
                initializationValue
              );

              ${updateSnippet}
            } else if (${filterWidthVec4Remainder === 3}) {
              vec4 values = vec4(
                getValue(batch, xD, xR, xC, ch),
                getValue(batch, xD, xR, xC + ${dilationWidth}, ch),
                getValue(batch, xD, xR, xC + 2 * ${dilationWidth}, ch),
                initializationValue
              );

              ${updateSnippet}
            }
          }
          setOutput(${returnValue});
        }
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function avgPool(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        assertNotComplex(x, 'avgPool');
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        const dilations = 1;
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in avgPool: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        const convInfo = computePool2DInfo(x.shape, filterSize, strides, dilations, pad, dimRoundingMode);
        if (convInfo.filterWidth === 1 && convInfo.filterHeight === 1 &&
            arraysEqual(convInfo.inShape, convInfo.outShape)) {
            return identity({ inputs: { x }, backend });
        }
        const avgPoolProgram = new Pool2DProgram(convInfo, 'avg', false);
        return backend.runWebGLProgram(avgPoolProgram, [x], 'float32');
    }
    const avgPoolConfig = {
        kernelName: AvgPool,
        backendName: 'webgl',
        kernelFunc: avgPool
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function avgPool3D(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { filterSize, strides, pad, dimRoundingMode, dataFormat } = attrs;
        const dilations = [1, 1, 1];
        const convInfo = computePool3DInfo(x.shape, filterSize, strides, dilations, pad, dimRoundingMode, dataFormat);
        const avgPoolProgram = new Pool3DProgram(convInfo, 'avg', false);
        return backend.runWebGLProgram(avgPoolProgram, [x], 'float32');
    }
    const avgPool3DConfig = {
        kernelName: AvgPool3D,
        backendName: 'webgl',
        kernelFunc: avgPool3D
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class AvgPool2DBackpropProgram {
        constructor(convInfo) {
            this.variableNames = ['dy'];
            this.outputShape = convInfo.inShape;
            const filterHeight = convInfo.filterHeight;
            const filterWidth = convInfo.filterWidth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const dilationHeight = convInfo.dilationHeight;
            const dilationWidth = convInfo.dilationWidth;
            const effectiveFilterHeight = convInfo.effectiveFilterHeight;
            const effectiveFilterWidth = convInfo.effectiveFilterWidth;
            const padTop = effectiveFilterHeight - 1 - convInfo.padInfo.top;
            const padLeft = effectiveFilterWidth - 1 - convInfo.padInfo.left;
            const avgMultiplier = 1 / (filterHeight * filterWidth);
            this.userCode = `
      const ivec2 pads = ivec2(${padTop}, ${padLeft});
      const float avgMultiplier = float(${avgMultiplier});

      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int d = coords[3];

        ivec2 dyRCCorner = coords.yz - pads;
        int dyRCorner = dyRCCorner.x;
        int dyCCorner = dyRCCorner.y;

        // Convolve dy(?, ?, d) with pos mask(:, :, d) to get dx(xR, xC, d).
        // ? = to be determined. : = across all values in that axis.
        float dotProd = 0.0;
        for (int wR = 0; wR < ${effectiveFilterHeight};
            wR += ${dilationHeight}) {
          float dyR = float(dyRCorner + wR) / ${strideHeight}.0;

          if (dyR < 0.0 || dyR >= ${convInfo.outHeight}.0 || fract(dyR) > 0.0) {
            continue;
          }
          int idyR = int(dyR);

          for (int wC = 0; wC < ${effectiveFilterWidth};
            wC+= ${dilationWidth}) {
            float dyC = float(dyCCorner + wC) / ${strideWidth}.0;

            if (dyC < 0.0 || dyC >= ${convInfo.outWidth}.0 ||
                fract(dyC) > 0.0) {
              continue;
            }
            int idyC = int(dyC);

            float dyValue = getDy(b, idyR, idyC, d);

            dotProd += dyValue * avgMultiplier;
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }
    class AvgPool3DBackpropProgram {
        constructor(convInfo) {
            this.variableNames = ['dy'];
            this.outputShape = convInfo.inShape;
            const filterDepth = convInfo.filterDepth;
            const filterHeight = convInfo.filterHeight;
            const filterWidth = convInfo.filterWidth;
            const strideDepth = convInfo.strideDepth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const dilationDepth = convInfo.dilationDepth;
            const dilationHeight = convInfo.dilationHeight;
            const dilationWidth = convInfo.dilationWidth;
            const effectiveFilterDepth = convInfo.effectiveFilterDepth;
            const effectiveFilterHeight = convInfo.effectiveFilterHeight;
            const effectiveFilterWidth = convInfo.effectiveFilterWidth;
            const padFront = effectiveFilterDepth - 1 - convInfo.padInfo.front;
            const padTop = effectiveFilterHeight - 1 - convInfo.padInfo.top;
            const padLeft = effectiveFilterWidth - 1 - convInfo.padInfo.left;
            const avgMultiplier = 1 / (filterDepth * filterHeight * filterWidth);
            this.userCode = `
      const ivec3 pads = ivec3(${padFront}, ${padTop}, ${padLeft});
      const float avgMultiplier = float(${avgMultiplier});

      void main() {
        ivec5 coords = getOutputCoords();
        int batch = coords.x;
        int ch = coords.u;

        ivec3 dyCorner = ivec3(coords.y, coords.z, coords.w) - pads;
        int dyDCorner = dyCorner.x;
        int dyRCorner = dyCorner.y;
        int dyCCorner = dyCorner.z;

        // Convolve dy(?, ?, ?, d) with pos mask(:, :, :, ch) to get
        // dx(xD, xR, xC, ch).
        // ? = to be determined. : = across all values in that axis.
        float dotProd = 0.0;

        for (int wD = 0; wD < ${effectiveFilterDepth};
            wD += ${dilationDepth}) {
          float dyD = float(dyDCorner + wD) / ${strideDepth}.0;

          if (dyD < 0.0 || dyD >= ${convInfo.outDepth}.0 || fract(dyD) > 0.0) {
            continue;
          }
          int idyD = int(dyD);

          for (int wR = 0; wR < ${effectiveFilterHeight};
              wR += ${dilationHeight}) {
            float dyR = float(dyRCorner + wR) / ${strideHeight}.0;

            if (dyR < 0.0 || dyR >= ${convInfo.outHeight}.0 ||
                fract(dyR) > 0.0) {
              continue;
            }
            int idyR = int(dyR);

            for (int wC = 0; wC < ${effectiveFilterWidth};
                wC += ${dilationWidth}) {
              float dyC = float(dyCCorner + wC) / ${strideWidth}.0;

              if (dyC < 0.0 || dyC >= ${convInfo.outWidth}.0 ||
                  fract(dyC) > 0.0) {
                continue;
              }
              int idyC = int(dyC);

              float dyValue = getDy(batch, idyD, idyR, idyC, ch);

              dotProd += dyValue * avgMultiplier;
            }
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function avgPool3DGrad(args) {
        const { inputs, backend, attrs } = args;
        const { dy, input } = inputs;
        const x = input;
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        const dilations = [1, 1, 1];
        const convInfo = computePool3DInfo(x.shape, filterSize, strides, dilations, pad, dimRoundingMode);
        const avgPoolBackpropProgram = new AvgPool3DBackpropProgram(convInfo);
        return backend.runWebGLProgram(avgPoolBackpropProgram, [dy], x.dtype);
    }
    const avgPool3DGradConfig = {
        kernelName: AvgPool3DGrad,
        backendName: 'webgl',
        kernelFunc: avgPool3DGrad
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function avgPoolGrad(args) {
        const { inputs, backend, attrs } = args;
        const { dy, input } = inputs;
        const x = input;
        assertNotComplex([dy, input], 'avgPoolGrad');
        const { filterSize, strides, pad } = attrs;
        const convInfo = computePool2DInfo(x.shape, filterSize, strides, 1 /* dilations */, pad);
        const avgPoolBackpropProgram = new AvgPool2DBackpropProgram(convInfo);
        return backend.runWebGLProgram(avgPoolBackpropProgram, [dy], x.dtype);
    }
    const avgPoolGradConfig = {
        kernelName: AvgPoolGrad,
        backendName: 'webgl',
        kernelFunc: avgPoolGrad
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function batchMatMul(args) {
        const { inputs, backend, attrs } = args;
        const { a, b } = inputs;
        const { transposeA, transposeB } = attrs;
        return batchMatMulImpl({ a, b, transposeA, transposeB, backend });
    }
    const batchMatMulConfig = {
        kernelName: BatchMatMul,
        backendName: 'webgl',
        kernelFunc: batchMatMul,
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class BatchNormProgram {
        constructor(xShape, meanShape, varianceShape, offsetShape, scaleShape, varianceEpsilon) {
            this.outputShape = [];
            this.variableNames = ['x', 'mean', 'variance'];
            assertAndGetBroadcastShape(xShape, meanShape);
            assertAndGetBroadcastShape(xShape, varianceShape);
            let offsetSnippet = '0.0';
            if (offsetShape != null) {
                assertAndGetBroadcastShape(xShape, offsetShape);
                this.variableNames.push('offset');
                offsetSnippet = 'getOffsetAtOutCoords()';
            }
            let scaleSnippet = '1.0';
            if (scaleShape != null) {
                assertAndGetBroadcastShape(xShape, scaleShape);
                this.variableNames.push('scale');
                scaleSnippet = 'getScaleAtOutCoords()';
            }
            this.outputShape = xShape;
            this.userCode = `
      void main() {
        float x = getXAtOutCoords();
        float mean = getMeanAtOutCoords();
        float variance = getVarianceAtOutCoords();
        float offset = ${offsetSnippet};
        float scale = ${scaleSnippet};
        float inv = scale * inversesqrt(variance + float(${varianceEpsilon}));
        setOutput(dot(vec3(x, -mean, offset), vec3(inv, inv, 1)));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class BatchNormPackedProgram {
        constructor(xShape, meanShape, varianceShape, offsetShape, scaleShape, varianceEpsilon) {
            this.packedInputs = true;
            this.packedOutput = true;
            this.variableNames = ['x', 'mean', 'variance'];
            assertAndGetBroadcastShape(xShape, meanShape);
            assertAndGetBroadcastShape(xShape, varianceShape);
            let offsetSnippet = 'vec4(0.0)';
            if (offsetShape != null) {
                assertAndGetBroadcastShape(xShape, offsetShape);
                this.variableNames.push('offset');
                offsetSnippet = 'getOffsetAtOutCoords()';
            }
            let scaleSnippet = 'vec4(1.0)';
            if (scaleShape != null) {
                assertAndGetBroadcastShape(xShape, scaleShape);
                this.variableNames.push('scale');
                scaleSnippet = 'getScaleAtOutCoords()';
            }
            this.outputShape = xShape;
            this.userCode = `
      void main() {
        vec4 offset = ${offsetSnippet};
        vec4 scale = ${scaleSnippet};

        vec4 x = getXAtOutCoords();
        vec4 mean = getMeanAtOutCoords();
        vec4 variance = getVarianceAtOutCoords();

        vec4 inv = scale * inversesqrt(variance + vec4(${varianceEpsilon}));

        setOutput((x - mean) * inv + offset);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const batchNorm = ({ inputs, backend, attrs }) => {
        const { x, mean, variance, offset, scale } = inputs;
        assert$4(mean.shape.length === variance.shape.length, () => 'Batch normalization gradient requires mean and variance to have ' +
            'equal ranks.');
        assert$4(offset == null || mean.shape.length === offset.shape.length, () => 'Batch normalization gradient requires mean and offset to have ' +
            'equal ranks.');
        assert$4(scale == null || mean.shape.length === scale.shape.length, () => 'Batch normalization gradient requires mean and scale to have ' +
            'equal ranks.');
        let { varianceEpsilon } = attrs;
        if (varianceEpsilon == null) {
            varianceEpsilon = 0.001;
        }
        const finalInputs = [x, mean, variance];
        let offsetShape = null;
        if (offset != null) {
            offsetShape = offset.shape;
            finalInputs.push(offset);
        }
        let scaleShape = null;
        if (scale != null) {
            scaleShape = scale.shape;
            finalInputs.push(scale);
        }
        const program = env$1().getBool('WEBGL_PACK_NORMALIZATION') ?
            new BatchNormPackedProgram(x.shape, mean.shape, variance.shape, offsetShape, scaleShape, varianceEpsilon) :
            new BatchNormProgram(x.shape, mean.shape, variance.shape, offsetShape, scaleShape, varianceEpsilon);
        const output = backend.runWebGLProgram(program, finalInputs, finalInputs[0].dtype);
        return output;
    };
    const batchNormConfig = {
        kernelName: FusedBatchNorm,
        backendName: 'webgl',
        kernelFunc: batchNorm,
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class SliceProgram {
        constructor(destSize) {
            this.variableNames = ['source'];
            this.outputShape = destSize;
            this.rank = destSize.length;
            const dtype = getCoordsDataType(this.rank);
            this.customUniforms = [{ name: 'start', arrayIndex: this.rank, type: 'int' }];
            const sourceCoords = getCoords$1(this.rank);
            let body;
            const coordSum = destSize.map((_, i) => {
                return `sourceLoc.${coords[i]} = start[${i}] + coords.${coords[i]};`;
            });
            body = `
        ${dtype} sourceLoc;
        ${dtype} coords = getOutputCoords();
        ${coordSum.join('\n')}
      `;
            this.userCode = `
      void main() {
        ${body}
        setOutput(getSource(${sourceCoords}));
      }
    `;
        }
    }
    const coords = ['x', 'y', 'z', 'w', 'u', 'v'];
    function getCoords$1(rank) {
        if (rank === 1) {
            return 'sourceLoc';
        }
        else if (rank <= 6) {
            return coords.slice(0, rank).map(x => 'sourceLoc.' + x).join(',');
        }
        else {
            throw Error(`Slicing for rank ${rank} is not yet supported`);
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class SlicePackedProgram {
        constructor(destSize) {
            this.variableNames = ['source'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.outputShape = destSize;
            this.rank = destSize.length;
            this.customUniforms = [{ name: 'start', arrayIndex: this.rank, type: 'int' }];
            const dtype = getCoordsDataType(this.rank);
            const coords = getChannels('coords', this.rank);
            const sourceLoc = getChannels('sourceLoc', this.rank);
            const innerDims = this.rank === 1 ? 'sourceLoc' : `vec2(${sourceLoc.slice(-2).join()})`;
            const getChannel = `getChannel(getSource(${sourceLoc.join()}), ${innerDims})`;
            const upperRow = `
      result.x = ${getChannel};
      if (++${coords[this.rank - 1]} < ${destSize[this.rank - 1]}) {
        ++${sourceLoc[this.rank - 1]};
        result.y = ${getChannel};
        --${sourceLoc[this.rank - 1]};
      }
    `;
            const lowerRow = this.rank === 1 ? '' : `
      --${coords[this.rank - 1]};
      if (++${coords[this.rank - 2]} < ${destSize[this.rank - 2]}) {
        ++${sourceLoc[this.rank - 2]};
        result.z = ${getChannel};
        if (++${coords[this.rank - 1]} < ${destSize[this.rank - 1]}) {
          ++${sourceLoc[this.rank - 1]};
          result.w = ${getChannel};
        }
      }
    `;
            const sourceLocSetup = this.rank <= 4 ?
                `sourceLoc = coords +
            ${dtype}(${destSize.map((_, i) => `start[${i}]`).join()});` :
                destSize.map((_, i) => `${sourceLoc[i]} = ${coords[i]} + start[${i}];`)
                    .join('\n');
            this.userCode = `
      void main() {
        ${dtype} coords = getOutputCoords();
        ${dtype} sourceLoc;
        ${sourceLocSetup}
        vec4 result = vec4(0.);
        ${upperRow}
        ${lowerRow}
        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function shallowSlice(x, begin, size, backend) {
        const xTexData = backend.texData.get(x.dataId);
        const t = backend.makeTensorInfo(size, x.dtype);
        const newTexData = backend.texData.get(t.dataId);
        // Copy texture data from the original tensor.
        Object.assign(newTexData, xTexData);
        newTexData.refCount = 1;
        newTexData.shape = size;
        newTexData.dtype = x.dtype;
        let flatOffset = computeFlatOffset(begin, computeStrides(x.shape));
        if (xTexData.slice) {
            // We are slicing an already sliced tensor, so we have to accumulate
            // the offset.
            flatOffset += xTexData.slice.flatOffset;
        }
        newTexData.slice = {
            flatOffset,
            // Point to the original dataId, which is used to do ref counting.
            origDataId: xTexData.slice && xTexData.slice.origDataId || x.dataId
        };
        // Increase the ref count for that data bucket.
        const refCount = backend.dataRefCount.get(newTexData.slice.origDataId) || 1;
        backend.dataRefCount.set(newTexData.slice.origDataId, refCount + 1);
        return t;
    }
    function slice(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { begin, size } = attrs;
        const [$begin, $size] = parseSliceParams(x, begin, size);
        assertParamsValid(x, $begin, $size);
        if (sizeFromShape($size) === 0) {
            return backend.makeTensorInfo($size, x.dtype, []);
        }
        // Run on cpu if dtype is string. For string, the backend represents it
        // as Uint8Array[], where each Uint8Array is a character. Given that the
        // computation is only on the outer array, uploading the whole data onto
        // gpu is wasteful. Also, currently webgl doesn't have a design to
        // upload and retrieve Uint8Array[] between cpu and gpu. Therefore, we
        // just run the kernel on cpu if dtype is string.
        if (backend.shouldExecuteOnCPU([x]) || x.dtype === 'string') {
            const xTexData = backend.texData.get(x.dataId);
            const outValues = sliceImplCPU(xTexData.values, $begin, $size, x.shape, x.dtype);
            return backend.makeTensorInfo($size, x.dtype, outValues);
        }
        const { isPacked } = backend.texData.get(x.dataId);
        const isContinous = isSliceContinous(x.shape, $begin, $size);
        if (isPacked || !isContinous) {
            const program = env$1().getBool('WEBGL_PACK_ARRAY_OPERATIONS') ?
                new SlicePackedProgram($size) :
                new SliceProgram($size);
            const customValues = [$begin];
            return backend.runWebGLProgram(program, [x], x.dtype, customValues);
        }
        backend.uploadToGPU(x.dataId);
        return shallowSlice(x, $begin, $size, backend);
    }
    const sliceConfig = {
        kernelName: Slice,
        backendName: 'webgl',
        kernelFunc: slice
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const batchToSpaceND = (args) => {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { blockShape, crops } = attrs;
        assert$4(x.shape.length <= 4, () => 'batchToSpaceND for rank > 4 with a WebGL backend not ' +
            'implemented yet');
        const prod = blockShape.reduce((a, b) => a * b);
        const reshaped = getReshaped(x.shape, blockShape, prod);
        const permuted = getPermuted(reshaped.length, blockShape.length);
        const reshapedPermuted = getReshapedPermuted(x.shape, blockShape, prod);
        const sliceBeginCoords = getSliceBeginCoords(crops, blockShape.length);
        const sliceSize = getSliceSize(reshapedPermuted, crops, blockShape.length);
        const toDispose = [];
        const reshapedIntermediate = reshape({ inputs: { x }, backend, attrs: { shape: reshaped } });
        const transposedIntermediate = transpose({ inputs: { x: reshapedIntermediate }, backend, attrs: { perm: permuted } });
        const reshapedIntermediate2 = reshape({
            inputs: { x: transposedIntermediate },
            backend,
            attrs: { shape: reshapedPermuted }
        });
        const sliced = slice({
            inputs: { x: reshapedIntermediate2 },
            backend,
            attrs: { begin: sliceBeginCoords, size: sliceSize }
        });
        toDispose.push(reshapedIntermediate);
        toDispose.push(transposedIntermediate);
        toDispose.push(reshapedIntermediate2);
        toDispose.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return sliced;
    };
    const batchToSpaceNDConfig = {
        kernelName: BatchToSpaceND,
        backendName: 'webgl',
        kernelFunc: batchToSpaceND
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function bincount(args) {
        const { inputs, backend, attrs } = args;
        const { x, weights } = inputs;
        const { size } = attrs;
        const xVals = backend.readSync(x.dataId);
        const weightsVals = backend.readSync(weights.dataId);
        const outVals = bincountImplCPU(xVals, weightsVals, weights.dtype, weights.shape, size);
        return backend.makeTensorInfo([size], weights.dtype, outVals);
    }
    const bincountConfig = {
        kernelName: Bincount,
        backendName: 'webgl',
        kernelFunc: bincount
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function broadcastArgs(args) {
        const { inputs, backend } = args;
        const { s0, s1 } = inputs;
        const s0Vals = backend.readSync(s0.dataId);
        const s1Vals = backend.readSync(s1.dataId);
        const broadcastShape = assertAndGetBroadcastShape(Array.from(s0Vals), Array.from(s1Vals));
        return backend.makeTensorInfo([broadcastShape.length], 'int32', Int32Array.from(broadcastShape));
    }
    const broadcastArgsConfig = {
        kernelName: BroadcastArgs,
        backendName: 'webgl',
        kernelFunc: broadcastArgs
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const NOT_EQUAL = `return float(a != b);`;
    const notEqual$1 = binaryKernelFunc({ opSnippet: NOT_EQUAL, cpuKernelImpl: notEqualImplCPU, dtype: 'bool' });
    const notEqualConfig = {
        kernelName: NotEqual,
        backendName: 'webgl',
        kernelFunc: notEqual$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function real(args) {
        const { inputs, backend } = args;
        const { input } = inputs;
        const inputData = backend.texData.get(input.dataId);
        return identity({ inputs: { x: inputData.complexTensorInfos.real }, backend });
    }
    const realConfig = {
        kernelName: Real,
        backendName: 'webgl',
        kernelFunc: real
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const TO_INT = `return float(int(x));`;
    function int(input, backend) {
        const program = new UnaryOpProgram(input.shape, TO_INT);
        const output = backend.runWebGLProgram(program, [input], 'int32');
        return { dataId: output.dataId, shape: output.shape, dtype: output.dtype };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function cast(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { dtype } = attrs;
        // Casting to complex64.
        if (dtype === 'complex64') {
            if (x.dtype === 'complex64') {
                return identity({ inputs: { x }, backend });
            }
            // TODO(annxingyuan): Import kernel function once zeros is modularized.
            const zerosTensor = zeros$2(x.shape);
            const floatX = cast({ inputs: { x }, backend, attrs: { dtype: 'float32' } });
            const result = complex({ inputs: { real: floatX, imag: zerosTensor }, backend });
            zerosTensor.dispose();
            backend.disposeIntermediateTensorInfo(floatX);
            return result;
        }
        // Casting from complex64
        if (x.dtype === 'complex64') {
            const realPart = real({ inputs: { input: x }, backend });
            const result = cast({ inputs: { x: realPart }, backend, attrs: { dtype } });
            backend.disposeIntermediateTensorInfo(realPart);
            return result;
        }
        if (!hasEncodingLoss(x.dtype, dtype)) {
            // We don't change the underlying data, since we cast to higher
            // precision.
            const result = identity({ inputs: { x }, backend });
            return { dataId: result.dataId, shape: result.shape, dtype };
        }
        if (dtype === 'int32') {
            return int(x, backend);
        }
        if (dtype === 'bool') {
            const zerosTensorInfo = backend.makeTensorInfo([], 'bool', getTypedArrayFromDType('bool', 1));
            const binaryInputs = { a: x, b: zerosTensorInfo };
            const result = notEqual$1({ inputs: binaryInputs, backend });
            backend.disposeIntermediateTensorInfo(zerosTensorInfo);
            return result;
        }
        throw new Error(`Error in Cast: failed to cast ${x.dtype} to ${dtype}`);
    }
    const castConfig = {
        kernelName: Cast,
        backendName: 'webgl',
        kernelFunc: cast
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const CEIL = `return ceil(x);`;
    const ceil = unaryKernelFunc({ opSnippet: CEIL, packedOpSnippet: CEIL, cpuKernelImpl: ceilImplCPU });
    const ceilConfig = {
        kernelName: Ceil,
        backendName: 'webgl',
        kernelFunc: ceil
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ClipProgram {
        constructor(aShape) {
            this.variableNames = ['A'];
            this.customUniforms = [
                { name: 'minVal', type: 'float' },
                { name: 'maxVal', type: 'float' }
            ];
            this.outputShape = aShape;
            this.userCode = `

      void main() {
        float value = getAAtOutCoords();
        if (isnan(value)) {
          setOutput(value);
          return;
        }

        setOutput(clamp(value, minVal, maxVal));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ClipPackedProgram {
        constructor(aShape) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.customUniforms = [
                { name: 'minVal', type: 'float' },
                { name: 'maxVal', type: 'float' }
            ];
            this.outputShape = aShape;
            this.userCode = `
      void main() {
        vec4 value = getAAtOutCoords();

        if (any(isnan(value))) {
          setOutput(value);
          return;
        }

        setOutput(clamp(value, vec4(minVal), vec4(maxVal)));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function clipByValue(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { clipValueMin, clipValueMax } = attrs;
        let program;
        if (env$1().getBool('WEBGL_PACK_CLIP')) {
            program = new ClipPackedProgram(x.shape);
        }
        else {
            program = new ClipProgram(x.shape);
        }
        const customValues = [[clipValueMin], [clipValueMax]];
        return backend.runWebGLProgram(program, [x], x.dtype, customValues);
    }
    const clipByValueConfig = {
        kernelName: ClipByValue,
        backendName: 'webgl',
        kernelFunc: clipByValue
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ComplexAbsProgram {
        constructor(shape) {
            this.variableNames = ['real', 'imag'];
            this.outputShape = shape;
            this.userCode = `
      void main() {
        float re = abs(getRealAtOutCoords());
        float im = abs(getImagAtOutCoords());
        float mx = max(re, im);

        // sadly the length function in glsl is not underflow-safe
        // (at least not on Intel GPUs). So the safe solution is
        // to ensure underflow-safety in all cases.
        setOutput(
          mx == 0.0 ? 0.0 : mx * length(vec2(1, min(re, im)/mx))
        );
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Returns a TensorInfo with the complex shape and the dataId of the
    // underlying part. We need to do this because a reshaped complex tensor is
    // not reflected in its parts.
    function makeComplexComponentTensorInfo(complexTensor, complexPart) {
        return {
            dataId: complexPart.dataId,
            dtype: complexPart.dtype,
            shape: complexTensor.shape
        };
    }
    function complexAbs(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        const xData = backend.texData.get(x.dataId);
        const program = new ComplexAbsProgram(x.shape);
        const programInputs = [
            makeComplexComponentTensorInfo(x, xData.complexTensorInfos.real),
            makeComplexComponentTensorInfo(x, xData.complexTensorInfos.imag),
        ];
        return backend.runWebGLProgram(program, programInputs, programInputs[0].dtype);
    }
    const complexAbsConfig = {
        kernelName: ComplexAbs,
        backendName: 'webgl',
        kernelFunc: complexAbs
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ConcatProgram {
        // Concats 2d tensors along axis=1. See comments in MathBackendWebGL.concat().
        constructor(shapes) {
            this.outputShape = [];
            this.outputShape = computeOutShape$1(shapes, 1 /* axis */);
            this.variableNames = shapes.map((_, i) => `T${i}`);
            const offsets = new Array(shapes.length - 1);
            offsets[0] = shapes[0][1];
            for (let i = 1; i < offsets.length; i++) {
                offsets[i] = offsets[i - 1] + shapes[i][1];
            }
            const snippets = [`if (yC < ${offsets[0]}) setOutput(getT0(yR, yC));`];
            for (let i = 1; i < offsets.length; i++) {
                const shift = offsets[i - 1];
                snippets.push(`else if (yC < ${offsets[i]}) ` +
                    `setOutput(getT${i}(yR, yC-${shift}));`);
            }
            const lastIndex = offsets.length;
            const lastShift = offsets[offsets.length - 1];
            snippets.push(`else setOutput(getT${lastIndex}(yR, yC-${lastShift}));`);
            this.userCode = `
      void main() {
        ivec2 coords = getOutputCoords();
        int yR = coords.x;
        int yC = coords.y;

        ${snippets.join('\n        ')}
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ConcatPackedProgram {
        constructor(shapes, axis) {
            this.packedInputs = true;
            this.packedOutput = true;
            this.outputShape = [];
            this.outputShape = computeOutShape$1(shapes, axis);
            const shape = this.outputShape;
            const rank = shape.length;
            const dtype = getCoordsDataType(rank);
            const coords = getChannels('coords', rank);
            const channels = ['x', 'y', 'z', 'w', 'u', 'v'].slice(0, rank);
            this.variableNames = shapes.map((_, i) => `T${i}`);
            const offsets = new Array(shapes.length - 1);
            offsets[0] = shapes[0][axis];
            for (let i = 1; i < offsets.length; i++) {
                offsets[i] = offsets[i - 1] + shapes[i][axis];
            }
            const channel = channels[axis];
            const lastChannels = channels.slice(-2);
            const allChannels = channels.join();
            let getValueSnippet = `if (${channel} < ${offsets[0]}) {
        return getChannel(
            getT0(${allChannels}), vec2(${lastChannels.join()}));
        }`;
            for (let i = 1; i < offsets.length; i++) {
                const shift = offsets[i - 1];
                // Note: the >= comparison below may seem unnecessary given the check
                // above but is needed to workaround branch execution issues on some
                // devices. It makes all the conditions exclusive without relying on
                // execution order.
                getValueSnippet += `
        if (${channel} < ${offsets[i]}  && ${channel} >= ${offsets[i - 1]}) {
          return getChannel(
            getT${i}(${shiftedChannels(channels, channel, shift)}),
            vec2(${shiftedChannels(lastChannels, channel, shift)}));
        }`;
            }
            const lastIndex = offsets.length;
            const shift = offsets[offsets.length - 1];
            getValueSnippet += `
        return getChannel(
          getT${lastIndex}(${shiftedChannels(channels, channel, shift)}),
          vec2(${shiftedChannels(lastChannels, channel, shift)}));`;
            this.userCode = `
      float getValue(${channels.map(x => 'int ' + x)}) {
        ${getValueSnippet}
      }

      void main() {
        ${dtype} coords = getOutputCoords();
        vec4 result = vec4(getValue(${coords}), 0., 0., 0.);

        ${coords[rank - 1]} = ${coords[rank - 1]} + 1;
        if (${coords[rank - 1]} < ${shape[rank - 1]}) {
          result.g = getValue(${coords});
        }

        ${coords[rank - 2]} = ${coords[rank - 2]} + 1;
        if (${coords[rank - 2]} < ${shape[rank - 2]}) {
          result.a = getValue(${coords});
        }

        ${coords[rank - 1]} = ${coords[rank - 1]} - 1;
        if (${coords[rank - 2]} < ${shape[rank - 2]} &&
            ${coords[rank - 1]} < ${shape[rank - 1]}) {
          result.b = getValue(${coords});
        }
        setOutput(result);
      }
    `;
        }
    }
    /**
     * Return an expression for coordinates into a vector where a given channel
     * will be offset by [shift].
     *
     * @param channels the channels to consider
     * @param channel the channel we want shifted
     * @param shift  the amount to subtract from the channel.
     *
     * @returns a string of the form 'x, y-[shift], z' where any one channel can
     * have the shift applied.
     */
    function shiftedChannels(channels, channel, shift) {
        const channelIdx = channels.indexOf(channel);
        const res = channels.map((c, idx) => {
            if (idx === channelIdx) {
                return `${c} - ${shift}`;
            }
            else {
                return c;
            }
        });
        return res.join();
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function imag(args) {
        const { inputs, backend } = args;
        const { input } = inputs;
        const inputData = backend.texData.get(input.dataId);
        return identity({ inputs: { x: inputData.complexTensorInfos.imag }, backend });
    }
    const imagConfig = {
        kernelName: Imag,
        backendName: 'webgl',
        kernelFunc: imag
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function concatImpl(inputs, axis, backend) {
        const dtype = inputs[0].dtype;
        if (dtype === 'complex64') {
            const reals = inputs.map((t) => real({ inputs: { input: t }, backend }));
            const imags = inputs.map((t) => imag({ inputs: { input: t }, backend }));
            const realConcated = concatImpl(reals, axis, backend);
            const imagConcated = concatImpl(imags, axis, backend);
            const result = complex({ inputs: { real: realConcated, imag: imagConcated }, backend });
            reals.forEach(r => backend.disposeIntermediateTensorInfo(r));
            imags.forEach(i => backend.disposeIntermediateTensorInfo(i));
            backend.disposeIntermediateTensorInfo(realConcated);
            backend.disposeIntermediateTensorInfo(imagConcated);
            return result;
        }
        let runOnCpu = backend.shouldExecuteOnCPU(inputs);
        // Run on cpu if dtype is string. For string, the backend represents it
        // as Uint8Array[], where each Uint8Array is a character. Given that the
        // computation is only on the outer array, uploading the whole data onto
        // gpu is wasteful. Also, currently webgl doesn't have a design to
        // upload and retrieve Uint8Array[] between cpu and gpu. Therefore, we
        // just run the kernel on cpu if dtype is string.
        if (dtype === 'string') {
            runOnCpu = true;
        }
        if (runOnCpu) {
            // Any concat of n-dimensional tensors across any axis can be reduced to
            // a concatenation of two-dimensional tensors across the axis 1 by first
            // partitioning the axes of the original tensors into those less than the
            // axis to be concatenated and the rest. Then reshape the tensors
            // into a two-dimensional tensor by collapsing these two sets of axes and
            // concatenate the resulting matrices across the axis 1, finally reshaping
            // the result to have the proper shape.
            const tensors2D = inputs.map(t => {
                const innerSize = sizeFromShape(t.shape.slice(axis));
                const shape = [-1, innerSize];
                return reshape({ inputs: { x: t }, backend, attrs: { shape } });
            });
            const inputsValShapes = tensors2D.map(t => {
                return { vals: backend.readSync(t.dataId), shape: t.shape };
            });
            // Concats 2d tensors along axis=1.
            const outShape = computeOutShape$1(tensors2D.map(t => t.shape), 1 /* axis */);
            const simplyConcat = tensors2D[0].shape[0] === 1;
            const outVals = concatImplCPU(inputsValShapes, outShape, dtype, simplyConcat);
            const finalOutShape = computeOutShape$1(inputs.map(t => t.shape), axis);
            const outInfo = backend.makeTensorInfo(finalOutShape, dtype, outVals);
            tensors2D.forEach(t => backend.disposeIntermediateTensorInfo(t));
            return outInfo;
        }
        if (inputs.length > env$1().getNumber('WEBGL_MAX_TEXTURES_IN_SHADER')) {
            const midIndex = Math.floor(inputs.length / 2);
            const leftSide = concatImpl(inputs.slice(0, midIndex), axis, backend);
            const rightSide = concatImpl(inputs.slice(midIndex), axis, backend);
            const result = concatImpl([leftSide, rightSide], axis, backend);
            backend.disposeIntermediateTensorInfo(leftSide);
            backend.disposeIntermediateTensorInfo(rightSide);
            return result;
        }
        if (env$1().getBool('WEBGL_PACK_ARRAY_OPERATIONS') &&
            inputs[0].shape.length > 1) {
            const program = new ConcatPackedProgram(inputs.map(t => t.shape), axis);
            return backend.runWebGLProgram(program, inputs, dtype);
        }
        const { tensors2D, outShape } = computeTensors2D(inputs, axis, backend);
        const program = new ConcatProgram(tensors2D.map(t => t.shape));
        const result = backend.runWebGLProgram(program, tensors2D, dtype);
        tensors2D.forEach(r => backend.disposeIntermediateTensorInfo(r));
        const reshapedResult = reshape({ inputs: { x: result }, attrs: { shape: outShape }, backend });
        backend.disposeIntermediateTensorInfo(result);
        return reshapedResult;
    }
    function computeTensors2D(inputs, axis, backend) {
        // Any concat of n-dimensional tensors across any axis can be reduced to
        // a concatenation of two-dimensional tensors across the axis 1 by first
        // partitioning the axes of the original tensors into those less than the
        // axis to be concatenated and the rest. Then reshape the tensors
        // into a two-dimensional tensor by collapsing these two sets of axes and
        // concatenate the resulting matrices across the axis 1, finally reshaping
        // the result to have the proper shape.
        const outShape = computeOutShape$1(inputs.map(t => t.shape), axis);
        const tensors2D = inputs.map(x => reshape({
            inputs: { x },
            attrs: { shape: [-1, sizeFromShape(x.shape.slice(axis))] },
            backend
        }));
        return { tensors2D, outShape };
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function concat(args) {
        const { inputs, backend, attrs } = args;
        const { axis } = attrs;
        const $axis = parseAxisParam(axis, inputs[0].shape)[0];
        const outShape = computeOutShape$1(inputs.map(t => t.shape), $axis);
        if (sizeFromShape(outShape) === 0) {
            return backend.makeTensorInfo(outShape, inputs[0].dtype, []);
        }
        // Keep only non-empty tensors (ignore tensors with 0 in their shape).
        const $inputs = inputs.filter(t => sizeFromShape(t.shape) > 0);
        if ($inputs.length === 1) {
            return identity({ inputs: { x: $inputs[0] }, backend });
        }
        const shapes = $inputs.map(t => t.shape);
        assertParamsConsistent(shapes, $axis);
        return concatImpl($inputs, $axis, backend);
    }
    const concatConfig = {
        kernelName: Concat,
        backendName: 'webgl',
        kernelFunc: concat
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class Conv2DProgram {
        constructor(convInfo, addBias = false, activation = null, hasPreluActivationWeights = false, hasLeakyreluAlpha = false) {
            this.variableNames = ['x', 'W'];
            this.outputShape = convInfo.outShape;
            const padTop = convInfo.padInfo.top;
            const padLeft = convInfo.padInfo.left;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const dilationHeight = convInfo.dilationHeight;
            const dilationWidth = convInfo.dilationWidth;
            const filterHeight = convInfo.filterHeight;
            const filterWidth = convInfo.filterWidth;
            const inputDepthNearestVec4 = Math.floor(convInfo.inChannels / 4) * 4;
            const inputDepthVec4Remainder = convInfo.inChannels % 4;
            const isChannelsLast = convInfo.dataFormat === 'channelsLast';
            const rowDim = isChannelsLast ? 1 : 2;
            const colDim = isChannelsLast ? 2 : 3;
            const channelDim = isChannelsLast ? 3 : 1;
            let activationSnippet = '', applyActivationSnippet = '';
            if (activation) {
                if (hasPreluActivationWeights) {
                    activationSnippet = `float activation(float a) {
          float b = getPreluActivationWeightsAtOutCoords();
          ${activation}
        }`;
                }
                else if (hasLeakyreluAlpha) {
                    activationSnippet = `float activation(float a) {
          float b = getLeakyreluAlphaAtOutCoords();
          ${activation}
        }`;
                }
                else {
                    activationSnippet = `
          float activation(float x) {
            ${activation}
          }
        `;
                }
                applyActivationSnippet = `result = activation(result);`;
            }
            const addBiasSnippet = addBias ? 'result += getBiasAtOutCoords();' : '';
            if (addBias) {
                this.variableNames.push('bias');
            }
            if (hasPreluActivationWeights) {
                this.variableNames.push('preluActivationWeights');
            }
            if (hasLeakyreluAlpha) {
                this.variableNames.push('leakyreluAlpha');
            }
            this.userCode = `
      ${activationSnippet}

      const ivec2 strides = ivec2(${strideHeight}, ${strideWidth});
      const ivec2 pads = ivec2(${padTop}, ${padLeft});

      void main() {
        ivec4 coords = getOutputCoords();
        int batch = coords[0];
        int d2 = coords[${channelDim}];

        ivec2 xRCCorner =
            ivec2(coords[${rowDim}], coords[${colDim}]) * strides - pads;
        int xRCorner = xRCCorner.x;
        int xCCorner = xRCCorner.y;

        // Convolve x(?, ?, d1) with w(:, :, d1, d2) to get y(yR, yC, d2).
        // ? = to be determined. : = across all values in that axis.
        float dotProd = 0.0;
        for (int wR = 0; wR < ${filterHeight}; wR++) {
          int xR = xRCorner + wR * ${dilationHeight};

          if (xR < 0 || xR >= ${convInfo.inHeight}) {
            continue;
          }

          for (int wC = 0; wC < ${filterWidth}; wC++) {
            int xC = xCCorner + wC * ${dilationWidth};

            if (xC < 0 || xC >= ${convInfo.inWidth}) {
              continue;
            }

            for (int d1 = 0; d1 < ${inputDepthNearestVec4}; d1 += 4) {
              vec4 wValues = vec4(
                getW(wR, wC, d1, d2),
                getW(wR, wC, d1 + 1, d2),
                getW(wR, wC, d1 + 2, d2),
                getW(wR, wC, d1 + 3, d2)
              );

              if (${isChannelsLast}) {
                vec4 xValues = vec4(
                  getX(batch, xR, xC, d1),
                  getX(batch, xR, xC, d1 + 1),
                  getX(batch, xR, xC, d1 + 2),
                  getX(batch, xR, xC, d1 + 3)
                );
                dotProd += dot(xValues, wValues);
              } else {
                vec4 xValues = vec4(
                  getX(batch, d1, xR, xC),
                  getX(batch, d1 + 1, xR, xC),
                  getX(batch, d1 + 2, xR, xC),
                  getX(batch, d1 + 3, xR, xC)
                );
                dotProd += dot(xValues, wValues);
              }
            }

            if (${inputDepthVec4Remainder === 1}) {

              if (${isChannelsLast}) {
                dotProd +=
                    getX(batch, xR, xC, ${inputDepthNearestVec4}) *
                    getW(wR, wC, ${inputDepthNearestVec4}, d2);
              } else {
                dotProd +=
                    getX(batch, ${inputDepthNearestVec4}, xR, xC) *
                    getW(wR, wC, ${inputDepthNearestVec4}, d2);
              }

            } else if (${inputDepthVec4Remainder === 2}) {
              vec2 wValues = vec2(
                getW(wR, wC, ${inputDepthNearestVec4}, d2),
                getW(wR, wC, ${inputDepthNearestVec4} + 1, d2)
              );

              if (${isChannelsLast}) {
                vec2 xValues = vec2(
                  getX(batch, xR, xC, ${inputDepthNearestVec4}),
                  getX(batch, xR, xC, ${inputDepthNearestVec4} + 1)
                );
                dotProd += dot(xValues, wValues);
              } else {
                vec2 xValues = vec2(
                  getX(batch, ${inputDepthNearestVec4}, xR, xC),
                  getX(batch, ${inputDepthNearestVec4} + 1, xR, xC)
                );
                dotProd += dot(xValues, wValues);
              }

            } else if (${inputDepthVec4Remainder === 3}) {
              vec3 wValues = vec3(
                getW(wR, wC, ${inputDepthNearestVec4}, d2),
                getW(wR, wC, ${inputDepthNearestVec4} + 1, d2),
                getW(wR, wC, ${inputDepthNearestVec4} + 2, d2)
              );

              if (${isChannelsLast}) {
                vec3 xValues = vec3(
                  getX(batch, xR, xC, ${inputDepthNearestVec4}),
                  getX(batch, xR, xC, ${inputDepthNearestVec4} + 1),
                  getX(batch, xR, xC, ${inputDepthNearestVec4} + 2)
                );
                dotProd += dot(xValues, wValues);
              } else {
                vec3 xValues = vec3(
                  getX(batch, ${inputDepthNearestVec4}, xR, xC),
                  getX(batch, ${inputDepthNearestVec4} + 1, xR, xC),
                  getX(batch, ${inputDepthNearestVec4} + 2, xR, xC)
                );
                dotProd += dot(xValues, wValues);
              }

            }
          }
        }

        float result = dotProd;
        ${addBiasSnippet}
        ${applyActivationSnippet}
        setOutput(result);
      }
    `;
        }
    }
    class Conv3DProgram {
        constructor(convInfo) {
            this.variableNames = ['x', 'W'];
            this.outputShape = convInfo.outShape;
            const padFront = convInfo.padInfo.front;
            const padTop = convInfo.padInfo.top;
            const padLeft = convInfo.padInfo.left;
            const strideDepth = convInfo.strideDepth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const dilationDepth = convInfo.dilationDepth;
            const dilationHeight = convInfo.dilationHeight;
            const dilationWidth = convInfo.dilationWidth;
            const filterDepth = convInfo.filterDepth;
            const filterHeight = convInfo.filterHeight;
            const filterWidth = convInfo.filterWidth;
            const inputDepthNearestVec4 = Math.floor(convInfo.inChannels / 4) * 4;
            const inputDepthVec4Remainder = convInfo.inChannels % 4;
            this.userCode = `
      const ivec3 strides = ivec3(${strideDepth}, ${strideHeight}, ${strideWidth});
      const ivec3 pads = ivec3(${padFront}, ${padTop}, ${padLeft});

      void main() {
        ivec5 coords = getOutputCoords();
        int batch = coords.x;
        int d2 = coords.u;

        ivec3 xFRCCorner = ivec3(coords.y, coords.z, coords.w) * strides - pads;
        int xFCorner = xFRCCorner.x;
        int xRCorner = xFRCCorner.y;
        int xCCorner = xFRCCorner.z;

        // Convolve x(?, ?, ?, d1) with w(:, :, :, d1, d2) to get
        // y(yF, yR, yC, d2). ? = to be determined. : = across all
        // values in that axis.
        float dotProd = 0.0;
        for (int wF = 0; wF < ${filterDepth}; wF++) {
          int xF = xFCorner + wF * ${dilationDepth};

          if (xF < 0 || xF >= ${convInfo.inDepth}) {
            continue;
          }

          for (int wR = 0; wR < ${filterHeight}; wR++) {
            int xR = xRCorner + wR * ${dilationHeight};

            if (xR < 0 || xR >= ${convInfo.inHeight}) {
              continue;
            }

            for (int wC = 0; wC < ${filterWidth}; wC++) {
              int xC = xCCorner + wC * ${dilationWidth};

              if (xC < 0 || xC >= ${convInfo.inWidth}) {
                continue;
              }

              for (int d1 = 0; d1 < ${inputDepthNearestVec4}; d1 += 4) {
                vec4 xValues = vec4(
                  getX(batch, xF, xR, xC, d1),
                  getX(batch, xF, xR, xC, d1 + 1),
                  getX(batch, xF, xR, xC, d1 + 2),
                  getX(batch, xF, xR, xC, d1 + 3)
                );
                vec4 wValues = vec4(
                  getW(wF, wR, wC, d1, d2),
                  getW(wF, wR, wC, d1 + 1, d2),
                  getW(wF, wR, wC, d1 + 2, d2),
                  getW(wF, wR, wC, d1 + 3, d2)
                );

                dotProd += dot(xValues, wValues);
              }

              if (${inputDepthVec4Remainder === 1}) {
                dotProd +=
                  getX(batch, xF, xR, xC, ${inputDepthNearestVec4}) *
                  getW(wF, wR, wC, ${inputDepthNearestVec4}, d2);
              } else if (${inputDepthVec4Remainder === 2}) {
                vec2 xValues = vec2(
                  getX(batch, xF, xR, xC, ${inputDepthNearestVec4}),
                  getX(batch, xF, xR, xC, ${inputDepthNearestVec4} + 1)
                );
                vec2 wValues = vec2(
                  getW(wF, wR, wC, ${inputDepthNearestVec4}, d2),
                  getW(wF, wR, wC, ${inputDepthNearestVec4} + 1, d2)
                );
                dotProd += dot(xValues, wValues);
              } else if (${inputDepthVec4Remainder === 3}) {
                vec3 xValues = vec3(
                  getX(batch, xF, xR, xC, ${inputDepthNearestVec4}),
                  getX(batch, xF, xR, xC, ${inputDepthNearestVec4} + 1),
                  getX(batch, xF, xR, xC, ${inputDepthNearestVec4} + 2)
                );
                vec3 wValues = vec3(
                  getW(wF, wR, wC, ${inputDepthNearestVec4}, d2),
                  getW(wF, wR, wC, ${inputDepthNearestVec4} + 1, d2),
                  getW(wF, wR, wC, ${inputDepthNearestVec4} + 2, d2)
                );
                dotProd += dot(xValues, wValues);
              }
            }
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class Im2ColPackedProgram {
        constructor(outputShape, convInfo) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.customUniforms = [
                { name: 'inputShape', type: 'ivec3' },
                { name: 'pad', type: 'ivec2' },
                { name: 'stride', type: 'ivec2' },
                { name: 'dilation', type: 'ivec2' },
                { name: 'inChannels', type: 'int' },
                { name: 'itemsPerBlockRow', type: 'int' },
                { name: 'outWidth', type: 'int' },
            ];
            this.outputShape = outputShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            const { dataFormat } = convInfo;
            const glsl = getGlslDifferences();
            const isChannelsLast = dataFormat === 'channelsLast';
            const rowDim = isChannelsLast ? 0 : 1;
            const colDim = isChannelsLast ? 1 : 2;
            const boundsCheckingSnippet = this.enableShapeUniforms ?
                'if(blockIndex < outShape[1] && pos < outShape[0]) {' :
                `if(blockIndex < ${outputShape[1]} && pos < ${outputShape[0]}) {`;
            let unrolled = ``;
            for (let row = 0; row <= 1; row++) {
                for (let col = 0; col <= 1; col++) {
                    unrolled += `
          blockIndex = rc.y + ${col};
          pos = rc.x + ${row};

          ${boundsCheckingSnippet}
            offsetY = int(blockIndex / outWidth) * stride[0] - pad[0];
            d0 = offsetY + dilation[0] * (pos / itemsPerBlockRow);

            if(d0 < inputShape[${rowDim}] && d0 >= 0) {
              // Use custom imod instead mod. On Intel GPU, mod may generate
              // unexpected value.
              // https://github.com/tensorflow/tfjs/issues/5447
              offsetX = imod(blockIndex, outWidth) * stride[1] - pad[1];
              d1 = offsetX + dilation[1] * (imod(pos, itemsPerBlockRow) /
                  inChannels);

              if(d1 < inputShape[${colDim}] && d1 >= 0) {

                ch = imod(pos, inChannels);

                if (${isChannelsLast}) {
                  innerDims = vec2(d1, ch);
                  result[${row * 2 + col}] = getChannel(
                    getA(d0, int(innerDims.x),
                    int(innerDims.y)), innerDims);
                } else {
                  innerDims = vec2(d0, d1);
                  result[${row * 2 + col}] = getChannel(
                    getA(ch, int(innerDims.x),
                    int(innerDims.y)), innerDims);
                }
              }
            }
          }
        `;
                }
            }
            this.userCode = `
      void main() {
        ivec2 rc = getOutputCoords();

        vec4 result = vec4(0);

        int blockIndex, pos, offsetY, d0, offsetX, d1, ch;
        vec2 innerDims;

        ${unrolled}

        ${glsl.output} = result;
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fitPreluActivationWeightsIntoNhwcFormat(alpha, outputShape, isChannelsLast, backend) {
        // PReLU's activation weights could be a scalar, a 1-D tensor or a 3-D
        // tensor.
        const alphaShape = alpha.shape;
        assert$4(alphaShape.length <= 1 || alphaShape.length === 3, () => `WebGL conv2d only supports scalar, 1-D Tensor or 3-D ` +
            `Tensor PReLU activation weights but got a tensor of ` +
            `rank-${alphaShape.length}.`);
        if (alphaShape.length === 1) {
            const outputChannels = isChannelsLast ? outputShape[3] : outputShape[1];
            assert$4(alphaShape[0] === 1 || alphaShape[0] === outputChannels, () => `WebGL conv2d PReLU activation weights (${alphaShape}) is ` +
                `not compatible with the number of output channels ` +
                `(${outputChannels}).`);
        }
        else if (alphaShape.length === 3) {
            try {
                assertAndGetBroadcastShape(alphaShape, outputShape);
            }
            catch (e) {
                const errMsg = `WebGL conv2d PReLU activation weights (${alphaShape}) ` +
                    `is not compatible with the output shape of the conv2d ` +
                    `(${outputShape}).`;
                throw Error(errMsg);
            }
            if (!isChannelsLast) {
                // If PReLU's activation weights is NCHW format, then convert it to NHWC.
                return transpose({ inputs: { x: alpha }, backend, attrs: { perm: [1, 2, 0] } });
            }
        }
        return alpha;
    }
    // For 1x1 kernels that iterate through every point in the input, convolution
    // can be expressed as matrix multiplication (without need for memory
    // remapping).
    function conv2dByMatMul({ x, filter, convInfo, backend, bias = null, preluActivationWeights = null, leakyreluAlpha = 0, activation = null }) {
        // Reshapes conv2D input to 2D tensors, uses matMul and then reshape the
        // result from 2D to 4D.
        const xShape = x.shape;
        const xTexData = backend.texData.get(x.dataId);
        const sharedMatMulDim = convInfo.inChannels;
        const outerShapeX = xShape[0] * xShape[1] * xShape[2];
        const outerShapeFilter = convInfo.outChannels;
        const isChannelsLast = convInfo.dataFormat === 'channelsLast';
        const transposeA = false;
        const transposeB = false;
        let out;
        const intermediates = [];
        if (preluActivationWeights != null) {
            const preluActivationWeightsInNhwcFormat = fitPreluActivationWeightsIntoNhwcFormat(preluActivationWeights, convInfo.outShape, isChannelsLast, backend);
            if (preluActivationWeightsInNhwcFormat.dataId !==
                preluActivationWeights.dataId) {
                // preluActivationWeightsInNhwcFormat is a new tensor, temporarily
                // generated to be compatible with the following matmul computation.
                intermediates.push(preluActivationWeightsInNhwcFormat);
                preluActivationWeights = preluActivationWeightsInNhwcFormat;
            }
        }
        // TODO: Once reduction ops are packed, batchMatMul will always be packed
        // and we can remove this condition.
        const batchMatMulWillBeUnpacked = (outerShapeX === 1 || outerShapeFilter === 1) &&
            sharedMatMulDim > MATMUL_SHARED_DIM_THRESHOLD;
        // The algorithm in the if condition assumes (1) the output will be packed,
        // (2) x is packed, (3) x isChannelsLast, (4)  x's packed texture is already
        // on GPU, (5) col is odd, (6) the width, height and inChannels are the same
        // for xTexData.shape and xShape.
        const canOptimize = !batchMatMulWillBeUnpacked && xTexData.isPacked &&
            isChannelsLast && xTexData.texture != null && xShape[2] % 2 !== 0 &&
            arraysEqual(xTexData.shape.slice(-3), xShape.slice(-3));
        if (canOptimize) {
            // We avoid expensive packed 2x2 reshape by padding col count to next,
            // even number. When col is odd, the result of packed batchMatMul is
            // the same (has the same texture layout and and values in the texture) as
            // it is for next even col. We make the odd-cols tensor to look like
            // even-cols tensor before the operation and, after the batchMatMul,
            // fix the even-cols result to have odd number of cols.
            const targetShape = xShape[0] * xShape[1] * (xShape[2] + 1);
            const xReshaped = {
                dataId: x.dataId,
                shape: [1, targetShape, convInfo.inChannels],
                dtype: x.dtype
            };
            // xTexData.shape gets referenced from GPGPUBinary.inShapeInfos.
            // Decrementing col count, after batchMatMul->...->compileProgram leads to
            // invalid col count within the reference in GPGPUBinary.inShapeInfos.
            // Alternative fix would be to provide a copy to GPGPUBinary.inShapeInfos
            // in compileProgram method, but that would affect compilation of all
            // programs - instead, provide a copy here, with even col count, before
            // calling batchMatMul->...->compileProgram and after that, the original
            // xTexData.shape is restored.
            const originalXTexDataShape = xTexData.shape;
            xTexData.shape = xTexData.shape.slice();
            xTexData.shape[xTexData.shape.length - 2]++;
            assert$4(isReshapeFree(xTexData.shape, xReshaped.shape), () => `packed reshape ${xTexData.shape} to ${xReshaped.shape} isn't free`);
            const filterReshaped = reshape({
                inputs: { x: filter },
                backend,
                attrs: { shape: [1, convInfo.inChannels, convInfo.outChannels] }
            });
            intermediates.push(filterReshaped);
            const pointwiseConv = batchMatMulImpl({
                a: xReshaped,
                b: filterReshaped,
                backend,
                transposeA,
                transposeB,
                bias,
                activation,
                preluActivationWeights,
                leakyreluAlpha
            });
            const pointwiseConvTexData = backend.texData.get(pointwiseConv.dataId);
            assert$4(pointwiseConvTexData.isPacked, () => 'batchMatMul result is expected to be packed');
            // Restore the input shape to original.
            xTexData.shape = originalXTexDataShape;
            // Set the output shape - there is no need for expensive reshape as data
            // layout is already correct.
            pointwiseConvTexData.shape = convInfo.outShape;
            out = identity({ inputs: { x: pointwiseConv }, backend });
            out.shape = convInfo.outShape;
            intermediates.push(pointwiseConv);
        }
        else {
            const xInNhwcFormat = isChannelsLast ?
                x :
                transpose({ inputs: { x }, backend, attrs: { perm: [0, 2, 3, 1] } });
            const xInNhwcFormatShape = xInNhwcFormat.shape;
            const targetShape = xInNhwcFormatShape[0] * xInNhwcFormatShape[1] * xInNhwcFormatShape[2];
            const xReshaped = reshape({
                inputs: { x: xInNhwcFormat },
                backend,
                attrs: { shape: [1, targetShape, convInfo.inChannels] }
            });
            const filterReshaped = reshape({
                inputs: { x: filter },
                backend,
                attrs: { shape: [1, convInfo.inChannels, convInfo.outChannels] }
            });
            const result = batchMatMulImpl({
                a: xReshaped,
                b: filterReshaped,
                transposeA,
                transposeB,
                backend,
                bias,
                activation,
                preluActivationWeights,
                leakyreluAlpha
            });
            const outInNHWCFormatShape = [
                convInfo.batchSize, convInfo.outHeight, convInfo.outWidth,
                convInfo.outChannels
            ];
            const outInNHWCFormat = reshape({ inputs: { x: result }, backend, attrs: { shape: outInNHWCFormatShape } });
            // If the data format is NCHW, then convert the output to be NCHW format.
            out = isChannelsLast ? outInNHWCFormat : transpose({
                inputs: { x: outInNHWCFormat },
                backend,
                attrs: { perm: [0, 3, 1, 2] }
            });
            if (!isChannelsLast) {
                intermediates.push(xInNhwcFormat);
                intermediates.push(outInNHWCFormat);
            }
            intermediates.push(xReshaped);
            intermediates.push(filterReshaped);
            intermediates.push(result);
        }
        for (const i of intermediates) {
            backend.disposeIntermediateTensorInfo(i);
        }
        return out;
    }
    // Implements the im2row algorithm as outlined in "High Performance
    // Convolutional Neural Networks for Document Processing" (Suvisoft, 2006)
    function conv2dWithIm2Row({ x, filter, convInfo, backend, bias = null, preluActivationWeights = null, leakyreluAlpha = 0, activation = null }) {
        // Rearranges conv2d input so each block to be convolved over forms the
        // column of a new matrix with shape [filterWidth * filterHeight *
        // inChannels, outHeight * outWidth]. The filter is also rearranged so each
        // output channel forms a row of a new matrix with shape [outChannels,
        // filterWidth * filterHeight * inChannels]. The convolution is then
        // computed by multiplying these matrices and reshaping the result.
        const { filterWidth, filterHeight, inChannels, outWidth, outHeight, dataFormat } = convInfo;
        const isChannelsLast = dataFormat === 'channelsLast';
        const sharedDim = filterWidth * filterHeight * inChannels;
        const numCols = outHeight * outWidth;
        const x2ColShape = [sharedDim, numCols];
        const transposeA = true;
        const transposeB = false;
        const intermediates = [];
        if (preluActivationWeights != null) {
            const preluActivationWeightsInNhwcFormat = fitPreluActivationWeightsIntoNhwcFormat(preluActivationWeights, convInfo.outShape, isChannelsLast, backend);
            if (preluActivationWeightsInNhwcFormat.dataId !==
                preluActivationWeights.dataId) {
                // preluActivationWeightsInNhwcFormat is a new tensor, temporarily
                // generated to be compatible with the following matmul computation.
                intermediates.push(preluActivationWeightsInNhwcFormat);
                preluActivationWeights = preluActivationWeightsInNhwcFormat;
            }
        }
        const xSqueezed = reshape({ inputs: { x }, backend, attrs: { shape: x.shape.slice(1) } });
        const w2Row = reshape({
            inputs: { x: filter },
            backend,
            attrs: { shape: [1, sharedDim, sizeFromShape(filter.shape) / sharedDim] }
        });
        intermediates.push(xSqueezed);
        intermediates.push(w2Row);
        const im2ColProgram = new Im2ColPackedProgram(x2ColShape, convInfo);
        const customValues = [
            xSqueezed.shape, [convInfo.padInfo.top, convInfo.padInfo.left],
            [convInfo.strideHeight, convInfo.strideWidth],
            [convInfo.dilationHeight, convInfo.dilationWidth], [convInfo.inChannels],
            [convInfo.filterWidth * convInfo.inChannels], [convInfo.outWidth]
        ];
        const im2Col = backend.runWebGLProgram(im2ColProgram, [xSqueezed], 'float32', customValues);
        const im2ColReshaped = reshape({
            inputs: { x: im2Col },
            backend,
            attrs: { shape: [1, x2ColShape[0], x2ColShape[1]] }
        });
        intermediates.push(im2Col);
        intermediates.push(im2ColReshaped);
        const hasBias = bias != null;
        const hasPreluActivationWeights = preluActivationWeights != null;
        const hasLeakyreluAlpha = activation === 'leakyrelu';
        const fusedActivation = activation ? mapActivationToShaderProgram(activation, true) : null;
        const matmulProgram = new MatMulPackedProgram(im2ColReshaped.shape, w2Row.shape, [1, numCols, convInfo.outChannels], transposeA, transposeB, hasBias, fusedActivation, hasPreluActivationWeights, hasLeakyreluAlpha);
        const inputs = [im2ColReshaped, w2Row];
        if (bias) {
            inputs.push(bias);
        }
        if (hasPreluActivationWeights) {
            inputs.push(preluActivationWeights);
        }
        if (hasLeakyreluAlpha) {
            const $leakyreluAlpha = backend.makeTensorInfo([], 'float32', createScalarValue(leakyreluAlpha, 'float32'));
            inputs.push($leakyreluAlpha);
            intermediates.push($leakyreluAlpha);
        }
        const product = backend.runWebGLProgram(matmulProgram, inputs, 'float32');
        const outInNHWCFormatShape = [1, outHeight, outWidth, convInfo.outChannels];
        const outInNHWCFormat = reshape({ inputs: { x: product }, backend, attrs: { shape: outInNHWCFormatShape } });
        // If the data format is NCHW, then convert the output to be NCHW format.
        const out = isChannelsLast ?
            outInNHWCFormat :
            transpose({ inputs: { x: outInNHWCFormat }, backend, attrs: { perm: [0, 3, 1, 2] } });
        if (!isChannelsLast) {
            intermediates.push(outInNHWCFormat);
        }
        intermediates.push(product);
        for (const i of intermediates) {
            backend.disposeIntermediateTensorInfo(i);
        }
        return out;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv2d(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter } = inputs;
        const { strides, pad, dataFormat, dilations, dimRoundingMode } = attrs;
        const $dataFormat = convertConv2DDataFormat(dataFormat);
        const convInfo = computeConv2DInfo(x.shape, filter.shape, strides, dilations, pad, dimRoundingMode, false /* depthwise */, $dataFormat);
        let out;
        if (convInfo.filterHeight === 1 && convInfo.filterWidth === 1 &&
            convInfo.dilationHeight === 1 && convInfo.dilationWidth === 1 &&
            convInfo.strideHeight === 1 && convInfo.strideWidth === 1 &&
            (convInfo.padInfo.type === 'SAME' || convInfo.padInfo.type === 'VALID')) {
            out = conv2dByMatMul({ x, filter, convInfo, backend });
        }
        else if (env$1().getBool('WEBGL_CONV_IM2COL') && x.shape[0] === 1) {
            out = conv2dWithIm2Row({ x, filter, convInfo, backend });
        }
        else {
            const program = new Conv2DProgram(convInfo);
            out = backend.runWebGLProgram(program, [x, filter], 'float32');
        }
        const outReshaped = reshape({ inputs: { x: out }, backend, attrs: { shape: convInfo.outShape } });
        backend.disposeIntermediateTensorInfo(out);
        return outReshaped;
    }
    const conv2DConfig = {
        kernelName: Conv2D$1,
        backendName: 'webgl',
        kernelFunc: conv2d,
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class Conv2DDerFilterProgram {
        constructor(convInfo) {
            this.variableNames = ['x', 'dy'];
            this.outputShape = convInfo.filterShape;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const padTop = convInfo.padInfo.top;
            const padLeft = convInfo.padInfo.left;
            const isChannelsLast = convInfo.dataFormat === 'channelsLast';
            this.userCode = `
      void main() {
        ivec4 coords = getOutputCoords();
        int wR = coords.x;
        int wC = coords.y;
        int d1 = coords.z;
        int d2 = coords.w;

        // Convolve x(?, ?, d1) with dy(:, :, d2) to get dw(wR, wC, d1, d2).
        // ? = to be determined. : = across all values in that axis.
        float dotProd = 0.0;

        for (int b = 0; b < ${convInfo.batchSize}; b++) {
          for (int yR = 0; yR < ${convInfo.outHeight}; yR++) {
            int xR = wR + yR * ${strideHeight} - ${padTop};

            if (xR < 0 || xR >= ${convInfo.inHeight}) {
              continue;
            }

            for (int yC = 0; yC < ${convInfo.outWidth}; yC++) {
              int xC = wC + yC * ${strideWidth} - ${padLeft};

              if (xC < 0 || xC >= ${convInfo.inWidth}) {
                continue;
              }

              if (${isChannelsLast}) {
                float dyValue = getDy(b, yR, yC, d2);
                float xValue = getX(b, xR, xC, d1);
                dotProd += (xValue * dyValue);
              } else {
                float dyValue = getDy(b, d2, yR, yC);
                float xValue = getX(b, d1, xR, xC);
                dotProd += (xValue * dyValue);
              }

            }
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }
    class Conv2DDerInputProgram {
        constructor(convInfo) {
            this.variableNames = ['dy', 'W'];
            this.outputShape = convInfo.inShape;
            const filterHeight = convInfo.filterHeight;
            const filterWidth = convInfo.filterWidth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const isChannelsLast = convInfo.dataFormat === 'channelsLast';
            const padTop = filterHeight - 1 - convInfo.padInfo.top;
            const padLeft = filterWidth - 1 - convInfo.padInfo.left;
            const rowDim = isChannelsLast ? 1 : 2;
            const colDim = isChannelsLast ? 2 : 3;
            const channelDim = isChannelsLast ? 3 : 1;
            this.userCode = `
      const ivec2 pads = ivec2(${padTop}, ${padLeft});

      void main() {
        ivec4 coords = getOutputCoords();
        int batch = coords[0];
        int d1 = coords[${channelDim}];

        ivec2 dyCorner = ivec2(coords[${rowDim}], coords[${colDim}]) - pads;
        int dyRCorner = dyCorner.x;
        int dyCCorner = dyCorner.y;

        // Convolve dy(?, ?, d2) with w(:, :, d1, d2) to compute dx(xR, xC, d1).
        // ? = to be determined. : = across all values in that axis.
        float dotProd = 0.0;
        for (int wR = 0; wR < ${filterHeight}; wR++) {
          float dyR = float(dyRCorner + wR) / ${strideHeight}.0;

          if (dyR < 0.0 || dyR >= ${convInfo.outHeight}.0 || fract(dyR) > 0.0) {
            continue;
          }
          int idyR = int(dyR);

          int wRPerm = ${filterHeight} - 1 - wR;

          for (int wC = 0; wC < ${filterWidth}; wC++) {
            float dyC = float(dyCCorner + wC) / ${strideWidth}.0;

            if (dyC < 0.0 || dyC >= ${convInfo.outWidth}.0 ||
                fract(dyC) > 0.0) {
              continue;
            }
            int idyC = int(dyC);

            int wCPerm = ${filterWidth} - 1 - wC;

            for (int d2 = 0; d2 < ${convInfo.outChannels}; d2++) {

              if (${isChannelsLast}) {
                float xValue = getDy(batch, idyR, idyC, d2);
                float wValue = getW(wRPerm, wCPerm, d1, d2);
                dotProd += xValue * wValue;
              } else {
                float xValue = getDy(batch, d2, idyR, idyC);
                float wValue = getW(wRPerm, wCPerm, d1, d2);
                dotProd += xValue * wValue;
              }

            }
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }
    class Conv3DDerFilterProgram {
        constructor(convInfo) {
            this.variableNames = ['x', 'dy'];
            this.outputShape = convInfo.filterShape;
            const strideDepth = convInfo.strideDepth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const padFront = convInfo.padInfo.front;
            const padTop = convInfo.padInfo.top;
            const padLeft = convInfo.padInfo.left;
            this.userCode = `
      void main() {
        ivec5 coords = getOutputCoords();
        int wF = coords.x;
        int wR = coords.y;
        int wC = coords.z;
        int d1 = coords.w;
        int d2 = coords.u;

        float dotProd = 0.0;

        for (int b = 0; b < ${convInfo.batchSize}; b++) {
          for (int yF = 0; yF < ${convInfo.outDepth}; yF++) {
            int xF = wF + yF * ${strideDepth} - ${padFront};

            if (xF < 0 || xF >= ${convInfo.inDepth}) {
              continue;
            }

            for (int yR = 0; yR < ${convInfo.outHeight}; yR++) {
              int xR = wR + yR * ${strideHeight} - ${padTop};

              if (xR < 0 || xR >= ${convInfo.inHeight}) {
                continue;
              }

              for (int yC = 0; yC < ${convInfo.outWidth}; yC++) {
                int xC = wC + yC * ${strideWidth} - ${padLeft};

                if (xC < 0 || xC >= ${convInfo.inWidth}) {
                  continue;
                }

                float dyValue = getDy(b, yF, yR, yC, d2);
                float xValue = getX(b, xF, xR, xC, d1);
                dotProd += (xValue * dyValue);
              }
            }
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }
    class Conv3DDerInputProgram {
        constructor(convInfo) {
            this.variableNames = ['dy', 'W'];
            this.outputShape = convInfo.inShape;
            const filterDepth = convInfo.filterDepth;
            const filterHeight = convInfo.filterHeight;
            const filterWidth = convInfo.filterWidth;
            const strideDepth = convInfo.strideDepth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const padFront = filterDepth - 1 - convInfo.padInfo.front;
            const padTop = filterHeight - 1 - convInfo.padInfo.top;
            const padLeft = filterWidth - 1 - convInfo.padInfo.left;
            this.userCode = `
      const ivec3 pads = ivec3(${padFront}, ${padTop}, ${padLeft});

      void main() {
        ivec5 coords = getOutputCoords();
        int batch = coords.x;
        int d1 = coords.u;


        ivec3 dyCorner = ivec3(coords.y, coords.z, coords.w) - pads;
        int dyFCorner = dyCorner.x;
        int dyRCorner = dyCorner.y;
        int dyCCorner = dyCorner.z;

        float dotProd = 0.0;
        for (int wF = 0; wF < ${filterDepth}; wF++) {
          float dyF = float(dyFCorner + wF) / ${strideDepth}.0;

          if (dyF < 0.0 || dyF >= ${convInfo.outDepth}.0 || fract(dyF) > 0.0) {
            continue;
          }
          int idyF = int(dyF);

          int wFPerm = ${filterDepth} - 1 - wF;

          for (int wR = 0; wR < ${filterHeight}; wR++) {
            float dyR = float(dyRCorner + wR) / ${strideHeight}.0;

            if (dyR < 0.0 || dyR >= ${convInfo.outHeight}.0 ||
              fract(dyR) > 0.0) {
              continue;
            }
            int idyR = int(dyR);

            int wRPerm = ${filterHeight} - 1 - wR;

            for (int wC = 0; wC < ${filterWidth}; wC++) {
              float dyC = float(dyCCorner + wC) / ${strideWidth}.0;

              if (dyC < 0.0 || dyC >= ${convInfo.outWidth}.0 ||
                  fract(dyC) > 0.0) {
                continue;
              }
              int idyC = int(dyC);

              int wCPerm = ${filterWidth} - 1 - wC;

              for (int d2 = 0; d2 < ${convInfo.outChannels}; d2++) {
                float xValue = getDy(batch, idyF, idyR, idyC, d2);
                float wValue = getW(wFPerm, wRPerm, wCPerm, d1, d2);
                dotProd += xValue * wValue;
              }
            }
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv2DBackpropFilter(args) {
        const { inputs, backend, attrs } = args;
        const { x, dy } = inputs;
        const { strides, pad, dataFormat, dimRoundingMode, filterShape } = attrs;
        const $dataFormat = convertConv2DDataFormat(dataFormat);
        const convInfo = computeConv2DInfo(x.shape, filterShape, strides, 1 /* dilations */, pad, dimRoundingMode, false /* depthwise */, $dataFormat);
        const program = new Conv2DDerFilterProgram(convInfo);
        return backend.runWebGLProgram(program, [x, dy], 'float32');
    }
    const conv2DBackpropFilterConfig = {
        kernelName: Conv2DBackpropFilter,
        backendName: 'webgl',
        kernelFunc: conv2DBackpropFilter,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv2DBackpropInput(args) {
        const { inputs, backend, attrs } = args;
        const { dy, filter } = inputs;
        const { inputShape, strides, pad, dataFormat, dimRoundingMode } = attrs;
        const $dataFormat = convertConv2DDataFormat(dataFormat);
        const convInfo = computeConv2DInfo(inputShape, filter.shape, strides, 1 /* dilations */, pad, dimRoundingMode, false, $dataFormat);
        const program = new Conv2DDerInputProgram(convInfo);
        return backend.runWebGLProgram(program, [dy, filter], 'float32');
    }
    const conv2DBackpropInputConfig = {
        kernelName: Conv2DBackpropInput,
        backendName: 'webgl',
        kernelFunc: conv2DBackpropInput,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv3D(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter } = inputs;
        const { strides, pad, dilations } = attrs;
        const convInfo = computeConv3DInfo(x.shape, filter.shape, strides, dilations, pad);
        const program = new Conv3DProgram(convInfo);
        return backend.runWebGLProgram(program, [x, filter], 'float32');
    }
    const conv3DConfig = {
        kernelName: Conv3D$1,
        backendName: 'webgl',
        kernelFunc: conv3D,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv3DBackpropFilterV2(args) {
        const { inputs, backend, attrs } = args;
        const { x, dy } = inputs;
        const { strides, pad, filterShape } = attrs;
        const convInfo = computeConv3DInfo(x.shape, filterShape, strides, 1 /* dilations */, pad);
        const program = new Conv3DDerFilterProgram(convInfo);
        return backend.runWebGLProgram(program, [x, dy], 'float32');
    }
    const conv3DBackpropFilterV2Config = {
        kernelName: Conv3DBackpropFilterV2,
        backendName: 'webgl',
        kernelFunc: conv3DBackpropFilterV2
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function conv3DBackpropInput(args) {
        const { inputs, backend, attrs } = args;
        const { dy, filter } = inputs;
        const { pad, strides, inputShape } = attrs;
        const convInfo = computeConv3DInfo(inputShape, filter.shape, strides, 1 /* dilations */, pad);
        const program = new Conv3DDerInputProgram(convInfo);
        return backend.runWebGLProgram(program, [dy, filter], 'float32');
    }
    const conv3DBackpropInputConfig = {
        kernelName: Conv3DBackpropInputV2,
        backendName: 'webgl',
        kernelFunc: conv3DBackpropInput,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const COS = CHECK_NAN_SNIPPET_UNARY + `
  return cos(x);
`;
    const cos = unaryKernelFunc({ opSnippet: COS });
    const cosConfig = {
        kernelName: Cos,
        backendName: 'webgl',
        kernelFunc: cos,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const COSH = `
  float e2x = exp(-x);
  return (e2x + 1.0 / e2x) / 2.0;
`;
    const cosh = unaryKernelFunc({ opSnippet: COSH });
    const coshConfig = {
        kernelName: Cosh,
        backendName: 'webgl',
        kernelFunc: cosh,
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class CropAndResizeProgram {
        constructor(imageShape, boxShape, cropSize, method, extrapolationValue) {
            this.variableNames = ['Image', 'Boxes', 'BoxInd'];
            this.outputShape = [];
            const [batch, imageHeight, imageWidth, depth] = imageShape;
            const [numBoxes,] = boxShape;
            const [cropHeight, cropWidth] = cropSize;
            this.outputShape = [numBoxes, cropHeight, cropWidth, depth];
            const methodId = method === 'bilinear' ? 1 : 0;
            const [inputHeightFloat, inputWidthFloat] = [`${imageHeight - 1}.0`, `${imageWidth - 1}.0`];
            const [heightRatio, heightScale, inY] = cropHeight > 1 ?
                [
                    `${(imageHeight - 1) / (cropHeight - 1)}`,
                    '(y2-y1) * height_ratio',
                    `y1*${inputHeightFloat} + float(y)*(height_scale)`,
                ] :
                [
                    '0.0',
                    '0.0',
                    `0.5 * (y1+y2) * ${inputHeightFloat}`,
                ];
            const [widthRatio, widthScale, inX] = cropWidth > 1 ?
                [
                    `${(imageWidth - 1) / (cropWidth - 1)}`,
                    '(x2-x1) * width_ratio',
                    `x1*${inputWidthFloat} + float(x)*(width_scale)`,
                ] :
                [
                    '0.0',
                    '0.0',
                    `0.5 * (x1+x2) * ${inputWidthFloat}`,
                ];
            // Reference implementation
            // tslint:disable-next-line:max-line-length
            // https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/kernels/crop_and_resize_op_gpu.cu.cc
            this.userCode = `
      const float height_ratio = float(${heightRatio});
      const float width_ratio = float(${widthRatio});
      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int y = coords[1];
        int x = coords[2];
        int d = coords[3];

        // get box vals
        float y1 = getBoxes(b,0);
        float x1 = getBoxes(b,1);
        float y2 = getBoxes(b,2);
        float x2 = getBoxes(b,3);

        // get image in batch index
        int bInd = round(getBoxInd(b));
        if(bInd < 0 || bInd >= ${batch}) {
          return;
        }

        float height_scale = ${heightScale};
        float width_scale = ${widthScale};

        float in_y = ${inY};
        if( in_y < 0.0 || in_y > ${inputHeightFloat} ) {
          setOutput(float(${extrapolationValue}));
          return;
        }
        float in_x = ${inX};
        if( in_x < 0.0 || in_x > ${inputWidthFloat} ) {
          setOutput(float(${extrapolationValue}));
          return;
        }

        vec2 sourceFracIndexCR = vec2(in_x,in_y);
        if(${methodId} == 1) {
          // Compute the four integer indices.
          ivec2 sourceFloorCR = ivec2(sourceFracIndexCR);
          ivec2 sourceCeilCR = ivec2(ceil(sourceFracIndexCR));

          float topLeft = getImage(b, sourceFloorCR.y, sourceFloorCR.x, d);
          float bottomLeft = getImage(b, sourceCeilCR.y, sourceFloorCR.x, d);
          float topRight = getImage(b, sourceFloorCR.y, sourceCeilCR.x, d);
          float bottomRight = getImage(b, sourceCeilCR.y, sourceCeilCR.x, d);

          vec2 fracCR = sourceFracIndexCR - vec2(sourceFloorCR);

          float top = topLeft + (topRight - topLeft) * fracCR.x;
          float bottom = bottomLeft + (bottomRight - bottomLeft) * fracCR.x;
          float newValue = top + (bottom - top) * fracCR.y;
          setOutput(newValue);
        } else {
          // Compute the coordinators of nearest neighbor point.
          ivec2 sourceNearestCR = ivec2(floor(
            sourceFracIndexCR + vec2(0.5,0.5)));
          float newValue = getImage(b, sourceNearestCR.y, sourceNearestCR.x, d);
          setOutput(newValue);
        }
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const cropAndResize = (args) => {
        const { inputs, backend, attrs } = args;
        const { image, boxes, boxInd } = inputs;
        const { cropSize, method, extrapolationValue } = attrs;
        const program = new CropAndResizeProgram(image.shape, boxes.shape, cropSize, method, extrapolationValue);
        return backend.runWebGLProgram(program, [image, boxes, boxInd], 'float32');
    };
    const cropAndResizeConfig = {
        kernelName: CropAndResize,
        backendName: 'webgl',
        kernelFunc: cropAndResize
    };

    var CumOpType;
    (function (CumOpType) {
        CumOpType["Prod"] = "*";
        CumOpType["Sum"] = "+";
    })(CumOpType || (CumOpType = {}));
    class CumProgram {
        constructor(op, shape, exclusive, reverse) {
            this.variableNames = ['x'];
            this.customUniforms = [{ name: 'index', type: 'float' }];
            this.op = op;
            this.outputShape = shape;
            const rank = shape.length;
            const initVal = this.op === CumOpType.Prod ? '1.0' : '0.0';
            const val = exclusive ? initVal : `getX(${getCoords(rank, 'coords', this.op)})`;
            const length = shape[shape.length - 1];
            let condition = '';
            let idxString = '';
            // When exclusive is set, the cum op becomes roll op that copies the
            // value from the previous index based on the direction specified by the
            // reverse flag.
            if (exclusive) {
                condition = reverse ? `end != ${length - 1}` : 'end != 0';
                idxString = reverse ? 'end + 1' : 'end - 1';
            }
            else {
                condition = reverse ? `end + pow2 < ${length}` : 'end >= pow2';
                idxString = (reverse ? 'end + pow2' : 'end - pow2');
            }
            this.userCode = `
      void main() {
        ${getCoordsDataType(rank)} coords = getOutputCoords();
        int end = ${getFinalCoord(rank, 'coords', this.op)};
        float val = ${val};
        int pow2 = int(pow(2.0, index));
        if (${condition}) {
          int idx = ${idxString};
          ${getFinalCoord(rank, 'coords', this.op)} = idx;
          val ${this.op}= getX(${getCoords(rank, 'coords', this.op)});
        }
        setOutput(val);
      }
    `;
        }
    }
    function getCoords(rank, name, op) {
        if (rank === 1) {
            return `${name}`;
        }
        else if (rank === 2) {
            return `${name}.x, ${name}.y`;
        }
        else if (rank === 3) {
            return `${name}.x, ${name}.y, ${name}.z`;
        }
        else if (rank === 4) {
            return `${name}.x, ${name}.y, ${name}.z, ${name}.w`;
        }
        else {
            throw Error(`Cumulative ${op} for rank ${rank} is not yet supported`);
        }
    }
    function getFinalCoord(rank, name, op) {
        if (rank === 1) {
            return `${name}`;
        }
        else if (rank === 2) {
            return `${name}.y`;
        }
        else if (rank === 3) {
            return `${name}.z`;
        }
        else if (rank === 4) {
            return `${name}.w`;
        }
        else {
            throw Error(`Cumulative ${op} for rank ${rank} is not yet supported`);
        }
    }

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function cumImpl(op, x, backend, axis, exclusive, reverse) {
        const xRank = x.shape.length;
        const permutation = getAxesPermutation([axis], xRank);
        let permutedX = x;
        if (permutation != null) {
            permutedX = transpose({ inputs: { x }, backend, attrs: { perm: permutation } });
        }
        const permutedAxis = getInnerMostAxes(1, xRank)[0];
        if (permutedAxis !== xRank - 1) {
            throw new Error(`WebGL cumprod shader expects an inner-most axis=${x.shape.length - 1} ` +
                `but got axis=${axis}`);
        }
        const size = permutedX.shape[permutedAxis];
        let result = identity({ inputs: { x: permutedX }, backend });
        // Use cum parallel algorithm, inspired by:
        // https://developer.nvidia.com/gpugems/gpugems3/part-vi-gpu-computing/chapter-39-parallel-prefix-sum-scan-cuda
        // Note: although the algorithm is called sum, it works for any associtative
        // operator with an identity.
        for (let i = 0; i <= Math.ceil(Math.log2(size)) - 1; i++) {
            const program = new CumProgram(op, permutedX.shape, false, reverse);
            const customValues = [[i]];
            const prevResult = result;
            result =
                backend.runWebGLProgram(program, [result], result.dtype, customValues);
            backend.disposeIntermediateTensorInfo(prevResult);
        }
        // For exclusive cum, shift the end result in the direction of product or sum
        // and add 1 for product or 0 for sum to the front index.
        if (exclusive) {
            const program = new CumProgram(op, permutedX.shape, exclusive, reverse);
            const prevResult = result;
            result = backend.runWebGLProgram(program, [result], result.dtype);
            backend.disposeIntermediateTensorInfo(prevResult);
        }
        if (permutation != null) {
            const reversePermutation = getUndoAxesPermutation(permutation);
            const reverseTransposedResult = transpose({ inputs: { x: result }, backend, attrs: { perm: reversePermutation } });
            backend.disposeIntermediateTensorInfo(result);
            backend.disposeIntermediateTensorInfo(permutedX);
            return reverseTransposedResult;
        }
        return result;
    }

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function cumprod(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, exclusive, reverse } = attrs;
        return cumImpl(CumOpType.Prod, x, backend, axis, exclusive, reverse);
    }
    const cumprodConfig = {
        kernelName: Cumprod,
        backendName: 'webgl',
        kernelFunc: cumprod
    };

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function cumsum(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, exclusive, reverse } = attrs;
        return cumImpl(CumOpType.Sum, x, backend, axis, exclusive, reverse);
    }
    const cumsumConfig = {
        kernelName: Cumsum,
        backendName: 'webgl',
        kernelFunc: cumsum
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function denseBincount(args) {
        const { inputs, backend, attrs } = args;
        const { x, weights } = inputs;
        const { size, binaryOutput } = attrs;
        if (x.shape.length === 1) {
            const xVals = backend.readSync(x.dataId);
            const weightsVals = backend.readSync(weights.dataId);
            const outVals = bincountImplCPU(xVals, weightsVals, weights.dtype, weights.shape, size);
            return backend.makeTensorInfo([size], weights.dtype, outVals);
        }
        else if (x.shape.length === 2) {
            const xBuf = backend.bufferSync(x);
            const weightsBuf = backend.bufferSync(weights);
            const outBuf = bincountReduceImplCPU(xBuf, weightsBuf, size, binaryOutput);
            return backend.makeTensorInfo(outBuf.shape, weights.dtype, outBuf.values);
        }
        throw new Error(`Error in denseBincount: input must be at most rank 2, but got rank` +
            `${x.shape.length}.`);
    }
    const denseBincountConfig = {
        kernelName: DenseBincount,
        backendName: 'webgl',
        kernelFunc: denseBincount
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class DepthToSpaceProgram {
        constructor(outputShape, blockSize, dataFormat) {
            this.variableNames = ['x'];
            this.outputShape = [];
            this.outputShape = outputShape;
            this.blockSize = blockSize;
            this.dataFormat = dataFormat;
            this.userCode = `
    void main() {
      ivec4 coords = getOutputCoords();
      int b = coords[0];
      int h = ${this.getHeightCoordString()};
      int w = ${this.getWidthCoordString()};
      int d = ${this.getDepthCoordString()};

      int in_h = h / ${blockSize};
      int offset_h = imod(h, ${blockSize});
      int in_w = w / ${blockSize};
      int offset_w = imod(w, ${blockSize});
      int offset_d = (offset_h * ${blockSize} + offset_w) *
        ${this.getOutputDepthSize()};
      int in_d = d + offset_d;

      float result = ${this.getInputSamplingString()};
      setOutput(result);
    }
  `;
        }
        getHeightCoordString() {
            if (this.dataFormat === 'NHWC') {
                return `coords[1]`;
            }
            else {
                return `coords[2]`;
            }
        }
        getWidthCoordString() {
            if (this.dataFormat === 'NHWC') {
                return `coords[2]`;
            }
            else {
                return `coords[3]`;
            }
        }
        getDepthCoordString() {
            if (this.dataFormat === 'NHWC') {
                return `coords[3]`;
            }
            else {
                return `coords[1]`;
            }
        }
        getOutputDepthSize() {
            if (this.dataFormat === 'NHWC') {
                return this.outputShape[3];
            }
            else {
                return this.outputShape[1];
            }
        }
        getInputSamplingString() {
            if (this.dataFormat === 'NHWC') {
                return `getX(b, in_h, in_w, in_d)`;
            }
            else {
                return `getX(b, in_d, in_h, in_w)`;
            }
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthToSpace(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { blockSize, dataFormat } = attrs;
        const batchSize = x.shape[0];
        const inputHeight = (dataFormat === 'NHWC') ? x.shape[1] : x.shape[2];
        const inputWidth = (dataFormat === 'NHWC') ? x.shape[2] : x.shape[3];
        const inputDepth = (dataFormat === 'NHWC') ? x.shape[3] : x.shape[1];
        const outputHeight = inputHeight * blockSize;
        const outputWidth = inputWidth * blockSize;
        const outputDepth = inputDepth / (blockSize * blockSize);
        const outputShape = (dataFormat === 'NHWC') ?
            [batchSize, outputHeight, outputWidth, outputDepth] :
            [batchSize, outputDepth, outputHeight, outputWidth];
        const program = new DepthToSpaceProgram(outputShape, blockSize, dataFormat);
        return backend.runWebGLProgram(program, [x], x.dtype);
    }
    const depthToSpaceConfig = {
        kernelName: DepthToSpace,
        backendName: 'webgl',
        kernelFunc: depthToSpace
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class DepthwiseConv2DProgram {
        constructor(convInfo, addBias = false, activation = null, hasPreluActivation = false, hasLeakyReluAlpha = false) {
            this.variableNames = ['x', 'W'];
            this.customUniforms = [
                { name: 'pads', type: 'ivec2' },
                { name: 'strides', type: 'ivec2' },
                { name: 'dilations', type: 'ivec2' },
                { name: 'inDims', type: 'ivec2' },
            ];
            this.outputShape = convInfo.outShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            const filterHeight = convInfo.filterHeight;
            const filterWidth = convInfo.filterWidth;
            const channelMul = convInfo.outChannels / convInfo.inChannels;
            let activationSnippet = '', applyActivationSnippet = '';
            if (activation) {
                if (hasPreluActivation) {
                    activationSnippet = `float activation(float a) {
          float b = getPreluActivationWeightsAtOutCoords();
          ${activation}
        }`;
                }
                else if (hasLeakyReluAlpha) {
                    activationSnippet = `float activation(float a) {
          float b = getLeakyreluAlphaAtOutCoords();
          ${activation}
        }`;
                }
                else {
                    activationSnippet = `
          float activation(float x) {
            ${activation}
          }
        `;
                }
                applyActivationSnippet = `result = activation(result);`;
            }
            const addBiasSnippet = addBias ? 'result += getBiasAtOutCoords();' : '';
            if (addBias) {
                this.variableNames.push('bias');
            }
            if (hasPreluActivation) {
                this.variableNames.push('preluActivationWeights');
            }
            if (hasLeakyReluAlpha) {
                this.variableNames.push('leakyreluAlpha');
            }
            this.userCode = `
      ${activationSnippet}

      void main() {
        ivec4 coords = getOutputCoords();
        int batch = coords.x;
        ivec2 xRCCorner = coords.yz * strides - pads;
        int d2 = coords.w;
        int d1 = d2 / ${channelMul};
        int q = d2 - d1 * ${channelMul};

        int xRCorner = xRCCorner.x;
        int xCCorner = xRCCorner.y;

        // Convolve x(?, ?, d1) with w(:, :, d1, q) to get y(yR, yC, d2).
        // ? = to be determined. : = across all values in that axis.
        float dotProd = 0.0;
        // TO DO(dsmilkov): Flatten the two for loops and vec4 the operations.
        for (int wR = 0; wR < ${filterHeight}; wR++) {
          int xR = xRCorner + wR * dilations[0];

          if (xR < 0 || xR >= inDims[0]) {
            continue;
          }

          for (int wC = 0; wC < ${filterWidth}; wC++) {
            int xC = xCCorner + wC * dilations[1];

            if (xC < 0 || xC >= inDims[1]) {
              continue;
            }

            float xVal = getX(batch, xR, xC, d1);
            float wVal = getW(wR, wC, d1, q);
            dotProd += xVal * wVal;
          }
        }

        float result = dotProd;
        ${addBiasSnippet}
        ${applyActivationSnippet}
        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class DepthwiseConvPacked2DProgram {
        constructor(convInfo, addBias = false, activation = null, hasPreluActivation = false, hasLeakyReluAlpha = false) {
            this.variableNames = ['x', 'W'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.customUniforms = [
                { name: 'pads', type: 'ivec2' },
                { name: 'strides', type: 'ivec2' },
                { name: 'dilations', type: 'ivec2' },
                { name: 'inDims', type: 'ivec2' },
            ];
            this.outputShape = convInfo.outShape;
            this.enableShapeUniforms = useShapeUniforms(this.outputShape.length);
            const channelMul = convInfo.outChannels / convInfo.inChannels;
            const padLeft = convInfo.padInfo.left;
            const strideWidth = convInfo.strideWidth;
            const dilationWidth = convInfo.dilationWidth;
            const filterHeight = convInfo.filterHeight;
            const filterWidth = convInfo.filterWidth;
            const texelsAcross = filterWidth;
            let mainLoop = `
      int xR; int xC; int xCOffset;
      vec4 wTexel; vec4 previous; vec4 final;`;
            for (let c = 0; c < filterWidth; c++) {
                mainLoop += `
          vec4 xTexelC${c * 2};
          int xTexelC${c * 2}Ready;
          vec4 xTexelC${c * 2 + 1};
          int xTexelC${c * 2 + 1}Ready;
          vec4 xC${c};`;
            }
            /**
             * This vectorized implementation works by gathering the values needed for
             * each output channel's dot product into vec4's and then multiplying them
             * all together (this happens in the final double for-loop below). Most of
             * the main loop consists of constructing these vec4's with the minimum
             * number of texture2D calls, which means making use of all four returned
             * values from a texture2D call at once.
             */
            mainLoop += `
    for (int r = 0; r < ${filterHeight}; r++) {
      `;
            for (let c = 0; c < filterWidth; c++) {
                mainLoop += `
          xTexelC${c * 2} = vec4(0.0);
          xTexelC${c * 2}Ready = 0;
          xTexelC${c * 2 + 1} = vec4(0.0);
          xTexelC${c * 2 + 1}Ready = 0;
          xC${c} = vec4(0.0);`;
            }
            mainLoop += `
        xR = xRCorner + r * dilations[0];
        if (xR >=0 && xR < inDims[0]) {
      `;
            for (let texelC = 0; texelC < (texelsAcross + 1) / 2; texelC++) {
                const colIndex = texelC * 2;
                mainLoop += `
          xC = xCCorner + ${colIndex * dilationWidth};
          `;
                if (strideWidth === 1) {
                    if (colIndex < filterWidth) {
                        // If padding is odd, the outer texels have to be composed.
                        if (padLeft % 2 === 1) {
                            // TODO: Ensure vec4 previous does not result in redundant sample,
                            // and avoid setting xTexelRC's that exceed the boundary in the
                            // first place rather than resetting them to vec4(0)).
                            // To compute xCOffset:
                            // - If padding is odd, we must add 1 to ensure we ask for an
                            // even-numbered row.
                            // - We subtract 2 to access the previous texel.
                            mainLoop += `
                xCOffset = xC + 1;
                if (xCOffset >= 0 && xCOffset < inDims[1] && xTexelC${colIndex}Ready == 0) {
                  xTexelC${colIndex} = getX(batch, xR, xCOffset, d1);

                  // Need to manually clear unused channels in case
                  // we're reading from recycled texture.
                  if (xCOffset + 1 >= inDims[1]) {
                    xTexelC${colIndex}.zw = vec2(0.0);
                  }
                  xTexelC${colIndex}Ready = 1;
                }
              `;
                            // This texel has been read in previous iteration if the dilation
                            // is 1.
                            if (dilationWidth === 1 && colIndex > 0) {
                                mainLoop += `
                xC${colIndex} = vec4(xTexelC${colIndex - 2}.zw, xTexelC${colIndex}.xy);
                `;
                            }
                            else {
                                mainLoop += `
                  xCOffset = xC + 1 - 2;

                  if (xCOffset >= 0 && xCOffset < inDims[1]) {
                    previous = getX(batch, xR, xCOffset, d1);

                    // Need to manually clear unused channels in case
                    // we're reading from recycled texture.
                    if (xCOffset + 1 >= inDims[1]) {
                      previous.zw = vec2(0.0);
                    }

                    xC${colIndex} = vec4(previous.zw, xTexelC${colIndex}.xy);
                  } else {
                    xC${colIndex} = vec4(0.0, 0.0, xTexelC${colIndex}.xy);
                  }
                  `;
                            }
                        }
                        else {
                            // Padding is even, so xRC corresponds to a single texel.
                            mainLoop += `
                if (xC >= 0 && xC < inDims[1] && xTexelC${colIndex}Ready == 0) {
                  xTexelC${colIndex} = getX(batch, xR, xC, d1);
                  if (xC + 1 >= inDims[1]) {
                    xTexelC${colIndex}.zw = vec2(0.0);
                  }
                  xTexelC${colIndex}Ready = 1;
                }

                xC${colIndex} = xTexelC${colIndex};
                `;
                        }
                        if (colIndex + 1 < filterWidth) {
                            // If dilation is even, the second entry should match the first
                            // (either both are composed or both are single samples). But if
                            // dilation is odd, then the second entry should be the opposite
                            // of the first (if the first is composed, the second is a single
                            // sample, and vice versa.)
                            const nextTexelOffset = padLeft % 2 === 0 ?
                                nearestLargerEven(dilationWidth) :
                                dilationWidth;
                            if ((dilationWidth % 2 === 0 && padLeft % 2 === 1) ||
                                (dilationWidth % 2 !== 0 && padLeft % 2 !== 1)) {
                                mainLoop += `
                  xCOffset = xC + imod(pads[1], 2) + ${nextTexelOffset};

                  if (xCOffset >= 0 && xCOffset < inDims[1] && xTexelC${colIndex + 1}Ready == 0) {
                    xTexelC${colIndex + 1} = getX(batch, xR, xCOffset, d1);

                    // Need to manually clear unused channels in case
                    // we're reading from recycled texture.
                    if (xCOffset + 1 >= inDims[1]) {
                      xTexelC${colIndex + 1}.zw = vec2(0.0);
                    }
                    xTexelC${colIndex + 1}Ready = 1;
                  }
                  `;
                                // If dilation > 1 then the xRC's will not be able to share any
                                // values, so each xRC will require two unique calls to getX.
                                if (dilationWidth > 1) {
                                    mainLoop += `
                    xCOffset -= 2;
                    if (xCOffset >= 0 && xCOffset < inDims[1] && xTexelC${colIndex}Ready == 0) {
                      xTexelC${colIndex} = getX(batch, xR, xCOffset, d1);
                      xTexelC${colIndex}Ready = 1;
                    }
                    `;
                                }
                                mainLoop += `
                  xC${colIndex + 1} = vec4(xTexelC${colIndex}.zw, xTexelC${colIndex + 1}.xy);
                  `;
                            }
                            else {
                                // If dilation is 1 and padding is odd, we have already read the
                                // texel when constructing the previous x value. Here we can
                                // simply skip the texture read.
                                if (nextTexelOffset === 1) {
                                    mainLoop += `
                    xC${colIndex + 1} = xTexelC${colIndex};
                    `;
                                }
                                else {
                                    mainLoop += `
                    xCOffset = xC + ${nextTexelOffset};

                    if (xCOffset >= 0 && xCOffset < inDims[1] && xTexelC${colIndex + 1}Ready == 0) {
                      xTexelC${colIndex + 1} = getX(batch, xR, xCOffset, d1);
                      if (xCOffset + 1 >= inDims[1]) {
                        xTexelC${colIndex + 1}.zw = vec2(0.0);
                      }
                      xTexelC${colIndex + 1}Ready = 1;
                    }

                    xC${colIndex + 1} = xTexelC${colIndex + 1};
                    `;
                                }
                            }
                        }
                    }
                }
                else { // stride === 2
                    if (colIndex < filterWidth) {
                        // Depending on whether padLeft is even or odd, we want either the
                        // xy or zw channels from X texels for xC${colIndex}. If padLeft is
                        // even, xC${colIndex +1} is simply the zw channels of texels we've
                        // already sampled. But if padLeft is odd, xC{$c + 1}.zw will
                        // need to come from the xy channels of a new texel, hence the `
                        // vec4
                        // final` initialized below.
                        if (padLeft % 2 === 1) {
                            mainLoop += `
                xCOffset = xC + 1 - strides[1];
                if(xCOffset >= 0 && xCOffset < inDims[1] && xTexelC${colIndex}Ready == 0) {
                  xTexelC${colIndex} = getX(batch, xR, xCOffset, d1);
                  // Need to manually clear unused channels in case
                  // we're reading from recycled texture.
                  if (xCOffset + 1 >= inDims[1]) {
                    xTexelC${colIndex}.zw = vec2(0.0);
                  }
                  xTexelC${colIndex}Ready = 1;
                }

                if(xC + 1 >= 0 && xC + 1 < inDims[1] && xTexelC${colIndex + 1}Ready == 0) {
                  xTexelC${colIndex + 1} = getX(batch, xR, xC + 1, d1);
                  // Need to manually clear unused channels in case
                  // we're reading from recycled texture.
                  if (xC + 2 >= inDims[1]) {
                    xTexelC${colIndex + 1}.zw = vec2(0.0);
                  }
                  xTexelC${colIndex + 1}Ready = 1;
                }

                xC${colIndex} = vec4(xTexelC${colIndex}.zw, xTexelC${colIndex + 1}.zw);
              `;
                            if (colIndex + 1 < filterWidth) {
                                mainLoop += `
                  final = vec4(0.0);
                  xCOffset = xC + 1 + strides[1];
                  if(xCOffset >= 0 && xCOffset < inDims[1]) {
                    final = getX(batch, xR, xCOffset, d1);
                  }
                  xC${colIndex + 1} = vec4(xTexelC${colIndex + 1}.xy, final.xy);
                `;
                            }
                        }
                        else {
                            mainLoop += `
                if(xC >= 0 && xC < inDims[1] && xTexelC${colIndex}Ready == 0) {
                  xTexelC${colIndex} = getX(batch, xR, xC, d1);
                  if (xC + 1 >= inDims[1]) {
                    xTexelC${colIndex}.zw = vec2(0.0);
                  }
                  xTexelC${colIndex}Ready = 1;
                }

                xCOffset = xC + strides[1];
                if(xCOffset >= 0 && xCOffset < inDims[1] && xTexelC${colIndex + 1}Ready == 0) {
                  xTexelC${colIndex + 1} = getX(batch, xR, xCOffset, d1);
                  if (xCOffset + 1 >= inDims[1]) {
                    xTexelC${colIndex + 1}.zw = vec2(0.);
                  }
                  xTexelC${colIndex + 1}Ready = 1;
                }

                xC${colIndex} = vec4(
                  xTexelC${colIndex}.xy, xTexelC${colIndex + 1}.xy);
              `;
                            if (colIndex + 1 < filterWidth) {
                                mainLoop += `
                  xC${colIndex + 1} = vec4(xTexelC${colIndex}.zw, xTexelC${colIndex + 1}.zw);
                `;
                            }
                        }
                    }
                }
                // localize the dotProd accumulation within the loop, the theory is for
                // GPU with limited cache, accumulate sum across large amount of
                // veriables will cause lots of cache misses. (i.e. 5x5 filter will have
                // 50 variables)
                if (colIndex < filterWidth) {
                    mainLoop += `
            wTexel = getW(r, ${colIndex}, d1, q);
            dotProd += xC${colIndex} * vec4(wTexel.xz, wTexel.xz);
          `;
                    if (colIndex + 1 < filterWidth) {
                        mainLoop += `
              wTexel = getW(r, ${colIndex + 1}, d1, q);
              dotProd += xC${colIndex + 1} * vec4(wTexel.xz, wTexel.xz);
            `;
                    }
                }
            }
            mainLoop += `
    }
  `;
            mainLoop += `
      }
    `;
            let activationSnippet = '', applyActivationSnippet = '';
            if (activation) {
                if (hasPreluActivation) {
                    activationSnippet = `vec4 activation(vec4 a) {
          vec4 b = getPreluActivationWeightsAtOutCoords();
          ${activation}
        }`;
                }
                else if (hasLeakyReluAlpha) {
                    activationSnippet = `vec4 activation(vec4 a) {
          vec4 b = getLeakyreluAlphaAtOutCoords();
          ${activation}
        }`;
                }
                else {
                    activationSnippet = `vec4 activation(vec4 x) {
          ${activation}
        }`;
                }
                applyActivationSnippet = `result = activation(result);`;
            }
            const addBiasSnippet = addBias ? 'result += getBiasAtOutCoords();' : '';
            if (addBias) {
                this.variableNames.push('bias');
            }
            if (hasPreluActivation) {
                this.variableNames.push('preluActivationWeights');
            }
            if (hasLeakyReluAlpha) {
                this.variableNames.push('leakyreluAlpha');
            }
            this.userCode = `
      ${activationSnippet}

      void main() {
        ivec4 coords = getOutputCoords();
        int batch = coords.x;
        ivec2 xRCCorner = coords.yz * strides - pads;
        int d2 = coords.w;
        int d1 = d2 / ${channelMul};
        int q = d2 - d1 * ${channelMul};
        int xRCorner = xRCCorner.x;
        int xCCorner = xRCCorner.y;

        //intialize dotProd with a small epsilon seems to reduce GPU accuracy loss.
        vec4 dotProd = vec4(0.000000000000001);

        ${mainLoop}

        vec4 result = dotProd - vec4(0.000000000000001);
        ${addBiasSnippet}
        ${applyActivationSnippet}
        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthwiseConv2dNative(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter } = inputs;
        const { strides, pad, dilations, dimRoundingMode } = attrs;
        let $dilations = dilations;
        if ($dilations == null) {
            $dilations = [1, 1];
        }
        assert$4(eitherStridesOrDilationsAreOne(strides, $dilations), () => 'Error in depthwiseConv2d: Either strides or dilations must be ' +
            `1. Got strides ${strides} and dilations '${$dilations}'`);
        const convInfo = computeConv2DInfo(x.shape, filter.shape, strides, $dilations, pad, dimRoundingMode, true /* depthwise */);
        let program;
        if (env$1().getBool('WEBGL_PACK_DEPTHWISECONV') && convInfo.strideWidth <= 2 &&
            convInfo.outChannels / convInfo.inChannels === 1) {
            program = new DepthwiseConvPacked2DProgram(convInfo);
        }
        else {
            program = new DepthwiseConv2DProgram(convInfo);
        }
        const customValues = [
            [convInfo.padInfo.top, convInfo.padInfo.left],
            [convInfo.strideHeight, convInfo.strideWidth],
            [convInfo.dilationHeight, convInfo.dilationWidth],
            [convInfo.inHeight, convInfo.inWidth]
        ];
        return backend.runWebGLProgram(program, [x, filter], 'float32', customValues);
    }
    const depthwiseConv2dNativeConfig = {
        kernelName: DepthwiseConv2dNative,
        backendName: 'webgl',
        kernelFunc: depthwiseConv2dNative,
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class DepthwiseConv2DDerFilterProgram {
        constructor(convInfo) {
            this.variableNames = ['x', 'dy'];
            this.outputShape = convInfo.filterShape;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const padTop = convInfo.padInfo.top;
            const padLeft = convInfo.padInfo.left;
            const channelMul = convInfo.outChannels / convInfo.inChannels;
            this.userCode = `
      void main() {
        ivec4 coords = getOutputCoords();
        int wR = coords.x;
        int wC = coords.y;
        int d1 = coords.z;
        int dm = coords.w;
        int d2 = d1 * ${channelMul} + dm;

        float dotProd = 0.0;

        // TO DO: Vec4 over the batch size
        for (int b = 0; b < ${convInfo.batchSize}; b++) {
          for (int yR = 0; yR < ${convInfo.outHeight}; yR++) {
            int xR = wR + yR * ${strideHeight} - ${padTop};

            if (xR < 0 || xR >= ${convInfo.inHeight}) {
              continue;
            }

            for (int yC = 0; yC < ${convInfo.outWidth}; yC++) {
              int xC = wC + yC * ${strideWidth} - ${padLeft};

              if (xC < 0 || xC >= ${convInfo.inWidth}) {
                continue;
              }

              float dyValue = getDy(b, yR, yC, d2);
              float xValue = getX(b, xR, xC, d1);
              dotProd += (xValue * dyValue);
            }
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }
    class DepthwiseConv2DDerInputProgram {
        constructor(convInfo) {
            this.variableNames = ['dy', 'W'];
            this.outputShape = convInfo.inShape;
            const filterHeight = convInfo.filterHeight;
            const filterWidth = convInfo.filterWidth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const padTop = filterHeight - 1 - convInfo.padInfo.top;
            const padLeft = filterWidth - 1 - convInfo.padInfo.left;
            const channelMul = convInfo.outChannels / convInfo.inChannels;
            this.userCode = `
      const ivec2 pads = ivec2(${padTop}, ${padLeft});

      void main() {
        ivec4 coords = getOutputCoords();
        int batch = coords[0];
        int d1 = coords[3];
        ivec2 dyCorner = coords.yz - pads;
        int dyRCorner = dyCorner.x;
        int dyCCorner = dyCorner.y;

        float dotProd = 0.0;

        for (int wR = 0; wR < ${filterHeight}; wR++) {
          float dyR = float(dyRCorner + wR) / ${strideHeight}.0;

          if (dyR < 0.0 || dyR >= ${convInfo.outHeight}.0 || fract(dyR) > 0.0) {
            continue;
          }
          int idyR = int(dyR);

          int wRPerm = ${filterHeight} - 1 - wR;

          for (int wC = 0; wC < ${filterWidth}; wC++) {
            float dyC = float(dyCCorner + wC) / ${strideWidth}.0;

            if (dyC < 0.0 || dyC >= ${convInfo.outWidth}.0 ||
                fract(dyC) > 0.0) {
              continue;
            }
            int idyC = int(dyC);

            int wCPerm = ${filterWidth} - 1 - wC;

            // TO DO: Vec4 over the channelMul
            for (int dm = 0; dm < ${channelMul}; dm++) {
              int d2 = d1 * ${channelMul} + dm;
              float xValue = getDy(batch, idyR, idyC, d2);
              float wValue = getW(wRPerm, wCPerm, d1, dm);
              dotProd += xValue * wValue;
            }
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthwiseConv2dNativeBackpropFilter(args) {
        const { inputs, backend, attrs } = args;
        const { x, dy } = inputs;
        const { strides, dilations, pad, dimRoundingMode, filterShape } = attrs;
        const convInfo = computeConv2DInfo(x.shape, filterShape, strides, dilations, pad, dimRoundingMode, true /* depthwise */);
        const program = new DepthwiseConv2DDerFilterProgram(convInfo);
        return backend.runWebGLProgram(program, [x, dy], 'float32');
    }
    const depthwiseConv2dNativeBackpropFilterConfig = {
        kernelName: DepthwiseConv2dNativeBackpropFilter,
        backendName: 'webgl',
        kernelFunc: depthwiseConv2dNativeBackpropFilter
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function depthwiseConv2dNativeBackpropInput(args) {
        const { inputs, backend, attrs } = args;
        const { dy, filter } = inputs;
        const { strides, dilations, pad, dimRoundingMode, inputShape } = attrs;
        const convInfo = computeConv2DInfo(inputShape, filter.shape, strides, dilations, pad, dimRoundingMode, true /* depthwise */);
        const program = new DepthwiseConv2DDerInputProgram(convInfo);
        return backend.runWebGLProgram(program, [dy, filter], 'float32');
    }
    const depthwiseConv2dNativeBackpropInputConfig = {
        kernelName: DepthwiseConv2dNativeBackpropInput,
        backendName: 'webgl',
        kernelFunc: depthwiseConv2dNativeBackpropInput
    };

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class DiagProgram {
        constructor(size) {
            this.variableNames = ['X'];
            this.outputShape = [size, size];
            this.userCode = `
      void main() {
          ivec2 coords = getOutputCoords();
          float val = coords[0] == coords[1] ? getX(coords[0]) : 0.0;
          setOutput(val);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function diag(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        const outShape = [...x.shape, ...x.shape];
        const xSize = sizeFromShape(x.shape);
        const flat = reshape({ inputs: { x }, backend, attrs: { shape: [xSize] } });
        const program = new DiagProgram(xSize);
        const res = backend.runWebGLProgram(program, [flat], flat.dtype);
        const out = reshape({ inputs: { x: res }, backend, attrs: { shape: outShape } });
        backend.disposeIntermediateTensorInfo(flat);
        backend.disposeIntermediateTensorInfo(res);
        return out;
    }
    const diagConfig = {
        kernelName: Diag,
        backendName: 'webgl',
        kernelFunc: diag
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class Dilation2DProgram {
        constructor(convInfo) {
            this.variableNames = ['x', 'W'];
            this.outputShape = convInfo.outShape;
            const { inHeight, inWidth, padInfo, strideHeight, strideWidth, filterHeight, filterWidth, dilationHeight, dilationWidth } = convInfo;
            const { top: padTop, left: padLeft } = padInfo;
            this.userCode = `
      const ivec2 strides = ivec2(${strideHeight}, ${strideWidth});
      const ivec2 pads = ivec2(${padTop}, ${padLeft});
      const float neg_infinity = -3.4e38;

      void main() {
        ivec4 coords = getOutputCoords();
        int batch = coords.x;
        int d1 = coords.w;
        ivec2 outTopLeftCorner =
            coords.yz * strides - pads;
        int hBeg = outTopLeftCorner.x;
        int wBeg = outTopLeftCorner.y;

        float curVal = neg_infinity;
        for (int h = 0; h < ${filterHeight}; h++) {
          int hIn = hBeg + h * ${dilationHeight};

          if (hIn >= 0 && hIn < ${inHeight}) {
            for (int w = 0; w < ${filterWidth}; w++) {
              int wIn = wBeg + w * ${dilationWidth};

              if (wIn >= 0 && wIn < ${inWidth}) {
                float xVal = getX(batch, hIn, wIn, d1);
                float wVal = getW(h, w, d1);

                float val = xVal + wVal;
                if (val > curVal) {
                  curVal = val;
                }
              }
            }
          }
        }

        float result = curVal;
        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function dilation2D(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter } = inputs;
        const { strides, pad, dilations } = attrs;
        const convInfo = computeDilation2DInfo(x.shape, filter.shape, strides, pad, 'NHWC' /* dataFormat */, dilations);
        let out;
        const program = new Dilation2DProgram(convInfo);
        out = backend.runWebGLProgram(program, [x, filter], 'float32');
        const outReshaped = reshape({ inputs: { x: out }, backend, attrs: { shape: convInfo.outShape } });
        backend.disposeIntermediateTensorInfo(out);
        return outReshaped;
    }
    const dilation2DConfig = {
        kernelName: Dilation2D,
        backendName: 'webgl',
        kernelFunc: dilation2D,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function einsum(args) {
        const { inputs, backend, attrs } = args;
        const { equation } = attrs;
        const tensors = inputs;
        const { allDims, summedDims, idDims } = decodeEinsumEquation(equation, tensors.length);
        checkEinsumDimSizes(allDims.length, idDims, tensors);
        const { path, steps } = getEinsumComputePath(summedDims, idDims);
        const nSteps = steps.length;
        let out = null;
        let numDimsRemaining = allDims.length;
        const tensorsToDispose = [];
        for (let i = 0; i < nSteps; ++i) {
            for (const idTerm of steps[i]) {
                const { permutationIndices: perm, expandDims: dimsToExpand } = getEinsumPermutation(numDimsRemaining, idDims[idTerm]);
                let x;
                if (isIdentityPermutation(perm)) {
                    x = tensors[idTerm];
                }
                else {
                    x = transpose({ inputs: { x: tensors[idTerm] }, backend, attrs: { perm } });
                    tensorsToDispose.push(x);
                }
                const targetShape = x.shape.slice();
                for (let k = 0; k < dimsToExpand.length; ++k) {
                    targetShape.splice(dimsToExpand[k], 0, 1);
                }
                if (!arraysEqual(x.shape, targetShape)) {
                    x = reshape({ inputs: { x }, backend, attrs: { shape: targetShape } });
                    tensorsToDispose.push(x);
                }
                if (out === null) {
                    out = x;
                }
                else {
                    // tslint:disable-next-line: no-unnecessary-type-assertion
                    out = multiply({ inputs: { a: x, b: out }, backend });
                    tensorsToDispose.push(out);
                }
            }
            if (i < nSteps - 1) {
                if (path[i] >= 0) {
                    out = sum({
                        inputs: { x: out },
                        backend,
                        attrs: {
                            axis: path[i] - (allDims.length - numDimsRemaining),
                            keepDims: false
                        }
                    });
                    tensorsToDispose.push(out);
                }
                numDimsRemaining--;
            }
        }
        // Clean up intermediate tensors.
        for (const tensorInfo of tensorsToDispose) {
            if (tensorInfo === out) {
                continue;
            }
            backend.disposeIntermediateTensorInfo(tensorInfo);
        }
        return out;
    }
    const einsumConfig = {
        kernelName: Einsum,
        backendName: 'webgl',
        kernelFunc: einsum
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ELU = `return (x >= 0.0) ? x : (exp(x) - 1.0);`;
    const ELU_PACKED = `
  vec4 result;

  result.r = (x.r >= 0.0) ? x.r : (exp(x.r) - 1.0);
  result.g = (x.g >= 0.0) ? x.g : (exp(x.g) - 1.0);
  result.b = (x.b >= 0.0) ? x.b : (exp(x.b) - 1.0);
  result.a = (x.a >= 0.0) ? x.a : (exp(x.a) - 1.0);

  return result;
`;
    const elu = unaryKernelFunc({ opSnippet: ELU, packedOpSnippet: ELU_PACKED });
    const eluConfig = {
        kernelName: Elu$1,
        backendName: 'webgl',
        kernelFunc: elu
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ELU_DER = `return (b >= 1.0) ? a : a * (b + 1.0);`;
    const ELU_DER_PACKED = `
  vec4 bGTEZero = vec4(greaterThanEqual(b, vec4(0.)));
  return (bGTEZero * a) + ((vec4(1.0) - bGTEZero) * (a * (b + vec4(1.0))));
`;
    const eluGrad = (args) => {
        const { inputs, backend } = args;
        const { dy, y } = inputs;
        const program = env$1().getBool('WEBGL_PACK_BINARY_OPERATIONS') ?
            new BinaryOpPackedProgram(ELU_DER_PACKED, dy.shape, y.shape) :
            new BinaryOpProgram(ELU_DER, dy.shape, y.shape);
        return backend.runWebGLProgram(program, [dy, y], dy.dtype);
    };
    const eluGradConfig = {
        kernelName: EluGrad,
        backendName: 'webgl',
        kernelFunc: eluGrad
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const PACKED_EQUAL = `
  return vec4(equal(a, b));
`;
    const EQUAL = `return float(a == b);`;
    const equal$1 = binaryKernelFunc({
        opSnippet: EQUAL,
        packedOpSnippet: PACKED_EQUAL,
        dtype: 'bool',
        cpuKernelImpl: equalImplCPU,
    });
    const equalConfig = {
        kernelName: Equal,
        backendName: 'webgl',
        kernelFunc: equal$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ERF = `
  // Error function is calculated approximately with elementary function.
  // See "Handbook of Mathematical Functions with Formulas,
  // Graphs, and Mathematical Tables", Abramowitz and Stegun.
  float p = ${ERF_P};
  float a1 = ${ERF_A1};
  float a2 = ${ERF_A2};
  float a3 = ${ERF_A3};
  float a4 = ${ERF_A4};
  float a5 = ${ERF_A5};

  float sign = sign(x);
  x = abs(x);
  float t = 1.0 / (1.0 + p * x);
  return sign * (1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*exp(-x*x));
`;
    const erf = unaryKernelFunc({ opSnippet: ERF });
    const erfConfig = {
        kernelName: Erf,
        backendName: 'webgl',
        kernelFunc: erf,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const EXP = CHECK_NAN_SNIPPET_UNARY + `
  return exp(x);
`;
    const EXP_PACKED = `
  vec4 result = exp(x);
  bvec4 isNaN = isnan(x);
  result.r = isNaN.r ? x.r : result.r;
  result.g = isNaN.g ? x.g : result.g;
  result.b = isNaN.b ? x.b : result.b;
  result.a = isNaN.a ? x.a : result.a;

  return result;
`;
    const exp = unaryKernelFunc({
        opSnippet: EXP,
        packedOpSnippet: EXP_PACKED,
        cpuKernelImpl: expImplCPU,
        dtype: 'float32',
    });
    const expConfig = {
        kernelName: Exp,
        backendName: 'webgl',
        kernelFunc: exp
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function expandDims(args) {
        const { inputs, attrs, backend } = args;
        const { dim } = attrs;
        const { input } = inputs;
        const inputRank = input.shape.length;
        const newShape = input.shape.slice();
        let $dim = dim;
        if (dim < 0) {
            // Negative value is counted from the tail of rank.
            assert$4(-(inputRank + 1) <= dim, () => `Axis must be in the interval [${-(inputRank + 1)}, ${inputRank}]`);
            $dim = inputRank + dim + 1;
        }
        newShape.splice($dim, 0, 1);
        return reshape({ inputs: { x: input }, backend, attrs: { shape: newShape } });
    }
    const expandDimsConfig = {
        kernelName: ExpandDims,
        backendName: 'webgl',
        kernelFunc: expandDims,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const EXPM1 = `return exp(x) - 1.0;`;
    const expm1 = unaryKernelFunc({ opSnippet: EXPM1, packedOpSnippet: EXPM1, cpuKernelImpl: expm1ImplCPU });
    const expm1Config = {
        kernelName: Expm1,
        backendName: 'webgl',
        kernelFunc: expm1
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class FFTProgram {
        constructor(component, inputShape, inverse) {
            this.variableNames = ['real', 'imag'];
            const innerDim = inputShape[1];
            this.outputShape = inputShape;
            const exponentMultiplierSnippet = inverse ? `2.0 * ${Math.PI}` : `-2.0 * ${Math.PI}`;
            const resultDenominator = inverse ? `${innerDim}.0` : '1.0';
            let opString;
            if (component === 'real') {
                opString = 'return real * expR - imag * expI;';
            }
            else if (component === 'imag') {
                opString = 'return real * expI + imag * expR;';
            }
            else {
                throw new Error(`FFT component must be either "real" or "imag", got ${component}.`);
            }
            this.userCode = `
      const float exponentMultiplier = ${exponentMultiplierSnippet};

      float unaryOpComplex(float real, float expR, float imag, float expI) {
        ${opString}
      }

      float mulMatDFT(int batch, int index) {
        float indexRatio = float(index) / float(${innerDim});
        float exponentMultiplierTimesIndexRatio =
            exponentMultiplier * indexRatio;

        float result = 0.0;

        for (int i = 0; i < ${innerDim}; i++) {
          // x = (-2|2 * PI / N) * index * i;
          float x = exponentMultiplierTimesIndexRatio * float(i);
          float expR = cos(x);
          float expI = sin(x);
          float real = getReal(batch, i);
          float imag = getImag(batch, i);

          result +=
              unaryOpComplex(real, expR, imag, expI) / ${resultDenominator};
        }

        return result;
      }

      void main() {
        ivec2 coords = getOutputCoords();
        setOutput(mulMatDFT(coords[0], coords[1]));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fftImpl(x, inverse, backend) {
        const xData = backend.texData.get(x.dataId);
        const inputSize = sizeFromShape(x.shape);
        // Collapse all outer dimensions to a single batch dimension.
        const innerDimensionSize = x.shape[x.shape.length - 1];
        const batch = inputSize / innerDimensionSize;
        const input2D = reshape({ inputs: { x }, backend, attrs: { shape: [batch, innerDimensionSize] } });
        const xShape = input2D.shape;
        const realProgram = new FFTProgram('real', xShape, inverse);
        const imagProgram = new FFTProgram('imag', xShape, inverse);
        const inputs = [
            {
                dataId: xData.complexTensorInfos.real.dataId,
                dtype: xData.complexTensorInfos.real.dtype,
                shape: xShape
            },
            {
                dataId: xData.complexTensorInfos.imag.dataId,
                dtype: xData.complexTensorInfos.imag.dtype,
                shape: xShape
            }
        ];
        const realPart = backend.runWebGLProgram(realProgram, inputs, 'float32');
        const imagPart = backend.runWebGLProgram(imagProgram, inputs, 'float32');
        const complexOutput = complex({ inputs: { real: realPart, imag: imagPart }, backend });
        backend.disposeIntermediateTensorInfo(realPart);
        backend.disposeIntermediateTensorInfo(imagPart);
        const complexOutputReshaped = reshape({ inputs: { x: complexOutput }, backend, attrs: { shape: x.shape } });
        backend.disposeIntermediateTensorInfo(input2D);
        backend.disposeIntermediateTensorInfo(complexOutput);
        return complexOutputReshaped;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fft(args) {
        const { inputs, backend } = args;
        const { input } = inputs;
        return fftImpl(input, false /* inverse */, backend);
    }
    const fftConfig = {
        kernelName: FFT,
        backendName: 'webgl',
        kernelFunc: fft
    };

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class FillProgram {
        constructor(shape, value) {
            this.outputShape = [];
            this.customUniforms = [{ name: 'value', type: 'float' }];
            this.variableNames = ['x'];
            this.outputShape = shape;
            this.userCode = `
      void main() {
        // Input can be obtained from uniform value.
        setOutput(value);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fill(args) {
        const { backend, attrs } = args;
        const { shape, value } = attrs;
        let { dtype } = attrs;
        dtype = dtype || inferDtype(value);
        if (dtype === 'string') {
            // String type should be handled in CPU memory.
            const values = getArrayFromDType(dtype, sizeFromShape(shape));
            values.fill(value);
            return backend.makeTensorInfo(shape, dtype, values);
        }
        else {
            const program = new FillProgram(shape, value);
            const customValues = [[value]];
            return backend.runWebGLProgram(program, [], dtype, customValues);
        }
    }
    const fillConfig = {
        kernelName: Fill,
        backendName: 'webgl',
        kernelFunc: fill
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class FlipLeftRightProgram {
        constructor(imageShape) {
            this.variableNames = ['Image'];
            this.outputShape = [];
            const imageWidth = imageShape[2];
            this.outputShape = imageShape;
            this.userCode = `
        void main() {
          ivec4 coords = getOutputCoords();
          int x = coords[2];

          int coordX = ${imageWidth} - x - 1;
          float outputValue;
          if(coordX >= 0 && coordX < ${imageWidth}) {
            outputValue = getImage(coords[0], coords[1], coordX, coords[3]);
          } else {
            outputValue = getImage(coords[0], coords[1], coords[2], coords[3]);
          }
          setOutput(outputValue);
        }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const flipLeftRightConfig = {
        kernelName: FlipLeftRight,
        backendName: 'webgl',
        kernelFunc: ({ inputs, backend }) => {
            const { image } = inputs;
            const webglBackend = backend;
            const program = new FlipLeftRightProgram(image.shape);
            const output = webglBackend.runWebGLProgram(program, [image], image.dtype);
            return output;
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const FLOOR = `return floor(x);`;
    const floor$1 = unaryKernelFunc({ opSnippet: FLOOR, packedOpSnippet: FLOOR, cpuKernelImpl: floorImplCPU });
    const floorConfig = {
        kernelName: Floor,
        backendName: 'webgl',
        kernelFunc: floor$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // We use native integer division to deal with floating point imprecision. Since
    // we implement floor division and glsl implements truncated division, we
    // correct for this by subtracting 1 from result when the result is negative and
    // there is a remainder.
    const INT_DIV = `
  float s = sign(a) * sign(b);
  int ia = round(a);
  int ib = round(b);
  if (ib != 0) {
    // Windows (D3D) wants guaranteed non-zero int division at compile-time.
    return float(idiv(ia, ib, s));
  } else {
    return NAN;
  }
`;
    const INT_DIV_PACKED = `
  ivec4 ia = round(a);
  ivec4 ib = round(b);
  bvec4 cond = notEqual(ib, ivec4(0));
  ivec4 result = ivec4(0);
  vec4 s = sign(a) * sign(b);

  // Windows (D3D) wants guaranteed non-zero int division at compile-time.
  if (cond[0]) {
    result[0] = idiv(ia[0], ib[0], s[0]);
  }
  if (cond[1]) {
    result[1] = idiv(ia[1], ib[1], s[1]);
  }
  if (cond[2]) {
    result[2] = idiv(ia[2], ib[2], s[2]);
  }
  if (cond[3]) {
    result[3] = idiv(ia[3], ib[3], s[3]);
  }
  return vec4(result);
`;
    const floorDiv = binaryKernelFunc({ opSnippet: INT_DIV, packedOpSnippet: INT_DIV_PACKED, dtype: 'int32' });
    const floorDivConfig = {
        kernelName: FloorDiv,
        backendName: 'webgl',
        kernelFunc: floorDiv
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class FromPixelsProgram {
        constructor(outputShape) {
            this.variableNames = ['A'];
            const glsl = getGlslDifferences();
            const [height, width,] = outputShape;
            this.outputShape = outputShape;
            this.userCode = `
      void main() {
        ivec3 coords = getOutputCoords();
        int texR = coords[0];
        int texC = coords[1];
        int depth = coords[2];
        vec2 uv = (vec2(texC, texR) + halfCR) / vec2(${width}.0, ${height}.0);

        vec4 values = ${glsl.texture2D}(A, uv);
        float value;
        if (depth == 0) {
          value = values.r;
        } else if (depth == 1) {
          value = values.g;
        } else if (depth == 2) {
          value = values.b;
        } else if (depth == 3) {
          value = values.a;
        }

        setOutput(floor(value * 255.0 + 0.5));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class FromPixelsPackedProgram {
        constructor(outputShape) {
            this.variableNames = ['A'];
            this.packedInputs = false;
            this.packedOutput = true;
            const glsl = getGlslDifferences();
            const [height, width,] = outputShape;
            this.outputShape = outputShape;
            this.userCode = `
      void main() {
        ivec3 coords = getOutputCoords();
        int texR = coords[0];
        int texC = coords[1];
        int depth = coords[2];

        vec4 result = vec4(0.);

        for(int row=0; row<=1; row++) {
          for(int col=0; col<=1; col++) {
            texC = coords[1] + row;
            depth = coords[2] + col;

            vec2 uv = (vec2(texC, texR) + halfCR) /
                       vec2(${width}.0, ${height}.0);
            vec4 values = ${glsl.texture2D}(A, uv);
            float value;
            if (depth == 0) {
              value = values.r;
            } else if (depth == 1) {
              value = values.g;
            } else if (depth == 2) {
              value = values.b;
            } else if (depth == 3) {
              value = values.a;
            }

            result[row * 2 + col] = floor(value * 255.0 + 0.5);
          }
        }

        ${glsl.output} = result;
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const fromPixelsConfig = {
        kernelName: FromPixels,
        backendName: 'webgl',
        kernelFunc: fromPixels,
    };
    let fromPixels2DContext;
    function fromPixels(args) {
        const { inputs, backend, attrs } = args;
        let { pixels } = inputs;
        const { numChannels } = attrs;
        const isVideo = typeof (HTMLVideoElement) !== 'undefined' &&
            pixels instanceof HTMLVideoElement;
        const isImage = typeof (HTMLImageElement) !== 'undefined' &&
            pixels instanceof HTMLImageElement;
        const [width, height] = isVideo ?
            [
                pixels.videoWidth,
                pixels.videoHeight
            ] :
            [pixels.width, pixels.height];
        const texShape = [height, width];
        const outShape = [height, width, numChannels];
        if (isImage || isVideo) {
            if (fromPixels2DContext == null) {
                fromPixels2DContext = document.createElement('canvas').getContext('2d');
            }
            fromPixels2DContext.canvas.width = width;
            fromPixels2DContext.canvas.height = height;
            fromPixels2DContext.drawImage(pixels, 0, 0, width, height);
            pixels = fromPixels2DContext.canvas;
        }
        const tempPixelHandle = backend.makeTensorInfo(texShape, 'int32');
        // This is a byte texture with pixels.
        backend.texData.get(tempPixelHandle.dataId).usage = TextureUsage.PIXELS;
        backend.gpgpu.uploadPixelDataToTexture(backend.getTexture(tempPixelHandle.dataId), pixels);
        const program = env$1().getBool('WEBGL_PACK') ?
            new FromPixelsPackedProgram(outShape) :
            new FromPixelsProgram(outShape);
        const res = backend.runWebGLProgram(program, [tempPixelHandle], 'int32');
        backend.disposeData(tempPixelHandle.dataId);
        return res;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fusedConv2d(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter, bias, preluActivationWeights } = inputs;
        const { strides, pad, dataFormat, dilations, dimRoundingMode, activation, leakyreluAlpha } = attrs;
        const $dataFormat = convertConv2DDataFormat(dataFormat);
        const convInfo = computeConv2DInfo(x.shape, filter.shape, strides, dilations, pad, dimRoundingMode, false /* depthwise */, $dataFormat);
        let out;
        const intermediates = [];
        if (convInfo.filterHeight === 1 && convInfo.filterWidth === 1 &&
            convInfo.dilationHeight === 1 && convInfo.dilationWidth === 1 &&
            convInfo.strideHeight === 1 && convInfo.strideWidth === 1 &&
            (convInfo.padInfo.type === 'SAME' || convInfo.padInfo.type === 'VALID')) {
            out = conv2dByMatMul({
                x,
                filter,
                convInfo,
                backend,
                bias,
                activation,
                preluActivationWeights,
                leakyreluAlpha
            });
        }
        else if (env$1().getBool('WEBGL_CONV_IM2COL') && x.shape[0] === 1) {
            out = conv2dWithIm2Row({
                x,
                filter,
                convInfo,
                backend,
                bias,
                activation,
                preluActivationWeights,
                leakyreluAlpha
            });
        }
        else {
            const hasBias = bias != null;
            const hasPreluActivationWeights = preluActivationWeights != null;
            const hasLeakyreluAlpha = activation === 'leakyrelu';
            const fusedActivation = activation ? mapActivationToShaderProgram(activation, false) : null;
            const program = new Conv2DProgram(convInfo, hasBias, fusedActivation, hasPreluActivationWeights, hasLeakyreluAlpha);
            const inputs = [x, filter];
            if (bias) {
                inputs.push(bias);
            }
            if (preluActivationWeights) {
                inputs.push(preluActivationWeights);
            }
            if (hasLeakyreluAlpha) {
                const $leakyreluAlpha = backend.makeTensorInfo([], 'float32', createScalarValue(leakyreluAlpha, 'float32'));
                inputs.push($leakyreluAlpha);
                intermediates.push($leakyreluAlpha);
            }
            out = backend.runWebGLProgram(program, inputs, 'float32');
        }
        const outReshaped = reshape({ inputs: { x: out }, backend, attrs: { shape: convInfo.outShape } });
        intermediates.push(out);
        intermediates.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return outReshaped;
    }
    const fusedConv2DConfig = {
        kernelName: FusedConv2D,
        backendName: 'webgl',
        kernelFunc: fusedConv2d,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function fusedDepthwiseConv2D(args) {
        const { inputs, backend, attrs } = args;
        const { x, filter, bias, preluActivationWeights } = inputs;
        const { strides, pad, dilations, dimRoundingMode, activation, leakyreluAlpha } = attrs;
        const intermediates = [];
        let $dilations = dilations;
        if ($dilations == null) {
            $dilations = [1, 1];
        }
        assert$4(eitherStridesOrDilationsAreOne(strides, $dilations), () => 'Error in depthwiseConv2d: Either strides or dilations must be ' +
            `1. Got strides ${strides} and dilations '${$dilations}'`);
        const convInfo = computeConv2DInfo(x.shape, filter.shape, strides, $dilations, pad, dimRoundingMode, true /* depthwise */);
        const shouldPackDepthwiseConv = env$1().getBool('WEBGL_PACK_DEPTHWISECONV') &&
            convInfo.strideWidth <= 2 &&
            convInfo.outChannels / convInfo.inChannels === 1;
        const fusedActivation = activation ?
            mapActivationToShaderProgram(activation, shouldPackDepthwiseConv) :
            null;
        const programInputs = [x, filter];
        const hasBias = bias != null;
        const hasPreluActivationWeights = preluActivationWeights != null;
        const hasLeakyreluAlpha = activation === 'leakyrelu';
        if (hasBias) {
            programInputs.push(bias);
        }
        if (hasPreluActivationWeights) {
            programInputs.push(preluActivationWeights);
        }
        if (hasLeakyreluAlpha) {
            const $leakyreluAlpha = backend.makeTensorInfo([], 'float32', createScalarValue(leakyreluAlpha, 'float32'));
            programInputs.push($leakyreluAlpha);
            intermediates.push($leakyreluAlpha);
        }
        let program;
        if (shouldPackDepthwiseConv) {
            program = new DepthwiseConvPacked2DProgram(convInfo, hasBias, fusedActivation, hasPreluActivationWeights, hasLeakyreluAlpha);
        }
        else {
            program = new DepthwiseConv2DProgram(convInfo, hasBias, fusedActivation, hasPreluActivationWeights, hasLeakyreluAlpha);
        }
        const customValues = [
            [convInfo.padInfo.top, convInfo.padInfo.left],
            [convInfo.strideHeight, convInfo.strideWidth],
            [convInfo.dilationHeight, convInfo.dilationWidth],
            [convInfo.inHeight, convInfo.inWidth]
        ];
        const result = backend.runWebGLProgram(program, programInputs, 'float32', customValues);
        intermediates.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return result;
    }
    const fusedDepthwiseConv2DConfig = {
        kernelName: FusedDepthwiseConv2D,
        backendName: 'webgl',
        kernelFunc: fusedDepthwiseConv2D,
    };

    class GatherNDProgram {
        constructor(sliceDim, strides, shape) {
            this.sliceDim = sliceDim;
            this.strides = strides;
            this.variableNames = ['x', 'indices'];
            this.outputShape = shape;
            const stridesType = getCoordsDataType(strides.length);
            const dtype = getCoordsDataType(shape.length);
            const strideString = this.sliceDim > 1 ? 'strides[j]' : 'strides';
            this.userCode = `
        ${stridesType} strides = ${stridesType}(${this.strides});
         void main() {
          ${dtype} coords = getOutputCoords();
          int flattenIndex = 0;
          for (int j = 0; j < ${this.sliceDim}; j++) {
            int index = round(getIndices(coords[0], j));
            flattenIndex += index * ${strideString};
          }
          setOutput(getX(flattenIndex, coords[1]));
        }
      `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function gatherNd(args) {
        const { inputs, backend } = args;
        const { params, indices } = inputs;
        const indicesShape = indices.shape;
        const sliceRank = indicesShape[indicesShape.length - 1];
        const paramsSize = sizeFromShape(params.shape);
        const [resultShape, numSlices, sliceSize, strides] = prepareAndValidate(params, indices);
        const flattenIndices = reshape({ inputs: { x: indices }, backend, attrs: { shape: [numSlices, sliceRank] } });
        const flattenX = reshape({
            inputs: { x: params },
            backend,
            attrs: { shape: [(sizeFromShape(params.shape) / sliceSize), sliceSize] }
        });
        if (backend.shouldExecuteOnCPU([params, indices]) ||
            params.dtype === 'string') {
            const indicesData = backend.readSync(indices.dataId);
            const paramsBuf = backend.bufferSync(params);
            const outValue = gatherNdImplCPU(indicesData, paramsBuf, params.dtype, numSlices, sliceRank, sliceSize, strides, params.shape, paramsSize);
            return backend.makeTensorInfo(resultShape, params.dtype, outValue.values);
        }
        const program = new GatherNDProgram(sliceRank, strides, [numSlices, sliceSize]);
        const res = backend.runWebGLProgram(program, [flattenX, flattenIndices], flattenX.dtype);
        const reshaped = reshape({ inputs: { x: res }, backend, attrs: { shape: resultShape } });
        backend.disposeIntermediateTensorInfo(flattenIndices);
        backend.disposeIntermediateTensorInfo(flattenX);
        backend.disposeIntermediateTensorInfo(res);
        return reshaped;
    }
    const gatherNdConfig = {
        kernelName: GatherNd,
        backendName: 'webgl',
        kernelFunc: gatherNd
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class GatherProgram {
        constructor(aShape, outputShape) {
            this.variableNames = ['A', 'indices'];
            this.outputShape = outputShape;
            this.rank = outputShape.length;
            const dtype = getCoordsDataType(this.rank);
            const sourceCoords = getSourceCoords$1(aShape);
            this.userCode = `
      void main() {
        ${dtype} resRC = getOutputCoords();
        int index = int(getIndices(resRC.x, resRC.z));
        float inBounds = (index >= 0) && (index < ${aShape[2]}) ? 1.0 : 0.0;
        setOutput(inBounds * getA(${sourceCoords}));
      }
    `;
        }
    }
    // The input and output are always flattened into rank 4 tensors.
    function getSourceCoords$1(aShape, axis) {
        const currentCoords = ['resRC.x', 'resRC.y', 'resRC.z', 'resRC.w'];
        const sourceCoords = [];
        for (let i = 0; i < aShape.length; i++) {
            if (i === 2) {
                sourceCoords.push('index');
            }
            else {
                sourceCoords.push(`${currentCoords[i]}`);
            }
        }
        return sourceCoords.join();
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function gatherV2(args) {
        const { inputs, backend, attrs } = args;
        const { x, indices } = inputs;
        const { axis, batchDims } = attrs;
        const parsedAxis = parseAxisParam(axis, x.shape)[0];
        if (env$1().get('DEBUG')) {
            // In debug mode, throw error when any index is out of bound.
            // Otherwise, just fill out of bounds with zeroes.
            const indicesVals = backend.readSync(indices.dataId);
            const axisDim = x.shape[parsedAxis];
            for (let i = 0; i < indicesVals.length; ++i) {
                const index = indicesVals[i];
                assert$4(index <= axisDim - 1 && index >= 0, () => `GatherV2: the index value ${index} is not in [0, ${axisDim - 1}]`);
            }
        }
        const shapeInfo = collectGatherOpShapeInfo(x, indices, parsedAxis, batchDims);
        const indicesSize = sizeFromShape(indices.shape);
        const toDispose = [];
        const flattenX = reshape({
            inputs: { x },
            backend,
            attrs: {
                shape: [
                    shapeInfo.batchSize, shapeInfo.outerSize, shapeInfo.dimSize,
                    shapeInfo.sliceSize
                ]
            }
        });
        const flattenIndex = reshape({
            inputs: { x: indices },
            backend,
            attrs: { shape: [shapeInfo.batchSize, indicesSize / shapeInfo.batchSize] }
        });
        toDispose.push(flattenX);
        toDispose.push(flattenIndex);
        const flattenOutputShape = [
            shapeInfo.batchSize, shapeInfo.outerSize, indicesSize / shapeInfo.batchSize,
            shapeInfo.sliceSize
        ];
        if (backend.shouldExecuteOnCPU([x, indices]) || x.dtype === 'string') {
            const indicesBuf = backend.bufferSync(flattenIndex);
            const xBuf = backend.bufferSync(flattenX);
            const outBuf = gatherV2ImplCPU(xBuf, indicesBuf, flattenOutputShape);
            toDispose.forEach(t => backend.disposeIntermediateTensorInfo(t));
            return backend.makeTensorInfo(shapeInfo.outputShape, outBuf.dtype, outBuf.values);
        }
        const program = new GatherProgram(flattenX.shape, flattenOutputShape);
        const res = backend.runWebGLProgram(program, [flattenX, flattenIndex], flattenX.dtype);
        toDispose.push(res);
        const reshaped = reshape({ inputs: { x: res }, backend, attrs: { shape: shapeInfo.outputShape } });
        toDispose.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return reshaped;
    }
    const gatherV2Config = {
        kernelName: GatherV2,
        backendName: 'webgl',
        kernelFunc: gatherV2
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const GREATER = `return float(a > b);`;
    const GREATER_PACKED = `
  return vec4(greaterThan(a, b));
`;
    const greater = binaryKernelFunc({
        opSnippet: GREATER,
        packedOpSnippet: GREATER_PACKED,
        cpuKernelImpl: greaterImplCPU,
        dtype: 'bool'
    });
    const greaterConfig = {
        kernelName: Greater,
        backendName: 'webgl',
        kernelFunc: greater
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const GREATER_EQUAL = `return float(a >= b);`;
    const GREATER_EQUAL_PACKED = `
  return vec4(greaterThanEqual(a, b));
`;
    const greaterEqual = binaryKernelFunc({
        opSnippet: GREATER_EQUAL,
        packedOpSnippet: GREATER_EQUAL_PACKED,
        dtype: 'bool',
        cpuKernelImpl: greaterEqualImplCPU
    });
    const greaterEqualConfig = {
        kernelName: GreaterEqual,
        backendName: 'webgl',
        kernelFunc: greaterEqual
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function ifft(args) {
        const { inputs, backend } = args;
        const { input } = inputs;
        return fftImpl(input, true /* inverse */, backend);
    }
    const ifftConfig = {
        kernelName: IFFT,
        backendName: 'webgl',
        kernelFunc: ifft
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const IS_FINITE = `return float(!isnan(x) && !isinf(x));`;
    const isFinite$1 = unaryKernelFunc({ opSnippet: IS_FINITE, dtype: 'bool' });
    const isFiniteConfig = {
        kernelName: IsFinite,
        backendName: 'webgl',
        kernelFunc: isFinite$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const IS_INF = `return float(isinf(x));`;
    const isInf = unaryKernelFunc({ opSnippet: IS_INF, dtype: 'bool' });
    const isInfConfig = {
        kernelName: IsInf,
        backendName: 'webgl',
        kernelFunc: isInf,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const IS_NAN = `return float(isnan(x));`;
    const isNaN$1 = unaryKernelFunc({ opSnippet: IS_NAN, dtype: 'bool' });
    const isNaNConfig = {
        kernelName: IsNan,
        backendName: 'webgl',
        kernelFunc: isNaN$1,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const LESS = `return float(a < b);`;
    const LESS_PACKED = `
  return vec4(lessThan(a, b));
`;
    const less = binaryKernelFunc({
        opSnippet: LESS,
        packedOpSnippet: LESS_PACKED,
        cpuKernelImpl: lessImplCPU,
        dtype: 'bool'
    });
    const lessConfig = {
        kernelName: Less,
        backendName: 'webgl',
        kernelFunc: less
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const LESS_EQUAL = `return float(a <= b);`;
    const LESS_EQUAL_PACKED = `
  return vec4(lessThanEqual(a, b));
`;
    const lessEqual = binaryKernelFunc({
        opSnippet: LESS_EQUAL,
        packedOpSnippet: LESS_EQUAL_PACKED,
        cpuKernelImpl: lessEqualImplCPU,
        dtype: 'bool'
    });
    const lessEqualConfig = {
        kernelName: LessEqual,
        backendName: 'webgl',
        kernelFunc: lessEqual
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function linSpace(args) {
        const { backend, attrs } = args;
        const { start, stop, num } = attrs;
        // TODO: Use CPU implementation due to the precision problem in Safari.
        const outVals = linSpaceImplCPU(start, stop, num);
        return backend.makeTensorInfo([outVals.length], 'float32', outVals);
    }
    const linSpaceConfig = {
        kernelName: LinSpace,
        backendName: 'webgl',
        kernelFunc: linSpace
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Windows chrome return 0 if the input is negative value. We will specifically
    // return NaN if the input is 0 to solve compatiblity issue.
    const LOG = CHECK_NAN_SNIPPET_UNARY + `
  return x < 0.0 ? 0./0. : log(x);
`;
    const LOG_PACKED = `
  vec4 result = log(x);
  bvec4 isNaN = isnan(x);
  result.r = isNaN.r ? x.r : (x.r < 0.0 ? 0./0. : result.r);
  result.g = isNaN.g ? x.g : (x.g < 0.0 ? 0./0. : result.g);
  result.b = isNaN.b ? x.b : (x.b < 0.0 ? 0./0. : result.b);
  result.a = isNaN.a ? x.a : (x.a < 0.0 ? 0./0. : result.a);
  return result;
`;
    const log$1 = unaryKernelFunc({ opSnippet: LOG, packedOpSnippet: LOG_PACKED, cpuKernelImpl: logImplCPU });
    const logConfig = {
        kernelName: Log,
        backendName: 'webgl',
        kernelFunc: log$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const LOG1P = CHECK_NAN_SNIPPET_UNARY + `
  return log(1.0 + x);
`;
    const log1p = unaryKernelFunc({ opSnippet: LOG1P });
    const log1pConfig = {
        kernelName: Log1p,
        backendName: 'webgl',
        kernelFunc: log1p,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const LOGICAL_AND = `return float(a >= 1.0 && b >= 1.0);`;
    const LOGICAL_AND_PACKED = `
  return vec4(
    vec4(greaterThanEqual(a, vec4(1.0))) *
    vec4(greaterThanEqual(b, vec4(1.0))));
`;
    const logicalAnd = binaryKernelFunc({
        opSnippet: LOGICAL_AND,
        packedOpSnippet: LOGICAL_AND_PACKED,
        dtype: 'bool'
    });
    const logicalAndConfig = {
        kernelName: LogicalAnd,
        backendName: 'webgl',
        kernelFunc: logicalAnd
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const LOGICAL_NOT = `return float(!(x >= 1.0));`;
    const logicalNot = unaryKernelFunc({ opSnippet: LOGICAL_NOT });
    const logicalNotConfig = {
        kernelName: LogicalNot,
        backendName: 'webgl',
        kernelFunc: logicalNot,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const LOGICAL_OR = `return float(a >= 1.0 || b >= 1.0);`;
    const LOGICAL_OR_PACKED = `
  return min(
    vec4(greaterThanEqual(a, vec4(1.0))) +
    vec4(greaterThanEqual(b, vec4(1.0))),
    vec4(1.0));
`;
    const logicalOr = binaryKernelFunc({ opSnippet: LOGICAL_OR, packedOpSnippet: LOGICAL_OR_PACKED, dtype: 'bool' });
    const logicalOrConfig = {
        kernelName: LogicalOr,
        backendName: 'webgl',
        kernelFunc: logicalOr
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class LRNProgram {
        constructor(xShape, radius, bias, alpha, beta) {
            this.variableNames = ['x'];
            this.outputShape = [];
            const rad = radius;
            const maxD = xShape[3] - 1;
            this.outputShape = xShape;
            // optimize pow(bias + alpha * sum, -beta)
            // src: https://github.com/tensorflow/tensorflow/..
            // blob/26033a1644a9c4a5fbe3170ab2e864b6a4ccd4ca/..
            // tensorflow/core/kernels/mkl_lrn_op.cc#L320
            let powOperator;
            const basis = `float(${bias}) + float(${alpha}) * sum`;
            if (beta === 0.5) {
                powOperator = `inversesqrt(${basis})`;
            }
            else if (beta === 1.0) {
                powOperator = `1.0/(${basis})`;
            }
            else {
                powOperator = `exp(log(${basis}) * float(-${beta}));`;
            }
            this.userCode = `
      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int r = coords[1];
        int c = coords[2];
        int d = coords[3];
        float x = getX(b, r, c, d);
        float sum = 0.0;
        for (int j = -${rad}; j <= ${rad}; j++) {
          int idx = d + j;
          if (idx >= 0 && idx <=  ${maxD}) {
            float z = getX(b, r, c, idx);
            sum += z * z;
          }
        }
        float val = x * ${powOperator};
        setOutput(val);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class LRNPackedProgram {
        constructor(xShape, radius, bias, alpha, beta) {
            this.variableNames = ['x'];
            this.outputShape = [];
            this.packedInputs = true;
            this.packedOutput = true;
            const rad = radius;
            const maxD = xShape[3] - 1;
            this.outputShape = xShape;
            // optimize pow(bias + alpha * sum, -beta)
            // src: https://github.com/tensorflow/tensorflow/..
            // blob/26033a1644a9c4a5fbe3170ab2e864b6a4ccd4ca/..
            // tensorflow/core/kernels/mkl_lrn_op.cc#L320
            let powOperator;
            const basis = `float(${bias}) + float(${alpha}) * sum`;
            if (beta === 0.5) {
                powOperator = `inversesqrt(${basis})`;
            }
            else if (beta === 1.0) {
                powOperator = `1.0/(${basis})`;
            }
            else {
                powOperator = `exp(log(${basis}) * float(-${beta}));`;
            }
            this.userCode = `
      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords.x;
        int r = coords.y;
        int c = coords.z;
        int d = coords.w;

        bool hasNextCol = d < ${this.outputShape[3]};
        bool hasNextRow = c < ${this.outputShape[2]};

        vec4 sum = vec4(0.);
        vec4 xFragAtOutputCoords = getX(b, r, c, d);

        vec4 xAtOutputCoords = vec4(
          getChannel(xFragAtOutputCoords, vec2(c, d)),
          hasNextCol ?
            getChannel(xFragAtOutputCoords, vec2(c, d + 1)) : 0.0,
          hasNextRow ?
            getChannel(xFragAtOutputCoords , vec2(c + 1, d)) : 0.0,
          (hasNextRow && hasNextCol) ?
            getChannel(xFragAtOutputCoords, vec2(c + 1, d + 1)) : 0.0
        );

        int firstChannel = d - ${rad};
        vec2 cache = vec2(0.);
        if(firstChannel >= 0){
          vec4 firstChannelFrag = getX(b, r, c, firstChannel);
          cache.x = getChannel(firstChannelFrag, vec2(c, firstChannel));
            if(hasNextRow){
              cache.y = getChannel(firstChannelFrag, vec2(c + 1, firstChannel));
            }
        }

        ivec2 depth = ivec2(d, d + 1);
        for (int j = - ${rad}; j <= ${rad}; j++) {
          ivec2 idx = depth + j;
          bvec2 aboveLowerBound = greaterThanEqual(idx, ivec2(0));
          bvec2 belowUpperBound = lessThanEqual(idx, ivec2(${maxD}));

          bool depthInRange = aboveLowerBound.x && belowUpperBound.x;
          bool depthPlusOneInRange = aboveLowerBound.y && belowUpperBound.y;

          if(depthInRange || depthPlusOneInRange){
            vec4 z = vec4(0.);
            vec4 xFragAtCurrentDepth;
            z.xz = cache.xy;
            if(depthPlusOneInRange && hasNextCol){
              xFragAtCurrentDepth = idx.y != d ?
                getX(b, r, c, idx.y) : xFragAtOutputCoords;
              z.y = getChannel(xFragAtCurrentDepth, vec2(c, idx.y));
              if(hasNextRow){
                z.w = getChannel(xFragAtCurrentDepth, vec2(c + 1, idx.y));
              }
            }
            cache.xy = z.yw;
            sum += z * z;
          }
        }
        vec4 result = xAtOutputCoords * ${powOperator};
        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const lrn = (args) => {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { depthRadius, bias, alpha, beta } = attrs;
        const program = env$1().getBool('WEBGL_PACK_NORMALIZATION') ?
            new LRNPackedProgram(x.shape, depthRadius, bias, alpha, beta) :
            new LRNProgram(x.shape, depthRadius, bias, alpha, beta);
        return backend.runWebGLProgram(program, [x], x.dtype);
    };
    // tslint:disable-next-line: variable-name
    const LRNConfig = {
        kernelName: LRN,
        backendName: 'webgl',
        kernelFunc: lrn
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class LRNGradProgram {
        constructor(inputShape, depthRadius, bias, alpha, beta) {
            this.variableNames = ['inputImage', 'outputImage', 'dy'];
            this.outputShape = [];
            this.outputShape = inputShape;
            this.depth = inputShape[3];
            this.depthRadius = depthRadius;
            this.bias = bias;
            this.alpha = alpha;
            this.beta = beta;
            this.userCode = `
      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int r = coords[1];
        int c = coords[2];

        float result = 0.0;
        for (int d = 0; d < ${this.depth}; ++d) {
          int depthBegin = int(max(0.0, float(d - ${depthRadius})));
          int depthEnd = int(min(float(${this.depth}),
              float(d + ${depthRadius} + 1)));

          const int MIN_DEPTH_BEGIN = 0;
          const int MAX_DEPTH_END = ${this.depth};

          float norm = 0.0;
          for (int k = MIN_DEPTH_BEGIN; k < MAX_DEPTH_END; ++k) {
            if (k < depthBegin){
              continue;
            }
            else if (k >= depthBegin && k < depthEnd) {
              norm += getInputImage(b, r, c, k) * getInputImage(b, r, c, k);
            }
            else {
              break;
            }
          }

          norm = float(${alpha}) * norm + float(${bias});

          for(int k = MIN_DEPTH_BEGIN; k < MAX_DEPTH_END; ++k){
            if (k < depthBegin){
              continue;
            }
            else if (k >= depthBegin && k < depthEnd){
              float dyi = -2.0 * float(${alpha})
                * float(${beta})
                * getInputImage(b ,r ,c, k) * getOutputImage(b, r, c, d)
                / norm;
              if (k == d) {
                dyi += pow(norm, -1.0 * ${beta});
              }
              if (k == coords[3]) {
                dyi *= getDy(b, r, c, d);
                result += dyi;
              }
            }
            else {
              break;
            }
          }
      }
      setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const lrnGrad = (args) => {
        const { inputs, backend, attrs } = args;
        const { x, y, dy } = inputs;
        const { depthRadius, bias, alpha, beta } = attrs;
        const program = new LRNGradProgram(x.shape, depthRadius, bias, alpha, beta);
        return backend.runWebGLProgram(program, [x, y, dy], x.dtype);
    };
    // tslint:disable-next-line: variable-name
    const LRNGradConfig = {
        kernelName: LRNGrad,
        backendName: 'webgl',
        kernelFunc: lrnGrad
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxImpl(x, reduceShape, outShape, backend) {
        const inSize = sizeFromShape(reduceShape);
        const xSize = sizeFromShape(x.shape);
        const batchSize = xSize / inSize;
        const reshapedInput = reshape({ inputs: { x }, attrs: { shape: [batchSize, inSize] }, backend });
        const reduced = reduce(reshapedInput, x.dtype, 'max', backend);
        const reshapedOutput = reshape({ inputs: { x: reduced }, attrs: { shape: outShape }, backend });
        backend.disposeIntermediateTensorInfo(reshapedInput);
        backend.disposeIntermediateTensorInfo(reduced);
        return reshapedOutput;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function max(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { reductionIndices, keepDims } = attrs;
        const xRank = x.shape.length;
        const origAxes = parseAxisParam(reductionIndices, x.shape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, xRank);
        const maxInputIsTransposed = permutedAxes != null;
        const shouldExecuteOnCPU = backend.shouldExecuteOnCPU([x]);
        let maxInput = x;
        if (maxInputIsTransposed) {
            if (shouldExecuteOnCPU) {
                const xTexData = backend.texData.get(maxInput.dataId);
                const values = xTexData.values;
                const newShape = new Array(xRank);
                for (let i = 0; i < newShape.length; i++) {
                    newShape[i] = x.shape[permutedAxes[i]];
                }
                const maxInputValues = transposeImplCPU(values, x.shape, x.dtype, permutedAxes, newShape);
                maxInput = backend.makeTensorInfo(newShape, x.dtype);
                const maxInputData = backend.texData.get(maxInput.dataId);
                maxInputData.values = maxInputValues;
            }
            else {
                maxInput = transposeImpl(x, permutedAxes, backend);
            }
            axes = getInnerMostAxes(axes.length, xRank);
        }
        assertAxesAreInnerMostDims('max', axes, xRank);
        const [maxOutShape, reduceShape] = computeOutAndReduceShapes(maxInput.shape, axes);
        let outShape = maxOutShape;
        if (keepDims) {
            // rather than reshape at the end, set the target shape here.
            outShape = expandShapeToKeepDim(maxOutShape, origAxes);
        }
        let out;
        if (shouldExecuteOnCPU) {
            const xTexData = backend.texData.get(maxInput.dataId);
            const values = xTexData.values;
            const outValues = maxImplCPU(values, sizeFromShape(reduceShape), outShape, x.dtype);
            out = backend.makeTensorInfo(outShape, x.dtype);
            const outData = backend.texData.get(out.dataId);
            outData.values = outValues;
        }
        else {
            out = maxImpl(maxInput, reduceShape, outShape, backend);
        }
        if (maxInputIsTransposed) {
            backend.disposeIntermediateTensorInfo(maxInput);
        }
        return out;
    }
    const maxConfig = {
        kernelName: Max,
        backendName: 'webgl',
        kernelFunc: max
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const MAXIMUM = CHECK_NAN_SNIPPET$1 + `
  return max(a, b);
`;
    const MAXIMUM_PACKED = `
  vec4 result = vec4(max(a, b));
  vec4 isNaN = min(vec4(isnan(a)) + vec4(isnan(b)), vec4(1.0));
  ` +
        CHECK_NAN_SNIPPET + `
  return result;
`;
    const maximum = binaryKernelFunc({
        opSnippet: MAXIMUM,
        packedOpSnippet: MAXIMUM_PACKED,
        cpuKernelImpl: maximumImplCPU
    });
    const maximumConfig = {
        kernelName: Maximum$1,
        backendName: 'webgl',
        kernelFunc: maximum
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPool(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        assertNotComplex(x, 'maxPool');
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        const dilations = 1;
        assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in maxPool: Either strides or dilations must be 1. ' +
            `Got strides ${strides} and dilations '${dilations}'`);
        const convInfo = computePool2DInfo(x.shape, filterSize, strides, dilations, pad, dimRoundingMode);
        if (convInfo.filterWidth === 1 && convInfo.filterHeight === 1 &&
            arraysEqual(convInfo.inShape, convInfo.outShape)) {
            return identity({ inputs: { x }, backend });
        }
        const maxPoolProgram = new Pool2DProgram(convInfo, 'max', false);
        return backend.runWebGLProgram(maxPoolProgram, [x], x.dtype);
    }
    const maxPoolConfig = {
        kernelName: MaxPool,
        backendName: 'webgl',
        kernelFunc: maxPool
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPool3d(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { filterSize, strides, pad, dataFormat, dimRoundingMode } = attrs;
        const dilations = [1, 1, 1];
        const convInfo = computePool3DInfo(x.shape, filterSize, strides, dilations, pad, dimRoundingMode, dataFormat);
        const maxPoolProgram = new Pool3DProgram(convInfo, 'max', false);
        return backend.runWebGLProgram(maxPoolProgram, [x], x.dtype);
    }
    const maxPool3DConfig = {
        kernelName: MaxPool3D,
        backendName: 'webgl',
        kernelFunc: maxPool3d
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class MaxPool2DBackpropProgram {
        constructor(convInfo) {
            this.variableNames = ['dy', 'maxPos'];
            this.outputShape = convInfo.inShape;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const dilationHeight = convInfo.dilationHeight;
            const effectiveFilterHeight = convInfo.effectiveFilterHeight;
            const effectiveFilterWidth = convInfo.effectiveFilterWidth;
            const padTop = effectiveFilterHeight - 1 - convInfo.padInfo.top;
            const padLeft = effectiveFilterWidth - 1 - convInfo.padInfo.left;
            const lastIndex = effectiveFilterHeight * effectiveFilterWidth - 1;
            this.userCode = `
      const ivec2 pads = ivec2(${padTop}, ${padLeft});

      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int d = coords[3];

        ivec2 dyRCCorner = coords.yz - pads;
        int dyRCorner = dyRCCorner.x;
        int dyCCorner = dyRCCorner.y;

        // Convolve dy(?, ?, d) with pos mask(:, :, d) to get dx(xR, xC, d).
        // ? = to be determined. : = across all values in that axis.
        float dotProd = 0.0;
        for (int wR = 0; wR < ${effectiveFilterHeight};
          wR += ${dilationHeight}) {
          float dyR = float(dyRCorner + wR) / ${strideHeight}.0;

          if (dyR < 0.0 || dyR >= ${convInfo.outHeight}.0 || fract(dyR) > 0.0) {
            continue;
          }
          int idyR = int(dyR);

          for (int wC = 0; wC < ${effectiveFilterWidth}; wC++) {
            float dyC = float(dyCCorner + wC) / ${strideWidth}.0;

            if (dyC < 0.0 || dyC >= ${convInfo.outWidth}.0 ||
                fract(dyC) > 0.0) {
              continue;
            }
            int idyC = int(dyC);

            float dyValue = getDy(b, idyR, idyC, d);
            int maxPosValue = ${lastIndex} - int(getMaxPos(b, idyR, idyC, d));

            // Get the current value, check it against the value from the
            // position matrix.
            int curPosValue = wR * ${effectiveFilterWidth} + wC;
            float mask = float(maxPosValue == curPosValue ? 1.0 : 0.0);

            dotProd += dyValue * mask;
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }
    class MaxPool3DBackpropProgram {
        constructor(convInfo) {
            this.variableNames = ['dy', 'maxPos'];
            this.outputShape = convInfo.inShape;
            const strideDepth = convInfo.strideDepth;
            const strideHeight = convInfo.strideHeight;
            const strideWidth = convInfo.strideWidth;
            const dilationDepth = convInfo.dilationDepth;
            const dilationHeight = convInfo.dilationHeight;
            const dilationWidth = convInfo.dilationWidth;
            const effectiveFilterDepth = convInfo.effectiveFilterDepth;
            const effectiveFilterHeight = convInfo.effectiveFilterHeight;
            const effectiveFilterWidth = convInfo.effectiveFilterWidth;
            const padFront = effectiveFilterDepth - 1 - convInfo.padInfo.front;
            const padTop = effectiveFilterHeight - 1 - convInfo.padInfo.top;
            const padLeft = effectiveFilterWidth - 1 - convInfo.padInfo.left;
            const lastIndex = effectiveFilterDepth * effectiveFilterHeight * effectiveFilterWidth - 1;
            this.userCode = `
      const ivec3 pads = ivec3(${padFront}, ${padTop}, ${padLeft});

      void main() {
        ivec5 coords = getOutputCoords();
        int batch = coords.x;
        int ch = coords.u;

        ivec3 dyCorner = ivec3(coords.y, coords.z, coords.w) - pads;
        int dyDCorner = dyCorner.x;
        int dyRCorner = dyCorner.y;
        int dyCCorner = dyCorner.z;

        // Convolve dy(?, ?, ?, ch) with pos mask(:, :, :, d) to get
        // dx(xD, xR, xC, ch).
        // ? = to be determined. : = across all values in that axis.
        float dotProd = 0.0;

        for (int wD = 0; wD < ${effectiveFilterDepth};
           wD += ${dilationDepth}) {
          float dyD = float(dyDCorner + wD) / ${strideDepth}.0;

          if (dyD < 0.0 || dyD >= ${convInfo.outDepth}.0 || fract(dyD) > 0.0) {
            continue;
          }
          int idyD = int(dyD);

          for (int wR = 0; wR < ${effectiveFilterHeight};
              wR += ${dilationHeight}) {
            float dyR = float(dyRCorner + wR) / ${strideHeight}.0;

            if (dyR < 0.0 || dyR >= ${convInfo.outHeight}.0 ||
                fract(dyR) > 0.0) {
              continue;
            }
            int idyR = int(dyR);

            for (int wC = 0; wC < ${effectiveFilterWidth};
                wC += ${dilationWidth}) {
              float dyC = float(dyCCorner + wC) / ${strideWidth}.0;

              if (dyC < 0.0 || dyC >= ${convInfo.outWidth}.0 ||
                  fract(dyC) > 0.0) {
                continue;
              }
              int idyC = int(dyC);

              float dyValue = getDy(batch, idyD, idyR, idyC, ch);
              int maxPosValue = ${lastIndex} -
                  int(getMaxPos(batch, idyD, idyR, idyC, ch));

              // Get the current value, check it against the value from the
              // position matrix.
              int curPosValue =
                  wD * ${effectiveFilterHeight} * ${effectiveFilterWidth} +
                  wR * ${effectiveFilterWidth} + wC;
              float mask = float(maxPosValue == curPosValue ? 1.0 : 0.0);

              dotProd += dyValue * mask;
            }
          }
        }
        setOutput(dotProd);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPool3DGrad(args) {
        const { inputs, backend, attrs } = args;
        const { dy, input } = inputs;
        const x = input;
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        const dilations = [1, 1, 1];
        const convInfo = computePool3DInfo(x.shape, filterSize, strides, dilations, pad, dimRoundingMode);
        const maxPool3dPositionsProgram = new Pool3DProgram(convInfo, 'max', true /* get positions */);
        const maxPool3dPositions = backend.runWebGLProgram(maxPool3dPositionsProgram, [x], x.dtype);
        const maxPoolBackpropProgram = new MaxPool3DBackpropProgram(convInfo);
        const result = backend.runWebGLProgram(maxPoolBackpropProgram, [dy, maxPool3dPositions], x.dtype);
        backend.disposeIntermediateTensorInfo(maxPool3dPositions);
        return result;
    }
    const maxPool3DGradConfig = {
        kernelName: MaxPool3DGrad,
        backendName: 'webgl',
        kernelFunc: maxPool3DGrad
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPoolGrad(args) {
        const { inputs, backend, attrs } = args;
        const { dy, input, output } = inputs;
        const x = input;
        assertNotComplex([input, output], 'maxPoolGrad');
        const { filterSize, strides, pad, dimRoundingMode } = attrs;
        const convInfo = computePool2DInfo(x.shape, filterSize, strides, 1 /* dilations */, pad, dimRoundingMode);
        const getPositions = true;
        const maxPoolPositionsProgram = new Pool2DProgram(convInfo, 'max', getPositions);
        const maxPoolPositions = backend.runWebGLProgram(maxPoolPositionsProgram, [x], x.dtype);
        const maxPoolBackPropProgram = new MaxPool2DBackpropProgram(convInfo);
        const result = backend.runWebGLProgram(maxPoolBackPropProgram, [dy, maxPoolPositions], x.dtype);
        backend.disposeIntermediateTensorInfo(maxPoolPositions);
        return result;
    }
    const maxPoolGradConfig = {
        kernelName: MaxPoolGrad,
        backendName: 'webgl',
        kernelFunc: maxPoolGrad
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function maxPoolWithArgmaxImpl(x, includeBatchInIndex, convInfo, backend) {
        let program = new Pool2DProgram(convInfo, 'max', false);
        const poolOutput = backend.runWebGLProgram(program, [x], 'float32');
        program = new Pool2DProgram(convInfo, 'max', true, true, includeBatchInIndex);
        const indexOutput = backend.runWebGLProgram(program, [x], 'float32');
        return [poolOutput, indexOutput];
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const maxPoolWithArgmaxConfig = {
        kernelName: MaxPoolWithArgmax,
        backendName: 'webgl',
        kernelFunc: ({ inputs, attrs, backend }) => {
            const { x } = inputs;
            const { filterSize, strides, pad, includeBatchInIndex } = attrs;
            const webglBackend = backend;
            assert$4(x.shape.length === 4, () => `Error in maxPool: input must be rank 4 but got rank ${x.shape.length}.`);
            const dilations = [1, 1];
            assert$4(eitherStridesOrDilationsAreOne(strides, dilations), () => 'Error in maxPool: Either strides or dilations must be 1. ' +
                `Got strides ${strides} and dilations '${dilations}'`);
            const convInfo = computePool2DInfo(x.shape, filterSize, strides, dilations, pad);
            const [result, indexes] = maxPoolWithArgmaxImpl(x, includeBatchInIndex, convInfo, webglBackend);
            return [result, indexes];
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function meanImpl(x, reduceShape, outShape, backend) {
        const inSize = sizeFromShape(reduceShape);
        const xSize = sizeFromShape(x.shape);
        const batchSize = xSize / inSize;
        const reshapedInput = reshape({ inputs: { x }, attrs: { shape: [batchSize, inSize] }, backend });
        const reduced = reduce(reshapedInput, 'float32', 'mean', backend);
        const reshapedOutput = reshape({ inputs: { x: reduced }, attrs: { shape: outShape }, backend });
        backend.disposeIntermediateTensorInfo(reshapedInput);
        backend.disposeIntermediateTensorInfo(reduced);
        return reshapedOutput;
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const meanConfig = {
        kernelName: Mean,
        backendName: 'webgl',
        kernelFunc: ({ inputs, attrs, backend }) => {
            const { x } = inputs;
            const { keepDims, axis } = attrs;
            const webglBackend = backend;
            const xRank = x.shape.length;
            const origAxes = parseAxisParam(axis, x.shape);
            let axes = origAxes;
            const permutedAxes = getAxesPermutation(axes, xRank);
            const meanInputIsTransposed = permutedAxes != null;
            const shouldExecuteOnCPU = webglBackend.shouldExecuteOnCPU([x]);
            const intermediates = [];
            let meanInput = x;
            if (meanInputIsTransposed) {
                if (shouldExecuteOnCPU) {
                    const xTexData = webglBackend.texData.get(meanInput.dataId);
                    const values = xTexData.values;
                    const newShape = new Array(xRank);
                    for (let i = 0; i < newShape.length; i++) {
                        newShape[i] = x.shape[permutedAxes[i]];
                    }
                    const meanInputValues = transposeImplCPU(values, x.shape, x.dtype, permutedAxes, newShape);
                    meanInput = webglBackend.makeTensorInfo(newShape, x.dtype);
                    const meanInputData = webglBackend.texData.get(meanInput.dataId);
                    meanInputData.values = meanInputValues;
                }
                else {
                    meanInput = transposeImpl(x, permutedAxes, webglBackend);
                }
                intermediates.push(meanInput);
                axes = getInnerMostAxes(axes.length, xRank);
            }
            assertAxesAreInnerMostDims('sum', axes, xRank);
            const [meanOutShape, reduceShape] = computeOutAndReduceShapes(meanInput.shape, axes);
            let outShape = meanOutShape;
            if (keepDims) {
                // rather than reshape at the end, set the target shape here.
                outShape = expandShapeToKeepDim(meanOutShape, origAxes);
            }
            const out = meanImpl(meanInput, reduceShape, outShape, webglBackend);
            for (const i of intermediates) {
                webglBackend.disposeIntermediateTensorInfo(i);
            }
            return out;
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function min(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        const xRank = x.shape.length;
        const origAxes = parseAxisParam(axis, x.shape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, xRank);
        let permutedX = x;
        if (permutedAxes != null) {
            permutedX = transpose({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            axes = getInnerMostAxes(axes.length, x.shape.length);
        }
        assertAxesAreInnerMostDims('min', axes, xRank);
        const [outShape, reduceShape] = computeOutAndReduceShapes(permutedX.shape, axes);
        const inSize = sizeFromShape(reduceShape);
        const a2D = reshape({ inputs: { x: permutedX }, backend, attrs: { shape: [-1, inSize] } });
        const reduced = reduce(a2D, a2D.dtype, 'min', backend);
        let res;
        if (keepDims) {
            const newShape = expandShapeToKeepDim(outShape, origAxes);
            res = reshape({ inputs: { x: reduced }, backend, attrs: { shape: newShape } });
        }
        else {
            res = reshape({ inputs: { x: reduced }, backend, attrs: { shape: outShape } });
        }
        backend.disposeIntermediateTensorInfo(a2D);
        backend.disposeIntermediateTensorInfo(reduced);
        if (permutedAxes != null) {
            backend.disposeIntermediateTensorInfo(permutedX);
        }
        return res;
    }
    const minConfig = {
        kernelName: Min,
        backendName: 'webgl',
        kernelFunc: min
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const MINIMUM = CHECK_NAN_SNIPPET$1 + `
  return min(a, b);
`;
    const MINIMUM_PACKED = `
  vec4 result = vec4(min(a, b));
  vec4 isNaN = min(vec4(isnan(a)) + vec4(isnan(b)), vec4(1.0));
  ` +
        CHECK_NAN_SNIPPET + `
  return result;
`;
    const minimum = binaryKernelFunc({
        opSnippet: MINIMUM,
        packedOpSnippet: MINIMUM_PACKED,
        cpuKernelImpl: minimumImplCPU
    });
    const minimumConfig = {
        kernelName: Minimum$1,
        backendName: 'webgl',
        kernelFunc: minimum
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class MirrorPadProgram {
        constructor(xShape, paddings, mode) {
            this.variableNames = ['x'];
            this.outputShape = paddings.map((p, i) => p[0] /* beforePad */ + xShape[i] + p[1] /* afterPad */);
            const rank = xShape.length;
            const dtype = getCoordsDataType(rank);
            const start = paddings.map(p => p[0]).join(',');
            const end = paddings.map((p, i) => p[0] + xShape[i]).join(',');
            const unpackedCoords = ['coords[0]', 'coords[1]', 'coords[2]', 'coords[3]'].slice(0, rank);
            const offset = mode === 'reflect' ? 0 : 1;
            if (rank === 1) {
                this.userCode = `
        int start = ${start};
        int end = ${end};

        void main() {
          int outC = getOutputCoords();
          if (outC < start) {
            outC = start * 2 - outC - ${offset};
          } else if(outC >= end) {
            outC = (end - 1) * 2 - outC + ${offset};
          }
          setOutput(getX(outC - start));
        }
      `;
                return;
            }
            this.userCode = `
      ${dtype} start = ${dtype}(${start});
      ${dtype} end = ${dtype}(${end});

      void main() {
        ${dtype} outC = getOutputCoords();
        for (int i = 0; i < ${rank}; i++) {
          if (outC[i] < start[i]) {
            outC[i] = start[i] * 2 - outC[i] - ${offset};
          } else if(outC[i] >= end[i]) {
            outC[i] = (end[i] - 1) * 2 - outC[i] + ${offset};
          }
        }
        ${dtype} coords = outC - start;
        setOutput(getX(${unpackedCoords}));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    /**
     * Example shader code for
     * `mirrorPad(tf.tensor1d([1, 2, 3], 'int32'), [[2, 2]], 'reflect')`
     * ```
     *    const int start = int(2);
     *    const int end = int(5);
     *
     *    void main() {
     *       int outputLoc = getOutputCoords();
     *       vec4 result = vec4(0.);
     *
     *       int rc = outputLoc;
     *
     *       int source = rc;
     *       if (source < start) {
     *         source = start * 2 - source - 0;
     *       } else if (source >= end) {
     *         source = (end - 1) * 2 - source + 0;
     *       }
     *       source -= start;
     *
     *       result[0] = getChannel(getX(source), source);
     *       rc += 1;
     *       if(rc < 6) {
     *          int source = rc;
     *          if (source < start) {
     *            source = start * 2 - source - 0;
     *          } else if (source >= end) {
     *            source = (end - 1) * 2 - source + 0;
     *          }
     *          source -= start;
     *
     *         result[1] = getChannel(getX(source), source);
     *       }
     *
     *       setOutput(result);
     *     }
     * ```
     */
    class MirrorPadPackedProgram {
        constructor(xShape, paddings, mode) {
            this.variableNames = ['x'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.outputShape = paddings.map((p, i) => p[0] /* beforePad */ + xShape[i] + p[1] /* afterPad */);
            const rank = xShape.length;
            const dtype = getCoordsDataType(rank);
            const start = paddings.map(p => p[0]).join(',');
            const end = paddings.map((p, i) => p[0] + xShape[i]).join(',');
            const coords = getChannels('rc', rank);
            const source = getChannels('source', rank);
            const cLimit = `${coords[rank - 1]} < ${this.outputShape[rank - 1]}`;
            const innerDims = rank === 1 ? 'source' : `vec2(${source.slice(-2).join()})`;
            const offset = mode === 'reflect' ? 0 : 1;
            let mainLoop = '';
            if (rank === 1) {
                const padSetup = `
        ${dtype} source = rc;
        if (source < start) {
          source = start * 2 - source - ${offset};
        } else if (source >= end) {
          source = (end - 1) * 2 - source + ${offset};
        }
        source -= start;
      `;
                mainLoop = `
        ${dtype} rc = outputLoc;
        ${padSetup}
        result[0] = getChannel(getX(${source.join()}), ${innerDims});
        ${coords[rank - 1]} += 1;
        if(${cLimit}) {
          ${padSetup}
          result[1] = getChannel(getX(${source.join()}), ${innerDims});
        }
      `;
            }
            else {
                const padSetup = `
        ${dtype} source = rc;
        ${dtype} lt = ${dtype}(lessThan(source, start));
        ${dtype} gte = ${dtype}(greaterThanEqual(source, end));
        ${dtype} orig = 1 - (lt + gte);
        source = orig * source +
                lt * (start * 2 - source - ${offset}) +
                gte * ((end - 1) * 2 - source + ${offset});
        source -= start;
      `;
                mainLoop = `
        ${dtype} rc = outputLoc;
        ${padSetup}
        result[0] = getChannel(getX(${source.join()}), ${innerDims});
        ${coords[rank - 1]} += 1;
        if(${cLimit}) {
          ${padSetup}
          result[1] = getChannel(getX(${source.join()}), ${innerDims});
        }
        rc = outputLoc;
        ${coords[rank - 2]} += 1;
        if(${coords[rank - 2]} < ${this.outputShape[rank - 2]}) {
          ${padSetup}
          result[2] = getChannel(getX(${source.join()}), ${innerDims});
          ${coords[rank - 1]} += 1;
          if(${cLimit}) {
            ${padSetup}
            result[3] = getChannel(getX(${source.join()}), ${innerDims});
          }
        }
      `;
            }
            this.userCode = `
      const ${dtype} start = ${dtype}(${start});
      const ${dtype} end = ${dtype}(${end});

      void main() {
        ${dtype} outputLoc = getOutputCoords();
        vec4 result = vec4(0.);
        ${mainLoop}
        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const mirrorPadKernelFunc = ({ inputs, backend, attrs }) => {
        const { x } = inputs;
        const { paddings, mode } = attrs;
        const program = env$1().getBool('WEBGL_PACK_ARRAY_OPERATIONS') ?
            new MirrorPadPackedProgram(x.shape, paddings, mode) :
            new MirrorPadProgram(x.shape, paddings, mode);
        const output = backend.runWebGLProgram(program, [x], x.dtype);
        return output;
    };
    const mirrorPadConfig = {
        kernelName: MirrorPad,
        backendName: 'webgl',
        kernelFunc: mirrorPadKernelFunc,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const MOD = `if (b == 0.0) return NAN;
  return mod(a, b);`;
    const MOD_PACKED = `
  vec4 result = mod(a, b);
  vec4 isNaN = vec4(equal(b, vec4(0.0)));
  ` +
        CHECK_NAN_SNIPPET + `
  return result;
`;
    const mod = binaryKernelFunc({
        opSnippet: MOD,
        packedOpSnippet: MOD_PACKED,
    });
    const modConfig = {
        kernelName: Mod,
        backendName: 'webgl',
        kernelFunc: mod
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class MultinomialProgram {
        constructor(batchSize, numOutcomes, numSamples) {
            this.variableNames = ['probs'];
            this.customUniforms = [{ name: 'seed', type: 'float' }];
            this.outputShape = [batchSize, numSamples];
            this.userCode = `
      void main() {
        ivec2 coords = getOutputCoords();
        int batch = coords[0];

        float r = random(seed);
        float cdf = 0.0;

        for (int i = 0; i < ${numOutcomes - 1}; i++) {
          cdf += getProbs(batch, i);

          if (r < cdf) {
            setOutput(float(i));
            return;
          }
        }

        // If no other event happened, last event happened.
        setOutput(float(${numOutcomes - 1}));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Without the equality check div produces 0.9999 for a = b, which when
    // floored can cause errors.
    const DIV = `
if (a == b) {
  return 1.0;
};
return a / b;`;
    // We do the same as in ./binaryop_gpu, with vec4 and ivec4.
    // On Linux, the vectorized implementation produces NaNs when a and b are 0.
    const DIV_PACKED = `
  // vec4 one = vec4(equal(a, b));
  // return one + (vec4(1.0) - one) * a / b;
  vec4 result = a / b;
  if(a.x == b.x) {
    result.x = 1.;
  }
  if(a.y == b.y) {
    result.y = 1.;
  }
  if(a.z == b.z) {
    result.z = 1.;
  }
  if(a.w == b.w) {
    result.w = 1.;
  }

  return result;
`;
    const realDiv = binaryKernelFunc({ opSnippet: DIV, packedOpSnippet: DIV_PACKED, checkOutOfBounds: true });
    const realDivConfig = {
        kernelName: RealDiv,
        backendName: 'webgl',
        kernelFunc: realDiv,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SUB = 'return a - b;';
    const sub = binaryKernelFunc({
        opSnippet: SUB,
        packedOpSnippet: SUB,
        supportsComplex: true,
        cpuKernelImpl: subImplCPU
    });
    const subConfig = {
        kernelName: Sub,
        backendName: 'webgl',
        kernelFunc: sub
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function softmax(args) {
        const { inputs, backend, attrs } = args;
        const { logits } = inputs;
        const { dim } = attrs;
        const axes = parseAxisParam([dim], logits.shape);
        const maxLogit = max({
            inputs: { x: logits },
            backend,
            attrs: { reductionIndices: axes, keepDims: false }
        });
        const expandedShape = expandShapeToKeepDim(maxLogit.shape, axes);
        const maxLogitsReshaped = reshape({ inputs: { x: maxLogit }, backend, attrs: { shape: expandedShape } });
        const a = sub({ inputs: { a: logits, b: maxLogitsReshaped }, backend });
        const b = exp({ inputs: { x: a }, backend });
        const sumExp = sum({ inputs: { x: b }, backend, attrs: { axis: axes, keepDims: false } });
        const sumExpReshaped = reshape({ inputs: { x: sumExp }, backend, attrs: { shape: expandedShape } });
        const res = realDiv({ inputs: { a: b, b: sumExpReshaped }, backend });
        backend.disposeIntermediateTensorInfo(maxLogit);
        backend.disposeIntermediateTensorInfo(maxLogitsReshaped);
        backend.disposeIntermediateTensorInfo(a);
        backend.disposeIntermediateTensorInfo(b);
        backend.disposeIntermediateTensorInfo(sumExp);
        backend.disposeIntermediateTensorInfo(sumExpReshaped);
        return res;
    }
    const softmaxConfig = {
        kernelName: Softmax$2,
        backendName: 'webgl',
        kernelFunc: softmax
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function multinomial(args) {
        const { inputs, backend, attrs } = args;
        const { logits } = inputs;
        const { numSamples, seed, normalized } = attrs;
        const probs = normalized ?
            logits :
            softmax({ inputs: { logits }, backend, attrs: { dim: logits.shape.length - 1 } });
        const batchSize = probs.shape[0];
        const numOutcomes = probs.shape[1];
        const program = new MultinomialProgram(batchSize, numOutcomes, numSamples);
        const customValues = [[seed]];
        const res = backend.runWebGLProgram(program, [probs], 'int32', customValues);
        if (!normalized) {
            backend.disposeIntermediateTensorInfo(probs);
        }
        return res;
    }
    const multinomialConfig = {
        kernelName: Multinomial,
        backendName: 'webgl',
        kernelFunc: multinomial
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const NEG = CHECK_NAN_SNIPPET$2 + `
  return -x;
`;
    const NEG_PACKED = `
  vec4 result = -x;
  bvec4 isNaN = isnan(x);

  result.r = isNaN.r ? x.r : result.r;
  result.g = isNaN.g ? x.g : result.g;
  result.b = isNaN.b ? x.b : result.b;
  result.a = isNaN.a ? x.a : result.a;

  return result;
`;
    // This doesn't use unaryKernelFunc because negImplCPU is not of type
    // SimpleUnaryKernelImplCPU.
    function neg(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        if (backend.shouldExecuteOnCPU([x])) {
            const xData = backend.texData.get(x.dataId);
            const [outValues, newShape] = negImplCPU(xData.values, x.shape, x.dtype);
            return backend.makeTensorInfo(newShape, x.dtype, outValues);
        }
        let program;
        if (env$1().getBool('WEBGL_PACK_UNARY_OPERATIONS')) {
            program = new UnaryOpPackedProgram(x.shape, NEG_PACKED);
        }
        else {
            program = new UnaryOpProgram(x.shape, NEG);
        }
        return backend.runWebGLProgram(program, [x], x.dtype);
    }
    const negConfig = {
        kernelName: Neg,
        backendName: 'webgl',
        kernelFunc: neg
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const nonMaxSuppressionV3Impl = nonMaxSuppressionV3Impl$2;
    function nonMaxSuppressionV3(args) {
        warn('tf.nonMaxSuppression() in webgl locks the UI thread. ' +
            'Call tf.nonMaxSuppressionAsync() instead');
        const { inputs, backend, attrs } = args;
        const { boxes, scores } = inputs;
        const { maxOutputSize, iouThreshold, scoreThreshold } = attrs;
        const boxesVals = backend.readSync(boxes.dataId);
        const scoresVals = backend.readSync(scores.dataId);
        const { selectedIndices } = nonMaxSuppressionV3Impl(boxesVals, scoresVals, maxOutputSize, iouThreshold, scoreThreshold);
        return backend.makeTensorInfo([selectedIndices.length], 'int32', new Int32Array(selectedIndices));
    }
    const nonMaxSuppressionV3Config = {
        kernelName: NonMaxSuppressionV3,
        backendName: 'webgl',
        kernelFunc: nonMaxSuppressionV3
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const nonMaxSuppressionV4Impl = nonMaxSuppressionV4Impl$2;
    function nonMaxSuppressionV4(args) {
        warn('tf.nonMaxSuppression() in webgl locks the UI thread. ' +
            'Call tf.nonMaxSuppressionAsync() instead');
        const { inputs, backend, attrs } = args;
        const { boxes, scores } = inputs;
        const { maxOutputSize, iouThreshold, scoreThreshold, padToMaxOutputSize } = attrs;
        const boxesVals = backend.readSync(boxes.dataId);
        const scoresVals = backend.readSync(scores.dataId);
        const { selectedIndices, validOutputs } = nonMaxSuppressionV4Impl(boxesVals, scoresVals, maxOutputSize, iouThreshold, scoreThreshold, padToMaxOutputSize);
        return [
            backend.makeTensorInfo([selectedIndices.length], 'int32', new Int32Array(selectedIndices)),
            backend.makeTensorInfo([], 'int32', new Int32Array([validOutputs]))
        ];
    }
    const nonMaxSuppressionV4Config = {
        kernelName: NonMaxSuppressionV4,
        backendName: 'webgl',
        kernelFunc: nonMaxSuppressionV4
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const nonMaxSuppressionV5Impl = nonMaxSuppressionV5Impl$2;
    function nonMaxSuppressionV5(args) {
        warn('tf.nonMaxSuppression() in webgl locks the UI thread. ' +
            'Call tf.nonMaxSuppressionAsync() instead');
        const { inputs, backend, attrs } = args;
        const { boxes, scores } = inputs;
        const { maxOutputSize, iouThreshold, scoreThreshold, softNmsSigma } = attrs;
        const boxesVals = backend.readSync(boxes.dataId);
        const scoresVals = backend.readSync(scores.dataId);
        const maxOutputSizeVal = maxOutputSize;
        const iouThresholdVal = iouThreshold;
        const scoreThresholdVal = scoreThreshold;
        const softNmsSigmaVal = softNmsSigma;
        const { selectedIndices, selectedScores } = nonMaxSuppressionV5Impl(boxesVals, scoresVals, maxOutputSizeVal, iouThresholdVal, scoreThresholdVal, softNmsSigmaVal);
        return [
            backend.makeTensorInfo([selectedIndices.length], 'int32', new Int32Array(selectedIndices)),
            backend.makeTensorInfo([selectedScores.length], 'float32', new Float32Array(selectedScores))
        ];
    }
    const nonMaxSuppressionV5Config = {
        kernelName: NonMaxSuppressionV5,
        backendName: 'webgl',
        kernelFunc: nonMaxSuppressionV5
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class OneHotProgram {
        constructor(numIndices, depth, onValue, offValue) {
            this.variableNames = ['indices'];
            this.outputShape = [numIndices, depth];
            this.userCode = `
      void main() {
        ivec2 coords = getOutputCoords();
        int index = round(getIndices(coords.x));
        setOutput(mix(float(${offValue}), float(${onValue}),
                      float(index == coords.y)));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const oneHot = (args) => {
        const { inputs, backend, attrs } = args;
        const { indices } = inputs;
        const { depth, onValue, offValue } = attrs;
        const indicesSize = sizeFromShape(indices.shape);
        const program = new OneHotProgram(indicesSize, depth, onValue, offValue);
        const reshaped = reshape({ inputs: { x: indices }, backend, attrs: { shape: [indicesSize] } });
        const result = backend.runWebGLProgram(program, [reshaped], indices.dtype);
        backend.disposeIntermediateTensorInfo(reshaped);
        const outShape = [...indices.shape, depth];
        const out = reshape({ inputs: { x: result }, backend, attrs: { shape: outShape } });
        backend.disposeIntermediateTensorInfo(result);
        return out;
    };
    const oneHotConfig = {
        kernelName: OneHot,
        backendName: 'webgl',
        kernelFunc: oneHot
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function zerosLike(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        if (x.dtype === 'complex64') {
            const realPart = real({ inputs: { input: x }, backend });
            const r = zerosLike({ inputs: { x: realPart }, backend });
            const imagPart = imag({ inputs: { input: x }, backend });
            const i = zerosLike({ inputs: { x: imagPart }, backend });
            const result = complex({ inputs: { real: r, imag: i }, backend });
            backend.disposeIntermediateTensorInfo(realPart);
            backend.disposeIntermediateTensorInfo(r);
            backend.disposeIntermediateTensorInfo(imagPart);
            backend.disposeIntermediateTensorInfo(i);
            return result;
        }
        else {
            return fill({
                attrs: {
                    shape: x.shape,
                    dtype: x.dtype,
                    value: x.dtype === 'string' ? '' : 0
                },
                backend
            });
        }
    }
    const zerosLikeConfig = {
        kernelName: ZerosLike,
        backendName: 'webgl',
        kernelFunc: zerosLike
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function onesLike(args) {
        const { inputs, backend } = args;
        const { x } = inputs;
        if (x.dtype === 'string') {
            throw new Error('onesLike is not supported under string dtype');
        }
        else if (x.dtype === 'complex64') {
            const realPart = real({ inputs: { input: x }, backend });
            const r = onesLike({ inputs: { x: realPart }, backend });
            const imagPart = imag({ inputs: { input: x }, backend });
            const i = zerosLike({ inputs: { x: imagPart }, backend });
            const result = complex({ inputs: { real: r, imag: i }, backend });
            backend.disposeIntermediateTensorInfo(realPart);
            backend.disposeIntermediateTensorInfo(r);
            backend.disposeIntermediateTensorInfo(imagPart);
            backend.disposeIntermediateTensorInfo(i);
            return result;
        }
        else {
            // TODO(cais, smilkov): Add WebGL shader for onesLike:
            //   https://github.com/tensorflow/tfjs/issues/1293
            return fill({ attrs: { shape: x.shape, dtype: x.dtype, value: 1 }, backend });
        }
    }
    const onesLikeConfig = {
        kernelName: OnesLike,
        backendName: 'webgl',
        kernelFunc: onesLike
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function pack(args) {
        const { inputs, backend, attrs } = args;
        const { axis } = attrs;
        if (inputs.length === 1) {
            return expandDims({ inputs: { input: inputs[0] }, backend, attrs: { dim: axis } });
        }
        const shape = inputs[0].shape;
        const dtype = inputs[0].dtype;
        inputs.forEach(t => {
            assertShapesMatch(shape, t.shape, 'All tensors passed to stack must have matching shapes');
            assert$4(dtype === t.dtype, () => 'All tensors passed to stack must have matching dtypes');
        });
        const intermediateTensorInfos = [];
        const expandedTensors = inputs.map(t => {
            const expandedT = expandDims({ inputs: { input: t }, backend, attrs: { dim: axis } });
            intermediateTensorInfos.push(expandedT);
            return expandedT;
        });
        const result = concat({ inputs: expandedTensors, backend, attrs: { axis } });
        intermediateTensorInfos.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return result;
    }
    const packConfig = {
        kernelName: Pack,
        backendName: 'webgl',
        kernelFunc: pack
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class PadProgram {
        constructor(xShape, paddings, constantValue) {
            this.variableNames = ['x'];
            this.customUniforms = [{ name: 'value', type: 'float' }];
            this.outputShape = paddings.map((p, i) => p[0] /* beforePad */ + xShape[i] + p[1] /* afterPad */);
            const rank = xShape.length;
            const type = getCoordsDataType(rank);
            const start = paddings.map(p => p[0]).join(',');
            const end = paddings.map((p, i) => p[0] + xShape[i]).join(',');
            const unpackedCoords = ['coords[0]', 'coords[1]', 'coords[2]', 'coords[3]'].slice(0, rank);
            if (rank === 1) {
                this.userCode = `
        int start = ${start};
        int end = ${end};

        void main() {
          int outC = getOutputCoords();
          if (outC < start || outC >= end) {
            setOutput(value);
          } else {
            setOutput(getX(outC - start));
          }
        }
      `;
                return;
            }
            this.userCode = `
      ${type} start = ${type}(${start});
      ${type} end = ${type}(${end});

      void main() {
        ${type} outC = getOutputCoords();
        if (any(lessThan(outC, start)) || any(greaterThanEqual(outC, end))) {
          setOutput(value);
        } else {
          ${type} coords = outC - start;
          setOutput(getX(${unpackedCoords}));
        }
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class PadPackedProgram {
        constructor(xShape, paddings, constantValue) {
            this.variableNames = ['x'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.customUniforms = [{ name: 'value', type: 'float' }];
            this.outputShape = paddings.map((p, i) => p[0] /* beforePad */ + xShape[i] + p[1] /* afterPad */);
            const rank = xShape.length;
            const dtype = getCoordsDataType(rank);
            const start = paddings.map(p => p[0]).join(',');
            const end = paddings.map((p, i) => p[0] + xShape[i]).join(',');
            const coords = getChannels('rc', rank);
            const source = getChannels('source', rank);
            const cLimit = `${coords[rank - 1]} < ${this.outputShape[rank - 1]}`;
            const innerDims = rank === 1 ? 'source' : `vec2(${source.slice(-2).join()})`;
            const componentSetup = [
                `${dtype} rc = outputLoc;`, `${coords[rank - 1]} += 1;
       if(${cLimit}) {
      `,
                rank === 1 ? '' : `}
       rc = outputLoc;
       ${coords[rank - 2]} += 1;
       if(${coords[rank - 2]} < ${this.outputShape[rank - 2]}) {`,
                rank === 1 ? '' : `  ${coords[rank - 1]} += 1;
         if(${cLimit}) {`
            ];
            const paddingArea = rank === 1 ?
                'rc < start || rc >= end' :
                'any(lessThan(rc, start)) || any(greaterThanEqual(rc, end))';
            let mainLoop = '';
            for (let i = 0, j = rank === 1 ? 2 : 4; i < j; i++) {
                mainLoop += `
        ${componentSetup[i]}
        if (${paddingArea}) {
          result[${i}] = float(value);
        } else {
          ${dtype} source = rc - start;
          result[${i}] = getChannel(getX(${source.join()}), ${innerDims});
        }
      `;
            }
            mainLoop += (rank === 1 ? `} ` : `}}`);
            this.userCode = `
      const ${dtype} start = ${dtype}(${start});
      const ${dtype} end = ${dtype}(${end});

      void main() {
        ${dtype} outputLoc = getOutputCoords();
        vec4 result = vec4(0.);
        ${mainLoop}
        setOutput(result);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const padV2 = (args) => {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { paddings, constantValue } = attrs;
        if (sizeFromShape(x.shape) === 0) {
            // Short-circuit the computation, since x doesn't have value, only
            // the shape is used to compute output shape to pad.
            const outputShape = paddings.map((p, i) => p[0] /* beforePad */ + x.shape[i] + p[1] /* afterPad */);
            return fill({
                backend,
                attrs: { shape: outputShape, value: constantValue, dtype: x.dtype }
            });
        }
        const program = env$1().getBool('WEBGL_PACK_ARRAY_OPERATIONS') ?
            new PadPackedProgram(x.shape, paddings, constantValue) :
            new PadProgram(x.shape, paddings, constantValue);
        const customValues = [[constantValue]];
        return backend.runWebGLProgram(program, [x], x.dtype, customValues);
    };
    const padV2Config = {
        kernelName: PadV2,
        backendName: 'webgl',
        kernelFunc: padV2
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const POW = `
  if(a < 0.0 && floor(b) < b){
    return NAN;
  }
  if (b == 0.0) {
    return 1.0;
  }
  return (round(mod(b, 2.0)) != 1) ?
      pow(abs(a), b) : sign(a) * pow(abs(a), b);
`;
    const POW_PACKED = `
  // isModRound1 has 1 for components with round(mod(b, 2.0)) == 1, 0 otherwise.
  vec4 isModRound1 = vec4(equal(round(mod(b, 2.0)), ivec4(1)));
  vec4 multiplier = sign(a) * isModRound1 + (vec4(1.0) - isModRound1);
  vec4 result = multiplier * pow(abs(a), b);

  // Ensure that a^0 = 1, including 0^0 = 1 as this correspond to TF and JS
  bvec4 isExpZero = equal(b, vec4(0.0));
  result.r = isExpZero.r ? 1.0 : result.r;
  result.g = isExpZero.g ? 1.0 : result.g;
  result.b = isExpZero.b ? 1.0 : result.b;
  result.a = isExpZero.a ? 1.0 : result.a;

  vec4 isNaN = vec4(lessThan(a, vec4(0.0))) * vec4(lessThan(floor(b), b));
  ` +
        CHECK_NAN_SNIPPET + `
  return result;
`;
    const pow = binaryKernelFunc({ opSnippet: POW, packedOpSnippet: POW_PACKED });
    const powConfig = {
        kernelName: Pow,
        backendName: 'webgl',
        kernelFunc: pow
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function prod(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { axis, keepDims } = attrs;
        const xRank = x.shape.length;
        const toDispose = [];
        const origAxes = parseAxisParam(axis, x.shape);
        let axes = origAxes;
        const permutedAxes = getAxesPermutation(axes, xRank);
        let permutedX = x;
        if (permutedAxes != null) {
            permutedX = transpose({ inputs: { x }, backend, attrs: { perm: permutedAxes } });
            axes = getInnerMostAxes(axes.length, xRank);
            toDispose.push(permutedX);
        }
        assertAxesAreInnerMostDims('prod', axes, xRank);
        let res;
        if (backend.shouldExecuteOnCPU([permutedX])) {
            const xVals = backend.texData.get(permutedX.dataId).values;
            const { outVals, outShape, outDtype } = prodImplCPU(permutedX.shape, permutedX.dtype, xVals, axes);
            res = backend.makeTensorInfo(outShape, outDtype, outVals);
        }
        else {
            const [outShape, reduceShape] = computeOutAndReduceShapes(permutedX.shape, axes);
            const inSize = sizeFromShape(reduceShape);
            const a2D = reshape({ inputs: { x: permutedX }, backend, attrs: { shape: [-1, inSize] } });
            const outputDType = sumOutType(x.dtype);
            const reduced = reduce(a2D, outputDType, 'prod', backend);
            res = reshape({ inputs: { x: reduced }, backend, attrs: { shape: outShape } });
            toDispose.push(a2D);
            toDispose.push(reduced);
        }
        if (keepDims) {
            toDispose.push(res);
            const newShape = expandShapeToKeepDim(res.shape, origAxes);
            res = reshape({ inputs: { x: res }, backend, attrs: { shape: newShape } });
        }
        toDispose.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return res;
    }
    const prodConfig = {
        kernelName: Prod,
        backendName: 'webgl',
        kernelFunc: prod
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const range$1 = (args) => {
        const { backend, attrs } = args;
        const { start, stop, step, dtype } = attrs;
        const values = rangeImplCPU(start, stop, step, dtype);
        return backend.makeTensorInfo([values.length], dtype, values);
    };
    const rangeConfig = {
        kernelName: Range,
        backendName: 'webgl',
        kernelFunc: range$1
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const RECIPROCAL = `return 1.0 / x;`;
    const reciprocal = unaryKernelFunc({ opSnippet: RECIPROCAL });
    const reciprocalConfig = {
        kernelName: Reciprocal,
        backendName: 'webgl',
        kernelFunc: reciprocal,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const RELU = CHECK_NAN_SNIPPET$2 + `
  return (x < 0.0) ? 0.0 : x;
`;
    const RELU_PACKED = `
  vec4 result = x * vec4(greaterThanEqual(x, vec4(0.0)));
  bvec4 isNaN = isnan(x);

  result.r = isNaN.r ? x.r : result.r;
  result.g = isNaN.g ? x.g : result.g;
  result.b = isNaN.b ? x.b : result.b;
  result.a = isNaN.a ? x.a : result.a;

  return result;
`;
    const relu = unaryKernelFunc({ opSnippet: RELU, packedOpSnippet: RELU_PACKED });
    const reluConfig = {
        kernelName: Relu$1,
        backendName: 'webgl',
        kernelFunc: relu
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const RELU6 = CHECK_NAN_SNIPPET$2 + `
  return (x < 0.0) ? 0.0 : min(6.0, x);
`;
    const RELU6_PACKED = `
  vec4 result = min(x, vec4(6.)) * vec4(greaterThanEqual(x, vec4(0.0)));
  bvec4 isNaN = isnan(x);

  result.r = isNaN.r ? x.r : result.r;
  result.g = isNaN.g ? x.g : result.g;
  result.b = isNaN.b ? x.b : result.b;
  result.a = isNaN.a ? x.a : result.a;

  return result;
`;
    const relu6 = unaryKernelFunc({ opSnippet: RELU6, packedOpSnippet: RELU6_PACKED });
    const relu6Config = {
        kernelName: Relu6$1,
        backendName: 'webgl',
        kernelFunc: relu6
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ResizeBilinearProgram {
        constructor(inputShape, newHeight, newWidth, alignCorners, halfPixelCenters) {
            this.variableNames = ['A'];
            this.outputShape = [];
            const [batch, oldHeight, oldWidth, depth] = inputShape;
            this.outputShape = [batch, newHeight, newWidth, depth];
            const effectiveInSize = [
                (alignCorners && newHeight > 1) ? oldHeight - 1 : oldHeight,
                (alignCorners && newWidth > 1) ? oldWidth - 1 : oldWidth
            ];
            const effectiveOutSize = [
                (alignCorners && newHeight > 1) ? newHeight - 1 : newHeight,
                (alignCorners && newWidth > 1) ? newWidth - 1 : newWidth
            ];
            let sourceFracIndexRC;
            if (halfPixelCenters) {
                sourceFracIndexRC =
                    `(vec2(yRC) + vec2(0.5)) * effectiveInputOverOutputRatioRC` +
                        ` - vec2(0.5)`;
            }
            else {
                sourceFracIndexRC = `vec2(yRC) * effectiveInputOverOutputRatioRC`;
            }
            this.userCode = `
      const vec2 effectiveInputOverOutputRatioRC = vec2(
          ${effectiveInSize[0] / effectiveOutSize[0]},
          ${effectiveInSize[1] / effectiveOutSize[1]});
      const vec2 inputShapeRC = vec2(${oldHeight}.0, ${oldWidth}.0);

      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int d = coords[3];
        ivec2 yRC = coords.yz;

        // Fractional source index.
        vec2 sourceFracIndexRC = ${sourceFracIndexRC};

        // Compute the four integer indices.
        ivec2 sourceFloorRC = ivec2(max(sourceFracIndexRC, vec2(0.0)));
        ivec2 sourceCeilRC = ivec2(
          min(inputShapeRC - 1.0, ceil(sourceFracIndexRC)));

        float topLeft = getA(b, sourceFloorRC.x, sourceFloorRC.y, d);
        float bottomLeft = getA(b, sourceCeilRC.x, sourceFloorRC.y, d);
        float topRight = getA(b, sourceFloorRC.x, sourceCeilRC.y, d);
        float bottomRight = getA(b, sourceCeilRC.x, sourceCeilRC.y, d);

        vec2 fracRC = sourceFracIndexRC - vec2(sourceFloorRC);

        float top = topLeft + (topRight - topLeft) * fracRC.y;
        float bottom = bottomLeft + (bottomRight - bottomLeft) * fracRC.y;
        float newValue = top + (bottom - top) * fracRC.x;

        setOutput(newValue);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ResizeBilinearPackedProgram {
        constructor(inputShape, newHeight, newWidth, alignCorners, halfPixelCenters) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.outputShape = [];
            const [batch, oldHeight, oldWidth, depth] = inputShape;
            this.outputShape = [batch, newHeight, newWidth, depth];
            const effectiveInSize = [
                (alignCorners && newHeight > 1) ? oldHeight - 1 : oldHeight,
                (alignCorners && newWidth > 1) ? oldWidth - 1 : oldWidth
            ];
            const effectiveOutSize = [
                (alignCorners && newHeight > 1) ? newHeight - 1 : newHeight,
                (alignCorners && newWidth > 1) ? newWidth - 1 : newWidth
            ];
            let sourceFracIndexRC;
            if (halfPixelCenters) {
                sourceFracIndexRC = `(vec3(yRC) + vec3(0.5)) * ` +
                    `effectiveInputOverOutputRatioRC - vec3(0.5)`;
            }
            else {
                sourceFracIndexRC = `vec3(yRC) * effectiveInputOverOutputRatioRC`;
            }
            this.userCode = `
      const vec3 effectiveInputOverOutputRatioRC = vec3(
          ${effectiveInSize[0] / effectiveOutSize[0]},
          ${effectiveInSize[1] / effectiveOutSize[1]},
          ${effectiveInSize[1] / effectiveOutSize[1]});
      const vec3 inputShapeRC = vec3(${oldHeight}.0, ${oldWidth}.0,
                                     ${oldWidth}.0);

      float getAValue(int b, int r, int c, int d) {
        return getChannel(getA(b, r, c, d), vec2(c, d));
      }

      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int d = coords[3];
        // Calculate values for next column in yRC.z.
        ivec3 yRC = coords.yzz + ivec3(0, 0, 1);

        // Fractional source index.
        vec3 sourceFracIndexRC = ${sourceFracIndexRC};

        // Compute the four integer indices.
        ivec3 sourceFloorRC = ivec3(max(sourceFracIndexRC, vec3(0.0)));
        ivec3 sourceCeilRC = ivec3(
          min(inputShapeRC - 1.0, ceil(sourceFracIndexRC)));

        // Should we calculate next column and row elements in 2x2 packed cell.
        bool hasNextCol = d < ${depth - 1};
        bool hasNextRow = coords.z < ${newWidth - 1};

        // In parallel, construct four corners for all four components in
        // packed 2x2 cell.
        vec4 topLeft = vec4(
          getAValue(b, sourceFloorRC.x, sourceFloorRC.y, d),
          hasNextCol ? getAValue(b, sourceFloorRC.x, sourceFloorRC.y, d + 1)
                     : 0.0,
          hasNextRow ? getAValue(b, sourceFloorRC.x, sourceFloorRC.z, d)
                     : 0.0,
          (hasNextRow && hasNextCol) ?
            getAValue(b, sourceFloorRC.x, sourceFloorRC.z, d + 1) : 0.0);

        vec4 bottomLeft = vec4(
          getAValue(b, sourceCeilRC.x, sourceFloorRC.y, d),
          hasNextCol ? getAValue(b, sourceCeilRC.x, sourceFloorRC.y, d + 1)
                     : 0.0,
          hasNextRow ? getAValue(b, sourceCeilRC.x, sourceFloorRC.z, d)
                     : 0.0,
          (hasNextRow && hasNextCol) ?
            getAValue(b, sourceCeilRC.x, sourceFloorRC.z, d + 1) : 0.0);

        vec4 topRight = vec4(
          getAValue(b, sourceFloorRC.x, sourceCeilRC.y, d),
          hasNextCol ? getAValue(b, sourceFloorRC.x, sourceCeilRC.y, d + 1)
                     : 0.0,
          hasNextRow ? getAValue(b, sourceFloorRC.x, sourceCeilRC.z, d)
                     : 0.0,
          (hasNextRow && hasNextCol) ?
            getAValue(b, sourceFloorRC.x, sourceCeilRC.z, d + 1) : 0.0);

        vec4 bottomRight = vec4(
          getAValue(b, sourceCeilRC.x, sourceCeilRC.y, d),
          hasNextCol ? getAValue(b, sourceCeilRC.x, sourceCeilRC.y, d + 1)
                     : 0.0,
          hasNextRow ? getAValue(b, sourceCeilRC.x, sourceCeilRC.z, d)
                     : 0.0,
          (hasNextRow && hasNextCol) ?
            getAValue(b, sourceCeilRC.x, sourceCeilRC.z, d + 1) : 0.0);

        vec3 fracRC = sourceFracIndexRC - vec3(sourceFloorRC);

        vec4 top = mix(topLeft, topRight, fracRC.yyzz);
        vec4 bottom = mix(bottomLeft, bottomRight, fracRC.yyzz);
        vec4 newValue = mix(top, bottom, fracRC.x);

        setOutput(newValue);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function resizeBilinear(args) {
        const { inputs, backend, attrs } = args;
        const { images } = inputs;
        const { alignCorners, halfPixelCenters, size } = attrs;
        const [newHeight, newWidth] = size;
        const program = env$1().getBool('WEBGL_PACK_IMAGE_OPERATIONS') ?
            new ResizeBilinearPackedProgram(images.shape, newHeight, newWidth, alignCorners, halfPixelCenters) :
            new ResizeBilinearProgram(images.shape, newHeight, newWidth, alignCorners, halfPixelCenters);
        return backend.runWebGLProgram(program, [images], 'float32');
    }
    const resizeBilinearConfig = {
        kernelName: ResizeBilinear,
        backendName: 'webgl',
        kernelFunc: resizeBilinear
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ResizeBilinearBackpropProgram {
        constructor(dyShape, inputShape, alignCorners) {
            this.variableNames = ['dy'];
            this.outputShape = [];
            this.outputShape = inputShape;
            const [, xHeight, xWidth,] = inputShape;
            const [, yHeight, yWidth] = dyShape;
            // In the backwards pass, we want to find the pixels that were generated for
            // each pixel in the input image the forward pass and add the corresponding
            // coefficient from dy to the gradient (with some interpolation).
            const effectiveXSize = [
                (alignCorners && yHeight > 1) ? xHeight - 1 : xHeight,
                (alignCorners && yWidth > 1) ? xWidth - 1 : xWidth
            ];
            const effectiveYSize = [
                (alignCorners && yHeight > 1) ? yHeight - 1 : yHeight,
                (alignCorners && yWidth > 1) ? yWidth - 1 : yWidth
            ];
            const heightScale = effectiveXSize[0] / effectiveYSize[0];
            const widthScale = effectiveXSize[1] / effectiveYSize[1];
            const invHeightScale = 1 / heightScale;
            const invWidthScale = 1 / widthScale;
            // This defines the size of the window of values around a particular
            // index in dy that we want to search for contributions to dx.
            const winHeight = (Math.ceil(invHeightScale) * 2) + 2;
            const winWidth = (Math.ceil(invWidthScale) * 2) + 2;
            this.userCode = `
      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int d = coords[3];
        int r = coords[1];
        int c = coords[2];

        float accumulator = 0.0;

        const float heightScale = float(${heightScale});
        const float widthScale = float(${widthScale});

        const float invHeightScale = float(${invHeightScale});
        const float invWidthScale = float(${invWidthScale});

        const int winHeight = int(${winHeight});
        const int winWidth = int(${winWidth});

        // Compute bounds for where in dy we will look
        float startRLerp = floor(float(r) * invHeightScale);
        int startDyR = int(startRLerp - float(winHeight / 2));

        float startCLerp = floor(float(c) * invWidthScale);
        int startDyC = int(startCLerp - float(winWidth / 2));

        // Loop over dy
        for (int dyROffset = 0; dyROffset < winHeight; dyROffset++) {
          int dyR = dyROffset + startDyR;

          // Guard against the window exceeding the bounds of dy
          if (dyR < 0 || dyR >= ${yHeight}) {
            continue;
          }

          for (int dyCOffset = 0; dyCOffset < winWidth; dyCOffset++) {
            int dyC = dyCOffset + startDyC;

            // Guard against the window exceeding the bounds of dy
            if (dyC < 0 || dyC >= ${yWidth}) {
              continue;
            }

            float dxR = float(dyR) * heightScale;
            int topDxRIndex = int(floor(dxR));
            int bottomDxRIndex = int(min(ceil(dxR), ${xHeight - 1}.0));
            float dxRLerp = dxR - float(topDxRIndex);
            float inverseDxRLerp = 1.0 - dxRLerp;

            float dxC = float(dyC) * widthScale;
            int leftDxCIndex = int(floor(dxC));
            int rightDxCIndex = int(min(ceil(dxC), ${xWidth - 1}.0));
            float dxCLerp = dxC - float(leftDxCIndex);
            float inverseDxCLerp = 1.0 - dxCLerp;

            if (r == topDxRIndex && c == leftDxCIndex) {
              // topLeft
              accumulator +=
                getDy(b, dyR, dyC, d) * inverseDxRLerp * inverseDxCLerp;
            }

            if (r == topDxRIndex && c == rightDxCIndex) {
              // topRight
              accumulator += getDy(b, dyR, dyC, d) * inverseDxRLerp * dxCLerp;
            }

            if (r == bottomDxRIndex && c == leftDxCIndex) {
              // bottomLeft
              accumulator += getDy(b, dyR, dyC, d) * dxRLerp * inverseDxCLerp;
            }

            if (r == bottomDxRIndex && c == rightDxCIndex) {
              // bottomRight
              accumulator += getDy(b, dyR, dyC, d) * dxRLerp * dxCLerp;
            }
          }
        }
        // End loop over dy

        setOutput(accumulator);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function resizeBilinearGrad(args) {
        const { inputs, backend, attrs } = args;
        const { images, dy } = inputs;
        const { alignCorners } = attrs;
        const program = new ResizeBilinearBackpropProgram(dy.shape, images.shape, alignCorners);
        return backend.runWebGLProgram(program, [dy], dy.dtype);
    }
    const resizeBilinearGradConfig = {
        kernelName: ResizeBilinearGrad,
        backendName: 'webgl',
        kernelFunc: resizeBilinearGrad
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ResizeNearestNeighborProgram {
        constructor(inputShape, newHeight, newWidth, alignCorners, halfPixelCenters) {
            this.variableNames = ['A'];
            this.outputShape = [];
            const [batch, oldHeight, oldWidth, depth] = inputShape;
            this.outputShape = [batch, newHeight, newWidth, depth];
            const effectiveInSize = [
                (alignCorners && newHeight > 1) ? oldHeight - 1 : oldHeight,
                (alignCorners && newWidth > 1) ? oldWidth - 1 : oldWidth
            ];
            const effectiveOutSize = [
                (alignCorners && newHeight > 1) ? newHeight - 1 : newHeight,
                (alignCorners && newWidth > 1) ? newWidth - 1 : newWidth
            ];
            // When align corners is false, we rounds the value with floor.
            const roundBase = alignCorners ? '0.5' : '0.0';
            let sourceFracIndexRC;
            if (halfPixelCenters) {
                sourceFracIndexRC =
                    `max((vec2(yRC) + vec2(0.5)) * effectiveInputOverOutputRatioRC` +
                        `, vec2(0.0))`;
            }
            else {
                sourceFracIndexRC = `vec2(yRC) * effectiveInputOverOutputRatioRC`;
            }
            this.userCode = `
      const vec2 effectiveInputOverOutputRatioRC = vec2(
          ${effectiveInSize[0] / effectiveOutSize[0]},
          ${effectiveInSize[1] / effectiveOutSize[1]});
      const vec2 inputShapeRC = vec2(${oldHeight}.0, ${oldWidth}.0);

      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int d = coords[3];
        ivec2 yRC = coords.yz;

        // Fractional source index.
        vec2 sourceFracIndexRC = ${sourceFracIndexRC};

        // Compute the coordinators of nearest neighbor point.
        ivec2 sourceNearestRC = ivec2(
          min(inputShapeRC - 1.0, floor(sourceFracIndexRC + ${roundBase})));
        float newValue = getA(b, sourceNearestRC.x, sourceNearestRC.y, d);

        setOutput(newValue);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ResizeNearestNeighborPackedProgram {
        constructor(inputShape, newHeight, newWidth, alignCorners, halfPixelCenters) {
            this.variableNames = ['A'];
            this.packedInputs = true;
            this.packedOutput = true;
            this.outputShape = [];
            const [batch, oldHeight, oldWidth, depth] = inputShape;
            this.outputShape = [batch, newHeight, newWidth, depth];
            const effectiveInSize = [
                (alignCorners && newHeight > 1) ? oldHeight - 1 : oldHeight,
                (alignCorners && newWidth > 1) ? oldWidth - 1 : oldWidth
            ];
            const effectiveOutSize = [
                (alignCorners && newHeight > 1) ? newHeight - 1 : newHeight,
                (alignCorners && newWidth > 1) ? newWidth - 1 : newWidth
            ];
            // When align corners is false, we rounds the value with floor.
            const roundBase = alignCorners ? '0.5' : '0.0';
            let sourceFracIndexRC;
            if (halfPixelCenters) {
                sourceFracIndexRC = `max((vec3(yRC) + vec3(0.5)) * ` +
                    `effectiveInputOverOutputRatioRC, vec3(0.0))`;
            }
            else {
                sourceFracIndexRC = `vec3(yRC) * effectiveInputOverOutputRatioRC`;
            }
            this.userCode = `
      const vec3 effectiveInputOverOutputRatioRC = vec3(
          ${effectiveInSize[0] / effectiveOutSize[0]},
          ${effectiveInSize[1] / effectiveOutSize[1]},
          ${effectiveInSize[1] / effectiveOutSize[1]});
      const vec3 inputShapeRC = vec3(${oldHeight}.0, ${oldWidth}.0,
                                     ${oldWidth}.0);

      float getAValue(int b, int r, int c, int d) {
        return getChannel(getA(b, r, c, d), vec2(c, d));
      }

      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int d = coords[3];
        // Calculate values for next column in yRC.z.
        ivec3 yRC = coords.yzz + ivec3(0, 0, 1);

        // Fractional source index.
        vec3 sourceFracIndexRC = ${sourceFracIndexRC};

        // Compute the coordinators of nearest neighbor point.
        ivec3 sourceNearestRC = ivec3(
          min(inputShapeRC - 1.0, floor(sourceFracIndexRC + ${roundBase})));

        // Should we calculate next column and row elements in 2x2 packed cell.
        bool hasNextCol = d < ${depth - 1};
        bool hasNextRow = coords.z < ${newWidth - 1};

        vec4 newValue = vec4(
          getAValue(b, sourceNearestRC.x, sourceNearestRC.y, d),
          hasNextCol ? getAValue(b, sourceNearestRC.x, sourceNearestRC.y, d + 1)
                     : 0.0,
          hasNextRow ? getAValue(b, sourceNearestRC.x, sourceNearestRC.z, d)
                     : 0.0,
          (hasNextRow && hasNextCol) ?
            getAValue(b, sourceNearestRC.x, sourceNearestRC.z, d + 1) : 0.0);

        setOutput(newValue);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function resizeNearestNeighbor(args) {
        const { inputs, backend, attrs } = args;
        const { images } = inputs;
        const { alignCorners, halfPixelCenters, size } = attrs;
        const [newHeight, newWidth] = size;
        const program = env$1().getBool('WEBGL_PACK_IMAGE_OPERATIONS') ?
            new ResizeNearestNeighborPackedProgram(images.shape, newHeight, newWidth, alignCorners, halfPixelCenters) :
            new ResizeNearestNeighborProgram(images.shape, newHeight, newWidth, alignCorners, halfPixelCenters);
        return backend.runWebGLProgram(program, [images], images.dtype);
    }
    const resizeNearestNeighborConfig = {
        kernelName: ResizeNearestNeighbor,
        backendName: 'webgl',
        kernelFunc: resizeNearestNeighbor
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ResizeNearestNeigborBackpropProgram {
        constructor(dyShape, inputShape, alignCorners) {
            this.variableNames = ['dy'];
            this.outputShape = [];
            this.outputShape = inputShape;
            const [, xHeight, xWidth,] = inputShape;
            const [, yHeight, yWidth] = dyShape;
            // In the backwards pass, we want to find the pixels that were generated for
            // each pixel in the input image the forward pass and add the corresponding
            // coefficient from dy to the gradient (with some interpolation).
            const effectiveXSize = [
                (alignCorners && yHeight > 1) ? xHeight - 1 : xHeight,
                (alignCorners && yWidth > 1) ? xWidth - 1 : xWidth
            ];
            const effectiveYSize = [
                (alignCorners && yHeight > 1) ? yHeight - 1 : yHeight,
                (alignCorners && yWidth > 1) ? yWidth - 1 : yWidth
            ];
            const heightScale = effectiveXSize[0] / effectiveYSize[0];
            const widthScale = effectiveXSize[1] / effectiveYSize[1];
            const invHeightScale = 1 / heightScale;
            const invWidthScale = 1 / widthScale;
            // This defines the size of the window of values around a particular
            // index in dy that we want to search for contributions to dx.
            const winHeight = (Math.ceil(invHeightScale) * 2) + 2;
            const winWidth = (Math.ceil(invWidthScale) * 2) + 2;
            this.userCode = `
      void main() {
        ivec4 coords = getOutputCoords();
        int b = coords[0];
        int d = coords[3];
        int r = coords[1];
        int c = coords[2];

        float accumulator = 0.0;

        const float heightScale = float(${heightScale});
        const float widthScale = float(${widthScale});

        const float invHeightScale = float(${invHeightScale});
        const float invWidthScale = float(${invWidthScale});

        const int winHeight = int(${winHeight});
        const int winWidth = int(${winWidth});

        // Compute bounds for where in dy we will look
        float startRLerp = floor(float(r) * invHeightScale);
        int startDyR = int(floor(startRLerp - float(winHeight / 2)));

        float startCLerp = floor(float(c) * invWidthScale);
        int startDyC = int(floor(startCLerp - float(winWidth / 2)));

        // Loop over dy
        for (int dyROffset = 0; dyROffset < winHeight; dyROffset++) {
          int dyR = dyROffset + startDyR;

          // Guard against the window exceeding the bounds of dy
          if (dyR < 0 || dyR >= ${yHeight}) {
            continue;
          }

          for (int dyCOffset = 0; dyCOffset < winWidth; dyCOffset++) {
            int dyC = dyCOffset + startDyC;

            // Guard against the window exceeding the bounds of dy
            if (dyC < 0 || dyC >= ${yWidth}) {
              continue;
            }

            float sourceFracRow =
              float(${effectiveXSize[0]}) *
                (float(dyR) / float(${effectiveYSize[0]}));

            float sourceFracCol =
                float(${effectiveXSize[1]}) *
                  (float(dyC) / float(${effectiveYSize[1]}));

            int sourceNearestRow = int(min(
                float(int(${xHeight}) - 1),
                ${alignCorners} ? float(round(sourceFracRow)) :
                                  float(floor(sourceFracRow))));

            int sourceNearestCol = int(min(
                float(int(${xWidth}) - 1),
                ${alignCorners} ? float(round(sourceFracCol)) :
                                  float(floor(sourceFracCol))));

            if (r == sourceNearestRow && c == sourceNearestCol) {
              accumulator += getDy(b, dyR, dyC, d);
            }
          }
        }
        // End loop over dy

        setOutput(accumulator);
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function resizeNearestNeighborGrad(args) {
        const { inputs, backend, attrs } = args;
        const { images, dy } = inputs;
        const { alignCorners } = attrs;
        const program = new ResizeNearestNeigborBackpropProgram(dy.shape, images.shape, alignCorners);
        return backend.runWebGLProgram(program, [dy], dy.dtype);
    }
    const resizeNearestNeighborGradConfig = {
        kernelName: ResizeNearestNeighborGrad,
        backendName: 'webgl',
        kernelFunc: resizeNearestNeighborGrad
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ReverseProgram {
        constructor(xShape, axis) {
            this.variableNames = ['x'];
            const rank = xShape.length;
            if (rank > 4) {
                throw new Error(`WebGL backend: Reverse of rank-${rank} tensor is not yet supported`);
            }
            this.outputShape = xShape;
            if (rank === 1) {
                this.userCode = `
        void main() {
          int coord = getOutputCoords();
          setOutput(getX(${xShape[0]} - coord - 1));
        }
      `;
                return;
            }
            const getInCoord = (i) => {
                if (axis.indexOf(i) !== -1 && xShape[i] !== 1) {
                    return `${xShape[i]} - coords[${i}] - 1`;
                }
                return `coords[${i}]`;
            };
            const inCoords = xShape.map((_, i) => getInCoord(i)).join(',');
            const type = getCoordsDataType(rank);
            this.userCode = `
      void main() {
        ${type} coords = getOutputCoords();
        setOutput(getX(${inCoords}));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ReversePackedProgram {
        constructor(xShape, axis) {
            this.variableNames = ['x'];
            this.packedInputs = true;
            this.packedOutput = true;
            const rank = xShape.length;
            if (rank > 4) {
                throw new Error(`WebGL backend: Reverse of rank-${rank} tensor is not yet supported`);
            }
            this.outputShape = xShape;
            const channels = getChannels('rc', rank);
            const nextColumn = `${channels[rank - 1]} + 1 < ${this.outputShape[rank - 1]}`;
            const nextRow = `${channels[rank - 2]} + 1 < ${this.outputShape[rank - 2]}`;
            const type = getCoordsDataType(rank);
            if (rank === 1) {
                this.userCode = `
        void main(){
          int rc = getOutputCoords();
          vec4 result = vec4(0.);
          result.r = getChannel(getX(${xShape[0]} - rc - 1),
            ${xShape[0]} - rc - 1);
          if(${nextColumn}){
              result.g = getChannel(getX(${xShape[0]} - (rc  + 1) - 1),
                ${xShape[0]} - (rc  + 1) - 1);
          }
          setOutput(result);
        }
      `;
            }
            else {
                this.userCode = `
        void main() {
          ${type} rc = getOutputCoords();
          vec4 result = vec4(0.);
          result.r = ${getR(channels.slice())};
          if(${nextColumn}){
            result.g = ${getG(channels.slice())};
          }
          if(${nextRow}) {
            result.b = ${getB(channels.slice())};
            if(${nextColumn}) {
              result.a = ${getA(channels.slice())};
            }
          }
          setOutput(result);
        }
    `;
            }
            function getR(channels) {
                return getChannel(channels);
            }
            function getG(channels) {
                channels[rank - 1] = '(' + channels[rank - 1] + ` + 1)`;
                return getChannel(channels);
            }
            function getB(channels) {
                channels[rank - 2] = '(' + channels[rank - 2] + ` + 1)`;
                return getChannel(channels);
            }
            function getA(channels) {
                channels[rank - 1] = '(' + channels[rank - 1] + ` + 1)`;
                channels[rank - 2] = '(' + channels[rank - 2] + ` + 1)`;
                return getChannel(channels);
            }
            function getChannel(channels) {
                const inCoordsArray = xShape.map((_, i) => getInCoord(i, channels));
                const inCoords = inCoordsArray.join(',');
                const innerDims = inCoordsArray.slice(-2).join(',');
                return `getChannel(getX(${inCoords}), vec2(${innerDims}))`;
            }
            function getInCoord(i, channels1) {
                if (axis.indexOf(i) !== -1 && xShape[i] !== 1) {
                    return `${xShape[i]} - ${channels1[i]} - 1`;
                }
                else {
                    return `${channels1[i]}`;
                }
            }
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function reverse(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { dims } = attrs;
        const xRank = x.shape.length;
        const $dims = parseAxisParam(dims, x.shape);
        if (xRank === 0) {
            return identity({ inputs: { x }, backend });
        }
        const program = env$1().getBool('WEBGL_PACK_ARRAY_OPERATIONS') ?
            new ReversePackedProgram(x.shape, $dims) :
            new ReverseProgram(x.shape, $dims);
        return backend.runWebGLProgram(program, [x], x.dtype);
    }
    const reverseConfig = {
        kernelName: Reverse,
        backendName: 'webgl',
        kernelFunc: reverse
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class RotateProgram {
        constructor(imageShape, fillValue) {
            this.variableNames = ['Image'];
            this.outputShape = [];
            this.customUniforms = [{ name: 'params', type: 'vec4' }];
            const imageHeight = imageShape[1];
            const imageWidth = imageShape[2];
            this.outputShape = imageShape;
            let fillSnippet = '';
            if (typeof fillValue === 'number') {
                fillSnippet = `float outputValue = ${fillValue.toFixed(2)};`;
            }
            else {
                fillSnippet = `
        vec3 fill = vec3(${fillValue.join(',')});
        float outputValue = fill[coords[3]];`;
            }
            this.userCode = `
        void main() {
          ivec4 coords = getOutputCoords();
          int x = coords[2];
          int y = coords[1];
          float coordXFloat = (float(x) - params[0]) * params[3] -
            (float(y) - params[1]) * params[2];
          float coordYFloat = (float(x) - params[0]) * params[2] +
            (float(y) - params[1]) * params[3];
          int coordX = int(round(coordXFloat + params[0]));
          int coordY = int(round(coordYFloat + params[1]));
          ${fillSnippet}
          if(coordX >= 0 && coordX < ${imageWidth} && coordY >= 0 && coordY < ${imageHeight}) {
            outputValue = getImage(coords[0], coordY, coordX, coords[3]);
          }
          setOutput(outputValue);
        }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const rotateWithOffsetConfig = {
        kernelName: RotateWithOffset,
        backendName: 'webgl',
        kernelFunc: ({ inputs, attrs, backend }) => {
            const { image } = inputs;
            const { radians, fillValue, center } = attrs;
            const webglBackend = backend;
            const program = new RotateProgram(image.shape, fillValue);
            const [centerX, centerY] = getImageCenter(center, image.shape[1], image.shape[2]);
            const customValues = [[centerX, centerY, Math.sin(radians), Math.cos(radians)]];
            const output = webglBackend.runWebGLProgram(program, [image], image.dtype, customValues);
            return output;
        }
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const ROUND = `
  // OpenGL ES does not support round function.
  // The algorithm is based on banker's rounding.
  float base = floor(x);
  if ((x - base) < 0.5) {
    return floor(x);
  } else if ((x - base) > 0.5) {
    return ceil(x);
  } else {
    if (mod(base, 2.0) == 0.0) {
      return base;
    } else {
      return base + 1.0;
    }
  }
`;
    const round = unaryKernelFunc({ opSnippet: ROUND });
    const roundConfig = {
        kernelName: Round,
        backendName: 'webgl',
        kernelFunc: round,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const RSQRT = `return inversesqrt(x);`;
    const rsqrt = unaryKernelFunc({ opSnippet: RSQRT, cpuKernelImpl: rsqrtImplCPU });
    const rsqrtConfig = {
        kernelName: Rsqrt,
        backendName: 'webgl',
        kernelFunc: rsqrt
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class ScatterProgram {
        constructor(updateSize, sliceDim, indicesRank, updatesRank, strides, shape, summingDupeIndex = true) {
            this.variableNames = ['updates', 'indices', 'defaultValue'];
            this.outputShape = shape;
            const stridesType = getCoordsDataType(strides.length);
            const dtype = getCoordsDataType(shape.length);
            let indicesString = '';
            if (indicesRank === 1) {
                indicesString = 'i';
            }
            else if (indicesRank === 2) {
                indicesString = 'i, j';
            }
            const indicesSnippet = `getIndices(${indicesString})`;
            let updatesString = '';
            if (updatesRank === 1) {
                updatesString = 'i';
            }
            else if (updatesRank === 2) {
                updatesString = 'i, coords[1]';
            }
            const updatesSnippet = `getUpdates(${updatesString})`;
            const strideString = sliceDim > 1 ? 'strides[j]' : 'strides';
            this.userCode = `
        ${stridesType} strides = ${stridesType}(${strides});

        void main() {
          ${dtype} coords = getOutputCoords();
          float sum = 0.0;
          bool found = false;
          for (int i = 0; i < ${updateSize}; i++) {
            int flattenedIndex = 0;
            for (int j = 0; j < ${sliceDim}; j++) {
              int index = round(${indicesSnippet});
              flattenedIndex += index * ${strideString};
            }
            if (flattenedIndex == coords[0]) {
              sum += ${updatesSnippet};
              found = true;
            }
          }
          setOutput(mix(getDefaultValue(), sum, float(found)));
        }
      `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function scatterNd(args) {
        const { inputs, backend, attrs } = args;
        const { indices, updates } = inputs;
        const { shape } = attrs;
        const { sliceRank, numUpdates, sliceSize, strides, outputSize } = calculateShapes(updates, indices, shape);
        const flattenShape = [outputSize / sliceSize, sliceSize];
        if (outputSize === 0) {
            return backend.makeTensorInfo(shape, indices.dtype);
        }
        const flattenIndices = reshape({ inputs: { x: indices }, backend, attrs: { shape: [numUpdates, sliceRank] } });
        const flattenX = reshape({ inputs: { x: updates }, backend, attrs: { shape: [numUpdates, sliceSize] } });
        const defaultValue = backend.makeTensorInfo([], 'float32', new Float32Array([0])); // scalar(0)
        const program = new ScatterProgram(numUpdates, sliceRank, flattenIndices.shape.length, flattenX.shape.length, strides, flattenShape);
        const res = backend.runWebGLProgram(program, [flattenX, flattenIndices, defaultValue], flattenX.dtype);
        const reshaped = reshape({ inputs: { x: res }, backend, attrs: { shape } });
        backend.disposeIntermediateTensorInfo(flattenIndices);
        backend.disposeIntermediateTensorInfo(flattenX);
        backend.disposeIntermediateTensorInfo(res);
        backend.disposeIntermediateTensorInfo(defaultValue);
        return reshaped;
    }
    const scatterNdConfig = {
        kernelName: ScatterNd,
        backendName: 'webgl',
        kernelFunc: scatterNd
    };

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class SearchSortedProgram {
        constructor(batchSize, numInputs, numValues, side) {
            this.variableNames = ['sortedSequence', 'values'];
            this.customUniforms = [{ name: 'numInputs', type: 'int' }];
            this.outputShape = [batchSize, numValues];
            const webGL2LoopHead = 'while (left < right) {';
            // WebGL1 doesn't accept non constant loop conditions, so upper bound loop
            // iterations.
            const webGL1LoopHead = `for (int i = 0; i < ${Math.ceil(Math.log2(numInputs + 1))}; ++i) { if (left >= right) break;`;
            const loopHead = env$1().getNumber('WEBGL_VERSION') === 2 ? webGL2LoopHead :
                webGL1LoopHead;
            // left corresponds to lower bound and right to upper bound.
            const boundComparator = side === 'left' ? '<' : '<=';
            this.userCode = `
       int findBound(int batch, float value) {
         int left = 0;
         int right = numInputs;
         int mid;
         ${loopHead}
           mid = (left + right) / 2;
           if (getSortedSequence(batch, mid) ${boundComparator} value) {
             left = mid + 1;
           } else {
             right = mid;
           }
         }
         return right;
       }

       void main() {
         ivec2 coords = getOutputCoords();
         int batch = coords[0];
         int valueIndex = coords[1];

         float value = getValues(batch, valueIndex);

         setOutput(float(findBound(batch, value)));
       }
     `;
        }
    }

    /**
     * @license
     * Copyright 2022 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function searchSorted(args) {
        const { inputs, backend, attrs } = args;
        const { sortedSequence, values } = inputs;
        const { side } = attrs;
        const program = new SearchSortedProgram(sortedSequence.shape[0], sortedSequence.shape[1], values.shape[1], side);
        const customValues = [[sortedSequence.shape[1]]];
        return backend.runWebGLProgram(program, [sortedSequence, values], 'int32', customValues);
    }
    const searchSortedConfig = {
        kernelName: SearchSorted,
        backendName: 'webgl',
        kernelFunc: searchSorted,
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class SelectProgram {
        constructor(cRank, shape, rank) {
            this.variableNames = ['c', 'a', 'b'];
            this.outputShape = shape;
            let cCoords;
            let abCoords;
            if (rank > 4) {
                throw Error(`Where for rank ${rank} is not yet supported`);
            }
            if (rank === 1) {
                abCoords = `resRC`;
                cCoords = `resRC`;
            }
            else {
                const currentCoords = ['resRC.x', 'resRC.y', 'resRC.z', 'resRC.w'];
                const cCoordVars = [];
                const abCoordVars = [];
                for (let i = 0; i < shape.length; i++) {
                    abCoordVars.push(`${currentCoords[i]}`);
                    if (i < cRank) {
                        cCoordVars.push(`${currentCoords[i]}`);
                    }
                }
                cCoords = cCoordVars.join();
                abCoords = abCoordVars.join();
            }
            const dtype = getCoordsDataType(rank);
            this.userCode = `
      void main() {
        ${dtype} resRC = getOutputCoords();
        float cVal = getC(${cCoords});
        if (cVal >= 1.0) {
          setOutput(getA(${abCoords}));
        } else {
          setOutput(getB(${abCoords}));
        }
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function select(args) {
        const { inputs, backend } = args;
        const { condition, t, e } = inputs;
        const program = new SelectProgram(condition.shape.length, t.shape, t.shape.length);
        return backend.runWebGLProgram(program, [condition, t, e], upcastType(t.dtype, e.dtype));
    }
    const selectConfig = {
        kernelName: Select,
        backendName: 'webgl',
        kernelFunc: select
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SELU = `
  // Stable and Attracting Fixed Point (0, 1) for Normalized Weights.
  // see: https://arxiv.org/abs/1706.02515
  float scaleAlpha = ${SELU_SCALEALPHA};
  float scale = ${SELU_SCALE};
  return (x >= 0.0) ? scale * x : scaleAlpha * (exp(x) - 1.0);
`;
    const selu = unaryKernelFunc({ opSnippet: SELU });
    const seluConfig = {
        kernelName: Selu$1,
        backendName: 'webgl',
        kernelFunc: selu,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SIGMOID = CHECK_NAN_SNIPPET_UNARY + `
  return 1.0 / (1.0 + exp(-1.0 * x));
`;
    const SIGMOID_PACKED = `
  vec4 result = 1.0 / (1.0 + exp(-1.0 * x));
  bvec4 isNaN = isnan(x);

  result.r = isNaN.r ? x.r : result.r;
  result.g = isNaN.g ? x.g : result.g;
  result.b = isNaN.b ? x.b : result.b;
  result.a = isNaN.a ? x.a : result.a;

  return result;
`;
    const sigmoid = unaryKernelFunc({
        opSnippet: SIGMOID,
        packedOpSnippet: SIGMOID_PACKED,
        cpuKernelImpl: sigmoidImplCPU
    });
    const sigmoidConfig = {
        kernelName: Sigmoid$1,
        backendName: 'webgl',
        kernelFunc: sigmoid,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // Sign does not propagate NANs.
    const SIGN = `
  if (isnan(x)) { return 0.0; }
  return sign(x);
`;
    const sign = unaryKernelFunc({ opSnippet: SIGN });
    const signConfig = {
        kernelName: Sign,
        backendName: 'webgl',
        kernelFunc: sign,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SIN = CHECK_NAN_SNIPPET_UNARY + `
  return sin(x);
`;
    const sin = unaryKernelFunc({ opSnippet: SIN });
    const sinConfig = {
        kernelName: Sin,
        backendName: 'webgl',
        kernelFunc: sin,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SINH = `
  float e2x = exp(x);
  return (e2x - 1.0 / e2x) / 2.0;
`;
    const sinh = unaryKernelFunc({ opSnippet: SINH });
    const sinhConfig = {
        kernelName: Sinh,
        backendName: 'webgl',
        kernelFunc: sinh,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SOFTPLUS = `
  float epsilon = 1.1920928955078125e-7;
  float threshold = log(epsilon) + 2.0;

  bool too_large = x > -threshold;
  bool too_small = x < threshold;

  float result;
  float exp_x = exp(x);

  if (too_large){
    result = x;
  }
  else if (too_small){
    result = exp_x;
  }
  else{
    result = log(exp_x + 1.0);
  }
  return result;
`;
    const softplus = unaryKernelFunc({ opSnippet: SOFTPLUS });
    const softplusConfig = {
        kernelName: Softplus$1,
        backendName: 'webgl',
        kernelFunc: softplus,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const spaceToBatchND = (args) => {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { blockShape, paddings } = attrs;
        assert$4(x.shape.length <= 4, () => 'spaceToBatchND for rank > 4 with a WebGL backend not ' +
            'implemented yet');
        const prod = blockShape.reduce((a, b) => a * b);
        const completePaddings = [[0, 0]];
        completePaddings.push(...paddings);
        for (let i = 1 + blockShape.length; i < x.shape.length; ++i) {
            completePaddings.push([0, 0]);
        }
        const toDispose = [];
        const paddedX = padV2({
            inputs: { x },
            backend,
            attrs: { paddings: completePaddings, constantValue: 0 }
        });
        const reshapedPaddedShape = getReshaped(paddedX.shape, blockShape, prod, false);
        const permutedReshapedPaddedPermutation = getPermuted(reshapedPaddedShape.length, blockShape.length, false);
        const flattenShape = getReshapedPermuted(paddedX.shape, blockShape, prod, false);
        const reshapedPaddedX = reshape({ inputs: { x: paddedX }, backend, attrs: { shape: reshapedPaddedShape } });
        const paddedXT = transpose({
            inputs: { x: reshapedPaddedX },
            backend,
            attrs: { perm: permutedReshapedPaddedPermutation }
        });
        const result = reshape({ inputs: { x: paddedXT }, backend, attrs: { shape: flattenShape } });
        toDispose.push(paddedX);
        toDispose.push(reshapedPaddedX);
        toDispose.push(paddedXT);
        toDispose.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return result;
    };
    const spaceToBatchNDConfig = {
        kernelName: SpaceToBatchND,
        backendName: 'webgl',
        kernelFunc: spaceToBatchND
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseFillEmptyRows(args) {
        const { inputs, backend } = args;
        const { indices, values, denseShape, defaultValue } = inputs;
        if (denseShape.shape.length !== 1) {
            throw new Error(`Dense shape must be a vector, saw:
         ${denseShape.shape}`);
        }
        if (indices.shape.length !== 2) {
            throw new Error(`Indices must be a matrix, saw:
         ${indices.shape}`);
        }
        if (values.shape.length !== 1) {
            throw new Error(`Values must be a vector, saw:
         ${values.shape}`);
        }
        if (defaultValue.shape.length !== 0) {
            throw new Error(`Default value must be a scalar, saw:
        ${defaultValue.shape}`);
        }
        const $indices = backend.readSync(indices.dataId);
        const $values = backend.readSync(values.dataId);
        const $denseShape = backend.readSync(denseShape.dataId);
        const $defaultValue = backend.readSync(defaultValue.dataId)[0];
        const [outputIndices, outputIndicesShape, outputValues, emptyRowIndicator, reverseIndexMap] = sparseFillEmptyRowsImplCPU($indices, indices.shape, indices.dtype, $values, values.dtype, $denseShape, $defaultValue);
        return [
            backend.makeTensorInfo(outputIndicesShape, indices.dtype, outputIndices),
            backend.makeTensorInfo([outputIndicesShape[0]], values.dtype, outputValues),
            backend.makeTensorInfo([emptyRowIndicator.length], 'bool', new Uint8Array(emptyRowIndicator.map((value) => Number(value)))),
            backend.makeTensorInfo([reverseIndexMap.length], indices.dtype, new Int32Array(reverseIndexMap)),
        ];
    }
    const sparseFillEmptyRowsConfig = {
        kernelName: SparseFillEmptyRows,
        backendName: 'webgl',
        kernelFunc: sparseFillEmptyRows,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseReshape(args) {
        const { inputs, backend } = args;
        const { inputIndices, inputShape, newShape } = inputs;
        if (inputIndices.shape.length !== 2) {
            throw new Error(`Input indices should be a matrix but received shape ${inputIndices.shape}`);
        }
        if (inputShape.shape.length !== 1) {
            throw new Error(`Input shape should be a vector but received shape ${inputShape.shape}`);
        }
        if (newShape.shape.length !== 1) {
            throw new Error(`Target shape should be a vector but received shape ${newShape.shape}`);
        }
        const $inputShape = Array.from(backend.readSync(inputShape.dataId));
        const $inputIndices = backend.readSync(inputIndices.dataId);
        const targetShape = Array.from(backend.readSync(newShape.dataId));
        const [newIndices, indicesShape, outputShape] = sparseReshapeImplCPU($inputIndices, inputIndices.shape, inputIndices.dtype, $inputShape, targetShape);
        return [
            backend.makeTensorInfo(indicesShape, inputIndices.dtype, newIndices),
            backend.makeTensorInfo([outputShape.length], newShape.dtype, new Int32Array(outputShape)),
        ];
    }
    const sparseReshapeConfig = {
        kernelName: SparseReshape,
        backendName: 'webgl',
        kernelFunc: sparseReshape,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseSegmentMean(args) {
        const { inputs, backend } = args;
        const { data, indices, segmentIds } = inputs;
        if (data.shape.length < 1) {
            throw new Error(`Data should be at least 1 dimensional but received scalar`);
        }
        if (indices.shape.length !== 1) {
            throw new Error(`Indices should be a vector but received shape
              ${indices.shape}`);
        }
        if (segmentIds.shape.length !== 1) {
            throw new Error(`Segment ids should be a vector but received shape
              ${segmentIds.shape}`);
        }
        const $data = backend.readSync(data.dataId);
        const $indices = backend.readSync(indices.dataId);
        const $segmentIds = backend.readSync(segmentIds.dataId);
        const [outputData, outputDataShape] = sparseSegmentReductionImplCPU($data, data.shape, data.dtype, $indices, $segmentIds, true);
        return backend.makeTensorInfo(outputDataShape, data.dtype, outputData);
    }
    const sparseSegmentMeanConfig = {
        kernelName: SparseSegmentMean,
        backendName: 'webgl',
        kernelFunc: sparseSegmentMean,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseSegmentSum(args) {
        const { inputs, backend } = args;
        const { data, indices, segmentIds } = inputs;
        if (data.shape.length < 1) {
            throw new Error(`Data should be at least 1 dimensional but received scalar`);
        }
        if (indices.shape.length !== 1) {
            throw new Error(`Indices should be a vector but received shape
             ${indices.shape}`);
        }
        if (segmentIds.shape.length !== 1) {
            throw new Error(`Segment ids should be a vector but received shape
             ${segmentIds.shape}`);
        }
        const $data = backend.readSync(data.dataId);
        const $indices = backend.readSync(indices.dataId);
        const $segmentIds = backend.readSync(segmentIds.dataId);
        const [outputData, outputDataShape] = sparseSegmentReductionImplCPU($data, data.shape, data.dtype, $indices, $segmentIds);
        return backend.makeTensorInfo(outputDataShape, data.dtype, outputData);
    }
    const sparseSegmentSumConfig = {
        kernelName: SparseSegmentSum,
        backendName: 'webgl',
        kernelFunc: sparseSegmentSum,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function sparseToDense(args) {
        const { inputs, backend, attrs } = args;
        const { sparseIndices, sparseValues, defaultValue } = inputs;
        const { outputShape } = attrs;
        const { sliceRank, numUpdates, sliceSize, strides, outputSize } = calculateShapes(sparseValues, sparseIndices, outputShape);
        const sumDupeIndices = false;
        if (sparseValues.dtype === 'string') {
            const indicesBuf = backend.bufferSync(sparseIndices);
            const updatesBuf = backend.bufferSync(sparseValues);
            const $defaultValue = decodeString(backend.readSync(defaultValue.dataId)[0]);
            const outBuf = scatterImplCPU(indicesBuf, updatesBuf, outputShape, outputSize, sliceSize, numUpdates, sliceRank, strides, $defaultValue, sumDupeIndices);
            return backend.makeTensorInfo(outputShape, outBuf.dtype, outBuf.values);
        }
        const program = new ScatterProgram(numUpdates, sliceRank, sparseIndices.shape.length, sparseValues.shape.length, strides, [outputSize, 1], sumDupeIndices);
        const res = backend.runWebGLProgram(program, [sparseValues, sparseIndices, defaultValue], sparseValues.dtype);
        const reshaped = reshape({ inputs: { x: res }, backend, attrs: { shape: outputShape } });
        backend.disposeIntermediateTensorInfo(res);
        return reshaped;
    }
    const sparseToDenseConfig = {
        kernelName: SparseToDense,
        backendName: 'webgl',
        kernelFunc: sparseToDense
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function splitV(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { numOrSizeSplits, axis } = attrs;
        const $axis = parseAxisParam(axis, x.shape)[0];
        const splitSizes = prepareSplitSize(x, numOrSizeSplits, $axis);
        const xRank = x.shape.length;
        const begin = new Array(xRank).fill(0);
        const size = x.shape.slice();
        return splitSizes.map(s => {
            const sliceSize = [...size];
            sliceSize[$axis] = s;
            const sliceT = slice({ inputs: { x }, backend, attrs: { begin, size: sliceSize } });
            begin[$axis] += s;
            return sliceT;
        });
    }
    const splitVConfig = {
        kernelName: SplitV,
        backendName: 'webgl',
        kernelFunc: splitV
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SQRT = `return sqrt(x);`;
    const sqrt = unaryKernelFunc({ opSnippet: SQRT, packedOpSnippet: SQRT, cpuKernelImpl: sqrtImplCPU });
    const sqrtConfig = {
        kernelName: Sqrt,
        backendName: 'webgl',
        kernelFunc: sqrt
    };

    /**
     * @license
     * Copyright 2019 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SQUARE = `return x * x;`;
    const square = unaryKernelFunc({ opSnippet: SQUARE });
    const squareConfig = {
        kernelName: Square,
        backendName: 'webgl',
        kernelFunc: square,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const SQUARED_DIFFERENCE = 'return (a - b) * (a - b);';
    const squaredDifference = binaryKernelFunc({ opSnippet: SQUARED_DIFFERENCE, packedOpSnippet: SQUARED_DIFFERENCE });
    const squaredDifferenceConfig = {
        kernelName: SquaredDifference,
        backendName: 'webgl',
        kernelFunc: squaredDifference,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function step({ inputs, attrs, backend }) {
        const { x } = inputs;
        const opSnippet = CHECK_NAN_SNIPPET$2 + `
    return x > 0.0 ? 1.0 : float(${attrs.alpha});
  `;
        const program = new UnaryOpProgram(x.shape, opSnippet);
        return backend.runWebGLProgram(program, [x], x.dtype);
    }
    const stepConfig = {
        kernelName: Step,
        backendName: 'webgl',
        kernelFunc: step,
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class StridedSliceProgram {
        constructor(begin, strides, size) {
            this.variableNames = ['x'];
            this.outputShape = size;
            const rank = size.length;
            const inputDtype = getCoordsDataType(size.length);
            const dtype = getCoordsDataType(size.length);
            let newCoords = '';
            if (rank === 1) {
                newCoords = 'coords * strides + begin';
            }
            else {
                let outputAxis = 0;
                newCoords =
                    size.map((_, i) => {
                        outputAxis++;
                        return size.length === 1 ?
                            `coords * strides[${i}] + begin[${i}]` :
                            `coords[${outputAxis - 1}] * strides[${i}] + begin[${i}]`;
                    })
                        .join(',');
            }
            this.userCode = `
      ${inputDtype} begin = ${inputDtype}(${begin});
      ${inputDtype} strides = ${inputDtype}(${strides});

      void main() {
        ${dtype} coords = getOutputCoords();
        setOutput(getX(${newCoords}));
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stridedSlice(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { begin, end, strides, beginMask, endMask, ellipsisMask, newAxisMask, shrinkAxisMask } = attrs;
        const { finalShapeSparse, finalShape, isIdentity, sliceDim0, isSimpleSlice, begin: $begin, end: $end, strides: $strides } = sliceInfo(x.shape, begin, end, strides, beginMask, endMask, ellipsisMask, newAxisMask, shrinkAxisMask);
        let result;
        if (isIdentity) {
            // Optimization #1, slice is a no-op plus reshape
            result = reshape({ inputs: { x }, backend, attrs: { shape: finalShape } });
        }
        else if (sliceDim0 || isSimpleSlice) {
            // Optimization #2, slice is memory contiguous (only occurs in dim 0)
            assert$4(x.shape.length >= 1, () => `Input must have rank at least 1, got: ${x.shape.length}`);
            const size = computeOutShape$2($begin, $end, $strides);
            // To tolerate begin[0] > end[0] (a 0-output slice), we min(begin, end).
            const sliced = slice({ inputs: { x }, backend, attrs: { begin: $begin, size } });
            result =
                reshape({ inputs: { x: sliced }, backend, attrs: { shape: finalShape } });
            backend.disposeIntermediateTensorInfo(sliced);
        }
        else {
            const shouldExecuteOnCPU = backend.shouldExecuteOnCPU([x]);
            if (shouldExecuteOnCPU) {
                // tslint:disable-next-line: no-unnecessary-type-assertion
                const values = backend.readSync(x.dataId);
                // tslint:disable-next-line: no-unnecessary-type-assertion
                const xBuf = buffer(x.shape, x.dtype, values);
                const resultValues = stridedSliceImplCPU(finalShapeSparse, xBuf, $strides, $begin);
                result = backend.makeTensorInfo(finalShape, x.dtype, resultValues.values);
            }
            else {
                const program = new StridedSliceProgram($begin, $strides, finalShapeSparse);
                result = backend.runWebGLProgram(program, [x], x.dtype);
            }
        }
        const resultReshaped = reshape({ inputs: { x: result }, backend, attrs: { shape: finalShape } });
        backend.disposeIntermediateTensorInfo(result);
        return resultReshaped;
    }
    const stridedSliceConfig = {
        kernelName: StridedSlice,
        backendName: 'webgl',
        kernelFunc: stridedSlice
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stringNGrams(args) {
        const { inputs, backend, attrs } = args;
        const { separator, nGramWidths, leftPad, rightPad, padWidth, preserveShortSequences } = attrs;
        const { data, dataSplits } = inputs;
        const $data = backend.readSync(data.dataId);
        const $dataSplits = backend.readSync(dataSplits.dataId);
        const [nGrams, nGramsSplits] = stringNGramsImplCPU($data, $dataSplits, separator, nGramWidths, leftPad, rightPad, padWidth, preserveShortSequences);
        return [
            backend.makeTensorInfo([nGrams.length], 'string', nGrams),
            backend.makeTensorInfo(dataSplits.shape, 'int32', nGramsSplits),
        ];
    }
    const stringNGramsConfig = {
        kernelName: StringNGrams,
        backendName: 'webgl',
        kernelFunc: stringNGrams,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stringSplit(args) {
        const { inputs, backend, attrs } = args;
        const { skipEmpty } = attrs;
        const { input, delimiter } = inputs;
        if (input.dtype !== 'string') {
            throw new Error('Input must be of datatype string');
        }
        if (input.shape.length !== 1) {
            throw new Error(`Input must be a vector, got shape: ${input.shape}`);
        }
        if (delimiter.shape.length !== 0) {
            throw new Error(`Delimiter must be a scalar, got shape: ${delimiter.shape}`);
        }
        const $input = backend.readSync(input.dataId);
        const $delimiter = backend.readSync(delimiter.dataId)[0];
        const [indices, values, shape] = stringSplitImplCPU($input, $delimiter, skipEmpty);
        const outputSize = values.length;
        return [
            backend.makeTensorInfo([outputSize, 2], 'int32', indices),
            backend.makeTensorInfo([outputSize], 'string', values),
            backend.makeTensorInfo([2], 'int32', new Int32Array(shape))
        ];
    }
    const stringSplitConfig = {
        kernelName: StringSplit,
        backendName: 'webgl',
        kernelFunc: stringSplit,
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function stringToHashBucketFast(args) {
        const { inputs, backend, attrs } = args;
        const { numBuckets } = attrs;
        const { input } = inputs;
        if (input.dtype !== 'string') {
            throw new Error('Input must be of datatype string');
        }
        if (numBuckets <= 0) {
            throw new Error(`Number of buckets must be at least 1`);
        }
        const $input = backend.readSync(input.dataId);
        const output = stringToHashBucketFastImplCPU($input, numBuckets);
        return backend.makeTensorInfo(input.shape, 'int32', output);
    }
    const stringToHashBucketFastConfig = {
        kernelName: StringToHashBucketFast,
        backendName: 'webgl',
        kernelFunc: stringToHashBucketFast,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const TAN = `return tan(x);`;
    const tan = unaryKernelFunc({ opSnippet: TAN });
    const tanConfig = {
        kernelName: Tan,
        backendName: 'webgl',
        kernelFunc: tan,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const TANH = `
  float e2x = exp(-2.0 * abs(x));
  return sign(x) * (1.0 - e2x) / (1.0 + e2x);
`;
    const tanh = unaryKernelFunc({ opSnippet: TANH });
    const tanhConfig = {
        kernelName: Tanh$1,
        backendName: 'webgl',
        kernelFunc: tanh,
    };

    /**
     * @license
     * Copyright 2017 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class TileProgram {
        constructor(aShape, reps) {
            this.variableNames = ['A'];
            const outputShape = new Array(aShape.length);
            for (let i = 0; i < outputShape.length; i++) {
                outputShape[i] = aShape[i] * reps[i];
            }
            this.outputShape = outputShape;
            this.rank = outputShape.length;
            const dtype = getCoordsDataType(this.rank);
            const sourceCoords = getSourceCoords(aShape);
            this.userCode = `
      void main() {
        ${dtype} resRC = getOutputCoords();
        setOutput(getA(${sourceCoords}));
      }
    `;
        }
    }
    function getSourceCoords(aShape) {
        const rank = aShape.length;
        if (rank > 5) {
            throw Error(`Tile for rank ${rank} is not yet supported`);
        }
        if (rank === 1) {
            return `imod(resRC, ${aShape[0]})`;
        }
        const currentCoords = ['resRC.x', 'resRC.y', 'resRC.z', 'resRC.w', 'resRC.u'];
        const sourceCoords = [];
        for (let i = 0; i < aShape.length; i++) {
            sourceCoords.push(`imod(${currentCoords[i]}, ${aShape[i]})`);
        }
        return sourceCoords.join();
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function tile(params) {
        const { inputs, backend, attrs } = params;
        const { x } = inputs;
        const { reps } = attrs;
        // tile gpu program cannot handle rank > 5 case.
        if (x.dtype === 'string' || x.shape.length > 5) {
            // Even thought string tensor is always on CPU, just to be consistent on how
            // to access tensor data.
            const data = backend.readSync(x.dataId);
            const value = x.dtype === 'string' ?
                data.map(d => decodeString(d)) :
                data;
            const buf = buffer(x.shape, x.dtype, value);
            const outBuf = tileImplCPU(buf, reps);
            return backend.makeTensorInfo(outBuf.shape, outBuf.dtype, outBuf.values);
        }
        const program = new TileProgram(x.shape, reps);
        const output = backend.runWebGLProgram(program, [x], x.dtype);
        return output;
    }
    const tileConfig = {
        kernelName: Tile,
        backendName: 'webgl',
        kernelFunc: tile,
    };

    // Based on Algorithm 2 of Bitonic Top K, ref:
    // https://anilshanbhag.in/static/papers/gputopk_sigmod18.pdf
    // The original algorithm is based on computing the top K only, however
    // since for TFJS we require the indices of the top K values as well then the
    // algorithm found here is a bit modified. Rather than producing the values
    // at each step, the indices containing the top K are generated instead.
    // The output values are not generated to reduce the number of outputs in the
    // GPU, the values can easily be retrieved from the indices using a gather
    // op.
    class SwapProgram {
        /**
         * @param shape desired output shape (can be larger than input shape, output
         *                                    will be padded with -Infinity)
         */
        constructor(shape) {
            this.variableNames = ['x', 'indices'];
            // |n| Size of the original input of TopK.
            // |firstPass|indicates if this is the first time swap is being used which
            // means no indices input containing the top K is present yet.
            // |inc| Swaps pairs of indices (0, inc), (1, inc + 1), (2, inc + 2) ...
            this.customUniforms = [
                { name: 'n', type: 'int' },
                { name: 'firstPass', type: 'int' },
                { name: 'negativeInf', type: 'float' },
                { name: 'dir', type: 'int' },
                { name: 'inc', type: 'int' }
            ];
            this.outputShape = shape;
            this.userCode = `
       void main() {
         ivec2 coords = getOutputCoords();
         int batch = coords[0];
         int elemIdx = coords[1];

         // We compare elements pair-wise within a group of size 2 * inc.
         // The comparing rule for each group alternates between ascending
         // and descending. Within each group, we compare each pair at
         // positions i and i+inc. To decide whether an element at position i
         // is x0 or x1, we mod it by 2 * inc, if the result is smaller than
         // inc, it is in the first half of the group, we denote it as x0,
         // otherwise we denote it as x1.
         // For example, as shown in the Bitonic top K paper referenced above,
         // Figure5(a) shows that element[1] is in the
         // second half of the group when group size is 2, but it is in the
         // first half of the group when group size is 4.

         bool isFirstInPair = imod(elemIdx, 2 * inc) < inc;
         int i = isFirstInPair ? elemIdx : elemIdx - inc;

         int i0 = firstPass == 1 ? i : int(getIndices(batch, i));
         int i1 = firstPass == 1 ? i + inc : int(getIndices(batch, i + inc));
         float x0 = i0 < n ? getX(batch, i0) : negativeInf;
         float x1 = i1 < n ? getX(batch, i1) : negativeInf;

         // Denotes which direction indices are in (ascending or descending).
         bool reverse = imod(elemIdx, 2 * dir) >= dir;
         bool isGreater = x0 > x1 || (x0 == x1 && i1 > i0);
         if (reverse == isGreater) { // Elements in opposite order of direction
           int iTemp = i0;
           i0 = i1;
           i1 = iTemp;
         }
         if (isFirstInPair) {
            setOutput(float(i0));
         } else {
            setOutput(float(i1));
         }
       }
     `;
        }
    }
    class MergeProgram {
        /**
         * @param shape desired output shape (must be half of the input size)
         */
        constructor(shape) {
            this.variableNames = ['x', 'indices'];
            // |n| Size of the original input of TopK
            // |firstPass| indicates if this is the first time swap is being used which
            // means no indices input containing the top K is present yet.
            // |k| Top k elements desired
            this.customUniforms = [
                { name: 'n', type: 'int' },
                { name: 'firstPass', type: 'int' },
                { name: 'k', type: 'int' }
            ];
            this.outputShape = shape;
            this.userCode = `
    void main() {
         // Takes max of indices (0, k), (1, k + 1), (2, k + 2) ...
         ivec2 coords = getOutputCoords();
         int batch = coords[0];
         int elemIdx = coords[1];

         // The output size is half of the previous size.
         // If the previous sequence is | | | | _ _ _ _  | | | |  _ _ _ _ (k=4),
         // we only need to output the indices at positions |, the indices at
         // positions _ can be thrown away, see Figure5(b) After Phase 2
         // (Merge phase) in the Bitonic Top K paper referenced above.
         // For example, the paper shows we only need to output the orange bars.
         // The output sequence should look like this | | | | | | | |.
         // Because the sequence is halved, to map the output index back
         // to the previous sequence to find the corresponding value,
         // we need to double the index. When we double the index,
         // we basically interpolate a position, so 2i looks like
         // | _ | _ | _ | _ | _ | _ | _. We move the | to the first k position
         // of each 2k positions by - elemIdx % k. E.g. for output at
         // index 4,5,6,7, we want to get the corresponding element at
         // original index 8,9,10,11, for output at index 8,9,10,11,
         // we want to get the corresponding element at original index
         // 16,17,18,19, so on and so forth.

         int i = elemIdx < k ? elemIdx : (elemIdx * 2 - imod(elemIdx, k));
         int i0 = firstPass == 1 ? i : int(getIndices(batch, i));
         int i1 = firstPass == 1 ? i + k : int(getIndices(batch, i + k));

         float x0 = getX(batch, i0);
         float x1 = i1 < n ? getX(batch, i1) : x0;

         setOutput(x0 >= x1 ? float(i0) : float(i1));
       }
     `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function disposeIntermediateTensorInfoOrNull(backend, tensorInfo) {
        if (tensorInfo !== null) {
            backend.disposeIntermediateTensorInfo(tensorInfo);
        }
    }
    function roundUpToPow2(num) {
        let pow2 = 1;
        while (pow2 < num) {
            pow2 *= 2;
        }
        return pow2;
    }
    // Based on Algorithm 2 of Bitonic Top K, ref:
    // https://anilshanbhag.in/static/papers/gputopk_sigmod18.pdf
    function topK(args) {
        const { inputs, backend, attrs } = args;
        const { x } = inputs;
        const { k, sorted } = attrs;
        // Empirically determined constant used to determine last dim threshold for
        // handing off execution to the CPU.
        const TOPK_LAST_DIM_CPU_HANDOFF_SIZE_THRESHOLD = env$1().getNumber('TOPK_LAST_DIM_CPU_HANDOFF_SIZE_THRESHOLD');
        // Empirically determined constant used to determine k threshold for handing
        // off execution to the CPU.
        const TOPK_K_CPU_HANDOFF_THRESHOLD = env$1().getNumber('TOPK_K_CPU_HANDOFF_THRESHOLD');
        const xShape = x.shape;
        const lastDim = xShape[xShape.length - 1];
        if (backend.shouldExecuteOnCPU([x]) ||
            lastDim < TOPK_LAST_DIM_CPU_HANDOFF_SIZE_THRESHOLD ||
            k > TOPK_K_CPU_HANDOFF_THRESHOLD) {
            const xVals = backend.readSync(x.dataId);
            const [allTopKVals, allTopKIndices] = topKImplCPU(xVals, xShape, x.dtype, k, sorted);
            return [
                backend.makeTensorInfo(allTopKVals.shape, allTopKVals.dtype, allTopKVals.values),
                backend.makeTensorInfo(allTopKIndices.shape, allTopKIndices.dtype, allTopKIndices.values)
            ];
        }
        if (k === 0) {
            xShape[xShape.length - 1] = 0;
            return [
                backend.makeTensorInfo(xShape, x.dtype, []),
                backend.makeTensorInfo(xShape, 'int32', [])
            ];
        }
        if (lastDim === 1 /* firstPass */) {
            return [
                x, fill({ attrs: { shape: xShape, dtype: 'int32', value: 0 }, backend })
            ];
        }
        // Eagerly unpack x input since it is passed in to all the shaders which
        // require unpacked inputs.
        const xtexData = backend.texData.get(x.dataId);
        const xIsPacked = xtexData !== null && xtexData.isPacked;
        const xUnPacked = xIsPacked ? backend.unpackTensor(x) : x;
        // Reshape into a 2d tensor [batch, lastDim] and compute topk along lastDim.
        const xSize = sizeFromShape(xShape);
        const batch = xSize / lastDim;
        const x2D = reshape({ inputs: { x: xUnPacked }, attrs: { shape: [batch, lastDim] }, backend });
        if (xIsPacked) {
            disposeIntermediateTensorInfoOrNull(backend, xUnPacked);
        }
        const kPow2 = roundUpToPow2(k);
        const lastDimPow2 = roundUpToPow2(lastDim);
        // Only the indices containing the top K are kept at every step to reduce
        // number of outputs in the GPU algorithms, so once the final set of indices
        // is computed then gather is used to grab the corresponding values
        // from the original input.
        let indices = null;
        // GPU algorithm always takes in an indices input but this input is not used
        // on the first run of a GPU algorithm, therefore if indices is null we simply
        // pass in x2D instead of it but the value will not actually be used
        const getInputs = () => indices === null ? [x2D, x2D] : [x2D, indices];
        const runSwap = (dir, inc, shape) => {
            const inputs = getInputs();
            const program = new SwapProgram(shape);
            const fistPass = indices === null ? 1 : 0;
            const customValues = [[lastDim], [fistPass], [Number.NEGATIVE_INFINITY], [dir], [inc]];
            const prevIndices = indices;
            indices = backend.runWebGLProgram(program, inputs, 'int32', customValues);
            disposeIntermediateTensorInfoOrNull(backend, prevIndices);
        };
        // Step 1: local sort
        for (let len = 1; len < kPow2; len *= 2) {
            const dir = len * 2;
            for (let inc = len; inc >= 1; inc /= 2) {
                runSwap(dir, inc, [batch, lastDimPow2]);
            }
        }
        // Step 2: merge
        for (let indicesSize = lastDimPow2; indicesSize > kPow2; indicesSize /= 2) {
            const inputs = getInputs();
            const mergeProgram = new MergeProgram([batch, indicesSize / 2]);
            const firstPass = indices === null ? 1 : 0;
            const customValues = [[lastDim], [firstPass], [kPow2]];
            const prevIndices = indices;
            indices =
                backend.runWebGLProgram(mergeProgram, inputs, 'int32', customValues);
            disposeIntermediateTensorInfoOrNull(backend, prevIndices);
            // Step 3: rebuild
            const len = kPow2 / 2;
            const dir = len * 2;
            for (let inc = len; inc >= 1; inc /= 2) {
                runSwap(dir, inc, indices.shape);
            }
        }
        // Keep only the requested top K results instead of kPow2
        let prevIndices = indices;
        indices = slice({ inputs: { x: indices }, backend, attrs: { begin: 0, size: [batch, k] } });
        disposeIntermediateTensorInfoOrNull(backend, prevIndices);
        // Gather values on last dimension
        let values = gatherV2({ inputs: { x: x2D, indices }, backend, attrs: { axis: 1, batchDims: 1 } });
        disposeIntermediateTensorInfoOrNull(backend, x2D);
        // Reshape back to the original input shape, except that the last
        // dimension is k.
        const newShape = xShape.slice(0, -1);
        newShape.push(k);
        prevIndices = indices;
        indices = reshape({ inputs: { x: indices }, attrs: { shape: newShape }, backend });
        disposeIntermediateTensorInfoOrNull(backend, prevIndices);
        const prevValues = values;
        values = reshape({ inputs: { x: values }, attrs: { shape: newShape }, backend });
        disposeIntermediateTensorInfoOrNull(backend, prevValues);
        return [values, indices];
    }
    const topKConfig = {
        kernelName: TopK,
        backendName: 'webgl',
        kernelFunc: topK
    };

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class TransformProgram {
        constructor(imageHeight, imageWidth, interpolation, fillMode, fillValue, outShape) {
            this.variableNames = ['Image', 'Transforms'];
            this.outputShape = outShape;
            const interpolationModeId = interpolation === 'nearest' ? 1 : 2;
            let fillModeId;
            switch (fillMode) {
                case 'constant':
                    fillModeId = 1;
                    break;
                case 'reflect':
                    fillModeId = 2;
                    break;
                case 'wrap':
                    fillModeId = 3;
                    break;
                case 'nearest':
                    fillModeId = 4;
                    break;
                default:
                    fillModeId = 1;
                    break;
            }
            this.userCode = `
            float mapCoord(float outCoord, float len) {
              float inCoord = outCoord;
              if(${fillModeId} == 2) {
                if (inCoord < 0.0) {
                  if (len <= 1.0) {
                    inCoord = 0.0;
                  } else {
                    float sz2 = 2.0 * len;
                    if (inCoord < sz2) {
                      inCoord = sz2 * float(int(float(-inCoord / sz2))) +
                      inCoord;
                    }
                    inCoord = inCoord < -len ? inCoord + sz2 : -inCoord - 1.0;
                  }
                } else if (inCoord > len - 1.0) {
                  if (len <= 1.0) {
                    inCoord = 0.0;
                  } else {
                    float sz2 = 2.0 * len;
                    inCoord -= sz2 * float(int(float(inCoord / sz2)));
                    if (inCoord >= len) {
                      inCoord = sz2 - inCoord - 1.0;
                    }
                  }
                }
                return clamp(inCoord, 0.0, len - 1.0);
              } else if (${fillModeId} == 3) {
                if (inCoord < 0.0) {
                  if (len <= 1.0) {
                    inCoord = 0.0;
                  } else {
                    float sz = len - 1.0;
                    inCoord += len * (float(int(float(-inCoord / sz))) + 1.0);
                  }
                } else if (inCoord > len - 1.0) {
                  if (len <= 1.0) {
                    inCoord = 0.0;
                  } else {
                    float sz = len - 1.0;
                    inCoord -= len * float(int(float(inCoord / sz)));
                  }
                }
                return clamp(inCoord, 0.0, len - 1.0);
              } else if (${fillModeId} == 4) {
                return clamp(outCoord, 0.0, len - 1.0);
              } else {
                return outCoord;
              }
            }

            float readWithFillValue(int batch, int coordY, int coordX,
              int channel) {
              float outputValue;
              if (0 <= coordY && coordY < ${imageHeight} && 0 <= coordX && coordX < ${imageWidth}) {
                  outputValue = getImage(batch, coordY, coordX, channel);
              } else {
                outputValue = float(${fillValue});
              }
              return outputValue;
            }

            void main() {
              ivec4 coords = getOutputCoords();
              float outputValue;
              int batch = coords[0];
              int x = coords[2];
              int y = coords[1];
              int channel = coords[3];
              float xf = float(x);
              float yf = float(y);
              float a1 = getTransforms(batch, 0);
              float a2 = getTransforms(batch, 1);
              float a3 = getTransforms(batch, 2);
              float b1 = getTransforms(batch, 3);
              float b2 = getTransforms(batch, 4);
              float b3 = getTransforms(batch, 5);
              float c1 = getTransforms(batch, 6);
              float c2 = getTransforms(batch, 7);
              float projection = c1 * xf + c2 * yf + 1.0;
              if (projection == 0.0) {
                outputValue = float(${fillValue});
              } else {
                float inX = (a1 * xf + a2 * yf + a3) / projection;
                float inY = (b1 * xf + b2 * yf + b3) / projection;
                float mapX = mapCoord(inX, float(${imageWidth}));
                float mapY = mapCoord(inY, float(${imageHeight}));

                if (${interpolationModeId} == 1) {
                  int coordY = int(round(mapY));
                  int coordX = int(round(mapX));
                  outputValue = readWithFillValue(batch, coordY, coordX,
                    channel);
                } else {
                  float yFloor = floor(mapY);
                  float xFloor = floor(mapX);
                  float yCeil = yFloor + 1.0;
                  float xCeil = xFloor + 1.0;
                  float valueYFloor = (xCeil - mapX) *
                  readWithFillValue(batch, int(yFloor), int(xFloor), channel) +
                  (mapX - xFloor) *
                  readWithFillValue(batch, int(yFloor), int(xCeil), channel);
                  float valueYCeil = (xCeil - mapX) *
                  readWithFillValue(batch, int(yCeil), int(xFloor), channel) +
                  (mapX - xFloor) *
                  readWithFillValue(batch, int(yCeil), int(xCeil), channel);
                  outputValue = (yCeil - mapY) * valueYFloor +
                  (mapY - yFloor) * valueYCeil;
                }
              }
              setOutput(outputValue);
            }
        `;
        }
    }

    /**
     * @license
     * Copyright 2021 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function transform(args) {
        const { inputs, backend, attrs } = args;
        const { image, transforms } = inputs;
        const { interpolation, fillMode, fillValue, outputShape } = attrs;
        const [batch, imageHeight, imageWidth, numChannels] = image.shape;
        const [outHeight, outWidth] = outputShape != null ? outputShape : [imageHeight, imageWidth];
        const outShape = [batch, outHeight, outWidth,
            numChannels];
        const program = new TransformProgram(imageHeight, imageWidth, interpolation, fillMode, fillValue, outShape);
        return backend.runWebGLProgram(program, [image, transforms], 'float32');
    }
    const transformConfig = {
        kernelName: Transform$1,
        backendName: 'webgl',
        kernelFunc: transform
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the License);
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an AS IS BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function unique(args) {
        const { inputs, attrs, backend } = args;
        const { axis } = attrs;
        const { x } = inputs;
        assertNotComplex(x, 'unique');
        // For now, always forward calculation to the CPU backend.
        console.warn('WARNING: ', 'UI might be locked temporarily as data is being downloaded');
        const values = backend.readSync(x.dataId);
        const { outputValues, outputShape, indices } = uniqueImplCPU(values, axis, x.shape, x.dtype);
        return [
            backend.makeTensorInfo(outputShape, x.dtype, outputValues),
            backend.makeTensorInfo([indices.length], 'int32', indices),
        ];
    }
    const uniqueConfig = {
        kernelName: Unique,
        backendName: 'webgl',
        kernelFunc: unique,
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function unpack(args) {
        const { inputs, backend, attrs } = args;
        const { value } = inputs;
        let { axis } = attrs;
        if (axis < 0) {
            axis += value.shape.length;
        }
        const x = value;
        const xRank = x.shape.length;
        const num = value.shape[axis];
        const outShape = new Array(xRank - 1);
        let outIndex = 0;
        for (let i = 0; i < xRank; i++) {
            if (i !== axis) {
                outShape[outIndex++] = x.shape[i];
            }
        }
        const toDispose = [];
        const begin = new Array(xRank).fill(0);
        const size = x.shape.slice();
        size[axis] = 1;
        const res = new Array(num);
        for (let i = 0; i < res.length; i++) {
            begin[axis] = i;
            const sliced = slice({ inputs: { x }, backend, attrs: { begin, size } });
            const reshaped = reshape({ inputs: { x: sliced }, backend, attrs: { shape: outShape } });
            res[i] = reshaped;
            toDispose.push(sliced);
        }
        toDispose.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return res;
    }
    const unpackConfig = {
        kernelName: Unpack,
        backendName: 'webgl',
        kernelFunc: unpack
    };

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    class SegmentOpProgram {
        constructor(segOpInfo, segOpType) {
            this.variableNames = ['x', 'segmentIds'];
            const windowSize = segOpInfo.windowSize;
            const batchSize = segOpInfo.batchSize;
            const inSize = segOpInfo.inSize;
            const numSegments = segOpInfo.numSegments;
            const outSize = numSegments * Math.ceil(inSize / windowSize);
            this.outputShape = [batchSize, outSize];
            const initializationValue = '0.0';
            const returnValue = `sumValue`;
            const windowSizeNearestVec4 = Math.floor(windowSize / 4) * 4;
            const windowSizeVec4Remainder = windowSize % 4;
            const updateSnippet = `
        sumValue += dot(values, segFilter);
    `;
            let checkValueOutOfBounds = '';
            if (inSize % windowSize > 0) {
                checkValueOutOfBounds = `
        if (inIdx < 0 || inIdx >= ${inSize}) {
          return initializationValue;
        }
      `;
            }
            let checkSegmentIdOutOfBounds = '';
            if (inSize % windowSize > 0) {
                checkSegmentIdOutOfBounds = `
        if (inIdx < 0 || inIdx >= ${inSize}) {
          return -1.0;
        }
      `;
            }
            this.userCode = `
      const float initializationValue = ${initializationValue};

      float getValue(int batch, int inIdx) {
        ${checkValueOutOfBounds}
        return getX(batch, inIdx);
      }

      float getSegmentIdAtIndex(int inIdx) {
        ${checkSegmentIdOutOfBounds}
        return getSegmentIds(inIdx);
      }

      void main() {
        ivec2 coords = getOutputCoords();
        int batch = coords[0];
        int outIdx = coords[1];
        int inOffset = int(floor(float(outIdx) / float(
          ${numSegments})) * float(${windowSize}));
        int currentSeg = int(mod(float(outIdx), float(${numSegments})));

        float sumValue = 0.0;

        for (int i = 0; i < ${windowSizeNearestVec4}; i += 4) {
          int inIdx = inOffset + i;
          vec4 values = vec4(
            getValue(batch, inIdx),
            getValue(batch, inIdx + 1),
            getValue(batch, inIdx + 2),
            getValue(batch, inIdx + 3)
          );

          vec4 segFilter = vec4(
            int(getSegmentIdAtIndex(inIdx)) == currentSeg ? 1 : 0,
            int(getSegmentIdAtIndex(inIdx + 1)) == currentSeg ? 1 : 0,
            int(getSegmentIdAtIndex(inIdx + 2)) == currentSeg ? 1 : 0,
            int(getSegmentIdAtIndex(inIdx + 3)) == currentSeg ? 1 : 0
          );

          ${updateSnippet}
        }

        int inIdx = inOffset + ${windowSizeNearestVec4};
        if (${windowSizeVec4Remainder === 1}) {
          vec4 values = vec4(
            getValue(batch, inIdx),
            initializationValue,
            initializationValue,
            initializationValue
          );

          int inIdxSeg = int(getSegmentIdAtIndex(inIdx));

          vec4 segFilter = vec4(
            int(getSegmentIdAtIndex(inIdx)) == currentSeg ? 1 : 0,
            0,
            0,
            0
          );

          ${updateSnippet}
        } else if (${windowSizeVec4Remainder === 2}) {
          vec4 values = vec4(
            getValue(batch, inIdx),
            getValue(batch, inIdx + 1),
            initializationValue,
            initializationValue
          );

          vec4 segFilter = vec4(
            int(getSegmentIdAtIndex(inIdx)) == currentSeg ? 1 : 0,
            int(getSegmentIdAtIndex(inIdx + 1)) == currentSeg ? 1 : 0,
              0,
              0
          );

          ${updateSnippet}
        } else if (${windowSizeVec4Remainder === 3}) {
          vec4 values = vec4(
            getValue(batch, inIdx),
            getValue(batch, inIdx + 1),
            getValue(batch, inIdx + 2),
            initializationValue
          );

          vec4 segFilter = vec4(
            int(getSegmentIdAtIndex(inIdx)) == currentSeg ? 1 : 0,
            int(getSegmentIdAtIndex(inIdx + 1)) == currentSeg ? 1 : 0,
            int(getSegmentIdAtIndex(inIdx + 2)) == currentSeg ? 1 : 0,
            0
          );

          ${updateSnippet}
        }
        setOutput(${returnValue});
      }
    `;
        }
    }

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    function unsortedSegmentSum(args) {
        const { inputs, backend, attrs } = args;
        const { x, segmentIds } = inputs;
        const { numSegments } = attrs;
        const xRank = x.shape.length;
        const toDispose = [];
        let axis = 0;
        const permutation = getAxesPermutation([axis], xRank);
        let permutedX = x;
        if (permutation != null) {
            permutedX = transpose({ inputs: { x }, backend, attrs: { perm: permutation } });
            toDispose.push(permutedX);
            axis = getInnerMostAxes(1, xRank)[0];
        }
        const outShape = computeOutShape(permutedX.shape, axis, numSegments);
        const inSize = sizeFromShape([permutedX.shape[axis]]);
        const a2D = reshape({ inputs: { x: permutedX }, backend, attrs: { shape: [-1, inSize] } });
        toDispose.push(a2D);
        const outputDType = sumOutType(x.dtype);
        const segOpCompute = (x, segOpType, segmentIds, dtype, numSegments) => {
            const batchSize = x.shape[0];
            const inSize = x.shape[1];
            const windowSize = segOpComputeOptimalWindowSize(inSize, numSegments);
            const segOpInfo = { windowSize, inSize, batchSize, numSegments };
            const program = new SegmentOpProgram(segOpInfo, segOpType);
            const output = backend.compileAndRun(program, [x, segmentIds], dtype);
            toDispose.push(output);
            // No need to run another GPGPU program.
            if (output.shape[1] === numSegments) {
                return output;
            }
            const rangeInfo = range$1({
                backend,
                attrs: { start: 0, stop: numSegments, step: 1, dtype: 'float32' }
            });
            const tileInfo = tile({
                inputs: { x: rangeInfo },
                backend,
                attrs: { reps: [inSize / windowSize] }
            });
            toDispose.push(rangeInfo);
            toDispose.push(tileInfo);
            const result = segOpCompute(output, segOpType, tileInfo, dtype, numSegments);
            return result;
        };
        const segOpResult = segOpCompute(a2D, 'unsortedSegmentSum', segmentIds, outputDType, numSegments);
        const reshaped = reshape({ inputs: { x: segOpResult }, backend, attrs: { shape: outShape } });
        let result = reshaped;
        if (permutation != null) {
            toDispose.push(reshaped);
            const perm = getUndoAxesPermutation(permutation);
            result = transpose({ inputs: { x: result }, backend, attrs: { perm } });
        }
        toDispose.forEach(t => backend.disposeIntermediateTensorInfo(t));
        return result;
    }
    const unsortedSegmentSumConfig = {
        kernelName: UnsortedSegmentSum,
        backendName: 'webgl',
        kernelFunc: unsortedSegmentSum
    };

    /**
     * @license
     * Copyright 2020 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    // List all kernel configs here
    const kernelConfigs = [
        _fusedMatMulConfig,
        absConfig,
        acosConfig,
        acoshConfig,
        addConfig,
        addNConfig,
        allConfig,
        anyConfig,
        argMaxConfig,
        argMinConfig,
        asinConfig,
        asinhConfig,
        atanConfig,
        atan2Config,
        atanhConfig,
        avgPoolConfig,
        avgPool3DConfig,
        avgPool3DGradConfig,
        avgPoolGradConfig,
        batchMatMulConfig,
        batchNormConfig,
        batchToSpaceNDConfig,
        bincountConfig,
        broadcastArgsConfig,
        castConfig,
        ceilConfig,
        clipByValueConfig,
        complexConfig,
        complexAbsConfig,
        concatConfig,
        conv2DConfig,
        conv2DBackpropFilterConfig,
        conv2DBackpropInputConfig,
        conv3DConfig,
        conv3DBackpropFilterV2Config,
        conv3DBackpropInputConfig,
        cosConfig,
        coshConfig,
        cropAndResizeConfig,
        cumprodConfig,
        cumsumConfig,
        denseBincountConfig,
        depthToSpaceConfig,
        depthwiseConv2dNativeConfig,
        depthwiseConv2dNativeBackpropFilterConfig,
        depthwiseConv2dNativeBackpropInputConfig,
        diagConfig,
        dilation2DConfig,
        einsumConfig,
        eluConfig,
        eluGradConfig,
        equalConfig,
        erfConfig,
        expConfig,
        expandDimsConfig,
        expm1Config,
        fftConfig,
        fillConfig,
        flipLeftRightConfig,
        floorConfig,
        floorDivConfig,
        fromPixelsConfig,
        fusedConv2DConfig,
        fusedDepthwiseConv2DConfig,
        gatherNdConfig,
        gatherV2Config,
        greaterConfig,
        greaterEqualConfig,
        identityConfig,
        ifftConfig,
        imagConfig,
        isFiniteConfig,
        isInfConfig,
        isNaNConfig,
        leakyReluConfig,
        lessConfig,
        lessEqualConfig,
        linSpaceConfig,
        logConfig,
        log1pConfig,
        logicalAndConfig,
        logicalNotConfig,
        logicalOrConfig,
        LRNConfig,
        LRNGradConfig,
        maxConfig,
        maximumConfig,
        maxPoolConfig,
        maxPool3DConfig,
        maxPool3DGradConfig,
        maxPoolGradConfig,
        maxPoolWithArgmaxConfig,
        meanConfig,
        minConfig,
        minimumConfig,
        mirrorPadConfig,
        modConfig,
        multinomialConfig,
        multiplyConfig,
        negConfig,
        nonMaxSuppressionV3Config,
        nonMaxSuppressionV4Config,
        nonMaxSuppressionV5Config,
        notEqualConfig,
        oneHotConfig,
        onesLikeConfig,
        packConfig,
        padV2Config,
        powConfig,
        preluConfig,
        prodConfig,
        rangeConfig,
        realConfig,
        realDivConfig,
        reciprocalConfig,
        reluConfig,
        relu6Config,
        reshapeConfig,
        resizeBilinearConfig,
        resizeBilinearGradConfig,
        resizeNearestNeighborConfig,
        resizeNearestNeighborGradConfig,
        reverseConfig,
        rotateWithOffsetConfig,
        roundConfig,
        rsqrtConfig,
        scatterNdConfig,
        searchSortedConfig,
        selectConfig,
        seluConfig,
        sigmoidConfig,
        signConfig,
        sinConfig,
        sinhConfig,
        sliceConfig,
        softmaxConfig,
        softplusConfig,
        spaceToBatchNDConfig,
        sparseFillEmptyRowsConfig,
        sparseReshapeConfig,
        sparseSegmentMeanConfig,
        sparseSegmentSumConfig,
        sparseToDenseConfig,
        splitVConfig,
        sqrtConfig,
        squareConfig,
        squaredDifferenceConfig,
        stepConfig,
        stridedSliceConfig,
        stringNGramsConfig,
        stringSplitConfig,
        stringToHashBucketFastConfig,
        subConfig,
        sumConfig,
        tanConfig,
        tanhConfig,
        tileConfig,
        topKConfig,
        transformConfig,
        transposeConfig,
        uniqueConfig,
        unpackConfig,
        unsortedSegmentSumConfig,
        zerosLikeConfig
    ];
    for (const kernelConfig of kernelConfigs) {
        registerKernel(kernelConfig);
    }

    /** @license See the LICENSE file. */
    // This code is auto-generated, do not modify this file!
    const version$1 = '3.17.0';

    /**
     * @license
     * Copyright 2018 Google LLC. All Rights Reserved.
     * Licensed under the Apache License, Version 2.0 (the "License");
     * you may not use this file except in compliance with the License.
     * You may obtain a copy of the License at
     *
     * http://www.apache.org/licenses/LICENSE-2.0
     *
     * Unless required by applicable law or agreed to in writing, software
     * distributed under the License is distributed on an "AS IS" BASIS,
     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     * See the License for the specific language governing permissions and
     * limitations under the License.
     * =============================================================================
     */
    const version = {
        'tfjs-core': version$7,
        'tfjs-backend-cpu': version$3,
        'tfjs-backend-webgl': version$2,
        'tfjs-data': version$4,
        'tfjs-layers': version$6,
        'tfjs-converter': version$5,
        'tfjs': version$1
    };

    var Tensorflow = /*#__PURE__*/Object.freeze({
        __proto__: null,
        data: index,
        version: version,
        AdadeltaOptimizer: AdadeltaOptimizer,
        AdagradOptimizer: AdagradOptimizer,
        AdamOptimizer: AdamOptimizer,
        AdamaxOptimizer: AdamaxOptimizer,
        MomentumOptimizer: MomentumOptimizer,
        Optimizer: Optimizer,
        OptimizerConstructors: OptimizerConstructors,
        RMSPropOptimizer: RMSPropOptimizer,
        SGDOptimizer: SGDOptimizer,
        Tensor: Tensor,
        TensorBuffer: TensorBuffer,
        Variable: Variable,
        get Rank () { return Rank; },
        sumOutType: sumOutType,
        upcastType: upcastType,
        get Reduction () { return Reduction; },
        customGrad: customGrad,
        grad: grad,
        grads: grads,
        valueAndGrad: valueAndGrad,
        valueAndGrads: valueAndGrads,
        variableGrads: variableGrads,
        Environment: Environment,
        env: env$1,
        get ENV () { return ENV$4; },
        nextFrame: nextFrame,
        KernelBackend: KernelBackend,
        DataStorage: DataStorage,
        abs: abs$2,
        acos: acos$2,
        acosh: acosh$2,
        add: add$2,
        addN: addN$2,
        all: all$2,
        any: any$2,
        argMax: argMax$2,
        argMin: argMin$2,
        asin: asin$2,
        asinh: asinh$2,
        atan: atan$2,
        atan2: atan2$2,
        atanh: atanh$2,
        avgPool: avgPool$2,
        avgPool3d: avgPool3d$1,
        basicLSTMCell: basicLSTMCell,
        batchToSpaceND: batchToSpaceND$2,
        batchNorm: batchNorm$2,
        batchNorm2d: batchNorm2d,
        batchNorm3d: batchNorm3d,
        batchNorm4d: batchNorm4d,
        bincount: bincount$2,
        broadcastArgs: broadcastArgs$2,
        broadcastTo: broadcastTo,
        buffer: buffer,
        cast: cast$3,
        ceil: ceil$2,
        clipByValue: clipByValue$2,
        clone: clone,
        complex: complex$2,
        concat: concat$2,
        concat1d: concat1d,
        concat2d: concat2d,
        concat3d: concat3d,
        concat4d: concat4d,
        conv1d: conv1d$1,
        conv2d: conv2d$3,
        conv2dTranspose: conv2dTranspose$1,
        conv3d: conv3d$1,
        conv3dTranspose: conv3dTranspose$1,
        cos: cos$2,
        cosh: cosh$2,
        cumprod: cumprod$2,
        cumsum: cumsum$2,
        denseBincount: denseBincount$2,
        depthToSpace: depthToSpace$2,
        depthwiseConv2d: depthwiseConv2d$3,
        diag: diag$2,
        dilation2d: dilation2d,
        div: div$1,
        divNoNan: divNoNan,
        dot: dot$2,
        einsum: einsum$2,
        elu: elu$4,
        equal: equal$3,
        erf: erf$2,
        euclideanNorm: euclideanNorm,
        exp: exp$2,
        expandDims: expandDims$3,
        expm1: expm1$2,
        eye: eye,
        fill: fill$2,
        floor: floor$3,
        floorDiv: floorDiv$2,
        gather: gather$1,
        greater: greater$3,
        greaterEqual: greaterEqual$2,
        imag: imag$2,
        isFinite: isFinite$3,
        isInf: isInf$2,
        isNaN: isNaN$3,
        leakyRelu: leakyRelu$2,
        less: less$3,
        lessEqual: lessEqual$2,
        linspace: linspace,
        localResponseNormalization: localResponseNormalization,
        log: log$3,
        log1p: log1p$2,
        logSigmoid: logSigmoid,
        logSoftmax: logSoftmax,
        logSumExp: logSumExp,
        logicalAnd: logicalAnd$2,
        logicalNot: logicalNot$2,
        logicalOr: logicalOr$2,
        logicalXor: logicalXor,
        lowerBound: lowerBound$1,
        matMul: matMul$1,
        max: max$3,
        maxPool: maxPool$2,
        maxPool3d: maxPool3d$1,
        maxPoolWithArgmax: maxPoolWithArgmax,
        maximum: maximum$3,
        mean: mean$1,
        meshgrid: meshgrid,
        min: min$3,
        minimum: minimum$3,
        mirrorPad: mirrorPad$1,
        mod: mod$2,
        moments: moments,
        mul: mul,
        multiRNNCell: multiRNNCell,
        multinomial: multinomial$2,
        neg: neg$2,
        notEqual: notEqual$3,
        oneHot: oneHot$2,
        ones: ones$1,
        onesLike: onesLike$2,
        outerProduct: outerProduct,
        pad: pad$1,
        pad1d: pad1d,
        pad2d: pad2d,
        pad3d: pad3d,
        pad4d: pad4d,
        pool: pool$1,
        pow: pow$2,
        prelu: prelu$3,
        print: print,
        prod: prod$2,
        rand: rand,
        randomGamma: randomGamma,
        randomNormal: randomNormal$2,
        randomUniform: randomUniform$1,
        range: range$4,
        real: real$2,
        reciprocal: reciprocal$2,
        relu: relu$2,
        relu6: relu6$2,
        reshape: reshape$3,
        reverse: reverse$2,
        reverse1d: reverse1d,
        reverse2d: reverse2d,
        reverse3d: reverse3d,
        reverse4d: reverse4d,
        round: round$2,
        rsqrt: rsqrt$2,
        scalar: scalar,
        selu: selu$2,
        separableConv2d: separableConv2d$1,
        setdiff1dAsync: setdiff1dAsync,
        sigmoid: sigmoid$2,
        sign: sign$2,
        sin: sin$2,
        sinh: sinh$2,
        slice: slice$2,
        slice1d: slice1d,
        slice2d: slice2d,
        slice3d: slice3d,
        slice4d: slice4d,
        softmax: softmax$3,
        softplus: softplus$2,
        spaceToBatchND: spaceToBatchND$2,
        fft: fft$2,
        ifft: ifft$2,
        irfft: irfft,
        rfft: rfft,
        split: split$2,
        sqrt: sqrt$2,
        square: square$2,
        squaredDifference: squaredDifference$2,
        squeeze: squeeze,
        stack: stack,
        step: step$2,
        stridedSlice: stridedSlice$2,
        sub: sub$2,
        sum: sum$2,
        tan: tan$2,
        tanh: tanh$2,
        tensor: tensor,
        tensor1d: tensor1d,
        tensor2d: tensor2d,
        tensor3d: tensor3d,
        tensor4d: tensor4d,
        tensor5d: tensor5d,
        tensor6d: tensor6d,
        tile: tile$3,
        topk: topk,
        truncatedNormal: truncatedNormal$1,
        unique: unique$3,
        unsortedSegmentSum: unsortedSegmentSum$2,
        unstack: unstack,
        upperBound: upperBound$1,
        variable: variable,
        where: where,
        whereAsync: whereAsync,
        zeros: zeros$2,
        zerosLike: zerosLike$2,
        op: op,
        OP_SCOPE_SUFFIX: OP_SCOPE_SUFFIX,
        booleanMaskAsync: booleanMaskAsync,
        transpose: transpose$2,
        norm: norm,
        movingAverage: movingAverage,
        scatterND: scatterND,
        searchSorted: searchSorted$2,
        sparseToDense: sparseToDense$2,
        gatherND: gatherND,
        dropout: dropout$2,
        enclosingPowerOfTwo: enclosingPowerOfTwo,
        cosineWindow: cosineWindow,
        inTopKAsync: inTopKAsync,
        image: image$1,
        linalg: linalg,
        losses: losses,
        spectral: spectral$1,
        fused: fused_ops,
        signal: signal,
        sparse: sparse$1,
        string: string$1,
        train: train,
        enableProdMode: enableProdMode,
        enableDebugMode: enableDebugMode,
        disableDeprecationWarnings: disableDeprecationWarnings,
        deprecationWarn: deprecationWarn,
        disposeVariables: disposeVariables,
        engine: engine,
        memory: memory,
        profile: profile,
        tidy: tidy,
        dispose: dispose,
        keep: keep,
        time: time,
        setBackend: setBackend,
        ready: ready,
        getBackend: getBackend,
        removeBackend: removeBackend,
        findBackend: findBackend,
        findBackendFactory: findBackendFactory,
        registerBackend: registerBackend,
        backend: backend,
        setPlatform: setPlatform,
        getKernel: getKernel,
        getGradient: getGradient,
        getKernelsForBackend: getKernelsForBackend,
        registerKernel: registerKernel,
        registerGradient: registerGradient,
        unregisterKernel: unregisterKernel,
        unregisterGradient: unregisterGradient,
        copyRegisteredKernels: copyRegisteredKernels,
        Abs: Abs,
        Acos: Acos,
        Acosh: Acosh,
        Add: Add$1,
        AddN: AddN,
        All: All,
        Any: Any,
        ArgMax: ArgMax,
        ArgMin: ArgMin,
        Asin: Asin,
        Asinh: Asinh,
        Atan: Atan,
        Atanh: Atanh,
        Atan2: Atan2,
        AvgPool: AvgPool,
        AvgPoolGrad: AvgPoolGrad,
        AvgPool3D: AvgPool3D,
        AvgPool3DGrad: AvgPool3DGrad,
        BatchMatMul: BatchMatMul,
        BatchToSpaceND: BatchToSpaceND,
        Bincount: Bincount,
        BroadcastTo: BroadcastTo,
        BroadcastArgs: BroadcastArgs,
        Cast: Cast,
        Ceil: Ceil,
        ClipByValue: ClipByValue,
        Complex: Complex,
        ComplexAbs: ComplexAbs,
        Concat: Concat,
        Conv2D: Conv2D$1,
        Conv2DBackpropFilter: Conv2DBackpropFilter,
        Conv2DBackpropInput: Conv2DBackpropInput,
        Conv3D: Conv3D$1,
        Conv3DBackpropFilterV2: Conv3DBackpropFilterV2,
        Conv3DBackpropInputV2: Conv3DBackpropInputV2,
        Cos: Cos,
        Cosh: Cosh,
        Cumprod: Cumprod,
        Cumsum: Cumsum,
        CropAndResize: CropAndResize,
        DenseBincount: DenseBincount,
        DepthToSpace: DepthToSpace,
        DepthwiseConv2dNative: DepthwiseConv2dNative,
        DepthwiseConv2dNativeBackpropFilter: DepthwiseConv2dNativeBackpropFilter,
        DepthwiseConv2dNativeBackpropInput: DepthwiseConv2dNativeBackpropInput,
        Diag: Diag,
        Dilation2D: Dilation2D,
        Dilation2DBackpropInput: Dilation2DBackpropInput,
        Dilation2DBackpropFilter: Dilation2DBackpropFilter,
        RealDiv: RealDiv,
        Einsum: Einsum,
        Elu: Elu$1,
        EluGrad: EluGrad,
        Erf: Erf,
        Equal: Equal,
        Exp: Exp,
        ExpandDims: ExpandDims,
        Expm1: Expm1,
        FFT: FFT,
        Fill: Fill,
        FlipLeftRight: FlipLeftRight,
        Floor: Floor,
        FloorDiv: FloorDiv,
        FusedBatchNorm: FusedBatchNorm,
        GatherV2: GatherV2,
        GatherNd: GatherNd,
        Greater: Greater,
        GreaterEqual: GreaterEqual,
        Identity: Identity$1,
        IFFT: IFFT,
        Imag: Imag,
        IsFinite: IsFinite,
        IsInf: IsInf,
        IsNan: IsNan,
        LeakyRelu: LeakyRelu,
        Less: Less,
        LessEqual: LessEqual,
        LinSpace: LinSpace,
        Log: Log,
        Log1p: Log1p,
        LogicalAnd: LogicalAnd,
        LogicalNot: LogicalNot,
        LogicalOr: LogicalOr,
        LogSoftmax: LogSoftmax$1,
        LowerBound: LowerBound,
        LRN: LRN,
        LRNGrad: LRNGrad,
        Max: Max,
        Maximum: Maximum$1,
        MaxPool: MaxPool,
        MaxPoolGrad: MaxPoolGrad,
        MaxPool3D: MaxPool3D,
        MaxPool3DGrad: MaxPool3DGrad,
        MaxPoolWithArgmax: MaxPoolWithArgmax,
        Mean: Mean,
        Min: Min,
        Minimum: Minimum$1,
        MirrorPad: MirrorPad,
        Mod: Mod,
        Multinomial: Multinomial,
        Multiply: Multiply$1,
        Neg: Neg,
        NotEqual: NotEqual,
        NonMaxSuppressionV3: NonMaxSuppressionV3,
        NonMaxSuppressionV4: NonMaxSuppressionV4,
        NonMaxSuppressionV5: NonMaxSuppressionV5,
        OnesLike: OnesLike,
        OneHot: OneHot,
        Pack: Pack,
        PadV2: PadV2,
        Pool: Pool,
        Pow: Pow,
        Prelu: Prelu,
        Prod: Prod,
        Range: Range,
        Real: Real,
        Reciprocal: Reciprocal,
        Relu: Relu$1,
        Reshape: Reshape$1,
        ResizeNearestNeighbor: ResizeNearestNeighbor,
        ResizeNearestNeighborGrad: ResizeNearestNeighborGrad,
        ResizeBilinear: ResizeBilinear,
        ResizeBilinearGrad: ResizeBilinearGrad,
        Relu6: Relu6$1,
        Reverse: Reverse,
        Round: Round,
        Rsqrt: Rsqrt,
        ScatterNd: ScatterNd,
        SearchSorted: SearchSorted,
        Select: Select,
        Selu: Selu$1,
        Slice: Slice,
        Sin: Sin,
        Sinh: Sinh,
        Sign: Sign,
        Sigmoid: Sigmoid$1,
        Softplus: Softplus$1,
        Sqrt: Sqrt,
        Sum: Sum,
        SpaceToBatchND: SpaceToBatchND,
        SplitV: SplitV,
        Softmax: Softmax$2,
        SparseFillEmptyRows: SparseFillEmptyRows,
        SparseReshape: SparseReshape,
        SparseSegmentMean: SparseSegmentMean,
        SparseSegmentSum: SparseSegmentSum,
        SparseToDense: SparseToDense,
        SquaredDifference: SquaredDifference,
        Square: Square,
        StridedSlice: StridedSlice,
        StringNGrams: StringNGrams,
        StringSplit: StringSplit,
        StringToHashBucketFast: StringToHashBucketFast,
        Sub: Sub,
        Tan: Tan,
        Tanh: Tanh$1,
        Tile: Tile,
        TopK: TopK,
        Transform: Transform$1,
        Transpose: Transpose,
        Unique: Unique,
        Unpack: Unpack,
        UnsortedSegmentSum: UnsortedSegmentSum,
        UpperBound: UpperBound,
        ZerosLike: ZerosLike,
        Step: Step,
        FromPixels: FromPixels,
        RotateWithOffset: RotateWithOffset,
        _FusedMatMul: _FusedMatMul,
        FusedConv2D: FusedConv2D,
        FusedDepthwiseConv2D: FusedDepthwiseConv2D,
        version_core: version$7,
        browser: browser$2,
        io: io,
        math: math,
        serialization: serialization,
        test_util: test_util,
        util: util$2,
        backend_util: backend_util,
        broadcast_util: broadcast_util,
        tensor_util: tensor_util,
        slice_util: slice_util,
        gather_util: gather_nd_util,
        scatter_util: scatter_nd_util,
        device_util: device_util,
        kernel_impls: kernel_impls,
        CallbackList: CallbackList,
        CustomCallback: CustomCallback,
        History: History,
        Callback: Callback,
        callbacks: callbacks,
        EarlyStopping: EarlyStopping,
        InputSpec: InputSpec,
        SymbolicTensor: SymbolicTensor,
        LayersModel: LayersModel,
        input: input,
        loadLayersModel: loadLayersModel,
        model: model$1,
        registerCallbackConstructor: registerCallbackConstructor,
        sequential: sequential,
        RNN: RNN,
        Sequential: Sequential,
        LayerVariable: LayerVariable,
        version_layers: version$6,
        constraints: exports_constraints,
        initializers: exports_initializers,
        layers: exports_layers,
        metrics: exports_metrics,
        models: exports_models,
        regularizers: exports_regularizers,
        GraphModel: GraphModel,
        loadGraphModel: loadGraphModel,
        deregisterOp: deregisterOp,
        registerOp: registerOp,
        version_converter: version$5
    });

    // import * as tensorflow from '@tensorflow/tfjs';



























































































































    /******************************************************************************
     * tensorflow.js lambda layer
     * written by twitter.com/benjaminwegener
     * license: MIT
     * @see https://benjamin-wegener.blogspot.com/2020/02/tensorflowjs-lambda-layer.html
     */
    class LambdaLayer extends Layer {
      
      
      //@ts-ignore
      
      constructor(config) {
        super(config);
        if (config.name === undefined) {
            config.name = ((+new Date) * Math.random()).toString(36); //random name from timestamp in case name hasn't been set
        }
        this.name = config.name;
        this.lambdaFunction = config.lambdaFunction;
        if(config.lambdaOutputShape) this.lambdaOutputShape = config.lambdaOutputShape;
      }

      call(input, kwargs) {
        // console.log({ input }, 'input[0].shape', input[0].shape)
        // input[0].data().then(inputData=>console.log)
        // console.log('input[0].data()', input[0].data())
        // return input;
        return tidy(() => {
          // return tf.mean(tf.tensor(input),1,true)
          let result = new Array();
          // eval(this.lambdaFunction);
          result = (new Function('input', 'tf', this.lambdaFunction))(input, Tensorflow);
          // result = tf.mean(input,1);
          return result;
        });
      }

      computeOutputShape(inputShape) {
        // console.log('computeOutputShape',{inputShape})
        if (this.lambdaOutputShape === undefined) { //if no outputshape provided, try to set as inputshape
          return inputShape[0];
        } else {
          return this.lambdaOutputShape;
        }
      }

      getConfig() {
        const config = {
          ...super.getConfig(),
          lambdaFunction: this.lambdaFunction,
          lambdaOutputShape: this.lambdaOutputShape,
        };
        return config;
      }

      static get className() {
          return 'LambdaLayer';
      }
    }

    /**
     * Base class for tensorscript models
     * @interface TensorScriptModelInterface
     * @property {Object} settings - tensorflow model hyperparameters
     * @property {Object} model - tensorflow model
     * @property {Object} tf - tensorflow / tensorflow-node / tensorflow-node-gpu
     * @property {Function} reshape - static reshape array function
     * @property {Function} getInputShape - static TensorScriptModelInterface
     */
    class TensorScriptModelInterface  {
      
      
      
      
      
      
      
      
      
      
      
      
      
      
      /**
       * @param {Object} options - tensorflow model hyperparameters
       * @param {Object} customTF - custom, overridale tensorflow / tensorflow-node / tensorflow-node-gpu
       * @param {{model:Object,tf:Object,}} properties - extra instance properties
       */
      constructor(options = {}, properties = {}) {
        // tf.setBackend('cpu');
        this.type = 'ModelInterface';
        /** @type {Object} */
        this.settings = Object.assign({  }, options);
        /** @type {Object} */
        this.model = properties.model;
        /** @type {Object} */
        this.tf = properties.tf || Tensorflow;
        /** @type {Boolean} */
        this.trained = false;
        this.compiled = false;
        /** @type {Function} */
        this.reshape = TensorScriptModelInterface.reshape;
        /** @type {Function} */
        this.getInputShape = TensorScriptModelInterface.getInputShape;
        if( this.tf && this.tf.serialization && this.tf.serialization.registerClass) this.tf.serialization.registerClass(LambdaLayer);
        return this;
      }
      /**
       * Reshapes an array
       * @function
       * @example 
       * const array = [ 0, 1, 1, 0, ];
       * const shape = [2,2];
       * TensorScriptModelInterface.reshape(array,shape) // => 
       * [
       *   [ 0, 1, ],
       *   [ 1, 0, ],
       * ];
       * @param {Array<number>} array - input array 
       * @param {Array<number>} shape - shape array 
       * @return {Array<Array<number>>} returns a matrix with the defined shape
       */
      /* istanbul ignore next */
      static reshape(array, shape) {
        const flatArray = flatten(array);
       

        function product (arr) {
          return arr.reduce((prev, curr) => prev * curr);
        }
      
        if (!Array.isArray(array) || !Array.isArray(shape)) {
          throw new TypeError('Array expected');
        }
      
        if (shape.length === 0) {
          //@ts-ignore
          throw new DimensionError(0, product(size(array)), '!=');
        }
        let newArray;
        let totalSize = 1;
        const rows = shape[ 0 ];
        for (let sizeIndex = 0; sizeIndex < shape.length; sizeIndex++) {
          totalSize *= shape[sizeIndex];
        }
      
        if (flatArray.length !== totalSize) {
          throw new DimensionError(
            product(shape),
            //@ts-ignore
            product(size(array)),
            '!='
          );
        }
      
        try {
          newArray = _reshape(flatArray, shape);
        } catch (e) {
          if (e instanceof DimensionError) {
            throw new DimensionError(
              product(shape),
              //@ts-ignore
              product(size(array)),
              '!='
            );
          }
          throw e;
        }
        if (newArray.length !== rows) throw new SyntaxError(`specified shape (${shape}) is compatible with input array or length (${array.length})`);

        // console.log({ newArray ,});
        //@ts-ignore
        return newArray;
      }
      /**
       * Returns the shape of an input matrix
       * @function
       * @example 
       * const input = [
       *   [ 0, 1, ],
       *   [ 1, 0, ],
       * ];
       * TensorScriptModelInterface.getInputShape(input) // => [2,2]
       * @see {https://stackoverflow.com/questions/10237615/get-size-of-dimensions-in-array}
       * @param {Array<Array<number>>} matrix - input matrix 
       * @return {Array<number>} returns the shape of a matrix (e.g. [2,2])
       */
      //@ts-ignore
      static getInputShape(matrix=[]) {
      // static getInputShape(matrix:NestedArray<V>=[]):Shape {
        if (Array.isArray(matrix) === false || !matrix[ 0 ] || !matrix[ 0 ].length || Array.isArray(matrix[ 0 ]) === false) throw new TypeError('input must be a matrix');
        const dim = [];
        const x_dimensions = matrix[ 0 ].length;
        let vectors = matrix;
        matrix.forEach((vector) => {
          if (vector.length !== x_dimensions) throw new SyntaxError('input must have the same length in each row');
        });
        for (;;) {
          dim.push(vectors.length);
          if (Array.isArray(vectors[0])) {
            vectors = vectors[0];
          } else {
            break;
          }
        }
        return dim;
      }
      exportConfiguration() {
        return {
          type: this.type,
          settings: this.settings,
          trained: this.trained,
          compiled: this.compiled,
          xShape: this.xShape,
          yShape: this.yShape,
          layers: this.layers,
        };
      }
      importConfiguration(configuration) {
        this.type = configuration.type || this.type;
        this.settings = {
          ...this.settings,
          ...configuration.settings,
        };
        this.trained = configuration.trained || this.trained;
        this.compiled = configuration.compiled || this.compiled;
        this.xShape = configuration.xShape || this.xShape;
        this.yShape = configuration.yShape || this.yShape;
        this.layers = configuration.layers || this.layers;
      }
      /**
       * Asynchronously trains tensorflow model, must be implemented by tensorscript class
       * @abstract 
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @return {Object} returns trained tensorflow model 
       */
      

      train(x_matrix, y_matrix) {
        throw new ReferenceError('train method is not implemented');
      }
      /**
       * Predicts new dependent variables
       * @abstract 
       * @param {Array<Array<number>>|Array<number>} matrix - new test independent variables
       * @return {{data: Promise}} returns tensorflow prediction 
       */
      calculate(matrix) {
        throw new ReferenceError('calculate method is not implemented');
      }
      /**
       * Loads a saved tensoflow / keras model, this is an alias for 
       * @param {Object} options - tensorflow load model options
       * @return {Object} tensorflow model
       * @see {@link https://www.tensorflow.org/js/guide/save_load#loading_a_tfmodel}
       */
      async loadModel(options) {
        this.model = await this.tf.loadLayersModel(options);
        this.xShape = this.model.inputs[0].shape;
        this.yShape = this.model.outputs[0].shape;
        this.trained = true;
        return this.model;
      }
      /**
       * saves a tensorflow model, this is an alias for 
       * @param {Object} options - tensorflow save model options
       * @return {Object} tensorflow model
       * @see {@link https://www.tensorflow.org/js/guide/save_load#save_a_tfmodel}
       */
      async saveModel(options) {
        const savedStatus = await this.model.save(options);
        return savedStatus;
      }

      /**
       * Returns prediction values from tensorflow model
       * @param {Array<Array<number>>|Array<number>} input_matrix - new test independent variables 
       * @param {Boolean} [options.json=true] - return object instead of typed array
       * @param {Boolean} [options.probability=true] - return real values instead of integers
       * @param {Boolean} [options.skip_matrix_check=false] - validate input is a matrix
       * @return {Array<number>|Array<Array<number>>} predicted model values
       */
      // async predict(options?:Matrix|Vector|InputTextArray|PredictionOptions):Promise<Matrix> 
      async predict(input_matrix, options) {
        if (!input_matrix || Array.isArray(input_matrix) === false) throw new Error('invalid input matrix');
        const config = {
          json: true,
          probability: true,
          ...options
        };
        const x_matrix = (Array.isArray(input_matrix ) || config.skip_matrix_check)
              ? input_matrix
          : [
            input_matrix,
          ];
        return this.calculate(x_matrix )
          .data()
          .then((predictions) => {
            // console.log({ predictions });
            if (config.json === false) {
              return predictions;
            } else {
              if (!this.yShape) throw new Error('Model is missing yShape');
              const shape = [(x_matrix ).length, this.yShape[ 1 ], ];
              const predictionValues = (config.probability === false) ? Array.from(predictions).map(Math.round) : Array.from(predictions);
              return this.reshape(predictionValues, shape);
            }
          })
          .catch((e) => {
            throw e; 
          });
      }
    }

    /**
     * Calculate the size of a multi dimensional array.
     * This function checks the size of the first entry, it does not validate
     * whether all dimensions match. (use function `validate` for that) (from math.js)
     * @param {Array} x
     * @see {https://github.com/josdejong/mathjs/blob/develop/src/utils/array.js}
     * @ignore
     * @return {Number[]} size
     */
    /* istanbul ignore next */
    function size (x) {
      let s = [];

      while (Array.isArray(x)) {
        s.push(x.length);
        x = x[0];
      }

      return s;
    }
    /**
     * Iteratively re-shape a multi dimensional array to fit the specified dimensions (from math.js)
     * @param {Array} array           Array to be reshaped
     * @param {Array.<number>} sizes  List of sizes for each dimension
     * @returns {Array}               Array whose data has been formatted to fit the
     *                                specified dimensions
     * @ignore
     * @see {https://github.com/josdejong/mathjs/blob/develop/src/utils/array.js}
     */
    /* istanbul ignore next */
    function _reshape(array, sizes) {
      // testing if there are enough elements for the requested shape
      var tmpArray = array;
      var tmpArray2;
      // for each dimensions starting by the last one and ignoring the first one
      for (var sizeIndex = sizes.length - 1; sizeIndex > 0; sizeIndex--) {
        var size = sizes[sizeIndex];
        tmpArray2 = [];

        // aggregate the elements of the current tmpArray in elements of the requested size
        var length = tmpArray.length / size;
        for (var i = 0; i < length; i++) {
          tmpArray2.push(tmpArray.slice(i * size, (i + 1) * size));
        }
        // set it as the new tmpArray for the next loop turn or for return
        //@ts-ignore
        tmpArray = tmpArray2;
      }
      //@ts-ignore
      return tmpArray;
    }

    /**
     * Create a range error with the message:
     *     'Dimension mismatch (<actual size> != <expected size>)' (from math.js)
     * @param {number | number[]} actual        The actual size
     * @param {number | number[]} expected      The expected size
     * @param {string} [relation='!=']          Optional relation between actual
     *                                          and expected size: '!=', '<', etc.
     * @extends RangeError
     * @ignore
     * @see {https://github.com/josdejong/mathjs/blob/develop/src/utils/array.js}
     */
    /* istanbul ignore next */
    class DimensionError extends RangeError {
      
      
      
      
      constructor(actual, expected, relation) {
        /* istanbul ignore next */
        const message = 'Dimension mismatch (' + (Array.isArray(actual) ? ('[' + actual.join(', ') + ']') : actual) + ' ' + ('!=') + ' ' + (Array.isArray(expected) ? ('[' + expected.join(', ') + ']') : expected) +  ')';
        super(message);
      
        this.actual = actual;
        this.expected = expected;
        this.relation = relation;
        // this.stack = (new Error()).stack
        this.message = message;
        this.name = 'DimensionError';
        this.isDimensionError = true;
      }
    }

    /**
     * Flatten a multi dimensional array, put all elements in a one dimensional
     * array
     * @param {Array} array   A multi dimensional array
     * @ignore
     * @see {https://github.com/josdejong/mathjs/blob/develop/src/utils/array.js}
     * @return {Array}        The flattened array (1 dimensional)
     */
    /* istanbul ignore next */
    function flatten(array) {
      /* istanbul ignore next */
      if (!Array.isArray(array)) {
        // if not an array, return as is
        /* istanbul ignore next */
        return array;
      }
      let flat = [];
      
      /* istanbul ignore next */
      array.forEach(function callback(value) {
        if (Array.isArray(value)) {
          value.forEach(callback); // traverse through sub-arrays recursively
        } else {
          flat.push(value);
        }
      });

      return flat;
    }


    async function asyncForEach(array, callback) {
      for (let index = 0; index < array.length; index++) {
        await callback(array[index], index, array);
      }
    }

    /**
     * Deep Learning with Tensorflow
     * @class BaseNeuralNetwork
     * @implements {TensorScriptModelInterface}
     */
    class BaseNeuralNetwork extends TensorScriptModelInterface {
      /**
       * @param {{layers:Array<Object>,compile:Object,fit:Object}} options - neural network configuration and tensorflow model hyperparameters
       * @param {{model:Object,tf:Object,}} properties - extra instance properties
       */
      constructor(options = {}, properties) {
        const config = Object.assign({
          layers: [],
          compile: {
            loss: 'meanSquaredError',
            optimizer: 'adam',
          },
          fit: {
            epochs: 100,
            batchSize: 5,
            verbose:0,
          },
        }, options);
        super(config, properties);
        this.type = 'BaseNeuralNetwork';
        return this;
      }
      /**
       * Adds dense layers to tensorflow model
       * @abstract 
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @param {Array<Object>} layers - model dense layer parameters
       */
       
      generateLayers( x_matrix, layers){
        throw new ReferenceError('generateLayers method is not implemented');
      }
      /**
       * Asynchronously trains tensorflow model
       * @override
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @param {Array<Object>} layers - array of model dense layer parameters
       * @param {Array<Array<number>>} x_text - validation data independent variables
       * @param {Array<Array<number>>} y_text - validation data dependent variables
       * @return {Object} returns trained tensorflow model 
       */
      async train(x_matrix, y_matrix, layers, x_test, y_test) {
        const xShape = this.getInputShape(x_matrix);
        const yShape = this.getInputShape(y_matrix);
        const xs = this.tf.tensor(x_matrix, xShape);
        const ys = this.tf.tensor(y_matrix, yShape);
        this.xShape = xShape;
        this.yShape = yShape;
        if (typeof this.compiled==='undefined' || this.compiled === false) {
          this.model = this.tf.sequential();
          this.generateLayers.call(this, x_matrix, y_matrix, layers || this.layers, x_test, y_test);
          this.model.compile(this.settings.compile);
          this.compiled = true;
        }
        await this.model.fit(xs, ys, this.settings.fit);
        this.trained = true;
        xs.dispose();
        ys.dispose();
        return this.model;
      }
      /**
       * Predicts new dependent variables
       * @override
       * @param {Array<Array<number>>|Array<number>} matrix - new test independent variables
       * @param {Object} options - model prediction options
       * @return {{data: Promise}} returns tensorflow prediction 
       */
      calculate(input_matrix, options) {
        if (!input_matrix || Array.isArray(input_matrix)===false) throw new Error('invalid input matrix');
        const predictionInput = (Array.isArray(input_matrix[ 0 ]))
          ? input_matrix
          : [
            input_matrix,
          ];
        const predictionTensor = this.tf.tensor(predictionInput);
        const prediction = this.model.predict(predictionTensor, options);
        predictionTensor.dispose();
        return prediction;
      }
    }

    /**
     * Deep Learning Regression with Tensorflow
     * @class DeepLearningRegression
     * @implements {BaseNeuralNetwork}
     */
    class DeepLearningRegression extends BaseNeuralNetwork {
      /**
       * @param {{layers:Array<Object>,compile:Object,fit:Object,layerPreference:String}} options - neural network configuration and tensorflow model hyperparameters
       * @param {{model:Object,tf:Object,}} properties - extra instance properties
       */
      constructor(options = {}, properties) {
        const config = Object.assign({
          layers: [],
          layerPreference:'deep',
          compile: {
            loss: 'meanSquaredError',
            optimizer: 'adam',
          },
          fit: {
            epochs: 100,
            batchSize: 5
          },
        }, options);
        super(config, properties);
        this.type = 'DeepLearningRegression';

        return this;
      }
      /**
       * Adds dense layers to tensorflow regression model
       * @override 
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @param {Array<Object>} layers - model dense layer parameters
       */
      generateLayers(x_matrix, y_matrix, layers) {
        const xShape = this.getInputShape(x_matrix);
        const yShape = this.getInputShape(y_matrix);
        const denseLayers = [];
        if (layers) {
          denseLayers.push(...layers);
        } else if(this.settings.layerPreference==='deep') {
          denseLayers.push({ units: xShape[ 1 ], inputShape: [xShape[1],], kernelInitializer: 'randomNormal', activation: 'relu', });
          denseLayers.push({ units: parseInt(String(Math.ceil(xShape[ 1 ] / 2)), 10), kernelInitializer: 'randomNormal', activation: 'relu', });
          denseLayers.push({ units: yShape[ 1 ], kernelInitializer: 'randomNormal', });
        } else {
          denseLayers.push({ units: (xShape[ 1 ] * 2), inputShape: [xShape[1],], kernelInitializer: 'randomNormal', activation: 'relu', });
          denseLayers.push({ units: yShape[ 1 ], kernelInitializer: 'randomNormal', });
        }
        this.layers = denseLayers;
        denseLayers.forEach(layer => {
          this.model.add(this.tf.layers.dense(layer));
        });
      }
    }

    /**
     * Deep Learning Classification with Tensorflow
     * @class DeepLearningClassification
     * @implements {BaseNeuralNetwork}
     */
    class DeepLearningClassification extends BaseNeuralNetwork{
      /**
       * @param {{layers:Array<Object>,compile:Object,fit:Object}} options - neural network configuration and tensorflow model hyperparameters
       * @param {{model:Object,tf:Object,}} properties - extra instance properties
       */
      constructor(options = {}, properties) {
        const config = Object.assign({
          layers: [],
          compile: {
            loss: 'categoricalCrossentropy',
            optimizer: 'adam',
          },
          fit: {
            epochs: 100,
            batchSize: 5,
          },
        }, options);
        super(config, properties);
        this.type = 'DeepLearningClassification';
        return this;
      }
      /**
       * Adds dense layers to tensorflow classification model
       * @override 
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @param {Array<Object>} layers - model dense layer parameters
       */
      generateLayers(x_matrix, y_matrix, layers) {
        const xShape = this.getInputShape(x_matrix);
        const yShape = this.getInputShape(y_matrix);
        this.yShape = yShape;
        this.xShape = xShape;
        const denseLayers = [];
        if (layers) {
          denseLayers.push(...layers);
        } else {
          denseLayers.push({ units: (xShape[ 1 ] * 2), inputDim: xShape[1],  activation: 'relu', });
          denseLayers.push({ units: yShape[ 1 ], activation: 'softmax', });
        }
        this.layers = denseLayers;
        denseLayers.forEach(layer => {
          this.model.add(this.tf.layers.dense(layer));
        });
      }
    }

    /**
     * Logistic Regression Classification with Tensorflow
     * @class LogisticRegression
     * @implements {BaseNeuralNetwork}
     */
    class LogisticRegression extends BaseNeuralNetwork {
      /**
       * @param {{layers:Array<Object>,compile:Object,fit:Object}} options - neural network configuration and tensorflow model hyperparameters
       * @param {{model:Object,tf:Object,}} properties - extra instance properties
       */
      constructor(options = {}, properties) {
        const config = Object.assign({
          layers: [],
          type:'simple',
          compile: {
            loss: 'meanSquaredError',
            optimizer: 'rmsprop',
          },
          fit: {
            epochs: 100,
            batchSize: 5,
          },
        }, options);
        super(config, properties);
        this.type = 'LogisticRegression';
        return this;
      }
      /**
       * Adds dense layers to tensorflow classification model
       * @override 
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @param {Array<Object>} layers - model dense layer parameters
       * @param {Array<Array<number>>} x_test - validation data independent variables
       * @param {Array<Array<number>>} y_test - validation data dependent variables
       */
      generateLayers(x_matrix, y_matrix, layers, x_test, y_test) {
        const xShape = this.getInputShape(x_matrix);
        const yShape = this.getInputShape(y_matrix);
        this.yShape = yShape;
        this.xShape = xShape;
        const denseLayers = [];
        if (layers) {
          denseLayers.push(...layers);
        } else if (this.settings.type==='class' && this.settings.compile) { 
          denseLayers.push({ units: 1, inputDim:  xShape[ 1 ], activation: 'sigmoid', });
          this.settings.compile.loss = 'binaryCrossentropy';
        } else if (this.settings.type === 'l1l2' && this.settings.compile) { 
          const kernelRegularizer = this.tf.regularizers.l1l2({ l1: 0.01, l2: 0.01, });
          denseLayers.push({ units: 1, inputDim:  xShape[ 1 ], activation: 'sigmoid', kernelRegularizer, });
          this.settings.compile.loss = 'binaryCrossentropy';
        } else {
          denseLayers.push({ units: 1, inputShape: [xShape[1], ], });
        }
        this.layers = denseLayers;
        denseLayers.forEach(layer => {
          this.model.add(this.tf.layers.dense(layer));
        });
        /* istanbul ignore next */
        if (x_test && y_test && this.settings && this.settings.fit) {
          this.settings.fit.validationData = [x_test, y_test];
        }
      }
    }

    /**
     * Mulitple Linear Regression with Tensorflow
     * @class MultipleLinearRegression
     * @implements {BaseNeuralNetwork}
     */
    class MultipleLinearRegression extends BaseNeuralNetwork {
      /**
       * @param {{layers:Array<Object>,compile:Object,fit:Object}} options - neural network configuration and tensorflow model hyperparameters
       * @param {{model:Object,tf:Object,}} properties - extra instance properties
       */
      /* istanbul ignore next */
      constructor(options = {}, properties = {}) {
        const config = Object.assign({
          layers: [],
          compile: {
            loss: 'meanSquaredError',
            optimizer: 'sgd',
          },
          fit: {
            epochs: 500,
            batchSize: 5,
          },
        }, options);
        super(config, properties);
        this.type = 'MultipleLinearRegression';

        return this;
      }
      /**
       * Adds dense layers to tensorflow regression model
       * @override 
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @param {Array<Object>} layers - model dense layer parameters
       */
      generateLayers(x_matrix, y_matrix, layers) {
        const xShape = this.getInputShape(x_matrix);
        const yShape = this.getInputShape(y_matrix);
        const denseLayers = [];
        if (layers) {
          denseLayers.push(...layers);
        } else {
          denseLayers.push({ units: yShape[1], inputShape: [xShape[1],], });
        }
        this.layers = denseLayers;
        denseLayers.forEach(layer => {
          this.model.add(this.tf.layers.dense(layer));
        });
      }
    }

    /**
     * lodash (Custom Build) <https://lodash.com/>
     * Build: `lodash modularize exports="npm" -o ./`
     * Copyright jQuery Foundation and other contributors <https://jquery.org/>
     * Released under MIT license <https://lodash.com/license>
     * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
     * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
     */

    /** Used as references for various `Number` constants. */
    var INFINITY = 1 / 0,
        MAX_SAFE_INTEGER = 9007199254740991,
        MAX_INTEGER = 1.7976931348623157e+308,
        NAN = 0 / 0;

    /** `Object#toString` result references. */
    var funcTag = '[object Function]',
        genTag = '[object GeneratorFunction]',
        symbolTag = '[object Symbol]';

    /** Used to match leading and trailing whitespace. */
    var reTrim = /^\s+|\s+$/g;

    /** Used to detect bad signed hexadecimal string values. */
    var reIsBadHex = /^[-+]0x[0-9a-f]+$/i;

    /** Used to detect binary string values. */
    var reIsBinary = /^0b[01]+$/i;

    /** Used to detect octal string values. */
    var reIsOctal = /^0o[0-7]+$/i;

    /** Used to detect unsigned integer values. */
    var reIsUint = /^(?:0|[1-9]\d*)$/;

    /** Built-in method references without a dependency on `root`. */
    var freeParseInt = parseInt;

    /** Used for built-in method references. */
    var objectProto = Object.prototype;

    /**
     * Used to resolve the
     * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
     * of values.
     */
    var objectToString$1 = objectProto.toString;

    /* Built-in method references for those with the same name as other `lodash` methods. */
    var nativeCeil = Math.ceil,
        nativeMax = Math.max;

    /**
     * The base implementation of `_.range` and `_.rangeRight` which doesn't
     * coerce arguments.
     *
     * @private
     * @param {number} start The start of the range.
     * @param {number} end The end of the range.
     * @param {number} step The value to increment or decrement by.
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {Array} Returns the range of numbers.
     */
    function baseRange(start, end, step, fromRight) {
      var index = -1,
          length = nativeMax(nativeCeil((end - start) / (step || 1)), 0),
          result = Array(length);

      while (length--) {
        result[fromRight ? length : ++index] = start;
        start += step;
      }
      return result;
    }

    /**
     * Creates a `_.range` or `_.rangeRight` function.
     *
     * @private
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {Function} Returns the new range function.
     */
    function createRange(fromRight) {
      return function(start, end, step) {
        if (step && typeof step != 'number' && isIterateeCall(start, end, step)) {
          end = step = undefined;
        }
        // Ensure the sign of `-0` is preserved.
        start = toFinite(start);
        if (end === undefined) {
          end = start;
          start = 0;
        } else {
          end = toFinite(end);
        }
        step = step === undefined ? (start < end ? 1 : -1) : toFinite(step);
        return baseRange(start, end, step, fromRight);
      };
    }

    /**
     * Checks if `value` is a valid array-like index.
     *
     * @private
     * @param {*} value The value to check.
     * @param {number} [length=MAX_SAFE_INTEGER] The upper bounds of a valid index.
     * @returns {boolean} Returns `true` if `value` is a valid index, else `false`.
     */
    function isIndex(value, length) {
      length = length == null ? MAX_SAFE_INTEGER : length;
      return !!length &&
        (typeof value == 'number' || reIsUint.test(value)) &&
        (value > -1 && value % 1 == 0 && value < length);
    }

    /**
     * Checks if the given arguments are from an iteratee call.
     *
     * @private
     * @param {*} value The potential iteratee value argument.
     * @param {*} index The potential iteratee index or key argument.
     * @param {*} object The potential iteratee object argument.
     * @returns {boolean} Returns `true` if the arguments are from an iteratee call,
     *  else `false`.
     */
    function isIterateeCall(value, index, object) {
      if (!isObject$3(object)) {
        return false;
      }
      var type = typeof index;
      if (type == 'number'
            ? (isArrayLike(object) && isIndex(index, object.length))
            : (type == 'string' && index in object)
          ) {
        return eq(object[index], value);
      }
      return false;
    }

    /**
     * Performs a
     * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * comparison between two values to determine if they are equivalent.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
     * @example
     *
     * var object = { 'a': 1 };
     * var other = { 'a': 1 };
     *
     * _.eq(object, object);
     * // => true
     *
     * _.eq(object, other);
     * // => false
     *
     * _.eq('a', 'a');
     * // => true
     *
     * _.eq('a', Object('a'));
     * // => false
     *
     * _.eq(NaN, NaN);
     * // => true
     */
    function eq(value, other) {
      return value === other || (value !== value && other !== other);
    }

    /**
     * Checks if `value` is array-like. A value is considered array-like if it's
     * not a function and has a `value.length` that's an integer greater than or
     * equal to `0` and less than or equal to `Number.MAX_SAFE_INTEGER`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is array-like, else `false`.
     * @example
     *
     * _.isArrayLike([1, 2, 3]);
     * // => true
     *
     * _.isArrayLike(document.body.children);
     * // => true
     *
     * _.isArrayLike('abc');
     * // => true
     *
     * _.isArrayLike(_.noop);
     * // => false
     */
    function isArrayLike(value) {
      return value != null && isLength(value.length) && !isFunction$4(value);
    }

    /**
     * Checks if `value` is classified as a `Function` object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a function, else `false`.
     * @example
     *
     * _.isFunction(_);
     * // => true
     *
     * _.isFunction(/abc/);
     * // => false
     */
    function isFunction$4(value) {
      // The use of `Object#toString` avoids issues with the `typeof` operator
      // in Safari 8-9 which returns 'object' for typed array and other constructors.
      var tag = isObject$3(value) ? objectToString$1.call(value) : '';
      return tag == funcTag || tag == genTag;
    }

    /**
     * Checks if `value` is a valid array-like length.
     *
     * **Note:** This method is loosely based on
     * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a valid length, else `false`.
     * @example
     *
     * _.isLength(3);
     * // => true
     *
     * _.isLength(Number.MIN_VALUE);
     * // => false
     *
     * _.isLength(Infinity);
     * // => false
     *
     * _.isLength('3');
     * // => false
     */
    function isLength(value) {
      return typeof value == 'number' &&
        value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER;
    }

    /**
     * Checks if `value` is the
     * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
     * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an object, else `false`.
     * @example
     *
     * _.isObject({});
     * // => true
     *
     * _.isObject([1, 2, 3]);
     * // => true
     *
     * _.isObject(_.noop);
     * // => true
     *
     * _.isObject(null);
     * // => false
     */
    function isObject$3(value) {
      var type = typeof value;
      return !!value && (type == 'object' || type == 'function');
    }

    /**
     * Checks if `value` is object-like. A value is object-like if it's not `null`
     * and has a `typeof` result of "object".
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
     * @example
     *
     * _.isObjectLike({});
     * // => true
     *
     * _.isObjectLike([1, 2, 3]);
     * // => true
     *
     * _.isObjectLike(_.noop);
     * // => false
     *
     * _.isObjectLike(null);
     * // => false
     */
    function isObjectLike(value) {
      return !!value && typeof value == 'object';
    }

    /**
     * Checks if `value` is classified as a `Symbol` primitive or object.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
     * @example
     *
     * _.isSymbol(Symbol.iterator);
     * // => true
     *
     * _.isSymbol('abc');
     * // => false
     */
    function isSymbol$1(value) {
      return typeof value == 'symbol' ||
        (isObjectLike(value) && objectToString$1.call(value) == symbolTag);
    }

    /**
     * Converts `value` to a finite number.
     *
     * @static
     * @memberOf _
     * @since 4.12.0
     * @category Lang
     * @param {*} value The value to convert.
     * @returns {number} Returns the converted number.
     * @example
     *
     * _.toFinite(3.2);
     * // => 3.2
     *
     * _.toFinite(Number.MIN_VALUE);
     * // => 5e-324
     *
     * _.toFinite(Infinity);
     * // => 1.7976931348623157e+308
     *
     * _.toFinite('3.2');
     * // => 3.2
     */
    function toFinite(value) {
      if (!value) {
        return value === 0 ? value : 0;
      }
      value = toNumber(value);
      if (value === INFINITY || value === -INFINITY) {
        var sign = (value < 0 ? -1 : 1);
        return sign * MAX_INTEGER;
      }
      return value === value ? value : 0;
    }

    /**
     * Converts `value` to a number.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to process.
     * @returns {number} Returns the number.
     * @example
     *
     * _.toNumber(3.2);
     * // => 3.2
     *
     * _.toNumber(Number.MIN_VALUE);
     * // => 5e-324
     *
     * _.toNumber(Infinity);
     * // => Infinity
     *
     * _.toNumber('3.2');
     * // => 3.2
     */
    function toNumber(value) {
      if (typeof value == 'number') {
        return value;
      }
      if (isSymbol$1(value)) {
        return NAN;
      }
      if (isObject$3(value)) {
        var other = typeof value.valueOf == 'function' ? value.valueOf() : value;
        value = isObject$3(other) ? (other + '') : other;
      }
      if (typeof value != 'string') {
        return value === 0 ? value : +value;
      }
      value = value.replace(reTrim, '');
      var isBinary = reIsBinary.test(value);
      return (isBinary || reIsOctal.test(value))
        ? freeParseInt(value.slice(2), isBinary ? 2 : 8)
        : (reIsBadHex.test(value) ? NAN : +value);
    }

    /**
     * Creates an array of numbers (positive and/or negative) progressing from
     * `start` up to, but not including, `end`. A step of `-1` is used if a negative
     * `start` is specified without an `end` or `step`. If `end` is not specified,
     * it's set to `start` with `start` then set to `0`.
     *
     * **Note:** JavaScript follows the IEEE-754 standard for resolving
     * floating-point values which can produce unexpected results.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Util
     * @param {number} [start=0] The start of the range.
     * @param {number} end The end of the range.
     * @param {number} [step=1] The value to increment or decrement by.
     * @returns {Array} Returns the range of numbers.
     * @see _.inRange, _.rangeRight
     * @example
     *
     * _.range(4);
     * // => [0, 1, 2, 3]
     *
     * _.range(-4);
     * // => [0, -1, -2, -3]
     *
     * _.range(1, 5);
     * // => [1, 2, 3, 4]
     *
     * _.range(0, 20, 5);
     * // => [0, 5, 10, 15]
     *
     * _.range(0, -4, -1);
     * // => [0, -1, -2, -3]
     *
     * _.range(1, 4, 0);
     * // => [1, 1, 1]
     *
     * _.range(0);
     * // => []
     */
    var range = createRange();

    var lodash_range = range;

    /**
     * Long Short Term Memory Time Series with Tensorflow
     * @class LSTMTimeSeries
     * @implements {BaseNeuralNetwork}
     */
    class LSTMTimeSeries extends BaseNeuralNetwork {
      
      // settings: TensorScriptOptions;

      /**
       * Creates dataset data
       * @example
       * LSTMTimeSeries.createDataset([ [ 1, ], [ 2, ], [ 3, ], [ 4, ], [ 5, ], [ 6, ], [ 7, ], [ 8, ], [ 9, ], [ 10, ], ], 3) // => 
          //  [ 
          //    [ 
          //      [ [ 1 ], [ 2 ], [ 3 ] ],
          //      [ [ 2 ], [ 3 ], [ 4 ] ],
          //      [ [ 3 ], [ 4 ], [ 5 ] ],
          //      [ [ 4 ], [ 5 ], [ 6 ] ],
          //      [ [ 5 ], [ 6 ], [ 7 ] ],
          //      [ [ 6 ], [ 7 ], [ 8 ] ], 
          //   ], //x_matrix
          //   [ [ 4 ], [ 5 ], [ 6 ], [ 7 ], [ 8 ], [ 9 ] ] //y_matrix
          // ]
       * @param {Array<Array<number>} dataset - array of values
       * @param {Number} look_back - number of values in each feature 
       * @return {[Array<Array<number>>,Array<number>]} returns x matrix and y matrix for model trainning
       */
      /* istanbul ignore next */
      static createDataset(dataset=[], look_back = 1) { 
        const dataX = new Array();
        const dataY = new Array();
        for (let index in lodash_range(dataset.length - look_back - 1)) {
          let i = parseInt(index);
          let a = dataset.slice(i, i + look_back);
          dataX.push(a);
          dataY.push(dataset[ i + look_back ]);
        }
        return [dataX, dataY, ];
      }
      /**
       * Reshape input to be [samples, time steps, features]
       * @example
       * LSTMTimeSeries.getTimeseriesShape([ 
          [ [ 1 ], [ 2 ], [ 3 ] ],
          [ [ 2 ], [ 3 ], [ 4 ] ],
          [ [ 3 ], [ 4 ], [ 5 ] ],
          [ [ 4 ], [ 5 ], [ 6 ] ],
          [ [ 5 ], [ 6 ], [ 7 ] ],
          [ [ 6 ], [ 7 ], [ 8 ] ], 
        ]) //=> [6, 1, 3,]
       * @param {Array<Array<number>} x_timeseries - dataset array of values
       * @return {Array<Array<number>>} returns proper timeseries forecasting shape
       */
      static getTimeseriesShape( x_timeseries) {
        const time_steps = this.settings.timeSteps;
        const xShape = this.getInputShape(x_timeseries);
        //@ts-ignore
        const _samples = xShape[ 0 ];
        const _timeSteps = time_steps;
        //@ts-ignore
        const _features = xShape[ 1 ];
        const newShape = (this.settings.mulitpleTimeSteps || this.settings.stateful)
          ? [_samples,  _features, _timeSteps, ]
          : [ _samples, _timeSteps, _features, ];
        // console.log({newShape})
        return newShape;
      }
      /**
       * Returns data for predicting values
       * @param timeseries 
       * @param look_back 
       */
      static getTimeseriesDataSet( timeseries, look_back) {
        const lookback = look_back || this.settings.lookback;
        const matrices = LSTMTimeSeries.createDataset.call(this, timeseries, lookback);
        const x_matrix = matrices[ 0 ];
        const y_matrix = matrices[ 1 ];
        // const timeseriesShape = LSTMTimeSeries.getTimeseriesShape.call(this,x_matrix);
        //@ts-ignore
        const x_matrix_timeseries = BaseNeuralNetwork.reshape(x_matrix, [x_matrix.length, lookback, ]);
        const xShape = BaseNeuralNetwork.getInputShape(x_matrix_timeseries);
        const yShape = BaseNeuralNetwork.getInputShape(y_matrix);
        return {
          yShape,
          xShape,
          y_matrix,
          x_matrix:x_matrix_timeseries,
        };
      }
      
      
      
      /**
       * @param {{layers:Array<Object>,compile:Object,fit:Object}} options - neural network configuration and tensorflow model hyperparameters
       * @param {{model:Object,tf:Object,}} properties - extra instance properties
       */
      constructor(options = {}, properties) {
        const config = Object.assign({
          layers: [],
          type: 'simple',
          stateful:false,
          stacked: false,
          mulitpleTimeSteps:false,
          lookback:1,
          batchSize:1,
          timeSteps:1,
          learningRate:0.1,
          compile: {
            loss: 'meanSquaredError',
            optimizer: 'adam',
          },
          fit: {
            epochs: 100,
            batchSize: 1,
          },
        }, options);
        super(config, properties);
        this.type = 'LSTMTimeSeries';

        this.createDataset = LSTMTimeSeries.createDataset;
        this.getTimeseriesDataSet = LSTMTimeSeries.getTimeseriesDataSet;
        this.getTimeseriesShape = LSTMTimeSeries.getTimeseriesShape;
        return this;
      }
      /**
       * Adds dense layers to tensorflow classification model
       * @override 
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @param {Array<Object>} layers - model dense layer parameters
       * @param {Array<Array<number>>} x_test - validation data independent variables
       * @param {Array<Array<number>>} y_test - validation data dependent variables
       */
      generateLayers(x_matrix, y_matrix, layers) {
        const xShape = this.getInputShape(x_matrix);
        const yShape = this.getInputShape(y_matrix);
        this.yShape = yShape;
        this.xShape = xShape;
        // const sgdoptimizer = this.tf.train.sgd(this.settings.learningRate);
        const lstmLayers = [];
        const rnnLayers = [];
        const denseLayers = [];
        /* istanbul ignore next */
        if (layers) {
          if(layers.lstmLayers && layers.lstmLayers.length) lstmLayers.push(...layers.lstmLayers);
          if(layers.rnnLayers && layers.rnnLayers.length) rnnLayers.push(...layers.rnnLayers);
          if(layers.denseLayers && layers.denseLayers.length) denseLayers.push(...layers.denseLayers);
        } else if (this.settings && this.settings.fit && this.settings.stateful) {
          const batchInputShape = [this.settings.fit.batchSize, this.settings.lookback, 1, ];
          rnnLayers.push({ units: 4, batchInputShape:batchInputShape,  returnSequences:true, });
          rnnLayers.push({ units: 4, batchInputShape:batchInputShape,  });
          denseLayers.push({ units: yShape[1], });
        // } else if(this.settings.stacked) {
        //   lstmLayers.push({ units: 4, inputShape: [ 1, this.settings.lookback ], });
        //   // model.add(LSTM(4, batch_input_shape=(batch_size, look_back, 1), stateful=True, return_sequences=True))
        //   // model.add(LSTM(4, batch_input_shape=(batch_size, look_back, 1), stateful=True))
        //   denseLayers.push({ units: yShape[1], });
        } else {
          const inputShape= [1, this.settings.lookback, ];
          // console.log('default timeseries', { inputShape, xShape, yShape ,  });
          lstmLayers.push({ units:4, inputShape,  });
          denseLayers.push({ units: yShape[1], });
        }
        // console.log('lstmLayers',lstmLayers)
        // console.log('denseLayers',denseLayers)
        if (lstmLayers.length) {
          lstmLayers.forEach(layer => {
            this.model.add(this.tf.layers.lstm(layer));
          });
        }
        if (rnnLayers.length) {
          /* istanbul ignore next */
          rnnLayers.forEach(layer => {
            this.model.add(this.tf.layers.simpleRNN(layer));
          });
        }
        if (denseLayers.length) {
          denseLayers.forEach(layer => {
            this.model.add(this.tf.layers.dense(layer));
          });
        }
        this.layers = {
          lstmLayers,
          rnnLayers,
          denseLayers,
        };
        // this.settings.compile.optimizer = sgdoptimizer;
      }
      async train(x_timeseries, y_timeseries, layers, x_test, y_test) {
        let yShape;
        let x_matrix;
        let y_matrix;
        const look_back = this.settings.lookback;
        if (y_timeseries) {
          x_matrix = x_timeseries;
          y_matrix = y_timeseries;
        } else {
          const matrices = this.createDataset(x_timeseries, look_back);
          x_matrix = matrices[ 0 ];
          y_matrix = matrices[ 1 ];
          yShape = this.getInputShape(y_matrix);
        }
        //_samples, _timeSteps, _features
        const timeseriesShape = this.getTimeseriesShape(x_matrix);
        const x_matrix_timeseries = BaseNeuralNetwork.reshape(x_matrix, timeseriesShape);
        const xs = this.tf.tensor(x_matrix_timeseries, timeseriesShape);
        const ys = this.tf.tensor(y_matrix, yShape);
        this.xShape = timeseriesShape;
        this.yShape = yShape;
        if (this.compiled === false) {
          this.model = this.tf.sequential();
          //@ts-ignore
          this.generateLayers.call(this, x_matrix_timeseries, y_matrix, layers || this.layers,
            // x_test, y_test
          );
          this.model.compile(this.settings.compile);
          if (this.settings.fit && this.settings.stateful) {
            this.settings.fit.shuffle = false;
          }
          this.compiled = true;
        }
        await this.model.fit(xs, ys, this.settings.fit);
        this.trained = true;

        // this.model.summary();
        xs.dispose();
        ys.dispose();
        return this.model;
      }
      calculate(x_matrix) {
        const timeseriesShape = this.getTimeseriesShape(x_matrix);
        //@ts-ignore
        const input_matrix = BaseNeuralNetwork.reshape(x_matrix, timeseriesShape);
        return super.calculate(input_matrix);
      }
      async predict(input_matrix, options = {}) {
        if (this.settings.stateful && input_matrix.length > 1) {
          //@ts-ignore
          return Promise.all(input_matrix.map((input)=>super.predict([input, ], options))) ;
        } else {
          return super.predict.call(this,input_matrix, options);
        }
      }
    }

    /**
     * Long Short Term Memory Multivariate Time Series with Tensorflow
     * @class LSTMMultivariateTimeSeries
     * @extends {LSTMTimeSeries}
     */
    //@ts-ignore
    class LSTMMultivariateTimeSeries extends LSTMTimeSeries {
      /**
       * Creates dataset data
       * @example
       * const ds = [
      [10, 20, 30, 40, 50, 60, 70, 80, 90,],
      [11, 21, 31, 41, 51, 61, 71, 81, 91,],
      [12, 22, 32, 42, 52, 62, 72, 82, 92,],
      [13, 23, 33, 43, 53, 63, 73, 83, 93,],
      [14, 24, 34, 44, 54, 64, 74, 84, 94,],
      [15, 25, 35, 45, 55, 65, 75, 85, 95,],
      [16, 26, 36, 46, 56, 66, 76, 86, 96,],
      [17, 27, 37, 47, 57, 67, 77, 87, 97,],
      [18, 28, 38, 48, 58, 68, 78, 88, 98,],
      [19, 29, 39, 49, 59, 69, 79, 89, 99,],
    ];
       * LSTMMultivariateTimeSeries.createDataset(ds,1) // => 
          //  [ 
          //   [ 
          //    [ 20, 30, 40, 50, 60, 70, 80, 90 ],
          //    [ 21, 31, 41, 51, 61, 71, 81, 91 ],
          //    [ 22, 32, 42, 52, 62, 72, 82, 92 ],
          //    [ 23, 33, 43, 53, 63, 73, 83, 93 ],
          //    [ 24, 34, 44, 54, 64, 74, 84, 94 ],
          //    [ 25, 35, 45, 55, 65, 75, 85, 95 ],
          //    [ 26, 36, 46, 56, 66, 76, 86, 96 ],
          //    [ 27, 37, 47, 57, 67, 77, 87, 97 ],
          //    [ 28, 38, 48, 58, 68, 78, 88, 98 ]    
          //   ], //x_matrix
          //   [ 11, 12, 13, 14, 15, 16, 17, 18, 19 ], //y_matrix
          //   8 //features
          // ]
       * @param {Array<Array<number>} dataset - array of values
       * @param {Number} look_back - number of values in each feature 
       * @override 
       * @return {[Array<Array<number>>,Array<number>]} returns x matrix and y matrix for model trainning
       */
      /* istanbul ignore next */
      static createDataset( dataset = [], look_back = 1) { 
        const features = (this.settings && this.settings.features) ? this.settings.features : dataset[ 0 ].length - 1;
        const n_in = look_back || this.settings.lookback ||1; //lookbacks
        const n_out = (this.settings && this.settings.outputs) ? this.settings.outputs : 1; //1;
        
        //@ts-ignore
        const series = LSTMMultivariateTimeSeries.seriesToSupervised(dataset, n_in, n_out);
        const dropped = LSTMMultivariateTimeSeries.getDropableColumns(features, n_in, n_out);
        //@ts-ignore
        const droppedColumns = LSTMMultivariateTimeSeries.drop(series, dropped);
        const original_dropped_c_columns = [ 0, droppedColumns[ 0 ].length - 1, ];
        
        // console.log({ series, dropped_c_columns,original_dropped_c_columns, dropped, });
        const y = pivotVector(droppedColumns)[ droppedColumns[0].length - 1 ];
        const x = LSTMMultivariateTimeSeries.drop(
          droppedColumns,
          original_dropped_c_columns,
          // [ 0, droppedColumns[ 0 ].length - 1, ]
        );
        //@ts-ignore
        return [ x, y, features, ];
      }
      /**
       * Drops columns by array index
       * @example
    const data = [ [ 10, 20, 30, 40, 50, 60, 70, 80, 90, 11, 21, 31, 41, 51, 61, 71, 81, 91 ],
         [ 11, 21, 31, 41, 51, 61, 71, 81, 91, 12, 22, 32, 42, 52, 62, 72, 82, 92 ],
         [ 12, 22, 32, 42, 52, 62, 72, 82, 92, 13, 23, 33, 43, 53, 63, 73, 83, 93 ],
         [ 13, 23, 33, 43, 53, 63, 73, 83, 93, 14, 24, 34, 44, 54, 64, 74, 84, 94 ],
         [ 14, 24, 34, 44, 54, 64, 74, 84, 94, 15, 25, 35, 45, 55, 65, 75, 85, 95 ],
         [ 15, 25, 35, 45, 55, 65, 75, 85, 95, 16, 26, 36, 46, 56, 66, 76, 86, 96 ],
         [ 16, 26, 36, 46, 56, 66, 76, 86, 96, 17, 27, 37, 47, 57, 67, 77, 87, 97 ],
         [ 17, 27, 37, 47, 57, 67, 77, 87, 97, 18, 28, 38, 48, 58, 68, 78, 88, 98 ],
         [ 18, 28, 38, 48, 58, 68, 78, 88, 98, 19, 29, 39, 49, 59, 69, 79, 89, 99 ] ];
    const n_in = 1; //lookbacks
    const n_out = 1;
    const dropColumns = getDropableColumns(8, n_in, n_out); // =>[ 10, 11, 12, 13, 14, 15, 16, 17 ]
    const newdata = drop(data,dropColumns); //=> [ 
        // [ 10, 20, 30, 40, 50, 60, 70, 80, 90, 11 ],
        // [ 11, 21, 31, 41, 51, 61, 71, 81, 91, 12 ],
        // [ 12, 22, 32, 42, 52, 62, 72, 82, 92, 13 ],
        // [ 13, 23, 33, 43, 53, 63, 73, 83, 93, 14 ],
        // [ 14, 24, 34, 44, 54, 64, 74, 84, 94, 15 ],
        // [ 15, 25, 35, 45, 55, 65, 75, 85, 95, 16 ],
        // [ 16, 26, 36, 46, 56, 66, 76, 86, 96, 17 ],
        // [ 17, 27, 37, 47, 57, 67, 77, 87, 97, 18 ],
        // [ 18, 28, 38, 48, 58, 68, 78, 88, 98, 19 ] 
        //]
      * @param {Array<Array<number>>} data - data set to drop columns 
      * @param {Array<number>} columns - array of column indexes
       * @returns {Array<Array<number>>} matrix with dropped columns
       */
      static drop(data, columns) {
        return data.reduce((cols, row, i) => { 
          cols[ i ] = [];
          row.forEach((col, idx) => {
            if (columns.indexOf(idx)===-1) {
              cols[ i ].push(col);
            }
          });
          return cols;
        }, []);
      }
      /**
       * Converts data set to supervised labels for forecasting, the first column must be the dependent variable
       * @example 
       const ds = [
        [10, 20, 30, 40, 50, 60, 70, 80, 90,],
        [11, 21, 31, 41, 51, 61, 71, 81, 91,],
        [12, 22, 32, 42, 52, 62, 72, 82, 92,],
        [13, 23, 33, 43, 53, 63, 73, 83, 93,],
        [14, 24, 34, 44, 54, 64, 74, 84, 94,],
        [15, 25, 35, 45, 55, 65, 75, 85, 95,],
        [16, 26, 36, 46, 56, 66, 76, 86, 96,],
        [17, 27, 37, 47, 57, 67, 77, 87, 97,],
        [18, 28, 38, 48, 58, 68, 78, 88, 98,],
        [19, 29, 39, 49, 59, 69, 79, 89, 99,],
      ]; 
      const n_in = 1; //lookbacks
      const n_out = 1;
      const series = seriesToSupervised(ds, n_in, n_out); //=> [ 
        // [ 10, 20, 30, 40, 50, 60, 70, 80, 90, 11, 21, 31, 41, 51, 61, 71, 81, 91 ],
        // [ 11, 21, 31, 41, 51, 61, 71, 81, 91, 12, 22, 32, 42, 52, 62, 72, 82, 92 ],
        // [ 12, 22, 32, 42, 52, 62, 72, 82, 92, 13, 23, 33, 43, 53, 63, 73, 83, 93 ],
        // [ 13, 23, 33, 43, 53, 63, 73, 83, 93, 14, 24, 34, 44, 54, 64, 74, 84, 94 ],
        // [ 14, 24, 34, 44, 54, 64, 74, 84, 94, 15, 25, 35, 45, 55, 65, 75, 85, 95 ],
        // [ 15, 25, 35, 45, 55, 65, 75, 85, 95, 16, 26, 36, 46, 56, 66, 76, 86, 96 ],
        // [ 16, 26, 36, 46, 56, 66, 76, 86, 96, 17, 27, 37, 47, 57, 67, 77, 87, 97 ],
        // [ 17, 27, 37, 47, 57, 67, 77, 87, 97, 18, 28, 38, 48, 58, 68, 78, 88, 98 ],
        // [ 18, 28, 38, 48, 58, 68, 78, 88, 98, 19, 29, 39, 49, 59, 69, 79, 89, 99 ] 
        //];
       * 
       * @param {Array<Array<number>>} data - data set 
       * @param {number} n_in - look backs 
       * @param {number} n_out - future iterations (only 1 supported) 
       * @todo support multiple future iterations
       * @returns {Array<Array<number>>} multivariate dataset for time series
       */
      static seriesToSupervised(data, n_in = 1, n_out = 1) {
        if (n_out !== 1) throw new RangeError('Only 1 future iteration supported');
        if (data && Array.isArray(data) && Array.isArray(data[0]) && data[ 0 ].length < 2) throw new RangeError('Must include at least two columns, the first column must be the dependent variable, followed by independent variables');
        // let n_vars = data[ 0 ].length;
        let cols = [];
        // let names = [];
        // input sequence (t-n, ... t-1)
        for (let x in data) {
          //@ts-ignore
          x = Number(parseInt(x));
          //@ts-ignore
          let maxIndex = x + n_in + n_out;
          if (maxIndex > data.length) break;
          //@ts-ignore
          cols[ x ] = [];
          // console.log({ x,maxIndex });
          for (let i in lodash_range(n_in)) {
            //@ts-ignore
            i = parseInt(i);
            //@ts-ignore
            cols[ x ].push(...data[x+i]);
            // console.log({ i, cols, });
          }
          for (let j in lodash_range(n_out)){
            //@ts-ignore
            j = parseInt(j);
            //@ts-ignore
            cols[ x ].push(...data[ x+j+n_in ]);
            // console.log({ j, cols, });
          }
        }
        return cols;
      }
      /**
       * Calculates which columns to drop by index
       * @todo support multiple iterations in the future, also only one output variable supported in column features * lookbacks -1
       * @example
    const ds = [
      [10, 20, 30, 40, 50, 60, 70, 80, 90,],
      [11, 21, 31, 41, 51, 61, 71, 81, 91,],
      [12, 22, 32, 42, 52, 62, 72, 82, 92,],
      [13, 23, 33, 43, 53, 63, 73, 83, 93,],
      [14, 24, 34, 44, 54, 64, 74, 84, 94,],
      [15, 25, 35, 45, 55, 65, 75, 85, 95,],
      [16, 26, 36, 46, 56, 66, 76, 86, 96,],
      [17, 27, 37, 47, 57, 67, 77, 87, 97,],
      [18, 28, 38, 48, 58, 68, 78, 88, 98,],
      [19, 29, 39, 49, 59, 69, 79, 89, 99,],
    ];
    const n_in = 1; //lookbacks
    const n_out = 1;
    const dropped = getDropableColumns(8, n_in, n_out); //=> [ 10, 11, 12, 13, 14, 15, 16, 17 ]
       * @param {number} features - number of independent variables
       * @param {number} n_in - look backs 
       * @param {number} n_out - future iterations (currently only 1 supported)
       * @returns {Array<number>} array indexes to drop
       */
      static getDropableColumns(features, n_in, n_out) {
        if (n_out !== 1) throw new RangeError('Only 1 future iteration supported');
        const cols = features + 1;
        const total_cols = cols * n_in + cols * n_out;
        // console.log({ cols, total_cols });
        return lodash_range(total_cols - cols +1, total_cols);
      }
      /**
       * Reshape input to be [samples, time steps, features]
       * @example
       * @override 
       * LSTMTimeSeries.getTimeseriesShape([ 
          [ [ 1 ], [ 2 ], [ 3 ] ],
          [ [ 2 ], [ 3 ], [ 4 ] ],
          [ [ 3 ], [ 4 ], [ 5 ] ],
          [ [ 4 ], [ 5 ], [ 6 ] ],
          [ [ 5 ], [ 6 ], [ 7 ] ],
          [ [ 6 ], [ 7 ], [ 8 ] ], 
        ]) //=> [6, 1, 3,]
       * @param {Array<Array<number>} x_timeseries - dataset array of values
       * @return {Array<Array<number>>} returns proper timeseries forecasting shape
       */
      static getTimeseriesShape( x_timeseries) {
        const time_steps = (this.settings && this.settings.lookback) ? this.settings.lookback : 1;
        const xShape = this.getInputShape(x_timeseries);
        const _samples = xShape[ 0 ];
        const _timeSteps = time_steps;
        const _features = (this.settings && this.settings.features) ? this.settings.features : xShape[ 1 ];
        // reshape input to be 3D [samples, timesteps, features]
        // train_X = train_X.reshape((train_X.shape[0], n_hours, n_features))
        const newShape = [ _samples, _timeSteps, _features, ];
        // console.log({newShape})
        return newShape;
      }
      /**
       * Returns data for predicting values
       * @param timeseries 
       * @param look_back 
       * @override 
       */
      static getTimeseriesDataSet( timeseries, look_back) {
        const lookBack = look_back || this.settings.lookback;
        const matrices = LSTMMultivariateTimeSeries.createDataset.call(this, timeseries, lookBack);
        const x_matrix = matrices[0];
        //@ts-ignore
        const y_matrix_m = LSTMMultivariateTimeSeries.reshape(matrices[ 1 ], [ matrices[ 1 ].length, 1 ]);
        //@ts-ignore
        const timeseriesShape = LSTMMultivariateTimeSeries.getTimeseriesShape.call(this,x_matrix);
        // const timeseriesShape = LSTMMultivariateTimeSeries.getTimeseriesShape(x_matrix);
        //@ts-ignore
        const x_matrix_timeseries = LSTMMultivariateTimeSeries.reshape(x_matrix, timeseriesShape);
        // const x_matrix_timeseries = LSTMMultivariateTimeSeries.reshape(x_matrix, [x_matrix.length, lookBack, ]);
        const xShape = LSTMMultivariateTimeSeries.getInputShape(x_matrix_timeseries);
        //@ts-ignore
        const y_matrix = pivotVector(y_matrix_m)[ y_matrix_m[0].length - 1 ];

        const yShape = LSTMMultivariateTimeSeries.getInputShape(y_matrix_m);
        return {
          yShape,
          xShape,
          y_matrix,
          x_matrix: x_matrix_timeseries,
        };
      }
      
      
      
      
      
      
      /**
       * @param {{layers:Array<Object>,compile:Object,fit:Object}} options - neural network configuration and tensorflow model hyperparameters
       * @param {{model:Object,tf:Object,}} properties - extra instance properties
       */
      constructor(options = {}, properties) {
        const config = Object.assign({
          layers: [],
          stateful: true,
          mulitpleTimeSteps:false,
          lookback: 1,
          features: undefined,
          outputs:1,
          learningRate: 0.1,
          compile: {
            loss: 'meanSquaredError',
            optimizer: 'adam',
          },
          fit: {
            epochs: 100,
            batchSize: 1,
            shuffle: false,
          },
        }, options);
        super(config, properties);
        this.type = 'LSTMMultivariateTimeSeries';

        this.createDataset = LSTMMultivariateTimeSeries.createDataset;
        this.seriesToSupervised = LSTMMultivariateTimeSeries.seriesToSupervised;
        this.getDropableColumns = LSTMMultivariateTimeSeries.getDropableColumns;
        this.drop = LSTMMultivariateTimeSeries.drop;
        this.getTimeseriesShape = LSTMMultivariateTimeSeries.getTimeseriesShape;
        this.getTimeseriesDataSet = LSTMMultivariateTimeSeries.getTimeseriesDataSet;
        return this;
      }
      /**
       * Adds dense layers to tensorflow classification model
       * @override 
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @param {Array<Object>} layers - model dense layer parameters
       * @param {Array<Array<number>>} x_test - validation data independent variables
       * @param {Array<Array<number>>} y_test - validation data dependent variables
       */
      generateLayers(x_matrix, y_matrix, layers) {
        const xShape = this.getInputShape(x_matrix);
        const yShape = this.getInputShape(y_matrix);
        this.yShape = yShape;
        this.xShape = xShape;
        const lstmLayers = [];
        const denseLayers = [];
        if (layers) {
          if(layers.lstmLayers && layers.lstmLayers.length) lstmLayers.push(...layers.lstmLayers);
          if(layers.denseLayers && layers.denseLayers.length) denseLayers.push(...layers.denseLayers);
       
        } else {
          const inputShape = [ xShape[ 1 ], xShape[ 2 ], ];
          // console.log('default timeseries', { inputShape, xShape, yShape, });
          // model.add(LSTM(50, input_shape=(train_X.shape[1], train_X.shape[2])))
          // model.add(Dense(1))

          lstmLayers.push({ units: 10, inputShape,  });
          denseLayers.push({ units: yShape[1], });
        }
        // console.log('lstmLayers',lstmLayers)
        // console.log('denseLayers',denseLayers)
        if (lstmLayers.length) {
          lstmLayers.forEach(layer => {
            this.model.add(this.tf.layers.lstm(layer));
          });
        }
        if (denseLayers.length) {
          denseLayers.forEach(layer => {
            this.model.add(this.tf.layers.dense(layer));
          });
        }
        this.layers = {
          lstmLayers,
          denseLayers,
        };
        // this.settings.compile.optimizer = sgdoptimizer;
      }
      /**
       * @override 
       * @param x_timeseries 
       * @param y_timeseries 
       * @param layers 
       * @param x_test 
       * @param y_test 
       */
      async train(x_timeseries, y_timeseries, layers, x_test, y_test) {
        let yShape;
        let x_matrix;
        let y_matrix;
        const look_back = this.settings.lookback;
        if (y_timeseries) {
          x_matrix = x_timeseries;
          y_matrix = y_timeseries;
          this.getInputShape(x_matrix);
          yShape = this.getInputShape(y_matrix);
        } else {
          const matrices = this.createDataset(x_timeseries, look_back);
          // console.log({matrices})
          x_matrix = matrices[0];
          //@ts-ignore
          y_matrix = this.reshape(matrices[ 1 ], [ matrices[ 1 ].length, 1 ]);
          this.getInputShape(x_matrix);
          yShape = this.getInputShape(y_matrix);
        }
        //_samples, _timeSteps, _features
        const timeseriesShape = this.getTimeseriesShape(x_matrix);
        // console.log({
        //   timeseriesShape, yShape, xShape,
        //   // y_matrix, x_matrix,
        // });
        const x_matrix_timeseries = LSTMMultivariateTimeSeries.reshape(x_matrix, timeseriesShape);
        // console.log('x_matrix_timeseries',JSON.stringify(x_matrix_timeseries));
        // console.log('x_matrix',JSON.stringify(x_matrix));
        // console.log('y_matrix',JSON.stringify(y_matrix));
        const xs = this.tf.tensor(x_matrix_timeseries, timeseriesShape);
        const ys = this.tf.tensor(y_matrix, yShape);
        this.xShape = timeseriesShape;
        this.yShape = yShape;
        if (this.compiled===false) {
          this.model = this.tf.sequential();
          //@ts-ignore
          this.generateLayers.call(this, x_matrix_timeseries, y_matrix, layers || this.layers,
          //  x_test, y_test
          );
          this.model.compile(this.settings.compile);
          this.compiled = true;
        }
        if (x_test && y_test && this.settings && this.settings.fit) {
          this.settings.fit.validation_data = [ x_test, y_test, ];
        }
        await this.model.fit(xs, ys, this.settings.fit);
        this.trained = true;

        // this.model.summary();
        xs.dispose();
        ys.dispose();
        return this.model;
      }
    }

    /**
     * Returns an array of vectors as an array of arrays (from modelscript)
     * @example
    const vectors = [ [1,2,3], [1,2,3], [3,3,4], [3,3,3] ];
    const arrays = pivotVector(vectors); // => [ [1,2,3,3], [2,2,3,3], [3,3,4,3] ];
     * @memberOf util
     * @param {Array[]} vectors 
     * @returns {Array[]}
     * @ignore
     * @see {https://github.com/repetere/modelscript/blob/master/src/util.js}
     */
    /* istanbul ignore next */
    function pivotVector(vectors=[]) {
      /* istanbul ignore next */
      return vectors.reduce((result, val, index/*, arr*/) => {
        val.forEach((vecVal, i) => {
          (index === 0)
            ? (result.push([vecVal,]))
            : (result[ i ].push(vecVal));
        });
        return result;
      }, []);
    }

    var universalSentenceEncoder = {exports: {}};

    var require$$0$4 = /*@__PURE__*/getAugmentedNamespace(dist$1);

    var require$$1$1 = /*@__PURE__*/getAugmentedNamespace(dist);

    /**
        * @license
        * Copyright 2021 Google LLC. All Rights Reserved.
        * Licensed under the Apache License, Version 2.0 (the "License");
        * you may not use this file except in compliance with the License.
        * You may obtain a copy of the License at
        *
        * http://www.apache.org/licenses/LICENSE-2.0
        *
        * Unless required by applicable law or agreed to in writing, software
        * distributed under the License is distributed on an "AS IS" BASIS,
        * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
        * See the License for the specific language governing permissions and
        * limitations under the License.
        * =============================================================================
        */

    (function (module, exports) {
    	(function (global, factory) {
    	    factory(exports, require$$0$4, require$$1$1) ;
    	}(commonjsGlobal, (function (exports,tf,tfconv) {
    	    /*! *****************************************************************************
    	    Copyright (c) Microsoft Corporation.

    	    Permission to use, copy, modify, and/or distribute this software for any
    	    purpose with or without fee is hereby granted.

    	    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    	    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    	    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    	    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    	    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    	    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    	    PERFORMANCE OF THIS SOFTWARE.
    	    ***************************************************************************** */

    	    function __awaiter(thisArg, _arguments, P, generator) {
    	        function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    	        return new (P || (P = Promise))(function (resolve, reject) {
    	            function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
    	            function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
    	            function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
    	            step((generator = generator.apply(thisArg, _arguments || [])).next());
    	        });
    	    }

    	    function __generator(thisArg, body) {
    	        var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    	        return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    	        function verb(n) { return function (v) { return step([n, v]); }; }
    	        function step(op) {
    	            if (f) throw new TypeError("Generator is already executing.");
    	            while (_) try {
    	                if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
    	                if (y = 0, t) op = [op[0] & 2, t.value];
    	                switch (op[0]) {
    	                    case 0: case 1: t = op; break;
    	                    case 4: _.label++; return { value: op[1], done: false };
    	                    case 5: _.label++; y = op[1]; op = [0]; continue;
    	                    case 7: op = _.ops.pop(); _.trys.pop(); continue;
    	                    default:
    	                        if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
    	                        if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
    	                        if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
    	                        if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
    	                        if (t[2]) _.ops.pop();
    	                        _.trys.pop(); continue;
    	                }
    	                op = body.call(thisArg, _);
    	            } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
    	            if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    	        }
    	    }

    	    /**
    	     * @license
    	     * Copyright 2019 Google LLC. All Rights Reserved.
    	     * Licensed under the Apache License, Version 2.0 (the "License");
    	     * you may not use this file except in compliance with the License.
    	     * You may obtain a copy of the License at
    	     *
    	     * http://www.apache.org/licenses/LICENSE-2.0
    	     *
    	     * Unless required by applicable law or agreed to in writing, software
    	     * distributed under the License is distributed on an "AS IS" BASIS,
    	     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    	     * See the License for the specific language governing permissions and
    	     * limitations under the License.
    	     * =============================================================================
    	     */
    	    // unicode-aware iteration
    	    var stringToChars = function (input) {
    	        var symbols = [];
    	        for (var _i = 0, input_1 = input; _i < input_1.length; _i++) {
    	            var symbol = input_1[_i];
    	            symbols.push(symbol);
    	        }
    	        return symbols;
    	    };

    	    /**
    	     * @license
    	     * Copyright 2019 Google LLC. All Rights Reserved.
    	     * Licensed under the Apache License, Version 2.0 (the "License");
    	     * you may not use this file except in compliance with the License.
    	     * You may obtain a copy of the License at
    	     *
    	     * http://www.apache.org/licenses/LICENSE-2.0
    	     *
    	     * Unless required by applicable law or agreed to in writing, software
    	     * distributed under the License is distributed on an "AS IS" BASIS,
    	     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    	     * See the License for the specific language governing permissions and
    	     * limitations under the License.
    	     * =============================================================================
    	     */
    	    var TrieNode = /** @class */ (function () {
    	        function TrieNode() {
    	            this.parent = null;
    	            this.children = {};
    	            this.end = false;
    	            this.word = [[], 0, 0];
    	        }
    	        return TrieNode;
    	    }());
    	    var Trie = /** @class */ (function () {
    	        function Trie() {
    	            this.root = new TrieNode();
    	        }
    	        /**
    	         * Inserts a token into the trie.
    	         */
    	        Trie.prototype.insert = function (word, score, index) {
    	            var node = this.root;
    	            var symbols = stringToChars(word);
    	            for (var i = 0; i < symbols.length; i++) {
    	                if (!node.children[symbols[i]]) {
    	                    node.children[symbols[i]] = new TrieNode();
    	                    node.children[symbols[i]].parent = node;
    	                    node.children[symbols[i]].word[0] = node.word[0].concat(symbols[i]);
    	                }
    	                node = node.children[symbols[i]];
    	                if (i === symbols.length - 1) {
    	                    node.end = true;
    	                    node.word[1] = score;
    	                    node.word[2] = index;
    	                }
    	            }
    	        };
    	        /**
    	         * Returns an array of all tokens starting with ss.
    	         *
    	         * @param ss The prefix to match on.
    	         */
    	        Trie.prototype.commonPrefixSearch = function (ss) {
    	            var output = [];
    	            var node = this.root.children[ss[0]];
    	            for (var i = 0; i < ss.length && node; i++) {
    	                if (node.end) {
    	                    output.push(node.word);
    	                }
    	                node = node.children[ss[i + 1]];
    	            }
    	            if (!output.length) {
    	                output.push([[ss[0]], 0, 0]);
    	            }
    	            return output;
    	        };
    	        return Trie;
    	    }());

    	    /**
    	     * @license
    	     * Copyright 2019 Google LLC. All Rights Reserved.
    	     * Licensed under the Apache License, Version 2.0 (the "License");
    	     * you may not use this file except in compliance with the License.
    	     * You may obtain a copy of the License at
    	     *
    	     * http://www.apache.org/licenses/LICENSE-2.0
    	     *
    	     * Unless required by applicable law or agreed to in writing, software
    	     * distributed under the License is distributed on an "AS IS" BASIS,
    	     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    	     * See the License for the specific language governing permissions and
    	     * limitations under the License.
    	     * =============================================================================
    	     */
    	    var separator = '\u2581'; // This is the unicode character 'lower one eighth block'.
    	    function processInput(str) {
    	        var normalized = str.normalize('NFKC');
    	        return normalized.length > 0 ?
    	            separator + normalized.replace(/ /g, separator) :
    	            normalized;
    	    }
    	    // The first tokens are reserved for unk, control symbols, and user-defined
    	    // symbols.
    	    var RESERVED_SYMBOLS_COUNT = 6;
    	    var Tokenizer = /** @class */ (function () {
    	        function Tokenizer(vocabulary, reservedSymbolsCount) {
    	            if (reservedSymbolsCount === void 0) { reservedSymbolsCount = RESERVED_SYMBOLS_COUNT; }
    	            this.vocabulary = vocabulary;
    	            this.reservedSymbolsCount = reservedSymbolsCount;
    	            this.trie = new Trie();
    	            for (var i = this.reservedSymbolsCount; i < this.vocabulary.length; i++) {
    	                this.trie.insert(this.vocabulary[i][0], this.vocabulary[i][1], i);
    	            }
    	        }
    	        Tokenizer.prototype.encode = function (input) {
    	            var nodes = [];
    	            var words = [];
    	            var best = [];
    	            input = processInput(input);
    	            var symbols = stringToChars(input);
    	            for (var i = 0; i <= symbols.length; i++) {
    	                nodes.push({});
    	                words.push(0);
    	                best.push(0);
    	            }
    	            // Construct the lattice.
    	            for (var i = 0; i < symbols.length; i++) {
    	                var matches = this.trie.commonPrefixSearch(symbols.slice(i));
    	                for (var j = 0; j < matches.length; j++) {
    	                    var piece = matches[j];
    	                    var obj = { key: piece[0], score: piece[1], index: piece[2] };
    	                    var endPos = piece[0].length;
    	                    if (nodes[i + endPos][i] == null) {
    	                        nodes[i + endPos][i] = [];
    	                    }
    	                    nodes[i + endPos][i].push(obj);
    	                }
    	            }
    	            for (var endPos = 0; endPos <= symbols.length; endPos++) {
    	                for (var startPos in nodes[endPos]) {
    	                    var arr = nodes[endPos][startPos];
    	                    for (var j = 0; j < arr.length; j++) {
    	                        var word = arr[j];
    	                        var score = word.score + best[endPos - word.key.length];
    	                        if (best[endPos] === 0 || score >= best[endPos]) {
    	                            best[endPos] = score;
    	                            words[endPos] = arr[j].index;
    	                        }
    	                    }
    	                }
    	            }
    	            var results = [];
    	            // Backward pass.
    	            var iter = words.length - 1;
    	            while (iter > 0) {
    	                results.push(words[iter]);
    	                iter -= this.vocabulary[words[iter]][0].length;
    	            }
    	            // Merge consecutive unks.
    	            var merged = [];
    	            var isPreviousUnk = false;
    	            for (var i = 0; i < results.length; i++) {
    	                var id = results[i];
    	                if (!(isPreviousUnk && id === 0)) {
    	                    merged.push(id);
    	                }
    	                isPreviousUnk = id === 0;
    	            }
    	            return merged.reverse();
    	        };
    	        return Tokenizer;
    	    }());
    	    /**
    	     * Load the Tokenizer for use independently from the UniversalSentenceEncoder.
    	     *
    	     * @param pathToVocabulary (optional) Provide a path to the vocabulary file.
    	     */
    	    function loadTokenizer(pathToVocabulary) {
    	        return __awaiter(this, void 0, void 0, function () {
    	            var vocabulary, tokenizer;
    	            return __generator(this, function (_a) {
    	                switch (_a.label) {
    	                    case 0: return [4 /*yield*/, loadVocabulary(pathToVocabulary)];
    	                    case 1:
    	                        vocabulary = _a.sent();
    	                        tokenizer = new Tokenizer(vocabulary);
    	                        return [2 /*return*/, tokenizer];
    	                }
    	            });
    	        });
    	    }
    	    /**
    	     * Load a vocabulary for the Tokenizer.
    	     *
    	     * @param pathToVocabulary Defaults to the path to the 8k vocabulary used by the
    	     * UniversalSentenceEncoder.
    	     */
    	    function loadVocabulary(pathToVocabulary) {
    	        return __awaiter(this, void 0, void 0, function () {
    	            var vocabulary;
    	            return __generator(this, function (_a) {
    	                switch (_a.label) {
    	                    case 0: return [4 /*yield*/, tf.util.fetch(pathToVocabulary)];
    	                    case 1:
    	                        vocabulary = _a.sent();
    	                        return [2 /*return*/, vocabulary.json()];
    	                }
    	            });
    	        });
    	    }

    	    /** @license See the LICENSE file. */
    	    // This code is auto-generated, do not modify this file!
    	    var version = '1.3.3';

    	    /**
    	     * @license
    	     * Copyright 2020 Google LLC. All Rights Reserved.
    	     * Licensed under the Apache License, Version 2.0 (the 'License');
    	     * you may not use this file except in compliance with the License.
    	     * You may obtain a copy of the License at
    	     *
    	     * http://www.apache.org/licenses/LICENSE-2.0
    	     *
    	     * Unless required by applicable law or agreed to in writing, software
    	     * distributed under the License is distributed on an 'AS IS' BASIS,
    	     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    	     * See the License for the specific language governing permissions and
    	     * limitations under the License.
    	     * =============================================================================
    	     */
    	    var BASE_PATH = 'https://tfhub.dev/google/tfjs-model/universal-sentence-encoder-qa-ondevice/1';
    	    // Index in the vocab file that needs to be skipped.
    	    var SKIP_VALUES = [0, 1, 2];
    	    // Offset value for skipped vocab index.
    	    var OFFSET = 3;
    	    // Input tensor size limit.
    	    var INPUT_LIMIT = 192;
    	    // Model node name for query.
    	    var QUERY_NODE_NAME = 'input_inp_text';
    	    // Model node name for query.
    	    var RESPONSE_CONTEXT_NODE_NAME = 'input_res_context';
    	    // Model node name for response.
    	    var RESPONSE_NODE_NAME = 'input_res_text';
    	    // Model node name for response result.
    	    var RESPONSE_RESULT_NODE_NAME = 'Final/EncodeResult/mul';
    	    // Model node name for query result.
    	    var QUERY_RESULT_NODE_NAME = 'Final/EncodeQuery/mul';
    	    // Reserved symbol count for tokenizer.
    	    var RESERVED_SYMBOLS_COUNT$1 = 3;
    	    // Value for token padding
    	    var TOKEN_PADDING = 2;
    	    // Start value for each token
    	    var TOKEN_START_VALUE = 1;
    	    function loadQnA() {
    	        return __awaiter(this, void 0, void 0, function () {
    	            var use;
    	            return __generator(this, function (_a) {
    	                switch (_a.label) {
    	                    case 0:
    	                        use = new UniversalSentenceEncoderQnA();
    	                        return [4 /*yield*/, use.load()];
    	                    case 1:
    	                        _a.sent();
    	                        return [2 /*return*/, use];
    	                }
    	            });
    	        });
    	    }
    	    var UniversalSentenceEncoderQnA = /** @class */ (function () {
    	        function UniversalSentenceEncoderQnA() {
    	        }
    	        UniversalSentenceEncoderQnA.prototype.loadModel = function () {
    	            return __awaiter(this, void 0, void 0, function () {
    	                return __generator(this, function (_a) {
    	                    return [2 /*return*/, tfconv.loadGraphModel(BASE_PATH, { fromTFHub: true })];
    	                });
    	            });
    	        };
    	        UniversalSentenceEncoderQnA.prototype.load = function () {
    	            return __awaiter(this, void 0, void 0, function () {
    	                var _a, model, vocabulary;
    	                return __generator(this, function (_b) {
    	                    switch (_b.label) {
    	                        case 0: return [4 /*yield*/, Promise.all([
    	                                this.loadModel(),
    	                                loadVocabulary(BASE_PATH + "/vocab.json?tfjs-format=file")
    	                            ])];
    	                        case 1:
    	                            _a = _b.sent(), model = _a[0], vocabulary = _a[1];
    	                            this.model = model;
    	                            this.tokenizer = new Tokenizer(vocabulary, RESERVED_SYMBOLS_COUNT$1);
    	                            return [2 /*return*/];
    	                    }
    	                });
    	            });
    	        };
    	        /**
    	         *
    	         * Returns a map of queryEmbedding and responseEmbedding
    	         *
    	         * @param input the ModelInput that contains queries and answers.
    	         */
    	        UniversalSentenceEncoderQnA.prototype.embed = function (input) {
    	            var _this = this;
    	            var embeddings = tf.tidy(function () {
    	                var queryEncoding = _this.tokenizeStrings(input.queries, INPUT_LIMIT);
    	                var responseEncoding = _this.tokenizeStrings(input.responses, INPUT_LIMIT);
    	                if (input.contexts != null) {
    	                    if (input.contexts.length !== input.responses.length) {
    	                        throw new Error('The length of response strings ' +
    	                            'and context strings need to match.');
    	                    }
    	                }
    	                var contexts = input.contexts || [];
    	                if (input.contexts == null) {
    	                    contexts.length = input.responses.length;
    	                    contexts.fill('');
    	                }
    	                var contextEncoding = _this.tokenizeStrings(contexts, INPUT_LIMIT);
    	                var modelInputs = {};
    	                modelInputs[QUERY_NODE_NAME] = queryEncoding;
    	                modelInputs[RESPONSE_NODE_NAME] = responseEncoding;
    	                modelInputs[RESPONSE_CONTEXT_NODE_NAME] = contextEncoding;
    	                return _this.model.execute(modelInputs, [QUERY_RESULT_NODE_NAME, RESPONSE_RESULT_NODE_NAME]);
    	            });
    	            var queryEmbedding = embeddings[0];
    	            var responseEmbedding = embeddings[1];
    	            return { queryEmbedding: queryEmbedding, responseEmbedding: responseEmbedding };
    	        };
    	        UniversalSentenceEncoderQnA.prototype.tokenizeStrings = function (strs, limit) {
    	            var _this = this;
    	            var tokens = strs.map(function (s) { return _this.shiftTokens(_this.tokenizer.encode(s), INPUT_LIMIT); });
    	            return tf.tensor2d(tokens, [strs.length, INPUT_LIMIT], 'int32');
    	        };
    	        UniversalSentenceEncoderQnA.prototype.shiftTokens = function (tokens, limit) {
    	            tokens.unshift(TOKEN_START_VALUE);
    	            for (var index = 0; index < limit; index++) {
    	                if (index >= tokens.length) {
    	                    tokens[index] = TOKEN_PADDING;
    	                }
    	                else if (!SKIP_VALUES.includes(tokens[index])) {
    	                    tokens[index] += OFFSET;
    	                }
    	            }
    	            return tokens.slice(0, limit);
    	        };
    	        return UniversalSentenceEncoderQnA;
    	    }());

    	    /**
    	     * @license
    	     * Copyright 2019 Google LLC. All Rights Reserved.
    	     * Licensed under the Apache License, Version 2.0 (the "License");
    	     * you may not use this file except in compliance with the License.
    	     * You may obtain a copy of the License at
    	     *
    	     * http://www.apache.org/licenses/LICENSE-2.0
    	     *
    	     * Unless required by applicable law or agreed to in writing, software
    	     * distributed under the License is distributed on an "AS IS" BASIS,
    	     * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    	     * See the License for the specific language governing permissions and
    	     * limitations under the License.
    	     * =============================================================================
    	     */
    	    var BASE_PATH$1 = 'https://storage.googleapis.com/tfjs-models/savedmodel/universal_sentence_encoder';
    	    function load(config) {
    	        return __awaiter(this, void 0, void 0, function () {
    	            var use;
    	            return __generator(this, function (_a) {
    	                switch (_a.label) {
    	                    case 0:
    	                        use = new UniversalSentenceEncoder();
    	                        return [4 /*yield*/, use.load(config)];
    	                    case 1:
    	                        _a.sent();
    	                        return [2 /*return*/, use];
    	                }
    	            });
    	        });
    	    }
    	    var UniversalSentenceEncoder = /** @class */ (function () {
    	        function UniversalSentenceEncoder() {
    	        }
    	        UniversalSentenceEncoder.prototype.loadModel = function (modelUrl) {
    	            return __awaiter(this, void 0, void 0, function () {
    	                return __generator(this, function (_a) {
    	                    return [2 /*return*/, modelUrl
    	                            ? tfconv.loadGraphModel(modelUrl)
    	                            : tfconv.loadGraphModel('https://tfhub.dev/tensorflow/tfjs-model/universal-sentence-encoder-lite/1/default/1', { fromTFHub: true })];
    	                });
    	            });
    	        };
    	        UniversalSentenceEncoder.prototype.load = function (config) {
    	            if (config === void 0) { config = {}; }
    	            return __awaiter(this, void 0, void 0, function () {
    	                var _a, model, vocabulary;
    	                return __generator(this, function (_b) {
    	                    switch (_b.label) {
    	                        case 0: return [4 /*yield*/, Promise.all([
    	                                this.loadModel(config.modelUrl),
    	                                loadVocabulary(config.vocabUrl || BASE_PATH$1 + "/vocab.json")
    	                            ])];
    	                        case 1:
    	                            _a = _b.sent(), model = _a[0], vocabulary = _a[1];
    	                            this.model = model;
    	                            this.tokenizer = new Tokenizer(vocabulary);
    	                            return [2 /*return*/];
    	                    }
    	                });
    	            });
    	        };
    	        /**
    	         *
    	         * Returns a 2D Tensor of shape [input.length, 512] that contains the
    	         * Universal Sentence Encoder embeddings for each input.
    	         *
    	         * @param inputs A string or an array of strings to embed.
    	         */
    	        UniversalSentenceEncoder.prototype.embed = function (inputs) {
    	            return __awaiter(this, void 0, void 0, function () {
    	                var encodings, indicesArr, flattenedIndicesArr, i, indices, values, modelInputs, embeddings;
    	                var _this = this;
    	                return __generator(this, function (_a) {
    	                    switch (_a.label) {
    	                        case 0:
    	                            if (typeof inputs === 'string') {
    	                                inputs = [inputs];
    	                            }
    	                            encodings = inputs.map(function (d) { return _this.tokenizer.encode(d); });
    	                            indicesArr = encodings.map(function (arr, i) { return arr.map(function (d, index) { return [i, index]; }); });
    	                            flattenedIndicesArr = [];
    	                            for (i = 0; i < indicesArr.length; i++) {
    	                                flattenedIndicesArr =
    	                                    flattenedIndicesArr.concat(indicesArr[i]);
    	                            }
    	                            indices = tf.tensor2d(flattenedIndicesArr, [flattenedIndicesArr.length, 2], 'int32');
    	                            values = tf.tensor1d(tf.util.flatten(encodings), 'int32');
    	                            modelInputs = { indices: indices, values: values };
    	                            return [4 /*yield*/, this.model.executeAsync(modelInputs)];
    	                        case 1:
    	                            embeddings = _a.sent();
    	                            indices.dispose();
    	                            values.dispose();
    	                            return [2 /*return*/, embeddings];
    	                    }
    	                });
    	            });
    	        };
    	        return UniversalSentenceEncoder;
    	    }());

    	    exports.load = load;
    	    exports.UniversalSentenceEncoder = UniversalSentenceEncoder;
    	    exports.Tokenizer = Tokenizer;
    	    exports.loadTokenizer = loadTokenizer;
    	    exports.loadQnA = loadQnA;
    	    exports.version = version;

    	    Object.defineProperty(exports, '__esModule', { value: true });

    	})));
    } (universalSentenceEncoder, universalSentenceEncoder.exports));

    var axios$3 = {exports: {}};

    var axios$2 = {exports: {}};

    var bind$2 = function bind(fn, thisArg) {
      return function wrap() {
        var args = new Array(arguments.length);
        for (var i = 0; i < args.length; i++) {
          args[i] = arguments[i];
        }
        return fn.apply(thisArg, args);
      };
    };

    var bind$1 = bind$2;

    // utils is a library of generic helper functions non-specific to axios

    var toString = Object.prototype.toString;

    // eslint-disable-next-line func-names
    var kindOf = (function(cache) {
      // eslint-disable-next-line func-names
      return function(thing) {
        var str = toString.call(thing);
        return cache[str] || (cache[str] = str.slice(8, -1).toLowerCase());
      };
    })(Object.create(null));

    function kindOfTest(type) {
      type = type.toLowerCase();
      return function isKindOf(thing) {
        return kindOf(thing) === type;
      };
    }

    /**
     * Determine if a value is an Array
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is an Array, otherwise false
     */
    function isArray$2(val) {
      return Array.isArray(val);
    }

    /**
     * Determine if a value is undefined
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if the value is undefined, otherwise false
     */
    function isUndefined$2(val) {
      return typeof val === 'undefined';
    }

    /**
     * Determine if a value is a Buffer
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Buffer, otherwise false
     */
    function isBuffer$1(val) {
      return val !== null && !isUndefined$2(val) && val.constructor !== null && !isUndefined$2(val.constructor)
        && typeof val.constructor.isBuffer === 'function' && val.constructor.isBuffer(val);
    }

    /**
     * Determine if a value is an ArrayBuffer
     *
     * @function
     * @param {Object} val The value to test
     * @returns {boolean} True if value is an ArrayBuffer, otherwise false
     */
    var isArrayBuffer = kindOfTest('ArrayBuffer');


    /**
     * Determine if a value is a view on an ArrayBuffer
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a view on an ArrayBuffer, otherwise false
     */
    function isArrayBufferView(val) {
      var result;
      if ((typeof ArrayBuffer !== 'undefined') && (ArrayBuffer.isView)) {
        result = ArrayBuffer.isView(val);
      } else {
        result = (val) && (val.buffer) && (isArrayBuffer(val.buffer));
      }
      return result;
    }

    /**
     * Determine if a value is a String
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a String, otherwise false
     */
    function isString$1(val) {
      return typeof val === 'string';
    }

    /**
     * Determine if a value is a Number
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Number, otherwise false
     */
    function isNumber$2(val) {
      return typeof val === 'number';
    }

    /**
     * Determine if a value is an Object
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is an Object, otherwise false
     */
    function isObject$2(val) {
      return val !== null && typeof val === 'object';
    }

    /**
     * Determine if a value is a plain Object
     *
     * @param {Object} val The value to test
     * @return {boolean} True if value is a plain Object, otherwise false
     */
    function isPlainObject(val) {
      if (kindOf(val) !== 'object') {
        return false;
      }

      var prototype = Object.getPrototypeOf(val);
      return prototype === null || prototype === Object.prototype;
    }

    /**
     * Determine if a value is a Date
     *
     * @function
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Date, otherwise false
     */
    var isDate$1 = kindOfTest('Date');

    /**
     * Determine if a value is a File
     *
     * @function
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a File, otherwise false
     */
    var isFile = kindOfTest('File');

    /**
     * Determine if a value is a Blob
     *
     * @function
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Blob, otherwise false
     */
    var isBlob = kindOfTest('Blob');

    /**
     * Determine if a value is a FileList
     *
     * @function
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a File, otherwise false
     */
    var isFileList = kindOfTest('FileList');

    /**
     * Determine if a value is a Function
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Function, otherwise false
     */
    function isFunction$3(val) {
      return toString.call(val) === '[object Function]';
    }

    /**
     * Determine if a value is a Stream
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Stream, otherwise false
     */
    function isStream(val) {
      return isObject$2(val) && isFunction$3(val.pipe);
    }

    /**
     * Determine if a value is a FormData
     *
     * @param {Object} thing The value to test
     * @returns {boolean} True if value is an FormData, otherwise false
     */
    function isFormData(thing) {
      var pattern = '[object FormData]';
      return thing && (
        (typeof FormData === 'function' && thing instanceof FormData) ||
        toString.call(thing) === pattern ||
        (isFunction$3(thing.toString) && thing.toString() === pattern)
      );
    }

    /**
     * Determine if a value is a URLSearchParams object
     * @function
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a URLSearchParams object, otherwise false
     */
    var isURLSearchParams = kindOfTest('URLSearchParams');

    /**
     * Trim excess whitespace off the beginning and end of a string
     *
     * @param {String} str The String to trim
     * @returns {String} The String freed of excess whitespace
     */
    function trim(str) {
      return str.trim ? str.trim() : str.replace(/^\s+|\s+$/g, '');
    }

    /**
     * Determine if we're running in a standard browser environment
     *
     * This allows axios to run in a web worker, and react-native.
     * Both environments support XMLHttpRequest, but not fully standard globals.
     *
     * web workers:
     *  typeof window -> undefined
     *  typeof document -> undefined
     *
     * react-native:
     *  navigator.product -> 'ReactNative'
     * nativescript
     *  navigator.product -> 'NativeScript' or 'NS'
     */
    function isStandardBrowserEnv() {
      if (typeof navigator !== 'undefined' && (navigator.product === 'ReactNative' ||
                                               navigator.product === 'NativeScript' ||
                                               navigator.product === 'NS')) {
        return false;
      }
      return (
        typeof window !== 'undefined' &&
        typeof document !== 'undefined'
      );
    }

    /**
     * Iterate over an Array or an Object invoking a function for each item.
     *
     * If `obj` is an Array callback will be called passing
     * the value, index, and complete array for each item.
     *
     * If 'obj' is an Object callback will be called passing
     * the value, key, and complete object for each property.
     *
     * @param {Object|Array} obj The object to iterate
     * @param {Function} fn The callback to invoke for each item
     */
    function forEach$1(obj, fn) {
      // Don't bother if no value provided
      if (obj === null || typeof obj === 'undefined') {
        return;
      }

      // Force an array if not already something iterable
      if (typeof obj !== 'object') {
        /*eslint no-param-reassign:0*/
        obj = [obj];
      }

      if (isArray$2(obj)) {
        // Iterate over array values
        for (var i = 0, l = obj.length; i < l; i++) {
          fn.call(null, obj[i], i, obj);
        }
      } else {
        // Iterate over object keys
        for (var key in obj) {
          if (Object.prototype.hasOwnProperty.call(obj, key)) {
            fn.call(null, obj[key], key, obj);
          }
        }
      }
    }

    /**
     * Accepts varargs expecting each argument to be an object, then
     * immutably merges the properties of each object and returns result.
     *
     * When multiple objects contain the same key the later object in
     * the arguments list will take precedence.
     *
     * Example:
     *
     * ```js
     * var result = merge({foo: 123}, {foo: 456});
     * console.log(result.foo); // outputs 456
     * ```
     *
     * @param {Object} obj1 Object to merge
     * @returns {Object} Result of all merge properties
     */
    function merge(/* obj1, obj2, obj3, ... */) {
      var result = {};
      function assignValue(val, key) {
        if (isPlainObject(result[key]) && isPlainObject(val)) {
          result[key] = merge(result[key], val);
        } else if (isPlainObject(val)) {
          result[key] = merge({}, val);
        } else if (isArray$2(val)) {
          result[key] = val.slice();
        } else {
          result[key] = val;
        }
      }

      for (var i = 0, l = arguments.length; i < l; i++) {
        forEach$1(arguments[i], assignValue);
      }
      return result;
    }

    /**
     * Extends object a by mutably adding to it the properties of object b.
     *
     * @param {Object} a The object to be extended
     * @param {Object} b The object to copy properties from
     * @param {Object} thisArg The object to bind function to
     * @return {Object} The resulting value of object a
     */
    function extend(a, b, thisArg) {
      forEach$1(b, function assignValue(val, key) {
        if (thisArg && typeof val === 'function') {
          a[key] = bind$1(val, thisArg);
        } else {
          a[key] = val;
        }
      });
      return a;
    }

    /**
     * Remove byte order marker. This catches EF BB BF (the UTF-8 BOM)
     *
     * @param {string} content with BOM
     * @return {string} content value without BOM
     */
    function stripBOM(content) {
      if (content.charCodeAt(0) === 0xFEFF) {
        content = content.slice(1);
      }
      return content;
    }

    /**
     * Inherit the prototype methods from one constructor into another
     * @param {function} constructor
     * @param {function} superConstructor
     * @param {object} [props]
     * @param {object} [descriptors]
     */

    function inherits$2(constructor, superConstructor, props, descriptors) {
      constructor.prototype = Object.create(superConstructor.prototype, descriptors);
      constructor.prototype.constructor = constructor;
      props && Object.assign(constructor.prototype, props);
    }

    /**
     * Resolve object with deep prototype chain to a flat object
     * @param {Object} sourceObj source object
     * @param {Object} [destObj]
     * @param {Function} [filter]
     * @returns {Object}
     */

    function toFlatObject(sourceObj, destObj, filter) {
      var props;
      var i;
      var prop;
      var merged = {};

      destObj = destObj || {};

      do {
        props = Object.getOwnPropertyNames(sourceObj);
        i = props.length;
        while (i-- > 0) {
          prop = props[i];
          if (!merged[prop]) {
            destObj[prop] = sourceObj[prop];
            merged[prop] = true;
          }
        }
        sourceObj = Object.getPrototypeOf(sourceObj);
      } while (sourceObj && (!filter || filter(sourceObj, destObj)) && sourceObj !== Object.prototype);

      return destObj;
    }

    /*
     * determines whether a string ends with the characters of a specified string
     * @param {String} str
     * @param {String} searchString
     * @param {Number} [position= 0]
     * @returns {boolean}
     */
    function endsWith(str, searchString, position) {
      str = String(str);
      if (position === undefined || position > str.length) {
        position = str.length;
      }
      position -= searchString.length;
      var lastIndex = str.indexOf(searchString, position);
      return lastIndex !== -1 && lastIndex === position;
    }


    /**
     * Returns new array from array like object
     * @param {*} [thing]
     * @returns {Array}
     */
    function toArray(thing) {
      if (!thing) return null;
      var i = thing.length;
      if (isUndefined$2(i)) return null;
      var arr = new Array(i);
      while (i-- > 0) {
        arr[i] = thing[i];
      }
      return arr;
    }

    // eslint-disable-next-line func-names
    var isTypedArray = (function(TypedArray) {
      // eslint-disable-next-line func-names
      return function(thing) {
        return TypedArray && thing instanceof TypedArray;
      };
    })(typeof Uint8Array !== 'undefined' && Object.getPrototypeOf(Uint8Array));

    var utils$9 = {
      isArray: isArray$2,
      isArrayBuffer: isArrayBuffer,
      isBuffer: isBuffer$1,
      isFormData: isFormData,
      isArrayBufferView: isArrayBufferView,
      isString: isString$1,
      isNumber: isNumber$2,
      isObject: isObject$2,
      isPlainObject: isPlainObject,
      isUndefined: isUndefined$2,
      isDate: isDate$1,
      isFile: isFile,
      isBlob: isBlob,
      isFunction: isFunction$3,
      isStream: isStream,
      isURLSearchParams: isURLSearchParams,
      isStandardBrowserEnv: isStandardBrowserEnv,
      forEach: forEach$1,
      merge: merge,
      extend: extend,
      trim: trim,
      stripBOM: stripBOM,
      inherits: inherits$2,
      toFlatObject: toFlatObject,
      kindOf: kindOf,
      kindOfTest: kindOfTest,
      endsWith: endsWith,
      toArray: toArray,
      isTypedArray: isTypedArray,
      isFileList: isFileList
    };

    var utils$8 = utils$9;

    function encode$1(val) {
      return encodeURIComponent(val).
        replace(/%3A/gi, ':').
        replace(/%24/g, '$').
        replace(/%2C/gi, ',').
        replace(/%20/g, '+').
        replace(/%5B/gi, '[').
        replace(/%5D/gi, ']');
    }

    /**
     * Build a URL by appending params to the end
     *
     * @param {string} url The base of the url (e.g., http://www.google.com)
     * @param {object} [params] The params to be appended
     * @returns {string} The formatted url
     */
    var buildURL$1 = function buildURL(url, params, paramsSerializer) {
      /*eslint no-param-reassign:0*/
      if (!params) {
        return url;
      }

      var serializedParams;
      if (paramsSerializer) {
        serializedParams = paramsSerializer(params);
      } else if (utils$8.isURLSearchParams(params)) {
        serializedParams = params.toString();
      } else {
        var parts = [];

        utils$8.forEach(params, function serialize(val, key) {
          if (val === null || typeof val === 'undefined') {
            return;
          }

          if (utils$8.isArray(val)) {
            key = key + '[]';
          } else {
            val = [val];
          }

          utils$8.forEach(val, function parseValue(v) {
            if (utils$8.isDate(v)) {
              v = v.toISOString();
            } else if (utils$8.isObject(v)) {
              v = JSON.stringify(v);
            }
            parts.push(encode$1(key) + '=' + encode$1(v));
          });
        });

        serializedParams = parts.join('&');
      }

      if (serializedParams) {
        var hashmarkIndex = url.indexOf('#');
        if (hashmarkIndex !== -1) {
          url = url.slice(0, hashmarkIndex);
        }

        url += (url.indexOf('?') === -1 ? '?' : '&') + serializedParams;
      }

      return url;
    };

    var utils$7 = utils$9;

    function InterceptorManager$1() {
      this.handlers = [];
    }

    /**
     * Add a new interceptor to the stack
     *
     * @param {Function} fulfilled The function to handle `then` for a `Promise`
     * @param {Function} rejected The function to handle `reject` for a `Promise`
     *
     * @return {Number} An ID used to remove interceptor later
     */
    InterceptorManager$1.prototype.use = function use(fulfilled, rejected, options) {
      this.handlers.push({
        fulfilled: fulfilled,
        rejected: rejected,
        synchronous: options ? options.synchronous : false,
        runWhen: options ? options.runWhen : null
      });
      return this.handlers.length - 1;
    };

    /**
     * Remove an interceptor from the stack
     *
     * @param {Number} id The ID that was returned by `use`
     */
    InterceptorManager$1.prototype.eject = function eject(id) {
      if (this.handlers[id]) {
        this.handlers[id] = null;
      }
    };

    /**
     * Iterate over all the registered interceptors
     *
     * This method is particularly useful for skipping over any
     * interceptors that may have become `null` calling `eject`.
     *
     * @param {Function} fn The function to call for each interceptor
     */
    InterceptorManager$1.prototype.forEach = function forEach(fn) {
      utils$7.forEach(this.handlers, function forEachHandler(h) {
        if (h !== null) {
          fn(h);
        }
      });
    };

    var InterceptorManager_1 = InterceptorManager$1;

    var utils$6 = utils$9;

    var normalizeHeaderName$1 = function normalizeHeaderName(headers, normalizedName) {
      utils$6.forEach(headers, function processHeader(value, name) {
        if (name !== normalizedName && name.toUpperCase() === normalizedName.toUpperCase()) {
          headers[normalizedName] = value;
          delete headers[name];
        }
      });
    };

    var AxiosError_1;
    var hasRequiredAxiosError;

    function requireAxiosError () {
    	if (hasRequiredAxiosError) return AxiosError_1;
    	hasRequiredAxiosError = 1;

    	var utils = utils$9;

    	/**
    	 * Create an Error with the specified message, config, error code, request and response.
    	 *
    	 * @param {string} message The error message.
    	 * @param {string} [code] The error code (for example, 'ECONNABORTED').
    	 * @param {Object} [config] The config.
    	 * @param {Object} [request] The request.
    	 * @param {Object} [response] The response.
    	 * @returns {Error} The created error.
    	 */
    	function AxiosError(message, code, config, request, response) {
    	  Error.call(this);
    	  this.message = message;
    	  this.name = 'AxiosError';
    	  code && (this.code = code);
    	  config && (this.config = config);
    	  request && (this.request = request);
    	  response && (this.response = response);
    	}

    	utils.inherits(AxiosError, Error, {
    	  toJSON: function toJSON() {
    	    return {
    	      // Standard
    	      message: this.message,
    	      name: this.name,
    	      // Microsoft
    	      description: this.description,
    	      number: this.number,
    	      // Mozilla
    	      fileName: this.fileName,
    	      lineNumber: this.lineNumber,
    	      columnNumber: this.columnNumber,
    	      stack: this.stack,
    	      // Axios
    	      config: this.config,
    	      code: this.code,
    	      status: this.response && this.response.status ? this.response.status : null
    	    };
    	  }
    	});

    	var prototype = AxiosError.prototype;
    	var descriptors = {};

    	[
    	  'ERR_BAD_OPTION_VALUE',
    	  'ERR_BAD_OPTION',
    	  'ECONNABORTED',
    	  'ETIMEDOUT',
    	  'ERR_NETWORK',
    	  'ERR_FR_TOO_MANY_REDIRECTS',
    	  'ERR_DEPRECATED',
    	  'ERR_BAD_RESPONSE',
    	  'ERR_BAD_REQUEST',
    	  'ERR_CANCELED'
    	// eslint-disable-next-line func-names
    	].forEach(function(code) {
    	  descriptors[code] = {value: code};
    	});

    	Object.defineProperties(AxiosError, descriptors);
    	Object.defineProperty(prototype, 'isAxiosError', {value: true});

    	// eslint-disable-next-line func-names
    	AxiosError.from = function(error, code, config, request, response, customProps) {
    	  var axiosError = Object.create(prototype);

    	  utils.toFlatObject(error, axiosError, function filter(obj) {
    	    return obj !== Error.prototype;
    	  });

    	  AxiosError.call(axiosError, error.message, code, config, request, response);

    	  axiosError.name = error.name;

    	  customProps && Object.assign(axiosError, customProps);

    	  return axiosError;
    	};

    	AxiosError_1 = AxiosError;
    	return AxiosError_1;
    }

    var transitional = {
      silentJSONParsing: true,
      forcedJSONParsing: true,
      clarifyTimeoutError: false
    };

    var toFormData_1;
    var hasRequiredToFormData;

    function requireToFormData () {
    	if (hasRequiredToFormData) return toFormData_1;
    	hasRequiredToFormData = 1;

    	var utils = utils$9;

    	/**
    	 * Convert a data object to FormData
    	 * @param {Object} obj
    	 * @param {?Object} [formData]
    	 * @returns {Object}
    	 **/

    	function toFormData(obj, formData) {
    	  // eslint-disable-next-line no-param-reassign
    	  formData = formData || new FormData();

    	  var stack = [];

    	  function convertValue(value) {
    	    if (value === null) return '';

    	    if (utils.isDate(value)) {
    	      return value.toISOString();
    	    }

    	    if (utils.isArrayBuffer(value) || utils.isTypedArray(value)) {
    	      return typeof Blob === 'function' ? new Blob([value]) : Buffer$1.from(value);
    	    }

    	    return value;
    	  }

    	  function build(data, parentKey) {
    	    if (utils.isPlainObject(data) || utils.isArray(data)) {
    	      if (stack.indexOf(data) !== -1) {
    	        throw Error('Circular reference detected in ' + parentKey);
    	      }

    	      stack.push(data);

    	      utils.forEach(data, function each(value, key) {
    	        if (utils.isUndefined(value)) return;
    	        var fullKey = parentKey ? parentKey + '.' + key : key;
    	        var arr;

    	        if (value && !parentKey && typeof value === 'object') {
    	          if (utils.endsWith(key, '{}')) {
    	            // eslint-disable-next-line no-param-reassign
    	            value = JSON.stringify(value);
    	          } else if (utils.endsWith(key, '[]') && (arr = utils.toArray(value))) {
    	            // eslint-disable-next-line func-names
    	            arr.forEach(function(el) {
    	              !utils.isUndefined(el) && formData.append(fullKey, convertValue(el));
    	            });
    	            return;
    	          }
    	        }

    	        build(value, fullKey);
    	      });

    	      stack.pop();
    	    } else {
    	      formData.append(parentKey, convertValue(data));
    	    }
    	  }

    	  build(obj);

    	  return formData;
    	}

    	toFormData_1 = toFormData;
    	return toFormData_1;
    }

    var settle;
    var hasRequiredSettle;

    function requireSettle () {
    	if (hasRequiredSettle) return settle;
    	hasRequiredSettle = 1;

    	var AxiosError = requireAxiosError();

    	/**
    	 * Resolve or reject a Promise based on response status.
    	 *
    	 * @param {Function} resolve A function that resolves the promise.
    	 * @param {Function} reject A function that rejects the promise.
    	 * @param {object} response The response.
    	 */
    	settle = function settle(resolve, reject, response) {
    	  var validateStatus = response.config.validateStatus;
    	  if (!response.status || !validateStatus || validateStatus(response.status)) {
    	    resolve(response);
    	  } else {
    	    reject(new AxiosError(
    	      'Request failed with status code ' + response.status,
    	      [AxiosError.ERR_BAD_REQUEST, AxiosError.ERR_BAD_RESPONSE][Math.floor(response.status / 100) - 4],
    	      response.config,
    	      response.request,
    	      response
    	    ));
    	  }
    	};
    	return settle;
    }

    var cookies;
    var hasRequiredCookies;

    function requireCookies () {
    	if (hasRequiredCookies) return cookies;
    	hasRequiredCookies = 1;

    	var utils = utils$9;

    	cookies = (
    	  utils.isStandardBrowserEnv() ?

    	  // Standard browser envs support document.cookie
    	    (function standardBrowserEnv() {
    	      return {
    	        write: function write(name, value, expires, path, domain, secure) {
    	          var cookie = [];
    	          cookie.push(name + '=' + encodeURIComponent(value));

    	          if (utils.isNumber(expires)) {
    	            cookie.push('expires=' + new Date(expires).toGMTString());
    	          }

    	          if (utils.isString(path)) {
    	            cookie.push('path=' + path);
    	          }

    	          if (utils.isString(domain)) {
    	            cookie.push('domain=' + domain);
    	          }

    	          if (secure === true) {
    	            cookie.push('secure');
    	          }

    	          document.cookie = cookie.join('; ');
    	        },

    	        read: function read(name) {
    	          var match = document.cookie.match(new RegExp('(^|;\\s*)(' + name + ')=([^;]*)'));
    	          return (match ? decodeURIComponent(match[3]) : null);
    	        },

    	        remove: function remove(name) {
    	          this.write(name, '', Date.now() - 86400000);
    	        }
    	      };
    	    })() :

    	  // Non standard browser env (web workers, react-native) lack needed support.
    	    (function nonStandardBrowserEnv() {
    	      return {
    	        write: function write() {},
    	        read: function read() { return null; },
    	        remove: function remove() {}
    	      };
    	    })()
    	);
    	return cookies;
    }

    /**
     * Determines whether the specified URL is absolute
     *
     * @param {string} url The URL to test
     * @returns {boolean} True if the specified URL is absolute, otherwise false
     */
    var isAbsoluteURL$1 = function isAbsoluteURL(url) {
      // A URL is considered absolute if it begins with "<scheme>://" or "//" (protocol-relative URL).
      // RFC 3986 defines scheme name as a sequence of characters beginning with a letter and followed
      // by any combination of letters, digits, plus, period, or hyphen.
      return /^([a-z][a-z\d+\-.]*:)?\/\//i.test(url);
    };

    /**
     * Creates a new URL by combining the specified URLs
     *
     * @param {string} baseURL The base URL
     * @param {string} relativeURL The relative URL
     * @returns {string} The combined URL
     */
    var combineURLs$1 = function combineURLs(baseURL, relativeURL) {
      return relativeURL
        ? baseURL.replace(/\/+$/, '') + '/' + relativeURL.replace(/^\/+/, '')
        : baseURL;
    };

    var isAbsoluteURL = isAbsoluteURL$1;
    var combineURLs = combineURLs$1;

    /**
     * Creates a new URL by combining the baseURL with the requestedURL,
     * only when the requestedURL is not already an absolute URL.
     * If the requestURL is absolute, this function returns the requestedURL untouched.
     *
     * @param {string} baseURL The base URL
     * @param {string} requestedURL Absolute or relative URL to combine
     * @returns {string} The combined full path
     */
    var buildFullPath$1 = function buildFullPath(baseURL, requestedURL) {
      if (baseURL && !isAbsoluteURL(requestedURL)) {
        return combineURLs(baseURL, requestedURL);
      }
      return requestedURL;
    };

    var parseHeaders;
    var hasRequiredParseHeaders;

    function requireParseHeaders () {
    	if (hasRequiredParseHeaders) return parseHeaders;
    	hasRequiredParseHeaders = 1;

    	var utils = utils$9;

    	// Headers whose duplicates are ignored by node
    	// c.f. https://nodejs.org/api/http.html#http_message_headers
    	var ignoreDuplicateOf = [
    	  'age', 'authorization', 'content-length', 'content-type', 'etag',
    	  'expires', 'from', 'host', 'if-modified-since', 'if-unmodified-since',
    	  'last-modified', 'location', 'max-forwards', 'proxy-authorization',
    	  'referer', 'retry-after', 'user-agent'
    	];

    	/**
    	 * Parse headers into an object
    	 *
    	 * ```
    	 * Date: Wed, 27 Aug 2014 08:58:49 GMT
    	 * Content-Type: application/json
    	 * Connection: keep-alive
    	 * Transfer-Encoding: chunked
    	 * ```
    	 *
    	 * @param {String} headers Headers needing to be parsed
    	 * @returns {Object} Headers parsed into an object
    	 */
    	parseHeaders = function parseHeaders(headers) {
    	  var parsed = {};
    	  var key;
    	  var val;
    	  var i;

    	  if (!headers) { return parsed; }

    	  utils.forEach(headers.split('\n'), function parser(line) {
    	    i = line.indexOf(':');
    	    key = utils.trim(line.substr(0, i)).toLowerCase();
    	    val = utils.trim(line.substr(i + 1));

    	    if (key) {
    	      if (parsed[key] && ignoreDuplicateOf.indexOf(key) >= 0) {
    	        return;
    	      }
    	      if (key === 'set-cookie') {
    	        parsed[key] = (parsed[key] ? parsed[key] : []).concat([val]);
    	      } else {
    	        parsed[key] = parsed[key] ? parsed[key] + ', ' + val : val;
    	      }
    	    }
    	  });

    	  return parsed;
    	};
    	return parseHeaders;
    }

    var isURLSameOrigin;
    var hasRequiredIsURLSameOrigin;

    function requireIsURLSameOrigin () {
    	if (hasRequiredIsURLSameOrigin) return isURLSameOrigin;
    	hasRequiredIsURLSameOrigin = 1;

    	var utils = utils$9;

    	isURLSameOrigin = (
    	  utils.isStandardBrowserEnv() ?

    	  // Standard browser envs have full support of the APIs needed to test
    	  // whether the request URL is of the same origin as current location.
    	    (function standardBrowserEnv() {
    	      var msie = /(msie|trident)/i.test(navigator.userAgent);
    	      var urlParsingNode = document.createElement('a');
    	      var originURL;

    	      /**
    	    * Parse a URL to discover it's components
    	    *
    	    * @param {String} url The URL to be parsed
    	    * @returns {Object}
    	    */
    	      function resolveURL(url) {
    	        var href = url;

    	        if (msie) {
    	        // IE needs attribute set twice to normalize properties
    	          urlParsingNode.setAttribute('href', href);
    	          href = urlParsingNode.href;
    	        }

    	        urlParsingNode.setAttribute('href', href);

    	        // urlParsingNode provides the UrlUtils interface - http://url.spec.whatwg.org/#urlutils
    	        return {
    	          href: urlParsingNode.href,
    	          protocol: urlParsingNode.protocol ? urlParsingNode.protocol.replace(/:$/, '') : '',
    	          host: urlParsingNode.host,
    	          search: urlParsingNode.search ? urlParsingNode.search.replace(/^\?/, '') : '',
    	          hash: urlParsingNode.hash ? urlParsingNode.hash.replace(/^#/, '') : '',
    	          hostname: urlParsingNode.hostname,
    	          port: urlParsingNode.port,
    	          pathname: (urlParsingNode.pathname.charAt(0) === '/') ?
    	            urlParsingNode.pathname :
    	            '/' + urlParsingNode.pathname
    	        };
    	      }

    	      originURL = resolveURL(window.location.href);

    	      /**
    	    * Determine if a URL shares the same origin as the current location
    	    *
    	    * @param {String} requestURL The URL to test
    	    * @returns {boolean} True if URL shares the same origin, otherwise false
    	    */
    	      return function isURLSameOrigin(requestURL) {
    	        var parsed = (utils.isString(requestURL)) ? resolveURL(requestURL) : requestURL;
    	        return (parsed.protocol === originURL.protocol &&
    	            parsed.host === originURL.host);
    	      };
    	    })() :

    	  // Non standard browser envs (web workers, react-native) lack needed support.
    	    (function nonStandardBrowserEnv() {
    	      return function isURLSameOrigin() {
    	        return true;
    	      };
    	    })()
    	);
    	return isURLSameOrigin;
    }

    var CanceledError_1;
    var hasRequiredCanceledError;

    function requireCanceledError () {
    	if (hasRequiredCanceledError) return CanceledError_1;
    	hasRequiredCanceledError = 1;

    	var AxiosError = requireAxiosError();
    	var utils = utils$9;

    	/**
    	 * A `CanceledError` is an object that is thrown when an operation is canceled.
    	 *
    	 * @class
    	 * @param {string=} message The message.
    	 */
    	function CanceledError(message) {
    	  // eslint-disable-next-line no-eq-null,eqeqeq
    	  AxiosError.call(this, message == null ? 'canceled' : message, AxiosError.ERR_CANCELED);
    	  this.name = 'CanceledError';
    	}

    	utils.inherits(CanceledError, AxiosError, {
    	  __CANCEL__: true
    	});

    	CanceledError_1 = CanceledError;
    	return CanceledError_1;
    }

    var parseProtocol;
    var hasRequiredParseProtocol;

    function requireParseProtocol () {
    	if (hasRequiredParseProtocol) return parseProtocol;
    	hasRequiredParseProtocol = 1;

    	parseProtocol = function parseProtocol(url) {
    	  var match = /^([-+\w]{1,25})(:?\/\/|:)/.exec(url);
    	  return match && match[1] || '';
    	};
    	return parseProtocol;
    }

    var xhr$1;
    var hasRequiredXhr;

    function requireXhr () {
    	if (hasRequiredXhr) return xhr$1;
    	hasRequiredXhr = 1;

    	var utils = utils$9;
    	var settle = requireSettle();
    	var cookies = requireCookies();
    	var buildURL = buildURL$1;
    	var buildFullPath = buildFullPath$1;
    	var parseHeaders = requireParseHeaders();
    	var isURLSameOrigin = requireIsURLSameOrigin();
    	var transitionalDefaults = transitional;
    	var AxiosError = requireAxiosError();
    	var CanceledError = requireCanceledError();
    	var parseProtocol = requireParseProtocol();

    	xhr$1 = function xhrAdapter(config) {
    	  return new Promise(function dispatchXhrRequest(resolve, reject) {
    	    var requestData = config.data;
    	    var requestHeaders = config.headers;
    	    var responseType = config.responseType;
    	    var onCanceled;
    	    function done() {
    	      if (config.cancelToken) {
    	        config.cancelToken.unsubscribe(onCanceled);
    	      }

    	      if (config.signal) {
    	        config.signal.removeEventListener('abort', onCanceled);
    	      }
    	    }

    	    if (utils.isFormData(requestData) && utils.isStandardBrowserEnv()) {
    	      delete requestHeaders['Content-Type']; // Let the browser set it
    	    }

    	    var request = new XMLHttpRequest();

    	    // HTTP basic authentication
    	    if (config.auth) {
    	      var username = config.auth.username || '';
    	      var password = config.auth.password ? unescape(encodeURIComponent(config.auth.password)) : '';
    	      requestHeaders.Authorization = 'Basic ' + btoa(username + ':' + password);
    	    }

    	    var fullPath = buildFullPath(config.baseURL, config.url);

    	    request.open(config.method.toUpperCase(), buildURL(fullPath, config.params, config.paramsSerializer), true);

    	    // Set the request timeout in MS
    	    request.timeout = config.timeout;

    	    function onloadend() {
    	      if (!request) {
    	        return;
    	      }
    	      // Prepare the response
    	      var responseHeaders = 'getAllResponseHeaders' in request ? parseHeaders(request.getAllResponseHeaders()) : null;
    	      var responseData = !responseType || responseType === 'text' ||  responseType === 'json' ?
    	        request.responseText : request.response;
    	      var response = {
    	        data: responseData,
    	        status: request.status,
    	        statusText: request.statusText,
    	        headers: responseHeaders,
    	        config: config,
    	        request: request
    	      };

    	      settle(function _resolve(value) {
    	        resolve(value);
    	        done();
    	      }, function _reject(err) {
    	        reject(err);
    	        done();
    	      }, response);

    	      // Clean up request
    	      request = null;
    	    }

    	    if ('onloadend' in request) {
    	      // Use onloadend if available
    	      request.onloadend = onloadend;
    	    } else {
    	      // Listen for ready state to emulate onloadend
    	      request.onreadystatechange = function handleLoad() {
    	        if (!request || request.readyState !== 4) {
    	          return;
    	        }

    	        // The request errored out and we didn't get a response, this will be
    	        // handled by onerror instead
    	        // With one exception: request that using file: protocol, most browsers
    	        // will return status as 0 even though it's a successful request
    	        if (request.status === 0 && !(request.responseURL && request.responseURL.indexOf('file:') === 0)) {
    	          return;
    	        }
    	        // readystate handler is calling before onerror or ontimeout handlers,
    	        // so we should call onloadend on the next 'tick'
    	        setTimeout(onloadend);
    	      };
    	    }

    	    // Handle browser request cancellation (as opposed to a manual cancellation)
    	    request.onabort = function handleAbort() {
    	      if (!request) {
    	        return;
    	      }

    	      reject(new AxiosError('Request aborted', AxiosError.ECONNABORTED, config, request));

    	      // Clean up request
    	      request = null;
    	    };

    	    // Handle low level network errors
    	    request.onerror = function handleError() {
    	      // Real errors are hidden from us by the browser
    	      // onerror should only fire if it's a network error
    	      reject(new AxiosError('Network Error', AxiosError.ERR_NETWORK, config, request, request));

    	      // Clean up request
    	      request = null;
    	    };

    	    // Handle timeout
    	    request.ontimeout = function handleTimeout() {
    	      var timeoutErrorMessage = config.timeout ? 'timeout of ' + config.timeout + 'ms exceeded' : 'timeout exceeded';
    	      var transitional = config.transitional || transitionalDefaults;
    	      if (config.timeoutErrorMessage) {
    	        timeoutErrorMessage = config.timeoutErrorMessage;
    	      }
    	      reject(new AxiosError(
    	        timeoutErrorMessage,
    	        transitional.clarifyTimeoutError ? AxiosError.ETIMEDOUT : AxiosError.ECONNABORTED,
    	        config,
    	        request));

    	      // Clean up request
    	      request = null;
    	    };

    	    // Add xsrf header
    	    // This is only done if running in a standard browser environment.
    	    // Specifically not if we're in a web worker, or react-native.
    	    if (utils.isStandardBrowserEnv()) {
    	      // Add xsrf header
    	      var xsrfValue = (config.withCredentials || isURLSameOrigin(fullPath)) && config.xsrfCookieName ?
    	        cookies.read(config.xsrfCookieName) :
    	        undefined;

    	      if (xsrfValue) {
    	        requestHeaders[config.xsrfHeaderName] = xsrfValue;
    	      }
    	    }

    	    // Add headers to the request
    	    if ('setRequestHeader' in request) {
    	      utils.forEach(requestHeaders, function setRequestHeader(val, key) {
    	        if (typeof requestData === 'undefined' && key.toLowerCase() === 'content-type') {
    	          // Remove Content-Type if data is undefined
    	          delete requestHeaders[key];
    	        } else {
    	          // Otherwise add header to the request
    	          request.setRequestHeader(key, val);
    	        }
    	      });
    	    }

    	    // Add withCredentials to request if needed
    	    if (!utils.isUndefined(config.withCredentials)) {
    	      request.withCredentials = !!config.withCredentials;
    	    }

    	    // Add responseType to request if needed
    	    if (responseType && responseType !== 'json') {
    	      request.responseType = config.responseType;
    	    }

    	    // Handle progress if needed
    	    if (typeof config.onDownloadProgress === 'function') {
    	      request.addEventListener('progress', config.onDownloadProgress);
    	    }

    	    // Not all browsers support upload events
    	    if (typeof config.onUploadProgress === 'function' && request.upload) {
    	      request.upload.addEventListener('progress', config.onUploadProgress);
    	    }

    	    if (config.cancelToken || config.signal) {
    	      // Handle cancellation
    	      // eslint-disable-next-line func-names
    	      onCanceled = function(cancel) {
    	        if (!request) {
    	          return;
    	        }
    	        reject(!cancel || (cancel && cancel.type) ? new CanceledError() : cancel);
    	        request.abort();
    	        request = null;
    	      };

    	      config.cancelToken && config.cancelToken.subscribe(onCanceled);
    	      if (config.signal) {
    	        config.signal.aborted ? onCanceled() : config.signal.addEventListener('abort', onCanceled);
    	      }
    	    }

    	    if (!requestData) {
    	      requestData = null;
    	    }

    	    var protocol = parseProtocol(fullPath);

    	    if (protocol && [ 'http', 'https', 'file' ].indexOf(protocol) === -1) {
    	      reject(new AxiosError('Unsupported protocol ' + protocol + ':', AxiosError.ERR_BAD_REQUEST, config));
    	      return;
    	    }


    	    // Send the request
    	    request.send(requestData);
    	  });
    	};
    	return xhr$1;
    }

    var hasFetch = isFunction$2(global$1.fetch) && isFunction$2(global$1.ReadableStream);

    var _blobConstructor;
    function blobConstructor() {
      if (typeof _blobConstructor !== 'undefined') {
        return _blobConstructor;
      }
      try {
        new global$1.Blob([new ArrayBuffer(1)]);
        _blobConstructor = true;
      } catch (e) {
        _blobConstructor = false;
      }
      return _blobConstructor
    }
    var xhr;

    function checkTypeSupport(type) {
      if (!xhr) {
        xhr = new global$1.XMLHttpRequest();
        // If location.host is empty, e.g. if this page/worker was loaded
        // from a Blob, then use example.com to avoid an error
        xhr.open('GET', global$1.location.host ? '/' : 'https://example.com');
      }
      try {
        xhr.responseType = type;
        return xhr.responseType === type
      } catch (e) {
        return false
      }

    }

    // For some strange reason, Safari 7.0 reports typeof global.ArrayBuffer === 'object'.
    // Safari 7.1 appears to have fixed this bug.
    var haveArrayBuffer = typeof global$1.ArrayBuffer !== 'undefined';
    var haveSlice = haveArrayBuffer && isFunction$2(global$1.ArrayBuffer.prototype.slice);

    var arraybuffer = haveArrayBuffer && checkTypeSupport('arraybuffer');
      // These next two tests unavoidably show warnings in Chrome. Since fetch will always
      // be used if it's available, just return false for these to avoid the warnings.
    var msstream = !hasFetch && haveSlice && checkTypeSupport('ms-stream');
    var mozchunkedarraybuffer = !hasFetch && haveArrayBuffer &&
      checkTypeSupport('moz-chunked-arraybuffer');
    var overrideMimeType = isFunction$2(xhr.overrideMimeType);
    var vbArray = isFunction$2(global$1.VBArray);

    function isFunction$2(value) {
      return typeof value === 'function'
    }

    xhr = null; // Help gc

    var inherits;
    if (typeof Object.create === 'function'){
      inherits = function inherits(ctor, superCtor) {
        // implementation from standard node.js 'util' module
        ctor.super_ = superCtor;
        ctor.prototype = Object.create(superCtor.prototype, {
          constructor: {
            value: ctor,
            enumerable: false,
            writable: true,
            configurable: true
          }
        });
      };
    } else {
      inherits = function inherits(ctor, superCtor) {
        ctor.super_ = superCtor;
        var TempCtor = function () {};
        TempCtor.prototype = superCtor.prototype;
        ctor.prototype = new TempCtor();
        ctor.prototype.constructor = ctor;
      };
    }
    var inherits$1 = inherits;

    var formatRegExp = /%[sdj%]/g;
    function format$1(f) {
      if (!isString(f)) {
        var objects = [];
        for (var i = 0; i < arguments.length; i++) {
          objects.push(inspect$1(arguments[i]));
        }
        return objects.join(' ');
      }

      var i = 1;
      var args = arguments;
      var len = args.length;
      var str = String(f).replace(formatRegExp, function(x) {
        if (x === '%%') return '%';
        if (i >= len) return x;
        switch (x) {
          case '%s': return String(args[i++]);
          case '%d': return Number(args[i++]);
          case '%j':
            try {
              return JSON.stringify(args[i++]);
            } catch (_) {
              return '[Circular]';
            }
          default:
            return x;
        }
      });
      for (var x = args[i]; i < len; x = args[++i]) {
        if (isNull(x) || !isObject$1(x)) {
          str += ' ' + x;
        } else {
          str += ' ' + inspect$1(x);
        }
      }
      return str;
    }

    // Mark that a method should not be used.
    // Returns a modified function which warns once by default.
    // If --no-deprecation is set, then it is a no-op.
    function deprecate(fn, msg) {
      // Allow for deprecating things in the process of starting up.
      if (isUndefined$1(global$1.process)) {
        return function() {
          return deprecate(fn, msg).apply(this, arguments);
        };
      }

      if (process.noDeprecation === true) {
        return fn;
      }

      var warned = false;
      function deprecated() {
        if (!warned) {
          if (process.throwDeprecation) {
            throw new Error(msg);
          } else if (process.traceDeprecation) {
            console.trace(msg);
          } else {
            console.error(msg);
          }
          warned = true;
        }
        return fn.apply(this, arguments);
      }

      return deprecated;
    }

    var debugs = {};
    var debugEnviron;
    function debuglog(set) {
      if (isUndefined$1(debugEnviron))
        debugEnviron = process.env.NODE_DEBUG || '';
      set = set.toUpperCase();
      if (!debugs[set]) {
        if (new RegExp('\\b' + set + '\\b', 'i').test(debugEnviron)) {
          var pid = 0;
          debugs[set] = function() {
            var msg = format$1.apply(null, arguments);
            console.error('%s %d: %s', set, pid, msg);
          };
        } else {
          debugs[set] = function() {};
        }
      }
      return debugs[set];
    }

    /**
     * Echos the value of a value. Trys to print the value out
     * in the best way possible given the different types.
     *
     * @param {Object} obj The object to print out.
     * @param {Object} opts Optional options object that alters the output.
     */
    /* legacy: obj, showHidden, depth, colors*/
    function inspect$1(obj, opts) {
      // default options
      var ctx = {
        seen: [],
        stylize: stylizeNoColor
      };
      // legacy...
      if (arguments.length >= 3) ctx.depth = arguments[2];
      if (arguments.length >= 4) ctx.colors = arguments[3];
      if (isBoolean(opts)) {
        // legacy...
        ctx.showHidden = opts;
      } else if (opts) {
        // got an "options" object
        _extend(ctx, opts);
      }
      // set default options
      if (isUndefined$1(ctx.showHidden)) ctx.showHidden = false;
      if (isUndefined$1(ctx.depth)) ctx.depth = 2;
      if (isUndefined$1(ctx.colors)) ctx.colors = false;
      if (isUndefined$1(ctx.customInspect)) ctx.customInspect = true;
      if (ctx.colors) ctx.stylize = stylizeWithColor;
      return formatValue(ctx, obj, ctx.depth);
    }

    // http://en.wikipedia.org/wiki/ANSI_escape_code#graphics
    inspect$1.colors = {
      'bold' : [1, 22],
      'italic' : [3, 23],
      'underline' : [4, 24],
      'inverse' : [7, 27],
      'white' : [37, 39],
      'grey' : [90, 39],
      'black' : [30, 39],
      'blue' : [34, 39],
      'cyan' : [36, 39],
      'green' : [32, 39],
      'magenta' : [35, 39],
      'red' : [31, 39],
      'yellow' : [33, 39]
    };

    // Don't use 'blue' not visible on cmd.exe
    inspect$1.styles = {
      'special': 'cyan',
      'number': 'yellow',
      'boolean': 'yellow',
      'undefined': 'grey',
      'null': 'bold',
      'string': 'green',
      'date': 'magenta',
      // "name": intentionally not styling
      'regexp': 'red'
    };


    function stylizeWithColor(str, styleType) {
      var style = inspect$1.styles[styleType];

      if (style) {
        return '\u001b[' + inspect$1.colors[style][0] + 'm' + str +
               '\u001b[' + inspect$1.colors[style][1] + 'm';
      } else {
        return str;
      }
    }


    function stylizeNoColor(str, styleType) {
      return str;
    }


    function arrayToHash(array) {
      var hash = {};

      array.forEach(function(val, idx) {
        hash[val] = true;
      });

      return hash;
    }


    function formatValue(ctx, value, recurseTimes) {
      // Provide a hook for user-specified inspect functions.
      // Check that value is an object with an inspect function on it
      if (ctx.customInspect &&
          value &&
          isFunction$1(value.inspect) &&
          // Filter out the util module, it's inspect function is special
          value.inspect !== inspect$1 &&
          // Also filter out any prototype objects using the circular check.
          !(value.constructor && value.constructor.prototype === value)) {
        var ret = value.inspect(recurseTimes, ctx);
        if (!isString(ret)) {
          ret = formatValue(ctx, ret, recurseTimes);
        }
        return ret;
      }

      // Primitive types cannot have properties
      var primitive = formatPrimitive(ctx, value);
      if (primitive) {
        return primitive;
      }

      // Look up the keys of the object.
      var keys = Object.keys(value);
      var visibleKeys = arrayToHash(keys);

      if (ctx.showHidden) {
        keys = Object.getOwnPropertyNames(value);
      }

      // IE doesn't make error fields non-enumerable
      // http://msdn.microsoft.com/en-us/library/ie/dww52sbt(v=vs.94).aspx
      if (isError(value)
          && (keys.indexOf('message') >= 0 || keys.indexOf('description') >= 0)) {
        return formatError(value);
      }

      // Some type of object without properties can be shortcutted.
      if (keys.length === 0) {
        if (isFunction$1(value)) {
          var name = value.name ? ': ' + value.name : '';
          return ctx.stylize('[Function' + name + ']', 'special');
        }
        if (isRegExp(value)) {
          return ctx.stylize(RegExp.prototype.toString.call(value), 'regexp');
        }
        if (isDate(value)) {
          return ctx.stylize(Date.prototype.toString.call(value), 'date');
        }
        if (isError(value)) {
          return formatError(value);
        }
      }

      var base = '', array = false, braces = ['{', '}'];

      // Make Array say that they are Array
      if (isArray$1(value)) {
        array = true;
        braces = ['[', ']'];
      }

      // Make functions say that they are functions
      if (isFunction$1(value)) {
        var n = value.name ? ': ' + value.name : '';
        base = ' [Function' + n + ']';
      }

      // Make RegExps say that they are RegExps
      if (isRegExp(value)) {
        base = ' ' + RegExp.prototype.toString.call(value);
      }

      // Make dates with properties first say the date
      if (isDate(value)) {
        base = ' ' + Date.prototype.toUTCString.call(value);
      }

      // Make error with message first say the error
      if (isError(value)) {
        base = ' ' + formatError(value);
      }

      if (keys.length === 0 && (!array || value.length == 0)) {
        return braces[0] + base + braces[1];
      }

      if (recurseTimes < 0) {
        if (isRegExp(value)) {
          return ctx.stylize(RegExp.prototype.toString.call(value), 'regexp');
        } else {
          return ctx.stylize('[Object]', 'special');
        }
      }

      ctx.seen.push(value);

      var output;
      if (array) {
        output = formatArray(ctx, value, recurseTimes, visibleKeys, keys);
      } else {
        output = keys.map(function(key) {
          return formatProperty(ctx, value, recurseTimes, visibleKeys, key, array);
        });
      }

      ctx.seen.pop();

      return reduceToSingleString(output, base, braces);
    }


    function formatPrimitive(ctx, value) {
      if (isUndefined$1(value))
        return ctx.stylize('undefined', 'undefined');
      if (isString(value)) {
        var simple = '\'' + JSON.stringify(value).replace(/^"|"$/g, '')
                                                 .replace(/'/g, "\\'")
                                                 .replace(/\\"/g, '"') + '\'';
        return ctx.stylize(simple, 'string');
      }
      if (isNumber$1(value))
        return ctx.stylize('' + value, 'number');
      if (isBoolean(value))
        return ctx.stylize('' + value, 'boolean');
      // For some reason typeof null is "object", so special case here.
      if (isNull(value))
        return ctx.stylize('null', 'null');
    }


    function formatError(value) {
      return '[' + Error.prototype.toString.call(value) + ']';
    }


    function formatArray(ctx, value, recurseTimes, visibleKeys, keys) {
      var output = [];
      for (var i = 0, l = value.length; i < l; ++i) {
        if (hasOwnProperty$1(value, String(i))) {
          output.push(formatProperty(ctx, value, recurseTimes, visibleKeys,
              String(i), true));
        } else {
          output.push('');
        }
      }
      keys.forEach(function(key) {
        if (!key.match(/^\d+$/)) {
          output.push(formatProperty(ctx, value, recurseTimes, visibleKeys,
              key, true));
        }
      });
      return output;
    }


    function formatProperty(ctx, value, recurseTimes, visibleKeys, key, array) {
      var name, str, desc;
      desc = Object.getOwnPropertyDescriptor(value, key) || { value: value[key] };
      if (desc.get) {
        if (desc.set) {
          str = ctx.stylize('[Getter/Setter]', 'special');
        } else {
          str = ctx.stylize('[Getter]', 'special');
        }
      } else {
        if (desc.set) {
          str = ctx.stylize('[Setter]', 'special');
        }
      }
      if (!hasOwnProperty$1(visibleKeys, key)) {
        name = '[' + key + ']';
      }
      if (!str) {
        if (ctx.seen.indexOf(desc.value) < 0) {
          if (isNull(recurseTimes)) {
            str = formatValue(ctx, desc.value, null);
          } else {
            str = formatValue(ctx, desc.value, recurseTimes - 1);
          }
          if (str.indexOf('\n') > -1) {
            if (array) {
              str = str.split('\n').map(function(line) {
                return '  ' + line;
              }).join('\n').substr(2);
            } else {
              str = '\n' + str.split('\n').map(function(line) {
                return '   ' + line;
              }).join('\n');
            }
          }
        } else {
          str = ctx.stylize('[Circular]', 'special');
        }
      }
      if (isUndefined$1(name)) {
        if (array && key.match(/^\d+$/)) {
          return str;
        }
        name = JSON.stringify('' + key);
        if (name.match(/^"([a-zA-Z_][a-zA-Z_0-9]*)"$/)) {
          name = name.substr(1, name.length - 2);
          name = ctx.stylize(name, 'name');
        } else {
          name = name.replace(/'/g, "\\'")
                     .replace(/\\"/g, '"')
                     .replace(/(^"|"$)/g, "'");
          name = ctx.stylize(name, 'string');
        }
      }

      return name + ': ' + str;
    }


    function reduceToSingleString(output, base, braces) {
      var length = output.reduce(function(prev, cur) {
        if (cur.indexOf('\n') >= 0) ;
        return prev + cur.replace(/\u001b\[\d\d?m/g, '').length + 1;
      }, 0);

      if (length > 60) {
        return braces[0] +
               (base === '' ? '' : base + '\n ') +
               ' ' +
               output.join(',\n  ') +
               ' ' +
               braces[1];
      }

      return braces[0] + base + ' ' + output.join(', ') + ' ' + braces[1];
    }


    // NOTE: These type checking functions intentionally don't use `instanceof`
    // because it is fragile and can be easily faked with `Object.create()`.
    function isArray$1(ar) {
      return Array.isArray(ar);
    }

    function isBoolean(arg) {
      return typeof arg === 'boolean';
    }

    function isNull(arg) {
      return arg === null;
    }

    function isNullOrUndefined(arg) {
      return arg == null;
    }

    function isNumber$1(arg) {
      return typeof arg === 'number';
    }

    function isString(arg) {
      return typeof arg === 'string';
    }

    function isSymbol(arg) {
      return typeof arg === 'symbol';
    }

    function isUndefined$1(arg) {
      return arg === void 0;
    }

    function isRegExp(re) {
      return isObject$1(re) && objectToString(re) === '[object RegExp]';
    }

    function isObject$1(arg) {
      return typeof arg === 'object' && arg !== null;
    }

    function isDate(d) {
      return isObject$1(d) && objectToString(d) === '[object Date]';
    }

    function isError(e) {
      return isObject$1(e) &&
          (objectToString(e) === '[object Error]' || e instanceof Error);
    }

    function isFunction$1(arg) {
      return typeof arg === 'function';
    }

    function isPrimitive(arg) {
      return arg === null ||
             typeof arg === 'boolean' ||
             typeof arg === 'number' ||
             typeof arg === 'string' ||
             typeof arg === 'symbol' ||  // ES6 symbol
             typeof arg === 'undefined';
    }

    function isBuffer(maybeBuf) {
      return isBuffer$2(maybeBuf);
    }

    function objectToString(o) {
      return Object.prototype.toString.call(o);
    }


    function pad(n) {
      return n < 10 ? '0' + n.toString(10) : n.toString(10);
    }


    var months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
                  'Oct', 'Nov', 'Dec'];

    // 26 Feb 16:19:34
    function timestamp() {
      var d = new Date();
      var time = [pad(d.getHours()),
                  pad(d.getMinutes()),
                  pad(d.getSeconds())].join(':');
      return [d.getDate(), months[d.getMonth()], time].join(' ');
    }


    // log is just a thin wrapper to console.log that prepends a timestamp
    function log() {
      console.log('%s - %s', timestamp(), format$1.apply(null, arguments));
    }

    function _extend(origin, add) {
      // Don't do anything if add isn't an object
      if (!add || !isObject$1(add)) return origin;

      var keys = Object.keys(add);
      var i = keys.length;
      while (i--) {
        origin[keys[i]] = add[keys[i]];
      }
      return origin;
    }
    function hasOwnProperty$1(obj, prop) {
      return Object.prototype.hasOwnProperty.call(obj, prop);
    }

    var util = {
      inherits: inherits$1,
      _extend: _extend,
      log: log,
      isBuffer: isBuffer,
      isPrimitive: isPrimitive,
      isFunction: isFunction$1,
      isError: isError,
      isDate: isDate,
      isObject: isObject$1,
      isRegExp: isRegExp,
      isUndefined: isUndefined$1,
      isSymbol: isSymbol,
      isString: isString,
      isNumber: isNumber$1,
      isNullOrUndefined: isNullOrUndefined,
      isNull: isNull,
      isBoolean: isBoolean,
      isArray: isArray$1,
      inspect: inspect$1,
      deprecate: deprecate,
      format: format$1,
      debuglog: debuglog
    };

    var util$1 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        format: format$1,
        deprecate: deprecate,
        debuglog: debuglog,
        inspect: inspect$1,
        isArray: isArray$1,
        isBoolean: isBoolean,
        isNull: isNull,
        isNullOrUndefined: isNullOrUndefined,
        isNumber: isNumber$1,
        isString: isString,
        isSymbol: isSymbol,
        isUndefined: isUndefined$1,
        isRegExp: isRegExp,
        isObject: isObject$1,
        isDate: isDate,
        isError: isError,
        isFunction: isFunction$1,
        isPrimitive: isPrimitive,
        isBuffer: isBuffer,
        log: log,
        inherits: inherits$1,
        _extend: _extend,
        'default': util
    });

    // Copyright Joyent, Inc. and other Node contributors.
    //
    // Permission is hereby granted, free of charge, to any person obtaining a
    // copy of this software and associated documentation files (the
    // "Software"), to deal in the Software without restriction, including
    // without limitation the rights to use, copy, modify, merge, publish,
    // distribute, sublicense, and/or sell copies of the Software, and to permit
    // persons to whom the Software is furnished to do so, subject to the
    // following conditions:
    //
    // The above copyright notice and this permission notice shall be included
    // in all copies or substantial portions of the Software.
    //
    // THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    // OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    // MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    // NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    // DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    // USE OR OTHER DEALINGS IN THE SOFTWARE.

    function EventEmitter() {
      this._events = this._events || {};
      this._maxListeners = this._maxListeners || undefined;
    }
    var events = EventEmitter;

    // Backwards-compat with node 0.10.x
    EventEmitter.EventEmitter = EventEmitter;

    EventEmitter.prototype._events = undefined;
    EventEmitter.prototype._maxListeners = undefined;

    // By default EventEmitters will print a warning if more than 10 listeners are
    // added to it. This is a useful default which helps finding memory leaks.
    EventEmitter.defaultMaxListeners = 10;

    // Obviously not all Emitters should be limited to 10. This function allows
    // that to be increased. Set to zero for unlimited.
    EventEmitter.prototype.setMaxListeners = function(n) {
      if (!isNumber(n) || n < 0 || isNaN(n))
        throw TypeError('n must be a positive number');
      this._maxListeners = n;
      return this;
    };

    EventEmitter.prototype.emit = function(type) {
      var er, handler, len, args, i, listeners;

      if (!this._events)
        this._events = {};

      // If there is no 'error' event listener then throw.
      if (type === 'error') {
        if (!this._events.error ||
            (isObject(this._events.error) && !this._events.error.length)) {
          er = arguments[1];
          if (er instanceof Error) {
            throw er; // Unhandled 'error' event
          } else {
            // At least give some kind of context to the user
            var err = new Error('Uncaught, unspecified "error" event. (' + er + ')');
            err.context = er;
            throw err;
          }
        }
      }

      handler = this._events[type];

      if (isUndefined(handler))
        return false;

      if (isFunction(handler)) {
        switch (arguments.length) {
          // fast cases
          case 1:
            handler.call(this);
            break;
          case 2:
            handler.call(this, arguments[1]);
            break;
          case 3:
            handler.call(this, arguments[1], arguments[2]);
            break;
          // slower
          default:
            args = Array.prototype.slice.call(arguments, 1);
            handler.apply(this, args);
        }
      } else if (isObject(handler)) {
        args = Array.prototype.slice.call(arguments, 1);
        listeners = handler.slice();
        len = listeners.length;
        for (i = 0; i < len; i++)
          listeners[i].apply(this, args);
      }

      return true;
    };

    EventEmitter.prototype.addListener = function(type, listener) {
      var m;

      if (!isFunction(listener))
        throw TypeError('listener must be a function');

      if (!this._events)
        this._events = {};

      // To avoid recursion in the case that type === "newListener"! Before
      // adding it to the listeners, first emit "newListener".
      if (this._events.newListener)
        this.emit('newListener', type,
                  isFunction(listener.listener) ?
                  listener.listener : listener);

      if (!this._events[type])
        // Optimize the case of one listener. Don't need the extra array object.
        this._events[type] = listener;
      else if (isObject(this._events[type]))
        // If we've already got an array, just append.
        this._events[type].push(listener);
      else
        // Adding the second element, need to change to array.
        this._events[type] = [this._events[type], listener];

      // Check for listener leak
      if (isObject(this._events[type]) && !this._events[type].warned) {
        if (!isUndefined(this._maxListeners)) {
          m = this._maxListeners;
        } else {
          m = EventEmitter.defaultMaxListeners;
        }

        if (m && m > 0 && this._events[type].length > m) {
          this._events[type].warned = true;
          console.error('(node) warning: possible EventEmitter memory ' +
                        'leak detected. %d listeners added. ' +
                        'Use emitter.setMaxListeners() to increase limit.',
                        this._events[type].length);
          if (typeof console.trace === 'function') {
            // not supported in IE 10
            console.trace();
          }
        }
      }

      return this;
    };

    EventEmitter.prototype.on = EventEmitter.prototype.addListener;

    EventEmitter.prototype.once = function(type, listener) {
      if (!isFunction(listener))
        throw TypeError('listener must be a function');

      var fired = false;

      function g() {
        this.removeListener(type, g);

        if (!fired) {
          fired = true;
          listener.apply(this, arguments);
        }
      }

      g.listener = listener;
      this.on(type, g);

      return this;
    };

    // emits a 'removeListener' event iff the listener was removed
    EventEmitter.prototype.removeListener = function(type, listener) {
      var list, position, length, i;

      if (!isFunction(listener))
        throw TypeError('listener must be a function');

      if (!this._events || !this._events[type])
        return this;

      list = this._events[type];
      length = list.length;
      position = -1;

      if (list === listener ||
          (isFunction(list.listener) && list.listener === listener)) {
        delete this._events[type];
        if (this._events.removeListener)
          this.emit('removeListener', type, listener);

      } else if (isObject(list)) {
        for (i = length; i-- > 0;) {
          if (list[i] === listener ||
              (list[i].listener && list[i].listener === listener)) {
            position = i;
            break;
          }
        }

        if (position < 0)
          return this;

        if (list.length === 1) {
          list.length = 0;
          delete this._events[type];
        } else {
          list.splice(position, 1);
        }

        if (this._events.removeListener)
          this.emit('removeListener', type, listener);
      }

      return this;
    };

    EventEmitter.prototype.removeAllListeners = function(type) {
      var key, listeners;

      if (!this._events)
        return this;

      // not listening for removeListener, no need to emit
      if (!this._events.removeListener) {
        if (arguments.length === 0)
          this._events = {};
        else if (this._events[type])
          delete this._events[type];
        return this;
      }

      // emit removeListener for all listeners on all events
      if (arguments.length === 0) {
        for (key in this._events) {
          if (key === 'removeListener') continue;
          this.removeAllListeners(key);
        }
        this.removeAllListeners('removeListener');
        this._events = {};
        return this;
      }

      listeners = this._events[type];

      if (isFunction(listeners)) {
        this.removeListener(type, listeners);
      } else if (listeners) {
        // LIFO order
        while (listeners.length)
          this.removeListener(type, listeners[listeners.length - 1]);
      }
      delete this._events[type];

      return this;
    };

    EventEmitter.prototype.listeners = function(type) {
      var ret;
      if (!this._events || !this._events[type])
        ret = [];
      else if (isFunction(this._events[type]))
        ret = [this._events[type]];
      else
        ret = this._events[type].slice();
      return ret;
    };

    EventEmitter.prototype.listenerCount = function(type) {
      if (this._events) {
        var evlistener = this._events[type];

        if (isFunction(evlistener))
          return 1;
        else if (evlistener)
          return evlistener.length;
      }
      return 0;
    };

    EventEmitter.listenerCount = function(emitter, type) {
      return emitter.listenerCount(type);
    };

    function isFunction(arg) {
      return typeof arg === 'function';
    }

    function isNumber(arg) {
      return typeof arg === 'number';
    }

    function isObject(arg) {
      return typeof arg === 'object' && arg !== null;
    }

    function isUndefined(arg) {
      return arg === void 0;
    }

    function BufferList() {
      this.head = null;
      this.tail = null;
      this.length = 0;
    }

    BufferList.prototype.push = function (v) {
      var entry = { data: v, next: null };
      if (this.length > 0) this.tail.next = entry;else this.head = entry;
      this.tail = entry;
      ++this.length;
    };

    BufferList.prototype.unshift = function (v) {
      var entry = { data: v, next: this.head };
      if (this.length === 0) this.tail = entry;
      this.head = entry;
      ++this.length;
    };

    BufferList.prototype.shift = function () {
      if (this.length === 0) return;
      var ret = this.head.data;
      if (this.length === 1) this.head = this.tail = null;else this.head = this.head.next;
      --this.length;
      return ret;
    };

    BufferList.prototype.clear = function () {
      this.head = this.tail = null;
      this.length = 0;
    };

    BufferList.prototype.join = function (s) {
      if (this.length === 0) return '';
      var p = this.head;
      var ret = '' + p.data;
      while (p = p.next) {
        ret += s + p.data;
      }return ret;
    };

    BufferList.prototype.concat = function (n) {
      if (this.length === 0) return Buffer$1.alloc(0);
      if (this.length === 1) return this.head.data;
      var ret = Buffer$1.allocUnsafe(n >>> 0);
      var p = this.head;
      var i = 0;
      while (p) {
        p.data.copy(ret, i);
        i += p.data.length;
        p = p.next;
      }
      return ret;
    };

    var safeBuffer = {exports: {}};

    var require$$0$3 = /*@__PURE__*/getAugmentedNamespace(bufferEs6);

    /* eslint-disable node/no-deprecated-api */

    (function (module, exports) {
    	var buffer = require$$0$3;
    	var Buffer = buffer.Buffer;

    	// alternative to using Object.keys for old browsers
    	function copyProps (src, dst) {
    	  for (var key in src) {
    	    dst[key] = src[key];
    	  }
    	}
    	if (Buffer.from && Buffer.alloc && Buffer.allocUnsafe && Buffer.allocUnsafeSlow) {
    	  module.exports = buffer;
    	} else {
    	  // Copy properties from require('buffer')
    	  copyProps(buffer, exports);
    	  exports.Buffer = SafeBuffer;
    	}

    	function SafeBuffer (arg, encodingOrOffset, length) {
    	  return Buffer(arg, encodingOrOffset, length)
    	}

    	// Copy static methods from Buffer
    	copyProps(Buffer, SafeBuffer);

    	SafeBuffer.from = function (arg, encodingOrOffset, length) {
    	  if (typeof arg === 'number') {
    	    throw new TypeError('Argument must not be a number')
    	  }
    	  return Buffer(arg, encodingOrOffset, length)
    	};

    	SafeBuffer.alloc = function (size, fill, encoding) {
    	  if (typeof size !== 'number') {
    	    throw new TypeError('Argument must be a number')
    	  }
    	  var buf = Buffer(size);
    	  if (fill !== undefined) {
    	    if (typeof encoding === 'string') {
    	      buf.fill(fill, encoding);
    	    } else {
    	      buf.fill(fill);
    	    }
    	  } else {
    	    buf.fill(0);
    	  }
    	  return buf
    	};

    	SafeBuffer.allocUnsafe = function (size) {
    	  if (typeof size !== 'number') {
    	    throw new TypeError('Argument must be a number')
    	  }
    	  return Buffer(size)
    	};

    	SafeBuffer.allocUnsafeSlow = function (size) {
    	  if (typeof size !== 'number') {
    	    throw new TypeError('Argument must be a number')
    	  }
    	  return buffer.SlowBuffer(size)
    	};
    } (safeBuffer, safeBuffer.exports));

    /*<replacement>*/

    var Buffer = safeBuffer.exports.Buffer;
    /*</replacement>*/

    var isEncoding = Buffer.isEncoding || function (encoding) {
      encoding = '' + encoding;
      switch (encoding && encoding.toLowerCase()) {
        case 'hex':case 'utf8':case 'utf-8':case 'ascii':case 'binary':case 'base64':case 'ucs2':case 'ucs-2':case 'utf16le':case 'utf-16le':case 'raw':
          return true;
        default:
          return false;
      }
    };

    function _normalizeEncoding(enc) {
      if (!enc) return 'utf8';
      var retried;
      while (true) {
        switch (enc) {
          case 'utf8':
          case 'utf-8':
            return 'utf8';
          case 'ucs2':
          case 'ucs-2':
          case 'utf16le':
          case 'utf-16le':
            return 'utf16le';
          case 'latin1':
          case 'binary':
            return 'latin1';
          case 'base64':
          case 'ascii':
          case 'hex':
            return enc;
          default:
            if (retried) return; // undefined
            enc = ('' + enc).toLowerCase();
            retried = true;
        }
      }
    }
    // Do not cache `Buffer.isEncoding` when checking encoding names as some
    // modules monkey-patch it to support additional encodings
    function normalizeEncoding(enc) {
      var nenc = _normalizeEncoding(enc);
      if (typeof nenc !== 'string' && (Buffer.isEncoding === isEncoding || !isEncoding(enc))) throw new Error('Unknown encoding: ' + enc);
      return nenc || enc;
    }

    // StringDecoder provides an interface for efficiently splitting a series of
    // buffers into a series of JS strings without breaking apart multi-byte
    // characters.
    var StringDecoder_1 = StringDecoder;
    function StringDecoder(encoding) {
      this.encoding = normalizeEncoding(encoding);
      var nb;
      switch (this.encoding) {
        case 'utf16le':
          this.text = utf16Text;
          this.end = utf16End;
          nb = 4;
          break;
        case 'utf8':
          this.fillLast = utf8FillLast;
          nb = 4;
          break;
        case 'base64':
          this.text = base64Text;
          this.end = base64End;
          nb = 3;
          break;
        default:
          this.write = simpleWrite;
          this.end = simpleEnd;
          return;
      }
      this.lastNeed = 0;
      this.lastTotal = 0;
      this.lastChar = Buffer.allocUnsafe(nb);
    }

    StringDecoder.prototype.write = function (buf) {
      if (buf.length === 0) return '';
      var r;
      var i;
      if (this.lastNeed) {
        r = this.fillLast(buf);
        if (r === undefined) return '';
        i = this.lastNeed;
        this.lastNeed = 0;
      } else {
        i = 0;
      }
      if (i < buf.length) return r ? r + this.text(buf, i) : this.text(buf, i);
      return r || '';
    };

    StringDecoder.prototype.end = utf8End;

    // Returns only complete characters in a Buffer
    StringDecoder.prototype.text = utf8Text;

    // Attempts to complete a partial non-UTF-8 character using bytes from a Buffer
    StringDecoder.prototype.fillLast = function (buf) {
      if (this.lastNeed <= buf.length) {
        buf.copy(this.lastChar, this.lastTotal - this.lastNeed, 0, this.lastNeed);
        return this.lastChar.toString(this.encoding, 0, this.lastTotal);
      }
      buf.copy(this.lastChar, this.lastTotal - this.lastNeed, 0, buf.length);
      this.lastNeed -= buf.length;
    };

    // Checks the type of a UTF-8 byte, whether it's ASCII, a leading byte, or a
    // continuation byte. If an invalid byte is detected, -2 is returned.
    function utf8CheckByte(byte) {
      if (byte <= 0x7F) return 0;else if (byte >> 5 === 0x06) return 2;else if (byte >> 4 === 0x0E) return 3;else if (byte >> 3 === 0x1E) return 4;
      return byte >> 6 === 0x02 ? -1 : -2;
    }

    // Checks at most 3 bytes at the end of a Buffer in order to detect an
    // incomplete multi-byte UTF-8 character. The total number of bytes (2, 3, or 4)
    // needed to complete the UTF-8 character (if applicable) are returned.
    function utf8CheckIncomplete(self, buf, i) {
      var j = buf.length - 1;
      if (j < i) return 0;
      var nb = utf8CheckByte(buf[j]);
      if (nb >= 0) {
        if (nb > 0) self.lastNeed = nb - 1;
        return nb;
      }
      if (--j < i || nb === -2) return 0;
      nb = utf8CheckByte(buf[j]);
      if (nb >= 0) {
        if (nb > 0) self.lastNeed = nb - 2;
        return nb;
      }
      if (--j < i || nb === -2) return 0;
      nb = utf8CheckByte(buf[j]);
      if (nb >= 0) {
        if (nb > 0) {
          if (nb === 2) nb = 0;else self.lastNeed = nb - 3;
        }
        return nb;
      }
      return 0;
    }

    // Validates as many continuation bytes for a multi-byte UTF-8 character as
    // needed or are available. If we see a non-continuation byte where we expect
    // one, we "replace" the validated continuation bytes we've seen so far with
    // a single UTF-8 replacement character ('\ufffd'), to match v8's UTF-8 decoding
    // behavior. The continuation byte check is included three times in the case
    // where all of the continuation bytes for a character exist in the same buffer.
    // It is also done this way as a slight performance increase instead of using a
    // loop.
    function utf8CheckExtraBytes(self, buf, p) {
      if ((buf[0] & 0xC0) !== 0x80) {
        self.lastNeed = 0;
        return '\ufffd';
      }
      if (self.lastNeed > 1 && buf.length > 1) {
        if ((buf[1] & 0xC0) !== 0x80) {
          self.lastNeed = 1;
          return '\ufffd';
        }
        if (self.lastNeed > 2 && buf.length > 2) {
          if ((buf[2] & 0xC0) !== 0x80) {
            self.lastNeed = 2;
            return '\ufffd';
          }
        }
      }
    }

    // Attempts to complete a multi-byte UTF-8 character using bytes from a Buffer.
    function utf8FillLast(buf) {
      var p = this.lastTotal - this.lastNeed;
      var r = utf8CheckExtraBytes(this, buf);
      if (r !== undefined) return r;
      if (this.lastNeed <= buf.length) {
        buf.copy(this.lastChar, p, 0, this.lastNeed);
        return this.lastChar.toString(this.encoding, 0, this.lastTotal);
      }
      buf.copy(this.lastChar, p, 0, buf.length);
      this.lastNeed -= buf.length;
    }

    // Returns all complete UTF-8 characters in a Buffer. If the Buffer ended on a
    // partial character, the character's bytes are buffered until the required
    // number of bytes are available.
    function utf8Text(buf, i) {
      var total = utf8CheckIncomplete(this, buf, i);
      if (!this.lastNeed) return buf.toString('utf8', i);
      this.lastTotal = total;
      var end = buf.length - (total - this.lastNeed);
      buf.copy(this.lastChar, 0, end);
      return buf.toString('utf8', i, end);
    }

    // For UTF-8, a replacement character is added when ending on a partial
    // character.
    function utf8End(buf) {
      var r = buf && buf.length ? this.write(buf) : '';
      if (this.lastNeed) return r + '\ufffd';
      return r;
    }

    // UTF-16LE typically needs two bytes per character, but even if we have an even
    // number of bytes available, we need to check if we end on a leading/high
    // surrogate. In that case, we need to wait for the next two bytes in order to
    // decode the last character properly.
    function utf16Text(buf, i) {
      if ((buf.length - i) % 2 === 0) {
        var r = buf.toString('utf16le', i);
        if (r) {
          var c = r.charCodeAt(r.length - 1);
          if (c >= 0xD800 && c <= 0xDBFF) {
            this.lastNeed = 2;
            this.lastTotal = 4;
            this.lastChar[0] = buf[buf.length - 2];
            this.lastChar[1] = buf[buf.length - 1];
            return r.slice(0, -1);
          }
        }
        return r;
      }
      this.lastNeed = 1;
      this.lastTotal = 2;
      this.lastChar[0] = buf[buf.length - 1];
      return buf.toString('utf16le', i, buf.length - 1);
    }

    // For UTF-16LE we do not explicitly append special replacement characters if we
    // end on a partial character, we simply let v8 handle that.
    function utf16End(buf) {
      var r = buf && buf.length ? this.write(buf) : '';
      if (this.lastNeed) {
        var end = this.lastTotal - this.lastNeed;
        return r + this.lastChar.toString('utf16le', 0, end);
      }
      return r;
    }

    function base64Text(buf, i) {
      var n = (buf.length - i) % 3;
      if (n === 0) return buf.toString('base64', i);
      this.lastNeed = 3 - n;
      this.lastTotal = 3;
      if (n === 1) {
        this.lastChar[0] = buf[buf.length - 1];
      } else {
        this.lastChar[0] = buf[buf.length - 2];
        this.lastChar[1] = buf[buf.length - 1];
      }
      return buf.toString('base64', i, buf.length - n);
    }

    function base64End(buf) {
      var r = buf && buf.length ? this.write(buf) : '';
      if (this.lastNeed) return r + this.lastChar.toString('base64', 0, 3 - this.lastNeed);
      return r;
    }

    // Pass bytes on through for single-byte encodings (e.g. ascii, latin1, hex)
    function simpleWrite(buf) {
      return buf.toString(this.encoding);
    }

    function simpleEnd(buf) {
      return buf && buf.length ? this.write(buf) : '';
    }

    Readable.ReadableState = ReadableState;

    var debug = debuglog('stream');
    inherits$1(Readable, events);

    function prependListener(emitter, event, fn) {
      // Sadly this is not cacheable as some libraries bundle their own
      // event emitter implementation with them.
      if (typeof emitter.prependListener === 'function') {
        return emitter.prependListener(event, fn);
      } else {
        // This is a hack to make sure that our error handler is attached before any
        // userland ones.  NEVER DO THIS. This is here only because this code needs
        // to continue to work with older versions of Node.js that do not include
        // the prependListener() method. The goal is to eventually remove this hack.
        if (!emitter._events || !emitter._events[event])
          emitter.on(event, fn);
        else if (Array.isArray(emitter._events[event]))
          emitter._events[event].unshift(fn);
        else
          emitter._events[event] = [fn, emitter._events[event]];
      }
    }
    function listenerCount (emitter, type) {
      return emitter.listeners(type).length;
    }
    function ReadableState(options, stream) {

      options = options || {};

      // object stream flag. Used to make read(n) ignore n and to
      // make all the buffer merging and length checks go away
      this.objectMode = !!options.objectMode;

      if (stream instanceof Duplex) this.objectMode = this.objectMode || !!options.readableObjectMode;

      // the point at which it stops calling _read() to fill the buffer
      // Note: 0 is a valid value, means "don't call _read preemptively ever"
      var hwm = options.highWaterMark;
      var defaultHwm = this.objectMode ? 16 : 16 * 1024;
      this.highWaterMark = hwm || hwm === 0 ? hwm : defaultHwm;

      // cast to ints.
      this.highWaterMark = ~ ~this.highWaterMark;

      // A linked list is used to store data chunks instead of an array because the
      // linked list can remove elements from the beginning faster than
      // array.shift()
      this.buffer = new BufferList();
      this.length = 0;
      this.pipes = null;
      this.pipesCount = 0;
      this.flowing = null;
      this.ended = false;
      this.endEmitted = false;
      this.reading = false;

      // a flag to be able to tell if the onwrite cb is called immediately,
      // or on a later tick.  We set this to true at first, because any
      // actions that shouldn't happen until "later" should generally also
      // not happen before the first write call.
      this.sync = true;

      // whenever we return null, then we set a flag to say
      // that we're awaiting a 'readable' event emission.
      this.needReadable = false;
      this.emittedReadable = false;
      this.readableListening = false;
      this.resumeScheduled = false;

      // Crypto is kind of old and crusty.  Historically, its default string
      // encoding is 'binary' so we have to make this configurable.
      // Everything else in the universe uses 'utf8', though.
      this.defaultEncoding = options.defaultEncoding || 'utf8';

      // when piping, we only care about 'readable' events that happen
      // after read()ing all the bytes and not getting any pushback.
      this.ranOut = false;

      // the number of writers that are awaiting a drain event in .pipe()s
      this.awaitDrain = 0;

      // if true, a maybeReadMore has been scheduled
      this.readingMore = false;

      this.decoder = null;
      this.encoding = null;
      if (options.encoding) {
        this.decoder = new StringDecoder_1(options.encoding);
        this.encoding = options.encoding;
      }
    }
    function Readable(options) {

      if (!(this instanceof Readable)) return new Readable(options);

      this._readableState = new ReadableState(options, this);

      // legacy
      this.readable = true;

      if (options && typeof options.read === 'function') this._read = options.read;

      events.call(this);
    }

    // Manually shove something into the read() buffer.
    // This returns true if the highWaterMark has not been hit yet,
    // similar to how Writable.write() returns true if you should
    // write() some more.
    Readable.prototype.push = function (chunk, encoding) {
      var state = this._readableState;

      if (!state.objectMode && typeof chunk === 'string') {
        encoding = encoding || state.defaultEncoding;
        if (encoding !== state.encoding) {
          chunk = Buffer$1.from(chunk, encoding);
          encoding = '';
        }
      }

      return readableAddChunk(this, state, chunk, encoding, false);
    };

    // Unshift should *always* be something directly out of read()
    Readable.prototype.unshift = function (chunk) {
      var state = this._readableState;
      return readableAddChunk(this, state, chunk, '', true);
    };

    Readable.prototype.isPaused = function () {
      return this._readableState.flowing === false;
    };

    function readableAddChunk(stream, state, chunk, encoding, addToFront) {
      var er = chunkInvalid(state, chunk);
      if (er) {
        stream.emit('error', er);
      } else if (chunk === null) {
        state.reading = false;
        onEofChunk(stream, state);
      } else if (state.objectMode || chunk && chunk.length > 0) {
        if (state.ended && !addToFront) {
          var e = new Error('stream.push() after EOF');
          stream.emit('error', e);
        } else if (state.endEmitted && addToFront) {
          var _e = new Error('stream.unshift() after end event');
          stream.emit('error', _e);
        } else {
          var skipAdd;
          if (state.decoder && !addToFront && !encoding) {
            chunk = state.decoder.write(chunk);
            skipAdd = !state.objectMode && chunk.length === 0;
          }

          if (!addToFront) state.reading = false;

          // Don't add to the buffer if we've decoded to an empty string chunk and
          // we're not in object mode
          if (!skipAdd) {
            // if we want the data now, just emit it.
            if (state.flowing && state.length === 0 && !state.sync) {
              stream.emit('data', chunk);
              stream.read(0);
            } else {
              // update the buffer info.
              state.length += state.objectMode ? 1 : chunk.length;
              if (addToFront) state.buffer.unshift(chunk);else state.buffer.push(chunk);

              if (state.needReadable) emitReadable(stream);
            }
          }

          maybeReadMore(stream, state);
        }
      } else if (!addToFront) {
        state.reading = false;
      }

      return needMoreData(state);
    }

    // if it's past the high water mark, we can push in some more.
    // Also, if we have no data yet, we can stand some
    // more bytes.  This is to work around cases where hwm=0,
    // such as the repl.  Also, if the push() triggered a
    // readable event, and the user called read(largeNumber) such that
    // needReadable was set, then we ought to push more, so that another
    // 'readable' event will be triggered.
    function needMoreData(state) {
      return !state.ended && (state.needReadable || state.length < state.highWaterMark || state.length === 0);
    }

    // backwards compatibility.
    Readable.prototype.setEncoding = function (enc) {
      this._readableState.decoder = new StringDecoder_1(enc);
      this._readableState.encoding = enc;
      return this;
    };

    // Don't raise the hwm > 8MB
    var MAX_HWM = 0x800000;
    function computeNewHighWaterMark(n) {
      if (n >= MAX_HWM) {
        n = MAX_HWM;
      } else {
        // Get the next highest power of 2 to prevent increasing hwm excessively in
        // tiny amounts
        n--;
        n |= n >>> 1;
        n |= n >>> 2;
        n |= n >>> 4;
        n |= n >>> 8;
        n |= n >>> 16;
        n++;
      }
      return n;
    }

    // This function is designed to be inlinable, so please take care when making
    // changes to the function body.
    function howMuchToRead(n, state) {
      if (n <= 0 || state.length === 0 && state.ended) return 0;
      if (state.objectMode) return 1;
      if (n !== n) {
        // Only flow one buffer at a time
        if (state.flowing && state.length) return state.buffer.head.data.length;else return state.length;
      }
      // If we're asking for more than the current hwm, then raise the hwm.
      if (n > state.highWaterMark) state.highWaterMark = computeNewHighWaterMark(n);
      if (n <= state.length) return n;
      // Don't have enough
      if (!state.ended) {
        state.needReadable = true;
        return 0;
      }
      return state.length;
    }

    // you can override either this method, or the async _read(n) below.
    Readable.prototype.read = function (n) {
      debug('read', n);
      n = parseInt(n, 10);
      var state = this._readableState;
      var nOrig = n;

      if (n !== 0) state.emittedReadable = false;

      // if we're doing read(0) to trigger a readable event, but we
      // already have a bunch of data in the buffer, then just trigger
      // the 'readable' event and move on.
      if (n === 0 && state.needReadable && (state.length >= state.highWaterMark || state.ended)) {
        debug('read: emitReadable', state.length, state.ended);
        if (state.length === 0 && state.ended) endReadable(this);else emitReadable(this);
        return null;
      }

      n = howMuchToRead(n, state);

      // if we've ended, and we're now clear, then finish it up.
      if (n === 0 && state.ended) {
        if (state.length === 0) endReadable(this);
        return null;
      }

      // All the actual chunk generation logic needs to be
      // *below* the call to _read.  The reason is that in certain
      // synthetic stream cases, such as passthrough streams, _read
      // may be a completely synchronous operation which may change
      // the state of the read buffer, providing enough data when
      // before there was *not* enough.
      //
      // So, the steps are:
      // 1. Figure out what the state of things will be after we do
      // a read from the buffer.
      //
      // 2. If that resulting state will trigger a _read, then call _read.
      // Note that this may be asynchronous, or synchronous.  Yes, it is
      // deeply ugly to write APIs this way, but that still doesn't mean
      // that the Readable class should behave improperly, as streams are
      // designed to be sync/async agnostic.
      // Take note if the _read call is sync or async (ie, if the read call
      // has returned yet), so that we know whether or not it's safe to emit
      // 'readable' etc.
      //
      // 3. Actually pull the requested chunks out of the buffer and return.

      // if we need a readable event, then we need to do some reading.
      var doRead = state.needReadable;
      debug('need readable', doRead);

      // if we currently have less than the highWaterMark, then also read some
      if (state.length === 0 || state.length - n < state.highWaterMark) {
        doRead = true;
        debug('length less than watermark', doRead);
      }

      // however, if we've ended, then there's no point, and if we're already
      // reading, then it's unnecessary.
      if (state.ended || state.reading) {
        doRead = false;
        debug('reading or ended', doRead);
      } else if (doRead) {
        debug('do read');
        state.reading = true;
        state.sync = true;
        // if the length is currently zero, then we *need* a readable event.
        if (state.length === 0) state.needReadable = true;
        // call internal read method
        this._read(state.highWaterMark);
        state.sync = false;
        // If _read pushed data synchronously, then `reading` will be false,
        // and we need to re-evaluate how much data we can return to the user.
        if (!state.reading) n = howMuchToRead(nOrig, state);
      }

      var ret;
      if (n > 0) ret = fromList(n, state);else ret = null;

      if (ret === null) {
        state.needReadable = true;
        n = 0;
      } else {
        state.length -= n;
      }

      if (state.length === 0) {
        // If we have nothing in the buffer, then we want to know
        // as soon as we *do* get something into the buffer.
        if (!state.ended) state.needReadable = true;

        // If we tried to read() past the EOF, then emit end on the next tick.
        if (nOrig !== n && state.ended) endReadable(this);
      }

      if (ret !== null) this.emit('data', ret);

      return ret;
    };

    function chunkInvalid(state, chunk) {
      var er = null;
      if (!isBuffer$2(chunk) && typeof chunk !== 'string' && chunk !== null && chunk !== undefined && !state.objectMode) {
        er = new TypeError('Invalid non-string/buffer chunk');
      }
      return er;
    }

    function onEofChunk(stream, state) {
      if (state.ended) return;
      if (state.decoder) {
        var chunk = state.decoder.end();
        if (chunk && chunk.length) {
          state.buffer.push(chunk);
          state.length += state.objectMode ? 1 : chunk.length;
        }
      }
      state.ended = true;

      // emit 'readable' now to make sure it gets picked up.
      emitReadable(stream);
    }

    // Don't emit readable right away in sync mode, because this can trigger
    // another read() call => stack overflow.  This way, it might trigger
    // a nextTick recursion warning, but that's not so bad.
    function emitReadable(stream) {
      var state = stream._readableState;
      state.needReadable = false;
      if (!state.emittedReadable) {
        debug('emitReadable', state.flowing);
        state.emittedReadable = true;
        if (state.sync) nextTick(emitReadable_, stream);else emitReadable_(stream);
      }
    }

    function emitReadable_(stream) {
      debug('emit readable');
      stream.emit('readable');
      flow(stream);
    }

    // at this point, the user has presumably seen the 'readable' event,
    // and called read() to consume some data.  that may have triggered
    // in turn another _read(n) call, in which case reading = true if
    // it's in progress.
    // However, if we're not ended, or reading, and the length < hwm,
    // then go ahead and try to read some more preemptively.
    function maybeReadMore(stream, state) {
      if (!state.readingMore) {
        state.readingMore = true;
        nextTick(maybeReadMore_, stream, state);
      }
    }

    function maybeReadMore_(stream, state) {
      var len = state.length;
      while (!state.reading && !state.flowing && !state.ended && state.length < state.highWaterMark) {
        debug('maybeReadMore read 0');
        stream.read(0);
        if (len === state.length)
          // didn't get any data, stop spinning.
          break;else len = state.length;
      }
      state.readingMore = false;
    }

    // abstract method.  to be overridden in specific implementation classes.
    // call cb(er, data) where data is <= n in length.
    // for virtual (non-string, non-buffer) streams, "length" is somewhat
    // arbitrary, and perhaps not very meaningful.
    Readable.prototype._read = function (n) {
      this.emit('error', new Error('not implemented'));
    };

    Readable.prototype.pipe = function (dest, pipeOpts) {
      var src = this;
      var state = this._readableState;

      switch (state.pipesCount) {
        case 0:
          state.pipes = dest;
          break;
        case 1:
          state.pipes = [state.pipes, dest];
          break;
        default:
          state.pipes.push(dest);
          break;
      }
      state.pipesCount += 1;
      debug('pipe count=%d opts=%j', state.pipesCount, pipeOpts);

      var doEnd = (!pipeOpts || pipeOpts.end !== false);

      var endFn = doEnd ? onend : cleanup;
      if (state.endEmitted) nextTick(endFn);else src.once('end', endFn);

      dest.on('unpipe', onunpipe);
      function onunpipe(readable) {
        debug('onunpipe');
        if (readable === src) {
          cleanup();
        }
      }

      function onend() {
        debug('onend');
        dest.end();
      }

      // when the dest drains, it reduces the awaitDrain counter
      // on the source.  This would be more elegant with a .once()
      // handler in flow(), but adding and removing repeatedly is
      // too slow.
      var ondrain = pipeOnDrain(src);
      dest.on('drain', ondrain);

      var cleanedUp = false;
      function cleanup() {
        debug('cleanup');
        // cleanup event handlers once the pipe is broken
        dest.removeListener('close', onclose);
        dest.removeListener('finish', onfinish);
        dest.removeListener('drain', ondrain);
        dest.removeListener('error', onerror);
        dest.removeListener('unpipe', onunpipe);
        src.removeListener('end', onend);
        src.removeListener('end', cleanup);
        src.removeListener('data', ondata);

        cleanedUp = true;

        // if the reader is waiting for a drain event from this
        // specific writer, then it would cause it to never start
        // flowing again.
        // So, if this is awaiting a drain, then we just call it now.
        // If we don't know, then assume that we are waiting for one.
        if (state.awaitDrain && (!dest._writableState || dest._writableState.needDrain)) ondrain();
      }

      // If the user pushes more data while we're writing to dest then we'll end up
      // in ondata again. However, we only want to increase awaitDrain once because
      // dest will only emit one 'drain' event for the multiple writes.
      // => Introduce a guard on increasing awaitDrain.
      var increasedAwaitDrain = false;
      src.on('data', ondata);
      function ondata(chunk) {
        debug('ondata');
        increasedAwaitDrain = false;
        var ret = dest.write(chunk);
        if (false === ret && !increasedAwaitDrain) {
          // If the user unpiped during `dest.write()`, it is possible
          // to get stuck in a permanently paused state if that write
          // also returned false.
          // => Check whether `dest` is still a piping destination.
          if ((state.pipesCount === 1 && state.pipes === dest || state.pipesCount > 1 && indexOf(state.pipes, dest) !== -1) && !cleanedUp) {
            debug('false write response, pause', src._readableState.awaitDrain);
            src._readableState.awaitDrain++;
            increasedAwaitDrain = true;
          }
          src.pause();
        }
      }

      // if the dest has an error, then stop piping into it.
      // however, don't suppress the throwing behavior for this.
      function onerror(er) {
        debug('onerror', er);
        unpipe();
        dest.removeListener('error', onerror);
        if (listenerCount(dest, 'error') === 0) dest.emit('error', er);
      }

      // Make sure our error handler is attached before userland ones.
      prependListener(dest, 'error', onerror);

      // Both close and finish should trigger unpipe, but only once.
      function onclose() {
        dest.removeListener('finish', onfinish);
        unpipe();
      }
      dest.once('close', onclose);
      function onfinish() {
        debug('onfinish');
        dest.removeListener('close', onclose);
        unpipe();
      }
      dest.once('finish', onfinish);

      function unpipe() {
        debug('unpipe');
        src.unpipe(dest);
      }

      // tell the dest that it's being piped to
      dest.emit('pipe', src);

      // start the flow if it hasn't been started already.
      if (!state.flowing) {
        debug('pipe resume');
        src.resume();
      }

      return dest;
    };

    function pipeOnDrain(src) {
      return function () {
        var state = src._readableState;
        debug('pipeOnDrain', state.awaitDrain);
        if (state.awaitDrain) state.awaitDrain--;
        if (state.awaitDrain === 0 && src.listeners('data').length) {
          state.flowing = true;
          flow(src);
        }
      };
    }

    Readable.prototype.unpipe = function (dest) {
      var state = this._readableState;

      // if we're not piping anywhere, then do nothing.
      if (state.pipesCount === 0) return this;

      // just one destination.  most common case.
      if (state.pipesCount === 1) {
        // passed in one, but it's not the right one.
        if (dest && dest !== state.pipes) return this;

        if (!dest) dest = state.pipes;

        // got a match.
        state.pipes = null;
        state.pipesCount = 0;
        state.flowing = false;
        if (dest) dest.emit('unpipe', this);
        return this;
      }

      // slow case. multiple pipe destinations.

      if (!dest) {
        // remove all.
        var dests = state.pipes;
        var len = state.pipesCount;
        state.pipes = null;
        state.pipesCount = 0;
        state.flowing = false;

        for (var _i = 0; _i < len; _i++) {
          dests[_i].emit('unpipe', this);
        }return this;
      }

      // try to find the right one.
      var i = indexOf(state.pipes, dest);
      if (i === -1) return this;

      state.pipes.splice(i, 1);
      state.pipesCount -= 1;
      if (state.pipesCount === 1) state.pipes = state.pipes[0];

      dest.emit('unpipe', this);

      return this;
    };

    // set up data events if they are asked for
    // Ensure readable listeners eventually get something
    Readable.prototype.on = function (ev, fn) {
      var res = events.prototype.on.call(this, ev, fn);

      if (ev === 'data') {
        // Start flowing on next tick if stream isn't explicitly paused
        if (this._readableState.flowing !== false) this.resume();
      } else if (ev === 'readable') {
        var state = this._readableState;
        if (!state.endEmitted && !state.readableListening) {
          state.readableListening = state.needReadable = true;
          state.emittedReadable = false;
          if (!state.reading) {
            nextTick(nReadingNextTick, this);
          } else if (state.length) {
            emitReadable(this);
          }
        }
      }

      return res;
    };
    Readable.prototype.addListener = Readable.prototype.on;

    function nReadingNextTick(self) {
      debug('readable nexttick read 0');
      self.read(0);
    }

    // pause() and resume() are remnants of the legacy readable stream API
    // If the user uses them, then switch into old mode.
    Readable.prototype.resume = function () {
      var state = this._readableState;
      if (!state.flowing) {
        debug('resume');
        state.flowing = true;
        resume(this, state);
      }
      return this;
    };

    function resume(stream, state) {
      if (!state.resumeScheduled) {
        state.resumeScheduled = true;
        nextTick(resume_, stream, state);
      }
    }

    function resume_(stream, state) {
      if (!state.reading) {
        debug('resume read 0');
        stream.read(0);
      }

      state.resumeScheduled = false;
      state.awaitDrain = 0;
      stream.emit('resume');
      flow(stream);
      if (state.flowing && !state.reading) stream.read(0);
    }

    Readable.prototype.pause = function () {
      debug('call pause flowing=%j', this._readableState.flowing);
      if (false !== this._readableState.flowing) {
        debug('pause');
        this._readableState.flowing = false;
        this.emit('pause');
      }
      return this;
    };

    function flow(stream) {
      var state = stream._readableState;
      debug('flow', state.flowing);
      while (state.flowing && stream.read() !== null) {}
    }

    // wrap an old-style stream as the async data source.
    // This is *not* part of the readable stream interface.
    // It is an ugly unfortunate mess of history.
    Readable.prototype.wrap = function (stream) {
      var state = this._readableState;
      var paused = false;

      var self = this;
      stream.on('end', function () {
        debug('wrapped end');
        if (state.decoder && !state.ended) {
          var chunk = state.decoder.end();
          if (chunk && chunk.length) self.push(chunk);
        }

        self.push(null);
      });

      stream.on('data', function (chunk) {
        debug('wrapped data');
        if (state.decoder) chunk = state.decoder.write(chunk);

        // don't skip over falsy values in objectMode
        if (state.objectMode && (chunk === null || chunk === undefined)) return;else if (!state.objectMode && (!chunk || !chunk.length)) return;

        var ret = self.push(chunk);
        if (!ret) {
          paused = true;
          stream.pause();
        }
      });

      // proxy all the other methods.
      // important when wrapping filters and duplexes.
      for (var i in stream) {
        if (this[i] === undefined && typeof stream[i] === 'function') {
          this[i] = function (method) {
            return function () {
              return stream[method].apply(stream, arguments);
            };
          }(i);
        }
      }

      // proxy certain important events.
      var events = ['error', 'close', 'destroy', 'pause', 'resume'];
      forEach(events, function (ev) {
        stream.on(ev, self.emit.bind(self, ev));
      });

      // when we try to consume some more bytes, simply unpause the
      // underlying stream.
      self._read = function (n) {
        debug('wrapped _read', n);
        if (paused) {
          paused = false;
          stream.resume();
        }
      };

      return self;
    };

    // exposed for testing purposes only.
    Readable._fromList = fromList;

    // Pluck off n bytes from an array of buffers.
    // Length is the combined lengths of all the buffers in the list.
    // This function is designed to be inlinable, so please take care when making
    // changes to the function body.
    function fromList(n, state) {
      // nothing buffered
      if (state.length === 0) return null;

      var ret;
      if (state.objectMode) ret = state.buffer.shift();else if (!n || n >= state.length) {
        // read it all, truncate the list
        if (state.decoder) ret = state.buffer.join('');else if (state.buffer.length === 1) ret = state.buffer.head.data;else ret = state.buffer.concat(state.length);
        state.buffer.clear();
      } else {
        // read part of list
        ret = fromListPartial(n, state.buffer, state.decoder);
      }

      return ret;
    }

    // Extracts only enough buffered data to satisfy the amount requested.
    // This function is designed to be inlinable, so please take care when making
    // changes to the function body.
    function fromListPartial(n, list, hasStrings) {
      var ret;
      if (n < list.head.data.length) {
        // slice is the same for buffers and strings
        ret = list.head.data.slice(0, n);
        list.head.data = list.head.data.slice(n);
      } else if (n === list.head.data.length) {
        // first chunk is a perfect match
        ret = list.shift();
      } else {
        // result spans more than one buffer
        ret = hasStrings ? copyFromBufferString(n, list) : copyFromBuffer(n, list);
      }
      return ret;
    }

    // Copies a specified amount of characters from the list of buffered data
    // chunks.
    // This function is designed to be inlinable, so please take care when making
    // changes to the function body.
    function copyFromBufferString(n, list) {
      var p = list.head;
      var c = 1;
      var ret = p.data;
      n -= ret.length;
      while (p = p.next) {
        var str = p.data;
        var nb = n > str.length ? str.length : n;
        if (nb === str.length) ret += str;else ret += str.slice(0, n);
        n -= nb;
        if (n === 0) {
          if (nb === str.length) {
            ++c;
            if (p.next) list.head = p.next;else list.head = list.tail = null;
          } else {
            list.head = p;
            p.data = str.slice(nb);
          }
          break;
        }
        ++c;
      }
      list.length -= c;
      return ret;
    }

    // Copies a specified amount of bytes from the list of buffered data chunks.
    // This function is designed to be inlinable, so please take care when making
    // changes to the function body.
    function copyFromBuffer(n, list) {
      var ret = Buffer$1.allocUnsafe(n);
      var p = list.head;
      var c = 1;
      p.data.copy(ret);
      n -= p.data.length;
      while (p = p.next) {
        var buf = p.data;
        var nb = n > buf.length ? buf.length : n;
        buf.copy(ret, ret.length - n, 0, nb);
        n -= nb;
        if (n === 0) {
          if (nb === buf.length) {
            ++c;
            if (p.next) list.head = p.next;else list.head = list.tail = null;
          } else {
            list.head = p;
            p.data = buf.slice(nb);
          }
          break;
        }
        ++c;
      }
      list.length -= c;
      return ret;
    }

    function endReadable(stream) {
      var state = stream._readableState;

      // If we get here before consuming all the bytes, then that is a
      // bug in node.  Should never happen.
      if (state.length > 0) throw new Error('"endReadable()" called on non-empty stream');

      if (!state.endEmitted) {
        state.ended = true;
        nextTick(endReadableNT, state, stream);
      }
    }

    function endReadableNT(state, stream) {
      // Check that we didn't get one last unshift.
      if (!state.endEmitted && state.length === 0) {
        state.endEmitted = true;
        stream.readable = false;
        stream.emit('end');
      }
    }

    function forEach(xs, f) {
      for (var i = 0, l = xs.length; i < l; i++) {
        f(xs[i], i);
      }
    }

    function indexOf(xs, x) {
      for (var i = 0, l = xs.length; i < l; i++) {
        if (xs[i] === x) return i;
      }
      return -1;
    }

    // A bit simpler than readable streams.
    Writable.WritableState = WritableState;
    inherits$1(Writable, events.EventEmitter);

    function nop() {}

    function WriteReq(chunk, encoding, cb) {
      this.chunk = chunk;
      this.encoding = encoding;
      this.callback = cb;
      this.next = null;
    }

    function WritableState(options, stream) {
      Object.defineProperty(this, 'buffer', {
        get: deprecate(function () {
          return this.getBuffer();
        }, '_writableState.buffer is deprecated. Use _writableState.getBuffer ' + 'instead.')
      });
      options = options || {};

      // object stream flag to indicate whether or not this stream
      // contains buffers or objects.
      this.objectMode = !!options.objectMode;

      if (stream instanceof Duplex) this.objectMode = this.objectMode || !!options.writableObjectMode;

      // the point at which write() starts returning false
      // Note: 0 is a valid value, means that we always return false if
      // the entire buffer is not flushed immediately on write()
      var hwm = options.highWaterMark;
      var defaultHwm = this.objectMode ? 16 : 16 * 1024;
      this.highWaterMark = hwm || hwm === 0 ? hwm : defaultHwm;

      // cast to ints.
      this.highWaterMark = ~ ~this.highWaterMark;

      this.needDrain = false;
      // at the start of calling end()
      this.ending = false;
      // when end() has been called, and returned
      this.ended = false;
      // when 'finish' is emitted
      this.finished = false;

      // should we decode strings into buffers before passing to _write?
      // this is here so that some node-core streams can optimize string
      // handling at a lower level.
      var noDecode = options.decodeStrings === false;
      this.decodeStrings = !noDecode;

      // Crypto is kind of old and crusty.  Historically, its default string
      // encoding is 'binary' so we have to make this configurable.
      // Everything else in the universe uses 'utf8', though.
      this.defaultEncoding = options.defaultEncoding || 'utf8';

      // not an actual buffer we keep track of, but a measurement
      // of how much we're waiting to get pushed to some underlying
      // socket or file.
      this.length = 0;

      // a flag to see when we're in the middle of a write.
      this.writing = false;

      // when true all writes will be buffered until .uncork() call
      this.corked = 0;

      // a flag to be able to tell if the onwrite cb is called immediately,
      // or on a later tick.  We set this to true at first, because any
      // actions that shouldn't happen until "later" should generally also
      // not happen before the first write call.
      this.sync = true;

      // a flag to know if we're processing previously buffered items, which
      // may call the _write() callback in the same tick, so that we don't
      // end up in an overlapped onwrite situation.
      this.bufferProcessing = false;

      // the callback that's passed to _write(chunk,cb)
      this.onwrite = function (er) {
        onwrite(stream, er);
      };

      // the callback that the user supplies to write(chunk,encoding,cb)
      this.writecb = null;

      // the amount that is being written when _write is called.
      this.writelen = 0;

      this.bufferedRequest = null;
      this.lastBufferedRequest = null;

      // number of pending user-supplied write callbacks
      // this must be 0 before 'finish' can be emitted
      this.pendingcb = 0;

      // emit prefinish if the only thing we're waiting for is _write cbs
      // This is relevant for synchronous Transform streams
      this.prefinished = false;

      // True if the error was already emitted and should not be thrown again
      this.errorEmitted = false;

      // count buffered requests
      this.bufferedRequestCount = 0;

      // allocate the first CorkedRequest, there is always
      // one allocated and free to use, and we maintain at most two
      this.corkedRequestsFree = new CorkedRequest(this);
    }

    WritableState.prototype.getBuffer = function writableStateGetBuffer() {
      var current = this.bufferedRequest;
      var out = [];
      while (current) {
        out.push(current);
        current = current.next;
      }
      return out;
    };
    function Writable(options) {

      // Writable ctor is applied to Duplexes, though they're not
      // instanceof Writable, they're instanceof Readable.
      if (!(this instanceof Writable) && !(this instanceof Duplex)) return new Writable(options);

      this._writableState = new WritableState(options, this);

      // legacy.
      this.writable = true;

      if (options) {
        if (typeof options.write === 'function') this._write = options.write;

        if (typeof options.writev === 'function') this._writev = options.writev;
      }

      events.EventEmitter.call(this);
    }

    // Otherwise people can pipe Writable streams, which is just wrong.
    Writable.prototype.pipe = function () {
      this.emit('error', new Error('Cannot pipe, not readable'));
    };

    function writeAfterEnd(stream, cb) {
      var er = new Error('write after end');
      // TODO: defer error events consistently everywhere, not just the cb
      stream.emit('error', er);
      nextTick(cb, er);
    }

    // If we get something that is not a buffer, string, null, or undefined,
    // and we're not in objectMode, then that's an error.
    // Otherwise stream chunks are all considered to be of length=1, and the
    // watermarks determine how many objects to keep in the buffer, rather than
    // how many bytes or characters.
    function validChunk(stream, state, chunk, cb) {
      var valid = true;
      var er = false;
      // Always throw error if a null is written
      // if we are not in object mode then throw
      // if it is not a buffer, string, or undefined.
      if (chunk === null) {
        er = new TypeError('May not write null values to stream');
      } else if (!Buffer$1.isBuffer(chunk) && typeof chunk !== 'string' && chunk !== undefined && !state.objectMode) {
        er = new TypeError('Invalid non-string/buffer chunk');
      }
      if (er) {
        stream.emit('error', er);
        nextTick(cb, er);
        valid = false;
      }
      return valid;
    }

    Writable.prototype.write = function (chunk, encoding, cb) {
      var state = this._writableState;
      var ret = false;

      if (typeof encoding === 'function') {
        cb = encoding;
        encoding = null;
      }

      if (Buffer$1.isBuffer(chunk)) encoding = 'buffer';else if (!encoding) encoding = state.defaultEncoding;

      if (typeof cb !== 'function') cb = nop;

      if (state.ended) writeAfterEnd(this, cb);else if (validChunk(this, state, chunk, cb)) {
        state.pendingcb++;
        ret = writeOrBuffer(this, state, chunk, encoding, cb);
      }

      return ret;
    };

    Writable.prototype.cork = function () {
      var state = this._writableState;

      state.corked++;
    };

    Writable.prototype.uncork = function () {
      var state = this._writableState;

      if (state.corked) {
        state.corked--;

        if (!state.writing && !state.corked && !state.finished && !state.bufferProcessing && state.bufferedRequest) clearBuffer(this, state);
      }
    };

    Writable.prototype.setDefaultEncoding = function setDefaultEncoding(encoding) {
      // node::ParseEncoding() requires lower case.
      if (typeof encoding === 'string') encoding = encoding.toLowerCase();
      if (!(['hex', 'utf8', 'utf-8', 'ascii', 'binary', 'base64', 'ucs2', 'ucs-2', 'utf16le', 'utf-16le', 'raw'].indexOf((encoding + '').toLowerCase()) > -1)) throw new TypeError('Unknown encoding: ' + encoding);
      this._writableState.defaultEncoding = encoding;
      return this;
    };

    function decodeChunk(state, chunk, encoding) {
      if (!state.objectMode && state.decodeStrings !== false && typeof chunk === 'string') {
        chunk = Buffer$1.from(chunk, encoding);
      }
      return chunk;
    }

    // if we're already writing something, then just put this
    // in the queue, and wait our turn.  Otherwise, call _write
    // If we return false, then we need a drain event, so set that flag.
    function writeOrBuffer(stream, state, chunk, encoding, cb) {
      chunk = decodeChunk(state, chunk, encoding);

      if (Buffer$1.isBuffer(chunk)) encoding = 'buffer';
      var len = state.objectMode ? 1 : chunk.length;

      state.length += len;

      var ret = state.length < state.highWaterMark;
      // we must ensure that previous needDrain will not be reset to false.
      if (!ret) state.needDrain = true;

      if (state.writing || state.corked) {
        var last = state.lastBufferedRequest;
        state.lastBufferedRequest = new WriteReq(chunk, encoding, cb);
        if (last) {
          last.next = state.lastBufferedRequest;
        } else {
          state.bufferedRequest = state.lastBufferedRequest;
        }
        state.bufferedRequestCount += 1;
      } else {
        doWrite(stream, state, false, len, chunk, encoding, cb);
      }

      return ret;
    }

    function doWrite(stream, state, writev, len, chunk, encoding, cb) {
      state.writelen = len;
      state.writecb = cb;
      state.writing = true;
      state.sync = true;
      if (writev) stream._writev(chunk, state.onwrite);else stream._write(chunk, encoding, state.onwrite);
      state.sync = false;
    }

    function onwriteError(stream, state, sync, er, cb) {
      --state.pendingcb;
      if (sync) nextTick(cb, er);else cb(er);

      stream._writableState.errorEmitted = true;
      stream.emit('error', er);
    }

    function onwriteStateUpdate(state) {
      state.writing = false;
      state.writecb = null;
      state.length -= state.writelen;
      state.writelen = 0;
    }

    function onwrite(stream, er) {
      var state = stream._writableState;
      var sync = state.sync;
      var cb = state.writecb;

      onwriteStateUpdate(state);

      if (er) onwriteError(stream, state, sync, er, cb);else {
        // Check if we're actually ready to finish, but don't emit yet
        var finished = needFinish(state);

        if (!finished && !state.corked && !state.bufferProcessing && state.bufferedRequest) {
          clearBuffer(stream, state);
        }

        if (sync) {
          /*<replacement>*/
            nextTick(afterWrite, stream, state, finished, cb);
          /*</replacement>*/
        } else {
            afterWrite(stream, state, finished, cb);
          }
      }
    }

    function afterWrite(stream, state, finished, cb) {
      if (!finished) onwriteDrain(stream, state);
      state.pendingcb--;
      cb();
      finishMaybe(stream, state);
    }

    // Must force callback to be called on nextTick, so that we don't
    // emit 'drain' before the write() consumer gets the 'false' return
    // value, and has a chance to attach a 'drain' listener.
    function onwriteDrain(stream, state) {
      if (state.length === 0 && state.needDrain) {
        state.needDrain = false;
        stream.emit('drain');
      }
    }

    // if there's something in the buffer waiting, then process it
    function clearBuffer(stream, state) {
      state.bufferProcessing = true;
      var entry = state.bufferedRequest;

      if (stream._writev && entry && entry.next) {
        // Fast case, write everything using _writev()
        var l = state.bufferedRequestCount;
        var buffer = new Array(l);
        var holder = state.corkedRequestsFree;
        holder.entry = entry;

        var count = 0;
        while (entry) {
          buffer[count] = entry;
          entry = entry.next;
          count += 1;
        }

        doWrite(stream, state, true, state.length, buffer, '', holder.finish);

        // doWrite is almost always async, defer these to save a bit of time
        // as the hot path ends with doWrite
        state.pendingcb++;
        state.lastBufferedRequest = null;
        if (holder.next) {
          state.corkedRequestsFree = holder.next;
          holder.next = null;
        } else {
          state.corkedRequestsFree = new CorkedRequest(state);
        }
      } else {
        // Slow case, write chunks one-by-one
        while (entry) {
          var chunk = entry.chunk;
          var encoding = entry.encoding;
          var cb = entry.callback;
          var len = state.objectMode ? 1 : chunk.length;

          doWrite(stream, state, false, len, chunk, encoding, cb);
          entry = entry.next;
          // if we didn't call the onwrite immediately, then
          // it means that we need to wait until it does.
          // also, that means that the chunk and cb are currently
          // being processed, so move the buffer counter past them.
          if (state.writing) {
            break;
          }
        }

        if (entry === null) state.lastBufferedRequest = null;
      }

      state.bufferedRequestCount = 0;
      state.bufferedRequest = entry;
      state.bufferProcessing = false;
    }

    Writable.prototype._write = function (chunk, encoding, cb) {
      cb(new Error('not implemented'));
    };

    Writable.prototype._writev = null;

    Writable.prototype.end = function (chunk, encoding, cb) {
      var state = this._writableState;

      if (typeof chunk === 'function') {
        cb = chunk;
        chunk = null;
        encoding = null;
      } else if (typeof encoding === 'function') {
        cb = encoding;
        encoding = null;
      }

      if (chunk !== null && chunk !== undefined) this.write(chunk, encoding);

      // .end() fully uncorks
      if (state.corked) {
        state.corked = 1;
        this.uncork();
      }

      // ignore unnecessary end() calls.
      if (!state.ending && !state.finished) endWritable(this, state, cb);
    };

    function needFinish(state) {
      return state.ending && state.length === 0 && state.bufferedRequest === null && !state.finished && !state.writing;
    }

    function prefinish(stream, state) {
      if (!state.prefinished) {
        state.prefinished = true;
        stream.emit('prefinish');
      }
    }

    function finishMaybe(stream, state) {
      var need = needFinish(state);
      if (need) {
        if (state.pendingcb === 0) {
          prefinish(stream, state);
          state.finished = true;
          stream.emit('finish');
        } else {
          prefinish(stream, state);
        }
      }
      return need;
    }

    function endWritable(stream, state, cb) {
      state.ending = true;
      finishMaybe(stream, state);
      if (cb) {
        if (state.finished) nextTick(cb);else stream.once('finish', cb);
      }
      state.ended = true;
      stream.writable = false;
    }

    // It seems a linked list but it is not
    // there will be only 2 of these for each stream
    function CorkedRequest(state) {
      var _this = this;

      this.next = null;
      this.entry = null;

      this.finish = function (err) {
        var entry = _this.entry;
        _this.entry = null;
        while (entry) {
          var cb = entry.callback;
          state.pendingcb--;
          cb(err);
          entry = entry.next;
        }
        if (state.corkedRequestsFree) {
          state.corkedRequestsFree.next = _this;
        } else {
          state.corkedRequestsFree = _this;
        }
      };
    }

    inherits$1(Duplex, Readable);

    var keys = Object.keys(Writable.prototype);
    for (var v = 0; v < keys.length; v++) {
      var method = keys[v];
      if (!Duplex.prototype[method]) Duplex.prototype[method] = Writable.prototype[method];
    }
    function Duplex(options) {
      if (!(this instanceof Duplex)) return new Duplex(options);

      Readable.call(this, options);
      Writable.call(this, options);

      if (options && options.readable === false) this.readable = false;

      if (options && options.writable === false) this.writable = false;

      this.allowHalfOpen = true;
      if (options && options.allowHalfOpen === false) this.allowHalfOpen = false;

      this.once('end', onend);
    }

    // the no-half-open enforcer
    function onend() {
      // if we allow half-open state, or if the writable side ended,
      // then we're ok.
      if (this.allowHalfOpen || this._writableState.ended) return;

      // no more data can be written.
      // But allow more writes to happen in this tick.
      nextTick(onEndNT, this);
    }

    function onEndNT(self) {
      self.end();
    }

    // a transform stream is a readable/writable stream where you do
    inherits$1(Transform, Duplex);

    function TransformState(stream) {
      this.afterTransform = function (er, data) {
        return afterTransform(stream, er, data);
      };

      this.needTransform = false;
      this.transforming = false;
      this.writecb = null;
      this.writechunk = null;
      this.writeencoding = null;
    }

    function afterTransform(stream, er, data) {
      var ts = stream._transformState;
      ts.transforming = false;

      var cb = ts.writecb;

      if (!cb) return stream.emit('error', new Error('no writecb in Transform class'));

      ts.writechunk = null;
      ts.writecb = null;

      if (data !== null && data !== undefined) stream.push(data);

      cb(er);

      var rs = stream._readableState;
      rs.reading = false;
      if (rs.needReadable || rs.length < rs.highWaterMark) {
        stream._read(rs.highWaterMark);
      }
    }
    function Transform(options) {
      if (!(this instanceof Transform)) return new Transform(options);

      Duplex.call(this, options);

      this._transformState = new TransformState(this);

      // when the writable side finishes, then flush out anything remaining.
      var stream = this;

      // start out asking for a readable event once data is transformed.
      this._readableState.needReadable = true;

      // we have implemented the _read method, and done the other things
      // that Readable wants before the first _read call, so unset the
      // sync guard flag.
      this._readableState.sync = false;

      if (options) {
        if (typeof options.transform === 'function') this._transform = options.transform;

        if (typeof options.flush === 'function') this._flush = options.flush;
      }

      this.once('prefinish', function () {
        if (typeof this._flush === 'function') this._flush(function (er) {
          done(stream, er);
        });else done(stream);
      });
    }

    Transform.prototype.push = function (chunk, encoding) {
      this._transformState.needTransform = false;
      return Duplex.prototype.push.call(this, chunk, encoding);
    };

    // This is the part where you do stuff!
    // override this function in implementation classes.
    // 'chunk' is an input chunk.
    //
    // Call `push(newChunk)` to pass along transformed output
    // to the readable side.  You may call 'push' zero or more times.
    //
    // Call `cb(err)` when you are done with this chunk.  If you pass
    // an error, then that'll put the hurt on the whole operation.  If you
    // never call cb(), then you'll never get another chunk.
    Transform.prototype._transform = function (chunk, encoding, cb) {
      throw new Error('Not implemented');
    };

    Transform.prototype._write = function (chunk, encoding, cb) {
      var ts = this._transformState;
      ts.writecb = cb;
      ts.writechunk = chunk;
      ts.writeencoding = encoding;
      if (!ts.transforming) {
        var rs = this._readableState;
        if (ts.needTransform || rs.needReadable || rs.length < rs.highWaterMark) this._read(rs.highWaterMark);
      }
    };

    // Doesn't matter what the args are here.
    // _transform does all the work.
    // That we got here means that the readable side wants more data.
    Transform.prototype._read = function (n) {
      var ts = this._transformState;

      if (ts.writechunk !== null && ts.writecb && !ts.transforming) {
        ts.transforming = true;
        this._transform(ts.writechunk, ts.writeencoding, ts.afterTransform);
      } else {
        // mark that we need a transform, so that any data that comes in
        // will get processed, now that we've asked for it.
        ts.needTransform = true;
      }
    };

    function done(stream, er) {
      if (er) return stream.emit('error', er);

      // if there's nothing in the write buffer, then that means
      // that nothing more will ever be provided
      var ws = stream._writableState;
      var ts = stream._transformState;

      if (ws.length) throw new Error('Calling transform done when ws.length != 0');

      if (ts.transforming) throw new Error('Calling transform done when still transforming');

      return stream.push(null);
    }

    inherits$1(PassThrough, Transform);
    function PassThrough(options) {
      if (!(this instanceof PassThrough)) return new PassThrough(options);

      Transform.call(this, options);
    }

    PassThrough.prototype._transform = function (chunk, encoding, cb) {
      cb(null, chunk);
    };

    inherits$1(Stream, events);
    Stream.Readable = Readable;
    Stream.Writable = Writable;
    Stream.Duplex = Duplex;
    Stream.Transform = Transform;
    Stream.PassThrough = PassThrough;

    // Backwards-compat with node 0.4.x
    Stream.Stream = Stream;

    // old-style streams.  Note that the pipe method (the only relevant
    // part of this class) is overridden in the Readable class.

    function Stream() {
      events.call(this);
    }

    Stream.prototype.pipe = function(dest, options) {
      var source = this;

      function ondata(chunk) {
        if (dest.writable) {
          if (false === dest.write(chunk) && source.pause) {
            source.pause();
          }
        }
      }

      source.on('data', ondata);

      function ondrain() {
        if (source.readable && source.resume) {
          source.resume();
        }
      }

      dest.on('drain', ondrain);

      // If the 'end' option is not supplied, dest.end() will be called when
      // source gets the 'end' or 'close' events.  Only dest.end() once.
      if (!dest._isStdio && (!options || options.end !== false)) {
        source.on('end', onend);
        source.on('close', onclose);
      }

      var didOnEnd = false;
      function onend() {
        if (didOnEnd) return;
        didOnEnd = true;

        dest.end();
      }


      function onclose() {
        if (didOnEnd) return;
        didOnEnd = true;

        if (typeof dest.destroy === 'function') dest.destroy();
      }

      // don't leave dangling pipes when there are errors.
      function onerror(er) {
        cleanup();
        if (events.listenerCount(this, 'error') === 0) {
          throw er; // Unhandled stream error in pipe.
        }
      }

      source.on('error', onerror);
      dest.on('error', onerror);

      // remove all the event listeners that were added.
      function cleanup() {
        source.removeListener('data', ondata);
        dest.removeListener('drain', ondrain);

        source.removeListener('end', onend);
        source.removeListener('close', onclose);

        source.removeListener('error', onerror);
        dest.removeListener('error', onerror);

        source.removeListener('end', cleanup);
        source.removeListener('close', cleanup);

        dest.removeListener('close', cleanup);
      }

      source.on('end', cleanup);
      source.on('close', cleanup);

      dest.on('close', cleanup);

      dest.emit('pipe', source);

      // Allow for unix-like usage: A.pipe(B).pipe(C)
      return dest;
    };

    var stream = /*#__PURE__*/Object.freeze({
        __proto__: null,
        'default': Stream,
        Readable: Readable,
        Writable: Writable,
        Duplex: Duplex,
        Transform: Transform,
        PassThrough: PassThrough,
        Stream: Stream
    });

    var rStates = {
      UNSENT: 0,
      OPENED: 1,
      HEADERS_RECEIVED: 2,
      LOADING: 3,
      DONE: 4
    };
    function IncomingMessage(xhr, response, mode) {
      var self = this;
      Readable.call(self);

      self._mode = mode;
      self.headers = {};
      self.rawHeaders = [];
      self.trailers = {};
      self.rawTrailers = [];

      // Fake the 'close' event, but only once 'end' fires
      self.on('end', function() {
        // The nextTick is necessary to prevent the 'request' module from causing an infinite loop
        nextTick(function() {
          self.emit('close');
        });
      });
      var read;
      if (mode === 'fetch') {
        self._fetchResponse = response;

        self.url = response.url;
        self.statusCode = response.status;
        self.statusMessage = response.statusText;
          // backwards compatible version of for (<item> of <iterable>):
          // for (var <item>,_i,_it = <iterable>[Symbol.iterator](); <item> = (_i = _it.next()).value,!_i.done;)
        for (var header, _i, _it = response.headers[Symbol.iterator](); header = (_i = _it.next()).value, !_i.done;) {
          self.headers[header[0].toLowerCase()] = header[1];
          self.rawHeaders.push(header[0], header[1]);
        }

        // TODO: this doesn't respect backpressure. Once WritableStream is available, this can be fixed
        var reader = response.body.getReader();

        read = function () {
          reader.read().then(function(result) {
            if (self._destroyed)
              return
            if (result.done) {
              self.push(null);
              return
            }
            self.push(new Buffer$1(result.value));
            read();
          });
        };
        read();

      } else {
        self._xhr = xhr;
        self._pos = 0;

        self.url = xhr.responseURL;
        self.statusCode = xhr.status;
        self.statusMessage = xhr.statusText;
        var headers = xhr.getAllResponseHeaders().split(/\r?\n/);
        headers.forEach(function(header) {
          var matches = header.match(/^([^:]+):\s*(.*)/);
          if (matches) {
            var key = matches[1].toLowerCase();
            if (key === 'set-cookie') {
              if (self.headers[key] === undefined) {
                self.headers[key] = [];
              }
              self.headers[key].push(matches[2]);
            } else if (self.headers[key] !== undefined) {
              self.headers[key] += ', ' + matches[2];
            } else {
              self.headers[key] = matches[2];
            }
            self.rawHeaders.push(matches[1], matches[2]);
          }
        });

        self._charset = 'x-user-defined';
        if (!overrideMimeType) {
          var mimeType = self.rawHeaders['mime-type'];
          if (mimeType) {
            var charsetMatch = mimeType.match(/;\s*charset=([^;])(;|$)/);
            if (charsetMatch) {
              self._charset = charsetMatch[1].toLowerCase();
            }
          }
          if (!self._charset)
            self._charset = 'utf-8'; // best guess
        }
      }
    }

    inherits$1(IncomingMessage, Readable);

    IncomingMessage.prototype._read = function() {};

    IncomingMessage.prototype._onXHRProgress = function() {
      var self = this;

      var xhr = self._xhr;

      var response = null;
      switch (self._mode) {
      case 'text:vbarray': // For IE9
        if (xhr.readyState !== rStates.DONE)
          break
        try {
          // This fails in IE8
          response = new global$1.VBArray(xhr.responseBody).toArray();
        } catch (e) {
          // pass
        }
        if (response !== null) {
          self.push(new Buffer$1(response));
          break
        }
        // Falls through in IE8
      case 'text':
        try { // This will fail when readyState = 3 in IE9. Switch mode and wait for readyState = 4
          response = xhr.responseText;
        } catch (e) {
          self._mode = 'text:vbarray';
          break
        }
        if (response.length > self._pos) {
          var newData = response.substr(self._pos);
          if (self._charset === 'x-user-defined') {
            var buffer = new Buffer$1(newData.length);
            for (var i = 0; i < newData.length; i++)
              buffer[i] = newData.charCodeAt(i) & 0xff;

            self.push(buffer);
          } else {
            self.push(newData, self._charset);
          }
          self._pos = response.length;
        }
        break
      case 'arraybuffer':
        if (xhr.readyState !== rStates.DONE || !xhr.response)
          break
        response = xhr.response;
        self.push(new Buffer$1(new Uint8Array(response)));
        break
      case 'moz-chunked-arraybuffer': // take whole
        response = xhr.response;
        if (xhr.readyState !== rStates.LOADING || !response)
          break
        self.push(new Buffer$1(new Uint8Array(response)));
        break
      case 'ms-stream':
        response = xhr.response;
        if (xhr.readyState !== rStates.LOADING)
          break
        var reader = new global$1.MSStreamReader();
        reader.onprogress = function() {
          if (reader.result.byteLength > self._pos) {
            self.push(new Buffer$1(new Uint8Array(reader.result.slice(self._pos))));
            self._pos = reader.result.byteLength;
          }
        };
        reader.onload = function() {
          self.push(null);
        };
          // reader.onerror = ??? // TODO: this
        reader.readAsArrayBuffer(response);
        break
      }

      // The ms-stream case handles end separately in reader.onload()
      if (self._xhr.readyState === rStates.DONE && self._mode !== 'ms-stream') {
        self.push(null);
      }
    };

    // from https://github.com/jhiesey/to-arraybuffer/blob/6502d9850e70ba7935a7df4ad86b358fc216f9f0/index.js
    function toArrayBuffer (buf) {
      // If the buffer is backed by a Uint8Array, a faster version will work
      if (buf instanceof Uint8Array) {
        // If the buffer isn't a subarray, return the underlying ArrayBuffer
        if (buf.byteOffset === 0 && buf.byteLength === buf.buffer.byteLength) {
          return buf.buffer
        } else if (typeof buf.buffer.slice === 'function') {
          // Otherwise we need to get a proper copy
          return buf.buffer.slice(buf.byteOffset, buf.byteOffset + buf.byteLength)
        }
      }

      if (isBuffer$2(buf)) {
        // This is the slow version that will work with any Buffer
        // implementation (even in old browsers)
        var arrayCopy = new Uint8Array(buf.length);
        var len = buf.length;
        for (var i = 0; i < len; i++) {
          arrayCopy[i] = buf[i];
        }
        return arrayCopy.buffer
      } else {
        throw new Error('Argument must be a Buffer')
      }
    }

    function decideMode(preferBinary, useFetch) {
      if (hasFetch && useFetch) {
        return 'fetch'
      } else if (mozchunkedarraybuffer) {
        return 'moz-chunked-arraybuffer'
      } else if (msstream) {
        return 'ms-stream'
      } else if (arraybuffer && preferBinary) {
        return 'arraybuffer'
      } else if (vbArray && preferBinary) {
        return 'text:vbarray'
      } else {
        return 'text'
      }
    }

    function ClientRequest(opts) {
      var self = this;
      Writable.call(self);

      self._opts = opts;
      self._body = [];
      self._headers = {};
      if (opts.auth)
        self.setHeader('Authorization', 'Basic ' + new Buffer$1(opts.auth).toString('base64'));
      Object.keys(opts.headers).forEach(function(name) {
        self.setHeader(name, opts.headers[name]);
      });

      var preferBinary;
      var useFetch = true;
      if (opts.mode === 'disable-fetch') {
        // If the use of XHR should be preferred and includes preserving the 'content-type' header
        useFetch = false;
        preferBinary = true;
      } else if (opts.mode === 'prefer-streaming') {
        // If streaming is a high priority but binary compatibility and
        // the accuracy of the 'content-type' header aren't
        preferBinary = false;
      } else if (opts.mode === 'allow-wrong-content-type') {
        // If streaming is more important than preserving the 'content-type' header
        preferBinary = !overrideMimeType;
      } else if (!opts.mode || opts.mode === 'default' || opts.mode === 'prefer-fast') {
        // Use binary if text streaming may corrupt data or the content-type header, or for speed
        preferBinary = true;
      } else {
        throw new Error('Invalid value for opts.mode')
      }
      self._mode = decideMode(preferBinary, useFetch);

      self.on('finish', function() {
        self._onFinish();
      });
    }

    inherits$1(ClientRequest, Writable);
    // Taken from http://www.w3.org/TR/XMLHttpRequest/#the-setrequestheader%28%29-method
    var unsafeHeaders = [
      'accept-charset',
      'accept-encoding',
      'access-control-request-headers',
      'access-control-request-method',
      'connection',
      'content-length',
      'cookie',
      'cookie2',
      'date',
      'dnt',
      'expect',
      'host',
      'keep-alive',
      'origin',
      'referer',
      'te',
      'trailer',
      'transfer-encoding',
      'upgrade',
      'user-agent',
      'via'
    ];
    ClientRequest.prototype.setHeader = function(name, value) {
      var self = this;
      var lowerName = name.toLowerCase();
        // This check is not necessary, but it prevents warnings from browsers about setting unsafe
        // headers. To be honest I'm not entirely sure hiding these warnings is a good thing, but
        // http-browserify did it, so I will too.
      if (unsafeHeaders.indexOf(lowerName) !== -1)
        return

      self._headers[lowerName] = {
        name: name,
        value: value
      };
    };

    ClientRequest.prototype.getHeader = function(name) {
      var self = this;
      return self._headers[name.toLowerCase()].value
    };

    ClientRequest.prototype.removeHeader = function(name) {
      var self = this;
      delete self._headers[name.toLowerCase()];
    };

    ClientRequest.prototype._onFinish = function() {
      var self = this;

      if (self._destroyed)
        return
      var opts = self._opts;

      var headersObj = self._headers;
      var body;
      if (opts.method === 'POST' || opts.method === 'PUT' || opts.method === 'PATCH') {
        if (blobConstructor()) {
          body = new global$1.Blob(self._body.map(function(buffer) {
            return toArrayBuffer(buffer)
          }), {
            type: (headersObj['content-type'] || {}).value || ''
          });
        } else {
          // get utf8 string
          body = Buffer$1.concat(self._body).toString();
        }
      }

      if (self._mode === 'fetch') {
        var headers = Object.keys(headersObj).map(function(name) {
          return [headersObj[name].name, headersObj[name].value]
        });

        global$1.fetch(self._opts.url, {
          method: self._opts.method,
          headers: headers,
          body: body,
          mode: 'cors',
          credentials: opts.withCredentials ? 'include' : 'same-origin'
        }).then(function(response) {
          self._fetchResponse = response;
          self._connect();
        }, function(reason) {
          self.emit('error', reason);
        });
      } else {
        var xhr = self._xhr = new global$1.XMLHttpRequest();
        try {
          xhr.open(self._opts.method, self._opts.url, true);
        } catch (err) {
          nextTick(function() {
            self.emit('error', err);
          });
          return
        }

        // Can't set responseType on really old browsers
        if ('responseType' in xhr)
          xhr.responseType = self._mode.split(':')[0];

        if ('withCredentials' in xhr)
          xhr.withCredentials = !!opts.withCredentials;

        if (self._mode === 'text' && 'overrideMimeType' in xhr)
          xhr.overrideMimeType('text/plain; charset=x-user-defined');

        Object.keys(headersObj).forEach(function(name) {
          xhr.setRequestHeader(headersObj[name].name, headersObj[name].value);
        });

        self._response = null;
        xhr.onreadystatechange = function() {
          switch (xhr.readyState) {
          case rStates.LOADING:
          case rStates.DONE:
            self._onXHRProgress();
            break
          }
        };
          // Necessary for streaming in Firefox, since xhr.response is ONLY defined
          // in onprogress, not in onreadystatechange with xhr.readyState = 3
        if (self._mode === 'moz-chunked-arraybuffer') {
          xhr.onprogress = function() {
            self._onXHRProgress();
          };
        }

        xhr.onerror = function() {
          if (self._destroyed)
            return
          self.emit('error', new Error('XHR error'));
        };

        try {
          xhr.send(body);
        } catch (err) {
          nextTick(function() {
            self.emit('error', err);
          });
          return
        }
      }
    };

    /**
     * Checks if xhr.status is readable and non-zero, indicating no error.
     * Even though the spec says it should be available in readyState 3,
     * accessing it throws an exception in IE8
     */
    function statusValid(xhr) {
      try {
        var status = xhr.status;
        return (status !== null && status !== 0)
      } catch (e) {
        return false
      }
    }

    ClientRequest.prototype._onXHRProgress = function() {
      var self = this;

      if (!statusValid(self._xhr) || self._destroyed)
        return

      if (!self._response)
        self._connect();

      self._response._onXHRProgress();
    };

    ClientRequest.prototype._connect = function() {
      var self = this;

      if (self._destroyed)
        return

      self._response = new IncomingMessage(self._xhr, self._fetchResponse, self._mode);
      self.emit('response', self._response);
    };

    ClientRequest.prototype._write = function(chunk, encoding, cb) {
      var self = this;

      self._body.push(chunk);
      cb();
    };

    ClientRequest.prototype.abort = ClientRequest.prototype.destroy = function() {
      var self = this;
      self._destroyed = true;
      if (self._response)
        self._response._destroyed = true;
      if (self._xhr)
        self._xhr.abort();
        // Currently, there isn't a way to truly abort a fetch.
        // If you like bikeshedding, see https://github.com/whatwg/fetch/issues/27
    };

    ClientRequest.prototype.end = function(data, encoding, cb) {
      var self = this;
      if (typeof data === 'function') {
        cb = data;
        data = undefined;
      }

      Writable.prototype.end.call(self, data, encoding, cb);
    };

    ClientRequest.prototype.flushHeaders = function() {};
    ClientRequest.prototype.setTimeout = function() {};
    ClientRequest.prototype.setNoDelay = function() {};
    ClientRequest.prototype.setSocketKeepAlive = function() {};

    /** Highest positive signed 32-bit float value */
    const maxInt = 2147483647; // aka. 0x7FFFFFFF or 2^31-1

    /** Bootstring parameters */
    const base = 36;
    const tMin = 1;
    const tMax = 26;
    const skew = 38;
    const damp = 700;
    const initialBias = 72;
    const initialN = 128; // 0x80
    const delimiter$1 = '-'; // '\x2D'
    const regexNonASCII = /[^\0-\x7E]/; // non-ASCII chars
    const regexSeparators = /[\x2E\u3002\uFF0E\uFF61]/g; // RFC 3490 separators

    /** Error messages */
    const errors = {
    	'overflow': 'Overflow: input needs wider integers to process',
    	'not-basic': 'Illegal input >= 0x80 (not a basic code point)',
    	'invalid-input': 'Invalid input'
    };

    /** Convenience shortcuts */
    const baseMinusTMin = base - tMin;
    const floor = Math.floor;
    const stringFromCharCode = String.fromCharCode;

    /*--------------------------------------------------------------------------*/

    /**
     * A generic error utility function.
     * @private
     * @param {String} type The error type.
     * @returns {Error} Throws a `RangeError` with the applicable error message.
     */
    function error(type) {
    	throw new RangeError(errors[type]);
    }

    /**
     * A generic `Array#map` utility function.
     * @private
     * @param {Array} array The array to iterate over.
     * @param {Function} callback The function that gets called for every array
     * item.
     * @returns {Array} A new array of values returned by the callback function.
     */
    function map$1(array, fn) {
    	const result = [];
    	let length = array.length;
    	while (length--) {
    		result[length] = fn(array[length]);
    	}
    	return result;
    }

    /**
     * A simple `Array#map`-like wrapper to work with domain name strings or email
     * addresses.
     * @private
     * @param {String} domain The domain name or email address.
     * @param {Function} callback The function that gets called for every
     * character.
     * @returns {Array} A new string of characters returned by the callback
     * function.
     */
    function mapDomain(string, fn) {
    	const parts = string.split('@');
    	let result = '';
    	if (parts.length > 1) {
    		// In email addresses, only the domain name should be punycoded. Leave
    		// the local part (i.e. everything up to `@`) intact.
    		result = parts[0] + '@';
    		string = parts[1];
    	}
    	// Avoid `split(regex)` for IE8 compatibility. See #17.
    	string = string.replace(regexSeparators, '\x2E');
    	const labels = string.split('.');
    	const encoded = map$1(labels, fn).join('.');
    	return result + encoded;
    }

    /**
     * Creates an array containing the numeric code points of each Unicode
     * character in the string. While JavaScript uses UCS-2 internally,
     * this function will convert a pair of surrogate halves (each of which
     * UCS-2 exposes as separate characters) into a single code point,
     * matching UTF-16.
     * @see `punycode.ucs2.encode`
     * @see <https://mathiasbynens.be/notes/javascript-encoding>
     * @memberOf punycode.ucs2
     * @name decode
     * @param {String} string The Unicode input string (UCS-2).
     * @returns {Array} The new array of code points.
     */
    function ucs2decode(string) {
    	const output = [];
    	let counter = 0;
    	const length = string.length;
    	while (counter < length) {
    		const value = string.charCodeAt(counter++);
    		if (value >= 0xD800 && value <= 0xDBFF && counter < length) {
    			// It's a high surrogate, and there is a next character.
    			const extra = string.charCodeAt(counter++);
    			if ((extra & 0xFC00) == 0xDC00) { // Low surrogate.
    				output.push(((value & 0x3FF) << 10) + (extra & 0x3FF) + 0x10000);
    			} else {
    				// It's an unmatched surrogate; only append this code unit, in case the
    				// next code unit is the high surrogate of a surrogate pair.
    				output.push(value);
    				counter--;
    			}
    		} else {
    			output.push(value);
    		}
    	}
    	return output;
    }

    /**
     * Converts a digit/integer into a basic code point.
     * @see `basicToDigit()`
     * @private
     * @param {Number} digit The numeric value of a basic code point.
     * @returns {Number} The basic code point whose value (when used for
     * representing integers) is `digit`, which needs to be in the range
     * `0` to `base - 1`. If `flag` is non-zero, the uppercase form is
     * used; else, the lowercase form is used. The behavior is undefined
     * if `flag` is non-zero and `digit` has no uppercase form.
     */
    const digitToBasic = function(digit, flag) {
    	//  0..25 map to ASCII a..z or A..Z
    	// 26..35 map to ASCII 0..9
    	return digit + 22 + 75 * (digit < 26) - ((flag != 0) << 5);
    };

    /**
     * Bias adaptation function as per section 3.4 of RFC 3492.
     * https://tools.ietf.org/html/rfc3492#section-3.4
     * @private
     */
    const adapt = function(delta, numPoints, firstTime) {
    	let k = 0;
    	delta = firstTime ? floor(delta / damp) : delta >> 1;
    	delta += floor(delta / numPoints);
    	for (/* no initialization */; delta > baseMinusTMin * tMax >> 1; k += base) {
    		delta = floor(delta / baseMinusTMin);
    	}
    	return floor(k + (baseMinusTMin + 1) * delta / (delta + skew));
    };

    /**
     * Converts a string of Unicode symbols (e.g. a domain name label) to a
     * Punycode string of ASCII-only symbols.
     * @memberOf punycode
     * @param {String} input The string of Unicode symbols.
     * @returns {String} The resulting Punycode string of ASCII-only symbols.
     */
    const encode = function(input) {
    	const output = [];

    	// Convert the input in UCS-2 to an array of Unicode code points.
    	input = ucs2decode(input);

    	// Cache the length.
    	let inputLength = input.length;

    	// Initialize the state.
    	let n = initialN;
    	let delta = 0;
    	let bias = initialBias;

    	// Handle the basic code points.
    	for (const currentValue of input) {
    		if (currentValue < 0x80) {
    			output.push(stringFromCharCode(currentValue));
    		}
    	}

    	let basicLength = output.length;
    	let handledCPCount = basicLength;

    	// `handledCPCount` is the number of code points that have been handled;
    	// `basicLength` is the number of basic code points.

    	// Finish the basic string with a delimiter unless it's empty.
    	if (basicLength) {
    		output.push(delimiter$1);
    	}

    	// Main encoding loop:
    	while (handledCPCount < inputLength) {

    		// All non-basic code points < n have been handled already. Find the next
    		// larger one:
    		let m = maxInt;
    		for (const currentValue of input) {
    			if (currentValue >= n && currentValue < m) {
    				m = currentValue;
    			}
    		}

    		// Increase `delta` enough to advance the decoder's <n,i> state to <m,0>,
    		// but guard against overflow.
    		const handledCPCountPlusOne = handledCPCount + 1;
    		if (m - n > floor((maxInt - delta) / handledCPCountPlusOne)) {
    			error('overflow');
    		}

    		delta += (m - n) * handledCPCountPlusOne;
    		n = m;

    		for (const currentValue of input) {
    			if (currentValue < n && ++delta > maxInt) {
    				error('overflow');
    			}
    			if (currentValue == n) {
    				// Represent delta as a generalized variable-length integer.
    				let q = delta;
    				for (let k = base; /* no condition */; k += base) {
    					const t = k <= bias ? tMin : (k >= bias + tMax ? tMax : k - bias);
    					if (q < t) {
    						break;
    					}
    					const qMinusT = q - t;
    					const baseMinusT = base - t;
    					output.push(
    						stringFromCharCode(digitToBasic(t + qMinusT % baseMinusT, 0))
    					);
    					q = floor(qMinusT / baseMinusT);
    				}

    				output.push(stringFromCharCode(digitToBasic(q, 0)));
    				bias = adapt(delta, handledCPCountPlusOne, handledCPCount == basicLength);
    				delta = 0;
    				++handledCPCount;
    			}
    		}

    		++delta;
    		++n;

    	}
    	return output.join('');
    };

    /**
     * Converts a Unicode string representing a domain name or an email address to
     * Punycode. Only the non-ASCII parts of the domain name will be converted,
     * i.e. it doesn't matter if you call it with a domain that's already in
     * ASCII.
     * @memberOf punycode
     * @param {String} input The domain name or email address to convert, as a
     * Unicode string.
     * @returns {String} The Punycode representation of the given domain name or
     * email address.
     */
    const toASCII = function(input) {
    	return mapDomain(input, function(string) {
    		return regexNonASCII.test(string)
    			? 'xn--' + encode(string)
    			: string;
    	});
    };

    // Copyright Joyent, Inc. and other Node contributors.
    //
    // Permission is hereby granted, free of charge, to any person obtaining a
    // copy of this software and associated documentation files (the
    // "Software"), to deal in the Software without restriction, including
    // without limitation the rights to use, copy, modify, merge, publish,
    // distribute, sublicense, and/or sell copies of the Software, and to permit
    // persons to whom the Software is furnished to do so, subject to the
    // following conditions:
    //
    // The above copyright notice and this permission notice shall be included
    // in all copies or substantial portions of the Software.
    //
    // THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    // OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    // MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    // NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    // DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    // USE OR OTHER DEALINGS IN THE SOFTWARE.


    // If obj.hasOwnProperty has been overridden, then calling
    // obj.hasOwnProperty(prop) will break.
    // See: https://github.com/joyent/node/issues/1707
    function hasOwnProperty(obj, prop) {
      return Object.prototype.hasOwnProperty.call(obj, prop);
    }
    var isArray = Array.isArray || function (xs) {
      return Object.prototype.toString.call(xs) === '[object Array]';
    };
    function stringifyPrimitive(v) {
      switch (typeof v) {
        case 'string':
          return v;

        case 'boolean':
          return v ? 'true' : 'false';

        case 'number':
          return isFinite(v) ? v : '';

        default:
          return '';
      }
    }

    function stringify (obj, sep, eq, name) {
      sep = sep || '&';
      eq = eq || '=';
      if (obj === null) {
        obj = undefined;
      }

      if (typeof obj === 'object') {
        return map(objectKeys$1(obj), function(k) {
          var ks = encodeURIComponent(stringifyPrimitive(k)) + eq;
          if (isArray(obj[k])) {
            return map(obj[k], function(v) {
              return ks + encodeURIComponent(stringifyPrimitive(v));
            }).join(sep);
          } else {
            return ks + encodeURIComponent(stringifyPrimitive(obj[k]));
          }
        }).join(sep);

      }

      if (!name) return '';
      return encodeURIComponent(stringifyPrimitive(name)) + eq +
             encodeURIComponent(stringifyPrimitive(obj));
    }
    function map (xs, f) {
      if (xs.map) return xs.map(f);
      var res = [];
      for (var i = 0; i < xs.length; i++) {
        res.push(f(xs[i], i));
      }
      return res;
    }

    var objectKeys$1 = Object.keys || function (obj) {
      var res = [];
      for (var key in obj) {
        if (Object.prototype.hasOwnProperty.call(obj, key)) res.push(key);
      }
      return res;
    };

    function parse$1(qs, sep, eq, options) {
      sep = sep || '&';
      eq = eq || '=';
      var obj = {};

      if (typeof qs !== 'string' || qs.length === 0) {
        return obj;
      }

      var regexp = /\+/g;
      qs = qs.split(sep);

      var maxKeys = 1000;
      if (options && typeof options.maxKeys === 'number') {
        maxKeys = options.maxKeys;
      }

      var len = qs.length;
      // maxKeys <= 0 means that we should not limit keys count
      if (maxKeys > 0 && len > maxKeys) {
        len = maxKeys;
      }

      for (var i = 0; i < len; ++i) {
        var x = qs[i].replace(regexp, '%20'),
            idx = x.indexOf(eq),
            kstr, vstr, k, v;

        if (idx >= 0) {
          kstr = x.substr(0, idx);
          vstr = x.substr(idx + 1);
        } else {
          kstr = x;
          vstr = '';
        }

        k = decodeURIComponent(kstr);
        v = decodeURIComponent(vstr);

        if (!hasOwnProperty(obj, k)) {
          obj[k] = v;
        } else if (isArray(obj[k])) {
          obj[k].push(v);
        } else {
          obj[k] = [obj[k], v];
        }
      }

      return obj;
    }

    // Copyright Joyent, Inc. and other Node contributors.
    var url = {
      parse: urlParse,
      resolve: urlResolve,
      resolveObject: urlResolveObject,
      format: urlFormat,
      Url: Url
    };
    function Url() {
      this.protocol = null;
      this.slashes = null;
      this.auth = null;
      this.host = null;
      this.port = null;
      this.hostname = null;
      this.hash = null;
      this.search = null;
      this.query = null;
      this.pathname = null;
      this.path = null;
      this.href = null;
    }

    // Reference: RFC 3986, RFC 1808, RFC 2396

    // define these here so at least they only have to be
    // compiled once on the first module load.
    var protocolPattern = /^([a-z0-9.+-]+:)/i,
      portPattern = /:[0-9]*$/,

      // Special case for a simple path URL
      simplePathPattern = /^(\/\/?(?!\/)[^\?\s]*)(\?[^\s]*)?$/,

      // RFC 2396: characters reserved for delimiting URLs.
      // We actually just auto-escape these.
      delims = ['<', '>', '"', '`', ' ', '\r', '\n', '\t'],

      // RFC 2396: characters not allowed for various reasons.
      unwise = ['{', '}', '|', '\\', '^', '`'].concat(delims),

      // Allowed by RFCs, but cause of XSS attacks.  Always escape these.
      autoEscape = ['\''].concat(unwise),
      // Characters that are never ever allowed in a hostname.
      // Note that any invalid chars are also handled, but these
      // are the ones that are *expected* to be seen, so we fast-path
      // them.
      nonHostChars = ['%', '/', '?', ';', '#'].concat(autoEscape),
      hostEndingChars = ['/', '?', '#'],
      hostnameMaxLen = 255,
      hostnamePartPattern = /^[+a-z0-9A-Z_-]{0,63}$/,
      hostnamePartStart = /^([+a-z0-9A-Z_-]{0,63})(.*)$/,
      // protocols that can allow "unsafe" and "unwise" chars.
      unsafeProtocol = {
        'javascript': true,
        'javascript:': true
      },
      // protocols that never have a hostname.
      hostlessProtocol = {
        'javascript': true,
        'javascript:': true
      },
      // protocols that always contain a // bit.
      slashedProtocol = {
        'http': true,
        'https': true,
        'ftp': true,
        'gopher': true,
        'file': true,
        'http:': true,
        'https:': true,
        'ftp:': true,
        'gopher:': true,
        'file:': true
      };

    function urlParse(url, parseQueryString, slashesDenoteHost) {
      if (url && isObject$1(url) && url instanceof Url) return url;

      var u = new Url;
      u.parse(url, parseQueryString, slashesDenoteHost);
      return u;
    }
    Url.prototype.parse = function(url, parseQueryString, slashesDenoteHost) {
      return parse(this, url, parseQueryString, slashesDenoteHost);
    };

    function parse(self, url, parseQueryString, slashesDenoteHost) {
      if (!isString(url)) {
        throw new TypeError('Parameter \'url\' must be a string, not ' + typeof url);
      }

      // Copy chrome, IE, opera backslash-handling behavior.
      // Back slashes before the query string get converted to forward slashes
      // See: https://code.google.com/p/chromium/issues/detail?id=25916
      var queryIndex = url.indexOf('?'),
        splitter =
        (queryIndex !== -1 && queryIndex < url.indexOf('#')) ? '?' : '#',
        uSplit = url.split(splitter),
        slashRegex = /\\/g;
      uSplit[0] = uSplit[0].replace(slashRegex, '/');
      url = uSplit.join(splitter);

      var rest = url;

      // trim before proceeding.
      // This is to support parse stuff like "  http://foo.com  \n"
      rest = rest.trim();

      if (!slashesDenoteHost && url.split('#').length === 1) {
        // Try fast path regexp
        var simplePath = simplePathPattern.exec(rest);
        if (simplePath) {
          self.path = rest;
          self.href = rest;
          self.pathname = simplePath[1];
          if (simplePath[2]) {
            self.search = simplePath[2];
            if (parseQueryString) {
              self.query = parse$1(self.search.substr(1));
            } else {
              self.query = self.search.substr(1);
            }
          } else if (parseQueryString) {
            self.search = '';
            self.query = {};
          }
          return self;
        }
      }

      var proto = protocolPattern.exec(rest);
      if (proto) {
        proto = proto[0];
        var lowerProto = proto.toLowerCase();
        self.protocol = lowerProto;
        rest = rest.substr(proto.length);
      }

      // figure out if it's got a host
      // user@server is *always* interpreted as a hostname, and url
      // resolution will treat //foo/bar as host=foo,path=bar because that's
      // how the browser resolves relative URLs.
      if (slashesDenoteHost || proto || rest.match(/^\/\/[^@\/]+@[^@\/]+/)) {
        var slashes = rest.substr(0, 2) === '//';
        if (slashes && !(proto && hostlessProtocol[proto])) {
          rest = rest.substr(2);
          self.slashes = true;
        }
      }
      var i, hec, l, p;
      if (!hostlessProtocol[proto] &&
        (slashes || (proto && !slashedProtocol[proto]))) {

        // there's a hostname.
        // the first instance of /, ?, ;, or # ends the host.
        //
        // If there is an @ in the hostname, then non-host chars *are* allowed
        // to the left of the last @ sign, unless some host-ending character
        // comes *before* the @-sign.
        // URLs are obnoxious.
        //
        // ex:
        // http://a@b@c/ => user:a@b host:c
        // http://a@b?@c => user:a host:c path:/?@c

        // v0.12 TODO(isaacs): This is not quite how Chrome does things.
        // Review our test case against browsers more comprehensively.

        // find the first instance of any hostEndingChars
        var hostEnd = -1;
        for (i = 0; i < hostEndingChars.length; i++) {
          hec = rest.indexOf(hostEndingChars[i]);
          if (hec !== -1 && (hostEnd === -1 || hec < hostEnd))
            hostEnd = hec;
        }

        // at this point, either we have an explicit point where the
        // auth portion cannot go past, or the last @ char is the decider.
        var auth, atSign;
        if (hostEnd === -1) {
          // atSign can be anywhere.
          atSign = rest.lastIndexOf('@');
        } else {
          // atSign must be in auth portion.
          // http://a@b/c@d => host:b auth:a path:/c@d
          atSign = rest.lastIndexOf('@', hostEnd);
        }

        // Now we have a portion which is definitely the auth.
        // Pull that off.
        if (atSign !== -1) {
          auth = rest.slice(0, atSign);
          rest = rest.slice(atSign + 1);
          self.auth = decodeURIComponent(auth);
        }

        // the host is the remaining to the left of the first non-host char
        hostEnd = -1;
        for (i = 0; i < nonHostChars.length; i++) {
          hec = rest.indexOf(nonHostChars[i]);
          if (hec !== -1 && (hostEnd === -1 || hec < hostEnd))
            hostEnd = hec;
        }
        // if we still have not hit it, then the entire thing is a host.
        if (hostEnd === -1)
          hostEnd = rest.length;

        self.host = rest.slice(0, hostEnd);
        rest = rest.slice(hostEnd);

        // pull out port.
        parseHost(self);

        // we've indicated that there is a hostname,
        // so even if it's empty, it has to be present.
        self.hostname = self.hostname || '';

        // if hostname begins with [ and ends with ]
        // assume that it's an IPv6 address.
        var ipv6Hostname = self.hostname[0] === '[' &&
          self.hostname[self.hostname.length - 1] === ']';

        // validate a little.
        if (!ipv6Hostname) {
          var hostparts = self.hostname.split(/\./);
          for (i = 0, l = hostparts.length; i < l; i++) {
            var part = hostparts[i];
            if (!part) continue;
            if (!part.match(hostnamePartPattern)) {
              var newpart = '';
              for (var j = 0, k = part.length; j < k; j++) {
                if (part.charCodeAt(j) > 127) {
                  // we replace non-ASCII char with a temporary placeholder
                  // we need this to make sure size of hostname is not
                  // broken by replacing non-ASCII by nothing
                  newpart += 'x';
                } else {
                  newpart += part[j];
                }
              }
              // we test again with ASCII char only
              if (!newpart.match(hostnamePartPattern)) {
                var validParts = hostparts.slice(0, i);
                var notHost = hostparts.slice(i + 1);
                var bit = part.match(hostnamePartStart);
                if (bit) {
                  validParts.push(bit[1]);
                  notHost.unshift(bit[2]);
                }
                if (notHost.length) {
                  rest = '/' + notHost.join('.') + rest;
                }
                self.hostname = validParts.join('.');
                break;
              }
            }
          }
        }

        if (self.hostname.length > hostnameMaxLen) {
          self.hostname = '';
        } else {
          // hostnames are always lower case.
          self.hostname = self.hostname.toLowerCase();
        }

        if (!ipv6Hostname) {
          // IDNA Support: Returns a punycoded representation of "domain".
          // It only converts parts of the domain name that
          // have non-ASCII characters, i.e. it doesn't matter if
          // you call it with a domain that already is ASCII-only.
          self.hostname = toASCII(self.hostname);
        }

        p = self.port ? ':' + self.port : '';
        var h = self.hostname || '';
        self.host = h + p;
        self.href += self.host;

        // strip [ and ] from the hostname
        // the host field still retains them, though
        if (ipv6Hostname) {
          self.hostname = self.hostname.substr(1, self.hostname.length - 2);
          if (rest[0] !== '/') {
            rest = '/' + rest;
          }
        }
      }

      // now rest is set to the post-host stuff.
      // chop off any delim chars.
      if (!unsafeProtocol[lowerProto]) {

        // First, make 100% sure that any "autoEscape" chars get
        // escaped, even if encodeURIComponent doesn't think they
        // need to be.
        for (i = 0, l = autoEscape.length; i < l; i++) {
          var ae = autoEscape[i];
          if (rest.indexOf(ae) === -1)
            continue;
          var esc = encodeURIComponent(ae);
          if (esc === ae) {
            esc = escape(ae);
          }
          rest = rest.split(ae).join(esc);
        }
      }


      // chop off from the tail first.
      var hash = rest.indexOf('#');
      if (hash !== -1) {
        // got a fragment string.
        self.hash = rest.substr(hash);
        rest = rest.slice(0, hash);
      }
      var qm = rest.indexOf('?');
      if (qm !== -1) {
        self.search = rest.substr(qm);
        self.query = rest.substr(qm + 1);
        if (parseQueryString) {
          self.query = parse$1(self.query);
        }
        rest = rest.slice(0, qm);
      } else if (parseQueryString) {
        // no query string, but parseQueryString still requested
        self.search = '';
        self.query = {};
      }
      if (rest) self.pathname = rest;
      if (slashedProtocol[lowerProto] &&
        self.hostname && !self.pathname) {
        self.pathname = '/';
      }

      //to support http.request
      if (self.pathname || self.search) {
        p = self.pathname || '';
        var s = self.search || '';
        self.path = p + s;
      }

      // finally, reconstruct the href based on what has been validated.
      self.href = format(self);
      return self;
    }

    // format a parsed object into a url string
    function urlFormat(obj) {
      // ensure it's an object, and not a string url.
      // If it's an obj, this is a no-op.
      // this way, you can call url_format() on strings
      // to clean up potentially wonky urls.
      if (isString(obj)) obj = parse({}, obj);
      return format(obj);
    }

    function format(self) {
      var auth = self.auth || '';
      if (auth) {
        auth = encodeURIComponent(auth);
        auth = auth.replace(/%3A/i, ':');
        auth += '@';
      }

      var protocol = self.protocol || '',
        pathname = self.pathname || '',
        hash = self.hash || '',
        host = false,
        query = '';

      if (self.host) {
        host = auth + self.host;
      } else if (self.hostname) {
        host = auth + (self.hostname.indexOf(':') === -1 ?
          self.hostname :
          '[' + this.hostname + ']');
        if (self.port) {
          host += ':' + self.port;
        }
      }

      if (self.query &&
        isObject$1(self.query) &&
        Object.keys(self.query).length) {
        query = stringify(self.query);
      }

      var search = self.search || (query && ('?' + query)) || '';

      if (protocol && protocol.substr(-1) !== ':') protocol += ':';

      // only the slashedProtocols get the //.  Not mailto:, xmpp:, etc.
      // unless they had them to begin with.
      if (self.slashes ||
        (!protocol || slashedProtocol[protocol]) && host !== false) {
        host = '//' + (host || '');
        if (pathname && pathname.charAt(0) !== '/') pathname = '/' + pathname;
      } else if (!host) {
        host = '';
      }

      if (hash && hash.charAt(0) !== '#') hash = '#' + hash;
      if (search && search.charAt(0) !== '?') search = '?' + search;

      pathname = pathname.replace(/[?#]/g, function(match) {
        return encodeURIComponent(match);
      });
      search = search.replace('#', '%23');

      return protocol + host + pathname + search + hash;
    }

    Url.prototype.format = function() {
      return format(this);
    };

    function urlResolve(source, relative) {
      return urlParse(source, false, true).resolve(relative);
    }

    Url.prototype.resolve = function(relative) {
      return this.resolveObject(urlParse(relative, false, true)).format();
    };

    function urlResolveObject(source, relative) {
      if (!source) return relative;
      return urlParse(source, false, true).resolveObject(relative);
    }

    Url.prototype.resolveObject = function(relative) {
      if (isString(relative)) {
        var rel = new Url();
        rel.parse(relative, false, true);
        relative = rel;
      }

      var result = new Url();
      var tkeys = Object.keys(this);
      for (var tk = 0; tk < tkeys.length; tk++) {
        var tkey = tkeys[tk];
        result[tkey] = this[tkey];
      }

      // hash is always overridden, no matter what.
      // even href="" will remove it.
      result.hash = relative.hash;

      // if the relative url is empty, then there's nothing left to do here.
      if (relative.href === '') {
        result.href = result.format();
        return result;
      }

      // hrefs like //foo/bar always cut to the protocol.
      if (relative.slashes && !relative.protocol) {
        // take everything except the protocol from relative
        var rkeys = Object.keys(relative);
        for (var rk = 0; rk < rkeys.length; rk++) {
          var rkey = rkeys[rk];
          if (rkey !== 'protocol')
            result[rkey] = relative[rkey];
        }

        //urlParse appends trailing / to urls like http://www.example.com
        if (slashedProtocol[result.protocol] &&
          result.hostname && !result.pathname) {
          result.path = result.pathname = '/';
        }

        result.href = result.format();
        return result;
      }
      var relPath;
      if (relative.protocol && relative.protocol !== result.protocol) {
        // if it's a known url protocol, then changing
        // the protocol does weird things
        // first, if it's not file:, then we MUST have a host,
        // and if there was a path
        // to begin with, then we MUST have a path.
        // if it is file:, then the host is dropped,
        // because that's known to be hostless.
        // anything else is assumed to be absolute.
        if (!slashedProtocol[relative.protocol]) {
          var keys = Object.keys(relative);
          for (var v = 0; v < keys.length; v++) {
            var k = keys[v];
            result[k] = relative[k];
          }
          result.href = result.format();
          return result;
        }

        result.protocol = relative.protocol;
        if (!relative.host && !hostlessProtocol[relative.protocol]) {
          relPath = (relative.pathname || '').split('/');
          while (relPath.length && !(relative.host = relPath.shift()));
          if (!relative.host) relative.host = '';
          if (!relative.hostname) relative.hostname = '';
          if (relPath[0] !== '') relPath.unshift('');
          if (relPath.length < 2) relPath.unshift('');
          result.pathname = relPath.join('/');
        } else {
          result.pathname = relative.pathname;
        }
        result.search = relative.search;
        result.query = relative.query;
        result.host = relative.host || '';
        result.auth = relative.auth;
        result.hostname = relative.hostname || relative.host;
        result.port = relative.port;
        // to support http.request
        if (result.pathname || result.search) {
          var p = result.pathname || '';
          var s = result.search || '';
          result.path = p + s;
        }
        result.slashes = result.slashes || relative.slashes;
        result.href = result.format();
        return result;
      }

      var isSourceAbs = (result.pathname && result.pathname.charAt(0) === '/'),
        isRelAbs = (
          relative.host ||
          relative.pathname && relative.pathname.charAt(0) === '/'
        ),
        mustEndAbs = (isRelAbs || isSourceAbs ||
          (result.host && relative.pathname)),
        removeAllDots = mustEndAbs,
        srcPath = result.pathname && result.pathname.split('/') || [],
        psychotic = result.protocol && !slashedProtocol[result.protocol];
      relPath = relative.pathname && relative.pathname.split('/') || [];
      // if the url is a non-slashed url, then relative
      // links like ../.. should be able
      // to crawl up to the hostname, as well.  This is strange.
      // result.protocol has already been set by now.
      // Later on, put the first path part into the host field.
      if (psychotic) {
        result.hostname = '';
        result.port = null;
        if (result.host) {
          if (srcPath[0] === '') srcPath[0] = result.host;
          else srcPath.unshift(result.host);
        }
        result.host = '';
        if (relative.protocol) {
          relative.hostname = null;
          relative.port = null;
          if (relative.host) {
            if (relPath[0] === '') relPath[0] = relative.host;
            else relPath.unshift(relative.host);
          }
          relative.host = null;
        }
        mustEndAbs = mustEndAbs && (relPath[0] === '' || srcPath[0] === '');
      }
      var authInHost;
      if (isRelAbs) {
        // it's absolute.
        result.host = (relative.host || relative.host === '') ?
          relative.host : result.host;
        result.hostname = (relative.hostname || relative.hostname === '') ?
          relative.hostname : result.hostname;
        result.search = relative.search;
        result.query = relative.query;
        srcPath = relPath;
        // fall through to the dot-handling below.
      } else if (relPath.length) {
        // it's relative
        // throw away the existing file, and take the new path instead.
        if (!srcPath) srcPath = [];
        srcPath.pop();
        srcPath = srcPath.concat(relPath);
        result.search = relative.search;
        result.query = relative.query;
      } else if (!isNullOrUndefined(relative.search)) {
        // just pull out the search.
        // like href='?foo'.
        // Put this after the other two cases because it simplifies the booleans
        if (psychotic) {
          result.hostname = result.host = srcPath.shift();
          //occationaly the auth can get stuck only in host
          //this especially happens in cases like
          //url.resolveObject('mailto:local1@domain1', 'local2@domain2')
          authInHost = result.host && result.host.indexOf('@') > 0 ?
            result.host.split('@') : false;
          if (authInHost) {
            result.auth = authInHost.shift();
            result.host = result.hostname = authInHost.shift();
          }
        }
        result.search = relative.search;
        result.query = relative.query;
        //to support http.request
        if (!isNull(result.pathname) || !isNull(result.search)) {
          result.path = (result.pathname ? result.pathname : '') +
            (result.search ? result.search : '');
        }
        result.href = result.format();
        return result;
      }

      if (!srcPath.length) {
        // no path at all.  easy.
        // we've already handled the other stuff above.
        result.pathname = null;
        //to support http.request
        if (result.search) {
          result.path = '/' + result.search;
        } else {
          result.path = null;
        }
        result.href = result.format();
        return result;
      }

      // if a url ENDs in . or .., then it must get a trailing slash.
      // however, if it ends in anything else non-slashy,
      // then it must NOT get a trailing slash.
      var last = srcPath.slice(-1)[0];
      var hasTrailingSlash = (
        (result.host || relative.host || srcPath.length > 1) &&
        (last === '.' || last === '..') || last === '');

      // strip single dots, resolve double dots to parent dir
      // if the path tries to go above the root, `up` ends up > 0
      var up = 0;
      for (var i = srcPath.length; i >= 0; i--) {
        last = srcPath[i];
        if (last === '.') {
          srcPath.splice(i, 1);
        } else if (last === '..') {
          srcPath.splice(i, 1);
          up++;
        } else if (up) {
          srcPath.splice(i, 1);
          up--;
        }
      }

      // if the path is allowed to go above the root, restore leading ..s
      if (!mustEndAbs && !removeAllDots) {
        for (; up--; up) {
          srcPath.unshift('..');
        }
      }

      if (mustEndAbs && srcPath[0] !== '' &&
        (!srcPath[0] || srcPath[0].charAt(0) !== '/')) {
        srcPath.unshift('');
      }

      if (hasTrailingSlash && (srcPath.join('/').substr(-1) !== '/')) {
        srcPath.push('');
      }

      var isAbsolute = srcPath[0] === '' ||
        (srcPath[0] && srcPath[0].charAt(0) === '/');

      // put the host back
      if (psychotic) {
        result.hostname = result.host = isAbsolute ? '' :
          srcPath.length ? srcPath.shift() : '';
        //occationaly the auth can get stuck only in host
        //this especially happens in cases like
        //url.resolveObject('mailto:local1@domain1', 'local2@domain2')
        authInHost = result.host && result.host.indexOf('@') > 0 ?
          result.host.split('@') : false;
        if (authInHost) {
          result.auth = authInHost.shift();
          result.host = result.hostname = authInHost.shift();
        }
      }

      mustEndAbs = mustEndAbs || (result.host && srcPath.length);

      if (mustEndAbs && !isAbsolute) {
        srcPath.unshift('');
      }

      if (!srcPath.length) {
        result.pathname = null;
        result.path = null;
      } else {
        result.pathname = srcPath.join('/');
      }

      //to support request.http
      if (!isNull(result.pathname) || !isNull(result.search)) {
        result.path = (result.pathname ? result.pathname : '') +
          (result.search ? result.search : '');
      }
      result.auth = relative.auth || result.auth;
      result.slashes = result.slashes || relative.slashes;
      result.href = result.format();
      return result;
    };

    Url.prototype.parseHost = function() {
      return parseHost(this);
    };

    function parseHost(self) {
      var host = self.host;
      var port = portPattern.exec(host);
      if (port) {
        port = port[0];
        if (port !== ':') {
          self.port = port.substr(1);
        }
        host = host.substr(0, host.length - port.length);
      }
      if (host) self.hostname = host;
    }

    var url$1 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        parse: urlParse,
        resolve: urlResolve,
        resolveObject: urlResolveObject,
        format: urlFormat,
        'default': url,
        Url: Url
    });

    function request(opts, cb) {
      if (typeof opts === 'string')
        opts = urlParse(opts);


      // Normally, the page is loaded from http or https, so not specifying a protocol
      // will result in a (valid) protocol-relative url. However, this won't work if
      // the protocol is something else, like 'file:'
      var defaultProtocol = global$1.location.protocol.search(/^https?:$/) === -1 ? 'http:' : '';

      var protocol = opts.protocol || defaultProtocol;
      var host = opts.hostname || opts.host;
      var port = opts.port;
      var path = opts.path || '/';

      // Necessary for IPv6 addresses
      if (host && host.indexOf(':') !== -1)
        host = '[' + host + ']';

      // This may be a relative url. The browser should always be able to interpret it correctly.
      opts.url = (host ? (protocol + '//' + host) : '') + (port ? ':' + port : '') + path;
      opts.method = (opts.method || 'GET').toUpperCase();
      opts.headers = opts.headers || {};

      // Also valid opts.auth, opts.mode

      var req = new ClientRequest(opts);
      if (cb)
        req.on('response', cb);
      return req
    }

    function get(opts, cb) {
      var req = request(opts, cb);
      req.end();
      return req
    }

    function Agent() {}
    Agent.defaultMaxSockets = 4;

    var METHODS = [
      'CHECKOUT',
      'CONNECT',
      'COPY',
      'DELETE',
      'GET',
      'HEAD',
      'LOCK',
      'M-SEARCH',
      'MERGE',
      'MKACTIVITY',
      'MKCOL',
      'MOVE',
      'NOTIFY',
      'OPTIONS',
      'PATCH',
      'POST',
      'PROPFIND',
      'PROPPATCH',
      'PURGE',
      'PUT',
      'REPORT',
      'SEARCH',
      'SUBSCRIBE',
      'TRACE',
      'UNLOCK',
      'UNSUBSCRIBE'
    ];
    var STATUS_CODES = {
      100: 'Continue',
      101: 'Switching Protocols',
      102: 'Processing', // RFC 2518, obsoleted by RFC 4918
      200: 'OK',
      201: 'Created',
      202: 'Accepted',
      203: 'Non-Authoritative Information',
      204: 'No Content',
      205: 'Reset Content',
      206: 'Partial Content',
      207: 'Multi-Status', // RFC 4918
      300: 'Multiple Choices',
      301: 'Moved Permanently',
      302: 'Moved Temporarily',
      303: 'See Other',
      304: 'Not Modified',
      305: 'Use Proxy',
      307: 'Temporary Redirect',
      400: 'Bad Request',
      401: 'Unauthorized',
      402: 'Payment Required',
      403: 'Forbidden',
      404: 'Not Found',
      405: 'Method Not Allowed',
      406: 'Not Acceptable',
      407: 'Proxy Authentication Required',
      408: 'Request Time-out',
      409: 'Conflict',
      410: 'Gone',
      411: 'Length Required',
      412: 'Precondition Failed',
      413: 'Request Entity Too Large',
      414: 'Request-URI Too Large',
      415: 'Unsupported Media Type',
      416: 'Requested Range Not Satisfiable',
      417: 'Expectation Failed',
      418: 'I\'m a teapot', // RFC 2324
      422: 'Unprocessable Entity', // RFC 4918
      423: 'Locked', // RFC 4918
      424: 'Failed Dependency', // RFC 4918
      425: 'Unordered Collection', // RFC 4918
      426: 'Upgrade Required', // RFC 2817
      428: 'Precondition Required', // RFC 6585
      429: 'Too Many Requests', // RFC 6585
      431: 'Request Header Fields Too Large', // RFC 6585
      500: 'Internal Server Error',
      501: 'Not Implemented',
      502: 'Bad Gateway',
      503: 'Service Unavailable',
      504: 'Gateway Time-out',
      505: 'HTTP Version Not Supported',
      506: 'Variant Also Negotiates', // RFC 2295
      507: 'Insufficient Storage', // RFC 4918
      509: 'Bandwidth Limit Exceeded',
      510: 'Not Extended', // RFC 2774
      511: 'Network Authentication Required' // RFC 6585
    };

    var http = {
      request,
      get,
      Agent,
      METHODS,
      STATUS_CODES
    };

    var http$1 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        request: request,
        get: get,
        Agent: Agent,
        METHODS: METHODS,
        STATUS_CODES: STATUS_CODES,
        'default': http
    });

    var require$$4$1 = /*@__PURE__*/getAugmentedNamespace(http$1);

    var followRedirects = {exports: {}};

    var require$$5 = /*@__PURE__*/getAugmentedNamespace(url$1);

    var require$$7 = /*@__PURE__*/getAugmentedNamespace(stream);

    function compare(a, b) {
      if (a === b) {
        return 0;
      }

      var x = a.length;
      var y = b.length;

      for (var i = 0, len = Math.min(x, y); i < len; ++i) {
        if (a[i] !== b[i]) {
          x = a[i];
          y = b[i];
          break;
        }
      }

      if (x < y) {
        return -1;
      }
      if (y < x) {
        return 1;
      }
      return 0;
    }
    var hasOwn = Object.prototype.hasOwnProperty;

    var objectKeys = Object.keys || function (obj) {
      var keys = [];
      for (var key in obj) {
        if (hasOwn.call(obj, key)) keys.push(key);
      }
      return keys;
    };
    var pSlice = Array.prototype.slice;
    var _functionsHaveNames;
    function functionsHaveNames() {
      if (typeof _functionsHaveNames !== 'undefined') {
        return _functionsHaveNames;
      }
      return _functionsHaveNames = (function () {
        return function foo() {}.name === 'foo';
      }());
    }
    function pToString (obj) {
      return Object.prototype.toString.call(obj);
    }
    function isView(arrbuf) {
      if (isBuffer$2(arrbuf)) {
        return false;
      }
      if (typeof global$1.ArrayBuffer !== 'function') {
        return false;
      }
      if (typeof ArrayBuffer.isView === 'function') {
        return ArrayBuffer.isView(arrbuf);
      }
      if (!arrbuf) {
        return false;
      }
      if (arrbuf instanceof DataView) {
        return true;
      }
      if (arrbuf.buffer && arrbuf.buffer instanceof ArrayBuffer) {
        return true;
      }
      return false;
    }
    // 1. The assert module provides functions that throw
    // AssertionError's when particular conditions are not met. The
    // assert module must conform to the following interface.

    function assert$1(value, message) {
      if (!value) fail(value, true, message, '==', ok);
    }

    // 2. The AssertionError is defined in assert.
    // new assert.AssertionError({ message: message,
    //                             actual: actual,
    //                             expected: expected })

    var regex = /\s*function\s+([^\(\s]*)\s*/;
    // based on https://github.com/ljharb/function.prototype.name/blob/adeeeec8bfcc6068b187d7d9fb3d5bb1d3a30899/implementation.js
    function getName(func) {
      if (!isFunction$1(func)) {
        return;
      }
      if (functionsHaveNames()) {
        return func.name;
      }
      var str = func.toString();
      var match = str.match(regex);
      return match && match[1];
    }
    assert$1.AssertionError = AssertionError;
    function AssertionError(options) {
      this.name = 'AssertionError';
      this.actual = options.actual;
      this.expected = options.expected;
      this.operator = options.operator;
      if (options.message) {
        this.message = options.message;
        this.generatedMessage = false;
      } else {
        this.message = getMessage(this);
        this.generatedMessage = true;
      }
      var stackStartFunction = options.stackStartFunction || fail;
      if (Error.captureStackTrace) {
        Error.captureStackTrace(this, stackStartFunction);
      } else {
        // non v8 browsers so we can have a stacktrace
        var err = new Error();
        if (err.stack) {
          var out = err.stack;

          // try to strip useless frames
          var fn_name = getName(stackStartFunction);
          var idx = out.indexOf('\n' + fn_name);
          if (idx >= 0) {
            // once we have located the function frame
            // we need to strip out everything before it (and its line)
            var next_line = out.indexOf('\n', idx + 1);
            out = out.substring(next_line + 1);
          }

          this.stack = out;
        }
      }
    }

    // assert.AssertionError instanceof Error
    inherits$1(AssertionError, Error);

    function truncate(s, n) {
      if (typeof s === 'string') {
        return s.length < n ? s : s.slice(0, n);
      } else {
        return s;
      }
    }
    function inspect(something) {
      if (functionsHaveNames() || !isFunction$1(something)) {
        return inspect$1(something);
      }
      var rawname = getName(something);
      var name = rawname ? ': ' + rawname : '';
      return '[Function' +  name + ']';
    }
    function getMessage(self) {
      return truncate(inspect(self.actual), 128) + ' ' +
             self.operator + ' ' +
             truncate(inspect(self.expected), 128);
    }

    // At present only the three keys mentioned above are used and
    // understood by the spec. Implementations or sub modules can pass
    // other keys to the AssertionError's constructor - they will be
    // ignored.

    // 3. All of the following functions must throw an AssertionError
    // when a corresponding condition is not met, with a message that
    // may be undefined if not provided.  All assertion methods provide
    // both the actual and expected values to the assertion error for
    // display purposes.

    function fail(actual, expected, message, operator, stackStartFunction) {
      throw new AssertionError({
        message: message,
        actual: actual,
        expected: expected,
        operator: operator,
        stackStartFunction: stackStartFunction
      });
    }

    // EXTENSION! allows for well behaved errors defined elsewhere.
    assert$1.fail = fail;

    // 4. Pure assertion tests whether a value is truthy, as determined
    // by !!guard.
    // assert.ok(guard, message_opt);
    // This statement is equivalent to assert.equal(true, !!guard,
    // message_opt);. To test strictly for the value true, use
    // assert.strictEqual(true, guard, message_opt);.

    function ok(value, message) {
      if (!value) fail(value, true, message, '==', ok);
    }
    assert$1.ok = ok;

    // 5. The equality assertion tests shallow, coercive equality with
    // ==.
    // assert.equal(actual, expected, message_opt);
    assert$1.equal = equal;
    function equal(actual, expected, message) {
      if (actual != expected) fail(actual, expected, message, '==', equal);
    }

    // 6. The non-equality assertion tests for whether two objects are not equal
    // with != assert.notEqual(actual, expected, message_opt);
    assert$1.notEqual = notEqual;
    function notEqual(actual, expected, message) {
      if (actual == expected) {
        fail(actual, expected, message, '!=', notEqual);
      }
    }

    // 7. The equivalence assertion tests a deep equality relation.
    // assert.deepEqual(actual, expected, message_opt);
    assert$1.deepEqual = deepEqual;
    function deepEqual(actual, expected, message) {
      if (!_deepEqual(actual, expected, false)) {
        fail(actual, expected, message, 'deepEqual', deepEqual);
      }
    }
    assert$1.deepStrictEqual = deepStrictEqual;
    function deepStrictEqual(actual, expected, message) {
      if (!_deepEqual(actual, expected, true)) {
        fail(actual, expected, message, 'deepStrictEqual', deepStrictEqual);
      }
    }

    function _deepEqual(actual, expected, strict, memos) {
      // 7.1. All identical values are equivalent, as determined by ===.
      if (actual === expected) {
        return true;
      } else if (isBuffer$2(actual) && isBuffer$2(expected)) {
        return compare(actual, expected) === 0;

      // 7.2. If the expected value is a Date object, the actual value is
      // equivalent if it is also a Date object that refers to the same time.
      } else if (isDate(actual) && isDate(expected)) {
        return actual.getTime() === expected.getTime();

      // 7.3 If the expected value is a RegExp object, the actual value is
      // equivalent if it is also a RegExp object with the same source and
      // properties (`global`, `multiline`, `lastIndex`, `ignoreCase`).
      } else if (isRegExp(actual) && isRegExp(expected)) {
        return actual.source === expected.source &&
               actual.global === expected.global &&
               actual.multiline === expected.multiline &&
               actual.lastIndex === expected.lastIndex &&
               actual.ignoreCase === expected.ignoreCase;

      // 7.4. Other pairs that do not both pass typeof value == 'object',
      // equivalence is determined by ==.
      } else if ((actual === null || typeof actual !== 'object') &&
                 (expected === null || typeof expected !== 'object')) {
        return strict ? actual === expected : actual == expected;

      // If both values are instances of typed arrays, wrap their underlying
      // ArrayBuffers in a Buffer each to increase performance
      // This optimization requires the arrays to have the same type as checked by
      // Object.prototype.toString (aka pToString). Never perform binary
      // comparisons for Float*Arrays, though, since e.g. +0 === -0 but their
      // bit patterns are not identical.
      } else if (isView(actual) && isView(expected) &&
                 pToString(actual) === pToString(expected) &&
                 !(actual instanceof Float32Array ||
                   actual instanceof Float64Array)) {
        return compare(new Uint8Array(actual.buffer),
                       new Uint8Array(expected.buffer)) === 0;

      // 7.5 For all other Object pairs, including Array objects, equivalence is
      // determined by having the same number of owned properties (as verified
      // with Object.prototype.hasOwnProperty.call), the same set of keys
      // (although not necessarily the same order), equivalent values for every
      // corresponding key, and an identical 'prototype' property. Note: this
      // accounts for both named and indexed properties on Arrays.
      } else if (isBuffer$2(actual) !== isBuffer$2(expected)) {
        return false;
      } else {
        memos = memos || {actual: [], expected: []};

        var actualIndex = memos.actual.indexOf(actual);
        if (actualIndex !== -1) {
          if (actualIndex === memos.expected.indexOf(expected)) {
            return true;
          }
        }

        memos.actual.push(actual);
        memos.expected.push(expected);

        return objEquiv(actual, expected, strict, memos);
      }
    }

    function isArguments(object) {
      return Object.prototype.toString.call(object) == '[object Arguments]';
    }

    function objEquiv(a, b, strict, actualVisitedObjects) {
      if (a === null || a === undefined || b === null || b === undefined)
        return false;
      // if one is a primitive, the other must be same
      if (isPrimitive(a) || isPrimitive(b))
        return a === b;
      if (strict && Object.getPrototypeOf(a) !== Object.getPrototypeOf(b))
        return false;
      var aIsArgs = isArguments(a);
      var bIsArgs = isArguments(b);
      if ((aIsArgs && !bIsArgs) || (!aIsArgs && bIsArgs))
        return false;
      if (aIsArgs) {
        a = pSlice.call(a);
        b = pSlice.call(b);
        return _deepEqual(a, b, strict);
      }
      var ka = objectKeys(a);
      var kb = objectKeys(b);
      var key, i;
      // having the same number of owned properties (keys incorporates
      // hasOwnProperty)
      if (ka.length !== kb.length)
        return false;
      //the same set of keys (although not necessarily the same order),
      ka.sort();
      kb.sort();
      //~~~cheap key test
      for (i = ka.length - 1; i >= 0; i--) {
        if (ka[i] !== kb[i])
          return false;
      }
      //equivalent values for every corresponding key, and
      //~~~possibly expensive deep test
      for (i = ka.length - 1; i >= 0; i--) {
        key = ka[i];
        if (!_deepEqual(a[key], b[key], strict, actualVisitedObjects))
          return false;
      }
      return true;
    }

    // 8. The non-equivalence assertion tests for any deep inequality.
    // assert.notDeepEqual(actual, expected, message_opt);
    assert$1.notDeepEqual = notDeepEqual;
    function notDeepEqual(actual, expected, message) {
      if (_deepEqual(actual, expected, false)) {
        fail(actual, expected, message, 'notDeepEqual', notDeepEqual);
      }
    }

    assert$1.notDeepStrictEqual = notDeepStrictEqual;
    function notDeepStrictEqual(actual, expected, message) {
      if (_deepEqual(actual, expected, true)) {
        fail(actual, expected, message, 'notDeepStrictEqual', notDeepStrictEqual);
      }
    }


    // 9. The strict equality assertion tests strict equality, as determined by ===.
    // assert.strictEqual(actual, expected, message_opt);
    assert$1.strictEqual = strictEqual;
    function strictEqual(actual, expected, message) {
      if (actual !== expected) {
        fail(actual, expected, message, '===', strictEqual);
      }
    }

    // 10. The strict non-equality assertion tests for strict inequality, as
    // determined by !==.  assert.notStrictEqual(actual, expected, message_opt);
    assert$1.notStrictEqual = notStrictEqual;
    function notStrictEqual(actual, expected, message) {
      if (actual === expected) {
        fail(actual, expected, message, '!==', notStrictEqual);
      }
    }

    function expectedException(actual, expected) {
      if (!actual || !expected) {
        return false;
      }

      if (Object.prototype.toString.call(expected) == '[object RegExp]') {
        return expected.test(actual);
      }

      try {
        if (actual instanceof expected) {
          return true;
        }
      } catch (e) {
        // Ignore.  The instanceof check doesn't work for arrow functions.
      }

      if (Error.isPrototypeOf(expected)) {
        return false;
      }

      return expected.call({}, actual) === true;
    }

    function _tryBlock(block) {
      var error;
      try {
        block();
      } catch (e) {
        error = e;
      }
      return error;
    }

    function _throws(shouldThrow, block, expected, message) {
      var actual;

      if (typeof block !== 'function') {
        throw new TypeError('"block" argument must be a function');
      }

      if (typeof expected === 'string') {
        message = expected;
        expected = null;
      }

      actual = _tryBlock(block);

      message = (expected && expected.name ? ' (' + expected.name + ').' : '.') +
                (message ? ' ' + message : '.');

      if (shouldThrow && !actual) {
        fail(actual, expected, 'Missing expected exception' + message);
      }

      var userProvidedMessage = typeof message === 'string';
      var isUnwantedException = !shouldThrow && isError(actual);
      var isUnexpectedException = !shouldThrow && actual && !expected;

      if ((isUnwantedException &&
          userProvidedMessage &&
          expectedException(actual, expected)) ||
          isUnexpectedException) {
        fail(actual, expected, 'Got unwanted exception' + message);
      }

      if ((shouldThrow && actual && expected &&
          !expectedException(actual, expected)) || (!shouldThrow && actual)) {
        throw actual;
      }
    }

    // 11. Expected to throw an error:
    // assert.throws(block, Error_opt, message_opt);
    assert$1.throws = throws;
    function throws(block, /*optional*/error, /*optional*/message) {
      _throws(true, block, error, message);
    }

    // EXTENSION! This is annoying to write outside this module.
    assert$1.doesNotThrow = doesNotThrow;
    function doesNotThrow(block, /*optional*/error, /*optional*/message) {
      _throws(false, block, error, message);
    }

    assert$1.ifError = ifError;
    function ifError(err) {
      if (err) throw err;
    }

    var assert$2 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        'default': assert$1,
        AssertionError: AssertionError,
        fail: fail,
        ok: ok,
        assert: ok,
        equal: equal,
        notEqual: notEqual,
        deepEqual: deepEqual,
        deepStrictEqual: deepStrictEqual,
        notDeepEqual: notDeepEqual,
        notDeepStrictEqual: notDeepStrictEqual,
        strictEqual: strictEqual,
        notStrictEqual: notStrictEqual,
        throws: throws,
        doesNotThrow: doesNotThrow,
        ifError: ifError
    });

    var require$$4 = /*@__PURE__*/getAugmentedNamespace(assert$2);

    var browser$1 = true;

    var src = {exports: {}};

    var browser = {exports: {}};

    /**
     * Helpers.
     */

    var ms;
    var hasRequiredMs;

    function requireMs () {
    	if (hasRequiredMs) return ms;
    	hasRequiredMs = 1;
    	var s = 1000;
    	var m = s * 60;
    	var h = m * 60;
    	var d = h * 24;
    	var w = d * 7;
    	var y = d * 365.25;

    	/**
    	 * Parse or format the given `val`.
    	 *
    	 * Options:
    	 *
    	 *  - `long` verbose formatting [false]
    	 *
    	 * @param {String|Number} val
    	 * @param {Object} [options]
    	 * @throws {Error} throw an error if val is not a non-empty string or a number
    	 * @return {String|Number}
    	 * @api public
    	 */

    	ms = function(val, options) {
    	  options = options || {};
    	  var type = typeof val;
    	  if (type === 'string' && val.length > 0) {
    	    return parse(val);
    	  } else if (type === 'number' && isFinite(val)) {
    	    return options.long ? fmtLong(val) : fmtShort(val);
    	  }
    	  throw new Error(
    	    'val is not a non-empty string or a valid number. val=' +
    	      JSON.stringify(val)
    	  );
    	};

    	/**
    	 * Parse the given `str` and return milliseconds.
    	 *
    	 * @param {String} str
    	 * @return {Number}
    	 * @api private
    	 */

    	function parse(str) {
    	  str = String(str);
    	  if (str.length > 100) {
    	    return;
    	  }
    	  var match = /^(-?(?:\d+)?\.?\d+) *(milliseconds?|msecs?|ms|seconds?|secs?|s|minutes?|mins?|m|hours?|hrs?|h|days?|d|weeks?|w|years?|yrs?|y)?$/i.exec(
    	    str
    	  );
    	  if (!match) {
    	    return;
    	  }
    	  var n = parseFloat(match[1]);
    	  var type = (match[2] || 'ms').toLowerCase();
    	  switch (type) {
    	    case 'years':
    	    case 'year':
    	    case 'yrs':
    	    case 'yr':
    	    case 'y':
    	      return n * y;
    	    case 'weeks':
    	    case 'week':
    	    case 'w':
    	      return n * w;
    	    case 'days':
    	    case 'day':
    	    case 'd':
    	      return n * d;
    	    case 'hours':
    	    case 'hour':
    	    case 'hrs':
    	    case 'hr':
    	    case 'h':
    	      return n * h;
    	    case 'minutes':
    	    case 'minute':
    	    case 'mins':
    	    case 'min':
    	    case 'm':
    	      return n * m;
    	    case 'seconds':
    	    case 'second':
    	    case 'secs':
    	    case 'sec':
    	    case 's':
    	      return n * s;
    	    case 'milliseconds':
    	    case 'millisecond':
    	    case 'msecs':
    	    case 'msec':
    	    case 'ms':
    	      return n;
    	    default:
    	      return undefined;
    	  }
    	}

    	/**
    	 * Short format for `ms`.
    	 *
    	 * @param {Number} ms
    	 * @return {String}
    	 * @api private
    	 */

    	function fmtShort(ms) {
    	  var msAbs = Math.abs(ms);
    	  if (msAbs >= d) {
    	    return Math.round(ms / d) + 'd';
    	  }
    	  if (msAbs >= h) {
    	    return Math.round(ms / h) + 'h';
    	  }
    	  if (msAbs >= m) {
    	    return Math.round(ms / m) + 'm';
    	  }
    	  if (msAbs >= s) {
    	    return Math.round(ms / s) + 's';
    	  }
    	  return ms + 'ms';
    	}

    	/**
    	 * Long format for `ms`.
    	 *
    	 * @param {Number} ms
    	 * @return {String}
    	 * @api private
    	 */

    	function fmtLong(ms) {
    	  var msAbs = Math.abs(ms);
    	  if (msAbs >= d) {
    	    return plural(ms, msAbs, d, 'day');
    	  }
    	  if (msAbs >= h) {
    	    return plural(ms, msAbs, h, 'hour');
    	  }
    	  if (msAbs >= m) {
    	    return plural(ms, msAbs, m, 'minute');
    	  }
    	  if (msAbs >= s) {
    	    return plural(ms, msAbs, s, 'second');
    	  }
    	  return ms + ' ms';
    	}

    	/**
    	 * Pluralization helper.
    	 */

    	function plural(ms, msAbs, n, name) {
    	  var isPlural = msAbs >= n * 1.5;
    	  return Math.round(ms / n) + ' ' + name + (isPlural ? 's' : '');
    	}
    	return ms;
    }

    var common;
    var hasRequiredCommon;

    function requireCommon () {
    	if (hasRequiredCommon) return common;
    	hasRequiredCommon = 1;
    	/**
    	 * This is the common logic for both the Node.js and web browser
    	 * implementations of `debug()`.
    	 */

    	function setup(env) {
    		createDebug.debug = createDebug;
    		createDebug.default = createDebug;
    		createDebug.coerce = coerce;
    		createDebug.disable = disable;
    		createDebug.enable = enable;
    		createDebug.enabled = enabled;
    		createDebug.humanize = requireMs();
    		createDebug.destroy = destroy;

    		Object.keys(env).forEach(key => {
    			createDebug[key] = env[key];
    		});

    		/**
    		* The currently active debug mode names, and names to skip.
    		*/

    		createDebug.names = [];
    		createDebug.skips = [];

    		/**
    		* Map of special "%n" handling functions, for the debug "format" argument.
    		*
    		* Valid key names are a single, lower or upper-case letter, i.e. "n" and "N".
    		*/
    		createDebug.formatters = {};

    		/**
    		* Selects a color for a debug namespace
    		* @param {String} namespace The namespace string for the debug instance to be colored
    		* @return {Number|String} An ANSI color code for the given namespace
    		* @api private
    		*/
    		function selectColor(namespace) {
    			let hash = 0;

    			for (let i = 0; i < namespace.length; i++) {
    				hash = ((hash << 5) - hash) + namespace.charCodeAt(i);
    				hash |= 0; // Convert to 32bit integer
    			}

    			return createDebug.colors[Math.abs(hash) % createDebug.colors.length];
    		}
    		createDebug.selectColor = selectColor;

    		/**
    		* Create a debugger with the given `namespace`.
    		*
    		* @param {String} namespace
    		* @return {Function}
    		* @api public
    		*/
    		function createDebug(namespace) {
    			let prevTime;
    			let enableOverride = null;
    			let namespacesCache;
    			let enabledCache;

    			function debug(...args) {
    				// Disabled?
    				if (!debug.enabled) {
    					return;
    				}

    				const self = debug;

    				// Set `diff` timestamp
    				const curr = Number(new Date());
    				const ms = curr - (prevTime || curr);
    				self.diff = ms;
    				self.prev = prevTime;
    				self.curr = curr;
    				prevTime = curr;

    				args[0] = createDebug.coerce(args[0]);

    				if (typeof args[0] !== 'string') {
    					// Anything else let's inspect with %O
    					args.unshift('%O');
    				}

    				// Apply any `formatters` transformations
    				let index = 0;
    				args[0] = args[0].replace(/%([a-zA-Z%])/g, (match, format) => {
    					// If we encounter an escaped % then don't increase the array index
    					if (match === '%%') {
    						return '%';
    					}
    					index++;
    					const formatter = createDebug.formatters[format];
    					if (typeof formatter === 'function') {
    						const val = args[index];
    						match = formatter.call(self, val);

    						// Now we need to remove `args[index]` since it's inlined in the `format`
    						args.splice(index, 1);
    						index--;
    					}
    					return match;
    				});

    				// Apply env-specific formatting (colors, etc.)
    				createDebug.formatArgs.call(self, args);

    				const logFn = self.log || createDebug.log;
    				logFn.apply(self, args);
    			}

    			debug.namespace = namespace;
    			debug.useColors = createDebug.useColors();
    			debug.color = createDebug.selectColor(namespace);
    			debug.extend = extend;
    			debug.destroy = createDebug.destroy; // XXX Temporary. Will be removed in the next major release.

    			Object.defineProperty(debug, 'enabled', {
    				enumerable: true,
    				configurable: false,
    				get: () => {
    					if (enableOverride !== null) {
    						return enableOverride;
    					}
    					if (namespacesCache !== createDebug.namespaces) {
    						namespacesCache = createDebug.namespaces;
    						enabledCache = createDebug.enabled(namespace);
    					}

    					return enabledCache;
    				},
    				set: v => {
    					enableOverride = v;
    				}
    			});

    			// Env-specific initialization logic for debug instances
    			if (typeof createDebug.init === 'function') {
    				createDebug.init(debug);
    			}

    			return debug;
    		}

    		function extend(namespace, delimiter) {
    			const newDebug = createDebug(this.namespace + (typeof delimiter === 'undefined' ? ':' : delimiter) + namespace);
    			newDebug.log = this.log;
    			return newDebug;
    		}

    		/**
    		* Enables a debug mode by namespaces. This can include modes
    		* separated by a colon and wildcards.
    		*
    		* @param {String} namespaces
    		* @api public
    		*/
    		function enable(namespaces) {
    			createDebug.save(namespaces);
    			createDebug.namespaces = namespaces;

    			createDebug.names = [];
    			createDebug.skips = [];

    			let i;
    			const split = (typeof namespaces === 'string' ? namespaces : '').split(/[\s,]+/);
    			const len = split.length;

    			for (i = 0; i < len; i++) {
    				if (!split[i]) {
    					// ignore empty strings
    					continue;
    				}

    				namespaces = split[i].replace(/\*/g, '.*?');

    				if (namespaces[0] === '-') {
    					createDebug.skips.push(new RegExp('^' + namespaces.slice(1) + '$'));
    				} else {
    					createDebug.names.push(new RegExp('^' + namespaces + '$'));
    				}
    			}
    		}

    		/**
    		* Disable debug output.
    		*
    		* @return {String} namespaces
    		* @api public
    		*/
    		function disable() {
    			const namespaces = [
    				...createDebug.names.map(toNamespace),
    				...createDebug.skips.map(toNamespace).map(namespace => '-' + namespace)
    			].join(',');
    			createDebug.enable('');
    			return namespaces;
    		}

    		/**
    		* Returns true if the given mode name is enabled, false otherwise.
    		*
    		* @param {String} name
    		* @return {Boolean}
    		* @api public
    		*/
    		function enabled(name) {
    			if (name[name.length - 1] === '*') {
    				return true;
    			}

    			let i;
    			let len;

    			for (i = 0, len = createDebug.skips.length; i < len; i++) {
    				if (createDebug.skips[i].test(name)) {
    					return false;
    				}
    			}

    			for (i = 0, len = createDebug.names.length; i < len; i++) {
    				if (createDebug.names[i].test(name)) {
    					return true;
    				}
    			}

    			return false;
    		}

    		/**
    		* Convert regexp to namespace
    		*
    		* @param {RegExp} regxep
    		* @return {String} namespace
    		* @api private
    		*/
    		function toNamespace(regexp) {
    			return regexp.toString()
    				.substring(2, regexp.toString().length - 2)
    				.replace(/\.\*\?$/, '*');
    		}

    		/**
    		* Coerce `val`.
    		*
    		* @param {Mixed} val
    		* @return {Mixed}
    		* @api private
    		*/
    		function coerce(val) {
    			if (val instanceof Error) {
    				return val.stack || val.message;
    			}
    			return val;
    		}

    		/**
    		* XXX DO NOT USE. This is a temporary stub function.
    		* XXX It WILL be removed in the next major release.
    		*/
    		function destroy() {
    			console.warn('Instance method `debug.destroy()` is deprecated and no longer does anything. It will be removed in the next major version of `debug`.');
    		}

    		createDebug.enable(createDebug.load());

    		return createDebug;
    	}

    	common = setup;
    	return common;
    }

    var hasRequiredBrowser;

    function requireBrowser () {
    	if (hasRequiredBrowser) return browser.exports;
    	hasRequiredBrowser = 1;
    	(function (module, exports) {
    		/**
    		 * This is the web browser implementation of `debug()`.
    		 */

    		exports.formatArgs = formatArgs;
    		exports.save = save;
    		exports.load = load;
    		exports.useColors = useColors;
    		exports.storage = localstorage();
    		exports.destroy = (() => {
    			let warned = false;

    			return () => {
    				if (!warned) {
    					warned = true;
    					console.warn('Instance method `debug.destroy()` is deprecated and no longer does anything. It will be removed in the next major version of `debug`.');
    				}
    			};
    		})();

    		/**
    		 * Colors.
    		 */

    		exports.colors = [
    			'#0000CC',
    			'#0000FF',
    			'#0033CC',
    			'#0033FF',
    			'#0066CC',
    			'#0066FF',
    			'#0099CC',
    			'#0099FF',
    			'#00CC00',
    			'#00CC33',
    			'#00CC66',
    			'#00CC99',
    			'#00CCCC',
    			'#00CCFF',
    			'#3300CC',
    			'#3300FF',
    			'#3333CC',
    			'#3333FF',
    			'#3366CC',
    			'#3366FF',
    			'#3399CC',
    			'#3399FF',
    			'#33CC00',
    			'#33CC33',
    			'#33CC66',
    			'#33CC99',
    			'#33CCCC',
    			'#33CCFF',
    			'#6600CC',
    			'#6600FF',
    			'#6633CC',
    			'#6633FF',
    			'#66CC00',
    			'#66CC33',
    			'#9900CC',
    			'#9900FF',
    			'#9933CC',
    			'#9933FF',
    			'#99CC00',
    			'#99CC33',
    			'#CC0000',
    			'#CC0033',
    			'#CC0066',
    			'#CC0099',
    			'#CC00CC',
    			'#CC00FF',
    			'#CC3300',
    			'#CC3333',
    			'#CC3366',
    			'#CC3399',
    			'#CC33CC',
    			'#CC33FF',
    			'#CC6600',
    			'#CC6633',
    			'#CC9900',
    			'#CC9933',
    			'#CCCC00',
    			'#CCCC33',
    			'#FF0000',
    			'#FF0033',
    			'#FF0066',
    			'#FF0099',
    			'#FF00CC',
    			'#FF00FF',
    			'#FF3300',
    			'#FF3333',
    			'#FF3366',
    			'#FF3399',
    			'#FF33CC',
    			'#FF33FF',
    			'#FF6600',
    			'#FF6633',
    			'#FF9900',
    			'#FF9933',
    			'#FFCC00',
    			'#FFCC33'
    		];

    		/**
    		 * Currently only WebKit-based Web Inspectors, Firefox >= v31,
    		 * and the Firebug extension (any Firefox version) are known
    		 * to support "%c" CSS customizations.
    		 *
    		 * TODO: add a `localStorage` variable to explicitly enable/disable colors
    		 */

    		// eslint-disable-next-line complexity
    		function useColors() {
    			// NB: In an Electron preload script, document will be defined but not fully
    			// initialized. Since we know we're in Chrome, we'll just detect this case
    			// explicitly
    			if (typeof window !== 'undefined' && window.process && (window.process.type === 'renderer' || window.process.__nwjs)) {
    				return true;
    			}

    			// Internet Explorer and Edge do not support colors.
    			if (typeof navigator !== 'undefined' && navigator.userAgent && navigator.userAgent.toLowerCase().match(/(edge|trident)\/(\d+)/)) {
    				return false;
    			}

    			// Is webkit? http://stackoverflow.com/a/16459606/376773
    			// document is undefined in react-native: https://github.com/facebook/react-native/pull/1632
    			return (typeof document !== 'undefined' && document.documentElement && document.documentElement.style && document.documentElement.style.WebkitAppearance) ||
    				// Is firebug? http://stackoverflow.com/a/398120/376773
    				(typeof window !== 'undefined' && window.console && (window.console.firebug || (window.console.exception && window.console.table))) ||
    				// Is firefox >= v31?
    				// https://developer.mozilla.org/en-US/docs/Tools/Web_Console#Styling_messages
    				(typeof navigator !== 'undefined' && navigator.userAgent && navigator.userAgent.toLowerCase().match(/firefox\/(\d+)/) && parseInt(RegExp.$1, 10) >= 31) ||
    				// Double check webkit in userAgent just in case we are in a worker
    				(typeof navigator !== 'undefined' && navigator.userAgent && navigator.userAgent.toLowerCase().match(/applewebkit\/(\d+)/));
    		}

    		/**
    		 * Colorize log arguments if enabled.
    		 *
    		 * @api public
    		 */

    		function formatArgs(args) {
    			args[0] = (this.useColors ? '%c' : '') +
    				this.namespace +
    				(this.useColors ? ' %c' : ' ') +
    				args[0] +
    				(this.useColors ? '%c ' : ' ') +
    				'+' + module.exports.humanize(this.diff);

    			if (!this.useColors) {
    				return;
    			}

    			const c = 'color: ' + this.color;
    			args.splice(1, 0, c, 'color: inherit');

    			// The final "%c" is somewhat tricky, because there could be other
    			// arguments passed either before or after the %c, so we need to
    			// figure out the correct index to insert the CSS into
    			let index = 0;
    			let lastC = 0;
    			args[0].replace(/%[a-zA-Z%]/g, match => {
    				if (match === '%%') {
    					return;
    				}
    				index++;
    				if (match === '%c') {
    					// We only are interested in the *last* %c
    					// (the user may have provided their own)
    					lastC = index;
    				}
    			});

    			args.splice(lastC, 0, c);
    		}

    		/**
    		 * Invokes `console.debug()` when available.
    		 * No-op when `console.debug` is not a "function".
    		 * If `console.debug` is not available, falls back
    		 * to `console.log`.
    		 *
    		 * @api public
    		 */
    		exports.log = console.debug || console.log || (() => {});

    		/**
    		 * Save `namespaces`.
    		 *
    		 * @param {String} namespaces
    		 * @api private
    		 */
    		function save(namespaces) {
    			try {
    				if (namespaces) {
    					exports.storage.setItem('debug', namespaces);
    				} else {
    					exports.storage.removeItem('debug');
    				}
    			} catch (error) {
    				// Swallow
    				// XXX (@Qix-) should we be logging these?
    			}
    		}

    		/**
    		 * Load `namespaces`.
    		 *
    		 * @return {String} returns the previously persisted debug modes
    		 * @api private
    		 */
    		function load() {
    			let r;
    			try {
    				r = exports.storage.getItem('debug');
    			} catch (error) {
    				// Swallow
    				// XXX (@Qix-) should we be logging these?
    			}

    			// If debug isn't set in LS, and we're in Electron, try to load $DEBUG
    			if (!r && typeof process !== 'undefined' && 'env' in process) {
    				r = process.env.DEBUG;
    			}

    			return r;
    		}

    		/**
    		 * Localstorage attempts to return the localstorage.
    		 *
    		 * This is necessary because safari throws
    		 * when a user disables cookies/localstorage
    		 * and you attempt to access it.
    		 *
    		 * @return {LocalStorage}
    		 * @api private
    		 */

    		function localstorage() {
    			try {
    				// TVMLKit (Apple TV JS Runtime) does not have a window object, just localStorage in the global context
    				// The Browser also has localStorage in the global context.
    				return localStorage;
    			} catch (error) {
    				// Swallow
    				// XXX (@Qix-) should we be logging these?
    			}
    		}

    		module.exports = requireCommon()(exports);

    		const {formatters} = module.exports;

    		/**
    		 * Map %j to `JSON.stringify()`, since no Web Inspectors do that by default.
    		 */

    		formatters.j = function (v) {
    			try {
    				return JSON.stringify(v);
    			} catch (error) {
    				return '[UnexpectedJSONParseError]: ' + error.message;
    			}
    		};
    } (browser, browser.exports));
    	return browser.exports;
    }

    var node = {exports: {}};

    // MIT lisence
    // from https://github.com/substack/tty-browserify/blob/1ba769a6429d242f36226538835b4034bf6b7886/index.js

    function isatty() {
      return false;
    }

    function ReadStream() {
      throw new Error('tty.ReadStream is not implemented');
    }

    function WriteStream() {
      throw new Error('tty.ReadStream is not implemented');
    }

    var tty = {
      isatty: isatty,
      ReadStream: ReadStream,
      WriteStream: WriteStream
    };

    var tty$1 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        isatty: isatty,
        ReadStream: ReadStream,
        WriteStream: WriteStream,
        'default': tty
    });

    var require$$0$2 = /*@__PURE__*/getAugmentedNamespace(tty$1);

    var require$$1 = /*@__PURE__*/getAugmentedNamespace(util$1);

    /*
    The MIT License (MIT)

    Copyright (c) 2016 CoderPuppy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

    */
    var _endianness;
    function endianness() {
      if (typeof _endianness === 'undefined') {
        var a = new ArrayBuffer(2);
        var b = new Uint8Array(a);
        var c = new Uint16Array(a);
        b[0] = 1;
        b[1] = 2;
        if (c[0] === 258) {
          _endianness = 'BE';
        } else if (c[0] === 513){
          _endianness = 'LE';
        } else {
          throw new Error('unable to figure out endianess');
        }
      }
      return _endianness;
    }

    function hostname() {
      if (typeof global$1.location !== 'undefined') {
        return global$1.location.hostname
      } else return '';
    }

    function loadavg() {
      return [];
    }

    function uptime() {
      return 0;
    }

    function freemem() {
      return Number.MAX_VALUE;
    }

    function totalmem() {
      return Number.MAX_VALUE;
    }

    function cpus() {
      return [];
    }

    function type() {
      return 'Browser';
    }

    function release () {
      if (typeof global$1.navigator !== 'undefined') {
        return global$1.navigator.appVersion;
      }
      return '';
    }

    function networkInterfaces(){}
    function getNetworkInterfaces(){}

    function arch() {
      return 'javascript';
    }

    function platform() {
      return 'browser';
    }

    function tmpDir() {
      return '/tmp';
    }
    var tmpdir = tmpDir;

    var EOL = '\n';
    var os = {
      EOL: EOL,
      tmpdir: tmpdir,
      tmpDir: tmpDir,
      networkInterfaces:networkInterfaces,
      getNetworkInterfaces: getNetworkInterfaces,
      release: release,
      type: type,
      cpus: cpus,
      totalmem: totalmem,
      freemem: freemem,
      uptime: uptime,
      loadavg: loadavg,
      hostname: hostname,
      endianness: endianness,
    };

    var os$1 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        endianness: endianness,
        hostname: hostname,
        loadavg: loadavg,
        uptime: uptime,
        freemem: freemem,
        totalmem: totalmem,
        cpus: cpus,
        type: type,
        release: release,
        networkInterfaces: networkInterfaces,
        getNetworkInterfaces: getNetworkInterfaces,
        arch: arch,
        platform: platform,
        tmpDir: tmpDir,
        tmpdir: tmpdir,
        EOL: EOL,
        'default': os
    });

    var require$$0$1 = /*@__PURE__*/getAugmentedNamespace(os$1);

    var hasFlag;
    var hasRequiredHasFlag;

    function requireHasFlag () {
    	if (hasRequiredHasFlag) return hasFlag;
    	hasRequiredHasFlag = 1;

    	hasFlag = (flag, argv = process.argv) => {
    		const prefix = flag.startsWith('-') ? '' : (flag.length === 1 ? '-' : '--');
    		const position = argv.indexOf(prefix + flag);
    		const terminatorPosition = argv.indexOf('--');
    		return position !== -1 && (terminatorPosition === -1 || position < terminatorPosition);
    	};
    	return hasFlag;
    }

    var supportsColor_1;
    var hasRequiredSupportsColor;

    function requireSupportsColor () {
    	if (hasRequiredSupportsColor) return supportsColor_1;
    	hasRequiredSupportsColor = 1;
    	const os = require$$0$1;
    	const tty = require$$0$2;
    	const hasFlag = requireHasFlag();

    	const {env} = process;

    	let forceColor;
    	if (hasFlag('no-color') ||
    		hasFlag('no-colors') ||
    		hasFlag('color=false') ||
    		hasFlag('color=never')) {
    		forceColor = 0;
    	} else if (hasFlag('color') ||
    		hasFlag('colors') ||
    		hasFlag('color=true') ||
    		hasFlag('color=always')) {
    		forceColor = 1;
    	}

    	if ('FORCE_COLOR' in env) {
    		if (env.FORCE_COLOR === 'true') {
    			forceColor = 1;
    		} else if (env.FORCE_COLOR === 'false') {
    			forceColor = 0;
    		} else {
    			forceColor = env.FORCE_COLOR.length === 0 ? 1 : Math.min(parseInt(env.FORCE_COLOR, 10), 3);
    		}
    	}

    	function translateLevel(level) {
    		if (level === 0) {
    			return false;
    		}

    		return {
    			level,
    			hasBasic: true,
    			has256: level >= 2,
    			has16m: level >= 3
    		};
    	}

    	function supportsColor(haveStream, streamIsTTY) {
    		if (forceColor === 0) {
    			return 0;
    		}

    		if (hasFlag('color=16m') ||
    			hasFlag('color=full') ||
    			hasFlag('color=truecolor')) {
    			return 3;
    		}

    		if (hasFlag('color=256')) {
    			return 2;
    		}

    		if (haveStream && !streamIsTTY && forceColor === undefined) {
    			return 0;
    		}

    		const min = forceColor || 0;

    		if (env.TERM === 'dumb') {
    			return min;
    		}

    		if (process.platform === 'win32') {
    			// Windows 10 build 10586 is the first Windows release that supports 256 colors.
    			// Windows 10 build 14931 is the first release that supports 16m/TrueColor.
    			const osRelease = os.release().split('.');
    			if (
    				Number(osRelease[0]) >= 10 &&
    				Number(osRelease[2]) >= 10586
    			) {
    				return Number(osRelease[2]) >= 14931 ? 3 : 2;
    			}

    			return 1;
    		}

    		if ('CI' in env) {
    			if (['TRAVIS', 'CIRCLECI', 'APPVEYOR', 'GITLAB_CI', 'GITHUB_ACTIONS', 'BUILDKITE'].some(sign => sign in env) || env.CI_NAME === 'codeship') {
    				return 1;
    			}

    			return min;
    		}

    		if ('TEAMCITY_VERSION' in env) {
    			return /^(9\.(0*[1-9]\d*)\.|\d{2,}\.)/.test(env.TEAMCITY_VERSION) ? 1 : 0;
    		}

    		if (env.COLORTERM === 'truecolor') {
    			return 3;
    		}

    		if ('TERM_PROGRAM' in env) {
    			const version = parseInt((env.TERM_PROGRAM_VERSION || '').split('.')[0], 10);

    			switch (env.TERM_PROGRAM) {
    				case 'iTerm.app':
    					return version >= 3 ? 3 : 2;
    				case 'Apple_Terminal':
    					return 2;
    				// No default
    			}
    		}

    		if (/-256(color)?$/i.test(env.TERM)) {
    			return 2;
    		}

    		if (/^screen|^xterm|^vt100|^vt220|^rxvt|color|ansi|cygwin|linux/i.test(env.TERM)) {
    			return 1;
    		}

    		if ('COLORTERM' in env) {
    			return 1;
    		}

    		return min;
    	}

    	function getSupportLevel(stream) {
    		const level = supportsColor(stream, stream && stream.isTTY);
    		return translateLevel(level);
    	}

    	supportsColor_1 = {
    		supportsColor: getSupportLevel,
    		stdout: translateLevel(supportsColor(true, tty.isatty(1))),
    		stderr: translateLevel(supportsColor(true, tty.isatty(2)))
    	};
    	return supportsColor_1;
    }

    var hasRequiredNode;

    function requireNode () {
    	if (hasRequiredNode) return node.exports;
    	hasRequiredNode = 1;
    	(function (module, exports) {
    		const tty = require$$0$2;
    		const util = require$$1;

    		/**
    		 * This is the Node.js implementation of `debug()`.
    		 */

    		exports.init = init;
    		exports.log = log;
    		exports.formatArgs = formatArgs;
    		exports.save = save;
    		exports.load = load;
    		exports.useColors = useColors;
    		exports.destroy = util.deprecate(
    			() => {},
    			'Instance method `debug.destroy()` is deprecated and no longer does anything. It will be removed in the next major version of `debug`.'
    		);

    		/**
    		 * Colors.
    		 */

    		exports.colors = [6, 2, 3, 4, 5, 1];

    		try {
    			// Optional dependency (as in, doesn't need to be installed, NOT like optionalDependencies in package.json)
    			// eslint-disable-next-line import/no-extraneous-dependencies
    			const supportsColor = requireSupportsColor();

    			if (supportsColor && (supportsColor.stderr || supportsColor).level >= 2) {
    				exports.colors = [
    					20,
    					21,
    					26,
    					27,
    					32,
    					33,
    					38,
    					39,
    					40,
    					41,
    					42,
    					43,
    					44,
    					45,
    					56,
    					57,
    					62,
    					63,
    					68,
    					69,
    					74,
    					75,
    					76,
    					77,
    					78,
    					79,
    					80,
    					81,
    					92,
    					93,
    					98,
    					99,
    					112,
    					113,
    					128,
    					129,
    					134,
    					135,
    					148,
    					149,
    					160,
    					161,
    					162,
    					163,
    					164,
    					165,
    					166,
    					167,
    					168,
    					169,
    					170,
    					171,
    					172,
    					173,
    					178,
    					179,
    					184,
    					185,
    					196,
    					197,
    					198,
    					199,
    					200,
    					201,
    					202,
    					203,
    					204,
    					205,
    					206,
    					207,
    					208,
    					209,
    					214,
    					215,
    					220,
    					221
    				];
    			}
    		} catch (error) {
    			// Swallow - we only care if `supports-color` is available; it doesn't have to be.
    		}

    		/**
    		 * Build up the default `inspectOpts` object from the environment variables.
    		 *
    		 *   $ DEBUG_COLORS=no DEBUG_DEPTH=10 DEBUG_SHOW_HIDDEN=enabled node script.js
    		 */

    		exports.inspectOpts = Object.keys(process.env).filter(key => {
    			return /^debug_/i.test(key);
    		}).reduce((obj, key) => {
    			// Camel-case
    			const prop = key
    				.substring(6)
    				.toLowerCase()
    				.replace(/_([a-z])/g, (_, k) => {
    					return k.toUpperCase();
    				});

    			// Coerce string value into JS value
    			let val = process.env[key];
    			if (/^(yes|on|true|enabled)$/i.test(val)) {
    				val = true;
    			} else if (/^(no|off|false|disabled)$/i.test(val)) {
    				val = false;
    			} else if (val === 'null') {
    				val = null;
    			} else {
    				val = Number(val);
    			}

    			obj[prop] = val;
    			return obj;
    		}, {});

    		/**
    		 * Is stdout a TTY? Colored output is enabled when `true`.
    		 */

    		function useColors() {
    			return 'colors' in exports.inspectOpts ?
    				Boolean(exports.inspectOpts.colors) :
    				tty.isatty(process.stderr.fd);
    		}

    		/**
    		 * Adds ANSI color escape codes if enabled.
    		 *
    		 * @api public
    		 */

    		function formatArgs(args) {
    			const {namespace: name, useColors} = this;

    			if (useColors) {
    				const c = this.color;
    				const colorCode = '\u001B[3' + (c < 8 ? c : '8;5;' + c);
    				const prefix = `  ${colorCode};1m${name} \u001B[0m`;

    				args[0] = prefix + args[0].split('\n').join('\n' + prefix);
    				args.push(colorCode + 'm+' + module.exports.humanize(this.diff) + '\u001B[0m');
    			} else {
    				args[0] = getDate() + name + ' ' + args[0];
    			}
    		}

    		function getDate() {
    			if (exports.inspectOpts.hideDate) {
    				return '';
    			}
    			return new Date().toISOString() + ' ';
    		}

    		/**
    		 * Invokes `util.format()` with the specified arguments and writes to stderr.
    		 */

    		function log(...args) {
    			return process.stderr.write(util.format(...args) + '\n');
    		}

    		/**
    		 * Save `namespaces`.
    		 *
    		 * @param {String} namespaces
    		 * @api private
    		 */
    		function save(namespaces) {
    			if (namespaces) {
    				process.env.DEBUG = namespaces;
    			} else {
    				// If you set a process.env field to null or undefined, it gets cast to the
    				// string 'null' or 'undefined'. Just delete instead.
    				delete process.env.DEBUG;
    			}
    		}

    		/**
    		 * Load `namespaces`.
    		 *
    		 * @return {String} returns the previously persisted debug modes
    		 * @api private
    		 */

    		function load() {
    			return process.env.DEBUG;
    		}

    		/**
    		 * Init logic for `debug` instances.
    		 *
    		 * Create a new `inspectOpts` object in case `useColors` is set
    		 * differently for a particular `debug` instance.
    		 */

    		function init(debug) {
    			debug.inspectOpts = {};

    			const keys = Object.keys(exports.inspectOpts);
    			for (let i = 0; i < keys.length; i++) {
    				debug.inspectOpts[keys[i]] = exports.inspectOpts[keys[i]];
    			}
    		}

    		module.exports = requireCommon()(exports);

    		const {formatters} = module.exports;

    		/**
    		 * Map %o to `util.inspect()`, all on a single line.
    		 */

    		formatters.o = function (v) {
    			this.inspectOpts.colors = this.useColors;
    			return util.inspect(v, this.inspectOpts)
    				.split('\n')
    				.map(str => str.trim())
    				.join(' ');
    		};

    		/**
    		 * Map %O to `util.inspect()`, allowing multiple lines if needed.
    		 */

    		formatters.O = function (v) {
    			this.inspectOpts.colors = this.useColors;
    			return util.inspect(v, this.inspectOpts);
    		};
    } (node, node.exports));
    	return node.exports;
    }

    var hasRequiredSrc;

    function requireSrc () {
    	if (hasRequiredSrc) return src.exports;
    	hasRequiredSrc = 1;
    	(function (module) {
    		if (typeof process === 'undefined' || process.type === 'renderer' || browser$1 === true || process.__nwjs) {
    			module.exports = requireBrowser();
    		} else {
    			module.exports = requireNode();
    		}
    } (src));
    	return src.exports;
    }

    var debug_1;
    var hasRequiredDebug;

    function requireDebug () {
    	if (hasRequiredDebug) return debug_1;
    	hasRequiredDebug = 1;
    	var debug;

    	debug_1 = function () {
    	  if (!debug) {
    	    try {
    	      /* eslint global-require: off */
    	      debug = requireSrc()("follow-redirects");
    	    }
    	    catch (error) { /* */ }
    	    if (typeof debug !== "function") {
    	      debug = function () { /* */ };
    	    }
    	  }
    	  debug.apply(null, arguments);
    	};
    	return debug_1;
    }

    var hasRequiredFollowRedirects;

    function requireFollowRedirects () {
    	if (hasRequiredFollowRedirects) return followRedirects.exports;
    	hasRequiredFollowRedirects = 1;
    	var url = require$$5;
    	var URL = url.URL;
    	var http = require$$4$1;
    	var https = require$$4$1;
    	var Writable = require$$7.Writable;
    	var assert = require$$4;
    	var debug = requireDebug();

    	// Create handlers that pass events from native requests
    	var events = ["abort", "aborted", "connect", "error", "socket", "timeout"];
    	var eventHandlers = Object.create(null);
    	events.forEach(function (event) {
    	  eventHandlers[event] = function (arg1, arg2, arg3) {
    	    this._redirectable.emit(event, arg1, arg2, arg3);
    	  };
    	});

    	// Error types with codes
    	var RedirectionError = createErrorType(
    	  "ERR_FR_REDIRECTION_FAILURE",
    	  "Redirected request failed"
    	);
    	var TooManyRedirectsError = createErrorType(
    	  "ERR_FR_TOO_MANY_REDIRECTS",
    	  "Maximum number of redirects exceeded"
    	);
    	var MaxBodyLengthExceededError = createErrorType(
    	  "ERR_FR_MAX_BODY_LENGTH_EXCEEDED",
    	  "Request body larger than maxBodyLength limit"
    	);
    	var WriteAfterEndError = createErrorType(
    	  "ERR_STREAM_WRITE_AFTER_END",
    	  "write after end"
    	);

    	// An HTTP(S) request that can be redirected
    	function RedirectableRequest(options, responseCallback) {
    	  // Initialize the request
    	  Writable.call(this);
    	  this._sanitizeOptions(options);
    	  this._options = options;
    	  this._ended = false;
    	  this._ending = false;
    	  this._redirectCount = 0;
    	  this._redirects = [];
    	  this._requestBodyLength = 0;
    	  this._requestBodyBuffers = [];

    	  // Attach a callback if passed
    	  if (responseCallback) {
    	    this.on("response", responseCallback);
    	  }

    	  // React to responses of native requests
    	  var self = this;
    	  this._onNativeResponse = function (response) {
    	    self._processResponse(response);
    	  };

    	  // Perform the first request
    	  this._performRequest();
    	}
    	RedirectableRequest.prototype = Object.create(Writable.prototype);

    	RedirectableRequest.prototype.abort = function () {
    	  abortRequest(this._currentRequest);
    	  this.emit("abort");
    	};

    	// Writes buffered data to the current native request
    	RedirectableRequest.prototype.write = function (data, encoding, callback) {
    	  // Writing is not allowed if end has been called
    	  if (this._ending) {
    	    throw new WriteAfterEndError();
    	  }

    	  // Validate input and shift parameters if necessary
    	  if (!(typeof data === "string" || typeof data === "object" && ("length" in data))) {
    	    throw new TypeError("data should be a string, Buffer or Uint8Array");
    	  }
    	  if (typeof encoding === "function") {
    	    callback = encoding;
    	    encoding = null;
    	  }

    	  // Ignore empty buffers, since writing them doesn't invoke the callback
    	  // https://github.com/nodejs/node/issues/22066
    	  if (data.length === 0) {
    	    if (callback) {
    	      callback();
    	    }
    	    return;
    	  }
    	  // Only write when we don't exceed the maximum body length
    	  if (this._requestBodyLength + data.length <= this._options.maxBodyLength) {
    	    this._requestBodyLength += data.length;
    	    this._requestBodyBuffers.push({ data: data, encoding: encoding });
    	    this._currentRequest.write(data, encoding, callback);
    	  }
    	  // Error when we exceed the maximum body length
    	  else {
    	    this.emit("error", new MaxBodyLengthExceededError());
    	    this.abort();
    	  }
    	};

    	// Ends the current native request
    	RedirectableRequest.prototype.end = function (data, encoding, callback) {
    	  // Shift parameters if necessary
    	  if (typeof data === "function") {
    	    callback = data;
    	    data = encoding = null;
    	  }
    	  else if (typeof encoding === "function") {
    	    callback = encoding;
    	    encoding = null;
    	  }

    	  // Write data if needed and end
    	  if (!data) {
    	    this._ended = this._ending = true;
    	    this._currentRequest.end(null, null, callback);
    	  }
    	  else {
    	    var self = this;
    	    var currentRequest = this._currentRequest;
    	    this.write(data, encoding, function () {
    	      self._ended = true;
    	      currentRequest.end(null, null, callback);
    	    });
    	    this._ending = true;
    	  }
    	};

    	// Sets a header value on the current native request
    	RedirectableRequest.prototype.setHeader = function (name, value) {
    	  this._options.headers[name] = value;
    	  this._currentRequest.setHeader(name, value);
    	};

    	// Clears a header value on the current native request
    	RedirectableRequest.prototype.removeHeader = function (name) {
    	  delete this._options.headers[name];
    	  this._currentRequest.removeHeader(name);
    	};

    	// Global timeout for all underlying requests
    	RedirectableRequest.prototype.setTimeout = function (msecs, callback) {
    	  var self = this;

    	  // Destroys the socket on timeout
    	  function destroyOnTimeout(socket) {
    	    socket.setTimeout(msecs);
    	    socket.removeListener("timeout", socket.destroy);
    	    socket.addListener("timeout", socket.destroy);
    	  }

    	  // Sets up a timer to trigger a timeout event
    	  function startTimer(socket) {
    	    if (self._timeout) {
    	      clearTimeout(self._timeout);
    	    }
    	    self._timeout = setTimeout(function () {
    	      self.emit("timeout");
    	      clearTimer();
    	    }, msecs);
    	    destroyOnTimeout(socket);
    	  }

    	  // Stops a timeout from triggering
    	  function clearTimer() {
    	    // Clear the timeout
    	    if (self._timeout) {
    	      clearTimeout(self._timeout);
    	      self._timeout = null;
    	    }

    	    // Clean up all attached listeners
    	    self.removeListener("abort", clearTimer);
    	    self.removeListener("error", clearTimer);
    	    self.removeListener("response", clearTimer);
    	    if (callback) {
    	      self.removeListener("timeout", callback);
    	    }
    	    if (!self.socket) {
    	      self._currentRequest.removeListener("socket", startTimer);
    	    }
    	  }

    	  // Attach callback if passed
    	  if (callback) {
    	    this.on("timeout", callback);
    	  }

    	  // Start the timer if or when the socket is opened
    	  if (this.socket) {
    	    startTimer(this.socket);
    	  }
    	  else {
    	    this._currentRequest.once("socket", startTimer);
    	  }

    	  // Clean up on events
    	  this.on("socket", destroyOnTimeout);
    	  this.on("abort", clearTimer);
    	  this.on("error", clearTimer);
    	  this.on("response", clearTimer);

    	  return this;
    	};

    	// Proxy all other public ClientRequest methods
    	[
    	  "flushHeaders", "getHeader",
    	  "setNoDelay", "setSocketKeepAlive",
    	].forEach(function (method) {
    	  RedirectableRequest.prototype[method] = function (a, b) {
    	    return this._currentRequest[method](a, b);
    	  };
    	});

    	// Proxy all public ClientRequest properties
    	["aborted", "connection", "socket"].forEach(function (property) {
    	  Object.defineProperty(RedirectableRequest.prototype, property, {
    	    get: function () { return this._currentRequest[property]; },
    	  });
    	});

    	RedirectableRequest.prototype._sanitizeOptions = function (options) {
    	  // Ensure headers are always present
    	  if (!options.headers) {
    	    options.headers = {};
    	  }

    	  // Since http.request treats host as an alias of hostname,
    	  // but the url module interprets host as hostname plus port,
    	  // eliminate the host property to avoid confusion.
    	  if (options.host) {
    	    // Use hostname if set, because it has precedence
    	    if (!options.hostname) {
    	      options.hostname = options.host;
    	    }
    	    delete options.host;
    	  }

    	  // Complete the URL object when necessary
    	  if (!options.pathname && options.path) {
    	    var searchPos = options.path.indexOf("?");
    	    if (searchPos < 0) {
    	      options.pathname = options.path;
    	    }
    	    else {
    	      options.pathname = options.path.substring(0, searchPos);
    	      options.search = options.path.substring(searchPos);
    	    }
    	  }
    	};


    	// Executes the next native request (initial or redirect)
    	RedirectableRequest.prototype._performRequest = function () {
    	  // Load the native protocol
    	  var protocol = this._options.protocol;
    	  var nativeProtocol = this._options.nativeProtocols[protocol];
    	  if (!nativeProtocol) {
    	    this.emit("error", new TypeError("Unsupported protocol " + protocol));
    	    return;
    	  }

    	  // If specified, use the agent corresponding to the protocol
    	  // (HTTP and HTTPS use different types of agents)
    	  if (this._options.agents) {
    	    var scheme = protocol.slice(0, -1);
    	    this._options.agent = this._options.agents[scheme];
    	  }

    	  // Create the native request
    	  var request = this._currentRequest =
    	        nativeProtocol.request(this._options, this._onNativeResponse);
    	  this._currentUrl = url.format(this._options);

    	  // Set up event handlers
    	  request._redirectable = this;
    	  for (var e = 0; e < events.length; e++) {
    	    request.on(events[e], eventHandlers[events[e]]);
    	  }

    	  // End a redirected request
    	  // (The first request must be ended explicitly with RedirectableRequest#end)
    	  if (this._isRedirect) {
    	    // Write the request entity and end.
    	    var i = 0;
    	    var self = this;
    	    var buffers = this._requestBodyBuffers;
    	    (function writeNext(error) {
    	      // Only write if this request has not been redirected yet
    	      /* istanbul ignore else */
    	      if (request === self._currentRequest) {
    	        // Report any write errors
    	        /* istanbul ignore if */
    	        if (error) {
    	          self.emit("error", error);
    	        }
    	        // Write the next buffer if there are still left
    	        else if (i < buffers.length) {
    	          var buffer = buffers[i++];
    	          /* istanbul ignore else */
    	          if (!request.finished) {
    	            request.write(buffer.data, buffer.encoding, writeNext);
    	          }
    	        }
    	        // End the request if `end` has been called on us
    	        else if (self._ended) {
    	          request.end();
    	        }
    	      }
    	    }());
    	  }
    	};

    	// Processes a response from the current native request
    	RedirectableRequest.prototype._processResponse = function (response) {
    	  // Store the redirected response
    	  var statusCode = response.statusCode;
    	  if (this._options.trackRedirects) {
    	    this._redirects.push({
    	      url: this._currentUrl,
    	      headers: response.headers,
    	      statusCode: statusCode,
    	    });
    	  }

    	  // RFC7231§6.4: The 3xx (Redirection) class of status code indicates
    	  // that further action needs to be taken by the user agent in order to
    	  // fulfill the request. If a Location header field is provided,
    	  // the user agent MAY automatically redirect its request to the URI
    	  // referenced by the Location field value,
    	  // even if the specific status code is not understood.

    	  // If the response is not a redirect; return it as-is
    	  var location = response.headers.location;
    	  if (!location || this._options.followRedirects === false ||
    	      statusCode < 300 || statusCode >= 400) {
    	    response.responseUrl = this._currentUrl;
    	    response.redirects = this._redirects;
    	    this.emit("response", response);

    	    // Clean up
    	    this._requestBodyBuffers = [];
    	    return;
    	  }

    	  // The response is a redirect, so abort the current request
    	  abortRequest(this._currentRequest);
    	  // Discard the remainder of the response to avoid waiting for data
    	  response.destroy();

    	  // RFC7231§6.4: A client SHOULD detect and intervene
    	  // in cyclical redirections (i.e., "infinite" redirection loops).
    	  if (++this._redirectCount > this._options.maxRedirects) {
    	    this.emit("error", new TooManyRedirectsError());
    	    return;
    	  }

    	  // Store the request headers if applicable
    	  var requestHeaders;
    	  var beforeRedirect = this._options.beforeRedirect;
    	  if (beforeRedirect) {
    	    requestHeaders = Object.assign({
    	      // The Host header was set by nativeProtocol.request
    	      Host: response.req.getHeader("host"),
    	    }, this._options.headers);
    	  }

    	  // RFC7231§6.4: Automatic redirection needs to done with
    	  // care for methods not known to be safe, […]
    	  // RFC7231§6.4.2–3: For historical reasons, a user agent MAY change
    	  // the request method from POST to GET for the subsequent request.
    	  var method = this._options.method;
    	  if ((statusCode === 301 || statusCode === 302) && this._options.method === "POST" ||
    	      // RFC7231§6.4.4: The 303 (See Other) status code indicates that
    	      // the server is redirecting the user agent to a different resource […]
    	      // A user agent can perform a retrieval request targeting that URI
    	      // (a GET or HEAD request if using HTTP) […]
    	      (statusCode === 303) && !/^(?:GET|HEAD)$/.test(this._options.method)) {
    	    this._options.method = "GET";
    	    // Drop a possible entity and headers related to it
    	    this._requestBodyBuffers = [];
    	    removeMatchingHeaders(/^content-/i, this._options.headers);
    	  }

    	  // Drop the Host header, as the redirect might lead to a different host
    	  var currentHostHeader = removeMatchingHeaders(/^host$/i, this._options.headers);

    	  // If the redirect is relative, carry over the host of the last request
    	  var currentUrlParts = url.parse(this._currentUrl);
    	  var currentHost = currentHostHeader || currentUrlParts.host;
    	  var currentUrl = /^\w+:/.test(location) ? this._currentUrl :
    	    url.format(Object.assign(currentUrlParts, { host: currentHost }));

    	  // Determine the URL of the redirection
    	  var redirectUrl;
    	  try {
    	    redirectUrl = url.resolve(currentUrl, location);
    	  }
    	  catch (cause) {
    	    this.emit("error", new RedirectionError(cause));
    	    return;
    	  }

    	  // Create the redirected request
    	  debug("redirecting to", redirectUrl);
    	  this._isRedirect = true;
    	  var redirectUrlParts = url.parse(redirectUrl);
    	  Object.assign(this._options, redirectUrlParts);

    	  // Drop confidential headers when redirecting to a less secure protocol
    	  // or to a different domain that is not a superdomain
    	  if (redirectUrlParts.protocol !== currentUrlParts.protocol &&
    	     redirectUrlParts.protocol !== "https:" ||
    	     redirectUrlParts.host !== currentHost &&
    	     !isSubdomain(redirectUrlParts.host, currentHost)) {
    	    removeMatchingHeaders(/^(?:authorization|cookie)$/i, this._options.headers);
    	  }

    	  // Evaluate the beforeRedirect callback
    	  if (typeof beforeRedirect === "function") {
    	    var responseDetails = {
    	      headers: response.headers,
    	      statusCode: statusCode,
    	    };
    	    var requestDetails = {
    	      url: currentUrl,
    	      method: method,
    	      headers: requestHeaders,
    	    };
    	    try {
    	      beforeRedirect(this._options, responseDetails, requestDetails);
    	    }
    	    catch (err) {
    	      this.emit("error", err);
    	      return;
    	    }
    	    this._sanitizeOptions(this._options);
    	  }

    	  // Perform the redirected request
    	  try {
    	    this._performRequest();
    	  }
    	  catch (cause) {
    	    this.emit("error", new RedirectionError(cause));
    	  }
    	};

    	// Wraps the key/value object of protocols with redirect functionality
    	function wrap(protocols) {
    	  // Default settings
    	  var exports = {
    	    maxRedirects: 21,
    	    maxBodyLength: 10 * 1024 * 1024,
    	  };

    	  // Wrap each protocol
    	  var nativeProtocols = {};
    	  Object.keys(protocols).forEach(function (scheme) {
    	    var protocol = scheme + ":";
    	    var nativeProtocol = nativeProtocols[protocol] = protocols[scheme];
    	    var wrappedProtocol = exports[scheme] = Object.create(nativeProtocol);

    	    // Executes a request, following redirects
    	    function request(input, options, callback) {
    	      // Parse parameters
    	      if (typeof input === "string") {
    	        var urlStr = input;
    	        try {
    	          input = urlToOptions(new URL(urlStr));
    	        }
    	        catch (err) {
    	          /* istanbul ignore next */
    	          input = url.parse(urlStr);
    	        }
    	      }
    	      else if (URL && (input instanceof URL)) {
    	        input = urlToOptions(input);
    	      }
    	      else {
    	        callback = options;
    	        options = input;
    	        input = { protocol: protocol };
    	      }
    	      if (typeof options === "function") {
    	        callback = options;
    	        options = null;
    	      }

    	      // Set defaults
    	      options = Object.assign({
    	        maxRedirects: exports.maxRedirects,
    	        maxBodyLength: exports.maxBodyLength,
    	      }, input, options);
    	      options.nativeProtocols = nativeProtocols;

    	      assert.equal(options.protocol, protocol, "protocol mismatch");
    	      debug("options", options);
    	      return new RedirectableRequest(options, callback);
    	    }

    	    // Executes a GET request, following redirects
    	    function get(input, options, callback) {
    	      var wrappedRequest = wrappedProtocol.request(input, options, callback);
    	      wrappedRequest.end();
    	      return wrappedRequest;
    	    }

    	    // Expose the properties on the wrapped protocol
    	    Object.defineProperties(wrappedProtocol, {
    	      request: { value: request, configurable: true, enumerable: true, writable: true },
    	      get: { value: get, configurable: true, enumerable: true, writable: true },
    	    });
    	  });
    	  return exports;
    	}

    	/* istanbul ignore next */
    	function noop() { /* empty */ }

    	// from https://github.com/nodejs/node/blob/master/lib/internal/url.js
    	function urlToOptions(urlObject) {
    	  var options = {
    	    protocol: urlObject.protocol,
    	    hostname: urlObject.hostname.startsWith("[") ?
    	      /* istanbul ignore next */
    	      urlObject.hostname.slice(1, -1) :
    	      urlObject.hostname,
    	    hash: urlObject.hash,
    	    search: urlObject.search,
    	    pathname: urlObject.pathname,
    	    path: urlObject.pathname + urlObject.search,
    	    href: urlObject.href,
    	  };
    	  if (urlObject.port !== "") {
    	    options.port = Number(urlObject.port);
    	  }
    	  return options;
    	}

    	function removeMatchingHeaders(regex, headers) {
    	  var lastValue;
    	  for (var header in headers) {
    	    if (regex.test(header)) {
    	      lastValue = headers[header];
    	      delete headers[header];
    	    }
    	  }
    	  return (lastValue === null || typeof lastValue === "undefined") ?
    	    undefined : String(lastValue).trim();
    	}

    	function createErrorType(code, defaultMessage) {
    	  function CustomError(cause) {
    	    Error.captureStackTrace(this, this.constructor);
    	    if (!cause) {
    	      this.message = defaultMessage;
    	    }
    	    else {
    	      this.message = defaultMessage + ": " + cause.message;
    	      this.cause = cause;
    	    }
    	  }
    	  CustomError.prototype = new Error();
    	  CustomError.prototype.constructor = CustomError;
    	  CustomError.prototype.name = "Error [" + code + "]";
    	  CustomError.prototype.code = code;
    	  return CustomError;
    	}

    	function abortRequest(request) {
    	  for (var e = 0; e < events.length; e++) {
    	    request.removeListener(events[e], eventHandlers[events[e]]);
    	  }
    	  request.on("error", noop);
    	  request.abort();
    	}

    	function isSubdomain(subdomain, domain) {
    	  const dot = subdomain.length - domain.length - 1;
    	  return dot > 0 && subdomain[dot] === "." && subdomain.endsWith(domain);
    	}

    	// Exports
    	followRedirects.exports = wrap({ http: http, https: https });
    	followRedirects.exports.wrap = wrap;
    	return followRedirects.exports;
    }

    var msg = {
      2:      'need dictionary',     /* Z_NEED_DICT       2  */
      1:      'stream end',          /* Z_STREAM_END      1  */
      0:      '',                    /* Z_OK              0  */
      '-1':   'file error',          /* Z_ERRNO         (-1) */
      '-2':   'stream error',        /* Z_STREAM_ERROR  (-2) */
      '-3':   'data error',          /* Z_DATA_ERROR    (-3) */
      '-4':   'insufficient memory', /* Z_MEM_ERROR     (-4) */
      '-5':   'buffer error',        /* Z_BUF_ERROR     (-5) */
      '-6':   'incompatible version' /* Z_VERSION_ERROR (-6) */
    };

    function ZStream() {
      /* next input byte */
      this.input = null; // JS specific, because we have no pointers
      this.next_in = 0;
      /* number of bytes available at input */
      this.avail_in = 0;
      /* total number of input bytes read so far */
      this.total_in = 0;
      /* next output byte should be put there */
      this.output = null; // JS specific, because we have no pointers
      this.next_out = 0;
      /* remaining free space at output */
      this.avail_out = 0;
      /* total number of bytes output so far */
      this.total_out = 0;
      /* last error message, NULL if no error */
      this.msg = ''/*Z_NULL*/;
      /* not visible by applications */
      this.state = null;
      /* best guess about the data type: binary or text */
      this.data_type = 2/*Z_UNKNOWN*/;
      /* adler32 value of the uncompressed data */
      this.adler = 0;
    }

    function arraySet(dest, src, src_offs, len, dest_offs) {
      if (src.subarray && dest.subarray) {
        dest.set(src.subarray(src_offs, src_offs + len), dest_offs);
        return;
      }
      // Fallback to ordinary array
      for (var i = 0; i < len; i++) {
        dest[dest_offs + i] = src[src_offs + i];
      }
    }


    var Buf8 = Uint8Array;
    var Buf16 = Uint16Array;
    var Buf32 = Int32Array;
    // Enable/Disable typed arrays use, for testing
    //

    /* Public constants ==========================================================*/
    /* ===========================================================================*/


    //var Z_FILTERED          = 1;
    //var Z_HUFFMAN_ONLY      = 2;
    //var Z_RLE               = 3;
    var Z_FIXED$2 = 4;
    //var Z_DEFAULT_STRATEGY  = 0;

    /* Possible values of the data_type field (though see inflate()) */
    var Z_BINARY$1 = 0;
    var Z_TEXT$1 = 1;
    //var Z_ASCII             = 1; // = Z_TEXT
    var Z_UNKNOWN$2 = 2;

    /*============================================================================*/


    function zero$1(buf) {
      var len = buf.length;
      while (--len >= 0) {
        buf[len] = 0;
      }
    }

    // From zutil.h

    var STORED_BLOCK = 0;
    var STATIC_TREES = 1;
    var DYN_TREES = 2;
    /* The three kinds of block type */

    var MIN_MATCH$1 = 3;
    var MAX_MATCH$1 = 258;
    /* The minimum and maximum match lengths */

    // From deflate.h
    /* ===========================================================================
     * Internal compression state.
     */

    var LENGTH_CODES$1 = 29;
    /* number of length codes, not counting the special END_BLOCK code */

    var LITERALS$1 = 256;
    /* number of literal bytes 0..255 */

    var L_CODES$1 = LITERALS$1 + 1 + LENGTH_CODES$1;
    /* number of Literal or Length codes, including the END_BLOCK code */

    var D_CODES$1 = 30;
    /* number of distance codes */

    var BL_CODES$1 = 19;
    /* number of codes used to transfer the bit lengths */

    var HEAP_SIZE$1 = 2 * L_CODES$1 + 1;
    /* maximum heap size */

    var MAX_BITS$1 = 15;
    /* All codes must not exceed MAX_BITS bits */

    var Buf_size = 16;
    /* size of bit buffer in bi_buf */


    /* ===========================================================================
     * Constants
     */

    var MAX_BL_BITS = 7;
    /* Bit length codes must not exceed MAX_BL_BITS bits */

    var END_BLOCK = 256;
    /* end of block literal code */

    var REP_3_6 = 16;
    /* repeat previous bit length 3-6 times (2 bits of repeat count) */

    var REPZ_3_10 = 17;
    /* repeat a zero length 3-10 times  (3 bits of repeat count) */

    var REPZ_11_138 = 18;
    /* repeat a zero length 11-138 times  (7 bits of repeat count) */

    /* eslint-disable comma-spacing,array-bracket-spacing */
    var extra_lbits = /* extra bits for each length code */ [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0];

    var extra_dbits = /* extra bits for each distance code */ [0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13];

    var extra_blbits = /* extra bits for each bit length code */ [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 7];

    var bl_order = [16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15];
    /* eslint-enable comma-spacing,array-bracket-spacing */

    /* The lengths of the bit length codes are sent in order of decreasing
     * probability, to avoid transmitting the lengths for unused bit length codes.
     */

    /* ===========================================================================
     * Local data. These are initialized only once.
     */

    // We pre-fill arrays with 0 to avoid uninitialized gaps

    var DIST_CODE_LEN = 512; /* see definition of array dist_code below */

    // !!!! Use flat array insdead of structure, Freq = i*2, Len = i*2+1
    var static_ltree = new Array((L_CODES$1 + 2) * 2);
    zero$1(static_ltree);
    /* The static literal tree. Since the bit lengths are imposed, there is no
     * need for the L_CODES extra codes used during heap construction. However
     * The codes 286 and 287 are needed to build a canonical tree (see _tr_init
     * below).
     */

    var static_dtree = new Array(D_CODES$1 * 2);
    zero$1(static_dtree);
    /* The static distance tree. (Actually a trivial tree since all codes use
     * 5 bits.)
     */

    var _dist_code = new Array(DIST_CODE_LEN);
    zero$1(_dist_code);
    /* Distance codes. The first 256 values correspond to the distances
     * 3 .. 258, the last 256 values correspond to the top 8 bits of
     * the 15 bit distances.
     */

    var _length_code = new Array(MAX_MATCH$1 - MIN_MATCH$1 + 1);
    zero$1(_length_code);
    /* length code for each normalized match length (0 == MIN_MATCH) */

    var base_length = new Array(LENGTH_CODES$1);
    zero$1(base_length);
    /* First normalized length for each code (0 = MIN_MATCH) */

    var base_dist = new Array(D_CODES$1);
    zero$1(base_dist);
    /* First normalized distance for each code (0 = distance of 1) */


    function StaticTreeDesc(static_tree, extra_bits, extra_base, elems, max_length) {

      this.static_tree = static_tree; /* static tree or NULL */
      this.extra_bits = extra_bits; /* extra bits for each code or NULL */
      this.extra_base = extra_base; /* base index for extra_bits */
      this.elems = elems; /* max number of elements in the tree */
      this.max_length = max_length; /* max bit length for the codes */

      // show if `static_tree` has data or dummy - needed for monomorphic objects
      this.has_stree = static_tree && static_tree.length;
    }


    var static_l_desc;
    var static_d_desc;
    var static_bl_desc;


    function TreeDesc(dyn_tree, stat_desc) {
      this.dyn_tree = dyn_tree; /* the dynamic tree */
      this.max_code = 0; /* largest code with non zero frequency */
      this.stat_desc = stat_desc; /* the corresponding static tree */
    }



    function d_code(dist) {
      return dist < 256 ? _dist_code[dist] : _dist_code[256 + (dist >>> 7)];
    }


    /* ===========================================================================
     * Output a short LSB first on the stream.
     * IN assertion: there is enough room in pendingBuf.
     */
    function put_short(s, w) {
      //    put_byte(s, (uch)((w) & 0xff));
      //    put_byte(s, (uch)((ush)(w) >> 8));
      s.pending_buf[s.pending++] = (w) & 0xff;
      s.pending_buf[s.pending++] = (w >>> 8) & 0xff;
    }


    /* ===========================================================================
     * Send a value on a given number of bits.
     * IN assertion: length <= 16 and value fits in length bits.
     */
    function send_bits(s, value, length) {
      if (s.bi_valid > (Buf_size - length)) {
        s.bi_buf |= (value << s.bi_valid) & 0xffff;
        put_short(s, s.bi_buf);
        s.bi_buf = value >> (Buf_size - s.bi_valid);
        s.bi_valid += length - Buf_size;
      } else {
        s.bi_buf |= (value << s.bi_valid) & 0xffff;
        s.bi_valid += length;
      }
    }


    function send_code(s, c, tree) {
      send_bits(s, tree[c * 2] /*.Code*/ , tree[c * 2 + 1] /*.Len*/ );
    }


    /* ===========================================================================
     * Reverse the first len bits of a code, using straightforward code (a faster
     * method would use a table)
     * IN assertion: 1 <= len <= 15
     */
    function bi_reverse(code, len) {
      var res = 0;
      do {
        res |= code & 1;
        code >>>= 1;
        res <<= 1;
      } while (--len > 0);
      return res >>> 1;
    }


    /* ===========================================================================
     * Flush the bit buffer, keeping at most 7 bits in it.
     */
    function bi_flush(s) {
      if (s.bi_valid === 16) {
        put_short(s, s.bi_buf);
        s.bi_buf = 0;
        s.bi_valid = 0;

      } else if (s.bi_valid >= 8) {
        s.pending_buf[s.pending++] = s.bi_buf & 0xff;
        s.bi_buf >>= 8;
        s.bi_valid -= 8;
      }
    }


    /* ===========================================================================
     * Compute the optimal bit lengths for a tree and update the total bit length
     * for the current block.
     * IN assertion: the fields freq and dad are set, heap[heap_max] and
     *    above are the tree nodes sorted by increasing frequency.
     * OUT assertions: the field len is set to the optimal bit length, the
     *     array bl_count contains the frequencies for each bit length.
     *     The length opt_len is updated; static_len is also updated if stree is
     *     not null.
     */
    function gen_bitlen(s, desc) {
    //    deflate_state *s;
    //    tree_desc *desc;    /* the tree descriptor */
      var tree = desc.dyn_tree;
      var max_code = desc.max_code;
      var stree = desc.stat_desc.static_tree;
      var has_stree = desc.stat_desc.has_stree;
      var extra = desc.stat_desc.extra_bits;
      var base = desc.stat_desc.extra_base;
      var max_length = desc.stat_desc.max_length;
      var h; /* heap index */
      var n, m; /* iterate over the tree elements */
      var bits; /* bit length */
      var xbits; /* extra bits */
      var f; /* frequency */
      var overflow = 0; /* number of elements with bit length too large */

      for (bits = 0; bits <= MAX_BITS$1; bits++) {
        s.bl_count[bits] = 0;
      }

      /* In a first pass, compute the optimal bit lengths (which may
       * overflow in the case of the bit length tree).
       */
      tree[s.heap[s.heap_max] * 2 + 1] /*.Len*/ = 0; /* root of the heap */

      for (h = s.heap_max + 1; h < HEAP_SIZE$1; h++) {
        n = s.heap[h];
        bits = tree[tree[n * 2 + 1] /*.Dad*/ * 2 + 1] /*.Len*/ + 1;
        if (bits > max_length) {
          bits = max_length;
          overflow++;
        }
        tree[n * 2 + 1] /*.Len*/ = bits;
        /* We overwrite tree[n].Dad which is no longer needed */

        if (n > max_code) {
          continue;
        } /* not a leaf node */

        s.bl_count[bits]++;
        xbits = 0;
        if (n >= base) {
          xbits = extra[n - base];
        }
        f = tree[n * 2] /*.Freq*/ ;
        s.opt_len += f * (bits + xbits);
        if (has_stree) {
          s.static_len += f * (stree[n * 2 + 1] /*.Len*/ + xbits);
        }
      }
      if (overflow === 0) {
        return;
      }

      // Trace((stderr,"\nbit length overflow\n"));
      /* This happens for example on obj2 and pic of the Calgary corpus */

      /* Find the first bit length which could increase: */
      do {
        bits = max_length - 1;
        while (s.bl_count[bits] === 0) {
          bits--;
        }
        s.bl_count[bits]--; /* move one leaf down the tree */
        s.bl_count[bits + 1] += 2; /* move one overflow item as its brother */
        s.bl_count[max_length]--;
        /* The brother of the overflow item also moves one step up,
         * but this does not affect bl_count[max_length]
         */
        overflow -= 2;
      } while (overflow > 0);

      /* Now recompute all bit lengths, scanning in increasing frequency.
       * h is still equal to HEAP_SIZE. (It is simpler to reconstruct all
       * lengths instead of fixing only the wrong ones. This idea is taken
       * from 'ar' written by Haruhiko Okumura.)
       */
      for (bits = max_length; bits !== 0; bits--) {
        n = s.bl_count[bits];
        while (n !== 0) {
          m = s.heap[--h];
          if (m > max_code) {
            continue;
          }
          if (tree[m * 2 + 1] /*.Len*/ !== bits) {
            // Trace((stderr,"code %d bits %d->%d\n", m, tree[m].Len, bits));
            s.opt_len += (bits - tree[m * 2 + 1] /*.Len*/ ) * tree[m * 2] /*.Freq*/ ;
            tree[m * 2 + 1] /*.Len*/ = bits;
          }
          n--;
        }
      }
    }


    /* ===========================================================================
     * Generate the codes for a given tree and bit counts (which need not be
     * optimal).
     * IN assertion: the array bl_count contains the bit length statistics for
     * the given tree and the field len is set for all tree elements.
     * OUT assertion: the field code is set for all tree elements of non
     *     zero code length.
     */
    function gen_codes(tree, max_code, bl_count) {
    //    ct_data *tree;             /* the tree to decorate */
    //    int max_code;              /* largest code with non zero frequency */
    //    ushf *bl_count;            /* number of codes at each bit length */

      var next_code = new Array(MAX_BITS$1 + 1); /* next code value for each bit length */
      var code = 0; /* running code value */
      var bits; /* bit index */
      var n; /* code index */

      /* The distribution counts are first used to generate the code values
       * without bit reversal.
       */
      for (bits = 1; bits <= MAX_BITS$1; bits++) {
        next_code[bits] = code = (code + bl_count[bits - 1]) << 1;
      }
      /* Check that the bit counts in bl_count are consistent. The last code
       * must be all ones.
       */
      //Assert (code + bl_count[MAX_BITS]-1 == (1<<MAX_BITS)-1,
      //        "inconsistent bit counts");
      //Tracev((stderr,"\ngen_codes: max_code %d ", max_code));

      for (n = 0; n <= max_code; n++) {
        var len = tree[n * 2 + 1] /*.Len*/ ;
        if (len === 0) {
          continue;
        }
        /* Now reverse the bits */
        tree[n * 2] /*.Code*/ = bi_reverse(next_code[len]++, len);

        //Tracecv(tree != static_ltree, (stderr,"\nn %3d %c l %2d c %4x (%x) ",
        //     n, (isgraph(n) ? n : ' '), len, tree[n].Code, next_code[len]-1));
      }
    }


    /* ===========================================================================
     * Initialize the various 'constant' tables.
     */
    function tr_static_init() {
      var n; /* iterates over tree elements */
      var bits; /* bit counter */
      var length; /* length value */
      var code; /* code value */
      var dist; /* distance index */
      var bl_count = new Array(MAX_BITS$1 + 1);
      /* number of codes at each bit length for an optimal tree */

      // do check in _tr_init()
      //if (static_init_done) return;

      /* For some embedded targets, global variables are not initialized: */
      /*#ifdef NO_INIT_GLOBAL_POINTERS
        static_l_desc.static_tree = static_ltree;
        static_l_desc.extra_bits = extra_lbits;
        static_d_desc.static_tree = static_dtree;
        static_d_desc.extra_bits = extra_dbits;
        static_bl_desc.extra_bits = extra_blbits;
      #endif*/

      /* Initialize the mapping length (0..255) -> length code (0..28) */
      length = 0;
      for (code = 0; code < LENGTH_CODES$1 - 1; code++) {
        base_length[code] = length;
        for (n = 0; n < (1 << extra_lbits[code]); n++) {
          _length_code[length++] = code;
        }
      }
      //Assert (length == 256, "tr_static_init: length != 256");
      /* Note that the length 255 (match length 258) can be represented
       * in two different ways: code 284 + 5 bits or code 285, so we
       * overwrite length_code[255] to use the best encoding:
       */
      _length_code[length - 1] = code;

      /* Initialize the mapping dist (0..32K) -> dist code (0..29) */
      dist = 0;
      for (code = 0; code < 16; code++) {
        base_dist[code] = dist;
        for (n = 0; n < (1 << extra_dbits[code]); n++) {
          _dist_code[dist++] = code;
        }
      }
      //Assert (dist == 256, "tr_static_init: dist != 256");
      dist >>= 7; /* from now on, all distances are divided by 128 */
      for (; code < D_CODES$1; code++) {
        base_dist[code] = dist << 7;
        for (n = 0; n < (1 << (extra_dbits[code] - 7)); n++) {
          _dist_code[256 + dist++] = code;
        }
      }
      //Assert (dist == 256, "tr_static_init: 256+dist != 512");

      /* Construct the codes of the static literal tree */
      for (bits = 0; bits <= MAX_BITS$1; bits++) {
        bl_count[bits] = 0;
      }

      n = 0;
      while (n <= 143) {
        static_ltree[n * 2 + 1] /*.Len*/ = 8;
        n++;
        bl_count[8]++;
      }
      while (n <= 255) {
        static_ltree[n * 2 + 1] /*.Len*/ = 9;
        n++;
        bl_count[9]++;
      }
      while (n <= 279) {
        static_ltree[n * 2 + 1] /*.Len*/ = 7;
        n++;
        bl_count[7]++;
      }
      while (n <= 287) {
        static_ltree[n * 2 + 1] /*.Len*/ = 8;
        n++;
        bl_count[8]++;
      }
      /* Codes 286 and 287 do not exist, but we must include them in the
       * tree construction to get a canonical Huffman tree (longest code
       * all ones)
       */
      gen_codes(static_ltree, L_CODES$1 + 1, bl_count);

      /* The static distance tree is trivial: */
      for (n = 0; n < D_CODES$1; n++) {
        static_dtree[n * 2 + 1] /*.Len*/ = 5;
        static_dtree[n * 2] /*.Code*/ = bi_reverse(n, 5);
      }

      // Now data ready and we can init static trees
      static_l_desc = new StaticTreeDesc(static_ltree, extra_lbits, LITERALS$1 + 1, L_CODES$1, MAX_BITS$1);
      static_d_desc = new StaticTreeDesc(static_dtree, extra_dbits, 0, D_CODES$1, MAX_BITS$1);
      static_bl_desc = new StaticTreeDesc(new Array(0), extra_blbits, 0, BL_CODES$1, MAX_BL_BITS);

      //static_init_done = true;
    }


    /* ===========================================================================
     * Initialize a new block.
     */
    function init_block(s) {
      var n; /* iterates over tree elements */

      /* Initialize the trees. */
      for (n = 0; n < L_CODES$1; n++) {
        s.dyn_ltree[n * 2] /*.Freq*/ = 0;
      }
      for (n = 0; n < D_CODES$1; n++) {
        s.dyn_dtree[n * 2] /*.Freq*/ = 0;
      }
      for (n = 0; n < BL_CODES$1; n++) {
        s.bl_tree[n * 2] /*.Freq*/ = 0;
      }

      s.dyn_ltree[END_BLOCK * 2] /*.Freq*/ = 1;
      s.opt_len = s.static_len = 0;
      s.last_lit = s.matches = 0;
    }


    /* ===========================================================================
     * Flush the bit buffer and align the output on a byte boundary
     */
    function bi_windup(s) {
      if (s.bi_valid > 8) {
        put_short(s, s.bi_buf);
      } else if (s.bi_valid > 0) {
        //put_byte(s, (Byte)s->bi_buf);
        s.pending_buf[s.pending++] = s.bi_buf;
      }
      s.bi_buf = 0;
      s.bi_valid = 0;
    }

    /* ===========================================================================
     * Copy a stored block, storing first the length and its
     * one's complement if requested.
     */
    function copy_block(s, buf, len, header) {
    //DeflateState *s;
    //charf    *buf;    /* the input data */
    //unsigned len;     /* its length */
    //int      header;  /* true if block header must be written */

      bi_windup(s); /* align on byte boundary */

      if (header) {
        put_short(s, len);
        put_short(s, ~len);
      }
      //  while (len--) {
      //    put_byte(s, *buf++);
      //  }
      arraySet(s.pending_buf, s.window, buf, len, s.pending);
      s.pending += len;
    }

    /* ===========================================================================
     * Compares to subtrees, using the tree depth as tie breaker when
     * the subtrees have equal frequency. This minimizes the worst case length.
     */
    function smaller(tree, n, m, depth) {
      var _n2 = n * 2;
      var _m2 = m * 2;
      return (tree[_n2] /*.Freq*/ < tree[_m2] /*.Freq*/ ||
        (tree[_n2] /*.Freq*/ === tree[_m2] /*.Freq*/ && depth[n] <= depth[m]));
    }

    /* ===========================================================================
     * Restore the heap property by moving down the tree starting at node k,
     * exchanging a node with the smallest of its two sons if necessary, stopping
     * when the heap property is re-established (each father smaller than its
     * two sons).
     */
    function pqdownheap(s, tree, k)
    //    deflate_state *s;
    //    ct_data *tree;  /* the tree to restore */
    //    int k;               /* node to move down */
    {
      var v = s.heap[k];
      var j = k << 1; /* left son of k */
      while (j <= s.heap_len) {
        /* Set j to the smallest of the two sons: */
        if (j < s.heap_len &&
          smaller(tree, s.heap[j + 1], s.heap[j], s.depth)) {
          j++;
        }
        /* Exit if v is smaller than both sons */
        if (smaller(tree, v, s.heap[j], s.depth)) {
          break;
        }

        /* Exchange v with the smallest son */
        s.heap[k] = s.heap[j];
        k = j;

        /* And continue down the tree, setting j to the left son of k */
        j <<= 1;
      }
      s.heap[k] = v;
    }


    // inlined manually
    // var SMALLEST = 1;

    /* ===========================================================================
     * Send the block data compressed using the given Huffman trees
     */
    function compress_block(s, ltree, dtree)
    //    deflate_state *s;
    //    const ct_data *ltree; /* literal tree */
    //    const ct_data *dtree; /* distance tree */
    {
      var dist; /* distance of matched string */
      var lc; /* match length or unmatched char (if dist == 0) */
      var lx = 0; /* running index in l_buf */
      var code; /* the code to send */
      var extra; /* number of extra bits to send */

      if (s.last_lit !== 0) {
        do {
          dist = (s.pending_buf[s.d_buf + lx * 2] << 8) | (s.pending_buf[s.d_buf + lx * 2 + 1]);
          lc = s.pending_buf[s.l_buf + lx];
          lx++;

          if (dist === 0) {
            send_code(s, lc, ltree); /* send a literal byte */
            //Tracecv(isgraph(lc), (stderr," '%c' ", lc));
          } else {
            /* Here, lc is the match length - MIN_MATCH */
            code = _length_code[lc];
            send_code(s, code + LITERALS$1 + 1, ltree); /* send the length code */
            extra = extra_lbits[code];
            if (extra !== 0) {
              lc -= base_length[code];
              send_bits(s, lc, extra); /* send the extra length bits */
            }
            dist--; /* dist is now the match distance - 1 */
            code = d_code(dist);
            //Assert (code < D_CODES, "bad d_code");

            send_code(s, code, dtree); /* send the distance code */
            extra = extra_dbits[code];
            if (extra !== 0) {
              dist -= base_dist[code];
              send_bits(s, dist, extra); /* send the extra distance bits */
            }
          } /* literal or match pair ? */

          /* Check that the overlay between pending_buf and d_buf+l_buf is ok: */
          //Assert((uInt)(s->pending) < s->lit_bufsize + 2*lx,
          //       "pendingBuf overflow");

        } while (lx < s.last_lit);
      }

      send_code(s, END_BLOCK, ltree);
    }


    /* ===========================================================================
     * Construct one Huffman tree and assigns the code bit strings and lengths.
     * Update the total bit length for the current block.
     * IN assertion: the field freq is set for all tree elements.
     * OUT assertions: the fields len and code are set to the optimal bit length
     *     and corresponding code. The length opt_len is updated; static_len is
     *     also updated if stree is not null. The field max_code is set.
     */
    function build_tree(s, desc)
    //    deflate_state *s;
    //    tree_desc *desc; /* the tree descriptor */
    {
      var tree = desc.dyn_tree;
      var stree = desc.stat_desc.static_tree;
      var has_stree = desc.stat_desc.has_stree;
      var elems = desc.stat_desc.elems;
      var n, m; /* iterate over heap elements */
      var max_code = -1; /* largest code with non zero frequency */
      var node; /* new node being created */

      /* Construct the initial heap, with least frequent element in
       * heap[SMALLEST]. The sons of heap[n] are heap[2*n] and heap[2*n+1].
       * heap[0] is not used.
       */
      s.heap_len = 0;
      s.heap_max = HEAP_SIZE$1;

      for (n = 0; n < elems; n++) {
        if (tree[n * 2] /*.Freq*/ !== 0) {
          s.heap[++s.heap_len] = max_code = n;
          s.depth[n] = 0;

        } else {
          tree[n * 2 + 1] /*.Len*/ = 0;
        }
      }

      /* The pkzip format requires that at least one distance code exists,
       * and that at least one bit should be sent even if there is only one
       * possible code. So to avoid special checks later on we force at least
       * two codes of non zero frequency.
       */
      while (s.heap_len < 2) {
        node = s.heap[++s.heap_len] = (max_code < 2 ? ++max_code : 0);
        tree[node * 2] /*.Freq*/ = 1;
        s.depth[node] = 0;
        s.opt_len--;

        if (has_stree) {
          s.static_len -= stree[node * 2 + 1] /*.Len*/ ;
        }
        /* node is 0 or 1 so it does not have extra bits */
      }
      desc.max_code = max_code;

      /* The elements heap[heap_len/2+1 .. heap_len] are leaves of the tree,
       * establish sub-heaps of increasing lengths:
       */
      for (n = (s.heap_len >> 1 /*int /2*/ ); n >= 1; n--) {
        pqdownheap(s, tree, n);
      }

      /* Construct the Huffman tree by repeatedly combining the least two
       * frequent nodes.
       */
      node = elems; /* next internal node of the tree */
      do {
        //pqremove(s, tree, n);  /* n = node of least frequency */
        /*** pqremove ***/
        n = s.heap[1 /*SMALLEST*/ ];
        s.heap[1 /*SMALLEST*/ ] = s.heap[s.heap_len--];
        pqdownheap(s, tree, 1 /*SMALLEST*/ );
        /***/

        m = s.heap[1 /*SMALLEST*/ ]; /* m = node of next least frequency */

        s.heap[--s.heap_max] = n; /* keep the nodes sorted by frequency */
        s.heap[--s.heap_max] = m;

        /* Create a new node father of n and m */
        tree[node * 2] /*.Freq*/ = tree[n * 2] /*.Freq*/ + tree[m * 2] /*.Freq*/ ;
        s.depth[node] = (s.depth[n] >= s.depth[m] ? s.depth[n] : s.depth[m]) + 1;
        tree[n * 2 + 1] /*.Dad*/ = tree[m * 2 + 1] /*.Dad*/ = node;

        /* and insert the new node in the heap */
        s.heap[1 /*SMALLEST*/ ] = node++;
        pqdownheap(s, tree, 1 /*SMALLEST*/ );

      } while (s.heap_len >= 2);

      s.heap[--s.heap_max] = s.heap[1 /*SMALLEST*/ ];

      /* At this point, the fields freq and dad are set. We can now
       * generate the bit lengths.
       */
      gen_bitlen(s, desc);

      /* The field len is now set, we can generate the bit codes */
      gen_codes(tree, max_code, s.bl_count);
    }


    /* ===========================================================================
     * Scan a literal or distance tree to determine the frequencies of the codes
     * in the bit length tree.
     */
    function scan_tree(s, tree, max_code)
    //    deflate_state *s;
    //    ct_data *tree;   /* the tree to be scanned */
    //    int max_code;    /* and its largest code of non zero frequency */
    {
      var n; /* iterates over all tree elements */
      var prevlen = -1; /* last emitted length */
      var curlen; /* length of current code */

      var nextlen = tree[0 * 2 + 1] /*.Len*/ ; /* length of next code */

      var count = 0; /* repeat count of the current code */
      var max_count = 7; /* max repeat count */
      var min_count = 4; /* min repeat count */

      if (nextlen === 0) {
        max_count = 138;
        min_count = 3;
      }
      tree[(max_code + 1) * 2 + 1] /*.Len*/ = 0xffff; /* guard */

      for (n = 0; n <= max_code; n++) {
        curlen = nextlen;
        nextlen = tree[(n + 1) * 2 + 1] /*.Len*/ ;

        if (++count < max_count && curlen === nextlen) {
          continue;

        } else if (count < min_count) {
          s.bl_tree[curlen * 2] /*.Freq*/ += count;

        } else if (curlen !== 0) {

          if (curlen !== prevlen) {
            s.bl_tree[curlen * 2] /*.Freq*/ ++;
          }
          s.bl_tree[REP_3_6 * 2] /*.Freq*/ ++;

        } else if (count <= 10) {
          s.bl_tree[REPZ_3_10 * 2] /*.Freq*/ ++;

        } else {
          s.bl_tree[REPZ_11_138 * 2] /*.Freq*/ ++;
        }

        count = 0;
        prevlen = curlen;

        if (nextlen === 0) {
          max_count = 138;
          min_count = 3;

        } else if (curlen === nextlen) {
          max_count = 6;
          min_count = 3;

        } else {
          max_count = 7;
          min_count = 4;
        }
      }
    }


    /* ===========================================================================
     * Send a literal or distance tree in compressed form, using the codes in
     * bl_tree.
     */
    function send_tree(s, tree, max_code)
    //    deflate_state *s;
    //    ct_data *tree; /* the tree to be scanned */
    //    int max_code;       /* and its largest code of non zero frequency */
    {
      var n; /* iterates over all tree elements */
      var prevlen = -1; /* last emitted length */
      var curlen; /* length of current code */

      var nextlen = tree[0 * 2 + 1] /*.Len*/ ; /* length of next code */

      var count = 0; /* repeat count of the current code */
      var max_count = 7; /* max repeat count */
      var min_count = 4; /* min repeat count */

      /* tree[max_code+1].Len = -1; */
      /* guard already set */
      if (nextlen === 0) {
        max_count = 138;
        min_count = 3;
      }

      for (n = 0; n <= max_code; n++) {
        curlen = nextlen;
        nextlen = tree[(n + 1) * 2 + 1] /*.Len*/ ;

        if (++count < max_count && curlen === nextlen) {
          continue;

        } else if (count < min_count) {
          do {
            send_code(s, curlen, s.bl_tree);
          } while (--count !== 0);

        } else if (curlen !== 0) {
          if (curlen !== prevlen) {
            send_code(s, curlen, s.bl_tree);
            count--;
          }
          //Assert(count >= 3 && count <= 6, " 3_6?");
          send_code(s, REP_3_6, s.bl_tree);
          send_bits(s, count - 3, 2);

        } else if (count <= 10) {
          send_code(s, REPZ_3_10, s.bl_tree);
          send_bits(s, count - 3, 3);

        } else {
          send_code(s, REPZ_11_138, s.bl_tree);
          send_bits(s, count - 11, 7);
        }

        count = 0;
        prevlen = curlen;
        if (nextlen === 0) {
          max_count = 138;
          min_count = 3;

        } else if (curlen === nextlen) {
          max_count = 6;
          min_count = 3;

        } else {
          max_count = 7;
          min_count = 4;
        }
      }
    }


    /* ===========================================================================
     * Construct the Huffman tree for the bit lengths and return the index in
     * bl_order of the last bit length code to send.
     */
    function build_bl_tree(s) {
      var max_blindex; /* index of last bit length code of non zero freq */

      /* Determine the bit length frequencies for literal and distance trees */
      scan_tree(s, s.dyn_ltree, s.l_desc.max_code);
      scan_tree(s, s.dyn_dtree, s.d_desc.max_code);

      /* Build the bit length tree: */
      build_tree(s, s.bl_desc);
      /* opt_len now includes the length of the tree representations, except
       * the lengths of the bit lengths codes and the 5+5+4 bits for the counts.
       */

      /* Determine the number of bit length codes to send. The pkzip format
       * requires that at least 4 bit length codes be sent. (appnote.txt says
       * 3 but the actual value used is 4.)
       */
      for (max_blindex = BL_CODES$1 - 1; max_blindex >= 3; max_blindex--) {
        if (s.bl_tree[bl_order[max_blindex] * 2 + 1] /*.Len*/ !== 0) {
          break;
        }
      }
      /* Update opt_len to include the bit length tree and counts */
      s.opt_len += 3 * (max_blindex + 1) + 5 + 5 + 4;
      //Tracev((stderr, "\ndyn trees: dyn %ld, stat %ld",
      //        s->opt_len, s->static_len));

      return max_blindex;
    }


    /* ===========================================================================
     * Send the header for a block using dynamic Huffman trees: the counts, the
     * lengths of the bit length codes, the literal tree and the distance tree.
     * IN assertion: lcodes >= 257, dcodes >= 1, blcodes >= 4.
     */
    function send_all_trees(s, lcodes, dcodes, blcodes)
    //    deflate_state *s;
    //    int lcodes, dcodes, blcodes; /* number of codes for each tree */
    {
      var rank; /* index in bl_order */

      //Assert (lcodes >= 257 && dcodes >= 1 && blcodes >= 4, "not enough codes");
      //Assert (lcodes <= L_CODES && dcodes <= D_CODES && blcodes <= BL_CODES,
      //        "too many codes");
      //Tracev((stderr, "\nbl counts: "));
      send_bits(s, lcodes - 257, 5); /* not +255 as stated in appnote.txt */
      send_bits(s, dcodes - 1, 5);
      send_bits(s, blcodes - 4, 4); /* not -3 as stated in appnote.txt */
      for (rank = 0; rank < blcodes; rank++) {
        //Tracev((stderr, "\nbl code %2d ", bl_order[rank]));
        send_bits(s, s.bl_tree[bl_order[rank] * 2 + 1] /*.Len*/ , 3);
      }
      //Tracev((stderr, "\nbl tree: sent %ld", s->bits_sent));

      send_tree(s, s.dyn_ltree, lcodes - 1); /* literal tree */
      //Tracev((stderr, "\nlit tree: sent %ld", s->bits_sent));

      send_tree(s, s.dyn_dtree, dcodes - 1); /* distance tree */
      //Tracev((stderr, "\ndist tree: sent %ld", s->bits_sent));
    }


    /* ===========================================================================
     * Check if the data type is TEXT or BINARY, using the following algorithm:
     * - TEXT if the two conditions below are satisfied:
     *    a) There are no non-portable control characters belonging to the
     *       "black list" (0..6, 14..25, 28..31).
     *    b) There is at least one printable character belonging to the
     *       "white list" (9 {TAB}, 10 {LF}, 13 {CR}, 32..255).
     * - BINARY otherwise.
     * - The following partially-portable control characters form a
     *   "gray list" that is ignored in this detection algorithm:
     *   (7 {BEL}, 8 {BS}, 11 {VT}, 12 {FF}, 26 {SUB}, 27 {ESC}).
     * IN assertion: the fields Freq of dyn_ltree are set.
     */
    function detect_data_type(s) {
      /* black_mask is the bit mask of black-listed bytes
       * set bits 0..6, 14..25, and 28..31
       * 0xf3ffc07f = binary 11110011111111111100000001111111
       */
      var black_mask = 0xf3ffc07f;
      var n;

      /* Check for non-textual ("black-listed") bytes. */
      for (n = 0; n <= 31; n++, black_mask >>>= 1) {
        if ((black_mask & 1) && (s.dyn_ltree[n * 2] /*.Freq*/ !== 0)) {
          return Z_BINARY$1;
        }
      }

      /* Check for textual ("white-listed") bytes. */
      if (s.dyn_ltree[9 * 2] /*.Freq*/ !== 0 || s.dyn_ltree[10 * 2] /*.Freq*/ !== 0 ||
        s.dyn_ltree[13 * 2] /*.Freq*/ !== 0) {
        return Z_TEXT$1;
      }
      for (n = 32; n < LITERALS$1; n++) {
        if (s.dyn_ltree[n * 2] /*.Freq*/ !== 0) {
          return Z_TEXT$1;
        }
      }

      /* There are no "black-listed" or "white-listed" bytes:
       * this stream either is empty or has tolerated ("gray-listed") bytes only.
       */
      return Z_BINARY$1;
    }


    var static_init_done = false;

    /* ===========================================================================
     * Initialize the tree data structures for a new zlib stream.
     */
    function _tr_init(s) {

      if (!static_init_done) {
        tr_static_init();
        static_init_done = true;
      }

      s.l_desc = new TreeDesc(s.dyn_ltree, static_l_desc);
      s.d_desc = new TreeDesc(s.dyn_dtree, static_d_desc);
      s.bl_desc = new TreeDesc(s.bl_tree, static_bl_desc);

      s.bi_buf = 0;
      s.bi_valid = 0;

      /* Initialize the first block of the first file: */
      init_block(s);
    }


    /* ===========================================================================
     * Send a stored block
     */
    function _tr_stored_block(s, buf, stored_len, last)
    //DeflateState *s;
    //charf *buf;       /* input block */
    //ulg stored_len;   /* length of input block */
    //int last;         /* one if this is the last block for a file */
    {
      send_bits(s, (STORED_BLOCK << 1) + (last ? 1 : 0), 3); /* send block type */
      copy_block(s, buf, stored_len, true); /* with header */
    }


    /* ===========================================================================
     * Send one empty static block to give enough lookahead for inflate.
     * This takes 10 bits, of which 7 may remain in the bit buffer.
     */
    function _tr_align(s) {
      send_bits(s, STATIC_TREES << 1, 3);
      send_code(s, END_BLOCK, static_ltree);
      bi_flush(s);
    }


    /* ===========================================================================
     * Determine the best encoding for the current block: dynamic trees, static
     * trees or store, and output the encoded block to the zip file.
     */
    function _tr_flush_block(s, buf, stored_len, last)
    //DeflateState *s;
    //charf *buf;       /* input block, or NULL if too old */
    //ulg stored_len;   /* length of input block */
    //int last;         /* one if this is the last block for a file */
    {
      var opt_lenb, static_lenb; /* opt_len and static_len in bytes */
      var max_blindex = 0; /* index of last bit length code of non zero freq */

      /* Build the Huffman trees unless a stored block is forced */
      if (s.level > 0) {

        /* Check if the file is binary or text */
        if (s.strm.data_type === Z_UNKNOWN$2) {
          s.strm.data_type = detect_data_type(s);
        }

        /* Construct the literal and distance trees */
        build_tree(s, s.l_desc);
        // Tracev((stderr, "\nlit data: dyn %ld, stat %ld", s->opt_len,
        //        s->static_len));

        build_tree(s, s.d_desc);
        // Tracev((stderr, "\ndist data: dyn %ld, stat %ld", s->opt_len,
        //        s->static_len));
        /* At this point, opt_len and static_len are the total bit lengths of
         * the compressed block data, excluding the tree representations.
         */

        /* Build the bit length tree for the above two trees, and get the index
         * in bl_order of the last bit length code to send.
         */
        max_blindex = build_bl_tree(s);

        /* Determine the best encoding. Compute the block lengths in bytes. */
        opt_lenb = (s.opt_len + 3 + 7) >>> 3;
        static_lenb = (s.static_len + 3 + 7) >>> 3;

        // Tracev((stderr, "\nopt %lu(%lu) stat %lu(%lu) stored %lu lit %u ",
        //        opt_lenb, s->opt_len, static_lenb, s->static_len, stored_len,
        //        s->last_lit));

        if (static_lenb <= opt_lenb) {
          opt_lenb = static_lenb;
        }

      } else {
        // Assert(buf != (char*)0, "lost buf");
        opt_lenb = static_lenb = stored_len + 5; /* force a stored block */
      }

      if ((stored_len + 4 <= opt_lenb) && (buf !== -1)) {
        /* 4: two words for the lengths */

        /* The test buf != NULL is only necessary if LIT_BUFSIZE > WSIZE.
         * Otherwise we can't have processed more than WSIZE input bytes since
         * the last block flush, because compression would have been
         * successful. If LIT_BUFSIZE <= WSIZE, it is never too late to
         * transform a block into a stored block.
         */
        _tr_stored_block(s, buf, stored_len, last);

      } else if (s.strategy === Z_FIXED$2 || static_lenb === opt_lenb) {

        send_bits(s, (STATIC_TREES << 1) + (last ? 1 : 0), 3);
        compress_block(s, static_ltree, static_dtree);

      } else {
        send_bits(s, (DYN_TREES << 1) + (last ? 1 : 0), 3);
        send_all_trees(s, s.l_desc.max_code + 1, s.d_desc.max_code + 1, max_blindex + 1);
        compress_block(s, s.dyn_ltree, s.dyn_dtree);
      }
      // Assert (s->compressed_len == s->bits_sent, "bad compressed size");
      /* The above check is made mod 2^32, for files larger than 512 MB
       * and uLong implemented on 32 bits.
       */
      init_block(s);

      if (last) {
        bi_windup(s);
      }
      // Tracev((stderr,"\ncomprlen %lu(%lu) ", s->compressed_len>>3,
      //       s->compressed_len-7*last));
    }

    /* ===========================================================================
     * Save the match info and tally the frequency counts. Return true if
     * the current block must be flushed.
     */
    function _tr_tally(s, dist, lc)
    //    deflate_state *s;
    //    unsigned dist;  /* distance of matched string */
    //    unsigned lc;    /* match length-MIN_MATCH or unmatched char (if dist==0) */
    {
      //var out_length, in_length, dcode;

      s.pending_buf[s.d_buf + s.last_lit * 2] = (dist >>> 8) & 0xff;
      s.pending_buf[s.d_buf + s.last_lit * 2 + 1] = dist & 0xff;

      s.pending_buf[s.l_buf + s.last_lit] = lc & 0xff;
      s.last_lit++;

      if (dist === 0) {
        /* lc is the unmatched char */
        s.dyn_ltree[lc * 2] /*.Freq*/ ++;
      } else {
        s.matches++;
        /* Here, lc is the match length - MIN_MATCH */
        dist--; /* dist = match distance - 1 */
        //Assert((ush)dist < (ush)MAX_DIST(s) &&
        //       (ush)lc <= (ush)(MAX_MATCH-MIN_MATCH) &&
        //       (ush)d_code(dist) < (ush)D_CODES,  "_tr_tally: bad match");

        s.dyn_ltree[(_length_code[lc] + LITERALS$1 + 1) * 2] /*.Freq*/ ++;
        s.dyn_dtree[d_code(dist) * 2] /*.Freq*/ ++;
      }

      // (!) This block is disabled in zlib defailts,
      // don't enable it for binary compatibility

      //#ifdef TRUNCATE_BLOCK
      //  /* Try to guess if it is profitable to stop the current block here */
      //  if ((s.last_lit & 0x1fff) === 0 && s.level > 2) {
      //    /* Compute an upper bound for the compressed length */
      //    out_length = s.last_lit*8;
      //    in_length = s.strstart - s.block_start;
      //
      //    for (dcode = 0; dcode < D_CODES; dcode++) {
      //      out_length += s.dyn_dtree[dcode*2]/*.Freq*/ * (5 + extra_dbits[dcode]);
      //    }
      //    out_length >>>= 3;
      //    //Tracev((stderr,"\nlast_lit %u, in %ld, out ~%ld(%ld%%) ",
      //    //       s->last_lit, in_length, out_length,
      //    //       100L - out_length*100L/in_length));
      //    if (s.matches < (s.last_lit>>1)/*int /2*/ && out_length < (in_length>>1)/*int /2*/) {
      //      return true;
      //    }
      //  }
      //#endif

      return (s.last_lit === s.lit_bufsize - 1);
      /* We avoid equality with lit_bufsize because of wraparound at 64K
       * on 16 bit machines and because stored blocks are restricted to
       * 64K-1 bytes.
       */
    }

    // Note: adler32 takes 12% for level 0 and 2% for level 6.
    // It doesn't worth to make additional optimizationa as in original.
    // Small size is preferable.

    function adler32(adler, buf, len, pos) {
      var s1 = (adler & 0xffff) |0,
          s2 = ((adler >>> 16) & 0xffff) |0,
          n = 0;

      while (len !== 0) {
        // Set limit ~ twice less than 5552, to keep
        // s2 in 31-bits, because we force signed ints.
        // in other case %= will fail.
        n = len > 2000 ? 2000 : len;
        len -= n;

        do {
          s1 = (s1 + buf[pos++]) |0;
          s2 = (s2 + s1) |0;
        } while (--n);

        s1 %= 65521;
        s2 %= 65521;
      }

      return (s1 | (s2 << 16)) |0;
    }

    // Note: we can't get significant speed boost here.
    // So write code to minimize size - no pregenerated tables
    // and array tools dependencies.


    // Use ordinary array, since untyped makes no boost here
    function makeTable() {
      var c, table = [];

      for (var n = 0; n < 256; n++) {
        c = n;
        for (var k = 0; k < 8; k++) {
          c = ((c & 1) ? (0xEDB88320 ^ (c >>> 1)) : (c >>> 1));
        }
        table[n] = c;
      }

      return table;
    }

    // Create table on load. Just 255 signed longs. Not a problem.
    var crcTable = makeTable();


    function crc32(crc, buf, len, pos) {
      var t = crcTable,
          end = pos + len;

      crc ^= -1;

      for (var i = pos; i < end; i++) {
        crc = (crc >>> 8) ^ t[(crc ^ buf[i]) & 0xFF];
      }

      return (crc ^ (-1)); // >>> 0;
    }

    /* Public constants ==========================================================*/
    /* ===========================================================================*/


    /* Allowed flush values; see deflate() and inflate() below for details */
    var Z_NO_FLUSH$1 = 0;
    var Z_PARTIAL_FLUSH$1 = 1;
    //var Z_SYNC_FLUSH    = 2;
    var Z_FULL_FLUSH$1 = 3;
    var Z_FINISH$2 = 4;
    var Z_BLOCK$2 = 5;
    //var Z_TREES         = 6;


    /* Return codes for the compression/decompression functions. Negative values
     * are errors, positive values are used for special but normal events.
     */
    var Z_OK$2 = 0;
    var Z_STREAM_END$2 = 1;
    //var Z_NEED_DICT     = 2;
    //var Z_ERRNO         = -1;
    var Z_STREAM_ERROR$2 = -2;
    var Z_DATA_ERROR$2 = -3;
    //var Z_MEM_ERROR     = -4;
    var Z_BUF_ERROR$2 = -5;
    //var Z_VERSION_ERROR = -6;


    /* compression levels */
    //var Z_NO_COMPRESSION      = 0;
    //var Z_BEST_SPEED          = 1;
    //var Z_BEST_COMPRESSION    = 9;
    var Z_DEFAULT_COMPRESSION$1 = -1;


    var Z_FILTERED$1 = 1;
    var Z_HUFFMAN_ONLY$1 = 2;
    var Z_RLE$1 = 3;
    var Z_FIXED$1 = 4;

    /* Possible values of the data_type field (though see inflate()) */
    //var Z_BINARY              = 0;
    //var Z_TEXT                = 1;
    //var Z_ASCII               = 1; // = Z_TEXT
    var Z_UNKNOWN$1 = 2;


    /* The deflate compression method */
    var Z_DEFLATED$2 = 8;

    /*============================================================================*/


    var MAX_MEM_LEVEL = 9;


    var LENGTH_CODES = 29;
    /* number of length codes, not counting the special END_BLOCK code */
    var LITERALS = 256;
    /* number of literal bytes 0..255 */
    var L_CODES = LITERALS + 1 + LENGTH_CODES;
    /* number of Literal or Length codes, including the END_BLOCK code */
    var D_CODES = 30;
    /* number of distance codes */
    var BL_CODES = 19;
    /* number of codes used to transfer the bit lengths */
    var HEAP_SIZE = 2 * L_CODES + 1;
    /* maximum heap size */
    var MAX_BITS = 15;
    /* All codes must not exceed MAX_BITS bits */

    var MIN_MATCH = 3;
    var MAX_MATCH = 258;
    var MIN_LOOKAHEAD = (MAX_MATCH + MIN_MATCH + 1);

    var PRESET_DICT = 0x20;

    var INIT_STATE = 42;
    var EXTRA_STATE = 69;
    var NAME_STATE = 73;
    var COMMENT_STATE = 91;
    var HCRC_STATE = 103;
    var BUSY_STATE = 113;
    var FINISH_STATE = 666;

    var BS_NEED_MORE = 1; /* block not completed, need more input or more output */
    var BS_BLOCK_DONE = 2; /* block flush performed */
    var BS_FINISH_STARTED = 3; /* finish started, need only more output at next deflate */
    var BS_FINISH_DONE = 4; /* finish done, accept no more input or output */

    var OS_CODE = 0x03; // Unix :) . Don't detect, use this default.

    function err(strm, errorCode) {
      strm.msg = msg[errorCode];
      return errorCode;
    }

    function rank(f) {
      return ((f) << 1) - ((f) > 4 ? 9 : 0);
    }

    function zero(buf) {
      var len = buf.length;
      while (--len >= 0) {
        buf[len] = 0;
      }
    }


    /* =========================================================================
     * Flush as much pending output as possible. All deflate() output goes
     * through this function so some applications may wish to modify it
     * to avoid allocating a large strm->output buffer and copying into it.
     * (See also read_buf()).
     */
    function flush_pending(strm) {
      var s = strm.state;

      //_tr_flush_bits(s);
      var len = s.pending;
      if (len > strm.avail_out) {
        len = strm.avail_out;
      }
      if (len === 0) {
        return;
      }

      arraySet(strm.output, s.pending_buf, s.pending_out, len, strm.next_out);
      strm.next_out += len;
      s.pending_out += len;
      strm.total_out += len;
      strm.avail_out -= len;
      s.pending -= len;
      if (s.pending === 0) {
        s.pending_out = 0;
      }
    }


    function flush_block_only(s, last) {
      _tr_flush_block(s, (s.block_start >= 0 ? s.block_start : -1), s.strstart - s.block_start, last);
      s.block_start = s.strstart;
      flush_pending(s.strm);
    }


    function put_byte(s, b) {
      s.pending_buf[s.pending++] = b;
    }


    /* =========================================================================
     * Put a short in the pending buffer. The 16-bit value is put in MSB order.
     * IN assertion: the stream state is correct and there is enough room in
     * pending_buf.
     */
    function putShortMSB(s, b) {
      //  put_byte(s, (Byte)(b >> 8));
      //  put_byte(s, (Byte)(b & 0xff));
      s.pending_buf[s.pending++] = (b >>> 8) & 0xff;
      s.pending_buf[s.pending++] = b & 0xff;
    }


    /* ===========================================================================
     * Read a new buffer from the current input stream, update the adler32
     * and total number of bytes read.  All deflate() input goes through
     * this function so some applications may wish to modify it to avoid
     * allocating a large strm->input buffer and copying from it.
     * (See also flush_pending()).
     */
    function read_buf(strm, buf, start, size) {
      var len = strm.avail_in;

      if (len > size) {
        len = size;
      }
      if (len === 0) {
        return 0;
      }

      strm.avail_in -= len;

      // zmemcpy(buf, strm->next_in, len);
      arraySet(buf, strm.input, strm.next_in, len, start);
      if (strm.state.wrap === 1) {
        strm.adler = adler32(strm.adler, buf, len, start);
      } else if (strm.state.wrap === 2) {
        strm.adler = crc32(strm.adler, buf, len, start);
      }

      strm.next_in += len;
      strm.total_in += len;

      return len;
    }


    /* ===========================================================================
     * Set match_start to the longest match starting at the given string and
     * return its length. Matches shorter or equal to prev_length are discarded,
     * in which case the result is equal to prev_length and match_start is
     * garbage.
     * IN assertions: cur_match is the head of the hash chain for the current
     *   string (strstart) and its distance is <= MAX_DIST, and prev_length >= 1
     * OUT assertion: the match length is not greater than s->lookahead.
     */
    function longest_match(s, cur_match) {
      var chain_length = s.max_chain_length; /* max hash chain length */
      var scan = s.strstart; /* current string */
      var match; /* matched string */
      var len; /* length of current match */
      var best_len = s.prev_length; /* best match length so far */
      var nice_match = s.nice_match; /* stop if match long enough */
      var limit = (s.strstart > (s.w_size - MIN_LOOKAHEAD)) ?
        s.strstart - (s.w_size - MIN_LOOKAHEAD) : 0 /*NIL*/ ;

      var _win = s.window; // shortcut

      var wmask = s.w_mask;
      var prev = s.prev;

      /* Stop when cur_match becomes <= limit. To simplify the code,
       * we prevent matches with the string of window index 0.
       */

      var strend = s.strstart + MAX_MATCH;
      var scan_end1 = _win[scan + best_len - 1];
      var scan_end = _win[scan + best_len];

      /* The code is optimized for HASH_BITS >= 8 and MAX_MATCH-2 multiple of 16.
       * It is easy to get rid of this optimization if necessary.
       */
      // Assert(s->hash_bits >= 8 && MAX_MATCH == 258, "Code too clever");

      /* Do not waste too much time if we already have a good match: */
      if (s.prev_length >= s.good_match) {
        chain_length >>= 2;
      }
      /* Do not look for matches beyond the end of the input. This is necessary
       * to make deflate deterministic.
       */
      if (nice_match > s.lookahead) {
        nice_match = s.lookahead;
      }

      // Assert((ulg)s->strstart <= s->window_size-MIN_LOOKAHEAD, "need lookahead");

      do {
        // Assert(cur_match < s->strstart, "no future");
        match = cur_match;

        /* Skip to next match if the match length cannot increase
         * or if the match length is less than 2.  Note that the checks below
         * for insufficient lookahead only occur occasionally for performance
         * reasons.  Therefore uninitialized memory will be accessed, and
         * conditional jumps will be made that depend on those values.
         * However the length of the match is limited to the lookahead, so
         * the output of deflate is not affected by the uninitialized values.
         */

        if (_win[match + best_len] !== scan_end ||
          _win[match + best_len - 1] !== scan_end1 ||
          _win[match] !== _win[scan] ||
          _win[++match] !== _win[scan + 1]) {
          continue;
        }

        /* The check at best_len-1 can be removed because it will be made
         * again later. (This heuristic is not always a win.)
         * It is not necessary to compare scan[2] and match[2] since they
         * are always equal when the other bytes match, given that
         * the hash keys are equal and that HASH_BITS >= 8.
         */
        scan += 2;
        match++;
        // Assert(*scan == *match, "match[2]?");

        /* We check for insufficient lookahead only every 8th comparison;
         * the 256th check will be made at strstart+258.
         */
        do {
          /*jshint noempty:false*/
        } while (_win[++scan] === _win[++match] && _win[++scan] === _win[++match] &&
          _win[++scan] === _win[++match] && _win[++scan] === _win[++match] &&
          _win[++scan] === _win[++match] && _win[++scan] === _win[++match] &&
          _win[++scan] === _win[++match] && _win[++scan] === _win[++match] &&
          scan < strend);

        // Assert(scan <= s->window+(unsigned)(s->window_size-1), "wild scan");

        len = MAX_MATCH - (strend - scan);
        scan = strend - MAX_MATCH;

        if (len > best_len) {
          s.match_start = cur_match;
          best_len = len;
          if (len >= nice_match) {
            break;
          }
          scan_end1 = _win[scan + best_len - 1];
          scan_end = _win[scan + best_len];
        }
      } while ((cur_match = prev[cur_match & wmask]) > limit && --chain_length !== 0);

      if (best_len <= s.lookahead) {
        return best_len;
      }
      return s.lookahead;
    }


    /* ===========================================================================
     * Fill the window when the lookahead becomes insufficient.
     * Updates strstart and lookahead.
     *
     * IN assertion: lookahead < MIN_LOOKAHEAD
     * OUT assertions: strstart <= window_size-MIN_LOOKAHEAD
     *    At least one byte has been read, or avail_in == 0; reads are
     *    performed for at least two bytes (required for the zip translate_eol
     *    option -- not supported here).
     */
    function fill_window(s) {
      var _w_size = s.w_size;
      var p, n, m, more, str;

      //Assert(s->lookahead < MIN_LOOKAHEAD, "already enough lookahead");

      do {
        more = s.window_size - s.lookahead - s.strstart;

        // JS ints have 32 bit, block below not needed
        /* Deal with !@#$% 64K limit: */
        //if (sizeof(int) <= 2) {
        //    if (more == 0 && s->strstart == 0 && s->lookahead == 0) {
        //        more = wsize;
        //
        //  } else if (more == (unsigned)(-1)) {
        //        /* Very unlikely, but possible on 16 bit machine if
        //         * strstart == 0 && lookahead == 1 (input done a byte at time)
        //         */
        //        more--;
        //    }
        //}


        /* If the window is almost full and there is insufficient lookahead,
         * move the upper half to the lower one to make room in the upper half.
         */
        if (s.strstart >= _w_size + (_w_size - MIN_LOOKAHEAD)) {

          arraySet(s.window, s.window, _w_size, _w_size, 0);
          s.match_start -= _w_size;
          s.strstart -= _w_size;
          /* we now have strstart >= MAX_DIST */
          s.block_start -= _w_size;

          /* Slide the hash table (could be avoided with 32 bit values
           at the expense of memory usage). We slide even when level == 0
           to keep the hash table consistent if we switch back to level > 0
           later. (Using level 0 permanently is not an optimal usage of
           zlib, so we don't care about this pathological case.)
           */

          n = s.hash_size;
          p = n;
          do {
            m = s.head[--p];
            s.head[p] = (m >= _w_size ? m - _w_size : 0);
          } while (--n);

          n = _w_size;
          p = n;
          do {
            m = s.prev[--p];
            s.prev[p] = (m >= _w_size ? m - _w_size : 0);
            /* If n is not on any hash chain, prev[n] is garbage but
             * its value will never be used.
             */
          } while (--n);

          more += _w_size;
        }
        if (s.strm.avail_in === 0) {
          break;
        }

        /* If there was no sliding:
         *    strstart <= WSIZE+MAX_DIST-1 && lookahead <= MIN_LOOKAHEAD - 1 &&
         *    more == window_size - lookahead - strstart
         * => more >= window_size - (MIN_LOOKAHEAD-1 + WSIZE + MAX_DIST-1)
         * => more >= window_size - 2*WSIZE + 2
         * In the BIG_MEM or MMAP case (not yet supported),
         *   window_size == input_size + MIN_LOOKAHEAD  &&
         *   strstart + s->lookahead <= input_size => more >= MIN_LOOKAHEAD.
         * Otherwise, window_size == 2*WSIZE so more >= 2.
         * If there was sliding, more >= WSIZE. So in all cases, more >= 2.
         */
        //Assert(more >= 2, "more < 2");
        n = read_buf(s.strm, s.window, s.strstart + s.lookahead, more);
        s.lookahead += n;

        /* Initialize the hash value now that we have some input: */
        if (s.lookahead + s.insert >= MIN_MATCH) {
          str = s.strstart - s.insert;
          s.ins_h = s.window[str];

          /* UPDATE_HASH(s, s->ins_h, s->window[str + 1]); */
          s.ins_h = ((s.ins_h << s.hash_shift) ^ s.window[str + 1]) & s.hash_mask;
          //#if MIN_MATCH != 3
          //        Call update_hash() MIN_MATCH-3 more times
          //#endif
          while (s.insert) {
            /* UPDATE_HASH(s, s->ins_h, s->window[str + MIN_MATCH-1]); */
            s.ins_h = ((s.ins_h << s.hash_shift) ^ s.window[str + MIN_MATCH - 1]) & s.hash_mask;

            s.prev[str & s.w_mask] = s.head[s.ins_h];
            s.head[s.ins_h] = str;
            str++;
            s.insert--;
            if (s.lookahead + s.insert < MIN_MATCH) {
              break;
            }
          }
        }
        /* If the whole input has less than MIN_MATCH bytes, ins_h is garbage,
         * but this is not important since only literal bytes will be emitted.
         */

      } while (s.lookahead < MIN_LOOKAHEAD && s.strm.avail_in !== 0);

      /* If the WIN_INIT bytes after the end of the current data have never been
       * written, then zero those bytes in order to avoid memory check reports of
       * the use of uninitialized (or uninitialised as Julian writes) bytes by
       * the longest match routines.  Update the high water mark for the next
       * time through here.  WIN_INIT is set to MAX_MATCH since the longest match
       * routines allow scanning to strstart + MAX_MATCH, ignoring lookahead.
       */
      //  if (s.high_water < s.window_size) {
      //    var curr = s.strstart + s.lookahead;
      //    var init = 0;
      //
      //    if (s.high_water < curr) {
      //      /* Previous high water mark below current data -- zero WIN_INIT
      //       * bytes or up to end of window, whichever is less.
      //       */
      //      init = s.window_size - curr;
      //      if (init > WIN_INIT)
      //        init = WIN_INIT;
      //      zmemzero(s->window + curr, (unsigned)init);
      //      s->high_water = curr + init;
      //    }
      //    else if (s->high_water < (ulg)curr + WIN_INIT) {
      //      /* High water mark at or above current data, but below current data
      //       * plus WIN_INIT -- zero out to current data plus WIN_INIT, or up
      //       * to end of window, whichever is less.
      //       */
      //      init = (ulg)curr + WIN_INIT - s->high_water;
      //      if (init > s->window_size - s->high_water)
      //        init = s->window_size - s->high_water;
      //      zmemzero(s->window + s->high_water, (unsigned)init);
      //      s->high_water += init;
      //    }
      //  }
      //
      //  Assert((ulg)s->strstart <= s->window_size - MIN_LOOKAHEAD,
      //    "not enough room for search");
    }

    /* ===========================================================================
     * Copy without compression as much as possible from the input stream, return
     * the current block state.
     * This function does not insert new strings in the dictionary since
     * uncompressible data is probably not useful. This function is used
     * only for the level=0 compression option.
     * NOTE: this function should be optimized to avoid extra copying from
     * window to pending_buf.
     */
    function deflate_stored(s, flush) {
      /* Stored blocks are limited to 0xffff bytes, pending_buf is limited
       * to pending_buf_size, and each stored block has a 5 byte header:
       */
      var max_block_size = 0xffff;

      if (max_block_size > s.pending_buf_size - 5) {
        max_block_size = s.pending_buf_size - 5;
      }

      /* Copy as much as possible from input to output: */
      for (;;) {
        /* Fill the window as much as possible: */
        if (s.lookahead <= 1) {

          //Assert(s->strstart < s->w_size+MAX_DIST(s) ||
          //  s->block_start >= (long)s->w_size, "slide too late");
          //      if (!(s.strstart < s.w_size + (s.w_size - MIN_LOOKAHEAD) ||
          //        s.block_start >= s.w_size)) {
          //        throw  new Error("slide too late");
          //      }

          fill_window(s);
          if (s.lookahead === 0 && flush === Z_NO_FLUSH$1) {
            return BS_NEED_MORE;
          }

          if (s.lookahead === 0) {
            break;
          }
          /* flush the current block */
        }
        //Assert(s->block_start >= 0L, "block gone");
        //    if (s.block_start < 0) throw new Error("block gone");

        s.strstart += s.lookahead;
        s.lookahead = 0;

        /* Emit a stored block if pending_buf will be full: */
        var max_start = s.block_start + max_block_size;

        if (s.strstart === 0 || s.strstart >= max_start) {
          /* strstart == 0 is possible when wraparound on 16-bit machine */
          s.lookahead = s.strstart - max_start;
          s.strstart = max_start;
          /*** FLUSH_BLOCK(s, 0); ***/
          flush_block_only(s, false);
          if (s.strm.avail_out === 0) {
            return BS_NEED_MORE;
          }
          /***/


        }
        /* Flush if we may have to slide, otherwise block_start may become
         * negative and the data will be gone:
         */
        if (s.strstart - s.block_start >= (s.w_size - MIN_LOOKAHEAD)) {
          /*** FLUSH_BLOCK(s, 0); ***/
          flush_block_only(s, false);
          if (s.strm.avail_out === 0) {
            return BS_NEED_MORE;
          }
          /***/
        }
      }

      s.insert = 0;

      if (flush === Z_FINISH$2) {
        /*** FLUSH_BLOCK(s, 1); ***/
        flush_block_only(s, true);
        if (s.strm.avail_out === 0) {
          return BS_FINISH_STARTED;
        }
        /***/
        return BS_FINISH_DONE;
      }

      if (s.strstart > s.block_start) {
        /*** FLUSH_BLOCK(s, 0); ***/
        flush_block_only(s, false);
        if (s.strm.avail_out === 0) {
          return BS_NEED_MORE;
        }
        /***/
      }

      return BS_NEED_MORE;
    }

    /* ===========================================================================
     * Compress as much as possible from the input stream, return the current
     * block state.
     * This function does not perform lazy evaluation of matches and inserts
     * new strings in the dictionary only for unmatched strings or for short
     * matches. It is used only for the fast compression options.
     */
    function deflate_fast(s, flush) {
      var hash_head; /* head of the hash chain */
      var bflush; /* set if current block must be flushed */

      for (;;) {
        /* Make sure that we always have enough lookahead, except
         * at the end of the input file. We need MAX_MATCH bytes
         * for the next match, plus MIN_MATCH bytes to insert the
         * string following the next match.
         */
        if (s.lookahead < MIN_LOOKAHEAD) {
          fill_window(s);
          if (s.lookahead < MIN_LOOKAHEAD && flush === Z_NO_FLUSH$1) {
            return BS_NEED_MORE;
          }
          if (s.lookahead === 0) {
            break; /* flush the current block */
          }
        }

        /* Insert the string window[strstart .. strstart+2] in the
         * dictionary, and set hash_head to the head of the hash chain:
         */
        hash_head = 0 /*NIL*/ ;
        if (s.lookahead >= MIN_MATCH) {
          /*** INSERT_STRING(s, s.strstart, hash_head); ***/
          s.ins_h = ((s.ins_h << s.hash_shift) ^ s.window[s.strstart + MIN_MATCH - 1]) & s.hash_mask;
          hash_head = s.prev[s.strstart & s.w_mask] = s.head[s.ins_h];
          s.head[s.ins_h] = s.strstart;
          /***/
        }

        /* Find the longest match, discarding those <= prev_length.
         * At this point we have always match_length < MIN_MATCH
         */
        if (hash_head !== 0 /*NIL*/ && ((s.strstart - hash_head) <= (s.w_size - MIN_LOOKAHEAD))) {
          /* To simplify the code, we prevent matches with the string
           * of window index 0 (in particular we have to avoid a match
           * of the string with itself at the start of the input file).
           */
          s.match_length = longest_match(s, hash_head);
          /* longest_match() sets match_start */
        }
        if (s.match_length >= MIN_MATCH) {
          // check_match(s, s.strstart, s.match_start, s.match_length); // for debug only

          /*** _tr_tally_dist(s, s.strstart - s.match_start,
                         s.match_length - MIN_MATCH, bflush); ***/
          bflush = _tr_tally(s, s.strstart - s.match_start, s.match_length - MIN_MATCH);

          s.lookahead -= s.match_length;

          /* Insert new strings in the hash table only if the match length
           * is not too large. This saves time but degrades compression.
           */
          if (s.match_length <= s.max_lazy_match /*max_insert_length*/ && s.lookahead >= MIN_MATCH) {
            s.match_length--; /* string at strstart already in table */
            do {
              s.strstart++;
              /*** INSERT_STRING(s, s.strstart, hash_head); ***/
              s.ins_h = ((s.ins_h << s.hash_shift) ^ s.window[s.strstart + MIN_MATCH - 1]) & s.hash_mask;
              hash_head = s.prev[s.strstart & s.w_mask] = s.head[s.ins_h];
              s.head[s.ins_h] = s.strstart;
              /***/
              /* strstart never exceeds WSIZE-MAX_MATCH, so there are
               * always MIN_MATCH bytes ahead.
               */
            } while (--s.match_length !== 0);
            s.strstart++;
          } else {
            s.strstart += s.match_length;
            s.match_length = 0;
            s.ins_h = s.window[s.strstart];
            /* UPDATE_HASH(s, s.ins_h, s.window[s.strstart+1]); */
            s.ins_h = ((s.ins_h << s.hash_shift) ^ s.window[s.strstart + 1]) & s.hash_mask;

            //#if MIN_MATCH != 3
            //                Call UPDATE_HASH() MIN_MATCH-3 more times
            //#endif
            /* If lookahead < MIN_MATCH, ins_h is garbage, but it does not
             * matter since it will be recomputed at next deflate call.
             */
          }
        } else {
          /* No match, output a literal byte */
          //Tracevv((stderr,"%c", s.window[s.strstart]));
          /*** _tr_tally_lit(s, s.window[s.strstart], bflush); ***/
          bflush = _tr_tally(s, 0, s.window[s.strstart]);

          s.lookahead--;
          s.strstart++;
        }
        if (bflush) {
          /*** FLUSH_BLOCK(s, 0); ***/
          flush_block_only(s, false);
          if (s.strm.avail_out === 0) {
            return BS_NEED_MORE;
          }
          /***/
        }
      }
      s.insert = ((s.strstart < (MIN_MATCH - 1)) ? s.strstart : MIN_MATCH - 1);
      if (flush === Z_FINISH$2) {
        /*** FLUSH_BLOCK(s, 1); ***/
        flush_block_only(s, true);
        if (s.strm.avail_out === 0) {
          return BS_FINISH_STARTED;
        }
        /***/
        return BS_FINISH_DONE;
      }
      if (s.last_lit) {
        /*** FLUSH_BLOCK(s, 0); ***/
        flush_block_only(s, false);
        if (s.strm.avail_out === 0) {
          return BS_NEED_MORE;
        }
        /***/
      }
      return BS_BLOCK_DONE;
    }

    /* ===========================================================================
     * Same as above, but achieves better compression. We use a lazy
     * evaluation for matches: a match is finally adopted only if there is
     * no better match at the next window position.
     */
    function deflate_slow(s, flush) {
      var hash_head; /* head of hash chain */
      var bflush; /* set if current block must be flushed */

      var max_insert;

      /* Process the input block. */
      for (;;) {
        /* Make sure that we always have enough lookahead, except
         * at the end of the input file. We need MAX_MATCH bytes
         * for the next match, plus MIN_MATCH bytes to insert the
         * string following the next match.
         */
        if (s.lookahead < MIN_LOOKAHEAD) {
          fill_window(s);
          if (s.lookahead < MIN_LOOKAHEAD && flush === Z_NO_FLUSH$1) {
            return BS_NEED_MORE;
          }
          if (s.lookahead === 0) {
            break;
          } /* flush the current block */
        }

        /* Insert the string window[strstart .. strstart+2] in the
         * dictionary, and set hash_head to the head of the hash chain:
         */
        hash_head = 0 /*NIL*/ ;
        if (s.lookahead >= MIN_MATCH) {
          /*** INSERT_STRING(s, s.strstart, hash_head); ***/
          s.ins_h = ((s.ins_h << s.hash_shift) ^ s.window[s.strstart + MIN_MATCH - 1]) & s.hash_mask;
          hash_head = s.prev[s.strstart & s.w_mask] = s.head[s.ins_h];
          s.head[s.ins_h] = s.strstart;
          /***/
        }

        /* Find the longest match, discarding those <= prev_length.
         */
        s.prev_length = s.match_length;
        s.prev_match = s.match_start;
        s.match_length = MIN_MATCH - 1;

        if (hash_head !== 0 /*NIL*/ && s.prev_length < s.max_lazy_match &&
          s.strstart - hash_head <= (s.w_size - MIN_LOOKAHEAD) /*MAX_DIST(s)*/ ) {
          /* To simplify the code, we prevent matches with the string
           * of window index 0 (in particular we have to avoid a match
           * of the string with itself at the start of the input file).
           */
          s.match_length = longest_match(s, hash_head);
          /* longest_match() sets match_start */

          if (s.match_length <= 5 &&
            (s.strategy === Z_FILTERED$1 || (s.match_length === MIN_MATCH && s.strstart - s.match_start > 4096 /*TOO_FAR*/ ))) {

            /* If prev_match is also MIN_MATCH, match_start is garbage
             * but we will ignore the current match anyway.
             */
            s.match_length = MIN_MATCH - 1;
          }
        }
        /* If there was a match at the previous step and the current
         * match is not better, output the previous match:
         */
        if (s.prev_length >= MIN_MATCH && s.match_length <= s.prev_length) {
          max_insert = s.strstart + s.lookahead - MIN_MATCH;
          /* Do not insert strings in hash table beyond this. */

          //check_match(s, s.strstart-1, s.prev_match, s.prev_length);

          /***_tr_tally_dist(s, s.strstart - 1 - s.prev_match,
                         s.prev_length - MIN_MATCH, bflush);***/
          bflush = _tr_tally(s, s.strstart - 1 - s.prev_match, s.prev_length - MIN_MATCH);
          /* Insert in hash table all strings up to the end of the match.
           * strstart-1 and strstart are already inserted. If there is not
           * enough lookahead, the last two strings are not inserted in
           * the hash table.
           */
          s.lookahead -= s.prev_length - 1;
          s.prev_length -= 2;
          do {
            if (++s.strstart <= max_insert) {
              /*** INSERT_STRING(s, s.strstart, hash_head); ***/
              s.ins_h = ((s.ins_h << s.hash_shift) ^ s.window[s.strstart + MIN_MATCH - 1]) & s.hash_mask;
              hash_head = s.prev[s.strstart & s.w_mask] = s.head[s.ins_h];
              s.head[s.ins_h] = s.strstart;
              /***/
            }
          } while (--s.prev_length !== 0);
          s.match_available = 0;
          s.match_length = MIN_MATCH - 1;
          s.strstart++;

          if (bflush) {
            /*** FLUSH_BLOCK(s, 0); ***/
            flush_block_only(s, false);
            if (s.strm.avail_out === 0) {
              return BS_NEED_MORE;
            }
            /***/
          }

        } else if (s.match_available) {
          /* If there was no match at the previous position, output a
           * single literal. If there was a match but the current match
           * is longer, truncate the previous match to a single literal.
           */
          //Tracevv((stderr,"%c", s->window[s->strstart-1]));
          /*** _tr_tally_lit(s, s.window[s.strstart-1], bflush); ***/
          bflush = _tr_tally(s, 0, s.window[s.strstart - 1]);

          if (bflush) {
            /*** FLUSH_BLOCK_ONLY(s, 0) ***/
            flush_block_only(s, false);
            /***/
          }
          s.strstart++;
          s.lookahead--;
          if (s.strm.avail_out === 0) {
            return BS_NEED_MORE;
          }
        } else {
          /* There is no previous match to compare with, wait for
           * the next step to decide.
           */
          s.match_available = 1;
          s.strstart++;
          s.lookahead--;
        }
      }
      //Assert (flush != Z_NO_FLUSH, "no flush?");
      if (s.match_available) {
        //Tracevv((stderr,"%c", s->window[s->strstart-1]));
        /*** _tr_tally_lit(s, s.window[s.strstart-1], bflush); ***/
        bflush = _tr_tally(s, 0, s.window[s.strstart - 1]);

        s.match_available = 0;
      }
      s.insert = s.strstart < MIN_MATCH - 1 ? s.strstart : MIN_MATCH - 1;
      if (flush === Z_FINISH$2) {
        /*** FLUSH_BLOCK(s, 1); ***/
        flush_block_only(s, true);
        if (s.strm.avail_out === 0) {
          return BS_FINISH_STARTED;
        }
        /***/
        return BS_FINISH_DONE;
      }
      if (s.last_lit) {
        /*** FLUSH_BLOCK(s, 0); ***/
        flush_block_only(s, false);
        if (s.strm.avail_out === 0) {
          return BS_NEED_MORE;
        }
        /***/
      }

      return BS_BLOCK_DONE;
    }


    /* ===========================================================================
     * For Z_RLE, simply look for runs of bytes, generate matches only of distance
     * one.  Do not maintain a hash table.  (It will be regenerated if this run of
     * deflate switches away from Z_RLE.)
     */
    function deflate_rle(s, flush) {
      var bflush; /* set if current block must be flushed */
      var prev; /* byte at distance one to match */
      var scan, strend; /* scan goes up to strend for length of run */

      var _win = s.window;

      for (;;) {
        /* Make sure that we always have enough lookahead, except
         * at the end of the input file. We need MAX_MATCH bytes
         * for the longest run, plus one for the unrolled loop.
         */
        if (s.lookahead <= MAX_MATCH) {
          fill_window(s);
          if (s.lookahead <= MAX_MATCH && flush === Z_NO_FLUSH$1) {
            return BS_NEED_MORE;
          }
          if (s.lookahead === 0) {
            break;
          } /* flush the current block */
        }

        /* See how many times the previous byte repeats */
        s.match_length = 0;
        if (s.lookahead >= MIN_MATCH && s.strstart > 0) {
          scan = s.strstart - 1;
          prev = _win[scan];
          if (prev === _win[++scan] && prev === _win[++scan] && prev === _win[++scan]) {
            strend = s.strstart + MAX_MATCH;
            do {
              /*jshint noempty:false*/
            } while (prev === _win[++scan] && prev === _win[++scan] &&
              prev === _win[++scan] && prev === _win[++scan] &&
              prev === _win[++scan] && prev === _win[++scan] &&
              prev === _win[++scan] && prev === _win[++scan] &&
              scan < strend);
            s.match_length = MAX_MATCH - (strend - scan);
            if (s.match_length > s.lookahead) {
              s.match_length = s.lookahead;
            }
          }
          //Assert(scan <= s->window+(uInt)(s->window_size-1), "wild scan");
        }

        /* Emit match if have run of MIN_MATCH or longer, else emit literal */
        if (s.match_length >= MIN_MATCH) {
          //check_match(s, s.strstart, s.strstart - 1, s.match_length);

          /*** _tr_tally_dist(s, 1, s.match_length - MIN_MATCH, bflush); ***/
          bflush = _tr_tally(s, 1, s.match_length - MIN_MATCH);

          s.lookahead -= s.match_length;
          s.strstart += s.match_length;
          s.match_length = 0;
        } else {
          /* No match, output a literal byte */
          //Tracevv((stderr,"%c", s->window[s->strstart]));
          /*** _tr_tally_lit(s, s.window[s.strstart], bflush); ***/
          bflush = _tr_tally(s, 0, s.window[s.strstart]);

          s.lookahead--;
          s.strstart++;
        }
        if (bflush) {
          /*** FLUSH_BLOCK(s, 0); ***/
          flush_block_only(s, false);
          if (s.strm.avail_out === 0) {
            return BS_NEED_MORE;
          }
          /***/
        }
      }
      s.insert = 0;
      if (flush === Z_FINISH$2) {
        /*** FLUSH_BLOCK(s, 1); ***/
        flush_block_only(s, true);
        if (s.strm.avail_out === 0) {
          return BS_FINISH_STARTED;
        }
        /***/
        return BS_FINISH_DONE;
      }
      if (s.last_lit) {
        /*** FLUSH_BLOCK(s, 0); ***/
        flush_block_only(s, false);
        if (s.strm.avail_out === 0) {
          return BS_NEED_MORE;
        }
        /***/
      }
      return BS_BLOCK_DONE;
    }

    /* ===========================================================================
     * For Z_HUFFMAN_ONLY, do not look for matches.  Do not maintain a hash table.
     * (It will be regenerated if this run of deflate switches away from Huffman.)
     */
    function deflate_huff(s, flush) {
      var bflush; /* set if current block must be flushed */

      for (;;) {
        /* Make sure that we have a literal to write. */
        if (s.lookahead === 0) {
          fill_window(s);
          if (s.lookahead === 0) {
            if (flush === Z_NO_FLUSH$1) {
              return BS_NEED_MORE;
            }
            break; /* flush the current block */
          }
        }

        /* Output a literal byte */
        s.match_length = 0;
        //Tracevv((stderr,"%c", s->window[s->strstart]));
        /*** _tr_tally_lit(s, s.window[s.strstart], bflush); ***/
        bflush = _tr_tally(s, 0, s.window[s.strstart]);
        s.lookahead--;
        s.strstart++;
        if (bflush) {
          /*** FLUSH_BLOCK(s, 0); ***/
          flush_block_only(s, false);
          if (s.strm.avail_out === 0) {
            return BS_NEED_MORE;
          }
          /***/
        }
      }
      s.insert = 0;
      if (flush === Z_FINISH$2) {
        /*** FLUSH_BLOCK(s, 1); ***/
        flush_block_only(s, true);
        if (s.strm.avail_out === 0) {
          return BS_FINISH_STARTED;
        }
        /***/
        return BS_FINISH_DONE;
      }
      if (s.last_lit) {
        /*** FLUSH_BLOCK(s, 0); ***/
        flush_block_only(s, false);
        if (s.strm.avail_out === 0) {
          return BS_NEED_MORE;
        }
        /***/
      }
      return BS_BLOCK_DONE;
    }

    /* Values for max_lazy_match, good_match and max_chain_length, depending on
     * the desired pack level (0..9). The values given below have been tuned to
     * exclude worst case performance for pathological files. Better values may be
     * found for specific files.
     */
    function Config(good_length, max_lazy, nice_length, max_chain, func) {
      this.good_length = good_length;
      this.max_lazy = max_lazy;
      this.nice_length = nice_length;
      this.max_chain = max_chain;
      this.func = func;
    }

    var configuration_table;

    configuration_table = [
      /*      good lazy nice chain */
      new Config(0, 0, 0, 0, deflate_stored), /* 0 store only */
      new Config(4, 4, 8, 4, deflate_fast), /* 1 max speed, no lazy matches */
      new Config(4, 5, 16, 8, deflate_fast), /* 2 */
      new Config(4, 6, 32, 32, deflate_fast), /* 3 */

      new Config(4, 4, 16, 16, deflate_slow), /* 4 lazy matches */
      new Config(8, 16, 32, 32, deflate_slow), /* 5 */
      new Config(8, 16, 128, 128, deflate_slow), /* 6 */
      new Config(8, 32, 128, 256, deflate_slow), /* 7 */
      new Config(32, 128, 258, 1024, deflate_slow), /* 8 */
      new Config(32, 258, 258, 4096, deflate_slow) /* 9 max compression */
    ];


    /* ===========================================================================
     * Initialize the "longest match" routines for a new zlib stream
     */
    function lm_init(s) {
      s.window_size = 2 * s.w_size;

      /*** CLEAR_HASH(s); ***/
      zero(s.head); // Fill with NIL (= 0);

      /* Set the default configuration parameters:
       */
      s.max_lazy_match = configuration_table[s.level].max_lazy;
      s.good_match = configuration_table[s.level].good_length;
      s.nice_match = configuration_table[s.level].nice_length;
      s.max_chain_length = configuration_table[s.level].max_chain;

      s.strstart = 0;
      s.block_start = 0;
      s.lookahead = 0;
      s.insert = 0;
      s.match_length = s.prev_length = MIN_MATCH - 1;
      s.match_available = 0;
      s.ins_h = 0;
    }


    function DeflateState() {
      this.strm = null; /* pointer back to this zlib stream */
      this.status = 0; /* as the name implies */
      this.pending_buf = null; /* output still pending */
      this.pending_buf_size = 0; /* size of pending_buf */
      this.pending_out = 0; /* next pending byte to output to the stream */
      this.pending = 0; /* nb of bytes in the pending buffer */
      this.wrap = 0; /* bit 0 true for zlib, bit 1 true for gzip */
      this.gzhead = null; /* gzip header information to write */
      this.gzindex = 0; /* where in extra, name, or comment */
      this.method = Z_DEFLATED$2; /* can only be DEFLATED */
      this.last_flush = -1; /* value of flush param for previous deflate call */

      this.w_size = 0; /* LZ77 window size (32K by default) */
      this.w_bits = 0; /* log2(w_size)  (8..16) */
      this.w_mask = 0; /* w_size - 1 */

      this.window = null;
      /* Sliding window. Input bytes are read into the second half of the window,
       * and move to the first half later to keep a dictionary of at least wSize
       * bytes. With this organization, matches are limited to a distance of
       * wSize-MAX_MATCH bytes, but this ensures that IO is always
       * performed with a length multiple of the block size.
       */

      this.window_size = 0;
      /* Actual size of window: 2*wSize, except when the user input buffer
       * is directly used as sliding window.
       */

      this.prev = null;
      /* Link to older string with same hash index. To limit the size of this
       * array to 64K, this link is maintained only for the last 32K strings.
       * An index in this array is thus a window index modulo 32K.
       */

      this.head = null; /* Heads of the hash chains or NIL. */

      this.ins_h = 0; /* hash index of string to be inserted */
      this.hash_size = 0; /* number of elements in hash table */
      this.hash_bits = 0; /* log2(hash_size) */
      this.hash_mask = 0; /* hash_size-1 */

      this.hash_shift = 0;
      /* Number of bits by which ins_h must be shifted at each input
       * step. It must be such that after MIN_MATCH steps, the oldest
       * byte no longer takes part in the hash key, that is:
       *   hash_shift * MIN_MATCH >= hash_bits
       */

      this.block_start = 0;
      /* Window position at the beginning of the current output block. Gets
       * negative when the window is moved backwards.
       */

      this.match_length = 0; /* length of best match */
      this.prev_match = 0; /* previous match */
      this.match_available = 0; /* set if previous match exists */
      this.strstart = 0; /* start of string to insert */
      this.match_start = 0; /* start of matching string */
      this.lookahead = 0; /* number of valid bytes ahead in window */

      this.prev_length = 0;
      /* Length of the best match at previous step. Matches not greater than this
       * are discarded. This is used in the lazy match evaluation.
       */

      this.max_chain_length = 0;
      /* To speed up deflation, hash chains are never searched beyond this
       * length.  A higher limit improves compression ratio but degrades the
       * speed.
       */

      this.max_lazy_match = 0;
      /* Attempt to find a better match only when the current match is strictly
       * smaller than this value. This mechanism is used only for compression
       * levels >= 4.
       */
      // That's alias to max_lazy_match, don't use directly
      //this.max_insert_length = 0;
      /* Insert new strings in the hash table only if the match length is not
       * greater than this length. This saves time but degrades compression.
       * max_insert_length is used only for compression levels <= 3.
       */

      this.level = 0; /* compression level (1..9) */
      this.strategy = 0; /* favor or force Huffman coding*/

      this.good_match = 0;
      /* Use a faster search when the previous match is longer than this */

      this.nice_match = 0; /* Stop searching when current match exceeds this */

      /* used by c: */

      /* Didn't use ct_data typedef below to suppress compiler warning */

      // struct ct_data_s dyn_ltree[HEAP_SIZE];   /* literal and length tree */
      // struct ct_data_s dyn_dtree[2*D_CODES+1]; /* distance tree */
      // struct ct_data_s bl_tree[2*BL_CODES+1];  /* Huffman tree for bit lengths */

      // Use flat array of DOUBLE size, with interleaved fata,
      // because JS does not support effective
      this.dyn_ltree = new Buf16(HEAP_SIZE * 2);
      this.dyn_dtree = new Buf16((2 * D_CODES + 1) * 2);
      this.bl_tree = new Buf16((2 * BL_CODES + 1) * 2);
      zero(this.dyn_ltree);
      zero(this.dyn_dtree);
      zero(this.bl_tree);

      this.l_desc = null; /* desc. for literal tree */
      this.d_desc = null; /* desc. for distance tree */
      this.bl_desc = null; /* desc. for bit length tree */

      //ush bl_count[MAX_BITS+1];
      this.bl_count = new Buf16(MAX_BITS + 1);
      /* number of codes at each bit length for an optimal tree */

      //int heap[2*L_CODES+1];      /* heap used to build the Huffman trees */
      this.heap = new Buf16(2 * L_CODES + 1); /* heap used to build the Huffman trees */
      zero(this.heap);

      this.heap_len = 0; /* number of elements in the heap */
      this.heap_max = 0; /* element of largest frequency */
      /* The sons of heap[n] are heap[2*n] and heap[2*n+1]. heap[0] is not used.
       * The same heap array is used to build all
       */

      this.depth = new Buf16(2 * L_CODES + 1); //uch depth[2*L_CODES+1];
      zero(this.depth);
      /* Depth of each subtree used as tie breaker for trees of equal frequency
       */

      this.l_buf = 0; /* buffer index for literals or lengths */

      this.lit_bufsize = 0;
      /* Size of match buffer for literals/lengths.  There are 4 reasons for
       * limiting lit_bufsize to 64K:
       *   - frequencies can be kept in 16 bit counters
       *   - if compression is not successful for the first block, all input
       *     data is still in the window so we can still emit a stored block even
       *     when input comes from standard input.  (This can also be done for
       *     all blocks if lit_bufsize is not greater than 32K.)
       *   - if compression is not successful for a file smaller than 64K, we can
       *     even emit a stored file instead of a stored block (saving 5 bytes).
       *     This is applicable only for zip (not gzip or zlib).
       *   - creating new Huffman trees less frequently may not provide fast
       *     adaptation to changes in the input data statistics. (Take for
       *     example a binary file with poorly compressible code followed by
       *     a highly compressible string table.) Smaller buffer sizes give
       *     fast adaptation but have of course the overhead of transmitting
       *     trees more frequently.
       *   - I can't count above 4
       */

      this.last_lit = 0; /* running index in l_buf */

      this.d_buf = 0;
      /* Buffer index for distances. To simplify the code, d_buf and l_buf have
       * the same number of elements. To use different lengths, an extra flag
       * array would be necessary.
       */

      this.opt_len = 0; /* bit length of current block with optimal trees */
      this.static_len = 0; /* bit length of current block with static trees */
      this.matches = 0; /* number of string matches in current block */
      this.insert = 0; /* bytes at end of window left to insert */


      this.bi_buf = 0;
      /* Output buffer. bits are inserted starting at the bottom (least
       * significant bits).
       */
      this.bi_valid = 0;
      /* Number of valid bits in bi_buf.  All bits above the last valid bit
       * are always zero.
       */

      // Used for window memory init. We safely ignore it for JS. That makes
      // sense only for pointers and memory check tools.
      //this.high_water = 0;
      /* High water mark offset in window for initialized bytes -- bytes above
       * this are set to zero in order to avoid memory check warnings when
       * longest match routines access bytes past the input.  This is then
       * updated to the new high water mark.
       */
    }


    function deflateResetKeep(strm) {
      var s;

      if (!strm || !strm.state) {
        return err(strm, Z_STREAM_ERROR$2);
      }

      strm.total_in = strm.total_out = 0;
      strm.data_type = Z_UNKNOWN$1;

      s = strm.state;
      s.pending = 0;
      s.pending_out = 0;

      if (s.wrap < 0) {
        s.wrap = -s.wrap;
        /* was made negative by deflate(..., Z_FINISH); */
      }
      s.status = (s.wrap ? INIT_STATE : BUSY_STATE);
      strm.adler = (s.wrap === 2) ?
        0 // crc32(0, Z_NULL, 0)
        :
        1; // adler32(0, Z_NULL, 0)
      s.last_flush = Z_NO_FLUSH$1;
      _tr_init(s);
      return Z_OK$2;
    }


    function deflateReset(strm) {
      var ret = deflateResetKeep(strm);
      if (ret === Z_OK$2) {
        lm_init(strm.state);
      }
      return ret;
    }


    function deflateInit2(strm, level, method, windowBits, memLevel, strategy) {
      if (!strm) { // === Z_NULL
        return Z_STREAM_ERROR$2;
      }
      var wrap = 1;

      if (level === Z_DEFAULT_COMPRESSION$1) {
        level = 6;
      }

      if (windowBits < 0) { /* suppress zlib wrapper */
        wrap = 0;
        windowBits = -windowBits;
      } else if (windowBits > 15) {
        wrap = 2; /* write gzip wrapper instead */
        windowBits -= 16;
      }


      if (memLevel < 1 || memLevel > MAX_MEM_LEVEL || method !== Z_DEFLATED$2 ||
        windowBits < 8 || windowBits > 15 || level < 0 || level > 9 ||
        strategy < 0 || strategy > Z_FIXED$1) {
        return err(strm, Z_STREAM_ERROR$2);
      }


      if (windowBits === 8) {
        windowBits = 9;
      }
      /* until 256-byte window bug fixed */

      var s = new DeflateState();

      strm.state = s;
      s.strm = strm;

      s.wrap = wrap;
      s.gzhead = null;
      s.w_bits = windowBits;
      s.w_size = 1 << s.w_bits;
      s.w_mask = s.w_size - 1;

      s.hash_bits = memLevel + 7;
      s.hash_size = 1 << s.hash_bits;
      s.hash_mask = s.hash_size - 1;
      s.hash_shift = ~~((s.hash_bits + MIN_MATCH - 1) / MIN_MATCH);

      s.window = new Buf8(s.w_size * 2);
      s.head = new Buf16(s.hash_size);
      s.prev = new Buf16(s.w_size);

      // Don't need mem init magic for JS.
      //s.high_water = 0;  /* nothing written to s->window yet */

      s.lit_bufsize = 1 << (memLevel + 6); /* 16K elements by default */

      s.pending_buf_size = s.lit_bufsize * 4;

      //overlay = (ushf *) ZALLOC(strm, s->lit_bufsize, sizeof(ush)+2);
      //s->pending_buf = (uchf *) overlay;
      s.pending_buf = new Buf8(s.pending_buf_size);

      // It is offset from `s.pending_buf` (size is `s.lit_bufsize * 2`)
      //s->d_buf = overlay + s->lit_bufsize/sizeof(ush);
      s.d_buf = 1 * s.lit_bufsize;

      //s->l_buf = s->pending_buf + (1+sizeof(ush))*s->lit_bufsize;
      s.l_buf = (1 + 2) * s.lit_bufsize;

      s.level = level;
      s.strategy = strategy;
      s.method = method;

      return deflateReset(strm);
    }


    function deflate$1(strm, flush) {
      var old_flush, s;
      var beg, val; // for gzip header write only

      if (!strm || !strm.state ||
        flush > Z_BLOCK$2 || flush < 0) {
        return strm ? err(strm, Z_STREAM_ERROR$2) : Z_STREAM_ERROR$2;
      }

      s = strm.state;

      if (!strm.output ||
        (!strm.input && strm.avail_in !== 0) ||
        (s.status === FINISH_STATE && flush !== Z_FINISH$2)) {
        return err(strm, (strm.avail_out === 0) ? Z_BUF_ERROR$2 : Z_STREAM_ERROR$2);
      }

      s.strm = strm; /* just in case */
      old_flush = s.last_flush;
      s.last_flush = flush;

      /* Write the header */
      if (s.status === INIT_STATE) {
        if (s.wrap === 2) {
          // GZIP header
          strm.adler = 0; //crc32(0L, Z_NULL, 0);
          put_byte(s, 31);
          put_byte(s, 139);
          put_byte(s, 8);
          if (!s.gzhead) { // s->gzhead == Z_NULL
            put_byte(s, 0);
            put_byte(s, 0);
            put_byte(s, 0);
            put_byte(s, 0);
            put_byte(s, 0);
            put_byte(s, s.level === 9 ? 2 :
              (s.strategy >= Z_HUFFMAN_ONLY$1 || s.level < 2 ?
                4 : 0));
            put_byte(s, OS_CODE);
            s.status = BUSY_STATE;
          } else {
            put_byte(s, (s.gzhead.text ? 1 : 0) +
              (s.gzhead.hcrc ? 2 : 0) +
              (!s.gzhead.extra ? 0 : 4) +
              (!s.gzhead.name ? 0 : 8) +
              (!s.gzhead.comment ? 0 : 16)
            );
            put_byte(s, s.gzhead.time & 0xff);
            put_byte(s, (s.gzhead.time >> 8) & 0xff);
            put_byte(s, (s.gzhead.time >> 16) & 0xff);
            put_byte(s, (s.gzhead.time >> 24) & 0xff);
            put_byte(s, s.level === 9 ? 2 :
              (s.strategy >= Z_HUFFMAN_ONLY$1 || s.level < 2 ?
                4 : 0));
            put_byte(s, s.gzhead.os & 0xff);
            if (s.gzhead.extra && s.gzhead.extra.length) {
              put_byte(s, s.gzhead.extra.length & 0xff);
              put_byte(s, (s.gzhead.extra.length >> 8) & 0xff);
            }
            if (s.gzhead.hcrc) {
              strm.adler = crc32(strm.adler, s.pending_buf, s.pending, 0);
            }
            s.gzindex = 0;
            s.status = EXTRA_STATE;
          }
        } else // DEFLATE header
        {
          var header = (Z_DEFLATED$2 + ((s.w_bits - 8) << 4)) << 8;
          var level_flags = -1;

          if (s.strategy >= Z_HUFFMAN_ONLY$1 || s.level < 2) {
            level_flags = 0;
          } else if (s.level < 6) {
            level_flags = 1;
          } else if (s.level === 6) {
            level_flags = 2;
          } else {
            level_flags = 3;
          }
          header |= (level_flags << 6);
          if (s.strstart !== 0) {
            header |= PRESET_DICT;
          }
          header += 31 - (header % 31);

          s.status = BUSY_STATE;
          putShortMSB(s, header);

          /* Save the adler32 of the preset dictionary: */
          if (s.strstart !== 0) {
            putShortMSB(s, strm.adler >>> 16);
            putShortMSB(s, strm.adler & 0xffff);
          }
          strm.adler = 1; // adler32(0L, Z_NULL, 0);
        }
      }

      //#ifdef GZIP
      if (s.status === EXTRA_STATE) {
        if (s.gzhead.extra /* != Z_NULL*/ ) {
          beg = s.pending; /* start of bytes to update crc */

          while (s.gzindex < (s.gzhead.extra.length & 0xffff)) {
            if (s.pending === s.pending_buf_size) {
              if (s.gzhead.hcrc && s.pending > beg) {
                strm.adler = crc32(strm.adler, s.pending_buf, s.pending - beg, beg);
              }
              flush_pending(strm);
              beg = s.pending;
              if (s.pending === s.pending_buf_size) {
                break;
              }
            }
            put_byte(s, s.gzhead.extra[s.gzindex] & 0xff);
            s.gzindex++;
          }
          if (s.gzhead.hcrc && s.pending > beg) {
            strm.adler = crc32(strm.adler, s.pending_buf, s.pending - beg, beg);
          }
          if (s.gzindex === s.gzhead.extra.length) {
            s.gzindex = 0;
            s.status = NAME_STATE;
          }
        } else {
          s.status = NAME_STATE;
        }
      }
      if (s.status === NAME_STATE) {
        if (s.gzhead.name /* != Z_NULL*/ ) {
          beg = s.pending; /* start of bytes to update crc */
          //int val;

          do {
            if (s.pending === s.pending_buf_size) {
              if (s.gzhead.hcrc && s.pending > beg) {
                strm.adler = crc32(strm.adler, s.pending_buf, s.pending - beg, beg);
              }
              flush_pending(strm);
              beg = s.pending;
              if (s.pending === s.pending_buf_size) {
                val = 1;
                break;
              }
            }
            // JS specific: little magic to add zero terminator to end of string
            if (s.gzindex < s.gzhead.name.length) {
              val = s.gzhead.name.charCodeAt(s.gzindex++) & 0xff;
            } else {
              val = 0;
            }
            put_byte(s, val);
          } while (val !== 0);

          if (s.gzhead.hcrc && s.pending > beg) {
            strm.adler = crc32(strm.adler, s.pending_buf, s.pending - beg, beg);
          }
          if (val === 0) {
            s.gzindex = 0;
            s.status = COMMENT_STATE;
          }
        } else {
          s.status = COMMENT_STATE;
        }
      }
      if (s.status === COMMENT_STATE) {
        if (s.gzhead.comment /* != Z_NULL*/ ) {
          beg = s.pending; /* start of bytes to update crc */
          //int val;

          do {
            if (s.pending === s.pending_buf_size) {
              if (s.gzhead.hcrc && s.pending > beg) {
                strm.adler = crc32(strm.adler, s.pending_buf, s.pending - beg, beg);
              }
              flush_pending(strm);
              beg = s.pending;
              if (s.pending === s.pending_buf_size) {
                val = 1;
                break;
              }
            }
            // JS specific: little magic to add zero terminator to end of string
            if (s.gzindex < s.gzhead.comment.length) {
              val = s.gzhead.comment.charCodeAt(s.gzindex++) & 0xff;
            } else {
              val = 0;
            }
            put_byte(s, val);
          } while (val !== 0);

          if (s.gzhead.hcrc && s.pending > beg) {
            strm.adler = crc32(strm.adler, s.pending_buf, s.pending - beg, beg);
          }
          if (val === 0) {
            s.status = HCRC_STATE;
          }
        } else {
          s.status = HCRC_STATE;
        }
      }
      if (s.status === HCRC_STATE) {
        if (s.gzhead.hcrc) {
          if (s.pending + 2 > s.pending_buf_size) {
            flush_pending(strm);
          }
          if (s.pending + 2 <= s.pending_buf_size) {
            put_byte(s, strm.adler & 0xff);
            put_byte(s, (strm.adler >> 8) & 0xff);
            strm.adler = 0; //crc32(0L, Z_NULL, 0);
            s.status = BUSY_STATE;
          }
        } else {
          s.status = BUSY_STATE;
        }
      }
      //#endif

      /* Flush as much pending output as possible */
      if (s.pending !== 0) {
        flush_pending(strm);
        if (strm.avail_out === 0) {
          /* Since avail_out is 0, deflate will be called again with
           * more output space, but possibly with both pending and
           * avail_in equal to zero. There won't be anything to do,
           * but this is not an error situation so make sure we
           * return OK instead of BUF_ERROR at next call of deflate:
           */
          s.last_flush = -1;
          return Z_OK$2;
        }

        /* Make sure there is something to do and avoid duplicate consecutive
         * flushes. For repeated and useless calls with Z_FINISH, we keep
         * returning Z_STREAM_END instead of Z_BUF_ERROR.
         */
      } else if (strm.avail_in === 0 && rank(flush) <= rank(old_flush) &&
        flush !== Z_FINISH$2) {
        return err(strm, Z_BUF_ERROR$2);
      }

      /* User must not provide more input after the first FINISH: */
      if (s.status === FINISH_STATE && strm.avail_in !== 0) {
        return err(strm, Z_BUF_ERROR$2);
      }

      /* Start a new block or continue the current one.
       */
      if (strm.avail_in !== 0 || s.lookahead !== 0 ||
        (flush !== Z_NO_FLUSH$1 && s.status !== FINISH_STATE)) {
        var bstate = (s.strategy === Z_HUFFMAN_ONLY$1) ? deflate_huff(s, flush) :
          (s.strategy === Z_RLE$1 ? deflate_rle(s, flush) :
            configuration_table[s.level].func(s, flush));

        if (bstate === BS_FINISH_STARTED || bstate === BS_FINISH_DONE) {
          s.status = FINISH_STATE;
        }
        if (bstate === BS_NEED_MORE || bstate === BS_FINISH_STARTED) {
          if (strm.avail_out === 0) {
            s.last_flush = -1;
            /* avoid BUF_ERROR next call, see above */
          }
          return Z_OK$2;
          /* If flush != Z_NO_FLUSH && avail_out == 0, the next call
           * of deflate should use the same flush parameter to make sure
           * that the flush is complete. So we don't have to output an
           * empty block here, this will be done at next call. This also
           * ensures that for a very small output buffer, we emit at most
           * one empty block.
           */
        }
        if (bstate === BS_BLOCK_DONE) {
          if (flush === Z_PARTIAL_FLUSH$1) {
            _tr_align(s);
          } else if (flush !== Z_BLOCK$2) { /* FULL_FLUSH or SYNC_FLUSH */

            _tr_stored_block(s, 0, 0, false);
            /* For a full flush, this empty block will be recognized
             * as a special marker by inflate_sync().
             */
            if (flush === Z_FULL_FLUSH$1) {
              /*** CLEAR_HASH(s); ***/
              /* forget history */
              zero(s.head); // Fill with NIL (= 0);

              if (s.lookahead === 0) {
                s.strstart = 0;
                s.block_start = 0;
                s.insert = 0;
              }
            }
          }
          flush_pending(strm);
          if (strm.avail_out === 0) {
            s.last_flush = -1; /* avoid BUF_ERROR at next call, see above */
            return Z_OK$2;
          }
        }
      }
      //Assert(strm->avail_out > 0, "bug2");
      //if (strm.avail_out <= 0) { throw new Error("bug2");}

      if (flush !== Z_FINISH$2) {
        return Z_OK$2;
      }
      if (s.wrap <= 0) {
        return Z_STREAM_END$2;
      }

      /* Write the trailer */
      if (s.wrap === 2) {
        put_byte(s, strm.adler & 0xff);
        put_byte(s, (strm.adler >> 8) & 0xff);
        put_byte(s, (strm.adler >> 16) & 0xff);
        put_byte(s, (strm.adler >> 24) & 0xff);
        put_byte(s, strm.total_in & 0xff);
        put_byte(s, (strm.total_in >> 8) & 0xff);
        put_byte(s, (strm.total_in >> 16) & 0xff);
        put_byte(s, (strm.total_in >> 24) & 0xff);
      } else {
        putShortMSB(s, strm.adler >>> 16);
        putShortMSB(s, strm.adler & 0xffff);
      }

      flush_pending(strm);
      /* If avail_out is zero, the application will call deflate again
       * to flush the rest.
       */
      if (s.wrap > 0) {
        s.wrap = -s.wrap;
      }
      /* write the trailer only once! */
      return s.pending !== 0 ? Z_OK$2 : Z_STREAM_END$2;
    }

    function deflateEnd(strm) {
      var status;

      if (!strm /*== Z_NULL*/ || !strm.state /*== Z_NULL*/ ) {
        return Z_STREAM_ERROR$2;
      }

      status = strm.state.status;
      if (status !== INIT_STATE &&
        status !== EXTRA_STATE &&
        status !== NAME_STATE &&
        status !== COMMENT_STATE &&
        status !== HCRC_STATE &&
        status !== BUSY_STATE &&
        status !== FINISH_STATE
      ) {
        return err(strm, Z_STREAM_ERROR$2);
      }

      strm.state = null;

      return status === BUSY_STATE ? err(strm, Z_DATA_ERROR$2) : Z_OK$2;
    }

    /* Not implemented
    exports.deflateBound = deflateBound;
    exports.deflateCopy = deflateCopy;
    exports.deflateParams = deflateParams;
    exports.deflatePending = deflatePending;
    exports.deflatePrime = deflatePrime;
    exports.deflateTune = deflateTune;
    */

    // See state defs from inflate.js
    var BAD$1 = 30;       /* got a data error -- remain here until reset */
    var TYPE$1 = 12;      /* i: waiting for type bits, including last-flag bit */

    /*
       Decode literal, length, and distance codes and write out the resulting
       literal and match bytes until either not enough input or output is
       available, an end-of-block is encountered, or a data error is encountered.
       When large enough input and output buffers are supplied to inflate(), for
       example, a 16K input buffer and a 64K output buffer, more than 95% of the
       inflate execution time is spent in this routine.

       Entry assumptions:

            state.mode === LEN
            strm.avail_in >= 6
            strm.avail_out >= 258
            start >= strm.avail_out
            state.bits < 8

       On return, state.mode is one of:

            LEN -- ran out of enough output space or enough available input
            TYPE -- reached end of block code, inflate() to interpret next block
            BAD -- error in block data

       Notes:

        - The maximum input bits used by a length/distance pair is 15 bits for the
          length code, 5 bits for the length extra, 15 bits for the distance code,
          and 13 bits for the distance extra.  This totals 48 bits, or six bytes.
          Therefore if strm.avail_in >= 6, then there is enough input to avoid
          checking for available input while decoding.

        - The maximum bytes that a single length/distance pair can output is 258
          bytes, which is the maximum length that can be coded.  inflate_fast()
          requires strm.avail_out >= 258 for each loop to avoid checking for
          output space.
     */
    function inflate_fast(strm, start) {
      var state;
      var _in;                    /* local strm.input */
      var last;                   /* have enough input while in < last */
      var _out;                   /* local strm.output */
      var beg;                    /* inflate()'s initial strm.output */
      var end;                    /* while out < end, enough space available */
    //#ifdef INFLATE_STRICT
      var dmax;                   /* maximum distance from zlib header */
    //#endif
      var wsize;                  /* window size or zero if not using window */
      var whave;                  /* valid bytes in the window */
      var wnext;                  /* window write index */
      // Use `s_window` instead `window`, avoid conflict with instrumentation tools
      var s_window;               /* allocated sliding window, if wsize != 0 */
      var hold;                   /* local strm.hold */
      var bits;                   /* local strm.bits */
      var lcode;                  /* local strm.lencode */
      var dcode;                  /* local strm.distcode */
      var lmask;                  /* mask for first level of length codes */
      var dmask;                  /* mask for first level of distance codes */
      var here;                   /* retrieved table entry */
      var op;                     /* code bits, operation, extra bits, or */
                                  /*  window position, window bytes to copy */
      var len;                    /* match length, unused bytes */
      var dist;                   /* match distance */
      var from;                   /* where to copy match from */
      var from_source;


      var input, output; // JS specific, because we have no pointers

      /* copy state to local variables */
      state = strm.state;
      //here = state.here;
      _in = strm.next_in;
      input = strm.input;
      last = _in + (strm.avail_in - 5);
      _out = strm.next_out;
      output = strm.output;
      beg = _out - (start - strm.avail_out);
      end = _out + (strm.avail_out - 257);
    //#ifdef INFLATE_STRICT
      dmax = state.dmax;
    //#endif
      wsize = state.wsize;
      whave = state.whave;
      wnext = state.wnext;
      s_window = state.window;
      hold = state.hold;
      bits = state.bits;
      lcode = state.lencode;
      dcode = state.distcode;
      lmask = (1 << state.lenbits) - 1;
      dmask = (1 << state.distbits) - 1;


      /* decode literals and length/distances until end-of-block or not enough
         input data or output space */

      top:
      do {
        if (bits < 15) {
          hold += input[_in++] << bits;
          bits += 8;
          hold += input[_in++] << bits;
          bits += 8;
        }

        here = lcode[hold & lmask];

        dolen:
        for (;;) { // Goto emulation
          op = here >>> 24/*here.bits*/;
          hold >>>= op;
          bits -= op;
          op = (here >>> 16) & 0xff/*here.op*/;
          if (op === 0) {                          /* literal */
            //Tracevv((stderr, here.val >= 0x20 && here.val < 0x7f ?
            //        "inflate:         literal '%c'\n" :
            //        "inflate:         literal 0x%02x\n", here.val));
            output[_out++] = here & 0xffff/*here.val*/;
          }
          else if (op & 16) {                     /* length base */
            len = here & 0xffff/*here.val*/;
            op &= 15;                           /* number of extra bits */
            if (op) {
              if (bits < op) {
                hold += input[_in++] << bits;
                bits += 8;
              }
              len += hold & ((1 << op) - 1);
              hold >>>= op;
              bits -= op;
            }
            //Tracevv((stderr, "inflate:         length %u\n", len));
            if (bits < 15) {
              hold += input[_in++] << bits;
              bits += 8;
              hold += input[_in++] << bits;
              bits += 8;
            }
            here = dcode[hold & dmask];

            dodist:
            for (;;) { // goto emulation
              op = here >>> 24/*here.bits*/;
              hold >>>= op;
              bits -= op;
              op = (here >>> 16) & 0xff/*here.op*/;

              if (op & 16) {                      /* distance base */
                dist = here & 0xffff/*here.val*/;
                op &= 15;                       /* number of extra bits */
                if (bits < op) {
                  hold += input[_in++] << bits;
                  bits += 8;
                  if (bits < op) {
                    hold += input[_in++] << bits;
                    bits += 8;
                  }
                }
                dist += hold & ((1 << op) - 1);
    //#ifdef INFLATE_STRICT
                if (dist > dmax) {
                  strm.msg = 'invalid distance too far back';
                  state.mode = BAD$1;
                  break top;
                }
    //#endif
                hold >>>= op;
                bits -= op;
                //Tracevv((stderr, "inflate:         distance %u\n", dist));
                op = _out - beg;                /* max distance in output */
                if (dist > op) {                /* see if copy from window */
                  op = dist - op;               /* distance back in window */
                  if (op > whave) {
                    if (state.sane) {
                      strm.msg = 'invalid distance too far back';
                      state.mode = BAD$1;
                      break top;
                    }

    // (!) This block is disabled in zlib defailts,
    // don't enable it for binary compatibility
    //#ifdef INFLATE_ALLOW_INVALID_DISTANCE_TOOFAR_ARRR
    //                if (len <= op - whave) {
    //                  do {
    //                    output[_out++] = 0;
    //                  } while (--len);
    //                  continue top;
    //                }
    //                len -= op - whave;
    //                do {
    //                  output[_out++] = 0;
    //                } while (--op > whave);
    //                if (op === 0) {
    //                  from = _out - dist;
    //                  do {
    //                    output[_out++] = output[from++];
    //                  } while (--len);
    //                  continue top;
    //                }
    //#endif
                  }
                  from = 0; // window index
                  from_source = s_window;
                  if (wnext === 0) {           /* very common case */
                    from += wsize - op;
                    if (op < len) {         /* some from window */
                      len -= op;
                      do {
                        output[_out++] = s_window[from++];
                      } while (--op);
                      from = _out - dist;  /* rest from output */
                      from_source = output;
                    }
                  }
                  else if (wnext < op) {      /* wrap around window */
                    from += wsize + wnext - op;
                    op -= wnext;
                    if (op < len) {         /* some from end of window */
                      len -= op;
                      do {
                        output[_out++] = s_window[from++];
                      } while (--op);
                      from = 0;
                      if (wnext < len) {  /* some from start of window */
                        op = wnext;
                        len -= op;
                        do {
                          output[_out++] = s_window[from++];
                        } while (--op);
                        from = _out - dist;      /* rest from output */
                        from_source = output;
                      }
                    }
                  }
                  else {                      /* contiguous in window */
                    from += wnext - op;
                    if (op < len) {         /* some from window */
                      len -= op;
                      do {
                        output[_out++] = s_window[from++];
                      } while (--op);
                      from = _out - dist;  /* rest from output */
                      from_source = output;
                    }
                  }
                  while (len > 2) {
                    output[_out++] = from_source[from++];
                    output[_out++] = from_source[from++];
                    output[_out++] = from_source[from++];
                    len -= 3;
                  }
                  if (len) {
                    output[_out++] = from_source[from++];
                    if (len > 1) {
                      output[_out++] = from_source[from++];
                    }
                  }
                }
                else {
                  from = _out - dist;          /* copy direct from output */
                  do {                        /* minimum length is three */
                    output[_out++] = output[from++];
                    output[_out++] = output[from++];
                    output[_out++] = output[from++];
                    len -= 3;
                  } while (len > 2);
                  if (len) {
                    output[_out++] = output[from++];
                    if (len > 1) {
                      output[_out++] = output[from++];
                    }
                  }
                }
              }
              else if ((op & 64) === 0) {          /* 2nd level distance code */
                here = dcode[(here & 0xffff)/*here.val*/ + (hold & ((1 << op) - 1))];
                continue dodist;
              }
              else {
                strm.msg = 'invalid distance code';
                state.mode = BAD$1;
                break top;
              }

              break; // need to emulate goto via "continue"
            }
          }
          else if ((op & 64) === 0) {              /* 2nd level length code */
            here = lcode[(here & 0xffff)/*here.val*/ + (hold & ((1 << op) - 1))];
            continue dolen;
          }
          else if (op & 32) {                     /* end-of-block */
            //Tracevv((stderr, "inflate:         end of block\n"));
            state.mode = TYPE$1;
            break top;
          }
          else {
            strm.msg = 'invalid literal/length code';
            state.mode = BAD$1;
            break top;
          }

          break; // need to emulate goto via "continue"
        }
      } while (_in < last && _out < end);

      /* return unused bytes (on entry, bits < 8, so in won't go too far back) */
      len = bits >> 3;
      _in -= len;
      bits -= len << 3;
      hold &= (1 << bits) - 1;

      /* update state and return */
      strm.next_in = _in;
      strm.next_out = _out;
      strm.avail_in = (_in < last ? 5 + (last - _in) : 5 - (_in - last));
      strm.avail_out = (_out < end ? 257 + (end - _out) : 257 - (_out - end));
      state.hold = hold;
      state.bits = bits;
      return;
    }

    var MAXBITS = 15;
    var ENOUGH_LENS$1 = 852;
    var ENOUGH_DISTS$1 = 592;
    //var ENOUGH = (ENOUGH_LENS+ENOUGH_DISTS);

    var CODES$1 = 0;
    var LENS$1 = 1;
    var DISTS$1 = 2;

    var lbase = [ /* Length codes 257..285 base */
      3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
      35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0, 0
    ];

    var lext = [ /* Length codes 257..285 extra */
      16, 16, 16, 16, 16, 16, 16, 16, 17, 17, 17, 17, 18, 18, 18, 18,
      19, 19, 19, 19, 20, 20, 20, 20, 21, 21, 21, 21, 16, 72, 78
    ];

    var dbase = [ /* Distance codes 0..29 base */
      1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
      257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
      8193, 12289, 16385, 24577, 0, 0
    ];

    var dext = [ /* Distance codes 0..29 extra */
      16, 16, 16, 16, 17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 22,
      23, 23, 24, 24, 25, 25, 26, 26, 27, 27,
      28, 28, 29, 29, 64, 64
    ];

    function inflate_table(type, lens, lens_index, codes, table, table_index, work, opts) {
      var bits = opts.bits;
      //here = opts.here; /* table entry for duplication */

      var len = 0; /* a code's length in bits */
      var sym = 0; /* index of code symbols */
      var min = 0,
        max = 0; /* minimum and maximum code lengths */
      var root = 0; /* number of index bits for root table */
      var curr = 0; /* number of index bits for current table */
      var drop = 0; /* code bits to drop for sub-table */
      var left = 0; /* number of prefix codes available */
      var used = 0; /* code entries in table used */
      var huff = 0; /* Huffman code */
      var incr; /* for incrementing code, index */
      var fill; /* index for replicating entries */
      var low; /* low bits for current root entry */
      var mask; /* mask for low root bits */
      var next; /* next available space in table */
      var base = null; /* base value table to use */
      var base_index = 0;
      //  var shoextra;    /* extra bits table to use */
      var end; /* use base and extra for symbol > end */
      var count = new Buf16(MAXBITS + 1); //[MAXBITS+1];    /* number of codes of each length */
      var offs = new Buf16(MAXBITS + 1); //[MAXBITS+1];     /* offsets in table for each length */
      var extra = null;
      var extra_index = 0;

      var here_bits, here_op, here_val;

      /*
       Process a set of code lengths to create a canonical Huffman code.  The
       code lengths are lens[0..codes-1].  Each length corresponds to the
       symbols 0..codes-1.  The Huffman code is generated by first sorting the
       symbols by length from short to long, and retaining the symbol order
       for codes with equal lengths.  Then the code starts with all zero bits
       for the first code of the shortest length, and the codes are integer
       increments for the same length, and zeros are appended as the length
       increases.  For the deflate format, these bits are stored backwards
       from their more natural integer increment ordering, and so when the
       decoding tables are built in the large loop below, the integer codes
       are incremented backwards.

       This routine assumes, but does not check, that all of the entries in
       lens[] are in the range 0..MAXBITS.  The caller must assure this.
       1..MAXBITS is interpreted as that code length.  zero means that that
       symbol does not occur in this code.

       The codes are sorted by computing a count of codes for each length,
       creating from that a table of starting indices for each length in the
       sorted table, and then entering the symbols in order in the sorted
       table.  The sorted table is work[], with that space being provided by
       the caller.

       The length counts are used for other purposes as well, i.e. finding
       the minimum and maximum length codes, determining if there are any
       codes at all, checking for a valid set of lengths, and looking ahead
       at length counts to determine sub-table sizes when building the
       decoding tables.
       */

      /* accumulate lengths for codes (assumes lens[] all in 0..MAXBITS) */
      for (len = 0; len <= MAXBITS; len++) {
        count[len] = 0;
      }
      for (sym = 0; sym < codes; sym++) {
        count[lens[lens_index + sym]]++;
      }

      /* bound code lengths, force root to be within code lengths */
      root = bits;
      for (max = MAXBITS; max >= 1; max--) {
        if (count[max] !== 0) {
          break;
        }
      }
      if (root > max) {
        root = max;
      }
      if (max === 0) { /* no symbols to code at all */
        //table.op[opts.table_index] = 64;  //here.op = (var char)64;    /* invalid code marker */
        //table.bits[opts.table_index] = 1;   //here.bits = (var char)1;
        //table.val[opts.table_index++] = 0;   //here.val = (var short)0;
        table[table_index++] = (1 << 24) | (64 << 16) | 0;


        //table.op[opts.table_index] = 64;
        //table.bits[opts.table_index] = 1;
        //table.val[opts.table_index++] = 0;
        table[table_index++] = (1 << 24) | (64 << 16) | 0;

        opts.bits = 1;
        return 0; /* no symbols, but wait for decoding to report error */
      }
      for (min = 1; min < max; min++) {
        if (count[min] !== 0) {
          break;
        }
      }
      if (root < min) {
        root = min;
      }

      /* check for an over-subscribed or incomplete set of lengths */
      left = 1;
      for (len = 1; len <= MAXBITS; len++) {
        left <<= 1;
        left -= count[len];
        if (left < 0) {
          return -1;
        } /* over-subscribed */
      }
      if (left > 0 && (type === CODES$1 || max !== 1)) {
        return -1; /* incomplete set */
      }

      /* generate offsets into symbol table for each length for sorting */
      offs[1] = 0;
      for (len = 1; len < MAXBITS; len++) {
        offs[len + 1] = offs[len] + count[len];
      }

      /* sort symbols by length, by symbol order within each length */
      for (sym = 0; sym < codes; sym++) {
        if (lens[lens_index + sym] !== 0) {
          work[offs[lens[lens_index + sym]]++] = sym;
        }
      }

      /*
       Create and fill in decoding tables.  In this loop, the table being
       filled is at next and has curr index bits.  The code being used is huff
       with length len.  That code is converted to an index by dropping drop
       bits off of the bottom.  For codes where len is less than drop + curr,
       those top drop + curr - len bits are incremented through all values to
       fill the table with replicated entries.

       root is the number of index bits for the root table.  When len exceeds
       root, sub-tables are created pointed to by the root entry with an index
       of the low root bits of huff.  This is saved in low to check for when a
       new sub-table should be started.  drop is zero when the root table is
       being filled, and drop is root when sub-tables are being filled.

       When a new sub-table is needed, it is necessary to look ahead in the
       code lengths to determine what size sub-table is needed.  The length
       counts are used for this, and so count[] is decremented as codes are
       entered in the tables.

       used keeps track of how many table entries have been allocated from the
       provided *table space.  It is checked for LENS and DIST tables against
       the constants ENOUGH_LENS and ENOUGH_DISTS to guard against changes in
       the initial root table size constants.  See the comments in inftrees.h
       for more information.

       sym increments through all symbols, and the loop terminates when
       all codes of length max, i.e. all codes, have been processed.  This
       routine permits incomplete codes, so another loop after this one fills
       in the rest of the decoding tables with invalid code markers.
       */

      /* set up for code type */
      // poor man optimization - use if-else instead of switch,
      // to avoid deopts in old v8
      if (type === CODES$1) {
        base = extra = work; /* dummy value--not used */
        end = 19;

      } else if (type === LENS$1) {
        base = lbase;
        base_index -= 257;
        extra = lext;
        extra_index -= 257;
        end = 256;

      } else { /* DISTS */
        base = dbase;
        extra = dext;
        end = -1;
      }

      /* initialize opts for loop */
      huff = 0; /* starting code */
      sym = 0; /* starting code symbol */
      len = min; /* starting code length */
      next = table_index; /* current table to fill in */
      curr = root; /* current table index bits */
      drop = 0; /* current bits to drop from code for index */
      low = -1; /* trigger new sub-table when len > root */
      used = 1 << root; /* use root table entries */
      mask = used - 1; /* mask for comparing low */

      /* check available table space */
      if ((type === LENS$1 && used > ENOUGH_LENS$1) ||
        (type === DISTS$1 && used > ENOUGH_DISTS$1)) {
        return 1;
      }
      /* process all codes and make table entries */
      for (;;) {
        /* create table entry */
        here_bits = len - drop;
        if (work[sym] < end) {
          here_op = 0;
          here_val = work[sym];
        } else if (work[sym] > end) {
          here_op = extra[extra_index + work[sym]];
          here_val = base[base_index + work[sym]];
        } else {
          here_op = 32 + 64; /* end of block */
          here_val = 0;
        }

        /* replicate for those indices with low len bits equal to huff */
        incr = 1 << (len - drop);
        fill = 1 << curr;
        min = fill; /* save offset to next table */
        do {
          fill -= incr;
          table[next + (huff >> drop) + fill] = (here_bits << 24) | (here_op << 16) | here_val | 0;
        } while (fill !== 0);

        /* backwards increment the len-bit code huff */
        incr = 1 << (len - 1);
        while (huff & incr) {
          incr >>= 1;
        }
        if (incr !== 0) {
          huff &= incr - 1;
          huff += incr;
        } else {
          huff = 0;
        }

        /* go to next symbol, update count, len */
        sym++;
        if (--count[len] === 0) {
          if (len === max) {
            break;
          }
          len = lens[lens_index + work[sym]];
        }

        /* create new sub-table if needed */
        if (len > root && (huff & mask) !== low) {
          /* if first time, transition to sub-tables */
          if (drop === 0) {
            drop = root;
          }

          /* increment past last table */
          next += min; /* here min is 1 << curr */

          /* determine length of next table */
          curr = len - drop;
          left = 1 << curr;
          while (curr + drop < max) {
            left -= count[curr + drop];
            if (left <= 0) {
              break;
            }
            curr++;
            left <<= 1;
          }

          /* check for enough space */
          used += 1 << curr;
          if ((type === LENS$1 && used > ENOUGH_LENS$1) ||
            (type === DISTS$1 && used > ENOUGH_DISTS$1)) {
            return 1;
          }

          /* point entry in root table to sub-table */
          low = huff & mask;
          /*table.op[low] = curr;
          table.bits[low] = root;
          table.val[low] = next - opts.table_index;*/
          table[low] = (root << 24) | (curr << 16) | (next - table_index) | 0;
        }
      }

      /* fill in remaining table entry if code is incomplete (guaranteed to have
       at most one remaining entry, since if the code is incomplete, the
       maximum code length that was allowed to get this far is one bit) */
      if (huff !== 0) {
        //table.op[next + huff] = 64;            /* invalid code marker */
        //table.bits[next + huff] = len - drop;
        //table.val[next + huff] = 0;
        table[next + huff] = ((len - drop) << 24) | (64 << 16) | 0;
      }

      /* set return parameters */
      //opts.table_index += used;
      opts.bits = root;
      return 0;
    }

    var CODES = 0;
    var LENS = 1;
    var DISTS = 2;

    /* Public constants ==========================================================*/
    /* ===========================================================================*/


    /* Allowed flush values; see deflate() and inflate() below for details */
    //var Z_NO_FLUSH      = 0;
    //var Z_PARTIAL_FLUSH = 1;
    //var Z_SYNC_FLUSH    = 2;
    //var Z_FULL_FLUSH    = 3;
    var Z_FINISH$1 = 4;
    var Z_BLOCK$1 = 5;
    var Z_TREES$1 = 6;


    /* Return codes for the compression/decompression functions. Negative values
     * are errors, positive values are used for special but normal events.
     */
    var Z_OK$1 = 0;
    var Z_STREAM_END$1 = 1;
    var Z_NEED_DICT$1 = 2;
    //var Z_ERRNO         = -1;
    var Z_STREAM_ERROR$1 = -2;
    var Z_DATA_ERROR$1 = -3;
    var Z_MEM_ERROR = -4;
    var Z_BUF_ERROR$1 = -5;
    //var Z_VERSION_ERROR = -6;

    /* The deflate compression method */
    var Z_DEFLATED$1 = 8;


    /* STATES ====================================================================*/
    /* ===========================================================================*/


    var HEAD = 1; /* i: waiting for magic header */
    var FLAGS = 2; /* i: waiting for method and flags (gzip) */
    var TIME = 3; /* i: waiting for modification time (gzip) */
    var OS = 4; /* i: waiting for extra flags and operating system (gzip) */
    var EXLEN = 5; /* i: waiting for extra length (gzip) */
    var EXTRA = 6; /* i: waiting for extra bytes (gzip) */
    var NAME = 7; /* i: waiting for end of file name (gzip) */
    var COMMENT = 8; /* i: waiting for end of comment (gzip) */
    var HCRC = 9; /* i: waiting for header crc (gzip) */
    var DICTID = 10; /* i: waiting for dictionary check value */
    var DICT = 11; /* waiting for inflateSetDictionary() call */
    var TYPE = 12; /* i: waiting for type bits, including last-flag bit */
    var TYPEDO = 13; /* i: same, but skip check to exit inflate on new block */
    var STORED = 14; /* i: waiting for stored size (length and complement) */
    var COPY_ = 15; /* i/o: same as COPY below, but only first time in */
    var COPY = 16; /* i/o: waiting for input or output to copy stored block */
    var TABLE = 17; /* i: waiting for dynamic block table lengths */
    var LENLENS = 18; /* i: waiting for code length code lengths */
    var CODELENS = 19; /* i: waiting for length/lit and distance code lengths */
    var LEN_ = 20; /* i: same as LEN below, but only first time in */
    var LEN = 21; /* i: waiting for length/lit/eob code */
    var LENEXT = 22; /* i: waiting for length extra bits */
    var DIST = 23; /* i: waiting for distance code */
    var DISTEXT = 24; /* i: waiting for distance extra bits */
    var MATCH = 25; /* o: waiting for output space to copy string */
    var LIT = 26; /* o: waiting for output space to write literal */
    var CHECK = 27; /* i: waiting for 32-bit check value */
    var LENGTH = 28; /* i: waiting for 32-bit length (gzip) */
    var DONE = 29; /* finished check, done -- remain here until reset */
    var BAD = 30; /* got a data error -- remain here until reset */
    var MEM = 31; /* got an inflate() memory error -- remain here until reset */
    var SYNC = 32; /* looking for synchronization bytes to restart inflate() */

    /* ===========================================================================*/



    var ENOUGH_LENS = 852;
    var ENOUGH_DISTS = 592;


    function zswap32(q) {
      return (((q >>> 24) & 0xff) +
        ((q >>> 8) & 0xff00) +
        ((q & 0xff00) << 8) +
        ((q & 0xff) << 24));
    }


    function InflateState() {
      this.mode = 0; /* current inflate mode */
      this.last = false; /* true if processing last block */
      this.wrap = 0; /* bit 0 true for zlib, bit 1 true for gzip */
      this.havedict = false; /* true if dictionary provided */
      this.flags = 0; /* gzip header method and flags (0 if zlib) */
      this.dmax = 0; /* zlib header max distance (INFLATE_STRICT) */
      this.check = 0; /* protected copy of check value */
      this.total = 0; /* protected copy of output count */
      // TODO: may be {}
      this.head = null; /* where to save gzip header information */

      /* sliding window */
      this.wbits = 0; /* log base 2 of requested window size */
      this.wsize = 0; /* window size or zero if not using window */
      this.whave = 0; /* valid bytes in the window */
      this.wnext = 0; /* window write index */
      this.window = null; /* allocated sliding window, if needed */

      /* bit accumulator */
      this.hold = 0; /* input bit accumulator */
      this.bits = 0; /* number of bits in "in" */

      /* for string and stored block copying */
      this.length = 0; /* literal or length of data to copy */
      this.offset = 0; /* distance back to copy string from */

      /* for table and code decoding */
      this.extra = 0; /* extra bits needed */

      /* fixed and dynamic code tables */
      this.lencode = null; /* starting table for length/literal codes */
      this.distcode = null; /* starting table for distance codes */
      this.lenbits = 0; /* index bits for lencode */
      this.distbits = 0; /* index bits for distcode */

      /* dynamic table building */
      this.ncode = 0; /* number of code length code lengths */
      this.nlen = 0; /* number of length code lengths */
      this.ndist = 0; /* number of distance code lengths */
      this.have = 0; /* number of code lengths in lens[] */
      this.next = null; /* next available space in codes[] */

      this.lens = new Buf16(320); /* temporary storage for code lengths */
      this.work = new Buf16(288); /* work area for code table building */

      /*
       because we don't have pointers in js, we use lencode and distcode directly
       as buffers so we don't need codes
      */
      //this.codes = new Buf32(ENOUGH);       /* space for code tables */
      this.lendyn = null; /* dynamic table for length/literal codes (JS specific) */
      this.distdyn = null; /* dynamic table for distance codes (JS specific) */
      this.sane = 0; /* if false, allow invalid distance too far */
      this.back = 0; /* bits back of last unprocessed length/lit */
      this.was = 0; /* initial length of match */
    }

    function inflateResetKeep(strm) {
      var state;

      if (!strm || !strm.state) {
        return Z_STREAM_ERROR$1;
      }
      state = strm.state;
      strm.total_in = strm.total_out = state.total = 0;
      strm.msg = ''; /*Z_NULL*/
      if (state.wrap) { /* to support ill-conceived Java test suite */
        strm.adler = state.wrap & 1;
      }
      state.mode = HEAD;
      state.last = 0;
      state.havedict = 0;
      state.dmax = 32768;
      state.head = null /*Z_NULL*/ ;
      state.hold = 0;
      state.bits = 0;
      //state.lencode = state.distcode = state.next = state.codes;
      state.lencode = state.lendyn = new Buf32(ENOUGH_LENS);
      state.distcode = state.distdyn = new Buf32(ENOUGH_DISTS);

      state.sane = 1;
      state.back = -1;
      //Tracev((stderr, "inflate: reset\n"));
      return Z_OK$1;
    }

    function inflateReset(strm) {
      var state;

      if (!strm || !strm.state) {
        return Z_STREAM_ERROR$1;
      }
      state = strm.state;
      state.wsize = 0;
      state.whave = 0;
      state.wnext = 0;
      return inflateResetKeep(strm);

    }

    function inflateReset2(strm, windowBits) {
      var wrap;
      var state;

      /* get the state */
      if (!strm || !strm.state) {
        return Z_STREAM_ERROR$1;
      }
      state = strm.state;

      /* extract wrap request from windowBits parameter */
      if (windowBits < 0) {
        wrap = 0;
        windowBits = -windowBits;
      } else {
        wrap = (windowBits >> 4) + 1;
        if (windowBits < 48) {
          windowBits &= 15;
        }
      }

      /* set number of window bits, free window if different */
      if (windowBits && (windowBits < 8 || windowBits > 15)) {
        return Z_STREAM_ERROR$1;
      }
      if (state.window !== null && state.wbits !== windowBits) {
        state.window = null;
      }

      /* update state and reset the rest of it */
      state.wrap = wrap;
      state.wbits = windowBits;
      return inflateReset(strm);
    }

    function inflateInit2(strm, windowBits) {
      var ret;
      var state;

      if (!strm) {
        return Z_STREAM_ERROR$1;
      }
      //strm.msg = Z_NULL;                 /* in case we return an error */

      state = new InflateState();

      //if (state === Z_NULL) return Z_MEM_ERROR;
      //Tracev((stderr, "inflate: allocated\n"));
      strm.state = state;
      state.window = null /*Z_NULL*/ ;
      ret = inflateReset2(strm, windowBits);
      if (ret !== Z_OK$1) {
        strm.state = null /*Z_NULL*/ ;
      }
      return ret;
    }


    /*
     Return state with length and distance decoding tables and index sizes set to
     fixed code decoding.  Normally this returns fixed tables from inffixed.h.
     If BUILDFIXED is defined, then instead this routine builds the tables the
     first time it's called, and returns those tables the first time and
     thereafter.  This reduces the size of the code by about 2K bytes, in
     exchange for a little execution time.  However, BUILDFIXED should not be
     used for threaded applications, since the rewriting of the tables and virgin
     may not be thread-safe.
     */
    var virgin = true;

    var lenfix, distfix; // We have no pointers in JS, so keep tables separate

    function fixedtables(state) {
      /* build fixed huffman tables if first call (may not be thread safe) */
      if (virgin) {
        var sym;

        lenfix = new Buf32(512);
        distfix = new Buf32(32);

        /* literal/length table */
        sym = 0;
        while (sym < 144) {
          state.lens[sym++] = 8;
        }
        while (sym < 256) {
          state.lens[sym++] = 9;
        }
        while (sym < 280) {
          state.lens[sym++] = 7;
        }
        while (sym < 288) {
          state.lens[sym++] = 8;
        }

        inflate_table(LENS, state.lens, 0, 288, lenfix, 0, state.work, {
          bits: 9
        });

        /* distance table */
        sym = 0;
        while (sym < 32) {
          state.lens[sym++] = 5;
        }

        inflate_table(DISTS, state.lens, 0, 32, distfix, 0, state.work, {
          bits: 5
        });

        /* do this just once */
        virgin = false;
      }

      state.lencode = lenfix;
      state.lenbits = 9;
      state.distcode = distfix;
      state.distbits = 5;
    }


    /*
     Update the window with the last wsize (normally 32K) bytes written before
     returning.  If window does not exist yet, create it.  This is only called
     when a window is already in use, or when output has been written during this
     inflate call, but the end of the deflate stream has not been reached yet.
     It is also called to create a window for dictionary data when a dictionary
     is loaded.

     Providing output buffers larger than 32K to inflate() should provide a speed
     advantage, since only the last 32K of output is copied to the sliding window
     upon return from inflate(), and since all distances after the first 32K of
     output will fall in the output data, making match copies simpler and faster.
     The advantage may be dependent on the size of the processor's data caches.
     */
    function updatewindow(strm, src, end, copy) {
      var dist;
      var state = strm.state;

      /* if it hasn't been done already, allocate space for the window */
      if (state.window === null) {
        state.wsize = 1 << state.wbits;
        state.wnext = 0;
        state.whave = 0;

        state.window = new Buf8(state.wsize);
      }

      /* copy state->wsize or less output bytes into the circular window */
      if (copy >= state.wsize) {
        arraySet(state.window, src, end - state.wsize, state.wsize, 0);
        state.wnext = 0;
        state.whave = state.wsize;
      } else {
        dist = state.wsize - state.wnext;
        if (dist > copy) {
          dist = copy;
        }
        //zmemcpy(state->window + state->wnext, end - copy, dist);
        arraySet(state.window, src, end - copy, dist, state.wnext);
        copy -= dist;
        if (copy) {
          //zmemcpy(state->window, end - copy, copy);
          arraySet(state.window, src, end - copy, copy, 0);
          state.wnext = copy;
          state.whave = state.wsize;
        } else {
          state.wnext += dist;
          if (state.wnext === state.wsize) {
            state.wnext = 0;
          }
          if (state.whave < state.wsize) {
            state.whave += dist;
          }
        }
      }
      return 0;
    }

    function inflate$1(strm, flush) {
      var state;
      var input, output; // input/output buffers
      var next; /* next input INDEX */
      var put; /* next output INDEX */
      var have, left; /* available input and output */
      var hold; /* bit buffer */
      var bits; /* bits in bit buffer */
      var _in, _out; /* save starting available input and output */
      var copy; /* number of stored or match bytes to copy */
      var from; /* where to copy match bytes from */
      var from_source;
      var here = 0; /* current decoding table entry */
      var here_bits, here_op, here_val; // paked "here" denormalized (JS specific)
      //var last;                   /* parent table entry */
      var last_bits, last_op, last_val; // paked "last" denormalized (JS specific)
      var len; /* length to copy for repeats, bits to drop */
      var ret; /* return code */
      var hbuf = new Buf8(4); /* buffer for gzip header crc calculation */
      var opts;

      var n; // temporary var for NEED_BITS

      var order = /* permutation of code lengths */ [16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15];


      if (!strm || !strm.state || !strm.output ||
        (!strm.input && strm.avail_in !== 0)) {
        return Z_STREAM_ERROR$1;
      }

      state = strm.state;
      if (state.mode === TYPE) {
        state.mode = TYPEDO;
      } /* skip check */


      //--- LOAD() ---
      put = strm.next_out;
      output = strm.output;
      left = strm.avail_out;
      next = strm.next_in;
      input = strm.input;
      have = strm.avail_in;
      hold = state.hold;
      bits = state.bits;
      //---

      _in = have;
      _out = left;
      ret = Z_OK$1;

      inf_leave: // goto emulation
        for (;;) {
          switch (state.mode) {
          case HEAD:
            if (state.wrap === 0) {
              state.mode = TYPEDO;
              break;
            }
            //=== NEEDBITS(16);
            while (bits < 16) {
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            if ((state.wrap & 2) && hold === 0x8b1f) { /* gzip header */
              state.check = 0 /*crc32(0L, Z_NULL, 0)*/ ;
              //=== CRC2(state.check, hold);
              hbuf[0] = hold & 0xff;
              hbuf[1] = (hold >>> 8) & 0xff;
              state.check = crc32(state.check, hbuf, 2, 0);
              //===//

              //=== INITBITS();
              hold = 0;
              bits = 0;
              //===//
              state.mode = FLAGS;
              break;
            }
            state.flags = 0; /* expect zlib header */
            if (state.head) {
              state.head.done = false;
            }
            if (!(state.wrap & 1) || /* check if zlib header allowed */
              (((hold & 0xff) /*BITS(8)*/ << 8) + (hold >> 8)) % 31) {
              strm.msg = 'incorrect header check';
              state.mode = BAD;
              break;
            }
            if ((hold & 0x0f) /*BITS(4)*/ !== Z_DEFLATED$1) {
              strm.msg = 'unknown compression method';
              state.mode = BAD;
              break;
            }
            //--- DROPBITS(4) ---//
            hold >>>= 4;
            bits -= 4;
            //---//
            len = (hold & 0x0f) /*BITS(4)*/ + 8;
            if (state.wbits === 0) {
              state.wbits = len;
            } else if (len > state.wbits) {
              strm.msg = 'invalid window size';
              state.mode = BAD;
              break;
            }
            state.dmax = 1 << len;
            //Tracev((stderr, "inflate:   zlib header ok\n"));
            strm.adler = state.check = 1 /*adler32(0L, Z_NULL, 0)*/ ;
            state.mode = hold & 0x200 ? DICTID : TYPE;
            //=== INITBITS();
            hold = 0;
            bits = 0;
            //===//
            break;
          case FLAGS:
            //=== NEEDBITS(16); */
            while (bits < 16) {
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            state.flags = hold;
            if ((state.flags & 0xff) !== Z_DEFLATED$1) {
              strm.msg = 'unknown compression method';
              state.mode = BAD;
              break;
            }
            if (state.flags & 0xe000) {
              strm.msg = 'unknown header flags set';
              state.mode = BAD;
              break;
            }
            if (state.head) {
              state.head.text = ((hold >> 8) & 1);
            }
            if (state.flags & 0x0200) {
              //=== CRC2(state.check, hold);
              hbuf[0] = hold & 0xff;
              hbuf[1] = (hold >>> 8) & 0xff;
              state.check = crc32(state.check, hbuf, 2, 0);
              //===//
            }
            //=== INITBITS();
            hold = 0;
            bits = 0;
            //===//
            state.mode = TIME;
            /* falls through */
          case TIME:
            //=== NEEDBITS(32); */
            while (bits < 32) {
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            if (state.head) {
              state.head.time = hold;
            }
            if (state.flags & 0x0200) {
              //=== CRC4(state.check, hold)
              hbuf[0] = hold & 0xff;
              hbuf[1] = (hold >>> 8) & 0xff;
              hbuf[2] = (hold >>> 16) & 0xff;
              hbuf[3] = (hold >>> 24) & 0xff;
              state.check = crc32(state.check, hbuf, 4, 0);
              //===
            }
            //=== INITBITS();
            hold = 0;
            bits = 0;
            //===//
            state.mode = OS;
            /* falls through */
          case OS:
            //=== NEEDBITS(16); */
            while (bits < 16) {
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            if (state.head) {
              state.head.xflags = (hold & 0xff);
              state.head.os = (hold >> 8);
            }
            if (state.flags & 0x0200) {
              //=== CRC2(state.check, hold);
              hbuf[0] = hold & 0xff;
              hbuf[1] = (hold >>> 8) & 0xff;
              state.check = crc32(state.check, hbuf, 2, 0);
              //===//
            }
            //=== INITBITS();
            hold = 0;
            bits = 0;
            //===//
            state.mode = EXLEN;
            /* falls through */
          case EXLEN:
            if (state.flags & 0x0400) {
              //=== NEEDBITS(16); */
              while (bits < 16) {
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                hold += input[next++] << bits;
                bits += 8;
              }
              //===//
              state.length = hold;
              if (state.head) {
                state.head.extra_len = hold;
              }
              if (state.flags & 0x0200) {
                //=== CRC2(state.check, hold);
                hbuf[0] = hold & 0xff;
                hbuf[1] = (hold >>> 8) & 0xff;
                state.check = crc32(state.check, hbuf, 2, 0);
                //===//
              }
              //=== INITBITS();
              hold = 0;
              bits = 0;
              //===//
            } else if (state.head) {
              state.head.extra = null /*Z_NULL*/ ;
            }
            state.mode = EXTRA;
            /* falls through */
          case EXTRA:
            if (state.flags & 0x0400) {
              copy = state.length;
              if (copy > have) {
                copy = have;
              }
              if (copy) {
                if (state.head) {
                  len = state.head.extra_len - state.length;
                  if (!state.head.extra) {
                    // Use untyped array for more conveniend processing later
                    state.head.extra = new Array(state.head.extra_len);
                  }
                  arraySet(
                    state.head.extra,
                    input,
                    next,
                    // extra field is limited to 65536 bytes
                    // - no need for additional size check
                    copy,
                    /*len + copy > state.head.extra_max - len ? state.head.extra_max : copy,*/
                    len
                  );
                  //zmemcpy(state.head.extra + len, next,
                  //        len + copy > state.head.extra_max ?
                  //        state.head.extra_max - len : copy);
                }
                if (state.flags & 0x0200) {
                  state.check = crc32(state.check, input, copy, next);
                }
                have -= copy;
                next += copy;
                state.length -= copy;
              }
              if (state.length) {
                break inf_leave;
              }
            }
            state.length = 0;
            state.mode = NAME;
            /* falls through */
          case NAME:
            if (state.flags & 0x0800) {
              if (have === 0) {
                break inf_leave;
              }
              copy = 0;
              do {
                // TODO: 2 or 1 bytes?
                len = input[next + copy++];
                /* use constant limit because in js we should not preallocate memory */
                if (state.head && len &&
                  (state.length < 65536 /*state.head.name_max*/ )) {
                  state.head.name += String.fromCharCode(len);
                }
              } while (len && copy < have);

              if (state.flags & 0x0200) {
                state.check = crc32(state.check, input, copy, next);
              }
              have -= copy;
              next += copy;
              if (len) {
                break inf_leave;
              }
            } else if (state.head) {
              state.head.name = null;
            }
            state.length = 0;
            state.mode = COMMENT;
            /* falls through */
          case COMMENT:
            if (state.flags & 0x1000) {
              if (have === 0) {
                break inf_leave;
              }
              copy = 0;
              do {
                len = input[next + copy++];
                /* use constant limit because in js we should not preallocate memory */
                if (state.head && len &&
                  (state.length < 65536 /*state.head.comm_max*/ )) {
                  state.head.comment += String.fromCharCode(len);
                }
              } while (len && copy < have);
              if (state.flags & 0x0200) {
                state.check = crc32(state.check, input, copy, next);
              }
              have -= copy;
              next += copy;
              if (len) {
                break inf_leave;
              }
            } else if (state.head) {
              state.head.comment = null;
            }
            state.mode = HCRC;
            /* falls through */
          case HCRC:
            if (state.flags & 0x0200) {
              //=== NEEDBITS(16); */
              while (bits < 16) {
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                hold += input[next++] << bits;
                bits += 8;
              }
              //===//
              if (hold !== (state.check & 0xffff)) {
                strm.msg = 'header crc mismatch';
                state.mode = BAD;
                break;
              }
              //=== INITBITS();
              hold = 0;
              bits = 0;
              //===//
            }
            if (state.head) {
              state.head.hcrc = ((state.flags >> 9) & 1);
              state.head.done = true;
            }
            strm.adler = state.check = 0;
            state.mode = TYPE;
            break;
          case DICTID:
            //=== NEEDBITS(32); */
            while (bits < 32) {
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            strm.adler = state.check = zswap32(hold);
            //=== INITBITS();
            hold = 0;
            bits = 0;
            //===//
            state.mode = DICT;
            /* falls through */
          case DICT:
            if (state.havedict === 0) {
              //--- RESTORE() ---
              strm.next_out = put;
              strm.avail_out = left;
              strm.next_in = next;
              strm.avail_in = have;
              state.hold = hold;
              state.bits = bits;
              //---
              return Z_NEED_DICT$1;
            }
            strm.adler = state.check = 1 /*adler32(0L, Z_NULL, 0)*/ ;
            state.mode = TYPE;
            /* falls through */
          case TYPE:
            if (flush === Z_BLOCK$1 || flush === Z_TREES$1) {
              break inf_leave;
            }
            /* falls through */
          case TYPEDO:
            if (state.last) {
              //--- BYTEBITS() ---//
              hold >>>= bits & 7;
              bits -= bits & 7;
              //---//
              state.mode = CHECK;
              break;
            }
            //=== NEEDBITS(3); */
            while (bits < 3) {
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            state.last = (hold & 0x01) /*BITS(1)*/ ;
            //--- DROPBITS(1) ---//
            hold >>>= 1;
            bits -= 1;
            //---//

            switch ((hold & 0x03) /*BITS(2)*/ ) {
            case 0:
              /* stored block */
              //Tracev((stderr, "inflate:     stored block%s\n",
              //        state.last ? " (last)" : ""));
              state.mode = STORED;
              break;
            case 1:
              /* fixed block */
              fixedtables(state);
              //Tracev((stderr, "inflate:     fixed codes block%s\n",
              //        state.last ? " (last)" : ""));
              state.mode = LEN_; /* decode codes */
              if (flush === Z_TREES$1) {
                //--- DROPBITS(2) ---//
                hold >>>= 2;
                bits -= 2;
                //---//
                break inf_leave;
              }
              break;
            case 2:
              /* dynamic block */
              //Tracev((stderr, "inflate:     dynamic codes block%s\n",
              //        state.last ? " (last)" : ""));
              state.mode = TABLE;
              break;
            case 3:
              strm.msg = 'invalid block type';
              state.mode = BAD;
            }
            //--- DROPBITS(2) ---//
            hold >>>= 2;
            bits -= 2;
            //---//
            break;
          case STORED:
            //--- BYTEBITS() ---// /* go to byte boundary */
            hold >>>= bits & 7;
            bits -= bits & 7;
            //---//
            //=== NEEDBITS(32); */
            while (bits < 32) {
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            if ((hold & 0xffff) !== ((hold >>> 16) ^ 0xffff)) {
              strm.msg = 'invalid stored block lengths';
              state.mode = BAD;
              break;
            }
            state.length = hold & 0xffff;
            //Tracev((stderr, "inflate:       stored length %u\n",
            //        state.length));
            //=== INITBITS();
            hold = 0;
            bits = 0;
            //===//
            state.mode = COPY_;
            if (flush === Z_TREES$1) {
              break inf_leave;
            }
            /* falls through */
          case COPY_:
            state.mode = COPY;
            /* falls through */
          case COPY:
            copy = state.length;
            if (copy) {
              if (copy > have) {
                copy = have;
              }
              if (copy > left) {
                copy = left;
              }
              if (copy === 0) {
                break inf_leave;
              }
              //--- zmemcpy(put, next, copy); ---
              arraySet(output, input, next, copy, put);
              //---//
              have -= copy;
              next += copy;
              left -= copy;
              put += copy;
              state.length -= copy;
              break;
            }
            //Tracev((stderr, "inflate:       stored end\n"));
            state.mode = TYPE;
            break;
          case TABLE:
            //=== NEEDBITS(14); */
            while (bits < 14) {
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            state.nlen = (hold & 0x1f) /*BITS(5)*/ + 257;
            //--- DROPBITS(5) ---//
            hold >>>= 5;
            bits -= 5;
            //---//
            state.ndist = (hold & 0x1f) /*BITS(5)*/ + 1;
            //--- DROPBITS(5) ---//
            hold >>>= 5;
            bits -= 5;
            //---//
            state.ncode = (hold & 0x0f) /*BITS(4)*/ + 4;
            //--- DROPBITS(4) ---//
            hold >>>= 4;
            bits -= 4;
            //---//
            //#ifndef PKZIP_BUG_WORKAROUND
            if (state.nlen > 286 || state.ndist > 30) {
              strm.msg = 'too many length or distance symbols';
              state.mode = BAD;
              break;
            }
            //#endif
            //Tracev((stderr, "inflate:       table sizes ok\n"));
            state.have = 0;
            state.mode = LENLENS;
            /* falls through */
          case LENLENS:
            while (state.have < state.ncode) {
              //=== NEEDBITS(3);
              while (bits < 3) {
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                hold += input[next++] << bits;
                bits += 8;
              }
              //===//
              state.lens[order[state.have++]] = (hold & 0x07); //BITS(3);
              //--- DROPBITS(3) ---//
              hold >>>= 3;
              bits -= 3;
              //---//
            }
            while (state.have < 19) {
              state.lens[order[state.have++]] = 0;
            }
            // We have separate tables & no pointers. 2 commented lines below not needed.
            //state.next = state.codes;
            //state.lencode = state.next;
            // Switch to use dynamic table
            state.lencode = state.lendyn;
            state.lenbits = 7;

            opts = {
              bits: state.lenbits
            };
            ret = inflate_table(CODES, state.lens, 0, 19, state.lencode, 0, state.work, opts);
            state.lenbits = opts.bits;

            if (ret) {
              strm.msg = 'invalid code lengths set';
              state.mode = BAD;
              break;
            }
            //Tracev((stderr, "inflate:       code lengths ok\n"));
            state.have = 0;
            state.mode = CODELENS;
            /* falls through */
          case CODELENS:
            while (state.have < state.nlen + state.ndist) {
              for (;;) {
                here = state.lencode[hold & ((1 << state.lenbits) - 1)]; /*BITS(state.lenbits)*/
                here_bits = here >>> 24;
                here_op = (here >>> 16) & 0xff;
                here_val = here & 0xffff;

                if ((here_bits) <= bits) {
                  break;
                }
                //--- PULLBYTE() ---//
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                hold += input[next++] << bits;
                bits += 8;
                //---//
              }
              if (here_val < 16) {
                //--- DROPBITS(here.bits) ---//
                hold >>>= here_bits;
                bits -= here_bits;
                //---//
                state.lens[state.have++] = here_val;
              } else {
                if (here_val === 16) {
                  //=== NEEDBITS(here.bits + 2);
                  n = here_bits + 2;
                  while (bits < n) {
                    if (have === 0) {
                      break inf_leave;
                    }
                    have--;
                    hold += input[next++] << bits;
                    bits += 8;
                  }
                  //===//
                  //--- DROPBITS(here.bits) ---//
                  hold >>>= here_bits;
                  bits -= here_bits;
                  //---//
                  if (state.have === 0) {
                    strm.msg = 'invalid bit length repeat';
                    state.mode = BAD;
                    break;
                  }
                  len = state.lens[state.have - 1];
                  copy = 3 + (hold & 0x03); //BITS(2);
                  //--- DROPBITS(2) ---//
                  hold >>>= 2;
                  bits -= 2;
                  //---//
                } else if (here_val === 17) {
                  //=== NEEDBITS(here.bits + 3);
                  n = here_bits + 3;
                  while (bits < n) {
                    if (have === 0) {
                      break inf_leave;
                    }
                    have--;
                    hold += input[next++] << bits;
                    bits += 8;
                  }
                  //===//
                  //--- DROPBITS(here.bits) ---//
                  hold >>>= here_bits;
                  bits -= here_bits;
                  //---//
                  len = 0;
                  copy = 3 + (hold & 0x07); //BITS(3);
                  //--- DROPBITS(3) ---//
                  hold >>>= 3;
                  bits -= 3;
                  //---//
                } else {
                  //=== NEEDBITS(here.bits + 7);
                  n = here_bits + 7;
                  while (bits < n) {
                    if (have === 0) {
                      break inf_leave;
                    }
                    have--;
                    hold += input[next++] << bits;
                    bits += 8;
                  }
                  //===//
                  //--- DROPBITS(here.bits) ---//
                  hold >>>= here_bits;
                  bits -= here_bits;
                  //---//
                  len = 0;
                  copy = 11 + (hold & 0x7f); //BITS(7);
                  //--- DROPBITS(7) ---//
                  hold >>>= 7;
                  bits -= 7;
                  //---//
                }
                if (state.have + copy > state.nlen + state.ndist) {
                  strm.msg = 'invalid bit length repeat';
                  state.mode = BAD;
                  break;
                }
                while (copy--) {
                  state.lens[state.have++] = len;
                }
              }
            }

            /* handle error breaks in while */
            if (state.mode === BAD) {
              break;
            }

            /* check for end-of-block code (better have one) */
            if (state.lens[256] === 0) {
              strm.msg = 'invalid code -- missing end-of-block';
              state.mode = BAD;
              break;
            }

            /* build code tables -- note: do not change the lenbits or distbits
               values here (9 and 6) without reading the comments in inftrees.h
               concerning the ENOUGH constants, which depend on those values */
            state.lenbits = 9;

            opts = {
              bits: state.lenbits
            };
            ret = inflate_table(LENS, state.lens, 0, state.nlen, state.lencode, 0, state.work, opts);
            // We have separate tables & no pointers. 2 commented lines below not needed.
            // state.next_index = opts.table_index;
            state.lenbits = opts.bits;
            // state.lencode = state.next;

            if (ret) {
              strm.msg = 'invalid literal/lengths set';
              state.mode = BAD;
              break;
            }

            state.distbits = 6;
            //state.distcode.copy(state.codes);
            // Switch to use dynamic table
            state.distcode = state.distdyn;
            opts = {
              bits: state.distbits
            };
            ret = inflate_table(DISTS, state.lens, state.nlen, state.ndist, state.distcode, 0, state.work, opts);
            // We have separate tables & no pointers. 2 commented lines below not needed.
            // state.next_index = opts.table_index;
            state.distbits = opts.bits;
            // state.distcode = state.next;

            if (ret) {
              strm.msg = 'invalid distances set';
              state.mode = BAD;
              break;
            }
            //Tracev((stderr, 'inflate:       codes ok\n'));
            state.mode = LEN_;
            if (flush === Z_TREES$1) {
              break inf_leave;
            }
            /* falls through */
          case LEN_:
            state.mode = LEN;
            /* falls through */
          case LEN:
            if (have >= 6 && left >= 258) {
              //--- RESTORE() ---
              strm.next_out = put;
              strm.avail_out = left;
              strm.next_in = next;
              strm.avail_in = have;
              state.hold = hold;
              state.bits = bits;
              //---
              inflate_fast(strm, _out);
              //--- LOAD() ---
              put = strm.next_out;
              output = strm.output;
              left = strm.avail_out;
              next = strm.next_in;
              input = strm.input;
              have = strm.avail_in;
              hold = state.hold;
              bits = state.bits;
              //---

              if (state.mode === TYPE) {
                state.back = -1;
              }
              break;
            }
            state.back = 0;
            for (;;) {
              here = state.lencode[hold & ((1 << state.lenbits) - 1)]; /*BITS(state.lenbits)*/
              here_bits = here >>> 24;
              here_op = (here >>> 16) & 0xff;
              here_val = here & 0xffff;

              if (here_bits <= bits) {
                break;
              }
              //--- PULLBYTE() ---//
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
              //---//
            }
            if (here_op && (here_op & 0xf0) === 0) {
              last_bits = here_bits;
              last_op = here_op;
              last_val = here_val;
              for (;;) {
                here = state.lencode[last_val +
                  ((hold & ((1 << (last_bits + last_op)) - 1)) /*BITS(last.bits + last.op)*/ >> last_bits)];
                here_bits = here >>> 24;
                here_op = (here >>> 16) & 0xff;
                here_val = here & 0xffff;

                if ((last_bits + here_bits) <= bits) {
                  break;
                }
                //--- PULLBYTE() ---//
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                hold += input[next++] << bits;
                bits += 8;
                //---//
              }
              //--- DROPBITS(last.bits) ---//
              hold >>>= last_bits;
              bits -= last_bits;
              //---//
              state.back += last_bits;
            }
            //--- DROPBITS(here.bits) ---//
            hold >>>= here_bits;
            bits -= here_bits;
            //---//
            state.back += here_bits;
            state.length = here_val;
            if (here_op === 0) {
              //Tracevv((stderr, here.val >= 0x20 && here.val < 0x7f ?
              //        "inflate:         literal '%c'\n" :
              //        "inflate:         literal 0x%02x\n", here.val));
              state.mode = LIT;
              break;
            }
            if (here_op & 32) {
              //Tracevv((stderr, "inflate:         end of block\n"));
              state.back = -1;
              state.mode = TYPE;
              break;
            }
            if (here_op & 64) {
              strm.msg = 'invalid literal/length code';
              state.mode = BAD;
              break;
            }
            state.extra = here_op & 15;
            state.mode = LENEXT;
            /* falls through */
          case LENEXT:
            if (state.extra) {
              //=== NEEDBITS(state.extra);
              n = state.extra;
              while (bits < n) {
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                hold += input[next++] << bits;
                bits += 8;
              }
              //===//
              state.length += hold & ((1 << state.extra) - 1) /*BITS(state.extra)*/ ;
              //--- DROPBITS(state.extra) ---//
              hold >>>= state.extra;
              bits -= state.extra;
              //---//
              state.back += state.extra;
            }
            //Tracevv((stderr, "inflate:         length %u\n", state.length));
            state.was = state.length;
            state.mode = DIST;
            /* falls through */
          case DIST:
            for (;;) {
              here = state.distcode[hold & ((1 << state.distbits) - 1)]; /*BITS(state.distbits)*/
              here_bits = here >>> 24;
              here_op = (here >>> 16) & 0xff;
              here_val = here & 0xffff;

              if ((here_bits) <= bits) {
                break;
              }
              //--- PULLBYTE() ---//
              if (have === 0) {
                break inf_leave;
              }
              have--;
              hold += input[next++] << bits;
              bits += 8;
              //---//
            }
            if ((here_op & 0xf0) === 0) {
              last_bits = here_bits;
              last_op = here_op;
              last_val = here_val;
              for (;;) {
                here = state.distcode[last_val +
                  ((hold & ((1 << (last_bits + last_op)) - 1)) /*BITS(last.bits + last.op)*/ >> last_bits)];
                here_bits = here >>> 24;
                here_op = (here >>> 16) & 0xff;
                here_val = here & 0xffff;

                if ((last_bits + here_bits) <= bits) {
                  break;
                }
                //--- PULLBYTE() ---//
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                hold += input[next++] << bits;
                bits += 8;
                //---//
              }
              //--- DROPBITS(last.bits) ---//
              hold >>>= last_bits;
              bits -= last_bits;
              //---//
              state.back += last_bits;
            }
            //--- DROPBITS(here.bits) ---//
            hold >>>= here_bits;
            bits -= here_bits;
            //---//
            state.back += here_bits;
            if (here_op & 64) {
              strm.msg = 'invalid distance code';
              state.mode = BAD;
              break;
            }
            state.offset = here_val;
            state.extra = (here_op) & 15;
            state.mode = DISTEXT;
            /* falls through */
          case DISTEXT:
            if (state.extra) {
              //=== NEEDBITS(state.extra);
              n = state.extra;
              while (bits < n) {
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                hold += input[next++] << bits;
                bits += 8;
              }
              //===//
              state.offset += hold & ((1 << state.extra) - 1) /*BITS(state.extra)*/ ;
              //--- DROPBITS(state.extra) ---//
              hold >>>= state.extra;
              bits -= state.extra;
              //---//
              state.back += state.extra;
            }
            //#ifdef INFLATE_STRICT
            if (state.offset > state.dmax) {
              strm.msg = 'invalid distance too far back';
              state.mode = BAD;
              break;
            }
            //#endif
            //Tracevv((stderr, "inflate:         distance %u\n", state.offset));
            state.mode = MATCH;
            /* falls through */
          case MATCH:
            if (left === 0) {
              break inf_leave;
            }
            copy = _out - left;
            if (state.offset > copy) { /* copy from window */
              copy = state.offset - copy;
              if (copy > state.whave) {
                if (state.sane) {
                  strm.msg = 'invalid distance too far back';
                  state.mode = BAD;
                  break;
                }
                // (!) This block is disabled in zlib defailts,
                // don't enable it for binary compatibility
                //#ifdef INFLATE_ALLOW_INVALID_DISTANCE_TOOFAR_ARRR
                //          Trace((stderr, "inflate.c too far\n"));
                //          copy -= state.whave;
                //          if (copy > state.length) { copy = state.length; }
                //          if (copy > left) { copy = left; }
                //          left -= copy;
                //          state.length -= copy;
                //          do {
                //            output[put++] = 0;
                //          } while (--copy);
                //          if (state.length === 0) { state.mode = LEN; }
                //          break;
                //#endif
              }
              if (copy > state.wnext) {
                copy -= state.wnext;
                from = state.wsize - copy;
              } else {
                from = state.wnext - copy;
              }
              if (copy > state.length) {
                copy = state.length;
              }
              from_source = state.window;
            } else { /* copy from output */
              from_source = output;
              from = put - state.offset;
              copy = state.length;
            }
            if (copy > left) {
              copy = left;
            }
            left -= copy;
            state.length -= copy;
            do {
              output[put++] = from_source[from++];
            } while (--copy);
            if (state.length === 0) {
              state.mode = LEN;
            }
            break;
          case LIT:
            if (left === 0) {
              break inf_leave;
            }
            output[put++] = state.length;
            left--;
            state.mode = LEN;
            break;
          case CHECK:
            if (state.wrap) {
              //=== NEEDBITS(32);
              while (bits < 32) {
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                // Use '|' insdead of '+' to make sure that result is signed
                hold |= input[next++] << bits;
                bits += 8;
              }
              //===//
              _out -= left;
              strm.total_out += _out;
              state.total += _out;
              if (_out) {
                strm.adler = state.check =
                  /*UPDATE(state.check, put - _out, _out);*/
                  (state.flags ? crc32(state.check, output, _out, put - _out) : adler32(state.check, output, _out, put - _out));

              }
              _out = left;
              // NB: crc32 stored as signed 32-bit int, zswap32 returns signed too
              if ((state.flags ? hold : zswap32(hold)) !== state.check) {
                strm.msg = 'incorrect data check';
                state.mode = BAD;
                break;
              }
              //=== INITBITS();
              hold = 0;
              bits = 0;
              //===//
              //Tracev((stderr, "inflate:   check matches trailer\n"));
            }
            state.mode = LENGTH;
            /* falls through */
          case LENGTH:
            if (state.wrap && state.flags) {
              //=== NEEDBITS(32);
              while (bits < 32) {
                if (have === 0) {
                  break inf_leave;
                }
                have--;
                hold += input[next++] << bits;
                bits += 8;
              }
              //===//
              if (hold !== (state.total & 0xffffffff)) {
                strm.msg = 'incorrect length check';
                state.mode = BAD;
                break;
              }
              //=== INITBITS();
              hold = 0;
              bits = 0;
              //===//
              //Tracev((stderr, "inflate:   length matches trailer\n"));
            }
            state.mode = DONE;
            /* falls through */
          case DONE:
            ret = Z_STREAM_END$1;
            break inf_leave;
          case BAD:
            ret = Z_DATA_ERROR$1;
            break inf_leave;
          case MEM:
            return Z_MEM_ERROR;
          case SYNC:
            /* falls through */
          default:
            return Z_STREAM_ERROR$1;
          }
        }

      // inf_leave <- here is real place for "goto inf_leave", emulated via "break inf_leave"

      /*
         Return from inflate(), updating the total counts and the check value.
         If there was no progress during the inflate() call, return a buffer
         error.  Call updatewindow() to create and/or update the window state.
         Note: a memory error from inflate() is non-recoverable.
       */

      //--- RESTORE() ---
      strm.next_out = put;
      strm.avail_out = left;
      strm.next_in = next;
      strm.avail_in = have;
      state.hold = hold;
      state.bits = bits;
      //---

      if (state.wsize || (_out !== strm.avail_out && state.mode < BAD &&
          (state.mode < CHECK || flush !== Z_FINISH$1))) {
        if (updatewindow(strm, strm.output, strm.next_out, _out - strm.avail_out)) ;
      }
      _in -= strm.avail_in;
      _out -= strm.avail_out;
      strm.total_in += _in;
      strm.total_out += _out;
      state.total += _out;
      if (state.wrap && _out) {
        strm.adler = state.check = /*UPDATE(state.check, strm.next_out - _out, _out);*/
          (state.flags ? crc32(state.check, output, _out, strm.next_out - _out) : adler32(state.check, output, _out, strm.next_out - _out));
      }
      strm.data_type = state.bits + (state.last ? 64 : 0) +
        (state.mode === TYPE ? 128 : 0) +
        (state.mode === LEN_ || state.mode === COPY_ ? 256 : 0);
      if (((_in === 0 && _out === 0) || flush === Z_FINISH$1) && ret === Z_OK$1) {
        ret = Z_BUF_ERROR$1;
      }
      return ret;
    }

    function inflateEnd(strm) {

      if (!strm || !strm.state /*|| strm->zfree == (free_func)0*/ ) {
        return Z_STREAM_ERROR$1;
      }

      var state = strm.state;
      if (state.window) {
        state.window = null;
      }
      strm.state = null;
      return Z_OK$1;
    }

    /* Not implemented
    exports.inflateCopy = inflateCopy;
    exports.inflateGetDictionary = inflateGetDictionary;
    exports.inflateMark = inflateMark;
    exports.inflatePrime = inflatePrime;
    exports.inflateSync = inflateSync;
    exports.inflateSyncPoint = inflateSyncPoint;
    exports.inflateUndermine = inflateUndermine;
    */

    // import constants from './constants';


    // zlib modes
    var NONE = 0;
    var DEFLATE = 1;
    var INFLATE = 2;
    var GZIP = 3;
    var GUNZIP = 4;
    var DEFLATERAW = 5;
    var INFLATERAW = 6;
    var UNZIP = 7;
    var Z_NO_FLUSH=         0,
      Z_PARTIAL_FLUSH=    1,
      Z_SYNC_FLUSH=    2,
      Z_FULL_FLUSH=       3,
      Z_FINISH=       4,
      Z_BLOCK=           5,
      Z_TREES=            6,

      /* Return codes for the compression/decompression functions. Negative values
      * are errors, positive values are used for special but normal events.
      */
      Z_OK=               0,
      Z_STREAM_END=       1,
      Z_NEED_DICT=      2,
      Z_ERRNO=       -1,
      Z_STREAM_ERROR=   -2,
      Z_DATA_ERROR=    -3,
      //Z_MEM_ERROR:     -4,
      Z_BUF_ERROR=    -5,
      //Z_VERSION_ERROR: -6,

      /* compression levels */
      Z_NO_COMPRESSION=         0,
      Z_BEST_SPEED=             1,
      Z_BEST_COMPRESSION=       9,
      Z_DEFAULT_COMPRESSION=   -1,


      Z_FILTERED=               1,
      Z_HUFFMAN_ONLY=           2,
      Z_RLE=                    3,
      Z_FIXED=                  4,
      Z_DEFAULT_STRATEGY=       0,

      /* Possible values of the data_type field (though see inflate()) */
      Z_BINARY=                 0,
      Z_TEXT=                   1,
      //Z_ASCII:                1, // = Z_TEXT (deprecated)
      Z_UNKNOWN=                2,

      /* The deflate compression method */
      Z_DEFLATED=               8;
    function Zlib$1(mode) {
      if (mode < DEFLATE || mode > UNZIP)
        throw new TypeError('Bad argument');

      this.mode = mode;
      this.init_done = false;
      this.write_in_progress = false;
      this.pending_close = false;
      this.windowBits = 0;
      this.level = 0;
      this.memLevel = 0;
      this.strategy = 0;
      this.dictionary = null;
    }

    Zlib$1.prototype.init = function(windowBits, level, memLevel, strategy, dictionary) {
      this.windowBits = windowBits;
      this.level = level;
      this.memLevel = memLevel;
      this.strategy = strategy;
      // dictionary not supported.

      if (this.mode === GZIP || this.mode === GUNZIP)
        this.windowBits += 16;

      if (this.mode === UNZIP)
        this.windowBits += 32;

      if (this.mode === DEFLATERAW || this.mode === INFLATERAW)
        this.windowBits = -this.windowBits;

      this.strm = new ZStream();
      var status;
      switch (this.mode) {
      case DEFLATE:
      case GZIP:
      case DEFLATERAW:
        status = deflateInit2(
          this.strm,
          this.level,
          Z_DEFLATED,
          this.windowBits,
          this.memLevel,
          this.strategy
        );
        break;
      case INFLATE:
      case GUNZIP:
      case INFLATERAW:
      case UNZIP:
        status  = inflateInit2(
          this.strm,
          this.windowBits
        );
        break;
      default:
        throw new Error('Unknown mode ' + this.mode);
      }

      if (status !== Z_OK) {
        this._error(status);
        return;
      }

      this.write_in_progress = false;
      this.init_done = true;
    };

    Zlib$1.prototype.params = function() {
      throw new Error('deflateParams Not supported');
    };

    Zlib$1.prototype._writeCheck = function() {
      if (!this.init_done)
        throw new Error('write before init');

      if (this.mode === NONE)
        throw new Error('already finalized');

      if (this.write_in_progress)
        throw new Error('write already in progress');

      if (this.pending_close)
        throw new Error('close is pending');
    };

    Zlib$1.prototype.write = function(flush, input, in_off, in_len, out, out_off, out_len) {
      this._writeCheck();
      this.write_in_progress = true;

      var self = this;
      nextTick(function() {
        self.write_in_progress = false;
        var res = self._write(flush, input, in_off, in_len, out, out_off, out_len);
        self.callback(res[0], res[1]);

        if (self.pending_close)
          self.close();
      });

      return this;
    };

    // set method for Node buffers, used by pako
    function bufferSet(data, offset) {
      for (var i = 0; i < data.length; i++) {
        this[offset + i] = data[i];
      }
    }

    Zlib$1.prototype.writeSync = function(flush, input, in_off, in_len, out, out_off, out_len) {
      this._writeCheck();
      return this._write(flush, input, in_off, in_len, out, out_off, out_len);
    };

    Zlib$1.prototype._write = function(flush, input, in_off, in_len, out, out_off, out_len) {
      this.write_in_progress = true;

      if (flush !== Z_NO_FLUSH &&
          flush !== Z_PARTIAL_FLUSH &&
          flush !== Z_SYNC_FLUSH &&
          flush !== Z_FULL_FLUSH &&
          flush !== Z_FINISH &&
          flush !== Z_BLOCK) {
        throw new Error('Invalid flush value');
      }

      if (input == null) {
        input = new Buffer$1(0);
        in_len = 0;
        in_off = 0;
      }

      if (out._set)
        out.set = out._set;
      else
        out.set = bufferSet;

      var strm = this.strm;
      strm.avail_in = in_len;
      strm.input = input;
      strm.next_in = in_off;
      strm.avail_out = out_len;
      strm.output = out;
      strm.next_out = out_off;
      var status;
      switch (this.mode) {
      case DEFLATE:
      case GZIP:
      case DEFLATERAW:
        status = deflate$1(strm, flush);
        break;
      case UNZIP:
      case INFLATE:
      case GUNZIP:
      case INFLATERAW:
        status = inflate$1(strm, flush);
        break;
      default:
        throw new Error('Unknown mode ' + this.mode);
      }

      if (status !== Z_STREAM_END && status !== Z_OK) {
        this._error(status);
      }

      this.write_in_progress = false;
      return [strm.avail_in, strm.avail_out];
    };

    Zlib$1.prototype.close = function() {
      if (this.write_in_progress) {
        this.pending_close = true;
        return;
      }

      this.pending_close = false;

      if (this.mode === DEFLATE || this.mode === GZIP || this.mode === DEFLATERAW) {
        deflateEnd(this.strm);
      } else {
        inflateEnd(this.strm);
      }

      this.mode = NONE;
    };
    var status;
    Zlib$1.prototype.reset = function() {
      switch (this.mode) {
      case DEFLATE:
      case DEFLATERAW:
        status = deflateReset(this.strm);
        break;
      case INFLATE:
      case INFLATERAW:
        status = inflateReset(this.strm);
        break;
      }

      if (status !== Z_OK) {
        this._error(status);
      }
    };

    Zlib$1.prototype._error = function(status) {
      this.onerror(msg[status] + ': ' + this.strm.msg, status);

      this.write_in_progress = false;
      if (this.pending_close)
        this.close();
    };

    var _binding = /*#__PURE__*/Object.freeze({
        __proto__: null,
        NONE: NONE,
        DEFLATE: DEFLATE,
        INFLATE: INFLATE,
        GZIP: GZIP,
        GUNZIP: GUNZIP,
        DEFLATERAW: DEFLATERAW,
        INFLATERAW: INFLATERAW,
        UNZIP: UNZIP,
        Z_NO_FLUSH: Z_NO_FLUSH,
        Z_PARTIAL_FLUSH: Z_PARTIAL_FLUSH,
        Z_SYNC_FLUSH: Z_SYNC_FLUSH,
        Z_FULL_FLUSH: Z_FULL_FLUSH,
        Z_FINISH: Z_FINISH,
        Z_BLOCK: Z_BLOCK,
        Z_TREES: Z_TREES,
        Z_OK: Z_OK,
        Z_STREAM_END: Z_STREAM_END,
        Z_NEED_DICT: Z_NEED_DICT,
        Z_ERRNO: Z_ERRNO,
        Z_STREAM_ERROR: Z_STREAM_ERROR,
        Z_DATA_ERROR: Z_DATA_ERROR,
        Z_BUF_ERROR: Z_BUF_ERROR,
        Z_NO_COMPRESSION: Z_NO_COMPRESSION,
        Z_BEST_SPEED: Z_BEST_SPEED,
        Z_BEST_COMPRESSION: Z_BEST_COMPRESSION,
        Z_DEFAULT_COMPRESSION: Z_DEFAULT_COMPRESSION,
        Z_FILTERED: Z_FILTERED,
        Z_HUFFMAN_ONLY: Z_HUFFMAN_ONLY,
        Z_RLE: Z_RLE,
        Z_FIXED: Z_FIXED,
        Z_DEFAULT_STRATEGY: Z_DEFAULT_STRATEGY,
        Z_BINARY: Z_BINARY,
        Z_TEXT: Z_TEXT,
        Z_UNKNOWN: Z_UNKNOWN,
        Z_DEFLATED: Z_DEFLATED,
        Zlib: Zlib$1
    });

    function assert (a, msg) {
      if (!a) {
        throw new Error(msg);
      }
    }
    var binding = {};
    Object.keys(_binding).forEach(function (key) {
      binding[key] = _binding[key];
    });
    // zlib doesn't provide these, so kludge them in following the same
    // const naming scheme zlib uses.
    binding.Z_MIN_WINDOWBITS = 8;
    binding.Z_MAX_WINDOWBITS = 15;
    binding.Z_DEFAULT_WINDOWBITS = 15;

    // fewer than 64 bytes per chunk is stupid.
    // technically it could work with as few as 8, but even 64 bytes
    // is absurdly low.  Usually a MB or more is best.
    binding.Z_MIN_CHUNK = 64;
    binding.Z_MAX_CHUNK = Infinity;
    binding.Z_DEFAULT_CHUNK = (16 * 1024);

    binding.Z_MIN_MEMLEVEL = 1;
    binding.Z_MAX_MEMLEVEL = 9;
    binding.Z_DEFAULT_MEMLEVEL = 8;

    binding.Z_MIN_LEVEL = -1;
    binding.Z_MAX_LEVEL = 9;
    binding.Z_DEFAULT_LEVEL = binding.Z_DEFAULT_COMPRESSION;


    // translation table for return codes.
    var codes = {
      Z_OK: binding.Z_OK,
      Z_STREAM_END: binding.Z_STREAM_END,
      Z_NEED_DICT: binding.Z_NEED_DICT,
      Z_ERRNO: binding.Z_ERRNO,
      Z_STREAM_ERROR: binding.Z_STREAM_ERROR,
      Z_DATA_ERROR: binding.Z_DATA_ERROR,
      Z_MEM_ERROR: binding.Z_MEM_ERROR,
      Z_BUF_ERROR: binding.Z_BUF_ERROR,
      Z_VERSION_ERROR: binding.Z_VERSION_ERROR
    };

    Object.keys(codes).forEach(function(k) {
      codes[codes[k]] = k;
    });

    function createDeflate(o) {
      return new Deflate(o);
    }

    function createInflate(o) {
      return new Inflate(o);
    }

    function createDeflateRaw(o) {
      return new DeflateRaw(o);
    }

    function createInflateRaw(o) {
      return new InflateRaw(o);
    }

    function createGzip(o) {
      return new Gzip(o);
    }

    function createGunzip(o) {
      return new Gunzip(o);
    }

    function createUnzip(o) {
      return new Unzip(o);
    }


    // Convenience methods.
    // compress/decompress a string or buffer in one step.
    function deflate(buffer, opts, callback) {
      if (typeof opts === 'function') {
        callback = opts;
        opts = {};
      }
      return zlibBuffer(new Deflate(opts), buffer, callback);
    }

    function deflateSync(buffer, opts) {
      return zlibBufferSync(new Deflate(opts), buffer);
    }

    function gzip(buffer, opts, callback) {
      if (typeof opts === 'function') {
        callback = opts;
        opts = {};
      }
      return zlibBuffer(new Gzip(opts), buffer, callback);
    }

    function gzipSync(buffer, opts) {
      return zlibBufferSync(new Gzip(opts), buffer);
    }

    function deflateRaw(buffer, opts, callback) {
      if (typeof opts === 'function') {
        callback = opts;
        opts = {};
      }
      return zlibBuffer(new DeflateRaw(opts), buffer, callback);
    }

    function deflateRawSync(buffer, opts) {
      return zlibBufferSync(new DeflateRaw(opts), buffer);
    }

    function unzip(buffer, opts, callback) {
      if (typeof opts === 'function') {
        callback = opts;
        opts = {};
      }
      return zlibBuffer(new Unzip(opts), buffer, callback);
    }

    function unzipSync(buffer, opts) {
      return zlibBufferSync(new Unzip(opts), buffer);
    }

    function inflate(buffer, opts, callback) {
      if (typeof opts === 'function') {
        callback = opts;
        opts = {};
      }
      return zlibBuffer(new Inflate(opts), buffer, callback);
    }

    function inflateSync(buffer, opts) {
      return zlibBufferSync(new Inflate(opts), buffer);
    }

    function gunzip(buffer, opts, callback) {
      if (typeof opts === 'function') {
        callback = opts;
        opts = {};
      }
      return zlibBuffer(new Gunzip(opts), buffer, callback);
    }

    function gunzipSync(buffer, opts) {
      return zlibBufferSync(new Gunzip(opts), buffer);
    }

    function inflateRaw(buffer, opts, callback) {
      if (typeof opts === 'function') {
        callback = opts;
        opts = {};
      }
      return zlibBuffer(new InflateRaw(opts), buffer, callback);
    }

    function inflateRawSync(buffer, opts) {
      return zlibBufferSync(new InflateRaw(opts), buffer);
    }

    function zlibBuffer(engine, buffer, callback) {
      var buffers = [];
      var nread = 0;

      engine.on('error', onError);
      engine.on('end', onEnd);

      engine.end(buffer);
      flow();

      function flow() {
        var chunk;
        while (null !== (chunk = engine.read())) {
          buffers.push(chunk);
          nread += chunk.length;
        }
        engine.once('readable', flow);
      }

      function onError(err) {
        engine.removeListener('end', onEnd);
        engine.removeListener('readable', flow);
        callback(err);
      }

      function onEnd() {
        var buf = Buffer$1.concat(buffers, nread);
        buffers = [];
        callback(null, buf);
        engine.close();
      }
    }

    function zlibBufferSync(engine, buffer) {
      if (typeof buffer === 'string')
        buffer = new Buffer$1(buffer);
      if (!isBuffer$2(buffer))
        throw new TypeError('Not a string or buffer');

      var flushFlag = binding.Z_FINISH;

      return engine._processChunk(buffer, flushFlag);
    }

    // generic zlib
    // minimal 2-byte header
    function Deflate(opts) {
      if (!(this instanceof Deflate)) return new Deflate(opts);
      Zlib.call(this, opts, binding.DEFLATE);
    }

    function Inflate(opts) {
      if (!(this instanceof Inflate)) return new Inflate(opts);
      Zlib.call(this, opts, binding.INFLATE);
    }



    // gzip - bigger header, same deflate compression
    function Gzip(opts) {
      if (!(this instanceof Gzip)) return new Gzip(opts);
      Zlib.call(this, opts, binding.GZIP);
    }

    function Gunzip(opts) {
      if (!(this instanceof Gunzip)) return new Gunzip(opts);
      Zlib.call(this, opts, binding.GUNZIP);
    }



    // raw - no header
    function DeflateRaw(opts) {
      if (!(this instanceof DeflateRaw)) return new DeflateRaw(opts);
      Zlib.call(this, opts, binding.DEFLATERAW);
    }

    function InflateRaw(opts) {
      if (!(this instanceof InflateRaw)) return new InflateRaw(opts);
      Zlib.call(this, opts, binding.INFLATERAW);
    }


    // auto-detect header.
    function Unzip(opts) {
      if (!(this instanceof Unzip)) return new Unzip(opts);
      Zlib.call(this, opts, binding.UNZIP);
    }


    // the Zlib class they all inherit from
    // This thing manages the queue of requests, and returns
    // true or false if there is anything in the queue when
    // you call the .write() method.

    function Zlib(opts, mode) {
      this._opts = opts = opts || {};
      this._chunkSize = opts.chunkSize || binding.Z_DEFAULT_CHUNK;

      Transform.call(this, opts);

      if (opts.flush) {
        if (opts.flush !== binding.Z_NO_FLUSH &&
            opts.flush !== binding.Z_PARTIAL_FLUSH &&
            opts.flush !== binding.Z_SYNC_FLUSH &&
            opts.flush !== binding.Z_FULL_FLUSH &&
            opts.flush !== binding.Z_FINISH &&
            opts.flush !== binding.Z_BLOCK) {
          throw new Error('Invalid flush flag: ' + opts.flush);
        }
      }
      this._flushFlag = opts.flush || binding.Z_NO_FLUSH;

      if (opts.chunkSize) {
        if (opts.chunkSize < binding.Z_MIN_CHUNK ||
            opts.chunkSize > binding.Z_MAX_CHUNK) {
          throw new Error('Invalid chunk size: ' + opts.chunkSize);
        }
      }

      if (opts.windowBits) {
        if (opts.windowBits < binding.Z_MIN_WINDOWBITS ||
            opts.windowBits > binding.Z_MAX_WINDOWBITS) {
          throw new Error('Invalid windowBits: ' + opts.windowBits);
        }
      }

      if (opts.level) {
        if (opts.level < binding.Z_MIN_LEVEL ||
            opts.level > binding.Z_MAX_LEVEL) {
          throw new Error('Invalid compression level: ' + opts.level);
        }
      }

      if (opts.memLevel) {
        if (opts.memLevel < binding.Z_MIN_MEMLEVEL ||
            opts.memLevel > binding.Z_MAX_MEMLEVEL) {
          throw new Error('Invalid memLevel: ' + opts.memLevel);
        }
      }

      if (opts.strategy) {
        if (opts.strategy != binding.Z_FILTERED &&
            opts.strategy != binding.Z_HUFFMAN_ONLY &&
            opts.strategy != binding.Z_RLE &&
            opts.strategy != binding.Z_FIXED &&
            opts.strategy != binding.Z_DEFAULT_STRATEGY) {
          throw new Error('Invalid strategy: ' + opts.strategy);
        }
      }

      if (opts.dictionary) {
        if (!isBuffer$2(opts.dictionary)) {
          throw new Error('Invalid dictionary: it should be a Buffer instance');
        }
      }

      this._binding = new binding.Zlib(mode);

      var self = this;
      this._hadError = false;
      this._binding.onerror = function(message, errno) {
        // there is no way to cleanly recover.
        // continuing only obscures problems.
        self._binding = null;
        self._hadError = true;

        var error = new Error(message);
        error.errno = errno;
        error.code = binding.codes[errno];
        self.emit('error', error);
      };

      var level = binding.Z_DEFAULT_COMPRESSION;
      if (typeof opts.level === 'number') level = opts.level;

      var strategy = binding.Z_DEFAULT_STRATEGY;
      if (typeof opts.strategy === 'number') strategy = opts.strategy;

      this._binding.init(opts.windowBits || binding.Z_DEFAULT_WINDOWBITS,
                         level,
                         opts.memLevel || binding.Z_DEFAULT_MEMLEVEL,
                         strategy,
                         opts.dictionary);

      this._buffer = new Buffer$1(this._chunkSize);
      this._offset = 0;
      this._closed = false;
      this._level = level;
      this._strategy = strategy;

      this.once('end', this.close);
    }

    inherits$1(Zlib, Transform);

    Zlib.prototype.params = function(level, strategy, callback) {
      if (level < binding.Z_MIN_LEVEL ||
          level > binding.Z_MAX_LEVEL) {
        throw new RangeError('Invalid compression level: ' + level);
      }
      if (strategy != binding.Z_FILTERED &&
          strategy != binding.Z_HUFFMAN_ONLY &&
          strategy != binding.Z_RLE &&
          strategy != binding.Z_FIXED &&
          strategy != binding.Z_DEFAULT_STRATEGY) {
        throw new TypeError('Invalid strategy: ' + strategy);
      }

      if (this._level !== level || this._strategy !== strategy) {
        var self = this;
        this.flush(binding.Z_SYNC_FLUSH, function() {
          self._binding.params(level, strategy);
          if (!self._hadError) {
            self._level = level;
            self._strategy = strategy;
            if (callback) callback();
          }
        });
      } else {
        nextTick(callback);
      }
    };

    Zlib.prototype.reset = function() {
      return this._binding.reset();
    };

    // This is the _flush function called by the transform class,
    // internally, when the last chunk has been written.
    Zlib.prototype._flush = function(callback) {
      this._transform(new Buffer$1(0), '', callback);
    };

    Zlib.prototype.flush = function(kind, callback) {
      var ws = this._writableState;

      if (typeof kind === 'function' || (kind === void 0 && !callback)) {
        callback = kind;
        kind = binding.Z_FULL_FLUSH;
      }

      if (ws.ended) {
        if (callback)
          nextTick(callback);
      } else if (ws.ending) {
        if (callback)
          this.once('end', callback);
      } else if (ws.needDrain) {
        var self = this;
        this.once('drain', function() {
          self.flush(callback);
        });
      } else {
        this._flushFlag = kind;
        this.write(new Buffer$1(0), '', callback);
      }
    };

    Zlib.prototype.close = function(callback) {
      if (callback)
        nextTick(callback);

      if (this._closed)
        return;

      this._closed = true;

      this._binding.close();

      var self = this;
      nextTick(function() {
        self.emit('close');
      });
    };

    Zlib.prototype._transform = function(chunk, encoding, cb) {
      var flushFlag;
      var ws = this._writableState;
      var ending = ws.ending || ws.ended;
      var last = ending && (!chunk || ws.length === chunk.length);

      if (!chunk === null && !isBuffer$2(chunk))
        return cb(new Error('invalid input'));

      // If it's the last chunk, or a final flush, we use the Z_FINISH flush flag.
      // If it's explicitly flushing at some other time, then we use
      // Z_FULL_FLUSH. Otherwise, use Z_NO_FLUSH for maximum compression
      // goodness.
      if (last)
        flushFlag = binding.Z_FINISH;
      else {
        flushFlag = this._flushFlag;
        // once we've flushed the last of the queue, stop flushing and
        // go back to the normal behavior.
        if (chunk.length >= ws.length) {
          this._flushFlag = this._opts.flush || binding.Z_NO_FLUSH;
        }
      }

      this._processChunk(chunk, flushFlag, cb);
    };

    Zlib.prototype._processChunk = function(chunk, flushFlag, cb) {
      var availInBefore = chunk && chunk.length;
      var availOutBefore = this._chunkSize - this._offset;
      var inOff = 0;

      var self = this;

      var async = typeof cb === 'function';

      if (!async) {
        var buffers = [];
        var nread = 0;

        var error;
        this.on('error', function(er) {
          error = er;
        });

        do {
          var res = this._binding.writeSync(flushFlag,
                                            chunk, // in
                                            inOff, // in_off
                                            availInBefore, // in_len
                                            this._buffer, // out
                                            this._offset, //out_off
                                            availOutBefore); // out_len
        } while (!this._hadError && callback(res[0], res[1]));

        if (this._hadError) {
          throw error;
        }

        var buf = Buffer$1.concat(buffers, nread);
        this.close();

        return buf;
      }

      var req = this._binding.write(flushFlag,
                                    chunk, // in
                                    inOff, // in_off
                                    availInBefore, // in_len
                                    this._buffer, // out
                                    this._offset, //out_off
                                    availOutBefore); // out_len

      req.buffer = chunk;
      req.callback = callback;

      function callback(availInAfter, availOutAfter) {
        if (self._hadError)
          return;

        var have = availOutBefore - availOutAfter;
        assert(have >= 0, 'have should not go down');

        if (have > 0) {
          var out = self._buffer.slice(self._offset, self._offset + have);
          self._offset += have;
          // serve some output to the consumer.
          if (async) {
            self.push(out);
          } else {
            buffers.push(out);
            nread += out.length;
          }
        }

        // exhausted the output buffer, or used all the input create a new one.
        if (availOutAfter === 0 || self._offset >= self._chunkSize) {
          availOutBefore = self._chunkSize;
          self._offset = 0;
          self._buffer = new Buffer$1(self._chunkSize);
        }

        if (availOutAfter === 0) {
          // Not actually done.  Need to reprocess.
          // Also, update the availInBefore to the availInAfter value,
          // so that if we have to hit it a third (fourth, etc.) time,
          // it'll have the correct byte counts.
          inOff += (availInBefore - availInAfter);
          availInBefore = availInAfter;

          if (!async)
            return true;

          var newReq = self._binding.write(flushFlag,
                                           chunk,
                                           inOff,
                                           availInBefore,
                                           self._buffer,
                                           self._offset,
                                           self._chunkSize);
          newReq.callback = callback; // this same function
          newReq.buffer = chunk;
          return;
        }

        if (!async)
          return false;

        // finished with the chunk.
        cb();
      }
    };

    inherits$1(Deflate, Zlib);
    inherits$1(Inflate, Zlib);
    inherits$1(Gzip, Zlib);
    inherits$1(Gunzip, Zlib);
    inherits$1(DeflateRaw, Zlib);
    inherits$1(InflateRaw, Zlib);
    inherits$1(Unzip, Zlib);
    var zlib = {
      codes: codes,
      createDeflate: createDeflate,
      createInflate: createInflate,
      createDeflateRaw: createDeflateRaw,
      createInflateRaw: createInflateRaw,
      createGzip: createGzip,
      createGunzip: createGunzip,
      createUnzip: createUnzip,
      deflate: deflate,
      deflateSync: deflateSync,
      gzip: gzip,
      gzipSync: gzipSync,
      deflateRaw: deflateRaw,
      deflateRawSync: deflateRawSync,
      unzip: unzip,
      unzipSync: unzipSync,
      inflate: inflate,
      inflateSync: inflateSync,
      gunzip: gunzip,
      gunzipSync: gunzipSync,
      inflateRaw: inflateRaw,
      inflateRawSync: inflateRawSync,
      Deflate: Deflate,
      Inflate: Inflate,
      Gzip: Gzip,
      Gunzip: Gunzip,
      DeflateRaw: DeflateRaw,
      InflateRaw: InflateRaw,
      Unzip: Unzip,
      Zlib: Zlib
    };

    var zlib$1 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        codes: codes,
        createDeflate: createDeflate,
        createInflate: createInflate,
        createDeflateRaw: createDeflateRaw,
        createInflateRaw: createInflateRaw,
        createGzip: createGzip,
        createGunzip: createGunzip,
        createUnzip: createUnzip,
        deflate: deflate,
        deflateSync: deflateSync,
        gzip: gzip,
        gzipSync: gzipSync,
        deflateRaw: deflateRaw,
        deflateRawSync: deflateRawSync,
        unzip: unzip,
        unzipSync: unzipSync,
        inflate: inflate,
        inflateSync: inflateSync,
        gunzip: gunzip,
        gunzipSync: gunzipSync,
        inflateRaw: inflateRaw,
        inflateRawSync: inflateRawSync,
        Deflate: Deflate,
        Inflate: Inflate,
        Gzip: Gzip,
        Gunzip: Gunzip,
        DeflateRaw: DeflateRaw,
        InflateRaw: InflateRaw,
        Unzip: Unzip,
        Zlib: Zlib,
        'default': zlib
    });

    var require$$8 = /*@__PURE__*/getAugmentedNamespace(zlib$1);

    var data;
    var hasRequiredData;

    function requireData () {
    	if (hasRequiredData) return data;
    	hasRequiredData = 1;
    	data = {
    	  "version": "0.27.2"
    	};
    	return data;
    }

    var http_1;
    var hasRequiredHttp;

    function requireHttp () {
    	if (hasRequiredHttp) return http_1;
    	hasRequiredHttp = 1;

    	var utils = utils$9;
    	var settle = requireSettle();
    	var buildFullPath = buildFullPath$1;
    	var buildURL = buildURL$1;
    	var http = require$$4$1;
    	var https = require$$4$1;
    	var httpFollow = requireFollowRedirects().http;
    	var httpsFollow = requireFollowRedirects().https;
    	var url = require$$5;
    	var zlib = require$$8;
    	var VERSION = requireData().version;
    	var transitionalDefaults = transitional;
    	var AxiosError = requireAxiosError();
    	var CanceledError = requireCanceledError();

    	var isHttps = /https:?/;

    	var supportedProtocols = [ 'http:', 'https:', 'file:' ];

    	/**
    	 *
    	 * @param {http.ClientRequestArgs} options
    	 * @param {AxiosProxyConfig} proxy
    	 * @param {string} location
    	 */
    	function setProxy(options, proxy, location) {
    	  options.hostname = proxy.host;
    	  options.host = proxy.host;
    	  options.port = proxy.port;
    	  options.path = location;

    	  // Basic proxy authorization
    	  if (proxy.auth) {
    	    var base64 = Buffer$1.from(proxy.auth.username + ':' + proxy.auth.password, 'utf8').toString('base64');
    	    options.headers['Proxy-Authorization'] = 'Basic ' + base64;
    	  }

    	  // If a proxy is used, any redirects must also pass through the proxy
    	  options.beforeRedirect = function beforeRedirect(redirection) {
    	    redirection.headers.host = redirection.host;
    	    setProxy(redirection, proxy, redirection.href);
    	  };
    	}

    	/*eslint consistent-return:0*/
    	http_1 = function httpAdapter(config) {
    	  return new Promise(function dispatchHttpRequest(resolvePromise, rejectPromise) {
    	    var onCanceled;
    	    function done() {
    	      if (config.cancelToken) {
    	        config.cancelToken.unsubscribe(onCanceled);
    	      }

    	      if (config.signal) {
    	        config.signal.removeEventListener('abort', onCanceled);
    	      }
    	    }
    	    var resolve = function resolve(value) {
    	      done();
    	      resolvePromise(value);
    	    };
    	    var rejected = false;
    	    var reject = function reject(value) {
    	      done();
    	      rejected = true;
    	      rejectPromise(value);
    	    };
    	    var data = config.data;
    	    var headers = config.headers;
    	    var headerNames = {};

    	    Object.keys(headers).forEach(function storeLowerName(name) {
    	      headerNames[name.toLowerCase()] = name;
    	    });

    	    // Set User-Agent (required by some servers)
    	    // See https://github.com/axios/axios/issues/69
    	    if ('user-agent' in headerNames) {
    	      // User-Agent is specified; handle case where no UA header is desired
    	      if (!headers[headerNames['user-agent']]) {
    	        delete headers[headerNames['user-agent']];
    	      }
    	      // Otherwise, use specified value
    	    } else {
    	      // Only set header if it hasn't been set in config
    	      headers['User-Agent'] = 'axios/' + VERSION;
    	    }

    	    // support for https://www.npmjs.com/package/form-data api
    	    if (utils.isFormData(data) && utils.isFunction(data.getHeaders)) {
    	      Object.assign(headers, data.getHeaders());
    	    } else if (data && !utils.isStream(data)) {
    	      if (isBuffer$2(data)) ; else if (utils.isArrayBuffer(data)) {
    	        data = Buffer$1.from(new Uint8Array(data));
    	      } else if (utils.isString(data)) {
    	        data = Buffer$1.from(data, 'utf-8');
    	      } else {
    	        return reject(new AxiosError(
    	          'Data after transformation must be a string, an ArrayBuffer, a Buffer, or a Stream',
    	          AxiosError.ERR_BAD_REQUEST,
    	          config
    	        ));
    	      }

    	      if (config.maxBodyLength > -1 && data.length > config.maxBodyLength) {
    	        return reject(new AxiosError(
    	          'Request body larger than maxBodyLength limit',
    	          AxiosError.ERR_BAD_REQUEST,
    	          config
    	        ));
    	      }

    	      // Add Content-Length header if data exists
    	      if (!headerNames['content-length']) {
    	        headers['Content-Length'] = data.length;
    	      }
    	    }

    	    // HTTP basic authentication
    	    var auth = undefined;
    	    if (config.auth) {
    	      var username = config.auth.username || '';
    	      var password = config.auth.password || '';
    	      auth = username + ':' + password;
    	    }

    	    // Parse url
    	    var fullPath = buildFullPath(config.baseURL, config.url);
    	    var parsed = url.parse(fullPath);
    	    var protocol = parsed.protocol || supportedProtocols[0];

    	    if (supportedProtocols.indexOf(protocol) === -1) {
    	      return reject(new AxiosError(
    	        'Unsupported protocol ' + protocol,
    	        AxiosError.ERR_BAD_REQUEST,
    	        config
    	      ));
    	    }

    	    if (!auth && parsed.auth) {
    	      var urlAuth = parsed.auth.split(':');
    	      var urlUsername = urlAuth[0] || '';
    	      var urlPassword = urlAuth[1] || '';
    	      auth = urlUsername + ':' + urlPassword;
    	    }

    	    if (auth && headerNames.authorization) {
    	      delete headers[headerNames.authorization];
    	    }

    	    var isHttpsRequest = isHttps.test(protocol);
    	    var agent = isHttpsRequest ? config.httpsAgent : config.httpAgent;

    	    try {
    	      buildURL(parsed.path, config.params, config.paramsSerializer).replace(/^\?/, '');
    	    } catch (err) {
    	      var customErr = new Error(err.message);
    	      customErr.config = config;
    	      customErr.url = config.url;
    	      customErr.exists = true;
    	      reject(customErr);
    	    }

    	    var options = {
    	      path: buildURL(parsed.path, config.params, config.paramsSerializer).replace(/^\?/, ''),
    	      method: config.method.toUpperCase(),
    	      headers: headers,
    	      agent: agent,
    	      agents: { http: config.httpAgent, https: config.httpsAgent },
    	      auth: auth
    	    };

    	    if (config.socketPath) {
    	      options.socketPath = config.socketPath;
    	    } else {
    	      options.hostname = parsed.hostname;
    	      options.port = parsed.port;
    	    }

    	    var proxy = config.proxy;
    	    if (!proxy && proxy !== false) {
    	      var proxyEnv = protocol.slice(0, -1) + '_proxy';
    	      var proxyUrl = process.env[proxyEnv] || process.env[proxyEnv.toUpperCase()];
    	      if (proxyUrl) {
    	        var parsedProxyUrl = url.parse(proxyUrl);
    	        var noProxyEnv = process.env.no_proxy || process.env.NO_PROXY;
    	        var shouldProxy = true;

    	        if (noProxyEnv) {
    	          var noProxy = noProxyEnv.split(',').map(function trim(s) {
    	            return s.trim();
    	          });

    	          shouldProxy = !noProxy.some(function proxyMatch(proxyElement) {
    	            if (!proxyElement) {
    	              return false;
    	            }
    	            if (proxyElement === '*') {
    	              return true;
    	            }
    	            if (proxyElement[0] === '.' &&
    	                parsed.hostname.substr(parsed.hostname.length - proxyElement.length) === proxyElement) {
    	              return true;
    	            }

    	            return parsed.hostname === proxyElement;
    	          });
    	        }

    	        if (shouldProxy) {
    	          proxy = {
    	            host: parsedProxyUrl.hostname,
    	            port: parsedProxyUrl.port,
    	            protocol: parsedProxyUrl.protocol
    	          };

    	          if (parsedProxyUrl.auth) {
    	            var proxyUrlAuth = parsedProxyUrl.auth.split(':');
    	            proxy.auth = {
    	              username: proxyUrlAuth[0],
    	              password: proxyUrlAuth[1]
    	            };
    	          }
    	        }
    	      }
    	    }

    	    if (proxy) {
    	      options.headers.host = parsed.hostname + (parsed.port ? ':' + parsed.port : '');
    	      setProxy(options, proxy, protocol + '//' + parsed.hostname + (parsed.port ? ':' + parsed.port : '') + options.path);
    	    }

    	    var transport;
    	    var isHttpsProxy = isHttpsRequest && (proxy ? isHttps.test(proxy.protocol) : true);
    	    if (config.transport) {
    	      transport = config.transport;
    	    } else if (config.maxRedirects === 0) {
    	      transport = isHttpsProxy ? https : http;
    	    } else {
    	      if (config.maxRedirects) {
    	        options.maxRedirects = config.maxRedirects;
    	      }
    	      if (config.beforeRedirect) {
    	        options.beforeRedirect = config.beforeRedirect;
    	      }
    	      transport = isHttpsProxy ? httpsFollow : httpFollow;
    	    }

    	    if (config.maxBodyLength > -1) {
    	      options.maxBodyLength = config.maxBodyLength;
    	    }

    	    if (config.insecureHTTPParser) {
    	      options.insecureHTTPParser = config.insecureHTTPParser;
    	    }

    	    // Create the request
    	    var req = transport.request(options, function handleResponse(res) {
    	      if (req.aborted) return;

    	      // uncompress the response body transparently if required
    	      var stream = res;

    	      // return the last request in case of redirects
    	      var lastRequest = res.req || req;


    	      // if no content, is HEAD request or decompress disabled we should not decompress
    	      if (res.statusCode !== 204 && lastRequest.method !== 'HEAD' && config.decompress !== false) {
    	        switch (res.headers['content-encoding']) {
    	        /*eslint default-case:0*/
    	        case 'gzip':
    	        case 'compress':
    	        case 'deflate':
    	        // add the unzipper to the body stream processing pipeline
    	          stream = stream.pipe(zlib.createUnzip());

    	          // remove the content-encoding in order to not confuse downstream operations
    	          delete res.headers['content-encoding'];
    	          break;
    	        }
    	      }

    	      var response = {
    	        status: res.statusCode,
    	        statusText: res.statusMessage,
    	        headers: res.headers,
    	        config: config,
    	        request: lastRequest
    	      };

    	      if (config.responseType === 'stream') {
    	        response.data = stream;
    	        settle(resolve, reject, response);
    	      } else {
    	        var responseBuffer = [];
    	        var totalResponseBytes = 0;
    	        stream.on('data', function handleStreamData(chunk) {
    	          responseBuffer.push(chunk);
    	          totalResponseBytes += chunk.length;

    	          // make sure the content length is not over the maxContentLength if specified
    	          if (config.maxContentLength > -1 && totalResponseBytes > config.maxContentLength) {
    	            // stream.destoy() emit aborted event before calling reject() on Node.js v16
    	            rejected = true;
    	            stream.destroy();
    	            reject(new AxiosError('maxContentLength size of ' + config.maxContentLength + ' exceeded',
    	              AxiosError.ERR_BAD_RESPONSE, config, lastRequest));
    	          }
    	        });

    	        stream.on('aborted', function handlerStreamAborted() {
    	          if (rejected) {
    	            return;
    	          }
    	          stream.destroy();
    	          reject(new AxiosError(
    	            'maxContentLength size of ' + config.maxContentLength + ' exceeded',
    	            AxiosError.ERR_BAD_RESPONSE,
    	            config,
    	            lastRequest
    	          ));
    	        });

    	        stream.on('error', function handleStreamError(err) {
    	          if (req.aborted) return;
    	          reject(AxiosError.from(err, null, config, lastRequest));
    	        });

    	        stream.on('end', function handleStreamEnd() {
    	          try {
    	            var responseData = responseBuffer.length === 1 ? responseBuffer[0] : Buffer$1.concat(responseBuffer);
    	            if (config.responseType !== 'arraybuffer') {
    	              responseData = responseData.toString(config.responseEncoding);
    	              if (!config.responseEncoding || config.responseEncoding === 'utf8') {
    	                responseData = utils.stripBOM(responseData);
    	              }
    	            }
    	            response.data = responseData;
    	          } catch (err) {
    	            reject(AxiosError.from(err, null, config, response.request, response));
    	          }
    	          settle(resolve, reject, response);
    	        });
    	      }
    	    });

    	    // Handle errors
    	    req.on('error', function handleRequestError(err) {
    	      // @todo remove
    	      // if (req.aborted && err.code !== AxiosError.ERR_FR_TOO_MANY_REDIRECTS) return;
    	      reject(AxiosError.from(err, null, config, req));
    	    });

    	    // set tcp keep alive to prevent drop connection by peer
    	    req.on('socket', function handleRequestSocket(socket) {
    	      // default interval of sending ack packet is 1 minute
    	      socket.setKeepAlive(true, 1000 * 60);
    	    });

    	    // Handle request timeout
    	    if (config.timeout) {
    	      // This is forcing a int timeout to avoid problems if the `req` interface doesn't handle other types.
    	      var timeout = parseInt(config.timeout, 10);

    	      if (isNaN(timeout)) {
    	        reject(new AxiosError(
    	          'error trying to parse `config.timeout` to int',
    	          AxiosError.ERR_BAD_OPTION_VALUE,
    	          config,
    	          req
    	        ));

    	        return;
    	      }

    	      // Sometime, the response will be very slow, and does not respond, the connect event will be block by event loop system.
    	      // And timer callback will be fired, and abort() will be invoked before connection, then get "socket hang up" and code ECONNRESET.
    	      // At this time, if we have a large number of request, nodejs will hang up some socket on background. and the number will up and up.
    	      // And then these socket which be hang up will devoring CPU little by little.
    	      // ClientRequest.setTimeout will be fired on the specify milliseconds, and can make sure that abort() will be fired after connect.
    	      req.setTimeout(timeout, function handleRequestTimeout() {
    	        req.abort();
    	        var transitional = config.transitional || transitionalDefaults;
    	        reject(new AxiosError(
    	          'timeout of ' + timeout + 'ms exceeded',
    	          transitional.clarifyTimeoutError ? AxiosError.ETIMEDOUT : AxiosError.ECONNABORTED,
    	          config,
    	          req
    	        ));
    	      });
    	    }

    	    if (config.cancelToken || config.signal) {
    	      // Handle cancellation
    	      // eslint-disable-next-line func-names
    	      onCanceled = function(cancel) {
    	        if (req.aborted) return;

    	        req.abort();
    	        reject(!cancel || (cancel && cancel.type) ? new CanceledError() : cancel);
    	      };

    	      config.cancelToken && config.cancelToken.subscribe(onCanceled);
    	      if (config.signal) {
    	        config.signal.aborted ? onCanceled() : config.signal.addEventListener('abort', onCanceled);
    	      }
    	    }


    	    // Send the request
    	    if (utils.isStream(data)) {
    	      data.on('error', function handleStreamError(err) {
    	        reject(AxiosError.from(err, config, null, req));
    	      }).pipe(req);
    	    } else {
    	      req.end(data);
    	    }
    	  });
    	};
    	return http_1;
    }

    var FormData$1 = {exports: {}};

    var delayed_stream;
    var hasRequiredDelayed_stream;

    function requireDelayed_stream () {
    	if (hasRequiredDelayed_stream) return delayed_stream;
    	hasRequiredDelayed_stream = 1;
    	var Stream = require$$7.Stream;
    	var util = require$$1;

    	delayed_stream = DelayedStream;
    	function DelayedStream() {
    	  this.source = null;
    	  this.dataSize = 0;
    	  this.maxDataSize = 1024 * 1024;
    	  this.pauseStream = true;

    	  this._maxDataSizeExceeded = false;
    	  this._released = false;
    	  this._bufferedEvents = [];
    	}
    	util.inherits(DelayedStream, Stream);

    	DelayedStream.create = function(source, options) {
    	  var delayedStream = new this();

    	  options = options || {};
    	  for (var option in options) {
    	    delayedStream[option] = options[option];
    	  }

    	  delayedStream.source = source;

    	  var realEmit = source.emit;
    	  source.emit = function() {
    	    delayedStream._handleEmit(arguments);
    	    return realEmit.apply(source, arguments);
    	  };

    	  source.on('error', function() {});
    	  if (delayedStream.pauseStream) {
    	    source.pause();
    	  }

    	  return delayedStream;
    	};

    	Object.defineProperty(DelayedStream.prototype, 'readable', {
    	  configurable: true,
    	  enumerable: true,
    	  get: function() {
    	    return this.source.readable;
    	  }
    	});

    	DelayedStream.prototype.setEncoding = function() {
    	  return this.source.setEncoding.apply(this.source, arguments);
    	};

    	DelayedStream.prototype.resume = function() {
    	  if (!this._released) {
    	    this.release();
    	  }

    	  this.source.resume();
    	};

    	DelayedStream.prototype.pause = function() {
    	  this.source.pause();
    	};

    	DelayedStream.prototype.release = function() {
    	  this._released = true;

    	  this._bufferedEvents.forEach(function(args) {
    	    this.emit.apply(this, args);
    	  }.bind(this));
    	  this._bufferedEvents = [];
    	};

    	DelayedStream.prototype.pipe = function() {
    	  var r = Stream.prototype.pipe.apply(this, arguments);
    	  this.resume();
    	  return r;
    	};

    	DelayedStream.prototype._handleEmit = function(args) {
    	  if (this._released) {
    	    this.emit.apply(this, args);
    	    return;
    	  }

    	  if (args[0] === 'data') {
    	    this.dataSize += args[1].length;
    	    this._checkIfMaxDataSizeExceeded();
    	  }

    	  this._bufferedEvents.push(args);
    	};

    	DelayedStream.prototype._checkIfMaxDataSizeExceeded = function() {
    	  if (this._maxDataSizeExceeded) {
    	    return;
    	  }

    	  if (this.dataSize <= this.maxDataSize) {
    	    return;
    	  }

    	  this._maxDataSizeExceeded = true;
    	  var message =
    	    'DelayedStream#maxDataSize of ' + this.maxDataSize + ' bytes exceeded.';
    	  this.emit('error', new Error(message));
    	};
    	return delayed_stream;
    }

    var combined_stream;
    var hasRequiredCombined_stream;

    function requireCombined_stream () {
    	if (hasRequiredCombined_stream) return combined_stream;
    	hasRequiredCombined_stream = 1;
    	var util = require$$1;
    	var Stream = require$$7.Stream;
    	var DelayedStream = requireDelayed_stream();

    	combined_stream = CombinedStream;
    	function CombinedStream() {
    	  this.writable = false;
    	  this.readable = true;
    	  this.dataSize = 0;
    	  this.maxDataSize = 2 * 1024 * 1024;
    	  this.pauseStreams = true;

    	  this._released = false;
    	  this._streams = [];
    	  this._currentStream = null;
    	  this._insideLoop = false;
    	  this._pendingNext = false;
    	}
    	util.inherits(CombinedStream, Stream);

    	CombinedStream.create = function(options) {
    	  var combinedStream = new this();

    	  options = options || {};
    	  for (var option in options) {
    	    combinedStream[option] = options[option];
    	  }

    	  return combinedStream;
    	};

    	CombinedStream.isStreamLike = function(stream) {
    	  return (typeof stream !== 'function')
    	    && (typeof stream !== 'string')
    	    && (typeof stream !== 'boolean')
    	    && (typeof stream !== 'number')
    	    && (!isBuffer$2(stream));
    	};

    	CombinedStream.prototype.append = function(stream) {
    	  var isStreamLike = CombinedStream.isStreamLike(stream);

    	  if (isStreamLike) {
    	    if (!(stream instanceof DelayedStream)) {
    	      var newStream = DelayedStream.create(stream, {
    	        maxDataSize: Infinity,
    	        pauseStream: this.pauseStreams,
    	      });
    	      stream.on('data', this._checkDataSize.bind(this));
    	      stream = newStream;
    	    }

    	    this._handleErrors(stream);

    	    if (this.pauseStreams) {
    	      stream.pause();
    	    }
    	  }

    	  this._streams.push(stream);
    	  return this;
    	};

    	CombinedStream.prototype.pipe = function(dest, options) {
    	  Stream.prototype.pipe.call(this, dest, options);
    	  this.resume();
    	  return dest;
    	};

    	CombinedStream.prototype._getNext = function() {
    	  this._currentStream = null;

    	  if (this._insideLoop) {
    	    this._pendingNext = true;
    	    return; // defer call
    	  }

    	  this._insideLoop = true;
    	  try {
    	    do {
    	      this._pendingNext = false;
    	      this._realGetNext();
    	    } while (this._pendingNext);
    	  } finally {
    	    this._insideLoop = false;
    	  }
    	};

    	CombinedStream.prototype._realGetNext = function() {
    	  var stream = this._streams.shift();


    	  if (typeof stream == 'undefined') {
    	    this.end();
    	    return;
    	  }

    	  if (typeof stream !== 'function') {
    	    this._pipeNext(stream);
    	    return;
    	  }

    	  var getStream = stream;
    	  getStream(function(stream) {
    	    var isStreamLike = CombinedStream.isStreamLike(stream);
    	    if (isStreamLike) {
    	      stream.on('data', this._checkDataSize.bind(this));
    	      this._handleErrors(stream);
    	    }

    	    this._pipeNext(stream);
    	  }.bind(this));
    	};

    	CombinedStream.prototype._pipeNext = function(stream) {
    	  this._currentStream = stream;

    	  var isStreamLike = CombinedStream.isStreamLike(stream);
    	  if (isStreamLike) {
    	    stream.on('end', this._getNext.bind(this));
    	    stream.pipe(this, {end: false});
    	    return;
    	  }

    	  var value = stream;
    	  this.write(value);
    	  this._getNext();
    	};

    	CombinedStream.prototype._handleErrors = function(stream) {
    	  var self = this;
    	  stream.on('error', function(err) {
    	    self._emitError(err);
    	  });
    	};

    	CombinedStream.prototype.write = function(data) {
    	  this.emit('data', data);
    	};

    	CombinedStream.prototype.pause = function() {
    	  if (!this.pauseStreams) {
    	    return;
    	  }

    	  if(this.pauseStreams && this._currentStream && typeof(this._currentStream.pause) == 'function') this._currentStream.pause();
    	  this.emit('pause');
    	};

    	CombinedStream.prototype.resume = function() {
    	  if (!this._released) {
    	    this._released = true;
    	    this.writable = true;
    	    this._getNext();
    	  }

    	  if(this.pauseStreams && this._currentStream && typeof(this._currentStream.resume) == 'function') this._currentStream.resume();
    	  this.emit('resume');
    	};

    	CombinedStream.prototype.end = function() {
    	  this._reset();
    	  this.emit('end');
    	};

    	CombinedStream.prototype.destroy = function() {
    	  this._reset();
    	  this.emit('close');
    	};

    	CombinedStream.prototype._reset = function() {
    	  this.writable = false;
    	  this._streams = [];
    	  this._currentStream = null;
    	};

    	CombinedStream.prototype._checkDataSize = function() {
    	  this._updateDataSize();
    	  if (this.dataSize <= this.maxDataSize) {
    	    return;
    	  }

    	  var message =
    	    'DelayedStream#maxDataSize of ' + this.maxDataSize + ' bytes exceeded.';
    	  this._emitError(new Error(message));
    	};

    	CombinedStream.prototype._updateDataSize = function() {
    	  this.dataSize = 0;

    	  var self = this;
    	  this._streams.forEach(function(stream) {
    	    if (!stream.dataSize) {
    	      return;
    	    }

    	    self.dataSize += stream.dataSize;
    	  });

    	  if (this._currentStream && this._currentStream.dataSize) {
    	    this.dataSize += this._currentStream.dataSize;
    	  }
    	};

    	CombinedStream.prototype._emitError = function(err) {
    	  this._reset();
    	  this.emit('error', err);
    	};
    	return combined_stream;
    }

    // Copyright Joyent, Inc. and other Node contributors.
    //
    // Permission is hereby granted, free of charge, to any person obtaining a
    // copy of this software and associated documentation files (the
    // "Software"), to deal in the Software without restriction, including
    // without limitation the rights to use, copy, modify, merge, publish,
    // distribute, sublicense, and/or sell copies of the Software, and to permit
    // persons to whom the Software is furnished to do so, subject to the
    // following conditions:
    //
    // The above copyright notice and this permission notice shall be included
    // in all copies or substantial portions of the Software.
    //
    // THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    // OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    // MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    // NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    // DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    // USE OR OTHER DEALINGS IN THE SOFTWARE.

    // resolves . and .. elements in a path array with directory names there
    // must be no slashes, empty elements, or device names (c:\) in the array
    // (so also no leading and trailing slashes - it does not distinguish
    // relative and absolute paths)
    function normalizeArray(parts, allowAboveRoot) {
      // if the path tries to go above the root, `up` ends up > 0
      var up = 0;
      for (var i = parts.length - 1; i >= 0; i--) {
        var last = parts[i];
        if (last === '.') {
          parts.splice(i, 1);
        } else if (last === '..') {
          parts.splice(i, 1);
          up++;
        } else if (up) {
          parts.splice(i, 1);
          up--;
        }
      }

      // if the path is allowed to go above the root, restore leading ..s
      if (allowAboveRoot) {
        for (; up--; up) {
          parts.unshift('..');
        }
      }

      return parts;
    }

    // Split a filename into [root, dir, basename, ext], unix version
    // 'root' is just a slash, or nothing.
    var splitPathRe =
        /^(\/?|)([\s\S]*?)((?:\.{1,2}|[^\/]+?|)(\.[^.\/]*|))(?:[\/]*)$/;
    var splitPath = function(filename) {
      return splitPathRe.exec(filename).slice(1);
    };

    // path.resolve([from ...], to)
    // posix version
    function resolve() {
      var resolvedPath = '',
          resolvedAbsolute = false;

      for (var i = arguments.length - 1; i >= -1 && !resolvedAbsolute; i--) {
        var path = (i >= 0) ? arguments[i] : '/';

        // Skip empty and invalid entries
        if (typeof path !== 'string') {
          throw new TypeError('Arguments to path.resolve must be strings');
        } else if (!path) {
          continue;
        }

        resolvedPath = path + '/' + resolvedPath;
        resolvedAbsolute = path.charAt(0) === '/';
      }

      // At this point the path should be resolved to a full absolute path, but
      // handle relative paths to be safe (might happen when process.cwd() fails)

      // Normalize the path
      resolvedPath = normalizeArray(filter(resolvedPath.split('/'), function(p) {
        return !!p;
      }), !resolvedAbsolute).join('/');

      return ((resolvedAbsolute ? '/' : '') + resolvedPath) || '.';
    }
    // path.normalize(path)
    // posix version
    function normalize(path) {
      var isPathAbsolute = isAbsolute(path),
          trailingSlash = substr(path, -1) === '/';

      // Normalize the path
      path = normalizeArray(filter(path.split('/'), function(p) {
        return !!p;
      }), !isPathAbsolute).join('/');

      if (!path && !isPathAbsolute) {
        path = '.';
      }
      if (path && trailingSlash) {
        path += '/';
      }

      return (isPathAbsolute ? '/' : '') + path;
    }
    // posix version
    function isAbsolute(path) {
      return path.charAt(0) === '/';
    }

    // posix version
    function join() {
      var paths = Array.prototype.slice.call(arguments, 0);
      return normalize(filter(paths, function(p, index) {
        if (typeof p !== 'string') {
          throw new TypeError('Arguments to path.join must be strings');
        }
        return p;
      }).join('/'));
    }


    // path.relative(from, to)
    // posix version
    function relative(from, to) {
      from = resolve(from).substr(1);
      to = resolve(to).substr(1);

      function trim(arr) {
        var start = 0;
        for (; start < arr.length; start++) {
          if (arr[start] !== '') break;
        }

        var end = arr.length - 1;
        for (; end >= 0; end--) {
          if (arr[end] !== '') break;
        }

        if (start > end) return [];
        return arr.slice(start, end - start + 1);
      }

      var fromParts = trim(from.split('/'));
      var toParts = trim(to.split('/'));

      var length = Math.min(fromParts.length, toParts.length);
      var samePartsLength = length;
      for (var i = 0; i < length; i++) {
        if (fromParts[i] !== toParts[i]) {
          samePartsLength = i;
          break;
        }
      }

      var outputParts = [];
      for (var i = samePartsLength; i < fromParts.length; i++) {
        outputParts.push('..');
      }

      outputParts = outputParts.concat(toParts.slice(samePartsLength));

      return outputParts.join('/');
    }

    var sep = '/';
    var delimiter = ':';

    function dirname(path) {
      var result = splitPath(path),
          root = result[0],
          dir = result[1];

      if (!root && !dir) {
        // No dirname whatsoever
        return '.';
      }

      if (dir) {
        // It has a dirname, strip trailing slash
        dir = dir.substr(0, dir.length - 1);
      }

      return root + dir;
    }

    function basename(path, ext) {
      var f = splitPath(path)[2];
      // TODO: make this comparison case-insensitive on windows?
      if (ext && f.substr(-1 * ext.length) === ext) {
        f = f.substr(0, f.length - ext.length);
      }
      return f;
    }


    function extname(path) {
      return splitPath(path)[3];
    }
    var path = {
      extname: extname,
      basename: basename,
      dirname: dirname,
      sep: sep,
      delimiter: delimiter,
      relative: relative,
      join: join,
      isAbsolute: isAbsolute,
      normalize: normalize,
      resolve: resolve
    };
    function filter (xs, f) {
        if (xs.filter) return xs.filter(f);
        var res = [];
        for (var i = 0; i < xs.length; i++) {
            if (f(xs[i], i, xs)) res.push(xs[i]);
        }
        return res;
    }

    // String.prototype.substr - negative index don't work in IE8
    var substr = 'ab'.substr(-1) === 'b' ?
        function (str, start, len) { return str.substr(start, len) } :
        function (str, start, len) {
            if (start < 0) start = str.length + start;
            return str.substr(start, len);
        }
    ;

    var path$1 = /*#__PURE__*/Object.freeze({
        __proto__: null,
        resolve: resolve,
        normalize: normalize,
        isAbsolute: isAbsolute,
        join: join,
        relative: relative,
        sep: sep,
        delimiter: delimiter,
        dirname: dirname,
        basename: basename,
        extname: extname,
        'default': path
    });

    var require$$2 = /*@__PURE__*/getAugmentedNamespace(path$1);

    var mimeTypes = {};

    var mimeDb = {exports: {}};

    var require$$0 = {
    	"application/1d-interleaved-parityfec": {
    	source: "iana"
    },
    	"application/3gpdash-qoe-report+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/3gpp-ims+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/3gpphal+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/3gpphalforms+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/a2l": {
    	source: "iana"
    },
    	"application/ace+cbor": {
    	source: "iana"
    },
    	"application/activemessage": {
    	source: "iana"
    },
    	"application/activity+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-costmap+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-costmapfilter+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-directory+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-endpointcost+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-endpointcostparams+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-endpointprop+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-endpointpropparams+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-error+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-networkmap+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-networkmapfilter+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-updatestreamcontrol+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/alto-updatestreamparams+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/aml": {
    	source: "iana"
    },
    	"application/andrew-inset": {
    	source: "iana",
    	extensions: [
    		"ez"
    	]
    },
    	"application/applefile": {
    	source: "iana"
    },
    	"application/applixware": {
    	source: "apache",
    	extensions: [
    		"aw"
    	]
    },
    	"application/at+jwt": {
    	source: "iana"
    },
    	"application/atf": {
    	source: "iana"
    },
    	"application/atfx": {
    	source: "iana"
    },
    	"application/atom+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"atom"
    	]
    },
    	"application/atomcat+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"atomcat"
    	]
    },
    	"application/atomdeleted+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"atomdeleted"
    	]
    },
    	"application/atomicmail": {
    	source: "iana"
    },
    	"application/atomsvc+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"atomsvc"
    	]
    },
    	"application/atsc-dwd+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"dwd"
    	]
    },
    	"application/atsc-dynamic-event-message": {
    	source: "iana"
    },
    	"application/atsc-held+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"held"
    	]
    },
    	"application/atsc-rdt+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/atsc-rsat+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rsat"
    	]
    },
    	"application/atxml": {
    	source: "iana"
    },
    	"application/auth-policy+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/bacnet-xdd+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/batch-smtp": {
    	source: "iana"
    },
    	"application/bdoc": {
    	compressible: false,
    	extensions: [
    		"bdoc"
    	]
    },
    	"application/beep+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/calendar+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/calendar+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xcs"
    	]
    },
    	"application/call-completion": {
    	source: "iana"
    },
    	"application/cals-1840": {
    	source: "iana"
    },
    	"application/captive+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/cbor": {
    	source: "iana"
    },
    	"application/cbor-seq": {
    	source: "iana"
    },
    	"application/cccex": {
    	source: "iana"
    },
    	"application/ccmp+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/ccxml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"ccxml"
    	]
    },
    	"application/cdfx+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"cdfx"
    	]
    },
    	"application/cdmi-capability": {
    	source: "iana",
    	extensions: [
    		"cdmia"
    	]
    },
    	"application/cdmi-container": {
    	source: "iana",
    	extensions: [
    		"cdmic"
    	]
    },
    	"application/cdmi-domain": {
    	source: "iana",
    	extensions: [
    		"cdmid"
    	]
    },
    	"application/cdmi-object": {
    	source: "iana",
    	extensions: [
    		"cdmio"
    	]
    },
    	"application/cdmi-queue": {
    	source: "iana",
    	extensions: [
    		"cdmiq"
    	]
    },
    	"application/cdni": {
    	source: "iana"
    },
    	"application/cea": {
    	source: "iana"
    },
    	"application/cea-2018+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/cellml+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/cfw": {
    	source: "iana"
    },
    	"application/city+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/clr": {
    	source: "iana"
    },
    	"application/clue+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/clue_info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/cms": {
    	source: "iana"
    },
    	"application/cnrp+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/coap-group+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/coap-payload": {
    	source: "iana"
    },
    	"application/commonground": {
    	source: "iana"
    },
    	"application/conference-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/cose": {
    	source: "iana"
    },
    	"application/cose-key": {
    	source: "iana"
    },
    	"application/cose-key-set": {
    	source: "iana"
    },
    	"application/cpl+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"cpl"
    	]
    },
    	"application/csrattrs": {
    	source: "iana"
    },
    	"application/csta+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/cstadata+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/csvm+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/cu-seeme": {
    	source: "apache",
    	extensions: [
    		"cu"
    	]
    },
    	"application/cwt": {
    	source: "iana"
    },
    	"application/cybercash": {
    	source: "iana"
    },
    	"application/dart": {
    	compressible: true
    },
    	"application/dash+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mpd"
    	]
    },
    	"application/dash-patch+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mpp"
    	]
    },
    	"application/dashdelta": {
    	source: "iana"
    },
    	"application/davmount+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"davmount"
    	]
    },
    	"application/dca-rft": {
    	source: "iana"
    },
    	"application/dcd": {
    	source: "iana"
    },
    	"application/dec-dx": {
    	source: "iana"
    },
    	"application/dialog-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/dicom": {
    	source: "iana"
    },
    	"application/dicom+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/dicom+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/dii": {
    	source: "iana"
    },
    	"application/dit": {
    	source: "iana"
    },
    	"application/dns": {
    	source: "iana"
    },
    	"application/dns+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/dns-message": {
    	source: "iana"
    },
    	"application/docbook+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"dbk"
    	]
    },
    	"application/dots+cbor": {
    	source: "iana"
    },
    	"application/dskpp+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/dssc+der": {
    	source: "iana",
    	extensions: [
    		"dssc"
    	]
    },
    	"application/dssc+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xdssc"
    	]
    },
    	"application/dvcs": {
    	source: "iana"
    },
    	"application/ecmascript": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"es",
    		"ecma"
    	]
    },
    	"application/edi-consent": {
    	source: "iana"
    },
    	"application/edi-x12": {
    	source: "iana",
    	compressible: false
    },
    	"application/edifact": {
    	source: "iana",
    	compressible: false
    },
    	"application/efi": {
    	source: "iana"
    },
    	"application/elm+json": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/elm+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/emergencycalldata.cap+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/emergencycalldata.comment+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/emergencycalldata.control+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/emergencycalldata.deviceinfo+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/emergencycalldata.ecall.msd": {
    	source: "iana"
    },
    	"application/emergencycalldata.providerinfo+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/emergencycalldata.serviceinfo+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/emergencycalldata.subscriberinfo+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/emergencycalldata.veds+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/emma+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"emma"
    	]
    },
    	"application/emotionml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"emotionml"
    	]
    },
    	"application/encaprtp": {
    	source: "iana"
    },
    	"application/epp+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/epub+zip": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"epub"
    	]
    },
    	"application/eshop": {
    	source: "iana"
    },
    	"application/exi": {
    	source: "iana",
    	extensions: [
    		"exi"
    	]
    },
    	"application/expect-ct-report+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/express": {
    	source: "iana",
    	extensions: [
    		"exp"
    	]
    },
    	"application/fastinfoset": {
    	source: "iana"
    },
    	"application/fastsoap": {
    	source: "iana"
    },
    	"application/fdt+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"fdt"
    	]
    },
    	"application/fhir+json": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/fhir+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/fido.trusted-apps+json": {
    	compressible: true
    },
    	"application/fits": {
    	source: "iana"
    },
    	"application/flexfec": {
    	source: "iana"
    },
    	"application/font-sfnt": {
    	source: "iana"
    },
    	"application/font-tdpfr": {
    	source: "iana",
    	extensions: [
    		"pfr"
    	]
    },
    	"application/font-woff": {
    	source: "iana",
    	compressible: false
    },
    	"application/framework-attributes+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/geo+json": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"geojson"
    	]
    },
    	"application/geo+json-seq": {
    	source: "iana"
    },
    	"application/geopackage+sqlite3": {
    	source: "iana"
    },
    	"application/geoxacml+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/gltf-buffer": {
    	source: "iana"
    },
    	"application/gml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"gml"
    	]
    },
    	"application/gpx+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"gpx"
    	]
    },
    	"application/gxf": {
    	source: "apache",
    	extensions: [
    		"gxf"
    	]
    },
    	"application/gzip": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"gz"
    	]
    },
    	"application/h224": {
    	source: "iana"
    },
    	"application/held+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/hjson": {
    	extensions: [
    		"hjson"
    	]
    },
    	"application/http": {
    	source: "iana"
    },
    	"application/hyperstudio": {
    	source: "iana",
    	extensions: [
    		"stk"
    	]
    },
    	"application/ibe-key-request+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/ibe-pkg-reply+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/ibe-pp-data": {
    	source: "iana"
    },
    	"application/iges": {
    	source: "iana"
    },
    	"application/im-iscomposing+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/index": {
    	source: "iana"
    },
    	"application/index.cmd": {
    	source: "iana"
    },
    	"application/index.obj": {
    	source: "iana"
    },
    	"application/index.response": {
    	source: "iana"
    },
    	"application/index.vnd": {
    	source: "iana"
    },
    	"application/inkml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"ink",
    		"inkml"
    	]
    },
    	"application/iotp": {
    	source: "iana"
    },
    	"application/ipfix": {
    	source: "iana",
    	extensions: [
    		"ipfix"
    	]
    },
    	"application/ipp": {
    	source: "iana"
    },
    	"application/isup": {
    	source: "iana"
    },
    	"application/its+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"its"
    	]
    },
    	"application/java-archive": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"jar",
    		"war",
    		"ear"
    	]
    },
    	"application/java-serialized-object": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"ser"
    	]
    },
    	"application/java-vm": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"class"
    	]
    },
    	"application/javascript": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true,
    	extensions: [
    		"js",
    		"mjs"
    	]
    },
    	"application/jf2feed+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/jose": {
    	source: "iana"
    },
    	"application/jose+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/jrd+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/jscalendar+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/json": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true,
    	extensions: [
    		"json",
    		"map"
    	]
    },
    	"application/json-patch+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/json-seq": {
    	source: "iana"
    },
    	"application/json5": {
    	extensions: [
    		"json5"
    	]
    },
    	"application/jsonml+json": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"jsonml"
    	]
    },
    	"application/jwk+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/jwk-set+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/jwt": {
    	source: "iana"
    },
    	"application/kpml-request+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/kpml-response+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/ld+json": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"jsonld"
    	]
    },
    	"application/lgr+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"lgr"
    	]
    },
    	"application/link-format": {
    	source: "iana"
    },
    	"application/load-control+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/lost+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"lostxml"
    	]
    },
    	"application/lostsync+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/lpf+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/lxf": {
    	source: "iana"
    },
    	"application/mac-binhex40": {
    	source: "iana",
    	extensions: [
    		"hqx"
    	]
    },
    	"application/mac-compactpro": {
    	source: "apache",
    	extensions: [
    		"cpt"
    	]
    },
    	"application/macwriteii": {
    	source: "iana"
    },
    	"application/mads+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mads"
    	]
    },
    	"application/manifest+json": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true,
    	extensions: [
    		"webmanifest"
    	]
    },
    	"application/marc": {
    	source: "iana",
    	extensions: [
    		"mrc"
    	]
    },
    	"application/marcxml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mrcx"
    	]
    },
    	"application/mathematica": {
    	source: "iana",
    	extensions: [
    		"ma",
    		"nb",
    		"mb"
    	]
    },
    	"application/mathml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mathml"
    	]
    },
    	"application/mathml-content+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mathml-presentation+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-associated-procedure-description+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-deregister+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-envelope+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-msk+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-msk-response+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-protection-description+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-reception-report+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-register+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-register-response+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-schedule+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbms-user-service-description+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mbox": {
    	source: "iana",
    	extensions: [
    		"mbox"
    	]
    },
    	"application/media-policy-dataset+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mpf"
    	]
    },
    	"application/media_control+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mediaservercontrol+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mscml"
    	]
    },
    	"application/merge-patch+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/metalink+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"metalink"
    	]
    },
    	"application/metalink4+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"meta4"
    	]
    },
    	"application/mets+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mets"
    	]
    },
    	"application/mf4": {
    	source: "iana"
    },
    	"application/mikey": {
    	source: "iana"
    },
    	"application/mipc": {
    	source: "iana"
    },
    	"application/missing-blocks+cbor-seq": {
    	source: "iana"
    },
    	"application/mmt-aei+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"maei"
    	]
    },
    	"application/mmt-usd+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"musd"
    	]
    },
    	"application/mods+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mods"
    	]
    },
    	"application/moss-keys": {
    	source: "iana"
    },
    	"application/moss-signature": {
    	source: "iana"
    },
    	"application/mosskey-data": {
    	source: "iana"
    },
    	"application/mosskey-request": {
    	source: "iana"
    },
    	"application/mp21": {
    	source: "iana",
    	extensions: [
    		"m21",
    		"mp21"
    	]
    },
    	"application/mp4": {
    	source: "iana",
    	extensions: [
    		"mp4s",
    		"m4p"
    	]
    },
    	"application/mpeg4-generic": {
    	source: "iana"
    },
    	"application/mpeg4-iod": {
    	source: "iana"
    },
    	"application/mpeg4-iod-xmt": {
    	source: "iana"
    },
    	"application/mrb-consumer+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/mrb-publish+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/msc-ivr+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/msc-mixer+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/msword": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"doc",
    		"dot"
    	]
    },
    	"application/mud+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/multipart-core": {
    	source: "iana"
    },
    	"application/mxf": {
    	source: "iana",
    	extensions: [
    		"mxf"
    	]
    },
    	"application/n-quads": {
    	source: "iana",
    	extensions: [
    		"nq"
    	]
    },
    	"application/n-triples": {
    	source: "iana",
    	extensions: [
    		"nt"
    	]
    },
    	"application/nasdata": {
    	source: "iana"
    },
    	"application/news-checkgroups": {
    	source: "iana",
    	charset: "US-ASCII"
    },
    	"application/news-groupinfo": {
    	source: "iana",
    	charset: "US-ASCII"
    },
    	"application/news-transmission": {
    	source: "iana"
    },
    	"application/nlsml+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/node": {
    	source: "iana",
    	extensions: [
    		"cjs"
    	]
    },
    	"application/nss": {
    	source: "iana"
    },
    	"application/oauth-authz-req+jwt": {
    	source: "iana"
    },
    	"application/oblivious-dns-message": {
    	source: "iana"
    },
    	"application/ocsp-request": {
    	source: "iana"
    },
    	"application/ocsp-response": {
    	source: "iana"
    },
    	"application/octet-stream": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"bin",
    		"dms",
    		"lrf",
    		"mar",
    		"so",
    		"dist",
    		"distz",
    		"pkg",
    		"bpk",
    		"dump",
    		"elc",
    		"deploy",
    		"exe",
    		"dll",
    		"deb",
    		"dmg",
    		"iso",
    		"img",
    		"msi",
    		"msp",
    		"msm",
    		"buffer"
    	]
    },
    	"application/oda": {
    	source: "iana",
    	extensions: [
    		"oda"
    	]
    },
    	"application/odm+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/odx": {
    	source: "iana"
    },
    	"application/oebps-package+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"opf"
    	]
    },
    	"application/ogg": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"ogx"
    	]
    },
    	"application/omdoc+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"omdoc"
    	]
    },
    	"application/onenote": {
    	source: "apache",
    	extensions: [
    		"onetoc",
    		"onetoc2",
    		"onetmp",
    		"onepkg"
    	]
    },
    	"application/opc-nodeset+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/oscore": {
    	source: "iana"
    },
    	"application/oxps": {
    	source: "iana",
    	extensions: [
    		"oxps"
    	]
    },
    	"application/p21": {
    	source: "iana"
    },
    	"application/p21+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/p2p-overlay+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"relo"
    	]
    },
    	"application/parityfec": {
    	source: "iana"
    },
    	"application/passport": {
    	source: "iana"
    },
    	"application/patch-ops-error+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xer"
    	]
    },
    	"application/pdf": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"pdf"
    	]
    },
    	"application/pdx": {
    	source: "iana"
    },
    	"application/pem-certificate-chain": {
    	source: "iana"
    },
    	"application/pgp-encrypted": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"pgp"
    	]
    },
    	"application/pgp-keys": {
    	source: "iana",
    	extensions: [
    		"asc"
    	]
    },
    	"application/pgp-signature": {
    	source: "iana",
    	extensions: [
    		"asc",
    		"sig"
    	]
    },
    	"application/pics-rules": {
    	source: "apache",
    	extensions: [
    		"prf"
    	]
    },
    	"application/pidf+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/pidf-diff+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/pkcs10": {
    	source: "iana",
    	extensions: [
    		"p10"
    	]
    },
    	"application/pkcs12": {
    	source: "iana"
    },
    	"application/pkcs7-mime": {
    	source: "iana",
    	extensions: [
    		"p7m",
    		"p7c"
    	]
    },
    	"application/pkcs7-signature": {
    	source: "iana",
    	extensions: [
    		"p7s"
    	]
    },
    	"application/pkcs8": {
    	source: "iana",
    	extensions: [
    		"p8"
    	]
    },
    	"application/pkcs8-encrypted": {
    	source: "iana"
    },
    	"application/pkix-attr-cert": {
    	source: "iana",
    	extensions: [
    		"ac"
    	]
    },
    	"application/pkix-cert": {
    	source: "iana",
    	extensions: [
    		"cer"
    	]
    },
    	"application/pkix-crl": {
    	source: "iana",
    	extensions: [
    		"crl"
    	]
    },
    	"application/pkix-pkipath": {
    	source: "iana",
    	extensions: [
    		"pkipath"
    	]
    },
    	"application/pkixcmp": {
    	source: "iana",
    	extensions: [
    		"pki"
    	]
    },
    	"application/pls+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"pls"
    	]
    },
    	"application/poc-settings+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/postscript": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"ai",
    		"eps",
    		"ps"
    	]
    },
    	"application/ppsp-tracker+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/problem+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/problem+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/provenance+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"provx"
    	]
    },
    	"application/prs.alvestrand.titrax-sheet": {
    	source: "iana"
    },
    	"application/prs.cww": {
    	source: "iana",
    	extensions: [
    		"cww"
    	]
    },
    	"application/prs.cyn": {
    	source: "iana",
    	charset: "7-BIT"
    },
    	"application/prs.hpub+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/prs.nprend": {
    	source: "iana"
    },
    	"application/prs.plucker": {
    	source: "iana"
    },
    	"application/prs.rdf-xml-crypt": {
    	source: "iana"
    },
    	"application/prs.xsf+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/pskc+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"pskcxml"
    	]
    },
    	"application/pvd+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/qsig": {
    	source: "iana"
    },
    	"application/raml+yaml": {
    	compressible: true,
    	extensions: [
    		"raml"
    	]
    },
    	"application/raptorfec": {
    	source: "iana"
    },
    	"application/rdap+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/rdf+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rdf",
    		"owl"
    	]
    },
    	"application/reginfo+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rif"
    	]
    },
    	"application/relax-ng-compact-syntax": {
    	source: "iana",
    	extensions: [
    		"rnc"
    	]
    },
    	"application/remote-printing": {
    	source: "iana"
    },
    	"application/reputon+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/resource-lists+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rl"
    	]
    },
    	"application/resource-lists-diff+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rld"
    	]
    },
    	"application/rfc+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/riscos": {
    	source: "iana"
    },
    	"application/rlmi+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/rls-services+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rs"
    	]
    },
    	"application/route-apd+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rapd"
    	]
    },
    	"application/route-s-tsid+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"sls"
    	]
    },
    	"application/route-usd+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rusd"
    	]
    },
    	"application/rpki-ghostbusters": {
    	source: "iana",
    	extensions: [
    		"gbr"
    	]
    },
    	"application/rpki-manifest": {
    	source: "iana",
    	extensions: [
    		"mft"
    	]
    },
    	"application/rpki-publication": {
    	source: "iana"
    },
    	"application/rpki-roa": {
    	source: "iana",
    	extensions: [
    		"roa"
    	]
    },
    	"application/rpki-updown": {
    	source: "iana"
    },
    	"application/rsd+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"rsd"
    	]
    },
    	"application/rss+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"rss"
    	]
    },
    	"application/rtf": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rtf"
    	]
    },
    	"application/rtploopback": {
    	source: "iana"
    },
    	"application/rtx": {
    	source: "iana"
    },
    	"application/samlassertion+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/samlmetadata+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/sarif+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/sarif-external-properties+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/sbe": {
    	source: "iana"
    },
    	"application/sbml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"sbml"
    	]
    },
    	"application/scaip+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/scim+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/scvp-cv-request": {
    	source: "iana",
    	extensions: [
    		"scq"
    	]
    },
    	"application/scvp-cv-response": {
    	source: "iana",
    	extensions: [
    		"scs"
    	]
    },
    	"application/scvp-vp-request": {
    	source: "iana",
    	extensions: [
    		"spq"
    	]
    },
    	"application/scvp-vp-response": {
    	source: "iana",
    	extensions: [
    		"spp"
    	]
    },
    	"application/sdp": {
    	source: "iana",
    	extensions: [
    		"sdp"
    	]
    },
    	"application/secevent+jwt": {
    	source: "iana"
    },
    	"application/senml+cbor": {
    	source: "iana"
    },
    	"application/senml+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/senml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"senmlx"
    	]
    },
    	"application/senml-etch+cbor": {
    	source: "iana"
    },
    	"application/senml-etch+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/senml-exi": {
    	source: "iana"
    },
    	"application/sensml+cbor": {
    	source: "iana"
    },
    	"application/sensml+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/sensml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"sensmlx"
    	]
    },
    	"application/sensml-exi": {
    	source: "iana"
    },
    	"application/sep+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/sep-exi": {
    	source: "iana"
    },
    	"application/session-info": {
    	source: "iana"
    },
    	"application/set-payment": {
    	source: "iana"
    },
    	"application/set-payment-initiation": {
    	source: "iana",
    	extensions: [
    		"setpay"
    	]
    },
    	"application/set-registration": {
    	source: "iana"
    },
    	"application/set-registration-initiation": {
    	source: "iana",
    	extensions: [
    		"setreg"
    	]
    },
    	"application/sgml": {
    	source: "iana"
    },
    	"application/sgml-open-catalog": {
    	source: "iana"
    },
    	"application/shf+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"shf"
    	]
    },
    	"application/sieve": {
    	source: "iana",
    	extensions: [
    		"siv",
    		"sieve"
    	]
    },
    	"application/simple-filter+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/simple-message-summary": {
    	source: "iana"
    },
    	"application/simplesymbolcontainer": {
    	source: "iana"
    },
    	"application/sipc": {
    	source: "iana"
    },
    	"application/slate": {
    	source: "iana"
    },
    	"application/smil": {
    	source: "iana"
    },
    	"application/smil+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"smi",
    		"smil"
    	]
    },
    	"application/smpte336m": {
    	source: "iana"
    },
    	"application/soap+fastinfoset": {
    	source: "iana"
    },
    	"application/soap+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/sparql-query": {
    	source: "iana",
    	extensions: [
    		"rq"
    	]
    },
    	"application/sparql-results+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"srx"
    	]
    },
    	"application/spdx+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/spirits-event+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/sql": {
    	source: "iana"
    },
    	"application/srgs": {
    	source: "iana",
    	extensions: [
    		"gram"
    	]
    },
    	"application/srgs+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"grxml"
    	]
    },
    	"application/sru+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"sru"
    	]
    },
    	"application/ssdl+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"ssdl"
    	]
    },
    	"application/ssml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"ssml"
    	]
    },
    	"application/stix+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/swid+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"swidtag"
    	]
    },
    	"application/tamp-apex-update": {
    	source: "iana"
    },
    	"application/tamp-apex-update-confirm": {
    	source: "iana"
    },
    	"application/tamp-community-update": {
    	source: "iana"
    },
    	"application/tamp-community-update-confirm": {
    	source: "iana"
    },
    	"application/tamp-error": {
    	source: "iana"
    },
    	"application/tamp-sequence-adjust": {
    	source: "iana"
    },
    	"application/tamp-sequence-adjust-confirm": {
    	source: "iana"
    },
    	"application/tamp-status-query": {
    	source: "iana"
    },
    	"application/tamp-status-response": {
    	source: "iana"
    },
    	"application/tamp-update": {
    	source: "iana"
    },
    	"application/tamp-update-confirm": {
    	source: "iana"
    },
    	"application/tar": {
    	compressible: true
    },
    	"application/taxii+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/td+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/tei+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"tei",
    		"teicorpus"
    	]
    },
    	"application/tetra_isi": {
    	source: "iana"
    },
    	"application/thraud+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"tfi"
    	]
    },
    	"application/timestamp-query": {
    	source: "iana"
    },
    	"application/timestamp-reply": {
    	source: "iana"
    },
    	"application/timestamped-data": {
    	source: "iana",
    	extensions: [
    		"tsd"
    	]
    },
    	"application/tlsrpt+gzip": {
    	source: "iana"
    },
    	"application/tlsrpt+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/tnauthlist": {
    	source: "iana"
    },
    	"application/token-introspection+jwt": {
    	source: "iana"
    },
    	"application/toml": {
    	compressible: true,
    	extensions: [
    		"toml"
    	]
    },
    	"application/trickle-ice-sdpfrag": {
    	source: "iana"
    },
    	"application/trig": {
    	source: "iana",
    	extensions: [
    		"trig"
    	]
    },
    	"application/ttml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"ttml"
    	]
    },
    	"application/tve-trigger": {
    	source: "iana"
    },
    	"application/tzif": {
    	source: "iana"
    },
    	"application/tzif-leap": {
    	source: "iana"
    },
    	"application/ubjson": {
    	compressible: false,
    	extensions: [
    		"ubj"
    	]
    },
    	"application/ulpfec": {
    	source: "iana"
    },
    	"application/urc-grpsheet+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/urc-ressheet+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rsheet"
    	]
    },
    	"application/urc-targetdesc+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"td"
    	]
    },
    	"application/urc-uisocketdesc+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vcard+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vcard+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vemmi": {
    	source: "iana"
    },
    	"application/vividence.scriptfile": {
    	source: "apache"
    },
    	"application/vnd.1000minds.decision-model+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"1km"
    	]
    },
    	"application/vnd.3gpp-prose+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp-prose-pc3ch+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp-v2x-local-service-information": {
    	source: "iana"
    },
    	"application/vnd.3gpp.5gnas": {
    	source: "iana"
    },
    	"application/vnd.3gpp.access-transfer-events+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.bsf+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.gmop+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.gtpc": {
    	source: "iana"
    },
    	"application/vnd.3gpp.interworking-data": {
    	source: "iana"
    },
    	"application/vnd.3gpp.lpp": {
    	source: "iana"
    },
    	"application/vnd.3gpp.mc-signalling-ear": {
    	source: "iana"
    },
    	"application/vnd.3gpp.mcdata-affiliation-command+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcdata-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcdata-payload": {
    	source: "iana"
    },
    	"application/vnd.3gpp.mcdata-service-config+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcdata-signalling": {
    	source: "iana"
    },
    	"application/vnd.3gpp.mcdata-ue-config+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcdata-user-profile+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-affiliation-command+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-floor-request+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-location-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-mbms-usage-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-service-config+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-signed+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-ue-config+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-ue-init-config+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcptt-user-profile+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcvideo-affiliation-command+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcvideo-affiliation-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcvideo-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcvideo-location-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcvideo-mbms-usage-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcvideo-service-config+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcvideo-transmission-request+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcvideo-ue-config+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mcvideo-user-profile+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.mid-call+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.ngap": {
    	source: "iana"
    },
    	"application/vnd.3gpp.pfcp": {
    	source: "iana"
    },
    	"application/vnd.3gpp.pic-bw-large": {
    	source: "iana",
    	extensions: [
    		"plb"
    	]
    },
    	"application/vnd.3gpp.pic-bw-small": {
    	source: "iana",
    	extensions: [
    		"psb"
    	]
    },
    	"application/vnd.3gpp.pic-bw-var": {
    	source: "iana",
    	extensions: [
    		"pvb"
    	]
    },
    	"application/vnd.3gpp.s1ap": {
    	source: "iana"
    },
    	"application/vnd.3gpp.sms": {
    	source: "iana"
    },
    	"application/vnd.3gpp.sms+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.srvcc-ext+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.srvcc-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.state-and-event-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp.ussd+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp2.bcmcsinfo+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.3gpp2.sms": {
    	source: "iana"
    },
    	"application/vnd.3gpp2.tcap": {
    	source: "iana",
    	extensions: [
    		"tcap"
    	]
    },
    	"application/vnd.3lightssoftware.imagescal": {
    	source: "iana"
    },
    	"application/vnd.3m.post-it-notes": {
    	source: "iana",
    	extensions: [
    		"pwn"
    	]
    },
    	"application/vnd.accpac.simply.aso": {
    	source: "iana",
    	extensions: [
    		"aso"
    	]
    },
    	"application/vnd.accpac.simply.imp": {
    	source: "iana",
    	extensions: [
    		"imp"
    	]
    },
    	"application/vnd.acucobol": {
    	source: "iana",
    	extensions: [
    		"acu"
    	]
    },
    	"application/vnd.acucorp": {
    	source: "iana",
    	extensions: [
    		"atc",
    		"acutc"
    	]
    },
    	"application/vnd.adobe.air-application-installer-package+zip": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"air"
    	]
    },
    	"application/vnd.adobe.flash.movie": {
    	source: "iana"
    },
    	"application/vnd.adobe.formscentral.fcdt": {
    	source: "iana",
    	extensions: [
    		"fcdt"
    	]
    },
    	"application/vnd.adobe.fxp": {
    	source: "iana",
    	extensions: [
    		"fxp",
    		"fxpl"
    	]
    },
    	"application/vnd.adobe.partial-upload": {
    	source: "iana"
    },
    	"application/vnd.adobe.xdp+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xdp"
    	]
    },
    	"application/vnd.adobe.xfdf": {
    	source: "iana",
    	extensions: [
    		"xfdf"
    	]
    },
    	"application/vnd.aether.imp": {
    	source: "iana"
    },
    	"application/vnd.afpc.afplinedata": {
    	source: "iana"
    },
    	"application/vnd.afpc.afplinedata-pagedef": {
    	source: "iana"
    },
    	"application/vnd.afpc.cmoca-cmresource": {
    	source: "iana"
    },
    	"application/vnd.afpc.foca-charset": {
    	source: "iana"
    },
    	"application/vnd.afpc.foca-codedfont": {
    	source: "iana"
    },
    	"application/vnd.afpc.foca-codepage": {
    	source: "iana"
    },
    	"application/vnd.afpc.modca": {
    	source: "iana"
    },
    	"application/vnd.afpc.modca-cmtable": {
    	source: "iana"
    },
    	"application/vnd.afpc.modca-formdef": {
    	source: "iana"
    },
    	"application/vnd.afpc.modca-mediummap": {
    	source: "iana"
    },
    	"application/vnd.afpc.modca-objectcontainer": {
    	source: "iana"
    },
    	"application/vnd.afpc.modca-overlay": {
    	source: "iana"
    },
    	"application/vnd.afpc.modca-pagesegment": {
    	source: "iana"
    },
    	"application/vnd.age": {
    	source: "iana",
    	extensions: [
    		"age"
    	]
    },
    	"application/vnd.ah-barcode": {
    	source: "iana"
    },
    	"application/vnd.ahead.space": {
    	source: "iana",
    	extensions: [
    		"ahead"
    	]
    },
    	"application/vnd.airzip.filesecure.azf": {
    	source: "iana",
    	extensions: [
    		"azf"
    	]
    },
    	"application/vnd.airzip.filesecure.azs": {
    	source: "iana",
    	extensions: [
    		"azs"
    	]
    },
    	"application/vnd.amadeus+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.amazon.ebook": {
    	source: "apache",
    	extensions: [
    		"azw"
    	]
    },
    	"application/vnd.amazon.mobi8-ebook": {
    	source: "iana"
    },
    	"application/vnd.americandynamics.acc": {
    	source: "iana",
    	extensions: [
    		"acc"
    	]
    },
    	"application/vnd.amiga.ami": {
    	source: "iana",
    	extensions: [
    		"ami"
    	]
    },
    	"application/vnd.amundsen.maze+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.android.ota": {
    	source: "iana"
    },
    	"application/vnd.android.package-archive": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"apk"
    	]
    },
    	"application/vnd.anki": {
    	source: "iana"
    },
    	"application/vnd.anser-web-certificate-issue-initiation": {
    	source: "iana",
    	extensions: [
    		"cii"
    	]
    },
    	"application/vnd.anser-web-funds-transfer-initiation": {
    	source: "apache",
    	extensions: [
    		"fti"
    	]
    },
    	"application/vnd.antix.game-component": {
    	source: "iana",
    	extensions: [
    		"atx"
    	]
    },
    	"application/vnd.apache.arrow.file": {
    	source: "iana"
    },
    	"application/vnd.apache.arrow.stream": {
    	source: "iana"
    },
    	"application/vnd.apache.thrift.binary": {
    	source: "iana"
    },
    	"application/vnd.apache.thrift.compact": {
    	source: "iana"
    },
    	"application/vnd.apache.thrift.json": {
    	source: "iana"
    },
    	"application/vnd.api+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.aplextor.warrp+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.apothekende.reservation+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.apple.installer+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mpkg"
    	]
    },
    	"application/vnd.apple.keynote": {
    	source: "iana",
    	extensions: [
    		"key"
    	]
    },
    	"application/vnd.apple.mpegurl": {
    	source: "iana",
    	extensions: [
    		"m3u8"
    	]
    },
    	"application/vnd.apple.numbers": {
    	source: "iana",
    	extensions: [
    		"numbers"
    	]
    },
    	"application/vnd.apple.pages": {
    	source: "iana",
    	extensions: [
    		"pages"
    	]
    },
    	"application/vnd.apple.pkpass": {
    	compressible: false,
    	extensions: [
    		"pkpass"
    	]
    },
    	"application/vnd.arastra.swi": {
    	source: "iana"
    },
    	"application/vnd.aristanetworks.swi": {
    	source: "iana",
    	extensions: [
    		"swi"
    	]
    },
    	"application/vnd.artisan+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.artsquare": {
    	source: "iana"
    },
    	"application/vnd.astraea-software.iota": {
    	source: "iana",
    	extensions: [
    		"iota"
    	]
    },
    	"application/vnd.audiograph": {
    	source: "iana",
    	extensions: [
    		"aep"
    	]
    },
    	"application/vnd.autopackage": {
    	source: "iana"
    },
    	"application/vnd.avalon+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.avistar+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.balsamiq.bmml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"bmml"
    	]
    },
    	"application/vnd.balsamiq.bmpr": {
    	source: "iana"
    },
    	"application/vnd.banana-accounting": {
    	source: "iana"
    },
    	"application/vnd.bbf.usp.error": {
    	source: "iana"
    },
    	"application/vnd.bbf.usp.msg": {
    	source: "iana"
    },
    	"application/vnd.bbf.usp.msg+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.bekitzur-stech+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.bint.med-content": {
    	source: "iana"
    },
    	"application/vnd.biopax.rdf+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.blink-idb-value-wrapper": {
    	source: "iana"
    },
    	"application/vnd.blueice.multipass": {
    	source: "iana",
    	extensions: [
    		"mpm"
    	]
    },
    	"application/vnd.bluetooth.ep.oob": {
    	source: "iana"
    },
    	"application/vnd.bluetooth.le.oob": {
    	source: "iana"
    },
    	"application/vnd.bmi": {
    	source: "iana",
    	extensions: [
    		"bmi"
    	]
    },
    	"application/vnd.bpf": {
    	source: "iana"
    },
    	"application/vnd.bpf3": {
    	source: "iana"
    },
    	"application/vnd.businessobjects": {
    	source: "iana",
    	extensions: [
    		"rep"
    	]
    },
    	"application/vnd.byu.uapi+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.cab-jscript": {
    	source: "iana"
    },
    	"application/vnd.canon-cpdl": {
    	source: "iana"
    },
    	"application/vnd.canon-lips": {
    	source: "iana"
    },
    	"application/vnd.capasystems-pg+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.cendio.thinlinc.clientconf": {
    	source: "iana"
    },
    	"application/vnd.century-systems.tcp_stream": {
    	source: "iana"
    },
    	"application/vnd.chemdraw+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"cdxml"
    	]
    },
    	"application/vnd.chess-pgn": {
    	source: "iana"
    },
    	"application/vnd.chipnuts.karaoke-mmd": {
    	source: "iana",
    	extensions: [
    		"mmd"
    	]
    },
    	"application/vnd.ciedi": {
    	source: "iana"
    },
    	"application/vnd.cinderella": {
    	source: "iana",
    	extensions: [
    		"cdy"
    	]
    },
    	"application/vnd.cirpack.isdn-ext": {
    	source: "iana"
    },
    	"application/vnd.citationstyles.style+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"csl"
    	]
    },
    	"application/vnd.claymore": {
    	source: "iana",
    	extensions: [
    		"cla"
    	]
    },
    	"application/vnd.cloanto.rp9": {
    	source: "iana",
    	extensions: [
    		"rp9"
    	]
    },
    	"application/vnd.clonk.c4group": {
    	source: "iana",
    	extensions: [
    		"c4g",
    		"c4d",
    		"c4f",
    		"c4p",
    		"c4u"
    	]
    },
    	"application/vnd.cluetrust.cartomobile-config": {
    	source: "iana",
    	extensions: [
    		"c11amc"
    	]
    },
    	"application/vnd.cluetrust.cartomobile-config-pkg": {
    	source: "iana",
    	extensions: [
    		"c11amz"
    	]
    },
    	"application/vnd.coffeescript": {
    	source: "iana"
    },
    	"application/vnd.collabio.xodocuments.document": {
    	source: "iana"
    },
    	"application/vnd.collabio.xodocuments.document-template": {
    	source: "iana"
    },
    	"application/vnd.collabio.xodocuments.presentation": {
    	source: "iana"
    },
    	"application/vnd.collabio.xodocuments.presentation-template": {
    	source: "iana"
    },
    	"application/vnd.collabio.xodocuments.spreadsheet": {
    	source: "iana"
    },
    	"application/vnd.collabio.xodocuments.spreadsheet-template": {
    	source: "iana"
    },
    	"application/vnd.collection+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.collection.doc+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.collection.next+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.comicbook+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.comicbook-rar": {
    	source: "iana"
    },
    	"application/vnd.commerce-battelle": {
    	source: "iana"
    },
    	"application/vnd.commonspace": {
    	source: "iana",
    	extensions: [
    		"csp"
    	]
    },
    	"application/vnd.contact.cmsg": {
    	source: "iana",
    	extensions: [
    		"cdbcmsg"
    	]
    },
    	"application/vnd.coreos.ignition+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.cosmocaller": {
    	source: "iana",
    	extensions: [
    		"cmc"
    	]
    },
    	"application/vnd.crick.clicker": {
    	source: "iana",
    	extensions: [
    		"clkx"
    	]
    },
    	"application/vnd.crick.clicker.keyboard": {
    	source: "iana",
    	extensions: [
    		"clkk"
    	]
    },
    	"application/vnd.crick.clicker.palette": {
    	source: "iana",
    	extensions: [
    		"clkp"
    	]
    },
    	"application/vnd.crick.clicker.template": {
    	source: "iana",
    	extensions: [
    		"clkt"
    	]
    },
    	"application/vnd.crick.clicker.wordbank": {
    	source: "iana",
    	extensions: [
    		"clkw"
    	]
    },
    	"application/vnd.criticaltools.wbs+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"wbs"
    	]
    },
    	"application/vnd.cryptii.pipe+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.crypto-shade-file": {
    	source: "iana"
    },
    	"application/vnd.cryptomator.encrypted": {
    	source: "iana"
    },
    	"application/vnd.cryptomator.vault": {
    	source: "iana"
    },
    	"application/vnd.ctc-posml": {
    	source: "iana",
    	extensions: [
    		"pml"
    	]
    },
    	"application/vnd.ctct.ws+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.cups-pdf": {
    	source: "iana"
    },
    	"application/vnd.cups-postscript": {
    	source: "iana"
    },
    	"application/vnd.cups-ppd": {
    	source: "iana",
    	extensions: [
    		"ppd"
    	]
    },
    	"application/vnd.cups-raster": {
    	source: "iana"
    },
    	"application/vnd.cups-raw": {
    	source: "iana"
    },
    	"application/vnd.curl": {
    	source: "iana"
    },
    	"application/vnd.curl.car": {
    	source: "apache",
    	extensions: [
    		"car"
    	]
    },
    	"application/vnd.curl.pcurl": {
    	source: "apache",
    	extensions: [
    		"pcurl"
    	]
    },
    	"application/vnd.cyan.dean.root+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.cybank": {
    	source: "iana"
    },
    	"application/vnd.cyclonedx+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.cyclonedx+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.d2l.coursepackage1p0+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.d3m-dataset": {
    	source: "iana"
    },
    	"application/vnd.d3m-problem": {
    	source: "iana"
    },
    	"application/vnd.dart": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"dart"
    	]
    },
    	"application/vnd.data-vision.rdz": {
    	source: "iana",
    	extensions: [
    		"rdz"
    	]
    },
    	"application/vnd.datapackage+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dataresource+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dbf": {
    	source: "iana",
    	extensions: [
    		"dbf"
    	]
    },
    	"application/vnd.debian.binary-package": {
    	source: "iana"
    },
    	"application/vnd.dece.data": {
    	source: "iana",
    	extensions: [
    		"uvf",
    		"uvvf",
    		"uvd",
    		"uvvd"
    	]
    },
    	"application/vnd.dece.ttml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"uvt",
    		"uvvt"
    	]
    },
    	"application/vnd.dece.unspecified": {
    	source: "iana",
    	extensions: [
    		"uvx",
    		"uvvx"
    	]
    },
    	"application/vnd.dece.zip": {
    	source: "iana",
    	extensions: [
    		"uvz",
    		"uvvz"
    	]
    },
    	"application/vnd.denovo.fcselayout-link": {
    	source: "iana",
    	extensions: [
    		"fe_launch"
    	]
    },
    	"application/vnd.desmume.movie": {
    	source: "iana"
    },
    	"application/vnd.dir-bi.plate-dl-nosuffix": {
    	source: "iana"
    },
    	"application/vnd.dm.delegation+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dna": {
    	source: "iana",
    	extensions: [
    		"dna"
    	]
    },
    	"application/vnd.document+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dolby.mlp": {
    	source: "apache",
    	extensions: [
    		"mlp"
    	]
    },
    	"application/vnd.dolby.mobile.1": {
    	source: "iana"
    },
    	"application/vnd.dolby.mobile.2": {
    	source: "iana"
    },
    	"application/vnd.doremir.scorecloud-binary-document": {
    	source: "iana"
    },
    	"application/vnd.dpgraph": {
    	source: "iana",
    	extensions: [
    		"dpg"
    	]
    },
    	"application/vnd.dreamfactory": {
    	source: "iana",
    	extensions: [
    		"dfac"
    	]
    },
    	"application/vnd.drive+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ds-keypoint": {
    	source: "apache",
    	extensions: [
    		"kpxx"
    	]
    },
    	"application/vnd.dtg.local": {
    	source: "iana"
    },
    	"application/vnd.dtg.local.flash": {
    	source: "iana"
    },
    	"application/vnd.dtg.local.html": {
    	source: "iana"
    },
    	"application/vnd.dvb.ait": {
    	source: "iana",
    	extensions: [
    		"ait"
    	]
    },
    	"application/vnd.dvb.dvbisl+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dvb.dvbj": {
    	source: "iana"
    },
    	"application/vnd.dvb.esgcontainer": {
    	source: "iana"
    },
    	"application/vnd.dvb.ipdcdftnotifaccess": {
    	source: "iana"
    },
    	"application/vnd.dvb.ipdcesgaccess": {
    	source: "iana"
    },
    	"application/vnd.dvb.ipdcesgaccess2": {
    	source: "iana"
    },
    	"application/vnd.dvb.ipdcesgpdd": {
    	source: "iana"
    },
    	"application/vnd.dvb.ipdcroaming": {
    	source: "iana"
    },
    	"application/vnd.dvb.iptv.alfec-base": {
    	source: "iana"
    },
    	"application/vnd.dvb.iptv.alfec-enhancement": {
    	source: "iana"
    },
    	"application/vnd.dvb.notif-aggregate-root+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dvb.notif-container+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dvb.notif-generic+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dvb.notif-ia-msglist+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dvb.notif-ia-registration-request+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dvb.notif-ia-registration-response+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dvb.notif-init+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.dvb.pfr": {
    	source: "iana"
    },
    	"application/vnd.dvb.service": {
    	source: "iana",
    	extensions: [
    		"svc"
    	]
    },
    	"application/vnd.dxr": {
    	source: "iana"
    },
    	"application/vnd.dynageo": {
    	source: "iana",
    	extensions: [
    		"geo"
    	]
    },
    	"application/vnd.dzr": {
    	source: "iana"
    },
    	"application/vnd.easykaraoke.cdgdownload": {
    	source: "iana"
    },
    	"application/vnd.ecdis-update": {
    	source: "iana"
    },
    	"application/vnd.ecip.rlp": {
    	source: "iana"
    },
    	"application/vnd.eclipse.ditto+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ecowin.chart": {
    	source: "iana",
    	extensions: [
    		"mag"
    	]
    },
    	"application/vnd.ecowin.filerequest": {
    	source: "iana"
    },
    	"application/vnd.ecowin.fileupdate": {
    	source: "iana"
    },
    	"application/vnd.ecowin.series": {
    	source: "iana"
    },
    	"application/vnd.ecowin.seriesrequest": {
    	source: "iana"
    },
    	"application/vnd.ecowin.seriesupdate": {
    	source: "iana"
    },
    	"application/vnd.efi.img": {
    	source: "iana"
    },
    	"application/vnd.efi.iso": {
    	source: "iana"
    },
    	"application/vnd.emclient.accessrequest+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.enliven": {
    	source: "iana",
    	extensions: [
    		"nml"
    	]
    },
    	"application/vnd.enphase.envoy": {
    	source: "iana"
    },
    	"application/vnd.eprints.data+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.epson.esf": {
    	source: "iana",
    	extensions: [
    		"esf"
    	]
    },
    	"application/vnd.epson.msf": {
    	source: "iana",
    	extensions: [
    		"msf"
    	]
    },
    	"application/vnd.epson.quickanime": {
    	source: "iana",
    	extensions: [
    		"qam"
    	]
    },
    	"application/vnd.epson.salt": {
    	source: "iana",
    	extensions: [
    		"slt"
    	]
    },
    	"application/vnd.epson.ssf": {
    	source: "iana",
    	extensions: [
    		"ssf"
    	]
    },
    	"application/vnd.ericsson.quickcall": {
    	source: "iana"
    },
    	"application/vnd.espass-espass+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.eszigno3+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"es3",
    		"et3"
    	]
    },
    	"application/vnd.etsi.aoc+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.asic-e+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.etsi.asic-s+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.etsi.cug+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.iptvcommand+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.iptvdiscovery+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.iptvprofile+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.iptvsad-bc+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.iptvsad-cod+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.iptvsad-npvr+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.iptvservice+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.iptvsync+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.iptvueprofile+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.mcid+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.mheg5": {
    	source: "iana"
    },
    	"application/vnd.etsi.overload-control-policy-dataset+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.pstn+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.sci+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.simservs+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.timestamp-token": {
    	source: "iana"
    },
    	"application/vnd.etsi.tsl+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.etsi.tsl.der": {
    	source: "iana"
    },
    	"application/vnd.eu.kasparian.car+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.eudora.data": {
    	source: "iana"
    },
    	"application/vnd.evolv.ecig.profile": {
    	source: "iana"
    },
    	"application/vnd.evolv.ecig.settings": {
    	source: "iana"
    },
    	"application/vnd.evolv.ecig.theme": {
    	source: "iana"
    },
    	"application/vnd.exstream-empower+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.exstream-package": {
    	source: "iana"
    },
    	"application/vnd.ezpix-album": {
    	source: "iana",
    	extensions: [
    		"ez2"
    	]
    },
    	"application/vnd.ezpix-package": {
    	source: "iana",
    	extensions: [
    		"ez3"
    	]
    },
    	"application/vnd.f-secure.mobile": {
    	source: "iana"
    },
    	"application/vnd.familysearch.gedcom+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.fastcopy-disk-image": {
    	source: "iana"
    },
    	"application/vnd.fdf": {
    	source: "iana",
    	extensions: [
    		"fdf"
    	]
    },
    	"application/vnd.fdsn.mseed": {
    	source: "iana",
    	extensions: [
    		"mseed"
    	]
    },
    	"application/vnd.fdsn.seed": {
    	source: "iana",
    	extensions: [
    		"seed",
    		"dataless"
    	]
    },
    	"application/vnd.ffsns": {
    	source: "iana"
    },
    	"application/vnd.ficlab.flb+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.filmit.zfc": {
    	source: "iana"
    },
    	"application/vnd.fints": {
    	source: "iana"
    },
    	"application/vnd.firemonkeys.cloudcell": {
    	source: "iana"
    },
    	"application/vnd.flographit": {
    	source: "iana",
    	extensions: [
    		"gph"
    	]
    },
    	"application/vnd.fluxtime.clip": {
    	source: "iana",
    	extensions: [
    		"ftc"
    	]
    },
    	"application/vnd.font-fontforge-sfd": {
    	source: "iana"
    },
    	"application/vnd.framemaker": {
    	source: "iana",
    	extensions: [
    		"fm",
    		"frame",
    		"maker",
    		"book"
    	]
    },
    	"application/vnd.frogans.fnc": {
    	source: "iana",
    	extensions: [
    		"fnc"
    	]
    },
    	"application/vnd.frogans.ltf": {
    	source: "iana",
    	extensions: [
    		"ltf"
    	]
    },
    	"application/vnd.fsc.weblaunch": {
    	source: "iana",
    	extensions: [
    		"fsc"
    	]
    },
    	"application/vnd.fujifilm.fb.docuworks": {
    	source: "iana"
    },
    	"application/vnd.fujifilm.fb.docuworks.binder": {
    	source: "iana"
    },
    	"application/vnd.fujifilm.fb.docuworks.container": {
    	source: "iana"
    },
    	"application/vnd.fujifilm.fb.jfi+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.fujitsu.oasys": {
    	source: "iana",
    	extensions: [
    		"oas"
    	]
    },
    	"application/vnd.fujitsu.oasys2": {
    	source: "iana",
    	extensions: [
    		"oa2"
    	]
    },
    	"application/vnd.fujitsu.oasys3": {
    	source: "iana",
    	extensions: [
    		"oa3"
    	]
    },
    	"application/vnd.fujitsu.oasysgp": {
    	source: "iana",
    	extensions: [
    		"fg5"
    	]
    },
    	"application/vnd.fujitsu.oasysprs": {
    	source: "iana",
    	extensions: [
    		"bh2"
    	]
    },
    	"application/vnd.fujixerox.art-ex": {
    	source: "iana"
    },
    	"application/vnd.fujixerox.art4": {
    	source: "iana"
    },
    	"application/vnd.fujixerox.ddd": {
    	source: "iana",
    	extensions: [
    		"ddd"
    	]
    },
    	"application/vnd.fujixerox.docuworks": {
    	source: "iana",
    	extensions: [
    		"xdw"
    	]
    },
    	"application/vnd.fujixerox.docuworks.binder": {
    	source: "iana",
    	extensions: [
    		"xbd"
    	]
    },
    	"application/vnd.fujixerox.docuworks.container": {
    	source: "iana"
    },
    	"application/vnd.fujixerox.hbpl": {
    	source: "iana"
    },
    	"application/vnd.fut-misnet": {
    	source: "iana"
    },
    	"application/vnd.futoin+cbor": {
    	source: "iana"
    },
    	"application/vnd.futoin+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.fuzzysheet": {
    	source: "iana",
    	extensions: [
    		"fzs"
    	]
    },
    	"application/vnd.genomatix.tuxedo": {
    	source: "iana",
    	extensions: [
    		"txd"
    	]
    },
    	"application/vnd.gentics.grd+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.geo+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.geocube+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.geogebra.file": {
    	source: "iana",
    	extensions: [
    		"ggb"
    	]
    },
    	"application/vnd.geogebra.slides": {
    	source: "iana"
    },
    	"application/vnd.geogebra.tool": {
    	source: "iana",
    	extensions: [
    		"ggt"
    	]
    },
    	"application/vnd.geometry-explorer": {
    	source: "iana",
    	extensions: [
    		"gex",
    		"gre"
    	]
    },
    	"application/vnd.geonext": {
    	source: "iana",
    	extensions: [
    		"gxt"
    	]
    },
    	"application/vnd.geoplan": {
    	source: "iana",
    	extensions: [
    		"g2w"
    	]
    },
    	"application/vnd.geospace": {
    	source: "iana",
    	extensions: [
    		"g3w"
    	]
    },
    	"application/vnd.gerber": {
    	source: "iana"
    },
    	"application/vnd.globalplatform.card-content-mgt": {
    	source: "iana"
    },
    	"application/vnd.globalplatform.card-content-mgt-response": {
    	source: "iana"
    },
    	"application/vnd.gmx": {
    	source: "iana",
    	extensions: [
    		"gmx"
    	]
    },
    	"application/vnd.google-apps.document": {
    	compressible: false,
    	extensions: [
    		"gdoc"
    	]
    },
    	"application/vnd.google-apps.presentation": {
    	compressible: false,
    	extensions: [
    		"gslides"
    	]
    },
    	"application/vnd.google-apps.spreadsheet": {
    	compressible: false,
    	extensions: [
    		"gsheet"
    	]
    },
    	"application/vnd.google-earth.kml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"kml"
    	]
    },
    	"application/vnd.google-earth.kmz": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"kmz"
    	]
    },
    	"application/vnd.gov.sk.e-form+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.gov.sk.e-form+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.gov.sk.xmldatacontainer+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.grafeq": {
    	source: "iana",
    	extensions: [
    		"gqf",
    		"gqs"
    	]
    },
    	"application/vnd.gridmp": {
    	source: "iana"
    },
    	"application/vnd.groove-account": {
    	source: "iana",
    	extensions: [
    		"gac"
    	]
    },
    	"application/vnd.groove-help": {
    	source: "iana",
    	extensions: [
    		"ghf"
    	]
    },
    	"application/vnd.groove-identity-message": {
    	source: "iana",
    	extensions: [
    		"gim"
    	]
    },
    	"application/vnd.groove-injector": {
    	source: "iana",
    	extensions: [
    		"grv"
    	]
    },
    	"application/vnd.groove-tool-message": {
    	source: "iana",
    	extensions: [
    		"gtm"
    	]
    },
    	"application/vnd.groove-tool-template": {
    	source: "iana",
    	extensions: [
    		"tpl"
    	]
    },
    	"application/vnd.groove-vcard": {
    	source: "iana",
    	extensions: [
    		"vcg"
    	]
    },
    	"application/vnd.hal+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.hal+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"hal"
    	]
    },
    	"application/vnd.handheld-entertainment+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"zmm"
    	]
    },
    	"application/vnd.hbci": {
    	source: "iana",
    	extensions: [
    		"hbci"
    	]
    },
    	"application/vnd.hc+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.hcl-bireports": {
    	source: "iana"
    },
    	"application/vnd.hdt": {
    	source: "iana"
    },
    	"application/vnd.heroku+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.hhe.lesson-player": {
    	source: "iana",
    	extensions: [
    		"les"
    	]
    },
    	"application/vnd.hl7cda+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/vnd.hl7v2+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/vnd.hp-hpgl": {
    	source: "iana",
    	extensions: [
    		"hpgl"
    	]
    },
    	"application/vnd.hp-hpid": {
    	source: "iana",
    	extensions: [
    		"hpid"
    	]
    },
    	"application/vnd.hp-hps": {
    	source: "iana",
    	extensions: [
    		"hps"
    	]
    },
    	"application/vnd.hp-jlyt": {
    	source: "iana",
    	extensions: [
    		"jlt"
    	]
    },
    	"application/vnd.hp-pcl": {
    	source: "iana",
    	extensions: [
    		"pcl"
    	]
    },
    	"application/vnd.hp-pclxl": {
    	source: "iana",
    	extensions: [
    		"pclxl"
    	]
    },
    	"application/vnd.httphone": {
    	source: "iana"
    },
    	"application/vnd.hydrostatix.sof-data": {
    	source: "iana",
    	extensions: [
    		"sfd-hdstx"
    	]
    },
    	"application/vnd.hyper+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.hyper-item+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.hyperdrive+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.hzn-3d-crossword": {
    	source: "iana"
    },
    	"application/vnd.ibm.afplinedata": {
    	source: "iana"
    },
    	"application/vnd.ibm.electronic-media": {
    	source: "iana"
    },
    	"application/vnd.ibm.minipay": {
    	source: "iana",
    	extensions: [
    		"mpy"
    	]
    },
    	"application/vnd.ibm.modcap": {
    	source: "iana",
    	extensions: [
    		"afp",
    		"listafp",
    		"list3820"
    	]
    },
    	"application/vnd.ibm.rights-management": {
    	source: "iana",
    	extensions: [
    		"irm"
    	]
    },
    	"application/vnd.ibm.secure-container": {
    	source: "iana",
    	extensions: [
    		"sc"
    	]
    },
    	"application/vnd.iccprofile": {
    	source: "iana",
    	extensions: [
    		"icc",
    		"icm"
    	]
    },
    	"application/vnd.ieee.1905": {
    	source: "iana"
    },
    	"application/vnd.igloader": {
    	source: "iana",
    	extensions: [
    		"igl"
    	]
    },
    	"application/vnd.imagemeter.folder+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.imagemeter.image+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.immervision-ivp": {
    	source: "iana",
    	extensions: [
    		"ivp"
    	]
    },
    	"application/vnd.immervision-ivu": {
    	source: "iana",
    	extensions: [
    		"ivu"
    	]
    },
    	"application/vnd.ims.imsccv1p1": {
    	source: "iana"
    },
    	"application/vnd.ims.imsccv1p2": {
    	source: "iana"
    },
    	"application/vnd.ims.imsccv1p3": {
    	source: "iana"
    },
    	"application/vnd.ims.lis.v2.result+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ims.lti.v2.toolconsumerprofile+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ims.lti.v2.toolproxy+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ims.lti.v2.toolproxy.id+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ims.lti.v2.toolsettings+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ims.lti.v2.toolsettings.simple+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.informedcontrol.rms+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.informix-visionary": {
    	source: "iana"
    },
    	"application/vnd.infotech.project": {
    	source: "iana"
    },
    	"application/vnd.infotech.project+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.innopath.wamp.notification": {
    	source: "iana"
    },
    	"application/vnd.insors.igm": {
    	source: "iana",
    	extensions: [
    		"igm"
    	]
    },
    	"application/vnd.intercon.formnet": {
    	source: "iana",
    	extensions: [
    		"xpw",
    		"xpx"
    	]
    },
    	"application/vnd.intergeo": {
    	source: "iana",
    	extensions: [
    		"i2g"
    	]
    },
    	"application/vnd.intertrust.digibox": {
    	source: "iana"
    },
    	"application/vnd.intertrust.nncp": {
    	source: "iana"
    },
    	"application/vnd.intu.qbo": {
    	source: "iana",
    	extensions: [
    		"qbo"
    	]
    },
    	"application/vnd.intu.qfx": {
    	source: "iana",
    	extensions: [
    		"qfx"
    	]
    },
    	"application/vnd.iptc.g2.catalogitem+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.iptc.g2.conceptitem+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.iptc.g2.knowledgeitem+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.iptc.g2.newsitem+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.iptc.g2.newsmessage+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.iptc.g2.packageitem+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.iptc.g2.planningitem+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ipunplugged.rcprofile": {
    	source: "iana",
    	extensions: [
    		"rcprofile"
    	]
    },
    	"application/vnd.irepository.package+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"irp"
    	]
    },
    	"application/vnd.is-xpr": {
    	source: "iana",
    	extensions: [
    		"xpr"
    	]
    },
    	"application/vnd.isac.fcs": {
    	source: "iana",
    	extensions: [
    		"fcs"
    	]
    },
    	"application/vnd.iso11783-10+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.jam": {
    	source: "iana",
    	extensions: [
    		"jam"
    	]
    },
    	"application/vnd.japannet-directory-service": {
    	source: "iana"
    },
    	"application/vnd.japannet-jpnstore-wakeup": {
    	source: "iana"
    },
    	"application/vnd.japannet-payment-wakeup": {
    	source: "iana"
    },
    	"application/vnd.japannet-registration": {
    	source: "iana"
    },
    	"application/vnd.japannet-registration-wakeup": {
    	source: "iana"
    },
    	"application/vnd.japannet-setstore-wakeup": {
    	source: "iana"
    },
    	"application/vnd.japannet-verification": {
    	source: "iana"
    },
    	"application/vnd.japannet-verification-wakeup": {
    	source: "iana"
    },
    	"application/vnd.jcp.javame.midlet-rms": {
    	source: "iana",
    	extensions: [
    		"rms"
    	]
    },
    	"application/vnd.jisp": {
    	source: "iana",
    	extensions: [
    		"jisp"
    	]
    },
    	"application/vnd.joost.joda-archive": {
    	source: "iana",
    	extensions: [
    		"joda"
    	]
    },
    	"application/vnd.jsk.isdn-ngn": {
    	source: "iana"
    },
    	"application/vnd.kahootz": {
    	source: "iana",
    	extensions: [
    		"ktz",
    		"ktr"
    	]
    },
    	"application/vnd.kde.karbon": {
    	source: "iana",
    	extensions: [
    		"karbon"
    	]
    },
    	"application/vnd.kde.kchart": {
    	source: "iana",
    	extensions: [
    		"chrt"
    	]
    },
    	"application/vnd.kde.kformula": {
    	source: "iana",
    	extensions: [
    		"kfo"
    	]
    },
    	"application/vnd.kde.kivio": {
    	source: "iana",
    	extensions: [
    		"flw"
    	]
    },
    	"application/vnd.kde.kontour": {
    	source: "iana",
    	extensions: [
    		"kon"
    	]
    },
    	"application/vnd.kde.kpresenter": {
    	source: "iana",
    	extensions: [
    		"kpr",
    		"kpt"
    	]
    },
    	"application/vnd.kde.kspread": {
    	source: "iana",
    	extensions: [
    		"ksp"
    	]
    },
    	"application/vnd.kde.kword": {
    	source: "iana",
    	extensions: [
    		"kwd",
    		"kwt"
    	]
    },
    	"application/vnd.kenameaapp": {
    	source: "iana",
    	extensions: [
    		"htke"
    	]
    },
    	"application/vnd.kidspiration": {
    	source: "iana",
    	extensions: [
    		"kia"
    	]
    },
    	"application/vnd.kinar": {
    	source: "iana",
    	extensions: [
    		"kne",
    		"knp"
    	]
    },
    	"application/vnd.koan": {
    	source: "iana",
    	extensions: [
    		"skp",
    		"skd",
    		"skt",
    		"skm"
    	]
    },
    	"application/vnd.kodak-descriptor": {
    	source: "iana",
    	extensions: [
    		"sse"
    	]
    },
    	"application/vnd.las": {
    	source: "iana"
    },
    	"application/vnd.las.las+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.las.las+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"lasxml"
    	]
    },
    	"application/vnd.laszip": {
    	source: "iana"
    },
    	"application/vnd.leap+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.liberty-request+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.llamagraphics.life-balance.desktop": {
    	source: "iana",
    	extensions: [
    		"lbd"
    	]
    },
    	"application/vnd.llamagraphics.life-balance.exchange+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"lbe"
    	]
    },
    	"application/vnd.logipipe.circuit+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.loom": {
    	source: "iana"
    },
    	"application/vnd.lotus-1-2-3": {
    	source: "iana",
    	extensions: [
    		"123"
    	]
    },
    	"application/vnd.lotus-approach": {
    	source: "iana",
    	extensions: [
    		"apr"
    	]
    },
    	"application/vnd.lotus-freelance": {
    	source: "iana",
    	extensions: [
    		"pre"
    	]
    },
    	"application/vnd.lotus-notes": {
    	source: "iana",
    	extensions: [
    		"nsf"
    	]
    },
    	"application/vnd.lotus-organizer": {
    	source: "iana",
    	extensions: [
    		"org"
    	]
    },
    	"application/vnd.lotus-screencam": {
    	source: "iana",
    	extensions: [
    		"scm"
    	]
    },
    	"application/vnd.lotus-wordpro": {
    	source: "iana",
    	extensions: [
    		"lwp"
    	]
    },
    	"application/vnd.macports.portpkg": {
    	source: "iana",
    	extensions: [
    		"portpkg"
    	]
    },
    	"application/vnd.mapbox-vector-tile": {
    	source: "iana",
    	extensions: [
    		"mvt"
    	]
    },
    	"application/vnd.marlin.drm.actiontoken+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.marlin.drm.conftoken+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.marlin.drm.license+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.marlin.drm.mdcf": {
    	source: "iana"
    },
    	"application/vnd.mason+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.maxar.archive.3tz+zip": {
    	source: "iana",
    	compressible: false
    },
    	"application/vnd.maxmind.maxmind-db": {
    	source: "iana"
    },
    	"application/vnd.mcd": {
    	source: "iana",
    	extensions: [
    		"mcd"
    	]
    },
    	"application/vnd.medcalcdata": {
    	source: "iana",
    	extensions: [
    		"mc1"
    	]
    },
    	"application/vnd.mediastation.cdkey": {
    	source: "iana",
    	extensions: [
    		"cdkey"
    	]
    },
    	"application/vnd.meridian-slingshot": {
    	source: "iana"
    },
    	"application/vnd.mfer": {
    	source: "iana",
    	extensions: [
    		"mwf"
    	]
    },
    	"application/vnd.mfmp": {
    	source: "iana",
    	extensions: [
    		"mfm"
    	]
    },
    	"application/vnd.micro+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.micrografx.flo": {
    	source: "iana",
    	extensions: [
    		"flo"
    	]
    },
    	"application/vnd.micrografx.igx": {
    	source: "iana",
    	extensions: [
    		"igx"
    	]
    },
    	"application/vnd.microsoft.portable-executable": {
    	source: "iana"
    },
    	"application/vnd.microsoft.windows.thumbnail-cache": {
    	source: "iana"
    },
    	"application/vnd.miele+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.mif": {
    	source: "iana",
    	extensions: [
    		"mif"
    	]
    },
    	"application/vnd.minisoft-hp3000-save": {
    	source: "iana"
    },
    	"application/vnd.mitsubishi.misty-guard.trustweb": {
    	source: "iana"
    },
    	"application/vnd.mobius.daf": {
    	source: "iana",
    	extensions: [
    		"daf"
    	]
    },
    	"application/vnd.mobius.dis": {
    	source: "iana",
    	extensions: [
    		"dis"
    	]
    },
    	"application/vnd.mobius.mbk": {
    	source: "iana",
    	extensions: [
    		"mbk"
    	]
    },
    	"application/vnd.mobius.mqy": {
    	source: "iana",
    	extensions: [
    		"mqy"
    	]
    },
    	"application/vnd.mobius.msl": {
    	source: "iana",
    	extensions: [
    		"msl"
    	]
    },
    	"application/vnd.mobius.plc": {
    	source: "iana",
    	extensions: [
    		"plc"
    	]
    },
    	"application/vnd.mobius.txf": {
    	source: "iana",
    	extensions: [
    		"txf"
    	]
    },
    	"application/vnd.mophun.application": {
    	source: "iana",
    	extensions: [
    		"mpn"
    	]
    },
    	"application/vnd.mophun.certificate": {
    	source: "iana",
    	extensions: [
    		"mpc"
    	]
    },
    	"application/vnd.motorola.flexsuite": {
    	source: "iana"
    },
    	"application/vnd.motorola.flexsuite.adsi": {
    	source: "iana"
    },
    	"application/vnd.motorola.flexsuite.fis": {
    	source: "iana"
    },
    	"application/vnd.motorola.flexsuite.gotap": {
    	source: "iana"
    },
    	"application/vnd.motorola.flexsuite.kmr": {
    	source: "iana"
    },
    	"application/vnd.motorola.flexsuite.ttc": {
    	source: "iana"
    },
    	"application/vnd.motorola.flexsuite.wem": {
    	source: "iana"
    },
    	"application/vnd.motorola.iprm": {
    	source: "iana"
    },
    	"application/vnd.mozilla.xul+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xul"
    	]
    },
    	"application/vnd.ms-3mfdocument": {
    	source: "iana"
    },
    	"application/vnd.ms-artgalry": {
    	source: "iana",
    	extensions: [
    		"cil"
    	]
    },
    	"application/vnd.ms-asf": {
    	source: "iana"
    },
    	"application/vnd.ms-cab-compressed": {
    	source: "iana",
    	extensions: [
    		"cab"
    	]
    },
    	"application/vnd.ms-color.iccprofile": {
    	source: "apache"
    },
    	"application/vnd.ms-excel": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"xls",
    		"xlm",
    		"xla",
    		"xlc",
    		"xlt",
    		"xlw"
    	]
    },
    	"application/vnd.ms-excel.addin.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"xlam"
    	]
    },
    	"application/vnd.ms-excel.sheet.binary.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"xlsb"
    	]
    },
    	"application/vnd.ms-excel.sheet.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"xlsm"
    	]
    },
    	"application/vnd.ms-excel.template.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"xltm"
    	]
    },
    	"application/vnd.ms-fontobject": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"eot"
    	]
    },
    	"application/vnd.ms-htmlhelp": {
    	source: "iana",
    	extensions: [
    		"chm"
    	]
    },
    	"application/vnd.ms-ims": {
    	source: "iana",
    	extensions: [
    		"ims"
    	]
    },
    	"application/vnd.ms-lrm": {
    	source: "iana",
    	extensions: [
    		"lrm"
    	]
    },
    	"application/vnd.ms-office.activex+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ms-officetheme": {
    	source: "iana",
    	extensions: [
    		"thmx"
    	]
    },
    	"application/vnd.ms-opentype": {
    	source: "apache",
    	compressible: true
    },
    	"application/vnd.ms-outlook": {
    	compressible: false,
    	extensions: [
    		"msg"
    	]
    },
    	"application/vnd.ms-package.obfuscated-opentype": {
    	source: "apache"
    },
    	"application/vnd.ms-pki.seccat": {
    	source: "apache",
    	extensions: [
    		"cat"
    	]
    },
    	"application/vnd.ms-pki.stl": {
    	source: "apache",
    	extensions: [
    		"stl"
    	]
    },
    	"application/vnd.ms-playready.initiator+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ms-powerpoint": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"ppt",
    		"pps",
    		"pot"
    	]
    },
    	"application/vnd.ms-powerpoint.addin.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"ppam"
    	]
    },
    	"application/vnd.ms-powerpoint.presentation.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"pptm"
    	]
    },
    	"application/vnd.ms-powerpoint.slide.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"sldm"
    	]
    },
    	"application/vnd.ms-powerpoint.slideshow.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"ppsm"
    	]
    },
    	"application/vnd.ms-powerpoint.template.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"potm"
    	]
    },
    	"application/vnd.ms-printdevicecapabilities+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ms-printing.printticket+xml": {
    	source: "apache",
    	compressible: true
    },
    	"application/vnd.ms-printschematicket+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ms-project": {
    	source: "iana",
    	extensions: [
    		"mpp",
    		"mpt"
    	]
    },
    	"application/vnd.ms-tnef": {
    	source: "iana"
    },
    	"application/vnd.ms-windows.devicepairing": {
    	source: "iana"
    },
    	"application/vnd.ms-windows.nwprinting.oob": {
    	source: "iana"
    },
    	"application/vnd.ms-windows.printerpairing": {
    	source: "iana"
    },
    	"application/vnd.ms-windows.wsd.oob": {
    	source: "iana"
    },
    	"application/vnd.ms-wmdrm.lic-chlg-req": {
    	source: "iana"
    },
    	"application/vnd.ms-wmdrm.lic-resp": {
    	source: "iana"
    },
    	"application/vnd.ms-wmdrm.meter-chlg-req": {
    	source: "iana"
    },
    	"application/vnd.ms-wmdrm.meter-resp": {
    	source: "iana"
    },
    	"application/vnd.ms-word.document.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"docm"
    	]
    },
    	"application/vnd.ms-word.template.macroenabled.12": {
    	source: "iana",
    	extensions: [
    		"dotm"
    	]
    },
    	"application/vnd.ms-works": {
    	source: "iana",
    	extensions: [
    		"wps",
    		"wks",
    		"wcm",
    		"wdb"
    	]
    },
    	"application/vnd.ms-wpl": {
    	source: "iana",
    	extensions: [
    		"wpl"
    	]
    },
    	"application/vnd.ms-xpsdocument": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"xps"
    	]
    },
    	"application/vnd.msa-disk-image": {
    	source: "iana"
    },
    	"application/vnd.mseq": {
    	source: "iana",
    	extensions: [
    		"mseq"
    	]
    },
    	"application/vnd.msign": {
    	source: "iana"
    },
    	"application/vnd.multiad.creator": {
    	source: "iana"
    },
    	"application/vnd.multiad.creator.cif": {
    	source: "iana"
    },
    	"application/vnd.music-niff": {
    	source: "iana"
    },
    	"application/vnd.musician": {
    	source: "iana",
    	extensions: [
    		"mus"
    	]
    },
    	"application/vnd.muvee.style": {
    	source: "iana",
    	extensions: [
    		"msty"
    	]
    },
    	"application/vnd.mynfc": {
    	source: "iana",
    	extensions: [
    		"taglet"
    	]
    },
    	"application/vnd.nacamar.ybrid+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.ncd.control": {
    	source: "iana"
    },
    	"application/vnd.ncd.reference": {
    	source: "iana"
    },
    	"application/vnd.nearst.inv+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.nebumind.line": {
    	source: "iana"
    },
    	"application/vnd.nervana": {
    	source: "iana"
    },
    	"application/vnd.netfpx": {
    	source: "iana"
    },
    	"application/vnd.neurolanguage.nlu": {
    	source: "iana",
    	extensions: [
    		"nlu"
    	]
    },
    	"application/vnd.nimn": {
    	source: "iana"
    },
    	"application/vnd.nintendo.nitro.rom": {
    	source: "iana"
    },
    	"application/vnd.nintendo.snes.rom": {
    	source: "iana"
    },
    	"application/vnd.nitf": {
    	source: "iana",
    	extensions: [
    		"ntf",
    		"nitf"
    	]
    },
    	"application/vnd.noblenet-directory": {
    	source: "iana",
    	extensions: [
    		"nnd"
    	]
    },
    	"application/vnd.noblenet-sealer": {
    	source: "iana",
    	extensions: [
    		"nns"
    	]
    },
    	"application/vnd.noblenet-web": {
    	source: "iana",
    	extensions: [
    		"nnw"
    	]
    },
    	"application/vnd.nokia.catalogs": {
    	source: "iana"
    },
    	"application/vnd.nokia.conml+wbxml": {
    	source: "iana"
    },
    	"application/vnd.nokia.conml+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.nokia.iptv.config+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.nokia.isds-radio-presets": {
    	source: "iana"
    },
    	"application/vnd.nokia.landmark+wbxml": {
    	source: "iana"
    },
    	"application/vnd.nokia.landmark+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.nokia.landmarkcollection+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.nokia.n-gage.ac+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"ac"
    	]
    },
    	"application/vnd.nokia.n-gage.data": {
    	source: "iana",
    	extensions: [
    		"ngdat"
    	]
    },
    	"application/vnd.nokia.n-gage.symbian.install": {
    	source: "iana",
    	extensions: [
    		"n-gage"
    	]
    },
    	"application/vnd.nokia.ncd": {
    	source: "iana"
    },
    	"application/vnd.nokia.pcd+wbxml": {
    	source: "iana"
    },
    	"application/vnd.nokia.pcd+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.nokia.radio-preset": {
    	source: "iana",
    	extensions: [
    		"rpst"
    	]
    },
    	"application/vnd.nokia.radio-presets": {
    	source: "iana",
    	extensions: [
    		"rpss"
    	]
    },
    	"application/vnd.novadigm.edm": {
    	source: "iana",
    	extensions: [
    		"edm"
    	]
    },
    	"application/vnd.novadigm.edx": {
    	source: "iana",
    	extensions: [
    		"edx"
    	]
    },
    	"application/vnd.novadigm.ext": {
    	source: "iana",
    	extensions: [
    		"ext"
    	]
    },
    	"application/vnd.ntt-local.content-share": {
    	source: "iana"
    },
    	"application/vnd.ntt-local.file-transfer": {
    	source: "iana"
    },
    	"application/vnd.ntt-local.ogw_remote-access": {
    	source: "iana"
    },
    	"application/vnd.ntt-local.sip-ta_remote": {
    	source: "iana"
    },
    	"application/vnd.ntt-local.sip-ta_tcp_stream": {
    	source: "iana"
    },
    	"application/vnd.oasis.opendocument.chart": {
    	source: "iana",
    	extensions: [
    		"odc"
    	]
    },
    	"application/vnd.oasis.opendocument.chart-template": {
    	source: "iana",
    	extensions: [
    		"otc"
    	]
    },
    	"application/vnd.oasis.opendocument.database": {
    	source: "iana",
    	extensions: [
    		"odb"
    	]
    },
    	"application/vnd.oasis.opendocument.formula": {
    	source: "iana",
    	extensions: [
    		"odf"
    	]
    },
    	"application/vnd.oasis.opendocument.formula-template": {
    	source: "iana",
    	extensions: [
    		"odft"
    	]
    },
    	"application/vnd.oasis.opendocument.graphics": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"odg"
    	]
    },
    	"application/vnd.oasis.opendocument.graphics-template": {
    	source: "iana",
    	extensions: [
    		"otg"
    	]
    },
    	"application/vnd.oasis.opendocument.image": {
    	source: "iana",
    	extensions: [
    		"odi"
    	]
    },
    	"application/vnd.oasis.opendocument.image-template": {
    	source: "iana",
    	extensions: [
    		"oti"
    	]
    },
    	"application/vnd.oasis.opendocument.presentation": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"odp"
    	]
    },
    	"application/vnd.oasis.opendocument.presentation-template": {
    	source: "iana",
    	extensions: [
    		"otp"
    	]
    },
    	"application/vnd.oasis.opendocument.spreadsheet": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"ods"
    	]
    },
    	"application/vnd.oasis.opendocument.spreadsheet-template": {
    	source: "iana",
    	extensions: [
    		"ots"
    	]
    },
    	"application/vnd.oasis.opendocument.text": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"odt"
    	]
    },
    	"application/vnd.oasis.opendocument.text-master": {
    	source: "iana",
    	extensions: [
    		"odm"
    	]
    },
    	"application/vnd.oasis.opendocument.text-template": {
    	source: "iana",
    	extensions: [
    		"ott"
    	]
    },
    	"application/vnd.oasis.opendocument.text-web": {
    	source: "iana",
    	extensions: [
    		"oth"
    	]
    },
    	"application/vnd.obn": {
    	source: "iana"
    },
    	"application/vnd.ocf+cbor": {
    	source: "iana"
    },
    	"application/vnd.oci.image.manifest.v1+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oftn.l10n+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oipf.contentaccessdownload+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oipf.contentaccessstreaming+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oipf.cspg-hexbinary": {
    	source: "iana"
    },
    	"application/vnd.oipf.dae.svg+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oipf.dae.xhtml+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oipf.mippvcontrolmessage+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oipf.pae.gem": {
    	source: "iana"
    },
    	"application/vnd.oipf.spdiscovery+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oipf.spdlist+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oipf.ueprofile+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oipf.userprofile+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.olpc-sugar": {
    	source: "iana",
    	extensions: [
    		"xo"
    	]
    },
    	"application/vnd.oma-scws-config": {
    	source: "iana"
    },
    	"application/vnd.oma-scws-http-request": {
    	source: "iana"
    },
    	"application/vnd.oma-scws-http-response": {
    	source: "iana"
    },
    	"application/vnd.oma.bcast.associated-procedure-parameter+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.bcast.drm-trigger+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.bcast.imd+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.bcast.ltkm": {
    	source: "iana"
    },
    	"application/vnd.oma.bcast.notification+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.bcast.provisioningtrigger": {
    	source: "iana"
    },
    	"application/vnd.oma.bcast.sgboot": {
    	source: "iana"
    },
    	"application/vnd.oma.bcast.sgdd+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.bcast.sgdu": {
    	source: "iana"
    },
    	"application/vnd.oma.bcast.simple-symbol-container": {
    	source: "iana"
    },
    	"application/vnd.oma.bcast.smartcard-trigger+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.bcast.sprov+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.bcast.stkm": {
    	source: "iana"
    },
    	"application/vnd.oma.cab-address-book+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.cab-feature-handler+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.cab-pcc+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.cab-subs-invite+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.cab-user-prefs+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.dcd": {
    	source: "iana"
    },
    	"application/vnd.oma.dcdc": {
    	source: "iana"
    },
    	"application/vnd.oma.dd2+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"dd2"
    	]
    },
    	"application/vnd.oma.drm.risd+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.group-usage-list+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.lwm2m+cbor": {
    	source: "iana"
    },
    	"application/vnd.oma.lwm2m+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.lwm2m+tlv": {
    	source: "iana"
    },
    	"application/vnd.oma.pal+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.poc.detailed-progress-report+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.poc.final-report+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.poc.groups+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.poc.invocation-descriptor+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.poc.optimized-progress-report+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.push": {
    	source: "iana"
    },
    	"application/vnd.oma.scidm.messages+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oma.xcap-directory+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.omads-email+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/vnd.omads-file+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/vnd.omads-folder+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/vnd.omaloc-supl-init": {
    	source: "iana"
    },
    	"application/vnd.onepager": {
    	source: "iana"
    },
    	"application/vnd.onepagertamp": {
    	source: "iana"
    },
    	"application/vnd.onepagertamx": {
    	source: "iana"
    },
    	"application/vnd.onepagertat": {
    	source: "iana"
    },
    	"application/vnd.onepagertatp": {
    	source: "iana"
    },
    	"application/vnd.onepagertatx": {
    	source: "iana"
    },
    	"application/vnd.openblox.game+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"obgx"
    	]
    },
    	"application/vnd.openblox.game-binary": {
    	source: "iana"
    },
    	"application/vnd.openeye.oeb": {
    	source: "iana"
    },
    	"application/vnd.openofficeorg.extension": {
    	source: "apache",
    	extensions: [
    		"oxt"
    	]
    },
    	"application/vnd.openstreetmap.data+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"osm"
    	]
    },
    	"application/vnd.opentimestamps.ots": {
    	source: "iana"
    },
    	"application/vnd.openxmlformats-officedocument.custom-properties+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.customxmlproperties+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.drawing+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.drawingml.chart+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.drawingml.chartshapes+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.drawingml.diagramcolors+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.drawingml.diagramdata+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.drawingml.diagramlayout+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.drawingml.diagramstyle+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.extended-properties+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.commentauthors+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.comments+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.handoutmaster+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.notesmaster+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.notesslide+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.presentation": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"pptx"
    	]
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.presentation.main+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.presprops+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.slide": {
    	source: "iana",
    	extensions: [
    		"sldx"
    	]
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.slide+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.slidelayout+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.slidemaster+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.slideshow": {
    	source: "iana",
    	extensions: [
    		"ppsx"
    	]
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.slideshow.main+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.slideupdateinfo+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.tablestyles+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.tags+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.template": {
    	source: "iana",
    	extensions: [
    		"potx"
    	]
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.template.main+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.presentationml.viewprops+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.calcchain+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.chartsheet+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.comments+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.connections+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.dialogsheet+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.externallink+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.pivotcachedefinition+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.pivotcacherecords+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.pivottable+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.querytable+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.revisionheaders+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.revisionlog+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.sharedstrings+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"xlsx"
    	]
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.sheetmetadata+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.styles+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.table+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.tablesinglecells+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.template": {
    	source: "iana",
    	extensions: [
    		"xltx"
    	]
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.template.main+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.usernames+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.volatiledependencies+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.theme+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.themeoverride+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.vmldrawing": {
    	source: "iana"
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.comments+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.document": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"docx"
    	]
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.document.glossary+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.endnotes+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.fonttable+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.footer+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.footnotes+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.numbering+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.settings+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.styles+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.template": {
    	source: "iana",
    	extensions: [
    		"dotx"
    	]
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.template.main+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-officedocument.wordprocessingml.websettings+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-package.core-properties+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-package.digital-signature-xmlsignature+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.openxmlformats-package.relationships+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oracle.resource+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.orange.indata": {
    	source: "iana"
    },
    	"application/vnd.osa.netdeploy": {
    	source: "iana"
    },
    	"application/vnd.osgeo.mapguide.package": {
    	source: "iana",
    	extensions: [
    		"mgp"
    	]
    },
    	"application/vnd.osgi.bundle": {
    	source: "iana"
    },
    	"application/vnd.osgi.dp": {
    	source: "iana",
    	extensions: [
    		"dp"
    	]
    },
    	"application/vnd.osgi.subsystem": {
    	source: "iana",
    	extensions: [
    		"esa"
    	]
    },
    	"application/vnd.otps.ct-kip+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.oxli.countgraph": {
    	source: "iana"
    },
    	"application/vnd.pagerduty+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.palm": {
    	source: "iana",
    	extensions: [
    		"pdb",
    		"pqa",
    		"oprc"
    	]
    },
    	"application/vnd.panoply": {
    	source: "iana"
    },
    	"application/vnd.paos.xml": {
    	source: "iana"
    },
    	"application/vnd.patentdive": {
    	source: "iana"
    },
    	"application/vnd.patientecommsdoc": {
    	source: "iana"
    },
    	"application/vnd.pawaafile": {
    	source: "iana",
    	extensions: [
    		"paw"
    	]
    },
    	"application/vnd.pcos": {
    	source: "iana"
    },
    	"application/vnd.pg.format": {
    	source: "iana",
    	extensions: [
    		"str"
    	]
    },
    	"application/vnd.pg.osasli": {
    	source: "iana",
    	extensions: [
    		"ei6"
    	]
    },
    	"application/vnd.piaccess.application-licence": {
    	source: "iana"
    },
    	"application/vnd.picsel": {
    	source: "iana",
    	extensions: [
    		"efif"
    	]
    },
    	"application/vnd.pmi.widget": {
    	source: "iana",
    	extensions: [
    		"wg"
    	]
    },
    	"application/vnd.poc.group-advertisement+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.pocketlearn": {
    	source: "iana",
    	extensions: [
    		"plf"
    	]
    },
    	"application/vnd.powerbuilder6": {
    	source: "iana",
    	extensions: [
    		"pbd"
    	]
    },
    	"application/vnd.powerbuilder6-s": {
    	source: "iana"
    },
    	"application/vnd.powerbuilder7": {
    	source: "iana"
    },
    	"application/vnd.powerbuilder7-s": {
    	source: "iana"
    },
    	"application/vnd.powerbuilder75": {
    	source: "iana"
    },
    	"application/vnd.powerbuilder75-s": {
    	source: "iana"
    },
    	"application/vnd.preminet": {
    	source: "iana"
    },
    	"application/vnd.previewsystems.box": {
    	source: "iana",
    	extensions: [
    		"box"
    	]
    },
    	"application/vnd.proteus.magazine": {
    	source: "iana",
    	extensions: [
    		"mgz"
    	]
    },
    	"application/vnd.psfs": {
    	source: "iana"
    },
    	"application/vnd.publishare-delta-tree": {
    	source: "iana",
    	extensions: [
    		"qps"
    	]
    },
    	"application/vnd.pvi.ptid1": {
    	source: "iana",
    	extensions: [
    		"ptid"
    	]
    },
    	"application/vnd.pwg-multiplexed": {
    	source: "iana"
    },
    	"application/vnd.pwg-xhtml-print+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.qualcomm.brew-app-res": {
    	source: "iana"
    },
    	"application/vnd.quarantainenet": {
    	source: "iana"
    },
    	"application/vnd.quark.quarkxpress": {
    	source: "iana",
    	extensions: [
    		"qxd",
    		"qxt",
    		"qwd",
    		"qwt",
    		"qxl",
    		"qxb"
    	]
    },
    	"application/vnd.quobject-quoxdocument": {
    	source: "iana"
    },
    	"application/vnd.radisys.moml+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-audit+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-audit-conf+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-audit-conn+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-audit-dialog+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-audit-stream+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-conf+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-dialog+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-dialog-base+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-dialog-fax-detect+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-dialog-fax-sendrecv+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-dialog-group+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-dialog-speech+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.radisys.msml-dialog-transform+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.rainstor.data": {
    	source: "iana"
    },
    	"application/vnd.rapid": {
    	source: "iana"
    },
    	"application/vnd.rar": {
    	source: "iana",
    	extensions: [
    		"rar"
    	]
    },
    	"application/vnd.realvnc.bed": {
    	source: "iana",
    	extensions: [
    		"bed"
    	]
    },
    	"application/vnd.recordare.musicxml": {
    	source: "iana",
    	extensions: [
    		"mxl"
    	]
    },
    	"application/vnd.recordare.musicxml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"musicxml"
    	]
    },
    	"application/vnd.renlearn.rlprint": {
    	source: "iana"
    },
    	"application/vnd.resilient.logic": {
    	source: "iana"
    },
    	"application/vnd.restful+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.rig.cryptonote": {
    	source: "iana",
    	extensions: [
    		"cryptonote"
    	]
    },
    	"application/vnd.rim.cod": {
    	source: "apache",
    	extensions: [
    		"cod"
    	]
    },
    	"application/vnd.rn-realmedia": {
    	source: "apache",
    	extensions: [
    		"rm"
    	]
    },
    	"application/vnd.rn-realmedia-vbr": {
    	source: "apache",
    	extensions: [
    		"rmvb"
    	]
    },
    	"application/vnd.route66.link66+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"link66"
    	]
    },
    	"application/vnd.rs-274x": {
    	source: "iana"
    },
    	"application/vnd.ruckus.download": {
    	source: "iana"
    },
    	"application/vnd.s3sms": {
    	source: "iana"
    },
    	"application/vnd.sailingtracker.track": {
    	source: "iana",
    	extensions: [
    		"st"
    	]
    },
    	"application/vnd.sar": {
    	source: "iana"
    },
    	"application/vnd.sbm.cid": {
    	source: "iana"
    },
    	"application/vnd.sbm.mid2": {
    	source: "iana"
    },
    	"application/vnd.scribus": {
    	source: "iana"
    },
    	"application/vnd.sealed.3df": {
    	source: "iana"
    },
    	"application/vnd.sealed.csf": {
    	source: "iana"
    },
    	"application/vnd.sealed.doc": {
    	source: "iana"
    },
    	"application/vnd.sealed.eml": {
    	source: "iana"
    },
    	"application/vnd.sealed.mht": {
    	source: "iana"
    },
    	"application/vnd.sealed.net": {
    	source: "iana"
    },
    	"application/vnd.sealed.ppt": {
    	source: "iana"
    },
    	"application/vnd.sealed.tiff": {
    	source: "iana"
    },
    	"application/vnd.sealed.xls": {
    	source: "iana"
    },
    	"application/vnd.sealedmedia.softseal.html": {
    	source: "iana"
    },
    	"application/vnd.sealedmedia.softseal.pdf": {
    	source: "iana"
    },
    	"application/vnd.seemail": {
    	source: "iana",
    	extensions: [
    		"see"
    	]
    },
    	"application/vnd.seis+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.sema": {
    	source: "iana",
    	extensions: [
    		"sema"
    	]
    },
    	"application/vnd.semd": {
    	source: "iana",
    	extensions: [
    		"semd"
    	]
    },
    	"application/vnd.semf": {
    	source: "iana",
    	extensions: [
    		"semf"
    	]
    },
    	"application/vnd.shade-save-file": {
    	source: "iana"
    },
    	"application/vnd.shana.informed.formdata": {
    	source: "iana",
    	extensions: [
    		"ifm"
    	]
    },
    	"application/vnd.shana.informed.formtemplate": {
    	source: "iana",
    	extensions: [
    		"itp"
    	]
    },
    	"application/vnd.shana.informed.interchange": {
    	source: "iana",
    	extensions: [
    		"iif"
    	]
    },
    	"application/vnd.shana.informed.package": {
    	source: "iana",
    	extensions: [
    		"ipk"
    	]
    },
    	"application/vnd.shootproof+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.shopkick+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.shp": {
    	source: "iana"
    },
    	"application/vnd.shx": {
    	source: "iana"
    },
    	"application/vnd.sigrok.session": {
    	source: "iana"
    },
    	"application/vnd.simtech-mindmapper": {
    	source: "iana",
    	extensions: [
    		"twd",
    		"twds"
    	]
    },
    	"application/vnd.siren+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.smaf": {
    	source: "iana",
    	extensions: [
    		"mmf"
    	]
    },
    	"application/vnd.smart.notebook": {
    	source: "iana"
    },
    	"application/vnd.smart.teacher": {
    	source: "iana",
    	extensions: [
    		"teacher"
    	]
    },
    	"application/vnd.snesdev-page-table": {
    	source: "iana"
    },
    	"application/vnd.software602.filler.form+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"fo"
    	]
    },
    	"application/vnd.software602.filler.form-xml-zip": {
    	source: "iana"
    },
    	"application/vnd.solent.sdkm+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"sdkm",
    		"sdkd"
    	]
    },
    	"application/vnd.spotfire.dxp": {
    	source: "iana",
    	extensions: [
    		"dxp"
    	]
    },
    	"application/vnd.spotfire.sfs": {
    	source: "iana",
    	extensions: [
    		"sfs"
    	]
    },
    	"application/vnd.sqlite3": {
    	source: "iana"
    },
    	"application/vnd.sss-cod": {
    	source: "iana"
    },
    	"application/vnd.sss-dtf": {
    	source: "iana"
    },
    	"application/vnd.sss-ntf": {
    	source: "iana"
    },
    	"application/vnd.stardivision.calc": {
    	source: "apache",
    	extensions: [
    		"sdc"
    	]
    },
    	"application/vnd.stardivision.draw": {
    	source: "apache",
    	extensions: [
    		"sda"
    	]
    },
    	"application/vnd.stardivision.impress": {
    	source: "apache",
    	extensions: [
    		"sdd"
    	]
    },
    	"application/vnd.stardivision.math": {
    	source: "apache",
    	extensions: [
    		"smf"
    	]
    },
    	"application/vnd.stardivision.writer": {
    	source: "apache",
    	extensions: [
    		"sdw",
    		"vor"
    	]
    },
    	"application/vnd.stardivision.writer-global": {
    	source: "apache",
    	extensions: [
    		"sgl"
    	]
    },
    	"application/vnd.stepmania.package": {
    	source: "iana",
    	extensions: [
    		"smzip"
    	]
    },
    	"application/vnd.stepmania.stepchart": {
    	source: "iana",
    	extensions: [
    		"sm"
    	]
    },
    	"application/vnd.street-stream": {
    	source: "iana"
    },
    	"application/vnd.sun.wadl+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"wadl"
    	]
    },
    	"application/vnd.sun.xml.calc": {
    	source: "apache",
    	extensions: [
    		"sxc"
    	]
    },
    	"application/vnd.sun.xml.calc.template": {
    	source: "apache",
    	extensions: [
    		"stc"
    	]
    },
    	"application/vnd.sun.xml.draw": {
    	source: "apache",
    	extensions: [
    		"sxd"
    	]
    },
    	"application/vnd.sun.xml.draw.template": {
    	source: "apache",
    	extensions: [
    		"std"
    	]
    },
    	"application/vnd.sun.xml.impress": {
    	source: "apache",
    	extensions: [
    		"sxi"
    	]
    },
    	"application/vnd.sun.xml.impress.template": {
    	source: "apache",
    	extensions: [
    		"sti"
    	]
    },
    	"application/vnd.sun.xml.math": {
    	source: "apache",
    	extensions: [
    		"sxm"
    	]
    },
    	"application/vnd.sun.xml.writer": {
    	source: "apache",
    	extensions: [
    		"sxw"
    	]
    },
    	"application/vnd.sun.xml.writer.global": {
    	source: "apache",
    	extensions: [
    		"sxg"
    	]
    },
    	"application/vnd.sun.xml.writer.template": {
    	source: "apache",
    	extensions: [
    		"stw"
    	]
    },
    	"application/vnd.sus-calendar": {
    	source: "iana",
    	extensions: [
    		"sus",
    		"susp"
    	]
    },
    	"application/vnd.svd": {
    	source: "iana",
    	extensions: [
    		"svd"
    	]
    },
    	"application/vnd.swiftview-ics": {
    	source: "iana"
    },
    	"application/vnd.sycle+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.syft+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.symbian.install": {
    	source: "apache",
    	extensions: [
    		"sis",
    		"sisx"
    	]
    },
    	"application/vnd.syncml+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true,
    	extensions: [
    		"xsm"
    	]
    },
    	"application/vnd.syncml.dm+wbxml": {
    	source: "iana",
    	charset: "UTF-8",
    	extensions: [
    		"bdm"
    	]
    },
    	"application/vnd.syncml.dm+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true,
    	extensions: [
    		"xdm"
    	]
    },
    	"application/vnd.syncml.dm.notification": {
    	source: "iana"
    },
    	"application/vnd.syncml.dmddf+wbxml": {
    	source: "iana"
    },
    	"application/vnd.syncml.dmddf+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true,
    	extensions: [
    		"ddf"
    	]
    },
    	"application/vnd.syncml.dmtnds+wbxml": {
    	source: "iana"
    },
    	"application/vnd.syncml.dmtnds+xml": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true
    },
    	"application/vnd.syncml.ds.notification": {
    	source: "iana"
    },
    	"application/vnd.tableschema+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.tao.intent-module-archive": {
    	source: "iana",
    	extensions: [
    		"tao"
    	]
    },
    	"application/vnd.tcpdump.pcap": {
    	source: "iana",
    	extensions: [
    		"pcap",
    		"cap",
    		"dmp"
    	]
    },
    	"application/vnd.think-cell.ppttc+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.tmd.mediaflex.api+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.tml": {
    	source: "iana"
    },
    	"application/vnd.tmobile-livetv": {
    	source: "iana",
    	extensions: [
    		"tmo"
    	]
    },
    	"application/vnd.tri.onesource": {
    	source: "iana"
    },
    	"application/vnd.trid.tpt": {
    	source: "iana",
    	extensions: [
    		"tpt"
    	]
    },
    	"application/vnd.triscape.mxs": {
    	source: "iana",
    	extensions: [
    		"mxs"
    	]
    },
    	"application/vnd.trueapp": {
    	source: "iana",
    	extensions: [
    		"tra"
    	]
    },
    	"application/vnd.truedoc": {
    	source: "iana"
    },
    	"application/vnd.ubisoft.webplayer": {
    	source: "iana"
    },
    	"application/vnd.ufdl": {
    	source: "iana",
    	extensions: [
    		"ufd",
    		"ufdl"
    	]
    },
    	"application/vnd.uiq.theme": {
    	source: "iana",
    	extensions: [
    		"utz"
    	]
    },
    	"application/vnd.umajin": {
    	source: "iana",
    	extensions: [
    		"umj"
    	]
    },
    	"application/vnd.unity": {
    	source: "iana",
    	extensions: [
    		"unityweb"
    	]
    },
    	"application/vnd.uoml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"uoml"
    	]
    },
    	"application/vnd.uplanet.alert": {
    	source: "iana"
    },
    	"application/vnd.uplanet.alert-wbxml": {
    	source: "iana"
    },
    	"application/vnd.uplanet.bearer-choice": {
    	source: "iana"
    },
    	"application/vnd.uplanet.bearer-choice-wbxml": {
    	source: "iana"
    },
    	"application/vnd.uplanet.cacheop": {
    	source: "iana"
    },
    	"application/vnd.uplanet.cacheop-wbxml": {
    	source: "iana"
    },
    	"application/vnd.uplanet.channel": {
    	source: "iana"
    },
    	"application/vnd.uplanet.channel-wbxml": {
    	source: "iana"
    },
    	"application/vnd.uplanet.list": {
    	source: "iana"
    },
    	"application/vnd.uplanet.list-wbxml": {
    	source: "iana"
    },
    	"application/vnd.uplanet.listcmd": {
    	source: "iana"
    },
    	"application/vnd.uplanet.listcmd-wbxml": {
    	source: "iana"
    },
    	"application/vnd.uplanet.signal": {
    	source: "iana"
    },
    	"application/vnd.uri-map": {
    	source: "iana"
    },
    	"application/vnd.valve.source.material": {
    	source: "iana"
    },
    	"application/vnd.vcx": {
    	source: "iana",
    	extensions: [
    		"vcx"
    	]
    },
    	"application/vnd.vd-study": {
    	source: "iana"
    },
    	"application/vnd.vectorworks": {
    	source: "iana"
    },
    	"application/vnd.vel+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.verimatrix.vcas": {
    	source: "iana"
    },
    	"application/vnd.veritone.aion+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.veryant.thin": {
    	source: "iana"
    },
    	"application/vnd.ves.encrypted": {
    	source: "iana"
    },
    	"application/vnd.vidsoft.vidconference": {
    	source: "iana"
    },
    	"application/vnd.visio": {
    	source: "iana",
    	extensions: [
    		"vsd",
    		"vst",
    		"vss",
    		"vsw"
    	]
    },
    	"application/vnd.visionary": {
    	source: "iana",
    	extensions: [
    		"vis"
    	]
    },
    	"application/vnd.vividence.scriptfile": {
    	source: "iana"
    },
    	"application/vnd.vsf": {
    	source: "iana",
    	extensions: [
    		"vsf"
    	]
    },
    	"application/vnd.wap.sic": {
    	source: "iana"
    },
    	"application/vnd.wap.slc": {
    	source: "iana"
    },
    	"application/vnd.wap.wbxml": {
    	source: "iana",
    	charset: "UTF-8",
    	extensions: [
    		"wbxml"
    	]
    },
    	"application/vnd.wap.wmlc": {
    	source: "iana",
    	extensions: [
    		"wmlc"
    	]
    },
    	"application/vnd.wap.wmlscriptc": {
    	source: "iana",
    	extensions: [
    		"wmlsc"
    	]
    },
    	"application/vnd.webturbo": {
    	source: "iana",
    	extensions: [
    		"wtb"
    	]
    },
    	"application/vnd.wfa.dpp": {
    	source: "iana"
    },
    	"application/vnd.wfa.p2p": {
    	source: "iana"
    },
    	"application/vnd.wfa.wsc": {
    	source: "iana"
    },
    	"application/vnd.windows.devicepairing": {
    	source: "iana"
    },
    	"application/vnd.wmc": {
    	source: "iana"
    },
    	"application/vnd.wmf.bootstrap": {
    	source: "iana"
    },
    	"application/vnd.wolfram.mathematica": {
    	source: "iana"
    },
    	"application/vnd.wolfram.mathematica.package": {
    	source: "iana"
    },
    	"application/vnd.wolfram.player": {
    	source: "iana",
    	extensions: [
    		"nbp"
    	]
    },
    	"application/vnd.wordperfect": {
    	source: "iana",
    	extensions: [
    		"wpd"
    	]
    },
    	"application/vnd.wqd": {
    	source: "iana",
    	extensions: [
    		"wqd"
    	]
    },
    	"application/vnd.wrq-hp3000-labelled": {
    	source: "iana"
    },
    	"application/vnd.wt.stf": {
    	source: "iana",
    	extensions: [
    		"stf"
    	]
    },
    	"application/vnd.wv.csp+wbxml": {
    	source: "iana"
    },
    	"application/vnd.wv.csp+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.wv.ssp+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.xacml+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.xara": {
    	source: "iana",
    	extensions: [
    		"xar"
    	]
    },
    	"application/vnd.xfdl": {
    	source: "iana",
    	extensions: [
    		"xfdl"
    	]
    },
    	"application/vnd.xfdl.webform": {
    	source: "iana"
    },
    	"application/vnd.xmi+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/vnd.xmpie.cpkg": {
    	source: "iana"
    },
    	"application/vnd.xmpie.dpkg": {
    	source: "iana"
    },
    	"application/vnd.xmpie.plan": {
    	source: "iana"
    },
    	"application/vnd.xmpie.ppkg": {
    	source: "iana"
    },
    	"application/vnd.xmpie.xlim": {
    	source: "iana"
    },
    	"application/vnd.yamaha.hv-dic": {
    	source: "iana",
    	extensions: [
    		"hvd"
    	]
    },
    	"application/vnd.yamaha.hv-script": {
    	source: "iana",
    	extensions: [
    		"hvs"
    	]
    },
    	"application/vnd.yamaha.hv-voice": {
    	source: "iana",
    	extensions: [
    		"hvp"
    	]
    },
    	"application/vnd.yamaha.openscoreformat": {
    	source: "iana",
    	extensions: [
    		"osf"
    	]
    },
    	"application/vnd.yamaha.openscoreformat.osfpvg+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"osfpvg"
    	]
    },
    	"application/vnd.yamaha.remote-setup": {
    	source: "iana"
    },
    	"application/vnd.yamaha.smaf-audio": {
    	source: "iana",
    	extensions: [
    		"saf"
    	]
    },
    	"application/vnd.yamaha.smaf-phrase": {
    	source: "iana",
    	extensions: [
    		"spf"
    	]
    },
    	"application/vnd.yamaha.through-ngn": {
    	source: "iana"
    },
    	"application/vnd.yamaha.tunnel-udpencap": {
    	source: "iana"
    },
    	"application/vnd.yaoweme": {
    	source: "iana"
    },
    	"application/vnd.yellowriver-custom-menu": {
    	source: "iana",
    	extensions: [
    		"cmp"
    	]
    },
    	"application/vnd.youtube.yt": {
    	source: "iana"
    },
    	"application/vnd.zul": {
    	source: "iana",
    	extensions: [
    		"zir",
    		"zirz"
    	]
    },
    	"application/vnd.zzazz.deck+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"zaz"
    	]
    },
    	"application/voicexml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"vxml"
    	]
    },
    	"application/voucher-cms+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/vq-rtcpxr": {
    	source: "iana"
    },
    	"application/wasm": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"wasm"
    	]
    },
    	"application/watcherinfo+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"wif"
    	]
    },
    	"application/webpush-options+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/whoispp-query": {
    	source: "iana"
    },
    	"application/whoispp-response": {
    	source: "iana"
    },
    	"application/widget": {
    	source: "iana",
    	extensions: [
    		"wgt"
    	]
    },
    	"application/winhlp": {
    	source: "apache",
    	extensions: [
    		"hlp"
    	]
    },
    	"application/wita": {
    	source: "iana"
    },
    	"application/wordperfect5.1": {
    	source: "iana"
    },
    	"application/wsdl+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"wsdl"
    	]
    },
    	"application/wspolicy+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"wspolicy"
    	]
    },
    	"application/x-7z-compressed": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"7z"
    	]
    },
    	"application/x-abiword": {
    	source: "apache",
    	extensions: [
    		"abw"
    	]
    },
    	"application/x-ace-compressed": {
    	source: "apache",
    	extensions: [
    		"ace"
    	]
    },
    	"application/x-amf": {
    	source: "apache"
    },
    	"application/x-apple-diskimage": {
    	source: "apache",
    	extensions: [
    		"dmg"
    	]
    },
    	"application/x-arj": {
    	compressible: false,
    	extensions: [
    		"arj"
    	]
    },
    	"application/x-authorware-bin": {
    	source: "apache",
    	extensions: [
    		"aab",
    		"x32",
    		"u32",
    		"vox"
    	]
    },
    	"application/x-authorware-map": {
    	source: "apache",
    	extensions: [
    		"aam"
    	]
    },
    	"application/x-authorware-seg": {
    	source: "apache",
    	extensions: [
    		"aas"
    	]
    },
    	"application/x-bcpio": {
    	source: "apache",
    	extensions: [
    		"bcpio"
    	]
    },
    	"application/x-bdoc": {
    	compressible: false,
    	extensions: [
    		"bdoc"
    	]
    },
    	"application/x-bittorrent": {
    	source: "apache",
    	extensions: [
    		"torrent"
    	]
    },
    	"application/x-blorb": {
    	source: "apache",
    	extensions: [
    		"blb",
    		"blorb"
    	]
    },
    	"application/x-bzip": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"bz"
    	]
    },
    	"application/x-bzip2": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"bz2",
    		"boz"
    	]
    },
    	"application/x-cbr": {
    	source: "apache",
    	extensions: [
    		"cbr",
    		"cba",
    		"cbt",
    		"cbz",
    		"cb7"
    	]
    },
    	"application/x-cdlink": {
    	source: "apache",
    	extensions: [
    		"vcd"
    	]
    },
    	"application/x-cfs-compressed": {
    	source: "apache",
    	extensions: [
    		"cfs"
    	]
    },
    	"application/x-chat": {
    	source: "apache",
    	extensions: [
    		"chat"
    	]
    },
    	"application/x-chess-pgn": {
    	source: "apache",
    	extensions: [
    		"pgn"
    	]
    },
    	"application/x-chrome-extension": {
    	extensions: [
    		"crx"
    	]
    },
    	"application/x-cocoa": {
    	source: "nginx",
    	extensions: [
    		"cco"
    	]
    },
    	"application/x-compress": {
    	source: "apache"
    },
    	"application/x-conference": {
    	source: "apache",
    	extensions: [
    		"nsc"
    	]
    },
    	"application/x-cpio": {
    	source: "apache",
    	extensions: [
    		"cpio"
    	]
    },
    	"application/x-csh": {
    	source: "apache",
    	extensions: [
    		"csh"
    	]
    },
    	"application/x-deb": {
    	compressible: false
    },
    	"application/x-debian-package": {
    	source: "apache",
    	extensions: [
    		"deb",
    		"udeb"
    	]
    },
    	"application/x-dgc-compressed": {
    	source: "apache",
    	extensions: [
    		"dgc"
    	]
    },
    	"application/x-director": {
    	source: "apache",
    	extensions: [
    		"dir",
    		"dcr",
    		"dxr",
    		"cst",
    		"cct",
    		"cxt",
    		"w3d",
    		"fgd",
    		"swa"
    	]
    },
    	"application/x-doom": {
    	source: "apache",
    	extensions: [
    		"wad"
    	]
    },
    	"application/x-dtbncx+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"ncx"
    	]
    },
    	"application/x-dtbook+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"dtb"
    	]
    },
    	"application/x-dtbresource+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"res"
    	]
    },
    	"application/x-dvi": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"dvi"
    	]
    },
    	"application/x-envoy": {
    	source: "apache",
    	extensions: [
    		"evy"
    	]
    },
    	"application/x-eva": {
    	source: "apache",
    	extensions: [
    		"eva"
    	]
    },
    	"application/x-font-bdf": {
    	source: "apache",
    	extensions: [
    		"bdf"
    	]
    },
    	"application/x-font-dos": {
    	source: "apache"
    },
    	"application/x-font-framemaker": {
    	source: "apache"
    },
    	"application/x-font-ghostscript": {
    	source: "apache",
    	extensions: [
    		"gsf"
    	]
    },
    	"application/x-font-libgrx": {
    	source: "apache"
    },
    	"application/x-font-linux-psf": {
    	source: "apache",
    	extensions: [
    		"psf"
    	]
    },
    	"application/x-font-pcf": {
    	source: "apache",
    	extensions: [
    		"pcf"
    	]
    },
    	"application/x-font-snf": {
    	source: "apache",
    	extensions: [
    		"snf"
    	]
    },
    	"application/x-font-speedo": {
    	source: "apache"
    },
    	"application/x-font-sunos-news": {
    	source: "apache"
    },
    	"application/x-font-type1": {
    	source: "apache",
    	extensions: [
    		"pfa",
    		"pfb",
    		"pfm",
    		"afm"
    	]
    },
    	"application/x-font-vfont": {
    	source: "apache"
    },
    	"application/x-freearc": {
    	source: "apache",
    	extensions: [
    		"arc"
    	]
    },
    	"application/x-futuresplash": {
    	source: "apache",
    	extensions: [
    		"spl"
    	]
    },
    	"application/x-gca-compressed": {
    	source: "apache",
    	extensions: [
    		"gca"
    	]
    },
    	"application/x-glulx": {
    	source: "apache",
    	extensions: [
    		"ulx"
    	]
    },
    	"application/x-gnumeric": {
    	source: "apache",
    	extensions: [
    		"gnumeric"
    	]
    },
    	"application/x-gramps-xml": {
    	source: "apache",
    	extensions: [
    		"gramps"
    	]
    },
    	"application/x-gtar": {
    	source: "apache",
    	extensions: [
    		"gtar"
    	]
    },
    	"application/x-gzip": {
    	source: "apache"
    },
    	"application/x-hdf": {
    	source: "apache",
    	extensions: [
    		"hdf"
    	]
    },
    	"application/x-httpd-php": {
    	compressible: true,
    	extensions: [
    		"php"
    	]
    },
    	"application/x-install-instructions": {
    	source: "apache",
    	extensions: [
    		"install"
    	]
    },
    	"application/x-iso9660-image": {
    	source: "apache",
    	extensions: [
    		"iso"
    	]
    },
    	"application/x-iwork-keynote-sffkey": {
    	extensions: [
    		"key"
    	]
    },
    	"application/x-iwork-numbers-sffnumbers": {
    	extensions: [
    		"numbers"
    	]
    },
    	"application/x-iwork-pages-sffpages": {
    	extensions: [
    		"pages"
    	]
    },
    	"application/x-java-archive-diff": {
    	source: "nginx",
    	extensions: [
    		"jardiff"
    	]
    },
    	"application/x-java-jnlp-file": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"jnlp"
    	]
    },
    	"application/x-javascript": {
    	compressible: true
    },
    	"application/x-keepass2": {
    	extensions: [
    		"kdbx"
    	]
    },
    	"application/x-latex": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"latex"
    	]
    },
    	"application/x-lua-bytecode": {
    	extensions: [
    		"luac"
    	]
    },
    	"application/x-lzh-compressed": {
    	source: "apache",
    	extensions: [
    		"lzh",
    		"lha"
    	]
    },
    	"application/x-makeself": {
    	source: "nginx",
    	extensions: [
    		"run"
    	]
    },
    	"application/x-mie": {
    	source: "apache",
    	extensions: [
    		"mie"
    	]
    },
    	"application/x-mobipocket-ebook": {
    	source: "apache",
    	extensions: [
    		"prc",
    		"mobi"
    	]
    },
    	"application/x-mpegurl": {
    	compressible: false
    },
    	"application/x-ms-application": {
    	source: "apache",
    	extensions: [
    		"application"
    	]
    },
    	"application/x-ms-shortcut": {
    	source: "apache",
    	extensions: [
    		"lnk"
    	]
    },
    	"application/x-ms-wmd": {
    	source: "apache",
    	extensions: [
    		"wmd"
    	]
    },
    	"application/x-ms-wmz": {
    	source: "apache",
    	extensions: [
    		"wmz"
    	]
    },
    	"application/x-ms-xbap": {
    	source: "apache",
    	extensions: [
    		"xbap"
    	]
    },
    	"application/x-msaccess": {
    	source: "apache",
    	extensions: [
    		"mdb"
    	]
    },
    	"application/x-msbinder": {
    	source: "apache",
    	extensions: [
    		"obd"
    	]
    },
    	"application/x-mscardfile": {
    	source: "apache",
    	extensions: [
    		"crd"
    	]
    },
    	"application/x-msclip": {
    	source: "apache",
    	extensions: [
    		"clp"
    	]
    },
    	"application/x-msdos-program": {
    	extensions: [
    		"exe"
    	]
    },
    	"application/x-msdownload": {
    	source: "apache",
    	extensions: [
    		"exe",
    		"dll",
    		"com",
    		"bat",
    		"msi"
    	]
    },
    	"application/x-msmediaview": {
    	source: "apache",
    	extensions: [
    		"mvb",
    		"m13",
    		"m14"
    	]
    },
    	"application/x-msmetafile": {
    	source: "apache",
    	extensions: [
    		"wmf",
    		"wmz",
    		"emf",
    		"emz"
    	]
    },
    	"application/x-msmoney": {
    	source: "apache",
    	extensions: [
    		"mny"
    	]
    },
    	"application/x-mspublisher": {
    	source: "apache",
    	extensions: [
    		"pub"
    	]
    },
    	"application/x-msschedule": {
    	source: "apache",
    	extensions: [
    		"scd"
    	]
    },
    	"application/x-msterminal": {
    	source: "apache",
    	extensions: [
    		"trm"
    	]
    },
    	"application/x-mswrite": {
    	source: "apache",
    	extensions: [
    		"wri"
    	]
    },
    	"application/x-netcdf": {
    	source: "apache",
    	extensions: [
    		"nc",
    		"cdf"
    	]
    },
    	"application/x-ns-proxy-autoconfig": {
    	compressible: true,
    	extensions: [
    		"pac"
    	]
    },
    	"application/x-nzb": {
    	source: "apache",
    	extensions: [
    		"nzb"
    	]
    },
    	"application/x-perl": {
    	source: "nginx",
    	extensions: [
    		"pl",
    		"pm"
    	]
    },
    	"application/x-pilot": {
    	source: "nginx",
    	extensions: [
    		"prc",
    		"pdb"
    	]
    },
    	"application/x-pkcs12": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"p12",
    		"pfx"
    	]
    },
    	"application/x-pkcs7-certificates": {
    	source: "apache",
    	extensions: [
    		"p7b",
    		"spc"
    	]
    },
    	"application/x-pkcs7-certreqresp": {
    	source: "apache",
    	extensions: [
    		"p7r"
    	]
    },
    	"application/x-pki-message": {
    	source: "iana"
    },
    	"application/x-rar-compressed": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"rar"
    	]
    },
    	"application/x-redhat-package-manager": {
    	source: "nginx",
    	extensions: [
    		"rpm"
    	]
    },
    	"application/x-research-info-systems": {
    	source: "apache",
    	extensions: [
    		"ris"
    	]
    },
    	"application/x-sea": {
    	source: "nginx",
    	extensions: [
    		"sea"
    	]
    },
    	"application/x-sh": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"sh"
    	]
    },
    	"application/x-shar": {
    	source: "apache",
    	extensions: [
    		"shar"
    	]
    },
    	"application/x-shockwave-flash": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"swf"
    	]
    },
    	"application/x-silverlight-app": {
    	source: "apache",
    	extensions: [
    		"xap"
    	]
    },
    	"application/x-sql": {
    	source: "apache",
    	extensions: [
    		"sql"
    	]
    },
    	"application/x-stuffit": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"sit"
    	]
    },
    	"application/x-stuffitx": {
    	source: "apache",
    	extensions: [
    		"sitx"
    	]
    },
    	"application/x-subrip": {
    	source: "apache",
    	extensions: [
    		"srt"
    	]
    },
    	"application/x-sv4cpio": {
    	source: "apache",
    	extensions: [
    		"sv4cpio"
    	]
    },
    	"application/x-sv4crc": {
    	source: "apache",
    	extensions: [
    		"sv4crc"
    	]
    },
    	"application/x-t3vm-image": {
    	source: "apache",
    	extensions: [
    		"t3"
    	]
    },
    	"application/x-tads": {
    	source: "apache",
    	extensions: [
    		"gam"
    	]
    },
    	"application/x-tar": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"tar"
    	]
    },
    	"application/x-tcl": {
    	source: "apache",
    	extensions: [
    		"tcl",
    		"tk"
    	]
    },
    	"application/x-tex": {
    	source: "apache",
    	extensions: [
    		"tex"
    	]
    },
    	"application/x-tex-tfm": {
    	source: "apache",
    	extensions: [
    		"tfm"
    	]
    },
    	"application/x-texinfo": {
    	source: "apache",
    	extensions: [
    		"texinfo",
    		"texi"
    	]
    },
    	"application/x-tgif": {
    	source: "apache",
    	extensions: [
    		"obj"
    	]
    },
    	"application/x-ustar": {
    	source: "apache",
    	extensions: [
    		"ustar"
    	]
    },
    	"application/x-virtualbox-hdd": {
    	compressible: true,
    	extensions: [
    		"hdd"
    	]
    },
    	"application/x-virtualbox-ova": {
    	compressible: true,
    	extensions: [
    		"ova"
    	]
    },
    	"application/x-virtualbox-ovf": {
    	compressible: true,
    	extensions: [
    		"ovf"
    	]
    },
    	"application/x-virtualbox-vbox": {
    	compressible: true,
    	extensions: [
    		"vbox"
    	]
    },
    	"application/x-virtualbox-vbox-extpack": {
    	compressible: false,
    	extensions: [
    		"vbox-extpack"
    	]
    },
    	"application/x-virtualbox-vdi": {
    	compressible: true,
    	extensions: [
    		"vdi"
    	]
    },
    	"application/x-virtualbox-vhd": {
    	compressible: true,
    	extensions: [
    		"vhd"
    	]
    },
    	"application/x-virtualbox-vmdk": {
    	compressible: true,
    	extensions: [
    		"vmdk"
    	]
    },
    	"application/x-wais-source": {
    	source: "apache",
    	extensions: [
    		"src"
    	]
    },
    	"application/x-web-app-manifest+json": {
    	compressible: true,
    	extensions: [
    		"webapp"
    	]
    },
    	"application/x-www-form-urlencoded": {
    	source: "iana",
    	compressible: true
    },
    	"application/x-x509-ca-cert": {
    	source: "iana",
    	extensions: [
    		"der",
    		"crt",
    		"pem"
    	]
    },
    	"application/x-x509-ca-ra-cert": {
    	source: "iana"
    },
    	"application/x-x509-next-ca-cert": {
    	source: "iana"
    },
    	"application/x-xfig": {
    	source: "apache",
    	extensions: [
    		"fig"
    	]
    },
    	"application/x-xliff+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"xlf"
    	]
    },
    	"application/x-xpinstall": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"xpi"
    	]
    },
    	"application/x-xz": {
    	source: "apache",
    	extensions: [
    		"xz"
    	]
    },
    	"application/x-zmachine": {
    	source: "apache",
    	extensions: [
    		"z1",
    		"z2",
    		"z3",
    		"z4",
    		"z5",
    		"z6",
    		"z7",
    		"z8"
    	]
    },
    	"application/x400-bp": {
    	source: "iana"
    },
    	"application/xacml+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/xaml+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"xaml"
    	]
    },
    	"application/xcap-att+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xav"
    	]
    },
    	"application/xcap-caps+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xca"
    	]
    },
    	"application/xcap-diff+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xdf"
    	]
    },
    	"application/xcap-el+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xel"
    	]
    },
    	"application/xcap-error+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/xcap-ns+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xns"
    	]
    },
    	"application/xcon-conference-info+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/xcon-conference-info-diff+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/xenc+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xenc"
    	]
    },
    	"application/xhtml+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xhtml",
    		"xht"
    	]
    },
    	"application/xhtml-voice+xml": {
    	source: "apache",
    	compressible: true
    },
    	"application/xliff+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xlf"
    	]
    },
    	"application/xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xml",
    		"xsl",
    		"xsd",
    		"rng"
    	]
    },
    	"application/xml-dtd": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"dtd"
    	]
    },
    	"application/xml-external-parsed-entity": {
    	source: "iana"
    },
    	"application/xml-patch+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/xmpp+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/xop+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xop"
    	]
    },
    	"application/xproc+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"xpl"
    	]
    },
    	"application/xslt+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xsl",
    		"xslt"
    	]
    },
    	"application/xspf+xml": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"xspf"
    	]
    },
    	"application/xv+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"mxml",
    		"xhvml",
    		"xvml",
    		"xvm"
    	]
    },
    	"application/yang": {
    	source: "iana",
    	extensions: [
    		"yang"
    	]
    },
    	"application/yang-data+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/yang-data+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/yang-patch+json": {
    	source: "iana",
    	compressible: true
    },
    	"application/yang-patch+xml": {
    	source: "iana",
    	compressible: true
    },
    	"application/yin+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"yin"
    	]
    },
    	"application/zip": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"zip"
    	]
    },
    	"application/zlib": {
    	source: "iana"
    },
    	"application/zstd": {
    	source: "iana"
    },
    	"audio/1d-interleaved-parityfec": {
    	source: "iana"
    },
    	"audio/32kadpcm": {
    	source: "iana"
    },
    	"audio/3gpp": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"3gpp"
    	]
    },
    	"audio/3gpp2": {
    	source: "iana"
    },
    	"audio/aac": {
    	source: "iana"
    },
    	"audio/ac3": {
    	source: "iana"
    },
    	"audio/adpcm": {
    	source: "apache",
    	extensions: [
    		"adp"
    	]
    },
    	"audio/amr": {
    	source: "iana",
    	extensions: [
    		"amr"
    	]
    },
    	"audio/amr-wb": {
    	source: "iana"
    },
    	"audio/amr-wb+": {
    	source: "iana"
    },
    	"audio/aptx": {
    	source: "iana"
    },
    	"audio/asc": {
    	source: "iana"
    },
    	"audio/atrac-advanced-lossless": {
    	source: "iana"
    },
    	"audio/atrac-x": {
    	source: "iana"
    },
    	"audio/atrac3": {
    	source: "iana"
    },
    	"audio/basic": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"au",
    		"snd"
    	]
    },
    	"audio/bv16": {
    	source: "iana"
    },
    	"audio/bv32": {
    	source: "iana"
    },
    	"audio/clearmode": {
    	source: "iana"
    },
    	"audio/cn": {
    	source: "iana"
    },
    	"audio/dat12": {
    	source: "iana"
    },
    	"audio/dls": {
    	source: "iana"
    },
    	"audio/dsr-es201108": {
    	source: "iana"
    },
    	"audio/dsr-es202050": {
    	source: "iana"
    },
    	"audio/dsr-es202211": {
    	source: "iana"
    },
    	"audio/dsr-es202212": {
    	source: "iana"
    },
    	"audio/dv": {
    	source: "iana"
    },
    	"audio/dvi4": {
    	source: "iana"
    },
    	"audio/eac3": {
    	source: "iana"
    },
    	"audio/encaprtp": {
    	source: "iana"
    },
    	"audio/evrc": {
    	source: "iana"
    },
    	"audio/evrc-qcp": {
    	source: "iana"
    },
    	"audio/evrc0": {
    	source: "iana"
    },
    	"audio/evrc1": {
    	source: "iana"
    },
    	"audio/evrcb": {
    	source: "iana"
    },
    	"audio/evrcb0": {
    	source: "iana"
    },
    	"audio/evrcb1": {
    	source: "iana"
    },
    	"audio/evrcnw": {
    	source: "iana"
    },
    	"audio/evrcnw0": {
    	source: "iana"
    },
    	"audio/evrcnw1": {
    	source: "iana"
    },
    	"audio/evrcwb": {
    	source: "iana"
    },
    	"audio/evrcwb0": {
    	source: "iana"
    },
    	"audio/evrcwb1": {
    	source: "iana"
    },
    	"audio/evs": {
    	source: "iana"
    },
    	"audio/flexfec": {
    	source: "iana"
    },
    	"audio/fwdred": {
    	source: "iana"
    },
    	"audio/g711-0": {
    	source: "iana"
    },
    	"audio/g719": {
    	source: "iana"
    },
    	"audio/g722": {
    	source: "iana"
    },
    	"audio/g7221": {
    	source: "iana"
    },
    	"audio/g723": {
    	source: "iana"
    },
    	"audio/g726-16": {
    	source: "iana"
    },
    	"audio/g726-24": {
    	source: "iana"
    },
    	"audio/g726-32": {
    	source: "iana"
    },
    	"audio/g726-40": {
    	source: "iana"
    },
    	"audio/g728": {
    	source: "iana"
    },
    	"audio/g729": {
    	source: "iana"
    },
    	"audio/g7291": {
    	source: "iana"
    },
    	"audio/g729d": {
    	source: "iana"
    },
    	"audio/g729e": {
    	source: "iana"
    },
    	"audio/gsm": {
    	source: "iana"
    },
    	"audio/gsm-efr": {
    	source: "iana"
    },
    	"audio/gsm-hr-08": {
    	source: "iana"
    },
    	"audio/ilbc": {
    	source: "iana"
    },
    	"audio/ip-mr_v2.5": {
    	source: "iana"
    },
    	"audio/isac": {
    	source: "apache"
    },
    	"audio/l16": {
    	source: "iana"
    },
    	"audio/l20": {
    	source: "iana"
    },
    	"audio/l24": {
    	source: "iana",
    	compressible: false
    },
    	"audio/l8": {
    	source: "iana"
    },
    	"audio/lpc": {
    	source: "iana"
    },
    	"audio/melp": {
    	source: "iana"
    },
    	"audio/melp1200": {
    	source: "iana"
    },
    	"audio/melp2400": {
    	source: "iana"
    },
    	"audio/melp600": {
    	source: "iana"
    },
    	"audio/mhas": {
    	source: "iana"
    },
    	"audio/midi": {
    	source: "apache",
    	extensions: [
    		"mid",
    		"midi",
    		"kar",
    		"rmi"
    	]
    },
    	"audio/mobile-xmf": {
    	source: "iana",
    	extensions: [
    		"mxmf"
    	]
    },
    	"audio/mp3": {
    	compressible: false,
    	extensions: [
    		"mp3"
    	]
    },
    	"audio/mp4": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"m4a",
    		"mp4a"
    	]
    },
    	"audio/mp4a-latm": {
    	source: "iana"
    },
    	"audio/mpa": {
    	source: "iana"
    },
    	"audio/mpa-robust": {
    	source: "iana"
    },
    	"audio/mpeg": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"mpga",
    		"mp2",
    		"mp2a",
    		"mp3",
    		"m2a",
    		"m3a"
    	]
    },
    	"audio/mpeg4-generic": {
    	source: "iana"
    },
    	"audio/musepack": {
    	source: "apache"
    },
    	"audio/ogg": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"oga",
    		"ogg",
    		"spx",
    		"opus"
    	]
    },
    	"audio/opus": {
    	source: "iana"
    },
    	"audio/parityfec": {
    	source: "iana"
    },
    	"audio/pcma": {
    	source: "iana"
    },
    	"audio/pcma-wb": {
    	source: "iana"
    },
    	"audio/pcmu": {
    	source: "iana"
    },
    	"audio/pcmu-wb": {
    	source: "iana"
    },
    	"audio/prs.sid": {
    	source: "iana"
    },
    	"audio/qcelp": {
    	source: "iana"
    },
    	"audio/raptorfec": {
    	source: "iana"
    },
    	"audio/red": {
    	source: "iana"
    },
    	"audio/rtp-enc-aescm128": {
    	source: "iana"
    },
    	"audio/rtp-midi": {
    	source: "iana"
    },
    	"audio/rtploopback": {
    	source: "iana"
    },
    	"audio/rtx": {
    	source: "iana"
    },
    	"audio/s3m": {
    	source: "apache",
    	extensions: [
    		"s3m"
    	]
    },
    	"audio/scip": {
    	source: "iana"
    },
    	"audio/silk": {
    	source: "apache",
    	extensions: [
    		"sil"
    	]
    },
    	"audio/smv": {
    	source: "iana"
    },
    	"audio/smv-qcp": {
    	source: "iana"
    },
    	"audio/smv0": {
    	source: "iana"
    },
    	"audio/sofa": {
    	source: "iana"
    },
    	"audio/sp-midi": {
    	source: "iana"
    },
    	"audio/speex": {
    	source: "iana"
    },
    	"audio/t140c": {
    	source: "iana"
    },
    	"audio/t38": {
    	source: "iana"
    },
    	"audio/telephone-event": {
    	source: "iana"
    },
    	"audio/tetra_acelp": {
    	source: "iana"
    },
    	"audio/tetra_acelp_bb": {
    	source: "iana"
    },
    	"audio/tone": {
    	source: "iana"
    },
    	"audio/tsvcis": {
    	source: "iana"
    },
    	"audio/uemclip": {
    	source: "iana"
    },
    	"audio/ulpfec": {
    	source: "iana"
    },
    	"audio/usac": {
    	source: "iana"
    },
    	"audio/vdvi": {
    	source: "iana"
    },
    	"audio/vmr-wb": {
    	source: "iana"
    },
    	"audio/vnd.3gpp.iufp": {
    	source: "iana"
    },
    	"audio/vnd.4sb": {
    	source: "iana"
    },
    	"audio/vnd.audiokoz": {
    	source: "iana"
    },
    	"audio/vnd.celp": {
    	source: "iana"
    },
    	"audio/vnd.cisco.nse": {
    	source: "iana"
    },
    	"audio/vnd.cmles.radio-events": {
    	source: "iana"
    },
    	"audio/vnd.cns.anp1": {
    	source: "iana"
    },
    	"audio/vnd.cns.inf1": {
    	source: "iana"
    },
    	"audio/vnd.dece.audio": {
    	source: "iana",
    	extensions: [
    		"uva",
    		"uvva"
    	]
    },
    	"audio/vnd.digital-winds": {
    	source: "iana",
    	extensions: [
    		"eol"
    	]
    },
    	"audio/vnd.dlna.adts": {
    	source: "iana"
    },
    	"audio/vnd.dolby.heaac.1": {
    	source: "iana"
    },
    	"audio/vnd.dolby.heaac.2": {
    	source: "iana"
    },
    	"audio/vnd.dolby.mlp": {
    	source: "iana"
    },
    	"audio/vnd.dolby.mps": {
    	source: "iana"
    },
    	"audio/vnd.dolby.pl2": {
    	source: "iana"
    },
    	"audio/vnd.dolby.pl2x": {
    	source: "iana"
    },
    	"audio/vnd.dolby.pl2z": {
    	source: "iana"
    },
    	"audio/vnd.dolby.pulse.1": {
    	source: "iana"
    },
    	"audio/vnd.dra": {
    	source: "iana",
    	extensions: [
    		"dra"
    	]
    },
    	"audio/vnd.dts": {
    	source: "iana",
    	extensions: [
    		"dts"
    	]
    },
    	"audio/vnd.dts.hd": {
    	source: "iana",
    	extensions: [
    		"dtshd"
    	]
    },
    	"audio/vnd.dts.uhd": {
    	source: "iana"
    },
    	"audio/vnd.dvb.file": {
    	source: "iana"
    },
    	"audio/vnd.everad.plj": {
    	source: "iana"
    },
    	"audio/vnd.hns.audio": {
    	source: "iana"
    },
    	"audio/vnd.lucent.voice": {
    	source: "iana",
    	extensions: [
    		"lvp"
    	]
    },
    	"audio/vnd.ms-playready.media.pya": {
    	source: "iana",
    	extensions: [
    		"pya"
    	]
    },
    	"audio/vnd.nokia.mobile-xmf": {
    	source: "iana"
    },
    	"audio/vnd.nortel.vbk": {
    	source: "iana"
    },
    	"audio/vnd.nuera.ecelp4800": {
    	source: "iana",
    	extensions: [
    		"ecelp4800"
    	]
    },
    	"audio/vnd.nuera.ecelp7470": {
    	source: "iana",
    	extensions: [
    		"ecelp7470"
    	]
    },
    	"audio/vnd.nuera.ecelp9600": {
    	source: "iana",
    	extensions: [
    		"ecelp9600"
    	]
    },
    	"audio/vnd.octel.sbc": {
    	source: "iana"
    },
    	"audio/vnd.presonus.multitrack": {
    	source: "iana"
    },
    	"audio/vnd.qcelp": {
    	source: "iana"
    },
    	"audio/vnd.rhetorex.32kadpcm": {
    	source: "iana"
    },
    	"audio/vnd.rip": {
    	source: "iana",
    	extensions: [
    		"rip"
    	]
    },
    	"audio/vnd.rn-realaudio": {
    	compressible: false
    },
    	"audio/vnd.sealedmedia.softseal.mpeg": {
    	source: "iana"
    },
    	"audio/vnd.vmx.cvsd": {
    	source: "iana"
    },
    	"audio/vnd.wave": {
    	compressible: false
    },
    	"audio/vorbis": {
    	source: "iana",
    	compressible: false
    },
    	"audio/vorbis-config": {
    	source: "iana"
    },
    	"audio/wav": {
    	compressible: false,
    	extensions: [
    		"wav"
    	]
    },
    	"audio/wave": {
    	compressible: false,
    	extensions: [
    		"wav"
    	]
    },
    	"audio/webm": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"weba"
    	]
    },
    	"audio/x-aac": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"aac"
    	]
    },
    	"audio/x-aiff": {
    	source: "apache",
    	extensions: [
    		"aif",
    		"aiff",
    		"aifc"
    	]
    },
    	"audio/x-caf": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"caf"
    	]
    },
    	"audio/x-flac": {
    	source: "apache",
    	extensions: [
    		"flac"
    	]
    },
    	"audio/x-m4a": {
    	source: "nginx",
    	extensions: [
    		"m4a"
    	]
    },
    	"audio/x-matroska": {
    	source: "apache",
    	extensions: [
    		"mka"
    	]
    },
    	"audio/x-mpegurl": {
    	source: "apache",
    	extensions: [
    		"m3u"
    	]
    },
    	"audio/x-ms-wax": {
    	source: "apache",
    	extensions: [
    		"wax"
    	]
    },
    	"audio/x-ms-wma": {
    	source: "apache",
    	extensions: [
    		"wma"
    	]
    },
    	"audio/x-pn-realaudio": {
    	source: "apache",
    	extensions: [
    		"ram",
    		"ra"
    	]
    },
    	"audio/x-pn-realaudio-plugin": {
    	source: "apache",
    	extensions: [
    		"rmp"
    	]
    },
    	"audio/x-realaudio": {
    	source: "nginx",
    	extensions: [
    		"ra"
    	]
    },
    	"audio/x-tta": {
    	source: "apache"
    },
    	"audio/x-wav": {
    	source: "apache",
    	extensions: [
    		"wav"
    	]
    },
    	"audio/xm": {
    	source: "apache",
    	extensions: [
    		"xm"
    	]
    },
    	"chemical/x-cdx": {
    	source: "apache",
    	extensions: [
    		"cdx"
    	]
    },
    	"chemical/x-cif": {
    	source: "apache",
    	extensions: [
    		"cif"
    	]
    },
    	"chemical/x-cmdf": {
    	source: "apache",
    	extensions: [
    		"cmdf"
    	]
    },
    	"chemical/x-cml": {
    	source: "apache",
    	extensions: [
    		"cml"
    	]
    },
    	"chemical/x-csml": {
    	source: "apache",
    	extensions: [
    		"csml"
    	]
    },
    	"chemical/x-pdb": {
    	source: "apache"
    },
    	"chemical/x-xyz": {
    	source: "apache",
    	extensions: [
    		"xyz"
    	]
    },
    	"font/collection": {
    	source: "iana",
    	extensions: [
    		"ttc"
    	]
    },
    	"font/otf": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"otf"
    	]
    },
    	"font/sfnt": {
    	source: "iana"
    },
    	"font/ttf": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"ttf"
    	]
    },
    	"font/woff": {
    	source: "iana",
    	extensions: [
    		"woff"
    	]
    },
    	"font/woff2": {
    	source: "iana",
    	extensions: [
    		"woff2"
    	]
    },
    	"image/aces": {
    	source: "iana",
    	extensions: [
    		"exr"
    	]
    },
    	"image/apng": {
    	compressible: false,
    	extensions: [
    		"apng"
    	]
    },
    	"image/avci": {
    	source: "iana",
    	extensions: [
    		"avci"
    	]
    },
    	"image/avcs": {
    	source: "iana",
    	extensions: [
    		"avcs"
    	]
    },
    	"image/avif": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"avif"
    	]
    },
    	"image/bmp": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"bmp"
    	]
    },
    	"image/cgm": {
    	source: "iana",
    	extensions: [
    		"cgm"
    	]
    },
    	"image/dicom-rle": {
    	source: "iana",
    	extensions: [
    		"drle"
    	]
    },
    	"image/emf": {
    	source: "iana",
    	extensions: [
    		"emf"
    	]
    },
    	"image/fits": {
    	source: "iana",
    	extensions: [
    		"fits"
    	]
    },
    	"image/g3fax": {
    	source: "iana",
    	extensions: [
    		"g3"
    	]
    },
    	"image/gif": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"gif"
    	]
    },
    	"image/heic": {
    	source: "iana",
    	extensions: [
    		"heic"
    	]
    },
    	"image/heic-sequence": {
    	source: "iana",
    	extensions: [
    		"heics"
    	]
    },
    	"image/heif": {
    	source: "iana",
    	extensions: [
    		"heif"
    	]
    },
    	"image/heif-sequence": {
    	source: "iana",
    	extensions: [
    		"heifs"
    	]
    },
    	"image/hej2k": {
    	source: "iana",
    	extensions: [
    		"hej2"
    	]
    },
    	"image/hsj2": {
    	source: "iana",
    	extensions: [
    		"hsj2"
    	]
    },
    	"image/ief": {
    	source: "iana",
    	extensions: [
    		"ief"
    	]
    },
    	"image/jls": {
    	source: "iana",
    	extensions: [
    		"jls"
    	]
    },
    	"image/jp2": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"jp2",
    		"jpg2"
    	]
    },
    	"image/jpeg": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"jpeg",
    		"jpg",
    		"jpe"
    	]
    },
    	"image/jph": {
    	source: "iana",
    	extensions: [
    		"jph"
    	]
    },
    	"image/jphc": {
    	source: "iana",
    	extensions: [
    		"jhc"
    	]
    },
    	"image/jpm": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"jpm"
    	]
    },
    	"image/jpx": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"jpx",
    		"jpf"
    	]
    },
    	"image/jxr": {
    	source: "iana",
    	extensions: [
    		"jxr"
    	]
    },
    	"image/jxra": {
    	source: "iana",
    	extensions: [
    		"jxra"
    	]
    },
    	"image/jxrs": {
    	source: "iana",
    	extensions: [
    		"jxrs"
    	]
    },
    	"image/jxs": {
    	source: "iana",
    	extensions: [
    		"jxs"
    	]
    },
    	"image/jxsc": {
    	source: "iana",
    	extensions: [
    		"jxsc"
    	]
    },
    	"image/jxsi": {
    	source: "iana",
    	extensions: [
    		"jxsi"
    	]
    },
    	"image/jxss": {
    	source: "iana",
    	extensions: [
    		"jxss"
    	]
    },
    	"image/ktx": {
    	source: "iana",
    	extensions: [
    		"ktx"
    	]
    },
    	"image/ktx2": {
    	source: "iana",
    	extensions: [
    		"ktx2"
    	]
    },
    	"image/naplps": {
    	source: "iana"
    },
    	"image/pjpeg": {
    	compressible: false
    },
    	"image/png": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"png"
    	]
    },
    	"image/prs.btif": {
    	source: "iana",
    	extensions: [
    		"btif"
    	]
    },
    	"image/prs.pti": {
    	source: "iana",
    	extensions: [
    		"pti"
    	]
    },
    	"image/pwg-raster": {
    	source: "iana"
    },
    	"image/sgi": {
    	source: "apache",
    	extensions: [
    		"sgi"
    	]
    },
    	"image/svg+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"svg",
    		"svgz"
    	]
    },
    	"image/t38": {
    	source: "iana",
    	extensions: [
    		"t38"
    	]
    },
    	"image/tiff": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"tif",
    		"tiff"
    	]
    },
    	"image/tiff-fx": {
    	source: "iana",
    	extensions: [
    		"tfx"
    	]
    },
    	"image/vnd.adobe.photoshop": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"psd"
    	]
    },
    	"image/vnd.airzip.accelerator.azv": {
    	source: "iana",
    	extensions: [
    		"azv"
    	]
    },
    	"image/vnd.cns.inf2": {
    	source: "iana"
    },
    	"image/vnd.dece.graphic": {
    	source: "iana",
    	extensions: [
    		"uvi",
    		"uvvi",
    		"uvg",
    		"uvvg"
    	]
    },
    	"image/vnd.djvu": {
    	source: "iana",
    	extensions: [
    		"djvu",
    		"djv"
    	]
    },
    	"image/vnd.dvb.subtitle": {
    	source: "iana",
    	extensions: [
    		"sub"
    	]
    },
    	"image/vnd.dwg": {
    	source: "iana",
    	extensions: [
    		"dwg"
    	]
    },
    	"image/vnd.dxf": {
    	source: "iana",
    	extensions: [
    		"dxf"
    	]
    },
    	"image/vnd.fastbidsheet": {
    	source: "iana",
    	extensions: [
    		"fbs"
    	]
    },
    	"image/vnd.fpx": {
    	source: "iana",
    	extensions: [
    		"fpx"
    	]
    },
    	"image/vnd.fst": {
    	source: "iana",
    	extensions: [
    		"fst"
    	]
    },
    	"image/vnd.fujixerox.edmics-mmr": {
    	source: "iana",
    	extensions: [
    		"mmr"
    	]
    },
    	"image/vnd.fujixerox.edmics-rlc": {
    	source: "iana",
    	extensions: [
    		"rlc"
    	]
    },
    	"image/vnd.globalgraphics.pgb": {
    	source: "iana"
    },
    	"image/vnd.microsoft.icon": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"ico"
    	]
    },
    	"image/vnd.mix": {
    	source: "iana"
    },
    	"image/vnd.mozilla.apng": {
    	source: "iana"
    },
    	"image/vnd.ms-dds": {
    	compressible: true,
    	extensions: [
    		"dds"
    	]
    },
    	"image/vnd.ms-modi": {
    	source: "iana",
    	extensions: [
    		"mdi"
    	]
    },
    	"image/vnd.ms-photo": {
    	source: "apache",
    	extensions: [
    		"wdp"
    	]
    },
    	"image/vnd.net-fpx": {
    	source: "iana",
    	extensions: [
    		"npx"
    	]
    },
    	"image/vnd.pco.b16": {
    	source: "iana",
    	extensions: [
    		"b16"
    	]
    },
    	"image/vnd.radiance": {
    	source: "iana"
    },
    	"image/vnd.sealed.png": {
    	source: "iana"
    },
    	"image/vnd.sealedmedia.softseal.gif": {
    	source: "iana"
    },
    	"image/vnd.sealedmedia.softseal.jpg": {
    	source: "iana"
    },
    	"image/vnd.svf": {
    	source: "iana"
    },
    	"image/vnd.tencent.tap": {
    	source: "iana",
    	extensions: [
    		"tap"
    	]
    },
    	"image/vnd.valve.source.texture": {
    	source: "iana",
    	extensions: [
    		"vtf"
    	]
    },
    	"image/vnd.wap.wbmp": {
    	source: "iana",
    	extensions: [
    		"wbmp"
    	]
    },
    	"image/vnd.xiff": {
    	source: "iana",
    	extensions: [
    		"xif"
    	]
    },
    	"image/vnd.zbrush.pcx": {
    	source: "iana",
    	extensions: [
    		"pcx"
    	]
    },
    	"image/webp": {
    	source: "apache",
    	extensions: [
    		"webp"
    	]
    },
    	"image/wmf": {
    	source: "iana",
    	extensions: [
    		"wmf"
    	]
    },
    	"image/x-3ds": {
    	source: "apache",
    	extensions: [
    		"3ds"
    	]
    },
    	"image/x-cmu-raster": {
    	source: "apache",
    	extensions: [
    		"ras"
    	]
    },
    	"image/x-cmx": {
    	source: "apache",
    	extensions: [
    		"cmx"
    	]
    },
    	"image/x-freehand": {
    	source: "apache",
    	extensions: [
    		"fh",
    		"fhc",
    		"fh4",
    		"fh5",
    		"fh7"
    	]
    },
    	"image/x-icon": {
    	source: "apache",
    	compressible: true,
    	extensions: [
    		"ico"
    	]
    },
    	"image/x-jng": {
    	source: "nginx",
    	extensions: [
    		"jng"
    	]
    },
    	"image/x-mrsid-image": {
    	source: "apache",
    	extensions: [
    		"sid"
    	]
    },
    	"image/x-ms-bmp": {
    	source: "nginx",
    	compressible: true,
    	extensions: [
    		"bmp"
    	]
    },
    	"image/x-pcx": {
    	source: "apache",
    	extensions: [
    		"pcx"
    	]
    },
    	"image/x-pict": {
    	source: "apache",
    	extensions: [
    		"pic",
    		"pct"
    	]
    },
    	"image/x-portable-anymap": {
    	source: "apache",
    	extensions: [
    		"pnm"
    	]
    },
    	"image/x-portable-bitmap": {
    	source: "apache",
    	extensions: [
    		"pbm"
    	]
    },
    	"image/x-portable-graymap": {
    	source: "apache",
    	extensions: [
    		"pgm"
    	]
    },
    	"image/x-portable-pixmap": {
    	source: "apache",
    	extensions: [
    		"ppm"
    	]
    },
    	"image/x-rgb": {
    	source: "apache",
    	extensions: [
    		"rgb"
    	]
    },
    	"image/x-tga": {
    	source: "apache",
    	extensions: [
    		"tga"
    	]
    },
    	"image/x-xbitmap": {
    	source: "apache",
    	extensions: [
    		"xbm"
    	]
    },
    	"image/x-xcf": {
    	compressible: false
    },
    	"image/x-xpixmap": {
    	source: "apache",
    	extensions: [
    		"xpm"
    	]
    },
    	"image/x-xwindowdump": {
    	source: "apache",
    	extensions: [
    		"xwd"
    	]
    },
    	"message/cpim": {
    	source: "iana"
    },
    	"message/delivery-status": {
    	source: "iana"
    },
    	"message/disposition-notification": {
    	source: "iana",
    	extensions: [
    		"disposition-notification"
    	]
    },
    	"message/external-body": {
    	source: "iana"
    },
    	"message/feedback-report": {
    	source: "iana"
    },
    	"message/global": {
    	source: "iana",
    	extensions: [
    		"u8msg"
    	]
    },
    	"message/global-delivery-status": {
    	source: "iana",
    	extensions: [
    		"u8dsn"
    	]
    },
    	"message/global-disposition-notification": {
    	source: "iana",
    	extensions: [
    		"u8mdn"
    	]
    },
    	"message/global-headers": {
    	source: "iana",
    	extensions: [
    		"u8hdr"
    	]
    },
    	"message/http": {
    	source: "iana",
    	compressible: false
    },
    	"message/imdn+xml": {
    	source: "iana",
    	compressible: true
    },
    	"message/news": {
    	source: "iana"
    },
    	"message/partial": {
    	source: "iana",
    	compressible: false
    },
    	"message/rfc822": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"eml",
    		"mime"
    	]
    },
    	"message/s-http": {
    	source: "iana"
    },
    	"message/sip": {
    	source: "iana"
    },
    	"message/sipfrag": {
    	source: "iana"
    },
    	"message/tracking-status": {
    	source: "iana"
    },
    	"message/vnd.si.simp": {
    	source: "iana"
    },
    	"message/vnd.wfa.wsc": {
    	source: "iana",
    	extensions: [
    		"wsc"
    	]
    },
    	"model/3mf": {
    	source: "iana",
    	extensions: [
    		"3mf"
    	]
    },
    	"model/e57": {
    	source: "iana"
    },
    	"model/gltf+json": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"gltf"
    	]
    },
    	"model/gltf-binary": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"glb"
    	]
    },
    	"model/iges": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"igs",
    		"iges"
    	]
    },
    	"model/mesh": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"msh",
    		"mesh",
    		"silo"
    	]
    },
    	"model/mtl": {
    	source: "iana",
    	extensions: [
    		"mtl"
    	]
    },
    	"model/obj": {
    	source: "iana",
    	extensions: [
    		"obj"
    	]
    },
    	"model/step": {
    	source: "iana"
    },
    	"model/step+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"stpx"
    	]
    },
    	"model/step+zip": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"stpz"
    	]
    },
    	"model/step-xml+zip": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"stpxz"
    	]
    },
    	"model/stl": {
    	source: "iana",
    	extensions: [
    		"stl"
    	]
    },
    	"model/vnd.collada+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"dae"
    	]
    },
    	"model/vnd.dwf": {
    	source: "iana",
    	extensions: [
    		"dwf"
    	]
    },
    	"model/vnd.flatland.3dml": {
    	source: "iana"
    },
    	"model/vnd.gdl": {
    	source: "iana",
    	extensions: [
    		"gdl"
    	]
    },
    	"model/vnd.gs-gdl": {
    	source: "apache"
    },
    	"model/vnd.gs.gdl": {
    	source: "iana"
    },
    	"model/vnd.gtw": {
    	source: "iana",
    	extensions: [
    		"gtw"
    	]
    },
    	"model/vnd.moml+xml": {
    	source: "iana",
    	compressible: true
    },
    	"model/vnd.mts": {
    	source: "iana",
    	extensions: [
    		"mts"
    	]
    },
    	"model/vnd.opengex": {
    	source: "iana",
    	extensions: [
    		"ogex"
    	]
    },
    	"model/vnd.parasolid.transmit.binary": {
    	source: "iana",
    	extensions: [
    		"x_b"
    	]
    },
    	"model/vnd.parasolid.transmit.text": {
    	source: "iana",
    	extensions: [
    		"x_t"
    	]
    },
    	"model/vnd.pytha.pyox": {
    	source: "iana"
    },
    	"model/vnd.rosette.annotated-data-model": {
    	source: "iana"
    },
    	"model/vnd.sap.vds": {
    	source: "iana",
    	extensions: [
    		"vds"
    	]
    },
    	"model/vnd.usdz+zip": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"usdz"
    	]
    },
    	"model/vnd.valve.source.compiled-map": {
    	source: "iana",
    	extensions: [
    		"bsp"
    	]
    },
    	"model/vnd.vtu": {
    	source: "iana",
    	extensions: [
    		"vtu"
    	]
    },
    	"model/vrml": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"wrl",
    		"vrml"
    	]
    },
    	"model/x3d+binary": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"x3db",
    		"x3dbz"
    	]
    },
    	"model/x3d+fastinfoset": {
    	source: "iana",
    	extensions: [
    		"x3db"
    	]
    },
    	"model/x3d+vrml": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"x3dv",
    		"x3dvz"
    	]
    },
    	"model/x3d+xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"x3d",
    		"x3dz"
    	]
    },
    	"model/x3d-vrml": {
    	source: "iana",
    	extensions: [
    		"x3dv"
    	]
    },
    	"multipart/alternative": {
    	source: "iana",
    	compressible: false
    },
    	"multipart/appledouble": {
    	source: "iana"
    },
    	"multipart/byteranges": {
    	source: "iana"
    },
    	"multipart/digest": {
    	source: "iana"
    },
    	"multipart/encrypted": {
    	source: "iana",
    	compressible: false
    },
    	"multipart/form-data": {
    	source: "iana",
    	compressible: false
    },
    	"multipart/header-set": {
    	source: "iana"
    },
    	"multipart/mixed": {
    	source: "iana"
    },
    	"multipart/multilingual": {
    	source: "iana"
    },
    	"multipart/parallel": {
    	source: "iana"
    },
    	"multipart/related": {
    	source: "iana",
    	compressible: false
    },
    	"multipart/report": {
    	source: "iana"
    },
    	"multipart/signed": {
    	source: "iana",
    	compressible: false
    },
    	"multipart/vnd.bint.med-plus": {
    	source: "iana"
    },
    	"multipart/voice-message": {
    	source: "iana"
    },
    	"multipart/x-mixed-replace": {
    	source: "iana"
    },
    	"text/1d-interleaved-parityfec": {
    	source: "iana"
    },
    	"text/cache-manifest": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"appcache",
    		"manifest"
    	]
    },
    	"text/calendar": {
    	source: "iana",
    	extensions: [
    		"ics",
    		"ifb"
    	]
    },
    	"text/calender": {
    	compressible: true
    },
    	"text/cmd": {
    	compressible: true
    },
    	"text/coffeescript": {
    	extensions: [
    		"coffee",
    		"litcoffee"
    	]
    },
    	"text/cql": {
    	source: "iana"
    },
    	"text/cql-expression": {
    	source: "iana"
    },
    	"text/cql-identifier": {
    	source: "iana"
    },
    	"text/css": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true,
    	extensions: [
    		"css"
    	]
    },
    	"text/csv": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"csv"
    	]
    },
    	"text/csv-schema": {
    	source: "iana"
    },
    	"text/directory": {
    	source: "iana"
    },
    	"text/dns": {
    	source: "iana"
    },
    	"text/ecmascript": {
    	source: "iana"
    },
    	"text/encaprtp": {
    	source: "iana"
    },
    	"text/enriched": {
    	source: "iana"
    },
    	"text/fhirpath": {
    	source: "iana"
    },
    	"text/flexfec": {
    	source: "iana"
    },
    	"text/fwdred": {
    	source: "iana"
    },
    	"text/gff3": {
    	source: "iana"
    },
    	"text/grammar-ref-list": {
    	source: "iana"
    },
    	"text/html": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"html",
    		"htm",
    		"shtml"
    	]
    },
    	"text/jade": {
    	extensions: [
    		"jade"
    	]
    },
    	"text/javascript": {
    	source: "iana",
    	compressible: true
    },
    	"text/jcr-cnd": {
    	source: "iana"
    },
    	"text/jsx": {
    	compressible: true,
    	extensions: [
    		"jsx"
    	]
    },
    	"text/less": {
    	compressible: true,
    	extensions: [
    		"less"
    	]
    },
    	"text/markdown": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"markdown",
    		"md"
    	]
    },
    	"text/mathml": {
    	source: "nginx",
    	extensions: [
    		"mml"
    	]
    },
    	"text/mdx": {
    	compressible: true,
    	extensions: [
    		"mdx"
    	]
    },
    	"text/mizar": {
    	source: "iana"
    },
    	"text/n3": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true,
    	extensions: [
    		"n3"
    	]
    },
    	"text/parameters": {
    	source: "iana",
    	charset: "UTF-8"
    },
    	"text/parityfec": {
    	source: "iana"
    },
    	"text/plain": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"txt",
    		"text",
    		"conf",
    		"def",
    		"list",
    		"log",
    		"in",
    		"ini"
    	]
    },
    	"text/provenance-notation": {
    	source: "iana",
    	charset: "UTF-8"
    },
    	"text/prs.fallenstein.rst": {
    	source: "iana"
    },
    	"text/prs.lines.tag": {
    	source: "iana",
    	extensions: [
    		"dsc"
    	]
    },
    	"text/prs.prop.logic": {
    	source: "iana"
    },
    	"text/raptorfec": {
    	source: "iana"
    },
    	"text/red": {
    	source: "iana"
    },
    	"text/rfc822-headers": {
    	source: "iana"
    },
    	"text/richtext": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rtx"
    	]
    },
    	"text/rtf": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"rtf"
    	]
    },
    	"text/rtp-enc-aescm128": {
    	source: "iana"
    },
    	"text/rtploopback": {
    	source: "iana"
    },
    	"text/rtx": {
    	source: "iana"
    },
    	"text/sgml": {
    	source: "iana",
    	extensions: [
    		"sgml",
    		"sgm"
    	]
    },
    	"text/shaclc": {
    	source: "iana"
    },
    	"text/shex": {
    	source: "iana",
    	extensions: [
    		"shex"
    	]
    },
    	"text/slim": {
    	extensions: [
    		"slim",
    		"slm"
    	]
    },
    	"text/spdx": {
    	source: "iana",
    	extensions: [
    		"spdx"
    	]
    },
    	"text/strings": {
    	source: "iana"
    },
    	"text/stylus": {
    	extensions: [
    		"stylus",
    		"styl"
    	]
    },
    	"text/t140": {
    	source: "iana"
    },
    	"text/tab-separated-values": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"tsv"
    	]
    },
    	"text/troff": {
    	source: "iana",
    	extensions: [
    		"t",
    		"tr",
    		"roff",
    		"man",
    		"me",
    		"ms"
    	]
    },
    	"text/turtle": {
    	source: "iana",
    	charset: "UTF-8",
    	extensions: [
    		"ttl"
    	]
    },
    	"text/ulpfec": {
    	source: "iana"
    },
    	"text/uri-list": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"uri",
    		"uris",
    		"urls"
    	]
    },
    	"text/vcard": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"vcard"
    	]
    },
    	"text/vnd.a": {
    	source: "iana"
    },
    	"text/vnd.abc": {
    	source: "iana"
    },
    	"text/vnd.ascii-art": {
    	source: "iana"
    },
    	"text/vnd.curl": {
    	source: "iana",
    	extensions: [
    		"curl"
    	]
    },
    	"text/vnd.curl.dcurl": {
    	source: "apache",
    	extensions: [
    		"dcurl"
    	]
    },
    	"text/vnd.curl.mcurl": {
    	source: "apache",
    	extensions: [
    		"mcurl"
    	]
    },
    	"text/vnd.curl.scurl": {
    	source: "apache",
    	extensions: [
    		"scurl"
    	]
    },
    	"text/vnd.debian.copyright": {
    	source: "iana",
    	charset: "UTF-8"
    },
    	"text/vnd.dmclientscript": {
    	source: "iana"
    },
    	"text/vnd.dvb.subtitle": {
    	source: "iana",
    	extensions: [
    		"sub"
    	]
    },
    	"text/vnd.esmertec.theme-descriptor": {
    	source: "iana",
    	charset: "UTF-8"
    },
    	"text/vnd.familysearch.gedcom": {
    	source: "iana",
    	extensions: [
    		"ged"
    	]
    },
    	"text/vnd.ficlab.flt": {
    	source: "iana"
    },
    	"text/vnd.fly": {
    	source: "iana",
    	extensions: [
    		"fly"
    	]
    },
    	"text/vnd.fmi.flexstor": {
    	source: "iana",
    	extensions: [
    		"flx"
    	]
    },
    	"text/vnd.gml": {
    	source: "iana"
    },
    	"text/vnd.graphviz": {
    	source: "iana",
    	extensions: [
    		"gv"
    	]
    },
    	"text/vnd.hans": {
    	source: "iana"
    },
    	"text/vnd.hgl": {
    	source: "iana"
    },
    	"text/vnd.in3d.3dml": {
    	source: "iana",
    	extensions: [
    		"3dml"
    	]
    },
    	"text/vnd.in3d.spot": {
    	source: "iana",
    	extensions: [
    		"spot"
    	]
    },
    	"text/vnd.iptc.newsml": {
    	source: "iana"
    },
    	"text/vnd.iptc.nitf": {
    	source: "iana"
    },
    	"text/vnd.latex-z": {
    	source: "iana"
    },
    	"text/vnd.motorola.reflex": {
    	source: "iana"
    },
    	"text/vnd.ms-mediapackage": {
    	source: "iana"
    },
    	"text/vnd.net2phone.commcenter.command": {
    	source: "iana"
    },
    	"text/vnd.radisys.msml-basic-layout": {
    	source: "iana"
    },
    	"text/vnd.senx.warpscript": {
    	source: "iana"
    },
    	"text/vnd.si.uricatalogue": {
    	source: "iana"
    },
    	"text/vnd.sosi": {
    	source: "iana"
    },
    	"text/vnd.sun.j2me.app-descriptor": {
    	source: "iana",
    	charset: "UTF-8",
    	extensions: [
    		"jad"
    	]
    },
    	"text/vnd.trolltech.linguist": {
    	source: "iana",
    	charset: "UTF-8"
    },
    	"text/vnd.wap.si": {
    	source: "iana"
    },
    	"text/vnd.wap.sl": {
    	source: "iana"
    },
    	"text/vnd.wap.wml": {
    	source: "iana",
    	extensions: [
    		"wml"
    	]
    },
    	"text/vnd.wap.wmlscript": {
    	source: "iana",
    	extensions: [
    		"wmls"
    	]
    },
    	"text/vtt": {
    	source: "iana",
    	charset: "UTF-8",
    	compressible: true,
    	extensions: [
    		"vtt"
    	]
    },
    	"text/x-asm": {
    	source: "apache",
    	extensions: [
    		"s",
    		"asm"
    	]
    },
    	"text/x-c": {
    	source: "apache",
    	extensions: [
    		"c",
    		"cc",
    		"cxx",
    		"cpp",
    		"h",
    		"hh",
    		"dic"
    	]
    },
    	"text/x-component": {
    	source: "nginx",
    	extensions: [
    		"htc"
    	]
    },
    	"text/x-fortran": {
    	source: "apache",
    	extensions: [
    		"f",
    		"for",
    		"f77",
    		"f90"
    	]
    },
    	"text/x-gwt-rpc": {
    	compressible: true
    },
    	"text/x-handlebars-template": {
    	extensions: [
    		"hbs"
    	]
    },
    	"text/x-java-source": {
    	source: "apache",
    	extensions: [
    		"java"
    	]
    },
    	"text/x-jquery-tmpl": {
    	compressible: true
    },
    	"text/x-lua": {
    	extensions: [
    		"lua"
    	]
    },
    	"text/x-markdown": {
    	compressible: true,
    	extensions: [
    		"mkd"
    	]
    },
    	"text/x-nfo": {
    	source: "apache",
    	extensions: [
    		"nfo"
    	]
    },
    	"text/x-opml": {
    	source: "apache",
    	extensions: [
    		"opml"
    	]
    },
    	"text/x-org": {
    	compressible: true,
    	extensions: [
    		"org"
    	]
    },
    	"text/x-pascal": {
    	source: "apache",
    	extensions: [
    		"p",
    		"pas"
    	]
    },
    	"text/x-processing": {
    	compressible: true,
    	extensions: [
    		"pde"
    	]
    },
    	"text/x-sass": {
    	extensions: [
    		"sass"
    	]
    },
    	"text/x-scss": {
    	extensions: [
    		"scss"
    	]
    },
    	"text/x-setext": {
    	source: "apache",
    	extensions: [
    		"etx"
    	]
    },
    	"text/x-sfv": {
    	source: "apache",
    	extensions: [
    		"sfv"
    	]
    },
    	"text/x-suse-ymp": {
    	compressible: true,
    	extensions: [
    		"ymp"
    	]
    },
    	"text/x-uuencode": {
    	source: "apache",
    	extensions: [
    		"uu"
    	]
    },
    	"text/x-vcalendar": {
    	source: "apache",
    	extensions: [
    		"vcs"
    	]
    },
    	"text/x-vcard": {
    	source: "apache",
    	extensions: [
    		"vcf"
    	]
    },
    	"text/xml": {
    	source: "iana",
    	compressible: true,
    	extensions: [
    		"xml"
    	]
    },
    	"text/xml-external-parsed-entity": {
    	source: "iana"
    },
    	"text/yaml": {
    	compressible: true,
    	extensions: [
    		"yaml",
    		"yml"
    	]
    },
    	"video/1d-interleaved-parityfec": {
    	source: "iana"
    },
    	"video/3gpp": {
    	source: "iana",
    	extensions: [
    		"3gp",
    		"3gpp"
    	]
    },
    	"video/3gpp-tt": {
    	source: "iana"
    },
    	"video/3gpp2": {
    	source: "iana",
    	extensions: [
    		"3g2"
    	]
    },
    	"video/av1": {
    	source: "iana"
    },
    	"video/bmpeg": {
    	source: "iana"
    },
    	"video/bt656": {
    	source: "iana"
    },
    	"video/celb": {
    	source: "iana"
    },
    	"video/dv": {
    	source: "iana"
    },
    	"video/encaprtp": {
    	source: "iana"
    },
    	"video/ffv1": {
    	source: "iana"
    },
    	"video/flexfec": {
    	source: "iana"
    },
    	"video/h261": {
    	source: "iana",
    	extensions: [
    		"h261"
    	]
    },
    	"video/h263": {
    	source: "iana",
    	extensions: [
    		"h263"
    	]
    },
    	"video/h263-1998": {
    	source: "iana"
    },
    	"video/h263-2000": {
    	source: "iana"
    },
    	"video/h264": {
    	source: "iana",
    	extensions: [
    		"h264"
    	]
    },
    	"video/h264-rcdo": {
    	source: "iana"
    },
    	"video/h264-svc": {
    	source: "iana"
    },
    	"video/h265": {
    	source: "iana"
    },
    	"video/iso.segment": {
    	source: "iana",
    	extensions: [
    		"m4s"
    	]
    },
    	"video/jpeg": {
    	source: "iana",
    	extensions: [
    		"jpgv"
    	]
    },
    	"video/jpeg2000": {
    	source: "iana"
    },
    	"video/jpm": {
    	source: "apache",
    	extensions: [
    		"jpm",
    		"jpgm"
    	]
    },
    	"video/jxsv": {
    	source: "iana"
    },
    	"video/mj2": {
    	source: "iana",
    	extensions: [
    		"mj2",
    		"mjp2"
    	]
    },
    	"video/mp1s": {
    	source: "iana"
    },
    	"video/mp2p": {
    	source: "iana"
    },
    	"video/mp2t": {
    	source: "iana",
    	extensions: [
    		"ts"
    	]
    },
    	"video/mp4": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"mp4",
    		"mp4v",
    		"mpg4"
    	]
    },
    	"video/mp4v-es": {
    	source: "iana"
    },
    	"video/mpeg": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"mpeg",
    		"mpg",
    		"mpe",
    		"m1v",
    		"m2v"
    	]
    },
    	"video/mpeg4-generic": {
    	source: "iana"
    },
    	"video/mpv": {
    	source: "iana"
    },
    	"video/nv": {
    	source: "iana"
    },
    	"video/ogg": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"ogv"
    	]
    },
    	"video/parityfec": {
    	source: "iana"
    },
    	"video/pointer": {
    	source: "iana"
    },
    	"video/quicktime": {
    	source: "iana",
    	compressible: false,
    	extensions: [
    		"qt",
    		"mov"
    	]
    },
    	"video/raptorfec": {
    	source: "iana"
    },
    	"video/raw": {
    	source: "iana"
    },
    	"video/rtp-enc-aescm128": {
    	source: "iana"
    },
    	"video/rtploopback": {
    	source: "iana"
    },
    	"video/rtx": {
    	source: "iana"
    },
    	"video/scip": {
    	source: "iana"
    },
    	"video/smpte291": {
    	source: "iana"
    },
    	"video/smpte292m": {
    	source: "iana"
    },
    	"video/ulpfec": {
    	source: "iana"
    },
    	"video/vc1": {
    	source: "iana"
    },
    	"video/vc2": {
    	source: "iana"
    },
    	"video/vnd.cctv": {
    	source: "iana"
    },
    	"video/vnd.dece.hd": {
    	source: "iana",
    	extensions: [
    		"uvh",
    		"uvvh"
    	]
    },
    	"video/vnd.dece.mobile": {
    	source: "iana",
    	extensions: [
    		"uvm",
    		"uvvm"
    	]
    },
    	"video/vnd.dece.mp4": {
    	source: "iana"
    },
    	"video/vnd.dece.pd": {
    	source: "iana",
    	extensions: [
    		"uvp",
    		"uvvp"
    	]
    },
    	"video/vnd.dece.sd": {
    	source: "iana",
    	extensions: [
    		"uvs",
    		"uvvs"
    	]
    },
    	"video/vnd.dece.video": {
    	source: "iana",
    	extensions: [
    		"uvv",
    		"uvvv"
    	]
    },
    	"video/vnd.directv.mpeg": {
    	source: "iana"
    },
    	"video/vnd.directv.mpeg-tts": {
    	source: "iana"
    },
    	"video/vnd.dlna.mpeg-tts": {
    	source: "iana"
    },
    	"video/vnd.dvb.file": {
    	source: "iana",
    	extensions: [
    		"dvb"
    	]
    },
    	"video/vnd.fvt": {
    	source: "iana",
    	extensions: [
    		"fvt"
    	]
    },
    	"video/vnd.hns.video": {
    	source: "iana"
    },
    	"video/vnd.iptvforum.1dparityfec-1010": {
    	source: "iana"
    },
    	"video/vnd.iptvforum.1dparityfec-2005": {
    	source: "iana"
    },
    	"video/vnd.iptvforum.2dparityfec-1010": {
    	source: "iana"
    },
    	"video/vnd.iptvforum.2dparityfec-2005": {
    	source: "iana"
    },
    	"video/vnd.iptvforum.ttsavc": {
    	source: "iana"
    },
    	"video/vnd.iptvforum.ttsmpeg2": {
    	source: "iana"
    },
    	"video/vnd.motorola.video": {
    	source: "iana"
    },
    	"video/vnd.motorola.videop": {
    	source: "iana"
    },
    	"video/vnd.mpegurl": {
    	source: "iana",
    	extensions: [
    		"mxu",
    		"m4u"
    	]
    },
    	"video/vnd.ms-playready.media.pyv": {
    	source: "iana",
    	extensions: [
    		"pyv"
    	]
    },
    	"video/vnd.nokia.interleaved-multimedia": {
    	source: "iana"
    },
    	"video/vnd.nokia.mp4vr": {
    	source: "iana"
    },
    	"video/vnd.nokia.videovoip": {
    	source: "iana"
    },
    	"video/vnd.objectvideo": {
    	source: "iana"
    },
    	"video/vnd.radgamettools.bink": {
    	source: "iana"
    },
    	"video/vnd.radgamettools.smacker": {
    	source: "iana"
    },
    	"video/vnd.sealed.mpeg1": {
    	source: "iana"
    },
    	"video/vnd.sealed.mpeg4": {
    	source: "iana"
    },
    	"video/vnd.sealed.swf": {
    	source: "iana"
    },
    	"video/vnd.sealedmedia.softseal.mov": {
    	source: "iana"
    },
    	"video/vnd.uvvu.mp4": {
    	source: "iana",
    	extensions: [
    		"uvu",
    		"uvvu"
    	]
    },
    	"video/vnd.vivo": {
    	source: "iana",
    	extensions: [
    		"viv"
    	]
    },
    	"video/vnd.youtube.yt": {
    	source: "iana"
    },
    	"video/vp8": {
    	source: "iana"
    },
    	"video/vp9": {
    	source: "iana"
    },
    	"video/webm": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"webm"
    	]
    },
    	"video/x-f4v": {
    	source: "apache",
    	extensions: [
    		"f4v"
    	]
    },
    	"video/x-fli": {
    	source: "apache",
    	extensions: [
    		"fli"
    	]
    },
    	"video/x-flv": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"flv"
    	]
    },
    	"video/x-m4v": {
    	source: "apache",
    	extensions: [
    		"m4v"
    	]
    },
    	"video/x-matroska": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"mkv",
    		"mk3d",
    		"mks"
    	]
    },
    	"video/x-mng": {
    	source: "apache",
    	extensions: [
    		"mng"
    	]
    },
    	"video/x-ms-asf": {
    	source: "apache",
    	extensions: [
    		"asf",
    		"asx"
    	]
    },
    	"video/x-ms-vob": {
    	source: "apache",
    	extensions: [
    		"vob"
    	]
    },
    	"video/x-ms-wm": {
    	source: "apache",
    	extensions: [
    		"wm"
    	]
    },
    	"video/x-ms-wmv": {
    	source: "apache",
    	compressible: false,
    	extensions: [
    		"wmv"
    	]
    },
    	"video/x-ms-wmx": {
    	source: "apache",
    	extensions: [
    		"wmx"
    	]
    },
    	"video/x-ms-wvx": {
    	source: "apache",
    	extensions: [
    		"wvx"
    	]
    },
    	"video/x-msvideo": {
    	source: "apache",
    	extensions: [
    		"avi"
    	]
    },
    	"video/x-sgi-movie": {
    	source: "apache",
    	extensions: [
    		"movie"
    	]
    },
    	"video/x-smv": {
    	source: "apache",
    	extensions: [
    		"smv"
    	]
    },
    	"x-conference/x-cooltalk": {
    	source: "apache",
    	extensions: [
    		"ice"
    	]
    },
    	"x-shader/x-fragment": {
    	compressible: true
    },
    	"x-shader/x-vertex": {
    	compressible: true
    }
    };

    /*!
     * mime-db
     * Copyright(c) 2014 Jonathan Ong
     * Copyright(c) 2015-2022 Douglas Christopher Wilson
     * MIT Licensed
     */

    var hasRequiredMimeDb;

    function requireMimeDb () {
    	if (hasRequiredMimeDb) return mimeDb.exports;
    	hasRequiredMimeDb = 1;
    	(function (module) {
    		/**
    		 * Module exports.
    		 */

    		module.exports = require$$0;
    } (mimeDb));
    	return mimeDb.exports;
    }

    /*!
     * mime-types
     * Copyright(c) 2014 Jonathan Ong
     * Copyright(c) 2015 Douglas Christopher Wilson
     * MIT Licensed
     */

    var hasRequiredMimeTypes;

    function requireMimeTypes () {
    	if (hasRequiredMimeTypes) return mimeTypes;
    	hasRequiredMimeTypes = 1;
    	(function (exports) {

    		/**
    		 * Module dependencies.
    		 * @private
    		 */

    		var db = requireMimeDb();
    		var extname = require$$2.extname;

    		/**
    		 * Module variables.
    		 * @private
    		 */

    		var EXTRACT_TYPE_REGEXP = /^\s*([^;\s]*)(?:;|\s|$)/;
    		var TEXT_TYPE_REGEXP = /^text\//i;

    		/**
    		 * Module exports.
    		 * @public
    		 */

    		exports.charset = charset;
    		exports.charsets = { lookup: charset };
    		exports.contentType = contentType;
    		exports.extension = extension;
    		exports.extensions = Object.create(null);
    		exports.lookup = lookup;
    		exports.types = Object.create(null);

    		// Populate the extensions/types maps
    		populateMaps(exports.extensions, exports.types);

    		/**
    		 * Get the default charset for a MIME type.
    		 *
    		 * @param {string} type
    		 * @return {boolean|string}
    		 */

    		function charset (type) {
    		  if (!type || typeof type !== 'string') {
    		    return false
    		  }

    		  // TODO: use media-typer
    		  var match = EXTRACT_TYPE_REGEXP.exec(type);
    		  var mime = match && db[match[1].toLowerCase()];

    		  if (mime && mime.charset) {
    		    return mime.charset
    		  }

    		  // default text/* to utf-8
    		  if (match && TEXT_TYPE_REGEXP.test(match[1])) {
    		    return 'UTF-8'
    		  }

    		  return false
    		}

    		/**
    		 * Create a full Content-Type header given a MIME type or extension.
    		 *
    		 * @param {string} str
    		 * @return {boolean|string}
    		 */

    		function contentType (str) {
    		  // TODO: should this even be in this module?
    		  if (!str || typeof str !== 'string') {
    		    return false
    		  }

    		  var mime = str.indexOf('/') === -1
    		    ? exports.lookup(str)
    		    : str;

    		  if (!mime) {
    		    return false
    		  }

    		  // TODO: use content-type or other module
    		  if (mime.indexOf('charset') === -1) {
    		    var charset = exports.charset(mime);
    		    if (charset) mime += '; charset=' + charset.toLowerCase();
    		  }

    		  return mime
    		}

    		/**
    		 * Get the default extension for a MIME type.
    		 *
    		 * @param {string} type
    		 * @return {boolean|string}
    		 */

    		function extension (type) {
    		  if (!type || typeof type !== 'string') {
    		    return false
    		  }

    		  // TODO: use media-typer
    		  var match = EXTRACT_TYPE_REGEXP.exec(type);

    		  // get extensions
    		  var exts = match && exports.extensions[match[1].toLowerCase()];

    		  if (!exts || !exts.length) {
    		    return false
    		  }

    		  return exts[0]
    		}

    		/**
    		 * Lookup the MIME type for a file path/extension.
    		 *
    		 * @param {string} path
    		 * @return {boolean|string}
    		 */

    		function lookup (path) {
    		  if (!path || typeof path !== 'string') {
    		    return false
    		  }

    		  // get the extension ("ext" or ".ext" or full path)
    		  var extension = extname('x.' + path)
    		    .toLowerCase()
    		    .substr(1);

    		  if (!extension) {
    		    return false
    		  }

    		  return exports.types[extension] || false
    		}

    		/**
    		 * Populate the extensions and types maps.
    		 * @private
    		 */

    		function populateMaps (extensions, types) {
    		  // source preference (least -> most)
    		  var preference = ['nginx', 'apache', undefined, 'iana'];

    		  Object.keys(db).forEach(function forEachMimeType (type) {
    		    var mime = db[type];
    		    var exts = mime.extensions;

    		    if (!exts || !exts.length) {
    		      return
    		    }

    		    // mime -> extensions
    		    extensions[type] = exts;

    		    // extension -> mime
    		    for (var i = 0; i < exts.length; i++) {
    		      var extension = exts[i];

    		      if (types[extension]) {
    		        var from = preference.indexOf(db[types[extension]].source);
    		        var to = preference.indexOf(mime.source);

    		        if (types[extension] !== 'application/octet-stream' &&
    		          (from > to || (from === to && types[extension].substr(0, 12) === 'application/'))) {
    		          // skip the remapping
    		          continue
    		        }
    		      }

    		      // set the extension -> mime
    		      types[extension] = type;
    		    }
    		  });
    		}
    } (mimeTypes));
    	return mimeTypes;
    }

    var defer_1;
    var hasRequiredDefer;

    function requireDefer () {
    	if (hasRequiredDefer) return defer_1;
    	hasRequiredDefer = 1;
    	defer_1 = defer;

    	/**
    	 * Runs provided function on next iteration of the event loop
    	 *
    	 * @param {function} fn - function to run
    	 */
    	function defer(fn)
    	{
    	  var nextTick$1 = typeof setImmediate == 'function'
    	    ? setImmediate
    	    : (
    	      typeof process == 'object' && typeof nextTick == 'function'
    	      ? nextTick
    	      : null
    	    );

    	  if (nextTick$1)
    	  {
    	    nextTick$1(fn);
    	  }
    	  else
    	  {
    	    setTimeout(fn, 0);
    	  }
    	}
    	return defer_1;
    }

    var async_1;
    var hasRequiredAsync;

    function requireAsync () {
    	if (hasRequiredAsync) return async_1;
    	hasRequiredAsync = 1;
    	var defer = requireDefer();

    	// API
    	async_1 = async;

    	/**
    	 * Runs provided callback asynchronously
    	 * even if callback itself is not
    	 *
    	 * @param   {function} callback - callback to invoke
    	 * @returns {function} - augmented callback
    	 */
    	function async(callback)
    	{
    	  var isAsync = false;

    	  // check if async happened
    	  defer(function() { isAsync = true; });

    	  return function async_callback(err, result)
    	  {
    	    if (isAsync)
    	    {
    	      callback(err, result);
    	    }
    	    else
    	    {
    	      defer(function nextTick_callback()
    	      {
    	        callback(err, result);
    	      });
    	    }
    	  };
    	}
    	return async_1;
    }

    var abort_1;
    var hasRequiredAbort;

    function requireAbort () {
    	if (hasRequiredAbort) return abort_1;
    	hasRequiredAbort = 1;
    	// API
    	abort_1 = abort;

    	/**
    	 * Aborts leftover active jobs
    	 *
    	 * @param {object} state - current state object
    	 */
    	function abort(state)
    	{
    	  Object.keys(state.jobs).forEach(clean.bind(state));

    	  // reset leftover jobs
    	  state.jobs = {};
    	}

    	/**
    	 * Cleans up leftover job by invoking abort function for the provided job id
    	 *
    	 * @this  state
    	 * @param {string|number} key - job id to abort
    	 */
    	function clean(key)
    	{
    	  if (typeof this.jobs[key] == 'function')
    	  {
    	    this.jobs[key]();
    	  }
    	}
    	return abort_1;
    }

    var iterate_1;
    var hasRequiredIterate;

    function requireIterate () {
    	if (hasRequiredIterate) return iterate_1;
    	hasRequiredIterate = 1;
    	var async = requireAsync()
    	  , abort = requireAbort()
    	  ;

    	// API
    	iterate_1 = iterate;

    	/**
    	 * Iterates over each job object
    	 *
    	 * @param {array|object} list - array or object (named list) to iterate over
    	 * @param {function} iterator - iterator to run
    	 * @param {object} state - current job status
    	 * @param {function} callback - invoked when all elements processed
    	 */
    	function iterate(list, iterator, state, callback)
    	{
    	  // store current index
    	  var key = state['keyedList'] ? state['keyedList'][state.index] : state.index;

    	  state.jobs[key] = runJob(iterator, key, list[key], function(error, output)
    	  {
    	    // don't repeat yourself
    	    // skip secondary callbacks
    	    if (!(key in state.jobs))
    	    {
    	      return;
    	    }

    	    // clean up jobs
    	    delete state.jobs[key];

    	    if (error)
    	    {
    	      // don't process rest of the results
    	      // stop still active jobs
    	      // and reset the list
    	      abort(state);
    	    }
    	    else
    	    {
    	      state.results[key] = output;
    	    }

    	    // return salvaged results
    	    callback(error, state.results);
    	  });
    	}

    	/**
    	 * Runs iterator over provided job element
    	 *
    	 * @param   {function} iterator - iterator to invoke
    	 * @param   {string|number} key - key/index of the element in the list of jobs
    	 * @param   {mixed} item - job description
    	 * @param   {function} callback - invoked after iterator is done with the job
    	 * @returns {function|mixed} - job abort function or something else
    	 */
    	function runJob(iterator, key, item, callback)
    	{
    	  var aborter;

    	  // allow shortcut if iterator expects only two arguments
    	  if (iterator.length == 2)
    	  {
    	    aborter = iterator(item, async(callback));
    	  }
    	  // otherwise go with full three arguments
    	  else
    	  {
    	    aborter = iterator(item, key, async(callback));
    	  }

    	  return aborter;
    	}
    	return iterate_1;
    }

    var state_1;
    var hasRequiredState;

    function requireState () {
    	if (hasRequiredState) return state_1;
    	hasRequiredState = 1;
    	// API
    	state_1 = state;

    	/**
    	 * Creates initial state object
    	 * for iteration over list
    	 *
    	 * @param   {array|object} list - list to iterate over
    	 * @param   {function|null} sortMethod - function to use for keys sort,
    	 *                                     or `null` to keep them as is
    	 * @returns {object} - initial state object
    	 */
    	function state(list, sortMethod)
    	{
    	  var isNamedList = !Array.isArray(list)
    	    , initState =
    	    {
    	      index    : 0,
    	      keyedList: isNamedList || sortMethod ? Object.keys(list) : null,
    	      jobs     : {},
    	      results  : isNamedList ? {} : [],
    	      size     : isNamedList ? Object.keys(list).length : list.length
    	    }
    	    ;

    	  if (sortMethod)
    	  {
    	    // sort array keys based on it's values
    	    // sort object's keys just on own merit
    	    initState.keyedList.sort(isNamedList ? sortMethod : function(a, b)
    	    {
    	      return sortMethod(list[a], list[b]);
    	    });
    	  }

    	  return initState;
    	}
    	return state_1;
    }

    var terminator_1;
    var hasRequiredTerminator;

    function requireTerminator () {
    	if (hasRequiredTerminator) return terminator_1;
    	hasRequiredTerminator = 1;
    	var abort = requireAbort()
    	  , async = requireAsync()
    	  ;

    	// API
    	terminator_1 = terminator;

    	/**
    	 * Terminates jobs in the attached state context
    	 *
    	 * @this  AsyncKitState#
    	 * @param {function} callback - final callback to invoke after termination
    	 */
    	function terminator(callback)
    	{
    	  if (!Object.keys(this.jobs).length)
    	  {
    	    return;
    	  }

    	  // fast forward iteration index
    	  this.index = this.size;

    	  // abort jobs
    	  abort(this);

    	  // send back results we have so far
    	  async(callback)(null, this.results);
    	}
    	return terminator_1;
    }

    var parallel_1;
    var hasRequiredParallel;

    function requireParallel () {
    	if (hasRequiredParallel) return parallel_1;
    	hasRequiredParallel = 1;
    	var iterate    = requireIterate()
    	  , initState  = requireState()
    	  , terminator = requireTerminator()
    	  ;

    	// Public API
    	parallel_1 = parallel;

    	/**
    	 * Runs iterator over provided array elements in parallel
    	 *
    	 * @param   {array|object} list - array or object (named list) to iterate over
    	 * @param   {function} iterator - iterator to run
    	 * @param   {function} callback - invoked when all elements processed
    	 * @returns {function} - jobs terminator
    	 */
    	function parallel(list, iterator, callback)
    	{
    	  var state = initState(list);

    	  while (state.index < (state['keyedList'] || list).length)
    	  {
    	    iterate(list, iterator, state, function(error, result)
    	    {
    	      if (error)
    	      {
    	        callback(error, result);
    	        return;
    	      }

    	      // looks like it's the last one
    	      if (Object.keys(state.jobs).length === 0)
    	      {
    	        callback(null, state.results);
    	        return;
    	      }
    	    });

    	    state.index++;
    	  }

    	  return terminator.bind(state, callback);
    	}
    	return parallel_1;
    }

    var serialOrdered = {exports: {}};

    var hasRequiredSerialOrdered;

    function requireSerialOrdered () {
    	if (hasRequiredSerialOrdered) return serialOrdered.exports;
    	hasRequiredSerialOrdered = 1;
    	var iterate    = requireIterate()
    	  , initState  = requireState()
    	  , terminator = requireTerminator()
    	  ;

    	// Public API
    	serialOrdered.exports = serialOrdered$1;
    	// sorting helpers
    	serialOrdered.exports.ascending  = ascending;
    	serialOrdered.exports.descending = descending;

    	/**
    	 * Runs iterator over provided sorted array elements in series
    	 *
    	 * @param   {array|object} list - array or object (named list) to iterate over
    	 * @param   {function} iterator - iterator to run
    	 * @param   {function} sortMethod - custom sort function
    	 * @param   {function} callback - invoked when all elements processed
    	 * @returns {function} - jobs terminator
    	 */
    	function serialOrdered$1(list, iterator, sortMethod, callback)
    	{
    	  var state = initState(list, sortMethod);

    	  iterate(list, iterator, state, function iteratorHandler(error, result)
    	  {
    	    if (error)
    	    {
    	      callback(error, result);
    	      return;
    	    }

    	    state.index++;

    	    // are we there yet?
    	    if (state.index < (state['keyedList'] || list).length)
    	    {
    	      iterate(list, iterator, state, iteratorHandler);
    	      return;
    	    }

    	    // done here
    	    callback(null, state.results);
    	  });

    	  return terminator.bind(state, callback);
    	}

    	/*
    	 * -- Sort methods
    	 */

    	/**
    	 * sort helper to sort array elements in ascending order
    	 *
    	 * @param   {mixed} a - an item to compare
    	 * @param   {mixed} b - an item to compare
    	 * @returns {number} - comparison result
    	 */
    	function ascending(a, b)
    	{
    	  return a < b ? -1 : a > b ? 1 : 0;
    	}

    	/**
    	 * sort helper to sort array elements in descending order
    	 *
    	 * @param   {mixed} a - an item to compare
    	 * @param   {mixed} b - an item to compare
    	 * @returns {number} - comparison result
    	 */
    	function descending(a, b)
    	{
    	  return -1 * ascending(a, b);
    	}
    	return serialOrdered.exports;
    }

    var serial_1;
    var hasRequiredSerial;

    function requireSerial () {
    	if (hasRequiredSerial) return serial_1;
    	hasRequiredSerial = 1;
    	var serialOrdered = requireSerialOrdered();

    	// Public API
    	serial_1 = serial;

    	/**
    	 * Runs iterator over provided array elements in series
    	 *
    	 * @param   {array|object} list - array or object (named list) to iterate over
    	 * @param   {function} iterator - iterator to run
    	 * @param   {function} callback - invoked when all elements processed
    	 * @returns {function} - jobs terminator
    	 */
    	function serial(list, iterator, callback)
    	{
    	  return serialOrdered(list, iterator, null, callback);
    	}
    	return serial_1;
    }

    var asynckit;
    var hasRequiredAsynckit;

    function requireAsynckit () {
    	if (hasRequiredAsynckit) return asynckit;
    	hasRequiredAsynckit = 1;
    	asynckit =
    	{
    	  parallel      : requireParallel(),
    	  serial        : requireSerial(),
    	  serialOrdered : requireSerialOrdered()
    	};
    	return asynckit;
    }

    var populate;
    var hasRequiredPopulate;

    function requirePopulate () {
    	if (hasRequiredPopulate) return populate;
    	hasRequiredPopulate = 1;
    	// populates missing values
    	populate = function(dst, src) {

    	  Object.keys(src).forEach(function(prop)
    	  {
    	    dst[prop] = dst[prop] || src[prop];
    	  });

    	  return dst;
    	};
    	return populate;
    }

    var form_data;
    var hasRequiredForm_data;

    function requireForm_data () {
    	if (hasRequiredForm_data) return form_data;
    	hasRequiredForm_data = 1;
    	var CombinedStream = requireCombined_stream();
    	var util = require$$1;
    	var path = require$$2;
    	var http = require$$4$1;
    	var https = require$$4$1;
    	var parseUrl = require$$5.parse;
    	var fs = require$$6;
    	var Stream = require$$7.Stream;
    	var mime = requireMimeTypes();
    	var asynckit = requireAsynckit();
    	var populate = requirePopulate();

    	// Public API
    	form_data = FormData;

    	// make it a Stream
    	util.inherits(FormData, CombinedStream);

    	/**
    	 * Create readable "multipart/form-data" streams.
    	 * Can be used to submit forms
    	 * and file uploads to other web applications.
    	 *
    	 * @constructor
    	 * @param {Object} options - Properties to be added/overriden for FormData and CombinedStream
    	 */
    	function FormData(options) {
    	  if (!(this instanceof FormData)) {
    	    return new FormData(options);
    	  }

    	  this._overheadLength = 0;
    	  this._valueLength = 0;
    	  this._valuesToMeasure = [];

    	  CombinedStream.call(this);

    	  options = options || {};
    	  for (var option in options) {
    	    this[option] = options[option];
    	  }
    	}

    	FormData.LINE_BREAK = '\r\n';
    	FormData.DEFAULT_CONTENT_TYPE = 'application/octet-stream';

    	FormData.prototype.append = function(field, value, options) {

    	  options = options || {};

    	  // allow filename as single option
    	  if (typeof options == 'string') {
    	    options = {filename: options};
    	  }

    	  var append = CombinedStream.prototype.append.bind(this);

    	  // all that streamy business can't handle numbers
    	  if (typeof value == 'number') {
    	    value = '' + value;
    	  }

    	  // https://github.com/felixge/node-form-data/issues/38
    	  if (util.isArray(value)) {
    	    // Please convert your array into string
    	    // the way web server expects it
    	    this._error(new Error('Arrays are not supported.'));
    	    return;
    	  }

    	  var header = this._multiPartHeader(field, value, options);
    	  var footer = this._multiPartFooter();

    	  append(header);
    	  append(value);
    	  append(footer);

    	  // pass along options.knownLength
    	  this._trackLength(header, value, options);
    	};

    	FormData.prototype._trackLength = function(header, value, options) {
    	  var valueLength = 0;

    	  // used w/ getLengthSync(), when length is known.
    	  // e.g. for streaming directly from a remote server,
    	  // w/ a known file a size, and not wanting to wait for
    	  // incoming file to finish to get its size.
    	  if (options.knownLength != null) {
    	    valueLength += +options.knownLength;
    	  } else if (isBuffer$2(value)) {
    	    valueLength = value.length;
    	  } else if (typeof value === 'string') {
    	    valueLength = Buffer$1.byteLength(value);
    	  }

    	  this._valueLength += valueLength;

    	  // @check why add CRLF? does this account for custom/multiple CRLFs?
    	  this._overheadLength +=
    	    Buffer$1.byteLength(header) +
    	    FormData.LINE_BREAK.length;

    	  // empty or either doesn't have path or not an http response or not a stream
    	  if (!value || ( !value.path && !(value.readable && value.hasOwnProperty('httpVersion')) && !(value instanceof Stream))) {
    	    return;
    	  }

    	  // no need to bother with the length
    	  if (!options.knownLength) {
    	    this._valuesToMeasure.push(value);
    	  }
    	};

    	FormData.prototype._lengthRetriever = function(value, callback) {

    	  if (value.hasOwnProperty('fd')) {

    	    // take read range into a account
    	    // `end` = Infinity –> read file till the end
    	    //
    	    // TODO: Looks like there is bug in Node fs.createReadStream
    	    // it doesn't respect `end` options without `start` options
    	    // Fix it when node fixes it.
    	    // https://github.com/joyent/node/issues/7819
    	    if (value.end != undefined && value.end != Infinity && value.start != undefined) {

    	      // when end specified
    	      // no need to calculate range
    	      // inclusive, starts with 0
    	      callback(null, value.end + 1 - (value.start ? value.start : 0));

    	    // not that fast snoopy
    	    } else {
    	      // still need to fetch file size from fs
    	      fs.stat(value.path, function(err, stat) {

    	        var fileSize;

    	        if (err) {
    	          callback(err);
    	          return;
    	        }

    	        // update final size based on the range options
    	        fileSize = stat.size - (value.start ? value.start : 0);
    	        callback(null, fileSize);
    	      });
    	    }

    	  // or http response
    	  } else if (value.hasOwnProperty('httpVersion')) {
    	    callback(null, +value.headers['content-length']);

    	  // or request stream http://github.com/mikeal/request
    	  } else if (value.hasOwnProperty('httpModule')) {
    	    // wait till response come back
    	    value.on('response', function(response) {
    	      value.pause();
    	      callback(null, +response.headers['content-length']);
    	    });
    	    value.resume();

    	  // something else
    	  } else {
    	    callback('Unknown stream');
    	  }
    	};

    	FormData.prototype._multiPartHeader = function(field, value, options) {
    	  // custom header specified (as string)?
    	  // it becomes responsible for boundary
    	  // (e.g. to handle extra CRLFs on .NET servers)
    	  if (typeof options.header == 'string') {
    	    return options.header;
    	  }

    	  var contentDisposition = this._getContentDisposition(value, options);
    	  var contentType = this._getContentType(value, options);

    	  var contents = '';
    	  var headers  = {
    	    // add custom disposition as third element or keep it two elements if not
    	    'Content-Disposition': ['form-data', 'name="' + field + '"'].concat(contentDisposition || []),
    	    // if no content type. allow it to be empty array
    	    'Content-Type': [].concat(contentType || [])
    	  };

    	  // allow custom headers.
    	  if (typeof options.header == 'object') {
    	    populate(headers, options.header);
    	  }

    	  var header;
    	  for (var prop in headers) {
    	    if (!headers.hasOwnProperty(prop)) continue;
    	    header = headers[prop];

    	    // skip nullish headers.
    	    if (header == null) {
    	      continue;
    	    }

    	    // convert all headers to arrays.
    	    if (!Array.isArray(header)) {
    	      header = [header];
    	    }

    	    // add non-empty headers.
    	    if (header.length) {
    	      contents += prop + ': ' + header.join('; ') + FormData.LINE_BREAK;
    	    }
    	  }

    	  return '--' + this.getBoundary() + FormData.LINE_BREAK + contents + FormData.LINE_BREAK;
    	};

    	FormData.prototype._getContentDisposition = function(value, options) {

    	  var filename
    	    , contentDisposition
    	    ;

    	  if (typeof options.filepath === 'string') {
    	    // custom filepath for relative paths
    	    filename = path.normalize(options.filepath).replace(/\\/g, '/');
    	  } else if (options.filename || value.name || value.path) {
    	    // custom filename take precedence
    	    // formidable and the browser add a name property
    	    // fs- and request- streams have path property
    	    filename = path.basename(options.filename || value.name || value.path);
    	  } else if (value.readable && value.hasOwnProperty('httpVersion')) {
    	    // or try http response
    	    filename = path.basename(value.client._httpMessage.path || '');
    	  }

    	  if (filename) {
    	    contentDisposition = 'filename="' + filename + '"';
    	  }

    	  return contentDisposition;
    	};

    	FormData.prototype._getContentType = function(value, options) {

    	  // use custom content-type above all
    	  var contentType = options.contentType;

    	  // or try `name` from formidable, browser
    	  if (!contentType && value.name) {
    	    contentType = mime.lookup(value.name);
    	  }

    	  // or try `path` from fs-, request- streams
    	  if (!contentType && value.path) {
    	    contentType = mime.lookup(value.path);
    	  }

    	  // or if it's http-reponse
    	  if (!contentType && value.readable && value.hasOwnProperty('httpVersion')) {
    	    contentType = value.headers['content-type'];
    	  }

    	  // or guess it from the filepath or filename
    	  if (!contentType && (options.filepath || options.filename)) {
    	    contentType = mime.lookup(options.filepath || options.filename);
    	  }

    	  // fallback to the default content type if `value` is not simple value
    	  if (!contentType && typeof value == 'object') {
    	    contentType = FormData.DEFAULT_CONTENT_TYPE;
    	  }

    	  return contentType;
    	};

    	FormData.prototype._multiPartFooter = function() {
    	  return function(next) {
    	    var footer = FormData.LINE_BREAK;

    	    var lastPart = (this._streams.length === 0);
    	    if (lastPart) {
    	      footer += this._lastBoundary();
    	    }

    	    next(footer);
    	  }.bind(this);
    	};

    	FormData.prototype._lastBoundary = function() {
    	  return '--' + this.getBoundary() + '--' + FormData.LINE_BREAK;
    	};

    	FormData.prototype.getHeaders = function(userHeaders) {
    	  var header;
    	  var formHeaders = {
    	    'content-type': 'multipart/form-data; boundary=' + this.getBoundary()
    	  };

    	  for (header in userHeaders) {
    	    if (userHeaders.hasOwnProperty(header)) {
    	      formHeaders[header.toLowerCase()] = userHeaders[header];
    	    }
    	  }

    	  return formHeaders;
    	};

    	FormData.prototype.setBoundary = function(boundary) {
    	  this._boundary = boundary;
    	};

    	FormData.prototype.getBoundary = function() {
    	  if (!this._boundary) {
    	    this._generateBoundary();
    	  }

    	  return this._boundary;
    	};

    	FormData.prototype.getBuffer = function() {
    	  var dataBuffer = new Buffer$1.alloc( 0 );
    	  var boundary = this.getBoundary();

    	  // Create the form content. Add Line breaks to the end of data.
    	  for (var i = 0, len = this._streams.length; i < len; i++) {
    	    if (typeof this._streams[i] !== 'function') {

    	      // Add content to the buffer.
    	      if(isBuffer$2(this._streams[i])) {
    	        dataBuffer = Buffer$1.concat( [dataBuffer, this._streams[i]]);
    	      }else {
    	        dataBuffer = Buffer$1.concat( [dataBuffer, Buffer$1.from(this._streams[i])]);
    	      }

    	      // Add break after content.
    	      if (typeof this._streams[i] !== 'string' || this._streams[i].substring( 2, boundary.length + 2 ) !== boundary) {
    	        dataBuffer = Buffer$1.concat( [dataBuffer, Buffer$1.from(FormData.LINE_BREAK)] );
    	      }
    	    }
    	  }

    	  // Add the footer and return the Buffer object.
    	  return Buffer$1.concat( [dataBuffer, Buffer$1.from(this._lastBoundary())] );
    	};

    	FormData.prototype._generateBoundary = function() {
    	  // This generates a 50 character boundary similar to those used by Firefox.
    	  // They are optimized for boyer-moore parsing.
    	  var boundary = '--------------------------';
    	  for (var i = 0; i < 24; i++) {
    	    boundary += Math.floor(Math.random() * 10).toString(16);
    	  }

    	  this._boundary = boundary;
    	};

    	// Note: getLengthSync DOESN'T calculate streams length
    	// As workaround one can calculate file size manually
    	// and add it as knownLength option
    	FormData.prototype.getLengthSync = function() {
    	  var knownLength = this._overheadLength + this._valueLength;

    	  // Don't get confused, there are 3 "internal" streams for each keyval pair
    	  // so it basically checks if there is any value added to the form
    	  if (this._streams.length) {
    	    knownLength += this._lastBoundary().length;
    	  }

    	  // https://github.com/form-data/form-data/issues/40
    	  if (!this.hasKnownLength()) {
    	    // Some async length retrievers are present
    	    // therefore synchronous length calculation is false.
    	    // Please use getLength(callback) to get proper length
    	    this._error(new Error('Cannot calculate proper length in synchronous way.'));
    	  }

    	  return knownLength;
    	};

    	// Public API to check if length of added values is known
    	// https://github.com/form-data/form-data/issues/196
    	// https://github.com/form-data/form-data/issues/262
    	FormData.prototype.hasKnownLength = function() {
    	  var hasKnownLength = true;

    	  if (this._valuesToMeasure.length) {
    	    hasKnownLength = false;
    	  }

    	  return hasKnownLength;
    	};

    	FormData.prototype.getLength = function(cb) {
    	  var knownLength = this._overheadLength + this._valueLength;

    	  if (this._streams.length) {
    	    knownLength += this._lastBoundary().length;
    	  }

    	  if (!this._valuesToMeasure.length) {
    	    nextTick(cb.bind(this, null, knownLength));
    	    return;
    	  }

    	  asynckit.parallel(this._valuesToMeasure, this._lengthRetriever, function(err, values) {
    	    if (err) {
    	      cb(err);
    	      return;
    	    }

    	    values.forEach(function(length) {
    	      knownLength += length;
    	    });

    	    cb(null, knownLength);
    	  });
    	};

    	FormData.prototype.submit = function(params, cb) {
    	  var request
    	    , options
    	    , defaults = {method: 'post'}
    	    ;

    	  // parse provided url if it's string
    	  // or treat it as options object
    	  if (typeof params == 'string') {

    	    params = parseUrl(params);
    	    options = populate({
    	      port: params.port,
    	      path: params.pathname,
    	      host: params.hostname,
    	      protocol: params.protocol
    	    }, defaults);

    	  // use custom params
    	  } else {

    	    options = populate(params, defaults);
    	    // if no port provided use default one
    	    if (!options.port) {
    	      options.port = options.protocol == 'https:' ? 443 : 80;
    	    }
    	  }

    	  // put that good code in getHeaders to some use
    	  options.headers = this.getHeaders(params.headers);

    	  // https if specified, fallback to http in any other case
    	  if (options.protocol == 'https:') {
    	    request = https.request(options);
    	  } else {
    	    request = http.request(options);
    	  }

    	  // get content length and fire away
    	  this.getLength(function(err, length) {
    	    if (err && err !== 'Unknown stream') {
    	      this._error(err);
    	      return;
    	    }

    	    // add content length
    	    if (length) {
    	      request.setHeader('Content-Length', length);
    	    }

    	    this.pipe(request);
    	    if (cb) {
    	      var onResponse;

    	      var callback = function (error, responce) {
    	        request.removeListener('error', callback);
    	        request.removeListener('response', onResponse);

    	        return cb.call(this, error, responce);
    	      };

    	      onResponse = callback.bind(this, null);

    	      request.on('error', callback);
    	      request.on('response', onResponse);
    	    }
    	  }.bind(this));

    	  return request;
    	};

    	FormData.prototype._error = function(err) {
    	  if (!this.error) {
    	    this.error = err;
    	    this.pause();
    	    this.emit('error', err);
    	  }
    	};

    	FormData.prototype.toString = function () {
    	  return '[object FormData]';
    	};
    	return form_data;
    }

    var hasRequiredFormData;

    function requireFormData () {
    	if (hasRequiredFormData) return FormData$1.exports;
    	hasRequiredFormData = 1;
    	(function (module) {
    		// eslint-disable-next-line strict
    		module.exports = requireForm_data();
    } (FormData$1));
    	return FormData$1.exports;
    }

    var utils$5 = utils$9;
    var normalizeHeaderName = normalizeHeaderName$1;
    var AxiosError$1 = requireAxiosError();
    var transitionalDefaults = transitional;
    var toFormData = requireToFormData();

    var DEFAULT_CONTENT_TYPE = {
      'Content-Type': 'application/x-www-form-urlencoded'
    };

    function setContentTypeIfUnset(headers, value) {
      if (!utils$5.isUndefined(headers) && utils$5.isUndefined(headers['Content-Type'])) {
        headers['Content-Type'] = value;
      }
    }

    function getDefaultAdapter() {
      var adapter;
      if (typeof XMLHttpRequest !== 'undefined') {
        // For browsers use XHR adapter
        adapter = requireXhr();
      } else if (typeof process !== 'undefined' && Object.prototype.toString.call(process) === '[object process]') {
        // For node use HTTP adapter
        adapter = requireHttp();
      }
      return adapter;
    }

    function stringifySafely(rawValue, parser, encoder) {
      if (utils$5.isString(rawValue)) {
        try {
          (parser || JSON.parse)(rawValue);
          return utils$5.trim(rawValue);
        } catch (e) {
          if (e.name !== 'SyntaxError') {
            throw e;
          }
        }
      }

      return (encoder || JSON.stringify)(rawValue);
    }

    var defaults$3 = {

      transitional: transitionalDefaults,

      adapter: getDefaultAdapter(),

      transformRequest: [function transformRequest(data, headers) {
        normalizeHeaderName(headers, 'Accept');
        normalizeHeaderName(headers, 'Content-Type');

        if (utils$5.isFormData(data) ||
          utils$5.isArrayBuffer(data) ||
          utils$5.isBuffer(data) ||
          utils$5.isStream(data) ||
          utils$5.isFile(data) ||
          utils$5.isBlob(data)
        ) {
          return data;
        }
        if (utils$5.isArrayBufferView(data)) {
          return data.buffer;
        }
        if (utils$5.isURLSearchParams(data)) {
          setContentTypeIfUnset(headers, 'application/x-www-form-urlencoded;charset=utf-8');
          return data.toString();
        }

        var isObjectPayload = utils$5.isObject(data);
        var contentType = headers && headers['Content-Type'];

        var isFileList;

        if ((isFileList = utils$5.isFileList(data)) || (isObjectPayload && contentType === 'multipart/form-data')) {
          var _FormData = this.env && this.env.FormData;
          return toFormData(isFileList ? {'files[]': data} : data, _FormData && new _FormData());
        } else if (isObjectPayload || contentType === 'application/json') {
          setContentTypeIfUnset(headers, 'application/json');
          return stringifySafely(data);
        }

        return data;
      }],

      transformResponse: [function transformResponse(data) {
        var transitional = this.transitional || defaults$3.transitional;
        var silentJSONParsing = transitional && transitional.silentJSONParsing;
        var forcedJSONParsing = transitional && transitional.forcedJSONParsing;
        var strictJSONParsing = !silentJSONParsing && this.responseType === 'json';

        if (strictJSONParsing || (forcedJSONParsing && utils$5.isString(data) && data.length)) {
          try {
            return JSON.parse(data);
          } catch (e) {
            if (strictJSONParsing) {
              if (e.name === 'SyntaxError') {
                throw AxiosError$1.from(e, AxiosError$1.ERR_BAD_RESPONSE, this, null, this.response);
              }
              throw e;
            }
          }
        }

        return data;
      }],

      /**
       * A timeout in milliseconds to abort a request. If set to 0 (default) a
       * timeout is not created.
       */
      timeout: 0,

      xsrfCookieName: 'XSRF-TOKEN',
      xsrfHeaderName: 'X-XSRF-TOKEN',

      maxContentLength: -1,
      maxBodyLength: -1,

      env: {
        FormData: requireFormData()
      },

      validateStatus: function validateStatus(status) {
        return status >= 200 && status < 300;
      },

      headers: {
        common: {
          'Accept': 'application/json, text/plain, */*'
        }
      }
    };

    utils$5.forEach(['delete', 'get', 'head'], function forEachMethodNoData(method) {
      defaults$3.headers[method] = {};
    });

    utils$5.forEach(['post', 'put', 'patch'], function forEachMethodWithData(method) {
      defaults$3.headers[method] = utils$5.merge(DEFAULT_CONTENT_TYPE);
    });

    var defaults_1 = defaults$3;

    var utils$4 = utils$9;
    var defaults$2 = defaults_1;

    /**
     * Transform the data for a request or a response
     *
     * @param {Object|String} data The data to be transformed
     * @param {Array} headers The headers for the request or response
     * @param {Array|Function} fns A single function or Array of functions
     * @returns {*} The resulting transformed data
     */
    var transformData$1 = function transformData(data, headers, fns) {
      var context = this || defaults$2;
      /*eslint no-param-reassign:0*/
      utils$4.forEach(fns, function transform(fn) {
        data = fn.call(context, data, headers);
      });

      return data;
    };

    var isCancel$1;
    var hasRequiredIsCancel;

    function requireIsCancel () {
    	if (hasRequiredIsCancel) return isCancel$1;
    	hasRequiredIsCancel = 1;

    	isCancel$1 = function isCancel(value) {
    	  return !!(value && value.__CANCEL__);
    	};
    	return isCancel$1;
    }

    var utils$3 = utils$9;
    var transformData = transformData$1;
    var isCancel = requireIsCancel();
    var defaults$1 = defaults_1;
    var CanceledError = requireCanceledError();

    /**
     * Throws a `CanceledError` if cancellation has been requested.
     */
    function throwIfCancellationRequested(config) {
      if (config.cancelToken) {
        config.cancelToken.throwIfRequested();
      }

      if (config.signal && config.signal.aborted) {
        throw new CanceledError();
      }
    }

    /**
     * Dispatch a request to the server using the configured adapter.
     *
     * @param {object} config The config that is to be used for the request
     * @returns {Promise} The Promise to be fulfilled
     */
    var dispatchRequest$1 = function dispatchRequest(config) {
      throwIfCancellationRequested(config);

      // Ensure headers exist
      config.headers = config.headers || {};

      // Transform request data
      config.data = transformData.call(
        config,
        config.data,
        config.headers,
        config.transformRequest
      );

      // Flatten headers
      config.headers = utils$3.merge(
        config.headers.common || {},
        config.headers[config.method] || {},
        config.headers
      );

      utils$3.forEach(
        ['delete', 'get', 'head', 'post', 'put', 'patch', 'common'],
        function cleanHeaderConfig(method) {
          delete config.headers[method];
        }
      );

      var adapter = config.adapter || defaults$1.adapter;

      return adapter(config).then(function onAdapterResolution(response) {
        throwIfCancellationRequested(config);

        // Transform response data
        response.data = transformData.call(
          config,
          response.data,
          response.headers,
          config.transformResponse
        );

        return response;
      }, function onAdapterRejection(reason) {
        if (!isCancel(reason)) {
          throwIfCancellationRequested(config);

          // Transform response data
          if (reason && reason.response) {
            reason.response.data = transformData.call(
              config,
              reason.response.data,
              reason.response.headers,
              config.transformResponse
            );
          }
        }

        return Promise.reject(reason);
      });
    };

    var utils$2 = utils$9;

    /**
     * Config-specific merge-function which creates a new config-object
     * by merging two configuration objects together.
     *
     * @param {Object} config1
     * @param {Object} config2
     * @returns {Object} New object resulting from merging config2 to config1
     */
    var mergeConfig$2 = function mergeConfig(config1, config2) {
      // eslint-disable-next-line no-param-reassign
      config2 = config2 || {};
      var config = {};

      function getMergedValue(target, source) {
        if (utils$2.isPlainObject(target) && utils$2.isPlainObject(source)) {
          return utils$2.merge(target, source);
        } else if (utils$2.isPlainObject(source)) {
          return utils$2.merge({}, source);
        } else if (utils$2.isArray(source)) {
          return source.slice();
        }
        return source;
      }

      // eslint-disable-next-line consistent-return
      function mergeDeepProperties(prop) {
        if (!utils$2.isUndefined(config2[prop])) {
          return getMergedValue(config1[prop], config2[prop]);
        } else if (!utils$2.isUndefined(config1[prop])) {
          return getMergedValue(undefined, config1[prop]);
        }
      }

      // eslint-disable-next-line consistent-return
      function valueFromConfig2(prop) {
        if (!utils$2.isUndefined(config2[prop])) {
          return getMergedValue(undefined, config2[prop]);
        }
      }

      // eslint-disable-next-line consistent-return
      function defaultToConfig2(prop) {
        if (!utils$2.isUndefined(config2[prop])) {
          return getMergedValue(undefined, config2[prop]);
        } else if (!utils$2.isUndefined(config1[prop])) {
          return getMergedValue(undefined, config1[prop]);
        }
      }

      // eslint-disable-next-line consistent-return
      function mergeDirectKeys(prop) {
        if (prop in config2) {
          return getMergedValue(config1[prop], config2[prop]);
        } else if (prop in config1) {
          return getMergedValue(undefined, config1[prop]);
        }
      }

      var mergeMap = {
        'url': valueFromConfig2,
        'method': valueFromConfig2,
        'data': valueFromConfig2,
        'baseURL': defaultToConfig2,
        'transformRequest': defaultToConfig2,
        'transformResponse': defaultToConfig2,
        'paramsSerializer': defaultToConfig2,
        'timeout': defaultToConfig2,
        'timeoutMessage': defaultToConfig2,
        'withCredentials': defaultToConfig2,
        'adapter': defaultToConfig2,
        'responseType': defaultToConfig2,
        'xsrfCookieName': defaultToConfig2,
        'xsrfHeaderName': defaultToConfig2,
        'onUploadProgress': defaultToConfig2,
        'onDownloadProgress': defaultToConfig2,
        'decompress': defaultToConfig2,
        'maxContentLength': defaultToConfig2,
        'maxBodyLength': defaultToConfig2,
        'beforeRedirect': defaultToConfig2,
        'transport': defaultToConfig2,
        'httpAgent': defaultToConfig2,
        'httpsAgent': defaultToConfig2,
        'cancelToken': defaultToConfig2,
        'socketPath': defaultToConfig2,
        'responseEncoding': defaultToConfig2,
        'validateStatus': mergeDirectKeys
      };

      utils$2.forEach(Object.keys(config1).concat(Object.keys(config2)), function computeConfigValue(prop) {
        var merge = mergeMap[prop] || mergeDeepProperties;
        var configValue = merge(prop);
        (utils$2.isUndefined(configValue) && merge !== mergeDirectKeys) || (config[prop] = configValue);
      });

      return config;
    };

    var VERSION = requireData().version;
    var AxiosError = requireAxiosError();

    var validators$1 = {};

    // eslint-disable-next-line func-names
    ['object', 'boolean', 'number', 'function', 'string', 'symbol'].forEach(function(type, i) {
      validators$1[type] = function validator(thing) {
        return typeof thing === type || 'a' + (i < 1 ? 'n ' : ' ') + type;
      };
    });

    var deprecatedWarnings = {};

    /**
     * Transitional option validator
     * @param {function|boolean?} validator - set to false if the transitional option has been removed
     * @param {string?} version - deprecated version / removed since version
     * @param {string?} message - some message with additional info
     * @returns {function}
     */
    validators$1.transitional = function transitional(validator, version, message) {
      function formatMessage(opt, desc) {
        return '[Axios v' + VERSION + '] Transitional option \'' + opt + '\'' + desc + (message ? '. ' + message : '');
      }

      // eslint-disable-next-line func-names
      return function(value, opt, opts) {
        if (validator === false) {
          throw new AxiosError(
            formatMessage(opt, ' has been removed' + (version ? ' in ' + version : '')),
            AxiosError.ERR_DEPRECATED
          );
        }

        if (version && !deprecatedWarnings[opt]) {
          deprecatedWarnings[opt] = true;
          // eslint-disable-next-line no-console
          console.warn(
            formatMessage(
              opt,
              ' has been deprecated since v' + version + ' and will be removed in the near future'
            )
          );
        }

        return validator ? validator(value, opt, opts) : true;
      };
    };

    /**
     * Assert object's properties type
     * @param {object} options
     * @param {object} schema
     * @param {boolean?} allowUnknown
     */

    function assertOptions(options, schema, allowUnknown) {
      if (typeof options !== 'object') {
        throw new AxiosError('options must be an object', AxiosError.ERR_BAD_OPTION_VALUE);
      }
      var keys = Object.keys(options);
      var i = keys.length;
      while (i-- > 0) {
        var opt = keys[i];
        var validator = schema[opt];
        if (validator) {
          var value = options[opt];
          var result = value === undefined || validator(value, opt, options);
          if (result !== true) {
            throw new AxiosError('option ' + opt + ' must be ' + result, AxiosError.ERR_BAD_OPTION_VALUE);
          }
          continue;
        }
        if (allowUnknown !== true) {
          throw new AxiosError('Unknown option ' + opt, AxiosError.ERR_BAD_OPTION);
        }
      }
    }

    var validator$1 = {
      assertOptions: assertOptions,
      validators: validators$1
    };

    var utils$1 = utils$9;
    var buildURL = buildURL$1;
    var InterceptorManager = InterceptorManager_1;
    var dispatchRequest = dispatchRequest$1;
    var mergeConfig$1 = mergeConfig$2;
    var buildFullPath = buildFullPath$1;
    var validator = validator$1;

    var validators = validator.validators;
    /**
     * Create a new instance of Axios
     *
     * @param {Object} instanceConfig The default config for the instance
     */
    function Axios$1(instanceConfig) {
      this.defaults = instanceConfig;
      this.interceptors = {
        request: new InterceptorManager(),
        response: new InterceptorManager()
      };
    }

    /**
     * Dispatch a request
     *
     * @param {Object} config The config specific for this request (merged with this.defaults)
     */
    Axios$1.prototype.request = function request(configOrUrl, config) {
      /*eslint no-param-reassign:0*/
      // Allow for axios('example/url'[, config]) a la fetch API
      if (typeof configOrUrl === 'string') {
        config = config || {};
        config.url = configOrUrl;
      } else {
        config = configOrUrl || {};
      }

      config = mergeConfig$1(this.defaults, config);

      // Set config.method
      if (config.method) {
        config.method = config.method.toLowerCase();
      } else if (this.defaults.method) {
        config.method = this.defaults.method.toLowerCase();
      } else {
        config.method = 'get';
      }

      var transitional = config.transitional;

      if (transitional !== undefined) {
        validator.assertOptions(transitional, {
          silentJSONParsing: validators.transitional(validators.boolean),
          forcedJSONParsing: validators.transitional(validators.boolean),
          clarifyTimeoutError: validators.transitional(validators.boolean)
        }, false);
      }

      // filter out skipped interceptors
      var requestInterceptorChain = [];
      var synchronousRequestInterceptors = true;
      this.interceptors.request.forEach(function unshiftRequestInterceptors(interceptor) {
        if (typeof interceptor.runWhen === 'function' && interceptor.runWhen(config) === false) {
          return;
        }

        synchronousRequestInterceptors = synchronousRequestInterceptors && interceptor.synchronous;

        requestInterceptorChain.unshift(interceptor.fulfilled, interceptor.rejected);
      });

      var responseInterceptorChain = [];
      this.interceptors.response.forEach(function pushResponseInterceptors(interceptor) {
        responseInterceptorChain.push(interceptor.fulfilled, interceptor.rejected);
      });

      var promise;

      if (!synchronousRequestInterceptors) {
        var chain = [dispatchRequest, undefined];

        Array.prototype.unshift.apply(chain, requestInterceptorChain);
        chain = chain.concat(responseInterceptorChain);

        promise = Promise.resolve(config);
        while (chain.length) {
          promise = promise.then(chain.shift(), chain.shift());
        }

        return promise;
      }


      var newConfig = config;
      while (requestInterceptorChain.length) {
        var onFulfilled = requestInterceptorChain.shift();
        var onRejected = requestInterceptorChain.shift();
        try {
          newConfig = onFulfilled(newConfig);
        } catch (error) {
          onRejected(error);
          break;
        }
      }

      try {
        promise = dispatchRequest(newConfig);
      } catch (error) {
        return Promise.reject(error);
      }

      while (responseInterceptorChain.length) {
        promise = promise.then(responseInterceptorChain.shift(), responseInterceptorChain.shift());
      }

      return promise;
    };

    Axios$1.prototype.getUri = function getUri(config) {
      config = mergeConfig$1(this.defaults, config);
      var fullPath = buildFullPath(config.baseURL, config.url);
      return buildURL(fullPath, config.params, config.paramsSerializer);
    };

    // Provide aliases for supported request methods
    utils$1.forEach(['delete', 'get', 'head', 'options'], function forEachMethodNoData(method) {
      /*eslint func-names:0*/
      Axios$1.prototype[method] = function(url, config) {
        return this.request(mergeConfig$1(config || {}, {
          method: method,
          url: url,
          data: (config || {}).data
        }));
      };
    });

    utils$1.forEach(['post', 'put', 'patch'], function forEachMethodWithData(method) {
      /*eslint func-names:0*/

      function generateHTTPMethod(isForm) {
        return function httpMethod(url, data, config) {
          return this.request(mergeConfig$1(config || {}, {
            method: method,
            headers: isForm ? {
              'Content-Type': 'multipart/form-data'
            } : {},
            url: url,
            data: data
          }));
        };
      }

      Axios$1.prototype[method] = generateHTTPMethod();

      Axios$1.prototype[method + 'Form'] = generateHTTPMethod(true);
    });

    var Axios_1 = Axios$1;

    var CancelToken_1;
    var hasRequiredCancelToken;

    function requireCancelToken () {
    	if (hasRequiredCancelToken) return CancelToken_1;
    	hasRequiredCancelToken = 1;

    	var CanceledError = requireCanceledError();

    	/**
    	 * A `CancelToken` is an object that can be used to request cancellation of an operation.
    	 *
    	 * @class
    	 * @param {Function} executor The executor function.
    	 */
    	function CancelToken(executor) {
    	  if (typeof executor !== 'function') {
    	    throw new TypeError('executor must be a function.');
    	  }

    	  var resolvePromise;

    	  this.promise = new Promise(function promiseExecutor(resolve) {
    	    resolvePromise = resolve;
    	  });

    	  var token = this;

    	  // eslint-disable-next-line func-names
    	  this.promise.then(function(cancel) {
    	    if (!token._listeners) return;

    	    var i;
    	    var l = token._listeners.length;

    	    for (i = 0; i < l; i++) {
    	      token._listeners[i](cancel);
    	    }
    	    token._listeners = null;
    	  });

    	  // eslint-disable-next-line func-names
    	  this.promise.then = function(onfulfilled) {
    	    var _resolve;
    	    // eslint-disable-next-line func-names
    	    var promise = new Promise(function(resolve) {
    	      token.subscribe(resolve);
    	      _resolve = resolve;
    	    }).then(onfulfilled);

    	    promise.cancel = function reject() {
    	      token.unsubscribe(_resolve);
    	    };

    	    return promise;
    	  };

    	  executor(function cancel(message) {
    	    if (token.reason) {
    	      // Cancellation has already been requested
    	      return;
    	    }

    	    token.reason = new CanceledError(message);
    	    resolvePromise(token.reason);
    	  });
    	}

    	/**
    	 * Throws a `CanceledError` if cancellation has been requested.
    	 */
    	CancelToken.prototype.throwIfRequested = function throwIfRequested() {
    	  if (this.reason) {
    	    throw this.reason;
    	  }
    	};

    	/**
    	 * Subscribe to the cancel signal
    	 */

    	CancelToken.prototype.subscribe = function subscribe(listener) {
    	  if (this.reason) {
    	    listener(this.reason);
    	    return;
    	  }

    	  if (this._listeners) {
    	    this._listeners.push(listener);
    	  } else {
    	    this._listeners = [listener];
    	  }
    	};

    	/**
    	 * Unsubscribe from the cancel signal
    	 */

    	CancelToken.prototype.unsubscribe = function unsubscribe(listener) {
    	  if (!this._listeners) {
    	    return;
    	  }
    	  var index = this._listeners.indexOf(listener);
    	  if (index !== -1) {
    	    this._listeners.splice(index, 1);
    	  }
    	};

    	/**
    	 * Returns an object that contains a new `CancelToken` and a function that, when called,
    	 * cancels the `CancelToken`.
    	 */
    	CancelToken.source = function source() {
    	  var cancel;
    	  var token = new CancelToken(function executor(c) {
    	    cancel = c;
    	  });
    	  return {
    	    token: token,
    	    cancel: cancel
    	  };
    	};

    	CancelToken_1 = CancelToken;
    	return CancelToken_1;
    }

    var spread;
    var hasRequiredSpread;

    function requireSpread () {
    	if (hasRequiredSpread) return spread;
    	hasRequiredSpread = 1;

    	/**
    	 * Syntactic sugar for invoking a function and expanding an array for arguments.
    	 *
    	 * Common use case would be to use `Function.prototype.apply`.
    	 *
    	 *  ```js
    	 *  function f(x, y, z) {}
    	 *  var args = [1, 2, 3];
    	 *  f.apply(null, args);
    	 *  ```
    	 *
    	 * With `spread` this example can be re-written.
    	 *
    	 *  ```js
    	 *  spread(function(x, y, z) {})([1, 2, 3]);
    	 *  ```
    	 *
    	 * @param {Function} callback
    	 * @returns {Function}
    	 */
    	spread = function spread(callback) {
    	  return function wrap(arr) {
    	    return callback.apply(null, arr);
    	  };
    	};
    	return spread;
    }

    var isAxiosError;
    var hasRequiredIsAxiosError;

    function requireIsAxiosError () {
    	if (hasRequiredIsAxiosError) return isAxiosError;
    	hasRequiredIsAxiosError = 1;

    	var utils = utils$9;

    	/**
    	 * Determines whether the payload is an error thrown by Axios
    	 *
    	 * @param {*} payload The value to test
    	 * @returns {boolean} True if the payload is an error thrown by Axios, otherwise false
    	 */
    	isAxiosError = function isAxiosError(payload) {
    	  return utils.isObject(payload) && (payload.isAxiosError === true);
    	};
    	return isAxiosError;
    }

    var utils = utils$9;
    var bind = bind$2;
    var Axios = Axios_1;
    var mergeConfig = mergeConfig$2;
    var defaults = defaults_1;

    /**
     * Create an instance of Axios
     *
     * @param {Object} defaultConfig The default config for the instance
     * @return {Axios} A new instance of Axios
     */
    function createInstance(defaultConfig) {
      var context = new Axios(defaultConfig);
      var instance = bind(Axios.prototype.request, context);

      // Copy axios.prototype to instance
      utils.extend(instance, Axios.prototype, context);

      // Copy context to instance
      utils.extend(instance, context);

      // Factory for creating new instances
      instance.create = function create(instanceConfig) {
        return createInstance(mergeConfig(defaultConfig, instanceConfig));
      };

      return instance;
    }

    // Create the default instance to be exported
    var axios$1 = createInstance(defaults);

    // Expose Axios class to allow class inheritance
    axios$1.Axios = Axios;

    // Expose Cancel & CancelToken
    axios$1.CanceledError = requireCanceledError();
    axios$1.CancelToken = requireCancelToken();
    axios$1.isCancel = requireIsCancel();
    axios$1.VERSION = requireData().version;
    axios$1.toFormData = requireToFormData();

    // Expose AxiosError class
    axios$1.AxiosError = requireAxiosError();

    // alias for CanceledError for backward compatibility
    axios$1.Cancel = axios$1.CanceledError;

    // Expose all/spread
    axios$1.all = function all(promises) {
      return Promise.all(promises);
    };
    axios$1.spread = requireSpread();

    // Expose isAxiosError
    axios$1.isAxiosError = requireIsAxiosError();

    axios$2.exports = axios$1;

    // Allow use of default import syntax in TypeScript
    axios$2.exports.default = axios$1;

    (function (module) {
    	module.exports = axios$2.exports;
    } (axios$3));

    var axios = /*@__PURE__*/getDefaultExportFromCjs(axios$3.exports);

    const BASE_PATH = 'https://storage.googleapis.com/tfjs-models/savedmodel/universal_sentence_encoder';

    // import {loadTokenizer} from '@tensorflow-models/universal-sentence-encoder/dist/tokenizer/index';
    /**
     * Load the Tokenizer for use independently from the UniversalSentenceEncoder.
     *
     * @param {string} pathToVocabulary (optional) Provide a path to the vocabulary file.
     */
    async function loadVocabulary(pathToVocabulary) {
      const vocabulary = await axios.get(pathToVocabulary);
      return vocabulary.data;
    }
    async function loadTokenizer() {
      const vocabulary = await (loadVocabulary(`${BASE_PATH}/vocab.json`));
      // console.log('vocabulary',vocabulary)
      const tokenizer = new universalSentenceEncoder.exports.Tokenizer(vocabulary);
      return tokenizer;
    }
    let model;
    let tokenizer;
    /**
     * Text Embedding with Tensorflow Universal Sentence Encoder (USE)
     * @class TextEmbedding
     * @implements {TensorScriptModelInterface}
     */
    class TextEmbedding extends TensorScriptModelInterface {
      /**
       * @param {Object} options - Options for USE
       * @param {{model:Object,tf:Object,}} properties - extra instance properties
       */
      constructor(options = {}, properties) {
        const config = Object.assign({
        }, options);
        super(config, properties);
        this.type = 'TextEmbedding';

        return this;
      }
      /**
       * Asynchronously loads Universal Sentence Encoder and tokenizer
       * @override
       * @return {Object} returns loaded UniversalSentenceEncoder model
       */
      async train() {
        const promises = [];
        if (!model) promises.push(universalSentenceEncoder.exports.load());
        else promises.push(Promise.resolve(model));
        if (!tokenizer) promises.push(loadTokenizer());
        else promises.push(Promise.resolve(tokenizer));
        const USE = await Promise.all(promises);
        if (!model) model = USE[ 0 ];
        if (!tokenizer) tokenizer = USE[ 1 ];
        // console.log({model,tokenizer})
        this.model = model;
        this.tokenizer = tokenizer;
        this.trained = true;
        this.compiled = true;

        return this.model;
      }
      /**
       * Calculates sentence embeddings
       * @override
       * @param {Array<Array<number>>|Array<number>} input_array - new test independent variables
       * @param {Object} options - model prediction options
       * @return {{data: Promise}} returns tensorflow prediction 
       */
      calculate(input_array, options = {}) {
        if (!input_array || Array.isArray(input_array) === false) throw new Error('invalid input array of sentences');
        const embeddings = this.model.embed(input_array);
        return embeddings;
      }
      /**
       * Returns prediction values from tensorflow model
       * @param {Array<string>} input_matrix - array of sentences to embed 
       * @param {Boolean} [options.json=true] - return object instead of typed array
       * @param {Boolean} [options.probability=true] - return real values instead of integers
       * @return {Array<Array<number>>} predicted model values
       */
      async predict(input_array, options = {}) {
        const config = Object.assign({
          json: true,
          probability: true,
        }, options);
        const embeddings = await this.calculate(input_array, options);
        const predictions = await embeddings.data(); 
        if (config.json === false) {
          return predictions;
        } else {
          const shape = [input_array.length, 512, ];
          const predictionValues = (options.probability === false)
            ? Array.from(predictions).map(Math.round)
            : Array.from(predictions);
          return this.reshape(predictionValues, shape);
        }
      }
    }

    function commonjsRequire(path) {
    	throw new Error('Could not dynamically require "' + path + '". Please configure the dynamicRequireTargets or/and ignoreDynamicRequires option of @rollup/plugin-commonjs appropriately for this require call to work.');
    }

    var tsne_min = {exports: {}};

    (function (module, exports) {
    	(function(f){{module.exports=f();}})(function(){return (function e(t,n,r){function s(o,u){if(!n[o]){if(!t[o]){var a=typeof commonjsRequire=="function"&&commonjsRequire;if(!u&&a)return a(o,!0);if(i)return i(o,!0);var f=new Error("Cannot find module '"+o+"'");throw f.code="MODULE_NOT_FOUND",f}var l=n[o]={exports:{}};t[o][0].call(l.exports,function(e){var n=t[o][1][e];return s(n?n:e)},l,l.exports,e,t,n,r);}return n[o].exports}var i=typeof commonjsRequire=="function"&&commonjsRequire;for(var o=0;o<r.length;o++)s(r[o]);return s})({1:[function(require,module,exports){
    function Procedure(){this.argTypes=[],this.shimArgs=[],this.arrayArgs=[],this.arrayBlockIndices=[],this.scalarArgs=[],this.offsetArgs=[],this.offsetArgIndex=[],this.indexArgs=[],this.shapeArgs=[],this.funcName="",this.pre=null,this.body=null,this.post=null,this.debug=!1;}function compileCwise(r){var e=new Procedure;e.pre=r.pre,e.body=r.body,e.post=r.post;var s=r.args.slice(0);e.argTypes=s;for(var o=0;o<s.length;++o){var a=s[o];if("array"===a||"object"==typeof a&&a.blockIndices){if(e.argTypes[o]="array",e.arrayArgs.push(o),e.arrayBlockIndices.push(a.blockIndices?a.blockIndices:0),e.shimArgs.push("array"+o),o<e.pre.args.length&&e.pre.args[o].count>0)throw new Error("cwise: pre() block may not reference array args");if(o<e.post.args.length&&e.post.args[o].count>0)throw new Error("cwise: post() block may not reference array args")}else if("scalar"===a)e.scalarArgs.push(o),e.shimArgs.push("scalar"+o);else if("index"===a){if(e.indexArgs.push(o),o<e.pre.args.length&&e.pre.args[o].count>0)throw new Error("cwise: pre() block may not reference array index");if(o<e.body.args.length&&e.body.args[o].lvalue)throw new Error("cwise: body() block may not write to array index");if(o<e.post.args.length&&e.post.args[o].count>0)throw new Error("cwise: post() block may not reference array index")}else if("shape"===a){if(e.shapeArgs.push(o),o<e.pre.args.length&&e.pre.args[o].lvalue)throw new Error("cwise: pre() block may not write to array shape");if(o<e.body.args.length&&e.body.args[o].lvalue)throw new Error("cwise: body() block may not write to array shape");if(o<e.post.args.length&&e.post.args[o].lvalue)throw new Error("cwise: post() block may not write to array shape")}else {if("object"!=typeof a||!a.offset)throw new Error("cwise: Unknown argument type "+s[o]);e.argTypes[o]="offset",e.offsetArgs.push({array:a.array,offset:a.offset}),e.offsetArgIndex.push(o);}}if(e.arrayArgs.length<=0)throw new Error("cwise: No array arguments specified");if(e.pre.args.length>s.length)throw new Error("cwise: Too many arguments in pre() block");if(e.body.args.length>s.length)throw new Error("cwise: Too many arguments in body() block");if(e.post.args.length>s.length)throw new Error("cwise: Too many arguments in post() block");return e.debug=!!r.printCode||!!r.debug,e.funcName=r.funcName||"cwise",e.blockSize=r.blockSize||64,createThunk(e)}var createThunk=require("./lib/thunk.js");module.exports=compileCwise;
    	},{"./lib/thunk.js":3}],2:[function(require,module,exports){
    function innerFill(r,n,e){var s,a,o=r.length,i=n.arrayArgs.length,t=n.indexArgs.length>0,p=[],h=[],u=0,c=0;for(s=0;o>s;++s)h.push(["i",s,"=0"].join(""));for(a=0;i>a;++a)for(s=0;o>s;++s)c=u,u=r[s],0===s?h.push(["d",a,"s",s,"=t",a,"p",u].join("")):h.push(["d",a,"s",s,"=(t",a,"p",u,"-s",c,"*t",a,"p",c,")"].join(""));for(p.push("var "+h.join(",")),s=o-1;s>=0;--s)u=r[s],p.push(["for(i",s,"=0;i",s,"<s",u,";++i",s,"){"].join(""));for(p.push(e),s=0;o>s;++s){for(c=u,u=r[s],a=0;i>a;++a)p.push(["p",a,"+=d",a,"s",s].join(""));t&&(s>0&&p.push(["index[",c,"]-=s",c].join("")),p.push(["++index[",u,"]"].join(""))),p.push("}");}return p.join("\n")}function outerFill(r,n,e,s){for(var a=n.length,o=e.arrayArgs.length,i=e.blockSize,t=e.indexArgs.length>0,p=[],h=0;o>h;++h)p.push(["var offset",h,"=p",h].join(""));for(var h=r;a>h;++h)p.push(["for(var j"+h+"=SS[",n[h],"]|0;j",h,">0;){"].join("")),p.push(["if(j",h,"<",i,"){"].join("")),p.push(["s",n[h],"=j",h].join("")),p.push(["j",h,"=0"].join("")),p.push(["}else{s",n[h],"=",i].join("")),p.push(["j",h,"-=",i,"}"].join("")),t&&p.push(["index[",n[h],"]=j",h].join(""));for(var h=0;o>h;++h){for(var u=["offset"+h],c=r;a>c;++c)u.push(["j",c,"*t",h,"p",n[c]].join(""));p.push(["p",h,"=(",u.join("+"),")"].join(""));}p.push(innerFill(n,e,s));for(var h=r;a>h;++h)p.push("}");return p.join("\n")}function countMatches(r){for(var n=0,e=r[0].length;e>n;){for(var s=1;s<r.length;++s)if(r[s][n]!==r[0][n])return n;++n;}return n}function processBlock(r,n,e){for(var s=r.body,a=[],o=[],i=0;i<r.args.length;++i){var t=r.args[i];if(!(t.count<=0)){var p=new RegExp(t.name,"g"),h="",u=n.arrayArgs.indexOf(i);switch(n.argTypes[i]){case"offset":var c=n.offsetArgIndex.indexOf(i),l=n.offsetArgs[c];u=l.array,h="+q"+c;case"array":h="p"+u+h;var f="l"+i,g="a"+u;if(0===n.arrayBlockIndices[u])1===t.count?"generic"===e[u]?t.lvalue?(a.push(["var ",f,"=",g,".get(",h,")"].join("")),s=s.replace(p,f),o.push([g,".set(",h,",",f,")"].join(""))):s=s.replace(p,[g,".get(",h,")"].join("")):s=s.replace(p,[g,"[",h,"]"].join("")):"generic"===e[u]?(a.push(["var ",f,"=",g,".get(",h,")"].join("")),s=s.replace(p,f),t.lvalue&&o.push([g,".set(",h,",",f,")"].join(""))):(a.push(["var ",f,"=",g,"[",h,"]"].join("")),s=s.replace(p,f),t.lvalue&&o.push([g,"[",h,"]=",f].join("")));else {for(var j=[t.name],v=[h],y=0;y<Math.abs(n.arrayBlockIndices[u]);y++)j.push("\\s*\\[([^\\]]+)\\]"),v.push("$"+(y+1)+"*t"+u+"b"+y);if(p=new RegExp(j.join(""),"g"),h=v.join("+"),"generic"===e[u])throw new Error("cwise: Generic arrays not supported in combination with blocks!");s=s.replace(p,[g,"[",h,"]"].join(""));}break;case"scalar":s=s.replace(p,"Y"+n.scalarArgs.indexOf(i));break;case"index":s=s.replace(p,"index");break;case"shape":s=s.replace(p,"shape");}}}return [a.join("\n"),s,o.join("\n")].join("\n").trim()}function typeSummary(r){for(var n=new Array(r.length),e=!0,s=0;s<r.length;++s){var a=r[s],o=a.match(/\d+/);o=o?o[0]:"",0===a.charAt(0)?n[s]="u"+a.charAt(1)+o:n[s]=a.charAt(0)+o,s>0&&(e=e&&n[s]===n[s-1]);}return e?n[0]:n.join("")}function generateCWiseOp(r,n){for(var e=n[1].length-Math.abs(r.arrayBlockIndices[0])|0,s=new Array(r.arrayArgs.length),a=new Array(r.arrayArgs.length),o=0;o<r.arrayArgs.length;++o)a[o]=n[2*o],s[o]=n[2*o+1];for(var i=[],t=[],p=[],h=[],u=[],o=0;o<r.arrayArgs.length;++o){r.arrayBlockIndices[o]<0?(p.push(0),h.push(e),i.push(e),t.push(e+r.arrayBlockIndices[o])):(p.push(r.arrayBlockIndices[o]),h.push(r.arrayBlockIndices[o]+e),i.push(0),t.push(r.arrayBlockIndices[o]));for(var c=[],l=0;l<s[o].length;l++)p[o]<=s[o][l]&&s[o][l]<h[o]&&c.push(s[o][l]-p[o]);u.push(c);}for(var f=["SS"],g=["'use strict'"],j=[],l=0;e>l;++l)j.push(["s",l,"=SS[",l,"]"].join(""));for(var o=0;o<r.arrayArgs.length;++o){f.push("a"+o),f.push("t"+o),f.push("p"+o);for(var l=0;e>l;++l)j.push(["t",o,"p",l,"=t",o,"[",p[o]+l,"]"].join(""));for(var l=0;l<Math.abs(r.arrayBlockIndices[o]);++l)j.push(["t",o,"b",l,"=t",o,"[",i[o]+l,"]"].join(""));}for(var o=0;o<r.scalarArgs.length;++o)f.push("Y"+o);if(r.shapeArgs.length>0&&j.push("shape=SS.slice(0)"),r.indexArgs.length>0){for(var v=new Array(e),o=0;e>o;++o)v[o]="0";j.push(["index=[",v.join(","),"]"].join(""));}for(var o=0;o<r.offsetArgs.length;++o){for(var y=r.offsetArgs[o],d=[],l=0;l<y.offset.length;++l)0!==y.offset[l]&&(1===y.offset[l]?d.push(["t",y.array,"p",l].join("")):d.push([y.offset[l],"*t",y.array,"p",l].join("")));0===d.length?j.push("q"+o+"=0"):j.push(["q",o,"=",d.join("+")].join(""));}var A=uniq([].concat(r.pre.thisVars).concat(r.body.thisVars).concat(r.post.thisVars));j=j.concat(A),g.push("var "+j.join(","));for(var o=0;o<r.arrayArgs.length;++o)g.push("p"+o+"|=0");r.pre.body.length>3&&g.push(processBlock(r.pre,r,a));var k=processBlock(r.body,r,a),b=countMatches(u);e>b?g.push(outerFill(b,u[0],r,k)):g.push(innerFill(u[0],r,k)),r.post.body.length>3&&g.push(processBlock(r.post,r,a)),r.debug&&console.log("-----Generated cwise routine for ",n,":\n"+g.join("\n")+"\n----------");var x=[r.funcName||"unnamed","_cwise_loop_",s[0].join("s"),"m",b,typeSummary(a)].join(""),w=new Function(["function ",x,"(",f.join(","),"){",g.join("\n"),"} return ",x].join(""));return w()}var uniq=require("uniq");module.exports=generateCWiseOp;
    	},{"uniq":17}],3:[function(require,module,exports){
    function createThunk(r){var a=["'use strict'","var CACHED={}"],e=[],s=r.funcName+"_cwise_thunk";a.push(["return function ",s,"(",r.shimArgs.join(","),"){"].join(""));for(var n=[],o=[],h=[["array",r.arrayArgs[0],".shape.slice(",Math.max(0,r.arrayBlockIndices[0]),r.arrayBlockIndices[0]<0?","+r.arrayBlockIndices[0]+")":")"].join("")],i=[],p=[],t=0;t<r.arrayArgs.length;++t){var c=r.arrayArgs[t];e.push(["t",c,"=array",c,".dtype,","r",c,"=array",c,".order"].join("")),n.push("t"+c),n.push("r"+c),o.push("t"+c),o.push("r"+c+".join()"),h.push("array"+c+".data"),h.push("array"+c+".stride"),h.push("array"+c+".offset|0"),t>0&&(i.push("array"+r.arrayArgs[0]+".shape.length===array"+c+".shape.length+"+(Math.abs(r.arrayBlockIndices[0])-Math.abs(r.arrayBlockIndices[t]))),p.push("array"+r.arrayArgs[0]+".shape[shapeIndex+"+Math.max(0,r.arrayBlockIndices[0])+"]===array"+c+".shape[shapeIndex+"+Math.max(0,r.arrayBlockIndices[t])+"]"));}r.arrayArgs.length>1&&(a.push("if (!("+i.join(" && ")+")) throw new Error('cwise: Arrays do not all have the same dimensionality!')"),a.push("for(var shapeIndex=array"+r.arrayArgs[0]+".shape.length-"+Math.abs(r.arrayBlockIndices[0])+"; shapeIndex-->0;) {"),a.push("if (!("+p.join(" && ")+")) throw new Error('cwise: Arrays do not all have the same shape!')"),a.push("}"));for(var t=0;t<r.scalarArgs.length;++t)h.push("scalar"+r.scalarArgs[t]);e.push(["type=[",o.join(","),"].join()"].join("")),e.push("proc=CACHED[type]"),a.push("var "+e.join(",")),a.push(["if(!proc){","CACHED[type]=proc=compile([",n.join(","),"])}","return proc(",h.join(","),")}"].join("")),r.debug&&console.log("-----Generated thunk:\n"+a.join("\n")+"\n----------");var u=new Function("compile",a.join("\n"));return u(compile.bind(void 0,r))}var compile=require("./compile.js");module.exports=createThunk;
    	},{"./compile.js":2}],4:[function(require,module,exports){
    function CompiledArgument(e,r,n){this.name=e,this.lvalue=r,this.rvalue=n,this.count=0;}function CompiledRoutine(e,r,n,t){this.body=e,this.args=r,this.thisVars=n,this.localVars=t;}function isGlobal(e){if("eval"===e)throw new Error("cwise-parser: eval() not allowed");return "undefined"!=typeof window?e in window:"undefined"!=typeof GLOBAL?e in GLOBAL:"undefined"!=typeof self?e in self:!1}function getArgNames(e){for(var r=e.body[0].expression.callee.params,n=new Array(r.length),t=0;t<r.length;++t)n[t]=r[t].name;return n}function preprocess(e){function r(e){var r=l+e.replace(/\_/g,"__");return y.push(r),r}function n(e){var r="this_"+e.replace(/\_/g,"__");return d.push(r),r}function t(e,r){for(var n=e.range[0],t=e.range[1],i=n+1;t>i;++i)g[i]="";g[n]=r;}function i(e){return "'"+e.replace(/\_/g,"\\_").replace(/\'/g,"'")+"'"}function o(e){return g.slice(e.range[0],e.range[1]).join("")}function a(e){return "AssignmentExpression"===e.parent.type&&e.parent.left===e?"="===e.parent.operator?m:m|v:"UpdateExpression"===e.parent.type?m|v:v}for(var s=["(",e,")()"].join(""),p=esprima.parse(s,{range:!0}),l="_inline_"+PREFIX_COUNTER++ +"_",u=getArgNames(p),f=new Array(u.length),c=0;c<u.length;++c)f[c]=new CompiledArgument([l,"arg",c,"_"].join(""),!1,!1);for(var g=new Array(s.length),c=0,h=s.length;h>c;++c)g[c]=s.charAt(c);var y=[],d=[],m=1,v=2;!function _(e,o){if(e.parent=o,"MemberExpression"===e.type)e.computed?(_(e.object,e),_(e.property,e)):"ThisExpression"===e.object.type?t(e,n(e.property.name)):_(e.object,e);else {if("ThisExpression"===e.type)throw new Error("cwise-parser: Computed this is not allowed");if("Identifier"===e.type){var s=e.name,p=u.indexOf(s);if(p>=0){var l=f[p],c=a(e);c&m&&(l.lvalue=!0),c&v&&(l.rvalue=!0),++l.count,t(e,l.name);}else isGlobal(s)||t(e,r(s));}else if("Literal"===e.type)"string"==typeof e.value&&t(e,i(e.value));else {if("WithStatement"===e.type)throw new Error("cwise-parser: with() statements not allowed");for(var g=Object.keys(e),h=0,y=g.length;y>h;++h)if("parent"!==g[h]){var d=e[g[h]];if(d)if(d instanceof Array)for(var w=0;w<d.length;++w)d[w]&&"string"==typeof d[w].type&&_(d[w],e);else "string"==typeof d.type&&_(d,e);}}}}(p.body[0].expression.callee.body,void 0),uniq(y),uniq(d);var w=new CompiledRoutine(o(p.body[0].expression.callee.body),f,d,y);return w}var esprima=require("esprima"),uniq=require("uniq"),PREFIX_COUNTER=0;module.exports=preprocess;
    	},{"esprima":8,"uniq":17}],5:[function(require,module,exports){
    function createCWise(e){for(var r in e)REQUIRED_FIELDS.indexOf(r)<0&&OPTIONAL_FIELDS.indexOf(r)<0&&console.warn("cwise: Unknown argument '"+r+"' passed to expression compiler");for(var o=0;o<REQUIRED_FIELDS.length;++o)if(!e[REQUIRED_FIELDS[o]])throw new Error("cwise: Missing argument: "+REQUIRED_FIELDS[o]);return compile({args:e.args,pre:parse(e.pre||function(){}),body:parse(e.body),post:parse(e.post||function(){}),debug:!!e.printCode,funcName:e.funcName||e.body.name||"cwise",blockSize:e.blockSize||64})}var parse=require("cwise-parser"),compile=require("cwise-compiler"),REQUIRED_FIELDS=["args","body"],OPTIONAL_FIELDS=["pre","post","printCode","funcName","blockSize"];module.exports=createCWise;
    	},{"cwise-compiler":1,"cwise-parser":4}],6:[function(require,module,exports){
    	module.exports=require("cwise-compiler");
    	},{"cwise-compiler":1}],7:[function(require,module,exports){
    function dupe_array(e,r,u){var n=0|e[u];if(0>=n)return [];var t,a=new Array(n);if(u===e.length-1)for(t=0;n>t;++t)a[t]=r;else for(t=0;n>t;++t)a[t]=dupe_array(e,r,u+1);return a}function dupe_number(e,r){var u,n;for(u=new Array(e),n=0;e>n;++n)u[n]=r;return u}function dupe(e,r){switch("undefined"==typeof r&&(r=0),typeof e){case"number":if(e>0)return dupe_number(0|e,r);break;case"object":if("number"==typeof e.length)return dupe_array(e,r,0)}return []}module.exports=dupe;
    	},{}],8:[function(require,module,exports){
    	!function(e,t){t("undefined"!=typeof exports?exports:e.esprima={});}(this,function(e){function t(e,t){if(!e)throw new Error("ASSERT: "+t)}function n(e){return e>=48&&57>=e}function r(e){return "0123456789abcdefABCDEF".indexOf(e)>=0}function a(e){return "01234567".indexOf(e)>=0}function i(e){return 32===e||9===e||11===e||12===e||160===e||e>=5760&&[5760,6158,8192,8193,8194,8195,8196,8197,8198,8199,8200,8201,8202,8239,8287,12288,65279].indexOf(e)>=0}function o(e){return 10===e||13===e||8232===e||8233===e}function c(e){return 36===e||95===e||e>=65&&90>=e||e>=97&&122>=e||92===e||e>=128&&it.NonAsciiIdentifierStart.test(String.fromCharCode(e))}function u(e){return 36===e||95===e||e>=65&&90>=e||e>=97&&122>=e||e>=48&&57>=e||92===e||e>=128&&it.NonAsciiIdentifierPart.test(String.fromCharCode(e))}function l(e){switch(e){case"class":case"enum":case"export":case"extends":case"import":case"super":return !0;default:return !1}}function s(e){switch(e){case"implements":case"interface":case"package":case"private":case"protected":case"public":case"static":case"yield":case"let":return !0;default:return !1}}function m(e){return "eval"===e||"arguments"===e}function d(e){if(ut&&s(e))return !0;switch(e.length){case 2:return "if"===e||"in"===e||"do"===e;case 3:return "var"===e||"for"===e||"new"===e||"try"===e||"let"===e;case 4:return "this"===e||"else"===e||"case"===e||"void"===e||"with"===e||"enum"===e;case 5:return "while"===e||"break"===e||"catch"===e||"throw"===e||"const"===e||"yield"===e||"class"===e||"super"===e;case 6:return "return"===e||"typeof"===e||"delete"===e||"switch"===e||"export"===e||"import"===e;case 7:return "default"===e||"finally"===e||"extends"===e;case 8:return "function"===e||"continue"===e||"debugger"===e;case 10:return "instanceof"===e;default:return !1}}function f(e,n,r,a,i){var o;t("number"==typeof r,"Comment must have valid position"),yt.lastCommentStart>=r||(yt.lastCommentStart=r,o={type:e,value:n},ht.range&&(o.range=[r,a]),ht.loc&&(o.loc=i),ht.comments.push(o),ht.attachComment&&(ht.leadingComments.push(o),ht.trailingComments.push(o)));}function p(e){var t,n,r,a;for(t=lt-e,n={start:{line:st,column:lt-mt-e}};dt>lt;)if(r=ct.charCodeAt(lt),++lt,o(r))return ht.comments&&(a=ct.slice(t+e,lt-1),n.end={line:st,column:lt-mt-1},f("Line",a,t,lt-1,n)),13===r&&10===ct.charCodeAt(lt)&&++lt,++st,void(mt=lt);ht.comments&&(a=ct.slice(t+e,lt),n.end={line:st,column:lt-mt},f("Line",a,t,lt,n));}function y(){var e,t,n,r;for(ht.comments&&(e=lt-2,t={start:{line:st,column:lt-mt-2}});dt>lt;)if(n=ct.charCodeAt(lt),o(n))13===n&&10===ct.charCodeAt(lt+1)&&++lt,++st,++lt,mt=lt,lt>=dt&&K({},at.UnexpectedToken,"ILLEGAL");else if(42===n){if(47===ct.charCodeAt(lt+1))return ++lt,++lt,void(ht.comments&&(r=ct.slice(e+2,lt-2),t.end={line:st,column:lt-mt},f("Block",r,e,lt,t)));++lt;}else ++lt;K({},at.UnexpectedToken,"ILLEGAL");}function h(){var e,t;for(t=0===lt;dt>lt;)if(e=ct.charCodeAt(lt),i(e))++lt;else if(o(e))++lt,13===e&&10===ct.charCodeAt(lt)&&++lt,++st,mt=lt,t=!0;else if(47===e)if(e=ct.charCodeAt(lt+1),47===e)++lt,++lt,p(2),t=!0;else {if(42!==e)break;++lt,++lt,y();}else if(t&&45===e){if(45!==ct.charCodeAt(lt+1)||62!==ct.charCodeAt(lt+2))break;lt+=3,p(3);}else {if(60!==e)break;if("!--"!==ct.slice(lt+1,lt+4))break;++lt,++lt,++lt,++lt,p(4);}}function S(e){var t,n,a,i=0;for(n="u"===e?4:2,t=0;n>t;++t){if(!(dt>lt&&r(ct[lt])))return "";a=ct[lt++],i=16*i+"0123456789abcdef".indexOf(a.toLowerCase());}return String.fromCharCode(i)}function g(){var e,t;for(e=ct.charCodeAt(lt++),t=String.fromCharCode(e),92===e&&(117!==ct.charCodeAt(lt)&&K({},at.UnexpectedToken,"ILLEGAL"),++lt,e=S("u"),e&&"\\"!==e&&c(e.charCodeAt(0))||K({},at.UnexpectedToken,"ILLEGAL"),t=e);dt>lt&&(e=ct.charCodeAt(lt),u(e));)++lt,t+=String.fromCharCode(e),92===e&&(t=t.substr(0,t.length-1),117!==ct.charCodeAt(lt)&&K({},at.UnexpectedToken,"ILLEGAL"),++lt,e=S("u"),e&&"\\"!==e&&u(e.charCodeAt(0))||K({},at.UnexpectedToken,"ILLEGAL"),t+=e);return t}function v(){var e,t;for(e=lt++;dt>lt;){if(t=ct.charCodeAt(lt),92===t)return lt=e,g();if(!u(t))break;++lt;}return ct.slice(e,lt)}function k(){var e,t,n;return e=lt,t=92===ct.charCodeAt(lt)?g():v(),n=1===t.length?_e.Identifier:d(t)?_e.Keyword:"null"===t?_e.NullLiteral:"true"===t||"false"===t?_e.BooleanLiteral:_e.Identifier,{type:n,value:t,lineNumber:st,lineStart:mt,start:e,end:lt}}function b(){var e,t,n,r,a=lt,i=ct.charCodeAt(lt),o=ct[lt];switch(i){case 46:case 40:case 41:case 59:case 44:case 123:case 125:case 91:case 93:case 58:case 63:case 126:return ++lt,ht.tokenize&&(40===i?ht.openParenToken=ht.tokens.length:123===i&&(ht.openCurlyToken=ht.tokens.length)),{type:_e.Punctuator,value:String.fromCharCode(i),lineNumber:st,lineStart:mt,start:a,end:lt};default:if(e=ct.charCodeAt(lt+1),61===e)switch(i){case 43:case 45:case 47:case 60:case 62:case 94:case 124:case 37:case 38:case 42:return lt+=2,{type:_e.Punctuator,value:String.fromCharCode(i)+String.fromCharCode(e),lineNumber:st,lineStart:mt,start:a,end:lt};case 33:case 61:return lt+=2,61===ct.charCodeAt(lt)&&++lt,{type:_e.Punctuator,value:ct.slice(a,lt),lineNumber:st,lineStart:mt,start:a,end:lt}}}return r=ct.substr(lt,4),">>>="===r?(lt+=4,{type:_e.Punctuator,value:r,lineNumber:st,lineStart:mt,start:a,end:lt}):(n=r.substr(0,3),">>>"===n||"<<="===n||">>="===n?(lt+=3,{type:_e.Punctuator,value:n,lineNumber:st,lineStart:mt,start:a,end:lt}):(t=n.substr(0,2),o===t[1]&&"+-<>&|".indexOf(o)>=0||"=>"===t?(lt+=2,{type:_e.Punctuator,value:t,lineNumber:st,lineStart:mt,start:a,end:lt}):"<>=!+-*%&|^/".indexOf(o)>=0?(++lt,{type:_e.Punctuator,value:o,lineNumber:st,lineStart:mt,start:a,end:lt}):void K({},at.UnexpectedToken,"ILLEGAL")))}function E(e){for(var t="";dt>lt&&r(ct[lt]);)t+=ct[lt++];return 0===t.length&&K({},at.UnexpectedToken,"ILLEGAL"),c(ct.charCodeAt(lt))&&K({},at.UnexpectedToken,"ILLEGAL"),{type:_e.NumericLiteral,value:parseInt("0x"+t,16),lineNumber:st,lineStart:mt,start:e,end:lt}}function x(e){for(var t="0"+ct[lt++];dt>lt&&a(ct[lt]);)t+=ct[lt++];return (c(ct.charCodeAt(lt))||n(ct.charCodeAt(lt)))&&K({},at.UnexpectedToken,"ILLEGAL"),{type:_e.NumericLiteral,value:parseInt(t,8),octal:!0,lineNumber:st,lineStart:mt,start:e,end:lt}}function C(){var e,t;for(e=lt+1;dt>e;++e){if(t=ct[e],"8"===t||"9"===t)return !1;if(!a(t))return !0}return !0}function w(){var e,r,i;if(i=ct[lt],t(n(i.charCodeAt(0))||"."===i,"Numeric literal must start with a decimal digit or a decimal point"),r=lt,e="","."!==i){if(e=ct[lt++],i=ct[lt],"0"===e){if("x"===i||"X"===i)return ++lt,E(r);if(a(i)&&C())return x(r)}for(;n(ct.charCodeAt(lt));)e+=ct[lt++];i=ct[lt];}if("."===i){for(e+=ct[lt++];n(ct.charCodeAt(lt));)e+=ct[lt++];i=ct[lt];}if("e"===i||"E"===i)if(e+=ct[lt++],i=ct[lt],("+"===i||"-"===i)&&(e+=ct[lt++]),n(ct.charCodeAt(lt)))for(;n(ct.charCodeAt(lt));)e+=ct[lt++];else K({},at.UnexpectedToken,"ILLEGAL");return c(ct.charCodeAt(lt))&&K({},at.UnexpectedToken,"ILLEGAL"),{type:_e.NumericLiteral,value:parseFloat(e),lineNumber:st,lineStart:mt,start:r,end:lt}}function I(){var e,n,r,i,c,u,l,s,m="",d=!1;for(l=st,s=mt,e=ct[lt],t("'"===e||'"'===e,"String literal must starts with a quote"),n=lt,++lt;dt>lt;){if(r=ct[lt++],r===e){e="";break}if("\\"===r)if(r=ct[lt++],r&&o(r.charCodeAt(0)))++st,"\r"===r&&"\n"===ct[lt]&&++lt,mt=lt;else switch(r){case"u":case"x":u=lt,c=S(r),c?m+=c:(lt=u,m+=r);break;case"n":m+="\n";break;case"r":m+="\r";break;case"t":m+="	";break;case"b":m+="\b";break;case"f":m+="\f";break;case"v":m+="\x0B";break;default:a(r)?(i="01234567".indexOf(r),0!==i&&(d=!0),dt>lt&&a(ct[lt])&&(d=!0,i=8*i+"01234567".indexOf(ct[lt++]),"0123".indexOf(r)>=0&&dt>lt&&a(ct[lt])&&(i=8*i+"01234567".indexOf(ct[lt++]))),m+=String.fromCharCode(i)):m+=r;}else {if(o(r.charCodeAt(0)))break;m+=r;}}return ""!==e&&K({},at.UnexpectedToken,"ILLEGAL"),{type:_e.StringLiteral,value:m,octal:d,startLineNumber:l,startLineStart:s,lineNumber:st,lineStart:mt,start:n,end:lt}}function L(e,t){var n;try{n=new RegExp(e,t);}catch(r){K({},at.InvalidRegExp);}return n}function A(){var e,n,r,a,i;for(e=ct[lt],t("/"===e,"Regular expression literal must start with a slash"),n=ct[lt++],r=!1,a=!1;dt>lt;)if(e=ct[lt++],n+=e,"\\"===e)e=ct[lt++],o(e.charCodeAt(0))&&K({},at.UnterminatedRegExp),n+=e;else if(o(e.charCodeAt(0)))K({},at.UnterminatedRegExp);else if(r)"]"===e&&(r=!1);else {if("/"===e){a=!0;break}"["===e&&(r=!0);}return a||K({},at.UnterminatedRegExp),i=n.substr(1,n.length-2),{value:i,literal:n}}function P(){var e,t,n,r;for(t="",n="";dt>lt&&(e=ct[lt],u(e.charCodeAt(0)));)if(++lt,"\\"===e&&dt>lt)if(e=ct[lt],"u"===e){if(++lt,r=lt,e=S("u"))for(n+=e,t+="\\u";lt>r;++r)t+=ct[r];else lt=r,n+="u",t+="\\u";V({},at.UnexpectedToken,"ILLEGAL");}else t+="\\",V({},at.UnexpectedToken,"ILLEGAL");else n+=e,t+=e;return {value:n,literal:t}}function N(){var e,t,n,r;return pt=null,h(),e=lt,t=A(),n=P(),r=L(t.value,n.value),ht.tokenize?{type:_e.RegularExpression,value:r,lineNumber:st,lineStart:mt,start:e,end:lt}:{literal:t.literal+n.literal,value:r,start:e,end:lt}}function O(){var e,t,n,r;return h(),e=lt,t={start:{line:st,column:lt-mt}},n=N(),t.end={line:st,column:lt-mt},ht.tokenize||(ht.tokens.length>0&&(r=ht.tokens[ht.tokens.length-1],r.range[0]===e&&"Punctuator"===r.type&&("/"===r.value||"/="===r.value)&&ht.tokens.pop()),ht.tokens.push({type:"RegularExpression",value:n.literal,range:[e,lt],loc:t})),n}function U(e){return e.type===_e.Identifier||e.type===_e.Keyword||e.type===_e.BooleanLiteral||e.type===_e.NullLiteral}function R(){var e,t;if(e=ht.tokens[ht.tokens.length-1],!e)return O();if("Punctuator"===e.type){if("]"===e.value)return b();if(")"===e.value)return t=ht.tokens[ht.openParenToken-1],!t||"Keyword"!==t.type||"if"!==t.value&&"while"!==t.value&&"for"!==t.value&&"with"!==t.value?b():O();if("}"===e.value){if(ht.tokens[ht.openCurlyToken-3]&&"Keyword"===ht.tokens[ht.openCurlyToken-3].type){if(t=ht.tokens[ht.openCurlyToken-4],!t)return b()}else {if(!ht.tokens[ht.openCurlyToken-4]||"Keyword"!==ht.tokens[ht.openCurlyToken-4].type)return b();if(t=ht.tokens[ht.openCurlyToken-5],!t)return O()}return tt.indexOf(t.value)>=0?b():O()}return O()}return "Keyword"===e.type&&"this"!==e.value?O():b()}function F(){var e;return h(),lt>=dt?{type:_e.EOF,lineNumber:st,lineStart:mt,start:lt,end:lt}:(e=ct.charCodeAt(lt),c(e)?k():40===e||41===e||59===e?b():39===e||34===e?I():46===e?n(ct.charCodeAt(lt+1))?w():b():n(e)?w():ht.tokenize&&47===e?R():b())}function T(){var e,t,n;return h(),e={start:{line:st,column:lt-mt}},t=F(),e.end={line:st,column:lt-mt},t.type!==_e.EOF&&(n=ct.slice(t.start,t.end),ht.tokens.push({type:et[t.type],value:n,range:[t.start,t.end],loc:e})),t}function D(){var e;return e=pt,lt=e.end,st=e.lineNumber,mt=e.lineStart,pt="undefined"!=typeof ht.tokens?T():F(),lt=e.end,st=e.lineNumber,mt=e.lineStart,e}function B(){var e,t,n;e=lt,t=st,n=mt,pt="undefined"!=typeof ht.tokens?T():F(),lt=e,st=t,mt=n;}function W(e,t){this.line=e,this.column=t;}function j(e,t,n,r){this.start=new W(e,t),this.end=new W(n,r);}function G(){var e,t,n,r;return e=lt,t=st,n=mt,h(),r=st!==t,lt=e,st=t,mt=n,r}function K(e,n){var r,a=Array.prototype.slice.call(arguments,2),i=n.replace(/%(\d)/g,function(e,n){return t(n<a.length,"Message reference must be in range"),a[n]});throw "number"==typeof e.lineNumber?(r=new Error("Line "+e.lineNumber+": "+i),r.index=e.start,r.lineNumber=e.lineNumber,r.column=e.start-mt+1):(r=new Error("Line "+st+": "+i),r.index=lt,r.lineNumber=st,r.column=lt-mt+1),r.description=i,r}function V(){try{K.apply(null,arguments);}catch(e){if(!ht.errors)throw e;ht.errors.push(e);}}function M(e){if(e.type===_e.EOF&&K(e,at.UnexpectedEOS),e.type===_e.NumericLiteral&&K(e,at.UnexpectedNumber),e.type===_e.StringLiteral&&K(e,at.UnexpectedString),e.type===_e.Identifier&&K(e,at.UnexpectedIdentifier),e.type===_e.Keyword){if(l(e.value))K(e,at.UnexpectedReserved);else if(ut&&s(e.value))return void V(e,at.StrictReservedWord);K(e,at.UnexpectedToken,e.value);}K(e,at.UnexpectedToken,e.value);}function H(e){var t=D();(t.type!==_e.Punctuator||t.value!==e)&&M(t);}function q(e){var t=D();(t.type!==_e.Keyword||t.value!==e)&&M(t);}function z(e){return pt.type===_e.Punctuator&&pt.value===e}function $(e){return pt.type===_e.Keyword&&pt.value===e}function X(){var e;return pt.type!==_e.Punctuator?!1:(e=pt.value,"="===e||"*="===e||"/="===e||"%="===e||"+="===e||"-="===e||"<<="===e||">>="===e||">>>="===e||"&="===e||"^="===e||"|="===e)}function J(){var e,t=lt,n=st,r=mt,a=pt;return 59===ct.charCodeAt(lt)||z(";")?void D():(e=st,h(),st!==e?(lt=t,st=n,mt=r,void(pt=a)):void(pt.type===_e.EOF||z("}")||M(pt)))}function Q(e){return e.type===nt.Identifier||e.type===nt.MemberExpression}function Y(){var e,t=[];for(e=pt,H("[");!z("]");)z(",")?(D(),t.push(null)):(t.push(he()),z("]")||H(","));return D(),ft.markEnd(ft.createArrayExpression(t),e)}function Z(e,t){var n,r,a;return n=ut,a=pt,r=Me(),t&&ut&&m(e[0].name)&&V(t,at.StrictParamName),ut=n,ft.markEnd(ft.createFunctionExpression(null,e,[],r),a)}function _(){var e,t;return t=pt,e=D(),e.type===_e.StringLiteral||e.type===_e.NumericLiteral?(ut&&e.octal&&V(e,at.StrictOctalLiteral),ft.markEnd(ft.createLiteral(e),t)):ft.markEnd(ft.createIdentifier(e.value),t)}function ee(){var e,t,n,r,a,i;return e=pt,i=pt,e.type===_e.Identifier?(n=_(),"get"!==e.value||z(":")?"set"!==e.value||z(":")?(H(":"),r=he(),ft.markEnd(ft.createProperty("init",n,r),i)):(t=_(),H("("),e=pt,e.type!==_e.Identifier?(H(")"),V(e,at.UnexpectedToken,e.value),r=Z([])):(a=[ke()],H(")"),r=Z(a,e)),ft.markEnd(ft.createProperty("set",t,r),i)):(t=_(),H("("),H(")"),r=Z([]),ft.markEnd(ft.createProperty("get",t,r),i))):e.type!==_e.EOF&&e.type!==_e.Punctuator?(t=_(),H(":"),r=he(),ft.markEnd(ft.createProperty("init",t,r),i)):void M(e)}function te(){var e,t,n,r,a,i=[],o={},c=String;for(a=pt,H("{");!z("}");)e=ee(),t=e.key.type===nt.Identifier?e.key.name:c(e.key.value),r="init"===e.kind?rt.Data:"get"===e.kind?rt.Get:rt.Set,n="$"+t,Object.prototype.hasOwnProperty.call(o,n)?(o[n]===rt.Data?ut&&r===rt.Data?V({},at.StrictDuplicateProperty):r!==rt.Data&&V({},at.AccessorDataProperty):r===rt.Data?V({},at.AccessorDataProperty):o[n]&r&&V({},at.AccessorGetSet),o[n]|=r):o[n]=r,i.push(e),z("}")||H(",");return H("}"),ft.markEnd(ft.createObjectExpression(i),a)}function ne(){var e;return H("("),e=Se(),H(")"),e}function re(){var e,t,n,r;if(z("("))return ne();if(z("["))return Y();if(z("{"))return te();if(e=pt.type,r=pt,e===_e.Identifier)n=ft.createIdentifier(D().value);else if(e===_e.StringLiteral||e===_e.NumericLiteral)ut&&pt.octal&&V(pt,at.StrictOctalLiteral),n=ft.createLiteral(D());else if(e===_e.Keyword){if($("function"))return ze();$("this")?(D(),n=ft.createThisExpression()):M(D());}else e===_e.BooleanLiteral?(t=D(),t.value="true"===t.value,n=ft.createLiteral(t)):e===_e.NullLiteral?(t=D(),t.value=null,n=ft.createLiteral(t)):z("/")||z("/=")?(n="undefined"!=typeof ht.tokens?ft.createLiteral(O()):ft.createLiteral(N()),B()):M(D());return ft.markEnd(n,r)}function ae(){var e=[];if(H("("),!z(")"))for(;dt>lt&&(e.push(he()),!z(")"));)H(",");return H(")"),e}function ie(){var e,t;return t=pt,e=D(),U(e)||M(e),ft.markEnd(ft.createIdentifier(e.value),t)}function oe(){return H("."),ie()}function ce(){var e;return H("["),e=Se(),H("]"),e}function ue(){var e,t,n;return n=pt,q("new"),e=se(),t=z("(")?ae():[],ft.markEnd(ft.createNewExpression(e,t),n)}function le(){var e,t,n,r,a=yt.allowIn;for(r=pt,yt.allowIn=!0,e=$("new")?ue():re();;){if(z("."))n=oe(),e=ft.createMemberExpression(".",e,n);else if(z("("))t=ae(),e=ft.createCallExpression(e,t);else {if(!z("["))break;n=ce(),e=ft.createMemberExpression("[",e,n);}ft.markEnd(e,r);}return yt.allowIn=a,e}function se(){var e,n,r;for(t(yt.allowIn,"callee of new expression always allow in keyword."),r=pt,e=$("new")?ue():re();z(".")||z("[");)z("[")?(n=ce(),e=ft.createMemberExpression("[",e,n)):(n=oe(),e=ft.createMemberExpression(".",e,n)),ft.markEnd(e,r);return e}function me(){var e,t,n=pt;return e=le(),pt.type===_e.Punctuator&&(!z("++")&&!z("--")||G()||(ut&&e.type===nt.Identifier&&m(e.name)&&V({},at.StrictLHSPostfix),Q(e)||V({},at.InvalidLHSInAssignment),t=D(),e=ft.markEnd(ft.createPostfixExpression(t.value,e),n))),e}function de(){var e,t,n;return pt.type!==_e.Punctuator&&pt.type!==_e.Keyword?t=me():z("++")||z("--")?(n=pt,e=D(),t=de(),ut&&t.type===nt.Identifier&&m(t.name)&&V({},at.StrictLHSPrefix),Q(t)||V({},at.InvalidLHSInAssignment),t=ft.createUnaryExpression(e.value,t),t=ft.markEnd(t,n)):z("+")||z("-")||z("~")||z("!")?(n=pt,e=D(),t=de(),t=ft.createUnaryExpression(e.value,t),t=ft.markEnd(t,n)):$("delete")||$("void")||$("typeof")?(n=pt,e=D(),t=de(),t=ft.createUnaryExpression(e.value,t),t=ft.markEnd(t,n),ut&&"delete"===t.operator&&t.argument.type===nt.Identifier&&V({},at.StrictDelete)):t=me(),t}function fe(e,t){var n=0;if(e.type!==_e.Punctuator&&e.type!==_e.Keyword)return 0;switch(e.value){case"||":n=1;break;case"&&":n=2;break;case"|":n=3;break;case"^":n=4;break;case"&":n=5;break;case"==":case"!=":case"===":case"!==":n=6;break;case"<":case">":case"<=":case">=":case"instanceof":n=7;break;case"in":n=t?7:0;break;case"<<":case">>":case">>>":n=8;break;case"+":case"-":n=9;break;case"*":case"/":case"%":n=11;}return n}function pe(){var e,t,n,r,a,i,o,c,u,l;if(e=pt,u=de(),r=pt,a=fe(r,yt.allowIn),0===a)return u;for(r.prec=a,D(),t=[e,pt],o=de(),i=[u,r,o];(a=fe(pt,yt.allowIn))>0;){for(;i.length>2&&a<=i[i.length-2].prec;)o=i.pop(),c=i.pop().value,u=i.pop(),n=ft.createBinaryExpression(c,u,o),t.pop(),e=t[t.length-1],ft.markEnd(n,e),i.push(n);r=D(),r.prec=a,i.push(r),t.push(pt),n=de(),i.push(n);}for(l=i.length-1,n=i[l],t.pop();l>1;)n=ft.createBinaryExpression(i[l-1].value,i[l-2],n),l-=2,e=t.pop(),ft.markEnd(n,e);return n}function ye(){var e,t,n,r,a;return a=pt,e=pe(),z("?")&&(D(),t=yt.allowIn,yt.allowIn=!0,n=he(),yt.allowIn=t,H(":"),r=he(),e=ft.createConditionalExpression(e,n,r),ft.markEnd(e,a)),e}function he(){var e,t,n,r,a;return e=pt,a=pt,r=t=ye(),X()&&(Q(t)||V({},at.InvalidLHSInAssignment),ut&&t.type===nt.Identifier&&m(t.name)&&V(e,at.StrictLHSAssignment),e=D(),n=he(),r=ft.markEnd(ft.createAssignmentExpression(e.value,t,n),a)),r}function Se(){var e,t=pt;if(e=he(),z(",")){for(e=ft.createSequenceExpression([e]);dt>lt&&z(",");)D(),e.expressions.push(he());ft.markEnd(e,t);}return e}function ge(){for(var e,t=[];dt>lt&&!z("}")&&(e=$e(),"undefined"!=typeof e);)t.push(e);return t}function ve(){var e,t;return t=pt,H("{"),e=ge(),H("}"),ft.markEnd(ft.createBlockStatement(e),t)}function ke(){var e,t;return t=pt,e=D(),e.type!==_e.Identifier&&M(e),ft.markEnd(ft.createIdentifier(e.value),t)}function be(e){var t,n,r=null;return n=pt,t=ke(),ut&&m(t.name)&&V({},at.StrictVarName),"const"===e?(H("="),r=he()):z("=")&&(D(),r=he()),ft.markEnd(ft.createVariableDeclarator(t,r),n)}function Ee(e){var t=[];do{if(t.push(be(e)),!z(","))break;D();}while(dt>lt);return t}function xe(){var e;return q("var"),e=Ee(),J(),ft.createVariableDeclaration(e,"var")}function Ce(e){var t,n;return n=pt,q(e),t=Ee(e),J(),ft.markEnd(ft.createVariableDeclaration(t,e),n)}function we(){return H(";"),ft.createEmptyStatement()}function Ie(){var e=Se();return J(),ft.createExpressionStatement(e)}function Le(){var e,t,n;return q("if"),H("("),e=Se(),H(")"),t=Ve(),$("else")?(D(),n=Ve()):n=null,ft.createIfStatement(e,t,n)}function Ae(){var e,t,n;return q("do"),n=yt.inIteration,yt.inIteration=!0,e=Ve(),yt.inIteration=n,q("while"),H("("),t=Se(),H(")"),z(";")&&D(),ft.createDoWhileStatement(e,t)}function Pe(){var e,t,n;return q("while"),H("("),e=Se(),H(")"),n=yt.inIteration,yt.inIteration=!0,t=Ve(),yt.inIteration=n,ft.createWhileStatement(e,t)}function Ne(){var e,t,n;return n=pt,e=D(),t=Ee(),ft.markEnd(ft.createVariableDeclaration(t,e.value),n)}function Oe(){var e,t,n,r,a,i,o,c=yt.allowIn;return e=t=n=null,q("for"),H("("),z(";")?D():($("var")||$("let")?(yt.allowIn=!1,e=Ne(),yt.allowIn=c,1===e.declarations.length&&$("in")&&(D(),r=e,a=Se(),e=null)):(yt.allowIn=!1,e=Se(),yt.allowIn=c,$("in")&&(Q(e)||V({},at.InvalidLHSInForIn),D(),r=e,a=Se(),e=null)),"undefined"==typeof r&&H(";")),"undefined"==typeof r&&(z(";")||(t=Se()),H(";"),z(")")||(n=Se())),H(")"),o=yt.inIteration,yt.inIteration=!0,i=Ve(),yt.inIteration=o,"undefined"==typeof r?ft.createForStatement(e,t,n,i):ft.createForInStatement(r,a,i)}function Ue(){var e,t=null;return q("continue"),59===ct.charCodeAt(lt)?(D(),yt.inIteration||K({},at.IllegalContinue),ft.createContinueStatement(null)):G()?(yt.inIteration||K({},at.IllegalContinue),ft.createContinueStatement(null)):(pt.type===_e.Identifier&&(t=ke(),e="$"+t.name,Object.prototype.hasOwnProperty.call(yt.labelSet,e)||K({},at.UnknownLabel,t.name)),J(),null!==t||yt.inIteration||K({},at.IllegalContinue),ft.createContinueStatement(t))}function Re(){var e,t=null;return q("break"),59===ct.charCodeAt(lt)?(D(),yt.inIteration||yt.inSwitch||K({},at.IllegalBreak),ft.createBreakStatement(null)):G()?(yt.inIteration||yt.inSwitch||K({},at.IllegalBreak),ft.createBreakStatement(null)):(pt.type===_e.Identifier&&(t=ke(),e="$"+t.name,Object.prototype.hasOwnProperty.call(yt.labelSet,e)||K({},at.UnknownLabel,t.name)),J(),null!==t||yt.inIteration||yt.inSwitch||K({},at.IllegalBreak),ft.createBreakStatement(t))}function Fe(){var e=null;return q("return"),yt.inFunctionBody||V({},at.IllegalReturn),32===ct.charCodeAt(lt)&&c(ct.charCodeAt(lt+1))?(e=Se(),J(),ft.createReturnStatement(e)):G()?ft.createReturnStatement(null):(z(";")||z("}")||pt.type===_e.EOF||(e=Se()),J(),ft.createReturnStatement(e))}function Te(){var e,t;return ut&&(h(),V({},at.StrictModeWith)),q("with"),H("("),e=Se(),H(")"),t=Ve(),ft.createWithStatement(e,t)}function De(){var e,t,n,r=[];for(n=pt,$("default")?(D(),e=null):(q("case"),e=Se()),H(":");dt>lt&&!(z("}")||$("default")||$("case"));)t=Ve(),r.push(t);return ft.markEnd(ft.createSwitchCase(e,r),n)}function Be(){var e,t,n,r,a;if(q("switch"),H("("),e=Se(),H(")"),H("{"),t=[],z("}"))return D(),ft.createSwitchStatement(e,t);for(r=yt.inSwitch,yt.inSwitch=!0,a=!1;dt>lt&&!z("}");)n=De(),null===n.test&&(a&&K({},at.MultipleDefaultsInSwitch),a=!0),t.push(n);return yt.inSwitch=r,H("}"),ft.createSwitchStatement(e,t)}function We(){var e;return q("throw"),G()&&K({},at.NewlineAfterThrow),e=Se(),J(),ft.createThrowStatement(e)}function je(){var e,t,n;return n=pt,q("catch"),H("("),z(")")&&M(pt),e=ke(),ut&&m(e.name)&&V({},at.StrictCatchVariable),H(")"),t=ve(),ft.markEnd(ft.createCatchClause(e,t),n)}function Ge(){var e,t=[],n=null;return q("try"),e=ve(),$("catch")&&t.push(je()),$("finally")&&(D(),n=ve()),0!==t.length||n||K({},at.NoCatchOrFinally),ft.createTryStatement(e,[],t,n)}function Ke(){return q("debugger"),J(),ft.createDebuggerStatement()}function Ve(){var e,t,n,r,a=pt.type;if(a===_e.EOF&&M(pt),a===_e.Punctuator&&"{"===pt.value)return ve();if(r=pt,a===_e.Punctuator)switch(pt.value){case";":return ft.markEnd(we(),r);case"(":return ft.markEnd(Ie(),r)}if(a===_e.Keyword)switch(pt.value){case"break":return ft.markEnd(Re(),r);case"continue":return ft.markEnd(Ue(),r);case"debugger":return ft.markEnd(Ke(),r);case"do":return ft.markEnd(Ae(),r);case"for":return ft.markEnd(Oe(),r);case"function":return ft.markEnd(qe(),r);case"if":return ft.markEnd(Le(),r);case"return":return ft.markEnd(Fe(),r);case"switch":return ft.markEnd(Be(),r);case"throw":return ft.markEnd(We(),r);case"try":return ft.markEnd(Ge(),r);case"var":return ft.markEnd(xe(),r);case"while":return ft.markEnd(Pe(),r);case"with":return ft.markEnd(Te(),r)}return e=Se(),e.type===nt.Identifier&&z(":")?(D(),n="$"+e.name,Object.prototype.hasOwnProperty.call(yt.labelSet,n)&&K({},at.Redeclaration,"Label",e.name),yt.labelSet[n]=!0,t=Ve(),delete yt.labelSet[n],ft.markEnd(ft.createLabeledStatement(e,t),r)):(J(),ft.markEnd(ft.createExpressionStatement(e),r))}function Me(){var e,t,n,r,a,i,o,c,u,l=[];for(u=pt,H("{");dt>lt&&pt.type===_e.StringLiteral&&(t=pt,e=$e(),l.push(e),e.expression.type===nt.Literal);)n=ct.slice(t.start+1,t.end-1),"use strict"===n?(ut=!0,r&&V(r,at.StrictOctalLiteral)):!r&&t.octal&&(r=t);for(a=yt.labelSet,i=yt.inIteration,o=yt.inSwitch,c=yt.inFunctionBody,yt.labelSet={},yt.inIteration=!1,yt.inSwitch=!1,yt.inFunctionBody=!0;dt>lt&&!z("}")&&(e=$e(),"undefined"!=typeof e);)l.push(e);return H("}"),yt.labelSet=a,yt.inIteration=i,yt.inSwitch=o,yt.inFunctionBody=c,ft.markEnd(ft.createBlockStatement(l),u)}function He(e){var t,n,r,a,i,o,c=[];if(H("("),!z(")"))for(a={};dt>lt&&(n=pt,t=ke(),i="$"+n.value,ut?(m(n.value)&&(r=n,o=at.StrictParamName),Object.prototype.hasOwnProperty.call(a,i)&&(r=n,o=at.StrictParamDupe)):e||(m(n.value)?(e=n,o=at.StrictParamName):s(n.value)?(e=n,o=at.StrictReservedWord):Object.prototype.hasOwnProperty.call(a,i)&&(e=n,o=at.StrictParamDupe)),c.push(t),a[i]=!0,!z(")"));)H(",");return H(")"),{params:c,stricted:r,firstRestricted:e,message:o}}function qe(){var e,t,n,r,a,i,o,c,u,l=[];return u=pt,q("function"),n=pt,e=ke(),ut?m(n.value)&&V(n,at.StrictFunctionName):m(n.value)?(i=n,o=at.StrictFunctionName):s(n.value)&&(i=n,o=at.StrictReservedWord),a=He(i),l=a.params,r=a.stricted,i=a.firstRestricted,a.message&&(o=a.message),c=ut,t=Me(),ut&&i&&K(i,o),ut&&r&&V(r,o),ut=c,ft.markEnd(ft.createFunctionDeclaration(e,l,[],t),u)}function ze(){var e,t,n,r,a,i,o,c,u=null,l=[];return c=pt,q("function"),z("(")||(e=pt,u=ke(),ut?m(e.value)&&V(e,at.StrictFunctionName):m(e.value)?(n=e,r=at.StrictFunctionName):s(e.value)&&(n=e,r=at.StrictReservedWord)),a=He(n),l=a.params,t=a.stricted,n=a.firstRestricted,a.message&&(r=a.message),o=ut,i=Me(),ut&&n&&K(n,r),ut&&t&&V(t,r),ut=o,ft.markEnd(ft.createFunctionExpression(u,l,[],i),c)}function $e(){if(pt.type===_e.Keyword)switch(pt.value){case"const":case"let":return Ce(pt.value);case"function":return qe();default:return Ve()}return pt.type!==_e.EOF?Ve():void 0}function Xe(){for(var e,t,n,r,a=[];dt>lt&&(t=pt,t.type===_e.StringLiteral)&&(e=$e(),a.push(e),e.expression.type===nt.Literal);)n=ct.slice(t.start+1,t.end-1),"use strict"===n?(ut=!0,r&&V(r,at.StrictOctalLiteral)):!r&&t.octal&&(r=t);for(;dt>lt&&(e=$e(),"undefined"!=typeof e);)a.push(e);return a}function Je(){var e,t;return h(),B(),t=pt,ut=!1,e=Xe(),ft.markEnd(ft.createProgram(e),t)}function Qe(){var e,t,n,r=[];for(e=0;e<ht.tokens.length;++e)t=ht.tokens[e],n={type:t.type,value:t.value},ht.range&&(n.range=t.range),ht.loc&&(n.loc=t.loc),r.push(n);ht.tokens=r;}function Ye(e,t){var n,r,a;n=String,"string"==typeof e||e instanceof String||(e=n(e)),ft=ot,ct=e,lt=0,st=ct.length>0?1:0,mt=0,dt=ct.length,pt=null,yt={allowIn:!0,labelSet:{},inFunctionBody:!1,inIteration:!1,inSwitch:!1,lastCommentStart:-1},ht={},t=t||{},t.tokens=!0,ht.tokens=[],ht.tokenize=!0,ht.openParenToken=-1,ht.openCurlyToken=-1,ht.range="boolean"==typeof t.range&&t.range,ht.loc="boolean"==typeof t.loc&&t.loc,"boolean"==typeof t.comment&&t.comment&&(ht.comments=[]),"boolean"==typeof t.tolerant&&t.tolerant&&(ht.errors=[]);try{if(B(),pt.type===_e.EOF)return ht.tokens;for(r=D();pt.type!==_e.EOF;)try{r=D();}catch(i){if(r=pt,ht.errors){ht.errors.push(i);break}throw i}Qe(),a=ht.tokens,"undefined"!=typeof ht.comments&&(a.comments=ht.comments),"undefined"!=typeof ht.errors&&(a.errors=ht.errors);}catch(o){throw o}finally{ht={};}return a}function Ze(e,t){var n,r;r=String,"string"==typeof e||e instanceof String||(e=r(e)),ft=ot,ct=e,lt=0,st=ct.length>0?1:0,mt=0,dt=ct.length,pt=null,yt={allowIn:!0,labelSet:{},inFunctionBody:!1,inIteration:!1,inSwitch:!1,lastCommentStart:-1},ht={},"undefined"!=typeof t&&(ht.range="boolean"==typeof t.range&&t.range,ht.loc="boolean"==typeof t.loc&&t.loc,ht.attachComment="boolean"==typeof t.attachComment&&t.attachComment,ht.loc&&null!==t.source&&void 0!==t.source&&(ht.source=r(t.source)),"boolean"==typeof t.tokens&&t.tokens&&(ht.tokens=[]),"boolean"==typeof t.comment&&t.comment&&(ht.comments=[]),"boolean"==typeof t.tolerant&&t.tolerant&&(ht.errors=[]),ht.attachComment&&(ht.range=!0,ht.comments=[],ht.bottomRightStack=[],ht.trailingComments=[],ht.leadingComments=[]));try{n=Je(),"undefined"!=typeof ht.comments&&(n.comments=ht.comments),"undefined"!=typeof ht.tokens&&(Qe(),n.tokens=ht.tokens),"undefined"!=typeof ht.errors&&(n.errors=ht.errors);}catch(a){throw a}finally{ht={};}return n}var _e,et,tt,nt,rt,at,it,ot,ct,ut,lt,st,mt,dt,ft,pt,yt,ht;_e={BooleanLiteral:1,EOF:2,Identifier:3,Keyword:4,NullLiteral:5,NumericLiteral:6,Punctuator:7,StringLiteral:8,RegularExpression:9},et={},et[_e.BooleanLiteral]="Boolean",et[_e.EOF]="<end>",et[_e.Identifier]="Identifier",et[_e.Keyword]="Keyword",et[_e.NullLiteral]="Null",et[_e.NumericLiteral]="Numeric",et[_e.Punctuator]="Punctuator",et[_e.StringLiteral]="String",et[_e.RegularExpression]="RegularExpression",tt=["(","{","[","in","typeof","instanceof","new","return","case","delete","throw","void","=","+=","-=","*=","/=","%=","<<=",">>=",">>>=","&=","|=","^=",",","+","-","*","/","%","++","--","<<",">>",">>>","&","|","^","!","~","&&","||","?",":","===","==",">=","<=","<",">","!=","!=="],nt={AssignmentExpression:"AssignmentExpression",ArrayExpression:"ArrayExpression",BlockStatement:"BlockStatement",BinaryExpression:"BinaryExpression",BreakStatement:"BreakStatement",CallExpression:"CallExpression",CatchClause:"CatchClause",ConditionalExpression:"ConditionalExpression",ContinueStatement:"ContinueStatement",DoWhileStatement:"DoWhileStatement",DebuggerStatement:"DebuggerStatement",EmptyStatement:"EmptyStatement",ExpressionStatement:"ExpressionStatement",ForStatement:"ForStatement",ForInStatement:"ForInStatement",FunctionDeclaration:"FunctionDeclaration",FunctionExpression:"FunctionExpression",Identifier:"Identifier",IfStatement:"IfStatement",Literal:"Literal",LabeledStatement:"LabeledStatement",LogicalExpression:"LogicalExpression",MemberExpression:"MemberExpression",NewExpression:"NewExpression",ObjectExpression:"ObjectExpression",Program:"Program",Property:"Property",ReturnStatement:"ReturnStatement",SequenceExpression:"SequenceExpression",SwitchStatement:"SwitchStatement",SwitchCase:"SwitchCase",ThisExpression:"ThisExpression",ThrowStatement:"ThrowStatement",TryStatement:"TryStatement",UnaryExpression:"UnaryExpression",UpdateExpression:"UpdateExpression",VariableDeclaration:"VariableDeclaration",VariableDeclarator:"VariableDeclarator",WhileStatement:"WhileStatement",WithStatement:"WithStatement"},rt={Data:1,Get:2,Set:4},at={UnexpectedToken:"Unexpected token %0",UnexpectedNumber:"Unexpected number",UnexpectedString:"Unexpected string",UnexpectedIdentifier:"Unexpected identifier",UnexpectedReserved:"Unexpected reserved word",UnexpectedEOS:"Unexpected end of input",NewlineAfterThrow:"Illegal newline after throw",InvalidRegExp:"Invalid regular expression",UnterminatedRegExp:"Invalid regular expression: missing /",InvalidLHSInAssignment:"Invalid left-hand side in assignment",InvalidLHSInForIn:"Invalid left-hand side in for-in",MultipleDefaultsInSwitch:"More than one default clause in switch statement",NoCatchOrFinally:"Missing catch or finally after try",UnknownLabel:"Undefined label '%0'",Redeclaration:"%0 '%1' has already been declared",IllegalContinue:"Illegal continue statement",IllegalBreak:"Illegal break statement",IllegalReturn:"Illegal return statement",StrictModeWith:"Strict mode code may not include a with statement",StrictCatchVariable:"Catch variable may not be eval or arguments in strict mode",StrictVarName:"Variable name may not be eval or arguments in strict mode",StrictParamName:"Parameter name eval or arguments is not allowed in strict mode",StrictParamDupe:"Strict mode function may not have duplicate parameter names",StrictFunctionName:"Function name may not be eval or arguments in strict mode",StrictOctalLiteral:"Octal literals are not allowed in strict mode.",StrictDelete:"Delete of an unqualified identifier in strict mode.",StrictDuplicateProperty:"Duplicate data property in object literal not allowed in strict mode",AccessorDataProperty:"Object literal may not have data and accessor property with the same name",AccessorGetSet:"Object literal may not have multiple get/set accessors with the same name",StrictLHSAssignment:"Assignment to eval or arguments is not allowed in strict mode",StrictLHSPostfix:"Postfix increment/decrement may not have eval or arguments operand in strict mode",StrictLHSPrefix:"Prefix increment/decrement may not have eval or arguments operand in strict mode",
    	StrictReservedWord:"Use of future reserved word in strict mode"},it={NonAsciiIdentifierStart:new RegExp("INVALIDE1"),NonAsciiIdentifierPart:new RegExp("INVALIDE2")},ot={name:"SyntaxTree",processComment:function(e){var t,n;if(!(e.type===nt.Program&&e.body.length>0)){for(ht.trailingComments.length>0?ht.trailingComments[0].range[0]>=e.range[1]?(n=ht.trailingComments,ht.trailingComments=[]):ht.trailingComments.length=0:ht.bottomRightStack.length>0&&ht.bottomRightStack[ht.bottomRightStack.length-1].trailingComments&&ht.bottomRightStack[ht.bottomRightStack.length-1].trailingComments[0].range[0]>=e.range[1]&&(n=ht.bottomRightStack[ht.bottomRightStack.length-1].trailingComments,delete ht.bottomRightStack[ht.bottomRightStack.length-1].trailingComments);ht.bottomRightStack.length>0&&ht.bottomRightStack[ht.bottomRightStack.length-1].range[0]>=e.range[0];)t=ht.bottomRightStack.pop();t?t.leadingComments&&t.leadingComments[t.leadingComments.length-1].range[1]<=e.range[0]&&(e.leadingComments=t.leadingComments,delete t.leadingComments):ht.leadingComments.length>0&&ht.leadingComments[ht.leadingComments.length-1].range[1]<=e.range[0]&&(e.leadingComments=ht.leadingComments,ht.leadingComments=[]),n&&(e.trailingComments=n),ht.bottomRightStack.push(e);}},markEnd:function(e,t){return ht.range&&(e.range=[t.start,lt]),ht.loc&&(e.loc=new j(void 0===t.startLineNumber?t.lineNumber:t.startLineNumber,t.start-(void 0===t.startLineStart?t.lineStart:t.startLineStart),st,lt-mt),this.postProcess(e)),ht.attachComment&&this.processComment(e),e},postProcess:function(e){return ht.source&&(e.loc.source=ht.source),e},createArrayExpression:function(e){return {type:nt.ArrayExpression,elements:e}},createAssignmentExpression:function(e,t,n){return {type:nt.AssignmentExpression,operator:e,left:t,right:n}},createBinaryExpression:function(e,t,n){var r="||"===e||"&&"===e?nt.LogicalExpression:nt.BinaryExpression;return {type:r,operator:e,left:t,right:n}},createBlockStatement:function(e){return {type:nt.BlockStatement,body:e}},createBreakStatement:function(e){return {type:nt.BreakStatement,label:e}},createCallExpression:function(e,t){return {type:nt.CallExpression,callee:e,arguments:t}},createCatchClause:function(e,t){return {type:nt.CatchClause,param:e,body:t}},createConditionalExpression:function(e,t,n){return {type:nt.ConditionalExpression,test:e,consequent:t,alternate:n}},createContinueStatement:function(e){return {type:nt.ContinueStatement,label:e}},createDebuggerStatement:function(){return {type:nt.DebuggerStatement}},createDoWhileStatement:function(e,t){return {type:nt.DoWhileStatement,body:e,test:t}},createEmptyStatement:function(){return {type:nt.EmptyStatement}},createExpressionStatement:function(e){return {type:nt.ExpressionStatement,expression:e}},createForStatement:function(e,t,n,r){return {type:nt.ForStatement,init:e,test:t,update:n,body:r}},createForInStatement:function(e,t,n){return {type:nt.ForInStatement,left:e,right:t,body:n,each:!1}},createFunctionDeclaration:function(e,t,n,r){return {type:nt.FunctionDeclaration,id:e,params:t,defaults:n,body:r,rest:null,generator:!1,expression:!1}},createFunctionExpression:function(e,t,n,r){return {type:nt.FunctionExpression,id:e,params:t,defaults:n,body:r,rest:null,generator:!1,expression:!1}},createIdentifier:function(e){return {type:nt.Identifier,name:e}},createIfStatement:function(e,t,n){return {type:nt.IfStatement,test:e,consequent:t,alternate:n}},createLabeledStatement:function(e,t){return {type:nt.LabeledStatement,label:e,body:t}},createLiteral:function(e){return {type:nt.Literal,value:e.value,raw:ct.slice(e.start,e.end)}},createMemberExpression:function(e,t,n){return {type:nt.MemberExpression,computed:"["===e,object:t,property:n}},createNewExpression:function(e,t){return {type:nt.NewExpression,callee:e,arguments:t}},createObjectExpression:function(e){return {type:nt.ObjectExpression,properties:e}},createPostfixExpression:function(e,t){return {type:nt.UpdateExpression,operator:e,argument:t,prefix:!1}},createProgram:function(e){return {type:nt.Program,body:e}},createProperty:function(e,t,n){return {type:nt.Property,key:t,value:n,kind:e}},createReturnStatement:function(e){return {type:nt.ReturnStatement,argument:e}},createSequenceExpression:function(e){return {type:nt.SequenceExpression,expressions:e}},createSwitchCase:function(e,t){return {type:nt.SwitchCase,test:e,consequent:t}},createSwitchStatement:function(e,t){return {type:nt.SwitchStatement,discriminant:e,cases:t}},createThisExpression:function(){return {type:nt.ThisExpression}},createThrowStatement:function(e){return {type:nt.ThrowStatement,argument:e}},createTryStatement:function(e,t,n,r){return {type:nt.TryStatement,block:e,guardedHandlers:t,handlers:n,finalizer:r}},createUnaryExpression:function(e,t){return "++"===e||"--"===e?{type:nt.UpdateExpression,operator:e,argument:t,prefix:!0}:{type:nt.UnaryExpression,operator:e,argument:t,prefix:!0}},createVariableDeclaration:function(e,t){return {type:nt.VariableDeclaration,declarations:e,kind:t}},createVariableDeclarator:function(e,t){return {type:nt.VariableDeclarator,id:e,init:t}},createWhileStatement:function(e,t){return {type:nt.WhileStatement,test:e,body:t}},createWithStatement:function(e,t){return {type:nt.WithStatement,object:e,body:t}}},e.version="1.2.5",e.tokenize=Ye,e.parse=Ze,e.Syntax=function(){var e,t={};"function"==typeof Object.create&&(t=Object.create(null));for(e in nt)nt.hasOwnProperty(e)&&(t[e]=nt[e]);return "function"==typeof Object.freeze&&Object.freeze(t),t}();});
    	},{}],9:[function(require,module,exports){
    	function EventEmitter(){this._events=this._events||{},this._maxListeners=this._maxListeners||void 0;}function isFunction(e){return "function"==typeof e}function isNumber(e){return "number"==typeof e}function isObject(e){return "object"==typeof e&&null!==e}function isUndefined(e){return void 0===e}module.exports=EventEmitter,EventEmitter.EventEmitter=EventEmitter,EventEmitter.prototype._events=void 0,EventEmitter.prototype._maxListeners=void 0,EventEmitter.defaultMaxListeners=10,EventEmitter.prototype.setMaxListeners=function(e){if(!isNumber(e)||0>e||isNaN(e))throw TypeError("n must be a positive number");return this._maxListeners=e,this},EventEmitter.prototype.emit=function(e){var t,n,i,s,r,o;if(this._events||(this._events={}),"error"===e&&(!this._events.error||isObject(this._events.error)&&!this._events.error.length)){if(t=arguments[1],t instanceof Error)throw t;throw TypeError('Uncaught, unspecified "error" event.')}if(n=this._events[e],isUndefined(n))return !1;if(isFunction(n))switch(arguments.length){case 1:n.call(this);break;case 2:n.call(this,arguments[1]);break;case 3:n.call(this,arguments[1],arguments[2]);break;default:s=Array.prototype.slice.call(arguments,1),n.apply(this,s);}else if(isObject(n))for(s=Array.prototype.slice.call(arguments,1),o=n.slice(),i=o.length,r=0;i>r;r++)o[r].apply(this,s);return !0},EventEmitter.prototype.addListener=function(e,t){var n;if(!isFunction(t))throw TypeError("listener must be a function");return this._events||(this._events={}),this._events.newListener&&this.emit("newListener",e,isFunction(t.listener)?t.listener:t),this._events[e]?isObject(this._events[e])?this._events[e].push(t):this._events[e]=[this._events[e],t]:this._events[e]=t,isObject(this._events[e])&&!this._events[e].warned&&(n=isUndefined(this._maxListeners)?EventEmitter.defaultMaxListeners:this._maxListeners,n&&n>0&&this._events[e].length>n&&(this._events[e].warned=!0,console.error("(node) warning: possible EventEmitter memory leak detected. %d listeners added. Use emitter.setMaxListeners() to increase limit.",this._events[e].length),"function"==typeof console.trace&&console.trace())),this},EventEmitter.prototype.on=EventEmitter.prototype.addListener,EventEmitter.prototype.once=function(e,t){function n(){this.removeListener(e,n),i||(i=!0,t.apply(this,arguments));}if(!isFunction(t))throw TypeError("listener must be a function");var i=!1;return n.listener=t,this.on(e,n),this},EventEmitter.prototype.removeListener=function(e,t){var n,i,s,r;if(!isFunction(t))throw TypeError("listener must be a function");if(!this._events||!this._events[e])return this;if(n=this._events[e],s=n.length,i=-1,n===t||isFunction(n.listener)&&n.listener===t)delete this._events[e],this._events.removeListener&&this.emit("removeListener",e,t);else if(isObject(n)){for(r=s;r-- >0;)if(n[r]===t||n[r].listener&&n[r].listener===t){i=r;break}if(0>i)return this;1===n.length?(n.length=0,delete this._events[e]):n.splice(i,1),this._events.removeListener&&this.emit("removeListener",e,t);}return this},EventEmitter.prototype.removeAllListeners=function(e){var t,n;if(!this._events)return this;if(!this._events.removeListener)return 0===arguments.length?this._events={}:this._events[e]&&delete this._events[e],this;if(0===arguments.length){for(t in this._events)"removeListener"!==t&&this.removeAllListeners(t);return this.removeAllListeners("removeListener"),this._events={},this}if(n=this._events[e],isFunction(n))this.removeListener(e,n);else if(n)for(;n.length;)this.removeListener(e,n[n.length-1]);return delete this._events[e],this},EventEmitter.prototype.listeners=function(e){return this._events&&this._events[e]?isFunction(this._events[e])?[this._events[e]]:this._events[e].slice():[]},EventEmitter.prototype.listenerCount=function(e){if(this._events){var t=this._events[e];if(isFunction(t))return 1;if(t)return t.length}return 0},EventEmitter.listenerCount=function(e,t){return e.listenerCount(t)};
    	},{}],10:[function(require,module,exports){
    function iota(r){for(var t=new Array(r),o=0;r>o;++o)t[o]=o;return t}module.exports=iota;
    	},{}],11:[function(require,module,exports){
    	module.exports=function(r){return !(null==r||!(r._isBuffer||r.constructor&&"function"==typeof r.constructor.isBuffer&&r.constructor.isBuffer(r)))};
    	},{}],12:[function(require,module,exports){
    function fixup(a){if(!a)return EmptyProc;for(var r=0;r<a.args.length;++r){var s=a.args[r];0===r?a.args[r]={name:s,lvalue:!0,rvalue:!!a.rvalue,count:a.count||1}:a.args[r]={name:s,lvalue:!1,rvalue:!0,count:1};}return a.thisVars||(a.thisVars=[]),a.localVars||(a.localVars=[]),a}function pcompile(a){return compile({args:a.args,pre:fixup(a.pre),body:fixup(a.body),post:fixup(a.proc),funcName:a.funcName})}function makeOp(a){for(var r=[],s=0;s<a.args.length;++s)r.push("a"+s);var i=new Function("P",["return function ",a.funcName,"_ndarrayops(",r.join(","),") {P(",r.join(","),");return a0}"].join(""));return i(pcompile(a))}var compile=require("cwise-compiler"),EmptyProc={body:"",args:[],thisVars:[],localVars:[]},assign_ops={add:"+",sub:"-",mul:"*",div:"/",mod:"%",band:"&",bor:"|",bxor:"^",lshift:"<<",rshift:">>",rrshift:">>>"};!function(){for(var a in assign_ops){var r=assign_ops[a];exports[a]=makeOp({args:["array","array","array"],body:{args:["a","b","c"],body:"a=b"+r+"c"},funcName:a}),exports[a+"eq"]=makeOp({args:["array","array"],body:{args:["a","b"],body:"a"+r+"=b"},rvalue:!0,funcName:a+"eq"}),exports[a+"s"]=makeOp({args:["array","array","scalar"],body:{args:["a","b","s"],body:"a=b"+r+"s"},funcName:a+"s"}),exports[a+"seq"]=makeOp({args:["array","scalar"],body:{args:["a","s"],body:"a"+r+"=s"},rvalue:!0,funcName:a+"seq"});}}();var unary_ops={not:"!",bnot:"~",neg:"-",recip:"1.0/"};!function(){for(var a in unary_ops){var r=unary_ops[a];exports[a]=makeOp({args:["array","array"],body:{args:["a","b"],body:"a="+r+"b"},funcName:a}),exports[a+"eq"]=makeOp({args:["array"],body:{args:["a"],body:"a="+r+"a"},rvalue:!0,count:2,funcName:a+"eq"});}}();var binary_ops={and:"&&",or:"||",eq:"===",neq:"!==",lt:"<",gt:">",leq:"<=",geq:">="};!function(){for(var a in binary_ops){var r=binary_ops[a];exports[a]=makeOp({args:["array","array","array"],body:{args:["a","b","c"],body:"a=b"+r+"c"},funcName:a}),exports[a+"s"]=makeOp({args:["array","array","scalar"],body:{args:["a","b","s"],body:"a=b"+r+"s"},funcName:a+"s"}),exports[a+"eq"]=makeOp({args:["array","array"],body:{args:["a","b"],body:"a=a"+r+"b"},rvalue:!0,count:2,funcName:a+"eq"}),exports[a+"seq"]=makeOp({args:["array","scalar"],body:{args:["a","s"],body:"a=a"+r+"s"},rvalue:!0,count:2,funcName:a+"seq"});}}();var math_unary=["abs","acos","asin","atan","ceil","cos","exp","floor","log","round","sin","sqrt","tan"];!function(){for(var a=0;a<math_unary.length;++a){var r=math_unary[a];exports[r]=makeOp({args:["array","array"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a","b"],body:"a=this_f(b)",thisVars:["this_f"]},funcName:r}),exports[r+"eq"]=makeOp({args:["array"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a"],body:"a=this_f(a)",thisVars:["this_f"]},rvalue:!0,count:2,funcName:r+"eq"});}}();var math_comm=["max","min","atan2","pow"];!function(){for(var a=0;a<math_comm.length;++a){var r=math_comm[a];exports[r]=makeOp({args:["array","array","array"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a","b","c"],body:"a=this_f(b,c)",thisVars:["this_f"]},funcName:r}),exports[r+"s"]=makeOp({args:["array","array","scalar"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a","b","c"],body:"a=this_f(b,c)",thisVars:["this_f"]},funcName:r+"s"}),exports[r+"eq"]=makeOp({args:["array","array"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a","b"],body:"a=this_f(a,b)",thisVars:["this_f"]},rvalue:!0,count:2,funcName:r+"eq"}),exports[r+"seq"]=makeOp({args:["array","scalar"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a","b"],body:"a=this_f(a,b)",thisVars:["this_f"]},rvalue:!0,count:2,funcName:r+"seq"});}}();var math_noncomm=["atan2","pow"];!function(){for(var a=0;a<math_noncomm.length;++a){var r=math_noncomm[a];exports[r+"op"]=makeOp({args:["array","array","array"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a","b","c"],body:"a=this_f(c,b)",thisVars:["this_f"]},funcName:r+"op"}),exports[r+"ops"]=makeOp({args:["array","array","scalar"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a","b","c"],body:"a=this_f(c,b)",thisVars:["this_f"]},funcName:r+"ops"}),exports[r+"opeq"]=makeOp({args:["array","array"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a","b"],body:"a=this_f(b,a)",thisVars:["this_f"]},rvalue:!0,count:2,funcName:r+"opeq"}),exports[r+"opseq"]=makeOp({args:["array","scalar"],pre:{args:[],body:"this_f=Math."+r,thisVars:["this_f"]},body:{args:["a","b"],body:"a=this_f(b,a)",thisVars:["this_f"]},rvalue:!0,count:2,funcName:r+"opseq"});}}(),exports.any=compile({args:["array"],pre:EmptyProc,body:{args:[{name:"a",lvalue:!1,rvalue:!0,count:1}],body:"if(a){return true}",localVars:[],thisVars:[]},post:{args:[],localVars:[],thisVars:[],body:"return false"},funcName:"any"}),exports.all=compile({args:["array"],pre:EmptyProc,body:{args:[{name:"x",lvalue:!1,rvalue:!0,count:1}],body:"if(!x){return false}",localVars:[],thisVars:[]},post:{args:[],localVars:[],thisVars:[],body:"return true"},funcName:"all"}),exports.sum=compile({args:["array"],pre:{args:[],localVars:[],thisVars:["this_s"],body:"this_s=0"},body:{args:[{name:"a",lvalue:!1,rvalue:!0,count:1}],body:"this_s+=a",localVars:[],thisVars:["this_s"]},post:{args:[],localVars:[],thisVars:["this_s"],body:"return this_s"},funcName:"sum"}),exports.prod=compile({args:["array"],pre:{args:[],localVars:[],thisVars:["this_s"],body:"this_s=1"},body:{args:[{name:"a",lvalue:!1,rvalue:!0,count:1}],body:"this_s*=a",localVars:[],thisVars:["this_s"]},post:{args:[],localVars:[],thisVars:["this_s"],body:"return this_s"},funcName:"prod"}),exports.norm2squared=compile({args:["array"],pre:{args:[],localVars:[],thisVars:["this_s"],body:"this_s=0"},body:{args:[{name:"a",lvalue:!1,rvalue:!0,count:2}],body:"this_s+=a*a",localVars:[],thisVars:["this_s"]},post:{args:[],localVars:[],thisVars:["this_s"],body:"return this_s"},funcName:"norm2squared"}),exports.norm2=compile({args:["array"],pre:{args:[],localVars:[],thisVars:["this_s"],body:"this_s=0"},body:{args:[{name:"a",lvalue:!1,rvalue:!0,count:2}],body:"this_s+=a*a",localVars:[],thisVars:["this_s"]},post:{args:[],localVars:[],thisVars:["this_s"],body:"return Math.sqrt(this_s)"},funcName:"norm2"}),exports.norminf=compile({args:["array"],pre:{args:[],localVars:[],thisVars:["this_s"],body:"this_s=0"},body:{args:[{name:"a",lvalue:!1,rvalue:!0,count:4}],body:"if(-a>this_s){this_s=-a}else if(a>this_s){this_s=a}",localVars:[],thisVars:["this_s"]},post:{args:[],localVars:[],thisVars:["this_s"],body:"return this_s"},funcName:"norminf"}),exports.norm1=compile({args:["array"],pre:{args:[],localVars:[],thisVars:["this_s"],body:"this_s=0"},body:{args:[{name:"a",lvalue:!1,rvalue:!0,count:3}],body:"this_s+=a<0?-a:a",localVars:[],thisVars:["this_s"]},post:{args:[],localVars:[],thisVars:["this_s"],body:"return this_s"},funcName:"norm1"}),exports.sup=compile({args:["array"],pre:{body:"this_h=-Infinity",args:[],thisVars:["this_h"],localVars:[]},body:{body:"if(_inline_1_arg0_>this_h)this_h=_inline_1_arg0_",args:[{name:"_inline_1_arg0_",lvalue:!1,rvalue:!0,count:2}],thisVars:["this_h"],localVars:[]},post:{body:"return this_h",args:[],thisVars:["this_h"],localVars:[]}}),exports.inf=compile({args:["array"],pre:{body:"this_h=Infinity",args:[],thisVars:["this_h"],localVars:[]},body:{body:"if(_inline_1_arg0_<this_h)this_h=_inline_1_arg0_",args:[{name:"_inline_1_arg0_",lvalue:!1,rvalue:!0,count:2}],thisVars:["this_h"],localVars:[]},post:{body:"return this_h",args:[],thisVars:["this_h"],localVars:[]}}),exports.argmin=compile({args:["index","array","shape"],pre:{body:"{this_v=Infinity;this_i=_inline_0_arg2_.slice(0)}",args:[{name:"_inline_0_arg0_",lvalue:!1,rvalue:!1,count:0},{name:"_inline_0_arg1_",lvalue:!1,rvalue:!1,count:0},{name:"_inline_0_arg2_",lvalue:!1,rvalue:!0,count:1}],thisVars:["this_i","this_v"],localVars:[]},body:{body:"{if(_inline_1_arg1_<this_v){this_v=_inline_1_arg1_;for(var _inline_1_k=0;_inline_1_k<_inline_1_arg0_.length;++_inline_1_k){this_i[_inline_1_k]=_inline_1_arg0_[_inline_1_k]}}}",args:[{name:"_inline_1_arg0_",lvalue:!1,rvalue:!0,count:2},{name:"_inline_1_arg1_",lvalue:!1,rvalue:!0,count:2}],thisVars:["this_i","this_v"],localVars:["_inline_1_k"]},post:{body:"{return this_i}",args:[],thisVars:["this_i"],localVars:[]}}),exports.argmax=compile({args:["index","array","shape"],pre:{body:"{this_v=-Infinity;this_i=_inline_0_arg2_.slice(0)}",args:[{name:"_inline_0_arg0_",lvalue:!1,rvalue:!1,count:0},{name:"_inline_0_arg1_",lvalue:!1,rvalue:!1,count:0},{name:"_inline_0_arg2_",lvalue:!1,rvalue:!0,count:1}],thisVars:["this_i","this_v"],localVars:[]},body:{body:"{if(_inline_1_arg1_>this_v){this_v=_inline_1_arg1_;for(var _inline_1_k=0;_inline_1_k<_inline_1_arg0_.length;++_inline_1_k){this_i[_inline_1_k]=_inline_1_arg0_[_inline_1_k]}}}",args:[{name:"_inline_1_arg0_",lvalue:!1,rvalue:!0,count:2},{name:"_inline_1_arg1_",lvalue:!1,rvalue:!0,count:2}],thisVars:["this_i","this_v"],localVars:["_inline_1_k"]},post:{body:"{return this_i}",args:[],thisVars:["this_i"],localVars:[]}}),exports.random=makeOp({args:["array"],pre:{args:[],body:"this_f=Math.random",thisVars:["this_f"]},body:{args:["a"],body:"a=this_f()",thisVars:["this_f"]},funcName:"random"}),exports.assign=makeOp({args:["array","array"],body:{args:["a","b"],body:"a=b"},funcName:"assign"}),exports.assigns=makeOp({args:["array","scalar"],body:{args:["a","b"],body:"a=b"},funcName:"assigns"}),exports.equals=compile({args:["array","array"],pre:EmptyProc,body:{args:[{name:"x",lvalue:!1,rvalue:!0,count:1},{name:"y",lvalue:!1,rvalue:!0,count:1}],body:"if(x!==y){return false}",localVars:[],thisVars:[]},post:{args:[],localVars:[],thisVars:[],body:"return true"},funcName:"equals"});
    	},{"cwise-compiler":1}],13:[function(require,module,exports){
    var ndarray=require("ndarray"),do_convert=require("./doConvert.js");module.exports=function(r,e){for(var n=[],a=r,t=1;a instanceof Array;)n.push(a.length),t*=a.length,a=a[0];return 0===n.length?ndarray():(e||(e=ndarray(new Float64Array(t),n)),do_convert(e,r),e)};
    	},{"./doConvert.js":14,"ndarray":16}],14:[function(require,module,exports){
    	module.exports=require("cwise-compiler")({args:["array","scalar","index"],pre:{body:"{}",args:[],thisVars:[],localVars:[]},body:{body:"{\nvar _inline_1_v=_inline_1_arg1_,_inline_1_i\nfor(_inline_1_i=0;_inline_1_i<_inline_1_arg2_.length-1;++_inline_1_i) {\n_inline_1_v=_inline_1_v[_inline_1_arg2_[_inline_1_i]]\n}\n_inline_1_arg0_=_inline_1_v[_inline_1_arg2_[_inline_1_arg2_.length-1]]\n}",args:[{name:"_inline_1_arg0_",lvalue:!0,rvalue:!1,count:1},{name:"_inline_1_arg1_",lvalue:!1,rvalue:!0,count:1},{name:"_inline_1_arg2_",lvalue:!1,rvalue:!0,count:4}],thisVars:[],localVars:["_inline_1_i","_inline_1_v"]},post:{body:"{}",args:[],thisVars:[],localVars:[]},funcName:"convert",blockSize:64});
    	},{"cwise-compiler":1}],15:[function(require,module,exports){
    var dup=require("dup"),do_unpack=require("cwise/lib/wrapper")({args:["array","scalar","index"],pre:{body:"{}",args:[],thisVars:[],localVars:[]},body:{body:"{var _inline_1_a,_inline_1_e=_inline_1_arg1_;for(_inline_1_a=0;_inline_1_a<_inline_1_arg2_.length-1;++_inline_1_a)_inline_1_e=_inline_1_e[_inline_1_arg2_[_inline_1_a]];_inline_1_e[_inline_1_arg2_[_inline_1_arg2_.length-1]]=_inline_1_arg0_}",args:[{name:"_inline_1_arg0_",lvalue:!1,rvalue:!0,count:1},{name:"_inline_1_arg1_",lvalue:!1,rvalue:!0,count:1},{name:"_inline_1_arg2_",lvalue:!1,rvalue:!0,count:4}],thisVars:[],localVars:["_inline_1_a","_inline_1_e"]},post:{body:"{}",args:[],thisVars:[],localVars:[]},debug:!1,funcName:"unpackCwise",blockSize:64});module.exports=function(_){var n=dup(_.shape);return do_unpack(_,n),n};
    	},{"cwise/lib/wrapper":6,"dup":7}],16:[function(require,module,exports){
    	function compare1st(t,r){return t[0]-r[0]}function order(){var t,r=this.stride,e=new Array(r.length);for(t=0;t<e.length;++t)e[t]=[Math.abs(r[t]),t];e.sort(compare1st);var n=new Array(e.length);for(t=0;t<n.length;++t)n[t]=e[t][1];return n}function compileConstructor(t,r){var e=["View",r,"d",t].join("");0>r&&(e="View_Nil"+t);var n="generic"===t;if(-1===r){var o="function "+e+"(a){this.data=a;};var proto="+e+".prototype;proto.dtype='"+t+"';proto.index=function(){return -1};proto.size=0;proto.dimension=-1;proto.shape=proto.stride=proto.order=[];proto.lo=proto.hi=proto.transpose=proto.step=function(){return new "+e+"(this.data);};proto.get=proto.set=function(){};proto.pick=function(){return null};return function construct_"+e+"(a){return new "+e+"(a);}",i=new Function(o);return i()}if(0===r){var o="function "+e+"(a,d) {this.data = a;this.offset = d};var proto="+e+".prototype;proto.dtype='"+t+"';proto.index=function(){return this.offset};proto.dimension=0;proto.size=1;proto.shape=proto.stride=proto.order=[];proto.lo=proto.hi=proto.transpose=proto.step=function "+e+"_copy() {return new "+e+"(this.data,this.offset)};proto.pick=function "+e+"_pick(){return TrivialArray(this.data);};proto.valueOf=proto.get=function "+e+"_get(){return "+(n?"this.data.get(this.offset)":"this.data[this.offset]")+"};proto.set=function "+e+"_set(v){return "+(n?"this.data.set(this.offset,v)":"this.data[this.offset]=v")+"};return function construct_"+e+"(a,b,c,d){return new "+e+"(a,d)}",i=new Function("TrivialArray",o);return i(CACHED_CONSTRUCTORS[t][0])}var o=["'use strict'"],a=iota(r),s=a.map(function(t){return "i"+t}),u="this.offset+"+a.map(function(t){return "this.stride["+t+"]*i"+t}).join("+"),p=a.map(function(t){return "b"+t}).join(","),f=a.map(function(t){return "c"+t}).join(",");o.push("function "+e+"(a,"+p+","+f+",d){this.data=a","this.shape=["+p+"]","this.stride=["+f+"]","this.offset=d|0}","var proto="+e+".prototype","proto.dtype='"+t+"'","proto.dimension="+r),o.push("Object.defineProperty(proto,'size',{get:function "+e+"_size(){return "+a.map(function(t){return "this.shape["+t+"]"}).join("*"),"}})"),1===r?o.push("proto.order=[0]"):(o.push("Object.defineProperty(proto,'order',{get:"),4>r?(o.push("function "+e+"_order(){"),2===r?o.push("return (Math.abs(this.stride[0])>Math.abs(this.stride[1]))?[1,0]:[0,1]}})"):3===r&&o.push("var s0=Math.abs(this.stride[0]),s1=Math.abs(this.stride[1]),s2=Math.abs(this.stride[2]);if(s0>s1){if(s1>s2){return [2,1,0];}else if(s0>s2){return [1,2,0];}else{return [1,0,2];}}else if(s0>s2){return [2,0,1];}else if(s2>s1){return [0,1,2];}else{return [0,2,1];}}})")):o.push("ORDER})")),o.push("proto.set=function "+e+"_set("+s.join(",")+",v){"),n?o.push("return this.data.set("+u+",v)}"):o.push("return this.data["+u+"]=v}"),o.push("proto.get=function "+e+"_get("+s.join(",")+"){"),n?o.push("return this.data.get("+u+")}"):o.push("return this.data["+u+"]}"),o.push("proto.index=function "+e+"_index(",s.join(),"){return "+u+"}"),o.push("proto.hi=function "+e+"_hi("+s.join(",")+"){return new "+e+"(this.data,"+a.map(function(t){return ["(typeof i",t,"!=='number'||i",t,"<0)?this.shape[",t,"]:i",t,"|0"].join("")}).join(",")+","+a.map(function(t){return "this.stride["+t+"]"}).join(",")+",this.offset)}");var h=a.map(function(t){return "a"+t+"=this.shape["+t+"]"}),c=a.map(function(t){return "c"+t+"=this.stride["+t+"]"});o.push("proto.lo=function "+e+"_lo("+s.join(",")+"){var b=this.offset,d=0,"+h.join(",")+","+c.join(","));for(var d=0;r>d;++d)o.push("if(typeof i"+d+"==='number'&&i"+d+">=0){d=i"+d+"|0;b+=c"+d+"*d;a"+d+"-=d}");o.push("return new "+e+"(this.data,"+a.map(function(t){return "a"+t}).join(",")+","+a.map(function(t){return "c"+t}).join(",")+",b)}"),o.push("proto.step=function "+e+"_step("+s.join(",")+"){var "+a.map(function(t){return "a"+t+"=this.shape["+t+"]"}).join(",")+","+a.map(function(t){return "b"+t+"=this.stride["+t+"]"}).join(",")+",c=this.offset,d=0,ceil=Math.ceil");for(var d=0;r>d;++d)o.push("if(typeof i"+d+"==='number'){d=i"+d+"|0;if(d<0){c+=b"+d+"*(a"+d+"-1);a"+d+"=ceil(-a"+d+"/d)}else{a"+d+"=ceil(a"+d+"/d)}b"+d+"*=d}");o.push("return new "+e+"(this.data,"+a.map(function(t){return "a"+t}).join(",")+","+a.map(function(t){return "b"+t}).join(",")+",c)}");for(var y=new Array(r),l=new Array(r),d=0;r>d;++d)y[d]="a[i"+d+"]",l[d]="b[i"+d+"]";o.push("proto.transpose=function "+e+"_transpose("+s+"){"+s.map(function(t,r){return t+"=("+t+"===undefined?"+r+":"+t+"|0)"}).join(";"),"var a=this.shape,b=this.stride;return new "+e+"(this.data,"+y.join(",")+","+l.join(",")+",this.offset)}"),o.push("proto.pick=function "+e+"_pick("+s+"){var a=[],b=[],c=this.offset");for(var d=0;r>d;++d)o.push("if(typeof i"+d+"==='number'&&i"+d+">=0){c=(c+this.stride["+d+"]*i"+d+")|0}else{a.push(this.shape["+d+"]);b.push(this.stride["+d+"])}");o.push("var ctor=CTOR_LIST[a.length+1];return ctor(this.data,a,b,c)}"),o.push("return function construct_"+e+"(data,shape,stride,offset){return new "+e+"(data,"+a.map(function(t){return "shape["+t+"]"}).join(",")+","+a.map(function(t){return "stride["+t+"]"}).join(",")+",offset)}");var i=new Function("CTOR_LIST","ORDER",o.join("\n"));return i(CACHED_CONSTRUCTORS[t],order)}function arrayDType(t){if(isBuffer(t))return "buffer";if(hasTypedArrays)switch(Object.prototype.toString.call(t)){case"[object Float64Array]":return "float64";case"[object Float32Array]":return "float32";case"[object Int8Array]":return "int8";case"[object Int16Array]":return "int16";case"[object Int32Array]":return "int32";case"[object Uint8Array]":return "uint8";case"[object Uint16Array]":return "uint16";case"[object Uint32Array]":return "uint32";case"[object Uint8ClampedArray]":return "uint8_clamped"}return Array.isArray(t)?"array":"generic"}function wrappedNDArrayCtor(t,r,e,n){if(void 0===t){var o=CACHED_CONSTRUCTORS.array[0];return o([])}"number"==typeof t&&(t=[t]),void 0===r&&(r=[t.length]);var i=r.length;if(void 0===e){e=new Array(i);for(var a=i-1,s=1;a>=0;--a)e[a]=s,s*=r[a];}if(void 0===n){n=0;for(var a=0;i>a;++a)e[a]<0&&(n-=(r[a]-1)*e[a]);}for(var u=arrayDType(t),p=CACHED_CONSTRUCTORS[u];p.length<=i+1;)p.push(compileConstructor(u,p.length-1));var o=p[i+1];return o(t,r,e,n)}var iota=require("iota-array"),isBuffer=require("is-buffer"),hasTypedArrays="undefined"!=typeof Float64Array,CACHED_CONSTRUCTORS={float32:[],float64:[],int8:[],int16:[],int32:[],uint8:[],uint16:[],uint32:[],array:[],uint8_clamped:[],buffer:[],generic:[]};module.exports=wrappedNDArrayCtor;
    	},{"iota-array":10,"is-buffer":11}],17:[function(require,module,exports){
    function unique_pred(n,e){for(var u=1,t=n.length,i=n[0],r=n[0],o=1;t>o;++o)if(r=i,i=n[o],e(i,r)){if(o===u){u++;continue}n[u++]=i;}return n.length=u,n}function unique_eq(n){for(var e=1,u=n.length,t=n[0],i=n[0],r=1;u>r;++r,i=t)if(i=t,t=n[r],t!==i){if(r===e){e++;continue}n[e++]=t;}return n.length=e,n}function unique(n,e,u){return 0===n.length?n:e?(u||n.sort(e),unique_pred(n,e)):(u||n.sort(),unique_eq(n))}module.exports=unique;
    	},{}],18:[function(require,module,exports){
    function _interopRequireDefault(e){return e&&e.__esModule?e:{"default":e}}function _toConsumableArray(e){if(Array.isArray(e)){for(var t=0,r=Array(e.length);t<e.length;t++)r[t]=e[t];return r}return Array.from(e)}function _classCallCheck(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function _possibleConstructorReturn(e,t){if(!e)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return !t||"object"!=typeof t&&"function"!=typeof t?e:t}function _inherits(e,t){if("function"!=typeof t&&null!==t)throw new TypeError("Super expression must either be null or a function, not "+typeof t);e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,enumerable:!1,writable:!0,configurable:!0}}),t&&(Object.setPrototypeOf?Object.setPrototypeOf(e,t):e.__proto__=t);}Object.defineProperty(exports,"__esModule",{value:!0});var _slicedToArray=function(){function e(e,t){var r=[],a=!0,i=!1,n=void 0;try{for(var u,s=e[Symbol.iterator]();!(a=(u=s.next()).done)&&(r.push(u.value),!t||r.length!==t);a=!0);}catch(o){i=!0,n=o;}finally{try{!a&&s["return"]&&s["return"]();}finally{if(i)throw n}}return r}return function(t,r){if(Array.isArray(t))return t;if(Symbol.iterator in Object(t))return e(t,r);throw new TypeError("Invalid attempt to destructure non-iterable instance")}}(),_createClass=function(){function e(e,t){for(var r=0;r<t.length;r++){var a=t[r];a.enumerable=a.enumerable||!1,a.configurable=!0,"value"in a&&(a.writable=!0),Object.defineProperty(e,a.key,a);}}return function(t,r,a){return r&&e(t.prototype,r),a&&e(t,a),t}}(),_events=require("events"),_ndarray=require("ndarray"),_ndarray2=_interopRequireDefault(_ndarray),_ndarrayOps=require("ndarray-ops"),_ndarrayOps2=_interopRequireDefault(_ndarrayOps),_ndarrayPack=require("ndarray-pack"),_ndarrayPack2=_interopRequireDefault(_ndarrayPack),_ndarrayUnpack=require("ndarray-unpack"),_ndarrayUnpack2=_interopRequireDefault(_ndarrayUnpack),_cwise=require("cwise"),_cwise2=_interopRequireDefault(_cwise),_randn=require("./randn"),_randn2=_interopRequireDefault(_randn),_pairwiseDistances=require("./pairwise-distances"),_pairwiseDistances2=_interopRequireDefault(_pairwiseDistances),_jointProbabilities=require("./joint-probabilities"),_jointProbabilities2=_interopRequireDefault(_jointProbabilities),_klDivergence=require("./kl-divergence"),_klDivergence2=_interopRequireDefault(_klDivergence),TSNE=function(e){function t(e){_classCallCheck(this,t);var r=_possibleConstructorReturn(this,Object.getPrototypeOf(t).call(this));return e=e||{},r.dim=e.dim||2,r.perplexity=e.perplexity||30,r.earlyExaggeration=e.earlyExaggeration||4,r.learningRate=e.learningRate||1e3,r.nIter=e.nIter||1e3,r.metric=e.metric||"euclidean",r.barneshut=e.barneshut||!1,r.inputData=null,r.outputEmbedding=null,r}return _inherits(t,e),_createClass(t,[{key:"init",value:function(e){e=e||{};var t=e.data||[],r=e.type||"dense";if("dense"===r)this.inputData=(0, _ndarrayPack2["default"])(t);else {if("sparse"!==r)throw new Error("input data type must be dense or sparse");for(var a=[],i=1,n=function(e){var r=Math.max.apply(null,t.map(function(t){return t[e]}))+1;a.push(r),i*=r;},u=0;u<t[0].length;u++)n(u);this.inputData=(0, _ndarray2["default"])(new Float64Array(i),a);var s=!0,o=!1,l=void 0;try{for(var d,p=t[Symbol.iterator]();!(s=(d=p.next()).done);s=!0){var c,f=d.value;(c=this.inputData).set.apply(c,_toConsumableArray(f).concat([1]));}}catch(y){o=!0,l=y;}finally{try{!s&&p["return"]&&p["return"]();}finally{if(o)throw l}}}this.outputEmbedding=(0, _randn2["default"])(this.inputData.shape[0],this.dim);}},{key:"run",value:function(){this.emit("progressStatus","Calculating pairwise distances"),this.distances=(0, _pairwiseDistances2["default"])(this.inputData,this.metric),this.emit("progressStatus","Calculating joint probabilities"),this.alpha=Math.max(this.dim-1,1),this.P=(0, _jointProbabilities2["default"])(this.distances,this.perplexity);var e=Number.MAX_VALUE,t=0;this.emit("progressStatus","Early exaggeration with momentum 0.5"),_ndarrayOps2["default"].mulseq(this.P,this.earlyExaggeration);var r=this._gradDesc(t,50,.5,0,0),a=_slicedToArray(r,2);e=a[0],t=a[1],this.emit("progressStatus","Early exaggeration with momentum 0.8");var i=this._gradDesc(t+1,100,.8,0,0),n=_slicedToArray(i,2);e=n[0],t=n[1],this.emit("progressStatus","Final optimization with momentum 0.8"),_ndarrayOps2["default"].divseq(this.P,this.earlyExaggeration);var u=this._gradDesc(t+1,this.nIter,.8,1e-6,1e-6),s=_slicedToArray(u,2);return e=s[0],t=s[1],this.emit("progressStatus","Optimization end"),[e,t]}},{key:"rerun",value:function(){this.outputEmbedding=(0, _randn2["default"])(this.inputData.shape[0],this.dim);var e=this.run(),t=_slicedToArray(e,2),r=t[0],a=t[1];return [r,a]}},{key:"getOutput",value:function(){return (0, _ndarrayUnpack2["default"])(this.outputEmbedding)}},{key:"getOutputScaled",value:function(){for(var e=(0, _ndarray2["default"])(new Float64Array(this.outputEmbedding.size),this.outputEmbedding.shape),t=(0, _ndarray2["default"])(new Float64Array(this.outputEmbedding.shape[0]),[this.outputEmbedding.shape[0]]),r=0;r<this.outputEmbedding.shape[1];r++){var a=_ndarrayOps2["default"].sup(_ndarrayOps2["default"].abs(t,this.outputEmbedding.pick(null,r)));_ndarrayOps2["default"].divs(e.pick(null,r),this.outputEmbedding.pick(null,r),a);}return (0, _ndarrayUnpack2["default"])(e)}},{key:"_gradDesc",value:function(e,t,r){for(var a=arguments.length<=3||void 0===arguments[3]?1e-6:arguments[3],i=arguments.length<=4||void 0===arguments[4]?1e-6:arguments[4],n=30,u=(0, _ndarray2["default"])(new Float64Array(this.outputEmbedding.size),this.outputEmbedding.shape),s=new Float64Array(this.outputEmbedding.size),o=0;o<s.length;o++)s[o]=1;var l=(0, _ndarray2["default"])(s,this.outputEmbedding.shape),d=Number.MAX_VALUE,p=Number.MAX_VALUE,c=0,f=(0, _cwise2["default"])({args:["array"],pre:function(e){this.sum=0;},body:function(e){this.sum+=e*e;},post:function(e){return Math.sqrt(this.sum)}}),y=(0, _cwise2["default"])({args:["array","array","array"],body:function(e,t,r){}}),h=void 0;for(h=e;t>h;h++){var _=(0, _klDivergence2["default"])(this.outputEmbedding,this.P,this.alpha),m=_slicedToArray(_,2),g=m[0],b=m[1],v=Math.abs(g-d);d=g;var w=f(b);if(this.emit("progressIter",[h,d,w]),p>d)p=d,c=h;else if(h-c>n)break;if(a>=w)break;if(i>=v)break;y(l,u,b),_ndarrayOps2["default"].muleq(b,l);var E=(0, _ndarray2["default"])(new Float64Array(b.size),b.shape);_ndarrayOps2["default"].muls(E,b,this.learningRate),_ndarrayOps2["default"].mulseq(u,r),_ndarrayOps2["default"].subeq(u,E),_ndarrayOps2["default"].addeq(this.outputEmbedding,u),this.emit("progressData",this.getOutputScaled());}return [d,h]}}]),t}(_events.EventEmitter);exports["default"]=TSNE,module.exports=exports["default"];
    	},{"./joint-probabilities":19,"./kl-divergence":20,"./pairwise-distances":21,"./randn":22,"cwise":5,"events":9,"ndarray":16,"ndarray-ops":12,"ndarray-pack":13,"ndarray-unpack":15}],19:[function(require,module,exports){
    function _interopRequireDefault(r){return r&&r.__esModule?r:{"default":r}}Object.defineProperty(exports,"__esModule",{value:!0}),exports["default"]=function(r,a){for(var e=100,t=r.shape[0],d=(0, _ndarray2["default"])(new Float64Array(t*t),[t,t]),o=(0, _ndarray2["default"])(new Float64Array(t*t),[t,t]),u=void 0,n=void 0,s=1/0,f=Math.log(a),i=void 0,l=void 0,p=void 0,y=void 0,E=0;t>E;E++){u=1,n=-(1/0),s=1/0;for(var O=0;e>O;O++){for(var v=0;t>v;v++)d.set(E,v,Math.exp(-r.get(E,v)*u));d.set(E,E,0),p=0;for(var v=0;t>v;v++)p+=d.get(E,v);0==p&&(p=EPSILON_DBL),y=0;for(var v=0;t>v;v++)d.set(E,v,d.get(E,v)/p),y+=r.get(E,v)*d.get(E,v);if(l=Math.log(p)+u*y,i=l-f,Math.abs(i)<=PERPLEXITY_TOLERANCE)break;i>0?(n=u,u=s==1/0?2*u:(u+s)/2):(s=u,n==-(1/0)?u/=2:u=(u+n)/2);}}_ndarrayOps2["default"].add(o,d,d.transpose(1,0));var L=Math.max(_ndarrayOps2["default"].sum(o),MACHINE_EPSILON);return _ndarrayOps2["default"].divseq(o,L),_ndarrayOps2["default"].maxseq(o,MACHINE_EPSILON),o};var _ndarray=require("ndarray"),_ndarray2=_interopRequireDefault(_ndarray),_ndarrayOps=require("ndarray-ops"),_ndarrayOps2=_interopRequireDefault(_ndarrayOps),EPSILON_DBL=1e-7,MACHINE_EPSILON=Number.EPSILON||2.220446049250313e-16,PERPLEXITY_TOLERANCE=1e-5;module.exports=exports["default"];
    	},{"ndarray":16,"ndarray-ops":12}],20:[function(require,module,exports){
    function _interopRequireDefault(a){return a&&a.__esModule?a:{"default":a}}Object.defineProperty(exports,"__esModule",{value:!0}),exports["default"]=function(a,r,e){var d=a.shape[0],s=a.shape[1],t=(0, _ndarray2["default"])(new Float64Array(d*d),[d,d]),u=(0, _pairwiseDistances2["default"])(a,"euclidean");square(u);var n=(e+1)/-2;_ndarrayOps2["default"].powseq(_ndarrayOps2["default"].divseq(_ndarrayOps2["default"].addseq(u,1),e),n);for(var l=0;d>l;l++)u.set(l,l,0);var _=Math.max(_ndarrayOps2["default"].sum(u),MACHINE_EPSILON);_ndarrayOps2["default"].divs(t,u,_),_ndarrayOps2["default"].maxseq(t,MACHINE_EPSILON);var i=(0, _ndarray2["default"])(new Float64Array(d*d),[d,d]),p=(0, _ndarray2["default"])(new Float64Array(d*d),[d,d]);_ndarrayOps2["default"].div(i,r,t),_ndarrayOps2["default"].logeq(i),_ndarrayOps2["default"].assign(p,r);var f=_ndarrayOps2["default"].sum(_ndarrayOps2["default"].muleq(i,p)),y=(0, _ndarray2["default"])(new Float64Array(a.size),a.shape),o=(0, _ndarray2["default"])(new Float64Array(d*d),[d,d]);_ndarrayOps2["default"].sub(o,r,t),_ndarrayOps2["default"].muleq(o,u);for(var l=0;d>l;l++)for(var O=0;s>O;O++){var q=(0, _ndarray2["default"])(new Float64Array(a.shape[0]),[a.shape[0]]);_ndarrayOps2["default"].assign(q,a.pick(null,O)),_ndarrayOps2["default"].addseq(_ndarrayOps2["default"].negeq(q),a.get(l,O)),_ndarrayOps2["default"].muleq(q,o.pick(l,null)),y.set(l,O,_ndarrayOps2["default"].sum(q));}var c=2*(e+1)/e;return _ndarrayOps2["default"].mulseq(y,c),[f,y]};var _ndarray=require("ndarray"),_ndarray2=_interopRequireDefault(_ndarray),_ndarrayOps=require("ndarray-ops"),_ndarrayOps2=_interopRequireDefault(_ndarrayOps),_cwise=require("cwise"),_cwise2=_interopRequireDefault(_cwise),_pairwiseDistances=require("./pairwise-distances"),_pairwiseDistances2=_interopRequireDefault(_pairwiseDistances),MACHINE_EPSILON=Number.EPSILON||2.220446049250313e-16,square=(0, _cwise2["default"])({args:["array"],body:function(a){}});module.exports=exports["default"];
    	},{"./pairwise-distances":21,"cwise":5,"ndarray":16,"ndarray-ops":12}],21:[function(require,module,exports){
    function _interopRequireDefault(t){return t&&t.__esModule?t:{"default":t}}Object.defineProperty(exports,"__esModule",{value:!0}),exports["default"]=function(t,r){var a=t.shape[0],e=(0, _ndarray2["default"])(new Float64Array(a*a),[a,a]);switch(r){case"euclidean":for(var i=0;a>i;i++)for(var s=i+1;a>s;s++){var n=euclidean(t.pick(i,null),t.pick(s,null));e.set(i,s,n),e.set(s,i,n);}break;case"manhattan":for(var i=0;a>i;i++)for(var s=i+1;a>s;s++){var n=manhattan(t.pick(i,null),t.pick(s,null));e.set(i,s,n),e.set(s,i,n);}break;case"jaccard":for(var i=0;a>i;i++)for(var s=i+1;a>s;s++){var n=jaccard(t.pick(i,null),t.pick(s,null));e.set(i,s,n),e.set(s,i,n);}break;case"dice":for(var i=0;a>i;i++)for(var s=i+1;a>s;s++){var n=dice(t.pick(i,null),t.pick(s,null));e.set(i,s,n),e.set(s,i,n);}}return e};var _ndarray=require("ndarray"),_ndarray2=_interopRequireDefault(_ndarray),_cwise=require("cwise"),_cwise2=_interopRequireDefault(_cwise),euclidean=(0, _cwise2["default"])({args:["array","array"],pre:function(t,r){this.sum=0;},body:function(t,r){var a=t-r;this.sum+=a*a;},post:function(t,r){return Math.sqrt(this.sum)}}),manhattan=(0, _cwise2["default"])({args:["array","array"],pre:function(t,r){this.sum=0;},body:function(t,r){this.sum+=Math.abs(t-r);},post:function(t,r){return this.sum}}),jaccard=(0, _cwise2["default"])({args:["array","array"],pre:function(t,r){this.tf=0,this.tt=0;},body:function(t,r){var a=Math.round(t),e=Math.round(r);this.tf+=+(a!=e),this.tt+=+(1==a&&1==e);},post:function(t,r){return this.tf+this.tt===0?1:this.tf/(this.tf+this.tt)}}),dice=(0, _cwise2["default"])({args:["array","array"],pre:function(t,r){this.tf=0,this.tt=0;},body:function(t,r){var a=Math.round(t),e=Math.round(r);this.tf+=+(a!=e),this.tt+=+(1==a&&1==e);},post:function(t,r){return this.tf+this.tt===0?1:this.tf/(this.tf+2*this.tt)}});module.exports=exports["default"];
    	},{"cwise":5,"ndarray":16}],22:[function(require,module,exports){
    function _interopRequireDefault(r){return r&&r.__esModule?r:{"default":r}}function gaussRandom(){var r=2*Math.random()-1,e=2*Math.random()-1,a=r*r+e*e;return 0==a||a>1?gaussRandom():r*Math.sqrt(-2*Math.log(a)/a)}Object.defineProperty(exports,"__esModule",{value:!0}),exports["default"]=function(r,e){for(var a=new Float64Array(r*e),t=0;t<a.length;t++)a[t]=1e-4*gaussRandom();return (0, _ndarray2["default"])(a,[r,e])};var _ndarray=require("ndarray"),_ndarray2=_interopRequireDefault(_ndarray);module.exports=exports["default"];
    	},{"ndarray":16}]},{},[18])(18)
    	});
    } (tsne_min));

    var TSNE = /*@__PURE__*/getDefaultExportFromCjs(tsne_min.exports);

    function _optionalChain(ops) { let lastAccessLHS = undefined; let value = ops[0]; let i = 1; while (i < ops.length) { const op = ops[i]; const fn = ops[i + 1]; i += 2; if ((op === 'optionalAccess' || op === 'optionalCall') && value == null) { return undefined; } if (op === 'access' || op === 'optionalAccess') { lastAccessLHS = value; value = fn(value); } else if (op === 'call' || op === 'optionalCall') { value = fn((...args) => value.call(lastAccessLHS, ...args)); lastAccessLHS = undefined; } } return value; }//https://towardsdatascience.com/understanding-feature-engineering-part-4-deep-learning-methods-for-text-data-96c44370bbfa






    /**
     * use a corpus to generate features from an embedding layer with Tensorflow
     * @class FeatureEmbedding
     * @implements {BaseNeuralNetwork}
     */
    class FeatureEmbedding extends BaseNeuralNetwork {
      
      
      
      
      
      
      
      // settings: TensorScriptOptions;
      static async getFeatureDataSet( { inputMatrixFeatures, PAD = 'PAD', initialIdToFeature, initialFeatureToId, }




    ) {
        let featIndex = initialFeatureToId ? Object.keys(initialFeatureToId).length : 1;
        const idToFeature = {
          0: this && this.settings && this.settings.PAD
            ? this.settings.PAD
            : PAD,
          ...initialIdToFeature,
        };
        const featureToId = inputMatrixFeatures.reduce((result, inputFeatureArray) => { 
          inputFeatureArray.forEach((inputFeature) => {
            if (!result[inputFeature]) {
              if (idToFeature[featIndex]) featIndex++;
              result[inputFeature] = featIndex;
              //@ts-ignore
              idToFeature[featIndex] = inputFeature;
              featIndex++;
            }
          });
          return result;
        }, {
            [PAD]: 0,
            ...initialFeatureToId,
        });
        //@ts-ignore
        const featureIds = inputMatrixFeatures.map(inputFeatureArray =>
          inputFeatureArray.map(inputFeature => featureToId[inputFeature])
        );
        // console.log('featureIds', featureIds);
        // console.log('featureIds.length', featureIds.length);
        // console.log('inputMatrixFeatures.length', inputMatrixFeatures.length);
        const numberOfFeatures = Object.keys(featureToId).length;
        return {
          featureToId, //word2id
          idToFeature, //id2word
          featureIds, //wids
          numberOfFeatures, //vocab_size
        };
      }
      static getMergedArray(base = [], merger= [], append=false, truncate=true) {
        let arr = new Array().concat(base);
        if (append) arr.splice(base.length-merger.length,merger.length,...merger);
        else arr.splice(0, merger.length, ...merger);
        if (truncate && append) return arr.slice(-1 * base.length);
        else if (truncate) return arr.slice(0, base.length);
        else return arr;
      }
      /**
       */
      static async getContextPairs( { inputMatrix, numberOfFeatures, window_size = 2, tf, }) {
        const tensorflow = this && this.tf ? this.tf : Tensorflow;
        const context_length = (this && this.settings && this.settings.windowSize ? this.settings.windowSize : window_size) * 2;
        const [emptyXVector, emptyYVector] = await Promise.all([
          tensorflow.zeros([context_length]).array(),
          tensorflow.zeros([numberOfFeatures]).array(),
        ]);
        const x = [];
        const y = [];
        inputMatrix.forEach((inputVector,inputVectorIndex) => {
          inputVector.forEach((word, index) => {
            if (word != 0) {
              const output = new Array().concat(emptyYVector);
              const inputMerger = new Array().concat(inputMatrix[inputVectorIndex]);
              inputMerger.splice(index, 1);
              const input = FeatureEmbedding.getMergedArray(emptyXVector, inputMerger,true);
              output[word] = 1;
              // x.push([[word],input]);
              x.push(input);
              y.push(output);
            }
          });
        });

        return {
          context_length,
          emptyXVector,
          emptyYVector,
          x,
          y,
        };
      }
      
      
      
      constructor(options = {}, properties) {
        const config = {
          layers: [],
          type: 'cbow',
          compile: {
            loss: 'categoricalCrossentropy',
            optimizer: 'rmsprop',
          },
          fit: {
            epochs: 15,
            batchSize: 1,
          },
          embedSize: 50,
          windowSize: 2,
          PAD: 'PAD',
          streamInputMatrix: true,
          initialLayerInitializerType: 'randomNormal',
          initialLayerInitializerOptions: { seed: 1 },
          ...options
        };
        super(config, properties);
        this.type = 'FeatureEmbedding';
        this.featureToId;
        this.idToFeature;
        this.featureIds;
        this.numberOfFeatures;
        this.getMergedArray = FeatureEmbedding.getMergedArray;
        this.getFeatureDataSet = FeatureEmbedding.getFeatureDataSet.bind(this);
        this.getContextPairs = FeatureEmbedding.getContextPairs.bind(this);
        return this;
      }
      /**
       * Adds dense layers to tensorflow classification model
       * @override 
       * @param {Array<Array<number>>} x_matrix - independent variables
       * @param {Array<Array<number>>} y_matrix - dependent variables
       * @param {Array<Object>} layers - model dense layer parameters
       */
      generateLayers( x_matrix, y_matrix, layers) {
        // const xShape = this.getInputShape(x_matrix);
        if (!this.numberOfFeatures) throw ReferenceError(`${this.settings.name} model is missing numberOfFeatures`);
        if (!this.settings.embedSize) throw ReferenceError(`${this.settings.name} model is missing embedSize`);
        const yShape = [this.numberOfFeatures, this.settings.embedSize,];// this.getInputShape(y_matrix);
        this.yShape = yShape;
        // this.xShape = xShape;
        const denseLayers = [];
        denseLayers.push({
          units: this.numberOfFeatures,
          inputDim: this.numberOfFeatures,
          outputDim: this.settings.embedSize,
          inputLength: (this.settings.windowSize || 2) * 2,
          embeddingsInitializer: this.settings.initialLayerInitializerType
            ? this.tf.initializers[this.settings.initialLayerInitializerType](this.settings.
              initialLayerInitializerOptions)
            : undefined,
        });
        //TODO:NOT USED:
        denseLayers.push({
          lambdaFunction: 'result = tf.mean(input,1,true)',
          lambdaOutputShape: [this.numberOfFeatures, this.settings.embedSize]
        });
        //TODO:END NOT USED:
        denseLayers.push({ units: this.numberOfFeatures, activation: 'softmax', });

        this.layers = denseLayers;
        this.model.add(this.tf.layers.embedding(denseLayers[0]));
          // this.model.add(new lambdaLayer(denseLayers[1]));
        this.model.add(this.tf.layers.flatten());
        this.model.add(this.tf.layers.dense(denseLayers[2]));

        if (layers && layers.length && layers[0].weights) {
          const originalModelWeights = this.model.getWeights();
          originalModelWeights[0] = layers[0].weights;
          this.model.setWeights(originalModelWeights);
          // const postOriginalModelWeights = this.model.getWeights();
          // const layerData = postOriginalModelWeights.map(w => w.dataSync());
          // console.log('layerData', layerData);
        }
        // console.log('this.model.layers',this.model.layers)
      }
      async trainOnBatch({ x_input_matrix, y_output_matrix, epoch, trainingLoss, inputVectorIndex, inputVectorLength,}) {
        let loss = Infinity;
        if (_optionalChain([this, 'access', _ => _.settings, 'access', _2 => _2.fit, 'optionalAccess', _3 => _3.callbacks, 'optionalAccess', _4 => _4.onEpochBegin])) _optionalChain([this, 'access', _5 => _5.settings, 'access', _6 => _6.fit, 'optionalAccess', _7 => _7.callbacks, 'optionalAccess', _8 => _8.onEpochBegin, 'call', _9 => _9(epoch, { loss:trainingLoss })]);
        await asyncForEach(x_input_matrix, async (x_input, xIndex) => {
          if (_optionalChain([this, 'access', _10 => _10.settings, 'access', _11 => _11.fit, 'optionalAccess', _12 => _12.callbacks, 'optionalAccess', _13 => _13.onBatchBegin])) _optionalChain([this, 'access', _14 => _14.settings, 'access', _15 => _15.fit, 'optionalAccess', _16 => _16.callbacks, 'optionalAccess', _17 => _17.onBatchBegin, 'call', _18 => _18(xIndex, { loss:trainingLoss, })]);
          const y_output = y_output_matrix[xIndex];
          const xShape = this.getInputShape([x_input]);
          const xs = this.tf.tensor(x_input, xShape);
          const yShape = this.getInputShape([y_output]);
          const ys = this.tf.tensor(y_output, yShape);
          // const xdata = await xs.data()
          // console.log({ xs, xdata, xShape });
          loss = await this.model.trainOnBatch(xs, ys);
          if (_optionalChain([this, 'access', _19 => _19.settings, 'access', _20 => _20.fit, 'optionalAccess', _21 => _21.callbacks, 'optionalAccess', _22 => _22.onYield])) _optionalChain([this, 'access', _23 => _23.settings, 'access', _24 => _24.fit, 'optionalAccess', _25 => _25.callbacks, 'optionalAccess', _26 => _26.onYield, 'call', _27 => _27(epoch, xIndex, { loss, 
            inputVectorIndex,
            inputVectorLength,
            completion: `${( 100 * ( ((inputVectorIndex||xIndex)+1) / (inputVectorLength||x_input_matrix.length) ) ).toFixed(2)}%`  })]);
          if (_optionalChain([this, 'access', _28 => _28.settings, 'access', _29 => _29.fit, 'optionalAccess', _30 => _30.callbacks, 'optionalAccess', _31 => _31.onBatchEnd])) _optionalChain([this, 'access', _32 => _32.settings, 'access', _33 => _33.fit, 'optionalAccess', _34 => _34.callbacks, 'optionalAccess', _35 => _35.onBatchEnd, 'call', _36 => _36(xIndex, { loss })]);
          // console.log({ x_input, xIndex, xShape, y_output, yShape, loss })
          xs.dispose();
          ys.dispose();
        });
        if (_optionalChain([this, 'access', _37 => _37.settings, 'access', _38 => _38.fit, 'optionalAccess', _39 => _39.callbacks, 'optionalAccess', _40 => _40.onEpochEnd])) _optionalChain([this, 'access', _41 => _41.settings, 'access', _42 => _42.fit, 'optionalAccess', _43 => _43.callbacks, 'optionalAccess', _44 => _44.onEpochEnd, 'call', _45 => _45(epoch, { loss })]);
        return {
          loss
        };
      }
      async generateBatch({ epoch }) {
        if (!this.featureIds) throw ReferenceError(`${this.settings.name} model is missing featureIds`);

        const preTransformedMatrix = this.featureIds;
        const context_length = (this && this.settings && this.settings.windowSize ? this.settings.windowSize : 2) * 2;
        const [emptyXVector, emptyYVector] = await Promise.all([
          this.tf.zeros([context_length]).array(),
          this.tf.zeros([this.numberOfFeatures]).array(),
        ]);
        let x_input_matrix = [];
        let y_output_matrix = [];
        let trainingLoss = Infinity;
        await asyncForEach(preTransformedMatrix, async (inputVector, inputVectorIndex) => {
          await asyncForEach(inputVector, async (word, index) => {
            if (this.settings.checkInputMatrix && this.numberOfFeatures&& word >= (this.numberOfFeatures)) {
              console.warn('invalid word in corpus', {
                trainingLoss,
                word,
                'this.idToFeature[word]': this.idToFeature&& this.idToFeature[word],
                'this.numberOfFeatures': this.numberOfFeatures,
              });
          } else if (word != 0) {
              const output = new Array().concat(emptyYVector);
              const inputMerger = new Array().concat(preTransformedMatrix[inputVectorIndex]);
              inputMerger.splice(index, 1);
              const input = FeatureEmbedding.getMergedArray(emptyXVector, inputMerger,true);
              output[word] = 1;
              // x.push([[word],input]);
              // x.push(input);
              // y.push(output);
              x_input_matrix = [input];
              y_output_matrix = [output];
              // console.log({ input, output });
              if (this.settings.checkInputMatrix && this.numberOfFeatures&& this.numberOfFeatures > 0 && input.filter((wordInput) => this.numberOfFeatures&&wordInput >= this.numberOfFeatures).length) {
                console.warn('Input matrix contains unknown weight', { input, 'this.numberOfFeatures': this.numberOfFeatures });
              } else {
                  x_input_matrix = [input];
                  y_output_matrix = [output];
                  // console.log({ input, output });
                  const modelStatus = await this.trainOnBatch({ x_input_matrix, y_output_matrix, epoch, trainingLoss , inputVectorIndex, inputVectorLength: preTransformedMatrix.length, });
                  trainingLoss = modelStatus.loss;
              }
            }
          });
        });
        return {
          loss: trainingLoss,
        };
      }
      async exportEmbeddings() {
        if (this.trained !== true) throw new ReferenceError('The model has to be trained before embeddings can be exported');
        
        const weights = await this.predict();
        const labeledWeights = this.labelWeights(weights);
        return {
          featureToId: this.featureToId,
          idToFeature: this.idToFeature,
          featureIds: this.featureIds,
          numberOfFeatures: this.numberOfFeatures,
          labeledWeights,
        };
      }
      async importEmbeddings({ featureToId, idToFeature, featureIds, numberOfFeatures, labeledWeights, addNewWeights = true, inputMatrixFeatures, fixImportedWeights = false, }








    ) {
        // console.log(this.settings.name,'before - labeledWeights', labeledWeights);
        this.model = undefined;
        let updatedModelProperties;
        // let newWeights:LabeledWeights = {};
        if (addNewWeights) {
          if (inputMatrixFeatures) {
            updatedModelProperties = await this.getFeatureDataSet({ inputMatrixFeatures, initialIdToFeature: idToFeature, initialFeatureToId: featureToId, });
            featureToId = _optionalChain([updatedModelProperties, 'optionalAccess', _46 => _46.featureToId]);
            idToFeature = _optionalChain([updatedModelProperties, 'optionalAccess', _47 => _47.idToFeature]);
            featureIds = _optionalChain([updatedModelProperties, 'optionalAccess', _48 => _48.featureIds]);
            numberOfFeatures = _optionalChain([updatedModelProperties, 'optionalAccess', _49 => _49.numberOfFeatures]);
            // console.log(this.settings.name,'updatedModelProperties',updatedModelProperties);
          }
          if (featureToId) {
            await asyncForEach(Object.keys(featureToId), async (weightLabel) => {
              if (!labeledWeights[weightLabel] || !labeledWeights[weightLabel].length) {
                // newWeights[weightLabel] = await this.tf.randomUniform([1, this.settings.embedSize], -1, 1).array();
                labeledWeights[weightLabel] = await this.tf.randomUniform([1, this.settings.embedSize], -1, 1).array();
                if (Array.isArray(labeledWeights[weightLabel][0])) {
                  if (typeof labeledWeights[weightLabel].flat ==='function') labeledWeights[weightLabel] = labeledWeights[weightLabel].flat();
                  else labeledWeights[weightLabel] = labeledWeights[weightLabel].reduce((acc, val) => acc.concat(val), []);
                }
              }
            });
          }
        }
        if (fixImportedWeights) {
          Object.keys(labeledWeights).forEach(weightLabel => {
            if (Array.isArray(labeledWeights[weightLabel][0])) {
              if (typeof labeledWeights[weightLabel].flat ==='function') labeledWeights[weightLabel] = labeledWeights[weightLabel].flat();
              else labeledWeights[weightLabel] = labeledWeights[weightLabel].reduce((acc, val) => acc.concat(val), []);
            }
          });
        }
        // console.log(this.settings.name,'newWeights', newWeights);
        // console.log(this.settings.name,'after - labeledWeights', labeledWeights);

        const firstLabeledWeight = Object.keys(labeledWeights)[0];
        if (!firstLabeledWeight || !labeledWeights[firstLabeledWeight] || labeledWeights[firstLabeledWeight].length !== this.settings.embedSize) throw new RangeError(`imported weights (${labeledWeights[firstLabeledWeight]?labeledWeights[firstLabeledWeight].length:'firstLabeledWeight:undefined'}) must have the same embedding size as model (${this.settings.embedSize})`);
        const trainedWeights = this.tf.variable(this.tf.tensor(Object.values(labeledWeights)));

        
        this.featureToId = featureToId;
        this.idToFeature = idToFeature;
        this.featureIds = featureIds;
        this.numberOfFeatures = numberOfFeatures;
        if (trainedWeights.shape[0] !== this.numberOfFeatures) {
          console.warn('INVALID NUMBER OF this.numberOfFeatures', {
            'trainedWeights.shape[0]': trainedWeights.shape[0],
            'this.numberOfFeatures': this.numberOfFeatures,
          });
          this.numberOfFeatures = trainedWeights.shape[0];
        }
        this.compileModel({ layers: [{ weights: trainedWeights }] });
        this.importedEmbeddings = true;
      }
      compileModel({ layers, } = {}) {
        this.model = undefined;
        this.model = this.tf.sequential();
        //@ts-ignore
        this.generateLayers.call(this, [], [], layers || this.layers, /* x_test, y_test */);
        this.model.compile(this.settings.compile);
        this.compiled = true;
      }
      async train(x_matrix, y_matrix, layers) {
        if (!this.featureToId || !this.idToFeature || !this.featureIds || !this.numberOfFeatures) {
          const featureEmbedDataSet = await this.getFeatureDataSet({ inputMatrixFeatures: x_matrix, });
          this.featureToId = featureEmbedDataSet.featureToId;
          this.idToFeature = featureEmbedDataSet.idToFeature;
          this.featureIds = featureEmbedDataSet.featureIds;
          this.numberOfFeatures = featureEmbedDataSet.numberOfFeatures;
        } else if (this.importedEmbeddings && x_matrix && x_matrix.length) this.featureIds = x_matrix;
        if (this.compiled === false) this.compileModel({layers});
        let loss = Infinity;
        if (_optionalChain([this, 'access', _50 => _50.settings, 'access', _51 => _51.fit, 'optionalAccess', _52 => _52.callbacks, 'optionalAccess', _53 => _53.onTrainBegin])) _optionalChain([this, 'access', _54 => _54.settings, 'access', _55 => _55.fit, 'optionalAccess', _56 => _56.callbacks, 'optionalAccess', _57 => _57.onTrainBegin, 'call', _58 => _58({ loss })]);
        await asyncForEach(lodash_range(0, _optionalChain([this, 'access', _59 => _59.settings, 'access', _60 => _60.fit, 'optionalAccess', _61 => _61.epochs])), async (epoch) => {
          if (this.settings.streamInputMatrix) {
            let modelStatus = await this.generateBatch({epoch,});
            loss = modelStatus.loss;
          } else {
            if (!this.numberOfFeatures) throw ReferenceError(`${this.settings.name} model is missing numberOfFeatures`);
            if (!this.featureIds) throw ReferenceError(`${this.settings.name} model is missing featureIds`);

            const cxt = await this.getContextPairs({ tf:this.tf, numberOfFeatures: this.numberOfFeatures, inputMatrix: this.featureIds });
            const x_input_matrix = cxt.x;
            const y_output_matrix = cxt.y;
            let modelStatus = await this.trainOnBatch({ x_input_matrix, y_output_matrix, epoch, trainingLoss:loss, });
            loss = modelStatus.loss;
          }
        });
        if (_optionalChain([this, 'access', _62 => _62.settings, 'access', _63 => _63.fit, 'optionalAccess', _64 => _64.callbacks, 'optionalAccess', _65 => _65.onTrainEnd])) _optionalChain([this, 'access', _66 => _66.settings, 'access', _67 => _67.fit, 'optionalAccess', _68 => _68.callbacks, 'optionalAccess', _69 => _69.onTrainEnd, 'call', _70 => _70({ loss })]);
        this.loss = loss;

        this.trained = true;
        return this.model;
      }
      // async calculate(x_matrix: Matrix | Vector | InputTextArray) {
      async calculate() {
        return this.model.getWeights()[0];
      }
      // async predict(input_matrix: any[], options: PredictionOptions | undefined) {
      async predict(options = {}) {
        const predictions = await this.calculate();
        if (options.json === false) {
          return await predictions.data();
        } else {
          // console.log({predictions})
          const arr = await predictions.array();
          if (!this.yShape) throw new Error('Model is missing yShape');
          return this.reshape(arr, predictions.shape);
        }
      }
      /**
       * Converts matrix of layer weights into labeled features
       * @example
    const weights = [
      [1.5,1,4,1.6,3.5],
      [4.3,3.2,5.5,6.5]
    ]
    FeatureEmbeddingInstance.labelWeights(weights) //=> 
    weights = {
      car:[1.5,1,4,1.6,3.5],
      boat:[4.3,3.2,5.5,6.5]
    }
       */
      labelWeights(weights) {
        return weights.reduce((result, weight, index) => { 
         if(this.idToFeature) result[this.idToFeature[index]] = weight;
          return result;
        }, {});
      }
      /**
       * Uses tSNE to reduce dimensionality of features
       * @example
    const weights = [
      [1.5,1,4,1.6,3.5],
      [4.3,3.2,5.5,6.5]
    ]
    FeatureEmbeddingInstance.reduceWeights(weights) //=> 
    [
      [1,2],
      [2,3],
    ]
       */
      async reduceWeights(weights, options) {
        let model = new TSNE({
          dim: 2,
          perplexity: 30.0,
          earlyExaggeration: 4.0,
          learningRate: 100.0,
          nIter: 1000,
          metric: 'euclidean',
          ...options
        });
        model.init({
          data: weights,
          type: options ? options.type : 'dense'
        });
        model.run();
        // console.log({ error, iter });
        let output = model.getOutput(); 
        return output;
      }
    /**
     * Uses either cosineProximity or Eucledian distance to rank similarity
    @example
    //weights = [ [1,2,3,], [1,2,2], [0,-1,3] ]
    //labeledWeights = [ {car:[1,2,3,],tesla:[1,2,2],boat:[0,-1,3]}]
    FeatureEmbeddingInstance.findSimilarFeatures(weights,{features:['car'], limit:2,}) //=> 
    {
      car:[
      {
        comparedFeature: 'tesla',
        proximity: -0.5087087154388428,
        distance: 0.03015853278338909
      },
      {
        comparedFeature: 'boat',
        proximity: -0.3032159209251404,
        distance: 0.036241017282009125
      },
      ]
    }
     */
      async findSimilarFeatures(weights, options = {}) {
        const tf = this.tf;
        const { features = [], limit=5, labeledWeights, metric='distance' } = options;
        const labeledFeatureWeights = labeledWeights || this.labelWeights(weights);
        if(this.settings && this.settings.PAD) delete labeledFeatureWeights[this.settings.PAD];
        return features.reduce((result,feature) => {
          const featureWeights = labeledFeatureWeights[feature];
          if(!featureWeights) throw new ReferenceError(`Invalid feature: ${feature}`);
         
          const sims = Object.keys(labeledFeatureWeights)
            .map(searchFeature => {
              const prox = tf.tidy(() => {
                const proximity = tf.metrics.cosineProximity(
                  tf.tensor(featureWeights),
                  tf.tensor(labeledFeatureWeights[searchFeature])
                );
                const distance = tf.metrics.meanSquaredError(
                  tf.tensor(featureWeights),
                  tf.tensor(labeledFeatureWeights[searchFeature])
                );
                return [
                  proximity.asScalar().dataSync()[0],
                  distance.asScalar().dataSync()[0],
                ];
              });
              return {
                comparedFeature: searchFeature,
                proximity: prox[0],
                distance: prox[1],
              }
            })
            .sort((a, b) => (metric === 'distance')
              ? a.distance - b.distance
              : a.proximity - b.proximity);
          // sims.shift()
          
          result[feature]= sims.slice(1, limit+1);
          return result;
        }, {});
      }
    }




    var SimilarityMetric; (function (SimilarityMetric) {
      const DISTANCE = 'distance'; SimilarityMetric["DISTANCE"] = DISTANCE;
      const PROXIMITY = 'proximity'; SimilarityMetric["PROXIMITY"] = PROXIMITY;
    })(SimilarityMetric || (SimilarityMetric = {}));

    exports.BaseNeuralNetwork = BaseNeuralNetwork;
    exports.DeepLearningClassification = DeepLearningClassification;
    exports.DeepLearningRegression = DeepLearningRegression;
    exports.FeatureEmbedding = FeatureEmbedding;
    exports.LSTMMultivariateTimeSeries = LSTMMultivariateTimeSeries;
    exports.LSTMTimeSeries = LSTMTimeSeries;
    exports.LogisticRegression = LogisticRegression;
    exports.MultipleLinearRegression = MultipleLinearRegression;
    exports.TensorScriptModelInterface = TensorScriptModelInterface;
    exports.TextEmbedding = TextEmbedding;

    Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=index.js.map
